#ifndef LIBRETRO_CORE_OPTIONS_INTL_H__
#define LIBRETRO_CORE_OPTIONS_INTL_H__

#if defined(_MSC_VER) && (_MSC_VER >= 1500 && _MSC_VER < 1900)
/* https://support.microsoft.com/en-us/kb/980263 */
#pragma execution_character_set("utf-8")
#pragma warning(disable:4566)
#endif

#include <libretro.h>

#include "libretro_options.h"

/*
 ********************************
 * VERSION: 1.3
 ********************************
 *
 * - 1.3: Move translations to libretro_core_options_intl.h
 *        - libretro_core_options_intl.h includes BOM and utf-8
 *          fix for MSVC 2010-2013
 *        - Added HAVE_NO_LANGEXTRA flag to disable translations
 *          on platforms/compilers without BOM support
 * - 1.2: Use core options v1 interface when
 *        RETRO_ENVIRONMENT_GET_CORE_OPTIONS_VERSION is >= 1
 *        (previously required RETRO_ENVIRONMENT_GET_CORE_OPTIONS_VERSION == 1)
 * - 1.1: Support generation of core options v0 retro_core_option_value
 *        arrays containing options with a single value
 * - 1.0: First commit
*/

#ifdef __cplusplus
extern "C" {
#endif

/*
 ********************************
 * Core Option Definitions
 ********************************
*/
/* RETRO_LANGUAGE_AR */

#define CATEGORY_SYSTEM_LABEL_AR "نظام"
#define CATEGORY_SYSTEM_INFO_0_AR NULL
#define CATEGORY_VIDEO_LABEL_AR "فيديو"
#define CATEGORY_VIDEO_INFO_0_AR NULL
#define CATEGORY_PGXP_LABEL_AR NULL
#define CATEGORY_PGXP_INFO_0_AR NULL
#define CATEGORY_OSD_LABEL_AR "العرض على الشاشة"
#define CATEGORY_OSD_INFO_0_AR NULL
#define CATEGORY_INPUT_LABEL_AR "الإدخال"
#define CATEGORY_INPUT_INFO_0_AR NULL
#define CATEGORY_MEMCARDS_LABEL_AR "بطاقة الذاكرة"
#define CATEGORY_MEMCARDS_INFO_0_AR NULL
#define CATEGORY_HACKS_LABEL_AR NULL
#define CATEGORY_HACKS_INFO_0_AR NULL
#define BEETLE_OPT_CPU_FREQ_SCALE_LABEL_AR NULL
#define BEETLE_OPT_CPU_FREQ_SCALE_INFO_0_AR NULL
#define OPTION_VAL_50_AR NULL
#define OPTION_VAL_60_AR NULL
#define OPTION_VAL_70_AR NULL
#define OPTION_VAL_80_AR NULL
#define OPTION_VAL_90_AR NULL
#define OPTION_VAL_100_AR NULL
#define OPTION_VAL_110_AR NULL
#define OPTION_VAL_120_AR NULL
#define OPTION_VAL_130_AR NULL
#define OPTION_VAL_140_AR NULL
#define OPTION_VAL_150_AR NULL
#define OPTION_VAL_160_AR NULL
#define OPTION_VAL_170_AR NULL
#define OPTION_VAL_180_AR NULL
#define OPTION_VAL_190_AR NULL
#define OPTION_VAL_200_AR NULL
#define OPTION_VAL_210_AR NULL
#define OPTION_VAL_220_AR NULL
#define OPTION_VAL_230_AR NULL
#define OPTION_VAL_240_AR NULL
#define OPTION_VAL_250_AR NULL
#define OPTION_VAL_260_AR NULL
#define OPTION_VAL_270_AR NULL
#define OPTION_VAL_280_AR NULL
#define OPTION_VAL_290_AR NULL
#define OPTION_VAL_300_AR NULL
#define OPTION_VAL_310_AR NULL
#define OPTION_VAL_320_AR NULL
#define OPTION_VAL_330_AR NULL
#define OPTION_VAL_340_AR NULL
#define OPTION_VAL_350_AR NULL
#define OPTION_VAL_360_AR NULL
#define OPTION_VAL_370_AR NULL
#define OPTION_VAL_380_AR NULL
#define OPTION_VAL_390_AR NULL
#define OPTION_VAL_400_AR NULL
#define OPTION_VAL_410_AR NULL
#define OPTION_VAL_420_AR NULL
#define OPTION_VAL_430_AR NULL
#define OPTION_VAL_440_AR NULL
#define OPTION_VAL_450_AR NULL
#define OPTION_VAL_460_AR NULL
#define OPTION_VAL_470_AR NULL
#define OPTION_VAL_480_AR NULL
#define OPTION_VAL_490_AR NULL
#define OPTION_VAL_500_AR NULL
#define OPTION_VAL_510_AR NULL
#define OPTION_VAL_520_AR NULL
#define OPTION_VAL_530_AR NULL
#define OPTION_VAL_540_AR NULL
#define OPTION_VAL_550_AR NULL
#define OPTION_VAL_560_AR NULL
#define OPTION_VAL_570_AR NULL
#define OPTION_VAL_580_AR NULL
#define OPTION_VAL_590_AR NULL
#define OPTION_VAL_600_AR NULL
#define OPTION_VAL_610_AR NULL
#define OPTION_VAL_620_AR NULL
#define OPTION_VAL_630_AR NULL
#define OPTION_VAL_640_AR NULL
#define OPTION_VAL_650_AR NULL
#define OPTION_VAL_660_AR NULL
#define OPTION_VAL_670_AR NULL
#define OPTION_VAL_680_AR NULL
#define OPTION_VAL_690_AR NULL
#define OPTION_VAL_700_AR NULL
#define OPTION_VAL_710_AR NULL
#define OPTION_VAL_720_AR NULL
#define OPTION_VAL_730_AR NULL
#define OPTION_VAL_740_AR NULL
#define OPTION_VAL_750_AR NULL
#define BEETLE_OPT_GPU_OVERCLOCK_LABEL_AR NULL
#define BEETLE_OPT_GPU_OVERCLOCK_INFO_0_AR NULL
#define OPTION_VAL_1X_NATIVE_AR "1x (أصلي)"
#define OPTION_VAL_2X_AR NULL
#define OPTION_VAL_4X_AR NULL
#define OPTION_VAL_8X_AR NULL
#define OPTION_VAL_16X_AR NULL
#define OPTION_VAL_32X_AR NULL
#define BEETLE_OPT_GTE_OVERCLOCK_LABEL_AR NULL
#define BEETLE_OPT_GTE_OVERCLOCK_INFO_0_AR NULL
#define BEETLE_OPT_CD_ACCESS_METHOD_LABEL_AR NULL
#define BEETLE_OPT_CD_ACCESS_METHOD_INFO_0_AR NULL
#define OPTION_VAL_SYNC_AR NULL
#define OPTION_VAL_ASYNC_AR NULL
#define OPTION_VAL_PRECACHE_AR NULL
#define BEETLE_OPT_CD_FASTLOAD_LABEL_AR NULL
#define BEETLE_OPT_CD_FASTLOAD_INFO_0_AR NULL
#define OPTION_VAL_2X_NATIVE_AR NULL
#define OPTION_VAL_6X_AR NULL
#define OPTION_VAL_10X_AR NULL
#define OPTION_VAL_12X_AR NULL
#define OPTION_VAL_14X_AR NULL
#define BEETLE_OPT_SKIP_BIOS_LABEL_AR NULL
#define BEETLE_OPT_SKIP_BIOS_INFO_0_AR NULL
#define BEETLE_OPT_OVERRIDE_BIOS_LABEL_AR NULL
#define BEETLE_OPT_OVERRIDE_BIOS_INFO_0_AR NULL
#define OPTION_VAL_PSXONPSP_AR NULL
#define OPTION_VAL_PS1_ROM_AR NULL
#define OPTION_VAL_OPENBIOS_AR NULL
#define BEETLE_OPT_REGION_LABEL_AR NULL
#define BEETLE_OPT_REGION_INFO_0_AR NULL
#define OPTION_VAL_AUTO_AR "تلقائي"
#define OPTION_VAL_NTSC_J_AR NULL
#define OPTION_VAL_NTSC_U_AR NULL
#define OPTION_VAL_PAL_AR NULL
#define BEETLE_OPT_INTERNAL_RESOLUTION_LABEL_AR NULL
#define BEETLE_OPT_INTERNAL_RESOLUTION_INFO_0_AR NULL
#define BEETLE_OPT_RENDERER_LABEL_AR NULL
#define BEETLE_OPT_RENDERER_INFO_0_AR NULL
#define OPTION_VAL_HARDWARE_AR NULL
#define OPTION_VAL_HARDWARE_GL_AR NULL
#define OPTION_VAL_HARDWARE_VK_AR NULL
#define OPTION_VAL_SOFTWARE_AR NULL
#define BEETLE_OPT_RENDERER_SOFTWARE_FB_LABEL_AR NULL
#define BEETLE_OPT_RENDERER_SOFTWARE_FB_INFO_0_AR NULL
#define BEETLE_OPT_DEPTH_LABEL_AR NULL
#define BEETLE_OPT_DEPTH_INFO_0_AR NULL
#define OPTION_VAL_16BPP_NATIVE_AR NULL
#define OPTION_VAL_32BPP_AR NULL
#define BEETLE_OPT_DITHER_MODE_LABEL_AR NULL
#define BEETLE_OPT_DITHER_MODE_INFO_0_AR NULL
#define OPTION_VAL_INTERNAL_RESOLUTION_AR NULL
#define BEETLE_OPT_SCALED_UV_OFFSET_LABEL_AR NULL
#define BEETLE_OPT_SCALED_UV_OFFSET_INFO_0_AR NULL
#define BEETLE_OPT_FILTER_LABEL_AR NULL
#define BEETLE_OPT_FILTER_INFO_0_AR NULL
#define OPTION_VAL_NEAREST_AR "أقرب"
#define OPTION_VAL_SABR_AR NULL
#define OPTION_VAL_XBR_AR NULL
#define OPTION_VAL_BILINEAR_AR "ثنائي"
#define OPTION_VAL_3_POINT_AR NULL
#define OPTION_VAL_JINC2_AR NULL
#define BEETLE_OPT_FILTER_EXCLUDE_SPRITE_LABEL_AR NULL
#define BEETLE_OPT_FILTER_EXCLUDE_SPRITE_INFO_0_AR NULL
#define OPTION_VAL_OPAQUE_AR NULL
#define OPTION_VAL_ALL_AR NULL
#define BEETLE_OPT_FILTER_EXCLUDE_2D_POLYGON_LABEL_AR NULL
#define BEETLE_OPT_FILTER_EXCLUDE_2D_POLYGON_INFO_0_AR NULL
#define BEETLE_OPT_ADAPTIVE_SMOOTHING_LABEL_AR NULL
#define BEETLE_OPT_ADAPTIVE_SMOOTHING_INFO_0_AR NULL
#define BEETLE_OPT_SUPER_SAMPLING_LABEL_AR NULL
#define BEETLE_OPT_SUPER_SAMPLING_INFO_0_AR NULL
#define BEETLE_OPT_MSAA_LABEL_AR NULL
#define BEETLE_OPT_MSAA_INFO_0_AR NULL
#define OPTION_VAL_1X_AR NULL
#define BEETLE_OPT_MDEC_YUV_LABEL_AR NULL
#define BEETLE_OPT_MDEC_YUV_INFO_0_AR NULL
#define BEETLE_OPT_TRACK_TEXTURES_LABEL_AR NULL
#define BEETLE_OPT_TRACK_TEXTURES_INFO_0_AR NULL
#define BEETLE_OPT_DUMP_TEXTURES_LABEL_AR NULL
#define BEETLE_OPT_DUMP_TEXTURES_INFO_0_AR NULL
#define BEETLE_OPT_REPLACE_TEXTURES_LABEL_AR NULL
#define BEETLE_OPT_REPLACE_TEXTURES_INFO_0_AR NULL
#define BEETLE_OPT_PGXP_MODE_LABEL_AR NULL
#define BEETLE_OPT_PGXP_MODE_INFO_0_AR NULL
#define OPTION_VAL_MEMORY_ONLY_AR NULL
#define OPTION_VAL_MEMORY_CPU_AR NULL
#define BEETLE_OPT_PGXP_2D_TOL_LABEL_AR NULL
#define BEETLE_OPT_PGXP_2D_TOL_INFO_0_AR NULL
#define OPTION_VAL_0PX_AR NULL
#define OPTION_VAL_1PX_AR NULL
#define OPTION_VAL_2PX_AR NULL
#define OPTION_VAL_3PX_AR NULL
#define OPTION_VAL_4PX_AR NULL
#define OPTION_VAL_5PX_AR NULL
#define OPTION_VAL_6PX_AR NULL
#define OPTION_VAL_7PX_AR NULL
#define OPTION_VAL_8PX_AR NULL
#define BEETLE_OPT_PGXP_NCLIP_LABEL_AR NULL
#define BEETLE_OPT_PGXP_NCLIP_INFO_0_AR NULL
#define BEETLE_OPT_PGXP_VERTEX_LABEL_AR NULL
#define BEETLE_OPT_PGXP_VERTEX_INFO_0_AR NULL
#define BEETLE_OPT_PGXP_TEXTURE_LABEL_AR NULL
#define BEETLE_OPT_PGXP_TEXTURE_INFO_0_AR NULL
#define BEETLE_OPT_DISPLAY_INTERNAL_FPS_LABEL_AR NULL
#define BEETLE_OPT_DISPLAY_INTERNAL_FPS_INFO_0_AR NULL
#define BEETLE_OPT_DISPLAY_OSD_LABEL_AR NULL
#define BEETLE_OPT_DISPLAY_OSD_INFO_0_AR NULL
#define BEETLE_OPT_DISPLAY_VRAM_LABEL_AR NULL
#define BEETLE_OPT_DISPLAY_VRAM_INFO_0_AR NULL
#define BEETLE_OPT_ANALOG_CALIBRATION_LABEL_AR NULL
#define BEETLE_OPT_ANALOG_CALIBRATION_INFO_0_AR NULL
#define BEETLE_OPT_ANALOG_TOGGLE_LABEL_AR NULL
#define BEETLE_OPT_ANALOG_TOGGLE_INFO_0_AR NULL
#define OPTION_VAL_ENABLED_ANALOG_AR NULL
#define BEETLE_OPT_ANALOG_TOGGLE_COMBO_LABEL_AR NULL
#define BEETLE_OPT_ANALOG_TOGGLE_COMBO_INFO_0_AR NULL
#define OPTION_VAL_L1_L2_R1_R2_START_SELECT_AR NULL
#define OPTION_VAL_L1_R1_SELECT_AR NULL
#define OPTION_VAL_L1_R1_START_AR NULL
#define OPTION_VAL_L1_R1_L3_AR NULL
#define OPTION_VAL_L1_R1_R3_AR NULL
#define OPTION_VAL_L2_R2_SELECT_AR NULL
#define OPTION_VAL_L2_R2_START_AR NULL
#define OPTION_VAL_L2_R2_L3_AR NULL
#define OPTION_VAL_L2_R2_R3_AR NULL
#define OPTION_VAL_L3_R3_AR NULL
#define BEETLE_OPT_ANALOG_TOGGLE_HOLD_LABEL_AR NULL
#define BEETLE_OPT_ANALOG_TOGGLE_HOLD_INFO_0_AR NULL
#define OPTION_VAL_0_AR NULL
#define OPTION_VAL_1_AR NULL
#define OPTION_VAL_2_AR NULL
#define OPTION_VAL_3_AR NULL
#define OPTION_VAL_4_AR NULL
#define OPTION_VAL_5_AR NULL
#define BEETLE_OPT_ENABLE_MULTITAP_PORT1_LABEL_AR NULL
#define BEETLE_OPT_ENABLE_MULTITAP_PORT1_INFO_0_AR NULL
#define BEETLE_OPT_ENABLE_MULTITAP_PORT2_LABEL_AR NULL
#define BEETLE_OPT_ENABLE_MULTITAP_PORT2_INFO_0_AR NULL
#define BEETLE_OPT_GUN_INPUT_MODE_LABEL_AR NULL
#define BEETLE_OPT_GUN_INPUT_MODE_INFO_0_AR NULL
#define OPTION_VAL_LIGHTGUN_AR NULL
#define OPTION_VAL_TOUCHSCREEN_AR NULL
#define BEETLE_OPT_GUN_CURSOR_LABEL_AR NULL
#define BEETLE_OPT_GUN_CURSOR_INFO_0_AR NULL
#define OPTION_VAL_CROSS_AR NULL
#define OPTION_VAL_DOT_AR NULL
#define BEETLE_OPT_CROSSHAIR_COLOR_P1_LABEL_AR NULL
#define BEETLE_OPT_CROSSHAIR_COLOR_P1_INFO_0_AR NULL
#define OPTION_VAL_RED_AR "أحمر"
#define OPTION_VAL_BLUE_AR "أزرق"
#define OPTION_VAL_GREEN_AR "أخضر"
#define OPTION_VAL_ORANGE_AR NULL
#define OPTION_VAL_YELLOW_AR "أصفر"
#define OPTION_VAL_CYAN_AR NULL
#define OPTION_VAL_PINK_AR NULL
#define OPTION_VAL_PURPLE_AR NULL
#define OPTION_VAL_BLACK_AR NULL
#define OPTION_VAL_WHITE_AR NULL
#define BEETLE_OPT_CROSSHAIR_COLOR_P2_LABEL_AR NULL
#define BEETLE_OPT_CROSSHAIR_COLOR_P2_INFO_0_AR NULL
#define BEETLE_OPT_MOUSE_SENSITIVITY_LABEL_AR NULL
#define BEETLE_OPT_MOUSE_SENSITIVITY_INFO_0_AR NULL
#define OPTION_VAL_5_O42_AR NULL
#define OPTION_VAL_10_AR NULL
#define OPTION_VAL_15_AR NULL
#define OPTION_VAL_20_AR NULL
#define OPTION_VAL_25_AR NULL
#define OPTION_VAL_30_AR NULL
#define OPTION_VAL_35_AR NULL
#define OPTION_VAL_40_AR NULL
#define OPTION_VAL_45_AR NULL
#define OPTION_VAL_55_AR NULL
#define OPTION_VAL_65_AR NULL
#define OPTION_VAL_75_AR NULL
#define OPTION_VAL_85_AR NULL
#define OPTION_VAL_95_AR NULL
#define OPTION_VAL_100_O42_AR NULL
#define OPTION_VAL_105_AR NULL
#define OPTION_VAL_115_AR NULL
#define OPTION_VAL_125_AR NULL
#define OPTION_VAL_135_AR NULL
#define OPTION_VAL_145_AR NULL
#define OPTION_VAL_155_AR NULL
#define OPTION_VAL_165_AR NULL
#define OPTION_VAL_175_AR NULL
#define OPTION_VAL_185_AR NULL
#define OPTION_VAL_195_AR NULL
#define BEETLE_OPT_NEGCON_RESPONSE_LABEL_AR NULL
#define BEETLE_OPT_NEGCON_RESPONSE_INFO_0_AR NULL
#define OPTION_VAL_LINEAR_AR "خطي"
#define OPTION_VAL_QUADRATIC_AR NULL
#define OPTION_VAL_CUBIC_AR NULL
#define BEETLE_OPT_NEGCON_DEADZONE_LABEL_AR NULL
#define BEETLE_OPT_NEGCON_DEADZONE_INFO_0_AR NULL
#define OPTION_VAL_0_O44_AR NULL
#define BEETLE_OPT_USE_MEDNAFEN_MEMCARD0_METHOD_LABEL_AR NULL
#define BEETLE_OPT_USE_MEDNAFEN_MEMCARD0_METHOD_INFO_0_AR NULL
#define OPTION_VAL_LIBRETRO_AR NULL
#define OPTION_VAL_MEDNAFEN_AR NULL
#define BEETLE_OPT_ENABLE_MEMCARD1_LABEL_AR NULL
#define BEETLE_OPT_ENABLE_MEMCARD1_INFO_0_AR NULL
#define BEETLE_OPT_SHARED_MEMORY_CARDS_LABEL_AR NULL
#define BEETLE_OPT_SHARED_MEMORY_CARDS_INFO_0_AR NULL
#define BEETLE_OPT_MEMCARD_LEFT_INDEX_LABEL_AR NULL
#define BEETLE_OPT_MEMCARD_LEFT_INDEX_INFO_0_AR NULL
#define OPTION_VAL_0_O48_AR NULL
#define BEETLE_OPT_MEMCARD_RIGHT_INDEX_LABEL_AR NULL
#define BEETLE_OPT_MEMCARD_RIGHT_INDEX_INFO_0_AR NULL
#define OPTION_VAL_1_O49_AR NULL
#define BEETLE_OPT_LINE_RENDER_LABEL_AR NULL
#define BEETLE_OPT_LINE_RENDER_INFO_0_AR NULL
#define OPTION_VAL_DEFAULT_AR "الافتراضي"
#define OPTION_VAL_AGGRESSIVE_AR NULL
#define BEETLE_OPT_WIDESCREEN_HACK_LABEL_AR NULL
#define BEETLE_OPT_WIDESCREEN_HACK_INFO_0_AR NULL
#define BEETLE_OPT_WIDESCREEN_HACK_ASPECT_RATIO_LABEL_AR NULL
#define BEETLE_OPT_WIDESCREEN_HACK_ASPECT_RATIO_INFO_0_AR NULL
#define OPTION_VAL_16_9_AR NULL
#define OPTION_VAL_16_10_AR NULL
#define OPTION_VAL_18_9_AR NULL
#define OPTION_VAL_19_9_AR NULL
#define OPTION_VAL_20_9_AR NULL
#define OPTION_VAL_21_9_AR NULL
#define OPTION_VAL_32_9_AR NULL
#define BEETLE_OPT_CPU_DYNAREC_LABEL_AR NULL
#define BEETLE_OPT_CPU_DYNAREC_INFO_0_AR NULL
#define OPTION_VAL_DISABLED_AR NULL
#define OPTION_VAL_EXECUTE_AR NULL
#define OPTION_VAL_EXECUTE_ONE_AR NULL
#define OPTION_VAL_RUN_INTERPRETER_AR NULL
#define BEETLE_OPT_DYNAREC_INVALIDATE_LABEL_AR NULL
#define BEETLE_OPT_DYNAREC_INVALIDATE_INFO_0_AR NULL
#define OPTION_VAL_FULL_AR NULL
#define OPTION_VAL_DMA_AR NULL
#define BEETLE_OPT_DYNAREC_EVENTCYCLES_LABEL_AR NULL
#define BEETLE_OPT_DYNAREC_EVENTCYCLES_INFO_0_AR NULL
#define OPTION_VAL_128_AR NULL
#define BEETLE_OPT_DYNAREC_SPU_SAMPLES_LABEL_AR NULL
#define BEETLE_OPT_DYNAREC_SPU_SAMPLES_INFO_0_AR NULL
#define BEETLE_OPT_CORE_TIMING_FPS_LABEL_AR NULL
#define BEETLE_OPT_CORE_TIMING_FPS_INFO_0_AR NULL
#define OPTION_VAL_AUTO_TOGGLE_AR NULL
#define OPTION_VAL_FORCE_PROGRESSIVE_AR NULL
#define OPTION_VAL_FORCE_INTERLACED_AR NULL
#define BEETLE_OPT_DEINTERLACER_LABEL_AR NULL
#define BEETLE_OPT_DEINTERLACER_INFO_0_AR NULL
#define OPTION_VAL_WEAVE_AR NULL
#define OPTION_VAL_BOB_AR NULL
#define OPTION_VAL_BOB_OFFSET_AR NULL
#define OPTION_VAL_FASTMAD_AR NULL
#define BEETLE_OPT_FRAME_DUPING_LABEL_AR NULL
#define BEETLE_OPT_FRAME_DUPING_INFO_0_AR NULL
#define BEETLE_OPT_PAL_VIDEO_TIMING_OVERRIDE_LABEL_AR NULL
#define BEETLE_OPT_PAL_VIDEO_TIMING_OVERRIDE_INFO_0_AR NULL
#define BEETLE_OPT_ASPECT_RATIO_LABEL_AR NULL
#define BEETLE_OPT_ASPECT_RATIO_INFO_0_AR NULL
#define OPTION_VAL_CORRECTED_AR NULL
#define OPTION_VAL_UNCORRECTED_AR NULL
#define OPTION_VAL_4_3_AR NULL
#define OPTION_VAL_NTSC_AR NULL
#define BEETLE_OPT_CROP_OVERSCAN_LABEL_AR NULL
#define BEETLE_OPT_CROP_OVERSCAN_INFO_0_AR NULL
#define OPTION_VAL_STATIC_AR NULL
#define OPTION_VAL_SMART_AR NULL
#define BEETLE_OPT_IMAGE_CROP_LABEL_AR NULL
#define BEETLE_OPT_IMAGE_CROP_INFO_0_AR NULL
#define OPTION_VAL_9PX_AR NULL
#define OPTION_VAL_10PX_AR NULL
#define OPTION_VAL_11PX_AR NULL
#define OPTION_VAL_12PX_AR NULL
#define OPTION_VAL_13PX_AR NULL
#define OPTION_VAL_14PX_AR NULL
#define OPTION_VAL_15PX_AR NULL
#define OPTION_VAL_16PX_AR NULL
#define OPTION_VAL_17PX_AR NULL
#define OPTION_VAL_18PX_AR NULL
#define OPTION_VAL_19PX_AR NULL
#define OPTION_VAL_20PX_AR NULL
#define BEETLE_OPT_IMAGE_OFFSET_LABEL_AR NULL
#define BEETLE_OPT_IMAGE_OFFSET_INFO_0_AR NULL
#define OPTION_VAL_12PX_O64_AR NULL
#define OPTION_VAL_11PX_O64_AR NULL
#define OPTION_VAL_10PX_O64_AR NULL
#define OPTION_VAL_9PX_O64_AR NULL
#define OPTION_VAL_8PX_O64_AR NULL
#define OPTION_VAL_7PX_O64_AR NULL
#define OPTION_VAL_6PX_O64_AR NULL
#define OPTION_VAL_5PX_O64_AR NULL
#define OPTION_VAL_4PX_O64_AR NULL
#define OPTION_VAL_3PX_O64_AR NULL
#define OPTION_VAL_2PX_O64_AR NULL
#define OPTION_VAL_1PX_O64_AR NULL
#define OPTION_VAL_1PX_O64_0_AR NULL
#define OPTION_VAL_2PX_O64_0_AR NULL
#define OPTION_VAL_3PX_O64_0_AR NULL
#define OPTION_VAL_4PX_O64_0_AR NULL
#define OPTION_VAL_5PX_O64_0_AR NULL
#define OPTION_VAL_6PX_O64_0_AR NULL
#define OPTION_VAL_7PX_O64_0_AR NULL
#define OPTION_VAL_8PX_O64_0_AR NULL
#define OPTION_VAL_9PX_O64_0_AR NULL
#define OPTION_VAL_10PX_O64_0_AR NULL
#define OPTION_VAL_11PX_O64_0_AR NULL
#define OPTION_VAL_12PX_O64_0_AR NULL
#define BEETLE_OPT_IMAGE_OFFSET_CYCLES_LABEL_AR NULL
#define BEETLE_OPT_IMAGE_OFFSET_CYCLES_INFO_0_AR NULL
#define BEETLE_OPT_INITIAL_SCANLINE_LABEL_AR NULL
#define BEETLE_OPT_INITIAL_SCANLINE_INFO_0_AR NULL
#define BEETLE_OPT_LAST_SCANLINE_LABEL_AR NULL
#define BEETLE_OPT_LAST_SCANLINE_INFO_0_AR NULL
#define OPTION_VAL_239_AR NULL
#define BEETLE_OPT_INITIAL_SCANLINE_PAL_LABEL_AR NULL
#define BEETLE_OPT_INITIAL_SCANLINE_PAL_INFO_0_AR NULL
#define BEETLE_OPT_LAST_SCANLINE_PAL_LABEL_AR NULL
#define BEETLE_OPT_LAST_SCANLINE_PAL_INFO_0_AR NULL
#define OPTION_VAL_287_AR NULL

struct retro_core_option_v2_category option_cats_ar[] = {
   {
      "system",
      CATEGORY_SYSTEM_LABEL_AR,
      CATEGORY_SYSTEM_INFO_0_AR
   },
   {
      "video",
      CATEGORY_VIDEO_LABEL_AR,
      CATEGORY_VIDEO_INFO_0_AR
   },
   {
      "pgxp",
      CATEGORY_PGXP_LABEL_AR,
      CATEGORY_PGXP_INFO_0_AR
   },
   {
      "osd",
      CATEGORY_OSD_LABEL_AR,
      CATEGORY_OSD_INFO_0_AR
   },
   {
      "input",
      CATEGORY_INPUT_LABEL_AR,
      CATEGORY_INPUT_INFO_0_AR
   },
   {
      "memcards",
      CATEGORY_MEMCARDS_LABEL_AR,
      CATEGORY_MEMCARDS_INFO_0_AR
   },
   {
      "hacks",
      CATEGORY_HACKS_LABEL_AR,
      CATEGORY_HACKS_INFO_0_AR
   },
   { NULL, NULL, NULL },
};
struct retro_core_option_v2_definition option_defs_ar[] = {
   {
      BEETLE_OPT(cpu_freq_scale),
      BEETLE_OPT_CPU_FREQ_SCALE_LABEL_AR,
      NULL,
      BEETLE_OPT_CPU_FREQ_SCALE_INFO_0_AR,
      NULL,
      "system",
      {
         { "50%",           OPTION_VAL_50_AR },
         { "60%",           OPTION_VAL_60_AR },
         { "70%",           OPTION_VAL_70_AR },
         { "80%",           OPTION_VAL_80_AR },
         { "90%",           OPTION_VAL_90_AR },
         { "100%",          OPTION_VAL_100_AR },
         { "110%",          OPTION_VAL_110_AR },
         { "120%",          OPTION_VAL_120_AR },
         { "130%",          OPTION_VAL_130_AR },
         { "140%",          OPTION_VAL_140_AR },
         { "150%",          OPTION_VAL_150_AR },
         { "160%",          OPTION_VAL_160_AR },
         { "170%",          OPTION_VAL_170_AR },
         { "180%",          OPTION_VAL_180_AR },
         { "190%",          OPTION_VAL_190_AR },
         { "200%",          OPTION_VAL_200_AR },
         { "210%",          OPTION_VAL_210_AR },
         { "220%",          OPTION_VAL_220_AR },
         { "230%",          OPTION_VAL_230_AR },
         { "240%",          OPTION_VAL_240_AR },
         { "250%",          OPTION_VAL_250_AR },
         { "260%",          OPTION_VAL_260_AR },
         { "270%",          OPTION_VAL_270_AR },
         { "280%",          OPTION_VAL_280_AR },
         { "290%",          OPTION_VAL_290_AR },
         { "300%",          OPTION_VAL_300_AR },
         { "310%",          OPTION_VAL_310_AR },
         { "320%",          OPTION_VAL_320_AR },
         { "330%",          OPTION_VAL_330_AR },
         { "340%",          OPTION_VAL_340_AR },
         { "350%",          OPTION_VAL_350_AR },
         { "360%",          OPTION_VAL_360_AR },
         { "370%",          OPTION_VAL_370_AR },
         { "380%",          OPTION_VAL_380_AR },
         { "390%",          OPTION_VAL_390_AR },
         { "400%",          OPTION_VAL_400_AR },
         { "410%",          OPTION_VAL_410_AR },
         { "420%",          OPTION_VAL_420_AR },
         { "430%",          OPTION_VAL_430_AR },
         { "440%",          OPTION_VAL_440_AR },
         { "450%",          OPTION_VAL_450_AR },
         { "460%",          OPTION_VAL_460_AR },
         { "470%",          OPTION_VAL_470_AR },
         { "480%",          OPTION_VAL_480_AR },
         { "490%",          OPTION_VAL_490_AR },
         { "500%",          OPTION_VAL_500_AR },
         { "510%",          OPTION_VAL_510_AR },
         { "520%",          OPTION_VAL_520_AR },
         { "530%",          OPTION_VAL_530_AR },
         { "540%",          OPTION_VAL_540_AR },
         { "550%",          OPTION_VAL_550_AR },
         { "560%",          OPTION_VAL_560_AR },
         { "570%",          OPTION_VAL_570_AR },
         { "580%",          OPTION_VAL_580_AR },
         { "590%",          OPTION_VAL_590_AR },
         { "600%",          OPTION_VAL_600_AR },
         { "610%",          OPTION_VAL_610_AR },
         { "620%",          OPTION_VAL_620_AR },
         { "630%",          OPTION_VAL_630_AR },
         { "640%",          OPTION_VAL_640_AR },
         { "650%",          OPTION_VAL_650_AR },
         { "660%",          OPTION_VAL_660_AR },
         { "670%",          OPTION_VAL_670_AR },
         { "680%",          OPTION_VAL_680_AR },
         { "690%",          OPTION_VAL_690_AR },
         { "700%",          OPTION_VAL_700_AR },
         { "710%",          OPTION_VAL_710_AR },
         { "720%",          OPTION_VAL_720_AR },
         { "730%",          OPTION_VAL_730_AR },
         { "740%",          OPTION_VAL_740_AR },
         { "750%",          OPTION_VAL_750_AR },
         { NULL, NULL },
      },
      "100%"
   },
   {
      BEETLE_OPT(gpu_overclock),
      BEETLE_OPT_GPU_OVERCLOCK_LABEL_AR,
      NULL,
      BEETLE_OPT_GPU_OVERCLOCK_INFO_0_AR,
      NULL,
      "system",
      {
         { "1x(native)", OPTION_VAL_1X_NATIVE_AR },
         { "2x",         OPTION_VAL_2X_AR },
         { "4x",         OPTION_VAL_4X_AR },
         { "8x",         OPTION_VAL_8X_AR },
         { "16x",        OPTION_VAL_16X_AR },
         { "32x",        OPTION_VAL_32X_AR },
         { NULL, NULL },
      },
      "1x(native)"
   },
   {
      BEETLE_OPT(gte_overclock),
      BEETLE_OPT_GTE_OVERCLOCK_LABEL_AR,
      NULL,
      BEETLE_OPT_GTE_OVERCLOCK_INFO_0_AR,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
#ifndef EMSCRIPTEN
   {
      BEETLE_OPT(cd_access_method),
      BEETLE_OPT_CD_ACCESS_METHOD_LABEL_AR,
      NULL,
      BEETLE_OPT_CD_ACCESS_METHOD_INFO_0_AR,
      NULL,
      "system",
      {
         { "sync",     OPTION_VAL_SYNC_AR },
         { "async",    OPTION_VAL_ASYNC_AR },
         { "precache", OPTION_VAL_PRECACHE_AR },
         { NULL, NULL },
      },
      "sync"
   },
#endif
   {
      BEETLE_OPT(cd_fastload),
      BEETLE_OPT_CD_FASTLOAD_LABEL_AR,
      NULL,
      BEETLE_OPT_CD_FASTLOAD_INFO_0_AR,
      NULL,
      "system",
      {
         { "2x(native)", OPTION_VAL_2X_NATIVE_AR },
         { "4x",          OPTION_VAL_4X_AR },
         { "6x",          OPTION_VAL_6X_AR },
         { "8x",          OPTION_VAL_8X_AR },
         { "10x",         OPTION_VAL_10X_AR },
         { "12x",         OPTION_VAL_12X_AR },
         { "14x",         OPTION_VAL_14X_AR },
         { NULL, NULL },
      },
      "2x(native)"
   },
   {
      BEETLE_OPT(skip_bios),
      BEETLE_OPT_SKIP_BIOS_LABEL_AR,
      NULL,
      BEETLE_OPT_SKIP_BIOS_INFO_0_AR,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(override_bios),
      BEETLE_OPT_OVERRIDE_BIOS_LABEL_AR,
      NULL,
      BEETLE_OPT_OVERRIDE_BIOS_INFO_0_AR,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "psxonpsp", OPTION_VAL_PSXONPSP_AR },
         { "ps1_rom",  OPTION_VAL_PS1_ROM_AR },
         { "openbios", OPTION_VAL_OPENBIOS_AR     },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(region),
      BEETLE_OPT_REGION_LABEL_AR,
      NULL,
      BEETLE_OPT_REGION_INFO_0_AR,
      NULL,
      "system",
      {
         { "auto",   OPTION_VAL_AUTO_AR },
         { "ntsc-j", OPTION_VAL_NTSC_J_AR },
         { "ntsc-u", OPTION_VAL_NTSC_U_AR },
         { "pal",    OPTION_VAL_PAL_AR },
         { NULL, NULL },
      },
      "auto"
   },
   {
      BEETLE_OPT(internal_resolution),
      BEETLE_OPT_INTERNAL_RESOLUTION_LABEL_AR,
      NULL,
      BEETLE_OPT_INTERNAL_RESOLUTION_INFO_0_AR,
      NULL,
      "video",
      {
         { "1x(native)", OPTION_VAL_1X_NATIVE_AR },
         { "2x",         OPTION_VAL_2X_AR },
         { "4x",         OPTION_VAL_4X_AR },
         { "8x",         OPTION_VAL_8X_AR },
         { "16x",        OPTION_VAL_16X_AR },
         { NULL, NULL },
      },
      "1x(native)"
   },
#if defined(HAVE_OPENGL) || defined(HAVE_OPENGLES) || defined(HAVE_VULKAN)
   {
      BEETLE_OPT(renderer),
      BEETLE_OPT_RENDERER_LABEL_AR,
      NULL,
      BEETLE_OPT_RENDERER_INFO_0_AR,
      NULL,
      "video",
      {
         { "hardware",    OPTION_VAL_HARDWARE_AR },
#if defined(HAVE_OPENGL) || defined(HAVE_OPENGLES)
         { "hardware_gl", OPTION_VAL_HARDWARE_GL_AR },
#endif
#if defined(HAVE_VULKAN)
         { "hardware_vk", OPTION_VAL_HARDWARE_VK_AR },
#endif
         { "software",    OPTION_VAL_SOFTWARE_AR },
         { NULL, NULL },
      },
      "hardware"
   },
   {
      BEETLE_OPT(renderer_software_fb),
      BEETLE_OPT_RENDERER_SOFTWARE_FB_LABEL_AR,
      NULL,
      BEETLE_OPT_RENDERER_SOFTWARE_FB_INFO_0_AR,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled"
   },
#endif
#if defined(HAVE_OPENGL) || defined(HAVE_OPENGLES)
   {
      BEETLE_OPT(depth),
      BEETLE_OPT_DEPTH_LABEL_AR,
      NULL,
      BEETLE_OPT_DEPTH_INFO_0_AR,
      NULL,
      "video",
      {
         { "16bpp(native)", OPTION_VAL_16BPP_NATIVE_AR },
         { "32bpp",         OPTION_VAL_32BPP_AR },
         { NULL, NULL },
      },
      "16bpp(native)"
   },
   // Sort of, it's more like 15-bit high color and 24-bit true color for visible output. The alpha channel is used for mask bit. Vulkan renderer uses ABGR1555_555 for 31 bits internal? FMVs are always 24-bit on all renderers like original hardware (BGR888, no alpha)
#endif
   {
      BEETLE_OPT(dither_mode),
      BEETLE_OPT_DITHER_MODE_LABEL_AR,
      NULL,
      BEETLE_OPT_DITHER_MODE_INFO_0_AR,
      NULL,
      "video",
      {
         { "1x(native)",          OPTION_VAL_1X_NATIVE_AR },
         { "internal resolution", OPTION_VAL_INTERNAL_RESOLUTION_AR },
         { "disabled",            NULL },
         { NULL, NULL },
      },
      "1x(native)"
   },
#ifdef HAVE_VULKAN
   {
      BEETLE_OPT(scaled_uv_offset),
      BEETLE_OPT_SCALED_UV_OFFSET_LABEL_AR,
      NULL,
      BEETLE_OPT_SCALED_UV_OFFSET_INFO_0_AR,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled"
   },
#endif
#if defined(HAVE_OPENGL) || defined(HAVE_OPENGLES) || defined(HAVE_VULKAN)
   {
      BEETLE_OPT(filter),
      BEETLE_OPT_FILTER_LABEL_AR,
      NULL,
      BEETLE_OPT_FILTER_INFO_0_AR,
      NULL,
      "video",
      {
         { "nearest",  OPTION_VAL_NEAREST_AR },
         { "SABR",     OPTION_VAL_SABR_AR },
         { "xBR",      OPTION_VAL_XBR_AR },
         { "bilinear", OPTION_VAL_BILINEAR_AR },
         { "3-point",  OPTION_VAL_3_POINT_AR },
         { "JINC2",    OPTION_VAL_JINC2_AR },
         { NULL, NULL },
      },
      "nearest"
   },
#ifdef HAVE_VULKAN
   {
      BEETLE_OPT(filter_exclude_sprite),
      BEETLE_OPT_FILTER_EXCLUDE_SPRITE_LABEL_AR,
      NULL,
      BEETLE_OPT_FILTER_EXCLUDE_SPRITE_INFO_0_AR,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "opaque", OPTION_VAL_OPAQUE_AR },
         { "all", OPTION_VAL_ALL_AR },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(filter_exclude_2d_polygon),
      BEETLE_OPT_FILTER_EXCLUDE_2D_POLYGON_LABEL_AR,
      NULL,
      BEETLE_OPT_FILTER_EXCLUDE_2D_POLYGON_INFO_0_AR,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "opaque", OPTION_VAL_OPAQUE_AR },
         { "all", OPTION_VAL_ALL_AR },
         { NULL, NULL },
      },
      "disabled"
   },
#endif
#endif
#ifdef HAVE_VULKAN
   {
      BEETLE_OPT(adaptive_smoothing),
      BEETLE_OPT_ADAPTIVE_SMOOTHING_LABEL_AR,
      NULL,
      BEETLE_OPT_ADAPTIVE_SMOOTHING_INFO_0_AR,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(super_sampling),
      BEETLE_OPT_SUPER_SAMPLING_LABEL_AR,
      NULL,
      BEETLE_OPT_SUPER_SAMPLING_INFO_0_AR,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(msaa),
      BEETLE_OPT_MSAA_LABEL_AR,
      NULL,
      BEETLE_OPT_MSAA_INFO_0_AR,
      NULL,
      "video",
      {
         { "1x",  OPTION_VAL_1X_AR },
         { "2x",  OPTION_VAL_2X_AR },
         { "4x",  OPTION_VAL_4X_AR },
         { "8x",  OPTION_VAL_8X_AR },
         { "16x", OPTION_VAL_16X_AR },
         { NULL, NULL },
      },
      "1x"
   },
   {
      BEETLE_OPT(mdec_yuv),
      BEETLE_OPT_MDEC_YUV_LABEL_AR,
      NULL,
      BEETLE_OPT_MDEC_YUV_INFO_0_AR,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(track_textures),
      BEETLE_OPT_TRACK_TEXTURES_LABEL_AR,
      NULL,
      BEETLE_OPT_TRACK_TEXTURES_INFO_0_AR,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
#ifdef TEXTURE_DUMPING_ENABLED
   {
      BEETLE_OPT(dump_textures),
      BEETLE_OPT_DUMP_TEXTURES_LABEL_AR,
      NULL,
      BEETLE_OPT_DUMP_TEXTURES_INFO_0_AR,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
#endif
   {
      BEETLE_OPT(replace_textures),
      BEETLE_OPT_REPLACE_TEXTURES_LABEL_AR,
      NULL,
      BEETLE_OPT_REPLACE_TEXTURES_INFO_0_AR,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
#endif
   {
      BEETLE_OPT(pgxp_mode),
      BEETLE_OPT_PGXP_MODE_LABEL_AR,
      NULL,
      BEETLE_OPT_PGXP_MODE_INFO_0_AR,
      NULL,
      "pgxp",
      {
         { "disabled",     NULL },
         { "memory only",  OPTION_VAL_MEMORY_ONLY_AR },
         { "memory + CPU", OPTION_VAL_MEMORY_CPU_AR },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(pgxp_2d_tol),
      BEETLE_OPT_PGXP_2D_TOL_LABEL_AR,
      NULL,
      BEETLE_OPT_PGXP_2D_TOL_INFO_0_AR,
      NULL,
      "pgxp",
      {
         { "disabled", NULL },
         { "0px", OPTION_VAL_0PX_AR },
         { "1px", OPTION_VAL_1PX_AR },
         { "2px", OPTION_VAL_2PX_AR },
         { "3px", OPTION_VAL_3PX_AR },
         { "4px", OPTION_VAL_4PX_AR },
         { "5px", OPTION_VAL_5PX_AR },
         { "6px", OPTION_VAL_6PX_AR },
         { "7px", OPTION_VAL_7PX_AR },
         { "8px", OPTION_VAL_8PX_AR },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(pgxp_nclip),
      BEETLE_OPT_PGXP_NCLIP_LABEL_AR,
      NULL,
      BEETLE_OPT_PGXP_NCLIP_INFO_0_AR,
      NULL,
      "pgxp",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
#if defined(HAVE_OPENGL) || defined(HAVE_OPENGLES) || defined(HAVE_VULKAN)
   {
      BEETLE_OPT(pgxp_vertex),
      BEETLE_OPT_PGXP_VERTEX_LABEL_AR,
      NULL,
      BEETLE_OPT_PGXP_VERTEX_INFO_0_AR,
      NULL,
      "pgxp",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
#endif
   {
      BEETLE_OPT(pgxp_texture),
      BEETLE_OPT_PGXP_TEXTURE_LABEL_AR,
      NULL,
      BEETLE_OPT_PGXP_TEXTURE_INFO_0_AR,
      NULL,
      "pgxp",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(display_internal_fps),
      BEETLE_OPT_DISPLAY_INTERNAL_FPS_LABEL_AR,
      NULL,
      BEETLE_OPT_DISPLAY_INTERNAL_FPS_INFO_0_AR,
      NULL,
      "osd",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(display_osd),
      BEETLE_OPT_DISPLAY_OSD_LABEL_AR,
      NULL,
      BEETLE_OPT_DISPLAY_OSD_INFO_0_AR,
      NULL,
      "osd",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled"
   },
#if defined(HAVE_OPENGL) || defined(HAVE_OPENGLES) || defined(HAVE_VULKAN)
   {
      BEETLE_OPT(display_vram),
      BEETLE_OPT_DISPLAY_VRAM_LABEL_AR,
      NULL,
      BEETLE_OPT_DISPLAY_VRAM_INFO_0_AR,
      NULL,
      "osd",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
#endif
   {
      BEETLE_OPT(analog_calibration),
      BEETLE_OPT_ANALOG_CALIBRATION_LABEL_AR,
      NULL,
      BEETLE_OPT_ANALOG_CALIBRATION_INFO_0_AR,
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(analog_toggle),
      BEETLE_OPT_ANALOG_TOGGLE_LABEL_AR,
      NULL,
      BEETLE_OPT_ANALOG_TOGGLE_INFO_0_AR,
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { "enabled-analog", OPTION_VAL_ENABLED_ANALOG_AR },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(analog_toggle_combo),
      BEETLE_OPT_ANALOG_TOGGLE_COMBO_LABEL_AR,
      NULL,
      BEETLE_OPT_ANALOG_TOGGLE_COMBO_INFO_0_AR,
      NULL,
      "input",
      {
         { "l1+l2+r1+r2+start+select", OPTION_VAL_L1_L2_R1_R2_START_SELECT_AR },
         { "l1+r1+select",             OPTION_VAL_L1_R1_SELECT_AR },
         { "l1+r1+start",              OPTION_VAL_L1_R1_START_AR },
         { "l1+r1+l3",                 OPTION_VAL_L1_R1_L3_AR },
         { "l1+r1+r3",                 OPTION_VAL_L1_R1_R3_AR },
         { "l2+r2+select",             OPTION_VAL_L2_R2_SELECT_AR },
         { "l2+r2+start",              OPTION_VAL_L2_R2_START_AR },
         { "l2+r2+l3",                 OPTION_VAL_L2_R2_L3_AR },
         { "l2+r2+r3",                 OPTION_VAL_L2_R2_R3_AR },
         { "l3+r3",                    OPTION_VAL_L3_R3_AR },
         { NULL, NULL },
      },
      "l1+r1+select"
   },
   {
      BEETLE_OPT(analog_toggle_hold),
      BEETLE_OPT_ANALOG_TOGGLE_HOLD_LABEL_AR,
      NULL,
      BEETLE_OPT_ANALOG_TOGGLE_HOLD_INFO_0_AR,
      NULL,
      "input",
      {
         { "0", OPTION_VAL_0_AR },
         { "1", OPTION_VAL_1_AR },
         { "2", OPTION_VAL_2_AR },
         { "3", OPTION_VAL_3_AR },
         { "4", OPTION_VAL_4_AR },
         { "5", OPTION_VAL_5_AR },
         { NULL, NULL },
      },
      "1"
   },
   {
      BEETLE_OPT(enable_multitap_port1),
      BEETLE_OPT_ENABLE_MULTITAP_PORT1_LABEL_AR,
      NULL,
      BEETLE_OPT_ENABLE_MULTITAP_PORT1_INFO_0_AR,
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(enable_multitap_port2),
      BEETLE_OPT_ENABLE_MULTITAP_PORT2_LABEL_AR,
      NULL,
      BEETLE_OPT_ENABLE_MULTITAP_PORT2_INFO_0_AR,
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(gun_input_mode),
      BEETLE_OPT_GUN_INPUT_MODE_LABEL_AR,
      NULL,
      BEETLE_OPT_GUN_INPUT_MODE_INFO_0_AR,
      NULL,
      "input",
      {
         { "lightgun",    OPTION_VAL_LIGHTGUN_AR },
         { "touchscreen", OPTION_VAL_TOUCHSCREEN_AR },
         { NULL, NULL },
      },
      "lightgun"
   },
   // Shouldn't the gun_input_mode just be Mouse vs. Touchscreen?
   {
      BEETLE_OPT(gun_cursor),
      BEETLE_OPT_GUN_CURSOR_LABEL_AR,
      NULL,
      BEETLE_OPT_GUN_CURSOR_INFO_0_AR,
      NULL,
      "input",
      {
         { "off",   "disabled" },
         { "cross", OPTION_VAL_CROSS_AR },
         { "dot",   OPTION_VAL_DOT_AR },
         { NULL, NULL },
      },
      "cross"
   },
   {
      BEETLE_OPT(crosshair_color_p1),
      BEETLE_OPT_CROSSHAIR_COLOR_P1_LABEL_AR,
      NULL,
      BEETLE_OPT_CROSSHAIR_COLOR_P1_INFO_0_AR,
      NULL,
      "input",
      {
         { "red",    OPTION_VAL_RED_AR },
         { "blue",   OPTION_VAL_BLUE_AR },
         { "green",  OPTION_VAL_GREEN_AR },
         { "orange", OPTION_VAL_ORANGE_AR },
         { "yellow", OPTION_VAL_YELLOW_AR },
         { "cyan",   OPTION_VAL_CYAN_AR },
         { "pink",   OPTION_VAL_PINK_AR },
         { "purple", OPTION_VAL_PURPLE_AR },
         { "black",  OPTION_VAL_BLACK_AR },
         { "white",  OPTION_VAL_WHITE_AR },
         { NULL, NULL },
      },
      "red"
   },
   {
      BEETLE_OPT(crosshair_color_p2),
      BEETLE_OPT_CROSSHAIR_COLOR_P2_LABEL_AR,
      NULL,
      BEETLE_OPT_CROSSHAIR_COLOR_P2_INFO_0_AR,
      NULL,
      "input",
      {
         { "blue",   OPTION_VAL_BLUE_AR },
         { "red",    OPTION_VAL_RED_AR },
         { "green",  OPTION_VAL_GREEN_AR },
         { "orange", OPTION_VAL_ORANGE_AR },
         { "yellow", OPTION_VAL_YELLOW_AR },
         { "cyan",   OPTION_VAL_CYAN_AR },
         { "pink",   OPTION_VAL_PINK_AR },
         { "purple", OPTION_VAL_PURPLE_AR },
         { "black",  OPTION_VAL_BLACK_AR },
         { "white",  OPTION_VAL_WHITE_AR },
         { NULL, NULL },
      },
      "blue"
   },
   {
      BEETLE_OPT(mouse_sensitivity),
      BEETLE_OPT_MOUSE_SENSITIVITY_LABEL_AR,
      NULL,
      BEETLE_OPT_MOUSE_SENSITIVITY_INFO_0_AR,
      NULL,
      "input",
      {
         { "5%",   OPTION_VAL_5_O42_AR },
         { "10%",  OPTION_VAL_10_AR },
         { "15%",  OPTION_VAL_15_AR },
         { "20%",  OPTION_VAL_20_AR },
         { "25%",  OPTION_VAL_25_AR },
         { "30%",  OPTION_VAL_30_AR },
         { "35%",  OPTION_VAL_35_AR },
         { "40%",  OPTION_VAL_40_AR },
         { "45%",  OPTION_VAL_45_AR },
         { "50%",  OPTION_VAL_50_AR },
         { "55%",  OPTION_VAL_55_AR },
         { "60%",  OPTION_VAL_60_AR },
         { "65%",  OPTION_VAL_65_AR },
         { "70%",  OPTION_VAL_70_AR },
         { "75%",  OPTION_VAL_75_AR },
         { "80%",  OPTION_VAL_80_AR },
         { "85%",  OPTION_VAL_85_AR },
         { "90%",  OPTION_VAL_90_AR },
         { "95%",  OPTION_VAL_95_AR },
         { "100%", OPTION_VAL_100_O42_AR },
         { "105%", OPTION_VAL_105_AR },
         { "110%", OPTION_VAL_110_AR },
         { "115%", OPTION_VAL_115_AR },
         { "120%", OPTION_VAL_120_AR },
         { "125%", OPTION_VAL_125_AR },
         { "130%", OPTION_VAL_130_AR },
         { "135%", OPTION_VAL_135_AR },
         { "140%", OPTION_VAL_140_AR },
         { "145%", OPTION_VAL_145_AR },
         { "150%", OPTION_VAL_150_AR },
         { "155%", OPTION_VAL_155_AR },
         { "160%", OPTION_VAL_160_AR },
         { "165%", OPTION_VAL_165_AR },
         { "170%", OPTION_VAL_170_AR },
         { "175%", OPTION_VAL_175_AR },
         { "180%", OPTION_VAL_180_AR },
         { "185%", OPTION_VAL_185_AR },
         { "190%", OPTION_VAL_190_AR },
         { "195%", OPTION_VAL_195_AR },
         { "200%", OPTION_VAL_200_AR },
         { NULL, NULL },
      },
      "100%"
   },
   {
      BEETLE_OPT(negcon_response),
      BEETLE_OPT_NEGCON_RESPONSE_LABEL_AR,
      NULL,
      BEETLE_OPT_NEGCON_RESPONSE_INFO_0_AR,
      NULL,
      "input",
      {
         { "linear",    OPTION_VAL_LINEAR_AR },
         { "quadratic", OPTION_VAL_QUADRATIC_AR },
         { "cubic",     OPTION_VAL_CUBIC_AR },
         { NULL, NULL },
      },
      "linear"
   },
   {
      BEETLE_OPT(negcon_deadzone),
      BEETLE_OPT_NEGCON_DEADZONE_LABEL_AR,
      NULL,
      BEETLE_OPT_NEGCON_DEADZONE_INFO_0_AR,
      NULL,
      "input",
      {
         { "0%",  OPTION_VAL_0_O44_AR },
         { "5%",  OPTION_VAL_5_O42_AR },
         { "10%", OPTION_VAL_10_AR },
         { "15%", OPTION_VAL_15_AR },
         { "20%", OPTION_VAL_20_AR },
         { "25%", OPTION_VAL_25_AR },
         { "30%", OPTION_VAL_30_AR },
         { NULL, NULL },
      },
      "0%"
   },
   {
      BEETLE_OPT(use_mednafen_memcard0_method),
      BEETLE_OPT_USE_MEDNAFEN_MEMCARD0_METHOD_LABEL_AR,
      NULL,
      BEETLE_OPT_USE_MEDNAFEN_MEMCARD0_METHOD_INFO_0_AR,
      NULL,
      "memcards",
      {
         { "libretro", OPTION_VAL_LIBRETRO_AR },
         { "mednafen", OPTION_VAL_MEDNAFEN_AR },
         { NULL, NULL },
      },
      "libretro"
   },
   {
      BEETLE_OPT(enable_memcard1),
      BEETLE_OPT_ENABLE_MEMCARD1_LABEL_AR,
      NULL,
      BEETLE_OPT_ENABLE_MEMCARD1_INFO_0_AR,
      NULL,
      "memcards",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(shared_memory_cards),
      BEETLE_OPT_SHARED_MEMORY_CARDS_LABEL_AR,
      NULL,
      BEETLE_OPT_SHARED_MEMORY_CARDS_INFO_0_AR,
      NULL,
      "memcards",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(memcard_left_index),
      BEETLE_OPT_MEMCARD_LEFT_INDEX_LABEL_AR,
      NULL,
      BEETLE_OPT_MEMCARD_LEFT_INDEX_INFO_0_AR,
      NULL,
      "memcards",
      {
         { "0",  OPTION_VAL_0_O48_AR },
         { "1",  NULL },
         { "2",  NULL },
         { "3",  NULL },
         { "4",  NULL },
         { "5",  NULL },
         { "6",  NULL },
         { "7",  NULL },
         { "8",  NULL },
         { "9",  NULL },
         { "10", NULL },
         { "11", NULL },
         { "12", NULL },
         { "13", NULL },
         { "14", NULL },
         { "15", NULL },
         { "16", NULL },
         { "17", NULL },
         { "18", NULL },
         { "19", NULL },
         { "20", NULL },
         { "21", NULL },
         { "22", NULL },
         { "23", NULL },
         { "24", NULL },
         { "25", NULL },
         { "26", NULL },
         { "27", NULL },
         { "28", NULL },
         { "29", NULL },
         { "30", NULL },
         { "31", NULL },
         { "32", NULL },
         { "33", NULL },
         { "34", NULL },
         { "35", NULL },
         { "36", NULL },
         { "37", NULL },
         { "38", NULL },
         { "39", NULL },
         { "40", NULL },
         { "41", NULL },
         { "42", NULL },
         { "43", NULL },
         { "44", NULL },
         { "45", NULL },
         { "46", NULL },
         { "47", NULL },
         { "48", NULL },
         { "49", NULL },
         { "50", NULL },
         { "51", NULL },
         { "52", NULL },
         { "53", NULL },
         { "54", NULL },
         { "55", NULL },
         { "56", NULL },
         { "57", NULL },
         { "58", NULL },
         { "59", NULL },
         { "60", NULL },
         { "61", NULL },
         { "62", NULL },
         { "63", NULL },
         { NULL, NULL },
      },
      "0"
   },
   {
      BEETLE_OPT(memcard_right_index),
      BEETLE_OPT_MEMCARD_RIGHT_INDEX_LABEL_AR,
      NULL,
      BEETLE_OPT_MEMCARD_RIGHT_INDEX_INFO_0_AR,
      NULL,
      "memcards",
      {
         { "0",  NULL },
         { "1",  OPTION_VAL_1_O49_AR },
         { "2",  NULL },
         { "3",  NULL },
         { "4",  NULL },
         { "5",  NULL },
         { "6",  NULL },
         { "7",  NULL },
         { "8",  NULL },
         { "9",  NULL },
         { "10", NULL },
         { "11", NULL },
         { "12", NULL },
         { "13", NULL },
         { "14", NULL },
         { "15", NULL },
         { "16", NULL },
         { "17", NULL },
         { "18", NULL },
         { "19", NULL },
         { "20", NULL },
         { "21", NULL },
         { "22", NULL },
         { "23", NULL },
         { "24", NULL },
         { "25", NULL },
         { "26", NULL },
         { "27", NULL },
         { "28", NULL },
         { "29", NULL },
         { "30", NULL },
         { "31", NULL },
         { "32", NULL },
         { "33", NULL },
         { "34", NULL },
         { "35", NULL },
         { "36", NULL },
         { "37", NULL },
         { "38", NULL },
         { "39", NULL },
         { "40", NULL },
         { "41", NULL },
         { "42", NULL },
         { "43", NULL },
         { "44", NULL },
         { "45", NULL },
         { "46", NULL },
         { "47", NULL },
         { "48", NULL },
         { "49", NULL },
         { "50", NULL },
         { "51", NULL },
         { "52", NULL },
         { "53", NULL },
         { "54", NULL },
         { "55", NULL },
         { "56", NULL },
         { "57", NULL },
         { "58", NULL },
         { "59", NULL },
         { "60", NULL },
         { "61", NULL },
         { "62", NULL },
         { "63", NULL },
         { NULL, NULL },
      },
      "1"
   },
   {
      BEETLE_OPT(line_render),
      BEETLE_OPT_LINE_RENDER_LABEL_AR,
      NULL,
      BEETLE_OPT_LINE_RENDER_INFO_0_AR,
      NULL,
      "hacks",
      {
         { "default",    OPTION_VAL_DEFAULT_AR },
         { "aggressive", OPTION_VAL_AGGRESSIVE_AR },
         { "disabled",   NULL },
         { NULL, NULL },
      },
      "default"
   },
   {
      BEETLE_OPT(widescreen_hack),
      BEETLE_OPT_WIDESCREEN_HACK_LABEL_AR,
      NULL,
      BEETLE_OPT_WIDESCREEN_HACK_INFO_0_AR,
      NULL,
      "hacks",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(widescreen_hack_aspect_ratio),
      BEETLE_OPT_WIDESCREEN_HACK_ASPECT_RATIO_LABEL_AR,
      NULL,
      BEETLE_OPT_WIDESCREEN_HACK_ASPECT_RATIO_INFO_0_AR,
      NULL,
      "hacks",
      {
         { "16:9",  OPTION_VAL_16_9_AR },
         { "16:10", OPTION_VAL_16_10_AR },
         { "18:9",  OPTION_VAL_18_9_AR },
         { "19:9",  OPTION_VAL_19_9_AR },
         { "20:9",  OPTION_VAL_20_9_AR },
         { "21:9",  OPTION_VAL_21_9_AR }, // 64:27
         { "32:9",  OPTION_VAL_32_9_AR },
         { NULL,    NULL },
      },
      "16:9"
   },
#if defined(HAVE_LIGHTREC)
   {
      BEETLE_OPT(cpu_dynarec),
      BEETLE_OPT_CPU_DYNAREC_LABEL_AR,
      NULL,
      BEETLE_OPT_CPU_DYNAREC_INFO_0_AR,
      NULL,
      "hacks",
      {
         { "disabled",        OPTION_VAL_DISABLED_AR },
         { "execute",         OPTION_VAL_EXECUTE_AR },
         { "execute_one",     OPTION_VAL_EXECUTE_ONE_AR },
         { "run_interpreter", OPTION_VAL_RUN_INTERPRETER_AR },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(dynarec_invalidate),
      BEETLE_OPT_DYNAREC_INVALIDATE_LABEL_AR,
      NULL,
      BEETLE_OPT_DYNAREC_INVALIDATE_INFO_0_AR,
      NULL,
      "hacks",
      {
         { "full", OPTION_VAL_FULL_AR },
         { "dma",  OPTION_VAL_DMA_AR },
         { NULL, NULL },
      },
      "full"
   },
   {
      BEETLE_OPT(dynarec_eventcycles),
      BEETLE_OPT_DYNAREC_EVENTCYCLES_LABEL_AR,
      NULL,
      BEETLE_OPT_DYNAREC_EVENTCYCLES_INFO_0_AR,
      NULL,
      "hacks",
      {
         { "128",  OPTION_VAL_128_AR },
         { "256",  NULL },
         { "384",  NULL },
         { "512",  NULL },
         { "640",  NULL },
         { "768",  NULL },
         { "896",  NULL },
         { "1024", NULL },
         { "1152", NULL },
         { "1280", NULL },
         { "1408", NULL },
         { "1536", NULL },
         { "1664", NULL },
         { "1792", NULL },
         { "1920", NULL },
         { "2048", NULL },
         { NULL, NULL },
      },
      "128"
   },
   {
      BEETLE_OPT(dynarec_spu_samples),
      BEETLE_OPT_DYNAREC_SPU_SAMPLES_LABEL_AR,
      NULL,
      BEETLE_OPT_DYNAREC_SPU_SAMPLES_INFO_0_AR,
      NULL,
      "hacks",
      {
         { "1",  OPTION_VAL_1_O49_AR },
         { "4",  NULL },
         { "16", NULL },
         { NULL, NULL },
      },
      "1"
   },
#endif
   {
      BEETLE_OPT(core_timing_fps),
      BEETLE_OPT_CORE_TIMING_FPS_LABEL_AR,
      NULL,
      BEETLE_OPT_CORE_TIMING_FPS_INFO_0_AR,
      NULL,
      "video",
      {
         { "auto_toggle",       OPTION_VAL_AUTO_TOGGLE_AR },
         { "force_progressive", OPTION_VAL_FORCE_PROGRESSIVE_AR },
         { "force_interlaced",  OPTION_VAL_FORCE_INTERLACED_AR },
      },
      "auto_toggle"
   },
   {
      BEETLE_OPT(deinterlacer),
      BEETLE_OPT_DEINTERLACER_LABEL_AR,
      NULL,
      BEETLE_OPT_DEINTERLACER_INFO_0_AR,
      NULL,
      "video",
      {
         { "weave",      OPTION_VAL_WEAVE_AR },
         { "bob",        OPTION_VAL_BOB_AR },
         { "bob_offset", OPTION_VAL_BOB_OFFSET_AR },
         { "fastmad",    OPTION_VAL_FASTMAD_AR },
         { "off",        "Off" },
         { NULL, NULL },
      },
      "weave"
   },
   {
      BEETLE_OPT(frame_duping),
      BEETLE_OPT_FRAME_DUPING_LABEL_AR,
      NULL,
      BEETLE_OPT_FRAME_DUPING_INFO_0_AR,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(pal_video_timing_override),
      BEETLE_OPT_PAL_VIDEO_TIMING_OVERRIDE_LABEL_AR,
      NULL,
      BEETLE_OPT_PAL_VIDEO_TIMING_OVERRIDE_INFO_0_AR,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(aspect_ratio),
      BEETLE_OPT_ASPECT_RATIO_LABEL_AR,
      NULL,
      BEETLE_OPT_ASPECT_RATIO_INFO_0_AR,
      NULL,
      "video",
      {
         { "corrected", OPTION_VAL_CORRECTED_AR },
         { "uncorrected", OPTION_VAL_UNCORRECTED_AR },
         { "4:3",  OPTION_VAL_4_3_AR },
         { "ntsc", OPTION_VAL_NTSC_AR },
      },
      "corrected"
   },
   {
      BEETLE_OPT(crop_overscan),
      BEETLE_OPT_CROP_OVERSCAN_LABEL_AR,
      NULL,
      BEETLE_OPT_CROP_OVERSCAN_INFO_0_AR,
      NULL,
      "video",
      {
         { "disabled", "Disabled" },
         { "static",   OPTION_VAL_STATIC_AR },
         { "smart",    OPTION_VAL_SMART_AR },
         { NULL, NULL },
      },
      "smart"
   },
   {
      BEETLE_OPT(image_crop),
      BEETLE_OPT_IMAGE_CROP_LABEL_AR,
      NULL,
      BEETLE_OPT_IMAGE_CROP_INFO_0_AR,
      NULL,
      "video",
      {
         { "disabled", "0" },
         { "1px",      OPTION_VAL_1PX_AR },
         { "2px",      OPTION_VAL_2PX_AR },
         { "3px",      OPTION_VAL_3PX_AR },
         { "4px",      OPTION_VAL_4PX_AR },
         { "5px",      OPTION_VAL_5PX_AR },
         { "6px",      OPTION_VAL_6PX_AR },
         { "7px",      OPTION_VAL_7PX_AR },
         { "8px",      OPTION_VAL_8PX_AR },
         { "9px",      OPTION_VAL_9PX_AR },
         { "10px",     OPTION_VAL_10PX_AR },
         { "11px",     OPTION_VAL_11PX_AR },
         { "12px",     OPTION_VAL_12PX_AR },
         { "13px",     OPTION_VAL_13PX_AR },
         { "14px",     OPTION_VAL_14PX_AR },
         { "15px",     OPTION_VAL_15PX_AR },
         { "16px",     OPTION_VAL_16PX_AR },
         { "17px",     OPTION_VAL_17PX_AR },
         { "18px",     OPTION_VAL_18PX_AR },
         { "19px",     OPTION_VAL_19PX_AR },
         { "20px",     OPTION_VAL_20PX_AR },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(image_offset),
      BEETLE_OPT_IMAGE_OFFSET_LABEL_AR,
      NULL,
      BEETLE_OPT_IMAGE_OFFSET_INFO_0_AR,
      NULL,
      "video",
      {
         { "-12px",    OPTION_VAL_12PX_O64_AR },
         { "-11px",    OPTION_VAL_11PX_O64_AR },
         { "-10px",    OPTION_VAL_10PX_O64_AR },
         { "-9px",     OPTION_VAL_9PX_O64_AR },
         { "-8px",     OPTION_VAL_8PX_O64_AR },
         { "-7px",     OPTION_VAL_7PX_O64_AR },
         { "-6px",     OPTION_VAL_6PX_O64_AR },
         { "-5px",     OPTION_VAL_5PX_O64_AR },
         { "-4px",     OPTION_VAL_4PX_O64_AR },
         { "-3px",     OPTION_VAL_3PX_O64_AR },
         { "-2px",     OPTION_VAL_2PX_O64_AR },
         { "-1px",     OPTION_VAL_1PX_O64_AR },
         { "disabled", "0" },
         { "+1px",     OPTION_VAL_1PX_O64_0_AR },
         { "+2px",     OPTION_VAL_2PX_O64_0_AR },
         { "+3px",     OPTION_VAL_3PX_O64_0_AR },
         { "+4px",     OPTION_VAL_4PX_O64_0_AR },
         { "+5px",     OPTION_VAL_5PX_O64_0_AR },
         { "+6px",     OPTION_VAL_6PX_O64_0_AR },
         { "+7px",     OPTION_VAL_7PX_O64_0_AR },
         { "+8px",     OPTION_VAL_8PX_O64_0_AR },
         { "+9px",     OPTION_VAL_9PX_O64_0_AR },
         { "+10px",    OPTION_VAL_10PX_O64_0_AR },
         { "+11px",    OPTION_VAL_11PX_O64_0_AR },
         { "+12px",    OPTION_VAL_12PX_O64_0_AR },
         { NULL, NULL },
      },
      "disabled"
   },
#if defined(HAVE_OPENGL) || defined(HAVE_OPENGLES) || defined(HAVE_VULKAN)
   {
      BEETLE_OPT(image_offset_cycles),
      BEETLE_OPT_IMAGE_OFFSET_CYCLES_LABEL_AR,
      NULL,
      BEETLE_OPT_IMAGE_OFFSET_CYCLES_INFO_0_AR,
      NULL,
      "video",
      {
         { "-40",      NULL },
         { "-39",      NULL },
         { "-38",      NULL },
         { "-37",      NULL },
         { "-36",      NULL },
         { "-35",      NULL },
         { "-34",      NULL },
         { "-33",      NULL },
         { "-32",      NULL },
         { "-31",      NULL },
         { "-30",      NULL },
         { "-29",      NULL },
         { "-28",      NULL },
         { "-27",      NULL },
         { "-26",      NULL },
         { "-25",      NULL },
         { "-24",      NULL },
         { "-23",      NULL },
         { "-22",      NULL },
         { "-21",      NULL },
         { "-20",      NULL },
         { "-19",      NULL },
         { "-18",      NULL },
         { "-17",      NULL },
         { "-16",      NULL },
         { "-15",      NULL },
         { "-14",      NULL },
         { "-13",      NULL },
         { "-12",      NULL },
         { "-11",      NULL },
         { "-10",      NULL },
         { "-9",       NULL },
         { "-8",       NULL },
         { "-7",       NULL },
         { "-6",       NULL },
         { "-5",       NULL },
         { "-4",       NULL },
         { "-3",       NULL },
         { "-2",       NULL },
         { "-1",       NULL },
         { "0",        "0" },
         { "+1",       NULL },
         { "+2",       NULL },
         { "+3",       NULL },
         { "+4",       NULL },
         { "+5",       NULL },
         { "+6",       NULL },
         { "+7",       NULL },
         { "+8",       NULL },
         { "+9",       NULL },
         { "+10",      NULL },
         { "+11",      NULL },
         { "+12",      NULL },
         { "+13",      NULL },
         { "+14",      NULL },
         { "+15",      NULL },
         { "+16",      NULL },
         { "+17",      NULL },
         { "+18",      NULL },
         { "+19",      NULL },
         { "+20",      NULL },
         { "+21",      NULL },
         { "+22",      NULL },
         { "+23",      NULL },
         { "+24",      NULL },
         { "+25",      NULL },
         { "+26",      NULL },
         { "+27",      NULL },
         { "+28",      NULL },
         { "+29",      NULL },
         { "+30",      NULL },
         { "+31",      NULL },
         { "+32",      NULL },
         { "+33",      NULL },
         { "+34",      NULL },
         { "+35",      NULL },
         { "+36",      NULL },
         { "+37",      NULL },
         { "+38",      NULL },
         { "+39",      NULL },
         { "+40",      NULL },
         { NULL, NULL},
      },
      "0"
   },
#endif
   {
      BEETLE_OPT(initial_scanline),
      BEETLE_OPT_INITIAL_SCANLINE_LABEL_AR,
      NULL,
      BEETLE_OPT_INITIAL_SCANLINE_INFO_0_AR,
      NULL,
      "video",
      {
         { "0",  OPTION_VAL_0_O48_AR },
         { "1",  NULL },
         { "2",  NULL },
         { "3",  NULL },
         { "4",  NULL },
         { "5",  NULL },
         { "6",  NULL },
         { "7",  NULL },
         { "8",  NULL },
         { "9",  NULL },
         { "10", NULL },
         { "11", NULL },
         { "12", NULL },
         { "13", NULL },
         { "14", NULL },
         { "15", NULL },
         { "16", NULL },
         { "17", NULL },
         { "18", NULL },
         { "19", NULL },
         { "20", NULL },
         { "21", NULL },
         { "22", NULL },
         { "23", NULL },
         { "24", NULL },
         { "25", NULL },
         { "26", NULL },
         { "27", NULL },
         { "28", NULL },
         { "29", NULL },
         { "30", NULL },
         { "31", NULL },
         { "32", NULL },
         { "33", NULL },
         { "34", NULL },
         { "35", NULL },
         { "36", NULL },
         { "37", NULL },
         { "38", NULL },
         { "39", NULL },
         { "40", NULL },
         { NULL, NULL },
      },
      "0"
   },
   {
      BEETLE_OPT(last_scanline),
      BEETLE_OPT_LAST_SCANLINE_LABEL_AR,
      NULL,
      BEETLE_OPT_LAST_SCANLINE_INFO_0_AR,
      NULL,
      "video",
      {
         { "210", NULL },
         { "211", NULL },
         { "212", NULL },
         { "213", NULL },
         { "214", NULL },
         { "215", NULL },
         { "216", NULL },
         { "217", NULL },
         { "218", NULL },
         { "219", NULL },
         { "220", NULL },
         { "221", NULL },
         { "222", NULL },
         { "223", NULL },
         { "224", NULL },
         { "225", NULL },
         { "226", NULL },
         { "227", NULL },
         { "228", NULL },
         { "229", NULL },
         { "230", NULL },
         { "231", NULL },
         { "232", NULL },
         { "233", NULL },
         { "234", NULL },
         { "235", NULL },
         { "236", NULL },
         { "237", NULL },
         { "238", NULL },
         { "239", OPTION_VAL_239_AR },
         { NULL, NULL },
      },
      "239"
   },
   {
      BEETLE_OPT(initial_scanline_pal),
      BEETLE_OPT_INITIAL_SCANLINE_PAL_LABEL_AR,
      NULL,
      BEETLE_OPT_INITIAL_SCANLINE_PAL_INFO_0_AR,
      NULL,
      "video",
      {
         { "0",  OPTION_VAL_0_O48_AR },
         { "1",  NULL },
         { "2",  NULL },
         { "3",  NULL },
         { "4",  NULL },
         { "5",  NULL },
         { "6",  NULL },
         { "7",  NULL },
         { "8",  NULL },
         { "9",  NULL },
         { "10", NULL },
         { "11", NULL },
         { "12", NULL },
         { "13", NULL },
         { "14", NULL },
         { "15", NULL },
         { "16", NULL },
         { "17", NULL },
         { "18", NULL },
         { "19", NULL },
         { "20", NULL },
         { "21", NULL },
         { "22", NULL },
         { "23", NULL },
         { "24", NULL },
         { "25", NULL },
         { "26", NULL },
         { "27", NULL },
         { "28", NULL },
         { "29", NULL },
         { "30", NULL },
         { "31", NULL },
         { "32", NULL },
         { "33", NULL },
         { "34", NULL },
         { "35", NULL },
         { "36", NULL },
         { "37", NULL },
         { "38", NULL },
         { "39", NULL },
         { "40", NULL },
         { NULL, NULL },
      },
      "0"
   },
   {
      BEETLE_OPT(last_scanline_pal),
      BEETLE_OPT_LAST_SCANLINE_PAL_LABEL_AR,
      NULL,
      BEETLE_OPT_LAST_SCANLINE_PAL_INFO_0_AR,
      NULL,
      "video",
      {
         { "230", NULL },
         { "231", NULL },
         { "232", NULL },
         { "233", NULL },
         { "234", NULL },
         { "235", NULL },
         { "236", NULL },
         { "237", NULL },
         { "238", NULL },
         { "239", NULL },
         { "240", NULL },
         { "241", NULL },
         { "242", NULL },
         { "243", NULL },
         { "244", NULL },
         { "245", NULL },
         { "246", NULL },
         { "247", NULL },
         { "248", NULL },
         { "249", NULL },
         { "250", NULL },
         { "251", NULL },
         { "252", NULL },
         { "253", NULL },
         { "254", NULL },
         { "255", NULL },
         { "256", NULL },
         { "257", NULL },
         { "258", NULL },
         { "259", NULL },
         { "260", NULL },
         { "261", NULL },
         { "262", NULL },
         { "263", NULL },
         { "264", NULL },
         { "265", NULL },
         { "266", NULL },
         { "267", NULL },
         { "268", NULL },
         { "269", NULL },
         { "270", NULL },
         { "271", NULL },
         { "272", NULL },
         { "273", NULL },
         { "274", NULL },
         { "275", NULL },
         { "276", NULL },
         { "277", NULL },
         { "278", NULL },
         { "279", NULL },
         { "280", NULL },
         { "281", NULL },
         { "282", NULL },
         { "283", NULL },
         { "284", NULL },
         { "285", NULL },
         { "286", NULL },
         { "287", OPTION_VAL_287_AR },
         { NULL, NULL },
      },
      "287"
   },
   { NULL, NULL, NULL, NULL, NULL, NULL, {{0}}, NULL },
};
struct retro_core_options_v2 options_ar = {
   option_cats_ar,
   option_defs_ar
};

/* RETRO_LANGUAGE_AST */

#define CATEGORY_SYSTEM_LABEL_AST "Sistema"
#define CATEGORY_SYSTEM_INFO_0_AST NULL
#define CATEGORY_VIDEO_LABEL_AST "Videu"
#define CATEGORY_VIDEO_INFO_0_AST NULL
#define CATEGORY_PGXP_LABEL_AST NULL
#define CATEGORY_PGXP_INFO_0_AST NULL
#define CATEGORY_OSD_LABEL_AST NULL
#define CATEGORY_OSD_INFO_0_AST NULL
#define CATEGORY_INPUT_LABEL_AST "Controles"
#define CATEGORY_INPUT_INFO_0_AST NULL
#define CATEGORY_MEMCARDS_LABEL_AST "Tarxeta de memoria"
#define CATEGORY_MEMCARDS_INFO_0_AST NULL
#define CATEGORY_HACKS_LABEL_AST "Hacks de la emulación"
#define CATEGORY_HACKS_INFO_0_AST NULL
#define BEETLE_OPT_CPU_FREQ_SCALE_LABEL_AST NULL
#define BEETLE_OPT_CPU_FREQ_SCALE_INFO_0_AST NULL
#define OPTION_VAL_50_AST NULL
#define OPTION_VAL_60_AST NULL
#define OPTION_VAL_70_AST NULL
#define OPTION_VAL_80_AST NULL
#define OPTION_VAL_90_AST NULL
#define OPTION_VAL_100_AST NULL
#define OPTION_VAL_110_AST NULL
#define OPTION_VAL_120_AST NULL
#define OPTION_VAL_130_AST NULL
#define OPTION_VAL_140_AST NULL
#define OPTION_VAL_150_AST NULL
#define OPTION_VAL_160_AST NULL
#define OPTION_VAL_170_AST NULL
#define OPTION_VAL_180_AST NULL
#define OPTION_VAL_190_AST NULL
#define OPTION_VAL_200_AST NULL
#define OPTION_VAL_210_AST NULL
#define OPTION_VAL_220_AST NULL
#define OPTION_VAL_230_AST NULL
#define OPTION_VAL_240_AST NULL
#define OPTION_VAL_250_AST NULL
#define OPTION_VAL_260_AST NULL
#define OPTION_VAL_270_AST NULL
#define OPTION_VAL_280_AST NULL
#define OPTION_VAL_290_AST NULL
#define OPTION_VAL_300_AST NULL
#define OPTION_VAL_310_AST NULL
#define OPTION_VAL_320_AST NULL
#define OPTION_VAL_330_AST NULL
#define OPTION_VAL_340_AST NULL
#define OPTION_VAL_350_AST NULL
#define OPTION_VAL_360_AST NULL
#define OPTION_VAL_370_AST NULL
#define OPTION_VAL_380_AST NULL
#define OPTION_VAL_390_AST NULL
#define OPTION_VAL_400_AST NULL
#define OPTION_VAL_410_AST NULL
#define OPTION_VAL_420_AST NULL
#define OPTION_VAL_430_AST NULL
#define OPTION_VAL_440_AST NULL
#define OPTION_VAL_450_AST NULL
#define OPTION_VAL_460_AST NULL
#define OPTION_VAL_470_AST NULL
#define OPTION_VAL_480_AST NULL
#define OPTION_VAL_490_AST NULL
#define OPTION_VAL_500_AST NULL
#define OPTION_VAL_510_AST NULL
#define OPTION_VAL_520_AST NULL
#define OPTION_VAL_530_AST NULL
#define OPTION_VAL_540_AST NULL
#define OPTION_VAL_550_AST NULL
#define OPTION_VAL_560_AST NULL
#define OPTION_VAL_570_AST NULL
#define OPTION_VAL_580_AST NULL
#define OPTION_VAL_590_AST NULL
#define OPTION_VAL_600_AST NULL
#define OPTION_VAL_610_AST NULL
#define OPTION_VAL_620_AST NULL
#define OPTION_VAL_630_AST NULL
#define OPTION_VAL_640_AST NULL
#define OPTION_VAL_650_AST NULL
#define OPTION_VAL_660_AST NULL
#define OPTION_VAL_670_AST NULL
#define OPTION_VAL_680_AST NULL
#define OPTION_VAL_690_AST NULL
#define OPTION_VAL_700_AST NULL
#define OPTION_VAL_710_AST NULL
#define OPTION_VAL_720_AST NULL
#define OPTION_VAL_730_AST NULL
#define OPTION_VAL_740_AST NULL
#define OPTION_VAL_750_AST NULL
#define BEETLE_OPT_GPU_OVERCLOCK_LABEL_AST NULL
#define BEETLE_OPT_GPU_OVERCLOCK_INFO_0_AST NULL
#define OPTION_VAL_1X_NATIVE_AST NULL
#define OPTION_VAL_2X_AST "x2"
#define OPTION_VAL_4X_AST "x4"
#define OPTION_VAL_8X_AST "x8"
#define OPTION_VAL_16X_AST "x16"
#define OPTION_VAL_32X_AST "x32"
#define BEETLE_OPT_GTE_OVERCLOCK_LABEL_AST NULL
#define BEETLE_OPT_GTE_OVERCLOCK_INFO_0_AST NULL
#define BEETLE_OPT_CD_ACCESS_METHOD_LABEL_AST NULL
#define BEETLE_OPT_CD_ACCESS_METHOD_INFO_0_AST NULL
#define OPTION_VAL_SYNC_AST "Síncronu"
#define OPTION_VAL_ASYNC_AST "Asíncronu"
#define OPTION_VAL_PRECACHE_AST NULL
#define BEETLE_OPT_CD_FASTLOAD_LABEL_AST NULL
#define BEETLE_OPT_CD_FASTLOAD_INFO_0_AST NULL
#define OPTION_VAL_2X_NATIVE_AST NULL
#define OPTION_VAL_6X_AST "x6"
#define OPTION_VAL_10X_AST "x10"
#define OPTION_VAL_12X_AST "x12"
#define OPTION_VAL_14X_AST "x14"
#define BEETLE_OPT_SKIP_BIOS_LABEL_AST NULL
#define BEETLE_OPT_SKIP_BIOS_INFO_0_AST "Salta l'animación d'arrique de la BIOS de PlayStation que suel apaecer al cargar conteníu. Nota: l'activación d'esta opción produz problemes de compatibilidá con dalgunos de xuegos (xuegos PAL con proteición anticopia, Saga Frontier, etc.)."
#define BEETLE_OPT_OVERRIDE_BIOS_LABEL_AST NULL
#define BEETLE_OPT_OVERRIDE_BIOS_INFO_0_AST NULL
#define OPTION_VAL_PSXONPSP_AST "BIOS de PS1 de la PSP"
#define OPTION_VAL_PS1_ROM_AST "BIOS de PS1 de la PS3"
#define OPTION_VAL_OPENBIOS_AST NULL
#define BEETLE_OPT_REGION_LABEL_AST "Rexón del sistema"
#define BEETLE_OPT_REGION_INFO_0_AST NULL
#define OPTION_VAL_AUTO_AST NULL
#define OPTION_VAL_NTSC_J_AST NULL
#define OPTION_VAL_NTSC_U_AST NULL
#define OPTION_VAL_PAL_AST NULL
#define BEETLE_OPT_INTERNAL_RESOLUTION_LABEL_AST NULL
#define BEETLE_OPT_INTERNAL_RESOLUTION_INFO_0_AST NULL
#define BEETLE_OPT_RENDERER_LABEL_AST NULL
#define BEETLE_OPT_RENDERER_INFO_0_AST NULL
#define OPTION_VAL_HARDWARE_AST NULL
#define OPTION_VAL_HARDWARE_GL_AST NULL
#define OPTION_VAL_HARDWARE_VK_AST NULL
#define OPTION_VAL_SOFTWARE_AST NULL
#define BEETLE_OPT_RENDERER_SOFTWARE_FB_LABEL_AST NULL
#define BEETLE_OPT_RENDERER_SOFTWARE_FB_INFO_0_AST NULL
#define BEETLE_OPT_DEPTH_LABEL_AST NULL
#define BEETLE_OPT_DEPTH_INFO_0_AST NULL
#define OPTION_VAL_16BPP_NATIVE_AST NULL
#define OPTION_VAL_32BPP_AST NULL
#define BEETLE_OPT_DITHER_MODE_LABEL_AST NULL
#define BEETLE_OPT_DITHER_MODE_INFO_0_AST NULL
#define OPTION_VAL_INTERNAL_RESOLUTION_AST "Resolución interna"
#define BEETLE_OPT_SCALED_UV_OFFSET_LABEL_AST NULL
#define BEETLE_OPT_SCALED_UV_OFFSET_INFO_0_AST NULL
#define BEETLE_OPT_FILTER_LABEL_AST NULL
#define BEETLE_OPT_FILTER_INFO_0_AST NULL
#define OPTION_VAL_NEAREST_AST NULL
#define OPTION_VAL_SABR_AST NULL
#define OPTION_VAL_XBR_AST NULL
#define OPTION_VAL_BILINEAR_AST "Billinial"
#define OPTION_VAL_3_POINT_AST NULL
#define OPTION_VAL_JINC2_AST NULL
#define BEETLE_OPT_FILTER_EXCLUDE_SPRITE_LABEL_AST NULL
#define BEETLE_OPT_FILTER_EXCLUDE_SPRITE_INFO_0_AST NULL
#define OPTION_VAL_OPAQUE_AST NULL
#define OPTION_VAL_ALL_AST NULL
#define BEETLE_OPT_FILTER_EXCLUDE_2D_POLYGON_LABEL_AST NULL
#define BEETLE_OPT_FILTER_EXCLUDE_2D_POLYGON_INFO_0_AST NULL
#define BEETLE_OPT_ADAPTIVE_SMOOTHING_LABEL_AST NULL
#define BEETLE_OPT_ADAPTIVE_SMOOTHING_INFO_0_AST NULL
#define BEETLE_OPT_SUPER_SAMPLING_LABEL_AST NULL
#define BEETLE_OPT_SUPER_SAMPLING_INFO_0_AST NULL
#define BEETLE_OPT_MSAA_LABEL_AST NULL
#define BEETLE_OPT_MSAA_INFO_0_AST NULL
#define OPTION_VAL_1X_AST "x1"
#define BEETLE_OPT_MDEC_YUV_LABEL_AST NULL
#define BEETLE_OPT_MDEC_YUV_INFO_0_AST NULL
#define BEETLE_OPT_TRACK_TEXTURES_LABEL_AST NULL
#define BEETLE_OPT_TRACK_TEXTURES_INFO_0_AST NULL
#define BEETLE_OPT_DUMP_TEXTURES_LABEL_AST NULL
#define BEETLE_OPT_DUMP_TEXTURES_INFO_0_AST NULL
#define BEETLE_OPT_REPLACE_TEXTURES_LABEL_AST NULL
#define BEETLE_OPT_REPLACE_TEXTURES_INFO_0_AST NULL
#define BEETLE_OPT_PGXP_MODE_LABEL_AST NULL
#define BEETLE_OPT_PGXP_MODE_INFO_0_AST NULL
#define OPTION_VAL_MEMORY_ONLY_AST NULL
#define OPTION_VAL_MEMORY_CPU_AST "Memoria + CPU (produz fallos)"
#define BEETLE_OPT_PGXP_2D_TOL_LABEL_AST NULL
#define BEETLE_OPT_PGXP_2D_TOL_INFO_0_AST NULL
#define OPTION_VAL_0PX_AST NULL
#define OPTION_VAL_1PX_AST NULL
#define OPTION_VAL_2PX_AST NULL
#define OPTION_VAL_3PX_AST NULL
#define OPTION_VAL_4PX_AST NULL
#define OPTION_VAL_5PX_AST NULL
#define OPTION_VAL_6PX_AST NULL
#define OPTION_VAL_7PX_AST NULL
#define OPTION_VAL_8PX_AST NULL
#define BEETLE_OPT_PGXP_NCLIP_LABEL_AST NULL
#define BEETLE_OPT_PGXP_NCLIP_INFO_0_AST NULL
#define BEETLE_OPT_PGXP_VERTEX_LABEL_AST NULL
#define BEETLE_OPT_PGXP_VERTEX_INFO_0_AST NULL
#define BEETLE_OPT_PGXP_TEXTURE_LABEL_AST NULL
#define BEETLE_OPT_PGXP_TEXTURE_INFO_0_AST NULL
#define BEETLE_OPT_DISPLAY_INTERNAL_FPS_LABEL_AST NULL
#define BEETLE_OPT_DISPLAY_INTERNAL_FPS_INFO_0_AST NULL
#define BEETLE_OPT_DISPLAY_OSD_LABEL_AST NULL
#define BEETLE_OPT_DISPLAY_OSD_INFO_0_AST NULL
#define BEETLE_OPT_DISPLAY_VRAM_LABEL_AST NULL
#define BEETLE_OPT_DISPLAY_VRAM_INFO_0_AST NULL
#define BEETLE_OPT_ANALOG_CALIBRATION_LABEL_AST NULL
#define BEETLE_OPT_ANALOG_CALIBRATION_INFO_0_AST NULL
#define BEETLE_OPT_ANALOG_TOGGLE_LABEL_AST NULL
#define BEETLE_OPT_ANALOG_TOGGLE_INFO_0_AST NULL
#define OPTION_VAL_ENABLED_ANALOG_AST NULL
#define BEETLE_OPT_ANALOG_TOGGLE_COMBO_LABEL_AST NULL
#define BEETLE_OPT_ANALOG_TOGGLE_COMBO_INFO_0_AST NULL
#define OPTION_VAL_L1_L2_R1_R2_START_SELECT_AST NULL
#define OPTION_VAL_L1_R1_SELECT_AST NULL
#define OPTION_VAL_L1_R1_START_AST NULL
#define OPTION_VAL_L1_R1_L3_AST NULL
#define OPTION_VAL_L1_R1_R3_AST NULL
#define OPTION_VAL_L2_R2_SELECT_AST NULL
#define OPTION_VAL_L2_R2_START_AST NULL
#define OPTION_VAL_L2_R2_L3_AST NULL
#define OPTION_VAL_L2_R2_R3_AST NULL
#define OPTION_VAL_L3_R3_AST NULL
#define BEETLE_OPT_ANALOG_TOGGLE_HOLD_LABEL_AST NULL
#define BEETLE_OPT_ANALOG_TOGGLE_HOLD_INFO_0_AST NULL
#define OPTION_VAL_0_AST NULL
#define OPTION_VAL_1_AST NULL
#define OPTION_VAL_2_AST NULL
#define OPTION_VAL_3_AST NULL
#define OPTION_VAL_4_AST NULL
#define OPTION_VAL_5_AST NULL
#define BEETLE_OPT_ENABLE_MULTITAP_PORT1_LABEL_AST NULL
#define BEETLE_OPT_ENABLE_MULTITAP_PORT1_INFO_0_AST NULL
#define BEETLE_OPT_ENABLE_MULTITAP_PORT2_LABEL_AST NULL
#define BEETLE_OPT_ENABLE_MULTITAP_PORT2_INFO_0_AST NULL
#define BEETLE_OPT_GUN_INPUT_MODE_LABEL_AST NULL
#define BEETLE_OPT_GUN_INPUT_MODE_INFO_0_AST NULL
#define OPTION_VAL_LIGHTGUN_AST NULL
#define OPTION_VAL_TOUCHSCREEN_AST "Pantalla táctil"
#define BEETLE_OPT_GUN_CURSOR_LABEL_AST NULL
#define BEETLE_OPT_GUN_CURSOR_INFO_0_AST NULL
#define OPTION_VAL_CROSS_AST NULL
#define OPTION_VAL_DOT_AST NULL
#define BEETLE_OPT_CROSSHAIR_COLOR_P1_LABEL_AST NULL
#define BEETLE_OPT_CROSSHAIR_COLOR_P1_INFO_0_AST NULL
#define OPTION_VAL_RED_AST "Coloráu"
#define OPTION_VAL_BLUE_AST "Azul"
#define OPTION_VAL_GREEN_AST "Verde"
#define OPTION_VAL_ORANGE_AST "Naranxa"
#define OPTION_VAL_YELLOW_AST "Mariellu"
#define OPTION_VAL_CYAN_AST "Cianu"
#define OPTION_VAL_PINK_AST "Rosa"
#define OPTION_VAL_PURPLE_AST "Moráu"
#define OPTION_VAL_BLACK_AST "Prietu"
#define OPTION_VAL_WHITE_AST "Blancu"
#define BEETLE_OPT_CROSSHAIR_COLOR_P2_LABEL_AST NULL
#define BEETLE_OPT_CROSSHAIR_COLOR_P2_INFO_0_AST NULL
#define BEETLE_OPT_MOUSE_SENSITIVITY_LABEL_AST "Sensibilidá del mur"
#define BEETLE_OPT_MOUSE_SENSITIVITY_INFO_0_AST NULL
#define OPTION_VAL_5_O42_AST NULL
#define OPTION_VAL_10_AST NULL
#define OPTION_VAL_15_AST NULL
#define OPTION_VAL_20_AST NULL
#define OPTION_VAL_25_AST NULL
#define OPTION_VAL_30_AST NULL
#define OPTION_VAL_35_AST NULL
#define OPTION_VAL_40_AST NULL
#define OPTION_VAL_45_AST NULL
#define OPTION_VAL_55_AST NULL
#define OPTION_VAL_65_AST NULL
#define OPTION_VAL_75_AST NULL
#define OPTION_VAL_85_AST NULL
#define OPTION_VAL_95_AST NULL
#define OPTION_VAL_100_O42_AST NULL
#define OPTION_VAL_105_AST NULL
#define OPTION_VAL_115_AST NULL
#define OPTION_VAL_125_AST NULL
#define OPTION_VAL_135_AST NULL
#define OPTION_VAL_145_AST NULL
#define OPTION_VAL_155_AST NULL
#define OPTION_VAL_165_AST NULL
#define OPTION_VAL_175_AST NULL
#define OPTION_VAL_185_AST NULL
#define OPTION_VAL_195_AST NULL
#define BEETLE_OPT_NEGCON_RESPONSE_LABEL_AST NULL
#define BEETLE_OPT_NEGCON_RESPONSE_INFO_0_AST NULL
#define OPTION_VAL_LINEAR_AST NULL
#define OPTION_VAL_QUADRATIC_AST NULL
#define OPTION_VAL_CUBIC_AST NULL
#define BEETLE_OPT_NEGCON_DEADZONE_LABEL_AST NULL
#define BEETLE_OPT_NEGCON_DEADZONE_INFO_0_AST NULL
#define OPTION_VAL_0_O44_AST NULL
#define BEETLE_OPT_USE_MEDNAFEN_MEMCARD0_METHOD_LABEL_AST NULL
#define BEETLE_OPT_USE_MEDNAFEN_MEMCARD0_METHOD_INFO_0_AST NULL
#define OPTION_VAL_LIBRETRO_AST NULL
#define OPTION_VAL_MEDNAFEN_AST NULL
#define BEETLE_OPT_ENABLE_MEMCARD1_LABEL_AST NULL
#define BEETLE_OPT_ENABLE_MEMCARD1_INFO_0_AST NULL
#define BEETLE_OPT_SHARED_MEMORY_CARDS_LABEL_AST NULL
#define BEETLE_OPT_SHARED_MEMORY_CARDS_INFO_0_AST NULL
#define BEETLE_OPT_MEMCARD_LEFT_INDEX_LABEL_AST NULL
#define BEETLE_OPT_MEMCARD_LEFT_INDEX_INFO_0_AST NULL
#define OPTION_VAL_0_O48_AST NULL
#define BEETLE_OPT_MEMCARD_RIGHT_INDEX_LABEL_AST NULL
#define BEETLE_OPT_MEMCARD_RIGHT_INDEX_INFO_0_AST NULL
#define OPTION_VAL_1_O49_AST NULL
#define BEETLE_OPT_LINE_RENDER_LABEL_AST NULL
#define BEETLE_OPT_LINE_RENDER_INFO_0_AST NULL
#define OPTION_VAL_DEFAULT_AST NULL
#define OPTION_VAL_AGGRESSIVE_AST "Agresivu"
#define BEETLE_OPT_WIDESCREEN_HACK_LABEL_AST NULL
#define BEETLE_OPT_WIDESCREEN_HACK_INFO_0_AST NULL
#define BEETLE_OPT_WIDESCREEN_HACK_ASPECT_RATIO_LABEL_AST NULL
#define BEETLE_OPT_WIDESCREEN_HACK_ASPECT_RATIO_INFO_0_AST NULL
#define OPTION_VAL_16_9_AST NULL
#define OPTION_VAL_16_10_AST NULL
#define OPTION_VAL_18_9_AST NULL
#define OPTION_VAL_19_9_AST NULL
#define OPTION_VAL_20_9_AST NULL
#define OPTION_VAL_21_9_AST NULL
#define OPTION_VAL_32_9_AST NULL
#define BEETLE_OPT_CPU_DYNAREC_LABEL_AST NULL
#define BEETLE_OPT_CPU_DYNAREC_INFO_0_AST NULL
#define OPTION_VAL_DISABLED_AST NULL
#define OPTION_VAL_EXECUTE_AST NULL
#define OPTION_VAL_EXECUTE_ONE_AST NULL
#define OPTION_VAL_RUN_INTERPRETER_AST NULL
#define BEETLE_OPT_DYNAREC_INVALIDATE_LABEL_AST NULL
#define BEETLE_OPT_DYNAREC_INVALIDATE_INFO_0_AST NULL
#define OPTION_VAL_FULL_AST NULL
#define OPTION_VAL_DMA_AST NULL
#define BEETLE_OPT_DYNAREC_EVENTCYCLES_LABEL_AST NULL
#define BEETLE_OPT_DYNAREC_EVENTCYCLES_INFO_0_AST NULL
#define OPTION_VAL_128_AST NULL
#define BEETLE_OPT_DYNAREC_SPU_SAMPLES_LABEL_AST NULL
#define BEETLE_OPT_DYNAREC_SPU_SAMPLES_INFO_0_AST NULL
#define BEETLE_OPT_CORE_TIMING_FPS_LABEL_AST NULL
#define BEETLE_OPT_CORE_TIMING_FPS_INFO_0_AST NULL
#define OPTION_VAL_AUTO_TOGGLE_AST NULL
#define OPTION_VAL_FORCE_PROGRESSIVE_AST NULL
#define OPTION_VAL_FORCE_INTERLACED_AST NULL
#define BEETLE_OPT_DEINTERLACER_LABEL_AST NULL
#define BEETLE_OPT_DEINTERLACER_INFO_0_AST NULL
#define OPTION_VAL_WEAVE_AST NULL
#define OPTION_VAL_BOB_AST NULL
#define OPTION_VAL_BOB_OFFSET_AST NULL
#define OPTION_VAL_FASTMAD_AST NULL
#define BEETLE_OPT_FRAME_DUPING_LABEL_AST NULL
#define BEETLE_OPT_FRAME_DUPING_INFO_0_AST NULL
#define BEETLE_OPT_PAL_VIDEO_TIMING_OVERRIDE_LABEL_AST NULL
#define BEETLE_OPT_PAL_VIDEO_TIMING_OVERRIDE_INFO_0_AST NULL
#define BEETLE_OPT_ASPECT_RATIO_LABEL_AST "Proporción d'aspeutu del nucleu"
#define BEETLE_OPT_ASPECT_RATIO_INFO_0_AST NULL
#define OPTION_VAL_CORRECTED_AST NULL
#define OPTION_VAL_UNCORRECTED_AST NULL
#define OPTION_VAL_4_3_AST NULL
#define OPTION_VAL_NTSC_AST NULL
#define BEETLE_OPT_CROP_OVERSCAN_LABEL_AST NULL
#define BEETLE_OPT_CROP_OVERSCAN_INFO_0_AST NULL
#define OPTION_VAL_STATIC_AST NULL
#define OPTION_VAL_SMART_AST NULL
#define BEETLE_OPT_IMAGE_CROP_LABEL_AST "Recorte adicional"
#define BEETLE_OPT_IMAGE_CROP_INFO_0_AST NULL
#define OPTION_VAL_9PX_AST NULL
#define OPTION_VAL_10PX_AST NULL
#define OPTION_VAL_11PX_AST NULL
#define OPTION_VAL_12PX_AST NULL
#define OPTION_VAL_13PX_AST NULL
#define OPTION_VAL_14PX_AST NULL
#define OPTION_VAL_15PX_AST NULL
#define OPTION_VAL_16PX_AST NULL
#define OPTION_VAL_17PX_AST NULL
#define OPTION_VAL_18PX_AST NULL
#define OPTION_VAL_19PX_AST NULL
#define OPTION_VAL_20PX_AST NULL
#define BEETLE_OPT_IMAGE_OFFSET_LABEL_AST NULL
#define BEETLE_OPT_IMAGE_OFFSET_INFO_0_AST NULL
#define OPTION_VAL_12PX_O64_AST NULL
#define OPTION_VAL_11PX_O64_AST NULL
#define OPTION_VAL_10PX_O64_AST NULL
#define OPTION_VAL_9PX_O64_AST NULL
#define OPTION_VAL_8PX_O64_AST NULL
#define OPTION_VAL_7PX_O64_AST NULL
#define OPTION_VAL_6PX_O64_AST NULL
#define OPTION_VAL_5PX_O64_AST NULL
#define OPTION_VAL_4PX_O64_AST NULL
#define OPTION_VAL_3PX_O64_AST NULL
#define OPTION_VAL_2PX_O64_AST NULL
#define OPTION_VAL_1PX_O64_AST NULL
#define OPTION_VAL_1PX_O64_0_AST NULL
#define OPTION_VAL_2PX_O64_0_AST NULL
#define OPTION_VAL_3PX_O64_0_AST NULL
#define OPTION_VAL_4PX_O64_0_AST NULL
#define OPTION_VAL_5PX_O64_0_AST NULL
#define OPTION_VAL_6PX_O64_0_AST NULL
#define OPTION_VAL_7PX_O64_0_AST NULL
#define OPTION_VAL_8PX_O64_0_AST NULL
#define OPTION_VAL_9PX_O64_0_AST NULL
#define OPTION_VAL_10PX_O64_0_AST NULL
#define OPTION_VAL_11PX_O64_0_AST NULL
#define OPTION_VAL_12PX_O64_0_AST NULL
#define BEETLE_OPT_IMAGE_OFFSET_CYCLES_LABEL_AST NULL
#define BEETLE_OPT_IMAGE_OFFSET_CYCLES_INFO_0_AST NULL
#define BEETLE_OPT_INITIAL_SCANLINE_LABEL_AST NULL
#define BEETLE_OPT_INITIAL_SCANLINE_INFO_0_AST NULL
#define BEETLE_OPT_LAST_SCANLINE_LABEL_AST NULL
#define BEETLE_OPT_LAST_SCANLINE_INFO_0_AST NULL
#define OPTION_VAL_239_AST "239 (por defeutu)"
#define BEETLE_OPT_INITIAL_SCANLINE_PAL_LABEL_AST NULL
#define BEETLE_OPT_INITIAL_SCANLINE_PAL_INFO_0_AST NULL
#define BEETLE_OPT_LAST_SCANLINE_PAL_LABEL_AST NULL
#define BEETLE_OPT_LAST_SCANLINE_PAL_INFO_0_AST NULL
#define OPTION_VAL_287_AST "287 (por defeutu)"

struct retro_core_option_v2_category option_cats_ast[] = {
   {
      "system",
      CATEGORY_SYSTEM_LABEL_AST,
      CATEGORY_SYSTEM_INFO_0_AST
   },
   {
      "video",
      CATEGORY_VIDEO_LABEL_AST,
      CATEGORY_VIDEO_INFO_0_AST
   },
   {
      "pgxp",
      CATEGORY_PGXP_LABEL_AST,
      CATEGORY_PGXP_INFO_0_AST
   },
   {
      "osd",
      CATEGORY_OSD_LABEL_AST,
      CATEGORY_OSD_INFO_0_AST
   },
   {
      "input",
      CATEGORY_INPUT_LABEL_AST,
      CATEGORY_INPUT_INFO_0_AST
   },
   {
      "memcards",
      CATEGORY_MEMCARDS_LABEL_AST,
      CATEGORY_MEMCARDS_INFO_0_AST
   },
   {
      "hacks",
      CATEGORY_HACKS_LABEL_AST,
      CATEGORY_HACKS_INFO_0_AST
   },
   { NULL, NULL, NULL },
};
struct retro_core_option_v2_definition option_defs_ast[] = {
   {
      BEETLE_OPT(cpu_freq_scale),
      BEETLE_OPT_CPU_FREQ_SCALE_LABEL_AST,
      NULL,
      BEETLE_OPT_CPU_FREQ_SCALE_INFO_0_AST,
      NULL,
      "system",
      {
         { "50%",           OPTION_VAL_50_AST },
         { "60%",           OPTION_VAL_60_AST },
         { "70%",           OPTION_VAL_70_AST },
         { "80%",           OPTION_VAL_80_AST },
         { "90%",           OPTION_VAL_90_AST },
         { "100%",          OPTION_VAL_100_AST },
         { "110%",          OPTION_VAL_110_AST },
         { "120%",          OPTION_VAL_120_AST },
         { "130%",          OPTION_VAL_130_AST },
         { "140%",          OPTION_VAL_140_AST },
         { "150%",          OPTION_VAL_150_AST },
         { "160%",          OPTION_VAL_160_AST },
         { "170%",          OPTION_VAL_170_AST },
         { "180%",          OPTION_VAL_180_AST },
         { "190%",          OPTION_VAL_190_AST },
         { "200%",          OPTION_VAL_200_AST },
         { "210%",          OPTION_VAL_210_AST },
         { "220%",          OPTION_VAL_220_AST },
         { "230%",          OPTION_VAL_230_AST },
         { "240%",          OPTION_VAL_240_AST },
         { "250%",          OPTION_VAL_250_AST },
         { "260%",          OPTION_VAL_260_AST },
         { "270%",          OPTION_VAL_270_AST },
         { "280%",          OPTION_VAL_280_AST },
         { "290%",          OPTION_VAL_290_AST },
         { "300%",          OPTION_VAL_300_AST },
         { "310%",          OPTION_VAL_310_AST },
         { "320%",          OPTION_VAL_320_AST },
         { "330%",          OPTION_VAL_330_AST },
         { "340%",          OPTION_VAL_340_AST },
         { "350%",          OPTION_VAL_350_AST },
         { "360%",          OPTION_VAL_360_AST },
         { "370%",          OPTION_VAL_370_AST },
         { "380%",          OPTION_VAL_380_AST },
         { "390%",          OPTION_VAL_390_AST },
         { "400%",          OPTION_VAL_400_AST },
         { "410%",          OPTION_VAL_410_AST },
         { "420%",          OPTION_VAL_420_AST },
         { "430%",          OPTION_VAL_430_AST },
         { "440%",          OPTION_VAL_440_AST },
         { "450%",          OPTION_VAL_450_AST },
         { "460%",          OPTION_VAL_460_AST },
         { "470%",          OPTION_VAL_470_AST },
         { "480%",          OPTION_VAL_480_AST },
         { "490%",          OPTION_VAL_490_AST },
         { "500%",          OPTION_VAL_500_AST },
         { "510%",          OPTION_VAL_510_AST },
         { "520%",          OPTION_VAL_520_AST },
         { "530%",          OPTION_VAL_530_AST },
         { "540%",          OPTION_VAL_540_AST },
         { "550%",          OPTION_VAL_550_AST },
         { "560%",          OPTION_VAL_560_AST },
         { "570%",          OPTION_VAL_570_AST },
         { "580%",          OPTION_VAL_580_AST },
         { "590%",          OPTION_VAL_590_AST },
         { "600%",          OPTION_VAL_600_AST },
         { "610%",          OPTION_VAL_610_AST },
         { "620%",          OPTION_VAL_620_AST },
         { "630%",          OPTION_VAL_630_AST },
         { "640%",          OPTION_VAL_640_AST },
         { "650%",          OPTION_VAL_650_AST },
         { "660%",          OPTION_VAL_660_AST },
         { "670%",          OPTION_VAL_670_AST },
         { "680%",          OPTION_VAL_680_AST },
         { "690%",          OPTION_VAL_690_AST },
         { "700%",          OPTION_VAL_700_AST },
         { "710%",          OPTION_VAL_710_AST },
         { "720%",          OPTION_VAL_720_AST },
         { "730%",          OPTION_VAL_730_AST },
         { "740%",          OPTION_VAL_740_AST },
         { "750%",          OPTION_VAL_750_AST },
         { NULL, NULL },
      },
      "100%"
   },
   {
      BEETLE_OPT(gpu_overclock),
      BEETLE_OPT_GPU_OVERCLOCK_LABEL_AST,
      NULL,
      BEETLE_OPT_GPU_OVERCLOCK_INFO_0_AST,
      NULL,
      "system",
      {
         { "1x(native)", OPTION_VAL_1X_NATIVE_AST },
         { "2x",         OPTION_VAL_2X_AST },
         { "4x",         OPTION_VAL_4X_AST },
         { "8x",         OPTION_VAL_8X_AST },
         { "16x",        OPTION_VAL_16X_AST },
         { "32x",        OPTION_VAL_32X_AST },
         { NULL, NULL },
      },
      "1x(native)"
   },
   {
      BEETLE_OPT(gte_overclock),
      BEETLE_OPT_GTE_OVERCLOCK_LABEL_AST,
      NULL,
      BEETLE_OPT_GTE_OVERCLOCK_INFO_0_AST,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
#ifndef EMSCRIPTEN
   {
      BEETLE_OPT(cd_access_method),
      BEETLE_OPT_CD_ACCESS_METHOD_LABEL_AST,
      NULL,
      BEETLE_OPT_CD_ACCESS_METHOD_INFO_0_AST,
      NULL,
      "system",
      {
         { "sync",     OPTION_VAL_SYNC_AST },
         { "async",    OPTION_VAL_ASYNC_AST },
         { "precache", OPTION_VAL_PRECACHE_AST },
         { NULL, NULL },
      },
      "sync"
   },
#endif
   {
      BEETLE_OPT(cd_fastload),
      BEETLE_OPT_CD_FASTLOAD_LABEL_AST,
      NULL,
      BEETLE_OPT_CD_FASTLOAD_INFO_0_AST,
      NULL,
      "system",
      {
         { "2x(native)", OPTION_VAL_2X_NATIVE_AST },
         { "4x",          OPTION_VAL_4X_AST },
         { "6x",          OPTION_VAL_6X_AST },
         { "8x",          OPTION_VAL_8X_AST },
         { "10x",         OPTION_VAL_10X_AST },
         { "12x",         OPTION_VAL_12X_AST },
         { "14x",         OPTION_VAL_14X_AST },
         { NULL, NULL },
      },
      "2x(native)"
   },
   {
      BEETLE_OPT(skip_bios),
      BEETLE_OPT_SKIP_BIOS_LABEL_AST,
      NULL,
      BEETLE_OPT_SKIP_BIOS_INFO_0_AST,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(override_bios),
      BEETLE_OPT_OVERRIDE_BIOS_LABEL_AST,
      NULL,
      BEETLE_OPT_OVERRIDE_BIOS_INFO_0_AST,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "psxonpsp", OPTION_VAL_PSXONPSP_AST },
         { "ps1_rom",  OPTION_VAL_PS1_ROM_AST },
         { "openbios", OPTION_VAL_OPENBIOS_AST     },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(region),
      BEETLE_OPT_REGION_LABEL_AST,
      NULL,
      BEETLE_OPT_REGION_INFO_0_AST,
      NULL,
      "system",
      {
         { "auto",   OPTION_VAL_AUTO_AST },
         { "ntsc-j", OPTION_VAL_NTSC_J_AST },
         { "ntsc-u", OPTION_VAL_NTSC_U_AST },
         { "pal",    OPTION_VAL_PAL_AST },
         { NULL, NULL },
      },
      "auto"
   },
   {
      BEETLE_OPT(internal_resolution),
      BEETLE_OPT_INTERNAL_RESOLUTION_LABEL_AST,
      NULL,
      BEETLE_OPT_INTERNAL_RESOLUTION_INFO_0_AST,
      NULL,
      "video",
      {
         { "1x(native)", OPTION_VAL_1X_NATIVE_AST },
         { "2x",         OPTION_VAL_2X_AST },
         { "4x",         OPTION_VAL_4X_AST },
         { "8x",         OPTION_VAL_8X_AST },
         { "16x",        OPTION_VAL_16X_AST },
         { NULL, NULL },
      },
      "1x(native)"
   },
#if defined(HAVE_OPENGL) || defined(HAVE_OPENGLES) || defined(HAVE_VULKAN)
   {
      BEETLE_OPT(renderer),
      BEETLE_OPT_RENDERER_LABEL_AST,
      NULL,
      BEETLE_OPT_RENDERER_INFO_0_AST,
      NULL,
      "video",
      {
         { "hardware",    OPTION_VAL_HARDWARE_AST },
#if defined(HAVE_OPENGL) || defined(HAVE_OPENGLES)
         { "hardware_gl", OPTION_VAL_HARDWARE_GL_AST },
#endif
#if defined(HAVE_VULKAN)
         { "hardware_vk", OPTION_VAL_HARDWARE_VK_AST },
#endif
         { "software",    OPTION_VAL_SOFTWARE_AST },
         { NULL, NULL },
      },
      "hardware"
   },
   {
      BEETLE_OPT(renderer_software_fb),
      BEETLE_OPT_RENDERER_SOFTWARE_FB_LABEL_AST,
      NULL,
      BEETLE_OPT_RENDERER_SOFTWARE_FB_INFO_0_AST,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled"
   },
#endif
#if defined(HAVE_OPENGL) || defined(HAVE_OPENGLES)
   {
      BEETLE_OPT(depth),
      BEETLE_OPT_DEPTH_LABEL_AST,
      NULL,
      BEETLE_OPT_DEPTH_INFO_0_AST,
      NULL,
      "video",
      {
         { "16bpp(native)", OPTION_VAL_16BPP_NATIVE_AST },
         { "32bpp",         OPTION_VAL_32BPP_AST },
         { NULL, NULL },
      },
      "16bpp(native)"
   },
   // Sort of, it's more like 15-bit high color and 24-bit true color for visible output. The alpha channel is used for mask bit. Vulkan renderer uses ABGR1555_555 for 31 bits internal? FMVs are always 24-bit on all renderers like original hardware (BGR888, no alpha)
#endif
   {
      BEETLE_OPT(dither_mode),
      BEETLE_OPT_DITHER_MODE_LABEL_AST,
      NULL,
      BEETLE_OPT_DITHER_MODE_INFO_0_AST,
      NULL,
      "video",
      {
         { "1x(native)",          OPTION_VAL_1X_NATIVE_AST },
         { "internal resolution", OPTION_VAL_INTERNAL_RESOLUTION_AST },
         { "disabled",            NULL },
         { NULL, NULL },
      },
      "1x(native)"
   },
#ifdef HAVE_VULKAN
   {
      BEETLE_OPT(scaled_uv_offset),
      BEETLE_OPT_SCALED_UV_OFFSET_LABEL_AST,
      NULL,
      BEETLE_OPT_SCALED_UV_OFFSET_INFO_0_AST,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled"
   },
#endif
#if defined(HAVE_OPENGL) || defined(HAVE_OPENGLES) || defined(HAVE_VULKAN)
   {
      BEETLE_OPT(filter),
      BEETLE_OPT_FILTER_LABEL_AST,
      NULL,
      BEETLE_OPT_FILTER_INFO_0_AST,
      NULL,
      "video",
      {
         { "nearest",  OPTION_VAL_NEAREST_AST },
         { "SABR",     OPTION_VAL_SABR_AST },
         { "xBR",      OPTION_VAL_XBR_AST },
         { "bilinear", OPTION_VAL_BILINEAR_AST },
         { "3-point",  OPTION_VAL_3_POINT_AST },
         { "JINC2",    OPTION_VAL_JINC2_AST },
         { NULL, NULL },
      },
      "nearest"
   },
#ifdef HAVE_VULKAN
   {
      BEETLE_OPT(filter_exclude_sprite),
      BEETLE_OPT_FILTER_EXCLUDE_SPRITE_LABEL_AST,
      NULL,
      BEETLE_OPT_FILTER_EXCLUDE_SPRITE_INFO_0_AST,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "opaque", OPTION_VAL_OPAQUE_AST },
         { "all", OPTION_VAL_ALL_AST },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(filter_exclude_2d_polygon),
      BEETLE_OPT_FILTER_EXCLUDE_2D_POLYGON_LABEL_AST,
      NULL,
      BEETLE_OPT_FILTER_EXCLUDE_2D_POLYGON_INFO_0_AST,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "opaque", OPTION_VAL_OPAQUE_AST },
         { "all", OPTION_VAL_ALL_AST },
         { NULL, NULL },
      },
      "disabled"
   },
#endif
#endif
#ifdef HAVE_VULKAN
   {
      BEETLE_OPT(adaptive_smoothing),
      BEETLE_OPT_ADAPTIVE_SMOOTHING_LABEL_AST,
      NULL,
      BEETLE_OPT_ADAPTIVE_SMOOTHING_INFO_0_AST,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(super_sampling),
      BEETLE_OPT_SUPER_SAMPLING_LABEL_AST,
      NULL,
      BEETLE_OPT_SUPER_SAMPLING_INFO_0_AST,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(msaa),
      BEETLE_OPT_MSAA_LABEL_AST,
      NULL,
      BEETLE_OPT_MSAA_INFO_0_AST,
      NULL,
      "video",
      {
         { "1x",  OPTION_VAL_1X_AST },
         { "2x",  OPTION_VAL_2X_AST },
         { "4x",  OPTION_VAL_4X_AST },
         { "8x",  OPTION_VAL_8X_AST },
         { "16x", OPTION_VAL_16X_AST },
         { NULL, NULL },
      },
      "1x"
   },
   {
      BEETLE_OPT(mdec_yuv),
      BEETLE_OPT_MDEC_YUV_LABEL_AST,
      NULL,
      BEETLE_OPT_MDEC_YUV_INFO_0_AST,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(track_textures),
      BEETLE_OPT_TRACK_TEXTURES_LABEL_AST,
      NULL,
      BEETLE_OPT_TRACK_TEXTURES_INFO_0_AST,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
#ifdef TEXTURE_DUMPING_ENABLED
   {
      BEETLE_OPT(dump_textures),
      BEETLE_OPT_DUMP_TEXTURES_LABEL_AST,
      NULL,
      BEETLE_OPT_DUMP_TEXTURES_INFO_0_AST,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
#endif
   {
      BEETLE_OPT(replace_textures),
      BEETLE_OPT_REPLACE_TEXTURES_LABEL_AST,
      NULL,
      BEETLE_OPT_REPLACE_TEXTURES_INFO_0_AST,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
#endif
   {
      BEETLE_OPT(pgxp_mode),
      BEETLE_OPT_PGXP_MODE_LABEL_AST,
      NULL,
      BEETLE_OPT_PGXP_MODE_INFO_0_AST,
      NULL,
      "pgxp",
      {
         { "disabled",     NULL },
         { "memory only",  OPTION_VAL_MEMORY_ONLY_AST },
         { "memory + CPU", OPTION_VAL_MEMORY_CPU_AST },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(pgxp_2d_tol),
      BEETLE_OPT_PGXP_2D_TOL_LABEL_AST,
      NULL,
      BEETLE_OPT_PGXP_2D_TOL_INFO_0_AST,
      NULL,
      "pgxp",
      {
         { "disabled", NULL },
         { "0px", OPTION_VAL_0PX_AST },
         { "1px", OPTION_VAL_1PX_AST },
         { "2px", OPTION_VAL_2PX_AST },
         { "3px", OPTION_VAL_3PX_AST },
         { "4px", OPTION_VAL_4PX_AST },
         { "5px", OPTION_VAL_5PX_AST },
         { "6px", OPTION_VAL_6PX_AST },
         { "7px", OPTION_VAL_7PX_AST },
         { "8px", OPTION_VAL_8PX_AST },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(pgxp_nclip),
      BEETLE_OPT_PGXP_NCLIP_LABEL_AST,
      NULL,
      BEETLE_OPT_PGXP_NCLIP_INFO_0_AST,
      NULL,
      "pgxp",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
#if defined(HAVE_OPENGL) || defined(HAVE_OPENGLES) || defined(HAVE_VULKAN)
   {
      BEETLE_OPT(pgxp_vertex),
      BEETLE_OPT_PGXP_VERTEX_LABEL_AST,
      NULL,
      BEETLE_OPT_PGXP_VERTEX_INFO_0_AST,
      NULL,
      "pgxp",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
#endif
   {
      BEETLE_OPT(pgxp_texture),
      BEETLE_OPT_PGXP_TEXTURE_LABEL_AST,
      NULL,
      BEETLE_OPT_PGXP_TEXTURE_INFO_0_AST,
      NULL,
      "pgxp",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(display_internal_fps),
      BEETLE_OPT_DISPLAY_INTERNAL_FPS_LABEL_AST,
      NULL,
      BEETLE_OPT_DISPLAY_INTERNAL_FPS_INFO_0_AST,
      NULL,
      "osd",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(display_osd),
      BEETLE_OPT_DISPLAY_OSD_LABEL_AST,
      NULL,
      BEETLE_OPT_DISPLAY_OSD_INFO_0_AST,
      NULL,
      "osd",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled"
   },
#if defined(HAVE_OPENGL) || defined(HAVE_OPENGLES) || defined(HAVE_VULKAN)
   {
      BEETLE_OPT(display_vram),
      BEETLE_OPT_DISPLAY_VRAM_LABEL_AST,
      NULL,
      BEETLE_OPT_DISPLAY_VRAM_INFO_0_AST,
      NULL,
      "osd",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
#endif
   {
      BEETLE_OPT(analog_calibration),
      BEETLE_OPT_ANALOG_CALIBRATION_LABEL_AST,
      NULL,
      BEETLE_OPT_ANALOG_CALIBRATION_INFO_0_AST,
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(analog_toggle),
      BEETLE_OPT_ANALOG_TOGGLE_LABEL_AST,
      NULL,
      BEETLE_OPT_ANALOG_TOGGLE_INFO_0_AST,
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { "enabled-analog", OPTION_VAL_ENABLED_ANALOG_AST },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(analog_toggle_combo),
      BEETLE_OPT_ANALOG_TOGGLE_COMBO_LABEL_AST,
      NULL,
      BEETLE_OPT_ANALOG_TOGGLE_COMBO_INFO_0_AST,
      NULL,
      "input",
      {
         { "l1+l2+r1+r2+start+select", OPTION_VAL_L1_L2_R1_R2_START_SELECT_AST },
         { "l1+r1+select",             OPTION_VAL_L1_R1_SELECT_AST },
         { "l1+r1+start",              OPTION_VAL_L1_R1_START_AST },
         { "l1+r1+l3",                 OPTION_VAL_L1_R1_L3_AST },
         { "l1+r1+r3",                 OPTION_VAL_L1_R1_R3_AST },
         { "l2+r2+select",             OPTION_VAL_L2_R2_SELECT_AST },
         { "l2+r2+start",              OPTION_VAL_L2_R2_START_AST },
         { "l2+r2+l3",                 OPTION_VAL_L2_R2_L3_AST },
         { "l2+r2+r3",                 OPTION_VAL_L2_R2_R3_AST },
         { "l3+r3",                    OPTION_VAL_L3_R3_AST },
         { NULL, NULL },
      },
      "l1+r1+select"
   },
   {
      BEETLE_OPT(analog_toggle_hold),
      BEETLE_OPT_ANALOG_TOGGLE_HOLD_LABEL_AST,
      NULL,
      BEETLE_OPT_ANALOG_TOGGLE_HOLD_INFO_0_AST,
      NULL,
      "input",
      {
         { "0", OPTION_VAL_0_AST },
         { "1", OPTION_VAL_1_AST },
         { "2", OPTION_VAL_2_AST },
         { "3", OPTION_VAL_3_AST },
         { "4", OPTION_VAL_4_AST },
         { "5", OPTION_VAL_5_AST },
         { NULL, NULL },
      },
      "1"
   },
   {
      BEETLE_OPT(enable_multitap_port1),
      BEETLE_OPT_ENABLE_MULTITAP_PORT1_LABEL_AST,
      NULL,
      BEETLE_OPT_ENABLE_MULTITAP_PORT1_INFO_0_AST,
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(enable_multitap_port2),
      BEETLE_OPT_ENABLE_MULTITAP_PORT2_LABEL_AST,
      NULL,
      BEETLE_OPT_ENABLE_MULTITAP_PORT2_INFO_0_AST,
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(gun_input_mode),
      BEETLE_OPT_GUN_INPUT_MODE_LABEL_AST,
      NULL,
      BEETLE_OPT_GUN_INPUT_MODE_INFO_0_AST,
      NULL,
      "input",
      {
         { "lightgun",    OPTION_VAL_LIGHTGUN_AST },
         { "touchscreen", OPTION_VAL_TOUCHSCREEN_AST },
         { NULL, NULL },
      },
      "lightgun"
   },
   // Shouldn't the gun_input_mode just be Mouse vs. Touchscreen?
   {
      BEETLE_OPT(gun_cursor),
      BEETLE_OPT_GUN_CURSOR_LABEL_AST,
      NULL,
      BEETLE_OPT_GUN_CURSOR_INFO_0_AST,
      NULL,
      "input",
      {
         { "off",   "disabled" },
         { "cross", OPTION_VAL_CROSS_AST },
         { "dot",   OPTION_VAL_DOT_AST },
         { NULL, NULL },
      },
      "cross"
   },
   {
      BEETLE_OPT(crosshair_color_p1),
      BEETLE_OPT_CROSSHAIR_COLOR_P1_LABEL_AST,
      NULL,
      BEETLE_OPT_CROSSHAIR_COLOR_P1_INFO_0_AST,
      NULL,
      "input",
      {
         { "red",    OPTION_VAL_RED_AST },
         { "blue",   OPTION_VAL_BLUE_AST },
         { "green",  OPTION_VAL_GREEN_AST },
         { "orange", OPTION_VAL_ORANGE_AST },
         { "yellow", OPTION_VAL_YELLOW_AST },
         { "cyan",   OPTION_VAL_CYAN_AST },
         { "pink",   OPTION_VAL_PINK_AST },
         { "purple", OPTION_VAL_PURPLE_AST },
         { "black",  OPTION_VAL_BLACK_AST },
         { "white",  OPTION_VAL_WHITE_AST },
         { NULL, NULL },
      },
      "red"
   },
   {
      BEETLE_OPT(crosshair_color_p2),
      BEETLE_OPT_CROSSHAIR_COLOR_P2_LABEL_AST,
      NULL,
      BEETLE_OPT_CROSSHAIR_COLOR_P2_INFO_0_AST,
      NULL,
      "input",
      {
         { "blue",   OPTION_VAL_BLUE_AST },
         { "red",    OPTION_VAL_RED_AST },
         { "green",  OPTION_VAL_GREEN_AST },
         { "orange", OPTION_VAL_ORANGE_AST },
         { "yellow", OPTION_VAL_YELLOW_AST },
         { "cyan",   OPTION_VAL_CYAN_AST },
         { "pink",   OPTION_VAL_PINK_AST },
         { "purple", OPTION_VAL_PURPLE_AST },
         { "black",  OPTION_VAL_BLACK_AST },
         { "white",  OPTION_VAL_WHITE_AST },
         { NULL, NULL },
      },
      "blue"
   },
   {
      BEETLE_OPT(mouse_sensitivity),
      BEETLE_OPT_MOUSE_SENSITIVITY_LABEL_AST,
      NULL,
      BEETLE_OPT_MOUSE_SENSITIVITY_INFO_0_AST,
      NULL,
      "input",
      {
         { "5%",   OPTION_VAL_5_O42_AST },
         { "10%",  OPTION_VAL_10_AST },
         { "15%",  OPTION_VAL_15_AST },
         { "20%",  OPTION_VAL_20_AST },
         { "25%",  OPTION_VAL_25_AST },
         { "30%",  OPTION_VAL_30_AST },
         { "35%",  OPTION_VAL_35_AST },
         { "40%",  OPTION_VAL_40_AST },
         { "45%",  OPTION_VAL_45_AST },
         { "50%",  OPTION_VAL_50_AST },
         { "55%",  OPTION_VAL_55_AST },
         { "60%",  OPTION_VAL_60_AST },
         { "65%",  OPTION_VAL_65_AST },
         { "70%",  OPTION_VAL_70_AST },
         { "75%",  OPTION_VAL_75_AST },
         { "80%",  OPTION_VAL_80_AST },
         { "85%",  OPTION_VAL_85_AST },
         { "90%",  OPTION_VAL_90_AST },
         { "95%",  OPTION_VAL_95_AST },
         { "100%", OPTION_VAL_100_O42_AST },
         { "105%", OPTION_VAL_105_AST },
         { "110%", OPTION_VAL_110_AST },
         { "115%", OPTION_VAL_115_AST },
         { "120%", OPTION_VAL_120_AST },
         { "125%", OPTION_VAL_125_AST },
         { "130%", OPTION_VAL_130_AST },
         { "135%", OPTION_VAL_135_AST },
         { "140%", OPTION_VAL_140_AST },
         { "145%", OPTION_VAL_145_AST },
         { "150%", OPTION_VAL_150_AST },
         { "155%", OPTION_VAL_155_AST },
         { "160%", OPTION_VAL_160_AST },
         { "165%", OPTION_VAL_165_AST },
         { "170%", OPTION_VAL_170_AST },
         { "175%", OPTION_VAL_175_AST },
         { "180%", OPTION_VAL_180_AST },
         { "185%", OPTION_VAL_185_AST },
         { "190%", OPTION_VAL_190_AST },
         { "195%", OPTION_VAL_195_AST },
         { "200%", OPTION_VAL_200_AST },
         { NULL, NULL },
      },
      "100%"
   },
   {
      BEETLE_OPT(negcon_response),
      BEETLE_OPT_NEGCON_RESPONSE_LABEL_AST,
      NULL,
      BEETLE_OPT_NEGCON_RESPONSE_INFO_0_AST,
      NULL,
      "input",
      {
         { "linear",    OPTION_VAL_LINEAR_AST },
         { "quadratic", OPTION_VAL_QUADRATIC_AST },
         { "cubic",     OPTION_VAL_CUBIC_AST },
         { NULL, NULL },
      },
      "linear"
   },
   {
      BEETLE_OPT(negcon_deadzone),
      BEETLE_OPT_NEGCON_DEADZONE_LABEL_AST,
      NULL,
      BEETLE_OPT_NEGCON_DEADZONE_INFO_0_AST,
      NULL,
      "input",
      {
         { "0%",  OPTION_VAL_0_O44_AST },
         { "5%",  OPTION_VAL_5_O42_AST },
         { "10%", OPTION_VAL_10_AST },
         { "15%", OPTION_VAL_15_AST },
         { "20%", OPTION_VAL_20_AST },
         { "25%", OPTION_VAL_25_AST },
         { "30%", OPTION_VAL_30_AST },
         { NULL, NULL },
      },
      "0%"
   },
   {
      BEETLE_OPT(use_mednafen_memcard0_method),
      BEETLE_OPT_USE_MEDNAFEN_MEMCARD0_METHOD_LABEL_AST,
      NULL,
      BEETLE_OPT_USE_MEDNAFEN_MEMCARD0_METHOD_INFO_0_AST,
      NULL,
      "memcards",
      {
         { "libretro", OPTION_VAL_LIBRETRO_AST },
         { "mednafen", OPTION_VAL_MEDNAFEN_AST },
         { NULL, NULL },
      },
      "libretro"
   },
   {
      BEETLE_OPT(enable_memcard1),
      BEETLE_OPT_ENABLE_MEMCARD1_LABEL_AST,
      NULL,
      BEETLE_OPT_ENABLE_MEMCARD1_INFO_0_AST,
      NULL,
      "memcards",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(shared_memory_cards),
      BEETLE_OPT_SHARED_MEMORY_CARDS_LABEL_AST,
      NULL,
      BEETLE_OPT_SHARED_MEMORY_CARDS_INFO_0_AST,
      NULL,
      "memcards",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(memcard_left_index),
      BEETLE_OPT_MEMCARD_LEFT_INDEX_LABEL_AST,
      NULL,
      BEETLE_OPT_MEMCARD_LEFT_INDEX_INFO_0_AST,
      NULL,
      "memcards",
      {
         { "0",  OPTION_VAL_0_O48_AST },
         { "1",  NULL },
         { "2",  NULL },
         { "3",  NULL },
         { "4",  NULL },
         { "5",  NULL },
         { "6",  NULL },
         { "7",  NULL },
         { "8",  NULL },
         { "9",  NULL },
         { "10", NULL },
         { "11", NULL },
         { "12", NULL },
         { "13", NULL },
         { "14", NULL },
         { "15", NULL },
         { "16", NULL },
         { "17", NULL },
         { "18", NULL },
         { "19", NULL },
         { "20", NULL },
         { "21", NULL },
         { "22", NULL },
         { "23", NULL },
         { "24", NULL },
         { "25", NULL },
         { "26", NULL },
         { "27", NULL },
         { "28", NULL },
         { "29", NULL },
         { "30", NULL },
         { "31", NULL },
         { "32", NULL },
         { "33", NULL },
         { "34", NULL },
         { "35", NULL },
         { "36", NULL },
         { "37", NULL },
         { "38", NULL },
         { "39", NULL },
         { "40", NULL },
         { "41", NULL },
         { "42", NULL },
         { "43", NULL },
         { "44", NULL },
         { "45", NULL },
         { "46", NULL },
         { "47", NULL },
         { "48", NULL },
         { "49", NULL },
         { "50", NULL },
         { "51", NULL },
         { "52", NULL },
         { "53", NULL },
         { "54", NULL },
         { "55", NULL },
         { "56", NULL },
         { "57", NULL },
         { "58", NULL },
         { "59", NULL },
         { "60", NULL },
         { "61", NULL },
         { "62", NULL },
         { "63", NULL },
         { NULL, NULL },
      },
      "0"
   },
   {
      BEETLE_OPT(memcard_right_index),
      BEETLE_OPT_MEMCARD_RIGHT_INDEX_LABEL_AST,
      NULL,
      BEETLE_OPT_MEMCARD_RIGHT_INDEX_INFO_0_AST,
      NULL,
      "memcards",
      {
         { "0",  NULL },
         { "1",  OPTION_VAL_1_O49_AST },
         { "2",  NULL },
         { "3",  NULL },
         { "4",  NULL },
         { "5",  NULL },
         { "6",  NULL },
         { "7",  NULL },
         { "8",  NULL },
         { "9",  NULL },
         { "10", NULL },
         { "11", NULL },
         { "12", NULL },
         { "13", NULL },
         { "14", NULL },
         { "15", NULL },
         { "16", NULL },
         { "17", NULL },
         { "18", NULL },
         { "19", NULL },
         { "20", NULL },
         { "21", NULL },
         { "22", NULL },
         { "23", NULL },
         { "24", NULL },
         { "25", NULL },
         { "26", NULL },
         { "27", NULL },
         { "28", NULL },
         { "29", NULL },
         { "30", NULL },
         { "31", NULL },
         { "32", NULL },
         { "33", NULL },
         { "34", NULL },
         { "35", NULL },
         { "36", NULL },
         { "37", NULL },
         { "38", NULL },
         { "39", NULL },
         { "40", NULL },
         { "41", NULL },
         { "42", NULL },
         { "43", NULL },
         { "44", NULL },
         { "45", NULL },
         { "46", NULL },
         { "47", NULL },
         { "48", NULL },
         { "49", NULL },
         { "50", NULL },
         { "51", NULL },
         { "52", NULL },
         { "53", NULL },
         { "54", NULL },
         { "55", NULL },
         { "56", NULL },
         { "57", NULL },
         { "58", NULL },
         { "59", NULL },
         { "60", NULL },
         { "61", NULL },
         { "62", NULL },
         { "63", NULL },
         { NULL, NULL },
      },
      "1"
   },
   {
      BEETLE_OPT(line_render),
      BEETLE_OPT_LINE_RENDER_LABEL_AST,
      NULL,
      BEETLE_OPT_LINE_RENDER_INFO_0_AST,
      NULL,
      "hacks",
      {
         { "default",    OPTION_VAL_DEFAULT_AST },
         { "aggressive", OPTION_VAL_AGGRESSIVE_AST },
         { "disabled",   NULL },
         { NULL, NULL },
      },
      "default"
   },
   {
      BEETLE_OPT(widescreen_hack),
      BEETLE_OPT_WIDESCREEN_HACK_LABEL_AST,
      NULL,
      BEETLE_OPT_WIDESCREEN_HACK_INFO_0_AST,
      NULL,
      "hacks",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(widescreen_hack_aspect_ratio),
      BEETLE_OPT_WIDESCREEN_HACK_ASPECT_RATIO_LABEL_AST,
      NULL,
      BEETLE_OPT_WIDESCREEN_HACK_ASPECT_RATIO_INFO_0_AST,
      NULL,
      "hacks",
      {
         { "16:9",  OPTION_VAL_16_9_AST },
         { "16:10", OPTION_VAL_16_10_AST },
         { "18:9",  OPTION_VAL_18_9_AST },
         { "19:9",  OPTION_VAL_19_9_AST },
         { "20:9",  OPTION_VAL_20_9_AST },
         { "21:9",  OPTION_VAL_21_9_AST }, // 64:27
         { "32:9",  OPTION_VAL_32_9_AST },
         { NULL,    NULL },
      },
      "16:9"
   },
#if defined(HAVE_LIGHTREC)
   {
      BEETLE_OPT(cpu_dynarec),
      BEETLE_OPT_CPU_DYNAREC_LABEL_AST,
      NULL,
      BEETLE_OPT_CPU_DYNAREC_INFO_0_AST,
      NULL,
      "hacks",
      {
         { "disabled",        OPTION_VAL_DISABLED_AST },
         { "execute",         OPTION_VAL_EXECUTE_AST },
         { "execute_one",     OPTION_VAL_EXECUTE_ONE_AST },
         { "run_interpreter", OPTION_VAL_RUN_INTERPRETER_AST },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(dynarec_invalidate),
      BEETLE_OPT_DYNAREC_INVALIDATE_LABEL_AST,
      NULL,
      BEETLE_OPT_DYNAREC_INVALIDATE_INFO_0_AST,
      NULL,
      "hacks",
      {
         { "full", OPTION_VAL_FULL_AST },
         { "dma",  OPTION_VAL_DMA_AST },
         { NULL, NULL },
      },
      "full"
   },
   {
      BEETLE_OPT(dynarec_eventcycles),
      BEETLE_OPT_DYNAREC_EVENTCYCLES_LABEL_AST,
      NULL,
      BEETLE_OPT_DYNAREC_EVENTCYCLES_INFO_0_AST,
      NULL,
      "hacks",
      {
         { "128",  OPTION_VAL_128_AST },
         { "256",  NULL },
         { "384",  NULL },
         { "512",  NULL },
         { "640",  NULL },
         { "768",  NULL },
         { "896",  NULL },
         { "1024", NULL },
         { "1152", NULL },
         { "1280", NULL },
         { "1408", NULL },
         { "1536", NULL },
         { "1664", NULL },
         { "1792", NULL },
         { "1920", NULL },
         { "2048", NULL },
         { NULL, NULL },
      },
      "128"
   },
   {
      BEETLE_OPT(dynarec_spu_samples),
      BEETLE_OPT_DYNAREC_SPU_SAMPLES_LABEL_AST,
      NULL,
      BEETLE_OPT_DYNAREC_SPU_SAMPLES_INFO_0_AST,
      NULL,
      "hacks",
      {
         { "1",  OPTION_VAL_1_O49_AST },
         { "4",  NULL },
         { "16", NULL },
         { NULL, NULL },
      },
      "1"
   },
#endif
   {
      BEETLE_OPT(core_timing_fps),
      BEETLE_OPT_CORE_TIMING_FPS_LABEL_AST,
      NULL,
      BEETLE_OPT_CORE_TIMING_FPS_INFO_0_AST,
      NULL,
      "video",
      {
         { "auto_toggle",       OPTION_VAL_AUTO_TOGGLE_AST },
         { "force_progressive", OPTION_VAL_FORCE_PROGRESSIVE_AST },
         { "force_interlaced",  OPTION_VAL_FORCE_INTERLACED_AST },
      },
      "auto_toggle"
   },
   {
      BEETLE_OPT(deinterlacer),
      BEETLE_OPT_DEINTERLACER_LABEL_AST,
      NULL,
      BEETLE_OPT_DEINTERLACER_INFO_0_AST,
      NULL,
      "video",
      {
         { "weave",      OPTION_VAL_WEAVE_AST },
         { "bob",        OPTION_VAL_BOB_AST },
         { "bob_offset", OPTION_VAL_BOB_OFFSET_AST },
         { "fastmad",    OPTION_VAL_FASTMAD_AST },
         { "off",        "Off" },
         { NULL, NULL },
      },
      "weave"
   },
   {
      BEETLE_OPT(frame_duping),
      BEETLE_OPT_FRAME_DUPING_LABEL_AST,
      NULL,
      BEETLE_OPT_FRAME_DUPING_INFO_0_AST,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(pal_video_timing_override),
      BEETLE_OPT_PAL_VIDEO_TIMING_OVERRIDE_LABEL_AST,
      NULL,
      BEETLE_OPT_PAL_VIDEO_TIMING_OVERRIDE_INFO_0_AST,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(aspect_ratio),
      BEETLE_OPT_ASPECT_RATIO_LABEL_AST,
      NULL,
      BEETLE_OPT_ASPECT_RATIO_INFO_0_AST,
      NULL,
      "video",
      {
         { "corrected", OPTION_VAL_CORRECTED_AST },
         { "uncorrected", OPTION_VAL_UNCORRECTED_AST },
         { "4:3",  OPTION_VAL_4_3_AST },
         { "ntsc", OPTION_VAL_NTSC_AST },
      },
      "corrected"
   },
   {
      BEETLE_OPT(crop_overscan),
      BEETLE_OPT_CROP_OVERSCAN_LABEL_AST,
      NULL,
      BEETLE_OPT_CROP_OVERSCAN_INFO_0_AST,
      NULL,
      "video",
      {
         { "disabled", "Disabled" },
         { "static",   OPTION_VAL_STATIC_AST },
         { "smart",    OPTION_VAL_SMART_AST },
         { NULL, NULL },
      },
      "smart"
   },
   {
      BEETLE_OPT(image_crop),
      BEETLE_OPT_IMAGE_CROP_LABEL_AST,
      NULL,
      BEETLE_OPT_IMAGE_CROP_INFO_0_AST,
      NULL,
      "video",
      {
         { "disabled", "0" },
         { "1px",      OPTION_VAL_1PX_AST },
         { "2px",      OPTION_VAL_2PX_AST },
         { "3px",      OPTION_VAL_3PX_AST },
         { "4px",      OPTION_VAL_4PX_AST },
         { "5px",      OPTION_VAL_5PX_AST },
         { "6px",      OPTION_VAL_6PX_AST },
         { "7px",      OPTION_VAL_7PX_AST },
         { "8px",      OPTION_VAL_8PX_AST },
         { "9px",      OPTION_VAL_9PX_AST },
         { "10px",     OPTION_VAL_10PX_AST },
         { "11px",     OPTION_VAL_11PX_AST },
         { "12px",     OPTION_VAL_12PX_AST },
         { "13px",     OPTION_VAL_13PX_AST },
         { "14px",     OPTION_VAL_14PX_AST },
         { "15px",     OPTION_VAL_15PX_AST },
         { "16px",     OPTION_VAL_16PX_AST },
         { "17px",     OPTION_VAL_17PX_AST },
         { "18px",     OPTION_VAL_18PX_AST },
         { "19px",     OPTION_VAL_19PX_AST },
         { "20px",     OPTION_VAL_20PX_AST },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(image_offset),
      BEETLE_OPT_IMAGE_OFFSET_LABEL_AST,
      NULL,
      BEETLE_OPT_IMAGE_OFFSET_INFO_0_AST,
      NULL,
      "video",
      {
         { "-12px",    OPTION_VAL_12PX_O64_AST },
         { "-11px",    OPTION_VAL_11PX_O64_AST },
         { "-10px",    OPTION_VAL_10PX_O64_AST },
         { "-9px",     OPTION_VAL_9PX_O64_AST },
         { "-8px",     OPTION_VAL_8PX_O64_AST },
         { "-7px",     OPTION_VAL_7PX_O64_AST },
         { "-6px",     OPTION_VAL_6PX_O64_AST },
         { "-5px",     OPTION_VAL_5PX_O64_AST },
         { "-4px",     OPTION_VAL_4PX_O64_AST },
         { "-3px",     OPTION_VAL_3PX_O64_AST },
         { "-2px",     OPTION_VAL_2PX_O64_AST },
         { "-1px",     OPTION_VAL_1PX_O64_AST },
         { "disabled", "0" },
         { "+1px",     OPTION_VAL_1PX_O64_0_AST },
         { "+2px",     OPTION_VAL_2PX_O64_0_AST },
         { "+3px",     OPTION_VAL_3PX_O64_0_AST },
         { "+4px",     OPTION_VAL_4PX_O64_0_AST },
         { "+5px",     OPTION_VAL_5PX_O64_0_AST },
         { "+6px",     OPTION_VAL_6PX_O64_0_AST },
         { "+7px",     OPTION_VAL_7PX_O64_0_AST },
         { "+8px",     OPTION_VAL_8PX_O64_0_AST },
         { "+9px",     OPTION_VAL_9PX_O64_0_AST },
         { "+10px",    OPTION_VAL_10PX_O64_0_AST },
         { "+11px",    OPTION_VAL_11PX_O64_0_AST },
         { "+12px",    OPTION_VAL_12PX_O64_0_AST },
         { NULL, NULL },
      },
      "disabled"
   },
#if defined(HAVE_OPENGL) || defined(HAVE_OPENGLES) || defined(HAVE_VULKAN)
   {
      BEETLE_OPT(image_offset_cycles),
      BEETLE_OPT_IMAGE_OFFSET_CYCLES_LABEL_AST,
      NULL,
      BEETLE_OPT_IMAGE_OFFSET_CYCLES_INFO_0_AST,
      NULL,
      "video",
      {
         { "-40",      NULL },
         { "-39",      NULL },
         { "-38",      NULL },
         { "-37",      NULL },
         { "-36",      NULL },
         { "-35",      NULL },
         { "-34",      NULL },
         { "-33",      NULL },
         { "-32",      NULL },
         { "-31",      NULL },
         { "-30",      NULL },
         { "-29",      NULL },
         { "-28",      NULL },
         { "-27",      NULL },
         { "-26",      NULL },
         { "-25",      NULL },
         { "-24",      NULL },
         { "-23",      NULL },
         { "-22",      NULL },
         { "-21",      NULL },
         { "-20",      NULL },
         { "-19",      NULL },
         { "-18",      NULL },
         { "-17",      NULL },
         { "-16",      NULL },
         { "-15",      NULL },
         { "-14",      NULL },
         { "-13",      NULL },
         { "-12",      NULL },
         { "-11",      NULL },
         { "-10",      NULL },
         { "-9",       NULL },
         { "-8",       NULL },
         { "-7",       NULL },
         { "-6",       NULL },
         { "-5",       NULL },
         { "-4",       NULL },
         { "-3",       NULL },
         { "-2",       NULL },
         { "-1",       NULL },
         { "0",        "0" },
         { "+1",       NULL },
         { "+2",       NULL },
         { "+3",       NULL },
         { "+4",       NULL },
         { "+5",       NULL },
         { "+6",       NULL },
         { "+7",       NULL },
         { "+8",       NULL },
         { "+9",       NULL },
         { "+10",      NULL },
         { "+11",      NULL },
         { "+12",      NULL },
         { "+13",      NULL },
         { "+14",      NULL },
         { "+15",      NULL },
         { "+16",      NULL },
         { "+17",      NULL },
         { "+18",      NULL },
         { "+19",      NULL },
         { "+20",      NULL },
         { "+21",      NULL },
         { "+22",      NULL },
         { "+23",      NULL },
         { "+24",      NULL },
         { "+25",      NULL },
         { "+26",      NULL },
         { "+27",      NULL },
         { "+28",      NULL },
         { "+29",      NULL },
         { "+30",      NULL },
         { "+31",      NULL },
         { "+32",      NULL },
         { "+33",      NULL },
         { "+34",      NULL },
         { "+35",      NULL },
         { "+36",      NULL },
         { "+37",      NULL },
         { "+38",      NULL },
         { "+39",      NULL },
         { "+40",      NULL },
         { NULL, NULL},
      },
      "0"
   },
#endif
   {
      BEETLE_OPT(initial_scanline),
      BEETLE_OPT_INITIAL_SCANLINE_LABEL_AST,
      NULL,
      BEETLE_OPT_INITIAL_SCANLINE_INFO_0_AST,
      NULL,
      "video",
      {
         { "0",  OPTION_VAL_0_O48_AST },
         { "1",  NULL },
         { "2",  NULL },
         { "3",  NULL },
         { "4",  NULL },
         { "5",  NULL },
         { "6",  NULL },
         { "7",  NULL },
         { "8",  NULL },
         { "9",  NULL },
         { "10", NULL },
         { "11", NULL },
         { "12", NULL },
         { "13", NULL },
         { "14", NULL },
         { "15", NULL },
         { "16", NULL },
         { "17", NULL },
         { "18", NULL },
         { "19", NULL },
         { "20", NULL },
         { "21", NULL },
         { "22", NULL },
         { "23", NULL },
         { "24", NULL },
         { "25", NULL },
         { "26", NULL },
         { "27", NULL },
         { "28", NULL },
         { "29", NULL },
         { "30", NULL },
         { "31", NULL },
         { "32", NULL },
         { "33", NULL },
         { "34", NULL },
         { "35", NULL },
         { "36", NULL },
         { "37", NULL },
         { "38", NULL },
         { "39", NULL },
         { "40", NULL },
         { NULL, NULL },
      },
      "0"
   },
   {
      BEETLE_OPT(last_scanline),
      BEETLE_OPT_LAST_SCANLINE_LABEL_AST,
      NULL,
      BEETLE_OPT_LAST_SCANLINE_INFO_0_AST,
      NULL,
      "video",
      {
         { "210", NULL },
         { "211", NULL },
         { "212", NULL },
         { "213", NULL },
         { "214", NULL },
         { "215", NULL },
         { "216", NULL },
         { "217", NULL },
         { "218", NULL },
         { "219", NULL },
         { "220", NULL },
         { "221", NULL },
         { "222", NULL },
         { "223", NULL },
         { "224", NULL },
         { "225", NULL },
         { "226", NULL },
         { "227", NULL },
         { "228", NULL },
         { "229", NULL },
         { "230", NULL },
         { "231", NULL },
         { "232", NULL },
         { "233", NULL },
         { "234", NULL },
         { "235", NULL },
         { "236", NULL },
         { "237", NULL },
         { "238", NULL },
         { "239", OPTION_VAL_239_AST },
         { NULL, NULL },
      },
      "239"
   },
   {
      BEETLE_OPT(initial_scanline_pal),
      BEETLE_OPT_INITIAL_SCANLINE_PAL_LABEL_AST,
      NULL,
      BEETLE_OPT_INITIAL_SCANLINE_PAL_INFO_0_AST,
      NULL,
      "video",
      {
         { "0",  OPTION_VAL_0_O48_AST },
         { "1",  NULL },
         { "2",  NULL },
         { "3",  NULL },
         { "4",  NULL },
         { "5",  NULL },
         { "6",  NULL },
         { "7",  NULL },
         { "8",  NULL },
         { "9",  NULL },
         { "10", NULL },
         { "11", NULL },
         { "12", NULL },
         { "13", NULL },
         { "14", NULL },
         { "15", NULL },
         { "16", NULL },
         { "17", NULL },
         { "18", NULL },
         { "19", NULL },
         { "20", NULL },
         { "21", NULL },
         { "22", NULL },
         { "23", NULL },
         { "24", NULL },
         { "25", NULL },
         { "26", NULL },
         { "27", NULL },
         { "28", NULL },
         { "29", NULL },
         { "30", NULL },
         { "31", NULL },
         { "32", NULL },
         { "33", NULL },
         { "34", NULL },
         { "35", NULL },
         { "36", NULL },
         { "37", NULL },
         { "38", NULL },
         { "39", NULL },
         { "40", NULL },
         { NULL, NULL },
      },
      "0"
   },
   {
      BEETLE_OPT(last_scanline_pal),
      BEETLE_OPT_LAST_SCANLINE_PAL_LABEL_AST,
      NULL,
      BEETLE_OPT_LAST_SCANLINE_PAL_INFO_0_AST,
      NULL,
      "video",
      {
         { "230", NULL },
         { "231", NULL },
         { "232", NULL },
         { "233", NULL },
         { "234", NULL },
         { "235", NULL },
         { "236", NULL },
         { "237", NULL },
         { "238", NULL },
         { "239", NULL },
         { "240", NULL },
         { "241", NULL },
         { "242", NULL },
         { "243", NULL },
         { "244", NULL },
         { "245", NULL },
         { "246", NULL },
         { "247", NULL },
         { "248", NULL },
         { "249", NULL },
         { "250", NULL },
         { "251", NULL },
         { "252", NULL },
         { "253", NULL },
         { "254", NULL },
         { "255", NULL },
         { "256", NULL },
         { "257", NULL },
         { "258", NULL },
         { "259", NULL },
         { "260", NULL },
         { "261", NULL },
         { "262", NULL },
         { "263", NULL },
         { "264", NULL },
         { "265", NULL },
         { "266", NULL },
         { "267", NULL },
         { "268", NULL },
         { "269", NULL },
         { "270", NULL },
         { "271", NULL },
         { "272", NULL },
         { "273", NULL },
         { "274", NULL },
         { "275", NULL },
         { "276", NULL },
         { "277", NULL },
         { "278", NULL },
         { "279", NULL },
         { "280", NULL },
         { "281", NULL },
         { "282", NULL },
         { "283", NULL },
         { "284", NULL },
         { "285", NULL },
         { "286", NULL },
         { "287", OPTION_VAL_287_AST },
         { NULL, NULL },
      },
      "287"
   },
   { NULL, NULL, NULL, NULL, NULL, NULL, {{0}}, NULL },
};
struct retro_core_options_v2 options_ast = {
   option_cats_ast,
   option_defs_ast
};

/* RETRO_LANGUAGE_BE */

#define CATEGORY_SYSTEM_LABEL_BE "Сістэма"
#define CATEGORY_SYSTEM_INFO_0_BE NULL
#define CATEGORY_VIDEO_LABEL_BE "Відэа"
#define CATEGORY_VIDEO_INFO_0_BE "Налады суадносін бакоў, кадравання, графічных фільтраў і пропуску кадраў."
#define CATEGORY_PGXP_LABEL_BE NULL
#define CATEGORY_PGXP_INFO_0_BE "Настройкі для паляпшэння якасці графікі вышэй за ўзровень арыгінальнай кансолі. PGXP дапамагае пазбавіцца ад Z-канфліктаў і дэфармацыі тэкстур."
#define CATEGORY_OSD_LABEL_BE "Адлюстраванне на экране"
#define CATEGORY_OSD_INFO_0_BE "Настройка паказу апавяшчэнняў на экране."
#define CATEGORY_INPUT_LABEL_BE "Увод"
#define CATEGORY_INPUT_INFO_0_BE "Настройкі светлавога пісталета, мышы і neGcon."
#define CATEGORY_MEMCARDS_LABEL_BE "Карта памяці"
#define CATEGORY_MEMCARDS_INFO_0_BE NULL
#define CATEGORY_HACKS_LABEL_BE "Хакі эмуляцыі"
#define CATEGORY_HACKS_INFO_0_BE NULL
#define BEETLE_OPT_CPU_FREQ_SCALE_LABEL_BE "Маштабаванне частаты CPU (разгон)"
#define BEETLE_OPT_CPU_FREQ_SCALE_INFO_0_BE "Разганяе (ці запавольвае) працэсар эмуляванай PSX. У некаторых гульнях дапамагае прыбраць запаволенні і падняць частату кадраў, але падвышае патрабаванні да рэсурсаў. З-за наяўнасці ўнутранага абмежавання частаты кадраў некаторыя гульні пры разгоне не атрымліваюць пераваг. Разгон таксама можа завышаць хуткасць некаторых эфектаў анімацыі."
#define OPTION_VAL_50_BE "50 %"
#define OPTION_VAL_60_BE "60 %"
#define OPTION_VAL_70_BE "70 %"
#define OPTION_VAL_80_BE "80 %"
#define OPTION_VAL_90_BE "90 %"
#define OPTION_VAL_100_BE NULL
#define OPTION_VAL_110_BE "110 %"
#define OPTION_VAL_120_BE "120 %"
#define OPTION_VAL_130_BE "130 %"
#define OPTION_VAL_140_BE "140 %"
#define OPTION_VAL_150_BE "150 %"
#define OPTION_VAL_160_BE "160 %"
#define OPTION_VAL_170_BE "170 %"
#define OPTION_VAL_180_BE "180 %"
#define OPTION_VAL_190_BE "190 %"
#define OPTION_VAL_200_BE "200 %"
#define OPTION_VAL_210_BE "210 %"
#define OPTION_VAL_220_BE "220 %"
#define OPTION_VAL_230_BE "230 %"
#define OPTION_VAL_240_BE "240 %"
#define OPTION_VAL_250_BE "250 %"
#define OPTION_VAL_260_BE "260 %"
#define OPTION_VAL_270_BE "270 %"
#define OPTION_VAL_280_BE "280 %"
#define OPTION_VAL_290_BE "290 %"
#define OPTION_VAL_300_BE "300 %"
#define OPTION_VAL_310_BE "310 %"
#define OPTION_VAL_320_BE "320 %"
#define OPTION_VAL_330_BE "330 %"
#define OPTION_VAL_340_BE "340 %"
#define OPTION_VAL_350_BE "350 %"
#define OPTION_VAL_360_BE "360 %"
#define OPTION_VAL_370_BE "370 %"
#define OPTION_VAL_380_BE "380 %"
#define OPTION_VAL_390_BE "390 %"
#define OPTION_VAL_400_BE "400 %"
#define OPTION_VAL_410_BE "410 %"
#define OPTION_VAL_420_BE "420 %"
#define OPTION_VAL_430_BE "430 %"
#define OPTION_VAL_440_BE "440 %"
#define OPTION_VAL_450_BE "450 %"
#define OPTION_VAL_460_BE "460 %"
#define OPTION_VAL_470_BE "470 %"
#define OPTION_VAL_480_BE "480 %"
#define OPTION_VAL_490_BE "490 %"
#define OPTION_VAL_500_BE "500 %"
#define OPTION_VAL_510_BE "510 %"
#define OPTION_VAL_520_BE "520 %"
#define OPTION_VAL_530_BE "530 %"
#define OPTION_VAL_540_BE "540 %"
#define OPTION_VAL_550_BE "550 %"
#define OPTION_VAL_560_BE "560 %"
#define OPTION_VAL_570_BE "570 %"
#define OPTION_VAL_580_BE "580 %"
#define OPTION_VAL_590_BE "590 %"
#define OPTION_VAL_600_BE "600 %"
#define OPTION_VAL_610_BE "610 %"
#define OPTION_VAL_620_BE "620 %"
#define OPTION_VAL_630_BE "630 %"
#define OPTION_VAL_640_BE "640 %"
#define OPTION_VAL_650_BE "650 %"
#define OPTION_VAL_660_BE "660 %"
#define OPTION_VAL_670_BE "670 %"
#define OPTION_VAL_680_BE "680 %"
#define OPTION_VAL_690_BE "690 %"
#define OPTION_VAL_700_BE "700 %"
#define OPTION_VAL_710_BE "710 %"
#define OPTION_VAL_720_BE "720 %"
#define OPTION_VAL_730_BE "730 %"
#define OPTION_VAL_740_BE "740 %"
#define OPTION_VAL_750_BE "750 %"
#define BEETLE_OPT_GPU_OVERCLOCK_LABEL_BE "Разгон GPU-растэрызатара"
#define BEETLE_OPT_GPU_OVERCLOCK_INFO_0_BE "Разганяе 2D-растэрызатар усярэдзіне GPU эмуляванай PSX. Не ўплывае на якасць 3D-рэндэрынгу і ў цэлым не аказвае значны эфект."
#define OPTION_VAL_1X_NATIVE_BE "1x (родная)"
#define OPTION_VAL_2X_BE NULL
#define OPTION_VAL_4X_BE NULL
#define OPTION_VAL_8X_BE NULL
#define OPTION_VAL_16X_BE NULL
#define OPTION_VAL_32X_BE NULL
#define BEETLE_OPT_GTE_OVERCLOCK_LABEL_BE "Разгон GTE"
#define BEETLE_OPT_GTE_OVERCLOCK_INFO_0_BE "Паніжае затрымку выканання аперацый якія эмулююцца GTE (супрацэсарам для 3D-графікі) да аднаго цыклу. Паднімае частату кадраў і стабілізуе час кадра ў гульнях, якія інтэнсіўна выкарыстоўваюць GTE."
#define BEETLE_OPT_CD_ACCESS_METHOD_LABEL_BE NULL
#define BEETLE_OPT_CD_ACCESS_METHOD_INFO_0_BE NULL
#define OPTION_VAL_SYNC_BE "Сінхронны"
#define OPTION_VAL_ASYNC_BE "Асінхронны"
#define OPTION_VAL_PRECACHE_BE "Перадзагрузка"
#define BEETLE_OPT_CD_FASTLOAD_LABEL_BE "Хуткасць чытання CD"
#define BEETLE_OPT_CD_FASTLOAD_INFO_0_BE NULL
#define OPTION_VAL_2X_NATIVE_BE "2x (родная)"
#define OPTION_VAL_6X_BE NULL
#define OPTION_VAL_10X_BE NULL
#define OPTION_VAL_12X_BE "x12"
#define OPTION_VAL_14X_BE "x14"
#define BEETLE_OPT_SKIP_BIOS_LABEL_BE "Прапускаць BIOS"
#define BEETLE_OPT_SKIP_BIOS_INFO_0_BE "Пропуск пачатковай анімацыі BIOS PlayStation, якая адлюстроўваецца пры запуску гульняў. Пры ўключэнні можа ўплываць на сумяшчальнасць з некаторымі гульнямі (Saga Frontier, PAL-гульні з абаронай ад капіявання і інш.)."
#define BEETLE_OPT_OVERRIDE_BIOS_LABEL_BE NULL
#define BEETLE_OPT_OVERRIDE_BIOS_INFO_0_BE NULL
#define OPTION_VAL_PSXONPSP_BE NULL
#define OPTION_VAL_PS1_ROM_BE NULL
#define OPTION_VAL_OPENBIOS_BE NULL
#define BEETLE_OPT_REGION_LABEL_BE "Рэгіён сістэмы"
#define BEETLE_OPT_REGION_INFO_0_BE NULL
#define OPTION_VAL_AUTO_BE "Аўта"
#define OPTION_VAL_NTSC_J_BE NULL
#define OPTION_VAL_NTSC_U_BE NULL
#define OPTION_VAL_PAL_BE NULL
#define BEETLE_OPT_INTERNAL_RESOLUTION_LABEL_BE "Унутраная раздзяляльнасць"
#define BEETLE_OPT_INTERNAL_RESOLUTION_INFO_0_BE NULL
#define BEETLE_OPT_RENDERER_LABEL_BE NULL
#define BEETLE_OPT_RENDERER_INFO_0_BE NULL
#define OPTION_VAL_HARDWARE_BE "Апаратны (аўта)"
#define OPTION_VAL_HARDWARE_GL_BE "Апаратны (OpenGL)"
#define OPTION_VAL_HARDWARE_VK_BE "Апаратны (Vulkan)"
#define OPTION_VAL_SOFTWARE_BE "Праграмнае забеспячэнне"
#define BEETLE_OPT_RENDERER_SOFTWARE_FB_LABEL_BE "Праграмны фрэймбуфер"
#define BEETLE_OPT_RENDERER_SOFTWARE_FB_INFO_0_BE "Уключае для апаратных рэндэраў дакладную эмуляцыю эфектаў фрэймбуфера (напрыклад размыццё руху, закручванне перад боем у FF7), запускаючы фонам копію праграмнага рэндэру ў натыўным раздзяленні. Пры адключэнні налады эфектаў прапускаюцца (OpenGL) або рэалізуюцца сродкамі GPU (Vulkan). Адключэнне падвышае прадукцыйнасць, але прыводзіць да з'яўлення мноства графічных багаў. Калі не ўпэўненыя, пакіньце уключанай."
#define BEETLE_OPT_DEPTH_LABEL_BE "Унутраная глыбіня колеру"
#define BEETLE_OPT_DEPTH_INFO_0_BE NULL
#define OPTION_VAL_16BPP_NATIVE_BE "16 біт/пікс (натыўна)"
#define OPTION_VAL_32BPP_BE NULL
#define BEETLE_OPT_DITHER_MODE_LABEL_BE "Шаблон дызерынгу"
#define BEETLE_OPT_DITHER_MODE_INFO_0_BE NULL
#define OPTION_VAL_INTERNAL_RESOLUTION_BE "Унутраная раздзяляльнасць"
#define BEETLE_OPT_SCALED_UV_OFFSET_LABEL_BE "Зрушэнне тэкстур па UV"
#define BEETLE_OPT_SCALED_UV_OFFSET_INFO_0_BE "Сэмпляваць тэкстуры для 3D-палігонаў са зрушэннем пры ўнутраным раздзяленні вышэй 1x. Памяншае бачнасць швоў тэкстур, але можа выклікаць адвольныя праблемы з графікай."
#define BEETLE_OPT_FILTER_LABEL_BE "Згладжванне тэкстур"
#define BEETLE_OPT_FILTER_INFO_0_BE NULL
#define OPTION_VAL_NEAREST_BE "Найбліжэйшая"
#define OPTION_VAL_SABR_BE NULL
#define OPTION_VAL_XBR_BE NULL
#define OPTION_VAL_BILINEAR_BE "Білінейны"
#define OPTION_VAL_3_POINT_BE "Па трох кропках"
#define OPTION_VAL_JINC2_BE NULL
#define BEETLE_OPT_FILTER_EXCLUDE_SPRITE_LABEL_BE "Выключаць спрайты з фільтрацыі"
#define BEETLE_OPT_FILTER_EXCLUDE_SPRITE_INFO_0_BE "Не ўжываць фільтраванне тэкстур да спрайтаў. Прадухіляе з'яўленне швоў у гульнях з заднім фонам, адмаляваным 2D-спрайтамі. Для максімальнай эфектыўнасці выкарыстоўвайце сумесна з адаптыўным згладжваннем ці іншым фільтрам постапрацоўкі."
#define OPTION_VAL_OPAQUE_BE "Толькі непразрыстыя"
#define OPTION_VAL_ALL_BE "Непразрыстыя і напаўпразрыстыя"
#define BEETLE_OPT_FILTER_EXCLUDE_2D_POLYGON_LABEL_BE "Выключаць 2D-палігоны з фільтрацыі"
#define BEETLE_OPT_FILTER_EXCLUDE_2D_POLYGON_INFO_0_BE "Не ўжываць фільтраванне тэкстур да 2D-палігонаў. 2D-палігоны вызначаюцца эўрыстычна, таму магчымыя скажэнні графікі. Для лепшага эфекту выкарыстоўвайце разам з опцыяй 'Адаптыўнае згладжванне' або іншым фільтрам пост-апрацоўкі."
#define BEETLE_OPT_ADAPTIVE_SMOOTHING_LABEL_BE "Адаптыўнае згладжванне"
#define BEETLE_OPT_ADAPTIVE_SMOOTHING_INFO_0_BE "Згладжвае 2D-элементы графікі і інтэрфейсу без страты выразнасці аб'ектаў у 3D. Толькі для рэндэрынгу на Vulkan."
#define BEETLE_OPT_SUPER_SAMPLING_LABEL_BE "Суперсэмплінг (паніжаць да натыўнага раздзялення)"
#define BEETLE_OPT_SUPER_SAMPLING_INFO_0_BE "Паніжае якасць адмаляванага кантэнту з падвышанага раздзялення да зыходнага. У камбінацыі з падвышаным раздзяленнем налада дазваляе атрымаць графіку ў арыгінальным раздзяленні, але са згладжаным 3D. Найбольш эфектыўна для гульняў, якія выкарыстоўваюць 2D і 3D элементы (напр. 3D-персанажы на адмаляваным заднім фоне), а таксама ў звязку з CRT-шэйдарамі. Падтрымліваецца толькі Vulkаn-рэндэрам. Заўвага: пры ўключэнні рэкамендуецца адключаць опцыю 'Шаблон дызерынгу'."
#define BEETLE_OPT_MSAA_LABEL_BE "Множная выбарка згладжвання"
#define BEETLE_OPT_MSAA_INFO_0_BE NULL
#define OPTION_VAL_1X_BE NULL
#define BEETLE_OPT_MDEC_YUV_LABEL_BE "Згладжванне YUV MDEC"
#define BEETLE_OPT_MDEC_YUV_INFO_0_BE "Паляпшае якасць відэазаставак шляхам памяншэння артэфактаў макраблакавання (квадраты/няроўныя краі). Толькі для рэндэрынгу на Vulkan."
#define BEETLE_OPT_TRACK_TEXTURES_LABEL_BE "Адсочваць тэкстуры"
#define BEETLE_OPT_TRACK_TEXTURES_INFO_0_BE NULL
#define BEETLE_OPT_DUMP_TEXTURES_LABEL_BE "Здымаць дамп тэкстур"
#define BEETLE_OPT_DUMP_TEXTURES_INFO_0_BE "Выгружаць тэкстуры якія выкарыстоўваюцца ў <cd>-texture-dump/"
#define BEETLE_OPT_REPLACE_TEXTURES_LABEL_BE "Замяняць тэкстуры"
#define BEETLE_OPT_REPLACE_TEXTURES_INFO_0_BE "Замяняць тэкстуры HD-копіямі з <cd>-texture-replacements/"
#define BEETLE_OPT_PGXP_MODE_LABEL_BE "Рэжым выканання PGXP"
#define BEETLE_OPT_PGXP_MODE_INFO_0_BE NULL
#define OPTION_VAL_MEMORY_ONLY_BE "Толькі памяць"
#define OPTION_VAL_MEMORY_CPU_BE "Памяць + CPU (нестабільна)"
#define BEETLE_OPT_PGXP_2D_TOL_LABEL_BE "Допуск PGXP для 2D-геаметрыі"
#define BEETLE_OPT_PGXP_2D_TOL_INFO_0_BE "Хавае больш выяўленых памылак PGXP: налада ўсталёўвае дапушчальнае адхіленне значэнняў PGXP для геаметрыі без дакладных дадзеных аб глыбіні."
#define OPTION_VAL_0PX_BE NULL
#define OPTION_VAL_1PX_BE NULL
#define OPTION_VAL_2PX_BE NULL
#define OPTION_VAL_3PX_BE NULL
#define OPTION_VAL_4PX_BE NULL
#define OPTION_VAL_5PX_BE NULL
#define OPTION_VAL_6PX_BE NULL
#define OPTION_VAL_7PX_BE NULL
#define OPTION_VAL_8PX_BE NULL
#define BEETLE_OPT_PGXP_NCLIP_LABEL_BE "Адбракоўка прымітываў PGXP"
#define BEETLE_OPT_PGXP_NCLIP_INFO_0_BE "Выкарыстоўваць PGXP-рэалізацыю NCLIP. Паляпшае графіку за кошт памяншэння правалаў у геаметрыі з дапамогай PGXP-каардынат. У некаторых выпадках можа прыводзіць да завісання гульняў."
#define BEETLE_OPT_PGXP_VERTEX_LABEL_BE "Вяршынны кэш PGXP"
#define BEETLE_OPT_PGXP_VERTEX_INFO_0_BE "Кэшаваць каардынаты вяршыняў, палепшаных з PGXP дзеля паўторнага выкарыстання пры адмалёўцы палігонаў. Паляпшае выраўноўванне аб'ектаў і памяншае бачнасць швоў пры рэндэрынгу тэкстур, але праз памылкі пры апытанні кэша можа выклікаць збоі графікі. Рэкамендуецца пакідаць выклычанай. Дадзеная опцыя прымяняецца толькі калі ўключаны 'Рэжым выканання PGXP'. Падтрымліваецца толькі апаратнымі рэндэрамі."
#define BEETLE_OPT_PGXP_TEXTURE_LABEL_BE "Перспектыўна-карэктнае тэкстураванне PGXP"
#define BEETLE_OPT_PGXP_TEXTURE_INFO_0_BE NULL
#define BEETLE_OPT_DISPLAY_INTERNAL_FPS_LABEL_BE "Паказваць частату кадраў"
#define BEETLE_OPT_DISPLAY_INTERNAL_FPS_INFO_0_BE "Адлюстроўваць унутраную частату кадраў, з якой эмуляваная PlayStation ажыццяўляе рэндэрынг. Заўвага: патрабуецца ўключэнне апавяшчэнняў у наладах франтэнда."
#define BEETLE_OPT_DISPLAY_OSD_LABEL_BE "Адлюстраванне паведамленняў"
#define BEETLE_OPT_DISPLAY_OSD_INFO_0_BE "Адлюстроўваць на экране паведамленні, утвораныя ядром."
#define BEETLE_OPT_DISPLAY_VRAM_LABEL_BE "Адлюстроўваць усю VRAM (наладка)"
#define BEETLE_OPT_DISPLAY_VRAM_INFO_0_BE "Выводзіць на экран поўнае змесціва відэапамяці якая эмулюецца кансоллю. Падтрымліваецца толькі для апаратных рэндэраў OpenGL і Vulkan. Выкарыстоўваецца для адладкі і ў звычайным рэжыме павінна быць адключана."
#define BEETLE_OPT_ANALOG_CALIBRATION_LABEL_BE "Аўтакаліброўка аналагавага ўводу"
#define BEETLE_OPT_ANALOG_CALIBRATION_INFO_0_BE NULL
#define BEETLE_OPT_ANALOG_TOGGLE_LABEL_BE "Перамыкач рэжыму DualShock"
#define BEETLE_OPT_ANALOG_TOGGLE_INFO_0_BE "Пры выбары прылады ўводу тыпу DualShock дадзеная опцыя дазваляе перамыкаць DualShock які эмулюецца паміж лічбавым і аналагавым рэжымамі, як на рэальнай кансолі. Рэжым таксама можа быць зменены адной з камбінацый кнопак."
#define OPTION_VAL_ENABLED_ANALOG_BE "Па змаўчанні-аналагавы"
#define BEETLE_OPT_ANALOG_TOGGLE_COMBO_LABEL_BE "Звязак для змены рэжыму DualShock"
#define BEETLE_OPT_ANALOG_TOGGLE_COMBO_INFO_0_BE NULL
#define OPTION_VAL_L1_L2_R1_R2_START_SELECT_BE NULL
#define OPTION_VAL_L1_R1_SELECT_BE NULL
#define OPTION_VAL_L1_R1_START_BE NULL
#define OPTION_VAL_L1_R1_L3_BE NULL
#define OPTION_VAL_L1_R1_R3_BE NULL
#define OPTION_VAL_L2_R2_SELECT_BE NULL
#define OPTION_VAL_L2_R2_START_BE NULL
#define OPTION_VAL_L2_R2_L3_BE NULL
#define OPTION_VAL_L2_R2_R3_BE NULL
#define OPTION_VAL_L3_R3_BE NULL
#define BEETLE_OPT_ANALOG_TOGGLE_HOLD_LABEL_BE "Затрымка націску для змены рэжыму DualShock"
#define BEETLE_OPT_ANALOG_TOGGLE_HOLD_INFO_0_BE NULL
#define OPTION_VAL_0_BE "Затрымка 0 секунд"
#define OPTION_VAL_1_BE "Затрымка 1 секунда"
#define OPTION_VAL_2_BE "Затрымка 2 секунды"
#define OPTION_VAL_3_BE "Затрымка 3 секунды"
#define OPTION_VAL_4_BE "Затрымка 4 секунды"
#define OPTION_VAL_5_BE "Затрымка 5 секунд"
#define BEETLE_OPT_ENABLE_MULTITAP_PORT1_LABEL_BE NULL
#define BEETLE_OPT_ENABLE_MULTITAP_PORT1_INFO_0_BE "Падлучае пашырэнне multitap да парта 1."
#define BEETLE_OPT_ENABLE_MULTITAP_PORT2_LABEL_BE NULL
#define BEETLE_OPT_ENABLE_MULTITAP_PORT2_INFO_0_BE "Падлучае пашырэнне multitap да парта 2."
#define BEETLE_OPT_GUN_INPUT_MODE_LABEL_BE "Рэжым светлавога пісталета"
#define BEETLE_OPT_GUN_INPUT_MODE_INFO_0_BE NULL
#define OPTION_VAL_LIGHTGUN_BE "Светлавы пісталет"
#define OPTION_VAL_TOUCHSCREEN_BE "Сэнсарны экран"
#define BEETLE_OPT_GUN_CURSOR_LABEL_BE "Курсор пісталета"
#define BEETLE_OPT_GUN_CURSOR_INFO_0_BE NULL
#define OPTION_VAL_CROSS_BE "Крыж"
#define OPTION_VAL_DOT_BE "Кропка"
#define BEETLE_OPT_CROSSHAIR_COLOR_P1_LABEL_BE NULL
#define BEETLE_OPT_CROSSHAIR_COLOR_P1_INFO_0_BE NULL
#define OPTION_VAL_RED_BE "Чырвоная"
#define OPTION_VAL_BLUE_BE "Блакітная"
#define OPTION_VAL_GREEN_BE "Зялёная"
#define OPTION_VAL_ORANGE_BE "Аранжавы"
#define OPTION_VAL_YELLOW_BE "Жоўтая"
#define OPTION_VAL_CYAN_BE "Марской хвалі"
#define OPTION_VAL_PINK_BE "Ружовы"
#define OPTION_VAL_PURPLE_BE "Фіялетавы"
#define OPTION_VAL_BLACK_BE "Чорны"
#define OPTION_VAL_WHITE_BE "Белы"
#define BEETLE_OPT_CROSSHAIR_COLOR_P2_LABEL_BE NULL
#define BEETLE_OPT_CROSSHAIR_COLOR_P2_INFO_0_BE NULL
#define BEETLE_OPT_MOUSE_SENSITIVITY_LABEL_BE "Чуласць Мышы"
#define BEETLE_OPT_MOUSE_SENSITIVITY_INFO_0_BE NULL
#define OPTION_VAL_5_O42_BE NULL
#define OPTION_VAL_10_BE "10 %"
#define OPTION_VAL_15_BE "15 %"
#define OPTION_VAL_20_BE "20 %"
#define OPTION_VAL_25_BE "25 %"
#define OPTION_VAL_30_BE "30 %"
#define OPTION_VAL_35_BE "35 %"
#define OPTION_VAL_40_BE "40 %"
#define OPTION_VAL_45_BE "45 %"
#define OPTION_VAL_55_BE "55 %"
#define OPTION_VAL_65_BE "65 %"
#define OPTION_VAL_75_BE "75 %"
#define OPTION_VAL_85_BE "85 %"
#define OPTION_VAL_95_BE "95 %"
#define OPTION_VAL_100_O42_BE "100 %"
#define OPTION_VAL_105_BE "105 %"
#define OPTION_VAL_115_BE "115 %"
#define OPTION_VAL_125_BE "125 %"
#define OPTION_VAL_135_BE "135 %"
#define OPTION_VAL_145_BE "145 %"
#define OPTION_VAL_155_BE "155 %"
#define OPTION_VAL_165_BE "165 %"
#define OPTION_VAL_175_BE "175 %"
#define OPTION_VAL_185_BE "185 %"
#define OPTION_VAL_195_BE "195 %"
#define BEETLE_OPT_NEGCON_RESPONSE_LABEL_BE "Водгук скручвання neGcon"
#define BEETLE_OPT_NEGCON_RESPONSE_INFO_0_BE NULL
#define OPTION_VAL_LINEAR_BE "Лінейны"
#define OPTION_VAL_QUADRATIC_BE "Квадратычны"
#define OPTION_VAL_CUBIC_BE "Кубічны"
#define BEETLE_OPT_NEGCON_DEADZONE_LABEL_BE "Мёртвая зона скручвання neGcon"
#define BEETLE_OPT_NEGCON_DEADZONE_INFO_0_BE NULL
#define OPTION_VAL_0_O44_BE NULL
#define BEETLE_OPT_USE_MEDNAFEN_MEMCARD0_METHOD_LABEL_BE NULL
#define BEETLE_OPT_USE_MEDNAFEN_MEMCARD0_METHOD_INFO_0_BE NULL
#define OPTION_VAL_LIBRETRO_BE NULL
#define OPTION_VAL_MEDNAFEN_BE NULL
#define BEETLE_OPT_ENABLE_MEMCARD1_LABEL_BE NULL
#define BEETLE_OPT_ENABLE_MEMCARD1_INFO_0_BE NULL
#define BEETLE_OPT_SHARED_MEMORY_CARDS_LABEL_BE NULL
#define BEETLE_OPT_SHARED_MEMORY_CARDS_INFO_0_BE NULL
#define BEETLE_OPT_MEMCARD_LEFT_INDEX_LABEL_BE NULL
#define BEETLE_OPT_MEMCARD_LEFT_INDEX_INFO_0_BE NULL
#define OPTION_VAL_0_O48_BE NULL
#define BEETLE_OPT_MEMCARD_RIGHT_INDEX_LABEL_BE NULL
#define BEETLE_OPT_MEMCARD_RIGHT_INDEX_INFO_0_BE NULL
#define OPTION_VAL_1_O49_BE NULL
#define BEETLE_OPT_LINE_RENDER_LABEL_BE "Хак лінія-у-чатырохкутнік"
#define BEETLE_OPT_LINE_RENDER_INFO_0_BE NULL
#define OPTION_VAL_DEFAULT_BE "Прадвызначана"
#define OPTION_VAL_AGGRESSIVE_BE "Агрэсіўна"
#define BEETLE_OPT_WIDESCREEN_HACK_LABEL_BE "Шырокафарматны хак"
#define BEETLE_OPT_WIDESCREEN_HACK_INFO_0_BE "Адмалёўваць 3D анаморфна і выводзіць змесціва фрэймбуфера ў шырокафарматным рэжыме. Забяспечвае найлепшыя вынікі з гульнямі ў поўным 3D. Элементы 2D будуць выцягнуты па гарызанталі і могуць перамяшчацца."
#define BEETLE_OPT_WIDESCREEN_HACK_ASPECT_RATIO_LABEL_BE "Прапорцыі экрана з шырокафарматным хакам"
#define BEETLE_OPT_WIDESCREEN_HACK_ASPECT_RATIO_INFO_0_BE NULL
#define OPTION_VAL_16_9_BE NULL
#define OPTION_VAL_16_10_BE NULL
#define OPTION_VAL_18_9_BE NULL
#define OPTION_VAL_19_9_BE NULL
#define OPTION_VAL_20_9_BE NULL
#define OPTION_VAL_21_9_BE NULL
#define OPTION_VAL_32_9_BE NULL
#define BEETLE_OPT_CPU_DYNAREC_LABEL_BE "Дынамічны рэкампілятар CPU"
#define BEETLE_OPT_CPU_DYNAREC_INFO_0_BE "Дынамічна пераўтварае інструкцыі CPU ў натыўныя. Істотна хутчэй інтэрпрэтатара, але забяспечвае меншую дакладнасць таймінгаў CPU і можа выклікаць памылкі."
#define OPTION_VAL_DISABLED_BE "Адключана (інтэрпрэтатар Beetle)"
#define OPTION_VAL_EXECUTE_BE "Макс. прадукцыйнасць"
#define OPTION_VAL_EXECUTE_ONE_BE "Праверка таймінгаў такту"
#define OPTION_VAL_RUN_INTERPRETER_BE "Інтэрпрэтатар Lightrec"
#define BEETLE_OPT_DYNAREC_INVALIDATE_LABEL_BE "Інвалідацыя кода dynarec"
#define BEETLE_OPT_DYNAREC_INVALIDATE_INFO_0_BE "Для некаторых гульняў патрабуецца 'Поўная' інвалідацыя, для іншых 'Толькі DMA'."
#define OPTION_VAL_FULL_BE "Цалкам"
#define OPTION_VAL_DMA_BE "Толькі DMA (крыху хутчэй)"
#define BEETLE_OPT_DYNAREC_EVENTCYCLES_LABEL_BE "Цыклы падзей DMA/MDEC/GPU у dynarec"
#define BEETLE_OPT_DYNAREC_EVENTCYCLES_INFO_0_BE "Максімальная колькасць цыклаў, якія выконваюцца працэсарам да праверкі абнаўлення таймераў/GPU/DMA/MDEC. Больш высокія значэнні падвышаюць хуткасць. Для інтэрпрэтатара beetle эфект менш выяўлены, чым для dynarec."
#define OPTION_VAL_128_BE "128 (па змаўчанні)"
#define BEETLE_OPT_DYNAREC_SPU_SAMPLES_LABEL_BE "Сэмплы SPU у dynarec"
#define BEETLE_OPT_DYNAREC_SPU_SAMPLES_INFO_0_BE "Максімальная колькасць сэмплаў, якія апрацоўваюцца да праверкі абнаўлення SPU. Больш высокія значэнні павялічваюць хуткасць, але ў некаторых гульнях адхіленне ад 1 прыводзіць да скажэння гуку."
#define BEETLE_OPT_CORE_TIMING_FPS_LABEL_BE "Таймінг частаты кадраў ядра"
#define BEETLE_OPT_CORE_TIMING_FPS_INFO_0_BE NULL
#define OPTION_VAL_AUTO_TOGGLE_BE NULL
#define OPTION_VAL_FORCE_PROGRESSIVE_BE NULL
#define OPTION_VAL_FORCE_INTERLACED_BE "Прымусовая празрадковая"
#define BEETLE_OPT_DEINTERLACER_LABEL_BE NULL
#define BEETLE_OPT_DEINTERLACER_INFO_0_BE NULL
#define OPTION_VAL_WEAVE_BE NULL
#define OPTION_VAL_BOB_BE NULL
#define OPTION_VAL_BOB_OFFSET_BE NULL
#define OPTION_VAL_FASTMAD_BE NULL
#define BEETLE_OPT_FRAME_DUPING_LABEL_BE NULL
#define BEETLE_OPT_FRAME_DUPING_INFO_0_BE NULL
#define BEETLE_OPT_PAL_VIDEO_TIMING_OVERRIDE_LABEL_BE NULL
#define BEETLE_OPT_PAL_VIDEO_TIMING_OVERRIDE_INFO_0_BE NULL
#define BEETLE_OPT_ASPECT_RATIO_LABEL_BE "Суадносіны бакоў ядра"
#define BEETLE_OPT_ASPECT_RATIO_INFO_0_BE NULL
#define OPTION_VAL_CORRECTED_BE "Скарэктаванае"
#define OPTION_VAL_UNCORRECTED_BE "Без карэкцыі"
#define OPTION_VAL_4_3_BE "Прымусова 4:3"
#define OPTION_VAL_NTSC_BE "Прымусова NTSC"
#define BEETLE_OPT_CROP_OVERSCAN_LABEL_BE "Кадраванне вылетаў разгорткі"
#define BEETLE_OPT_CROP_OVERSCAN_INFO_0_BE NULL
#define OPTION_VAL_STATIC_BE "Гарызантальная"
#define OPTION_VAL_SMART_BE NULL
#define BEETLE_OPT_IMAGE_CROP_LABEL_BE "Дадатковае кадраванне"
#define BEETLE_OPT_IMAGE_CROP_INFO_0_BE NULL
#define OPTION_VAL_9PX_BE NULL
#define OPTION_VAL_10PX_BE NULL
#define OPTION_VAL_11PX_BE NULL
#define OPTION_VAL_12PX_BE NULL
#define OPTION_VAL_13PX_BE NULL
#define OPTION_VAL_14PX_BE NULL
#define OPTION_VAL_15PX_BE NULL
#define OPTION_VAL_16PX_BE NULL
#define OPTION_VAL_17PX_BE NULL
#define OPTION_VAL_18PX_BE NULL
#define OPTION_VAL_19PX_BE NULL
#define OPTION_VAL_20PX_BE NULL
#define BEETLE_OPT_IMAGE_OFFSET_LABEL_BE "Зрух малюнка пасля абрэзкі"
#define BEETLE_OPT_IMAGE_OFFSET_INFO_0_BE NULL
#define OPTION_VAL_12PX_O64_BE NULL
#define OPTION_VAL_11PX_O64_BE NULL
#define OPTION_VAL_10PX_O64_BE NULL
#define OPTION_VAL_9PX_O64_BE NULL
#define OPTION_VAL_8PX_O64_BE NULL
#define OPTION_VAL_7PX_O64_BE NULL
#define OPTION_VAL_6PX_O64_BE NULL
#define OPTION_VAL_5PX_O64_BE NULL
#define OPTION_VAL_4PX_O64_BE NULL
#define OPTION_VAL_3PX_O64_BE NULL
#define OPTION_VAL_2PX_O64_BE NULL
#define OPTION_VAL_1PX_O64_BE NULL
#define OPTION_VAL_1PX_O64_0_BE NULL
#define OPTION_VAL_2PX_O64_0_BE NULL
#define OPTION_VAL_3PX_O64_0_BE NULL
#define OPTION_VAL_4PX_O64_0_BE NULL
#define OPTION_VAL_5PX_O64_0_BE NULL
#define OPTION_VAL_6PX_O64_0_BE NULL
#define OPTION_VAL_7PX_O64_0_BE NULL
#define OPTION_VAL_8PX_O64_0_BE NULL
#define OPTION_VAL_9PX_O64_0_BE NULL
#define OPTION_VAL_10PX_O64_0_BE NULL
#define OPTION_VAL_11PX_O64_0_BE NULL
#define OPTION_VAL_12PX_O64_0_BE NULL
#define BEETLE_OPT_IMAGE_OFFSET_CYCLES_LABEL_BE "Зрух выявы па гарызанталі (у тактах GPU)"
#define BEETLE_OPT_IMAGE_OFFSET_CYCLES_INFO_0_BE NULL
#define BEETLE_OPT_INITIAL_SCANLINE_LABEL_BE "Пачатковы радок разгорткі - NTSC"
#define BEETLE_OPT_INITIAL_SCANLINE_INFO_0_BE NULL
#define BEETLE_OPT_LAST_SCANLINE_LABEL_BE "Апошні радок разгорткі - NTSC"
#define BEETLE_OPT_LAST_SCANLINE_INFO_0_BE NULL
#define OPTION_VAL_239_BE "239 (прадвызначана)"
#define BEETLE_OPT_INITIAL_SCANLINE_PAL_LABEL_BE "Пачатковы радок разгорткі - NTSC"
#define BEETLE_OPT_INITIAL_SCANLINE_PAL_INFO_0_BE NULL
#define BEETLE_OPT_LAST_SCANLINE_PAL_LABEL_BE "Апошні радок разгорткі - PAL"
#define BEETLE_OPT_LAST_SCANLINE_PAL_INFO_0_BE NULL
#define OPTION_VAL_287_BE "287 (прадвызначана)"

struct retro_core_option_v2_category option_cats_be[] = {
   {
      "system",
      CATEGORY_SYSTEM_LABEL_BE,
      CATEGORY_SYSTEM_INFO_0_BE
   },
   {
      "video",
      CATEGORY_VIDEO_LABEL_BE,
      CATEGORY_VIDEO_INFO_0_BE
   },
   {
      "pgxp",
      CATEGORY_PGXP_LABEL_BE,
      CATEGORY_PGXP_INFO_0_BE
   },
   {
      "osd",
      CATEGORY_OSD_LABEL_BE,
      CATEGORY_OSD_INFO_0_BE
   },
   {
      "input",
      CATEGORY_INPUT_LABEL_BE,
      CATEGORY_INPUT_INFO_0_BE
   },
   {
      "memcards",
      CATEGORY_MEMCARDS_LABEL_BE,
      CATEGORY_MEMCARDS_INFO_0_BE
   },
   {
      "hacks",
      CATEGORY_HACKS_LABEL_BE,
      CATEGORY_HACKS_INFO_0_BE
   },
   { NULL, NULL, NULL },
};
struct retro_core_option_v2_definition option_defs_be[] = {
   {
      BEETLE_OPT(cpu_freq_scale),
      BEETLE_OPT_CPU_FREQ_SCALE_LABEL_BE,
      NULL,
      BEETLE_OPT_CPU_FREQ_SCALE_INFO_0_BE,
      NULL,
      "system",
      {
         { "50%",           OPTION_VAL_50_BE },
         { "60%",           OPTION_VAL_60_BE },
         { "70%",           OPTION_VAL_70_BE },
         { "80%",           OPTION_VAL_80_BE },
         { "90%",           OPTION_VAL_90_BE },
         { "100%",          OPTION_VAL_100_BE },
         { "110%",          OPTION_VAL_110_BE },
         { "120%",          OPTION_VAL_120_BE },
         { "130%",          OPTION_VAL_130_BE },
         { "140%",          OPTION_VAL_140_BE },
         { "150%",          OPTION_VAL_150_BE },
         { "160%",          OPTION_VAL_160_BE },
         { "170%",          OPTION_VAL_170_BE },
         { "180%",          OPTION_VAL_180_BE },
         { "190%",          OPTION_VAL_190_BE },
         { "200%",          OPTION_VAL_200_BE },
         { "210%",          OPTION_VAL_210_BE },
         { "220%",          OPTION_VAL_220_BE },
         { "230%",          OPTION_VAL_230_BE },
         { "240%",          OPTION_VAL_240_BE },
         { "250%",          OPTION_VAL_250_BE },
         { "260%",          OPTION_VAL_260_BE },
         { "270%",          OPTION_VAL_270_BE },
         { "280%",          OPTION_VAL_280_BE },
         { "290%",          OPTION_VAL_290_BE },
         { "300%",          OPTION_VAL_300_BE },
         { "310%",          OPTION_VAL_310_BE },
         { "320%",          OPTION_VAL_320_BE },
         { "330%",          OPTION_VAL_330_BE },
         { "340%",          OPTION_VAL_340_BE },
         { "350%",          OPTION_VAL_350_BE },
         { "360%",          OPTION_VAL_360_BE },
         { "370%",          OPTION_VAL_370_BE },
         { "380%",          OPTION_VAL_380_BE },
         { "390%",          OPTION_VAL_390_BE },
         { "400%",          OPTION_VAL_400_BE },
         { "410%",          OPTION_VAL_410_BE },
         { "420%",          OPTION_VAL_420_BE },
         { "430%",          OPTION_VAL_430_BE },
         { "440%",          OPTION_VAL_440_BE },
         { "450%",          OPTION_VAL_450_BE },
         { "460%",          OPTION_VAL_460_BE },
         { "470%",          OPTION_VAL_470_BE },
         { "480%",          OPTION_VAL_480_BE },
         { "490%",          OPTION_VAL_490_BE },
         { "500%",          OPTION_VAL_500_BE },
         { "510%",          OPTION_VAL_510_BE },
         { "520%",          OPTION_VAL_520_BE },
         { "530%",          OPTION_VAL_530_BE },
         { "540%",          OPTION_VAL_540_BE },
         { "550%",          OPTION_VAL_550_BE },
         { "560%",          OPTION_VAL_560_BE },
         { "570%",          OPTION_VAL_570_BE },
         { "580%",          OPTION_VAL_580_BE },
         { "590%",          OPTION_VAL_590_BE },
         { "600%",          OPTION_VAL_600_BE },
         { "610%",          OPTION_VAL_610_BE },
         { "620%",          OPTION_VAL_620_BE },
         { "630%",          OPTION_VAL_630_BE },
         { "640%",          OPTION_VAL_640_BE },
         { "650%",          OPTION_VAL_650_BE },
         { "660%",          OPTION_VAL_660_BE },
         { "670%",          OPTION_VAL_670_BE },
         { "680%",          OPTION_VAL_680_BE },
         { "690%",          OPTION_VAL_690_BE },
         { "700%",          OPTION_VAL_700_BE },
         { "710%",          OPTION_VAL_710_BE },
         { "720%",          OPTION_VAL_720_BE },
         { "730%",          OPTION_VAL_730_BE },
         { "740%",          OPTION_VAL_740_BE },
         { "750%",          OPTION_VAL_750_BE },
         { NULL, NULL },
      },
      "100%"
   },
   {
      BEETLE_OPT(gpu_overclock),
      BEETLE_OPT_GPU_OVERCLOCK_LABEL_BE,
      NULL,
      BEETLE_OPT_GPU_OVERCLOCK_INFO_0_BE,
      NULL,
      "system",
      {
         { "1x(native)", OPTION_VAL_1X_NATIVE_BE },
         { "2x",         OPTION_VAL_2X_BE },
         { "4x",         OPTION_VAL_4X_BE },
         { "8x",         OPTION_VAL_8X_BE },
         { "16x",        OPTION_VAL_16X_BE },
         { "32x",        OPTION_VAL_32X_BE },
         { NULL, NULL },
      },
      "1x(native)"
   },
   {
      BEETLE_OPT(gte_overclock),
      BEETLE_OPT_GTE_OVERCLOCK_LABEL_BE,
      NULL,
      BEETLE_OPT_GTE_OVERCLOCK_INFO_0_BE,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
#ifndef EMSCRIPTEN
   {
      BEETLE_OPT(cd_access_method),
      BEETLE_OPT_CD_ACCESS_METHOD_LABEL_BE,
      NULL,
      BEETLE_OPT_CD_ACCESS_METHOD_INFO_0_BE,
      NULL,
      "system",
      {
         { "sync",     OPTION_VAL_SYNC_BE },
         { "async",    OPTION_VAL_ASYNC_BE },
         { "precache", OPTION_VAL_PRECACHE_BE },
         { NULL, NULL },
      },
      "sync"
   },
#endif
   {
      BEETLE_OPT(cd_fastload),
      BEETLE_OPT_CD_FASTLOAD_LABEL_BE,
      NULL,
      BEETLE_OPT_CD_FASTLOAD_INFO_0_BE,
      NULL,
      "system",
      {
         { "2x(native)", OPTION_VAL_2X_NATIVE_BE },
         { "4x",          OPTION_VAL_4X_BE },
         { "6x",          OPTION_VAL_6X_BE },
         { "8x",          OPTION_VAL_8X_BE },
         { "10x",         OPTION_VAL_10X_BE },
         { "12x",         OPTION_VAL_12X_BE },
         { "14x",         OPTION_VAL_14X_BE },
         { NULL, NULL },
      },
      "2x(native)"
   },
   {
      BEETLE_OPT(skip_bios),
      BEETLE_OPT_SKIP_BIOS_LABEL_BE,
      NULL,
      BEETLE_OPT_SKIP_BIOS_INFO_0_BE,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(override_bios),
      BEETLE_OPT_OVERRIDE_BIOS_LABEL_BE,
      NULL,
      BEETLE_OPT_OVERRIDE_BIOS_INFO_0_BE,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "psxonpsp", OPTION_VAL_PSXONPSP_BE },
         { "ps1_rom",  OPTION_VAL_PS1_ROM_BE },
         { "openbios", OPTION_VAL_OPENBIOS_BE     },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(region),
      BEETLE_OPT_REGION_LABEL_BE,
      NULL,
      BEETLE_OPT_REGION_INFO_0_BE,
      NULL,
      "system",
      {
         { "auto",   OPTION_VAL_AUTO_BE },
         { "ntsc-j", OPTION_VAL_NTSC_J_BE },
         { "ntsc-u", OPTION_VAL_NTSC_U_BE },
         { "pal",    OPTION_VAL_PAL_BE },
         { NULL, NULL },
      },
      "auto"
   },
   {
      BEETLE_OPT(internal_resolution),
      BEETLE_OPT_INTERNAL_RESOLUTION_LABEL_BE,
      NULL,
      BEETLE_OPT_INTERNAL_RESOLUTION_INFO_0_BE,
      NULL,
      "video",
      {
         { "1x(native)", OPTION_VAL_1X_NATIVE_BE },
         { "2x",         OPTION_VAL_2X_BE },
         { "4x",         OPTION_VAL_4X_BE },
         { "8x",         OPTION_VAL_8X_BE },
         { "16x",        OPTION_VAL_16X_BE },
         { NULL, NULL },
      },
      "1x(native)"
   },
#if defined(HAVE_OPENGL) || defined(HAVE_OPENGLES) || defined(HAVE_VULKAN)
   {
      BEETLE_OPT(renderer),
      BEETLE_OPT_RENDERER_LABEL_BE,
      NULL,
      BEETLE_OPT_RENDERER_INFO_0_BE,
      NULL,
      "video",
      {
         { "hardware",    OPTION_VAL_HARDWARE_BE },
#if defined(HAVE_OPENGL) || defined(HAVE_OPENGLES)
         { "hardware_gl", OPTION_VAL_HARDWARE_GL_BE },
#endif
#if defined(HAVE_VULKAN)
         { "hardware_vk", OPTION_VAL_HARDWARE_VK_BE },
#endif
         { "software",    OPTION_VAL_SOFTWARE_BE },
         { NULL, NULL },
      },
      "hardware"
   },
   {
      BEETLE_OPT(renderer_software_fb),
      BEETLE_OPT_RENDERER_SOFTWARE_FB_LABEL_BE,
      NULL,
      BEETLE_OPT_RENDERER_SOFTWARE_FB_INFO_0_BE,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled"
   },
#endif
#if defined(HAVE_OPENGL) || defined(HAVE_OPENGLES)
   {
      BEETLE_OPT(depth),
      BEETLE_OPT_DEPTH_LABEL_BE,
      NULL,
      BEETLE_OPT_DEPTH_INFO_0_BE,
      NULL,
      "video",
      {
         { "16bpp(native)", OPTION_VAL_16BPP_NATIVE_BE },
         { "32bpp",         OPTION_VAL_32BPP_BE },
         { NULL, NULL },
      },
      "16bpp(native)"
   },
   // Sort of, it's more like 15-bit high color and 24-bit true color for visible output. The alpha channel is used for mask bit. Vulkan renderer uses ABGR1555_555 for 31 bits internal? FMVs are always 24-bit on all renderers like original hardware (BGR888, no alpha)
#endif
   {
      BEETLE_OPT(dither_mode),
      BEETLE_OPT_DITHER_MODE_LABEL_BE,
      NULL,
      BEETLE_OPT_DITHER_MODE_INFO_0_BE,
      NULL,
      "video",
      {
         { "1x(native)",          OPTION_VAL_1X_NATIVE_BE },
         { "internal resolution", OPTION_VAL_INTERNAL_RESOLUTION_BE },
         { "disabled",            NULL },
         { NULL, NULL },
      },
      "1x(native)"
   },
#ifdef HAVE_VULKAN
   {
      BEETLE_OPT(scaled_uv_offset),
      BEETLE_OPT_SCALED_UV_OFFSET_LABEL_BE,
      NULL,
      BEETLE_OPT_SCALED_UV_OFFSET_INFO_0_BE,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled"
   },
#endif
#if defined(HAVE_OPENGL) || defined(HAVE_OPENGLES) || defined(HAVE_VULKAN)
   {
      BEETLE_OPT(filter),
      BEETLE_OPT_FILTER_LABEL_BE,
      NULL,
      BEETLE_OPT_FILTER_INFO_0_BE,
      NULL,
      "video",
      {
         { "nearest",  OPTION_VAL_NEAREST_BE },
         { "SABR",     OPTION_VAL_SABR_BE },
         { "xBR",      OPTION_VAL_XBR_BE },
         { "bilinear", OPTION_VAL_BILINEAR_BE },
         { "3-point",  OPTION_VAL_3_POINT_BE },
         { "JINC2",    OPTION_VAL_JINC2_BE },
         { NULL, NULL },
      },
      "nearest"
   },
#ifdef HAVE_VULKAN
   {
      BEETLE_OPT(filter_exclude_sprite),
      BEETLE_OPT_FILTER_EXCLUDE_SPRITE_LABEL_BE,
      NULL,
      BEETLE_OPT_FILTER_EXCLUDE_SPRITE_INFO_0_BE,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "opaque", OPTION_VAL_OPAQUE_BE },
         { "all", OPTION_VAL_ALL_BE },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(filter_exclude_2d_polygon),
      BEETLE_OPT_FILTER_EXCLUDE_2D_POLYGON_LABEL_BE,
      NULL,
      BEETLE_OPT_FILTER_EXCLUDE_2D_POLYGON_INFO_0_BE,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "opaque", OPTION_VAL_OPAQUE_BE },
         { "all", OPTION_VAL_ALL_BE },
         { NULL, NULL },
      },
      "disabled"
   },
#endif
#endif
#ifdef HAVE_VULKAN
   {
      BEETLE_OPT(adaptive_smoothing),
      BEETLE_OPT_ADAPTIVE_SMOOTHING_LABEL_BE,
      NULL,
      BEETLE_OPT_ADAPTIVE_SMOOTHING_INFO_0_BE,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(super_sampling),
      BEETLE_OPT_SUPER_SAMPLING_LABEL_BE,
      NULL,
      BEETLE_OPT_SUPER_SAMPLING_INFO_0_BE,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(msaa),
      BEETLE_OPT_MSAA_LABEL_BE,
      NULL,
      BEETLE_OPT_MSAA_INFO_0_BE,
      NULL,
      "video",
      {
         { "1x",  OPTION_VAL_1X_BE },
         { "2x",  OPTION_VAL_2X_BE },
         { "4x",  OPTION_VAL_4X_BE },
         { "8x",  OPTION_VAL_8X_BE },
         { "16x", OPTION_VAL_16X_BE },
         { NULL, NULL },
      },
      "1x"
   },
   {
      BEETLE_OPT(mdec_yuv),
      BEETLE_OPT_MDEC_YUV_LABEL_BE,
      NULL,
      BEETLE_OPT_MDEC_YUV_INFO_0_BE,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(track_textures),
      BEETLE_OPT_TRACK_TEXTURES_LABEL_BE,
      NULL,
      BEETLE_OPT_TRACK_TEXTURES_INFO_0_BE,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
#ifdef TEXTURE_DUMPING_ENABLED
   {
      BEETLE_OPT(dump_textures),
      BEETLE_OPT_DUMP_TEXTURES_LABEL_BE,
      NULL,
      BEETLE_OPT_DUMP_TEXTURES_INFO_0_BE,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
#endif
   {
      BEETLE_OPT(replace_textures),
      BEETLE_OPT_REPLACE_TEXTURES_LABEL_BE,
      NULL,
      BEETLE_OPT_REPLACE_TEXTURES_INFO_0_BE,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
#endif
   {
      BEETLE_OPT(pgxp_mode),
      BEETLE_OPT_PGXP_MODE_LABEL_BE,
      NULL,
      BEETLE_OPT_PGXP_MODE_INFO_0_BE,
      NULL,
      "pgxp",
      {
         { "disabled",     NULL },
         { "memory only",  OPTION_VAL_MEMORY_ONLY_BE },
         { "memory + CPU", OPTION_VAL_MEMORY_CPU_BE },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(pgxp_2d_tol),
      BEETLE_OPT_PGXP_2D_TOL_LABEL_BE,
      NULL,
      BEETLE_OPT_PGXP_2D_TOL_INFO_0_BE,
      NULL,
      "pgxp",
      {
         { "disabled", NULL },
         { "0px", OPTION_VAL_0PX_BE },
         { "1px", OPTION_VAL_1PX_BE },
         { "2px", OPTION_VAL_2PX_BE },
         { "3px", OPTION_VAL_3PX_BE },
         { "4px", OPTION_VAL_4PX_BE },
         { "5px", OPTION_VAL_5PX_BE },
         { "6px", OPTION_VAL_6PX_BE },
         { "7px", OPTION_VAL_7PX_BE },
         { "8px", OPTION_VAL_8PX_BE },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(pgxp_nclip),
      BEETLE_OPT_PGXP_NCLIP_LABEL_BE,
      NULL,
      BEETLE_OPT_PGXP_NCLIP_INFO_0_BE,
      NULL,
      "pgxp",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
#if defined(HAVE_OPENGL) || defined(HAVE_OPENGLES) || defined(HAVE_VULKAN)
   {
      BEETLE_OPT(pgxp_vertex),
      BEETLE_OPT_PGXP_VERTEX_LABEL_BE,
      NULL,
      BEETLE_OPT_PGXP_VERTEX_INFO_0_BE,
      NULL,
      "pgxp",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
#endif
   {
      BEETLE_OPT(pgxp_texture),
      BEETLE_OPT_PGXP_TEXTURE_LABEL_BE,
      NULL,
      BEETLE_OPT_PGXP_TEXTURE_INFO_0_BE,
      NULL,
      "pgxp",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(display_internal_fps),
      BEETLE_OPT_DISPLAY_INTERNAL_FPS_LABEL_BE,
      NULL,
      BEETLE_OPT_DISPLAY_INTERNAL_FPS_INFO_0_BE,
      NULL,
      "osd",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(display_osd),
      BEETLE_OPT_DISPLAY_OSD_LABEL_BE,
      NULL,
      BEETLE_OPT_DISPLAY_OSD_INFO_0_BE,
      NULL,
      "osd",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled"
   },
#if defined(HAVE_OPENGL) || defined(HAVE_OPENGLES) || defined(HAVE_VULKAN)
   {
      BEETLE_OPT(display_vram),
      BEETLE_OPT_DISPLAY_VRAM_LABEL_BE,
      NULL,
      BEETLE_OPT_DISPLAY_VRAM_INFO_0_BE,
      NULL,
      "osd",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
#endif
   {
      BEETLE_OPT(analog_calibration),
      BEETLE_OPT_ANALOG_CALIBRATION_LABEL_BE,
      NULL,
      BEETLE_OPT_ANALOG_CALIBRATION_INFO_0_BE,
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(analog_toggle),
      BEETLE_OPT_ANALOG_TOGGLE_LABEL_BE,
      NULL,
      BEETLE_OPT_ANALOG_TOGGLE_INFO_0_BE,
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { "enabled-analog", OPTION_VAL_ENABLED_ANALOG_BE },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(analog_toggle_combo),
      BEETLE_OPT_ANALOG_TOGGLE_COMBO_LABEL_BE,
      NULL,
      BEETLE_OPT_ANALOG_TOGGLE_COMBO_INFO_0_BE,
      NULL,
      "input",
      {
         { "l1+l2+r1+r2+start+select", OPTION_VAL_L1_L2_R1_R2_START_SELECT_BE },
         { "l1+r1+select",             OPTION_VAL_L1_R1_SELECT_BE },
         { "l1+r1+start",              OPTION_VAL_L1_R1_START_BE },
         { "l1+r1+l3",                 OPTION_VAL_L1_R1_L3_BE },
         { "l1+r1+r3",                 OPTION_VAL_L1_R1_R3_BE },
         { "l2+r2+select",             OPTION_VAL_L2_R2_SELECT_BE },
         { "l2+r2+start",              OPTION_VAL_L2_R2_START_BE },
         { "l2+r2+l3",                 OPTION_VAL_L2_R2_L3_BE },
         { "l2+r2+r3",                 OPTION_VAL_L2_R2_R3_BE },
         { "l3+r3",                    OPTION_VAL_L3_R3_BE },
         { NULL, NULL },
      },
      "l1+r1+select"
   },
   {
      BEETLE_OPT(analog_toggle_hold),
      BEETLE_OPT_ANALOG_TOGGLE_HOLD_LABEL_BE,
      NULL,
      BEETLE_OPT_ANALOG_TOGGLE_HOLD_INFO_0_BE,
      NULL,
      "input",
      {
         { "0", OPTION_VAL_0_BE },
         { "1", OPTION_VAL_1_BE },
         { "2", OPTION_VAL_2_BE },
         { "3", OPTION_VAL_3_BE },
         { "4", OPTION_VAL_4_BE },
         { "5", OPTION_VAL_5_BE },
         { NULL, NULL },
      },
      "1"
   },
   {
      BEETLE_OPT(enable_multitap_port1),
      BEETLE_OPT_ENABLE_MULTITAP_PORT1_LABEL_BE,
      NULL,
      BEETLE_OPT_ENABLE_MULTITAP_PORT1_INFO_0_BE,
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(enable_multitap_port2),
      BEETLE_OPT_ENABLE_MULTITAP_PORT2_LABEL_BE,
      NULL,
      BEETLE_OPT_ENABLE_MULTITAP_PORT2_INFO_0_BE,
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(gun_input_mode),
      BEETLE_OPT_GUN_INPUT_MODE_LABEL_BE,
      NULL,
      BEETLE_OPT_GUN_INPUT_MODE_INFO_0_BE,
      NULL,
      "input",
      {
         { "lightgun",    OPTION_VAL_LIGHTGUN_BE },
         { "touchscreen", OPTION_VAL_TOUCHSCREEN_BE },
         { NULL, NULL },
      },
      "lightgun"
   },
   // Shouldn't the gun_input_mode just be Mouse vs. Touchscreen?
   {
      BEETLE_OPT(gun_cursor),
      BEETLE_OPT_GUN_CURSOR_LABEL_BE,
      NULL,
      BEETLE_OPT_GUN_CURSOR_INFO_0_BE,
      NULL,
      "input",
      {
         { "off",   "disabled" },
         { "cross", OPTION_VAL_CROSS_BE },
         { "dot",   OPTION_VAL_DOT_BE },
         { NULL, NULL },
      },
      "cross"
   },
   {
      BEETLE_OPT(crosshair_color_p1),
      BEETLE_OPT_CROSSHAIR_COLOR_P1_LABEL_BE,
      NULL,
      BEETLE_OPT_CROSSHAIR_COLOR_P1_INFO_0_BE,
      NULL,
      "input",
      {
         { "red",    OPTION_VAL_RED_BE },
         { "blue",   OPTION_VAL_BLUE_BE },
         { "green",  OPTION_VAL_GREEN_BE },
         { "orange", OPTION_VAL_ORANGE_BE },
         { "yellow", OPTION_VAL_YELLOW_BE },
         { "cyan",   OPTION_VAL_CYAN_BE },
         { "pink",   OPTION_VAL_PINK_BE },
         { "purple", OPTION_VAL_PURPLE_BE },
         { "black",  OPTION_VAL_BLACK_BE },
         { "white",  OPTION_VAL_WHITE_BE },
         { NULL, NULL },
      },
      "red"
   },
   {
      BEETLE_OPT(crosshair_color_p2),
      BEETLE_OPT_CROSSHAIR_COLOR_P2_LABEL_BE,
      NULL,
      BEETLE_OPT_CROSSHAIR_COLOR_P2_INFO_0_BE,
      NULL,
      "input",
      {
         { "blue",   OPTION_VAL_BLUE_BE },
         { "red",    OPTION_VAL_RED_BE },
         { "green",  OPTION_VAL_GREEN_BE },
         { "orange", OPTION_VAL_ORANGE_BE },
         { "yellow", OPTION_VAL_YELLOW_BE },
         { "cyan",   OPTION_VAL_CYAN_BE },
         { "pink",   OPTION_VAL_PINK_BE },
         { "purple", OPTION_VAL_PURPLE_BE },
         { "black",  OPTION_VAL_BLACK_BE },
         { "white",  OPTION_VAL_WHITE_BE },
         { NULL, NULL },
      },
      "blue"
   },
   {
      BEETLE_OPT(mouse_sensitivity),
      BEETLE_OPT_MOUSE_SENSITIVITY_LABEL_BE,
      NULL,
      BEETLE_OPT_MOUSE_SENSITIVITY_INFO_0_BE,
      NULL,
      "input",
      {
         { "5%",   OPTION_VAL_5_O42_BE },
         { "10%",  OPTION_VAL_10_BE },
         { "15%",  OPTION_VAL_15_BE },
         { "20%",  OPTION_VAL_20_BE },
         { "25%",  OPTION_VAL_25_BE },
         { "30%",  OPTION_VAL_30_BE },
         { "35%",  OPTION_VAL_35_BE },
         { "40%",  OPTION_VAL_40_BE },
         { "45%",  OPTION_VAL_45_BE },
         { "50%",  OPTION_VAL_50_BE },
         { "55%",  OPTION_VAL_55_BE },
         { "60%",  OPTION_VAL_60_BE },
         { "65%",  OPTION_VAL_65_BE },
         { "70%",  OPTION_VAL_70_BE },
         { "75%",  OPTION_VAL_75_BE },
         { "80%",  OPTION_VAL_80_BE },
         { "85%",  OPTION_VAL_85_BE },
         { "90%",  OPTION_VAL_90_BE },
         { "95%",  OPTION_VAL_95_BE },
         { "100%", OPTION_VAL_100_O42_BE },
         { "105%", OPTION_VAL_105_BE },
         { "110%", OPTION_VAL_110_BE },
         { "115%", OPTION_VAL_115_BE },
         { "120%", OPTION_VAL_120_BE },
         { "125%", OPTION_VAL_125_BE },
         { "130%", OPTION_VAL_130_BE },
         { "135%", OPTION_VAL_135_BE },
         { "140%", OPTION_VAL_140_BE },
         { "145%", OPTION_VAL_145_BE },
         { "150%", OPTION_VAL_150_BE },
         { "155%", OPTION_VAL_155_BE },
         { "160%", OPTION_VAL_160_BE },
         { "165%", OPTION_VAL_165_BE },
         { "170%", OPTION_VAL_170_BE },
         { "175%", OPTION_VAL_175_BE },
         { "180%", OPTION_VAL_180_BE },
         { "185%", OPTION_VAL_185_BE },
         { "190%", OPTION_VAL_190_BE },
         { "195%", OPTION_VAL_195_BE },
         { "200%", OPTION_VAL_200_BE },
         { NULL, NULL },
      },
      "100%"
   },
   {
      BEETLE_OPT(negcon_response),
      BEETLE_OPT_NEGCON_RESPONSE_LABEL_BE,
      NULL,
      BEETLE_OPT_NEGCON_RESPONSE_INFO_0_BE,
      NULL,
      "input",
      {
         { "linear",    OPTION_VAL_LINEAR_BE },
         { "quadratic", OPTION_VAL_QUADRATIC_BE },
         { "cubic",     OPTION_VAL_CUBIC_BE },
         { NULL, NULL },
      },
      "linear"
   },
   {
      BEETLE_OPT(negcon_deadzone),
      BEETLE_OPT_NEGCON_DEADZONE_LABEL_BE,
      NULL,
      BEETLE_OPT_NEGCON_DEADZONE_INFO_0_BE,
      NULL,
      "input",
      {
         { "0%",  OPTION_VAL_0_O44_BE },
         { "5%",  OPTION_VAL_5_O42_BE },
         { "10%", OPTION_VAL_10_BE },
         { "15%", OPTION_VAL_15_BE },
         { "20%", OPTION_VAL_20_BE },
         { "25%", OPTION_VAL_25_BE },
         { "30%", OPTION_VAL_30_BE },
         { NULL, NULL },
      },
      "0%"
   },
   {
      BEETLE_OPT(use_mednafen_memcard0_method),
      BEETLE_OPT_USE_MEDNAFEN_MEMCARD0_METHOD_LABEL_BE,
      NULL,
      BEETLE_OPT_USE_MEDNAFEN_MEMCARD0_METHOD_INFO_0_BE,
      NULL,
      "memcards",
      {
         { "libretro", OPTION_VAL_LIBRETRO_BE },
         { "mednafen", OPTION_VAL_MEDNAFEN_BE },
         { NULL, NULL },
      },
      "libretro"
   },
   {
      BEETLE_OPT(enable_memcard1),
      BEETLE_OPT_ENABLE_MEMCARD1_LABEL_BE,
      NULL,
      BEETLE_OPT_ENABLE_MEMCARD1_INFO_0_BE,
      NULL,
      "memcards",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(shared_memory_cards),
      BEETLE_OPT_SHARED_MEMORY_CARDS_LABEL_BE,
      NULL,
      BEETLE_OPT_SHARED_MEMORY_CARDS_INFO_0_BE,
      NULL,
      "memcards",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(memcard_left_index),
      BEETLE_OPT_MEMCARD_LEFT_INDEX_LABEL_BE,
      NULL,
      BEETLE_OPT_MEMCARD_LEFT_INDEX_INFO_0_BE,
      NULL,
      "memcards",
      {
         { "0",  OPTION_VAL_0_O48_BE },
         { "1",  NULL },
         { "2",  NULL },
         { "3",  NULL },
         { "4",  NULL },
         { "5",  NULL },
         { "6",  NULL },
         { "7",  NULL },
         { "8",  NULL },
         { "9",  NULL },
         { "10", NULL },
         { "11", NULL },
         { "12", NULL },
         { "13", NULL },
         { "14", NULL },
         { "15", NULL },
         { "16", NULL },
         { "17", NULL },
         { "18", NULL },
         { "19", NULL },
         { "20", NULL },
         { "21", NULL },
         { "22", NULL },
         { "23", NULL },
         { "24", NULL },
         { "25", NULL },
         { "26", NULL },
         { "27", NULL },
         { "28", NULL },
         { "29", NULL },
         { "30", NULL },
         { "31", NULL },
         { "32", NULL },
         { "33", NULL },
         { "34", NULL },
         { "35", NULL },
         { "36", NULL },
         { "37", NULL },
         { "38", NULL },
         { "39", NULL },
         { "40", NULL },
         { "41", NULL },
         { "42", NULL },
         { "43", NULL },
         { "44", NULL },
         { "45", NULL },
         { "46", NULL },
         { "47", NULL },
         { "48", NULL },
         { "49", NULL },
         { "50", NULL },
         { "51", NULL },
         { "52", NULL },
         { "53", NULL },
         { "54", NULL },
         { "55", NULL },
         { "56", NULL },
         { "57", NULL },
         { "58", NULL },
         { "59", NULL },
         { "60", NULL },
         { "61", NULL },
         { "62", NULL },
         { "63", NULL },
         { NULL, NULL },
      },
      "0"
   },
   {
      BEETLE_OPT(memcard_right_index),
      BEETLE_OPT_MEMCARD_RIGHT_INDEX_LABEL_BE,
      NULL,
      BEETLE_OPT_MEMCARD_RIGHT_INDEX_INFO_0_BE,
      NULL,
      "memcards",
      {
         { "0",  NULL },
         { "1",  OPTION_VAL_1_O49_BE },
         { "2",  NULL },
         { "3",  NULL },
         { "4",  NULL },
         { "5",  NULL },
         { "6",  NULL },
         { "7",  NULL },
         { "8",  NULL },
         { "9",  NULL },
         { "10", NULL },
         { "11", NULL },
         { "12", NULL },
         { "13", NULL },
         { "14", NULL },
         { "15", NULL },
         { "16", NULL },
         { "17", NULL },
         { "18", NULL },
         { "19", NULL },
         { "20", NULL },
         { "21", NULL },
         { "22", NULL },
         { "23", NULL },
         { "24", NULL },
         { "25", NULL },
         { "26", NULL },
         { "27", NULL },
         { "28", NULL },
         { "29", NULL },
         { "30", NULL },
         { "31", NULL },
         { "32", NULL },
         { "33", NULL },
         { "34", NULL },
         { "35", NULL },
         { "36", NULL },
         { "37", NULL },
         { "38", NULL },
         { "39", NULL },
         { "40", NULL },
         { "41", NULL },
         { "42", NULL },
         { "43", NULL },
         { "44", NULL },
         { "45", NULL },
         { "46", NULL },
         { "47", NULL },
         { "48", NULL },
         { "49", NULL },
         { "50", NULL },
         { "51", NULL },
         { "52", NULL },
         { "53", NULL },
         { "54", NULL },
         { "55", NULL },
         { "56", NULL },
         { "57", NULL },
         { "58", NULL },
         { "59", NULL },
         { "60", NULL },
         { "61", NULL },
         { "62", NULL },
         { "63", NULL },
         { NULL, NULL },
      },
      "1"
   },
   {
      BEETLE_OPT(line_render),
      BEETLE_OPT_LINE_RENDER_LABEL_BE,
      NULL,
      BEETLE_OPT_LINE_RENDER_INFO_0_BE,
      NULL,
      "hacks",
      {
         { "default",    OPTION_VAL_DEFAULT_BE },
         { "aggressive", OPTION_VAL_AGGRESSIVE_BE },
         { "disabled",   NULL },
         { NULL, NULL },
      },
      "default"
   },
   {
      BEETLE_OPT(widescreen_hack),
      BEETLE_OPT_WIDESCREEN_HACK_LABEL_BE,
      NULL,
      BEETLE_OPT_WIDESCREEN_HACK_INFO_0_BE,
      NULL,
      "hacks",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(widescreen_hack_aspect_ratio),
      BEETLE_OPT_WIDESCREEN_HACK_ASPECT_RATIO_LABEL_BE,
      NULL,
      BEETLE_OPT_WIDESCREEN_HACK_ASPECT_RATIO_INFO_0_BE,
      NULL,
      "hacks",
      {
         { "16:9",  OPTION_VAL_16_9_BE },
         { "16:10", OPTION_VAL_16_10_BE },
         { "18:9",  OPTION_VAL_18_9_BE },
         { "19:9",  OPTION_VAL_19_9_BE },
         { "20:9",  OPTION_VAL_20_9_BE },
         { "21:9",  OPTION_VAL_21_9_BE }, // 64:27
         { "32:9",  OPTION_VAL_32_9_BE },
         { NULL,    NULL },
      },
      "16:9"
   },
#if defined(HAVE_LIGHTREC)
   {
      BEETLE_OPT(cpu_dynarec),
      BEETLE_OPT_CPU_DYNAREC_LABEL_BE,
      NULL,
      BEETLE_OPT_CPU_DYNAREC_INFO_0_BE,
      NULL,
      "hacks",
      {
         { "disabled",        OPTION_VAL_DISABLED_BE },
         { "execute",         OPTION_VAL_EXECUTE_BE },
         { "execute_one",     OPTION_VAL_EXECUTE_ONE_BE },
         { "run_interpreter", OPTION_VAL_RUN_INTERPRETER_BE },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(dynarec_invalidate),
      BEETLE_OPT_DYNAREC_INVALIDATE_LABEL_BE,
      NULL,
      BEETLE_OPT_DYNAREC_INVALIDATE_INFO_0_BE,
      NULL,
      "hacks",
      {
         { "full", OPTION_VAL_FULL_BE },
         { "dma",  OPTION_VAL_DMA_BE },
         { NULL, NULL },
      },
      "full"
   },
   {
      BEETLE_OPT(dynarec_eventcycles),
      BEETLE_OPT_DYNAREC_EVENTCYCLES_LABEL_BE,
      NULL,
      BEETLE_OPT_DYNAREC_EVENTCYCLES_INFO_0_BE,
      NULL,
      "hacks",
      {
         { "128",  OPTION_VAL_128_BE },
         { "256",  NULL },
         { "384",  NULL },
         { "512",  NULL },
         { "640",  NULL },
         { "768",  NULL },
         { "896",  NULL },
         { "1024", NULL },
         { "1152", NULL },
         { "1280", NULL },
         { "1408", NULL },
         { "1536", NULL },
         { "1664", NULL },
         { "1792", NULL },
         { "1920", NULL },
         { "2048", NULL },
         { NULL, NULL },
      },
      "128"
   },
   {
      BEETLE_OPT(dynarec_spu_samples),
      BEETLE_OPT_DYNAREC_SPU_SAMPLES_LABEL_BE,
      NULL,
      BEETLE_OPT_DYNAREC_SPU_SAMPLES_INFO_0_BE,
      NULL,
      "hacks",
      {
         { "1",  OPTION_VAL_1_O49_BE },
         { "4",  NULL },
         { "16", NULL },
         { NULL, NULL },
      },
      "1"
   },
#endif
   {
      BEETLE_OPT(core_timing_fps),
      BEETLE_OPT_CORE_TIMING_FPS_LABEL_BE,
      NULL,
      BEETLE_OPT_CORE_TIMING_FPS_INFO_0_BE,
      NULL,
      "video",
      {
         { "auto_toggle",       OPTION_VAL_AUTO_TOGGLE_BE },
         { "force_progressive", OPTION_VAL_FORCE_PROGRESSIVE_BE },
         { "force_interlaced",  OPTION_VAL_FORCE_INTERLACED_BE },
      },
      "auto_toggle"
   },
   {
      BEETLE_OPT(deinterlacer),
      BEETLE_OPT_DEINTERLACER_LABEL_BE,
      NULL,
      BEETLE_OPT_DEINTERLACER_INFO_0_BE,
      NULL,
      "video",
      {
         { "weave",      OPTION_VAL_WEAVE_BE },
         { "bob",        OPTION_VAL_BOB_BE },
         { "bob_offset", OPTION_VAL_BOB_OFFSET_BE },
         { "fastmad",    OPTION_VAL_FASTMAD_BE },
         { "off",        "Off" },
         { NULL, NULL },
      },
      "weave"
   },
   {
      BEETLE_OPT(frame_duping),
      BEETLE_OPT_FRAME_DUPING_LABEL_BE,
      NULL,
      BEETLE_OPT_FRAME_DUPING_INFO_0_BE,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(pal_video_timing_override),
      BEETLE_OPT_PAL_VIDEO_TIMING_OVERRIDE_LABEL_BE,
      NULL,
      BEETLE_OPT_PAL_VIDEO_TIMING_OVERRIDE_INFO_0_BE,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(aspect_ratio),
      BEETLE_OPT_ASPECT_RATIO_LABEL_BE,
      NULL,
      BEETLE_OPT_ASPECT_RATIO_INFO_0_BE,
      NULL,
      "video",
      {
         { "corrected", OPTION_VAL_CORRECTED_BE },
         { "uncorrected", OPTION_VAL_UNCORRECTED_BE },
         { "4:3",  OPTION_VAL_4_3_BE },
         { "ntsc", OPTION_VAL_NTSC_BE },
      },
      "corrected"
   },
   {
      BEETLE_OPT(crop_overscan),
      BEETLE_OPT_CROP_OVERSCAN_LABEL_BE,
      NULL,
      BEETLE_OPT_CROP_OVERSCAN_INFO_0_BE,
      NULL,
      "video",
      {
         { "disabled", "Disabled" },
         { "static",   OPTION_VAL_STATIC_BE },
         { "smart",    OPTION_VAL_SMART_BE },
         { NULL, NULL },
      },
      "smart"
   },
   {
      BEETLE_OPT(image_crop),
      BEETLE_OPT_IMAGE_CROP_LABEL_BE,
      NULL,
      BEETLE_OPT_IMAGE_CROP_INFO_0_BE,
      NULL,
      "video",
      {
         { "disabled", "0" },
         { "1px",      OPTION_VAL_1PX_BE },
         { "2px",      OPTION_VAL_2PX_BE },
         { "3px",      OPTION_VAL_3PX_BE },
         { "4px",      OPTION_VAL_4PX_BE },
         { "5px",      OPTION_VAL_5PX_BE },
         { "6px",      OPTION_VAL_6PX_BE },
         { "7px",      OPTION_VAL_7PX_BE },
         { "8px",      OPTION_VAL_8PX_BE },
         { "9px",      OPTION_VAL_9PX_BE },
         { "10px",     OPTION_VAL_10PX_BE },
         { "11px",     OPTION_VAL_11PX_BE },
         { "12px",     OPTION_VAL_12PX_BE },
         { "13px",     OPTION_VAL_13PX_BE },
         { "14px",     OPTION_VAL_14PX_BE },
         { "15px",     OPTION_VAL_15PX_BE },
         { "16px",     OPTION_VAL_16PX_BE },
         { "17px",     OPTION_VAL_17PX_BE },
         { "18px",     OPTION_VAL_18PX_BE },
         { "19px",     OPTION_VAL_19PX_BE },
         { "20px",     OPTION_VAL_20PX_BE },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(image_offset),
      BEETLE_OPT_IMAGE_OFFSET_LABEL_BE,
      NULL,
      BEETLE_OPT_IMAGE_OFFSET_INFO_0_BE,
      NULL,
      "video",
      {
         { "-12px",    OPTION_VAL_12PX_O64_BE },
         { "-11px",    OPTION_VAL_11PX_O64_BE },
         { "-10px",    OPTION_VAL_10PX_O64_BE },
         { "-9px",     OPTION_VAL_9PX_O64_BE },
         { "-8px",     OPTION_VAL_8PX_O64_BE },
         { "-7px",     OPTION_VAL_7PX_O64_BE },
         { "-6px",     OPTION_VAL_6PX_O64_BE },
         { "-5px",     OPTION_VAL_5PX_O64_BE },
         { "-4px",     OPTION_VAL_4PX_O64_BE },
         { "-3px",     OPTION_VAL_3PX_O64_BE },
         { "-2px",     OPTION_VAL_2PX_O64_BE },
         { "-1px",     OPTION_VAL_1PX_O64_BE },
         { "disabled", "0" },
         { "+1px",     OPTION_VAL_1PX_O64_0_BE },
         { "+2px",     OPTION_VAL_2PX_O64_0_BE },
         { "+3px",     OPTION_VAL_3PX_O64_0_BE },
         { "+4px",     OPTION_VAL_4PX_O64_0_BE },
         { "+5px",     OPTION_VAL_5PX_O64_0_BE },
         { "+6px",     OPTION_VAL_6PX_O64_0_BE },
         { "+7px",     OPTION_VAL_7PX_O64_0_BE },
         { "+8px",     OPTION_VAL_8PX_O64_0_BE },
         { "+9px",     OPTION_VAL_9PX_O64_0_BE },
         { "+10px",    OPTION_VAL_10PX_O64_0_BE },
         { "+11px",    OPTION_VAL_11PX_O64_0_BE },
         { "+12px",    OPTION_VAL_12PX_O64_0_BE },
         { NULL, NULL },
      },
      "disabled"
   },
#if defined(HAVE_OPENGL) || defined(HAVE_OPENGLES) || defined(HAVE_VULKAN)
   {
      BEETLE_OPT(image_offset_cycles),
      BEETLE_OPT_IMAGE_OFFSET_CYCLES_LABEL_BE,
      NULL,
      BEETLE_OPT_IMAGE_OFFSET_CYCLES_INFO_0_BE,
      NULL,
      "video",
      {
         { "-40",      NULL },
         { "-39",      NULL },
         { "-38",      NULL },
         { "-37",      NULL },
         { "-36",      NULL },
         { "-35",      NULL },
         { "-34",      NULL },
         { "-33",      NULL },
         { "-32",      NULL },
         { "-31",      NULL },
         { "-30",      NULL },
         { "-29",      NULL },
         { "-28",      NULL },
         { "-27",      NULL },
         { "-26",      NULL },
         { "-25",      NULL },
         { "-24",      NULL },
         { "-23",      NULL },
         { "-22",      NULL },
         { "-21",      NULL },
         { "-20",      NULL },
         { "-19",      NULL },
         { "-18",      NULL },
         { "-17",      NULL },
         { "-16",      NULL },
         { "-15",      NULL },
         { "-14",      NULL },
         { "-13",      NULL },
         { "-12",      NULL },
         { "-11",      NULL },
         { "-10",      NULL },
         { "-9",       NULL },
         { "-8",       NULL },
         { "-7",       NULL },
         { "-6",       NULL },
         { "-5",       NULL },
         { "-4",       NULL },
         { "-3",       NULL },
         { "-2",       NULL },
         { "-1",       NULL },
         { "0",        "0" },
         { "+1",       NULL },
         { "+2",       NULL },
         { "+3",       NULL },
         { "+4",       NULL },
         { "+5",       NULL },
         { "+6",       NULL },
         { "+7",       NULL },
         { "+8",       NULL },
         { "+9",       NULL },
         { "+10",      NULL },
         { "+11",      NULL },
         { "+12",      NULL },
         { "+13",      NULL },
         { "+14",      NULL },
         { "+15",      NULL },
         { "+16",      NULL },
         { "+17",      NULL },
         { "+18",      NULL },
         { "+19",      NULL },
         { "+20",      NULL },
         { "+21",      NULL },
         { "+22",      NULL },
         { "+23",      NULL },
         { "+24",      NULL },
         { "+25",      NULL },
         { "+26",      NULL },
         { "+27",      NULL },
         { "+28",      NULL },
         { "+29",      NULL },
         { "+30",      NULL },
         { "+31",      NULL },
         { "+32",      NULL },
         { "+33",      NULL },
         { "+34",      NULL },
         { "+35",      NULL },
         { "+36",      NULL },
         { "+37",      NULL },
         { "+38",      NULL },
         { "+39",      NULL },
         { "+40",      NULL },
         { NULL, NULL},
      },
      "0"
   },
#endif
   {
      BEETLE_OPT(initial_scanline),
      BEETLE_OPT_INITIAL_SCANLINE_LABEL_BE,
      NULL,
      BEETLE_OPT_INITIAL_SCANLINE_INFO_0_BE,
      NULL,
      "video",
      {
         { "0",  OPTION_VAL_0_O48_BE },
         { "1",  NULL },
         { "2",  NULL },
         { "3",  NULL },
         { "4",  NULL },
         { "5",  NULL },
         { "6",  NULL },
         { "7",  NULL },
         { "8",  NULL },
         { "9",  NULL },
         { "10", NULL },
         { "11", NULL },
         { "12", NULL },
         { "13", NULL },
         { "14", NULL },
         { "15", NULL },
         { "16", NULL },
         { "17", NULL },
         { "18", NULL },
         { "19", NULL },
         { "20", NULL },
         { "21", NULL },
         { "22", NULL },
         { "23", NULL },
         { "24", NULL },
         { "25", NULL },
         { "26", NULL },
         { "27", NULL },
         { "28", NULL },
         { "29", NULL },
         { "30", NULL },
         { "31", NULL },
         { "32", NULL },
         { "33", NULL },
         { "34", NULL },
         { "35", NULL },
         { "36", NULL },
         { "37", NULL },
         { "38", NULL },
         { "39", NULL },
         { "40", NULL },
         { NULL, NULL },
      },
      "0"
   },
   {
      BEETLE_OPT(last_scanline),
      BEETLE_OPT_LAST_SCANLINE_LABEL_BE,
      NULL,
      BEETLE_OPT_LAST_SCANLINE_INFO_0_BE,
      NULL,
      "video",
      {
         { "210", NULL },
         { "211", NULL },
         { "212", NULL },
         { "213", NULL },
         { "214", NULL },
         { "215", NULL },
         { "216", NULL },
         { "217", NULL },
         { "218", NULL },
         { "219", NULL },
         { "220", NULL },
         { "221", NULL },
         { "222", NULL },
         { "223", NULL },
         { "224", NULL },
         { "225", NULL },
         { "226", NULL },
         { "227", NULL },
         { "228", NULL },
         { "229", NULL },
         { "230", NULL },
         { "231", NULL },
         { "232", NULL },
         { "233", NULL },
         { "234", NULL },
         { "235", NULL },
         { "236", NULL },
         { "237", NULL },
         { "238", NULL },
         { "239", OPTION_VAL_239_BE },
         { NULL, NULL },
      },
      "239"
   },
   {
      BEETLE_OPT(initial_scanline_pal),
      BEETLE_OPT_INITIAL_SCANLINE_PAL_LABEL_BE,
      NULL,
      BEETLE_OPT_INITIAL_SCANLINE_PAL_INFO_0_BE,
      NULL,
      "video",
      {
         { "0",  OPTION_VAL_0_O48_BE },
         { "1",  NULL },
         { "2",  NULL },
         { "3",  NULL },
         { "4",  NULL },
         { "5",  NULL },
         { "6",  NULL },
         { "7",  NULL },
         { "8",  NULL },
         { "9",  NULL },
         { "10", NULL },
         { "11", NULL },
         { "12", NULL },
         { "13", NULL },
         { "14", NULL },
         { "15", NULL },
         { "16", NULL },
         { "17", NULL },
         { "18", NULL },
         { "19", NULL },
         { "20", NULL },
         { "21", NULL },
         { "22", NULL },
         { "23", NULL },
         { "24", NULL },
         { "25", NULL },
         { "26", NULL },
         { "27", NULL },
         { "28", NULL },
         { "29", NULL },
         { "30", NULL },
         { "31", NULL },
         { "32", NULL },
         { "33", NULL },
         { "34", NULL },
         { "35", NULL },
         { "36", NULL },
         { "37", NULL },
         { "38", NULL },
         { "39", NULL },
         { "40", NULL },
         { NULL, NULL },
      },
      "0"
   },
   {
      BEETLE_OPT(last_scanline_pal),
      BEETLE_OPT_LAST_SCANLINE_PAL_LABEL_BE,
      NULL,
      BEETLE_OPT_LAST_SCANLINE_PAL_INFO_0_BE,
      NULL,
      "video",
      {
         { "230", NULL },
         { "231", NULL },
         { "232", NULL },
         { "233", NULL },
         { "234", NULL },
         { "235", NULL },
         { "236", NULL },
         { "237", NULL },
         { "238", NULL },
         { "239", NULL },
         { "240", NULL },
         { "241", NULL },
         { "242", NULL },
         { "243", NULL },
         { "244", NULL },
         { "245", NULL },
         { "246", NULL },
         { "247", NULL },
         { "248", NULL },
         { "249", NULL },
         { "250", NULL },
         { "251", NULL },
         { "252", NULL },
         { "253", NULL },
         { "254", NULL },
         { "255", NULL },
         { "256", NULL },
         { "257", NULL },
         { "258", NULL },
         { "259", NULL },
         { "260", NULL },
         { "261", NULL },
         { "262", NULL },
         { "263", NULL },
         { "264", NULL },
         { "265", NULL },
         { "266", NULL },
         { "267", NULL },
         { "268", NULL },
         { "269", NULL },
         { "270", NULL },
         { "271", NULL },
         { "272", NULL },
         { "273", NULL },
         { "274", NULL },
         { "275", NULL },
         { "276", NULL },
         { "277", NULL },
         { "278", NULL },
         { "279", NULL },
         { "280", NULL },
         { "281", NULL },
         { "282", NULL },
         { "283", NULL },
         { "284", NULL },
         { "285", NULL },
         { "286", NULL },
         { "287", OPTION_VAL_287_BE },
         { NULL, NULL },
      },
      "287"
   },
   { NULL, NULL, NULL, NULL, NULL, NULL, {{0}}, NULL },
};
struct retro_core_options_v2 options_be = {
   option_cats_be,
   option_defs_be
};

/* RETRO_LANGUAGE_BG */

#define CATEGORY_SYSTEM_LABEL_BG NULL
#define CATEGORY_SYSTEM_INFO_0_BG NULL
#define CATEGORY_VIDEO_LABEL_BG "Видео"
#define CATEGORY_VIDEO_INFO_0_BG NULL
#define CATEGORY_PGXP_LABEL_BG NULL
#define CATEGORY_PGXP_INFO_0_BG NULL
#define CATEGORY_OSD_LABEL_BG NULL
#define CATEGORY_OSD_INFO_0_BG NULL
#define CATEGORY_INPUT_LABEL_BG "Вход"
#define CATEGORY_INPUT_INFO_0_BG NULL
#define CATEGORY_MEMCARDS_LABEL_BG NULL
#define CATEGORY_MEMCARDS_INFO_0_BG NULL
#define CATEGORY_HACKS_LABEL_BG NULL
#define CATEGORY_HACKS_INFO_0_BG NULL
#define BEETLE_OPT_CPU_FREQ_SCALE_LABEL_BG NULL
#define BEETLE_OPT_CPU_FREQ_SCALE_INFO_0_BG NULL
#define OPTION_VAL_50_BG NULL
#define OPTION_VAL_60_BG NULL
#define OPTION_VAL_70_BG NULL
#define OPTION_VAL_80_BG NULL
#define OPTION_VAL_90_BG NULL
#define OPTION_VAL_100_BG NULL
#define OPTION_VAL_110_BG NULL
#define OPTION_VAL_120_BG NULL
#define OPTION_VAL_130_BG NULL
#define OPTION_VAL_140_BG NULL
#define OPTION_VAL_150_BG NULL
#define OPTION_VAL_160_BG NULL
#define OPTION_VAL_170_BG NULL
#define OPTION_VAL_180_BG NULL
#define OPTION_VAL_190_BG NULL
#define OPTION_VAL_200_BG NULL
#define OPTION_VAL_210_BG NULL
#define OPTION_VAL_220_BG NULL
#define OPTION_VAL_230_BG NULL
#define OPTION_VAL_240_BG NULL
#define OPTION_VAL_250_BG NULL
#define OPTION_VAL_260_BG NULL
#define OPTION_VAL_270_BG NULL
#define OPTION_VAL_280_BG NULL
#define OPTION_VAL_290_BG NULL
#define OPTION_VAL_300_BG NULL
#define OPTION_VAL_310_BG NULL
#define OPTION_VAL_320_BG NULL
#define OPTION_VAL_330_BG NULL
#define OPTION_VAL_340_BG NULL
#define OPTION_VAL_350_BG NULL
#define OPTION_VAL_360_BG NULL
#define OPTION_VAL_370_BG NULL
#define OPTION_VAL_380_BG NULL
#define OPTION_VAL_390_BG NULL
#define OPTION_VAL_400_BG NULL
#define OPTION_VAL_410_BG NULL
#define OPTION_VAL_420_BG NULL
#define OPTION_VAL_430_BG NULL
#define OPTION_VAL_440_BG NULL
#define OPTION_VAL_450_BG NULL
#define OPTION_VAL_460_BG NULL
#define OPTION_VAL_470_BG NULL
#define OPTION_VAL_480_BG NULL
#define OPTION_VAL_490_BG NULL
#define OPTION_VAL_500_BG NULL
#define OPTION_VAL_510_BG NULL
#define OPTION_VAL_520_BG NULL
#define OPTION_VAL_530_BG NULL
#define OPTION_VAL_540_BG NULL
#define OPTION_VAL_550_BG NULL
#define OPTION_VAL_560_BG NULL
#define OPTION_VAL_570_BG NULL
#define OPTION_VAL_580_BG NULL
#define OPTION_VAL_590_BG NULL
#define OPTION_VAL_600_BG NULL
#define OPTION_VAL_610_BG NULL
#define OPTION_VAL_620_BG NULL
#define OPTION_VAL_630_BG NULL
#define OPTION_VAL_640_BG NULL
#define OPTION_VAL_650_BG NULL
#define OPTION_VAL_660_BG NULL
#define OPTION_VAL_670_BG NULL
#define OPTION_VAL_680_BG NULL
#define OPTION_VAL_690_BG NULL
#define OPTION_VAL_700_BG NULL
#define OPTION_VAL_710_BG NULL
#define OPTION_VAL_720_BG NULL
#define OPTION_VAL_730_BG NULL
#define OPTION_VAL_740_BG NULL
#define OPTION_VAL_750_BG NULL
#define BEETLE_OPT_GPU_OVERCLOCK_LABEL_BG NULL
#define BEETLE_OPT_GPU_OVERCLOCK_INFO_0_BG NULL
#define OPTION_VAL_1X_NATIVE_BG NULL
#define OPTION_VAL_2X_BG NULL
#define OPTION_VAL_4X_BG NULL
#define OPTION_VAL_8X_BG NULL
#define OPTION_VAL_16X_BG NULL
#define OPTION_VAL_32X_BG NULL
#define BEETLE_OPT_GTE_OVERCLOCK_LABEL_BG NULL
#define BEETLE_OPT_GTE_OVERCLOCK_INFO_0_BG NULL
#define BEETLE_OPT_CD_ACCESS_METHOD_LABEL_BG NULL
#define BEETLE_OPT_CD_ACCESS_METHOD_INFO_0_BG NULL
#define OPTION_VAL_SYNC_BG NULL
#define OPTION_VAL_ASYNC_BG NULL
#define OPTION_VAL_PRECACHE_BG NULL
#define BEETLE_OPT_CD_FASTLOAD_LABEL_BG NULL
#define BEETLE_OPT_CD_FASTLOAD_INFO_0_BG NULL
#define OPTION_VAL_2X_NATIVE_BG NULL
#define OPTION_VAL_6X_BG NULL
#define OPTION_VAL_10X_BG NULL
#define OPTION_VAL_12X_BG NULL
#define OPTION_VAL_14X_BG NULL
#define BEETLE_OPT_SKIP_BIOS_LABEL_BG NULL
#define BEETLE_OPT_SKIP_BIOS_INFO_0_BG NULL
#define BEETLE_OPT_OVERRIDE_BIOS_LABEL_BG NULL
#define BEETLE_OPT_OVERRIDE_BIOS_INFO_0_BG NULL
#define OPTION_VAL_PSXONPSP_BG NULL
#define OPTION_VAL_PS1_ROM_BG NULL
#define OPTION_VAL_OPENBIOS_BG NULL
#define BEETLE_OPT_REGION_LABEL_BG NULL
#define BEETLE_OPT_REGION_INFO_0_BG NULL
#define OPTION_VAL_AUTO_BG NULL
#define OPTION_VAL_NTSC_J_BG NULL
#define OPTION_VAL_NTSC_U_BG NULL
#define OPTION_VAL_PAL_BG NULL
#define BEETLE_OPT_INTERNAL_RESOLUTION_LABEL_BG NULL
#define BEETLE_OPT_INTERNAL_RESOLUTION_INFO_0_BG NULL
#define BEETLE_OPT_RENDERER_LABEL_BG NULL
#define BEETLE_OPT_RENDERER_INFO_0_BG NULL
#define OPTION_VAL_HARDWARE_BG NULL
#define OPTION_VAL_HARDWARE_GL_BG NULL
#define OPTION_VAL_HARDWARE_VK_BG NULL
#define OPTION_VAL_SOFTWARE_BG NULL
#define BEETLE_OPT_RENDERER_SOFTWARE_FB_LABEL_BG NULL
#define BEETLE_OPT_RENDERER_SOFTWARE_FB_INFO_0_BG NULL
#define BEETLE_OPT_DEPTH_LABEL_BG NULL
#define BEETLE_OPT_DEPTH_INFO_0_BG NULL
#define OPTION_VAL_16BPP_NATIVE_BG NULL
#define OPTION_VAL_32BPP_BG NULL
#define BEETLE_OPT_DITHER_MODE_LABEL_BG NULL
#define BEETLE_OPT_DITHER_MODE_INFO_0_BG NULL
#define OPTION_VAL_INTERNAL_RESOLUTION_BG NULL
#define BEETLE_OPT_SCALED_UV_OFFSET_LABEL_BG NULL
#define BEETLE_OPT_SCALED_UV_OFFSET_INFO_0_BG NULL
#define BEETLE_OPT_FILTER_LABEL_BG NULL
#define BEETLE_OPT_FILTER_INFO_0_BG NULL
#define OPTION_VAL_NEAREST_BG NULL
#define OPTION_VAL_SABR_BG NULL
#define OPTION_VAL_XBR_BG NULL
#define OPTION_VAL_BILINEAR_BG NULL
#define OPTION_VAL_3_POINT_BG NULL
#define OPTION_VAL_JINC2_BG NULL
#define BEETLE_OPT_FILTER_EXCLUDE_SPRITE_LABEL_BG NULL
#define BEETLE_OPT_FILTER_EXCLUDE_SPRITE_INFO_0_BG NULL
#define OPTION_VAL_OPAQUE_BG NULL
#define OPTION_VAL_ALL_BG NULL
#define BEETLE_OPT_FILTER_EXCLUDE_2D_POLYGON_LABEL_BG NULL
#define BEETLE_OPT_FILTER_EXCLUDE_2D_POLYGON_INFO_0_BG NULL
#define BEETLE_OPT_ADAPTIVE_SMOOTHING_LABEL_BG NULL
#define BEETLE_OPT_ADAPTIVE_SMOOTHING_INFO_0_BG NULL
#define BEETLE_OPT_SUPER_SAMPLING_LABEL_BG NULL
#define BEETLE_OPT_SUPER_SAMPLING_INFO_0_BG NULL
#define BEETLE_OPT_MSAA_LABEL_BG NULL
#define BEETLE_OPT_MSAA_INFO_0_BG NULL
#define OPTION_VAL_1X_BG NULL
#define BEETLE_OPT_MDEC_YUV_LABEL_BG NULL
#define BEETLE_OPT_MDEC_YUV_INFO_0_BG NULL
#define BEETLE_OPT_TRACK_TEXTURES_LABEL_BG NULL
#define BEETLE_OPT_TRACK_TEXTURES_INFO_0_BG NULL
#define BEETLE_OPT_DUMP_TEXTURES_LABEL_BG NULL
#define BEETLE_OPT_DUMP_TEXTURES_INFO_0_BG NULL
#define BEETLE_OPT_REPLACE_TEXTURES_LABEL_BG NULL
#define BEETLE_OPT_REPLACE_TEXTURES_INFO_0_BG NULL
#define BEETLE_OPT_PGXP_MODE_LABEL_BG NULL
#define BEETLE_OPT_PGXP_MODE_INFO_0_BG NULL
#define OPTION_VAL_MEMORY_ONLY_BG NULL
#define OPTION_VAL_MEMORY_CPU_BG NULL
#define BEETLE_OPT_PGXP_2D_TOL_LABEL_BG NULL
#define BEETLE_OPT_PGXP_2D_TOL_INFO_0_BG NULL
#define OPTION_VAL_0PX_BG NULL
#define OPTION_VAL_1PX_BG NULL
#define OPTION_VAL_2PX_BG NULL
#define OPTION_VAL_3PX_BG NULL
#define OPTION_VAL_4PX_BG NULL
#define OPTION_VAL_5PX_BG NULL
#define OPTION_VAL_6PX_BG NULL
#define OPTION_VAL_7PX_BG NULL
#define OPTION_VAL_8PX_BG NULL
#define BEETLE_OPT_PGXP_NCLIP_LABEL_BG NULL
#define BEETLE_OPT_PGXP_NCLIP_INFO_0_BG NULL
#define BEETLE_OPT_PGXP_VERTEX_LABEL_BG NULL
#define BEETLE_OPT_PGXP_VERTEX_INFO_0_BG NULL
#define BEETLE_OPT_PGXP_TEXTURE_LABEL_BG NULL
#define BEETLE_OPT_PGXP_TEXTURE_INFO_0_BG NULL
#define BEETLE_OPT_DISPLAY_INTERNAL_FPS_LABEL_BG NULL
#define BEETLE_OPT_DISPLAY_INTERNAL_FPS_INFO_0_BG NULL
#define BEETLE_OPT_DISPLAY_OSD_LABEL_BG NULL
#define BEETLE_OPT_DISPLAY_OSD_INFO_0_BG NULL
#define BEETLE_OPT_DISPLAY_VRAM_LABEL_BG NULL
#define BEETLE_OPT_DISPLAY_VRAM_INFO_0_BG NULL
#define BEETLE_OPT_ANALOG_CALIBRATION_LABEL_BG NULL
#define BEETLE_OPT_ANALOG_CALIBRATION_INFO_0_BG NULL
#define BEETLE_OPT_ANALOG_TOGGLE_LABEL_BG NULL
#define BEETLE_OPT_ANALOG_TOGGLE_INFO_0_BG NULL
#define OPTION_VAL_ENABLED_ANALOG_BG NULL
#define BEETLE_OPT_ANALOG_TOGGLE_COMBO_LABEL_BG NULL
#define BEETLE_OPT_ANALOG_TOGGLE_COMBO_INFO_0_BG NULL
#define OPTION_VAL_L1_L2_R1_R2_START_SELECT_BG NULL
#define OPTION_VAL_L1_R1_SELECT_BG NULL
#define OPTION_VAL_L1_R1_START_BG NULL
#define OPTION_VAL_L1_R1_L3_BG NULL
#define OPTION_VAL_L1_R1_R3_BG NULL
#define OPTION_VAL_L2_R2_SELECT_BG NULL
#define OPTION_VAL_L2_R2_START_BG NULL
#define OPTION_VAL_L2_R2_L3_BG NULL
#define OPTION_VAL_L2_R2_R3_BG NULL
#define OPTION_VAL_L3_R3_BG NULL
#define BEETLE_OPT_ANALOG_TOGGLE_HOLD_LABEL_BG NULL
#define BEETLE_OPT_ANALOG_TOGGLE_HOLD_INFO_0_BG NULL
#define OPTION_VAL_0_BG NULL
#define OPTION_VAL_1_BG NULL
#define OPTION_VAL_2_BG NULL
#define OPTION_VAL_3_BG NULL
#define OPTION_VAL_4_BG NULL
#define OPTION_VAL_5_BG NULL
#define BEETLE_OPT_ENABLE_MULTITAP_PORT1_LABEL_BG NULL
#define BEETLE_OPT_ENABLE_MULTITAP_PORT1_INFO_0_BG NULL
#define BEETLE_OPT_ENABLE_MULTITAP_PORT2_LABEL_BG NULL
#define BEETLE_OPT_ENABLE_MULTITAP_PORT2_INFO_0_BG NULL
#define BEETLE_OPT_GUN_INPUT_MODE_LABEL_BG NULL
#define BEETLE_OPT_GUN_INPUT_MODE_INFO_0_BG NULL
#define OPTION_VAL_LIGHTGUN_BG NULL
#define OPTION_VAL_TOUCHSCREEN_BG NULL
#define BEETLE_OPT_GUN_CURSOR_LABEL_BG NULL
#define BEETLE_OPT_GUN_CURSOR_INFO_0_BG NULL
#define OPTION_VAL_CROSS_BG NULL
#define OPTION_VAL_DOT_BG NULL
#define BEETLE_OPT_CROSSHAIR_COLOR_P1_LABEL_BG NULL
#define BEETLE_OPT_CROSSHAIR_COLOR_P1_INFO_0_BG NULL
#define OPTION_VAL_RED_BG NULL
#define OPTION_VAL_BLUE_BG NULL
#define OPTION_VAL_GREEN_BG NULL
#define OPTION_VAL_ORANGE_BG NULL
#define OPTION_VAL_YELLOW_BG NULL
#define OPTION_VAL_CYAN_BG NULL
#define OPTION_VAL_PINK_BG NULL
#define OPTION_VAL_PURPLE_BG NULL
#define OPTION_VAL_BLACK_BG NULL
#define OPTION_VAL_WHITE_BG NULL
#define BEETLE_OPT_CROSSHAIR_COLOR_P2_LABEL_BG NULL
#define BEETLE_OPT_CROSSHAIR_COLOR_P2_INFO_0_BG NULL
#define BEETLE_OPT_MOUSE_SENSITIVITY_LABEL_BG NULL
#define BEETLE_OPT_MOUSE_SENSITIVITY_INFO_0_BG NULL
#define OPTION_VAL_5_O42_BG NULL
#define OPTION_VAL_10_BG NULL
#define OPTION_VAL_15_BG NULL
#define OPTION_VAL_20_BG NULL
#define OPTION_VAL_25_BG NULL
#define OPTION_VAL_30_BG NULL
#define OPTION_VAL_35_BG NULL
#define OPTION_VAL_40_BG NULL
#define OPTION_VAL_45_BG NULL
#define OPTION_VAL_55_BG NULL
#define OPTION_VAL_65_BG NULL
#define OPTION_VAL_75_BG NULL
#define OPTION_VAL_85_BG NULL
#define OPTION_VAL_95_BG NULL
#define OPTION_VAL_100_O42_BG NULL
#define OPTION_VAL_105_BG NULL
#define OPTION_VAL_115_BG NULL
#define OPTION_VAL_125_BG NULL
#define OPTION_VAL_135_BG NULL
#define OPTION_VAL_145_BG NULL
#define OPTION_VAL_155_BG NULL
#define OPTION_VAL_165_BG NULL
#define OPTION_VAL_175_BG NULL
#define OPTION_VAL_185_BG NULL
#define OPTION_VAL_195_BG NULL
#define BEETLE_OPT_NEGCON_RESPONSE_LABEL_BG NULL
#define BEETLE_OPT_NEGCON_RESPONSE_INFO_0_BG NULL
#define OPTION_VAL_LINEAR_BG NULL
#define OPTION_VAL_QUADRATIC_BG NULL
#define OPTION_VAL_CUBIC_BG NULL
#define BEETLE_OPT_NEGCON_DEADZONE_LABEL_BG NULL
#define BEETLE_OPT_NEGCON_DEADZONE_INFO_0_BG NULL
#define OPTION_VAL_0_O44_BG NULL
#define BEETLE_OPT_USE_MEDNAFEN_MEMCARD0_METHOD_LABEL_BG NULL
#define BEETLE_OPT_USE_MEDNAFEN_MEMCARD0_METHOD_INFO_0_BG NULL
#define OPTION_VAL_LIBRETRO_BG NULL
#define OPTION_VAL_MEDNAFEN_BG NULL
#define BEETLE_OPT_ENABLE_MEMCARD1_LABEL_BG NULL
#define BEETLE_OPT_ENABLE_MEMCARD1_INFO_0_BG NULL
#define BEETLE_OPT_SHARED_MEMORY_CARDS_LABEL_BG NULL
#define BEETLE_OPT_SHARED_MEMORY_CARDS_INFO_0_BG NULL
#define BEETLE_OPT_MEMCARD_LEFT_INDEX_LABEL_BG NULL
#define BEETLE_OPT_MEMCARD_LEFT_INDEX_INFO_0_BG NULL
#define OPTION_VAL_0_O48_BG NULL
#define BEETLE_OPT_MEMCARD_RIGHT_INDEX_LABEL_BG NULL
#define BEETLE_OPT_MEMCARD_RIGHT_INDEX_INFO_0_BG NULL
#define OPTION_VAL_1_O49_BG NULL
#define BEETLE_OPT_LINE_RENDER_LABEL_BG NULL
#define BEETLE_OPT_LINE_RENDER_INFO_0_BG NULL
#define OPTION_VAL_DEFAULT_BG NULL
#define OPTION_VAL_AGGRESSIVE_BG NULL
#define BEETLE_OPT_WIDESCREEN_HACK_LABEL_BG NULL
#define BEETLE_OPT_WIDESCREEN_HACK_INFO_0_BG NULL
#define BEETLE_OPT_WIDESCREEN_HACK_ASPECT_RATIO_LABEL_BG NULL
#define BEETLE_OPT_WIDESCREEN_HACK_ASPECT_RATIO_INFO_0_BG NULL
#define OPTION_VAL_16_9_BG NULL
#define OPTION_VAL_16_10_BG NULL
#define OPTION_VAL_18_9_BG NULL
#define OPTION_VAL_19_9_BG NULL
#define OPTION_VAL_20_9_BG NULL
#define OPTION_VAL_21_9_BG NULL
#define OPTION_VAL_32_9_BG NULL
#define BEETLE_OPT_CPU_DYNAREC_LABEL_BG NULL
#define BEETLE_OPT_CPU_DYNAREC_INFO_0_BG NULL
#define OPTION_VAL_DISABLED_BG NULL
#define OPTION_VAL_EXECUTE_BG NULL
#define OPTION_VAL_EXECUTE_ONE_BG NULL
#define OPTION_VAL_RUN_INTERPRETER_BG NULL
#define BEETLE_OPT_DYNAREC_INVALIDATE_LABEL_BG NULL
#define BEETLE_OPT_DYNAREC_INVALIDATE_INFO_0_BG NULL
#define OPTION_VAL_FULL_BG NULL
#define OPTION_VAL_DMA_BG NULL
#define BEETLE_OPT_DYNAREC_EVENTCYCLES_LABEL_BG NULL
#define BEETLE_OPT_DYNAREC_EVENTCYCLES_INFO_0_BG NULL
#define OPTION_VAL_128_BG NULL
#define BEETLE_OPT_DYNAREC_SPU_SAMPLES_LABEL_BG NULL
#define BEETLE_OPT_DYNAREC_SPU_SAMPLES_INFO_0_BG NULL
#define BEETLE_OPT_CORE_TIMING_FPS_LABEL_BG NULL
#define BEETLE_OPT_CORE_TIMING_FPS_INFO_0_BG NULL
#define OPTION_VAL_AUTO_TOGGLE_BG NULL
#define OPTION_VAL_FORCE_PROGRESSIVE_BG NULL
#define OPTION_VAL_FORCE_INTERLACED_BG NULL
#define BEETLE_OPT_DEINTERLACER_LABEL_BG NULL
#define BEETLE_OPT_DEINTERLACER_INFO_0_BG NULL
#define OPTION_VAL_WEAVE_BG NULL
#define OPTION_VAL_BOB_BG NULL
#define OPTION_VAL_BOB_OFFSET_BG NULL
#define OPTION_VAL_FASTMAD_BG NULL
#define BEETLE_OPT_FRAME_DUPING_LABEL_BG NULL
#define BEETLE_OPT_FRAME_DUPING_INFO_0_BG NULL
#define BEETLE_OPT_PAL_VIDEO_TIMING_OVERRIDE_LABEL_BG NULL
#define BEETLE_OPT_PAL_VIDEO_TIMING_OVERRIDE_INFO_0_BG NULL
#define BEETLE_OPT_ASPECT_RATIO_LABEL_BG NULL
#define BEETLE_OPT_ASPECT_RATIO_INFO_0_BG NULL
#define OPTION_VAL_CORRECTED_BG NULL
#define OPTION_VAL_UNCORRECTED_BG NULL
#define OPTION_VAL_4_3_BG NULL
#define OPTION_VAL_NTSC_BG NULL
#define BEETLE_OPT_CROP_OVERSCAN_LABEL_BG NULL
#define BEETLE_OPT_CROP_OVERSCAN_INFO_0_BG NULL
#define OPTION_VAL_STATIC_BG NULL
#define OPTION_VAL_SMART_BG NULL
#define BEETLE_OPT_IMAGE_CROP_LABEL_BG NULL
#define BEETLE_OPT_IMAGE_CROP_INFO_0_BG NULL
#define OPTION_VAL_9PX_BG NULL
#define OPTION_VAL_10PX_BG NULL
#define OPTION_VAL_11PX_BG NULL
#define OPTION_VAL_12PX_BG NULL
#define OPTION_VAL_13PX_BG NULL
#define OPTION_VAL_14PX_BG NULL
#define OPTION_VAL_15PX_BG NULL
#define OPTION_VAL_16PX_BG NULL
#define OPTION_VAL_17PX_BG NULL
#define OPTION_VAL_18PX_BG NULL
#define OPTION_VAL_19PX_BG NULL
#define OPTION_VAL_20PX_BG NULL
#define BEETLE_OPT_IMAGE_OFFSET_LABEL_BG NULL
#define BEETLE_OPT_IMAGE_OFFSET_INFO_0_BG NULL
#define OPTION_VAL_12PX_O64_BG NULL
#define OPTION_VAL_11PX_O64_BG NULL
#define OPTION_VAL_10PX_O64_BG NULL
#define OPTION_VAL_9PX_O64_BG NULL
#define OPTION_VAL_8PX_O64_BG NULL
#define OPTION_VAL_7PX_O64_BG NULL
#define OPTION_VAL_6PX_O64_BG NULL
#define OPTION_VAL_5PX_O64_BG NULL
#define OPTION_VAL_4PX_O64_BG NULL
#define OPTION_VAL_3PX_O64_BG NULL
#define OPTION_VAL_2PX_O64_BG NULL
#define OPTION_VAL_1PX_O64_BG NULL
#define OPTION_VAL_1PX_O64_0_BG NULL
#define OPTION_VAL_2PX_O64_0_BG NULL
#define OPTION_VAL_3PX_O64_0_BG NULL
#define OPTION_VAL_4PX_O64_0_BG NULL
#define OPTION_VAL_5PX_O64_0_BG NULL
#define OPTION_VAL_6PX_O64_0_BG NULL
#define OPTION_VAL_7PX_O64_0_BG NULL
#define OPTION_VAL_8PX_O64_0_BG NULL
#define OPTION_VAL_9PX_O64_0_BG NULL
#define OPTION_VAL_10PX_O64_0_BG NULL
#define OPTION_VAL_11PX_O64_0_BG NULL
#define OPTION_VAL_12PX_O64_0_BG NULL
#define BEETLE_OPT_IMAGE_OFFSET_CYCLES_LABEL_BG NULL
#define BEETLE_OPT_IMAGE_OFFSET_CYCLES_INFO_0_BG NULL
#define BEETLE_OPT_INITIAL_SCANLINE_LABEL_BG NULL
#define BEETLE_OPT_INITIAL_SCANLINE_INFO_0_BG NULL
#define BEETLE_OPT_LAST_SCANLINE_LABEL_BG NULL
#define BEETLE_OPT_LAST_SCANLINE_INFO_0_BG NULL
#define OPTION_VAL_239_BG NULL
#define BEETLE_OPT_INITIAL_SCANLINE_PAL_LABEL_BG NULL
#define BEETLE_OPT_INITIAL_SCANLINE_PAL_INFO_0_BG NULL
#define BEETLE_OPT_LAST_SCANLINE_PAL_LABEL_BG NULL
#define BEETLE_OPT_LAST_SCANLINE_PAL_INFO_0_BG NULL
#define OPTION_VAL_287_BG NULL

struct retro_core_option_v2_category option_cats_bg[] = {
   {
      "system",
      CATEGORY_SYSTEM_LABEL_BG,
      CATEGORY_SYSTEM_INFO_0_BG
   },
   {
      "video",
      CATEGORY_VIDEO_LABEL_BG,
      CATEGORY_VIDEO_INFO_0_BG
   },
   {
      "pgxp",
      CATEGORY_PGXP_LABEL_BG,
      CATEGORY_PGXP_INFO_0_BG
   },
   {
      "osd",
      CATEGORY_OSD_LABEL_BG,
      CATEGORY_OSD_INFO_0_BG
   },
   {
      "input",
      CATEGORY_INPUT_LABEL_BG,
      CATEGORY_INPUT_INFO_0_BG
   },
   {
      "memcards",
      CATEGORY_MEMCARDS_LABEL_BG,
      CATEGORY_MEMCARDS_INFO_0_BG
   },
   {
      "hacks",
      CATEGORY_HACKS_LABEL_BG,
      CATEGORY_HACKS_INFO_0_BG
   },
   { NULL, NULL, NULL },
};
struct retro_core_option_v2_definition option_defs_bg[] = {
   {
      BEETLE_OPT(cpu_freq_scale),
      BEETLE_OPT_CPU_FREQ_SCALE_LABEL_BG,
      NULL,
      BEETLE_OPT_CPU_FREQ_SCALE_INFO_0_BG,
      NULL,
      "system",
      {
         { "50%",           OPTION_VAL_50_BG },
         { "60%",           OPTION_VAL_60_BG },
         { "70%",           OPTION_VAL_70_BG },
         { "80%",           OPTION_VAL_80_BG },
         { "90%",           OPTION_VAL_90_BG },
         { "100%",          OPTION_VAL_100_BG },
         { "110%",          OPTION_VAL_110_BG },
         { "120%",          OPTION_VAL_120_BG },
         { "130%",          OPTION_VAL_130_BG },
         { "140%",          OPTION_VAL_140_BG },
         { "150%",          OPTION_VAL_150_BG },
         { "160%",          OPTION_VAL_160_BG },
         { "170%",          OPTION_VAL_170_BG },
         { "180%",          OPTION_VAL_180_BG },
         { "190%",          OPTION_VAL_190_BG },
         { "200%",          OPTION_VAL_200_BG },
         { "210%",          OPTION_VAL_210_BG },
         { "220%",          OPTION_VAL_220_BG },
         { "230%",          OPTION_VAL_230_BG },
         { "240%",          OPTION_VAL_240_BG },
         { "250%",          OPTION_VAL_250_BG },
         { "260%",          OPTION_VAL_260_BG },
         { "270%",          OPTION_VAL_270_BG },
         { "280%",          OPTION_VAL_280_BG },
         { "290%",          OPTION_VAL_290_BG },
         { "300%",          OPTION_VAL_300_BG },
         { "310%",          OPTION_VAL_310_BG },
         { "320%",          OPTION_VAL_320_BG },
         { "330%",          OPTION_VAL_330_BG },
         { "340%",          OPTION_VAL_340_BG },
         { "350%",          OPTION_VAL_350_BG },
         { "360%",          OPTION_VAL_360_BG },
         { "370%",          OPTION_VAL_370_BG },
         { "380%",          OPTION_VAL_380_BG },
         { "390%",          OPTION_VAL_390_BG },
         { "400%",          OPTION_VAL_400_BG },
         { "410%",          OPTION_VAL_410_BG },
         { "420%",          OPTION_VAL_420_BG },
         { "430%",          OPTION_VAL_430_BG },
         { "440%",          OPTION_VAL_440_BG },
         { "450%",          OPTION_VAL_450_BG },
         { "460%",          OPTION_VAL_460_BG },
         { "470%",          OPTION_VAL_470_BG },
         { "480%",          OPTION_VAL_480_BG },
         { "490%",          OPTION_VAL_490_BG },
         { "500%",          OPTION_VAL_500_BG },
         { "510%",          OPTION_VAL_510_BG },
         { "520%",          OPTION_VAL_520_BG },
         { "530%",          OPTION_VAL_530_BG },
         { "540%",          OPTION_VAL_540_BG },
         { "550%",          OPTION_VAL_550_BG },
         { "560%",          OPTION_VAL_560_BG },
         { "570%",          OPTION_VAL_570_BG },
         { "580%",          OPTION_VAL_580_BG },
         { "590%",          OPTION_VAL_590_BG },
         { "600%",          OPTION_VAL_600_BG },
         { "610%",          OPTION_VAL_610_BG },
         { "620%",          OPTION_VAL_620_BG },
         { "630%",          OPTION_VAL_630_BG },
         { "640%",          OPTION_VAL_640_BG },
         { "650%",          OPTION_VAL_650_BG },
         { "660%",          OPTION_VAL_660_BG },
         { "670%",          OPTION_VAL_670_BG },
         { "680%",          OPTION_VAL_680_BG },
         { "690%",          OPTION_VAL_690_BG },
         { "700%",          OPTION_VAL_700_BG },
         { "710%",          OPTION_VAL_710_BG },
         { "720%",          OPTION_VAL_720_BG },
         { "730%",          OPTION_VAL_730_BG },
         { "740%",          OPTION_VAL_740_BG },
         { "750%",          OPTION_VAL_750_BG },
         { NULL, NULL },
      },
      "100%"
   },
   {
      BEETLE_OPT(gpu_overclock),
      BEETLE_OPT_GPU_OVERCLOCK_LABEL_BG,
      NULL,
      BEETLE_OPT_GPU_OVERCLOCK_INFO_0_BG,
      NULL,
      "system",
      {
         { "1x(native)", OPTION_VAL_1X_NATIVE_BG },
         { "2x",         OPTION_VAL_2X_BG },
         { "4x",         OPTION_VAL_4X_BG },
         { "8x",         OPTION_VAL_8X_BG },
         { "16x",        OPTION_VAL_16X_BG },
         { "32x",        OPTION_VAL_32X_BG },
         { NULL, NULL },
      },
      "1x(native)"
   },
   {
      BEETLE_OPT(gte_overclock),
      BEETLE_OPT_GTE_OVERCLOCK_LABEL_BG,
      NULL,
      BEETLE_OPT_GTE_OVERCLOCK_INFO_0_BG,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
#ifndef EMSCRIPTEN
   {
      BEETLE_OPT(cd_access_method),
      BEETLE_OPT_CD_ACCESS_METHOD_LABEL_BG,
      NULL,
      BEETLE_OPT_CD_ACCESS_METHOD_INFO_0_BG,
      NULL,
      "system",
      {
         { "sync",     OPTION_VAL_SYNC_BG },
         { "async",    OPTION_VAL_ASYNC_BG },
         { "precache", OPTION_VAL_PRECACHE_BG },
         { NULL, NULL },
      },
      "sync"
   },
#endif
   {
      BEETLE_OPT(cd_fastload),
      BEETLE_OPT_CD_FASTLOAD_LABEL_BG,
      NULL,
      BEETLE_OPT_CD_FASTLOAD_INFO_0_BG,
      NULL,
      "system",
      {
         { "2x(native)", OPTION_VAL_2X_NATIVE_BG },
         { "4x",          OPTION_VAL_4X_BG },
         { "6x",          OPTION_VAL_6X_BG },
         { "8x",          OPTION_VAL_8X_BG },
         { "10x",         OPTION_VAL_10X_BG },
         { "12x",         OPTION_VAL_12X_BG },
         { "14x",         OPTION_VAL_14X_BG },
         { NULL, NULL },
      },
      "2x(native)"
   },
   {
      BEETLE_OPT(skip_bios),
      BEETLE_OPT_SKIP_BIOS_LABEL_BG,
      NULL,
      BEETLE_OPT_SKIP_BIOS_INFO_0_BG,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(override_bios),
      BEETLE_OPT_OVERRIDE_BIOS_LABEL_BG,
      NULL,
      BEETLE_OPT_OVERRIDE_BIOS_INFO_0_BG,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "psxonpsp", OPTION_VAL_PSXONPSP_BG },
         { "ps1_rom",  OPTION_VAL_PS1_ROM_BG },
         { "openbios", OPTION_VAL_OPENBIOS_BG     },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(region),
      BEETLE_OPT_REGION_LABEL_BG,
      NULL,
      BEETLE_OPT_REGION_INFO_0_BG,
      NULL,
      "system",
      {
         { "auto",   OPTION_VAL_AUTO_BG },
         { "ntsc-j", OPTION_VAL_NTSC_J_BG },
         { "ntsc-u", OPTION_VAL_NTSC_U_BG },
         { "pal",    OPTION_VAL_PAL_BG },
         { NULL, NULL },
      },
      "auto"
   },
   {
      BEETLE_OPT(internal_resolution),
      BEETLE_OPT_INTERNAL_RESOLUTION_LABEL_BG,
      NULL,
      BEETLE_OPT_INTERNAL_RESOLUTION_INFO_0_BG,
      NULL,
      "video",
      {
         { "1x(native)", OPTION_VAL_1X_NATIVE_BG },
         { "2x",         OPTION_VAL_2X_BG },
         { "4x",         OPTION_VAL_4X_BG },
         { "8x",         OPTION_VAL_8X_BG },
         { "16x",        OPTION_VAL_16X_BG },
         { NULL, NULL },
      },
      "1x(native)"
   },
#if defined(HAVE_OPENGL) || defined(HAVE_OPENGLES) || defined(HAVE_VULKAN)
   {
      BEETLE_OPT(renderer),
      BEETLE_OPT_RENDERER_LABEL_BG,
      NULL,
      BEETLE_OPT_RENDERER_INFO_0_BG,
      NULL,
      "video",
      {
         { "hardware",    OPTION_VAL_HARDWARE_BG },
#if defined(HAVE_OPENGL) || defined(HAVE_OPENGLES)
         { "hardware_gl", OPTION_VAL_HARDWARE_GL_BG },
#endif
#if defined(HAVE_VULKAN)
         { "hardware_vk", OPTION_VAL_HARDWARE_VK_BG },
#endif
         { "software",    OPTION_VAL_SOFTWARE_BG },
         { NULL, NULL },
      },
      "hardware"
   },
   {
      BEETLE_OPT(renderer_software_fb),
      BEETLE_OPT_RENDERER_SOFTWARE_FB_LABEL_BG,
      NULL,
      BEETLE_OPT_RENDERER_SOFTWARE_FB_INFO_0_BG,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled"
   },
#endif
#if defined(HAVE_OPENGL) || defined(HAVE_OPENGLES)
   {
      BEETLE_OPT(depth),
      BEETLE_OPT_DEPTH_LABEL_BG,
      NULL,
      BEETLE_OPT_DEPTH_INFO_0_BG,
      NULL,
      "video",
      {
         { "16bpp(native)", OPTION_VAL_16BPP_NATIVE_BG },
         { "32bpp",         OPTION_VAL_32BPP_BG },
         { NULL, NULL },
      },
      "16bpp(native)"
   },
   // Sort of, it's more like 15-bit high color and 24-bit true color for visible output. The alpha channel is used for mask bit. Vulkan renderer uses ABGR1555_555 for 31 bits internal? FMVs are always 24-bit on all renderers like original hardware (BGR888, no alpha)
#endif
   {
      BEETLE_OPT(dither_mode),
      BEETLE_OPT_DITHER_MODE_LABEL_BG,
      NULL,
      BEETLE_OPT_DITHER_MODE_INFO_0_BG,
      NULL,
      "video",
      {
         { "1x(native)",          OPTION_VAL_1X_NATIVE_BG },
         { "internal resolution", OPTION_VAL_INTERNAL_RESOLUTION_BG },
         { "disabled",            NULL },
         { NULL, NULL },
      },
      "1x(native)"
   },
#ifdef HAVE_VULKAN
   {
      BEETLE_OPT(scaled_uv_offset),
      BEETLE_OPT_SCALED_UV_OFFSET_LABEL_BG,
      NULL,
      BEETLE_OPT_SCALED_UV_OFFSET_INFO_0_BG,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled"
   },
#endif
#if defined(HAVE_OPENGL) || defined(HAVE_OPENGLES) || defined(HAVE_VULKAN)
   {
      BEETLE_OPT(filter),
      BEETLE_OPT_FILTER_LABEL_BG,
      NULL,
      BEETLE_OPT_FILTER_INFO_0_BG,
      NULL,
      "video",
      {
         { "nearest",  OPTION_VAL_NEAREST_BG },
         { "SABR",     OPTION_VAL_SABR_BG },
         { "xBR",      OPTION_VAL_XBR_BG },
         { "bilinear", OPTION_VAL_BILINEAR_BG },
         { "3-point",  OPTION_VAL_3_POINT_BG },
         { "JINC2",    OPTION_VAL_JINC2_BG },
         { NULL, NULL },
      },
      "nearest"
   },
#ifdef HAVE_VULKAN
   {
      BEETLE_OPT(filter_exclude_sprite),
      BEETLE_OPT_FILTER_EXCLUDE_SPRITE_LABEL_BG,
      NULL,
      BEETLE_OPT_FILTER_EXCLUDE_SPRITE_INFO_0_BG,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "opaque", OPTION_VAL_OPAQUE_BG },
         { "all", OPTION_VAL_ALL_BG },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(filter_exclude_2d_polygon),
      BEETLE_OPT_FILTER_EXCLUDE_2D_POLYGON_LABEL_BG,
      NULL,
      BEETLE_OPT_FILTER_EXCLUDE_2D_POLYGON_INFO_0_BG,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "opaque", OPTION_VAL_OPAQUE_BG },
         { "all", OPTION_VAL_ALL_BG },
         { NULL, NULL },
      },
      "disabled"
   },
#endif
#endif
#ifdef HAVE_VULKAN
   {
      BEETLE_OPT(adaptive_smoothing),
      BEETLE_OPT_ADAPTIVE_SMOOTHING_LABEL_BG,
      NULL,
      BEETLE_OPT_ADAPTIVE_SMOOTHING_INFO_0_BG,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(super_sampling),
      BEETLE_OPT_SUPER_SAMPLING_LABEL_BG,
      NULL,
      BEETLE_OPT_SUPER_SAMPLING_INFO_0_BG,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(msaa),
      BEETLE_OPT_MSAA_LABEL_BG,
      NULL,
      BEETLE_OPT_MSAA_INFO_0_BG,
      NULL,
      "video",
      {
         { "1x",  OPTION_VAL_1X_BG },
         { "2x",  OPTION_VAL_2X_BG },
         { "4x",  OPTION_VAL_4X_BG },
         { "8x",  OPTION_VAL_8X_BG },
         { "16x", OPTION_VAL_16X_BG },
         { NULL, NULL },
      },
      "1x"
   },
   {
      BEETLE_OPT(mdec_yuv),
      BEETLE_OPT_MDEC_YUV_LABEL_BG,
      NULL,
      BEETLE_OPT_MDEC_YUV_INFO_0_BG,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(track_textures),
      BEETLE_OPT_TRACK_TEXTURES_LABEL_BG,
      NULL,
      BEETLE_OPT_TRACK_TEXTURES_INFO_0_BG,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
#ifdef TEXTURE_DUMPING_ENABLED
   {
      BEETLE_OPT(dump_textures),
      BEETLE_OPT_DUMP_TEXTURES_LABEL_BG,
      NULL,
      BEETLE_OPT_DUMP_TEXTURES_INFO_0_BG,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
#endif
   {
      BEETLE_OPT(replace_textures),
      BEETLE_OPT_REPLACE_TEXTURES_LABEL_BG,
      NULL,
      BEETLE_OPT_REPLACE_TEXTURES_INFO_0_BG,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
#endif
   {
      BEETLE_OPT(pgxp_mode),
      BEETLE_OPT_PGXP_MODE_LABEL_BG,
      NULL,
      BEETLE_OPT_PGXP_MODE_INFO_0_BG,
      NULL,
      "pgxp",
      {
         { "disabled",     NULL },
         { "memory only",  OPTION_VAL_MEMORY_ONLY_BG },
         { "memory + CPU", OPTION_VAL_MEMORY_CPU_BG },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(pgxp_2d_tol),
      BEETLE_OPT_PGXP_2D_TOL_LABEL_BG,
      NULL,
      BEETLE_OPT_PGXP_2D_TOL_INFO_0_BG,
      NULL,
      "pgxp",
      {
         { "disabled", NULL },
         { "0px", OPTION_VAL_0PX_BG },
         { "1px", OPTION_VAL_1PX_BG },
         { "2px", OPTION_VAL_2PX_BG },
         { "3px", OPTION_VAL_3PX_BG },
         { "4px", OPTION_VAL_4PX_BG },
         { "5px", OPTION_VAL_5PX_BG },
         { "6px", OPTION_VAL_6PX_BG },
         { "7px", OPTION_VAL_7PX_BG },
         { "8px", OPTION_VAL_8PX_BG },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(pgxp_nclip),
      BEETLE_OPT_PGXP_NCLIP_LABEL_BG,
      NULL,
      BEETLE_OPT_PGXP_NCLIP_INFO_0_BG,
      NULL,
      "pgxp",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
#if defined(HAVE_OPENGL) || defined(HAVE_OPENGLES) || defined(HAVE_VULKAN)
   {
      BEETLE_OPT(pgxp_vertex),
      BEETLE_OPT_PGXP_VERTEX_LABEL_BG,
      NULL,
      BEETLE_OPT_PGXP_VERTEX_INFO_0_BG,
      NULL,
      "pgxp",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
#endif
   {
      BEETLE_OPT(pgxp_texture),
      BEETLE_OPT_PGXP_TEXTURE_LABEL_BG,
      NULL,
      BEETLE_OPT_PGXP_TEXTURE_INFO_0_BG,
      NULL,
      "pgxp",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(display_internal_fps),
      BEETLE_OPT_DISPLAY_INTERNAL_FPS_LABEL_BG,
      NULL,
      BEETLE_OPT_DISPLAY_INTERNAL_FPS_INFO_0_BG,
      NULL,
      "osd",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(display_osd),
      BEETLE_OPT_DISPLAY_OSD_LABEL_BG,
      NULL,
      BEETLE_OPT_DISPLAY_OSD_INFO_0_BG,
      NULL,
      "osd",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled"
   },
#if defined(HAVE_OPENGL) || defined(HAVE_OPENGLES) || defined(HAVE_VULKAN)
   {
      BEETLE_OPT(display_vram),
      BEETLE_OPT_DISPLAY_VRAM_LABEL_BG,
      NULL,
      BEETLE_OPT_DISPLAY_VRAM_INFO_0_BG,
      NULL,
      "osd",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
#endif
   {
      BEETLE_OPT(analog_calibration),
      BEETLE_OPT_ANALOG_CALIBRATION_LABEL_BG,
      NULL,
      BEETLE_OPT_ANALOG_CALIBRATION_INFO_0_BG,
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(analog_toggle),
      BEETLE_OPT_ANALOG_TOGGLE_LABEL_BG,
      NULL,
      BEETLE_OPT_ANALOG_TOGGLE_INFO_0_BG,
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { "enabled-analog", OPTION_VAL_ENABLED_ANALOG_BG },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(analog_toggle_combo),
      BEETLE_OPT_ANALOG_TOGGLE_COMBO_LABEL_BG,
      NULL,
      BEETLE_OPT_ANALOG_TOGGLE_COMBO_INFO_0_BG,
      NULL,
      "input",
      {
         { "l1+l2+r1+r2+start+select", OPTION_VAL_L1_L2_R1_R2_START_SELECT_BG },
         { "l1+r1+select",             OPTION_VAL_L1_R1_SELECT_BG },
         { "l1+r1+start",              OPTION_VAL_L1_R1_START_BG },
         { "l1+r1+l3",                 OPTION_VAL_L1_R1_L3_BG },
         { "l1+r1+r3",                 OPTION_VAL_L1_R1_R3_BG },
         { "l2+r2+select",             OPTION_VAL_L2_R2_SELECT_BG },
         { "l2+r2+start",              OPTION_VAL_L2_R2_START_BG },
         { "l2+r2+l3",                 OPTION_VAL_L2_R2_L3_BG },
         { "l2+r2+r3",                 OPTION_VAL_L2_R2_R3_BG },
         { "l3+r3",                    OPTION_VAL_L3_R3_BG },
         { NULL, NULL },
      },
      "l1+r1+select"
   },
   {
      BEETLE_OPT(analog_toggle_hold),
      BEETLE_OPT_ANALOG_TOGGLE_HOLD_LABEL_BG,
      NULL,
      BEETLE_OPT_ANALOG_TOGGLE_HOLD_INFO_0_BG,
      NULL,
      "input",
      {
         { "0", OPTION_VAL_0_BG },
         { "1", OPTION_VAL_1_BG },
         { "2", OPTION_VAL_2_BG },
         { "3", OPTION_VAL_3_BG },
         { "4", OPTION_VAL_4_BG },
         { "5", OPTION_VAL_5_BG },
         { NULL, NULL },
      },
      "1"
   },
   {
      BEETLE_OPT(enable_multitap_port1),
      BEETLE_OPT_ENABLE_MULTITAP_PORT1_LABEL_BG,
      NULL,
      BEETLE_OPT_ENABLE_MULTITAP_PORT1_INFO_0_BG,
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(enable_multitap_port2),
      BEETLE_OPT_ENABLE_MULTITAP_PORT2_LABEL_BG,
      NULL,
      BEETLE_OPT_ENABLE_MULTITAP_PORT2_INFO_0_BG,
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(gun_input_mode),
      BEETLE_OPT_GUN_INPUT_MODE_LABEL_BG,
      NULL,
      BEETLE_OPT_GUN_INPUT_MODE_INFO_0_BG,
      NULL,
      "input",
      {
         { "lightgun",    OPTION_VAL_LIGHTGUN_BG },
         { "touchscreen", OPTION_VAL_TOUCHSCREEN_BG },
         { NULL, NULL },
      },
      "lightgun"
   },
   // Shouldn't the gun_input_mode just be Mouse vs. Touchscreen?
   {
      BEETLE_OPT(gun_cursor),
      BEETLE_OPT_GUN_CURSOR_LABEL_BG,
      NULL,
      BEETLE_OPT_GUN_CURSOR_INFO_0_BG,
      NULL,
      "input",
      {
         { "off",   "disabled" },
         { "cross", OPTION_VAL_CROSS_BG },
         { "dot",   OPTION_VAL_DOT_BG },
         { NULL, NULL },
      },
      "cross"
   },
   {
      BEETLE_OPT(crosshair_color_p1),
      BEETLE_OPT_CROSSHAIR_COLOR_P1_LABEL_BG,
      NULL,
      BEETLE_OPT_CROSSHAIR_COLOR_P1_INFO_0_BG,
      NULL,
      "input",
      {
         { "red",    OPTION_VAL_RED_BG },
         { "blue",   OPTION_VAL_BLUE_BG },
         { "green",  OPTION_VAL_GREEN_BG },
         { "orange", OPTION_VAL_ORANGE_BG },
         { "yellow", OPTION_VAL_YELLOW_BG },
         { "cyan",   OPTION_VAL_CYAN_BG },
         { "pink",   OPTION_VAL_PINK_BG },
         { "purple", OPTION_VAL_PURPLE_BG },
         { "black",  OPTION_VAL_BLACK_BG },
         { "white",  OPTION_VAL_WHITE_BG },
         { NULL, NULL },
      },
      "red"
   },
   {
      BEETLE_OPT(crosshair_color_p2),
      BEETLE_OPT_CROSSHAIR_COLOR_P2_LABEL_BG,
      NULL,
      BEETLE_OPT_CROSSHAIR_COLOR_P2_INFO_0_BG,
      NULL,
      "input",
      {
         { "blue",   OPTION_VAL_BLUE_BG },
         { "red",    OPTION_VAL_RED_BG },
         { "green",  OPTION_VAL_GREEN_BG },
         { "orange", OPTION_VAL_ORANGE_BG },
         { "yellow", OPTION_VAL_YELLOW_BG },
         { "cyan",   OPTION_VAL_CYAN_BG },
         { "pink",   OPTION_VAL_PINK_BG },
         { "purple", OPTION_VAL_PURPLE_BG },
         { "black",  OPTION_VAL_BLACK_BG },
         { "white",  OPTION_VAL_WHITE_BG },
         { NULL, NULL },
      },
      "blue"
   },
   {
      BEETLE_OPT(mouse_sensitivity),
      BEETLE_OPT_MOUSE_SENSITIVITY_LABEL_BG,
      NULL,
      BEETLE_OPT_MOUSE_SENSITIVITY_INFO_0_BG,
      NULL,
      "input",
      {
         { "5%",   OPTION_VAL_5_O42_BG },
         { "10%",  OPTION_VAL_10_BG },
         { "15%",  OPTION_VAL_15_BG },
         { "20%",  OPTION_VAL_20_BG },
         { "25%",  OPTION_VAL_25_BG },
         { "30%",  OPTION_VAL_30_BG },
         { "35%",  OPTION_VAL_35_BG },
         { "40%",  OPTION_VAL_40_BG },
         { "45%",  OPTION_VAL_45_BG },
         { "50%",  OPTION_VAL_50_BG },
         { "55%",  OPTION_VAL_55_BG },
         { "60%",  OPTION_VAL_60_BG },
         { "65%",  OPTION_VAL_65_BG },
         { "70%",  OPTION_VAL_70_BG },
         { "75%",  OPTION_VAL_75_BG },
         { "80%",  OPTION_VAL_80_BG },
         { "85%",  OPTION_VAL_85_BG },
         { "90%",  OPTION_VAL_90_BG },
         { "95%",  OPTION_VAL_95_BG },
         { "100%", OPTION_VAL_100_O42_BG },
         { "105%", OPTION_VAL_105_BG },
         { "110%", OPTION_VAL_110_BG },
         { "115%", OPTION_VAL_115_BG },
         { "120%", OPTION_VAL_120_BG },
         { "125%", OPTION_VAL_125_BG },
         { "130%", OPTION_VAL_130_BG },
         { "135%", OPTION_VAL_135_BG },
         { "140%", OPTION_VAL_140_BG },
         { "145%", OPTION_VAL_145_BG },
         { "150%", OPTION_VAL_150_BG },
         { "155%", OPTION_VAL_155_BG },
         { "160%", OPTION_VAL_160_BG },
         { "165%", OPTION_VAL_165_BG },
         { "170%", OPTION_VAL_170_BG },
         { "175%", OPTION_VAL_175_BG },
         { "180%", OPTION_VAL_180_BG },
         { "185%", OPTION_VAL_185_BG },
         { "190%", OPTION_VAL_190_BG },
         { "195%", OPTION_VAL_195_BG },
         { "200%", OPTION_VAL_200_BG },
         { NULL, NULL },
      },
      "100%"
   },
   {
      BEETLE_OPT(negcon_response),
      BEETLE_OPT_NEGCON_RESPONSE_LABEL_BG,
      NULL,
      BEETLE_OPT_NEGCON_RESPONSE_INFO_0_BG,
      NULL,
      "input",
      {
         { "linear",    OPTION_VAL_LINEAR_BG },
         { "quadratic", OPTION_VAL_QUADRATIC_BG },
         { "cubic",     OPTION_VAL_CUBIC_BG },
         { NULL, NULL },
      },
      "linear"
   },
   {
      BEETLE_OPT(negcon_deadzone),
      BEETLE_OPT_NEGCON_DEADZONE_LABEL_BG,
      NULL,
      BEETLE_OPT_NEGCON_DEADZONE_INFO_0_BG,
      NULL,
      "input",
      {
         { "0%",  OPTION_VAL_0_O44_BG },
         { "5%",  OPTION_VAL_5_O42_BG },
         { "10%", OPTION_VAL_10_BG },
         { "15%", OPTION_VAL_15_BG },
         { "20%", OPTION_VAL_20_BG },
         { "25%", OPTION_VAL_25_BG },
         { "30%", OPTION_VAL_30_BG },
         { NULL, NULL },
      },
      "0%"
   },
   {
      BEETLE_OPT(use_mednafen_memcard0_method),
      BEETLE_OPT_USE_MEDNAFEN_MEMCARD0_METHOD_LABEL_BG,
      NULL,
      BEETLE_OPT_USE_MEDNAFEN_MEMCARD0_METHOD_INFO_0_BG,
      NULL,
      "memcards",
      {
         { "libretro", OPTION_VAL_LIBRETRO_BG },
         { "mednafen", OPTION_VAL_MEDNAFEN_BG },
         { NULL, NULL },
      },
      "libretro"
   },
   {
      BEETLE_OPT(enable_memcard1),
      BEETLE_OPT_ENABLE_MEMCARD1_LABEL_BG,
      NULL,
      BEETLE_OPT_ENABLE_MEMCARD1_INFO_0_BG,
      NULL,
      "memcards",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(shared_memory_cards),
      BEETLE_OPT_SHARED_MEMORY_CARDS_LABEL_BG,
      NULL,
      BEETLE_OPT_SHARED_MEMORY_CARDS_INFO_0_BG,
      NULL,
      "memcards",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(memcard_left_index),
      BEETLE_OPT_MEMCARD_LEFT_INDEX_LABEL_BG,
      NULL,
      BEETLE_OPT_MEMCARD_LEFT_INDEX_INFO_0_BG,
      NULL,
      "memcards",
      {
         { "0",  OPTION_VAL_0_O48_BG },
         { "1",  NULL },
         { "2",  NULL },
         { "3",  NULL },
         { "4",  NULL },
         { "5",  NULL },
         { "6",  NULL },
         { "7",  NULL },
         { "8",  NULL },
         { "9",  NULL },
         { "10", NULL },
         { "11", NULL },
         { "12", NULL },
         { "13", NULL },
         { "14", NULL },
         { "15", NULL },
         { "16", NULL },
         { "17", NULL },
         { "18", NULL },
         { "19", NULL },
         { "20", NULL },
         { "21", NULL },
         { "22", NULL },
         { "23", NULL },
         { "24", NULL },
         { "25", NULL },
         { "26", NULL },
         { "27", NULL },
         { "28", NULL },
         { "29", NULL },
         { "30", NULL },
         { "31", NULL },
         { "32", NULL },
         { "33", NULL },
         { "34", NULL },
         { "35", NULL },
         { "36", NULL },
         { "37", NULL },
         { "38", NULL },
         { "39", NULL },
         { "40", NULL },
         { "41", NULL },
         { "42", NULL },
         { "43", NULL },
         { "44", NULL },
         { "45", NULL },
         { "46", NULL },
         { "47", NULL },
         { "48", NULL },
         { "49", NULL },
         { "50", NULL },
         { "51", NULL },
         { "52", NULL },
         { "53", NULL },
         { "54", NULL },
         { "55", NULL },
         { "56", NULL },
         { "57", NULL },
         { "58", NULL },
         { "59", NULL },
         { "60", NULL },
         { "61", NULL },
         { "62", NULL },
         { "63", NULL },
         { NULL, NULL },
      },
      "0"
   },
   {
      BEETLE_OPT(memcard_right_index),
      BEETLE_OPT_MEMCARD_RIGHT_INDEX_LABEL_BG,
      NULL,
      BEETLE_OPT_MEMCARD_RIGHT_INDEX_INFO_0_BG,
      NULL,
      "memcards",
      {
         { "0",  NULL },
         { "1",  OPTION_VAL_1_O49_BG },
         { "2",  NULL },
         { "3",  NULL },
         { "4",  NULL },
         { "5",  NULL },
         { "6",  NULL },
         { "7",  NULL },
         { "8",  NULL },
         { "9",  NULL },
         { "10", NULL },
         { "11", NULL },
         { "12", NULL },
         { "13", NULL },
         { "14", NULL },
         { "15", NULL },
         { "16", NULL },
         { "17", NULL },
         { "18", NULL },
         { "19", NULL },
         { "20", NULL },
         { "21", NULL },
         { "22", NULL },
         { "23", NULL },
         { "24", NULL },
         { "25", NULL },
         { "26", NULL },
         { "27", NULL },
         { "28", NULL },
         { "29", NULL },
         { "30", NULL },
         { "31", NULL },
         { "32", NULL },
         { "33", NULL },
         { "34", NULL },
         { "35", NULL },
         { "36", NULL },
         { "37", NULL },
         { "38", NULL },
         { "39", NULL },
         { "40", NULL },
         { "41", NULL },
         { "42", NULL },
         { "43", NULL },
         { "44", NULL },
         { "45", NULL },
         { "46", NULL },
         { "47", NULL },
         { "48", NULL },
         { "49", NULL },
         { "50", NULL },
         { "51", NULL },
         { "52", NULL },
         { "53", NULL },
         { "54", NULL },
         { "55", NULL },
         { "56", NULL },
         { "57", NULL },
         { "58", NULL },
         { "59", NULL },
         { "60", NULL },
         { "61", NULL },
         { "62", NULL },
         { "63", NULL },
         { NULL, NULL },
      },
      "1"
   },
   {
      BEETLE_OPT(line_render),
      BEETLE_OPT_LINE_RENDER_LABEL_BG,
      NULL,
      BEETLE_OPT_LINE_RENDER_INFO_0_BG,
      NULL,
      "hacks",
      {
         { "default",    OPTION_VAL_DEFAULT_BG },
         { "aggressive", OPTION_VAL_AGGRESSIVE_BG },
         { "disabled",   NULL },
         { NULL, NULL },
      },
      "default"
   },
   {
      BEETLE_OPT(widescreen_hack),
      BEETLE_OPT_WIDESCREEN_HACK_LABEL_BG,
      NULL,
      BEETLE_OPT_WIDESCREEN_HACK_INFO_0_BG,
      NULL,
      "hacks",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(widescreen_hack_aspect_ratio),
      BEETLE_OPT_WIDESCREEN_HACK_ASPECT_RATIO_LABEL_BG,
      NULL,
      BEETLE_OPT_WIDESCREEN_HACK_ASPECT_RATIO_INFO_0_BG,
      NULL,
      "hacks",
      {
         { "16:9",  OPTION_VAL_16_9_BG },
         { "16:10", OPTION_VAL_16_10_BG },
         { "18:9",  OPTION_VAL_18_9_BG },
         { "19:9",  OPTION_VAL_19_9_BG },
         { "20:9",  OPTION_VAL_20_9_BG },
         { "21:9",  OPTION_VAL_21_9_BG }, // 64:27
         { "32:9",  OPTION_VAL_32_9_BG },
         { NULL,    NULL },
      },
      "16:9"
   },
#if defined(HAVE_LIGHTREC)
   {
      BEETLE_OPT(cpu_dynarec),
      BEETLE_OPT_CPU_DYNAREC_LABEL_BG,
      NULL,
      BEETLE_OPT_CPU_DYNAREC_INFO_0_BG,
      NULL,
      "hacks",
      {
         { "disabled",        OPTION_VAL_DISABLED_BG },
         { "execute",         OPTION_VAL_EXECUTE_BG },
         { "execute_one",     OPTION_VAL_EXECUTE_ONE_BG },
         { "run_interpreter", OPTION_VAL_RUN_INTERPRETER_BG },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(dynarec_invalidate),
      BEETLE_OPT_DYNAREC_INVALIDATE_LABEL_BG,
      NULL,
      BEETLE_OPT_DYNAREC_INVALIDATE_INFO_0_BG,
      NULL,
      "hacks",
      {
         { "full", OPTION_VAL_FULL_BG },
         { "dma",  OPTION_VAL_DMA_BG },
         { NULL, NULL },
      },
      "full"
   },
   {
      BEETLE_OPT(dynarec_eventcycles),
      BEETLE_OPT_DYNAREC_EVENTCYCLES_LABEL_BG,
      NULL,
      BEETLE_OPT_DYNAREC_EVENTCYCLES_INFO_0_BG,
      NULL,
      "hacks",
      {
         { "128",  OPTION_VAL_128_BG },
         { "256",  NULL },
         { "384",  NULL },
         { "512",  NULL },
         { "640",  NULL },
         { "768",  NULL },
         { "896",  NULL },
         { "1024", NULL },
         { "1152", NULL },
         { "1280", NULL },
         { "1408", NULL },
         { "1536", NULL },
         { "1664", NULL },
         { "1792", NULL },
         { "1920", NULL },
         { "2048", NULL },
         { NULL, NULL },
      },
      "128"
   },
   {
      BEETLE_OPT(dynarec_spu_samples),
      BEETLE_OPT_DYNAREC_SPU_SAMPLES_LABEL_BG,
      NULL,
      BEETLE_OPT_DYNAREC_SPU_SAMPLES_INFO_0_BG,
      NULL,
      "hacks",
      {
         { "1",  OPTION_VAL_1_O49_BG },
         { "4",  NULL },
         { "16", NULL },
         { NULL, NULL },
      },
      "1"
   },
#endif
   {
      BEETLE_OPT(core_timing_fps),
      BEETLE_OPT_CORE_TIMING_FPS_LABEL_BG,
      NULL,
      BEETLE_OPT_CORE_TIMING_FPS_INFO_0_BG,
      NULL,
      "video",
      {
         { "auto_toggle",       OPTION_VAL_AUTO_TOGGLE_BG },
         { "force_progressive", OPTION_VAL_FORCE_PROGRESSIVE_BG },
         { "force_interlaced",  OPTION_VAL_FORCE_INTERLACED_BG },
      },
      "auto_toggle"
   },
   {
      BEETLE_OPT(deinterlacer),
      BEETLE_OPT_DEINTERLACER_LABEL_BG,
      NULL,
      BEETLE_OPT_DEINTERLACER_INFO_0_BG,
      NULL,
      "video",
      {
         { "weave",      OPTION_VAL_WEAVE_BG },
         { "bob",        OPTION_VAL_BOB_BG },
         { "bob_offset", OPTION_VAL_BOB_OFFSET_BG },
         { "fastmad",    OPTION_VAL_FASTMAD_BG },
         { "off",        "Off" },
         { NULL, NULL },
      },
      "weave"
   },
   {
      BEETLE_OPT(frame_duping),
      BEETLE_OPT_FRAME_DUPING_LABEL_BG,
      NULL,
      BEETLE_OPT_FRAME_DUPING_INFO_0_BG,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(pal_video_timing_override),
      BEETLE_OPT_PAL_VIDEO_TIMING_OVERRIDE_LABEL_BG,
      NULL,
      BEETLE_OPT_PAL_VIDEO_TIMING_OVERRIDE_INFO_0_BG,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(aspect_ratio),
      BEETLE_OPT_ASPECT_RATIO_LABEL_BG,
      NULL,
      BEETLE_OPT_ASPECT_RATIO_INFO_0_BG,
      NULL,
      "video",
      {
         { "corrected", OPTION_VAL_CORRECTED_BG },
         { "uncorrected", OPTION_VAL_UNCORRECTED_BG },
         { "4:3",  OPTION_VAL_4_3_BG },
         { "ntsc", OPTION_VAL_NTSC_BG },
      },
      "corrected"
   },
   {
      BEETLE_OPT(crop_overscan),
      BEETLE_OPT_CROP_OVERSCAN_LABEL_BG,
      NULL,
      BEETLE_OPT_CROP_OVERSCAN_INFO_0_BG,
      NULL,
      "video",
      {
         { "disabled", "Disabled" },
         { "static",   OPTION_VAL_STATIC_BG },
         { "smart",    OPTION_VAL_SMART_BG },
         { NULL, NULL },
      },
      "smart"
   },
   {
      BEETLE_OPT(image_crop),
      BEETLE_OPT_IMAGE_CROP_LABEL_BG,
      NULL,
      BEETLE_OPT_IMAGE_CROP_INFO_0_BG,
      NULL,
      "video",
      {
         { "disabled", "0" },
         { "1px",      OPTION_VAL_1PX_BG },
         { "2px",      OPTION_VAL_2PX_BG },
         { "3px",      OPTION_VAL_3PX_BG },
         { "4px",      OPTION_VAL_4PX_BG },
         { "5px",      OPTION_VAL_5PX_BG },
         { "6px",      OPTION_VAL_6PX_BG },
         { "7px",      OPTION_VAL_7PX_BG },
         { "8px",      OPTION_VAL_8PX_BG },
         { "9px",      OPTION_VAL_9PX_BG },
         { "10px",     OPTION_VAL_10PX_BG },
         { "11px",     OPTION_VAL_11PX_BG },
         { "12px",     OPTION_VAL_12PX_BG },
         { "13px",     OPTION_VAL_13PX_BG },
         { "14px",     OPTION_VAL_14PX_BG },
         { "15px",     OPTION_VAL_15PX_BG },
         { "16px",     OPTION_VAL_16PX_BG },
         { "17px",     OPTION_VAL_17PX_BG },
         { "18px",     OPTION_VAL_18PX_BG },
         { "19px",     OPTION_VAL_19PX_BG },
         { "20px",     OPTION_VAL_20PX_BG },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(image_offset),
      BEETLE_OPT_IMAGE_OFFSET_LABEL_BG,
      NULL,
      BEETLE_OPT_IMAGE_OFFSET_INFO_0_BG,
      NULL,
      "video",
      {
         { "-12px",    OPTION_VAL_12PX_O64_BG },
         { "-11px",    OPTION_VAL_11PX_O64_BG },
         { "-10px",    OPTION_VAL_10PX_O64_BG },
         { "-9px",     OPTION_VAL_9PX_O64_BG },
         { "-8px",     OPTION_VAL_8PX_O64_BG },
         { "-7px",     OPTION_VAL_7PX_O64_BG },
         { "-6px",     OPTION_VAL_6PX_O64_BG },
         { "-5px",     OPTION_VAL_5PX_O64_BG },
         { "-4px",     OPTION_VAL_4PX_O64_BG },
         { "-3px",     OPTION_VAL_3PX_O64_BG },
         { "-2px",     OPTION_VAL_2PX_O64_BG },
         { "-1px",     OPTION_VAL_1PX_O64_BG },
         { "disabled", "0" },
         { "+1px",     OPTION_VAL_1PX_O64_0_BG },
         { "+2px",     OPTION_VAL_2PX_O64_0_BG },
         { "+3px",     OPTION_VAL_3PX_O64_0_BG },
         { "+4px",     OPTION_VAL_4PX_O64_0_BG },
         { "+5px",     OPTION_VAL_5PX_O64_0_BG },
         { "+6px",     OPTION_VAL_6PX_O64_0_BG },
         { "+7px",     OPTION_VAL_7PX_O64_0_BG },
         { "+8px",     OPTION_VAL_8PX_O64_0_BG },
         { "+9px",     OPTION_VAL_9PX_O64_0_BG },
         { "+10px",    OPTION_VAL_10PX_O64_0_BG },
         { "+11px",    OPTION_VAL_11PX_O64_0_BG },
         { "+12px",    OPTION_VAL_12PX_O64_0_BG },
         { NULL, NULL },
      },
      "disabled"
   },
#if defined(HAVE_OPENGL) || defined(HAVE_OPENGLES) || defined(HAVE_VULKAN)
   {
      BEETLE_OPT(image_offset_cycles),
      BEETLE_OPT_IMAGE_OFFSET_CYCLES_LABEL_BG,
      NULL,
      BEETLE_OPT_IMAGE_OFFSET_CYCLES_INFO_0_BG,
      NULL,
      "video",
      {
         { "-40",      NULL },
         { "-39",      NULL },
         { "-38",      NULL },
         { "-37",      NULL },
         { "-36",      NULL },
         { "-35",      NULL },
         { "-34",      NULL },
         { "-33",      NULL },
         { "-32",      NULL },
         { "-31",      NULL },
         { "-30",      NULL },
         { "-29",      NULL },
         { "-28",      NULL },
         { "-27",      NULL },
         { "-26",      NULL },
         { "-25",      NULL },
         { "-24",      NULL },
         { "-23",      NULL },
         { "-22",      NULL },
         { "-21",      NULL },
         { "-20",      NULL },
         { "-19",      NULL },
         { "-18",      NULL },
         { "-17",      NULL },
         { "-16",      NULL },
         { "-15",      NULL },
         { "-14",      NULL },
         { "-13",      NULL },
         { "-12",      NULL },
         { "-11",      NULL },
         { "-10",      NULL },
         { "-9",       NULL },
         { "-8",       NULL },
         { "-7",       NULL },
         { "-6",       NULL },
         { "-5",       NULL },
         { "-4",       NULL },
         { "-3",       NULL },
         { "-2",       NULL },
         { "-1",       NULL },
         { "0",        "0" },
         { "+1",       NULL },
         { "+2",       NULL },
         { "+3",       NULL },
         { "+4",       NULL },
         { "+5",       NULL },
         { "+6",       NULL },
         { "+7",       NULL },
         { "+8",       NULL },
         { "+9",       NULL },
         { "+10",      NULL },
         { "+11",      NULL },
         { "+12",      NULL },
         { "+13",      NULL },
         { "+14",      NULL },
         { "+15",      NULL },
         { "+16",      NULL },
         { "+17",      NULL },
         { "+18",      NULL },
         { "+19",      NULL },
         { "+20",      NULL },
         { "+21",      NULL },
         { "+22",      NULL },
         { "+23",      NULL },
         { "+24",      NULL },
         { "+25",      NULL },
         { "+26",      NULL },
         { "+27",      NULL },
         { "+28",      NULL },
         { "+29",      NULL },
         { "+30",      NULL },
         { "+31",      NULL },
         { "+32",      NULL },
         { "+33",      NULL },
         { "+34",      NULL },
         { "+35",      NULL },
         { "+36",      NULL },
         { "+37",      NULL },
         { "+38",      NULL },
         { "+39",      NULL },
         { "+40",      NULL },
         { NULL, NULL},
      },
      "0"
   },
#endif
   {
      BEETLE_OPT(initial_scanline),
      BEETLE_OPT_INITIAL_SCANLINE_LABEL_BG,
      NULL,
      BEETLE_OPT_INITIAL_SCANLINE_INFO_0_BG,
      NULL,
      "video",
      {
         { "0",  OPTION_VAL_0_O48_BG },
         { "1",  NULL },
         { "2",  NULL },
         { "3",  NULL },
         { "4",  NULL },
         { "5",  NULL },
         { "6",  NULL },
         { "7",  NULL },
         { "8",  NULL },
         { "9",  NULL },
         { "10", NULL },
         { "11", NULL },
         { "12", NULL },
         { "13", NULL },
         { "14", NULL },
         { "15", NULL },
         { "16", NULL },
         { "17", NULL },
         { "18", NULL },
         { "19", NULL },
         { "20", NULL },
         { "21", NULL },
         { "22", NULL },
         { "23", NULL },
         { "24", NULL },
         { "25", NULL },
         { "26", NULL },
         { "27", NULL },
         { "28", NULL },
         { "29", NULL },
         { "30", NULL },
         { "31", NULL },
         { "32", NULL },
         { "33", NULL },
         { "34", NULL },
         { "35", NULL },
         { "36", NULL },
         { "37", NULL },
         { "38", NULL },
         { "39", NULL },
         { "40", NULL },
         { NULL, NULL },
      },
      "0"
   },
   {
      BEETLE_OPT(last_scanline),
      BEETLE_OPT_LAST_SCANLINE_LABEL_BG,
      NULL,
      BEETLE_OPT_LAST_SCANLINE_INFO_0_BG,
      NULL,
      "video",
      {
         { "210", NULL },
         { "211", NULL },
         { "212", NULL },
         { "213", NULL },
         { "214", NULL },
         { "215", NULL },
         { "216", NULL },
         { "217", NULL },
         { "218", NULL },
         { "219", NULL },
         { "220", NULL },
         { "221", NULL },
         { "222", NULL },
         { "223", NULL },
         { "224", NULL },
         { "225", NULL },
         { "226", NULL },
         { "227", NULL },
         { "228", NULL },
         { "229", NULL },
         { "230", NULL },
         { "231", NULL },
         { "232", NULL },
         { "233", NULL },
         { "234", NULL },
         { "235", NULL },
         { "236", NULL },
         { "237", NULL },
         { "238", NULL },
         { "239", OPTION_VAL_239_BG },
         { NULL, NULL },
      },
      "239"
   },
   {
      BEETLE_OPT(initial_scanline_pal),
      BEETLE_OPT_INITIAL_SCANLINE_PAL_LABEL_BG,
      NULL,
      BEETLE_OPT_INITIAL_SCANLINE_PAL_INFO_0_BG,
      NULL,
      "video",
      {
         { "0",  OPTION_VAL_0_O48_BG },
         { "1",  NULL },
         { "2",  NULL },
         { "3",  NULL },
         { "4",  NULL },
         { "5",  NULL },
         { "6",  NULL },
         { "7",  NULL },
         { "8",  NULL },
         { "9",  NULL },
         { "10", NULL },
         { "11", NULL },
         { "12", NULL },
         { "13", NULL },
         { "14", NULL },
         { "15", NULL },
         { "16", NULL },
         { "17", NULL },
         { "18", NULL },
         { "19", NULL },
         { "20", NULL },
         { "21", NULL },
         { "22", NULL },
         { "23", NULL },
         { "24", NULL },
         { "25", NULL },
         { "26", NULL },
         { "27", NULL },
         { "28", NULL },
         { "29", NULL },
         { "30", NULL },
         { "31", NULL },
         { "32", NULL },
         { "33", NULL },
         { "34", NULL },
         { "35", NULL },
         { "36", NULL },
         { "37", NULL },
         { "38", NULL },
         { "39", NULL },
         { "40", NULL },
         { NULL, NULL },
      },
      "0"
   },
   {
      BEETLE_OPT(last_scanline_pal),
      BEETLE_OPT_LAST_SCANLINE_PAL_LABEL_BG,
      NULL,
      BEETLE_OPT_LAST_SCANLINE_PAL_INFO_0_BG,
      NULL,
      "video",
      {
         { "230", NULL },
         { "231", NULL },
         { "232", NULL },
         { "233", NULL },
         { "234", NULL },
         { "235", NULL },
         { "236", NULL },
         { "237", NULL },
         { "238", NULL },
         { "239", NULL },
         { "240", NULL },
         { "241", NULL },
         { "242", NULL },
         { "243", NULL },
         { "244", NULL },
         { "245", NULL },
         { "246", NULL },
         { "247", NULL },
         { "248", NULL },
         { "249", NULL },
         { "250", NULL },
         { "251", NULL },
         { "252", NULL },
         { "253", NULL },
         { "254", NULL },
         { "255", NULL },
         { "256", NULL },
         { "257", NULL },
         { "258", NULL },
         { "259", NULL },
         { "260", NULL },
         { "261", NULL },
         { "262", NULL },
         { "263", NULL },
         { "264", NULL },
         { "265", NULL },
         { "266", NULL },
         { "267", NULL },
         { "268", NULL },
         { "269", NULL },
         { "270", NULL },
         { "271", NULL },
         { "272", NULL },
         { "273", NULL },
         { "274", NULL },
         { "275", NULL },
         { "276", NULL },
         { "277", NULL },
         { "278", NULL },
         { "279", NULL },
         { "280", NULL },
         { "281", NULL },
         { "282", NULL },
         { "283", NULL },
         { "284", NULL },
         { "285", NULL },
         { "286", NULL },
         { "287", OPTION_VAL_287_BG },
         { NULL, NULL },
      },
      "287"
   },
   { NULL, NULL, NULL, NULL, NULL, NULL, {{0}}, NULL },
};
struct retro_core_options_v2 options_bg = {
   option_cats_bg,
   option_defs_bg
};

/* RETRO_LANGUAGE_CA */

#define CATEGORY_SYSTEM_LABEL_CA "Sistema"
#define CATEGORY_SYSTEM_INFO_0_CA "Canvia les opcions de velocitat del processador."
#define CATEGORY_VIDEO_LABEL_CA "Vídeo"
#define CATEGORY_VIDEO_INFO_0_CA "Canvia la relació d'aspecte, retalls de la imatge, filtre de vídeo i l'omissió de fotogrames."
#define CATEGORY_PGXP_LABEL_CA "PGXP (Procés de transformació precisa de la geometria)"
#define CATEGORY_PGXP_INFO_0_CA "Aquestes opcions permeten millorar els gràfics per sobre del que permet la consola real, eliminant distorsions entre les textures i el Z-fighting (distorsió entre polígons)."
#define CATEGORY_OSD_LABEL_CA "Visualització en pantalla"
#define CATEGORY_OSD_INFO_0_CA "Canvia les notificacions que es mostraran en pantalla."
#define CATEGORY_INPUT_LABEL_CA "Entrada"
#define CATEGORY_INPUT_INFO_0_CA "Canvia pistola de llum, ratolí i configuració neGcon."
#define CATEGORY_MEMCARDS_LABEL_CA "Targeta de memòria"
#define CATEGORY_MEMCARDS_INFO_0_CA "Canvia la configuració de la targeta de memòria virtual."
#define CATEGORY_HACKS_LABEL_CA "Correccions de l'emulació"
#define CATEGORY_HACKS_INFO_0_CA "Canvia les opcions de precisió de l'emulació, que afecten el rendiment a baix nivell i a la compatibilitat."
#define BEETLE_OPT_CPU_FREQ_SCALE_LABEL_CA "Escalat de freqüència de la CPU (augmentar velocitat)"
#define BEETLE_OPT_CPU_FREQ_SCALE_INFO_0_CA "Augmenta (o disminueix) la velocitat la CPU emulada de PSX. Augmentar la velocitat pot eliminar les estrabades i millorar la taxa de fotogrames en alguns jocs a canvi d'incrementar els requisits de maquinari. Cal tenir en compte que alguns jocs tenen un limitador de fotogrames intern i no hi ha cap benefici d'augmentar la velocitat. Pot provocar que alguns jocs s'executin més ràpid."
#define OPTION_VAL_50_CA NULL
#define OPTION_VAL_60_CA NULL
#define OPTION_VAL_70_CA NULL
#define OPTION_VAL_80_CA NULL
#define OPTION_VAL_90_CA NULL
#define OPTION_VAL_100_CA "100% (Nadiu)"
#define OPTION_VAL_110_CA NULL
#define OPTION_VAL_120_CA NULL
#define OPTION_VAL_130_CA NULL
#define OPTION_VAL_140_CA NULL
#define OPTION_VAL_150_CA NULL
#define OPTION_VAL_160_CA NULL
#define OPTION_VAL_170_CA NULL
#define OPTION_VAL_180_CA NULL
#define OPTION_VAL_190_CA NULL
#define OPTION_VAL_200_CA NULL
#define OPTION_VAL_210_CA NULL
#define OPTION_VAL_220_CA NULL
#define OPTION_VAL_230_CA NULL
#define OPTION_VAL_240_CA NULL
#define OPTION_VAL_250_CA NULL
#define OPTION_VAL_260_CA NULL
#define OPTION_VAL_270_CA NULL
#define OPTION_VAL_280_CA NULL
#define OPTION_VAL_290_CA NULL
#define OPTION_VAL_300_CA NULL
#define OPTION_VAL_310_CA NULL
#define OPTION_VAL_320_CA NULL
#define OPTION_VAL_330_CA NULL
#define OPTION_VAL_340_CA NULL
#define OPTION_VAL_350_CA NULL
#define OPTION_VAL_360_CA NULL
#define OPTION_VAL_370_CA NULL
#define OPTION_VAL_380_CA NULL
#define OPTION_VAL_390_CA NULL
#define OPTION_VAL_400_CA NULL
#define OPTION_VAL_410_CA NULL
#define OPTION_VAL_420_CA NULL
#define OPTION_VAL_430_CA NULL
#define OPTION_VAL_440_CA NULL
#define OPTION_VAL_450_CA NULL
#define OPTION_VAL_460_CA NULL
#define OPTION_VAL_470_CA NULL
#define OPTION_VAL_480_CA NULL
#define OPTION_VAL_490_CA NULL
#define OPTION_VAL_500_CA NULL
#define OPTION_VAL_510_CA NULL
#define OPTION_VAL_520_CA NULL
#define OPTION_VAL_530_CA NULL
#define OPTION_VAL_540_CA NULL
#define OPTION_VAL_550_CA NULL
#define OPTION_VAL_560_CA NULL
#define OPTION_VAL_570_CA NULL
#define OPTION_VAL_580_CA NULL
#define OPTION_VAL_590_CA NULL
#define OPTION_VAL_600_CA NULL
#define OPTION_VAL_610_CA NULL
#define OPTION_VAL_620_CA NULL
#define OPTION_VAL_630_CA NULL
#define OPTION_VAL_640_CA NULL
#define OPTION_VAL_650_CA NULL
#define OPTION_VAL_660_CA NULL
#define OPTION_VAL_670_CA NULL
#define OPTION_VAL_680_CA NULL
#define OPTION_VAL_690_CA NULL
#define OPTION_VAL_700_CA NULL
#define OPTION_VAL_710_CA NULL
#define OPTION_VAL_720_CA NULL
#define OPTION_VAL_730_CA NULL
#define OPTION_VAL_740_CA NULL
#define OPTION_VAL_750_CA NULL
#define BEETLE_OPT_GPU_OVERCLOCK_LABEL_CA "Augmentar velocitat del rasteritzador de la GPU"
#define BEETLE_OPT_GPU_OVERCLOCK_INFO_0_CA "Augmenta la velocitat del rasteritzador 2D que es troba a la GPU emulada del PSX. No millora el renderitzat 3D i el seu efecte acostuma a ser limitat."
#define OPTION_VAL_1X_NATIVE_CA "1x (Natiu)"
#define OPTION_VAL_2X_CA NULL
#define OPTION_VAL_4X_CA NULL
#define OPTION_VAL_8X_CA NULL
#define OPTION_VAL_16X_CA NULL
#define OPTION_VAL_32X_CA NULL
#define BEETLE_OPT_GTE_OVERCLOCK_LABEL_CA "Força el processador GTE"
#define BEETLE_OPT_GTE_OVERCLOCK_INFO_0_CA "Redueix la velocitat de totes les operacions emulades del GTE (el processador de la CPU per gràfics 3D) a una latència constant d'un cicle. Aquells jocs que utilitzin molt el GTE poden veure un augment de la velocitat dels fotogrames i una millora en l'estabilitat."
#define BEETLE_OPT_CD_ACCESS_METHOD_LABEL_CA "Mètode d'accés al CD"
#define BEETLE_OPT_CD_ACCESS_METHOD_INFO_0_CA "'Sincrònic' emula el maquinari original, 'Asincrònic' pot reduir les estrebades en els dispositius que tinguin un dispositiu d'emmagatzematge lent i 'Precarregat' carrega la imatge de disc sencera en memòria en executar contingut, això pot millorar el temps de càrrega del joc a costa de retardar l'arrancada inicial del joc. 'Precarregat' pot produir problemes en sistemes amb poca memòria RAM."
#define OPTION_VAL_SYNC_CA "Sincrònic"
#define OPTION_VAL_ASYNC_CA "Asíncron"
#define OPTION_VAL_PRECACHE_CA "Precarregar"
#define BEETLE_OPT_CD_FASTLOAD_LABEL_CA "Velocitat de càrrega del CD"
#define BEETLE_OPT_CD_FASTLOAD_INFO_0_CA "Valors superiors a '2x (Nadiu)' poden reduir el temps de càrrega, però pot provocar errors de sincronització. Alguns jocs poden no funcionar correctament si aquest valor és molt alt."
#define OPTION_VAL_2X_NATIVE_CA "2x (Natiu)"
#define OPTION_VAL_6X_CA NULL
#define OPTION_VAL_10X_CA NULL
#define OPTION_VAL_12X_CA NULL
#define OPTION_VAL_14X_CA NULL
#define BEETLE_OPT_SKIP_BIOS_LABEL_CA "Omet la BIOS"
#define BEETLE_OPT_SKIP_BIOS_INFO_0_CA "Salta l'animació d'arrancada de la BIOS de PlayStation que sol aparèixer al carregar contingut. Nota: Activar aquesta opció pot provocar errors de compatibilitat (Jocs amb protecció anticòpia PAL, Saga Frontier,...)."
#define BEETLE_OPT_OVERRIDE_BIOS_LABEL_CA "Sobreescriu la BIOS"
#define BEETLE_OPT_OVERRIDE_BIOS_INFO_0_CA "Sobreescriu la regió estàndard específica de la BIOS per una sense regió. Cal reiniciar."
#define OPTION_VAL_PSXONPSP_CA NULL
#define OPTION_VAL_PS1_ROM_CA NULL
#define OPTION_VAL_OPENBIOS_CA NULL
#define BEETLE_OPT_REGION_LABEL_CA "Regió del sistema"
#define BEETLE_OPT_REGION_INFO_0_CA "Estableix la regió que es farà servir quan no es pot detectar el contingut automàticament (per exemple fitxers PS-X EXE). Els jocs basats en disc encara fan servir la regió incrustada en el disc. Es requereix reinici."
#define OPTION_VAL_AUTO_CA "Automàtic"
#define OPTION_VAL_NTSC_J_CA "NTSC-J (Japó)"
#define OPTION_VAL_NTSC_U_CA "NTSC-U (Amèrica del nord)"
#define OPTION_VAL_PAL_CA "PAL (Europa)"
#define BEETLE_OPT_INTERNAL_RESOLUTION_LABEL_CA "Resolució interna de la GPU"
#define BEETLE_OPT_INTERNAL_RESOLUTION_INFO_0_CA "Resolucions superiors a '1x (Natiu)' millora la definició dels models 3D a canvi d'incrementar els requisits de maquinari. Els models 2D no solen veure's afectats per aquest paràmetre."
#define BEETLE_OPT_RENDERER_LABEL_CA "Renderitzador"
#define BEETLE_OPT_RENDERER_INFO_0_CA "El renderitzador per programari és més precís, però té uns requeriments alts si s'executa el joc a una resolució interna superior a la de la targeta gràfica. Els renderitzadors per maquinari son menys precisos, però tenen un rendiment superior si es fan servir resolucions internes superiors i permeten millores gràfiques. 'Maquinari (Automàtic)' selecciona automàticament Vulkan o OpenGl segons el controlador de vídeo de libretro. Si el controlador no és compatible amb Vulkan o OpenGl 3.3, es farà servir el renderitzador per programari."
#define OPTION_VAL_HARDWARE_CA "Maquinari (Automàtic)"
#define OPTION_VAL_HARDWARE_GL_CA "Maquinari (OpenGL)"
#define OPTION_VAL_HARDWARE_VK_CA "Maquinari (Vulkan)"
#define OPTION_VAL_SOFTWARE_CA "Programari"
#define BEETLE_OPT_RENDERER_SOFTWARE_FB_LABEL_CA "Memòria intermèdia per programari"
#define BEETLE_OPT_RENDERER_SOFTWARE_FB_INFO_0_CA "Permet emular amb precisió els efectes de la memòria intermèdia de fotogrames (desenfocament de moviment, les batalles de FF7) en utilitzar els renderitzadors per maquinari executant una còpia del renderitzador per programari a resolució nativa en segon pla. En desactivar aquesta opció, s'ometran aquestes operacions (OpenGL) o es renderitzadaran a la GPU (Vulkan). Desactivar aquesta opció pot millorar el rendiment a costa de generar errors gràfics greus. En cas de dubte, deixa activada aquesta opció."
#define BEETLE_OPT_DEPTH_LABEL_CA "Profunditat interna dels colors"
#define BEETLE_OPT_DEPTH_INFO_0_CA "Una profunditat més gran pot reduir l'efecte de les bandes de color sense haver de fer servir l'efecte de tramatge. 16 bpp emula el maquinari original, però pot generar franges de color si l'opció de tramatge no està activada. Es recomana desactivar l'opció 'Tramatge' en seleccionar 32 bpp en aquesta opció."
#define OPTION_VAL_16BPP_NATIVE_CA "16 bpp (Natiu)"
#define OPTION_VAL_32BPP_CA NULL
#define BEETLE_OPT_DITHER_MODE_LABEL_CA "Tramatge"
#define BEETLE_OPT_DITHER_MODE_INFO_0_CA "'1x (Natiu)' simula el tramatge natiu a baixa resolució que feia servir el maquinari original per suavitzar els defectes de bandes de color visibles en la profunditat de colors nativa. 'Resolució interna' canvia l'escala del tramatge de la resolució interna configurada per produir un resultat més net. Es recomana desactivar aquesta opció en utilitzar una profunditat de colors de 32 bpp. Nota: Activar aquesta opció a Vulkan desactivarà la reducció de colors a la profunditat nativa i activarà una profunditat interna superior."
#define OPTION_VAL_INTERNAL_RESOLUTION_CA "Resolució Interna"
#define BEETLE_OPT_SCALED_UV_OFFSET_LABEL_CA "Compensació UV de textures"
#define BEETLE_OPT_SCALED_UV_OFFSET_INFO_0_CA "Mostra les textures dels polígons 3D amb una compensació per resolucions superiors a la resolució nativa. Redueix els pics en les textures, però pot produir altres errors gràfics."
#define BEETLE_OPT_FILTER_LABEL_CA "Filtratge de textures"
#define BEETLE_OPT_FILTER_INFO_0_CA "'Més proper' emula el maquinari original. 'Bilineal' i 'De tres punts' son filtres de suavitzat que redueixen la pixelació. 'SABR', 'xBR' i 'JIMC2' son filtres d'escalat que poden millorar la definició de les textures a canvi d'augmentar els requisits de maquinari. Només funciona en renderitzats per maquinari."
#define OPTION_VAL_NEAREST_CA "Més proper"
#define OPTION_VAL_SABR_CA NULL
#define OPTION_VAL_XBR_CA NULL
#define OPTION_VAL_BILINEAR_CA "Bilineal"
#define OPTION_VAL_3_POINT_CA "De 3 punts"
#define OPTION_VAL_JINC2_CA NULL
#define BEETLE_OPT_FILTER_EXCLUDE_SPRITE_LABEL_CA "Ometre sprites del filtratge"
#define BEETLE_OPT_FILTER_EXCLUDE_SPRITE_INFO_0_CA "No s'apliquen els filtres de textura als sprites. Permet prevenir errors gràfics en jocs amb el fons creat per sprites. Es pot utilitzar al mateix temps que el Suavitzat adaptatiu o un altre filtre de post-processat per obtenir millors resultats."
#define OPTION_VAL_OPAQUE_CA "Només els sprites opacs"
#define OPTION_VAL_ALL_CA "Opac i semitransparent"
#define BEETLE_OPT_FILTER_EXCLUDE_2D_POLYGON_LABEL_CA "Excloure polígons 2D del filtre"
#define BEETLE_OPT_FILTER_EXCLUDE_2D_POLYGON_INFO_0_CA "No s'apliquen els filtres de textura als polígons 2D. Els polígons 2D es detecten amb un mètode heurístic i pot haver-hi alguns errors gràfics. Es pot utilitzar conjuntament el Suavitzat Adaptatiu o un altre filtre de post-processat per obtenir millors resultats."
#define BEETLE_OPT_ADAPTIVE_SMOOTHING_LABEL_CA "Suavitzat adaptatiu"
#define BEETLE_OPT_ADAPTIVE_SMOOTHING_INFO_0_CA "Suavitza els gràfics i els elements de la interfície en 2D sense difuminar els objectes renderitzats en 3D. Només funciona amb el renderitzador Vulkan."
#define BEETLE_OPT_SUPER_SAMPLING_LABEL_CA "Sobremostreig (reducció a resolució nativa)"
#define BEETLE_OPT_SUPER_SAMPLING_INFO_0_CA "Redueix el contingut renderitzat a una resolució interna superior a la resolució nativa. En combinar aquest efecte amb els multiplicadors per augmentar la resolució interna, els jocs es poden mostrar amb objectes 3D amb les vores suavitzades en la resolució nativa. Genera millors resultats en aquells jocs que combinen 2D i 3D (per exemple personatges 3D sobre sobre fons prerenderitzats) i acompanyats de shaders de simulació de pantalles CRT. Només funciona amb el renderitzador Vulkan. Nota: Es recomana desactivar el tramatge en fer servir aquesta opció."
#define BEETLE_OPT_MSAA_LABEL_CA "Suavitzat de vores MSAA"
#define BEETLE_OPT_MSAA_INFO_0_CA "MSAA millora l'aspecte dels objectes 3D. Només funciona amb el renderitzador Vulkan."
#define OPTION_VAL_1X_CA NULL
#define BEETLE_OPT_MDEC_YUV_LABEL_CA "Filtre Croma MDEC YUV"
#define BEETLE_OPT_MDEC_YUV_INFO_0_CA "Millora la qualitat de la reproducció de les seqüències FMV reduint els artefactes dels macroblocs (quadrat i vores de serra). Només funciona amb el renderitzador Vulkan."
#define BEETLE_OPT_TRACK_TEXTURES_LABEL_CA "Rastreig de textures"
#define BEETLE_OPT_TRACK_TEXTURES_INFO_0_CA "És un requisit per poder abocar i canviar textures. Pot provocar errors en alguns jocs."
#define BEETLE_OPT_DUMP_TEXTURES_LABEL_CA "Abocar textures"
#define BEETLE_OPT_DUMP_TEXTURES_INFO_0_CA "Abocar les textures actuals a <cd>-texture-dump/"
#define BEETLE_OPT_REPLACE_TEXTURES_LABEL_CA "Substitueix les textures"
#define BEETLE_OPT_REPLACE_TEXTURES_INFO_0_CA "Reemplaça les textures fent servir les versions en alta qualitat (HD) que es troben a <cd>-texture-replacements/"
#define BEETLE_OPT_PGXP_MODE_LABEL_CA "Mode d'operació PGXP"
#define BEETLE_OPT_PGXP_MODE_INFO_0_CA "Renderitza els objectes 3D amb una precisió subpíxel, utilitzant coordenades dels vèrtexs de coma fixa per minimitzar la distorsió i tremolors dels objectes 3D que hi havia en el maquinari original. 'Només en memòria' dona menys problemes de compatibilitat i es recomana per ús general. 'Memòria + CPU (inestable)' pot reduir més les distorsions visuals, però augmenten els requisits de maquinari."
#define OPTION_VAL_MEMORY_ONLY_CA "Només en memòria"
#define OPTION_VAL_MEMORY_CPU_CA "Memòria + CPU (inestable)"
#define BEETLE_OPT_PGXP_2D_TOL_LABEL_CA "Tolerància de la geometria plana del PGXP"
#define BEETLE_OPT_PGXP_2D_TOL_INFO_0_CA "Amaga els errors més evidents en les operacions del PGXP: el valor especifica la tolerància en la qual es mantindrà els valors de PGXP en cas de geometries sense la informació de profunditat."
#define OPTION_VAL_0PX_CA NULL
#define OPTION_VAL_1PX_CA NULL
#define OPTION_VAL_2PX_CA NULL
#define OPTION_VAL_3PX_CA NULL
#define OPTION_VAL_4PX_CA NULL
#define OPTION_VAL_5PX_CA NULL
#define OPTION_VAL_6PX_CA NULL
#define OPTION_VAL_7PX_CA NULL
#define OPTION_VAL_8PX_CA NULL
#define BEETLE_OPT_PGXP_NCLIP_LABEL_CA "Selecció de primitives del PGXP"
#define BEETLE_OPT_PGXP_NCLIP_INFO_0_CA "Fes servir la implementació NCLIP de PGXP. Millora l'aparença reduint els forats en les geometries que tenen coordenades PGXP. Pot provocar errors en alguns jocs i en certes circumstàncies."
#define BEETLE_OPT_PGXP_VERTEX_LABEL_CA "Memòria cau dels vèrtexs del PGXP"
#define BEETLE_OPT_PGXP_VERTEX_INFO_0_CA "Escorcolla les posicions dels vèrtexs millorades pel PGXP perquè siguin escorcollades per dibuixar polígons. Pot millorar l'alineació dels objectes i reduir les vores, però pot produir defectes visuals. Ara mateix, es recomana desactivar aquesta opció. Només s'aplicarà si s'activa el mode d'operació del PGXP. Només funciona amb els renderitzadors per maquinari."
#define BEETLE_OPT_PGXP_TEXTURE_LABEL_CA "Texturitzat fidel a la perspectiva del PGXP"
#define BEETLE_OPT_PGXP_TEXTURE_INFO_0_CA "Canvia el mapeig de textures natiu de PSX per un mapeig de textures amb la perspectiva correcta. Elimina les distorsions per posició i la deformació de textures, produïnt textures ben alineades. Només s'aplica quan s'activa el mode d'operació del PGXP. Funciona en els renderitzadors per maquinari o per programari."
#define BEETLE_OPT_DISPLAY_INTERNAL_FPS_LABEL_CA "Mostrar FPS interns"
#define BEETLE_OPT_DISPLAY_INTERNAL_FPS_INFO_0_CA "Mostra la velocitat de fotogrames interna a la qual el sistema PlayStations emulat renderitza el contingut. Nota: És necessari activar les notificacions en pantalla."
#define BEETLE_OPT_DISPLAY_OSD_LABEL_CA "Mostrar missatges en pantalla"
#define BEETLE_OPT_DISPLAY_OSD_INFO_0_CA "Mostra els missatges en pantalla generats pel nucli."
#define BEETLE_OPT_DISPLAY_VRAM_LABEL_CA "Mostra la VRAM completa (depuració)"
#define BEETLE_OPT_DISPLAY_VRAM_INFO_0_CA "Mostra la memòria VRAM emulada de la consola. Només disponible en els renderitzadors OpenGL i Vulkan. Nota: Només per motius de desenvolupament, normalment ha d'estar desactivat."
#define BEETLE_OPT_ANALOG_CALIBRATION_LABEL_CA "Calibració automàtica dels controls analògics"
#define BEETLE_OPT_ANALOG_CALIBRATION_INFO_0_CA "Es calibraran en temps real els valors màxims d'entrada que es registrin per escalar els valors màxims que s'enviïn a l'emulador. Això serveix per jocs com Mega Man Legends 2, que esperen uns valors superiors als que poden enviar els controladors actuals. Es recomana que els joystics analògics es girin en tota l'extensió possible en executar el joc per a que el sistema de calibració funcioni el millor possible."
#define BEETLE_OPT_ANALOG_TOGGLE_LABEL_CA "Alternar el mode analògic del DualShock"
#define BEETLE_OPT_ANALOG_TOGGLE_INFO_0_CA "Si el dispositiu d'entrada està seleccionat com a DualShock, aquesta opció permet al controlador DualShock emulat alternar entre el mode digital o analògic del controlador original. El mode també es pot canviar amb una combinació de botons."
#define OPTION_VAL_ENABLED_ANALOG_CA "Analògic per defecte"
#define BEETLE_OPT_ANALOG_TOGGLE_COMBO_LABEL_CA "Combinació de botons del mode analògic del DualShock"
#define BEETLE_OPT_ANALOG_TOGGLE_COMBO_INFO_0_CA "Selecciona la combinació de botons que es farà servir per alternar entre el mode DIGITAL i ANALÒGIC pel controlador emulat DualShock."
#define OPTION_VAL_L1_L2_R1_R2_START_SELECT_CA "L1 + L2 + R1 + R2 + Start + Select "
#define OPTION_VAL_L1_R1_SELECT_CA NULL
#define OPTION_VAL_L1_R1_START_CA NULL
#define OPTION_VAL_L1_R1_L3_CA NULL
#define OPTION_VAL_L1_R1_R3_CA NULL
#define OPTION_VAL_L2_R2_SELECT_CA NULL
#define OPTION_VAL_L2_R2_START_CA NULL
#define OPTION_VAL_L2_R2_L3_CA NULL
#define OPTION_VAL_L2_R2_R3_CA NULL
#define OPTION_VAL_L3_R3_CA NULL
#define BEETLE_OPT_ANALOG_TOGGLE_HOLD_LABEL_CA "Temps d'espera per a les combinacions de botons en el mode analògic"
#define BEETLE_OPT_ANALOG_TOGGLE_HOLD_INFO_0_CA "Estableix el temps que cal mantenir premuts els botons de la combinació del mode analògic."
#define OPTION_VAL_0_CA "Espera de 0 segons"
#define OPTION_VAL_1_CA "Espera de 1 segon"
#define OPTION_VAL_2_CA "Espera de 2 segons"
#define OPTION_VAL_3_CA "Espera de 3 segons"
#define OPTION_VAL_4_CA "Espera de 4 segons"
#define OPTION_VAL_5_CA "Espera de 5 segons"
#define BEETLE_OPT_ENABLE_MULTITAP_PORT1_LABEL_CA "Toc múltiple en el port 1"
#define BEETLE_OPT_ENABLE_MULTITAP_PORT1_INFO_0_CA "Habilitar el sistema Multitap en el port 1."
#define BEETLE_OPT_ENABLE_MULTITAP_PORT2_LABEL_CA "Toc múltiple en el port 2"
#define BEETLE_OPT_ENABLE_MULTITAP_PORT2_INFO_0_CA "Habilitat el sistema Multitap en el port 2."
#define BEETLE_OPT_GUN_INPUT_MODE_LABEL_CA "Mode d'entrada de les pistoles"
#define BEETLE_OPT_GUN_INPUT_MODE_INFO_0_CA "Fes servir l'entrada del ratolí per 'Light Gun' o 'Touchscreen' si el dispositiu seleccionat és 'Guncon/G-Con 45' o 'Justifier'."
#define OPTION_VAL_LIGHTGUN_CA "Pistola de llum"
#define OPTION_VAL_TOUCHSCREEN_CA "Pantalla tàctil"
#define BEETLE_OPT_GUN_CURSOR_LABEL_CA "Mira de arma"
#define BEETLE_OPT_GUN_CURSOR_INFO_0_CA "Mostra un cursor a la pantalla si es fan servir els dispositius 'Guncon/G-Con 45' i 'Justifier'."
#define OPTION_VAL_CROSS_CA "Creu"
#define OPTION_VAL_DOT_CA "Punt"
#define BEETLE_OPT_CROSSHAIR_COLOR_P1_LABEL_CA "Color de la creueta de la pistola: Port 1"
#define BEETLE_OPT_CROSSHAIR_COLOR_P1_INFO_0_CA "Color de la creueta de la pistola de llum del port 1."
#define OPTION_VAL_RED_CA "Vermell"
#define OPTION_VAL_BLUE_CA "Blau"
#define OPTION_VAL_GREEN_CA "Verd"
#define OPTION_VAL_ORANGE_CA "Taronja"
#define OPTION_VAL_YELLOW_CA "Groc"
#define OPTION_VAL_CYAN_CA "Cian"
#define OPTION_VAL_PINK_CA "Rosa"
#define OPTION_VAL_PURPLE_CA "Lila"
#define OPTION_VAL_BLACK_CA "Negre"
#define OPTION_VAL_WHITE_CA "Blanc"
#define BEETLE_OPT_CROSSHAIR_COLOR_P2_LABEL_CA "Color de la creueta de la pistola: Port 2"
#define BEETLE_OPT_CROSSHAIR_COLOR_P2_INFO_0_CA "Color de la creueta de la pistola de llum del port 2."
#define BEETLE_OPT_MOUSE_SENSITIVITY_LABEL_CA "Sensibilitat del ratolí"
#define BEETLE_OPT_MOUSE_SENSITIVITY_INFO_0_CA "Selecciona la sensibilitat del dispositiu d'entrada ratolí."
#define OPTION_VAL_5_O42_CA NULL
#define OPTION_VAL_10_CA NULL
#define OPTION_VAL_15_CA NULL
#define OPTION_VAL_20_CA NULL
#define OPTION_VAL_25_CA NULL
#define OPTION_VAL_30_CA NULL
#define OPTION_VAL_35_CA NULL
#define OPTION_VAL_40_CA NULL
#define OPTION_VAL_45_CA NULL
#define OPTION_VAL_55_CA NULL
#define OPTION_VAL_65_CA NULL
#define OPTION_VAL_75_CA NULL
#define OPTION_VAL_85_CA NULL
#define OPTION_VAL_95_CA NULL
#define OPTION_VAL_100_O42_CA NULL
#define OPTION_VAL_105_CA NULL
#define OPTION_VAL_115_CA NULL
#define OPTION_VAL_125_CA NULL
#define OPTION_VAL_135_CA NULL
#define OPTION_VAL_145_CA NULL
#define OPTION_VAL_155_CA NULL
#define OPTION_VAL_165_CA NULL
#define OPTION_VAL_175_CA NULL
#define OPTION_VAL_185_CA NULL
#define OPTION_VAL_195_CA NULL
#define BEETLE_OPT_NEGCON_RESPONSE_LABEL_CA "Resposta neGcon Twist"
#define BEETLE_OPT_NEGCON_RESPONSE_INFO_0_CA "Selecciona el tipus de resposta de la palanca esquerra del RetroPad en simular l'acció de tòrcer un dispositiu d'entrada neGcon. El desplaçament de la palanca analògica pot ser assignada a la rotació del neGcon de forma lineal, quadràtica o cúbica. 'Quadràtica' dona més precisió que 'Lineal' en moviments més petits. 'Cúbica' augmenta la precisió en els moviments més petits, però exagera els moviments grans. Nota: Només es recomana fer servir 'Lineal' amb volants de carrera. Els controladors convencionals gestionene l'entrada analògica de forma molt diferent al mecanisme de rotació del neGcon, de manera que l'assignació 'Lineal' sobreamplifica els moviments petits, el que afecta a la precisió. En la majoria de jocs, 'Quadràtica' ofereix la millor aproximació al maquinari original."
#define OPTION_VAL_LINEAR_CA "Lineal"
#define OPTION_VAL_QUADRATIC_CA "Quadràtic"
#define OPTION_VAL_CUBIC_CA "Cúbic"
#define BEETLE_OPT_NEGCON_DEADZONE_LABEL_CA "Zona morta neGcon Twist"
#define BEETLE_OPT_NEGCON_DEADZONE_INFO_0_CA "Selecciona la zona morta del joystick analògic esquerre del RetroPad en simular l'acció de tòrcer un dispositiu d'entrada neGcon emulat. Serveix per evitar que el controlador registri moviments no realitzats. Nota: Molts jocs compatibles amb neGcon inclouen opcions dins del jocs per ajustar un valor de la zona morta en tòrcer. Aquest valor ha de configurar-se com a 0 o neutral en tot moment per a què no hi hagi un pèrdua de precisió. Qualsevol ajustament que sigui necessari només ha d'aplicar-se mitjançant aquesta opció del nucli, sobretot si 'Resposta en tòrcer el neGcon' està configurada com 'Quadràtica' o 'Cúbica'."
#define OPTION_VAL_0_O44_CA NULL
#define BEETLE_OPT_USE_MEDNAFEN_MEMCARD0_METHOD_LABEL_CA "Mètode de la targeta de memòria"
#define BEETLE_OPT_USE_MEDNAFEN_MEMCARD0_METHOD_INFO_0_CA "Desa el format de desat que es farà servir a la targeta de memòria 0. 'Mednafen' pot millorar la compatibilitat amb la versió independent de Mednafen. En utilitzar els fitxers desats amb Beetle PSC, els fitxers de libretro (.srm) i Mednafen (.mcr) son iguals internament, així poden intercanviar entre ells mateixos, canviant els noms."
#define OPTION_VAL_LIBRETRO_CA NULL
#define OPTION_VAL_MEDNAFEN_CA NULL
#define BEETLE_OPT_ENABLE_MEMCARD1_LABEL_CA "Ranura de la targeta de de memòria 2"
#define BEETLE_OPT_ENABLE_MEMCARD1_INFO_0_CA "Simula una segona targeta de memòria a la ranura . Quan està desactivada, els jocs només poden accedir a la targeta de memòria de la ranura . Nota: Alguns jocs necessiten aquesta opció per funcionar correctament (per exemple Codename Tenka). Nota: La targeta de memòria 2 es desa en el format Mednafen (.mcr)."
#define BEETLE_OPT_SHARED_MEMORY_CARDS_LABEL_CA "Targetes de memòria compartides"
#define BEETLE_OPT_SHARED_MEMORY_CARDS_INFO_0_CA "Fes servir la memòria compartida en compte de targetes de memòria separades per cada contingut. Nota: si el mètode de targeta de memòria està establert en 'Libretro', només afecta a la targeta de memòria de la ranura 2. Es requereix reinici."
#define BEETLE_OPT_MEMCARD_LEFT_INDEX_LABEL_CA "Índex de la targeta de memòria 1"
#define BEETLE_OPT_MEMCARD_LEFT_INDEX_INFO_0_CA "Hi ha una targeta de memòria carregada a la ranura esquerra. Aquesta opció només funciona si el mètode de la targeta de memòria està establert en Mednafen."
#define OPTION_VAL_0_O48_CA "0 (Per defecte)"
#define BEETLE_OPT_MEMCARD_RIGHT_INDEX_LABEL_CA "Índex de la targeta de memòria 2"
#define BEETLE_OPT_MEMCARD_RIGHT_INDEX_INFO_0_CA "Hi ha una targeta de memòria carregada a la ranura dreta. Aquesta opció només funciona si la ranura de la targeta de memòria 2 està activada."
#define OPTION_VAL_1_O49_CA "1 (Per defecte)"
#define BEETLE_OPT_LINE_RENDER_LABEL_CA "Correcció Line-to-Quad"
#define BEETLE_OPT_LINE_RENDER_INFO_0_CA "Alguns jocs (Doom, Hexen, Soul Blade,...) dibuixen línies horitzontal estirant triangles de 1 píxel d'alçada al llarg de la pantalla, que després son renderitzats com una columna de píxels en el maquinari original. Aquesta correcció detecta aquests triangles i els converteix en rectangles si és necessari, permeten que es mostrin correctament en els renderitzadors per maquinari i en resolucions internes superiors a la nativa. L'opció 'Agressiu' és necessària per alguns jocs (per exemple: Dark Forces, Duke Nukem) però pot produir defectes visuals en altres jocs. En cas de dubte, seleccionar 'Predeterminat'."
#define OPTION_VAL_DEFAULT_CA "Per defecte"
#define OPTION_VAL_AGGRESSIVE_CA "Agressiu"
#define BEETLE_OPT_WIDESCREEN_HACK_LABEL_CA "Correcció de pantalla panoràmica"
#define BEETLE_OPT_WIDESCREEN_HACK_INFO_0_CA "Renderitza el contingut 3D en format anamòrfic i fes que la sortida de la memòria intermèdia de fotogrames tingui un format panoràmic. Produeix millors resultats en jocs fets completament en 3D. Els elements 2D s'estiraran horitzontalment i deixaran d'estar alineats."
#define BEETLE_OPT_WIDESCREEN_HACK_ASPECT_RATIO_LABEL_CA "Correcció per a pantalles panoràmiques"
#define BEETLE_OPT_WIDESCREEN_HACK_ASPECT_RATIO_INFO_0_CA "Selecciona la relació d'aspecte que farà servir la correcció de pantalles panoràmiques."
#define OPTION_VAL_16_9_CA NULL
#define OPTION_VAL_16_10_CA NULL
#define OPTION_VAL_18_9_CA NULL
#define OPTION_VAL_19_9_CA NULL
#define OPTION_VAL_20_9_CA NULL
#define OPTION_VAL_21_9_CA NULL
#define OPTION_VAL_32_9_CA NULL
#define BEETLE_OPT_CPU_DYNAREC_LABEL_CA NULL
#define BEETLE_OPT_CPU_DYNAREC_INFO_0_CA "Recompila de forma dinàmica les instruccions de la CPU a instruccions natives. Més ràpid que l'intèrpret, pels intervals de la CPU són menys precisos i pot provocar errors."
#define OPTION_VAL_DISABLED_CA "Deshabilitat (intèrpret Beetle)"
#define OPTION_VAL_EXECUTE_CA "Màxim rendiment"
#define OPTION_VAL_EXECUTE_ONE_CA "Verificació dels intervals entre cicles"
#define OPTION_VAL_RUN_INTERPRETER_CA "Intèrpret Lightrec"
#define BEETLE_OPT_DYNAREC_INVALIDATE_LABEL_CA "Invalidar el codi Dynarec"
#define BEETLE_OPT_DYNAREC_INVALIDATE_INFO_0_CA "Alguns jocs necessiten invalidació 'Full', altres necessiten 'Només DMA'."
#define OPTION_VAL_FULL_CA "Complet"
#define OPTION_VAL_DMA_CA "Només DMA (Lleugerament més ràpid)"
#define BEETLE_OPT_DYNAREC_EVENTCYCLES_LABEL_CA "Cicles d'esdeveniments de DMA/GPU/MDEC/Cronòmetre de Dynarec"
#define BEETLE_OPT_DYNAREC_EVENTCYCLES_INFO_0_CA "Indica el nombre màxim de cicles executats per la CPU abans de comprovar la DMA, la GPU, el MDEC o el temporitzador. Un valor alt serà més ràpid, té menys impacte en l'intèrpret Beetle que en el dynarec."
#define OPTION_VAL_128_CA "128 (Per defecte)"
#define BEETLE_OPT_DYNAREC_SPU_SAMPLES_LABEL_CA "Mostres de la SPU mitjançant dynarec"
#define BEETLE_OPT_DYNAREC_SPU_SAMPLES_INFO_0_CA "Mostra el nombre màxim de mostres de la SPU que s'executaran abans de comprovar les actualitzacions de la SPU. Un valor alt serà més ràpid, però un valor que no sigui 1 pot provocar errors en el àudio d'alguns jocs."
#define BEETLE_OPT_CORE_TIMING_FPS_LABEL_CA "Cronòmetre de fotogrames per segon segons el nucli"
#define BEETLE_OPT_CORE_TIMING_FPS_INFO_0_CA "El canvi automàtic permet que el nucli canviï entre els senyals progressiu i entrellaçat."
#define OPTION_VAL_AUTO_TOGGLE_CA "Habilitar el canvi automàtic"
#define OPTION_VAL_FORCE_PROGRESSIVE_CA "Força la taxa progressiva"
#define OPTION_VAL_FORCE_INTERLACED_CA "Forçar trama entrellaçada"
#define BEETLE_OPT_DEINTERLACER_LABEL_CA "Mètode de desentrellaçat"
#define BEETLE_OPT_DEINTERLACER_INFO_0_CA "'Weave' mostra els dos camps en el mateix fotograma per tenir una resolució alta encara que la imatge no ho sigui, 'Bob' sacrifica claredat de resolució per claredat de moviment. 'Bob (Offset)' és Bob amb la duplicació del camp actual a tota la superfície com una compensació entre camps, això fa que es mantingui la resolució de sortida de libretro i es conserva la claredat de moviment. 'FastMAD (adaptatiu al moviment)' entrellaça els pixels fixs i interpola els píxels en moviment en l'eix vertical, ofereix una imatge amb l'altura complerta sense els efectes de 'Weave' o la pèrdua de resolució de 'Bob'. Si es desactiva aquesta opció, el renderitzador per programari farà servir el mateix comportament que el renderitzador per maquinari (Vulkan/OpenGL) rasteritzant totes les línies de la memòria virtual en cada fotograma i retrasant l'escaneig fins que acabi el fotograma, això elimina els efectes en alguns jocs entrellaçats. Nota: aquells programes que facin servir el DMA durant l'escaneig (com algunes demos PAL de més de 512 línies que fan servir trucs en la memòria intermèdia de fotogrames) poden mostrar línies paralitzades si aquesta opció està desactivada, actívala en aquests casos."
#define OPTION_VAL_WEAVE_CA NULL
#define OPTION_VAL_BOB_CA NULL
#define OPTION_VAL_BOB_OFFSET_CA NULL
#define OPTION_VAL_FASTMAD_CA "FastMAD (Moviment adaptatiu)"
#define BEETLE_OPT_FRAME_DUPING_LABEL_CA "Omissió de fotogrames"
#define BEETLE_OPT_FRAME_DUPING_INFO_0_CA "Si el nucli no té res nou a mostrar, repeteix l'últim fotograma."
#define BEETLE_OPT_PAL_VIDEO_TIMING_OVERRIDE_LABEL_CA "Anul·la la velocitat de sincronització PAL"
#define BEETLE_OPT_PAL_VIDEO_TIMING_OVERRIDE_INFO_0_CA "A causa de la diferència entre estàndards, els jocs PAL solen semblar més lents que les seves versions en NTSC, tant americanes com japoneses. Aquesta opció pot anul·lar la velocitat del vídeo PAL per intentar executar aquests jocs amb la velocitat de fotogrames NTSC."
#define BEETLE_OPT_ASPECT_RATIO_LABEL_CA "Relació d'aspecte del nucli"
#define BEETLE_OPT_ASPECT_RATIO_INFO_0_CA "Aquesta opció és ignorada si la correcció de pantalla panoràmica o mostrar VRAM completa està activada."
#define OPTION_VAL_CORRECTED_CA "Corregit"
#define OPTION_VAL_UNCORRECTED_CA "Sense corregir"
#define OPTION_VAL_4_3_CA "Força 4:3"
#define OPTION_VAL_NTSC_CA "Força NTSC"
#define BEETLE_OPT_CROP_OVERSCAN_LABEL_CA "Retallar el sobreescaneig"
#define BEETLE_OPT_CROP_OVERSCAN_INFO_0_CA "Imatge nativa o sense vores."
#define OPTION_VAL_STATIC_CA "Horitzontal"
#define OPTION_VAL_SMART_CA "Horitzontal + Vertical"
#define BEETLE_OPT_IMAGE_CROP_LABEL_CA "Retallada addicional"
#define BEETLE_OPT_IMAGE_CROP_INFO_0_CA "Redueix l'amplada de la imatge en un nombre concret de píxels."
#define OPTION_VAL_9PX_CA NULL
#define OPTION_VAL_10PX_CA NULL
#define OPTION_VAL_11PX_CA NULL
#define OPTION_VAL_12PX_CA NULL
#define OPTION_VAL_13PX_CA NULL
#define OPTION_VAL_14PX_CA NULL
#define OPTION_VAL_15PX_CA NULL
#define OPTION_VAL_16PX_CA NULL
#define OPTION_VAL_17PX_CA NULL
#define OPTION_VAL_18PX_CA NULL
#define OPTION_VAL_19PX_CA NULL
#define OPTION_VAL_20PX_CA NULL
#define BEETLE_OPT_IMAGE_OFFSET_LABEL_CA "Desplaçament de la imatge retallada"
#define BEETLE_OPT_IMAGE_OFFSET_INFO_0_CA "Mou horitzontalment cap a la dreta (positiu) o esquerra (negatiu) per un nombre específic de píxels. Pot corregir errors d'alineació. Només funciona en el renderitzador per programari."
#define OPTION_VAL_12PX_O64_CA NULL
#define OPTION_VAL_11PX_O64_CA NULL
#define OPTION_VAL_10PX_O64_CA NULL
#define OPTION_VAL_9PX_O64_CA NULL
#define OPTION_VAL_8PX_O64_CA NULL
#define OPTION_VAL_7PX_O64_CA NULL
#define OPTION_VAL_6PX_O64_CA NULL
#define OPTION_VAL_5PX_O64_CA NULL
#define OPTION_VAL_4PX_O64_CA NULL
#define OPTION_VAL_3PX_O64_CA NULL
#define OPTION_VAL_2PX_O64_CA NULL
#define OPTION_VAL_1PX_O64_CA NULL
#define OPTION_VAL_1PX_O64_0_CA NULL
#define OPTION_VAL_2PX_O64_0_CA NULL
#define OPTION_VAL_3PX_O64_0_CA NULL
#define OPTION_VAL_4PX_O64_0_CA NULL
#define OPTION_VAL_5PX_O64_0_CA NULL
#define OPTION_VAL_6PX_O64_0_CA NULL
#define OPTION_VAL_7PX_O64_0_CA NULL
#define OPTION_VAL_8PX_O64_0_CA NULL
#define OPTION_VAL_9PX_O64_0_CA NULL
#define OPTION_VAL_10PX_O64_0_CA NULL
#define OPTION_VAL_11PX_O64_0_CA NULL
#define OPTION_VAL_12PX_O64_0_CA NULL
#define BEETLE_OPT_IMAGE_OFFSET_CYCLES_LABEL_CA "Desplaçament horitzontal de la imatge (cicles de la GPU)"
#define BEETLE_OPT_IMAGE_OFFSET_CYCLES_INFO_0_CA "Els valors positius mouen la imatge cap a la dreta, els valors negatius mouen la imatge cap a l'esquerra. Només funciona amb el renderitzat per maquinari."
#define BEETLE_OPT_INITIAL_SCANLINE_LABEL_CA "Línia escanejada inicial - NTSC"
#define BEETLE_OPT_INITIAL_SCANLINE_INFO_0_CA "Primera línia d'escombrat que es mostrarà en executar contingut NTSC. Un valor superior a 0 reduirà l'alçada de les imatges generades, retallant píxels de la part superior. Pot servir per evitar l'efecte 'letterboxing'."
#define BEETLE_OPT_LAST_SCANLINE_LABEL_CA "Última linia escanejada - NTSC"
#define BEETLE_OPT_LAST_SCANLINE_INFO_0_CA "Última línia d'escombrat que es mostrarà en executar contingut NTSC. Un valor inferior a 239 reduirà l'alçada de les imatges generades, retallant píxels de la part inferior. Pot servir per evitar l'efecte 'letterboxing'."
#define OPTION_VAL_239_CA "239 (Per defecte)"
#define BEETLE_OPT_INITIAL_SCANLINE_PAL_LABEL_CA "Primera línia escanejada - PAL"
#define BEETLE_OPT_INITIAL_SCANLINE_PAL_INFO_0_CA "Primera línia d'escombrat que es mostrarà en executar contingut PAL. Un valor superior a 0 reduirà l'alçada de les imatges generades, retallant píxels de la part superior. Pot servir per evitar l'efecte 'letterboxing'."
#define BEETLE_OPT_LAST_SCANLINE_PAL_LABEL_CA "Última linia escanejada - PAL"
#define BEETLE_OPT_LAST_SCANLINE_PAL_INFO_0_CA "Última línia d'escombrat que es mostrarà en executar contingut PAL. Un valor inferior a 287 reduirà l'alçada de les imatges generades, retallant píxels de la part inferior. Pot servir per evitar l'efecte 'letterboxing'."
#define OPTION_VAL_287_CA "287 (Per defecte)"

struct retro_core_option_v2_category option_cats_ca[] = {
   {
      "system",
      CATEGORY_SYSTEM_LABEL_CA,
      CATEGORY_SYSTEM_INFO_0_CA
   },
   {
      "video",
      CATEGORY_VIDEO_LABEL_CA,
      CATEGORY_VIDEO_INFO_0_CA
   },
   {
      "pgxp",
      CATEGORY_PGXP_LABEL_CA,
      CATEGORY_PGXP_INFO_0_CA
   },
   {
      "osd",
      CATEGORY_OSD_LABEL_CA,
      CATEGORY_OSD_INFO_0_CA
   },
   {
      "input",
      CATEGORY_INPUT_LABEL_CA,
      CATEGORY_INPUT_INFO_0_CA
   },
   {
      "memcards",
      CATEGORY_MEMCARDS_LABEL_CA,
      CATEGORY_MEMCARDS_INFO_0_CA
   },
   {
      "hacks",
      CATEGORY_HACKS_LABEL_CA,
      CATEGORY_HACKS_INFO_0_CA
   },
   { NULL, NULL, NULL },
};
struct retro_core_option_v2_definition option_defs_ca[] = {
   {
      BEETLE_OPT(cpu_freq_scale),
      BEETLE_OPT_CPU_FREQ_SCALE_LABEL_CA,
      NULL,
      BEETLE_OPT_CPU_FREQ_SCALE_INFO_0_CA,
      NULL,
      "system",
      {
         { "50%",           OPTION_VAL_50_CA },
         { "60%",           OPTION_VAL_60_CA },
         { "70%",           OPTION_VAL_70_CA },
         { "80%",           OPTION_VAL_80_CA },
         { "90%",           OPTION_VAL_90_CA },
         { "100%",          OPTION_VAL_100_CA },
         { "110%",          OPTION_VAL_110_CA },
         { "120%",          OPTION_VAL_120_CA },
         { "130%",          OPTION_VAL_130_CA },
         { "140%",          OPTION_VAL_140_CA },
         { "150%",          OPTION_VAL_150_CA },
         { "160%",          OPTION_VAL_160_CA },
         { "170%",          OPTION_VAL_170_CA },
         { "180%",          OPTION_VAL_180_CA },
         { "190%",          OPTION_VAL_190_CA },
         { "200%",          OPTION_VAL_200_CA },
         { "210%",          OPTION_VAL_210_CA },
         { "220%",          OPTION_VAL_220_CA },
         { "230%",          OPTION_VAL_230_CA },
         { "240%",          OPTION_VAL_240_CA },
         { "250%",          OPTION_VAL_250_CA },
         { "260%",          OPTION_VAL_260_CA },
         { "270%",          OPTION_VAL_270_CA },
         { "280%",          OPTION_VAL_280_CA },
         { "290%",          OPTION_VAL_290_CA },
         { "300%",          OPTION_VAL_300_CA },
         { "310%",          OPTION_VAL_310_CA },
         { "320%",          OPTION_VAL_320_CA },
         { "330%",          OPTION_VAL_330_CA },
         { "340%",          OPTION_VAL_340_CA },
         { "350%",          OPTION_VAL_350_CA },
         { "360%",          OPTION_VAL_360_CA },
         { "370%",          OPTION_VAL_370_CA },
         { "380%",          OPTION_VAL_380_CA },
         { "390%",          OPTION_VAL_390_CA },
         { "400%",          OPTION_VAL_400_CA },
         { "410%",          OPTION_VAL_410_CA },
         { "420%",          OPTION_VAL_420_CA },
         { "430%",          OPTION_VAL_430_CA },
         { "440%",          OPTION_VAL_440_CA },
         { "450%",          OPTION_VAL_450_CA },
         { "460%",          OPTION_VAL_460_CA },
         { "470%",          OPTION_VAL_470_CA },
         { "480%",          OPTION_VAL_480_CA },
         { "490%",          OPTION_VAL_490_CA },
         { "500%",          OPTION_VAL_500_CA },
         { "510%",          OPTION_VAL_510_CA },
         { "520%",          OPTION_VAL_520_CA },
         { "530%",          OPTION_VAL_530_CA },
         { "540%",          OPTION_VAL_540_CA },
         { "550%",          OPTION_VAL_550_CA },
         { "560%",          OPTION_VAL_560_CA },
         { "570%",          OPTION_VAL_570_CA },
         { "580%",          OPTION_VAL_580_CA },
         { "590%",          OPTION_VAL_590_CA },
         { "600%",          OPTION_VAL_600_CA },
         { "610%",          OPTION_VAL_610_CA },
         { "620%",          OPTION_VAL_620_CA },
         { "630%",          OPTION_VAL_630_CA },
         { "640%",          OPTION_VAL_640_CA },
         { "650%",          OPTION_VAL_650_CA },
         { "660%",          OPTION_VAL_660_CA },
         { "670%",          OPTION_VAL_670_CA },
         { "680%",          OPTION_VAL_680_CA },
         { "690%",          OPTION_VAL_690_CA },
         { "700%",          OPTION_VAL_700_CA },
         { "710%",          OPTION_VAL_710_CA },
         { "720%",          OPTION_VAL_720_CA },
         { "730%",          OPTION_VAL_730_CA },
         { "740%",          OPTION_VAL_740_CA },
         { "750%",          OPTION_VAL_750_CA },
         { NULL, NULL },
      },
      "100%"
   },
   {
      BEETLE_OPT(gpu_overclock),
      BEETLE_OPT_GPU_OVERCLOCK_LABEL_CA,
      NULL,
      BEETLE_OPT_GPU_OVERCLOCK_INFO_0_CA,
      NULL,
      "system",
      {
         { "1x(native)", OPTION_VAL_1X_NATIVE_CA },
         { "2x",         OPTION_VAL_2X_CA },
         { "4x",         OPTION_VAL_4X_CA },
         { "8x",         OPTION_VAL_8X_CA },
         { "16x",        OPTION_VAL_16X_CA },
         { "32x",        OPTION_VAL_32X_CA },
         { NULL, NULL },
      },
      "1x(native)"
   },
   {
      BEETLE_OPT(gte_overclock),
      BEETLE_OPT_GTE_OVERCLOCK_LABEL_CA,
      NULL,
      BEETLE_OPT_GTE_OVERCLOCK_INFO_0_CA,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
#ifndef EMSCRIPTEN
   {
      BEETLE_OPT(cd_access_method),
      BEETLE_OPT_CD_ACCESS_METHOD_LABEL_CA,
      NULL,
      BEETLE_OPT_CD_ACCESS_METHOD_INFO_0_CA,
      NULL,
      "system",
      {
         { "sync",     OPTION_VAL_SYNC_CA },
         { "async",    OPTION_VAL_ASYNC_CA },
         { "precache", OPTION_VAL_PRECACHE_CA },
         { NULL, NULL },
      },
      "sync"
   },
#endif
   {
      BEETLE_OPT(cd_fastload),
      BEETLE_OPT_CD_FASTLOAD_LABEL_CA,
      NULL,
      BEETLE_OPT_CD_FASTLOAD_INFO_0_CA,
      NULL,
      "system",
      {
         { "2x(native)", OPTION_VAL_2X_NATIVE_CA },
         { "4x",          OPTION_VAL_4X_CA },
         { "6x",          OPTION_VAL_6X_CA },
         { "8x",          OPTION_VAL_8X_CA },
         { "10x",         OPTION_VAL_10X_CA },
         { "12x",         OPTION_VAL_12X_CA },
         { "14x",         OPTION_VAL_14X_CA },
         { NULL, NULL },
      },
      "2x(native)"
   },
   {
      BEETLE_OPT(skip_bios),
      BEETLE_OPT_SKIP_BIOS_LABEL_CA,
      NULL,
      BEETLE_OPT_SKIP_BIOS_INFO_0_CA,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(override_bios),
      BEETLE_OPT_OVERRIDE_BIOS_LABEL_CA,
      NULL,
      BEETLE_OPT_OVERRIDE_BIOS_INFO_0_CA,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "psxonpsp", OPTION_VAL_PSXONPSP_CA },
         { "ps1_rom",  OPTION_VAL_PS1_ROM_CA },
         { "openbios", OPTION_VAL_OPENBIOS_CA     },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(region),
      BEETLE_OPT_REGION_LABEL_CA,
      NULL,
      BEETLE_OPT_REGION_INFO_0_CA,
      NULL,
      "system",
      {
         { "auto",   OPTION_VAL_AUTO_CA },
         { "ntsc-j", OPTION_VAL_NTSC_J_CA },
         { "ntsc-u", OPTION_VAL_NTSC_U_CA },
         { "pal",    OPTION_VAL_PAL_CA },
         { NULL, NULL },
      },
      "auto"
   },
   {
      BEETLE_OPT(internal_resolution),
      BEETLE_OPT_INTERNAL_RESOLUTION_LABEL_CA,
      NULL,
      BEETLE_OPT_INTERNAL_RESOLUTION_INFO_0_CA,
      NULL,
      "video",
      {
         { "1x(native)", OPTION_VAL_1X_NATIVE_CA },
         { "2x",         OPTION_VAL_2X_CA },
         { "4x",         OPTION_VAL_4X_CA },
         { "8x",         OPTION_VAL_8X_CA },
         { "16x",        OPTION_VAL_16X_CA },
         { NULL, NULL },
      },
      "1x(native)"
   },
#if defined(HAVE_OPENGL) || defined(HAVE_OPENGLES) || defined(HAVE_VULKAN)
   {
      BEETLE_OPT(renderer),
      BEETLE_OPT_RENDERER_LABEL_CA,
      NULL,
      BEETLE_OPT_RENDERER_INFO_0_CA,
      NULL,
      "video",
      {
         { "hardware",    OPTION_VAL_HARDWARE_CA },
#if defined(HAVE_OPENGL) || defined(HAVE_OPENGLES)
         { "hardware_gl", OPTION_VAL_HARDWARE_GL_CA },
#endif
#if defined(HAVE_VULKAN)
         { "hardware_vk", OPTION_VAL_HARDWARE_VK_CA },
#endif
         { "software",    OPTION_VAL_SOFTWARE_CA },
         { NULL, NULL },
      },
      "hardware"
   },
   {
      BEETLE_OPT(renderer_software_fb),
      BEETLE_OPT_RENDERER_SOFTWARE_FB_LABEL_CA,
      NULL,
      BEETLE_OPT_RENDERER_SOFTWARE_FB_INFO_0_CA,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled"
   },
#endif
#if defined(HAVE_OPENGL) || defined(HAVE_OPENGLES)
   {
      BEETLE_OPT(depth),
      BEETLE_OPT_DEPTH_LABEL_CA,
      NULL,
      BEETLE_OPT_DEPTH_INFO_0_CA,
      NULL,
      "video",
      {
         { "16bpp(native)", OPTION_VAL_16BPP_NATIVE_CA },
         { "32bpp",         OPTION_VAL_32BPP_CA },
         { NULL, NULL },
      },
      "16bpp(native)"
   },
   // Sort of, it's more like 15-bit high color and 24-bit true color for visible output. The alpha channel is used for mask bit. Vulkan renderer uses ABGR1555_555 for 31 bits internal? FMVs are always 24-bit on all renderers like original hardware (BGR888, no alpha)
#endif
   {
      BEETLE_OPT(dither_mode),
      BEETLE_OPT_DITHER_MODE_LABEL_CA,
      NULL,
      BEETLE_OPT_DITHER_MODE_INFO_0_CA,
      NULL,
      "video",
      {
         { "1x(native)",          OPTION_VAL_1X_NATIVE_CA },
         { "internal resolution", OPTION_VAL_INTERNAL_RESOLUTION_CA },
         { "disabled",            NULL },
         { NULL, NULL },
      },
      "1x(native)"
   },
#ifdef HAVE_VULKAN
   {
      BEETLE_OPT(scaled_uv_offset),
      BEETLE_OPT_SCALED_UV_OFFSET_LABEL_CA,
      NULL,
      BEETLE_OPT_SCALED_UV_OFFSET_INFO_0_CA,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled"
   },
#endif
#if defined(HAVE_OPENGL) || defined(HAVE_OPENGLES) || defined(HAVE_VULKAN)
   {
      BEETLE_OPT(filter),
      BEETLE_OPT_FILTER_LABEL_CA,
      NULL,
      BEETLE_OPT_FILTER_INFO_0_CA,
      NULL,
      "video",
      {
         { "nearest",  OPTION_VAL_NEAREST_CA },
         { "SABR",     OPTION_VAL_SABR_CA },
         { "xBR",      OPTION_VAL_XBR_CA },
         { "bilinear", OPTION_VAL_BILINEAR_CA },
         { "3-point",  OPTION_VAL_3_POINT_CA },
         { "JINC2",    OPTION_VAL_JINC2_CA },
         { NULL, NULL },
      },
      "nearest"
   },
#ifdef HAVE_VULKAN
   {
      BEETLE_OPT(filter_exclude_sprite),
      BEETLE_OPT_FILTER_EXCLUDE_SPRITE_LABEL_CA,
      NULL,
      BEETLE_OPT_FILTER_EXCLUDE_SPRITE_INFO_0_CA,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "opaque", OPTION_VAL_OPAQUE_CA },
         { "all", OPTION_VAL_ALL_CA },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(filter_exclude_2d_polygon),
      BEETLE_OPT_FILTER_EXCLUDE_2D_POLYGON_LABEL_CA,
      NULL,
      BEETLE_OPT_FILTER_EXCLUDE_2D_POLYGON_INFO_0_CA,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "opaque", OPTION_VAL_OPAQUE_CA },
         { "all", OPTION_VAL_ALL_CA },
         { NULL, NULL },
      },
      "disabled"
   },
#endif
#endif
#ifdef HAVE_VULKAN
   {
      BEETLE_OPT(adaptive_smoothing),
      BEETLE_OPT_ADAPTIVE_SMOOTHING_LABEL_CA,
      NULL,
      BEETLE_OPT_ADAPTIVE_SMOOTHING_INFO_0_CA,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(super_sampling),
      BEETLE_OPT_SUPER_SAMPLING_LABEL_CA,
      NULL,
      BEETLE_OPT_SUPER_SAMPLING_INFO_0_CA,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(msaa),
      BEETLE_OPT_MSAA_LABEL_CA,
      NULL,
      BEETLE_OPT_MSAA_INFO_0_CA,
      NULL,
      "video",
      {
         { "1x",  OPTION_VAL_1X_CA },
         { "2x",  OPTION_VAL_2X_CA },
         { "4x",  OPTION_VAL_4X_CA },
         { "8x",  OPTION_VAL_8X_CA },
         { "16x", OPTION_VAL_16X_CA },
         { NULL, NULL },
      },
      "1x"
   },
   {
      BEETLE_OPT(mdec_yuv),
      BEETLE_OPT_MDEC_YUV_LABEL_CA,
      NULL,
      BEETLE_OPT_MDEC_YUV_INFO_0_CA,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(track_textures),
      BEETLE_OPT_TRACK_TEXTURES_LABEL_CA,
      NULL,
      BEETLE_OPT_TRACK_TEXTURES_INFO_0_CA,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
#ifdef TEXTURE_DUMPING_ENABLED
   {
      BEETLE_OPT(dump_textures),
      BEETLE_OPT_DUMP_TEXTURES_LABEL_CA,
      NULL,
      BEETLE_OPT_DUMP_TEXTURES_INFO_0_CA,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
#endif
   {
      BEETLE_OPT(replace_textures),
      BEETLE_OPT_REPLACE_TEXTURES_LABEL_CA,
      NULL,
      BEETLE_OPT_REPLACE_TEXTURES_INFO_0_CA,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
#endif
   {
      BEETLE_OPT(pgxp_mode),
      BEETLE_OPT_PGXP_MODE_LABEL_CA,
      NULL,
      BEETLE_OPT_PGXP_MODE_INFO_0_CA,
      NULL,
      "pgxp",
      {
         { "disabled",     NULL },
         { "memory only",  OPTION_VAL_MEMORY_ONLY_CA },
         { "memory + CPU", OPTION_VAL_MEMORY_CPU_CA },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(pgxp_2d_tol),
      BEETLE_OPT_PGXP_2D_TOL_LABEL_CA,
      NULL,
      BEETLE_OPT_PGXP_2D_TOL_INFO_0_CA,
      NULL,
      "pgxp",
      {
         { "disabled", NULL },
         { "0px", OPTION_VAL_0PX_CA },
         { "1px", OPTION_VAL_1PX_CA },
         { "2px", OPTION_VAL_2PX_CA },
         { "3px", OPTION_VAL_3PX_CA },
         { "4px", OPTION_VAL_4PX_CA },
         { "5px", OPTION_VAL_5PX_CA },
         { "6px", OPTION_VAL_6PX_CA },
         { "7px", OPTION_VAL_7PX_CA },
         { "8px", OPTION_VAL_8PX_CA },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(pgxp_nclip),
      BEETLE_OPT_PGXP_NCLIP_LABEL_CA,
      NULL,
      BEETLE_OPT_PGXP_NCLIP_INFO_0_CA,
      NULL,
      "pgxp",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
#if defined(HAVE_OPENGL) || defined(HAVE_OPENGLES) || defined(HAVE_VULKAN)
   {
      BEETLE_OPT(pgxp_vertex),
      BEETLE_OPT_PGXP_VERTEX_LABEL_CA,
      NULL,
      BEETLE_OPT_PGXP_VERTEX_INFO_0_CA,
      NULL,
      "pgxp",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
#endif
   {
      BEETLE_OPT(pgxp_texture),
      BEETLE_OPT_PGXP_TEXTURE_LABEL_CA,
      NULL,
      BEETLE_OPT_PGXP_TEXTURE_INFO_0_CA,
      NULL,
      "pgxp",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(display_internal_fps),
      BEETLE_OPT_DISPLAY_INTERNAL_FPS_LABEL_CA,
      NULL,
      BEETLE_OPT_DISPLAY_INTERNAL_FPS_INFO_0_CA,
      NULL,
      "osd",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(display_osd),
      BEETLE_OPT_DISPLAY_OSD_LABEL_CA,
      NULL,
      BEETLE_OPT_DISPLAY_OSD_INFO_0_CA,
      NULL,
      "osd",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled"
   },
#if defined(HAVE_OPENGL) || defined(HAVE_OPENGLES) || defined(HAVE_VULKAN)
   {
      BEETLE_OPT(display_vram),
      BEETLE_OPT_DISPLAY_VRAM_LABEL_CA,
      NULL,
      BEETLE_OPT_DISPLAY_VRAM_INFO_0_CA,
      NULL,
      "osd",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
#endif
   {
      BEETLE_OPT(analog_calibration),
      BEETLE_OPT_ANALOG_CALIBRATION_LABEL_CA,
      NULL,
      BEETLE_OPT_ANALOG_CALIBRATION_INFO_0_CA,
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(analog_toggle),
      BEETLE_OPT_ANALOG_TOGGLE_LABEL_CA,
      NULL,
      BEETLE_OPT_ANALOG_TOGGLE_INFO_0_CA,
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { "enabled-analog", OPTION_VAL_ENABLED_ANALOG_CA },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(analog_toggle_combo),
      BEETLE_OPT_ANALOG_TOGGLE_COMBO_LABEL_CA,
      NULL,
      BEETLE_OPT_ANALOG_TOGGLE_COMBO_INFO_0_CA,
      NULL,
      "input",
      {
         { "l1+l2+r1+r2+start+select", OPTION_VAL_L1_L2_R1_R2_START_SELECT_CA },
         { "l1+r1+select",             OPTION_VAL_L1_R1_SELECT_CA },
         { "l1+r1+start",              OPTION_VAL_L1_R1_START_CA },
         { "l1+r1+l3",                 OPTION_VAL_L1_R1_L3_CA },
         { "l1+r1+r3",                 OPTION_VAL_L1_R1_R3_CA },
         { "l2+r2+select",             OPTION_VAL_L2_R2_SELECT_CA },
         { "l2+r2+start",              OPTION_VAL_L2_R2_START_CA },
         { "l2+r2+l3",                 OPTION_VAL_L2_R2_L3_CA },
         { "l2+r2+r3",                 OPTION_VAL_L2_R2_R3_CA },
         { "l3+r3",                    OPTION_VAL_L3_R3_CA },
         { NULL, NULL },
      },
      "l1+r1+select"
   },
   {
      BEETLE_OPT(analog_toggle_hold),
      BEETLE_OPT_ANALOG_TOGGLE_HOLD_LABEL_CA,
      NULL,
      BEETLE_OPT_ANALOG_TOGGLE_HOLD_INFO_0_CA,
      NULL,
      "input",
      {
         { "0", OPTION_VAL_0_CA },
         { "1", OPTION_VAL_1_CA },
         { "2", OPTION_VAL_2_CA },
         { "3", OPTION_VAL_3_CA },
         { "4", OPTION_VAL_4_CA },
         { "5", OPTION_VAL_5_CA },
         { NULL, NULL },
      },
      "1"
   },
   {
      BEETLE_OPT(enable_multitap_port1),
      BEETLE_OPT_ENABLE_MULTITAP_PORT1_LABEL_CA,
      NULL,
      BEETLE_OPT_ENABLE_MULTITAP_PORT1_INFO_0_CA,
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(enable_multitap_port2),
      BEETLE_OPT_ENABLE_MULTITAP_PORT2_LABEL_CA,
      NULL,
      BEETLE_OPT_ENABLE_MULTITAP_PORT2_INFO_0_CA,
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(gun_input_mode),
      BEETLE_OPT_GUN_INPUT_MODE_LABEL_CA,
      NULL,
      BEETLE_OPT_GUN_INPUT_MODE_INFO_0_CA,
      NULL,
      "input",
      {
         { "lightgun",    OPTION_VAL_LIGHTGUN_CA },
         { "touchscreen", OPTION_VAL_TOUCHSCREEN_CA },
         { NULL, NULL },
      },
      "lightgun"
   },
   // Shouldn't the gun_input_mode just be Mouse vs. Touchscreen?
   {
      BEETLE_OPT(gun_cursor),
      BEETLE_OPT_GUN_CURSOR_LABEL_CA,
      NULL,
      BEETLE_OPT_GUN_CURSOR_INFO_0_CA,
      NULL,
      "input",
      {
         { "off",   "disabled" },
         { "cross", OPTION_VAL_CROSS_CA },
         { "dot",   OPTION_VAL_DOT_CA },
         { NULL, NULL },
      },
      "cross"
   },
   {
      BEETLE_OPT(crosshair_color_p1),
      BEETLE_OPT_CROSSHAIR_COLOR_P1_LABEL_CA,
      NULL,
      BEETLE_OPT_CROSSHAIR_COLOR_P1_INFO_0_CA,
      NULL,
      "input",
      {
         { "red",    OPTION_VAL_RED_CA },
         { "blue",   OPTION_VAL_BLUE_CA },
         { "green",  OPTION_VAL_GREEN_CA },
         { "orange", OPTION_VAL_ORANGE_CA },
         { "yellow", OPTION_VAL_YELLOW_CA },
         { "cyan",   OPTION_VAL_CYAN_CA },
         { "pink",   OPTION_VAL_PINK_CA },
         { "purple", OPTION_VAL_PURPLE_CA },
         { "black",  OPTION_VAL_BLACK_CA },
         { "white",  OPTION_VAL_WHITE_CA },
         { NULL, NULL },
      },
      "red"
   },
   {
      BEETLE_OPT(crosshair_color_p2),
      BEETLE_OPT_CROSSHAIR_COLOR_P2_LABEL_CA,
      NULL,
      BEETLE_OPT_CROSSHAIR_COLOR_P2_INFO_0_CA,
      NULL,
      "input",
      {
         { "blue",   OPTION_VAL_BLUE_CA },
         { "red",    OPTION_VAL_RED_CA },
         { "green",  OPTION_VAL_GREEN_CA },
         { "orange", OPTION_VAL_ORANGE_CA },
         { "yellow", OPTION_VAL_YELLOW_CA },
         { "cyan",   OPTION_VAL_CYAN_CA },
         { "pink",   OPTION_VAL_PINK_CA },
         { "purple", OPTION_VAL_PURPLE_CA },
         { "black",  OPTION_VAL_BLACK_CA },
         { "white",  OPTION_VAL_WHITE_CA },
         { NULL, NULL },
      },
      "blue"
   },
   {
      BEETLE_OPT(mouse_sensitivity),
      BEETLE_OPT_MOUSE_SENSITIVITY_LABEL_CA,
      NULL,
      BEETLE_OPT_MOUSE_SENSITIVITY_INFO_0_CA,
      NULL,
      "input",
      {
         { "5%",   OPTION_VAL_5_O42_CA },
         { "10%",  OPTION_VAL_10_CA },
         { "15%",  OPTION_VAL_15_CA },
         { "20%",  OPTION_VAL_20_CA },
         { "25%",  OPTION_VAL_25_CA },
         { "30%",  OPTION_VAL_30_CA },
         { "35%",  OPTION_VAL_35_CA },
         { "40%",  OPTION_VAL_40_CA },
         { "45%",  OPTION_VAL_45_CA },
         { "50%",  OPTION_VAL_50_CA },
         { "55%",  OPTION_VAL_55_CA },
         { "60%",  OPTION_VAL_60_CA },
         { "65%",  OPTION_VAL_65_CA },
         { "70%",  OPTION_VAL_70_CA },
         { "75%",  OPTION_VAL_75_CA },
         { "80%",  OPTION_VAL_80_CA },
         { "85%",  OPTION_VAL_85_CA },
         { "90%",  OPTION_VAL_90_CA },
         { "95%",  OPTION_VAL_95_CA },
         { "100%", OPTION_VAL_100_O42_CA },
         { "105%", OPTION_VAL_105_CA },
         { "110%", OPTION_VAL_110_CA },
         { "115%", OPTION_VAL_115_CA },
         { "120%", OPTION_VAL_120_CA },
         { "125%", OPTION_VAL_125_CA },
         { "130%", OPTION_VAL_130_CA },
         { "135%", OPTION_VAL_135_CA },
         { "140%", OPTION_VAL_140_CA },
         { "145%", OPTION_VAL_145_CA },
         { "150%", OPTION_VAL_150_CA },
         { "155%", OPTION_VAL_155_CA },
         { "160%", OPTION_VAL_160_CA },
         { "165%", OPTION_VAL_165_CA },
         { "170%", OPTION_VAL_170_CA },
         { "175%", OPTION_VAL_175_CA },
         { "180%", OPTION_VAL_180_CA },
         { "185%", OPTION_VAL_185_CA },
         { "190%", OPTION_VAL_190_CA },
         { "195%", OPTION_VAL_195_CA },
         { "200%", OPTION_VAL_200_CA },
         { NULL, NULL },
      },
      "100%"
   },
   {
      BEETLE_OPT(negcon_response),
      BEETLE_OPT_NEGCON_RESPONSE_LABEL_CA,
      NULL,
      BEETLE_OPT_NEGCON_RESPONSE_INFO_0_CA,
      NULL,
      "input",
      {
         { "linear",    OPTION_VAL_LINEAR_CA },
         { "quadratic", OPTION_VAL_QUADRATIC_CA },
         { "cubic",     OPTION_VAL_CUBIC_CA },
         { NULL, NULL },
      },
      "linear"
   },
   {
      BEETLE_OPT(negcon_deadzone),
      BEETLE_OPT_NEGCON_DEADZONE_LABEL_CA,
      NULL,
      BEETLE_OPT_NEGCON_DEADZONE_INFO_0_CA,
      NULL,
      "input",
      {
         { "0%",  OPTION_VAL_0_O44_CA },
         { "5%",  OPTION_VAL_5_O42_CA },
         { "10%", OPTION_VAL_10_CA },
         { "15%", OPTION_VAL_15_CA },
         { "20%", OPTION_VAL_20_CA },
         { "25%", OPTION_VAL_25_CA },
         { "30%", OPTION_VAL_30_CA },
         { NULL, NULL },
      },
      "0%"
   },
   {
      BEETLE_OPT(use_mednafen_memcard0_method),
      BEETLE_OPT_USE_MEDNAFEN_MEMCARD0_METHOD_LABEL_CA,
      NULL,
      BEETLE_OPT_USE_MEDNAFEN_MEMCARD0_METHOD_INFO_0_CA,
      NULL,
      "memcards",
      {
         { "libretro", OPTION_VAL_LIBRETRO_CA },
         { "mednafen", OPTION_VAL_MEDNAFEN_CA },
         { NULL, NULL },
      },
      "libretro"
   },
   {
      BEETLE_OPT(enable_memcard1),
      BEETLE_OPT_ENABLE_MEMCARD1_LABEL_CA,
      NULL,
      BEETLE_OPT_ENABLE_MEMCARD1_INFO_0_CA,
      NULL,
      "memcards",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(shared_memory_cards),
      BEETLE_OPT_SHARED_MEMORY_CARDS_LABEL_CA,
      NULL,
      BEETLE_OPT_SHARED_MEMORY_CARDS_INFO_0_CA,
      NULL,
      "memcards",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(memcard_left_index),
      BEETLE_OPT_MEMCARD_LEFT_INDEX_LABEL_CA,
      NULL,
      BEETLE_OPT_MEMCARD_LEFT_INDEX_INFO_0_CA,
      NULL,
      "memcards",
      {
         { "0",  OPTION_VAL_0_O48_CA },
         { "1",  NULL },
         { "2",  NULL },
         { "3",  NULL },
         { "4",  NULL },
         { "5",  NULL },
         { "6",  NULL },
         { "7",  NULL },
         { "8",  NULL },
         { "9",  NULL },
         { "10", NULL },
         { "11", NULL },
         { "12", NULL },
         { "13", NULL },
         { "14", NULL },
         { "15", NULL },
         { "16", NULL },
         { "17", NULL },
         { "18", NULL },
         { "19", NULL },
         { "20", NULL },
         { "21", NULL },
         { "22", NULL },
         { "23", NULL },
         { "24", NULL },
         { "25", NULL },
         { "26", NULL },
         { "27", NULL },
         { "28", NULL },
         { "29", NULL },
         { "30", NULL },
         { "31", NULL },
         { "32", NULL },
         { "33", NULL },
         { "34", NULL },
         { "35", NULL },
         { "36", NULL },
         { "37", NULL },
         { "38", NULL },
         { "39", NULL },
         { "40", NULL },
         { "41", NULL },
         { "42", NULL },
         { "43", NULL },
         { "44", NULL },
         { "45", NULL },
         { "46", NULL },
         { "47", NULL },
         { "48", NULL },
         { "49", NULL },
         { "50", NULL },
         { "51", NULL },
         { "52", NULL },
         { "53", NULL },
         { "54", NULL },
         { "55", NULL },
         { "56", NULL },
         { "57", NULL },
         { "58", NULL },
         { "59", NULL },
         { "60", NULL },
         { "61", NULL },
         { "62", NULL },
         { "63", NULL },
         { NULL, NULL },
      },
      "0"
   },
   {
      BEETLE_OPT(memcard_right_index),
      BEETLE_OPT_MEMCARD_RIGHT_INDEX_LABEL_CA,
      NULL,
      BEETLE_OPT_MEMCARD_RIGHT_INDEX_INFO_0_CA,
      NULL,
      "memcards",
      {
         { "0",  NULL },
         { "1",  OPTION_VAL_1_O49_CA },
         { "2",  NULL },
         { "3",  NULL },
         { "4",  NULL },
         { "5",  NULL },
         { "6",  NULL },
         { "7",  NULL },
         { "8",  NULL },
         { "9",  NULL },
         { "10", NULL },
         { "11", NULL },
         { "12", NULL },
         { "13", NULL },
         { "14", NULL },
         { "15", NULL },
         { "16", NULL },
         { "17", NULL },
         { "18", NULL },
         { "19", NULL },
         { "20", NULL },
         { "21", NULL },
         { "22", NULL },
         { "23", NULL },
         { "24", NULL },
         { "25", NULL },
         { "26", NULL },
         { "27", NULL },
         { "28", NULL },
         { "29", NULL },
         { "30", NULL },
         { "31", NULL },
         { "32", NULL },
         { "33", NULL },
         { "34", NULL },
         { "35", NULL },
         { "36", NULL },
         { "37", NULL },
         { "38", NULL },
         { "39", NULL },
         { "40", NULL },
         { "41", NULL },
         { "42", NULL },
         { "43", NULL },
         { "44", NULL },
         { "45", NULL },
         { "46", NULL },
         { "47", NULL },
         { "48", NULL },
         { "49", NULL },
         { "50", NULL },
         { "51", NULL },
         { "52", NULL },
         { "53", NULL },
         { "54", NULL },
         { "55", NULL },
         { "56", NULL },
         { "57", NULL },
         { "58", NULL },
         { "59", NULL },
         { "60", NULL },
         { "61", NULL },
         { "62", NULL },
         { "63", NULL },
         { NULL, NULL },
      },
      "1"
   },
   {
      BEETLE_OPT(line_render),
      BEETLE_OPT_LINE_RENDER_LABEL_CA,
      NULL,
      BEETLE_OPT_LINE_RENDER_INFO_0_CA,
      NULL,
      "hacks",
      {
         { "default",    OPTION_VAL_DEFAULT_CA },
         { "aggressive", OPTION_VAL_AGGRESSIVE_CA },
         { "disabled",   NULL },
         { NULL, NULL },
      },
      "default"
   },
   {
      BEETLE_OPT(widescreen_hack),
      BEETLE_OPT_WIDESCREEN_HACK_LABEL_CA,
      NULL,
      BEETLE_OPT_WIDESCREEN_HACK_INFO_0_CA,
      NULL,
      "hacks",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(widescreen_hack_aspect_ratio),
      BEETLE_OPT_WIDESCREEN_HACK_ASPECT_RATIO_LABEL_CA,
      NULL,
      BEETLE_OPT_WIDESCREEN_HACK_ASPECT_RATIO_INFO_0_CA,
      NULL,
      "hacks",
      {
         { "16:9",  OPTION_VAL_16_9_CA },
         { "16:10", OPTION_VAL_16_10_CA },
         { "18:9",  OPTION_VAL_18_9_CA },
         { "19:9",  OPTION_VAL_19_9_CA },
         { "20:9",  OPTION_VAL_20_9_CA },
         { "21:9",  OPTION_VAL_21_9_CA }, // 64:27
         { "32:9",  OPTION_VAL_32_9_CA },
         { NULL,    NULL },
      },
      "16:9"
   },
#if defined(HAVE_LIGHTREC)
   {
      BEETLE_OPT(cpu_dynarec),
      BEETLE_OPT_CPU_DYNAREC_LABEL_CA,
      NULL,
      BEETLE_OPT_CPU_DYNAREC_INFO_0_CA,
      NULL,
      "hacks",
      {
         { "disabled",        OPTION_VAL_DISABLED_CA },
         { "execute",         OPTION_VAL_EXECUTE_CA },
         { "execute_one",     OPTION_VAL_EXECUTE_ONE_CA },
         { "run_interpreter", OPTION_VAL_RUN_INTERPRETER_CA },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(dynarec_invalidate),
      BEETLE_OPT_DYNAREC_INVALIDATE_LABEL_CA,
      NULL,
      BEETLE_OPT_DYNAREC_INVALIDATE_INFO_0_CA,
      NULL,
      "hacks",
      {
         { "full", OPTION_VAL_FULL_CA },
         { "dma",  OPTION_VAL_DMA_CA },
         { NULL, NULL },
      },
      "full"
   },
   {
      BEETLE_OPT(dynarec_eventcycles),
      BEETLE_OPT_DYNAREC_EVENTCYCLES_LABEL_CA,
      NULL,
      BEETLE_OPT_DYNAREC_EVENTCYCLES_INFO_0_CA,
      NULL,
      "hacks",
      {
         { "128",  OPTION_VAL_128_CA },
         { "256",  NULL },
         { "384",  NULL },
         { "512",  NULL },
         { "640",  NULL },
         { "768",  NULL },
         { "896",  NULL },
         { "1024", NULL },
         { "1152", NULL },
         { "1280", NULL },
         { "1408", NULL },
         { "1536", NULL },
         { "1664", NULL },
         { "1792", NULL },
         { "1920", NULL },
         { "2048", NULL },
         { NULL, NULL },
      },
      "128"
   },
   {
      BEETLE_OPT(dynarec_spu_samples),
      BEETLE_OPT_DYNAREC_SPU_SAMPLES_LABEL_CA,
      NULL,
      BEETLE_OPT_DYNAREC_SPU_SAMPLES_INFO_0_CA,
      NULL,
      "hacks",
      {
         { "1",  OPTION_VAL_1_O49_CA },
         { "4",  NULL },
         { "16", NULL },
         { NULL, NULL },
      },
      "1"
   },
#endif
   {
      BEETLE_OPT(core_timing_fps),
      BEETLE_OPT_CORE_TIMING_FPS_LABEL_CA,
      NULL,
      BEETLE_OPT_CORE_TIMING_FPS_INFO_0_CA,
      NULL,
      "video",
      {
         { "auto_toggle",       OPTION_VAL_AUTO_TOGGLE_CA },
         { "force_progressive", OPTION_VAL_FORCE_PROGRESSIVE_CA },
         { "force_interlaced",  OPTION_VAL_FORCE_INTERLACED_CA },
      },
      "auto_toggle"
   },
   {
      BEETLE_OPT(deinterlacer),
      BEETLE_OPT_DEINTERLACER_LABEL_CA,
      NULL,
      BEETLE_OPT_DEINTERLACER_INFO_0_CA,
      NULL,
      "video",
      {
         { "weave",      OPTION_VAL_WEAVE_CA },
         { "bob",        OPTION_VAL_BOB_CA },
         { "bob_offset", OPTION_VAL_BOB_OFFSET_CA },
         { "fastmad",    OPTION_VAL_FASTMAD_CA },
         { "off",        "Off" },
         { NULL, NULL },
      },
      "weave"
   },
   {
      BEETLE_OPT(frame_duping),
      BEETLE_OPT_FRAME_DUPING_LABEL_CA,
      NULL,
      BEETLE_OPT_FRAME_DUPING_INFO_0_CA,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(pal_video_timing_override),
      BEETLE_OPT_PAL_VIDEO_TIMING_OVERRIDE_LABEL_CA,
      NULL,
      BEETLE_OPT_PAL_VIDEO_TIMING_OVERRIDE_INFO_0_CA,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(aspect_ratio),
      BEETLE_OPT_ASPECT_RATIO_LABEL_CA,
      NULL,
      BEETLE_OPT_ASPECT_RATIO_INFO_0_CA,
      NULL,
      "video",
      {
         { "corrected", OPTION_VAL_CORRECTED_CA },
         { "uncorrected", OPTION_VAL_UNCORRECTED_CA },
         { "4:3",  OPTION_VAL_4_3_CA },
         { "ntsc", OPTION_VAL_NTSC_CA },
      },
      "corrected"
   },
   {
      BEETLE_OPT(crop_overscan),
      BEETLE_OPT_CROP_OVERSCAN_LABEL_CA,
      NULL,
      BEETLE_OPT_CROP_OVERSCAN_INFO_0_CA,
      NULL,
      "video",
      {
         { "disabled", "Disabled" },
         { "static",   OPTION_VAL_STATIC_CA },
         { "smart",    OPTION_VAL_SMART_CA },
         { NULL, NULL },
      },
      "smart"
   },
   {
      BEETLE_OPT(image_crop),
      BEETLE_OPT_IMAGE_CROP_LABEL_CA,
      NULL,
      BEETLE_OPT_IMAGE_CROP_INFO_0_CA,
      NULL,
      "video",
      {
         { "disabled", "0" },
         { "1px",      OPTION_VAL_1PX_CA },
         { "2px",      OPTION_VAL_2PX_CA },
         { "3px",      OPTION_VAL_3PX_CA },
         { "4px",      OPTION_VAL_4PX_CA },
         { "5px",      OPTION_VAL_5PX_CA },
         { "6px",      OPTION_VAL_6PX_CA },
         { "7px",      OPTION_VAL_7PX_CA },
         { "8px",      OPTION_VAL_8PX_CA },
         { "9px",      OPTION_VAL_9PX_CA },
         { "10px",     OPTION_VAL_10PX_CA },
         { "11px",     OPTION_VAL_11PX_CA },
         { "12px",     OPTION_VAL_12PX_CA },
         { "13px",     OPTION_VAL_13PX_CA },
         { "14px",     OPTION_VAL_14PX_CA },
         { "15px",     OPTION_VAL_15PX_CA },
         { "16px",     OPTION_VAL_16PX_CA },
         { "17px",     OPTION_VAL_17PX_CA },
         { "18px",     OPTION_VAL_18PX_CA },
         { "19px",     OPTION_VAL_19PX_CA },
         { "20px",     OPTION_VAL_20PX_CA },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(image_offset),
      BEETLE_OPT_IMAGE_OFFSET_LABEL_CA,
      NULL,
      BEETLE_OPT_IMAGE_OFFSET_INFO_0_CA,
      NULL,
      "video",
      {
         { "-12px",    OPTION_VAL_12PX_O64_CA },
         { "-11px",    OPTION_VAL_11PX_O64_CA },
         { "-10px",    OPTION_VAL_10PX_O64_CA },
         { "-9px",     OPTION_VAL_9PX_O64_CA },
         { "-8px",     OPTION_VAL_8PX_O64_CA },
         { "-7px",     OPTION_VAL_7PX_O64_CA },
         { "-6px",     OPTION_VAL_6PX_O64_CA },
         { "-5px",     OPTION_VAL_5PX_O64_CA },
         { "-4px",     OPTION_VAL_4PX_O64_CA },
         { "-3px",     OPTION_VAL_3PX_O64_CA },
         { "-2px",     OPTION_VAL_2PX_O64_CA },
         { "-1px",     OPTION_VAL_1PX_O64_CA },
         { "disabled", "0" },
         { "+1px",     OPTION_VAL_1PX_O64_0_CA },
         { "+2px",     OPTION_VAL_2PX_O64_0_CA },
         { "+3px",     OPTION_VAL_3PX_O64_0_CA },
         { "+4px",     OPTION_VAL_4PX_O64_0_CA },
         { "+5px",     OPTION_VAL_5PX_O64_0_CA },
         { "+6px",     OPTION_VAL_6PX_O64_0_CA },
         { "+7px",     OPTION_VAL_7PX_O64_0_CA },
         { "+8px",     OPTION_VAL_8PX_O64_0_CA },
         { "+9px",     OPTION_VAL_9PX_O64_0_CA },
         { "+10px",    OPTION_VAL_10PX_O64_0_CA },
         { "+11px",    OPTION_VAL_11PX_O64_0_CA },
         { "+12px",    OPTION_VAL_12PX_O64_0_CA },
         { NULL, NULL },
      },
      "disabled"
   },
#if defined(HAVE_OPENGL) || defined(HAVE_OPENGLES) || defined(HAVE_VULKAN)
   {
      BEETLE_OPT(image_offset_cycles),
      BEETLE_OPT_IMAGE_OFFSET_CYCLES_LABEL_CA,
      NULL,
      BEETLE_OPT_IMAGE_OFFSET_CYCLES_INFO_0_CA,
      NULL,
      "video",
      {
         { "-40",      NULL },
         { "-39",      NULL },
         { "-38",      NULL },
         { "-37",      NULL },
         { "-36",      NULL },
         { "-35",      NULL },
         { "-34",      NULL },
         { "-33",      NULL },
         { "-32",      NULL },
         { "-31",      NULL },
         { "-30",      NULL },
         { "-29",      NULL },
         { "-28",      NULL },
         { "-27",      NULL },
         { "-26",      NULL },
         { "-25",      NULL },
         { "-24",      NULL },
         { "-23",      NULL },
         { "-22",      NULL },
         { "-21",      NULL },
         { "-20",      NULL },
         { "-19",      NULL },
         { "-18",      NULL },
         { "-17",      NULL },
         { "-16",      NULL },
         { "-15",      NULL },
         { "-14",      NULL },
         { "-13",      NULL },
         { "-12",      NULL },
         { "-11",      NULL },
         { "-10",      NULL },
         { "-9",       NULL },
         { "-8",       NULL },
         { "-7",       NULL },
         { "-6",       NULL },
         { "-5",       NULL },
         { "-4",       NULL },
         { "-3",       NULL },
         { "-2",       NULL },
         { "-1",       NULL },
         { "0",        "0" },
         { "+1",       NULL },
         { "+2",       NULL },
         { "+3",       NULL },
         { "+4",       NULL },
         { "+5",       NULL },
         { "+6",       NULL },
         { "+7",       NULL },
         { "+8",       NULL },
         { "+9",       NULL },
         { "+10",      NULL },
         { "+11",      NULL },
         { "+12",      NULL },
         { "+13",      NULL },
         { "+14",      NULL },
         { "+15",      NULL },
         { "+16",      NULL },
         { "+17",      NULL },
         { "+18",      NULL },
         { "+19",      NULL },
         { "+20",      NULL },
         { "+21",      NULL },
         { "+22",      NULL },
         { "+23",      NULL },
         { "+24",      NULL },
         { "+25",      NULL },
         { "+26",      NULL },
         { "+27",      NULL },
         { "+28",      NULL },
         { "+29",      NULL },
         { "+30",      NULL },
         { "+31",      NULL },
         { "+32",      NULL },
         { "+33",      NULL },
         { "+34",      NULL },
         { "+35",      NULL },
         { "+36",      NULL },
         { "+37",      NULL },
         { "+38",      NULL },
         { "+39",      NULL },
         { "+40",      NULL },
         { NULL, NULL},
      },
      "0"
   },
#endif
   {
      BEETLE_OPT(initial_scanline),
      BEETLE_OPT_INITIAL_SCANLINE_LABEL_CA,
      NULL,
      BEETLE_OPT_INITIAL_SCANLINE_INFO_0_CA,
      NULL,
      "video",
      {
         { "0",  OPTION_VAL_0_O48_CA },
         { "1",  NULL },
         { "2",  NULL },
         { "3",  NULL },
         { "4",  NULL },
         { "5",  NULL },
         { "6",  NULL },
         { "7",  NULL },
         { "8",  NULL },
         { "9",  NULL },
         { "10", NULL },
         { "11", NULL },
         { "12", NULL },
         { "13", NULL },
         { "14", NULL },
         { "15", NULL },
         { "16", NULL },
         { "17", NULL },
         { "18", NULL },
         { "19", NULL },
         { "20", NULL },
         { "21", NULL },
         { "22", NULL },
         { "23", NULL },
         { "24", NULL },
         { "25", NULL },
         { "26", NULL },
         { "27", NULL },
         { "28", NULL },
         { "29", NULL },
         { "30", NULL },
         { "31", NULL },
         { "32", NULL },
         { "33", NULL },
         { "34", NULL },
         { "35", NULL },
         { "36", NULL },
         { "37", NULL },
         { "38", NULL },
         { "39", NULL },
         { "40", NULL },
         { NULL, NULL },
      },
      "0"
   },
   {
      BEETLE_OPT(last_scanline),
      BEETLE_OPT_LAST_SCANLINE_LABEL_CA,
      NULL,
      BEETLE_OPT_LAST_SCANLINE_INFO_0_CA,
      NULL,
      "video",
      {
         { "210", NULL },
         { "211", NULL },
         { "212", NULL },
         { "213", NULL },
         { "214", NULL },
         { "215", NULL },
         { "216", NULL },
         { "217", NULL },
         { "218", NULL },
         { "219", NULL },
         { "220", NULL },
         { "221", NULL },
         { "222", NULL },
         { "223", NULL },
         { "224", NULL },
         { "225", NULL },
         { "226", NULL },
         { "227", NULL },
         { "228", NULL },
         { "229", NULL },
         { "230", NULL },
         { "231", NULL },
         { "232", NULL },
         { "233", NULL },
         { "234", NULL },
         { "235", NULL },
         { "236", NULL },
         { "237", NULL },
         { "238", NULL },
         { "239", OPTION_VAL_239_CA },
         { NULL, NULL },
      },
      "239"
   },
   {
      BEETLE_OPT(initial_scanline_pal),
      BEETLE_OPT_INITIAL_SCANLINE_PAL_LABEL_CA,
      NULL,
      BEETLE_OPT_INITIAL_SCANLINE_PAL_INFO_0_CA,
      NULL,
      "video",
      {
         { "0",  OPTION_VAL_0_O48_CA },
         { "1",  NULL },
         { "2",  NULL },
         { "3",  NULL },
         { "4",  NULL },
         { "5",  NULL },
         { "6",  NULL },
         { "7",  NULL },
         { "8",  NULL },
         { "9",  NULL },
         { "10", NULL },
         { "11", NULL },
         { "12", NULL },
         { "13", NULL },
         { "14", NULL },
         { "15", NULL },
         { "16", NULL },
         { "17", NULL },
         { "18", NULL },
         { "19", NULL },
         { "20", NULL },
         { "21", NULL },
         { "22", NULL },
         { "23", NULL },
         { "24", NULL },
         { "25", NULL },
         { "26", NULL },
         { "27", NULL },
         { "28", NULL },
         { "29", NULL },
         { "30", NULL },
         { "31", NULL },
         { "32", NULL },
         { "33", NULL },
         { "34", NULL },
         { "35", NULL },
         { "36", NULL },
         { "37", NULL },
         { "38", NULL },
         { "39", NULL },
         { "40", NULL },
         { NULL, NULL },
      },
      "0"
   },
   {
      BEETLE_OPT(last_scanline_pal),
      BEETLE_OPT_LAST_SCANLINE_PAL_LABEL_CA,
      NULL,
      BEETLE_OPT_LAST_SCANLINE_PAL_INFO_0_CA,
      NULL,
      "video",
      {
         { "230", NULL },
         { "231", NULL },
         { "232", NULL },
         { "233", NULL },
         { "234", NULL },
         { "235", NULL },
         { "236", NULL },
         { "237", NULL },
         { "238", NULL },
         { "239", NULL },
         { "240", NULL },
         { "241", NULL },
         { "242", NULL },
         { "243", NULL },
         { "244", NULL },
         { "245", NULL },
         { "246", NULL },
         { "247", NULL },
         { "248", NULL },
         { "249", NULL },
         { "250", NULL },
         { "251", NULL },
         { "252", NULL },
         { "253", NULL },
         { "254", NULL },
         { "255", NULL },
         { "256", NULL },
         { "257", NULL },
         { "258", NULL },
         { "259", NULL },
         { "260", NULL },
         { "261", NULL },
         { "262", NULL },
         { "263", NULL },
         { "264", NULL },
         { "265", NULL },
         { "266", NULL },
         { "267", NULL },
         { "268", NULL },
         { "269", NULL },
         { "270", NULL },
         { "271", NULL },
         { "272", NULL },
         { "273", NULL },
         { "274", NULL },
         { "275", NULL },
         { "276", NULL },
         { "277", NULL },
         { "278", NULL },
         { "279", NULL },
         { "280", NULL },
         { "281", NULL },
         { "282", NULL },
         { "283", NULL },
         { "284", NULL },
         { "285", NULL },
         { "286", NULL },
         { "287", OPTION_VAL_287_CA },
         { NULL, NULL },
      },
      "287"
   },
   { NULL, NULL, NULL, NULL, NULL, NULL, {{0}}, NULL },
};
struct retro_core_options_v2 options_ca = {
   option_cats_ca,
   option_defs_ca
};

/* RETRO_LANGUAGE_CHS */

#define CATEGORY_SYSTEM_LABEL_CHS "系统"
#define CATEGORY_SYSTEM_INFO_0_CHS NULL
#define CATEGORY_VIDEO_LABEL_CHS "视频"
#define CATEGORY_VIDEO_INFO_0_CHS "更改长宽比、显示裁剪、视频过滤和帧跳过设置。"
#define CATEGORY_PGXP_LABEL_CHS "PGXP（精确几何变换管线）"
#define CATEGORY_PGXP_INFO_0_CHS "这些选项控制可提升图形效果的增强功能（相比原版主机）。PGXP 可以消除纹理扭曲和 Z-冲突问题。"
#define CATEGORY_OSD_LABEL_CHS "屏幕显示"
#define CATEGORY_OSD_INFO_0_CHS "更改屏幕上显示的通知。"
#define CATEGORY_INPUT_LABEL_CHS "输入"
#define CATEGORY_INPUT_INFO_0_CHS "更改光枪、鼠标和 neGcon 设置。"
#define CATEGORY_MEMCARDS_LABEL_CHS "记忆卡"
#define CATEGORY_MEMCARDS_INFO_0_CHS NULL
#define CATEGORY_HACKS_LABEL_CHS "模拟破解"
#define CATEGORY_HACKS_INFO_0_CHS NULL
#define BEETLE_OPT_CPU_FREQ_SCALE_LABEL_CHS "CPU 频率缩放（超频）"
#define BEETLE_OPT_CPU_FREQ_SCALE_INFO_0_CHS "对模拟的 PSX CPU 进行超频（或降频）。超频可以在某些游戏中消除卡顿并提高帧率，但会增加性能需求。请注意，某些游戏内置帧率限制器，可能无法从超频中受益。超频可能导致某些特效在某些游戏中动画速度比预期更快。"
#define OPTION_VAL_50_CHS NULL
#define OPTION_VAL_60_CHS NULL
#define OPTION_VAL_70_CHS NULL
#define OPTION_VAL_80_CHS NULL
#define OPTION_VAL_90_CHS NULL
#define OPTION_VAL_100_CHS NULL
#define OPTION_VAL_110_CHS NULL
#define OPTION_VAL_120_CHS NULL
#define OPTION_VAL_130_CHS NULL
#define OPTION_VAL_140_CHS NULL
#define OPTION_VAL_150_CHS NULL
#define OPTION_VAL_160_CHS NULL
#define OPTION_VAL_170_CHS NULL
#define OPTION_VAL_180_CHS NULL
#define OPTION_VAL_190_CHS NULL
#define OPTION_VAL_200_CHS NULL
#define OPTION_VAL_210_CHS NULL
#define OPTION_VAL_220_CHS NULL
#define OPTION_VAL_230_CHS NULL
#define OPTION_VAL_240_CHS NULL
#define OPTION_VAL_250_CHS NULL
#define OPTION_VAL_260_CHS NULL
#define OPTION_VAL_270_CHS NULL
#define OPTION_VAL_280_CHS NULL
#define OPTION_VAL_290_CHS NULL
#define OPTION_VAL_300_CHS NULL
#define OPTION_VAL_310_CHS NULL
#define OPTION_VAL_320_CHS NULL
#define OPTION_VAL_330_CHS NULL
#define OPTION_VAL_340_CHS NULL
#define OPTION_VAL_350_CHS NULL
#define OPTION_VAL_360_CHS NULL
#define OPTION_VAL_370_CHS NULL
#define OPTION_VAL_380_CHS NULL
#define OPTION_VAL_390_CHS NULL
#define OPTION_VAL_400_CHS NULL
#define OPTION_VAL_410_CHS NULL
#define OPTION_VAL_420_CHS NULL
#define OPTION_VAL_430_CHS NULL
#define OPTION_VAL_440_CHS NULL
#define OPTION_VAL_450_CHS NULL
#define OPTION_VAL_460_CHS NULL
#define OPTION_VAL_470_CHS NULL
#define OPTION_VAL_480_CHS NULL
#define OPTION_VAL_490_CHS NULL
#define OPTION_VAL_500_CHS NULL
#define OPTION_VAL_510_CHS NULL
#define OPTION_VAL_520_CHS NULL
#define OPTION_VAL_530_CHS NULL
#define OPTION_VAL_540_CHS NULL
#define OPTION_VAL_550_CHS NULL
#define OPTION_VAL_560_CHS NULL
#define OPTION_VAL_570_CHS NULL
#define OPTION_VAL_580_CHS NULL
#define OPTION_VAL_590_CHS NULL
#define OPTION_VAL_600_CHS NULL
#define OPTION_VAL_610_CHS NULL
#define OPTION_VAL_620_CHS NULL
#define OPTION_VAL_630_CHS NULL
#define OPTION_VAL_640_CHS NULL
#define OPTION_VAL_650_CHS NULL
#define OPTION_VAL_660_CHS NULL
#define OPTION_VAL_670_CHS NULL
#define OPTION_VAL_680_CHS NULL
#define OPTION_VAL_690_CHS NULL
#define OPTION_VAL_700_CHS NULL
#define OPTION_VAL_710_CHS NULL
#define OPTION_VAL_720_CHS NULL
#define OPTION_VAL_730_CHS NULL
#define OPTION_VAL_740_CHS NULL
#define OPTION_VAL_750_CHS NULL
#define BEETLE_OPT_GPU_OVERCLOCK_LABEL_CHS "GPU 光栅器超频"
#define BEETLE_OPT_GPU_OVERCLOCK_INFO_0_CHS "启用对模拟 PSX GPU 内置的 2D 光栅化器进行超频。此选项不会提升 3D 渲染效果，通常影响甚微。32x"
#define OPTION_VAL_1X_NATIVE_CHS "1x（原生）"
#define OPTION_VAL_2X_CHS NULL
#define OPTION_VAL_4X_CHS NULL
#define OPTION_VAL_8X_CHS NULL
#define OPTION_VAL_16X_CHS NULL
#define OPTION_VAL_32X_CHS NULL
#define BEETLE_OPT_GTE_OVERCLOCK_LABEL_CHS "GTE 超频"
#define BEETLE_OPT_GTE_OVERCLOCK_INFO_0_CHS "将所有模拟的 GTE（用于 3D 图形的 CPU 协处理器）操作降低至恒定的单周期延迟。对于重度使用 GTE 的游戏，这可以显著提升帧率和帧时间稳定性。"
#define BEETLE_OPT_CD_ACCESS_METHOD_LABEL_CHS NULL
#define BEETLE_OPT_CD_ACCESS_METHOD_INFO_0_CHS NULL
#define OPTION_VAL_SYNC_CHS "同步"
#define OPTION_VAL_ASYNC_CHS "异步"
#define OPTION_VAL_PRECACHE_CHS "预缓存"
#define BEETLE_OPT_CD_FASTLOAD_LABEL_CHS "CD 加载速度"
#define BEETLE_OPT_CD_FASTLOAD_INFO_0_CHS NULL
#define OPTION_VAL_2X_NATIVE_CHS "2x（原生）"
#define OPTION_VAL_6X_CHS NULL
#define OPTION_VAL_10X_CHS NULL
#define OPTION_VAL_12X_CHS NULL
#define OPTION_VAL_14X_CHS NULL
#define BEETLE_OPT_SKIP_BIOS_LABEL_CHS "跳过 BIOS"
#define BEETLE_OPT_SKIP_BIOS_INFO_0_CHS "跳过加载内容时通常显示的 PlayStation BIOS 启动动画。注意：启用此选项会导致与一些游戏（如 PAL 版复制保护游戏、Saga Frontier 等）出现兼容性问题。"
#define BEETLE_OPT_OVERRIDE_BIOS_LABEL_CHS NULL
#define BEETLE_OPT_OVERRIDE_BIOS_INFO_0_CHS NULL
#define OPTION_VAL_PSXONPSP_CHS NULL
#define OPTION_VAL_PS1_ROM_CHS NULL
#define OPTION_VAL_OPENBIOS_CHS NULL
#define BEETLE_OPT_REGION_LABEL_CHS "系统区域"
#define BEETLE_OPT_REGION_INFO_0_CHS NULL
#define OPTION_VAL_AUTO_CHS "自动"
#define OPTION_VAL_NTSC_J_CHS NULL
#define OPTION_VAL_NTSC_U_CHS NULL
#define OPTION_VAL_PAL_CHS NULL
#define BEETLE_OPT_INTERNAL_RESOLUTION_LABEL_CHS "内部 GPU 分辨率"
#define BEETLE_OPT_INTERNAL_RESOLUTION_INFO_0_CHS NULL
#define BEETLE_OPT_RENDERER_LABEL_CHS NULL
#define BEETLE_OPT_RENDERER_INFO_0_CHS NULL
#define OPTION_VAL_HARDWARE_CHS "硬件（自动）"
#define OPTION_VAL_HARDWARE_GL_CHS "硬件（OpenGL）"
#define OPTION_VAL_HARDWARE_VK_CHS "硬件（Vulkan）"
#define OPTION_VAL_SOFTWARE_CHS "软件"
#define BEETLE_OPT_RENDERER_SOFTWARE_FB_LABEL_CHS "软件帧缓冲区"
#define BEETLE_OPT_RENDERER_SOFTWARE_FB_INFO_0_CHS "当使用硬件渲染器时，通过在后台以原生分辨率运行一份软件渲染器副本，来实现对帧缓冲区效果（例如运动模糊、FF7 战斗漩涡）的精确模拟。如果禁用，这些操作会被省略（OpenGL）或在 GPU 上渲染（Vulkan）。禁用可以提升性能，但可能导致严重的图形错误。如果不确定，请保持启用状态。"
#define BEETLE_OPT_DEPTH_LABEL_CHS "内部颜色深度"
#define BEETLE_OPT_DEPTH_INFO_0_CHS NULL
#define OPTION_VAL_16BPP_NATIVE_CHS "16 bpp（原生）"
#define OPTION_VAL_32BPP_CHS NULL
#define BEETLE_OPT_DITHER_MODE_LABEL_CHS "抖动模式"
#define BEETLE_OPT_DITHER_MODE_INFO_0_CHS NULL
#define OPTION_VAL_INTERNAL_RESOLUTION_CHS "内部分辨率"
#define BEETLE_OPT_SCALED_UV_OFFSET_LABEL_CHS "纹理 UV 偏移"
#define BEETLE_OPT_SCALED_UV_OFFSET_INFO_0_CHS "对 3D 多边形的纹理进行偏移采样，用于高于 1x 的内部分辨率。可减少纹理接缝，但可能导致意外的图形故障。"
#define BEETLE_OPT_FILTER_LABEL_CHS "纹理过滤"
#define BEETLE_OPT_FILTER_INFO_0_CHS NULL
#define OPTION_VAL_NEAREST_CHS "最近邻"
#define OPTION_VAL_SABR_CHS NULL
#define OPTION_VAL_XBR_CHS NULL
#define OPTION_VAL_BILINEAR_CHS "双线性"
#define OPTION_VAL_3_POINT_CHS "三线性"
#define OPTION_VAL_JINC2_CHS NULL
#define BEETLE_OPT_FILTER_EXCLUDE_SPRITE_LABEL_CHS "排除精灵图过滤"
#define BEETLE_OPT_FILTER_EXCLUDE_SPRITE_INFO_0_CHS "不对精灵图应用纹理过滤。可防止各种带有 2D 精灵图背景的游戏出现接缝。与自适应平滑或其他后处理滤镜配合使用效果最佳。"
#define OPTION_VAL_OPAQUE_CHS "仅不透明"
#define OPTION_VAL_ALL_CHS "不透明和半透明"
#define BEETLE_OPT_FILTER_EXCLUDE_2D_POLYGON_LABEL_CHS "排除 2D 多边形过滤"
#define BEETLE_OPT_FILTER_EXCLUDE_2D_POLYGON_INFO_0_CHS "不对 2D 多边形应用纹理过滤。2D 多边形通过启发式算法检测，可能会出现故障。与自适应平滑或其他后处理滤镜配合使用效果最佳。"
#define BEETLE_OPT_ADAPTIVE_SMOOTHING_LABEL_CHS "自适应平滑"
#define BEETLE_OPT_ADAPTIVE_SMOOTHING_INFO_0_CHS "平滑 2D 图像和 UI 元素，同时不模糊 3D 渲染对象。仅由 Vulkan 渲染器支持。"
#define BEETLE_OPT_SUPER_SAMPLING_LABEL_CHS "超采样（降采样至原生分辨率）"
#define BEETLE_OPT_SUPER_SAMPLING_INFO_0_CHS "将从升高的内部分辨率渲染的内容降采样至原生分辨率。结合更高的内部分辨率倍数，允许游戏以抗锯齿的 3D 对象在原生低分辨率下显示。在混合 2D 和 3D 元素（例如预渲染背景上的 3D 角色）的游戏上效果最佳，并且与 CRT 着色器配合使用效果良好。仅由 Vulkan 渲染器支持。注意：启用此选项时，建议禁用“抖动模式”。"
#define BEETLE_OPT_MSAA_LABEL_CHS "多重采样抗锯齿"
#define BEETLE_OPT_MSAA_INFO_0_CHS NULL
#define OPTION_VAL_1X_CHS NULL
#define BEETLE_OPT_MDEC_YUV_LABEL_CHS "MDEC YUV 色度滤波器"
#define BEETLE_OPT_MDEC_YUV_INFO_0_CHS "通过减少“宏块”伪影（方块/锯齿边缘）来提升 FMV 播放质量。仅由 Vulkan 渲染器支持。"
#define BEETLE_OPT_TRACK_TEXTURES_LABEL_CHS "追踪纹理"
#define BEETLE_OPT_TRACK_TEXTURES_INFO_0_CHS NULL
#define BEETLE_OPT_DUMP_TEXTURES_LABEL_CHS "导出纹理"
#define BEETLE_OPT_DUMP_TEXTURES_INFO_0_CHS "将使用的纹理导出到 <cd>-texture-dump/ 目录。"
#define BEETLE_OPT_REPLACE_TEXTURES_LABEL_CHS "替换纹理"
#define BEETLE_OPT_REPLACE_TEXTURES_INFO_0_CHS "使用 <cd>-texture-replacements/ 目录中的高清版本替换纹理。"
#define BEETLE_OPT_PGXP_MODE_LABEL_CHS "PGXP 运行模式"
#define BEETLE_OPT_PGXP_MODE_INFO_0_CHS NULL
#define OPTION_VAL_MEMORY_ONLY_CHS "仅内存"
#define OPTION_VAL_MEMORY_CPU_CHS "内存 + CPU（有缺陷）"
#define BEETLE_OPT_PGXP_2D_TOL_LABEL_CHS "PGXP 2D 几何容差"
#define BEETLE_OPT_PGXP_2D_TOL_INFO_0_CHS "隐藏 PGXP 操作中更明显的错误：该值指定了在没有正确深度信息的情况下，PGXP 值将保留的容差。"
#define OPTION_VAL_0PX_CHS NULL
#define OPTION_VAL_1PX_CHS NULL
#define OPTION_VAL_2PX_CHS NULL
#define OPTION_VAL_3PX_CHS NULL
#define OPTION_VAL_4PX_CHS NULL
#define OPTION_VAL_5PX_CHS NULL
#define OPTION_VAL_6PX_CHS NULL
#define OPTION_VAL_7PX_CHS NULL
#define OPTION_VAL_8PX_CHS NULL
#define BEETLE_OPT_PGXP_NCLIP_LABEL_CHS "PGXP 图元剔除"
#define BEETLE_OPT_PGXP_NCLIP_INFO_0_CHS "使用 PGXP 的 NCLIP 实现。通过减少带有 PGXP 坐标的几何体中的孔洞来改善外观。已知在某些情况下会导致一些游戏卡死。"
#define BEETLE_OPT_PGXP_VERTEX_LABEL_CHS "PGXP 顶点缓存"
#define BEETLE_OPT_PGXP_VERTEX_INFO_0_CHS "缓存 PGXP 增强的顶点位置，以便在多边形绘制中重复使用。在渲染纹理时可能改善物体对齐并减少可见接缝，但在查询缓存时可能产生误报导致图形故障。目前建议保持此选项禁用。仅由硬件渲染器支持。"
#define BEETLE_OPT_PGXP_TEXTURE_LABEL_CHS "PGXP 透视校正纹理映射"
#define BEETLE_OPT_PGXP_TEXTURE_INFO_0_CHS NULL
#define BEETLE_OPT_DISPLAY_INTERNAL_FPS_LABEL_CHS "显示内部 FPS"
#define BEETLE_OPT_DISPLAY_INTERNAL_FPS_INFO_0_CHS "显示模拟的 PlayStation 系统渲染内容时的内部帧率。注意：需要在 libretro 前端中启用屏幕通知。"
#define BEETLE_OPT_DISPLAY_OSD_LABEL_CHS "显示 OSD 消息"
#define BEETLE_OPT_DISPLAY_OSD_INFO_0_CHS "显示由核心生成的 OSD 消息。"
#define BEETLE_OPT_DISPLAY_VRAM_LABEL_CHS "显示完整 VRAM（调试）"
#define BEETLE_OPT_DISPLAY_VRAM_INFO_0_CHS "可视化整个模拟主机的 VRAM。仅由 OpenGL 和 Vulkan 硬件渲染器支持。注意：此功能用于调试目的，通常应禁用。"
#define BEETLE_OPT_ANALOG_CALIBRATION_LABEL_CHS "模拟摇杆自动校准"
#define BEETLE_OPT_ANALOG_CALIBRATION_INFO_0_CHS NULL
#define BEETLE_OPT_ANALOG_TOGGLE_LABEL_CHS "DualShock 模拟模式切换"
#define BEETLE_OPT_ANALOG_TOGGLE_INFO_0_CHS "当输入设备类型为 DualShock 时，此选项允许模拟的 DualShock 在 DIGITAL 和 ANALOG 模式之间切换，如同原始硬件一样。模式也可以通过选择的按钮组合进行切换。"
#define OPTION_VAL_ENABLED_ANALOG_CHS "默认模拟"
#define BEETLE_OPT_ANALOG_TOGGLE_COMBO_LABEL_CHS "DualShock 模拟模式切换组合键"
#define BEETLE_OPT_ANALOG_TOGGLE_COMBO_INFO_0_CHS NULL
#define OPTION_VAL_L1_L2_R1_R2_START_SELECT_CHS "L1 + L2 + R1 + R2 + 开始键 + 选择键"
#define OPTION_VAL_L1_R1_SELECT_CHS "L1 + R1 + 选择键"
#define OPTION_VAL_L1_R1_START_CHS "L1 + R1 + 开始键"
#define OPTION_VAL_L1_R1_L3_CHS NULL
#define OPTION_VAL_L1_R1_R3_CHS NULL
#define OPTION_VAL_L2_R2_SELECT_CHS "L2 + R2 + 选择键"
#define OPTION_VAL_L2_R2_START_CHS "L2 + R2 + 开始键"
#define OPTION_VAL_L2_R2_L3_CHS NULL
#define OPTION_VAL_L2_R2_R3_CHS NULL
#define OPTION_VAL_L3_R3_CHS NULL
#define BEETLE_OPT_ANALOG_TOGGLE_HOLD_LABEL_CHS "DualShock 模拟模式组合按键保持时长"
#define BEETLE_OPT_ANALOG_TOGGLE_HOLD_INFO_0_CHS NULL
#define OPTION_VAL_0_CHS "0 秒延迟"
#define OPTION_VAL_1_CHS "1 秒延迟"
#define OPTION_VAL_2_CHS "2 秒延迟"
#define OPTION_VAL_3_CHS "3 秒延迟"
#define OPTION_VAL_4_CHS "4 秒延迟"
#define OPTION_VAL_5_CHS "5 秒延迟"
#define BEETLE_OPT_ENABLE_MULTITAP_PORT1_LABEL_CHS NULL
#define BEETLE_OPT_ENABLE_MULTITAP_PORT1_INFO_0_CHS "启用端口 1 的多击功能。"
#define BEETLE_OPT_ENABLE_MULTITAP_PORT2_LABEL_CHS NULL
#define BEETLE_OPT_ENABLE_MULTITAP_PORT2_INFO_0_CHS "启用端口 1 的多击功能。"
#define BEETLE_OPT_GUN_INPUT_MODE_LABEL_CHS "光枪输入模式"
#define BEETLE_OPT_GUN_INPUT_MODE_INFO_0_CHS NULL
#define OPTION_VAL_LIGHTGUN_CHS "光枪"
#define OPTION_VAL_TOUCHSCREEN_CHS "触摸屏"
#define BEETLE_OPT_GUN_CURSOR_LABEL_CHS "光枪准星"
#define BEETLE_OPT_GUN_CURSOR_INFO_0_CHS NULL
#define OPTION_VAL_CROSS_CHS "十字准星"
#define OPTION_VAL_DOT_CHS "点"
#define BEETLE_OPT_CROSSHAIR_COLOR_P1_LABEL_CHS NULL
#define BEETLE_OPT_CROSSHAIR_COLOR_P1_INFO_0_CHS NULL
#define OPTION_VAL_RED_CHS "红色"
#define OPTION_VAL_BLUE_CHS "蓝色"
#define OPTION_VAL_GREEN_CHS "绿色"
#define OPTION_VAL_ORANGE_CHS "橙色"
#define OPTION_VAL_YELLOW_CHS "黄色"
#define OPTION_VAL_CYAN_CHS "青色"
#define OPTION_VAL_PINK_CHS "粉"
#define OPTION_VAL_PURPLE_CHS "紫色"
#define OPTION_VAL_BLACK_CHS "黑色"
#define OPTION_VAL_WHITE_CHS "白色"
#define BEETLE_OPT_CROSSHAIR_COLOR_P2_LABEL_CHS NULL
#define BEETLE_OPT_CROSSHAIR_COLOR_P2_INFO_0_CHS NULL
#define BEETLE_OPT_MOUSE_SENSITIVITY_LABEL_CHS "鼠标灵敏度"
#define BEETLE_OPT_MOUSE_SENSITIVITY_INFO_0_CHS NULL
#define OPTION_VAL_5_O42_CHS NULL
#define OPTION_VAL_10_CHS NULL
#define OPTION_VAL_15_CHS NULL
#define OPTION_VAL_20_CHS NULL
#define OPTION_VAL_25_CHS NULL
#define OPTION_VAL_30_CHS NULL
#define OPTION_VAL_35_CHS NULL
#define OPTION_VAL_40_CHS NULL
#define OPTION_VAL_45_CHS NULL
#define OPTION_VAL_55_CHS NULL
#define OPTION_VAL_65_CHS NULL
#define OPTION_VAL_75_CHS NULL
#define OPTION_VAL_85_CHS NULL
#define OPTION_VAL_95_CHS NULL
#define OPTION_VAL_100_O42_CHS NULL
#define OPTION_VAL_105_CHS NULL
#define OPTION_VAL_115_CHS NULL
#define OPTION_VAL_125_CHS NULL
#define OPTION_VAL_135_CHS NULL
#define OPTION_VAL_145_CHS "145%"
#define OPTION_VAL_155_CHS NULL
#define OPTION_VAL_165_CHS NULL
#define OPTION_VAL_175_CHS NULL
#define OPTION_VAL_185_CHS NULL
#define OPTION_VAL_195_CHS NULL
#define BEETLE_OPT_NEGCON_RESPONSE_LABEL_CHS "neGcon 扭转响应"
#define BEETLE_OPT_NEGCON_RESPONSE_INFO_0_CHS NULL
#define OPTION_VAL_LINEAR_CHS "线性"
#define OPTION_VAL_QUADRATIC_CHS "二次方"
#define OPTION_VAL_CUBIC_CHS "三次方"
#define BEETLE_OPT_NEGCON_DEADZONE_LABEL_CHS "neGcon 扭转死区"
#define BEETLE_OPT_NEGCON_DEADZONE_INFO_0_CHS NULL
#define OPTION_VAL_0_O44_CHS NULL
#define BEETLE_OPT_USE_MEDNAFEN_MEMCARD0_METHOD_LABEL_CHS NULL
#define BEETLE_OPT_USE_MEDNAFEN_MEMCARD0_METHOD_INFO_0_CHS NULL
#define OPTION_VAL_LIBRETRO_CHS NULL
#define OPTION_VAL_MEDNAFEN_CHS NULL
#define BEETLE_OPT_ENABLE_MEMCARD1_LABEL_CHS NULL
#define BEETLE_OPT_ENABLE_MEMCARD1_INFO_0_CHS NULL
#define BEETLE_OPT_SHARED_MEMORY_CARDS_LABEL_CHS NULL
#define BEETLE_OPT_SHARED_MEMORY_CARDS_INFO_0_CHS NULL
#define BEETLE_OPT_MEMCARD_LEFT_INDEX_LABEL_CHS NULL
#define BEETLE_OPT_MEMCARD_LEFT_INDEX_INFO_0_CHS NULL
#define OPTION_VAL_0_O48_CHS NULL
#define BEETLE_OPT_MEMCARD_RIGHT_INDEX_LABEL_CHS NULL
#define BEETLE_OPT_MEMCARD_RIGHT_INDEX_INFO_0_CHS NULL
#define OPTION_VAL_1_O49_CHS NULL
#define BEETLE_OPT_LINE_RENDER_LABEL_CHS "线转四边形破解"
#define BEETLE_OPT_LINE_RENDER_INFO_0_CHS NULL
#define OPTION_VAL_DEFAULT_CHS "默认"
#define OPTION_VAL_AGGRESSIVE_CHS "激进"
#define BEETLE_OPT_WIDESCREEN_HACK_LABEL_CHS "宽屏模式增强"
#define BEETLE_OPT_WIDESCREEN_HACK_INFO_0_CHS "以变形方式渲染 3D 内容，并以宽屏宽高比输出模拟的帧缓冲区。在纯 3D 游戏上效果最佳。2D 元素将被水平拉伸，可能导致错位。"
#define BEETLE_OPT_WIDESCREEN_HACK_ASPECT_RATIO_LABEL_CHS "宽屏模式破解宽高比"
#define BEETLE_OPT_WIDESCREEN_HACK_ASPECT_RATIO_INFO_0_CHS NULL
#define OPTION_VAL_16_9_CHS NULL
#define OPTION_VAL_16_10_CHS NULL
#define OPTION_VAL_18_9_CHS NULL
#define OPTION_VAL_19_9_CHS NULL
#define OPTION_VAL_20_9_CHS NULL
#define OPTION_VAL_21_9_CHS NULL
#define OPTION_VAL_32_9_CHS NULL
#define BEETLE_OPT_CPU_DYNAREC_LABEL_CHS NULL
#define BEETLE_OPT_CPU_DYNAREC_INFO_0_CHS "将 CPU 指令动态重新编译为原生指令。比解释器快得多，但 CPU 计时不够精确，且可能存在 Bug。"
#define OPTION_VAL_DISABLED_CHS "已禁用（Beetle 解释器）"
#define OPTION_VAL_EXECUTE_CHS "最大性能"
#define OPTION_VAL_EXECUTE_ONE_CHS "循环计时检查"
#define OPTION_VAL_RUN_INTERPRETER_CHS "轻量级解释器"
#define BEETLE_OPT_DYNAREC_INVALIDATE_LABEL_CHS "Dynarec 代码失效模式"
#define BEETLE_OPT_DYNAREC_INVALIDATE_INFO_0_CHS "某些游戏需要“完整”失效，有些则需要“仅 DMA”。"
#define OPTION_VAL_FULL_CHS "全双工"
#define OPTION_VAL_DMA_CHS "仅 DMA（稍快）"
#define BEETLE_OPT_DYNAREC_EVENTCYCLES_LABEL_CHS "Dynarec DMA/GPU/MDEC/定时器事件周期"
#define BEETLE_OPT_DYNAREC_EVENTCYCLES_INFO_0_CHS "CPU 在检查 GPU/DMA/MDEC/定时器更新前运行的最大周期数。数值越高速度越快，但对 beetle 解释器的影响远小于 dynarec。"
#define OPTION_VAL_128_CHS "128（默认）"
#define BEETLE_OPT_DYNAREC_SPU_SAMPLES_LABEL_CHS "Dynarec SPU 采样数"
#define BEETLE_OPT_DYNAREC_SPU_SAMPLES_INFO_0_CHS "SPU 更新前运行的最大采样数。数值越高速度越快，但在除 1 以外的游戏上会导致声音故障。"
#define BEETLE_OPT_CORE_TIMING_FPS_LABEL_CHS "核心报告的 FPS 计时"
#define BEETLE_OPT_CORE_TIMING_FPS_INFO_0_CHS NULL
#define OPTION_VAL_AUTO_TOGGLE_CHS NULL
#define OPTION_VAL_FORCE_PROGRESSIVE_CHS NULL
#define OPTION_VAL_FORCE_INTERLACED_CHS "强制隔行速率"
#define BEETLE_OPT_DEINTERLACER_LABEL_CHS "去隔行方法"
#define BEETLE_OPT_DEINTERLACER_INFO_0_CHS NULL
#define OPTION_VAL_WEAVE_CHS "交织"
#define OPTION_VAL_BOB_CHS NULL
#define OPTION_VAL_BOB_OFFSET_CHS NULL
#define OPTION_VAL_FASTMAD_CHS NULL
#define BEETLE_OPT_FRAME_DUPING_LABEL_CHS NULL
#define BEETLE_OPT_FRAME_DUPING_INFO_0_CHS NULL
#define BEETLE_OPT_PAL_VIDEO_TIMING_OVERRIDE_LABEL_CHS NULL
#define BEETLE_OPT_PAL_VIDEO_TIMING_OVERRIDE_INFO_0_CHS NULL
#define BEETLE_OPT_ASPECT_RATIO_LABEL_CHS "核心宽高比"
#define BEETLE_OPT_ASPECT_RATIO_INFO_0_CHS NULL
#define OPTION_VAL_CORRECTED_CHS "已校正"
#define OPTION_VAL_UNCORRECTED_CHS "未校正"
#define OPTION_VAL_4_3_CHS "强制 4:3"
#define OPTION_VAL_NTSC_CHS "强制 NTSC"
#define BEETLE_OPT_CROP_OVERSCAN_LABEL_CHS "裁剪过扫描"
#define BEETLE_OPT_CROP_OVERSCAN_INFO_0_CHS NULL
#define OPTION_VAL_STATIC_CHS "水平"
#define OPTION_VAL_SMART_CHS NULL
#define BEETLE_OPT_IMAGE_CROP_LABEL_CHS "额外裁剪"
#define BEETLE_OPT_IMAGE_CROP_INFO_0_CHS NULL
#define OPTION_VAL_9PX_CHS NULL
#define OPTION_VAL_10PX_CHS NULL
#define OPTION_VAL_11PX_CHS NULL
#define OPTION_VAL_12PX_CHS NULL
#define OPTION_VAL_13PX_CHS NULL
#define OPTION_VAL_14PX_CHS NULL
#define OPTION_VAL_15PX_CHS NULL
#define OPTION_VAL_16PX_CHS NULL
#define OPTION_VAL_17PX_CHS NULL
#define OPTION_VAL_18PX_CHS NULL
#define OPTION_VAL_19PX_CHS NULL
#define OPTION_VAL_20PX_CHS NULL
#define BEETLE_OPT_IMAGE_OFFSET_LABEL_CHS "偏移裁剪图像"
#define BEETLE_OPT_IMAGE_OFFSET_INFO_0_CHS NULL
#define OPTION_VAL_12PX_O64_CHS NULL
#define OPTION_VAL_11PX_O64_CHS NULL
#define OPTION_VAL_10PX_O64_CHS NULL
#define OPTION_VAL_9PX_O64_CHS NULL
#define OPTION_VAL_8PX_O64_CHS NULL
#define OPTION_VAL_7PX_O64_CHS NULL
#define OPTION_VAL_6PX_O64_CHS NULL
#define OPTION_VAL_5PX_O64_CHS NULL
#define OPTION_VAL_4PX_O64_CHS NULL
#define OPTION_VAL_3PX_O64_CHS NULL
#define OPTION_VAL_2PX_O64_CHS NULL
#define OPTION_VAL_1PX_O64_CHS NULL
#define OPTION_VAL_1PX_O64_0_CHS NULL
#define OPTION_VAL_2PX_O64_0_CHS NULL
#define OPTION_VAL_3PX_O64_0_CHS NULL
#define OPTION_VAL_4PX_O64_0_CHS NULL
#define OPTION_VAL_5PX_O64_0_CHS NULL
#define OPTION_VAL_6PX_O64_0_CHS NULL
#define OPTION_VAL_7PX_O64_0_CHS NULL
#define OPTION_VAL_8PX_O64_0_CHS NULL
#define OPTION_VAL_9PX_O64_0_CHS NULL
#define OPTION_VAL_10PX_O64_0_CHS NULL
#define OPTION_VAL_11PX_O64_0_CHS NULL
#define OPTION_VAL_12PX_O64_0_CHS NULL
#define BEETLE_OPT_IMAGE_OFFSET_CYCLES_LABEL_CHS "水平图像偏移（GPU 周期）"
#define BEETLE_OPT_IMAGE_OFFSET_CYCLES_INFO_0_CHS NULL
#define BEETLE_OPT_INITIAL_SCANLINE_LABEL_CHS "NTSC 初始扫描线"
#define BEETLE_OPT_INITIAL_SCANLINE_INFO_0_CHS NULL
#define BEETLE_OPT_LAST_SCANLINE_LABEL_CHS "最后扫描线 - NTSC"
#define BEETLE_OPT_LAST_SCANLINE_INFO_0_CHS NULL
#define OPTION_VAL_239_CHS "239（默认）"
#define BEETLE_OPT_INITIAL_SCANLINE_PAL_LABEL_CHS "初始扫描线 - PAL"
#define BEETLE_OPT_INITIAL_SCANLINE_PAL_INFO_0_CHS NULL
#define BEETLE_OPT_LAST_SCANLINE_PAL_LABEL_CHS "最后扫描线 - PAL"
#define BEETLE_OPT_LAST_SCANLINE_PAL_INFO_0_CHS NULL
#define OPTION_VAL_287_CHS "287（默认）"

struct retro_core_option_v2_category option_cats_chs[] = {
   {
      "system",
      CATEGORY_SYSTEM_LABEL_CHS,
      CATEGORY_SYSTEM_INFO_0_CHS
   },
   {
      "video",
      CATEGORY_VIDEO_LABEL_CHS,
      CATEGORY_VIDEO_INFO_0_CHS
   },
   {
      "pgxp",
      CATEGORY_PGXP_LABEL_CHS,
      CATEGORY_PGXP_INFO_0_CHS
   },
   {
      "osd",
      CATEGORY_OSD_LABEL_CHS,
      CATEGORY_OSD_INFO_0_CHS
   },
   {
      "input",
      CATEGORY_INPUT_LABEL_CHS,
      CATEGORY_INPUT_INFO_0_CHS
   },
   {
      "memcards",
      CATEGORY_MEMCARDS_LABEL_CHS,
      CATEGORY_MEMCARDS_INFO_0_CHS
   },
   {
      "hacks",
      CATEGORY_HACKS_LABEL_CHS,
      CATEGORY_HACKS_INFO_0_CHS
   },
   { NULL, NULL, NULL },
};
struct retro_core_option_v2_definition option_defs_chs[] = {
   {
      BEETLE_OPT(cpu_freq_scale),
      BEETLE_OPT_CPU_FREQ_SCALE_LABEL_CHS,
      NULL,
      BEETLE_OPT_CPU_FREQ_SCALE_INFO_0_CHS,
      NULL,
      "system",
      {
         { "50%",           OPTION_VAL_50_CHS },
         { "60%",           OPTION_VAL_60_CHS },
         { "70%",           OPTION_VAL_70_CHS },
         { "80%",           OPTION_VAL_80_CHS },
         { "90%",           OPTION_VAL_90_CHS },
         { "100%",          OPTION_VAL_100_CHS },
         { "110%",          OPTION_VAL_110_CHS },
         { "120%",          OPTION_VAL_120_CHS },
         { "130%",          OPTION_VAL_130_CHS },
         { "140%",          OPTION_VAL_140_CHS },
         { "150%",          OPTION_VAL_150_CHS },
         { "160%",          OPTION_VAL_160_CHS },
         { "170%",          OPTION_VAL_170_CHS },
         { "180%",          OPTION_VAL_180_CHS },
         { "190%",          OPTION_VAL_190_CHS },
         { "200%",          OPTION_VAL_200_CHS },
         { "210%",          OPTION_VAL_210_CHS },
         { "220%",          OPTION_VAL_220_CHS },
         { "230%",          OPTION_VAL_230_CHS },
         { "240%",          OPTION_VAL_240_CHS },
         { "250%",          OPTION_VAL_250_CHS },
         { "260%",          OPTION_VAL_260_CHS },
         { "270%",          OPTION_VAL_270_CHS },
         { "280%",          OPTION_VAL_280_CHS },
         { "290%",          OPTION_VAL_290_CHS },
         { "300%",          OPTION_VAL_300_CHS },
         { "310%",          OPTION_VAL_310_CHS },
         { "320%",          OPTION_VAL_320_CHS },
         { "330%",          OPTION_VAL_330_CHS },
         { "340%",          OPTION_VAL_340_CHS },
         { "350%",          OPTION_VAL_350_CHS },
         { "360%",          OPTION_VAL_360_CHS },
         { "370%",          OPTION_VAL_370_CHS },
         { "380%",          OPTION_VAL_380_CHS },
         { "390%",          OPTION_VAL_390_CHS },
         { "400%",          OPTION_VAL_400_CHS },
         { "410%",          OPTION_VAL_410_CHS },
         { "420%",          OPTION_VAL_420_CHS },
         { "430%",          OPTION_VAL_430_CHS },
         { "440%",          OPTION_VAL_440_CHS },
         { "450%",          OPTION_VAL_450_CHS },
         { "460%",          OPTION_VAL_460_CHS },
         { "470%",          OPTION_VAL_470_CHS },
         { "480%",          OPTION_VAL_480_CHS },
         { "490%",          OPTION_VAL_490_CHS },
         { "500%",          OPTION_VAL_500_CHS },
         { "510%",          OPTION_VAL_510_CHS },
         { "520%",          OPTION_VAL_520_CHS },
         { "530%",          OPTION_VAL_530_CHS },
         { "540%",          OPTION_VAL_540_CHS },
         { "550%",          OPTION_VAL_550_CHS },
         { "560%",          OPTION_VAL_560_CHS },
         { "570%",          OPTION_VAL_570_CHS },
         { "580%",          OPTION_VAL_580_CHS },
         { "590%",          OPTION_VAL_590_CHS },
         { "600%",          OPTION_VAL_600_CHS },
         { "610%",          OPTION_VAL_610_CHS },
         { "620%",          OPTION_VAL_620_CHS },
         { "630%",          OPTION_VAL_630_CHS },
         { "640%",          OPTION_VAL_640_CHS },
         { "650%",          OPTION_VAL_650_CHS },
         { "660%",          OPTION_VAL_660_CHS },
         { "670%",          OPTION_VAL_670_CHS },
         { "680%",          OPTION_VAL_680_CHS },
         { "690%",          OPTION_VAL_690_CHS },
         { "700%",          OPTION_VAL_700_CHS },
         { "710%",          OPTION_VAL_710_CHS },
         { "720%",          OPTION_VAL_720_CHS },
         { "730%",          OPTION_VAL_730_CHS },
         { "740%",          OPTION_VAL_740_CHS },
         { "750%",          OPTION_VAL_750_CHS },
         { NULL, NULL },
      },
      "100%"
   },
   {
      BEETLE_OPT(gpu_overclock),
      BEETLE_OPT_GPU_OVERCLOCK_LABEL_CHS,
      NULL,
      BEETLE_OPT_GPU_OVERCLOCK_INFO_0_CHS,
      NULL,
      "system",
      {
         { "1x(native)", OPTION_VAL_1X_NATIVE_CHS },
         { "2x",         OPTION_VAL_2X_CHS },
         { "4x",         OPTION_VAL_4X_CHS },
         { "8x",         OPTION_VAL_8X_CHS },
         { "16x",        OPTION_VAL_16X_CHS },
         { "32x",        OPTION_VAL_32X_CHS },
         { NULL, NULL },
      },
      "1x(native)"
   },
   {
      BEETLE_OPT(gte_overclock),
      BEETLE_OPT_GTE_OVERCLOCK_LABEL_CHS,
      NULL,
      BEETLE_OPT_GTE_OVERCLOCK_INFO_0_CHS,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
#ifndef EMSCRIPTEN
   {
      BEETLE_OPT(cd_access_method),
      BEETLE_OPT_CD_ACCESS_METHOD_LABEL_CHS,
      NULL,
      BEETLE_OPT_CD_ACCESS_METHOD_INFO_0_CHS,
      NULL,
      "system",
      {
         { "sync",     OPTION_VAL_SYNC_CHS },
         { "async",    OPTION_VAL_ASYNC_CHS },
         { "precache", OPTION_VAL_PRECACHE_CHS },
         { NULL, NULL },
      },
      "sync"
   },
#endif
   {
      BEETLE_OPT(cd_fastload),
      BEETLE_OPT_CD_FASTLOAD_LABEL_CHS,
      NULL,
      BEETLE_OPT_CD_FASTLOAD_INFO_0_CHS,
      NULL,
      "system",
      {
         { "2x(native)", OPTION_VAL_2X_NATIVE_CHS },
         { "4x",          OPTION_VAL_4X_CHS },
         { "6x",          OPTION_VAL_6X_CHS },
         { "8x",          OPTION_VAL_8X_CHS },
         { "10x",         OPTION_VAL_10X_CHS },
         { "12x",         OPTION_VAL_12X_CHS },
         { "14x",         OPTION_VAL_14X_CHS },
         { NULL, NULL },
      },
      "2x(native)"
   },
   {
      BEETLE_OPT(skip_bios),
      BEETLE_OPT_SKIP_BIOS_LABEL_CHS,
      NULL,
      BEETLE_OPT_SKIP_BIOS_INFO_0_CHS,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(override_bios),
      BEETLE_OPT_OVERRIDE_BIOS_LABEL_CHS,
      NULL,
      BEETLE_OPT_OVERRIDE_BIOS_INFO_0_CHS,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "psxonpsp", OPTION_VAL_PSXONPSP_CHS },
         { "ps1_rom",  OPTION_VAL_PS1_ROM_CHS },
         { "openbios", OPTION_VAL_OPENBIOS_CHS     },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(region),
      BEETLE_OPT_REGION_LABEL_CHS,
      NULL,
      BEETLE_OPT_REGION_INFO_0_CHS,
      NULL,
      "system",
      {
         { "auto",   OPTION_VAL_AUTO_CHS },
         { "ntsc-j", OPTION_VAL_NTSC_J_CHS },
         { "ntsc-u", OPTION_VAL_NTSC_U_CHS },
         { "pal",    OPTION_VAL_PAL_CHS },
         { NULL, NULL },
      },
      "auto"
   },
   {
      BEETLE_OPT(internal_resolution),
      BEETLE_OPT_INTERNAL_RESOLUTION_LABEL_CHS,
      NULL,
      BEETLE_OPT_INTERNAL_RESOLUTION_INFO_0_CHS,
      NULL,
      "video",
      {
         { "1x(native)", OPTION_VAL_1X_NATIVE_CHS },
         { "2x",         OPTION_VAL_2X_CHS },
         { "4x",         OPTION_VAL_4X_CHS },
         { "8x",         OPTION_VAL_8X_CHS },
         { "16x",        OPTION_VAL_16X_CHS },
         { NULL, NULL },
      },
      "1x(native)"
   },
#if defined(HAVE_OPENGL) || defined(HAVE_OPENGLES) || defined(HAVE_VULKAN)
   {
      BEETLE_OPT(renderer),
      BEETLE_OPT_RENDERER_LABEL_CHS,
      NULL,
      BEETLE_OPT_RENDERER_INFO_0_CHS,
      NULL,
      "video",
      {
         { "hardware",    OPTION_VAL_HARDWARE_CHS },
#if defined(HAVE_OPENGL) || defined(HAVE_OPENGLES)
         { "hardware_gl", OPTION_VAL_HARDWARE_GL_CHS },
#endif
#if defined(HAVE_VULKAN)
         { "hardware_vk", OPTION_VAL_HARDWARE_VK_CHS },
#endif
         { "software",    OPTION_VAL_SOFTWARE_CHS },
         { NULL, NULL },
      },
      "hardware"
   },
   {
      BEETLE_OPT(renderer_software_fb),
      BEETLE_OPT_RENDERER_SOFTWARE_FB_LABEL_CHS,
      NULL,
      BEETLE_OPT_RENDERER_SOFTWARE_FB_INFO_0_CHS,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled"
   },
#endif
#if defined(HAVE_OPENGL) || defined(HAVE_OPENGLES)
   {
      BEETLE_OPT(depth),
      BEETLE_OPT_DEPTH_LABEL_CHS,
      NULL,
      BEETLE_OPT_DEPTH_INFO_0_CHS,
      NULL,
      "video",
      {
         { "16bpp(native)", OPTION_VAL_16BPP_NATIVE_CHS },
         { "32bpp",         OPTION_VAL_32BPP_CHS },
         { NULL, NULL },
      },
      "16bpp(native)"
   },
   // Sort of, it's more like 15-bit high color and 24-bit true color for visible output. The alpha channel is used for mask bit. Vulkan renderer uses ABGR1555_555 for 31 bits internal? FMVs are always 24-bit on all renderers like original hardware (BGR888, no alpha)
#endif
   {
      BEETLE_OPT(dither_mode),
      BEETLE_OPT_DITHER_MODE_LABEL_CHS,
      NULL,
      BEETLE_OPT_DITHER_MODE_INFO_0_CHS,
      NULL,
      "video",
      {
         { "1x(native)",          OPTION_VAL_1X_NATIVE_CHS },
         { "internal resolution", OPTION_VAL_INTERNAL_RESOLUTION_CHS },
         { "disabled",            NULL },
         { NULL, NULL },
      },
      "1x(native)"
   },
#ifdef HAVE_VULKAN
   {
      BEETLE_OPT(scaled_uv_offset),
      BEETLE_OPT_SCALED_UV_OFFSET_LABEL_CHS,
      NULL,
      BEETLE_OPT_SCALED_UV_OFFSET_INFO_0_CHS,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled"
   },
#endif
#if defined(HAVE_OPENGL) || defined(HAVE_OPENGLES) || defined(HAVE_VULKAN)
   {
      BEETLE_OPT(filter),
      BEETLE_OPT_FILTER_LABEL_CHS,
      NULL,
      BEETLE_OPT_FILTER_INFO_0_CHS,
      NULL,
      "video",
      {
         { "nearest",  OPTION_VAL_NEAREST_CHS },
         { "SABR",     OPTION_VAL_SABR_CHS },
         { "xBR",      OPTION_VAL_XBR_CHS },
         { "bilinear", OPTION_VAL_BILINEAR_CHS },
         { "3-point",  OPTION_VAL_3_POINT_CHS },
         { "JINC2",    OPTION_VAL_JINC2_CHS },
         { NULL, NULL },
      },
      "nearest"
   },
#ifdef HAVE_VULKAN
   {
      BEETLE_OPT(filter_exclude_sprite),
      BEETLE_OPT_FILTER_EXCLUDE_SPRITE_LABEL_CHS,
      NULL,
      BEETLE_OPT_FILTER_EXCLUDE_SPRITE_INFO_0_CHS,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "opaque", OPTION_VAL_OPAQUE_CHS },
         { "all", OPTION_VAL_ALL_CHS },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(filter_exclude_2d_polygon),
      BEETLE_OPT_FILTER_EXCLUDE_2D_POLYGON_LABEL_CHS,
      NULL,
      BEETLE_OPT_FILTER_EXCLUDE_2D_POLYGON_INFO_0_CHS,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "opaque", OPTION_VAL_OPAQUE_CHS },
         { "all", OPTION_VAL_ALL_CHS },
         { NULL, NULL },
      },
      "disabled"
   },
#endif
#endif
#ifdef HAVE_VULKAN
   {
      BEETLE_OPT(adaptive_smoothing),
      BEETLE_OPT_ADAPTIVE_SMOOTHING_LABEL_CHS,
      NULL,
      BEETLE_OPT_ADAPTIVE_SMOOTHING_INFO_0_CHS,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(super_sampling),
      BEETLE_OPT_SUPER_SAMPLING_LABEL_CHS,
      NULL,
      BEETLE_OPT_SUPER_SAMPLING_INFO_0_CHS,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(msaa),
      BEETLE_OPT_MSAA_LABEL_CHS,
      NULL,
      BEETLE_OPT_MSAA_INFO_0_CHS,
      NULL,
      "video",
      {
         { "1x",  OPTION_VAL_1X_CHS },
         { "2x",  OPTION_VAL_2X_CHS },
         { "4x",  OPTION_VAL_4X_CHS },
         { "8x",  OPTION_VAL_8X_CHS },
         { "16x", OPTION_VAL_16X_CHS },
         { NULL, NULL },
      },
      "1x"
   },
   {
      BEETLE_OPT(mdec_yuv),
      BEETLE_OPT_MDEC_YUV_LABEL_CHS,
      NULL,
      BEETLE_OPT_MDEC_YUV_INFO_0_CHS,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(track_textures),
      BEETLE_OPT_TRACK_TEXTURES_LABEL_CHS,
      NULL,
      BEETLE_OPT_TRACK_TEXTURES_INFO_0_CHS,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
#ifdef TEXTURE_DUMPING_ENABLED
   {
      BEETLE_OPT(dump_textures),
      BEETLE_OPT_DUMP_TEXTURES_LABEL_CHS,
      NULL,
      BEETLE_OPT_DUMP_TEXTURES_INFO_0_CHS,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
#endif
   {
      BEETLE_OPT(replace_textures),
      BEETLE_OPT_REPLACE_TEXTURES_LABEL_CHS,
      NULL,
      BEETLE_OPT_REPLACE_TEXTURES_INFO_0_CHS,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
#endif
   {
      BEETLE_OPT(pgxp_mode),
      BEETLE_OPT_PGXP_MODE_LABEL_CHS,
      NULL,
      BEETLE_OPT_PGXP_MODE_INFO_0_CHS,
      NULL,
      "pgxp",
      {
         { "disabled",     NULL },
         { "memory only",  OPTION_VAL_MEMORY_ONLY_CHS },
         { "memory + CPU", OPTION_VAL_MEMORY_CPU_CHS },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(pgxp_2d_tol),
      BEETLE_OPT_PGXP_2D_TOL_LABEL_CHS,
      NULL,
      BEETLE_OPT_PGXP_2D_TOL_INFO_0_CHS,
      NULL,
      "pgxp",
      {
         { "disabled", NULL },
         { "0px", OPTION_VAL_0PX_CHS },
         { "1px", OPTION_VAL_1PX_CHS },
         { "2px", OPTION_VAL_2PX_CHS },
         { "3px", OPTION_VAL_3PX_CHS },
         { "4px", OPTION_VAL_4PX_CHS },
         { "5px", OPTION_VAL_5PX_CHS },
         { "6px", OPTION_VAL_6PX_CHS },
         { "7px", OPTION_VAL_7PX_CHS },
         { "8px", OPTION_VAL_8PX_CHS },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(pgxp_nclip),
      BEETLE_OPT_PGXP_NCLIP_LABEL_CHS,
      NULL,
      BEETLE_OPT_PGXP_NCLIP_INFO_0_CHS,
      NULL,
      "pgxp",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
#if defined(HAVE_OPENGL) || defined(HAVE_OPENGLES) || defined(HAVE_VULKAN)
   {
      BEETLE_OPT(pgxp_vertex),
      BEETLE_OPT_PGXP_VERTEX_LABEL_CHS,
      NULL,
      BEETLE_OPT_PGXP_VERTEX_INFO_0_CHS,
      NULL,
      "pgxp",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
#endif
   {
      BEETLE_OPT(pgxp_texture),
      BEETLE_OPT_PGXP_TEXTURE_LABEL_CHS,
      NULL,
      BEETLE_OPT_PGXP_TEXTURE_INFO_0_CHS,
      NULL,
      "pgxp",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(display_internal_fps),
      BEETLE_OPT_DISPLAY_INTERNAL_FPS_LABEL_CHS,
      NULL,
      BEETLE_OPT_DISPLAY_INTERNAL_FPS_INFO_0_CHS,
      NULL,
      "osd",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(display_osd),
      BEETLE_OPT_DISPLAY_OSD_LABEL_CHS,
      NULL,
      BEETLE_OPT_DISPLAY_OSD_INFO_0_CHS,
      NULL,
      "osd",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled"
   },
#if defined(HAVE_OPENGL) || defined(HAVE_OPENGLES) || defined(HAVE_VULKAN)
   {
      BEETLE_OPT(display_vram),
      BEETLE_OPT_DISPLAY_VRAM_LABEL_CHS,
      NULL,
      BEETLE_OPT_DISPLAY_VRAM_INFO_0_CHS,
      NULL,
      "osd",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
#endif
   {
      BEETLE_OPT(analog_calibration),
      BEETLE_OPT_ANALOG_CALIBRATION_LABEL_CHS,
      NULL,
      BEETLE_OPT_ANALOG_CALIBRATION_INFO_0_CHS,
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(analog_toggle),
      BEETLE_OPT_ANALOG_TOGGLE_LABEL_CHS,
      NULL,
      BEETLE_OPT_ANALOG_TOGGLE_INFO_0_CHS,
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { "enabled-analog", OPTION_VAL_ENABLED_ANALOG_CHS },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(analog_toggle_combo),
      BEETLE_OPT_ANALOG_TOGGLE_COMBO_LABEL_CHS,
      NULL,
      BEETLE_OPT_ANALOG_TOGGLE_COMBO_INFO_0_CHS,
      NULL,
      "input",
      {
         { "l1+l2+r1+r2+start+select", OPTION_VAL_L1_L2_R1_R2_START_SELECT_CHS },
         { "l1+r1+select",             OPTION_VAL_L1_R1_SELECT_CHS },
         { "l1+r1+start",              OPTION_VAL_L1_R1_START_CHS },
         { "l1+r1+l3",                 OPTION_VAL_L1_R1_L3_CHS },
         { "l1+r1+r3",                 OPTION_VAL_L1_R1_R3_CHS },
         { "l2+r2+select",             OPTION_VAL_L2_R2_SELECT_CHS },
         { "l2+r2+start",              OPTION_VAL_L2_R2_START_CHS },
         { "l2+r2+l3",                 OPTION_VAL_L2_R2_L3_CHS },
         { "l2+r2+r3",                 OPTION_VAL_L2_R2_R3_CHS },
         { "l3+r3",                    OPTION_VAL_L3_R3_CHS },
         { NULL, NULL },
      },
      "l1+r1+select"
   },
   {
      BEETLE_OPT(analog_toggle_hold),
      BEETLE_OPT_ANALOG_TOGGLE_HOLD_LABEL_CHS,
      NULL,
      BEETLE_OPT_ANALOG_TOGGLE_HOLD_INFO_0_CHS,
      NULL,
      "input",
      {
         { "0", OPTION_VAL_0_CHS },
         { "1", OPTION_VAL_1_CHS },
         { "2", OPTION_VAL_2_CHS },
         { "3", OPTION_VAL_3_CHS },
         { "4", OPTION_VAL_4_CHS },
         { "5", OPTION_VAL_5_CHS },
         { NULL, NULL },
      },
      "1"
   },
   {
      BEETLE_OPT(enable_multitap_port1),
      BEETLE_OPT_ENABLE_MULTITAP_PORT1_LABEL_CHS,
      NULL,
      BEETLE_OPT_ENABLE_MULTITAP_PORT1_INFO_0_CHS,
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(enable_multitap_port2),
      BEETLE_OPT_ENABLE_MULTITAP_PORT2_LABEL_CHS,
      NULL,
      BEETLE_OPT_ENABLE_MULTITAP_PORT2_INFO_0_CHS,
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(gun_input_mode),
      BEETLE_OPT_GUN_INPUT_MODE_LABEL_CHS,
      NULL,
      BEETLE_OPT_GUN_INPUT_MODE_INFO_0_CHS,
      NULL,
      "input",
      {
         { "lightgun",    OPTION_VAL_LIGHTGUN_CHS },
         { "touchscreen", OPTION_VAL_TOUCHSCREEN_CHS },
         { NULL, NULL },
      },
      "lightgun"
   },
   // Shouldn't the gun_input_mode just be Mouse vs. Touchscreen?
   {
      BEETLE_OPT(gun_cursor),
      BEETLE_OPT_GUN_CURSOR_LABEL_CHS,
      NULL,
      BEETLE_OPT_GUN_CURSOR_INFO_0_CHS,
      NULL,
      "input",
      {
         { "off",   "disabled" },
         { "cross", OPTION_VAL_CROSS_CHS },
         { "dot",   OPTION_VAL_DOT_CHS },
         { NULL, NULL },
      },
      "cross"
   },
   {
      BEETLE_OPT(crosshair_color_p1),
      BEETLE_OPT_CROSSHAIR_COLOR_P1_LABEL_CHS,
      NULL,
      BEETLE_OPT_CROSSHAIR_COLOR_P1_INFO_0_CHS,
      NULL,
      "input",
      {
         { "red",    OPTION_VAL_RED_CHS },
         { "blue",   OPTION_VAL_BLUE_CHS },
         { "green",  OPTION_VAL_GREEN_CHS },
         { "orange", OPTION_VAL_ORANGE_CHS },
         { "yellow", OPTION_VAL_YELLOW_CHS },
         { "cyan",   OPTION_VAL_CYAN_CHS },
         { "pink",   OPTION_VAL_PINK_CHS },
         { "purple", OPTION_VAL_PURPLE_CHS },
         { "black",  OPTION_VAL_BLACK_CHS },
         { "white",  OPTION_VAL_WHITE_CHS },
         { NULL, NULL },
      },
      "red"
   },
   {
      BEETLE_OPT(crosshair_color_p2),
      BEETLE_OPT_CROSSHAIR_COLOR_P2_LABEL_CHS,
      NULL,
      BEETLE_OPT_CROSSHAIR_COLOR_P2_INFO_0_CHS,
      NULL,
      "input",
      {
         { "blue",   OPTION_VAL_BLUE_CHS },
         { "red",    OPTION_VAL_RED_CHS },
         { "green",  OPTION_VAL_GREEN_CHS },
         { "orange", OPTION_VAL_ORANGE_CHS },
         { "yellow", OPTION_VAL_YELLOW_CHS },
         { "cyan",   OPTION_VAL_CYAN_CHS },
         { "pink",   OPTION_VAL_PINK_CHS },
         { "purple", OPTION_VAL_PURPLE_CHS },
         { "black",  OPTION_VAL_BLACK_CHS },
         { "white",  OPTION_VAL_WHITE_CHS },
         { NULL, NULL },
      },
      "blue"
   },
   {
      BEETLE_OPT(mouse_sensitivity),
      BEETLE_OPT_MOUSE_SENSITIVITY_LABEL_CHS,
      NULL,
      BEETLE_OPT_MOUSE_SENSITIVITY_INFO_0_CHS,
      NULL,
      "input",
      {
         { "5%",   OPTION_VAL_5_O42_CHS },
         { "10%",  OPTION_VAL_10_CHS },
         { "15%",  OPTION_VAL_15_CHS },
         { "20%",  OPTION_VAL_20_CHS },
         { "25%",  OPTION_VAL_25_CHS },
         { "30%",  OPTION_VAL_30_CHS },
         { "35%",  OPTION_VAL_35_CHS },
         { "40%",  OPTION_VAL_40_CHS },
         { "45%",  OPTION_VAL_45_CHS },
         { "50%",  OPTION_VAL_50_CHS },
         { "55%",  OPTION_VAL_55_CHS },
         { "60%",  OPTION_VAL_60_CHS },
         { "65%",  OPTION_VAL_65_CHS },
         { "70%",  OPTION_VAL_70_CHS },
         { "75%",  OPTION_VAL_75_CHS },
         { "80%",  OPTION_VAL_80_CHS },
         { "85%",  OPTION_VAL_85_CHS },
         { "90%",  OPTION_VAL_90_CHS },
         { "95%",  OPTION_VAL_95_CHS },
         { "100%", OPTION_VAL_100_O42_CHS },
         { "105%", OPTION_VAL_105_CHS },
         { "110%", OPTION_VAL_110_CHS },
         { "115%", OPTION_VAL_115_CHS },
         { "120%", OPTION_VAL_120_CHS },
         { "125%", OPTION_VAL_125_CHS },
         { "130%", OPTION_VAL_130_CHS },
         { "135%", OPTION_VAL_135_CHS },
         { "140%", OPTION_VAL_140_CHS },
         { "145%", OPTION_VAL_145_CHS },
         { "150%", OPTION_VAL_150_CHS },
         { "155%", OPTION_VAL_155_CHS },
         { "160%", OPTION_VAL_160_CHS },
         { "165%", OPTION_VAL_165_CHS },
         { "170%", OPTION_VAL_170_CHS },
         { "175%", OPTION_VAL_175_CHS },
         { "180%", OPTION_VAL_180_CHS },
         { "185%", OPTION_VAL_185_CHS },
         { "190%", OPTION_VAL_190_CHS },
         { "195%", OPTION_VAL_195_CHS },
         { "200%", OPTION_VAL_200_CHS },
         { NULL, NULL },
      },
      "100%"
   },
   {
      BEETLE_OPT(negcon_response),
      BEETLE_OPT_NEGCON_RESPONSE_LABEL_CHS,
      NULL,
      BEETLE_OPT_NEGCON_RESPONSE_INFO_0_CHS,
      NULL,
      "input",
      {
         { "linear",    OPTION_VAL_LINEAR_CHS },
         { "quadratic", OPTION_VAL_QUADRATIC_CHS },
         { "cubic",     OPTION_VAL_CUBIC_CHS },
         { NULL, NULL },
      },
      "linear"
   },
   {
      BEETLE_OPT(negcon_deadzone),
      BEETLE_OPT_NEGCON_DEADZONE_LABEL_CHS,
      NULL,
      BEETLE_OPT_NEGCON_DEADZONE_INFO_0_CHS,
      NULL,
      "input",
      {
         { "0%",  OPTION_VAL_0_O44_CHS },
         { "5%",  OPTION_VAL_5_O42_CHS },
         { "10%", OPTION_VAL_10_CHS },
         { "15%", OPTION_VAL_15_CHS },
         { "20%", OPTION_VAL_20_CHS },
         { "25%", OPTION_VAL_25_CHS },
         { "30%", OPTION_VAL_30_CHS },
         { NULL, NULL },
      },
      "0%"
   },
   {
      BEETLE_OPT(use_mednafen_memcard0_method),
      BEETLE_OPT_USE_MEDNAFEN_MEMCARD0_METHOD_LABEL_CHS,
      NULL,
      BEETLE_OPT_USE_MEDNAFEN_MEMCARD0_METHOD_INFO_0_CHS,
      NULL,
      "memcards",
      {
         { "libretro", OPTION_VAL_LIBRETRO_CHS },
         { "mednafen", OPTION_VAL_MEDNAFEN_CHS },
         { NULL, NULL },
      },
      "libretro"
   },
   {
      BEETLE_OPT(enable_memcard1),
      BEETLE_OPT_ENABLE_MEMCARD1_LABEL_CHS,
      NULL,
      BEETLE_OPT_ENABLE_MEMCARD1_INFO_0_CHS,
      NULL,
      "memcards",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(shared_memory_cards),
      BEETLE_OPT_SHARED_MEMORY_CARDS_LABEL_CHS,
      NULL,
      BEETLE_OPT_SHARED_MEMORY_CARDS_INFO_0_CHS,
      NULL,
      "memcards",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(memcard_left_index),
      BEETLE_OPT_MEMCARD_LEFT_INDEX_LABEL_CHS,
      NULL,
      BEETLE_OPT_MEMCARD_LEFT_INDEX_INFO_0_CHS,
      NULL,
      "memcards",
      {
         { "0",  OPTION_VAL_0_O48_CHS },
         { "1",  NULL },
         { "2",  NULL },
         { "3",  NULL },
         { "4",  NULL },
         { "5",  NULL },
         { "6",  NULL },
         { "7",  NULL },
         { "8",  NULL },
         { "9",  NULL },
         { "10", NULL },
         { "11", NULL },
         { "12", NULL },
         { "13", NULL },
         { "14", NULL },
         { "15", NULL },
         { "16", NULL },
         { "17", NULL },
         { "18", NULL },
         { "19", NULL },
         { "20", NULL },
         { "21", NULL },
         { "22", NULL },
         { "23", NULL },
         { "24", NULL },
         { "25", NULL },
         { "26", NULL },
         { "27", NULL },
         { "28", NULL },
         { "29", NULL },
         { "30", NULL },
         { "31", NULL },
         { "32", NULL },
         { "33", NULL },
         { "34", NULL },
         { "35", NULL },
         { "36", NULL },
         { "37", NULL },
         { "38", NULL },
         { "39", NULL },
         { "40", NULL },
         { "41", NULL },
         { "42", NULL },
         { "43", NULL },
         { "44", NULL },
         { "45", NULL },
         { "46", NULL },
         { "47", NULL },
         { "48", NULL },
         { "49", NULL },
         { "50", NULL },
         { "51", NULL },
         { "52", NULL },
         { "53", NULL },
         { "54", NULL },
         { "55", NULL },
         { "56", NULL },
         { "57", NULL },
         { "58", NULL },
         { "59", NULL },
         { "60", NULL },
         { "61", NULL },
         { "62", NULL },
         { "63", NULL },
         { NULL, NULL },
      },
      "0"
   },
   {
      BEETLE_OPT(memcard_right_index),
      BEETLE_OPT_MEMCARD_RIGHT_INDEX_LABEL_CHS,
      NULL,
      BEETLE_OPT_MEMCARD_RIGHT_INDEX_INFO_0_CHS,
      NULL,
      "memcards",
      {
         { "0",  NULL },
         { "1",  OPTION_VAL_1_O49_CHS },
         { "2",  NULL },
         { "3",  NULL },
         { "4",  NULL },
         { "5",  NULL },
         { "6",  NULL },
         { "7",  NULL },
         { "8",  NULL },
         { "9",  NULL },
         { "10", NULL },
         { "11", NULL },
         { "12", NULL },
         { "13", NULL },
         { "14", NULL },
         { "15", NULL },
         { "16", NULL },
         { "17", NULL },
         { "18", NULL },
         { "19", NULL },
         { "20", NULL },
         { "21", NULL },
         { "22", NULL },
         { "23", NULL },
         { "24", NULL },
         { "25", NULL },
         { "26", NULL },
         { "27", NULL },
         { "28", NULL },
         { "29", NULL },
         { "30", NULL },
         { "31", NULL },
         { "32", NULL },
         { "33", NULL },
         { "34", NULL },
         { "35", NULL },
         { "36", NULL },
         { "37", NULL },
         { "38", NULL },
         { "39", NULL },
         { "40", NULL },
         { "41", NULL },
         { "42", NULL },
         { "43", NULL },
         { "44", NULL },
         { "45", NULL },
         { "46", NULL },
         { "47", NULL },
         { "48", NULL },
         { "49", NULL },
         { "50", NULL },
         { "51", NULL },
         { "52", NULL },
         { "53", NULL },
         { "54", NULL },
         { "55", NULL },
         { "56", NULL },
         { "57", NULL },
         { "58", NULL },
         { "59", NULL },
         { "60", NULL },
         { "61", NULL },
         { "62", NULL },
         { "63", NULL },
         { NULL, NULL },
      },
      "1"
   },
   {
      BEETLE_OPT(line_render),
      BEETLE_OPT_LINE_RENDER_LABEL_CHS,
      NULL,
      BEETLE_OPT_LINE_RENDER_INFO_0_CHS,
      NULL,
      "hacks",
      {
         { "default",    OPTION_VAL_DEFAULT_CHS },
         { "aggressive", OPTION_VAL_AGGRESSIVE_CHS },
         { "disabled",   NULL },
         { NULL, NULL },
      },
      "default"
   },
   {
      BEETLE_OPT(widescreen_hack),
      BEETLE_OPT_WIDESCREEN_HACK_LABEL_CHS,
      NULL,
      BEETLE_OPT_WIDESCREEN_HACK_INFO_0_CHS,
      NULL,
      "hacks",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(widescreen_hack_aspect_ratio),
      BEETLE_OPT_WIDESCREEN_HACK_ASPECT_RATIO_LABEL_CHS,
      NULL,
      BEETLE_OPT_WIDESCREEN_HACK_ASPECT_RATIO_INFO_0_CHS,
      NULL,
      "hacks",
      {
         { "16:9",  OPTION_VAL_16_9_CHS },
         { "16:10", OPTION_VAL_16_10_CHS },
         { "18:9",  OPTION_VAL_18_9_CHS },
         { "19:9",  OPTION_VAL_19_9_CHS },
         { "20:9",  OPTION_VAL_20_9_CHS },
         { "21:9",  OPTION_VAL_21_9_CHS }, // 64:27
         { "32:9",  OPTION_VAL_32_9_CHS },
         { NULL,    NULL },
      },
      "16:9"
   },
#if defined(HAVE_LIGHTREC)
   {
      BEETLE_OPT(cpu_dynarec),
      BEETLE_OPT_CPU_DYNAREC_LABEL_CHS,
      NULL,
      BEETLE_OPT_CPU_DYNAREC_INFO_0_CHS,
      NULL,
      "hacks",
      {
         { "disabled",        OPTION_VAL_DISABLED_CHS },
         { "execute",         OPTION_VAL_EXECUTE_CHS },
         { "execute_one",     OPTION_VAL_EXECUTE_ONE_CHS },
         { "run_interpreter", OPTION_VAL_RUN_INTERPRETER_CHS },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(dynarec_invalidate),
      BEETLE_OPT_DYNAREC_INVALIDATE_LABEL_CHS,
      NULL,
      BEETLE_OPT_DYNAREC_INVALIDATE_INFO_0_CHS,
      NULL,
      "hacks",
      {
         { "full", OPTION_VAL_FULL_CHS },
         { "dma",  OPTION_VAL_DMA_CHS },
         { NULL, NULL },
      },
      "full"
   },
   {
      BEETLE_OPT(dynarec_eventcycles),
      BEETLE_OPT_DYNAREC_EVENTCYCLES_LABEL_CHS,
      NULL,
      BEETLE_OPT_DYNAREC_EVENTCYCLES_INFO_0_CHS,
      NULL,
      "hacks",
      {
         { "128",  OPTION_VAL_128_CHS },
         { "256",  NULL },
         { "384",  NULL },
         { "512",  NULL },
         { "640",  NULL },
         { "768",  NULL },
         { "896",  NULL },
         { "1024", NULL },
         { "1152", NULL },
         { "1280", NULL },
         { "1408", NULL },
         { "1536", NULL },
         { "1664", NULL },
         { "1792", NULL },
         { "1920", NULL },
         { "2048", NULL },
         { NULL, NULL },
      },
      "128"
   },
   {
      BEETLE_OPT(dynarec_spu_samples),
      BEETLE_OPT_DYNAREC_SPU_SAMPLES_LABEL_CHS,
      NULL,
      BEETLE_OPT_DYNAREC_SPU_SAMPLES_INFO_0_CHS,
      NULL,
      "hacks",
      {
         { "1",  OPTION_VAL_1_O49_CHS },
         { "4",  NULL },
         { "16", NULL },
         { NULL, NULL },
      },
      "1"
   },
#endif
   {
      BEETLE_OPT(core_timing_fps),
      BEETLE_OPT_CORE_TIMING_FPS_LABEL_CHS,
      NULL,
      BEETLE_OPT_CORE_TIMING_FPS_INFO_0_CHS,
      NULL,
      "video",
      {
         { "auto_toggle",       OPTION_VAL_AUTO_TOGGLE_CHS },
         { "force_progressive", OPTION_VAL_FORCE_PROGRESSIVE_CHS },
         { "force_interlaced",  OPTION_VAL_FORCE_INTERLACED_CHS },
      },
      "auto_toggle"
   },
   {
      BEETLE_OPT(deinterlacer),
      BEETLE_OPT_DEINTERLACER_LABEL_CHS,
      NULL,
      BEETLE_OPT_DEINTERLACER_INFO_0_CHS,
      NULL,
      "video",
      {
         { "weave",      OPTION_VAL_WEAVE_CHS },
         { "bob",        OPTION_VAL_BOB_CHS },
         { "bob_offset", OPTION_VAL_BOB_OFFSET_CHS },
         { "fastmad",    OPTION_VAL_FASTMAD_CHS },
         { "off",        "Off" },
         { NULL, NULL },
      },
      "weave"
   },
   {
      BEETLE_OPT(frame_duping),
      BEETLE_OPT_FRAME_DUPING_LABEL_CHS,
      NULL,
      BEETLE_OPT_FRAME_DUPING_INFO_0_CHS,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(pal_video_timing_override),
      BEETLE_OPT_PAL_VIDEO_TIMING_OVERRIDE_LABEL_CHS,
      NULL,
      BEETLE_OPT_PAL_VIDEO_TIMING_OVERRIDE_INFO_0_CHS,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(aspect_ratio),
      BEETLE_OPT_ASPECT_RATIO_LABEL_CHS,
      NULL,
      BEETLE_OPT_ASPECT_RATIO_INFO_0_CHS,
      NULL,
      "video",
      {
         { "corrected", OPTION_VAL_CORRECTED_CHS },
         { "uncorrected", OPTION_VAL_UNCORRECTED_CHS },
         { "4:3",  OPTION_VAL_4_3_CHS },
         { "ntsc", OPTION_VAL_NTSC_CHS },
      },
      "corrected"
   },
   {
      BEETLE_OPT(crop_overscan),
      BEETLE_OPT_CROP_OVERSCAN_LABEL_CHS,
      NULL,
      BEETLE_OPT_CROP_OVERSCAN_INFO_0_CHS,
      NULL,
      "video",
      {
         { "disabled", "Disabled" },
         { "static",   OPTION_VAL_STATIC_CHS },
         { "smart",    OPTION_VAL_SMART_CHS },
         { NULL, NULL },
      },
      "smart"
   },
   {
      BEETLE_OPT(image_crop),
      BEETLE_OPT_IMAGE_CROP_LABEL_CHS,
      NULL,
      BEETLE_OPT_IMAGE_CROP_INFO_0_CHS,
      NULL,
      "video",
      {
         { "disabled", "0" },
         { "1px",      OPTION_VAL_1PX_CHS },
         { "2px",      OPTION_VAL_2PX_CHS },
         { "3px",      OPTION_VAL_3PX_CHS },
         { "4px",      OPTION_VAL_4PX_CHS },
         { "5px",      OPTION_VAL_5PX_CHS },
         { "6px",      OPTION_VAL_6PX_CHS },
         { "7px",      OPTION_VAL_7PX_CHS },
         { "8px",      OPTION_VAL_8PX_CHS },
         { "9px",      OPTION_VAL_9PX_CHS },
         { "10px",     OPTION_VAL_10PX_CHS },
         { "11px",     OPTION_VAL_11PX_CHS },
         { "12px",     OPTION_VAL_12PX_CHS },
         { "13px",     OPTION_VAL_13PX_CHS },
         { "14px",     OPTION_VAL_14PX_CHS },
         { "15px",     OPTION_VAL_15PX_CHS },
         { "16px",     OPTION_VAL_16PX_CHS },
         { "17px",     OPTION_VAL_17PX_CHS },
         { "18px",     OPTION_VAL_18PX_CHS },
         { "19px",     OPTION_VAL_19PX_CHS },
         { "20px",     OPTION_VAL_20PX_CHS },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(image_offset),
      BEETLE_OPT_IMAGE_OFFSET_LABEL_CHS,
      NULL,
      BEETLE_OPT_IMAGE_OFFSET_INFO_0_CHS,
      NULL,
      "video",
      {
         { "-12px",    OPTION_VAL_12PX_O64_CHS },
         { "-11px",    OPTION_VAL_11PX_O64_CHS },
         { "-10px",    OPTION_VAL_10PX_O64_CHS },
         { "-9px",     OPTION_VAL_9PX_O64_CHS },
         { "-8px",     OPTION_VAL_8PX_O64_CHS },
         { "-7px",     OPTION_VAL_7PX_O64_CHS },
         { "-6px",     OPTION_VAL_6PX_O64_CHS },
         { "-5px",     OPTION_VAL_5PX_O64_CHS },
         { "-4px",     OPTION_VAL_4PX_O64_CHS },
         { "-3px",     OPTION_VAL_3PX_O64_CHS },
         { "-2px",     OPTION_VAL_2PX_O64_CHS },
         { "-1px",     OPTION_VAL_1PX_O64_CHS },
         { "disabled", "0" },
         { "+1px",     OPTION_VAL_1PX_O64_0_CHS },
         { "+2px",     OPTION_VAL_2PX_O64_0_CHS },
         { "+3px",     OPTION_VAL_3PX_O64_0_CHS },
         { "+4px",     OPTION_VAL_4PX_O64_0_CHS },
         { "+5px",     OPTION_VAL_5PX_O64_0_CHS },
         { "+6px",     OPTION_VAL_6PX_O64_0_CHS },
         { "+7px",     OPTION_VAL_7PX_O64_0_CHS },
         { "+8px",     OPTION_VAL_8PX_O64_0_CHS },
         { "+9px",     OPTION_VAL_9PX_O64_0_CHS },
         { "+10px",    OPTION_VAL_10PX_O64_0_CHS },
         { "+11px",    OPTION_VAL_11PX_O64_0_CHS },
         { "+12px",    OPTION_VAL_12PX_O64_0_CHS },
         { NULL, NULL },
      },
      "disabled"
   },
#if defined(HAVE_OPENGL) || defined(HAVE_OPENGLES) || defined(HAVE_VULKAN)
   {
      BEETLE_OPT(image_offset_cycles),
      BEETLE_OPT_IMAGE_OFFSET_CYCLES_LABEL_CHS,
      NULL,
      BEETLE_OPT_IMAGE_OFFSET_CYCLES_INFO_0_CHS,
      NULL,
      "video",
      {
         { "-40",      NULL },
         { "-39",      NULL },
         { "-38",      NULL },
         { "-37",      NULL },
         { "-36",      NULL },
         { "-35",      NULL },
         { "-34",      NULL },
         { "-33",      NULL },
         { "-32",      NULL },
         { "-31",      NULL },
         { "-30",      NULL },
         { "-29",      NULL },
         { "-28",      NULL },
         { "-27",      NULL },
         { "-26",      NULL },
         { "-25",      NULL },
         { "-24",      NULL },
         { "-23",      NULL },
         { "-22",      NULL },
         { "-21",      NULL },
         { "-20",      NULL },
         { "-19",      NULL },
         { "-18",      NULL },
         { "-17",      NULL },
         { "-16",      NULL },
         { "-15",      NULL },
         { "-14",      NULL },
         { "-13",      NULL },
         { "-12",      NULL },
         { "-11",      NULL },
         { "-10",      NULL },
         { "-9",       NULL },
         { "-8",       NULL },
         { "-7",       NULL },
         { "-6",       NULL },
         { "-5",       NULL },
         { "-4",       NULL },
         { "-3",       NULL },
         { "-2",       NULL },
         { "-1",       NULL },
         { "0",        "0" },
         { "+1",       NULL },
         { "+2",       NULL },
         { "+3",       NULL },
         { "+4",       NULL },
         { "+5",       NULL },
         { "+6",       NULL },
         { "+7",       NULL },
         { "+8",       NULL },
         { "+9",       NULL },
         { "+10",      NULL },
         { "+11",      NULL },
         { "+12",      NULL },
         { "+13",      NULL },
         { "+14",      NULL },
         { "+15",      NULL },
         { "+16",      NULL },
         { "+17",      NULL },
         { "+18",      NULL },
         { "+19",      NULL },
         { "+20",      NULL },
         { "+21",      NULL },
         { "+22",      NULL },
         { "+23",      NULL },
         { "+24",      NULL },
         { "+25",      NULL },
         { "+26",      NULL },
         { "+27",      NULL },
         { "+28",      NULL },
         { "+29",      NULL },
         { "+30",      NULL },
         { "+31",      NULL },
         { "+32",      NULL },
         { "+33",      NULL },
         { "+34",      NULL },
         { "+35",      NULL },
         { "+36",      NULL },
         { "+37",      NULL },
         { "+38",      NULL },
         { "+39",      NULL },
         { "+40",      NULL },
         { NULL, NULL},
      },
      "0"
   },
#endif
   {
      BEETLE_OPT(initial_scanline),
      BEETLE_OPT_INITIAL_SCANLINE_LABEL_CHS,
      NULL,
      BEETLE_OPT_INITIAL_SCANLINE_INFO_0_CHS,
      NULL,
      "video",
      {
         { "0",  OPTION_VAL_0_O48_CHS },
         { "1",  NULL },
         { "2",  NULL },
         { "3",  NULL },
         { "4",  NULL },
         { "5",  NULL },
         { "6",  NULL },
         { "7",  NULL },
         { "8",  NULL },
         { "9",  NULL },
         { "10", NULL },
         { "11", NULL },
         { "12", NULL },
         { "13", NULL },
         { "14", NULL },
         { "15", NULL },
         { "16", NULL },
         { "17", NULL },
         { "18", NULL },
         { "19", NULL },
         { "20", NULL },
         { "21", NULL },
         { "22", NULL },
         { "23", NULL },
         { "24", NULL },
         { "25", NULL },
         { "26", NULL },
         { "27", NULL },
         { "28", NULL },
         { "29", NULL },
         { "30", NULL },
         { "31", NULL },
         { "32", NULL },
         { "33", NULL },
         { "34", NULL },
         { "35", NULL },
         { "36", NULL },
         { "37", NULL },
         { "38", NULL },
         { "39", NULL },
         { "40", NULL },
         { NULL, NULL },
      },
      "0"
   },
   {
      BEETLE_OPT(last_scanline),
      BEETLE_OPT_LAST_SCANLINE_LABEL_CHS,
      NULL,
      BEETLE_OPT_LAST_SCANLINE_INFO_0_CHS,
      NULL,
      "video",
      {
         { "210", NULL },
         { "211", NULL },
         { "212", NULL },
         { "213", NULL },
         { "214", NULL },
         { "215", NULL },
         { "216", NULL },
         { "217", NULL },
         { "218", NULL },
         { "219", NULL },
         { "220", NULL },
         { "221", NULL },
         { "222", NULL },
         { "223", NULL },
         { "224", NULL },
         { "225", NULL },
         { "226", NULL },
         { "227", NULL },
         { "228", NULL },
         { "229", NULL },
         { "230", NULL },
         { "231", NULL },
         { "232", NULL },
         { "233", NULL },
         { "234", NULL },
         { "235", NULL },
         { "236", NULL },
         { "237", NULL },
         { "238", NULL },
         { "239", OPTION_VAL_239_CHS },
         { NULL, NULL },
      },
      "239"
   },
   {
      BEETLE_OPT(initial_scanline_pal),
      BEETLE_OPT_INITIAL_SCANLINE_PAL_LABEL_CHS,
      NULL,
      BEETLE_OPT_INITIAL_SCANLINE_PAL_INFO_0_CHS,
      NULL,
      "video",
      {
         { "0",  OPTION_VAL_0_O48_CHS },
         { "1",  NULL },
         { "2",  NULL },
         { "3",  NULL },
         { "4",  NULL },
         { "5",  NULL },
         { "6",  NULL },
         { "7",  NULL },
         { "8",  NULL },
         { "9",  NULL },
         { "10", NULL },
         { "11", NULL },
         { "12", NULL },
         { "13", NULL },
         { "14", NULL },
         { "15", NULL },
         { "16", NULL },
         { "17", NULL },
         { "18", NULL },
         { "19", NULL },
         { "20", NULL },
         { "21", NULL },
         { "22", NULL },
         { "23", NULL },
         { "24", NULL },
         { "25", NULL },
         { "26", NULL },
         { "27", NULL },
         { "28", NULL },
         { "29", NULL },
         { "30", NULL },
         { "31", NULL },
         { "32", NULL },
         { "33", NULL },
         { "34", NULL },
         { "35", NULL },
         { "36", NULL },
         { "37", NULL },
         { "38", NULL },
         { "39", NULL },
         { "40", NULL },
         { NULL, NULL },
      },
      "0"
   },
   {
      BEETLE_OPT(last_scanline_pal),
      BEETLE_OPT_LAST_SCANLINE_PAL_LABEL_CHS,
      NULL,
      BEETLE_OPT_LAST_SCANLINE_PAL_INFO_0_CHS,
      NULL,
      "video",
      {
         { "230", NULL },
         { "231", NULL },
         { "232", NULL },
         { "233", NULL },
         { "234", NULL },
         { "235", NULL },
         { "236", NULL },
         { "237", NULL },
         { "238", NULL },
         { "239", NULL },
         { "240", NULL },
         { "241", NULL },
         { "242", NULL },
         { "243", NULL },
         { "244", NULL },
         { "245", NULL },
         { "246", NULL },
         { "247", NULL },
         { "248", NULL },
         { "249", NULL },
         { "250", NULL },
         { "251", NULL },
         { "252", NULL },
         { "253", NULL },
         { "254", NULL },
         { "255", NULL },
         { "256", NULL },
         { "257", NULL },
         { "258", NULL },
         { "259", NULL },
         { "260", NULL },
         { "261", NULL },
         { "262", NULL },
         { "263", NULL },
         { "264", NULL },
         { "265", NULL },
         { "266", NULL },
         { "267", NULL },
         { "268", NULL },
         { "269", NULL },
         { "270", NULL },
         { "271", NULL },
         { "272", NULL },
         { "273", NULL },
         { "274", NULL },
         { "275", NULL },
         { "276", NULL },
         { "277", NULL },
         { "278", NULL },
         { "279", NULL },
         { "280", NULL },
         { "281", NULL },
         { "282", NULL },
         { "283", NULL },
         { "284", NULL },
         { "285", NULL },
         { "286", NULL },
         { "287", OPTION_VAL_287_CHS },
         { NULL, NULL },
      },
      "287"
   },
   { NULL, NULL, NULL, NULL, NULL, NULL, {{0}}, NULL },
};
struct retro_core_options_v2 options_chs = {
   option_cats_chs,
   option_defs_chs
};

/* RETRO_LANGUAGE_CHT */

#define CATEGORY_SYSTEM_LABEL_CHT "系統"
#define CATEGORY_SYSTEM_INFO_0_CHT NULL
#define CATEGORY_VIDEO_LABEL_CHT "顯示"
#define CATEGORY_VIDEO_INFO_0_CHT "變更長寬比、畫面裁剪、影像濾鏡和跳幀的設定。"
#define CATEGORY_PGXP_LABEL_CHT "精密幾何轉換管線 (PGXP)"
#define CATEGORY_PGXP_INFO_0_CHT "變更改進圖形功能的設定, PGXP可消除扭曲紋理和高度軸衝突問題。"
#define CATEGORY_OSD_LABEL_CHT "螢幕選單"
#define CATEGORY_OSD_INFO_0_CHT "變更螢幕上顯示資訊的設定。"
#define CATEGORY_INPUT_LABEL_CHT "輸入"
#define CATEGORY_INPUT_INFO_0_CHT "變更轉接插座、光線槍、滑鼠和類比搖桿的設定。"
#define CATEGORY_MEMCARDS_LABEL_CHT "記憶卡"
#define CATEGORY_MEMCARDS_INFO_0_CHT NULL
#define CATEGORY_HACKS_LABEL_CHT "進階"
#define CATEGORY_HACKS_INFO_0_CHT NULL
#define BEETLE_OPT_CPU_FREQ_SCALE_LABEL_CHT NULL
#define BEETLE_OPT_CPU_FREQ_SCALE_INFO_0_CHT NULL
#define OPTION_VAL_50_CHT NULL
#define OPTION_VAL_60_CHT NULL
#define OPTION_VAL_70_CHT NULL
#define OPTION_VAL_80_CHT NULL
#define OPTION_VAL_90_CHT NULL
#define OPTION_VAL_100_CHT NULL
#define OPTION_VAL_110_CHT NULL
#define OPTION_VAL_120_CHT NULL
#define OPTION_VAL_130_CHT NULL
#define OPTION_VAL_140_CHT NULL
#define OPTION_VAL_150_CHT NULL
#define OPTION_VAL_160_CHT NULL
#define OPTION_VAL_170_CHT NULL
#define OPTION_VAL_180_CHT NULL
#define OPTION_VAL_190_CHT NULL
#define OPTION_VAL_200_CHT NULL
#define OPTION_VAL_210_CHT NULL
#define OPTION_VAL_220_CHT NULL
#define OPTION_VAL_230_CHT NULL
#define OPTION_VAL_240_CHT NULL
#define OPTION_VAL_250_CHT NULL
#define OPTION_VAL_260_CHT NULL
#define OPTION_VAL_270_CHT NULL
#define OPTION_VAL_280_CHT NULL
#define OPTION_VAL_290_CHT NULL
#define OPTION_VAL_300_CHT NULL
#define OPTION_VAL_310_CHT NULL
#define OPTION_VAL_320_CHT NULL
#define OPTION_VAL_330_CHT NULL
#define OPTION_VAL_340_CHT NULL
#define OPTION_VAL_350_CHT NULL
#define OPTION_VAL_360_CHT NULL
#define OPTION_VAL_370_CHT NULL
#define OPTION_VAL_380_CHT NULL
#define OPTION_VAL_390_CHT NULL
#define OPTION_VAL_400_CHT NULL
#define OPTION_VAL_410_CHT NULL
#define OPTION_VAL_420_CHT NULL
#define OPTION_VAL_430_CHT NULL
#define OPTION_VAL_440_CHT NULL
#define OPTION_VAL_450_CHT NULL
#define OPTION_VAL_460_CHT NULL
#define OPTION_VAL_470_CHT NULL
#define OPTION_VAL_480_CHT NULL
#define OPTION_VAL_490_CHT NULL
#define OPTION_VAL_500_CHT NULL
#define OPTION_VAL_510_CHT NULL
#define OPTION_VAL_520_CHT NULL
#define OPTION_VAL_530_CHT NULL
#define OPTION_VAL_540_CHT NULL
#define OPTION_VAL_550_CHT NULL
#define OPTION_VAL_560_CHT NULL
#define OPTION_VAL_570_CHT NULL
#define OPTION_VAL_580_CHT NULL
#define OPTION_VAL_590_CHT NULL
#define OPTION_VAL_600_CHT NULL
#define OPTION_VAL_610_CHT NULL
#define OPTION_VAL_620_CHT NULL
#define OPTION_VAL_630_CHT NULL
#define OPTION_VAL_640_CHT NULL
#define OPTION_VAL_650_CHT NULL
#define OPTION_VAL_660_CHT NULL
#define OPTION_VAL_670_CHT NULL
#define OPTION_VAL_680_CHT NULL
#define OPTION_VAL_690_CHT NULL
#define OPTION_VAL_700_CHT NULL
#define OPTION_VAL_710_CHT NULL
#define OPTION_VAL_720_CHT NULL
#define OPTION_VAL_730_CHT NULL
#define OPTION_VAL_740_CHT NULL
#define OPTION_VAL_750_CHT NULL
#define BEETLE_OPT_GPU_OVERCLOCK_LABEL_CHT NULL
#define BEETLE_OPT_GPU_OVERCLOCK_INFO_0_CHT NULL
#define OPTION_VAL_1X_NATIVE_CHT "1倍 (原廠)"
#define OPTION_VAL_2X_CHT "2倍"
#define OPTION_VAL_4X_CHT "4倍"
#define OPTION_VAL_8X_CHT "8倍"
#define OPTION_VAL_16X_CHT "16倍"
#define OPTION_VAL_32X_CHT "32倍"
#define BEETLE_OPT_GTE_OVERCLOCK_LABEL_CHT NULL
#define BEETLE_OPT_GTE_OVERCLOCK_INFO_0_CHT NULL
#define BEETLE_OPT_CD_ACCESS_METHOD_LABEL_CHT NULL
#define BEETLE_OPT_CD_ACCESS_METHOD_INFO_0_CHT NULL
#define OPTION_VAL_SYNC_CHT NULL
#define OPTION_VAL_ASYNC_CHT NULL
#define OPTION_VAL_PRECACHE_CHT NULL
#define BEETLE_OPT_CD_FASTLOAD_LABEL_CHT NULL
#define BEETLE_OPT_CD_FASTLOAD_INFO_0_CHT NULL
#define OPTION_VAL_2X_NATIVE_CHT "2倍 (原廠)"
#define OPTION_VAL_6X_CHT "6倍"
#define OPTION_VAL_10X_CHT "10倍"
#define OPTION_VAL_12X_CHT "12倍"
#define OPTION_VAL_14X_CHT "14倍"
#define BEETLE_OPT_SKIP_BIOS_LABEL_CHT "跳過BIOS"
#define BEETLE_OPT_SKIP_BIOS_INFO_0_CHT NULL
#define BEETLE_OPT_OVERRIDE_BIOS_LABEL_CHT NULL
#define BEETLE_OPT_OVERRIDE_BIOS_INFO_0_CHT NULL
#define OPTION_VAL_PSXONPSP_CHT "PSP內建 (psxonpsp660.bin)"
#define OPTION_VAL_PS1_ROM_CHT "PS3內建 (ps1_rom.bin)"
#define OPTION_VAL_OPENBIOS_CHT NULL
#define BEETLE_OPT_REGION_LABEL_CHT "系統區域"
#define BEETLE_OPT_REGION_INFO_0_CHT NULL
#define OPTION_VAL_AUTO_CHT "自動"
#define OPTION_VAL_NTSC_J_CHT NULL
#define OPTION_VAL_NTSC_U_CHT NULL
#define OPTION_VAL_PAL_CHT NULL
#define BEETLE_OPT_INTERNAL_RESOLUTION_LABEL_CHT "內部解析度 (GPU)"
#define BEETLE_OPT_INTERNAL_RESOLUTION_INFO_0_CHT NULL
#define BEETLE_OPT_RENDERER_LABEL_CHT NULL
#define BEETLE_OPT_RENDERER_INFO_0_CHT NULL
#define OPTION_VAL_HARDWARE_CHT "自動 (硬體)"
#define OPTION_VAL_HARDWARE_GL_CHT "OpenGL (硬體)"
#define OPTION_VAL_HARDWARE_VK_CHT "Vulkan (硬體)"
#define OPTION_VAL_SOFTWARE_CHT "軟體渲染"
#define BEETLE_OPT_RENDERER_SOFTWARE_FB_LABEL_CHT NULL
#define BEETLE_OPT_RENDERER_SOFTWARE_FB_INFO_0_CHT NULL
#define BEETLE_OPT_DEPTH_LABEL_CHT "內部色彩深度"
#define BEETLE_OPT_DEPTH_INFO_0_CHT NULL
#define OPTION_VAL_16BPP_NATIVE_CHT "每像素16位元 (原廠)"
#define OPTION_VAL_32BPP_CHT "每像素32位元"
#define BEETLE_OPT_DITHER_MODE_LABEL_CHT "抖動色彩效果"
#define BEETLE_OPT_DITHER_MODE_INFO_0_CHT NULL
#define OPTION_VAL_INTERNAL_RESOLUTION_CHT "內部解析度"
#define BEETLE_OPT_SCALED_UV_OFFSET_LABEL_CHT NULL
#define BEETLE_OPT_SCALED_UV_OFFSET_INFO_0_CHT NULL
#define BEETLE_OPT_FILTER_LABEL_CHT "材質過濾"
#define BEETLE_OPT_FILTER_INFO_0_CHT NULL
#define OPTION_VAL_NEAREST_CHT "最接近"
#define OPTION_VAL_SABR_CHT NULL
#define OPTION_VAL_XBR_CHT NULL
#define OPTION_VAL_BILINEAR_CHT "雙線性插值"
#define OPTION_VAL_3_POINT_CHT NULL
#define OPTION_VAL_JINC2_CHT NULL
#define BEETLE_OPT_FILTER_EXCLUDE_SPRITE_LABEL_CHT NULL
#define BEETLE_OPT_FILTER_EXCLUDE_SPRITE_INFO_0_CHT NULL
#define OPTION_VAL_OPAQUE_CHT "僅限不透明"
#define OPTION_VAL_ALL_CHT "不透明與半透明"
#define BEETLE_OPT_FILTER_EXCLUDE_2D_POLYGON_LABEL_CHT NULL
#define BEETLE_OPT_FILTER_EXCLUDE_2D_POLYGON_INFO_0_CHT "請勿在2D多邊形使用材質過濾。2D多邊形使用啟發法來偵測並可能造成小故障。搭配適應平滑法或其他後期處理濾鏡來達到最佳效果。"
#define BEETLE_OPT_ADAPTIVE_SMOOTHING_LABEL_CHT NULL
#define BEETLE_OPT_ADAPTIVE_SMOOTHING_INFO_0_CHT NULL
#define BEETLE_OPT_SUPER_SAMPLING_LABEL_CHT "超級取樣 (降採樣至原生解析度)"
#define BEETLE_OPT_SUPER_SAMPLING_INFO_0_CHT NULL
#define BEETLE_OPT_MSAA_LABEL_CHT "多重取樣反鋸齒"
#define BEETLE_OPT_MSAA_INFO_0_CHT NULL
#define OPTION_VAL_1X_CHT "1倍"
#define BEETLE_OPT_MDEC_YUV_LABEL_CHT NULL
#define BEETLE_OPT_MDEC_YUV_INFO_0_CHT NULL
#define BEETLE_OPT_TRACK_TEXTURES_LABEL_CHT NULL
#define BEETLE_OPT_TRACK_TEXTURES_INFO_0_CHT NULL
#define BEETLE_OPT_DUMP_TEXTURES_LABEL_CHT NULL
#define BEETLE_OPT_DUMP_TEXTURES_INFO_0_CHT NULL
#define BEETLE_OPT_REPLACE_TEXTURES_LABEL_CHT NULL
#define BEETLE_OPT_REPLACE_TEXTURES_INFO_0_CHT NULL
#define BEETLE_OPT_PGXP_MODE_LABEL_CHT NULL
#define BEETLE_OPT_PGXP_MODE_INFO_0_CHT NULL
#define OPTION_VAL_MEMORY_ONLY_CHT "記憶體"
#define OPTION_VAL_MEMORY_CPU_CHT "記憶體 + 中央處理器 (錯誤)"
#define BEETLE_OPT_PGXP_2D_TOL_LABEL_CHT NULL
#define BEETLE_OPT_PGXP_2D_TOL_INFO_0_CHT NULL
#define OPTION_VAL_0PX_CHT "0像素"
#define OPTION_VAL_1PX_CHT "1像素"
#define OPTION_VAL_2PX_CHT "2像素"
#define OPTION_VAL_3PX_CHT "3像素"
#define OPTION_VAL_4PX_CHT "4像素"
#define OPTION_VAL_5PX_CHT "5像素"
#define OPTION_VAL_6PX_CHT "6像素"
#define OPTION_VAL_7PX_CHT "7像素"
#define OPTION_VAL_8PX_CHT "8像素"
#define BEETLE_OPT_PGXP_NCLIP_LABEL_CHT NULL
#define BEETLE_OPT_PGXP_NCLIP_INFO_0_CHT NULL
#define BEETLE_OPT_PGXP_VERTEX_LABEL_CHT NULL
#define BEETLE_OPT_PGXP_VERTEX_INFO_0_CHT NULL
#define BEETLE_OPT_PGXP_TEXTURE_LABEL_CHT NULL
#define BEETLE_OPT_PGXP_TEXTURE_INFO_0_CHT NULL
#define BEETLE_OPT_DISPLAY_INTERNAL_FPS_LABEL_CHT NULL
#define BEETLE_OPT_DISPLAY_INTERNAL_FPS_INFO_0_CHT NULL
#define BEETLE_OPT_DISPLAY_OSD_LABEL_CHT NULL
#define BEETLE_OPT_DISPLAY_OSD_INFO_0_CHT NULL
#define BEETLE_OPT_DISPLAY_VRAM_LABEL_CHT NULL
#define BEETLE_OPT_DISPLAY_VRAM_INFO_0_CHT NULL
#define BEETLE_OPT_ANALOG_CALIBRATION_LABEL_CHT "類比自動校正"
#define BEETLE_OPT_ANALOG_CALIBRATION_INFO_0_CHT NULL
#define BEETLE_OPT_ANALOG_TOGGLE_LABEL_CHT NULL
#define BEETLE_OPT_ANALOG_TOGGLE_INFO_0_CHT NULL
#define OPTION_VAL_ENABLED_ANALOG_CHT NULL
#define BEETLE_OPT_ANALOG_TOGGLE_COMBO_LABEL_CHT NULL
#define BEETLE_OPT_ANALOG_TOGGLE_COMBO_INFO_0_CHT NULL
#define OPTION_VAL_L1_L2_R1_R2_START_SELECT_CHT NULL
#define OPTION_VAL_L1_R1_SELECT_CHT NULL
#define OPTION_VAL_L1_R1_START_CHT NULL
#define OPTION_VAL_L1_R1_L3_CHT NULL
#define OPTION_VAL_L1_R1_R3_CHT NULL
#define OPTION_VAL_L2_R2_SELECT_CHT NULL
#define OPTION_VAL_L2_R2_START_CHT NULL
#define OPTION_VAL_L2_R2_L3_CHT NULL
#define OPTION_VAL_L2_R2_R3_CHT NULL
#define OPTION_VAL_L3_R3_CHT NULL
#define BEETLE_OPT_ANALOG_TOGGLE_HOLD_LABEL_CHT NULL
#define BEETLE_OPT_ANALOG_TOGGLE_HOLD_INFO_0_CHT NULL
#define OPTION_VAL_0_CHT NULL
#define OPTION_VAL_1_CHT NULL
#define OPTION_VAL_2_CHT NULL
#define OPTION_VAL_3_CHT NULL
#define OPTION_VAL_4_CHT NULL
#define OPTION_VAL_5_CHT NULL
#define BEETLE_OPT_ENABLE_MULTITAP_PORT1_LABEL_CHT NULL
#define BEETLE_OPT_ENABLE_MULTITAP_PORT1_INFO_0_CHT "開啟連接埠1轉接插座可對應四位玩家。"
#define BEETLE_OPT_ENABLE_MULTITAP_PORT2_LABEL_CHT NULL
#define BEETLE_OPT_ENABLE_MULTITAP_PORT2_INFO_0_CHT "開啟連接埠2轉接插座可對應四位玩家。"
#define BEETLE_OPT_GUN_INPUT_MODE_LABEL_CHT "光線槍輸入模式"
#define BEETLE_OPT_GUN_INPUT_MODE_INFO_0_CHT NULL
#define OPTION_VAL_LIGHTGUN_CHT "光線槍"
#define OPTION_VAL_TOUCHSCREEN_CHT "觸控螢幕"
#define BEETLE_OPT_GUN_CURSOR_LABEL_CHT "光線槍準心"
#define BEETLE_OPT_GUN_CURSOR_INFO_0_CHT NULL
#define OPTION_VAL_CROSS_CHT "十字"
#define OPTION_VAL_DOT_CHT "圓點"
#define BEETLE_OPT_CROSSHAIR_COLOR_P1_LABEL_CHT NULL
#define BEETLE_OPT_CROSSHAIR_COLOR_P1_INFO_0_CHT NULL
#define OPTION_VAL_RED_CHT "紅色"
#define OPTION_VAL_BLUE_CHT "藍色"
#define OPTION_VAL_GREEN_CHT "綠色"
#define OPTION_VAL_ORANGE_CHT "橙色"
#define OPTION_VAL_YELLOW_CHT "黃色"
#define OPTION_VAL_CYAN_CHT "青色"
#define OPTION_VAL_PINK_CHT "粉紅"
#define OPTION_VAL_PURPLE_CHT "紫色"
#define OPTION_VAL_BLACK_CHT "黑色"
#define OPTION_VAL_WHITE_CHT "白色"
#define BEETLE_OPT_CROSSHAIR_COLOR_P2_LABEL_CHT NULL
#define BEETLE_OPT_CROSSHAIR_COLOR_P2_INFO_0_CHT NULL
#define BEETLE_OPT_MOUSE_SENSITIVITY_LABEL_CHT "滑鼠靈敏度"
#define BEETLE_OPT_MOUSE_SENSITIVITY_INFO_0_CHT NULL
#define OPTION_VAL_5_O42_CHT NULL
#define OPTION_VAL_10_CHT NULL
#define OPTION_VAL_15_CHT NULL
#define OPTION_VAL_20_CHT NULL
#define OPTION_VAL_25_CHT NULL
#define OPTION_VAL_30_CHT NULL
#define OPTION_VAL_35_CHT NULL
#define OPTION_VAL_40_CHT NULL
#define OPTION_VAL_45_CHT NULL
#define OPTION_VAL_55_CHT NULL
#define OPTION_VAL_65_CHT NULL
#define OPTION_VAL_75_CHT NULL
#define OPTION_VAL_85_CHT NULL
#define OPTION_VAL_95_CHT NULL
#define OPTION_VAL_100_O42_CHT NULL
#define OPTION_VAL_105_CHT NULL
#define OPTION_VAL_115_CHT NULL
#define OPTION_VAL_125_CHT NULL
#define OPTION_VAL_135_CHT NULL
#define OPTION_VAL_145_CHT NULL
#define OPTION_VAL_155_CHT NULL
#define OPTION_VAL_165_CHT NULL
#define OPTION_VAL_175_CHT NULL
#define OPTION_VAL_185_CHT NULL
#define OPTION_VAL_195_CHT NULL
#define BEETLE_OPT_NEGCON_RESPONSE_LABEL_CHT "扭轉控制器靈敏度 (neGcon)"
#define BEETLE_OPT_NEGCON_RESPONSE_INFO_0_CHT NULL
#define OPTION_VAL_LINEAR_CHT "線性"
#define OPTION_VAL_QUADRATIC_CHT NULL
#define OPTION_VAL_CUBIC_CHT NULL
#define BEETLE_OPT_NEGCON_DEADZONE_LABEL_CHT "扭轉控制器作用區 (neGcon)"
#define BEETLE_OPT_NEGCON_DEADZONE_INFO_0_CHT NULL
#define OPTION_VAL_0_O44_CHT NULL
#define BEETLE_OPT_USE_MEDNAFEN_MEMCARD0_METHOD_LABEL_CHT NULL
#define BEETLE_OPT_USE_MEDNAFEN_MEMCARD0_METHOD_INFO_0_CHT NULL
#define OPTION_VAL_LIBRETRO_CHT "SRM格式 (Libretro模擬器)"
#define OPTION_VAL_MEDNAFEN_CHT "MCR格式 (Mednafen模擬器)"
#define BEETLE_OPT_ENABLE_MEMCARD1_LABEL_CHT NULL
#define BEETLE_OPT_ENABLE_MEMCARD1_INFO_0_CHT NULL
#define BEETLE_OPT_SHARED_MEMORY_CARDS_LABEL_CHT NULL
#define BEETLE_OPT_SHARED_MEMORY_CARDS_INFO_0_CHT NULL
#define BEETLE_OPT_MEMCARD_LEFT_INDEX_LABEL_CHT NULL
#define BEETLE_OPT_MEMCARD_LEFT_INDEX_INFO_0_CHT NULL
#define OPTION_VAL_0_O48_CHT NULL
#define BEETLE_OPT_MEMCARD_RIGHT_INDEX_LABEL_CHT NULL
#define BEETLE_OPT_MEMCARD_RIGHT_INDEX_INFO_0_CHT NULL
#define OPTION_VAL_1_O49_CHT NULL
#define BEETLE_OPT_LINE_RENDER_LABEL_CHT NULL
#define BEETLE_OPT_LINE_RENDER_INFO_0_CHT NULL
#define OPTION_VAL_DEFAULT_CHT "預設"
#define OPTION_VAL_AGGRESSIVE_CHT "強制"
#define BEETLE_OPT_WIDESCREEN_HACK_LABEL_CHT "寬螢幕模式 (破解)"
#define BEETLE_OPT_WIDESCREEN_HACK_INFO_0_CHT NULL
#define BEETLE_OPT_WIDESCREEN_HACK_ASPECT_RATIO_LABEL_CHT "寬螢幕模式比例"
#define BEETLE_OPT_WIDESCREEN_HACK_ASPECT_RATIO_INFO_0_CHT NULL
#define OPTION_VAL_16_9_CHT NULL
#define OPTION_VAL_16_10_CHT NULL
#define OPTION_VAL_18_9_CHT NULL
#define OPTION_VAL_19_9_CHT NULL
#define OPTION_VAL_20_9_CHT NULL
#define OPTION_VAL_21_9_CHT NULL
#define OPTION_VAL_32_9_CHT NULL
#define BEETLE_OPT_CPU_DYNAREC_LABEL_CHT NULL
#define BEETLE_OPT_CPU_DYNAREC_INFO_0_CHT NULL
#define OPTION_VAL_DISABLED_CHT NULL
#define OPTION_VAL_EXECUTE_CHT "最高效能"
#define OPTION_VAL_EXECUTE_ONE_CHT NULL
#define OPTION_VAL_RUN_INTERPRETER_CHT NULL
#define BEETLE_OPT_DYNAREC_INVALIDATE_LABEL_CHT NULL
#define BEETLE_OPT_DYNAREC_INVALIDATE_INFO_0_CHT NULL
#define OPTION_VAL_FULL_CHT "全部"
#define OPTION_VAL_DMA_CHT NULL
#define BEETLE_OPT_DYNAREC_EVENTCYCLES_LABEL_CHT NULL
#define BEETLE_OPT_DYNAREC_EVENTCYCLES_INFO_0_CHT NULL
#define OPTION_VAL_128_CHT "128 (預設)"
#define BEETLE_OPT_DYNAREC_SPU_SAMPLES_LABEL_CHT NULL
#define BEETLE_OPT_DYNAREC_SPU_SAMPLES_INFO_0_CHT NULL
#define BEETLE_OPT_CORE_TIMING_FPS_LABEL_CHT NULL
#define BEETLE_OPT_CORE_TIMING_FPS_INFO_0_CHT NULL
#define OPTION_VAL_AUTO_TOGGLE_CHT NULL
#define OPTION_VAL_FORCE_PROGRESSIVE_CHT NULL
#define OPTION_VAL_FORCE_INTERLACED_CHT NULL
#define BEETLE_OPT_DEINTERLACER_LABEL_CHT NULL
#define BEETLE_OPT_DEINTERLACER_INFO_0_CHT NULL
#define OPTION_VAL_WEAVE_CHT NULL
#define OPTION_VAL_BOB_CHT NULL
#define OPTION_VAL_BOB_OFFSET_CHT NULL
#define OPTION_VAL_FASTMAD_CHT NULL
#define BEETLE_OPT_FRAME_DUPING_LABEL_CHT NULL
#define BEETLE_OPT_FRAME_DUPING_INFO_0_CHT NULL
#define BEETLE_OPT_PAL_VIDEO_TIMING_OVERRIDE_LABEL_CHT NULL
#define BEETLE_OPT_PAL_VIDEO_TIMING_OVERRIDE_INFO_0_CHT NULL
#define BEETLE_OPT_ASPECT_RATIO_LABEL_CHT NULL
#define BEETLE_OPT_ASPECT_RATIO_INFO_0_CHT NULL
#define OPTION_VAL_CORRECTED_CHT NULL
#define OPTION_VAL_UNCORRECTED_CHT "未調整"
#define OPTION_VAL_4_3_CHT NULL
#define OPTION_VAL_NTSC_CHT NULL
#define BEETLE_OPT_CROP_OVERSCAN_LABEL_CHT "裁剪過度掃描"
#define BEETLE_OPT_CROP_OVERSCAN_INFO_0_CHT NULL
#define OPTION_VAL_STATIC_CHT NULL
#define OPTION_VAL_SMART_CHT NULL
#define BEETLE_OPT_IMAGE_CROP_LABEL_CHT NULL
#define BEETLE_OPT_IMAGE_CROP_INFO_0_CHT NULL
#define OPTION_VAL_9PX_CHT "9像素"
#define OPTION_VAL_10PX_CHT "10像素"
#define OPTION_VAL_11PX_CHT "11像素"
#define OPTION_VAL_12PX_CHT "12像素"
#define OPTION_VAL_13PX_CHT "13像素"
#define OPTION_VAL_14PX_CHT "14像素"
#define OPTION_VAL_15PX_CHT "15像素"
#define OPTION_VAL_16PX_CHT "16像素"
#define OPTION_VAL_17PX_CHT "17像素"
#define OPTION_VAL_18PX_CHT "18像素"
#define OPTION_VAL_19PX_CHT "19像素"
#define OPTION_VAL_20PX_CHT "20像素"
#define BEETLE_OPT_IMAGE_OFFSET_LABEL_CHT NULL
#define BEETLE_OPT_IMAGE_OFFSET_INFO_0_CHT NULL
#define OPTION_VAL_12PX_O64_CHT NULL
#define OPTION_VAL_11PX_O64_CHT NULL
#define OPTION_VAL_10PX_O64_CHT NULL
#define OPTION_VAL_9PX_O64_CHT NULL
#define OPTION_VAL_8PX_O64_CHT NULL
#define OPTION_VAL_7PX_O64_CHT NULL
#define OPTION_VAL_6PX_O64_CHT NULL
#define OPTION_VAL_5PX_O64_CHT NULL
#define OPTION_VAL_4PX_O64_CHT NULL
#define OPTION_VAL_3PX_O64_CHT NULL
#define OPTION_VAL_2PX_O64_CHT NULL
#define OPTION_VAL_1PX_O64_CHT NULL
#define OPTION_VAL_1PX_O64_0_CHT NULL
#define OPTION_VAL_2PX_O64_0_CHT NULL
#define OPTION_VAL_3PX_O64_0_CHT NULL
#define OPTION_VAL_4PX_O64_0_CHT NULL
#define OPTION_VAL_5PX_O64_0_CHT NULL
#define OPTION_VAL_6PX_O64_0_CHT NULL
#define OPTION_VAL_7PX_O64_0_CHT NULL
#define OPTION_VAL_8PX_O64_0_CHT NULL
#define OPTION_VAL_9PX_O64_0_CHT NULL
#define OPTION_VAL_10PX_O64_0_CHT NULL
#define OPTION_VAL_11PX_O64_0_CHT NULL
#define OPTION_VAL_12PX_O64_0_CHT NULL
#define BEETLE_OPT_IMAGE_OFFSET_CYCLES_LABEL_CHT NULL
#define BEETLE_OPT_IMAGE_OFFSET_CYCLES_INFO_0_CHT NULL
#define BEETLE_OPT_INITIAL_SCANLINE_LABEL_CHT NULL
#define BEETLE_OPT_INITIAL_SCANLINE_INFO_0_CHT NULL
#define BEETLE_OPT_LAST_SCANLINE_LABEL_CHT NULL
#define BEETLE_OPT_LAST_SCANLINE_INFO_0_CHT NULL
#define OPTION_VAL_239_CHT "239 (預設)"
#define BEETLE_OPT_INITIAL_SCANLINE_PAL_LABEL_CHT NULL
#define BEETLE_OPT_INITIAL_SCANLINE_PAL_INFO_0_CHT NULL
#define BEETLE_OPT_LAST_SCANLINE_PAL_LABEL_CHT NULL
#define BEETLE_OPT_LAST_SCANLINE_PAL_INFO_0_CHT NULL
#define OPTION_VAL_287_CHT "287 (預設)"

struct retro_core_option_v2_category option_cats_cht[] = {
   {
      "system",
      CATEGORY_SYSTEM_LABEL_CHT,
      CATEGORY_SYSTEM_INFO_0_CHT
   },
   {
      "video",
      CATEGORY_VIDEO_LABEL_CHT,
      CATEGORY_VIDEO_INFO_0_CHT
   },
   {
      "pgxp",
      CATEGORY_PGXP_LABEL_CHT,
      CATEGORY_PGXP_INFO_0_CHT
   },
   {
      "osd",
      CATEGORY_OSD_LABEL_CHT,
      CATEGORY_OSD_INFO_0_CHT
   },
   {
      "input",
      CATEGORY_INPUT_LABEL_CHT,
      CATEGORY_INPUT_INFO_0_CHT
   },
   {
      "memcards",
      CATEGORY_MEMCARDS_LABEL_CHT,
      CATEGORY_MEMCARDS_INFO_0_CHT
   },
   {
      "hacks",
      CATEGORY_HACKS_LABEL_CHT,
      CATEGORY_HACKS_INFO_0_CHT
   },
   { NULL, NULL, NULL },
};
struct retro_core_option_v2_definition option_defs_cht[] = {
   {
      BEETLE_OPT(cpu_freq_scale),
      BEETLE_OPT_CPU_FREQ_SCALE_LABEL_CHT,
      NULL,
      BEETLE_OPT_CPU_FREQ_SCALE_INFO_0_CHT,
      NULL,
      "system",
      {
         { "50%",           OPTION_VAL_50_CHT },
         { "60%",           OPTION_VAL_60_CHT },
         { "70%",           OPTION_VAL_70_CHT },
         { "80%",           OPTION_VAL_80_CHT },
         { "90%",           OPTION_VAL_90_CHT },
         { "100%",          OPTION_VAL_100_CHT },
         { "110%",          OPTION_VAL_110_CHT },
         { "120%",          OPTION_VAL_120_CHT },
         { "130%",          OPTION_VAL_130_CHT },
         { "140%",          OPTION_VAL_140_CHT },
         { "150%",          OPTION_VAL_150_CHT },
         { "160%",          OPTION_VAL_160_CHT },
         { "170%",          OPTION_VAL_170_CHT },
         { "180%",          OPTION_VAL_180_CHT },
         { "190%",          OPTION_VAL_190_CHT },
         { "200%",          OPTION_VAL_200_CHT },
         { "210%",          OPTION_VAL_210_CHT },
         { "220%",          OPTION_VAL_220_CHT },
         { "230%",          OPTION_VAL_230_CHT },
         { "240%",          OPTION_VAL_240_CHT },
         { "250%",          OPTION_VAL_250_CHT },
         { "260%",          OPTION_VAL_260_CHT },
         { "270%",          OPTION_VAL_270_CHT },
         { "280%",          OPTION_VAL_280_CHT },
         { "290%",          OPTION_VAL_290_CHT },
         { "300%",          OPTION_VAL_300_CHT },
         { "310%",          OPTION_VAL_310_CHT },
         { "320%",          OPTION_VAL_320_CHT },
         { "330%",          OPTION_VAL_330_CHT },
         { "340%",          OPTION_VAL_340_CHT },
         { "350%",          OPTION_VAL_350_CHT },
         { "360%",          OPTION_VAL_360_CHT },
         { "370%",          OPTION_VAL_370_CHT },
         { "380%",          OPTION_VAL_380_CHT },
         { "390%",          OPTION_VAL_390_CHT },
         { "400%",          OPTION_VAL_400_CHT },
         { "410%",          OPTION_VAL_410_CHT },
         { "420%",          OPTION_VAL_420_CHT },
         { "430%",          OPTION_VAL_430_CHT },
         { "440%",          OPTION_VAL_440_CHT },
         { "450%",          OPTION_VAL_450_CHT },
         { "460%",          OPTION_VAL_460_CHT },
         { "470%",          OPTION_VAL_470_CHT },
         { "480%",          OPTION_VAL_480_CHT },
         { "490%",          OPTION_VAL_490_CHT },
         { "500%",          OPTION_VAL_500_CHT },
         { "510%",          OPTION_VAL_510_CHT },
         { "520%",          OPTION_VAL_520_CHT },
         { "530%",          OPTION_VAL_530_CHT },
         { "540%",          OPTION_VAL_540_CHT },
         { "550%",          OPTION_VAL_550_CHT },
         { "560%",          OPTION_VAL_560_CHT },
         { "570%",          OPTION_VAL_570_CHT },
         { "580%",          OPTION_VAL_580_CHT },
         { "590%",          OPTION_VAL_590_CHT },
         { "600%",          OPTION_VAL_600_CHT },
         { "610%",          OPTION_VAL_610_CHT },
         { "620%",          OPTION_VAL_620_CHT },
         { "630%",          OPTION_VAL_630_CHT },
         { "640%",          OPTION_VAL_640_CHT },
         { "650%",          OPTION_VAL_650_CHT },
         { "660%",          OPTION_VAL_660_CHT },
         { "670%",          OPTION_VAL_670_CHT },
         { "680%",          OPTION_VAL_680_CHT },
         { "690%",          OPTION_VAL_690_CHT },
         { "700%",          OPTION_VAL_700_CHT },
         { "710%",          OPTION_VAL_710_CHT },
         { "720%",          OPTION_VAL_720_CHT },
         { "730%",          OPTION_VAL_730_CHT },
         { "740%",          OPTION_VAL_740_CHT },
         { "750%",          OPTION_VAL_750_CHT },
         { NULL, NULL },
      },
      "100%"
   },
   {
      BEETLE_OPT(gpu_overclock),
      BEETLE_OPT_GPU_OVERCLOCK_LABEL_CHT,
      NULL,
      BEETLE_OPT_GPU_OVERCLOCK_INFO_0_CHT,
      NULL,
      "system",
      {
         { "1x(native)", OPTION_VAL_1X_NATIVE_CHT },
         { "2x",         OPTION_VAL_2X_CHT },
         { "4x",         OPTION_VAL_4X_CHT },
         { "8x",         OPTION_VAL_8X_CHT },
         { "16x",        OPTION_VAL_16X_CHT },
         { "32x",        OPTION_VAL_32X_CHT },
         { NULL, NULL },
      },
      "1x(native)"
   },
   {
      BEETLE_OPT(gte_overclock),
      BEETLE_OPT_GTE_OVERCLOCK_LABEL_CHT,
      NULL,
      BEETLE_OPT_GTE_OVERCLOCK_INFO_0_CHT,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
#ifndef EMSCRIPTEN
   {
      BEETLE_OPT(cd_access_method),
      BEETLE_OPT_CD_ACCESS_METHOD_LABEL_CHT,
      NULL,
      BEETLE_OPT_CD_ACCESS_METHOD_INFO_0_CHT,
      NULL,
      "system",
      {
         { "sync",     OPTION_VAL_SYNC_CHT },
         { "async",    OPTION_VAL_ASYNC_CHT },
         { "precache", OPTION_VAL_PRECACHE_CHT },
         { NULL, NULL },
      },
      "sync"
   },
#endif
   {
      BEETLE_OPT(cd_fastload),
      BEETLE_OPT_CD_FASTLOAD_LABEL_CHT,
      NULL,
      BEETLE_OPT_CD_FASTLOAD_INFO_0_CHT,
      NULL,
      "system",
      {
         { "2x(native)", OPTION_VAL_2X_NATIVE_CHT },
         { "4x",          OPTION_VAL_4X_CHT },
         { "6x",          OPTION_VAL_6X_CHT },
         { "8x",          OPTION_VAL_8X_CHT },
         { "10x",         OPTION_VAL_10X_CHT },
         { "12x",         OPTION_VAL_12X_CHT },
         { "14x",         OPTION_VAL_14X_CHT },
         { NULL, NULL },
      },
      "2x(native)"
   },
   {
      BEETLE_OPT(skip_bios),
      BEETLE_OPT_SKIP_BIOS_LABEL_CHT,
      NULL,
      BEETLE_OPT_SKIP_BIOS_INFO_0_CHT,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(override_bios),
      BEETLE_OPT_OVERRIDE_BIOS_LABEL_CHT,
      NULL,
      BEETLE_OPT_OVERRIDE_BIOS_INFO_0_CHT,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "psxonpsp", OPTION_VAL_PSXONPSP_CHT },
         { "ps1_rom",  OPTION_VAL_PS1_ROM_CHT },
         { "openbios", OPTION_VAL_OPENBIOS_CHT     },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(region),
      BEETLE_OPT_REGION_LABEL_CHT,
      NULL,
      BEETLE_OPT_REGION_INFO_0_CHT,
      NULL,
      "system",
      {
         { "auto",   OPTION_VAL_AUTO_CHT },
         { "ntsc-j", OPTION_VAL_NTSC_J_CHT },
         { "ntsc-u", OPTION_VAL_NTSC_U_CHT },
         { "pal",    OPTION_VAL_PAL_CHT },
         { NULL, NULL },
      },
      "auto"
   },
   {
      BEETLE_OPT(internal_resolution),
      BEETLE_OPT_INTERNAL_RESOLUTION_LABEL_CHT,
      NULL,
      BEETLE_OPT_INTERNAL_RESOLUTION_INFO_0_CHT,
      NULL,
      "video",
      {
         { "1x(native)", OPTION_VAL_1X_NATIVE_CHT },
         { "2x",         OPTION_VAL_2X_CHT },
         { "4x",         OPTION_VAL_4X_CHT },
         { "8x",         OPTION_VAL_8X_CHT },
         { "16x",        OPTION_VAL_16X_CHT },
         { NULL, NULL },
      },
      "1x(native)"
   },
#if defined(HAVE_OPENGL) || defined(HAVE_OPENGLES) || defined(HAVE_VULKAN)
   {
      BEETLE_OPT(renderer),
      BEETLE_OPT_RENDERER_LABEL_CHT,
      NULL,
      BEETLE_OPT_RENDERER_INFO_0_CHT,
      NULL,
      "video",
      {
         { "hardware",    OPTION_VAL_HARDWARE_CHT },
#if defined(HAVE_OPENGL) || defined(HAVE_OPENGLES)
         { "hardware_gl", OPTION_VAL_HARDWARE_GL_CHT },
#endif
#if defined(HAVE_VULKAN)
         { "hardware_vk", OPTION_VAL_HARDWARE_VK_CHT },
#endif
         { "software",    OPTION_VAL_SOFTWARE_CHT },
         { NULL, NULL },
      },
      "hardware"
   },
   {
      BEETLE_OPT(renderer_software_fb),
      BEETLE_OPT_RENDERER_SOFTWARE_FB_LABEL_CHT,
      NULL,
      BEETLE_OPT_RENDERER_SOFTWARE_FB_INFO_0_CHT,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled"
   },
#endif
#if defined(HAVE_OPENGL) || defined(HAVE_OPENGLES)
   {
      BEETLE_OPT(depth),
      BEETLE_OPT_DEPTH_LABEL_CHT,
      NULL,
      BEETLE_OPT_DEPTH_INFO_0_CHT,
      NULL,
      "video",
      {
         { "16bpp(native)", OPTION_VAL_16BPP_NATIVE_CHT },
         { "32bpp",         OPTION_VAL_32BPP_CHT },
         { NULL, NULL },
      },
      "16bpp(native)"
   },
   // Sort of, it's more like 15-bit high color and 24-bit true color for visible output. The alpha channel is used for mask bit. Vulkan renderer uses ABGR1555_555 for 31 bits internal? FMVs are always 24-bit on all renderers like original hardware (BGR888, no alpha)
#endif
   {
      BEETLE_OPT(dither_mode),
      BEETLE_OPT_DITHER_MODE_LABEL_CHT,
      NULL,
      BEETLE_OPT_DITHER_MODE_INFO_0_CHT,
      NULL,
      "video",
      {
         { "1x(native)",          OPTION_VAL_1X_NATIVE_CHT },
         { "internal resolution", OPTION_VAL_INTERNAL_RESOLUTION_CHT },
         { "disabled",            NULL },
         { NULL, NULL },
      },
      "1x(native)"
   },
#ifdef HAVE_VULKAN
   {
      BEETLE_OPT(scaled_uv_offset),
      BEETLE_OPT_SCALED_UV_OFFSET_LABEL_CHT,
      NULL,
      BEETLE_OPT_SCALED_UV_OFFSET_INFO_0_CHT,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled"
   },
#endif
#if defined(HAVE_OPENGL) || defined(HAVE_OPENGLES) || defined(HAVE_VULKAN)
   {
      BEETLE_OPT(filter),
      BEETLE_OPT_FILTER_LABEL_CHT,
      NULL,
      BEETLE_OPT_FILTER_INFO_0_CHT,
      NULL,
      "video",
      {
         { "nearest",  OPTION_VAL_NEAREST_CHT },
         { "SABR",     OPTION_VAL_SABR_CHT },
         { "xBR",      OPTION_VAL_XBR_CHT },
         { "bilinear", OPTION_VAL_BILINEAR_CHT },
         { "3-point",  OPTION_VAL_3_POINT_CHT },
         { "JINC2",    OPTION_VAL_JINC2_CHT },
         { NULL, NULL },
      },
      "nearest"
   },
#ifdef HAVE_VULKAN
   {
      BEETLE_OPT(filter_exclude_sprite),
      BEETLE_OPT_FILTER_EXCLUDE_SPRITE_LABEL_CHT,
      NULL,
      BEETLE_OPT_FILTER_EXCLUDE_SPRITE_INFO_0_CHT,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "opaque", OPTION_VAL_OPAQUE_CHT },
         { "all", OPTION_VAL_ALL_CHT },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(filter_exclude_2d_polygon),
      BEETLE_OPT_FILTER_EXCLUDE_2D_POLYGON_LABEL_CHT,
      NULL,
      BEETLE_OPT_FILTER_EXCLUDE_2D_POLYGON_INFO_0_CHT,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "opaque", OPTION_VAL_OPAQUE_CHT },
         { "all", OPTION_VAL_ALL_CHT },
         { NULL, NULL },
      },
      "disabled"
   },
#endif
#endif
#ifdef HAVE_VULKAN
   {
      BEETLE_OPT(adaptive_smoothing),
      BEETLE_OPT_ADAPTIVE_SMOOTHING_LABEL_CHT,
      NULL,
      BEETLE_OPT_ADAPTIVE_SMOOTHING_INFO_0_CHT,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(super_sampling),
      BEETLE_OPT_SUPER_SAMPLING_LABEL_CHT,
      NULL,
      BEETLE_OPT_SUPER_SAMPLING_INFO_0_CHT,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(msaa),
      BEETLE_OPT_MSAA_LABEL_CHT,
      NULL,
      BEETLE_OPT_MSAA_INFO_0_CHT,
      NULL,
      "video",
      {
         { "1x",  OPTION_VAL_1X_CHT },
         { "2x",  OPTION_VAL_2X_CHT },
         { "4x",  OPTION_VAL_4X_CHT },
         { "8x",  OPTION_VAL_8X_CHT },
         { "16x", OPTION_VAL_16X_CHT },
         { NULL, NULL },
      },
      "1x"
   },
   {
      BEETLE_OPT(mdec_yuv),
      BEETLE_OPT_MDEC_YUV_LABEL_CHT,
      NULL,
      BEETLE_OPT_MDEC_YUV_INFO_0_CHT,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(track_textures),
      BEETLE_OPT_TRACK_TEXTURES_LABEL_CHT,
      NULL,
      BEETLE_OPT_TRACK_TEXTURES_INFO_0_CHT,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
#ifdef TEXTURE_DUMPING_ENABLED
   {
      BEETLE_OPT(dump_textures),
      BEETLE_OPT_DUMP_TEXTURES_LABEL_CHT,
      NULL,
      BEETLE_OPT_DUMP_TEXTURES_INFO_0_CHT,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
#endif
   {
      BEETLE_OPT(replace_textures),
      BEETLE_OPT_REPLACE_TEXTURES_LABEL_CHT,
      NULL,
      BEETLE_OPT_REPLACE_TEXTURES_INFO_0_CHT,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
#endif
   {
      BEETLE_OPT(pgxp_mode),
      BEETLE_OPT_PGXP_MODE_LABEL_CHT,
      NULL,
      BEETLE_OPT_PGXP_MODE_INFO_0_CHT,
      NULL,
      "pgxp",
      {
         { "disabled",     NULL },
         { "memory only",  OPTION_VAL_MEMORY_ONLY_CHT },
         { "memory + CPU", OPTION_VAL_MEMORY_CPU_CHT },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(pgxp_2d_tol),
      BEETLE_OPT_PGXP_2D_TOL_LABEL_CHT,
      NULL,
      BEETLE_OPT_PGXP_2D_TOL_INFO_0_CHT,
      NULL,
      "pgxp",
      {
         { "disabled", NULL },
         { "0px", OPTION_VAL_0PX_CHT },
         { "1px", OPTION_VAL_1PX_CHT },
         { "2px", OPTION_VAL_2PX_CHT },
         { "3px", OPTION_VAL_3PX_CHT },
         { "4px", OPTION_VAL_4PX_CHT },
         { "5px", OPTION_VAL_5PX_CHT },
         { "6px", OPTION_VAL_6PX_CHT },
         { "7px", OPTION_VAL_7PX_CHT },
         { "8px", OPTION_VAL_8PX_CHT },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(pgxp_nclip),
      BEETLE_OPT_PGXP_NCLIP_LABEL_CHT,
      NULL,
      BEETLE_OPT_PGXP_NCLIP_INFO_0_CHT,
      NULL,
      "pgxp",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
#if defined(HAVE_OPENGL) || defined(HAVE_OPENGLES) || defined(HAVE_VULKAN)
   {
      BEETLE_OPT(pgxp_vertex),
      BEETLE_OPT_PGXP_VERTEX_LABEL_CHT,
      NULL,
      BEETLE_OPT_PGXP_VERTEX_INFO_0_CHT,
      NULL,
      "pgxp",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
#endif
   {
      BEETLE_OPT(pgxp_texture),
      BEETLE_OPT_PGXP_TEXTURE_LABEL_CHT,
      NULL,
      BEETLE_OPT_PGXP_TEXTURE_INFO_0_CHT,
      NULL,
      "pgxp",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(display_internal_fps),
      BEETLE_OPT_DISPLAY_INTERNAL_FPS_LABEL_CHT,
      NULL,
      BEETLE_OPT_DISPLAY_INTERNAL_FPS_INFO_0_CHT,
      NULL,
      "osd",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(display_osd),
      BEETLE_OPT_DISPLAY_OSD_LABEL_CHT,
      NULL,
      BEETLE_OPT_DISPLAY_OSD_INFO_0_CHT,
      NULL,
      "osd",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled"
   },
#if defined(HAVE_OPENGL) || defined(HAVE_OPENGLES) || defined(HAVE_VULKAN)
   {
      BEETLE_OPT(display_vram),
      BEETLE_OPT_DISPLAY_VRAM_LABEL_CHT,
      NULL,
      BEETLE_OPT_DISPLAY_VRAM_INFO_0_CHT,
      NULL,
      "osd",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
#endif
   {
      BEETLE_OPT(analog_calibration),
      BEETLE_OPT_ANALOG_CALIBRATION_LABEL_CHT,
      NULL,
      BEETLE_OPT_ANALOG_CALIBRATION_INFO_0_CHT,
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(analog_toggle),
      BEETLE_OPT_ANALOG_TOGGLE_LABEL_CHT,
      NULL,
      BEETLE_OPT_ANALOG_TOGGLE_INFO_0_CHT,
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { "enabled-analog", OPTION_VAL_ENABLED_ANALOG_CHT },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(analog_toggle_combo),
      BEETLE_OPT_ANALOG_TOGGLE_COMBO_LABEL_CHT,
      NULL,
      BEETLE_OPT_ANALOG_TOGGLE_COMBO_INFO_0_CHT,
      NULL,
      "input",
      {
         { "l1+l2+r1+r2+start+select", OPTION_VAL_L1_L2_R1_R2_START_SELECT_CHT },
         { "l1+r1+select",             OPTION_VAL_L1_R1_SELECT_CHT },
         { "l1+r1+start",              OPTION_VAL_L1_R1_START_CHT },
         { "l1+r1+l3",                 OPTION_VAL_L1_R1_L3_CHT },
         { "l1+r1+r3",                 OPTION_VAL_L1_R1_R3_CHT },
         { "l2+r2+select",             OPTION_VAL_L2_R2_SELECT_CHT },
         { "l2+r2+start",              OPTION_VAL_L2_R2_START_CHT },
         { "l2+r2+l3",                 OPTION_VAL_L2_R2_L3_CHT },
         { "l2+r2+r3",                 OPTION_VAL_L2_R2_R3_CHT },
         { "l3+r3",                    OPTION_VAL_L3_R3_CHT },
         { NULL, NULL },
      },
      "l1+r1+select"
   },
   {
      BEETLE_OPT(analog_toggle_hold),
      BEETLE_OPT_ANALOG_TOGGLE_HOLD_LABEL_CHT,
      NULL,
      BEETLE_OPT_ANALOG_TOGGLE_HOLD_INFO_0_CHT,
      NULL,
      "input",
      {
         { "0", OPTION_VAL_0_CHT },
         { "1", OPTION_VAL_1_CHT },
         { "2", OPTION_VAL_2_CHT },
         { "3", OPTION_VAL_3_CHT },
         { "4", OPTION_VAL_4_CHT },
         { "5", OPTION_VAL_5_CHT },
         { NULL, NULL },
      },
      "1"
   },
   {
      BEETLE_OPT(enable_multitap_port1),
      BEETLE_OPT_ENABLE_MULTITAP_PORT1_LABEL_CHT,
      NULL,
      BEETLE_OPT_ENABLE_MULTITAP_PORT1_INFO_0_CHT,
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(enable_multitap_port2),
      BEETLE_OPT_ENABLE_MULTITAP_PORT2_LABEL_CHT,
      NULL,
      BEETLE_OPT_ENABLE_MULTITAP_PORT2_INFO_0_CHT,
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(gun_input_mode),
      BEETLE_OPT_GUN_INPUT_MODE_LABEL_CHT,
      NULL,
      BEETLE_OPT_GUN_INPUT_MODE_INFO_0_CHT,
      NULL,
      "input",
      {
         { "lightgun",    OPTION_VAL_LIGHTGUN_CHT },
         { "touchscreen", OPTION_VAL_TOUCHSCREEN_CHT },
         { NULL, NULL },
      },
      "lightgun"
   },
   // Shouldn't the gun_input_mode just be Mouse vs. Touchscreen?
   {
      BEETLE_OPT(gun_cursor),
      BEETLE_OPT_GUN_CURSOR_LABEL_CHT,
      NULL,
      BEETLE_OPT_GUN_CURSOR_INFO_0_CHT,
      NULL,
      "input",
      {
         { "off",   "disabled" },
         { "cross", OPTION_VAL_CROSS_CHT },
         { "dot",   OPTION_VAL_DOT_CHT },
         { NULL, NULL },
      },
      "cross"
   },
   {
      BEETLE_OPT(crosshair_color_p1),
      BEETLE_OPT_CROSSHAIR_COLOR_P1_LABEL_CHT,
      NULL,
      BEETLE_OPT_CROSSHAIR_COLOR_P1_INFO_0_CHT,
      NULL,
      "input",
      {
         { "red",    OPTION_VAL_RED_CHT },
         { "blue",   OPTION_VAL_BLUE_CHT },
         { "green",  OPTION_VAL_GREEN_CHT },
         { "orange", OPTION_VAL_ORANGE_CHT },
         { "yellow", OPTION_VAL_YELLOW_CHT },
         { "cyan",   OPTION_VAL_CYAN_CHT },
         { "pink",   OPTION_VAL_PINK_CHT },
         { "purple", OPTION_VAL_PURPLE_CHT },
         { "black",  OPTION_VAL_BLACK_CHT },
         { "white",  OPTION_VAL_WHITE_CHT },
         { NULL, NULL },
      },
      "red"
   },
   {
      BEETLE_OPT(crosshair_color_p2),
      BEETLE_OPT_CROSSHAIR_COLOR_P2_LABEL_CHT,
      NULL,
      BEETLE_OPT_CROSSHAIR_COLOR_P2_INFO_0_CHT,
      NULL,
      "input",
      {
         { "blue",   OPTION_VAL_BLUE_CHT },
         { "red",    OPTION_VAL_RED_CHT },
         { "green",  OPTION_VAL_GREEN_CHT },
         { "orange", OPTION_VAL_ORANGE_CHT },
         { "yellow", OPTION_VAL_YELLOW_CHT },
         { "cyan",   OPTION_VAL_CYAN_CHT },
         { "pink",   OPTION_VAL_PINK_CHT },
         { "purple", OPTION_VAL_PURPLE_CHT },
         { "black",  OPTION_VAL_BLACK_CHT },
         { "white",  OPTION_VAL_WHITE_CHT },
         { NULL, NULL },
      },
      "blue"
   },
   {
      BEETLE_OPT(mouse_sensitivity),
      BEETLE_OPT_MOUSE_SENSITIVITY_LABEL_CHT,
      NULL,
      BEETLE_OPT_MOUSE_SENSITIVITY_INFO_0_CHT,
      NULL,
      "input",
      {
         { "5%",   OPTION_VAL_5_O42_CHT },
         { "10%",  OPTION_VAL_10_CHT },
         { "15%",  OPTION_VAL_15_CHT },
         { "20%",  OPTION_VAL_20_CHT },
         { "25%",  OPTION_VAL_25_CHT },
         { "30%",  OPTION_VAL_30_CHT },
         { "35%",  OPTION_VAL_35_CHT },
         { "40%",  OPTION_VAL_40_CHT },
         { "45%",  OPTION_VAL_45_CHT },
         { "50%",  OPTION_VAL_50_CHT },
         { "55%",  OPTION_VAL_55_CHT },
         { "60%",  OPTION_VAL_60_CHT },
         { "65%",  OPTION_VAL_65_CHT },
         { "70%",  OPTION_VAL_70_CHT },
         { "75%",  OPTION_VAL_75_CHT },
         { "80%",  OPTION_VAL_80_CHT },
         { "85%",  OPTION_VAL_85_CHT },
         { "90%",  OPTION_VAL_90_CHT },
         { "95%",  OPTION_VAL_95_CHT },
         { "100%", OPTION_VAL_100_O42_CHT },
         { "105%", OPTION_VAL_105_CHT },
         { "110%", OPTION_VAL_110_CHT },
         { "115%", OPTION_VAL_115_CHT },
         { "120%", OPTION_VAL_120_CHT },
         { "125%", OPTION_VAL_125_CHT },
         { "130%", OPTION_VAL_130_CHT },
         { "135%", OPTION_VAL_135_CHT },
         { "140%", OPTION_VAL_140_CHT },
         { "145%", OPTION_VAL_145_CHT },
         { "150%", OPTION_VAL_150_CHT },
         { "155%", OPTION_VAL_155_CHT },
         { "160%", OPTION_VAL_160_CHT },
         { "165%", OPTION_VAL_165_CHT },
         { "170%", OPTION_VAL_170_CHT },
         { "175%", OPTION_VAL_175_CHT },
         { "180%", OPTION_VAL_180_CHT },
         { "185%", OPTION_VAL_185_CHT },
         { "190%", OPTION_VAL_190_CHT },
         { "195%", OPTION_VAL_195_CHT },
         { "200%", OPTION_VAL_200_CHT },
         { NULL, NULL },
      },
      "100%"
   },
   {
      BEETLE_OPT(negcon_response),
      BEETLE_OPT_NEGCON_RESPONSE_LABEL_CHT,
      NULL,
      BEETLE_OPT_NEGCON_RESPONSE_INFO_0_CHT,
      NULL,
      "input",
      {
         { "linear",    OPTION_VAL_LINEAR_CHT },
         { "quadratic", OPTION_VAL_QUADRATIC_CHT },
         { "cubic",     OPTION_VAL_CUBIC_CHT },
         { NULL, NULL },
      },
      "linear"
   },
   {
      BEETLE_OPT(negcon_deadzone),
      BEETLE_OPT_NEGCON_DEADZONE_LABEL_CHT,
      NULL,
      BEETLE_OPT_NEGCON_DEADZONE_INFO_0_CHT,
      NULL,
      "input",
      {
         { "0%",  OPTION_VAL_0_O44_CHT },
         { "5%",  OPTION_VAL_5_O42_CHT },
         { "10%", OPTION_VAL_10_CHT },
         { "15%", OPTION_VAL_15_CHT },
         { "20%", OPTION_VAL_20_CHT },
         { "25%", OPTION_VAL_25_CHT },
         { "30%", OPTION_VAL_30_CHT },
         { NULL, NULL },
      },
      "0%"
   },
   {
      BEETLE_OPT(use_mednafen_memcard0_method),
      BEETLE_OPT_USE_MEDNAFEN_MEMCARD0_METHOD_LABEL_CHT,
      NULL,
      BEETLE_OPT_USE_MEDNAFEN_MEMCARD0_METHOD_INFO_0_CHT,
      NULL,
      "memcards",
      {
         { "libretro", OPTION_VAL_LIBRETRO_CHT },
         { "mednafen", OPTION_VAL_MEDNAFEN_CHT },
         { NULL, NULL },
      },
      "libretro"
   },
   {
      BEETLE_OPT(enable_memcard1),
      BEETLE_OPT_ENABLE_MEMCARD1_LABEL_CHT,
      NULL,
      BEETLE_OPT_ENABLE_MEMCARD1_INFO_0_CHT,
      NULL,
      "memcards",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(shared_memory_cards),
      BEETLE_OPT_SHARED_MEMORY_CARDS_LABEL_CHT,
      NULL,
      BEETLE_OPT_SHARED_MEMORY_CARDS_INFO_0_CHT,
      NULL,
      "memcards",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(memcard_left_index),
      BEETLE_OPT_MEMCARD_LEFT_INDEX_LABEL_CHT,
      NULL,
      BEETLE_OPT_MEMCARD_LEFT_INDEX_INFO_0_CHT,
      NULL,
      "memcards",
      {
         { "0",  OPTION_VAL_0_O48_CHT },
         { "1",  NULL },
         { "2",  NULL },
         { "3",  NULL },
         { "4",  NULL },
         { "5",  NULL },
         { "6",  NULL },
         { "7",  NULL },
         { "8",  NULL },
         { "9",  NULL },
         { "10", NULL },
         { "11", NULL },
         { "12", NULL },
         { "13", NULL },
         { "14", NULL },
         { "15", NULL },
         { "16", NULL },
         { "17", NULL },
         { "18", NULL },
         { "19", NULL },
         { "20", NULL },
         { "21", NULL },
         { "22", NULL },
         { "23", NULL },
         { "24", NULL },
         { "25", NULL },
         { "26", NULL },
         { "27", NULL },
         { "28", NULL },
         { "29", NULL },
         { "30", NULL },
         { "31", NULL },
         { "32", NULL },
         { "33", NULL },
         { "34", NULL },
         { "35", NULL },
         { "36", NULL },
         { "37", NULL },
         { "38", NULL },
         { "39", NULL },
         { "40", NULL },
         { "41", NULL },
         { "42", NULL },
         { "43", NULL },
         { "44", NULL },
         { "45", NULL },
         { "46", NULL },
         { "47", NULL },
         { "48", NULL },
         { "49", NULL },
         { "50", NULL },
         { "51", NULL },
         { "52", NULL },
         { "53", NULL },
         { "54", NULL },
         { "55", NULL },
         { "56", NULL },
         { "57", NULL },
         { "58", NULL },
         { "59", NULL },
         { "60", NULL },
         { "61", NULL },
         { "62", NULL },
         { "63", NULL },
         { NULL, NULL },
      },
      "0"
   },
   {
      BEETLE_OPT(memcard_right_index),
      BEETLE_OPT_MEMCARD_RIGHT_INDEX_LABEL_CHT,
      NULL,
      BEETLE_OPT_MEMCARD_RIGHT_INDEX_INFO_0_CHT,
      NULL,
      "memcards",
      {
         { "0",  NULL },
         { "1",  OPTION_VAL_1_O49_CHT },
         { "2",  NULL },
         { "3",  NULL },
         { "4",  NULL },
         { "5",  NULL },
         { "6",  NULL },
         { "7",  NULL },
         { "8",  NULL },
         { "9",  NULL },
         { "10", NULL },
         { "11", NULL },
         { "12", NULL },
         { "13", NULL },
         { "14", NULL },
         { "15", NULL },
         { "16", NULL },
         { "17", NULL },
         { "18", NULL },
         { "19", NULL },
         { "20", NULL },
         { "21", NULL },
         { "22", NULL },
         { "23", NULL },
         { "24", NULL },
         { "25", NULL },
         { "26", NULL },
         { "27", NULL },
         { "28", NULL },
         { "29", NULL },
         { "30", NULL },
         { "31", NULL },
         { "32", NULL },
         { "33", NULL },
         { "34", NULL },
         { "35", NULL },
         { "36", NULL },
         { "37", NULL },
         { "38", NULL },
         { "39", NULL },
         { "40", NULL },
         { "41", NULL },
         { "42", NULL },
         { "43", NULL },
         { "44", NULL },
         { "45", NULL },
         { "46", NULL },
         { "47", NULL },
         { "48", NULL },
         { "49", NULL },
         { "50", NULL },
         { "51", NULL },
         { "52", NULL },
         { "53", NULL },
         { "54", NULL },
         { "55", NULL },
         { "56", NULL },
         { "57", NULL },
         { "58", NULL },
         { "59", NULL },
         { "60", NULL },
         { "61", NULL },
         { "62", NULL },
         { "63", NULL },
         { NULL, NULL },
      },
      "1"
   },
   {
      BEETLE_OPT(line_render),
      BEETLE_OPT_LINE_RENDER_LABEL_CHT,
      NULL,
      BEETLE_OPT_LINE_RENDER_INFO_0_CHT,
      NULL,
      "hacks",
      {
         { "default",    OPTION_VAL_DEFAULT_CHT },
         { "aggressive", OPTION_VAL_AGGRESSIVE_CHT },
         { "disabled",   NULL },
         { NULL, NULL },
      },
      "default"
   },
   {
      BEETLE_OPT(widescreen_hack),
      BEETLE_OPT_WIDESCREEN_HACK_LABEL_CHT,
      NULL,
      BEETLE_OPT_WIDESCREEN_HACK_INFO_0_CHT,
      NULL,
      "hacks",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(widescreen_hack_aspect_ratio),
      BEETLE_OPT_WIDESCREEN_HACK_ASPECT_RATIO_LABEL_CHT,
      NULL,
      BEETLE_OPT_WIDESCREEN_HACK_ASPECT_RATIO_INFO_0_CHT,
      NULL,
      "hacks",
      {
         { "16:9",  OPTION_VAL_16_9_CHT },
         { "16:10", OPTION_VAL_16_10_CHT },
         { "18:9",  OPTION_VAL_18_9_CHT },
         { "19:9",  OPTION_VAL_19_9_CHT },
         { "20:9",  OPTION_VAL_20_9_CHT },
         { "21:9",  OPTION_VAL_21_9_CHT }, // 64:27
         { "32:9",  OPTION_VAL_32_9_CHT },
         { NULL,    NULL },
      },
      "16:9"
   },
#if defined(HAVE_LIGHTREC)
   {
      BEETLE_OPT(cpu_dynarec),
      BEETLE_OPT_CPU_DYNAREC_LABEL_CHT,
      NULL,
      BEETLE_OPT_CPU_DYNAREC_INFO_0_CHT,
      NULL,
      "hacks",
      {
         { "disabled",        OPTION_VAL_DISABLED_CHT },
         { "execute",         OPTION_VAL_EXECUTE_CHT },
         { "execute_one",     OPTION_VAL_EXECUTE_ONE_CHT },
         { "run_interpreter", OPTION_VAL_RUN_INTERPRETER_CHT },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(dynarec_invalidate),
      BEETLE_OPT_DYNAREC_INVALIDATE_LABEL_CHT,
      NULL,
      BEETLE_OPT_DYNAREC_INVALIDATE_INFO_0_CHT,
      NULL,
      "hacks",
      {
         { "full", OPTION_VAL_FULL_CHT },
         { "dma",  OPTION_VAL_DMA_CHT },
         { NULL, NULL },
      },
      "full"
   },
   {
      BEETLE_OPT(dynarec_eventcycles),
      BEETLE_OPT_DYNAREC_EVENTCYCLES_LABEL_CHT,
      NULL,
      BEETLE_OPT_DYNAREC_EVENTCYCLES_INFO_0_CHT,
      NULL,
      "hacks",
      {
         { "128",  OPTION_VAL_128_CHT },
         { "256",  NULL },
         { "384",  NULL },
         { "512",  NULL },
         { "640",  NULL },
         { "768",  NULL },
         { "896",  NULL },
         { "1024", NULL },
         { "1152", NULL },
         { "1280", NULL },
         { "1408", NULL },
         { "1536", NULL },
         { "1664", NULL },
         { "1792", NULL },
         { "1920", NULL },
         { "2048", NULL },
         { NULL, NULL },
      },
      "128"
   },
   {
      BEETLE_OPT(dynarec_spu_samples),
      BEETLE_OPT_DYNAREC_SPU_SAMPLES_LABEL_CHT,
      NULL,
      BEETLE_OPT_DYNAREC_SPU_SAMPLES_INFO_0_CHT,
      NULL,
      "hacks",
      {
         { "1",  OPTION_VAL_1_O49_CHT },
         { "4",  NULL },
         { "16", NULL },
         { NULL, NULL },
      },
      "1"
   },
#endif
   {
      BEETLE_OPT(core_timing_fps),
      BEETLE_OPT_CORE_TIMING_FPS_LABEL_CHT,
      NULL,
      BEETLE_OPT_CORE_TIMING_FPS_INFO_0_CHT,
      NULL,
      "video",
      {
         { "auto_toggle",       OPTION_VAL_AUTO_TOGGLE_CHT },
         { "force_progressive", OPTION_VAL_FORCE_PROGRESSIVE_CHT },
         { "force_interlaced",  OPTION_VAL_FORCE_INTERLACED_CHT },
      },
      "auto_toggle"
   },
   {
      BEETLE_OPT(deinterlacer),
      BEETLE_OPT_DEINTERLACER_LABEL_CHT,
      NULL,
      BEETLE_OPT_DEINTERLACER_INFO_0_CHT,
      NULL,
      "video",
      {
         { "weave",      OPTION_VAL_WEAVE_CHT },
         { "bob",        OPTION_VAL_BOB_CHT },
         { "bob_offset", OPTION_VAL_BOB_OFFSET_CHT },
         { "fastmad",    OPTION_VAL_FASTMAD_CHT },
         { "off",        "Off" },
         { NULL, NULL },
      },
      "weave"
   },
   {
      BEETLE_OPT(frame_duping),
      BEETLE_OPT_FRAME_DUPING_LABEL_CHT,
      NULL,
      BEETLE_OPT_FRAME_DUPING_INFO_0_CHT,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(pal_video_timing_override),
      BEETLE_OPT_PAL_VIDEO_TIMING_OVERRIDE_LABEL_CHT,
      NULL,
      BEETLE_OPT_PAL_VIDEO_TIMING_OVERRIDE_INFO_0_CHT,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(aspect_ratio),
      BEETLE_OPT_ASPECT_RATIO_LABEL_CHT,
      NULL,
      BEETLE_OPT_ASPECT_RATIO_INFO_0_CHT,
      NULL,
      "video",
      {
         { "corrected", OPTION_VAL_CORRECTED_CHT },
         { "uncorrected", OPTION_VAL_UNCORRECTED_CHT },
         { "4:3",  OPTION_VAL_4_3_CHT },
         { "ntsc", OPTION_VAL_NTSC_CHT },
      },
      "corrected"
   },
   {
      BEETLE_OPT(crop_overscan),
      BEETLE_OPT_CROP_OVERSCAN_LABEL_CHT,
      NULL,
      BEETLE_OPT_CROP_OVERSCAN_INFO_0_CHT,
      NULL,
      "video",
      {
         { "disabled", "Disabled" },
         { "static",   OPTION_VAL_STATIC_CHT },
         { "smart",    OPTION_VAL_SMART_CHT },
         { NULL, NULL },
      },
      "smart"
   },
   {
      BEETLE_OPT(image_crop),
      BEETLE_OPT_IMAGE_CROP_LABEL_CHT,
      NULL,
      BEETLE_OPT_IMAGE_CROP_INFO_0_CHT,
      NULL,
      "video",
      {
         { "disabled", "0" },
         { "1px",      OPTION_VAL_1PX_CHT },
         { "2px",      OPTION_VAL_2PX_CHT },
         { "3px",      OPTION_VAL_3PX_CHT },
         { "4px",      OPTION_VAL_4PX_CHT },
         { "5px",      OPTION_VAL_5PX_CHT },
         { "6px",      OPTION_VAL_6PX_CHT },
         { "7px",      OPTION_VAL_7PX_CHT },
         { "8px",      OPTION_VAL_8PX_CHT },
         { "9px",      OPTION_VAL_9PX_CHT },
         { "10px",     OPTION_VAL_10PX_CHT },
         { "11px",     OPTION_VAL_11PX_CHT },
         { "12px",     OPTION_VAL_12PX_CHT },
         { "13px",     OPTION_VAL_13PX_CHT },
         { "14px",     OPTION_VAL_14PX_CHT },
         { "15px",     OPTION_VAL_15PX_CHT },
         { "16px",     OPTION_VAL_16PX_CHT },
         { "17px",     OPTION_VAL_17PX_CHT },
         { "18px",     OPTION_VAL_18PX_CHT },
         { "19px",     OPTION_VAL_19PX_CHT },
         { "20px",     OPTION_VAL_20PX_CHT },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(image_offset),
      BEETLE_OPT_IMAGE_OFFSET_LABEL_CHT,
      NULL,
      BEETLE_OPT_IMAGE_OFFSET_INFO_0_CHT,
      NULL,
      "video",
      {
         { "-12px",    OPTION_VAL_12PX_O64_CHT },
         { "-11px",    OPTION_VAL_11PX_O64_CHT },
         { "-10px",    OPTION_VAL_10PX_O64_CHT },
         { "-9px",     OPTION_VAL_9PX_O64_CHT },
         { "-8px",     OPTION_VAL_8PX_O64_CHT },
         { "-7px",     OPTION_VAL_7PX_O64_CHT },
         { "-6px",     OPTION_VAL_6PX_O64_CHT },
         { "-5px",     OPTION_VAL_5PX_O64_CHT },
         { "-4px",     OPTION_VAL_4PX_O64_CHT },
         { "-3px",     OPTION_VAL_3PX_O64_CHT },
         { "-2px",     OPTION_VAL_2PX_O64_CHT },
         { "-1px",     OPTION_VAL_1PX_O64_CHT },
         { "disabled", "0" },
         { "+1px",     OPTION_VAL_1PX_O64_0_CHT },
         { "+2px",     OPTION_VAL_2PX_O64_0_CHT },
         { "+3px",     OPTION_VAL_3PX_O64_0_CHT },
         { "+4px",     OPTION_VAL_4PX_O64_0_CHT },
         { "+5px",     OPTION_VAL_5PX_O64_0_CHT },
         { "+6px",     OPTION_VAL_6PX_O64_0_CHT },
         { "+7px",     OPTION_VAL_7PX_O64_0_CHT },
         { "+8px",     OPTION_VAL_8PX_O64_0_CHT },
         { "+9px",     OPTION_VAL_9PX_O64_0_CHT },
         { "+10px",    OPTION_VAL_10PX_O64_0_CHT },
         { "+11px",    OPTION_VAL_11PX_O64_0_CHT },
         { "+12px",    OPTION_VAL_12PX_O64_0_CHT },
         { NULL, NULL },
      },
      "disabled"
   },
#if defined(HAVE_OPENGL) || defined(HAVE_OPENGLES) || defined(HAVE_VULKAN)
   {
      BEETLE_OPT(image_offset_cycles),
      BEETLE_OPT_IMAGE_OFFSET_CYCLES_LABEL_CHT,
      NULL,
      BEETLE_OPT_IMAGE_OFFSET_CYCLES_INFO_0_CHT,
      NULL,
      "video",
      {
         { "-40",      NULL },
         { "-39",      NULL },
         { "-38",      NULL },
         { "-37",      NULL },
         { "-36",      NULL },
         { "-35",      NULL },
         { "-34",      NULL },
         { "-33",      NULL },
         { "-32",      NULL },
         { "-31",      NULL },
         { "-30",      NULL },
         { "-29",      NULL },
         { "-28",      NULL },
         { "-27",      NULL },
         { "-26",      NULL },
         { "-25",      NULL },
         { "-24",      NULL },
         { "-23",      NULL },
         { "-22",      NULL },
         { "-21",      NULL },
         { "-20",      NULL },
         { "-19",      NULL },
         { "-18",      NULL },
         { "-17",      NULL },
         { "-16",      NULL },
         { "-15",      NULL },
         { "-14",      NULL },
         { "-13",      NULL },
         { "-12",      NULL },
         { "-11",      NULL },
         { "-10",      NULL },
         { "-9",       NULL },
         { "-8",       NULL },
         { "-7",       NULL },
         { "-6",       NULL },
         { "-5",       NULL },
         { "-4",       NULL },
         { "-3",       NULL },
         { "-2",       NULL },
         { "-1",       NULL },
         { "0",        "0" },
         { "+1",       NULL },
         { "+2",       NULL },
         { "+3",       NULL },
         { "+4",       NULL },
         { "+5",       NULL },
         { "+6",       NULL },
         { "+7",       NULL },
         { "+8",       NULL },
         { "+9",       NULL },
         { "+10",      NULL },
         { "+11",      NULL },
         { "+12",      NULL },
         { "+13",      NULL },
         { "+14",      NULL },
         { "+15",      NULL },
         { "+16",      NULL },
         { "+17",      NULL },
         { "+18",      NULL },
         { "+19",      NULL },
         { "+20",      NULL },
         { "+21",      NULL },
         { "+22",      NULL },
         { "+23",      NULL },
         { "+24",      NULL },
         { "+25",      NULL },
         { "+26",      NULL },
         { "+27",      NULL },
         { "+28",      NULL },
         { "+29",      NULL },
         { "+30",      NULL },
         { "+31",      NULL },
         { "+32",      NULL },
         { "+33",      NULL },
         { "+34",      NULL },
         { "+35",      NULL },
         { "+36",      NULL },
         { "+37",      NULL },
         { "+38",      NULL },
         { "+39",      NULL },
         { "+40",      NULL },
         { NULL, NULL},
      },
      "0"
   },
#endif
   {
      BEETLE_OPT(initial_scanline),
      BEETLE_OPT_INITIAL_SCANLINE_LABEL_CHT,
      NULL,
      BEETLE_OPT_INITIAL_SCANLINE_INFO_0_CHT,
      NULL,
      "video",
      {
         { "0",  OPTION_VAL_0_O48_CHT },
         { "1",  NULL },
         { "2",  NULL },
         { "3",  NULL },
         { "4",  NULL },
         { "5",  NULL },
         { "6",  NULL },
         { "7",  NULL },
         { "8",  NULL },
         { "9",  NULL },
         { "10", NULL },
         { "11", NULL },
         { "12", NULL },
         { "13", NULL },
         { "14", NULL },
         { "15", NULL },
         { "16", NULL },
         { "17", NULL },
         { "18", NULL },
         { "19", NULL },
         { "20", NULL },
         { "21", NULL },
         { "22", NULL },
         { "23", NULL },
         { "24", NULL },
         { "25", NULL },
         { "26", NULL },
         { "27", NULL },
         { "28", NULL },
         { "29", NULL },
         { "30", NULL },
         { "31", NULL },
         { "32", NULL },
         { "33", NULL },
         { "34", NULL },
         { "35", NULL },
         { "36", NULL },
         { "37", NULL },
         { "38", NULL },
         { "39", NULL },
         { "40", NULL },
         { NULL, NULL },
      },
      "0"
   },
   {
      BEETLE_OPT(last_scanline),
      BEETLE_OPT_LAST_SCANLINE_LABEL_CHT,
      NULL,
      BEETLE_OPT_LAST_SCANLINE_INFO_0_CHT,
      NULL,
      "video",
      {
         { "210", NULL },
         { "211", NULL },
         { "212", NULL },
         { "213", NULL },
         { "214", NULL },
         { "215", NULL },
         { "216", NULL },
         { "217", NULL },
         { "218", NULL },
         { "219", NULL },
         { "220", NULL },
         { "221", NULL },
         { "222", NULL },
         { "223", NULL },
         { "224", NULL },
         { "225", NULL },
         { "226", NULL },
         { "227", NULL },
         { "228", NULL },
         { "229", NULL },
         { "230", NULL },
         { "231", NULL },
         { "232", NULL },
         { "233", NULL },
         { "234", NULL },
         { "235", NULL },
         { "236", NULL },
         { "237", NULL },
         { "238", NULL },
         { "239", OPTION_VAL_239_CHT },
         { NULL, NULL },
      },
      "239"
   },
   {
      BEETLE_OPT(initial_scanline_pal),
      BEETLE_OPT_INITIAL_SCANLINE_PAL_LABEL_CHT,
      NULL,
      BEETLE_OPT_INITIAL_SCANLINE_PAL_INFO_0_CHT,
      NULL,
      "video",
      {
         { "0",  OPTION_VAL_0_O48_CHT },
         { "1",  NULL },
         { "2",  NULL },
         { "3",  NULL },
         { "4",  NULL },
         { "5",  NULL },
         { "6",  NULL },
         { "7",  NULL },
         { "8",  NULL },
         { "9",  NULL },
         { "10", NULL },
         { "11", NULL },
         { "12", NULL },
         { "13", NULL },
         { "14", NULL },
         { "15", NULL },
         { "16", NULL },
         { "17", NULL },
         { "18", NULL },
         { "19", NULL },
         { "20", NULL },
         { "21", NULL },
         { "22", NULL },
         { "23", NULL },
         { "24", NULL },
         { "25", NULL },
         { "26", NULL },
         { "27", NULL },
         { "28", NULL },
         { "29", NULL },
         { "30", NULL },
         { "31", NULL },
         { "32", NULL },
         { "33", NULL },
         { "34", NULL },
         { "35", NULL },
         { "36", NULL },
         { "37", NULL },
         { "38", NULL },
         { "39", NULL },
         { "40", NULL },
         { NULL, NULL },
      },
      "0"
   },
   {
      BEETLE_OPT(last_scanline_pal),
      BEETLE_OPT_LAST_SCANLINE_PAL_LABEL_CHT,
      NULL,
      BEETLE_OPT_LAST_SCANLINE_PAL_INFO_0_CHT,
      NULL,
      "video",
      {
         { "230", NULL },
         { "231", NULL },
         { "232", NULL },
         { "233", NULL },
         { "234", NULL },
         { "235", NULL },
         { "236", NULL },
         { "237", NULL },
         { "238", NULL },
         { "239", NULL },
         { "240", NULL },
         { "241", NULL },
         { "242", NULL },
         { "243", NULL },
         { "244", NULL },
         { "245", NULL },
         { "246", NULL },
         { "247", NULL },
         { "248", NULL },
         { "249", NULL },
         { "250", NULL },
         { "251", NULL },
         { "252", NULL },
         { "253", NULL },
         { "254", NULL },
         { "255", NULL },
         { "256", NULL },
         { "257", NULL },
         { "258", NULL },
         { "259", NULL },
         { "260", NULL },
         { "261", NULL },
         { "262", NULL },
         { "263", NULL },
         { "264", NULL },
         { "265", NULL },
         { "266", NULL },
         { "267", NULL },
         { "268", NULL },
         { "269", NULL },
         { "270", NULL },
         { "271", NULL },
         { "272", NULL },
         { "273", NULL },
         { "274", NULL },
         { "275", NULL },
         { "276", NULL },
         { "277", NULL },
         { "278", NULL },
         { "279", NULL },
         { "280", NULL },
         { "281", NULL },
         { "282", NULL },
         { "283", NULL },
         { "284", NULL },
         { "285", NULL },
         { "286", NULL },
         { "287", OPTION_VAL_287_CHT },
         { NULL, NULL },
      },
      "287"
   },
   { NULL, NULL, NULL, NULL, NULL, NULL, {{0}}, NULL },
};
struct retro_core_options_v2 options_cht = {
   option_cats_cht,
   option_defs_cht
};

/* RETRO_LANGUAGE_CS */

#define CATEGORY_SYSTEM_LABEL_CS "Systém"
#define CATEGORY_SYSTEM_INFO_0_CS "Změňte systémové a přetaktovací možnosti."
#define CATEGORY_VIDEO_LABEL_CS NULL
#define CATEGORY_VIDEO_INFO_0_CS "Změňte poměr stran, oříznutí displeje, filtr videa a nastavení přeskakování snímků."
#define CATEGORY_PGXP_LABEL_CS "PGXP (Přesná transformace geometrie)"
#define CATEGORY_PGXP_INFO_0_CS "Tyto možnosti ovládají vylepšení, která mohou zlepšit grafiku ve srovnání s původní konzolí. PGXP dokáže odstranit deformace textur a problémy Z-fightingu."
#define CATEGORY_OSD_LABEL_CS "Zobrazení na obrazovce"
#define CATEGORY_OSD_INFO_0_CS "Změna oznámení zobrazovaných na obrazovce."
#define CATEGORY_INPUT_LABEL_CS "Vstup"
#define CATEGORY_INPUT_INFO_0_CS "Změna nastavení světelné pistole, myši a neGconu."
#define CATEGORY_MEMCARDS_LABEL_CS "Paměťová karta"
#define CATEGORY_MEMCARDS_INFO_0_CS "Změňte nastavení týkající se virtuální paměťové karty."
#define CATEGORY_HACKS_LABEL_CS "Emulace hacků"
#define CATEGORY_HACKS_INFO_0_CS "Změňte nastavení přesnosti emulace, které ovlivňuje výkon na nízké úrovni a kompatibilitu."
#define BEETLE_OPT_CPU_FREQ_SCALE_LABEL_CS "Škálování Frekvence CPU (Přetaktování)"
#define BEETLE_OPT_CPU_FREQ_SCALE_INFO_0_CS "Přetaktování (nebo podtaktování) emulovaného procesoru PSX. Přetaktování může eliminovat zpomalení a zlepšit snímkovou frekvenci v některých hrách na úkor zvýšených nároků na výkon. Upozorňujeme, že některé hry mají interní omezovač snímkové frekvence a přetaktování pro ně nemusí být přínosem. V některých titulech může přetaktování způsobit, že se některé efekty animují rychleji, než je zamýšleno."
#define OPTION_VAL_50_CS NULL
#define OPTION_VAL_60_CS NULL
#define OPTION_VAL_70_CS NULL
#define OPTION_VAL_80_CS NULL
#define OPTION_VAL_90_CS NULL
#define OPTION_VAL_100_CS "100 % (nativní)"
#define OPTION_VAL_110_CS NULL
#define OPTION_VAL_120_CS NULL
#define OPTION_VAL_130_CS NULL
#define OPTION_VAL_140_CS NULL
#define OPTION_VAL_150_CS NULL
#define OPTION_VAL_160_CS NULL
#define OPTION_VAL_170_CS NULL
#define OPTION_VAL_180_CS NULL
#define OPTION_VAL_190_CS NULL
#define OPTION_VAL_200_CS NULL
#define OPTION_VAL_210_CS NULL
#define OPTION_VAL_220_CS NULL
#define OPTION_VAL_230_CS NULL
#define OPTION_VAL_240_CS NULL
#define OPTION_VAL_250_CS NULL
#define OPTION_VAL_260_CS NULL
#define OPTION_VAL_270_CS NULL
#define OPTION_VAL_280_CS NULL
#define OPTION_VAL_290_CS NULL
#define OPTION_VAL_300_CS NULL
#define OPTION_VAL_310_CS NULL
#define OPTION_VAL_320_CS NULL
#define OPTION_VAL_330_CS NULL
#define OPTION_VAL_340_CS NULL
#define OPTION_VAL_350_CS NULL
#define OPTION_VAL_360_CS NULL
#define OPTION_VAL_370_CS NULL
#define OPTION_VAL_380_CS NULL
#define OPTION_VAL_390_CS NULL
#define OPTION_VAL_400_CS NULL
#define OPTION_VAL_410_CS NULL
#define OPTION_VAL_420_CS NULL
#define OPTION_VAL_430_CS NULL
#define OPTION_VAL_440_CS NULL
#define OPTION_VAL_450_CS NULL
#define OPTION_VAL_460_CS NULL
#define OPTION_VAL_470_CS NULL
#define OPTION_VAL_480_CS NULL
#define OPTION_VAL_490_CS NULL
#define OPTION_VAL_500_CS NULL
#define OPTION_VAL_510_CS NULL
#define OPTION_VAL_520_CS NULL
#define OPTION_VAL_530_CS NULL
#define OPTION_VAL_540_CS NULL
#define OPTION_VAL_550_CS NULL
#define OPTION_VAL_560_CS NULL
#define OPTION_VAL_570_CS NULL
#define OPTION_VAL_580_CS NULL
#define OPTION_VAL_590_CS NULL
#define OPTION_VAL_600_CS NULL
#define OPTION_VAL_610_CS NULL
#define OPTION_VAL_620_CS NULL
#define OPTION_VAL_630_CS NULL
#define OPTION_VAL_640_CS NULL
#define OPTION_VAL_650_CS NULL
#define OPTION_VAL_660_CS NULL
#define OPTION_VAL_670_CS NULL
#define OPTION_VAL_680_CS NULL
#define OPTION_VAL_690_CS NULL
#define OPTION_VAL_700_CS NULL
#define OPTION_VAL_710_CS NULL
#define OPTION_VAL_720_CS NULL
#define OPTION_VAL_730_CS NULL
#define OPTION_VAL_740_CS NULL
#define OPTION_VAL_750_CS NULL
#define BEETLE_OPT_GPU_OVERCLOCK_LABEL_CS "Přetaktování GPU Rasterizer"
#define BEETLE_OPT_GPU_OVERCLOCK_INFO_0_CS "Povolení přetaktování 2D rasterizéru obsaženého v grafickém procesoru emulované konzole PSX. Nezlepšuje 3D vykreslování a obecně má malý vliv."
#define OPTION_VAL_1X_NATIVE_CS "1x (Nativní)"
#define OPTION_VAL_2X_CS NULL
#define OPTION_VAL_4X_CS NULL
#define OPTION_VAL_8X_CS NULL
#define OPTION_VAL_16X_CS NULL
#define OPTION_VAL_32X_CS NULL
#define BEETLE_OPT_GTE_OVERCLOCK_LABEL_CS "Přetaktování GTE"
#define BEETLE_OPT_GTE_OVERCLOCK_INFO_0_CS "Snižte všechny emulované operace GTE (koprocesor CPU pro 3D grafiku) na konstantní latenci jednoho cyklu. U her, které GTE hojně využívají, to může výrazně zlepšit stabilitu snímkové frekvence a času snímků."
#define BEETLE_OPT_CD_ACCESS_METHOD_LABEL_CS "Metoda přístupu k CD"
#define BEETLE_OPT_CD_ACCESS_METHOD_INFO_0_CS "„Synchronní“ napodobuje původní hardware. „Asynchronní“ může snížit zadrhávání na zařízeních s pomalým úložištěm. „Pre-Cache“ načte celý obraz disku do paměti při spuštění obsahu, což může zlepšit dobu načítání ve hře na úkor počátečního zpoždění při spuštění. „Pre-Cache“ může způsobit problémy na systémech s nízkou pamětí RAM a u fyzických médií se vrátí k „Synchronnímu“ režimu. Je nutný restart."
#define OPTION_VAL_SYNC_CS "Synchronní"
#define OPTION_VAL_ASYNC_CS "Asynchronní"
#define OPTION_VAL_PRECACHE_CS NULL
#define BEETLE_OPT_CD_FASTLOAD_LABEL_CS "Rychlost Načítání CD"
#define BEETLE_OPT_CD_FASTLOAD_INFO_0_CS "Hodnoty vyšší než „2x (nativní)“ mohou výrazně zkrátit dobu načítání ve hře, ale mohou také způsobit chyby v načasování. Některé hry nemusí nad určitou hodnotou fungovat správně."
#define OPTION_VAL_2X_NATIVE_CS "2x (Nativní)"
#define OPTION_VAL_6X_CS NULL
#define OPTION_VAL_10X_CS NULL
#define OPTION_VAL_12X_CS NULL
#define OPTION_VAL_14X_CS NULL
#define BEETLE_OPT_SKIP_BIOS_LABEL_CS "Přeskočit BIOS"
#define BEETLE_OPT_SKIP_BIOS_INFO_0_CS "Přeskočení zaváděcí animace systému PlayStation BIOS, která se obvykle zobrazuje při načítání obsahu. Poznámka: Povolení této funkce způsobuje problémy s kompatibilitou u řady her (hry chráněné proti kopírování PAL, Saga Frontier atd.)."
#define BEETLE_OPT_OVERRIDE_BIOS_LABEL_CS NULL
#define BEETLE_OPT_OVERRIDE_BIOS_INFO_0_CS NULL
#define OPTION_VAL_PSXONPSP_CS NULL
#define OPTION_VAL_PS1_ROM_CS NULL
#define OPTION_VAL_OPENBIOS_CS NULL
#define BEETLE_OPT_REGION_LABEL_CS "Systémový Region"
#define BEETLE_OPT_REGION_INFO_0_CS NULL
#define OPTION_VAL_AUTO_CS NULL
#define OPTION_VAL_NTSC_J_CS NULL
#define OPTION_VAL_NTSC_U_CS NULL
#define OPTION_VAL_PAL_CS NULL
#define BEETLE_OPT_INTERNAL_RESOLUTION_LABEL_CS "Interní rozlišení GPU"
#define BEETLE_OPT_INTERNAL_RESOLUTION_INFO_0_CS NULL
#define BEETLE_OPT_RENDERER_LABEL_CS NULL
#define BEETLE_OPT_RENDERER_INFO_0_CS NULL
#define OPTION_VAL_HARDWARE_CS NULL
#define OPTION_VAL_HARDWARE_GL_CS NULL
#define OPTION_VAL_HARDWARE_VK_CS NULL
#define OPTION_VAL_SOFTWARE_CS NULL
#define BEETLE_OPT_RENDERER_SOFTWARE_FB_LABEL_CS NULL
#define BEETLE_OPT_RENDERER_SOFTWARE_FB_INFO_0_CS "Umožňuje přesnou emulaci efektů framebufferu (např. motion blur, FF7 battle swirl) při použití hardwarových rendererů spuštěním kopie softwarového rendereru v nativním rozlišení na pozadí. Jsou-li tyto operace zakázány, jsou vynechány (OpenGL) nebo vykreslovány na GPU (Vulkan). Zakázání může zvýšit výkon, ale může způsobit závažné grafické chyby. Pokud si nejste jisti, ponechte povoleno."
#define BEETLE_OPT_DEPTH_LABEL_CS "Vnitřní hloubka barev"
#define BEETLE_OPT_DEPTH_INFO_0_CS NULL
#define OPTION_VAL_16BPP_NATIVE_CS "16 bpp (nativní)"
#define OPTION_VAL_32BPP_CS NULL
#define BEETLE_OPT_DITHER_MODE_LABEL_CS "Vzor Ditheringu"
#define BEETLE_OPT_DITHER_MODE_INFO_0_CS NULL
#define OPTION_VAL_INTERNAL_RESOLUTION_CS "Vnitřní rozlišení"
#define BEETLE_OPT_SCALED_UV_OFFSET_LABEL_CS "Odsazení UV textur"
#define BEETLE_OPT_SCALED_UV_OFFSET_INFO_0_CS "Vzorové textury pro 3D polygony s posunem pro vyšší než 1x interní rozlišení. Snižuje počet švů textur, ale může způsobit nechtěné grafické závady."
#define BEETLE_OPT_FILTER_LABEL_CS "Filtrování textur"
#define BEETLE_OPT_FILTER_INFO_0_CS NULL
#define OPTION_VAL_NEAREST_CS "Nejbližší"
#define OPTION_VAL_SABR_CS NULL
#define OPTION_VAL_XBR_CS NULL
#define OPTION_VAL_BILINEAR_CS "Bilineární"
#define OPTION_VAL_3_POINT_CS NULL
#define OPTION_VAL_JINC2_CS NULL
#define BEETLE_OPT_FILTER_EXCLUDE_SPRITE_LABEL_CS "Vyloučit sprity z filtrování"
#define BEETLE_OPT_FILTER_EXCLUDE_SPRITE_INFO_0_CS "Na sprity se nepoužije filtrování textur. Zabraňuje vzniku švů v různých hrách s pozadím vykresleným 2D sprity. Pro dosažení nejlepšího účinku použijte společně s adaptivním vyhlazováním nebo jiným filtrem následného zpracování."
#define OPTION_VAL_OPAQUE_CS "Pouze neprůhledné"
#define OPTION_VAL_ALL_CS "Neprůhledné a poloprůhledné provedení"
#define BEETLE_OPT_FILTER_EXCLUDE_2D_POLYGON_LABEL_CS "Vyloučení 2D polygonů z filtrování"
#define BEETLE_OPT_FILTER_EXCLUDE_2D_POLYGON_INFO_0_CS "Na 2D polygony nepoužívejte filtrování textur. 2D polygony jsou detekovány pomocí heuristiky a může docházet k chybám. Pro dosažení nejlepšího účinku použijte společně s adaptivním vyhlazováním nebo jiným filtrem následného zpracování."
#define BEETLE_OPT_ADAPTIVE_SMOOTHING_LABEL_CS "Adaptivní vyhlazování"
#define BEETLE_OPT_ADAPTIVE_SMOOTHING_INFO_0_CS "Vyhlazení 2D grafiky a prvků uživatelského rozhraní bez rozmazání 3D vykreslených objektů. Podporováno pouze vykreslovačem Vulkan."
#define BEETLE_OPT_SUPER_SAMPLING_LABEL_CS "Supersampling (převzorkování na nativní rozlišení)"
#define BEETLE_OPT_SUPER_SAMPLING_INFO_0_CS "Snížení vykreslovaného obsahu z interního rozlišení se zvýšeným rozlišením na nativní rozlišení. Kombinace tohoto postupu s vyššími násobiči interního rozlišení umožňuje zobrazovat hry s 3D objekty s vyhlazováním v nativním nízkém rozlišení. Nejlepších výsledků dosahuje při použití u titulů, které kombinují 2D a 3D prvky (např. 3D postavy na předem vykresleném pozadí), a dobře funguje ve spojení s CRT shadery. Podporováno pouze vykreslovačem Vulkan. Poznámka: Při zapnutí této možnosti se doporučuje vypnout možnost 'Vzor Ditheringu'."
#define BEETLE_OPT_MSAA_LABEL_CS "Vícenásobný Anti Aliasing"
#define BEETLE_OPT_MSAA_INFO_0_CS NULL
#define OPTION_VAL_1X_CS NULL
#define BEETLE_OPT_MDEC_YUV_LABEL_CS "MDEC YUV Chroma Filtr"
#define BEETLE_OPT_MDEC_YUV_INFO_0_CS "Zlepšení kvality přehrávání FMV omezením artefaktů 'makroblockingu' (čtverečky/roztřepené hrany). Podporováno pouze vykreslovačem Vulkan."
#define BEETLE_OPT_TRACK_TEXTURES_LABEL_CS "Sledování Textur"
#define BEETLE_OPT_TRACK_TEXTURES_INFO_0_CS NULL
#define BEETLE_OPT_DUMP_TEXTURES_LABEL_CS "Skladiště Textur"
#define BEETLE_OPT_DUMP_TEXTURES_INFO_0_CS "Výpis použitých textur do souboru <cd>-texture-dump/"
#define BEETLE_OPT_REPLACE_TEXTURES_LABEL_CS "Nahrazení Textur"
#define BEETLE_OPT_REPLACE_TEXTURES_INFO_0_CS "Nahrazení textur pomocí HD verzí z <cd>-texture-replacements/"
#define BEETLE_OPT_PGXP_MODE_LABEL_CS "Provozní Režim PGXP"
#define BEETLE_OPT_PGXP_MODE_INFO_0_CS NULL
#define OPTION_VAL_MEMORY_ONLY_CS "Pouze Paměť"
#define OPTION_VAL_MEMORY_CPU_CS "Paměť + CPU (Chyby)"
#define BEETLE_OPT_PGXP_2D_TOL_LABEL_CS "PGXP 2D Tolerance Geometrie"
#define BEETLE_OPT_PGXP_2D_TOL_INFO_0_CS "Skrytí zjevnějších chyb při operacích PGXP: hodnota určuje toleranci, v níž budou hodnoty PGXP zachovány v případě geometrií bez správných informací o hloubce."
#define OPTION_VAL_0PX_CS NULL
#define OPTION_VAL_1PX_CS NULL
#define OPTION_VAL_2PX_CS NULL
#define OPTION_VAL_3PX_CS NULL
#define OPTION_VAL_4PX_CS NULL
#define OPTION_VAL_5PX_CS NULL
#define OPTION_VAL_6PX_CS NULL
#define OPTION_VAL_7PX_CS NULL
#define OPTION_VAL_8PX_CS NULL
#define BEETLE_OPT_PGXP_NCLIP_LABEL_CS "PGXP Primitivní Culling"
#define BEETLE_OPT_PGXP_NCLIP_INFO_0_CS "Použijte implementaci NCLIP společnosti PGXP. Zlepšuje vzhled zmenšením děr v geometriích se souřadnicemi PGXP. Je známo, že za různých okolností způsobuje zablokování některých her."
#define BEETLE_OPT_PGXP_VERTEX_LABEL_CS NULL
#define BEETLE_OPT_PGXP_VERTEX_INFO_0_CS "Ukládání pozic vrcholů s podporou PGXP do mezipaměti pro opakované použití při kreslení polygonů. Může potenciálně zlepšit zarovnání objektů a omezit viditelné švy při vykreslování textur, ale falešně pozitivní dotazy na mezipaměť mohou způsobit grafické závady. V současné době se doporučuje ponechat tuto možnost vypnutou. Tato volba se použije pouze tehdy, když je povolen operační režim PGXP. Podporováno pouze hardwarovými vykreslovači."
#define BEETLE_OPT_PGXP_TEXTURE_LABEL_CS "PGXP Perspektivní Správné Texturování"
#define BEETLE_OPT_PGXP_TEXTURE_INFO_0_CS NULL
#define BEETLE_OPT_DISPLAY_INTERNAL_FPS_LABEL_CS "Interní Zobrazení FPS"
#define BEETLE_OPT_DISPLAY_INTERNAL_FPS_INFO_0_CS "Zobrazení interní snímkové frekvence, při které emulovaný systém PlayStation vykresluje obsah. Poznámka: Vyžaduje, aby byla ve frontendu libretro povolena oznámení na obrazovce."
#define BEETLE_OPT_DISPLAY_OSD_LABEL_CS "Zobrazení OSD zpráv"
#define BEETLE_OPT_DISPLAY_OSD_INFO_0_CS "Zobrazení OSD zpráv generovaných jádrem."
#define BEETLE_OPT_DISPLAY_VRAM_LABEL_CS "Zobrazení Plné Paměti VRAM (Ladění)"
#define BEETLE_OPT_DISPLAY_VRAM_INFO_0_CS "Vizualizace celé paměti VRAM emulované konzole. Podporováno pouze hardwarovými vykreslovači OpenGL a Vulkan. Poznámka: Tato funkce slouží k ladění a měla by být normálně vypnuta."
#define BEETLE_OPT_ANALOG_CALIBRATION_LABEL_CS "Analogová Vlastní Kalibrace"
#define BEETLE_OPT_ANALOG_CALIBRATION_INFO_0_CS NULL
#define BEETLE_OPT_ANALOG_TOGGLE_LABEL_CS "Přepínání analogového režimu DualShock"
#define BEETLE_OPT_ANALOG_TOGGLE_INFO_0_CS "Pokud je typem vstupního zařízení DualShock, umožňuje tato možnost přepínat emulovaný DualShock mezi DIGITÁLNÍM a ANALOGOVÝM režimem jako původní hardware. Režim lze přepínat také pomocí vybrané kombinace tlačítek."
#define OPTION_VAL_ENABLED_ANALOG_CS "Výchozí analogový"
#define BEETLE_OPT_ANALOG_TOGGLE_COMBO_LABEL_CS "Kombinace analogového režimu DualShock"
#define BEETLE_OPT_ANALOG_TOGGLE_COMBO_INFO_0_CS NULL
#define OPTION_VAL_L1_L2_R1_R2_START_SELECT_CS NULL
#define OPTION_VAL_L1_R1_SELECT_CS NULL
#define OPTION_VAL_L1_R1_START_CS NULL
#define OPTION_VAL_L1_R1_L3_CS NULL
#define OPTION_VAL_L1_R1_R3_CS NULL
#define OPTION_VAL_L2_R2_SELECT_CS NULL
#define OPTION_VAL_L2_R2_START_CS NULL
#define OPTION_VAL_L2_R2_L3_CS NULL
#define OPTION_VAL_L2_R2_R3_CS NULL
#define OPTION_VAL_L3_R3_CS NULL
#define BEETLE_OPT_ANALOG_TOGGLE_HOLD_LABEL_CS "Zpoždění podržení analogového režimu DualShock"
#define BEETLE_OPT_ANALOG_TOGGLE_HOLD_INFO_0_CS NULL
#define OPTION_VAL_0_CS "0 sekundové zpoždění"
#define OPTION_VAL_1_CS "1 sekundové zpoždění"
#define OPTION_VAL_2_CS "2 sekundové zpoždění"
#define OPTION_VAL_3_CS "3 sekundové zpoždění"
#define OPTION_VAL_4_CS "4 sekundové zpoždění"
#define OPTION_VAL_5_CS "5 sekundové zpoždění"
#define BEETLE_OPT_ENABLE_MULTITAP_PORT1_LABEL_CS NULL
#define BEETLE_OPT_ENABLE_MULTITAP_PORT1_INFO_0_CS "Povolit funkci multitap na portu 1."
#define BEETLE_OPT_ENABLE_MULTITAP_PORT2_LABEL_CS NULL
#define BEETLE_OPT_ENABLE_MULTITAP_PORT2_INFO_0_CS "Povolit funkci multitap na portu 2."
#define BEETLE_OPT_GUN_INPUT_MODE_LABEL_CS "Režim Vstupu Zbraně"
#define BEETLE_OPT_GUN_INPUT_MODE_INFO_0_CS NULL
#define OPTION_VAL_LIGHTGUN_CS NULL
#define OPTION_VAL_TOUCHSCREEN_CS "Dotyková Obrazovka"
#define BEETLE_OPT_GUN_CURSOR_LABEL_CS "Kurzor Zbraně"
#define BEETLE_OPT_GUN_CURSOR_INFO_0_CS NULL
#define OPTION_VAL_CROSS_CS "Kříž"
#define OPTION_VAL_DOT_CS "Tečka"
#define BEETLE_OPT_CROSSHAIR_COLOR_P1_LABEL_CS NULL
#define BEETLE_OPT_CROSSHAIR_COLOR_P1_INFO_0_CS NULL
#define OPTION_VAL_RED_CS "Červený"
#define OPTION_VAL_BLUE_CS "Modrý"
#define OPTION_VAL_GREEN_CS "Zelený"
#define OPTION_VAL_ORANGE_CS "Oranžová"
#define OPTION_VAL_YELLOW_CS "Žlutá"
#define OPTION_VAL_CYAN_CS "Azurová"
#define OPTION_VAL_PINK_CS "Růžová"
#define OPTION_VAL_PURPLE_CS "Fialová"
#define OPTION_VAL_BLACK_CS "Černá"
#define OPTION_VAL_WHITE_CS "Bílá"
#define BEETLE_OPT_CROSSHAIR_COLOR_P2_LABEL_CS NULL
#define BEETLE_OPT_CROSSHAIR_COLOR_P2_INFO_0_CS NULL
#define BEETLE_OPT_MOUSE_SENSITIVITY_LABEL_CS "Citlivost Myši"
#define BEETLE_OPT_MOUSE_SENSITIVITY_INFO_0_CS NULL
#define OPTION_VAL_5_O42_CS NULL
#define OPTION_VAL_10_CS NULL
#define OPTION_VAL_15_CS NULL
#define OPTION_VAL_20_CS NULL
#define OPTION_VAL_25_CS NULL
#define OPTION_VAL_30_CS NULL
#define OPTION_VAL_35_CS NULL
#define OPTION_VAL_40_CS NULL
#define OPTION_VAL_45_CS NULL
#define OPTION_VAL_55_CS NULL
#define OPTION_VAL_65_CS NULL
#define OPTION_VAL_75_CS NULL
#define OPTION_VAL_85_CS NULL
#define OPTION_VAL_95_CS NULL
#define OPTION_VAL_100_O42_CS NULL
#define OPTION_VAL_105_CS NULL
#define OPTION_VAL_115_CS NULL
#define OPTION_VAL_125_CS NULL
#define OPTION_VAL_135_CS NULL
#define OPTION_VAL_145_CS NULL
#define OPTION_VAL_155_CS NULL
#define OPTION_VAL_165_CS NULL
#define OPTION_VAL_175_CS NULL
#define OPTION_VAL_185_CS NULL
#define OPTION_VAL_195_CS NULL
#define BEETLE_OPT_NEGCON_RESPONSE_LABEL_CS NULL
#define BEETLE_OPT_NEGCON_RESPONSE_INFO_0_CS NULL
#define OPTION_VAL_LINEAR_CS "Lineární"
#define OPTION_VAL_QUADRATIC_CS "Kvadratická"
#define OPTION_VAL_CUBIC_CS "Kubická"
#define BEETLE_OPT_NEGCON_DEADZONE_LABEL_CS "mrtvá Zóna neGcon Twist"
#define BEETLE_OPT_NEGCON_DEADZONE_INFO_0_CS NULL
#define OPTION_VAL_0_O44_CS NULL
#define BEETLE_OPT_USE_MEDNAFEN_MEMCARD0_METHOD_LABEL_CS NULL
#define BEETLE_OPT_USE_MEDNAFEN_MEMCARD0_METHOD_INFO_0_CS NULL
#define OPTION_VAL_LIBRETRO_CS NULL
#define OPTION_VAL_MEDNAFEN_CS NULL
#define BEETLE_OPT_ENABLE_MEMCARD1_LABEL_CS NULL
#define BEETLE_OPT_ENABLE_MEMCARD1_INFO_0_CS NULL
#define BEETLE_OPT_SHARED_MEMORY_CARDS_LABEL_CS NULL
#define BEETLE_OPT_SHARED_MEMORY_CARDS_INFO_0_CS NULL
#define BEETLE_OPT_MEMCARD_LEFT_INDEX_LABEL_CS NULL
#define BEETLE_OPT_MEMCARD_LEFT_INDEX_INFO_0_CS NULL
#define OPTION_VAL_0_O48_CS NULL
#define BEETLE_OPT_MEMCARD_RIGHT_INDEX_LABEL_CS NULL
#define BEETLE_OPT_MEMCARD_RIGHT_INDEX_INFO_0_CS NULL
#define OPTION_VAL_1_O49_CS NULL
#define BEETLE_OPT_LINE_RENDER_LABEL_CS NULL
#define BEETLE_OPT_LINE_RENDER_INFO_0_CS NULL
#define OPTION_VAL_DEFAULT_CS "Výchozí"
#define OPTION_VAL_AGGRESSIVE_CS "Agresivní"
#define BEETLE_OPT_WIDESCREEN_HACK_LABEL_CS "Hack Režim Celé Obrazovky"
#define BEETLE_OPT_WIDESCREEN_HACK_INFO_0_CS "Vykresluje 3D obsah anamorfně a výstupem emulovaného framebufferu s širokoúhlým poměrem stran. Nejlepších výsledků dosahuje u plně 3D her. 2D prvky budou horizontálně roztaženy a mohou být špatně zarovnány."
#define BEETLE_OPT_WIDESCREEN_HACK_ASPECT_RATIO_LABEL_CS "Poměr Stran Hacku v Širokoúhlem Režimu"
#define BEETLE_OPT_WIDESCREEN_HACK_ASPECT_RATIO_INFO_0_CS NULL
#define OPTION_VAL_16_9_CS NULL
#define OPTION_VAL_16_10_CS NULL
#define OPTION_VAL_18_9_CS NULL
#define OPTION_VAL_19_9_CS NULL
#define OPTION_VAL_20_9_CS NULL
#define OPTION_VAL_21_9_CS NULL
#define OPTION_VAL_32_9_CS NULL
#define BEETLE_OPT_CPU_DYNAREC_LABEL_CS NULL
#define BEETLE_OPT_CPU_DYNAREC_INFO_0_CS "Dynamická překompilace instrukce procesoru na nativní instrukce. Mnohem rychlejší než interpret, ale časování CPU je méně přesné a může obsahovat chyby."
#define OPTION_VAL_DISABLED_CS "Zakázáno (Beetle Interpreter)"
#define OPTION_VAL_EXECUTE_CS "Maximální Výkon"
#define OPTION_VAL_EXECUTE_ONE_CS "Kontrola Časování Cyklu"
#define OPTION_VAL_RUN_INTERPRETER_CS "Interpret Lightrec"
#define BEETLE_OPT_DYNAREC_INVALIDATE_LABEL_CS "Zrušení Kódu Dynarec"
#define BEETLE_OPT_DYNAREC_INVALIDATE_INFO_0_CS "Některé hry vyžadují 'Plné' zneplatnění, jiné pouze 'DMA'."
#define OPTION_VAL_FULL_CS "Plné"
#define OPTION_VAL_DMA_CS "Pouze DMA (Mírně Rychlejší)"
#define BEETLE_OPT_DYNAREC_EVENTCYCLES_LABEL_CS "Cykly událostí Dynarec DMA/GPU/MDEC/Timer"
#define BEETLE_OPT_DYNAREC_EVENTCYCLES_INFO_0_CS "Maximální počet cyklů spuštěných CPU před kontrolou aktualizace GPU/DMA/MDEC/Timer, vyšší číslo bude rychlejší, má mnohem menší vliv na interpreter beetle než dynarec."
#define OPTION_VAL_128_CS "128 (Výchozí)"
#define BEETLE_OPT_DYNAREC_SPU_SAMPLES_LABEL_CS "Vzorky Dynarec SPU"
#define BEETLE_OPT_DYNAREC_SPU_SAMPLES_INFO_0_CS "Maximální počet vzorků SPU, které mají být spuštěny před kontrolou aktualizace SPU, vyšší číslo bude rychlejší, ale v některých hrách s jiným číslem než 1 bude způsobovat závady zvuku."
#define BEETLE_OPT_CORE_TIMING_FPS_LABEL_CS "Nahlášené-Jádro Časování FPS"
#define BEETLE_OPT_CORE_TIMING_FPS_INFO_0_CS NULL
#define OPTION_VAL_AUTO_TOGGLE_CS NULL
#define OPTION_VAL_FORCE_PROGRESSIVE_CS NULL
#define OPTION_VAL_FORCE_INTERLACED_CS "Vynutit Prokládanou Hodnotu"
#define BEETLE_OPT_DEINTERLACER_LABEL_CS NULL
#define BEETLE_OPT_DEINTERLACER_INFO_0_CS NULL
#define OPTION_VAL_WEAVE_CS NULL
#define OPTION_VAL_BOB_CS NULL
#define OPTION_VAL_BOB_OFFSET_CS NULL
#define OPTION_VAL_FASTMAD_CS NULL
#define BEETLE_OPT_FRAME_DUPING_LABEL_CS NULL
#define BEETLE_OPT_FRAME_DUPING_INFO_0_CS NULL
#define BEETLE_OPT_PAL_VIDEO_TIMING_OVERRIDE_LABEL_CS NULL
#define BEETLE_OPT_PAL_VIDEO_TIMING_OVERRIDE_INFO_0_CS NULL
#define BEETLE_OPT_ASPECT_RATIO_LABEL_CS "Poměr Stran Jádra"
#define BEETLE_OPT_ASPECT_RATIO_INFO_0_CS NULL
#define OPTION_VAL_CORRECTED_CS "Opraveno"
#define OPTION_VAL_UNCORRECTED_CS "Neopraveno"
#define OPTION_VAL_4_3_CS "Vunutit 4:3"
#define OPTION_VAL_NTSC_CS "Vynutit NTSC"
#define BEETLE_OPT_CROP_OVERSCAN_LABEL_CS "Overscan Oříznutí"
#define BEETLE_OPT_CROP_OVERSCAN_INFO_0_CS NULL
#define OPTION_VAL_STATIC_CS "Horizontální"
#define OPTION_VAL_SMART_CS NULL
#define BEETLE_OPT_IMAGE_CROP_LABEL_CS "Dodatečné Oříznutí"
#define BEETLE_OPT_IMAGE_CROP_INFO_0_CS NULL
#define OPTION_VAL_9PX_CS NULL
#define OPTION_VAL_10PX_CS NULL
#define OPTION_VAL_11PX_CS NULL
#define OPTION_VAL_12PX_CS NULL
#define OPTION_VAL_13PX_CS NULL
#define OPTION_VAL_14PX_CS NULL
#define OPTION_VAL_15PX_CS NULL
#define OPTION_VAL_16PX_CS NULL
#define OPTION_VAL_17PX_CS NULL
#define OPTION_VAL_18PX_CS NULL
#define OPTION_VAL_19PX_CS NULL
#define OPTION_VAL_20PX_CS NULL
#define BEETLE_OPT_IMAGE_OFFSET_LABEL_CS "Posun Oříznutého Snímku"
#define BEETLE_OPT_IMAGE_OFFSET_INFO_0_CS NULL
#define OPTION_VAL_12PX_O64_CS NULL
#define OPTION_VAL_11PX_O64_CS NULL
#define OPTION_VAL_10PX_O64_CS NULL
#define OPTION_VAL_9PX_O64_CS NULL
#define OPTION_VAL_8PX_O64_CS NULL
#define OPTION_VAL_7PX_O64_CS NULL
#define OPTION_VAL_6PX_O64_CS NULL
#define OPTION_VAL_5PX_O64_CS NULL
#define OPTION_VAL_4PX_O64_CS NULL
#define OPTION_VAL_3PX_O64_CS NULL
#define OPTION_VAL_2PX_O64_CS NULL
#define OPTION_VAL_1PX_O64_CS NULL
#define OPTION_VAL_1PX_O64_0_CS NULL
#define OPTION_VAL_2PX_O64_0_CS NULL
#define OPTION_VAL_3PX_O64_0_CS NULL
#define OPTION_VAL_4PX_O64_0_CS NULL
#define OPTION_VAL_5PX_O64_0_CS NULL
#define OPTION_VAL_6PX_O64_0_CS NULL
#define OPTION_VAL_7PX_O64_0_CS NULL
#define OPTION_VAL_8PX_O64_0_CS NULL
#define OPTION_VAL_9PX_O64_0_CS NULL
#define OPTION_VAL_10PX_O64_0_CS NULL
#define OPTION_VAL_11PX_O64_0_CS NULL
#define OPTION_VAL_12PX_O64_0_CS NULL
#define BEETLE_OPT_IMAGE_OFFSET_CYCLES_LABEL_CS "Horizontální Posun Obrazu (Cykly GPU)"
#define BEETLE_OPT_IMAGE_OFFSET_CYCLES_INFO_0_CS NULL
#define BEETLE_OPT_INITIAL_SCANLINE_LABEL_CS "Počáteční Skenovací Řádek - NTSC"
#define BEETLE_OPT_INITIAL_SCANLINE_INFO_0_CS NULL
#define BEETLE_OPT_LAST_SCANLINE_LABEL_CS "Poslední Skenovací Řádek - NTSC"
#define BEETLE_OPT_LAST_SCANLINE_INFO_0_CS NULL
#define OPTION_VAL_239_CS "239 (Výchozí)"
#define BEETLE_OPT_INITIAL_SCANLINE_PAL_LABEL_CS "Počáteční Skenovací Řádek - PAL"
#define BEETLE_OPT_INITIAL_SCANLINE_PAL_INFO_0_CS NULL
#define BEETLE_OPT_LAST_SCANLINE_PAL_LABEL_CS "Poslední Skenovací Řádek - PAL"
#define BEETLE_OPT_LAST_SCANLINE_PAL_INFO_0_CS NULL
#define OPTION_VAL_287_CS "287 (Výchozí)"

struct retro_core_option_v2_category option_cats_cs[] = {
   {
      "system",
      CATEGORY_SYSTEM_LABEL_CS,
      CATEGORY_SYSTEM_INFO_0_CS
   },
   {
      "video",
      CATEGORY_VIDEO_LABEL_CS,
      CATEGORY_VIDEO_INFO_0_CS
   },
   {
      "pgxp",
      CATEGORY_PGXP_LABEL_CS,
      CATEGORY_PGXP_INFO_0_CS
   },
   {
      "osd",
      CATEGORY_OSD_LABEL_CS,
      CATEGORY_OSD_INFO_0_CS
   },
   {
      "input",
      CATEGORY_INPUT_LABEL_CS,
      CATEGORY_INPUT_INFO_0_CS
   },
   {
      "memcards",
      CATEGORY_MEMCARDS_LABEL_CS,
      CATEGORY_MEMCARDS_INFO_0_CS
   },
   {
      "hacks",
      CATEGORY_HACKS_LABEL_CS,
      CATEGORY_HACKS_INFO_0_CS
   },
   { NULL, NULL, NULL },
};
struct retro_core_option_v2_definition option_defs_cs[] = {
   {
      BEETLE_OPT(cpu_freq_scale),
      BEETLE_OPT_CPU_FREQ_SCALE_LABEL_CS,
      NULL,
      BEETLE_OPT_CPU_FREQ_SCALE_INFO_0_CS,
      NULL,
      "system",
      {
         { "50%",           OPTION_VAL_50_CS },
         { "60%",           OPTION_VAL_60_CS },
         { "70%",           OPTION_VAL_70_CS },
         { "80%",           OPTION_VAL_80_CS },
         { "90%",           OPTION_VAL_90_CS },
         { "100%",          OPTION_VAL_100_CS },
         { "110%",          OPTION_VAL_110_CS },
         { "120%",          OPTION_VAL_120_CS },
         { "130%",          OPTION_VAL_130_CS },
         { "140%",          OPTION_VAL_140_CS },
         { "150%",          OPTION_VAL_150_CS },
         { "160%",          OPTION_VAL_160_CS },
         { "170%",          OPTION_VAL_170_CS },
         { "180%",          OPTION_VAL_180_CS },
         { "190%",          OPTION_VAL_190_CS },
         { "200%",          OPTION_VAL_200_CS },
         { "210%",          OPTION_VAL_210_CS },
         { "220%",          OPTION_VAL_220_CS },
         { "230%",          OPTION_VAL_230_CS },
         { "240%",          OPTION_VAL_240_CS },
         { "250%",          OPTION_VAL_250_CS },
         { "260%",          OPTION_VAL_260_CS },
         { "270%",          OPTION_VAL_270_CS },
         { "280%",          OPTION_VAL_280_CS },
         { "290%",          OPTION_VAL_290_CS },
         { "300%",          OPTION_VAL_300_CS },
         { "310%",          OPTION_VAL_310_CS },
         { "320%",          OPTION_VAL_320_CS },
         { "330%",          OPTION_VAL_330_CS },
         { "340%",          OPTION_VAL_340_CS },
         { "350%",          OPTION_VAL_350_CS },
         { "360%",          OPTION_VAL_360_CS },
         { "370%",          OPTION_VAL_370_CS },
         { "380%",          OPTION_VAL_380_CS },
         { "390%",          OPTION_VAL_390_CS },
         { "400%",          OPTION_VAL_400_CS },
         { "410%",          OPTION_VAL_410_CS },
         { "420%",          OPTION_VAL_420_CS },
         { "430%",          OPTION_VAL_430_CS },
         { "440%",          OPTION_VAL_440_CS },
         { "450%",          OPTION_VAL_450_CS },
         { "460%",          OPTION_VAL_460_CS },
         { "470%",          OPTION_VAL_470_CS },
         { "480%",          OPTION_VAL_480_CS },
         { "490%",          OPTION_VAL_490_CS },
         { "500%",          OPTION_VAL_500_CS },
         { "510%",          OPTION_VAL_510_CS },
         { "520%",          OPTION_VAL_520_CS },
         { "530%",          OPTION_VAL_530_CS },
         { "540%",          OPTION_VAL_540_CS },
         { "550%",          OPTION_VAL_550_CS },
         { "560%",          OPTION_VAL_560_CS },
         { "570%",          OPTION_VAL_570_CS },
         { "580%",          OPTION_VAL_580_CS },
         { "590%",          OPTION_VAL_590_CS },
         { "600%",          OPTION_VAL_600_CS },
         { "610%",          OPTION_VAL_610_CS },
         { "620%",          OPTION_VAL_620_CS },
         { "630%",          OPTION_VAL_630_CS },
         { "640%",          OPTION_VAL_640_CS },
         { "650%",          OPTION_VAL_650_CS },
         { "660%",          OPTION_VAL_660_CS },
         { "670%",          OPTION_VAL_670_CS },
         { "680%",          OPTION_VAL_680_CS },
         { "690%",          OPTION_VAL_690_CS },
         { "700%",          OPTION_VAL_700_CS },
         { "710%",          OPTION_VAL_710_CS },
         { "720%",          OPTION_VAL_720_CS },
         { "730%",          OPTION_VAL_730_CS },
         { "740%",          OPTION_VAL_740_CS },
         { "750%",          OPTION_VAL_750_CS },
         { NULL, NULL },
      },
      "100%"
   },
   {
      BEETLE_OPT(gpu_overclock),
      BEETLE_OPT_GPU_OVERCLOCK_LABEL_CS,
      NULL,
      BEETLE_OPT_GPU_OVERCLOCK_INFO_0_CS,
      NULL,
      "system",
      {
         { "1x(native)", OPTION_VAL_1X_NATIVE_CS },
         { "2x",         OPTION_VAL_2X_CS },
         { "4x",         OPTION_VAL_4X_CS },
         { "8x",         OPTION_VAL_8X_CS },
         { "16x",        OPTION_VAL_16X_CS },
         { "32x",        OPTION_VAL_32X_CS },
         { NULL, NULL },
      },
      "1x(native)"
   },
   {
      BEETLE_OPT(gte_overclock),
      BEETLE_OPT_GTE_OVERCLOCK_LABEL_CS,
      NULL,
      BEETLE_OPT_GTE_OVERCLOCK_INFO_0_CS,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
#ifndef EMSCRIPTEN
   {
      BEETLE_OPT(cd_access_method),
      BEETLE_OPT_CD_ACCESS_METHOD_LABEL_CS,
      NULL,
      BEETLE_OPT_CD_ACCESS_METHOD_INFO_0_CS,
      NULL,
      "system",
      {
         { "sync",     OPTION_VAL_SYNC_CS },
         { "async",    OPTION_VAL_ASYNC_CS },
         { "precache", OPTION_VAL_PRECACHE_CS },
         { NULL, NULL },
      },
      "sync"
   },
#endif
   {
      BEETLE_OPT(cd_fastload),
      BEETLE_OPT_CD_FASTLOAD_LABEL_CS,
      NULL,
      BEETLE_OPT_CD_FASTLOAD_INFO_0_CS,
      NULL,
      "system",
      {
         { "2x(native)", OPTION_VAL_2X_NATIVE_CS },
         { "4x",          OPTION_VAL_4X_CS },
         { "6x",          OPTION_VAL_6X_CS },
         { "8x",          OPTION_VAL_8X_CS },
         { "10x",         OPTION_VAL_10X_CS },
         { "12x",         OPTION_VAL_12X_CS },
         { "14x",         OPTION_VAL_14X_CS },
         { NULL, NULL },
      },
      "2x(native)"
   },
   {
      BEETLE_OPT(skip_bios),
      BEETLE_OPT_SKIP_BIOS_LABEL_CS,
      NULL,
      BEETLE_OPT_SKIP_BIOS_INFO_0_CS,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(override_bios),
      BEETLE_OPT_OVERRIDE_BIOS_LABEL_CS,
      NULL,
      BEETLE_OPT_OVERRIDE_BIOS_INFO_0_CS,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "psxonpsp", OPTION_VAL_PSXONPSP_CS },
         { "ps1_rom",  OPTION_VAL_PS1_ROM_CS },
         { "openbios", OPTION_VAL_OPENBIOS_CS     },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(region),
      BEETLE_OPT_REGION_LABEL_CS,
      NULL,
      BEETLE_OPT_REGION_INFO_0_CS,
      NULL,
      "system",
      {
         { "auto",   OPTION_VAL_AUTO_CS },
         { "ntsc-j", OPTION_VAL_NTSC_J_CS },
         { "ntsc-u", OPTION_VAL_NTSC_U_CS },
         { "pal",    OPTION_VAL_PAL_CS },
         { NULL, NULL },
      },
      "auto"
   },
   {
      BEETLE_OPT(internal_resolution),
      BEETLE_OPT_INTERNAL_RESOLUTION_LABEL_CS,
      NULL,
      BEETLE_OPT_INTERNAL_RESOLUTION_INFO_0_CS,
      NULL,
      "video",
      {
         { "1x(native)", OPTION_VAL_1X_NATIVE_CS },
         { "2x",         OPTION_VAL_2X_CS },
         { "4x",         OPTION_VAL_4X_CS },
         { "8x",         OPTION_VAL_8X_CS },
         { "16x",        OPTION_VAL_16X_CS },
         { NULL, NULL },
      },
      "1x(native)"
   },
#if defined(HAVE_OPENGL) || defined(HAVE_OPENGLES) || defined(HAVE_VULKAN)
   {
      BEETLE_OPT(renderer),
      BEETLE_OPT_RENDERER_LABEL_CS,
      NULL,
      BEETLE_OPT_RENDERER_INFO_0_CS,
      NULL,
      "video",
      {
         { "hardware",    OPTION_VAL_HARDWARE_CS },
#if defined(HAVE_OPENGL) || defined(HAVE_OPENGLES)
         { "hardware_gl", OPTION_VAL_HARDWARE_GL_CS },
#endif
#if defined(HAVE_VULKAN)
         { "hardware_vk", OPTION_VAL_HARDWARE_VK_CS },
#endif
         { "software",    OPTION_VAL_SOFTWARE_CS },
         { NULL, NULL },
      },
      "hardware"
   },
   {
      BEETLE_OPT(renderer_software_fb),
      BEETLE_OPT_RENDERER_SOFTWARE_FB_LABEL_CS,
      NULL,
      BEETLE_OPT_RENDERER_SOFTWARE_FB_INFO_0_CS,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled"
   },
#endif
#if defined(HAVE_OPENGL) || defined(HAVE_OPENGLES)
   {
      BEETLE_OPT(depth),
      BEETLE_OPT_DEPTH_LABEL_CS,
      NULL,
      BEETLE_OPT_DEPTH_INFO_0_CS,
      NULL,
      "video",
      {
         { "16bpp(native)", OPTION_VAL_16BPP_NATIVE_CS },
         { "32bpp",         OPTION_VAL_32BPP_CS },
         { NULL, NULL },
      },
      "16bpp(native)"
   },
   // Sort of, it's more like 15-bit high color and 24-bit true color for visible output. The alpha channel is used for mask bit. Vulkan renderer uses ABGR1555_555 for 31 bits internal? FMVs are always 24-bit on all renderers like original hardware (BGR888, no alpha)
#endif
   {
      BEETLE_OPT(dither_mode),
      BEETLE_OPT_DITHER_MODE_LABEL_CS,
      NULL,
      BEETLE_OPT_DITHER_MODE_INFO_0_CS,
      NULL,
      "video",
      {
         { "1x(native)",          OPTION_VAL_1X_NATIVE_CS },
         { "internal resolution", OPTION_VAL_INTERNAL_RESOLUTION_CS },
         { "disabled",            NULL },
         { NULL, NULL },
      },
      "1x(native)"
   },
#ifdef HAVE_VULKAN
   {
      BEETLE_OPT(scaled_uv_offset),
      BEETLE_OPT_SCALED_UV_OFFSET_LABEL_CS,
      NULL,
      BEETLE_OPT_SCALED_UV_OFFSET_INFO_0_CS,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled"
   },
#endif
#if defined(HAVE_OPENGL) || defined(HAVE_OPENGLES) || defined(HAVE_VULKAN)
   {
      BEETLE_OPT(filter),
      BEETLE_OPT_FILTER_LABEL_CS,
      NULL,
      BEETLE_OPT_FILTER_INFO_0_CS,
      NULL,
      "video",
      {
         { "nearest",  OPTION_VAL_NEAREST_CS },
         { "SABR",     OPTION_VAL_SABR_CS },
         { "xBR",      OPTION_VAL_XBR_CS },
         { "bilinear", OPTION_VAL_BILINEAR_CS },
         { "3-point",  OPTION_VAL_3_POINT_CS },
         { "JINC2",    OPTION_VAL_JINC2_CS },
         { NULL, NULL },
      },
      "nearest"
   },
#ifdef HAVE_VULKAN
   {
      BEETLE_OPT(filter_exclude_sprite),
      BEETLE_OPT_FILTER_EXCLUDE_SPRITE_LABEL_CS,
      NULL,
      BEETLE_OPT_FILTER_EXCLUDE_SPRITE_INFO_0_CS,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "opaque", OPTION_VAL_OPAQUE_CS },
         { "all", OPTION_VAL_ALL_CS },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(filter_exclude_2d_polygon),
      BEETLE_OPT_FILTER_EXCLUDE_2D_POLYGON_LABEL_CS,
      NULL,
      BEETLE_OPT_FILTER_EXCLUDE_2D_POLYGON_INFO_0_CS,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "opaque", OPTION_VAL_OPAQUE_CS },
         { "all", OPTION_VAL_ALL_CS },
         { NULL, NULL },
      },
      "disabled"
   },
#endif
#endif
#ifdef HAVE_VULKAN
   {
      BEETLE_OPT(adaptive_smoothing),
      BEETLE_OPT_ADAPTIVE_SMOOTHING_LABEL_CS,
      NULL,
      BEETLE_OPT_ADAPTIVE_SMOOTHING_INFO_0_CS,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(super_sampling),
      BEETLE_OPT_SUPER_SAMPLING_LABEL_CS,
      NULL,
      BEETLE_OPT_SUPER_SAMPLING_INFO_0_CS,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(msaa),
      BEETLE_OPT_MSAA_LABEL_CS,
      NULL,
      BEETLE_OPT_MSAA_INFO_0_CS,
      NULL,
      "video",
      {
         { "1x",  OPTION_VAL_1X_CS },
         { "2x",  OPTION_VAL_2X_CS },
         { "4x",  OPTION_VAL_4X_CS },
         { "8x",  OPTION_VAL_8X_CS },
         { "16x", OPTION_VAL_16X_CS },
         { NULL, NULL },
      },
      "1x"
   },
   {
      BEETLE_OPT(mdec_yuv),
      BEETLE_OPT_MDEC_YUV_LABEL_CS,
      NULL,
      BEETLE_OPT_MDEC_YUV_INFO_0_CS,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(track_textures),
      BEETLE_OPT_TRACK_TEXTURES_LABEL_CS,
      NULL,
      BEETLE_OPT_TRACK_TEXTURES_INFO_0_CS,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
#ifdef TEXTURE_DUMPING_ENABLED
   {
      BEETLE_OPT(dump_textures),
      BEETLE_OPT_DUMP_TEXTURES_LABEL_CS,
      NULL,
      BEETLE_OPT_DUMP_TEXTURES_INFO_0_CS,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
#endif
   {
      BEETLE_OPT(replace_textures),
      BEETLE_OPT_REPLACE_TEXTURES_LABEL_CS,
      NULL,
      BEETLE_OPT_REPLACE_TEXTURES_INFO_0_CS,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
#endif
   {
      BEETLE_OPT(pgxp_mode),
      BEETLE_OPT_PGXP_MODE_LABEL_CS,
      NULL,
      BEETLE_OPT_PGXP_MODE_INFO_0_CS,
      NULL,
      "pgxp",
      {
         { "disabled",     NULL },
         { "memory only",  OPTION_VAL_MEMORY_ONLY_CS },
         { "memory + CPU", OPTION_VAL_MEMORY_CPU_CS },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(pgxp_2d_tol),
      BEETLE_OPT_PGXP_2D_TOL_LABEL_CS,
      NULL,
      BEETLE_OPT_PGXP_2D_TOL_INFO_0_CS,
      NULL,
      "pgxp",
      {
         { "disabled", NULL },
         { "0px", OPTION_VAL_0PX_CS },
         { "1px", OPTION_VAL_1PX_CS },
         { "2px", OPTION_VAL_2PX_CS },
         { "3px", OPTION_VAL_3PX_CS },
         { "4px", OPTION_VAL_4PX_CS },
         { "5px", OPTION_VAL_5PX_CS },
         { "6px", OPTION_VAL_6PX_CS },
         { "7px", OPTION_VAL_7PX_CS },
         { "8px", OPTION_VAL_8PX_CS },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(pgxp_nclip),
      BEETLE_OPT_PGXP_NCLIP_LABEL_CS,
      NULL,
      BEETLE_OPT_PGXP_NCLIP_INFO_0_CS,
      NULL,
      "pgxp",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
#if defined(HAVE_OPENGL) || defined(HAVE_OPENGLES) || defined(HAVE_VULKAN)
   {
      BEETLE_OPT(pgxp_vertex),
      BEETLE_OPT_PGXP_VERTEX_LABEL_CS,
      NULL,
      BEETLE_OPT_PGXP_VERTEX_INFO_0_CS,
      NULL,
      "pgxp",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
#endif
   {
      BEETLE_OPT(pgxp_texture),
      BEETLE_OPT_PGXP_TEXTURE_LABEL_CS,
      NULL,
      BEETLE_OPT_PGXP_TEXTURE_INFO_0_CS,
      NULL,
      "pgxp",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(display_internal_fps),
      BEETLE_OPT_DISPLAY_INTERNAL_FPS_LABEL_CS,
      NULL,
      BEETLE_OPT_DISPLAY_INTERNAL_FPS_INFO_0_CS,
      NULL,
      "osd",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(display_osd),
      BEETLE_OPT_DISPLAY_OSD_LABEL_CS,
      NULL,
      BEETLE_OPT_DISPLAY_OSD_INFO_0_CS,
      NULL,
      "osd",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled"
   },
#if defined(HAVE_OPENGL) || defined(HAVE_OPENGLES) || defined(HAVE_VULKAN)
   {
      BEETLE_OPT(display_vram),
      BEETLE_OPT_DISPLAY_VRAM_LABEL_CS,
      NULL,
      BEETLE_OPT_DISPLAY_VRAM_INFO_0_CS,
      NULL,
      "osd",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
#endif
   {
      BEETLE_OPT(analog_calibration),
      BEETLE_OPT_ANALOG_CALIBRATION_LABEL_CS,
      NULL,
      BEETLE_OPT_ANALOG_CALIBRATION_INFO_0_CS,
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(analog_toggle),
      BEETLE_OPT_ANALOG_TOGGLE_LABEL_CS,
      NULL,
      BEETLE_OPT_ANALOG_TOGGLE_INFO_0_CS,
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { "enabled-analog", OPTION_VAL_ENABLED_ANALOG_CS },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(analog_toggle_combo),
      BEETLE_OPT_ANALOG_TOGGLE_COMBO_LABEL_CS,
      NULL,
      BEETLE_OPT_ANALOG_TOGGLE_COMBO_INFO_0_CS,
      NULL,
      "input",
      {
         { "l1+l2+r1+r2+start+select", OPTION_VAL_L1_L2_R1_R2_START_SELECT_CS },
         { "l1+r1+select",             OPTION_VAL_L1_R1_SELECT_CS },
         { "l1+r1+start",              OPTION_VAL_L1_R1_START_CS },
         { "l1+r1+l3",                 OPTION_VAL_L1_R1_L3_CS },
         { "l1+r1+r3",                 OPTION_VAL_L1_R1_R3_CS },
         { "l2+r2+select",             OPTION_VAL_L2_R2_SELECT_CS },
         { "l2+r2+start",              OPTION_VAL_L2_R2_START_CS },
         { "l2+r2+l3",                 OPTION_VAL_L2_R2_L3_CS },
         { "l2+r2+r3",                 OPTION_VAL_L2_R2_R3_CS },
         { "l3+r3",                    OPTION_VAL_L3_R3_CS },
         { NULL, NULL },
      },
      "l1+r1+select"
   },
   {
      BEETLE_OPT(analog_toggle_hold),
      BEETLE_OPT_ANALOG_TOGGLE_HOLD_LABEL_CS,
      NULL,
      BEETLE_OPT_ANALOG_TOGGLE_HOLD_INFO_0_CS,
      NULL,
      "input",
      {
         { "0", OPTION_VAL_0_CS },
         { "1", OPTION_VAL_1_CS },
         { "2", OPTION_VAL_2_CS },
         { "3", OPTION_VAL_3_CS },
         { "4", OPTION_VAL_4_CS },
         { "5", OPTION_VAL_5_CS },
         { NULL, NULL },
      },
      "1"
   },
   {
      BEETLE_OPT(enable_multitap_port1),
      BEETLE_OPT_ENABLE_MULTITAP_PORT1_LABEL_CS,
      NULL,
      BEETLE_OPT_ENABLE_MULTITAP_PORT1_INFO_0_CS,
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(enable_multitap_port2),
      BEETLE_OPT_ENABLE_MULTITAP_PORT2_LABEL_CS,
      NULL,
      BEETLE_OPT_ENABLE_MULTITAP_PORT2_INFO_0_CS,
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(gun_input_mode),
      BEETLE_OPT_GUN_INPUT_MODE_LABEL_CS,
      NULL,
      BEETLE_OPT_GUN_INPUT_MODE_INFO_0_CS,
      NULL,
      "input",
      {
         { "lightgun",    OPTION_VAL_LIGHTGUN_CS },
         { "touchscreen", OPTION_VAL_TOUCHSCREEN_CS },
         { NULL, NULL },
      },
      "lightgun"
   },
   // Shouldn't the gun_input_mode just be Mouse vs. Touchscreen?
   {
      BEETLE_OPT(gun_cursor),
      BEETLE_OPT_GUN_CURSOR_LABEL_CS,
      NULL,
      BEETLE_OPT_GUN_CURSOR_INFO_0_CS,
      NULL,
      "input",
      {
         { "off",   "disabled" },
         { "cross", OPTION_VAL_CROSS_CS },
         { "dot",   OPTION_VAL_DOT_CS },
         { NULL, NULL },
      },
      "cross"
   },
   {
      BEETLE_OPT(crosshair_color_p1),
      BEETLE_OPT_CROSSHAIR_COLOR_P1_LABEL_CS,
      NULL,
      BEETLE_OPT_CROSSHAIR_COLOR_P1_INFO_0_CS,
      NULL,
      "input",
      {
         { "red",    OPTION_VAL_RED_CS },
         { "blue",   OPTION_VAL_BLUE_CS },
         { "green",  OPTION_VAL_GREEN_CS },
         { "orange", OPTION_VAL_ORANGE_CS },
         { "yellow", OPTION_VAL_YELLOW_CS },
         { "cyan",   OPTION_VAL_CYAN_CS },
         { "pink",   OPTION_VAL_PINK_CS },
         { "purple", OPTION_VAL_PURPLE_CS },
         { "black",  OPTION_VAL_BLACK_CS },
         { "white",  OPTION_VAL_WHITE_CS },
         { NULL, NULL },
      },
      "red"
   },
   {
      BEETLE_OPT(crosshair_color_p2),
      BEETLE_OPT_CROSSHAIR_COLOR_P2_LABEL_CS,
      NULL,
      BEETLE_OPT_CROSSHAIR_COLOR_P2_INFO_0_CS,
      NULL,
      "input",
      {
         { "blue",   OPTION_VAL_BLUE_CS },
         { "red",    OPTION_VAL_RED_CS },
         { "green",  OPTION_VAL_GREEN_CS },
         { "orange", OPTION_VAL_ORANGE_CS },
         { "yellow", OPTION_VAL_YELLOW_CS },
         { "cyan",   OPTION_VAL_CYAN_CS },
         { "pink",   OPTION_VAL_PINK_CS },
         { "purple", OPTION_VAL_PURPLE_CS },
         { "black",  OPTION_VAL_BLACK_CS },
         { "white",  OPTION_VAL_WHITE_CS },
         { NULL, NULL },
      },
      "blue"
   },
   {
      BEETLE_OPT(mouse_sensitivity),
      BEETLE_OPT_MOUSE_SENSITIVITY_LABEL_CS,
      NULL,
      BEETLE_OPT_MOUSE_SENSITIVITY_INFO_0_CS,
      NULL,
      "input",
      {
         { "5%",   OPTION_VAL_5_O42_CS },
         { "10%",  OPTION_VAL_10_CS },
         { "15%",  OPTION_VAL_15_CS },
         { "20%",  OPTION_VAL_20_CS },
         { "25%",  OPTION_VAL_25_CS },
         { "30%",  OPTION_VAL_30_CS },
         { "35%",  OPTION_VAL_35_CS },
         { "40%",  OPTION_VAL_40_CS },
         { "45%",  OPTION_VAL_45_CS },
         { "50%",  OPTION_VAL_50_CS },
         { "55%",  OPTION_VAL_55_CS },
         { "60%",  OPTION_VAL_60_CS },
         { "65%",  OPTION_VAL_65_CS },
         { "70%",  OPTION_VAL_70_CS },
         { "75%",  OPTION_VAL_75_CS },
         { "80%",  OPTION_VAL_80_CS },
         { "85%",  OPTION_VAL_85_CS },
         { "90%",  OPTION_VAL_90_CS },
         { "95%",  OPTION_VAL_95_CS },
         { "100%", OPTION_VAL_100_O42_CS },
         { "105%", OPTION_VAL_105_CS },
         { "110%", OPTION_VAL_110_CS },
         { "115%", OPTION_VAL_115_CS },
         { "120%", OPTION_VAL_120_CS },
         { "125%", OPTION_VAL_125_CS },
         { "130%", OPTION_VAL_130_CS },
         { "135%", OPTION_VAL_135_CS },
         { "140%", OPTION_VAL_140_CS },
         { "145%", OPTION_VAL_145_CS },
         { "150%", OPTION_VAL_150_CS },
         { "155%", OPTION_VAL_155_CS },
         { "160%", OPTION_VAL_160_CS },
         { "165%", OPTION_VAL_165_CS },
         { "170%", OPTION_VAL_170_CS },
         { "175%", OPTION_VAL_175_CS },
         { "180%", OPTION_VAL_180_CS },
         { "185%", OPTION_VAL_185_CS },
         { "190%", OPTION_VAL_190_CS },
         { "195%", OPTION_VAL_195_CS },
         { "200%", OPTION_VAL_200_CS },
         { NULL, NULL },
      },
      "100%"
   },
   {
      BEETLE_OPT(negcon_response),
      BEETLE_OPT_NEGCON_RESPONSE_LABEL_CS,
      NULL,
      BEETLE_OPT_NEGCON_RESPONSE_INFO_0_CS,
      NULL,
      "input",
      {
         { "linear",    OPTION_VAL_LINEAR_CS },
         { "quadratic", OPTION_VAL_QUADRATIC_CS },
         { "cubic",     OPTION_VAL_CUBIC_CS },
         { NULL, NULL },
      },
      "linear"
   },
   {
      BEETLE_OPT(negcon_deadzone),
      BEETLE_OPT_NEGCON_DEADZONE_LABEL_CS,
      NULL,
      BEETLE_OPT_NEGCON_DEADZONE_INFO_0_CS,
      NULL,
      "input",
      {
         { "0%",  OPTION_VAL_0_O44_CS },
         { "5%",  OPTION_VAL_5_O42_CS },
         { "10%", OPTION_VAL_10_CS },
         { "15%", OPTION_VAL_15_CS },
         { "20%", OPTION_VAL_20_CS },
         { "25%", OPTION_VAL_25_CS },
         { "30%", OPTION_VAL_30_CS },
         { NULL, NULL },
      },
      "0%"
   },
   {
      BEETLE_OPT(use_mednafen_memcard0_method),
      BEETLE_OPT_USE_MEDNAFEN_MEMCARD0_METHOD_LABEL_CS,
      NULL,
      BEETLE_OPT_USE_MEDNAFEN_MEMCARD0_METHOD_INFO_0_CS,
      NULL,
      "memcards",
      {
         { "libretro", OPTION_VAL_LIBRETRO_CS },
         { "mednafen", OPTION_VAL_MEDNAFEN_CS },
         { NULL, NULL },
      },
      "libretro"
   },
   {
      BEETLE_OPT(enable_memcard1),
      BEETLE_OPT_ENABLE_MEMCARD1_LABEL_CS,
      NULL,
      BEETLE_OPT_ENABLE_MEMCARD1_INFO_0_CS,
      NULL,
      "memcards",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(shared_memory_cards),
      BEETLE_OPT_SHARED_MEMORY_CARDS_LABEL_CS,
      NULL,
      BEETLE_OPT_SHARED_MEMORY_CARDS_INFO_0_CS,
      NULL,
      "memcards",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(memcard_left_index),
      BEETLE_OPT_MEMCARD_LEFT_INDEX_LABEL_CS,
      NULL,
      BEETLE_OPT_MEMCARD_LEFT_INDEX_INFO_0_CS,
      NULL,
      "memcards",
      {
         { "0",  OPTION_VAL_0_O48_CS },
         { "1",  NULL },
         { "2",  NULL },
         { "3",  NULL },
         { "4",  NULL },
         { "5",  NULL },
         { "6",  NULL },
         { "7",  NULL },
         { "8",  NULL },
         { "9",  NULL },
         { "10", NULL },
         { "11", NULL },
         { "12", NULL },
         { "13", NULL },
         { "14", NULL },
         { "15", NULL },
         { "16", NULL },
         { "17", NULL },
         { "18", NULL },
         { "19", NULL },
         { "20", NULL },
         { "21", NULL },
         { "22", NULL },
         { "23", NULL },
         { "24", NULL },
         { "25", NULL },
         { "26", NULL },
         { "27", NULL },
         { "28", NULL },
         { "29", NULL },
         { "30", NULL },
         { "31", NULL },
         { "32", NULL },
         { "33", NULL },
         { "34", NULL },
         { "35", NULL },
         { "36", NULL },
         { "37", NULL },
         { "38", NULL },
         { "39", NULL },
         { "40", NULL },
         { "41", NULL },
         { "42", NULL },
         { "43", NULL },
         { "44", NULL },
         { "45", NULL },
         { "46", NULL },
         { "47", NULL },
         { "48", NULL },
         { "49", NULL },
         { "50", NULL },
         { "51", NULL },
         { "52", NULL },
         { "53", NULL },
         { "54", NULL },
         { "55", NULL },
         { "56", NULL },
         { "57", NULL },
         { "58", NULL },
         { "59", NULL },
         { "60", NULL },
         { "61", NULL },
         { "62", NULL },
         { "63", NULL },
         { NULL, NULL },
      },
      "0"
   },
   {
      BEETLE_OPT(memcard_right_index),
      BEETLE_OPT_MEMCARD_RIGHT_INDEX_LABEL_CS,
      NULL,
      BEETLE_OPT_MEMCARD_RIGHT_INDEX_INFO_0_CS,
      NULL,
      "memcards",
      {
         { "0",  NULL },
         { "1",  OPTION_VAL_1_O49_CS },
         { "2",  NULL },
         { "3",  NULL },
         { "4",  NULL },
         { "5",  NULL },
         { "6",  NULL },
         { "7",  NULL },
         { "8",  NULL },
         { "9",  NULL },
         { "10", NULL },
         { "11", NULL },
         { "12", NULL },
         { "13", NULL },
         { "14", NULL },
         { "15", NULL },
         { "16", NULL },
         { "17", NULL },
         { "18", NULL },
         { "19", NULL },
         { "20", NULL },
         { "21", NULL },
         { "22", NULL },
         { "23", NULL },
         { "24", NULL },
         { "25", NULL },
         { "26", NULL },
         { "27", NULL },
         { "28", NULL },
         { "29", NULL },
         { "30", NULL },
         { "31", NULL },
         { "32", NULL },
         { "33", NULL },
         { "34", NULL },
         { "35", NULL },
         { "36", NULL },
         { "37", NULL },
         { "38", NULL },
         { "39", NULL },
         { "40", NULL },
         { "41", NULL },
         { "42", NULL },
         { "43", NULL },
         { "44", NULL },
         { "45", NULL },
         { "46", NULL },
         { "47", NULL },
         { "48", NULL },
         { "49", NULL },
         { "50", NULL },
         { "51", NULL },
         { "52", NULL },
         { "53", NULL },
         { "54", NULL },
         { "55", NULL },
         { "56", NULL },
         { "57", NULL },
         { "58", NULL },
         { "59", NULL },
         { "60", NULL },
         { "61", NULL },
         { "62", NULL },
         { "63", NULL },
         { NULL, NULL },
      },
      "1"
   },
   {
      BEETLE_OPT(line_render),
      BEETLE_OPT_LINE_RENDER_LABEL_CS,
      NULL,
      BEETLE_OPT_LINE_RENDER_INFO_0_CS,
      NULL,
      "hacks",
      {
         { "default",    OPTION_VAL_DEFAULT_CS },
         { "aggressive", OPTION_VAL_AGGRESSIVE_CS },
         { "disabled",   NULL },
         { NULL, NULL },
      },
      "default"
   },
   {
      BEETLE_OPT(widescreen_hack),
      BEETLE_OPT_WIDESCREEN_HACK_LABEL_CS,
      NULL,
      BEETLE_OPT_WIDESCREEN_HACK_INFO_0_CS,
      NULL,
      "hacks",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(widescreen_hack_aspect_ratio),
      BEETLE_OPT_WIDESCREEN_HACK_ASPECT_RATIO_LABEL_CS,
      NULL,
      BEETLE_OPT_WIDESCREEN_HACK_ASPECT_RATIO_INFO_0_CS,
      NULL,
      "hacks",
      {
         { "16:9",  OPTION_VAL_16_9_CS },
         { "16:10", OPTION_VAL_16_10_CS },
         { "18:9",  OPTION_VAL_18_9_CS },
         { "19:9",  OPTION_VAL_19_9_CS },
         { "20:9",  OPTION_VAL_20_9_CS },
         { "21:9",  OPTION_VAL_21_9_CS }, // 64:27
         { "32:9",  OPTION_VAL_32_9_CS },
         { NULL,    NULL },
      },
      "16:9"
   },
#if defined(HAVE_LIGHTREC)
   {
      BEETLE_OPT(cpu_dynarec),
      BEETLE_OPT_CPU_DYNAREC_LABEL_CS,
      NULL,
      BEETLE_OPT_CPU_DYNAREC_INFO_0_CS,
      NULL,
      "hacks",
      {
         { "disabled",        OPTION_VAL_DISABLED_CS },
         { "execute",         OPTION_VAL_EXECUTE_CS },
         { "execute_one",     OPTION_VAL_EXECUTE_ONE_CS },
         { "run_interpreter", OPTION_VAL_RUN_INTERPRETER_CS },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(dynarec_invalidate),
      BEETLE_OPT_DYNAREC_INVALIDATE_LABEL_CS,
      NULL,
      BEETLE_OPT_DYNAREC_INVALIDATE_INFO_0_CS,
      NULL,
      "hacks",
      {
         { "full", OPTION_VAL_FULL_CS },
         { "dma",  OPTION_VAL_DMA_CS },
         { NULL, NULL },
      },
      "full"
   },
   {
      BEETLE_OPT(dynarec_eventcycles),
      BEETLE_OPT_DYNAREC_EVENTCYCLES_LABEL_CS,
      NULL,
      BEETLE_OPT_DYNAREC_EVENTCYCLES_INFO_0_CS,
      NULL,
      "hacks",
      {
         { "128",  OPTION_VAL_128_CS },
         { "256",  NULL },
         { "384",  NULL },
         { "512",  NULL },
         { "640",  NULL },
         { "768",  NULL },
         { "896",  NULL },
         { "1024", NULL },
         { "1152", NULL },
         { "1280", NULL },
         { "1408", NULL },
         { "1536", NULL },
         { "1664", NULL },
         { "1792", NULL },
         { "1920", NULL },
         { "2048", NULL },
         { NULL, NULL },
      },
      "128"
   },
   {
      BEETLE_OPT(dynarec_spu_samples),
      BEETLE_OPT_DYNAREC_SPU_SAMPLES_LABEL_CS,
      NULL,
      BEETLE_OPT_DYNAREC_SPU_SAMPLES_INFO_0_CS,
      NULL,
      "hacks",
      {
         { "1",  OPTION_VAL_1_O49_CS },
         { "4",  NULL },
         { "16", NULL },
         { NULL, NULL },
      },
      "1"
   },
#endif
   {
      BEETLE_OPT(core_timing_fps),
      BEETLE_OPT_CORE_TIMING_FPS_LABEL_CS,
      NULL,
      BEETLE_OPT_CORE_TIMING_FPS_INFO_0_CS,
      NULL,
      "video",
      {
         { "auto_toggle",       OPTION_VAL_AUTO_TOGGLE_CS },
         { "force_progressive", OPTION_VAL_FORCE_PROGRESSIVE_CS },
         { "force_interlaced",  OPTION_VAL_FORCE_INTERLACED_CS },
      },
      "auto_toggle"
   },
   {
      BEETLE_OPT(deinterlacer),
      BEETLE_OPT_DEINTERLACER_LABEL_CS,
      NULL,
      BEETLE_OPT_DEINTERLACER_INFO_0_CS,
      NULL,
      "video",
      {
         { "weave",      OPTION_VAL_WEAVE_CS },
         { "bob",        OPTION_VAL_BOB_CS },
         { "bob_offset", OPTION_VAL_BOB_OFFSET_CS },
         { "fastmad",    OPTION_VAL_FASTMAD_CS },
         { "off",        "Off" },
         { NULL, NULL },
      },
      "weave"
   },
   {
      BEETLE_OPT(frame_duping),
      BEETLE_OPT_FRAME_DUPING_LABEL_CS,
      NULL,
      BEETLE_OPT_FRAME_DUPING_INFO_0_CS,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(pal_video_timing_override),
      BEETLE_OPT_PAL_VIDEO_TIMING_OVERRIDE_LABEL_CS,
      NULL,
      BEETLE_OPT_PAL_VIDEO_TIMING_OVERRIDE_INFO_0_CS,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(aspect_ratio),
      BEETLE_OPT_ASPECT_RATIO_LABEL_CS,
      NULL,
      BEETLE_OPT_ASPECT_RATIO_INFO_0_CS,
      NULL,
      "video",
      {
         { "corrected", OPTION_VAL_CORRECTED_CS },
         { "uncorrected", OPTION_VAL_UNCORRECTED_CS },
         { "4:3",  OPTION_VAL_4_3_CS },
         { "ntsc", OPTION_VAL_NTSC_CS },
      },
      "corrected"
   },
   {
      BEETLE_OPT(crop_overscan),
      BEETLE_OPT_CROP_OVERSCAN_LABEL_CS,
      NULL,
      BEETLE_OPT_CROP_OVERSCAN_INFO_0_CS,
      NULL,
      "video",
      {
         { "disabled", "Disabled" },
         { "static",   OPTION_VAL_STATIC_CS },
         { "smart",    OPTION_VAL_SMART_CS },
         { NULL, NULL },
      },
      "smart"
   },
   {
      BEETLE_OPT(image_crop),
      BEETLE_OPT_IMAGE_CROP_LABEL_CS,
      NULL,
      BEETLE_OPT_IMAGE_CROP_INFO_0_CS,
      NULL,
      "video",
      {
         { "disabled", "0" },
         { "1px",      OPTION_VAL_1PX_CS },
         { "2px",      OPTION_VAL_2PX_CS },
         { "3px",      OPTION_VAL_3PX_CS },
         { "4px",      OPTION_VAL_4PX_CS },
         { "5px",      OPTION_VAL_5PX_CS },
         { "6px",      OPTION_VAL_6PX_CS },
         { "7px",      OPTION_VAL_7PX_CS },
         { "8px",      OPTION_VAL_8PX_CS },
         { "9px",      OPTION_VAL_9PX_CS },
         { "10px",     OPTION_VAL_10PX_CS },
         { "11px",     OPTION_VAL_11PX_CS },
         { "12px",     OPTION_VAL_12PX_CS },
         { "13px",     OPTION_VAL_13PX_CS },
         { "14px",     OPTION_VAL_14PX_CS },
         { "15px",     OPTION_VAL_15PX_CS },
         { "16px",     OPTION_VAL_16PX_CS },
         { "17px",     OPTION_VAL_17PX_CS },
         { "18px",     OPTION_VAL_18PX_CS },
         { "19px",     OPTION_VAL_19PX_CS },
         { "20px",     OPTION_VAL_20PX_CS },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(image_offset),
      BEETLE_OPT_IMAGE_OFFSET_LABEL_CS,
      NULL,
      BEETLE_OPT_IMAGE_OFFSET_INFO_0_CS,
      NULL,
      "video",
      {
         { "-12px",    OPTION_VAL_12PX_O64_CS },
         { "-11px",    OPTION_VAL_11PX_O64_CS },
         { "-10px",    OPTION_VAL_10PX_O64_CS },
         { "-9px",     OPTION_VAL_9PX_O64_CS },
         { "-8px",     OPTION_VAL_8PX_O64_CS },
         { "-7px",     OPTION_VAL_7PX_O64_CS },
         { "-6px",     OPTION_VAL_6PX_O64_CS },
         { "-5px",     OPTION_VAL_5PX_O64_CS },
         { "-4px",     OPTION_VAL_4PX_O64_CS },
         { "-3px",     OPTION_VAL_3PX_O64_CS },
         { "-2px",     OPTION_VAL_2PX_O64_CS },
         { "-1px",     OPTION_VAL_1PX_O64_CS },
         { "disabled", "0" },
         { "+1px",     OPTION_VAL_1PX_O64_0_CS },
         { "+2px",     OPTION_VAL_2PX_O64_0_CS },
         { "+3px",     OPTION_VAL_3PX_O64_0_CS },
         { "+4px",     OPTION_VAL_4PX_O64_0_CS },
         { "+5px",     OPTION_VAL_5PX_O64_0_CS },
         { "+6px",     OPTION_VAL_6PX_O64_0_CS },
         { "+7px",     OPTION_VAL_7PX_O64_0_CS },
         { "+8px",     OPTION_VAL_8PX_O64_0_CS },
         { "+9px",     OPTION_VAL_9PX_O64_0_CS },
         { "+10px",    OPTION_VAL_10PX_O64_0_CS },
         { "+11px",    OPTION_VAL_11PX_O64_0_CS },
         { "+12px",    OPTION_VAL_12PX_O64_0_CS },
         { NULL, NULL },
      },
      "disabled"
   },
#if defined(HAVE_OPENGL) || defined(HAVE_OPENGLES) || defined(HAVE_VULKAN)
   {
      BEETLE_OPT(image_offset_cycles),
      BEETLE_OPT_IMAGE_OFFSET_CYCLES_LABEL_CS,
      NULL,
      BEETLE_OPT_IMAGE_OFFSET_CYCLES_INFO_0_CS,
      NULL,
      "video",
      {
         { "-40",      NULL },
         { "-39",      NULL },
         { "-38",      NULL },
         { "-37",      NULL },
         { "-36",      NULL },
         { "-35",      NULL },
         { "-34",      NULL },
         { "-33",      NULL },
         { "-32",      NULL },
         { "-31",      NULL },
         { "-30",      NULL },
         { "-29",      NULL },
         { "-28",      NULL },
         { "-27",      NULL },
         { "-26",      NULL },
         { "-25",      NULL },
         { "-24",      NULL },
         { "-23",      NULL },
         { "-22",      NULL },
         { "-21",      NULL },
         { "-20",      NULL },
         { "-19",      NULL },
         { "-18",      NULL },
         { "-17",      NULL },
         { "-16",      NULL },
         { "-15",      NULL },
         { "-14",      NULL },
         { "-13",      NULL },
         { "-12",      NULL },
         { "-11",      NULL },
         { "-10",      NULL },
         { "-9",       NULL },
         { "-8",       NULL },
         { "-7",       NULL },
         { "-6",       NULL },
         { "-5",       NULL },
         { "-4",       NULL },
         { "-3",       NULL },
         { "-2",       NULL },
         { "-1",       NULL },
         { "0",        "0" },
         { "+1",       NULL },
         { "+2",       NULL },
         { "+3",       NULL },
         { "+4",       NULL },
         { "+5",       NULL },
         { "+6",       NULL },
         { "+7",       NULL },
         { "+8",       NULL },
         { "+9",       NULL },
         { "+10",      NULL },
         { "+11",      NULL },
         { "+12",      NULL },
         { "+13",      NULL },
         { "+14",      NULL },
         { "+15",      NULL },
         { "+16",      NULL },
         { "+17",      NULL },
         { "+18",      NULL },
         { "+19",      NULL },
         { "+20",      NULL },
         { "+21",      NULL },
         { "+22",      NULL },
         { "+23",      NULL },
         { "+24",      NULL },
         { "+25",      NULL },
         { "+26",      NULL },
         { "+27",      NULL },
         { "+28",      NULL },
         { "+29",      NULL },
         { "+30",      NULL },
         { "+31",      NULL },
         { "+32",      NULL },
         { "+33",      NULL },
         { "+34",      NULL },
         { "+35",      NULL },
         { "+36",      NULL },
         { "+37",      NULL },
         { "+38",      NULL },
         { "+39",      NULL },
         { "+40",      NULL },
         { NULL, NULL},
      },
      "0"
   },
#endif
   {
      BEETLE_OPT(initial_scanline),
      BEETLE_OPT_INITIAL_SCANLINE_LABEL_CS,
      NULL,
      BEETLE_OPT_INITIAL_SCANLINE_INFO_0_CS,
      NULL,
      "video",
      {
         { "0",  OPTION_VAL_0_O48_CS },
         { "1",  NULL },
         { "2",  NULL },
         { "3",  NULL },
         { "4",  NULL },
         { "5",  NULL },
         { "6",  NULL },
         { "7",  NULL },
         { "8",  NULL },
         { "9",  NULL },
         { "10", NULL },
         { "11", NULL },
         { "12", NULL },
         { "13", NULL },
         { "14", NULL },
         { "15", NULL },
         { "16", NULL },
         { "17", NULL },
         { "18", NULL },
         { "19", NULL },
         { "20", NULL },
         { "21", NULL },
         { "22", NULL },
         { "23", NULL },
         { "24", NULL },
         { "25", NULL },
         { "26", NULL },
         { "27", NULL },
         { "28", NULL },
         { "29", NULL },
         { "30", NULL },
         { "31", NULL },
         { "32", NULL },
         { "33", NULL },
         { "34", NULL },
         { "35", NULL },
         { "36", NULL },
         { "37", NULL },
         { "38", NULL },
         { "39", NULL },
         { "40", NULL },
         { NULL, NULL },
      },
      "0"
   },
   {
      BEETLE_OPT(last_scanline),
      BEETLE_OPT_LAST_SCANLINE_LABEL_CS,
      NULL,
      BEETLE_OPT_LAST_SCANLINE_INFO_0_CS,
      NULL,
      "video",
      {
         { "210", NULL },
         { "211", NULL },
         { "212", NULL },
         { "213", NULL },
         { "214", NULL },
         { "215", NULL },
         { "216", NULL },
         { "217", NULL },
         { "218", NULL },
         { "219", NULL },
         { "220", NULL },
         { "221", NULL },
         { "222", NULL },
         { "223", NULL },
         { "224", NULL },
         { "225", NULL },
         { "226", NULL },
         { "227", NULL },
         { "228", NULL },
         { "229", NULL },
         { "230", NULL },
         { "231", NULL },
         { "232", NULL },
         { "233", NULL },
         { "234", NULL },
         { "235", NULL },
         { "236", NULL },
         { "237", NULL },
         { "238", NULL },
         { "239", OPTION_VAL_239_CS },
         { NULL, NULL },
      },
      "239"
   },
   {
      BEETLE_OPT(initial_scanline_pal),
      BEETLE_OPT_INITIAL_SCANLINE_PAL_LABEL_CS,
      NULL,
      BEETLE_OPT_INITIAL_SCANLINE_PAL_INFO_0_CS,
      NULL,
      "video",
      {
         { "0",  OPTION_VAL_0_O48_CS },
         { "1",  NULL },
         { "2",  NULL },
         { "3",  NULL },
         { "4",  NULL },
         { "5",  NULL },
         { "6",  NULL },
         { "7",  NULL },
         { "8",  NULL },
         { "9",  NULL },
         { "10", NULL },
         { "11", NULL },
         { "12", NULL },
         { "13", NULL },
         { "14", NULL },
         { "15", NULL },
         { "16", NULL },
         { "17", NULL },
         { "18", NULL },
         { "19", NULL },
         { "20", NULL },
         { "21", NULL },
         { "22", NULL },
         { "23", NULL },
         { "24", NULL },
         { "25", NULL },
         { "26", NULL },
         { "27", NULL },
         { "28", NULL },
         { "29", NULL },
         { "30", NULL },
         { "31", NULL },
         { "32", NULL },
         { "33", NULL },
         { "34", NULL },
         { "35", NULL },
         { "36", NULL },
         { "37", NULL },
         { "38", NULL },
         { "39", NULL },
         { "40", NULL },
         { NULL, NULL },
      },
      "0"
   },
   {
      BEETLE_OPT(last_scanline_pal),
      BEETLE_OPT_LAST_SCANLINE_PAL_LABEL_CS,
      NULL,
      BEETLE_OPT_LAST_SCANLINE_PAL_INFO_0_CS,
      NULL,
      "video",
      {
         { "230", NULL },
         { "231", NULL },
         { "232", NULL },
         { "233", NULL },
         { "234", NULL },
         { "235", NULL },
         { "236", NULL },
         { "237", NULL },
         { "238", NULL },
         { "239", NULL },
         { "240", NULL },
         { "241", NULL },
         { "242", NULL },
         { "243", NULL },
         { "244", NULL },
         { "245", NULL },
         { "246", NULL },
         { "247", NULL },
         { "248", NULL },
         { "249", NULL },
         { "250", NULL },
         { "251", NULL },
         { "252", NULL },
         { "253", NULL },
         { "254", NULL },
         { "255", NULL },
         { "256", NULL },
         { "257", NULL },
         { "258", NULL },
         { "259", NULL },
         { "260", NULL },
         { "261", NULL },
         { "262", NULL },
         { "263", NULL },
         { "264", NULL },
         { "265", NULL },
         { "266", NULL },
         { "267", NULL },
         { "268", NULL },
         { "269", NULL },
         { "270", NULL },
         { "271", NULL },
         { "272", NULL },
         { "273", NULL },
         { "274", NULL },
         { "275", NULL },
         { "276", NULL },
         { "277", NULL },
         { "278", NULL },
         { "279", NULL },
         { "280", NULL },
         { "281", NULL },
         { "282", NULL },
         { "283", NULL },
         { "284", NULL },
         { "285", NULL },
         { "286", NULL },
         { "287", OPTION_VAL_287_CS },
         { NULL, NULL },
      },
      "287"
   },
   { NULL, NULL, NULL, NULL, NULL, NULL, {{0}}, NULL },
};
struct retro_core_options_v2 options_cs = {
   option_cats_cs,
   option_defs_cs
};

/* RETRO_LANGUAGE_CY */

#define CATEGORY_SYSTEM_LABEL_CY NULL
#define CATEGORY_SYSTEM_INFO_0_CY NULL
#define CATEGORY_VIDEO_LABEL_CY NULL
#define CATEGORY_VIDEO_INFO_0_CY NULL
#define CATEGORY_PGXP_LABEL_CY NULL
#define CATEGORY_PGXP_INFO_0_CY NULL
#define CATEGORY_OSD_LABEL_CY NULL
#define CATEGORY_OSD_INFO_0_CY NULL
#define CATEGORY_INPUT_LABEL_CY "Mewnbwn"
#define CATEGORY_INPUT_INFO_0_CY NULL
#define CATEGORY_MEMCARDS_LABEL_CY NULL
#define CATEGORY_MEMCARDS_INFO_0_CY NULL
#define CATEGORY_HACKS_LABEL_CY NULL
#define CATEGORY_HACKS_INFO_0_CY NULL
#define BEETLE_OPT_CPU_FREQ_SCALE_LABEL_CY NULL
#define BEETLE_OPT_CPU_FREQ_SCALE_INFO_0_CY NULL
#define OPTION_VAL_50_CY NULL
#define OPTION_VAL_60_CY NULL
#define OPTION_VAL_70_CY NULL
#define OPTION_VAL_80_CY NULL
#define OPTION_VAL_90_CY NULL
#define OPTION_VAL_100_CY NULL
#define OPTION_VAL_110_CY NULL
#define OPTION_VAL_120_CY NULL
#define OPTION_VAL_130_CY NULL
#define OPTION_VAL_140_CY NULL
#define OPTION_VAL_150_CY NULL
#define OPTION_VAL_160_CY NULL
#define OPTION_VAL_170_CY NULL
#define OPTION_VAL_180_CY NULL
#define OPTION_VAL_190_CY NULL
#define OPTION_VAL_200_CY NULL
#define OPTION_VAL_210_CY NULL
#define OPTION_VAL_220_CY NULL
#define OPTION_VAL_230_CY NULL
#define OPTION_VAL_240_CY NULL
#define OPTION_VAL_250_CY NULL
#define OPTION_VAL_260_CY NULL
#define OPTION_VAL_270_CY NULL
#define OPTION_VAL_280_CY NULL
#define OPTION_VAL_290_CY NULL
#define OPTION_VAL_300_CY NULL
#define OPTION_VAL_310_CY NULL
#define OPTION_VAL_320_CY NULL
#define OPTION_VAL_330_CY NULL
#define OPTION_VAL_340_CY NULL
#define OPTION_VAL_350_CY NULL
#define OPTION_VAL_360_CY NULL
#define OPTION_VAL_370_CY NULL
#define OPTION_VAL_380_CY NULL
#define OPTION_VAL_390_CY NULL
#define OPTION_VAL_400_CY NULL
#define OPTION_VAL_410_CY NULL
#define OPTION_VAL_420_CY NULL
#define OPTION_VAL_430_CY NULL
#define OPTION_VAL_440_CY NULL
#define OPTION_VAL_450_CY NULL
#define OPTION_VAL_460_CY NULL
#define OPTION_VAL_470_CY NULL
#define OPTION_VAL_480_CY NULL
#define OPTION_VAL_490_CY NULL
#define OPTION_VAL_500_CY NULL
#define OPTION_VAL_510_CY NULL
#define OPTION_VAL_520_CY NULL
#define OPTION_VAL_530_CY NULL
#define OPTION_VAL_540_CY NULL
#define OPTION_VAL_550_CY NULL
#define OPTION_VAL_560_CY NULL
#define OPTION_VAL_570_CY NULL
#define OPTION_VAL_580_CY NULL
#define OPTION_VAL_590_CY NULL
#define OPTION_VAL_600_CY NULL
#define OPTION_VAL_610_CY NULL
#define OPTION_VAL_620_CY NULL
#define OPTION_VAL_630_CY NULL
#define OPTION_VAL_640_CY NULL
#define OPTION_VAL_650_CY NULL
#define OPTION_VAL_660_CY NULL
#define OPTION_VAL_670_CY NULL
#define OPTION_VAL_680_CY NULL
#define OPTION_VAL_690_CY NULL
#define OPTION_VAL_700_CY NULL
#define OPTION_VAL_710_CY NULL
#define OPTION_VAL_720_CY NULL
#define OPTION_VAL_730_CY NULL
#define OPTION_VAL_740_CY NULL
#define OPTION_VAL_750_CY NULL
#define BEETLE_OPT_GPU_OVERCLOCK_LABEL_CY NULL
#define BEETLE_OPT_GPU_OVERCLOCK_INFO_0_CY NULL
#define OPTION_VAL_1X_NATIVE_CY NULL
#define OPTION_VAL_2X_CY NULL
#define OPTION_VAL_4X_CY NULL
#define OPTION_VAL_8X_CY NULL
#define OPTION_VAL_16X_CY NULL
#define OPTION_VAL_32X_CY NULL
#define BEETLE_OPT_GTE_OVERCLOCK_LABEL_CY NULL
#define BEETLE_OPT_GTE_OVERCLOCK_INFO_0_CY NULL
#define BEETLE_OPT_CD_ACCESS_METHOD_LABEL_CY NULL
#define BEETLE_OPT_CD_ACCESS_METHOD_INFO_0_CY NULL
#define OPTION_VAL_SYNC_CY NULL
#define OPTION_VAL_ASYNC_CY NULL
#define OPTION_VAL_PRECACHE_CY NULL
#define BEETLE_OPT_CD_FASTLOAD_LABEL_CY NULL
#define BEETLE_OPT_CD_FASTLOAD_INFO_0_CY NULL
#define OPTION_VAL_2X_NATIVE_CY NULL
#define OPTION_VAL_6X_CY NULL
#define OPTION_VAL_10X_CY NULL
#define OPTION_VAL_12X_CY NULL
#define OPTION_VAL_14X_CY NULL
#define BEETLE_OPT_SKIP_BIOS_LABEL_CY NULL
#define BEETLE_OPT_SKIP_BIOS_INFO_0_CY NULL
#define BEETLE_OPT_OVERRIDE_BIOS_LABEL_CY NULL
#define BEETLE_OPT_OVERRIDE_BIOS_INFO_0_CY NULL
#define OPTION_VAL_PSXONPSP_CY NULL
#define OPTION_VAL_PS1_ROM_CY NULL
#define OPTION_VAL_OPENBIOS_CY NULL
#define BEETLE_OPT_REGION_LABEL_CY NULL
#define BEETLE_OPT_REGION_INFO_0_CY NULL
#define OPTION_VAL_AUTO_CY NULL
#define OPTION_VAL_NTSC_J_CY NULL
#define OPTION_VAL_NTSC_U_CY NULL
#define OPTION_VAL_PAL_CY NULL
#define BEETLE_OPT_INTERNAL_RESOLUTION_LABEL_CY NULL
#define BEETLE_OPT_INTERNAL_RESOLUTION_INFO_0_CY NULL
#define BEETLE_OPT_RENDERER_LABEL_CY NULL
#define BEETLE_OPT_RENDERER_INFO_0_CY NULL
#define OPTION_VAL_HARDWARE_CY NULL
#define OPTION_VAL_HARDWARE_GL_CY NULL
#define OPTION_VAL_HARDWARE_VK_CY NULL
#define OPTION_VAL_SOFTWARE_CY NULL
#define BEETLE_OPT_RENDERER_SOFTWARE_FB_LABEL_CY NULL
#define BEETLE_OPT_RENDERER_SOFTWARE_FB_INFO_0_CY NULL
#define BEETLE_OPT_DEPTH_LABEL_CY NULL
#define BEETLE_OPT_DEPTH_INFO_0_CY NULL
#define OPTION_VAL_16BPP_NATIVE_CY NULL
#define OPTION_VAL_32BPP_CY NULL
#define BEETLE_OPT_DITHER_MODE_LABEL_CY NULL
#define BEETLE_OPT_DITHER_MODE_INFO_0_CY NULL
#define OPTION_VAL_INTERNAL_RESOLUTION_CY NULL
#define BEETLE_OPT_SCALED_UV_OFFSET_LABEL_CY NULL
#define BEETLE_OPT_SCALED_UV_OFFSET_INFO_0_CY NULL
#define BEETLE_OPT_FILTER_LABEL_CY NULL
#define BEETLE_OPT_FILTER_INFO_0_CY NULL
#define OPTION_VAL_NEAREST_CY NULL
#define OPTION_VAL_SABR_CY NULL
#define OPTION_VAL_XBR_CY NULL
#define OPTION_VAL_BILINEAR_CY NULL
#define OPTION_VAL_3_POINT_CY NULL
#define OPTION_VAL_JINC2_CY NULL
#define BEETLE_OPT_FILTER_EXCLUDE_SPRITE_LABEL_CY NULL
#define BEETLE_OPT_FILTER_EXCLUDE_SPRITE_INFO_0_CY NULL
#define OPTION_VAL_OPAQUE_CY NULL
#define OPTION_VAL_ALL_CY NULL
#define BEETLE_OPT_FILTER_EXCLUDE_2D_POLYGON_LABEL_CY NULL
#define BEETLE_OPT_FILTER_EXCLUDE_2D_POLYGON_INFO_0_CY NULL
#define BEETLE_OPT_ADAPTIVE_SMOOTHING_LABEL_CY NULL
#define BEETLE_OPT_ADAPTIVE_SMOOTHING_INFO_0_CY NULL
#define BEETLE_OPT_SUPER_SAMPLING_LABEL_CY NULL
#define BEETLE_OPT_SUPER_SAMPLING_INFO_0_CY NULL
#define BEETLE_OPT_MSAA_LABEL_CY NULL
#define BEETLE_OPT_MSAA_INFO_0_CY NULL
#define OPTION_VAL_1X_CY NULL
#define BEETLE_OPT_MDEC_YUV_LABEL_CY NULL
#define BEETLE_OPT_MDEC_YUV_INFO_0_CY NULL
#define BEETLE_OPT_TRACK_TEXTURES_LABEL_CY NULL
#define BEETLE_OPT_TRACK_TEXTURES_INFO_0_CY NULL
#define BEETLE_OPT_DUMP_TEXTURES_LABEL_CY NULL
#define BEETLE_OPT_DUMP_TEXTURES_INFO_0_CY NULL
#define BEETLE_OPT_REPLACE_TEXTURES_LABEL_CY NULL
#define BEETLE_OPT_REPLACE_TEXTURES_INFO_0_CY NULL
#define BEETLE_OPT_PGXP_MODE_LABEL_CY NULL
#define BEETLE_OPT_PGXP_MODE_INFO_0_CY NULL
#define OPTION_VAL_MEMORY_ONLY_CY NULL
#define OPTION_VAL_MEMORY_CPU_CY NULL
#define BEETLE_OPT_PGXP_2D_TOL_LABEL_CY NULL
#define BEETLE_OPT_PGXP_2D_TOL_INFO_0_CY NULL
#define OPTION_VAL_0PX_CY NULL
#define OPTION_VAL_1PX_CY NULL
#define OPTION_VAL_2PX_CY NULL
#define OPTION_VAL_3PX_CY NULL
#define OPTION_VAL_4PX_CY NULL
#define OPTION_VAL_5PX_CY NULL
#define OPTION_VAL_6PX_CY NULL
#define OPTION_VAL_7PX_CY NULL
#define OPTION_VAL_8PX_CY NULL
#define BEETLE_OPT_PGXP_NCLIP_LABEL_CY NULL
#define BEETLE_OPT_PGXP_NCLIP_INFO_0_CY NULL
#define BEETLE_OPT_PGXP_VERTEX_LABEL_CY NULL
#define BEETLE_OPT_PGXP_VERTEX_INFO_0_CY NULL
#define BEETLE_OPT_PGXP_TEXTURE_LABEL_CY NULL
#define BEETLE_OPT_PGXP_TEXTURE_INFO_0_CY NULL
#define BEETLE_OPT_DISPLAY_INTERNAL_FPS_LABEL_CY NULL
#define BEETLE_OPT_DISPLAY_INTERNAL_FPS_INFO_0_CY NULL
#define BEETLE_OPT_DISPLAY_OSD_LABEL_CY NULL
#define BEETLE_OPT_DISPLAY_OSD_INFO_0_CY NULL
#define BEETLE_OPT_DISPLAY_VRAM_LABEL_CY NULL
#define BEETLE_OPT_DISPLAY_VRAM_INFO_0_CY NULL
#define BEETLE_OPT_ANALOG_CALIBRATION_LABEL_CY NULL
#define BEETLE_OPT_ANALOG_CALIBRATION_INFO_0_CY NULL
#define BEETLE_OPT_ANALOG_TOGGLE_LABEL_CY NULL
#define BEETLE_OPT_ANALOG_TOGGLE_INFO_0_CY NULL
#define OPTION_VAL_ENABLED_ANALOG_CY NULL
#define BEETLE_OPT_ANALOG_TOGGLE_COMBO_LABEL_CY NULL
#define BEETLE_OPT_ANALOG_TOGGLE_COMBO_INFO_0_CY NULL
#define OPTION_VAL_L1_L2_R1_R2_START_SELECT_CY NULL
#define OPTION_VAL_L1_R1_SELECT_CY NULL
#define OPTION_VAL_L1_R1_START_CY NULL
#define OPTION_VAL_L1_R1_L3_CY NULL
#define OPTION_VAL_L1_R1_R3_CY NULL
#define OPTION_VAL_L2_R2_SELECT_CY NULL
#define OPTION_VAL_L2_R2_START_CY NULL
#define OPTION_VAL_L2_R2_L3_CY NULL
#define OPTION_VAL_L2_R2_R3_CY NULL
#define OPTION_VAL_L3_R3_CY NULL
#define BEETLE_OPT_ANALOG_TOGGLE_HOLD_LABEL_CY NULL
#define BEETLE_OPT_ANALOG_TOGGLE_HOLD_INFO_0_CY NULL
#define OPTION_VAL_0_CY NULL
#define OPTION_VAL_1_CY NULL
#define OPTION_VAL_2_CY NULL
#define OPTION_VAL_3_CY NULL
#define OPTION_VAL_4_CY NULL
#define OPTION_VAL_5_CY NULL
#define BEETLE_OPT_ENABLE_MULTITAP_PORT1_LABEL_CY NULL
#define BEETLE_OPT_ENABLE_MULTITAP_PORT1_INFO_0_CY NULL
#define BEETLE_OPT_ENABLE_MULTITAP_PORT2_LABEL_CY NULL
#define BEETLE_OPT_ENABLE_MULTITAP_PORT2_INFO_0_CY NULL
#define BEETLE_OPT_GUN_INPUT_MODE_LABEL_CY NULL
#define BEETLE_OPT_GUN_INPUT_MODE_INFO_0_CY NULL
#define OPTION_VAL_LIGHTGUN_CY NULL
#define OPTION_VAL_TOUCHSCREEN_CY NULL
#define BEETLE_OPT_GUN_CURSOR_LABEL_CY NULL
#define BEETLE_OPT_GUN_CURSOR_INFO_0_CY NULL
#define OPTION_VAL_CROSS_CY NULL
#define OPTION_VAL_DOT_CY NULL
#define BEETLE_OPT_CROSSHAIR_COLOR_P1_LABEL_CY NULL
#define BEETLE_OPT_CROSSHAIR_COLOR_P1_INFO_0_CY NULL
#define OPTION_VAL_RED_CY "Coch"
#define OPTION_VAL_BLUE_CY NULL
#define OPTION_VAL_GREEN_CY NULL
#define OPTION_VAL_ORANGE_CY NULL
#define OPTION_VAL_YELLOW_CY NULL
#define OPTION_VAL_CYAN_CY NULL
#define OPTION_VAL_PINK_CY NULL
#define OPTION_VAL_PURPLE_CY NULL
#define OPTION_VAL_BLACK_CY NULL
#define OPTION_VAL_WHITE_CY NULL
#define BEETLE_OPT_CROSSHAIR_COLOR_P2_LABEL_CY NULL
#define BEETLE_OPT_CROSSHAIR_COLOR_P2_INFO_0_CY NULL
#define BEETLE_OPT_MOUSE_SENSITIVITY_LABEL_CY NULL
#define BEETLE_OPT_MOUSE_SENSITIVITY_INFO_0_CY NULL
#define OPTION_VAL_5_O42_CY NULL
#define OPTION_VAL_10_CY NULL
#define OPTION_VAL_15_CY NULL
#define OPTION_VAL_20_CY NULL
#define OPTION_VAL_25_CY NULL
#define OPTION_VAL_30_CY NULL
#define OPTION_VAL_35_CY NULL
#define OPTION_VAL_40_CY NULL
#define OPTION_VAL_45_CY NULL
#define OPTION_VAL_55_CY NULL
#define OPTION_VAL_65_CY NULL
#define OPTION_VAL_75_CY NULL
#define OPTION_VAL_85_CY NULL
#define OPTION_VAL_95_CY NULL
#define OPTION_VAL_100_O42_CY NULL
#define OPTION_VAL_105_CY NULL
#define OPTION_VAL_115_CY NULL
#define OPTION_VAL_125_CY NULL
#define OPTION_VAL_135_CY NULL
#define OPTION_VAL_145_CY NULL
#define OPTION_VAL_155_CY NULL
#define OPTION_VAL_165_CY NULL
#define OPTION_VAL_175_CY NULL
#define OPTION_VAL_185_CY NULL
#define OPTION_VAL_195_CY NULL
#define BEETLE_OPT_NEGCON_RESPONSE_LABEL_CY NULL
#define BEETLE_OPT_NEGCON_RESPONSE_INFO_0_CY NULL
#define OPTION_VAL_LINEAR_CY NULL
#define OPTION_VAL_QUADRATIC_CY NULL
#define OPTION_VAL_CUBIC_CY NULL
#define BEETLE_OPT_NEGCON_DEADZONE_LABEL_CY NULL
#define BEETLE_OPT_NEGCON_DEADZONE_INFO_0_CY NULL
#define OPTION_VAL_0_O44_CY NULL
#define BEETLE_OPT_USE_MEDNAFEN_MEMCARD0_METHOD_LABEL_CY NULL
#define BEETLE_OPT_USE_MEDNAFEN_MEMCARD0_METHOD_INFO_0_CY NULL
#define OPTION_VAL_LIBRETRO_CY NULL
#define OPTION_VAL_MEDNAFEN_CY NULL
#define BEETLE_OPT_ENABLE_MEMCARD1_LABEL_CY NULL
#define BEETLE_OPT_ENABLE_MEMCARD1_INFO_0_CY NULL
#define BEETLE_OPT_SHARED_MEMORY_CARDS_LABEL_CY NULL
#define BEETLE_OPT_SHARED_MEMORY_CARDS_INFO_0_CY NULL
#define BEETLE_OPT_MEMCARD_LEFT_INDEX_LABEL_CY NULL
#define BEETLE_OPT_MEMCARD_LEFT_INDEX_INFO_0_CY NULL
#define OPTION_VAL_0_O48_CY NULL
#define BEETLE_OPT_MEMCARD_RIGHT_INDEX_LABEL_CY NULL
#define BEETLE_OPT_MEMCARD_RIGHT_INDEX_INFO_0_CY NULL
#define OPTION_VAL_1_O49_CY NULL
#define BEETLE_OPT_LINE_RENDER_LABEL_CY NULL
#define BEETLE_OPT_LINE_RENDER_INFO_0_CY NULL
#define OPTION_VAL_DEFAULT_CY NULL
#define OPTION_VAL_AGGRESSIVE_CY NULL
#define BEETLE_OPT_WIDESCREEN_HACK_LABEL_CY NULL
#define BEETLE_OPT_WIDESCREEN_HACK_INFO_0_CY NULL
#define BEETLE_OPT_WIDESCREEN_HACK_ASPECT_RATIO_LABEL_CY NULL
#define BEETLE_OPT_WIDESCREEN_HACK_ASPECT_RATIO_INFO_0_CY NULL
#define OPTION_VAL_16_9_CY NULL
#define OPTION_VAL_16_10_CY NULL
#define OPTION_VAL_18_9_CY NULL
#define OPTION_VAL_19_9_CY NULL
#define OPTION_VAL_20_9_CY NULL
#define OPTION_VAL_21_9_CY NULL
#define OPTION_VAL_32_9_CY NULL
#define BEETLE_OPT_CPU_DYNAREC_LABEL_CY NULL
#define BEETLE_OPT_CPU_DYNAREC_INFO_0_CY NULL
#define OPTION_VAL_DISABLED_CY NULL
#define OPTION_VAL_EXECUTE_CY NULL
#define OPTION_VAL_EXECUTE_ONE_CY NULL
#define OPTION_VAL_RUN_INTERPRETER_CY NULL
#define BEETLE_OPT_DYNAREC_INVALIDATE_LABEL_CY NULL
#define BEETLE_OPT_DYNAREC_INVALIDATE_INFO_0_CY NULL
#define OPTION_VAL_FULL_CY NULL
#define OPTION_VAL_DMA_CY NULL
#define BEETLE_OPT_DYNAREC_EVENTCYCLES_LABEL_CY NULL
#define BEETLE_OPT_DYNAREC_EVENTCYCLES_INFO_0_CY NULL
#define OPTION_VAL_128_CY NULL
#define BEETLE_OPT_DYNAREC_SPU_SAMPLES_LABEL_CY NULL
#define BEETLE_OPT_DYNAREC_SPU_SAMPLES_INFO_0_CY NULL
#define BEETLE_OPT_CORE_TIMING_FPS_LABEL_CY NULL
#define BEETLE_OPT_CORE_TIMING_FPS_INFO_0_CY NULL
#define OPTION_VAL_AUTO_TOGGLE_CY NULL
#define OPTION_VAL_FORCE_PROGRESSIVE_CY NULL
#define OPTION_VAL_FORCE_INTERLACED_CY NULL
#define BEETLE_OPT_DEINTERLACER_LABEL_CY NULL
#define BEETLE_OPT_DEINTERLACER_INFO_0_CY NULL
#define OPTION_VAL_WEAVE_CY NULL
#define OPTION_VAL_BOB_CY NULL
#define OPTION_VAL_BOB_OFFSET_CY NULL
#define OPTION_VAL_FASTMAD_CY NULL
#define BEETLE_OPT_FRAME_DUPING_LABEL_CY NULL
#define BEETLE_OPT_FRAME_DUPING_INFO_0_CY NULL
#define BEETLE_OPT_PAL_VIDEO_TIMING_OVERRIDE_LABEL_CY NULL
#define BEETLE_OPT_PAL_VIDEO_TIMING_OVERRIDE_INFO_0_CY NULL
#define BEETLE_OPT_ASPECT_RATIO_LABEL_CY NULL
#define BEETLE_OPT_ASPECT_RATIO_INFO_0_CY NULL
#define OPTION_VAL_CORRECTED_CY NULL
#define OPTION_VAL_UNCORRECTED_CY NULL
#define OPTION_VAL_4_3_CY NULL
#define OPTION_VAL_NTSC_CY NULL
#define BEETLE_OPT_CROP_OVERSCAN_LABEL_CY NULL
#define BEETLE_OPT_CROP_OVERSCAN_INFO_0_CY NULL
#define OPTION_VAL_STATIC_CY NULL
#define OPTION_VAL_SMART_CY NULL
#define BEETLE_OPT_IMAGE_CROP_LABEL_CY NULL
#define BEETLE_OPT_IMAGE_CROP_INFO_0_CY NULL
#define OPTION_VAL_9PX_CY NULL
#define OPTION_VAL_10PX_CY NULL
#define OPTION_VAL_11PX_CY NULL
#define OPTION_VAL_12PX_CY NULL
#define OPTION_VAL_13PX_CY NULL
#define OPTION_VAL_14PX_CY NULL
#define OPTION_VAL_15PX_CY NULL
#define OPTION_VAL_16PX_CY NULL
#define OPTION_VAL_17PX_CY NULL
#define OPTION_VAL_18PX_CY NULL
#define OPTION_VAL_19PX_CY NULL
#define OPTION_VAL_20PX_CY NULL
#define BEETLE_OPT_IMAGE_OFFSET_LABEL_CY NULL
#define BEETLE_OPT_IMAGE_OFFSET_INFO_0_CY NULL
#define OPTION_VAL_12PX_O64_CY NULL
#define OPTION_VAL_11PX_O64_CY NULL
#define OPTION_VAL_10PX_O64_CY NULL
#define OPTION_VAL_9PX_O64_CY NULL
#define OPTION_VAL_8PX_O64_CY NULL
#define OPTION_VAL_7PX_O64_CY NULL
#define OPTION_VAL_6PX_O64_CY NULL
#define OPTION_VAL_5PX_O64_CY NULL
#define OPTION_VAL_4PX_O64_CY NULL
#define OPTION_VAL_3PX_O64_CY NULL
#define OPTION_VAL_2PX_O64_CY NULL
#define OPTION_VAL_1PX_O64_CY NULL
#define OPTION_VAL_1PX_O64_0_CY NULL
#define OPTION_VAL_2PX_O64_0_CY NULL
#define OPTION_VAL_3PX_O64_0_CY NULL
#define OPTION_VAL_4PX_O64_0_CY NULL
#define OPTION_VAL_5PX_O64_0_CY NULL
#define OPTION_VAL_6PX_O64_0_CY NULL
#define OPTION_VAL_7PX_O64_0_CY NULL
#define OPTION_VAL_8PX_O64_0_CY NULL
#define OPTION_VAL_9PX_O64_0_CY NULL
#define OPTION_VAL_10PX_O64_0_CY NULL
#define OPTION_VAL_11PX_O64_0_CY NULL
#define OPTION_VAL_12PX_O64_0_CY NULL
#define BEETLE_OPT_IMAGE_OFFSET_CYCLES_LABEL_CY NULL
#define BEETLE_OPT_IMAGE_OFFSET_CYCLES_INFO_0_CY NULL
#define BEETLE_OPT_INITIAL_SCANLINE_LABEL_CY NULL
#define BEETLE_OPT_INITIAL_SCANLINE_INFO_0_CY NULL
#define BEETLE_OPT_LAST_SCANLINE_LABEL_CY NULL
#define BEETLE_OPT_LAST_SCANLINE_INFO_0_CY NULL
#define OPTION_VAL_239_CY NULL
#define BEETLE_OPT_INITIAL_SCANLINE_PAL_LABEL_CY NULL
#define BEETLE_OPT_INITIAL_SCANLINE_PAL_INFO_0_CY NULL
#define BEETLE_OPT_LAST_SCANLINE_PAL_LABEL_CY NULL
#define BEETLE_OPT_LAST_SCANLINE_PAL_INFO_0_CY NULL
#define OPTION_VAL_287_CY NULL

struct retro_core_option_v2_category option_cats_cy[] = {
   {
      "system",
      CATEGORY_SYSTEM_LABEL_CY,
      CATEGORY_SYSTEM_INFO_0_CY
   },
   {
      "video",
      CATEGORY_VIDEO_LABEL_CY,
      CATEGORY_VIDEO_INFO_0_CY
   },
   {
      "pgxp",
      CATEGORY_PGXP_LABEL_CY,
      CATEGORY_PGXP_INFO_0_CY
   },
   {
      "osd",
      CATEGORY_OSD_LABEL_CY,
      CATEGORY_OSD_INFO_0_CY
   },
   {
      "input",
      CATEGORY_INPUT_LABEL_CY,
      CATEGORY_INPUT_INFO_0_CY
   },
   {
      "memcards",
      CATEGORY_MEMCARDS_LABEL_CY,
      CATEGORY_MEMCARDS_INFO_0_CY
   },
   {
      "hacks",
      CATEGORY_HACKS_LABEL_CY,
      CATEGORY_HACKS_INFO_0_CY
   },
   { NULL, NULL, NULL },
};
struct retro_core_option_v2_definition option_defs_cy[] = {
   {
      BEETLE_OPT(cpu_freq_scale),
      BEETLE_OPT_CPU_FREQ_SCALE_LABEL_CY,
      NULL,
      BEETLE_OPT_CPU_FREQ_SCALE_INFO_0_CY,
      NULL,
      "system",
      {
         { "50%",           OPTION_VAL_50_CY },
         { "60%",           OPTION_VAL_60_CY },
         { "70%",           OPTION_VAL_70_CY },
         { "80%",           OPTION_VAL_80_CY },
         { "90%",           OPTION_VAL_90_CY },
         { "100%",          OPTION_VAL_100_CY },
         { "110%",          OPTION_VAL_110_CY },
         { "120%",          OPTION_VAL_120_CY },
         { "130%",          OPTION_VAL_130_CY },
         { "140%",          OPTION_VAL_140_CY },
         { "150%",          OPTION_VAL_150_CY },
         { "160%",          OPTION_VAL_160_CY },
         { "170%",          OPTION_VAL_170_CY },
         { "180%",          OPTION_VAL_180_CY },
         { "190%",          OPTION_VAL_190_CY },
         { "200%",          OPTION_VAL_200_CY },
         { "210%",          OPTION_VAL_210_CY },
         { "220%",          OPTION_VAL_220_CY },
         { "230%",          OPTION_VAL_230_CY },
         { "240%",          OPTION_VAL_240_CY },
         { "250%",          OPTION_VAL_250_CY },
         { "260%",          OPTION_VAL_260_CY },
         { "270%",          OPTION_VAL_270_CY },
         { "280%",          OPTION_VAL_280_CY },
         { "290%",          OPTION_VAL_290_CY },
         { "300%",          OPTION_VAL_300_CY },
         { "310%",          OPTION_VAL_310_CY },
         { "320%",          OPTION_VAL_320_CY },
         { "330%",          OPTION_VAL_330_CY },
         { "340%",          OPTION_VAL_340_CY },
         { "350%",          OPTION_VAL_350_CY },
         { "360%",          OPTION_VAL_360_CY },
         { "370%",          OPTION_VAL_370_CY },
         { "380%",          OPTION_VAL_380_CY },
         { "390%",          OPTION_VAL_390_CY },
         { "400%",          OPTION_VAL_400_CY },
         { "410%",          OPTION_VAL_410_CY },
         { "420%",          OPTION_VAL_420_CY },
         { "430%",          OPTION_VAL_430_CY },
         { "440%",          OPTION_VAL_440_CY },
         { "450%",          OPTION_VAL_450_CY },
         { "460%",          OPTION_VAL_460_CY },
         { "470%",          OPTION_VAL_470_CY },
         { "480%",          OPTION_VAL_480_CY },
         { "490%",          OPTION_VAL_490_CY },
         { "500%",          OPTION_VAL_500_CY },
         { "510%",          OPTION_VAL_510_CY },
         { "520%",          OPTION_VAL_520_CY },
         { "530%",          OPTION_VAL_530_CY },
         { "540%",          OPTION_VAL_540_CY },
         { "550%",          OPTION_VAL_550_CY },
         { "560%",          OPTION_VAL_560_CY },
         { "570%",          OPTION_VAL_570_CY },
         { "580%",          OPTION_VAL_580_CY },
         { "590%",          OPTION_VAL_590_CY },
         { "600%",          OPTION_VAL_600_CY },
         { "610%",          OPTION_VAL_610_CY },
         { "620%",          OPTION_VAL_620_CY },
         { "630%",          OPTION_VAL_630_CY },
         { "640%",          OPTION_VAL_640_CY },
         { "650%",          OPTION_VAL_650_CY },
         { "660%",          OPTION_VAL_660_CY },
         { "670%",          OPTION_VAL_670_CY },
         { "680%",          OPTION_VAL_680_CY },
         { "690%",          OPTION_VAL_690_CY },
         { "700%",          OPTION_VAL_700_CY },
         { "710%",          OPTION_VAL_710_CY },
         { "720%",          OPTION_VAL_720_CY },
         { "730%",          OPTION_VAL_730_CY },
         { "740%",          OPTION_VAL_740_CY },
         { "750%",          OPTION_VAL_750_CY },
         { NULL, NULL },
      },
      "100%"
   },
   {
      BEETLE_OPT(gpu_overclock),
      BEETLE_OPT_GPU_OVERCLOCK_LABEL_CY,
      NULL,
      BEETLE_OPT_GPU_OVERCLOCK_INFO_0_CY,
      NULL,
      "system",
      {
         { "1x(native)", OPTION_VAL_1X_NATIVE_CY },
         { "2x",         OPTION_VAL_2X_CY },
         { "4x",         OPTION_VAL_4X_CY },
         { "8x",         OPTION_VAL_8X_CY },
         { "16x",        OPTION_VAL_16X_CY },
         { "32x",        OPTION_VAL_32X_CY },
         { NULL, NULL },
      },
      "1x(native)"
   },
   {
      BEETLE_OPT(gte_overclock),
      BEETLE_OPT_GTE_OVERCLOCK_LABEL_CY,
      NULL,
      BEETLE_OPT_GTE_OVERCLOCK_INFO_0_CY,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
#ifndef EMSCRIPTEN
   {
      BEETLE_OPT(cd_access_method),
      BEETLE_OPT_CD_ACCESS_METHOD_LABEL_CY,
      NULL,
      BEETLE_OPT_CD_ACCESS_METHOD_INFO_0_CY,
      NULL,
      "system",
      {
         { "sync",     OPTION_VAL_SYNC_CY },
         { "async",    OPTION_VAL_ASYNC_CY },
         { "precache", OPTION_VAL_PRECACHE_CY },
         { NULL, NULL },
      },
      "sync"
   },
#endif
   {
      BEETLE_OPT(cd_fastload),
      BEETLE_OPT_CD_FASTLOAD_LABEL_CY,
      NULL,
      BEETLE_OPT_CD_FASTLOAD_INFO_0_CY,
      NULL,
      "system",
      {
         { "2x(native)", OPTION_VAL_2X_NATIVE_CY },
         { "4x",          OPTION_VAL_4X_CY },
         { "6x",          OPTION_VAL_6X_CY },
         { "8x",          OPTION_VAL_8X_CY },
         { "10x",         OPTION_VAL_10X_CY },
         { "12x",         OPTION_VAL_12X_CY },
         { "14x",         OPTION_VAL_14X_CY },
         { NULL, NULL },
      },
      "2x(native)"
   },
   {
      BEETLE_OPT(skip_bios),
      BEETLE_OPT_SKIP_BIOS_LABEL_CY,
      NULL,
      BEETLE_OPT_SKIP_BIOS_INFO_0_CY,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(override_bios),
      BEETLE_OPT_OVERRIDE_BIOS_LABEL_CY,
      NULL,
      BEETLE_OPT_OVERRIDE_BIOS_INFO_0_CY,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "psxonpsp", OPTION_VAL_PSXONPSP_CY },
         { "ps1_rom",  OPTION_VAL_PS1_ROM_CY },
         { "openbios", OPTION_VAL_OPENBIOS_CY     },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(region),
      BEETLE_OPT_REGION_LABEL_CY,
      NULL,
      BEETLE_OPT_REGION_INFO_0_CY,
      NULL,
      "system",
      {
         { "auto",   OPTION_VAL_AUTO_CY },
         { "ntsc-j", OPTION_VAL_NTSC_J_CY },
         { "ntsc-u", OPTION_VAL_NTSC_U_CY },
         { "pal",    OPTION_VAL_PAL_CY },
         { NULL, NULL },
      },
      "auto"
   },
   {
      BEETLE_OPT(internal_resolution),
      BEETLE_OPT_INTERNAL_RESOLUTION_LABEL_CY,
      NULL,
      BEETLE_OPT_INTERNAL_RESOLUTION_INFO_0_CY,
      NULL,
      "video",
      {
         { "1x(native)", OPTION_VAL_1X_NATIVE_CY },
         { "2x",         OPTION_VAL_2X_CY },
         { "4x",         OPTION_VAL_4X_CY },
         { "8x",         OPTION_VAL_8X_CY },
         { "16x",        OPTION_VAL_16X_CY },
         { NULL, NULL },
      },
      "1x(native)"
   },
#if defined(HAVE_OPENGL) || defined(HAVE_OPENGLES) || defined(HAVE_VULKAN)
   {
      BEETLE_OPT(renderer),
      BEETLE_OPT_RENDERER_LABEL_CY,
      NULL,
      BEETLE_OPT_RENDERER_INFO_0_CY,
      NULL,
      "video",
      {
         { "hardware",    OPTION_VAL_HARDWARE_CY },
#if defined(HAVE_OPENGL) || defined(HAVE_OPENGLES)
         { "hardware_gl", OPTION_VAL_HARDWARE_GL_CY },
#endif
#if defined(HAVE_VULKAN)
         { "hardware_vk", OPTION_VAL_HARDWARE_VK_CY },
#endif
         { "software",    OPTION_VAL_SOFTWARE_CY },
         { NULL, NULL },
      },
      "hardware"
   },
   {
      BEETLE_OPT(renderer_software_fb),
      BEETLE_OPT_RENDERER_SOFTWARE_FB_LABEL_CY,
      NULL,
      BEETLE_OPT_RENDERER_SOFTWARE_FB_INFO_0_CY,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled"
   },
#endif
#if defined(HAVE_OPENGL) || defined(HAVE_OPENGLES)
   {
      BEETLE_OPT(depth),
      BEETLE_OPT_DEPTH_LABEL_CY,
      NULL,
      BEETLE_OPT_DEPTH_INFO_0_CY,
      NULL,
      "video",
      {
         { "16bpp(native)", OPTION_VAL_16BPP_NATIVE_CY },
         { "32bpp",         OPTION_VAL_32BPP_CY },
         { NULL, NULL },
      },
      "16bpp(native)"
   },
   // Sort of, it's more like 15-bit high color and 24-bit true color for visible output. The alpha channel is used for mask bit. Vulkan renderer uses ABGR1555_555 for 31 bits internal? FMVs are always 24-bit on all renderers like original hardware (BGR888, no alpha)
#endif
   {
      BEETLE_OPT(dither_mode),
      BEETLE_OPT_DITHER_MODE_LABEL_CY,
      NULL,
      BEETLE_OPT_DITHER_MODE_INFO_0_CY,
      NULL,
      "video",
      {
         { "1x(native)",          OPTION_VAL_1X_NATIVE_CY },
         { "internal resolution", OPTION_VAL_INTERNAL_RESOLUTION_CY },
         { "disabled",            NULL },
         { NULL, NULL },
      },
      "1x(native)"
   },
#ifdef HAVE_VULKAN
   {
      BEETLE_OPT(scaled_uv_offset),
      BEETLE_OPT_SCALED_UV_OFFSET_LABEL_CY,
      NULL,
      BEETLE_OPT_SCALED_UV_OFFSET_INFO_0_CY,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled"
   },
#endif
#if defined(HAVE_OPENGL) || defined(HAVE_OPENGLES) || defined(HAVE_VULKAN)
   {
      BEETLE_OPT(filter),
      BEETLE_OPT_FILTER_LABEL_CY,
      NULL,
      BEETLE_OPT_FILTER_INFO_0_CY,
      NULL,
      "video",
      {
         { "nearest",  OPTION_VAL_NEAREST_CY },
         { "SABR",     OPTION_VAL_SABR_CY },
         { "xBR",      OPTION_VAL_XBR_CY },
         { "bilinear", OPTION_VAL_BILINEAR_CY },
         { "3-point",  OPTION_VAL_3_POINT_CY },
         { "JINC2",    OPTION_VAL_JINC2_CY },
         { NULL, NULL },
      },
      "nearest"
   },
#ifdef HAVE_VULKAN
   {
      BEETLE_OPT(filter_exclude_sprite),
      BEETLE_OPT_FILTER_EXCLUDE_SPRITE_LABEL_CY,
      NULL,
      BEETLE_OPT_FILTER_EXCLUDE_SPRITE_INFO_0_CY,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "opaque", OPTION_VAL_OPAQUE_CY },
         { "all", OPTION_VAL_ALL_CY },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(filter_exclude_2d_polygon),
      BEETLE_OPT_FILTER_EXCLUDE_2D_POLYGON_LABEL_CY,
      NULL,
      BEETLE_OPT_FILTER_EXCLUDE_2D_POLYGON_INFO_0_CY,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "opaque", OPTION_VAL_OPAQUE_CY },
         { "all", OPTION_VAL_ALL_CY },
         { NULL, NULL },
      },
      "disabled"
   },
#endif
#endif
#ifdef HAVE_VULKAN
   {
      BEETLE_OPT(adaptive_smoothing),
      BEETLE_OPT_ADAPTIVE_SMOOTHING_LABEL_CY,
      NULL,
      BEETLE_OPT_ADAPTIVE_SMOOTHING_INFO_0_CY,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(super_sampling),
      BEETLE_OPT_SUPER_SAMPLING_LABEL_CY,
      NULL,
      BEETLE_OPT_SUPER_SAMPLING_INFO_0_CY,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(msaa),
      BEETLE_OPT_MSAA_LABEL_CY,
      NULL,
      BEETLE_OPT_MSAA_INFO_0_CY,
      NULL,
      "video",
      {
         { "1x",  OPTION_VAL_1X_CY },
         { "2x",  OPTION_VAL_2X_CY },
         { "4x",  OPTION_VAL_4X_CY },
         { "8x",  OPTION_VAL_8X_CY },
         { "16x", OPTION_VAL_16X_CY },
         { NULL, NULL },
      },
      "1x"
   },
   {
      BEETLE_OPT(mdec_yuv),
      BEETLE_OPT_MDEC_YUV_LABEL_CY,
      NULL,
      BEETLE_OPT_MDEC_YUV_INFO_0_CY,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(track_textures),
      BEETLE_OPT_TRACK_TEXTURES_LABEL_CY,
      NULL,
      BEETLE_OPT_TRACK_TEXTURES_INFO_0_CY,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
#ifdef TEXTURE_DUMPING_ENABLED
   {
      BEETLE_OPT(dump_textures),
      BEETLE_OPT_DUMP_TEXTURES_LABEL_CY,
      NULL,
      BEETLE_OPT_DUMP_TEXTURES_INFO_0_CY,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
#endif
   {
      BEETLE_OPT(replace_textures),
      BEETLE_OPT_REPLACE_TEXTURES_LABEL_CY,
      NULL,
      BEETLE_OPT_REPLACE_TEXTURES_INFO_0_CY,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
#endif
   {
      BEETLE_OPT(pgxp_mode),
      BEETLE_OPT_PGXP_MODE_LABEL_CY,
      NULL,
      BEETLE_OPT_PGXP_MODE_INFO_0_CY,
      NULL,
      "pgxp",
      {
         { "disabled",     NULL },
         { "memory only",  OPTION_VAL_MEMORY_ONLY_CY },
         { "memory + CPU", OPTION_VAL_MEMORY_CPU_CY },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(pgxp_2d_tol),
      BEETLE_OPT_PGXP_2D_TOL_LABEL_CY,
      NULL,
      BEETLE_OPT_PGXP_2D_TOL_INFO_0_CY,
      NULL,
      "pgxp",
      {
         { "disabled", NULL },
         { "0px", OPTION_VAL_0PX_CY },
         { "1px", OPTION_VAL_1PX_CY },
         { "2px", OPTION_VAL_2PX_CY },
         { "3px", OPTION_VAL_3PX_CY },
         { "4px", OPTION_VAL_4PX_CY },
         { "5px", OPTION_VAL_5PX_CY },
         { "6px", OPTION_VAL_6PX_CY },
         { "7px", OPTION_VAL_7PX_CY },
         { "8px", OPTION_VAL_8PX_CY },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(pgxp_nclip),
      BEETLE_OPT_PGXP_NCLIP_LABEL_CY,
      NULL,
      BEETLE_OPT_PGXP_NCLIP_INFO_0_CY,
      NULL,
      "pgxp",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
#if defined(HAVE_OPENGL) || defined(HAVE_OPENGLES) || defined(HAVE_VULKAN)
   {
      BEETLE_OPT(pgxp_vertex),
      BEETLE_OPT_PGXP_VERTEX_LABEL_CY,
      NULL,
      BEETLE_OPT_PGXP_VERTEX_INFO_0_CY,
      NULL,
      "pgxp",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
#endif
   {
      BEETLE_OPT(pgxp_texture),
      BEETLE_OPT_PGXP_TEXTURE_LABEL_CY,
      NULL,
      BEETLE_OPT_PGXP_TEXTURE_INFO_0_CY,
      NULL,
      "pgxp",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(display_internal_fps),
      BEETLE_OPT_DISPLAY_INTERNAL_FPS_LABEL_CY,
      NULL,
      BEETLE_OPT_DISPLAY_INTERNAL_FPS_INFO_0_CY,
      NULL,
      "osd",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(display_osd),
      BEETLE_OPT_DISPLAY_OSD_LABEL_CY,
      NULL,
      BEETLE_OPT_DISPLAY_OSD_INFO_0_CY,
      NULL,
      "osd",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled"
   },
#if defined(HAVE_OPENGL) || defined(HAVE_OPENGLES) || defined(HAVE_VULKAN)
   {
      BEETLE_OPT(display_vram),
      BEETLE_OPT_DISPLAY_VRAM_LABEL_CY,
      NULL,
      BEETLE_OPT_DISPLAY_VRAM_INFO_0_CY,
      NULL,
      "osd",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
#endif
   {
      BEETLE_OPT(analog_calibration),
      BEETLE_OPT_ANALOG_CALIBRATION_LABEL_CY,
      NULL,
      BEETLE_OPT_ANALOG_CALIBRATION_INFO_0_CY,
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(analog_toggle),
      BEETLE_OPT_ANALOG_TOGGLE_LABEL_CY,
      NULL,
      BEETLE_OPT_ANALOG_TOGGLE_INFO_0_CY,
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { "enabled-analog", OPTION_VAL_ENABLED_ANALOG_CY },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(analog_toggle_combo),
      BEETLE_OPT_ANALOG_TOGGLE_COMBO_LABEL_CY,
      NULL,
      BEETLE_OPT_ANALOG_TOGGLE_COMBO_INFO_0_CY,
      NULL,
      "input",
      {
         { "l1+l2+r1+r2+start+select", OPTION_VAL_L1_L2_R1_R2_START_SELECT_CY },
         { "l1+r1+select",             OPTION_VAL_L1_R1_SELECT_CY },
         { "l1+r1+start",              OPTION_VAL_L1_R1_START_CY },
         { "l1+r1+l3",                 OPTION_VAL_L1_R1_L3_CY },
         { "l1+r1+r3",                 OPTION_VAL_L1_R1_R3_CY },
         { "l2+r2+select",             OPTION_VAL_L2_R2_SELECT_CY },
         { "l2+r2+start",              OPTION_VAL_L2_R2_START_CY },
         { "l2+r2+l3",                 OPTION_VAL_L2_R2_L3_CY },
         { "l2+r2+r3",                 OPTION_VAL_L2_R2_R3_CY },
         { "l3+r3",                    OPTION_VAL_L3_R3_CY },
         { NULL, NULL },
      },
      "l1+r1+select"
   },
   {
      BEETLE_OPT(analog_toggle_hold),
      BEETLE_OPT_ANALOG_TOGGLE_HOLD_LABEL_CY,
      NULL,
      BEETLE_OPT_ANALOG_TOGGLE_HOLD_INFO_0_CY,
      NULL,
      "input",
      {
         { "0", OPTION_VAL_0_CY },
         { "1", OPTION_VAL_1_CY },
         { "2", OPTION_VAL_2_CY },
         { "3", OPTION_VAL_3_CY },
         { "4", OPTION_VAL_4_CY },
         { "5", OPTION_VAL_5_CY },
         { NULL, NULL },
      },
      "1"
   },
   {
      BEETLE_OPT(enable_multitap_port1),
      BEETLE_OPT_ENABLE_MULTITAP_PORT1_LABEL_CY,
      NULL,
      BEETLE_OPT_ENABLE_MULTITAP_PORT1_INFO_0_CY,
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(enable_multitap_port2),
      BEETLE_OPT_ENABLE_MULTITAP_PORT2_LABEL_CY,
      NULL,
      BEETLE_OPT_ENABLE_MULTITAP_PORT2_INFO_0_CY,
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(gun_input_mode),
      BEETLE_OPT_GUN_INPUT_MODE_LABEL_CY,
      NULL,
      BEETLE_OPT_GUN_INPUT_MODE_INFO_0_CY,
      NULL,
      "input",
      {
         { "lightgun",    OPTION_VAL_LIGHTGUN_CY },
         { "touchscreen", OPTION_VAL_TOUCHSCREEN_CY },
         { NULL, NULL },
      },
      "lightgun"
   },
   // Shouldn't the gun_input_mode just be Mouse vs. Touchscreen?
   {
      BEETLE_OPT(gun_cursor),
      BEETLE_OPT_GUN_CURSOR_LABEL_CY,
      NULL,
      BEETLE_OPT_GUN_CURSOR_INFO_0_CY,
      NULL,
      "input",
      {
         { "off",   "disabled" },
         { "cross", OPTION_VAL_CROSS_CY },
         { "dot",   OPTION_VAL_DOT_CY },
         { NULL, NULL },
      },
      "cross"
   },
   {
      BEETLE_OPT(crosshair_color_p1),
      BEETLE_OPT_CROSSHAIR_COLOR_P1_LABEL_CY,
      NULL,
      BEETLE_OPT_CROSSHAIR_COLOR_P1_INFO_0_CY,
      NULL,
      "input",
      {
         { "red",    OPTION_VAL_RED_CY },
         { "blue",   OPTION_VAL_BLUE_CY },
         { "green",  OPTION_VAL_GREEN_CY },
         { "orange", OPTION_VAL_ORANGE_CY },
         { "yellow", OPTION_VAL_YELLOW_CY },
         { "cyan",   OPTION_VAL_CYAN_CY },
         { "pink",   OPTION_VAL_PINK_CY },
         { "purple", OPTION_VAL_PURPLE_CY },
         { "black",  OPTION_VAL_BLACK_CY },
         { "white",  OPTION_VAL_WHITE_CY },
         { NULL, NULL },
      },
      "red"
   },
   {
      BEETLE_OPT(crosshair_color_p2),
      BEETLE_OPT_CROSSHAIR_COLOR_P2_LABEL_CY,
      NULL,
      BEETLE_OPT_CROSSHAIR_COLOR_P2_INFO_0_CY,
      NULL,
      "input",
      {
         { "blue",   OPTION_VAL_BLUE_CY },
         { "red",    OPTION_VAL_RED_CY },
         { "green",  OPTION_VAL_GREEN_CY },
         { "orange", OPTION_VAL_ORANGE_CY },
         { "yellow", OPTION_VAL_YELLOW_CY },
         { "cyan",   OPTION_VAL_CYAN_CY },
         { "pink",   OPTION_VAL_PINK_CY },
         { "purple", OPTION_VAL_PURPLE_CY },
         { "black",  OPTION_VAL_BLACK_CY },
         { "white",  OPTION_VAL_WHITE_CY },
         { NULL, NULL },
      },
      "blue"
   },
   {
      BEETLE_OPT(mouse_sensitivity),
      BEETLE_OPT_MOUSE_SENSITIVITY_LABEL_CY,
      NULL,
      BEETLE_OPT_MOUSE_SENSITIVITY_INFO_0_CY,
      NULL,
      "input",
      {
         { "5%",   OPTION_VAL_5_O42_CY },
         { "10%",  OPTION_VAL_10_CY },
         { "15%",  OPTION_VAL_15_CY },
         { "20%",  OPTION_VAL_20_CY },
         { "25%",  OPTION_VAL_25_CY },
         { "30%",  OPTION_VAL_30_CY },
         { "35%",  OPTION_VAL_35_CY },
         { "40%",  OPTION_VAL_40_CY },
         { "45%",  OPTION_VAL_45_CY },
         { "50%",  OPTION_VAL_50_CY },
         { "55%",  OPTION_VAL_55_CY },
         { "60%",  OPTION_VAL_60_CY },
         { "65%",  OPTION_VAL_65_CY },
         { "70%",  OPTION_VAL_70_CY },
         { "75%",  OPTION_VAL_75_CY },
         { "80%",  OPTION_VAL_80_CY },
         { "85%",  OPTION_VAL_85_CY },
         { "90%",  OPTION_VAL_90_CY },
         { "95%",  OPTION_VAL_95_CY },
         { "100%", OPTION_VAL_100_O42_CY },
         { "105%", OPTION_VAL_105_CY },
         { "110%", OPTION_VAL_110_CY },
         { "115%", OPTION_VAL_115_CY },
         { "120%", OPTION_VAL_120_CY },
         { "125%", OPTION_VAL_125_CY },
         { "130%", OPTION_VAL_130_CY },
         { "135%", OPTION_VAL_135_CY },
         { "140%", OPTION_VAL_140_CY },
         { "145%", OPTION_VAL_145_CY },
         { "150%", OPTION_VAL_150_CY },
         { "155%", OPTION_VAL_155_CY },
         { "160%", OPTION_VAL_160_CY },
         { "165%", OPTION_VAL_165_CY },
         { "170%", OPTION_VAL_170_CY },
         { "175%", OPTION_VAL_175_CY },
         { "180%", OPTION_VAL_180_CY },
         { "185%", OPTION_VAL_185_CY },
         { "190%", OPTION_VAL_190_CY },
         { "195%", OPTION_VAL_195_CY },
         { "200%", OPTION_VAL_200_CY },
         { NULL, NULL },
      },
      "100%"
   },
   {
      BEETLE_OPT(negcon_response),
      BEETLE_OPT_NEGCON_RESPONSE_LABEL_CY,
      NULL,
      BEETLE_OPT_NEGCON_RESPONSE_INFO_0_CY,
      NULL,
      "input",
      {
         { "linear",    OPTION_VAL_LINEAR_CY },
         { "quadratic", OPTION_VAL_QUADRATIC_CY },
         { "cubic",     OPTION_VAL_CUBIC_CY },
         { NULL, NULL },
      },
      "linear"
   },
   {
      BEETLE_OPT(negcon_deadzone),
      BEETLE_OPT_NEGCON_DEADZONE_LABEL_CY,
      NULL,
      BEETLE_OPT_NEGCON_DEADZONE_INFO_0_CY,
      NULL,
      "input",
      {
         { "0%",  OPTION_VAL_0_O44_CY },
         { "5%",  OPTION_VAL_5_O42_CY },
         { "10%", OPTION_VAL_10_CY },
         { "15%", OPTION_VAL_15_CY },
         { "20%", OPTION_VAL_20_CY },
         { "25%", OPTION_VAL_25_CY },
         { "30%", OPTION_VAL_30_CY },
         { NULL, NULL },
      },
      "0%"
   },
   {
      BEETLE_OPT(use_mednafen_memcard0_method),
      BEETLE_OPT_USE_MEDNAFEN_MEMCARD0_METHOD_LABEL_CY,
      NULL,
      BEETLE_OPT_USE_MEDNAFEN_MEMCARD0_METHOD_INFO_0_CY,
      NULL,
      "memcards",
      {
         { "libretro", OPTION_VAL_LIBRETRO_CY },
         { "mednafen", OPTION_VAL_MEDNAFEN_CY },
         { NULL, NULL },
      },
      "libretro"
   },
   {
      BEETLE_OPT(enable_memcard1),
      BEETLE_OPT_ENABLE_MEMCARD1_LABEL_CY,
      NULL,
      BEETLE_OPT_ENABLE_MEMCARD1_INFO_0_CY,
      NULL,
      "memcards",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(shared_memory_cards),
      BEETLE_OPT_SHARED_MEMORY_CARDS_LABEL_CY,
      NULL,
      BEETLE_OPT_SHARED_MEMORY_CARDS_INFO_0_CY,
      NULL,
      "memcards",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(memcard_left_index),
      BEETLE_OPT_MEMCARD_LEFT_INDEX_LABEL_CY,
      NULL,
      BEETLE_OPT_MEMCARD_LEFT_INDEX_INFO_0_CY,
      NULL,
      "memcards",
      {
         { "0",  OPTION_VAL_0_O48_CY },
         { "1",  NULL },
         { "2",  NULL },
         { "3",  NULL },
         { "4",  NULL },
         { "5",  NULL },
         { "6",  NULL },
         { "7",  NULL },
         { "8",  NULL },
         { "9",  NULL },
         { "10", NULL },
         { "11", NULL },
         { "12", NULL },
         { "13", NULL },
         { "14", NULL },
         { "15", NULL },
         { "16", NULL },
         { "17", NULL },
         { "18", NULL },
         { "19", NULL },
         { "20", NULL },
         { "21", NULL },
         { "22", NULL },
         { "23", NULL },
         { "24", NULL },
         { "25", NULL },
         { "26", NULL },
         { "27", NULL },
         { "28", NULL },
         { "29", NULL },
         { "30", NULL },
         { "31", NULL },
         { "32", NULL },
         { "33", NULL },
         { "34", NULL },
         { "35", NULL },
         { "36", NULL },
         { "37", NULL },
         { "38", NULL },
         { "39", NULL },
         { "40", NULL },
         { "41", NULL },
         { "42", NULL },
         { "43", NULL },
         { "44", NULL },
         { "45", NULL },
         { "46", NULL },
         { "47", NULL },
         { "48", NULL },
         { "49", NULL },
         { "50", NULL },
         { "51", NULL },
         { "52", NULL },
         { "53", NULL },
         { "54", NULL },
         { "55", NULL },
         { "56", NULL },
         { "57", NULL },
         { "58", NULL },
         { "59", NULL },
         { "60", NULL },
         { "61", NULL },
         { "62", NULL },
         { "63", NULL },
         { NULL, NULL },
      },
      "0"
   },
   {
      BEETLE_OPT(memcard_right_index),
      BEETLE_OPT_MEMCARD_RIGHT_INDEX_LABEL_CY,
      NULL,
      BEETLE_OPT_MEMCARD_RIGHT_INDEX_INFO_0_CY,
      NULL,
      "memcards",
      {
         { "0",  NULL },
         { "1",  OPTION_VAL_1_O49_CY },
         { "2",  NULL },
         { "3",  NULL },
         { "4",  NULL },
         { "5",  NULL },
         { "6",  NULL },
         { "7",  NULL },
         { "8",  NULL },
         { "9",  NULL },
         { "10", NULL },
         { "11", NULL },
         { "12", NULL },
         { "13", NULL },
         { "14", NULL },
         { "15", NULL },
         { "16", NULL },
         { "17", NULL },
         { "18", NULL },
         { "19", NULL },
         { "20", NULL },
         { "21", NULL },
         { "22", NULL },
         { "23", NULL },
         { "24", NULL },
         { "25", NULL },
         { "26", NULL },
         { "27", NULL },
         { "28", NULL },
         { "29", NULL },
         { "30", NULL },
         { "31", NULL },
         { "32", NULL },
         { "33", NULL },
         { "34", NULL },
         { "35", NULL },
         { "36", NULL },
         { "37", NULL },
         { "38", NULL },
         { "39", NULL },
         { "40", NULL },
         { "41", NULL },
         { "42", NULL },
         { "43", NULL },
         { "44", NULL },
         { "45", NULL },
         { "46", NULL },
         { "47", NULL },
         { "48", NULL },
         { "49", NULL },
         { "50", NULL },
         { "51", NULL },
         { "52", NULL },
         { "53", NULL },
         { "54", NULL },
         { "55", NULL },
         { "56", NULL },
         { "57", NULL },
         { "58", NULL },
         { "59", NULL },
         { "60", NULL },
         { "61", NULL },
         { "62", NULL },
         { "63", NULL },
         { NULL, NULL },
      },
      "1"
   },
   {
      BEETLE_OPT(line_render),
      BEETLE_OPT_LINE_RENDER_LABEL_CY,
      NULL,
      BEETLE_OPT_LINE_RENDER_INFO_0_CY,
      NULL,
      "hacks",
      {
         { "default",    OPTION_VAL_DEFAULT_CY },
         { "aggressive", OPTION_VAL_AGGRESSIVE_CY },
         { "disabled",   NULL },
         { NULL, NULL },
      },
      "default"
   },
   {
      BEETLE_OPT(widescreen_hack),
      BEETLE_OPT_WIDESCREEN_HACK_LABEL_CY,
      NULL,
      BEETLE_OPT_WIDESCREEN_HACK_INFO_0_CY,
      NULL,
      "hacks",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(widescreen_hack_aspect_ratio),
      BEETLE_OPT_WIDESCREEN_HACK_ASPECT_RATIO_LABEL_CY,
      NULL,
      BEETLE_OPT_WIDESCREEN_HACK_ASPECT_RATIO_INFO_0_CY,
      NULL,
      "hacks",
      {
         { "16:9",  OPTION_VAL_16_9_CY },
         { "16:10", OPTION_VAL_16_10_CY },
         { "18:9",  OPTION_VAL_18_9_CY },
         { "19:9",  OPTION_VAL_19_9_CY },
         { "20:9",  OPTION_VAL_20_9_CY },
         { "21:9",  OPTION_VAL_21_9_CY }, // 64:27
         { "32:9",  OPTION_VAL_32_9_CY },
         { NULL,    NULL },
      },
      "16:9"
   },
#if defined(HAVE_LIGHTREC)
   {
      BEETLE_OPT(cpu_dynarec),
      BEETLE_OPT_CPU_DYNAREC_LABEL_CY,
      NULL,
      BEETLE_OPT_CPU_DYNAREC_INFO_0_CY,
      NULL,
      "hacks",
      {
         { "disabled",        OPTION_VAL_DISABLED_CY },
         { "execute",         OPTION_VAL_EXECUTE_CY },
         { "execute_one",     OPTION_VAL_EXECUTE_ONE_CY },
         { "run_interpreter", OPTION_VAL_RUN_INTERPRETER_CY },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(dynarec_invalidate),
      BEETLE_OPT_DYNAREC_INVALIDATE_LABEL_CY,
      NULL,
      BEETLE_OPT_DYNAREC_INVALIDATE_INFO_0_CY,
      NULL,
      "hacks",
      {
         { "full", OPTION_VAL_FULL_CY },
         { "dma",  OPTION_VAL_DMA_CY },
         { NULL, NULL },
      },
      "full"
   },
   {
      BEETLE_OPT(dynarec_eventcycles),
      BEETLE_OPT_DYNAREC_EVENTCYCLES_LABEL_CY,
      NULL,
      BEETLE_OPT_DYNAREC_EVENTCYCLES_INFO_0_CY,
      NULL,
      "hacks",
      {
         { "128",  OPTION_VAL_128_CY },
         { "256",  NULL },
         { "384",  NULL },
         { "512",  NULL },
         { "640",  NULL },
         { "768",  NULL },
         { "896",  NULL },
         { "1024", NULL },
         { "1152", NULL },
         { "1280", NULL },
         { "1408", NULL },
         { "1536", NULL },
         { "1664", NULL },
         { "1792", NULL },
         { "1920", NULL },
         { "2048", NULL },
         { NULL, NULL },
      },
      "128"
   },
   {
      BEETLE_OPT(dynarec_spu_samples),
      BEETLE_OPT_DYNAREC_SPU_SAMPLES_LABEL_CY,
      NULL,
      BEETLE_OPT_DYNAREC_SPU_SAMPLES_INFO_0_CY,
      NULL,
      "hacks",
      {
         { "1",  OPTION_VAL_1_O49_CY },
         { "4",  NULL },
         { "16", NULL },
         { NULL, NULL },
      },
      "1"
   },
#endif
   {
      BEETLE_OPT(core_timing_fps),
      BEETLE_OPT_CORE_TIMING_FPS_LABEL_CY,
      NULL,
      BEETLE_OPT_CORE_TIMING_FPS_INFO_0_CY,
      NULL,
      "video",
      {
         { "auto_toggle",       OPTION_VAL_AUTO_TOGGLE_CY },
         { "force_progressive", OPTION_VAL_FORCE_PROGRESSIVE_CY },
         { "force_interlaced",  OPTION_VAL_FORCE_INTERLACED_CY },
      },
      "auto_toggle"
   },
   {
      BEETLE_OPT(deinterlacer),
      BEETLE_OPT_DEINTERLACER_LABEL_CY,
      NULL,
      BEETLE_OPT_DEINTERLACER_INFO_0_CY,
      NULL,
      "video",
      {
         { "weave",      OPTION_VAL_WEAVE_CY },
         { "bob",        OPTION_VAL_BOB_CY },
         { "bob_offset", OPTION_VAL_BOB_OFFSET_CY },
         { "fastmad",    OPTION_VAL_FASTMAD_CY },
         { "off",        "Off" },
         { NULL, NULL },
      },
      "weave"
   },
   {
      BEETLE_OPT(frame_duping),
      BEETLE_OPT_FRAME_DUPING_LABEL_CY,
      NULL,
      BEETLE_OPT_FRAME_DUPING_INFO_0_CY,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(pal_video_timing_override),
      BEETLE_OPT_PAL_VIDEO_TIMING_OVERRIDE_LABEL_CY,
      NULL,
      BEETLE_OPT_PAL_VIDEO_TIMING_OVERRIDE_INFO_0_CY,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(aspect_ratio),
      BEETLE_OPT_ASPECT_RATIO_LABEL_CY,
      NULL,
      BEETLE_OPT_ASPECT_RATIO_INFO_0_CY,
      NULL,
      "video",
      {
         { "corrected", OPTION_VAL_CORRECTED_CY },
         { "uncorrected", OPTION_VAL_UNCORRECTED_CY },
         { "4:3",  OPTION_VAL_4_3_CY },
         { "ntsc", OPTION_VAL_NTSC_CY },
      },
      "corrected"
   },
   {
      BEETLE_OPT(crop_overscan),
      BEETLE_OPT_CROP_OVERSCAN_LABEL_CY,
      NULL,
      BEETLE_OPT_CROP_OVERSCAN_INFO_0_CY,
      NULL,
      "video",
      {
         { "disabled", "Disabled" },
         { "static",   OPTION_VAL_STATIC_CY },
         { "smart",    OPTION_VAL_SMART_CY },
         { NULL, NULL },
      },
      "smart"
   },
   {
      BEETLE_OPT(image_crop),
      BEETLE_OPT_IMAGE_CROP_LABEL_CY,
      NULL,
      BEETLE_OPT_IMAGE_CROP_INFO_0_CY,
      NULL,
      "video",
      {
         { "disabled", "0" },
         { "1px",      OPTION_VAL_1PX_CY },
         { "2px",      OPTION_VAL_2PX_CY },
         { "3px",      OPTION_VAL_3PX_CY },
         { "4px",      OPTION_VAL_4PX_CY },
         { "5px",      OPTION_VAL_5PX_CY },
         { "6px",      OPTION_VAL_6PX_CY },
         { "7px",      OPTION_VAL_7PX_CY },
         { "8px",      OPTION_VAL_8PX_CY },
         { "9px",      OPTION_VAL_9PX_CY },
         { "10px",     OPTION_VAL_10PX_CY },
         { "11px",     OPTION_VAL_11PX_CY },
         { "12px",     OPTION_VAL_12PX_CY },
         { "13px",     OPTION_VAL_13PX_CY },
         { "14px",     OPTION_VAL_14PX_CY },
         { "15px",     OPTION_VAL_15PX_CY },
         { "16px",     OPTION_VAL_16PX_CY },
         { "17px",     OPTION_VAL_17PX_CY },
         { "18px",     OPTION_VAL_18PX_CY },
         { "19px",     OPTION_VAL_19PX_CY },
         { "20px",     OPTION_VAL_20PX_CY },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(image_offset),
      BEETLE_OPT_IMAGE_OFFSET_LABEL_CY,
      NULL,
      BEETLE_OPT_IMAGE_OFFSET_INFO_0_CY,
      NULL,
      "video",
      {
         { "-12px",    OPTION_VAL_12PX_O64_CY },
         { "-11px",    OPTION_VAL_11PX_O64_CY },
         { "-10px",    OPTION_VAL_10PX_O64_CY },
         { "-9px",     OPTION_VAL_9PX_O64_CY },
         { "-8px",     OPTION_VAL_8PX_O64_CY },
         { "-7px",     OPTION_VAL_7PX_O64_CY },
         { "-6px",     OPTION_VAL_6PX_O64_CY },
         { "-5px",     OPTION_VAL_5PX_O64_CY },
         { "-4px",     OPTION_VAL_4PX_O64_CY },
         { "-3px",     OPTION_VAL_3PX_O64_CY },
         { "-2px",     OPTION_VAL_2PX_O64_CY },
         { "-1px",     OPTION_VAL_1PX_O64_CY },
         { "disabled", "0" },
         { "+1px",     OPTION_VAL_1PX_O64_0_CY },
         { "+2px",     OPTION_VAL_2PX_O64_0_CY },
         { "+3px",     OPTION_VAL_3PX_O64_0_CY },
         { "+4px",     OPTION_VAL_4PX_O64_0_CY },
         { "+5px",     OPTION_VAL_5PX_O64_0_CY },
         { "+6px",     OPTION_VAL_6PX_O64_0_CY },
         { "+7px",     OPTION_VAL_7PX_O64_0_CY },
         { "+8px",     OPTION_VAL_8PX_O64_0_CY },
         { "+9px",     OPTION_VAL_9PX_O64_0_CY },
         { "+10px",    OPTION_VAL_10PX_O64_0_CY },
         { "+11px",    OPTION_VAL_11PX_O64_0_CY },
         { "+12px",    OPTION_VAL_12PX_O64_0_CY },
         { NULL, NULL },
      },
      "disabled"
   },
#if defined(HAVE_OPENGL) || defined(HAVE_OPENGLES) || defined(HAVE_VULKAN)
   {
      BEETLE_OPT(image_offset_cycles),
      BEETLE_OPT_IMAGE_OFFSET_CYCLES_LABEL_CY,
      NULL,
      BEETLE_OPT_IMAGE_OFFSET_CYCLES_INFO_0_CY,
      NULL,
      "video",
      {
         { "-40",      NULL },
         { "-39",      NULL },
         { "-38",      NULL },
         { "-37",      NULL },
         { "-36",      NULL },
         { "-35",      NULL },
         { "-34",      NULL },
         { "-33",      NULL },
         { "-32",      NULL },
         { "-31",      NULL },
         { "-30",      NULL },
         { "-29",      NULL },
         { "-28",      NULL },
         { "-27",      NULL },
         { "-26",      NULL },
         { "-25",      NULL },
         { "-24",      NULL },
         { "-23",      NULL },
         { "-22",      NULL },
         { "-21",      NULL },
         { "-20",      NULL },
         { "-19",      NULL },
         { "-18",      NULL },
         { "-17",      NULL },
         { "-16",      NULL },
         { "-15",      NULL },
         { "-14",      NULL },
         { "-13",      NULL },
         { "-12",      NULL },
         { "-11",      NULL },
         { "-10",      NULL },
         { "-9",       NULL },
         { "-8",       NULL },
         { "-7",       NULL },
         { "-6",       NULL },
         { "-5",       NULL },
         { "-4",       NULL },
         { "-3",       NULL },
         { "-2",       NULL },
         { "-1",       NULL },
         { "0",        "0" },
         { "+1",       NULL },
         { "+2",       NULL },
         { "+3",       NULL },
         { "+4",       NULL },
         { "+5",       NULL },
         { "+6",       NULL },
         { "+7",       NULL },
         { "+8",       NULL },
         { "+9",       NULL },
         { "+10",      NULL },
         { "+11",      NULL },
         { "+12",      NULL },
         { "+13",      NULL },
         { "+14",      NULL },
         { "+15",      NULL },
         { "+16",      NULL },
         { "+17",      NULL },
         { "+18",      NULL },
         { "+19",      NULL },
         { "+20",      NULL },
         { "+21",      NULL },
         { "+22",      NULL },
         { "+23",      NULL },
         { "+24",      NULL },
         { "+25",      NULL },
         { "+26",      NULL },
         { "+27",      NULL },
         { "+28",      NULL },
         { "+29",      NULL },
         { "+30",      NULL },
         { "+31",      NULL },
         { "+32",      NULL },
         { "+33",      NULL },
         { "+34",      NULL },
         { "+35",      NULL },
         { "+36",      NULL },
         { "+37",      NULL },
         { "+38",      NULL },
         { "+39",      NULL },
         { "+40",      NULL },
         { NULL, NULL},
      },
      "0"
   },
#endif
   {
      BEETLE_OPT(initial_scanline),
      BEETLE_OPT_INITIAL_SCANLINE_LABEL_CY,
      NULL,
      BEETLE_OPT_INITIAL_SCANLINE_INFO_0_CY,
      NULL,
      "video",
      {
         { "0",  OPTION_VAL_0_O48_CY },
         { "1",  NULL },
         { "2",  NULL },
         { "3",  NULL },
         { "4",  NULL },
         { "5",  NULL },
         { "6",  NULL },
         { "7",  NULL },
         { "8",  NULL },
         { "9",  NULL },
         { "10", NULL },
         { "11", NULL },
         { "12", NULL },
         { "13", NULL },
         { "14", NULL },
         { "15", NULL },
         { "16", NULL },
         { "17", NULL },
         { "18", NULL },
         { "19", NULL },
         { "20", NULL },
         { "21", NULL },
         { "22", NULL },
         { "23", NULL },
         { "24", NULL },
         { "25", NULL },
         { "26", NULL },
         { "27", NULL },
         { "28", NULL },
         { "29", NULL },
         { "30", NULL },
         { "31", NULL },
         { "32", NULL },
         { "33", NULL },
         { "34", NULL },
         { "35", NULL },
         { "36", NULL },
         { "37", NULL },
         { "38", NULL },
         { "39", NULL },
         { "40", NULL },
         { NULL, NULL },
      },
      "0"
   },
   {
      BEETLE_OPT(last_scanline),
      BEETLE_OPT_LAST_SCANLINE_LABEL_CY,
      NULL,
      BEETLE_OPT_LAST_SCANLINE_INFO_0_CY,
      NULL,
      "video",
      {
         { "210", NULL },
         { "211", NULL },
         { "212", NULL },
         { "213", NULL },
         { "214", NULL },
         { "215", NULL },
         { "216", NULL },
         { "217", NULL },
         { "218", NULL },
         { "219", NULL },
         { "220", NULL },
         { "221", NULL },
         { "222", NULL },
         { "223", NULL },
         { "224", NULL },
         { "225", NULL },
         { "226", NULL },
         { "227", NULL },
         { "228", NULL },
         { "229", NULL },
         { "230", NULL },
         { "231", NULL },
         { "232", NULL },
         { "233", NULL },
         { "234", NULL },
         { "235", NULL },
         { "236", NULL },
         { "237", NULL },
         { "238", NULL },
         { "239", OPTION_VAL_239_CY },
         { NULL, NULL },
      },
      "239"
   },
   {
      BEETLE_OPT(initial_scanline_pal),
      BEETLE_OPT_INITIAL_SCANLINE_PAL_LABEL_CY,
      NULL,
      BEETLE_OPT_INITIAL_SCANLINE_PAL_INFO_0_CY,
      NULL,
      "video",
      {
         { "0",  OPTION_VAL_0_O48_CY },
         { "1",  NULL },
         { "2",  NULL },
         { "3",  NULL },
         { "4",  NULL },
         { "5",  NULL },
         { "6",  NULL },
         { "7",  NULL },
         { "8",  NULL },
         { "9",  NULL },
         { "10", NULL },
         { "11", NULL },
         { "12", NULL },
         { "13", NULL },
         { "14", NULL },
         { "15", NULL },
         { "16", NULL },
         { "17", NULL },
         { "18", NULL },
         { "19", NULL },
         { "20", NULL },
         { "21", NULL },
         { "22", NULL },
         { "23", NULL },
         { "24", NULL },
         { "25", NULL },
         { "26", NULL },
         { "27", NULL },
         { "28", NULL },
         { "29", NULL },
         { "30", NULL },
         { "31", NULL },
         { "32", NULL },
         { "33", NULL },
         { "34", NULL },
         { "35", NULL },
         { "36", NULL },
         { "37", NULL },
         { "38", NULL },
         { "39", NULL },
         { "40", NULL },
         { NULL, NULL },
      },
      "0"
   },
   {
      BEETLE_OPT(last_scanline_pal),
      BEETLE_OPT_LAST_SCANLINE_PAL_LABEL_CY,
      NULL,
      BEETLE_OPT_LAST_SCANLINE_PAL_INFO_0_CY,
      NULL,
      "video",
      {
         { "230", NULL },
         { "231", NULL },
         { "232", NULL },
         { "233", NULL },
         { "234", NULL },
         { "235", NULL },
         { "236", NULL },
         { "237", NULL },
         { "238", NULL },
         { "239", NULL },
         { "240", NULL },
         { "241", NULL },
         { "242", NULL },
         { "243", NULL },
         { "244", NULL },
         { "245", NULL },
         { "246", NULL },
         { "247", NULL },
         { "248", NULL },
         { "249", NULL },
         { "250", NULL },
         { "251", NULL },
         { "252", NULL },
         { "253", NULL },
         { "254", NULL },
         { "255", NULL },
         { "256", NULL },
         { "257", NULL },
         { "258", NULL },
         { "259", NULL },
         { "260", NULL },
         { "261", NULL },
         { "262", NULL },
         { "263", NULL },
         { "264", NULL },
         { "265", NULL },
         { "266", NULL },
         { "267", NULL },
         { "268", NULL },
         { "269", NULL },
         { "270", NULL },
         { "271", NULL },
         { "272", NULL },
         { "273", NULL },
         { "274", NULL },
         { "275", NULL },
         { "276", NULL },
         { "277", NULL },
         { "278", NULL },
         { "279", NULL },
         { "280", NULL },
         { "281", NULL },
         { "282", NULL },
         { "283", NULL },
         { "284", NULL },
         { "285", NULL },
         { "286", NULL },
         { "287", OPTION_VAL_287_CY },
         { NULL, NULL },
      },
      "287"
   },
   { NULL, NULL, NULL, NULL, NULL, NULL, {{0}}, NULL },
};
struct retro_core_options_v2 options_cy = {
   option_cats_cy,
   option_defs_cy
};

/* RETRO_LANGUAGE_DA */

#define CATEGORY_SYSTEM_LABEL_DA NULL
#define CATEGORY_SYSTEM_INFO_0_DA NULL
#define CATEGORY_VIDEO_LABEL_DA NULL
#define CATEGORY_VIDEO_INFO_0_DA NULL
#define CATEGORY_PGXP_LABEL_DA NULL
#define CATEGORY_PGXP_INFO_0_DA NULL
#define CATEGORY_OSD_LABEL_DA NULL
#define CATEGORY_OSD_INFO_0_DA NULL
#define CATEGORY_INPUT_LABEL_DA NULL
#define CATEGORY_INPUT_INFO_0_DA NULL
#define CATEGORY_MEMCARDS_LABEL_DA NULL
#define CATEGORY_MEMCARDS_INFO_0_DA NULL
#define CATEGORY_HACKS_LABEL_DA NULL
#define CATEGORY_HACKS_INFO_0_DA NULL
#define BEETLE_OPT_CPU_FREQ_SCALE_LABEL_DA NULL
#define BEETLE_OPT_CPU_FREQ_SCALE_INFO_0_DA NULL
#define OPTION_VAL_50_DA NULL
#define OPTION_VAL_60_DA NULL
#define OPTION_VAL_70_DA NULL
#define OPTION_VAL_80_DA NULL
#define OPTION_VAL_90_DA NULL
#define OPTION_VAL_100_DA NULL
#define OPTION_VAL_110_DA NULL
#define OPTION_VAL_120_DA NULL
#define OPTION_VAL_130_DA NULL
#define OPTION_VAL_140_DA NULL
#define OPTION_VAL_150_DA NULL
#define OPTION_VAL_160_DA NULL
#define OPTION_VAL_170_DA NULL
#define OPTION_VAL_180_DA NULL
#define OPTION_VAL_190_DA NULL
#define OPTION_VAL_200_DA NULL
#define OPTION_VAL_210_DA NULL
#define OPTION_VAL_220_DA NULL
#define OPTION_VAL_230_DA NULL
#define OPTION_VAL_240_DA NULL
#define OPTION_VAL_250_DA NULL
#define OPTION_VAL_260_DA NULL
#define OPTION_VAL_270_DA NULL
#define OPTION_VAL_280_DA NULL
#define OPTION_VAL_290_DA NULL
#define OPTION_VAL_300_DA NULL
#define OPTION_VAL_310_DA NULL
#define OPTION_VAL_320_DA NULL
#define OPTION_VAL_330_DA NULL
#define OPTION_VAL_340_DA NULL
#define OPTION_VAL_350_DA NULL
#define OPTION_VAL_360_DA NULL
#define OPTION_VAL_370_DA NULL
#define OPTION_VAL_380_DA NULL
#define OPTION_VAL_390_DA NULL
#define OPTION_VAL_400_DA NULL
#define OPTION_VAL_410_DA NULL
#define OPTION_VAL_420_DA NULL
#define OPTION_VAL_430_DA NULL
#define OPTION_VAL_440_DA NULL
#define OPTION_VAL_450_DA NULL
#define OPTION_VAL_460_DA NULL
#define OPTION_VAL_470_DA NULL
#define OPTION_VAL_480_DA NULL
#define OPTION_VAL_490_DA NULL
#define OPTION_VAL_500_DA NULL
#define OPTION_VAL_510_DA NULL
#define OPTION_VAL_520_DA NULL
#define OPTION_VAL_530_DA NULL
#define OPTION_VAL_540_DA NULL
#define OPTION_VAL_550_DA NULL
#define OPTION_VAL_560_DA NULL
#define OPTION_VAL_570_DA NULL
#define OPTION_VAL_580_DA NULL
#define OPTION_VAL_590_DA NULL
#define OPTION_VAL_600_DA NULL
#define OPTION_VAL_610_DA NULL
#define OPTION_VAL_620_DA NULL
#define OPTION_VAL_630_DA NULL
#define OPTION_VAL_640_DA NULL
#define OPTION_VAL_650_DA NULL
#define OPTION_VAL_660_DA NULL
#define OPTION_VAL_670_DA NULL
#define OPTION_VAL_680_DA NULL
#define OPTION_VAL_690_DA NULL
#define OPTION_VAL_700_DA NULL
#define OPTION_VAL_710_DA NULL
#define OPTION_VAL_720_DA NULL
#define OPTION_VAL_730_DA NULL
#define OPTION_VAL_740_DA NULL
#define OPTION_VAL_750_DA NULL
#define BEETLE_OPT_GPU_OVERCLOCK_LABEL_DA NULL
#define BEETLE_OPT_GPU_OVERCLOCK_INFO_0_DA NULL
#define OPTION_VAL_1X_NATIVE_DA NULL
#define OPTION_VAL_2X_DA NULL
#define OPTION_VAL_4X_DA NULL
#define OPTION_VAL_8X_DA NULL
#define OPTION_VAL_16X_DA NULL
#define OPTION_VAL_32X_DA NULL
#define BEETLE_OPT_GTE_OVERCLOCK_LABEL_DA NULL
#define BEETLE_OPT_GTE_OVERCLOCK_INFO_0_DA NULL
#define BEETLE_OPT_CD_ACCESS_METHOD_LABEL_DA NULL
#define BEETLE_OPT_CD_ACCESS_METHOD_INFO_0_DA NULL
#define OPTION_VAL_SYNC_DA NULL
#define OPTION_VAL_ASYNC_DA NULL
#define OPTION_VAL_PRECACHE_DA NULL
#define BEETLE_OPT_CD_FASTLOAD_LABEL_DA NULL
#define BEETLE_OPT_CD_FASTLOAD_INFO_0_DA NULL
#define OPTION_VAL_2X_NATIVE_DA NULL
#define OPTION_VAL_6X_DA NULL
#define OPTION_VAL_10X_DA NULL
#define OPTION_VAL_12X_DA NULL
#define OPTION_VAL_14X_DA NULL
#define BEETLE_OPT_SKIP_BIOS_LABEL_DA NULL
#define BEETLE_OPT_SKIP_BIOS_INFO_0_DA NULL
#define BEETLE_OPT_OVERRIDE_BIOS_LABEL_DA NULL
#define BEETLE_OPT_OVERRIDE_BIOS_INFO_0_DA NULL
#define OPTION_VAL_PSXONPSP_DA NULL
#define OPTION_VAL_PS1_ROM_DA NULL
#define OPTION_VAL_OPENBIOS_DA NULL
#define BEETLE_OPT_REGION_LABEL_DA NULL
#define BEETLE_OPT_REGION_INFO_0_DA NULL
#define OPTION_VAL_AUTO_DA NULL
#define OPTION_VAL_NTSC_J_DA NULL
#define OPTION_VAL_NTSC_U_DA NULL
#define OPTION_VAL_PAL_DA NULL
#define BEETLE_OPT_INTERNAL_RESOLUTION_LABEL_DA NULL
#define BEETLE_OPT_INTERNAL_RESOLUTION_INFO_0_DA NULL
#define BEETLE_OPT_RENDERER_LABEL_DA NULL
#define BEETLE_OPT_RENDERER_INFO_0_DA NULL
#define OPTION_VAL_HARDWARE_DA NULL
#define OPTION_VAL_HARDWARE_GL_DA NULL
#define OPTION_VAL_HARDWARE_VK_DA NULL
#define OPTION_VAL_SOFTWARE_DA NULL
#define BEETLE_OPT_RENDERER_SOFTWARE_FB_LABEL_DA NULL
#define BEETLE_OPT_RENDERER_SOFTWARE_FB_INFO_0_DA NULL
#define BEETLE_OPT_DEPTH_LABEL_DA NULL
#define BEETLE_OPT_DEPTH_INFO_0_DA NULL
#define OPTION_VAL_16BPP_NATIVE_DA NULL
#define OPTION_VAL_32BPP_DA NULL
#define BEETLE_OPT_DITHER_MODE_LABEL_DA NULL
#define BEETLE_OPT_DITHER_MODE_INFO_0_DA NULL
#define OPTION_VAL_INTERNAL_RESOLUTION_DA NULL
#define BEETLE_OPT_SCALED_UV_OFFSET_LABEL_DA NULL
#define BEETLE_OPT_SCALED_UV_OFFSET_INFO_0_DA NULL
#define BEETLE_OPT_FILTER_LABEL_DA NULL
#define BEETLE_OPT_FILTER_INFO_0_DA NULL
#define OPTION_VAL_NEAREST_DA NULL
#define OPTION_VAL_SABR_DA NULL
#define OPTION_VAL_XBR_DA NULL
#define OPTION_VAL_BILINEAR_DA NULL
#define OPTION_VAL_3_POINT_DA NULL
#define OPTION_VAL_JINC2_DA NULL
#define BEETLE_OPT_FILTER_EXCLUDE_SPRITE_LABEL_DA NULL
#define BEETLE_OPT_FILTER_EXCLUDE_SPRITE_INFO_0_DA NULL
#define OPTION_VAL_OPAQUE_DA NULL
#define OPTION_VAL_ALL_DA NULL
#define BEETLE_OPT_FILTER_EXCLUDE_2D_POLYGON_LABEL_DA NULL
#define BEETLE_OPT_FILTER_EXCLUDE_2D_POLYGON_INFO_0_DA NULL
#define BEETLE_OPT_ADAPTIVE_SMOOTHING_LABEL_DA NULL
#define BEETLE_OPT_ADAPTIVE_SMOOTHING_INFO_0_DA NULL
#define BEETLE_OPT_SUPER_SAMPLING_LABEL_DA NULL
#define BEETLE_OPT_SUPER_SAMPLING_INFO_0_DA NULL
#define BEETLE_OPT_MSAA_LABEL_DA NULL
#define BEETLE_OPT_MSAA_INFO_0_DA NULL
#define OPTION_VAL_1X_DA NULL
#define BEETLE_OPT_MDEC_YUV_LABEL_DA NULL
#define BEETLE_OPT_MDEC_YUV_INFO_0_DA NULL
#define BEETLE_OPT_TRACK_TEXTURES_LABEL_DA NULL
#define BEETLE_OPT_TRACK_TEXTURES_INFO_0_DA NULL
#define BEETLE_OPT_DUMP_TEXTURES_LABEL_DA NULL
#define BEETLE_OPT_DUMP_TEXTURES_INFO_0_DA NULL
#define BEETLE_OPT_REPLACE_TEXTURES_LABEL_DA NULL
#define BEETLE_OPT_REPLACE_TEXTURES_INFO_0_DA NULL
#define BEETLE_OPT_PGXP_MODE_LABEL_DA NULL
#define BEETLE_OPT_PGXP_MODE_INFO_0_DA NULL
#define OPTION_VAL_MEMORY_ONLY_DA NULL
#define OPTION_VAL_MEMORY_CPU_DA NULL
#define BEETLE_OPT_PGXP_2D_TOL_LABEL_DA NULL
#define BEETLE_OPT_PGXP_2D_TOL_INFO_0_DA NULL
#define OPTION_VAL_0PX_DA NULL
#define OPTION_VAL_1PX_DA NULL
#define OPTION_VAL_2PX_DA NULL
#define OPTION_VAL_3PX_DA NULL
#define OPTION_VAL_4PX_DA NULL
#define OPTION_VAL_5PX_DA NULL
#define OPTION_VAL_6PX_DA NULL
#define OPTION_VAL_7PX_DA NULL
#define OPTION_VAL_8PX_DA NULL
#define BEETLE_OPT_PGXP_NCLIP_LABEL_DA NULL
#define BEETLE_OPT_PGXP_NCLIP_INFO_0_DA NULL
#define BEETLE_OPT_PGXP_VERTEX_LABEL_DA NULL
#define BEETLE_OPT_PGXP_VERTEX_INFO_0_DA NULL
#define BEETLE_OPT_PGXP_TEXTURE_LABEL_DA NULL
#define BEETLE_OPT_PGXP_TEXTURE_INFO_0_DA NULL
#define BEETLE_OPT_DISPLAY_INTERNAL_FPS_LABEL_DA NULL
#define BEETLE_OPT_DISPLAY_INTERNAL_FPS_INFO_0_DA NULL
#define BEETLE_OPT_DISPLAY_OSD_LABEL_DA NULL
#define BEETLE_OPT_DISPLAY_OSD_INFO_0_DA NULL
#define BEETLE_OPT_DISPLAY_VRAM_LABEL_DA NULL
#define BEETLE_OPT_DISPLAY_VRAM_INFO_0_DA NULL
#define BEETLE_OPT_ANALOG_CALIBRATION_LABEL_DA NULL
#define BEETLE_OPT_ANALOG_CALIBRATION_INFO_0_DA NULL
#define BEETLE_OPT_ANALOG_TOGGLE_LABEL_DA NULL
#define BEETLE_OPT_ANALOG_TOGGLE_INFO_0_DA NULL
#define OPTION_VAL_ENABLED_ANALOG_DA NULL
#define BEETLE_OPT_ANALOG_TOGGLE_COMBO_LABEL_DA NULL
#define BEETLE_OPT_ANALOG_TOGGLE_COMBO_INFO_0_DA NULL
#define OPTION_VAL_L1_L2_R1_R2_START_SELECT_DA NULL
#define OPTION_VAL_L1_R1_SELECT_DA NULL
#define OPTION_VAL_L1_R1_START_DA NULL
#define OPTION_VAL_L1_R1_L3_DA NULL
#define OPTION_VAL_L1_R1_R3_DA NULL
#define OPTION_VAL_L2_R2_SELECT_DA NULL
#define OPTION_VAL_L2_R2_START_DA NULL
#define OPTION_VAL_L2_R2_L3_DA NULL
#define OPTION_VAL_L2_R2_R3_DA NULL
#define OPTION_VAL_L3_R3_DA NULL
#define BEETLE_OPT_ANALOG_TOGGLE_HOLD_LABEL_DA NULL
#define BEETLE_OPT_ANALOG_TOGGLE_HOLD_INFO_0_DA NULL
#define OPTION_VAL_0_DA NULL
#define OPTION_VAL_1_DA NULL
#define OPTION_VAL_2_DA NULL
#define OPTION_VAL_3_DA NULL
#define OPTION_VAL_4_DA NULL
#define OPTION_VAL_5_DA NULL
#define BEETLE_OPT_ENABLE_MULTITAP_PORT1_LABEL_DA NULL
#define BEETLE_OPT_ENABLE_MULTITAP_PORT1_INFO_0_DA NULL
#define BEETLE_OPT_ENABLE_MULTITAP_PORT2_LABEL_DA NULL
#define BEETLE_OPT_ENABLE_MULTITAP_PORT2_INFO_0_DA NULL
#define BEETLE_OPT_GUN_INPUT_MODE_LABEL_DA NULL
#define BEETLE_OPT_GUN_INPUT_MODE_INFO_0_DA NULL
#define OPTION_VAL_LIGHTGUN_DA NULL
#define OPTION_VAL_TOUCHSCREEN_DA NULL
#define BEETLE_OPT_GUN_CURSOR_LABEL_DA NULL
#define BEETLE_OPT_GUN_CURSOR_INFO_0_DA NULL
#define OPTION_VAL_CROSS_DA NULL
#define OPTION_VAL_DOT_DA NULL
#define BEETLE_OPT_CROSSHAIR_COLOR_P1_LABEL_DA NULL
#define BEETLE_OPT_CROSSHAIR_COLOR_P1_INFO_0_DA NULL
#define OPTION_VAL_RED_DA NULL
#define OPTION_VAL_BLUE_DA NULL
#define OPTION_VAL_GREEN_DA NULL
#define OPTION_VAL_ORANGE_DA NULL
#define OPTION_VAL_YELLOW_DA NULL
#define OPTION_VAL_CYAN_DA NULL
#define OPTION_VAL_PINK_DA NULL
#define OPTION_VAL_PURPLE_DA NULL
#define OPTION_VAL_BLACK_DA NULL
#define OPTION_VAL_WHITE_DA NULL
#define BEETLE_OPT_CROSSHAIR_COLOR_P2_LABEL_DA NULL
#define BEETLE_OPT_CROSSHAIR_COLOR_P2_INFO_0_DA NULL
#define BEETLE_OPT_MOUSE_SENSITIVITY_LABEL_DA NULL
#define BEETLE_OPT_MOUSE_SENSITIVITY_INFO_0_DA NULL
#define OPTION_VAL_5_O42_DA NULL
#define OPTION_VAL_10_DA NULL
#define OPTION_VAL_15_DA NULL
#define OPTION_VAL_20_DA NULL
#define OPTION_VAL_25_DA NULL
#define OPTION_VAL_30_DA NULL
#define OPTION_VAL_35_DA NULL
#define OPTION_VAL_40_DA NULL
#define OPTION_VAL_45_DA NULL
#define OPTION_VAL_55_DA NULL
#define OPTION_VAL_65_DA NULL
#define OPTION_VAL_75_DA NULL
#define OPTION_VAL_85_DA NULL
#define OPTION_VAL_95_DA NULL
#define OPTION_VAL_100_O42_DA NULL
#define OPTION_VAL_105_DA NULL
#define OPTION_VAL_115_DA NULL
#define OPTION_VAL_125_DA NULL
#define OPTION_VAL_135_DA NULL
#define OPTION_VAL_145_DA NULL
#define OPTION_VAL_155_DA NULL
#define OPTION_VAL_165_DA NULL
#define OPTION_VAL_175_DA NULL
#define OPTION_VAL_185_DA NULL
#define OPTION_VAL_195_DA NULL
#define BEETLE_OPT_NEGCON_RESPONSE_LABEL_DA NULL
#define BEETLE_OPT_NEGCON_RESPONSE_INFO_0_DA NULL
#define OPTION_VAL_LINEAR_DA NULL
#define OPTION_VAL_QUADRATIC_DA NULL
#define OPTION_VAL_CUBIC_DA NULL
#define BEETLE_OPT_NEGCON_DEADZONE_LABEL_DA NULL
#define BEETLE_OPT_NEGCON_DEADZONE_INFO_0_DA NULL
#define OPTION_VAL_0_O44_DA NULL
#define BEETLE_OPT_USE_MEDNAFEN_MEMCARD0_METHOD_LABEL_DA NULL
#define BEETLE_OPT_USE_MEDNAFEN_MEMCARD0_METHOD_INFO_0_DA NULL
#define OPTION_VAL_LIBRETRO_DA NULL
#define OPTION_VAL_MEDNAFEN_DA NULL
#define BEETLE_OPT_ENABLE_MEMCARD1_LABEL_DA NULL
#define BEETLE_OPT_ENABLE_MEMCARD1_INFO_0_DA NULL
#define BEETLE_OPT_SHARED_MEMORY_CARDS_LABEL_DA NULL
#define BEETLE_OPT_SHARED_MEMORY_CARDS_INFO_0_DA NULL
#define BEETLE_OPT_MEMCARD_LEFT_INDEX_LABEL_DA NULL
#define BEETLE_OPT_MEMCARD_LEFT_INDEX_INFO_0_DA NULL
#define OPTION_VAL_0_O48_DA NULL
#define BEETLE_OPT_MEMCARD_RIGHT_INDEX_LABEL_DA NULL
#define BEETLE_OPT_MEMCARD_RIGHT_INDEX_INFO_0_DA NULL
#define OPTION_VAL_1_O49_DA NULL
#define BEETLE_OPT_LINE_RENDER_LABEL_DA NULL
#define BEETLE_OPT_LINE_RENDER_INFO_0_DA NULL
#define OPTION_VAL_DEFAULT_DA NULL
#define OPTION_VAL_AGGRESSIVE_DA NULL
#define BEETLE_OPT_WIDESCREEN_HACK_LABEL_DA NULL
#define BEETLE_OPT_WIDESCREEN_HACK_INFO_0_DA NULL
#define BEETLE_OPT_WIDESCREEN_HACK_ASPECT_RATIO_LABEL_DA NULL
#define BEETLE_OPT_WIDESCREEN_HACK_ASPECT_RATIO_INFO_0_DA NULL
#define OPTION_VAL_16_9_DA NULL
#define OPTION_VAL_16_10_DA NULL
#define OPTION_VAL_18_9_DA NULL
#define OPTION_VAL_19_9_DA NULL
#define OPTION_VAL_20_9_DA NULL
#define OPTION_VAL_21_9_DA NULL
#define OPTION_VAL_32_9_DA NULL
#define BEETLE_OPT_CPU_DYNAREC_LABEL_DA NULL
#define BEETLE_OPT_CPU_DYNAREC_INFO_0_DA NULL
#define OPTION_VAL_DISABLED_DA NULL
#define OPTION_VAL_EXECUTE_DA NULL
#define OPTION_VAL_EXECUTE_ONE_DA NULL
#define OPTION_VAL_RUN_INTERPRETER_DA NULL
#define BEETLE_OPT_DYNAREC_INVALIDATE_LABEL_DA NULL
#define BEETLE_OPT_DYNAREC_INVALIDATE_INFO_0_DA NULL
#define OPTION_VAL_FULL_DA NULL
#define OPTION_VAL_DMA_DA NULL
#define BEETLE_OPT_DYNAREC_EVENTCYCLES_LABEL_DA NULL
#define BEETLE_OPT_DYNAREC_EVENTCYCLES_INFO_0_DA NULL
#define OPTION_VAL_128_DA NULL
#define BEETLE_OPT_DYNAREC_SPU_SAMPLES_LABEL_DA NULL
#define BEETLE_OPT_DYNAREC_SPU_SAMPLES_INFO_0_DA NULL
#define BEETLE_OPT_CORE_TIMING_FPS_LABEL_DA NULL
#define BEETLE_OPT_CORE_TIMING_FPS_INFO_0_DA NULL
#define OPTION_VAL_AUTO_TOGGLE_DA NULL
#define OPTION_VAL_FORCE_PROGRESSIVE_DA NULL
#define OPTION_VAL_FORCE_INTERLACED_DA NULL
#define BEETLE_OPT_DEINTERLACER_LABEL_DA NULL
#define BEETLE_OPT_DEINTERLACER_INFO_0_DA NULL
#define OPTION_VAL_WEAVE_DA NULL
#define OPTION_VAL_BOB_DA NULL
#define OPTION_VAL_BOB_OFFSET_DA NULL
#define OPTION_VAL_FASTMAD_DA NULL
#define BEETLE_OPT_FRAME_DUPING_LABEL_DA NULL
#define BEETLE_OPT_FRAME_DUPING_INFO_0_DA NULL
#define BEETLE_OPT_PAL_VIDEO_TIMING_OVERRIDE_LABEL_DA NULL
#define BEETLE_OPT_PAL_VIDEO_TIMING_OVERRIDE_INFO_0_DA NULL
#define BEETLE_OPT_ASPECT_RATIO_LABEL_DA NULL
#define BEETLE_OPT_ASPECT_RATIO_INFO_0_DA NULL
#define OPTION_VAL_CORRECTED_DA NULL
#define OPTION_VAL_UNCORRECTED_DA NULL
#define OPTION_VAL_4_3_DA NULL
#define OPTION_VAL_NTSC_DA NULL
#define BEETLE_OPT_CROP_OVERSCAN_LABEL_DA NULL
#define BEETLE_OPT_CROP_OVERSCAN_INFO_0_DA NULL
#define OPTION_VAL_STATIC_DA NULL
#define OPTION_VAL_SMART_DA NULL
#define BEETLE_OPT_IMAGE_CROP_LABEL_DA NULL
#define BEETLE_OPT_IMAGE_CROP_INFO_0_DA NULL
#define OPTION_VAL_9PX_DA NULL
#define OPTION_VAL_10PX_DA NULL
#define OPTION_VAL_11PX_DA NULL
#define OPTION_VAL_12PX_DA NULL
#define OPTION_VAL_13PX_DA NULL
#define OPTION_VAL_14PX_DA NULL
#define OPTION_VAL_15PX_DA NULL
#define OPTION_VAL_16PX_DA NULL
#define OPTION_VAL_17PX_DA NULL
#define OPTION_VAL_18PX_DA NULL
#define OPTION_VAL_19PX_DA NULL
#define OPTION_VAL_20PX_DA NULL
#define BEETLE_OPT_IMAGE_OFFSET_LABEL_DA NULL
#define BEETLE_OPT_IMAGE_OFFSET_INFO_0_DA NULL
#define OPTION_VAL_12PX_O64_DA NULL
#define OPTION_VAL_11PX_O64_DA NULL
#define OPTION_VAL_10PX_O64_DA NULL
#define OPTION_VAL_9PX_O64_DA NULL
#define OPTION_VAL_8PX_O64_DA NULL
#define OPTION_VAL_7PX_O64_DA NULL
#define OPTION_VAL_6PX_O64_DA NULL
#define OPTION_VAL_5PX_O64_DA NULL
#define OPTION_VAL_4PX_O64_DA NULL
#define OPTION_VAL_3PX_O64_DA NULL
#define OPTION_VAL_2PX_O64_DA NULL
#define OPTION_VAL_1PX_O64_DA NULL
#define OPTION_VAL_1PX_O64_0_DA NULL
#define OPTION_VAL_2PX_O64_0_DA NULL
#define OPTION_VAL_3PX_O64_0_DA NULL
#define OPTION_VAL_4PX_O64_0_DA NULL
#define OPTION_VAL_5PX_O64_0_DA NULL
#define OPTION_VAL_6PX_O64_0_DA NULL
#define OPTION_VAL_7PX_O64_0_DA NULL
#define OPTION_VAL_8PX_O64_0_DA NULL
#define OPTION_VAL_9PX_O64_0_DA NULL
#define OPTION_VAL_10PX_O64_0_DA NULL
#define OPTION_VAL_11PX_O64_0_DA NULL
#define OPTION_VAL_12PX_O64_0_DA NULL
#define BEETLE_OPT_IMAGE_OFFSET_CYCLES_LABEL_DA NULL
#define BEETLE_OPT_IMAGE_OFFSET_CYCLES_INFO_0_DA NULL
#define BEETLE_OPT_INITIAL_SCANLINE_LABEL_DA NULL
#define BEETLE_OPT_INITIAL_SCANLINE_INFO_0_DA NULL
#define BEETLE_OPT_LAST_SCANLINE_LABEL_DA NULL
#define BEETLE_OPT_LAST_SCANLINE_INFO_0_DA NULL
#define OPTION_VAL_239_DA NULL
#define BEETLE_OPT_INITIAL_SCANLINE_PAL_LABEL_DA NULL
#define BEETLE_OPT_INITIAL_SCANLINE_PAL_INFO_0_DA NULL
#define BEETLE_OPT_LAST_SCANLINE_PAL_LABEL_DA NULL
#define BEETLE_OPT_LAST_SCANLINE_PAL_INFO_0_DA NULL
#define OPTION_VAL_287_DA NULL

struct retro_core_option_v2_category option_cats_da[] = {
   {
      "system",
      CATEGORY_SYSTEM_LABEL_DA,
      CATEGORY_SYSTEM_INFO_0_DA
   },
   {
      "video",
      CATEGORY_VIDEO_LABEL_DA,
      CATEGORY_VIDEO_INFO_0_DA
   },
   {
      "pgxp",
      CATEGORY_PGXP_LABEL_DA,
      CATEGORY_PGXP_INFO_0_DA
   },
   {
      "osd",
      CATEGORY_OSD_LABEL_DA,
      CATEGORY_OSD_INFO_0_DA
   },
   {
      "input",
      CATEGORY_INPUT_LABEL_DA,
      CATEGORY_INPUT_INFO_0_DA
   },
   {
      "memcards",
      CATEGORY_MEMCARDS_LABEL_DA,
      CATEGORY_MEMCARDS_INFO_0_DA
   },
   {
      "hacks",
      CATEGORY_HACKS_LABEL_DA,
      CATEGORY_HACKS_INFO_0_DA
   },
   { NULL, NULL, NULL },
};
struct retro_core_option_v2_definition option_defs_da[] = {
   {
      BEETLE_OPT(cpu_freq_scale),
      BEETLE_OPT_CPU_FREQ_SCALE_LABEL_DA,
      NULL,
      BEETLE_OPT_CPU_FREQ_SCALE_INFO_0_DA,
      NULL,
      "system",
      {
         { "50%",           OPTION_VAL_50_DA },
         { "60%",           OPTION_VAL_60_DA },
         { "70%",           OPTION_VAL_70_DA },
         { "80%",           OPTION_VAL_80_DA },
         { "90%",           OPTION_VAL_90_DA },
         { "100%",          OPTION_VAL_100_DA },
         { "110%",          OPTION_VAL_110_DA },
         { "120%",          OPTION_VAL_120_DA },
         { "130%",          OPTION_VAL_130_DA },
         { "140%",          OPTION_VAL_140_DA },
         { "150%",          OPTION_VAL_150_DA },
         { "160%",          OPTION_VAL_160_DA },
         { "170%",          OPTION_VAL_170_DA },
         { "180%",          OPTION_VAL_180_DA },
         { "190%",          OPTION_VAL_190_DA },
         { "200%",          OPTION_VAL_200_DA },
         { "210%",          OPTION_VAL_210_DA },
         { "220%",          OPTION_VAL_220_DA },
         { "230%",          OPTION_VAL_230_DA },
         { "240%",          OPTION_VAL_240_DA },
         { "250%",          OPTION_VAL_250_DA },
         { "260%",          OPTION_VAL_260_DA },
         { "270%",          OPTION_VAL_270_DA },
         { "280%",          OPTION_VAL_280_DA },
         { "290%",          OPTION_VAL_290_DA },
         { "300%",          OPTION_VAL_300_DA },
         { "310%",          OPTION_VAL_310_DA },
         { "320%",          OPTION_VAL_320_DA },
         { "330%",          OPTION_VAL_330_DA },
         { "340%",          OPTION_VAL_340_DA },
         { "350%",          OPTION_VAL_350_DA },
         { "360%",          OPTION_VAL_360_DA },
         { "370%",          OPTION_VAL_370_DA },
         { "380%",          OPTION_VAL_380_DA },
         { "390%",          OPTION_VAL_390_DA },
         { "400%",          OPTION_VAL_400_DA },
         { "410%",          OPTION_VAL_410_DA },
         { "420%",          OPTION_VAL_420_DA },
         { "430%",          OPTION_VAL_430_DA },
         { "440%",          OPTION_VAL_440_DA },
         { "450%",          OPTION_VAL_450_DA },
         { "460%",          OPTION_VAL_460_DA },
         { "470%",          OPTION_VAL_470_DA },
         { "480%",          OPTION_VAL_480_DA },
         { "490%",          OPTION_VAL_490_DA },
         { "500%",          OPTION_VAL_500_DA },
         { "510%",          OPTION_VAL_510_DA },
         { "520%",          OPTION_VAL_520_DA },
         { "530%",          OPTION_VAL_530_DA },
         { "540%",          OPTION_VAL_540_DA },
         { "550%",          OPTION_VAL_550_DA },
         { "560%",          OPTION_VAL_560_DA },
         { "570%",          OPTION_VAL_570_DA },
         { "580%",          OPTION_VAL_580_DA },
         { "590%",          OPTION_VAL_590_DA },
         { "600%",          OPTION_VAL_600_DA },
         { "610%",          OPTION_VAL_610_DA },
         { "620%",          OPTION_VAL_620_DA },
         { "630%",          OPTION_VAL_630_DA },
         { "640%",          OPTION_VAL_640_DA },
         { "650%",          OPTION_VAL_650_DA },
         { "660%",          OPTION_VAL_660_DA },
         { "670%",          OPTION_VAL_670_DA },
         { "680%",          OPTION_VAL_680_DA },
         { "690%",          OPTION_VAL_690_DA },
         { "700%",          OPTION_VAL_700_DA },
         { "710%",          OPTION_VAL_710_DA },
         { "720%",          OPTION_VAL_720_DA },
         { "730%",          OPTION_VAL_730_DA },
         { "740%",          OPTION_VAL_740_DA },
         { "750%",          OPTION_VAL_750_DA },
         { NULL, NULL },
      },
      "100%"
   },
   {
      BEETLE_OPT(gpu_overclock),
      BEETLE_OPT_GPU_OVERCLOCK_LABEL_DA,
      NULL,
      BEETLE_OPT_GPU_OVERCLOCK_INFO_0_DA,
      NULL,
      "system",
      {
         { "1x(native)", OPTION_VAL_1X_NATIVE_DA },
         { "2x",         OPTION_VAL_2X_DA },
         { "4x",         OPTION_VAL_4X_DA },
         { "8x",         OPTION_VAL_8X_DA },
         { "16x",        OPTION_VAL_16X_DA },
         { "32x",        OPTION_VAL_32X_DA },
         { NULL, NULL },
      },
      "1x(native)"
   },
   {
      BEETLE_OPT(gte_overclock),
      BEETLE_OPT_GTE_OVERCLOCK_LABEL_DA,
      NULL,
      BEETLE_OPT_GTE_OVERCLOCK_INFO_0_DA,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
#ifndef EMSCRIPTEN
   {
      BEETLE_OPT(cd_access_method),
      BEETLE_OPT_CD_ACCESS_METHOD_LABEL_DA,
      NULL,
      BEETLE_OPT_CD_ACCESS_METHOD_INFO_0_DA,
      NULL,
      "system",
      {
         { "sync",     OPTION_VAL_SYNC_DA },
         { "async",    OPTION_VAL_ASYNC_DA },
         { "precache", OPTION_VAL_PRECACHE_DA },
         { NULL, NULL },
      },
      "sync"
   },
#endif
   {
      BEETLE_OPT(cd_fastload),
      BEETLE_OPT_CD_FASTLOAD_LABEL_DA,
      NULL,
      BEETLE_OPT_CD_FASTLOAD_INFO_0_DA,
      NULL,
      "system",
      {
         { "2x(native)", OPTION_VAL_2X_NATIVE_DA },
         { "4x",          OPTION_VAL_4X_DA },
         { "6x",          OPTION_VAL_6X_DA },
         { "8x",          OPTION_VAL_8X_DA },
         { "10x",         OPTION_VAL_10X_DA },
         { "12x",         OPTION_VAL_12X_DA },
         { "14x",         OPTION_VAL_14X_DA },
         { NULL, NULL },
      },
      "2x(native)"
   },
   {
      BEETLE_OPT(skip_bios),
      BEETLE_OPT_SKIP_BIOS_LABEL_DA,
      NULL,
      BEETLE_OPT_SKIP_BIOS_INFO_0_DA,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(override_bios),
      BEETLE_OPT_OVERRIDE_BIOS_LABEL_DA,
      NULL,
      BEETLE_OPT_OVERRIDE_BIOS_INFO_0_DA,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "psxonpsp", OPTION_VAL_PSXONPSP_DA },
         { "ps1_rom",  OPTION_VAL_PS1_ROM_DA },
         { "openbios", OPTION_VAL_OPENBIOS_DA     },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(region),
      BEETLE_OPT_REGION_LABEL_DA,
      NULL,
      BEETLE_OPT_REGION_INFO_0_DA,
      NULL,
      "system",
      {
         { "auto",   OPTION_VAL_AUTO_DA },
         { "ntsc-j", OPTION_VAL_NTSC_J_DA },
         { "ntsc-u", OPTION_VAL_NTSC_U_DA },
         { "pal",    OPTION_VAL_PAL_DA },
         { NULL, NULL },
      },
      "auto"
   },
   {
      BEETLE_OPT(internal_resolution),
      BEETLE_OPT_INTERNAL_RESOLUTION_LABEL_DA,
      NULL,
      BEETLE_OPT_INTERNAL_RESOLUTION_INFO_0_DA,
      NULL,
      "video",
      {
         { "1x(native)", OPTION_VAL_1X_NATIVE_DA },
         { "2x",         OPTION_VAL_2X_DA },
         { "4x",         OPTION_VAL_4X_DA },
         { "8x",         OPTION_VAL_8X_DA },
         { "16x",        OPTION_VAL_16X_DA },
         { NULL, NULL },
      },
      "1x(native)"
   },
#if defined(HAVE_OPENGL) || defined(HAVE_OPENGLES) || defined(HAVE_VULKAN)
   {
      BEETLE_OPT(renderer),
      BEETLE_OPT_RENDERER_LABEL_DA,
      NULL,
      BEETLE_OPT_RENDERER_INFO_0_DA,
      NULL,
      "video",
      {
         { "hardware",    OPTION_VAL_HARDWARE_DA },
#if defined(HAVE_OPENGL) || defined(HAVE_OPENGLES)
         { "hardware_gl", OPTION_VAL_HARDWARE_GL_DA },
#endif
#if defined(HAVE_VULKAN)
         { "hardware_vk", OPTION_VAL_HARDWARE_VK_DA },
#endif
         { "software",    OPTION_VAL_SOFTWARE_DA },
         { NULL, NULL },
      },
      "hardware"
   },
   {
      BEETLE_OPT(renderer_software_fb),
      BEETLE_OPT_RENDERER_SOFTWARE_FB_LABEL_DA,
      NULL,
      BEETLE_OPT_RENDERER_SOFTWARE_FB_INFO_0_DA,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled"
   },
#endif
#if defined(HAVE_OPENGL) || defined(HAVE_OPENGLES)
   {
      BEETLE_OPT(depth),
      BEETLE_OPT_DEPTH_LABEL_DA,
      NULL,
      BEETLE_OPT_DEPTH_INFO_0_DA,
      NULL,
      "video",
      {
         { "16bpp(native)", OPTION_VAL_16BPP_NATIVE_DA },
         { "32bpp",         OPTION_VAL_32BPP_DA },
         { NULL, NULL },
      },
      "16bpp(native)"
   },
   // Sort of, it's more like 15-bit high color and 24-bit true color for visible output. The alpha channel is used for mask bit. Vulkan renderer uses ABGR1555_555 for 31 bits internal? FMVs are always 24-bit on all renderers like original hardware (BGR888, no alpha)
#endif
   {
      BEETLE_OPT(dither_mode),
      BEETLE_OPT_DITHER_MODE_LABEL_DA,
      NULL,
      BEETLE_OPT_DITHER_MODE_INFO_0_DA,
      NULL,
      "video",
      {
         { "1x(native)",          OPTION_VAL_1X_NATIVE_DA },
         { "internal resolution", OPTION_VAL_INTERNAL_RESOLUTION_DA },
         { "disabled",            NULL },
         { NULL, NULL },
      },
      "1x(native)"
   },
#ifdef HAVE_VULKAN
   {
      BEETLE_OPT(scaled_uv_offset),
      BEETLE_OPT_SCALED_UV_OFFSET_LABEL_DA,
      NULL,
      BEETLE_OPT_SCALED_UV_OFFSET_INFO_0_DA,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled"
   },
#endif
#if defined(HAVE_OPENGL) || defined(HAVE_OPENGLES) || defined(HAVE_VULKAN)
   {
      BEETLE_OPT(filter),
      BEETLE_OPT_FILTER_LABEL_DA,
      NULL,
      BEETLE_OPT_FILTER_INFO_0_DA,
      NULL,
      "video",
      {
         { "nearest",  OPTION_VAL_NEAREST_DA },
         { "SABR",     OPTION_VAL_SABR_DA },
         { "xBR",      OPTION_VAL_XBR_DA },
         { "bilinear", OPTION_VAL_BILINEAR_DA },
         { "3-point",  OPTION_VAL_3_POINT_DA },
         { "JINC2",    OPTION_VAL_JINC2_DA },
         { NULL, NULL },
      },
      "nearest"
   },
#ifdef HAVE_VULKAN
   {
      BEETLE_OPT(filter_exclude_sprite),
      BEETLE_OPT_FILTER_EXCLUDE_SPRITE_LABEL_DA,
      NULL,
      BEETLE_OPT_FILTER_EXCLUDE_SPRITE_INFO_0_DA,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "opaque", OPTION_VAL_OPAQUE_DA },
         { "all", OPTION_VAL_ALL_DA },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(filter_exclude_2d_polygon),
      BEETLE_OPT_FILTER_EXCLUDE_2D_POLYGON_LABEL_DA,
      NULL,
      BEETLE_OPT_FILTER_EXCLUDE_2D_POLYGON_INFO_0_DA,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "opaque", OPTION_VAL_OPAQUE_DA },
         { "all", OPTION_VAL_ALL_DA },
         { NULL, NULL },
      },
      "disabled"
   },
#endif
#endif
#ifdef HAVE_VULKAN
   {
      BEETLE_OPT(adaptive_smoothing),
      BEETLE_OPT_ADAPTIVE_SMOOTHING_LABEL_DA,
      NULL,
      BEETLE_OPT_ADAPTIVE_SMOOTHING_INFO_0_DA,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(super_sampling),
      BEETLE_OPT_SUPER_SAMPLING_LABEL_DA,
      NULL,
      BEETLE_OPT_SUPER_SAMPLING_INFO_0_DA,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(msaa),
      BEETLE_OPT_MSAA_LABEL_DA,
      NULL,
      BEETLE_OPT_MSAA_INFO_0_DA,
      NULL,
      "video",
      {
         { "1x",  OPTION_VAL_1X_DA },
         { "2x",  OPTION_VAL_2X_DA },
         { "4x",  OPTION_VAL_4X_DA },
         { "8x",  OPTION_VAL_8X_DA },
         { "16x", OPTION_VAL_16X_DA },
         { NULL, NULL },
      },
      "1x"
   },
   {
      BEETLE_OPT(mdec_yuv),
      BEETLE_OPT_MDEC_YUV_LABEL_DA,
      NULL,
      BEETLE_OPT_MDEC_YUV_INFO_0_DA,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(track_textures),
      BEETLE_OPT_TRACK_TEXTURES_LABEL_DA,
      NULL,
      BEETLE_OPT_TRACK_TEXTURES_INFO_0_DA,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
#ifdef TEXTURE_DUMPING_ENABLED
   {
      BEETLE_OPT(dump_textures),
      BEETLE_OPT_DUMP_TEXTURES_LABEL_DA,
      NULL,
      BEETLE_OPT_DUMP_TEXTURES_INFO_0_DA,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
#endif
   {
      BEETLE_OPT(replace_textures),
      BEETLE_OPT_REPLACE_TEXTURES_LABEL_DA,
      NULL,
      BEETLE_OPT_REPLACE_TEXTURES_INFO_0_DA,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
#endif
   {
      BEETLE_OPT(pgxp_mode),
      BEETLE_OPT_PGXP_MODE_LABEL_DA,
      NULL,
      BEETLE_OPT_PGXP_MODE_INFO_0_DA,
      NULL,
      "pgxp",
      {
         { "disabled",     NULL },
         { "memory only",  OPTION_VAL_MEMORY_ONLY_DA },
         { "memory + CPU", OPTION_VAL_MEMORY_CPU_DA },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(pgxp_2d_tol),
      BEETLE_OPT_PGXP_2D_TOL_LABEL_DA,
      NULL,
      BEETLE_OPT_PGXP_2D_TOL_INFO_0_DA,
      NULL,
      "pgxp",
      {
         { "disabled", NULL },
         { "0px", OPTION_VAL_0PX_DA },
         { "1px", OPTION_VAL_1PX_DA },
         { "2px", OPTION_VAL_2PX_DA },
         { "3px", OPTION_VAL_3PX_DA },
         { "4px", OPTION_VAL_4PX_DA },
         { "5px", OPTION_VAL_5PX_DA },
         { "6px", OPTION_VAL_6PX_DA },
         { "7px", OPTION_VAL_7PX_DA },
         { "8px", OPTION_VAL_8PX_DA },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(pgxp_nclip),
      BEETLE_OPT_PGXP_NCLIP_LABEL_DA,
      NULL,
      BEETLE_OPT_PGXP_NCLIP_INFO_0_DA,
      NULL,
      "pgxp",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
#if defined(HAVE_OPENGL) || defined(HAVE_OPENGLES) || defined(HAVE_VULKAN)
   {
      BEETLE_OPT(pgxp_vertex),
      BEETLE_OPT_PGXP_VERTEX_LABEL_DA,
      NULL,
      BEETLE_OPT_PGXP_VERTEX_INFO_0_DA,
      NULL,
      "pgxp",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
#endif
   {
      BEETLE_OPT(pgxp_texture),
      BEETLE_OPT_PGXP_TEXTURE_LABEL_DA,
      NULL,
      BEETLE_OPT_PGXP_TEXTURE_INFO_0_DA,
      NULL,
      "pgxp",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(display_internal_fps),
      BEETLE_OPT_DISPLAY_INTERNAL_FPS_LABEL_DA,
      NULL,
      BEETLE_OPT_DISPLAY_INTERNAL_FPS_INFO_0_DA,
      NULL,
      "osd",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(display_osd),
      BEETLE_OPT_DISPLAY_OSD_LABEL_DA,
      NULL,
      BEETLE_OPT_DISPLAY_OSD_INFO_0_DA,
      NULL,
      "osd",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled"
   },
#if defined(HAVE_OPENGL) || defined(HAVE_OPENGLES) || defined(HAVE_VULKAN)
   {
      BEETLE_OPT(display_vram),
      BEETLE_OPT_DISPLAY_VRAM_LABEL_DA,
      NULL,
      BEETLE_OPT_DISPLAY_VRAM_INFO_0_DA,
      NULL,
      "osd",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
#endif
   {
      BEETLE_OPT(analog_calibration),
      BEETLE_OPT_ANALOG_CALIBRATION_LABEL_DA,
      NULL,
      BEETLE_OPT_ANALOG_CALIBRATION_INFO_0_DA,
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(analog_toggle),
      BEETLE_OPT_ANALOG_TOGGLE_LABEL_DA,
      NULL,
      BEETLE_OPT_ANALOG_TOGGLE_INFO_0_DA,
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { "enabled-analog", OPTION_VAL_ENABLED_ANALOG_DA },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(analog_toggle_combo),
      BEETLE_OPT_ANALOG_TOGGLE_COMBO_LABEL_DA,
      NULL,
      BEETLE_OPT_ANALOG_TOGGLE_COMBO_INFO_0_DA,
      NULL,
      "input",
      {
         { "l1+l2+r1+r2+start+select", OPTION_VAL_L1_L2_R1_R2_START_SELECT_DA },
         { "l1+r1+select",             OPTION_VAL_L1_R1_SELECT_DA },
         { "l1+r1+start",              OPTION_VAL_L1_R1_START_DA },
         { "l1+r1+l3",                 OPTION_VAL_L1_R1_L3_DA },
         { "l1+r1+r3",                 OPTION_VAL_L1_R1_R3_DA },
         { "l2+r2+select",             OPTION_VAL_L2_R2_SELECT_DA },
         { "l2+r2+start",              OPTION_VAL_L2_R2_START_DA },
         { "l2+r2+l3",                 OPTION_VAL_L2_R2_L3_DA },
         { "l2+r2+r3",                 OPTION_VAL_L2_R2_R3_DA },
         { "l3+r3",                    OPTION_VAL_L3_R3_DA },
         { NULL, NULL },
      },
      "l1+r1+select"
   },
   {
      BEETLE_OPT(analog_toggle_hold),
      BEETLE_OPT_ANALOG_TOGGLE_HOLD_LABEL_DA,
      NULL,
      BEETLE_OPT_ANALOG_TOGGLE_HOLD_INFO_0_DA,
      NULL,
      "input",
      {
         { "0", OPTION_VAL_0_DA },
         { "1", OPTION_VAL_1_DA },
         { "2", OPTION_VAL_2_DA },
         { "3", OPTION_VAL_3_DA },
         { "4", OPTION_VAL_4_DA },
         { "5", OPTION_VAL_5_DA },
         { NULL, NULL },
      },
      "1"
   },
   {
      BEETLE_OPT(enable_multitap_port1),
      BEETLE_OPT_ENABLE_MULTITAP_PORT1_LABEL_DA,
      NULL,
      BEETLE_OPT_ENABLE_MULTITAP_PORT1_INFO_0_DA,
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(enable_multitap_port2),
      BEETLE_OPT_ENABLE_MULTITAP_PORT2_LABEL_DA,
      NULL,
      BEETLE_OPT_ENABLE_MULTITAP_PORT2_INFO_0_DA,
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(gun_input_mode),
      BEETLE_OPT_GUN_INPUT_MODE_LABEL_DA,
      NULL,
      BEETLE_OPT_GUN_INPUT_MODE_INFO_0_DA,
      NULL,
      "input",
      {
         { "lightgun",    OPTION_VAL_LIGHTGUN_DA },
         { "touchscreen", OPTION_VAL_TOUCHSCREEN_DA },
         { NULL, NULL },
      },
      "lightgun"
   },
   // Shouldn't the gun_input_mode just be Mouse vs. Touchscreen?
   {
      BEETLE_OPT(gun_cursor),
      BEETLE_OPT_GUN_CURSOR_LABEL_DA,
      NULL,
      BEETLE_OPT_GUN_CURSOR_INFO_0_DA,
      NULL,
      "input",
      {
         { "off",   "disabled" },
         { "cross", OPTION_VAL_CROSS_DA },
         { "dot",   OPTION_VAL_DOT_DA },
         { NULL, NULL },
      },
      "cross"
   },
   {
      BEETLE_OPT(crosshair_color_p1),
      BEETLE_OPT_CROSSHAIR_COLOR_P1_LABEL_DA,
      NULL,
      BEETLE_OPT_CROSSHAIR_COLOR_P1_INFO_0_DA,
      NULL,
      "input",
      {
         { "red",    OPTION_VAL_RED_DA },
         { "blue",   OPTION_VAL_BLUE_DA },
         { "green",  OPTION_VAL_GREEN_DA },
         { "orange", OPTION_VAL_ORANGE_DA },
         { "yellow", OPTION_VAL_YELLOW_DA },
         { "cyan",   OPTION_VAL_CYAN_DA },
         { "pink",   OPTION_VAL_PINK_DA },
         { "purple", OPTION_VAL_PURPLE_DA },
         { "black",  OPTION_VAL_BLACK_DA },
         { "white",  OPTION_VAL_WHITE_DA },
         { NULL, NULL },
      },
      "red"
   },
   {
      BEETLE_OPT(crosshair_color_p2),
      BEETLE_OPT_CROSSHAIR_COLOR_P2_LABEL_DA,
      NULL,
      BEETLE_OPT_CROSSHAIR_COLOR_P2_INFO_0_DA,
      NULL,
      "input",
      {
         { "blue",   OPTION_VAL_BLUE_DA },
         { "red",    OPTION_VAL_RED_DA },
         { "green",  OPTION_VAL_GREEN_DA },
         { "orange", OPTION_VAL_ORANGE_DA },
         { "yellow", OPTION_VAL_YELLOW_DA },
         { "cyan",   OPTION_VAL_CYAN_DA },
         { "pink",   OPTION_VAL_PINK_DA },
         { "purple", OPTION_VAL_PURPLE_DA },
         { "black",  OPTION_VAL_BLACK_DA },
         { "white",  OPTION_VAL_WHITE_DA },
         { NULL, NULL },
      },
      "blue"
   },
   {
      BEETLE_OPT(mouse_sensitivity),
      BEETLE_OPT_MOUSE_SENSITIVITY_LABEL_DA,
      NULL,
      BEETLE_OPT_MOUSE_SENSITIVITY_INFO_0_DA,
      NULL,
      "input",
      {
         { "5%",   OPTION_VAL_5_O42_DA },
         { "10%",  OPTION_VAL_10_DA },
         { "15%",  OPTION_VAL_15_DA },
         { "20%",  OPTION_VAL_20_DA },
         { "25%",  OPTION_VAL_25_DA },
         { "30%",  OPTION_VAL_30_DA },
         { "35%",  OPTION_VAL_35_DA },
         { "40%",  OPTION_VAL_40_DA },
         { "45%",  OPTION_VAL_45_DA },
         { "50%",  OPTION_VAL_50_DA },
         { "55%",  OPTION_VAL_55_DA },
         { "60%",  OPTION_VAL_60_DA },
         { "65%",  OPTION_VAL_65_DA },
         { "70%",  OPTION_VAL_70_DA },
         { "75%",  OPTION_VAL_75_DA },
         { "80%",  OPTION_VAL_80_DA },
         { "85%",  OPTION_VAL_85_DA },
         { "90%",  OPTION_VAL_90_DA },
         { "95%",  OPTION_VAL_95_DA },
         { "100%", OPTION_VAL_100_O42_DA },
         { "105%", OPTION_VAL_105_DA },
         { "110%", OPTION_VAL_110_DA },
         { "115%", OPTION_VAL_115_DA },
         { "120%", OPTION_VAL_120_DA },
         { "125%", OPTION_VAL_125_DA },
         { "130%", OPTION_VAL_130_DA },
         { "135%", OPTION_VAL_135_DA },
         { "140%", OPTION_VAL_140_DA },
         { "145%", OPTION_VAL_145_DA },
         { "150%", OPTION_VAL_150_DA },
         { "155%", OPTION_VAL_155_DA },
         { "160%", OPTION_VAL_160_DA },
         { "165%", OPTION_VAL_165_DA },
         { "170%", OPTION_VAL_170_DA },
         { "175%", OPTION_VAL_175_DA },
         { "180%", OPTION_VAL_180_DA },
         { "185%", OPTION_VAL_185_DA },
         { "190%", OPTION_VAL_190_DA },
         { "195%", OPTION_VAL_195_DA },
         { "200%", OPTION_VAL_200_DA },
         { NULL, NULL },
      },
      "100%"
   },
   {
      BEETLE_OPT(negcon_response),
      BEETLE_OPT_NEGCON_RESPONSE_LABEL_DA,
      NULL,
      BEETLE_OPT_NEGCON_RESPONSE_INFO_0_DA,
      NULL,
      "input",
      {
         { "linear",    OPTION_VAL_LINEAR_DA },
         { "quadratic", OPTION_VAL_QUADRATIC_DA },
         { "cubic",     OPTION_VAL_CUBIC_DA },
         { NULL, NULL },
      },
      "linear"
   },
   {
      BEETLE_OPT(negcon_deadzone),
      BEETLE_OPT_NEGCON_DEADZONE_LABEL_DA,
      NULL,
      BEETLE_OPT_NEGCON_DEADZONE_INFO_0_DA,
      NULL,
      "input",
      {
         { "0%",  OPTION_VAL_0_O44_DA },
         { "5%",  OPTION_VAL_5_O42_DA },
         { "10%", OPTION_VAL_10_DA },
         { "15%", OPTION_VAL_15_DA },
         { "20%", OPTION_VAL_20_DA },
         { "25%", OPTION_VAL_25_DA },
         { "30%", OPTION_VAL_30_DA },
         { NULL, NULL },
      },
      "0%"
   },
   {
      BEETLE_OPT(use_mednafen_memcard0_method),
      BEETLE_OPT_USE_MEDNAFEN_MEMCARD0_METHOD_LABEL_DA,
      NULL,
      BEETLE_OPT_USE_MEDNAFEN_MEMCARD0_METHOD_INFO_0_DA,
      NULL,
      "memcards",
      {
         { "libretro", OPTION_VAL_LIBRETRO_DA },
         { "mednafen", OPTION_VAL_MEDNAFEN_DA },
         { NULL, NULL },
      },
      "libretro"
   },
   {
      BEETLE_OPT(enable_memcard1),
      BEETLE_OPT_ENABLE_MEMCARD1_LABEL_DA,
      NULL,
      BEETLE_OPT_ENABLE_MEMCARD1_INFO_0_DA,
      NULL,
      "memcards",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(shared_memory_cards),
      BEETLE_OPT_SHARED_MEMORY_CARDS_LABEL_DA,
      NULL,
      BEETLE_OPT_SHARED_MEMORY_CARDS_INFO_0_DA,
      NULL,
      "memcards",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(memcard_left_index),
      BEETLE_OPT_MEMCARD_LEFT_INDEX_LABEL_DA,
      NULL,
      BEETLE_OPT_MEMCARD_LEFT_INDEX_INFO_0_DA,
      NULL,
      "memcards",
      {
         { "0",  OPTION_VAL_0_O48_DA },
         { "1",  NULL },
         { "2",  NULL },
         { "3",  NULL },
         { "4",  NULL },
         { "5",  NULL },
         { "6",  NULL },
         { "7",  NULL },
         { "8",  NULL },
         { "9",  NULL },
         { "10", NULL },
         { "11", NULL },
         { "12", NULL },
         { "13", NULL },
         { "14", NULL },
         { "15", NULL },
         { "16", NULL },
         { "17", NULL },
         { "18", NULL },
         { "19", NULL },
         { "20", NULL },
         { "21", NULL },
         { "22", NULL },
         { "23", NULL },
         { "24", NULL },
         { "25", NULL },
         { "26", NULL },
         { "27", NULL },
         { "28", NULL },
         { "29", NULL },
         { "30", NULL },
         { "31", NULL },
         { "32", NULL },
         { "33", NULL },
         { "34", NULL },
         { "35", NULL },
         { "36", NULL },
         { "37", NULL },
         { "38", NULL },
         { "39", NULL },
         { "40", NULL },
         { "41", NULL },
         { "42", NULL },
         { "43", NULL },
         { "44", NULL },
         { "45", NULL },
         { "46", NULL },
         { "47", NULL },
         { "48", NULL },
         { "49", NULL },
         { "50", NULL },
         { "51", NULL },
         { "52", NULL },
         { "53", NULL },
         { "54", NULL },
         { "55", NULL },
         { "56", NULL },
         { "57", NULL },
         { "58", NULL },
         { "59", NULL },
         { "60", NULL },
         { "61", NULL },
         { "62", NULL },
         { "63", NULL },
         { NULL, NULL },
      },
      "0"
   },
   {
      BEETLE_OPT(memcard_right_index),
      BEETLE_OPT_MEMCARD_RIGHT_INDEX_LABEL_DA,
      NULL,
      BEETLE_OPT_MEMCARD_RIGHT_INDEX_INFO_0_DA,
      NULL,
      "memcards",
      {
         { "0",  NULL },
         { "1",  OPTION_VAL_1_O49_DA },
         { "2",  NULL },
         { "3",  NULL },
         { "4",  NULL },
         { "5",  NULL },
         { "6",  NULL },
         { "7",  NULL },
         { "8",  NULL },
         { "9",  NULL },
         { "10", NULL },
         { "11", NULL },
         { "12", NULL },
         { "13", NULL },
         { "14", NULL },
         { "15", NULL },
         { "16", NULL },
         { "17", NULL },
         { "18", NULL },
         { "19", NULL },
         { "20", NULL },
         { "21", NULL },
         { "22", NULL },
         { "23", NULL },
         { "24", NULL },
         { "25", NULL },
         { "26", NULL },
         { "27", NULL },
         { "28", NULL },
         { "29", NULL },
         { "30", NULL },
         { "31", NULL },
         { "32", NULL },
         { "33", NULL },
         { "34", NULL },
         { "35", NULL },
         { "36", NULL },
         { "37", NULL },
         { "38", NULL },
         { "39", NULL },
         { "40", NULL },
         { "41", NULL },
         { "42", NULL },
         { "43", NULL },
         { "44", NULL },
         { "45", NULL },
         { "46", NULL },
         { "47", NULL },
         { "48", NULL },
         { "49", NULL },
         { "50", NULL },
         { "51", NULL },
         { "52", NULL },
         { "53", NULL },
         { "54", NULL },
         { "55", NULL },
         { "56", NULL },
         { "57", NULL },
         { "58", NULL },
         { "59", NULL },
         { "60", NULL },
         { "61", NULL },
         { "62", NULL },
         { "63", NULL },
         { NULL, NULL },
      },
      "1"
   },
   {
      BEETLE_OPT(line_render),
      BEETLE_OPT_LINE_RENDER_LABEL_DA,
      NULL,
      BEETLE_OPT_LINE_RENDER_INFO_0_DA,
      NULL,
      "hacks",
      {
         { "default",    OPTION_VAL_DEFAULT_DA },
         { "aggressive", OPTION_VAL_AGGRESSIVE_DA },
         { "disabled",   NULL },
         { NULL, NULL },
      },
      "default"
   },
   {
      BEETLE_OPT(widescreen_hack),
      BEETLE_OPT_WIDESCREEN_HACK_LABEL_DA,
      NULL,
      BEETLE_OPT_WIDESCREEN_HACK_INFO_0_DA,
      NULL,
      "hacks",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(widescreen_hack_aspect_ratio),
      BEETLE_OPT_WIDESCREEN_HACK_ASPECT_RATIO_LABEL_DA,
      NULL,
      BEETLE_OPT_WIDESCREEN_HACK_ASPECT_RATIO_INFO_0_DA,
      NULL,
      "hacks",
      {
         { "16:9",  OPTION_VAL_16_9_DA },
         { "16:10", OPTION_VAL_16_10_DA },
         { "18:9",  OPTION_VAL_18_9_DA },
         { "19:9",  OPTION_VAL_19_9_DA },
         { "20:9",  OPTION_VAL_20_9_DA },
         { "21:9",  OPTION_VAL_21_9_DA }, // 64:27
         { "32:9",  OPTION_VAL_32_9_DA },
         { NULL,    NULL },
      },
      "16:9"
   },
#if defined(HAVE_LIGHTREC)
   {
      BEETLE_OPT(cpu_dynarec),
      BEETLE_OPT_CPU_DYNAREC_LABEL_DA,
      NULL,
      BEETLE_OPT_CPU_DYNAREC_INFO_0_DA,
      NULL,
      "hacks",
      {
         { "disabled",        OPTION_VAL_DISABLED_DA },
         { "execute",         OPTION_VAL_EXECUTE_DA },
         { "execute_one",     OPTION_VAL_EXECUTE_ONE_DA },
         { "run_interpreter", OPTION_VAL_RUN_INTERPRETER_DA },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(dynarec_invalidate),
      BEETLE_OPT_DYNAREC_INVALIDATE_LABEL_DA,
      NULL,
      BEETLE_OPT_DYNAREC_INVALIDATE_INFO_0_DA,
      NULL,
      "hacks",
      {
         { "full", OPTION_VAL_FULL_DA },
         { "dma",  OPTION_VAL_DMA_DA },
         { NULL, NULL },
      },
      "full"
   },
   {
      BEETLE_OPT(dynarec_eventcycles),
      BEETLE_OPT_DYNAREC_EVENTCYCLES_LABEL_DA,
      NULL,
      BEETLE_OPT_DYNAREC_EVENTCYCLES_INFO_0_DA,
      NULL,
      "hacks",
      {
         { "128",  OPTION_VAL_128_DA },
         { "256",  NULL },
         { "384",  NULL },
         { "512",  NULL },
         { "640",  NULL },
         { "768",  NULL },
         { "896",  NULL },
         { "1024", NULL },
         { "1152", NULL },
         { "1280", NULL },
         { "1408", NULL },
         { "1536", NULL },
         { "1664", NULL },
         { "1792", NULL },
         { "1920", NULL },
         { "2048", NULL },
         { NULL, NULL },
      },
      "128"
   },
   {
      BEETLE_OPT(dynarec_spu_samples),
      BEETLE_OPT_DYNAREC_SPU_SAMPLES_LABEL_DA,
      NULL,
      BEETLE_OPT_DYNAREC_SPU_SAMPLES_INFO_0_DA,
      NULL,
      "hacks",
      {
         { "1",  OPTION_VAL_1_O49_DA },
         { "4",  NULL },
         { "16", NULL },
         { NULL, NULL },
      },
      "1"
   },
#endif
   {
      BEETLE_OPT(core_timing_fps),
      BEETLE_OPT_CORE_TIMING_FPS_LABEL_DA,
      NULL,
      BEETLE_OPT_CORE_TIMING_FPS_INFO_0_DA,
      NULL,
      "video",
      {
         { "auto_toggle",       OPTION_VAL_AUTO_TOGGLE_DA },
         { "force_progressive", OPTION_VAL_FORCE_PROGRESSIVE_DA },
         { "force_interlaced",  OPTION_VAL_FORCE_INTERLACED_DA },
      },
      "auto_toggle"
   },
   {
      BEETLE_OPT(deinterlacer),
      BEETLE_OPT_DEINTERLACER_LABEL_DA,
      NULL,
      BEETLE_OPT_DEINTERLACER_INFO_0_DA,
      NULL,
      "video",
      {
         { "weave",      OPTION_VAL_WEAVE_DA },
         { "bob",        OPTION_VAL_BOB_DA },
         { "bob_offset", OPTION_VAL_BOB_OFFSET_DA },
         { "fastmad",    OPTION_VAL_FASTMAD_DA },
         { "off",        "Off" },
         { NULL, NULL },
      },
      "weave"
   },
   {
      BEETLE_OPT(frame_duping),
      BEETLE_OPT_FRAME_DUPING_LABEL_DA,
      NULL,
      BEETLE_OPT_FRAME_DUPING_INFO_0_DA,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(pal_video_timing_override),
      BEETLE_OPT_PAL_VIDEO_TIMING_OVERRIDE_LABEL_DA,
      NULL,
      BEETLE_OPT_PAL_VIDEO_TIMING_OVERRIDE_INFO_0_DA,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(aspect_ratio),
      BEETLE_OPT_ASPECT_RATIO_LABEL_DA,
      NULL,
      BEETLE_OPT_ASPECT_RATIO_INFO_0_DA,
      NULL,
      "video",
      {
         { "corrected", OPTION_VAL_CORRECTED_DA },
         { "uncorrected", OPTION_VAL_UNCORRECTED_DA },
         { "4:3",  OPTION_VAL_4_3_DA },
         { "ntsc", OPTION_VAL_NTSC_DA },
      },
      "corrected"
   },
   {
      BEETLE_OPT(crop_overscan),
      BEETLE_OPT_CROP_OVERSCAN_LABEL_DA,
      NULL,
      BEETLE_OPT_CROP_OVERSCAN_INFO_0_DA,
      NULL,
      "video",
      {
         { "disabled", "Disabled" },
         { "static",   OPTION_VAL_STATIC_DA },
         { "smart",    OPTION_VAL_SMART_DA },
         { NULL, NULL },
      },
      "smart"
   },
   {
      BEETLE_OPT(image_crop),
      BEETLE_OPT_IMAGE_CROP_LABEL_DA,
      NULL,
      BEETLE_OPT_IMAGE_CROP_INFO_0_DA,
      NULL,
      "video",
      {
         { "disabled", "0" },
         { "1px",      OPTION_VAL_1PX_DA },
         { "2px",      OPTION_VAL_2PX_DA },
         { "3px",      OPTION_VAL_3PX_DA },
         { "4px",      OPTION_VAL_4PX_DA },
         { "5px",      OPTION_VAL_5PX_DA },
         { "6px",      OPTION_VAL_6PX_DA },
         { "7px",      OPTION_VAL_7PX_DA },
         { "8px",      OPTION_VAL_8PX_DA },
         { "9px",      OPTION_VAL_9PX_DA },
         { "10px",     OPTION_VAL_10PX_DA },
         { "11px",     OPTION_VAL_11PX_DA },
         { "12px",     OPTION_VAL_12PX_DA },
         { "13px",     OPTION_VAL_13PX_DA },
         { "14px",     OPTION_VAL_14PX_DA },
         { "15px",     OPTION_VAL_15PX_DA },
         { "16px",     OPTION_VAL_16PX_DA },
         { "17px",     OPTION_VAL_17PX_DA },
         { "18px",     OPTION_VAL_18PX_DA },
         { "19px",     OPTION_VAL_19PX_DA },
         { "20px",     OPTION_VAL_20PX_DA },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(image_offset),
      BEETLE_OPT_IMAGE_OFFSET_LABEL_DA,
      NULL,
      BEETLE_OPT_IMAGE_OFFSET_INFO_0_DA,
      NULL,
      "video",
      {
         { "-12px",    OPTION_VAL_12PX_O64_DA },
         { "-11px",    OPTION_VAL_11PX_O64_DA },
         { "-10px",    OPTION_VAL_10PX_O64_DA },
         { "-9px",     OPTION_VAL_9PX_O64_DA },
         { "-8px",     OPTION_VAL_8PX_O64_DA },
         { "-7px",     OPTION_VAL_7PX_O64_DA },
         { "-6px",     OPTION_VAL_6PX_O64_DA },
         { "-5px",     OPTION_VAL_5PX_O64_DA },
         { "-4px",     OPTION_VAL_4PX_O64_DA },
         { "-3px",     OPTION_VAL_3PX_O64_DA },
         { "-2px",     OPTION_VAL_2PX_O64_DA },
         { "-1px",     OPTION_VAL_1PX_O64_DA },
         { "disabled", "0" },
         { "+1px",     OPTION_VAL_1PX_O64_0_DA },
         { "+2px",     OPTION_VAL_2PX_O64_0_DA },
         { "+3px",     OPTION_VAL_3PX_O64_0_DA },
         { "+4px",     OPTION_VAL_4PX_O64_0_DA },
         { "+5px",     OPTION_VAL_5PX_O64_0_DA },
         { "+6px",     OPTION_VAL_6PX_O64_0_DA },
         { "+7px",     OPTION_VAL_7PX_O64_0_DA },
         { "+8px",     OPTION_VAL_8PX_O64_0_DA },
         { "+9px",     OPTION_VAL_9PX_O64_0_DA },
         { "+10px",    OPTION_VAL_10PX_O64_0_DA },
         { "+11px",    OPTION_VAL_11PX_O64_0_DA },
         { "+12px",    OPTION_VAL_12PX_O64_0_DA },
         { NULL, NULL },
      },
      "disabled"
   },
#if defined(HAVE_OPENGL) || defined(HAVE_OPENGLES) || defined(HAVE_VULKAN)
   {
      BEETLE_OPT(image_offset_cycles),
      BEETLE_OPT_IMAGE_OFFSET_CYCLES_LABEL_DA,
      NULL,
      BEETLE_OPT_IMAGE_OFFSET_CYCLES_INFO_0_DA,
      NULL,
      "video",
      {
         { "-40",      NULL },
         { "-39",      NULL },
         { "-38",      NULL },
         { "-37",      NULL },
         { "-36",      NULL },
         { "-35",      NULL },
         { "-34",      NULL },
         { "-33",      NULL },
         { "-32",      NULL },
         { "-31",      NULL },
         { "-30",      NULL },
         { "-29",      NULL },
         { "-28",      NULL },
         { "-27",      NULL },
         { "-26",      NULL },
         { "-25",      NULL },
         { "-24",      NULL },
         { "-23",      NULL },
         { "-22",      NULL },
         { "-21",      NULL },
         { "-20",      NULL },
         { "-19",      NULL },
         { "-18",      NULL },
         { "-17",      NULL },
         { "-16",      NULL },
         { "-15",      NULL },
         { "-14",      NULL },
         { "-13",      NULL },
         { "-12",      NULL },
         { "-11",      NULL },
         { "-10",      NULL },
         { "-9",       NULL },
         { "-8",       NULL },
         { "-7",       NULL },
         { "-6",       NULL },
         { "-5",       NULL },
         { "-4",       NULL },
         { "-3",       NULL },
         { "-2",       NULL },
         { "-1",       NULL },
         { "0",        "0" },
         { "+1",       NULL },
         { "+2",       NULL },
         { "+3",       NULL },
         { "+4",       NULL },
         { "+5",       NULL },
         { "+6",       NULL },
         { "+7",       NULL },
         { "+8",       NULL },
         { "+9",       NULL },
         { "+10",      NULL },
         { "+11",      NULL },
         { "+12",      NULL },
         { "+13",      NULL },
         { "+14",      NULL },
         { "+15",      NULL },
         { "+16",      NULL },
         { "+17",      NULL },
         { "+18",      NULL },
         { "+19",      NULL },
         { "+20",      NULL },
         { "+21",      NULL },
         { "+22",      NULL },
         { "+23",      NULL },
         { "+24",      NULL },
         { "+25",      NULL },
         { "+26",      NULL },
         { "+27",      NULL },
         { "+28",      NULL },
         { "+29",      NULL },
         { "+30",      NULL },
         { "+31",      NULL },
         { "+32",      NULL },
         { "+33",      NULL },
         { "+34",      NULL },
         { "+35",      NULL },
         { "+36",      NULL },
         { "+37",      NULL },
         { "+38",      NULL },
         { "+39",      NULL },
         { "+40",      NULL },
         { NULL, NULL},
      },
      "0"
   },
#endif
   {
      BEETLE_OPT(initial_scanline),
      BEETLE_OPT_INITIAL_SCANLINE_LABEL_DA,
      NULL,
      BEETLE_OPT_INITIAL_SCANLINE_INFO_0_DA,
      NULL,
      "video",
      {
         { "0",  OPTION_VAL_0_O48_DA },
         { "1",  NULL },
         { "2",  NULL },
         { "3",  NULL },
         { "4",  NULL },
         { "5",  NULL },
         { "6",  NULL },
         { "7",  NULL },
         { "8",  NULL },
         { "9",  NULL },
         { "10", NULL },
         { "11", NULL },
         { "12", NULL },
         { "13", NULL },
         { "14", NULL },
         { "15", NULL },
         { "16", NULL },
         { "17", NULL },
         { "18", NULL },
         { "19", NULL },
         { "20", NULL },
         { "21", NULL },
         { "22", NULL },
         { "23", NULL },
         { "24", NULL },
         { "25", NULL },
         { "26", NULL },
         { "27", NULL },
         { "28", NULL },
         { "29", NULL },
         { "30", NULL },
         { "31", NULL },
         { "32", NULL },
         { "33", NULL },
         { "34", NULL },
         { "35", NULL },
         { "36", NULL },
         { "37", NULL },
         { "38", NULL },
         { "39", NULL },
         { "40", NULL },
         { NULL, NULL },
      },
      "0"
   },
   {
      BEETLE_OPT(last_scanline),
      BEETLE_OPT_LAST_SCANLINE_LABEL_DA,
      NULL,
      BEETLE_OPT_LAST_SCANLINE_INFO_0_DA,
      NULL,
      "video",
      {
         { "210", NULL },
         { "211", NULL },
         { "212", NULL },
         { "213", NULL },
         { "214", NULL },
         { "215", NULL },
         { "216", NULL },
         { "217", NULL },
         { "218", NULL },
         { "219", NULL },
         { "220", NULL },
         { "221", NULL },
         { "222", NULL },
         { "223", NULL },
         { "224", NULL },
         { "225", NULL },
         { "226", NULL },
         { "227", NULL },
         { "228", NULL },
         { "229", NULL },
         { "230", NULL },
         { "231", NULL },
         { "232", NULL },
         { "233", NULL },
         { "234", NULL },
         { "235", NULL },
         { "236", NULL },
         { "237", NULL },
         { "238", NULL },
         { "239", OPTION_VAL_239_DA },
         { NULL, NULL },
      },
      "239"
   },
   {
      BEETLE_OPT(initial_scanline_pal),
      BEETLE_OPT_INITIAL_SCANLINE_PAL_LABEL_DA,
      NULL,
      BEETLE_OPT_INITIAL_SCANLINE_PAL_INFO_0_DA,
      NULL,
      "video",
      {
         { "0",  OPTION_VAL_0_O48_DA },
         { "1",  NULL },
         { "2",  NULL },
         { "3",  NULL },
         { "4",  NULL },
         { "5",  NULL },
         { "6",  NULL },
         { "7",  NULL },
         { "8",  NULL },
         { "9",  NULL },
         { "10", NULL },
         { "11", NULL },
         { "12", NULL },
         { "13", NULL },
         { "14", NULL },
         { "15", NULL },
         { "16", NULL },
         { "17", NULL },
         { "18", NULL },
         { "19", NULL },
         { "20", NULL },
         { "21", NULL },
         { "22", NULL },
         { "23", NULL },
         { "24", NULL },
         { "25", NULL },
         { "26", NULL },
         { "27", NULL },
         { "28", NULL },
         { "29", NULL },
         { "30", NULL },
         { "31", NULL },
         { "32", NULL },
         { "33", NULL },
         { "34", NULL },
         { "35", NULL },
         { "36", NULL },
         { "37", NULL },
         { "38", NULL },
         { "39", NULL },
         { "40", NULL },
         { NULL, NULL },
      },
      "0"
   },
   {
      BEETLE_OPT(last_scanline_pal),
      BEETLE_OPT_LAST_SCANLINE_PAL_LABEL_DA,
      NULL,
      BEETLE_OPT_LAST_SCANLINE_PAL_INFO_0_DA,
      NULL,
      "video",
      {
         { "230", NULL },
         { "231", NULL },
         { "232", NULL },
         { "233", NULL },
         { "234", NULL },
         { "235", NULL },
         { "236", NULL },
         { "237", NULL },
         { "238", NULL },
         { "239", NULL },
         { "240", NULL },
         { "241", NULL },
         { "242", NULL },
         { "243", NULL },
         { "244", NULL },
         { "245", NULL },
         { "246", NULL },
         { "247", NULL },
         { "248", NULL },
         { "249", NULL },
         { "250", NULL },
         { "251", NULL },
         { "252", NULL },
         { "253", NULL },
         { "254", NULL },
         { "255", NULL },
         { "256", NULL },
         { "257", NULL },
         { "258", NULL },
         { "259", NULL },
         { "260", NULL },
         { "261", NULL },
         { "262", NULL },
         { "263", NULL },
         { "264", NULL },
         { "265", NULL },
         { "266", NULL },
         { "267", NULL },
         { "268", NULL },
         { "269", NULL },
         { "270", NULL },
         { "271", NULL },
         { "272", NULL },
         { "273", NULL },
         { "274", NULL },
         { "275", NULL },
         { "276", NULL },
         { "277", NULL },
         { "278", NULL },
         { "279", NULL },
         { "280", NULL },
         { "281", NULL },
         { "282", NULL },
         { "283", NULL },
         { "284", NULL },
         { "285", NULL },
         { "286", NULL },
         { "287", OPTION_VAL_287_DA },
         { NULL, NULL },
      },
      "287"
   },
   { NULL, NULL, NULL, NULL, NULL, NULL, {{0}}, NULL },
};
struct retro_core_options_v2 options_da = {
   option_cats_da,
   option_defs_da
};

/* RETRO_LANGUAGE_DE */

#define CATEGORY_SYSTEM_LABEL_DE NULL
#define CATEGORY_SYSTEM_INFO_0_DE "System- und Übertaktungsoptionen ändern."
#define CATEGORY_VIDEO_LABEL_DE NULL
#define CATEGORY_VIDEO_INFO_0_DE "Einstellungen für Seitenverhältnis, Bildausschnitt, Videofilter und Frameskipping ändern."
#define CATEGORY_PGXP_LABEL_DE NULL
#define CATEGORY_PGXP_INFO_0_DE "Diese Optionen steuern Erweiterungen, die die Grafik im Vergleich zur Originalkonsole verbessern können. PGXP kann verzerrte Texturen und Z-Fighting-Probleme beseitigen."
#define CATEGORY_OSD_LABEL_DE "Bildschirmanzeige"
#define CATEGORY_OSD_INFO_0_DE "Bildschirm-Benachrichtigungen ändern."
#define CATEGORY_INPUT_LABEL_DE "Eingabe"
#define CATEGORY_INPUT_INFO_0_DE "Einstellungen für Lightgun, Maus und neGcon ändern."
#define CATEGORY_MEMCARDS_LABEL_DE NULL
#define CATEGORY_MEMCARDS_INFO_0_DE "Die Einstellungen für die virtuelle Speicherkarte ändern."
#define CATEGORY_HACKS_LABEL_DE "Emulations-Hacks"
#define CATEGORY_HACKS_INFO_0_DE "Einstellungen für die Emulationsgenauigkeit ändern, die sich auf die Low-Level-Leistung und Kompatibilität auswirken."
#define BEETLE_OPT_CPU_FREQ_SCALE_LABEL_DE "CPU-Frequenzskalierung (Übertaktung)"
#define BEETLE_OPT_CPU_FREQ_SCALE_INFO_0_DE "Emulierte PSX-CPU übertakten (oder untertakten). Durch die Übertaktung können Slowdowns beseitigt und die Bildrate in bestimmten Spielen verbessert werden, allerdings auf Kosten von höheren Leistungsansprüchen. Zu beachten ist, dass einige Spiele eine interne Bildratenbegrenzung haben und möglicherweise nicht von einer Übertaktung profitieren. Bei einigen Spielen kann die Übertaktung dazu führen, dass bestimmte Effekte schneller als vorgesehen animiert werden."
#define OPTION_VAL_50_DE "50 %"
#define OPTION_VAL_60_DE "60 %"
#define OPTION_VAL_70_DE "70 %"
#define OPTION_VAL_80_DE "80 %"
#define OPTION_VAL_90_DE "90 %"
#define OPTION_VAL_100_DE "100 % (Nativ)"
#define OPTION_VAL_110_DE "110 %"
#define OPTION_VAL_120_DE "120 %"
#define OPTION_VAL_130_DE "130 %"
#define OPTION_VAL_140_DE "140 %"
#define OPTION_VAL_150_DE "150 %"
#define OPTION_VAL_160_DE "160 %"
#define OPTION_VAL_170_DE "170 %"
#define OPTION_VAL_180_DE "180 %"
#define OPTION_VAL_190_DE "190 %"
#define OPTION_VAL_200_DE "200 %"
#define OPTION_VAL_210_DE "210 %"
#define OPTION_VAL_220_DE "220 %"
#define OPTION_VAL_230_DE "230 %"
#define OPTION_VAL_240_DE "240 %"
#define OPTION_VAL_250_DE "250 %"
#define OPTION_VAL_260_DE "260 %"
#define OPTION_VAL_270_DE "270 %"
#define OPTION_VAL_280_DE "280 %"
#define OPTION_VAL_290_DE "290 %"
#define OPTION_VAL_300_DE "300 %"
#define OPTION_VAL_310_DE "310 %"
#define OPTION_VAL_320_DE "320 %"
#define OPTION_VAL_330_DE "330 %"
#define OPTION_VAL_340_DE "340 %"
#define OPTION_VAL_350_DE "350 %"
#define OPTION_VAL_360_DE "360 %"
#define OPTION_VAL_370_DE "370 %"
#define OPTION_VAL_380_DE "380 %"
#define OPTION_VAL_390_DE "390 %"
#define OPTION_VAL_400_DE "400 %"
#define OPTION_VAL_410_DE "410 %"
#define OPTION_VAL_420_DE "420 %"
#define OPTION_VAL_430_DE "430 %"
#define OPTION_VAL_440_DE "440 %"
#define OPTION_VAL_450_DE "450 %"
#define OPTION_VAL_460_DE "460 %"
#define OPTION_VAL_470_DE "470 %"
#define OPTION_VAL_480_DE "480 %"
#define OPTION_VAL_490_DE "490 %"
#define OPTION_VAL_500_DE "500 %"
#define OPTION_VAL_510_DE "510 %"
#define OPTION_VAL_520_DE "520 %"
#define OPTION_VAL_530_DE "530 %"
#define OPTION_VAL_540_DE "540 %"
#define OPTION_VAL_550_DE "550 %"
#define OPTION_VAL_560_DE "560 %"
#define OPTION_VAL_570_DE "570 %"
#define OPTION_VAL_580_DE "580 %"
#define OPTION_VAL_590_DE "590 %"
#define OPTION_VAL_600_DE "600 %"
#define OPTION_VAL_610_DE "610 %"
#define OPTION_VAL_620_DE "620 %"
#define OPTION_VAL_630_DE "630 %"
#define OPTION_VAL_640_DE "640 %"
#define OPTION_VAL_650_DE "650 %"
#define OPTION_VAL_660_DE "660 %"
#define OPTION_VAL_670_DE "670 %"
#define OPTION_VAL_680_DE "680 %"
#define OPTION_VAL_690_DE "690 %"
#define OPTION_VAL_700_DE "700 %"
#define OPTION_VAL_710_DE "710 %"
#define OPTION_VAL_720_DE "720 %"
#define OPTION_VAL_730_DE "730 %"
#define OPTION_VAL_740_DE "740 %"
#define OPTION_VAL_750_DE "750 %"
#define BEETLE_OPT_GPU_OVERCLOCK_LABEL_DE "GPU-Rasterübertaktung"
#define BEETLE_OPT_GPU_OVERCLOCK_INFO_0_DE "Aktiviert Übertaktung des 2D-Rasters, welcher in der GPU-Emulation der PSX enthalten ist. Verbessert nicht die 3D-Darstellung und hat im Allgemeinen wenig Effekt."
#define OPTION_VAL_1X_NATIVE_DE "1x (Nativ)"
#define OPTION_VAL_2X_DE NULL
#define OPTION_VAL_4X_DE NULL
#define OPTION_VAL_8X_DE NULL
#define OPTION_VAL_16X_DE NULL
#define OPTION_VAL_32X_DE NULL
#define BEETLE_OPT_GTE_OVERCLOCK_LABEL_DE "GTE-Übertaktung"
#define BEETLE_OPT_GTE_OVERCLOCK_INFO_0_DE "Verringert alle emulierten GTE (CPU-Koprozessor für 3D-Grafiken) Operationen zu einer konstanten 1-Zyklus-Latenz. Bei Spielen, die die GTE stark nutzen, kann dies die Bildrate und die Bildstabilität erheblich verbessern."
#define BEETLE_OPT_CD_ACCESS_METHOD_LABEL_DE "CD-Zugriffsmethode"
#define BEETLE_OPT_CD_ACCESS_METHOD_INFO_0_DE "„Synchron“ ahmt die Originalhardware nach. „Asynchron“ kann das Stottern auf Geräten mit langsamem Speicher verringern. „Vorab laden“ lädt beim Starten von Inhalten das gesamte CD-Abbildes in den Speicher, was die Ladezeiten im Spiel auf Kosten einer anfänglichen Verzögerung beim Starten verbessern kann. „Vorab laden“ kann auf Systemen mit wenig Arbeitsspeicher zu Problemen führen, und wird bei physischen Medien auf „Synchron“ zurückgreifen. Neustart erforderlich."
#define OPTION_VAL_SYNC_DE "Synchron"
#define OPTION_VAL_ASYNC_DE "Asynchron"
#define OPTION_VAL_PRECACHE_DE "Vorab laden"
#define BEETLE_OPT_CD_FASTLOAD_LABEL_DE "CD-Ladegeschwindigkeit"
#define BEETLE_OPT_CD_FASTLOAD_INFO_0_DE "Werte über „2x (Nativ)“ können die Ladezeiten im Spiel erheblich verkürzen, jedoch zu Timing-Fehlern führen. Einige Spiele funktionieren ab einem bestimmten Wert möglicherweise nicht mehr ordnungsgemäß."
#define OPTION_VAL_2X_NATIVE_DE "2x (Nativ)"
#define OPTION_VAL_6X_DE "6 x"
#define OPTION_VAL_10X_DE NULL
#define OPTION_VAL_12X_DE NULL
#define OPTION_VAL_14X_DE NULL
#define BEETLE_OPT_SKIP_BIOS_LABEL_DE "BIOS überspringen"
#define BEETLE_OPT_SKIP_BIOS_INFO_0_DE "Die PlayStation BIOS-Boot-Animation, die normalerweise beim Laden von Inhalten angezeigt wird, überspringen. Hinweis: Diese Funktion führt zu Kompatibilitätsproblemen bei einer Reihe von Spielen (PAL-kopiergeschützte Spiele, Saga Frontier usw.)."
#define BEETLE_OPT_OVERRIDE_BIOS_LABEL_DE "BIOS überschreiben"
#define BEETLE_OPT_OVERRIDE_BIOS_INFO_0_DE "Das Standard-regionsspezifische BIOS mit einem regionenfreien BIOS überschreiben, falls vorhanden. Neustart erforderlich."
#define OPTION_VAL_PSXONPSP_DE NULL
#define OPTION_VAL_PS1_ROM_DE NULL
#define OPTION_VAL_OPENBIOS_DE NULL
#define BEETLE_OPT_REGION_LABEL_DE "Systemregion"
#define BEETLE_OPT_REGION_INFO_0_DE "Legt den Fallback-Bereich fest, der verwendet wird, wenn Inhalte nicht automatisch erkannt werden können (z. B. rohe PS-X EXE-Dateien). Disc-basierte Spiele verwenden immer noch die auf der Disc hinterlegten Region. Neustart erforderlich."
#define OPTION_VAL_AUTO_DE "Automatisch"
#define OPTION_VAL_NTSC_J_DE NULL
#define OPTION_VAL_NTSC_U_DE "NTSC-U (Nordamerika)"
#define OPTION_VAL_PAL_DE "PAL (Europa)"
#define BEETLE_OPT_INTERNAL_RESOLUTION_LABEL_DE "Interne GPU-Auflösung"
#define BEETLE_OPT_INTERNAL_RESOLUTION_INFO_0_DE "Auflösungen, höher als „1x (Native)“, verbessern die Genauigkeit von 3D-Modellen, erhöhen jedoch die Leistungsanforderungen. 2D-Elemente sind von dieser Einstellung in der Regel nicht betroffen."
#define BEETLE_OPT_RENDERER_LABEL_DE NULL
#define BEETLE_OPT_RENDERER_INFO_0_DE "Der Software-Renderer ist der genaueste, hat aber hohe Leistungsanforderungen, wenn er bei erhöhten internen GPU-Auflösungen ausgeführt wird. Die Hardware-Renderer verbessern die Leistung gegenüber dem Software-Renderer bei hoheren internen Auflösungen und ermöglichen verschiedene grafische Verbesserungen. „Hardware (Automatisch)“ wählt automatisch die Renderer Vulkan oder OpenGL aus, abhängig vom aktuellen libretro-Videotreiber. Wenn der mitgelieferte Videotreiber nicht mit Vulkan oder OpenGL 3.3 kompatibel ist, greift der Core auf den Software-Renderer zurück. Neustart erforderlich."
#define OPTION_VAL_HARDWARE_DE "Hardware (Automatisch)"
#define OPTION_VAL_HARDWARE_GL_DE NULL
#define OPTION_VAL_HARDWARE_VK_DE NULL
#define OPTION_VAL_SOFTWARE_DE NULL
#define BEETLE_OPT_RENDERER_SOFTWARE_FB_LABEL_DE "Software-Framebuffer"
#define BEETLE_OPT_RENDERER_SOFTWARE_FB_INFO_0_DE "Aktiviere genaue Emulation von Bildspeicher-Effekten (z. Motion Blur, FF7 Battle swirl), wenn Sie Hardware-Renderer verwenden, indem Sie eine Kopie des Software-Renderer mit nativer Auflösung im Hintergrund ausführen. Falls deaktiviert, werden diese Operationen (OpenGL) ausgelassen oder auf der GPU (Vulkan) gerendert. Das Deaktivieren kann die Leistung verbessern, kann aber schwerwiegende graphische Fehler verursachen. Aktiviert lassen wenn unsicher."
#define BEETLE_OPT_DEPTH_LABEL_DE "Interne Farbtiefe"
#define BEETLE_OPT_DEPTH_INFO_0_DE "Eine höhere Farbtiefe kann ohne Dithering Farbstreifen reduzieren. 16 bpp emuliert die Originalhardware, kann jedoch sichtbare Streifen verursachen, wenn Dithering nicht aktiviert ist. Es wird empfohlen, „Ditheringmuster” zu deaktivieren, wenn diese Option auf 32 bpp eingestellt ist."
#define OPTION_VAL_16BPP_NATIVE_DE "16 bpp (Nativ)"
#define OPTION_VAL_32BPP_DE NULL
#define BEETLE_OPT_DITHER_MODE_LABEL_DE "Dithering-Muster"
#define BEETLE_OPT_DITHER_MODE_INFO_0_DE "„1x (Nativ)“ emuliert natives, niedrigaufgelöstes Dithering, wie es von der Originalhardware zum Glätten von Farbband-Artefakten bei nativer Farbtiefe verwendet wird. „Interne Auflösung“ skaliert die Dithering-Granularität auf die ausgewählte interne Auflösung für sauberere Ergebnisse. Bei 32 bpp Farbtiefe wird empfohlen, diese Option zu deaktivieren. Hinweis: Bei Vulkan erzwingt die Aktivierung dieser Option ein Downsampling auf die native Farbtiefe, während die Deaktivierung die Ausgabe mit einer höheren Farbtiefe automatisch aktiviert."
#define OPTION_VAL_INTERNAL_RESOLUTION_DE "Interne Auflösung"
#define BEETLE_OPT_SCALED_UV_OFFSET_LABEL_DE "Texturen-UV-Versatz"
#define BEETLE_OPT_SCALED_UV_OFFSET_INFO_0_DE "Texturen für 3D-Polygone bei höherer interner Auflösung mit einem Versatz abtasten. Verringert Texturnähte, kann aber unbeabsichtigte grafische Glitches verursachen."
#define BEETLE_OPT_FILTER_LABEL_DE "Texturfilterung"
#define BEETLE_OPT_FILTER_INFO_0_DE "„Nächste“ emuliert originale Hardware. „Bilinear“ und „3-Punkt“ sind Glättungsfilter, die die Pixelierung durch Unschärfe reduzieren. „SABR“, „xBR“ und „JINC2“ sind Hochskalierungsfilter, die die Texturtreue/-schärfe auf Kosten erhöhter Leistungsanforderungen verbessern können. Sie werden nur von den Hardware-Renderern unterstützt."
#define OPTION_VAL_NEAREST_DE "Nächste"
#define OPTION_VAL_SABR_DE NULL
#define OPTION_VAL_XBR_DE NULL
#define OPTION_VAL_BILINEAR_DE NULL
#define OPTION_VAL_3_POINT_DE "3-Punkt"
#define OPTION_VAL_JINC2_DE NULL
#define BEETLE_OPT_FILTER_EXCLUDE_SPRITE_LABEL_DE "Sprites von der Filterung ausschließen"
#define BEETLE_OPT_FILTER_EXCLUDE_SPRITE_INFO_0_DE "Keine Texturfilterung auf Sprites anwenden. Verhindert Nahtstellen in verschiedenen Spielen mit 2D-Sprite-gerenderten Hintergründen. Zusammen mit \"Adaptive Glättung\" oder einem anderen Nachbearbeitungsfilter verwenden, um die beste Wirkung zu erzielen."
#define OPTION_VAL_OPAQUE_DE "Nur undurchsichtig"
#define OPTION_VAL_ALL_DE "Undurchsichtig und Halbtransparent"
#define BEETLE_OPT_FILTER_EXCLUDE_2D_POLYGON_LABEL_DE "2D-Polygone von der Filterung ausschließen"
#define BEETLE_OPT_FILTER_EXCLUDE_2D_POLYGON_INFO_0_DE "Keine Texturfilterung auf 2D-Polygone anwenden. 2D-Polygone werden mit einer Heuristik erkannt und es kann zu Glitches kommen. Texturfilterung zusammen mit \"Adaptive Glättung\" oder einem anderen Nachbearbeitungsfilter verwenden, um eine optimale Wirkung zu erzielen."
#define BEETLE_OPT_ADAPTIVE_SMOOTHING_LABEL_DE "Adaptive Glättung"
#define BEETLE_OPT_ADAPTIVE_SMOOTHING_INFO_0_DE "2D-Grafiken und UI-Elemente glätten, ohne 3D-Objekte zu verwischen. Wird nur vom Vulkan-Renderer unterstützt."
#define BEETLE_OPT_SUPER_SAMPLING_LABEL_DE "Supersampling (Downsampling zur nativen Auflösung)"
#define BEETLE_OPT_SUPER_SAMPLING_INFO_0_DE "Heruntertakten gerenderter Inhalte von hochskalierter interner Auflösung bis hin zur nativen Auflösung. Die Kombination mit einer höheren internen Auflösung ermöglicht die Darstellung von Spielen mit geglätteten 3D-Objekten mit niedriger nativer Auflösung. Erzeugt beste Ergebnisse bei Titeln, die 2D- und 3D-Elemente mischen (z. 3D-Zeichen auf vorgerenderten Hintergründen) und funktioniert gut in Verbindung mit CRT-Shadern. Nur vom Vulkan Renderer unterstützt. Hinweis: 'Dithering-Muster' wird empfohlen, wenn diese Option aktiviert wird."
#define BEETLE_OPT_MSAA_LABEL_DE "Multisampled-Antialiasing"
#define BEETLE_OPT_MSAA_INFO_0_DE "MSAA verbessert die Darstellung von 3D-Objekten. Wird nur vom Vulkan-Renderer unterstützt."
#define OPTION_VAL_1X_DE NULL
#define BEETLE_OPT_MDEC_YUV_LABEL_DE "MDEC YUV Chromafilter"
#define BEETLE_OPT_MDEC_YUV_INFO_0_DE "Die Qualität der FMV-Wiedergabe durch Reduzierung von „Macroblocking“-Artefakten (Quadrate/gezackte Kanten) verbessern. Wird nur vom Vulkan-Renderer unterstützt."
#define BEETLE_OPT_TRACK_TEXTURES_LABEL_DE "Texturenverfolgung"
#define BEETLE_OPT_TRACK_TEXTURES_INFO_0_DE "Voraussetzung für Dumping und Ersetzen von Texturen. Wird wahrscheinlich in den meisten Spielen abstürzen."
#define BEETLE_OPT_DUMP_TEXTURES_LABEL_DE "Texturen dumpen"
#define BEETLE_OPT_DUMP_TEXTURES_INFO_0_DE "Verwendete Texturen nach <cd>-texture-dump/ ausgeben"
#define BEETLE_OPT_REPLACE_TEXTURES_LABEL_DE "Texturen ersetzen"
#define BEETLE_OPT_REPLACE_TEXTURES_INFO_0_DE "Texturen durch HD-Versionen aus <cd>-texture-replacements/ ersetzen"
#define BEETLE_OPT_PGXP_MODE_LABEL_DE "PGXP-Betriebsart"
#define BEETLE_OPT_PGXP_MODE_INFO_0_DE "Darstellung von 3D-Objekten mit Unterpixelgenauigkeit, minimiert Verzerrung und Ruckler von 3D-Objekten, die auf Originalhardware durch die Verwendung von fixen Punktvertex-Koordinaten gesehen werden. Der Modus „Nur Speicher“ hat minimale Kompatibilitätsprobleme und wird für den allgemeinen Gebrauch empfohlen. „Speicher und CPU (fehlerbehaftet)“-Modus kann Ruckler noch weiter reduzieren, hat aber hohe Leistungsanforderungen und kann zu verschiedenen Geometriefehlern führen."
#define OPTION_VAL_MEMORY_ONLY_DE "Nur Speicher"
#define OPTION_VAL_MEMORY_CPU_DE "Speicher und CPU (fehlerbehaftet)"
#define BEETLE_OPT_PGXP_2D_TOL_LABEL_DE "PGXP 2D-Geometrietoleranz"
#define BEETLE_OPT_PGXP_2D_TOL_INFO_0_DE "Versteckt weitere eklatante Fehler in PGXP-Operationen: Der Wert legt fest, in welcher Toleranz PGXP-Werte im Falle von Geometrien ohne korrekte Tiefeninformationen beibehalten werden."
#define OPTION_VAL_0PX_DE "0 px"
#define OPTION_VAL_1PX_DE "1 px"
#define OPTION_VAL_2PX_DE "2 px"
#define OPTION_VAL_3PX_DE "3 px"
#define OPTION_VAL_4PX_DE "4 px"
#define OPTION_VAL_5PX_DE "5 px"
#define OPTION_VAL_6PX_DE "6 px"
#define OPTION_VAL_7PX_DE "7 px"
#define OPTION_VAL_8PX_DE "8 px"
#define BEETLE_OPT_PGXP_NCLIP_LABEL_DE "PGXP Primitive Keulung"
#define BEETLE_OPT_PGXP_NCLIP_INFO_0_DE "PGXP-NCLIP-Implementierung verwenden. Verbessert die Darstellung durch Reduzierung von Lücken in Geometrien mit PGXP-Koordinaten. Kann dazu führen, dass sich einige Spiele unter Umständen aufhängen."
#define BEETLE_OPT_PGXP_VERTEX_LABEL_DE "PGXP-Vertex-Cache"
#define BEETLE_OPT_PGXP_VERTEX_INFO_0_DE "Cache-PGXP-erweiterte Vertex-Positionen zur Wiederverwendung über Polygon-Zeichnungen hinweg. Kann möglicherweise die Objektorientierung verbessern und die sichtbaren Nähte beim Rendern von Texturen reduzieren, aber falsche Positiven, wenn der Cache abfragt wird, können zu grafischen Fehlern führen. Es wird derzeit empfohlen, diese Option deaktiviert zu lassen. Diese Option wird nur angewendet, wenn der PGXP-Operationsmodus aktiviert ist. Unterstützt nur von Hardware-Renderern."
#define BEETLE_OPT_PGXP_TEXTURE_LABEL_DE "PGXP-Perspektivische Texturkorrektur"
#define BEETLE_OPT_PGXP_TEXTURE_INFO_0_DE "Ersetzt das native affine Textur-Mapping der PSX durch perspektivkorrektes Textur-Mapping. Beseitigt positionsabhängige Verzerrungen und Verformungen von Texturen, wodurch korrekt ausgerichtete Texturen entstehen. Diese Option wird nur angewendet, wenn der PGXP-Betriebsmodus aktiviert ist. Wird von den Hardware-Renderern und dem Software-Renderer unterstützt."
#define BEETLE_OPT_DISPLAY_INTERNAL_FPS_LABEL_DE "Interne FPS anzeigen"
#define BEETLE_OPT_DISPLAY_INTERNAL_FPS_INFO_0_DE "Die interne Bildrate anzeigen, mit der das emulierte PlayStation-System Inhalte rendert. Hinweis: Benötigt, dass Bildschirm-Benachrichtigungen im Libretro-Frontend aktiviert sind."
#define BEETLE_OPT_DISPLAY_OSD_LABEL_DE "OSD-Meldungen anzeigen"
#define BEETLE_OPT_DISPLAY_OSD_INFO_0_DE "OSD-Meldungen anzeigen, die vom Core generiert wurden."
#define BEETLE_OPT_DISPLAY_VRAM_LABEL_DE "Vollständigen VRAM anzeigen (Debugging)"
#define BEETLE_OPT_DISPLAY_VRAM_INFO_0_DE "Den gesamten VRAM der emulierten Konsole visualisieren. Wird nur von den OpenGL- und Vulkan-Hardware-Renderern unterstützt. Hinweis: Diese Funktion ist für Debugging-Zwecke gedacht und sollte normalerweise deaktiviert werden."
#define BEETLE_OPT_ANALOG_CALIBRATION_LABEL_DE "Analogeingaben-Selbstkalibrierung"
#define BEETLE_OPT_ANALOG_CALIBRATION_INFO_0_DE "Aktiviert die dynamische Kalibrierung der analogen Eingaben, wenn das Eingabegerät auf DualShock, Analog-Controller, Analog-Joystick oder neGcon eingestellt ist. Die maximalen registrierten Eingangswerte werden in Echtzeit überwacht und zur Skalierung der an den Emulator übergebenen analogen Koordinaten verwendet. Diese Option sollte für Spiele wie Mega Man Legends 2 verwendet werden, die größere Werte erwarten, als moderne Controller liefern. Um optimale Ergebnisse zu erzielen, sollten die Analogsticks jedes Mal, wenn Inhalte geladen werden, vollständig rotiert werden, um den Kalibrierungsalgorithmus einzustellen."
#define BEETLE_OPT_ANALOG_TOGGLE_LABEL_DE "DualShock-Analogmodusumschalter"
#define BEETLE_OPT_ANALOG_TOGGLE_INFO_0_DE "Wenn der Typ des Eingabegeräts DualShock ist, kann mit dieser Option der emulierte DualShock wie die Originalhardware zwischen DIGITAL- und ANALOG-Modus umgeschaltet werden. Der Modus kann auch über die ausgewählte Tastenkombination umgeschaltet werden."
#define OPTION_VAL_ENABLED_ANALOG_DE "Standardanalog"
#define BEETLE_OPT_ANALOG_TOGGLE_COMBO_LABEL_DE "DualShock Analogmodus-Combo"
#define BEETLE_OPT_ANALOG_TOGGLE_COMBO_INFO_0_DE "Die Tastenkombination, mit der zwischen DIGITAL- und ANALOG-Modus für den emulierten DualShock umgeschaltet werden soll."
#define OPTION_VAL_L1_L2_R1_R2_START_SELECT_DE NULL
#define OPTION_VAL_L1_R1_SELECT_DE NULL
#define OPTION_VAL_L1_R1_START_DE NULL
#define OPTION_VAL_L1_R1_L3_DE NULL
#define OPTION_VAL_L1_R1_R3_DE NULL
#define OPTION_VAL_L2_R2_SELECT_DE NULL
#define OPTION_VAL_L2_R2_START_DE NULL
#define OPTION_VAL_L2_R2_L3_DE NULL
#define OPTION_VAL_L2_R2_R3_DE NULL
#define OPTION_VAL_L3_R3_DE NULL
#define BEETLE_OPT_ANALOG_TOGGLE_HOLD_LABEL_DE "DualShock Analogmodus-Combo Halteverzögerung"
#define BEETLE_OPT_ANALOG_TOGGLE_HOLD_INFO_0_DE "Haltezeit für die Kombitasten im Analogmodus."
#define OPTION_VAL_0_DE "0 Sekunden Verzögerung"
#define OPTION_VAL_1_DE "1 Sekunde Verzögerung"
#define OPTION_VAL_2_DE "2 Sekunden Verzögerung"
#define OPTION_VAL_3_DE "3 Sekunden Verzögerung"
#define OPTION_VAL_4_DE "4 Sekunden Verzögerung"
#define OPTION_VAL_5_DE "5 Sekunden Verzögerung"
#define BEETLE_OPT_ENABLE_MULTITAP_PORT1_LABEL_DE "MultiTap auf Port 1"
#define BEETLE_OPT_ENABLE_MULTITAP_PORT1_INFO_0_DE "Multitap-Funktionalität auf Port 1 aktivieren."
#define BEETLE_OPT_ENABLE_MULTITAP_PORT2_LABEL_DE "MultiTap auf Port 2"
#define BEETLE_OPT_ENABLE_MULTITAP_PORT2_INFO_0_DE "Multitap-Funktionalität auf Port 2 aktivieren."
#define BEETLE_OPT_GUN_INPUT_MODE_LABEL_DE "Waffeneingabemodus"
#define BEETLE_OPT_GUN_INPUT_MODE_INFO_0_DE "Wenn der Gerätetyp auf „Guncon/G-Con 45” oder „Justifier” eingestellt ist, die mausgesteuerte Eingabe „Light Gun” oder „Touchscreen” verwenden."
#define OPTION_VAL_LIGHTGUN_DE "Lightgun"
#define OPTION_VAL_TOUCHSCREEN_DE NULL
#define BEETLE_OPT_GUN_CURSOR_LABEL_DE "Waffencursor"
#define BEETLE_OPT_GUN_CURSOR_INFO_0_DE "Beim Einsatz der Eingabegeräte „Guncon/G-Con 45” und „Justifier” einen Cursor auf dem Bildschirm zeichnen."
#define OPTION_VAL_CROSS_DE "Kreuz"
#define OPTION_VAL_DOT_DE "Punkt"
#define BEETLE_OPT_CROSSHAIR_COLOR_P1_LABEL_DE "Fadenkreuzfarbe: Port 1"
#define BEETLE_OPT_CROSSHAIR_COLOR_P1_INFO_0_DE "Fadenkreuzfarbe der Lightgun für Port 1."
#define OPTION_VAL_RED_DE "Rot"
#define OPTION_VAL_BLUE_DE "Blau"
#define OPTION_VAL_GREEN_DE "Grün"
#define OPTION_VAL_ORANGE_DE NULL
#define OPTION_VAL_YELLOW_DE "Gelb"
#define OPTION_VAL_CYAN_DE NULL
#define OPTION_VAL_PINK_DE NULL
#define OPTION_VAL_PURPLE_DE "Lila"
#define OPTION_VAL_BLACK_DE "Schwarz"
#define OPTION_VAL_WHITE_DE "Weiß"
#define BEETLE_OPT_CROSSHAIR_COLOR_P2_LABEL_DE "Fadenkreuzfarbe: Port 2"
#define BEETLE_OPT_CROSSHAIR_COLOR_P2_INFO_0_DE "Fadenkreuzfarbe der Lightgun für Port 2."
#define BEETLE_OPT_MOUSE_SENSITIVITY_LABEL_DE "Mausempfindlichkeit"
#define BEETLE_OPT_MOUSE_SENSITIVITY_INFO_0_DE "Empfindlichkeit des Eingabegerätetyps „Maus“."
#define OPTION_VAL_5_O42_DE "5 %"
#define OPTION_VAL_10_DE "10 %"
#define OPTION_VAL_15_DE "15 %"
#define OPTION_VAL_20_DE "20 %"
#define OPTION_VAL_25_DE "25 %"
#define OPTION_VAL_30_DE "30 %"
#define OPTION_VAL_35_DE "35 %"
#define OPTION_VAL_40_DE "40 %"
#define OPTION_VAL_45_DE "45 %"
#define OPTION_VAL_55_DE "55 %"
#define OPTION_VAL_65_DE "65 %"
#define OPTION_VAL_75_DE "75 %"
#define OPTION_VAL_85_DE "85 %"
#define OPTION_VAL_95_DE "95 %"
#define OPTION_VAL_100_O42_DE "100 %"
#define OPTION_VAL_105_DE "105 %"
#define OPTION_VAL_115_DE "115 %"
#define OPTION_VAL_125_DE "125 %"
#define OPTION_VAL_135_DE "135 %"
#define OPTION_VAL_145_DE "145 %"
#define OPTION_VAL_155_DE "155 %"
#define OPTION_VAL_165_DE "165 %"
#define OPTION_VAL_175_DE "175 %"
#define OPTION_VAL_185_DE "185 %"
#define OPTION_VAL_195_DE "195 %"
#define BEETLE_OPT_NEGCON_RESPONSE_LABEL_DE "neGcon-Verdrehungs-Rückmeldung"
#define BEETLE_OPT_NEGCON_RESPONSE_INFO_0_DE "Rückmeldungstyp des linken analogen Sticks des RetroPads, wenn die „Verdrehungs“-Aktion von emulierten „neGcon“-Eingabegeräten simuliert wird. Analogstick-Auslenkung kann entweder linear, quadratisch oder kubisch auf neGcon-Drehwinkel zugeordnet werden. „Quadratisch“ erlaubt eine höhere Präzision als „Linear“ bei kleinen Bewegungen. „Kubisch“ erhöht die kleine Bewegungsgenauigkeit weiter, aber „übertreibt“ größere Bewegungen. Hinweis: „Linear“ wird nur empfohlen bei Verwendung von Rennlenkrad-Peripheriegeräten. Konventionelle Controller implementieren die analoge Eingabe in einer Weise, die sich grundlegend von dem neGcon „Verdrehungs“-Mechanismus unterscheidet, so dass die lineare Zuordnung kleine Bewegungen übersteigt, wodurch die Feinkontrolle beeinträchtigt wird. In den meisten Fällen bietet „Quadratisch“ die engste Annäherung der realen Hardware."
#define OPTION_VAL_LINEAR_DE NULL
#define OPTION_VAL_QUADRATIC_DE "Quadratisch"
#define OPTION_VAL_CUBIC_DE "Kubisch"
#define BEETLE_OPT_NEGCON_DEADZONE_LABEL_DE "neGcon-Verdrehungs-Totzone"
#define BEETLE_OPT_NEGCON_DEADZONE_INFO_0_DE "Totzone des linken analogen Sticks des RetroPads bei der Simulation der „Verdrehungs“-Aktion von emulierten „neGcon“ Eingabegeräten. Wird verwendet, um den Controller-Drift-Effekt zu beseitigen. Hinweis: Die meisten neGcon-kompatiblen Titel bieten in-game-Optionen, um einen Totzonenwert für die „Verdrehung“ zu setzen. Um einen Verlust der Genauigkeit zu vermeiden, sollte die Totzone im Spiel *immer* auf Null gesetzt werden. Alle erforderlichen Anpassungen sollten *nur* über diese Coreoption angewendet werden. Dies ist besonders wichtig, wenn „neGcon-Verdrehungs-Rückmeldung“ auf „Quadratisch“ oder „Kubisch“ gesetzt wird."
#define OPTION_VAL_0_O44_DE "0 %"
#define BEETLE_OPT_USE_MEDNAFEN_MEMCARD0_METHOD_LABEL_DE "Speicherkartenmethode"
#define BEETLE_OPT_USE_MEDNAFEN_MEMCARD0_METHOD_INFO_0_DE "Das Speicherformat für den linken Speicherkartenslot. „Mednafen“ kann zur Kompatibilität mit der Stand-alone-Version von Mednafen verwendet werden. Bei Beetle PSX, Libretro (.srm) und Mednafen (.mcr) haben Speicherstände intern identische Formate und können über Umbenennen untereinander konvertiert werden. Neustart erforderlich."
#define OPTION_VAL_LIBRETRO_DE NULL
#define OPTION_VAL_MEDNAFEN_DE NULL
#define BEETLE_OPT_ENABLE_MEMCARD1_LABEL_DE "Speicherkartenslot 2"
#define BEETLE_OPT_ENABLE_MEMCARD1_INFO_0_DE "Eine zweite Speicherkarte im rechten Speicherkartenslot emulieren. Hinweis: Bei einigen Spielen muss diese Option für den korrekten Betrieb deaktiviert werden (z. B. Codename Tenka). Hinweis: Speicherkartenslot 2 verwendet das Mednafen (.mcr) Speicherformat. Neustart erforderlich."
#define BEETLE_OPT_SHARED_MEMORY_CARDS_LABEL_DE "Gemeinsame Speicherkarten"
#define BEETLE_OPT_SHARED_MEMORY_CARDS_INFO_0_DE "Gemeinsam genutzte Speicherkarten anstelle separater Karten für jeden Inhalt verwenden. Hinweis: Wenn „Speicherkartenmethode“ auf „Libretro“ eingestellt ist, ist nur Speicherkartenslot 2 betroffen. Neustart erforderlich."
#define BEETLE_OPT_MEMCARD_LEFT_INDEX_LABEL_DE "Index Speicherkartenslot 1"
#define BEETLE_OPT_MEMCARD_LEFT_INDEX_INFO_0_DE "Speicherkarte wird aktuell im linken Slot geladen. Diese Option funktioniert nur, wenn die Speicherkartenmethode auf Mednafen eingestellt ist."
#define OPTION_VAL_0_O48_DE "0 (Standard)"
#define BEETLE_OPT_MEMCARD_RIGHT_INDEX_LABEL_DE "Index Speicherkartenslot 2"
#define BEETLE_OPT_MEMCARD_RIGHT_INDEX_INFO_0_DE "Speicherkarte wird aktuell im rechten Slot geladen. Diese Option funktioniert nur, wenn Speicherkartenslot 2 aktiviert ist."
#define OPTION_VAL_1_O49_DE "1 (Standard)"
#define BEETLE_OPT_LINE_RENDER_LABEL_DE "Linie-zu-Viereck Hack"
#define BEETLE_OPT_LINE_RENDER_INFO_0_DE "Einige Spiele (z. B. Doom, Hexen, Soul Blade usw.) zeichnen horizontale Linien, indem sie einpixelhohe Dreiecke über den Bildschirm dehnen, die auf der Originalhardware als Pixelreihe gerastert werden. Dieser Hack erkennt diese kleinen Dreiecke und wandelt sie bei Bedarf in Vierecke um, sodass sie auf den Hardware-Renderern und in hochskalierten internen Auflösungen korrekt dargestellt werden können. Die Einstellung „Aggressiv“ ist für einige Titel (z. B. Dark Forces, Duke Nukem) erforderlich, kann aber andernfalls Grafikfehler verursachen. Im Zweifelsfall auf „Standard“ belassen."
#define OPTION_VAL_DEFAULT_DE "Standard"
#define OPTION_VAL_AGGRESSIVE_DE "Aggressiv"
#define BEETLE_OPT_WIDESCREEN_HACK_LABEL_DE "Breitbildmodus-Hack"
#define BEETLE_OPT_WIDESCREEN_HACK_INFO_0_DE "3D-Inhalte anamorph ausgeben und den emulierten Bildspeicher in einem Breitbild-Seitenverhältnis ausgeben. Erstellt beste Ergebnisse mit vollwertigen 3D-Spielen. 2D-Elemente werden horizontal gestreckt und können falsch ausgerichtet werden."
#define BEETLE_OPT_WIDESCREEN_HACK_ASPECT_RATIO_LABEL_DE "Seitenverhältnis für Breitbildmodus-Hack"
#define BEETLE_OPT_WIDESCREEN_HACK_ASPECT_RATIO_INFO_0_DE "Seitenverhältnis, das vom Breitbildmodus-Hack verwendet werden soll."
#define OPTION_VAL_16_9_DE NULL
#define OPTION_VAL_16_10_DE NULL
#define OPTION_VAL_18_9_DE NULL
#define OPTION_VAL_19_9_DE NULL
#define OPTION_VAL_20_9_DE NULL
#define OPTION_VAL_21_9_DE NULL
#define OPTION_VAL_32_9_DE NULL
#define BEETLE_OPT_CPU_DYNAREC_LABEL_DE "CPU-Dynarec"
#define BEETLE_OPT_CPU_DYNAREC_INFO_0_DE "CPU-Anweisungen dynamisch neu zu native Anweisungen kompilieren. Viel schneller als Interpreter, aber das CPU-Timing ist weniger genau und kann Fehler enthalten."
#define OPTION_VAL_DISABLED_DE "Deaktiviert (Beetle-Interpreter)"
#define OPTION_VAL_EXECUTE_DE "Max. Leistung"
#define OPTION_VAL_EXECUTE_ONE_DE "Zykluszeitprüfung"
#define OPTION_VAL_RUN_INTERPRETER_DE "Lightrec-Interpreter"
#define BEETLE_OPT_DYNAREC_INVALIDATE_LABEL_DE "Dynarec-Code-Invalidierung"
#define BEETLE_OPT_DYNAREC_INVALIDATE_INFO_0_DE "Einige Spiele erfordern „volle“ Invalidierung, andere benötigen „Nur DMA“."
#define OPTION_VAL_FULL_DE "Voll"
#define OPTION_VAL_DMA_DE "Nur DMA (etwas schneller)"
#define BEETLE_OPT_DYNAREC_EVENTCYCLES_LABEL_DE "Dynarec DMA/GPU/MDEC/Timer Event Zyklen"
#define BEETLE_OPT_DYNAREC_EVENTCYCLES_INFO_0_DE "Maximale Zyklen, die von der CPU ausgeführt werden, bevor ein GPU/DMA/MDEC/Timer Update überprüft wird. Eine höhere Zahl wird schneller sein, hat viel weniger Auswirkungen auf den beetle interpreter als dynarec."
#define OPTION_VAL_128_DE "128 (Standard)"
#define BEETLE_OPT_DYNAREC_SPU_SAMPLES_LABEL_DE "Dynarec SPU-Samples"
#define BEETLE_OPT_DYNAREC_SPU_SAMPLES_INFO_0_DE "Maximal laufende SPU-Samples bevor ein SPU-Update geprüft wird; eine höhere Zahl ist schneller, wird aber bei einigen Spielen Soundfehler mit etwas anderem als „1“ verursachen."
#define BEETLE_OPT_CORE_TIMING_FPS_LABEL_DE "Von Core gemeldetes FPS-Timing"
#define BEETLE_OPT_CORE_TIMING_FPS_INFO_0_DE "Automatische Umschaltung ermöglicht es dem Core, zwischen progressiver und interlaced Bildrate umzuschalten."
#define OPTION_VAL_AUTO_TOGGLE_DE "Automatisches Umschalten"
#define OPTION_VAL_FORCE_PROGRESSIVE_DE "Progressive Rate erzwingen"
#define OPTION_VAL_FORCE_INTERLACED_DE "„Interlaced“ Frequenz erzwingen"
#define BEETLE_OPT_DEINTERLACER_LABEL_DE "Deinterlacing-Methode"
#define BEETLE_OPT_DEINTERLACER_INFO_0_DE "„Weave“ zeigt beide Halbbilder im selben Bild an und liefert so ein Standbild in voller hoher Auflösung, während „Bob“ Abstriche bei der Bildschärfe zugunsten einer klareren Bewegungsdarstellung macht. „Bob (Offset)“ ist Bob, wobei das aktuelle Halbbild auf die volle Bildhöhe verdoppelt wird, mit einem Offset von einer Zeile zwischen den Halbbildern, wodurch die Libretro-Ausgabeauflösung beibehalten wird, während die flüssige Bewegung im Bob-Stil erhalten bleibt. „FastMAD (Motion Adaptive)“ verwebt Standpixel und interpoliert bewegte Pixel vertikal pixelweise, bei voller Ausgabehöhe, ohne den Kammeffekt bei Bewegung von Weave oder die halbierte Auflösung von Bob. „Off“ passt das Verhalten des Software-Renderers an das des Hardware-Renderers (Vulkan/OpenGL) an, indem jeder Frame auf alle VRAM-Zeilen rasterisiert und das Auslesen auf das Ende des Frames verschoben wird, wodurch Kammartefakte bei „interlaced“ Spielen beseitigt werden. Hinweis: Programme, die während des Scanouts DMA-zu-VRAM durchführen (einige Demos, die >512-Zeilen-PAL-Framebuffer-Wrap-Tricks verwenden) oder die bewegende Sprites über CPU-zu-VRAM-Bildübertragungen aktualisieren, die an das angezeigte Feld gebunden sind (z. B. Pro Pinball), können bei „Aus“ obsolete Zeilen oder Sprite-Ghosting anzeigen; in solchen Fällen „Weave“, „Bob“, „Bob (Offset)“ oder „FastMAD“ verwenden."
#define OPTION_VAL_WEAVE_DE NULL
#define OPTION_VAL_BOB_DE NULL
#define OPTION_VAL_BOB_OFFSET_DE NULL
#define OPTION_VAL_FASTMAD_DE "FastMAD (Bewegungsadaptiv)"
#define BEETLE_OPT_FRAME_DUPING_LABEL_DE "Bild-Duplizierung"
#define BEETLE_OPT_FRAME_DUPING_INFO_0_DE "Wiederholt das vorherige Bild, wenn der Core nichts Neues zum Anzeigen hat."
#define BEETLE_OPT_PAL_VIDEO_TIMING_OVERRIDE_LABEL_DE "PAL-Video-Timing-Übersteuerung"
#define BEETLE_OPT_PAL_VIDEO_TIMING_OVERRIDE_INFO_0_DE "Aufgrund unterschiedlicher Standards erscheinen PAL-Spiele im Vergleich zu den amerikanischen oder japanischen NTSC-Releases oft verlangsamt. Diese Option kann genutzt werden, um PAL-Timings zu übersteuern und zu versuchen, diese Spiele mit der NTSC-Bildwiederholrate auszuführen."
#define BEETLE_OPT_ASPECT_RATIO_LABEL_DE "Core-Seitenverhältnis"
#define BEETLE_OPT_ASPECT_RATIO_INFO_0_DE "Diese Einstellung wird ignoriert, wenn die Optionen „Breitbildmodus-Hack“ oder „Vollständigen VRAM anzeigen“ aktiviert sind."
#define OPTION_VAL_CORRECTED_DE "Korrigiert"
#define OPTION_VAL_UNCORRECTED_DE "Unkorrigiert"
#define OPTION_VAL_4_3_DE "4:3 erzwingen"
#define OPTION_VAL_NTSC_DE "NTSC erzwingen"
#define BEETLE_OPT_CROP_OVERSCAN_LABEL_DE "Overscan zuschneiden"
#define BEETLE_OPT_CROP_OVERSCAN_INFO_0_DE "Vollständiges natives Bild oder keine Balken."
#define OPTION_VAL_STATIC_DE NULL
#define OPTION_VAL_SMART_DE "Horizontal und Vertikal"
#define BEETLE_OPT_IMAGE_CROP_LABEL_DE "Zusätzliches Zuschneiden"
#define BEETLE_OPT_IMAGE_CROP_INFO_0_DE "Reduziert die Breite des zugeschnittenen Bildes um die angegebene Anzahl von Pixeln."
#define OPTION_VAL_9PX_DE "9 px"
#define OPTION_VAL_10PX_DE "10 px"
#define OPTION_VAL_11PX_DE "11 px"
#define OPTION_VAL_12PX_DE "12 px"
#define OPTION_VAL_13PX_DE "13 px"
#define OPTION_VAL_14PX_DE "14 px"
#define OPTION_VAL_15PX_DE "15 px"
#define OPTION_VAL_16PX_DE "16 px"
#define OPTION_VAL_17PX_DE "17 px"
#define OPTION_VAL_18PX_DE "18 px"
#define OPTION_VAL_19PX_DE "19 px"
#define OPTION_VAL_20PX_DE "20 px"
#define BEETLE_OPT_IMAGE_OFFSET_LABEL_DE "Versatz zugeschnittenes Bild"
#define BEETLE_OPT_IMAGE_OFFSET_INFO_0_DE "Versatz horizontal nach rechts (positiv) oder links (negativ) um die angegebene Anzahl Pixel. Kann zur Korrektur von Ausrichtungsproblemen verwendet werden. Wird nur vom Software-Renderer unterstützt."
#define OPTION_VAL_12PX_O64_DE "-12 px"
#define OPTION_VAL_11PX_O64_DE "-11 px"
#define OPTION_VAL_10PX_O64_DE "-10 px"
#define OPTION_VAL_9PX_O64_DE "-9 px"
#define OPTION_VAL_8PX_O64_DE "-8 px"
#define OPTION_VAL_7PX_O64_DE "-7 px"
#define OPTION_VAL_6PX_O64_DE "-6 px"
#define OPTION_VAL_5PX_O64_DE "-5 px"
#define OPTION_VAL_4PX_O64_DE "-4 px"
#define OPTION_VAL_3PX_O64_DE "-3 px"
#define OPTION_VAL_2PX_O64_DE "-2 px"
#define OPTION_VAL_1PX_O64_DE "-1 px"
#define OPTION_VAL_1PX_O64_0_DE "+1 px"
#define OPTION_VAL_2PX_O64_0_DE "+2 px"
#define OPTION_VAL_3PX_O64_0_DE "+3 px"
#define OPTION_VAL_4PX_O64_0_DE "+4 px"
#define OPTION_VAL_5PX_O64_0_DE "+5 px"
#define OPTION_VAL_6PX_O64_0_DE "+6 px"
#define OPTION_VAL_7PX_O64_0_DE "+7 px"
#define OPTION_VAL_8PX_O64_0_DE "+8 px"
#define OPTION_VAL_9PX_O64_0_DE "+9 px"
#define OPTION_VAL_10PX_O64_0_DE "+10 px"
#define OPTION_VAL_11PX_O64_0_DE "+11 px"
#define OPTION_VAL_12PX_O64_0_DE "+12 px"
#define BEETLE_OPT_IMAGE_OFFSET_CYCLES_LABEL_DE "Horizontaler Bildversatz (GPU-Zyklen)"
#define BEETLE_OPT_IMAGE_OFFSET_CYCLES_INFO_0_DE "Positive Werte verschieben das Bild nach rechts, negative Werte verschieben das Bild nach links. Wird nur von Hardware-Renderern unterstützt."
#define BEETLE_OPT_INITIAL_SCANLINE_LABEL_DE "Erste Abtastzeile - NTSC"
#define BEETLE_OPT_INITIAL_SCANLINE_INFO_0_DE "Die erste angezeigte Scanline beim Ausführen von NTSC-Inhalten. Werte größer als Null verringern die Höhe der Ausgabebilder durch Zuschneiden von Pixeln am oberen Rand. Kann verwendet werden, um schwarze Balken entgegenzuwirken."
#define BEETLE_OPT_LAST_SCANLINE_LABEL_DE "Letzte Abtastzeile - NTSC"
#define BEETLE_OPT_LAST_SCANLINE_INFO_0_DE "Die letzte angezeigte Scanline beim Ausführen von NTSC-Inhalten. Werte kleiner als 239 verringern die Höhe der Ausgabebilder durch Abschneiden von Pixeln vom unteren Rand. Kann verwendet werden, um schwarze Balken entgegenzuwirken."
#define OPTION_VAL_239_DE "239 (Standard)"
#define BEETLE_OPT_INITIAL_SCANLINE_PAL_LABEL_DE "Erste Abtastzeile - PAL"
#define BEETLE_OPT_INITIAL_SCANLINE_PAL_INFO_0_DE "Die erste angezeigte Scanline beim Ausführen von PAL-Inhalten. Werte größer als Null verringern die Höhe der Ausgabebilder durch Abschneiden von Pixeln am oberen Rand. Kann verwendet werden, um schwarze Balken entgegenzuwirken."
#define BEETLE_OPT_LAST_SCANLINE_PAL_LABEL_DE "Letzte Abtastzeile - PAL"
#define BEETLE_OPT_LAST_SCANLINE_PAL_INFO_0_DE "Zuletzt angezeigte Scanlinie bei der Wiedergabe von PAL-Inhalten. Werte unter 287 reduzieren die Höhe der Ausgabebilder, indem Pixel vom unteren Rand abgeschnitten werden. Kann verwendet werden, um Letterboxing entgegenzuwirken."
#define OPTION_VAL_287_DE "287 (Standard)"

struct retro_core_option_v2_category option_cats_de[] = {
   {
      "system",
      CATEGORY_SYSTEM_LABEL_DE,
      CATEGORY_SYSTEM_INFO_0_DE
   },
   {
      "video",
      CATEGORY_VIDEO_LABEL_DE,
      CATEGORY_VIDEO_INFO_0_DE
   },
   {
      "pgxp",
      CATEGORY_PGXP_LABEL_DE,
      CATEGORY_PGXP_INFO_0_DE
   },
   {
      "osd",
      CATEGORY_OSD_LABEL_DE,
      CATEGORY_OSD_INFO_0_DE
   },
   {
      "input",
      CATEGORY_INPUT_LABEL_DE,
      CATEGORY_INPUT_INFO_0_DE
   },
   {
      "memcards",
      CATEGORY_MEMCARDS_LABEL_DE,
      CATEGORY_MEMCARDS_INFO_0_DE
   },
   {
      "hacks",
      CATEGORY_HACKS_LABEL_DE,
      CATEGORY_HACKS_INFO_0_DE
   },
   { NULL, NULL, NULL },
};
struct retro_core_option_v2_definition option_defs_de[] = {
   {
      BEETLE_OPT(cpu_freq_scale),
      BEETLE_OPT_CPU_FREQ_SCALE_LABEL_DE,
      NULL,
      BEETLE_OPT_CPU_FREQ_SCALE_INFO_0_DE,
      NULL,
      "system",
      {
         { "50%",           OPTION_VAL_50_DE },
         { "60%",           OPTION_VAL_60_DE },
         { "70%",           OPTION_VAL_70_DE },
         { "80%",           OPTION_VAL_80_DE },
         { "90%",           OPTION_VAL_90_DE },
         { "100%",          OPTION_VAL_100_DE },
         { "110%",          OPTION_VAL_110_DE },
         { "120%",          OPTION_VAL_120_DE },
         { "130%",          OPTION_VAL_130_DE },
         { "140%",          OPTION_VAL_140_DE },
         { "150%",          OPTION_VAL_150_DE },
         { "160%",          OPTION_VAL_160_DE },
         { "170%",          OPTION_VAL_170_DE },
         { "180%",          OPTION_VAL_180_DE },
         { "190%",          OPTION_VAL_190_DE },
         { "200%",          OPTION_VAL_200_DE },
         { "210%",          OPTION_VAL_210_DE },
         { "220%",          OPTION_VAL_220_DE },
         { "230%",          OPTION_VAL_230_DE },
         { "240%",          OPTION_VAL_240_DE },
         { "250%",          OPTION_VAL_250_DE },
         { "260%",          OPTION_VAL_260_DE },
         { "270%",          OPTION_VAL_270_DE },
         { "280%",          OPTION_VAL_280_DE },
         { "290%",          OPTION_VAL_290_DE },
         { "300%",          OPTION_VAL_300_DE },
         { "310%",          OPTION_VAL_310_DE },
         { "320%",          OPTION_VAL_320_DE },
         { "330%",          OPTION_VAL_330_DE },
         { "340%",          OPTION_VAL_340_DE },
         { "350%",          OPTION_VAL_350_DE },
         { "360%",          OPTION_VAL_360_DE },
         { "370%",          OPTION_VAL_370_DE },
         { "380%",          OPTION_VAL_380_DE },
         { "390%",          OPTION_VAL_390_DE },
         { "400%",          OPTION_VAL_400_DE },
         { "410%",          OPTION_VAL_410_DE },
         { "420%",          OPTION_VAL_420_DE },
         { "430%",          OPTION_VAL_430_DE },
         { "440%",          OPTION_VAL_440_DE },
         { "450%",          OPTION_VAL_450_DE },
         { "460%",          OPTION_VAL_460_DE },
         { "470%",          OPTION_VAL_470_DE },
         { "480%",          OPTION_VAL_480_DE },
         { "490%",          OPTION_VAL_490_DE },
         { "500%",          OPTION_VAL_500_DE },
         { "510%",          OPTION_VAL_510_DE },
         { "520%",          OPTION_VAL_520_DE },
         { "530%",          OPTION_VAL_530_DE },
         { "540%",          OPTION_VAL_540_DE },
         { "550%",          OPTION_VAL_550_DE },
         { "560%",          OPTION_VAL_560_DE },
         { "570%",          OPTION_VAL_570_DE },
         { "580%",          OPTION_VAL_580_DE },
         { "590%",          OPTION_VAL_590_DE },
         { "600%",          OPTION_VAL_600_DE },
         { "610%",          OPTION_VAL_610_DE },
         { "620%",          OPTION_VAL_620_DE },
         { "630%",          OPTION_VAL_630_DE },
         { "640%",          OPTION_VAL_640_DE },
         { "650%",          OPTION_VAL_650_DE },
         { "660%",          OPTION_VAL_660_DE },
         { "670%",          OPTION_VAL_670_DE },
         { "680%",          OPTION_VAL_680_DE },
         { "690%",          OPTION_VAL_690_DE },
         { "700%",          OPTION_VAL_700_DE },
         { "710%",          OPTION_VAL_710_DE },
         { "720%",          OPTION_VAL_720_DE },
         { "730%",          OPTION_VAL_730_DE },
         { "740%",          OPTION_VAL_740_DE },
         { "750%",          OPTION_VAL_750_DE },
         { NULL, NULL },
      },
      "100%"
   },
   {
      BEETLE_OPT(gpu_overclock),
      BEETLE_OPT_GPU_OVERCLOCK_LABEL_DE,
      NULL,
      BEETLE_OPT_GPU_OVERCLOCK_INFO_0_DE,
      NULL,
      "system",
      {
         { "1x(native)", OPTION_VAL_1X_NATIVE_DE },
         { "2x",         OPTION_VAL_2X_DE },
         { "4x",         OPTION_VAL_4X_DE },
         { "8x",         OPTION_VAL_8X_DE },
         { "16x",        OPTION_VAL_16X_DE },
         { "32x",        OPTION_VAL_32X_DE },
         { NULL, NULL },
      },
      "1x(native)"
   },
   {
      BEETLE_OPT(gte_overclock),
      BEETLE_OPT_GTE_OVERCLOCK_LABEL_DE,
      NULL,
      BEETLE_OPT_GTE_OVERCLOCK_INFO_0_DE,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
#ifndef EMSCRIPTEN
   {
      BEETLE_OPT(cd_access_method),
      BEETLE_OPT_CD_ACCESS_METHOD_LABEL_DE,
      NULL,
      BEETLE_OPT_CD_ACCESS_METHOD_INFO_0_DE,
      NULL,
      "system",
      {
         { "sync",     OPTION_VAL_SYNC_DE },
         { "async",    OPTION_VAL_ASYNC_DE },
         { "precache", OPTION_VAL_PRECACHE_DE },
         { NULL, NULL },
      },
      "sync"
   },
#endif
   {
      BEETLE_OPT(cd_fastload),
      BEETLE_OPT_CD_FASTLOAD_LABEL_DE,
      NULL,
      BEETLE_OPT_CD_FASTLOAD_INFO_0_DE,
      NULL,
      "system",
      {
         { "2x(native)", OPTION_VAL_2X_NATIVE_DE },
         { "4x",          OPTION_VAL_4X_DE },
         { "6x",          OPTION_VAL_6X_DE },
         { "8x",          OPTION_VAL_8X_DE },
         { "10x",         OPTION_VAL_10X_DE },
         { "12x",         OPTION_VAL_12X_DE },
         { "14x",         OPTION_VAL_14X_DE },
         { NULL, NULL },
      },
      "2x(native)"
   },
   {
      BEETLE_OPT(skip_bios),
      BEETLE_OPT_SKIP_BIOS_LABEL_DE,
      NULL,
      BEETLE_OPT_SKIP_BIOS_INFO_0_DE,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(override_bios),
      BEETLE_OPT_OVERRIDE_BIOS_LABEL_DE,
      NULL,
      BEETLE_OPT_OVERRIDE_BIOS_INFO_0_DE,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "psxonpsp", OPTION_VAL_PSXONPSP_DE },
         { "ps1_rom",  OPTION_VAL_PS1_ROM_DE },
         { "openbios", OPTION_VAL_OPENBIOS_DE     },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(region),
      BEETLE_OPT_REGION_LABEL_DE,
      NULL,
      BEETLE_OPT_REGION_INFO_0_DE,
      NULL,
      "system",
      {
         { "auto",   OPTION_VAL_AUTO_DE },
         { "ntsc-j", OPTION_VAL_NTSC_J_DE },
         { "ntsc-u", OPTION_VAL_NTSC_U_DE },
         { "pal",    OPTION_VAL_PAL_DE },
         { NULL, NULL },
      },
      "auto"
   },
   {
      BEETLE_OPT(internal_resolution),
      BEETLE_OPT_INTERNAL_RESOLUTION_LABEL_DE,
      NULL,
      BEETLE_OPT_INTERNAL_RESOLUTION_INFO_0_DE,
      NULL,
      "video",
      {
         { "1x(native)", OPTION_VAL_1X_NATIVE_DE },
         { "2x",         OPTION_VAL_2X_DE },
         { "4x",         OPTION_VAL_4X_DE },
         { "8x",         OPTION_VAL_8X_DE },
         { "16x",        OPTION_VAL_16X_DE },
         { NULL, NULL },
      },
      "1x(native)"
   },
#if defined(HAVE_OPENGL) || defined(HAVE_OPENGLES) || defined(HAVE_VULKAN)
   {
      BEETLE_OPT(renderer),
      BEETLE_OPT_RENDERER_LABEL_DE,
      NULL,
      BEETLE_OPT_RENDERER_INFO_0_DE,
      NULL,
      "video",
      {
         { "hardware",    OPTION_VAL_HARDWARE_DE },
#if defined(HAVE_OPENGL) || defined(HAVE_OPENGLES)
         { "hardware_gl", OPTION_VAL_HARDWARE_GL_DE },
#endif
#if defined(HAVE_VULKAN)
         { "hardware_vk", OPTION_VAL_HARDWARE_VK_DE },
#endif
         { "software",    OPTION_VAL_SOFTWARE_DE },
         { NULL, NULL },
      },
      "hardware"
   },
   {
      BEETLE_OPT(renderer_software_fb),
      BEETLE_OPT_RENDERER_SOFTWARE_FB_LABEL_DE,
      NULL,
      BEETLE_OPT_RENDERER_SOFTWARE_FB_INFO_0_DE,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled"
   },
#endif
#if defined(HAVE_OPENGL) || defined(HAVE_OPENGLES)
   {
      BEETLE_OPT(depth),
      BEETLE_OPT_DEPTH_LABEL_DE,
      NULL,
      BEETLE_OPT_DEPTH_INFO_0_DE,
      NULL,
      "video",
      {
         { "16bpp(native)", OPTION_VAL_16BPP_NATIVE_DE },
         { "32bpp",         OPTION_VAL_32BPP_DE },
         { NULL, NULL },
      },
      "16bpp(native)"
   },
   // Sort of, it's more like 15-bit high color and 24-bit true color for visible output. The alpha channel is used for mask bit. Vulkan renderer uses ABGR1555_555 for 31 bits internal? FMVs are always 24-bit on all renderers like original hardware (BGR888, no alpha)
#endif
   {
      BEETLE_OPT(dither_mode),
      BEETLE_OPT_DITHER_MODE_LABEL_DE,
      NULL,
      BEETLE_OPT_DITHER_MODE_INFO_0_DE,
      NULL,
      "video",
      {
         { "1x(native)",          OPTION_VAL_1X_NATIVE_DE },
         { "internal resolution", OPTION_VAL_INTERNAL_RESOLUTION_DE },
         { "disabled",            NULL },
         { NULL, NULL },
      },
      "1x(native)"
   },
#ifdef HAVE_VULKAN
   {
      BEETLE_OPT(scaled_uv_offset),
      BEETLE_OPT_SCALED_UV_OFFSET_LABEL_DE,
      NULL,
      BEETLE_OPT_SCALED_UV_OFFSET_INFO_0_DE,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled"
   },
#endif
#if defined(HAVE_OPENGL) || defined(HAVE_OPENGLES) || defined(HAVE_VULKAN)
   {
      BEETLE_OPT(filter),
      BEETLE_OPT_FILTER_LABEL_DE,
      NULL,
      BEETLE_OPT_FILTER_INFO_0_DE,
      NULL,
      "video",
      {
         { "nearest",  OPTION_VAL_NEAREST_DE },
         { "SABR",     OPTION_VAL_SABR_DE },
         { "xBR",      OPTION_VAL_XBR_DE },
         { "bilinear", OPTION_VAL_BILINEAR_DE },
         { "3-point",  OPTION_VAL_3_POINT_DE },
         { "JINC2",    OPTION_VAL_JINC2_DE },
         { NULL, NULL },
      },
      "nearest"
   },
#ifdef HAVE_VULKAN
   {
      BEETLE_OPT(filter_exclude_sprite),
      BEETLE_OPT_FILTER_EXCLUDE_SPRITE_LABEL_DE,
      NULL,
      BEETLE_OPT_FILTER_EXCLUDE_SPRITE_INFO_0_DE,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "opaque", OPTION_VAL_OPAQUE_DE },
         { "all", OPTION_VAL_ALL_DE },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(filter_exclude_2d_polygon),
      BEETLE_OPT_FILTER_EXCLUDE_2D_POLYGON_LABEL_DE,
      NULL,
      BEETLE_OPT_FILTER_EXCLUDE_2D_POLYGON_INFO_0_DE,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "opaque", OPTION_VAL_OPAQUE_DE },
         { "all", OPTION_VAL_ALL_DE },
         { NULL, NULL },
      },
      "disabled"
   },
#endif
#endif
#ifdef HAVE_VULKAN
   {
      BEETLE_OPT(adaptive_smoothing),
      BEETLE_OPT_ADAPTIVE_SMOOTHING_LABEL_DE,
      NULL,
      BEETLE_OPT_ADAPTIVE_SMOOTHING_INFO_0_DE,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(super_sampling),
      BEETLE_OPT_SUPER_SAMPLING_LABEL_DE,
      NULL,
      BEETLE_OPT_SUPER_SAMPLING_INFO_0_DE,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(msaa),
      BEETLE_OPT_MSAA_LABEL_DE,
      NULL,
      BEETLE_OPT_MSAA_INFO_0_DE,
      NULL,
      "video",
      {
         { "1x",  OPTION_VAL_1X_DE },
         { "2x",  OPTION_VAL_2X_DE },
         { "4x",  OPTION_VAL_4X_DE },
         { "8x",  OPTION_VAL_8X_DE },
         { "16x", OPTION_VAL_16X_DE },
         { NULL, NULL },
      },
      "1x"
   },
   {
      BEETLE_OPT(mdec_yuv),
      BEETLE_OPT_MDEC_YUV_LABEL_DE,
      NULL,
      BEETLE_OPT_MDEC_YUV_INFO_0_DE,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(track_textures),
      BEETLE_OPT_TRACK_TEXTURES_LABEL_DE,
      NULL,
      BEETLE_OPT_TRACK_TEXTURES_INFO_0_DE,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
#ifdef TEXTURE_DUMPING_ENABLED
   {
      BEETLE_OPT(dump_textures),
      BEETLE_OPT_DUMP_TEXTURES_LABEL_DE,
      NULL,
      BEETLE_OPT_DUMP_TEXTURES_INFO_0_DE,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
#endif
   {
      BEETLE_OPT(replace_textures),
      BEETLE_OPT_REPLACE_TEXTURES_LABEL_DE,
      NULL,
      BEETLE_OPT_REPLACE_TEXTURES_INFO_0_DE,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
#endif
   {
      BEETLE_OPT(pgxp_mode),
      BEETLE_OPT_PGXP_MODE_LABEL_DE,
      NULL,
      BEETLE_OPT_PGXP_MODE_INFO_0_DE,
      NULL,
      "pgxp",
      {
         { "disabled",     NULL },
         { "memory only",  OPTION_VAL_MEMORY_ONLY_DE },
         { "memory + CPU", OPTION_VAL_MEMORY_CPU_DE },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(pgxp_2d_tol),
      BEETLE_OPT_PGXP_2D_TOL_LABEL_DE,
      NULL,
      BEETLE_OPT_PGXP_2D_TOL_INFO_0_DE,
      NULL,
      "pgxp",
      {
         { "disabled", NULL },
         { "0px", OPTION_VAL_0PX_DE },
         { "1px", OPTION_VAL_1PX_DE },
         { "2px", OPTION_VAL_2PX_DE },
         { "3px", OPTION_VAL_3PX_DE },
         { "4px", OPTION_VAL_4PX_DE },
         { "5px", OPTION_VAL_5PX_DE },
         { "6px", OPTION_VAL_6PX_DE },
         { "7px", OPTION_VAL_7PX_DE },
         { "8px", OPTION_VAL_8PX_DE },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(pgxp_nclip),
      BEETLE_OPT_PGXP_NCLIP_LABEL_DE,
      NULL,
      BEETLE_OPT_PGXP_NCLIP_INFO_0_DE,
      NULL,
      "pgxp",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
#if defined(HAVE_OPENGL) || defined(HAVE_OPENGLES) || defined(HAVE_VULKAN)
   {
      BEETLE_OPT(pgxp_vertex),
      BEETLE_OPT_PGXP_VERTEX_LABEL_DE,
      NULL,
      BEETLE_OPT_PGXP_VERTEX_INFO_0_DE,
      NULL,
      "pgxp",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
#endif
   {
      BEETLE_OPT(pgxp_texture),
      BEETLE_OPT_PGXP_TEXTURE_LABEL_DE,
      NULL,
      BEETLE_OPT_PGXP_TEXTURE_INFO_0_DE,
      NULL,
      "pgxp",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(display_internal_fps),
      BEETLE_OPT_DISPLAY_INTERNAL_FPS_LABEL_DE,
      NULL,
      BEETLE_OPT_DISPLAY_INTERNAL_FPS_INFO_0_DE,
      NULL,
      "osd",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(display_osd),
      BEETLE_OPT_DISPLAY_OSD_LABEL_DE,
      NULL,
      BEETLE_OPT_DISPLAY_OSD_INFO_0_DE,
      NULL,
      "osd",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled"
   },
#if defined(HAVE_OPENGL) || defined(HAVE_OPENGLES) || defined(HAVE_VULKAN)
   {
      BEETLE_OPT(display_vram),
      BEETLE_OPT_DISPLAY_VRAM_LABEL_DE,
      NULL,
      BEETLE_OPT_DISPLAY_VRAM_INFO_0_DE,
      NULL,
      "osd",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
#endif
   {
      BEETLE_OPT(analog_calibration),
      BEETLE_OPT_ANALOG_CALIBRATION_LABEL_DE,
      NULL,
      BEETLE_OPT_ANALOG_CALIBRATION_INFO_0_DE,
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(analog_toggle),
      BEETLE_OPT_ANALOG_TOGGLE_LABEL_DE,
      NULL,
      BEETLE_OPT_ANALOG_TOGGLE_INFO_0_DE,
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { "enabled-analog", OPTION_VAL_ENABLED_ANALOG_DE },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(analog_toggle_combo),
      BEETLE_OPT_ANALOG_TOGGLE_COMBO_LABEL_DE,
      NULL,
      BEETLE_OPT_ANALOG_TOGGLE_COMBO_INFO_0_DE,
      NULL,
      "input",
      {
         { "l1+l2+r1+r2+start+select", OPTION_VAL_L1_L2_R1_R2_START_SELECT_DE },
         { "l1+r1+select",             OPTION_VAL_L1_R1_SELECT_DE },
         { "l1+r1+start",              OPTION_VAL_L1_R1_START_DE },
         { "l1+r1+l3",                 OPTION_VAL_L1_R1_L3_DE },
         { "l1+r1+r3",                 OPTION_VAL_L1_R1_R3_DE },
         { "l2+r2+select",             OPTION_VAL_L2_R2_SELECT_DE },
         { "l2+r2+start",              OPTION_VAL_L2_R2_START_DE },
         { "l2+r2+l3",                 OPTION_VAL_L2_R2_L3_DE },
         { "l2+r2+r3",                 OPTION_VAL_L2_R2_R3_DE },
         { "l3+r3",                    OPTION_VAL_L3_R3_DE },
         { NULL, NULL },
      },
      "l1+r1+select"
   },
   {
      BEETLE_OPT(analog_toggle_hold),
      BEETLE_OPT_ANALOG_TOGGLE_HOLD_LABEL_DE,
      NULL,
      BEETLE_OPT_ANALOG_TOGGLE_HOLD_INFO_0_DE,
      NULL,
      "input",
      {
         { "0", OPTION_VAL_0_DE },
         { "1", OPTION_VAL_1_DE },
         { "2", OPTION_VAL_2_DE },
         { "3", OPTION_VAL_3_DE },
         { "4", OPTION_VAL_4_DE },
         { "5", OPTION_VAL_5_DE },
         { NULL, NULL },
      },
      "1"
   },
   {
      BEETLE_OPT(enable_multitap_port1),
      BEETLE_OPT_ENABLE_MULTITAP_PORT1_LABEL_DE,
      NULL,
      BEETLE_OPT_ENABLE_MULTITAP_PORT1_INFO_0_DE,
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(enable_multitap_port2),
      BEETLE_OPT_ENABLE_MULTITAP_PORT2_LABEL_DE,
      NULL,
      BEETLE_OPT_ENABLE_MULTITAP_PORT2_INFO_0_DE,
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(gun_input_mode),
      BEETLE_OPT_GUN_INPUT_MODE_LABEL_DE,
      NULL,
      BEETLE_OPT_GUN_INPUT_MODE_INFO_0_DE,
      NULL,
      "input",
      {
         { "lightgun",    OPTION_VAL_LIGHTGUN_DE },
         { "touchscreen", OPTION_VAL_TOUCHSCREEN_DE },
         { NULL, NULL },
      },
      "lightgun"
   },
   // Shouldn't the gun_input_mode just be Mouse vs. Touchscreen?
   {
      BEETLE_OPT(gun_cursor),
      BEETLE_OPT_GUN_CURSOR_LABEL_DE,
      NULL,
      BEETLE_OPT_GUN_CURSOR_INFO_0_DE,
      NULL,
      "input",
      {
         { "off",   "disabled" },
         { "cross", OPTION_VAL_CROSS_DE },
         { "dot",   OPTION_VAL_DOT_DE },
         { NULL, NULL },
      },
      "cross"
   },
   {
      BEETLE_OPT(crosshair_color_p1),
      BEETLE_OPT_CROSSHAIR_COLOR_P1_LABEL_DE,
      NULL,
      BEETLE_OPT_CROSSHAIR_COLOR_P1_INFO_0_DE,
      NULL,
      "input",
      {
         { "red",    OPTION_VAL_RED_DE },
         { "blue",   OPTION_VAL_BLUE_DE },
         { "green",  OPTION_VAL_GREEN_DE },
         { "orange", OPTION_VAL_ORANGE_DE },
         { "yellow", OPTION_VAL_YELLOW_DE },
         { "cyan",   OPTION_VAL_CYAN_DE },
         { "pink",   OPTION_VAL_PINK_DE },
         { "purple", OPTION_VAL_PURPLE_DE },
         { "black",  OPTION_VAL_BLACK_DE },
         { "white",  OPTION_VAL_WHITE_DE },
         { NULL, NULL },
      },
      "red"
   },
   {
      BEETLE_OPT(crosshair_color_p2),
      BEETLE_OPT_CROSSHAIR_COLOR_P2_LABEL_DE,
      NULL,
      BEETLE_OPT_CROSSHAIR_COLOR_P2_INFO_0_DE,
      NULL,
      "input",
      {
         { "blue",   OPTION_VAL_BLUE_DE },
         { "red",    OPTION_VAL_RED_DE },
         { "green",  OPTION_VAL_GREEN_DE },
         { "orange", OPTION_VAL_ORANGE_DE },
         { "yellow", OPTION_VAL_YELLOW_DE },
         { "cyan",   OPTION_VAL_CYAN_DE },
         { "pink",   OPTION_VAL_PINK_DE },
         { "purple", OPTION_VAL_PURPLE_DE },
         { "black",  OPTION_VAL_BLACK_DE },
         { "white",  OPTION_VAL_WHITE_DE },
         { NULL, NULL },
      },
      "blue"
   },
   {
      BEETLE_OPT(mouse_sensitivity),
      BEETLE_OPT_MOUSE_SENSITIVITY_LABEL_DE,
      NULL,
      BEETLE_OPT_MOUSE_SENSITIVITY_INFO_0_DE,
      NULL,
      "input",
      {
         { "5%",   OPTION_VAL_5_O42_DE },
         { "10%",  OPTION_VAL_10_DE },
         { "15%",  OPTION_VAL_15_DE },
         { "20%",  OPTION_VAL_20_DE },
         { "25%",  OPTION_VAL_25_DE },
         { "30%",  OPTION_VAL_30_DE },
         { "35%",  OPTION_VAL_35_DE },
         { "40%",  OPTION_VAL_40_DE },
         { "45%",  OPTION_VAL_45_DE },
         { "50%",  OPTION_VAL_50_DE },
         { "55%",  OPTION_VAL_55_DE },
         { "60%",  OPTION_VAL_60_DE },
         { "65%",  OPTION_VAL_65_DE },
         { "70%",  OPTION_VAL_70_DE },
         { "75%",  OPTION_VAL_75_DE },
         { "80%",  OPTION_VAL_80_DE },
         { "85%",  OPTION_VAL_85_DE },
         { "90%",  OPTION_VAL_90_DE },
         { "95%",  OPTION_VAL_95_DE },
         { "100%", OPTION_VAL_100_O42_DE },
         { "105%", OPTION_VAL_105_DE },
         { "110%", OPTION_VAL_110_DE },
         { "115%", OPTION_VAL_115_DE },
         { "120%", OPTION_VAL_120_DE },
         { "125%", OPTION_VAL_125_DE },
         { "130%", OPTION_VAL_130_DE },
         { "135%", OPTION_VAL_135_DE },
         { "140%", OPTION_VAL_140_DE },
         { "145%", OPTION_VAL_145_DE },
         { "150%", OPTION_VAL_150_DE },
         { "155%", OPTION_VAL_155_DE },
         { "160%", OPTION_VAL_160_DE },
         { "165%", OPTION_VAL_165_DE },
         { "170%", OPTION_VAL_170_DE },
         { "175%", OPTION_VAL_175_DE },
         { "180%", OPTION_VAL_180_DE },
         { "185%", OPTION_VAL_185_DE },
         { "190%", OPTION_VAL_190_DE },
         { "195%", OPTION_VAL_195_DE },
         { "200%", OPTION_VAL_200_DE },
         { NULL, NULL },
      },
      "100%"
   },
   {
      BEETLE_OPT(negcon_response),
      BEETLE_OPT_NEGCON_RESPONSE_LABEL_DE,
      NULL,
      BEETLE_OPT_NEGCON_RESPONSE_INFO_0_DE,
      NULL,
      "input",
      {
         { "linear",    OPTION_VAL_LINEAR_DE },
         { "quadratic", OPTION_VAL_QUADRATIC_DE },
         { "cubic",     OPTION_VAL_CUBIC_DE },
         { NULL, NULL },
      },
      "linear"
   },
   {
      BEETLE_OPT(negcon_deadzone),
      BEETLE_OPT_NEGCON_DEADZONE_LABEL_DE,
      NULL,
      BEETLE_OPT_NEGCON_DEADZONE_INFO_0_DE,
      NULL,
      "input",
      {
         { "0%",  OPTION_VAL_0_O44_DE },
         { "5%",  OPTION_VAL_5_O42_DE },
         { "10%", OPTION_VAL_10_DE },
         { "15%", OPTION_VAL_15_DE },
         { "20%", OPTION_VAL_20_DE },
         { "25%", OPTION_VAL_25_DE },
         { "30%", OPTION_VAL_30_DE },
         { NULL, NULL },
      },
      "0%"
   },
   {
      BEETLE_OPT(use_mednafen_memcard0_method),
      BEETLE_OPT_USE_MEDNAFEN_MEMCARD0_METHOD_LABEL_DE,
      NULL,
      BEETLE_OPT_USE_MEDNAFEN_MEMCARD0_METHOD_INFO_0_DE,
      NULL,
      "memcards",
      {
         { "libretro", OPTION_VAL_LIBRETRO_DE },
         { "mednafen", OPTION_VAL_MEDNAFEN_DE },
         { NULL, NULL },
      },
      "libretro"
   },
   {
      BEETLE_OPT(enable_memcard1),
      BEETLE_OPT_ENABLE_MEMCARD1_LABEL_DE,
      NULL,
      BEETLE_OPT_ENABLE_MEMCARD1_INFO_0_DE,
      NULL,
      "memcards",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(shared_memory_cards),
      BEETLE_OPT_SHARED_MEMORY_CARDS_LABEL_DE,
      NULL,
      BEETLE_OPT_SHARED_MEMORY_CARDS_INFO_0_DE,
      NULL,
      "memcards",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(memcard_left_index),
      BEETLE_OPT_MEMCARD_LEFT_INDEX_LABEL_DE,
      NULL,
      BEETLE_OPT_MEMCARD_LEFT_INDEX_INFO_0_DE,
      NULL,
      "memcards",
      {
         { "0",  OPTION_VAL_0_O48_DE },
         { "1",  NULL },
         { "2",  NULL },
         { "3",  NULL },
         { "4",  NULL },
         { "5",  NULL },
         { "6",  NULL },
         { "7",  NULL },
         { "8",  NULL },
         { "9",  NULL },
         { "10", NULL },
         { "11", NULL },
         { "12", NULL },
         { "13", NULL },
         { "14", NULL },
         { "15", NULL },
         { "16", NULL },
         { "17", NULL },
         { "18", NULL },
         { "19", NULL },
         { "20", NULL },
         { "21", NULL },
         { "22", NULL },
         { "23", NULL },
         { "24", NULL },
         { "25", NULL },
         { "26", NULL },
         { "27", NULL },
         { "28", NULL },
         { "29", NULL },
         { "30", NULL },
         { "31", NULL },
         { "32", NULL },
         { "33", NULL },
         { "34", NULL },
         { "35", NULL },
         { "36", NULL },
         { "37", NULL },
         { "38", NULL },
         { "39", NULL },
         { "40", NULL },
         { "41", NULL },
         { "42", NULL },
         { "43", NULL },
         { "44", NULL },
         { "45", NULL },
         { "46", NULL },
         { "47", NULL },
         { "48", NULL },
         { "49", NULL },
         { "50", NULL },
         { "51", NULL },
         { "52", NULL },
         { "53", NULL },
         { "54", NULL },
         { "55", NULL },
         { "56", NULL },
         { "57", NULL },
         { "58", NULL },
         { "59", NULL },
         { "60", NULL },
         { "61", NULL },
         { "62", NULL },
         { "63", NULL },
         { NULL, NULL },
      },
      "0"
   },
   {
      BEETLE_OPT(memcard_right_index),
      BEETLE_OPT_MEMCARD_RIGHT_INDEX_LABEL_DE,
      NULL,
      BEETLE_OPT_MEMCARD_RIGHT_INDEX_INFO_0_DE,
      NULL,
      "memcards",
      {
         { "0",  NULL },
         { "1",  OPTION_VAL_1_O49_DE },
         { "2",  NULL },
         { "3",  NULL },
         { "4",  NULL },
         { "5",  NULL },
         { "6",  NULL },
         { "7",  NULL },
         { "8",  NULL },
         { "9",  NULL },
         { "10", NULL },
         { "11", NULL },
         { "12", NULL },
         { "13", NULL },
         { "14", NULL },
         { "15", NULL },
         { "16", NULL },
         { "17", NULL },
         { "18", NULL },
         { "19", NULL },
         { "20", NULL },
         { "21", NULL },
         { "22", NULL },
         { "23", NULL },
         { "24", NULL },
         { "25", NULL },
         { "26", NULL },
         { "27", NULL },
         { "28", NULL },
         { "29", NULL },
         { "30", NULL },
         { "31", NULL },
         { "32", NULL },
         { "33", NULL },
         { "34", NULL },
         { "35", NULL },
         { "36", NULL },
         { "37", NULL },
         { "38", NULL },
         { "39", NULL },
         { "40", NULL },
         { "41", NULL },
         { "42", NULL },
         { "43", NULL },
         { "44", NULL },
         { "45", NULL },
         { "46", NULL },
         { "47", NULL },
         { "48", NULL },
         { "49", NULL },
         { "50", NULL },
         { "51", NULL },
         { "52", NULL },
         { "53", NULL },
         { "54", NULL },
         { "55", NULL },
         { "56", NULL },
         { "57", NULL },
         { "58", NULL },
         { "59", NULL },
         { "60", NULL },
         { "61", NULL },
         { "62", NULL },
         { "63", NULL },
         { NULL, NULL },
      },
      "1"
   },
   {
      BEETLE_OPT(line_render),
      BEETLE_OPT_LINE_RENDER_LABEL_DE,
      NULL,
      BEETLE_OPT_LINE_RENDER_INFO_0_DE,
      NULL,
      "hacks",
      {
         { "default",    OPTION_VAL_DEFAULT_DE },
         { "aggressive", OPTION_VAL_AGGRESSIVE_DE },
         { "disabled",   NULL },
         { NULL, NULL },
      },
      "default"
   },
   {
      BEETLE_OPT(widescreen_hack),
      BEETLE_OPT_WIDESCREEN_HACK_LABEL_DE,
      NULL,
      BEETLE_OPT_WIDESCREEN_HACK_INFO_0_DE,
      NULL,
      "hacks",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(widescreen_hack_aspect_ratio),
      BEETLE_OPT_WIDESCREEN_HACK_ASPECT_RATIO_LABEL_DE,
      NULL,
      BEETLE_OPT_WIDESCREEN_HACK_ASPECT_RATIO_INFO_0_DE,
      NULL,
      "hacks",
      {
         { "16:9",  OPTION_VAL_16_9_DE },
         { "16:10", OPTION_VAL_16_10_DE },
         { "18:9",  OPTION_VAL_18_9_DE },
         { "19:9",  OPTION_VAL_19_9_DE },
         { "20:9",  OPTION_VAL_20_9_DE },
         { "21:9",  OPTION_VAL_21_9_DE }, // 64:27
         { "32:9",  OPTION_VAL_32_9_DE },
         { NULL,    NULL },
      },
      "16:9"
   },
#if defined(HAVE_LIGHTREC)
   {
      BEETLE_OPT(cpu_dynarec),
      BEETLE_OPT_CPU_DYNAREC_LABEL_DE,
      NULL,
      BEETLE_OPT_CPU_DYNAREC_INFO_0_DE,
      NULL,
      "hacks",
      {
         { "disabled",        OPTION_VAL_DISABLED_DE },
         { "execute",         OPTION_VAL_EXECUTE_DE },
         { "execute_one",     OPTION_VAL_EXECUTE_ONE_DE },
         { "run_interpreter", OPTION_VAL_RUN_INTERPRETER_DE },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(dynarec_invalidate),
      BEETLE_OPT_DYNAREC_INVALIDATE_LABEL_DE,
      NULL,
      BEETLE_OPT_DYNAREC_INVALIDATE_INFO_0_DE,
      NULL,
      "hacks",
      {
         { "full", OPTION_VAL_FULL_DE },
         { "dma",  OPTION_VAL_DMA_DE },
         { NULL, NULL },
      },
      "full"
   },
   {
      BEETLE_OPT(dynarec_eventcycles),
      BEETLE_OPT_DYNAREC_EVENTCYCLES_LABEL_DE,
      NULL,
      BEETLE_OPT_DYNAREC_EVENTCYCLES_INFO_0_DE,
      NULL,
      "hacks",
      {
         { "128",  OPTION_VAL_128_DE },
         { "256",  NULL },
         { "384",  NULL },
         { "512",  NULL },
         { "640",  NULL },
         { "768",  NULL },
         { "896",  NULL },
         { "1024", NULL },
         { "1152", NULL },
         { "1280", NULL },
         { "1408", NULL },
         { "1536", NULL },
         { "1664", NULL },
         { "1792", NULL },
         { "1920", NULL },
         { "2048", NULL },
         { NULL, NULL },
      },
      "128"
   },
   {
      BEETLE_OPT(dynarec_spu_samples),
      BEETLE_OPT_DYNAREC_SPU_SAMPLES_LABEL_DE,
      NULL,
      BEETLE_OPT_DYNAREC_SPU_SAMPLES_INFO_0_DE,
      NULL,
      "hacks",
      {
         { "1",  OPTION_VAL_1_O49_DE },
         { "4",  NULL },
         { "16", NULL },
         { NULL, NULL },
      },
      "1"
   },
#endif
   {
      BEETLE_OPT(core_timing_fps),
      BEETLE_OPT_CORE_TIMING_FPS_LABEL_DE,
      NULL,
      BEETLE_OPT_CORE_TIMING_FPS_INFO_0_DE,
      NULL,
      "video",
      {
         { "auto_toggle",       OPTION_VAL_AUTO_TOGGLE_DE },
         { "force_progressive", OPTION_VAL_FORCE_PROGRESSIVE_DE },
         { "force_interlaced",  OPTION_VAL_FORCE_INTERLACED_DE },
      },
      "auto_toggle"
   },
   {
      BEETLE_OPT(deinterlacer),
      BEETLE_OPT_DEINTERLACER_LABEL_DE,
      NULL,
      BEETLE_OPT_DEINTERLACER_INFO_0_DE,
      NULL,
      "video",
      {
         { "weave",      OPTION_VAL_WEAVE_DE },
         { "bob",        OPTION_VAL_BOB_DE },
         { "bob_offset", OPTION_VAL_BOB_OFFSET_DE },
         { "fastmad",    OPTION_VAL_FASTMAD_DE },
         { "off",        "Off" },
         { NULL, NULL },
      },
      "weave"
   },
   {
      BEETLE_OPT(frame_duping),
      BEETLE_OPT_FRAME_DUPING_LABEL_DE,
      NULL,
      BEETLE_OPT_FRAME_DUPING_INFO_0_DE,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(pal_video_timing_override),
      BEETLE_OPT_PAL_VIDEO_TIMING_OVERRIDE_LABEL_DE,
      NULL,
      BEETLE_OPT_PAL_VIDEO_TIMING_OVERRIDE_INFO_0_DE,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(aspect_ratio),
      BEETLE_OPT_ASPECT_RATIO_LABEL_DE,
      NULL,
      BEETLE_OPT_ASPECT_RATIO_INFO_0_DE,
      NULL,
      "video",
      {
         { "corrected", OPTION_VAL_CORRECTED_DE },
         { "uncorrected", OPTION_VAL_UNCORRECTED_DE },
         { "4:3",  OPTION_VAL_4_3_DE },
         { "ntsc", OPTION_VAL_NTSC_DE },
      },
      "corrected"
   },
   {
      BEETLE_OPT(crop_overscan),
      BEETLE_OPT_CROP_OVERSCAN_LABEL_DE,
      NULL,
      BEETLE_OPT_CROP_OVERSCAN_INFO_0_DE,
      NULL,
      "video",
      {
         { "disabled", "Disabled" },
         { "static",   OPTION_VAL_STATIC_DE },
         { "smart",    OPTION_VAL_SMART_DE },
         { NULL, NULL },
      },
      "smart"
   },
   {
      BEETLE_OPT(image_crop),
      BEETLE_OPT_IMAGE_CROP_LABEL_DE,
      NULL,
      BEETLE_OPT_IMAGE_CROP_INFO_0_DE,
      NULL,
      "video",
      {
         { "disabled", "0" },
         { "1px",      OPTION_VAL_1PX_DE },
         { "2px",      OPTION_VAL_2PX_DE },
         { "3px",      OPTION_VAL_3PX_DE },
         { "4px",      OPTION_VAL_4PX_DE },
         { "5px",      OPTION_VAL_5PX_DE },
         { "6px",      OPTION_VAL_6PX_DE },
         { "7px",      OPTION_VAL_7PX_DE },
         { "8px",      OPTION_VAL_8PX_DE },
         { "9px",      OPTION_VAL_9PX_DE },
         { "10px",     OPTION_VAL_10PX_DE },
         { "11px",     OPTION_VAL_11PX_DE },
         { "12px",     OPTION_VAL_12PX_DE },
         { "13px",     OPTION_VAL_13PX_DE },
         { "14px",     OPTION_VAL_14PX_DE },
         { "15px",     OPTION_VAL_15PX_DE },
         { "16px",     OPTION_VAL_16PX_DE },
         { "17px",     OPTION_VAL_17PX_DE },
         { "18px",     OPTION_VAL_18PX_DE },
         { "19px",     OPTION_VAL_19PX_DE },
         { "20px",     OPTION_VAL_20PX_DE },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(image_offset),
      BEETLE_OPT_IMAGE_OFFSET_LABEL_DE,
      NULL,
      BEETLE_OPT_IMAGE_OFFSET_INFO_0_DE,
      NULL,
      "video",
      {
         { "-12px",    OPTION_VAL_12PX_O64_DE },
         { "-11px",    OPTION_VAL_11PX_O64_DE },
         { "-10px",    OPTION_VAL_10PX_O64_DE },
         { "-9px",     OPTION_VAL_9PX_O64_DE },
         { "-8px",     OPTION_VAL_8PX_O64_DE },
         { "-7px",     OPTION_VAL_7PX_O64_DE },
         { "-6px",     OPTION_VAL_6PX_O64_DE },
         { "-5px",     OPTION_VAL_5PX_O64_DE },
         { "-4px",     OPTION_VAL_4PX_O64_DE },
         { "-3px",     OPTION_VAL_3PX_O64_DE },
         { "-2px",     OPTION_VAL_2PX_O64_DE },
         { "-1px",     OPTION_VAL_1PX_O64_DE },
         { "disabled", "0" },
         { "+1px",     OPTION_VAL_1PX_O64_0_DE },
         { "+2px",     OPTION_VAL_2PX_O64_0_DE },
         { "+3px",     OPTION_VAL_3PX_O64_0_DE },
         { "+4px",     OPTION_VAL_4PX_O64_0_DE },
         { "+5px",     OPTION_VAL_5PX_O64_0_DE },
         { "+6px",     OPTION_VAL_6PX_O64_0_DE },
         { "+7px",     OPTION_VAL_7PX_O64_0_DE },
         { "+8px",     OPTION_VAL_8PX_O64_0_DE },
         { "+9px",     OPTION_VAL_9PX_O64_0_DE },
         { "+10px",    OPTION_VAL_10PX_O64_0_DE },
         { "+11px",    OPTION_VAL_11PX_O64_0_DE },
         { "+12px",    OPTION_VAL_12PX_O64_0_DE },
         { NULL, NULL },
      },
      "disabled"
   },
#if defined(HAVE_OPENGL) || defined(HAVE_OPENGLES) || defined(HAVE_VULKAN)
   {
      BEETLE_OPT(image_offset_cycles),
      BEETLE_OPT_IMAGE_OFFSET_CYCLES_LABEL_DE,
      NULL,
      BEETLE_OPT_IMAGE_OFFSET_CYCLES_INFO_0_DE,
      NULL,
      "video",
      {
         { "-40",      NULL },
         { "-39",      NULL },
         { "-38",      NULL },
         { "-37",      NULL },
         { "-36",      NULL },
         { "-35",      NULL },
         { "-34",      NULL },
         { "-33",      NULL },
         { "-32",      NULL },
         { "-31",      NULL },
         { "-30",      NULL },
         { "-29",      NULL },
         { "-28",      NULL },
         { "-27",      NULL },
         { "-26",      NULL },
         { "-25",      NULL },
         { "-24",      NULL },
         { "-23",      NULL },
         { "-22",      NULL },
         { "-21",      NULL },
         { "-20",      NULL },
         { "-19",      NULL },
         { "-18",      NULL },
         { "-17",      NULL },
         { "-16",      NULL },
         { "-15",      NULL },
         { "-14",      NULL },
         { "-13",      NULL },
         { "-12",      NULL },
         { "-11",      NULL },
         { "-10",      NULL },
         { "-9",       NULL },
         { "-8",       NULL },
         { "-7",       NULL },
         { "-6",       NULL },
         { "-5",       NULL },
         { "-4",       NULL },
         { "-3",       NULL },
         { "-2",       NULL },
         { "-1",       NULL },
         { "0",        "0" },
         { "+1",       NULL },
         { "+2",       NULL },
         { "+3",       NULL },
         { "+4",       NULL },
         { "+5",       NULL },
         { "+6",       NULL },
         { "+7",       NULL },
         { "+8",       NULL },
         { "+9",       NULL },
         { "+10",      NULL },
         { "+11",      NULL },
         { "+12",      NULL },
         { "+13",      NULL },
         { "+14",      NULL },
         { "+15",      NULL },
         { "+16",      NULL },
         { "+17",      NULL },
         { "+18",      NULL },
         { "+19",      NULL },
         { "+20",      NULL },
         { "+21",      NULL },
         { "+22",      NULL },
         { "+23",      NULL },
         { "+24",      NULL },
         { "+25",      NULL },
         { "+26",      NULL },
         { "+27",      NULL },
         { "+28",      NULL },
         { "+29",      NULL },
         { "+30",      NULL },
         { "+31",      NULL },
         { "+32",      NULL },
         { "+33",      NULL },
         { "+34",      NULL },
         { "+35",      NULL },
         { "+36",      NULL },
         { "+37",      NULL },
         { "+38",      NULL },
         { "+39",      NULL },
         { "+40",      NULL },
         { NULL, NULL},
      },
      "0"
   },
#endif
   {
      BEETLE_OPT(initial_scanline),
      BEETLE_OPT_INITIAL_SCANLINE_LABEL_DE,
      NULL,
      BEETLE_OPT_INITIAL_SCANLINE_INFO_0_DE,
      NULL,
      "video",
      {
         { "0",  OPTION_VAL_0_O48_DE },
         { "1",  NULL },
         { "2",  NULL },
         { "3",  NULL },
         { "4",  NULL },
         { "5",  NULL },
         { "6",  NULL },
         { "7",  NULL },
         { "8",  NULL },
         { "9",  NULL },
         { "10", NULL },
         { "11", NULL },
         { "12", NULL },
         { "13", NULL },
         { "14", NULL },
         { "15", NULL },
         { "16", NULL },
         { "17", NULL },
         { "18", NULL },
         { "19", NULL },
         { "20", NULL },
         { "21", NULL },
         { "22", NULL },
         { "23", NULL },
         { "24", NULL },
         { "25", NULL },
         { "26", NULL },
         { "27", NULL },
         { "28", NULL },
         { "29", NULL },
         { "30", NULL },
         { "31", NULL },
         { "32", NULL },
         { "33", NULL },
         { "34", NULL },
         { "35", NULL },
         { "36", NULL },
         { "37", NULL },
         { "38", NULL },
         { "39", NULL },
         { "40", NULL },
         { NULL, NULL },
      },
      "0"
   },
   {
      BEETLE_OPT(last_scanline),
      BEETLE_OPT_LAST_SCANLINE_LABEL_DE,
      NULL,
      BEETLE_OPT_LAST_SCANLINE_INFO_0_DE,
      NULL,
      "video",
      {
         { "210", NULL },
         { "211", NULL },
         { "212", NULL },
         { "213", NULL },
         { "214", NULL },
         { "215", NULL },
         { "216", NULL },
         { "217", NULL },
         { "218", NULL },
         { "219", NULL },
         { "220", NULL },
         { "221", NULL },
         { "222", NULL },
         { "223", NULL },
         { "224", NULL },
         { "225", NULL },
         { "226", NULL },
         { "227", NULL },
         { "228", NULL },
         { "229", NULL },
         { "230", NULL },
         { "231", NULL },
         { "232", NULL },
         { "233", NULL },
         { "234", NULL },
         { "235", NULL },
         { "236", NULL },
         { "237", NULL },
         { "238", NULL },
         { "239", OPTION_VAL_239_DE },
         { NULL, NULL },
      },
      "239"
   },
   {
      BEETLE_OPT(initial_scanline_pal),
      BEETLE_OPT_INITIAL_SCANLINE_PAL_LABEL_DE,
      NULL,
      BEETLE_OPT_INITIAL_SCANLINE_PAL_INFO_0_DE,
      NULL,
      "video",
      {
         { "0",  OPTION_VAL_0_O48_DE },
         { "1",  NULL },
         { "2",  NULL },
         { "3",  NULL },
         { "4",  NULL },
         { "5",  NULL },
         { "6",  NULL },
         { "7",  NULL },
         { "8",  NULL },
         { "9",  NULL },
         { "10", NULL },
         { "11", NULL },
         { "12", NULL },
         { "13", NULL },
         { "14", NULL },
         { "15", NULL },
         { "16", NULL },
         { "17", NULL },
         { "18", NULL },
         { "19", NULL },
         { "20", NULL },
         { "21", NULL },
         { "22", NULL },
         { "23", NULL },
         { "24", NULL },
         { "25", NULL },
         { "26", NULL },
         { "27", NULL },
         { "28", NULL },
         { "29", NULL },
         { "30", NULL },
         { "31", NULL },
         { "32", NULL },
         { "33", NULL },
         { "34", NULL },
         { "35", NULL },
         { "36", NULL },
         { "37", NULL },
         { "38", NULL },
         { "39", NULL },
         { "40", NULL },
         { NULL, NULL },
      },
      "0"
   },
   {
      BEETLE_OPT(last_scanline_pal),
      BEETLE_OPT_LAST_SCANLINE_PAL_LABEL_DE,
      NULL,
      BEETLE_OPT_LAST_SCANLINE_PAL_INFO_0_DE,
      NULL,
      "video",
      {
         { "230", NULL },
         { "231", NULL },
         { "232", NULL },
         { "233", NULL },
         { "234", NULL },
         { "235", NULL },
         { "236", NULL },
         { "237", NULL },
         { "238", NULL },
         { "239", NULL },
         { "240", NULL },
         { "241", NULL },
         { "242", NULL },
         { "243", NULL },
         { "244", NULL },
         { "245", NULL },
         { "246", NULL },
         { "247", NULL },
         { "248", NULL },
         { "249", NULL },
         { "250", NULL },
         { "251", NULL },
         { "252", NULL },
         { "253", NULL },
         { "254", NULL },
         { "255", NULL },
         { "256", NULL },
         { "257", NULL },
         { "258", NULL },
         { "259", NULL },
         { "260", NULL },
         { "261", NULL },
         { "262", NULL },
         { "263", NULL },
         { "264", NULL },
         { "265", NULL },
         { "266", NULL },
         { "267", NULL },
         { "268", NULL },
         { "269", NULL },
         { "270", NULL },
         { "271", NULL },
         { "272", NULL },
         { "273", NULL },
         { "274", NULL },
         { "275", NULL },
         { "276", NULL },
         { "277", NULL },
         { "278", NULL },
         { "279", NULL },
         { "280", NULL },
         { "281", NULL },
         { "282", NULL },
         { "283", NULL },
         { "284", NULL },
         { "285", NULL },
         { "286", NULL },
         { "287", OPTION_VAL_287_DE },
         { NULL, NULL },
      },
      "287"
   },
   { NULL, NULL, NULL, NULL, NULL, NULL, {{0}}, NULL },
};
struct retro_core_options_v2 options_de = {
   option_cats_de,
   option_defs_de
};

/* RETRO_LANGUAGE_EL */

#define CATEGORY_SYSTEM_LABEL_EL "Σύστημα"
#define CATEGORY_SYSTEM_INFO_0_EL NULL
#define CATEGORY_VIDEO_LABEL_EL "Οδηγός Βίντεο"
#define CATEGORY_VIDEO_INFO_0_EL NULL
#define CATEGORY_PGXP_LABEL_EL NULL
#define CATEGORY_PGXP_INFO_0_EL NULL
#define CATEGORY_OSD_LABEL_EL "Οθόνη Απεικόνισης"
#define CATEGORY_OSD_INFO_0_EL NULL
#define CATEGORY_INPUT_LABEL_EL "Οδηγός Εισαγωγής"
#define CATEGORY_INPUT_INFO_0_EL NULL
#define CATEGORY_MEMCARDS_LABEL_EL NULL
#define CATEGORY_MEMCARDS_INFO_0_EL NULL
#define CATEGORY_HACKS_LABEL_EL NULL
#define CATEGORY_HACKS_INFO_0_EL NULL
#define BEETLE_OPT_CPU_FREQ_SCALE_LABEL_EL NULL
#define BEETLE_OPT_CPU_FREQ_SCALE_INFO_0_EL NULL
#define OPTION_VAL_50_EL NULL
#define OPTION_VAL_60_EL NULL
#define OPTION_VAL_70_EL NULL
#define OPTION_VAL_80_EL "το 80%"
#define OPTION_VAL_90_EL NULL
#define OPTION_VAL_100_EL NULL
#define OPTION_VAL_110_EL NULL
#define OPTION_VAL_120_EL NULL
#define OPTION_VAL_130_EL NULL
#define OPTION_VAL_140_EL NULL
#define OPTION_VAL_150_EL NULL
#define OPTION_VAL_160_EL NULL
#define OPTION_VAL_170_EL NULL
#define OPTION_VAL_180_EL NULL
#define OPTION_VAL_190_EL NULL
#define OPTION_VAL_200_EL NULL
#define OPTION_VAL_210_EL NULL
#define OPTION_VAL_220_EL NULL
#define OPTION_VAL_230_EL NULL
#define OPTION_VAL_240_EL NULL
#define OPTION_VAL_250_EL NULL
#define OPTION_VAL_260_EL NULL
#define OPTION_VAL_270_EL NULL
#define OPTION_VAL_280_EL NULL
#define OPTION_VAL_290_EL NULL
#define OPTION_VAL_300_EL NULL
#define OPTION_VAL_310_EL NULL
#define OPTION_VAL_320_EL NULL
#define OPTION_VAL_330_EL NULL
#define OPTION_VAL_340_EL NULL
#define OPTION_VAL_350_EL NULL
#define OPTION_VAL_360_EL NULL
#define OPTION_VAL_370_EL NULL
#define OPTION_VAL_380_EL NULL
#define OPTION_VAL_390_EL NULL
#define OPTION_VAL_400_EL NULL
#define OPTION_VAL_410_EL NULL
#define OPTION_VAL_420_EL NULL
#define OPTION_VAL_430_EL NULL
#define OPTION_VAL_440_EL NULL
#define OPTION_VAL_450_EL NULL
#define OPTION_VAL_460_EL NULL
#define OPTION_VAL_470_EL NULL
#define OPTION_VAL_480_EL NULL
#define OPTION_VAL_490_EL NULL
#define OPTION_VAL_500_EL NULL
#define OPTION_VAL_510_EL NULL
#define OPTION_VAL_520_EL NULL
#define OPTION_VAL_530_EL NULL
#define OPTION_VAL_540_EL NULL
#define OPTION_VAL_550_EL NULL
#define OPTION_VAL_560_EL NULL
#define OPTION_VAL_570_EL NULL
#define OPTION_VAL_580_EL NULL
#define OPTION_VAL_590_EL NULL
#define OPTION_VAL_600_EL NULL
#define OPTION_VAL_610_EL NULL
#define OPTION_VAL_620_EL NULL
#define OPTION_VAL_630_EL NULL
#define OPTION_VAL_640_EL NULL
#define OPTION_VAL_650_EL NULL
#define OPTION_VAL_660_EL NULL
#define OPTION_VAL_670_EL NULL
#define OPTION_VAL_680_EL NULL
#define OPTION_VAL_690_EL NULL
#define OPTION_VAL_700_EL NULL
#define OPTION_VAL_710_EL NULL
#define OPTION_VAL_720_EL NULL
#define OPTION_VAL_730_EL NULL
#define OPTION_VAL_740_EL NULL
#define OPTION_VAL_750_EL NULL
#define BEETLE_OPT_GPU_OVERCLOCK_LABEL_EL NULL
#define BEETLE_OPT_GPU_OVERCLOCK_INFO_0_EL NULL
#define OPTION_VAL_1X_NATIVE_EL NULL
#define OPTION_VAL_2X_EL NULL
#define OPTION_VAL_4X_EL NULL
#define OPTION_VAL_8X_EL NULL
#define OPTION_VAL_16X_EL NULL
#define OPTION_VAL_32X_EL NULL
#define BEETLE_OPT_GTE_OVERCLOCK_LABEL_EL NULL
#define BEETLE_OPT_GTE_OVERCLOCK_INFO_0_EL NULL
#define BEETLE_OPT_CD_ACCESS_METHOD_LABEL_EL NULL
#define BEETLE_OPT_CD_ACCESS_METHOD_INFO_0_EL NULL
#define OPTION_VAL_SYNC_EL NULL
#define OPTION_VAL_ASYNC_EL NULL
#define OPTION_VAL_PRECACHE_EL NULL
#define BEETLE_OPT_CD_FASTLOAD_LABEL_EL NULL
#define BEETLE_OPT_CD_FASTLOAD_INFO_0_EL NULL
#define OPTION_VAL_2X_NATIVE_EL NULL
#define OPTION_VAL_6X_EL NULL
#define OPTION_VAL_10X_EL NULL
#define OPTION_VAL_12X_EL NULL
#define OPTION_VAL_14X_EL NULL
#define BEETLE_OPT_SKIP_BIOS_LABEL_EL NULL
#define BEETLE_OPT_SKIP_BIOS_INFO_0_EL NULL
#define BEETLE_OPT_OVERRIDE_BIOS_LABEL_EL NULL
#define BEETLE_OPT_OVERRIDE_BIOS_INFO_0_EL NULL
#define OPTION_VAL_PSXONPSP_EL NULL
#define OPTION_VAL_PS1_ROM_EL NULL
#define OPTION_VAL_OPENBIOS_EL NULL
#define BEETLE_OPT_REGION_LABEL_EL NULL
#define BEETLE_OPT_REGION_INFO_0_EL NULL
#define OPTION_VAL_AUTO_EL "Αυτόματο"
#define OPTION_VAL_NTSC_J_EL NULL
#define OPTION_VAL_NTSC_U_EL NULL
#define OPTION_VAL_PAL_EL NULL
#define BEETLE_OPT_INTERNAL_RESOLUTION_LABEL_EL NULL
#define BEETLE_OPT_INTERNAL_RESOLUTION_INFO_0_EL NULL
#define BEETLE_OPT_RENDERER_LABEL_EL NULL
#define BEETLE_OPT_RENDERER_INFO_0_EL NULL
#define OPTION_VAL_HARDWARE_EL NULL
#define OPTION_VAL_HARDWARE_GL_EL NULL
#define OPTION_VAL_HARDWARE_VK_EL NULL
#define OPTION_VAL_SOFTWARE_EL NULL
#define BEETLE_OPT_RENDERER_SOFTWARE_FB_LABEL_EL NULL
#define BEETLE_OPT_RENDERER_SOFTWARE_FB_INFO_0_EL NULL
#define BEETLE_OPT_DEPTH_LABEL_EL NULL
#define BEETLE_OPT_DEPTH_INFO_0_EL NULL
#define OPTION_VAL_16BPP_NATIVE_EL NULL
#define OPTION_VAL_32BPP_EL NULL
#define BEETLE_OPT_DITHER_MODE_LABEL_EL NULL
#define BEETLE_OPT_DITHER_MODE_INFO_0_EL NULL
#define OPTION_VAL_INTERNAL_RESOLUTION_EL NULL
#define BEETLE_OPT_SCALED_UV_OFFSET_LABEL_EL NULL
#define BEETLE_OPT_SCALED_UV_OFFSET_INFO_0_EL NULL
#define BEETLE_OPT_FILTER_LABEL_EL NULL
#define BEETLE_OPT_FILTER_INFO_0_EL NULL
#define OPTION_VAL_NEAREST_EL "Κοντινότερο"
#define OPTION_VAL_SABR_EL NULL
#define OPTION_VAL_XBR_EL NULL
#define OPTION_VAL_BILINEAR_EL NULL
#define OPTION_VAL_3_POINT_EL NULL
#define OPTION_VAL_JINC2_EL NULL
#define BEETLE_OPT_FILTER_EXCLUDE_SPRITE_LABEL_EL NULL
#define BEETLE_OPT_FILTER_EXCLUDE_SPRITE_INFO_0_EL NULL
#define OPTION_VAL_OPAQUE_EL NULL
#define OPTION_VAL_ALL_EL NULL
#define BEETLE_OPT_FILTER_EXCLUDE_2D_POLYGON_LABEL_EL NULL
#define BEETLE_OPT_FILTER_EXCLUDE_2D_POLYGON_INFO_0_EL NULL
#define BEETLE_OPT_ADAPTIVE_SMOOTHING_LABEL_EL NULL
#define BEETLE_OPT_ADAPTIVE_SMOOTHING_INFO_0_EL NULL
#define BEETLE_OPT_SUPER_SAMPLING_LABEL_EL NULL
#define BEETLE_OPT_SUPER_SAMPLING_INFO_0_EL NULL
#define BEETLE_OPT_MSAA_LABEL_EL NULL
#define BEETLE_OPT_MSAA_INFO_0_EL NULL
#define OPTION_VAL_1X_EL NULL
#define BEETLE_OPT_MDEC_YUV_LABEL_EL NULL
#define BEETLE_OPT_MDEC_YUV_INFO_0_EL NULL
#define BEETLE_OPT_TRACK_TEXTURES_LABEL_EL NULL
#define BEETLE_OPT_TRACK_TEXTURES_INFO_0_EL NULL
#define BEETLE_OPT_DUMP_TEXTURES_LABEL_EL NULL
#define BEETLE_OPT_DUMP_TEXTURES_INFO_0_EL NULL
#define BEETLE_OPT_REPLACE_TEXTURES_LABEL_EL NULL
#define BEETLE_OPT_REPLACE_TEXTURES_INFO_0_EL NULL
#define BEETLE_OPT_PGXP_MODE_LABEL_EL NULL
#define BEETLE_OPT_PGXP_MODE_INFO_0_EL NULL
#define OPTION_VAL_MEMORY_ONLY_EL NULL
#define OPTION_VAL_MEMORY_CPU_EL NULL
#define BEETLE_OPT_PGXP_2D_TOL_LABEL_EL NULL
#define BEETLE_OPT_PGXP_2D_TOL_INFO_0_EL NULL
#define OPTION_VAL_0PX_EL NULL
#define OPTION_VAL_1PX_EL NULL
#define OPTION_VAL_2PX_EL NULL
#define OPTION_VAL_3PX_EL NULL
#define OPTION_VAL_4PX_EL NULL
#define OPTION_VAL_5PX_EL NULL
#define OPTION_VAL_6PX_EL NULL
#define OPTION_VAL_7PX_EL NULL
#define OPTION_VAL_8PX_EL NULL
#define BEETLE_OPT_PGXP_NCLIP_LABEL_EL NULL
#define BEETLE_OPT_PGXP_NCLIP_INFO_0_EL NULL
#define BEETLE_OPT_PGXP_VERTEX_LABEL_EL NULL
#define BEETLE_OPT_PGXP_VERTEX_INFO_0_EL NULL
#define BEETLE_OPT_PGXP_TEXTURE_LABEL_EL NULL
#define BEETLE_OPT_PGXP_TEXTURE_INFO_0_EL NULL
#define BEETLE_OPT_DISPLAY_INTERNAL_FPS_LABEL_EL NULL
#define BEETLE_OPT_DISPLAY_INTERNAL_FPS_INFO_0_EL NULL
#define BEETLE_OPT_DISPLAY_OSD_LABEL_EL NULL
#define BEETLE_OPT_DISPLAY_OSD_INFO_0_EL NULL
#define BEETLE_OPT_DISPLAY_VRAM_LABEL_EL NULL
#define BEETLE_OPT_DISPLAY_VRAM_INFO_0_EL NULL
#define BEETLE_OPT_ANALOG_CALIBRATION_LABEL_EL NULL
#define BEETLE_OPT_ANALOG_CALIBRATION_INFO_0_EL NULL
#define BEETLE_OPT_ANALOG_TOGGLE_LABEL_EL NULL
#define BEETLE_OPT_ANALOG_TOGGLE_INFO_0_EL NULL
#define OPTION_VAL_ENABLED_ANALOG_EL NULL
#define BEETLE_OPT_ANALOG_TOGGLE_COMBO_LABEL_EL NULL
#define BEETLE_OPT_ANALOG_TOGGLE_COMBO_INFO_0_EL NULL
#define OPTION_VAL_L1_L2_R1_R2_START_SELECT_EL NULL
#define OPTION_VAL_L1_R1_SELECT_EL NULL
#define OPTION_VAL_L1_R1_START_EL NULL
#define OPTION_VAL_L1_R1_L3_EL NULL
#define OPTION_VAL_L1_R1_R3_EL NULL
#define OPTION_VAL_L2_R2_SELECT_EL NULL
#define OPTION_VAL_L2_R2_START_EL NULL
#define OPTION_VAL_L2_R2_L3_EL NULL
#define OPTION_VAL_L2_R2_R3_EL NULL
#define OPTION_VAL_L3_R3_EL NULL
#define BEETLE_OPT_ANALOG_TOGGLE_HOLD_LABEL_EL NULL
#define BEETLE_OPT_ANALOG_TOGGLE_HOLD_INFO_0_EL NULL
#define OPTION_VAL_0_EL NULL
#define OPTION_VAL_1_EL NULL
#define OPTION_VAL_2_EL NULL
#define OPTION_VAL_3_EL NULL
#define OPTION_VAL_4_EL NULL
#define OPTION_VAL_5_EL NULL
#define BEETLE_OPT_ENABLE_MULTITAP_PORT1_LABEL_EL NULL
#define BEETLE_OPT_ENABLE_MULTITAP_PORT1_INFO_0_EL NULL
#define BEETLE_OPT_ENABLE_MULTITAP_PORT2_LABEL_EL NULL
#define BEETLE_OPT_ENABLE_MULTITAP_PORT2_INFO_0_EL NULL
#define BEETLE_OPT_GUN_INPUT_MODE_LABEL_EL NULL
#define BEETLE_OPT_GUN_INPUT_MODE_INFO_0_EL NULL
#define OPTION_VAL_LIGHTGUN_EL NULL
#define OPTION_VAL_TOUCHSCREEN_EL NULL
#define BEETLE_OPT_GUN_CURSOR_LABEL_EL NULL
#define BEETLE_OPT_GUN_CURSOR_INFO_0_EL NULL
#define OPTION_VAL_CROSS_EL NULL
#define OPTION_VAL_DOT_EL NULL
#define BEETLE_OPT_CROSSHAIR_COLOR_P1_LABEL_EL NULL
#define BEETLE_OPT_CROSSHAIR_COLOR_P1_INFO_0_EL NULL
#define OPTION_VAL_RED_EL "Κόκκινο"
#define OPTION_VAL_BLUE_EL "Μπλε"
#define OPTION_VAL_GREEN_EL "Πράσινο"
#define OPTION_VAL_ORANGE_EL NULL
#define OPTION_VAL_YELLOW_EL "Κίτρινο"
#define OPTION_VAL_CYAN_EL NULL
#define OPTION_VAL_PINK_EL NULL
#define OPTION_VAL_PURPLE_EL NULL
#define OPTION_VAL_BLACK_EL NULL
#define OPTION_VAL_WHITE_EL NULL
#define BEETLE_OPT_CROSSHAIR_COLOR_P2_LABEL_EL NULL
#define BEETLE_OPT_CROSSHAIR_COLOR_P2_INFO_0_EL NULL
#define BEETLE_OPT_MOUSE_SENSITIVITY_LABEL_EL "Ευαισθησία Ποντικιού"
#define BEETLE_OPT_MOUSE_SENSITIVITY_INFO_0_EL NULL
#define OPTION_VAL_5_O42_EL NULL
#define OPTION_VAL_10_EL NULL
#define OPTION_VAL_15_EL NULL
#define OPTION_VAL_20_EL NULL
#define OPTION_VAL_25_EL NULL
#define OPTION_VAL_30_EL NULL
#define OPTION_VAL_35_EL NULL
#define OPTION_VAL_40_EL NULL
#define OPTION_VAL_45_EL NULL
#define OPTION_VAL_55_EL NULL
#define OPTION_VAL_65_EL NULL
#define OPTION_VAL_75_EL NULL
#define OPTION_VAL_85_EL NULL
#define OPTION_VAL_95_EL NULL
#define OPTION_VAL_100_O42_EL NULL
#define OPTION_VAL_105_EL NULL
#define OPTION_VAL_115_EL NULL
#define OPTION_VAL_125_EL NULL
#define OPTION_VAL_135_EL NULL
#define OPTION_VAL_145_EL NULL
#define OPTION_VAL_155_EL NULL
#define OPTION_VAL_165_EL NULL
#define OPTION_VAL_175_EL NULL
#define OPTION_VAL_185_EL NULL
#define OPTION_VAL_195_EL NULL
#define BEETLE_OPT_NEGCON_RESPONSE_LABEL_EL NULL
#define BEETLE_OPT_NEGCON_RESPONSE_INFO_0_EL NULL
#define OPTION_VAL_LINEAR_EL "Γραμμικός"
#define OPTION_VAL_QUADRATIC_EL NULL
#define OPTION_VAL_CUBIC_EL NULL
#define BEETLE_OPT_NEGCON_DEADZONE_LABEL_EL NULL
#define BEETLE_OPT_NEGCON_DEADZONE_INFO_0_EL NULL
#define OPTION_VAL_0_O44_EL NULL
#define BEETLE_OPT_USE_MEDNAFEN_MEMCARD0_METHOD_LABEL_EL NULL
#define BEETLE_OPT_USE_MEDNAFEN_MEMCARD0_METHOD_INFO_0_EL NULL
#define OPTION_VAL_LIBRETRO_EL NULL
#define OPTION_VAL_MEDNAFEN_EL NULL
#define BEETLE_OPT_ENABLE_MEMCARD1_LABEL_EL NULL
#define BEETLE_OPT_ENABLE_MEMCARD1_INFO_0_EL NULL
#define BEETLE_OPT_SHARED_MEMORY_CARDS_LABEL_EL NULL
#define BEETLE_OPT_SHARED_MEMORY_CARDS_INFO_0_EL NULL
#define BEETLE_OPT_MEMCARD_LEFT_INDEX_LABEL_EL NULL
#define BEETLE_OPT_MEMCARD_LEFT_INDEX_INFO_0_EL NULL
#define OPTION_VAL_0_O48_EL NULL
#define BEETLE_OPT_MEMCARD_RIGHT_INDEX_LABEL_EL NULL
#define BEETLE_OPT_MEMCARD_RIGHT_INDEX_INFO_0_EL NULL
#define OPTION_VAL_1_O49_EL NULL
#define BEETLE_OPT_LINE_RENDER_LABEL_EL NULL
#define BEETLE_OPT_LINE_RENDER_INFO_0_EL NULL
#define OPTION_VAL_DEFAULT_EL "Προκαθορισμένο"
#define OPTION_VAL_AGGRESSIVE_EL NULL
#define BEETLE_OPT_WIDESCREEN_HACK_LABEL_EL NULL
#define BEETLE_OPT_WIDESCREEN_HACK_INFO_0_EL NULL
#define BEETLE_OPT_WIDESCREEN_HACK_ASPECT_RATIO_LABEL_EL NULL
#define BEETLE_OPT_WIDESCREEN_HACK_ASPECT_RATIO_INFO_0_EL NULL
#define OPTION_VAL_16_9_EL NULL
#define OPTION_VAL_16_10_EL NULL
#define OPTION_VAL_18_9_EL NULL
#define OPTION_VAL_19_9_EL NULL
#define OPTION_VAL_20_9_EL NULL
#define OPTION_VAL_21_9_EL NULL
#define OPTION_VAL_32_9_EL NULL
#define BEETLE_OPT_CPU_DYNAREC_LABEL_EL NULL
#define BEETLE_OPT_CPU_DYNAREC_INFO_0_EL NULL
#define OPTION_VAL_DISABLED_EL NULL
#define OPTION_VAL_EXECUTE_EL NULL
#define OPTION_VAL_EXECUTE_ONE_EL NULL
#define OPTION_VAL_RUN_INTERPRETER_EL NULL
#define BEETLE_OPT_DYNAREC_INVALIDATE_LABEL_EL NULL
#define BEETLE_OPT_DYNAREC_INVALIDATE_INFO_0_EL NULL
#define OPTION_VAL_FULL_EL NULL
#define OPTION_VAL_DMA_EL NULL
#define BEETLE_OPT_DYNAREC_EVENTCYCLES_LABEL_EL NULL
#define BEETLE_OPT_DYNAREC_EVENTCYCLES_INFO_0_EL NULL
#define OPTION_VAL_128_EL NULL
#define BEETLE_OPT_DYNAREC_SPU_SAMPLES_LABEL_EL NULL
#define BEETLE_OPT_DYNAREC_SPU_SAMPLES_INFO_0_EL NULL
#define BEETLE_OPT_CORE_TIMING_FPS_LABEL_EL NULL
#define BEETLE_OPT_CORE_TIMING_FPS_INFO_0_EL NULL
#define OPTION_VAL_AUTO_TOGGLE_EL NULL
#define OPTION_VAL_FORCE_PROGRESSIVE_EL NULL
#define OPTION_VAL_FORCE_INTERLACED_EL NULL
#define BEETLE_OPT_DEINTERLACER_LABEL_EL NULL
#define BEETLE_OPT_DEINTERLACER_INFO_0_EL NULL
#define OPTION_VAL_WEAVE_EL NULL
#define OPTION_VAL_BOB_EL NULL
#define OPTION_VAL_BOB_OFFSET_EL NULL
#define OPTION_VAL_FASTMAD_EL NULL
#define BEETLE_OPT_FRAME_DUPING_LABEL_EL NULL
#define BEETLE_OPT_FRAME_DUPING_INFO_0_EL NULL
#define BEETLE_OPT_PAL_VIDEO_TIMING_OVERRIDE_LABEL_EL NULL
#define BEETLE_OPT_PAL_VIDEO_TIMING_OVERRIDE_INFO_0_EL NULL
#define BEETLE_OPT_ASPECT_RATIO_LABEL_EL NULL
#define BEETLE_OPT_ASPECT_RATIO_INFO_0_EL NULL
#define OPTION_VAL_CORRECTED_EL NULL
#define OPTION_VAL_UNCORRECTED_EL NULL
#define OPTION_VAL_4_3_EL NULL
#define OPTION_VAL_NTSC_EL NULL
#define BEETLE_OPT_CROP_OVERSCAN_LABEL_EL NULL
#define BEETLE_OPT_CROP_OVERSCAN_INFO_0_EL NULL
#define OPTION_VAL_STATIC_EL NULL
#define OPTION_VAL_SMART_EL NULL
#define BEETLE_OPT_IMAGE_CROP_LABEL_EL NULL
#define BEETLE_OPT_IMAGE_CROP_INFO_0_EL NULL
#define OPTION_VAL_9PX_EL NULL
#define OPTION_VAL_10PX_EL NULL
#define OPTION_VAL_11PX_EL NULL
#define OPTION_VAL_12PX_EL NULL
#define OPTION_VAL_13PX_EL NULL
#define OPTION_VAL_14PX_EL NULL
#define OPTION_VAL_15PX_EL NULL
#define OPTION_VAL_16PX_EL NULL
#define OPTION_VAL_17PX_EL NULL
#define OPTION_VAL_18PX_EL NULL
#define OPTION_VAL_19PX_EL NULL
#define OPTION_VAL_20PX_EL NULL
#define BEETLE_OPT_IMAGE_OFFSET_LABEL_EL NULL
#define BEETLE_OPT_IMAGE_OFFSET_INFO_0_EL NULL
#define OPTION_VAL_12PX_O64_EL NULL
#define OPTION_VAL_11PX_O64_EL NULL
#define OPTION_VAL_10PX_O64_EL NULL
#define OPTION_VAL_9PX_O64_EL NULL
#define OPTION_VAL_8PX_O64_EL NULL
#define OPTION_VAL_7PX_O64_EL NULL
#define OPTION_VAL_6PX_O64_EL NULL
#define OPTION_VAL_5PX_O64_EL NULL
#define OPTION_VAL_4PX_O64_EL NULL
#define OPTION_VAL_3PX_O64_EL NULL
#define OPTION_VAL_2PX_O64_EL NULL
#define OPTION_VAL_1PX_O64_EL NULL
#define OPTION_VAL_1PX_O64_0_EL NULL
#define OPTION_VAL_2PX_O64_0_EL NULL
#define OPTION_VAL_3PX_O64_0_EL NULL
#define OPTION_VAL_4PX_O64_0_EL NULL
#define OPTION_VAL_5PX_O64_0_EL NULL
#define OPTION_VAL_6PX_O64_0_EL NULL
#define OPTION_VAL_7PX_O64_0_EL NULL
#define OPTION_VAL_8PX_O64_0_EL NULL
#define OPTION_VAL_9PX_O64_0_EL NULL
#define OPTION_VAL_10PX_O64_0_EL NULL
#define OPTION_VAL_11PX_O64_0_EL NULL
#define OPTION_VAL_12PX_O64_0_EL NULL
#define BEETLE_OPT_IMAGE_OFFSET_CYCLES_LABEL_EL NULL
#define BEETLE_OPT_IMAGE_OFFSET_CYCLES_INFO_0_EL NULL
#define BEETLE_OPT_INITIAL_SCANLINE_LABEL_EL NULL
#define BEETLE_OPT_INITIAL_SCANLINE_INFO_0_EL NULL
#define BEETLE_OPT_LAST_SCANLINE_LABEL_EL NULL
#define BEETLE_OPT_LAST_SCANLINE_INFO_0_EL NULL
#define OPTION_VAL_239_EL "239 (Προεπιλογή)"
#define BEETLE_OPT_INITIAL_SCANLINE_PAL_LABEL_EL NULL
#define BEETLE_OPT_INITIAL_SCANLINE_PAL_INFO_0_EL NULL
#define BEETLE_OPT_LAST_SCANLINE_PAL_LABEL_EL NULL
#define BEETLE_OPT_LAST_SCANLINE_PAL_INFO_0_EL NULL
#define OPTION_VAL_287_EL "287 (Προεπιλογή)"

struct retro_core_option_v2_category option_cats_el[] = {
   {
      "system",
      CATEGORY_SYSTEM_LABEL_EL,
      CATEGORY_SYSTEM_INFO_0_EL
   },
   {
      "video",
      CATEGORY_VIDEO_LABEL_EL,
      CATEGORY_VIDEO_INFO_0_EL
   },
   {
      "pgxp",
      CATEGORY_PGXP_LABEL_EL,
      CATEGORY_PGXP_INFO_0_EL
   },
   {
      "osd",
      CATEGORY_OSD_LABEL_EL,
      CATEGORY_OSD_INFO_0_EL
   },
   {
      "input",
      CATEGORY_INPUT_LABEL_EL,
      CATEGORY_INPUT_INFO_0_EL
   },
   {
      "memcards",
      CATEGORY_MEMCARDS_LABEL_EL,
      CATEGORY_MEMCARDS_INFO_0_EL
   },
   {
      "hacks",
      CATEGORY_HACKS_LABEL_EL,
      CATEGORY_HACKS_INFO_0_EL
   },
   { NULL, NULL, NULL },
};
struct retro_core_option_v2_definition option_defs_el[] = {
   {
      BEETLE_OPT(cpu_freq_scale),
      BEETLE_OPT_CPU_FREQ_SCALE_LABEL_EL,
      NULL,
      BEETLE_OPT_CPU_FREQ_SCALE_INFO_0_EL,
      NULL,
      "system",
      {
         { "50%",           OPTION_VAL_50_EL },
         { "60%",           OPTION_VAL_60_EL },
         { "70%",           OPTION_VAL_70_EL },
         { "80%",           OPTION_VAL_80_EL },
         { "90%",           OPTION_VAL_90_EL },
         { "100%",          OPTION_VAL_100_EL },
         { "110%",          OPTION_VAL_110_EL },
         { "120%",          OPTION_VAL_120_EL },
         { "130%",          OPTION_VAL_130_EL },
         { "140%",          OPTION_VAL_140_EL },
         { "150%",          OPTION_VAL_150_EL },
         { "160%",          OPTION_VAL_160_EL },
         { "170%",          OPTION_VAL_170_EL },
         { "180%",          OPTION_VAL_180_EL },
         { "190%",          OPTION_VAL_190_EL },
         { "200%",          OPTION_VAL_200_EL },
         { "210%",          OPTION_VAL_210_EL },
         { "220%",          OPTION_VAL_220_EL },
         { "230%",          OPTION_VAL_230_EL },
         { "240%",          OPTION_VAL_240_EL },
         { "250%",          OPTION_VAL_250_EL },
         { "260%",          OPTION_VAL_260_EL },
         { "270%",          OPTION_VAL_270_EL },
         { "280%",          OPTION_VAL_280_EL },
         { "290%",          OPTION_VAL_290_EL },
         { "300%",          OPTION_VAL_300_EL },
         { "310%",          OPTION_VAL_310_EL },
         { "320%",          OPTION_VAL_320_EL },
         { "330%",          OPTION_VAL_330_EL },
         { "340%",          OPTION_VAL_340_EL },
         { "350%",          OPTION_VAL_350_EL },
         { "360%",          OPTION_VAL_360_EL },
         { "370%",          OPTION_VAL_370_EL },
         { "380%",          OPTION_VAL_380_EL },
         { "390%",          OPTION_VAL_390_EL },
         { "400%",          OPTION_VAL_400_EL },
         { "410%",          OPTION_VAL_410_EL },
         { "420%",          OPTION_VAL_420_EL },
         { "430%",          OPTION_VAL_430_EL },
         { "440%",          OPTION_VAL_440_EL },
         { "450%",          OPTION_VAL_450_EL },
         { "460%",          OPTION_VAL_460_EL },
         { "470%",          OPTION_VAL_470_EL },
         { "480%",          OPTION_VAL_480_EL },
         { "490%",          OPTION_VAL_490_EL },
         { "500%",          OPTION_VAL_500_EL },
         { "510%",          OPTION_VAL_510_EL },
         { "520%",          OPTION_VAL_520_EL },
         { "530%",          OPTION_VAL_530_EL },
         { "540%",          OPTION_VAL_540_EL },
         { "550%",          OPTION_VAL_550_EL },
         { "560%",          OPTION_VAL_560_EL },
         { "570%",          OPTION_VAL_570_EL },
         { "580%",          OPTION_VAL_580_EL },
         { "590%",          OPTION_VAL_590_EL },
         { "600%",          OPTION_VAL_600_EL },
         { "610%",          OPTION_VAL_610_EL },
         { "620%",          OPTION_VAL_620_EL },
         { "630%",          OPTION_VAL_630_EL },
         { "640%",          OPTION_VAL_640_EL },
         { "650%",          OPTION_VAL_650_EL },
         { "660%",          OPTION_VAL_660_EL },
         { "670%",          OPTION_VAL_670_EL },
         { "680%",          OPTION_VAL_680_EL },
         { "690%",          OPTION_VAL_690_EL },
         { "700%",          OPTION_VAL_700_EL },
         { "710%",          OPTION_VAL_710_EL },
         { "720%",          OPTION_VAL_720_EL },
         { "730%",          OPTION_VAL_730_EL },
         { "740%",          OPTION_VAL_740_EL },
         { "750%",          OPTION_VAL_750_EL },
         { NULL, NULL },
      },
      "100%"
   },
   {
      BEETLE_OPT(gpu_overclock),
      BEETLE_OPT_GPU_OVERCLOCK_LABEL_EL,
      NULL,
      BEETLE_OPT_GPU_OVERCLOCK_INFO_0_EL,
      NULL,
      "system",
      {
         { "1x(native)", OPTION_VAL_1X_NATIVE_EL },
         { "2x",         OPTION_VAL_2X_EL },
         { "4x",         OPTION_VAL_4X_EL },
         { "8x",         OPTION_VAL_8X_EL },
         { "16x",        OPTION_VAL_16X_EL },
         { "32x",        OPTION_VAL_32X_EL },
         { NULL, NULL },
      },
      "1x(native)"
   },
   {
      BEETLE_OPT(gte_overclock),
      BEETLE_OPT_GTE_OVERCLOCK_LABEL_EL,
      NULL,
      BEETLE_OPT_GTE_OVERCLOCK_INFO_0_EL,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
#ifndef EMSCRIPTEN
   {
      BEETLE_OPT(cd_access_method),
      BEETLE_OPT_CD_ACCESS_METHOD_LABEL_EL,
      NULL,
      BEETLE_OPT_CD_ACCESS_METHOD_INFO_0_EL,
      NULL,
      "system",
      {
         { "sync",     OPTION_VAL_SYNC_EL },
         { "async",    OPTION_VAL_ASYNC_EL },
         { "precache", OPTION_VAL_PRECACHE_EL },
         { NULL, NULL },
      },
      "sync"
   },
#endif
   {
      BEETLE_OPT(cd_fastload),
      BEETLE_OPT_CD_FASTLOAD_LABEL_EL,
      NULL,
      BEETLE_OPT_CD_FASTLOAD_INFO_0_EL,
      NULL,
      "system",
      {
         { "2x(native)", OPTION_VAL_2X_NATIVE_EL },
         { "4x",          OPTION_VAL_4X_EL },
         { "6x",          OPTION_VAL_6X_EL },
         { "8x",          OPTION_VAL_8X_EL },
         { "10x",         OPTION_VAL_10X_EL },
         { "12x",         OPTION_VAL_12X_EL },
         { "14x",         OPTION_VAL_14X_EL },
         { NULL, NULL },
      },
      "2x(native)"
   },
   {
      BEETLE_OPT(skip_bios),
      BEETLE_OPT_SKIP_BIOS_LABEL_EL,
      NULL,
      BEETLE_OPT_SKIP_BIOS_INFO_0_EL,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(override_bios),
      BEETLE_OPT_OVERRIDE_BIOS_LABEL_EL,
      NULL,
      BEETLE_OPT_OVERRIDE_BIOS_INFO_0_EL,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "psxonpsp", OPTION_VAL_PSXONPSP_EL },
         { "ps1_rom",  OPTION_VAL_PS1_ROM_EL },
         { "openbios", OPTION_VAL_OPENBIOS_EL     },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(region),
      BEETLE_OPT_REGION_LABEL_EL,
      NULL,
      BEETLE_OPT_REGION_INFO_0_EL,
      NULL,
      "system",
      {
         { "auto",   OPTION_VAL_AUTO_EL },
         { "ntsc-j", OPTION_VAL_NTSC_J_EL },
         { "ntsc-u", OPTION_VAL_NTSC_U_EL },
         { "pal",    OPTION_VAL_PAL_EL },
         { NULL, NULL },
      },
      "auto"
   },
   {
      BEETLE_OPT(internal_resolution),
      BEETLE_OPT_INTERNAL_RESOLUTION_LABEL_EL,
      NULL,
      BEETLE_OPT_INTERNAL_RESOLUTION_INFO_0_EL,
      NULL,
      "video",
      {
         { "1x(native)", OPTION_VAL_1X_NATIVE_EL },
         { "2x",         OPTION_VAL_2X_EL },
         { "4x",         OPTION_VAL_4X_EL },
         { "8x",         OPTION_VAL_8X_EL },
         { "16x",        OPTION_VAL_16X_EL },
         { NULL, NULL },
      },
      "1x(native)"
   },
#if defined(HAVE_OPENGL) || defined(HAVE_OPENGLES) || defined(HAVE_VULKAN)
   {
      BEETLE_OPT(renderer),
      BEETLE_OPT_RENDERER_LABEL_EL,
      NULL,
      BEETLE_OPT_RENDERER_INFO_0_EL,
      NULL,
      "video",
      {
         { "hardware",    OPTION_VAL_HARDWARE_EL },
#if defined(HAVE_OPENGL) || defined(HAVE_OPENGLES)
         { "hardware_gl", OPTION_VAL_HARDWARE_GL_EL },
#endif
#if defined(HAVE_VULKAN)
         { "hardware_vk", OPTION_VAL_HARDWARE_VK_EL },
#endif
         { "software",    OPTION_VAL_SOFTWARE_EL },
         { NULL, NULL },
      },
      "hardware"
   },
   {
      BEETLE_OPT(renderer_software_fb),
      BEETLE_OPT_RENDERER_SOFTWARE_FB_LABEL_EL,
      NULL,
      BEETLE_OPT_RENDERER_SOFTWARE_FB_INFO_0_EL,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled"
   },
#endif
#if defined(HAVE_OPENGL) || defined(HAVE_OPENGLES)
   {
      BEETLE_OPT(depth),
      BEETLE_OPT_DEPTH_LABEL_EL,
      NULL,
      BEETLE_OPT_DEPTH_INFO_0_EL,
      NULL,
      "video",
      {
         { "16bpp(native)", OPTION_VAL_16BPP_NATIVE_EL },
         { "32bpp",         OPTION_VAL_32BPP_EL },
         { NULL, NULL },
      },
      "16bpp(native)"
   },
   // Sort of, it's more like 15-bit high color and 24-bit true color for visible output. The alpha channel is used for mask bit. Vulkan renderer uses ABGR1555_555 for 31 bits internal? FMVs are always 24-bit on all renderers like original hardware (BGR888, no alpha)
#endif
   {
      BEETLE_OPT(dither_mode),
      BEETLE_OPT_DITHER_MODE_LABEL_EL,
      NULL,
      BEETLE_OPT_DITHER_MODE_INFO_0_EL,
      NULL,
      "video",
      {
         { "1x(native)",          OPTION_VAL_1X_NATIVE_EL },
         { "internal resolution", OPTION_VAL_INTERNAL_RESOLUTION_EL },
         { "disabled",            NULL },
         { NULL, NULL },
      },
      "1x(native)"
   },
#ifdef HAVE_VULKAN
   {
      BEETLE_OPT(scaled_uv_offset),
      BEETLE_OPT_SCALED_UV_OFFSET_LABEL_EL,
      NULL,
      BEETLE_OPT_SCALED_UV_OFFSET_INFO_0_EL,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled"
   },
#endif
#if defined(HAVE_OPENGL) || defined(HAVE_OPENGLES) || defined(HAVE_VULKAN)
   {
      BEETLE_OPT(filter),
      BEETLE_OPT_FILTER_LABEL_EL,
      NULL,
      BEETLE_OPT_FILTER_INFO_0_EL,
      NULL,
      "video",
      {
         { "nearest",  OPTION_VAL_NEAREST_EL },
         { "SABR",     OPTION_VAL_SABR_EL },
         { "xBR",      OPTION_VAL_XBR_EL },
         { "bilinear", OPTION_VAL_BILINEAR_EL },
         { "3-point",  OPTION_VAL_3_POINT_EL },
         { "JINC2",    OPTION_VAL_JINC2_EL },
         { NULL, NULL },
      },
      "nearest"
   },
#ifdef HAVE_VULKAN
   {
      BEETLE_OPT(filter_exclude_sprite),
      BEETLE_OPT_FILTER_EXCLUDE_SPRITE_LABEL_EL,
      NULL,
      BEETLE_OPT_FILTER_EXCLUDE_SPRITE_INFO_0_EL,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "opaque", OPTION_VAL_OPAQUE_EL },
         { "all", OPTION_VAL_ALL_EL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(filter_exclude_2d_polygon),
      BEETLE_OPT_FILTER_EXCLUDE_2D_POLYGON_LABEL_EL,
      NULL,
      BEETLE_OPT_FILTER_EXCLUDE_2D_POLYGON_INFO_0_EL,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "opaque", OPTION_VAL_OPAQUE_EL },
         { "all", OPTION_VAL_ALL_EL },
         { NULL, NULL },
      },
      "disabled"
   },
#endif
#endif
#ifdef HAVE_VULKAN
   {
      BEETLE_OPT(adaptive_smoothing),
      BEETLE_OPT_ADAPTIVE_SMOOTHING_LABEL_EL,
      NULL,
      BEETLE_OPT_ADAPTIVE_SMOOTHING_INFO_0_EL,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(super_sampling),
      BEETLE_OPT_SUPER_SAMPLING_LABEL_EL,
      NULL,
      BEETLE_OPT_SUPER_SAMPLING_INFO_0_EL,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(msaa),
      BEETLE_OPT_MSAA_LABEL_EL,
      NULL,
      BEETLE_OPT_MSAA_INFO_0_EL,
      NULL,
      "video",
      {
         { "1x",  OPTION_VAL_1X_EL },
         { "2x",  OPTION_VAL_2X_EL },
         { "4x",  OPTION_VAL_4X_EL },
         { "8x",  OPTION_VAL_8X_EL },
         { "16x", OPTION_VAL_16X_EL },
         { NULL, NULL },
      },
      "1x"
   },
   {
      BEETLE_OPT(mdec_yuv),
      BEETLE_OPT_MDEC_YUV_LABEL_EL,
      NULL,
      BEETLE_OPT_MDEC_YUV_INFO_0_EL,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(track_textures),
      BEETLE_OPT_TRACK_TEXTURES_LABEL_EL,
      NULL,
      BEETLE_OPT_TRACK_TEXTURES_INFO_0_EL,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
#ifdef TEXTURE_DUMPING_ENABLED
   {
      BEETLE_OPT(dump_textures),
      BEETLE_OPT_DUMP_TEXTURES_LABEL_EL,
      NULL,
      BEETLE_OPT_DUMP_TEXTURES_INFO_0_EL,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
#endif
   {
      BEETLE_OPT(replace_textures),
      BEETLE_OPT_REPLACE_TEXTURES_LABEL_EL,
      NULL,
      BEETLE_OPT_REPLACE_TEXTURES_INFO_0_EL,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
#endif
   {
      BEETLE_OPT(pgxp_mode),
      BEETLE_OPT_PGXP_MODE_LABEL_EL,
      NULL,
      BEETLE_OPT_PGXP_MODE_INFO_0_EL,
      NULL,
      "pgxp",
      {
         { "disabled",     NULL },
         { "memory only",  OPTION_VAL_MEMORY_ONLY_EL },
         { "memory + CPU", OPTION_VAL_MEMORY_CPU_EL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(pgxp_2d_tol),
      BEETLE_OPT_PGXP_2D_TOL_LABEL_EL,
      NULL,
      BEETLE_OPT_PGXP_2D_TOL_INFO_0_EL,
      NULL,
      "pgxp",
      {
         { "disabled", NULL },
         { "0px", OPTION_VAL_0PX_EL },
         { "1px", OPTION_VAL_1PX_EL },
         { "2px", OPTION_VAL_2PX_EL },
         { "3px", OPTION_VAL_3PX_EL },
         { "4px", OPTION_VAL_4PX_EL },
         { "5px", OPTION_VAL_5PX_EL },
         { "6px", OPTION_VAL_6PX_EL },
         { "7px", OPTION_VAL_7PX_EL },
         { "8px", OPTION_VAL_8PX_EL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(pgxp_nclip),
      BEETLE_OPT_PGXP_NCLIP_LABEL_EL,
      NULL,
      BEETLE_OPT_PGXP_NCLIP_INFO_0_EL,
      NULL,
      "pgxp",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
#if defined(HAVE_OPENGL) || defined(HAVE_OPENGLES) || defined(HAVE_VULKAN)
   {
      BEETLE_OPT(pgxp_vertex),
      BEETLE_OPT_PGXP_VERTEX_LABEL_EL,
      NULL,
      BEETLE_OPT_PGXP_VERTEX_INFO_0_EL,
      NULL,
      "pgxp",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
#endif
   {
      BEETLE_OPT(pgxp_texture),
      BEETLE_OPT_PGXP_TEXTURE_LABEL_EL,
      NULL,
      BEETLE_OPT_PGXP_TEXTURE_INFO_0_EL,
      NULL,
      "pgxp",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(display_internal_fps),
      BEETLE_OPT_DISPLAY_INTERNAL_FPS_LABEL_EL,
      NULL,
      BEETLE_OPT_DISPLAY_INTERNAL_FPS_INFO_0_EL,
      NULL,
      "osd",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(display_osd),
      BEETLE_OPT_DISPLAY_OSD_LABEL_EL,
      NULL,
      BEETLE_OPT_DISPLAY_OSD_INFO_0_EL,
      NULL,
      "osd",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled"
   },
#if defined(HAVE_OPENGL) || defined(HAVE_OPENGLES) || defined(HAVE_VULKAN)
   {
      BEETLE_OPT(display_vram),
      BEETLE_OPT_DISPLAY_VRAM_LABEL_EL,
      NULL,
      BEETLE_OPT_DISPLAY_VRAM_INFO_0_EL,
      NULL,
      "osd",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
#endif
   {
      BEETLE_OPT(analog_calibration),
      BEETLE_OPT_ANALOG_CALIBRATION_LABEL_EL,
      NULL,
      BEETLE_OPT_ANALOG_CALIBRATION_INFO_0_EL,
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(analog_toggle),
      BEETLE_OPT_ANALOG_TOGGLE_LABEL_EL,
      NULL,
      BEETLE_OPT_ANALOG_TOGGLE_INFO_0_EL,
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { "enabled-analog", OPTION_VAL_ENABLED_ANALOG_EL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(analog_toggle_combo),
      BEETLE_OPT_ANALOG_TOGGLE_COMBO_LABEL_EL,
      NULL,
      BEETLE_OPT_ANALOG_TOGGLE_COMBO_INFO_0_EL,
      NULL,
      "input",
      {
         { "l1+l2+r1+r2+start+select", OPTION_VAL_L1_L2_R1_R2_START_SELECT_EL },
         { "l1+r1+select",             OPTION_VAL_L1_R1_SELECT_EL },
         { "l1+r1+start",              OPTION_VAL_L1_R1_START_EL },
         { "l1+r1+l3",                 OPTION_VAL_L1_R1_L3_EL },
         { "l1+r1+r3",                 OPTION_VAL_L1_R1_R3_EL },
         { "l2+r2+select",             OPTION_VAL_L2_R2_SELECT_EL },
         { "l2+r2+start",              OPTION_VAL_L2_R2_START_EL },
         { "l2+r2+l3",                 OPTION_VAL_L2_R2_L3_EL },
         { "l2+r2+r3",                 OPTION_VAL_L2_R2_R3_EL },
         { "l3+r3",                    OPTION_VAL_L3_R3_EL },
         { NULL, NULL },
      },
      "l1+r1+select"
   },
   {
      BEETLE_OPT(analog_toggle_hold),
      BEETLE_OPT_ANALOG_TOGGLE_HOLD_LABEL_EL,
      NULL,
      BEETLE_OPT_ANALOG_TOGGLE_HOLD_INFO_0_EL,
      NULL,
      "input",
      {
         { "0", OPTION_VAL_0_EL },
         { "1", OPTION_VAL_1_EL },
         { "2", OPTION_VAL_2_EL },
         { "3", OPTION_VAL_3_EL },
         { "4", OPTION_VAL_4_EL },
         { "5", OPTION_VAL_5_EL },
         { NULL, NULL },
      },
      "1"
   },
   {
      BEETLE_OPT(enable_multitap_port1),
      BEETLE_OPT_ENABLE_MULTITAP_PORT1_LABEL_EL,
      NULL,
      BEETLE_OPT_ENABLE_MULTITAP_PORT1_INFO_0_EL,
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(enable_multitap_port2),
      BEETLE_OPT_ENABLE_MULTITAP_PORT2_LABEL_EL,
      NULL,
      BEETLE_OPT_ENABLE_MULTITAP_PORT2_INFO_0_EL,
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(gun_input_mode),
      BEETLE_OPT_GUN_INPUT_MODE_LABEL_EL,
      NULL,
      BEETLE_OPT_GUN_INPUT_MODE_INFO_0_EL,
      NULL,
      "input",
      {
         { "lightgun",    OPTION_VAL_LIGHTGUN_EL },
         { "touchscreen", OPTION_VAL_TOUCHSCREEN_EL },
         { NULL, NULL },
      },
      "lightgun"
   },
   // Shouldn't the gun_input_mode just be Mouse vs. Touchscreen?
   {
      BEETLE_OPT(gun_cursor),
      BEETLE_OPT_GUN_CURSOR_LABEL_EL,
      NULL,
      BEETLE_OPT_GUN_CURSOR_INFO_0_EL,
      NULL,
      "input",
      {
         { "off",   "disabled" },
         { "cross", OPTION_VAL_CROSS_EL },
         { "dot",   OPTION_VAL_DOT_EL },
         { NULL, NULL },
      },
      "cross"
   },
   {
      BEETLE_OPT(crosshair_color_p1),
      BEETLE_OPT_CROSSHAIR_COLOR_P1_LABEL_EL,
      NULL,
      BEETLE_OPT_CROSSHAIR_COLOR_P1_INFO_0_EL,
      NULL,
      "input",
      {
         { "red",    OPTION_VAL_RED_EL },
         { "blue",   OPTION_VAL_BLUE_EL },
         { "green",  OPTION_VAL_GREEN_EL },
         { "orange", OPTION_VAL_ORANGE_EL },
         { "yellow", OPTION_VAL_YELLOW_EL },
         { "cyan",   OPTION_VAL_CYAN_EL },
         { "pink",   OPTION_VAL_PINK_EL },
         { "purple", OPTION_VAL_PURPLE_EL },
         { "black",  OPTION_VAL_BLACK_EL },
         { "white",  OPTION_VAL_WHITE_EL },
         { NULL, NULL },
      },
      "red"
   },
   {
      BEETLE_OPT(crosshair_color_p2),
      BEETLE_OPT_CROSSHAIR_COLOR_P2_LABEL_EL,
      NULL,
      BEETLE_OPT_CROSSHAIR_COLOR_P2_INFO_0_EL,
      NULL,
      "input",
      {
         { "blue",   OPTION_VAL_BLUE_EL },
         { "red",    OPTION_VAL_RED_EL },
         { "green",  OPTION_VAL_GREEN_EL },
         { "orange", OPTION_VAL_ORANGE_EL },
         { "yellow", OPTION_VAL_YELLOW_EL },
         { "cyan",   OPTION_VAL_CYAN_EL },
         { "pink",   OPTION_VAL_PINK_EL },
         { "purple", OPTION_VAL_PURPLE_EL },
         { "black",  OPTION_VAL_BLACK_EL },
         { "white",  OPTION_VAL_WHITE_EL },
         { NULL, NULL },
      },
      "blue"
   },
   {
      BEETLE_OPT(mouse_sensitivity),
      BEETLE_OPT_MOUSE_SENSITIVITY_LABEL_EL,
      NULL,
      BEETLE_OPT_MOUSE_SENSITIVITY_INFO_0_EL,
      NULL,
      "input",
      {
         { "5%",   OPTION_VAL_5_O42_EL },
         { "10%",  OPTION_VAL_10_EL },
         { "15%",  OPTION_VAL_15_EL },
         { "20%",  OPTION_VAL_20_EL },
         { "25%",  OPTION_VAL_25_EL },
         { "30%",  OPTION_VAL_30_EL },
         { "35%",  OPTION_VAL_35_EL },
         { "40%",  OPTION_VAL_40_EL },
         { "45%",  OPTION_VAL_45_EL },
         { "50%",  OPTION_VAL_50_EL },
         { "55%",  OPTION_VAL_55_EL },
         { "60%",  OPTION_VAL_60_EL },
         { "65%",  OPTION_VAL_65_EL },
         { "70%",  OPTION_VAL_70_EL },
         { "75%",  OPTION_VAL_75_EL },
         { "80%",  OPTION_VAL_80_EL },
         { "85%",  OPTION_VAL_85_EL },
         { "90%",  OPTION_VAL_90_EL },
         { "95%",  OPTION_VAL_95_EL },
         { "100%", OPTION_VAL_100_O42_EL },
         { "105%", OPTION_VAL_105_EL },
         { "110%", OPTION_VAL_110_EL },
         { "115%", OPTION_VAL_115_EL },
         { "120%", OPTION_VAL_120_EL },
         { "125%", OPTION_VAL_125_EL },
         { "130%", OPTION_VAL_130_EL },
         { "135%", OPTION_VAL_135_EL },
         { "140%", OPTION_VAL_140_EL },
         { "145%", OPTION_VAL_145_EL },
         { "150%", OPTION_VAL_150_EL },
         { "155%", OPTION_VAL_155_EL },
         { "160%", OPTION_VAL_160_EL },
         { "165%", OPTION_VAL_165_EL },
         { "170%", OPTION_VAL_170_EL },
         { "175%", OPTION_VAL_175_EL },
         { "180%", OPTION_VAL_180_EL },
         { "185%", OPTION_VAL_185_EL },
         { "190%", OPTION_VAL_190_EL },
         { "195%", OPTION_VAL_195_EL },
         { "200%", OPTION_VAL_200_EL },
         { NULL, NULL },
      },
      "100%"
   },
   {
      BEETLE_OPT(negcon_response),
      BEETLE_OPT_NEGCON_RESPONSE_LABEL_EL,
      NULL,
      BEETLE_OPT_NEGCON_RESPONSE_INFO_0_EL,
      NULL,
      "input",
      {
         { "linear",    OPTION_VAL_LINEAR_EL },
         { "quadratic", OPTION_VAL_QUADRATIC_EL },
         { "cubic",     OPTION_VAL_CUBIC_EL },
         { NULL, NULL },
      },
      "linear"
   },
   {
      BEETLE_OPT(negcon_deadzone),
      BEETLE_OPT_NEGCON_DEADZONE_LABEL_EL,
      NULL,
      BEETLE_OPT_NEGCON_DEADZONE_INFO_0_EL,
      NULL,
      "input",
      {
         { "0%",  OPTION_VAL_0_O44_EL },
         { "5%",  OPTION_VAL_5_O42_EL },
         { "10%", OPTION_VAL_10_EL },
         { "15%", OPTION_VAL_15_EL },
         { "20%", OPTION_VAL_20_EL },
         { "25%", OPTION_VAL_25_EL },
         { "30%", OPTION_VAL_30_EL },
         { NULL, NULL },
      },
      "0%"
   },
   {
      BEETLE_OPT(use_mednafen_memcard0_method),
      BEETLE_OPT_USE_MEDNAFEN_MEMCARD0_METHOD_LABEL_EL,
      NULL,
      BEETLE_OPT_USE_MEDNAFEN_MEMCARD0_METHOD_INFO_0_EL,
      NULL,
      "memcards",
      {
         { "libretro", OPTION_VAL_LIBRETRO_EL },
         { "mednafen", OPTION_VAL_MEDNAFEN_EL },
         { NULL, NULL },
      },
      "libretro"
   },
   {
      BEETLE_OPT(enable_memcard1),
      BEETLE_OPT_ENABLE_MEMCARD1_LABEL_EL,
      NULL,
      BEETLE_OPT_ENABLE_MEMCARD1_INFO_0_EL,
      NULL,
      "memcards",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(shared_memory_cards),
      BEETLE_OPT_SHARED_MEMORY_CARDS_LABEL_EL,
      NULL,
      BEETLE_OPT_SHARED_MEMORY_CARDS_INFO_0_EL,
      NULL,
      "memcards",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(memcard_left_index),
      BEETLE_OPT_MEMCARD_LEFT_INDEX_LABEL_EL,
      NULL,
      BEETLE_OPT_MEMCARD_LEFT_INDEX_INFO_0_EL,
      NULL,
      "memcards",
      {
         { "0",  OPTION_VAL_0_O48_EL },
         { "1",  NULL },
         { "2",  NULL },
         { "3",  NULL },
         { "4",  NULL },
         { "5",  NULL },
         { "6",  NULL },
         { "7",  NULL },
         { "8",  NULL },
         { "9",  NULL },
         { "10", NULL },
         { "11", NULL },
         { "12", NULL },
         { "13", NULL },
         { "14", NULL },
         { "15", NULL },
         { "16", NULL },
         { "17", NULL },
         { "18", NULL },
         { "19", NULL },
         { "20", NULL },
         { "21", NULL },
         { "22", NULL },
         { "23", NULL },
         { "24", NULL },
         { "25", NULL },
         { "26", NULL },
         { "27", NULL },
         { "28", NULL },
         { "29", NULL },
         { "30", NULL },
         { "31", NULL },
         { "32", NULL },
         { "33", NULL },
         { "34", NULL },
         { "35", NULL },
         { "36", NULL },
         { "37", NULL },
         { "38", NULL },
         { "39", NULL },
         { "40", NULL },
         { "41", NULL },
         { "42", NULL },
         { "43", NULL },
         { "44", NULL },
         { "45", NULL },
         { "46", NULL },
         { "47", NULL },
         { "48", NULL },
         { "49", NULL },
         { "50", NULL },
         { "51", NULL },
         { "52", NULL },
         { "53", NULL },
         { "54", NULL },
         { "55", NULL },
         { "56", NULL },
         { "57", NULL },
         { "58", NULL },
         { "59", NULL },
         { "60", NULL },
         { "61", NULL },
         { "62", NULL },
         { "63", NULL },
         { NULL, NULL },
      },
      "0"
   },
   {
      BEETLE_OPT(memcard_right_index),
      BEETLE_OPT_MEMCARD_RIGHT_INDEX_LABEL_EL,
      NULL,
      BEETLE_OPT_MEMCARD_RIGHT_INDEX_INFO_0_EL,
      NULL,
      "memcards",
      {
         { "0",  NULL },
         { "1",  OPTION_VAL_1_O49_EL },
         { "2",  NULL },
         { "3",  NULL },
         { "4",  NULL },
         { "5",  NULL },
         { "6",  NULL },
         { "7",  NULL },
         { "8",  NULL },
         { "9",  NULL },
         { "10", NULL },
         { "11", NULL },
         { "12", NULL },
         { "13", NULL },
         { "14", NULL },
         { "15", NULL },
         { "16", NULL },
         { "17", NULL },
         { "18", NULL },
         { "19", NULL },
         { "20", NULL },
         { "21", NULL },
         { "22", NULL },
         { "23", NULL },
         { "24", NULL },
         { "25", NULL },
         { "26", NULL },
         { "27", NULL },
         { "28", NULL },
         { "29", NULL },
         { "30", NULL },
         { "31", NULL },
         { "32", NULL },
         { "33", NULL },
         { "34", NULL },
         { "35", NULL },
         { "36", NULL },
         { "37", NULL },
         { "38", NULL },
         { "39", NULL },
         { "40", NULL },
         { "41", NULL },
         { "42", NULL },
         { "43", NULL },
         { "44", NULL },
         { "45", NULL },
         { "46", NULL },
         { "47", NULL },
         { "48", NULL },
         { "49", NULL },
         { "50", NULL },
         { "51", NULL },
         { "52", NULL },
         { "53", NULL },
         { "54", NULL },
         { "55", NULL },
         { "56", NULL },
         { "57", NULL },
         { "58", NULL },
         { "59", NULL },
         { "60", NULL },
         { "61", NULL },
         { "62", NULL },
         { "63", NULL },
         { NULL, NULL },
      },
      "1"
   },
   {
      BEETLE_OPT(line_render),
      BEETLE_OPT_LINE_RENDER_LABEL_EL,
      NULL,
      BEETLE_OPT_LINE_RENDER_INFO_0_EL,
      NULL,
      "hacks",
      {
         { "default",    OPTION_VAL_DEFAULT_EL },
         { "aggressive", OPTION_VAL_AGGRESSIVE_EL },
         { "disabled",   NULL },
         { NULL, NULL },
      },
      "default"
   },
   {
      BEETLE_OPT(widescreen_hack),
      BEETLE_OPT_WIDESCREEN_HACK_LABEL_EL,
      NULL,
      BEETLE_OPT_WIDESCREEN_HACK_INFO_0_EL,
      NULL,
      "hacks",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(widescreen_hack_aspect_ratio),
      BEETLE_OPT_WIDESCREEN_HACK_ASPECT_RATIO_LABEL_EL,
      NULL,
      BEETLE_OPT_WIDESCREEN_HACK_ASPECT_RATIO_INFO_0_EL,
      NULL,
      "hacks",
      {
         { "16:9",  OPTION_VAL_16_9_EL },
         { "16:10", OPTION_VAL_16_10_EL },
         { "18:9",  OPTION_VAL_18_9_EL },
         { "19:9",  OPTION_VAL_19_9_EL },
         { "20:9",  OPTION_VAL_20_9_EL },
         { "21:9",  OPTION_VAL_21_9_EL }, // 64:27
         { "32:9",  OPTION_VAL_32_9_EL },
         { NULL,    NULL },
      },
      "16:9"
   },
#if defined(HAVE_LIGHTREC)
   {
      BEETLE_OPT(cpu_dynarec),
      BEETLE_OPT_CPU_DYNAREC_LABEL_EL,
      NULL,
      BEETLE_OPT_CPU_DYNAREC_INFO_0_EL,
      NULL,
      "hacks",
      {
         { "disabled",        OPTION_VAL_DISABLED_EL },
         { "execute",         OPTION_VAL_EXECUTE_EL },
         { "execute_one",     OPTION_VAL_EXECUTE_ONE_EL },
         { "run_interpreter", OPTION_VAL_RUN_INTERPRETER_EL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(dynarec_invalidate),
      BEETLE_OPT_DYNAREC_INVALIDATE_LABEL_EL,
      NULL,
      BEETLE_OPT_DYNAREC_INVALIDATE_INFO_0_EL,
      NULL,
      "hacks",
      {
         { "full", OPTION_VAL_FULL_EL },
         { "dma",  OPTION_VAL_DMA_EL },
         { NULL, NULL },
      },
      "full"
   },
   {
      BEETLE_OPT(dynarec_eventcycles),
      BEETLE_OPT_DYNAREC_EVENTCYCLES_LABEL_EL,
      NULL,
      BEETLE_OPT_DYNAREC_EVENTCYCLES_INFO_0_EL,
      NULL,
      "hacks",
      {
         { "128",  OPTION_VAL_128_EL },
         { "256",  NULL },
         { "384",  NULL },
         { "512",  NULL },
         { "640",  NULL },
         { "768",  NULL },
         { "896",  NULL },
         { "1024", NULL },
         { "1152", NULL },
         { "1280", NULL },
         { "1408", NULL },
         { "1536", NULL },
         { "1664", NULL },
         { "1792", NULL },
         { "1920", NULL },
         { "2048", NULL },
         { NULL, NULL },
      },
      "128"
   },
   {
      BEETLE_OPT(dynarec_spu_samples),
      BEETLE_OPT_DYNAREC_SPU_SAMPLES_LABEL_EL,
      NULL,
      BEETLE_OPT_DYNAREC_SPU_SAMPLES_INFO_0_EL,
      NULL,
      "hacks",
      {
         { "1",  OPTION_VAL_1_O49_EL },
         { "4",  NULL },
         { "16", NULL },
         { NULL, NULL },
      },
      "1"
   },
#endif
   {
      BEETLE_OPT(core_timing_fps),
      BEETLE_OPT_CORE_TIMING_FPS_LABEL_EL,
      NULL,
      BEETLE_OPT_CORE_TIMING_FPS_INFO_0_EL,
      NULL,
      "video",
      {
         { "auto_toggle",       OPTION_VAL_AUTO_TOGGLE_EL },
         { "force_progressive", OPTION_VAL_FORCE_PROGRESSIVE_EL },
         { "force_interlaced",  OPTION_VAL_FORCE_INTERLACED_EL },
      },
      "auto_toggle"
   },
   {
      BEETLE_OPT(deinterlacer),
      BEETLE_OPT_DEINTERLACER_LABEL_EL,
      NULL,
      BEETLE_OPT_DEINTERLACER_INFO_0_EL,
      NULL,
      "video",
      {
         { "weave",      OPTION_VAL_WEAVE_EL },
         { "bob",        OPTION_VAL_BOB_EL },
         { "bob_offset", OPTION_VAL_BOB_OFFSET_EL },
         { "fastmad",    OPTION_VAL_FASTMAD_EL },
         { "off",        "Off" },
         { NULL, NULL },
      },
      "weave"
   },
   {
      BEETLE_OPT(frame_duping),
      BEETLE_OPT_FRAME_DUPING_LABEL_EL,
      NULL,
      BEETLE_OPT_FRAME_DUPING_INFO_0_EL,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(pal_video_timing_override),
      BEETLE_OPT_PAL_VIDEO_TIMING_OVERRIDE_LABEL_EL,
      NULL,
      BEETLE_OPT_PAL_VIDEO_TIMING_OVERRIDE_INFO_0_EL,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(aspect_ratio),
      BEETLE_OPT_ASPECT_RATIO_LABEL_EL,
      NULL,
      BEETLE_OPT_ASPECT_RATIO_INFO_0_EL,
      NULL,
      "video",
      {
         { "corrected", OPTION_VAL_CORRECTED_EL },
         { "uncorrected", OPTION_VAL_UNCORRECTED_EL },
         { "4:3",  OPTION_VAL_4_3_EL },
         { "ntsc", OPTION_VAL_NTSC_EL },
      },
      "corrected"
   },
   {
      BEETLE_OPT(crop_overscan),
      BEETLE_OPT_CROP_OVERSCAN_LABEL_EL,
      NULL,
      BEETLE_OPT_CROP_OVERSCAN_INFO_0_EL,
      NULL,
      "video",
      {
         { "disabled", "Disabled" },
         { "static",   OPTION_VAL_STATIC_EL },
         { "smart",    OPTION_VAL_SMART_EL },
         { NULL, NULL },
      },
      "smart"
   },
   {
      BEETLE_OPT(image_crop),
      BEETLE_OPT_IMAGE_CROP_LABEL_EL,
      NULL,
      BEETLE_OPT_IMAGE_CROP_INFO_0_EL,
      NULL,
      "video",
      {
         { "disabled", "0" },
         { "1px",      OPTION_VAL_1PX_EL },
         { "2px",      OPTION_VAL_2PX_EL },
         { "3px",      OPTION_VAL_3PX_EL },
         { "4px",      OPTION_VAL_4PX_EL },
         { "5px",      OPTION_VAL_5PX_EL },
         { "6px",      OPTION_VAL_6PX_EL },
         { "7px",      OPTION_VAL_7PX_EL },
         { "8px",      OPTION_VAL_8PX_EL },
         { "9px",      OPTION_VAL_9PX_EL },
         { "10px",     OPTION_VAL_10PX_EL },
         { "11px",     OPTION_VAL_11PX_EL },
         { "12px",     OPTION_VAL_12PX_EL },
         { "13px",     OPTION_VAL_13PX_EL },
         { "14px",     OPTION_VAL_14PX_EL },
         { "15px",     OPTION_VAL_15PX_EL },
         { "16px",     OPTION_VAL_16PX_EL },
         { "17px",     OPTION_VAL_17PX_EL },
         { "18px",     OPTION_VAL_18PX_EL },
         { "19px",     OPTION_VAL_19PX_EL },
         { "20px",     OPTION_VAL_20PX_EL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(image_offset),
      BEETLE_OPT_IMAGE_OFFSET_LABEL_EL,
      NULL,
      BEETLE_OPT_IMAGE_OFFSET_INFO_0_EL,
      NULL,
      "video",
      {
         { "-12px",    OPTION_VAL_12PX_O64_EL },
         { "-11px",    OPTION_VAL_11PX_O64_EL },
         { "-10px",    OPTION_VAL_10PX_O64_EL },
         { "-9px",     OPTION_VAL_9PX_O64_EL },
         { "-8px",     OPTION_VAL_8PX_O64_EL },
         { "-7px",     OPTION_VAL_7PX_O64_EL },
         { "-6px",     OPTION_VAL_6PX_O64_EL },
         { "-5px",     OPTION_VAL_5PX_O64_EL },
         { "-4px",     OPTION_VAL_4PX_O64_EL },
         { "-3px",     OPTION_VAL_3PX_O64_EL },
         { "-2px",     OPTION_VAL_2PX_O64_EL },
         { "-1px",     OPTION_VAL_1PX_O64_EL },
         { "disabled", "0" },
         { "+1px",     OPTION_VAL_1PX_O64_0_EL },
         { "+2px",     OPTION_VAL_2PX_O64_0_EL },
         { "+3px",     OPTION_VAL_3PX_O64_0_EL },
         { "+4px",     OPTION_VAL_4PX_O64_0_EL },
         { "+5px",     OPTION_VAL_5PX_O64_0_EL },
         { "+6px",     OPTION_VAL_6PX_O64_0_EL },
         { "+7px",     OPTION_VAL_7PX_O64_0_EL },
         { "+8px",     OPTION_VAL_8PX_O64_0_EL },
         { "+9px",     OPTION_VAL_9PX_O64_0_EL },
         { "+10px",    OPTION_VAL_10PX_O64_0_EL },
         { "+11px",    OPTION_VAL_11PX_O64_0_EL },
         { "+12px",    OPTION_VAL_12PX_O64_0_EL },
         { NULL, NULL },
      },
      "disabled"
   },
#if defined(HAVE_OPENGL) || defined(HAVE_OPENGLES) || defined(HAVE_VULKAN)
   {
      BEETLE_OPT(image_offset_cycles),
      BEETLE_OPT_IMAGE_OFFSET_CYCLES_LABEL_EL,
      NULL,
      BEETLE_OPT_IMAGE_OFFSET_CYCLES_INFO_0_EL,
      NULL,
      "video",
      {
         { "-40",      NULL },
         { "-39",      NULL },
         { "-38",      NULL },
         { "-37",      NULL },
         { "-36",      NULL },
         { "-35",      NULL },
         { "-34",      NULL },
         { "-33",      NULL },
         { "-32",      NULL },
         { "-31",      NULL },
         { "-30",      NULL },
         { "-29",      NULL },
         { "-28",      NULL },
         { "-27",      NULL },
         { "-26",      NULL },
         { "-25",      NULL },
         { "-24",      NULL },
         { "-23",      NULL },
         { "-22",      NULL },
         { "-21",      NULL },
         { "-20",      NULL },
         { "-19",      NULL },
         { "-18",      NULL },
         { "-17",      NULL },
         { "-16",      NULL },
         { "-15",      NULL },
         { "-14",      NULL },
         { "-13",      NULL },
         { "-12",      NULL },
         { "-11",      NULL },
         { "-10",      NULL },
         { "-9",       NULL },
         { "-8",       NULL },
         { "-7",       NULL },
         { "-6",       NULL },
         { "-5",       NULL },
         { "-4",       NULL },
         { "-3",       NULL },
         { "-2",       NULL },
         { "-1",       NULL },
         { "0",        "0" },
         { "+1",       NULL },
         { "+2",       NULL },
         { "+3",       NULL },
         { "+4",       NULL },
         { "+5",       NULL },
         { "+6",       NULL },
         { "+7",       NULL },
         { "+8",       NULL },
         { "+9",       NULL },
         { "+10",      NULL },
         { "+11",      NULL },
         { "+12",      NULL },
         { "+13",      NULL },
         { "+14",      NULL },
         { "+15",      NULL },
         { "+16",      NULL },
         { "+17",      NULL },
         { "+18",      NULL },
         { "+19",      NULL },
         { "+20",      NULL },
         { "+21",      NULL },
         { "+22",      NULL },
         { "+23",      NULL },
         { "+24",      NULL },
         { "+25",      NULL },
         { "+26",      NULL },
         { "+27",      NULL },
         { "+28",      NULL },
         { "+29",      NULL },
         { "+30",      NULL },
         { "+31",      NULL },
         { "+32",      NULL },
         { "+33",      NULL },
         { "+34",      NULL },
         { "+35",      NULL },
         { "+36",      NULL },
         { "+37",      NULL },
         { "+38",      NULL },
         { "+39",      NULL },
         { "+40",      NULL },
         { NULL, NULL},
      },
      "0"
   },
#endif
   {
      BEETLE_OPT(initial_scanline),
      BEETLE_OPT_INITIAL_SCANLINE_LABEL_EL,
      NULL,
      BEETLE_OPT_INITIAL_SCANLINE_INFO_0_EL,
      NULL,
      "video",
      {
         { "0",  OPTION_VAL_0_O48_EL },
         { "1",  NULL },
         { "2",  NULL },
         { "3",  NULL },
         { "4",  NULL },
         { "5",  NULL },
         { "6",  NULL },
         { "7",  NULL },
         { "8",  NULL },
         { "9",  NULL },
         { "10", NULL },
         { "11", NULL },
         { "12", NULL },
         { "13", NULL },
         { "14", NULL },
         { "15", NULL },
         { "16", NULL },
         { "17", NULL },
         { "18", NULL },
         { "19", NULL },
         { "20", NULL },
         { "21", NULL },
         { "22", NULL },
         { "23", NULL },
         { "24", NULL },
         { "25", NULL },
         { "26", NULL },
         { "27", NULL },
         { "28", NULL },
         { "29", NULL },
         { "30", NULL },
         { "31", NULL },
         { "32", NULL },
         { "33", NULL },
         { "34", NULL },
         { "35", NULL },
         { "36", NULL },
         { "37", NULL },
         { "38", NULL },
         { "39", NULL },
         { "40", NULL },
         { NULL, NULL },
      },
      "0"
   },
   {
      BEETLE_OPT(last_scanline),
      BEETLE_OPT_LAST_SCANLINE_LABEL_EL,
      NULL,
      BEETLE_OPT_LAST_SCANLINE_INFO_0_EL,
      NULL,
      "video",
      {
         { "210", NULL },
         { "211", NULL },
         { "212", NULL },
         { "213", NULL },
         { "214", NULL },
         { "215", NULL },
         { "216", NULL },
         { "217", NULL },
         { "218", NULL },
         { "219", NULL },
         { "220", NULL },
         { "221", NULL },
         { "222", NULL },
         { "223", NULL },
         { "224", NULL },
         { "225", NULL },
         { "226", NULL },
         { "227", NULL },
         { "228", NULL },
         { "229", NULL },
         { "230", NULL },
         { "231", NULL },
         { "232", NULL },
         { "233", NULL },
         { "234", NULL },
         { "235", NULL },
         { "236", NULL },
         { "237", NULL },
         { "238", NULL },
         { "239", OPTION_VAL_239_EL },
         { NULL, NULL },
      },
      "239"
   },
   {
      BEETLE_OPT(initial_scanline_pal),
      BEETLE_OPT_INITIAL_SCANLINE_PAL_LABEL_EL,
      NULL,
      BEETLE_OPT_INITIAL_SCANLINE_PAL_INFO_0_EL,
      NULL,
      "video",
      {
         { "0",  OPTION_VAL_0_O48_EL },
         { "1",  NULL },
         { "2",  NULL },
         { "3",  NULL },
         { "4",  NULL },
         { "5",  NULL },
         { "6",  NULL },
         { "7",  NULL },
         { "8",  NULL },
         { "9",  NULL },
         { "10", NULL },
         { "11", NULL },
         { "12", NULL },
         { "13", NULL },
         { "14", NULL },
         { "15", NULL },
         { "16", NULL },
         { "17", NULL },
         { "18", NULL },
         { "19", NULL },
         { "20", NULL },
         { "21", NULL },
         { "22", NULL },
         { "23", NULL },
         { "24", NULL },
         { "25", NULL },
         { "26", NULL },
         { "27", NULL },
         { "28", NULL },
         { "29", NULL },
         { "30", NULL },
         { "31", NULL },
         { "32", NULL },
         { "33", NULL },
         { "34", NULL },
         { "35", NULL },
         { "36", NULL },
         { "37", NULL },
         { "38", NULL },
         { "39", NULL },
         { "40", NULL },
         { NULL, NULL },
      },
      "0"
   },
   {
      BEETLE_OPT(last_scanline_pal),
      BEETLE_OPT_LAST_SCANLINE_PAL_LABEL_EL,
      NULL,
      BEETLE_OPT_LAST_SCANLINE_PAL_INFO_0_EL,
      NULL,
      "video",
      {
         { "230", NULL },
         { "231", NULL },
         { "232", NULL },
         { "233", NULL },
         { "234", NULL },
         { "235", NULL },
         { "236", NULL },
         { "237", NULL },
         { "238", NULL },
         { "239", NULL },
         { "240", NULL },
         { "241", NULL },
         { "242", NULL },
         { "243", NULL },
         { "244", NULL },
         { "245", NULL },
         { "246", NULL },
         { "247", NULL },
         { "248", NULL },
         { "249", NULL },
         { "250", NULL },
         { "251", NULL },
         { "252", NULL },
         { "253", NULL },
         { "254", NULL },
         { "255", NULL },
         { "256", NULL },
         { "257", NULL },
         { "258", NULL },
         { "259", NULL },
         { "260", NULL },
         { "261", NULL },
         { "262", NULL },
         { "263", NULL },
         { "264", NULL },
         { "265", NULL },
         { "266", NULL },
         { "267", NULL },
         { "268", NULL },
         { "269", NULL },
         { "270", NULL },
         { "271", NULL },
         { "272", NULL },
         { "273", NULL },
         { "274", NULL },
         { "275", NULL },
         { "276", NULL },
         { "277", NULL },
         { "278", NULL },
         { "279", NULL },
         { "280", NULL },
         { "281", NULL },
         { "282", NULL },
         { "283", NULL },
         { "284", NULL },
         { "285", NULL },
         { "286", NULL },
         { "287", OPTION_VAL_287_EL },
         { NULL, NULL },
      },
      "287"
   },
   { NULL, NULL, NULL, NULL, NULL, NULL, {{0}}, NULL },
};
struct retro_core_options_v2 options_el = {
   option_cats_el,
   option_defs_el
};

/* RETRO_LANGUAGE_EN */

#define CATEGORY_SYSTEM_LABEL_EN NULL
#define CATEGORY_SYSTEM_INFO_0_EN NULL
#define CATEGORY_VIDEO_LABEL_EN NULL
#define CATEGORY_VIDEO_INFO_0_EN NULL
#define CATEGORY_PGXP_LABEL_EN NULL
#define CATEGORY_PGXP_INFO_0_EN NULL
#define CATEGORY_OSD_LABEL_EN NULL
#define CATEGORY_OSD_INFO_0_EN NULL
#define CATEGORY_INPUT_LABEL_EN NULL
#define CATEGORY_INPUT_INFO_0_EN NULL
#define CATEGORY_MEMCARDS_LABEL_EN NULL
#define CATEGORY_MEMCARDS_INFO_0_EN NULL
#define CATEGORY_HACKS_LABEL_EN NULL
#define CATEGORY_HACKS_INFO_0_EN NULL
#define BEETLE_OPT_CPU_FREQ_SCALE_LABEL_EN "CPU Frequency Scaling (Over-clock)"
#define BEETLE_OPT_CPU_FREQ_SCALE_INFO_0_EN "Over-clock (or under-clock) the emulated PS1 CPU. Overclocking can eliminate slowdown and improve frame rates in certain games at the expense of increased performance requirements. Note that some games have an internal frame rate limiter and may not benefit from overclocking. May cause certain effects to animate faster than intended in some titles when overclocked."
#define OPTION_VAL_50_EN NULL
#define OPTION_VAL_60_EN NULL
#define OPTION_VAL_70_EN NULL
#define OPTION_VAL_80_EN NULL
#define OPTION_VAL_90_EN NULL
#define OPTION_VAL_100_EN NULL
#define OPTION_VAL_110_EN NULL
#define OPTION_VAL_120_EN NULL
#define OPTION_VAL_130_EN NULL
#define OPTION_VAL_140_EN NULL
#define OPTION_VAL_150_EN NULL
#define OPTION_VAL_160_EN NULL
#define OPTION_VAL_170_EN NULL
#define OPTION_VAL_180_EN NULL
#define OPTION_VAL_190_EN NULL
#define OPTION_VAL_200_EN NULL
#define OPTION_VAL_210_EN NULL
#define OPTION_VAL_220_EN NULL
#define OPTION_VAL_230_EN NULL
#define OPTION_VAL_240_EN NULL
#define OPTION_VAL_250_EN NULL
#define OPTION_VAL_260_EN NULL
#define OPTION_VAL_270_EN NULL
#define OPTION_VAL_280_EN NULL
#define OPTION_VAL_290_EN NULL
#define OPTION_VAL_300_EN NULL
#define OPTION_VAL_310_EN NULL
#define OPTION_VAL_320_EN NULL
#define OPTION_VAL_330_EN NULL
#define OPTION_VAL_340_EN NULL
#define OPTION_VAL_350_EN NULL
#define OPTION_VAL_360_EN NULL
#define OPTION_VAL_370_EN NULL
#define OPTION_VAL_380_EN NULL
#define OPTION_VAL_390_EN NULL
#define OPTION_VAL_400_EN NULL
#define OPTION_VAL_410_EN NULL
#define OPTION_VAL_420_EN NULL
#define OPTION_VAL_430_EN NULL
#define OPTION_VAL_440_EN NULL
#define OPTION_VAL_450_EN NULL
#define OPTION_VAL_460_EN NULL
#define OPTION_VAL_470_EN NULL
#define OPTION_VAL_480_EN NULL
#define OPTION_VAL_490_EN NULL
#define OPTION_VAL_500_EN NULL
#define OPTION_VAL_510_EN NULL
#define OPTION_VAL_520_EN NULL
#define OPTION_VAL_530_EN NULL
#define OPTION_VAL_540_EN NULL
#define OPTION_VAL_550_EN NULL
#define OPTION_VAL_560_EN NULL
#define OPTION_VAL_570_EN NULL
#define OPTION_VAL_580_EN NULL
#define OPTION_VAL_590_EN NULL
#define OPTION_VAL_600_EN NULL
#define OPTION_VAL_610_EN NULL
#define OPTION_VAL_620_EN NULL
#define OPTION_VAL_630_EN NULL
#define OPTION_VAL_640_EN NULL
#define OPTION_VAL_650_EN NULL
#define OPTION_VAL_660_EN NULL
#define OPTION_VAL_670_EN NULL
#define OPTION_VAL_680_EN NULL
#define OPTION_VAL_690_EN NULL
#define OPTION_VAL_700_EN NULL
#define OPTION_VAL_710_EN NULL
#define OPTION_VAL_720_EN NULL
#define OPTION_VAL_730_EN NULL
#define OPTION_VAL_740_EN NULL
#define OPTION_VAL_750_EN NULL
#define BEETLE_OPT_GPU_OVERCLOCK_LABEL_EN "GPU Rasteriser Over-clock"
#define BEETLE_OPT_GPU_OVERCLOCK_INFO_0_EN "Enable over-clocking of the 2D rasteriser contained within the emulated PS1's GPU. Does not improve 3D rendering, and in general has little effect."
#define OPTION_VAL_1X_NATIVE_EN NULL
#define OPTION_VAL_2X_EN NULL
#define OPTION_VAL_4X_EN NULL
#define OPTION_VAL_8X_EN NULL
#define OPTION_VAL_16X_EN NULL
#define OPTION_VAL_32X_EN NULL
#define BEETLE_OPT_GTE_OVERCLOCK_LABEL_EN "GTE Over-clock"
#define BEETLE_OPT_GTE_OVERCLOCK_INFO_0_EN NULL
#define BEETLE_OPT_CD_ACCESS_METHOD_LABEL_EN NULL
#define BEETLE_OPT_CD_ACCESS_METHOD_INFO_0_EN NULL
#define OPTION_VAL_SYNC_EN NULL
#define OPTION_VAL_ASYNC_EN NULL
#define OPTION_VAL_PRECACHE_EN NULL
#define BEETLE_OPT_CD_FASTLOAD_LABEL_EN NULL
#define BEETLE_OPT_CD_FASTLOAD_INFO_0_EN NULL
#define OPTION_VAL_2X_NATIVE_EN NULL
#define OPTION_VAL_6X_EN NULL
#define OPTION_VAL_10X_EN NULL
#define OPTION_VAL_12X_EN NULL
#define OPTION_VAL_14X_EN NULL
#define BEETLE_OPT_SKIP_BIOS_LABEL_EN NULL
#define BEETLE_OPT_SKIP_BIOS_INFO_0_EN NULL
#define BEETLE_OPT_OVERRIDE_BIOS_LABEL_EN NULL
#define BEETLE_OPT_OVERRIDE_BIOS_INFO_0_EN NULL
#define OPTION_VAL_PSXONPSP_EN NULL
#define OPTION_VAL_PS1_ROM_EN NULL
#define OPTION_VAL_OPENBIOS_EN NULL
#define BEETLE_OPT_REGION_LABEL_EN NULL
#define BEETLE_OPT_REGION_INFO_0_EN NULL
#define OPTION_VAL_AUTO_EN NULL
#define OPTION_VAL_NTSC_J_EN NULL
#define OPTION_VAL_NTSC_U_EN NULL
#define OPTION_VAL_PAL_EN NULL
#define BEETLE_OPT_INTERNAL_RESOLUTION_LABEL_EN NULL
#define BEETLE_OPT_INTERNAL_RESOLUTION_INFO_0_EN NULL
#define BEETLE_OPT_RENDERER_LABEL_EN NULL
#define BEETLE_OPT_RENDERER_INFO_0_EN NULL
#define OPTION_VAL_HARDWARE_EN NULL
#define OPTION_VAL_HARDWARE_GL_EN NULL
#define OPTION_VAL_HARDWARE_VK_EN NULL
#define OPTION_VAL_SOFTWARE_EN NULL
#define BEETLE_OPT_RENDERER_SOFTWARE_FB_LABEL_EN "Software Frame-buffer"
#define BEETLE_OPT_RENDERER_SOFTWARE_FB_INFO_0_EN "Enable accurate emulation of frame-buffer effects (e.g. motion blur, FF7 battle swirl) when using hardware renderers by running a copy of the software renderer at native resolution in the background. If disabled, these operations are omitted (OpenGL) or rendered on the GPU (Vulkan). Disabling can improve performance but may cause severe graphical errors. Leave enabled if unsure."
#define BEETLE_OPT_DEPTH_LABEL_EN "Internal Colour Depth"
#define BEETLE_OPT_DEPTH_INFO_0_EN "Higher colour depth can reduce colour banding effects without the use of dithering. 16 bpp emulates original hardware but may have visible banding if dithering is not enabled. 'Dithering Pattern' is recommended to be disabled when this option is set to 32 bpp."
#define OPTION_VAL_16BPP_NATIVE_EN NULL
#define OPTION_VAL_32BPP_EN NULL
#define BEETLE_OPT_DITHER_MODE_LABEL_EN NULL
#define BEETLE_OPT_DITHER_MODE_INFO_0_EN "'1x (Native)' emulates native low resolution dithering used by original hardware to smooth out colour banding artifacts visible at native colour depth. 'Internal Resolution' scales dithering granularity to the configured internal resolution for cleaner results. Recommended to be disabled when running at 32 bpp colour depth. Note: On Vulkan, enabling this will force downsampling to native colour depth, while disabling will automatically enable output at higher colour depth."
#define OPTION_VAL_INTERNAL_RESOLUTION_EN NULL
#define BEETLE_OPT_SCALED_UV_OFFSET_LABEL_EN NULL
#define BEETLE_OPT_SCALED_UV_OFFSET_INFO_0_EN NULL
#define BEETLE_OPT_FILTER_LABEL_EN NULL
#define BEETLE_OPT_FILTER_INFO_0_EN NULL
#define OPTION_VAL_NEAREST_EN NULL
#define OPTION_VAL_SABR_EN NULL
#define OPTION_VAL_XBR_EN NULL
#define OPTION_VAL_BILINEAR_EN NULL
#define OPTION_VAL_3_POINT_EN NULL
#define OPTION_VAL_JINC2_EN NULL
#define BEETLE_OPT_FILTER_EXCLUDE_SPRITE_LABEL_EN NULL
#define BEETLE_OPT_FILTER_EXCLUDE_SPRITE_INFO_0_EN "Do not apply texture filtering to sprites. That prevents seams in various games with 2D sprite-rendered backgrounds. Use together with Adaptive Smoothing or another post-processing filter for best effect."
#define OPTION_VAL_OPAQUE_EN NULL
#define OPTION_VAL_ALL_EN NULL
#define BEETLE_OPT_FILTER_EXCLUDE_2D_POLYGON_LABEL_EN NULL
#define BEETLE_OPT_FILTER_EXCLUDE_2D_POLYGON_INFO_0_EN NULL
#define BEETLE_OPT_ADAPTIVE_SMOOTHING_LABEL_EN NULL
#define BEETLE_OPT_ADAPTIVE_SMOOTHING_INFO_0_EN NULL
#define BEETLE_OPT_SUPER_SAMPLING_LABEL_EN "Super-sampling (Downsample to Native Resolution)"
#define BEETLE_OPT_SUPER_SAMPLING_INFO_0_EN NULL
#define BEETLE_OPT_MSAA_LABEL_EN NULL
#define BEETLE_OPT_MSAA_INFO_0_EN NULL
#define OPTION_VAL_1X_EN NULL
#define BEETLE_OPT_MDEC_YUV_LABEL_EN NULL
#define BEETLE_OPT_MDEC_YUV_INFO_0_EN "Improve the quality of FMV playback by reducing 'macro-blocking' artefacts (squares/jagged edges). Only supported by the Vulkan renderer."
#define BEETLE_OPT_TRACK_TEXTURES_LABEL_EN NULL
#define BEETLE_OPT_TRACK_TEXTURES_INFO_0_EN NULL
#define BEETLE_OPT_DUMP_TEXTURES_LABEL_EN NULL
#define BEETLE_OPT_DUMP_TEXTURES_INFO_0_EN NULL
#define BEETLE_OPT_REPLACE_TEXTURES_LABEL_EN NULL
#define BEETLE_OPT_REPLACE_TEXTURES_INFO_0_EN NULL
#define BEETLE_OPT_PGXP_MODE_LABEL_EN NULL
#define BEETLE_OPT_PGXP_MODE_INFO_0_EN NULL
#define OPTION_VAL_MEMORY_ONLY_EN NULL
#define OPTION_VAL_MEMORY_CPU_EN NULL
#define BEETLE_OPT_PGXP_2D_TOL_LABEL_EN NULL
#define BEETLE_OPT_PGXP_2D_TOL_INFO_0_EN NULL
#define OPTION_VAL_0PX_EN NULL
#define OPTION_VAL_1PX_EN NULL
#define OPTION_VAL_2PX_EN NULL
#define OPTION_VAL_3PX_EN NULL
#define OPTION_VAL_4PX_EN NULL
#define OPTION_VAL_5PX_EN NULL
#define OPTION_VAL_6PX_EN NULL
#define OPTION_VAL_7PX_EN NULL
#define OPTION_VAL_8PX_EN NULL
#define BEETLE_OPT_PGXP_NCLIP_LABEL_EN NULL
#define BEETLE_OPT_PGXP_NCLIP_INFO_0_EN NULL
#define BEETLE_OPT_PGXP_VERTEX_LABEL_EN NULL
#define BEETLE_OPT_PGXP_VERTEX_INFO_0_EN NULL
#define BEETLE_OPT_PGXP_TEXTURE_LABEL_EN NULL
#define BEETLE_OPT_PGXP_TEXTURE_INFO_0_EN NULL
#define BEETLE_OPT_DISPLAY_INTERNAL_FPS_LABEL_EN NULL
#define BEETLE_OPT_DISPLAY_INTERNAL_FPS_INFO_0_EN "Display the internal frame rate at which the emulated PlayStation system is rendering content. Note: Requires on-screen notifications to be enabled in the libretro front-end."
#define BEETLE_OPT_DISPLAY_OSD_LABEL_EN NULL
#define BEETLE_OPT_DISPLAY_OSD_INFO_0_EN NULL
#define BEETLE_OPT_DISPLAY_VRAM_LABEL_EN NULL
#define BEETLE_OPT_DISPLAY_VRAM_INFO_0_EN "Visualise the entire emulated console's VRAM. Only supported by the OpenGL and Vulkan hardware renderers. Note: This is for debugging purposes, and should normally be disabled."
#define BEETLE_OPT_ANALOG_CALIBRATION_LABEL_EN "Analogue Self-Calibration"
#define BEETLE_OPT_ANALOG_CALIBRATION_INFO_0_EN NULL
#define BEETLE_OPT_ANALOG_TOGGLE_LABEL_EN "DualShock Analogue Mode Toggle"
#define BEETLE_OPT_ANALOG_TOGGLE_INFO_0_EN NULL
#define OPTION_VAL_ENABLED_ANALOG_EN NULL
#define BEETLE_OPT_ANALOG_TOGGLE_COMBO_LABEL_EN NULL
#define BEETLE_OPT_ANALOG_TOGGLE_COMBO_INFO_0_EN NULL
#define OPTION_VAL_L1_L2_R1_R2_START_SELECT_EN NULL
#define OPTION_VAL_L1_R1_SELECT_EN NULL
#define OPTION_VAL_L1_R1_START_EN NULL
#define OPTION_VAL_L1_R1_L3_EN NULL
#define OPTION_VAL_L1_R1_R3_EN NULL
#define OPTION_VAL_L2_R2_SELECT_EN NULL
#define OPTION_VAL_L2_R2_START_EN NULL
#define OPTION_VAL_L2_R2_L3_EN NULL
#define OPTION_VAL_L2_R2_R3_EN NULL
#define OPTION_VAL_L3_R3_EN NULL
#define BEETLE_OPT_ANALOG_TOGGLE_HOLD_LABEL_EN NULL
#define BEETLE_OPT_ANALOG_TOGGLE_HOLD_INFO_0_EN NULL
#define OPTION_VAL_0_EN NULL
#define OPTION_VAL_1_EN NULL
#define OPTION_VAL_2_EN NULL
#define OPTION_VAL_3_EN NULL
#define OPTION_VAL_4_EN NULL
#define OPTION_VAL_5_EN NULL
#define BEETLE_OPT_ENABLE_MULTITAP_PORT1_LABEL_EN NULL
#define BEETLE_OPT_ENABLE_MULTITAP_PORT1_INFO_0_EN "Enable multi-tap functionality on port 1."
#define BEETLE_OPT_ENABLE_MULTITAP_PORT2_LABEL_EN NULL
#define BEETLE_OPT_ENABLE_MULTITAP_PORT2_INFO_0_EN "Enable multi-tap functionality on port 2."
#define BEETLE_OPT_GUN_INPUT_MODE_LABEL_EN NULL
#define BEETLE_OPT_GUN_INPUT_MODE_INFO_0_EN NULL
#define OPTION_VAL_LIGHTGUN_EN NULL
#define OPTION_VAL_TOUCHSCREEN_EN NULL
#define BEETLE_OPT_GUN_CURSOR_LABEL_EN NULL
#define BEETLE_OPT_GUN_CURSOR_INFO_0_EN NULL
#define OPTION_VAL_CROSS_EN NULL
#define OPTION_VAL_DOT_EN NULL
#define BEETLE_OPT_CROSSHAIR_COLOR_P1_LABEL_EN NULL
#define BEETLE_OPT_CROSSHAIR_COLOR_P1_INFO_0_EN NULL
#define OPTION_VAL_RED_EN NULL
#define OPTION_VAL_BLUE_EN NULL
#define OPTION_VAL_GREEN_EN NULL
#define OPTION_VAL_ORANGE_EN NULL
#define OPTION_VAL_YELLOW_EN NULL
#define OPTION_VAL_CYAN_EN NULL
#define OPTION_VAL_PINK_EN NULL
#define OPTION_VAL_PURPLE_EN NULL
#define OPTION_VAL_BLACK_EN NULL
#define OPTION_VAL_WHITE_EN NULL
#define BEETLE_OPT_CROSSHAIR_COLOR_P2_LABEL_EN NULL
#define BEETLE_OPT_CROSSHAIR_COLOR_P2_INFO_0_EN NULL
#define BEETLE_OPT_MOUSE_SENSITIVITY_LABEL_EN NULL
#define BEETLE_OPT_MOUSE_SENSITIVITY_INFO_0_EN NULL
#define OPTION_VAL_5_O42_EN NULL
#define OPTION_VAL_10_EN NULL
#define OPTION_VAL_15_EN NULL
#define OPTION_VAL_20_EN NULL
#define OPTION_VAL_25_EN NULL
#define OPTION_VAL_30_EN NULL
#define OPTION_VAL_35_EN NULL
#define OPTION_VAL_40_EN NULL
#define OPTION_VAL_45_EN NULL
#define OPTION_VAL_55_EN NULL
#define OPTION_VAL_65_EN NULL
#define OPTION_VAL_75_EN NULL
#define OPTION_VAL_85_EN NULL
#define OPTION_VAL_95_EN NULL
#define OPTION_VAL_100_O42_EN NULL
#define OPTION_VAL_105_EN NULL
#define OPTION_VAL_115_EN NULL
#define OPTION_VAL_125_EN NULL
#define OPTION_VAL_135_EN NULL
#define OPTION_VAL_145_EN NULL
#define OPTION_VAL_155_EN NULL
#define OPTION_VAL_165_EN NULL
#define OPTION_VAL_175_EN NULL
#define OPTION_VAL_185_EN NULL
#define OPTION_VAL_195_EN NULL
#define BEETLE_OPT_NEGCON_RESPONSE_LABEL_EN NULL
#define BEETLE_OPT_NEGCON_RESPONSE_INFO_0_EN NULL
#define OPTION_VAL_LINEAR_EN NULL
#define OPTION_VAL_QUADRATIC_EN NULL
#define OPTION_VAL_CUBIC_EN NULL
#define BEETLE_OPT_NEGCON_DEADZONE_LABEL_EN "neGcon Twist Dead-zone"
#define BEETLE_OPT_NEGCON_DEADZONE_INFO_0_EN NULL
#define OPTION_VAL_0_O44_EN NULL
#define BEETLE_OPT_USE_MEDNAFEN_MEMCARD0_METHOD_LABEL_EN NULL
#define BEETLE_OPT_USE_MEDNAFEN_MEMCARD0_METHOD_INFO_0_EN NULL
#define OPTION_VAL_LIBRETRO_EN NULL
#define OPTION_VAL_MEDNAFEN_EN NULL
#define BEETLE_OPT_ENABLE_MEMCARD1_LABEL_EN NULL
#define BEETLE_OPT_ENABLE_MEMCARD1_INFO_0_EN NULL
#define BEETLE_OPT_SHARED_MEMORY_CARDS_LABEL_EN NULL
#define BEETLE_OPT_SHARED_MEMORY_CARDS_INFO_0_EN NULL
#define BEETLE_OPT_MEMCARD_LEFT_INDEX_LABEL_EN NULL
#define BEETLE_OPT_MEMCARD_LEFT_INDEX_INFO_0_EN NULL
#define OPTION_VAL_0_O48_EN NULL
#define BEETLE_OPT_MEMCARD_RIGHT_INDEX_LABEL_EN NULL
#define BEETLE_OPT_MEMCARD_RIGHT_INDEX_INFO_0_EN NULL
#define OPTION_VAL_1_O49_EN NULL
#define BEETLE_OPT_LINE_RENDER_LABEL_EN NULL
#define BEETLE_OPT_LINE_RENDER_INFO_0_EN NULL
#define OPTION_VAL_DEFAULT_EN NULL
#define OPTION_VAL_AGGRESSIVE_EN NULL
#define BEETLE_OPT_WIDESCREEN_HACK_LABEL_EN NULL
#define BEETLE_OPT_WIDESCREEN_HACK_INFO_0_EN "Render 3D content in an anamorphic format and output the emulated frame-buffer at a widescreen aspect ratio. Produces the best results with fully 3D games. 2D elements will be horizontally stretched and may be misaligned."
#define BEETLE_OPT_WIDESCREEN_HACK_ASPECT_RATIO_LABEL_EN NULL
#define BEETLE_OPT_WIDESCREEN_HACK_ASPECT_RATIO_INFO_0_EN NULL
#define OPTION_VAL_16_9_EN NULL
#define OPTION_VAL_16_10_EN NULL
#define OPTION_VAL_18_9_EN NULL
#define OPTION_VAL_19_9_EN NULL
#define OPTION_VAL_20_9_EN NULL
#define OPTION_VAL_21_9_EN NULL
#define OPTION_VAL_32_9_EN NULL
#define BEETLE_OPT_CPU_DYNAREC_LABEL_EN "CPU Dyna-rec"
#define BEETLE_OPT_CPU_DYNAREC_INFO_0_EN "Dynamically recompile CPU instructions to native instructions. Much faster than the interpreter, but CPU timing is less accurate, and may have bugs."
#define OPTION_VAL_DISABLED_EN NULL
#define OPTION_VAL_EXECUTE_EN "Maximum Performance"
#define OPTION_VAL_EXECUTE_ONE_EN NULL
#define OPTION_VAL_RUN_INTERPRETER_EN "Light-rec Interpreter"
#define BEETLE_OPT_DYNAREC_INVALIDATE_LABEL_EN "Dyna-rec Code Invalidation"
#define BEETLE_OPT_DYNAREC_INVALIDATE_INFO_0_EN NULL
#define OPTION_VAL_FULL_EN NULL
#define OPTION_VAL_DMA_EN NULL
#define BEETLE_OPT_DYNAREC_EVENTCYCLES_LABEL_EN NULL
#define BEETLE_OPT_DYNAREC_EVENTCYCLES_INFO_0_EN NULL
#define OPTION_VAL_128_EN NULL
#define BEETLE_OPT_DYNAREC_SPU_SAMPLES_LABEL_EN NULL
#define BEETLE_OPT_DYNAREC_SPU_SAMPLES_INFO_0_EN NULL
#define BEETLE_OPT_CORE_TIMING_FPS_LABEL_EN NULL
#define BEETLE_OPT_CORE_TIMING_FPS_INFO_0_EN NULL
#define OPTION_VAL_AUTO_TOGGLE_EN NULL
#define OPTION_VAL_FORCE_PROGRESSIVE_EN NULL
#define OPTION_VAL_FORCE_INTERLACED_EN NULL
#define BEETLE_OPT_DEINTERLACER_LABEL_EN NULL
#define BEETLE_OPT_DEINTERLACER_INFO_0_EN NULL
#define OPTION_VAL_WEAVE_EN NULL
#define OPTION_VAL_BOB_EN NULL
#define OPTION_VAL_BOB_OFFSET_EN NULL
#define OPTION_VAL_FASTMAD_EN NULL
#define BEETLE_OPT_FRAME_DUPING_LABEL_EN NULL
#define BEETLE_OPT_FRAME_DUPING_INFO_0_EN NULL
#define BEETLE_OPT_PAL_VIDEO_TIMING_OVERRIDE_LABEL_EN NULL
#define BEETLE_OPT_PAL_VIDEO_TIMING_OVERRIDE_INFO_0_EN NULL
#define BEETLE_OPT_ASPECT_RATIO_LABEL_EN NULL
#define BEETLE_OPT_ASPECT_RATIO_INFO_0_EN NULL
#define OPTION_VAL_CORRECTED_EN NULL
#define OPTION_VAL_UNCORRECTED_EN NULL
#define OPTION_VAL_4_3_EN NULL
#define OPTION_VAL_NTSC_EN NULL
#define BEETLE_OPT_CROP_OVERSCAN_LABEL_EN NULL
#define BEETLE_OPT_CROP_OVERSCAN_INFO_0_EN NULL
#define OPTION_VAL_STATIC_EN NULL
#define OPTION_VAL_SMART_EN NULL
#define BEETLE_OPT_IMAGE_CROP_LABEL_EN NULL
#define BEETLE_OPT_IMAGE_CROP_INFO_0_EN NULL
#define OPTION_VAL_9PX_EN NULL
#define OPTION_VAL_10PX_EN NULL
#define OPTION_VAL_11PX_EN NULL
#define OPTION_VAL_12PX_EN NULL
#define OPTION_VAL_13PX_EN NULL
#define OPTION_VAL_14PX_EN NULL
#define OPTION_VAL_15PX_EN NULL
#define OPTION_VAL_16PX_EN NULL
#define OPTION_VAL_17PX_EN NULL
#define OPTION_VAL_18PX_EN NULL
#define OPTION_VAL_19PX_EN NULL
#define OPTION_VAL_20PX_EN NULL
#define BEETLE_OPT_IMAGE_OFFSET_LABEL_EN NULL
#define BEETLE_OPT_IMAGE_OFFSET_INFO_0_EN NULL
#define OPTION_VAL_12PX_O64_EN NULL
#define OPTION_VAL_11PX_O64_EN NULL
#define OPTION_VAL_10PX_O64_EN NULL
#define OPTION_VAL_9PX_O64_EN NULL
#define OPTION_VAL_8PX_O64_EN NULL
#define OPTION_VAL_7PX_O64_EN NULL
#define OPTION_VAL_6PX_O64_EN NULL
#define OPTION_VAL_5PX_O64_EN NULL
#define OPTION_VAL_4PX_O64_EN NULL
#define OPTION_VAL_3PX_O64_EN NULL
#define OPTION_VAL_2PX_O64_EN NULL
#define OPTION_VAL_1PX_O64_EN NULL
#define OPTION_VAL_1PX_O64_0_EN NULL
#define OPTION_VAL_2PX_O64_0_EN NULL
#define OPTION_VAL_3PX_O64_0_EN NULL
#define OPTION_VAL_4PX_O64_0_EN NULL
#define OPTION_VAL_5PX_O64_0_EN NULL
#define OPTION_VAL_6PX_O64_0_EN NULL
#define OPTION_VAL_7PX_O64_0_EN NULL
#define OPTION_VAL_8PX_O64_0_EN NULL
#define OPTION_VAL_9PX_O64_0_EN NULL
#define OPTION_VAL_10PX_O64_0_EN NULL
#define OPTION_VAL_11PX_O64_0_EN NULL
#define OPTION_VAL_12PX_O64_0_EN NULL
#define BEETLE_OPT_IMAGE_OFFSET_CYCLES_LABEL_EN NULL
#define BEETLE_OPT_IMAGE_OFFSET_CYCLES_INFO_0_EN NULL
#define BEETLE_OPT_INITIAL_SCANLINE_LABEL_EN NULL
#define BEETLE_OPT_INITIAL_SCANLINE_INFO_0_EN NULL
#define BEETLE_OPT_LAST_SCANLINE_LABEL_EN NULL
#define BEETLE_OPT_LAST_SCANLINE_INFO_0_EN NULL
#define OPTION_VAL_239_EN NULL
#define BEETLE_OPT_INITIAL_SCANLINE_PAL_LABEL_EN NULL
#define BEETLE_OPT_INITIAL_SCANLINE_PAL_INFO_0_EN NULL
#define BEETLE_OPT_LAST_SCANLINE_PAL_LABEL_EN NULL
#define BEETLE_OPT_LAST_SCANLINE_PAL_INFO_0_EN NULL
#define OPTION_VAL_287_EN NULL

struct retro_core_option_v2_category option_cats_en[] = {
   {
      "system",
      CATEGORY_SYSTEM_LABEL_EN,
      CATEGORY_SYSTEM_INFO_0_EN
   },
   {
      "video",
      CATEGORY_VIDEO_LABEL_EN,
      CATEGORY_VIDEO_INFO_0_EN
   },
   {
      "pgxp",
      CATEGORY_PGXP_LABEL_EN,
      CATEGORY_PGXP_INFO_0_EN
   },
   {
      "osd",
      CATEGORY_OSD_LABEL_EN,
      CATEGORY_OSD_INFO_0_EN
   },
   {
      "input",
      CATEGORY_INPUT_LABEL_EN,
      CATEGORY_INPUT_INFO_0_EN
   },
   {
      "memcards",
      CATEGORY_MEMCARDS_LABEL_EN,
      CATEGORY_MEMCARDS_INFO_0_EN
   },
   {
      "hacks",
      CATEGORY_HACKS_LABEL_EN,
      CATEGORY_HACKS_INFO_0_EN
   },
   { NULL, NULL, NULL },
};
struct retro_core_option_v2_definition option_defs_en[] = {
   {
      BEETLE_OPT(cpu_freq_scale),
      BEETLE_OPT_CPU_FREQ_SCALE_LABEL_EN,
      NULL,
      BEETLE_OPT_CPU_FREQ_SCALE_INFO_0_EN,
      NULL,
      "system",
      {
         { "50%",           OPTION_VAL_50_EN },
         { "60%",           OPTION_VAL_60_EN },
         { "70%",           OPTION_VAL_70_EN },
         { "80%",           OPTION_VAL_80_EN },
         { "90%",           OPTION_VAL_90_EN },
         { "100%",          OPTION_VAL_100_EN },
         { "110%",          OPTION_VAL_110_EN },
         { "120%",          OPTION_VAL_120_EN },
         { "130%",          OPTION_VAL_130_EN },
         { "140%",          OPTION_VAL_140_EN },
         { "150%",          OPTION_VAL_150_EN },
         { "160%",          OPTION_VAL_160_EN },
         { "170%",          OPTION_VAL_170_EN },
         { "180%",          OPTION_VAL_180_EN },
         { "190%",          OPTION_VAL_190_EN },
         { "200%",          OPTION_VAL_200_EN },
         { "210%",          OPTION_VAL_210_EN },
         { "220%",          OPTION_VAL_220_EN },
         { "230%",          OPTION_VAL_230_EN },
         { "240%",          OPTION_VAL_240_EN },
         { "250%",          OPTION_VAL_250_EN },
         { "260%",          OPTION_VAL_260_EN },
         { "270%",          OPTION_VAL_270_EN },
         { "280%",          OPTION_VAL_280_EN },
         { "290%",          OPTION_VAL_290_EN },
         { "300%",          OPTION_VAL_300_EN },
         { "310%",          OPTION_VAL_310_EN },
         { "320%",          OPTION_VAL_320_EN },
         { "330%",          OPTION_VAL_330_EN },
         { "340%",          OPTION_VAL_340_EN },
         { "350%",          OPTION_VAL_350_EN },
         { "360%",          OPTION_VAL_360_EN },
         { "370%",          OPTION_VAL_370_EN },
         { "380%",          OPTION_VAL_380_EN },
         { "390%",          OPTION_VAL_390_EN },
         { "400%",          OPTION_VAL_400_EN },
         { "410%",          OPTION_VAL_410_EN },
         { "420%",          OPTION_VAL_420_EN },
         { "430%",          OPTION_VAL_430_EN },
         { "440%",          OPTION_VAL_440_EN },
         { "450%",          OPTION_VAL_450_EN },
         { "460%",          OPTION_VAL_460_EN },
         { "470%",          OPTION_VAL_470_EN },
         { "480%",          OPTION_VAL_480_EN },
         { "490%",          OPTION_VAL_490_EN },
         { "500%",          OPTION_VAL_500_EN },
         { "510%",          OPTION_VAL_510_EN },
         { "520%",          OPTION_VAL_520_EN },
         { "530%",          OPTION_VAL_530_EN },
         { "540%",          OPTION_VAL_540_EN },
         { "550%",          OPTION_VAL_550_EN },
         { "560%",          OPTION_VAL_560_EN },
         { "570%",          OPTION_VAL_570_EN },
         { "580%",          OPTION_VAL_580_EN },
         { "590%",          OPTION_VAL_590_EN },
         { "600%",          OPTION_VAL_600_EN },
         { "610%",          OPTION_VAL_610_EN },
         { "620%",          OPTION_VAL_620_EN },
         { "630%",          OPTION_VAL_630_EN },
         { "640%",          OPTION_VAL_640_EN },
         { "650%",          OPTION_VAL_650_EN },
         { "660%",          OPTION_VAL_660_EN },
         { "670%",          OPTION_VAL_670_EN },
         { "680%",          OPTION_VAL_680_EN },
         { "690%",          OPTION_VAL_690_EN },
         { "700%",          OPTION_VAL_700_EN },
         { "710%",          OPTION_VAL_710_EN },
         { "720%",          OPTION_VAL_720_EN },
         { "730%",          OPTION_VAL_730_EN },
         { "740%",          OPTION_VAL_740_EN },
         { "750%",          OPTION_VAL_750_EN },
         { NULL, NULL },
      },
      "100%"
   },
   {
      BEETLE_OPT(gpu_overclock),
      BEETLE_OPT_GPU_OVERCLOCK_LABEL_EN,
      NULL,
      BEETLE_OPT_GPU_OVERCLOCK_INFO_0_EN,
      NULL,
      "system",
      {
         { "1x(native)", OPTION_VAL_1X_NATIVE_EN },
         { "2x",         OPTION_VAL_2X_EN },
         { "4x",         OPTION_VAL_4X_EN },
         { "8x",         OPTION_VAL_8X_EN },
         { "16x",        OPTION_VAL_16X_EN },
         { "32x",        OPTION_VAL_32X_EN },
         { NULL, NULL },
      },
      "1x(native)"
   },
   {
      BEETLE_OPT(gte_overclock),
      BEETLE_OPT_GTE_OVERCLOCK_LABEL_EN,
      NULL,
      BEETLE_OPT_GTE_OVERCLOCK_INFO_0_EN,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
#ifndef EMSCRIPTEN
   {
      BEETLE_OPT(cd_access_method),
      BEETLE_OPT_CD_ACCESS_METHOD_LABEL_EN,
      NULL,
      BEETLE_OPT_CD_ACCESS_METHOD_INFO_0_EN,
      NULL,
      "system",
      {
         { "sync",     OPTION_VAL_SYNC_EN },
         { "async",    OPTION_VAL_ASYNC_EN },
         { "precache", OPTION_VAL_PRECACHE_EN },
         { NULL, NULL },
      },
      "sync"
   },
#endif
   {
      BEETLE_OPT(cd_fastload),
      BEETLE_OPT_CD_FASTLOAD_LABEL_EN,
      NULL,
      BEETLE_OPT_CD_FASTLOAD_INFO_0_EN,
      NULL,
      "system",
      {
         { "2x(native)", OPTION_VAL_2X_NATIVE_EN },
         { "4x",          OPTION_VAL_4X_EN },
         { "6x",          OPTION_VAL_6X_EN },
         { "8x",          OPTION_VAL_8X_EN },
         { "10x",         OPTION_VAL_10X_EN },
         { "12x",         OPTION_VAL_12X_EN },
         { "14x",         OPTION_VAL_14X_EN },
         { NULL, NULL },
      },
      "2x(native)"
   },
   {
      BEETLE_OPT(skip_bios),
      BEETLE_OPT_SKIP_BIOS_LABEL_EN,
      NULL,
      BEETLE_OPT_SKIP_BIOS_INFO_0_EN,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(override_bios),
      BEETLE_OPT_OVERRIDE_BIOS_LABEL_EN,
      NULL,
      BEETLE_OPT_OVERRIDE_BIOS_INFO_0_EN,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "psxonpsp", OPTION_VAL_PSXONPSP_EN },
         { "ps1_rom",  OPTION_VAL_PS1_ROM_EN },
         { "openbios", OPTION_VAL_OPENBIOS_EN     },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(region),
      BEETLE_OPT_REGION_LABEL_EN,
      NULL,
      BEETLE_OPT_REGION_INFO_0_EN,
      NULL,
      "system",
      {
         { "auto",   OPTION_VAL_AUTO_EN },
         { "ntsc-j", OPTION_VAL_NTSC_J_EN },
         { "ntsc-u", OPTION_VAL_NTSC_U_EN },
         { "pal",    OPTION_VAL_PAL_EN },
         { NULL, NULL },
      },
      "auto"
   },
   {
      BEETLE_OPT(internal_resolution),
      BEETLE_OPT_INTERNAL_RESOLUTION_LABEL_EN,
      NULL,
      BEETLE_OPT_INTERNAL_RESOLUTION_INFO_0_EN,
      NULL,
      "video",
      {
         { "1x(native)", OPTION_VAL_1X_NATIVE_EN },
         { "2x",         OPTION_VAL_2X_EN },
         { "4x",         OPTION_VAL_4X_EN },
         { "8x",         OPTION_VAL_8X_EN },
         { "16x",        OPTION_VAL_16X_EN },
         { NULL, NULL },
      },
      "1x(native)"
   },
#if defined(HAVE_OPENGL) || defined(HAVE_OPENGLES) || defined(HAVE_VULKAN)
   {
      BEETLE_OPT(renderer),
      BEETLE_OPT_RENDERER_LABEL_EN,
      NULL,
      BEETLE_OPT_RENDERER_INFO_0_EN,
      NULL,
      "video",
      {
         { "hardware",    OPTION_VAL_HARDWARE_EN },
#if defined(HAVE_OPENGL) || defined(HAVE_OPENGLES)
         { "hardware_gl", OPTION_VAL_HARDWARE_GL_EN },
#endif
#if defined(HAVE_VULKAN)
         { "hardware_vk", OPTION_VAL_HARDWARE_VK_EN },
#endif
         { "software",    OPTION_VAL_SOFTWARE_EN },
         { NULL, NULL },
      },
      "hardware"
   },
   {
      BEETLE_OPT(renderer_software_fb),
      BEETLE_OPT_RENDERER_SOFTWARE_FB_LABEL_EN,
      NULL,
      BEETLE_OPT_RENDERER_SOFTWARE_FB_INFO_0_EN,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled"
   },
#endif
#if defined(HAVE_OPENGL) || defined(HAVE_OPENGLES)
   {
      BEETLE_OPT(depth),
      BEETLE_OPT_DEPTH_LABEL_EN,
      NULL,
      BEETLE_OPT_DEPTH_INFO_0_EN,
      NULL,
      "video",
      {
         { "16bpp(native)", OPTION_VAL_16BPP_NATIVE_EN },
         { "32bpp",         OPTION_VAL_32BPP_EN },
         { NULL, NULL },
      },
      "16bpp(native)"
   },
   // Sort of, it's more like 15-bit high color and 24-bit true color for visible output. The alpha channel is used for mask bit. Vulkan renderer uses ABGR1555_555 for 31 bits internal? FMVs are always 24-bit on all renderers like original hardware (BGR888, no alpha)
#endif
   {
      BEETLE_OPT(dither_mode),
      BEETLE_OPT_DITHER_MODE_LABEL_EN,
      NULL,
      BEETLE_OPT_DITHER_MODE_INFO_0_EN,
      NULL,
      "video",
      {
         { "1x(native)",          OPTION_VAL_1X_NATIVE_EN },
         { "internal resolution", OPTION_VAL_INTERNAL_RESOLUTION_EN },
         { "disabled",            NULL },
         { NULL, NULL },
      },
      "1x(native)"
   },
#ifdef HAVE_VULKAN
   {
      BEETLE_OPT(scaled_uv_offset),
      BEETLE_OPT_SCALED_UV_OFFSET_LABEL_EN,
      NULL,
      BEETLE_OPT_SCALED_UV_OFFSET_INFO_0_EN,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled"
   },
#endif
#if defined(HAVE_OPENGL) || defined(HAVE_OPENGLES) || defined(HAVE_VULKAN)
   {
      BEETLE_OPT(filter),
      BEETLE_OPT_FILTER_LABEL_EN,
      NULL,
      BEETLE_OPT_FILTER_INFO_0_EN,
      NULL,
      "video",
      {
         { "nearest",  OPTION_VAL_NEAREST_EN },
         { "SABR",     OPTION_VAL_SABR_EN },
         { "xBR",      OPTION_VAL_XBR_EN },
         { "bilinear", OPTION_VAL_BILINEAR_EN },
         { "3-point",  OPTION_VAL_3_POINT_EN },
         { "JINC2",    OPTION_VAL_JINC2_EN },
         { NULL, NULL },
      },
      "nearest"
   },
#ifdef HAVE_VULKAN
   {
      BEETLE_OPT(filter_exclude_sprite),
      BEETLE_OPT_FILTER_EXCLUDE_SPRITE_LABEL_EN,
      NULL,
      BEETLE_OPT_FILTER_EXCLUDE_SPRITE_INFO_0_EN,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "opaque", OPTION_VAL_OPAQUE_EN },
         { "all", OPTION_VAL_ALL_EN },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(filter_exclude_2d_polygon),
      BEETLE_OPT_FILTER_EXCLUDE_2D_POLYGON_LABEL_EN,
      NULL,
      BEETLE_OPT_FILTER_EXCLUDE_2D_POLYGON_INFO_0_EN,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "opaque", OPTION_VAL_OPAQUE_EN },
         { "all", OPTION_VAL_ALL_EN },
         { NULL, NULL },
      },
      "disabled"
   },
#endif
#endif
#ifdef HAVE_VULKAN
   {
      BEETLE_OPT(adaptive_smoothing),
      BEETLE_OPT_ADAPTIVE_SMOOTHING_LABEL_EN,
      NULL,
      BEETLE_OPT_ADAPTIVE_SMOOTHING_INFO_0_EN,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(super_sampling),
      BEETLE_OPT_SUPER_SAMPLING_LABEL_EN,
      NULL,
      BEETLE_OPT_SUPER_SAMPLING_INFO_0_EN,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(msaa),
      BEETLE_OPT_MSAA_LABEL_EN,
      NULL,
      BEETLE_OPT_MSAA_INFO_0_EN,
      NULL,
      "video",
      {
         { "1x",  OPTION_VAL_1X_EN },
         { "2x",  OPTION_VAL_2X_EN },
         { "4x",  OPTION_VAL_4X_EN },
         { "8x",  OPTION_VAL_8X_EN },
         { "16x", OPTION_VAL_16X_EN },
         { NULL, NULL },
      },
      "1x"
   },
   {
      BEETLE_OPT(mdec_yuv),
      BEETLE_OPT_MDEC_YUV_LABEL_EN,
      NULL,
      BEETLE_OPT_MDEC_YUV_INFO_0_EN,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(track_textures),
      BEETLE_OPT_TRACK_TEXTURES_LABEL_EN,
      NULL,
      BEETLE_OPT_TRACK_TEXTURES_INFO_0_EN,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
#ifdef TEXTURE_DUMPING_ENABLED
   {
      BEETLE_OPT(dump_textures),
      BEETLE_OPT_DUMP_TEXTURES_LABEL_EN,
      NULL,
      BEETLE_OPT_DUMP_TEXTURES_INFO_0_EN,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
#endif
   {
      BEETLE_OPT(replace_textures),
      BEETLE_OPT_REPLACE_TEXTURES_LABEL_EN,
      NULL,
      BEETLE_OPT_REPLACE_TEXTURES_INFO_0_EN,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
#endif
   {
      BEETLE_OPT(pgxp_mode),
      BEETLE_OPT_PGXP_MODE_LABEL_EN,
      NULL,
      BEETLE_OPT_PGXP_MODE_INFO_0_EN,
      NULL,
      "pgxp",
      {
         { "disabled",     NULL },
         { "memory only",  OPTION_VAL_MEMORY_ONLY_EN },
         { "memory + CPU", OPTION_VAL_MEMORY_CPU_EN },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(pgxp_2d_tol),
      BEETLE_OPT_PGXP_2D_TOL_LABEL_EN,
      NULL,
      BEETLE_OPT_PGXP_2D_TOL_INFO_0_EN,
      NULL,
      "pgxp",
      {
         { "disabled", NULL },
         { "0px", OPTION_VAL_0PX_EN },
         { "1px", OPTION_VAL_1PX_EN },
         { "2px", OPTION_VAL_2PX_EN },
         { "3px", OPTION_VAL_3PX_EN },
         { "4px", OPTION_VAL_4PX_EN },
         { "5px", OPTION_VAL_5PX_EN },
         { "6px", OPTION_VAL_6PX_EN },
         { "7px", OPTION_VAL_7PX_EN },
         { "8px", OPTION_VAL_8PX_EN },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(pgxp_nclip),
      BEETLE_OPT_PGXP_NCLIP_LABEL_EN,
      NULL,
      BEETLE_OPT_PGXP_NCLIP_INFO_0_EN,
      NULL,
      "pgxp",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
#if defined(HAVE_OPENGL) || defined(HAVE_OPENGLES) || defined(HAVE_VULKAN)
   {
      BEETLE_OPT(pgxp_vertex),
      BEETLE_OPT_PGXP_VERTEX_LABEL_EN,
      NULL,
      BEETLE_OPT_PGXP_VERTEX_INFO_0_EN,
      NULL,
      "pgxp",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
#endif
   {
      BEETLE_OPT(pgxp_texture),
      BEETLE_OPT_PGXP_TEXTURE_LABEL_EN,
      NULL,
      BEETLE_OPT_PGXP_TEXTURE_INFO_0_EN,
      NULL,
      "pgxp",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(display_internal_fps),
      BEETLE_OPT_DISPLAY_INTERNAL_FPS_LABEL_EN,
      NULL,
      BEETLE_OPT_DISPLAY_INTERNAL_FPS_INFO_0_EN,
      NULL,
      "osd",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(display_osd),
      BEETLE_OPT_DISPLAY_OSD_LABEL_EN,
      NULL,
      BEETLE_OPT_DISPLAY_OSD_INFO_0_EN,
      NULL,
      "osd",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled"
   },
#if defined(HAVE_OPENGL) || defined(HAVE_OPENGLES) || defined(HAVE_VULKAN)
   {
      BEETLE_OPT(display_vram),
      BEETLE_OPT_DISPLAY_VRAM_LABEL_EN,
      NULL,
      BEETLE_OPT_DISPLAY_VRAM_INFO_0_EN,
      NULL,
      "osd",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
#endif
   {
      BEETLE_OPT(analog_calibration),
      BEETLE_OPT_ANALOG_CALIBRATION_LABEL_EN,
      NULL,
      BEETLE_OPT_ANALOG_CALIBRATION_INFO_0_EN,
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(analog_toggle),
      BEETLE_OPT_ANALOG_TOGGLE_LABEL_EN,
      NULL,
      BEETLE_OPT_ANALOG_TOGGLE_INFO_0_EN,
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { "enabled-analog", OPTION_VAL_ENABLED_ANALOG_EN },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(analog_toggle_combo),
      BEETLE_OPT_ANALOG_TOGGLE_COMBO_LABEL_EN,
      NULL,
      BEETLE_OPT_ANALOG_TOGGLE_COMBO_INFO_0_EN,
      NULL,
      "input",
      {
         { "l1+l2+r1+r2+start+select", OPTION_VAL_L1_L2_R1_R2_START_SELECT_EN },
         { "l1+r1+select",             OPTION_VAL_L1_R1_SELECT_EN },
         { "l1+r1+start",              OPTION_VAL_L1_R1_START_EN },
         { "l1+r1+l3",                 OPTION_VAL_L1_R1_L3_EN },
         { "l1+r1+r3",                 OPTION_VAL_L1_R1_R3_EN },
         { "l2+r2+select",             OPTION_VAL_L2_R2_SELECT_EN },
         { "l2+r2+start",              OPTION_VAL_L2_R2_START_EN },
         { "l2+r2+l3",                 OPTION_VAL_L2_R2_L3_EN },
         { "l2+r2+r3",                 OPTION_VAL_L2_R2_R3_EN },
         { "l3+r3",                    OPTION_VAL_L3_R3_EN },
         { NULL, NULL },
      },
      "l1+r1+select"
   },
   {
      BEETLE_OPT(analog_toggle_hold),
      BEETLE_OPT_ANALOG_TOGGLE_HOLD_LABEL_EN,
      NULL,
      BEETLE_OPT_ANALOG_TOGGLE_HOLD_INFO_0_EN,
      NULL,
      "input",
      {
         { "0", OPTION_VAL_0_EN },
         { "1", OPTION_VAL_1_EN },
         { "2", OPTION_VAL_2_EN },
         { "3", OPTION_VAL_3_EN },
         { "4", OPTION_VAL_4_EN },
         { "5", OPTION_VAL_5_EN },
         { NULL, NULL },
      },
      "1"
   },
   {
      BEETLE_OPT(enable_multitap_port1),
      BEETLE_OPT_ENABLE_MULTITAP_PORT1_LABEL_EN,
      NULL,
      BEETLE_OPT_ENABLE_MULTITAP_PORT1_INFO_0_EN,
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(enable_multitap_port2),
      BEETLE_OPT_ENABLE_MULTITAP_PORT2_LABEL_EN,
      NULL,
      BEETLE_OPT_ENABLE_MULTITAP_PORT2_INFO_0_EN,
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(gun_input_mode),
      BEETLE_OPT_GUN_INPUT_MODE_LABEL_EN,
      NULL,
      BEETLE_OPT_GUN_INPUT_MODE_INFO_0_EN,
      NULL,
      "input",
      {
         { "lightgun",    OPTION_VAL_LIGHTGUN_EN },
         { "touchscreen", OPTION_VAL_TOUCHSCREEN_EN },
         { NULL, NULL },
      },
      "lightgun"
   },
   // Shouldn't the gun_input_mode just be Mouse vs. Touchscreen?
   {
      BEETLE_OPT(gun_cursor),
      BEETLE_OPT_GUN_CURSOR_LABEL_EN,
      NULL,
      BEETLE_OPT_GUN_CURSOR_INFO_0_EN,
      NULL,
      "input",
      {
         { "off",   "disabled" },
         { "cross", OPTION_VAL_CROSS_EN },
         { "dot",   OPTION_VAL_DOT_EN },
         { NULL, NULL },
      },
      "cross"
   },
   {
      BEETLE_OPT(crosshair_color_p1),
      BEETLE_OPT_CROSSHAIR_COLOR_P1_LABEL_EN,
      NULL,
      BEETLE_OPT_CROSSHAIR_COLOR_P1_INFO_0_EN,
      NULL,
      "input",
      {
         { "red",    OPTION_VAL_RED_EN },
         { "blue",   OPTION_VAL_BLUE_EN },
         { "green",  OPTION_VAL_GREEN_EN },
         { "orange", OPTION_VAL_ORANGE_EN },
         { "yellow", OPTION_VAL_YELLOW_EN },
         { "cyan",   OPTION_VAL_CYAN_EN },
         { "pink",   OPTION_VAL_PINK_EN },
         { "purple", OPTION_VAL_PURPLE_EN },
         { "black",  OPTION_VAL_BLACK_EN },
         { "white",  OPTION_VAL_WHITE_EN },
         { NULL, NULL },
      },
      "red"
   },
   {
      BEETLE_OPT(crosshair_color_p2),
      BEETLE_OPT_CROSSHAIR_COLOR_P2_LABEL_EN,
      NULL,
      BEETLE_OPT_CROSSHAIR_COLOR_P2_INFO_0_EN,
      NULL,
      "input",
      {
         { "blue",   OPTION_VAL_BLUE_EN },
         { "red",    OPTION_VAL_RED_EN },
         { "green",  OPTION_VAL_GREEN_EN },
         { "orange", OPTION_VAL_ORANGE_EN },
         { "yellow", OPTION_VAL_YELLOW_EN },
         { "cyan",   OPTION_VAL_CYAN_EN },
         { "pink",   OPTION_VAL_PINK_EN },
         { "purple", OPTION_VAL_PURPLE_EN },
         { "black",  OPTION_VAL_BLACK_EN },
         { "white",  OPTION_VAL_WHITE_EN },
         { NULL, NULL },
      },
      "blue"
   },
   {
      BEETLE_OPT(mouse_sensitivity),
      BEETLE_OPT_MOUSE_SENSITIVITY_LABEL_EN,
      NULL,
      BEETLE_OPT_MOUSE_SENSITIVITY_INFO_0_EN,
      NULL,
      "input",
      {
         { "5%",   OPTION_VAL_5_O42_EN },
         { "10%",  OPTION_VAL_10_EN },
         { "15%",  OPTION_VAL_15_EN },
         { "20%",  OPTION_VAL_20_EN },
         { "25%",  OPTION_VAL_25_EN },
         { "30%",  OPTION_VAL_30_EN },
         { "35%",  OPTION_VAL_35_EN },
         { "40%",  OPTION_VAL_40_EN },
         { "45%",  OPTION_VAL_45_EN },
         { "50%",  OPTION_VAL_50_EN },
         { "55%",  OPTION_VAL_55_EN },
         { "60%",  OPTION_VAL_60_EN },
         { "65%",  OPTION_VAL_65_EN },
         { "70%",  OPTION_VAL_70_EN },
         { "75%",  OPTION_VAL_75_EN },
         { "80%",  OPTION_VAL_80_EN },
         { "85%",  OPTION_VAL_85_EN },
         { "90%",  OPTION_VAL_90_EN },
         { "95%",  OPTION_VAL_95_EN },
         { "100%", OPTION_VAL_100_O42_EN },
         { "105%", OPTION_VAL_105_EN },
         { "110%", OPTION_VAL_110_EN },
         { "115%", OPTION_VAL_115_EN },
         { "120%", OPTION_VAL_120_EN },
         { "125%", OPTION_VAL_125_EN },
         { "130%", OPTION_VAL_130_EN },
         { "135%", OPTION_VAL_135_EN },
         { "140%", OPTION_VAL_140_EN },
         { "145%", OPTION_VAL_145_EN },
         { "150%", OPTION_VAL_150_EN },
         { "155%", OPTION_VAL_155_EN },
         { "160%", OPTION_VAL_160_EN },
         { "165%", OPTION_VAL_165_EN },
         { "170%", OPTION_VAL_170_EN },
         { "175%", OPTION_VAL_175_EN },
         { "180%", OPTION_VAL_180_EN },
         { "185%", OPTION_VAL_185_EN },
         { "190%", OPTION_VAL_190_EN },
         { "195%", OPTION_VAL_195_EN },
         { "200%", OPTION_VAL_200_EN },
         { NULL, NULL },
      },
      "100%"
   },
   {
      BEETLE_OPT(negcon_response),
      BEETLE_OPT_NEGCON_RESPONSE_LABEL_EN,
      NULL,
      BEETLE_OPT_NEGCON_RESPONSE_INFO_0_EN,
      NULL,
      "input",
      {
         { "linear",    OPTION_VAL_LINEAR_EN },
         { "quadratic", OPTION_VAL_QUADRATIC_EN },
         { "cubic",     OPTION_VAL_CUBIC_EN },
         { NULL, NULL },
      },
      "linear"
   },
   {
      BEETLE_OPT(negcon_deadzone),
      BEETLE_OPT_NEGCON_DEADZONE_LABEL_EN,
      NULL,
      BEETLE_OPT_NEGCON_DEADZONE_INFO_0_EN,
      NULL,
      "input",
      {
         { "0%",  OPTION_VAL_0_O44_EN },
         { "5%",  OPTION_VAL_5_O42_EN },
         { "10%", OPTION_VAL_10_EN },
         { "15%", OPTION_VAL_15_EN },
         { "20%", OPTION_VAL_20_EN },
         { "25%", OPTION_VAL_25_EN },
         { "30%", OPTION_VAL_30_EN },
         { NULL, NULL },
      },
      "0%"
   },
   {
      BEETLE_OPT(use_mednafen_memcard0_method),
      BEETLE_OPT_USE_MEDNAFEN_MEMCARD0_METHOD_LABEL_EN,
      NULL,
      BEETLE_OPT_USE_MEDNAFEN_MEMCARD0_METHOD_INFO_0_EN,
      NULL,
      "memcards",
      {
         { "libretro", OPTION_VAL_LIBRETRO_EN },
         { "mednafen", OPTION_VAL_MEDNAFEN_EN },
         { NULL, NULL },
      },
      "libretro"
   },
   {
      BEETLE_OPT(enable_memcard1),
      BEETLE_OPT_ENABLE_MEMCARD1_LABEL_EN,
      NULL,
      BEETLE_OPT_ENABLE_MEMCARD1_INFO_0_EN,
      NULL,
      "memcards",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(shared_memory_cards),
      BEETLE_OPT_SHARED_MEMORY_CARDS_LABEL_EN,
      NULL,
      BEETLE_OPT_SHARED_MEMORY_CARDS_INFO_0_EN,
      NULL,
      "memcards",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(memcard_left_index),
      BEETLE_OPT_MEMCARD_LEFT_INDEX_LABEL_EN,
      NULL,
      BEETLE_OPT_MEMCARD_LEFT_INDEX_INFO_0_EN,
      NULL,
      "memcards",
      {
         { "0",  OPTION_VAL_0_O48_EN },
         { "1",  NULL },
         { "2",  NULL },
         { "3",  NULL },
         { "4",  NULL },
         { "5",  NULL },
         { "6",  NULL },
         { "7",  NULL },
         { "8",  NULL },
         { "9",  NULL },
         { "10", NULL },
         { "11", NULL },
         { "12", NULL },
         { "13", NULL },
         { "14", NULL },
         { "15", NULL },
         { "16", NULL },
         { "17", NULL },
         { "18", NULL },
         { "19", NULL },
         { "20", NULL },
         { "21", NULL },
         { "22", NULL },
         { "23", NULL },
         { "24", NULL },
         { "25", NULL },
         { "26", NULL },
         { "27", NULL },
         { "28", NULL },
         { "29", NULL },
         { "30", NULL },
         { "31", NULL },
         { "32", NULL },
         { "33", NULL },
         { "34", NULL },
         { "35", NULL },
         { "36", NULL },
         { "37", NULL },
         { "38", NULL },
         { "39", NULL },
         { "40", NULL },
         { "41", NULL },
         { "42", NULL },
         { "43", NULL },
         { "44", NULL },
         { "45", NULL },
         { "46", NULL },
         { "47", NULL },
         { "48", NULL },
         { "49", NULL },
         { "50", NULL },
         { "51", NULL },
         { "52", NULL },
         { "53", NULL },
         { "54", NULL },
         { "55", NULL },
         { "56", NULL },
         { "57", NULL },
         { "58", NULL },
         { "59", NULL },
         { "60", NULL },
         { "61", NULL },
         { "62", NULL },
         { "63", NULL },
         { NULL, NULL },
      },
      "0"
   },
   {
      BEETLE_OPT(memcard_right_index),
      BEETLE_OPT_MEMCARD_RIGHT_INDEX_LABEL_EN,
      NULL,
      BEETLE_OPT_MEMCARD_RIGHT_INDEX_INFO_0_EN,
      NULL,
      "memcards",
      {
         { "0",  NULL },
         { "1",  OPTION_VAL_1_O49_EN },
         { "2",  NULL },
         { "3",  NULL },
         { "4",  NULL },
         { "5",  NULL },
         { "6",  NULL },
         { "7",  NULL },
         { "8",  NULL },
         { "9",  NULL },
         { "10", NULL },
         { "11", NULL },
         { "12", NULL },
         { "13", NULL },
         { "14", NULL },
         { "15", NULL },
         { "16", NULL },
         { "17", NULL },
         { "18", NULL },
         { "19", NULL },
         { "20", NULL },
         { "21", NULL },
         { "22", NULL },
         { "23", NULL },
         { "24", NULL },
         { "25", NULL },
         { "26", NULL },
         { "27", NULL },
         { "28", NULL },
         { "29", NULL },
         { "30", NULL },
         { "31", NULL },
         { "32", NULL },
         { "33", NULL },
         { "34", NULL },
         { "35", NULL },
         { "36", NULL },
         { "37", NULL },
         { "38", NULL },
         { "39", NULL },
         { "40", NULL },
         { "41", NULL },
         { "42", NULL },
         { "43", NULL },
         { "44", NULL },
         { "45", NULL },
         { "46", NULL },
         { "47", NULL },
         { "48", NULL },
         { "49", NULL },
         { "50", NULL },
         { "51", NULL },
         { "52", NULL },
         { "53", NULL },
         { "54", NULL },
         { "55", NULL },
         { "56", NULL },
         { "57", NULL },
         { "58", NULL },
         { "59", NULL },
         { "60", NULL },
         { "61", NULL },
         { "62", NULL },
         { "63", NULL },
         { NULL, NULL },
      },
      "1"
   },
   {
      BEETLE_OPT(line_render),
      BEETLE_OPT_LINE_RENDER_LABEL_EN,
      NULL,
      BEETLE_OPT_LINE_RENDER_INFO_0_EN,
      NULL,
      "hacks",
      {
         { "default",    OPTION_VAL_DEFAULT_EN },
         { "aggressive", OPTION_VAL_AGGRESSIVE_EN },
         { "disabled",   NULL },
         { NULL, NULL },
      },
      "default"
   },
   {
      BEETLE_OPT(widescreen_hack),
      BEETLE_OPT_WIDESCREEN_HACK_LABEL_EN,
      NULL,
      BEETLE_OPT_WIDESCREEN_HACK_INFO_0_EN,
      NULL,
      "hacks",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(widescreen_hack_aspect_ratio),
      BEETLE_OPT_WIDESCREEN_HACK_ASPECT_RATIO_LABEL_EN,
      NULL,
      BEETLE_OPT_WIDESCREEN_HACK_ASPECT_RATIO_INFO_0_EN,
      NULL,
      "hacks",
      {
         { "16:9",  OPTION_VAL_16_9_EN },
         { "16:10", OPTION_VAL_16_10_EN },
         { "18:9",  OPTION_VAL_18_9_EN },
         { "19:9",  OPTION_VAL_19_9_EN },
         { "20:9",  OPTION_VAL_20_9_EN },
         { "21:9",  OPTION_VAL_21_9_EN }, // 64:27
         { "32:9",  OPTION_VAL_32_9_EN },
         { NULL,    NULL },
      },
      "16:9"
   },
#if defined(HAVE_LIGHTREC)
   {
      BEETLE_OPT(cpu_dynarec),
      BEETLE_OPT_CPU_DYNAREC_LABEL_EN,
      NULL,
      BEETLE_OPT_CPU_DYNAREC_INFO_0_EN,
      NULL,
      "hacks",
      {
         { "disabled",        OPTION_VAL_DISABLED_EN },
         { "execute",         OPTION_VAL_EXECUTE_EN },
         { "execute_one",     OPTION_VAL_EXECUTE_ONE_EN },
         { "run_interpreter", OPTION_VAL_RUN_INTERPRETER_EN },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(dynarec_invalidate),
      BEETLE_OPT_DYNAREC_INVALIDATE_LABEL_EN,
      NULL,
      BEETLE_OPT_DYNAREC_INVALIDATE_INFO_0_EN,
      NULL,
      "hacks",
      {
         { "full", OPTION_VAL_FULL_EN },
         { "dma",  OPTION_VAL_DMA_EN },
         { NULL, NULL },
      },
      "full"
   },
   {
      BEETLE_OPT(dynarec_eventcycles),
      BEETLE_OPT_DYNAREC_EVENTCYCLES_LABEL_EN,
      NULL,
      BEETLE_OPT_DYNAREC_EVENTCYCLES_INFO_0_EN,
      NULL,
      "hacks",
      {
         { "128",  OPTION_VAL_128_EN },
         { "256",  NULL },
         { "384",  NULL },
         { "512",  NULL },
         { "640",  NULL },
         { "768",  NULL },
         { "896",  NULL },
         { "1024", NULL },
         { "1152", NULL },
         { "1280", NULL },
         { "1408", NULL },
         { "1536", NULL },
         { "1664", NULL },
         { "1792", NULL },
         { "1920", NULL },
         { "2048", NULL },
         { NULL, NULL },
      },
      "128"
   },
   {
      BEETLE_OPT(dynarec_spu_samples),
      BEETLE_OPT_DYNAREC_SPU_SAMPLES_LABEL_EN,
      NULL,
      BEETLE_OPT_DYNAREC_SPU_SAMPLES_INFO_0_EN,
      NULL,
      "hacks",
      {
         { "1",  OPTION_VAL_1_O49_EN },
         { "4",  NULL },
         { "16", NULL },
         { NULL, NULL },
      },
      "1"
   },
#endif
   {
      BEETLE_OPT(core_timing_fps),
      BEETLE_OPT_CORE_TIMING_FPS_LABEL_EN,
      NULL,
      BEETLE_OPT_CORE_TIMING_FPS_INFO_0_EN,
      NULL,
      "video",
      {
         { "auto_toggle",       OPTION_VAL_AUTO_TOGGLE_EN },
         { "force_progressive", OPTION_VAL_FORCE_PROGRESSIVE_EN },
         { "force_interlaced",  OPTION_VAL_FORCE_INTERLACED_EN },
      },
      "auto_toggle"
   },
   {
      BEETLE_OPT(deinterlacer),
      BEETLE_OPT_DEINTERLACER_LABEL_EN,
      NULL,
      BEETLE_OPT_DEINTERLACER_INFO_0_EN,
      NULL,
      "video",
      {
         { "weave",      OPTION_VAL_WEAVE_EN },
         { "bob",        OPTION_VAL_BOB_EN },
         { "bob_offset", OPTION_VAL_BOB_OFFSET_EN },
         { "fastmad",    OPTION_VAL_FASTMAD_EN },
         { "off",        "Off" },
         { NULL, NULL },
      },
      "weave"
   },
   {
      BEETLE_OPT(frame_duping),
      BEETLE_OPT_FRAME_DUPING_LABEL_EN,
      NULL,
      BEETLE_OPT_FRAME_DUPING_INFO_0_EN,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(pal_video_timing_override),
      BEETLE_OPT_PAL_VIDEO_TIMING_OVERRIDE_LABEL_EN,
      NULL,
      BEETLE_OPT_PAL_VIDEO_TIMING_OVERRIDE_INFO_0_EN,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(aspect_ratio),
      BEETLE_OPT_ASPECT_RATIO_LABEL_EN,
      NULL,
      BEETLE_OPT_ASPECT_RATIO_INFO_0_EN,
      NULL,
      "video",
      {
         { "corrected", OPTION_VAL_CORRECTED_EN },
         { "uncorrected", OPTION_VAL_UNCORRECTED_EN },
         { "4:3",  OPTION_VAL_4_3_EN },
         { "ntsc", OPTION_VAL_NTSC_EN },
      },
      "corrected"
   },
   {
      BEETLE_OPT(crop_overscan),
      BEETLE_OPT_CROP_OVERSCAN_LABEL_EN,
      NULL,
      BEETLE_OPT_CROP_OVERSCAN_INFO_0_EN,
      NULL,
      "video",
      {
         { "disabled", "Disabled" },
         { "static",   OPTION_VAL_STATIC_EN },
         { "smart",    OPTION_VAL_SMART_EN },
         { NULL, NULL },
      },
      "smart"
   },
   {
      BEETLE_OPT(image_crop),
      BEETLE_OPT_IMAGE_CROP_LABEL_EN,
      NULL,
      BEETLE_OPT_IMAGE_CROP_INFO_0_EN,
      NULL,
      "video",
      {
         { "disabled", "0" },
         { "1px",      OPTION_VAL_1PX_EN },
         { "2px",      OPTION_VAL_2PX_EN },
         { "3px",      OPTION_VAL_3PX_EN },
         { "4px",      OPTION_VAL_4PX_EN },
         { "5px",      OPTION_VAL_5PX_EN },
         { "6px",      OPTION_VAL_6PX_EN },
         { "7px",      OPTION_VAL_7PX_EN },
         { "8px",      OPTION_VAL_8PX_EN },
         { "9px",      OPTION_VAL_9PX_EN },
         { "10px",     OPTION_VAL_10PX_EN },
         { "11px",     OPTION_VAL_11PX_EN },
         { "12px",     OPTION_VAL_12PX_EN },
         { "13px",     OPTION_VAL_13PX_EN },
         { "14px",     OPTION_VAL_14PX_EN },
         { "15px",     OPTION_VAL_15PX_EN },
         { "16px",     OPTION_VAL_16PX_EN },
         { "17px",     OPTION_VAL_17PX_EN },
         { "18px",     OPTION_VAL_18PX_EN },
         { "19px",     OPTION_VAL_19PX_EN },
         { "20px",     OPTION_VAL_20PX_EN },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(image_offset),
      BEETLE_OPT_IMAGE_OFFSET_LABEL_EN,
      NULL,
      BEETLE_OPT_IMAGE_OFFSET_INFO_0_EN,
      NULL,
      "video",
      {
         { "-12px",    OPTION_VAL_12PX_O64_EN },
         { "-11px",    OPTION_VAL_11PX_O64_EN },
         { "-10px",    OPTION_VAL_10PX_O64_EN },
         { "-9px",     OPTION_VAL_9PX_O64_EN },
         { "-8px",     OPTION_VAL_8PX_O64_EN },
         { "-7px",     OPTION_VAL_7PX_O64_EN },
         { "-6px",     OPTION_VAL_6PX_O64_EN },
         { "-5px",     OPTION_VAL_5PX_O64_EN },
         { "-4px",     OPTION_VAL_4PX_O64_EN },
         { "-3px",     OPTION_VAL_3PX_O64_EN },
         { "-2px",     OPTION_VAL_2PX_O64_EN },
         { "-1px",     OPTION_VAL_1PX_O64_EN },
         { "disabled", "0" },
         { "+1px",     OPTION_VAL_1PX_O64_0_EN },
         { "+2px",     OPTION_VAL_2PX_O64_0_EN },
         { "+3px",     OPTION_VAL_3PX_O64_0_EN },
         { "+4px",     OPTION_VAL_4PX_O64_0_EN },
         { "+5px",     OPTION_VAL_5PX_O64_0_EN },
         { "+6px",     OPTION_VAL_6PX_O64_0_EN },
         { "+7px",     OPTION_VAL_7PX_O64_0_EN },
         { "+8px",     OPTION_VAL_8PX_O64_0_EN },
         { "+9px",     OPTION_VAL_9PX_O64_0_EN },
         { "+10px",    OPTION_VAL_10PX_O64_0_EN },
         { "+11px",    OPTION_VAL_11PX_O64_0_EN },
         { "+12px",    OPTION_VAL_12PX_O64_0_EN },
         { NULL, NULL },
      },
      "disabled"
   },
#if defined(HAVE_OPENGL) || defined(HAVE_OPENGLES) || defined(HAVE_VULKAN)
   {
      BEETLE_OPT(image_offset_cycles),
      BEETLE_OPT_IMAGE_OFFSET_CYCLES_LABEL_EN,
      NULL,
      BEETLE_OPT_IMAGE_OFFSET_CYCLES_INFO_0_EN,
      NULL,
      "video",
      {
         { "-40",      NULL },
         { "-39",      NULL },
         { "-38",      NULL },
         { "-37",      NULL },
         { "-36",      NULL },
         { "-35",      NULL },
         { "-34",      NULL },
         { "-33",      NULL },
         { "-32",      NULL },
         { "-31",      NULL },
         { "-30",      NULL },
         { "-29",      NULL },
         { "-28",      NULL },
         { "-27",      NULL },
         { "-26",      NULL },
         { "-25",      NULL },
         { "-24",      NULL },
         { "-23",      NULL },
         { "-22",      NULL },
         { "-21",      NULL },
         { "-20",      NULL },
         { "-19",      NULL },
         { "-18",      NULL },
         { "-17",      NULL },
         { "-16",      NULL },
         { "-15",      NULL },
         { "-14",      NULL },
         { "-13",      NULL },
         { "-12",      NULL },
         { "-11",      NULL },
         { "-10",      NULL },
         { "-9",       NULL },
         { "-8",       NULL },
         { "-7",       NULL },
         { "-6",       NULL },
         { "-5",       NULL },
         { "-4",       NULL },
         { "-3",       NULL },
         { "-2",       NULL },
         { "-1",       NULL },
         { "0",        "0" },
         { "+1",       NULL },
         { "+2",       NULL },
         { "+3",       NULL },
         { "+4",       NULL },
         { "+5",       NULL },
         { "+6",       NULL },
         { "+7",       NULL },
         { "+8",       NULL },
         { "+9",       NULL },
         { "+10",      NULL },
         { "+11",      NULL },
         { "+12",      NULL },
         { "+13",      NULL },
         { "+14",      NULL },
         { "+15",      NULL },
         { "+16",      NULL },
         { "+17",      NULL },
         { "+18",      NULL },
         { "+19",      NULL },
         { "+20",      NULL },
         { "+21",      NULL },
         { "+22",      NULL },
         { "+23",      NULL },
         { "+24",      NULL },
         { "+25",      NULL },
         { "+26",      NULL },
         { "+27",      NULL },
         { "+28",      NULL },
         { "+29",      NULL },
         { "+30",      NULL },
         { "+31",      NULL },
         { "+32",      NULL },
         { "+33",      NULL },
         { "+34",      NULL },
         { "+35",      NULL },
         { "+36",      NULL },
         { "+37",      NULL },
         { "+38",      NULL },
         { "+39",      NULL },
         { "+40",      NULL },
         { NULL, NULL},
      },
      "0"
   },
#endif
   {
      BEETLE_OPT(initial_scanline),
      BEETLE_OPT_INITIAL_SCANLINE_LABEL_EN,
      NULL,
      BEETLE_OPT_INITIAL_SCANLINE_INFO_0_EN,
      NULL,
      "video",
      {
         { "0",  OPTION_VAL_0_O48_EN },
         { "1",  NULL },
         { "2",  NULL },
         { "3",  NULL },
         { "4",  NULL },
         { "5",  NULL },
         { "6",  NULL },
         { "7",  NULL },
         { "8",  NULL },
         { "9",  NULL },
         { "10", NULL },
         { "11", NULL },
         { "12", NULL },
         { "13", NULL },
         { "14", NULL },
         { "15", NULL },
         { "16", NULL },
         { "17", NULL },
         { "18", NULL },
         { "19", NULL },
         { "20", NULL },
         { "21", NULL },
         { "22", NULL },
         { "23", NULL },
         { "24", NULL },
         { "25", NULL },
         { "26", NULL },
         { "27", NULL },
         { "28", NULL },
         { "29", NULL },
         { "30", NULL },
         { "31", NULL },
         { "32", NULL },
         { "33", NULL },
         { "34", NULL },
         { "35", NULL },
         { "36", NULL },
         { "37", NULL },
         { "38", NULL },
         { "39", NULL },
         { "40", NULL },
         { NULL, NULL },
      },
      "0"
   },
   {
      BEETLE_OPT(last_scanline),
      BEETLE_OPT_LAST_SCANLINE_LABEL_EN,
      NULL,
      BEETLE_OPT_LAST_SCANLINE_INFO_0_EN,
      NULL,
      "video",
      {
         { "210", NULL },
         { "211", NULL },
         { "212", NULL },
         { "213", NULL },
         { "214", NULL },
         { "215", NULL },
         { "216", NULL },
         { "217", NULL },
         { "218", NULL },
         { "219", NULL },
         { "220", NULL },
         { "221", NULL },
         { "222", NULL },
         { "223", NULL },
         { "224", NULL },
         { "225", NULL },
         { "226", NULL },
         { "227", NULL },
         { "228", NULL },
         { "229", NULL },
         { "230", NULL },
         { "231", NULL },
         { "232", NULL },
         { "233", NULL },
         { "234", NULL },
         { "235", NULL },
         { "236", NULL },
         { "237", NULL },
         { "238", NULL },
         { "239", OPTION_VAL_239_EN },
         { NULL, NULL },
      },
      "239"
   },
   {
      BEETLE_OPT(initial_scanline_pal),
      BEETLE_OPT_INITIAL_SCANLINE_PAL_LABEL_EN,
      NULL,
      BEETLE_OPT_INITIAL_SCANLINE_PAL_INFO_0_EN,
      NULL,
      "video",
      {
         { "0",  OPTION_VAL_0_O48_EN },
         { "1",  NULL },
         { "2",  NULL },
         { "3",  NULL },
         { "4",  NULL },
         { "5",  NULL },
         { "6",  NULL },
         { "7",  NULL },
         { "8",  NULL },
         { "9",  NULL },
         { "10", NULL },
         { "11", NULL },
         { "12", NULL },
         { "13", NULL },
         { "14", NULL },
         { "15", NULL },
         { "16", NULL },
         { "17", NULL },
         { "18", NULL },
         { "19", NULL },
         { "20", NULL },
         { "21", NULL },
         { "22", NULL },
         { "23", NULL },
         { "24", NULL },
         { "25", NULL },
         { "26", NULL },
         { "27", NULL },
         { "28", NULL },
         { "29", NULL },
         { "30", NULL },
         { "31", NULL },
         { "32", NULL },
         { "33", NULL },
         { "34", NULL },
         { "35", NULL },
         { "36", NULL },
         { "37", NULL },
         { "38", NULL },
         { "39", NULL },
         { "40", NULL },
         { NULL, NULL },
      },
      "0"
   },
   {
      BEETLE_OPT(last_scanline_pal),
      BEETLE_OPT_LAST_SCANLINE_PAL_LABEL_EN,
      NULL,
      BEETLE_OPT_LAST_SCANLINE_PAL_INFO_0_EN,
      NULL,
      "video",
      {
         { "230", NULL },
         { "231", NULL },
         { "232", NULL },
         { "233", NULL },
         { "234", NULL },
         { "235", NULL },
         { "236", NULL },
         { "237", NULL },
         { "238", NULL },
         { "239", NULL },
         { "240", NULL },
         { "241", NULL },
         { "242", NULL },
         { "243", NULL },
         { "244", NULL },
         { "245", NULL },
         { "246", NULL },
         { "247", NULL },
         { "248", NULL },
         { "249", NULL },
         { "250", NULL },
         { "251", NULL },
         { "252", NULL },
         { "253", NULL },
         { "254", NULL },
         { "255", NULL },
         { "256", NULL },
         { "257", NULL },
         { "258", NULL },
         { "259", NULL },
         { "260", NULL },
         { "261", NULL },
         { "262", NULL },
         { "263", NULL },
         { "264", NULL },
         { "265", NULL },
         { "266", NULL },
         { "267", NULL },
         { "268", NULL },
         { "269", NULL },
         { "270", NULL },
         { "271", NULL },
         { "272", NULL },
         { "273", NULL },
         { "274", NULL },
         { "275", NULL },
         { "276", NULL },
         { "277", NULL },
         { "278", NULL },
         { "279", NULL },
         { "280", NULL },
         { "281", NULL },
         { "282", NULL },
         { "283", NULL },
         { "284", NULL },
         { "285", NULL },
         { "286", NULL },
         { "287", OPTION_VAL_287_EN },
         { NULL, NULL },
      },
      "287"
   },
   { NULL, NULL, NULL, NULL, NULL, NULL, {{0}}, NULL },
};
struct retro_core_options_v2 options_en = {
   option_cats_en,
   option_defs_en
};

/* RETRO_LANGUAGE_EO */

#define CATEGORY_SYSTEM_LABEL_EO NULL
#define CATEGORY_SYSTEM_INFO_0_EO NULL
#define CATEGORY_VIDEO_LABEL_EO "Video Driver"
#define CATEGORY_VIDEO_INFO_0_EO NULL
#define CATEGORY_PGXP_LABEL_EO NULL
#define CATEGORY_PGXP_INFO_0_EO NULL
#define CATEGORY_OSD_LABEL_EO NULL
#define CATEGORY_OSD_INFO_0_EO NULL
#define CATEGORY_INPUT_LABEL_EO "Input Driver"
#define CATEGORY_INPUT_INFO_0_EO NULL
#define CATEGORY_MEMCARDS_LABEL_EO NULL
#define CATEGORY_MEMCARDS_INFO_0_EO NULL
#define CATEGORY_HACKS_LABEL_EO NULL
#define CATEGORY_HACKS_INFO_0_EO NULL
#define BEETLE_OPT_CPU_FREQ_SCALE_LABEL_EO NULL
#define BEETLE_OPT_CPU_FREQ_SCALE_INFO_0_EO NULL
#define OPTION_VAL_50_EO NULL
#define OPTION_VAL_60_EO NULL
#define OPTION_VAL_70_EO NULL
#define OPTION_VAL_80_EO NULL
#define OPTION_VAL_90_EO NULL
#define OPTION_VAL_100_EO NULL
#define OPTION_VAL_110_EO NULL
#define OPTION_VAL_120_EO NULL
#define OPTION_VAL_130_EO NULL
#define OPTION_VAL_140_EO NULL
#define OPTION_VAL_150_EO NULL
#define OPTION_VAL_160_EO NULL
#define OPTION_VAL_170_EO NULL
#define OPTION_VAL_180_EO NULL
#define OPTION_VAL_190_EO NULL
#define OPTION_VAL_200_EO NULL
#define OPTION_VAL_210_EO NULL
#define OPTION_VAL_220_EO NULL
#define OPTION_VAL_230_EO NULL
#define OPTION_VAL_240_EO NULL
#define OPTION_VAL_250_EO NULL
#define OPTION_VAL_260_EO NULL
#define OPTION_VAL_270_EO NULL
#define OPTION_VAL_280_EO NULL
#define OPTION_VAL_290_EO NULL
#define OPTION_VAL_300_EO NULL
#define OPTION_VAL_310_EO NULL
#define OPTION_VAL_320_EO NULL
#define OPTION_VAL_330_EO NULL
#define OPTION_VAL_340_EO NULL
#define OPTION_VAL_350_EO NULL
#define OPTION_VAL_360_EO NULL
#define OPTION_VAL_370_EO NULL
#define OPTION_VAL_380_EO NULL
#define OPTION_VAL_390_EO NULL
#define OPTION_VAL_400_EO NULL
#define OPTION_VAL_410_EO NULL
#define OPTION_VAL_420_EO NULL
#define OPTION_VAL_430_EO NULL
#define OPTION_VAL_440_EO NULL
#define OPTION_VAL_450_EO NULL
#define OPTION_VAL_460_EO NULL
#define OPTION_VAL_470_EO NULL
#define OPTION_VAL_480_EO NULL
#define OPTION_VAL_490_EO NULL
#define OPTION_VAL_500_EO NULL
#define OPTION_VAL_510_EO NULL
#define OPTION_VAL_520_EO NULL
#define OPTION_VAL_530_EO NULL
#define OPTION_VAL_540_EO NULL
#define OPTION_VAL_550_EO NULL
#define OPTION_VAL_560_EO NULL
#define OPTION_VAL_570_EO NULL
#define OPTION_VAL_580_EO NULL
#define OPTION_VAL_590_EO NULL
#define OPTION_VAL_600_EO NULL
#define OPTION_VAL_610_EO NULL
#define OPTION_VAL_620_EO NULL
#define OPTION_VAL_630_EO NULL
#define OPTION_VAL_640_EO NULL
#define OPTION_VAL_650_EO NULL
#define OPTION_VAL_660_EO NULL
#define OPTION_VAL_670_EO NULL
#define OPTION_VAL_680_EO NULL
#define OPTION_VAL_690_EO NULL
#define OPTION_VAL_700_EO NULL
#define OPTION_VAL_710_EO NULL
#define OPTION_VAL_720_EO NULL
#define OPTION_VAL_730_EO NULL
#define OPTION_VAL_740_EO NULL
#define OPTION_VAL_750_EO NULL
#define BEETLE_OPT_GPU_OVERCLOCK_LABEL_EO NULL
#define BEETLE_OPT_GPU_OVERCLOCK_INFO_0_EO NULL
#define OPTION_VAL_1X_NATIVE_EO NULL
#define OPTION_VAL_2X_EO NULL
#define OPTION_VAL_4X_EO NULL
#define OPTION_VAL_8X_EO NULL
#define OPTION_VAL_16X_EO NULL
#define OPTION_VAL_32X_EO NULL
#define BEETLE_OPT_GTE_OVERCLOCK_LABEL_EO NULL
#define BEETLE_OPT_GTE_OVERCLOCK_INFO_0_EO NULL
#define BEETLE_OPT_CD_ACCESS_METHOD_LABEL_EO NULL
#define BEETLE_OPT_CD_ACCESS_METHOD_INFO_0_EO NULL
#define OPTION_VAL_SYNC_EO NULL
#define OPTION_VAL_ASYNC_EO NULL
#define OPTION_VAL_PRECACHE_EO NULL
#define BEETLE_OPT_CD_FASTLOAD_LABEL_EO NULL
#define BEETLE_OPT_CD_FASTLOAD_INFO_0_EO NULL
#define OPTION_VAL_2X_NATIVE_EO NULL
#define OPTION_VAL_6X_EO NULL
#define OPTION_VAL_10X_EO NULL
#define OPTION_VAL_12X_EO NULL
#define OPTION_VAL_14X_EO NULL
#define BEETLE_OPT_SKIP_BIOS_LABEL_EO NULL
#define BEETLE_OPT_SKIP_BIOS_INFO_0_EO NULL
#define BEETLE_OPT_OVERRIDE_BIOS_LABEL_EO NULL
#define BEETLE_OPT_OVERRIDE_BIOS_INFO_0_EO NULL
#define OPTION_VAL_PSXONPSP_EO NULL
#define OPTION_VAL_PS1_ROM_EO NULL
#define OPTION_VAL_OPENBIOS_EO NULL
#define BEETLE_OPT_REGION_LABEL_EO NULL
#define BEETLE_OPT_REGION_INFO_0_EO NULL
#define OPTION_VAL_AUTO_EO NULL
#define OPTION_VAL_NTSC_J_EO NULL
#define OPTION_VAL_NTSC_U_EO NULL
#define OPTION_VAL_PAL_EO NULL
#define BEETLE_OPT_INTERNAL_RESOLUTION_LABEL_EO NULL
#define BEETLE_OPT_INTERNAL_RESOLUTION_INFO_0_EO NULL
#define BEETLE_OPT_RENDERER_LABEL_EO NULL
#define BEETLE_OPT_RENDERER_INFO_0_EO NULL
#define OPTION_VAL_HARDWARE_EO NULL
#define OPTION_VAL_HARDWARE_GL_EO NULL
#define OPTION_VAL_HARDWARE_VK_EO NULL
#define OPTION_VAL_SOFTWARE_EO NULL
#define BEETLE_OPT_RENDERER_SOFTWARE_FB_LABEL_EO NULL
#define BEETLE_OPT_RENDERER_SOFTWARE_FB_INFO_0_EO NULL
#define BEETLE_OPT_DEPTH_LABEL_EO NULL
#define BEETLE_OPT_DEPTH_INFO_0_EO NULL
#define OPTION_VAL_16BPP_NATIVE_EO NULL
#define OPTION_VAL_32BPP_EO NULL
#define BEETLE_OPT_DITHER_MODE_LABEL_EO NULL
#define BEETLE_OPT_DITHER_MODE_INFO_0_EO NULL
#define OPTION_VAL_INTERNAL_RESOLUTION_EO "Interna Rezolucio"
#define BEETLE_OPT_SCALED_UV_OFFSET_LABEL_EO NULL
#define BEETLE_OPT_SCALED_UV_OFFSET_INFO_0_EO NULL
#define BEETLE_OPT_FILTER_LABEL_EO NULL
#define BEETLE_OPT_FILTER_INFO_0_EO NULL
#define OPTION_VAL_NEAREST_EO NULL
#define OPTION_VAL_SABR_EO NULL
#define OPTION_VAL_XBR_EO NULL
#define OPTION_VAL_BILINEAR_EO NULL
#define OPTION_VAL_3_POINT_EO NULL
#define OPTION_VAL_JINC2_EO NULL
#define BEETLE_OPT_FILTER_EXCLUDE_SPRITE_LABEL_EO NULL
#define BEETLE_OPT_FILTER_EXCLUDE_SPRITE_INFO_0_EO NULL
#define OPTION_VAL_OPAQUE_EO NULL
#define OPTION_VAL_ALL_EO NULL
#define BEETLE_OPT_FILTER_EXCLUDE_2D_POLYGON_LABEL_EO NULL
#define BEETLE_OPT_FILTER_EXCLUDE_2D_POLYGON_INFO_0_EO NULL
#define BEETLE_OPT_ADAPTIVE_SMOOTHING_LABEL_EO NULL
#define BEETLE_OPT_ADAPTIVE_SMOOTHING_INFO_0_EO NULL
#define BEETLE_OPT_SUPER_SAMPLING_LABEL_EO NULL
#define BEETLE_OPT_SUPER_SAMPLING_INFO_0_EO NULL
#define BEETLE_OPT_MSAA_LABEL_EO NULL
#define BEETLE_OPT_MSAA_INFO_0_EO NULL
#define OPTION_VAL_1X_EO NULL
#define BEETLE_OPT_MDEC_YUV_LABEL_EO NULL
#define BEETLE_OPT_MDEC_YUV_INFO_0_EO NULL
#define BEETLE_OPT_TRACK_TEXTURES_LABEL_EO NULL
#define BEETLE_OPT_TRACK_TEXTURES_INFO_0_EO NULL
#define BEETLE_OPT_DUMP_TEXTURES_LABEL_EO NULL
#define BEETLE_OPT_DUMP_TEXTURES_INFO_0_EO NULL
#define BEETLE_OPT_REPLACE_TEXTURES_LABEL_EO NULL
#define BEETLE_OPT_REPLACE_TEXTURES_INFO_0_EO NULL
#define BEETLE_OPT_PGXP_MODE_LABEL_EO NULL
#define BEETLE_OPT_PGXP_MODE_INFO_0_EO NULL
#define OPTION_VAL_MEMORY_ONLY_EO NULL
#define OPTION_VAL_MEMORY_CPU_EO NULL
#define BEETLE_OPT_PGXP_2D_TOL_LABEL_EO NULL
#define BEETLE_OPT_PGXP_2D_TOL_INFO_0_EO NULL
#define OPTION_VAL_0PX_EO NULL
#define OPTION_VAL_1PX_EO NULL
#define OPTION_VAL_2PX_EO NULL
#define OPTION_VAL_3PX_EO NULL
#define OPTION_VAL_4PX_EO NULL
#define OPTION_VAL_5PX_EO NULL
#define OPTION_VAL_6PX_EO NULL
#define OPTION_VAL_7PX_EO NULL
#define OPTION_VAL_8PX_EO NULL
#define BEETLE_OPT_PGXP_NCLIP_LABEL_EO NULL
#define BEETLE_OPT_PGXP_NCLIP_INFO_0_EO NULL
#define BEETLE_OPT_PGXP_VERTEX_LABEL_EO NULL
#define BEETLE_OPT_PGXP_VERTEX_INFO_0_EO NULL
#define BEETLE_OPT_PGXP_TEXTURE_LABEL_EO NULL
#define BEETLE_OPT_PGXP_TEXTURE_INFO_0_EO NULL
#define BEETLE_OPT_DISPLAY_INTERNAL_FPS_LABEL_EO NULL
#define BEETLE_OPT_DISPLAY_INTERNAL_FPS_INFO_0_EO NULL
#define BEETLE_OPT_DISPLAY_OSD_LABEL_EO NULL
#define BEETLE_OPT_DISPLAY_OSD_INFO_0_EO NULL
#define BEETLE_OPT_DISPLAY_VRAM_LABEL_EO NULL
#define BEETLE_OPT_DISPLAY_VRAM_INFO_0_EO NULL
#define BEETLE_OPT_ANALOG_CALIBRATION_LABEL_EO NULL
#define BEETLE_OPT_ANALOG_CALIBRATION_INFO_0_EO NULL
#define BEETLE_OPT_ANALOG_TOGGLE_LABEL_EO NULL
#define BEETLE_OPT_ANALOG_TOGGLE_INFO_0_EO NULL
#define OPTION_VAL_ENABLED_ANALOG_EO NULL
#define BEETLE_OPT_ANALOG_TOGGLE_COMBO_LABEL_EO NULL
#define BEETLE_OPT_ANALOG_TOGGLE_COMBO_INFO_0_EO NULL
#define OPTION_VAL_L1_L2_R1_R2_START_SELECT_EO NULL
#define OPTION_VAL_L1_R1_SELECT_EO NULL
#define OPTION_VAL_L1_R1_START_EO NULL
#define OPTION_VAL_L1_R1_L3_EO NULL
#define OPTION_VAL_L1_R1_R3_EO NULL
#define OPTION_VAL_L2_R2_SELECT_EO NULL
#define OPTION_VAL_L2_R2_START_EO NULL
#define OPTION_VAL_L2_R2_L3_EO NULL
#define OPTION_VAL_L2_R2_R3_EO NULL
#define OPTION_VAL_L3_R3_EO NULL
#define BEETLE_OPT_ANALOG_TOGGLE_HOLD_LABEL_EO NULL
#define BEETLE_OPT_ANALOG_TOGGLE_HOLD_INFO_0_EO NULL
#define OPTION_VAL_0_EO NULL
#define OPTION_VAL_1_EO NULL
#define OPTION_VAL_2_EO NULL
#define OPTION_VAL_3_EO NULL
#define OPTION_VAL_4_EO NULL
#define OPTION_VAL_5_EO NULL
#define BEETLE_OPT_ENABLE_MULTITAP_PORT1_LABEL_EO NULL
#define BEETLE_OPT_ENABLE_MULTITAP_PORT1_INFO_0_EO NULL
#define BEETLE_OPT_ENABLE_MULTITAP_PORT2_LABEL_EO NULL
#define BEETLE_OPT_ENABLE_MULTITAP_PORT2_INFO_0_EO NULL
#define BEETLE_OPT_GUN_INPUT_MODE_LABEL_EO NULL
#define BEETLE_OPT_GUN_INPUT_MODE_INFO_0_EO NULL
#define OPTION_VAL_LIGHTGUN_EO NULL
#define OPTION_VAL_TOUCHSCREEN_EO NULL
#define BEETLE_OPT_GUN_CURSOR_LABEL_EO NULL
#define BEETLE_OPT_GUN_CURSOR_INFO_0_EO NULL
#define OPTION_VAL_CROSS_EO NULL
#define OPTION_VAL_DOT_EO NULL
#define BEETLE_OPT_CROSSHAIR_COLOR_P1_LABEL_EO NULL
#define BEETLE_OPT_CROSSHAIR_COLOR_P1_INFO_0_EO NULL
#define OPTION_VAL_RED_EO NULL
#define OPTION_VAL_BLUE_EO NULL
#define OPTION_VAL_GREEN_EO NULL
#define OPTION_VAL_ORANGE_EO NULL
#define OPTION_VAL_YELLOW_EO NULL
#define OPTION_VAL_CYAN_EO NULL
#define OPTION_VAL_PINK_EO NULL
#define OPTION_VAL_PURPLE_EO NULL
#define OPTION_VAL_BLACK_EO NULL
#define OPTION_VAL_WHITE_EO NULL
#define BEETLE_OPT_CROSSHAIR_COLOR_P2_LABEL_EO NULL
#define BEETLE_OPT_CROSSHAIR_COLOR_P2_INFO_0_EO NULL
#define BEETLE_OPT_MOUSE_SENSITIVITY_LABEL_EO NULL
#define BEETLE_OPT_MOUSE_SENSITIVITY_INFO_0_EO NULL
#define OPTION_VAL_5_O42_EO NULL
#define OPTION_VAL_10_EO NULL
#define OPTION_VAL_15_EO NULL
#define OPTION_VAL_20_EO NULL
#define OPTION_VAL_25_EO NULL
#define OPTION_VAL_30_EO NULL
#define OPTION_VAL_35_EO NULL
#define OPTION_VAL_40_EO NULL
#define OPTION_VAL_45_EO NULL
#define OPTION_VAL_55_EO NULL
#define OPTION_VAL_65_EO NULL
#define OPTION_VAL_75_EO NULL
#define OPTION_VAL_85_EO NULL
#define OPTION_VAL_95_EO NULL
#define OPTION_VAL_100_O42_EO NULL
#define OPTION_VAL_105_EO NULL
#define OPTION_VAL_115_EO NULL
#define OPTION_VAL_125_EO NULL
#define OPTION_VAL_135_EO NULL
#define OPTION_VAL_145_EO NULL
#define OPTION_VAL_155_EO NULL
#define OPTION_VAL_165_EO NULL
#define OPTION_VAL_175_EO NULL
#define OPTION_VAL_185_EO NULL
#define OPTION_VAL_195_EO NULL
#define BEETLE_OPT_NEGCON_RESPONSE_LABEL_EO NULL
#define BEETLE_OPT_NEGCON_RESPONSE_INFO_0_EO NULL
#define OPTION_VAL_LINEAR_EO NULL
#define OPTION_VAL_QUADRATIC_EO NULL
#define OPTION_VAL_CUBIC_EO NULL
#define BEETLE_OPT_NEGCON_DEADZONE_LABEL_EO NULL
#define BEETLE_OPT_NEGCON_DEADZONE_INFO_0_EO NULL
#define OPTION_VAL_0_O44_EO NULL
#define BEETLE_OPT_USE_MEDNAFEN_MEMCARD0_METHOD_LABEL_EO NULL
#define BEETLE_OPT_USE_MEDNAFEN_MEMCARD0_METHOD_INFO_0_EO NULL
#define OPTION_VAL_LIBRETRO_EO NULL
#define OPTION_VAL_MEDNAFEN_EO NULL
#define BEETLE_OPT_ENABLE_MEMCARD1_LABEL_EO NULL
#define BEETLE_OPT_ENABLE_MEMCARD1_INFO_0_EO NULL
#define BEETLE_OPT_SHARED_MEMORY_CARDS_LABEL_EO NULL
#define BEETLE_OPT_SHARED_MEMORY_CARDS_INFO_0_EO NULL
#define BEETLE_OPT_MEMCARD_LEFT_INDEX_LABEL_EO NULL
#define BEETLE_OPT_MEMCARD_LEFT_INDEX_INFO_0_EO NULL
#define OPTION_VAL_0_O48_EO NULL
#define BEETLE_OPT_MEMCARD_RIGHT_INDEX_LABEL_EO NULL
#define BEETLE_OPT_MEMCARD_RIGHT_INDEX_INFO_0_EO NULL
#define OPTION_VAL_1_O49_EO NULL
#define BEETLE_OPT_LINE_RENDER_LABEL_EO NULL
#define BEETLE_OPT_LINE_RENDER_INFO_0_EO NULL
#define OPTION_VAL_DEFAULT_EO NULL
#define OPTION_VAL_AGGRESSIVE_EO NULL
#define BEETLE_OPT_WIDESCREEN_HACK_LABEL_EO NULL
#define BEETLE_OPT_WIDESCREEN_HACK_INFO_0_EO NULL
#define BEETLE_OPT_WIDESCREEN_HACK_ASPECT_RATIO_LABEL_EO NULL
#define BEETLE_OPT_WIDESCREEN_HACK_ASPECT_RATIO_INFO_0_EO NULL
#define OPTION_VAL_16_9_EO NULL
#define OPTION_VAL_16_10_EO NULL
#define OPTION_VAL_18_9_EO NULL
#define OPTION_VAL_19_9_EO NULL
#define OPTION_VAL_20_9_EO NULL
#define OPTION_VAL_21_9_EO NULL
#define OPTION_VAL_32_9_EO NULL
#define BEETLE_OPT_CPU_DYNAREC_LABEL_EO NULL
#define BEETLE_OPT_CPU_DYNAREC_INFO_0_EO NULL
#define OPTION_VAL_DISABLED_EO NULL
#define OPTION_VAL_EXECUTE_EO NULL
#define OPTION_VAL_EXECUTE_ONE_EO NULL
#define OPTION_VAL_RUN_INTERPRETER_EO NULL
#define BEETLE_OPT_DYNAREC_INVALIDATE_LABEL_EO NULL
#define BEETLE_OPT_DYNAREC_INVALIDATE_INFO_0_EO NULL
#define OPTION_VAL_FULL_EO NULL
#define OPTION_VAL_DMA_EO NULL
#define BEETLE_OPT_DYNAREC_EVENTCYCLES_LABEL_EO NULL
#define BEETLE_OPT_DYNAREC_EVENTCYCLES_INFO_0_EO NULL
#define OPTION_VAL_128_EO NULL
#define BEETLE_OPT_DYNAREC_SPU_SAMPLES_LABEL_EO NULL
#define BEETLE_OPT_DYNAREC_SPU_SAMPLES_INFO_0_EO NULL
#define BEETLE_OPT_CORE_TIMING_FPS_LABEL_EO NULL
#define BEETLE_OPT_CORE_TIMING_FPS_INFO_0_EO NULL
#define OPTION_VAL_AUTO_TOGGLE_EO NULL
#define OPTION_VAL_FORCE_PROGRESSIVE_EO NULL
#define OPTION_VAL_FORCE_INTERLACED_EO NULL
#define BEETLE_OPT_DEINTERLACER_LABEL_EO NULL
#define BEETLE_OPT_DEINTERLACER_INFO_0_EO NULL
#define OPTION_VAL_WEAVE_EO NULL
#define OPTION_VAL_BOB_EO NULL
#define OPTION_VAL_BOB_OFFSET_EO NULL
#define OPTION_VAL_FASTMAD_EO NULL
#define BEETLE_OPT_FRAME_DUPING_LABEL_EO NULL
#define BEETLE_OPT_FRAME_DUPING_INFO_0_EO NULL
#define BEETLE_OPT_PAL_VIDEO_TIMING_OVERRIDE_LABEL_EO NULL
#define BEETLE_OPT_PAL_VIDEO_TIMING_OVERRIDE_INFO_0_EO NULL
#define BEETLE_OPT_ASPECT_RATIO_LABEL_EO NULL
#define BEETLE_OPT_ASPECT_RATIO_INFO_0_EO NULL
#define OPTION_VAL_CORRECTED_EO NULL
#define OPTION_VAL_UNCORRECTED_EO NULL
#define OPTION_VAL_4_3_EO NULL
#define OPTION_VAL_NTSC_EO NULL
#define BEETLE_OPT_CROP_OVERSCAN_LABEL_EO NULL
#define BEETLE_OPT_CROP_OVERSCAN_INFO_0_EO NULL
#define OPTION_VAL_STATIC_EO NULL
#define OPTION_VAL_SMART_EO NULL
#define BEETLE_OPT_IMAGE_CROP_LABEL_EO NULL
#define BEETLE_OPT_IMAGE_CROP_INFO_0_EO NULL
#define OPTION_VAL_9PX_EO NULL
#define OPTION_VAL_10PX_EO NULL
#define OPTION_VAL_11PX_EO NULL
#define OPTION_VAL_12PX_EO NULL
#define OPTION_VAL_13PX_EO NULL
#define OPTION_VAL_14PX_EO NULL
#define OPTION_VAL_15PX_EO NULL
#define OPTION_VAL_16PX_EO NULL
#define OPTION_VAL_17PX_EO NULL
#define OPTION_VAL_18PX_EO NULL
#define OPTION_VAL_19PX_EO NULL
#define OPTION_VAL_20PX_EO NULL
#define BEETLE_OPT_IMAGE_OFFSET_LABEL_EO NULL
#define BEETLE_OPT_IMAGE_OFFSET_INFO_0_EO NULL
#define OPTION_VAL_12PX_O64_EO NULL
#define OPTION_VAL_11PX_O64_EO NULL
#define OPTION_VAL_10PX_O64_EO NULL
#define OPTION_VAL_9PX_O64_EO NULL
#define OPTION_VAL_8PX_O64_EO NULL
#define OPTION_VAL_7PX_O64_EO NULL
#define OPTION_VAL_6PX_O64_EO NULL
#define OPTION_VAL_5PX_O64_EO NULL
#define OPTION_VAL_4PX_O64_EO NULL
#define OPTION_VAL_3PX_O64_EO NULL
#define OPTION_VAL_2PX_O64_EO NULL
#define OPTION_VAL_1PX_O64_EO NULL
#define OPTION_VAL_1PX_O64_0_EO NULL
#define OPTION_VAL_2PX_O64_0_EO NULL
#define OPTION_VAL_3PX_O64_0_EO NULL
#define OPTION_VAL_4PX_O64_0_EO NULL
#define OPTION_VAL_5PX_O64_0_EO NULL
#define OPTION_VAL_6PX_O64_0_EO NULL
#define OPTION_VAL_7PX_O64_0_EO NULL
#define OPTION_VAL_8PX_O64_0_EO NULL
#define OPTION_VAL_9PX_O64_0_EO NULL
#define OPTION_VAL_10PX_O64_0_EO NULL
#define OPTION_VAL_11PX_O64_0_EO NULL
#define OPTION_VAL_12PX_O64_0_EO NULL
#define BEETLE_OPT_IMAGE_OFFSET_CYCLES_LABEL_EO NULL
#define BEETLE_OPT_IMAGE_OFFSET_CYCLES_INFO_0_EO NULL
#define BEETLE_OPT_INITIAL_SCANLINE_LABEL_EO NULL
#define BEETLE_OPT_INITIAL_SCANLINE_INFO_0_EO NULL
#define BEETLE_OPT_LAST_SCANLINE_LABEL_EO NULL
#define BEETLE_OPT_LAST_SCANLINE_INFO_0_EO NULL
#define OPTION_VAL_239_EO NULL
#define BEETLE_OPT_INITIAL_SCANLINE_PAL_LABEL_EO NULL
#define BEETLE_OPT_INITIAL_SCANLINE_PAL_INFO_0_EO NULL
#define BEETLE_OPT_LAST_SCANLINE_PAL_LABEL_EO NULL
#define BEETLE_OPT_LAST_SCANLINE_PAL_INFO_0_EO NULL
#define OPTION_VAL_287_EO NULL

struct retro_core_option_v2_category option_cats_eo[] = {
   {
      "system",
      CATEGORY_SYSTEM_LABEL_EO,
      CATEGORY_SYSTEM_INFO_0_EO
   },
   {
      "video",
      CATEGORY_VIDEO_LABEL_EO,
      CATEGORY_VIDEO_INFO_0_EO
   },
   {
      "pgxp",
      CATEGORY_PGXP_LABEL_EO,
      CATEGORY_PGXP_INFO_0_EO
   },
   {
      "osd",
      CATEGORY_OSD_LABEL_EO,
      CATEGORY_OSD_INFO_0_EO
   },
   {
      "input",
      CATEGORY_INPUT_LABEL_EO,
      CATEGORY_INPUT_INFO_0_EO
   },
   {
      "memcards",
      CATEGORY_MEMCARDS_LABEL_EO,
      CATEGORY_MEMCARDS_INFO_0_EO
   },
   {
      "hacks",
      CATEGORY_HACKS_LABEL_EO,
      CATEGORY_HACKS_INFO_0_EO
   },
   { NULL, NULL, NULL },
};
struct retro_core_option_v2_definition option_defs_eo[] = {
   {
      BEETLE_OPT(cpu_freq_scale),
      BEETLE_OPT_CPU_FREQ_SCALE_LABEL_EO,
      NULL,
      BEETLE_OPT_CPU_FREQ_SCALE_INFO_0_EO,
      NULL,
      "system",
      {
         { "50%",           OPTION_VAL_50_EO },
         { "60%",           OPTION_VAL_60_EO },
         { "70%",           OPTION_VAL_70_EO },
         { "80%",           OPTION_VAL_80_EO },
         { "90%",           OPTION_VAL_90_EO },
         { "100%",          OPTION_VAL_100_EO },
         { "110%",          OPTION_VAL_110_EO },
         { "120%",          OPTION_VAL_120_EO },
         { "130%",          OPTION_VAL_130_EO },
         { "140%",          OPTION_VAL_140_EO },
         { "150%",          OPTION_VAL_150_EO },
         { "160%",          OPTION_VAL_160_EO },
         { "170%",          OPTION_VAL_170_EO },
         { "180%",          OPTION_VAL_180_EO },
         { "190%",          OPTION_VAL_190_EO },
         { "200%",          OPTION_VAL_200_EO },
         { "210%",          OPTION_VAL_210_EO },
         { "220%",          OPTION_VAL_220_EO },
         { "230%",          OPTION_VAL_230_EO },
         { "240%",          OPTION_VAL_240_EO },
         { "250%",          OPTION_VAL_250_EO },
         { "260%",          OPTION_VAL_260_EO },
         { "270%",          OPTION_VAL_270_EO },
         { "280%",          OPTION_VAL_280_EO },
         { "290%",          OPTION_VAL_290_EO },
         { "300%",          OPTION_VAL_300_EO },
         { "310%",          OPTION_VAL_310_EO },
         { "320%",          OPTION_VAL_320_EO },
         { "330%",          OPTION_VAL_330_EO },
         { "340%",          OPTION_VAL_340_EO },
         { "350%",          OPTION_VAL_350_EO },
         { "360%",          OPTION_VAL_360_EO },
         { "370%",          OPTION_VAL_370_EO },
         { "380%",          OPTION_VAL_380_EO },
         { "390%",          OPTION_VAL_390_EO },
         { "400%",          OPTION_VAL_400_EO },
         { "410%",          OPTION_VAL_410_EO },
         { "420%",          OPTION_VAL_420_EO },
         { "430%",          OPTION_VAL_430_EO },
         { "440%",          OPTION_VAL_440_EO },
         { "450%",          OPTION_VAL_450_EO },
         { "460%",          OPTION_VAL_460_EO },
         { "470%",          OPTION_VAL_470_EO },
         { "480%",          OPTION_VAL_480_EO },
         { "490%",          OPTION_VAL_490_EO },
         { "500%",          OPTION_VAL_500_EO },
         { "510%",          OPTION_VAL_510_EO },
         { "520%",          OPTION_VAL_520_EO },
         { "530%",          OPTION_VAL_530_EO },
         { "540%",          OPTION_VAL_540_EO },
         { "550%",          OPTION_VAL_550_EO },
         { "560%",          OPTION_VAL_560_EO },
         { "570%",          OPTION_VAL_570_EO },
         { "580%",          OPTION_VAL_580_EO },
         { "590%",          OPTION_VAL_590_EO },
         { "600%",          OPTION_VAL_600_EO },
         { "610%",          OPTION_VAL_610_EO },
         { "620%",          OPTION_VAL_620_EO },
         { "630%",          OPTION_VAL_630_EO },
         { "640%",          OPTION_VAL_640_EO },
         { "650%",          OPTION_VAL_650_EO },
         { "660%",          OPTION_VAL_660_EO },
         { "670%",          OPTION_VAL_670_EO },
         { "680%",          OPTION_VAL_680_EO },
         { "690%",          OPTION_VAL_690_EO },
         { "700%",          OPTION_VAL_700_EO },
         { "710%",          OPTION_VAL_710_EO },
         { "720%",          OPTION_VAL_720_EO },
         { "730%",          OPTION_VAL_730_EO },
         { "740%",          OPTION_VAL_740_EO },
         { "750%",          OPTION_VAL_750_EO },
         { NULL, NULL },
      },
      "100%"
   },
   {
      BEETLE_OPT(gpu_overclock),
      BEETLE_OPT_GPU_OVERCLOCK_LABEL_EO,
      NULL,
      BEETLE_OPT_GPU_OVERCLOCK_INFO_0_EO,
      NULL,
      "system",
      {
         { "1x(native)", OPTION_VAL_1X_NATIVE_EO },
         { "2x",         OPTION_VAL_2X_EO },
         { "4x",         OPTION_VAL_4X_EO },
         { "8x",         OPTION_VAL_8X_EO },
         { "16x",        OPTION_VAL_16X_EO },
         { "32x",        OPTION_VAL_32X_EO },
         { NULL, NULL },
      },
      "1x(native)"
   },
   {
      BEETLE_OPT(gte_overclock),
      BEETLE_OPT_GTE_OVERCLOCK_LABEL_EO,
      NULL,
      BEETLE_OPT_GTE_OVERCLOCK_INFO_0_EO,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
#ifndef EMSCRIPTEN
   {
      BEETLE_OPT(cd_access_method),
      BEETLE_OPT_CD_ACCESS_METHOD_LABEL_EO,
      NULL,
      BEETLE_OPT_CD_ACCESS_METHOD_INFO_0_EO,
      NULL,
      "system",
      {
         { "sync",     OPTION_VAL_SYNC_EO },
         { "async",    OPTION_VAL_ASYNC_EO },
         { "precache", OPTION_VAL_PRECACHE_EO },
         { NULL, NULL },
      },
      "sync"
   },
#endif
   {
      BEETLE_OPT(cd_fastload),
      BEETLE_OPT_CD_FASTLOAD_LABEL_EO,
      NULL,
      BEETLE_OPT_CD_FASTLOAD_INFO_0_EO,
      NULL,
      "system",
      {
         { "2x(native)", OPTION_VAL_2X_NATIVE_EO },
         { "4x",          OPTION_VAL_4X_EO },
         { "6x",          OPTION_VAL_6X_EO },
         { "8x",          OPTION_VAL_8X_EO },
         { "10x",         OPTION_VAL_10X_EO },
         { "12x",         OPTION_VAL_12X_EO },
         { "14x",         OPTION_VAL_14X_EO },
         { NULL, NULL },
      },
      "2x(native)"
   },
   {
      BEETLE_OPT(skip_bios),
      BEETLE_OPT_SKIP_BIOS_LABEL_EO,
      NULL,
      BEETLE_OPT_SKIP_BIOS_INFO_0_EO,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(override_bios),
      BEETLE_OPT_OVERRIDE_BIOS_LABEL_EO,
      NULL,
      BEETLE_OPT_OVERRIDE_BIOS_INFO_0_EO,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "psxonpsp", OPTION_VAL_PSXONPSP_EO },
         { "ps1_rom",  OPTION_VAL_PS1_ROM_EO },
         { "openbios", OPTION_VAL_OPENBIOS_EO     },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(region),
      BEETLE_OPT_REGION_LABEL_EO,
      NULL,
      BEETLE_OPT_REGION_INFO_0_EO,
      NULL,
      "system",
      {
         { "auto",   OPTION_VAL_AUTO_EO },
         { "ntsc-j", OPTION_VAL_NTSC_J_EO },
         { "ntsc-u", OPTION_VAL_NTSC_U_EO },
         { "pal",    OPTION_VAL_PAL_EO },
         { NULL, NULL },
      },
      "auto"
   },
   {
      BEETLE_OPT(internal_resolution),
      BEETLE_OPT_INTERNAL_RESOLUTION_LABEL_EO,
      NULL,
      BEETLE_OPT_INTERNAL_RESOLUTION_INFO_0_EO,
      NULL,
      "video",
      {
         { "1x(native)", OPTION_VAL_1X_NATIVE_EO },
         { "2x",         OPTION_VAL_2X_EO },
         { "4x",         OPTION_VAL_4X_EO },
         { "8x",         OPTION_VAL_8X_EO },
         { "16x",        OPTION_VAL_16X_EO },
         { NULL, NULL },
      },
      "1x(native)"
   },
#if defined(HAVE_OPENGL) || defined(HAVE_OPENGLES) || defined(HAVE_VULKAN)
   {
      BEETLE_OPT(renderer),
      BEETLE_OPT_RENDERER_LABEL_EO,
      NULL,
      BEETLE_OPT_RENDERER_INFO_0_EO,
      NULL,
      "video",
      {
         { "hardware",    OPTION_VAL_HARDWARE_EO },
#if defined(HAVE_OPENGL) || defined(HAVE_OPENGLES)
         { "hardware_gl", OPTION_VAL_HARDWARE_GL_EO },
#endif
#if defined(HAVE_VULKAN)
         { "hardware_vk", OPTION_VAL_HARDWARE_VK_EO },
#endif
         { "software",    OPTION_VAL_SOFTWARE_EO },
         { NULL, NULL },
      },
      "hardware"
   },
   {
      BEETLE_OPT(renderer_software_fb),
      BEETLE_OPT_RENDERER_SOFTWARE_FB_LABEL_EO,
      NULL,
      BEETLE_OPT_RENDERER_SOFTWARE_FB_INFO_0_EO,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled"
   },
#endif
#if defined(HAVE_OPENGL) || defined(HAVE_OPENGLES)
   {
      BEETLE_OPT(depth),
      BEETLE_OPT_DEPTH_LABEL_EO,
      NULL,
      BEETLE_OPT_DEPTH_INFO_0_EO,
      NULL,
      "video",
      {
         { "16bpp(native)", OPTION_VAL_16BPP_NATIVE_EO },
         { "32bpp",         OPTION_VAL_32BPP_EO },
         { NULL, NULL },
      },
      "16bpp(native)"
   },
   // Sort of, it's more like 15-bit high color and 24-bit true color for visible output. The alpha channel is used for mask bit. Vulkan renderer uses ABGR1555_555 for 31 bits internal? FMVs are always 24-bit on all renderers like original hardware (BGR888, no alpha)
#endif
   {
      BEETLE_OPT(dither_mode),
      BEETLE_OPT_DITHER_MODE_LABEL_EO,
      NULL,
      BEETLE_OPT_DITHER_MODE_INFO_0_EO,
      NULL,
      "video",
      {
         { "1x(native)",          OPTION_VAL_1X_NATIVE_EO },
         { "internal resolution", OPTION_VAL_INTERNAL_RESOLUTION_EO },
         { "disabled",            NULL },
         { NULL, NULL },
      },
      "1x(native)"
   },
#ifdef HAVE_VULKAN
   {
      BEETLE_OPT(scaled_uv_offset),
      BEETLE_OPT_SCALED_UV_OFFSET_LABEL_EO,
      NULL,
      BEETLE_OPT_SCALED_UV_OFFSET_INFO_0_EO,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled"
   },
#endif
#if defined(HAVE_OPENGL) || defined(HAVE_OPENGLES) || defined(HAVE_VULKAN)
   {
      BEETLE_OPT(filter),
      BEETLE_OPT_FILTER_LABEL_EO,
      NULL,
      BEETLE_OPT_FILTER_INFO_0_EO,
      NULL,
      "video",
      {
         { "nearest",  OPTION_VAL_NEAREST_EO },
         { "SABR",     OPTION_VAL_SABR_EO },
         { "xBR",      OPTION_VAL_XBR_EO },
         { "bilinear", OPTION_VAL_BILINEAR_EO },
         { "3-point",  OPTION_VAL_3_POINT_EO },
         { "JINC2",    OPTION_VAL_JINC2_EO },
         { NULL, NULL },
      },
      "nearest"
   },
#ifdef HAVE_VULKAN
   {
      BEETLE_OPT(filter_exclude_sprite),
      BEETLE_OPT_FILTER_EXCLUDE_SPRITE_LABEL_EO,
      NULL,
      BEETLE_OPT_FILTER_EXCLUDE_SPRITE_INFO_0_EO,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "opaque", OPTION_VAL_OPAQUE_EO },
         { "all", OPTION_VAL_ALL_EO },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(filter_exclude_2d_polygon),
      BEETLE_OPT_FILTER_EXCLUDE_2D_POLYGON_LABEL_EO,
      NULL,
      BEETLE_OPT_FILTER_EXCLUDE_2D_POLYGON_INFO_0_EO,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "opaque", OPTION_VAL_OPAQUE_EO },
         { "all", OPTION_VAL_ALL_EO },
         { NULL, NULL },
      },
      "disabled"
   },
#endif
#endif
#ifdef HAVE_VULKAN
   {
      BEETLE_OPT(adaptive_smoothing),
      BEETLE_OPT_ADAPTIVE_SMOOTHING_LABEL_EO,
      NULL,
      BEETLE_OPT_ADAPTIVE_SMOOTHING_INFO_0_EO,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(super_sampling),
      BEETLE_OPT_SUPER_SAMPLING_LABEL_EO,
      NULL,
      BEETLE_OPT_SUPER_SAMPLING_INFO_0_EO,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(msaa),
      BEETLE_OPT_MSAA_LABEL_EO,
      NULL,
      BEETLE_OPT_MSAA_INFO_0_EO,
      NULL,
      "video",
      {
         { "1x",  OPTION_VAL_1X_EO },
         { "2x",  OPTION_VAL_2X_EO },
         { "4x",  OPTION_VAL_4X_EO },
         { "8x",  OPTION_VAL_8X_EO },
         { "16x", OPTION_VAL_16X_EO },
         { NULL, NULL },
      },
      "1x"
   },
   {
      BEETLE_OPT(mdec_yuv),
      BEETLE_OPT_MDEC_YUV_LABEL_EO,
      NULL,
      BEETLE_OPT_MDEC_YUV_INFO_0_EO,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(track_textures),
      BEETLE_OPT_TRACK_TEXTURES_LABEL_EO,
      NULL,
      BEETLE_OPT_TRACK_TEXTURES_INFO_0_EO,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
#ifdef TEXTURE_DUMPING_ENABLED
   {
      BEETLE_OPT(dump_textures),
      BEETLE_OPT_DUMP_TEXTURES_LABEL_EO,
      NULL,
      BEETLE_OPT_DUMP_TEXTURES_INFO_0_EO,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
#endif
   {
      BEETLE_OPT(replace_textures),
      BEETLE_OPT_REPLACE_TEXTURES_LABEL_EO,
      NULL,
      BEETLE_OPT_REPLACE_TEXTURES_INFO_0_EO,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
#endif
   {
      BEETLE_OPT(pgxp_mode),
      BEETLE_OPT_PGXP_MODE_LABEL_EO,
      NULL,
      BEETLE_OPT_PGXP_MODE_INFO_0_EO,
      NULL,
      "pgxp",
      {
         { "disabled",     NULL },
         { "memory only",  OPTION_VAL_MEMORY_ONLY_EO },
         { "memory + CPU", OPTION_VAL_MEMORY_CPU_EO },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(pgxp_2d_tol),
      BEETLE_OPT_PGXP_2D_TOL_LABEL_EO,
      NULL,
      BEETLE_OPT_PGXP_2D_TOL_INFO_0_EO,
      NULL,
      "pgxp",
      {
         { "disabled", NULL },
         { "0px", OPTION_VAL_0PX_EO },
         { "1px", OPTION_VAL_1PX_EO },
         { "2px", OPTION_VAL_2PX_EO },
         { "3px", OPTION_VAL_3PX_EO },
         { "4px", OPTION_VAL_4PX_EO },
         { "5px", OPTION_VAL_5PX_EO },
         { "6px", OPTION_VAL_6PX_EO },
         { "7px", OPTION_VAL_7PX_EO },
         { "8px", OPTION_VAL_8PX_EO },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(pgxp_nclip),
      BEETLE_OPT_PGXP_NCLIP_LABEL_EO,
      NULL,
      BEETLE_OPT_PGXP_NCLIP_INFO_0_EO,
      NULL,
      "pgxp",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
#if defined(HAVE_OPENGL) || defined(HAVE_OPENGLES) || defined(HAVE_VULKAN)
   {
      BEETLE_OPT(pgxp_vertex),
      BEETLE_OPT_PGXP_VERTEX_LABEL_EO,
      NULL,
      BEETLE_OPT_PGXP_VERTEX_INFO_0_EO,
      NULL,
      "pgxp",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
#endif
   {
      BEETLE_OPT(pgxp_texture),
      BEETLE_OPT_PGXP_TEXTURE_LABEL_EO,
      NULL,
      BEETLE_OPT_PGXP_TEXTURE_INFO_0_EO,
      NULL,
      "pgxp",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(display_internal_fps),
      BEETLE_OPT_DISPLAY_INTERNAL_FPS_LABEL_EO,
      NULL,
      BEETLE_OPT_DISPLAY_INTERNAL_FPS_INFO_0_EO,
      NULL,
      "osd",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(display_osd),
      BEETLE_OPT_DISPLAY_OSD_LABEL_EO,
      NULL,
      BEETLE_OPT_DISPLAY_OSD_INFO_0_EO,
      NULL,
      "osd",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled"
   },
#if defined(HAVE_OPENGL) || defined(HAVE_OPENGLES) || defined(HAVE_VULKAN)
   {
      BEETLE_OPT(display_vram),
      BEETLE_OPT_DISPLAY_VRAM_LABEL_EO,
      NULL,
      BEETLE_OPT_DISPLAY_VRAM_INFO_0_EO,
      NULL,
      "osd",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
#endif
   {
      BEETLE_OPT(analog_calibration),
      BEETLE_OPT_ANALOG_CALIBRATION_LABEL_EO,
      NULL,
      BEETLE_OPT_ANALOG_CALIBRATION_INFO_0_EO,
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(analog_toggle),
      BEETLE_OPT_ANALOG_TOGGLE_LABEL_EO,
      NULL,
      BEETLE_OPT_ANALOG_TOGGLE_INFO_0_EO,
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { "enabled-analog", OPTION_VAL_ENABLED_ANALOG_EO },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(analog_toggle_combo),
      BEETLE_OPT_ANALOG_TOGGLE_COMBO_LABEL_EO,
      NULL,
      BEETLE_OPT_ANALOG_TOGGLE_COMBO_INFO_0_EO,
      NULL,
      "input",
      {
         { "l1+l2+r1+r2+start+select", OPTION_VAL_L1_L2_R1_R2_START_SELECT_EO },
         { "l1+r1+select",             OPTION_VAL_L1_R1_SELECT_EO },
         { "l1+r1+start",              OPTION_VAL_L1_R1_START_EO },
         { "l1+r1+l3",                 OPTION_VAL_L1_R1_L3_EO },
         { "l1+r1+r3",                 OPTION_VAL_L1_R1_R3_EO },
         { "l2+r2+select",             OPTION_VAL_L2_R2_SELECT_EO },
         { "l2+r2+start",              OPTION_VAL_L2_R2_START_EO },
         { "l2+r2+l3",                 OPTION_VAL_L2_R2_L3_EO },
         { "l2+r2+r3",                 OPTION_VAL_L2_R2_R3_EO },
         { "l3+r3",                    OPTION_VAL_L3_R3_EO },
         { NULL, NULL },
      },
      "l1+r1+select"
   },
   {
      BEETLE_OPT(analog_toggle_hold),
      BEETLE_OPT_ANALOG_TOGGLE_HOLD_LABEL_EO,
      NULL,
      BEETLE_OPT_ANALOG_TOGGLE_HOLD_INFO_0_EO,
      NULL,
      "input",
      {
         { "0", OPTION_VAL_0_EO },
         { "1", OPTION_VAL_1_EO },
         { "2", OPTION_VAL_2_EO },
         { "3", OPTION_VAL_3_EO },
         { "4", OPTION_VAL_4_EO },
         { "5", OPTION_VAL_5_EO },
         { NULL, NULL },
      },
      "1"
   },
   {
      BEETLE_OPT(enable_multitap_port1),
      BEETLE_OPT_ENABLE_MULTITAP_PORT1_LABEL_EO,
      NULL,
      BEETLE_OPT_ENABLE_MULTITAP_PORT1_INFO_0_EO,
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(enable_multitap_port2),
      BEETLE_OPT_ENABLE_MULTITAP_PORT2_LABEL_EO,
      NULL,
      BEETLE_OPT_ENABLE_MULTITAP_PORT2_INFO_0_EO,
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(gun_input_mode),
      BEETLE_OPT_GUN_INPUT_MODE_LABEL_EO,
      NULL,
      BEETLE_OPT_GUN_INPUT_MODE_INFO_0_EO,
      NULL,
      "input",
      {
         { "lightgun",    OPTION_VAL_LIGHTGUN_EO },
         { "touchscreen", OPTION_VAL_TOUCHSCREEN_EO },
         { NULL, NULL },
      },
      "lightgun"
   },
   // Shouldn't the gun_input_mode just be Mouse vs. Touchscreen?
   {
      BEETLE_OPT(gun_cursor),
      BEETLE_OPT_GUN_CURSOR_LABEL_EO,
      NULL,
      BEETLE_OPT_GUN_CURSOR_INFO_0_EO,
      NULL,
      "input",
      {
         { "off",   "disabled" },
         { "cross", OPTION_VAL_CROSS_EO },
         { "dot",   OPTION_VAL_DOT_EO },
         { NULL, NULL },
      },
      "cross"
   },
   {
      BEETLE_OPT(crosshair_color_p1),
      BEETLE_OPT_CROSSHAIR_COLOR_P1_LABEL_EO,
      NULL,
      BEETLE_OPT_CROSSHAIR_COLOR_P1_INFO_0_EO,
      NULL,
      "input",
      {
         { "red",    OPTION_VAL_RED_EO },
         { "blue",   OPTION_VAL_BLUE_EO },
         { "green",  OPTION_VAL_GREEN_EO },
         { "orange", OPTION_VAL_ORANGE_EO },
         { "yellow", OPTION_VAL_YELLOW_EO },
         { "cyan",   OPTION_VAL_CYAN_EO },
         { "pink",   OPTION_VAL_PINK_EO },
         { "purple", OPTION_VAL_PURPLE_EO },
         { "black",  OPTION_VAL_BLACK_EO },
         { "white",  OPTION_VAL_WHITE_EO },
         { NULL, NULL },
      },
      "red"
   },
   {
      BEETLE_OPT(crosshair_color_p2),
      BEETLE_OPT_CROSSHAIR_COLOR_P2_LABEL_EO,
      NULL,
      BEETLE_OPT_CROSSHAIR_COLOR_P2_INFO_0_EO,
      NULL,
      "input",
      {
         { "blue",   OPTION_VAL_BLUE_EO },
         { "red",    OPTION_VAL_RED_EO },
         { "green",  OPTION_VAL_GREEN_EO },
         { "orange", OPTION_VAL_ORANGE_EO },
         { "yellow", OPTION_VAL_YELLOW_EO },
         { "cyan",   OPTION_VAL_CYAN_EO },
         { "pink",   OPTION_VAL_PINK_EO },
         { "purple", OPTION_VAL_PURPLE_EO },
         { "black",  OPTION_VAL_BLACK_EO },
         { "white",  OPTION_VAL_WHITE_EO },
         { NULL, NULL },
      },
      "blue"
   },
   {
      BEETLE_OPT(mouse_sensitivity),
      BEETLE_OPT_MOUSE_SENSITIVITY_LABEL_EO,
      NULL,
      BEETLE_OPT_MOUSE_SENSITIVITY_INFO_0_EO,
      NULL,
      "input",
      {
         { "5%",   OPTION_VAL_5_O42_EO },
         { "10%",  OPTION_VAL_10_EO },
         { "15%",  OPTION_VAL_15_EO },
         { "20%",  OPTION_VAL_20_EO },
         { "25%",  OPTION_VAL_25_EO },
         { "30%",  OPTION_VAL_30_EO },
         { "35%",  OPTION_VAL_35_EO },
         { "40%",  OPTION_VAL_40_EO },
         { "45%",  OPTION_VAL_45_EO },
         { "50%",  OPTION_VAL_50_EO },
         { "55%",  OPTION_VAL_55_EO },
         { "60%",  OPTION_VAL_60_EO },
         { "65%",  OPTION_VAL_65_EO },
         { "70%",  OPTION_VAL_70_EO },
         { "75%",  OPTION_VAL_75_EO },
         { "80%",  OPTION_VAL_80_EO },
         { "85%",  OPTION_VAL_85_EO },
         { "90%",  OPTION_VAL_90_EO },
         { "95%",  OPTION_VAL_95_EO },
         { "100%", OPTION_VAL_100_O42_EO },
         { "105%", OPTION_VAL_105_EO },
         { "110%", OPTION_VAL_110_EO },
         { "115%", OPTION_VAL_115_EO },
         { "120%", OPTION_VAL_120_EO },
         { "125%", OPTION_VAL_125_EO },
         { "130%", OPTION_VAL_130_EO },
         { "135%", OPTION_VAL_135_EO },
         { "140%", OPTION_VAL_140_EO },
         { "145%", OPTION_VAL_145_EO },
         { "150%", OPTION_VAL_150_EO },
         { "155%", OPTION_VAL_155_EO },
         { "160%", OPTION_VAL_160_EO },
         { "165%", OPTION_VAL_165_EO },
         { "170%", OPTION_VAL_170_EO },
         { "175%", OPTION_VAL_175_EO },
         { "180%", OPTION_VAL_180_EO },
         { "185%", OPTION_VAL_185_EO },
         { "190%", OPTION_VAL_190_EO },
         { "195%", OPTION_VAL_195_EO },
         { "200%", OPTION_VAL_200_EO },
         { NULL, NULL },
      },
      "100%"
   },
   {
      BEETLE_OPT(negcon_response),
      BEETLE_OPT_NEGCON_RESPONSE_LABEL_EO,
      NULL,
      BEETLE_OPT_NEGCON_RESPONSE_INFO_0_EO,
      NULL,
      "input",
      {
         { "linear",    OPTION_VAL_LINEAR_EO },
         { "quadratic", OPTION_VAL_QUADRATIC_EO },
         { "cubic",     OPTION_VAL_CUBIC_EO },
         { NULL, NULL },
      },
      "linear"
   },
   {
      BEETLE_OPT(negcon_deadzone),
      BEETLE_OPT_NEGCON_DEADZONE_LABEL_EO,
      NULL,
      BEETLE_OPT_NEGCON_DEADZONE_INFO_0_EO,
      NULL,
      "input",
      {
         { "0%",  OPTION_VAL_0_O44_EO },
         { "5%",  OPTION_VAL_5_O42_EO },
         { "10%", OPTION_VAL_10_EO },
         { "15%", OPTION_VAL_15_EO },
         { "20%", OPTION_VAL_20_EO },
         { "25%", OPTION_VAL_25_EO },
         { "30%", OPTION_VAL_30_EO },
         { NULL, NULL },
      },
      "0%"
   },
   {
      BEETLE_OPT(use_mednafen_memcard0_method),
      BEETLE_OPT_USE_MEDNAFEN_MEMCARD0_METHOD_LABEL_EO,
      NULL,
      BEETLE_OPT_USE_MEDNAFEN_MEMCARD0_METHOD_INFO_0_EO,
      NULL,
      "memcards",
      {
         { "libretro", OPTION_VAL_LIBRETRO_EO },
         { "mednafen", OPTION_VAL_MEDNAFEN_EO },
         { NULL, NULL },
      },
      "libretro"
   },
   {
      BEETLE_OPT(enable_memcard1),
      BEETLE_OPT_ENABLE_MEMCARD1_LABEL_EO,
      NULL,
      BEETLE_OPT_ENABLE_MEMCARD1_INFO_0_EO,
      NULL,
      "memcards",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(shared_memory_cards),
      BEETLE_OPT_SHARED_MEMORY_CARDS_LABEL_EO,
      NULL,
      BEETLE_OPT_SHARED_MEMORY_CARDS_INFO_0_EO,
      NULL,
      "memcards",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(memcard_left_index),
      BEETLE_OPT_MEMCARD_LEFT_INDEX_LABEL_EO,
      NULL,
      BEETLE_OPT_MEMCARD_LEFT_INDEX_INFO_0_EO,
      NULL,
      "memcards",
      {
         { "0",  OPTION_VAL_0_O48_EO },
         { "1",  NULL },
         { "2",  NULL },
         { "3",  NULL },
         { "4",  NULL },
         { "5",  NULL },
         { "6",  NULL },
         { "7",  NULL },
         { "8",  NULL },
         { "9",  NULL },
         { "10", NULL },
         { "11", NULL },
         { "12", NULL },
         { "13", NULL },
         { "14", NULL },
         { "15", NULL },
         { "16", NULL },
         { "17", NULL },
         { "18", NULL },
         { "19", NULL },
         { "20", NULL },
         { "21", NULL },
         { "22", NULL },
         { "23", NULL },
         { "24", NULL },
         { "25", NULL },
         { "26", NULL },
         { "27", NULL },
         { "28", NULL },
         { "29", NULL },
         { "30", NULL },
         { "31", NULL },
         { "32", NULL },
         { "33", NULL },
         { "34", NULL },
         { "35", NULL },
         { "36", NULL },
         { "37", NULL },
         { "38", NULL },
         { "39", NULL },
         { "40", NULL },
         { "41", NULL },
         { "42", NULL },
         { "43", NULL },
         { "44", NULL },
         { "45", NULL },
         { "46", NULL },
         { "47", NULL },
         { "48", NULL },
         { "49", NULL },
         { "50", NULL },
         { "51", NULL },
         { "52", NULL },
         { "53", NULL },
         { "54", NULL },
         { "55", NULL },
         { "56", NULL },
         { "57", NULL },
         { "58", NULL },
         { "59", NULL },
         { "60", NULL },
         { "61", NULL },
         { "62", NULL },
         { "63", NULL },
         { NULL, NULL },
      },
      "0"
   },
   {
      BEETLE_OPT(memcard_right_index),
      BEETLE_OPT_MEMCARD_RIGHT_INDEX_LABEL_EO,
      NULL,
      BEETLE_OPT_MEMCARD_RIGHT_INDEX_INFO_0_EO,
      NULL,
      "memcards",
      {
         { "0",  NULL },
         { "1",  OPTION_VAL_1_O49_EO },
         { "2",  NULL },
         { "3",  NULL },
         { "4",  NULL },
         { "5",  NULL },
         { "6",  NULL },
         { "7",  NULL },
         { "8",  NULL },
         { "9",  NULL },
         { "10", NULL },
         { "11", NULL },
         { "12", NULL },
         { "13", NULL },
         { "14", NULL },
         { "15", NULL },
         { "16", NULL },
         { "17", NULL },
         { "18", NULL },
         { "19", NULL },
         { "20", NULL },
         { "21", NULL },
         { "22", NULL },
         { "23", NULL },
         { "24", NULL },
         { "25", NULL },
         { "26", NULL },
         { "27", NULL },
         { "28", NULL },
         { "29", NULL },
         { "30", NULL },
         { "31", NULL },
         { "32", NULL },
         { "33", NULL },
         { "34", NULL },
         { "35", NULL },
         { "36", NULL },
         { "37", NULL },
         { "38", NULL },
         { "39", NULL },
         { "40", NULL },
         { "41", NULL },
         { "42", NULL },
         { "43", NULL },
         { "44", NULL },
         { "45", NULL },
         { "46", NULL },
         { "47", NULL },
         { "48", NULL },
         { "49", NULL },
         { "50", NULL },
         { "51", NULL },
         { "52", NULL },
         { "53", NULL },
         { "54", NULL },
         { "55", NULL },
         { "56", NULL },
         { "57", NULL },
         { "58", NULL },
         { "59", NULL },
         { "60", NULL },
         { "61", NULL },
         { "62", NULL },
         { "63", NULL },
         { NULL, NULL },
      },
      "1"
   },
   {
      BEETLE_OPT(line_render),
      BEETLE_OPT_LINE_RENDER_LABEL_EO,
      NULL,
      BEETLE_OPT_LINE_RENDER_INFO_0_EO,
      NULL,
      "hacks",
      {
         { "default",    OPTION_VAL_DEFAULT_EO },
         { "aggressive", OPTION_VAL_AGGRESSIVE_EO },
         { "disabled",   NULL },
         { NULL, NULL },
      },
      "default"
   },
   {
      BEETLE_OPT(widescreen_hack),
      BEETLE_OPT_WIDESCREEN_HACK_LABEL_EO,
      NULL,
      BEETLE_OPT_WIDESCREEN_HACK_INFO_0_EO,
      NULL,
      "hacks",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(widescreen_hack_aspect_ratio),
      BEETLE_OPT_WIDESCREEN_HACK_ASPECT_RATIO_LABEL_EO,
      NULL,
      BEETLE_OPT_WIDESCREEN_HACK_ASPECT_RATIO_INFO_0_EO,
      NULL,
      "hacks",
      {
         { "16:9",  OPTION_VAL_16_9_EO },
         { "16:10", OPTION_VAL_16_10_EO },
         { "18:9",  OPTION_VAL_18_9_EO },
         { "19:9",  OPTION_VAL_19_9_EO },
         { "20:9",  OPTION_VAL_20_9_EO },
         { "21:9",  OPTION_VAL_21_9_EO }, // 64:27
         { "32:9",  OPTION_VAL_32_9_EO },
         { NULL,    NULL },
      },
      "16:9"
   },
#if defined(HAVE_LIGHTREC)
   {
      BEETLE_OPT(cpu_dynarec),
      BEETLE_OPT_CPU_DYNAREC_LABEL_EO,
      NULL,
      BEETLE_OPT_CPU_DYNAREC_INFO_0_EO,
      NULL,
      "hacks",
      {
         { "disabled",        OPTION_VAL_DISABLED_EO },
         { "execute",         OPTION_VAL_EXECUTE_EO },
         { "execute_one",     OPTION_VAL_EXECUTE_ONE_EO },
         { "run_interpreter", OPTION_VAL_RUN_INTERPRETER_EO },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(dynarec_invalidate),
      BEETLE_OPT_DYNAREC_INVALIDATE_LABEL_EO,
      NULL,
      BEETLE_OPT_DYNAREC_INVALIDATE_INFO_0_EO,
      NULL,
      "hacks",
      {
         { "full", OPTION_VAL_FULL_EO },
         { "dma",  OPTION_VAL_DMA_EO },
         { NULL, NULL },
      },
      "full"
   },
   {
      BEETLE_OPT(dynarec_eventcycles),
      BEETLE_OPT_DYNAREC_EVENTCYCLES_LABEL_EO,
      NULL,
      BEETLE_OPT_DYNAREC_EVENTCYCLES_INFO_0_EO,
      NULL,
      "hacks",
      {
         { "128",  OPTION_VAL_128_EO },
         { "256",  NULL },
         { "384",  NULL },
         { "512",  NULL },
         { "640",  NULL },
         { "768",  NULL },
         { "896",  NULL },
         { "1024", NULL },
         { "1152", NULL },
         { "1280", NULL },
         { "1408", NULL },
         { "1536", NULL },
         { "1664", NULL },
         { "1792", NULL },
         { "1920", NULL },
         { "2048", NULL },
         { NULL, NULL },
      },
      "128"
   },
   {
      BEETLE_OPT(dynarec_spu_samples),
      BEETLE_OPT_DYNAREC_SPU_SAMPLES_LABEL_EO,
      NULL,
      BEETLE_OPT_DYNAREC_SPU_SAMPLES_INFO_0_EO,
      NULL,
      "hacks",
      {
         { "1",  OPTION_VAL_1_O49_EO },
         { "4",  NULL },
         { "16", NULL },
         { NULL, NULL },
      },
      "1"
   },
#endif
   {
      BEETLE_OPT(core_timing_fps),
      BEETLE_OPT_CORE_TIMING_FPS_LABEL_EO,
      NULL,
      BEETLE_OPT_CORE_TIMING_FPS_INFO_0_EO,
      NULL,
      "video",
      {
         { "auto_toggle",       OPTION_VAL_AUTO_TOGGLE_EO },
         { "force_progressive", OPTION_VAL_FORCE_PROGRESSIVE_EO },
         { "force_interlaced",  OPTION_VAL_FORCE_INTERLACED_EO },
      },
      "auto_toggle"
   },
   {
      BEETLE_OPT(deinterlacer),
      BEETLE_OPT_DEINTERLACER_LABEL_EO,
      NULL,
      BEETLE_OPT_DEINTERLACER_INFO_0_EO,
      NULL,
      "video",
      {
         { "weave",      OPTION_VAL_WEAVE_EO },
         { "bob",        OPTION_VAL_BOB_EO },
         { "bob_offset", OPTION_VAL_BOB_OFFSET_EO },
         { "fastmad",    OPTION_VAL_FASTMAD_EO },
         { "off",        "Off" },
         { NULL, NULL },
      },
      "weave"
   },
   {
      BEETLE_OPT(frame_duping),
      BEETLE_OPT_FRAME_DUPING_LABEL_EO,
      NULL,
      BEETLE_OPT_FRAME_DUPING_INFO_0_EO,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(pal_video_timing_override),
      BEETLE_OPT_PAL_VIDEO_TIMING_OVERRIDE_LABEL_EO,
      NULL,
      BEETLE_OPT_PAL_VIDEO_TIMING_OVERRIDE_INFO_0_EO,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(aspect_ratio),
      BEETLE_OPT_ASPECT_RATIO_LABEL_EO,
      NULL,
      BEETLE_OPT_ASPECT_RATIO_INFO_0_EO,
      NULL,
      "video",
      {
         { "corrected", OPTION_VAL_CORRECTED_EO },
         { "uncorrected", OPTION_VAL_UNCORRECTED_EO },
         { "4:3",  OPTION_VAL_4_3_EO },
         { "ntsc", OPTION_VAL_NTSC_EO },
      },
      "corrected"
   },
   {
      BEETLE_OPT(crop_overscan),
      BEETLE_OPT_CROP_OVERSCAN_LABEL_EO,
      NULL,
      BEETLE_OPT_CROP_OVERSCAN_INFO_0_EO,
      NULL,
      "video",
      {
         { "disabled", "Disabled" },
         { "static",   OPTION_VAL_STATIC_EO },
         { "smart",    OPTION_VAL_SMART_EO },
         { NULL, NULL },
      },
      "smart"
   },
   {
      BEETLE_OPT(image_crop),
      BEETLE_OPT_IMAGE_CROP_LABEL_EO,
      NULL,
      BEETLE_OPT_IMAGE_CROP_INFO_0_EO,
      NULL,
      "video",
      {
         { "disabled", "0" },
         { "1px",      OPTION_VAL_1PX_EO },
         { "2px",      OPTION_VAL_2PX_EO },
         { "3px",      OPTION_VAL_3PX_EO },
         { "4px",      OPTION_VAL_4PX_EO },
         { "5px",      OPTION_VAL_5PX_EO },
         { "6px",      OPTION_VAL_6PX_EO },
         { "7px",      OPTION_VAL_7PX_EO },
         { "8px",      OPTION_VAL_8PX_EO },
         { "9px",      OPTION_VAL_9PX_EO },
         { "10px",     OPTION_VAL_10PX_EO },
         { "11px",     OPTION_VAL_11PX_EO },
         { "12px",     OPTION_VAL_12PX_EO },
         { "13px",     OPTION_VAL_13PX_EO },
         { "14px",     OPTION_VAL_14PX_EO },
         { "15px",     OPTION_VAL_15PX_EO },
         { "16px",     OPTION_VAL_16PX_EO },
         { "17px",     OPTION_VAL_17PX_EO },
         { "18px",     OPTION_VAL_18PX_EO },
         { "19px",     OPTION_VAL_19PX_EO },
         { "20px",     OPTION_VAL_20PX_EO },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(image_offset),
      BEETLE_OPT_IMAGE_OFFSET_LABEL_EO,
      NULL,
      BEETLE_OPT_IMAGE_OFFSET_INFO_0_EO,
      NULL,
      "video",
      {
         { "-12px",    OPTION_VAL_12PX_O64_EO },
         { "-11px",    OPTION_VAL_11PX_O64_EO },
         { "-10px",    OPTION_VAL_10PX_O64_EO },
         { "-9px",     OPTION_VAL_9PX_O64_EO },
         { "-8px",     OPTION_VAL_8PX_O64_EO },
         { "-7px",     OPTION_VAL_7PX_O64_EO },
         { "-6px",     OPTION_VAL_6PX_O64_EO },
         { "-5px",     OPTION_VAL_5PX_O64_EO },
         { "-4px",     OPTION_VAL_4PX_O64_EO },
         { "-3px",     OPTION_VAL_3PX_O64_EO },
         { "-2px",     OPTION_VAL_2PX_O64_EO },
         { "-1px",     OPTION_VAL_1PX_O64_EO },
         { "disabled", "0" },
         { "+1px",     OPTION_VAL_1PX_O64_0_EO },
         { "+2px",     OPTION_VAL_2PX_O64_0_EO },
         { "+3px",     OPTION_VAL_3PX_O64_0_EO },
         { "+4px",     OPTION_VAL_4PX_O64_0_EO },
         { "+5px",     OPTION_VAL_5PX_O64_0_EO },
         { "+6px",     OPTION_VAL_6PX_O64_0_EO },
         { "+7px",     OPTION_VAL_7PX_O64_0_EO },
         { "+8px",     OPTION_VAL_8PX_O64_0_EO },
         { "+9px",     OPTION_VAL_9PX_O64_0_EO },
         { "+10px",    OPTION_VAL_10PX_O64_0_EO },
         { "+11px",    OPTION_VAL_11PX_O64_0_EO },
         { "+12px",    OPTION_VAL_12PX_O64_0_EO },
         { NULL, NULL },
      },
      "disabled"
   },
#if defined(HAVE_OPENGL) || defined(HAVE_OPENGLES) || defined(HAVE_VULKAN)
   {
      BEETLE_OPT(image_offset_cycles),
      BEETLE_OPT_IMAGE_OFFSET_CYCLES_LABEL_EO,
      NULL,
      BEETLE_OPT_IMAGE_OFFSET_CYCLES_INFO_0_EO,
      NULL,
      "video",
      {
         { "-40",      NULL },
         { "-39",      NULL },
         { "-38",      NULL },
         { "-37",      NULL },
         { "-36",      NULL },
         { "-35",      NULL },
         { "-34",      NULL },
         { "-33",      NULL },
         { "-32",      NULL },
         { "-31",      NULL },
         { "-30",      NULL },
         { "-29",      NULL },
         { "-28",      NULL },
         { "-27",      NULL },
         { "-26",      NULL },
         { "-25",      NULL },
         { "-24",      NULL },
         { "-23",      NULL },
         { "-22",      NULL },
         { "-21",      NULL },
         { "-20",      NULL },
         { "-19",      NULL },
         { "-18",      NULL },
         { "-17",      NULL },
         { "-16",      NULL },
         { "-15",      NULL },
         { "-14",      NULL },
         { "-13",      NULL },
         { "-12",      NULL },
         { "-11",      NULL },
         { "-10",      NULL },
         { "-9",       NULL },
         { "-8",       NULL },
         { "-7",       NULL },
         { "-6",       NULL },
         { "-5",       NULL },
         { "-4",       NULL },
         { "-3",       NULL },
         { "-2",       NULL },
         { "-1",       NULL },
         { "0",        "0" },
         { "+1",       NULL },
         { "+2",       NULL },
         { "+3",       NULL },
         { "+4",       NULL },
         { "+5",       NULL },
         { "+6",       NULL },
         { "+7",       NULL },
         { "+8",       NULL },
         { "+9",       NULL },
         { "+10",      NULL },
         { "+11",      NULL },
         { "+12",      NULL },
         { "+13",      NULL },
         { "+14",      NULL },
         { "+15",      NULL },
         { "+16",      NULL },
         { "+17",      NULL },
         { "+18",      NULL },
         { "+19",      NULL },
         { "+20",      NULL },
         { "+21",      NULL },
         { "+22",      NULL },
         { "+23",      NULL },
         { "+24",      NULL },
         { "+25",      NULL },
         { "+26",      NULL },
         { "+27",      NULL },
         { "+28",      NULL },
         { "+29",      NULL },
         { "+30",      NULL },
         { "+31",      NULL },
         { "+32",      NULL },
         { "+33",      NULL },
         { "+34",      NULL },
         { "+35",      NULL },
         { "+36",      NULL },
         { "+37",      NULL },
         { "+38",      NULL },
         { "+39",      NULL },
         { "+40",      NULL },
         { NULL, NULL},
      },
      "0"
   },
#endif
   {
      BEETLE_OPT(initial_scanline),
      BEETLE_OPT_INITIAL_SCANLINE_LABEL_EO,
      NULL,
      BEETLE_OPT_INITIAL_SCANLINE_INFO_0_EO,
      NULL,
      "video",
      {
         { "0",  OPTION_VAL_0_O48_EO },
         { "1",  NULL },
         { "2",  NULL },
         { "3",  NULL },
         { "4",  NULL },
         { "5",  NULL },
         { "6",  NULL },
         { "7",  NULL },
         { "8",  NULL },
         { "9",  NULL },
         { "10", NULL },
         { "11", NULL },
         { "12", NULL },
         { "13", NULL },
         { "14", NULL },
         { "15", NULL },
         { "16", NULL },
         { "17", NULL },
         { "18", NULL },
         { "19", NULL },
         { "20", NULL },
         { "21", NULL },
         { "22", NULL },
         { "23", NULL },
         { "24", NULL },
         { "25", NULL },
         { "26", NULL },
         { "27", NULL },
         { "28", NULL },
         { "29", NULL },
         { "30", NULL },
         { "31", NULL },
         { "32", NULL },
         { "33", NULL },
         { "34", NULL },
         { "35", NULL },
         { "36", NULL },
         { "37", NULL },
         { "38", NULL },
         { "39", NULL },
         { "40", NULL },
         { NULL, NULL },
      },
      "0"
   },
   {
      BEETLE_OPT(last_scanline),
      BEETLE_OPT_LAST_SCANLINE_LABEL_EO,
      NULL,
      BEETLE_OPT_LAST_SCANLINE_INFO_0_EO,
      NULL,
      "video",
      {
         { "210", NULL },
         { "211", NULL },
         { "212", NULL },
         { "213", NULL },
         { "214", NULL },
         { "215", NULL },
         { "216", NULL },
         { "217", NULL },
         { "218", NULL },
         { "219", NULL },
         { "220", NULL },
         { "221", NULL },
         { "222", NULL },
         { "223", NULL },
         { "224", NULL },
         { "225", NULL },
         { "226", NULL },
         { "227", NULL },
         { "228", NULL },
         { "229", NULL },
         { "230", NULL },
         { "231", NULL },
         { "232", NULL },
         { "233", NULL },
         { "234", NULL },
         { "235", NULL },
         { "236", NULL },
         { "237", NULL },
         { "238", NULL },
         { "239", OPTION_VAL_239_EO },
         { NULL, NULL },
      },
      "239"
   },
   {
      BEETLE_OPT(initial_scanline_pal),
      BEETLE_OPT_INITIAL_SCANLINE_PAL_LABEL_EO,
      NULL,
      BEETLE_OPT_INITIAL_SCANLINE_PAL_INFO_0_EO,
      NULL,
      "video",
      {
         { "0",  OPTION_VAL_0_O48_EO },
         { "1",  NULL },
         { "2",  NULL },
         { "3",  NULL },
         { "4",  NULL },
         { "5",  NULL },
         { "6",  NULL },
         { "7",  NULL },
         { "8",  NULL },
         { "9",  NULL },
         { "10", NULL },
         { "11", NULL },
         { "12", NULL },
         { "13", NULL },
         { "14", NULL },
         { "15", NULL },
         { "16", NULL },
         { "17", NULL },
         { "18", NULL },
         { "19", NULL },
         { "20", NULL },
         { "21", NULL },
         { "22", NULL },
         { "23", NULL },
         { "24", NULL },
         { "25", NULL },
         { "26", NULL },
         { "27", NULL },
         { "28", NULL },
         { "29", NULL },
         { "30", NULL },
         { "31", NULL },
         { "32", NULL },
         { "33", NULL },
         { "34", NULL },
         { "35", NULL },
         { "36", NULL },
         { "37", NULL },
         { "38", NULL },
         { "39", NULL },
         { "40", NULL },
         { NULL, NULL },
      },
      "0"
   },
   {
      BEETLE_OPT(last_scanline_pal),
      BEETLE_OPT_LAST_SCANLINE_PAL_LABEL_EO,
      NULL,
      BEETLE_OPT_LAST_SCANLINE_PAL_INFO_0_EO,
      NULL,
      "video",
      {
         { "230", NULL },
         { "231", NULL },
         { "232", NULL },
         { "233", NULL },
         { "234", NULL },
         { "235", NULL },
         { "236", NULL },
         { "237", NULL },
         { "238", NULL },
         { "239", NULL },
         { "240", NULL },
         { "241", NULL },
         { "242", NULL },
         { "243", NULL },
         { "244", NULL },
         { "245", NULL },
         { "246", NULL },
         { "247", NULL },
         { "248", NULL },
         { "249", NULL },
         { "250", NULL },
         { "251", NULL },
         { "252", NULL },
         { "253", NULL },
         { "254", NULL },
         { "255", NULL },
         { "256", NULL },
         { "257", NULL },
         { "258", NULL },
         { "259", NULL },
         { "260", NULL },
         { "261", NULL },
         { "262", NULL },
         { "263", NULL },
         { "264", NULL },
         { "265", NULL },
         { "266", NULL },
         { "267", NULL },
         { "268", NULL },
         { "269", NULL },
         { "270", NULL },
         { "271", NULL },
         { "272", NULL },
         { "273", NULL },
         { "274", NULL },
         { "275", NULL },
         { "276", NULL },
         { "277", NULL },
         { "278", NULL },
         { "279", NULL },
         { "280", NULL },
         { "281", NULL },
         { "282", NULL },
         { "283", NULL },
         { "284", NULL },
         { "285", NULL },
         { "286", NULL },
         { "287", OPTION_VAL_287_EO },
         { NULL, NULL },
      },
      "287"
   },
   { NULL, NULL, NULL, NULL, NULL, NULL, {{0}}, NULL },
};
struct retro_core_options_v2 options_eo = {
   option_cats_eo,
   option_defs_eo
};

/* RETRO_LANGUAGE_ES */

#define CATEGORY_SYSTEM_LABEL_ES "Sistema"
#define CATEGORY_SYSTEM_INFO_0_ES "Cambia las opciones del sistema y de velocidad de procesamiento («overclock»)."
#define CATEGORY_VIDEO_LABEL_ES "Vídeo"
#define CATEGORY_VIDEO_INFO_0_ES "Cambia las opciones de relación de aspecto, recorte de imagen, filtros de vídeo y omisión de fotogramas."
#define CATEGORY_PGXP_LABEL_ES "PGXP (Proceso de transformación precisa de geometría)"
#define CATEGORY_PGXP_INFO_0_ES "Estas opciones permiten mejorar los gráficos por encima de lo que puede mostrar la consola real, eliminando las distorsiones de texturas y el «Z-fighting» (distorsión entre polígonos)."
#define CATEGORY_OSD_LABEL_ES "Presentación en pantalla"
#define CATEGORY_OSD_INFO_0_ES "Cambia las notificaciones que se mostrarán en pantalla."
#define CATEGORY_INPUT_LABEL_ES "Entrada"
#define CATEGORY_INPUT_INFO_0_ES "Cambia los ajustes de las pistolas de luz, el ratón y el neGcon."
#define CATEGORY_MEMCARDS_LABEL_ES "Memory Cards"
#define CATEGORY_MEMCARDS_INFO_0_ES "Cambia las opciones de las Memory Card virtuales."
#define CATEGORY_HACKS_LABEL_ES "Arreglos de emulación"
#define CATEGORY_HACKS_INFO_0_ES "Cambia las opciones de precisión de la emulación, las cuales afectan al rendimiento a bajo nivel y a la compatibilidad."
#define BEETLE_OPT_CPU_FREQ_SCALE_LABEL_ES "Escalado de frecuencia de la CPU (velocidad)"
#define BEETLE_OPT_CPU_FREQ_SCALE_INFO_0_ES "Aumenta (o disminuye) la velocidad del reloj emulado de la CPU de PSX. Una velocidad superior puede eliminar las ralentizaciones y mejorar la velocidad de fotogramas en ciertos juegos a costa de aumentar los requisitos de rendimiento. Ten en cuenta que ciertos juegos incluyen un limitador de fotogramas interno y no ganarán beneficio alguno al aumentar la velocidad. Puede provocar que ciertos efectos vayan más rápido de lo normal en ciertos juegos."
#define OPTION_VAL_50_ES "50 %"
#define OPTION_VAL_60_ES "60 %"
#define OPTION_VAL_70_ES "70 %"
#define OPTION_VAL_80_ES "80 %"
#define OPTION_VAL_90_ES "90 %"
#define OPTION_VAL_100_ES "100 % (nativa)"
#define OPTION_VAL_110_ES "110 %"
#define OPTION_VAL_120_ES "120 %"
#define OPTION_VAL_130_ES "130 %"
#define OPTION_VAL_140_ES "140 %"
#define OPTION_VAL_150_ES "150 %"
#define OPTION_VAL_160_ES "160 %"
#define OPTION_VAL_170_ES "170 %"
#define OPTION_VAL_180_ES "180 %"
#define OPTION_VAL_190_ES "190 %"
#define OPTION_VAL_200_ES "200 %"
#define OPTION_VAL_210_ES "210 %"
#define OPTION_VAL_220_ES "220 %"
#define OPTION_VAL_230_ES "230 %"
#define OPTION_VAL_240_ES "240 %"
#define OPTION_VAL_250_ES "250 %"
#define OPTION_VAL_260_ES "260 %"
#define OPTION_VAL_270_ES "270 %"
#define OPTION_VAL_280_ES "280 %"
#define OPTION_VAL_290_ES "290 %"
#define OPTION_VAL_300_ES "300 %"
#define OPTION_VAL_310_ES "310 %"
#define OPTION_VAL_320_ES "320 %"
#define OPTION_VAL_330_ES "330 %"
#define OPTION_VAL_340_ES "340 %"
#define OPTION_VAL_350_ES "350 %"
#define OPTION_VAL_360_ES "360 %"
#define OPTION_VAL_370_ES "370 %"
#define OPTION_VAL_380_ES "380 %"
#define OPTION_VAL_390_ES "390 %"
#define OPTION_VAL_400_ES "400 %"
#define OPTION_VAL_410_ES "410 %"
#define OPTION_VAL_420_ES "420 %"
#define OPTION_VAL_430_ES "430 %"
#define OPTION_VAL_440_ES "440 %"
#define OPTION_VAL_450_ES "450 %"
#define OPTION_VAL_460_ES "460 %"
#define OPTION_VAL_470_ES "470 %"
#define OPTION_VAL_480_ES "480 %"
#define OPTION_VAL_490_ES "490 %"
#define OPTION_VAL_500_ES "500 %"
#define OPTION_VAL_510_ES "510 %"
#define OPTION_VAL_520_ES "520 %"
#define OPTION_VAL_530_ES "530 %"
#define OPTION_VAL_540_ES "540 %"
#define OPTION_VAL_550_ES "550 %"
#define OPTION_VAL_560_ES "560 %"
#define OPTION_VAL_570_ES "570 %"
#define OPTION_VAL_580_ES "580 %"
#define OPTION_VAL_590_ES "590 %"
#define OPTION_VAL_600_ES "600 %"
#define OPTION_VAL_610_ES "610 %"
#define OPTION_VAL_620_ES "620 %"
#define OPTION_VAL_630_ES "630 %"
#define OPTION_VAL_640_ES "640 %"
#define OPTION_VAL_650_ES "650 %"
#define OPTION_VAL_660_ES "660 %"
#define OPTION_VAL_670_ES "670 %"
#define OPTION_VAL_680_ES "680 %"
#define OPTION_VAL_690_ES "690 %"
#define OPTION_VAL_700_ES "700 %"
#define OPTION_VAL_710_ES "710 %"
#define OPTION_VAL_720_ES "720 %"
#define OPTION_VAL_730_ES "730 %"
#define OPTION_VAL_740_ES "740 %"
#define OPTION_VAL_750_ES "750 %"
#define BEETLE_OPT_GPU_OVERCLOCK_LABEL_ES "Acelerar el rasterizador de la GPU"
#define BEETLE_OPT_GPU_OVERCLOCK_INFO_0_ES "Aumenta la velocidad del rasterizador 2D incluido en la GPU emulada de PSX. No mejora el renderizado 3D y su efecto suele ser limitado."
#define OPTION_VAL_1X_NATIVE_ES "1x (nativa)"
#define OPTION_VAL_2X_ES NULL
#define OPTION_VAL_4X_ES NULL
#define OPTION_VAL_8X_ES NULL
#define OPTION_VAL_16X_ES NULL
#define OPTION_VAL_32X_ES NULL
#define BEETLE_OPT_GTE_OVERCLOCK_LABEL_ES "Acelerar el procesador GTE"
#define BEETLE_OPT_GTE_OVERCLOCK_INFO_0_ES "Frena todas las operaciones emuladas del GTE (el coprocesador de la CPU para gráficos 3D) a una latencia constante de un ciclo. Aquellos juegos que utilicen en gran medida el GTE podrían mejorar su velocidad de fotogramas y la estabilidad en la duración de los mismos."
#define BEETLE_OPT_CD_ACCESS_METHOD_LABEL_ES "Método de acceso al CD"
#define BEETLE_OPT_CD_ACCESS_METHOD_INFO_0_ES "«Sincrónico» simula el hardware original, «Asincrónico» puede reducir los tirones en aquellos dispositivos que tengan un almacenamiento lento y «Precacheado» carga la imagen de disco entera en memoria al ejecutar el contenido, lo que puede mejorar los tiempos de carga del juego a costa de retrasar el arranque inicial. «Precacheado» puede producir problemas en sistemas con poca RAM. Es necesario reiniciar."
#define OPTION_VAL_SYNC_ES "Sincrónico"
#define OPTION_VAL_ASYNC_ES "Asincrónico"
#define OPTION_VAL_PRECACHE_ES "Precacheado"
#define BEETLE_OPT_CD_FASTLOAD_LABEL_ES "Velocidad de lectura del CD"
#define BEETLE_OPT_CD_FASTLOAD_INFO_0_ES "Un valor superior a «2x (nativa)» puede reducir en gran medida los tiempos de carga de los juegos, pero produciendo errores de sincronía. Algunos juegos podrían no funcionar correctamente si este valor es superior a cierta cifra."
#define OPTION_VAL_2X_NATIVE_ES "2x (nativa)"
#define OPTION_VAL_6X_ES NULL
#define OPTION_VAL_10X_ES NULL
#define OPTION_VAL_12X_ES NULL
#define OPTION_VAL_14X_ES NULL
#define BEETLE_OPT_SKIP_BIOS_LABEL_ES "Omitir BIOS"
#define BEETLE_OPT_SKIP_BIOS_INFO_0_ES "Omite la animación de arranque de la BIOS de PlayStation que suele aparecer al cargar un contenido. Nota: Esta opción puede generar problemas de compatibilidad en varios juegos (juegos PAL con protección anticopia, Saga Frontier, etc.)."
#define BEETLE_OPT_OVERRIDE_BIOS_LABEL_ES "Reemplazar BIOS"
#define BEETLE_OPT_OVERRIDE_BIOS_INFO_0_ES "Reemplaza la BIOS correspondiente a la región del contenido con una de región libre (si está disponible). Es necesario reiniciar."
#define OPTION_VAL_PSXONPSP_ES "BIOS de PS1 usada en PSP"
#define OPTION_VAL_PS1_ROM_ES "BIOS de PS1 usada en PS3"
#define OPTION_VAL_OPENBIOS_ES NULL
#define BEETLE_OPT_REGION_LABEL_ES "Región del sistema"
#define BEETLE_OPT_REGION_INFO_0_ES "Establece la región alternativa para cuando no se pueda detectar automáticamente el contenido (p. ej.: con archivos PS-X EXE en bruto). Los juegos en formato de disco utilizarán la región incluida en el mismo. Es necesario reiniciar."
#define OPTION_VAL_AUTO_ES "Automático"
#define OPTION_VAL_NTSC_J_ES "NTSC-J (Japón)"
#define OPTION_VAL_NTSC_U_ES "NTSC-U (Norteamérica)"
#define OPTION_VAL_PAL_ES "PAL (Europa)"
#define BEETLE_OPT_INTERNAL_RESOLUTION_LABEL_ES "Resolución interna de la GPU"
#define BEETLE_OPT_INTERNAL_RESOLUTION_INFO_0_ES "Toda resolución superior a «1x (Nativa)» mejorarán la fidelidad de los modelos 3D a costa de aumentar los requisitos de rendimiento. Los elementos 2D no suelen verse afectados por este ajuste."
#define BEETLE_OPT_RENDERER_LABEL_ES "Renderizador"
#define BEETLE_OPT_RENDERER_INFO_0_ES "El renderizador por software es el más preciso, pero sus requisitos son muy elevados si se ejecuta a una resolución interna superior a la de la GPU. Los renderizadores por hardware son menos precisos, pero tienen un mejor rendimiento al usar resoluciones internas superiores y permiten varias mejoras gráficas. «Hardware (Automático)» seleccionará automáticamente el renderizador Vulkan u OpenGL según el controlador de vídeo actual del front-end de libretro. Si el controlador no es compatible con Vulkan o con OpenGL 3.3, se utilizará el renderizador por software como reserva del núcleo. Es necesario reiniciar."
#define OPTION_VAL_HARDWARE_ES "Hardware (automático)"
#define OPTION_VAL_HARDWARE_GL_ES NULL
#define OPTION_VAL_HARDWARE_VK_ES NULL
#define OPTION_VAL_SOFTWARE_ES NULL
#define BEETLE_OPT_RENDERER_SOFTWARE_FB_LABEL_ES "Framebuffer por software"
#define BEETLE_OPT_RENDERER_SOFTWARE_FB_INFO_0_ES "Permite emular con precisión los efectos del framebuffer (p. ej.: desenfoque de movimiento, la espiral de batallas de FF7) al utilizar los renderizadores por hardware ejecutando una copia del renderizador por software a resolución nativa en un segundo plano. Al desactivar esta opción se omitirán estas operaciones (OpenGL) o se renderizarán en la GPU (Vulkan). Desactivar esta opción puede mejorar el rendimiento a costa de generar errores gráficos graves. En caso de duda, deja esta opción activada."
#define BEETLE_OPT_DEPTH_LABEL_ES "Profundidad interna de colores"
#define BEETLE_OPT_DEPTH_INFO_0_ES "Una mayor profundidad puede reducir el efecto de bandas de color sin tener que recurrir al efecto de tramado. 16 bpp simula el hardware original, pero puede generar bandas de color si la opción de tramado no está activada. Se recomienda desactivar la opción «Tramado» al seleccionar 32 bpp en esta opción."
#define OPTION_VAL_16BPP_NATIVE_ES "16 bpp (nativa)"
#define OPTION_VAL_32BPP_ES NULL
#define BEETLE_OPT_DITHER_MODE_LABEL_ES "Tramado"
#define BEETLE_OPT_DITHER_MODE_INFO_0_ES "«1x (Nativo)» simula el tramado nativo a baja resolución que utilizaba el hardware original para suavizar los defectos de bandas de color visibles en la profundidad de colores nativa. «Resolución interna» cambia la escala del tramado a la resolución interna configurada para producir un resultado más limpio. Se recomienda desactivar esta opción al utilizar una profundidad de colores de 32 bpp. Nota: Activar esta opción en Vulkan desactivará la reducción de colores a la profundidad nativa y activará una profundidad interna superior."
#define OPTION_VAL_INTERNAL_RESOLUTION_ES "Resolución interna"
#define BEETLE_OPT_SCALED_UV_OFFSET_LABEL_ES "Compensación de UV de texturas"
#define BEETLE_OPT_SCALED_UV_OFFSET_INFO_0_ES "Muestrea las texturas de los polígonos 3D con una compensación para resoluciones superiores a la nativa. Reduce los salientes en las texturas, pero puede producir fallos gráficos accidentales."
#define BEETLE_OPT_FILTER_LABEL_ES "Filtrado de texturas"
#define BEETLE_OPT_FILTER_INFO_0_ES "«Más cercano» simula el hardware original. «Bilineal» y «De tres puntos» son filtros de suavizado que reducen la pixelación difuminando la imagen. «SABR», «xBR» y «JINC2» son filtros de escalado que podrían mejorar la fidelidad/definición de las texturas a costa de aumentar los requisitos de rendimiento. Solo funciona con los renderizadores por hardware."
#define OPTION_VAL_NEAREST_ES "Más cercano"
#define OPTION_VAL_SABR_ES NULL
#define OPTION_VAL_XBR_ES NULL
#define OPTION_VAL_BILINEAR_ES "Bilineal"
#define OPTION_VAL_3_POINT_ES "De tres puntos"
#define OPTION_VAL_JINC2_ES NULL
#define BEETLE_OPT_FILTER_EXCLUDE_SPRITE_LABEL_ES "No filtrar sprites"
#define BEETLE_OPT_FILTER_EXCLUDE_SPRITE_INFO_0_ES "No aplicará el filtrado de texturas a los sprites. Evita los bordes en varios juegos con fondos renderizados en 2D. Se recomienda activar también la opción de Suavizado adaptativo u otro efecto de posprocesado para conseguir el mejor efecto."
#define OPTION_VAL_OPAQUE_ES "Solo sprites opacos"
#define OPTION_VAL_ALL_ES "Sprites opacos y semitransparentes"
#define BEETLE_OPT_FILTER_EXCLUDE_2D_POLYGON_LABEL_ES "No filtrar polígonos 2D"
#define BEETLE_OPT_FILTER_EXCLUDE_2D_POLYGON_INFO_0_ES "No aplicará el filtrado de texturas a los polígonos en 2D. Estos son detectados mediante un proceso heurístico, por lo que puede haber fallos visuales. Se recomienda activar también la opción «Suavizado adaptativo» u otro efecto de posprocesado para conseguir el mejor efecto."
#define BEETLE_OPT_ADAPTIVE_SMOOTHING_LABEL_ES "Suavizado adaptativo"
#define BEETLE_OPT_ADAPTIVE_SMOOTHING_INFO_0_ES "Suaviza los gráficos y elementos de interfaz en 2D sin difuminar los objetos renderizados en 3D. Solo funciona con el renderizador Vulkan."
#define BEETLE_OPT_SUPER_SAMPLING_LABEL_ES "Supermuestreo (reducción a resolución nativa)"
#define BEETLE_OPT_SUPER_SAMPLING_INFO_0_ES "Reduce el contenido renderizado de una resolución interna superior a la nativa. Al combinar este efecto con multiplicadores para aumentar la resolución interna, los juegos pueden mostrarse con objetos 3D con bordes suavizados en la resolución nativa. Genera los mejores resultados en aquellos títulos que combinan elementos 2D y 3D (p. ej.: personajes 3D sobre fondos prerenderizados) y acompañado de shaders de simulación de CRT. Solo funciona con el renderizador Vulkan. Nota: Se recomienda desactivar el tramado al utilizar esta opción."
#define BEETLE_OPT_MSAA_LABEL_ES "Suavizado de bordes MSAA"
#define BEETLE_OPT_MSAA_INFO_0_ES "MSAA mejorará el aspecto de los objetos 3D. Solo funciona con el renderizador Vulkan."
#define OPTION_VAL_1X_ES NULL
#define BEETLE_OPT_MDEC_YUV_LABEL_ES "Filtro cromático YUV para MDEC"
#define BEETLE_OPT_MDEC_YUV_INFO_0_ES "Mejora la calidad de la reproducción de secuencias FMV reduciendo los artefactos de los «macrobloques» (cuadrados y bordes de sierra). Solo funciona con el renderizador Vulkan."
#define BEETLE_OPT_TRACK_TEXTURES_LABEL_ES "Rastreo de texturas"
#define BEETLE_OPT_TRACK_TEXTURES_INFO_0_ES "Necesario para el volcado y sustitución de texturas. Probablemente provocará que la mayoría de juegos se queden colgados."
#define BEETLE_OPT_DUMP_TEXTURES_LABEL_ES "Volcar texturas"
#define BEETLE_OPT_DUMP_TEXTURES_INFO_0_ES "Vuelca las texturas que se utilicen a <cd>-texture-dump/"
#define BEETLE_OPT_REPLACE_TEXTURES_LABEL_ES "Sustituir texturas"
#define BEETLE_OPT_REPLACE_TEXTURES_INFO_0_ES "Sustituye las texturas con las versiones en alta definición que se encuentren en <cd>-texture-replacements/"
#define BEETLE_OPT_PGXP_MODE_LABEL_ES "Modo de operación del PGXP"
#define BEETLE_OPT_PGXP_MODE_INFO_0_ES "Renderiza los objetos 3D con una precisión subpíxel, utilizando coordenadas de vértices de coma fija para minimizar la distorsión y temblores en objetos 3D que tenía el hardware original. «Solo en memoria» apenas da problemas de compatibilidad y se recomienda su uso general. «Memoria + CPU (inestable)» puede reducir aún más las distorsiones visuales, pero tiene unos requisitos de rendimiento elevados y puede provocar fallos en la geometría."
#define OPTION_VAL_MEMORY_ONLY_ES "Solo en memoria"
#define OPTION_VAL_MEMORY_CPU_ES "Memoria + CPU (inestable)"
#define BEETLE_OPT_PGXP_2D_TOL_LABEL_ES "Tolerancia a la geometría 2D del PGXP"
#define BEETLE_OPT_PGXP_2D_TOL_INFO_0_ES "Oculta los errores más evidentes de las operaciones del PGXP: este valor especifica la tolerancia con la que se conservarán los valores del PGXP en aquellas geometrías que no dispongan de la información de profundidad necesaria."
#define OPTION_VAL_0PX_ES "0 px"
#define OPTION_VAL_1PX_ES "1 px"
#define OPTION_VAL_2PX_ES "2 px"
#define OPTION_VAL_3PX_ES "3 px"
#define OPTION_VAL_4PX_ES "4 px"
#define OPTION_VAL_5PX_ES "5 px"
#define OPTION_VAL_6PX_ES "6 px"
#define OPTION_VAL_7PX_ES "7 px"
#define OPTION_VAL_8PX_ES "8 px"
#define BEETLE_OPT_PGXP_NCLIP_LABEL_ES "Selección de primitivos del PGXP"
#define BEETLE_OPT_PGXP_NCLIP_INFO_0_ES "Utiliza la implementación de NCLIP del PGXP. Mejora el aspecto visual tapando los agujeros de aquellas geometrías que tengan coordenadas PGXP. Provoca cuelgues en ciertos juegos y circunstancias."
#define BEETLE_OPT_PGXP_VERTEX_LABEL_ES "Caché de vértices del PGXP"
#define BEETLE_OPT_PGXP_VERTEX_INFO_0_ES "Cachea las posiciones de vértices mejoradas por el PGXP sean cacheadas para reutilizarlas al dibujar polígonos. Puede mejorar la alineación de los objetos y reducir los bordes, pero los falsos positivos a la hora de recurrir a la caché pueden producir defectos visuales. Actualmente se recomienda desactivar esta opción. Solo se aplicará cuando se active el modo de operación del PGXP. Solo funciona con los renderizadores por hardware."
#define BEETLE_OPT_PGXP_TEXTURE_LABEL_ES "Texturizado fiel a la perspectiva del PGXP"
#define BEETLE_OPT_PGXP_TEXTURE_INFO_0_ES "Sustituye el mapeado de texturas afín nativo de PSX con un mapeado de texturas fiel a la perspectiva. Elimina las distorsiones posicionales y la deformación de texturas, produciendo texturas correctamente alineadas. Solo se aplicará cuando se active el modo de operación del PGXP. Compatible con los renderizadores por hardware y software."
#define BEETLE_OPT_DISPLAY_INTERNAL_FPS_LABEL_ES "Mostrar FPS internos"
#define BEETLE_OPT_DISPLAY_INTERNAL_FPS_INFO_0_ES "Muestra la velocidad de fotogramas interna con la que el sistema PlayStation emulado renderiza el contenido. Nota: Es necesario activar las notificaciones en pantalla en el front-end libretro."
#define BEETLE_OPT_DISPLAY_OSD_LABEL_ES "Mostrar mensajes en pantalla"
#define BEETLE_OPT_DISPLAY_OSD_INFO_0_ES "Muestra en pantalla los mensajes que genere el núcleo."
#define BEETLE_OPT_DISPLAY_VRAM_LABEL_ES "Mostrar la VRAM completa (depuración)"
#define BEETLE_OPT_DISPLAY_VRAM_INFO_0_ES "Visualiza toda la VRAM emulada. Solo funciona con los renderizadores por hardware de OpenGL y Vulkan. Nota: Esta opción solo se utiliza para depuración y debería mantenerse desactivada por norma general."
#define BEETLE_OPT_ANALOG_CALIBRATION_LABEL_ES "Autocalibrar controles analógicos"
#define BEETLE_OPT_ANALOG_CALIBRATION_INFO_0_ES "Cuando el dispositivo de entrada esté configurado como DualShock, mando analógico, joystick analógico o neGcon, permite calibrar de forma dinámica las entradas analógicas. Se controlarán en tiempo real los valores de entrada máximos registrados para escalar las coordenadas analógicas que se transmitan al emulador. Esto sirve para juegos como Mega Man Legends 2, que esperan unos valores superiores a los que pueden transmitir los mandos modernos. Se recomienda que los sticks analógicos sean girados en toda su extensión nada más cargar un contenido para que el algoritmo de calibración funcione lo mejor posible."
#define BEETLE_OPT_ANALOG_TOGGLE_LABEL_ES "Alternar modo analógico del DualShock"
#define BEETLE_OPT_ANALOG_TOGGLE_INFO_0_ES "Cuando el tipo de dispositivo de entrada sea DualShock, esta opción permite que el DualShock emulado pueda alternar entre los modos digital y analógico, como hace el hardware original. También se puede alternar el modo analógico con la combinación de botones seleccionada."
#define OPTION_VAL_ENABLED_ANALOG_ES "Analógico (predeterminado)"
#define BEETLE_OPT_ANALOG_TOGGLE_COMBO_LABEL_ES "Combinación de botones del modo analógico del DualShock"
#define BEETLE_OPT_ANALOG_TOGGLE_COMBO_INFO_0_ES "La combinación de botones con la que se alternarán los modos digital y analógico del DualShock emulado."
#define OPTION_VAL_L1_L2_R1_R2_START_SELECT_ES NULL
#define OPTION_VAL_L1_R1_SELECT_ES NULL
#define OPTION_VAL_L1_R1_START_ES NULL
#define OPTION_VAL_L1_R1_L3_ES NULL
#define OPTION_VAL_L1_R1_R3_ES NULL
#define OPTION_VAL_L2_R2_SELECT_ES NULL
#define OPTION_VAL_L2_R2_START_ES NULL
#define OPTION_VAL_L2_R2_L3_ES NULL
#define OPTION_VAL_L2_R2_R3_ES NULL
#define OPTION_VAL_L3_R3_ES NULL
#define BEETLE_OPT_ANALOG_TOGGLE_HOLD_LABEL_ES "Tiempo de espera para la combinación del modo analógico"
#define BEETLE_OPT_ANALOG_TOGGLE_HOLD_INFO_0_ES "El tiempo que se deben mantener pulsados los botones de la combinación del modo analógico."
#define OPTION_VAL_0_ES "Espera de 0 segundos"
#define OPTION_VAL_1_ES "Espera de 1 segundo"
#define OPTION_VAL_2_ES "Espera de 2 segundos"
#define OPTION_VAL_3_ES "Espera de 3 segundos"
#define OPTION_VAL_4_ES "Espera de 4 segundos"
#define OPTION_VAL_5_ES "Espera de 5 segundos"
#define BEETLE_OPT_ENABLE_MULTITAP_PORT1_LABEL_ES "Multitap del puerto 1"
#define BEETLE_OPT_ENABLE_MULTITAP_PORT1_INFO_0_ES "Activa el sistema de Multitap en el puerto 1."
#define BEETLE_OPT_ENABLE_MULTITAP_PORT2_LABEL_ES "Multitap del puerto 2"
#define BEETLE_OPT_ENABLE_MULTITAP_PORT2_INFO_0_ES "Activa el sistema de Multitap en el puerto 2."
#define BEETLE_OPT_GUN_INPUT_MODE_LABEL_ES "Modo de entrada de pistolas de luz"
#define BEETLE_OPT_GUN_INPUT_MODE_INFO_0_ES "Utiliza una entrada de ratón o de pantalla táctil cuando el tipo de dispositivo esté configurado como «Guncon/G-Con 45» o «Justifier»."
#define OPTION_VAL_LIGHTGUN_ES "Pistola de luz"
#define OPTION_VAL_TOUCHSCREEN_ES "Pantalla táctil"
#define BEETLE_OPT_GUN_CURSOR_LABEL_ES "Cursor del arma"
#define BEETLE_OPT_GUN_CURSOR_INFO_0_ES "Muestra un cursor en pantalla al utilizar los tipos de dispositivo de entrada «Guncon/G-Con 45» o «Justifier»."
#define OPTION_VAL_CROSS_ES "Cruz"
#define OPTION_VAL_DOT_ES "Punto"
#define BEETLE_OPT_CROSSHAIR_COLOR_P1_LABEL_ES "Color de punto de mira de pistola: puerto 1"
#define BEETLE_OPT_CROSSHAIR_COLOR_P1_INFO_0_ES "El color del punto de mira de la pistola de luz conectada en el puerto 1."
#define OPTION_VAL_RED_ES "Rojo"
#define OPTION_VAL_BLUE_ES "Azul"
#define OPTION_VAL_GREEN_ES "Verde"
#define OPTION_VAL_ORANGE_ES "Naranja"
#define OPTION_VAL_YELLOW_ES "Amarillo"
#define OPTION_VAL_CYAN_ES "Cian"
#define OPTION_VAL_PINK_ES "Rosa"
#define OPTION_VAL_PURPLE_ES "Violeta"
#define OPTION_VAL_BLACK_ES "Negro"
#define OPTION_VAL_WHITE_ES "Blanco"
#define BEETLE_OPT_CROSSHAIR_COLOR_P2_LABEL_ES "Color de punto de mira de pistola: puerto 2"
#define BEETLE_OPT_CROSSHAIR_COLOR_P2_INFO_0_ES "El color del punto de mira de la pistola de luz conectada en el puerto 2."
#define BEETLE_OPT_MOUSE_SENSITIVITY_LABEL_ES "Sensibilidad del ratón"
#define BEETLE_OPT_MOUSE_SENSITIVITY_INFO_0_ES "La respuesta del dispositivo de entrada «Mouse (ratón)»."
#define OPTION_VAL_5_O42_ES "5 %"
#define OPTION_VAL_10_ES "10 %"
#define OPTION_VAL_15_ES "15 %"
#define OPTION_VAL_20_ES "20 %"
#define OPTION_VAL_25_ES "25 %"
#define OPTION_VAL_30_ES "30 %"
#define OPTION_VAL_35_ES "35 %"
#define OPTION_VAL_40_ES "40 %"
#define OPTION_VAL_45_ES "45 %"
#define OPTION_VAL_55_ES "55 %"
#define OPTION_VAL_65_ES "65 %"
#define OPTION_VAL_75_ES "75 %"
#define OPTION_VAL_85_ES "85 %"
#define OPTION_VAL_95_ES "95 %"
#define OPTION_VAL_100_O42_ES "100 %"
#define OPTION_VAL_105_ES "105 %"
#define OPTION_VAL_115_ES "115 %"
#define OPTION_VAL_125_ES "125 %"
#define OPTION_VAL_135_ES "135 %"
#define OPTION_VAL_145_ES "145 %"
#define OPTION_VAL_155_ES "155 %"
#define OPTION_VAL_165_ES "165 %"
#define OPTION_VAL_175_ES "175 %"
#define OPTION_VAL_185_ES "185 %"
#define OPTION_VAL_195_ES "195 %"
#define BEETLE_OPT_NEGCON_RESPONSE_LABEL_ES "Respuesta al torcer el neGcon"
#define BEETLE_OPT_NEGCON_RESPONSE_INFO_0_ES "La respuesta del stick analógico izquierdo al simular la acción de torcer un dispositivo de entrada neGcon emulado. El desplazamiento del stick analógico puede ser asignado a la rotación del neGcon de forma lineal, cuadrática o cúbica. «Cuadrática» da una mayor precisión que «Lineal» en movimientos más sutiles. «Cúbica» aumenta más la precisión de los movimientos más sutiles, pero exagera los más largos. Nota: solo se recomienda usar «Lineal» con periféricos tipo volante de carreras. Los mandos convencionales gestionan la entrada analógica de una forma muy distinta al mecanismo de rotación del neGcon de tal forma que la asignación lineal sobreamplifica los movimientos sutiles, lo que afecta al control preciso. En la mayoría de casos, la opción «Cuadrática» produce la mejor aproximación posible al hardware real."
#define OPTION_VAL_LINEAR_ES "Lineal"
#define OPTION_VAL_QUADRATIC_ES "Cuadrática"
#define OPTION_VAL_CUBIC_ES "Cúbica"
#define BEETLE_OPT_NEGCON_DEADZONE_LABEL_ES "Zona muerta al torcer el neGcon"
#define BEETLE_OPT_NEGCON_DEADZONE_INFO_0_ES "La zona muerta del stick analógico izquierdo al simular la acción de torcer un dispositivo de entrada neGcon emulado. Sirve para evitar que el mando registre movimientos no realizados. Nota: la mayoría de juegos compatibles con neGcon incluyen opciones dentro del juego para ajustar un valor de zona muerta al torcerlo. Este valor debe configurarse como cero o neutral en todo momento para que no haya una pérdida de precisión. Cualquier ajuste que sea necesario solo debe aplicarse mediante esta opción del núcleo, sobre todo si «Respuesta al torcer el neGcon» está configurada como «Cuadrática» o «Cúbica»."
#define OPTION_VAL_0_O44_ES "0 %"
#define BEETLE_OPT_USE_MEDNAFEN_MEMCARD0_METHOD_LABEL_ES "Método de guardado de Memory Cards"
#define BEETLE_OPT_USE_MEDNAFEN_MEMCARD0_METHOD_INFO_0_ES "Selecciona el formato de datos guardados que se utilizará en la Memory Card izquierda. «Mednafen» puede mejorar la compatibilidad con la versión independiente de Mednafen. Al utilizar los archivos guardados con Beetle PSX, los archivos de libretro (.srm) y Mednafen (.mcr) son idénticos internamente, así que pueden intercambiarse entre sí modificando sus nombres. Es necesario reiniciar."
#define OPTION_VAL_LIBRETRO_ES NULL
#define OPTION_VAL_MEDNAFEN_ES NULL
#define BEETLE_OPT_ENABLE_MEMCARD1_LABEL_ES "Ranura para Memory Card 2"
#define BEETLE_OPT_ENABLE_MEMCARD1_INFO_0_ES "Emula una segunda Memory Card en la ranura derecha. Nota: algunos juegos necesitan que esta opción esté desactivada para funcionar correctamente (p. ej.: Codename Tenka). Nota: la Memory Card 2 utiliza el formato de guardado Mednafen (.mcr). Es necesario reiniciar."
#define BEETLE_OPT_SHARED_MEMORY_CARDS_LABEL_ES "Memory Cards compartidas"
#define BEETLE_OPT_SHARED_MEMORY_CARDS_INFO_0_ES "Utiliza Memory Cards compartidas en vez de individuales para cada contenido. Nota: si la opción «Método de guardado de Memory Cards» está configurada en «Libretro», esta opción solo hará efecto a la Memory Card de la ranura 2. Es necesario reiniciar."
#define BEETLE_OPT_MEMCARD_LEFT_INDEX_LABEL_ES "Índice de la Memory Card de la ranura 1"
#define BEETLE_OPT_MEMCARD_LEFT_INDEX_INFO_0_ES "La Memory Card que está cargada en la ranura izquierda. Esta opción solo funcionará si la opción «Método de la Memory Card» está configurada como «Mednafen»."
#define OPTION_VAL_0_O48_ES "0 (valor predeterminado)"
#define BEETLE_OPT_MEMCARD_RIGHT_INDEX_LABEL_ES "Índice de la Memory Card de la ranura 2"
#define BEETLE_OPT_MEMCARD_RIGHT_INDEX_INFO_0_ES "La Memory Card que está cargada en la ranura derecha. Solo funcionará si la ranura para Memory Card 2 está activada."
#define OPTION_VAL_1_O49_ES "1 (valor predeterminado)"
#define BEETLE_OPT_LINE_RENDER_LABEL_ES "Arreglo para el efecto «Line-to-Quad»"
#define BEETLE_OPT_LINE_RENDER_INFO_0_ES "Ciertos juegos (Doom, Hexen, Soul Blade, etcétera) dibujan líneas horizontales estirando triángulos de un píxel de altura a lo largo de la pantalla, que luego son rasterizados como una hilera de píxeles en el hardware original. Este hack detecta esos triángulos y los convierte en rectángulos según sea necesario, permitiendo que se muestren correctamente en los renderizadores por hardware y en resoluciones internas superiores a la nativa. La opción «Agresivo» es necesaria en ciertos títulos (p. ej.: Dark Forces, Duke Nukem) pero puede producir defectos visuales en otros. En caso de duda, seleccionar «Predeterminado»."
#define OPTION_VAL_DEFAULT_ES "Predeterminado"
#define OPTION_VAL_AGGRESSIVE_ES "Agresivo"
#define BEETLE_OPT_WIDESCREEN_HACK_LABEL_ES "Modificación para pantallas panorámicas"
#define BEETLE_OPT_WIDESCREEN_HACK_INFO_0_ES "Renderiza los contenidos 3D en formato anamórfico y hace que el framebuffer emulado tenga una relación de aspecto panorámica. Produce los mejores resultados con juegos hechos completamente en 3D. Los elementos 2D se estirarán horizontalmente y podrían dejar de estar alineados."
#define BEETLE_OPT_WIDESCREEN_HACK_ASPECT_RATIO_LABEL_ES "Relación de aspecto de la modificación para pantallas panorámicas"
#define BEETLE_OPT_WIDESCREEN_HACK_ASPECT_RATIO_INFO_0_ES "Indica la relación de aspecto usada por la modificación para pantallas panorámicas."
#define OPTION_VAL_16_9_ES NULL
#define OPTION_VAL_16_10_ES NULL
#define OPTION_VAL_18_9_ES NULL
#define OPTION_VAL_19_9_ES NULL
#define OPTION_VAL_20_9_ES NULL
#define OPTION_VAL_21_9_ES NULL
#define OPTION_VAL_32_9_ES NULL
#define BEETLE_OPT_CPU_DYNAREC_LABEL_ES "Dynarec de CPU"
#define BEETLE_OPT_CPU_DYNAREC_INFO_0_ES "Recompila de forma dinámica las instrucciones de la CPU a instrucciones nativas. Es más rápido que el intérprete, pero los intervalos de la CPU son menos precisos y puede provocar fallos."
#define OPTION_VAL_DISABLED_ES "Desactivada (intérprete Beetle)"
#define OPTION_VAL_EXECUTE_ES "Máximo rendimiento"
#define OPTION_VAL_EXECUTE_ONE_ES "Con control de los intervalos entre ciclos"
#define OPTION_VAL_RUN_INTERPRETER_ES "Intérprete lightrec"
#define BEETLE_OPT_DYNAREC_INVALIDATE_LABEL_ES "Invalidación de código del dynarec"
#define BEETLE_OPT_DYNAREC_INVALIDATE_INFO_0_ES "Algunos juegos necesitan una invalidación completa y otros solo necesitan invalidar la DMA."
#define OPTION_VAL_FULL_ES "Completa"
#define OPTION_VAL_DMA_ES "Solo DMA (algo más rápida)"
#define BEETLE_OPT_DYNAREC_EVENTCYCLES_LABEL_ES "Control de ciclos de eventos de DMA/GPU/MDEC/Timer del dynarec"
#define BEETLE_OPT_DYNAREC_EVENTCYCLES_INFO_0_ES "Indica el número máximo de ciclos que debe ejecutar la CPU antes de comprobar el DMA, la GPU, el MDEC o el temporizador (timer). Un valor superior será más rápido. Tiene menos repercusión en el intérprete Beetle que en el dynarec."
#define OPTION_VAL_128_ES "128 (valor predeterminado)"
#define BEETLE_OPT_DYNAREC_SPU_SAMPLES_LABEL_ES "Control de muestras de la SPU del dynarec"
#define BEETLE_OPT_DYNAREC_SPU_SAMPLES_INFO_0_ES "Indica el número máximo de muestras que debe ejecutar la SPU antes de comprobar las actualizaciones de la SPU. Un valor superior será más rápido, pero un valor que no sea 1 provocará fallos de sonido en algunos juegos."
#define BEETLE_OPT_CORE_TIMING_FPS_LABEL_ES "Barrido de señal según núcleo"
#define BEETLE_OPT_CORE_TIMING_FPS_INFO_0_ES "El cambio automático permite que el núcleo cambie entre las señales progresiva y entrelazada."
#define OPTION_VAL_AUTO_TOGGLE_ES "Cambiar automáticamente"
#define OPTION_VAL_FORCE_PROGRESSIVE_ES "Forzar imagen progresiva"
#define OPTION_VAL_FORCE_INTERLACED_ES "Forzar barrido entrelazado"
#define BEETLE_OPT_DEINTERLACER_LABEL_ES "Método de desentrelazado"
#define BEETLE_OPT_DEINTERLACER_INFO_0_ES "«Weave» muestra ambos campos en el mismo fotograma para producir imágenes fijas a la resolución máxima, «Bob» sacrifica claridad en la resolución para ganarla en los movimientos. «Bob (compensado)» aplica el desentrelazado Bob duplicando el campo actual a toda la superficie con una compensación de una línea entre campos, lo que preservará la resolución de salida de libretro y conservará el suavizado de movimientos del desentrelazado Bob. «FastMAD (adaptativo a movimientos)» entrelaza los píxeles fijos e interpola los píxeles en movimiento en el eje vertical, ofreciendo una altura de imagen de salida completa sin el efecto peine de Weave ni la pérdida de resolución de Bob. Al desactivar esta opción, el renderizador por software respetará el mismo comportamiento del renderizador por hardware (Vulkan/OpenGL) rasterizando todas las líneas de la VRAM en cada fotograma y aplazando el «scan-out» a que acabe el fotograma, lo que eliminará el efecto de peine en juegos entrelazados. NOTA: aquellos programas que hagan DMA a la VRAM durante el «scan-out» (como ciertas demos PAL de más de 512 líneas que utilizan trucos de ajuste del búfer de fotogramas) podrían mostrar líneas paralizadas con la opción desactivada, utiliza «Weave», «Bob», «Bob (compensado)» o «FastMAD (adaptativo a movimientos)» para estos casos."
#define OPTION_VAL_WEAVE_ES "«Weave»"
#define OPTION_VAL_BOB_ES NULL
#define OPTION_VAL_BOB_OFFSET_ES "Bob (compensado)"
#define OPTION_VAL_FASTMAD_ES "FastMAD (adaptativo a movimientos)"
#define BEETLE_OPT_FRAME_DUPING_LABEL_ES "Duplicado de fotogramas"
#define BEETLE_OPT_FRAME_DUPING_INFO_0_ES "Si el núcleo no tiene nada nuevo que mostrar en la imagen, repetirá el fotograma anterior."
#define BEETLE_OPT_PAL_VIDEO_TIMING_OVERRIDE_LABEL_ES "Anular velocidad de vídeo PAL"
#define BEETLE_OPT_PAL_VIDEO_TIMING_OVERRIDE_INFO_0_ES "Debido a la diferencia entre estándares, los juegos PAL suelen parecer más lentos que sus versiones NTSC, tanto estadounidenses como japonesas. Esta opción puede anular la velocidad de vídeo PAL para intentar ejecutar dichos juegos con la velocidad de fotogramas NTSC."
#define BEETLE_OPT_ASPECT_RATIO_LABEL_ES "Relación de aspecto del núcleo"
#define BEETLE_OPT_ASPECT_RATIO_INFO_0_ES "Este ajuste será ignorado si se activan las opciones «Modificación para pantallas panorámicas» o «Mostrar la VRAM completa»."
#define OPTION_VAL_CORRECTED_ES "Corregida"
#define OPTION_VAL_UNCORRECTED_ES "Sin corregir"
#define OPTION_VAL_4_3_ES "Forzar 4:3"
#define OPTION_VAL_NTSC_ES "Forzar NTSC"
#define BEETLE_OPT_CROP_OVERSCAN_LABEL_ES "Recortar sobrebarrido horizontal"
#define BEETLE_OPT_CROP_OVERSCAN_INFO_0_ES "La imagen nativa completa, o sin bordes."
#define OPTION_VAL_STATIC_ES NULL
#define OPTION_VAL_SMART_ES NULL
#define BEETLE_OPT_IMAGE_CROP_LABEL_ES "Recorte adicional"
#define BEETLE_OPT_IMAGE_CROP_INFO_0_ES "Reduce el ancho de la imagen recortada usando un valor concreto en píxeles."
#define OPTION_VAL_9PX_ES "9 px"
#define OPTION_VAL_10PX_ES "10 px"
#define OPTION_VAL_11PX_ES "11 px"
#define OPTION_VAL_12PX_ES "12 px"
#define OPTION_VAL_13PX_ES "13 px"
#define OPTION_VAL_14PX_ES "14 px"
#define OPTION_VAL_15PX_ES "15 px"
#define OPTION_VAL_16PX_ES "16 px"
#define OPTION_VAL_17PX_ES "17 px"
#define OPTION_VAL_18PX_ES "18 px"
#define OPTION_VAL_19PX_ES "19 px"
#define OPTION_VAL_20PX_ES "20 px"
#define BEETLE_OPT_IMAGE_OFFSET_LABEL_ES "Desplazar imagen recortada"
#define BEETLE_OPT_IMAGE_OFFSET_INFO_0_ES "Desplaza la imagen horizontalmewnte hacia la derecha (valores positivos) o la izquierda (negativos) usando un valor concreto en píxeles. Solo compatible con el renderizador por software."
#define OPTION_VAL_12PX_O64_ES "-12 px"
#define OPTION_VAL_11PX_O64_ES "-11 px"
#define OPTION_VAL_10PX_O64_ES "-10 px"
#define OPTION_VAL_9PX_O64_ES "-9 px"
#define OPTION_VAL_8PX_O64_ES "-8 px"
#define OPTION_VAL_7PX_O64_ES "-7 px"
#define OPTION_VAL_6PX_O64_ES "-6 px"
#define OPTION_VAL_5PX_O64_ES "-5 px"
#define OPTION_VAL_4PX_O64_ES "-4 px"
#define OPTION_VAL_3PX_O64_ES "-3 px"
#define OPTION_VAL_2PX_O64_ES "-2 px"
#define OPTION_VAL_1PX_O64_ES "-1 px"
#define OPTION_VAL_1PX_O64_0_ES "+1 px"
#define OPTION_VAL_2PX_O64_0_ES "+2 px"
#define OPTION_VAL_3PX_O64_0_ES "+3 px"
#define OPTION_VAL_4PX_O64_0_ES "+4 px"
#define OPTION_VAL_5PX_O64_0_ES "+5 px"
#define OPTION_VAL_6PX_O64_0_ES "+6 px"
#define OPTION_VAL_7PX_O64_0_ES "+7 px"
#define OPTION_VAL_8PX_O64_0_ES "+8 px"
#define OPTION_VAL_9PX_O64_0_ES "+9 px"
#define OPTION_VAL_10PX_O64_0_ES "+10 px"
#define OPTION_VAL_11PX_O64_0_ES "+11 px"
#define OPTION_VAL_12PX_O64_0_ES "+12 px"
#define BEETLE_OPT_IMAGE_OFFSET_CYCLES_LABEL_ES "Desplazamiento horizontal de imagen (ciclos de GPU)"
#define BEETLE_OPT_IMAGE_OFFSET_CYCLES_INFO_0_ES "Los valores positivos mueven la imagen hacia la derecha y los negativos hacia la izquierda. Solo funciona con los renderizadores por hardware."
#define BEETLE_OPT_INITIAL_SCANLINE_LABEL_ES "Línea de barrido inicial en formato NTSC"
#define BEETLE_OPT_INITIAL_SCANLINE_INFO_0_ES "La primera línea de barrido que se mostrará al ejecutar contenidos NTSC. Un valor superior a cero reducirá la altura de las imágenes generadas recortando los píxeles de la parte superior. Puede servir para evitar el efecto «letterboxing»."
#define BEETLE_OPT_LAST_SCANLINE_LABEL_ES "Línea de barrido final en formato NTSC"
#define BEETLE_OPT_LAST_SCANLINE_INFO_0_ES "La última línea de barrido que se mostrará al ejecutar contenidos NTSC. Un valor inferior a 239 reducirá la altura de las imágenes generadas recortando los píxeles de la parte inferior. Puede servir para evitar el efecto «letterboxing»."
#define OPTION_VAL_239_ES "239 (por defecto)"
#define BEETLE_OPT_INITIAL_SCANLINE_PAL_LABEL_ES "Línea de barrido inicial en formato PAL"
#define BEETLE_OPT_INITIAL_SCANLINE_PAL_INFO_0_ES "La primera línea de barrido que se mostrará al ejecutar contenidos PAL. Un valor superior a cero reducirá la altura de las imágenes generadas recortando los píxeles de la parte superior. Puede servir para evitar el efecto «letterboxing»."
#define BEETLE_OPT_LAST_SCANLINE_PAL_LABEL_ES "Línea de barrido final en formato PAL"
#define BEETLE_OPT_LAST_SCANLINE_PAL_INFO_0_ES "La última línea de barrido que se mostrará al ejecutar contenidos PAL. Un valor inferior a 287 reducirá la altura de las imágenes generadas recortando los píxeles de la parte inferior. Puede servir para evitar el efecto «letterboxing»."
#define OPTION_VAL_287_ES "287 (por defecto)"

struct retro_core_option_v2_category option_cats_es[] = {
   {
      "system",
      CATEGORY_SYSTEM_LABEL_ES,
      CATEGORY_SYSTEM_INFO_0_ES
   },
   {
      "video",
      CATEGORY_VIDEO_LABEL_ES,
      CATEGORY_VIDEO_INFO_0_ES
   },
   {
      "pgxp",
      CATEGORY_PGXP_LABEL_ES,
      CATEGORY_PGXP_INFO_0_ES
   },
   {
      "osd",
      CATEGORY_OSD_LABEL_ES,
      CATEGORY_OSD_INFO_0_ES
   },
   {
      "input",
      CATEGORY_INPUT_LABEL_ES,
      CATEGORY_INPUT_INFO_0_ES
   },
   {
      "memcards",
      CATEGORY_MEMCARDS_LABEL_ES,
      CATEGORY_MEMCARDS_INFO_0_ES
   },
   {
      "hacks",
      CATEGORY_HACKS_LABEL_ES,
      CATEGORY_HACKS_INFO_0_ES
   },
   { NULL, NULL, NULL },
};
struct retro_core_option_v2_definition option_defs_es[] = {
   {
      BEETLE_OPT(cpu_freq_scale),
      BEETLE_OPT_CPU_FREQ_SCALE_LABEL_ES,
      NULL,
      BEETLE_OPT_CPU_FREQ_SCALE_INFO_0_ES,
      NULL,
      "system",
      {
         { "50%",           OPTION_VAL_50_ES },
         { "60%",           OPTION_VAL_60_ES },
         { "70%",           OPTION_VAL_70_ES },
         { "80%",           OPTION_VAL_80_ES },
         { "90%",           OPTION_VAL_90_ES },
         { "100%",          OPTION_VAL_100_ES },
         { "110%",          OPTION_VAL_110_ES },
         { "120%",          OPTION_VAL_120_ES },
         { "130%",          OPTION_VAL_130_ES },
         { "140%",          OPTION_VAL_140_ES },
         { "150%",          OPTION_VAL_150_ES },
         { "160%",          OPTION_VAL_160_ES },
         { "170%",          OPTION_VAL_170_ES },
         { "180%",          OPTION_VAL_180_ES },
         { "190%",          OPTION_VAL_190_ES },
         { "200%",          OPTION_VAL_200_ES },
         { "210%",          OPTION_VAL_210_ES },
         { "220%",          OPTION_VAL_220_ES },
         { "230%",          OPTION_VAL_230_ES },
         { "240%",          OPTION_VAL_240_ES },
         { "250%",          OPTION_VAL_250_ES },
         { "260%",          OPTION_VAL_260_ES },
         { "270%",          OPTION_VAL_270_ES },
         { "280%",          OPTION_VAL_280_ES },
         { "290%",          OPTION_VAL_290_ES },
         { "300%",          OPTION_VAL_300_ES },
         { "310%",          OPTION_VAL_310_ES },
         { "320%",          OPTION_VAL_320_ES },
         { "330%",          OPTION_VAL_330_ES },
         { "340%",          OPTION_VAL_340_ES },
         { "350%",          OPTION_VAL_350_ES },
         { "360%",          OPTION_VAL_360_ES },
         { "370%",          OPTION_VAL_370_ES },
         { "380%",          OPTION_VAL_380_ES },
         { "390%",          OPTION_VAL_390_ES },
         { "400%",          OPTION_VAL_400_ES },
         { "410%",          OPTION_VAL_410_ES },
         { "420%",          OPTION_VAL_420_ES },
         { "430%",          OPTION_VAL_430_ES },
         { "440%",          OPTION_VAL_440_ES },
         { "450%",          OPTION_VAL_450_ES },
         { "460%",          OPTION_VAL_460_ES },
         { "470%",          OPTION_VAL_470_ES },
         { "480%",          OPTION_VAL_480_ES },
         { "490%",          OPTION_VAL_490_ES },
         { "500%",          OPTION_VAL_500_ES },
         { "510%",          OPTION_VAL_510_ES },
         { "520%",          OPTION_VAL_520_ES },
         { "530%",          OPTION_VAL_530_ES },
         { "540%",          OPTION_VAL_540_ES },
         { "550%",          OPTION_VAL_550_ES },
         { "560%",          OPTION_VAL_560_ES },
         { "570%",          OPTION_VAL_570_ES },
         { "580%",          OPTION_VAL_580_ES },
         { "590%",          OPTION_VAL_590_ES },
         { "600%",          OPTION_VAL_600_ES },
         { "610%",          OPTION_VAL_610_ES },
         { "620%",          OPTION_VAL_620_ES },
         { "630%",          OPTION_VAL_630_ES },
         { "640%",          OPTION_VAL_640_ES },
         { "650%",          OPTION_VAL_650_ES },
         { "660%",          OPTION_VAL_660_ES },
         { "670%",          OPTION_VAL_670_ES },
         { "680%",          OPTION_VAL_680_ES },
         { "690%",          OPTION_VAL_690_ES },
         { "700%",          OPTION_VAL_700_ES },
         { "710%",          OPTION_VAL_710_ES },
         { "720%",          OPTION_VAL_720_ES },
         { "730%",          OPTION_VAL_730_ES },
         { "740%",          OPTION_VAL_740_ES },
         { "750%",          OPTION_VAL_750_ES },
         { NULL, NULL },
      },
      "100%"
   },
   {
      BEETLE_OPT(gpu_overclock),
      BEETLE_OPT_GPU_OVERCLOCK_LABEL_ES,
      NULL,
      BEETLE_OPT_GPU_OVERCLOCK_INFO_0_ES,
      NULL,
      "system",
      {
         { "1x(native)", OPTION_VAL_1X_NATIVE_ES },
         { "2x",         OPTION_VAL_2X_ES },
         { "4x",         OPTION_VAL_4X_ES },
         { "8x",         OPTION_VAL_8X_ES },
         { "16x",        OPTION_VAL_16X_ES },
         { "32x",        OPTION_VAL_32X_ES },
         { NULL, NULL },
      },
      "1x(native)"
   },
   {
      BEETLE_OPT(gte_overclock),
      BEETLE_OPT_GTE_OVERCLOCK_LABEL_ES,
      NULL,
      BEETLE_OPT_GTE_OVERCLOCK_INFO_0_ES,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
#ifndef EMSCRIPTEN
   {
      BEETLE_OPT(cd_access_method),
      BEETLE_OPT_CD_ACCESS_METHOD_LABEL_ES,
      NULL,
      BEETLE_OPT_CD_ACCESS_METHOD_INFO_0_ES,
      NULL,
      "system",
      {
         { "sync",     OPTION_VAL_SYNC_ES },
         { "async",    OPTION_VAL_ASYNC_ES },
         { "precache", OPTION_VAL_PRECACHE_ES },
         { NULL, NULL },
      },
      "sync"
   },
#endif
   {
      BEETLE_OPT(cd_fastload),
      BEETLE_OPT_CD_FASTLOAD_LABEL_ES,
      NULL,
      BEETLE_OPT_CD_FASTLOAD_INFO_0_ES,
      NULL,
      "system",
      {
         { "2x(native)", OPTION_VAL_2X_NATIVE_ES },
         { "4x",          OPTION_VAL_4X_ES },
         { "6x",          OPTION_VAL_6X_ES },
         { "8x",          OPTION_VAL_8X_ES },
         { "10x",         OPTION_VAL_10X_ES },
         { "12x",         OPTION_VAL_12X_ES },
         { "14x",         OPTION_VAL_14X_ES },
         { NULL, NULL },
      },
      "2x(native)"
   },
   {
      BEETLE_OPT(skip_bios),
      BEETLE_OPT_SKIP_BIOS_LABEL_ES,
      NULL,
      BEETLE_OPT_SKIP_BIOS_INFO_0_ES,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(override_bios),
      BEETLE_OPT_OVERRIDE_BIOS_LABEL_ES,
      NULL,
      BEETLE_OPT_OVERRIDE_BIOS_INFO_0_ES,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "psxonpsp", OPTION_VAL_PSXONPSP_ES },
         { "ps1_rom",  OPTION_VAL_PS1_ROM_ES },
         { "openbios", OPTION_VAL_OPENBIOS_ES     },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(region),
      BEETLE_OPT_REGION_LABEL_ES,
      NULL,
      BEETLE_OPT_REGION_INFO_0_ES,
      NULL,
      "system",
      {
         { "auto",   OPTION_VAL_AUTO_ES },
         { "ntsc-j", OPTION_VAL_NTSC_J_ES },
         { "ntsc-u", OPTION_VAL_NTSC_U_ES },
         { "pal",    OPTION_VAL_PAL_ES },
         { NULL, NULL },
      },
      "auto"
   },
   {
      BEETLE_OPT(internal_resolution),
      BEETLE_OPT_INTERNAL_RESOLUTION_LABEL_ES,
      NULL,
      BEETLE_OPT_INTERNAL_RESOLUTION_INFO_0_ES,
      NULL,
      "video",
      {
         { "1x(native)", OPTION_VAL_1X_NATIVE_ES },
         { "2x",         OPTION_VAL_2X_ES },
         { "4x",         OPTION_VAL_4X_ES },
         { "8x",         OPTION_VAL_8X_ES },
         { "16x",        OPTION_VAL_16X_ES },
         { NULL, NULL },
      },
      "1x(native)"
   },
#if defined(HAVE_OPENGL) || defined(HAVE_OPENGLES) || defined(HAVE_VULKAN)
   {
      BEETLE_OPT(renderer),
      BEETLE_OPT_RENDERER_LABEL_ES,
      NULL,
      BEETLE_OPT_RENDERER_INFO_0_ES,
      NULL,
      "video",
      {
         { "hardware",    OPTION_VAL_HARDWARE_ES },
#if defined(HAVE_OPENGL) || defined(HAVE_OPENGLES)
         { "hardware_gl", OPTION_VAL_HARDWARE_GL_ES },
#endif
#if defined(HAVE_VULKAN)
         { "hardware_vk", OPTION_VAL_HARDWARE_VK_ES },
#endif
         { "software",    OPTION_VAL_SOFTWARE_ES },
         { NULL, NULL },
      },
      "hardware"
   },
   {
      BEETLE_OPT(renderer_software_fb),
      BEETLE_OPT_RENDERER_SOFTWARE_FB_LABEL_ES,
      NULL,
      BEETLE_OPT_RENDERER_SOFTWARE_FB_INFO_0_ES,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled"
   },
#endif
#if defined(HAVE_OPENGL) || defined(HAVE_OPENGLES)
   {
      BEETLE_OPT(depth),
      BEETLE_OPT_DEPTH_LABEL_ES,
      NULL,
      BEETLE_OPT_DEPTH_INFO_0_ES,
      NULL,
      "video",
      {
         { "16bpp(native)", OPTION_VAL_16BPP_NATIVE_ES },
         { "32bpp",         OPTION_VAL_32BPP_ES },
         { NULL, NULL },
      },
      "16bpp(native)"
   },
   // Sort of, it's more like 15-bit high color and 24-bit true color for visible output. The alpha channel is used for mask bit. Vulkan renderer uses ABGR1555_555 for 31 bits internal? FMVs are always 24-bit on all renderers like original hardware (BGR888, no alpha)
#endif
   {
      BEETLE_OPT(dither_mode),
      BEETLE_OPT_DITHER_MODE_LABEL_ES,
      NULL,
      BEETLE_OPT_DITHER_MODE_INFO_0_ES,
      NULL,
      "video",
      {
         { "1x(native)",          OPTION_VAL_1X_NATIVE_ES },
         { "internal resolution", OPTION_VAL_INTERNAL_RESOLUTION_ES },
         { "disabled",            NULL },
         { NULL, NULL },
      },
      "1x(native)"
   },
#ifdef HAVE_VULKAN
   {
      BEETLE_OPT(scaled_uv_offset),
      BEETLE_OPT_SCALED_UV_OFFSET_LABEL_ES,
      NULL,
      BEETLE_OPT_SCALED_UV_OFFSET_INFO_0_ES,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled"
   },
#endif
#if defined(HAVE_OPENGL) || defined(HAVE_OPENGLES) || defined(HAVE_VULKAN)
   {
      BEETLE_OPT(filter),
      BEETLE_OPT_FILTER_LABEL_ES,
      NULL,
      BEETLE_OPT_FILTER_INFO_0_ES,
      NULL,
      "video",
      {
         { "nearest",  OPTION_VAL_NEAREST_ES },
         { "SABR",     OPTION_VAL_SABR_ES },
         { "xBR",      OPTION_VAL_XBR_ES },
         { "bilinear", OPTION_VAL_BILINEAR_ES },
         { "3-point",  OPTION_VAL_3_POINT_ES },
         { "JINC2",    OPTION_VAL_JINC2_ES },
         { NULL, NULL },
      },
      "nearest"
   },
#ifdef HAVE_VULKAN
   {
      BEETLE_OPT(filter_exclude_sprite),
      BEETLE_OPT_FILTER_EXCLUDE_SPRITE_LABEL_ES,
      NULL,
      BEETLE_OPT_FILTER_EXCLUDE_SPRITE_INFO_0_ES,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "opaque", OPTION_VAL_OPAQUE_ES },
         { "all", OPTION_VAL_ALL_ES },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(filter_exclude_2d_polygon),
      BEETLE_OPT_FILTER_EXCLUDE_2D_POLYGON_LABEL_ES,
      NULL,
      BEETLE_OPT_FILTER_EXCLUDE_2D_POLYGON_INFO_0_ES,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "opaque", OPTION_VAL_OPAQUE_ES },
         { "all", OPTION_VAL_ALL_ES },
         { NULL, NULL },
      },
      "disabled"
   },
#endif
#endif
#ifdef HAVE_VULKAN
   {
      BEETLE_OPT(adaptive_smoothing),
      BEETLE_OPT_ADAPTIVE_SMOOTHING_LABEL_ES,
      NULL,
      BEETLE_OPT_ADAPTIVE_SMOOTHING_INFO_0_ES,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(super_sampling),
      BEETLE_OPT_SUPER_SAMPLING_LABEL_ES,
      NULL,
      BEETLE_OPT_SUPER_SAMPLING_INFO_0_ES,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(msaa),
      BEETLE_OPT_MSAA_LABEL_ES,
      NULL,
      BEETLE_OPT_MSAA_INFO_0_ES,
      NULL,
      "video",
      {
         { "1x",  OPTION_VAL_1X_ES },
         { "2x",  OPTION_VAL_2X_ES },
         { "4x",  OPTION_VAL_4X_ES },
         { "8x",  OPTION_VAL_8X_ES },
         { "16x", OPTION_VAL_16X_ES },
         { NULL, NULL },
      },
      "1x"
   },
   {
      BEETLE_OPT(mdec_yuv),
      BEETLE_OPT_MDEC_YUV_LABEL_ES,
      NULL,
      BEETLE_OPT_MDEC_YUV_INFO_0_ES,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(track_textures),
      BEETLE_OPT_TRACK_TEXTURES_LABEL_ES,
      NULL,
      BEETLE_OPT_TRACK_TEXTURES_INFO_0_ES,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
#ifdef TEXTURE_DUMPING_ENABLED
   {
      BEETLE_OPT(dump_textures),
      BEETLE_OPT_DUMP_TEXTURES_LABEL_ES,
      NULL,
      BEETLE_OPT_DUMP_TEXTURES_INFO_0_ES,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
#endif
   {
      BEETLE_OPT(replace_textures),
      BEETLE_OPT_REPLACE_TEXTURES_LABEL_ES,
      NULL,
      BEETLE_OPT_REPLACE_TEXTURES_INFO_0_ES,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
#endif
   {
      BEETLE_OPT(pgxp_mode),
      BEETLE_OPT_PGXP_MODE_LABEL_ES,
      NULL,
      BEETLE_OPT_PGXP_MODE_INFO_0_ES,
      NULL,
      "pgxp",
      {
         { "disabled",     NULL },
         { "memory only",  OPTION_VAL_MEMORY_ONLY_ES },
         { "memory + CPU", OPTION_VAL_MEMORY_CPU_ES },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(pgxp_2d_tol),
      BEETLE_OPT_PGXP_2D_TOL_LABEL_ES,
      NULL,
      BEETLE_OPT_PGXP_2D_TOL_INFO_0_ES,
      NULL,
      "pgxp",
      {
         { "disabled", NULL },
         { "0px", OPTION_VAL_0PX_ES },
         { "1px", OPTION_VAL_1PX_ES },
         { "2px", OPTION_VAL_2PX_ES },
         { "3px", OPTION_VAL_3PX_ES },
         { "4px", OPTION_VAL_4PX_ES },
         { "5px", OPTION_VAL_5PX_ES },
         { "6px", OPTION_VAL_6PX_ES },
         { "7px", OPTION_VAL_7PX_ES },
         { "8px", OPTION_VAL_8PX_ES },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(pgxp_nclip),
      BEETLE_OPT_PGXP_NCLIP_LABEL_ES,
      NULL,
      BEETLE_OPT_PGXP_NCLIP_INFO_0_ES,
      NULL,
      "pgxp",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
#if defined(HAVE_OPENGL) || defined(HAVE_OPENGLES) || defined(HAVE_VULKAN)
   {
      BEETLE_OPT(pgxp_vertex),
      BEETLE_OPT_PGXP_VERTEX_LABEL_ES,
      NULL,
      BEETLE_OPT_PGXP_VERTEX_INFO_0_ES,
      NULL,
      "pgxp",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
#endif
   {
      BEETLE_OPT(pgxp_texture),
      BEETLE_OPT_PGXP_TEXTURE_LABEL_ES,
      NULL,
      BEETLE_OPT_PGXP_TEXTURE_INFO_0_ES,
      NULL,
      "pgxp",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(display_internal_fps),
      BEETLE_OPT_DISPLAY_INTERNAL_FPS_LABEL_ES,
      NULL,
      BEETLE_OPT_DISPLAY_INTERNAL_FPS_INFO_0_ES,
      NULL,
      "osd",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(display_osd),
      BEETLE_OPT_DISPLAY_OSD_LABEL_ES,
      NULL,
      BEETLE_OPT_DISPLAY_OSD_INFO_0_ES,
      NULL,
      "osd",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled"
   },
#if defined(HAVE_OPENGL) || defined(HAVE_OPENGLES) || defined(HAVE_VULKAN)
   {
      BEETLE_OPT(display_vram),
      BEETLE_OPT_DISPLAY_VRAM_LABEL_ES,
      NULL,
      BEETLE_OPT_DISPLAY_VRAM_INFO_0_ES,
      NULL,
      "osd",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
#endif
   {
      BEETLE_OPT(analog_calibration),
      BEETLE_OPT_ANALOG_CALIBRATION_LABEL_ES,
      NULL,
      BEETLE_OPT_ANALOG_CALIBRATION_INFO_0_ES,
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(analog_toggle),
      BEETLE_OPT_ANALOG_TOGGLE_LABEL_ES,
      NULL,
      BEETLE_OPT_ANALOG_TOGGLE_INFO_0_ES,
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { "enabled-analog", OPTION_VAL_ENABLED_ANALOG_ES },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(analog_toggle_combo),
      BEETLE_OPT_ANALOG_TOGGLE_COMBO_LABEL_ES,
      NULL,
      BEETLE_OPT_ANALOG_TOGGLE_COMBO_INFO_0_ES,
      NULL,
      "input",
      {
         { "l1+l2+r1+r2+start+select", OPTION_VAL_L1_L2_R1_R2_START_SELECT_ES },
         { "l1+r1+select",             OPTION_VAL_L1_R1_SELECT_ES },
         { "l1+r1+start",              OPTION_VAL_L1_R1_START_ES },
         { "l1+r1+l3",                 OPTION_VAL_L1_R1_L3_ES },
         { "l1+r1+r3",                 OPTION_VAL_L1_R1_R3_ES },
         { "l2+r2+select",             OPTION_VAL_L2_R2_SELECT_ES },
         { "l2+r2+start",              OPTION_VAL_L2_R2_START_ES },
         { "l2+r2+l3",                 OPTION_VAL_L2_R2_L3_ES },
         { "l2+r2+r3",                 OPTION_VAL_L2_R2_R3_ES },
         { "l3+r3",                    OPTION_VAL_L3_R3_ES },
         { NULL, NULL },
      },
      "l1+r1+select"
   },
   {
      BEETLE_OPT(analog_toggle_hold),
      BEETLE_OPT_ANALOG_TOGGLE_HOLD_LABEL_ES,
      NULL,
      BEETLE_OPT_ANALOG_TOGGLE_HOLD_INFO_0_ES,
      NULL,
      "input",
      {
         { "0", OPTION_VAL_0_ES },
         { "1", OPTION_VAL_1_ES },
         { "2", OPTION_VAL_2_ES },
         { "3", OPTION_VAL_3_ES },
         { "4", OPTION_VAL_4_ES },
         { "5", OPTION_VAL_5_ES },
         { NULL, NULL },
      },
      "1"
   },
   {
      BEETLE_OPT(enable_multitap_port1),
      BEETLE_OPT_ENABLE_MULTITAP_PORT1_LABEL_ES,
      NULL,
      BEETLE_OPT_ENABLE_MULTITAP_PORT1_INFO_0_ES,
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(enable_multitap_port2),
      BEETLE_OPT_ENABLE_MULTITAP_PORT2_LABEL_ES,
      NULL,
      BEETLE_OPT_ENABLE_MULTITAP_PORT2_INFO_0_ES,
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(gun_input_mode),
      BEETLE_OPT_GUN_INPUT_MODE_LABEL_ES,
      NULL,
      BEETLE_OPT_GUN_INPUT_MODE_INFO_0_ES,
      NULL,
      "input",
      {
         { "lightgun",    OPTION_VAL_LIGHTGUN_ES },
         { "touchscreen", OPTION_VAL_TOUCHSCREEN_ES },
         { NULL, NULL },
      },
      "lightgun"
   },
   // Shouldn't the gun_input_mode just be Mouse vs. Touchscreen?
   {
      BEETLE_OPT(gun_cursor),
      BEETLE_OPT_GUN_CURSOR_LABEL_ES,
      NULL,
      BEETLE_OPT_GUN_CURSOR_INFO_0_ES,
      NULL,
      "input",
      {
         { "off",   "disabled" },
         { "cross", OPTION_VAL_CROSS_ES },
         { "dot",   OPTION_VAL_DOT_ES },
         { NULL, NULL },
      },
      "cross"
   },
   {
      BEETLE_OPT(crosshair_color_p1),
      BEETLE_OPT_CROSSHAIR_COLOR_P1_LABEL_ES,
      NULL,
      BEETLE_OPT_CROSSHAIR_COLOR_P1_INFO_0_ES,
      NULL,
      "input",
      {
         { "red",    OPTION_VAL_RED_ES },
         { "blue",   OPTION_VAL_BLUE_ES },
         { "green",  OPTION_VAL_GREEN_ES },
         { "orange", OPTION_VAL_ORANGE_ES },
         { "yellow", OPTION_VAL_YELLOW_ES },
         { "cyan",   OPTION_VAL_CYAN_ES },
         { "pink",   OPTION_VAL_PINK_ES },
         { "purple", OPTION_VAL_PURPLE_ES },
         { "black",  OPTION_VAL_BLACK_ES },
         { "white",  OPTION_VAL_WHITE_ES },
         { NULL, NULL },
      },
      "red"
   },
   {
      BEETLE_OPT(crosshair_color_p2),
      BEETLE_OPT_CROSSHAIR_COLOR_P2_LABEL_ES,
      NULL,
      BEETLE_OPT_CROSSHAIR_COLOR_P2_INFO_0_ES,
      NULL,
      "input",
      {
         { "blue",   OPTION_VAL_BLUE_ES },
         { "red",    OPTION_VAL_RED_ES },
         { "green",  OPTION_VAL_GREEN_ES },
         { "orange", OPTION_VAL_ORANGE_ES },
         { "yellow", OPTION_VAL_YELLOW_ES },
         { "cyan",   OPTION_VAL_CYAN_ES },
         { "pink",   OPTION_VAL_PINK_ES },
         { "purple", OPTION_VAL_PURPLE_ES },
         { "black",  OPTION_VAL_BLACK_ES },
         { "white",  OPTION_VAL_WHITE_ES },
         { NULL, NULL },
      },
      "blue"
   },
   {
      BEETLE_OPT(mouse_sensitivity),
      BEETLE_OPT_MOUSE_SENSITIVITY_LABEL_ES,
      NULL,
      BEETLE_OPT_MOUSE_SENSITIVITY_INFO_0_ES,
      NULL,
      "input",
      {
         { "5%",   OPTION_VAL_5_O42_ES },
         { "10%",  OPTION_VAL_10_ES },
         { "15%",  OPTION_VAL_15_ES },
         { "20%",  OPTION_VAL_20_ES },
         { "25%",  OPTION_VAL_25_ES },
         { "30%",  OPTION_VAL_30_ES },
         { "35%",  OPTION_VAL_35_ES },
         { "40%",  OPTION_VAL_40_ES },
         { "45%",  OPTION_VAL_45_ES },
         { "50%",  OPTION_VAL_50_ES },
         { "55%",  OPTION_VAL_55_ES },
         { "60%",  OPTION_VAL_60_ES },
         { "65%",  OPTION_VAL_65_ES },
         { "70%",  OPTION_VAL_70_ES },
         { "75%",  OPTION_VAL_75_ES },
         { "80%",  OPTION_VAL_80_ES },
         { "85%",  OPTION_VAL_85_ES },
         { "90%",  OPTION_VAL_90_ES },
         { "95%",  OPTION_VAL_95_ES },
         { "100%", OPTION_VAL_100_O42_ES },
         { "105%", OPTION_VAL_105_ES },
         { "110%", OPTION_VAL_110_ES },
         { "115%", OPTION_VAL_115_ES },
         { "120%", OPTION_VAL_120_ES },
         { "125%", OPTION_VAL_125_ES },
         { "130%", OPTION_VAL_130_ES },
         { "135%", OPTION_VAL_135_ES },
         { "140%", OPTION_VAL_140_ES },
         { "145%", OPTION_VAL_145_ES },
         { "150%", OPTION_VAL_150_ES },
         { "155%", OPTION_VAL_155_ES },
         { "160%", OPTION_VAL_160_ES },
         { "165%", OPTION_VAL_165_ES },
         { "170%", OPTION_VAL_170_ES },
         { "175%", OPTION_VAL_175_ES },
         { "180%", OPTION_VAL_180_ES },
         { "185%", OPTION_VAL_185_ES },
         { "190%", OPTION_VAL_190_ES },
         { "195%", OPTION_VAL_195_ES },
         { "200%", OPTION_VAL_200_ES },
         { NULL, NULL },
      },
      "100%"
   },
   {
      BEETLE_OPT(negcon_response),
      BEETLE_OPT_NEGCON_RESPONSE_LABEL_ES,
      NULL,
      BEETLE_OPT_NEGCON_RESPONSE_INFO_0_ES,
      NULL,
      "input",
      {
         { "linear",    OPTION_VAL_LINEAR_ES },
         { "quadratic", OPTION_VAL_QUADRATIC_ES },
         { "cubic",     OPTION_VAL_CUBIC_ES },
         { NULL, NULL },
      },
      "linear"
   },
   {
      BEETLE_OPT(negcon_deadzone),
      BEETLE_OPT_NEGCON_DEADZONE_LABEL_ES,
      NULL,
      BEETLE_OPT_NEGCON_DEADZONE_INFO_0_ES,
      NULL,
      "input",
      {
         { "0%",  OPTION_VAL_0_O44_ES },
         { "5%",  OPTION_VAL_5_O42_ES },
         { "10%", OPTION_VAL_10_ES },
         { "15%", OPTION_VAL_15_ES },
         { "20%", OPTION_VAL_20_ES },
         { "25%", OPTION_VAL_25_ES },
         { "30%", OPTION_VAL_30_ES },
         { NULL, NULL },
      },
      "0%"
   },
   {
      BEETLE_OPT(use_mednafen_memcard0_method),
      BEETLE_OPT_USE_MEDNAFEN_MEMCARD0_METHOD_LABEL_ES,
      NULL,
      BEETLE_OPT_USE_MEDNAFEN_MEMCARD0_METHOD_INFO_0_ES,
      NULL,
      "memcards",
      {
         { "libretro", OPTION_VAL_LIBRETRO_ES },
         { "mednafen", OPTION_VAL_MEDNAFEN_ES },
         { NULL, NULL },
      },
      "libretro"
   },
   {
      BEETLE_OPT(enable_memcard1),
      BEETLE_OPT_ENABLE_MEMCARD1_LABEL_ES,
      NULL,
      BEETLE_OPT_ENABLE_MEMCARD1_INFO_0_ES,
      NULL,
      "memcards",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(shared_memory_cards),
      BEETLE_OPT_SHARED_MEMORY_CARDS_LABEL_ES,
      NULL,
      BEETLE_OPT_SHARED_MEMORY_CARDS_INFO_0_ES,
      NULL,
      "memcards",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(memcard_left_index),
      BEETLE_OPT_MEMCARD_LEFT_INDEX_LABEL_ES,
      NULL,
      BEETLE_OPT_MEMCARD_LEFT_INDEX_INFO_0_ES,
      NULL,
      "memcards",
      {
         { "0",  OPTION_VAL_0_O48_ES },
         { "1",  NULL },
         { "2",  NULL },
         { "3",  NULL },
         { "4",  NULL },
         { "5",  NULL },
         { "6",  NULL },
         { "7",  NULL },
         { "8",  NULL },
         { "9",  NULL },
         { "10", NULL },
         { "11", NULL },
         { "12", NULL },
         { "13", NULL },
         { "14", NULL },
         { "15", NULL },
         { "16", NULL },
         { "17", NULL },
         { "18", NULL },
         { "19", NULL },
         { "20", NULL },
         { "21", NULL },
         { "22", NULL },
         { "23", NULL },
         { "24", NULL },
         { "25", NULL },
         { "26", NULL },
         { "27", NULL },
         { "28", NULL },
         { "29", NULL },
         { "30", NULL },
         { "31", NULL },
         { "32", NULL },
         { "33", NULL },
         { "34", NULL },
         { "35", NULL },
         { "36", NULL },
         { "37", NULL },
         { "38", NULL },
         { "39", NULL },
         { "40", NULL },
         { "41", NULL },
         { "42", NULL },
         { "43", NULL },
         { "44", NULL },
         { "45", NULL },
         { "46", NULL },
         { "47", NULL },
         { "48", NULL },
         { "49", NULL },
         { "50", NULL },
         { "51", NULL },
         { "52", NULL },
         { "53", NULL },
         { "54", NULL },
         { "55", NULL },
         { "56", NULL },
         { "57", NULL },
         { "58", NULL },
         { "59", NULL },
         { "60", NULL },
         { "61", NULL },
         { "62", NULL },
         { "63", NULL },
         { NULL, NULL },
      },
      "0"
   },
   {
      BEETLE_OPT(memcard_right_index),
      BEETLE_OPT_MEMCARD_RIGHT_INDEX_LABEL_ES,
      NULL,
      BEETLE_OPT_MEMCARD_RIGHT_INDEX_INFO_0_ES,
      NULL,
      "memcards",
      {
         { "0",  NULL },
         { "1",  OPTION_VAL_1_O49_ES },
         { "2",  NULL },
         { "3",  NULL },
         { "4",  NULL },
         { "5",  NULL },
         { "6",  NULL },
         { "7",  NULL },
         { "8",  NULL },
         { "9",  NULL },
         { "10", NULL },
         { "11", NULL },
         { "12", NULL },
         { "13", NULL },
         { "14", NULL },
         { "15", NULL },
         { "16", NULL },
         { "17", NULL },
         { "18", NULL },
         { "19", NULL },
         { "20", NULL },
         { "21", NULL },
         { "22", NULL },
         { "23", NULL },
         { "24", NULL },
         { "25", NULL },
         { "26", NULL },
         { "27", NULL },
         { "28", NULL },
         { "29", NULL },
         { "30", NULL },
         { "31", NULL },
         { "32", NULL },
         { "33", NULL },
         { "34", NULL },
         { "35", NULL },
         { "36", NULL },
         { "37", NULL },
         { "38", NULL },
         { "39", NULL },
         { "40", NULL },
         { "41", NULL },
         { "42", NULL },
         { "43", NULL },
         { "44", NULL },
         { "45", NULL },
         { "46", NULL },
         { "47", NULL },
         { "48", NULL },
         { "49", NULL },
         { "50", NULL },
         { "51", NULL },
         { "52", NULL },
         { "53", NULL },
         { "54", NULL },
         { "55", NULL },
         { "56", NULL },
         { "57", NULL },
         { "58", NULL },
         { "59", NULL },
         { "60", NULL },
         { "61", NULL },
         { "62", NULL },
         { "63", NULL },
         { NULL, NULL },
      },
      "1"
   },
   {
      BEETLE_OPT(line_render),
      BEETLE_OPT_LINE_RENDER_LABEL_ES,
      NULL,
      BEETLE_OPT_LINE_RENDER_INFO_0_ES,
      NULL,
      "hacks",
      {
         { "default",    OPTION_VAL_DEFAULT_ES },
         { "aggressive", OPTION_VAL_AGGRESSIVE_ES },
         { "disabled",   NULL },
         { NULL, NULL },
      },
      "default"
   },
   {
      BEETLE_OPT(widescreen_hack),
      BEETLE_OPT_WIDESCREEN_HACK_LABEL_ES,
      NULL,
      BEETLE_OPT_WIDESCREEN_HACK_INFO_0_ES,
      NULL,
      "hacks",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(widescreen_hack_aspect_ratio),
      BEETLE_OPT_WIDESCREEN_HACK_ASPECT_RATIO_LABEL_ES,
      NULL,
      BEETLE_OPT_WIDESCREEN_HACK_ASPECT_RATIO_INFO_0_ES,
      NULL,
      "hacks",
      {
         { "16:9",  OPTION_VAL_16_9_ES },
         { "16:10", OPTION_VAL_16_10_ES },
         { "18:9",  OPTION_VAL_18_9_ES },
         { "19:9",  OPTION_VAL_19_9_ES },
         { "20:9",  OPTION_VAL_20_9_ES },
         { "21:9",  OPTION_VAL_21_9_ES }, // 64:27
         { "32:9",  OPTION_VAL_32_9_ES },
         { NULL,    NULL },
      },
      "16:9"
   },
#if defined(HAVE_LIGHTREC)
   {
      BEETLE_OPT(cpu_dynarec),
      BEETLE_OPT_CPU_DYNAREC_LABEL_ES,
      NULL,
      BEETLE_OPT_CPU_DYNAREC_INFO_0_ES,
      NULL,
      "hacks",
      {
         { "disabled",        OPTION_VAL_DISABLED_ES },
         { "execute",         OPTION_VAL_EXECUTE_ES },
         { "execute_one",     OPTION_VAL_EXECUTE_ONE_ES },
         { "run_interpreter", OPTION_VAL_RUN_INTERPRETER_ES },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(dynarec_invalidate),
      BEETLE_OPT_DYNAREC_INVALIDATE_LABEL_ES,
      NULL,
      BEETLE_OPT_DYNAREC_INVALIDATE_INFO_0_ES,
      NULL,
      "hacks",
      {
         { "full", OPTION_VAL_FULL_ES },
         { "dma",  OPTION_VAL_DMA_ES },
         { NULL, NULL },
      },
      "full"
   },
   {
      BEETLE_OPT(dynarec_eventcycles),
      BEETLE_OPT_DYNAREC_EVENTCYCLES_LABEL_ES,
      NULL,
      BEETLE_OPT_DYNAREC_EVENTCYCLES_INFO_0_ES,
      NULL,
      "hacks",
      {
         { "128",  OPTION_VAL_128_ES },
         { "256",  NULL },
         { "384",  NULL },
         { "512",  NULL },
         { "640",  NULL },
         { "768",  NULL },
         { "896",  NULL },
         { "1024", NULL },
         { "1152", NULL },
         { "1280", NULL },
         { "1408", NULL },
         { "1536", NULL },
         { "1664", NULL },
         { "1792", NULL },
         { "1920", NULL },
         { "2048", NULL },
         { NULL, NULL },
      },
      "128"
   },
   {
      BEETLE_OPT(dynarec_spu_samples),
      BEETLE_OPT_DYNAREC_SPU_SAMPLES_LABEL_ES,
      NULL,
      BEETLE_OPT_DYNAREC_SPU_SAMPLES_INFO_0_ES,
      NULL,
      "hacks",
      {
         { "1",  OPTION_VAL_1_O49_ES },
         { "4",  NULL },
         { "16", NULL },
         { NULL, NULL },
      },
      "1"
   },
#endif
   {
      BEETLE_OPT(core_timing_fps),
      BEETLE_OPT_CORE_TIMING_FPS_LABEL_ES,
      NULL,
      BEETLE_OPT_CORE_TIMING_FPS_INFO_0_ES,
      NULL,
      "video",
      {
         { "auto_toggle",       OPTION_VAL_AUTO_TOGGLE_ES },
         { "force_progressive", OPTION_VAL_FORCE_PROGRESSIVE_ES },
         { "force_interlaced",  OPTION_VAL_FORCE_INTERLACED_ES },
      },
      "auto_toggle"
   },
   {
      BEETLE_OPT(deinterlacer),
      BEETLE_OPT_DEINTERLACER_LABEL_ES,
      NULL,
      BEETLE_OPT_DEINTERLACER_INFO_0_ES,
      NULL,
      "video",
      {
         { "weave",      OPTION_VAL_WEAVE_ES },
         { "bob",        OPTION_VAL_BOB_ES },
         { "bob_offset", OPTION_VAL_BOB_OFFSET_ES },
         { "fastmad",    OPTION_VAL_FASTMAD_ES },
         { "off",        "Off" },
         { NULL, NULL },
      },
      "weave"
   },
   {
      BEETLE_OPT(frame_duping),
      BEETLE_OPT_FRAME_DUPING_LABEL_ES,
      NULL,
      BEETLE_OPT_FRAME_DUPING_INFO_0_ES,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(pal_video_timing_override),
      BEETLE_OPT_PAL_VIDEO_TIMING_OVERRIDE_LABEL_ES,
      NULL,
      BEETLE_OPT_PAL_VIDEO_TIMING_OVERRIDE_INFO_0_ES,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(aspect_ratio),
      BEETLE_OPT_ASPECT_RATIO_LABEL_ES,
      NULL,
      BEETLE_OPT_ASPECT_RATIO_INFO_0_ES,
      NULL,
      "video",
      {
         { "corrected", OPTION_VAL_CORRECTED_ES },
         { "uncorrected", OPTION_VAL_UNCORRECTED_ES },
         { "4:3",  OPTION_VAL_4_3_ES },
         { "ntsc", OPTION_VAL_NTSC_ES },
      },
      "corrected"
   },
   {
      BEETLE_OPT(crop_overscan),
      BEETLE_OPT_CROP_OVERSCAN_LABEL_ES,
      NULL,
      BEETLE_OPT_CROP_OVERSCAN_INFO_0_ES,
      NULL,
      "video",
      {
         { "disabled", "Disabled" },
         { "static",   OPTION_VAL_STATIC_ES },
         { "smart",    OPTION_VAL_SMART_ES },
         { NULL, NULL },
      },
      "smart"
   },
   {
      BEETLE_OPT(image_crop),
      BEETLE_OPT_IMAGE_CROP_LABEL_ES,
      NULL,
      BEETLE_OPT_IMAGE_CROP_INFO_0_ES,
      NULL,
      "video",
      {
         { "disabled", "0" },
         { "1px",      OPTION_VAL_1PX_ES },
         { "2px",      OPTION_VAL_2PX_ES },
         { "3px",      OPTION_VAL_3PX_ES },
         { "4px",      OPTION_VAL_4PX_ES },
         { "5px",      OPTION_VAL_5PX_ES },
         { "6px",      OPTION_VAL_6PX_ES },
         { "7px",      OPTION_VAL_7PX_ES },
         { "8px",      OPTION_VAL_8PX_ES },
         { "9px",      OPTION_VAL_9PX_ES },
         { "10px",     OPTION_VAL_10PX_ES },
         { "11px",     OPTION_VAL_11PX_ES },
         { "12px",     OPTION_VAL_12PX_ES },
         { "13px",     OPTION_VAL_13PX_ES },
         { "14px",     OPTION_VAL_14PX_ES },
         { "15px",     OPTION_VAL_15PX_ES },
         { "16px",     OPTION_VAL_16PX_ES },
         { "17px",     OPTION_VAL_17PX_ES },
         { "18px",     OPTION_VAL_18PX_ES },
         { "19px",     OPTION_VAL_19PX_ES },
         { "20px",     OPTION_VAL_20PX_ES },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(image_offset),
      BEETLE_OPT_IMAGE_OFFSET_LABEL_ES,
      NULL,
      BEETLE_OPT_IMAGE_OFFSET_INFO_0_ES,
      NULL,
      "video",
      {
         { "-12px",    OPTION_VAL_12PX_O64_ES },
         { "-11px",    OPTION_VAL_11PX_O64_ES },
         { "-10px",    OPTION_VAL_10PX_O64_ES },
         { "-9px",     OPTION_VAL_9PX_O64_ES },
         { "-8px",     OPTION_VAL_8PX_O64_ES },
         { "-7px",     OPTION_VAL_7PX_O64_ES },
         { "-6px",     OPTION_VAL_6PX_O64_ES },
         { "-5px",     OPTION_VAL_5PX_O64_ES },
         { "-4px",     OPTION_VAL_4PX_O64_ES },
         { "-3px",     OPTION_VAL_3PX_O64_ES },
         { "-2px",     OPTION_VAL_2PX_O64_ES },
         { "-1px",     OPTION_VAL_1PX_O64_ES },
         { "disabled", "0" },
         { "+1px",     OPTION_VAL_1PX_O64_0_ES },
         { "+2px",     OPTION_VAL_2PX_O64_0_ES },
         { "+3px",     OPTION_VAL_3PX_O64_0_ES },
         { "+4px",     OPTION_VAL_4PX_O64_0_ES },
         { "+5px",     OPTION_VAL_5PX_O64_0_ES },
         { "+6px",     OPTION_VAL_6PX_O64_0_ES },
         { "+7px",     OPTION_VAL_7PX_O64_0_ES },
         { "+8px",     OPTION_VAL_8PX_O64_0_ES },
         { "+9px",     OPTION_VAL_9PX_O64_0_ES },
         { "+10px",    OPTION_VAL_10PX_O64_0_ES },
         { "+11px",    OPTION_VAL_11PX_O64_0_ES },
         { "+12px",    OPTION_VAL_12PX_O64_0_ES },
         { NULL, NULL },
      },
      "disabled"
   },
#if defined(HAVE_OPENGL) || defined(HAVE_OPENGLES) || defined(HAVE_VULKAN)
   {
      BEETLE_OPT(image_offset_cycles),
      BEETLE_OPT_IMAGE_OFFSET_CYCLES_LABEL_ES,
      NULL,
      BEETLE_OPT_IMAGE_OFFSET_CYCLES_INFO_0_ES,
      NULL,
      "video",
      {
         { "-40",      NULL },
         { "-39",      NULL },
         { "-38",      NULL },
         { "-37",      NULL },
         { "-36",      NULL },
         { "-35",      NULL },
         { "-34",      NULL },
         { "-33",      NULL },
         { "-32",      NULL },
         { "-31",      NULL },
         { "-30",      NULL },
         { "-29",      NULL },
         { "-28",      NULL },
         { "-27",      NULL },
         { "-26",      NULL },
         { "-25",      NULL },
         { "-24",      NULL },
         { "-23",      NULL },
         { "-22",      NULL },
         { "-21",      NULL },
         { "-20",      NULL },
         { "-19",      NULL },
         { "-18",      NULL },
         { "-17",      NULL },
         { "-16",      NULL },
         { "-15",      NULL },
         { "-14",      NULL },
         { "-13",      NULL },
         { "-12",      NULL },
         { "-11",      NULL },
         { "-10",      NULL },
         { "-9",       NULL },
         { "-8",       NULL },
         { "-7",       NULL },
         { "-6",       NULL },
         { "-5",       NULL },
         { "-4",       NULL },
         { "-3",       NULL },
         { "-2",       NULL },
         { "-1",       NULL },
         { "0",        "0" },
         { "+1",       NULL },
         { "+2",       NULL },
         { "+3",       NULL },
         { "+4",       NULL },
         { "+5",       NULL },
         { "+6",       NULL },
         { "+7",       NULL },
         { "+8",       NULL },
         { "+9",       NULL },
         { "+10",      NULL },
         { "+11",      NULL },
         { "+12",      NULL },
         { "+13",      NULL },
         { "+14",      NULL },
         { "+15",      NULL },
         { "+16",      NULL },
         { "+17",      NULL },
         { "+18",      NULL },
         { "+19",      NULL },
         { "+20",      NULL },
         { "+21",      NULL },
         { "+22",      NULL },
         { "+23",      NULL },
         { "+24",      NULL },
         { "+25",      NULL },
         { "+26",      NULL },
         { "+27",      NULL },
         { "+28",      NULL },
         { "+29",      NULL },
         { "+30",      NULL },
         { "+31",      NULL },
         { "+32",      NULL },
         { "+33",      NULL },
         { "+34",      NULL },
         { "+35",      NULL },
         { "+36",      NULL },
         { "+37",      NULL },
         { "+38",      NULL },
         { "+39",      NULL },
         { "+40",      NULL },
         { NULL, NULL},
      },
      "0"
   },
#endif
   {
      BEETLE_OPT(initial_scanline),
      BEETLE_OPT_INITIAL_SCANLINE_LABEL_ES,
      NULL,
      BEETLE_OPT_INITIAL_SCANLINE_INFO_0_ES,
      NULL,
      "video",
      {
         { "0",  OPTION_VAL_0_O48_ES },
         { "1",  NULL },
         { "2",  NULL },
         { "3",  NULL },
         { "4",  NULL },
         { "5",  NULL },
         { "6",  NULL },
         { "7",  NULL },
         { "8",  NULL },
         { "9",  NULL },
         { "10", NULL },
         { "11", NULL },
         { "12", NULL },
         { "13", NULL },
         { "14", NULL },
         { "15", NULL },
         { "16", NULL },
         { "17", NULL },
         { "18", NULL },
         { "19", NULL },
         { "20", NULL },
         { "21", NULL },
         { "22", NULL },
         { "23", NULL },
         { "24", NULL },
         { "25", NULL },
         { "26", NULL },
         { "27", NULL },
         { "28", NULL },
         { "29", NULL },
         { "30", NULL },
         { "31", NULL },
         { "32", NULL },
         { "33", NULL },
         { "34", NULL },
         { "35", NULL },
         { "36", NULL },
         { "37", NULL },
         { "38", NULL },
         { "39", NULL },
         { "40", NULL },
         { NULL, NULL },
      },
      "0"
   },
   {
      BEETLE_OPT(last_scanline),
      BEETLE_OPT_LAST_SCANLINE_LABEL_ES,
      NULL,
      BEETLE_OPT_LAST_SCANLINE_INFO_0_ES,
      NULL,
      "video",
      {
         { "210", NULL },
         { "211", NULL },
         { "212", NULL },
         { "213", NULL },
         { "214", NULL },
         { "215", NULL },
         { "216", NULL },
         { "217", NULL },
         { "218", NULL },
         { "219", NULL },
         { "220", NULL },
         { "221", NULL },
         { "222", NULL },
         { "223", NULL },
         { "224", NULL },
         { "225", NULL },
         { "226", NULL },
         { "227", NULL },
         { "228", NULL },
         { "229", NULL },
         { "230", NULL },
         { "231", NULL },
         { "232", NULL },
         { "233", NULL },
         { "234", NULL },
         { "235", NULL },
         { "236", NULL },
         { "237", NULL },
         { "238", NULL },
         { "239", OPTION_VAL_239_ES },
         { NULL, NULL },
      },
      "239"
   },
   {
      BEETLE_OPT(initial_scanline_pal),
      BEETLE_OPT_INITIAL_SCANLINE_PAL_LABEL_ES,
      NULL,
      BEETLE_OPT_INITIAL_SCANLINE_PAL_INFO_0_ES,
      NULL,
      "video",
      {
         { "0",  OPTION_VAL_0_O48_ES },
         { "1",  NULL },
         { "2",  NULL },
         { "3",  NULL },
         { "4",  NULL },
         { "5",  NULL },
         { "6",  NULL },
         { "7",  NULL },
         { "8",  NULL },
         { "9",  NULL },
         { "10", NULL },
         { "11", NULL },
         { "12", NULL },
         { "13", NULL },
         { "14", NULL },
         { "15", NULL },
         { "16", NULL },
         { "17", NULL },
         { "18", NULL },
         { "19", NULL },
         { "20", NULL },
         { "21", NULL },
         { "22", NULL },
         { "23", NULL },
         { "24", NULL },
         { "25", NULL },
         { "26", NULL },
         { "27", NULL },
         { "28", NULL },
         { "29", NULL },
         { "30", NULL },
         { "31", NULL },
         { "32", NULL },
         { "33", NULL },
         { "34", NULL },
         { "35", NULL },
         { "36", NULL },
         { "37", NULL },
         { "38", NULL },
         { "39", NULL },
         { "40", NULL },
         { NULL, NULL },
      },
      "0"
   },
   {
      BEETLE_OPT(last_scanline_pal),
      BEETLE_OPT_LAST_SCANLINE_PAL_LABEL_ES,
      NULL,
      BEETLE_OPT_LAST_SCANLINE_PAL_INFO_0_ES,
      NULL,
      "video",
      {
         { "230", NULL },
         { "231", NULL },
         { "232", NULL },
         { "233", NULL },
         { "234", NULL },
         { "235", NULL },
         { "236", NULL },
         { "237", NULL },
         { "238", NULL },
         { "239", NULL },
         { "240", NULL },
         { "241", NULL },
         { "242", NULL },
         { "243", NULL },
         { "244", NULL },
         { "245", NULL },
         { "246", NULL },
         { "247", NULL },
         { "248", NULL },
         { "249", NULL },
         { "250", NULL },
         { "251", NULL },
         { "252", NULL },
         { "253", NULL },
         { "254", NULL },
         { "255", NULL },
         { "256", NULL },
         { "257", NULL },
         { "258", NULL },
         { "259", NULL },
         { "260", NULL },
         { "261", NULL },
         { "262", NULL },
         { "263", NULL },
         { "264", NULL },
         { "265", NULL },
         { "266", NULL },
         { "267", NULL },
         { "268", NULL },
         { "269", NULL },
         { "270", NULL },
         { "271", NULL },
         { "272", NULL },
         { "273", NULL },
         { "274", NULL },
         { "275", NULL },
         { "276", NULL },
         { "277", NULL },
         { "278", NULL },
         { "279", NULL },
         { "280", NULL },
         { "281", NULL },
         { "282", NULL },
         { "283", NULL },
         { "284", NULL },
         { "285", NULL },
         { "286", NULL },
         { "287", OPTION_VAL_287_ES },
         { NULL, NULL },
      },
      "287"
   },
   { NULL, NULL, NULL, NULL, NULL, NULL, {{0}}, NULL },
};
struct retro_core_options_v2 options_es = {
   option_cats_es,
   option_defs_es
};

/* RETRO_LANGUAGE_FA */

#define CATEGORY_SYSTEM_LABEL_FA "سیستم"
#define CATEGORY_SYSTEM_INFO_0_FA NULL
#define CATEGORY_VIDEO_LABEL_FA "ویدیو"
#define CATEGORY_VIDEO_INFO_0_FA NULL
#define CATEGORY_PGXP_LABEL_FA NULL
#define CATEGORY_PGXP_INFO_0_FA NULL
#define CATEGORY_OSD_LABEL_FA "نمایش روی صفحه"
#define CATEGORY_OSD_INFO_0_FA NULL
#define CATEGORY_INPUT_LABEL_FA "داده"
#define CATEGORY_INPUT_INFO_0_FA NULL
#define CATEGORY_MEMCARDS_LABEL_FA NULL
#define CATEGORY_MEMCARDS_INFO_0_FA NULL
#define CATEGORY_HACKS_LABEL_FA NULL
#define CATEGORY_HACKS_INFO_0_FA NULL
#define BEETLE_OPT_CPU_FREQ_SCALE_LABEL_FA NULL
#define BEETLE_OPT_CPU_FREQ_SCALE_INFO_0_FA NULL
#define OPTION_VAL_50_FA NULL
#define OPTION_VAL_60_FA NULL
#define OPTION_VAL_70_FA NULL
#define OPTION_VAL_80_FA NULL
#define OPTION_VAL_90_FA NULL
#define OPTION_VAL_100_FA NULL
#define OPTION_VAL_110_FA NULL
#define OPTION_VAL_120_FA NULL
#define OPTION_VAL_130_FA NULL
#define OPTION_VAL_140_FA NULL
#define OPTION_VAL_150_FA NULL
#define OPTION_VAL_160_FA NULL
#define OPTION_VAL_170_FA NULL
#define OPTION_VAL_180_FA NULL
#define OPTION_VAL_190_FA NULL
#define OPTION_VAL_200_FA NULL
#define OPTION_VAL_210_FA NULL
#define OPTION_VAL_220_FA NULL
#define OPTION_VAL_230_FA NULL
#define OPTION_VAL_240_FA NULL
#define OPTION_VAL_250_FA NULL
#define OPTION_VAL_260_FA NULL
#define OPTION_VAL_270_FA NULL
#define OPTION_VAL_280_FA NULL
#define OPTION_VAL_290_FA NULL
#define OPTION_VAL_300_FA NULL
#define OPTION_VAL_310_FA NULL
#define OPTION_VAL_320_FA NULL
#define OPTION_VAL_330_FA NULL
#define OPTION_VAL_340_FA NULL
#define OPTION_VAL_350_FA NULL
#define OPTION_VAL_360_FA NULL
#define OPTION_VAL_370_FA NULL
#define OPTION_VAL_380_FA NULL
#define OPTION_VAL_390_FA NULL
#define OPTION_VAL_400_FA NULL
#define OPTION_VAL_410_FA NULL
#define OPTION_VAL_420_FA NULL
#define OPTION_VAL_430_FA NULL
#define OPTION_VAL_440_FA NULL
#define OPTION_VAL_450_FA NULL
#define OPTION_VAL_460_FA NULL
#define OPTION_VAL_470_FA NULL
#define OPTION_VAL_480_FA NULL
#define OPTION_VAL_490_FA NULL
#define OPTION_VAL_500_FA NULL
#define OPTION_VAL_510_FA NULL
#define OPTION_VAL_520_FA NULL
#define OPTION_VAL_530_FA NULL
#define OPTION_VAL_540_FA NULL
#define OPTION_VAL_550_FA NULL
#define OPTION_VAL_560_FA NULL
#define OPTION_VAL_570_FA NULL
#define OPTION_VAL_580_FA NULL
#define OPTION_VAL_590_FA NULL
#define OPTION_VAL_600_FA NULL
#define OPTION_VAL_610_FA NULL
#define OPTION_VAL_620_FA NULL
#define OPTION_VAL_630_FA NULL
#define OPTION_VAL_640_FA NULL
#define OPTION_VAL_650_FA NULL
#define OPTION_VAL_660_FA NULL
#define OPTION_VAL_670_FA NULL
#define OPTION_VAL_680_FA NULL
#define OPTION_VAL_690_FA NULL
#define OPTION_VAL_700_FA NULL
#define OPTION_VAL_710_FA NULL
#define OPTION_VAL_720_FA NULL
#define OPTION_VAL_730_FA NULL
#define OPTION_VAL_740_FA NULL
#define OPTION_VAL_750_FA NULL
#define BEETLE_OPT_GPU_OVERCLOCK_LABEL_FA NULL
#define BEETLE_OPT_GPU_OVERCLOCK_INFO_0_FA NULL
#define OPTION_VAL_1X_NATIVE_FA NULL
#define OPTION_VAL_2X_FA NULL
#define OPTION_VAL_4X_FA NULL
#define OPTION_VAL_8X_FA NULL
#define OPTION_VAL_16X_FA NULL
#define OPTION_VAL_32X_FA NULL
#define BEETLE_OPT_GTE_OVERCLOCK_LABEL_FA NULL
#define BEETLE_OPT_GTE_OVERCLOCK_INFO_0_FA NULL
#define BEETLE_OPT_CD_ACCESS_METHOD_LABEL_FA NULL
#define BEETLE_OPT_CD_ACCESS_METHOD_INFO_0_FA NULL
#define OPTION_VAL_SYNC_FA NULL
#define OPTION_VAL_ASYNC_FA NULL
#define OPTION_VAL_PRECACHE_FA NULL
#define BEETLE_OPT_CD_FASTLOAD_LABEL_FA NULL
#define BEETLE_OPT_CD_FASTLOAD_INFO_0_FA NULL
#define OPTION_VAL_2X_NATIVE_FA NULL
#define OPTION_VAL_6X_FA NULL
#define OPTION_VAL_10X_FA NULL
#define OPTION_VAL_12X_FA NULL
#define OPTION_VAL_14X_FA NULL
#define BEETLE_OPT_SKIP_BIOS_LABEL_FA NULL
#define BEETLE_OPT_SKIP_BIOS_INFO_0_FA NULL
#define BEETLE_OPT_OVERRIDE_BIOS_LABEL_FA NULL
#define BEETLE_OPT_OVERRIDE_BIOS_INFO_0_FA NULL
#define OPTION_VAL_PSXONPSP_FA NULL
#define OPTION_VAL_PS1_ROM_FA NULL
#define OPTION_VAL_OPENBIOS_FA NULL
#define BEETLE_OPT_REGION_LABEL_FA NULL
#define BEETLE_OPT_REGION_INFO_0_FA NULL
#define OPTION_VAL_AUTO_FA "خودکار"
#define OPTION_VAL_NTSC_J_FA NULL
#define OPTION_VAL_NTSC_U_FA NULL
#define OPTION_VAL_PAL_FA NULL
#define BEETLE_OPT_INTERNAL_RESOLUTION_LABEL_FA NULL
#define BEETLE_OPT_INTERNAL_RESOLUTION_INFO_0_FA NULL
#define BEETLE_OPT_RENDERER_LABEL_FA NULL
#define BEETLE_OPT_RENDERER_INFO_0_FA NULL
#define OPTION_VAL_HARDWARE_FA NULL
#define OPTION_VAL_HARDWARE_GL_FA NULL
#define OPTION_VAL_HARDWARE_VK_FA NULL
#define OPTION_VAL_SOFTWARE_FA NULL
#define BEETLE_OPT_RENDERER_SOFTWARE_FB_LABEL_FA NULL
#define BEETLE_OPT_RENDERER_SOFTWARE_FB_INFO_0_FA NULL
#define BEETLE_OPT_DEPTH_LABEL_FA NULL
#define BEETLE_OPT_DEPTH_INFO_0_FA NULL
#define OPTION_VAL_16BPP_NATIVE_FA NULL
#define OPTION_VAL_32BPP_FA NULL
#define BEETLE_OPT_DITHER_MODE_LABEL_FA NULL
#define BEETLE_OPT_DITHER_MODE_INFO_0_FA NULL
#define OPTION_VAL_INTERNAL_RESOLUTION_FA NULL
#define BEETLE_OPT_SCALED_UV_OFFSET_LABEL_FA NULL
#define BEETLE_OPT_SCALED_UV_OFFSET_INFO_0_FA NULL
#define BEETLE_OPT_FILTER_LABEL_FA NULL
#define BEETLE_OPT_FILTER_INFO_0_FA NULL
#define OPTION_VAL_NEAREST_FA NULL
#define OPTION_VAL_SABR_FA NULL
#define OPTION_VAL_XBR_FA NULL
#define OPTION_VAL_BILINEAR_FA NULL
#define OPTION_VAL_3_POINT_FA NULL
#define OPTION_VAL_JINC2_FA NULL
#define BEETLE_OPT_FILTER_EXCLUDE_SPRITE_LABEL_FA NULL
#define BEETLE_OPT_FILTER_EXCLUDE_SPRITE_INFO_0_FA NULL
#define OPTION_VAL_OPAQUE_FA NULL
#define OPTION_VAL_ALL_FA NULL
#define BEETLE_OPT_FILTER_EXCLUDE_2D_POLYGON_LABEL_FA NULL
#define BEETLE_OPT_FILTER_EXCLUDE_2D_POLYGON_INFO_0_FA NULL
#define BEETLE_OPT_ADAPTIVE_SMOOTHING_LABEL_FA NULL
#define BEETLE_OPT_ADAPTIVE_SMOOTHING_INFO_0_FA NULL
#define BEETLE_OPT_SUPER_SAMPLING_LABEL_FA NULL
#define BEETLE_OPT_SUPER_SAMPLING_INFO_0_FA NULL
#define BEETLE_OPT_MSAA_LABEL_FA NULL
#define BEETLE_OPT_MSAA_INFO_0_FA NULL
#define OPTION_VAL_1X_FA NULL
#define BEETLE_OPT_MDEC_YUV_LABEL_FA NULL
#define BEETLE_OPT_MDEC_YUV_INFO_0_FA NULL
#define BEETLE_OPT_TRACK_TEXTURES_LABEL_FA NULL
#define BEETLE_OPT_TRACK_TEXTURES_INFO_0_FA NULL
#define BEETLE_OPT_DUMP_TEXTURES_LABEL_FA NULL
#define BEETLE_OPT_DUMP_TEXTURES_INFO_0_FA NULL
#define BEETLE_OPT_REPLACE_TEXTURES_LABEL_FA NULL
#define BEETLE_OPT_REPLACE_TEXTURES_INFO_0_FA NULL
#define BEETLE_OPT_PGXP_MODE_LABEL_FA NULL
#define BEETLE_OPT_PGXP_MODE_INFO_0_FA NULL
#define OPTION_VAL_MEMORY_ONLY_FA NULL
#define OPTION_VAL_MEMORY_CPU_FA NULL
#define BEETLE_OPT_PGXP_2D_TOL_LABEL_FA NULL
#define BEETLE_OPT_PGXP_2D_TOL_INFO_0_FA NULL
#define OPTION_VAL_0PX_FA NULL
#define OPTION_VAL_1PX_FA NULL
#define OPTION_VAL_2PX_FA NULL
#define OPTION_VAL_3PX_FA NULL
#define OPTION_VAL_4PX_FA NULL
#define OPTION_VAL_5PX_FA NULL
#define OPTION_VAL_6PX_FA NULL
#define OPTION_VAL_7PX_FA NULL
#define OPTION_VAL_8PX_FA NULL
#define BEETLE_OPT_PGXP_NCLIP_LABEL_FA NULL
#define BEETLE_OPT_PGXP_NCLIP_INFO_0_FA NULL
#define BEETLE_OPT_PGXP_VERTEX_LABEL_FA NULL
#define BEETLE_OPT_PGXP_VERTEX_INFO_0_FA NULL
#define BEETLE_OPT_PGXP_TEXTURE_LABEL_FA NULL
#define BEETLE_OPT_PGXP_TEXTURE_INFO_0_FA NULL
#define BEETLE_OPT_DISPLAY_INTERNAL_FPS_LABEL_FA NULL
#define BEETLE_OPT_DISPLAY_INTERNAL_FPS_INFO_0_FA NULL
#define BEETLE_OPT_DISPLAY_OSD_LABEL_FA NULL
#define BEETLE_OPT_DISPLAY_OSD_INFO_0_FA NULL
#define BEETLE_OPT_DISPLAY_VRAM_LABEL_FA NULL
#define BEETLE_OPT_DISPLAY_VRAM_INFO_0_FA NULL
#define BEETLE_OPT_ANALOG_CALIBRATION_LABEL_FA NULL
#define BEETLE_OPT_ANALOG_CALIBRATION_INFO_0_FA NULL
#define BEETLE_OPT_ANALOG_TOGGLE_LABEL_FA NULL
#define BEETLE_OPT_ANALOG_TOGGLE_INFO_0_FA NULL
#define OPTION_VAL_ENABLED_ANALOG_FA NULL
#define BEETLE_OPT_ANALOG_TOGGLE_COMBO_LABEL_FA NULL
#define BEETLE_OPT_ANALOG_TOGGLE_COMBO_INFO_0_FA NULL
#define OPTION_VAL_L1_L2_R1_R2_START_SELECT_FA NULL
#define OPTION_VAL_L1_R1_SELECT_FA NULL
#define OPTION_VAL_L1_R1_START_FA NULL
#define OPTION_VAL_L1_R1_L3_FA NULL
#define OPTION_VAL_L1_R1_R3_FA NULL
#define OPTION_VAL_L2_R2_SELECT_FA NULL
#define OPTION_VAL_L2_R2_START_FA NULL
#define OPTION_VAL_L2_R2_L3_FA NULL
#define OPTION_VAL_L2_R2_R3_FA NULL
#define OPTION_VAL_L3_R3_FA NULL
#define BEETLE_OPT_ANALOG_TOGGLE_HOLD_LABEL_FA NULL
#define BEETLE_OPT_ANALOG_TOGGLE_HOLD_INFO_0_FA NULL
#define OPTION_VAL_0_FA NULL
#define OPTION_VAL_1_FA NULL
#define OPTION_VAL_2_FA NULL
#define OPTION_VAL_3_FA NULL
#define OPTION_VAL_4_FA NULL
#define OPTION_VAL_5_FA NULL
#define BEETLE_OPT_ENABLE_MULTITAP_PORT1_LABEL_FA NULL
#define BEETLE_OPT_ENABLE_MULTITAP_PORT1_INFO_0_FA NULL
#define BEETLE_OPT_ENABLE_MULTITAP_PORT2_LABEL_FA NULL
#define BEETLE_OPT_ENABLE_MULTITAP_PORT2_INFO_0_FA NULL
#define BEETLE_OPT_GUN_INPUT_MODE_LABEL_FA NULL
#define BEETLE_OPT_GUN_INPUT_MODE_INFO_0_FA NULL
#define OPTION_VAL_LIGHTGUN_FA NULL
#define OPTION_VAL_TOUCHSCREEN_FA NULL
#define BEETLE_OPT_GUN_CURSOR_LABEL_FA NULL
#define BEETLE_OPT_GUN_CURSOR_INFO_0_FA NULL
#define OPTION_VAL_CROSS_FA NULL
#define OPTION_VAL_DOT_FA NULL
#define BEETLE_OPT_CROSSHAIR_COLOR_P1_LABEL_FA NULL
#define BEETLE_OPT_CROSSHAIR_COLOR_P1_INFO_0_FA NULL
#define OPTION_VAL_RED_FA NULL
#define OPTION_VAL_BLUE_FA NULL
#define OPTION_VAL_GREEN_FA NULL
#define OPTION_VAL_ORANGE_FA NULL
#define OPTION_VAL_YELLOW_FA NULL
#define OPTION_VAL_CYAN_FA NULL
#define OPTION_VAL_PINK_FA NULL
#define OPTION_VAL_PURPLE_FA NULL
#define OPTION_VAL_BLACK_FA NULL
#define OPTION_VAL_WHITE_FA NULL
#define BEETLE_OPT_CROSSHAIR_COLOR_P2_LABEL_FA NULL
#define BEETLE_OPT_CROSSHAIR_COLOR_P2_INFO_0_FA NULL
#define BEETLE_OPT_MOUSE_SENSITIVITY_LABEL_FA NULL
#define BEETLE_OPT_MOUSE_SENSITIVITY_INFO_0_FA NULL
#define OPTION_VAL_5_O42_FA NULL
#define OPTION_VAL_10_FA NULL
#define OPTION_VAL_15_FA NULL
#define OPTION_VAL_20_FA NULL
#define OPTION_VAL_25_FA NULL
#define OPTION_VAL_30_FA NULL
#define OPTION_VAL_35_FA NULL
#define OPTION_VAL_40_FA NULL
#define OPTION_VAL_45_FA NULL
#define OPTION_VAL_55_FA NULL
#define OPTION_VAL_65_FA NULL
#define OPTION_VAL_75_FA NULL
#define OPTION_VAL_85_FA NULL
#define OPTION_VAL_95_FA NULL
#define OPTION_VAL_100_O42_FA NULL
#define OPTION_VAL_105_FA NULL
#define OPTION_VAL_115_FA NULL
#define OPTION_VAL_125_FA NULL
#define OPTION_VAL_135_FA NULL
#define OPTION_VAL_145_FA NULL
#define OPTION_VAL_155_FA NULL
#define OPTION_VAL_165_FA NULL
#define OPTION_VAL_175_FA NULL
#define OPTION_VAL_185_FA NULL
#define OPTION_VAL_195_FA NULL
#define BEETLE_OPT_NEGCON_RESPONSE_LABEL_FA NULL
#define BEETLE_OPT_NEGCON_RESPONSE_INFO_0_FA NULL
#define OPTION_VAL_LINEAR_FA NULL
#define OPTION_VAL_QUADRATIC_FA NULL
#define OPTION_VAL_CUBIC_FA NULL
#define BEETLE_OPT_NEGCON_DEADZONE_LABEL_FA NULL
#define BEETLE_OPT_NEGCON_DEADZONE_INFO_0_FA NULL
#define OPTION_VAL_0_O44_FA NULL
#define BEETLE_OPT_USE_MEDNAFEN_MEMCARD0_METHOD_LABEL_FA NULL
#define BEETLE_OPT_USE_MEDNAFEN_MEMCARD0_METHOD_INFO_0_FA NULL
#define OPTION_VAL_LIBRETRO_FA NULL
#define OPTION_VAL_MEDNAFEN_FA NULL
#define BEETLE_OPT_ENABLE_MEMCARD1_LABEL_FA NULL
#define BEETLE_OPT_ENABLE_MEMCARD1_INFO_0_FA NULL
#define BEETLE_OPT_SHARED_MEMORY_CARDS_LABEL_FA NULL
#define BEETLE_OPT_SHARED_MEMORY_CARDS_INFO_0_FA NULL
#define BEETLE_OPT_MEMCARD_LEFT_INDEX_LABEL_FA NULL
#define BEETLE_OPT_MEMCARD_LEFT_INDEX_INFO_0_FA NULL
#define OPTION_VAL_0_O48_FA NULL
#define BEETLE_OPT_MEMCARD_RIGHT_INDEX_LABEL_FA NULL
#define BEETLE_OPT_MEMCARD_RIGHT_INDEX_INFO_0_FA NULL
#define OPTION_VAL_1_O49_FA NULL
#define BEETLE_OPT_LINE_RENDER_LABEL_FA NULL
#define BEETLE_OPT_LINE_RENDER_INFO_0_FA NULL
#define OPTION_VAL_DEFAULT_FA NULL
#define OPTION_VAL_AGGRESSIVE_FA NULL
#define BEETLE_OPT_WIDESCREEN_HACK_LABEL_FA NULL
#define BEETLE_OPT_WIDESCREEN_HACK_INFO_0_FA NULL
#define BEETLE_OPT_WIDESCREEN_HACK_ASPECT_RATIO_LABEL_FA NULL
#define BEETLE_OPT_WIDESCREEN_HACK_ASPECT_RATIO_INFO_0_FA NULL
#define OPTION_VAL_16_9_FA NULL
#define OPTION_VAL_16_10_FA NULL
#define OPTION_VAL_18_9_FA NULL
#define OPTION_VAL_19_9_FA NULL
#define OPTION_VAL_20_9_FA NULL
#define OPTION_VAL_21_9_FA NULL
#define OPTION_VAL_32_9_FA NULL
#define BEETLE_OPT_CPU_DYNAREC_LABEL_FA NULL
#define BEETLE_OPT_CPU_DYNAREC_INFO_0_FA NULL
#define OPTION_VAL_DISABLED_FA NULL
#define OPTION_VAL_EXECUTE_FA NULL
#define OPTION_VAL_EXECUTE_ONE_FA NULL
#define OPTION_VAL_RUN_INTERPRETER_FA NULL
#define BEETLE_OPT_DYNAREC_INVALIDATE_LABEL_FA NULL
#define BEETLE_OPT_DYNAREC_INVALIDATE_INFO_0_FA NULL
#define OPTION_VAL_FULL_FA NULL
#define OPTION_VAL_DMA_FA NULL
#define BEETLE_OPT_DYNAREC_EVENTCYCLES_LABEL_FA NULL
#define BEETLE_OPT_DYNAREC_EVENTCYCLES_INFO_0_FA NULL
#define OPTION_VAL_128_FA NULL
#define BEETLE_OPT_DYNAREC_SPU_SAMPLES_LABEL_FA NULL
#define BEETLE_OPT_DYNAREC_SPU_SAMPLES_INFO_0_FA NULL
#define BEETLE_OPT_CORE_TIMING_FPS_LABEL_FA NULL
#define BEETLE_OPT_CORE_TIMING_FPS_INFO_0_FA NULL
#define OPTION_VAL_AUTO_TOGGLE_FA NULL
#define OPTION_VAL_FORCE_PROGRESSIVE_FA NULL
#define OPTION_VAL_FORCE_INTERLACED_FA NULL
#define BEETLE_OPT_DEINTERLACER_LABEL_FA NULL
#define BEETLE_OPT_DEINTERLACER_INFO_0_FA NULL
#define OPTION_VAL_WEAVE_FA NULL
#define OPTION_VAL_BOB_FA NULL
#define OPTION_VAL_BOB_OFFSET_FA NULL
#define OPTION_VAL_FASTMAD_FA NULL
#define BEETLE_OPT_FRAME_DUPING_LABEL_FA NULL
#define BEETLE_OPT_FRAME_DUPING_INFO_0_FA NULL
#define BEETLE_OPT_PAL_VIDEO_TIMING_OVERRIDE_LABEL_FA NULL
#define BEETLE_OPT_PAL_VIDEO_TIMING_OVERRIDE_INFO_0_FA NULL
#define BEETLE_OPT_ASPECT_RATIO_LABEL_FA NULL
#define BEETLE_OPT_ASPECT_RATIO_INFO_0_FA NULL
#define OPTION_VAL_CORRECTED_FA NULL
#define OPTION_VAL_UNCORRECTED_FA NULL
#define OPTION_VAL_4_3_FA NULL
#define OPTION_VAL_NTSC_FA NULL
#define BEETLE_OPT_CROP_OVERSCAN_LABEL_FA NULL
#define BEETLE_OPT_CROP_OVERSCAN_INFO_0_FA NULL
#define OPTION_VAL_STATIC_FA NULL
#define OPTION_VAL_SMART_FA NULL
#define BEETLE_OPT_IMAGE_CROP_LABEL_FA NULL
#define BEETLE_OPT_IMAGE_CROP_INFO_0_FA NULL
#define OPTION_VAL_9PX_FA NULL
#define OPTION_VAL_10PX_FA NULL
#define OPTION_VAL_11PX_FA NULL
#define OPTION_VAL_12PX_FA NULL
#define OPTION_VAL_13PX_FA NULL
#define OPTION_VAL_14PX_FA NULL
#define OPTION_VAL_15PX_FA NULL
#define OPTION_VAL_16PX_FA NULL
#define OPTION_VAL_17PX_FA NULL
#define OPTION_VAL_18PX_FA NULL
#define OPTION_VAL_19PX_FA NULL
#define OPTION_VAL_20PX_FA NULL
#define BEETLE_OPT_IMAGE_OFFSET_LABEL_FA NULL
#define BEETLE_OPT_IMAGE_OFFSET_INFO_0_FA NULL
#define OPTION_VAL_12PX_O64_FA NULL
#define OPTION_VAL_11PX_O64_FA NULL
#define OPTION_VAL_10PX_O64_FA NULL
#define OPTION_VAL_9PX_O64_FA NULL
#define OPTION_VAL_8PX_O64_FA NULL
#define OPTION_VAL_7PX_O64_FA NULL
#define OPTION_VAL_6PX_O64_FA NULL
#define OPTION_VAL_5PX_O64_FA NULL
#define OPTION_VAL_4PX_O64_FA NULL
#define OPTION_VAL_3PX_O64_FA NULL
#define OPTION_VAL_2PX_O64_FA NULL
#define OPTION_VAL_1PX_O64_FA NULL
#define OPTION_VAL_1PX_O64_0_FA NULL
#define OPTION_VAL_2PX_O64_0_FA NULL
#define OPTION_VAL_3PX_O64_0_FA NULL
#define OPTION_VAL_4PX_O64_0_FA NULL
#define OPTION_VAL_5PX_O64_0_FA NULL
#define OPTION_VAL_6PX_O64_0_FA NULL
#define OPTION_VAL_7PX_O64_0_FA NULL
#define OPTION_VAL_8PX_O64_0_FA NULL
#define OPTION_VAL_9PX_O64_0_FA NULL
#define OPTION_VAL_10PX_O64_0_FA NULL
#define OPTION_VAL_11PX_O64_0_FA NULL
#define OPTION_VAL_12PX_O64_0_FA NULL
#define BEETLE_OPT_IMAGE_OFFSET_CYCLES_LABEL_FA NULL
#define BEETLE_OPT_IMAGE_OFFSET_CYCLES_INFO_0_FA NULL
#define BEETLE_OPT_INITIAL_SCANLINE_LABEL_FA NULL
#define BEETLE_OPT_INITIAL_SCANLINE_INFO_0_FA NULL
#define BEETLE_OPT_LAST_SCANLINE_LABEL_FA NULL
#define BEETLE_OPT_LAST_SCANLINE_INFO_0_FA NULL
#define OPTION_VAL_239_FA NULL
#define BEETLE_OPT_INITIAL_SCANLINE_PAL_LABEL_FA NULL
#define BEETLE_OPT_INITIAL_SCANLINE_PAL_INFO_0_FA NULL
#define BEETLE_OPT_LAST_SCANLINE_PAL_LABEL_FA NULL
#define BEETLE_OPT_LAST_SCANLINE_PAL_INFO_0_FA NULL
#define OPTION_VAL_287_FA NULL

struct retro_core_option_v2_category option_cats_fa[] = {
   {
      "system",
      CATEGORY_SYSTEM_LABEL_FA,
      CATEGORY_SYSTEM_INFO_0_FA
   },
   {
      "video",
      CATEGORY_VIDEO_LABEL_FA,
      CATEGORY_VIDEO_INFO_0_FA
   },
   {
      "pgxp",
      CATEGORY_PGXP_LABEL_FA,
      CATEGORY_PGXP_INFO_0_FA
   },
   {
      "osd",
      CATEGORY_OSD_LABEL_FA,
      CATEGORY_OSD_INFO_0_FA
   },
   {
      "input",
      CATEGORY_INPUT_LABEL_FA,
      CATEGORY_INPUT_INFO_0_FA
   },
   {
      "memcards",
      CATEGORY_MEMCARDS_LABEL_FA,
      CATEGORY_MEMCARDS_INFO_0_FA
   },
   {
      "hacks",
      CATEGORY_HACKS_LABEL_FA,
      CATEGORY_HACKS_INFO_0_FA
   },
   { NULL, NULL, NULL },
};
struct retro_core_option_v2_definition option_defs_fa[] = {
   {
      BEETLE_OPT(cpu_freq_scale),
      BEETLE_OPT_CPU_FREQ_SCALE_LABEL_FA,
      NULL,
      BEETLE_OPT_CPU_FREQ_SCALE_INFO_0_FA,
      NULL,
      "system",
      {
         { "50%",           OPTION_VAL_50_FA },
         { "60%",           OPTION_VAL_60_FA },
         { "70%",           OPTION_VAL_70_FA },
         { "80%",           OPTION_VAL_80_FA },
         { "90%",           OPTION_VAL_90_FA },
         { "100%",          OPTION_VAL_100_FA },
         { "110%",          OPTION_VAL_110_FA },
         { "120%",          OPTION_VAL_120_FA },
         { "130%",          OPTION_VAL_130_FA },
         { "140%",          OPTION_VAL_140_FA },
         { "150%",          OPTION_VAL_150_FA },
         { "160%",          OPTION_VAL_160_FA },
         { "170%",          OPTION_VAL_170_FA },
         { "180%",          OPTION_VAL_180_FA },
         { "190%",          OPTION_VAL_190_FA },
         { "200%",          OPTION_VAL_200_FA },
         { "210%",          OPTION_VAL_210_FA },
         { "220%",          OPTION_VAL_220_FA },
         { "230%",          OPTION_VAL_230_FA },
         { "240%",          OPTION_VAL_240_FA },
         { "250%",          OPTION_VAL_250_FA },
         { "260%",          OPTION_VAL_260_FA },
         { "270%",          OPTION_VAL_270_FA },
         { "280%",          OPTION_VAL_280_FA },
         { "290%",          OPTION_VAL_290_FA },
         { "300%",          OPTION_VAL_300_FA },
         { "310%",          OPTION_VAL_310_FA },
         { "320%",          OPTION_VAL_320_FA },
         { "330%",          OPTION_VAL_330_FA },
         { "340%",          OPTION_VAL_340_FA },
         { "350%",          OPTION_VAL_350_FA },
         { "360%",          OPTION_VAL_360_FA },
         { "370%",          OPTION_VAL_370_FA },
         { "380%",          OPTION_VAL_380_FA },
         { "390%",          OPTION_VAL_390_FA },
         { "400%",          OPTION_VAL_400_FA },
         { "410%",          OPTION_VAL_410_FA },
         { "420%",          OPTION_VAL_420_FA },
         { "430%",          OPTION_VAL_430_FA },
         { "440%",          OPTION_VAL_440_FA },
         { "450%",          OPTION_VAL_450_FA },
         { "460%",          OPTION_VAL_460_FA },
         { "470%",          OPTION_VAL_470_FA },
         { "480%",          OPTION_VAL_480_FA },
         { "490%",          OPTION_VAL_490_FA },
         { "500%",          OPTION_VAL_500_FA },
         { "510%",          OPTION_VAL_510_FA },
         { "520%",          OPTION_VAL_520_FA },
         { "530%",          OPTION_VAL_530_FA },
         { "540%",          OPTION_VAL_540_FA },
         { "550%",          OPTION_VAL_550_FA },
         { "560%",          OPTION_VAL_560_FA },
         { "570%",          OPTION_VAL_570_FA },
         { "580%",          OPTION_VAL_580_FA },
         { "590%",          OPTION_VAL_590_FA },
         { "600%",          OPTION_VAL_600_FA },
         { "610%",          OPTION_VAL_610_FA },
         { "620%",          OPTION_VAL_620_FA },
         { "630%",          OPTION_VAL_630_FA },
         { "640%",          OPTION_VAL_640_FA },
         { "650%",          OPTION_VAL_650_FA },
         { "660%",          OPTION_VAL_660_FA },
         { "670%",          OPTION_VAL_670_FA },
         { "680%",          OPTION_VAL_680_FA },
         { "690%",          OPTION_VAL_690_FA },
         { "700%",          OPTION_VAL_700_FA },
         { "710%",          OPTION_VAL_710_FA },
         { "720%",          OPTION_VAL_720_FA },
         { "730%",          OPTION_VAL_730_FA },
         { "740%",          OPTION_VAL_740_FA },
         { "750%",          OPTION_VAL_750_FA },
         { NULL, NULL },
      },
      "100%"
   },
   {
      BEETLE_OPT(gpu_overclock),
      BEETLE_OPT_GPU_OVERCLOCK_LABEL_FA,
      NULL,
      BEETLE_OPT_GPU_OVERCLOCK_INFO_0_FA,
      NULL,
      "system",
      {
         { "1x(native)", OPTION_VAL_1X_NATIVE_FA },
         { "2x",         OPTION_VAL_2X_FA },
         { "4x",         OPTION_VAL_4X_FA },
         { "8x",         OPTION_VAL_8X_FA },
         { "16x",        OPTION_VAL_16X_FA },
         { "32x",        OPTION_VAL_32X_FA },
         { NULL, NULL },
      },
      "1x(native)"
   },
   {
      BEETLE_OPT(gte_overclock),
      BEETLE_OPT_GTE_OVERCLOCK_LABEL_FA,
      NULL,
      BEETLE_OPT_GTE_OVERCLOCK_INFO_0_FA,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
#ifndef EMSCRIPTEN
   {
      BEETLE_OPT(cd_access_method),
      BEETLE_OPT_CD_ACCESS_METHOD_LABEL_FA,
      NULL,
      BEETLE_OPT_CD_ACCESS_METHOD_INFO_0_FA,
      NULL,
      "system",
      {
         { "sync",     OPTION_VAL_SYNC_FA },
         { "async",    OPTION_VAL_ASYNC_FA },
         { "precache", OPTION_VAL_PRECACHE_FA },
         { NULL, NULL },
      },
      "sync"
   },
#endif
   {
      BEETLE_OPT(cd_fastload),
      BEETLE_OPT_CD_FASTLOAD_LABEL_FA,
      NULL,
      BEETLE_OPT_CD_FASTLOAD_INFO_0_FA,
      NULL,
      "system",
      {
         { "2x(native)", OPTION_VAL_2X_NATIVE_FA },
         { "4x",          OPTION_VAL_4X_FA },
         { "6x",          OPTION_VAL_6X_FA },
         { "8x",          OPTION_VAL_8X_FA },
         { "10x",         OPTION_VAL_10X_FA },
         { "12x",         OPTION_VAL_12X_FA },
         { "14x",         OPTION_VAL_14X_FA },
         { NULL, NULL },
      },
      "2x(native)"
   },
   {
      BEETLE_OPT(skip_bios),
      BEETLE_OPT_SKIP_BIOS_LABEL_FA,
      NULL,
      BEETLE_OPT_SKIP_BIOS_INFO_0_FA,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(override_bios),
      BEETLE_OPT_OVERRIDE_BIOS_LABEL_FA,
      NULL,
      BEETLE_OPT_OVERRIDE_BIOS_INFO_0_FA,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "psxonpsp", OPTION_VAL_PSXONPSP_FA },
         { "ps1_rom",  OPTION_VAL_PS1_ROM_FA },
         { "openbios", OPTION_VAL_OPENBIOS_FA     },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(region),
      BEETLE_OPT_REGION_LABEL_FA,
      NULL,
      BEETLE_OPT_REGION_INFO_0_FA,
      NULL,
      "system",
      {
         { "auto",   OPTION_VAL_AUTO_FA },
         { "ntsc-j", OPTION_VAL_NTSC_J_FA },
         { "ntsc-u", OPTION_VAL_NTSC_U_FA },
         { "pal",    OPTION_VAL_PAL_FA },
         { NULL, NULL },
      },
      "auto"
   },
   {
      BEETLE_OPT(internal_resolution),
      BEETLE_OPT_INTERNAL_RESOLUTION_LABEL_FA,
      NULL,
      BEETLE_OPT_INTERNAL_RESOLUTION_INFO_0_FA,
      NULL,
      "video",
      {
         { "1x(native)", OPTION_VAL_1X_NATIVE_FA },
         { "2x",         OPTION_VAL_2X_FA },
         { "4x",         OPTION_VAL_4X_FA },
         { "8x",         OPTION_VAL_8X_FA },
         { "16x",        OPTION_VAL_16X_FA },
         { NULL, NULL },
      },
      "1x(native)"
   },
#if defined(HAVE_OPENGL) || defined(HAVE_OPENGLES) || defined(HAVE_VULKAN)
   {
      BEETLE_OPT(renderer),
      BEETLE_OPT_RENDERER_LABEL_FA,
      NULL,
      BEETLE_OPT_RENDERER_INFO_0_FA,
      NULL,
      "video",
      {
         { "hardware",    OPTION_VAL_HARDWARE_FA },
#if defined(HAVE_OPENGL) || defined(HAVE_OPENGLES)
         { "hardware_gl", OPTION_VAL_HARDWARE_GL_FA },
#endif
#if defined(HAVE_VULKAN)
         { "hardware_vk", OPTION_VAL_HARDWARE_VK_FA },
#endif
         { "software",    OPTION_VAL_SOFTWARE_FA },
         { NULL, NULL },
      },
      "hardware"
   },
   {
      BEETLE_OPT(renderer_software_fb),
      BEETLE_OPT_RENDERER_SOFTWARE_FB_LABEL_FA,
      NULL,
      BEETLE_OPT_RENDERER_SOFTWARE_FB_INFO_0_FA,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled"
   },
#endif
#if defined(HAVE_OPENGL) || defined(HAVE_OPENGLES)
   {
      BEETLE_OPT(depth),
      BEETLE_OPT_DEPTH_LABEL_FA,
      NULL,
      BEETLE_OPT_DEPTH_INFO_0_FA,
      NULL,
      "video",
      {
         { "16bpp(native)", OPTION_VAL_16BPP_NATIVE_FA },
         { "32bpp",         OPTION_VAL_32BPP_FA },
         { NULL, NULL },
      },
      "16bpp(native)"
   },
   // Sort of, it's more like 15-bit high color and 24-bit true color for visible output. The alpha channel is used for mask bit. Vulkan renderer uses ABGR1555_555 for 31 bits internal? FMVs are always 24-bit on all renderers like original hardware (BGR888, no alpha)
#endif
   {
      BEETLE_OPT(dither_mode),
      BEETLE_OPT_DITHER_MODE_LABEL_FA,
      NULL,
      BEETLE_OPT_DITHER_MODE_INFO_0_FA,
      NULL,
      "video",
      {
         { "1x(native)",          OPTION_VAL_1X_NATIVE_FA },
         { "internal resolution", OPTION_VAL_INTERNAL_RESOLUTION_FA },
         { "disabled",            NULL },
         { NULL, NULL },
      },
      "1x(native)"
   },
#ifdef HAVE_VULKAN
   {
      BEETLE_OPT(scaled_uv_offset),
      BEETLE_OPT_SCALED_UV_OFFSET_LABEL_FA,
      NULL,
      BEETLE_OPT_SCALED_UV_OFFSET_INFO_0_FA,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled"
   },
#endif
#if defined(HAVE_OPENGL) || defined(HAVE_OPENGLES) || defined(HAVE_VULKAN)
   {
      BEETLE_OPT(filter),
      BEETLE_OPT_FILTER_LABEL_FA,
      NULL,
      BEETLE_OPT_FILTER_INFO_0_FA,
      NULL,
      "video",
      {
         { "nearest",  OPTION_VAL_NEAREST_FA },
         { "SABR",     OPTION_VAL_SABR_FA },
         { "xBR",      OPTION_VAL_XBR_FA },
         { "bilinear", OPTION_VAL_BILINEAR_FA },
         { "3-point",  OPTION_VAL_3_POINT_FA },
         { "JINC2",    OPTION_VAL_JINC2_FA },
         { NULL, NULL },
      },
      "nearest"
   },
#ifdef HAVE_VULKAN
   {
      BEETLE_OPT(filter_exclude_sprite),
      BEETLE_OPT_FILTER_EXCLUDE_SPRITE_LABEL_FA,
      NULL,
      BEETLE_OPT_FILTER_EXCLUDE_SPRITE_INFO_0_FA,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "opaque", OPTION_VAL_OPAQUE_FA },
         { "all", OPTION_VAL_ALL_FA },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(filter_exclude_2d_polygon),
      BEETLE_OPT_FILTER_EXCLUDE_2D_POLYGON_LABEL_FA,
      NULL,
      BEETLE_OPT_FILTER_EXCLUDE_2D_POLYGON_INFO_0_FA,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "opaque", OPTION_VAL_OPAQUE_FA },
         { "all", OPTION_VAL_ALL_FA },
         { NULL, NULL },
      },
      "disabled"
   },
#endif
#endif
#ifdef HAVE_VULKAN
   {
      BEETLE_OPT(adaptive_smoothing),
      BEETLE_OPT_ADAPTIVE_SMOOTHING_LABEL_FA,
      NULL,
      BEETLE_OPT_ADAPTIVE_SMOOTHING_INFO_0_FA,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(super_sampling),
      BEETLE_OPT_SUPER_SAMPLING_LABEL_FA,
      NULL,
      BEETLE_OPT_SUPER_SAMPLING_INFO_0_FA,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(msaa),
      BEETLE_OPT_MSAA_LABEL_FA,
      NULL,
      BEETLE_OPT_MSAA_INFO_0_FA,
      NULL,
      "video",
      {
         { "1x",  OPTION_VAL_1X_FA },
         { "2x",  OPTION_VAL_2X_FA },
         { "4x",  OPTION_VAL_4X_FA },
         { "8x",  OPTION_VAL_8X_FA },
         { "16x", OPTION_VAL_16X_FA },
         { NULL, NULL },
      },
      "1x"
   },
   {
      BEETLE_OPT(mdec_yuv),
      BEETLE_OPT_MDEC_YUV_LABEL_FA,
      NULL,
      BEETLE_OPT_MDEC_YUV_INFO_0_FA,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(track_textures),
      BEETLE_OPT_TRACK_TEXTURES_LABEL_FA,
      NULL,
      BEETLE_OPT_TRACK_TEXTURES_INFO_0_FA,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
#ifdef TEXTURE_DUMPING_ENABLED
   {
      BEETLE_OPT(dump_textures),
      BEETLE_OPT_DUMP_TEXTURES_LABEL_FA,
      NULL,
      BEETLE_OPT_DUMP_TEXTURES_INFO_0_FA,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
#endif
   {
      BEETLE_OPT(replace_textures),
      BEETLE_OPT_REPLACE_TEXTURES_LABEL_FA,
      NULL,
      BEETLE_OPT_REPLACE_TEXTURES_INFO_0_FA,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
#endif
   {
      BEETLE_OPT(pgxp_mode),
      BEETLE_OPT_PGXP_MODE_LABEL_FA,
      NULL,
      BEETLE_OPT_PGXP_MODE_INFO_0_FA,
      NULL,
      "pgxp",
      {
         { "disabled",     NULL },
         { "memory only",  OPTION_VAL_MEMORY_ONLY_FA },
         { "memory + CPU", OPTION_VAL_MEMORY_CPU_FA },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(pgxp_2d_tol),
      BEETLE_OPT_PGXP_2D_TOL_LABEL_FA,
      NULL,
      BEETLE_OPT_PGXP_2D_TOL_INFO_0_FA,
      NULL,
      "pgxp",
      {
         { "disabled", NULL },
         { "0px", OPTION_VAL_0PX_FA },
         { "1px", OPTION_VAL_1PX_FA },
         { "2px", OPTION_VAL_2PX_FA },
         { "3px", OPTION_VAL_3PX_FA },
         { "4px", OPTION_VAL_4PX_FA },
         { "5px", OPTION_VAL_5PX_FA },
         { "6px", OPTION_VAL_6PX_FA },
         { "7px", OPTION_VAL_7PX_FA },
         { "8px", OPTION_VAL_8PX_FA },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(pgxp_nclip),
      BEETLE_OPT_PGXP_NCLIP_LABEL_FA,
      NULL,
      BEETLE_OPT_PGXP_NCLIP_INFO_0_FA,
      NULL,
      "pgxp",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
#if defined(HAVE_OPENGL) || defined(HAVE_OPENGLES) || defined(HAVE_VULKAN)
   {
      BEETLE_OPT(pgxp_vertex),
      BEETLE_OPT_PGXP_VERTEX_LABEL_FA,
      NULL,
      BEETLE_OPT_PGXP_VERTEX_INFO_0_FA,
      NULL,
      "pgxp",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
#endif
   {
      BEETLE_OPT(pgxp_texture),
      BEETLE_OPT_PGXP_TEXTURE_LABEL_FA,
      NULL,
      BEETLE_OPT_PGXP_TEXTURE_INFO_0_FA,
      NULL,
      "pgxp",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(display_internal_fps),
      BEETLE_OPT_DISPLAY_INTERNAL_FPS_LABEL_FA,
      NULL,
      BEETLE_OPT_DISPLAY_INTERNAL_FPS_INFO_0_FA,
      NULL,
      "osd",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(display_osd),
      BEETLE_OPT_DISPLAY_OSD_LABEL_FA,
      NULL,
      BEETLE_OPT_DISPLAY_OSD_INFO_0_FA,
      NULL,
      "osd",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled"
   },
#if defined(HAVE_OPENGL) || defined(HAVE_OPENGLES) || defined(HAVE_VULKAN)
   {
      BEETLE_OPT(display_vram),
      BEETLE_OPT_DISPLAY_VRAM_LABEL_FA,
      NULL,
      BEETLE_OPT_DISPLAY_VRAM_INFO_0_FA,
      NULL,
      "osd",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
#endif
   {
      BEETLE_OPT(analog_calibration),
      BEETLE_OPT_ANALOG_CALIBRATION_LABEL_FA,
      NULL,
      BEETLE_OPT_ANALOG_CALIBRATION_INFO_0_FA,
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(analog_toggle),
      BEETLE_OPT_ANALOG_TOGGLE_LABEL_FA,
      NULL,
      BEETLE_OPT_ANALOG_TOGGLE_INFO_0_FA,
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { "enabled-analog", OPTION_VAL_ENABLED_ANALOG_FA },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(analog_toggle_combo),
      BEETLE_OPT_ANALOG_TOGGLE_COMBO_LABEL_FA,
      NULL,
      BEETLE_OPT_ANALOG_TOGGLE_COMBO_INFO_0_FA,
      NULL,
      "input",
      {
         { "l1+l2+r1+r2+start+select", OPTION_VAL_L1_L2_R1_R2_START_SELECT_FA },
         { "l1+r1+select",             OPTION_VAL_L1_R1_SELECT_FA },
         { "l1+r1+start",              OPTION_VAL_L1_R1_START_FA },
         { "l1+r1+l3",                 OPTION_VAL_L1_R1_L3_FA },
         { "l1+r1+r3",                 OPTION_VAL_L1_R1_R3_FA },
         { "l2+r2+select",             OPTION_VAL_L2_R2_SELECT_FA },
         { "l2+r2+start",              OPTION_VAL_L2_R2_START_FA },
         { "l2+r2+l3",                 OPTION_VAL_L2_R2_L3_FA },
         { "l2+r2+r3",                 OPTION_VAL_L2_R2_R3_FA },
         { "l3+r3",                    OPTION_VAL_L3_R3_FA },
         { NULL, NULL },
      },
      "l1+r1+select"
   },
   {
      BEETLE_OPT(analog_toggle_hold),
      BEETLE_OPT_ANALOG_TOGGLE_HOLD_LABEL_FA,
      NULL,
      BEETLE_OPT_ANALOG_TOGGLE_HOLD_INFO_0_FA,
      NULL,
      "input",
      {
         { "0", OPTION_VAL_0_FA },
         { "1", OPTION_VAL_1_FA },
         { "2", OPTION_VAL_2_FA },
         { "3", OPTION_VAL_3_FA },
         { "4", OPTION_VAL_4_FA },
         { "5", OPTION_VAL_5_FA },
         { NULL, NULL },
      },
      "1"
   },
   {
      BEETLE_OPT(enable_multitap_port1),
      BEETLE_OPT_ENABLE_MULTITAP_PORT1_LABEL_FA,
      NULL,
      BEETLE_OPT_ENABLE_MULTITAP_PORT1_INFO_0_FA,
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(enable_multitap_port2),
      BEETLE_OPT_ENABLE_MULTITAP_PORT2_LABEL_FA,
      NULL,
      BEETLE_OPT_ENABLE_MULTITAP_PORT2_INFO_0_FA,
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(gun_input_mode),
      BEETLE_OPT_GUN_INPUT_MODE_LABEL_FA,
      NULL,
      BEETLE_OPT_GUN_INPUT_MODE_INFO_0_FA,
      NULL,
      "input",
      {
         { "lightgun",    OPTION_VAL_LIGHTGUN_FA },
         { "touchscreen", OPTION_VAL_TOUCHSCREEN_FA },
         { NULL, NULL },
      },
      "lightgun"
   },
   // Shouldn't the gun_input_mode just be Mouse vs. Touchscreen?
   {
      BEETLE_OPT(gun_cursor),
      BEETLE_OPT_GUN_CURSOR_LABEL_FA,
      NULL,
      BEETLE_OPT_GUN_CURSOR_INFO_0_FA,
      NULL,
      "input",
      {
         { "off",   "disabled" },
         { "cross", OPTION_VAL_CROSS_FA },
         { "dot",   OPTION_VAL_DOT_FA },
         { NULL, NULL },
      },
      "cross"
   },
   {
      BEETLE_OPT(crosshair_color_p1),
      BEETLE_OPT_CROSSHAIR_COLOR_P1_LABEL_FA,
      NULL,
      BEETLE_OPT_CROSSHAIR_COLOR_P1_INFO_0_FA,
      NULL,
      "input",
      {
         { "red",    OPTION_VAL_RED_FA },
         { "blue",   OPTION_VAL_BLUE_FA },
         { "green",  OPTION_VAL_GREEN_FA },
         { "orange", OPTION_VAL_ORANGE_FA },
         { "yellow", OPTION_VAL_YELLOW_FA },
         { "cyan",   OPTION_VAL_CYAN_FA },
         { "pink",   OPTION_VAL_PINK_FA },
         { "purple", OPTION_VAL_PURPLE_FA },
         { "black",  OPTION_VAL_BLACK_FA },
         { "white",  OPTION_VAL_WHITE_FA },
         { NULL, NULL },
      },
      "red"
   },
   {
      BEETLE_OPT(crosshair_color_p2),
      BEETLE_OPT_CROSSHAIR_COLOR_P2_LABEL_FA,
      NULL,
      BEETLE_OPT_CROSSHAIR_COLOR_P2_INFO_0_FA,
      NULL,
      "input",
      {
         { "blue",   OPTION_VAL_BLUE_FA },
         { "red",    OPTION_VAL_RED_FA },
         { "green",  OPTION_VAL_GREEN_FA },
         { "orange", OPTION_VAL_ORANGE_FA },
         { "yellow", OPTION_VAL_YELLOW_FA },
         { "cyan",   OPTION_VAL_CYAN_FA },
         { "pink",   OPTION_VAL_PINK_FA },
         { "purple", OPTION_VAL_PURPLE_FA },
         { "black",  OPTION_VAL_BLACK_FA },
         { "white",  OPTION_VAL_WHITE_FA },
         { NULL, NULL },
      },
      "blue"
   },
   {
      BEETLE_OPT(mouse_sensitivity),
      BEETLE_OPT_MOUSE_SENSITIVITY_LABEL_FA,
      NULL,
      BEETLE_OPT_MOUSE_SENSITIVITY_INFO_0_FA,
      NULL,
      "input",
      {
         { "5%",   OPTION_VAL_5_O42_FA },
         { "10%",  OPTION_VAL_10_FA },
         { "15%",  OPTION_VAL_15_FA },
         { "20%",  OPTION_VAL_20_FA },
         { "25%",  OPTION_VAL_25_FA },
         { "30%",  OPTION_VAL_30_FA },
         { "35%",  OPTION_VAL_35_FA },
         { "40%",  OPTION_VAL_40_FA },
         { "45%",  OPTION_VAL_45_FA },
         { "50%",  OPTION_VAL_50_FA },
         { "55%",  OPTION_VAL_55_FA },
         { "60%",  OPTION_VAL_60_FA },
         { "65%",  OPTION_VAL_65_FA },
         { "70%",  OPTION_VAL_70_FA },
         { "75%",  OPTION_VAL_75_FA },
         { "80%",  OPTION_VAL_80_FA },
         { "85%",  OPTION_VAL_85_FA },
         { "90%",  OPTION_VAL_90_FA },
         { "95%",  OPTION_VAL_95_FA },
         { "100%", OPTION_VAL_100_O42_FA },
         { "105%", OPTION_VAL_105_FA },
         { "110%", OPTION_VAL_110_FA },
         { "115%", OPTION_VAL_115_FA },
         { "120%", OPTION_VAL_120_FA },
         { "125%", OPTION_VAL_125_FA },
         { "130%", OPTION_VAL_130_FA },
         { "135%", OPTION_VAL_135_FA },
         { "140%", OPTION_VAL_140_FA },
         { "145%", OPTION_VAL_145_FA },
         { "150%", OPTION_VAL_150_FA },
         { "155%", OPTION_VAL_155_FA },
         { "160%", OPTION_VAL_160_FA },
         { "165%", OPTION_VAL_165_FA },
         { "170%", OPTION_VAL_170_FA },
         { "175%", OPTION_VAL_175_FA },
         { "180%", OPTION_VAL_180_FA },
         { "185%", OPTION_VAL_185_FA },
         { "190%", OPTION_VAL_190_FA },
         { "195%", OPTION_VAL_195_FA },
         { "200%", OPTION_VAL_200_FA },
         { NULL, NULL },
      },
      "100%"
   },
   {
      BEETLE_OPT(negcon_response),
      BEETLE_OPT_NEGCON_RESPONSE_LABEL_FA,
      NULL,
      BEETLE_OPT_NEGCON_RESPONSE_INFO_0_FA,
      NULL,
      "input",
      {
         { "linear",    OPTION_VAL_LINEAR_FA },
         { "quadratic", OPTION_VAL_QUADRATIC_FA },
         { "cubic",     OPTION_VAL_CUBIC_FA },
         { NULL, NULL },
      },
      "linear"
   },
   {
      BEETLE_OPT(negcon_deadzone),
      BEETLE_OPT_NEGCON_DEADZONE_LABEL_FA,
      NULL,
      BEETLE_OPT_NEGCON_DEADZONE_INFO_0_FA,
      NULL,
      "input",
      {
         { "0%",  OPTION_VAL_0_O44_FA },
         { "5%",  OPTION_VAL_5_O42_FA },
         { "10%", OPTION_VAL_10_FA },
         { "15%", OPTION_VAL_15_FA },
         { "20%", OPTION_VAL_20_FA },
         { "25%", OPTION_VAL_25_FA },
         { "30%", OPTION_VAL_30_FA },
         { NULL, NULL },
      },
      "0%"
   },
   {
      BEETLE_OPT(use_mednafen_memcard0_method),
      BEETLE_OPT_USE_MEDNAFEN_MEMCARD0_METHOD_LABEL_FA,
      NULL,
      BEETLE_OPT_USE_MEDNAFEN_MEMCARD0_METHOD_INFO_0_FA,
      NULL,
      "memcards",
      {
         { "libretro", OPTION_VAL_LIBRETRO_FA },
         { "mednafen", OPTION_VAL_MEDNAFEN_FA },
         { NULL, NULL },
      },
      "libretro"
   },
   {
      BEETLE_OPT(enable_memcard1),
      BEETLE_OPT_ENABLE_MEMCARD1_LABEL_FA,
      NULL,
      BEETLE_OPT_ENABLE_MEMCARD1_INFO_0_FA,
      NULL,
      "memcards",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(shared_memory_cards),
      BEETLE_OPT_SHARED_MEMORY_CARDS_LABEL_FA,
      NULL,
      BEETLE_OPT_SHARED_MEMORY_CARDS_INFO_0_FA,
      NULL,
      "memcards",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(memcard_left_index),
      BEETLE_OPT_MEMCARD_LEFT_INDEX_LABEL_FA,
      NULL,
      BEETLE_OPT_MEMCARD_LEFT_INDEX_INFO_0_FA,
      NULL,
      "memcards",
      {
         { "0",  OPTION_VAL_0_O48_FA },
         { "1",  NULL },
         { "2",  NULL },
         { "3",  NULL },
         { "4",  NULL },
         { "5",  NULL },
         { "6",  NULL },
         { "7",  NULL },
         { "8",  NULL },
         { "9",  NULL },
         { "10", NULL },
         { "11", NULL },
         { "12", NULL },
         { "13", NULL },
         { "14", NULL },
         { "15", NULL },
         { "16", NULL },
         { "17", NULL },
         { "18", NULL },
         { "19", NULL },
         { "20", NULL },
         { "21", NULL },
         { "22", NULL },
         { "23", NULL },
         { "24", NULL },
         { "25", NULL },
         { "26", NULL },
         { "27", NULL },
         { "28", NULL },
         { "29", NULL },
         { "30", NULL },
         { "31", NULL },
         { "32", NULL },
         { "33", NULL },
         { "34", NULL },
         { "35", NULL },
         { "36", NULL },
         { "37", NULL },
         { "38", NULL },
         { "39", NULL },
         { "40", NULL },
         { "41", NULL },
         { "42", NULL },
         { "43", NULL },
         { "44", NULL },
         { "45", NULL },
         { "46", NULL },
         { "47", NULL },
         { "48", NULL },
         { "49", NULL },
         { "50", NULL },
         { "51", NULL },
         { "52", NULL },
         { "53", NULL },
         { "54", NULL },
         { "55", NULL },
         { "56", NULL },
         { "57", NULL },
         { "58", NULL },
         { "59", NULL },
         { "60", NULL },
         { "61", NULL },
         { "62", NULL },
         { "63", NULL },
         { NULL, NULL },
      },
      "0"
   },
   {
      BEETLE_OPT(memcard_right_index),
      BEETLE_OPT_MEMCARD_RIGHT_INDEX_LABEL_FA,
      NULL,
      BEETLE_OPT_MEMCARD_RIGHT_INDEX_INFO_0_FA,
      NULL,
      "memcards",
      {
         { "0",  NULL },
         { "1",  OPTION_VAL_1_O49_FA },
         { "2",  NULL },
         { "3",  NULL },
         { "4",  NULL },
         { "5",  NULL },
         { "6",  NULL },
         { "7",  NULL },
         { "8",  NULL },
         { "9",  NULL },
         { "10", NULL },
         { "11", NULL },
         { "12", NULL },
         { "13", NULL },
         { "14", NULL },
         { "15", NULL },
         { "16", NULL },
         { "17", NULL },
         { "18", NULL },
         { "19", NULL },
         { "20", NULL },
         { "21", NULL },
         { "22", NULL },
         { "23", NULL },
         { "24", NULL },
         { "25", NULL },
         { "26", NULL },
         { "27", NULL },
         { "28", NULL },
         { "29", NULL },
         { "30", NULL },
         { "31", NULL },
         { "32", NULL },
         { "33", NULL },
         { "34", NULL },
         { "35", NULL },
         { "36", NULL },
         { "37", NULL },
         { "38", NULL },
         { "39", NULL },
         { "40", NULL },
         { "41", NULL },
         { "42", NULL },
         { "43", NULL },
         { "44", NULL },
         { "45", NULL },
         { "46", NULL },
         { "47", NULL },
         { "48", NULL },
         { "49", NULL },
         { "50", NULL },
         { "51", NULL },
         { "52", NULL },
         { "53", NULL },
         { "54", NULL },
         { "55", NULL },
         { "56", NULL },
         { "57", NULL },
         { "58", NULL },
         { "59", NULL },
         { "60", NULL },
         { "61", NULL },
         { "62", NULL },
         { "63", NULL },
         { NULL, NULL },
      },
      "1"
   },
   {
      BEETLE_OPT(line_render),
      BEETLE_OPT_LINE_RENDER_LABEL_FA,
      NULL,
      BEETLE_OPT_LINE_RENDER_INFO_0_FA,
      NULL,
      "hacks",
      {
         { "default",    OPTION_VAL_DEFAULT_FA },
         { "aggressive", OPTION_VAL_AGGRESSIVE_FA },
         { "disabled",   NULL },
         { NULL, NULL },
      },
      "default"
   },
   {
      BEETLE_OPT(widescreen_hack),
      BEETLE_OPT_WIDESCREEN_HACK_LABEL_FA,
      NULL,
      BEETLE_OPT_WIDESCREEN_HACK_INFO_0_FA,
      NULL,
      "hacks",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(widescreen_hack_aspect_ratio),
      BEETLE_OPT_WIDESCREEN_HACK_ASPECT_RATIO_LABEL_FA,
      NULL,
      BEETLE_OPT_WIDESCREEN_HACK_ASPECT_RATIO_INFO_0_FA,
      NULL,
      "hacks",
      {
         { "16:9",  OPTION_VAL_16_9_FA },
         { "16:10", OPTION_VAL_16_10_FA },
         { "18:9",  OPTION_VAL_18_9_FA },
         { "19:9",  OPTION_VAL_19_9_FA },
         { "20:9",  OPTION_VAL_20_9_FA },
         { "21:9",  OPTION_VAL_21_9_FA }, // 64:27
         { "32:9",  OPTION_VAL_32_9_FA },
         { NULL,    NULL },
      },
      "16:9"
   },
#if defined(HAVE_LIGHTREC)
   {
      BEETLE_OPT(cpu_dynarec),
      BEETLE_OPT_CPU_DYNAREC_LABEL_FA,
      NULL,
      BEETLE_OPT_CPU_DYNAREC_INFO_0_FA,
      NULL,
      "hacks",
      {
         { "disabled",        OPTION_VAL_DISABLED_FA },
         { "execute",         OPTION_VAL_EXECUTE_FA },
         { "execute_one",     OPTION_VAL_EXECUTE_ONE_FA },
         { "run_interpreter", OPTION_VAL_RUN_INTERPRETER_FA },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(dynarec_invalidate),
      BEETLE_OPT_DYNAREC_INVALIDATE_LABEL_FA,
      NULL,
      BEETLE_OPT_DYNAREC_INVALIDATE_INFO_0_FA,
      NULL,
      "hacks",
      {
         { "full", OPTION_VAL_FULL_FA },
         { "dma",  OPTION_VAL_DMA_FA },
         { NULL, NULL },
      },
      "full"
   },
   {
      BEETLE_OPT(dynarec_eventcycles),
      BEETLE_OPT_DYNAREC_EVENTCYCLES_LABEL_FA,
      NULL,
      BEETLE_OPT_DYNAREC_EVENTCYCLES_INFO_0_FA,
      NULL,
      "hacks",
      {
         { "128",  OPTION_VAL_128_FA },
         { "256",  NULL },
         { "384",  NULL },
         { "512",  NULL },
         { "640",  NULL },
         { "768",  NULL },
         { "896",  NULL },
         { "1024", NULL },
         { "1152", NULL },
         { "1280", NULL },
         { "1408", NULL },
         { "1536", NULL },
         { "1664", NULL },
         { "1792", NULL },
         { "1920", NULL },
         { "2048", NULL },
         { NULL, NULL },
      },
      "128"
   },
   {
      BEETLE_OPT(dynarec_spu_samples),
      BEETLE_OPT_DYNAREC_SPU_SAMPLES_LABEL_FA,
      NULL,
      BEETLE_OPT_DYNAREC_SPU_SAMPLES_INFO_0_FA,
      NULL,
      "hacks",
      {
         { "1",  OPTION_VAL_1_O49_FA },
         { "4",  NULL },
         { "16", NULL },
         { NULL, NULL },
      },
      "1"
   },
#endif
   {
      BEETLE_OPT(core_timing_fps),
      BEETLE_OPT_CORE_TIMING_FPS_LABEL_FA,
      NULL,
      BEETLE_OPT_CORE_TIMING_FPS_INFO_0_FA,
      NULL,
      "video",
      {
         { "auto_toggle",       OPTION_VAL_AUTO_TOGGLE_FA },
         { "force_progressive", OPTION_VAL_FORCE_PROGRESSIVE_FA },
         { "force_interlaced",  OPTION_VAL_FORCE_INTERLACED_FA },
      },
      "auto_toggle"
   },
   {
      BEETLE_OPT(deinterlacer),
      BEETLE_OPT_DEINTERLACER_LABEL_FA,
      NULL,
      BEETLE_OPT_DEINTERLACER_INFO_0_FA,
      NULL,
      "video",
      {
         { "weave",      OPTION_VAL_WEAVE_FA },
         { "bob",        OPTION_VAL_BOB_FA },
         { "bob_offset", OPTION_VAL_BOB_OFFSET_FA },
         { "fastmad",    OPTION_VAL_FASTMAD_FA },
         { "off",        "Off" },
         { NULL, NULL },
      },
      "weave"
   },
   {
      BEETLE_OPT(frame_duping),
      BEETLE_OPT_FRAME_DUPING_LABEL_FA,
      NULL,
      BEETLE_OPT_FRAME_DUPING_INFO_0_FA,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(pal_video_timing_override),
      BEETLE_OPT_PAL_VIDEO_TIMING_OVERRIDE_LABEL_FA,
      NULL,
      BEETLE_OPT_PAL_VIDEO_TIMING_OVERRIDE_INFO_0_FA,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(aspect_ratio),
      BEETLE_OPT_ASPECT_RATIO_LABEL_FA,
      NULL,
      BEETLE_OPT_ASPECT_RATIO_INFO_0_FA,
      NULL,
      "video",
      {
         { "corrected", OPTION_VAL_CORRECTED_FA },
         { "uncorrected", OPTION_VAL_UNCORRECTED_FA },
         { "4:3",  OPTION_VAL_4_3_FA },
         { "ntsc", OPTION_VAL_NTSC_FA },
      },
      "corrected"
   },
   {
      BEETLE_OPT(crop_overscan),
      BEETLE_OPT_CROP_OVERSCAN_LABEL_FA,
      NULL,
      BEETLE_OPT_CROP_OVERSCAN_INFO_0_FA,
      NULL,
      "video",
      {
         { "disabled", "Disabled" },
         { "static",   OPTION_VAL_STATIC_FA },
         { "smart",    OPTION_VAL_SMART_FA },
         { NULL, NULL },
      },
      "smart"
   },
   {
      BEETLE_OPT(image_crop),
      BEETLE_OPT_IMAGE_CROP_LABEL_FA,
      NULL,
      BEETLE_OPT_IMAGE_CROP_INFO_0_FA,
      NULL,
      "video",
      {
         { "disabled", "0" },
         { "1px",      OPTION_VAL_1PX_FA },
         { "2px",      OPTION_VAL_2PX_FA },
         { "3px",      OPTION_VAL_3PX_FA },
         { "4px",      OPTION_VAL_4PX_FA },
         { "5px",      OPTION_VAL_5PX_FA },
         { "6px",      OPTION_VAL_6PX_FA },
         { "7px",      OPTION_VAL_7PX_FA },
         { "8px",      OPTION_VAL_8PX_FA },
         { "9px",      OPTION_VAL_9PX_FA },
         { "10px",     OPTION_VAL_10PX_FA },
         { "11px",     OPTION_VAL_11PX_FA },
         { "12px",     OPTION_VAL_12PX_FA },
         { "13px",     OPTION_VAL_13PX_FA },
         { "14px",     OPTION_VAL_14PX_FA },
         { "15px",     OPTION_VAL_15PX_FA },
         { "16px",     OPTION_VAL_16PX_FA },
         { "17px",     OPTION_VAL_17PX_FA },
         { "18px",     OPTION_VAL_18PX_FA },
         { "19px",     OPTION_VAL_19PX_FA },
         { "20px",     OPTION_VAL_20PX_FA },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(image_offset),
      BEETLE_OPT_IMAGE_OFFSET_LABEL_FA,
      NULL,
      BEETLE_OPT_IMAGE_OFFSET_INFO_0_FA,
      NULL,
      "video",
      {
         { "-12px",    OPTION_VAL_12PX_O64_FA },
         { "-11px",    OPTION_VAL_11PX_O64_FA },
         { "-10px",    OPTION_VAL_10PX_O64_FA },
         { "-9px",     OPTION_VAL_9PX_O64_FA },
         { "-8px",     OPTION_VAL_8PX_O64_FA },
         { "-7px",     OPTION_VAL_7PX_O64_FA },
         { "-6px",     OPTION_VAL_6PX_O64_FA },
         { "-5px",     OPTION_VAL_5PX_O64_FA },
         { "-4px",     OPTION_VAL_4PX_O64_FA },
         { "-3px",     OPTION_VAL_3PX_O64_FA },
         { "-2px",     OPTION_VAL_2PX_O64_FA },
         { "-1px",     OPTION_VAL_1PX_O64_FA },
         { "disabled", "0" },
         { "+1px",     OPTION_VAL_1PX_O64_0_FA },
         { "+2px",     OPTION_VAL_2PX_O64_0_FA },
         { "+3px",     OPTION_VAL_3PX_O64_0_FA },
         { "+4px",     OPTION_VAL_4PX_O64_0_FA },
         { "+5px",     OPTION_VAL_5PX_O64_0_FA },
         { "+6px",     OPTION_VAL_6PX_O64_0_FA },
         { "+7px",     OPTION_VAL_7PX_O64_0_FA },
         { "+8px",     OPTION_VAL_8PX_O64_0_FA },
         { "+9px",     OPTION_VAL_9PX_O64_0_FA },
         { "+10px",    OPTION_VAL_10PX_O64_0_FA },
         { "+11px",    OPTION_VAL_11PX_O64_0_FA },
         { "+12px",    OPTION_VAL_12PX_O64_0_FA },
         { NULL, NULL },
      },
      "disabled"
   },
#if defined(HAVE_OPENGL) || defined(HAVE_OPENGLES) || defined(HAVE_VULKAN)
   {
      BEETLE_OPT(image_offset_cycles),
      BEETLE_OPT_IMAGE_OFFSET_CYCLES_LABEL_FA,
      NULL,
      BEETLE_OPT_IMAGE_OFFSET_CYCLES_INFO_0_FA,
      NULL,
      "video",
      {
         { "-40",      NULL },
         { "-39",      NULL },
         { "-38",      NULL },
         { "-37",      NULL },
         { "-36",      NULL },
         { "-35",      NULL },
         { "-34",      NULL },
         { "-33",      NULL },
         { "-32",      NULL },
         { "-31",      NULL },
         { "-30",      NULL },
         { "-29",      NULL },
         { "-28",      NULL },
         { "-27",      NULL },
         { "-26",      NULL },
         { "-25",      NULL },
         { "-24",      NULL },
         { "-23",      NULL },
         { "-22",      NULL },
         { "-21",      NULL },
         { "-20",      NULL },
         { "-19",      NULL },
         { "-18",      NULL },
         { "-17",      NULL },
         { "-16",      NULL },
         { "-15",      NULL },
         { "-14",      NULL },
         { "-13",      NULL },
         { "-12",      NULL },
         { "-11",      NULL },
         { "-10",      NULL },
         { "-9",       NULL },
         { "-8",       NULL },
         { "-7",       NULL },
         { "-6",       NULL },
         { "-5",       NULL },
         { "-4",       NULL },
         { "-3",       NULL },
         { "-2",       NULL },
         { "-1",       NULL },
         { "0",        "0" },
         { "+1",       NULL },
         { "+2",       NULL },
         { "+3",       NULL },
         { "+4",       NULL },
         { "+5",       NULL },
         { "+6",       NULL },
         { "+7",       NULL },
         { "+8",       NULL },
         { "+9",       NULL },
         { "+10",      NULL },
         { "+11",      NULL },
         { "+12",      NULL },
         { "+13",      NULL },
         { "+14",      NULL },
         { "+15",      NULL },
         { "+16",      NULL },
         { "+17",      NULL },
         { "+18",      NULL },
         { "+19",      NULL },
         { "+20",      NULL },
         { "+21",      NULL },
         { "+22",      NULL },
         { "+23",      NULL },
         { "+24",      NULL },
         { "+25",      NULL },
         { "+26",      NULL },
         { "+27",      NULL },
         { "+28",      NULL },
         { "+29",      NULL },
         { "+30",      NULL },
         { "+31",      NULL },
         { "+32",      NULL },
         { "+33",      NULL },
         { "+34",      NULL },
         { "+35",      NULL },
         { "+36",      NULL },
         { "+37",      NULL },
         { "+38",      NULL },
         { "+39",      NULL },
         { "+40",      NULL },
         { NULL, NULL},
      },
      "0"
   },
#endif
   {
      BEETLE_OPT(initial_scanline),
      BEETLE_OPT_INITIAL_SCANLINE_LABEL_FA,
      NULL,
      BEETLE_OPT_INITIAL_SCANLINE_INFO_0_FA,
      NULL,
      "video",
      {
         { "0",  OPTION_VAL_0_O48_FA },
         { "1",  NULL },
         { "2",  NULL },
         { "3",  NULL },
         { "4",  NULL },
         { "5",  NULL },
         { "6",  NULL },
         { "7",  NULL },
         { "8",  NULL },
         { "9",  NULL },
         { "10", NULL },
         { "11", NULL },
         { "12", NULL },
         { "13", NULL },
         { "14", NULL },
         { "15", NULL },
         { "16", NULL },
         { "17", NULL },
         { "18", NULL },
         { "19", NULL },
         { "20", NULL },
         { "21", NULL },
         { "22", NULL },
         { "23", NULL },
         { "24", NULL },
         { "25", NULL },
         { "26", NULL },
         { "27", NULL },
         { "28", NULL },
         { "29", NULL },
         { "30", NULL },
         { "31", NULL },
         { "32", NULL },
         { "33", NULL },
         { "34", NULL },
         { "35", NULL },
         { "36", NULL },
         { "37", NULL },
         { "38", NULL },
         { "39", NULL },
         { "40", NULL },
         { NULL, NULL },
      },
      "0"
   },
   {
      BEETLE_OPT(last_scanline),
      BEETLE_OPT_LAST_SCANLINE_LABEL_FA,
      NULL,
      BEETLE_OPT_LAST_SCANLINE_INFO_0_FA,
      NULL,
      "video",
      {
         { "210", NULL },
         { "211", NULL },
         { "212", NULL },
         { "213", NULL },
         { "214", NULL },
         { "215", NULL },
         { "216", NULL },
         { "217", NULL },
         { "218", NULL },
         { "219", NULL },
         { "220", NULL },
         { "221", NULL },
         { "222", NULL },
         { "223", NULL },
         { "224", NULL },
         { "225", NULL },
         { "226", NULL },
         { "227", NULL },
         { "228", NULL },
         { "229", NULL },
         { "230", NULL },
         { "231", NULL },
         { "232", NULL },
         { "233", NULL },
         { "234", NULL },
         { "235", NULL },
         { "236", NULL },
         { "237", NULL },
         { "238", NULL },
         { "239", OPTION_VAL_239_FA },
         { NULL, NULL },
      },
      "239"
   },
   {
      BEETLE_OPT(initial_scanline_pal),
      BEETLE_OPT_INITIAL_SCANLINE_PAL_LABEL_FA,
      NULL,
      BEETLE_OPT_INITIAL_SCANLINE_PAL_INFO_0_FA,
      NULL,
      "video",
      {
         { "0",  OPTION_VAL_0_O48_FA },
         { "1",  NULL },
         { "2",  NULL },
         { "3",  NULL },
         { "4",  NULL },
         { "5",  NULL },
         { "6",  NULL },
         { "7",  NULL },
         { "8",  NULL },
         { "9",  NULL },
         { "10", NULL },
         { "11", NULL },
         { "12", NULL },
         { "13", NULL },
         { "14", NULL },
         { "15", NULL },
         { "16", NULL },
         { "17", NULL },
         { "18", NULL },
         { "19", NULL },
         { "20", NULL },
         { "21", NULL },
         { "22", NULL },
         { "23", NULL },
         { "24", NULL },
         { "25", NULL },
         { "26", NULL },
         { "27", NULL },
         { "28", NULL },
         { "29", NULL },
         { "30", NULL },
         { "31", NULL },
         { "32", NULL },
         { "33", NULL },
         { "34", NULL },
         { "35", NULL },
         { "36", NULL },
         { "37", NULL },
         { "38", NULL },
         { "39", NULL },
         { "40", NULL },
         { NULL, NULL },
      },
      "0"
   },
   {
      BEETLE_OPT(last_scanline_pal),
      BEETLE_OPT_LAST_SCANLINE_PAL_LABEL_FA,
      NULL,
      BEETLE_OPT_LAST_SCANLINE_PAL_INFO_0_FA,
      NULL,
      "video",
      {
         { "230", NULL },
         { "231", NULL },
         { "232", NULL },
         { "233", NULL },
         { "234", NULL },
         { "235", NULL },
         { "236", NULL },
         { "237", NULL },
         { "238", NULL },
         { "239", NULL },
         { "240", NULL },
         { "241", NULL },
         { "242", NULL },
         { "243", NULL },
         { "244", NULL },
         { "245", NULL },
         { "246", NULL },
         { "247", NULL },
         { "248", NULL },
         { "249", NULL },
         { "250", NULL },
         { "251", NULL },
         { "252", NULL },
         { "253", NULL },
         { "254", NULL },
         { "255", NULL },
         { "256", NULL },
         { "257", NULL },
         { "258", NULL },
         { "259", NULL },
         { "260", NULL },
         { "261", NULL },
         { "262", NULL },
         { "263", NULL },
         { "264", NULL },
         { "265", NULL },
         { "266", NULL },
         { "267", NULL },
         { "268", NULL },
         { "269", NULL },
         { "270", NULL },
         { "271", NULL },
         { "272", NULL },
         { "273", NULL },
         { "274", NULL },
         { "275", NULL },
         { "276", NULL },
         { "277", NULL },
         { "278", NULL },
         { "279", NULL },
         { "280", NULL },
         { "281", NULL },
         { "282", NULL },
         { "283", NULL },
         { "284", NULL },
         { "285", NULL },
         { "286", NULL },
         { "287", OPTION_VAL_287_FA },
         { NULL, NULL },
      },
      "287"
   },
   { NULL, NULL, NULL, NULL, NULL, NULL, {{0}}, NULL },
};
struct retro_core_options_v2 options_fa = {
   option_cats_fa,
   option_defs_fa
};

/* RETRO_LANGUAGE_FI */

#define CATEGORY_SYSTEM_LABEL_FI "Järjestelmä"
#define CATEGORY_SYSTEM_INFO_0_FI "Vaihda järjestelmän ja ylikelloitus asetuksia."
#define CATEGORY_VIDEO_LABEL_FI NULL
#define CATEGORY_VIDEO_INFO_0_FI "Muuta kuvasuhdetta, näytön rajausta, videosuodatusta ja kuvanohitusasetuksia."
#define CATEGORY_PGXP_LABEL_FI NULL
#define CATEGORY_PGXP_INFO_0_FI "Nämä asetukset ohjaavat parannuksia, jotka voivat parantaa grafiikkaa alkuperäiseen konsoliin verrattuna. PGXP saattaa poistaa vääristelevät tekstuurit ja Z-kiistelyyn liittyvät ongelmat."
#define CATEGORY_OSD_LABEL_FI "Näyttöruutu"
#define CATEGORY_OSD_INFO_0_FI "Muuta näytöllä näytettäviä ilmoituksia."
#define CATEGORY_INPUT_LABEL_FI "Syöte"
#define CATEGORY_INPUT_INFO_0_FI "Muokkaa valoaseen, hiiren ja neGcon:in asetuksia."
#define CATEGORY_MEMCARDS_LABEL_FI "Muistikortti"
#define CATEGORY_MEMCARDS_INFO_0_FI "Vaihda virtuaalisen muistikortin asetuksia."
#define CATEGORY_HACKS_LABEL_FI "Emulointikikat"
#define CATEGORY_HACKS_INFO_0_FI "Muuta prosessorin ylikellotus- ja emulointitarkkuusasetuksia, jotka vaikuttavat matalan tason suorituskykyyn ja yhteensopivuuteen."
#define BEETLE_OPT_CPU_FREQ_SCALE_LABEL_FI "Prosessorin taajuuden skaalaus (ylikelloitus)"
#define BEETLE_OPT_CPU_FREQ_SCALE_INFO_0_FI "Ylikelloita (tai alikelloita) emuloidun PS1 prosessoria. Ylikelloitus voi poistaa hitauden ja parantaa kuvataajuutta tietyissä peleissä, mutta suorituskykyvaatimuksien kustannuksella. Huomaa että joissakin peleissä on sisäinen kuvataajuus rajoitin, jolloin ne ei hyödy ylikelloittamisesta. Ylikelloitus voi nopeuttaa tiettyjä animaatioita ja tehosteita joissakin peleissä."
#define OPTION_VAL_50_FI "50 %"
#define OPTION_VAL_60_FI "60 %"
#define OPTION_VAL_70_FI "70 %"
#define OPTION_VAL_80_FI "80 %"
#define OPTION_VAL_90_FI "90 %"
#define OPTION_VAL_100_FI "100 % (Natiivi)"
#define OPTION_VAL_110_FI "110 %"
#define OPTION_VAL_120_FI "120 %"
#define OPTION_VAL_130_FI "130 %"
#define OPTION_VAL_140_FI "140 %"
#define OPTION_VAL_150_FI "150 %"
#define OPTION_VAL_160_FI "160 %"
#define OPTION_VAL_170_FI "170 %"
#define OPTION_VAL_180_FI "180 %"
#define OPTION_VAL_190_FI "190 %"
#define OPTION_VAL_200_FI "200 %"
#define OPTION_VAL_210_FI "210 %"
#define OPTION_VAL_220_FI "220 %"
#define OPTION_VAL_230_FI "230 %"
#define OPTION_VAL_240_FI "240 %"
#define OPTION_VAL_250_FI "250 %"
#define OPTION_VAL_260_FI "260 %"
#define OPTION_VAL_270_FI "270 %"
#define OPTION_VAL_280_FI "280 %"
#define OPTION_VAL_290_FI "290 %"
#define OPTION_VAL_300_FI "300 %"
#define OPTION_VAL_310_FI "310 %"
#define OPTION_VAL_320_FI "320 %"
#define OPTION_VAL_330_FI "330 %"
#define OPTION_VAL_340_FI "340 %"
#define OPTION_VAL_350_FI "350 %"
#define OPTION_VAL_360_FI "360 %"
#define OPTION_VAL_370_FI "370 %"
#define OPTION_VAL_380_FI "380 %"
#define OPTION_VAL_390_FI "390 %"
#define OPTION_VAL_400_FI "400 %"
#define OPTION_VAL_410_FI "410 %"
#define OPTION_VAL_420_FI "420 %"
#define OPTION_VAL_430_FI "430 %"
#define OPTION_VAL_440_FI "440 %"
#define OPTION_VAL_450_FI "450 %"
#define OPTION_VAL_460_FI "460 %"
#define OPTION_VAL_470_FI "470 %"
#define OPTION_VAL_480_FI "480 %"
#define OPTION_VAL_490_FI "490 %"
#define OPTION_VAL_500_FI "500 %"
#define OPTION_VAL_510_FI "510 %"
#define OPTION_VAL_520_FI "520 %"
#define OPTION_VAL_530_FI "530 %"
#define OPTION_VAL_540_FI "540 %"
#define OPTION_VAL_550_FI "550 %"
#define OPTION_VAL_560_FI "560 %"
#define OPTION_VAL_570_FI "570 %"
#define OPTION_VAL_580_FI "580 %"
#define OPTION_VAL_590_FI "590 %"
#define OPTION_VAL_600_FI "600 %"
#define OPTION_VAL_610_FI "610 %"
#define OPTION_VAL_620_FI "620 %"
#define OPTION_VAL_630_FI "630 %"
#define OPTION_VAL_640_FI "640 %"
#define OPTION_VAL_650_FI "650 %"
#define OPTION_VAL_660_FI "660 %"
#define OPTION_VAL_670_FI "670 %"
#define OPTION_VAL_680_FI "680 %"
#define OPTION_VAL_690_FI "690 %"
#define OPTION_VAL_700_FI "700 %"
#define OPTION_VAL_710_FI "710 %"
#define OPTION_VAL_720_FI "720 %"
#define OPTION_VAL_730_FI "730 %"
#define OPTION_VAL_740_FI "740 %"
#define OPTION_VAL_750_FI "750 %"
#define BEETLE_OPT_GPU_OVERCLOCK_LABEL_FI "Näytönohjaimen rasteroijan ylikelloitus"
#define BEETLE_OPT_GPU_OVERCLOCK_INFO_0_FI "Ota käyttöön emuloidun PS1:n näytönohjaimen 2D rasteroijan ylikelloitus. Ei paranna 3D-renderointintia, ja yleensä sillä on vähän vaikutusta."
#define OPTION_VAL_1X_NATIVE_FI "1x (Natiivi)"
#define OPTION_VAL_2X_FI NULL
#define OPTION_VAL_4X_FI NULL
#define OPTION_VAL_8X_FI NULL
#define OPTION_VAL_16X_FI NULL
#define OPTION_VAL_32X_FI NULL
#define BEETLE_OPT_GTE_OVERCLOCK_LABEL_FI "GTE ylikelloitus"
#define BEETLE_OPT_GTE_OVERCLOCK_INFO_0_FI "Pienentää kaikki emuloidut GTE (3D grafiikan apuprosessori) toiminnan jatkuvaksi yhden-syklin viiveeksi. Peleissä jotka käyttää GTE: tä paljon, tämä voi mahdollisesti parantaa kuvataajuutta ja vakautta."
#define BEETLE_OPT_CD_ACCESS_METHOD_LABEL_FI "CD:n käyttötapa"
#define BEETLE_OPT_CD_ACCESS_METHOD_INFO_0_FI "\"Tahdistettu\" jäljittelee alkuperäistä konsolia. \"Epä-tahdistettu\" voi vähentää jumiutumista hitailla laitteilla. \"Välimuistissa\" lataa koko levyn muistiin käynnistettäessä, mikä voi nopeuttaa pelin sisäisiä latausaikoja aloitusvaiheen viivästymisen kustannuksella. \"Välimuistissa\" voi aiheuttaa ongelmia laitteilla joilla ei ole paljoa RAM-muistia, ja palaa takaisin \"Tahdistettu\" asetukseen. Uudelleenkäynnistys vaaditaan."
#define OPTION_VAL_SYNC_FI "Synkroninen"
#define OPTION_VAL_ASYNC_FI "Asynkroninen"
#define OPTION_VAL_PRECACHE_FI "Esivälimuisti"
#define BEETLE_OPT_CD_FASTLOAD_LABEL_FI "CD:n latausnopeus"
#define BEETLE_OPT_CD_FASTLOAD_INFO_0_FI "Arvot, jotka ovat korkeammat kuin 2x (natiivi) voivat vähentää latausaikoja, mutta saattavat aiheuttaa ajoitusvirheitä. Jotkin pelit ei toimi kunnolla tiettyjen arvojen yläpuolella."
#define OPTION_VAL_2X_NATIVE_FI "2x (Natiivi)"
#define OPTION_VAL_6X_FI NULL
#define OPTION_VAL_10X_FI NULL
#define OPTION_VAL_12X_FI NULL
#define OPTION_VAL_14X_FI NULL
#define BEETLE_OPT_SKIP_BIOS_LABEL_FI "Ohita BIOS"
#define BEETLE_OPT_SKIP_BIOS_INFO_0_FI "Ohittaa Playstationin BIOS:sin käynnistysanimaation. Huom, tämä asetus voi aiheuttaa yhteensopivusongelmia tietyissä peleissä (PAL-kopiosuojatut pelit, Saga Frontier, jne)."
#define BEETLE_OPT_OVERRIDE_BIOS_LABEL_FI "Ohita BIOS"
#define BEETLE_OPT_OVERRIDE_BIOS_INFO_0_FI "Korvaa vakioaluekohtaisen BIOS:sin aluevapaalla, mikäli sellainen löytyy. Uudelleenkäynnistys vaaditaan."
#define OPTION_VAL_PSXONPSP_FI "PSP:n PS1 BIOS"
#define OPTION_VAL_PS1_ROM_FI "PS3:n PS1 BIOS"
#define OPTION_VAL_OPENBIOS_FI NULL
#define BEETLE_OPT_REGION_LABEL_FI "Järjestelmän alue"
#define BEETLE_OPT_REGION_INFO_0_FI NULL
#define OPTION_VAL_AUTO_FI "Automaattinen"
#define OPTION_VAL_NTSC_J_FI "NTSC-J (Japani)"
#define OPTION_VAL_NTSC_U_FI "NTSC-U (Pohjois-Amerikka)"
#define OPTION_VAL_PAL_FI "PAL (Eurooppa)"
#define BEETLE_OPT_INTERNAL_RESOLUTION_LABEL_FI "Sisäisen näytönohjaimen resoluutio"
#define BEETLE_OPT_INTERNAL_RESOLUTION_INFO_0_FI "Korkeammat resoluutiot kuin 1x (Natiivi) parantaa 3D-mallien tarkkuutta lisääntyvät suorituskukuvaatimuksien kustannuksella. Tämä asetus ei yleensä vaikuta 2D-elementteihin."
#define BEETLE_OPT_RENDERER_LABEL_FI "Renderoija"
#define BEETLE_OPT_RENDERER_INFO_0_FI "Ohjelmistorenderoija on tarkka, mutta sillä on jyrkät järjestelmävaatimukset kun käytetään suurempaa sisäistä näytönohjaimen resoluutioita. Laitteistorenderoija on vähemmän tarkempi, mutta parantaa suorituskykykä ja sen kanssa voidaan käyttää erinlaisia grafiikan parannuksia. \"Laitteisto (automaattinen)\" automaattisesti valitsee Vulkan tai OpenGL renderoijan, riippuen nykyisestä libretro käyttöliittymän ajurista. Jos annettu video-ajuri ei ole Vulkan tai OpenGL 3.3-yhteensopiva, sitten ydin palaa takaisin ohjelmistorenderoijaan. Uudelleenkäynnistys vaaditaan."
#define OPTION_VAL_HARDWARE_FI "Laitteisto (Auto)"
#define OPTION_VAL_HARDWARE_GL_FI "Laitteisto (OpenGL)"
#define OPTION_VAL_HARDWARE_VK_FI "Laitteisto (Vulkan)"
#define OPTION_VAL_SOFTWARE_FI "Ohjelmisto"
#define BEETLE_OPT_RENDERER_SOFTWARE_FB_LABEL_FI "Ohjelmisto kehyspuskuri"
#define BEETLE_OPT_RENDERER_SOFTWARE_FB_INFO_0_FI "Ota käyttöön kehyspuskuri-tehosteiden (esim. liikkeen hämärtyminen, FF7-taistelun pyörre) tarkka emulointi laitteistopohjaista renderöijöitää käytettäessä ajamalla taustalla ohjelmistopohjaisen renderöijän kopio alkuperäisellä tarkkuudella. Jos tämä asetus on pois käytöstä, nämä toiminnot ohitetaan (OpenGL) tai renderöidään näytönohjaimella (Vulkan). Asetuksen poistaminen käytöstä voi parantaa suorituskykyä, mutta se voi aiheuttaa vakavia graafisia virheitä. Jätä asetus käyttöön, jos et ole varma."
#define BEETLE_OPT_DEPTH_LABEL_FI "Sisäinen värisyvyys"
#define BEETLE_OPT_DEPTH_INFO_0_FI NULL
#define OPTION_VAL_16BPP_NATIVE_FI "16 bpp (Natiivi)"
#define OPTION_VAL_32BPP_FI NULL
#define BEETLE_OPT_DITHER_MODE_LABEL_FI NULL
#define BEETLE_OPT_DITHER_MODE_INFO_0_FI NULL
#define OPTION_VAL_INTERNAL_RESOLUTION_FI "Sisäinen resoluutio"
#define BEETLE_OPT_SCALED_UV_OFFSET_LABEL_FI NULL
#define BEETLE_OPT_SCALED_UV_OFFSET_INFO_0_FI NULL
#define BEETLE_OPT_FILTER_LABEL_FI "Tekstuurisuodatus"
#define BEETLE_OPT_FILTER_INFO_0_FI NULL
#define OPTION_VAL_NEAREST_FI "Lähin"
#define OPTION_VAL_SABR_FI NULL
#define OPTION_VAL_XBR_FI NULL
#define OPTION_VAL_BILINEAR_FI "Bilineaarinen"
#define OPTION_VAL_3_POINT_FI "Kolmipiste"
#define OPTION_VAL_JINC2_FI NULL
#define BEETLE_OPT_FILTER_EXCLUDE_SPRITE_LABEL_FI NULL
#define BEETLE_OPT_FILTER_EXCLUDE_SPRITE_INFO_0_FI NULL
#define OPTION_VAL_OPAQUE_FI NULL
#define OPTION_VAL_ALL_FI NULL
#define BEETLE_OPT_FILTER_EXCLUDE_2D_POLYGON_LABEL_FI NULL
#define BEETLE_OPT_FILTER_EXCLUDE_2D_POLYGON_INFO_0_FI NULL
#define BEETLE_OPT_ADAPTIVE_SMOOTHING_LABEL_FI "Mukautuva pehmennys"
#define BEETLE_OPT_ADAPTIVE_SMOOTHING_INFO_0_FI NULL
#define BEETLE_OPT_SUPER_SAMPLING_LABEL_FI NULL
#define BEETLE_OPT_SUPER_SAMPLING_INFO_0_FI NULL
#define BEETLE_OPT_MSAA_LABEL_FI NULL
#define BEETLE_OPT_MSAA_INFO_0_FI NULL
#define OPTION_VAL_1X_FI NULL
#define BEETLE_OPT_MDEC_YUV_LABEL_FI NULL
#define BEETLE_OPT_MDEC_YUV_INFO_0_FI NULL
#define BEETLE_OPT_TRACK_TEXTURES_LABEL_FI NULL
#define BEETLE_OPT_TRACK_TEXTURES_INFO_0_FI NULL
#define BEETLE_OPT_DUMP_TEXTURES_LABEL_FI NULL
#define BEETLE_OPT_DUMP_TEXTURES_INFO_0_FI NULL
#define BEETLE_OPT_REPLACE_TEXTURES_LABEL_FI "Korvaa tekstuureja"
#define BEETLE_OPT_REPLACE_TEXTURES_INFO_0_FI NULL
#define BEETLE_OPT_PGXP_MODE_LABEL_FI NULL
#define BEETLE_OPT_PGXP_MODE_INFO_0_FI NULL
#define OPTION_VAL_MEMORY_ONLY_FI NULL
#define OPTION_VAL_MEMORY_CPU_FI NULL
#define BEETLE_OPT_PGXP_2D_TOL_LABEL_FI NULL
#define BEETLE_OPT_PGXP_2D_TOL_INFO_0_FI "Piilota räikeämmät virheet PGXP-operaatioissa: arvo määrittää toleranssin, jossa PGXP-arvot säilytetään, jos geometriat on ilman asianmukaista syvyystietoa."
#define OPTION_VAL_0PX_FI "0 px"
#define OPTION_VAL_1PX_FI "1 px"
#define OPTION_VAL_2PX_FI "2 px"
#define OPTION_VAL_3PX_FI "3 px"
#define OPTION_VAL_4PX_FI "4 px"
#define OPTION_VAL_5PX_FI "5 px"
#define OPTION_VAL_6PX_FI "6 px"
#define OPTION_VAL_7PX_FI "7 px"
#define OPTION_VAL_8PX_FI "8 px"
#define BEETLE_OPT_PGXP_NCLIP_LABEL_FI NULL
#define BEETLE_OPT_PGXP_NCLIP_INFO_0_FI NULL
#define BEETLE_OPT_PGXP_VERTEX_LABEL_FI NULL
#define BEETLE_OPT_PGXP_VERTEX_INFO_0_FI NULL
#define BEETLE_OPT_PGXP_TEXTURE_LABEL_FI NULL
#define BEETLE_OPT_PGXP_TEXTURE_INFO_0_FI NULL
#define BEETLE_OPT_DISPLAY_INTERNAL_FPS_LABEL_FI "Näytä sisäinen FPS"
#define BEETLE_OPT_DISPLAY_INTERNAL_FPS_INFO_0_FI NULL
#define BEETLE_OPT_DISPLAY_OSD_LABEL_FI NULL
#define BEETLE_OPT_DISPLAY_OSD_INFO_0_FI NULL
#define BEETLE_OPT_DISPLAY_VRAM_LABEL_FI "Näytä täysi VRAM (vianetsintä)"
#define BEETLE_OPT_DISPLAY_VRAM_INFO_0_FI NULL
#define BEETLE_OPT_ANALOG_CALIBRATION_LABEL_FI NULL
#define BEETLE_OPT_ANALOG_CALIBRATION_INFO_0_FI NULL
#define BEETLE_OPT_ANALOG_TOGGLE_LABEL_FI NULL
#define BEETLE_OPT_ANALOG_TOGGLE_INFO_0_FI NULL
#define OPTION_VAL_ENABLED_ANALOG_FI NULL
#define BEETLE_OPT_ANALOG_TOGGLE_COMBO_LABEL_FI NULL
#define BEETLE_OPT_ANALOG_TOGGLE_COMBO_INFO_0_FI NULL
#define OPTION_VAL_L1_L2_R1_R2_START_SELECT_FI NULL
#define OPTION_VAL_L1_R1_SELECT_FI NULL
#define OPTION_VAL_L1_R1_START_FI NULL
#define OPTION_VAL_L1_R1_L3_FI NULL
#define OPTION_VAL_L1_R1_R3_FI NULL
#define OPTION_VAL_L2_R2_SELECT_FI NULL
#define OPTION_VAL_L2_R2_START_FI NULL
#define OPTION_VAL_L2_R2_L3_FI NULL
#define OPTION_VAL_L2_R2_R3_FI NULL
#define OPTION_VAL_L3_R3_FI NULL
#define BEETLE_OPT_ANALOG_TOGGLE_HOLD_LABEL_FI "DualShockin analogisen tilan combon pitämisen viive"
#define BEETLE_OPT_ANALOG_TOGGLE_HOLD_INFO_0_FI NULL
#define OPTION_VAL_0_FI "0 sekunnin viive"
#define OPTION_VAL_1_FI "1 sekunnin viive"
#define OPTION_VAL_2_FI "2 sekunnin viive"
#define OPTION_VAL_3_FI "3 sekunnin viive"
#define OPTION_VAL_4_FI "4 sekunnin viive"
#define OPTION_VAL_5_FI "5 sekunnin viive"
#define BEETLE_OPT_ENABLE_MULTITAP_PORT1_LABEL_FI NULL
#define BEETLE_OPT_ENABLE_MULTITAP_PORT1_INFO_0_FI NULL
#define BEETLE_OPT_ENABLE_MULTITAP_PORT2_LABEL_FI NULL
#define BEETLE_OPT_ENABLE_MULTITAP_PORT2_INFO_0_FI NULL
#define BEETLE_OPT_GUN_INPUT_MODE_LABEL_FI NULL
#define BEETLE_OPT_GUN_INPUT_MODE_INFO_0_FI NULL
#define OPTION_VAL_LIGHTGUN_FI "Valoase"
#define OPTION_VAL_TOUCHSCREEN_FI "Kosketusnäyttö"
#define BEETLE_OPT_GUN_CURSOR_LABEL_FI "Aseen tähtäin"
#define BEETLE_OPT_GUN_CURSOR_INFO_0_FI NULL
#define OPTION_VAL_CROSS_FI "Risti"
#define OPTION_VAL_DOT_FI "Piste"
#define BEETLE_OPT_CROSSHAIR_COLOR_P1_LABEL_FI NULL
#define BEETLE_OPT_CROSSHAIR_COLOR_P1_INFO_0_FI NULL
#define OPTION_VAL_RED_FI "Punainen"
#define OPTION_VAL_BLUE_FI "Sininen"
#define OPTION_VAL_GREEN_FI "Vihreä"
#define OPTION_VAL_ORANGE_FI "Oranssi"
#define OPTION_VAL_YELLOW_FI "Keltainen"
#define OPTION_VAL_CYAN_FI "Syaani"
#define OPTION_VAL_PINK_FI "Vaaleanpunainen"
#define OPTION_VAL_PURPLE_FI "Violetti"
#define OPTION_VAL_BLACK_FI "Musta"
#define OPTION_VAL_WHITE_FI "Valkoinen"
#define BEETLE_OPT_CROSSHAIR_COLOR_P2_LABEL_FI NULL
#define BEETLE_OPT_CROSSHAIR_COLOR_P2_INFO_0_FI NULL
#define BEETLE_OPT_MOUSE_SENSITIVITY_LABEL_FI "Hiiren herkkyys"
#define BEETLE_OPT_MOUSE_SENSITIVITY_INFO_0_FI NULL
#define OPTION_VAL_5_O42_FI "5 %"
#define OPTION_VAL_10_FI "10 %"
#define OPTION_VAL_15_FI "15 %"
#define OPTION_VAL_20_FI "20 %"
#define OPTION_VAL_25_FI "25 %"
#define OPTION_VAL_30_FI "30 %"
#define OPTION_VAL_35_FI "35 %"
#define OPTION_VAL_40_FI "40 %"
#define OPTION_VAL_45_FI "45 %"
#define OPTION_VAL_55_FI "55 %"
#define OPTION_VAL_65_FI "65 %"
#define OPTION_VAL_75_FI "75 %"
#define OPTION_VAL_85_FI "85 %"
#define OPTION_VAL_95_FI "95 %"
#define OPTION_VAL_100_O42_FI "100 %"
#define OPTION_VAL_105_FI "105 %"
#define OPTION_VAL_115_FI "115 %"
#define OPTION_VAL_125_FI "125 %"
#define OPTION_VAL_135_FI "135 %"
#define OPTION_VAL_145_FI "145 %"
#define OPTION_VAL_155_FI "155 %"
#define OPTION_VAL_165_FI "165 %"
#define OPTION_VAL_175_FI "175 %"
#define OPTION_VAL_185_FI "185 %"
#define OPTION_VAL_195_FI "195 %"
#define BEETLE_OPT_NEGCON_RESPONSE_LABEL_FI NULL
#define BEETLE_OPT_NEGCON_RESPONSE_INFO_0_FI NULL
#define OPTION_VAL_LINEAR_FI "Lineaarinen"
#define OPTION_VAL_QUADRATIC_FI NULL
#define OPTION_VAL_CUBIC_FI NULL
#define BEETLE_OPT_NEGCON_DEADZONE_LABEL_FI NULL
#define BEETLE_OPT_NEGCON_DEADZONE_INFO_0_FI NULL
#define OPTION_VAL_0_O44_FI "0 %"
#define BEETLE_OPT_USE_MEDNAFEN_MEMCARD0_METHOD_LABEL_FI "Muistikortin menetelmä"
#define BEETLE_OPT_USE_MEDNAFEN_MEMCARD0_METHOD_INFO_0_FI NULL
#define OPTION_VAL_LIBRETRO_FI NULL
#define OPTION_VAL_MEDNAFEN_FI NULL
#define BEETLE_OPT_ENABLE_MEMCARD1_LABEL_FI "Muistikortin paikka 2"
#define BEETLE_OPT_ENABLE_MEMCARD1_INFO_0_FI NULL
#define BEETLE_OPT_SHARED_MEMORY_CARDS_LABEL_FI "Jaetut muistikortit"
#define BEETLE_OPT_SHARED_MEMORY_CARDS_INFO_0_FI NULL
#define BEETLE_OPT_MEMCARD_LEFT_INDEX_LABEL_FI NULL
#define BEETLE_OPT_MEMCARD_LEFT_INDEX_INFO_0_FI NULL
#define OPTION_VAL_0_O48_FI "0 (Oletus)"
#define BEETLE_OPT_MEMCARD_RIGHT_INDEX_LABEL_FI NULL
#define BEETLE_OPT_MEMCARD_RIGHT_INDEX_INFO_0_FI NULL
#define OPTION_VAL_1_O49_FI "1 (Oletus)"
#define BEETLE_OPT_LINE_RENDER_LABEL_FI NULL
#define BEETLE_OPT_LINE_RENDER_INFO_0_FI NULL
#define OPTION_VAL_DEFAULT_FI "Oletus"
#define OPTION_VAL_AGGRESSIVE_FI NULL
#define BEETLE_OPT_WIDESCREEN_HACK_LABEL_FI NULL
#define BEETLE_OPT_WIDESCREEN_HACK_INFO_0_FI NULL
#define BEETLE_OPT_WIDESCREEN_HACK_ASPECT_RATIO_LABEL_FI NULL
#define BEETLE_OPT_WIDESCREEN_HACK_ASPECT_RATIO_INFO_0_FI NULL
#define OPTION_VAL_16_9_FI NULL
#define OPTION_VAL_16_10_FI NULL
#define OPTION_VAL_18_9_FI NULL
#define OPTION_VAL_19_9_FI NULL
#define OPTION_VAL_20_9_FI NULL
#define OPTION_VAL_21_9_FI NULL
#define OPTION_VAL_32_9_FI NULL
#define BEETLE_OPT_CPU_DYNAREC_LABEL_FI NULL
#define BEETLE_OPT_CPU_DYNAREC_INFO_0_FI NULL
#define OPTION_VAL_DISABLED_FI NULL
#define OPTION_VAL_EXECUTE_FI NULL
#define OPTION_VAL_EXECUTE_ONE_FI NULL
#define OPTION_VAL_RUN_INTERPRETER_FI NULL
#define BEETLE_OPT_DYNAREC_INVALIDATE_LABEL_FI NULL
#define BEETLE_OPT_DYNAREC_INVALIDATE_INFO_0_FI "Jotkin pelit vaativat 'Täysi'-mitätöintiä, toiset vaativat 'Vain DMA'."
#define OPTION_VAL_FULL_FI "Täysi"
#define OPTION_VAL_DMA_FI "Vain DMA (hieman nopeampi)"
#define BEETLE_OPT_DYNAREC_EVENTCYCLES_LABEL_FI NULL
#define BEETLE_OPT_DYNAREC_EVENTCYCLES_INFO_0_FI NULL
#define OPTION_VAL_128_FI "128 (Oletus)"
#define BEETLE_OPT_DYNAREC_SPU_SAMPLES_LABEL_FI NULL
#define BEETLE_OPT_DYNAREC_SPU_SAMPLES_INFO_0_FI NULL
#define BEETLE_OPT_CORE_TIMING_FPS_LABEL_FI NULL
#define BEETLE_OPT_CORE_TIMING_FPS_INFO_0_FI NULL
#define OPTION_VAL_AUTO_TOGGLE_FI NULL
#define OPTION_VAL_FORCE_PROGRESSIVE_FI NULL
#define OPTION_VAL_FORCE_INTERLACED_FI NULL
#define BEETLE_OPT_DEINTERLACER_LABEL_FI NULL
#define BEETLE_OPT_DEINTERLACER_INFO_0_FI NULL
#define OPTION_VAL_WEAVE_FI NULL
#define OPTION_VAL_BOB_FI NULL
#define OPTION_VAL_BOB_OFFSET_FI NULL
#define OPTION_VAL_FASTMAD_FI NULL
#define BEETLE_OPT_FRAME_DUPING_LABEL_FI NULL
#define BEETLE_OPT_FRAME_DUPING_INFO_0_FI NULL
#define BEETLE_OPT_PAL_VIDEO_TIMING_OVERRIDE_LABEL_FI NULL
#define BEETLE_OPT_PAL_VIDEO_TIMING_OVERRIDE_INFO_0_FI NULL
#define BEETLE_OPT_ASPECT_RATIO_LABEL_FI "Ytimen kuvasuhde"
#define BEETLE_OPT_ASPECT_RATIO_INFO_0_FI NULL
#define OPTION_VAL_CORRECTED_FI "Korjattu"
#define OPTION_VAL_UNCORRECTED_FI "Ei korjattu"
#define OPTION_VAL_4_3_FI "Pakota 4:3"
#define OPTION_VAL_NTSC_FI "Pakota NTSC"
#define BEETLE_OPT_CROP_OVERSCAN_LABEL_FI NULL
#define BEETLE_OPT_CROP_OVERSCAN_INFO_0_FI NULL
#define OPTION_VAL_STATIC_FI NULL
#define OPTION_VAL_SMART_FI NULL
#define BEETLE_OPT_IMAGE_CROP_LABEL_FI NULL
#define BEETLE_OPT_IMAGE_CROP_INFO_0_FI NULL
#define OPTION_VAL_9PX_FI "9 px"
#define OPTION_VAL_10PX_FI "10 px"
#define OPTION_VAL_11PX_FI "11 px"
#define OPTION_VAL_12PX_FI "12 px"
#define OPTION_VAL_13PX_FI "13 px"
#define OPTION_VAL_14PX_FI "14 px"
#define OPTION_VAL_15PX_FI "15 px"
#define OPTION_VAL_16PX_FI "16 px"
#define OPTION_VAL_17PX_FI "17 px"
#define OPTION_VAL_18PX_FI "18 px"
#define OPTION_VAL_19PX_FI "19 px"
#define OPTION_VAL_20PX_FI "20 px"
#define BEETLE_OPT_IMAGE_OFFSET_LABEL_FI NULL
#define BEETLE_OPT_IMAGE_OFFSET_INFO_0_FI NULL
#define OPTION_VAL_12PX_O64_FI NULL
#define OPTION_VAL_11PX_O64_FI NULL
#define OPTION_VAL_10PX_O64_FI NULL
#define OPTION_VAL_9PX_O64_FI NULL
#define OPTION_VAL_8PX_O64_FI NULL
#define OPTION_VAL_7PX_O64_FI NULL
#define OPTION_VAL_6PX_O64_FI NULL
#define OPTION_VAL_5PX_O64_FI NULL
#define OPTION_VAL_4PX_O64_FI NULL
#define OPTION_VAL_3PX_O64_FI NULL
#define OPTION_VAL_2PX_O64_FI NULL
#define OPTION_VAL_1PX_O64_FI NULL
#define OPTION_VAL_1PX_O64_0_FI NULL
#define OPTION_VAL_2PX_O64_0_FI NULL
#define OPTION_VAL_3PX_O64_0_FI NULL
#define OPTION_VAL_4PX_O64_0_FI NULL
#define OPTION_VAL_5PX_O64_0_FI NULL
#define OPTION_VAL_6PX_O64_0_FI NULL
#define OPTION_VAL_7PX_O64_0_FI NULL
#define OPTION_VAL_8PX_O64_0_FI NULL
#define OPTION_VAL_9PX_O64_0_FI NULL
#define OPTION_VAL_10PX_O64_0_FI NULL
#define OPTION_VAL_11PX_O64_0_FI NULL
#define OPTION_VAL_12PX_O64_0_FI NULL
#define BEETLE_OPT_IMAGE_OFFSET_CYCLES_LABEL_FI NULL
#define BEETLE_OPT_IMAGE_OFFSET_CYCLES_INFO_0_FI NULL
#define BEETLE_OPT_INITIAL_SCANLINE_LABEL_FI NULL
#define BEETLE_OPT_INITIAL_SCANLINE_INFO_0_FI NULL
#define BEETLE_OPT_LAST_SCANLINE_LABEL_FI NULL
#define BEETLE_OPT_LAST_SCANLINE_INFO_0_FI NULL
#define OPTION_VAL_239_FI "239 (Oletus)"
#define BEETLE_OPT_INITIAL_SCANLINE_PAL_LABEL_FI NULL
#define BEETLE_OPT_INITIAL_SCANLINE_PAL_INFO_0_FI NULL
#define BEETLE_OPT_LAST_SCANLINE_PAL_LABEL_FI NULL
#define BEETLE_OPT_LAST_SCANLINE_PAL_INFO_0_FI NULL
#define OPTION_VAL_287_FI "287 (Oletus)"

struct retro_core_option_v2_category option_cats_fi[] = {
   {
      "system",
      CATEGORY_SYSTEM_LABEL_FI,
      CATEGORY_SYSTEM_INFO_0_FI
   },
   {
      "video",
      CATEGORY_VIDEO_LABEL_FI,
      CATEGORY_VIDEO_INFO_0_FI
   },
   {
      "pgxp",
      CATEGORY_PGXP_LABEL_FI,
      CATEGORY_PGXP_INFO_0_FI
   },
   {
      "osd",
      CATEGORY_OSD_LABEL_FI,
      CATEGORY_OSD_INFO_0_FI
   },
   {
      "input",
      CATEGORY_INPUT_LABEL_FI,
      CATEGORY_INPUT_INFO_0_FI
   },
   {
      "memcards",
      CATEGORY_MEMCARDS_LABEL_FI,
      CATEGORY_MEMCARDS_INFO_0_FI
   },
   {
      "hacks",
      CATEGORY_HACKS_LABEL_FI,
      CATEGORY_HACKS_INFO_0_FI
   },
   { NULL, NULL, NULL },
};
struct retro_core_option_v2_definition option_defs_fi[] = {
   {
      BEETLE_OPT(cpu_freq_scale),
      BEETLE_OPT_CPU_FREQ_SCALE_LABEL_FI,
      NULL,
      BEETLE_OPT_CPU_FREQ_SCALE_INFO_0_FI,
      NULL,
      "system",
      {
         { "50%",           OPTION_VAL_50_FI },
         { "60%",           OPTION_VAL_60_FI },
         { "70%",           OPTION_VAL_70_FI },
         { "80%",           OPTION_VAL_80_FI },
         { "90%",           OPTION_VAL_90_FI },
         { "100%",          OPTION_VAL_100_FI },
         { "110%",          OPTION_VAL_110_FI },
         { "120%",          OPTION_VAL_120_FI },
         { "130%",          OPTION_VAL_130_FI },
         { "140%",          OPTION_VAL_140_FI },
         { "150%",          OPTION_VAL_150_FI },
         { "160%",          OPTION_VAL_160_FI },
         { "170%",          OPTION_VAL_170_FI },
         { "180%",          OPTION_VAL_180_FI },
         { "190%",          OPTION_VAL_190_FI },
         { "200%",          OPTION_VAL_200_FI },
         { "210%",          OPTION_VAL_210_FI },
         { "220%",          OPTION_VAL_220_FI },
         { "230%",          OPTION_VAL_230_FI },
         { "240%",          OPTION_VAL_240_FI },
         { "250%",          OPTION_VAL_250_FI },
         { "260%",          OPTION_VAL_260_FI },
         { "270%",          OPTION_VAL_270_FI },
         { "280%",          OPTION_VAL_280_FI },
         { "290%",          OPTION_VAL_290_FI },
         { "300%",          OPTION_VAL_300_FI },
         { "310%",          OPTION_VAL_310_FI },
         { "320%",          OPTION_VAL_320_FI },
         { "330%",          OPTION_VAL_330_FI },
         { "340%",          OPTION_VAL_340_FI },
         { "350%",          OPTION_VAL_350_FI },
         { "360%",          OPTION_VAL_360_FI },
         { "370%",          OPTION_VAL_370_FI },
         { "380%",          OPTION_VAL_380_FI },
         { "390%",          OPTION_VAL_390_FI },
         { "400%",          OPTION_VAL_400_FI },
         { "410%",          OPTION_VAL_410_FI },
         { "420%",          OPTION_VAL_420_FI },
         { "430%",          OPTION_VAL_430_FI },
         { "440%",          OPTION_VAL_440_FI },
         { "450%",          OPTION_VAL_450_FI },
         { "460%",          OPTION_VAL_460_FI },
         { "470%",          OPTION_VAL_470_FI },
         { "480%",          OPTION_VAL_480_FI },
         { "490%",          OPTION_VAL_490_FI },
         { "500%",          OPTION_VAL_500_FI },
         { "510%",          OPTION_VAL_510_FI },
         { "520%",          OPTION_VAL_520_FI },
         { "530%",          OPTION_VAL_530_FI },
         { "540%",          OPTION_VAL_540_FI },
         { "550%",          OPTION_VAL_550_FI },
         { "560%",          OPTION_VAL_560_FI },
         { "570%",          OPTION_VAL_570_FI },
         { "580%",          OPTION_VAL_580_FI },
         { "590%",          OPTION_VAL_590_FI },
         { "600%",          OPTION_VAL_600_FI },
         { "610%",          OPTION_VAL_610_FI },
         { "620%",          OPTION_VAL_620_FI },
         { "630%",          OPTION_VAL_630_FI },
         { "640%",          OPTION_VAL_640_FI },
         { "650%",          OPTION_VAL_650_FI },
         { "660%",          OPTION_VAL_660_FI },
         { "670%",          OPTION_VAL_670_FI },
         { "680%",          OPTION_VAL_680_FI },
         { "690%",          OPTION_VAL_690_FI },
         { "700%",          OPTION_VAL_700_FI },
         { "710%",          OPTION_VAL_710_FI },
         { "720%",          OPTION_VAL_720_FI },
         { "730%",          OPTION_VAL_730_FI },
         { "740%",          OPTION_VAL_740_FI },
         { "750%",          OPTION_VAL_750_FI },
         { NULL, NULL },
      },
      "100%"
   },
   {
      BEETLE_OPT(gpu_overclock),
      BEETLE_OPT_GPU_OVERCLOCK_LABEL_FI,
      NULL,
      BEETLE_OPT_GPU_OVERCLOCK_INFO_0_FI,
      NULL,
      "system",
      {
         { "1x(native)", OPTION_VAL_1X_NATIVE_FI },
         { "2x",         OPTION_VAL_2X_FI },
         { "4x",         OPTION_VAL_4X_FI },
         { "8x",         OPTION_VAL_8X_FI },
         { "16x",        OPTION_VAL_16X_FI },
         { "32x",        OPTION_VAL_32X_FI },
         { NULL, NULL },
      },
      "1x(native)"
   },
   {
      BEETLE_OPT(gte_overclock),
      BEETLE_OPT_GTE_OVERCLOCK_LABEL_FI,
      NULL,
      BEETLE_OPT_GTE_OVERCLOCK_INFO_0_FI,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
#ifndef EMSCRIPTEN
   {
      BEETLE_OPT(cd_access_method),
      BEETLE_OPT_CD_ACCESS_METHOD_LABEL_FI,
      NULL,
      BEETLE_OPT_CD_ACCESS_METHOD_INFO_0_FI,
      NULL,
      "system",
      {
         { "sync",     OPTION_VAL_SYNC_FI },
         { "async",    OPTION_VAL_ASYNC_FI },
         { "precache", OPTION_VAL_PRECACHE_FI },
         { NULL, NULL },
      },
      "sync"
   },
#endif
   {
      BEETLE_OPT(cd_fastload),
      BEETLE_OPT_CD_FASTLOAD_LABEL_FI,
      NULL,
      BEETLE_OPT_CD_FASTLOAD_INFO_0_FI,
      NULL,
      "system",
      {
         { "2x(native)", OPTION_VAL_2X_NATIVE_FI },
         { "4x",          OPTION_VAL_4X_FI },
         { "6x",          OPTION_VAL_6X_FI },
         { "8x",          OPTION_VAL_8X_FI },
         { "10x",         OPTION_VAL_10X_FI },
         { "12x",         OPTION_VAL_12X_FI },
         { "14x",         OPTION_VAL_14X_FI },
         { NULL, NULL },
      },
      "2x(native)"
   },
   {
      BEETLE_OPT(skip_bios),
      BEETLE_OPT_SKIP_BIOS_LABEL_FI,
      NULL,
      BEETLE_OPT_SKIP_BIOS_INFO_0_FI,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(override_bios),
      BEETLE_OPT_OVERRIDE_BIOS_LABEL_FI,
      NULL,
      BEETLE_OPT_OVERRIDE_BIOS_INFO_0_FI,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "psxonpsp", OPTION_VAL_PSXONPSP_FI },
         { "ps1_rom",  OPTION_VAL_PS1_ROM_FI },
         { "openbios", OPTION_VAL_OPENBIOS_FI     },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(region),
      BEETLE_OPT_REGION_LABEL_FI,
      NULL,
      BEETLE_OPT_REGION_INFO_0_FI,
      NULL,
      "system",
      {
         { "auto",   OPTION_VAL_AUTO_FI },
         { "ntsc-j", OPTION_VAL_NTSC_J_FI },
         { "ntsc-u", OPTION_VAL_NTSC_U_FI },
         { "pal",    OPTION_VAL_PAL_FI },
         { NULL, NULL },
      },
      "auto"
   },
   {
      BEETLE_OPT(internal_resolution),
      BEETLE_OPT_INTERNAL_RESOLUTION_LABEL_FI,
      NULL,
      BEETLE_OPT_INTERNAL_RESOLUTION_INFO_0_FI,
      NULL,
      "video",
      {
         { "1x(native)", OPTION_VAL_1X_NATIVE_FI },
         { "2x",         OPTION_VAL_2X_FI },
         { "4x",         OPTION_VAL_4X_FI },
         { "8x",         OPTION_VAL_8X_FI },
         { "16x",        OPTION_VAL_16X_FI },
         { NULL, NULL },
      },
      "1x(native)"
   },
#if defined(HAVE_OPENGL) || defined(HAVE_OPENGLES) || defined(HAVE_VULKAN)
   {
      BEETLE_OPT(renderer),
      BEETLE_OPT_RENDERER_LABEL_FI,
      NULL,
      BEETLE_OPT_RENDERER_INFO_0_FI,
      NULL,
      "video",
      {
         { "hardware",    OPTION_VAL_HARDWARE_FI },
#if defined(HAVE_OPENGL) || defined(HAVE_OPENGLES)
         { "hardware_gl", OPTION_VAL_HARDWARE_GL_FI },
#endif
#if defined(HAVE_VULKAN)
         { "hardware_vk", OPTION_VAL_HARDWARE_VK_FI },
#endif
         { "software",    OPTION_VAL_SOFTWARE_FI },
         { NULL, NULL },
      },
      "hardware"
   },
   {
      BEETLE_OPT(renderer_software_fb),
      BEETLE_OPT_RENDERER_SOFTWARE_FB_LABEL_FI,
      NULL,
      BEETLE_OPT_RENDERER_SOFTWARE_FB_INFO_0_FI,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled"
   },
#endif
#if defined(HAVE_OPENGL) || defined(HAVE_OPENGLES)
   {
      BEETLE_OPT(depth),
      BEETLE_OPT_DEPTH_LABEL_FI,
      NULL,
      BEETLE_OPT_DEPTH_INFO_0_FI,
      NULL,
      "video",
      {
         { "16bpp(native)", OPTION_VAL_16BPP_NATIVE_FI },
         { "32bpp",         OPTION_VAL_32BPP_FI },
         { NULL, NULL },
      },
      "16bpp(native)"
   },
   // Sort of, it's more like 15-bit high color and 24-bit true color for visible output. The alpha channel is used for mask bit. Vulkan renderer uses ABGR1555_555 for 31 bits internal? FMVs are always 24-bit on all renderers like original hardware (BGR888, no alpha)
#endif
   {
      BEETLE_OPT(dither_mode),
      BEETLE_OPT_DITHER_MODE_LABEL_FI,
      NULL,
      BEETLE_OPT_DITHER_MODE_INFO_0_FI,
      NULL,
      "video",
      {
         { "1x(native)",          OPTION_VAL_1X_NATIVE_FI },
         { "internal resolution", OPTION_VAL_INTERNAL_RESOLUTION_FI },
         { "disabled",            NULL },
         { NULL, NULL },
      },
      "1x(native)"
   },
#ifdef HAVE_VULKAN
   {
      BEETLE_OPT(scaled_uv_offset),
      BEETLE_OPT_SCALED_UV_OFFSET_LABEL_FI,
      NULL,
      BEETLE_OPT_SCALED_UV_OFFSET_INFO_0_FI,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled"
   },
#endif
#if defined(HAVE_OPENGL) || defined(HAVE_OPENGLES) || defined(HAVE_VULKAN)
   {
      BEETLE_OPT(filter),
      BEETLE_OPT_FILTER_LABEL_FI,
      NULL,
      BEETLE_OPT_FILTER_INFO_0_FI,
      NULL,
      "video",
      {
         { "nearest",  OPTION_VAL_NEAREST_FI },
         { "SABR",     OPTION_VAL_SABR_FI },
         { "xBR",      OPTION_VAL_XBR_FI },
         { "bilinear", OPTION_VAL_BILINEAR_FI },
         { "3-point",  OPTION_VAL_3_POINT_FI },
         { "JINC2",    OPTION_VAL_JINC2_FI },
         { NULL, NULL },
      },
      "nearest"
   },
#ifdef HAVE_VULKAN
   {
      BEETLE_OPT(filter_exclude_sprite),
      BEETLE_OPT_FILTER_EXCLUDE_SPRITE_LABEL_FI,
      NULL,
      BEETLE_OPT_FILTER_EXCLUDE_SPRITE_INFO_0_FI,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "opaque", OPTION_VAL_OPAQUE_FI },
         { "all", OPTION_VAL_ALL_FI },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(filter_exclude_2d_polygon),
      BEETLE_OPT_FILTER_EXCLUDE_2D_POLYGON_LABEL_FI,
      NULL,
      BEETLE_OPT_FILTER_EXCLUDE_2D_POLYGON_INFO_0_FI,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "opaque", OPTION_VAL_OPAQUE_FI },
         { "all", OPTION_VAL_ALL_FI },
         { NULL, NULL },
      },
      "disabled"
   },
#endif
#endif
#ifdef HAVE_VULKAN
   {
      BEETLE_OPT(adaptive_smoothing),
      BEETLE_OPT_ADAPTIVE_SMOOTHING_LABEL_FI,
      NULL,
      BEETLE_OPT_ADAPTIVE_SMOOTHING_INFO_0_FI,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(super_sampling),
      BEETLE_OPT_SUPER_SAMPLING_LABEL_FI,
      NULL,
      BEETLE_OPT_SUPER_SAMPLING_INFO_0_FI,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(msaa),
      BEETLE_OPT_MSAA_LABEL_FI,
      NULL,
      BEETLE_OPT_MSAA_INFO_0_FI,
      NULL,
      "video",
      {
         { "1x",  OPTION_VAL_1X_FI },
         { "2x",  OPTION_VAL_2X_FI },
         { "4x",  OPTION_VAL_4X_FI },
         { "8x",  OPTION_VAL_8X_FI },
         { "16x", OPTION_VAL_16X_FI },
         { NULL, NULL },
      },
      "1x"
   },
   {
      BEETLE_OPT(mdec_yuv),
      BEETLE_OPT_MDEC_YUV_LABEL_FI,
      NULL,
      BEETLE_OPT_MDEC_YUV_INFO_0_FI,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(track_textures),
      BEETLE_OPT_TRACK_TEXTURES_LABEL_FI,
      NULL,
      BEETLE_OPT_TRACK_TEXTURES_INFO_0_FI,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
#ifdef TEXTURE_DUMPING_ENABLED
   {
      BEETLE_OPT(dump_textures),
      BEETLE_OPT_DUMP_TEXTURES_LABEL_FI,
      NULL,
      BEETLE_OPT_DUMP_TEXTURES_INFO_0_FI,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
#endif
   {
      BEETLE_OPT(replace_textures),
      BEETLE_OPT_REPLACE_TEXTURES_LABEL_FI,
      NULL,
      BEETLE_OPT_REPLACE_TEXTURES_INFO_0_FI,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
#endif
   {
      BEETLE_OPT(pgxp_mode),
      BEETLE_OPT_PGXP_MODE_LABEL_FI,
      NULL,
      BEETLE_OPT_PGXP_MODE_INFO_0_FI,
      NULL,
      "pgxp",
      {
         { "disabled",     NULL },
         { "memory only",  OPTION_VAL_MEMORY_ONLY_FI },
         { "memory + CPU", OPTION_VAL_MEMORY_CPU_FI },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(pgxp_2d_tol),
      BEETLE_OPT_PGXP_2D_TOL_LABEL_FI,
      NULL,
      BEETLE_OPT_PGXP_2D_TOL_INFO_0_FI,
      NULL,
      "pgxp",
      {
         { "disabled", NULL },
         { "0px", OPTION_VAL_0PX_FI },
         { "1px", OPTION_VAL_1PX_FI },
         { "2px", OPTION_VAL_2PX_FI },
         { "3px", OPTION_VAL_3PX_FI },
         { "4px", OPTION_VAL_4PX_FI },
         { "5px", OPTION_VAL_5PX_FI },
         { "6px", OPTION_VAL_6PX_FI },
         { "7px", OPTION_VAL_7PX_FI },
         { "8px", OPTION_VAL_8PX_FI },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(pgxp_nclip),
      BEETLE_OPT_PGXP_NCLIP_LABEL_FI,
      NULL,
      BEETLE_OPT_PGXP_NCLIP_INFO_0_FI,
      NULL,
      "pgxp",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
#if defined(HAVE_OPENGL) || defined(HAVE_OPENGLES) || defined(HAVE_VULKAN)
   {
      BEETLE_OPT(pgxp_vertex),
      BEETLE_OPT_PGXP_VERTEX_LABEL_FI,
      NULL,
      BEETLE_OPT_PGXP_VERTEX_INFO_0_FI,
      NULL,
      "pgxp",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
#endif
   {
      BEETLE_OPT(pgxp_texture),
      BEETLE_OPT_PGXP_TEXTURE_LABEL_FI,
      NULL,
      BEETLE_OPT_PGXP_TEXTURE_INFO_0_FI,
      NULL,
      "pgxp",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(display_internal_fps),
      BEETLE_OPT_DISPLAY_INTERNAL_FPS_LABEL_FI,
      NULL,
      BEETLE_OPT_DISPLAY_INTERNAL_FPS_INFO_0_FI,
      NULL,
      "osd",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(display_osd),
      BEETLE_OPT_DISPLAY_OSD_LABEL_FI,
      NULL,
      BEETLE_OPT_DISPLAY_OSD_INFO_0_FI,
      NULL,
      "osd",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled"
   },
#if defined(HAVE_OPENGL) || defined(HAVE_OPENGLES) || defined(HAVE_VULKAN)
   {
      BEETLE_OPT(display_vram),
      BEETLE_OPT_DISPLAY_VRAM_LABEL_FI,
      NULL,
      BEETLE_OPT_DISPLAY_VRAM_INFO_0_FI,
      NULL,
      "osd",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
#endif
   {
      BEETLE_OPT(analog_calibration),
      BEETLE_OPT_ANALOG_CALIBRATION_LABEL_FI,
      NULL,
      BEETLE_OPT_ANALOG_CALIBRATION_INFO_0_FI,
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(analog_toggle),
      BEETLE_OPT_ANALOG_TOGGLE_LABEL_FI,
      NULL,
      BEETLE_OPT_ANALOG_TOGGLE_INFO_0_FI,
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { "enabled-analog", OPTION_VAL_ENABLED_ANALOG_FI },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(analog_toggle_combo),
      BEETLE_OPT_ANALOG_TOGGLE_COMBO_LABEL_FI,
      NULL,
      BEETLE_OPT_ANALOG_TOGGLE_COMBO_INFO_0_FI,
      NULL,
      "input",
      {
         { "l1+l2+r1+r2+start+select", OPTION_VAL_L1_L2_R1_R2_START_SELECT_FI },
         { "l1+r1+select",             OPTION_VAL_L1_R1_SELECT_FI },
         { "l1+r1+start",              OPTION_VAL_L1_R1_START_FI },
         { "l1+r1+l3",                 OPTION_VAL_L1_R1_L3_FI },
         { "l1+r1+r3",                 OPTION_VAL_L1_R1_R3_FI },
         { "l2+r2+select",             OPTION_VAL_L2_R2_SELECT_FI },
         { "l2+r2+start",              OPTION_VAL_L2_R2_START_FI },
         { "l2+r2+l3",                 OPTION_VAL_L2_R2_L3_FI },
         { "l2+r2+r3",                 OPTION_VAL_L2_R2_R3_FI },
         { "l3+r3",                    OPTION_VAL_L3_R3_FI },
         { NULL, NULL },
      },
      "l1+r1+select"
   },
   {
      BEETLE_OPT(analog_toggle_hold),
      BEETLE_OPT_ANALOG_TOGGLE_HOLD_LABEL_FI,
      NULL,
      BEETLE_OPT_ANALOG_TOGGLE_HOLD_INFO_0_FI,
      NULL,
      "input",
      {
         { "0", OPTION_VAL_0_FI },
         { "1", OPTION_VAL_1_FI },
         { "2", OPTION_VAL_2_FI },
         { "3", OPTION_VAL_3_FI },
         { "4", OPTION_VAL_4_FI },
         { "5", OPTION_VAL_5_FI },
         { NULL, NULL },
      },
      "1"
   },
   {
      BEETLE_OPT(enable_multitap_port1),
      BEETLE_OPT_ENABLE_MULTITAP_PORT1_LABEL_FI,
      NULL,
      BEETLE_OPT_ENABLE_MULTITAP_PORT1_INFO_0_FI,
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(enable_multitap_port2),
      BEETLE_OPT_ENABLE_MULTITAP_PORT2_LABEL_FI,
      NULL,
      BEETLE_OPT_ENABLE_MULTITAP_PORT2_INFO_0_FI,
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(gun_input_mode),
      BEETLE_OPT_GUN_INPUT_MODE_LABEL_FI,
      NULL,
      BEETLE_OPT_GUN_INPUT_MODE_INFO_0_FI,
      NULL,
      "input",
      {
         { "lightgun",    OPTION_VAL_LIGHTGUN_FI },
         { "touchscreen", OPTION_VAL_TOUCHSCREEN_FI },
         { NULL, NULL },
      },
      "lightgun"
   },
   // Shouldn't the gun_input_mode just be Mouse vs. Touchscreen?
   {
      BEETLE_OPT(gun_cursor),
      BEETLE_OPT_GUN_CURSOR_LABEL_FI,
      NULL,
      BEETLE_OPT_GUN_CURSOR_INFO_0_FI,
      NULL,
      "input",
      {
         { "off",   "disabled" },
         { "cross", OPTION_VAL_CROSS_FI },
         { "dot",   OPTION_VAL_DOT_FI },
         { NULL, NULL },
      },
      "cross"
   },
   {
      BEETLE_OPT(crosshair_color_p1),
      BEETLE_OPT_CROSSHAIR_COLOR_P1_LABEL_FI,
      NULL,
      BEETLE_OPT_CROSSHAIR_COLOR_P1_INFO_0_FI,
      NULL,
      "input",
      {
         { "red",    OPTION_VAL_RED_FI },
         { "blue",   OPTION_VAL_BLUE_FI },
         { "green",  OPTION_VAL_GREEN_FI },
         { "orange", OPTION_VAL_ORANGE_FI },
         { "yellow", OPTION_VAL_YELLOW_FI },
         { "cyan",   OPTION_VAL_CYAN_FI },
         { "pink",   OPTION_VAL_PINK_FI },
         { "purple", OPTION_VAL_PURPLE_FI },
         { "black",  OPTION_VAL_BLACK_FI },
         { "white",  OPTION_VAL_WHITE_FI },
         { NULL, NULL },
      },
      "red"
   },
   {
      BEETLE_OPT(crosshair_color_p2),
      BEETLE_OPT_CROSSHAIR_COLOR_P2_LABEL_FI,
      NULL,
      BEETLE_OPT_CROSSHAIR_COLOR_P2_INFO_0_FI,
      NULL,
      "input",
      {
         { "blue",   OPTION_VAL_BLUE_FI },
         { "red",    OPTION_VAL_RED_FI },
         { "green",  OPTION_VAL_GREEN_FI },
         { "orange", OPTION_VAL_ORANGE_FI },
         { "yellow", OPTION_VAL_YELLOW_FI },
         { "cyan",   OPTION_VAL_CYAN_FI },
         { "pink",   OPTION_VAL_PINK_FI },
         { "purple", OPTION_VAL_PURPLE_FI },
         { "black",  OPTION_VAL_BLACK_FI },
         { "white",  OPTION_VAL_WHITE_FI },
         { NULL, NULL },
      },
      "blue"
   },
   {
      BEETLE_OPT(mouse_sensitivity),
      BEETLE_OPT_MOUSE_SENSITIVITY_LABEL_FI,
      NULL,
      BEETLE_OPT_MOUSE_SENSITIVITY_INFO_0_FI,
      NULL,
      "input",
      {
         { "5%",   OPTION_VAL_5_O42_FI },
         { "10%",  OPTION_VAL_10_FI },
         { "15%",  OPTION_VAL_15_FI },
         { "20%",  OPTION_VAL_20_FI },
         { "25%",  OPTION_VAL_25_FI },
         { "30%",  OPTION_VAL_30_FI },
         { "35%",  OPTION_VAL_35_FI },
         { "40%",  OPTION_VAL_40_FI },
         { "45%",  OPTION_VAL_45_FI },
         { "50%",  OPTION_VAL_50_FI },
         { "55%",  OPTION_VAL_55_FI },
         { "60%",  OPTION_VAL_60_FI },
         { "65%",  OPTION_VAL_65_FI },
         { "70%",  OPTION_VAL_70_FI },
         { "75%",  OPTION_VAL_75_FI },
         { "80%",  OPTION_VAL_80_FI },
         { "85%",  OPTION_VAL_85_FI },
         { "90%",  OPTION_VAL_90_FI },
         { "95%",  OPTION_VAL_95_FI },
         { "100%", OPTION_VAL_100_O42_FI },
         { "105%", OPTION_VAL_105_FI },
         { "110%", OPTION_VAL_110_FI },
         { "115%", OPTION_VAL_115_FI },
         { "120%", OPTION_VAL_120_FI },
         { "125%", OPTION_VAL_125_FI },
         { "130%", OPTION_VAL_130_FI },
         { "135%", OPTION_VAL_135_FI },
         { "140%", OPTION_VAL_140_FI },
         { "145%", OPTION_VAL_145_FI },
         { "150%", OPTION_VAL_150_FI },
         { "155%", OPTION_VAL_155_FI },
         { "160%", OPTION_VAL_160_FI },
         { "165%", OPTION_VAL_165_FI },
         { "170%", OPTION_VAL_170_FI },
         { "175%", OPTION_VAL_175_FI },
         { "180%", OPTION_VAL_180_FI },
         { "185%", OPTION_VAL_185_FI },
         { "190%", OPTION_VAL_190_FI },
         { "195%", OPTION_VAL_195_FI },
         { "200%", OPTION_VAL_200_FI },
         { NULL, NULL },
      },
      "100%"
   },
   {
      BEETLE_OPT(negcon_response),
      BEETLE_OPT_NEGCON_RESPONSE_LABEL_FI,
      NULL,
      BEETLE_OPT_NEGCON_RESPONSE_INFO_0_FI,
      NULL,
      "input",
      {
         { "linear",    OPTION_VAL_LINEAR_FI },
         { "quadratic", OPTION_VAL_QUADRATIC_FI },
         { "cubic",     OPTION_VAL_CUBIC_FI },
         { NULL, NULL },
      },
      "linear"
   },
   {
      BEETLE_OPT(negcon_deadzone),
      BEETLE_OPT_NEGCON_DEADZONE_LABEL_FI,
      NULL,
      BEETLE_OPT_NEGCON_DEADZONE_INFO_0_FI,
      NULL,
      "input",
      {
         { "0%",  OPTION_VAL_0_O44_FI },
         { "5%",  OPTION_VAL_5_O42_FI },
         { "10%", OPTION_VAL_10_FI },
         { "15%", OPTION_VAL_15_FI },
         { "20%", OPTION_VAL_20_FI },
         { "25%", OPTION_VAL_25_FI },
         { "30%", OPTION_VAL_30_FI },
         { NULL, NULL },
      },
      "0%"
   },
   {
      BEETLE_OPT(use_mednafen_memcard0_method),
      BEETLE_OPT_USE_MEDNAFEN_MEMCARD0_METHOD_LABEL_FI,
      NULL,
      BEETLE_OPT_USE_MEDNAFEN_MEMCARD0_METHOD_INFO_0_FI,
      NULL,
      "memcards",
      {
         { "libretro", OPTION_VAL_LIBRETRO_FI },
         { "mednafen", OPTION_VAL_MEDNAFEN_FI },
         { NULL, NULL },
      },
      "libretro"
   },
   {
      BEETLE_OPT(enable_memcard1),
      BEETLE_OPT_ENABLE_MEMCARD1_LABEL_FI,
      NULL,
      BEETLE_OPT_ENABLE_MEMCARD1_INFO_0_FI,
      NULL,
      "memcards",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(shared_memory_cards),
      BEETLE_OPT_SHARED_MEMORY_CARDS_LABEL_FI,
      NULL,
      BEETLE_OPT_SHARED_MEMORY_CARDS_INFO_0_FI,
      NULL,
      "memcards",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(memcard_left_index),
      BEETLE_OPT_MEMCARD_LEFT_INDEX_LABEL_FI,
      NULL,
      BEETLE_OPT_MEMCARD_LEFT_INDEX_INFO_0_FI,
      NULL,
      "memcards",
      {
         { "0",  OPTION_VAL_0_O48_FI },
         { "1",  NULL },
         { "2",  NULL },
         { "3",  NULL },
         { "4",  NULL },
         { "5",  NULL },
         { "6",  NULL },
         { "7",  NULL },
         { "8",  NULL },
         { "9",  NULL },
         { "10", NULL },
         { "11", NULL },
         { "12", NULL },
         { "13", NULL },
         { "14", NULL },
         { "15", NULL },
         { "16", NULL },
         { "17", NULL },
         { "18", NULL },
         { "19", NULL },
         { "20", NULL },
         { "21", NULL },
         { "22", NULL },
         { "23", NULL },
         { "24", NULL },
         { "25", NULL },
         { "26", NULL },
         { "27", NULL },
         { "28", NULL },
         { "29", NULL },
         { "30", NULL },
         { "31", NULL },
         { "32", NULL },
         { "33", NULL },
         { "34", NULL },
         { "35", NULL },
         { "36", NULL },
         { "37", NULL },
         { "38", NULL },
         { "39", NULL },
         { "40", NULL },
         { "41", NULL },
         { "42", NULL },
         { "43", NULL },
         { "44", NULL },
         { "45", NULL },
         { "46", NULL },
         { "47", NULL },
         { "48", NULL },
         { "49", NULL },
         { "50", NULL },
         { "51", NULL },
         { "52", NULL },
         { "53", NULL },
         { "54", NULL },
         { "55", NULL },
         { "56", NULL },
         { "57", NULL },
         { "58", NULL },
         { "59", NULL },
         { "60", NULL },
         { "61", NULL },
         { "62", NULL },
         { "63", NULL },
         { NULL, NULL },
      },
      "0"
   },
   {
      BEETLE_OPT(memcard_right_index),
      BEETLE_OPT_MEMCARD_RIGHT_INDEX_LABEL_FI,
      NULL,
      BEETLE_OPT_MEMCARD_RIGHT_INDEX_INFO_0_FI,
      NULL,
      "memcards",
      {
         { "0",  NULL },
         { "1",  OPTION_VAL_1_O49_FI },
         { "2",  NULL },
         { "3",  NULL },
         { "4",  NULL },
         { "5",  NULL },
         { "6",  NULL },
         { "7",  NULL },
         { "8",  NULL },
         { "9",  NULL },
         { "10", NULL },
         { "11", NULL },
         { "12", NULL },
         { "13", NULL },
         { "14", NULL },
         { "15", NULL },
         { "16", NULL },
         { "17", NULL },
         { "18", NULL },
         { "19", NULL },
         { "20", NULL },
         { "21", NULL },
         { "22", NULL },
         { "23", NULL },
         { "24", NULL },
         { "25", NULL },
         { "26", NULL },
         { "27", NULL },
         { "28", NULL },
         { "29", NULL },
         { "30", NULL },
         { "31", NULL },
         { "32", NULL },
         { "33", NULL },
         { "34", NULL },
         { "35", NULL },
         { "36", NULL },
         { "37", NULL },
         { "38", NULL },
         { "39", NULL },
         { "40", NULL },
         { "41", NULL },
         { "42", NULL },
         { "43", NULL },
         { "44", NULL },
         { "45", NULL },
         { "46", NULL },
         { "47", NULL },
         { "48", NULL },
         { "49", NULL },
         { "50", NULL },
         { "51", NULL },
         { "52", NULL },
         { "53", NULL },
         { "54", NULL },
         { "55", NULL },
         { "56", NULL },
         { "57", NULL },
         { "58", NULL },
         { "59", NULL },
         { "60", NULL },
         { "61", NULL },
         { "62", NULL },
         { "63", NULL },
         { NULL, NULL },
      },
      "1"
   },
   {
      BEETLE_OPT(line_render),
      BEETLE_OPT_LINE_RENDER_LABEL_FI,
      NULL,
      BEETLE_OPT_LINE_RENDER_INFO_0_FI,
      NULL,
      "hacks",
      {
         { "default",    OPTION_VAL_DEFAULT_FI },
         { "aggressive", OPTION_VAL_AGGRESSIVE_FI },
         { "disabled",   NULL },
         { NULL, NULL },
      },
      "default"
   },
   {
      BEETLE_OPT(widescreen_hack),
      BEETLE_OPT_WIDESCREEN_HACK_LABEL_FI,
      NULL,
      BEETLE_OPT_WIDESCREEN_HACK_INFO_0_FI,
      NULL,
      "hacks",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(widescreen_hack_aspect_ratio),
      BEETLE_OPT_WIDESCREEN_HACK_ASPECT_RATIO_LABEL_FI,
      NULL,
      BEETLE_OPT_WIDESCREEN_HACK_ASPECT_RATIO_INFO_0_FI,
      NULL,
      "hacks",
      {
         { "16:9",  OPTION_VAL_16_9_FI },
         { "16:10", OPTION_VAL_16_10_FI },
         { "18:9",  OPTION_VAL_18_9_FI },
         { "19:9",  OPTION_VAL_19_9_FI },
         { "20:9",  OPTION_VAL_20_9_FI },
         { "21:9",  OPTION_VAL_21_9_FI }, // 64:27
         { "32:9",  OPTION_VAL_32_9_FI },
         { NULL,    NULL },
      },
      "16:9"
   },
#if defined(HAVE_LIGHTREC)
   {
      BEETLE_OPT(cpu_dynarec),
      BEETLE_OPT_CPU_DYNAREC_LABEL_FI,
      NULL,
      BEETLE_OPT_CPU_DYNAREC_INFO_0_FI,
      NULL,
      "hacks",
      {
         { "disabled",        OPTION_VAL_DISABLED_FI },
         { "execute",         OPTION_VAL_EXECUTE_FI },
         { "execute_one",     OPTION_VAL_EXECUTE_ONE_FI },
         { "run_interpreter", OPTION_VAL_RUN_INTERPRETER_FI },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(dynarec_invalidate),
      BEETLE_OPT_DYNAREC_INVALIDATE_LABEL_FI,
      NULL,
      BEETLE_OPT_DYNAREC_INVALIDATE_INFO_0_FI,
      NULL,
      "hacks",
      {
         { "full", OPTION_VAL_FULL_FI },
         { "dma",  OPTION_VAL_DMA_FI },
         { NULL, NULL },
      },
      "full"
   },
   {
      BEETLE_OPT(dynarec_eventcycles),
      BEETLE_OPT_DYNAREC_EVENTCYCLES_LABEL_FI,
      NULL,
      BEETLE_OPT_DYNAREC_EVENTCYCLES_INFO_0_FI,
      NULL,
      "hacks",
      {
         { "128",  OPTION_VAL_128_FI },
         { "256",  NULL },
         { "384",  NULL },
         { "512",  NULL },
         { "640",  NULL },
         { "768",  NULL },
         { "896",  NULL },
         { "1024", NULL },
         { "1152", NULL },
         { "1280", NULL },
         { "1408", NULL },
         { "1536", NULL },
         { "1664", NULL },
         { "1792", NULL },
         { "1920", NULL },
         { "2048", NULL },
         { NULL, NULL },
      },
      "128"
   },
   {
      BEETLE_OPT(dynarec_spu_samples),
      BEETLE_OPT_DYNAREC_SPU_SAMPLES_LABEL_FI,
      NULL,
      BEETLE_OPT_DYNAREC_SPU_SAMPLES_INFO_0_FI,
      NULL,
      "hacks",
      {
         { "1",  OPTION_VAL_1_O49_FI },
         { "4",  NULL },
         { "16", NULL },
         { NULL, NULL },
      },
      "1"
   },
#endif
   {
      BEETLE_OPT(core_timing_fps),
      BEETLE_OPT_CORE_TIMING_FPS_LABEL_FI,
      NULL,
      BEETLE_OPT_CORE_TIMING_FPS_INFO_0_FI,
      NULL,
      "video",
      {
         { "auto_toggle",       OPTION_VAL_AUTO_TOGGLE_FI },
         { "force_progressive", OPTION_VAL_FORCE_PROGRESSIVE_FI },
         { "force_interlaced",  OPTION_VAL_FORCE_INTERLACED_FI },
      },
      "auto_toggle"
   },
   {
      BEETLE_OPT(deinterlacer),
      BEETLE_OPT_DEINTERLACER_LABEL_FI,
      NULL,
      BEETLE_OPT_DEINTERLACER_INFO_0_FI,
      NULL,
      "video",
      {
         { "weave",      OPTION_VAL_WEAVE_FI },
         { "bob",        OPTION_VAL_BOB_FI },
         { "bob_offset", OPTION_VAL_BOB_OFFSET_FI },
         { "fastmad",    OPTION_VAL_FASTMAD_FI },
         { "off",        "Off" },
         { NULL, NULL },
      },
      "weave"
   },
   {
      BEETLE_OPT(frame_duping),
      BEETLE_OPT_FRAME_DUPING_LABEL_FI,
      NULL,
      BEETLE_OPT_FRAME_DUPING_INFO_0_FI,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(pal_video_timing_override),
      BEETLE_OPT_PAL_VIDEO_TIMING_OVERRIDE_LABEL_FI,
      NULL,
      BEETLE_OPT_PAL_VIDEO_TIMING_OVERRIDE_INFO_0_FI,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(aspect_ratio),
      BEETLE_OPT_ASPECT_RATIO_LABEL_FI,
      NULL,
      BEETLE_OPT_ASPECT_RATIO_INFO_0_FI,
      NULL,
      "video",
      {
         { "corrected", OPTION_VAL_CORRECTED_FI },
         { "uncorrected", OPTION_VAL_UNCORRECTED_FI },
         { "4:3",  OPTION_VAL_4_3_FI },
         { "ntsc", OPTION_VAL_NTSC_FI },
      },
      "corrected"
   },
   {
      BEETLE_OPT(crop_overscan),
      BEETLE_OPT_CROP_OVERSCAN_LABEL_FI,
      NULL,
      BEETLE_OPT_CROP_OVERSCAN_INFO_0_FI,
      NULL,
      "video",
      {
         { "disabled", "Disabled" },
         { "static",   OPTION_VAL_STATIC_FI },
         { "smart",    OPTION_VAL_SMART_FI },
         { NULL, NULL },
      },
      "smart"
   },
   {
      BEETLE_OPT(image_crop),
      BEETLE_OPT_IMAGE_CROP_LABEL_FI,
      NULL,
      BEETLE_OPT_IMAGE_CROP_INFO_0_FI,
      NULL,
      "video",
      {
         { "disabled", "0" },
         { "1px",      OPTION_VAL_1PX_FI },
         { "2px",      OPTION_VAL_2PX_FI },
         { "3px",      OPTION_VAL_3PX_FI },
         { "4px",      OPTION_VAL_4PX_FI },
         { "5px",      OPTION_VAL_5PX_FI },
         { "6px",      OPTION_VAL_6PX_FI },
         { "7px",      OPTION_VAL_7PX_FI },
         { "8px",      OPTION_VAL_8PX_FI },
         { "9px",      OPTION_VAL_9PX_FI },
         { "10px",     OPTION_VAL_10PX_FI },
         { "11px",     OPTION_VAL_11PX_FI },
         { "12px",     OPTION_VAL_12PX_FI },
         { "13px",     OPTION_VAL_13PX_FI },
         { "14px",     OPTION_VAL_14PX_FI },
         { "15px",     OPTION_VAL_15PX_FI },
         { "16px",     OPTION_VAL_16PX_FI },
         { "17px",     OPTION_VAL_17PX_FI },
         { "18px",     OPTION_VAL_18PX_FI },
         { "19px",     OPTION_VAL_19PX_FI },
         { "20px",     OPTION_VAL_20PX_FI },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(image_offset),
      BEETLE_OPT_IMAGE_OFFSET_LABEL_FI,
      NULL,
      BEETLE_OPT_IMAGE_OFFSET_INFO_0_FI,
      NULL,
      "video",
      {
         { "-12px",    OPTION_VAL_12PX_O64_FI },
         { "-11px",    OPTION_VAL_11PX_O64_FI },
         { "-10px",    OPTION_VAL_10PX_O64_FI },
         { "-9px",     OPTION_VAL_9PX_O64_FI },
         { "-8px",     OPTION_VAL_8PX_O64_FI },
         { "-7px",     OPTION_VAL_7PX_O64_FI },
         { "-6px",     OPTION_VAL_6PX_O64_FI },
         { "-5px",     OPTION_VAL_5PX_O64_FI },
         { "-4px",     OPTION_VAL_4PX_O64_FI },
         { "-3px",     OPTION_VAL_3PX_O64_FI },
         { "-2px",     OPTION_VAL_2PX_O64_FI },
         { "-1px",     OPTION_VAL_1PX_O64_FI },
         { "disabled", "0" },
         { "+1px",     OPTION_VAL_1PX_O64_0_FI },
         { "+2px",     OPTION_VAL_2PX_O64_0_FI },
         { "+3px",     OPTION_VAL_3PX_O64_0_FI },
         { "+4px",     OPTION_VAL_4PX_O64_0_FI },
         { "+5px",     OPTION_VAL_5PX_O64_0_FI },
         { "+6px",     OPTION_VAL_6PX_O64_0_FI },
         { "+7px",     OPTION_VAL_7PX_O64_0_FI },
         { "+8px",     OPTION_VAL_8PX_O64_0_FI },
         { "+9px",     OPTION_VAL_9PX_O64_0_FI },
         { "+10px",    OPTION_VAL_10PX_O64_0_FI },
         { "+11px",    OPTION_VAL_11PX_O64_0_FI },
         { "+12px",    OPTION_VAL_12PX_O64_0_FI },
         { NULL, NULL },
      },
      "disabled"
   },
#if defined(HAVE_OPENGL) || defined(HAVE_OPENGLES) || defined(HAVE_VULKAN)
   {
      BEETLE_OPT(image_offset_cycles),
      BEETLE_OPT_IMAGE_OFFSET_CYCLES_LABEL_FI,
      NULL,
      BEETLE_OPT_IMAGE_OFFSET_CYCLES_INFO_0_FI,
      NULL,
      "video",
      {
         { "-40",      NULL },
         { "-39",      NULL },
         { "-38",      NULL },
         { "-37",      NULL },
         { "-36",      NULL },
         { "-35",      NULL },
         { "-34",      NULL },
         { "-33",      NULL },
         { "-32",      NULL },
         { "-31",      NULL },
         { "-30",      NULL },
         { "-29",      NULL },
         { "-28",      NULL },
         { "-27",      NULL },
         { "-26",      NULL },
         { "-25",      NULL },
         { "-24",      NULL },
         { "-23",      NULL },
         { "-22",      NULL },
         { "-21",      NULL },
         { "-20",      NULL },
         { "-19",      NULL },
         { "-18",      NULL },
         { "-17",      NULL },
         { "-16",      NULL },
         { "-15",      NULL },
         { "-14",      NULL },
         { "-13",      NULL },
         { "-12",      NULL },
         { "-11",      NULL },
         { "-10",      NULL },
         { "-9",       NULL },
         { "-8",       NULL },
         { "-7",       NULL },
         { "-6",       NULL },
         { "-5",       NULL },
         { "-4",       NULL },
         { "-3",       NULL },
         { "-2",       NULL },
         { "-1",       NULL },
         { "0",        "0" },
         { "+1",       NULL },
         { "+2",       NULL },
         { "+3",       NULL },
         { "+4",       NULL },
         { "+5",       NULL },
         { "+6",       NULL },
         { "+7",       NULL },
         { "+8",       NULL },
         { "+9",       NULL },
         { "+10",      NULL },
         { "+11",      NULL },
         { "+12",      NULL },
         { "+13",      NULL },
         { "+14",      NULL },
         { "+15",      NULL },
         { "+16",      NULL },
         { "+17",      NULL },
         { "+18",      NULL },
         { "+19",      NULL },
         { "+20",      NULL },
         { "+21",      NULL },
         { "+22",      NULL },
         { "+23",      NULL },
         { "+24",      NULL },
         { "+25",      NULL },
         { "+26",      NULL },
         { "+27",      NULL },
         { "+28",      NULL },
         { "+29",      NULL },
         { "+30",      NULL },
         { "+31",      NULL },
         { "+32",      NULL },
         { "+33",      NULL },
         { "+34",      NULL },
         { "+35",      NULL },
         { "+36",      NULL },
         { "+37",      NULL },
         { "+38",      NULL },
         { "+39",      NULL },
         { "+40",      NULL },
         { NULL, NULL},
      },
      "0"
   },
#endif
   {
      BEETLE_OPT(initial_scanline),
      BEETLE_OPT_INITIAL_SCANLINE_LABEL_FI,
      NULL,
      BEETLE_OPT_INITIAL_SCANLINE_INFO_0_FI,
      NULL,
      "video",
      {
         { "0",  OPTION_VAL_0_O48_FI },
         { "1",  NULL },
         { "2",  NULL },
         { "3",  NULL },
         { "4",  NULL },
         { "5",  NULL },
         { "6",  NULL },
         { "7",  NULL },
         { "8",  NULL },
         { "9",  NULL },
         { "10", NULL },
         { "11", NULL },
         { "12", NULL },
         { "13", NULL },
         { "14", NULL },
         { "15", NULL },
         { "16", NULL },
         { "17", NULL },
         { "18", NULL },
         { "19", NULL },
         { "20", NULL },
         { "21", NULL },
         { "22", NULL },
         { "23", NULL },
         { "24", NULL },
         { "25", NULL },
         { "26", NULL },
         { "27", NULL },
         { "28", NULL },
         { "29", NULL },
         { "30", NULL },
         { "31", NULL },
         { "32", NULL },
         { "33", NULL },
         { "34", NULL },
         { "35", NULL },
         { "36", NULL },
         { "37", NULL },
         { "38", NULL },
         { "39", NULL },
         { "40", NULL },
         { NULL, NULL },
      },
      "0"
   },
   {
      BEETLE_OPT(last_scanline),
      BEETLE_OPT_LAST_SCANLINE_LABEL_FI,
      NULL,
      BEETLE_OPT_LAST_SCANLINE_INFO_0_FI,
      NULL,
      "video",
      {
         { "210", NULL },
         { "211", NULL },
         { "212", NULL },
         { "213", NULL },
         { "214", NULL },
         { "215", NULL },
         { "216", NULL },
         { "217", NULL },
         { "218", NULL },
         { "219", NULL },
         { "220", NULL },
         { "221", NULL },
         { "222", NULL },
         { "223", NULL },
         { "224", NULL },
         { "225", NULL },
         { "226", NULL },
         { "227", NULL },
         { "228", NULL },
         { "229", NULL },
         { "230", NULL },
         { "231", NULL },
         { "232", NULL },
         { "233", NULL },
         { "234", NULL },
         { "235", NULL },
         { "236", NULL },
         { "237", NULL },
         { "238", NULL },
         { "239", OPTION_VAL_239_FI },
         { NULL, NULL },
      },
      "239"
   },
   {
      BEETLE_OPT(initial_scanline_pal),
      BEETLE_OPT_INITIAL_SCANLINE_PAL_LABEL_FI,
      NULL,
      BEETLE_OPT_INITIAL_SCANLINE_PAL_INFO_0_FI,
      NULL,
      "video",
      {
         { "0",  OPTION_VAL_0_O48_FI },
         { "1",  NULL },
         { "2",  NULL },
         { "3",  NULL },
         { "4",  NULL },
         { "5",  NULL },
         { "6",  NULL },
         { "7",  NULL },
         { "8",  NULL },
         { "9",  NULL },
         { "10", NULL },
         { "11", NULL },
         { "12", NULL },
         { "13", NULL },
         { "14", NULL },
         { "15", NULL },
         { "16", NULL },
         { "17", NULL },
         { "18", NULL },
         { "19", NULL },
         { "20", NULL },
         { "21", NULL },
         { "22", NULL },
         { "23", NULL },
         { "24", NULL },
         { "25", NULL },
         { "26", NULL },
         { "27", NULL },
         { "28", NULL },
         { "29", NULL },
         { "30", NULL },
         { "31", NULL },
         { "32", NULL },
         { "33", NULL },
         { "34", NULL },
         { "35", NULL },
         { "36", NULL },
         { "37", NULL },
         { "38", NULL },
         { "39", NULL },
         { "40", NULL },
         { NULL, NULL },
      },
      "0"
   },
   {
      BEETLE_OPT(last_scanline_pal),
      BEETLE_OPT_LAST_SCANLINE_PAL_LABEL_FI,
      NULL,
      BEETLE_OPT_LAST_SCANLINE_PAL_INFO_0_FI,
      NULL,
      "video",
      {
         { "230", NULL },
         { "231", NULL },
         { "232", NULL },
         { "233", NULL },
         { "234", NULL },
         { "235", NULL },
         { "236", NULL },
         { "237", NULL },
         { "238", NULL },
         { "239", NULL },
         { "240", NULL },
         { "241", NULL },
         { "242", NULL },
         { "243", NULL },
         { "244", NULL },
         { "245", NULL },
         { "246", NULL },
         { "247", NULL },
         { "248", NULL },
         { "249", NULL },
         { "250", NULL },
         { "251", NULL },
         { "252", NULL },
         { "253", NULL },
         { "254", NULL },
         { "255", NULL },
         { "256", NULL },
         { "257", NULL },
         { "258", NULL },
         { "259", NULL },
         { "260", NULL },
         { "261", NULL },
         { "262", NULL },
         { "263", NULL },
         { "264", NULL },
         { "265", NULL },
         { "266", NULL },
         { "267", NULL },
         { "268", NULL },
         { "269", NULL },
         { "270", NULL },
         { "271", NULL },
         { "272", NULL },
         { "273", NULL },
         { "274", NULL },
         { "275", NULL },
         { "276", NULL },
         { "277", NULL },
         { "278", NULL },
         { "279", NULL },
         { "280", NULL },
         { "281", NULL },
         { "282", NULL },
         { "283", NULL },
         { "284", NULL },
         { "285", NULL },
         { "286", NULL },
         { "287", OPTION_VAL_287_FI },
         { NULL, NULL },
      },
      "287"
   },
   { NULL, NULL, NULL, NULL, NULL, NULL, {{0}}, NULL },
};
struct retro_core_options_v2 options_fi = {
   option_cats_fi,
   option_defs_fi
};

/* RETRO_LANGUAGE_FR */

#define CATEGORY_SYSTEM_LABEL_FR "Système"
#define CATEGORY_SYSTEM_INFO_0_FR "Modifier les options système et d'overclocking."
#define CATEGORY_VIDEO_LABEL_FR "Vidéo"
#define CATEGORY_VIDEO_INFO_0_FR "Modifier les réglages du rapport d'aspect, du recadrage de l'affichage, du filtre vidéo et du saut d'images."
#define CATEGORY_PGXP_LABEL_FR "PGXP (Pipeline de transformation géométrique précise)"
#define CATEGORY_PGXP_INFO_0_FR "Ces options contrôlent des améliorations qui peuvent améliorer les graphismes par rapport à la console originale. PGXP peut se débarrasser des déformations de textures et des problèmes de Z-fighting (chevauchement de textures)."
#define CATEGORY_OSD_LABEL_FR "Affichage à l'écran"
#define CATEGORY_OSD_INFO_0_FR "Modifier les notifications affichées à l'écran."
#define CATEGORY_INPUT_LABEL_FR "Entrées"
#define CATEGORY_INPUT_INFO_0_FR "Modifier les réglages du pistolet, de la souris et du neGcon."
#define CATEGORY_MEMCARDS_LABEL_FR "Memory Card (sauvegarde)"
#define CATEGORY_MEMCARDS_INFO_0_FR "Modifier les réglages liés à la carte mémoire virtuelle."
#define CATEGORY_HACKS_LABEL_FR "Hacks d'émulation"
#define CATEGORY_HACKS_INFO_0_FR "Modifier les réglages de la précision d'émulation qui affectent les performances de bas niveau et la compatibilité."
#define BEETLE_OPT_CPU_FREQ_SCALE_LABEL_FR "Échelle de la fréquence du processeur (overclock)"
#define BEETLE_OPT_CPU_FREQ_SCALE_INFO_0_FR "Overclocker (ou underclocker) le processeur émulé de la PSX. L'overclocking peut éliminer le ralentissement et améliorer la cadence d'images dans certains jeux au détriment des exigences de performances accrues. Notez que certains jeux ont un limiteur de fréquence d'images interne et peuvent ne pas bénéficier de l'overclocking. Peut causer l'animation plus rapide que prévu de certains effets dans certains titres lors de l'overclocking."
#define OPTION_VAL_50_FR NULL
#define OPTION_VAL_60_FR NULL
#define OPTION_VAL_70_FR NULL
#define OPTION_VAL_80_FR NULL
#define OPTION_VAL_90_FR NULL
#define OPTION_VAL_100_FR "100 % (native)"
#define OPTION_VAL_110_FR NULL
#define OPTION_VAL_120_FR NULL
#define OPTION_VAL_130_FR NULL
#define OPTION_VAL_140_FR NULL
#define OPTION_VAL_150_FR NULL
#define OPTION_VAL_160_FR NULL
#define OPTION_VAL_170_FR NULL
#define OPTION_VAL_180_FR NULL
#define OPTION_VAL_190_FR NULL
#define OPTION_VAL_200_FR NULL
#define OPTION_VAL_210_FR NULL
#define OPTION_VAL_220_FR NULL
#define OPTION_VAL_230_FR NULL
#define OPTION_VAL_240_FR NULL
#define OPTION_VAL_250_FR NULL
#define OPTION_VAL_260_FR NULL
#define OPTION_VAL_270_FR NULL
#define OPTION_VAL_280_FR NULL
#define OPTION_VAL_290_FR NULL
#define OPTION_VAL_300_FR NULL
#define OPTION_VAL_310_FR NULL
#define OPTION_VAL_320_FR NULL
#define OPTION_VAL_330_FR NULL
#define OPTION_VAL_340_FR NULL
#define OPTION_VAL_350_FR NULL
#define OPTION_VAL_360_FR NULL
#define OPTION_VAL_370_FR NULL
#define OPTION_VAL_380_FR NULL
#define OPTION_VAL_390_FR NULL
#define OPTION_VAL_400_FR NULL
#define OPTION_VAL_410_FR NULL
#define OPTION_VAL_420_FR NULL
#define OPTION_VAL_430_FR NULL
#define OPTION_VAL_440_FR NULL
#define OPTION_VAL_450_FR NULL
#define OPTION_VAL_460_FR NULL
#define OPTION_VAL_470_FR NULL
#define OPTION_VAL_480_FR NULL
#define OPTION_VAL_490_FR NULL
#define OPTION_VAL_500_FR NULL
#define OPTION_VAL_510_FR NULL
#define OPTION_VAL_520_FR NULL
#define OPTION_VAL_530_FR NULL
#define OPTION_VAL_540_FR NULL
#define OPTION_VAL_550_FR NULL
#define OPTION_VAL_560_FR NULL
#define OPTION_VAL_570_FR NULL
#define OPTION_VAL_580_FR NULL
#define OPTION_VAL_590_FR NULL
#define OPTION_VAL_600_FR NULL
#define OPTION_VAL_610_FR NULL
#define OPTION_VAL_620_FR NULL
#define OPTION_VAL_630_FR NULL
#define OPTION_VAL_640_FR NULL
#define OPTION_VAL_650_FR NULL
#define OPTION_VAL_660_FR NULL
#define OPTION_VAL_670_FR NULL
#define OPTION_VAL_680_FR NULL
#define OPTION_VAL_690_FR NULL
#define OPTION_VAL_700_FR NULL
#define OPTION_VAL_710_FR NULL
#define OPTION_VAL_720_FR NULL
#define OPTION_VAL_730_FR NULL
#define OPTION_VAL_740_FR NULL
#define OPTION_VAL_750_FR NULL
#define BEETLE_OPT_GPU_OVERCLOCK_LABEL_FR "Overclocking du rastériseur processeur graphique"
#define BEETLE_OPT_GPU_OVERCLOCK_INFO_0_FR "Active l'overclocking du rastériseur 2D contenu dans le processeur graphique de la PSX émulée. N'améliore pas le rendu 3D, et en général n'a que peu d'effet."
#define OPTION_VAL_1X_NATIVE_FR "x1 (native)"
#define OPTION_VAL_2X_FR "x2"
#define OPTION_VAL_4X_FR "x4"
#define OPTION_VAL_8X_FR "x8"
#define OPTION_VAL_16X_FR "x16"
#define OPTION_VAL_32X_FR "x32"
#define BEETLE_OPT_GTE_OVERCLOCK_LABEL_FR "Overclocker le GTE"
#define BEETLE_OPT_GTE_OVERCLOCK_INFO_0_FR "Réduire toutes les opérations du GTE émulé (coprocesseur du CPU pour les graphismes 3D) à une latence constante d'un cycle. Pour les jeux qui font un usage intensif du GTE, cela peut grandement améliorer la fréquence d'images et la stabilité du temps d'images."
#define BEETLE_OPT_CD_ACCESS_METHOD_LABEL_FR "Méthode d'accès au CD"
#define BEETLE_OPT_CD_ACCESS_METHOD_INFO_0_FR "'Synchrone' imite le matériel original. 'Asynchrone' peut réduire les saccades sur les appareils avec un stockage lent. 'Pré-cachée' charge la totalité de l'image disque en mémoire lors du lancement de contenu, ce qui peut améliorer les temps de chargement en jeu au prix d'un délai initial au démarrage. 'Pré-cachée' peut causer des problèmes sur les systèmes avec peu de mémoire vive (RAM), et se repliera sur 'Synchrone' pour les médias physiques. Redémarrage requis."
#define OPTION_VAL_SYNC_FR "Synchrone"
#define OPTION_VAL_ASYNC_FR "Asynchrone"
#define OPTION_VAL_PRECACHE_FR "Pré-cachée"
#define BEETLE_OPT_CD_FASTLOAD_LABEL_FR "Vitesse de chargement du CD"
#define BEETLE_OPT_CD_FASTLOAD_INFO_0_FR "Les valeurs supérieures à 'x2 (native)' peuvent considérablement réduire le temps de chargement en jeu, mais peuvent introduire des erreurs de timing. Certains jeux peuvent ne pas fonctionner correctement au-dessus d'une certaine valeur."
#define OPTION_VAL_2X_NATIVE_FR "x2 (native)"
#define OPTION_VAL_6X_FR "x6"
#define OPTION_VAL_10X_FR "x10"
#define OPTION_VAL_12X_FR "x12"
#define OPTION_VAL_14X_FR "x14"
#define BEETLE_OPT_SKIP_BIOS_LABEL_FR "Ignorer le BIOS"
#define BEETLE_OPT_SKIP_BIOS_INFO_0_FR "Ignorer l'animation de démarrage du BIOS PlayStation normalement affichée lors du chargement de contenu. Remarque : l'activation de cette option pose des problèmes de compatibilité avec un certain nombre de jeux (jeux protégés contre la copie PAL, Saga Frontier, etc.)."
#define BEETLE_OPT_OVERRIDE_BIOS_LABEL_FR "Remplacer le BIOS"
#define BEETLE_OPT_OVERRIDE_BIOS_INFO_0_FR "Remplacer le BIOS spécifique à la région standard par un BIOS dézoné s'il est trouvé. Redémarrage requis."
#define OPTION_VAL_PSXONPSP_FR "BIOS PS1 de la PSP"
#define OPTION_VAL_PS1_ROM_FR "BIOS PS1 de la PS3"
#define OPTION_VAL_OPENBIOS_FR NULL
#define BEETLE_OPT_REGION_LABEL_FR "Région du système"
#define BEETLE_OPT_REGION_INFO_0_FR "Définit la région par défaut utilisée lorsque le contenu ne peut pas être détecté automatiquement (les fichiers PS-X EXE bruts par exemple). Les jeux sur disque utilisent toujours la région intégrée sur le disque. Redémarrage requis."
#define OPTION_VAL_AUTO_FR NULL
#define OPTION_VAL_NTSC_J_FR "NTSC-J (Japon)"
#define OPTION_VAL_NTSC_U_FR "NTSC-U (Amérique du Nord)"
#define OPTION_VAL_PAL_FR NULL
#define BEETLE_OPT_INTERNAL_RESOLUTION_LABEL_FR "Résolution du processeur graphique interne"
#define BEETLE_OPT_INTERNAL_RESOLUTION_INFO_0_FR "Les résolutions supérieures à 'x1 (native)' améliorent la fidélité des modèles 3D au détriment des exigences de performance accrues. Les éléments 2D ne sont généralement pas affectés par ce réglage."
#define BEETLE_OPT_RENDERER_LABEL_FR "Moteur de rendu"
#define BEETLE_OPT_RENDERER_INFO_0_FR "Le moteur de rendu logiciel est le plus précis mais a des exigences en performances élevées lorsqu'il fonctionne avec des résolutions de carte graphique interne accrues. Les moteurs de rendu matériels, bien que moins précis, améliorent les performances par rapport au rendu logiciel à des résolutions internes accrues et permettent diverses améliorations graphiques. 'Matériel (auto)' sélectionne automatiquement le moteur de rendu Vulkan ou OpenGL, selon le pilote vidéo de l'interface libretro actuel. Si le pilote vidéo fourni n'est pas compatible avec Vulkan ou OpenGL 3.3, alors le cœur se repliera sur le moteur de rendu logiciel. Redémarrage requis."
#define OPTION_VAL_HARDWARE_FR "Matériel (auto)"
#define OPTION_VAL_HARDWARE_GL_FR "Matériel (OpenGL)"
#define OPTION_VAL_HARDWARE_VK_FR "Matériel (Vulkan)"
#define OPTION_VAL_SOFTWARE_FR "Logiciel"
#define BEETLE_OPT_RENDERER_SOFTWARE_FB_LABEL_FR "Tampon d'images logiciel"
#define BEETLE_OPT_RENDERER_SOFTWARE_FB_INFO_0_FR "Activer l'émulation précise des effets de tampon d'image (flou de mouvement, tourbillon de combat dans FF7 par exemple) lors de l'utilisation des moteurs de rendu matériels en exécutant une copie du rendu logiciel à la résolution native en arrière-plan. Si cette option est désactivée, ces opérations sont omises (OpenGL) ou rendues par le processeur graphique (Vulkan). La désactivation peut améliorer les performances mais peut causer des erreurs graphiques graves. Laisser activé si vous n'êtes pas sûr."
#define BEETLE_OPT_DEPTH_LABEL_FR "Profondeur de couleurs interne"
#define BEETLE_OPT_DEPTH_INFO_0_FR "Une profondeur de couleurs plus élevée peut réduire les effets de bande de couleurs sans utiliser de tramage. 16 bpp émule le matériel original mais peut avoir un bandage visible si le tramage n'est pas activé. 'Motif de tramage (dithering)' est recommandé d'être désactivé lorsque cette option est réglée sur 32 bpp."
#define OPTION_VAL_16BPP_NATIVE_FR "16 bpp (native)"
#define OPTION_VAL_32BPP_FR NULL
#define BEETLE_OPT_DITHER_MODE_LABEL_FR "Motif de tramage (dithering)"
#define BEETLE_OPT_DITHER_MODE_INFO_0_FR "'x1 (native)' émule le tramage natif de basse résolution utilisé par le matériel original pour lisser les artefacts de bandes de couleurs visibles à la profondeur de couleurs native. La 'Résolution interne' met à l'échelle la granularité du tramage sur la résolution interne configurée pour des résultats plus propres. Devrait être désactivé lors de l'exécution à une profondeur de couleurs sur 32 bpp. Remarque : Avec Vulkan, activer ceci forcera l'échantillonnage à la profondeur de couleur native, tandis que la désactivation activera automatiquement la sortie à une plus grande profondeur de couleurs."
#define OPTION_VAL_INTERNAL_RESOLUTION_FR "Résolution interne"
#define BEETLE_OPT_SCALED_UV_OFFSET_LABEL_FR "Décalage UV de la texture"
#define BEETLE_OPT_SCALED_UV_OFFSET_INFO_0_FR "Échantillonner les textures des polygones 3D avec un décalage pour les résolutions internes supérieures à x1. Réduit les discontinuités de textures mais peut causer des bugs graphiques involontaires."
#define BEETLE_OPT_FILTER_LABEL_FR "Filtrage des textures"
#define BEETLE_OPT_FILTER_INFO_0_FR "'Au plus proche' émule le matériel original. 'Bilinéaire' et 'À 3 points' sont des filtres de lissage, qui réduisent la pixelisation via le flou. Les filtres 'SABR', 'xBR' et 'JINC2' sont des filtres d'agrandissement qui peuvent améliorer la fidélité/netteté des textures au détriment des exigences de performances accrues. Uniquement pris en charge par les moteurs de rendu matériels."
#define OPTION_VAL_NEAREST_FR "Au plus proche"
#define OPTION_VAL_SABR_FR NULL
#define OPTION_VAL_XBR_FR NULL
#define OPTION_VAL_BILINEAR_FR "Bilinéaire"
#define OPTION_VAL_3_POINT_FR "À 3 points"
#define OPTION_VAL_JINC2_FR NULL
#define BEETLE_OPT_FILTER_EXCLUDE_SPRITE_LABEL_FR "Exclure les sprites du filtrage"
#define BEETLE_OPT_FILTER_EXCLUDE_SPRITE_INFO_0_FR "Ne pas appliquer le filtrage de textures aux sprites. Empêche les discontinuités dans divers jeux avec des fonds rendus sur des sprites 2D. Utiliser avec le lissage adaptatif ou un autre filtre de post-traitement pour un meilleur effet."
#define OPTION_VAL_OPAQUE_FR "Opaque uniquement"
#define OPTION_VAL_ALL_FR "Opaque et semi-transparent"
#define BEETLE_OPT_FILTER_EXCLUDE_2D_POLYGON_LABEL_FR "Exclure les polygones 2D du filtrage"
#define BEETLE_OPT_FILTER_EXCLUDE_2D_POLYGON_INFO_0_FR "Ne pas appliquer le filtrage de textures aux polygones 2D. Les polygones 2D sont détectés avec une heuristique et il peut y avoir des bugs. Utiliser avec le lissage adaptatif ou un autre filtre de post-traitement pour un meilleur effet."
#define BEETLE_OPT_ADAPTIVE_SMOOTHING_LABEL_FR "Lissage adaptatif"
#define BEETLE_OPT_ADAPTIVE_SMOOTHING_INFO_0_FR "Lisser les images 2D et les éléments de l'interface utilisateur sans flouter les objets rendus en 3D. Uniquement pris en charge par le moteur de rendu Vulkan."
#define BEETLE_OPT_SUPER_SAMPLING_LABEL_FR "Super échantillonnage (sous-échantillonnage à la résolution native)"
#define BEETLE_OPT_SUPER_SAMPLING_INFO_0_FR "Sous-échantillonne le rendu depuis la résolution interne supérieure à l'échelle de la résolution native. La combinaison de ceci avec des multiplicateurs de résolution interne plus élevés permet d'afficher des jeux avec des objets 3D anti-crénelés en basse résolution native. Produit les meilleurs résultats lorsqu'elle est appliquée à des titres mélangeant des éléments 2D et 3D (par ex. les personnages en 3D sur des arrière-plans pré-rendus), et fonctionne bien en conjonction avec les shaders CRT. Uniquement pris en charge par le moteur de rendu Vulkan. Remarque : désactiver 'Motif de tramage' est recommandé lorsque cette option est activée."
#define BEETLE_OPT_MSAA_LABEL_FR "Anti-crénelage multi-échantillonné"
#define BEETLE_OPT_MSAA_INFO_0_FR "MSAA améliore l'apparence des objets 3D. Uniquement pris en charge par le moteur de rendu Vulkan."
#define OPTION_VAL_1X_FR "x1"
#define BEETLE_OPT_MDEC_YUV_LABEL_FR "Filtre MDEC YUV chroma"
#define BEETLE_OPT_MDEC_YUV_INFO_0_FR "Améliore la qualité de la lecture des vidéos (FMV) en réduisant les artefacts de 'macroblocs' (carrés/bords dentelés). Uniquement pris en charge par le moteur de rendu Vulkan."
#define BEETLE_OPT_TRACK_TEXTURES_LABEL_FR "Suivre les textures"
#define BEETLE_OPT_TRACK_TEXTURES_INFO_0_FR "Pré-requis pour l'importation et le remplacement de textures. Plantera probablement dans la plupart des jeux."
#define BEETLE_OPT_DUMP_TEXTURES_LABEL_FR "Importer les textures"
#define BEETLE_OPT_DUMP_TEXTURES_INFO_0_FR "Importer les textures utilisées vers <cd>-texture-dump/"
#define BEETLE_OPT_REPLACE_TEXTURES_LABEL_FR "Remplacer les textures"
#define BEETLE_OPT_REPLACE_TEXTURES_INFO_0_FR "Remplacer les textures par des versions HD depuis <cd>-texture-replacements/"
#define BEETLE_OPT_PGXP_MODE_LABEL_FR "Mode d'opération de PGXP"
#define BEETLE_OPT_PGXP_MODE_INFO_0_FR "Rendu des objets 3D avec une précision aux sous-pixels, minimisant la distorsion et les variations des objets 3D vues sur le matériel d'origine en raison de l'utilisation de coordonnées de vertex sur un point fixe. Le mode 'Mémoire uniquement' a un minimum de problèmes de compatibilité et est recommandé pour une utilisation générale. Le mode 'Mémoire + processeur (buggé)' peut réduire encore plus les variations, mais il a des exigences de performances élevées et peut causer diverses erreurs de géométrie."
#define OPTION_VAL_MEMORY_ONLY_FR "Mémoire uniquement"
#define OPTION_VAL_MEMORY_CPU_FR "Mémoire + processeur (buggé)"
#define BEETLE_OPT_PGXP_2D_TOL_LABEL_FR "Tolérance à la géométrie 2D de PGXP"
#define BEETLE_OPT_PGXP_2D_TOL_INFO_0_FR "Masquer les erreurs les plus flagrantes dans les opérations de PGXP : la valeur spécifie la tolérance avec laquelle les valeurs de PGXP seront conservées en cas de géométries sans informations de profondeur appropriées."
#define OPTION_VAL_0PX_FR "0 px"
#define OPTION_VAL_1PX_FR "1 px"
#define OPTION_VAL_2PX_FR "2 px"
#define OPTION_VAL_3PX_FR "3 px"
#define OPTION_VAL_4PX_FR "4 px"
#define OPTION_VAL_5PX_FR "5 px"
#define OPTION_VAL_6PX_FR "6 px"
#define OPTION_VAL_7PX_FR "7 px"
#define OPTION_VAL_8PX_FR "8 px"
#define BEETLE_OPT_PGXP_NCLIP_LABEL_FR "Élimination primitive des faces cachées PGXP"
#define BEETLE_OPT_PGXP_NCLIP_INFO_0_FR "Utiliser l'implémentation NCLIP de PGXP. Améliore l'apparence en réduisant les trous dans les géométries avec les coordonnées de PGXP. Connu pour causer des bloquages de certains jeux dans diverses circonstances."
#define BEETLE_OPT_PGXP_VERTEX_LABEL_FR "Cache des vertex de PGXP"
#define BEETLE_OPT_PGXP_VERTEX_INFO_0_FR "Mettre en cache la position des vertex améliorée par PGXP pour une réutilisation entre les rendus de polygones. Peut potentiellement améliorer l'alignement des objets et réduire les discontinuités visibles lors du rendu des textures, mais les faux positifs lors des requêtes du cache peuvent produire des bugs graphiques. Il est actuellement recommandé de laisser cette option désactivée. Cette option n'est appliquée que lorsque le Mode d'opération de PGXP est activé. Uniquement pris en charge par les moteurs de rendu matériels."
#define BEETLE_OPT_PGXP_TEXTURE_LABEL_FR "Textures correctes en perspective de PGXP"
#define BEETLE_OPT_PGXP_TEXTURE_INFO_0_FR "Remplacer le mappage de textures affiné natif de la PSX par le mappage de texture correct pour la perspective. Élimine la distorsion et la déformation dépendant de la position des textures, ce qui entraîne un alignement correct des textures. Cette option n'est appliquée que lorsque le Mode d'opération de PGXP est activé. Pris en charge par les moteurs de rendu matériels et le moteur de rendu logiciel."
#define BEETLE_OPT_DISPLAY_INTERNAL_FPS_LABEL_FR "Afficher les i/s internes"
#define BEETLE_OPT_DISPLAY_INTERNAL_FPS_INFO_0_FR "Affiche la fréquence d'images interne à laquelle le système PlayStation émulé rend du contenu. Remarque : Nécessite que les notifications à l'écran soient activées dans l'interface de libretro."
#define BEETLE_OPT_DISPLAY_OSD_LABEL_FR "Afficher les messages à l'écran"
#define BEETLE_OPT_DISPLAY_OSD_INFO_0_FR "Affiche les messages texte à l'écran générés par le cœur."
#define BEETLE_OPT_DISPLAY_VRAM_LABEL_FR "Afficher la VRAM complète (diagnostic)"
#define BEETLE_OPT_DISPLAY_VRAM_INFO_0_FR "Visualiser entièrement la VRAM de la console émulée. Uniquement pris en charge par les moteurs de rendu matériels OpenGL et Vulkan. Remarque : Ceci est à des fins de diagnostic, et devrait normalement être désactivé."
#define BEETLE_OPT_ANALOG_CALIBRATION_LABEL_FR "Auto-étalonnage analogique"
#define BEETLE_OPT_ANALOG_CALIBRATION_INFO_0_FR "Active l'étalonnage dynamique des entrées analogiques lorsque le périphérique d'entrée est réglé sur DualShock, Manette analogique, Joystick analogique ou neGcon. Les valeurs maximales enregistrées sont surveillées en temps réel et utilisées pour mettre à l'échelle les coordonnées analogiques passées à l'émulateur. Cela devrait être utilisé pour des jeux tels que Mega Man Legends 2 qui s'attendent à des valeurs plus grandes que ce que les manettes modernes fournissent. Pour de meilleurs résultats, les sticks analogiques devraient être tournés dans toute la mesure du possible pour régler l'algorithme de calibration chaque fois que du contenu est chargé."
#define BEETLE_OPT_ANALOG_TOGGLE_LABEL_FR "Activer/désactiver le mode analogique DualShock"
#define BEETLE_OPT_ANALOG_TOGGLE_INFO_0_FR "Lorsque le type de périphérique d'entrée est DualShock, cette option permet au DualShock émulé d'être basculé entre le mode NUMÉRIQUE et le mode ANALOGIQUE comme le matériel original. Le mode peut également être basculé en utilisant la combinaison de touches sélectionnée."
#define OPTION_VAL_ENABLED_ANALOG_FR "Analogique par défaut"
#define BEETLE_OPT_ANALOG_TOGGLE_COMBO_LABEL_FR "Combinaison de touches du mode analogique DualShock"
#define BEETLE_OPT_ANALOG_TOGGLE_COMBO_INFO_0_FR "Combinaison de touches qui sera utilisée pour basculer entre le mode DIGITAL et le mode ANALOG pour le DualShock émulé."
#define OPTION_VAL_L1_L2_R1_R2_START_SELECT_FR NULL
#define OPTION_VAL_L1_R1_SELECT_FR NULL
#define OPTION_VAL_L1_R1_START_FR NULL
#define OPTION_VAL_L1_R1_L3_FR NULL
#define OPTION_VAL_L1_R1_R3_FR NULL
#define OPTION_VAL_L2_R2_SELECT_FR NULL
#define OPTION_VAL_L2_R2_START_FR NULL
#define OPTION_VAL_L2_R2_L3_FR NULL
#define OPTION_VAL_L2_R2_R3_FR NULL
#define OPTION_VAL_L3_R3_FR NULL
#define BEETLE_OPT_ANALOG_TOGGLE_HOLD_LABEL_FR "Délai de maintien de la combinaison de touches du mode analogique DualShock"
#define BEETLE_OPT_ANALOG_TOGGLE_HOLD_INFO_0_FR "Temps de maintien de la combinaison de touches pour le mode analogique."
#define OPTION_VAL_0_FR "0 seconde"
#define OPTION_VAL_1_FR "1 seconde"
#define OPTION_VAL_2_FR "2 secondes"
#define OPTION_VAL_3_FR "3 secondes"
#define OPTION_VAL_4_FR "4 secondes"
#define OPTION_VAL_5_FR "5 secondes"
#define BEETLE_OPT_ENABLE_MULTITAP_PORT1_LABEL_FR "Multitap sur le port 1"
#define BEETLE_OPT_ENABLE_MULTITAP_PORT1_INFO_0_FR "Activer la fonctionnalité multitap sur le port 1."
#define BEETLE_OPT_ENABLE_MULTITAP_PORT2_LABEL_FR "Multitap sur le port 2"
#define BEETLE_OPT_ENABLE_MULTITAP_PORT2_INFO_0_FR "Activer la fonctionnalité multitap sur le port 2."
#define BEETLE_OPT_GUN_INPUT_MODE_LABEL_FR "Mode d'entrée du pistolet"
#define BEETLE_OPT_GUN_INPUT_MODE_INFO_0_FR "Utiliser une entrée 'Pistolet' contrôlée par la souris ou 'Écran tactile' lorsque le type de périphérique est réglé sur 'Guncon/G-Con 45' ou 'Justifier'."
#define OPTION_VAL_LIGHTGUN_FR "Pistolet"
#define OPTION_VAL_TOUCHSCREEN_FR "Écran tactile"
#define BEETLE_OPT_GUN_CURSOR_LABEL_FR "Viseur du pistolet"
#define BEETLE_OPT_GUN_CURSOR_INFO_0_FR "Dessine un curseur à l'écran lors de l'utilisation des types de périphériques d'entrée 'Guncon/G-Con 45' et 'Justifier'."
#define OPTION_VAL_CROSS_FR "Croix"
#define OPTION_VAL_DOT_FR "Point"
#define BEETLE_OPT_CROSSHAIR_COLOR_P1_LABEL_FR "Couleur de viseur pistolet : Port 1"
#define BEETLE_OPT_CROSSHAIR_COLOR_P1_INFO_0_FR "Couleur de viseur du pistolet pour le port 1."
#define OPTION_VAL_RED_FR "Rouge"
#define OPTION_VAL_BLUE_FR "Bleue"
#define OPTION_VAL_GREEN_FR "Verte"
#define OPTION_VAL_ORANGE_FR NULL
#define OPTION_VAL_YELLOW_FR "Jaune"
#define OPTION_VAL_CYAN_FR NULL
#define OPTION_VAL_PINK_FR "Rose"
#define OPTION_VAL_PURPLE_FR "Violette"
#define OPTION_VAL_BLACK_FR "Noire"
#define OPTION_VAL_WHITE_FR "Blanche"
#define BEETLE_OPT_CROSSHAIR_COLOR_P2_LABEL_FR "Couleur de viseur pistolet : Port 2"
#define BEETLE_OPT_CROSSHAIR_COLOR_P2_INFO_0_FR "Couleur de viseur du pistolet pour le port 2."
#define BEETLE_OPT_MOUSE_SENSITIVITY_LABEL_FR "Sensibilité de la souris"
#define BEETLE_OPT_MOUSE_SENSITIVITY_INFO_0_FR "Réactivité du type de périphérique d'entrée 'Souris'."
#define OPTION_VAL_5_O42_FR "5 %"
#define OPTION_VAL_10_FR NULL
#define OPTION_VAL_15_FR NULL
#define OPTION_VAL_20_FR NULL
#define OPTION_VAL_25_FR NULL
#define OPTION_VAL_30_FR NULL
#define OPTION_VAL_35_FR NULL
#define OPTION_VAL_40_FR NULL
#define OPTION_VAL_45_FR NULL
#define OPTION_VAL_55_FR NULL
#define OPTION_VAL_65_FR NULL
#define OPTION_VAL_75_FR NULL
#define OPTION_VAL_85_FR NULL
#define OPTION_VAL_95_FR NULL
#define OPTION_VAL_100_O42_FR NULL
#define OPTION_VAL_105_FR NULL
#define OPTION_VAL_115_FR NULL
#define OPTION_VAL_125_FR NULL
#define OPTION_VAL_135_FR NULL
#define OPTION_VAL_145_FR NULL
#define OPTION_VAL_155_FR NULL
#define OPTION_VAL_165_FR NULL
#define OPTION_VAL_175_FR NULL
#define OPTION_VAL_185_FR NULL
#define OPTION_VAL_195_FR NULL
#define BEETLE_OPT_NEGCON_RESPONSE_LABEL_FR "Réponse de torsion du neGcon"
#define BEETLE_OPT_NEGCON_RESPONSE_INFO_0_FR "Type de réponse du stick analogique de gauche lors de la simulation de l'action de 'torsion' des périphériques d'entrée 'neGcon' émulés. Le déplacement des sticks analogiques peut être mappé à un angle de rotation neGcon linéaire, quadratique ou cubique. 'Quadratique' permet une plus grande précision que 'Linéaire' lors de petits déplacements. 'Cubique' augmente davantage la précision des petits déplacements, mais 'exagère' les mouvements plus grands. Remarque : 'Linéaire' n'est recommandé que lorsque vous utilisez des périphériques de volants de course. Les manettes conventionnelles implémentent l'entrée analogique d'une manière fondamentalement différente du mécanisme de 'torsion' du neGcon, de telle sorte que le mappage linéaire amplifie trop les petits mouvements, entravant le contrôle précis. Dans la plupart des cas, 'Quadratique' fournit l'approximation la plus proche du matériel réel."
#define OPTION_VAL_LINEAR_FR "Linéaire"
#define OPTION_VAL_QUADRATIC_FR "Quadratique"
#define OPTION_VAL_CUBIC_FR "Cubique"
#define BEETLE_OPT_NEGCON_DEADZONE_LABEL_FR "Deadzone de torsion du negCon"
#define BEETLE_OPT_NEGCON_DEADZONE_INFO_0_FR "Deadzone du stick analogique gauche lors de la simulation de l'action 'torsion' des périphériques d'entrée 'neGcon' émulés. Utilisé pour éliminer la dérive de la manette. Remarque : la plupart des titres compatibles avec le neGcon fournissent des options en jeu pour définir une valeur de deadzone de la 'torsion'. Pour éviter la perte de précision, la deadzone réglée en jeu devrait *toujours* être réglée à zéro. Tous les ajustements requis devraient *uniquement* être appliqués via cette option de coeur. Ceci est particulièrement important lorsque la valeur de 'Réponse de torsion du neGcon' est définie sur 'Quadratique' ou 'Cubique'."
#define OPTION_VAL_0_O44_FR "0 %"
#define BEETLE_OPT_USE_MEDNAFEN_MEMCARD0_METHOD_LABEL_FR "Méthode de carte mémoire"
#define BEETLE_OPT_USE_MEDNAFEN_MEMCARD0_METHOD_INFO_0_FR "Format des données de sauvegarde utilisé pour l'emplacement carte mémoire de gauche. 'Mednafen' peut être utilisé pour la compatibilité avec la version autonome de Mednafen. Lors de l'utilisation avec Beetle PSX, Libretro (.srm) et Mednafen (.mcr) les sauvegardes ont des formats internes identiques et peuvent être converties les unes vers les autres via le renommage. Redémarrage requis."
#define OPTION_VAL_LIBRETRO_FR NULL
#define OPTION_VAL_MEDNAFEN_FR NULL
#define BEETLE_OPT_ENABLE_MEMCARD1_LABEL_FR "Emplacement de carte mémoire 2"
#define BEETLE_OPT_ENABLE_MEMCARD1_INFO_0_FR "Émuler une deuxième carte mémoire dans l'emplacement carte mémoire de droite. Remarque : certains jeux nécessitent que cette option soit désactivée pour un fonctionnement correct (par exemple, Codename Tenka). Remarque : l'emplacement de carte mémoire 2 utilise le format de sauvegarde Mednafen (.mcr)."
#define BEETLE_OPT_SHARED_MEMORY_CARDS_LABEL_FR "Cartes mémoire partagées"
#define BEETLE_OPT_SHARED_MEMORY_CARDS_INFO_0_FR "Utiliser des cartes mémoire partagées au lieu de cartes séparées pour chaque contenu. Remarque : si la 'Méthode de carte mémoire' est définie sur 'Libretro', seul l'emplacement de carte mémoire 2 sera affecté. Redémarrage requis."
#define BEETLE_OPT_MEMCARD_LEFT_INDEX_LABEL_FR "Index de l'emplacement de carte mémoire 1"
#define BEETLE_OPT_MEMCARD_LEFT_INDEX_INFO_0_FR "Carte mémoire actuellement chargée dans l'emplacement de gauche. Cette option ne fonctionnera que si la méthode de carte mémoire est définie sur Mednafen."
#define OPTION_VAL_0_O48_FR "0 (par défaut)"
#define BEETLE_OPT_MEMCARD_RIGHT_INDEX_LABEL_FR "Index de l'emplacement de carte mémoire 2"
#define BEETLE_OPT_MEMCARD_RIGHT_INDEX_INFO_0_FR "Carte mémoire actuellement chargée dans l'emplacement de droite. Cette option ne fonctionnera que si l'emplacement de carte mémoire 2 est activé."
#define OPTION_VAL_1_O49_FR "1 (par défaut)"
#define BEETLE_OPT_LINE_RENDER_LABEL_FR "Hack des lignes en quadrilatères"
#define BEETLE_OPT_LINE_RENDER_INFO_0_FR "Certains jeux (comme Doom, Hexen, Soul Blade, etc. par exemple) rendent des lignes horizontales en étirant des triangles d'une hauteur d'un pixel sur l'écran, qui sont rastérisés comme une rangée de pixels sur le matériel original. Ce hack détecte ces petits triangles et les convertit en quadrilatères si nécessaire, leur permettant d'être affichés correctement sur les moteurs de rendu matériels et à des résolutions internes agrandies. 'Agressif' est requis pour certains titres (Dark Forces et Duke Nukem par exemple) mais peut autrement introduire des bugs graphiques. Laisser sur 'Par défaut' en cas d'incertitude."
#define OPTION_VAL_DEFAULT_FR "Par défaut"
#define OPTION_VAL_AGGRESSIVE_FR "Agressif"
#define BEETLE_OPT_WIDESCREEN_HACK_LABEL_FR "Hack de mode écran large"
#define BEETLE_OPT_WIDESCREEN_HACK_INFO_0_FR "Rendre le contenu 3D de manière anamorphique et afficher le tampon d'image émulé à un rapport d'aspect d'écran large. Produit de meilleurs résultats avec des jeux entièrement en 3D. Les éléments 2D seront étirés horizontalement et peuvent être mal alignés."
#define BEETLE_OPT_WIDESCREEN_HACK_ASPECT_RATIO_LABEL_FR "Rapport d'aspect du hack de mode écran large"
#define BEETLE_OPT_WIDESCREEN_HACK_ASPECT_RATIO_INFO_0_FR "Rapport d'aspect à utiliser pour le hack de mode écran large."
#define OPTION_VAL_16_9_FR NULL
#define OPTION_VAL_16_10_FR NULL
#define OPTION_VAL_18_9_FR NULL
#define OPTION_VAL_19_9_FR NULL
#define OPTION_VAL_20_9_FR NULL
#define OPTION_VAL_21_9_FR NULL
#define OPTION_VAL_32_9_FR NULL
#define BEETLE_OPT_CPU_DYNAREC_LABEL_FR "Dynarec du processeur"
#define BEETLE_OPT_CPU_DYNAREC_INFO_0_FR "Recompiler dynamiquement les instructions du processeur en mode natif. Bien plus rapide que l'interpréteur, mais la cadence du processeur est moins précise et peut avoir des bugs."
#define OPTION_VAL_DISABLED_FR "Désactivé (interpréteur Beetle)"
#define OPTION_VAL_EXECUTE_FR "Performances maximales"
#define OPTION_VAL_EXECUTE_ONE_FR "Contrôle du temps de cycle"
#define OPTION_VAL_RUN_INTERPRETER_FR "Interpréteur Lightrec"
#define BEETLE_OPT_DYNAREC_INVALIDATE_LABEL_FR "Invalidation du code Dynarec"
#define BEETLE_OPT_DYNAREC_INVALIDATE_INFO_0_FR "Certains jeux requièrent une invalidation 'Complète', certains requièrent 'DMA uniquement'."
#define OPTION_VAL_FULL_FR "Complète"
#define OPTION_VAL_DMA_FR "DMA uniquement (légèrement plus rapide)"
#define BEETLE_OPT_DYNAREC_EVENTCYCLES_LABEL_FR "Cycles d'événements Dynarec DMA/processeur graphique/MDEC/Timer"
#define BEETLE_OPT_DYNAREC_EVENTCYCLES_INFO_0_FR "Cycles maximum exécutés par le processeur avant la vérification d'une mise à jour du processeur graphique/DMA/MDEC/Timer, un nombre plus élevé sera plus rapide, a beaucoup moins d'impact sur l'interpréteur Beetle que sur le Dynarec."
#define OPTION_VAL_128_FR "128 (par défaut)"
#define BEETLE_OPT_DYNAREC_SPU_SAMPLES_LABEL_FR "Échantillons SPU Dynarec"
#define BEETLE_OPT_DYNAREC_SPU_SAMPLES_INFO_0_FR "Nombre maximum d'échantillons SPU à exécuter avant qu'une mise à jour SPU ne soit vérifiée, un nombre plus élevé sera plus rapide, mais causera des problèmes de son dans certains jeux avec une valeur autre que 1."
#define BEETLE_OPT_CORE_TIMING_FPS_LABEL_FR "Cadence des i/s communiquée par le cœur"
#define BEETLE_OPT_CORE_TIMING_FPS_INFO_0_FR "Le basculement automatique permet au cœur de basculer entre les débits progressifs et entrelacés."
#define OPTION_VAL_AUTO_TOGGLE_FR "Basculement automatique"
#define OPTION_VAL_FORCE_PROGRESSIVE_FR "Forcer le débit progressif"
#define OPTION_VAL_FORCE_INTERLACED_FR "Forcer le débit entrelacé"
#define BEETLE_OPT_DEINTERLACER_LABEL_FR "Méthode de désentrelacement"
#define BEETLE_OPT_DEINTERLACER_INFO_0_FR "'Weave' affiche les deux champs sur la même image pour une image fixe en pleine haute résolution, 'Bob' sacrifie la clarté de la résolution pour la clarté du mouvement. 'Bob (décalé)' est Bob avec le champ actuel doublé dans la hauteur de la surface complète et un décalage d'une ligne entre les champs, conservant la résolution de libretro en sortie tout en gardant la clarté du mouvement de style Bob. 'FastMAD (Adaptif au mouvement)' trame des pixels fixes et interpole verticalement les pixels en mouvement pour chaque pixel, à la pleine hauteur de sortie sans l'effet de peigne sur mouvement de Weave ni la résolution diminuée de moitié de Bob. 'Désactivé' fait correspondre le moteur de rendu logiciel au comportement du rendu matériel (Vulkan/OpenGL) en rastérisant sur toutes les lignes VRAM à chaque image et en reportant la fin du balayage à la fin de l'image, en éliminant l'effet de peigne sur mouvement dans les jeux entrelacés. Remarque : les programmes qui font du DMA vers VRAM pendant la fin du balayage (certaines démos utilisant des astuces de réenroulement du tampon d'image PAL >512 lignes), ou qui mettent à jour les sprites en mouvement via des transferts d'image processeur vers VRAM liés au champ affiché (Pro Pinball par exemple), peuvent montrer des lignes obsolètes ou une rémanence des sprites avec l'option 'Désactivé' ; utiliser 'Weave', 'Bob', 'Bob (décalé)' ou 'FastMAD (adaptatif au mouvement)' pour ceux-ci."
#define OPTION_VAL_WEAVE_FR NULL
#define OPTION_VAL_BOB_FR NULL
#define OPTION_VAL_BOB_OFFSET_FR "Bob (décalé)"
#define OPTION_VAL_FASTMAD_FR "FastMAD (adaptatif au mouvement)"
#define BEETLE_OPT_FRAME_DUPING_LABEL_FR "Duplication d'images"
#define BEETLE_OPT_FRAME_DUPING_INFO_0_FR "Répéter l'image précédente si le cœur n'a rien de nouveau à afficher."
#define BEETLE_OPT_PAL_VIDEO_TIMING_OVERRIDE_LABEL_FR "Substitution de la cadence vidéo PAL"
#define BEETLE_OPT_PAL_VIDEO_TIMING_OVERRIDE_INFO_0_FR "En raison de normes différentes, les jeux PAL semblent souvent ralentis par rapport aux versions américaines ou japonaises à la norme NTSC. Cette option peut être utilisée pour remplacer les timings PAL afin d'essayer d'exécuter ces jeux avec le débit d'image NTSC."
#define BEETLE_OPT_ASPECT_RATIO_LABEL_FR "Rapport d'aspect du cœur"
#define BEETLE_OPT_ASPECT_RATIO_INFO_0_FR "Ce réglage est ignoré lorsque les options Hack mode écran large ou Afficher la VRAM complète sont activées."
#define OPTION_VAL_CORRECTED_FR "Corrigé"
#define OPTION_VAL_UNCORRECTED_FR "Non corrigé"
#define OPTION_VAL_4_3_FR "Forcer 4:3"
#define OPTION_VAL_NTSC_FR "Forcer NTSC"
#define BEETLE_OPT_CROP_OVERSCAN_LABEL_FR "Recadrage du surbalayage"
#define BEETLE_OPT_CROP_OVERSCAN_INFO_0_FR "Image native complète ou pas de bordures."
#define OPTION_VAL_STATIC_FR "Horizontale"
#define OPTION_VAL_SMART_FR "Horizontale + verticale"
#define BEETLE_OPT_IMAGE_CROP_LABEL_FR "Recadrage supplémentaire"
#define BEETLE_OPT_IMAGE_CROP_INFO_0_FR "Réduire la largeur de l'image recadrée par le nombre spécifié de pixels."
#define OPTION_VAL_9PX_FR "9 px"
#define OPTION_VAL_10PX_FR "10 px"
#define OPTION_VAL_11PX_FR "11 px"
#define OPTION_VAL_12PX_FR "12 px"
#define OPTION_VAL_13PX_FR "13 px"
#define OPTION_VAL_14PX_FR "14 px"
#define OPTION_VAL_15PX_FR "15 px"
#define OPTION_VAL_16PX_FR "16 px"
#define OPTION_VAL_17PX_FR "17 px"
#define OPTION_VAL_18PX_FR "18 px"
#define OPTION_VAL_19PX_FR "19 px"
#define OPTION_VAL_20PX_FR "20 px"
#define BEETLE_OPT_IMAGE_OFFSET_LABEL_FR "Décalage de l'image recadrée"
#define BEETLE_OPT_IMAGE_OFFSET_INFO_0_FR "Décaler horizontalement vers la droite (positif) ou vers la gauche (négatif) par le nombre spécifié de pixels. Peut être utilisé pour corriger les problèmes d'alignement. Uniquement pris en charge par le moteur de rendu logiciel."
#define OPTION_VAL_12PX_O64_FR "-12 px"
#define OPTION_VAL_11PX_O64_FR "-11 px"
#define OPTION_VAL_10PX_O64_FR "-10 px"
#define OPTION_VAL_9PX_O64_FR "-9 px"
#define OPTION_VAL_8PX_O64_FR "-8 px"
#define OPTION_VAL_7PX_O64_FR "-7 px"
#define OPTION_VAL_6PX_O64_FR "-6 px"
#define OPTION_VAL_5PX_O64_FR "-5 px"
#define OPTION_VAL_4PX_O64_FR "-4 px"
#define OPTION_VAL_3PX_O64_FR "-3 px"
#define OPTION_VAL_2PX_O64_FR "-2 px"
#define OPTION_VAL_1PX_O64_FR "-1 px"
#define OPTION_VAL_1PX_O64_0_FR "+1 px"
#define OPTION_VAL_2PX_O64_0_FR "+2 px"
#define OPTION_VAL_3PX_O64_0_FR "+3 px"
#define OPTION_VAL_4PX_O64_0_FR "+4 px"
#define OPTION_VAL_5PX_O64_0_FR "+5 px"
#define OPTION_VAL_6PX_O64_0_FR "+6 px"
#define OPTION_VAL_7PX_O64_0_FR "+7 px"
#define OPTION_VAL_8PX_O64_0_FR "+8 px"
#define OPTION_VAL_9PX_O64_0_FR "+9 px"
#define OPTION_VAL_10PX_O64_0_FR "+10 px"
#define OPTION_VAL_11PX_O64_0_FR "+11 px"
#define OPTION_VAL_12PX_O64_0_FR "+12 px"
#define BEETLE_OPT_IMAGE_OFFSET_CYCLES_LABEL_FR "Décalage horizontal de l'image (cycles du processeur graphique)"
#define BEETLE_OPT_IMAGE_OFFSET_CYCLES_INFO_0_FR "Les valeurs positives décalent l'image vers la droite, les valeurs négatives décalent l'image vers la gauche. Uniquement pris en charge par les moteurs de rendu matériels."
#define BEETLE_OPT_INITIAL_SCANLINE_LABEL_FR "Ligne de balayage initiale - NTSC"
#define BEETLE_OPT_INITIAL_SCANLINE_INFO_0_FR "Première ligne de balayage affichée lors de l'exécution de contenu NTSC. Les valeurs supérieures à zéro réduiront la hauteur des images en sortie en recadrant les pixels à partir du bord le plus haut. Peut être utilisé pour contrecarrer l'effet de letterboxing."
#define BEETLE_OPT_LAST_SCANLINE_LABEL_FR "Dernière ligne de balayage - NTSC"
#define BEETLE_OPT_LAST_SCANLINE_INFO_0_FR "Dernière ligne de balayage affichée lors de l'exécution de contenu NTSC. Les valeurs inférieures à 239 réduiront la hauteur des images en sortie en recadrant les pixels à partir du bord le plus bas. Peut être utilisé pour contrecarrer l'effet de letterboxing."
#define OPTION_VAL_239_FR "239 (par défaut)"
#define BEETLE_OPT_INITIAL_SCANLINE_PAL_LABEL_FR "Ligne de balayage initiale - PAL"
#define BEETLE_OPT_INITIAL_SCANLINE_PAL_INFO_0_FR "Première ligne de balayage affichée lors de l'exécution de contenu PAL. Les valeurs supérieures à zéro réduiront la hauteur des images en sortie en recadrant les pixels à partir du bord le plus haut. Peut être utilisé pour contrecarrer l'effet de letterboxing."
#define BEETLE_OPT_LAST_SCANLINE_PAL_LABEL_FR "Dernière ligne de balayage - PAL"
#define BEETLE_OPT_LAST_SCANLINE_PAL_INFO_0_FR "Dernière ligne de balayage affichée lors de l'exécution de contenu PAL. Les valeurs inférieures à 287 réduiront la hauteur des images en sortie en recadrant les pixels à partir du bord le plus bas. Peut être utilisé pour contrecarrer l'effet de letterboxing."
#define OPTION_VAL_287_FR "287 (par défaut)"

struct retro_core_option_v2_category option_cats_fr[] = {
   {
      "system",
      CATEGORY_SYSTEM_LABEL_FR,
      CATEGORY_SYSTEM_INFO_0_FR
   },
   {
      "video",
      CATEGORY_VIDEO_LABEL_FR,
      CATEGORY_VIDEO_INFO_0_FR
   },
   {
      "pgxp",
      CATEGORY_PGXP_LABEL_FR,
      CATEGORY_PGXP_INFO_0_FR
   },
   {
      "osd",
      CATEGORY_OSD_LABEL_FR,
      CATEGORY_OSD_INFO_0_FR
   },
   {
      "input",
      CATEGORY_INPUT_LABEL_FR,
      CATEGORY_INPUT_INFO_0_FR
   },
   {
      "memcards",
      CATEGORY_MEMCARDS_LABEL_FR,
      CATEGORY_MEMCARDS_INFO_0_FR
   },
   {
      "hacks",
      CATEGORY_HACKS_LABEL_FR,
      CATEGORY_HACKS_INFO_0_FR
   },
   { NULL, NULL, NULL },
};
struct retro_core_option_v2_definition option_defs_fr[] = {
   {
      BEETLE_OPT(cpu_freq_scale),
      BEETLE_OPT_CPU_FREQ_SCALE_LABEL_FR,
      NULL,
      BEETLE_OPT_CPU_FREQ_SCALE_INFO_0_FR,
      NULL,
      "system",
      {
         { "50%",           OPTION_VAL_50_FR },
         { "60%",           OPTION_VAL_60_FR },
         { "70%",           OPTION_VAL_70_FR },
         { "80%",           OPTION_VAL_80_FR },
         { "90%",           OPTION_VAL_90_FR },
         { "100%",          OPTION_VAL_100_FR },
         { "110%",          OPTION_VAL_110_FR },
         { "120%",          OPTION_VAL_120_FR },
         { "130%",          OPTION_VAL_130_FR },
         { "140%",          OPTION_VAL_140_FR },
         { "150%",          OPTION_VAL_150_FR },
         { "160%",          OPTION_VAL_160_FR },
         { "170%",          OPTION_VAL_170_FR },
         { "180%",          OPTION_VAL_180_FR },
         { "190%",          OPTION_VAL_190_FR },
         { "200%",          OPTION_VAL_200_FR },
         { "210%",          OPTION_VAL_210_FR },
         { "220%",          OPTION_VAL_220_FR },
         { "230%",          OPTION_VAL_230_FR },
         { "240%",          OPTION_VAL_240_FR },
         { "250%",          OPTION_VAL_250_FR },
         { "260%",          OPTION_VAL_260_FR },
         { "270%",          OPTION_VAL_270_FR },
         { "280%",          OPTION_VAL_280_FR },
         { "290%",          OPTION_VAL_290_FR },
         { "300%",          OPTION_VAL_300_FR },
         { "310%",          OPTION_VAL_310_FR },
         { "320%",          OPTION_VAL_320_FR },
         { "330%",          OPTION_VAL_330_FR },
         { "340%",          OPTION_VAL_340_FR },
         { "350%",          OPTION_VAL_350_FR },
         { "360%",          OPTION_VAL_360_FR },
         { "370%",          OPTION_VAL_370_FR },
         { "380%",          OPTION_VAL_380_FR },
         { "390%",          OPTION_VAL_390_FR },
         { "400%",          OPTION_VAL_400_FR },
         { "410%",          OPTION_VAL_410_FR },
         { "420%",          OPTION_VAL_420_FR },
         { "430%",          OPTION_VAL_430_FR },
         { "440%",          OPTION_VAL_440_FR },
         { "450%",          OPTION_VAL_450_FR },
         { "460%",          OPTION_VAL_460_FR },
         { "470%",          OPTION_VAL_470_FR },
         { "480%",          OPTION_VAL_480_FR },
         { "490%",          OPTION_VAL_490_FR },
         { "500%",          OPTION_VAL_500_FR },
         { "510%",          OPTION_VAL_510_FR },
         { "520%",          OPTION_VAL_520_FR },
         { "530%",          OPTION_VAL_530_FR },
         { "540%",          OPTION_VAL_540_FR },
         { "550%",          OPTION_VAL_550_FR },
         { "560%",          OPTION_VAL_560_FR },
         { "570%",          OPTION_VAL_570_FR },
         { "580%",          OPTION_VAL_580_FR },
         { "590%",          OPTION_VAL_590_FR },
         { "600%",          OPTION_VAL_600_FR },
         { "610%",          OPTION_VAL_610_FR },
         { "620%",          OPTION_VAL_620_FR },
         { "630%",          OPTION_VAL_630_FR },
         { "640%",          OPTION_VAL_640_FR },
         { "650%",          OPTION_VAL_650_FR },
         { "660%",          OPTION_VAL_660_FR },
         { "670%",          OPTION_VAL_670_FR },
         { "680%",          OPTION_VAL_680_FR },
         { "690%",          OPTION_VAL_690_FR },
         { "700%",          OPTION_VAL_700_FR },
         { "710%",          OPTION_VAL_710_FR },
         { "720%",          OPTION_VAL_720_FR },
         { "730%",          OPTION_VAL_730_FR },
         { "740%",          OPTION_VAL_740_FR },
         { "750%",          OPTION_VAL_750_FR },
         { NULL, NULL },
      },
      "100%"
   },
   {
      BEETLE_OPT(gpu_overclock),
      BEETLE_OPT_GPU_OVERCLOCK_LABEL_FR,
      NULL,
      BEETLE_OPT_GPU_OVERCLOCK_INFO_0_FR,
      NULL,
      "system",
      {
         { "1x(native)", OPTION_VAL_1X_NATIVE_FR },
         { "2x",         OPTION_VAL_2X_FR },
         { "4x",         OPTION_VAL_4X_FR },
         { "8x",         OPTION_VAL_8X_FR },
         { "16x",        OPTION_VAL_16X_FR },
         { "32x",        OPTION_VAL_32X_FR },
         { NULL, NULL },
      },
      "1x(native)"
   },
   {
      BEETLE_OPT(gte_overclock),
      BEETLE_OPT_GTE_OVERCLOCK_LABEL_FR,
      NULL,
      BEETLE_OPT_GTE_OVERCLOCK_INFO_0_FR,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
#ifndef EMSCRIPTEN
   {
      BEETLE_OPT(cd_access_method),
      BEETLE_OPT_CD_ACCESS_METHOD_LABEL_FR,
      NULL,
      BEETLE_OPT_CD_ACCESS_METHOD_INFO_0_FR,
      NULL,
      "system",
      {
         { "sync",     OPTION_VAL_SYNC_FR },
         { "async",    OPTION_VAL_ASYNC_FR },
         { "precache", OPTION_VAL_PRECACHE_FR },
         { NULL, NULL },
      },
      "sync"
   },
#endif
   {
      BEETLE_OPT(cd_fastload),
      BEETLE_OPT_CD_FASTLOAD_LABEL_FR,
      NULL,
      BEETLE_OPT_CD_FASTLOAD_INFO_0_FR,
      NULL,
      "system",
      {
         { "2x(native)", OPTION_VAL_2X_NATIVE_FR },
         { "4x",          OPTION_VAL_4X_FR },
         { "6x",          OPTION_VAL_6X_FR },
         { "8x",          OPTION_VAL_8X_FR },
         { "10x",         OPTION_VAL_10X_FR },
         { "12x",         OPTION_VAL_12X_FR },
         { "14x",         OPTION_VAL_14X_FR },
         { NULL, NULL },
      },
      "2x(native)"
   },
   {
      BEETLE_OPT(skip_bios),
      BEETLE_OPT_SKIP_BIOS_LABEL_FR,
      NULL,
      BEETLE_OPT_SKIP_BIOS_INFO_0_FR,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(override_bios),
      BEETLE_OPT_OVERRIDE_BIOS_LABEL_FR,
      NULL,
      BEETLE_OPT_OVERRIDE_BIOS_INFO_0_FR,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "psxonpsp", OPTION_VAL_PSXONPSP_FR },
         { "ps1_rom",  OPTION_VAL_PS1_ROM_FR },
         { "openbios", OPTION_VAL_OPENBIOS_FR     },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(region),
      BEETLE_OPT_REGION_LABEL_FR,
      NULL,
      BEETLE_OPT_REGION_INFO_0_FR,
      NULL,
      "system",
      {
         { "auto",   OPTION_VAL_AUTO_FR },
         { "ntsc-j", OPTION_VAL_NTSC_J_FR },
         { "ntsc-u", OPTION_VAL_NTSC_U_FR },
         { "pal",    OPTION_VAL_PAL_FR },
         { NULL, NULL },
      },
      "auto"
   },
   {
      BEETLE_OPT(internal_resolution),
      BEETLE_OPT_INTERNAL_RESOLUTION_LABEL_FR,
      NULL,
      BEETLE_OPT_INTERNAL_RESOLUTION_INFO_0_FR,
      NULL,
      "video",
      {
         { "1x(native)", OPTION_VAL_1X_NATIVE_FR },
         { "2x",         OPTION_VAL_2X_FR },
         { "4x",         OPTION_VAL_4X_FR },
         { "8x",         OPTION_VAL_8X_FR },
         { "16x",        OPTION_VAL_16X_FR },
         { NULL, NULL },
      },
      "1x(native)"
   },
#if defined(HAVE_OPENGL) || defined(HAVE_OPENGLES) || defined(HAVE_VULKAN)
   {
      BEETLE_OPT(renderer),
      BEETLE_OPT_RENDERER_LABEL_FR,
      NULL,
      BEETLE_OPT_RENDERER_INFO_0_FR,
      NULL,
      "video",
      {
         { "hardware",    OPTION_VAL_HARDWARE_FR },
#if defined(HAVE_OPENGL) || defined(HAVE_OPENGLES)
         { "hardware_gl", OPTION_VAL_HARDWARE_GL_FR },
#endif
#if defined(HAVE_VULKAN)
         { "hardware_vk", OPTION_VAL_HARDWARE_VK_FR },
#endif
         { "software",    OPTION_VAL_SOFTWARE_FR },
         { NULL, NULL },
      },
      "hardware"
   },
   {
      BEETLE_OPT(renderer_software_fb),
      BEETLE_OPT_RENDERER_SOFTWARE_FB_LABEL_FR,
      NULL,
      BEETLE_OPT_RENDERER_SOFTWARE_FB_INFO_0_FR,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled"
   },
#endif
#if defined(HAVE_OPENGL) || defined(HAVE_OPENGLES)
   {
      BEETLE_OPT(depth),
      BEETLE_OPT_DEPTH_LABEL_FR,
      NULL,
      BEETLE_OPT_DEPTH_INFO_0_FR,
      NULL,
      "video",
      {
         { "16bpp(native)", OPTION_VAL_16BPP_NATIVE_FR },
         { "32bpp",         OPTION_VAL_32BPP_FR },
         { NULL, NULL },
      },
      "16bpp(native)"
   },
   // Sort of, it's more like 15-bit high color and 24-bit true color for visible output. The alpha channel is used for mask bit. Vulkan renderer uses ABGR1555_555 for 31 bits internal? FMVs are always 24-bit on all renderers like original hardware (BGR888, no alpha)
#endif
   {
      BEETLE_OPT(dither_mode),
      BEETLE_OPT_DITHER_MODE_LABEL_FR,
      NULL,
      BEETLE_OPT_DITHER_MODE_INFO_0_FR,
      NULL,
      "video",
      {
         { "1x(native)",          OPTION_VAL_1X_NATIVE_FR },
         { "internal resolution", OPTION_VAL_INTERNAL_RESOLUTION_FR },
         { "disabled",            NULL },
         { NULL, NULL },
      },
      "1x(native)"
   },
#ifdef HAVE_VULKAN
   {
      BEETLE_OPT(scaled_uv_offset),
      BEETLE_OPT_SCALED_UV_OFFSET_LABEL_FR,
      NULL,
      BEETLE_OPT_SCALED_UV_OFFSET_INFO_0_FR,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled"
   },
#endif
#if defined(HAVE_OPENGL) || defined(HAVE_OPENGLES) || defined(HAVE_VULKAN)
   {
      BEETLE_OPT(filter),
      BEETLE_OPT_FILTER_LABEL_FR,
      NULL,
      BEETLE_OPT_FILTER_INFO_0_FR,
      NULL,
      "video",
      {
         { "nearest",  OPTION_VAL_NEAREST_FR },
         { "SABR",     OPTION_VAL_SABR_FR },
         { "xBR",      OPTION_VAL_XBR_FR },
         { "bilinear", OPTION_VAL_BILINEAR_FR },
         { "3-point",  OPTION_VAL_3_POINT_FR },
         { "JINC2",    OPTION_VAL_JINC2_FR },
         { NULL, NULL },
      },
      "nearest"
   },
#ifdef HAVE_VULKAN
   {
      BEETLE_OPT(filter_exclude_sprite),
      BEETLE_OPT_FILTER_EXCLUDE_SPRITE_LABEL_FR,
      NULL,
      BEETLE_OPT_FILTER_EXCLUDE_SPRITE_INFO_0_FR,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "opaque", OPTION_VAL_OPAQUE_FR },
         { "all", OPTION_VAL_ALL_FR },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(filter_exclude_2d_polygon),
      BEETLE_OPT_FILTER_EXCLUDE_2D_POLYGON_LABEL_FR,
      NULL,
      BEETLE_OPT_FILTER_EXCLUDE_2D_POLYGON_INFO_0_FR,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "opaque", OPTION_VAL_OPAQUE_FR },
         { "all", OPTION_VAL_ALL_FR },
         { NULL, NULL },
      },
      "disabled"
   },
#endif
#endif
#ifdef HAVE_VULKAN
   {
      BEETLE_OPT(adaptive_smoothing),
      BEETLE_OPT_ADAPTIVE_SMOOTHING_LABEL_FR,
      NULL,
      BEETLE_OPT_ADAPTIVE_SMOOTHING_INFO_0_FR,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(super_sampling),
      BEETLE_OPT_SUPER_SAMPLING_LABEL_FR,
      NULL,
      BEETLE_OPT_SUPER_SAMPLING_INFO_0_FR,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(msaa),
      BEETLE_OPT_MSAA_LABEL_FR,
      NULL,
      BEETLE_OPT_MSAA_INFO_0_FR,
      NULL,
      "video",
      {
         { "1x",  OPTION_VAL_1X_FR },
         { "2x",  OPTION_VAL_2X_FR },
         { "4x",  OPTION_VAL_4X_FR },
         { "8x",  OPTION_VAL_8X_FR },
         { "16x", OPTION_VAL_16X_FR },
         { NULL, NULL },
      },
      "1x"
   },
   {
      BEETLE_OPT(mdec_yuv),
      BEETLE_OPT_MDEC_YUV_LABEL_FR,
      NULL,
      BEETLE_OPT_MDEC_YUV_INFO_0_FR,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(track_textures),
      BEETLE_OPT_TRACK_TEXTURES_LABEL_FR,
      NULL,
      BEETLE_OPT_TRACK_TEXTURES_INFO_0_FR,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
#ifdef TEXTURE_DUMPING_ENABLED
   {
      BEETLE_OPT(dump_textures),
      BEETLE_OPT_DUMP_TEXTURES_LABEL_FR,
      NULL,
      BEETLE_OPT_DUMP_TEXTURES_INFO_0_FR,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
#endif
   {
      BEETLE_OPT(replace_textures),
      BEETLE_OPT_REPLACE_TEXTURES_LABEL_FR,
      NULL,
      BEETLE_OPT_REPLACE_TEXTURES_INFO_0_FR,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
#endif
   {
      BEETLE_OPT(pgxp_mode),
      BEETLE_OPT_PGXP_MODE_LABEL_FR,
      NULL,
      BEETLE_OPT_PGXP_MODE_INFO_0_FR,
      NULL,
      "pgxp",
      {
         { "disabled",     NULL },
         { "memory only",  OPTION_VAL_MEMORY_ONLY_FR },
         { "memory + CPU", OPTION_VAL_MEMORY_CPU_FR },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(pgxp_2d_tol),
      BEETLE_OPT_PGXP_2D_TOL_LABEL_FR,
      NULL,
      BEETLE_OPT_PGXP_2D_TOL_INFO_0_FR,
      NULL,
      "pgxp",
      {
         { "disabled", NULL },
         { "0px", OPTION_VAL_0PX_FR },
         { "1px", OPTION_VAL_1PX_FR },
         { "2px", OPTION_VAL_2PX_FR },
         { "3px", OPTION_VAL_3PX_FR },
         { "4px", OPTION_VAL_4PX_FR },
         { "5px", OPTION_VAL_5PX_FR },
         { "6px", OPTION_VAL_6PX_FR },
         { "7px", OPTION_VAL_7PX_FR },
         { "8px", OPTION_VAL_8PX_FR },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(pgxp_nclip),
      BEETLE_OPT_PGXP_NCLIP_LABEL_FR,
      NULL,
      BEETLE_OPT_PGXP_NCLIP_INFO_0_FR,
      NULL,
      "pgxp",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
#if defined(HAVE_OPENGL) || defined(HAVE_OPENGLES) || defined(HAVE_VULKAN)
   {
      BEETLE_OPT(pgxp_vertex),
      BEETLE_OPT_PGXP_VERTEX_LABEL_FR,
      NULL,
      BEETLE_OPT_PGXP_VERTEX_INFO_0_FR,
      NULL,
      "pgxp",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
#endif
   {
      BEETLE_OPT(pgxp_texture),
      BEETLE_OPT_PGXP_TEXTURE_LABEL_FR,
      NULL,
      BEETLE_OPT_PGXP_TEXTURE_INFO_0_FR,
      NULL,
      "pgxp",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(display_internal_fps),
      BEETLE_OPT_DISPLAY_INTERNAL_FPS_LABEL_FR,
      NULL,
      BEETLE_OPT_DISPLAY_INTERNAL_FPS_INFO_0_FR,
      NULL,
      "osd",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(display_osd),
      BEETLE_OPT_DISPLAY_OSD_LABEL_FR,
      NULL,
      BEETLE_OPT_DISPLAY_OSD_INFO_0_FR,
      NULL,
      "osd",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled"
   },
#if defined(HAVE_OPENGL) || defined(HAVE_OPENGLES) || defined(HAVE_VULKAN)
   {
      BEETLE_OPT(display_vram),
      BEETLE_OPT_DISPLAY_VRAM_LABEL_FR,
      NULL,
      BEETLE_OPT_DISPLAY_VRAM_INFO_0_FR,
      NULL,
      "osd",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
#endif
   {
      BEETLE_OPT(analog_calibration),
      BEETLE_OPT_ANALOG_CALIBRATION_LABEL_FR,
      NULL,
      BEETLE_OPT_ANALOG_CALIBRATION_INFO_0_FR,
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(analog_toggle),
      BEETLE_OPT_ANALOG_TOGGLE_LABEL_FR,
      NULL,
      BEETLE_OPT_ANALOG_TOGGLE_INFO_0_FR,
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { "enabled-analog", OPTION_VAL_ENABLED_ANALOG_FR },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(analog_toggle_combo),
      BEETLE_OPT_ANALOG_TOGGLE_COMBO_LABEL_FR,
      NULL,
      BEETLE_OPT_ANALOG_TOGGLE_COMBO_INFO_0_FR,
      NULL,
      "input",
      {
         { "l1+l2+r1+r2+start+select", OPTION_VAL_L1_L2_R1_R2_START_SELECT_FR },
         { "l1+r1+select",             OPTION_VAL_L1_R1_SELECT_FR },
         { "l1+r1+start",              OPTION_VAL_L1_R1_START_FR },
         { "l1+r1+l3",                 OPTION_VAL_L1_R1_L3_FR },
         { "l1+r1+r3",                 OPTION_VAL_L1_R1_R3_FR },
         { "l2+r2+select",             OPTION_VAL_L2_R2_SELECT_FR },
         { "l2+r2+start",              OPTION_VAL_L2_R2_START_FR },
         { "l2+r2+l3",                 OPTION_VAL_L2_R2_L3_FR },
         { "l2+r2+r3",                 OPTION_VAL_L2_R2_R3_FR },
         { "l3+r3",                    OPTION_VAL_L3_R3_FR },
         { NULL, NULL },
      },
      "l1+r1+select"
   },
   {
      BEETLE_OPT(analog_toggle_hold),
      BEETLE_OPT_ANALOG_TOGGLE_HOLD_LABEL_FR,
      NULL,
      BEETLE_OPT_ANALOG_TOGGLE_HOLD_INFO_0_FR,
      NULL,
      "input",
      {
         { "0", OPTION_VAL_0_FR },
         { "1", OPTION_VAL_1_FR },
         { "2", OPTION_VAL_2_FR },
         { "3", OPTION_VAL_3_FR },
         { "4", OPTION_VAL_4_FR },
         { "5", OPTION_VAL_5_FR },
         { NULL, NULL },
      },
      "1"
   },
   {
      BEETLE_OPT(enable_multitap_port1),
      BEETLE_OPT_ENABLE_MULTITAP_PORT1_LABEL_FR,
      NULL,
      BEETLE_OPT_ENABLE_MULTITAP_PORT1_INFO_0_FR,
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(enable_multitap_port2),
      BEETLE_OPT_ENABLE_MULTITAP_PORT2_LABEL_FR,
      NULL,
      BEETLE_OPT_ENABLE_MULTITAP_PORT2_INFO_0_FR,
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(gun_input_mode),
      BEETLE_OPT_GUN_INPUT_MODE_LABEL_FR,
      NULL,
      BEETLE_OPT_GUN_INPUT_MODE_INFO_0_FR,
      NULL,
      "input",
      {
         { "lightgun",    OPTION_VAL_LIGHTGUN_FR },
         { "touchscreen", OPTION_VAL_TOUCHSCREEN_FR },
         { NULL, NULL },
      },
      "lightgun"
   },
   // Shouldn't the gun_input_mode just be Mouse vs. Touchscreen?
   {
      BEETLE_OPT(gun_cursor),
      BEETLE_OPT_GUN_CURSOR_LABEL_FR,
      NULL,
      BEETLE_OPT_GUN_CURSOR_INFO_0_FR,
      NULL,
      "input",
      {
         { "off",   "disabled" },
         { "cross", OPTION_VAL_CROSS_FR },
         { "dot",   OPTION_VAL_DOT_FR },
         { NULL, NULL },
      },
      "cross"
   },
   {
      BEETLE_OPT(crosshair_color_p1),
      BEETLE_OPT_CROSSHAIR_COLOR_P1_LABEL_FR,
      NULL,
      BEETLE_OPT_CROSSHAIR_COLOR_P1_INFO_0_FR,
      NULL,
      "input",
      {
         { "red",    OPTION_VAL_RED_FR },
         { "blue",   OPTION_VAL_BLUE_FR },
         { "green",  OPTION_VAL_GREEN_FR },
         { "orange", OPTION_VAL_ORANGE_FR },
         { "yellow", OPTION_VAL_YELLOW_FR },
         { "cyan",   OPTION_VAL_CYAN_FR },
         { "pink",   OPTION_VAL_PINK_FR },
         { "purple", OPTION_VAL_PURPLE_FR },
         { "black",  OPTION_VAL_BLACK_FR },
         { "white",  OPTION_VAL_WHITE_FR },
         { NULL, NULL },
      },
      "red"
   },
   {
      BEETLE_OPT(crosshair_color_p2),
      BEETLE_OPT_CROSSHAIR_COLOR_P2_LABEL_FR,
      NULL,
      BEETLE_OPT_CROSSHAIR_COLOR_P2_INFO_0_FR,
      NULL,
      "input",
      {
         { "blue",   OPTION_VAL_BLUE_FR },
         { "red",    OPTION_VAL_RED_FR },
         { "green",  OPTION_VAL_GREEN_FR },
         { "orange", OPTION_VAL_ORANGE_FR },
         { "yellow", OPTION_VAL_YELLOW_FR },
         { "cyan",   OPTION_VAL_CYAN_FR },
         { "pink",   OPTION_VAL_PINK_FR },
         { "purple", OPTION_VAL_PURPLE_FR },
         { "black",  OPTION_VAL_BLACK_FR },
         { "white",  OPTION_VAL_WHITE_FR },
         { NULL, NULL },
      },
      "blue"
   },
   {
      BEETLE_OPT(mouse_sensitivity),
      BEETLE_OPT_MOUSE_SENSITIVITY_LABEL_FR,
      NULL,
      BEETLE_OPT_MOUSE_SENSITIVITY_INFO_0_FR,
      NULL,
      "input",
      {
         { "5%",   OPTION_VAL_5_O42_FR },
         { "10%",  OPTION_VAL_10_FR },
         { "15%",  OPTION_VAL_15_FR },
         { "20%",  OPTION_VAL_20_FR },
         { "25%",  OPTION_VAL_25_FR },
         { "30%",  OPTION_VAL_30_FR },
         { "35%",  OPTION_VAL_35_FR },
         { "40%",  OPTION_VAL_40_FR },
         { "45%",  OPTION_VAL_45_FR },
         { "50%",  OPTION_VAL_50_FR },
         { "55%",  OPTION_VAL_55_FR },
         { "60%",  OPTION_VAL_60_FR },
         { "65%",  OPTION_VAL_65_FR },
         { "70%",  OPTION_VAL_70_FR },
         { "75%",  OPTION_VAL_75_FR },
         { "80%",  OPTION_VAL_80_FR },
         { "85%",  OPTION_VAL_85_FR },
         { "90%",  OPTION_VAL_90_FR },
         { "95%",  OPTION_VAL_95_FR },
         { "100%", OPTION_VAL_100_O42_FR },
         { "105%", OPTION_VAL_105_FR },
         { "110%", OPTION_VAL_110_FR },
         { "115%", OPTION_VAL_115_FR },
         { "120%", OPTION_VAL_120_FR },
         { "125%", OPTION_VAL_125_FR },
         { "130%", OPTION_VAL_130_FR },
         { "135%", OPTION_VAL_135_FR },
         { "140%", OPTION_VAL_140_FR },
         { "145%", OPTION_VAL_145_FR },
         { "150%", OPTION_VAL_150_FR },
         { "155%", OPTION_VAL_155_FR },
         { "160%", OPTION_VAL_160_FR },
         { "165%", OPTION_VAL_165_FR },
         { "170%", OPTION_VAL_170_FR },
         { "175%", OPTION_VAL_175_FR },
         { "180%", OPTION_VAL_180_FR },
         { "185%", OPTION_VAL_185_FR },
         { "190%", OPTION_VAL_190_FR },
         { "195%", OPTION_VAL_195_FR },
         { "200%", OPTION_VAL_200_FR },
         { NULL, NULL },
      },
      "100%"
   },
   {
      BEETLE_OPT(negcon_response),
      BEETLE_OPT_NEGCON_RESPONSE_LABEL_FR,
      NULL,
      BEETLE_OPT_NEGCON_RESPONSE_INFO_0_FR,
      NULL,
      "input",
      {
         { "linear",    OPTION_VAL_LINEAR_FR },
         { "quadratic", OPTION_VAL_QUADRATIC_FR },
         { "cubic",     OPTION_VAL_CUBIC_FR },
         { NULL, NULL },
      },
      "linear"
   },
   {
      BEETLE_OPT(negcon_deadzone),
      BEETLE_OPT_NEGCON_DEADZONE_LABEL_FR,
      NULL,
      BEETLE_OPT_NEGCON_DEADZONE_INFO_0_FR,
      NULL,
      "input",
      {
         { "0%",  OPTION_VAL_0_O44_FR },
         { "5%",  OPTION_VAL_5_O42_FR },
         { "10%", OPTION_VAL_10_FR },
         { "15%", OPTION_VAL_15_FR },
         { "20%", OPTION_VAL_20_FR },
         { "25%", OPTION_VAL_25_FR },
         { "30%", OPTION_VAL_30_FR },
         { NULL, NULL },
      },
      "0%"
   },
   {
      BEETLE_OPT(use_mednafen_memcard0_method),
      BEETLE_OPT_USE_MEDNAFEN_MEMCARD0_METHOD_LABEL_FR,
      NULL,
      BEETLE_OPT_USE_MEDNAFEN_MEMCARD0_METHOD_INFO_0_FR,
      NULL,
      "memcards",
      {
         { "libretro", OPTION_VAL_LIBRETRO_FR },
         { "mednafen", OPTION_VAL_MEDNAFEN_FR },
         { NULL, NULL },
      },
      "libretro"
   },
   {
      BEETLE_OPT(enable_memcard1),
      BEETLE_OPT_ENABLE_MEMCARD1_LABEL_FR,
      NULL,
      BEETLE_OPT_ENABLE_MEMCARD1_INFO_0_FR,
      NULL,
      "memcards",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(shared_memory_cards),
      BEETLE_OPT_SHARED_MEMORY_CARDS_LABEL_FR,
      NULL,
      BEETLE_OPT_SHARED_MEMORY_CARDS_INFO_0_FR,
      NULL,
      "memcards",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(memcard_left_index),
      BEETLE_OPT_MEMCARD_LEFT_INDEX_LABEL_FR,
      NULL,
      BEETLE_OPT_MEMCARD_LEFT_INDEX_INFO_0_FR,
      NULL,
      "memcards",
      {
         { "0",  OPTION_VAL_0_O48_FR },
         { "1",  NULL },
         { "2",  NULL },
         { "3",  NULL },
         { "4",  NULL },
         { "5",  NULL },
         { "6",  NULL },
         { "7",  NULL },
         { "8",  NULL },
         { "9",  NULL },
         { "10", NULL },
         { "11", NULL },
         { "12", NULL },
         { "13", NULL },
         { "14", NULL },
         { "15", NULL },
         { "16", NULL },
         { "17", NULL },
         { "18", NULL },
         { "19", NULL },
         { "20", NULL },
         { "21", NULL },
         { "22", NULL },
         { "23", NULL },
         { "24", NULL },
         { "25", NULL },
         { "26", NULL },
         { "27", NULL },
         { "28", NULL },
         { "29", NULL },
         { "30", NULL },
         { "31", NULL },
         { "32", NULL },
         { "33", NULL },
         { "34", NULL },
         { "35", NULL },
         { "36", NULL },
         { "37", NULL },
         { "38", NULL },
         { "39", NULL },
         { "40", NULL },
         { "41", NULL },
         { "42", NULL },
         { "43", NULL },
         { "44", NULL },
         { "45", NULL },
         { "46", NULL },
         { "47", NULL },
         { "48", NULL },
         { "49", NULL },
         { "50", NULL },
         { "51", NULL },
         { "52", NULL },
         { "53", NULL },
         { "54", NULL },
         { "55", NULL },
         { "56", NULL },
         { "57", NULL },
         { "58", NULL },
         { "59", NULL },
         { "60", NULL },
         { "61", NULL },
         { "62", NULL },
         { "63", NULL },
         { NULL, NULL },
      },
      "0"
   },
   {
      BEETLE_OPT(memcard_right_index),
      BEETLE_OPT_MEMCARD_RIGHT_INDEX_LABEL_FR,
      NULL,
      BEETLE_OPT_MEMCARD_RIGHT_INDEX_INFO_0_FR,
      NULL,
      "memcards",
      {
         { "0",  NULL },
         { "1",  OPTION_VAL_1_O49_FR },
         { "2",  NULL },
         { "3",  NULL },
         { "4",  NULL },
         { "5",  NULL },
         { "6",  NULL },
         { "7",  NULL },
         { "8",  NULL },
         { "9",  NULL },
         { "10", NULL },
         { "11", NULL },
         { "12", NULL },
         { "13", NULL },
         { "14", NULL },
         { "15", NULL },
         { "16", NULL },
         { "17", NULL },
         { "18", NULL },
         { "19", NULL },
         { "20", NULL },
         { "21", NULL },
         { "22", NULL },
         { "23", NULL },
         { "24", NULL },
         { "25", NULL },
         { "26", NULL },
         { "27", NULL },
         { "28", NULL },
         { "29", NULL },
         { "30", NULL },
         { "31", NULL },
         { "32", NULL },
         { "33", NULL },
         { "34", NULL },
         { "35", NULL },
         { "36", NULL },
         { "37", NULL },
         { "38", NULL },
         { "39", NULL },
         { "40", NULL },
         { "41", NULL },
         { "42", NULL },
         { "43", NULL },
         { "44", NULL },
         { "45", NULL },
         { "46", NULL },
         { "47", NULL },
         { "48", NULL },
         { "49", NULL },
         { "50", NULL },
         { "51", NULL },
         { "52", NULL },
         { "53", NULL },
         { "54", NULL },
         { "55", NULL },
         { "56", NULL },
         { "57", NULL },
         { "58", NULL },
         { "59", NULL },
         { "60", NULL },
         { "61", NULL },
         { "62", NULL },
         { "63", NULL },
         { NULL, NULL },
      },
      "1"
   },
   {
      BEETLE_OPT(line_render),
      BEETLE_OPT_LINE_RENDER_LABEL_FR,
      NULL,
      BEETLE_OPT_LINE_RENDER_INFO_0_FR,
      NULL,
      "hacks",
      {
         { "default",    OPTION_VAL_DEFAULT_FR },
         { "aggressive", OPTION_VAL_AGGRESSIVE_FR },
         { "disabled",   NULL },
         { NULL, NULL },
      },
      "default"
   },
   {
      BEETLE_OPT(widescreen_hack),
      BEETLE_OPT_WIDESCREEN_HACK_LABEL_FR,
      NULL,
      BEETLE_OPT_WIDESCREEN_HACK_INFO_0_FR,
      NULL,
      "hacks",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(widescreen_hack_aspect_ratio),
      BEETLE_OPT_WIDESCREEN_HACK_ASPECT_RATIO_LABEL_FR,
      NULL,
      BEETLE_OPT_WIDESCREEN_HACK_ASPECT_RATIO_INFO_0_FR,
      NULL,
      "hacks",
      {
         { "16:9",  OPTION_VAL_16_9_FR },
         { "16:10", OPTION_VAL_16_10_FR },
         { "18:9",  OPTION_VAL_18_9_FR },
         { "19:9",  OPTION_VAL_19_9_FR },
         { "20:9",  OPTION_VAL_20_9_FR },
         { "21:9",  OPTION_VAL_21_9_FR }, // 64:27
         { "32:9",  OPTION_VAL_32_9_FR },
         { NULL,    NULL },
      },
      "16:9"
   },
#if defined(HAVE_LIGHTREC)
   {
      BEETLE_OPT(cpu_dynarec),
      BEETLE_OPT_CPU_DYNAREC_LABEL_FR,
      NULL,
      BEETLE_OPT_CPU_DYNAREC_INFO_0_FR,
      NULL,
      "hacks",
      {
         { "disabled",        OPTION_VAL_DISABLED_FR },
         { "execute",         OPTION_VAL_EXECUTE_FR },
         { "execute_one",     OPTION_VAL_EXECUTE_ONE_FR },
         { "run_interpreter", OPTION_VAL_RUN_INTERPRETER_FR },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(dynarec_invalidate),
      BEETLE_OPT_DYNAREC_INVALIDATE_LABEL_FR,
      NULL,
      BEETLE_OPT_DYNAREC_INVALIDATE_INFO_0_FR,
      NULL,
      "hacks",
      {
         { "full", OPTION_VAL_FULL_FR },
         { "dma",  OPTION_VAL_DMA_FR },
         { NULL, NULL },
      },
      "full"
   },
   {
      BEETLE_OPT(dynarec_eventcycles),
      BEETLE_OPT_DYNAREC_EVENTCYCLES_LABEL_FR,
      NULL,
      BEETLE_OPT_DYNAREC_EVENTCYCLES_INFO_0_FR,
      NULL,
      "hacks",
      {
         { "128",  OPTION_VAL_128_FR },
         { "256",  NULL },
         { "384",  NULL },
         { "512",  NULL },
         { "640",  NULL },
         { "768",  NULL },
         { "896",  NULL },
         { "1024", NULL },
         { "1152", NULL },
         { "1280", NULL },
         { "1408", NULL },
         { "1536", NULL },
         { "1664", NULL },
         { "1792", NULL },
         { "1920", NULL },
         { "2048", NULL },
         { NULL, NULL },
      },
      "128"
   },
   {
      BEETLE_OPT(dynarec_spu_samples),
      BEETLE_OPT_DYNAREC_SPU_SAMPLES_LABEL_FR,
      NULL,
      BEETLE_OPT_DYNAREC_SPU_SAMPLES_INFO_0_FR,
      NULL,
      "hacks",
      {
         { "1",  OPTION_VAL_1_O49_FR },
         { "4",  NULL },
         { "16", NULL },
         { NULL, NULL },
      },
      "1"
   },
#endif
   {
      BEETLE_OPT(core_timing_fps),
      BEETLE_OPT_CORE_TIMING_FPS_LABEL_FR,
      NULL,
      BEETLE_OPT_CORE_TIMING_FPS_INFO_0_FR,
      NULL,
      "video",
      {
         { "auto_toggle",       OPTION_VAL_AUTO_TOGGLE_FR },
         { "force_progressive", OPTION_VAL_FORCE_PROGRESSIVE_FR },
         { "force_interlaced",  OPTION_VAL_FORCE_INTERLACED_FR },
      },
      "auto_toggle"
   },
   {
      BEETLE_OPT(deinterlacer),
      BEETLE_OPT_DEINTERLACER_LABEL_FR,
      NULL,
      BEETLE_OPT_DEINTERLACER_INFO_0_FR,
      NULL,
      "video",
      {
         { "weave",      OPTION_VAL_WEAVE_FR },
         { "bob",        OPTION_VAL_BOB_FR },
         { "bob_offset", OPTION_VAL_BOB_OFFSET_FR },
         { "fastmad",    OPTION_VAL_FASTMAD_FR },
         { "off",        "Off" },
         { NULL, NULL },
      },
      "weave"
   },
   {
      BEETLE_OPT(frame_duping),
      BEETLE_OPT_FRAME_DUPING_LABEL_FR,
      NULL,
      BEETLE_OPT_FRAME_DUPING_INFO_0_FR,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(pal_video_timing_override),
      BEETLE_OPT_PAL_VIDEO_TIMING_OVERRIDE_LABEL_FR,
      NULL,
      BEETLE_OPT_PAL_VIDEO_TIMING_OVERRIDE_INFO_0_FR,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(aspect_ratio),
      BEETLE_OPT_ASPECT_RATIO_LABEL_FR,
      NULL,
      BEETLE_OPT_ASPECT_RATIO_INFO_0_FR,
      NULL,
      "video",
      {
         { "corrected", OPTION_VAL_CORRECTED_FR },
         { "uncorrected", OPTION_VAL_UNCORRECTED_FR },
         { "4:3",  OPTION_VAL_4_3_FR },
         { "ntsc", OPTION_VAL_NTSC_FR },
      },
      "corrected"
   },
   {
      BEETLE_OPT(crop_overscan),
      BEETLE_OPT_CROP_OVERSCAN_LABEL_FR,
      NULL,
      BEETLE_OPT_CROP_OVERSCAN_INFO_0_FR,
      NULL,
      "video",
      {
         { "disabled", "Disabled" },
         { "static",   OPTION_VAL_STATIC_FR },
         { "smart",    OPTION_VAL_SMART_FR },
         { NULL, NULL },
      },
      "smart"
   },
   {
      BEETLE_OPT(image_crop),
      BEETLE_OPT_IMAGE_CROP_LABEL_FR,
      NULL,
      BEETLE_OPT_IMAGE_CROP_INFO_0_FR,
      NULL,
      "video",
      {
         { "disabled", "0" },
         { "1px",      OPTION_VAL_1PX_FR },
         { "2px",      OPTION_VAL_2PX_FR },
         { "3px",      OPTION_VAL_3PX_FR },
         { "4px",      OPTION_VAL_4PX_FR },
         { "5px",      OPTION_VAL_5PX_FR },
         { "6px",      OPTION_VAL_6PX_FR },
         { "7px",      OPTION_VAL_7PX_FR },
         { "8px",      OPTION_VAL_8PX_FR },
         { "9px",      OPTION_VAL_9PX_FR },
         { "10px",     OPTION_VAL_10PX_FR },
         { "11px",     OPTION_VAL_11PX_FR },
         { "12px",     OPTION_VAL_12PX_FR },
         { "13px",     OPTION_VAL_13PX_FR },
         { "14px",     OPTION_VAL_14PX_FR },
         { "15px",     OPTION_VAL_15PX_FR },
         { "16px",     OPTION_VAL_16PX_FR },
         { "17px",     OPTION_VAL_17PX_FR },
         { "18px",     OPTION_VAL_18PX_FR },
         { "19px",     OPTION_VAL_19PX_FR },
         { "20px",     OPTION_VAL_20PX_FR },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(image_offset),
      BEETLE_OPT_IMAGE_OFFSET_LABEL_FR,
      NULL,
      BEETLE_OPT_IMAGE_OFFSET_INFO_0_FR,
      NULL,
      "video",
      {
         { "-12px",    OPTION_VAL_12PX_O64_FR },
         { "-11px",    OPTION_VAL_11PX_O64_FR },
         { "-10px",    OPTION_VAL_10PX_O64_FR },
         { "-9px",     OPTION_VAL_9PX_O64_FR },
         { "-8px",     OPTION_VAL_8PX_O64_FR },
         { "-7px",     OPTION_VAL_7PX_O64_FR },
         { "-6px",     OPTION_VAL_6PX_O64_FR },
         { "-5px",     OPTION_VAL_5PX_O64_FR },
         { "-4px",     OPTION_VAL_4PX_O64_FR },
         { "-3px",     OPTION_VAL_3PX_O64_FR },
         { "-2px",     OPTION_VAL_2PX_O64_FR },
         { "-1px",     OPTION_VAL_1PX_O64_FR },
         { "disabled", "0" },
         { "+1px",     OPTION_VAL_1PX_O64_0_FR },
         { "+2px",     OPTION_VAL_2PX_O64_0_FR },
         { "+3px",     OPTION_VAL_3PX_O64_0_FR },
         { "+4px",     OPTION_VAL_4PX_O64_0_FR },
         { "+5px",     OPTION_VAL_5PX_O64_0_FR },
         { "+6px",     OPTION_VAL_6PX_O64_0_FR },
         { "+7px",     OPTION_VAL_7PX_O64_0_FR },
         { "+8px",     OPTION_VAL_8PX_O64_0_FR },
         { "+9px",     OPTION_VAL_9PX_O64_0_FR },
         { "+10px",    OPTION_VAL_10PX_O64_0_FR },
         { "+11px",    OPTION_VAL_11PX_O64_0_FR },
         { "+12px",    OPTION_VAL_12PX_O64_0_FR },
         { NULL, NULL },
      },
      "disabled"
   },
#if defined(HAVE_OPENGL) || defined(HAVE_OPENGLES) || defined(HAVE_VULKAN)
   {
      BEETLE_OPT(image_offset_cycles),
      BEETLE_OPT_IMAGE_OFFSET_CYCLES_LABEL_FR,
      NULL,
      BEETLE_OPT_IMAGE_OFFSET_CYCLES_INFO_0_FR,
      NULL,
      "video",
      {
         { "-40",      NULL },
         { "-39",      NULL },
         { "-38",      NULL },
         { "-37",      NULL },
         { "-36",      NULL },
         { "-35",      NULL },
         { "-34",      NULL },
         { "-33",      NULL },
         { "-32",      NULL },
         { "-31",      NULL },
         { "-30",      NULL },
         { "-29",      NULL },
         { "-28",      NULL },
         { "-27",      NULL },
         { "-26",      NULL },
         { "-25",      NULL },
         { "-24",      NULL },
         { "-23",      NULL },
         { "-22",      NULL },
         { "-21",      NULL },
         { "-20",      NULL },
         { "-19",      NULL },
         { "-18",      NULL },
         { "-17",      NULL },
         { "-16",      NULL },
         { "-15",      NULL },
         { "-14",      NULL },
         { "-13",      NULL },
         { "-12",      NULL },
         { "-11",      NULL },
         { "-10",      NULL },
         { "-9",       NULL },
         { "-8",       NULL },
         { "-7",       NULL },
         { "-6",       NULL },
         { "-5",       NULL },
         { "-4",       NULL },
         { "-3",       NULL },
         { "-2",       NULL },
         { "-1",       NULL },
         { "0",        "0" },
         { "+1",       NULL },
         { "+2",       NULL },
         { "+3",       NULL },
         { "+4",       NULL },
         { "+5",       NULL },
         { "+6",       NULL },
         { "+7",       NULL },
         { "+8",       NULL },
         { "+9",       NULL },
         { "+10",      NULL },
         { "+11",      NULL },
         { "+12",      NULL },
         { "+13",      NULL },
         { "+14",      NULL },
         { "+15",      NULL },
         { "+16",      NULL },
         { "+17",      NULL },
         { "+18",      NULL },
         { "+19",      NULL },
         { "+20",      NULL },
         { "+21",      NULL },
         { "+22",      NULL },
         { "+23",      NULL },
         { "+24",      NULL },
         { "+25",      NULL },
         { "+26",      NULL },
         { "+27",      NULL },
         { "+28",      NULL },
         { "+29",      NULL },
         { "+30",      NULL },
         { "+31",      NULL },
         { "+32",      NULL },
         { "+33",      NULL },
         { "+34",      NULL },
         { "+35",      NULL },
         { "+36",      NULL },
         { "+37",      NULL },
         { "+38",      NULL },
         { "+39",      NULL },
         { "+40",      NULL },
         { NULL, NULL},
      },
      "0"
   },
#endif
   {
      BEETLE_OPT(initial_scanline),
      BEETLE_OPT_INITIAL_SCANLINE_LABEL_FR,
      NULL,
      BEETLE_OPT_INITIAL_SCANLINE_INFO_0_FR,
      NULL,
      "video",
      {
         { "0",  OPTION_VAL_0_O48_FR },
         { "1",  NULL },
         { "2",  NULL },
         { "3",  NULL },
         { "4",  NULL },
         { "5",  NULL },
         { "6",  NULL },
         { "7",  NULL },
         { "8",  NULL },
         { "9",  NULL },
         { "10", NULL },
         { "11", NULL },
         { "12", NULL },
         { "13", NULL },
         { "14", NULL },
         { "15", NULL },
         { "16", NULL },
         { "17", NULL },
         { "18", NULL },
         { "19", NULL },
         { "20", NULL },
         { "21", NULL },
         { "22", NULL },
         { "23", NULL },
         { "24", NULL },
         { "25", NULL },
         { "26", NULL },
         { "27", NULL },
         { "28", NULL },
         { "29", NULL },
         { "30", NULL },
         { "31", NULL },
         { "32", NULL },
         { "33", NULL },
         { "34", NULL },
         { "35", NULL },
         { "36", NULL },
         { "37", NULL },
         { "38", NULL },
         { "39", NULL },
         { "40", NULL },
         { NULL, NULL },
      },
      "0"
   },
   {
      BEETLE_OPT(last_scanline),
      BEETLE_OPT_LAST_SCANLINE_LABEL_FR,
      NULL,
      BEETLE_OPT_LAST_SCANLINE_INFO_0_FR,
      NULL,
      "video",
      {
         { "210", NULL },
         { "211", NULL },
         { "212", NULL },
         { "213", NULL },
         { "214", NULL },
         { "215", NULL },
         { "216", NULL },
         { "217", NULL },
         { "218", NULL },
         { "219", NULL },
         { "220", NULL },
         { "221", NULL },
         { "222", NULL },
         { "223", NULL },
         { "224", NULL },
         { "225", NULL },
         { "226", NULL },
         { "227", NULL },
         { "228", NULL },
         { "229", NULL },
         { "230", NULL },
         { "231", NULL },
         { "232", NULL },
         { "233", NULL },
         { "234", NULL },
         { "235", NULL },
         { "236", NULL },
         { "237", NULL },
         { "238", NULL },
         { "239", OPTION_VAL_239_FR },
         { NULL, NULL },
      },
      "239"
   },
   {
      BEETLE_OPT(initial_scanline_pal),
      BEETLE_OPT_INITIAL_SCANLINE_PAL_LABEL_FR,
      NULL,
      BEETLE_OPT_INITIAL_SCANLINE_PAL_INFO_0_FR,
      NULL,
      "video",
      {
         { "0",  OPTION_VAL_0_O48_FR },
         { "1",  NULL },
         { "2",  NULL },
         { "3",  NULL },
         { "4",  NULL },
         { "5",  NULL },
         { "6",  NULL },
         { "7",  NULL },
         { "8",  NULL },
         { "9",  NULL },
         { "10", NULL },
         { "11", NULL },
         { "12", NULL },
         { "13", NULL },
         { "14", NULL },
         { "15", NULL },
         { "16", NULL },
         { "17", NULL },
         { "18", NULL },
         { "19", NULL },
         { "20", NULL },
         { "21", NULL },
         { "22", NULL },
         { "23", NULL },
         { "24", NULL },
         { "25", NULL },
         { "26", NULL },
         { "27", NULL },
         { "28", NULL },
         { "29", NULL },
         { "30", NULL },
         { "31", NULL },
         { "32", NULL },
         { "33", NULL },
         { "34", NULL },
         { "35", NULL },
         { "36", NULL },
         { "37", NULL },
         { "38", NULL },
         { "39", NULL },
         { "40", NULL },
         { NULL, NULL },
      },
      "0"
   },
   {
      BEETLE_OPT(last_scanline_pal),
      BEETLE_OPT_LAST_SCANLINE_PAL_LABEL_FR,
      NULL,
      BEETLE_OPT_LAST_SCANLINE_PAL_INFO_0_FR,
      NULL,
      "video",
      {
         { "230", NULL },
         { "231", NULL },
         { "232", NULL },
         { "233", NULL },
         { "234", NULL },
         { "235", NULL },
         { "236", NULL },
         { "237", NULL },
         { "238", NULL },
         { "239", NULL },
         { "240", NULL },
         { "241", NULL },
         { "242", NULL },
         { "243", NULL },
         { "244", NULL },
         { "245", NULL },
         { "246", NULL },
         { "247", NULL },
         { "248", NULL },
         { "249", NULL },
         { "250", NULL },
         { "251", NULL },
         { "252", NULL },
         { "253", NULL },
         { "254", NULL },
         { "255", NULL },
         { "256", NULL },
         { "257", NULL },
         { "258", NULL },
         { "259", NULL },
         { "260", NULL },
         { "261", NULL },
         { "262", NULL },
         { "263", NULL },
         { "264", NULL },
         { "265", NULL },
         { "266", NULL },
         { "267", NULL },
         { "268", NULL },
         { "269", NULL },
         { "270", NULL },
         { "271", NULL },
         { "272", NULL },
         { "273", NULL },
         { "274", NULL },
         { "275", NULL },
         { "276", NULL },
         { "277", NULL },
         { "278", NULL },
         { "279", NULL },
         { "280", NULL },
         { "281", NULL },
         { "282", NULL },
         { "283", NULL },
         { "284", NULL },
         { "285", NULL },
         { "286", NULL },
         { "287", OPTION_VAL_287_FR },
         { NULL, NULL },
      },
      "287"
   },
   { NULL, NULL, NULL, NULL, NULL, NULL, {{0}}, NULL },
};
struct retro_core_options_v2 options_fr = {
   option_cats_fr,
   option_defs_fr
};

/* RETRO_LANGUAGE_GA */

#define CATEGORY_SYSTEM_LABEL_GA "Córas"
#define CATEGORY_SYSTEM_INFO_0_GA "Athraigh roghanna córais agus róchlogála."
#define CATEGORY_VIDEO_LABEL_GA "Físeán"
#define CATEGORY_VIDEO_INFO_0_GA "Athraigh cóimheas gné, bearradh taispeána, scagaire físe agus socruithe scipeála frámaí."
#define CATEGORY_PGXP_LABEL_GA "PGXP (Píblíne Claochlaithe Geoiméadrachta Beacht)"
#define CATEGORY_PGXP_INFO_0_GA "Rialaíonn na roghanna seo feabhsuithe ar féidir leo grafaicí a fheabhsú i gcomparáid leis an gconsól bunaidh. Is féidir le PGXP fáil réidh le huigeachtaí casta agus fadhbanna troid-Z."
#define CATEGORY_OSD_LABEL_GA "Taispeántas ar an Scáileán"
#define CATEGORY_OSD_INFO_0_GA "Athraigh na fógraí atá á thaispeáint ar an scáileán."
#define CATEGORY_INPUT_LABEL_GA "Ionchur"
#define CATEGORY_INPUT_INFO_0_GA "Athraigh socruithe an ghunna solais, na luiche agus an neGcon."
#define CATEGORY_MEMCARDS_LABEL_GA "Cárta Cuimhne"
#define CATEGORY_MEMCARDS_INFO_0_GA "Athraigh socruithe a bhaineann le cárta cuimhne fíorúil."
#define CATEGORY_HACKS_LABEL_GA "Cleasanna Aithrise"
#define CATEGORY_HACKS_INFO_0_GA "Athraigh socruithe cruinneas aithrise a mbíonn tionchar acu ar fheidhmíocht agus comhoiriúnacht íseal-leibhéil."
#define BEETLE_OPT_CPU_FREQ_SCALE_LABEL_GA "Scálú Minicíochta LAP (Róchlogáil)"
#define BEETLE_OPT_CPU_FREQ_SCALE_INFO_0_GA "Róchlogáil (nó fochlogáil) an LAP PSX aithrisithe. Is féidir le róchlogáil moilliú a dhíchur agus rátaí fráma a fheabhsú i gcluichí áirithe ar chostas riachtanais feidhmíochta méadaithe. Tabhair faoi deara go bhfuil teorannóir ráta fráma inmheánach ag roinnt cluichí agus nach bhféadfadh róchlogáil leas a bhaint astu. D’fhéadfadh sé go mbeadh éifeachtaí áirithe beochan níos tapúla ná mar a bhí beartaithe i roinnt teidil nuair a bhíonn siad róchlogáilte."
#define OPTION_VAL_50_GA NULL
#define OPTION_VAL_60_GA NULL
#define OPTION_VAL_70_GA NULL
#define OPTION_VAL_80_GA NULL
#define OPTION_VAL_90_GA NULL
#define OPTION_VAL_100_GA "100% (Dúchasach)"
#define OPTION_VAL_110_GA NULL
#define OPTION_VAL_120_GA NULL
#define OPTION_VAL_130_GA NULL
#define OPTION_VAL_140_GA NULL
#define OPTION_VAL_150_GA NULL
#define OPTION_VAL_160_GA NULL
#define OPTION_VAL_170_GA NULL
#define OPTION_VAL_180_GA NULL
#define OPTION_VAL_190_GA NULL
#define OPTION_VAL_200_GA NULL
#define OPTION_VAL_210_GA NULL
#define OPTION_VAL_220_GA NULL
#define OPTION_VAL_230_GA NULL
#define OPTION_VAL_240_GA NULL
#define OPTION_VAL_250_GA NULL
#define OPTION_VAL_260_GA NULL
#define OPTION_VAL_270_GA NULL
#define OPTION_VAL_280_GA NULL
#define OPTION_VAL_290_GA NULL
#define OPTION_VAL_300_GA NULL
#define OPTION_VAL_310_GA NULL
#define OPTION_VAL_320_GA NULL
#define OPTION_VAL_330_GA NULL
#define OPTION_VAL_340_GA NULL
#define OPTION_VAL_350_GA NULL
#define OPTION_VAL_360_GA NULL
#define OPTION_VAL_370_GA NULL
#define OPTION_VAL_380_GA NULL
#define OPTION_VAL_390_GA NULL
#define OPTION_VAL_400_GA NULL
#define OPTION_VAL_410_GA NULL
#define OPTION_VAL_420_GA NULL
#define OPTION_VAL_430_GA NULL
#define OPTION_VAL_440_GA NULL
#define OPTION_VAL_450_GA NULL
#define OPTION_VAL_460_GA NULL
#define OPTION_VAL_470_GA NULL
#define OPTION_VAL_480_GA NULL
#define OPTION_VAL_490_GA NULL
#define OPTION_VAL_500_GA NULL
#define OPTION_VAL_510_GA NULL
#define OPTION_VAL_520_GA NULL
#define OPTION_VAL_530_GA NULL
#define OPTION_VAL_540_GA NULL
#define OPTION_VAL_550_GA NULL
#define OPTION_VAL_560_GA NULL
#define OPTION_VAL_570_GA NULL
#define OPTION_VAL_580_GA NULL
#define OPTION_VAL_590_GA NULL
#define OPTION_VAL_600_GA NULL
#define OPTION_VAL_610_GA NULL
#define OPTION_VAL_620_GA NULL
#define OPTION_VAL_630_GA NULL
#define OPTION_VAL_640_GA NULL
#define OPTION_VAL_650_GA NULL
#define OPTION_VAL_660_GA NULL
#define OPTION_VAL_670_GA NULL
#define OPTION_VAL_680_GA NULL
#define OPTION_VAL_690_GA NULL
#define OPTION_VAL_700_GA NULL
#define OPTION_VAL_710_GA NULL
#define OPTION_VAL_720_GA NULL
#define OPTION_VAL_730_GA NULL
#define OPTION_VAL_740_GA NULL
#define OPTION_VAL_750_GA NULL
#define BEETLE_OPT_GPU_OVERCLOCK_LABEL_GA "Róchlogáil Rasterizer GPU"
#define BEETLE_OPT_GPU_OVERCLOCK_INFO_0_GA "Cumasaigh róchlogáil an rastairitheora 2T atá laistigh de GPU an PSX aithrisithe. Ní fheabhsaíonn sé rindreáil 3T, agus i gcoitinne níl mórán éifeacht aige."
#define OPTION_VAL_1X_NATIVE_GA "1x (Dúchasach)"
#define OPTION_VAL_2X_GA NULL
#define OPTION_VAL_4X_GA NULL
#define OPTION_VAL_8X_GA NULL
#define OPTION_VAL_16X_GA NULL
#define OPTION_VAL_32X_GA NULL
#define BEETLE_OPT_GTE_OVERCLOCK_LABEL_GA "Róchlogáil GTE"
#define BEETLE_OPT_GTE_OVERCLOCK_INFO_0_GA "Ísligh gach oibríocht GTE aithriste (comhphróiseálaí LAP le haghaidh grafaicí 3T) go dtí latency tairiseach aon-timthriall. I gcás cluichí a úsáideann an GTE go mór, is féidir leis seo feabhas mór a chur ar an ráta fráma agus ar chobhsaíocht ama fráma."
#define BEETLE_OPT_CD_ACCESS_METHOD_LABEL_GA "Modh Rochtana CD"
#define BEETLE_OPT_CD_ACCESS_METHOD_INFO_0_GA "Déanann 'Sioncrónach' aithris ar chrua-earraí bunaidh. Is féidir le 'Ashioncrónach' stad a laghdú ar ghléasanna le stóráil mhall. Lódálann 'Réamh-Thaisce' an íomhá diosca iomlán isteach sa chuimhne agus ábhar á sheoladh, rud a d'fhéadfadh feabhas a chur ar amanna lódála sa chluiche ar chostas moille tosaigh ag an am tosaithe. D'fhéadfadh 'Réamh-Thaisce' fadhbanna a chruthú ar chórais le RAM íseal, agus titeann sé ar ais go 'Sioncrónach' le haghaidh meán fisiceach. Atosú ag teastáil."
#define OPTION_VAL_SYNC_GA "Sioncrónach"
#define OPTION_VAL_ASYNC_GA "Asincrónach"
#define OPTION_VAL_PRECACHE_GA "Réamh-Thaisce"
#define BEETLE_OPT_CD_FASTLOAD_LABEL_GA "Luas Luchtaithe CD"
#define BEETLE_OPT_CD_FASTLOAD_INFO_0_GA "Is féidir le luachanna níos airde ná '2x (Dúchasach)' amanna lódála sa chluiche a laghdú go mór, ach d'fhéadfadh earráidí ama a bheith mar thoradh orthu. B'fhéidir nach bhfeidhmeoidh roinnt cluichí i gceart os cionn luacha áirithe."
#define OPTION_VAL_2X_NATIVE_GA "2x (Dúchasach)"
#define OPTION_VAL_6X_GA NULL
#define OPTION_VAL_10X_GA NULL
#define OPTION_VAL_12X_GA NULL
#define OPTION_VAL_14X_GA NULL
#define BEETLE_OPT_SKIP_BIOS_LABEL_GA "Léim thar an BIOS"
#define BEETLE_OPT_SKIP_BIOS_INFO_0_GA "Seachain beochan tosaithe BIOS PlayStation a thaispeántar de ghnáth agus ábhar á luchtú. Tabhair faoi deara: Má chumasaítear é seo, bíonn fadhbanna comhoiriúnachta ann le roinnt cluichí (cluichí atá cosanta ag cóip PAL, Saga Frontier, srl.)."
#define BEETLE_OPT_OVERRIDE_BIOS_LABEL_GA "Sáraigh BIOS"
#define BEETLE_OPT_OVERRIDE_BIOS_INFO_0_GA "Sáraigh an BIOS caighdeánach atá sainiúil don réigiún le ceann saor ó réigiún má aimsítear é. Atosú ag teastáil."
#define OPTION_VAL_PSXONPSP_GA NULL
#define OPTION_VAL_PS1_ROM_GA NULL
#define OPTION_VAL_OPENBIOS_GA NULL
#define BEETLE_OPT_REGION_LABEL_GA "Réigiún an Chórais"
#define BEETLE_OPT_REGION_INFO_0_GA "Socraíonn sé an réigiún cúltaca a úsáidtear nuair nach féidir ábhar a bhrath go huathoibríoch (m.sh. comhaid PS-X EXE amha). Úsáideann cluichí diosca-bhunaithe an réigiún atá leabaithe ar an diosca fós. Atosú ag teastáil."
#define OPTION_VAL_AUTO_GA "Uathoibríoch"
#define OPTION_VAL_NTSC_J_GA "NTSC-J (An tSeapáin)"
#define OPTION_VAL_NTSC_U_GA "NTSC-U (Meiriceá Thuaidh)"
#define OPTION_VAL_PAL_GA "PAL (An Eoraip)"
#define BEETLE_OPT_INTERNAL_RESOLUTION_LABEL_GA "Rún GPU Inmheánach"
#define BEETLE_OPT_INTERNAL_RESOLUTION_INFO_0_GA "Feabhsaíonn taifeach níos airde ná '1x (Dúchasach)' dílseacht na samhlacha 3T ar chostas riachtanais feidhmíochta méadaithe. Ní bhíonn tionchar ag an socrú seo ar eilimintí 2T go ginearálta."
#define BEETLE_OPT_RENDERER_LABEL_GA "Rindreálaí"
#define BEETLE_OPT_RENDERER_INFO_0_GA "Is é an rindreálaí bogearraí an ceann is cruinne ach tá riachtanais feidhmíochta géara aige nuair a bhíonn sé ag rith ag taifeach inmheánach GPU méadaithe. Cé nach bhfuil na rindreálaithe crua-earraí chomh cruinn céanna, feabhsaíonn siad feidhmíocht thar an rindreálaí bogearraí ag taifeach inmheánach méadaithe agus cuireann siad feabhsuithe grafacha éagsúla ar chumas. Roghnaíonn 'Crua-earraí (Uathoibríoch)' an rindreálaí Vulkan nó OpenGL go huathoibríoch, ag brath ar thiománaí físe tosaigh reatha an libretro. Mura bhfuil an tiománaí físe a chuirtear ar fáil comhoiriúnach le Vulkan nó OpenGL 3.3, ansin titeann an croí ar ais chuig an rindreálaí bogearraí. Atosú ag teastáil."
#define OPTION_VAL_HARDWARE_GA "Crua-earraí (Uathoibríoch)"
#define OPTION_VAL_HARDWARE_GL_GA "Crua-earraí (OpenGL)"
#define OPTION_VAL_HARDWARE_VK_GA "Crua-earraí (Vulkan)"
#define OPTION_VAL_SOFTWARE_GA "Bogearraí"
#define BEETLE_OPT_RENDERER_SOFTWARE_FB_LABEL_GA "Maolán Fráma Bogearraí"
#define BEETLE_OPT_RENDERER_SOFTWARE_FB_INFO_0_GA "Cumasaigh aithris chruinn ar éifeachtaí maolán fráma (m.sh. doiléire gluaisne, corraíl chatha FF7) agus rindreálaithe crua-earraí in úsáid trí chóip den rindreálaí bogearraí a rith ag réiteach dúchasach sa chúlra. Má tá sé díchumasaithe, fágtar na hoibríochtaí seo ar lár (OpenGL) nó rindreáiltear iad ar an GPU (Vulkan). Is féidir le díchumasú feidhmíocht a fheabhsú ach d'fhéadfadh earráidí grafacha tromchúiseacha a bheith mar thoradh air. Fág cumasaithe mura bhfuil tú cinnte."
#define BEETLE_OPT_DEPTH_LABEL_GA "Doimhneacht Dath Inmheánach"
#define BEETLE_OPT_DEPTH_INFO_0_GA "Is féidir le doimhneacht datha níos airde éifeachtaí bandaí datha a laghdú gan úsáid a bhaint as díthering. Déanann 16 bpp aithris ar chrua-earraí bunaidh ach d'fhéadfadh bandaí a bheith le feiceáil mura bhfuil díthering cumasaithe. Moltar 'Patrún Díthering' a dhíchumasú nuair a shocraítear an rogha seo go 32 bpp."
#define OPTION_VAL_16BPP_NATIVE_GA "16 bpp (Dúchasach)"
#define OPTION_VAL_32BPP_GA NULL
#define BEETLE_OPT_DITHER_MODE_LABEL_GA "Patrún Díthrithe"
#define BEETLE_OPT_DITHER_MODE_INFO_0_GA "Déanann '1x (Dúchasach)' aithris ar dhíthriú ísealtaifigh dúchasach a úsáideann crua-earraí bunaidh chun déantáin bandála datha atá le feiceáil ag doimhneacht datha dúchasach a mhaolú. Déanann 'Rún Inmheánach' gráinneacht dhíthriú a scála go dtí an rún inmheánach cumraithe le haghaidh torthaí níos glaine. Moltar é a dhíchumasú nuair a bhíonn doimhneacht datha 32 bpp á rith. Tabhair faoi deara: Ar Vulkan, cuirfidh cumasú seo iallach ar shampláil síos go doimhneacht datha dúchasach, agus cuirfidh díchumasú aschur ar chumas go huathoibríoch ag doimhneacht datha níos airde."
#define OPTION_VAL_INTERNAL_RESOLUTION_GA "Rún Inmheánach"
#define BEETLE_OPT_SCALED_UV_OFFSET_LABEL_GA "Fritháireamh UV Uigeachta"
#define BEETLE_OPT_SCALED_UV_OFFSET_INFO_0_GA "Uigeachtaí samplacha do pholagáin 3T ag fritháireamh le haghaidh taifeach inmheánach níos airde ná 1x. Laghdaíonn sé seo fuaiseanna uigeachta ach d’fhéadfadh sé go mbeadh teipeanna grafacha neamhbheartaithe mar thoradh air."
#define BEETLE_OPT_FILTER_LABEL_GA "Scagadh Uigeachta"
#define BEETLE_OPT_FILTER_INFO_0_GA "Déanann 'Nearest' aithris ar chrua-earraí bunaidh. Is scagairí réidhithe iad 'Bilinear' agus '3-Point', a laghdaíonn picteilíniú trí doiléiriú. Is scagairí uasghrádaithe iad 'SABR', 'xBR', agus 'JINC2' a d'fhéadfadh feabhas a chur ar dhílseacht/géire uigeachta ar chostas riachtanais feidhmíochta méadaithe. Ní thacaítear leis seo ach ag na rindreálaithe crua-earraí."
#define OPTION_VAL_NEAREST_GA "Is cóngaraí"
#define OPTION_VAL_SABR_GA NULL
#define OPTION_VAL_XBR_GA NULL
#define OPTION_VAL_BILINEAR_GA "Délíneach"
#define OPTION_VAL_3_POINT_GA "3-Phointe"
#define OPTION_VAL_JINC2_GA NULL
#define BEETLE_OPT_FILTER_EXCLUDE_SPRITE_LABEL_GA "Eisiamh Sprideanna ó Scagadh"
#define BEETLE_OPT_FILTER_EXCLUDE_SPRITE_INFO_0_GA "Ná cuir scagadh uigeachta i bhfeidhm ar sprideanna. Cuireann sé cosc ​​ar shíomaí i gcluichí éagsúla le cúlraí rindreáilte le sprideanna 2T. Úsáid i dteannta le Smúdáil Oiriúnaitheach nó scagaire iarphróiseála eile chun an éifeacht is fearr a fháil."
#define OPTION_VAL_OPAQUE_GA "Teimhneach Amháin"
#define OPTION_VAL_ALL_GA "Teimhneach agus Leath-Trédhearcach"
#define BEETLE_OPT_FILTER_EXCLUDE_2D_POLYGON_LABEL_GA "Eisiamh Polagáin 2T ó Scagadh"
#define BEETLE_OPT_FILTER_EXCLUDE_2D_POLYGON_INFO_0_GA "Ná cuir scagadh uigeachta i bhfeidhm ar pholagáin 2T. Braitear polagáin 2T le heorastach agus d'fhéadfadh go mbeadh teipeanna ann. Bain úsáid as i dteannta le Smúdáil Oiriúnaitheach nó scagaire iarphróiseála eile chun an éifeacht is fearr a bhaint amach."
#define BEETLE_OPT_ADAPTIVE_SMOOTHING_LABEL_GA "Smúdáil Oiriúnaitheach"
#define BEETLE_OPT_ADAPTIVE_SMOOTHING_INFO_0_GA "Réidh amach saothar ealaíne 2T agus eilimintí UI gan réada rindreáilte 3T a dhéanamh doiléir. Ní thacaítear leis seo ach leis an rindreálaí Vulkan."
#define BEETLE_OPT_SUPER_SAMPLING_LABEL_GA "Super-sampláil (Síossamplú go dtí an Taifeach Dúchasach)"
#define BEETLE_OPT_SUPER_SAMPLING_INFO_0_GA "Laghdaigh samplaíocht an ábhair rindreáilte ó réiteach inmheánach uasghrádaithe síos go réiteach dúchasach. Trí é seo a chomhcheangal le hiolraitheoirí réitigh inmheánacha níos airde, is féidir cluichí a thaispeáint le rudaí 3T frith-ailíse ag réiteach íseal dúchasach. Tugann sé na torthaí is fearr nuair a chuirtear i bhfeidhm é ar theidil a mheascann eilimintí 2T agus 3T (e.g. carachtair 3T ar chúlraí réamh-rindreáilte), agus oibríonn sé go maith i gcomhar le scáthaitheoirí CRT. Ní thacaítear leis ach amháin ag an rindreálaí Vulkan. Tabhair faoi deara: Moltar 'Patrún Díthrithe' a dhíchumasú agus an rogha seo á cumasú."
#define BEETLE_OPT_MSAA_LABEL_GA "Frith-Ailiasú Il-Shampláilte"
#define BEETLE_OPT_MSAA_INFO_0_GA "Feabhsaíonn MSAA cuma réad 3T. Ní thacaítear leis ach amháin ag an rindreálaí Vulkan."
#define OPTION_VAL_1X_GA NULL
#define BEETLE_OPT_MDEC_YUV_LABEL_GA "Scagaire Chroma MDEC YUV"
#define BEETLE_OPT_MDEC_YUV_INFO_0_GA "Feabhas a chur ar cháilíocht athsheinm FMV trí airteafóin 'macrabhlocáil' (cearnóga/imill gharbha) a laghdú. Ní thacaítear leis seo ach leis an rindreálaí Vulkan."
#define BEETLE_OPT_TRACK_TEXTURES_LABEL_GA "Uigeachtaí Rian"
#define BEETLE_OPT_TRACK_TEXTURES_INFO_0_GA "Réamhriachtanas le haghaidh dumpáil agus athsholáthar uigeachta. Is dócha go dtitfidh sé i bhformhór na gcluichí."
#define BEETLE_OPT_DUMP_TEXTURES_LABEL_GA "Uigeachtaí Dumpála"
#define BEETLE_OPT_DUMP_TEXTURES_INFO_0_GA "Dumpáil uigeachtaí úsáidte chuig <cd>-texture-dump/"
#define BEETLE_OPT_REPLACE_TEXTURES_LABEL_GA "Cuir Uigeachtaí in Ionad"
#define BEETLE_OPT_REPLACE_TEXTURES_INFO_0_GA "Cuir uigeachtaí in ionad a chéile ag baint úsáide as leaganacha HD ó <cd>-texture-replacements/"
#define BEETLE_OPT_PGXP_MODE_LABEL_GA "Mód Oibríochta PGXP"
#define BEETLE_OPT_PGXP_MODE_INFO_0_GA "Déan rudaí 3T a rindreáil le cruinneas fo-phicteilíní, ag íoslaghdú saobhadh agus crith rudaí 3T a fheictear ar chrua-earraí bunaidh mar gheall ar úsáid chomhordanáidí buaicphointí seasta. Tá saincheisteanna comhoiriúnachta íosta ag baint leis an mód 'Cuimhne Amháin' agus moltar é le haghaidh úsáide ginearálta. Is féidir leis an mód 'Cuimhne + LAP (Buggy)' crith a laghdú níos mó fós ach tá ceanglais ardfheidhmíochta aige agus d'fhéadfadh sé earráidí geoiméadrachta éagsúla a chur faoi deara."
#define OPTION_VAL_MEMORY_ONLY_GA "Cuimhne Amháin"
#define OPTION_VAL_MEMORY_CPU_GA "Cuimhne + LAP (Buggy)"
#define BEETLE_OPT_PGXP_2D_TOL_LABEL_GA "Caoinfhulaingt Geoiméadrachta 2T PGXP"
#define BEETLE_OPT_PGXP_2D_TOL_INFO_0_GA "Folaigh earráidí níos soiléire in oibríochtaí PGXP: sonraíonn an luach an lamháltas ina gcoimeádfar luachanna PGXP i gcás geoiméadrachtaí gan faisnéis cheart faoi dhoimhneacht."
#define OPTION_VAL_0PX_GA NULL
#define OPTION_VAL_1PX_GA NULL
#define OPTION_VAL_2PX_GA NULL
#define OPTION_VAL_3PX_GA NULL
#define OPTION_VAL_4PX_GA NULL
#define OPTION_VAL_5PX_GA NULL
#define OPTION_VAL_6PX_GA NULL
#define OPTION_VAL_7PX_GA NULL
#define OPTION_VAL_8PX_GA NULL
#define BEETLE_OPT_PGXP_NCLIP_LABEL_GA "Marú Prímitiúil PGXP"
#define BEETLE_OPT_PGXP_NCLIP_INFO_0_GA "Bain úsáid as cur i bhfeidhm NCLIP PGXP. Feabhsaíonn sé an chuma trí phoill i ngeoiméadrachtaí a laghdú le comhordanáidí PGXP. Is eol go mbíonn roinnt cluichí faoi ghlas i gcúinsí éagsúla dá bharr."
#define BEETLE_OPT_PGXP_VERTEX_LABEL_GA "Taisce Buaicphointe PGXP"
#define BEETLE_OPT_PGXP_VERTEX_INFO_0_GA "Cuir suíomhanna buaicphointí feabhsaithe PGXP i dtaisce lena n-athúsáid ar fud líníochtaí polagán. Is féidir ailíniú réad a fheabhsú agus fuálacha infheicthe a laghdú agus uigeachtaí á rindreáil, ach d'fhéadfadh dearfacha bréagacha agus fiosrúcháin á ndéanamh ar an taisce fabhtanna grafacha a chruthú. Moltar faoi láthair an rogha seo a fhágáil díchumasaithe. Ní chuirtear an rogha seo i bhfeidhm ach amháin nuair a bhíonn Mód Oibríochta PGXP cumasaithe. Ní thacaítear leis ach ag na rindreálaithe crua-earraí."
#define BEETLE_OPT_PGXP_TEXTURE_LABEL_GA "Uigeacht Cheart Peirspictíochta PGXP"
#define BEETLE_OPT_PGXP_TEXTURE_INFO_0_GA "Cuir mapáil uigeachta atá ceart ó thaobh peirspictíochta in ionad mapáil uigeachta afíne dúchais PSX. Cuireann sé deireadh le saobhadh agus le lúbadh uigeachtaí atá ag brath ar shuíomh, rud a fhágann go mbíonn uigeachtaí ailínithe i gceart. Ní chuirtear an rogha seo i bhfeidhm ach amháin nuair a bhíonn Mód Oibríochta PGXP cumasaithe. Tacaítear leis ag na rindreálaithe crua-earraí agus an rindreálaí bogearraí."
#define BEETLE_OPT_DISPLAY_INTERNAL_FPS_LABEL_GA "Taispeáin FPS Inmheánach"
#define BEETLE_OPT_DISPLAY_INTERNAL_FPS_INFO_0_GA "Taispeáin an ráta fráma inmheánach ag a bhfuil an córas PlayStation aithrisithe ag rindreáil ábhair. Nóta: Éilíonn sé seo go mbeidh fógraí ar an scáileán cumasaithe i dtosach libretro."
#define BEETLE_OPT_DISPLAY_OSD_LABEL_GA "Taispeáin Teachtaireachtaí OSD"
#define BEETLE_OPT_DISPLAY_OSD_INFO_0_GA "Taispeáin teachtaireachtaí OSD a ghineann an croí."
#define BEETLE_OPT_DISPLAY_VRAM_LABEL_GA "Taispeáin VRAM Iomlán (Dífhabhtú)"
#define BEETLE_OPT_DISPLAY_VRAM_INFO_0_GA "Amharcléirigh VRAM an chonsóil aithrisithe ar fad. Ní thacaítear leis seo ach ag rindreálaithe crua-earraí OpenGL agus Vulkan. Tabhair faoi deara: Chun críocha dífhabhtaithe atá sé seo, agus ba chóir é a dhíchumasú de ghnáth."
#define BEETLE_OPT_ANALOG_CALIBRATION_LABEL_GA "Féin-Chalabrú Analógach"
#define BEETLE_OPT_ANALOG_CALIBRATION_INFO_0_GA "Cumasaigh calabrú dinimiciúil ionchuir analógacha nuair a bhíonn an gléas socraithe go DualShock, Rialaitheoir Analógach, Luamhán Luamhán Analógach, nó neGcon. Déantar monatóireacht ar uasluachanna ionchuir cláraithe i bhfíor-am agus úsáidtear iad chun comhordanáidí analógacha a chuirtear chuig an aithriseoir a scála. Ba chóir é seo a úsáid le haghaidh cluichí ar nós Mega Man Legends 2 a bhfuiltear ag súil leo luachanna níos mó ná mar a sholáthraíonn rialtóirí nua-aimseartha. Chun na torthaí is fearr a fháil, ba chóir na bataí analógacha a rothlú go hiomlán chun an algartam calabrúcháin a choigeartú gach uair a luchtaítear ábhar."
#define BEETLE_OPT_ANALOG_TOGGLE_LABEL_GA "Mód Analógach DualShock a Athrú"
#define BEETLE_OPT_ANALOG_TOGGLE_INFO_0_GA "Nuair is é DualShock an cineál feiste ionchuir, ceadaíonn an rogha seo don DualShock aithriste a bheith idir mód DIGITEACH agus mód ANALÓGACH cosúil leis an gcrua-earraí bunaidh. Is féidir an mód a athrú freisin trí úsáid a bhaint as an teaglaim cnaipe roghnaithe."
#define OPTION_VAL_ENABLED_ANALOG_GA "Réamhshocrú-Analógach"
#define BEETLE_OPT_ANALOG_TOGGLE_COMBO_LABEL_GA "Comhcheangal Mód Analógach DualShock"
#define BEETLE_OPT_ANALOG_TOGGLE_COMBO_INFO_0_GA "An teaglaim cnaipí a úsáidfear chun aistriú idir mód DIGITEACH agus ANALÓGACH don DualShock aithrisithe."
#define OPTION_VAL_L1_L2_R1_R2_START_SELECT_GA NULL
#define OPTION_VAL_L1_R1_SELECT_GA NULL
#define OPTION_VAL_L1_R1_START_GA NULL
#define OPTION_VAL_L1_R1_L3_GA NULL
#define OPTION_VAL_L1_R1_R3_GA NULL
#define OPTION_VAL_L2_R2_SELECT_GA NULL
#define OPTION_VAL_L2_R2_START_GA NULL
#define OPTION_VAL_L2_R2_L3_GA NULL
#define OPTION_VAL_L2_R2_R3_GA NULL
#define OPTION_VAL_L3_R3_GA NULL
#define BEETLE_OPT_ANALOG_TOGGLE_HOLD_LABEL_GA "Moill Shealbhaíochta i Mód Analógach DualShock Combo"
#define BEETLE_OPT_ANALOG_TOGGLE_HOLD_INFO_0_GA "Coinnigh am do na cnaipí teaglama mód analógach."
#define OPTION_VAL_0_GA "Moill 0 Soicind"
#define OPTION_VAL_1_GA "Moill 1 Soicind"
#define OPTION_VAL_2_GA "Moill 2 Shoicind"
#define OPTION_VAL_3_GA "Moill 3 Shoicind"
#define OPTION_VAL_4_GA "Moill 4 Soicind"
#define OPTION_VAL_5_GA "Moill 5 Soicind"
#define BEETLE_OPT_ENABLE_MULTITAP_PORT1_LABEL_GA "Il-thapáil ar Port 1"
#define BEETLE_OPT_ENABLE_MULTITAP_PORT1_INFO_0_GA "Cumasaigh feidhmiúlacht il-thapála ar phort 1."
#define BEETLE_OPT_ENABLE_MULTITAP_PORT2_LABEL_GA "Il-thapáil ar Port 2"
#define BEETLE_OPT_ENABLE_MULTITAP_PORT2_INFO_0_GA "Cumasaigh feidhmiúlacht il-thapála ar phort 2."
#define BEETLE_OPT_GUN_INPUT_MODE_LABEL_GA "Mód Ionchuir Gunna"
#define BEETLE_OPT_GUN_INPUT_MODE_INFO_0_GA "Bain úsáid as ionchur 'Gunna Solais' nó 'Scáileán Tadhaill' atá rialaithe ag an luch nuair a bhíonn an cineál gléis socraithe go 'Guncon/G-Con 45' nó 'Justifier'."
#define OPTION_VAL_LIGHTGUN_GA "Gunna Solais"
#define OPTION_VAL_TOUCHSCREEN_GA "Scáileán tadhaill"
#define BEETLE_OPT_GUN_CURSOR_LABEL_GA "Cúrsóir Gunna"
#define BEETLE_OPT_GUN_CURSOR_INFO_0_GA "Tarraing cúrsóir ar an scáileán agus na cineálacha gléasanna ionchuir 'Guncon/G-Con 45' agus 'Justifier' in úsáid agat."
#define OPTION_VAL_CROSS_GA "Cros"
#define OPTION_VAL_DOT_GA "Ponc"
#define BEETLE_OPT_CROSSHAIR_COLOR_P1_LABEL_GA "Dath Crosaire Gunna: Port 1"
#define BEETLE_OPT_CROSSHAIR_COLOR_P1_INFO_0_GA "Dath crosaire gunna solais do phort 1."
#define OPTION_VAL_RED_GA "Dearg"
#define OPTION_VAL_BLUE_GA "Gorm"
#define OPTION_VAL_GREEN_GA "Glas"
#define OPTION_VAL_ORANGE_GA "Oráiste"
#define OPTION_VAL_YELLOW_GA "Buí"
#define OPTION_VAL_CYAN_GA "Cian"
#define OPTION_VAL_PINK_GA "Bándearg"
#define OPTION_VAL_PURPLE_GA "Corcra"
#define OPTION_VAL_BLACK_GA "Dubh"
#define OPTION_VAL_WHITE_GA "Bán"
#define BEETLE_OPT_CROSSHAIR_COLOR_P2_LABEL_GA "Dath Crosaire Gunna: Port 2"
#define BEETLE_OPT_CROSSHAIR_COLOR_P2_INFO_0_GA "Dath crosaire gunna solais do phort 2."
#define BEETLE_OPT_MOUSE_SENSITIVITY_LABEL_GA "Íogaireacht Luiche"
#define BEETLE_OPT_MOUSE_SENSITIVITY_INFO_0_GA "Freagrúlacht an chineáil gléis ionchuir 'Luch'."
#define OPTION_VAL_5_O42_GA NULL
#define OPTION_VAL_10_GA NULL
#define OPTION_VAL_15_GA NULL
#define OPTION_VAL_20_GA NULL
#define OPTION_VAL_25_GA NULL
#define OPTION_VAL_30_GA NULL
#define OPTION_VAL_35_GA NULL
#define OPTION_VAL_40_GA NULL
#define OPTION_VAL_45_GA NULL
#define OPTION_VAL_55_GA NULL
#define OPTION_VAL_65_GA NULL
#define OPTION_VAL_75_GA NULL
#define OPTION_VAL_85_GA NULL
#define OPTION_VAL_95_GA NULL
#define OPTION_VAL_100_O42_GA NULL
#define OPTION_VAL_105_GA NULL
#define OPTION_VAL_115_GA NULL
#define OPTION_VAL_125_GA NULL
#define OPTION_VAL_135_GA NULL
#define OPTION_VAL_145_GA NULL
#define OPTION_VAL_155_GA NULL
#define OPTION_VAL_165_GA NULL
#define OPTION_VAL_175_GA NULL
#define OPTION_VAL_185_GA NULL
#define OPTION_VAL_195_GA NULL
#define BEETLE_OPT_NEGCON_RESPONSE_LABEL_GA "Freagra Casadh neGcon"
#define BEETLE_OPT_NEGCON_RESPONSE_INFO_0_GA "Cineál freagartha an mhaide analógaigh chlé agus gníomh 'casadh' gléasanna ionchuir 'neGcon' aithriste á insamhladh. Is féidir díláithriú an mhaide analógaigh a mhapáil chuig uillinn rothlaithe neGcon go líneach, go cearnachach nó go ciúbach. Ceadaíonn 'Cuadrach' cruinneas níos mó ná 'Líneach' agus gluaiseachtaí beaga á ndéanamh. Méadaíonn 'Ciúbach' cruinneas gluaiseachtaí beaga a thuilleadh, ach 'áibhéilíonn' sé gluaiseachtaí níos mó. Tabhair faoi deara: Ní mholtar 'Líneach' ach amháin agus forimeallaigh roth rásaíochta in úsáid. Cuireann rialtóirí traidisiúnta ionchur analógach i bhfeidhm ar bhealach atá go bunúsach difriúil ó mheicníocht 'casadh' neGcon, sa chaoi is go ndéanann mapáil líneach ró-aimpliú ar ghluaiseachtaí beaga, rud a chuireann isteach ar rialú mín. I bhformhór na gcásanna, soláthraíonn 'Cuadrach' an garmheastachán is gaire do chrua-earraí fíor."
#define OPTION_VAL_LINEAR_GA "Líneach"
#define OPTION_VAL_QUADRATIC_GA "Cearnógach"
#define OPTION_VAL_CUBIC_GA "Ciúbach"
#define BEETLE_OPT_NEGCON_DEADZONE_LABEL_GA "crios Marbh Twist neGcon"
#define BEETLE_OPT_NEGCON_DEADZONE_INFO_0_GA "Crios marbh an mhaide analógaigh chlé agus gníomh 'casadh' gléasanna ionchuir 'neGcon' aithriste á insamhladh. Úsáidtear é chun drift rialtóra a dhíchur. Tabhair faoi deara: Soláthraíonn formhór na dteideal atá comhoiriúnach le neGcon roghanna sa chluiche chun luach crios marbh 'casadh' a shocrú. Chun cailliúint cruinnis a sheachaint, ba chóir an crios marbh sa chluiche a shocrú go náid *i gcónaí*. Níor chóir aon choigeartuithe riachtanacha a chur i bhfeidhm *ach* tríd an rogha lárnach seo. Tá sé seo thar a bheith tábhachtach nuair a bhíonn 'Freagra Casadh neGcon' socraithe go 'Cearnáiteach' nó 'Ciúbach'."
#define OPTION_VAL_0_O44_GA NULL
#define BEETLE_OPT_USE_MEDNAFEN_MEMCARD0_METHOD_LABEL_GA "Modh Cárta Cuimhne"
#define BEETLE_OPT_USE_MEDNAFEN_MEMCARD0_METHOD_INFO_0_GA "Formáid sonraí sábhála a úsáidtear don sliotán cárta cuimhne ar chlé. Féadfar 'Mednafen' a úsáid le haghaidh comhoiriúnachta leis an leagan neamhspleách de Mednafen. Nuair a úsáidtear é le Beetle PSX, bíonn formáidí comhionanna inmheánacha ag sábhálacha Libretro (.srm) agus Mednafen (.mcr) agus is féidir iad a thiontú eatarthu trí athainmniú. Atosú ag teastáil."
#define OPTION_VAL_LIBRETRO_GA NULL
#define OPTION_VAL_MEDNAFEN_GA NULL
#define BEETLE_OPT_ENABLE_MEMCARD1_LABEL_GA "Sliotán Cárta Cuimhne 2"
#define BEETLE_OPT_ENABLE_MEMCARD1_INFO_0_GA "Déan aithris ar an dara cárta cuimhne sa sliotán cárta cuimhne ar dheis. Tabhair faoi deara: Éilíonn roinnt cluichí go ndíchumasófar an rogha seo le go n-oibreoidh sé i gceart (e.g. Códainm Tenka). Tabhair faoi deara: Úsáideann Sliotán Cárta Cuimhne 2 an fhormáid sábhála Mednafen (.mcr). Atosú ag teastáil."
#define BEETLE_OPT_SHARED_MEMORY_CARDS_LABEL_GA "Cártaí Cuimhne Comhroinnte"
#define BEETLE_OPT_SHARED_MEMORY_CARDS_INFO_0_GA "Bain úsáid as cártaí cuimhne comhroinnte in ionad cártaí ar leithligh do gach ábhar. Tabhair faoi deara: má tá 'Modh Cárta Cuimhne' socraithe go 'Libretro', ní bheidh tionchar ach ar shliotán cárta cuimhne 2. Atosú ag teastáil."
#define BEETLE_OPT_MEMCARD_LEFT_INDEX_LABEL_GA "Sliotán Cárta Cuimhne 1 Innéacs"
#define BEETLE_OPT_MEMCARD_LEFT_INDEX_INFO_0_GA "Tá cárta cuimhne luchtaithe sa sliotán clé faoi láthair. Ní oibreoidh an rogha seo ach amháin má tá Modh Cárta Cuimhne socraithe go Mednafen."
#define OPTION_VAL_0_O48_GA "0 (Réamhshocrú)"
#define BEETLE_OPT_MEMCARD_RIGHT_INDEX_LABEL_GA "Sliotán Cárta Cuimhne 2 Innéacs"
#define BEETLE_OPT_MEMCARD_RIGHT_INDEX_INFO_0_GA "Tá cárta cuimhne luchtaithe sa sliotán ceart faoi láthair. Ní oibreoidh an rogha seo ach amháin má tá Sliotán Cárta Cuimhne 2 cumasaithe."
#define OPTION_VAL_1_O49_GA "1 (Réamhshocrú)"
#define BEETLE_OPT_LINE_RENDER_LABEL_GA "Haic Líne-go-Ceathair"
#define BEETLE_OPT_LINE_RENDER_INFO_0_GA "Tarraingíonn roinnt cluichí (e.g. Doom, Hexen, Soul Blade, srl.) línte cothrománacha trí thriantáin aon-phicteilín ar airde a shíneadh trasna an scáileáin, atá rastraithe mar shraith picteilíní ar an gcrua-earraí bunaidh. Braitheann an haca seo na triantáin bheaga seo agus athraíonn sé iad go ceathairshníomhacha de réir mar is gá, rud a ligeann dóibh a bheith taispeánta i gceart ar na rindreálaithe crua-earraí agus ag taifeach inmheánach uasghrádaithe. Tá 'Ionsaitheach' riachtanach do roinnt teidil (m.sh. Dark Forces, Duke Nukem) ach d'fhéadfadh sé fabhtanna grafacha a thabhairt isteach murach sin. Fág ag 'Réamhshocrú' mura bhfuil tú cinnte."
#define OPTION_VAL_DEFAULT_GA "Réamhshocrú"
#define OPTION_VAL_AGGRESSIVE_GA "Ionsaitheach"
#define BEETLE_OPT_WIDESCREEN_HACK_LABEL_GA "Haic Mód Leathan-Scáileáin"
#define BEETLE_OPT_WIDESCREEN_HACK_INFO_0_GA "Déan ábhar 3T a rindreáil go hanamorfach agus aschur an maolán fráma aithrisithe ag cóimheas gné scáileáin leathan. Tugann sé na torthaí is fearr le cluichí lán-3T. Beidh eilimintí 2T sínte go cothrománach agus d'fhéadfadh siad a bheith mí-ailínithe."
#define BEETLE_OPT_WIDESCREEN_HACK_ASPECT_RATIO_LABEL_GA "Cóimheas Gné Haic Mód Leathan-Scáileáin"
#define BEETLE_OPT_WIDESCREEN_HACK_ASPECT_RATIO_INFO_0_GA "Cóimheas gné le húsáid ag an Hack Mód Scáileáin Leathan."
#define OPTION_VAL_16_9_GA NULL
#define OPTION_VAL_16_10_GA NULL
#define OPTION_VAL_18_9_GA NULL
#define OPTION_VAL_19_9_GA NULL
#define OPTION_VAL_20_9_GA NULL
#define OPTION_VAL_21_9_GA NULL
#define OPTION_VAL_32_9_GA NULL
#define BEETLE_OPT_CPU_DYNAREC_LABEL_GA "LAP Dynarec"
#define BEETLE_OPT_CPU_DYNAREC_INFO_0_GA "Athchóimeáil treoracha LAP go dinimiciúil chuig treoracha dúchasacha. I bhfad níos tapúla ná an léirmhínitheoir, ach níl an t-amchlár LAP chomh cruinn céanna, agus d'fhéadfadh fabhtanna a bheith ann."
#define OPTION_VAL_DISABLED_GA "Míchumasaithe (Ateangaire Ciaróg)"
#define OPTION_VAL_EXECUTE_GA "Uasfheidhmíocht"
#define OPTION_VAL_EXECUTE_ONE_GA "Seiceáil Ama Timthrialla"
#define OPTION_VAL_RUN_INTERPRETER_GA "Ateangaire Lightrec"
#define BEETLE_OPT_DYNAREC_INVALIDATE_LABEL_GA "Neamhbhailíochtú Cód Dynarec"
#define BEETLE_OPT_DYNAREC_INVALIDATE_INFO_0_GA "Éilíonn roinnt cluichí neamhbhailíochtú 'Iomlán', agus teastaíonn 'DMA Amháin' ó chuid eile."
#define OPTION_VAL_FULL_GA "Lán"
#define OPTION_VAL_DMA_GA "DMA Amháin (Beagán níos Tapúla)"
#define BEETLE_OPT_DYNAREC_EVENTCYCLES_LABEL_GA "Timthriallta Imeachta Dynarec DMA/GPU/MDEC/Amadóir"
#define BEETLE_OPT_DYNAREC_EVENTCYCLES_INFO_0_GA "Uasmhéid timthriallta a ritheann an LAP sula ndéantar Nuashonrú GPU/DMA/MDEC/Amaire a sheiceáil, beidh uimhir níos airde níos tapúla, agus bíonn tionchar i bhfad níos lú aige ar léirmhínitheoir beetle ná ar dynarec."
#define OPTION_VAL_128_GA "128 (Réamhshocrú)"
#define BEETLE_OPT_DYNAREC_SPU_SAMPLES_LABEL_GA "Samplaí SPU Dynarec"
#define BEETLE_OPT_DYNAREC_SPU_SAMPLES_INFO_0_GA "Uasmhéid samplaí SPU le rith sula ndéantar Nuashonrú SPU a sheiceáil, beidh uimhir níos airde níos tapúla, ach beidh teipeanna fuaime i roinnt cluichí le rud ar bith seachas 1."
#define BEETLE_OPT_CORE_TIMING_FPS_LABEL_GA "Amchlár FPS Tuairiscithe ag an gCroí"
#define BEETLE_OPT_CORE_TIMING_FPS_INFO_0_GA "Le Toggling Uathoibríoch is féidir leis an gcroílár aistriú idir rátaí forásacha agus idirnasctha."
#define OPTION_VAL_AUTO_TOGGLE_GA "Athrú Uathoibríoch"
#define OPTION_VAL_FORCE_PROGRESSIVE_GA "Ráta Forásach Fórsa"
#define OPTION_VAL_FORCE_INTERLACED_GA "Ráta Idirfhite Fórsa"
#define BEETLE_OPT_DEINTERLACER_LABEL_GA "Modh Dí-idirfhite"
#define BEETLE_OPT_DEINTERLACER_INFO_0_GA "Taispeánann 'Weave' an dá réimse ar an bhfráma céanna le haghaidh íomhá neamhghluaiste ardtaifigh iomlán, agus íobairtíonn 'Bob' soiléireacht taifigh ar mhaithe le soiléireacht gluaisne. Is é 'Bob (Offset)' Bob agus an réimse reatha dúbailte isteach san airde dromchla iomlán le fritháireamh aonlíne idir réimsí, ag caomhnú taifeach aschuir an libretro agus gluaiseacht réidh stíl Bob á choinneáil. Fíonn 'FastMAD (Motion Adaptive)' picteilíní neamhghluaiste agus idirshuíonn sé picteilíní gluaisteacha go hingearach in aghaidh an phicteilín, airde aschuir iomlán gan cíor an fhí ar ghluaiseacht ná taifeach leathnaithe Bob. Déanann 'Off' an rindreálaí bogearraí a mheaitseáil le hiompar an rindreálaí crua-earraí (Vulkan/OpenGL) trí rastar a dhéanamh chuig gach líne VRAM i ngach fráma agus scanadh amach a chur siar go dtí deireadh an fhráma, rud a chuireann deireadh le cíor-ar-ghluaiseacht i gcluichí idirnasctha. Tabhair faoi deara: féadfaidh cláir a dhéanann DMA go VRAM le linn scanadh amach (roinnt taispeántóirí ag baint úsáide as cleasanna fillte maolán fráma PAL >512 líne), nó a dhéanann nuashonrú ar sprites gluaisteacha trí aistrithe íomhá LAP-go-VRAM atá ceangailte leis an réimse taispeánta (e.g. Pro Pinball), sraitheanna seanchaite nó taibhsí sprites a thaispeáint faoi 'As'; bain úsáid as 'Weave', 'Bob', 'Bob (Offset)', nó 'FastMAD' dóibh siúd."
#define OPTION_VAL_WEAVE_GA "Fí"
#define OPTION_VAL_BOB_GA NULL
#define OPTION_VAL_BOB_OFFSET_GA "Bob (Fritháireamh)"
#define OPTION_VAL_FASTMAD_GA "FastMAD (Oiriúnaitheach Gluaiseachta)"
#define BEETLE_OPT_FRAME_DUPING_LABEL_GA "Dupáil Fráma"
#define BEETLE_OPT_FRAME_DUPING_INFO_0_GA "Déan an fráma roimhe seo arís mura bhfuil aon rud nua le taispeáint ag an gcroílár."
#define BEETLE_OPT_PAL_VIDEO_TIMING_OVERRIDE_LABEL_GA "Sáraigh Ama Físe PAL"
#define BEETLE_OPT_PAL_VIDEO_TIMING_OVERRIDE_INFO_0_GA "Mar gheall ar chaighdeáin éagsúla, is minic a bhíonn cuma mhoillithe ar chluichí PAL i gcomparáid leis na heisiúintí NTSC Meiriceánacha nó Seapánacha. Is féidir an rogha seo a úsáid chun amanna PAL a shárú d’fhonn iarracht a dhéanamh na cluichí seo a rith leis an ráta fráma NTSC."
#define BEETLE_OPT_ASPECT_RATIO_LABEL_GA "Cóimheas Gné Croí"
#define BEETLE_OPT_ASPECT_RATIO_INFO_0_GA "Déantar neamhaird den socrú seo nuair a bhíonn na roghanna Widescreen Mode Hack nó Display Full VRAM cumasaithe."
#define OPTION_VAL_CORRECTED_GA "Ceartaithe"
#define OPTION_VAL_UNCORRECTED_GA "Gan cheartú"
#define OPTION_VAL_4_3_GA "Fórsa 4:3"
#define OPTION_VAL_NTSC_GA "Fórsa NTSC"
#define BEETLE_OPT_CROP_OVERSCAN_LABEL_GA "Overscan Barraí"
#define BEETLE_OPT_CROP_OVERSCAN_INFO_0_GA "Íomhá dhúchasach iomlán nó gan teorainneacha."
#define OPTION_VAL_STATIC_GA "Cothrománach"
#define OPTION_VAL_SMART_GA "Cothrománach + Ingearach"
#define BEETLE_OPT_IMAGE_CROP_LABEL_GA "Bearradh Breise"
#define BEETLE_OPT_IMAGE_CROP_INFO_0_GA "Laghdaigh leithead na híomhá bearrtha faoin líon picteilíní sonraithe."
#define OPTION_VAL_9PX_GA NULL
#define OPTION_VAL_10PX_GA NULL
#define OPTION_VAL_11PX_GA NULL
#define OPTION_VAL_12PX_GA NULL
#define OPTION_VAL_13PX_GA NULL
#define OPTION_VAL_14PX_GA NULL
#define OPTION_VAL_15PX_GA NULL
#define OPTION_VAL_16PX_GA NULL
#define OPTION_VAL_17PX_GA NULL
#define OPTION_VAL_18PX_GA NULL
#define OPTION_VAL_19PX_GA NULL
#define OPTION_VAL_20PX_GA NULL
#define BEETLE_OPT_IMAGE_OFFSET_LABEL_GA "Íomhá Bearrtha Fritháireamh"
#define BEETLE_OPT_IMAGE_OFFSET_INFO_0_GA "Fritháireamh go cothrománach ar dheis (dearfach) nó ar chlé (diúltach) faoin líon picteilíní sonraithe. Féadfar é seo a úsáid chun fadhbanna ailínithe a cheartú. Ní thacaítear leis ach amháin ag an rindreálaí bogearraí."
#define OPTION_VAL_12PX_O64_GA NULL
#define OPTION_VAL_11PX_O64_GA NULL
#define OPTION_VAL_10PX_O64_GA NULL
#define OPTION_VAL_9PX_O64_GA NULL
#define OPTION_VAL_8PX_O64_GA NULL
#define OPTION_VAL_7PX_O64_GA NULL
#define OPTION_VAL_6PX_O64_GA NULL
#define OPTION_VAL_5PX_O64_GA NULL
#define OPTION_VAL_4PX_O64_GA NULL
#define OPTION_VAL_3PX_O64_GA NULL
#define OPTION_VAL_2PX_O64_GA NULL
#define OPTION_VAL_1PX_O64_GA NULL
#define OPTION_VAL_1PX_O64_0_GA NULL
#define OPTION_VAL_2PX_O64_0_GA NULL
#define OPTION_VAL_3PX_O64_0_GA NULL
#define OPTION_VAL_4PX_O64_0_GA NULL
#define OPTION_VAL_5PX_O64_0_GA NULL
#define OPTION_VAL_6PX_O64_0_GA NULL
#define OPTION_VAL_7PX_O64_0_GA NULL
#define OPTION_VAL_8PX_O64_0_GA NULL
#define OPTION_VAL_9PX_O64_0_GA NULL
#define OPTION_VAL_10PX_O64_0_GA NULL
#define OPTION_VAL_11PX_O64_0_GA NULL
#define OPTION_VAL_12PX_O64_0_GA NULL
#define BEETLE_OPT_IMAGE_OFFSET_CYCLES_LABEL_GA "Fritháireamh Íomhá Cothrománach (Timthriallta GPU)"
#define BEETLE_OPT_IMAGE_OFFSET_CYCLES_INFO_0_GA "Bogann luachanna dearfacha an íomhá ar dheis, bogann luachanna diúltacha an íomhá ar chlé. Ní thacaítear leis seo ach amháin ag na rindreálaithe crua-earraí."
#define BEETLE_OPT_INITIAL_SCANLINE_LABEL_GA "Líne Scanadh Tosaigh - NTSC"
#define BEETLE_OPT_INITIAL_SCANLINE_INFO_0_GA "An chéad líne scanadh a thaispeántar agus ábhar NTSC á rith. Laghdóidh luachanna níos mó ná náid airde na n-íomhánna aschuir trí phicteilíní a bhearradh ón imeall is airde. Féadfar é a úsáid chun dul i ngleic le boscaí litreacha."
#define BEETLE_OPT_LAST_SCANLINE_LABEL_GA "Líne Scanadh Deiridh - NTSC"
#define BEETLE_OPT_LAST_SCANLINE_INFO_0_GA "An líne scanadh dheireanach a thaispeántar agus ábhar NTSC á rith. Laghdóidh luachanna níos lú ná 239 airde na n-íomhánna aschuir trí phicteilíní a bhearradh ón imeall is ísle. Féadfar é a úsáid chun dul i ngleic le boscaí litreacha."
#define OPTION_VAL_239_GA "239 (Réamhshocrú)"
#define BEETLE_OPT_INITIAL_SCANLINE_PAL_LABEL_GA "Líne Scanadh Tosaigh - PAL"
#define BEETLE_OPT_INITIAL_SCANLINE_PAL_INFO_0_GA "An chéad líne scanadh a thaispeántar agus ábhar PAL á rith. Laghdóidh luachanna níos mó ná náid airde na n-íomhánna aschuir trí phicteilíní a bhearradh ón imeall is airde. Féadfar é a úsáid chun dul i ngleic le boscaí litreacha."
#define BEETLE_OPT_LAST_SCANLINE_PAL_LABEL_GA "Líne Scanadh Deiridh - PAL"
#define BEETLE_OPT_LAST_SCANLINE_PAL_INFO_0_GA "An líne scanadh dheireanach a thaispeántar agus ábhar PAL á rith. Laghdóidh luachanna faoi bhun 287 airde na n-íomhánna aschuir trí phicteilíní a bhearradh ón imeall is ísle. Féadfar é a úsáid chun dul i ngleic le boscaí litreacha."
#define OPTION_VAL_287_GA "287 (Réamhshocrú)"

struct retro_core_option_v2_category option_cats_ga[] = {
   {
      "system",
      CATEGORY_SYSTEM_LABEL_GA,
      CATEGORY_SYSTEM_INFO_0_GA
   },
   {
      "video",
      CATEGORY_VIDEO_LABEL_GA,
      CATEGORY_VIDEO_INFO_0_GA
   },
   {
      "pgxp",
      CATEGORY_PGXP_LABEL_GA,
      CATEGORY_PGXP_INFO_0_GA
   },
   {
      "osd",
      CATEGORY_OSD_LABEL_GA,
      CATEGORY_OSD_INFO_0_GA
   },
   {
      "input",
      CATEGORY_INPUT_LABEL_GA,
      CATEGORY_INPUT_INFO_0_GA
   },
   {
      "memcards",
      CATEGORY_MEMCARDS_LABEL_GA,
      CATEGORY_MEMCARDS_INFO_0_GA
   },
   {
      "hacks",
      CATEGORY_HACKS_LABEL_GA,
      CATEGORY_HACKS_INFO_0_GA
   },
   { NULL, NULL, NULL },
};
struct retro_core_option_v2_definition option_defs_ga[] = {
   {
      BEETLE_OPT(cpu_freq_scale),
      BEETLE_OPT_CPU_FREQ_SCALE_LABEL_GA,
      NULL,
      BEETLE_OPT_CPU_FREQ_SCALE_INFO_0_GA,
      NULL,
      "system",
      {
         { "50%",           OPTION_VAL_50_GA },
         { "60%",           OPTION_VAL_60_GA },
         { "70%",           OPTION_VAL_70_GA },
         { "80%",           OPTION_VAL_80_GA },
         { "90%",           OPTION_VAL_90_GA },
         { "100%",          OPTION_VAL_100_GA },
         { "110%",          OPTION_VAL_110_GA },
         { "120%",          OPTION_VAL_120_GA },
         { "130%",          OPTION_VAL_130_GA },
         { "140%",          OPTION_VAL_140_GA },
         { "150%",          OPTION_VAL_150_GA },
         { "160%",          OPTION_VAL_160_GA },
         { "170%",          OPTION_VAL_170_GA },
         { "180%",          OPTION_VAL_180_GA },
         { "190%",          OPTION_VAL_190_GA },
         { "200%",          OPTION_VAL_200_GA },
         { "210%",          OPTION_VAL_210_GA },
         { "220%",          OPTION_VAL_220_GA },
         { "230%",          OPTION_VAL_230_GA },
         { "240%",          OPTION_VAL_240_GA },
         { "250%",          OPTION_VAL_250_GA },
         { "260%",          OPTION_VAL_260_GA },
         { "270%",          OPTION_VAL_270_GA },
         { "280%",          OPTION_VAL_280_GA },
         { "290%",          OPTION_VAL_290_GA },
         { "300%",          OPTION_VAL_300_GA },
         { "310%",          OPTION_VAL_310_GA },
         { "320%",          OPTION_VAL_320_GA },
         { "330%",          OPTION_VAL_330_GA },
         { "340%",          OPTION_VAL_340_GA },
         { "350%",          OPTION_VAL_350_GA },
         { "360%",          OPTION_VAL_360_GA },
         { "370%",          OPTION_VAL_370_GA },
         { "380%",          OPTION_VAL_380_GA },
         { "390%",          OPTION_VAL_390_GA },
         { "400%",          OPTION_VAL_400_GA },
         { "410%",          OPTION_VAL_410_GA },
         { "420%",          OPTION_VAL_420_GA },
         { "430%",          OPTION_VAL_430_GA },
         { "440%",          OPTION_VAL_440_GA },
         { "450%",          OPTION_VAL_450_GA },
         { "460%",          OPTION_VAL_460_GA },
         { "470%",          OPTION_VAL_470_GA },
         { "480%",          OPTION_VAL_480_GA },
         { "490%",          OPTION_VAL_490_GA },
         { "500%",          OPTION_VAL_500_GA },
         { "510%",          OPTION_VAL_510_GA },
         { "520%",          OPTION_VAL_520_GA },
         { "530%",          OPTION_VAL_530_GA },
         { "540%",          OPTION_VAL_540_GA },
         { "550%",          OPTION_VAL_550_GA },
         { "560%",          OPTION_VAL_560_GA },
         { "570%",          OPTION_VAL_570_GA },
         { "580%",          OPTION_VAL_580_GA },
         { "590%",          OPTION_VAL_590_GA },
         { "600%",          OPTION_VAL_600_GA },
         { "610%",          OPTION_VAL_610_GA },
         { "620%",          OPTION_VAL_620_GA },
         { "630%",          OPTION_VAL_630_GA },
         { "640%",          OPTION_VAL_640_GA },
         { "650%",          OPTION_VAL_650_GA },
         { "660%",          OPTION_VAL_660_GA },
         { "670%",          OPTION_VAL_670_GA },
         { "680%",          OPTION_VAL_680_GA },
         { "690%",          OPTION_VAL_690_GA },
         { "700%",          OPTION_VAL_700_GA },
         { "710%",          OPTION_VAL_710_GA },
         { "720%",          OPTION_VAL_720_GA },
         { "730%",          OPTION_VAL_730_GA },
         { "740%",          OPTION_VAL_740_GA },
         { "750%",          OPTION_VAL_750_GA },
         { NULL, NULL },
      },
      "100%"
   },
   {
      BEETLE_OPT(gpu_overclock),
      BEETLE_OPT_GPU_OVERCLOCK_LABEL_GA,
      NULL,
      BEETLE_OPT_GPU_OVERCLOCK_INFO_0_GA,
      NULL,
      "system",
      {
         { "1x(native)", OPTION_VAL_1X_NATIVE_GA },
         { "2x",         OPTION_VAL_2X_GA },
         { "4x",         OPTION_VAL_4X_GA },
         { "8x",         OPTION_VAL_8X_GA },
         { "16x",        OPTION_VAL_16X_GA },
         { "32x",        OPTION_VAL_32X_GA },
         { NULL, NULL },
      },
      "1x(native)"
   },
   {
      BEETLE_OPT(gte_overclock),
      BEETLE_OPT_GTE_OVERCLOCK_LABEL_GA,
      NULL,
      BEETLE_OPT_GTE_OVERCLOCK_INFO_0_GA,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
#ifndef EMSCRIPTEN
   {
      BEETLE_OPT(cd_access_method),
      BEETLE_OPT_CD_ACCESS_METHOD_LABEL_GA,
      NULL,
      BEETLE_OPT_CD_ACCESS_METHOD_INFO_0_GA,
      NULL,
      "system",
      {
         { "sync",     OPTION_VAL_SYNC_GA },
         { "async",    OPTION_VAL_ASYNC_GA },
         { "precache", OPTION_VAL_PRECACHE_GA },
         { NULL, NULL },
      },
      "sync"
   },
#endif
   {
      BEETLE_OPT(cd_fastload),
      BEETLE_OPT_CD_FASTLOAD_LABEL_GA,
      NULL,
      BEETLE_OPT_CD_FASTLOAD_INFO_0_GA,
      NULL,
      "system",
      {
         { "2x(native)", OPTION_VAL_2X_NATIVE_GA },
         { "4x",          OPTION_VAL_4X_GA },
         { "6x",          OPTION_VAL_6X_GA },
         { "8x",          OPTION_VAL_8X_GA },
         { "10x",         OPTION_VAL_10X_GA },
         { "12x",         OPTION_VAL_12X_GA },
         { "14x",         OPTION_VAL_14X_GA },
         { NULL, NULL },
      },
      "2x(native)"
   },
   {
      BEETLE_OPT(skip_bios),
      BEETLE_OPT_SKIP_BIOS_LABEL_GA,
      NULL,
      BEETLE_OPT_SKIP_BIOS_INFO_0_GA,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(override_bios),
      BEETLE_OPT_OVERRIDE_BIOS_LABEL_GA,
      NULL,
      BEETLE_OPT_OVERRIDE_BIOS_INFO_0_GA,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "psxonpsp", OPTION_VAL_PSXONPSP_GA },
         { "ps1_rom",  OPTION_VAL_PS1_ROM_GA },
         { "openbios", OPTION_VAL_OPENBIOS_GA     },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(region),
      BEETLE_OPT_REGION_LABEL_GA,
      NULL,
      BEETLE_OPT_REGION_INFO_0_GA,
      NULL,
      "system",
      {
         { "auto",   OPTION_VAL_AUTO_GA },
         { "ntsc-j", OPTION_VAL_NTSC_J_GA },
         { "ntsc-u", OPTION_VAL_NTSC_U_GA },
         { "pal",    OPTION_VAL_PAL_GA },
         { NULL, NULL },
      },
      "auto"
   },
   {
      BEETLE_OPT(internal_resolution),
      BEETLE_OPT_INTERNAL_RESOLUTION_LABEL_GA,
      NULL,
      BEETLE_OPT_INTERNAL_RESOLUTION_INFO_0_GA,
      NULL,
      "video",
      {
         { "1x(native)", OPTION_VAL_1X_NATIVE_GA },
         { "2x",         OPTION_VAL_2X_GA },
         { "4x",         OPTION_VAL_4X_GA },
         { "8x",         OPTION_VAL_8X_GA },
         { "16x",        OPTION_VAL_16X_GA },
         { NULL, NULL },
      },
      "1x(native)"
   },
#if defined(HAVE_OPENGL) || defined(HAVE_OPENGLES) || defined(HAVE_VULKAN)
   {
      BEETLE_OPT(renderer),
      BEETLE_OPT_RENDERER_LABEL_GA,
      NULL,
      BEETLE_OPT_RENDERER_INFO_0_GA,
      NULL,
      "video",
      {
         { "hardware",    OPTION_VAL_HARDWARE_GA },
#if defined(HAVE_OPENGL) || defined(HAVE_OPENGLES)
         { "hardware_gl", OPTION_VAL_HARDWARE_GL_GA },
#endif
#if defined(HAVE_VULKAN)
         { "hardware_vk", OPTION_VAL_HARDWARE_VK_GA },
#endif
         { "software",    OPTION_VAL_SOFTWARE_GA },
         { NULL, NULL },
      },
      "hardware"
   },
   {
      BEETLE_OPT(renderer_software_fb),
      BEETLE_OPT_RENDERER_SOFTWARE_FB_LABEL_GA,
      NULL,
      BEETLE_OPT_RENDERER_SOFTWARE_FB_INFO_0_GA,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled"
   },
#endif
#if defined(HAVE_OPENGL) || defined(HAVE_OPENGLES)
   {
      BEETLE_OPT(depth),
      BEETLE_OPT_DEPTH_LABEL_GA,
      NULL,
      BEETLE_OPT_DEPTH_INFO_0_GA,
      NULL,
      "video",
      {
         { "16bpp(native)", OPTION_VAL_16BPP_NATIVE_GA },
         { "32bpp",         OPTION_VAL_32BPP_GA },
         { NULL, NULL },
      },
      "16bpp(native)"
   },
   // Sort of, it's more like 15-bit high color and 24-bit true color for visible output. The alpha channel is used for mask bit. Vulkan renderer uses ABGR1555_555 for 31 bits internal? FMVs are always 24-bit on all renderers like original hardware (BGR888, no alpha)
#endif
   {
      BEETLE_OPT(dither_mode),
      BEETLE_OPT_DITHER_MODE_LABEL_GA,
      NULL,
      BEETLE_OPT_DITHER_MODE_INFO_0_GA,
      NULL,
      "video",
      {
         { "1x(native)",          OPTION_VAL_1X_NATIVE_GA },
         { "internal resolution", OPTION_VAL_INTERNAL_RESOLUTION_GA },
         { "disabled",            NULL },
         { NULL, NULL },
      },
      "1x(native)"
   },
#ifdef HAVE_VULKAN
   {
      BEETLE_OPT(scaled_uv_offset),
      BEETLE_OPT_SCALED_UV_OFFSET_LABEL_GA,
      NULL,
      BEETLE_OPT_SCALED_UV_OFFSET_INFO_0_GA,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled"
   },
#endif
#if defined(HAVE_OPENGL) || defined(HAVE_OPENGLES) || defined(HAVE_VULKAN)
   {
      BEETLE_OPT(filter),
      BEETLE_OPT_FILTER_LABEL_GA,
      NULL,
      BEETLE_OPT_FILTER_INFO_0_GA,
      NULL,
      "video",
      {
         { "nearest",  OPTION_VAL_NEAREST_GA },
         { "SABR",     OPTION_VAL_SABR_GA },
         { "xBR",      OPTION_VAL_XBR_GA },
         { "bilinear", OPTION_VAL_BILINEAR_GA },
         { "3-point",  OPTION_VAL_3_POINT_GA },
         { "JINC2",    OPTION_VAL_JINC2_GA },
         { NULL, NULL },
      },
      "nearest"
   },
#ifdef HAVE_VULKAN
   {
      BEETLE_OPT(filter_exclude_sprite),
      BEETLE_OPT_FILTER_EXCLUDE_SPRITE_LABEL_GA,
      NULL,
      BEETLE_OPT_FILTER_EXCLUDE_SPRITE_INFO_0_GA,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "opaque", OPTION_VAL_OPAQUE_GA },
         { "all", OPTION_VAL_ALL_GA },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(filter_exclude_2d_polygon),
      BEETLE_OPT_FILTER_EXCLUDE_2D_POLYGON_LABEL_GA,
      NULL,
      BEETLE_OPT_FILTER_EXCLUDE_2D_POLYGON_INFO_0_GA,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "opaque", OPTION_VAL_OPAQUE_GA },
         { "all", OPTION_VAL_ALL_GA },
         { NULL, NULL },
      },
      "disabled"
   },
#endif
#endif
#ifdef HAVE_VULKAN
   {
      BEETLE_OPT(adaptive_smoothing),
      BEETLE_OPT_ADAPTIVE_SMOOTHING_LABEL_GA,
      NULL,
      BEETLE_OPT_ADAPTIVE_SMOOTHING_INFO_0_GA,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(super_sampling),
      BEETLE_OPT_SUPER_SAMPLING_LABEL_GA,
      NULL,
      BEETLE_OPT_SUPER_SAMPLING_INFO_0_GA,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(msaa),
      BEETLE_OPT_MSAA_LABEL_GA,
      NULL,
      BEETLE_OPT_MSAA_INFO_0_GA,
      NULL,
      "video",
      {
         { "1x",  OPTION_VAL_1X_GA },
         { "2x",  OPTION_VAL_2X_GA },
         { "4x",  OPTION_VAL_4X_GA },
         { "8x",  OPTION_VAL_8X_GA },
         { "16x", OPTION_VAL_16X_GA },
         { NULL, NULL },
      },
      "1x"
   },
   {
      BEETLE_OPT(mdec_yuv),
      BEETLE_OPT_MDEC_YUV_LABEL_GA,
      NULL,
      BEETLE_OPT_MDEC_YUV_INFO_0_GA,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(track_textures),
      BEETLE_OPT_TRACK_TEXTURES_LABEL_GA,
      NULL,
      BEETLE_OPT_TRACK_TEXTURES_INFO_0_GA,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
#ifdef TEXTURE_DUMPING_ENABLED
   {
      BEETLE_OPT(dump_textures),
      BEETLE_OPT_DUMP_TEXTURES_LABEL_GA,
      NULL,
      BEETLE_OPT_DUMP_TEXTURES_INFO_0_GA,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
#endif
   {
      BEETLE_OPT(replace_textures),
      BEETLE_OPT_REPLACE_TEXTURES_LABEL_GA,
      NULL,
      BEETLE_OPT_REPLACE_TEXTURES_INFO_0_GA,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
#endif
   {
      BEETLE_OPT(pgxp_mode),
      BEETLE_OPT_PGXP_MODE_LABEL_GA,
      NULL,
      BEETLE_OPT_PGXP_MODE_INFO_0_GA,
      NULL,
      "pgxp",
      {
         { "disabled",     NULL },
         { "memory only",  OPTION_VAL_MEMORY_ONLY_GA },
         { "memory + CPU", OPTION_VAL_MEMORY_CPU_GA },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(pgxp_2d_tol),
      BEETLE_OPT_PGXP_2D_TOL_LABEL_GA,
      NULL,
      BEETLE_OPT_PGXP_2D_TOL_INFO_0_GA,
      NULL,
      "pgxp",
      {
         { "disabled", NULL },
         { "0px", OPTION_VAL_0PX_GA },
         { "1px", OPTION_VAL_1PX_GA },
         { "2px", OPTION_VAL_2PX_GA },
         { "3px", OPTION_VAL_3PX_GA },
         { "4px", OPTION_VAL_4PX_GA },
         { "5px", OPTION_VAL_5PX_GA },
         { "6px", OPTION_VAL_6PX_GA },
         { "7px", OPTION_VAL_7PX_GA },
         { "8px", OPTION_VAL_8PX_GA },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(pgxp_nclip),
      BEETLE_OPT_PGXP_NCLIP_LABEL_GA,
      NULL,
      BEETLE_OPT_PGXP_NCLIP_INFO_0_GA,
      NULL,
      "pgxp",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
#if defined(HAVE_OPENGL) || defined(HAVE_OPENGLES) || defined(HAVE_VULKAN)
   {
      BEETLE_OPT(pgxp_vertex),
      BEETLE_OPT_PGXP_VERTEX_LABEL_GA,
      NULL,
      BEETLE_OPT_PGXP_VERTEX_INFO_0_GA,
      NULL,
      "pgxp",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
#endif
   {
      BEETLE_OPT(pgxp_texture),
      BEETLE_OPT_PGXP_TEXTURE_LABEL_GA,
      NULL,
      BEETLE_OPT_PGXP_TEXTURE_INFO_0_GA,
      NULL,
      "pgxp",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(display_internal_fps),
      BEETLE_OPT_DISPLAY_INTERNAL_FPS_LABEL_GA,
      NULL,
      BEETLE_OPT_DISPLAY_INTERNAL_FPS_INFO_0_GA,
      NULL,
      "osd",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(display_osd),
      BEETLE_OPT_DISPLAY_OSD_LABEL_GA,
      NULL,
      BEETLE_OPT_DISPLAY_OSD_INFO_0_GA,
      NULL,
      "osd",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled"
   },
#if defined(HAVE_OPENGL) || defined(HAVE_OPENGLES) || defined(HAVE_VULKAN)
   {
      BEETLE_OPT(display_vram),
      BEETLE_OPT_DISPLAY_VRAM_LABEL_GA,
      NULL,
      BEETLE_OPT_DISPLAY_VRAM_INFO_0_GA,
      NULL,
      "osd",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
#endif
   {
      BEETLE_OPT(analog_calibration),
      BEETLE_OPT_ANALOG_CALIBRATION_LABEL_GA,
      NULL,
      BEETLE_OPT_ANALOG_CALIBRATION_INFO_0_GA,
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(analog_toggle),
      BEETLE_OPT_ANALOG_TOGGLE_LABEL_GA,
      NULL,
      BEETLE_OPT_ANALOG_TOGGLE_INFO_0_GA,
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { "enabled-analog", OPTION_VAL_ENABLED_ANALOG_GA },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(analog_toggle_combo),
      BEETLE_OPT_ANALOG_TOGGLE_COMBO_LABEL_GA,
      NULL,
      BEETLE_OPT_ANALOG_TOGGLE_COMBO_INFO_0_GA,
      NULL,
      "input",
      {
         { "l1+l2+r1+r2+start+select", OPTION_VAL_L1_L2_R1_R2_START_SELECT_GA },
         { "l1+r1+select",             OPTION_VAL_L1_R1_SELECT_GA },
         { "l1+r1+start",              OPTION_VAL_L1_R1_START_GA },
         { "l1+r1+l3",                 OPTION_VAL_L1_R1_L3_GA },
         { "l1+r1+r3",                 OPTION_VAL_L1_R1_R3_GA },
         { "l2+r2+select",             OPTION_VAL_L2_R2_SELECT_GA },
         { "l2+r2+start",              OPTION_VAL_L2_R2_START_GA },
         { "l2+r2+l3",                 OPTION_VAL_L2_R2_L3_GA },
         { "l2+r2+r3",                 OPTION_VAL_L2_R2_R3_GA },
         { "l3+r3",                    OPTION_VAL_L3_R3_GA },
         { NULL, NULL },
      },
      "l1+r1+select"
   },
   {
      BEETLE_OPT(analog_toggle_hold),
      BEETLE_OPT_ANALOG_TOGGLE_HOLD_LABEL_GA,
      NULL,
      BEETLE_OPT_ANALOG_TOGGLE_HOLD_INFO_0_GA,
      NULL,
      "input",
      {
         { "0", OPTION_VAL_0_GA },
         { "1", OPTION_VAL_1_GA },
         { "2", OPTION_VAL_2_GA },
         { "3", OPTION_VAL_3_GA },
         { "4", OPTION_VAL_4_GA },
         { "5", OPTION_VAL_5_GA },
         { NULL, NULL },
      },
      "1"
   },
   {
      BEETLE_OPT(enable_multitap_port1),
      BEETLE_OPT_ENABLE_MULTITAP_PORT1_LABEL_GA,
      NULL,
      BEETLE_OPT_ENABLE_MULTITAP_PORT1_INFO_0_GA,
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(enable_multitap_port2),
      BEETLE_OPT_ENABLE_MULTITAP_PORT2_LABEL_GA,
      NULL,
      BEETLE_OPT_ENABLE_MULTITAP_PORT2_INFO_0_GA,
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(gun_input_mode),
      BEETLE_OPT_GUN_INPUT_MODE_LABEL_GA,
      NULL,
      BEETLE_OPT_GUN_INPUT_MODE_INFO_0_GA,
      NULL,
      "input",
      {
         { "lightgun",    OPTION_VAL_LIGHTGUN_GA },
         { "touchscreen", OPTION_VAL_TOUCHSCREEN_GA },
         { NULL, NULL },
      },
      "lightgun"
   },
   // Shouldn't the gun_input_mode just be Mouse vs. Touchscreen?
   {
      BEETLE_OPT(gun_cursor),
      BEETLE_OPT_GUN_CURSOR_LABEL_GA,
      NULL,
      BEETLE_OPT_GUN_CURSOR_INFO_0_GA,
      NULL,
      "input",
      {
         { "off",   "disabled" },
         { "cross", OPTION_VAL_CROSS_GA },
         { "dot",   OPTION_VAL_DOT_GA },
         { NULL, NULL },
      },
      "cross"
   },
   {
      BEETLE_OPT(crosshair_color_p1),
      BEETLE_OPT_CROSSHAIR_COLOR_P1_LABEL_GA,
      NULL,
      BEETLE_OPT_CROSSHAIR_COLOR_P1_INFO_0_GA,
      NULL,
      "input",
      {
         { "red",    OPTION_VAL_RED_GA },
         { "blue",   OPTION_VAL_BLUE_GA },
         { "green",  OPTION_VAL_GREEN_GA },
         { "orange", OPTION_VAL_ORANGE_GA },
         { "yellow", OPTION_VAL_YELLOW_GA },
         { "cyan",   OPTION_VAL_CYAN_GA },
         { "pink",   OPTION_VAL_PINK_GA },
         { "purple", OPTION_VAL_PURPLE_GA },
         { "black",  OPTION_VAL_BLACK_GA },
         { "white",  OPTION_VAL_WHITE_GA },
         { NULL, NULL },
      },
      "red"
   },
   {
      BEETLE_OPT(crosshair_color_p2),
      BEETLE_OPT_CROSSHAIR_COLOR_P2_LABEL_GA,
      NULL,
      BEETLE_OPT_CROSSHAIR_COLOR_P2_INFO_0_GA,
      NULL,
      "input",
      {
         { "blue",   OPTION_VAL_BLUE_GA },
         { "red",    OPTION_VAL_RED_GA },
         { "green",  OPTION_VAL_GREEN_GA },
         { "orange", OPTION_VAL_ORANGE_GA },
         { "yellow", OPTION_VAL_YELLOW_GA },
         { "cyan",   OPTION_VAL_CYAN_GA },
         { "pink",   OPTION_VAL_PINK_GA },
         { "purple", OPTION_VAL_PURPLE_GA },
         { "black",  OPTION_VAL_BLACK_GA },
         { "white",  OPTION_VAL_WHITE_GA },
         { NULL, NULL },
      },
      "blue"
   },
   {
      BEETLE_OPT(mouse_sensitivity),
      BEETLE_OPT_MOUSE_SENSITIVITY_LABEL_GA,
      NULL,
      BEETLE_OPT_MOUSE_SENSITIVITY_INFO_0_GA,
      NULL,
      "input",
      {
         { "5%",   OPTION_VAL_5_O42_GA },
         { "10%",  OPTION_VAL_10_GA },
         { "15%",  OPTION_VAL_15_GA },
         { "20%",  OPTION_VAL_20_GA },
         { "25%",  OPTION_VAL_25_GA },
         { "30%",  OPTION_VAL_30_GA },
         { "35%",  OPTION_VAL_35_GA },
         { "40%",  OPTION_VAL_40_GA },
         { "45%",  OPTION_VAL_45_GA },
         { "50%",  OPTION_VAL_50_GA },
         { "55%",  OPTION_VAL_55_GA },
         { "60%",  OPTION_VAL_60_GA },
         { "65%",  OPTION_VAL_65_GA },
         { "70%",  OPTION_VAL_70_GA },
         { "75%",  OPTION_VAL_75_GA },
         { "80%",  OPTION_VAL_80_GA },
         { "85%",  OPTION_VAL_85_GA },
         { "90%",  OPTION_VAL_90_GA },
         { "95%",  OPTION_VAL_95_GA },
         { "100%", OPTION_VAL_100_O42_GA },
         { "105%", OPTION_VAL_105_GA },
         { "110%", OPTION_VAL_110_GA },
         { "115%", OPTION_VAL_115_GA },
         { "120%", OPTION_VAL_120_GA },
         { "125%", OPTION_VAL_125_GA },
         { "130%", OPTION_VAL_130_GA },
         { "135%", OPTION_VAL_135_GA },
         { "140%", OPTION_VAL_140_GA },
         { "145%", OPTION_VAL_145_GA },
         { "150%", OPTION_VAL_150_GA },
         { "155%", OPTION_VAL_155_GA },
         { "160%", OPTION_VAL_160_GA },
         { "165%", OPTION_VAL_165_GA },
         { "170%", OPTION_VAL_170_GA },
         { "175%", OPTION_VAL_175_GA },
         { "180%", OPTION_VAL_180_GA },
         { "185%", OPTION_VAL_185_GA },
         { "190%", OPTION_VAL_190_GA },
         { "195%", OPTION_VAL_195_GA },
         { "200%", OPTION_VAL_200_GA },
         { NULL, NULL },
      },
      "100%"
   },
   {
      BEETLE_OPT(negcon_response),
      BEETLE_OPT_NEGCON_RESPONSE_LABEL_GA,
      NULL,
      BEETLE_OPT_NEGCON_RESPONSE_INFO_0_GA,
      NULL,
      "input",
      {
         { "linear",    OPTION_VAL_LINEAR_GA },
         { "quadratic", OPTION_VAL_QUADRATIC_GA },
         { "cubic",     OPTION_VAL_CUBIC_GA },
         { NULL, NULL },
      },
      "linear"
   },
   {
      BEETLE_OPT(negcon_deadzone),
      BEETLE_OPT_NEGCON_DEADZONE_LABEL_GA,
      NULL,
      BEETLE_OPT_NEGCON_DEADZONE_INFO_0_GA,
      NULL,
      "input",
      {
         { "0%",  OPTION_VAL_0_O44_GA },
         { "5%",  OPTION_VAL_5_O42_GA },
         { "10%", OPTION_VAL_10_GA },
         { "15%", OPTION_VAL_15_GA },
         { "20%", OPTION_VAL_20_GA },
         { "25%", OPTION_VAL_25_GA },
         { "30%", OPTION_VAL_30_GA },
         { NULL, NULL },
      },
      "0%"
   },
   {
      BEETLE_OPT(use_mednafen_memcard0_method),
      BEETLE_OPT_USE_MEDNAFEN_MEMCARD0_METHOD_LABEL_GA,
      NULL,
      BEETLE_OPT_USE_MEDNAFEN_MEMCARD0_METHOD_INFO_0_GA,
      NULL,
      "memcards",
      {
         { "libretro", OPTION_VAL_LIBRETRO_GA },
         { "mednafen", OPTION_VAL_MEDNAFEN_GA },
         { NULL, NULL },
      },
      "libretro"
   },
   {
      BEETLE_OPT(enable_memcard1),
      BEETLE_OPT_ENABLE_MEMCARD1_LABEL_GA,
      NULL,
      BEETLE_OPT_ENABLE_MEMCARD1_INFO_0_GA,
      NULL,
      "memcards",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(shared_memory_cards),
      BEETLE_OPT_SHARED_MEMORY_CARDS_LABEL_GA,
      NULL,
      BEETLE_OPT_SHARED_MEMORY_CARDS_INFO_0_GA,
      NULL,
      "memcards",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(memcard_left_index),
      BEETLE_OPT_MEMCARD_LEFT_INDEX_LABEL_GA,
      NULL,
      BEETLE_OPT_MEMCARD_LEFT_INDEX_INFO_0_GA,
      NULL,
      "memcards",
      {
         { "0",  OPTION_VAL_0_O48_GA },
         { "1",  NULL },
         { "2",  NULL },
         { "3",  NULL },
         { "4",  NULL },
         { "5",  NULL },
         { "6",  NULL },
         { "7",  NULL },
         { "8",  NULL },
         { "9",  NULL },
         { "10", NULL },
         { "11", NULL },
         { "12", NULL },
         { "13", NULL },
         { "14", NULL },
         { "15", NULL },
         { "16", NULL },
         { "17", NULL },
         { "18", NULL },
         { "19", NULL },
         { "20", NULL },
         { "21", NULL },
         { "22", NULL },
         { "23", NULL },
         { "24", NULL },
         { "25", NULL },
         { "26", NULL },
         { "27", NULL },
         { "28", NULL },
         { "29", NULL },
         { "30", NULL },
         { "31", NULL },
         { "32", NULL },
         { "33", NULL },
         { "34", NULL },
         { "35", NULL },
         { "36", NULL },
         { "37", NULL },
         { "38", NULL },
         { "39", NULL },
         { "40", NULL },
         { "41", NULL },
         { "42", NULL },
         { "43", NULL },
         { "44", NULL },
         { "45", NULL },
         { "46", NULL },
         { "47", NULL },
         { "48", NULL },
         { "49", NULL },
         { "50", NULL },
         { "51", NULL },
         { "52", NULL },
         { "53", NULL },
         { "54", NULL },
         { "55", NULL },
         { "56", NULL },
         { "57", NULL },
         { "58", NULL },
         { "59", NULL },
         { "60", NULL },
         { "61", NULL },
         { "62", NULL },
         { "63", NULL },
         { NULL, NULL },
      },
      "0"
   },
   {
      BEETLE_OPT(memcard_right_index),
      BEETLE_OPT_MEMCARD_RIGHT_INDEX_LABEL_GA,
      NULL,
      BEETLE_OPT_MEMCARD_RIGHT_INDEX_INFO_0_GA,
      NULL,
      "memcards",
      {
         { "0",  NULL },
         { "1",  OPTION_VAL_1_O49_GA },
         { "2",  NULL },
         { "3",  NULL },
         { "4",  NULL },
         { "5",  NULL },
         { "6",  NULL },
         { "7",  NULL },
         { "8",  NULL },
         { "9",  NULL },
         { "10", NULL },
         { "11", NULL },
         { "12", NULL },
         { "13", NULL },
         { "14", NULL },
         { "15", NULL },
         { "16", NULL },
         { "17", NULL },
         { "18", NULL },
         { "19", NULL },
         { "20", NULL },
         { "21", NULL },
         { "22", NULL },
         { "23", NULL },
         { "24", NULL },
         { "25", NULL },
         { "26", NULL },
         { "27", NULL },
         { "28", NULL },
         { "29", NULL },
         { "30", NULL },
         { "31", NULL },
         { "32", NULL },
         { "33", NULL },
         { "34", NULL },
         { "35", NULL },
         { "36", NULL },
         { "37", NULL },
         { "38", NULL },
         { "39", NULL },
         { "40", NULL },
         { "41", NULL },
         { "42", NULL },
         { "43", NULL },
         { "44", NULL },
         { "45", NULL },
         { "46", NULL },
         { "47", NULL },
         { "48", NULL },
         { "49", NULL },
         { "50", NULL },
         { "51", NULL },
         { "52", NULL },
         { "53", NULL },
         { "54", NULL },
         { "55", NULL },
         { "56", NULL },
         { "57", NULL },
         { "58", NULL },
         { "59", NULL },
         { "60", NULL },
         { "61", NULL },
         { "62", NULL },
         { "63", NULL },
         { NULL, NULL },
      },
      "1"
   },
   {
      BEETLE_OPT(line_render),
      BEETLE_OPT_LINE_RENDER_LABEL_GA,
      NULL,
      BEETLE_OPT_LINE_RENDER_INFO_0_GA,
      NULL,
      "hacks",
      {
         { "default",    OPTION_VAL_DEFAULT_GA },
         { "aggressive", OPTION_VAL_AGGRESSIVE_GA },
         { "disabled",   NULL },
         { NULL, NULL },
      },
      "default"
   },
   {
      BEETLE_OPT(widescreen_hack),
      BEETLE_OPT_WIDESCREEN_HACK_LABEL_GA,
      NULL,
      BEETLE_OPT_WIDESCREEN_HACK_INFO_0_GA,
      NULL,
      "hacks",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(widescreen_hack_aspect_ratio),
      BEETLE_OPT_WIDESCREEN_HACK_ASPECT_RATIO_LABEL_GA,
      NULL,
      BEETLE_OPT_WIDESCREEN_HACK_ASPECT_RATIO_INFO_0_GA,
      NULL,
      "hacks",
      {
         { "16:9",  OPTION_VAL_16_9_GA },
         { "16:10", OPTION_VAL_16_10_GA },
         { "18:9",  OPTION_VAL_18_9_GA },
         { "19:9",  OPTION_VAL_19_9_GA },
         { "20:9",  OPTION_VAL_20_9_GA },
         { "21:9",  OPTION_VAL_21_9_GA }, // 64:27
         { "32:9",  OPTION_VAL_32_9_GA },
         { NULL,    NULL },
      },
      "16:9"
   },
#if defined(HAVE_LIGHTREC)
   {
      BEETLE_OPT(cpu_dynarec),
      BEETLE_OPT_CPU_DYNAREC_LABEL_GA,
      NULL,
      BEETLE_OPT_CPU_DYNAREC_INFO_0_GA,
      NULL,
      "hacks",
      {
         { "disabled",        OPTION_VAL_DISABLED_GA },
         { "execute",         OPTION_VAL_EXECUTE_GA },
         { "execute_one",     OPTION_VAL_EXECUTE_ONE_GA },
         { "run_interpreter", OPTION_VAL_RUN_INTERPRETER_GA },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(dynarec_invalidate),
      BEETLE_OPT_DYNAREC_INVALIDATE_LABEL_GA,
      NULL,
      BEETLE_OPT_DYNAREC_INVALIDATE_INFO_0_GA,
      NULL,
      "hacks",
      {
         { "full", OPTION_VAL_FULL_GA },
         { "dma",  OPTION_VAL_DMA_GA },
         { NULL, NULL },
      },
      "full"
   },
   {
      BEETLE_OPT(dynarec_eventcycles),
      BEETLE_OPT_DYNAREC_EVENTCYCLES_LABEL_GA,
      NULL,
      BEETLE_OPT_DYNAREC_EVENTCYCLES_INFO_0_GA,
      NULL,
      "hacks",
      {
         { "128",  OPTION_VAL_128_GA },
         { "256",  NULL },
         { "384",  NULL },
         { "512",  NULL },
         { "640",  NULL },
         { "768",  NULL },
         { "896",  NULL },
         { "1024", NULL },
         { "1152", NULL },
         { "1280", NULL },
         { "1408", NULL },
         { "1536", NULL },
         { "1664", NULL },
         { "1792", NULL },
         { "1920", NULL },
         { "2048", NULL },
         { NULL, NULL },
      },
      "128"
   },
   {
      BEETLE_OPT(dynarec_spu_samples),
      BEETLE_OPT_DYNAREC_SPU_SAMPLES_LABEL_GA,
      NULL,
      BEETLE_OPT_DYNAREC_SPU_SAMPLES_INFO_0_GA,
      NULL,
      "hacks",
      {
         { "1",  OPTION_VAL_1_O49_GA },
         { "4",  NULL },
         { "16", NULL },
         { NULL, NULL },
      },
      "1"
   },
#endif
   {
      BEETLE_OPT(core_timing_fps),
      BEETLE_OPT_CORE_TIMING_FPS_LABEL_GA,
      NULL,
      BEETLE_OPT_CORE_TIMING_FPS_INFO_0_GA,
      NULL,
      "video",
      {
         { "auto_toggle",       OPTION_VAL_AUTO_TOGGLE_GA },
         { "force_progressive", OPTION_VAL_FORCE_PROGRESSIVE_GA },
         { "force_interlaced",  OPTION_VAL_FORCE_INTERLACED_GA },
      },
      "auto_toggle"
   },
   {
      BEETLE_OPT(deinterlacer),
      BEETLE_OPT_DEINTERLACER_LABEL_GA,
      NULL,
      BEETLE_OPT_DEINTERLACER_INFO_0_GA,
      NULL,
      "video",
      {
         { "weave",      OPTION_VAL_WEAVE_GA },
         { "bob",        OPTION_VAL_BOB_GA },
         { "bob_offset", OPTION_VAL_BOB_OFFSET_GA },
         { "fastmad",    OPTION_VAL_FASTMAD_GA },
         { "off",        "Off" },
         { NULL, NULL },
      },
      "weave"
   },
   {
      BEETLE_OPT(frame_duping),
      BEETLE_OPT_FRAME_DUPING_LABEL_GA,
      NULL,
      BEETLE_OPT_FRAME_DUPING_INFO_0_GA,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(pal_video_timing_override),
      BEETLE_OPT_PAL_VIDEO_TIMING_OVERRIDE_LABEL_GA,
      NULL,
      BEETLE_OPT_PAL_VIDEO_TIMING_OVERRIDE_INFO_0_GA,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(aspect_ratio),
      BEETLE_OPT_ASPECT_RATIO_LABEL_GA,
      NULL,
      BEETLE_OPT_ASPECT_RATIO_INFO_0_GA,
      NULL,
      "video",
      {
         { "corrected", OPTION_VAL_CORRECTED_GA },
         { "uncorrected", OPTION_VAL_UNCORRECTED_GA },
         { "4:3",  OPTION_VAL_4_3_GA },
         { "ntsc", OPTION_VAL_NTSC_GA },
      },
      "corrected"
   },
   {
      BEETLE_OPT(crop_overscan),
      BEETLE_OPT_CROP_OVERSCAN_LABEL_GA,
      NULL,
      BEETLE_OPT_CROP_OVERSCAN_INFO_0_GA,
      NULL,
      "video",
      {
         { "disabled", "Disabled" },
         { "static",   OPTION_VAL_STATIC_GA },
         { "smart",    OPTION_VAL_SMART_GA },
         { NULL, NULL },
      },
      "smart"
   },
   {
      BEETLE_OPT(image_crop),
      BEETLE_OPT_IMAGE_CROP_LABEL_GA,
      NULL,
      BEETLE_OPT_IMAGE_CROP_INFO_0_GA,
      NULL,
      "video",
      {
         { "disabled", "0" },
         { "1px",      OPTION_VAL_1PX_GA },
         { "2px",      OPTION_VAL_2PX_GA },
         { "3px",      OPTION_VAL_3PX_GA },
         { "4px",      OPTION_VAL_4PX_GA },
         { "5px",      OPTION_VAL_5PX_GA },
         { "6px",      OPTION_VAL_6PX_GA },
         { "7px",      OPTION_VAL_7PX_GA },
         { "8px",      OPTION_VAL_8PX_GA },
         { "9px",      OPTION_VAL_9PX_GA },
         { "10px",     OPTION_VAL_10PX_GA },
         { "11px",     OPTION_VAL_11PX_GA },
         { "12px",     OPTION_VAL_12PX_GA },
         { "13px",     OPTION_VAL_13PX_GA },
         { "14px",     OPTION_VAL_14PX_GA },
         { "15px",     OPTION_VAL_15PX_GA },
         { "16px",     OPTION_VAL_16PX_GA },
         { "17px",     OPTION_VAL_17PX_GA },
         { "18px",     OPTION_VAL_18PX_GA },
         { "19px",     OPTION_VAL_19PX_GA },
         { "20px",     OPTION_VAL_20PX_GA },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(image_offset),
      BEETLE_OPT_IMAGE_OFFSET_LABEL_GA,
      NULL,
      BEETLE_OPT_IMAGE_OFFSET_INFO_0_GA,
      NULL,
      "video",
      {
         { "-12px",    OPTION_VAL_12PX_O64_GA },
         { "-11px",    OPTION_VAL_11PX_O64_GA },
         { "-10px",    OPTION_VAL_10PX_O64_GA },
         { "-9px",     OPTION_VAL_9PX_O64_GA },
         { "-8px",     OPTION_VAL_8PX_O64_GA },
         { "-7px",     OPTION_VAL_7PX_O64_GA },
         { "-6px",     OPTION_VAL_6PX_O64_GA },
         { "-5px",     OPTION_VAL_5PX_O64_GA },
         { "-4px",     OPTION_VAL_4PX_O64_GA },
         { "-3px",     OPTION_VAL_3PX_O64_GA },
         { "-2px",     OPTION_VAL_2PX_O64_GA },
         { "-1px",     OPTION_VAL_1PX_O64_GA },
         { "disabled", "0" },
         { "+1px",     OPTION_VAL_1PX_O64_0_GA },
         { "+2px",     OPTION_VAL_2PX_O64_0_GA },
         { "+3px",     OPTION_VAL_3PX_O64_0_GA },
         { "+4px",     OPTION_VAL_4PX_O64_0_GA },
         { "+5px",     OPTION_VAL_5PX_O64_0_GA },
         { "+6px",     OPTION_VAL_6PX_O64_0_GA },
         { "+7px",     OPTION_VAL_7PX_O64_0_GA },
         { "+8px",     OPTION_VAL_8PX_O64_0_GA },
         { "+9px",     OPTION_VAL_9PX_O64_0_GA },
         { "+10px",    OPTION_VAL_10PX_O64_0_GA },
         { "+11px",    OPTION_VAL_11PX_O64_0_GA },
         { "+12px",    OPTION_VAL_12PX_O64_0_GA },
         { NULL, NULL },
      },
      "disabled"
   },
#if defined(HAVE_OPENGL) || defined(HAVE_OPENGLES) || defined(HAVE_VULKAN)
   {
      BEETLE_OPT(image_offset_cycles),
      BEETLE_OPT_IMAGE_OFFSET_CYCLES_LABEL_GA,
      NULL,
      BEETLE_OPT_IMAGE_OFFSET_CYCLES_INFO_0_GA,
      NULL,
      "video",
      {
         { "-40",      NULL },
         { "-39",      NULL },
         { "-38",      NULL },
         { "-37",      NULL },
         { "-36",      NULL },
         { "-35",      NULL },
         { "-34",      NULL },
         { "-33",      NULL },
         { "-32",      NULL },
         { "-31",      NULL },
         { "-30",      NULL },
         { "-29",      NULL },
         { "-28",      NULL },
         { "-27",      NULL },
         { "-26",      NULL },
         { "-25",      NULL },
         { "-24",      NULL },
         { "-23",      NULL },
         { "-22",      NULL },
         { "-21",      NULL },
         { "-20",      NULL },
         { "-19",      NULL },
         { "-18",      NULL },
         { "-17",      NULL },
         { "-16",      NULL },
         { "-15",      NULL },
         { "-14",      NULL },
         { "-13",      NULL },
         { "-12",      NULL },
         { "-11",      NULL },
         { "-10",      NULL },
         { "-9",       NULL },
         { "-8",       NULL },
         { "-7",       NULL },
         { "-6",       NULL },
         { "-5",       NULL },
         { "-4",       NULL },
         { "-3",       NULL },
         { "-2",       NULL },
         { "-1",       NULL },
         { "0",        "0" },
         { "+1",       NULL },
         { "+2",       NULL },
         { "+3",       NULL },
         { "+4",       NULL },
         { "+5",       NULL },
         { "+6",       NULL },
         { "+7",       NULL },
         { "+8",       NULL },
         { "+9",       NULL },
         { "+10",      NULL },
         { "+11",      NULL },
         { "+12",      NULL },
         { "+13",      NULL },
         { "+14",      NULL },
         { "+15",      NULL },
         { "+16",      NULL },
         { "+17",      NULL },
         { "+18",      NULL },
         { "+19",      NULL },
         { "+20",      NULL },
         { "+21",      NULL },
         { "+22",      NULL },
         { "+23",      NULL },
         { "+24",      NULL },
         { "+25",      NULL },
         { "+26",      NULL },
         { "+27",      NULL },
         { "+28",      NULL },
         { "+29",      NULL },
         { "+30",      NULL },
         { "+31",      NULL },
         { "+32",      NULL },
         { "+33",      NULL },
         { "+34",      NULL },
         { "+35",      NULL },
         { "+36",      NULL },
         { "+37",      NULL },
         { "+38",      NULL },
         { "+39",      NULL },
         { "+40",      NULL },
         { NULL, NULL},
      },
      "0"
   },
#endif
   {
      BEETLE_OPT(initial_scanline),
      BEETLE_OPT_INITIAL_SCANLINE_LABEL_GA,
      NULL,
      BEETLE_OPT_INITIAL_SCANLINE_INFO_0_GA,
      NULL,
      "video",
      {
         { "0",  OPTION_VAL_0_O48_GA },
         { "1",  NULL },
         { "2",  NULL },
         { "3",  NULL },
         { "4",  NULL },
         { "5",  NULL },
         { "6",  NULL },
         { "7",  NULL },
         { "8",  NULL },
         { "9",  NULL },
         { "10", NULL },
         { "11", NULL },
         { "12", NULL },
         { "13", NULL },
         { "14", NULL },
         { "15", NULL },
         { "16", NULL },
         { "17", NULL },
         { "18", NULL },
         { "19", NULL },
         { "20", NULL },
         { "21", NULL },
         { "22", NULL },
         { "23", NULL },
         { "24", NULL },
         { "25", NULL },
         { "26", NULL },
         { "27", NULL },
         { "28", NULL },
         { "29", NULL },
         { "30", NULL },
         { "31", NULL },
         { "32", NULL },
         { "33", NULL },
         { "34", NULL },
         { "35", NULL },
         { "36", NULL },
         { "37", NULL },
         { "38", NULL },
         { "39", NULL },
         { "40", NULL },
         { NULL, NULL },
      },
      "0"
   },
   {
      BEETLE_OPT(last_scanline),
      BEETLE_OPT_LAST_SCANLINE_LABEL_GA,
      NULL,
      BEETLE_OPT_LAST_SCANLINE_INFO_0_GA,
      NULL,
      "video",
      {
         { "210", NULL },
         { "211", NULL },
         { "212", NULL },
         { "213", NULL },
         { "214", NULL },
         { "215", NULL },
         { "216", NULL },
         { "217", NULL },
         { "218", NULL },
         { "219", NULL },
         { "220", NULL },
         { "221", NULL },
         { "222", NULL },
         { "223", NULL },
         { "224", NULL },
         { "225", NULL },
         { "226", NULL },
         { "227", NULL },
         { "228", NULL },
         { "229", NULL },
         { "230", NULL },
         { "231", NULL },
         { "232", NULL },
         { "233", NULL },
         { "234", NULL },
         { "235", NULL },
         { "236", NULL },
         { "237", NULL },
         { "238", NULL },
         { "239", OPTION_VAL_239_GA },
         { NULL, NULL },
      },
      "239"
   },
   {
      BEETLE_OPT(initial_scanline_pal),
      BEETLE_OPT_INITIAL_SCANLINE_PAL_LABEL_GA,
      NULL,
      BEETLE_OPT_INITIAL_SCANLINE_PAL_INFO_0_GA,
      NULL,
      "video",
      {
         { "0",  OPTION_VAL_0_O48_GA },
         { "1",  NULL },
         { "2",  NULL },
         { "3",  NULL },
         { "4",  NULL },
         { "5",  NULL },
         { "6",  NULL },
         { "7",  NULL },
         { "8",  NULL },
         { "9",  NULL },
         { "10", NULL },
         { "11", NULL },
         { "12", NULL },
         { "13", NULL },
         { "14", NULL },
         { "15", NULL },
         { "16", NULL },
         { "17", NULL },
         { "18", NULL },
         { "19", NULL },
         { "20", NULL },
         { "21", NULL },
         { "22", NULL },
         { "23", NULL },
         { "24", NULL },
         { "25", NULL },
         { "26", NULL },
         { "27", NULL },
         { "28", NULL },
         { "29", NULL },
         { "30", NULL },
         { "31", NULL },
         { "32", NULL },
         { "33", NULL },
         { "34", NULL },
         { "35", NULL },
         { "36", NULL },
         { "37", NULL },
         { "38", NULL },
         { "39", NULL },
         { "40", NULL },
         { NULL, NULL },
      },
      "0"
   },
   {
      BEETLE_OPT(last_scanline_pal),
      BEETLE_OPT_LAST_SCANLINE_PAL_LABEL_GA,
      NULL,
      BEETLE_OPT_LAST_SCANLINE_PAL_INFO_0_GA,
      NULL,
      "video",
      {
         { "230", NULL },
         { "231", NULL },
         { "232", NULL },
         { "233", NULL },
         { "234", NULL },
         { "235", NULL },
         { "236", NULL },
         { "237", NULL },
         { "238", NULL },
         { "239", NULL },
         { "240", NULL },
         { "241", NULL },
         { "242", NULL },
         { "243", NULL },
         { "244", NULL },
         { "245", NULL },
         { "246", NULL },
         { "247", NULL },
         { "248", NULL },
         { "249", NULL },
         { "250", NULL },
         { "251", NULL },
         { "252", NULL },
         { "253", NULL },
         { "254", NULL },
         { "255", NULL },
         { "256", NULL },
         { "257", NULL },
         { "258", NULL },
         { "259", NULL },
         { "260", NULL },
         { "261", NULL },
         { "262", NULL },
         { "263", NULL },
         { "264", NULL },
         { "265", NULL },
         { "266", NULL },
         { "267", NULL },
         { "268", NULL },
         { "269", NULL },
         { "270", NULL },
         { "271", NULL },
         { "272", NULL },
         { "273", NULL },
         { "274", NULL },
         { "275", NULL },
         { "276", NULL },
         { "277", NULL },
         { "278", NULL },
         { "279", NULL },
         { "280", NULL },
         { "281", NULL },
         { "282", NULL },
         { "283", NULL },
         { "284", NULL },
         { "285", NULL },
         { "286", NULL },
         { "287", OPTION_VAL_287_GA },
         { NULL, NULL },
      },
      "287"
   },
   { NULL, NULL, NULL, NULL, NULL, NULL, {{0}}, NULL },
};
struct retro_core_options_v2 options_ga = {
   option_cats_ga,
   option_defs_ga
};

/* RETRO_LANGUAGE_GL */

#define CATEGORY_SYSTEM_LABEL_GL "Sistema"
#define CATEGORY_SYSTEM_INFO_0_GL "Cambiar opcións do sistema e de overclocking."
#define CATEGORY_VIDEO_LABEL_GL "Vídeo"
#define CATEGORY_VIDEO_INFO_0_GL "Cambia a relación de aspecto, o recorte da pantalla, o filtro de vídeo e a configuración de saltar de fotogramas."
#define CATEGORY_PGXP_LABEL_GL NULL
#define CATEGORY_PGXP_INFO_0_GL "Estas opcións controlan qué parámetros poden millorar os gráficos comparados coa consola orixinal PGXP pode desbotar texturas deformadas e problemas co eixe Z nos xogos de loita."
#define CATEGORY_OSD_LABEL_GL "Información en pantalla (OSD)"
#define CATEGORY_OSD_INFO_0_GL "Configurar as notificacións que se amosan na pantalla."
#define CATEGORY_INPUT_LABEL_GL "Entrada"
#define CATEGORY_INPUT_INFO_0_GL "Cambia a configuración de pistola lixeira, rato e neGcon."
#define CATEGORY_MEMCARDS_LABEL_GL "Tarxeta de memoria"
#define CATEGORY_MEMCARDS_INFO_0_GL "Cambiar configuracións relacionadas coa tarxeta de memoria virtual."
#define CATEGORY_HACKS_LABEL_GL "Hacks de emulación"
#define CATEGORY_HACKS_INFO_0_GL "Cambiar a precisión da emulación que afecta ao rendemento a baixo nivel e á compatibilidade."
#define BEETLE_OPT_CPU_FREQ_SCALE_LABEL_GL "Escala de frecuencia da CPU (overclock)"
#define BEETLE_OPT_CPU_FREQ_SCALE_INFO_0_GL "Overclock (ou underclock) a CPU PSX emulada. O overclocking pode eliminar a desaceleración e mellorar as taxas de cadros en determinados xogos a costa de aumentar os requisitos de rendemento. Teña en conta que algúns xogos teñen un limitador de frecuencia de cadros interno e poden non beneficiarse do overclocking. Pode provocar que certos efectos se animen máis rápido do previsto nalgúns títulos cando se fai overclock."
#define OPTION_VAL_50_GL NULL
#define OPTION_VAL_60_GL NULL
#define OPTION_VAL_70_GL NULL
#define OPTION_VAL_80_GL NULL
#define OPTION_VAL_90_GL NULL
#define OPTION_VAL_100_GL "100% (Nativo)"
#define OPTION_VAL_110_GL NULL
#define OPTION_VAL_120_GL NULL
#define OPTION_VAL_130_GL NULL
#define OPTION_VAL_140_GL NULL
#define OPTION_VAL_150_GL NULL
#define OPTION_VAL_160_GL NULL
#define OPTION_VAL_170_GL NULL
#define OPTION_VAL_180_GL NULL
#define OPTION_VAL_190_GL NULL
#define OPTION_VAL_200_GL NULL
#define OPTION_VAL_210_GL NULL
#define OPTION_VAL_220_GL NULL
#define OPTION_VAL_230_GL NULL
#define OPTION_VAL_240_GL NULL
#define OPTION_VAL_250_GL NULL
#define OPTION_VAL_260_GL NULL
#define OPTION_VAL_270_GL NULL
#define OPTION_VAL_280_GL NULL
#define OPTION_VAL_290_GL NULL
#define OPTION_VAL_300_GL NULL
#define OPTION_VAL_310_GL NULL
#define OPTION_VAL_320_GL NULL
#define OPTION_VAL_330_GL NULL
#define OPTION_VAL_340_GL NULL
#define OPTION_VAL_350_GL NULL
#define OPTION_VAL_360_GL NULL
#define OPTION_VAL_370_GL NULL
#define OPTION_VAL_380_GL NULL
#define OPTION_VAL_390_GL NULL
#define OPTION_VAL_400_GL NULL
#define OPTION_VAL_410_GL NULL
#define OPTION_VAL_420_GL NULL
#define OPTION_VAL_430_GL NULL
#define OPTION_VAL_440_GL NULL
#define OPTION_VAL_450_GL NULL
#define OPTION_VAL_460_GL NULL
#define OPTION_VAL_470_GL NULL
#define OPTION_VAL_480_GL NULL
#define OPTION_VAL_490_GL NULL
#define OPTION_VAL_500_GL NULL
#define OPTION_VAL_510_GL NULL
#define OPTION_VAL_520_GL NULL
#define OPTION_VAL_530_GL NULL
#define OPTION_VAL_540_GL NULL
#define OPTION_VAL_550_GL NULL
#define OPTION_VAL_560_GL NULL
#define OPTION_VAL_570_GL NULL
#define OPTION_VAL_580_GL NULL
#define OPTION_VAL_590_GL NULL
#define OPTION_VAL_600_GL NULL
#define OPTION_VAL_610_GL NULL
#define OPTION_VAL_620_GL NULL
#define OPTION_VAL_630_GL NULL
#define OPTION_VAL_640_GL NULL
#define OPTION_VAL_650_GL NULL
#define OPTION_VAL_660_GL NULL
#define OPTION_VAL_670_GL NULL
#define OPTION_VAL_680_GL NULL
#define OPTION_VAL_690_GL NULL
#define OPTION_VAL_700_GL NULL
#define OPTION_VAL_710_GL NULL
#define OPTION_VAL_720_GL NULL
#define OPTION_VAL_730_GL NULL
#define OPTION_VAL_740_GL NULL
#define OPTION_VAL_750_GL NULL
#define BEETLE_OPT_GPU_OVERCLOCK_LABEL_GL "Overclock de GPU Rasterizer"
#define BEETLE_OPT_GPU_OVERCLOCK_INFO_0_GL "Activa o overclocking do rasterizador 2D contido na GPU da PSX emulada. Non mellora o renderizado 3D e, en xeral, ten pouco efecto."
#define OPTION_VAL_1X_NATIVE_GL "1x (Nativo)"
#define OPTION_VAL_2X_GL NULL
#define OPTION_VAL_4X_GL NULL
#define OPTION_VAL_8X_GL NULL
#define OPTION_VAL_16X_GL NULL
#define OPTION_VAL_32X_GL NULL
#define BEETLE_OPT_GTE_OVERCLOCK_LABEL_GL "Overclock GTE"
#define BEETLE_OPT_GTE_OVERCLOCK_INFO_0_GL "Reduce todas as operacións emuladas de GTE (coprocesador de CPU para gráficos 3D) a unha latencia constante dun ciclo. Para os xogos que fan un uso intensivo do GTE, isto pode mellorar moito a taxa de cadros e a estabilidade do tempo de cadros."
#define BEETLE_OPT_CD_ACCESS_METHOD_LABEL_GL "Método de acceso ao CD"
#define BEETLE_OPT_CD_ACCESS_METHOD_INFO_0_GL "“Síncrono” imita o soporte físico orixinal. “Asíncrono” pode reducir tartamudeos en dispositivos con almacenamento lento. “Pre-carga” carga toda a imaxe do disco na memoria ao iniciar o contido, o que pode mellorar os tempos de carga durante o xogo a cambio dun atraso inicial no inicio. “Pre-carga” pode causar problemas en sistemas con pouca RAM e volverá a “Síncrono” para medios físicos. É necesario reiniciar."
#define OPTION_VAL_SYNC_GL "Sincrónico"
#define OPTION_VAL_ASYNC_GL "Asíncrono"
#define OPTION_VAL_PRECACHE_GL "Pre-caché"
#define BEETLE_OPT_CD_FASTLOAD_LABEL_GL "Velocidade de carga de CD"
#define BEETLE_OPT_CD_FASTLOAD_INFO_0_GL "Valores superiores a “2x (Nativo)” poden reducir considerablemente os tempos de carga no xogo, pero poden introducir erros de sincronización. Algúns xogos poden non funcionar correctamente por riba dun certo valor."
#define OPTION_VAL_2X_NATIVE_GL "2x (nativo)"
#define OPTION_VAL_6X_GL NULL
#define OPTION_VAL_10X_GL NULL
#define OPTION_VAL_12X_GL NULL
#define OPTION_VAL_14X_GL NULL
#define BEETLE_OPT_SKIP_BIOS_LABEL_GL "Saltar a BIOS"
#define BEETLE_OPT_SKIP_BIOS_INFO_0_GL "Omita a animación de inicio da BIOS de PlayStation que se mostra normalmente ao cargar contido. Nota: Activar isto provoca problemas de compatibilidade con varios xogos (xogos protexidos contra copia PAL, Saga Frontier, etc.)."
#define BEETLE_OPT_OVERRIDE_BIOS_LABEL_GL "Sobrescribir BIOS"
#define BEETLE_OPT_OVERRIDE_BIOS_INFO_0_GL "Sobrescribir o BIOS estándar específico dunha rexión cun BIOS libre de rexión se se atopa. É necesario reiniciar."
#define OPTION_VAL_PSXONPSP_GL NULL
#define OPTION_VAL_PS1_ROM_GL NULL
#define OPTION_VAL_OPENBIOS_GL NULL
#define BEETLE_OPT_REGION_LABEL_GL "Rexión do sistema"
#define BEETLE_OPT_REGION_INFO_0_GL NULL
#define OPTION_VAL_AUTO_GL NULL
#define OPTION_VAL_NTSC_J_GL NULL
#define OPTION_VAL_NTSC_U_GL NULL
#define OPTION_VAL_PAL_GL NULL
#define BEETLE_OPT_INTERNAL_RESOLUTION_LABEL_GL "Resolución interna da GPU"
#define BEETLE_OPT_INTERNAL_RESOLUTION_INFO_0_GL "Resolucións superiores a “1x (Nativo)” melloran a fidelidade dos modelos 3D a cambio de maiores requirimentos de rendemento. Os elementos 2D xeralmente non se ven afectados por esta configuración."
#define BEETLE_OPT_RENDERER_LABEL_GL "Motor de renderizado"
#define BEETLE_OPT_RENDERER_INFO_0_GL "O motor de renderizado por soporte lóxico é o máis preciso, pero ten grandes requirimentos de rendemento ao funcionar a resolucións internas de GPU aumentadas. Os motores de soporte físico, aínda que menos precisos, melloran o rendemento fronte ao motor de soporte lóxico en resolucións internas aumentadas e permiten varias melloras gráficas. “Soporte físico (Auto)” selecciona automaticamente o motor de Vulkan ou OpenGL, dependendo do controlador de vídeo do frontend libretro actual. Se o controlador de vídeo non é compatible con Vulkan ou OpenGL 3.3, o núcleo volverá ao motor de soporte lóxico. É necesario reiniciar."
#define OPTION_VAL_HARDWARE_GL NULL
#define OPTION_VAL_HARDWARE_GL_GL NULL
#define OPTION_VAL_HARDWARE_VK_GL NULL
#define OPTION_VAL_SOFTWARE_GL NULL
#define BEETLE_OPT_RENDERER_SOFTWARE_FB_LABEL_GL NULL
#define BEETLE_OPT_RENDERER_SOFTWARE_FB_INFO_0_GL "Activa a emulación precisa dos efectos do framebuffer (por exemplo, desenfoque de movemento, remolino de batalla FF7) ao utilizar renderizadores de hardware executando unha copia do renderizador de software con resolución nativa en segundo plano. Se está desactivada, estas operacións omítense (OpenGL) ou represéntanse na GPU (Vulkan). A desactivación pode mellorar o rendemento, pero pode provocar erros gráficos graves. Deixa activado se non está seguro."
#define BEETLE_OPT_DEPTH_LABEL_GL "Profundidade de Cor Interna"
#define BEETLE_OPT_DEPTH_INFO_0_GL "Unha maior profundidade de cor pode reducir os efectos de bandas de cor sen usar dithering. 16 bpp emula o soporte físico orixinal, pero pode mostrar bandas visibles se o dithering non está activado. Recoméndase desactivar o “Patrón de dithering” cando esta opción se configure a 32 bpp."
#define OPTION_VAL_16BPP_NATIVE_GL "16 bpp (Nativo)"
#define OPTION_VAL_32BPP_GL NULL
#define BEETLE_OPT_DITHER_MODE_LABEL_GL "Patrón Dithering"
#define BEETLE_OPT_DITHER_MODE_INFO_0_GL "“1x (Nativo)” emula o dithering en baixa resolución usado polo soporte físico orixinal para suavizar os artefactos de bandas de cor visibles na profundidade de cor nativa. “Resolución interna” axusta a granularidade do dithering á resolución interna configurada para obter resultados máis limpos. Recoméndase desactivar esta opción cando se use unha profundidade de cor de 32 bpp. Nota: En Vulkan, activar isto forzará a redución á profundidade de cor nativa, mentres que desactivalo activará automaticamente a saída a maior profundidade de cor."
#define OPTION_VAL_INTERNAL_RESOLUTION_GL "Resolución interna"
#define BEETLE_OPT_SCALED_UV_OFFSET_LABEL_GL "Desprazamento UV da textura"
#define BEETLE_OPT_SCALED_UV_OFFSET_INFO_0_GL "Mostra de texturas para polígonos 3D cunha compensación para unha resolución interna superior a 1x. Reduce as costuras de textura pero pode causar fallos gráficos non desexados."
#define BEETLE_OPT_FILTER_LABEL_GL "Filtrado de Texturas"
#define BEETLE_OPT_FILTER_INFO_0_GL "“Nearest” emula o soporte físico orixinal. “Bilinear” e “3-Point” son filtros de suavizado que reducen a pixelación mediante desenfoque. “SABR”, “xBR” e “JINC2” son filtros de ampliación que poden mellorar a fidelidade ou nitidez das texturas a cambio de maiores requirimentos de rendemento. Só son compatibles cos motores de soporte físico."
#define OPTION_VAL_NEAREST_GL NULL
#define OPTION_VAL_SABR_GL NULL
#define OPTION_VAL_XBR_GL NULL
#define OPTION_VAL_BILINEAR_GL NULL
#define OPTION_VAL_3_POINT_GL "3-puntos"
#define OPTION_VAL_JINC2_GL NULL
#define BEETLE_OPT_FILTER_EXCLUDE_SPRITE_LABEL_GL "Exclúe os sprites do filtrado"
#define BEETLE_OPT_FILTER_EXCLUDE_SPRITE_INFO_0_GL "Non aplique filtrado de textura aos sprites. Evita as costuras en varios xogos con fondos representados en sprites en 2D. Utilízao xunto con Adaptive Smoothing ou outro filtro de post-procesamento para obter o mellor efecto."
#define OPTION_VAL_OPAQUE_GL "Opaco únicamente"
#define OPTION_VAL_ALL_GL "Opaco e semi-transparente"
#define BEETLE_OPT_FILTER_EXCLUDE_2D_POLYGON_LABEL_GL "Exclúe os polígonos 2D do filtrado"
#define BEETLE_OPT_FILTER_EXCLUDE_2D_POLYGON_INFO_0_GL "Non aplica o filtrado de texturas ós polígonos 2D. Os polígonos 2D son detectados cunha heurística e pode haber erros gráficos. Usar esta opción xunto co Suavizado Adaptativo ou con outro filtro de post-procesado para un millor efecto."
#define BEETLE_OPT_ADAPTIVE_SMOOTHING_LABEL_GL "Suavizado Adaptativo"
#define BEETLE_OPT_ADAPTIVE_SMOOTHING_INFO_0_GL "Suavizar as obras de arte en 2D e os elementos da IU sen difuminar os obxectos renderizados en 3D. Só admitido polo renderizador Vulkan."
#define BEETLE_OPT_SUPER_SAMPLING_LABEL_GL "Supermostreo (redución da mostra á resolución nativa)"
#define BEETLE_OPT_SUPER_SAMPLING_INFO_0_GL "Reducir o contido renderizado desde a resolución interna aumentada ata a resolución nativa. A combinación destes con multiplicadores de resolución interna máis altas permite que os xogos se mostren con obxectos 3D con antialias a baixa resolución nativa. Produce os mellores resultados cando se aplica a títulos que mesturan elementos 2D e 3D (por exemplo, caracteres 3D sobre fondos renderizados previamente) e funciona ben xunto cos sombreadores CRT. Só admitido polo renderizador Vulkan. Nota: Recoméndase que se desactive o \"Patrón de tramado\" ao activar esta opción."
#define BEETLE_OPT_MSAA_LABEL_GL NULL
#define BEETLE_OPT_MSAA_INFO_0_GL "MSAA mellora a aparencia dos obxectos 3D. Só é compatible co motor Vulkan."
#define OPTION_VAL_1X_GL NULL
#define BEETLE_OPT_MDEC_YUV_LABEL_GL "Filtro cromático MDEC YUV"
#define BEETLE_OPT_MDEC_YUV_INFO_0_GL "Mellora a calidade da reprodución FMV reducindo os artefactos de \"macrobloqueo\" (cadrados/bordes irregulares). Só admitido polo renderizador Vulkan."
#define BEETLE_OPT_TRACK_TEXTURES_LABEL_GL "Seguimento de texturas"
#define BEETLE_OPT_TRACK_TEXTURES_INFO_0_GL "Requisito previo para volcado e substitución de texturas. Probablemente provocará fallos na maioría dos xogos."
#define BEETLE_OPT_DUMP_TEXTURES_LABEL_GL "Texturas de volcado"
#define BEETLE_OPT_DUMP_TEXTURES_INFO_0_GL "Volcar as texturas usadas a <cd>-texture-dump/"
#define BEETLE_OPT_REPLACE_TEXTURES_LABEL_GL "Substituír texturas"
#define BEETLE_OPT_REPLACE_TEXTURES_INFO_0_GL "Substitúe texturas usando versións HD de <cd>-texture-replacements/"
#define BEETLE_OPT_PGXP_MODE_LABEL_GL "Modo de operación PGXP"
#define BEETLE_OPT_PGXP_MODE_INFO_0_GL "Renderizar obxectos 3D con precisión subpixel, minimizando distorsións e tremer de obxectos 3D que se observan no soporte físico orixinal debido ao uso de coordenadas de vértice en punto fixo. O modo “Só memoria” ten problemas de compatibilidade mínimos e recoméndase para uso xeral. O modo “Memoria + CPU (Inestable)” pode reducir aínda máis o tremer, pero ten altos requirimentos de rendemento e pode causar varios erros de xeometría."
#define OPTION_VAL_MEMORY_ONLY_GL "Só memoria"
#define OPTION_VAL_MEMORY_CPU_GL "Memoria + CPU (Buggy)"
#define BEETLE_OPT_PGXP_2D_TOL_LABEL_GL "Tolerancia á xeometría 2D PGXP"
#define BEETLE_OPT_PGXP_2D_TOL_INFO_0_GL "Ocultar erros máis evidentes nas operacións PGXP: o valor especifica a tolerancia na que se manterán os valores PGXP no caso de xeometrías sen a información de profundidade adecuada."
#define OPTION_VAL_0PX_GL NULL
#define OPTION_VAL_1PX_GL NULL
#define OPTION_VAL_2PX_GL NULL
#define OPTION_VAL_3PX_GL NULL
#define OPTION_VAL_4PX_GL NULL
#define OPTION_VAL_5PX_GL NULL
#define OPTION_VAL_6PX_GL NULL
#define OPTION_VAL_7PX_GL NULL
#define OPTION_VAL_8PX_GL NULL
#define BEETLE_OPT_PGXP_NCLIP_LABEL_GL NULL
#define BEETLE_OPT_PGXP_NCLIP_INFO_0_GL "Use a implementación NCLIP de PGXP. Mellora a aparencia reducindo os buracos nas xeometrías con coordenadas PGXP. Sábese que provoca que algúns xogos se bloqueen en varias circunstancias."
#define BEETLE_OPT_PGXP_VERTEX_LABEL_GL NULL
#define BEETLE_OPT_PGXP_VERTEX_INFO_0_GL "Almacene en caché as posicións dos vértices melloradas con PGXP para reutilizalas en debuxos de polígonos. Pode mellorar o aliñamento dos obxectos e reducir as costuras visibles ao renderizar texturas, pero os falsos positivos ao consultar a caché poden producir fallos gráficos. Actualmente recoméndase deixar esta opción desactivada. Esta opción só se aplica cando o Modo de operación PGXP está activado. Só admitido polos renderizadores de hardware."
#define BEETLE_OPT_PGXP_TEXTURE_LABEL_GL "PGXP Perspectiva Texturización correcta"
#define BEETLE_OPT_PGXP_TEXTURE_INFO_0_GL NULL
#define BEETLE_OPT_DISPLAY_INTERNAL_FPS_LABEL_GL NULL
#define BEETLE_OPT_DISPLAY_INTERNAL_FPS_INFO_0_GL "Mostra a velocidade de fotogramas interna á que o sistema PlayStation emulado está a renderizar contido. Nota: Require que as notificacións en pantalla estean activadas na interface de libretro."
#define BEETLE_OPT_DISPLAY_OSD_LABEL_GL "Amosar mensaxes OSD"
#define BEETLE_OPT_DISPLAY_OSD_INFO_0_GL "Amosar mensaxes OSD xeradas polo núcleo."
#define BEETLE_OPT_DISPLAY_VRAM_LABEL_GL "Mostrar VRAM completa (depuración)"
#define BEETLE_OPT_DISPLAY_VRAM_INFO_0_GL "Visualiza a VRAM de toda a consola emulada. Só admitido polos renderizadores de hardware OpenGL e Vulkan. Nota: Isto é para fins de depuración e normalmente debería estar desactivado."
#define BEETLE_OPT_ANALOG_CALIBRATION_LABEL_GL "Autocalibración analóxica"
#define BEETLE_OPT_ANALOG_CALIBRATION_INFO_0_GL "Activar a calibración dinámica das entradas analóxicas cando o dispositivo está configurado como DualShock, control analóxico, panca de mando analóxica ou neGcon. Os valores máximos rexistrados das entradas monitorízanse en tempo real e úsanse para escalar as coordenadas analóxicas enviadas ao emulador. Isto debería usarse en xogos como Mega Man Legends 2 que esperan valores maiores dos que proporcionan os mandos modernos. Para obter os mellores resultados, os sticks analóxicos deben xirar ao máximo para axustar o algoritmo de calibración cada vez que se carga contido."
#define BEETLE_OPT_ANALOG_TOGGLE_LABEL_GL "Alternador de modo analóxico de DualShock"
#define BEETLE_OPT_ANALOG_TOGGLE_INFO_0_GL "Cando o tipo de dispositivo de entrada é DualShock, esta opción permite que o DualShock emulado se altere entre os modos DIXITAL e ANALÓXICO como o hardware orixinal. O modo tamén se pode cambiar usando a combinación de botóns seleccionada."
#define OPTION_VAL_ENABLED_ANALOG_GL "Predeterminado: analóxico"
#define BEETLE_OPT_ANALOG_TOGGLE_COMBO_LABEL_GL "Combinación de modo analóxico DualShock"
#define BEETLE_OPT_ANALOG_TOGGLE_COMBO_INFO_0_GL "A combinación de botóns que se usará para cambiar entre os modos DIXITAL e ANALÓXICO no DualShock emulado."
#define OPTION_VAL_L1_L2_R1_R2_START_SELECT_GL NULL
#define OPTION_VAL_L1_R1_SELECT_GL NULL
#define OPTION_VAL_L1_R1_START_GL NULL
#define OPTION_VAL_L1_R1_L3_GL NULL
#define OPTION_VAL_L1_R1_R3_GL NULL
#define OPTION_VAL_L2_R2_SELECT_GL NULL
#define OPTION_VAL_L2_R2_START_GL NULL
#define OPTION_VAL_L2_R2_L3_GL NULL
#define OPTION_VAL_L2_R2_R3_GL NULL
#define OPTION_VAL_L3_R3_GL NULL
#define BEETLE_OPT_ANALOG_TOGGLE_HOLD_LABEL_GL "Retardo de retención combinado de modo analóxico DualShock"
#define BEETLE_OPT_ANALOG_TOGGLE_HOLD_INFO_0_GL "Tempo de mantemento para os botóns da combinación do modo analóxico."
#define OPTION_VAL_0_GL "Retraso de 0 segundos"
#define OPTION_VAL_1_GL "Retraso de 1 segundo"
#define OPTION_VAL_2_GL "Retraso de 2 segundos"
#define OPTION_VAL_3_GL "Retraso de 3 segundos"
#define OPTION_VAL_4_GL "Retraso de 4 segundo"
#define OPTION_VAL_5_GL "Retraso de 5 segundo"
#define BEETLE_OPT_ENABLE_MULTITAP_PORT1_LABEL_GL "Multitap no Porto 1"
#define BEETLE_OPT_ENABLE_MULTITAP_PORT1_INFO_0_GL "Activa a función de multitoque no porto 1."
#define BEETLE_OPT_ENABLE_MULTITAP_PORT2_LABEL_GL "Multitap no Porto 2"
#define BEETLE_OPT_ENABLE_MULTITAP_PORT2_INFO_0_GL "Activa a funcionalidade de múltiples toques no porto 2."
#define BEETLE_OPT_GUN_INPUT_MODE_LABEL_GL "Modo de entrada de pistola"
#define BEETLE_OPT_GUN_INPUT_MODE_INFO_0_GL "Usar entrada de “Light Gun” ou “Pantalla táctil” controlada por rato cando o tipo de dispositivo estea configurado como “Guncon/G-Con 45” ou “Justifier”."
#define OPTION_VAL_LIGHTGUN_GL "Pistola lixeira"
#define OPTION_VAL_TOUCHSCREEN_GL "Pantalla táctil"
#define BEETLE_OPT_GUN_CURSOR_LABEL_GL "Cursor de pistola"
#define BEETLE_OPT_GUN_CURSOR_INFO_0_GL "Debuxar un cursor na pantalla mentres se usan os tipos de dispositivo “Guncon/G-Con 45” e “Justifier”."
#define OPTION_VAL_CROSS_GL "Cruz"
#define OPTION_VAL_DOT_GL "Punto"
#define BEETLE_OPT_CROSSHAIR_COLOR_P1_LABEL_GL "Cor da mira da pistola: Porto 1"
#define BEETLE_OPT_CROSSHAIR_COLOR_P1_INFO_0_GL "Cor da mira da pistola para o porto 1."
#define OPTION_VAL_RED_GL "Vermello"
#define OPTION_VAL_BLUE_GL "Azul"
#define OPTION_VAL_GREEN_GL "Verde"
#define OPTION_VAL_ORANGE_GL "Laranxa"
#define OPTION_VAL_YELLOW_GL "Amarelo"
#define OPTION_VAL_CYAN_GL "Azul ceo"
#define OPTION_VAL_PINK_GL "Rosa"
#define OPTION_VAL_PURPLE_GL "Malva"
#define OPTION_VAL_BLACK_GL "Negro"
#define OPTION_VAL_WHITE_GL "Branco"
#define BEETLE_OPT_CROSSHAIR_COLOR_P2_LABEL_GL "Cor da mira da pistola: Porto 2"
#define BEETLE_OPT_CROSSHAIR_COLOR_P2_INFO_0_GL "Cor da mira da pistola para o porto 2."
#define BEETLE_OPT_MOUSE_SENSITIVITY_LABEL_GL "Sensibilidade do rato"
#define BEETLE_OPT_MOUSE_SENSITIVITY_INFO_0_GL "Sensibilidade do tipo de dispositivo “Rato”."
#define OPTION_VAL_5_O42_GL NULL
#define OPTION_VAL_10_GL NULL
#define OPTION_VAL_15_GL NULL
#define OPTION_VAL_20_GL NULL
#define OPTION_VAL_25_GL NULL
#define OPTION_VAL_30_GL NULL
#define OPTION_VAL_35_GL NULL
#define OPTION_VAL_40_GL NULL
#define OPTION_VAL_45_GL NULL
#define OPTION_VAL_55_GL NULL
#define OPTION_VAL_65_GL NULL
#define OPTION_VAL_75_GL NULL
#define OPTION_VAL_85_GL NULL
#define OPTION_VAL_95_GL NULL
#define OPTION_VAL_100_O42_GL NULL
#define OPTION_VAL_105_GL NULL
#define OPTION_VAL_115_GL NULL
#define OPTION_VAL_125_GL NULL
#define OPTION_VAL_135_GL NULL
#define OPTION_VAL_145_GL NULL
#define OPTION_VAL_155_GL NULL
#define OPTION_VAL_165_GL NULL
#define OPTION_VAL_175_GL NULL
#define OPTION_VAL_185_GL NULL
#define OPTION_VAL_195_GL NULL
#define BEETLE_OPT_NEGCON_RESPONSE_LABEL_GL NULL
#define BEETLE_OPT_NEGCON_RESPONSE_INFO_0_GL "Tipo de resposta do stick analóxico esquerdo ao simular a acción de “xirar/rotar” dos dispositivos emulados “neGcon”. O desprazamento do stick analóxico pode mapearse ao ángulo de rotación do neGcon de forma lineal, cuadrática ou cúbica. “Cuadrática” permite maior precisión que “Lineal” ao facer pequenos movementos. “Cúbica” aumenta aínda máis a precisión dos pequenos movementos, pero “exaxera” os movementos maiores. Nota: Recoméndase “Lineal” só ao usar volantes de carreiras. Os mandos convencionais implementan a entrada analóxica dun xeito fundamentalmente diferente ao mecanismo de “xirar/rotar” do neGcon, de xeito co mapeado lineal amplifica demasiado os pequenos movementos, dificultando o control fino. Na maioría dos casos, “Cuadrática” proporciona a aproximación máis próxima ao hardware real."
#define OPTION_VAL_LINEAR_GL "Lineal"
#define OPTION_VAL_QUADRATIC_GL "Cuadrático"
#define OPTION_VAL_CUBIC_GL "Cuadrático"
#define BEETLE_OPT_NEGCON_DEADZONE_LABEL_GL NULL
#define BEETLE_OPT_NEGCON_DEADZONE_INFO_0_GL "Zona morta do stick analóxico esquerdo ao simular a acción de “xirar/rotar” dos dispositivos emulados “neGcon”. Usada para eliminar o desvío do mando. Nota: A maioría dos xogos compatibles con neGcon ofrecen opcións no xogo para configurar o valor da zona morta do “xirar/rotar”. Para evitar perda de precisión, a zona morta no xogo debe establecerse *sempre* a cero. Calquera axuste necesario debe aplicarse *só* a través desta opción do núcleo. Isto é especialmente importante cando a opción “Resposta de xirar/rotar do neGcon” está configurada como “Cuadrática” ou “Cúbica”."
#define OPTION_VAL_0_O44_GL NULL
#define BEETLE_OPT_USE_MEDNAFEN_MEMCARD0_METHOD_LABEL_GL "Método de tarxeta de memoria"
#define BEETLE_OPT_USE_MEDNAFEN_MEMCARD0_METHOD_INFO_0_GL "Formato de datos de gardado usado para a rañura da tarxeta de memoria esquerda. “Mednafen” pode usarse para compatibilidade coa versión independente de Mednafen. Ao usalo con Beetle PSX, os gardados de Libretro (.srm) e Mednafen (.mcr) teñen formatos internamente idénticos e poden converterse entre si mediante cambio de nome. É necesario reiniciar."
#define OPTION_VAL_LIBRETRO_GL NULL
#define OPTION_VAL_MEDNAFEN_GL NULL
#define BEETLE_OPT_ENABLE_MEMCARD1_LABEL_GL "Rañura de tarxeta de memoria 2"
#define BEETLE_OPT_ENABLE_MEMCARD1_INFO_0_GL "Emular unha segunda tarxeta de memoria na rañura dereita. Nota: Algúns xogos requiren que esta opción estea desactivada para funcionar correctamente (por exemplo, Codename Tenka). Nota: A rañura de tarxeta de memoria 2 usa o formato de gardado Mednafen (.mcr). É necesario reiniciar."
#define BEETLE_OPT_SHARED_MEMORY_CARDS_LABEL_GL "Tarxetas de memoria compartidas"
#define BEETLE_OPT_SHARED_MEMORY_CARDS_INFO_0_GL "Usar tarxetas de memoria compartidas en lugar de tarxetas separadas para cada contido. Nota: se o “Método de tarxeta de memoria” está configurado como “Libretro”, só se verá afectada a rañura de tarxeta 2. É necesario reiniciar."
#define BEETLE_OPT_MEMCARD_LEFT_INDEX_LABEL_GL "Índice da rañura de tarxeta 1"
#define BEETLE_OPT_MEMCARD_LEFT_INDEX_INFO_0_GL "Tarxeta de memoria actualmente cargada na rañura esquerda. Esta opción só funcionará se o “Método de tarxeta de memoria” está configurado como Mednafen."
#define OPTION_VAL_0_O48_GL NULL
#define BEETLE_OPT_MEMCARD_RIGHT_INDEX_LABEL_GL "Índice da rañura de tarxeta 2"
#define BEETLE_OPT_MEMCARD_RIGHT_INDEX_INFO_0_GL "Tarxeta de memoria actualmente cargada na rañura dereita. Esta opción só funcionará se a rañura de tarxeta de memoria 2 está activada."
#define OPTION_VAL_1_O49_GL NULL
#define BEETLE_OPT_LINE_RENDER_LABEL_GL "Corte de liña a cuádruple"
#define BEETLE_OPT_LINE_RENDER_INFO_0_GL "Algúns xogos (por exemplo, Doom, Hexen, Soul Blade, etc.) debuxan liñas horizontais estirando triángulos dunha altura dun só píxel a través da pantalla, que se rasterizan como unha fila de píxeles no soporte físico orixinal. Este hack detecta estes pequenos triángulos e convérteos en cuadriláteros segundo se requira, permitindo que se mostren correctamente nos motores de soporte físco e a resolucións internas aumentadas. “Agresivo” é necesario para algúns títulos (por exemplo, Dark Forces, Duke Nukem), pero doutra forma pode introducir erros gráficos. Déixeo en “Predeterminado” se non estás seguro."
#define OPTION_VAL_DEFAULT_GL "Por defecto"
#define OPTION_VAL_AGGRESSIVE_GL "Agresiva"
#define BEETLE_OPT_WIDESCREEN_HACK_LABEL_GL "Modo de pantalla ancha Hack"
#define BEETLE_OPT_WIDESCREEN_HACK_INFO_0_GL "Renderiza o contido 3D de forma anamórfica e mostra o framebuffer emulado cunha relación de aspecto de pantalla panorámica. Produce os mellores resultados con xogos totalmente en 3D. Os elementos 2D estiraranse horizontalmente e poden estar mal aliñados."
#define BEETLE_OPT_WIDESCREEN_HACK_ASPECT_RATIO_LABEL_GL "Modo de pantalla ancha cortar a relación de aspecto"
#define BEETLE_OPT_WIDESCREEN_HACK_ASPECT_RATIO_INFO_0_GL "Relación de aspecto que se usará no hack de modo panorámico."
#define OPTION_VAL_16_9_GL NULL
#define OPTION_VAL_16_10_GL NULL
#define OPTION_VAL_18_9_GL NULL
#define OPTION_VAL_19_9_GL NULL
#define OPTION_VAL_20_9_GL NULL
#define OPTION_VAL_21_9_GL NULL
#define OPTION_VAL_32_9_GL NULL
#define BEETLE_OPT_CPU_DYNAREC_LABEL_GL NULL
#define BEETLE_OPT_CPU_DYNAREC_INFO_0_GL "Recompila dinámicamente as instrucións da CPU a instrucións nativas. Moito máis rápido que o intérprete, pero a sincronización da CPU é menos precisa e pode ter erros."
#define OPTION_VAL_DISABLED_GL "Desactivado (intérprete de Beetle)"
#define OPTION_VAL_EXECUTE_GL "Máximo rendemento"
#define OPTION_VAL_EXECUTE_ONE_GL "Comprobación do tempo de ciclo"
#define OPTION_VAL_RUN_INTERPRETER_GL "Intérprete Lightrec"
#define BEETLE_OPT_DYNAREC_INVALIDATE_LABEL_GL "Invalidación do código Dynarec"
#define BEETLE_OPT_DYNAREC_INVALIDATE_INFO_0_GL "Algúns xogos requiren a invalidación \"completa\", outros requiren \"Só DMA\"."
#define OPTION_VAL_FULL_GL "Cheo"
#define OPTION_VAL_DMA_GL "Só DMA (un pouco máis rápido)"
#define BEETLE_OPT_DYNAREC_EVENTCYCLES_LABEL_GL "Ciclos de eventos Dynarec DMA/GPU/MDEC/Timer"
#define BEETLE_OPT_DYNAREC_EVENTCYCLES_INFO_0_GL "Os ciclos máximos executados pola CPU antes de que se comprobe unha actualización de GPU/DMA/MDEC/temporizador, un número maior será máis rápido, ten moito menos impacto no intérprete do escaravello que o dynarec."
#define OPTION_VAL_128_GL "128 (predeterminado)"
#define BEETLE_OPT_DYNAREC_SPU_SAMPLES_LABEL_GL "Mostras de SPU de Dynarec"
#define BEETLE_OPT_DYNAREC_SPU_SAMPLES_INFO_0_GL "Máximo de mostras de SPU para executar antes de que se comprobe unha actualización de SPU, o número máis alto será máis rápido, pero provocará fallos de son nalgúns xogos con calquera cousa que non sexa 1."
#define BEETLE_OPT_CORE_TIMING_FPS_LABEL_GL "Cronometraxe FPS informada polo núcleo"
#define BEETLE_OPT_CORE_TIMING_FPS_INFO_0_GL "A alternancia automática permite ao núcleo alternar entre taxas progresivas e entrelazadas."
#define OPTION_VAL_AUTO_TOGGLE_GL "Alternancia automática"
#define OPTION_VAL_FORCE_PROGRESSIVE_GL "Forzar taxa progresiva"
#define OPTION_VAL_FORCE_INTERLACED_GL "Forzar taxa entrelazada"
#define BEETLE_OPT_DEINTERLACER_LABEL_GL "Método de desentrelazado"
#define BEETLE_OPT_DEINTERLACER_INFO_0_GL NULL
#define OPTION_VAL_WEAVE_GL "Entretecido"
#define OPTION_VAL_BOB_GL NULL
#define OPTION_VAL_BOB_OFFSET_GL NULL
#define OPTION_VAL_FASTMAD_GL NULL
#define BEETLE_OPT_FRAME_DUPING_LABEL_GL "Duplicación de fotogramas"
#define BEETLE_OPT_FRAME_DUPING_INFO_0_GL "Repetir o fotograma anterior se o núcleo non ten nada novo que mostrar."
#define BEETLE_OPT_PAL_VIDEO_TIMING_OVERRIDE_LABEL_GL "Sobrescribir temporización de vídeo PAL"
#define BEETLE_OPT_PAL_VIDEO_TIMING_OVERRIDE_INFO_0_GL "Debido a diferentes estándares, os xogos PAL adoitan parecer máis lentos en comparación coas versións NTSC estadounidenses ou xaponesas. Esta opción pódese usar para sobrescribir a temporización PAL e intentar executar estes xogos á taxa de fotogramas NTSC."
#define BEETLE_OPT_ASPECT_RATIO_LABEL_GL "Relación de aspecto do núcleo"
#define BEETLE_OPT_ASPECT_RATIO_INFO_0_GL "Esta configuración ignórase cando están activadas as opcións de hack de modo panorámico ou mostrar toda a VRAM."
#define OPTION_VAL_CORRECTED_GL "Corrixido"
#define OPTION_VAL_UNCORRECTED_GL "Sen corrixir"
#define OPTION_VAL_4_3_GL "Forza 4:3"
#define OPTION_VAL_NTSC_GL "Forzar NTSC"
#define BEETLE_OPT_CROP_OVERSCAN_LABEL_GL "Recortar Overscan"
#define BEETLE_OPT_CROP_OVERSCAN_INFO_0_GL "Imaxe totalmente nativa ou sen bordos."
#define OPTION_VAL_STATIC_GL NULL
#define OPTION_VAL_SMART_GL NULL
#define BEETLE_OPT_IMAGE_CROP_LABEL_GL "Recorte adicional"
#define BEETLE_OPT_IMAGE_CROP_INFO_0_GL "Reducir a anchura da imaxe recortada polo número de píxeles especificado."
#define OPTION_VAL_9PX_GL NULL
#define OPTION_VAL_10PX_GL NULL
#define OPTION_VAL_11PX_GL NULL
#define OPTION_VAL_12PX_GL NULL
#define OPTION_VAL_13PX_GL NULL
#define OPTION_VAL_14PX_GL NULL
#define OPTION_VAL_15PX_GL NULL
#define OPTION_VAL_16PX_GL NULL
#define OPTION_VAL_17PX_GL NULL
#define OPTION_VAL_18PX_GL NULL
#define OPTION_VAL_19PX_GL NULL
#define OPTION_VAL_20PX_GL NULL
#define BEETLE_OPT_IMAGE_OFFSET_LABEL_GL "Imaxe recortada en compensación"
#define BEETLE_OPT_IMAGE_OFFSET_INFO_0_GL "Desprazamento horizontal cara á dereita (positivo) ou á esquerda (negativo) polo número de píxeles especificado. Pode usarse para corrixir problemas de aliñamento. Só é compatible co motor de soporte lóxico."
#define OPTION_VAL_12PX_O64_GL NULL
#define OPTION_VAL_11PX_O64_GL NULL
#define OPTION_VAL_10PX_O64_GL NULL
#define OPTION_VAL_9PX_O64_GL NULL
#define OPTION_VAL_8PX_O64_GL NULL
#define OPTION_VAL_7PX_O64_GL NULL
#define OPTION_VAL_6PX_O64_GL NULL
#define OPTION_VAL_5PX_O64_GL NULL
#define OPTION_VAL_4PX_O64_GL NULL
#define OPTION_VAL_3PX_O64_GL NULL
#define OPTION_VAL_2PX_O64_GL NULL
#define OPTION_VAL_1PX_O64_GL NULL
#define OPTION_VAL_1PX_O64_0_GL NULL
#define OPTION_VAL_2PX_O64_0_GL NULL
#define OPTION_VAL_3PX_O64_0_GL NULL
#define OPTION_VAL_4PX_O64_0_GL NULL
#define OPTION_VAL_5PX_O64_0_GL NULL
#define OPTION_VAL_6PX_O64_0_GL NULL
#define OPTION_VAL_7PX_O64_0_GL NULL
#define OPTION_VAL_8PX_O64_0_GL NULL
#define OPTION_VAL_9PX_O64_0_GL NULL
#define OPTION_VAL_10PX_O64_0_GL NULL
#define OPTION_VAL_11PX_O64_0_GL NULL
#define OPTION_VAL_12PX_O64_0_GL NULL
#define BEETLE_OPT_IMAGE_OFFSET_CYCLES_LABEL_GL "Desfase horizontal da imaxe (ciclos de GPU)"
#define BEETLE_OPT_IMAGE_OFFSET_CYCLES_INFO_0_GL "Valores positivos moven a imaxe cara á dereita, valores negativos moven a imaxe cara á esquerda. Só é compatible cos motores de soporte físico."
#define BEETLE_OPT_INITIAL_SCANLINE_LABEL_GL "Liña de exploración inicial - NTSC"
#define BEETLE_OPT_INITIAL_SCANLINE_INFO_0_GL "Primeira liña escaneada mostrada ao executar contido NTSC. Valores maiores que cero reducirán a altura das imaxes de saída recortando píxeles desde o bordo superior. Pode usarse para compensar a aparición de bandas negras (letterboxing)."
#define BEETLE_OPT_LAST_SCANLINE_LABEL_GL "Última liña de exploración - NTSC"
#define BEETLE_OPT_LAST_SCANLINE_INFO_0_GL "Última liña escaneada mostrada ao executar contido NTSC. Valores menores que 239 reducirán a altura das imaxes de saída recortando píxeles desde o bordo inferior. Pode usarse para compensar a aparición de bandas negras (letterboxing)."
#define OPTION_VAL_239_GL NULL
#define BEETLE_OPT_INITIAL_SCANLINE_PAL_LABEL_GL "Liña de exploración inicial - PAL"
#define BEETLE_OPT_INITIAL_SCANLINE_PAL_INFO_0_GL "Primeira liña escaneada mostrada ao executar contido PAL. Valores maiores que cero reducirán a altura das imaxes de saída recortando píxeles desde o bordo superior. Pode usarse para compensar a aparición de bandas negras (letterboxing)."
#define BEETLE_OPT_LAST_SCANLINE_PAL_LABEL_GL "Última liña de exploración - PAL"
#define BEETLE_OPT_LAST_SCANLINE_PAL_INFO_0_GL "Última liña escaneada mostrada ao executar contido PAL. Valores menores que 287 reducirán a altura das imaxes de saída recortando píxeles desde o bordo inferior. Pode usarse para compensar a aparición de bandas negras (letterboxing)."
#define OPTION_VAL_287_GL "287 (predeterminado)"

struct retro_core_option_v2_category option_cats_gl[] = {
   {
      "system",
      CATEGORY_SYSTEM_LABEL_GL,
      CATEGORY_SYSTEM_INFO_0_GL
   },
   {
      "video",
      CATEGORY_VIDEO_LABEL_GL,
      CATEGORY_VIDEO_INFO_0_GL
   },
   {
      "pgxp",
      CATEGORY_PGXP_LABEL_GL,
      CATEGORY_PGXP_INFO_0_GL
   },
   {
      "osd",
      CATEGORY_OSD_LABEL_GL,
      CATEGORY_OSD_INFO_0_GL
   },
   {
      "input",
      CATEGORY_INPUT_LABEL_GL,
      CATEGORY_INPUT_INFO_0_GL
   },
   {
      "memcards",
      CATEGORY_MEMCARDS_LABEL_GL,
      CATEGORY_MEMCARDS_INFO_0_GL
   },
   {
      "hacks",
      CATEGORY_HACKS_LABEL_GL,
      CATEGORY_HACKS_INFO_0_GL
   },
   { NULL, NULL, NULL },
};
struct retro_core_option_v2_definition option_defs_gl[] = {
   {
      BEETLE_OPT(cpu_freq_scale),
      BEETLE_OPT_CPU_FREQ_SCALE_LABEL_GL,
      NULL,
      BEETLE_OPT_CPU_FREQ_SCALE_INFO_0_GL,
      NULL,
      "system",
      {
         { "50%",           OPTION_VAL_50_GL },
         { "60%",           OPTION_VAL_60_GL },
         { "70%",           OPTION_VAL_70_GL },
         { "80%",           OPTION_VAL_80_GL },
         { "90%",           OPTION_VAL_90_GL },
         { "100%",          OPTION_VAL_100_GL },
         { "110%",          OPTION_VAL_110_GL },
         { "120%",          OPTION_VAL_120_GL },
         { "130%",          OPTION_VAL_130_GL },
         { "140%",          OPTION_VAL_140_GL },
         { "150%",          OPTION_VAL_150_GL },
         { "160%",          OPTION_VAL_160_GL },
         { "170%",          OPTION_VAL_170_GL },
         { "180%",          OPTION_VAL_180_GL },
         { "190%",          OPTION_VAL_190_GL },
         { "200%",          OPTION_VAL_200_GL },
         { "210%",          OPTION_VAL_210_GL },
         { "220%",          OPTION_VAL_220_GL },
         { "230%",          OPTION_VAL_230_GL },
         { "240%",          OPTION_VAL_240_GL },
         { "250%",          OPTION_VAL_250_GL },
         { "260%",          OPTION_VAL_260_GL },
         { "270%",          OPTION_VAL_270_GL },
         { "280%",          OPTION_VAL_280_GL },
         { "290%",          OPTION_VAL_290_GL },
         { "300%",          OPTION_VAL_300_GL },
         { "310%",          OPTION_VAL_310_GL },
         { "320%",          OPTION_VAL_320_GL },
         { "330%",          OPTION_VAL_330_GL },
         { "340%",          OPTION_VAL_340_GL },
         { "350%",          OPTION_VAL_350_GL },
         { "360%",          OPTION_VAL_360_GL },
         { "370%",          OPTION_VAL_370_GL },
         { "380%",          OPTION_VAL_380_GL },
         { "390%",          OPTION_VAL_390_GL },
         { "400%",          OPTION_VAL_400_GL },
         { "410%",          OPTION_VAL_410_GL },
         { "420%",          OPTION_VAL_420_GL },
         { "430%",          OPTION_VAL_430_GL },
         { "440%",          OPTION_VAL_440_GL },
         { "450%",          OPTION_VAL_450_GL },
         { "460%",          OPTION_VAL_460_GL },
         { "470%",          OPTION_VAL_470_GL },
         { "480%",          OPTION_VAL_480_GL },
         { "490%",          OPTION_VAL_490_GL },
         { "500%",          OPTION_VAL_500_GL },
         { "510%",          OPTION_VAL_510_GL },
         { "520%",          OPTION_VAL_520_GL },
         { "530%",          OPTION_VAL_530_GL },
         { "540%",          OPTION_VAL_540_GL },
         { "550%",          OPTION_VAL_550_GL },
         { "560%",          OPTION_VAL_560_GL },
         { "570%",          OPTION_VAL_570_GL },
         { "580%",          OPTION_VAL_580_GL },
         { "590%",          OPTION_VAL_590_GL },
         { "600%",          OPTION_VAL_600_GL },
         { "610%",          OPTION_VAL_610_GL },
         { "620%",          OPTION_VAL_620_GL },
         { "630%",          OPTION_VAL_630_GL },
         { "640%",          OPTION_VAL_640_GL },
         { "650%",          OPTION_VAL_650_GL },
         { "660%",          OPTION_VAL_660_GL },
         { "670%",          OPTION_VAL_670_GL },
         { "680%",          OPTION_VAL_680_GL },
         { "690%",          OPTION_VAL_690_GL },
         { "700%",          OPTION_VAL_700_GL },
         { "710%",          OPTION_VAL_710_GL },
         { "720%",          OPTION_VAL_720_GL },
         { "730%",          OPTION_VAL_730_GL },
         { "740%",          OPTION_VAL_740_GL },
         { "750%",          OPTION_VAL_750_GL },
         { NULL, NULL },
      },
      "100%"
   },
   {
      BEETLE_OPT(gpu_overclock),
      BEETLE_OPT_GPU_OVERCLOCK_LABEL_GL,
      NULL,
      BEETLE_OPT_GPU_OVERCLOCK_INFO_0_GL,
      NULL,
      "system",
      {
         { "1x(native)", OPTION_VAL_1X_NATIVE_GL },
         { "2x",         OPTION_VAL_2X_GL },
         { "4x",         OPTION_VAL_4X_GL },
         { "8x",         OPTION_VAL_8X_GL },
         { "16x",        OPTION_VAL_16X_GL },
         { "32x",        OPTION_VAL_32X_GL },
         { NULL, NULL },
      },
      "1x(native)"
   },
   {
      BEETLE_OPT(gte_overclock),
      BEETLE_OPT_GTE_OVERCLOCK_LABEL_GL,
      NULL,
      BEETLE_OPT_GTE_OVERCLOCK_INFO_0_GL,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
#ifndef EMSCRIPTEN
   {
      BEETLE_OPT(cd_access_method),
      BEETLE_OPT_CD_ACCESS_METHOD_LABEL_GL,
      NULL,
      BEETLE_OPT_CD_ACCESS_METHOD_INFO_0_GL,
      NULL,
      "system",
      {
         { "sync",     OPTION_VAL_SYNC_GL },
         { "async",    OPTION_VAL_ASYNC_GL },
         { "precache", OPTION_VAL_PRECACHE_GL },
         { NULL, NULL },
      },
      "sync"
   },
#endif
   {
      BEETLE_OPT(cd_fastload),
      BEETLE_OPT_CD_FASTLOAD_LABEL_GL,
      NULL,
      BEETLE_OPT_CD_FASTLOAD_INFO_0_GL,
      NULL,
      "system",
      {
         { "2x(native)", OPTION_VAL_2X_NATIVE_GL },
         { "4x",          OPTION_VAL_4X_GL },
         { "6x",          OPTION_VAL_6X_GL },
         { "8x",          OPTION_VAL_8X_GL },
         { "10x",         OPTION_VAL_10X_GL },
         { "12x",         OPTION_VAL_12X_GL },
         { "14x",         OPTION_VAL_14X_GL },
         { NULL, NULL },
      },
      "2x(native)"
   },
   {
      BEETLE_OPT(skip_bios),
      BEETLE_OPT_SKIP_BIOS_LABEL_GL,
      NULL,
      BEETLE_OPT_SKIP_BIOS_INFO_0_GL,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(override_bios),
      BEETLE_OPT_OVERRIDE_BIOS_LABEL_GL,
      NULL,
      BEETLE_OPT_OVERRIDE_BIOS_INFO_0_GL,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "psxonpsp", OPTION_VAL_PSXONPSP_GL },
         { "ps1_rom",  OPTION_VAL_PS1_ROM_GL },
         { "openbios", OPTION_VAL_OPENBIOS_GL     },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(region),
      BEETLE_OPT_REGION_LABEL_GL,
      NULL,
      BEETLE_OPT_REGION_INFO_0_GL,
      NULL,
      "system",
      {
         { "auto",   OPTION_VAL_AUTO_GL },
         { "ntsc-j", OPTION_VAL_NTSC_J_GL },
         { "ntsc-u", OPTION_VAL_NTSC_U_GL },
         { "pal",    OPTION_VAL_PAL_GL },
         { NULL, NULL },
      },
      "auto"
   },
   {
      BEETLE_OPT(internal_resolution),
      BEETLE_OPT_INTERNAL_RESOLUTION_LABEL_GL,
      NULL,
      BEETLE_OPT_INTERNAL_RESOLUTION_INFO_0_GL,
      NULL,
      "video",
      {
         { "1x(native)", OPTION_VAL_1X_NATIVE_GL },
         { "2x",         OPTION_VAL_2X_GL },
         { "4x",         OPTION_VAL_4X_GL },
         { "8x",         OPTION_VAL_8X_GL },
         { "16x",        OPTION_VAL_16X_GL },
         { NULL, NULL },
      },
      "1x(native)"
   },
#if defined(HAVE_OPENGL) || defined(HAVE_OPENGLES) || defined(HAVE_VULKAN)
   {
      BEETLE_OPT(renderer),
      BEETLE_OPT_RENDERER_LABEL_GL,
      NULL,
      BEETLE_OPT_RENDERER_INFO_0_GL,
      NULL,
      "video",
      {
         { "hardware",    OPTION_VAL_HARDWARE_GL },
#if defined(HAVE_OPENGL) || defined(HAVE_OPENGLES)
         { "hardware_gl", OPTION_VAL_HARDWARE_GL_GL },
#endif
#if defined(HAVE_VULKAN)
         { "hardware_vk", OPTION_VAL_HARDWARE_VK_GL },
#endif
         { "software",    OPTION_VAL_SOFTWARE_GL },
         { NULL, NULL },
      },
      "hardware"
   },
   {
      BEETLE_OPT(renderer_software_fb),
      BEETLE_OPT_RENDERER_SOFTWARE_FB_LABEL_GL,
      NULL,
      BEETLE_OPT_RENDERER_SOFTWARE_FB_INFO_0_GL,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled"
   },
#endif
#if defined(HAVE_OPENGL) || defined(HAVE_OPENGLES)
   {
      BEETLE_OPT(depth),
      BEETLE_OPT_DEPTH_LABEL_GL,
      NULL,
      BEETLE_OPT_DEPTH_INFO_0_GL,
      NULL,
      "video",
      {
         { "16bpp(native)", OPTION_VAL_16BPP_NATIVE_GL },
         { "32bpp",         OPTION_VAL_32BPP_GL },
         { NULL, NULL },
      },
      "16bpp(native)"
   },
   // Sort of, it's more like 15-bit high color and 24-bit true color for visible output. The alpha channel is used for mask bit. Vulkan renderer uses ABGR1555_555 for 31 bits internal? FMVs are always 24-bit on all renderers like original hardware (BGR888, no alpha)
#endif
   {
      BEETLE_OPT(dither_mode),
      BEETLE_OPT_DITHER_MODE_LABEL_GL,
      NULL,
      BEETLE_OPT_DITHER_MODE_INFO_0_GL,
      NULL,
      "video",
      {
         { "1x(native)",          OPTION_VAL_1X_NATIVE_GL },
         { "internal resolution", OPTION_VAL_INTERNAL_RESOLUTION_GL },
         { "disabled",            NULL },
         { NULL, NULL },
      },
      "1x(native)"
   },
#ifdef HAVE_VULKAN
   {
      BEETLE_OPT(scaled_uv_offset),
      BEETLE_OPT_SCALED_UV_OFFSET_LABEL_GL,
      NULL,
      BEETLE_OPT_SCALED_UV_OFFSET_INFO_0_GL,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled"
   },
#endif
#if defined(HAVE_OPENGL) || defined(HAVE_OPENGLES) || defined(HAVE_VULKAN)
   {
      BEETLE_OPT(filter),
      BEETLE_OPT_FILTER_LABEL_GL,
      NULL,
      BEETLE_OPT_FILTER_INFO_0_GL,
      NULL,
      "video",
      {
         { "nearest",  OPTION_VAL_NEAREST_GL },
         { "SABR",     OPTION_VAL_SABR_GL },
         { "xBR",      OPTION_VAL_XBR_GL },
         { "bilinear", OPTION_VAL_BILINEAR_GL },
         { "3-point",  OPTION_VAL_3_POINT_GL },
         { "JINC2",    OPTION_VAL_JINC2_GL },
         { NULL, NULL },
      },
      "nearest"
   },
#ifdef HAVE_VULKAN
   {
      BEETLE_OPT(filter_exclude_sprite),
      BEETLE_OPT_FILTER_EXCLUDE_SPRITE_LABEL_GL,
      NULL,
      BEETLE_OPT_FILTER_EXCLUDE_SPRITE_INFO_0_GL,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "opaque", OPTION_VAL_OPAQUE_GL },
         { "all", OPTION_VAL_ALL_GL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(filter_exclude_2d_polygon),
      BEETLE_OPT_FILTER_EXCLUDE_2D_POLYGON_LABEL_GL,
      NULL,
      BEETLE_OPT_FILTER_EXCLUDE_2D_POLYGON_INFO_0_GL,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "opaque", OPTION_VAL_OPAQUE_GL },
         { "all", OPTION_VAL_ALL_GL },
         { NULL, NULL },
      },
      "disabled"
   },
#endif
#endif
#ifdef HAVE_VULKAN
   {
      BEETLE_OPT(adaptive_smoothing),
      BEETLE_OPT_ADAPTIVE_SMOOTHING_LABEL_GL,
      NULL,
      BEETLE_OPT_ADAPTIVE_SMOOTHING_INFO_0_GL,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(super_sampling),
      BEETLE_OPT_SUPER_SAMPLING_LABEL_GL,
      NULL,
      BEETLE_OPT_SUPER_SAMPLING_INFO_0_GL,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(msaa),
      BEETLE_OPT_MSAA_LABEL_GL,
      NULL,
      BEETLE_OPT_MSAA_INFO_0_GL,
      NULL,
      "video",
      {
         { "1x",  OPTION_VAL_1X_GL },
         { "2x",  OPTION_VAL_2X_GL },
         { "4x",  OPTION_VAL_4X_GL },
         { "8x",  OPTION_VAL_8X_GL },
         { "16x", OPTION_VAL_16X_GL },
         { NULL, NULL },
      },
      "1x"
   },
   {
      BEETLE_OPT(mdec_yuv),
      BEETLE_OPT_MDEC_YUV_LABEL_GL,
      NULL,
      BEETLE_OPT_MDEC_YUV_INFO_0_GL,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(track_textures),
      BEETLE_OPT_TRACK_TEXTURES_LABEL_GL,
      NULL,
      BEETLE_OPT_TRACK_TEXTURES_INFO_0_GL,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
#ifdef TEXTURE_DUMPING_ENABLED
   {
      BEETLE_OPT(dump_textures),
      BEETLE_OPT_DUMP_TEXTURES_LABEL_GL,
      NULL,
      BEETLE_OPT_DUMP_TEXTURES_INFO_0_GL,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
#endif
   {
      BEETLE_OPT(replace_textures),
      BEETLE_OPT_REPLACE_TEXTURES_LABEL_GL,
      NULL,
      BEETLE_OPT_REPLACE_TEXTURES_INFO_0_GL,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
#endif
   {
      BEETLE_OPT(pgxp_mode),
      BEETLE_OPT_PGXP_MODE_LABEL_GL,
      NULL,
      BEETLE_OPT_PGXP_MODE_INFO_0_GL,
      NULL,
      "pgxp",
      {
         { "disabled",     NULL },
         { "memory only",  OPTION_VAL_MEMORY_ONLY_GL },
         { "memory + CPU", OPTION_VAL_MEMORY_CPU_GL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(pgxp_2d_tol),
      BEETLE_OPT_PGXP_2D_TOL_LABEL_GL,
      NULL,
      BEETLE_OPT_PGXP_2D_TOL_INFO_0_GL,
      NULL,
      "pgxp",
      {
         { "disabled", NULL },
         { "0px", OPTION_VAL_0PX_GL },
         { "1px", OPTION_VAL_1PX_GL },
         { "2px", OPTION_VAL_2PX_GL },
         { "3px", OPTION_VAL_3PX_GL },
         { "4px", OPTION_VAL_4PX_GL },
         { "5px", OPTION_VAL_5PX_GL },
         { "6px", OPTION_VAL_6PX_GL },
         { "7px", OPTION_VAL_7PX_GL },
         { "8px", OPTION_VAL_8PX_GL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(pgxp_nclip),
      BEETLE_OPT_PGXP_NCLIP_LABEL_GL,
      NULL,
      BEETLE_OPT_PGXP_NCLIP_INFO_0_GL,
      NULL,
      "pgxp",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
#if defined(HAVE_OPENGL) || defined(HAVE_OPENGLES) || defined(HAVE_VULKAN)
   {
      BEETLE_OPT(pgxp_vertex),
      BEETLE_OPT_PGXP_VERTEX_LABEL_GL,
      NULL,
      BEETLE_OPT_PGXP_VERTEX_INFO_0_GL,
      NULL,
      "pgxp",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
#endif
   {
      BEETLE_OPT(pgxp_texture),
      BEETLE_OPT_PGXP_TEXTURE_LABEL_GL,
      NULL,
      BEETLE_OPT_PGXP_TEXTURE_INFO_0_GL,
      NULL,
      "pgxp",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(display_internal_fps),
      BEETLE_OPT_DISPLAY_INTERNAL_FPS_LABEL_GL,
      NULL,
      BEETLE_OPT_DISPLAY_INTERNAL_FPS_INFO_0_GL,
      NULL,
      "osd",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(display_osd),
      BEETLE_OPT_DISPLAY_OSD_LABEL_GL,
      NULL,
      BEETLE_OPT_DISPLAY_OSD_INFO_0_GL,
      NULL,
      "osd",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled"
   },
#if defined(HAVE_OPENGL) || defined(HAVE_OPENGLES) || defined(HAVE_VULKAN)
   {
      BEETLE_OPT(display_vram),
      BEETLE_OPT_DISPLAY_VRAM_LABEL_GL,
      NULL,
      BEETLE_OPT_DISPLAY_VRAM_INFO_0_GL,
      NULL,
      "osd",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
#endif
   {
      BEETLE_OPT(analog_calibration),
      BEETLE_OPT_ANALOG_CALIBRATION_LABEL_GL,
      NULL,
      BEETLE_OPT_ANALOG_CALIBRATION_INFO_0_GL,
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(analog_toggle),
      BEETLE_OPT_ANALOG_TOGGLE_LABEL_GL,
      NULL,
      BEETLE_OPT_ANALOG_TOGGLE_INFO_0_GL,
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { "enabled-analog", OPTION_VAL_ENABLED_ANALOG_GL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(analog_toggle_combo),
      BEETLE_OPT_ANALOG_TOGGLE_COMBO_LABEL_GL,
      NULL,
      BEETLE_OPT_ANALOG_TOGGLE_COMBO_INFO_0_GL,
      NULL,
      "input",
      {
         { "l1+l2+r1+r2+start+select", OPTION_VAL_L1_L2_R1_R2_START_SELECT_GL },
         { "l1+r1+select",             OPTION_VAL_L1_R1_SELECT_GL },
         { "l1+r1+start",              OPTION_VAL_L1_R1_START_GL },
         { "l1+r1+l3",                 OPTION_VAL_L1_R1_L3_GL },
         { "l1+r1+r3",                 OPTION_VAL_L1_R1_R3_GL },
         { "l2+r2+select",             OPTION_VAL_L2_R2_SELECT_GL },
         { "l2+r2+start",              OPTION_VAL_L2_R2_START_GL },
         { "l2+r2+l3",                 OPTION_VAL_L2_R2_L3_GL },
         { "l2+r2+r3",                 OPTION_VAL_L2_R2_R3_GL },
         { "l3+r3",                    OPTION_VAL_L3_R3_GL },
         { NULL, NULL },
      },
      "l1+r1+select"
   },
   {
      BEETLE_OPT(analog_toggle_hold),
      BEETLE_OPT_ANALOG_TOGGLE_HOLD_LABEL_GL,
      NULL,
      BEETLE_OPT_ANALOG_TOGGLE_HOLD_INFO_0_GL,
      NULL,
      "input",
      {
         { "0", OPTION_VAL_0_GL },
         { "1", OPTION_VAL_1_GL },
         { "2", OPTION_VAL_2_GL },
         { "3", OPTION_VAL_3_GL },
         { "4", OPTION_VAL_4_GL },
         { "5", OPTION_VAL_5_GL },
         { NULL, NULL },
      },
      "1"
   },
   {
      BEETLE_OPT(enable_multitap_port1),
      BEETLE_OPT_ENABLE_MULTITAP_PORT1_LABEL_GL,
      NULL,
      BEETLE_OPT_ENABLE_MULTITAP_PORT1_INFO_0_GL,
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(enable_multitap_port2),
      BEETLE_OPT_ENABLE_MULTITAP_PORT2_LABEL_GL,
      NULL,
      BEETLE_OPT_ENABLE_MULTITAP_PORT2_INFO_0_GL,
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(gun_input_mode),
      BEETLE_OPT_GUN_INPUT_MODE_LABEL_GL,
      NULL,
      BEETLE_OPT_GUN_INPUT_MODE_INFO_0_GL,
      NULL,
      "input",
      {
         { "lightgun",    OPTION_VAL_LIGHTGUN_GL },
         { "touchscreen", OPTION_VAL_TOUCHSCREEN_GL },
         { NULL, NULL },
      },
      "lightgun"
   },
   // Shouldn't the gun_input_mode just be Mouse vs. Touchscreen?
   {
      BEETLE_OPT(gun_cursor),
      BEETLE_OPT_GUN_CURSOR_LABEL_GL,
      NULL,
      BEETLE_OPT_GUN_CURSOR_INFO_0_GL,
      NULL,
      "input",
      {
         { "off",   "disabled" },
         { "cross", OPTION_VAL_CROSS_GL },
         { "dot",   OPTION_VAL_DOT_GL },
         { NULL, NULL },
      },
      "cross"
   },
   {
      BEETLE_OPT(crosshair_color_p1),
      BEETLE_OPT_CROSSHAIR_COLOR_P1_LABEL_GL,
      NULL,
      BEETLE_OPT_CROSSHAIR_COLOR_P1_INFO_0_GL,
      NULL,
      "input",
      {
         { "red",    OPTION_VAL_RED_GL },
         { "blue",   OPTION_VAL_BLUE_GL },
         { "green",  OPTION_VAL_GREEN_GL },
         { "orange", OPTION_VAL_ORANGE_GL },
         { "yellow", OPTION_VAL_YELLOW_GL },
         { "cyan",   OPTION_VAL_CYAN_GL },
         { "pink",   OPTION_VAL_PINK_GL },
         { "purple", OPTION_VAL_PURPLE_GL },
         { "black",  OPTION_VAL_BLACK_GL },
         { "white",  OPTION_VAL_WHITE_GL },
         { NULL, NULL },
      },
      "red"
   },
   {
      BEETLE_OPT(crosshair_color_p2),
      BEETLE_OPT_CROSSHAIR_COLOR_P2_LABEL_GL,
      NULL,
      BEETLE_OPT_CROSSHAIR_COLOR_P2_INFO_0_GL,
      NULL,
      "input",
      {
         { "blue",   OPTION_VAL_BLUE_GL },
         { "red",    OPTION_VAL_RED_GL },
         { "green",  OPTION_VAL_GREEN_GL },
         { "orange", OPTION_VAL_ORANGE_GL },
         { "yellow", OPTION_VAL_YELLOW_GL },
         { "cyan",   OPTION_VAL_CYAN_GL },
         { "pink",   OPTION_VAL_PINK_GL },
         { "purple", OPTION_VAL_PURPLE_GL },
         { "black",  OPTION_VAL_BLACK_GL },
         { "white",  OPTION_VAL_WHITE_GL },
         { NULL, NULL },
      },
      "blue"
   },
   {
      BEETLE_OPT(mouse_sensitivity),
      BEETLE_OPT_MOUSE_SENSITIVITY_LABEL_GL,
      NULL,
      BEETLE_OPT_MOUSE_SENSITIVITY_INFO_0_GL,
      NULL,
      "input",
      {
         { "5%",   OPTION_VAL_5_O42_GL },
         { "10%",  OPTION_VAL_10_GL },
         { "15%",  OPTION_VAL_15_GL },
         { "20%",  OPTION_VAL_20_GL },
         { "25%",  OPTION_VAL_25_GL },
         { "30%",  OPTION_VAL_30_GL },
         { "35%",  OPTION_VAL_35_GL },
         { "40%",  OPTION_VAL_40_GL },
         { "45%",  OPTION_VAL_45_GL },
         { "50%",  OPTION_VAL_50_GL },
         { "55%",  OPTION_VAL_55_GL },
         { "60%",  OPTION_VAL_60_GL },
         { "65%",  OPTION_VAL_65_GL },
         { "70%",  OPTION_VAL_70_GL },
         { "75%",  OPTION_VAL_75_GL },
         { "80%",  OPTION_VAL_80_GL },
         { "85%",  OPTION_VAL_85_GL },
         { "90%",  OPTION_VAL_90_GL },
         { "95%",  OPTION_VAL_95_GL },
         { "100%", OPTION_VAL_100_O42_GL },
         { "105%", OPTION_VAL_105_GL },
         { "110%", OPTION_VAL_110_GL },
         { "115%", OPTION_VAL_115_GL },
         { "120%", OPTION_VAL_120_GL },
         { "125%", OPTION_VAL_125_GL },
         { "130%", OPTION_VAL_130_GL },
         { "135%", OPTION_VAL_135_GL },
         { "140%", OPTION_VAL_140_GL },
         { "145%", OPTION_VAL_145_GL },
         { "150%", OPTION_VAL_150_GL },
         { "155%", OPTION_VAL_155_GL },
         { "160%", OPTION_VAL_160_GL },
         { "165%", OPTION_VAL_165_GL },
         { "170%", OPTION_VAL_170_GL },
         { "175%", OPTION_VAL_175_GL },
         { "180%", OPTION_VAL_180_GL },
         { "185%", OPTION_VAL_185_GL },
         { "190%", OPTION_VAL_190_GL },
         { "195%", OPTION_VAL_195_GL },
         { "200%", OPTION_VAL_200_GL },
         { NULL, NULL },
      },
      "100%"
   },
   {
      BEETLE_OPT(negcon_response),
      BEETLE_OPT_NEGCON_RESPONSE_LABEL_GL,
      NULL,
      BEETLE_OPT_NEGCON_RESPONSE_INFO_0_GL,
      NULL,
      "input",
      {
         { "linear",    OPTION_VAL_LINEAR_GL },
         { "quadratic", OPTION_VAL_QUADRATIC_GL },
         { "cubic",     OPTION_VAL_CUBIC_GL },
         { NULL, NULL },
      },
      "linear"
   },
   {
      BEETLE_OPT(negcon_deadzone),
      BEETLE_OPT_NEGCON_DEADZONE_LABEL_GL,
      NULL,
      BEETLE_OPT_NEGCON_DEADZONE_INFO_0_GL,
      NULL,
      "input",
      {
         { "0%",  OPTION_VAL_0_O44_GL },
         { "5%",  OPTION_VAL_5_O42_GL },
         { "10%", OPTION_VAL_10_GL },
         { "15%", OPTION_VAL_15_GL },
         { "20%", OPTION_VAL_20_GL },
         { "25%", OPTION_VAL_25_GL },
         { "30%", OPTION_VAL_30_GL },
         { NULL, NULL },
      },
      "0%"
   },
   {
      BEETLE_OPT(use_mednafen_memcard0_method),
      BEETLE_OPT_USE_MEDNAFEN_MEMCARD0_METHOD_LABEL_GL,
      NULL,
      BEETLE_OPT_USE_MEDNAFEN_MEMCARD0_METHOD_INFO_0_GL,
      NULL,
      "memcards",
      {
         { "libretro", OPTION_VAL_LIBRETRO_GL },
         { "mednafen", OPTION_VAL_MEDNAFEN_GL },
         { NULL, NULL },
      },
      "libretro"
   },
   {
      BEETLE_OPT(enable_memcard1),
      BEETLE_OPT_ENABLE_MEMCARD1_LABEL_GL,
      NULL,
      BEETLE_OPT_ENABLE_MEMCARD1_INFO_0_GL,
      NULL,
      "memcards",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(shared_memory_cards),
      BEETLE_OPT_SHARED_MEMORY_CARDS_LABEL_GL,
      NULL,
      BEETLE_OPT_SHARED_MEMORY_CARDS_INFO_0_GL,
      NULL,
      "memcards",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(memcard_left_index),
      BEETLE_OPT_MEMCARD_LEFT_INDEX_LABEL_GL,
      NULL,
      BEETLE_OPT_MEMCARD_LEFT_INDEX_INFO_0_GL,
      NULL,
      "memcards",
      {
         { "0",  OPTION_VAL_0_O48_GL },
         { "1",  NULL },
         { "2",  NULL },
         { "3",  NULL },
         { "4",  NULL },
         { "5",  NULL },
         { "6",  NULL },
         { "7",  NULL },
         { "8",  NULL },
         { "9",  NULL },
         { "10", NULL },
         { "11", NULL },
         { "12", NULL },
         { "13", NULL },
         { "14", NULL },
         { "15", NULL },
         { "16", NULL },
         { "17", NULL },
         { "18", NULL },
         { "19", NULL },
         { "20", NULL },
         { "21", NULL },
         { "22", NULL },
         { "23", NULL },
         { "24", NULL },
         { "25", NULL },
         { "26", NULL },
         { "27", NULL },
         { "28", NULL },
         { "29", NULL },
         { "30", NULL },
         { "31", NULL },
         { "32", NULL },
         { "33", NULL },
         { "34", NULL },
         { "35", NULL },
         { "36", NULL },
         { "37", NULL },
         { "38", NULL },
         { "39", NULL },
         { "40", NULL },
         { "41", NULL },
         { "42", NULL },
         { "43", NULL },
         { "44", NULL },
         { "45", NULL },
         { "46", NULL },
         { "47", NULL },
         { "48", NULL },
         { "49", NULL },
         { "50", NULL },
         { "51", NULL },
         { "52", NULL },
         { "53", NULL },
         { "54", NULL },
         { "55", NULL },
         { "56", NULL },
         { "57", NULL },
         { "58", NULL },
         { "59", NULL },
         { "60", NULL },
         { "61", NULL },
         { "62", NULL },
         { "63", NULL },
         { NULL, NULL },
      },
      "0"
   },
   {
      BEETLE_OPT(memcard_right_index),
      BEETLE_OPT_MEMCARD_RIGHT_INDEX_LABEL_GL,
      NULL,
      BEETLE_OPT_MEMCARD_RIGHT_INDEX_INFO_0_GL,
      NULL,
      "memcards",
      {
         { "0",  NULL },
         { "1",  OPTION_VAL_1_O49_GL },
         { "2",  NULL },
         { "3",  NULL },
         { "4",  NULL },
         { "5",  NULL },
         { "6",  NULL },
         { "7",  NULL },
         { "8",  NULL },
         { "9",  NULL },
         { "10", NULL },
         { "11", NULL },
         { "12", NULL },
         { "13", NULL },
         { "14", NULL },
         { "15", NULL },
         { "16", NULL },
         { "17", NULL },
         { "18", NULL },
         { "19", NULL },
         { "20", NULL },
         { "21", NULL },
         { "22", NULL },
         { "23", NULL },
         { "24", NULL },
         { "25", NULL },
         { "26", NULL },
         { "27", NULL },
         { "28", NULL },
         { "29", NULL },
         { "30", NULL },
         { "31", NULL },
         { "32", NULL },
         { "33", NULL },
         { "34", NULL },
         { "35", NULL },
         { "36", NULL },
         { "37", NULL },
         { "38", NULL },
         { "39", NULL },
         { "40", NULL },
         { "41", NULL },
         { "42", NULL },
         { "43", NULL },
         { "44", NULL },
         { "45", NULL },
         { "46", NULL },
         { "47", NULL },
         { "48", NULL },
         { "49", NULL },
         { "50", NULL },
         { "51", NULL },
         { "52", NULL },
         { "53", NULL },
         { "54", NULL },
         { "55", NULL },
         { "56", NULL },
         { "57", NULL },
         { "58", NULL },
         { "59", NULL },
         { "60", NULL },
         { "61", NULL },
         { "62", NULL },
         { "63", NULL },
         { NULL, NULL },
      },
      "1"
   },
   {
      BEETLE_OPT(line_render),
      BEETLE_OPT_LINE_RENDER_LABEL_GL,
      NULL,
      BEETLE_OPT_LINE_RENDER_INFO_0_GL,
      NULL,
      "hacks",
      {
         { "default",    OPTION_VAL_DEFAULT_GL },
         { "aggressive", OPTION_VAL_AGGRESSIVE_GL },
         { "disabled",   NULL },
         { NULL, NULL },
      },
      "default"
   },
   {
      BEETLE_OPT(widescreen_hack),
      BEETLE_OPT_WIDESCREEN_HACK_LABEL_GL,
      NULL,
      BEETLE_OPT_WIDESCREEN_HACK_INFO_0_GL,
      NULL,
      "hacks",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(widescreen_hack_aspect_ratio),
      BEETLE_OPT_WIDESCREEN_HACK_ASPECT_RATIO_LABEL_GL,
      NULL,
      BEETLE_OPT_WIDESCREEN_HACK_ASPECT_RATIO_INFO_0_GL,
      NULL,
      "hacks",
      {
         { "16:9",  OPTION_VAL_16_9_GL },
         { "16:10", OPTION_VAL_16_10_GL },
         { "18:9",  OPTION_VAL_18_9_GL },
         { "19:9",  OPTION_VAL_19_9_GL },
         { "20:9",  OPTION_VAL_20_9_GL },
         { "21:9",  OPTION_VAL_21_9_GL }, // 64:27
         { "32:9",  OPTION_VAL_32_9_GL },
         { NULL,    NULL },
      },
      "16:9"
   },
#if defined(HAVE_LIGHTREC)
   {
      BEETLE_OPT(cpu_dynarec),
      BEETLE_OPT_CPU_DYNAREC_LABEL_GL,
      NULL,
      BEETLE_OPT_CPU_DYNAREC_INFO_0_GL,
      NULL,
      "hacks",
      {
         { "disabled",        OPTION_VAL_DISABLED_GL },
         { "execute",         OPTION_VAL_EXECUTE_GL },
         { "execute_one",     OPTION_VAL_EXECUTE_ONE_GL },
         { "run_interpreter", OPTION_VAL_RUN_INTERPRETER_GL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(dynarec_invalidate),
      BEETLE_OPT_DYNAREC_INVALIDATE_LABEL_GL,
      NULL,
      BEETLE_OPT_DYNAREC_INVALIDATE_INFO_0_GL,
      NULL,
      "hacks",
      {
         { "full", OPTION_VAL_FULL_GL },
         { "dma",  OPTION_VAL_DMA_GL },
         { NULL, NULL },
      },
      "full"
   },
   {
      BEETLE_OPT(dynarec_eventcycles),
      BEETLE_OPT_DYNAREC_EVENTCYCLES_LABEL_GL,
      NULL,
      BEETLE_OPT_DYNAREC_EVENTCYCLES_INFO_0_GL,
      NULL,
      "hacks",
      {
         { "128",  OPTION_VAL_128_GL },
         { "256",  NULL },
         { "384",  NULL },
         { "512",  NULL },
         { "640",  NULL },
         { "768",  NULL },
         { "896",  NULL },
         { "1024", NULL },
         { "1152", NULL },
         { "1280", NULL },
         { "1408", NULL },
         { "1536", NULL },
         { "1664", NULL },
         { "1792", NULL },
         { "1920", NULL },
         { "2048", NULL },
         { NULL, NULL },
      },
      "128"
   },
   {
      BEETLE_OPT(dynarec_spu_samples),
      BEETLE_OPT_DYNAREC_SPU_SAMPLES_LABEL_GL,
      NULL,
      BEETLE_OPT_DYNAREC_SPU_SAMPLES_INFO_0_GL,
      NULL,
      "hacks",
      {
         { "1",  OPTION_VAL_1_O49_GL },
         { "4",  NULL },
         { "16", NULL },
         { NULL, NULL },
      },
      "1"
   },
#endif
   {
      BEETLE_OPT(core_timing_fps),
      BEETLE_OPT_CORE_TIMING_FPS_LABEL_GL,
      NULL,
      BEETLE_OPT_CORE_TIMING_FPS_INFO_0_GL,
      NULL,
      "video",
      {
         { "auto_toggle",       OPTION_VAL_AUTO_TOGGLE_GL },
         { "force_progressive", OPTION_VAL_FORCE_PROGRESSIVE_GL },
         { "force_interlaced",  OPTION_VAL_FORCE_INTERLACED_GL },
      },
      "auto_toggle"
   },
   {
      BEETLE_OPT(deinterlacer),
      BEETLE_OPT_DEINTERLACER_LABEL_GL,
      NULL,
      BEETLE_OPT_DEINTERLACER_INFO_0_GL,
      NULL,
      "video",
      {
         { "weave",      OPTION_VAL_WEAVE_GL },
         { "bob",        OPTION_VAL_BOB_GL },
         { "bob_offset", OPTION_VAL_BOB_OFFSET_GL },
         { "fastmad",    OPTION_VAL_FASTMAD_GL },
         { "off",        "Off" },
         { NULL, NULL },
      },
      "weave"
   },
   {
      BEETLE_OPT(frame_duping),
      BEETLE_OPT_FRAME_DUPING_LABEL_GL,
      NULL,
      BEETLE_OPT_FRAME_DUPING_INFO_0_GL,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(pal_video_timing_override),
      BEETLE_OPT_PAL_VIDEO_TIMING_OVERRIDE_LABEL_GL,
      NULL,
      BEETLE_OPT_PAL_VIDEO_TIMING_OVERRIDE_INFO_0_GL,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(aspect_ratio),
      BEETLE_OPT_ASPECT_RATIO_LABEL_GL,
      NULL,
      BEETLE_OPT_ASPECT_RATIO_INFO_0_GL,
      NULL,
      "video",
      {
         { "corrected", OPTION_VAL_CORRECTED_GL },
         { "uncorrected", OPTION_VAL_UNCORRECTED_GL },
         { "4:3",  OPTION_VAL_4_3_GL },
         { "ntsc", OPTION_VAL_NTSC_GL },
      },
      "corrected"
   },
   {
      BEETLE_OPT(crop_overscan),
      BEETLE_OPT_CROP_OVERSCAN_LABEL_GL,
      NULL,
      BEETLE_OPT_CROP_OVERSCAN_INFO_0_GL,
      NULL,
      "video",
      {
         { "disabled", "Disabled" },
         { "static",   OPTION_VAL_STATIC_GL },
         { "smart",    OPTION_VAL_SMART_GL },
         { NULL, NULL },
      },
      "smart"
   },
   {
      BEETLE_OPT(image_crop),
      BEETLE_OPT_IMAGE_CROP_LABEL_GL,
      NULL,
      BEETLE_OPT_IMAGE_CROP_INFO_0_GL,
      NULL,
      "video",
      {
         { "disabled", "0" },
         { "1px",      OPTION_VAL_1PX_GL },
         { "2px",      OPTION_VAL_2PX_GL },
         { "3px",      OPTION_VAL_3PX_GL },
         { "4px",      OPTION_VAL_4PX_GL },
         { "5px",      OPTION_VAL_5PX_GL },
         { "6px",      OPTION_VAL_6PX_GL },
         { "7px",      OPTION_VAL_7PX_GL },
         { "8px",      OPTION_VAL_8PX_GL },
         { "9px",      OPTION_VAL_9PX_GL },
         { "10px",     OPTION_VAL_10PX_GL },
         { "11px",     OPTION_VAL_11PX_GL },
         { "12px",     OPTION_VAL_12PX_GL },
         { "13px",     OPTION_VAL_13PX_GL },
         { "14px",     OPTION_VAL_14PX_GL },
         { "15px",     OPTION_VAL_15PX_GL },
         { "16px",     OPTION_VAL_16PX_GL },
         { "17px",     OPTION_VAL_17PX_GL },
         { "18px",     OPTION_VAL_18PX_GL },
         { "19px",     OPTION_VAL_19PX_GL },
         { "20px",     OPTION_VAL_20PX_GL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(image_offset),
      BEETLE_OPT_IMAGE_OFFSET_LABEL_GL,
      NULL,
      BEETLE_OPT_IMAGE_OFFSET_INFO_0_GL,
      NULL,
      "video",
      {
         { "-12px",    OPTION_VAL_12PX_O64_GL },
         { "-11px",    OPTION_VAL_11PX_O64_GL },
         { "-10px",    OPTION_VAL_10PX_O64_GL },
         { "-9px",     OPTION_VAL_9PX_O64_GL },
         { "-8px",     OPTION_VAL_8PX_O64_GL },
         { "-7px",     OPTION_VAL_7PX_O64_GL },
         { "-6px",     OPTION_VAL_6PX_O64_GL },
         { "-5px",     OPTION_VAL_5PX_O64_GL },
         { "-4px",     OPTION_VAL_4PX_O64_GL },
         { "-3px",     OPTION_VAL_3PX_O64_GL },
         { "-2px",     OPTION_VAL_2PX_O64_GL },
         { "-1px",     OPTION_VAL_1PX_O64_GL },
         { "disabled", "0" },
         { "+1px",     OPTION_VAL_1PX_O64_0_GL },
         { "+2px",     OPTION_VAL_2PX_O64_0_GL },
         { "+3px",     OPTION_VAL_3PX_O64_0_GL },
         { "+4px",     OPTION_VAL_4PX_O64_0_GL },
         { "+5px",     OPTION_VAL_5PX_O64_0_GL },
         { "+6px",     OPTION_VAL_6PX_O64_0_GL },
         { "+7px",     OPTION_VAL_7PX_O64_0_GL },
         { "+8px",     OPTION_VAL_8PX_O64_0_GL },
         { "+9px",     OPTION_VAL_9PX_O64_0_GL },
         { "+10px",    OPTION_VAL_10PX_O64_0_GL },
         { "+11px",    OPTION_VAL_11PX_O64_0_GL },
         { "+12px",    OPTION_VAL_12PX_O64_0_GL },
         { NULL, NULL },
      },
      "disabled"
   },
#if defined(HAVE_OPENGL) || defined(HAVE_OPENGLES) || defined(HAVE_VULKAN)
   {
      BEETLE_OPT(image_offset_cycles),
      BEETLE_OPT_IMAGE_OFFSET_CYCLES_LABEL_GL,
      NULL,
      BEETLE_OPT_IMAGE_OFFSET_CYCLES_INFO_0_GL,
      NULL,
      "video",
      {
         { "-40",      NULL },
         { "-39",      NULL },
         { "-38",      NULL },
         { "-37",      NULL },
         { "-36",      NULL },
         { "-35",      NULL },
         { "-34",      NULL },
         { "-33",      NULL },
         { "-32",      NULL },
         { "-31",      NULL },
         { "-30",      NULL },
         { "-29",      NULL },
         { "-28",      NULL },
         { "-27",      NULL },
         { "-26",      NULL },
         { "-25",      NULL },
         { "-24",      NULL },
         { "-23",      NULL },
         { "-22",      NULL },
         { "-21",      NULL },
         { "-20",      NULL },
         { "-19",      NULL },
         { "-18",      NULL },
         { "-17",      NULL },
         { "-16",      NULL },
         { "-15",      NULL },
         { "-14",      NULL },
         { "-13",      NULL },
         { "-12",      NULL },
         { "-11",      NULL },
         { "-10",      NULL },
         { "-9",       NULL },
         { "-8",       NULL },
         { "-7",       NULL },
         { "-6",       NULL },
         { "-5",       NULL },
         { "-4",       NULL },
         { "-3",       NULL },
         { "-2",       NULL },
         { "-1",       NULL },
         { "0",        "0" },
         { "+1",       NULL },
         { "+2",       NULL },
         { "+3",       NULL },
         { "+4",       NULL },
         { "+5",       NULL },
         { "+6",       NULL },
         { "+7",       NULL },
         { "+8",       NULL },
         { "+9",       NULL },
         { "+10",      NULL },
         { "+11",      NULL },
         { "+12",      NULL },
         { "+13",      NULL },
         { "+14",      NULL },
         { "+15",      NULL },
         { "+16",      NULL },
         { "+17",      NULL },
         { "+18",      NULL },
         { "+19",      NULL },
         { "+20",      NULL },
         { "+21",      NULL },
         { "+22",      NULL },
         { "+23",      NULL },
         { "+24",      NULL },
         { "+25",      NULL },
         { "+26",      NULL },
         { "+27",      NULL },
         { "+28",      NULL },
         { "+29",      NULL },
         { "+30",      NULL },
         { "+31",      NULL },
         { "+32",      NULL },
         { "+33",      NULL },
         { "+34",      NULL },
         { "+35",      NULL },
         { "+36",      NULL },
         { "+37",      NULL },
         { "+38",      NULL },
         { "+39",      NULL },
         { "+40",      NULL },
         { NULL, NULL},
      },
      "0"
   },
#endif
   {
      BEETLE_OPT(initial_scanline),
      BEETLE_OPT_INITIAL_SCANLINE_LABEL_GL,
      NULL,
      BEETLE_OPT_INITIAL_SCANLINE_INFO_0_GL,
      NULL,
      "video",
      {
         { "0",  OPTION_VAL_0_O48_GL },
         { "1",  NULL },
         { "2",  NULL },
         { "3",  NULL },
         { "4",  NULL },
         { "5",  NULL },
         { "6",  NULL },
         { "7",  NULL },
         { "8",  NULL },
         { "9",  NULL },
         { "10", NULL },
         { "11", NULL },
         { "12", NULL },
         { "13", NULL },
         { "14", NULL },
         { "15", NULL },
         { "16", NULL },
         { "17", NULL },
         { "18", NULL },
         { "19", NULL },
         { "20", NULL },
         { "21", NULL },
         { "22", NULL },
         { "23", NULL },
         { "24", NULL },
         { "25", NULL },
         { "26", NULL },
         { "27", NULL },
         { "28", NULL },
         { "29", NULL },
         { "30", NULL },
         { "31", NULL },
         { "32", NULL },
         { "33", NULL },
         { "34", NULL },
         { "35", NULL },
         { "36", NULL },
         { "37", NULL },
         { "38", NULL },
         { "39", NULL },
         { "40", NULL },
         { NULL, NULL },
      },
      "0"
   },
   {
      BEETLE_OPT(last_scanline),
      BEETLE_OPT_LAST_SCANLINE_LABEL_GL,
      NULL,
      BEETLE_OPT_LAST_SCANLINE_INFO_0_GL,
      NULL,
      "video",
      {
         { "210", NULL },
         { "211", NULL },
         { "212", NULL },
         { "213", NULL },
         { "214", NULL },
         { "215", NULL },
         { "216", NULL },
         { "217", NULL },
         { "218", NULL },
         { "219", NULL },
         { "220", NULL },
         { "221", NULL },
         { "222", NULL },
         { "223", NULL },
         { "224", NULL },
         { "225", NULL },
         { "226", NULL },
         { "227", NULL },
         { "228", NULL },
         { "229", NULL },
         { "230", NULL },
         { "231", NULL },
         { "232", NULL },
         { "233", NULL },
         { "234", NULL },
         { "235", NULL },
         { "236", NULL },
         { "237", NULL },
         { "238", NULL },
         { "239", OPTION_VAL_239_GL },
         { NULL, NULL },
      },
      "239"
   },
   {
      BEETLE_OPT(initial_scanline_pal),
      BEETLE_OPT_INITIAL_SCANLINE_PAL_LABEL_GL,
      NULL,
      BEETLE_OPT_INITIAL_SCANLINE_PAL_INFO_0_GL,
      NULL,
      "video",
      {
         { "0",  OPTION_VAL_0_O48_GL },
         { "1",  NULL },
         { "2",  NULL },
         { "3",  NULL },
         { "4",  NULL },
         { "5",  NULL },
         { "6",  NULL },
         { "7",  NULL },
         { "8",  NULL },
         { "9",  NULL },
         { "10", NULL },
         { "11", NULL },
         { "12", NULL },
         { "13", NULL },
         { "14", NULL },
         { "15", NULL },
         { "16", NULL },
         { "17", NULL },
         { "18", NULL },
         { "19", NULL },
         { "20", NULL },
         { "21", NULL },
         { "22", NULL },
         { "23", NULL },
         { "24", NULL },
         { "25", NULL },
         { "26", NULL },
         { "27", NULL },
         { "28", NULL },
         { "29", NULL },
         { "30", NULL },
         { "31", NULL },
         { "32", NULL },
         { "33", NULL },
         { "34", NULL },
         { "35", NULL },
         { "36", NULL },
         { "37", NULL },
         { "38", NULL },
         { "39", NULL },
         { "40", NULL },
         { NULL, NULL },
      },
      "0"
   },
   {
      BEETLE_OPT(last_scanline_pal),
      BEETLE_OPT_LAST_SCANLINE_PAL_LABEL_GL,
      NULL,
      BEETLE_OPT_LAST_SCANLINE_PAL_INFO_0_GL,
      NULL,
      "video",
      {
         { "230", NULL },
         { "231", NULL },
         { "232", NULL },
         { "233", NULL },
         { "234", NULL },
         { "235", NULL },
         { "236", NULL },
         { "237", NULL },
         { "238", NULL },
         { "239", NULL },
         { "240", NULL },
         { "241", NULL },
         { "242", NULL },
         { "243", NULL },
         { "244", NULL },
         { "245", NULL },
         { "246", NULL },
         { "247", NULL },
         { "248", NULL },
         { "249", NULL },
         { "250", NULL },
         { "251", NULL },
         { "252", NULL },
         { "253", NULL },
         { "254", NULL },
         { "255", NULL },
         { "256", NULL },
         { "257", NULL },
         { "258", NULL },
         { "259", NULL },
         { "260", NULL },
         { "261", NULL },
         { "262", NULL },
         { "263", NULL },
         { "264", NULL },
         { "265", NULL },
         { "266", NULL },
         { "267", NULL },
         { "268", NULL },
         { "269", NULL },
         { "270", NULL },
         { "271", NULL },
         { "272", NULL },
         { "273", NULL },
         { "274", NULL },
         { "275", NULL },
         { "276", NULL },
         { "277", NULL },
         { "278", NULL },
         { "279", NULL },
         { "280", NULL },
         { "281", NULL },
         { "282", NULL },
         { "283", NULL },
         { "284", NULL },
         { "285", NULL },
         { "286", NULL },
         { "287", OPTION_VAL_287_GL },
         { NULL, NULL },
      },
      "287"
   },
   { NULL, NULL, NULL, NULL, NULL, NULL, {{0}}, NULL },
};
struct retro_core_options_v2 options_gl = {
   option_cats_gl,
   option_defs_gl
};

/* RETRO_LANGUAGE_HE */

#define CATEGORY_SYSTEM_LABEL_HE NULL
#define CATEGORY_SYSTEM_INFO_0_HE NULL
#define CATEGORY_VIDEO_LABEL_HE "וידאו"
#define CATEGORY_VIDEO_INFO_0_HE NULL
#define CATEGORY_PGXP_LABEL_HE NULL
#define CATEGORY_PGXP_INFO_0_HE NULL
#define CATEGORY_OSD_LABEL_HE NULL
#define CATEGORY_OSD_INFO_0_HE NULL
#define CATEGORY_INPUT_LABEL_HE "קלט"
#define CATEGORY_INPUT_INFO_0_HE NULL
#define CATEGORY_MEMCARDS_LABEL_HE NULL
#define CATEGORY_MEMCARDS_INFO_0_HE NULL
#define CATEGORY_HACKS_LABEL_HE NULL
#define CATEGORY_HACKS_INFO_0_HE NULL
#define BEETLE_OPT_CPU_FREQ_SCALE_LABEL_HE NULL
#define BEETLE_OPT_CPU_FREQ_SCALE_INFO_0_HE NULL
#define OPTION_VAL_50_HE NULL
#define OPTION_VAL_60_HE NULL
#define OPTION_VAL_70_HE NULL
#define OPTION_VAL_80_HE NULL
#define OPTION_VAL_90_HE NULL
#define OPTION_VAL_100_HE NULL
#define OPTION_VAL_110_HE NULL
#define OPTION_VAL_120_HE NULL
#define OPTION_VAL_130_HE NULL
#define OPTION_VAL_140_HE NULL
#define OPTION_VAL_150_HE NULL
#define OPTION_VAL_160_HE NULL
#define OPTION_VAL_170_HE NULL
#define OPTION_VAL_180_HE NULL
#define OPTION_VAL_190_HE NULL
#define OPTION_VAL_200_HE NULL
#define OPTION_VAL_210_HE NULL
#define OPTION_VAL_220_HE NULL
#define OPTION_VAL_230_HE NULL
#define OPTION_VAL_240_HE NULL
#define OPTION_VAL_250_HE NULL
#define OPTION_VAL_260_HE NULL
#define OPTION_VAL_270_HE NULL
#define OPTION_VAL_280_HE NULL
#define OPTION_VAL_290_HE NULL
#define OPTION_VAL_300_HE NULL
#define OPTION_VAL_310_HE NULL
#define OPTION_VAL_320_HE NULL
#define OPTION_VAL_330_HE NULL
#define OPTION_VAL_340_HE NULL
#define OPTION_VAL_350_HE NULL
#define OPTION_VAL_360_HE NULL
#define OPTION_VAL_370_HE NULL
#define OPTION_VAL_380_HE NULL
#define OPTION_VAL_390_HE NULL
#define OPTION_VAL_400_HE NULL
#define OPTION_VAL_410_HE NULL
#define OPTION_VAL_420_HE NULL
#define OPTION_VAL_430_HE NULL
#define OPTION_VAL_440_HE NULL
#define OPTION_VAL_450_HE NULL
#define OPTION_VAL_460_HE NULL
#define OPTION_VAL_470_HE NULL
#define OPTION_VAL_480_HE NULL
#define OPTION_VAL_490_HE NULL
#define OPTION_VAL_500_HE NULL
#define OPTION_VAL_510_HE NULL
#define OPTION_VAL_520_HE NULL
#define OPTION_VAL_530_HE NULL
#define OPTION_VAL_540_HE NULL
#define OPTION_VAL_550_HE NULL
#define OPTION_VAL_560_HE NULL
#define OPTION_VAL_570_HE NULL
#define OPTION_VAL_580_HE NULL
#define OPTION_VAL_590_HE NULL
#define OPTION_VAL_600_HE NULL
#define OPTION_VAL_610_HE NULL
#define OPTION_VAL_620_HE NULL
#define OPTION_VAL_630_HE NULL
#define OPTION_VAL_640_HE NULL
#define OPTION_VAL_650_HE NULL
#define OPTION_VAL_660_HE NULL
#define OPTION_VAL_670_HE NULL
#define OPTION_VAL_680_HE NULL
#define OPTION_VAL_690_HE NULL
#define OPTION_VAL_700_HE NULL
#define OPTION_VAL_710_HE NULL
#define OPTION_VAL_720_HE NULL
#define OPTION_VAL_730_HE NULL
#define OPTION_VAL_740_HE NULL
#define OPTION_VAL_750_HE NULL
#define BEETLE_OPT_GPU_OVERCLOCK_LABEL_HE NULL
#define BEETLE_OPT_GPU_OVERCLOCK_INFO_0_HE NULL
#define OPTION_VAL_1X_NATIVE_HE NULL
#define OPTION_VAL_2X_HE NULL
#define OPTION_VAL_4X_HE NULL
#define OPTION_VAL_8X_HE NULL
#define OPTION_VAL_16X_HE NULL
#define OPTION_VAL_32X_HE NULL
#define BEETLE_OPT_GTE_OVERCLOCK_LABEL_HE NULL
#define BEETLE_OPT_GTE_OVERCLOCK_INFO_0_HE NULL
#define BEETLE_OPT_CD_ACCESS_METHOD_LABEL_HE NULL
#define BEETLE_OPT_CD_ACCESS_METHOD_INFO_0_HE NULL
#define OPTION_VAL_SYNC_HE NULL
#define OPTION_VAL_ASYNC_HE NULL
#define OPTION_VAL_PRECACHE_HE NULL
#define BEETLE_OPT_CD_FASTLOAD_LABEL_HE NULL
#define BEETLE_OPT_CD_FASTLOAD_INFO_0_HE NULL
#define OPTION_VAL_2X_NATIVE_HE NULL
#define OPTION_VAL_6X_HE NULL
#define OPTION_VAL_10X_HE NULL
#define OPTION_VAL_12X_HE NULL
#define OPTION_VAL_14X_HE NULL
#define BEETLE_OPT_SKIP_BIOS_LABEL_HE NULL
#define BEETLE_OPT_SKIP_BIOS_INFO_0_HE NULL
#define BEETLE_OPT_OVERRIDE_BIOS_LABEL_HE NULL
#define BEETLE_OPT_OVERRIDE_BIOS_INFO_0_HE NULL
#define OPTION_VAL_PSXONPSP_HE NULL
#define OPTION_VAL_PS1_ROM_HE NULL
#define OPTION_VAL_OPENBIOS_HE NULL
#define BEETLE_OPT_REGION_LABEL_HE NULL
#define BEETLE_OPT_REGION_INFO_0_HE NULL
#define OPTION_VAL_AUTO_HE NULL
#define OPTION_VAL_NTSC_J_HE NULL
#define OPTION_VAL_NTSC_U_HE NULL
#define OPTION_VAL_PAL_HE NULL
#define BEETLE_OPT_INTERNAL_RESOLUTION_LABEL_HE NULL
#define BEETLE_OPT_INTERNAL_RESOLUTION_INFO_0_HE NULL
#define BEETLE_OPT_RENDERER_LABEL_HE NULL
#define BEETLE_OPT_RENDERER_INFO_0_HE NULL
#define OPTION_VAL_HARDWARE_HE NULL
#define OPTION_VAL_HARDWARE_GL_HE NULL
#define OPTION_VAL_HARDWARE_VK_HE NULL
#define OPTION_VAL_SOFTWARE_HE NULL
#define BEETLE_OPT_RENDERER_SOFTWARE_FB_LABEL_HE NULL
#define BEETLE_OPT_RENDERER_SOFTWARE_FB_INFO_0_HE NULL
#define BEETLE_OPT_DEPTH_LABEL_HE NULL
#define BEETLE_OPT_DEPTH_INFO_0_HE NULL
#define OPTION_VAL_16BPP_NATIVE_HE NULL
#define OPTION_VAL_32BPP_HE NULL
#define BEETLE_OPT_DITHER_MODE_LABEL_HE NULL
#define BEETLE_OPT_DITHER_MODE_INFO_0_HE NULL
#define OPTION_VAL_INTERNAL_RESOLUTION_HE NULL
#define BEETLE_OPT_SCALED_UV_OFFSET_LABEL_HE NULL
#define BEETLE_OPT_SCALED_UV_OFFSET_INFO_0_HE NULL
#define BEETLE_OPT_FILTER_LABEL_HE NULL
#define BEETLE_OPT_FILTER_INFO_0_HE NULL
#define OPTION_VAL_NEAREST_HE NULL
#define OPTION_VAL_SABR_HE NULL
#define OPTION_VAL_XBR_HE NULL
#define OPTION_VAL_BILINEAR_HE NULL
#define OPTION_VAL_3_POINT_HE NULL
#define OPTION_VAL_JINC2_HE NULL
#define BEETLE_OPT_FILTER_EXCLUDE_SPRITE_LABEL_HE NULL
#define BEETLE_OPT_FILTER_EXCLUDE_SPRITE_INFO_0_HE NULL
#define OPTION_VAL_OPAQUE_HE NULL
#define OPTION_VAL_ALL_HE NULL
#define BEETLE_OPT_FILTER_EXCLUDE_2D_POLYGON_LABEL_HE NULL
#define BEETLE_OPT_FILTER_EXCLUDE_2D_POLYGON_INFO_0_HE NULL
#define BEETLE_OPT_ADAPTIVE_SMOOTHING_LABEL_HE NULL
#define BEETLE_OPT_ADAPTIVE_SMOOTHING_INFO_0_HE NULL
#define BEETLE_OPT_SUPER_SAMPLING_LABEL_HE NULL
#define BEETLE_OPT_SUPER_SAMPLING_INFO_0_HE NULL
#define BEETLE_OPT_MSAA_LABEL_HE NULL
#define BEETLE_OPT_MSAA_INFO_0_HE NULL
#define OPTION_VAL_1X_HE NULL
#define BEETLE_OPT_MDEC_YUV_LABEL_HE NULL
#define BEETLE_OPT_MDEC_YUV_INFO_0_HE NULL
#define BEETLE_OPT_TRACK_TEXTURES_LABEL_HE NULL
#define BEETLE_OPT_TRACK_TEXTURES_INFO_0_HE NULL
#define BEETLE_OPT_DUMP_TEXTURES_LABEL_HE NULL
#define BEETLE_OPT_DUMP_TEXTURES_INFO_0_HE NULL
#define BEETLE_OPT_REPLACE_TEXTURES_LABEL_HE NULL
#define BEETLE_OPT_REPLACE_TEXTURES_INFO_0_HE NULL
#define BEETLE_OPT_PGXP_MODE_LABEL_HE NULL
#define BEETLE_OPT_PGXP_MODE_INFO_0_HE NULL
#define OPTION_VAL_MEMORY_ONLY_HE NULL
#define OPTION_VAL_MEMORY_CPU_HE NULL
#define BEETLE_OPT_PGXP_2D_TOL_LABEL_HE NULL
#define BEETLE_OPT_PGXP_2D_TOL_INFO_0_HE NULL
#define OPTION_VAL_0PX_HE NULL
#define OPTION_VAL_1PX_HE NULL
#define OPTION_VAL_2PX_HE NULL
#define OPTION_VAL_3PX_HE NULL
#define OPTION_VAL_4PX_HE NULL
#define OPTION_VAL_5PX_HE NULL
#define OPTION_VAL_6PX_HE NULL
#define OPTION_VAL_7PX_HE NULL
#define OPTION_VAL_8PX_HE NULL
#define BEETLE_OPT_PGXP_NCLIP_LABEL_HE NULL
#define BEETLE_OPT_PGXP_NCLIP_INFO_0_HE NULL
#define BEETLE_OPT_PGXP_VERTEX_LABEL_HE NULL
#define BEETLE_OPT_PGXP_VERTEX_INFO_0_HE NULL
#define BEETLE_OPT_PGXP_TEXTURE_LABEL_HE NULL
#define BEETLE_OPT_PGXP_TEXTURE_INFO_0_HE NULL
#define BEETLE_OPT_DISPLAY_INTERNAL_FPS_LABEL_HE NULL
#define BEETLE_OPT_DISPLAY_INTERNAL_FPS_INFO_0_HE NULL
#define BEETLE_OPT_DISPLAY_OSD_LABEL_HE NULL
#define BEETLE_OPT_DISPLAY_OSD_INFO_0_HE NULL
#define BEETLE_OPT_DISPLAY_VRAM_LABEL_HE NULL
#define BEETLE_OPT_DISPLAY_VRAM_INFO_0_HE NULL
#define BEETLE_OPT_ANALOG_CALIBRATION_LABEL_HE NULL
#define BEETLE_OPT_ANALOG_CALIBRATION_INFO_0_HE NULL
#define BEETLE_OPT_ANALOG_TOGGLE_LABEL_HE NULL
#define BEETLE_OPT_ANALOG_TOGGLE_INFO_0_HE NULL
#define OPTION_VAL_ENABLED_ANALOG_HE NULL
#define BEETLE_OPT_ANALOG_TOGGLE_COMBO_LABEL_HE NULL
#define BEETLE_OPT_ANALOG_TOGGLE_COMBO_INFO_0_HE NULL
#define OPTION_VAL_L1_L2_R1_R2_START_SELECT_HE NULL
#define OPTION_VAL_L1_R1_SELECT_HE NULL
#define OPTION_VAL_L1_R1_START_HE NULL
#define OPTION_VAL_L1_R1_L3_HE NULL
#define OPTION_VAL_L1_R1_R3_HE NULL
#define OPTION_VAL_L2_R2_SELECT_HE NULL
#define OPTION_VAL_L2_R2_START_HE NULL
#define OPTION_VAL_L2_R2_L3_HE NULL
#define OPTION_VAL_L2_R2_R3_HE NULL
#define OPTION_VAL_L3_R3_HE NULL
#define BEETLE_OPT_ANALOG_TOGGLE_HOLD_LABEL_HE NULL
#define BEETLE_OPT_ANALOG_TOGGLE_HOLD_INFO_0_HE NULL
#define OPTION_VAL_0_HE NULL
#define OPTION_VAL_1_HE NULL
#define OPTION_VAL_2_HE NULL
#define OPTION_VAL_3_HE NULL
#define OPTION_VAL_4_HE NULL
#define OPTION_VAL_5_HE NULL
#define BEETLE_OPT_ENABLE_MULTITAP_PORT1_LABEL_HE NULL
#define BEETLE_OPT_ENABLE_MULTITAP_PORT1_INFO_0_HE NULL
#define BEETLE_OPT_ENABLE_MULTITAP_PORT2_LABEL_HE NULL
#define BEETLE_OPT_ENABLE_MULTITAP_PORT2_INFO_0_HE NULL
#define BEETLE_OPT_GUN_INPUT_MODE_LABEL_HE NULL
#define BEETLE_OPT_GUN_INPUT_MODE_INFO_0_HE NULL
#define OPTION_VAL_LIGHTGUN_HE NULL
#define OPTION_VAL_TOUCHSCREEN_HE NULL
#define BEETLE_OPT_GUN_CURSOR_LABEL_HE NULL
#define BEETLE_OPT_GUN_CURSOR_INFO_0_HE NULL
#define OPTION_VAL_CROSS_HE NULL
#define OPTION_VAL_DOT_HE NULL
#define BEETLE_OPT_CROSSHAIR_COLOR_P1_LABEL_HE NULL
#define BEETLE_OPT_CROSSHAIR_COLOR_P1_INFO_0_HE NULL
#define OPTION_VAL_RED_HE NULL
#define OPTION_VAL_BLUE_HE NULL
#define OPTION_VAL_GREEN_HE NULL
#define OPTION_VAL_ORANGE_HE NULL
#define OPTION_VAL_YELLOW_HE NULL
#define OPTION_VAL_CYAN_HE NULL
#define OPTION_VAL_PINK_HE NULL
#define OPTION_VAL_PURPLE_HE NULL
#define OPTION_VAL_BLACK_HE NULL
#define OPTION_VAL_WHITE_HE NULL
#define BEETLE_OPT_CROSSHAIR_COLOR_P2_LABEL_HE NULL
#define BEETLE_OPT_CROSSHAIR_COLOR_P2_INFO_0_HE NULL
#define BEETLE_OPT_MOUSE_SENSITIVITY_LABEL_HE NULL
#define BEETLE_OPT_MOUSE_SENSITIVITY_INFO_0_HE NULL
#define OPTION_VAL_5_O42_HE NULL
#define OPTION_VAL_10_HE NULL
#define OPTION_VAL_15_HE NULL
#define OPTION_VAL_20_HE NULL
#define OPTION_VAL_25_HE NULL
#define OPTION_VAL_30_HE NULL
#define OPTION_VAL_35_HE NULL
#define OPTION_VAL_40_HE NULL
#define OPTION_VAL_45_HE NULL
#define OPTION_VAL_55_HE NULL
#define OPTION_VAL_65_HE NULL
#define OPTION_VAL_75_HE NULL
#define OPTION_VAL_85_HE NULL
#define OPTION_VAL_95_HE NULL
#define OPTION_VAL_100_O42_HE NULL
#define OPTION_VAL_105_HE NULL
#define OPTION_VAL_115_HE NULL
#define OPTION_VAL_125_HE NULL
#define OPTION_VAL_135_HE NULL
#define OPTION_VAL_145_HE NULL
#define OPTION_VAL_155_HE NULL
#define OPTION_VAL_165_HE NULL
#define OPTION_VAL_175_HE NULL
#define OPTION_VAL_185_HE NULL
#define OPTION_VAL_195_HE NULL
#define BEETLE_OPT_NEGCON_RESPONSE_LABEL_HE NULL
#define BEETLE_OPT_NEGCON_RESPONSE_INFO_0_HE NULL
#define OPTION_VAL_LINEAR_HE NULL
#define OPTION_VAL_QUADRATIC_HE NULL
#define OPTION_VAL_CUBIC_HE NULL
#define BEETLE_OPT_NEGCON_DEADZONE_LABEL_HE NULL
#define BEETLE_OPT_NEGCON_DEADZONE_INFO_0_HE NULL
#define OPTION_VAL_0_O44_HE NULL
#define BEETLE_OPT_USE_MEDNAFEN_MEMCARD0_METHOD_LABEL_HE NULL
#define BEETLE_OPT_USE_MEDNAFEN_MEMCARD0_METHOD_INFO_0_HE NULL
#define OPTION_VAL_LIBRETRO_HE NULL
#define OPTION_VAL_MEDNAFEN_HE NULL
#define BEETLE_OPT_ENABLE_MEMCARD1_LABEL_HE NULL
#define BEETLE_OPT_ENABLE_MEMCARD1_INFO_0_HE NULL
#define BEETLE_OPT_SHARED_MEMORY_CARDS_LABEL_HE NULL
#define BEETLE_OPT_SHARED_MEMORY_CARDS_INFO_0_HE NULL
#define BEETLE_OPT_MEMCARD_LEFT_INDEX_LABEL_HE NULL
#define BEETLE_OPT_MEMCARD_LEFT_INDEX_INFO_0_HE NULL
#define OPTION_VAL_0_O48_HE NULL
#define BEETLE_OPT_MEMCARD_RIGHT_INDEX_LABEL_HE NULL
#define BEETLE_OPT_MEMCARD_RIGHT_INDEX_INFO_0_HE NULL
#define OPTION_VAL_1_O49_HE NULL
#define BEETLE_OPT_LINE_RENDER_LABEL_HE NULL
#define BEETLE_OPT_LINE_RENDER_INFO_0_HE NULL
#define OPTION_VAL_DEFAULT_HE NULL
#define OPTION_VAL_AGGRESSIVE_HE NULL
#define BEETLE_OPT_WIDESCREEN_HACK_LABEL_HE NULL
#define BEETLE_OPT_WIDESCREEN_HACK_INFO_0_HE NULL
#define BEETLE_OPT_WIDESCREEN_HACK_ASPECT_RATIO_LABEL_HE NULL
#define BEETLE_OPT_WIDESCREEN_HACK_ASPECT_RATIO_INFO_0_HE NULL
#define OPTION_VAL_16_9_HE NULL
#define OPTION_VAL_16_10_HE NULL
#define OPTION_VAL_18_9_HE NULL
#define OPTION_VAL_19_9_HE NULL
#define OPTION_VAL_20_9_HE NULL
#define OPTION_VAL_21_9_HE NULL
#define OPTION_VAL_32_9_HE NULL
#define BEETLE_OPT_CPU_DYNAREC_LABEL_HE NULL
#define BEETLE_OPT_CPU_DYNAREC_INFO_0_HE NULL
#define OPTION_VAL_DISABLED_HE NULL
#define OPTION_VAL_EXECUTE_HE NULL
#define OPTION_VAL_EXECUTE_ONE_HE NULL
#define OPTION_VAL_RUN_INTERPRETER_HE NULL
#define BEETLE_OPT_DYNAREC_INVALIDATE_LABEL_HE NULL
#define BEETLE_OPT_DYNAREC_INVALIDATE_INFO_0_HE NULL
#define OPTION_VAL_FULL_HE NULL
#define OPTION_VAL_DMA_HE NULL
#define BEETLE_OPT_DYNAREC_EVENTCYCLES_LABEL_HE NULL
#define BEETLE_OPT_DYNAREC_EVENTCYCLES_INFO_0_HE NULL
#define OPTION_VAL_128_HE NULL
#define BEETLE_OPT_DYNAREC_SPU_SAMPLES_LABEL_HE NULL
#define BEETLE_OPT_DYNAREC_SPU_SAMPLES_INFO_0_HE NULL
#define BEETLE_OPT_CORE_TIMING_FPS_LABEL_HE NULL
#define BEETLE_OPT_CORE_TIMING_FPS_INFO_0_HE NULL
#define OPTION_VAL_AUTO_TOGGLE_HE NULL
#define OPTION_VAL_FORCE_PROGRESSIVE_HE NULL
#define OPTION_VAL_FORCE_INTERLACED_HE NULL
#define BEETLE_OPT_DEINTERLACER_LABEL_HE NULL
#define BEETLE_OPT_DEINTERLACER_INFO_0_HE NULL
#define OPTION_VAL_WEAVE_HE NULL
#define OPTION_VAL_BOB_HE NULL
#define OPTION_VAL_BOB_OFFSET_HE NULL
#define OPTION_VAL_FASTMAD_HE NULL
#define BEETLE_OPT_FRAME_DUPING_LABEL_HE NULL
#define BEETLE_OPT_FRAME_DUPING_INFO_0_HE NULL
#define BEETLE_OPT_PAL_VIDEO_TIMING_OVERRIDE_LABEL_HE NULL
#define BEETLE_OPT_PAL_VIDEO_TIMING_OVERRIDE_INFO_0_HE NULL
#define BEETLE_OPT_ASPECT_RATIO_LABEL_HE NULL
#define BEETLE_OPT_ASPECT_RATIO_INFO_0_HE NULL
#define OPTION_VAL_CORRECTED_HE NULL
#define OPTION_VAL_UNCORRECTED_HE NULL
#define OPTION_VAL_4_3_HE NULL
#define OPTION_VAL_NTSC_HE NULL
#define BEETLE_OPT_CROP_OVERSCAN_LABEL_HE NULL
#define BEETLE_OPT_CROP_OVERSCAN_INFO_0_HE NULL
#define OPTION_VAL_STATIC_HE NULL
#define OPTION_VAL_SMART_HE NULL
#define BEETLE_OPT_IMAGE_CROP_LABEL_HE NULL
#define BEETLE_OPT_IMAGE_CROP_INFO_0_HE NULL
#define OPTION_VAL_9PX_HE NULL
#define OPTION_VAL_10PX_HE NULL
#define OPTION_VAL_11PX_HE NULL
#define OPTION_VAL_12PX_HE NULL
#define OPTION_VAL_13PX_HE NULL
#define OPTION_VAL_14PX_HE NULL
#define OPTION_VAL_15PX_HE NULL
#define OPTION_VAL_16PX_HE NULL
#define OPTION_VAL_17PX_HE NULL
#define OPTION_VAL_18PX_HE NULL
#define OPTION_VAL_19PX_HE NULL
#define OPTION_VAL_20PX_HE NULL
#define BEETLE_OPT_IMAGE_OFFSET_LABEL_HE NULL
#define BEETLE_OPT_IMAGE_OFFSET_INFO_0_HE NULL
#define OPTION_VAL_12PX_O64_HE NULL
#define OPTION_VAL_11PX_O64_HE NULL
#define OPTION_VAL_10PX_O64_HE NULL
#define OPTION_VAL_9PX_O64_HE NULL
#define OPTION_VAL_8PX_O64_HE NULL
#define OPTION_VAL_7PX_O64_HE NULL
#define OPTION_VAL_6PX_O64_HE NULL
#define OPTION_VAL_5PX_O64_HE NULL
#define OPTION_VAL_4PX_O64_HE NULL
#define OPTION_VAL_3PX_O64_HE NULL
#define OPTION_VAL_2PX_O64_HE NULL
#define OPTION_VAL_1PX_O64_HE NULL
#define OPTION_VAL_1PX_O64_0_HE NULL
#define OPTION_VAL_2PX_O64_0_HE NULL
#define OPTION_VAL_3PX_O64_0_HE NULL
#define OPTION_VAL_4PX_O64_0_HE NULL
#define OPTION_VAL_5PX_O64_0_HE NULL
#define OPTION_VAL_6PX_O64_0_HE NULL
#define OPTION_VAL_7PX_O64_0_HE NULL
#define OPTION_VAL_8PX_O64_0_HE NULL
#define OPTION_VAL_9PX_O64_0_HE NULL
#define OPTION_VAL_10PX_O64_0_HE NULL
#define OPTION_VAL_11PX_O64_0_HE NULL
#define OPTION_VAL_12PX_O64_0_HE NULL
#define BEETLE_OPT_IMAGE_OFFSET_CYCLES_LABEL_HE NULL
#define BEETLE_OPT_IMAGE_OFFSET_CYCLES_INFO_0_HE NULL
#define BEETLE_OPT_INITIAL_SCANLINE_LABEL_HE NULL
#define BEETLE_OPT_INITIAL_SCANLINE_INFO_0_HE NULL
#define BEETLE_OPT_LAST_SCANLINE_LABEL_HE NULL
#define BEETLE_OPT_LAST_SCANLINE_INFO_0_HE NULL
#define OPTION_VAL_239_HE NULL
#define BEETLE_OPT_INITIAL_SCANLINE_PAL_LABEL_HE NULL
#define BEETLE_OPT_INITIAL_SCANLINE_PAL_INFO_0_HE NULL
#define BEETLE_OPT_LAST_SCANLINE_PAL_LABEL_HE NULL
#define BEETLE_OPT_LAST_SCANLINE_PAL_INFO_0_HE NULL
#define OPTION_VAL_287_HE NULL

struct retro_core_option_v2_category option_cats_he[] = {
   {
      "system",
      CATEGORY_SYSTEM_LABEL_HE,
      CATEGORY_SYSTEM_INFO_0_HE
   },
   {
      "video",
      CATEGORY_VIDEO_LABEL_HE,
      CATEGORY_VIDEO_INFO_0_HE
   },
   {
      "pgxp",
      CATEGORY_PGXP_LABEL_HE,
      CATEGORY_PGXP_INFO_0_HE
   },
   {
      "osd",
      CATEGORY_OSD_LABEL_HE,
      CATEGORY_OSD_INFO_0_HE
   },
   {
      "input",
      CATEGORY_INPUT_LABEL_HE,
      CATEGORY_INPUT_INFO_0_HE
   },
   {
      "memcards",
      CATEGORY_MEMCARDS_LABEL_HE,
      CATEGORY_MEMCARDS_INFO_0_HE
   },
   {
      "hacks",
      CATEGORY_HACKS_LABEL_HE,
      CATEGORY_HACKS_INFO_0_HE
   },
   { NULL, NULL, NULL },
};
struct retro_core_option_v2_definition option_defs_he[] = {
   {
      BEETLE_OPT(cpu_freq_scale),
      BEETLE_OPT_CPU_FREQ_SCALE_LABEL_HE,
      NULL,
      BEETLE_OPT_CPU_FREQ_SCALE_INFO_0_HE,
      NULL,
      "system",
      {
         { "50%",           OPTION_VAL_50_HE },
         { "60%",           OPTION_VAL_60_HE },
         { "70%",           OPTION_VAL_70_HE },
         { "80%",           OPTION_VAL_80_HE },
         { "90%",           OPTION_VAL_90_HE },
         { "100%",          OPTION_VAL_100_HE },
         { "110%",          OPTION_VAL_110_HE },
         { "120%",          OPTION_VAL_120_HE },
         { "130%",          OPTION_VAL_130_HE },
         { "140%",          OPTION_VAL_140_HE },
         { "150%",          OPTION_VAL_150_HE },
         { "160%",          OPTION_VAL_160_HE },
         { "170%",          OPTION_VAL_170_HE },
         { "180%",          OPTION_VAL_180_HE },
         { "190%",          OPTION_VAL_190_HE },
         { "200%",          OPTION_VAL_200_HE },
         { "210%",          OPTION_VAL_210_HE },
         { "220%",          OPTION_VAL_220_HE },
         { "230%",          OPTION_VAL_230_HE },
         { "240%",          OPTION_VAL_240_HE },
         { "250%",          OPTION_VAL_250_HE },
         { "260%",          OPTION_VAL_260_HE },
         { "270%",          OPTION_VAL_270_HE },
         { "280%",          OPTION_VAL_280_HE },
         { "290%",          OPTION_VAL_290_HE },
         { "300%",          OPTION_VAL_300_HE },
         { "310%",          OPTION_VAL_310_HE },
         { "320%",          OPTION_VAL_320_HE },
         { "330%",          OPTION_VAL_330_HE },
         { "340%",          OPTION_VAL_340_HE },
         { "350%",          OPTION_VAL_350_HE },
         { "360%",          OPTION_VAL_360_HE },
         { "370%",          OPTION_VAL_370_HE },
         { "380%",          OPTION_VAL_380_HE },
         { "390%",          OPTION_VAL_390_HE },
         { "400%",          OPTION_VAL_400_HE },
         { "410%",          OPTION_VAL_410_HE },
         { "420%",          OPTION_VAL_420_HE },
         { "430%",          OPTION_VAL_430_HE },
         { "440%",          OPTION_VAL_440_HE },
         { "450%",          OPTION_VAL_450_HE },
         { "460%",          OPTION_VAL_460_HE },
         { "470%",          OPTION_VAL_470_HE },
         { "480%",          OPTION_VAL_480_HE },
         { "490%",          OPTION_VAL_490_HE },
         { "500%",          OPTION_VAL_500_HE },
         { "510%",          OPTION_VAL_510_HE },
         { "520%",          OPTION_VAL_520_HE },
         { "530%",          OPTION_VAL_530_HE },
         { "540%",          OPTION_VAL_540_HE },
         { "550%",          OPTION_VAL_550_HE },
         { "560%",          OPTION_VAL_560_HE },
         { "570%",          OPTION_VAL_570_HE },
         { "580%",          OPTION_VAL_580_HE },
         { "590%",          OPTION_VAL_590_HE },
         { "600%",          OPTION_VAL_600_HE },
         { "610%",          OPTION_VAL_610_HE },
         { "620%",          OPTION_VAL_620_HE },
         { "630%",          OPTION_VAL_630_HE },
         { "640%",          OPTION_VAL_640_HE },
         { "650%",          OPTION_VAL_650_HE },
         { "660%",          OPTION_VAL_660_HE },
         { "670%",          OPTION_VAL_670_HE },
         { "680%",          OPTION_VAL_680_HE },
         { "690%",          OPTION_VAL_690_HE },
         { "700%",          OPTION_VAL_700_HE },
         { "710%",          OPTION_VAL_710_HE },
         { "720%",          OPTION_VAL_720_HE },
         { "730%",          OPTION_VAL_730_HE },
         { "740%",          OPTION_VAL_740_HE },
         { "750%",          OPTION_VAL_750_HE },
         { NULL, NULL },
      },
      "100%"
   },
   {
      BEETLE_OPT(gpu_overclock),
      BEETLE_OPT_GPU_OVERCLOCK_LABEL_HE,
      NULL,
      BEETLE_OPT_GPU_OVERCLOCK_INFO_0_HE,
      NULL,
      "system",
      {
         { "1x(native)", OPTION_VAL_1X_NATIVE_HE },
         { "2x",         OPTION_VAL_2X_HE },
         { "4x",         OPTION_VAL_4X_HE },
         { "8x",         OPTION_VAL_8X_HE },
         { "16x",        OPTION_VAL_16X_HE },
         { "32x",        OPTION_VAL_32X_HE },
         { NULL, NULL },
      },
      "1x(native)"
   },
   {
      BEETLE_OPT(gte_overclock),
      BEETLE_OPT_GTE_OVERCLOCK_LABEL_HE,
      NULL,
      BEETLE_OPT_GTE_OVERCLOCK_INFO_0_HE,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
#ifndef EMSCRIPTEN
   {
      BEETLE_OPT(cd_access_method),
      BEETLE_OPT_CD_ACCESS_METHOD_LABEL_HE,
      NULL,
      BEETLE_OPT_CD_ACCESS_METHOD_INFO_0_HE,
      NULL,
      "system",
      {
         { "sync",     OPTION_VAL_SYNC_HE },
         { "async",    OPTION_VAL_ASYNC_HE },
         { "precache", OPTION_VAL_PRECACHE_HE },
         { NULL, NULL },
      },
      "sync"
   },
#endif
   {
      BEETLE_OPT(cd_fastload),
      BEETLE_OPT_CD_FASTLOAD_LABEL_HE,
      NULL,
      BEETLE_OPT_CD_FASTLOAD_INFO_0_HE,
      NULL,
      "system",
      {
         { "2x(native)", OPTION_VAL_2X_NATIVE_HE },
         { "4x",          OPTION_VAL_4X_HE },
         { "6x",          OPTION_VAL_6X_HE },
         { "8x",          OPTION_VAL_8X_HE },
         { "10x",         OPTION_VAL_10X_HE },
         { "12x",         OPTION_VAL_12X_HE },
         { "14x",         OPTION_VAL_14X_HE },
         { NULL, NULL },
      },
      "2x(native)"
   },
   {
      BEETLE_OPT(skip_bios),
      BEETLE_OPT_SKIP_BIOS_LABEL_HE,
      NULL,
      BEETLE_OPT_SKIP_BIOS_INFO_0_HE,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(override_bios),
      BEETLE_OPT_OVERRIDE_BIOS_LABEL_HE,
      NULL,
      BEETLE_OPT_OVERRIDE_BIOS_INFO_0_HE,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "psxonpsp", OPTION_VAL_PSXONPSP_HE },
         { "ps1_rom",  OPTION_VAL_PS1_ROM_HE },
         { "openbios", OPTION_VAL_OPENBIOS_HE     },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(region),
      BEETLE_OPT_REGION_LABEL_HE,
      NULL,
      BEETLE_OPT_REGION_INFO_0_HE,
      NULL,
      "system",
      {
         { "auto",   OPTION_VAL_AUTO_HE },
         { "ntsc-j", OPTION_VAL_NTSC_J_HE },
         { "ntsc-u", OPTION_VAL_NTSC_U_HE },
         { "pal",    OPTION_VAL_PAL_HE },
         { NULL, NULL },
      },
      "auto"
   },
   {
      BEETLE_OPT(internal_resolution),
      BEETLE_OPT_INTERNAL_RESOLUTION_LABEL_HE,
      NULL,
      BEETLE_OPT_INTERNAL_RESOLUTION_INFO_0_HE,
      NULL,
      "video",
      {
         { "1x(native)", OPTION_VAL_1X_NATIVE_HE },
         { "2x",         OPTION_VAL_2X_HE },
         { "4x",         OPTION_VAL_4X_HE },
         { "8x",         OPTION_VAL_8X_HE },
         { "16x",        OPTION_VAL_16X_HE },
         { NULL, NULL },
      },
      "1x(native)"
   },
#if defined(HAVE_OPENGL) || defined(HAVE_OPENGLES) || defined(HAVE_VULKAN)
   {
      BEETLE_OPT(renderer),
      BEETLE_OPT_RENDERER_LABEL_HE,
      NULL,
      BEETLE_OPT_RENDERER_INFO_0_HE,
      NULL,
      "video",
      {
         { "hardware",    OPTION_VAL_HARDWARE_HE },
#if defined(HAVE_OPENGL) || defined(HAVE_OPENGLES)
         { "hardware_gl", OPTION_VAL_HARDWARE_GL_HE },
#endif
#if defined(HAVE_VULKAN)
         { "hardware_vk", OPTION_VAL_HARDWARE_VK_HE },
#endif
         { "software",    OPTION_VAL_SOFTWARE_HE },
         { NULL, NULL },
      },
      "hardware"
   },
   {
      BEETLE_OPT(renderer_software_fb),
      BEETLE_OPT_RENDERER_SOFTWARE_FB_LABEL_HE,
      NULL,
      BEETLE_OPT_RENDERER_SOFTWARE_FB_INFO_0_HE,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled"
   },
#endif
#if defined(HAVE_OPENGL) || defined(HAVE_OPENGLES)
   {
      BEETLE_OPT(depth),
      BEETLE_OPT_DEPTH_LABEL_HE,
      NULL,
      BEETLE_OPT_DEPTH_INFO_0_HE,
      NULL,
      "video",
      {
         { "16bpp(native)", OPTION_VAL_16BPP_NATIVE_HE },
         { "32bpp",         OPTION_VAL_32BPP_HE },
         { NULL, NULL },
      },
      "16bpp(native)"
   },
   // Sort of, it's more like 15-bit high color and 24-bit true color for visible output. The alpha channel is used for mask bit. Vulkan renderer uses ABGR1555_555 for 31 bits internal? FMVs are always 24-bit on all renderers like original hardware (BGR888, no alpha)
#endif
   {
      BEETLE_OPT(dither_mode),
      BEETLE_OPT_DITHER_MODE_LABEL_HE,
      NULL,
      BEETLE_OPT_DITHER_MODE_INFO_0_HE,
      NULL,
      "video",
      {
         { "1x(native)",          OPTION_VAL_1X_NATIVE_HE },
         { "internal resolution", OPTION_VAL_INTERNAL_RESOLUTION_HE },
         { "disabled",            NULL },
         { NULL, NULL },
      },
      "1x(native)"
   },
#ifdef HAVE_VULKAN
   {
      BEETLE_OPT(scaled_uv_offset),
      BEETLE_OPT_SCALED_UV_OFFSET_LABEL_HE,
      NULL,
      BEETLE_OPT_SCALED_UV_OFFSET_INFO_0_HE,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled"
   },
#endif
#if defined(HAVE_OPENGL) || defined(HAVE_OPENGLES) || defined(HAVE_VULKAN)
   {
      BEETLE_OPT(filter),
      BEETLE_OPT_FILTER_LABEL_HE,
      NULL,
      BEETLE_OPT_FILTER_INFO_0_HE,
      NULL,
      "video",
      {
         { "nearest",  OPTION_VAL_NEAREST_HE },
         { "SABR",     OPTION_VAL_SABR_HE },
         { "xBR",      OPTION_VAL_XBR_HE },
         { "bilinear", OPTION_VAL_BILINEAR_HE },
         { "3-point",  OPTION_VAL_3_POINT_HE },
         { "JINC2",    OPTION_VAL_JINC2_HE },
         { NULL, NULL },
      },
      "nearest"
   },
#ifdef HAVE_VULKAN
   {
      BEETLE_OPT(filter_exclude_sprite),
      BEETLE_OPT_FILTER_EXCLUDE_SPRITE_LABEL_HE,
      NULL,
      BEETLE_OPT_FILTER_EXCLUDE_SPRITE_INFO_0_HE,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "opaque", OPTION_VAL_OPAQUE_HE },
         { "all", OPTION_VAL_ALL_HE },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(filter_exclude_2d_polygon),
      BEETLE_OPT_FILTER_EXCLUDE_2D_POLYGON_LABEL_HE,
      NULL,
      BEETLE_OPT_FILTER_EXCLUDE_2D_POLYGON_INFO_0_HE,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "opaque", OPTION_VAL_OPAQUE_HE },
         { "all", OPTION_VAL_ALL_HE },
         { NULL, NULL },
      },
      "disabled"
   },
#endif
#endif
#ifdef HAVE_VULKAN
   {
      BEETLE_OPT(adaptive_smoothing),
      BEETLE_OPT_ADAPTIVE_SMOOTHING_LABEL_HE,
      NULL,
      BEETLE_OPT_ADAPTIVE_SMOOTHING_INFO_0_HE,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(super_sampling),
      BEETLE_OPT_SUPER_SAMPLING_LABEL_HE,
      NULL,
      BEETLE_OPT_SUPER_SAMPLING_INFO_0_HE,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(msaa),
      BEETLE_OPT_MSAA_LABEL_HE,
      NULL,
      BEETLE_OPT_MSAA_INFO_0_HE,
      NULL,
      "video",
      {
         { "1x",  OPTION_VAL_1X_HE },
         { "2x",  OPTION_VAL_2X_HE },
         { "4x",  OPTION_VAL_4X_HE },
         { "8x",  OPTION_VAL_8X_HE },
         { "16x", OPTION_VAL_16X_HE },
         { NULL, NULL },
      },
      "1x"
   },
   {
      BEETLE_OPT(mdec_yuv),
      BEETLE_OPT_MDEC_YUV_LABEL_HE,
      NULL,
      BEETLE_OPT_MDEC_YUV_INFO_0_HE,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(track_textures),
      BEETLE_OPT_TRACK_TEXTURES_LABEL_HE,
      NULL,
      BEETLE_OPT_TRACK_TEXTURES_INFO_0_HE,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
#ifdef TEXTURE_DUMPING_ENABLED
   {
      BEETLE_OPT(dump_textures),
      BEETLE_OPT_DUMP_TEXTURES_LABEL_HE,
      NULL,
      BEETLE_OPT_DUMP_TEXTURES_INFO_0_HE,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
#endif
   {
      BEETLE_OPT(replace_textures),
      BEETLE_OPT_REPLACE_TEXTURES_LABEL_HE,
      NULL,
      BEETLE_OPT_REPLACE_TEXTURES_INFO_0_HE,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
#endif
   {
      BEETLE_OPT(pgxp_mode),
      BEETLE_OPT_PGXP_MODE_LABEL_HE,
      NULL,
      BEETLE_OPT_PGXP_MODE_INFO_0_HE,
      NULL,
      "pgxp",
      {
         { "disabled",     NULL },
         { "memory only",  OPTION_VAL_MEMORY_ONLY_HE },
         { "memory + CPU", OPTION_VAL_MEMORY_CPU_HE },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(pgxp_2d_tol),
      BEETLE_OPT_PGXP_2D_TOL_LABEL_HE,
      NULL,
      BEETLE_OPT_PGXP_2D_TOL_INFO_0_HE,
      NULL,
      "pgxp",
      {
         { "disabled", NULL },
         { "0px", OPTION_VAL_0PX_HE },
         { "1px", OPTION_VAL_1PX_HE },
         { "2px", OPTION_VAL_2PX_HE },
         { "3px", OPTION_VAL_3PX_HE },
         { "4px", OPTION_VAL_4PX_HE },
         { "5px", OPTION_VAL_5PX_HE },
         { "6px", OPTION_VAL_6PX_HE },
         { "7px", OPTION_VAL_7PX_HE },
         { "8px", OPTION_VAL_8PX_HE },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(pgxp_nclip),
      BEETLE_OPT_PGXP_NCLIP_LABEL_HE,
      NULL,
      BEETLE_OPT_PGXP_NCLIP_INFO_0_HE,
      NULL,
      "pgxp",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
#if defined(HAVE_OPENGL) || defined(HAVE_OPENGLES) || defined(HAVE_VULKAN)
   {
      BEETLE_OPT(pgxp_vertex),
      BEETLE_OPT_PGXP_VERTEX_LABEL_HE,
      NULL,
      BEETLE_OPT_PGXP_VERTEX_INFO_0_HE,
      NULL,
      "pgxp",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
#endif
   {
      BEETLE_OPT(pgxp_texture),
      BEETLE_OPT_PGXP_TEXTURE_LABEL_HE,
      NULL,
      BEETLE_OPT_PGXP_TEXTURE_INFO_0_HE,
      NULL,
      "pgxp",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(display_internal_fps),
      BEETLE_OPT_DISPLAY_INTERNAL_FPS_LABEL_HE,
      NULL,
      BEETLE_OPT_DISPLAY_INTERNAL_FPS_INFO_0_HE,
      NULL,
      "osd",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(display_osd),
      BEETLE_OPT_DISPLAY_OSD_LABEL_HE,
      NULL,
      BEETLE_OPT_DISPLAY_OSD_INFO_0_HE,
      NULL,
      "osd",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled"
   },
#if defined(HAVE_OPENGL) || defined(HAVE_OPENGLES) || defined(HAVE_VULKAN)
   {
      BEETLE_OPT(display_vram),
      BEETLE_OPT_DISPLAY_VRAM_LABEL_HE,
      NULL,
      BEETLE_OPT_DISPLAY_VRAM_INFO_0_HE,
      NULL,
      "osd",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
#endif
   {
      BEETLE_OPT(analog_calibration),
      BEETLE_OPT_ANALOG_CALIBRATION_LABEL_HE,
      NULL,
      BEETLE_OPT_ANALOG_CALIBRATION_INFO_0_HE,
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(analog_toggle),
      BEETLE_OPT_ANALOG_TOGGLE_LABEL_HE,
      NULL,
      BEETLE_OPT_ANALOG_TOGGLE_INFO_0_HE,
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { "enabled-analog", OPTION_VAL_ENABLED_ANALOG_HE },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(analog_toggle_combo),
      BEETLE_OPT_ANALOG_TOGGLE_COMBO_LABEL_HE,
      NULL,
      BEETLE_OPT_ANALOG_TOGGLE_COMBO_INFO_0_HE,
      NULL,
      "input",
      {
         { "l1+l2+r1+r2+start+select", OPTION_VAL_L1_L2_R1_R2_START_SELECT_HE },
         { "l1+r1+select",             OPTION_VAL_L1_R1_SELECT_HE },
         { "l1+r1+start",              OPTION_VAL_L1_R1_START_HE },
         { "l1+r1+l3",                 OPTION_VAL_L1_R1_L3_HE },
         { "l1+r1+r3",                 OPTION_VAL_L1_R1_R3_HE },
         { "l2+r2+select",             OPTION_VAL_L2_R2_SELECT_HE },
         { "l2+r2+start",              OPTION_VAL_L2_R2_START_HE },
         { "l2+r2+l3",                 OPTION_VAL_L2_R2_L3_HE },
         { "l2+r2+r3",                 OPTION_VAL_L2_R2_R3_HE },
         { "l3+r3",                    OPTION_VAL_L3_R3_HE },
         { NULL, NULL },
      },
      "l1+r1+select"
   },
   {
      BEETLE_OPT(analog_toggle_hold),
      BEETLE_OPT_ANALOG_TOGGLE_HOLD_LABEL_HE,
      NULL,
      BEETLE_OPT_ANALOG_TOGGLE_HOLD_INFO_0_HE,
      NULL,
      "input",
      {
         { "0", OPTION_VAL_0_HE },
         { "1", OPTION_VAL_1_HE },
         { "2", OPTION_VAL_2_HE },
         { "3", OPTION_VAL_3_HE },
         { "4", OPTION_VAL_4_HE },
         { "5", OPTION_VAL_5_HE },
         { NULL, NULL },
      },
      "1"
   },
   {
      BEETLE_OPT(enable_multitap_port1),
      BEETLE_OPT_ENABLE_MULTITAP_PORT1_LABEL_HE,
      NULL,
      BEETLE_OPT_ENABLE_MULTITAP_PORT1_INFO_0_HE,
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(enable_multitap_port2),
      BEETLE_OPT_ENABLE_MULTITAP_PORT2_LABEL_HE,
      NULL,
      BEETLE_OPT_ENABLE_MULTITAP_PORT2_INFO_0_HE,
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(gun_input_mode),
      BEETLE_OPT_GUN_INPUT_MODE_LABEL_HE,
      NULL,
      BEETLE_OPT_GUN_INPUT_MODE_INFO_0_HE,
      NULL,
      "input",
      {
         { "lightgun",    OPTION_VAL_LIGHTGUN_HE },
         { "touchscreen", OPTION_VAL_TOUCHSCREEN_HE },
         { NULL, NULL },
      },
      "lightgun"
   },
   // Shouldn't the gun_input_mode just be Mouse vs. Touchscreen?
   {
      BEETLE_OPT(gun_cursor),
      BEETLE_OPT_GUN_CURSOR_LABEL_HE,
      NULL,
      BEETLE_OPT_GUN_CURSOR_INFO_0_HE,
      NULL,
      "input",
      {
         { "off",   "disabled" },
         { "cross", OPTION_VAL_CROSS_HE },
         { "dot",   OPTION_VAL_DOT_HE },
         { NULL, NULL },
      },
      "cross"
   },
   {
      BEETLE_OPT(crosshair_color_p1),
      BEETLE_OPT_CROSSHAIR_COLOR_P1_LABEL_HE,
      NULL,
      BEETLE_OPT_CROSSHAIR_COLOR_P1_INFO_0_HE,
      NULL,
      "input",
      {
         { "red",    OPTION_VAL_RED_HE },
         { "blue",   OPTION_VAL_BLUE_HE },
         { "green",  OPTION_VAL_GREEN_HE },
         { "orange", OPTION_VAL_ORANGE_HE },
         { "yellow", OPTION_VAL_YELLOW_HE },
         { "cyan",   OPTION_VAL_CYAN_HE },
         { "pink",   OPTION_VAL_PINK_HE },
         { "purple", OPTION_VAL_PURPLE_HE },
         { "black",  OPTION_VAL_BLACK_HE },
         { "white",  OPTION_VAL_WHITE_HE },
         { NULL, NULL },
      },
      "red"
   },
   {
      BEETLE_OPT(crosshair_color_p2),
      BEETLE_OPT_CROSSHAIR_COLOR_P2_LABEL_HE,
      NULL,
      BEETLE_OPT_CROSSHAIR_COLOR_P2_INFO_0_HE,
      NULL,
      "input",
      {
         { "blue",   OPTION_VAL_BLUE_HE },
         { "red",    OPTION_VAL_RED_HE },
         { "green",  OPTION_VAL_GREEN_HE },
         { "orange", OPTION_VAL_ORANGE_HE },
         { "yellow", OPTION_VAL_YELLOW_HE },
         { "cyan",   OPTION_VAL_CYAN_HE },
         { "pink",   OPTION_VAL_PINK_HE },
         { "purple", OPTION_VAL_PURPLE_HE },
         { "black",  OPTION_VAL_BLACK_HE },
         { "white",  OPTION_VAL_WHITE_HE },
         { NULL, NULL },
      },
      "blue"
   },
   {
      BEETLE_OPT(mouse_sensitivity),
      BEETLE_OPT_MOUSE_SENSITIVITY_LABEL_HE,
      NULL,
      BEETLE_OPT_MOUSE_SENSITIVITY_INFO_0_HE,
      NULL,
      "input",
      {
         { "5%",   OPTION_VAL_5_O42_HE },
         { "10%",  OPTION_VAL_10_HE },
         { "15%",  OPTION_VAL_15_HE },
         { "20%",  OPTION_VAL_20_HE },
         { "25%",  OPTION_VAL_25_HE },
         { "30%",  OPTION_VAL_30_HE },
         { "35%",  OPTION_VAL_35_HE },
         { "40%",  OPTION_VAL_40_HE },
         { "45%",  OPTION_VAL_45_HE },
         { "50%",  OPTION_VAL_50_HE },
         { "55%",  OPTION_VAL_55_HE },
         { "60%",  OPTION_VAL_60_HE },
         { "65%",  OPTION_VAL_65_HE },
         { "70%",  OPTION_VAL_70_HE },
         { "75%",  OPTION_VAL_75_HE },
         { "80%",  OPTION_VAL_80_HE },
         { "85%",  OPTION_VAL_85_HE },
         { "90%",  OPTION_VAL_90_HE },
         { "95%",  OPTION_VAL_95_HE },
         { "100%", OPTION_VAL_100_O42_HE },
         { "105%", OPTION_VAL_105_HE },
         { "110%", OPTION_VAL_110_HE },
         { "115%", OPTION_VAL_115_HE },
         { "120%", OPTION_VAL_120_HE },
         { "125%", OPTION_VAL_125_HE },
         { "130%", OPTION_VAL_130_HE },
         { "135%", OPTION_VAL_135_HE },
         { "140%", OPTION_VAL_140_HE },
         { "145%", OPTION_VAL_145_HE },
         { "150%", OPTION_VAL_150_HE },
         { "155%", OPTION_VAL_155_HE },
         { "160%", OPTION_VAL_160_HE },
         { "165%", OPTION_VAL_165_HE },
         { "170%", OPTION_VAL_170_HE },
         { "175%", OPTION_VAL_175_HE },
         { "180%", OPTION_VAL_180_HE },
         { "185%", OPTION_VAL_185_HE },
         { "190%", OPTION_VAL_190_HE },
         { "195%", OPTION_VAL_195_HE },
         { "200%", OPTION_VAL_200_HE },
         { NULL, NULL },
      },
      "100%"
   },
   {
      BEETLE_OPT(negcon_response),
      BEETLE_OPT_NEGCON_RESPONSE_LABEL_HE,
      NULL,
      BEETLE_OPT_NEGCON_RESPONSE_INFO_0_HE,
      NULL,
      "input",
      {
         { "linear",    OPTION_VAL_LINEAR_HE },
         { "quadratic", OPTION_VAL_QUADRATIC_HE },
         { "cubic",     OPTION_VAL_CUBIC_HE },
         { NULL, NULL },
      },
      "linear"
   },
   {
      BEETLE_OPT(negcon_deadzone),
      BEETLE_OPT_NEGCON_DEADZONE_LABEL_HE,
      NULL,
      BEETLE_OPT_NEGCON_DEADZONE_INFO_0_HE,
      NULL,
      "input",
      {
         { "0%",  OPTION_VAL_0_O44_HE },
         { "5%",  OPTION_VAL_5_O42_HE },
         { "10%", OPTION_VAL_10_HE },
         { "15%", OPTION_VAL_15_HE },
         { "20%", OPTION_VAL_20_HE },
         { "25%", OPTION_VAL_25_HE },
         { "30%", OPTION_VAL_30_HE },
         { NULL, NULL },
      },
      "0%"
   },
   {
      BEETLE_OPT(use_mednafen_memcard0_method),
      BEETLE_OPT_USE_MEDNAFEN_MEMCARD0_METHOD_LABEL_HE,
      NULL,
      BEETLE_OPT_USE_MEDNAFEN_MEMCARD0_METHOD_INFO_0_HE,
      NULL,
      "memcards",
      {
         { "libretro", OPTION_VAL_LIBRETRO_HE },
         { "mednafen", OPTION_VAL_MEDNAFEN_HE },
         { NULL, NULL },
      },
      "libretro"
   },
   {
      BEETLE_OPT(enable_memcard1),
      BEETLE_OPT_ENABLE_MEMCARD1_LABEL_HE,
      NULL,
      BEETLE_OPT_ENABLE_MEMCARD1_INFO_0_HE,
      NULL,
      "memcards",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(shared_memory_cards),
      BEETLE_OPT_SHARED_MEMORY_CARDS_LABEL_HE,
      NULL,
      BEETLE_OPT_SHARED_MEMORY_CARDS_INFO_0_HE,
      NULL,
      "memcards",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(memcard_left_index),
      BEETLE_OPT_MEMCARD_LEFT_INDEX_LABEL_HE,
      NULL,
      BEETLE_OPT_MEMCARD_LEFT_INDEX_INFO_0_HE,
      NULL,
      "memcards",
      {
         { "0",  OPTION_VAL_0_O48_HE },
         { "1",  NULL },
         { "2",  NULL },
         { "3",  NULL },
         { "4",  NULL },
         { "5",  NULL },
         { "6",  NULL },
         { "7",  NULL },
         { "8",  NULL },
         { "9",  NULL },
         { "10", NULL },
         { "11", NULL },
         { "12", NULL },
         { "13", NULL },
         { "14", NULL },
         { "15", NULL },
         { "16", NULL },
         { "17", NULL },
         { "18", NULL },
         { "19", NULL },
         { "20", NULL },
         { "21", NULL },
         { "22", NULL },
         { "23", NULL },
         { "24", NULL },
         { "25", NULL },
         { "26", NULL },
         { "27", NULL },
         { "28", NULL },
         { "29", NULL },
         { "30", NULL },
         { "31", NULL },
         { "32", NULL },
         { "33", NULL },
         { "34", NULL },
         { "35", NULL },
         { "36", NULL },
         { "37", NULL },
         { "38", NULL },
         { "39", NULL },
         { "40", NULL },
         { "41", NULL },
         { "42", NULL },
         { "43", NULL },
         { "44", NULL },
         { "45", NULL },
         { "46", NULL },
         { "47", NULL },
         { "48", NULL },
         { "49", NULL },
         { "50", NULL },
         { "51", NULL },
         { "52", NULL },
         { "53", NULL },
         { "54", NULL },
         { "55", NULL },
         { "56", NULL },
         { "57", NULL },
         { "58", NULL },
         { "59", NULL },
         { "60", NULL },
         { "61", NULL },
         { "62", NULL },
         { "63", NULL },
         { NULL, NULL },
      },
      "0"
   },
   {
      BEETLE_OPT(memcard_right_index),
      BEETLE_OPT_MEMCARD_RIGHT_INDEX_LABEL_HE,
      NULL,
      BEETLE_OPT_MEMCARD_RIGHT_INDEX_INFO_0_HE,
      NULL,
      "memcards",
      {
         { "0",  NULL },
         { "1",  OPTION_VAL_1_O49_HE },
         { "2",  NULL },
         { "3",  NULL },
         { "4",  NULL },
         { "5",  NULL },
         { "6",  NULL },
         { "7",  NULL },
         { "8",  NULL },
         { "9",  NULL },
         { "10", NULL },
         { "11", NULL },
         { "12", NULL },
         { "13", NULL },
         { "14", NULL },
         { "15", NULL },
         { "16", NULL },
         { "17", NULL },
         { "18", NULL },
         { "19", NULL },
         { "20", NULL },
         { "21", NULL },
         { "22", NULL },
         { "23", NULL },
         { "24", NULL },
         { "25", NULL },
         { "26", NULL },
         { "27", NULL },
         { "28", NULL },
         { "29", NULL },
         { "30", NULL },
         { "31", NULL },
         { "32", NULL },
         { "33", NULL },
         { "34", NULL },
         { "35", NULL },
         { "36", NULL },
         { "37", NULL },
         { "38", NULL },
         { "39", NULL },
         { "40", NULL },
         { "41", NULL },
         { "42", NULL },
         { "43", NULL },
         { "44", NULL },
         { "45", NULL },
         { "46", NULL },
         { "47", NULL },
         { "48", NULL },
         { "49", NULL },
         { "50", NULL },
         { "51", NULL },
         { "52", NULL },
         { "53", NULL },
         { "54", NULL },
         { "55", NULL },
         { "56", NULL },
         { "57", NULL },
         { "58", NULL },
         { "59", NULL },
         { "60", NULL },
         { "61", NULL },
         { "62", NULL },
         { "63", NULL },
         { NULL, NULL },
      },
      "1"
   },
   {
      BEETLE_OPT(line_render),
      BEETLE_OPT_LINE_RENDER_LABEL_HE,
      NULL,
      BEETLE_OPT_LINE_RENDER_INFO_0_HE,
      NULL,
      "hacks",
      {
         { "default",    OPTION_VAL_DEFAULT_HE },
         { "aggressive", OPTION_VAL_AGGRESSIVE_HE },
         { "disabled",   NULL },
         { NULL, NULL },
      },
      "default"
   },
   {
      BEETLE_OPT(widescreen_hack),
      BEETLE_OPT_WIDESCREEN_HACK_LABEL_HE,
      NULL,
      BEETLE_OPT_WIDESCREEN_HACK_INFO_0_HE,
      NULL,
      "hacks",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(widescreen_hack_aspect_ratio),
      BEETLE_OPT_WIDESCREEN_HACK_ASPECT_RATIO_LABEL_HE,
      NULL,
      BEETLE_OPT_WIDESCREEN_HACK_ASPECT_RATIO_INFO_0_HE,
      NULL,
      "hacks",
      {
         { "16:9",  OPTION_VAL_16_9_HE },
         { "16:10", OPTION_VAL_16_10_HE },
         { "18:9",  OPTION_VAL_18_9_HE },
         { "19:9",  OPTION_VAL_19_9_HE },
         { "20:9",  OPTION_VAL_20_9_HE },
         { "21:9",  OPTION_VAL_21_9_HE }, // 64:27
         { "32:9",  OPTION_VAL_32_9_HE },
         { NULL,    NULL },
      },
      "16:9"
   },
#if defined(HAVE_LIGHTREC)
   {
      BEETLE_OPT(cpu_dynarec),
      BEETLE_OPT_CPU_DYNAREC_LABEL_HE,
      NULL,
      BEETLE_OPT_CPU_DYNAREC_INFO_0_HE,
      NULL,
      "hacks",
      {
         { "disabled",        OPTION_VAL_DISABLED_HE },
         { "execute",         OPTION_VAL_EXECUTE_HE },
         { "execute_one",     OPTION_VAL_EXECUTE_ONE_HE },
         { "run_interpreter", OPTION_VAL_RUN_INTERPRETER_HE },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(dynarec_invalidate),
      BEETLE_OPT_DYNAREC_INVALIDATE_LABEL_HE,
      NULL,
      BEETLE_OPT_DYNAREC_INVALIDATE_INFO_0_HE,
      NULL,
      "hacks",
      {
         { "full", OPTION_VAL_FULL_HE },
         { "dma",  OPTION_VAL_DMA_HE },
         { NULL, NULL },
      },
      "full"
   },
   {
      BEETLE_OPT(dynarec_eventcycles),
      BEETLE_OPT_DYNAREC_EVENTCYCLES_LABEL_HE,
      NULL,
      BEETLE_OPT_DYNAREC_EVENTCYCLES_INFO_0_HE,
      NULL,
      "hacks",
      {
         { "128",  OPTION_VAL_128_HE },
         { "256",  NULL },
         { "384",  NULL },
         { "512",  NULL },
         { "640",  NULL },
         { "768",  NULL },
         { "896",  NULL },
         { "1024", NULL },
         { "1152", NULL },
         { "1280", NULL },
         { "1408", NULL },
         { "1536", NULL },
         { "1664", NULL },
         { "1792", NULL },
         { "1920", NULL },
         { "2048", NULL },
         { NULL, NULL },
      },
      "128"
   },
   {
      BEETLE_OPT(dynarec_spu_samples),
      BEETLE_OPT_DYNAREC_SPU_SAMPLES_LABEL_HE,
      NULL,
      BEETLE_OPT_DYNAREC_SPU_SAMPLES_INFO_0_HE,
      NULL,
      "hacks",
      {
         { "1",  OPTION_VAL_1_O49_HE },
         { "4",  NULL },
         { "16", NULL },
         { NULL, NULL },
      },
      "1"
   },
#endif
   {
      BEETLE_OPT(core_timing_fps),
      BEETLE_OPT_CORE_TIMING_FPS_LABEL_HE,
      NULL,
      BEETLE_OPT_CORE_TIMING_FPS_INFO_0_HE,
      NULL,
      "video",
      {
         { "auto_toggle",       OPTION_VAL_AUTO_TOGGLE_HE },
         { "force_progressive", OPTION_VAL_FORCE_PROGRESSIVE_HE },
         { "force_interlaced",  OPTION_VAL_FORCE_INTERLACED_HE },
      },
      "auto_toggle"
   },
   {
      BEETLE_OPT(deinterlacer),
      BEETLE_OPT_DEINTERLACER_LABEL_HE,
      NULL,
      BEETLE_OPT_DEINTERLACER_INFO_0_HE,
      NULL,
      "video",
      {
         { "weave",      OPTION_VAL_WEAVE_HE },
         { "bob",        OPTION_VAL_BOB_HE },
         { "bob_offset", OPTION_VAL_BOB_OFFSET_HE },
         { "fastmad",    OPTION_VAL_FASTMAD_HE },
         { "off",        "Off" },
         { NULL, NULL },
      },
      "weave"
   },
   {
      BEETLE_OPT(frame_duping),
      BEETLE_OPT_FRAME_DUPING_LABEL_HE,
      NULL,
      BEETLE_OPT_FRAME_DUPING_INFO_0_HE,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(pal_video_timing_override),
      BEETLE_OPT_PAL_VIDEO_TIMING_OVERRIDE_LABEL_HE,
      NULL,
      BEETLE_OPT_PAL_VIDEO_TIMING_OVERRIDE_INFO_0_HE,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(aspect_ratio),
      BEETLE_OPT_ASPECT_RATIO_LABEL_HE,
      NULL,
      BEETLE_OPT_ASPECT_RATIO_INFO_0_HE,
      NULL,
      "video",
      {
         { "corrected", OPTION_VAL_CORRECTED_HE },
         { "uncorrected", OPTION_VAL_UNCORRECTED_HE },
         { "4:3",  OPTION_VAL_4_3_HE },
         { "ntsc", OPTION_VAL_NTSC_HE },
      },
      "corrected"
   },
   {
      BEETLE_OPT(crop_overscan),
      BEETLE_OPT_CROP_OVERSCAN_LABEL_HE,
      NULL,
      BEETLE_OPT_CROP_OVERSCAN_INFO_0_HE,
      NULL,
      "video",
      {
         { "disabled", "Disabled" },
         { "static",   OPTION_VAL_STATIC_HE },
         { "smart",    OPTION_VAL_SMART_HE },
         { NULL, NULL },
      },
      "smart"
   },
   {
      BEETLE_OPT(image_crop),
      BEETLE_OPT_IMAGE_CROP_LABEL_HE,
      NULL,
      BEETLE_OPT_IMAGE_CROP_INFO_0_HE,
      NULL,
      "video",
      {
         { "disabled", "0" },
         { "1px",      OPTION_VAL_1PX_HE },
         { "2px",      OPTION_VAL_2PX_HE },
         { "3px",      OPTION_VAL_3PX_HE },
         { "4px",      OPTION_VAL_4PX_HE },
         { "5px",      OPTION_VAL_5PX_HE },
         { "6px",      OPTION_VAL_6PX_HE },
         { "7px",      OPTION_VAL_7PX_HE },
         { "8px",      OPTION_VAL_8PX_HE },
         { "9px",      OPTION_VAL_9PX_HE },
         { "10px",     OPTION_VAL_10PX_HE },
         { "11px",     OPTION_VAL_11PX_HE },
         { "12px",     OPTION_VAL_12PX_HE },
         { "13px",     OPTION_VAL_13PX_HE },
         { "14px",     OPTION_VAL_14PX_HE },
         { "15px",     OPTION_VAL_15PX_HE },
         { "16px",     OPTION_VAL_16PX_HE },
         { "17px",     OPTION_VAL_17PX_HE },
         { "18px",     OPTION_VAL_18PX_HE },
         { "19px",     OPTION_VAL_19PX_HE },
         { "20px",     OPTION_VAL_20PX_HE },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(image_offset),
      BEETLE_OPT_IMAGE_OFFSET_LABEL_HE,
      NULL,
      BEETLE_OPT_IMAGE_OFFSET_INFO_0_HE,
      NULL,
      "video",
      {
         { "-12px",    OPTION_VAL_12PX_O64_HE },
         { "-11px",    OPTION_VAL_11PX_O64_HE },
         { "-10px",    OPTION_VAL_10PX_O64_HE },
         { "-9px",     OPTION_VAL_9PX_O64_HE },
         { "-8px",     OPTION_VAL_8PX_O64_HE },
         { "-7px",     OPTION_VAL_7PX_O64_HE },
         { "-6px",     OPTION_VAL_6PX_O64_HE },
         { "-5px",     OPTION_VAL_5PX_O64_HE },
         { "-4px",     OPTION_VAL_4PX_O64_HE },
         { "-3px",     OPTION_VAL_3PX_O64_HE },
         { "-2px",     OPTION_VAL_2PX_O64_HE },
         { "-1px",     OPTION_VAL_1PX_O64_HE },
         { "disabled", "0" },
         { "+1px",     OPTION_VAL_1PX_O64_0_HE },
         { "+2px",     OPTION_VAL_2PX_O64_0_HE },
         { "+3px",     OPTION_VAL_3PX_O64_0_HE },
         { "+4px",     OPTION_VAL_4PX_O64_0_HE },
         { "+5px",     OPTION_VAL_5PX_O64_0_HE },
         { "+6px",     OPTION_VAL_6PX_O64_0_HE },
         { "+7px",     OPTION_VAL_7PX_O64_0_HE },
         { "+8px",     OPTION_VAL_8PX_O64_0_HE },
         { "+9px",     OPTION_VAL_9PX_O64_0_HE },
         { "+10px",    OPTION_VAL_10PX_O64_0_HE },
         { "+11px",    OPTION_VAL_11PX_O64_0_HE },
         { "+12px",    OPTION_VAL_12PX_O64_0_HE },
         { NULL, NULL },
      },
      "disabled"
   },
#if defined(HAVE_OPENGL) || defined(HAVE_OPENGLES) || defined(HAVE_VULKAN)
   {
      BEETLE_OPT(image_offset_cycles),
      BEETLE_OPT_IMAGE_OFFSET_CYCLES_LABEL_HE,
      NULL,
      BEETLE_OPT_IMAGE_OFFSET_CYCLES_INFO_0_HE,
      NULL,
      "video",
      {
         { "-40",      NULL },
         { "-39",      NULL },
         { "-38",      NULL },
         { "-37",      NULL },
         { "-36",      NULL },
         { "-35",      NULL },
         { "-34",      NULL },
         { "-33",      NULL },
         { "-32",      NULL },
         { "-31",      NULL },
         { "-30",      NULL },
         { "-29",      NULL },
         { "-28",      NULL },
         { "-27",      NULL },
         { "-26",      NULL },
         { "-25",      NULL },
         { "-24",      NULL },
         { "-23",      NULL },
         { "-22",      NULL },
         { "-21",      NULL },
         { "-20",      NULL },
         { "-19",      NULL },
         { "-18",      NULL },
         { "-17",      NULL },
         { "-16",      NULL },
         { "-15",      NULL },
         { "-14",      NULL },
         { "-13",      NULL },
         { "-12",      NULL },
         { "-11",      NULL },
         { "-10",      NULL },
         { "-9",       NULL },
         { "-8",       NULL },
         { "-7",       NULL },
         { "-6",       NULL },
         { "-5",       NULL },
         { "-4",       NULL },
         { "-3",       NULL },
         { "-2",       NULL },
         { "-1",       NULL },
         { "0",        "0" },
         { "+1",       NULL },
         { "+2",       NULL },
         { "+3",       NULL },
         { "+4",       NULL },
         { "+5",       NULL },
         { "+6",       NULL },
         { "+7",       NULL },
         { "+8",       NULL },
         { "+9",       NULL },
         { "+10",      NULL },
         { "+11",      NULL },
         { "+12",      NULL },
         { "+13",      NULL },
         { "+14",      NULL },
         { "+15",      NULL },
         { "+16",      NULL },
         { "+17",      NULL },
         { "+18",      NULL },
         { "+19",      NULL },
         { "+20",      NULL },
         { "+21",      NULL },
         { "+22",      NULL },
         { "+23",      NULL },
         { "+24",      NULL },
         { "+25",      NULL },
         { "+26",      NULL },
         { "+27",      NULL },
         { "+28",      NULL },
         { "+29",      NULL },
         { "+30",      NULL },
         { "+31",      NULL },
         { "+32",      NULL },
         { "+33",      NULL },
         { "+34",      NULL },
         { "+35",      NULL },
         { "+36",      NULL },
         { "+37",      NULL },
         { "+38",      NULL },
         { "+39",      NULL },
         { "+40",      NULL },
         { NULL, NULL},
      },
      "0"
   },
#endif
   {
      BEETLE_OPT(initial_scanline),
      BEETLE_OPT_INITIAL_SCANLINE_LABEL_HE,
      NULL,
      BEETLE_OPT_INITIAL_SCANLINE_INFO_0_HE,
      NULL,
      "video",
      {
         { "0",  OPTION_VAL_0_O48_HE },
         { "1",  NULL },
         { "2",  NULL },
         { "3",  NULL },
         { "4",  NULL },
         { "5",  NULL },
         { "6",  NULL },
         { "7",  NULL },
         { "8",  NULL },
         { "9",  NULL },
         { "10", NULL },
         { "11", NULL },
         { "12", NULL },
         { "13", NULL },
         { "14", NULL },
         { "15", NULL },
         { "16", NULL },
         { "17", NULL },
         { "18", NULL },
         { "19", NULL },
         { "20", NULL },
         { "21", NULL },
         { "22", NULL },
         { "23", NULL },
         { "24", NULL },
         { "25", NULL },
         { "26", NULL },
         { "27", NULL },
         { "28", NULL },
         { "29", NULL },
         { "30", NULL },
         { "31", NULL },
         { "32", NULL },
         { "33", NULL },
         { "34", NULL },
         { "35", NULL },
         { "36", NULL },
         { "37", NULL },
         { "38", NULL },
         { "39", NULL },
         { "40", NULL },
         { NULL, NULL },
      },
      "0"
   },
   {
      BEETLE_OPT(last_scanline),
      BEETLE_OPT_LAST_SCANLINE_LABEL_HE,
      NULL,
      BEETLE_OPT_LAST_SCANLINE_INFO_0_HE,
      NULL,
      "video",
      {
         { "210", NULL },
         { "211", NULL },
         { "212", NULL },
         { "213", NULL },
         { "214", NULL },
         { "215", NULL },
         { "216", NULL },
         { "217", NULL },
         { "218", NULL },
         { "219", NULL },
         { "220", NULL },
         { "221", NULL },
         { "222", NULL },
         { "223", NULL },
         { "224", NULL },
         { "225", NULL },
         { "226", NULL },
         { "227", NULL },
         { "228", NULL },
         { "229", NULL },
         { "230", NULL },
         { "231", NULL },
         { "232", NULL },
         { "233", NULL },
         { "234", NULL },
         { "235", NULL },
         { "236", NULL },
         { "237", NULL },
         { "238", NULL },
         { "239", OPTION_VAL_239_HE },
         { NULL, NULL },
      },
      "239"
   },
   {
      BEETLE_OPT(initial_scanline_pal),
      BEETLE_OPT_INITIAL_SCANLINE_PAL_LABEL_HE,
      NULL,
      BEETLE_OPT_INITIAL_SCANLINE_PAL_INFO_0_HE,
      NULL,
      "video",
      {
         { "0",  OPTION_VAL_0_O48_HE },
         { "1",  NULL },
         { "2",  NULL },
         { "3",  NULL },
         { "4",  NULL },
         { "5",  NULL },
         { "6",  NULL },
         { "7",  NULL },
         { "8",  NULL },
         { "9",  NULL },
         { "10", NULL },
         { "11", NULL },
         { "12", NULL },
         { "13", NULL },
         { "14", NULL },
         { "15", NULL },
         { "16", NULL },
         { "17", NULL },
         { "18", NULL },
         { "19", NULL },
         { "20", NULL },
         { "21", NULL },
         { "22", NULL },
         { "23", NULL },
         { "24", NULL },
         { "25", NULL },
         { "26", NULL },
         { "27", NULL },
         { "28", NULL },
         { "29", NULL },
         { "30", NULL },
         { "31", NULL },
         { "32", NULL },
         { "33", NULL },
         { "34", NULL },
         { "35", NULL },
         { "36", NULL },
         { "37", NULL },
         { "38", NULL },
         { "39", NULL },
         { "40", NULL },
         { NULL, NULL },
      },
      "0"
   },
   {
      BEETLE_OPT(last_scanline_pal),
      BEETLE_OPT_LAST_SCANLINE_PAL_LABEL_HE,
      NULL,
      BEETLE_OPT_LAST_SCANLINE_PAL_INFO_0_HE,
      NULL,
      "video",
      {
         { "230", NULL },
         { "231", NULL },
         { "232", NULL },
         { "233", NULL },
         { "234", NULL },
         { "235", NULL },
         { "236", NULL },
         { "237", NULL },
         { "238", NULL },
         { "239", NULL },
         { "240", NULL },
         { "241", NULL },
         { "242", NULL },
         { "243", NULL },
         { "244", NULL },
         { "245", NULL },
         { "246", NULL },
         { "247", NULL },
         { "248", NULL },
         { "249", NULL },
         { "250", NULL },
         { "251", NULL },
         { "252", NULL },
         { "253", NULL },
         { "254", NULL },
         { "255", NULL },
         { "256", NULL },
         { "257", NULL },
         { "258", NULL },
         { "259", NULL },
         { "260", NULL },
         { "261", NULL },
         { "262", NULL },
         { "263", NULL },
         { "264", NULL },
         { "265", NULL },
         { "266", NULL },
         { "267", NULL },
         { "268", NULL },
         { "269", NULL },
         { "270", NULL },
         { "271", NULL },
         { "272", NULL },
         { "273", NULL },
         { "274", NULL },
         { "275", NULL },
         { "276", NULL },
         { "277", NULL },
         { "278", NULL },
         { "279", NULL },
         { "280", NULL },
         { "281", NULL },
         { "282", NULL },
         { "283", NULL },
         { "284", NULL },
         { "285", NULL },
         { "286", NULL },
         { "287", OPTION_VAL_287_HE },
         { NULL, NULL },
      },
      "287"
   },
   { NULL, NULL, NULL, NULL, NULL, NULL, {{0}}, NULL },
};
struct retro_core_options_v2 options_he = {
   option_cats_he,
   option_defs_he
};

/* RETRO_LANGUAGE_HR */

#define CATEGORY_SYSTEM_LABEL_HR NULL
#define CATEGORY_SYSTEM_INFO_0_HR NULL
#define CATEGORY_VIDEO_LABEL_HR NULL
#define CATEGORY_VIDEO_INFO_0_HR NULL
#define CATEGORY_PGXP_LABEL_HR NULL
#define CATEGORY_PGXP_INFO_0_HR NULL
#define CATEGORY_OSD_LABEL_HR NULL
#define CATEGORY_OSD_INFO_0_HR NULL
#define CATEGORY_INPUT_LABEL_HR NULL
#define CATEGORY_INPUT_INFO_0_HR NULL
#define CATEGORY_MEMCARDS_LABEL_HR NULL
#define CATEGORY_MEMCARDS_INFO_0_HR NULL
#define CATEGORY_HACKS_LABEL_HR NULL
#define CATEGORY_HACKS_INFO_0_HR NULL
#define BEETLE_OPT_CPU_FREQ_SCALE_LABEL_HR NULL
#define BEETLE_OPT_CPU_FREQ_SCALE_INFO_0_HR NULL
#define OPTION_VAL_50_HR NULL
#define OPTION_VAL_60_HR NULL
#define OPTION_VAL_70_HR NULL
#define OPTION_VAL_80_HR NULL
#define OPTION_VAL_90_HR NULL
#define OPTION_VAL_100_HR NULL
#define OPTION_VAL_110_HR NULL
#define OPTION_VAL_120_HR NULL
#define OPTION_VAL_130_HR NULL
#define OPTION_VAL_140_HR NULL
#define OPTION_VAL_150_HR NULL
#define OPTION_VAL_160_HR NULL
#define OPTION_VAL_170_HR NULL
#define OPTION_VAL_180_HR NULL
#define OPTION_VAL_190_HR NULL
#define OPTION_VAL_200_HR NULL
#define OPTION_VAL_210_HR NULL
#define OPTION_VAL_220_HR NULL
#define OPTION_VAL_230_HR NULL
#define OPTION_VAL_240_HR NULL
#define OPTION_VAL_250_HR NULL
#define OPTION_VAL_260_HR NULL
#define OPTION_VAL_270_HR NULL
#define OPTION_VAL_280_HR NULL
#define OPTION_VAL_290_HR NULL
#define OPTION_VAL_300_HR NULL
#define OPTION_VAL_310_HR NULL
#define OPTION_VAL_320_HR NULL
#define OPTION_VAL_330_HR NULL
#define OPTION_VAL_340_HR NULL
#define OPTION_VAL_350_HR NULL
#define OPTION_VAL_360_HR NULL
#define OPTION_VAL_370_HR NULL
#define OPTION_VAL_380_HR NULL
#define OPTION_VAL_390_HR NULL
#define OPTION_VAL_400_HR NULL
#define OPTION_VAL_410_HR NULL
#define OPTION_VAL_420_HR NULL
#define OPTION_VAL_430_HR NULL
#define OPTION_VAL_440_HR NULL
#define OPTION_VAL_450_HR NULL
#define OPTION_VAL_460_HR NULL
#define OPTION_VAL_470_HR NULL
#define OPTION_VAL_480_HR NULL
#define OPTION_VAL_490_HR NULL
#define OPTION_VAL_500_HR NULL
#define OPTION_VAL_510_HR NULL
#define OPTION_VAL_520_HR NULL
#define OPTION_VAL_530_HR NULL
#define OPTION_VAL_540_HR NULL
#define OPTION_VAL_550_HR NULL
#define OPTION_VAL_560_HR NULL
#define OPTION_VAL_570_HR NULL
#define OPTION_VAL_580_HR NULL
#define OPTION_VAL_590_HR NULL
#define OPTION_VAL_600_HR NULL
#define OPTION_VAL_610_HR NULL
#define OPTION_VAL_620_HR NULL
#define OPTION_VAL_630_HR NULL
#define OPTION_VAL_640_HR NULL
#define OPTION_VAL_650_HR NULL
#define OPTION_VAL_660_HR NULL
#define OPTION_VAL_670_HR NULL
#define OPTION_VAL_680_HR NULL
#define OPTION_VAL_690_HR NULL
#define OPTION_VAL_700_HR NULL
#define OPTION_VAL_710_HR NULL
#define OPTION_VAL_720_HR NULL
#define OPTION_VAL_730_HR NULL
#define OPTION_VAL_740_HR NULL
#define OPTION_VAL_750_HR NULL
#define BEETLE_OPT_GPU_OVERCLOCK_LABEL_HR NULL
#define BEETLE_OPT_GPU_OVERCLOCK_INFO_0_HR NULL
#define OPTION_VAL_1X_NATIVE_HR NULL
#define OPTION_VAL_2X_HR NULL
#define OPTION_VAL_4X_HR NULL
#define OPTION_VAL_8X_HR NULL
#define OPTION_VAL_16X_HR NULL
#define OPTION_VAL_32X_HR NULL
#define BEETLE_OPT_GTE_OVERCLOCK_LABEL_HR NULL
#define BEETLE_OPT_GTE_OVERCLOCK_INFO_0_HR NULL
#define BEETLE_OPT_CD_ACCESS_METHOD_LABEL_HR NULL
#define BEETLE_OPT_CD_ACCESS_METHOD_INFO_0_HR NULL
#define OPTION_VAL_SYNC_HR NULL
#define OPTION_VAL_ASYNC_HR NULL
#define OPTION_VAL_PRECACHE_HR NULL
#define BEETLE_OPT_CD_FASTLOAD_LABEL_HR NULL
#define BEETLE_OPT_CD_FASTLOAD_INFO_0_HR NULL
#define OPTION_VAL_2X_NATIVE_HR NULL
#define OPTION_VAL_6X_HR NULL
#define OPTION_VAL_10X_HR NULL
#define OPTION_VAL_12X_HR NULL
#define OPTION_VAL_14X_HR NULL
#define BEETLE_OPT_SKIP_BIOS_LABEL_HR NULL
#define BEETLE_OPT_SKIP_BIOS_INFO_0_HR NULL
#define BEETLE_OPT_OVERRIDE_BIOS_LABEL_HR NULL
#define BEETLE_OPT_OVERRIDE_BIOS_INFO_0_HR NULL
#define OPTION_VAL_PSXONPSP_HR NULL
#define OPTION_VAL_PS1_ROM_HR NULL
#define OPTION_VAL_OPENBIOS_HR NULL
#define BEETLE_OPT_REGION_LABEL_HR "Regija sustava"
#define BEETLE_OPT_REGION_INFO_0_HR NULL
#define OPTION_VAL_AUTO_HR NULL
#define OPTION_VAL_NTSC_J_HR NULL
#define OPTION_VAL_NTSC_U_HR NULL
#define OPTION_VAL_PAL_HR NULL
#define BEETLE_OPT_INTERNAL_RESOLUTION_LABEL_HR NULL
#define BEETLE_OPT_INTERNAL_RESOLUTION_INFO_0_HR NULL
#define BEETLE_OPT_RENDERER_LABEL_HR NULL
#define BEETLE_OPT_RENDERER_INFO_0_HR NULL
#define OPTION_VAL_HARDWARE_HR NULL
#define OPTION_VAL_HARDWARE_GL_HR NULL
#define OPTION_VAL_HARDWARE_VK_HR NULL
#define OPTION_VAL_SOFTWARE_HR NULL
#define BEETLE_OPT_RENDERER_SOFTWARE_FB_LABEL_HR NULL
#define BEETLE_OPT_RENDERER_SOFTWARE_FB_INFO_0_HR NULL
#define BEETLE_OPT_DEPTH_LABEL_HR NULL
#define BEETLE_OPT_DEPTH_INFO_0_HR NULL
#define OPTION_VAL_16BPP_NATIVE_HR NULL
#define OPTION_VAL_32BPP_HR NULL
#define BEETLE_OPT_DITHER_MODE_LABEL_HR NULL
#define BEETLE_OPT_DITHER_MODE_INFO_0_HR NULL
#define OPTION_VAL_INTERNAL_RESOLUTION_HR NULL
#define BEETLE_OPT_SCALED_UV_OFFSET_LABEL_HR NULL
#define BEETLE_OPT_SCALED_UV_OFFSET_INFO_0_HR NULL
#define BEETLE_OPT_FILTER_LABEL_HR NULL
#define BEETLE_OPT_FILTER_INFO_0_HR NULL
#define OPTION_VAL_NEAREST_HR NULL
#define OPTION_VAL_SABR_HR NULL
#define OPTION_VAL_XBR_HR NULL
#define OPTION_VAL_BILINEAR_HR NULL
#define OPTION_VAL_3_POINT_HR NULL
#define OPTION_VAL_JINC2_HR NULL
#define BEETLE_OPT_FILTER_EXCLUDE_SPRITE_LABEL_HR NULL
#define BEETLE_OPT_FILTER_EXCLUDE_SPRITE_INFO_0_HR NULL
#define OPTION_VAL_OPAQUE_HR NULL
#define OPTION_VAL_ALL_HR NULL
#define BEETLE_OPT_FILTER_EXCLUDE_2D_POLYGON_LABEL_HR NULL
#define BEETLE_OPT_FILTER_EXCLUDE_2D_POLYGON_INFO_0_HR NULL
#define BEETLE_OPT_ADAPTIVE_SMOOTHING_LABEL_HR NULL
#define BEETLE_OPT_ADAPTIVE_SMOOTHING_INFO_0_HR NULL
#define BEETLE_OPT_SUPER_SAMPLING_LABEL_HR NULL
#define BEETLE_OPT_SUPER_SAMPLING_INFO_0_HR NULL
#define BEETLE_OPT_MSAA_LABEL_HR NULL
#define BEETLE_OPT_MSAA_INFO_0_HR NULL
#define OPTION_VAL_1X_HR NULL
#define BEETLE_OPT_MDEC_YUV_LABEL_HR NULL
#define BEETLE_OPT_MDEC_YUV_INFO_0_HR NULL
#define BEETLE_OPT_TRACK_TEXTURES_LABEL_HR NULL
#define BEETLE_OPT_TRACK_TEXTURES_INFO_0_HR NULL
#define BEETLE_OPT_DUMP_TEXTURES_LABEL_HR NULL
#define BEETLE_OPT_DUMP_TEXTURES_INFO_0_HR NULL
#define BEETLE_OPT_REPLACE_TEXTURES_LABEL_HR NULL
#define BEETLE_OPT_REPLACE_TEXTURES_INFO_0_HR NULL
#define BEETLE_OPT_PGXP_MODE_LABEL_HR NULL
#define BEETLE_OPT_PGXP_MODE_INFO_0_HR NULL
#define OPTION_VAL_MEMORY_ONLY_HR NULL
#define OPTION_VAL_MEMORY_CPU_HR NULL
#define BEETLE_OPT_PGXP_2D_TOL_LABEL_HR NULL
#define BEETLE_OPT_PGXP_2D_TOL_INFO_0_HR NULL
#define OPTION_VAL_0PX_HR NULL
#define OPTION_VAL_1PX_HR NULL
#define OPTION_VAL_2PX_HR NULL
#define OPTION_VAL_3PX_HR NULL
#define OPTION_VAL_4PX_HR NULL
#define OPTION_VAL_5PX_HR NULL
#define OPTION_VAL_6PX_HR NULL
#define OPTION_VAL_7PX_HR NULL
#define OPTION_VAL_8PX_HR NULL
#define BEETLE_OPT_PGXP_NCLIP_LABEL_HR NULL
#define BEETLE_OPT_PGXP_NCLIP_INFO_0_HR NULL
#define BEETLE_OPT_PGXP_VERTEX_LABEL_HR NULL
#define BEETLE_OPT_PGXP_VERTEX_INFO_0_HR NULL
#define BEETLE_OPT_PGXP_TEXTURE_LABEL_HR NULL
#define BEETLE_OPT_PGXP_TEXTURE_INFO_0_HR NULL
#define BEETLE_OPT_DISPLAY_INTERNAL_FPS_LABEL_HR NULL
#define BEETLE_OPT_DISPLAY_INTERNAL_FPS_INFO_0_HR NULL
#define BEETLE_OPT_DISPLAY_OSD_LABEL_HR NULL
#define BEETLE_OPT_DISPLAY_OSD_INFO_0_HR NULL
#define BEETLE_OPT_DISPLAY_VRAM_LABEL_HR NULL
#define BEETLE_OPT_DISPLAY_VRAM_INFO_0_HR NULL
#define BEETLE_OPT_ANALOG_CALIBRATION_LABEL_HR NULL
#define BEETLE_OPT_ANALOG_CALIBRATION_INFO_0_HR NULL
#define BEETLE_OPT_ANALOG_TOGGLE_LABEL_HR NULL
#define BEETLE_OPT_ANALOG_TOGGLE_INFO_0_HR NULL
#define OPTION_VAL_ENABLED_ANALOG_HR NULL
#define BEETLE_OPT_ANALOG_TOGGLE_COMBO_LABEL_HR NULL
#define BEETLE_OPT_ANALOG_TOGGLE_COMBO_INFO_0_HR NULL
#define OPTION_VAL_L1_L2_R1_R2_START_SELECT_HR NULL
#define OPTION_VAL_L1_R1_SELECT_HR NULL
#define OPTION_VAL_L1_R1_START_HR NULL
#define OPTION_VAL_L1_R1_L3_HR NULL
#define OPTION_VAL_L1_R1_R3_HR NULL
#define OPTION_VAL_L2_R2_SELECT_HR NULL
#define OPTION_VAL_L2_R2_START_HR NULL
#define OPTION_VAL_L2_R2_L3_HR NULL
#define OPTION_VAL_L2_R2_R3_HR NULL
#define OPTION_VAL_L3_R3_HR NULL
#define BEETLE_OPT_ANALOG_TOGGLE_HOLD_LABEL_HR NULL
#define BEETLE_OPT_ANALOG_TOGGLE_HOLD_INFO_0_HR NULL
#define OPTION_VAL_0_HR NULL
#define OPTION_VAL_1_HR NULL
#define OPTION_VAL_2_HR NULL
#define OPTION_VAL_3_HR NULL
#define OPTION_VAL_4_HR NULL
#define OPTION_VAL_5_HR NULL
#define BEETLE_OPT_ENABLE_MULTITAP_PORT1_LABEL_HR NULL
#define BEETLE_OPT_ENABLE_MULTITAP_PORT1_INFO_0_HR NULL
#define BEETLE_OPT_ENABLE_MULTITAP_PORT2_LABEL_HR NULL
#define BEETLE_OPT_ENABLE_MULTITAP_PORT2_INFO_0_HR NULL
#define BEETLE_OPT_GUN_INPUT_MODE_LABEL_HR NULL
#define BEETLE_OPT_GUN_INPUT_MODE_INFO_0_HR NULL
#define OPTION_VAL_LIGHTGUN_HR NULL
#define OPTION_VAL_TOUCHSCREEN_HR NULL
#define BEETLE_OPT_GUN_CURSOR_LABEL_HR NULL
#define BEETLE_OPT_GUN_CURSOR_INFO_0_HR NULL
#define OPTION_VAL_CROSS_HR NULL
#define OPTION_VAL_DOT_HR NULL
#define BEETLE_OPT_CROSSHAIR_COLOR_P1_LABEL_HR NULL
#define BEETLE_OPT_CROSSHAIR_COLOR_P1_INFO_0_HR NULL
#define OPTION_VAL_RED_HR NULL
#define OPTION_VAL_BLUE_HR NULL
#define OPTION_VAL_GREEN_HR NULL
#define OPTION_VAL_ORANGE_HR NULL
#define OPTION_VAL_YELLOW_HR NULL
#define OPTION_VAL_CYAN_HR NULL
#define OPTION_VAL_PINK_HR NULL
#define OPTION_VAL_PURPLE_HR NULL
#define OPTION_VAL_BLACK_HR NULL
#define OPTION_VAL_WHITE_HR NULL
#define BEETLE_OPT_CROSSHAIR_COLOR_P2_LABEL_HR NULL
#define BEETLE_OPT_CROSSHAIR_COLOR_P2_INFO_0_HR NULL
#define BEETLE_OPT_MOUSE_SENSITIVITY_LABEL_HR NULL
#define BEETLE_OPT_MOUSE_SENSITIVITY_INFO_0_HR NULL
#define OPTION_VAL_5_O42_HR NULL
#define OPTION_VAL_10_HR NULL
#define OPTION_VAL_15_HR NULL
#define OPTION_VAL_20_HR NULL
#define OPTION_VAL_25_HR NULL
#define OPTION_VAL_30_HR NULL
#define OPTION_VAL_35_HR NULL
#define OPTION_VAL_40_HR NULL
#define OPTION_VAL_45_HR NULL
#define OPTION_VAL_55_HR NULL
#define OPTION_VAL_65_HR NULL
#define OPTION_VAL_75_HR NULL
#define OPTION_VAL_85_HR NULL
#define OPTION_VAL_95_HR NULL
#define OPTION_VAL_100_O42_HR NULL
#define OPTION_VAL_105_HR NULL
#define OPTION_VAL_115_HR NULL
#define OPTION_VAL_125_HR NULL
#define OPTION_VAL_135_HR NULL
#define OPTION_VAL_145_HR NULL
#define OPTION_VAL_155_HR NULL
#define OPTION_VAL_165_HR NULL
#define OPTION_VAL_175_HR NULL
#define OPTION_VAL_185_HR NULL
#define OPTION_VAL_195_HR NULL
#define BEETLE_OPT_NEGCON_RESPONSE_LABEL_HR NULL
#define BEETLE_OPT_NEGCON_RESPONSE_INFO_0_HR NULL
#define OPTION_VAL_LINEAR_HR NULL
#define OPTION_VAL_QUADRATIC_HR NULL
#define OPTION_VAL_CUBIC_HR NULL
#define BEETLE_OPT_NEGCON_DEADZONE_LABEL_HR NULL
#define BEETLE_OPT_NEGCON_DEADZONE_INFO_0_HR NULL
#define OPTION_VAL_0_O44_HR NULL
#define BEETLE_OPT_USE_MEDNAFEN_MEMCARD0_METHOD_LABEL_HR NULL
#define BEETLE_OPT_USE_MEDNAFEN_MEMCARD0_METHOD_INFO_0_HR NULL
#define OPTION_VAL_LIBRETRO_HR NULL
#define OPTION_VAL_MEDNAFEN_HR NULL
#define BEETLE_OPT_ENABLE_MEMCARD1_LABEL_HR NULL
#define BEETLE_OPT_ENABLE_MEMCARD1_INFO_0_HR NULL
#define BEETLE_OPT_SHARED_MEMORY_CARDS_LABEL_HR NULL
#define BEETLE_OPT_SHARED_MEMORY_CARDS_INFO_0_HR NULL
#define BEETLE_OPT_MEMCARD_LEFT_INDEX_LABEL_HR NULL
#define BEETLE_OPT_MEMCARD_LEFT_INDEX_INFO_0_HR NULL
#define OPTION_VAL_0_O48_HR NULL
#define BEETLE_OPT_MEMCARD_RIGHT_INDEX_LABEL_HR NULL
#define BEETLE_OPT_MEMCARD_RIGHT_INDEX_INFO_0_HR NULL
#define OPTION_VAL_1_O49_HR NULL
#define BEETLE_OPT_LINE_RENDER_LABEL_HR NULL
#define BEETLE_OPT_LINE_RENDER_INFO_0_HR NULL
#define OPTION_VAL_DEFAULT_HR NULL
#define OPTION_VAL_AGGRESSIVE_HR NULL
#define BEETLE_OPT_WIDESCREEN_HACK_LABEL_HR NULL
#define BEETLE_OPT_WIDESCREEN_HACK_INFO_0_HR NULL
#define BEETLE_OPT_WIDESCREEN_HACK_ASPECT_RATIO_LABEL_HR NULL
#define BEETLE_OPT_WIDESCREEN_HACK_ASPECT_RATIO_INFO_0_HR NULL
#define OPTION_VAL_16_9_HR NULL
#define OPTION_VAL_16_10_HR NULL
#define OPTION_VAL_18_9_HR NULL
#define OPTION_VAL_19_9_HR NULL
#define OPTION_VAL_20_9_HR NULL
#define OPTION_VAL_21_9_HR NULL
#define OPTION_VAL_32_9_HR NULL
#define BEETLE_OPT_CPU_DYNAREC_LABEL_HR NULL
#define BEETLE_OPT_CPU_DYNAREC_INFO_0_HR NULL
#define OPTION_VAL_DISABLED_HR NULL
#define OPTION_VAL_EXECUTE_HR NULL
#define OPTION_VAL_EXECUTE_ONE_HR NULL
#define OPTION_VAL_RUN_INTERPRETER_HR NULL
#define BEETLE_OPT_DYNAREC_INVALIDATE_LABEL_HR NULL
#define BEETLE_OPT_DYNAREC_INVALIDATE_INFO_0_HR NULL
#define OPTION_VAL_FULL_HR NULL
#define OPTION_VAL_DMA_HR NULL
#define BEETLE_OPT_DYNAREC_EVENTCYCLES_LABEL_HR NULL
#define BEETLE_OPT_DYNAREC_EVENTCYCLES_INFO_0_HR NULL
#define OPTION_VAL_128_HR NULL
#define BEETLE_OPT_DYNAREC_SPU_SAMPLES_LABEL_HR NULL
#define BEETLE_OPT_DYNAREC_SPU_SAMPLES_INFO_0_HR NULL
#define BEETLE_OPT_CORE_TIMING_FPS_LABEL_HR NULL
#define BEETLE_OPT_CORE_TIMING_FPS_INFO_0_HR NULL
#define OPTION_VAL_AUTO_TOGGLE_HR NULL
#define OPTION_VAL_FORCE_PROGRESSIVE_HR NULL
#define OPTION_VAL_FORCE_INTERLACED_HR NULL
#define BEETLE_OPT_DEINTERLACER_LABEL_HR NULL
#define BEETLE_OPT_DEINTERLACER_INFO_0_HR NULL
#define OPTION_VAL_WEAVE_HR NULL
#define OPTION_VAL_BOB_HR NULL
#define OPTION_VAL_BOB_OFFSET_HR NULL
#define OPTION_VAL_FASTMAD_HR NULL
#define BEETLE_OPT_FRAME_DUPING_LABEL_HR NULL
#define BEETLE_OPT_FRAME_DUPING_INFO_0_HR NULL
#define BEETLE_OPT_PAL_VIDEO_TIMING_OVERRIDE_LABEL_HR NULL
#define BEETLE_OPT_PAL_VIDEO_TIMING_OVERRIDE_INFO_0_HR NULL
#define BEETLE_OPT_ASPECT_RATIO_LABEL_HR NULL
#define BEETLE_OPT_ASPECT_RATIO_INFO_0_HR NULL
#define OPTION_VAL_CORRECTED_HR NULL
#define OPTION_VAL_UNCORRECTED_HR NULL
#define OPTION_VAL_4_3_HR NULL
#define OPTION_VAL_NTSC_HR NULL
#define BEETLE_OPT_CROP_OVERSCAN_LABEL_HR NULL
#define BEETLE_OPT_CROP_OVERSCAN_INFO_0_HR NULL
#define OPTION_VAL_STATIC_HR NULL
#define OPTION_VAL_SMART_HR NULL
#define BEETLE_OPT_IMAGE_CROP_LABEL_HR NULL
#define BEETLE_OPT_IMAGE_CROP_INFO_0_HR NULL
#define OPTION_VAL_9PX_HR NULL
#define OPTION_VAL_10PX_HR NULL
#define OPTION_VAL_11PX_HR NULL
#define OPTION_VAL_12PX_HR NULL
#define OPTION_VAL_13PX_HR NULL
#define OPTION_VAL_14PX_HR NULL
#define OPTION_VAL_15PX_HR NULL
#define OPTION_VAL_16PX_HR NULL
#define OPTION_VAL_17PX_HR NULL
#define OPTION_VAL_18PX_HR NULL
#define OPTION_VAL_19PX_HR NULL
#define OPTION_VAL_20PX_HR NULL
#define BEETLE_OPT_IMAGE_OFFSET_LABEL_HR NULL
#define BEETLE_OPT_IMAGE_OFFSET_INFO_0_HR NULL
#define OPTION_VAL_12PX_O64_HR NULL
#define OPTION_VAL_11PX_O64_HR NULL
#define OPTION_VAL_10PX_O64_HR NULL
#define OPTION_VAL_9PX_O64_HR NULL
#define OPTION_VAL_8PX_O64_HR NULL
#define OPTION_VAL_7PX_O64_HR NULL
#define OPTION_VAL_6PX_O64_HR NULL
#define OPTION_VAL_5PX_O64_HR NULL
#define OPTION_VAL_4PX_O64_HR NULL
#define OPTION_VAL_3PX_O64_HR NULL
#define OPTION_VAL_2PX_O64_HR NULL
#define OPTION_VAL_1PX_O64_HR NULL
#define OPTION_VAL_1PX_O64_0_HR NULL
#define OPTION_VAL_2PX_O64_0_HR NULL
#define OPTION_VAL_3PX_O64_0_HR NULL
#define OPTION_VAL_4PX_O64_0_HR NULL
#define OPTION_VAL_5PX_O64_0_HR NULL
#define OPTION_VAL_6PX_O64_0_HR NULL
#define OPTION_VAL_7PX_O64_0_HR NULL
#define OPTION_VAL_8PX_O64_0_HR NULL
#define OPTION_VAL_9PX_O64_0_HR NULL
#define OPTION_VAL_10PX_O64_0_HR NULL
#define OPTION_VAL_11PX_O64_0_HR NULL
#define OPTION_VAL_12PX_O64_0_HR NULL
#define BEETLE_OPT_IMAGE_OFFSET_CYCLES_LABEL_HR NULL
#define BEETLE_OPT_IMAGE_OFFSET_CYCLES_INFO_0_HR NULL
#define BEETLE_OPT_INITIAL_SCANLINE_LABEL_HR NULL
#define BEETLE_OPT_INITIAL_SCANLINE_INFO_0_HR NULL
#define BEETLE_OPT_LAST_SCANLINE_LABEL_HR NULL
#define BEETLE_OPT_LAST_SCANLINE_INFO_0_HR NULL
#define OPTION_VAL_239_HR NULL
#define BEETLE_OPT_INITIAL_SCANLINE_PAL_LABEL_HR NULL
#define BEETLE_OPT_INITIAL_SCANLINE_PAL_INFO_0_HR NULL
#define BEETLE_OPT_LAST_SCANLINE_PAL_LABEL_HR NULL
#define BEETLE_OPT_LAST_SCANLINE_PAL_INFO_0_HR NULL
#define OPTION_VAL_287_HR NULL

struct retro_core_option_v2_category option_cats_hr[] = {
   {
      "system",
      CATEGORY_SYSTEM_LABEL_HR,
      CATEGORY_SYSTEM_INFO_0_HR
   },
   {
      "video",
      CATEGORY_VIDEO_LABEL_HR,
      CATEGORY_VIDEO_INFO_0_HR
   },
   {
      "pgxp",
      CATEGORY_PGXP_LABEL_HR,
      CATEGORY_PGXP_INFO_0_HR
   },
   {
      "osd",
      CATEGORY_OSD_LABEL_HR,
      CATEGORY_OSD_INFO_0_HR
   },
   {
      "input",
      CATEGORY_INPUT_LABEL_HR,
      CATEGORY_INPUT_INFO_0_HR
   },
   {
      "memcards",
      CATEGORY_MEMCARDS_LABEL_HR,
      CATEGORY_MEMCARDS_INFO_0_HR
   },
   {
      "hacks",
      CATEGORY_HACKS_LABEL_HR,
      CATEGORY_HACKS_INFO_0_HR
   },
   { NULL, NULL, NULL },
};
struct retro_core_option_v2_definition option_defs_hr[] = {
   {
      BEETLE_OPT(cpu_freq_scale),
      BEETLE_OPT_CPU_FREQ_SCALE_LABEL_HR,
      NULL,
      BEETLE_OPT_CPU_FREQ_SCALE_INFO_0_HR,
      NULL,
      "system",
      {
         { "50%",           OPTION_VAL_50_HR },
         { "60%",           OPTION_VAL_60_HR },
         { "70%",           OPTION_VAL_70_HR },
         { "80%",           OPTION_VAL_80_HR },
         { "90%",           OPTION_VAL_90_HR },
         { "100%",          OPTION_VAL_100_HR },
         { "110%",          OPTION_VAL_110_HR },
         { "120%",          OPTION_VAL_120_HR },
         { "130%",          OPTION_VAL_130_HR },
         { "140%",          OPTION_VAL_140_HR },
         { "150%",          OPTION_VAL_150_HR },
         { "160%",          OPTION_VAL_160_HR },
         { "170%",          OPTION_VAL_170_HR },
         { "180%",          OPTION_VAL_180_HR },
         { "190%",          OPTION_VAL_190_HR },
         { "200%",          OPTION_VAL_200_HR },
         { "210%",          OPTION_VAL_210_HR },
         { "220%",          OPTION_VAL_220_HR },
         { "230%",          OPTION_VAL_230_HR },
         { "240%",          OPTION_VAL_240_HR },
         { "250%",          OPTION_VAL_250_HR },
         { "260%",          OPTION_VAL_260_HR },
         { "270%",          OPTION_VAL_270_HR },
         { "280%",          OPTION_VAL_280_HR },
         { "290%",          OPTION_VAL_290_HR },
         { "300%",          OPTION_VAL_300_HR },
         { "310%",          OPTION_VAL_310_HR },
         { "320%",          OPTION_VAL_320_HR },
         { "330%",          OPTION_VAL_330_HR },
         { "340%",          OPTION_VAL_340_HR },
         { "350%",          OPTION_VAL_350_HR },
         { "360%",          OPTION_VAL_360_HR },
         { "370%",          OPTION_VAL_370_HR },
         { "380%",          OPTION_VAL_380_HR },
         { "390%",          OPTION_VAL_390_HR },
         { "400%",          OPTION_VAL_400_HR },
         { "410%",          OPTION_VAL_410_HR },
         { "420%",          OPTION_VAL_420_HR },
         { "430%",          OPTION_VAL_430_HR },
         { "440%",          OPTION_VAL_440_HR },
         { "450%",          OPTION_VAL_450_HR },
         { "460%",          OPTION_VAL_460_HR },
         { "470%",          OPTION_VAL_470_HR },
         { "480%",          OPTION_VAL_480_HR },
         { "490%",          OPTION_VAL_490_HR },
         { "500%",          OPTION_VAL_500_HR },
         { "510%",          OPTION_VAL_510_HR },
         { "520%",          OPTION_VAL_520_HR },
         { "530%",          OPTION_VAL_530_HR },
         { "540%",          OPTION_VAL_540_HR },
         { "550%",          OPTION_VAL_550_HR },
         { "560%",          OPTION_VAL_560_HR },
         { "570%",          OPTION_VAL_570_HR },
         { "580%",          OPTION_VAL_580_HR },
         { "590%",          OPTION_VAL_590_HR },
         { "600%",          OPTION_VAL_600_HR },
         { "610%",          OPTION_VAL_610_HR },
         { "620%",          OPTION_VAL_620_HR },
         { "630%",          OPTION_VAL_630_HR },
         { "640%",          OPTION_VAL_640_HR },
         { "650%",          OPTION_VAL_650_HR },
         { "660%",          OPTION_VAL_660_HR },
         { "670%",          OPTION_VAL_670_HR },
         { "680%",          OPTION_VAL_680_HR },
         { "690%",          OPTION_VAL_690_HR },
         { "700%",          OPTION_VAL_700_HR },
         { "710%",          OPTION_VAL_710_HR },
         { "720%",          OPTION_VAL_720_HR },
         { "730%",          OPTION_VAL_730_HR },
         { "740%",          OPTION_VAL_740_HR },
         { "750%",          OPTION_VAL_750_HR },
         { NULL, NULL },
      },
      "100%"
   },
   {
      BEETLE_OPT(gpu_overclock),
      BEETLE_OPT_GPU_OVERCLOCK_LABEL_HR,
      NULL,
      BEETLE_OPT_GPU_OVERCLOCK_INFO_0_HR,
      NULL,
      "system",
      {
         { "1x(native)", OPTION_VAL_1X_NATIVE_HR },
         { "2x",         OPTION_VAL_2X_HR },
         { "4x",         OPTION_VAL_4X_HR },
         { "8x",         OPTION_VAL_8X_HR },
         { "16x",        OPTION_VAL_16X_HR },
         { "32x",        OPTION_VAL_32X_HR },
         { NULL, NULL },
      },
      "1x(native)"
   },
   {
      BEETLE_OPT(gte_overclock),
      BEETLE_OPT_GTE_OVERCLOCK_LABEL_HR,
      NULL,
      BEETLE_OPT_GTE_OVERCLOCK_INFO_0_HR,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
#ifndef EMSCRIPTEN
   {
      BEETLE_OPT(cd_access_method),
      BEETLE_OPT_CD_ACCESS_METHOD_LABEL_HR,
      NULL,
      BEETLE_OPT_CD_ACCESS_METHOD_INFO_0_HR,
      NULL,
      "system",
      {
         { "sync",     OPTION_VAL_SYNC_HR },
         { "async",    OPTION_VAL_ASYNC_HR },
         { "precache", OPTION_VAL_PRECACHE_HR },
         { NULL, NULL },
      },
      "sync"
   },
#endif
   {
      BEETLE_OPT(cd_fastload),
      BEETLE_OPT_CD_FASTLOAD_LABEL_HR,
      NULL,
      BEETLE_OPT_CD_FASTLOAD_INFO_0_HR,
      NULL,
      "system",
      {
         { "2x(native)", OPTION_VAL_2X_NATIVE_HR },
         { "4x",          OPTION_VAL_4X_HR },
         { "6x",          OPTION_VAL_6X_HR },
         { "8x",          OPTION_VAL_8X_HR },
         { "10x",         OPTION_VAL_10X_HR },
         { "12x",         OPTION_VAL_12X_HR },
         { "14x",         OPTION_VAL_14X_HR },
         { NULL, NULL },
      },
      "2x(native)"
   },
   {
      BEETLE_OPT(skip_bios),
      BEETLE_OPT_SKIP_BIOS_LABEL_HR,
      NULL,
      BEETLE_OPT_SKIP_BIOS_INFO_0_HR,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(override_bios),
      BEETLE_OPT_OVERRIDE_BIOS_LABEL_HR,
      NULL,
      BEETLE_OPT_OVERRIDE_BIOS_INFO_0_HR,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "psxonpsp", OPTION_VAL_PSXONPSP_HR },
         { "ps1_rom",  OPTION_VAL_PS1_ROM_HR },
         { "openbios", OPTION_VAL_OPENBIOS_HR     },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(region),
      BEETLE_OPT_REGION_LABEL_HR,
      NULL,
      BEETLE_OPT_REGION_INFO_0_HR,
      NULL,
      "system",
      {
         { "auto",   OPTION_VAL_AUTO_HR },
         { "ntsc-j", OPTION_VAL_NTSC_J_HR },
         { "ntsc-u", OPTION_VAL_NTSC_U_HR },
         { "pal",    OPTION_VAL_PAL_HR },
         { NULL, NULL },
      },
      "auto"
   },
   {
      BEETLE_OPT(internal_resolution),
      BEETLE_OPT_INTERNAL_RESOLUTION_LABEL_HR,
      NULL,
      BEETLE_OPT_INTERNAL_RESOLUTION_INFO_0_HR,
      NULL,
      "video",
      {
         { "1x(native)", OPTION_VAL_1X_NATIVE_HR },
         { "2x",         OPTION_VAL_2X_HR },
         { "4x",         OPTION_VAL_4X_HR },
         { "8x",         OPTION_VAL_8X_HR },
         { "16x",        OPTION_VAL_16X_HR },
         { NULL, NULL },
      },
      "1x(native)"
   },
#if defined(HAVE_OPENGL) || defined(HAVE_OPENGLES) || defined(HAVE_VULKAN)
   {
      BEETLE_OPT(renderer),
      BEETLE_OPT_RENDERER_LABEL_HR,
      NULL,
      BEETLE_OPT_RENDERER_INFO_0_HR,
      NULL,
      "video",
      {
         { "hardware",    OPTION_VAL_HARDWARE_HR },
#if defined(HAVE_OPENGL) || defined(HAVE_OPENGLES)
         { "hardware_gl", OPTION_VAL_HARDWARE_GL_HR },
#endif
#if defined(HAVE_VULKAN)
         { "hardware_vk", OPTION_VAL_HARDWARE_VK_HR },
#endif
         { "software",    OPTION_VAL_SOFTWARE_HR },
         { NULL, NULL },
      },
      "hardware"
   },
   {
      BEETLE_OPT(renderer_software_fb),
      BEETLE_OPT_RENDERER_SOFTWARE_FB_LABEL_HR,
      NULL,
      BEETLE_OPT_RENDERER_SOFTWARE_FB_INFO_0_HR,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled"
   },
#endif
#if defined(HAVE_OPENGL) || defined(HAVE_OPENGLES)
   {
      BEETLE_OPT(depth),
      BEETLE_OPT_DEPTH_LABEL_HR,
      NULL,
      BEETLE_OPT_DEPTH_INFO_0_HR,
      NULL,
      "video",
      {
         { "16bpp(native)", OPTION_VAL_16BPP_NATIVE_HR },
         { "32bpp",         OPTION_VAL_32BPP_HR },
         { NULL, NULL },
      },
      "16bpp(native)"
   },
   // Sort of, it's more like 15-bit high color and 24-bit true color for visible output. The alpha channel is used for mask bit. Vulkan renderer uses ABGR1555_555 for 31 bits internal? FMVs are always 24-bit on all renderers like original hardware (BGR888, no alpha)
#endif
   {
      BEETLE_OPT(dither_mode),
      BEETLE_OPT_DITHER_MODE_LABEL_HR,
      NULL,
      BEETLE_OPT_DITHER_MODE_INFO_0_HR,
      NULL,
      "video",
      {
         { "1x(native)",          OPTION_VAL_1X_NATIVE_HR },
         { "internal resolution", OPTION_VAL_INTERNAL_RESOLUTION_HR },
         { "disabled",            NULL },
         { NULL, NULL },
      },
      "1x(native)"
   },
#ifdef HAVE_VULKAN
   {
      BEETLE_OPT(scaled_uv_offset),
      BEETLE_OPT_SCALED_UV_OFFSET_LABEL_HR,
      NULL,
      BEETLE_OPT_SCALED_UV_OFFSET_INFO_0_HR,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled"
   },
#endif
#if defined(HAVE_OPENGL) || defined(HAVE_OPENGLES) || defined(HAVE_VULKAN)
   {
      BEETLE_OPT(filter),
      BEETLE_OPT_FILTER_LABEL_HR,
      NULL,
      BEETLE_OPT_FILTER_INFO_0_HR,
      NULL,
      "video",
      {
         { "nearest",  OPTION_VAL_NEAREST_HR },
         { "SABR",     OPTION_VAL_SABR_HR },
         { "xBR",      OPTION_VAL_XBR_HR },
         { "bilinear", OPTION_VAL_BILINEAR_HR },
         { "3-point",  OPTION_VAL_3_POINT_HR },
         { "JINC2",    OPTION_VAL_JINC2_HR },
         { NULL, NULL },
      },
      "nearest"
   },
#ifdef HAVE_VULKAN
   {
      BEETLE_OPT(filter_exclude_sprite),
      BEETLE_OPT_FILTER_EXCLUDE_SPRITE_LABEL_HR,
      NULL,
      BEETLE_OPT_FILTER_EXCLUDE_SPRITE_INFO_0_HR,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "opaque", OPTION_VAL_OPAQUE_HR },
         { "all", OPTION_VAL_ALL_HR },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(filter_exclude_2d_polygon),
      BEETLE_OPT_FILTER_EXCLUDE_2D_POLYGON_LABEL_HR,
      NULL,
      BEETLE_OPT_FILTER_EXCLUDE_2D_POLYGON_INFO_0_HR,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "opaque", OPTION_VAL_OPAQUE_HR },
         { "all", OPTION_VAL_ALL_HR },
         { NULL, NULL },
      },
      "disabled"
   },
#endif
#endif
#ifdef HAVE_VULKAN
   {
      BEETLE_OPT(adaptive_smoothing),
      BEETLE_OPT_ADAPTIVE_SMOOTHING_LABEL_HR,
      NULL,
      BEETLE_OPT_ADAPTIVE_SMOOTHING_INFO_0_HR,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(super_sampling),
      BEETLE_OPT_SUPER_SAMPLING_LABEL_HR,
      NULL,
      BEETLE_OPT_SUPER_SAMPLING_INFO_0_HR,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(msaa),
      BEETLE_OPT_MSAA_LABEL_HR,
      NULL,
      BEETLE_OPT_MSAA_INFO_0_HR,
      NULL,
      "video",
      {
         { "1x",  OPTION_VAL_1X_HR },
         { "2x",  OPTION_VAL_2X_HR },
         { "4x",  OPTION_VAL_4X_HR },
         { "8x",  OPTION_VAL_8X_HR },
         { "16x", OPTION_VAL_16X_HR },
         { NULL, NULL },
      },
      "1x"
   },
   {
      BEETLE_OPT(mdec_yuv),
      BEETLE_OPT_MDEC_YUV_LABEL_HR,
      NULL,
      BEETLE_OPT_MDEC_YUV_INFO_0_HR,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(track_textures),
      BEETLE_OPT_TRACK_TEXTURES_LABEL_HR,
      NULL,
      BEETLE_OPT_TRACK_TEXTURES_INFO_0_HR,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
#ifdef TEXTURE_DUMPING_ENABLED
   {
      BEETLE_OPT(dump_textures),
      BEETLE_OPT_DUMP_TEXTURES_LABEL_HR,
      NULL,
      BEETLE_OPT_DUMP_TEXTURES_INFO_0_HR,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
#endif
   {
      BEETLE_OPT(replace_textures),
      BEETLE_OPT_REPLACE_TEXTURES_LABEL_HR,
      NULL,
      BEETLE_OPT_REPLACE_TEXTURES_INFO_0_HR,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
#endif
   {
      BEETLE_OPT(pgxp_mode),
      BEETLE_OPT_PGXP_MODE_LABEL_HR,
      NULL,
      BEETLE_OPT_PGXP_MODE_INFO_0_HR,
      NULL,
      "pgxp",
      {
         { "disabled",     NULL },
         { "memory only",  OPTION_VAL_MEMORY_ONLY_HR },
         { "memory + CPU", OPTION_VAL_MEMORY_CPU_HR },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(pgxp_2d_tol),
      BEETLE_OPT_PGXP_2D_TOL_LABEL_HR,
      NULL,
      BEETLE_OPT_PGXP_2D_TOL_INFO_0_HR,
      NULL,
      "pgxp",
      {
         { "disabled", NULL },
         { "0px", OPTION_VAL_0PX_HR },
         { "1px", OPTION_VAL_1PX_HR },
         { "2px", OPTION_VAL_2PX_HR },
         { "3px", OPTION_VAL_3PX_HR },
         { "4px", OPTION_VAL_4PX_HR },
         { "5px", OPTION_VAL_5PX_HR },
         { "6px", OPTION_VAL_6PX_HR },
         { "7px", OPTION_VAL_7PX_HR },
         { "8px", OPTION_VAL_8PX_HR },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(pgxp_nclip),
      BEETLE_OPT_PGXP_NCLIP_LABEL_HR,
      NULL,
      BEETLE_OPT_PGXP_NCLIP_INFO_0_HR,
      NULL,
      "pgxp",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
#if defined(HAVE_OPENGL) || defined(HAVE_OPENGLES) || defined(HAVE_VULKAN)
   {
      BEETLE_OPT(pgxp_vertex),
      BEETLE_OPT_PGXP_VERTEX_LABEL_HR,
      NULL,
      BEETLE_OPT_PGXP_VERTEX_INFO_0_HR,
      NULL,
      "pgxp",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
#endif
   {
      BEETLE_OPT(pgxp_texture),
      BEETLE_OPT_PGXP_TEXTURE_LABEL_HR,
      NULL,
      BEETLE_OPT_PGXP_TEXTURE_INFO_0_HR,
      NULL,
      "pgxp",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(display_internal_fps),
      BEETLE_OPT_DISPLAY_INTERNAL_FPS_LABEL_HR,
      NULL,
      BEETLE_OPT_DISPLAY_INTERNAL_FPS_INFO_0_HR,
      NULL,
      "osd",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(display_osd),
      BEETLE_OPT_DISPLAY_OSD_LABEL_HR,
      NULL,
      BEETLE_OPT_DISPLAY_OSD_INFO_0_HR,
      NULL,
      "osd",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled"
   },
#if defined(HAVE_OPENGL) || defined(HAVE_OPENGLES) || defined(HAVE_VULKAN)
   {
      BEETLE_OPT(display_vram),
      BEETLE_OPT_DISPLAY_VRAM_LABEL_HR,
      NULL,
      BEETLE_OPT_DISPLAY_VRAM_INFO_0_HR,
      NULL,
      "osd",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
#endif
   {
      BEETLE_OPT(analog_calibration),
      BEETLE_OPT_ANALOG_CALIBRATION_LABEL_HR,
      NULL,
      BEETLE_OPT_ANALOG_CALIBRATION_INFO_0_HR,
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(analog_toggle),
      BEETLE_OPT_ANALOG_TOGGLE_LABEL_HR,
      NULL,
      BEETLE_OPT_ANALOG_TOGGLE_INFO_0_HR,
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { "enabled-analog", OPTION_VAL_ENABLED_ANALOG_HR },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(analog_toggle_combo),
      BEETLE_OPT_ANALOG_TOGGLE_COMBO_LABEL_HR,
      NULL,
      BEETLE_OPT_ANALOG_TOGGLE_COMBO_INFO_0_HR,
      NULL,
      "input",
      {
         { "l1+l2+r1+r2+start+select", OPTION_VAL_L1_L2_R1_R2_START_SELECT_HR },
         { "l1+r1+select",             OPTION_VAL_L1_R1_SELECT_HR },
         { "l1+r1+start",              OPTION_VAL_L1_R1_START_HR },
         { "l1+r1+l3",                 OPTION_VAL_L1_R1_L3_HR },
         { "l1+r1+r3",                 OPTION_VAL_L1_R1_R3_HR },
         { "l2+r2+select",             OPTION_VAL_L2_R2_SELECT_HR },
         { "l2+r2+start",              OPTION_VAL_L2_R2_START_HR },
         { "l2+r2+l3",                 OPTION_VAL_L2_R2_L3_HR },
         { "l2+r2+r3",                 OPTION_VAL_L2_R2_R3_HR },
         { "l3+r3",                    OPTION_VAL_L3_R3_HR },
         { NULL, NULL },
      },
      "l1+r1+select"
   },
   {
      BEETLE_OPT(analog_toggle_hold),
      BEETLE_OPT_ANALOG_TOGGLE_HOLD_LABEL_HR,
      NULL,
      BEETLE_OPT_ANALOG_TOGGLE_HOLD_INFO_0_HR,
      NULL,
      "input",
      {
         { "0", OPTION_VAL_0_HR },
         { "1", OPTION_VAL_1_HR },
         { "2", OPTION_VAL_2_HR },
         { "3", OPTION_VAL_3_HR },
         { "4", OPTION_VAL_4_HR },
         { "5", OPTION_VAL_5_HR },
         { NULL, NULL },
      },
      "1"
   },
   {
      BEETLE_OPT(enable_multitap_port1),
      BEETLE_OPT_ENABLE_MULTITAP_PORT1_LABEL_HR,
      NULL,
      BEETLE_OPT_ENABLE_MULTITAP_PORT1_INFO_0_HR,
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(enable_multitap_port2),
      BEETLE_OPT_ENABLE_MULTITAP_PORT2_LABEL_HR,
      NULL,
      BEETLE_OPT_ENABLE_MULTITAP_PORT2_INFO_0_HR,
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(gun_input_mode),
      BEETLE_OPT_GUN_INPUT_MODE_LABEL_HR,
      NULL,
      BEETLE_OPT_GUN_INPUT_MODE_INFO_0_HR,
      NULL,
      "input",
      {
         { "lightgun",    OPTION_VAL_LIGHTGUN_HR },
         { "touchscreen", OPTION_VAL_TOUCHSCREEN_HR },
         { NULL, NULL },
      },
      "lightgun"
   },
   // Shouldn't the gun_input_mode just be Mouse vs. Touchscreen?
   {
      BEETLE_OPT(gun_cursor),
      BEETLE_OPT_GUN_CURSOR_LABEL_HR,
      NULL,
      BEETLE_OPT_GUN_CURSOR_INFO_0_HR,
      NULL,
      "input",
      {
         { "off",   "disabled" },
         { "cross", OPTION_VAL_CROSS_HR },
         { "dot",   OPTION_VAL_DOT_HR },
         { NULL, NULL },
      },
      "cross"
   },
   {
      BEETLE_OPT(crosshair_color_p1),
      BEETLE_OPT_CROSSHAIR_COLOR_P1_LABEL_HR,
      NULL,
      BEETLE_OPT_CROSSHAIR_COLOR_P1_INFO_0_HR,
      NULL,
      "input",
      {
         { "red",    OPTION_VAL_RED_HR },
         { "blue",   OPTION_VAL_BLUE_HR },
         { "green",  OPTION_VAL_GREEN_HR },
         { "orange", OPTION_VAL_ORANGE_HR },
         { "yellow", OPTION_VAL_YELLOW_HR },
         { "cyan",   OPTION_VAL_CYAN_HR },
         { "pink",   OPTION_VAL_PINK_HR },
         { "purple", OPTION_VAL_PURPLE_HR },
         { "black",  OPTION_VAL_BLACK_HR },
         { "white",  OPTION_VAL_WHITE_HR },
         { NULL, NULL },
      },
      "red"
   },
   {
      BEETLE_OPT(crosshair_color_p2),
      BEETLE_OPT_CROSSHAIR_COLOR_P2_LABEL_HR,
      NULL,
      BEETLE_OPT_CROSSHAIR_COLOR_P2_INFO_0_HR,
      NULL,
      "input",
      {
         { "blue",   OPTION_VAL_BLUE_HR },
         { "red",    OPTION_VAL_RED_HR },
         { "green",  OPTION_VAL_GREEN_HR },
         { "orange", OPTION_VAL_ORANGE_HR },
         { "yellow", OPTION_VAL_YELLOW_HR },
         { "cyan",   OPTION_VAL_CYAN_HR },
         { "pink",   OPTION_VAL_PINK_HR },
         { "purple", OPTION_VAL_PURPLE_HR },
         { "black",  OPTION_VAL_BLACK_HR },
         { "white",  OPTION_VAL_WHITE_HR },
         { NULL, NULL },
      },
      "blue"
   },
   {
      BEETLE_OPT(mouse_sensitivity),
      BEETLE_OPT_MOUSE_SENSITIVITY_LABEL_HR,
      NULL,
      BEETLE_OPT_MOUSE_SENSITIVITY_INFO_0_HR,
      NULL,
      "input",
      {
         { "5%",   OPTION_VAL_5_O42_HR },
         { "10%",  OPTION_VAL_10_HR },
         { "15%",  OPTION_VAL_15_HR },
         { "20%",  OPTION_VAL_20_HR },
         { "25%",  OPTION_VAL_25_HR },
         { "30%",  OPTION_VAL_30_HR },
         { "35%",  OPTION_VAL_35_HR },
         { "40%",  OPTION_VAL_40_HR },
         { "45%",  OPTION_VAL_45_HR },
         { "50%",  OPTION_VAL_50_HR },
         { "55%",  OPTION_VAL_55_HR },
         { "60%",  OPTION_VAL_60_HR },
         { "65%",  OPTION_VAL_65_HR },
         { "70%",  OPTION_VAL_70_HR },
         { "75%",  OPTION_VAL_75_HR },
         { "80%",  OPTION_VAL_80_HR },
         { "85%",  OPTION_VAL_85_HR },
         { "90%",  OPTION_VAL_90_HR },
         { "95%",  OPTION_VAL_95_HR },
         { "100%", OPTION_VAL_100_O42_HR },
         { "105%", OPTION_VAL_105_HR },
         { "110%", OPTION_VAL_110_HR },
         { "115%", OPTION_VAL_115_HR },
         { "120%", OPTION_VAL_120_HR },
         { "125%", OPTION_VAL_125_HR },
         { "130%", OPTION_VAL_130_HR },
         { "135%", OPTION_VAL_135_HR },
         { "140%", OPTION_VAL_140_HR },
         { "145%", OPTION_VAL_145_HR },
         { "150%", OPTION_VAL_150_HR },
         { "155%", OPTION_VAL_155_HR },
         { "160%", OPTION_VAL_160_HR },
         { "165%", OPTION_VAL_165_HR },
         { "170%", OPTION_VAL_170_HR },
         { "175%", OPTION_VAL_175_HR },
         { "180%", OPTION_VAL_180_HR },
         { "185%", OPTION_VAL_185_HR },
         { "190%", OPTION_VAL_190_HR },
         { "195%", OPTION_VAL_195_HR },
         { "200%", OPTION_VAL_200_HR },
         { NULL, NULL },
      },
      "100%"
   },
   {
      BEETLE_OPT(negcon_response),
      BEETLE_OPT_NEGCON_RESPONSE_LABEL_HR,
      NULL,
      BEETLE_OPT_NEGCON_RESPONSE_INFO_0_HR,
      NULL,
      "input",
      {
         { "linear",    OPTION_VAL_LINEAR_HR },
         { "quadratic", OPTION_VAL_QUADRATIC_HR },
         { "cubic",     OPTION_VAL_CUBIC_HR },
         { NULL, NULL },
      },
      "linear"
   },
   {
      BEETLE_OPT(negcon_deadzone),
      BEETLE_OPT_NEGCON_DEADZONE_LABEL_HR,
      NULL,
      BEETLE_OPT_NEGCON_DEADZONE_INFO_0_HR,
      NULL,
      "input",
      {
         { "0%",  OPTION_VAL_0_O44_HR },
         { "5%",  OPTION_VAL_5_O42_HR },
         { "10%", OPTION_VAL_10_HR },
         { "15%", OPTION_VAL_15_HR },
         { "20%", OPTION_VAL_20_HR },
         { "25%", OPTION_VAL_25_HR },
         { "30%", OPTION_VAL_30_HR },
         { NULL, NULL },
      },
      "0%"
   },
   {
      BEETLE_OPT(use_mednafen_memcard0_method),
      BEETLE_OPT_USE_MEDNAFEN_MEMCARD0_METHOD_LABEL_HR,
      NULL,
      BEETLE_OPT_USE_MEDNAFEN_MEMCARD0_METHOD_INFO_0_HR,
      NULL,
      "memcards",
      {
         { "libretro", OPTION_VAL_LIBRETRO_HR },
         { "mednafen", OPTION_VAL_MEDNAFEN_HR },
         { NULL, NULL },
      },
      "libretro"
   },
   {
      BEETLE_OPT(enable_memcard1),
      BEETLE_OPT_ENABLE_MEMCARD1_LABEL_HR,
      NULL,
      BEETLE_OPT_ENABLE_MEMCARD1_INFO_0_HR,
      NULL,
      "memcards",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(shared_memory_cards),
      BEETLE_OPT_SHARED_MEMORY_CARDS_LABEL_HR,
      NULL,
      BEETLE_OPT_SHARED_MEMORY_CARDS_INFO_0_HR,
      NULL,
      "memcards",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(memcard_left_index),
      BEETLE_OPT_MEMCARD_LEFT_INDEX_LABEL_HR,
      NULL,
      BEETLE_OPT_MEMCARD_LEFT_INDEX_INFO_0_HR,
      NULL,
      "memcards",
      {
         { "0",  OPTION_VAL_0_O48_HR },
         { "1",  NULL },
         { "2",  NULL },
         { "3",  NULL },
         { "4",  NULL },
         { "5",  NULL },
         { "6",  NULL },
         { "7",  NULL },
         { "8",  NULL },
         { "9",  NULL },
         { "10", NULL },
         { "11", NULL },
         { "12", NULL },
         { "13", NULL },
         { "14", NULL },
         { "15", NULL },
         { "16", NULL },
         { "17", NULL },
         { "18", NULL },
         { "19", NULL },
         { "20", NULL },
         { "21", NULL },
         { "22", NULL },
         { "23", NULL },
         { "24", NULL },
         { "25", NULL },
         { "26", NULL },
         { "27", NULL },
         { "28", NULL },
         { "29", NULL },
         { "30", NULL },
         { "31", NULL },
         { "32", NULL },
         { "33", NULL },
         { "34", NULL },
         { "35", NULL },
         { "36", NULL },
         { "37", NULL },
         { "38", NULL },
         { "39", NULL },
         { "40", NULL },
         { "41", NULL },
         { "42", NULL },
         { "43", NULL },
         { "44", NULL },
         { "45", NULL },
         { "46", NULL },
         { "47", NULL },
         { "48", NULL },
         { "49", NULL },
         { "50", NULL },
         { "51", NULL },
         { "52", NULL },
         { "53", NULL },
         { "54", NULL },
         { "55", NULL },
         { "56", NULL },
         { "57", NULL },
         { "58", NULL },
         { "59", NULL },
         { "60", NULL },
         { "61", NULL },
         { "62", NULL },
         { "63", NULL },
         { NULL, NULL },
      },
      "0"
   },
   {
      BEETLE_OPT(memcard_right_index),
      BEETLE_OPT_MEMCARD_RIGHT_INDEX_LABEL_HR,
      NULL,
      BEETLE_OPT_MEMCARD_RIGHT_INDEX_INFO_0_HR,
      NULL,
      "memcards",
      {
         { "0",  NULL },
         { "1",  OPTION_VAL_1_O49_HR },
         { "2",  NULL },
         { "3",  NULL },
         { "4",  NULL },
         { "5",  NULL },
         { "6",  NULL },
         { "7",  NULL },
         { "8",  NULL },
         { "9",  NULL },
         { "10", NULL },
         { "11", NULL },
         { "12", NULL },
         { "13", NULL },
         { "14", NULL },
         { "15", NULL },
         { "16", NULL },
         { "17", NULL },
         { "18", NULL },
         { "19", NULL },
         { "20", NULL },
         { "21", NULL },
         { "22", NULL },
         { "23", NULL },
         { "24", NULL },
         { "25", NULL },
         { "26", NULL },
         { "27", NULL },
         { "28", NULL },
         { "29", NULL },
         { "30", NULL },
         { "31", NULL },
         { "32", NULL },
         { "33", NULL },
         { "34", NULL },
         { "35", NULL },
         { "36", NULL },
         { "37", NULL },
         { "38", NULL },
         { "39", NULL },
         { "40", NULL },
         { "41", NULL },
         { "42", NULL },
         { "43", NULL },
         { "44", NULL },
         { "45", NULL },
         { "46", NULL },
         { "47", NULL },
         { "48", NULL },
         { "49", NULL },
         { "50", NULL },
         { "51", NULL },
         { "52", NULL },
         { "53", NULL },
         { "54", NULL },
         { "55", NULL },
         { "56", NULL },
         { "57", NULL },
         { "58", NULL },
         { "59", NULL },
         { "60", NULL },
         { "61", NULL },
         { "62", NULL },
         { "63", NULL },
         { NULL, NULL },
      },
      "1"
   },
   {
      BEETLE_OPT(line_render),
      BEETLE_OPT_LINE_RENDER_LABEL_HR,
      NULL,
      BEETLE_OPT_LINE_RENDER_INFO_0_HR,
      NULL,
      "hacks",
      {
         { "default",    OPTION_VAL_DEFAULT_HR },
         { "aggressive", OPTION_VAL_AGGRESSIVE_HR },
         { "disabled",   NULL },
         { NULL, NULL },
      },
      "default"
   },
   {
      BEETLE_OPT(widescreen_hack),
      BEETLE_OPT_WIDESCREEN_HACK_LABEL_HR,
      NULL,
      BEETLE_OPT_WIDESCREEN_HACK_INFO_0_HR,
      NULL,
      "hacks",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(widescreen_hack_aspect_ratio),
      BEETLE_OPT_WIDESCREEN_HACK_ASPECT_RATIO_LABEL_HR,
      NULL,
      BEETLE_OPT_WIDESCREEN_HACK_ASPECT_RATIO_INFO_0_HR,
      NULL,
      "hacks",
      {
         { "16:9",  OPTION_VAL_16_9_HR },
         { "16:10", OPTION_VAL_16_10_HR },
         { "18:9",  OPTION_VAL_18_9_HR },
         { "19:9",  OPTION_VAL_19_9_HR },
         { "20:9",  OPTION_VAL_20_9_HR },
         { "21:9",  OPTION_VAL_21_9_HR }, // 64:27
         { "32:9",  OPTION_VAL_32_9_HR },
         { NULL,    NULL },
      },
      "16:9"
   },
#if defined(HAVE_LIGHTREC)
   {
      BEETLE_OPT(cpu_dynarec),
      BEETLE_OPT_CPU_DYNAREC_LABEL_HR,
      NULL,
      BEETLE_OPT_CPU_DYNAREC_INFO_0_HR,
      NULL,
      "hacks",
      {
         { "disabled",        OPTION_VAL_DISABLED_HR },
         { "execute",         OPTION_VAL_EXECUTE_HR },
         { "execute_one",     OPTION_VAL_EXECUTE_ONE_HR },
         { "run_interpreter", OPTION_VAL_RUN_INTERPRETER_HR },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(dynarec_invalidate),
      BEETLE_OPT_DYNAREC_INVALIDATE_LABEL_HR,
      NULL,
      BEETLE_OPT_DYNAREC_INVALIDATE_INFO_0_HR,
      NULL,
      "hacks",
      {
         { "full", OPTION_VAL_FULL_HR },
         { "dma",  OPTION_VAL_DMA_HR },
         { NULL, NULL },
      },
      "full"
   },
   {
      BEETLE_OPT(dynarec_eventcycles),
      BEETLE_OPT_DYNAREC_EVENTCYCLES_LABEL_HR,
      NULL,
      BEETLE_OPT_DYNAREC_EVENTCYCLES_INFO_0_HR,
      NULL,
      "hacks",
      {
         { "128",  OPTION_VAL_128_HR },
         { "256",  NULL },
         { "384",  NULL },
         { "512",  NULL },
         { "640",  NULL },
         { "768",  NULL },
         { "896",  NULL },
         { "1024", NULL },
         { "1152", NULL },
         { "1280", NULL },
         { "1408", NULL },
         { "1536", NULL },
         { "1664", NULL },
         { "1792", NULL },
         { "1920", NULL },
         { "2048", NULL },
         { NULL, NULL },
      },
      "128"
   },
   {
      BEETLE_OPT(dynarec_spu_samples),
      BEETLE_OPT_DYNAREC_SPU_SAMPLES_LABEL_HR,
      NULL,
      BEETLE_OPT_DYNAREC_SPU_SAMPLES_INFO_0_HR,
      NULL,
      "hacks",
      {
         { "1",  OPTION_VAL_1_O49_HR },
         { "4",  NULL },
         { "16", NULL },
         { NULL, NULL },
      },
      "1"
   },
#endif
   {
      BEETLE_OPT(core_timing_fps),
      BEETLE_OPT_CORE_TIMING_FPS_LABEL_HR,
      NULL,
      BEETLE_OPT_CORE_TIMING_FPS_INFO_0_HR,
      NULL,
      "video",
      {
         { "auto_toggle",       OPTION_VAL_AUTO_TOGGLE_HR },
         { "force_progressive", OPTION_VAL_FORCE_PROGRESSIVE_HR },
         { "force_interlaced",  OPTION_VAL_FORCE_INTERLACED_HR },
      },
      "auto_toggle"
   },
   {
      BEETLE_OPT(deinterlacer),
      BEETLE_OPT_DEINTERLACER_LABEL_HR,
      NULL,
      BEETLE_OPT_DEINTERLACER_INFO_0_HR,
      NULL,
      "video",
      {
         { "weave",      OPTION_VAL_WEAVE_HR },
         { "bob",        OPTION_VAL_BOB_HR },
         { "bob_offset", OPTION_VAL_BOB_OFFSET_HR },
         { "fastmad",    OPTION_VAL_FASTMAD_HR },
         { "off",        "Off" },
         { NULL, NULL },
      },
      "weave"
   },
   {
      BEETLE_OPT(frame_duping),
      BEETLE_OPT_FRAME_DUPING_LABEL_HR,
      NULL,
      BEETLE_OPT_FRAME_DUPING_INFO_0_HR,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(pal_video_timing_override),
      BEETLE_OPT_PAL_VIDEO_TIMING_OVERRIDE_LABEL_HR,
      NULL,
      BEETLE_OPT_PAL_VIDEO_TIMING_OVERRIDE_INFO_0_HR,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(aspect_ratio),
      BEETLE_OPT_ASPECT_RATIO_LABEL_HR,
      NULL,
      BEETLE_OPT_ASPECT_RATIO_INFO_0_HR,
      NULL,
      "video",
      {
         { "corrected", OPTION_VAL_CORRECTED_HR },
         { "uncorrected", OPTION_VAL_UNCORRECTED_HR },
         { "4:3",  OPTION_VAL_4_3_HR },
         { "ntsc", OPTION_VAL_NTSC_HR },
      },
      "corrected"
   },
   {
      BEETLE_OPT(crop_overscan),
      BEETLE_OPT_CROP_OVERSCAN_LABEL_HR,
      NULL,
      BEETLE_OPT_CROP_OVERSCAN_INFO_0_HR,
      NULL,
      "video",
      {
         { "disabled", "Disabled" },
         { "static",   OPTION_VAL_STATIC_HR },
         { "smart",    OPTION_VAL_SMART_HR },
         { NULL, NULL },
      },
      "smart"
   },
   {
      BEETLE_OPT(image_crop),
      BEETLE_OPT_IMAGE_CROP_LABEL_HR,
      NULL,
      BEETLE_OPT_IMAGE_CROP_INFO_0_HR,
      NULL,
      "video",
      {
         { "disabled", "0" },
         { "1px",      OPTION_VAL_1PX_HR },
         { "2px",      OPTION_VAL_2PX_HR },
         { "3px",      OPTION_VAL_3PX_HR },
         { "4px",      OPTION_VAL_4PX_HR },
         { "5px",      OPTION_VAL_5PX_HR },
         { "6px",      OPTION_VAL_6PX_HR },
         { "7px",      OPTION_VAL_7PX_HR },
         { "8px",      OPTION_VAL_8PX_HR },
         { "9px",      OPTION_VAL_9PX_HR },
         { "10px",     OPTION_VAL_10PX_HR },
         { "11px",     OPTION_VAL_11PX_HR },
         { "12px",     OPTION_VAL_12PX_HR },
         { "13px",     OPTION_VAL_13PX_HR },
         { "14px",     OPTION_VAL_14PX_HR },
         { "15px",     OPTION_VAL_15PX_HR },
         { "16px",     OPTION_VAL_16PX_HR },
         { "17px",     OPTION_VAL_17PX_HR },
         { "18px",     OPTION_VAL_18PX_HR },
         { "19px",     OPTION_VAL_19PX_HR },
         { "20px",     OPTION_VAL_20PX_HR },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(image_offset),
      BEETLE_OPT_IMAGE_OFFSET_LABEL_HR,
      NULL,
      BEETLE_OPT_IMAGE_OFFSET_INFO_0_HR,
      NULL,
      "video",
      {
         { "-12px",    OPTION_VAL_12PX_O64_HR },
         { "-11px",    OPTION_VAL_11PX_O64_HR },
         { "-10px",    OPTION_VAL_10PX_O64_HR },
         { "-9px",     OPTION_VAL_9PX_O64_HR },
         { "-8px",     OPTION_VAL_8PX_O64_HR },
         { "-7px",     OPTION_VAL_7PX_O64_HR },
         { "-6px",     OPTION_VAL_6PX_O64_HR },
         { "-5px",     OPTION_VAL_5PX_O64_HR },
         { "-4px",     OPTION_VAL_4PX_O64_HR },
         { "-3px",     OPTION_VAL_3PX_O64_HR },
         { "-2px",     OPTION_VAL_2PX_O64_HR },
         { "-1px",     OPTION_VAL_1PX_O64_HR },
         { "disabled", "0" },
         { "+1px",     OPTION_VAL_1PX_O64_0_HR },
         { "+2px",     OPTION_VAL_2PX_O64_0_HR },
         { "+3px",     OPTION_VAL_3PX_O64_0_HR },
         { "+4px",     OPTION_VAL_4PX_O64_0_HR },
         { "+5px",     OPTION_VAL_5PX_O64_0_HR },
         { "+6px",     OPTION_VAL_6PX_O64_0_HR },
         { "+7px",     OPTION_VAL_7PX_O64_0_HR },
         { "+8px",     OPTION_VAL_8PX_O64_0_HR },
         { "+9px",     OPTION_VAL_9PX_O64_0_HR },
         { "+10px",    OPTION_VAL_10PX_O64_0_HR },
         { "+11px",    OPTION_VAL_11PX_O64_0_HR },
         { "+12px",    OPTION_VAL_12PX_O64_0_HR },
         { NULL, NULL },
      },
      "disabled"
   },
#if defined(HAVE_OPENGL) || defined(HAVE_OPENGLES) || defined(HAVE_VULKAN)
   {
      BEETLE_OPT(image_offset_cycles),
      BEETLE_OPT_IMAGE_OFFSET_CYCLES_LABEL_HR,
      NULL,
      BEETLE_OPT_IMAGE_OFFSET_CYCLES_INFO_0_HR,
      NULL,
      "video",
      {
         { "-40",      NULL },
         { "-39",      NULL },
         { "-38",      NULL },
         { "-37",      NULL },
         { "-36",      NULL },
         { "-35",      NULL },
         { "-34",      NULL },
         { "-33",      NULL },
         { "-32",      NULL },
         { "-31",      NULL },
         { "-30",      NULL },
         { "-29",      NULL },
         { "-28",      NULL },
         { "-27",      NULL },
         { "-26",      NULL },
         { "-25",      NULL },
         { "-24",      NULL },
         { "-23",      NULL },
         { "-22",      NULL },
         { "-21",      NULL },
         { "-20",      NULL },
         { "-19",      NULL },
         { "-18",      NULL },
         { "-17",      NULL },
         { "-16",      NULL },
         { "-15",      NULL },
         { "-14",      NULL },
         { "-13",      NULL },
         { "-12",      NULL },
         { "-11",      NULL },
         { "-10",      NULL },
         { "-9",       NULL },
         { "-8",       NULL },
         { "-7",       NULL },
         { "-6",       NULL },
         { "-5",       NULL },
         { "-4",       NULL },
         { "-3",       NULL },
         { "-2",       NULL },
         { "-1",       NULL },
         { "0",        "0" },
         { "+1",       NULL },
         { "+2",       NULL },
         { "+3",       NULL },
         { "+4",       NULL },
         { "+5",       NULL },
         { "+6",       NULL },
         { "+7",       NULL },
         { "+8",       NULL },
         { "+9",       NULL },
         { "+10",      NULL },
         { "+11",      NULL },
         { "+12",      NULL },
         { "+13",      NULL },
         { "+14",      NULL },
         { "+15",      NULL },
         { "+16",      NULL },
         { "+17",      NULL },
         { "+18",      NULL },
         { "+19",      NULL },
         { "+20",      NULL },
         { "+21",      NULL },
         { "+22",      NULL },
         { "+23",      NULL },
         { "+24",      NULL },
         { "+25",      NULL },
         { "+26",      NULL },
         { "+27",      NULL },
         { "+28",      NULL },
         { "+29",      NULL },
         { "+30",      NULL },
         { "+31",      NULL },
         { "+32",      NULL },
         { "+33",      NULL },
         { "+34",      NULL },
         { "+35",      NULL },
         { "+36",      NULL },
         { "+37",      NULL },
         { "+38",      NULL },
         { "+39",      NULL },
         { "+40",      NULL },
         { NULL, NULL},
      },
      "0"
   },
#endif
   {
      BEETLE_OPT(initial_scanline),
      BEETLE_OPT_INITIAL_SCANLINE_LABEL_HR,
      NULL,
      BEETLE_OPT_INITIAL_SCANLINE_INFO_0_HR,
      NULL,
      "video",
      {
         { "0",  OPTION_VAL_0_O48_HR },
         { "1",  NULL },
         { "2",  NULL },
         { "3",  NULL },
         { "4",  NULL },
         { "5",  NULL },
         { "6",  NULL },
         { "7",  NULL },
         { "8",  NULL },
         { "9",  NULL },
         { "10", NULL },
         { "11", NULL },
         { "12", NULL },
         { "13", NULL },
         { "14", NULL },
         { "15", NULL },
         { "16", NULL },
         { "17", NULL },
         { "18", NULL },
         { "19", NULL },
         { "20", NULL },
         { "21", NULL },
         { "22", NULL },
         { "23", NULL },
         { "24", NULL },
         { "25", NULL },
         { "26", NULL },
         { "27", NULL },
         { "28", NULL },
         { "29", NULL },
         { "30", NULL },
         { "31", NULL },
         { "32", NULL },
         { "33", NULL },
         { "34", NULL },
         { "35", NULL },
         { "36", NULL },
         { "37", NULL },
         { "38", NULL },
         { "39", NULL },
         { "40", NULL },
         { NULL, NULL },
      },
      "0"
   },
   {
      BEETLE_OPT(last_scanline),
      BEETLE_OPT_LAST_SCANLINE_LABEL_HR,
      NULL,
      BEETLE_OPT_LAST_SCANLINE_INFO_0_HR,
      NULL,
      "video",
      {
         { "210", NULL },
         { "211", NULL },
         { "212", NULL },
         { "213", NULL },
         { "214", NULL },
         { "215", NULL },
         { "216", NULL },
         { "217", NULL },
         { "218", NULL },
         { "219", NULL },
         { "220", NULL },
         { "221", NULL },
         { "222", NULL },
         { "223", NULL },
         { "224", NULL },
         { "225", NULL },
         { "226", NULL },
         { "227", NULL },
         { "228", NULL },
         { "229", NULL },
         { "230", NULL },
         { "231", NULL },
         { "232", NULL },
         { "233", NULL },
         { "234", NULL },
         { "235", NULL },
         { "236", NULL },
         { "237", NULL },
         { "238", NULL },
         { "239", OPTION_VAL_239_HR },
         { NULL, NULL },
      },
      "239"
   },
   {
      BEETLE_OPT(initial_scanline_pal),
      BEETLE_OPT_INITIAL_SCANLINE_PAL_LABEL_HR,
      NULL,
      BEETLE_OPT_INITIAL_SCANLINE_PAL_INFO_0_HR,
      NULL,
      "video",
      {
         { "0",  OPTION_VAL_0_O48_HR },
         { "1",  NULL },
         { "2",  NULL },
         { "3",  NULL },
         { "4",  NULL },
         { "5",  NULL },
         { "6",  NULL },
         { "7",  NULL },
         { "8",  NULL },
         { "9",  NULL },
         { "10", NULL },
         { "11", NULL },
         { "12", NULL },
         { "13", NULL },
         { "14", NULL },
         { "15", NULL },
         { "16", NULL },
         { "17", NULL },
         { "18", NULL },
         { "19", NULL },
         { "20", NULL },
         { "21", NULL },
         { "22", NULL },
         { "23", NULL },
         { "24", NULL },
         { "25", NULL },
         { "26", NULL },
         { "27", NULL },
         { "28", NULL },
         { "29", NULL },
         { "30", NULL },
         { "31", NULL },
         { "32", NULL },
         { "33", NULL },
         { "34", NULL },
         { "35", NULL },
         { "36", NULL },
         { "37", NULL },
         { "38", NULL },
         { "39", NULL },
         { "40", NULL },
         { NULL, NULL },
      },
      "0"
   },
   {
      BEETLE_OPT(last_scanline_pal),
      BEETLE_OPT_LAST_SCANLINE_PAL_LABEL_HR,
      NULL,
      BEETLE_OPT_LAST_SCANLINE_PAL_INFO_0_HR,
      NULL,
      "video",
      {
         { "230", NULL },
         { "231", NULL },
         { "232", NULL },
         { "233", NULL },
         { "234", NULL },
         { "235", NULL },
         { "236", NULL },
         { "237", NULL },
         { "238", NULL },
         { "239", NULL },
         { "240", NULL },
         { "241", NULL },
         { "242", NULL },
         { "243", NULL },
         { "244", NULL },
         { "245", NULL },
         { "246", NULL },
         { "247", NULL },
         { "248", NULL },
         { "249", NULL },
         { "250", NULL },
         { "251", NULL },
         { "252", NULL },
         { "253", NULL },
         { "254", NULL },
         { "255", NULL },
         { "256", NULL },
         { "257", NULL },
         { "258", NULL },
         { "259", NULL },
         { "260", NULL },
         { "261", NULL },
         { "262", NULL },
         { "263", NULL },
         { "264", NULL },
         { "265", NULL },
         { "266", NULL },
         { "267", NULL },
         { "268", NULL },
         { "269", NULL },
         { "270", NULL },
         { "271", NULL },
         { "272", NULL },
         { "273", NULL },
         { "274", NULL },
         { "275", NULL },
         { "276", NULL },
         { "277", NULL },
         { "278", NULL },
         { "279", NULL },
         { "280", NULL },
         { "281", NULL },
         { "282", NULL },
         { "283", NULL },
         { "284", NULL },
         { "285", NULL },
         { "286", NULL },
         { "287", OPTION_VAL_287_HR },
         { NULL, NULL },
      },
      "287"
   },
   { NULL, NULL, NULL, NULL, NULL, NULL, {{0}}, NULL },
};
struct retro_core_options_v2 options_hr = {
   option_cats_hr,
   option_defs_hr
};

/* RETRO_LANGUAGE_HU */

#define CATEGORY_SYSTEM_LABEL_HU "Rendszer"
#define CATEGORY_SYSTEM_INFO_0_HU NULL
#define CATEGORY_VIDEO_LABEL_HU "Kép"
#define CATEGORY_VIDEO_INFO_0_HU "A képarány, képvágás, videoszűrő és képkocka kihagyás beállításai."
#define CATEGORY_PGXP_LABEL_HU "PGXP (precíziós geometria feldolgozó)"
#define CATEGORY_PGXP_INFO_0_HU "Az eredeti konzolhoz képest grafikai javulást hozó fejlesztések. A PGXP eltüntetheti a görbült textúrákat és a Z-ütközést."
#define CATEGORY_OSD_LABEL_HU "Képernyőn megjelenő elemek (OSD)"
#define CATEGORY_OSD_INFO_0_HU "A képernyőn megjelenő értesítések beállítása."
#define CATEGORY_INPUT_LABEL_HU "Bevitel"
#define CATEGORY_INPUT_INFO_0_HU "A fénypisztoly, az egér és a neGcon beállításai."
#define CATEGORY_MEMCARDS_LABEL_HU "Memóriakártya"
#define CATEGORY_MEMCARDS_INFO_0_HU NULL
#define CATEGORY_HACKS_LABEL_HU "Emulációs trükkök"
#define CATEGORY_HACKS_INFO_0_HU NULL
#define BEETLE_OPT_CPU_FREQ_SCALE_LABEL_HU "CPU frekvencia skálázás (túlhajtás)"
#define BEETLE_OPT_CPU_FREQ_SCALE_INFO_0_HU "Az emulált PSX CPU túlhajtása (vagy alulhajtása). A túlhajtás megszüntetheti a lelassulást és bizonyos játékoknál javíthatja a képsebességet, nagyobb teljesítményigény árán. Figyelem, néhány játékban beépített képsebesség korlátozás van, ezeken nem segít a túlhajtás. Túlhajtásnál bizonyos effektek néhány játékban a kívántnál gyorsabban mozoghatnak."
#define OPTION_VAL_50_HU NULL
#define OPTION_VAL_60_HU NULL
#define OPTION_VAL_70_HU NULL
#define OPTION_VAL_80_HU NULL
#define OPTION_VAL_90_HU NULL
#define OPTION_VAL_100_HU NULL
#define OPTION_VAL_110_HU NULL
#define OPTION_VAL_120_HU NULL
#define OPTION_VAL_130_HU NULL
#define OPTION_VAL_140_HU NULL
#define OPTION_VAL_150_HU NULL
#define OPTION_VAL_160_HU NULL
#define OPTION_VAL_170_HU NULL
#define OPTION_VAL_180_HU NULL
#define OPTION_VAL_190_HU NULL
#define OPTION_VAL_200_HU NULL
#define OPTION_VAL_210_HU NULL
#define OPTION_VAL_220_HU NULL
#define OPTION_VAL_230_HU NULL
#define OPTION_VAL_240_HU NULL
#define OPTION_VAL_250_HU NULL
#define OPTION_VAL_260_HU NULL
#define OPTION_VAL_270_HU NULL
#define OPTION_VAL_280_HU NULL
#define OPTION_VAL_290_HU NULL
#define OPTION_VAL_300_HU NULL
#define OPTION_VAL_310_HU NULL
#define OPTION_VAL_320_HU NULL
#define OPTION_VAL_330_HU NULL
#define OPTION_VAL_340_HU NULL
#define OPTION_VAL_350_HU NULL
#define OPTION_VAL_360_HU NULL
#define OPTION_VAL_370_HU NULL
#define OPTION_VAL_380_HU NULL
#define OPTION_VAL_390_HU NULL
#define OPTION_VAL_400_HU NULL
#define OPTION_VAL_410_HU NULL
#define OPTION_VAL_420_HU NULL
#define OPTION_VAL_430_HU NULL
#define OPTION_VAL_440_HU NULL
#define OPTION_VAL_450_HU NULL
#define OPTION_VAL_460_HU NULL
#define OPTION_VAL_470_HU NULL
#define OPTION_VAL_480_HU NULL
#define OPTION_VAL_490_HU NULL
#define OPTION_VAL_500_HU NULL
#define OPTION_VAL_510_HU NULL
#define OPTION_VAL_520_HU NULL
#define OPTION_VAL_530_HU NULL
#define OPTION_VAL_540_HU NULL
#define OPTION_VAL_550_HU NULL
#define OPTION_VAL_560_HU NULL
#define OPTION_VAL_570_HU NULL
#define OPTION_VAL_580_HU NULL
#define OPTION_VAL_590_HU NULL
#define OPTION_VAL_600_HU NULL
#define OPTION_VAL_610_HU NULL
#define OPTION_VAL_620_HU NULL
#define OPTION_VAL_630_HU NULL
#define OPTION_VAL_640_HU NULL
#define OPTION_VAL_650_HU NULL
#define OPTION_VAL_660_HU NULL
#define OPTION_VAL_670_HU NULL
#define OPTION_VAL_680_HU NULL
#define OPTION_VAL_690_HU NULL
#define OPTION_VAL_700_HU NULL
#define OPTION_VAL_710_HU NULL
#define OPTION_VAL_720_HU NULL
#define OPTION_VAL_730_HU NULL
#define OPTION_VAL_740_HU NULL
#define OPTION_VAL_750_HU NULL
#define BEETLE_OPT_GPU_OVERCLOCK_LABEL_HU "GPU rasterizer túlhajtás"
#define BEETLE_OPT_GPU_OVERCLOCK_INFO_0_HU "Az emulált PSX GPU-jához tartozó 2D rasterizer túlhajtása. A 3D megjelenítést nem javítja, és általánosságban kevés hatása van."
#define OPTION_VAL_1X_NATIVE_HU "1x (eredeti)"
#define OPTION_VAL_2X_HU NULL
#define OPTION_VAL_4X_HU NULL
#define OPTION_VAL_8X_HU NULL
#define OPTION_VAL_16X_HU NULL
#define OPTION_VAL_32X_HU NULL
#define BEETLE_OPT_GTE_OVERCLOCK_LABEL_HU "GTE túlhajtás"
#define BEETLE_OPT_GTE_OVERCLOCK_INFO_0_HU "Az emulált GTE (a CPU 3D coprocesszora) minden művelete fix késleltetés, egy órajelciklus alatt végbemegy. Ha a játék intenzíven használja a GTE-t, ez nagyban segít a képkockasebességen és a képkocka idejének stabilitásán."
#define BEETLE_OPT_CD_ACCESS_METHOD_LABEL_HU NULL
#define BEETLE_OPT_CD_ACCESS_METHOD_INFO_0_HU NULL
#define OPTION_VAL_SYNC_HU "Szinkron"
#define OPTION_VAL_ASYNC_HU "Aszinkron"
#define OPTION_VAL_PRECACHE_HU "Előre beolvasott"
#define BEETLE_OPT_CD_FASTLOAD_LABEL_HU "CD betöltés sebessége"
#define BEETLE_OPT_CD_FASTLOAD_INFO_0_HU NULL
#define OPTION_VAL_2X_NATIVE_HU "2x (eredeti)"
#define OPTION_VAL_6X_HU NULL
#define OPTION_VAL_10X_HU NULL
#define OPTION_VAL_12X_HU NULL
#define OPTION_VAL_14X_HU NULL
#define BEETLE_OPT_SKIP_BIOS_LABEL_HU "BIOS átugrása"
#define BEETLE_OPT_SKIP_BIOS_INFO_0_HU "A PlayStation BIOS indító animáció kihagyása, ami normál esetben megjelenik tartalom betöltésekor. Figyelem: kompatibilitási gondokat okozhat számos játéknál (PAL másolásvédett játékok, Saga Frontier, stb.)."
#define BEETLE_OPT_OVERRIDE_BIOS_LABEL_HU NULL
#define BEETLE_OPT_OVERRIDE_BIOS_INFO_0_HU NULL
#define OPTION_VAL_PSXONPSP_HU NULL
#define OPTION_VAL_PS1_ROM_HU NULL
#define OPTION_VAL_OPENBIOS_HU NULL
#define BEETLE_OPT_REGION_LABEL_HU "Régió"
#define BEETLE_OPT_REGION_INFO_0_HU NULL
#define OPTION_VAL_AUTO_HU "Automatikus"
#define OPTION_VAL_NTSC_J_HU NULL
#define OPTION_VAL_NTSC_U_HU NULL
#define OPTION_VAL_PAL_HU NULL
#define BEETLE_OPT_INTERNAL_RESOLUTION_LABEL_HU "A GPU belső felbontása"
#define BEETLE_OPT_INTERNAL_RESOLUTION_INFO_0_HU NULL
#define BEETLE_OPT_RENDERER_LABEL_HU NULL
#define BEETLE_OPT_RENDERER_INFO_0_HU NULL
#define OPTION_VAL_HARDWARE_HU "Hardveres (automatikus)"
#define OPTION_VAL_HARDWARE_GL_HU "Hardveres (OpenGL)"
#define OPTION_VAL_HARDWARE_VK_HU "Hardveres (Vulkan)"
#define OPTION_VAL_SOFTWARE_HU "Szoftveres"
#define BEETLE_OPT_RENDERER_SOFTWARE_FB_LABEL_HU "Szoftveres framebuffer"
#define BEETLE_OPT_RENDERER_SOFTWARE_FB_INFO_0_HU "A framebuffer effektek (mozgási elmosódás, FF7 csata örvény) pontos emulálása hardveres megjelenítőknél is, egy normál felbontású szoftveres megjelenítő futtatásával a háttérben. Kikapcsolva ezek a műveletek elmaradnak (OpenGL) vagy a GPU számolja őket (Vulkan). A kikapcsolás javíthatja a teljesítményt, de súlyos grafikus hibákat okozhat. Kétség esetén maradjon bekapcsolva."
#define BEETLE_OPT_DEPTH_LABEL_HU "Belső színmélység"
#define BEETLE_OPT_DEPTH_INFO_0_HU NULL
#define OPTION_VAL_16BPP_NATIVE_HU "16 bpp (eredeti)"
#define OPTION_VAL_32BPP_HU NULL
#define BEETLE_OPT_DITHER_MODE_LABEL_HU "Dithering minta"
#define BEETLE_OPT_DITHER_MODE_INFO_0_HU NULL
#define OPTION_VAL_INTERNAL_RESOLUTION_HU "Belső felbontás"
#define BEETLE_OPT_SCALED_UV_OFFSET_LABEL_HU "Textúrák UV eltolása"
#define BEETLE_OPT_SCALED_UV_OFFSET_INFO_0_HU "A 3D sokszögek textúráinak mintavétele eltolással az egyszeresnél nagyobb belső felbontásokhoz. Csökkenti a textúraillesztési hibákat, de okozhat grafikus problémákat."
#define BEETLE_OPT_FILTER_LABEL_HU "Textúraszűrés"
#define BEETLE_OPT_FILTER_INFO_0_HU NULL
#define OPTION_VAL_NEAREST_HU "Legközelebbi"
#define OPTION_VAL_SABR_HU NULL
#define OPTION_VAL_XBR_HU NULL
#define OPTION_VAL_BILINEAR_HU "Bilineáris"
#define OPTION_VAL_3_POINT_HU "Hárompontos"
#define OPTION_VAL_JINC2_HU NULL
#define BEETLE_OPT_FILTER_EXCLUDE_SPRITE_LABEL_HU "Sprite-ok kihagyása a szűrésből"
#define BEETLE_OPT_FILTER_EXCLUDE_SPRITE_INFO_0_HU "A textúraszűrés nem lesz érvényes a sprite-okra. Megakadályozza az illesztési hibákat 2D sprite alapú hátterekkel rendelkező játékoknál. Az adaptív élsimítással vagy más utófeldolgozó szűrővel adja a legjobb eredményt."
#define OPTION_VAL_OPAQUE_HU "Csak az átlátszatlanokat"
#define OPTION_VAL_ALL_HU "Az átlátszatlanokat és a félig átlátszóakat"
#define BEETLE_OPT_FILTER_EXCLUDE_2D_POLYGON_LABEL_HU "2D sokszögek kihagyása a szűrésből"
#define BEETLE_OPT_FILTER_EXCLUDE_2D_POLYGON_INFO_0_HU "A textúraszűrés nem lesz érvényes a 2D sokszögekre. A 2D sokszögeket egy heurisztika érzékeli és lehetnek hibák. Az adaptív élsimítással vagy más utófeldolgozó szűrővel adja a legjobb eredményt."
#define BEETLE_OPT_ADAPTIVE_SMOOTHING_LABEL_HU "Adaptív élsimítás"
#define BEETLE_OPT_ADAPTIVE_SMOOTHING_INFO_0_HU "A 2D rajzok és és a felhasználói felület elsimítása, a 3D objektumok elhomályosítása nélkül. Csak a Vulkan meghajtó támogatja."
#define BEETLE_OPT_SUPER_SAMPLING_LABEL_HU "Túlmintavételezés (alulmintavételezéssel az eredeti felbontásra)"
#define BEETLE_OPT_SUPER_SAMPLING_INFO_0_HU "A felskálázott belső felbontáson generált tartalom alulmintavételezése az eredeti felbontáshoz. Nagyobb belső felbontással együtt így az eredeti, alacsony felbontáson is élsimított 3D objektumok tudnak megjelenni a játékban. A legjobb eredményt olyan címeknél adja, amelyek keverik a 2D és 3D elemeket (pl. 3D szereplők előre megrajzolt hátterek előtt), és jól működik a CRT shaderekkel együtt. Csak a Vulkan meghajtó támogatja. Figyelem: a Dithering mintázatot ajánlott kikapcsolni, ha ez az opció aktív."
#define BEETLE_OPT_MSAA_LABEL_HU "Többszörös mintavételezésű élsimítés (MSAA)"
#define BEETLE_OPT_MSAA_INFO_0_HU NULL
#define OPTION_VAL_1X_HU NULL
#define BEETLE_OPT_MDEC_YUV_LABEL_HU "MDEC YUV színjel szűrő"
#define BEETLE_OPT_MDEC_YUV_INFO_0_HU "Az FMV lejátszás minőségét javítja a \"macroblocking\" képhibák (négyzetek/cakkos élek) csökkentésével. Csak a Vulkan megjelenítő támogatja."
#define BEETLE_OPT_TRACK_TEXTURES_LABEL_HU "Textúrák nyomkövetése"
#define BEETLE_OPT_TRACK_TEXTURES_INFO_0_HU NULL
#define BEETLE_OPT_DUMP_TEXTURES_LABEL_HU "Textúrák lementése"
#define BEETLE_OPT_DUMP_TEXTURES_INFO_0_HU "A használt textúrák kimentése a <cd>texture-dump/ alá"
#define BEETLE_OPT_REPLACE_TEXTURES_LABEL_HU "Textúrák lecserélése"
#define BEETLE_OPT_REPLACE_TEXTURES_INFO_0_HU "A textúrák lecserélése HD verzióra a <cd>texture-replacements/ alól"
#define BEETLE_OPT_PGXP_MODE_LABEL_HU "PGXP működési módja"
#define BEETLE_OPT_PGXP_MODE_INFO_0_HU NULL
#define OPTION_VAL_MEMORY_ONLY_HU "Csak memóriában"
#define OPTION_VAL_MEMORY_CPU_HU "Memória + CPU (hibás)"
#define BEETLE_OPT_PGXP_2D_TOL_LABEL_HU "PGXP 2D geometria tűrés"
#define BEETLE_OPT_PGXP_2D_TOL_INFO_0_HU "A PGXP működés feltűnőbb hibáinak elrejtése: az érték azt a tűrést adja meg, amin belül maradnak a PGXP értékek megfelelő mélységi információval nem rendelkező geometriák esetén."
#define OPTION_VAL_0PX_HU NULL
#define OPTION_VAL_1PX_HU NULL
#define OPTION_VAL_2PX_HU NULL
#define OPTION_VAL_3PX_HU NULL
#define OPTION_VAL_4PX_HU NULL
#define OPTION_VAL_5PX_HU NULL
#define OPTION_VAL_6PX_HU NULL
#define OPTION_VAL_7PX_HU NULL
#define OPTION_VAL_8PX_HU NULL
#define BEETLE_OPT_PGXP_NCLIP_LABEL_HU NULL
#define BEETLE_OPT_PGXP_NCLIP_INFO_0_HU "A PGXP NCLIP implementációjának használata. Javítja a kinézetet azáltal, hogy csökkenti a lukakat PGXP koordinátával rendelkező geometriáknál. Néhány játék esetében ismerten fagyást okoz bizonyos körülmények közt."
#define BEETLE_OPT_PGXP_VERTEX_LABEL_HU "PGXP vertex gyorsítótár"
#define BEETLE_OPT_PGXP_VERTEX_INFO_0_HU "A PXGP által javított vertex pozíciók eltárolása és újrahasználata újabb poligonoknál. Javíthatja az objektumok illeszkedését és csökkenti a textúrák illesztési hézagait, de a gyorsítótár fals pozitív találatai grafikai hibákat okozhatnak. Jelenleg ajánlott kikapcsolva hagyni. Csak PGXP műveleti módnál érvényes, és csak a hardveres megjelenítők támogatják."
#define BEETLE_OPT_PGXP_TEXTURE_LABEL_HU NULL
#define BEETLE_OPT_PGXP_TEXTURE_INFO_0_HU NULL
#define BEETLE_OPT_DISPLAY_INTERNAL_FPS_LABEL_HU "Belső FPS megjelenítése"
#define BEETLE_OPT_DISPLAY_INTERNAL_FPS_INFO_0_HU "A belső képkockasebesség megjelenítése, amivel az emulált PlayStation rendszer előállítja a képet. Figyelem: a képernyőn megjelenő értesítések be kell legyenek kapcsolva a libretro frontenden."
#define BEETLE_OPT_DISPLAY_OSD_LABEL_HU "OSD üzenetek megjelenítése"
#define BEETLE_OPT_DISPLAY_OSD_INFO_0_HU "A mag által generált OSD üzenetek megjelenítése."
#define BEETLE_OPT_DISPLAY_VRAM_LABEL_HU "Az egész VRAM megjelenítése (debug)"
#define BEETLE_OPT_DISPLAY_VRAM_INFO_0_HU "Az emulált konzol teljes VRAM tartalmának megjelenítése. Csak az OpenGL és Vulkan hardveres megjelenítők támogatják. Figyelem: ez csak hibakeresési célokat szolgál, és normál esetben ki kell legyen kapcsolva."
#define BEETLE_OPT_ANALOG_CALIBRATION_LABEL_HU "Analóg önkalibráció"
#define BEETLE_OPT_ANALOG_CALIBRATION_INFO_0_HU NULL
#define BEETLE_OPT_ANALOG_TOGGLE_LABEL_HU "DualShock analóg mód átkapcsolás"
#define BEETLE_OPT_ANALOG_TOGGLE_INFO_0_HU "Ha a bemeneti eszköz típusa DualShock, ez az opció engedélyezi az emulált DualShock átkapcsolását a DIGITAL és ANALOG mód közt, az eredeti hardverhez hűen. A mód a kiválasztott gombkombinációval is átkapcsolható."
#define OPTION_VAL_ENABLED_ANALOG_HU "Alapértelmezett - Analóg"
#define BEETLE_OPT_ANALOG_TOGGLE_COMBO_LABEL_HU "DualShock analóg mód gombkombináció"
#define BEETLE_OPT_ANALOG_TOGGLE_COMBO_INFO_0_HU NULL
#define OPTION_VAL_L1_L2_R1_R2_START_SELECT_HU NULL
#define OPTION_VAL_L1_R1_SELECT_HU NULL
#define OPTION_VAL_L1_R1_START_HU NULL
#define OPTION_VAL_L1_R1_L3_HU NULL
#define OPTION_VAL_L1_R1_R3_HU NULL
#define OPTION_VAL_L2_R2_SELECT_HU NULL
#define OPTION_VAL_L2_R2_START_HU NULL
#define OPTION_VAL_L2_R2_L3_HU NULL
#define OPTION_VAL_L2_R2_R3_HU NULL
#define OPTION_VAL_L3_R3_HU NULL
#define BEETLE_OPT_ANALOG_TOGGLE_HOLD_LABEL_HU "DualShock analóg mód gombkombináció késleltetés"
#define BEETLE_OPT_ANALOG_TOGGLE_HOLD_INFO_0_HU NULL
#define OPTION_VAL_0_HU "0 másodperc"
#define OPTION_VAL_1_HU "1 másodperc"
#define OPTION_VAL_2_HU "2 másodperc"
#define OPTION_VAL_3_HU "3 másodperc"
#define OPTION_VAL_4_HU "4 másodperc"
#define OPTION_VAL_5_HU "5 másodperc"
#define BEETLE_OPT_ENABLE_MULTITAP_PORT1_LABEL_HU NULL
#define BEETLE_OPT_ENABLE_MULTITAP_PORT1_INFO_0_HU "A multitap funkció engedélyezése az 1. csatlakozón."
#define BEETLE_OPT_ENABLE_MULTITAP_PORT2_LABEL_HU NULL
#define BEETLE_OPT_ENABLE_MULTITAP_PORT2_INFO_0_HU "A multitap funkció engedélyezése a 2. csatlakozón."
#define BEETLE_OPT_GUN_INPUT_MODE_LABEL_HU "Fénypisztoly bemenet módja"
#define BEETLE_OPT_GUN_INPUT_MODE_INFO_0_HU NULL
#define OPTION_VAL_LIGHTGUN_HU "Fénypisztoly"
#define OPTION_VAL_TOUCHSCREEN_HU "Érintőképernyő"
#define BEETLE_OPT_GUN_CURSOR_LABEL_HU "Fénypisztoly kurzora"
#define BEETLE_OPT_GUN_CURSOR_INFO_0_HU NULL
#define OPTION_VAL_CROSS_HU "Kereszt"
#define OPTION_VAL_DOT_HU "Pont"
#define BEETLE_OPT_CROSSHAIR_COLOR_P1_LABEL_HU NULL
#define BEETLE_OPT_CROSSHAIR_COLOR_P1_INFO_0_HU NULL
#define OPTION_VAL_RED_HU "Vörös"
#define OPTION_VAL_BLUE_HU "Kék"
#define OPTION_VAL_GREEN_HU "Zöld"
#define OPTION_VAL_ORANGE_HU "Narancssárga"
#define OPTION_VAL_YELLOW_HU "Sárga"
#define OPTION_VAL_CYAN_HU "Ciánkék"
#define OPTION_VAL_PINK_HU "Rózsaszín"
#define OPTION_VAL_PURPLE_HU "Lila"
#define OPTION_VAL_BLACK_HU "Fekete"
#define OPTION_VAL_WHITE_HU "Fehér"
#define BEETLE_OPT_CROSSHAIR_COLOR_P2_LABEL_HU NULL
#define BEETLE_OPT_CROSSHAIR_COLOR_P2_INFO_0_HU NULL
#define BEETLE_OPT_MOUSE_SENSITIVITY_LABEL_HU "Egér érzékenysége"
#define BEETLE_OPT_MOUSE_SENSITIVITY_INFO_0_HU NULL
#define OPTION_VAL_5_O42_HU NULL
#define OPTION_VAL_10_HU NULL
#define OPTION_VAL_15_HU NULL
#define OPTION_VAL_20_HU NULL
#define OPTION_VAL_25_HU NULL
#define OPTION_VAL_30_HU NULL
#define OPTION_VAL_35_HU NULL
#define OPTION_VAL_40_HU NULL
#define OPTION_VAL_45_HU NULL
#define OPTION_VAL_55_HU NULL
#define OPTION_VAL_65_HU NULL
#define OPTION_VAL_75_HU NULL
#define OPTION_VAL_85_HU NULL
#define OPTION_VAL_95_HU NULL
#define OPTION_VAL_100_O42_HU NULL
#define OPTION_VAL_105_HU NULL
#define OPTION_VAL_115_HU NULL
#define OPTION_VAL_125_HU NULL
#define OPTION_VAL_135_HU NULL
#define OPTION_VAL_145_HU NULL
#define OPTION_VAL_155_HU NULL
#define OPTION_VAL_165_HU NULL
#define OPTION_VAL_175_HU NULL
#define OPTION_VAL_185_HU NULL
#define OPTION_VAL_195_HU NULL
#define BEETLE_OPT_NEGCON_RESPONSE_LABEL_HU "neGcon csavarodás érzékelés"
#define BEETLE_OPT_NEGCON_RESPONSE_INFO_0_HU NULL
#define OPTION_VAL_LINEAR_HU "Lineáris"
#define OPTION_VAL_QUADRATIC_HU "Négyzetes"
#define OPTION_VAL_CUBIC_HU "Köbös"
#define BEETLE_OPT_NEGCON_DEADZONE_LABEL_HU "neGcon csavarás holtsávja"
#define BEETLE_OPT_NEGCON_DEADZONE_INFO_0_HU NULL
#define OPTION_VAL_0_O44_HU NULL
#define BEETLE_OPT_USE_MEDNAFEN_MEMCARD0_METHOD_LABEL_HU NULL
#define BEETLE_OPT_USE_MEDNAFEN_MEMCARD0_METHOD_INFO_0_HU NULL
#define OPTION_VAL_LIBRETRO_HU NULL
#define OPTION_VAL_MEDNAFEN_HU NULL
#define BEETLE_OPT_ENABLE_MEMCARD1_LABEL_HU NULL
#define BEETLE_OPT_ENABLE_MEMCARD1_INFO_0_HU NULL
#define BEETLE_OPT_SHARED_MEMORY_CARDS_LABEL_HU NULL
#define BEETLE_OPT_SHARED_MEMORY_CARDS_INFO_0_HU NULL
#define BEETLE_OPT_MEMCARD_LEFT_INDEX_LABEL_HU NULL
#define BEETLE_OPT_MEMCARD_LEFT_INDEX_INFO_0_HU NULL
#define OPTION_VAL_0_O48_HU NULL
#define BEETLE_OPT_MEMCARD_RIGHT_INDEX_LABEL_HU NULL
#define BEETLE_OPT_MEMCARD_RIGHT_INDEX_INFO_0_HU NULL
#define OPTION_VAL_1_O49_HU NULL
#define BEETLE_OPT_LINE_RENDER_LABEL_HU "Sor-négyes trükk"
#define BEETLE_OPT_LINE_RENDER_INFO_0_HU NULL
#define OPTION_VAL_DEFAULT_HU "Alapértelmezett"
#define OPTION_VAL_AGGRESSIVE_HU "Agresszív"
#define BEETLE_OPT_WIDESCREEN_HACK_LABEL_HU "Szélesvásznú trükk"
#define BEETLE_OPT_WIDESCREEN_HACK_INFO_0_HU "A 3D tartalom anamorf előállítása, és az emulált framebuffer továbbítása szélesvásznú képaránnyal. A teljesen 3D játékoknál működik a legjobban. A 2D elemek vízszintesen megnyúlnak és esetleg nem jól illeszkednek."
#define BEETLE_OPT_WIDESCREEN_HACK_ASPECT_RATIO_LABEL_HU "Szélesvásznú trükk képaránya"
#define BEETLE_OPT_WIDESCREEN_HACK_ASPECT_RATIO_INFO_0_HU NULL
#define OPTION_VAL_16_9_HU NULL
#define OPTION_VAL_16_10_HU NULL
#define OPTION_VAL_18_9_HU NULL
#define OPTION_VAL_19_9_HU NULL
#define OPTION_VAL_20_9_HU NULL
#define OPTION_VAL_21_9_HU NULL
#define OPTION_VAL_32_9_HU NULL
#define BEETLE_OPT_CPU_DYNAREC_LABEL_HU NULL
#define BEETLE_OPT_CPU_DYNAREC_INFO_0_HU "A CPU utasítások dinamikus újrafordítása a futtató CPU saját utasításaira. Sokkal gyorsabb, mint az értelmező, de a CPU időzítés kevésbé pontos, és tartalmazhat hibákat."
#define OPTION_VAL_DISABLED_HU "Kikapcsolva (Beetle értelmező)"
#define OPTION_VAL_EXECUTE_HU "Maximális teljesítmény"
#define OPTION_VAL_EXECUTE_ONE_HU "Ciklusidőzítés ellenőrző"
#define OPTION_VAL_RUN_INTERPRETER_HU "Lightrec értelmező"
#define BEETLE_OPT_DYNAREC_INVALIDATE_LABEL_HU "Dynarec kód érvénytelenítés"
#define BEETLE_OPT_DYNAREC_INVALIDATE_INFO_0_HU "Néhány játéknak a Teljes, másoknak a Csak DMA beállításra van szüksége."
#define OPTION_VAL_FULL_HU "Teljes"
#define OPTION_VAL_DMA_HU "Csak DMA (kicsit gyorsabb)"
#define BEETLE_OPT_DYNAREC_EVENTCYCLES_LABEL_HU "Dynarec DMA/GPU/MDEC/időzítő esemény ciklusok"
#define BEETLE_OPT_DYNAREC_EVENTCYCLES_INFO_0_HU "A CPU maximális ciklusainak száma, mielőtt a DMA/GPU/MDEC/időzítő eseményeket ellenőrzi. Magasabb értéknél gyorsabb, a Beetle értelmezőre sokkal kisebb hatása van, mint a Dynarec módra."
#define OPTION_VAL_128_HU "128 (alapértelmezett)"
#define BEETLE_OPT_DYNAREC_SPU_SAMPLES_LABEL_HU "Dynarec SPU minták"
#define BEETLE_OPT_DYNAREC_SPU_SAMPLES_INFO_0_HU "Az SPU mintáinak maximális száma az SPU frissítés ellenőrzése előtt. Nagyobb értékek gyorsabbak, de néhány játéknál hanghibát okozhatnak 1-től eltérő beállításon."
#define BEETLE_OPT_CORE_TIMING_FPS_LABEL_HU "Mag által jelentett FPS időzítése"
#define BEETLE_OPT_CORE_TIMING_FPS_INFO_0_HU NULL
#define OPTION_VAL_AUTO_TOGGLE_HU NULL
#define OPTION_VAL_FORCE_PROGRESSIVE_HU NULL
#define OPTION_VAL_FORCE_INTERLACED_HU "Mindenképpen interlaced"
#define BEETLE_OPT_DEINTERLACER_LABEL_HU "Deinterlace módszer"
#define BEETLE_OPT_DEINTERLACER_INFO_0_HU NULL
#define OPTION_VAL_WEAVE_HU "Szövés"
#define OPTION_VAL_BOB_HU NULL
#define OPTION_VAL_BOB_OFFSET_HU NULL
#define OPTION_VAL_FASTMAD_HU NULL
#define BEETLE_OPT_FRAME_DUPING_LABEL_HU NULL
#define BEETLE_OPT_FRAME_DUPING_INFO_0_HU NULL
#define BEETLE_OPT_PAL_VIDEO_TIMING_OVERRIDE_LABEL_HU NULL
#define BEETLE_OPT_PAL_VIDEO_TIMING_OVERRIDE_INFO_0_HU NULL
#define BEETLE_OPT_ASPECT_RATIO_LABEL_HU "A mag képaránya"
#define BEETLE_OPT_ASPECT_RATIO_INFO_0_HU NULL
#define OPTION_VAL_CORRECTED_HU "Korrigált"
#define OPTION_VAL_UNCORRECTED_HU "Korrigálatlan"
#define OPTION_VAL_4_3_HU "Mindenképpen 4:3"
#define OPTION_VAL_NTSC_HU "Mindenképpen NTSC"
#define BEETLE_OPT_CROP_OVERSCAN_LABEL_HU "Overscan levágása"
#define BEETLE_OPT_CROP_OVERSCAN_INFO_0_HU NULL
#define OPTION_VAL_STATIC_HU "Vízszintes"
#define OPTION_VAL_SMART_HU NULL
#define BEETLE_OPT_IMAGE_CROP_LABEL_HU "További képvágás"
#define BEETLE_OPT_IMAGE_CROP_INFO_0_HU NULL
#define OPTION_VAL_9PX_HU NULL
#define OPTION_VAL_10PX_HU NULL
#define OPTION_VAL_11PX_HU NULL
#define OPTION_VAL_12PX_HU NULL
#define OPTION_VAL_13PX_HU NULL
#define OPTION_VAL_14PX_HU NULL
#define OPTION_VAL_15PX_HU NULL
#define OPTION_VAL_16PX_HU NULL
#define OPTION_VAL_17PX_HU NULL
#define OPTION_VAL_18PX_HU NULL
#define OPTION_VAL_19PX_HU NULL
#define OPTION_VAL_20PX_HU NULL
#define BEETLE_OPT_IMAGE_OFFSET_LABEL_HU "Vágott kép eltolása"
#define BEETLE_OPT_IMAGE_OFFSET_INFO_0_HU NULL
#define OPTION_VAL_12PX_O64_HU NULL
#define OPTION_VAL_11PX_O64_HU NULL
#define OPTION_VAL_10PX_O64_HU NULL
#define OPTION_VAL_9PX_O64_HU NULL
#define OPTION_VAL_8PX_O64_HU NULL
#define OPTION_VAL_7PX_O64_HU NULL
#define OPTION_VAL_6PX_O64_HU NULL
#define OPTION_VAL_5PX_O64_HU NULL
#define OPTION_VAL_4PX_O64_HU NULL
#define OPTION_VAL_3PX_O64_HU NULL
#define OPTION_VAL_2PX_O64_HU NULL
#define OPTION_VAL_1PX_O64_HU NULL
#define OPTION_VAL_1PX_O64_0_HU NULL
#define OPTION_VAL_2PX_O64_0_HU NULL
#define OPTION_VAL_3PX_O64_0_HU NULL
#define OPTION_VAL_4PX_O64_0_HU NULL
#define OPTION_VAL_5PX_O64_0_HU NULL
#define OPTION_VAL_6PX_O64_0_HU NULL
#define OPTION_VAL_7PX_O64_0_HU NULL
#define OPTION_VAL_8PX_O64_0_HU NULL
#define OPTION_VAL_9PX_O64_0_HU NULL
#define OPTION_VAL_10PX_O64_0_HU NULL
#define OPTION_VAL_11PX_O64_0_HU NULL
#define OPTION_VAL_12PX_O64_0_HU NULL
#define BEETLE_OPT_IMAGE_OFFSET_CYCLES_LABEL_HU "Kép vízszintes eltolása (GPU ciklus)"
#define BEETLE_OPT_IMAGE_OFFSET_CYCLES_INFO_0_HU NULL
#define BEETLE_OPT_INITIAL_SCANLINE_LABEL_HU "Kezdő scanline - NTSC"
#define BEETLE_OPT_INITIAL_SCANLINE_INFO_0_HU NULL
#define BEETLE_OPT_LAST_SCANLINE_LABEL_HU "Utolsó scanline - NTSC"
#define BEETLE_OPT_LAST_SCANLINE_INFO_0_HU NULL
#define OPTION_VAL_239_HU "239 (alapértelmezett)"
#define BEETLE_OPT_INITIAL_SCANLINE_PAL_LABEL_HU "Kezdő scanline - PAL"
#define BEETLE_OPT_INITIAL_SCANLINE_PAL_INFO_0_HU NULL
#define BEETLE_OPT_LAST_SCANLINE_PAL_LABEL_HU "Utolsó scanline - PAL"
#define BEETLE_OPT_LAST_SCANLINE_PAL_INFO_0_HU NULL
#define OPTION_VAL_287_HU "287 (alapértelmezett)"

struct retro_core_option_v2_category option_cats_hu[] = {
   {
      "system",
      CATEGORY_SYSTEM_LABEL_HU,
      CATEGORY_SYSTEM_INFO_0_HU
   },
   {
      "video",
      CATEGORY_VIDEO_LABEL_HU,
      CATEGORY_VIDEO_INFO_0_HU
   },
   {
      "pgxp",
      CATEGORY_PGXP_LABEL_HU,
      CATEGORY_PGXP_INFO_0_HU
   },
   {
      "osd",
      CATEGORY_OSD_LABEL_HU,
      CATEGORY_OSD_INFO_0_HU
   },
   {
      "input",
      CATEGORY_INPUT_LABEL_HU,
      CATEGORY_INPUT_INFO_0_HU
   },
   {
      "memcards",
      CATEGORY_MEMCARDS_LABEL_HU,
      CATEGORY_MEMCARDS_INFO_0_HU
   },
   {
      "hacks",
      CATEGORY_HACKS_LABEL_HU,
      CATEGORY_HACKS_INFO_0_HU
   },
   { NULL, NULL, NULL },
};
struct retro_core_option_v2_definition option_defs_hu[] = {
   {
      BEETLE_OPT(cpu_freq_scale),
      BEETLE_OPT_CPU_FREQ_SCALE_LABEL_HU,
      NULL,
      BEETLE_OPT_CPU_FREQ_SCALE_INFO_0_HU,
      NULL,
      "system",
      {
         { "50%",           OPTION_VAL_50_HU },
         { "60%",           OPTION_VAL_60_HU },
         { "70%",           OPTION_VAL_70_HU },
         { "80%",           OPTION_VAL_80_HU },
         { "90%",           OPTION_VAL_90_HU },
         { "100%",          OPTION_VAL_100_HU },
         { "110%",          OPTION_VAL_110_HU },
         { "120%",          OPTION_VAL_120_HU },
         { "130%",          OPTION_VAL_130_HU },
         { "140%",          OPTION_VAL_140_HU },
         { "150%",          OPTION_VAL_150_HU },
         { "160%",          OPTION_VAL_160_HU },
         { "170%",          OPTION_VAL_170_HU },
         { "180%",          OPTION_VAL_180_HU },
         { "190%",          OPTION_VAL_190_HU },
         { "200%",          OPTION_VAL_200_HU },
         { "210%",          OPTION_VAL_210_HU },
         { "220%",          OPTION_VAL_220_HU },
         { "230%",          OPTION_VAL_230_HU },
         { "240%",          OPTION_VAL_240_HU },
         { "250%",          OPTION_VAL_250_HU },
         { "260%",          OPTION_VAL_260_HU },
         { "270%",          OPTION_VAL_270_HU },
         { "280%",          OPTION_VAL_280_HU },
         { "290%",          OPTION_VAL_290_HU },
         { "300%",          OPTION_VAL_300_HU },
         { "310%",          OPTION_VAL_310_HU },
         { "320%",          OPTION_VAL_320_HU },
         { "330%",          OPTION_VAL_330_HU },
         { "340%",          OPTION_VAL_340_HU },
         { "350%",          OPTION_VAL_350_HU },
         { "360%",          OPTION_VAL_360_HU },
         { "370%",          OPTION_VAL_370_HU },
         { "380%",          OPTION_VAL_380_HU },
         { "390%",          OPTION_VAL_390_HU },
         { "400%",          OPTION_VAL_400_HU },
         { "410%",          OPTION_VAL_410_HU },
         { "420%",          OPTION_VAL_420_HU },
         { "430%",          OPTION_VAL_430_HU },
         { "440%",          OPTION_VAL_440_HU },
         { "450%",          OPTION_VAL_450_HU },
         { "460%",          OPTION_VAL_460_HU },
         { "470%",          OPTION_VAL_470_HU },
         { "480%",          OPTION_VAL_480_HU },
         { "490%",          OPTION_VAL_490_HU },
         { "500%",          OPTION_VAL_500_HU },
         { "510%",          OPTION_VAL_510_HU },
         { "520%",          OPTION_VAL_520_HU },
         { "530%",          OPTION_VAL_530_HU },
         { "540%",          OPTION_VAL_540_HU },
         { "550%",          OPTION_VAL_550_HU },
         { "560%",          OPTION_VAL_560_HU },
         { "570%",          OPTION_VAL_570_HU },
         { "580%",          OPTION_VAL_580_HU },
         { "590%",          OPTION_VAL_590_HU },
         { "600%",          OPTION_VAL_600_HU },
         { "610%",          OPTION_VAL_610_HU },
         { "620%",          OPTION_VAL_620_HU },
         { "630%",          OPTION_VAL_630_HU },
         { "640%",          OPTION_VAL_640_HU },
         { "650%",          OPTION_VAL_650_HU },
         { "660%",          OPTION_VAL_660_HU },
         { "670%",          OPTION_VAL_670_HU },
         { "680%",          OPTION_VAL_680_HU },
         { "690%",          OPTION_VAL_690_HU },
         { "700%",          OPTION_VAL_700_HU },
         { "710%",          OPTION_VAL_710_HU },
         { "720%",          OPTION_VAL_720_HU },
         { "730%",          OPTION_VAL_730_HU },
         { "740%",          OPTION_VAL_740_HU },
         { "750%",          OPTION_VAL_750_HU },
         { NULL, NULL },
      },
      "100%"
   },
   {
      BEETLE_OPT(gpu_overclock),
      BEETLE_OPT_GPU_OVERCLOCK_LABEL_HU,
      NULL,
      BEETLE_OPT_GPU_OVERCLOCK_INFO_0_HU,
      NULL,
      "system",
      {
         { "1x(native)", OPTION_VAL_1X_NATIVE_HU },
         { "2x",         OPTION_VAL_2X_HU },
         { "4x",         OPTION_VAL_4X_HU },
         { "8x",         OPTION_VAL_8X_HU },
         { "16x",        OPTION_VAL_16X_HU },
         { "32x",        OPTION_VAL_32X_HU },
         { NULL, NULL },
      },
      "1x(native)"
   },
   {
      BEETLE_OPT(gte_overclock),
      BEETLE_OPT_GTE_OVERCLOCK_LABEL_HU,
      NULL,
      BEETLE_OPT_GTE_OVERCLOCK_INFO_0_HU,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
#ifndef EMSCRIPTEN
   {
      BEETLE_OPT(cd_access_method),
      BEETLE_OPT_CD_ACCESS_METHOD_LABEL_HU,
      NULL,
      BEETLE_OPT_CD_ACCESS_METHOD_INFO_0_HU,
      NULL,
      "system",
      {
         { "sync",     OPTION_VAL_SYNC_HU },
         { "async",    OPTION_VAL_ASYNC_HU },
         { "precache", OPTION_VAL_PRECACHE_HU },
         { NULL, NULL },
      },
      "sync"
   },
#endif
   {
      BEETLE_OPT(cd_fastload),
      BEETLE_OPT_CD_FASTLOAD_LABEL_HU,
      NULL,
      BEETLE_OPT_CD_FASTLOAD_INFO_0_HU,
      NULL,
      "system",
      {
         { "2x(native)", OPTION_VAL_2X_NATIVE_HU },
         { "4x",          OPTION_VAL_4X_HU },
         { "6x",          OPTION_VAL_6X_HU },
         { "8x",          OPTION_VAL_8X_HU },
         { "10x",         OPTION_VAL_10X_HU },
         { "12x",         OPTION_VAL_12X_HU },
         { "14x",         OPTION_VAL_14X_HU },
         { NULL, NULL },
      },
      "2x(native)"
   },
   {
      BEETLE_OPT(skip_bios),
      BEETLE_OPT_SKIP_BIOS_LABEL_HU,
      NULL,
      BEETLE_OPT_SKIP_BIOS_INFO_0_HU,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(override_bios),
      BEETLE_OPT_OVERRIDE_BIOS_LABEL_HU,
      NULL,
      BEETLE_OPT_OVERRIDE_BIOS_INFO_0_HU,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "psxonpsp", OPTION_VAL_PSXONPSP_HU },
         { "ps1_rom",  OPTION_VAL_PS1_ROM_HU },
         { "openbios", OPTION_VAL_OPENBIOS_HU     },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(region),
      BEETLE_OPT_REGION_LABEL_HU,
      NULL,
      BEETLE_OPT_REGION_INFO_0_HU,
      NULL,
      "system",
      {
         { "auto",   OPTION_VAL_AUTO_HU },
         { "ntsc-j", OPTION_VAL_NTSC_J_HU },
         { "ntsc-u", OPTION_VAL_NTSC_U_HU },
         { "pal",    OPTION_VAL_PAL_HU },
         { NULL, NULL },
      },
      "auto"
   },
   {
      BEETLE_OPT(internal_resolution),
      BEETLE_OPT_INTERNAL_RESOLUTION_LABEL_HU,
      NULL,
      BEETLE_OPT_INTERNAL_RESOLUTION_INFO_0_HU,
      NULL,
      "video",
      {
         { "1x(native)", OPTION_VAL_1X_NATIVE_HU },
         { "2x",         OPTION_VAL_2X_HU },
         { "4x",         OPTION_VAL_4X_HU },
         { "8x",         OPTION_VAL_8X_HU },
         { "16x",        OPTION_VAL_16X_HU },
         { NULL, NULL },
      },
      "1x(native)"
   },
#if defined(HAVE_OPENGL) || defined(HAVE_OPENGLES) || defined(HAVE_VULKAN)
   {
      BEETLE_OPT(renderer),
      BEETLE_OPT_RENDERER_LABEL_HU,
      NULL,
      BEETLE_OPT_RENDERER_INFO_0_HU,
      NULL,
      "video",
      {
         { "hardware",    OPTION_VAL_HARDWARE_HU },
#if defined(HAVE_OPENGL) || defined(HAVE_OPENGLES)
         { "hardware_gl", OPTION_VAL_HARDWARE_GL_HU },
#endif
#if defined(HAVE_VULKAN)
         { "hardware_vk", OPTION_VAL_HARDWARE_VK_HU },
#endif
         { "software",    OPTION_VAL_SOFTWARE_HU },
         { NULL, NULL },
      },
      "hardware"
   },
   {
      BEETLE_OPT(renderer_software_fb),
      BEETLE_OPT_RENDERER_SOFTWARE_FB_LABEL_HU,
      NULL,
      BEETLE_OPT_RENDERER_SOFTWARE_FB_INFO_0_HU,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled"
   },
#endif
#if defined(HAVE_OPENGL) || defined(HAVE_OPENGLES)
   {
      BEETLE_OPT(depth),
      BEETLE_OPT_DEPTH_LABEL_HU,
      NULL,
      BEETLE_OPT_DEPTH_INFO_0_HU,
      NULL,
      "video",
      {
         { "16bpp(native)", OPTION_VAL_16BPP_NATIVE_HU },
         { "32bpp",         OPTION_VAL_32BPP_HU },
         { NULL, NULL },
      },
      "16bpp(native)"
   },
   // Sort of, it's more like 15-bit high color and 24-bit true color for visible output. The alpha channel is used for mask bit. Vulkan renderer uses ABGR1555_555 for 31 bits internal? FMVs are always 24-bit on all renderers like original hardware (BGR888, no alpha)
#endif
   {
      BEETLE_OPT(dither_mode),
      BEETLE_OPT_DITHER_MODE_LABEL_HU,
      NULL,
      BEETLE_OPT_DITHER_MODE_INFO_0_HU,
      NULL,
      "video",
      {
         { "1x(native)",          OPTION_VAL_1X_NATIVE_HU },
         { "internal resolution", OPTION_VAL_INTERNAL_RESOLUTION_HU },
         { "disabled",            NULL },
         { NULL, NULL },
      },
      "1x(native)"
   },
#ifdef HAVE_VULKAN
   {
      BEETLE_OPT(scaled_uv_offset),
      BEETLE_OPT_SCALED_UV_OFFSET_LABEL_HU,
      NULL,
      BEETLE_OPT_SCALED_UV_OFFSET_INFO_0_HU,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled"
   },
#endif
#if defined(HAVE_OPENGL) || defined(HAVE_OPENGLES) || defined(HAVE_VULKAN)
   {
      BEETLE_OPT(filter),
      BEETLE_OPT_FILTER_LABEL_HU,
      NULL,
      BEETLE_OPT_FILTER_INFO_0_HU,
      NULL,
      "video",
      {
         { "nearest",  OPTION_VAL_NEAREST_HU },
         { "SABR",     OPTION_VAL_SABR_HU },
         { "xBR",      OPTION_VAL_XBR_HU },
         { "bilinear", OPTION_VAL_BILINEAR_HU },
         { "3-point",  OPTION_VAL_3_POINT_HU },
         { "JINC2",    OPTION_VAL_JINC2_HU },
         { NULL, NULL },
      },
      "nearest"
   },
#ifdef HAVE_VULKAN
   {
      BEETLE_OPT(filter_exclude_sprite),
      BEETLE_OPT_FILTER_EXCLUDE_SPRITE_LABEL_HU,
      NULL,
      BEETLE_OPT_FILTER_EXCLUDE_SPRITE_INFO_0_HU,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "opaque", OPTION_VAL_OPAQUE_HU },
         { "all", OPTION_VAL_ALL_HU },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(filter_exclude_2d_polygon),
      BEETLE_OPT_FILTER_EXCLUDE_2D_POLYGON_LABEL_HU,
      NULL,
      BEETLE_OPT_FILTER_EXCLUDE_2D_POLYGON_INFO_0_HU,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "opaque", OPTION_VAL_OPAQUE_HU },
         { "all", OPTION_VAL_ALL_HU },
         { NULL, NULL },
      },
      "disabled"
   },
#endif
#endif
#ifdef HAVE_VULKAN
   {
      BEETLE_OPT(adaptive_smoothing),
      BEETLE_OPT_ADAPTIVE_SMOOTHING_LABEL_HU,
      NULL,
      BEETLE_OPT_ADAPTIVE_SMOOTHING_INFO_0_HU,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(super_sampling),
      BEETLE_OPT_SUPER_SAMPLING_LABEL_HU,
      NULL,
      BEETLE_OPT_SUPER_SAMPLING_INFO_0_HU,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(msaa),
      BEETLE_OPT_MSAA_LABEL_HU,
      NULL,
      BEETLE_OPT_MSAA_INFO_0_HU,
      NULL,
      "video",
      {
         { "1x",  OPTION_VAL_1X_HU },
         { "2x",  OPTION_VAL_2X_HU },
         { "4x",  OPTION_VAL_4X_HU },
         { "8x",  OPTION_VAL_8X_HU },
         { "16x", OPTION_VAL_16X_HU },
         { NULL, NULL },
      },
      "1x"
   },
   {
      BEETLE_OPT(mdec_yuv),
      BEETLE_OPT_MDEC_YUV_LABEL_HU,
      NULL,
      BEETLE_OPT_MDEC_YUV_INFO_0_HU,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(track_textures),
      BEETLE_OPT_TRACK_TEXTURES_LABEL_HU,
      NULL,
      BEETLE_OPT_TRACK_TEXTURES_INFO_0_HU,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
#ifdef TEXTURE_DUMPING_ENABLED
   {
      BEETLE_OPT(dump_textures),
      BEETLE_OPT_DUMP_TEXTURES_LABEL_HU,
      NULL,
      BEETLE_OPT_DUMP_TEXTURES_INFO_0_HU,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
#endif
   {
      BEETLE_OPT(replace_textures),
      BEETLE_OPT_REPLACE_TEXTURES_LABEL_HU,
      NULL,
      BEETLE_OPT_REPLACE_TEXTURES_INFO_0_HU,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
#endif
   {
      BEETLE_OPT(pgxp_mode),
      BEETLE_OPT_PGXP_MODE_LABEL_HU,
      NULL,
      BEETLE_OPT_PGXP_MODE_INFO_0_HU,
      NULL,
      "pgxp",
      {
         { "disabled",     NULL },
         { "memory only",  OPTION_VAL_MEMORY_ONLY_HU },
         { "memory + CPU", OPTION_VAL_MEMORY_CPU_HU },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(pgxp_2d_tol),
      BEETLE_OPT_PGXP_2D_TOL_LABEL_HU,
      NULL,
      BEETLE_OPT_PGXP_2D_TOL_INFO_0_HU,
      NULL,
      "pgxp",
      {
         { "disabled", NULL },
         { "0px", OPTION_VAL_0PX_HU },
         { "1px", OPTION_VAL_1PX_HU },
         { "2px", OPTION_VAL_2PX_HU },
         { "3px", OPTION_VAL_3PX_HU },
         { "4px", OPTION_VAL_4PX_HU },
         { "5px", OPTION_VAL_5PX_HU },
         { "6px", OPTION_VAL_6PX_HU },
         { "7px", OPTION_VAL_7PX_HU },
         { "8px", OPTION_VAL_8PX_HU },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(pgxp_nclip),
      BEETLE_OPT_PGXP_NCLIP_LABEL_HU,
      NULL,
      BEETLE_OPT_PGXP_NCLIP_INFO_0_HU,
      NULL,
      "pgxp",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
#if defined(HAVE_OPENGL) || defined(HAVE_OPENGLES) || defined(HAVE_VULKAN)
   {
      BEETLE_OPT(pgxp_vertex),
      BEETLE_OPT_PGXP_VERTEX_LABEL_HU,
      NULL,
      BEETLE_OPT_PGXP_VERTEX_INFO_0_HU,
      NULL,
      "pgxp",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
#endif
   {
      BEETLE_OPT(pgxp_texture),
      BEETLE_OPT_PGXP_TEXTURE_LABEL_HU,
      NULL,
      BEETLE_OPT_PGXP_TEXTURE_INFO_0_HU,
      NULL,
      "pgxp",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(display_internal_fps),
      BEETLE_OPT_DISPLAY_INTERNAL_FPS_LABEL_HU,
      NULL,
      BEETLE_OPT_DISPLAY_INTERNAL_FPS_INFO_0_HU,
      NULL,
      "osd",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(display_osd),
      BEETLE_OPT_DISPLAY_OSD_LABEL_HU,
      NULL,
      BEETLE_OPT_DISPLAY_OSD_INFO_0_HU,
      NULL,
      "osd",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled"
   },
#if defined(HAVE_OPENGL) || defined(HAVE_OPENGLES) || defined(HAVE_VULKAN)
   {
      BEETLE_OPT(display_vram),
      BEETLE_OPT_DISPLAY_VRAM_LABEL_HU,
      NULL,
      BEETLE_OPT_DISPLAY_VRAM_INFO_0_HU,
      NULL,
      "osd",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
#endif
   {
      BEETLE_OPT(analog_calibration),
      BEETLE_OPT_ANALOG_CALIBRATION_LABEL_HU,
      NULL,
      BEETLE_OPT_ANALOG_CALIBRATION_INFO_0_HU,
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(analog_toggle),
      BEETLE_OPT_ANALOG_TOGGLE_LABEL_HU,
      NULL,
      BEETLE_OPT_ANALOG_TOGGLE_INFO_0_HU,
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { "enabled-analog", OPTION_VAL_ENABLED_ANALOG_HU },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(analog_toggle_combo),
      BEETLE_OPT_ANALOG_TOGGLE_COMBO_LABEL_HU,
      NULL,
      BEETLE_OPT_ANALOG_TOGGLE_COMBO_INFO_0_HU,
      NULL,
      "input",
      {
         { "l1+l2+r1+r2+start+select", OPTION_VAL_L1_L2_R1_R2_START_SELECT_HU },
         { "l1+r1+select",             OPTION_VAL_L1_R1_SELECT_HU },
         { "l1+r1+start",              OPTION_VAL_L1_R1_START_HU },
         { "l1+r1+l3",                 OPTION_VAL_L1_R1_L3_HU },
         { "l1+r1+r3",                 OPTION_VAL_L1_R1_R3_HU },
         { "l2+r2+select",             OPTION_VAL_L2_R2_SELECT_HU },
         { "l2+r2+start",              OPTION_VAL_L2_R2_START_HU },
         { "l2+r2+l3",                 OPTION_VAL_L2_R2_L3_HU },
         { "l2+r2+r3",                 OPTION_VAL_L2_R2_R3_HU },
         { "l3+r3",                    OPTION_VAL_L3_R3_HU },
         { NULL, NULL },
      },
      "l1+r1+select"
   },
   {
      BEETLE_OPT(analog_toggle_hold),
      BEETLE_OPT_ANALOG_TOGGLE_HOLD_LABEL_HU,
      NULL,
      BEETLE_OPT_ANALOG_TOGGLE_HOLD_INFO_0_HU,
      NULL,
      "input",
      {
         { "0", OPTION_VAL_0_HU },
         { "1", OPTION_VAL_1_HU },
         { "2", OPTION_VAL_2_HU },
         { "3", OPTION_VAL_3_HU },
         { "4", OPTION_VAL_4_HU },
         { "5", OPTION_VAL_5_HU },
         { NULL, NULL },
      },
      "1"
   },
   {
      BEETLE_OPT(enable_multitap_port1),
      BEETLE_OPT_ENABLE_MULTITAP_PORT1_LABEL_HU,
      NULL,
      BEETLE_OPT_ENABLE_MULTITAP_PORT1_INFO_0_HU,
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(enable_multitap_port2),
      BEETLE_OPT_ENABLE_MULTITAP_PORT2_LABEL_HU,
      NULL,
      BEETLE_OPT_ENABLE_MULTITAP_PORT2_INFO_0_HU,
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(gun_input_mode),
      BEETLE_OPT_GUN_INPUT_MODE_LABEL_HU,
      NULL,
      BEETLE_OPT_GUN_INPUT_MODE_INFO_0_HU,
      NULL,
      "input",
      {
         { "lightgun",    OPTION_VAL_LIGHTGUN_HU },
         { "touchscreen", OPTION_VAL_TOUCHSCREEN_HU },
         { NULL, NULL },
      },
      "lightgun"
   },
   // Shouldn't the gun_input_mode just be Mouse vs. Touchscreen?
   {
      BEETLE_OPT(gun_cursor),
      BEETLE_OPT_GUN_CURSOR_LABEL_HU,
      NULL,
      BEETLE_OPT_GUN_CURSOR_INFO_0_HU,
      NULL,
      "input",
      {
         { "off",   "disabled" },
         { "cross", OPTION_VAL_CROSS_HU },
         { "dot",   OPTION_VAL_DOT_HU },
         { NULL, NULL },
      },
      "cross"
   },
   {
      BEETLE_OPT(crosshair_color_p1),
      BEETLE_OPT_CROSSHAIR_COLOR_P1_LABEL_HU,
      NULL,
      BEETLE_OPT_CROSSHAIR_COLOR_P1_INFO_0_HU,
      NULL,
      "input",
      {
         { "red",    OPTION_VAL_RED_HU },
         { "blue",   OPTION_VAL_BLUE_HU },
         { "green",  OPTION_VAL_GREEN_HU },
         { "orange", OPTION_VAL_ORANGE_HU },
         { "yellow", OPTION_VAL_YELLOW_HU },
         { "cyan",   OPTION_VAL_CYAN_HU },
         { "pink",   OPTION_VAL_PINK_HU },
         { "purple", OPTION_VAL_PURPLE_HU },
         { "black",  OPTION_VAL_BLACK_HU },
         { "white",  OPTION_VAL_WHITE_HU },
         { NULL, NULL },
      },
      "red"
   },
   {
      BEETLE_OPT(crosshair_color_p2),
      BEETLE_OPT_CROSSHAIR_COLOR_P2_LABEL_HU,
      NULL,
      BEETLE_OPT_CROSSHAIR_COLOR_P2_INFO_0_HU,
      NULL,
      "input",
      {
         { "blue",   OPTION_VAL_BLUE_HU },
         { "red",    OPTION_VAL_RED_HU },
         { "green",  OPTION_VAL_GREEN_HU },
         { "orange", OPTION_VAL_ORANGE_HU },
         { "yellow", OPTION_VAL_YELLOW_HU },
         { "cyan",   OPTION_VAL_CYAN_HU },
         { "pink",   OPTION_VAL_PINK_HU },
         { "purple", OPTION_VAL_PURPLE_HU },
         { "black",  OPTION_VAL_BLACK_HU },
         { "white",  OPTION_VAL_WHITE_HU },
         { NULL, NULL },
      },
      "blue"
   },
   {
      BEETLE_OPT(mouse_sensitivity),
      BEETLE_OPT_MOUSE_SENSITIVITY_LABEL_HU,
      NULL,
      BEETLE_OPT_MOUSE_SENSITIVITY_INFO_0_HU,
      NULL,
      "input",
      {
         { "5%",   OPTION_VAL_5_O42_HU },
         { "10%",  OPTION_VAL_10_HU },
         { "15%",  OPTION_VAL_15_HU },
         { "20%",  OPTION_VAL_20_HU },
         { "25%",  OPTION_VAL_25_HU },
         { "30%",  OPTION_VAL_30_HU },
         { "35%",  OPTION_VAL_35_HU },
         { "40%",  OPTION_VAL_40_HU },
         { "45%",  OPTION_VAL_45_HU },
         { "50%",  OPTION_VAL_50_HU },
         { "55%",  OPTION_VAL_55_HU },
         { "60%",  OPTION_VAL_60_HU },
         { "65%",  OPTION_VAL_65_HU },
         { "70%",  OPTION_VAL_70_HU },
         { "75%",  OPTION_VAL_75_HU },
         { "80%",  OPTION_VAL_80_HU },
         { "85%",  OPTION_VAL_85_HU },
         { "90%",  OPTION_VAL_90_HU },
         { "95%",  OPTION_VAL_95_HU },
         { "100%", OPTION_VAL_100_O42_HU },
         { "105%", OPTION_VAL_105_HU },
         { "110%", OPTION_VAL_110_HU },
         { "115%", OPTION_VAL_115_HU },
         { "120%", OPTION_VAL_120_HU },
         { "125%", OPTION_VAL_125_HU },
         { "130%", OPTION_VAL_130_HU },
         { "135%", OPTION_VAL_135_HU },
         { "140%", OPTION_VAL_140_HU },
         { "145%", OPTION_VAL_145_HU },
         { "150%", OPTION_VAL_150_HU },
         { "155%", OPTION_VAL_155_HU },
         { "160%", OPTION_VAL_160_HU },
         { "165%", OPTION_VAL_165_HU },
         { "170%", OPTION_VAL_170_HU },
         { "175%", OPTION_VAL_175_HU },
         { "180%", OPTION_VAL_180_HU },
         { "185%", OPTION_VAL_185_HU },
         { "190%", OPTION_VAL_190_HU },
         { "195%", OPTION_VAL_195_HU },
         { "200%", OPTION_VAL_200_HU },
         { NULL, NULL },
      },
      "100%"
   },
   {
      BEETLE_OPT(negcon_response),
      BEETLE_OPT_NEGCON_RESPONSE_LABEL_HU,
      NULL,
      BEETLE_OPT_NEGCON_RESPONSE_INFO_0_HU,
      NULL,
      "input",
      {
         { "linear",    OPTION_VAL_LINEAR_HU },
         { "quadratic", OPTION_VAL_QUADRATIC_HU },
         { "cubic",     OPTION_VAL_CUBIC_HU },
         { NULL, NULL },
      },
      "linear"
   },
   {
      BEETLE_OPT(negcon_deadzone),
      BEETLE_OPT_NEGCON_DEADZONE_LABEL_HU,
      NULL,
      BEETLE_OPT_NEGCON_DEADZONE_INFO_0_HU,
      NULL,
      "input",
      {
         { "0%",  OPTION_VAL_0_O44_HU },
         { "5%",  OPTION_VAL_5_O42_HU },
         { "10%", OPTION_VAL_10_HU },
         { "15%", OPTION_VAL_15_HU },
         { "20%", OPTION_VAL_20_HU },
         { "25%", OPTION_VAL_25_HU },
         { "30%", OPTION_VAL_30_HU },
         { NULL, NULL },
      },
      "0%"
   },
   {
      BEETLE_OPT(use_mednafen_memcard0_method),
      BEETLE_OPT_USE_MEDNAFEN_MEMCARD0_METHOD_LABEL_HU,
      NULL,
      BEETLE_OPT_USE_MEDNAFEN_MEMCARD0_METHOD_INFO_0_HU,
      NULL,
      "memcards",
      {
         { "libretro", OPTION_VAL_LIBRETRO_HU },
         { "mednafen", OPTION_VAL_MEDNAFEN_HU },
         { NULL, NULL },
      },
      "libretro"
   },
   {
      BEETLE_OPT(enable_memcard1),
      BEETLE_OPT_ENABLE_MEMCARD1_LABEL_HU,
      NULL,
      BEETLE_OPT_ENABLE_MEMCARD1_INFO_0_HU,
      NULL,
      "memcards",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(shared_memory_cards),
      BEETLE_OPT_SHARED_MEMORY_CARDS_LABEL_HU,
      NULL,
      BEETLE_OPT_SHARED_MEMORY_CARDS_INFO_0_HU,
      NULL,
      "memcards",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(memcard_left_index),
      BEETLE_OPT_MEMCARD_LEFT_INDEX_LABEL_HU,
      NULL,
      BEETLE_OPT_MEMCARD_LEFT_INDEX_INFO_0_HU,
      NULL,
      "memcards",
      {
         { "0",  OPTION_VAL_0_O48_HU },
         { "1",  NULL },
         { "2",  NULL },
         { "3",  NULL },
         { "4",  NULL },
         { "5",  NULL },
         { "6",  NULL },
         { "7",  NULL },
         { "8",  NULL },
         { "9",  NULL },
         { "10", NULL },
         { "11", NULL },
         { "12", NULL },
         { "13", NULL },
         { "14", NULL },
         { "15", NULL },
         { "16", NULL },
         { "17", NULL },
         { "18", NULL },
         { "19", NULL },
         { "20", NULL },
         { "21", NULL },
         { "22", NULL },
         { "23", NULL },
         { "24", NULL },
         { "25", NULL },
         { "26", NULL },
         { "27", NULL },
         { "28", NULL },
         { "29", NULL },
         { "30", NULL },
         { "31", NULL },
         { "32", NULL },
         { "33", NULL },
         { "34", NULL },
         { "35", NULL },
         { "36", NULL },
         { "37", NULL },
         { "38", NULL },
         { "39", NULL },
         { "40", NULL },
         { "41", NULL },
         { "42", NULL },
         { "43", NULL },
         { "44", NULL },
         { "45", NULL },
         { "46", NULL },
         { "47", NULL },
         { "48", NULL },
         { "49", NULL },
         { "50", NULL },
         { "51", NULL },
         { "52", NULL },
         { "53", NULL },
         { "54", NULL },
         { "55", NULL },
         { "56", NULL },
         { "57", NULL },
         { "58", NULL },
         { "59", NULL },
         { "60", NULL },
         { "61", NULL },
         { "62", NULL },
         { "63", NULL },
         { NULL, NULL },
      },
      "0"
   },
   {
      BEETLE_OPT(memcard_right_index),
      BEETLE_OPT_MEMCARD_RIGHT_INDEX_LABEL_HU,
      NULL,
      BEETLE_OPT_MEMCARD_RIGHT_INDEX_INFO_0_HU,
      NULL,
      "memcards",
      {
         { "0",  NULL },
         { "1",  OPTION_VAL_1_O49_HU },
         { "2",  NULL },
         { "3",  NULL },
         { "4",  NULL },
         { "5",  NULL },
         { "6",  NULL },
         { "7",  NULL },
         { "8",  NULL },
         { "9",  NULL },
         { "10", NULL },
         { "11", NULL },
         { "12", NULL },
         { "13", NULL },
         { "14", NULL },
         { "15", NULL },
         { "16", NULL },
         { "17", NULL },
         { "18", NULL },
         { "19", NULL },
         { "20", NULL },
         { "21", NULL },
         { "22", NULL },
         { "23", NULL },
         { "24", NULL },
         { "25", NULL },
         { "26", NULL },
         { "27", NULL },
         { "28", NULL },
         { "29", NULL },
         { "30", NULL },
         { "31", NULL },
         { "32", NULL },
         { "33", NULL },
         { "34", NULL },
         { "35", NULL },
         { "36", NULL },
         { "37", NULL },
         { "38", NULL },
         { "39", NULL },
         { "40", NULL },
         { "41", NULL },
         { "42", NULL },
         { "43", NULL },
         { "44", NULL },
         { "45", NULL },
         { "46", NULL },
         { "47", NULL },
         { "48", NULL },
         { "49", NULL },
         { "50", NULL },
         { "51", NULL },
         { "52", NULL },
         { "53", NULL },
         { "54", NULL },
         { "55", NULL },
         { "56", NULL },
         { "57", NULL },
         { "58", NULL },
         { "59", NULL },
         { "60", NULL },
         { "61", NULL },
         { "62", NULL },
         { "63", NULL },
         { NULL, NULL },
      },
      "1"
   },
   {
      BEETLE_OPT(line_render),
      BEETLE_OPT_LINE_RENDER_LABEL_HU,
      NULL,
      BEETLE_OPT_LINE_RENDER_INFO_0_HU,
      NULL,
      "hacks",
      {
         { "default",    OPTION_VAL_DEFAULT_HU },
         { "aggressive", OPTION_VAL_AGGRESSIVE_HU },
         { "disabled",   NULL },
         { NULL, NULL },
      },
      "default"
   },
   {
      BEETLE_OPT(widescreen_hack),
      BEETLE_OPT_WIDESCREEN_HACK_LABEL_HU,
      NULL,
      BEETLE_OPT_WIDESCREEN_HACK_INFO_0_HU,
      NULL,
      "hacks",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(widescreen_hack_aspect_ratio),
      BEETLE_OPT_WIDESCREEN_HACK_ASPECT_RATIO_LABEL_HU,
      NULL,
      BEETLE_OPT_WIDESCREEN_HACK_ASPECT_RATIO_INFO_0_HU,
      NULL,
      "hacks",
      {
         { "16:9",  OPTION_VAL_16_9_HU },
         { "16:10", OPTION_VAL_16_10_HU },
         { "18:9",  OPTION_VAL_18_9_HU },
         { "19:9",  OPTION_VAL_19_9_HU },
         { "20:9",  OPTION_VAL_20_9_HU },
         { "21:9",  OPTION_VAL_21_9_HU }, // 64:27
         { "32:9",  OPTION_VAL_32_9_HU },
         { NULL,    NULL },
      },
      "16:9"
   },
#if defined(HAVE_LIGHTREC)
   {
      BEETLE_OPT(cpu_dynarec),
      BEETLE_OPT_CPU_DYNAREC_LABEL_HU,
      NULL,
      BEETLE_OPT_CPU_DYNAREC_INFO_0_HU,
      NULL,
      "hacks",
      {
         { "disabled",        OPTION_VAL_DISABLED_HU },
         { "execute",         OPTION_VAL_EXECUTE_HU },
         { "execute_one",     OPTION_VAL_EXECUTE_ONE_HU },
         { "run_interpreter", OPTION_VAL_RUN_INTERPRETER_HU },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(dynarec_invalidate),
      BEETLE_OPT_DYNAREC_INVALIDATE_LABEL_HU,
      NULL,
      BEETLE_OPT_DYNAREC_INVALIDATE_INFO_0_HU,
      NULL,
      "hacks",
      {
         { "full", OPTION_VAL_FULL_HU },
         { "dma",  OPTION_VAL_DMA_HU },
         { NULL, NULL },
      },
      "full"
   },
   {
      BEETLE_OPT(dynarec_eventcycles),
      BEETLE_OPT_DYNAREC_EVENTCYCLES_LABEL_HU,
      NULL,
      BEETLE_OPT_DYNAREC_EVENTCYCLES_INFO_0_HU,
      NULL,
      "hacks",
      {
         { "128",  OPTION_VAL_128_HU },
         { "256",  NULL },
         { "384",  NULL },
         { "512",  NULL },
         { "640",  NULL },
         { "768",  NULL },
         { "896",  NULL },
         { "1024", NULL },
         { "1152", NULL },
         { "1280", NULL },
         { "1408", NULL },
         { "1536", NULL },
         { "1664", NULL },
         { "1792", NULL },
         { "1920", NULL },
         { "2048", NULL },
         { NULL, NULL },
      },
      "128"
   },
   {
      BEETLE_OPT(dynarec_spu_samples),
      BEETLE_OPT_DYNAREC_SPU_SAMPLES_LABEL_HU,
      NULL,
      BEETLE_OPT_DYNAREC_SPU_SAMPLES_INFO_0_HU,
      NULL,
      "hacks",
      {
         { "1",  OPTION_VAL_1_O49_HU },
         { "4",  NULL },
         { "16", NULL },
         { NULL, NULL },
      },
      "1"
   },
#endif
   {
      BEETLE_OPT(core_timing_fps),
      BEETLE_OPT_CORE_TIMING_FPS_LABEL_HU,
      NULL,
      BEETLE_OPT_CORE_TIMING_FPS_INFO_0_HU,
      NULL,
      "video",
      {
         { "auto_toggle",       OPTION_VAL_AUTO_TOGGLE_HU },
         { "force_progressive", OPTION_VAL_FORCE_PROGRESSIVE_HU },
         { "force_interlaced",  OPTION_VAL_FORCE_INTERLACED_HU },
      },
      "auto_toggle"
   },
   {
      BEETLE_OPT(deinterlacer),
      BEETLE_OPT_DEINTERLACER_LABEL_HU,
      NULL,
      BEETLE_OPT_DEINTERLACER_INFO_0_HU,
      NULL,
      "video",
      {
         { "weave",      OPTION_VAL_WEAVE_HU },
         { "bob",        OPTION_VAL_BOB_HU },
         { "bob_offset", OPTION_VAL_BOB_OFFSET_HU },
         { "fastmad",    OPTION_VAL_FASTMAD_HU },
         { "off",        "Off" },
         { NULL, NULL },
      },
      "weave"
   },
   {
      BEETLE_OPT(frame_duping),
      BEETLE_OPT_FRAME_DUPING_LABEL_HU,
      NULL,
      BEETLE_OPT_FRAME_DUPING_INFO_0_HU,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(pal_video_timing_override),
      BEETLE_OPT_PAL_VIDEO_TIMING_OVERRIDE_LABEL_HU,
      NULL,
      BEETLE_OPT_PAL_VIDEO_TIMING_OVERRIDE_INFO_0_HU,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(aspect_ratio),
      BEETLE_OPT_ASPECT_RATIO_LABEL_HU,
      NULL,
      BEETLE_OPT_ASPECT_RATIO_INFO_0_HU,
      NULL,
      "video",
      {
         { "corrected", OPTION_VAL_CORRECTED_HU },
         { "uncorrected", OPTION_VAL_UNCORRECTED_HU },
         { "4:3",  OPTION_VAL_4_3_HU },
         { "ntsc", OPTION_VAL_NTSC_HU },
      },
      "corrected"
   },
   {
      BEETLE_OPT(crop_overscan),
      BEETLE_OPT_CROP_OVERSCAN_LABEL_HU,
      NULL,
      BEETLE_OPT_CROP_OVERSCAN_INFO_0_HU,
      NULL,
      "video",
      {
         { "disabled", "Disabled" },
         { "static",   OPTION_VAL_STATIC_HU },
         { "smart",    OPTION_VAL_SMART_HU },
         { NULL, NULL },
      },
      "smart"
   },
   {
      BEETLE_OPT(image_crop),
      BEETLE_OPT_IMAGE_CROP_LABEL_HU,
      NULL,
      BEETLE_OPT_IMAGE_CROP_INFO_0_HU,
      NULL,
      "video",
      {
         { "disabled", "0" },
         { "1px",      OPTION_VAL_1PX_HU },
         { "2px",      OPTION_VAL_2PX_HU },
         { "3px",      OPTION_VAL_3PX_HU },
         { "4px",      OPTION_VAL_4PX_HU },
         { "5px",      OPTION_VAL_5PX_HU },
         { "6px",      OPTION_VAL_6PX_HU },
         { "7px",      OPTION_VAL_7PX_HU },
         { "8px",      OPTION_VAL_8PX_HU },
         { "9px",      OPTION_VAL_9PX_HU },
         { "10px",     OPTION_VAL_10PX_HU },
         { "11px",     OPTION_VAL_11PX_HU },
         { "12px",     OPTION_VAL_12PX_HU },
         { "13px",     OPTION_VAL_13PX_HU },
         { "14px",     OPTION_VAL_14PX_HU },
         { "15px",     OPTION_VAL_15PX_HU },
         { "16px",     OPTION_VAL_16PX_HU },
         { "17px",     OPTION_VAL_17PX_HU },
         { "18px",     OPTION_VAL_18PX_HU },
         { "19px",     OPTION_VAL_19PX_HU },
         { "20px",     OPTION_VAL_20PX_HU },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(image_offset),
      BEETLE_OPT_IMAGE_OFFSET_LABEL_HU,
      NULL,
      BEETLE_OPT_IMAGE_OFFSET_INFO_0_HU,
      NULL,
      "video",
      {
         { "-12px",    OPTION_VAL_12PX_O64_HU },
         { "-11px",    OPTION_VAL_11PX_O64_HU },
         { "-10px",    OPTION_VAL_10PX_O64_HU },
         { "-9px",     OPTION_VAL_9PX_O64_HU },
         { "-8px",     OPTION_VAL_8PX_O64_HU },
         { "-7px",     OPTION_VAL_7PX_O64_HU },
         { "-6px",     OPTION_VAL_6PX_O64_HU },
         { "-5px",     OPTION_VAL_5PX_O64_HU },
         { "-4px",     OPTION_VAL_4PX_O64_HU },
         { "-3px",     OPTION_VAL_3PX_O64_HU },
         { "-2px",     OPTION_VAL_2PX_O64_HU },
         { "-1px",     OPTION_VAL_1PX_O64_HU },
         { "disabled", "0" },
         { "+1px",     OPTION_VAL_1PX_O64_0_HU },
         { "+2px",     OPTION_VAL_2PX_O64_0_HU },
         { "+3px",     OPTION_VAL_3PX_O64_0_HU },
         { "+4px",     OPTION_VAL_4PX_O64_0_HU },
         { "+5px",     OPTION_VAL_5PX_O64_0_HU },
         { "+6px",     OPTION_VAL_6PX_O64_0_HU },
         { "+7px",     OPTION_VAL_7PX_O64_0_HU },
         { "+8px",     OPTION_VAL_8PX_O64_0_HU },
         { "+9px",     OPTION_VAL_9PX_O64_0_HU },
         { "+10px",    OPTION_VAL_10PX_O64_0_HU },
         { "+11px",    OPTION_VAL_11PX_O64_0_HU },
         { "+12px",    OPTION_VAL_12PX_O64_0_HU },
         { NULL, NULL },
      },
      "disabled"
   },
#if defined(HAVE_OPENGL) || defined(HAVE_OPENGLES) || defined(HAVE_VULKAN)
   {
      BEETLE_OPT(image_offset_cycles),
      BEETLE_OPT_IMAGE_OFFSET_CYCLES_LABEL_HU,
      NULL,
      BEETLE_OPT_IMAGE_OFFSET_CYCLES_INFO_0_HU,
      NULL,
      "video",
      {
         { "-40",      NULL },
         { "-39",      NULL },
         { "-38",      NULL },
         { "-37",      NULL },
         { "-36",      NULL },
         { "-35",      NULL },
         { "-34",      NULL },
         { "-33",      NULL },
         { "-32",      NULL },
         { "-31",      NULL },
         { "-30",      NULL },
         { "-29",      NULL },
         { "-28",      NULL },
         { "-27",      NULL },
         { "-26",      NULL },
         { "-25",      NULL },
         { "-24",      NULL },
         { "-23",      NULL },
         { "-22",      NULL },
         { "-21",      NULL },
         { "-20",      NULL },
         { "-19",      NULL },
         { "-18",      NULL },
         { "-17",      NULL },
         { "-16",      NULL },
         { "-15",      NULL },
         { "-14",      NULL },
         { "-13",      NULL },
         { "-12",      NULL },
         { "-11",      NULL },
         { "-10",      NULL },
         { "-9",       NULL },
         { "-8",       NULL },
         { "-7",       NULL },
         { "-6",       NULL },
         { "-5",       NULL },
         { "-4",       NULL },
         { "-3",       NULL },
         { "-2",       NULL },
         { "-1",       NULL },
         { "0",        "0" },
         { "+1",       NULL },
         { "+2",       NULL },
         { "+3",       NULL },
         { "+4",       NULL },
         { "+5",       NULL },
         { "+6",       NULL },
         { "+7",       NULL },
         { "+8",       NULL },
         { "+9",       NULL },
         { "+10",      NULL },
         { "+11",      NULL },
         { "+12",      NULL },
         { "+13",      NULL },
         { "+14",      NULL },
         { "+15",      NULL },
         { "+16",      NULL },
         { "+17",      NULL },
         { "+18",      NULL },
         { "+19",      NULL },
         { "+20",      NULL },
         { "+21",      NULL },
         { "+22",      NULL },
         { "+23",      NULL },
         { "+24",      NULL },
         { "+25",      NULL },
         { "+26",      NULL },
         { "+27",      NULL },
         { "+28",      NULL },
         { "+29",      NULL },
         { "+30",      NULL },
         { "+31",      NULL },
         { "+32",      NULL },
         { "+33",      NULL },
         { "+34",      NULL },
         { "+35",      NULL },
         { "+36",      NULL },
         { "+37",      NULL },
         { "+38",      NULL },
         { "+39",      NULL },
         { "+40",      NULL },
         { NULL, NULL},
      },
      "0"
   },
#endif
   {
      BEETLE_OPT(initial_scanline),
      BEETLE_OPT_INITIAL_SCANLINE_LABEL_HU,
      NULL,
      BEETLE_OPT_INITIAL_SCANLINE_INFO_0_HU,
      NULL,
      "video",
      {
         { "0",  OPTION_VAL_0_O48_HU },
         { "1",  NULL },
         { "2",  NULL },
         { "3",  NULL },
         { "4",  NULL },
         { "5",  NULL },
         { "6",  NULL },
         { "7",  NULL },
         { "8",  NULL },
         { "9",  NULL },
         { "10", NULL },
         { "11", NULL },
         { "12", NULL },
         { "13", NULL },
         { "14", NULL },
         { "15", NULL },
         { "16", NULL },
         { "17", NULL },
         { "18", NULL },
         { "19", NULL },
         { "20", NULL },
         { "21", NULL },
         { "22", NULL },
         { "23", NULL },
         { "24", NULL },
         { "25", NULL },
         { "26", NULL },
         { "27", NULL },
         { "28", NULL },
         { "29", NULL },
         { "30", NULL },
         { "31", NULL },
         { "32", NULL },
         { "33", NULL },
         { "34", NULL },
         { "35", NULL },
         { "36", NULL },
         { "37", NULL },
         { "38", NULL },
         { "39", NULL },
         { "40", NULL },
         { NULL, NULL },
      },
      "0"
   },
   {
      BEETLE_OPT(last_scanline),
      BEETLE_OPT_LAST_SCANLINE_LABEL_HU,
      NULL,
      BEETLE_OPT_LAST_SCANLINE_INFO_0_HU,
      NULL,
      "video",
      {
         { "210", NULL },
         { "211", NULL },
         { "212", NULL },
         { "213", NULL },
         { "214", NULL },
         { "215", NULL },
         { "216", NULL },
         { "217", NULL },
         { "218", NULL },
         { "219", NULL },
         { "220", NULL },
         { "221", NULL },
         { "222", NULL },
         { "223", NULL },
         { "224", NULL },
         { "225", NULL },
         { "226", NULL },
         { "227", NULL },
         { "228", NULL },
         { "229", NULL },
         { "230", NULL },
         { "231", NULL },
         { "232", NULL },
         { "233", NULL },
         { "234", NULL },
         { "235", NULL },
         { "236", NULL },
         { "237", NULL },
         { "238", NULL },
         { "239", OPTION_VAL_239_HU },
         { NULL, NULL },
      },
      "239"
   },
   {
      BEETLE_OPT(initial_scanline_pal),
      BEETLE_OPT_INITIAL_SCANLINE_PAL_LABEL_HU,
      NULL,
      BEETLE_OPT_INITIAL_SCANLINE_PAL_INFO_0_HU,
      NULL,
      "video",
      {
         { "0",  OPTION_VAL_0_O48_HU },
         { "1",  NULL },
         { "2",  NULL },
         { "3",  NULL },
         { "4",  NULL },
         { "5",  NULL },
         { "6",  NULL },
         { "7",  NULL },
         { "8",  NULL },
         { "9",  NULL },
         { "10", NULL },
         { "11", NULL },
         { "12", NULL },
         { "13", NULL },
         { "14", NULL },
         { "15", NULL },
         { "16", NULL },
         { "17", NULL },
         { "18", NULL },
         { "19", NULL },
         { "20", NULL },
         { "21", NULL },
         { "22", NULL },
         { "23", NULL },
         { "24", NULL },
         { "25", NULL },
         { "26", NULL },
         { "27", NULL },
         { "28", NULL },
         { "29", NULL },
         { "30", NULL },
         { "31", NULL },
         { "32", NULL },
         { "33", NULL },
         { "34", NULL },
         { "35", NULL },
         { "36", NULL },
         { "37", NULL },
         { "38", NULL },
         { "39", NULL },
         { "40", NULL },
         { NULL, NULL },
      },
      "0"
   },
   {
      BEETLE_OPT(last_scanline_pal),
      BEETLE_OPT_LAST_SCANLINE_PAL_LABEL_HU,
      NULL,
      BEETLE_OPT_LAST_SCANLINE_PAL_INFO_0_HU,
      NULL,
      "video",
      {
         { "230", NULL },
         { "231", NULL },
         { "232", NULL },
         { "233", NULL },
         { "234", NULL },
         { "235", NULL },
         { "236", NULL },
         { "237", NULL },
         { "238", NULL },
         { "239", NULL },
         { "240", NULL },
         { "241", NULL },
         { "242", NULL },
         { "243", NULL },
         { "244", NULL },
         { "245", NULL },
         { "246", NULL },
         { "247", NULL },
         { "248", NULL },
         { "249", NULL },
         { "250", NULL },
         { "251", NULL },
         { "252", NULL },
         { "253", NULL },
         { "254", NULL },
         { "255", NULL },
         { "256", NULL },
         { "257", NULL },
         { "258", NULL },
         { "259", NULL },
         { "260", NULL },
         { "261", NULL },
         { "262", NULL },
         { "263", NULL },
         { "264", NULL },
         { "265", NULL },
         { "266", NULL },
         { "267", NULL },
         { "268", NULL },
         { "269", NULL },
         { "270", NULL },
         { "271", NULL },
         { "272", NULL },
         { "273", NULL },
         { "274", NULL },
         { "275", NULL },
         { "276", NULL },
         { "277", NULL },
         { "278", NULL },
         { "279", NULL },
         { "280", NULL },
         { "281", NULL },
         { "282", NULL },
         { "283", NULL },
         { "284", NULL },
         { "285", NULL },
         { "286", NULL },
         { "287", OPTION_VAL_287_HU },
         { NULL, NULL },
      },
      "287"
   },
   { NULL, NULL, NULL, NULL, NULL, NULL, {{0}}, NULL },
};
struct retro_core_options_v2 options_hu = {
   option_cats_hu,
   option_defs_hu
};

/* RETRO_LANGUAGE_ID */

#define CATEGORY_SYSTEM_LABEL_ID "Sistem"
#define CATEGORY_SYSTEM_INFO_0_ID NULL
#define CATEGORY_VIDEO_LABEL_ID NULL
#define CATEGORY_VIDEO_INFO_0_ID "Mengubah pengaturan rasio aspek, pangkas layar, saring video, dan pelewatan bingkai."
#define CATEGORY_PGXP_LABEL_ID "PGXP (Pipeline Alihragam Geometri Presisi)"
#define CATEGORY_PGXP_INFO_0_ID "Opsi berikut mengatur penyempurnaan yang dapat meningkatkan grafis dibanding konsol aslinya. PGXP dapat menghilangkan tekstur meleding dan isu Z-fighting."
#define CATEGORY_OSD_LABEL_ID "Tampilan Muka Layar"
#define CATEGORY_OSD_INFO_0_ID "Mengubah tampilan pemberitahuan di layar."
#define CATEGORY_INPUT_LABEL_ID "Masukan"
#define CATEGORY_INPUT_INFO_0_ID "Mengubah pengaturan light gun, tetikus, dan neGcon."
#define CATEGORY_MEMCARDS_LABEL_ID "Kartu Memori"
#define CATEGORY_MEMCARDS_INFO_0_ID NULL
#define CATEGORY_HACKS_LABEL_ID "Peretasan Emulasi"
#define CATEGORY_HACKS_INFO_0_ID NULL
#define BEETLE_OPT_CPU_FREQ_SCALE_LABEL_ID NULL
#define BEETLE_OPT_CPU_FREQ_SCALE_INFO_0_ID NULL
#define OPTION_VAL_50_ID NULL
#define OPTION_VAL_60_ID NULL
#define OPTION_VAL_70_ID NULL
#define OPTION_VAL_80_ID NULL
#define OPTION_VAL_90_ID NULL
#define OPTION_VAL_100_ID NULL
#define OPTION_VAL_110_ID NULL
#define OPTION_VAL_120_ID NULL
#define OPTION_VAL_130_ID NULL
#define OPTION_VAL_140_ID NULL
#define OPTION_VAL_150_ID NULL
#define OPTION_VAL_160_ID NULL
#define OPTION_VAL_170_ID NULL
#define OPTION_VAL_180_ID NULL
#define OPTION_VAL_190_ID NULL
#define OPTION_VAL_200_ID NULL
#define OPTION_VAL_210_ID NULL
#define OPTION_VAL_220_ID NULL
#define OPTION_VAL_230_ID NULL
#define OPTION_VAL_240_ID NULL
#define OPTION_VAL_250_ID NULL
#define OPTION_VAL_260_ID NULL
#define OPTION_VAL_270_ID NULL
#define OPTION_VAL_280_ID NULL
#define OPTION_VAL_290_ID NULL
#define OPTION_VAL_300_ID NULL
#define OPTION_VAL_310_ID NULL
#define OPTION_VAL_320_ID NULL
#define OPTION_VAL_330_ID NULL
#define OPTION_VAL_340_ID NULL
#define OPTION_VAL_350_ID NULL
#define OPTION_VAL_360_ID NULL
#define OPTION_VAL_370_ID NULL
#define OPTION_VAL_380_ID NULL
#define OPTION_VAL_390_ID NULL
#define OPTION_VAL_400_ID NULL
#define OPTION_VAL_410_ID NULL
#define OPTION_VAL_420_ID NULL
#define OPTION_VAL_430_ID NULL
#define OPTION_VAL_440_ID NULL
#define OPTION_VAL_450_ID NULL
#define OPTION_VAL_460_ID NULL
#define OPTION_VAL_470_ID NULL
#define OPTION_VAL_480_ID NULL
#define OPTION_VAL_490_ID NULL
#define OPTION_VAL_500_ID NULL
#define OPTION_VAL_510_ID NULL
#define OPTION_VAL_520_ID NULL
#define OPTION_VAL_530_ID NULL
#define OPTION_VAL_540_ID NULL
#define OPTION_VAL_550_ID NULL
#define OPTION_VAL_560_ID NULL
#define OPTION_VAL_570_ID NULL
#define OPTION_VAL_580_ID NULL
#define OPTION_VAL_590_ID NULL
#define OPTION_VAL_600_ID NULL
#define OPTION_VAL_610_ID NULL
#define OPTION_VAL_620_ID NULL
#define OPTION_VAL_630_ID NULL
#define OPTION_VAL_640_ID NULL
#define OPTION_VAL_650_ID NULL
#define OPTION_VAL_660_ID NULL
#define OPTION_VAL_670_ID NULL
#define OPTION_VAL_680_ID NULL
#define OPTION_VAL_690_ID NULL
#define OPTION_VAL_700_ID NULL
#define OPTION_VAL_710_ID NULL
#define OPTION_VAL_720_ID NULL
#define OPTION_VAL_730_ID NULL
#define OPTION_VAL_740_ID NULL
#define OPTION_VAL_750_ID NULL
#define BEETLE_OPT_GPU_OVERCLOCK_LABEL_ID NULL
#define BEETLE_OPT_GPU_OVERCLOCK_INFO_0_ID NULL
#define OPTION_VAL_1X_NATIVE_ID "1x (Lokal)"
#define OPTION_VAL_2X_ID NULL
#define OPTION_VAL_4X_ID NULL
#define OPTION_VAL_8X_ID NULL
#define OPTION_VAL_16X_ID NULL
#define OPTION_VAL_32X_ID NULL
#define BEETLE_OPT_GTE_OVERCLOCK_LABEL_ID NULL
#define BEETLE_OPT_GTE_OVERCLOCK_INFO_0_ID NULL
#define BEETLE_OPT_CD_ACCESS_METHOD_LABEL_ID NULL
#define BEETLE_OPT_CD_ACCESS_METHOD_INFO_0_ID NULL
#define OPTION_VAL_SYNC_ID NULL
#define OPTION_VAL_ASYNC_ID NULL
#define OPTION_VAL_PRECACHE_ID NULL
#define BEETLE_OPT_CD_FASTLOAD_LABEL_ID NULL
#define BEETLE_OPT_CD_FASTLOAD_INFO_0_ID NULL
#define OPTION_VAL_2X_NATIVE_ID "2x (Lokal)"
#define OPTION_VAL_6X_ID NULL
#define OPTION_VAL_10X_ID NULL
#define OPTION_VAL_12X_ID NULL
#define OPTION_VAL_14X_ID NULL
#define BEETLE_OPT_SKIP_BIOS_LABEL_ID NULL
#define BEETLE_OPT_SKIP_BIOS_INFO_0_ID NULL
#define BEETLE_OPT_OVERRIDE_BIOS_LABEL_ID NULL
#define BEETLE_OPT_OVERRIDE_BIOS_INFO_0_ID NULL
#define OPTION_VAL_PSXONPSP_ID NULL
#define OPTION_VAL_PS1_ROM_ID NULL
#define OPTION_VAL_OPENBIOS_ID NULL
#define BEETLE_OPT_REGION_LABEL_ID NULL
#define BEETLE_OPT_REGION_INFO_0_ID NULL
#define OPTION_VAL_AUTO_ID "Otomatis"
#define OPTION_VAL_NTSC_J_ID NULL
#define OPTION_VAL_NTSC_U_ID NULL
#define OPTION_VAL_PAL_ID NULL
#define BEETLE_OPT_INTERNAL_RESOLUTION_LABEL_ID "Resolusi Internal GPU"
#define BEETLE_OPT_INTERNAL_RESOLUTION_INFO_0_ID NULL
#define BEETLE_OPT_RENDERER_LABEL_ID NULL
#define BEETLE_OPT_RENDERER_INFO_0_ID NULL
#define OPTION_VAL_HARDWARE_ID "Peranti Keras (Auto)"
#define OPTION_VAL_HARDWARE_GL_ID "Peranti Keras (OpenGL)"
#define OPTION_VAL_HARDWARE_VK_ID "Peranti Keras (Vulkan)"
#define OPTION_VAL_SOFTWARE_ID NULL
#define BEETLE_OPT_RENDERER_SOFTWARE_FB_LABEL_ID "Dapar Bingkai Digital"
#define BEETLE_OPT_RENDERER_SOFTWARE_FB_INFO_0_ID NULL
#define BEETLE_OPT_DEPTH_LABEL_ID "Kedalaman Warna Internal"
#define BEETLE_OPT_DEPTH_INFO_0_ID NULL
#define OPTION_VAL_16BPP_NATIVE_ID NULL
#define OPTION_VAL_32BPP_ID NULL
#define BEETLE_OPT_DITHER_MODE_LABEL_ID "Pola Baur Galat"
#define BEETLE_OPT_DITHER_MODE_INFO_0_ID NULL
#define OPTION_VAL_INTERNAL_RESOLUTION_ID "Resolusi Internal"
#define BEETLE_OPT_SCALED_UV_OFFSET_LABEL_ID "Kompensasi Tekstur UV"
#define BEETLE_OPT_SCALED_UV_OFFSET_INFO_0_ID "Sampel tekstur untuk poligon 3D dengan kompensasi lebih tinggi dari resolusi internal 1x. Mengurangi jahitan tekstur tetapi dapat menyebabkan cacat grafis yang tidak diinginkan."
#define BEETLE_OPT_FILTER_LABEL_ID "Penyaringan Tekstur"
#define BEETLE_OPT_FILTER_INFO_0_ID NULL
#define OPTION_VAL_NEAREST_ID NULL
#define OPTION_VAL_SABR_ID NULL
#define OPTION_VAL_XBR_ID NULL
#define OPTION_VAL_BILINEAR_ID NULL
#define OPTION_VAL_3_POINT_ID NULL
#define OPTION_VAL_JINC2_ID NULL
#define BEETLE_OPT_FILTER_EXCLUDE_SPRITE_LABEL_ID "Lepas Penyaringan Sprite"
#define BEETLE_OPT_FILTER_EXCLUDE_SPRITE_INFO_0_ID "Tidak menerapkan saring tekstur ke sprite. Mencegah sprite 2D pecah-pecah pada hasil render latar belakang di banyak permainan. Gunakan dengan Penghalusan Adaptif atau saring pasca-pengolahan lain untuk efek terbaik."
#define OPTION_VAL_OPAQUE_ID "Hanya Buram"
#define OPTION_VAL_ALL_ID "Buram dan Setengah Tembus Pandang"
#define BEETLE_OPT_FILTER_EXCLUDE_2D_POLYGON_LABEL_ID "Lepas Penyaringan Poligon 2D"
#define BEETLE_OPT_FILTER_EXCLUDE_2D_POLYGON_INFO_0_ID "Tidak menerapkan penyaringan tekstur ke poligon 2D. Poligon 2D terdapat heuristik dan kemungkinan akan 'glitch'. Gunakan dengan Penghalusan Adaptif atau saring pasca-pengolahan lain untuk efek terbaik."
#define BEETLE_OPT_ADAPTIVE_SMOOTHING_LABEL_ID "Penghalusan Adaptif"
#define BEETLE_OPT_ADAPTIVE_SMOOTHING_INFO_0_ID "Menghaluskan gambar 2D dan unsur UI tanpa memburamkan objek 3D yang di-render. Hanya didukung perender Vulkan."
#define BEETLE_OPT_SUPER_SAMPLING_LABEL_ID "Supersampling (Downsample ke Resolusi Native)"
#define BEETLE_OPT_SUPER_SAMPLING_INFO_0_ID "Men-'downsample' konten yang di-render dari resolusi naik skala internal ke resolusi lokal lagi. Menggabungkan ini dengan pengkalian resolusi internal lebih tinggi bisa menampilkan objek 3D pada permainan dengan objek 3D ter-'antialias' di resolusi rendah lokal. Memberikan hasil terbaik jika diterapkan ke permainan bercampuran unsur 2D dan 3D (misal: karakter 3D di latar belakang pra-render), dan juga berfungsi dengan 'shader' CRT. Hanya didukung perender Vulkan. Catatan: 'Pola Baur Galat' lebih baik dinonaktifkan saat opsi ini aktif."
#define BEETLE_OPT_MSAA_LABEL_ID NULL
#define BEETLE_OPT_MSAA_INFO_0_ID NULL
#define OPTION_VAL_1X_ID NULL
#define BEETLE_OPT_MDEC_YUV_LABEL_ID NULL
#define BEETLE_OPT_MDEC_YUV_INFO_0_ID "Meningkatkan mutu pemutaran FMV dengan mengurangi artefak 'macroblocking' (tepi kotak/bergerigi). Hanya didukung perender Vulkan."
#define BEETLE_OPT_TRACK_TEXTURES_LABEL_ID "Lacak Tekstur"
#define BEETLE_OPT_TRACK_TEXTURES_INFO_0_ID NULL
#define BEETLE_OPT_DUMP_TEXTURES_LABEL_ID "Dump Tekstur"
#define BEETLE_OPT_DUMP_TEXTURES_INFO_0_ID "Dump tekstur yang digunakan ke <cd>-dump-tekstur/"
#define BEETLE_OPT_REPLACE_TEXTURES_LABEL_ID "Ganti Tekstur"
#define BEETLE_OPT_REPLACE_TEXTURES_INFO_0_ID "Ganti tekstur dengan versi HD dari <cd>-pengganti-tekstur/"
#define BEETLE_OPT_PGXP_MODE_LABEL_ID NULL
#define BEETLE_OPT_PGXP_MODE_INFO_0_ID NULL
#define OPTION_VAL_MEMORY_ONLY_ID NULL
#define OPTION_VAL_MEMORY_CPU_ID NULL
#define BEETLE_OPT_PGXP_2D_TOL_LABEL_ID NULL
#define BEETLE_OPT_PGXP_2D_TOL_INFO_0_ID NULL
#define OPTION_VAL_0PX_ID NULL
#define OPTION_VAL_1PX_ID NULL
#define OPTION_VAL_2PX_ID NULL
#define OPTION_VAL_3PX_ID NULL
#define OPTION_VAL_4PX_ID NULL
#define OPTION_VAL_5PX_ID NULL
#define OPTION_VAL_6PX_ID NULL
#define OPTION_VAL_7PX_ID NULL
#define OPTION_VAL_8PX_ID NULL
#define BEETLE_OPT_PGXP_NCLIP_LABEL_ID NULL
#define BEETLE_OPT_PGXP_NCLIP_INFO_0_ID NULL
#define BEETLE_OPT_PGXP_VERTEX_LABEL_ID NULL
#define BEETLE_OPT_PGXP_VERTEX_INFO_0_ID NULL
#define BEETLE_OPT_PGXP_TEXTURE_LABEL_ID NULL
#define BEETLE_OPT_PGXP_TEXTURE_INFO_0_ID NULL
#define BEETLE_OPT_DISPLAY_INTERNAL_FPS_LABEL_ID "Tampilkan FPS Internal"
#define BEETLE_OPT_DISPLAY_INTERNAL_FPS_INFO_0_ID "Tampilkan kecepatan bingkai internal dimana sistem PlayStation yang diemulasi merender konten. Catatan: Membutuhkan pemberitahuan dalam layar untuk aktif di frontend libretro."
#define BEETLE_OPT_DISPLAY_OSD_LABEL_ID NULL
#define BEETLE_OPT_DISPLAY_OSD_INFO_0_ID NULL
#define BEETLE_OPT_DISPLAY_VRAM_LABEL_ID "Tampilkan VRAM Penuh (Awakutu)"
#define BEETLE_OPT_DISPLAY_VRAM_INFO_0_ID "Memvisualkan semua VRAM konsol yang diemulasi. Hanya didukung perender peranti keras OpenGL dan Vulkan. Catatan: Ini hanya untuk awakutu, dan biasanya harus nonaktif."
#define BEETLE_OPT_ANALOG_CALIBRATION_LABEL_ID "Pengkalibrasian Sendiri Analog"
#define BEETLE_OPT_ANALOG_CALIBRATION_INFO_0_ID NULL
#define BEETLE_OPT_ANALOG_TOGGLE_LABEL_ID "Alihkan Mode Analog DualShock"
#define BEETLE_OPT_ANALOG_TOGGLE_INFO_0_ID "Saat jenis perangkat masukannya DualShock, opsi ini mengizinkan emulasi DualShock diaktifkan ke mode DIGITAL atau ANALOG seperti fisik asli. Mode juga dapat diaktifkan menggunakan kombinasi tombol yang dipilih."
#define OPTION_VAL_ENABLED_ANALOG_ID "Analog-Asali"
#define BEETLE_OPT_ANALOG_TOGGLE_COMBO_LABEL_ID "Mode Kombo Analog DualShock"
#define BEETLE_OPT_ANALOG_TOGGLE_COMBO_INFO_0_ID NULL
#define OPTION_VAL_L1_L2_R1_R2_START_SELECT_ID NULL
#define OPTION_VAL_L1_R1_SELECT_ID NULL
#define OPTION_VAL_L1_R1_START_ID NULL
#define OPTION_VAL_L1_R1_L3_ID NULL
#define OPTION_VAL_L1_R1_R3_ID NULL
#define OPTION_VAL_L2_R2_SELECT_ID NULL
#define OPTION_VAL_L2_R2_START_ID NULL
#define OPTION_VAL_L2_R2_L3_ID NULL
#define OPTION_VAL_L2_R2_R3_ID NULL
#define OPTION_VAL_L3_R3_ID NULL
#define BEETLE_OPT_ANALOG_TOGGLE_HOLD_LABEL_ID "Tahan Penundaan Mode Kombo Analog DualShock"
#define BEETLE_OPT_ANALOG_TOGGLE_HOLD_INFO_0_ID NULL
#define OPTION_VAL_0_ID "Bertunda 0 Detik"
#define OPTION_VAL_1_ID "Bertunda 1 Detik"
#define OPTION_VAL_2_ID "Bertunda 2 Detik"
#define OPTION_VAL_3_ID "Bertunda 3 Detik"
#define OPTION_VAL_4_ID "Bertunda 4 Detik"
#define OPTION_VAL_5_ID "Bertunda 5 Detik"
#define BEETLE_OPT_ENABLE_MULTITAP_PORT1_LABEL_ID NULL
#define BEETLE_OPT_ENABLE_MULTITAP_PORT1_INFO_0_ID "Aktifkan fungsionalitas multitap pada port 1."
#define BEETLE_OPT_ENABLE_MULTITAP_PORT2_LABEL_ID NULL
#define BEETLE_OPT_ENABLE_MULTITAP_PORT2_INFO_0_ID "Aktifkan fungsionalitas multitap pada port 2."
#define BEETLE_OPT_GUN_INPUT_MODE_LABEL_ID "Mode Masukan Senjata"
#define BEETLE_OPT_GUN_INPUT_MODE_INFO_0_ID NULL
#define OPTION_VAL_LIGHTGUN_ID NULL
#define OPTION_VAL_TOUCHSCREEN_ID "Layar Sentuh"
#define BEETLE_OPT_GUN_CURSOR_LABEL_ID "Kursor Senjata"
#define BEETLE_OPT_GUN_CURSOR_INFO_0_ID NULL
#define OPTION_VAL_CROSS_ID "Silang"
#define OPTION_VAL_DOT_ID NULL
#define BEETLE_OPT_CROSSHAIR_COLOR_P1_LABEL_ID NULL
#define BEETLE_OPT_CROSSHAIR_COLOR_P1_INFO_0_ID NULL
#define OPTION_VAL_RED_ID "Merah"
#define OPTION_VAL_BLUE_ID "Biru"
#define OPTION_VAL_GREEN_ID "Hijau"
#define OPTION_VAL_ORANGE_ID "Oranye"
#define OPTION_VAL_YELLOW_ID "Kuning"
#define OPTION_VAL_CYAN_ID NULL
#define OPTION_VAL_PINK_ID "Merah Muda"
#define OPTION_VAL_PURPLE_ID "Ungu"
#define OPTION_VAL_BLACK_ID "Hitam"
#define OPTION_VAL_WHITE_ID "Putih"
#define BEETLE_OPT_CROSSHAIR_COLOR_P2_LABEL_ID NULL
#define BEETLE_OPT_CROSSHAIR_COLOR_P2_INFO_0_ID NULL
#define BEETLE_OPT_MOUSE_SENSITIVITY_LABEL_ID "Sensivitas Mouse"
#define BEETLE_OPT_MOUSE_SENSITIVITY_INFO_0_ID NULL
#define OPTION_VAL_5_O42_ID NULL
#define OPTION_VAL_10_ID NULL
#define OPTION_VAL_15_ID NULL
#define OPTION_VAL_20_ID NULL
#define OPTION_VAL_25_ID NULL
#define OPTION_VAL_30_ID NULL
#define OPTION_VAL_35_ID NULL
#define OPTION_VAL_40_ID NULL
#define OPTION_VAL_45_ID NULL
#define OPTION_VAL_55_ID NULL
#define OPTION_VAL_65_ID NULL
#define OPTION_VAL_75_ID NULL
#define OPTION_VAL_85_ID NULL
#define OPTION_VAL_95_ID NULL
#define OPTION_VAL_100_O42_ID NULL
#define OPTION_VAL_105_ID NULL
#define OPTION_VAL_115_ID NULL
#define OPTION_VAL_125_ID NULL
#define OPTION_VAL_135_ID NULL
#define OPTION_VAL_145_ID NULL
#define OPTION_VAL_155_ID NULL
#define OPTION_VAL_165_ID NULL
#define OPTION_VAL_175_ID NULL
#define OPTION_VAL_185_ID NULL
#define OPTION_VAL_195_ID NULL
#define BEETLE_OPT_NEGCON_RESPONSE_LABEL_ID "Respon Putaran neGcon"
#define BEETLE_OPT_NEGCON_RESPONSE_INFO_0_ID NULL
#define OPTION_VAL_LINEAR_ID NULL
#define OPTION_VAL_QUADRATIC_ID NULL
#define OPTION_VAL_CUBIC_ID NULL
#define BEETLE_OPT_NEGCON_DEADZONE_LABEL_ID NULL
#define BEETLE_OPT_NEGCON_DEADZONE_INFO_0_ID NULL
#define OPTION_VAL_0_O44_ID NULL
#define BEETLE_OPT_USE_MEDNAFEN_MEMCARD0_METHOD_LABEL_ID NULL
#define BEETLE_OPT_USE_MEDNAFEN_MEMCARD0_METHOD_INFO_0_ID NULL
#define OPTION_VAL_LIBRETRO_ID NULL
#define OPTION_VAL_MEDNAFEN_ID NULL
#define BEETLE_OPT_ENABLE_MEMCARD1_LABEL_ID NULL
#define BEETLE_OPT_ENABLE_MEMCARD1_INFO_0_ID NULL
#define BEETLE_OPT_SHARED_MEMORY_CARDS_LABEL_ID NULL
#define BEETLE_OPT_SHARED_MEMORY_CARDS_INFO_0_ID NULL
#define BEETLE_OPT_MEMCARD_LEFT_INDEX_LABEL_ID NULL
#define BEETLE_OPT_MEMCARD_LEFT_INDEX_INFO_0_ID NULL
#define OPTION_VAL_0_O48_ID NULL
#define BEETLE_OPT_MEMCARD_RIGHT_INDEX_LABEL_ID NULL
#define BEETLE_OPT_MEMCARD_RIGHT_INDEX_INFO_0_ID NULL
#define OPTION_VAL_1_O49_ID NULL
#define BEETLE_OPT_LINE_RENDER_LABEL_ID NULL
#define BEETLE_OPT_LINE_RENDER_INFO_0_ID NULL
#define OPTION_VAL_DEFAULT_ID NULL
#define OPTION_VAL_AGGRESSIVE_ID NULL
#define BEETLE_OPT_WIDESCREEN_HACK_LABEL_ID NULL
#define BEETLE_OPT_WIDESCREEN_HACK_INFO_0_ID NULL
#define BEETLE_OPT_WIDESCREEN_HACK_ASPECT_RATIO_LABEL_ID NULL
#define BEETLE_OPT_WIDESCREEN_HACK_ASPECT_RATIO_INFO_0_ID NULL
#define OPTION_VAL_16_9_ID NULL
#define OPTION_VAL_16_10_ID NULL
#define OPTION_VAL_18_9_ID NULL
#define OPTION_VAL_19_9_ID NULL
#define OPTION_VAL_20_9_ID NULL
#define OPTION_VAL_21_9_ID NULL
#define OPTION_VAL_32_9_ID NULL
#define BEETLE_OPT_CPU_DYNAREC_LABEL_ID NULL
#define BEETLE_OPT_CPU_DYNAREC_INFO_0_ID NULL
#define OPTION_VAL_DISABLED_ID NULL
#define OPTION_VAL_EXECUTE_ID NULL
#define OPTION_VAL_EXECUTE_ONE_ID "Periksa Pewaktuan Daur"
#define OPTION_VAL_RUN_INTERPRETER_ID NULL
#define BEETLE_OPT_DYNAREC_INVALIDATE_LABEL_ID NULL
#define BEETLE_OPT_DYNAREC_INVALIDATE_INFO_0_ID NULL
#define OPTION_VAL_FULL_ID NULL
#define OPTION_VAL_DMA_ID NULL
#define BEETLE_OPT_DYNAREC_EVENTCYCLES_LABEL_ID NULL
#define BEETLE_OPT_DYNAREC_EVENTCYCLES_INFO_0_ID NULL
#define OPTION_VAL_128_ID NULL
#define BEETLE_OPT_DYNAREC_SPU_SAMPLES_LABEL_ID NULL
#define BEETLE_OPT_DYNAREC_SPU_SAMPLES_INFO_0_ID NULL
#define BEETLE_OPT_CORE_TIMING_FPS_LABEL_ID NULL
#define BEETLE_OPT_CORE_TIMING_FPS_INFO_0_ID NULL
#define OPTION_VAL_AUTO_TOGGLE_ID NULL
#define OPTION_VAL_FORCE_PROGRESSIVE_ID NULL
#define OPTION_VAL_FORCE_INTERLACED_ID NULL
#define BEETLE_OPT_DEINTERLACER_LABEL_ID NULL
#define BEETLE_OPT_DEINTERLACER_INFO_0_ID NULL
#define OPTION_VAL_WEAVE_ID NULL
#define OPTION_VAL_BOB_ID NULL
#define OPTION_VAL_BOB_OFFSET_ID NULL
#define OPTION_VAL_FASTMAD_ID NULL
#define BEETLE_OPT_FRAME_DUPING_LABEL_ID NULL
#define BEETLE_OPT_FRAME_DUPING_INFO_0_ID NULL
#define BEETLE_OPT_PAL_VIDEO_TIMING_OVERRIDE_LABEL_ID NULL
#define BEETLE_OPT_PAL_VIDEO_TIMING_OVERRIDE_INFO_0_ID NULL
#define BEETLE_OPT_ASPECT_RATIO_LABEL_ID NULL
#define BEETLE_OPT_ASPECT_RATIO_INFO_0_ID NULL
#define OPTION_VAL_CORRECTED_ID NULL
#define OPTION_VAL_UNCORRECTED_ID NULL
#define OPTION_VAL_4_3_ID NULL
#define OPTION_VAL_NTSC_ID NULL
#define BEETLE_OPT_CROP_OVERSCAN_LABEL_ID "Pangkas Lebihan Pindai"
#define BEETLE_OPT_CROP_OVERSCAN_INFO_0_ID NULL
#define OPTION_VAL_STATIC_ID NULL
#define OPTION_VAL_SMART_ID NULL
#define BEETLE_OPT_IMAGE_CROP_LABEL_ID "Pangkas Tambahan"
#define BEETLE_OPT_IMAGE_CROP_INFO_0_ID NULL
#define OPTION_VAL_9PX_ID NULL
#define OPTION_VAL_10PX_ID NULL
#define OPTION_VAL_11PX_ID NULL
#define OPTION_VAL_12PX_ID NULL
#define OPTION_VAL_13PX_ID NULL
#define OPTION_VAL_14PX_ID NULL
#define OPTION_VAL_15PX_ID NULL
#define OPTION_VAL_16PX_ID NULL
#define OPTION_VAL_17PX_ID NULL
#define OPTION_VAL_18PX_ID NULL
#define OPTION_VAL_19PX_ID NULL
#define OPTION_VAL_20PX_ID NULL
#define BEETLE_OPT_IMAGE_OFFSET_LABEL_ID NULL
#define BEETLE_OPT_IMAGE_OFFSET_INFO_0_ID NULL
#define OPTION_VAL_12PX_O64_ID NULL
#define OPTION_VAL_11PX_O64_ID NULL
#define OPTION_VAL_10PX_O64_ID NULL
#define OPTION_VAL_9PX_O64_ID NULL
#define OPTION_VAL_8PX_O64_ID NULL
#define OPTION_VAL_7PX_O64_ID NULL
#define OPTION_VAL_6PX_O64_ID NULL
#define OPTION_VAL_5PX_O64_ID NULL
#define OPTION_VAL_4PX_O64_ID NULL
#define OPTION_VAL_3PX_O64_ID NULL
#define OPTION_VAL_2PX_O64_ID NULL
#define OPTION_VAL_1PX_O64_ID NULL
#define OPTION_VAL_1PX_O64_0_ID NULL
#define OPTION_VAL_2PX_O64_0_ID NULL
#define OPTION_VAL_3PX_O64_0_ID NULL
#define OPTION_VAL_4PX_O64_0_ID NULL
#define OPTION_VAL_5PX_O64_0_ID NULL
#define OPTION_VAL_6PX_O64_0_ID NULL
#define OPTION_VAL_7PX_O64_0_ID NULL
#define OPTION_VAL_8PX_O64_0_ID NULL
#define OPTION_VAL_9PX_O64_0_ID NULL
#define OPTION_VAL_10PX_O64_0_ID NULL
#define OPTION_VAL_11PX_O64_0_ID NULL
#define OPTION_VAL_12PX_O64_0_ID NULL
#define BEETLE_OPT_IMAGE_OFFSET_CYCLES_LABEL_ID NULL
#define BEETLE_OPT_IMAGE_OFFSET_CYCLES_INFO_0_ID NULL
#define BEETLE_OPT_INITIAL_SCANLINE_LABEL_ID NULL
#define BEETLE_OPT_INITIAL_SCANLINE_INFO_0_ID NULL
#define BEETLE_OPT_LAST_SCANLINE_LABEL_ID NULL
#define BEETLE_OPT_LAST_SCANLINE_INFO_0_ID NULL
#define OPTION_VAL_239_ID NULL
#define BEETLE_OPT_INITIAL_SCANLINE_PAL_LABEL_ID NULL
#define BEETLE_OPT_INITIAL_SCANLINE_PAL_INFO_0_ID NULL
#define BEETLE_OPT_LAST_SCANLINE_PAL_LABEL_ID NULL
#define BEETLE_OPT_LAST_SCANLINE_PAL_INFO_0_ID NULL
#define OPTION_VAL_287_ID NULL

struct retro_core_option_v2_category option_cats_id[] = {
   {
      "system",
      CATEGORY_SYSTEM_LABEL_ID,
      CATEGORY_SYSTEM_INFO_0_ID
   },
   {
      "video",
      CATEGORY_VIDEO_LABEL_ID,
      CATEGORY_VIDEO_INFO_0_ID
   },
   {
      "pgxp",
      CATEGORY_PGXP_LABEL_ID,
      CATEGORY_PGXP_INFO_0_ID
   },
   {
      "osd",
      CATEGORY_OSD_LABEL_ID,
      CATEGORY_OSD_INFO_0_ID
   },
   {
      "input",
      CATEGORY_INPUT_LABEL_ID,
      CATEGORY_INPUT_INFO_0_ID
   },
   {
      "memcards",
      CATEGORY_MEMCARDS_LABEL_ID,
      CATEGORY_MEMCARDS_INFO_0_ID
   },
   {
      "hacks",
      CATEGORY_HACKS_LABEL_ID,
      CATEGORY_HACKS_INFO_0_ID
   },
   { NULL, NULL, NULL },
};
struct retro_core_option_v2_definition option_defs_id[] = {
   {
      BEETLE_OPT(cpu_freq_scale),
      BEETLE_OPT_CPU_FREQ_SCALE_LABEL_ID,
      NULL,
      BEETLE_OPT_CPU_FREQ_SCALE_INFO_0_ID,
      NULL,
      "system",
      {
         { "50%",           OPTION_VAL_50_ID },
         { "60%",           OPTION_VAL_60_ID },
         { "70%",           OPTION_VAL_70_ID },
         { "80%",           OPTION_VAL_80_ID },
         { "90%",           OPTION_VAL_90_ID },
         { "100%",          OPTION_VAL_100_ID },
         { "110%",          OPTION_VAL_110_ID },
         { "120%",          OPTION_VAL_120_ID },
         { "130%",          OPTION_VAL_130_ID },
         { "140%",          OPTION_VAL_140_ID },
         { "150%",          OPTION_VAL_150_ID },
         { "160%",          OPTION_VAL_160_ID },
         { "170%",          OPTION_VAL_170_ID },
         { "180%",          OPTION_VAL_180_ID },
         { "190%",          OPTION_VAL_190_ID },
         { "200%",          OPTION_VAL_200_ID },
         { "210%",          OPTION_VAL_210_ID },
         { "220%",          OPTION_VAL_220_ID },
         { "230%",          OPTION_VAL_230_ID },
         { "240%",          OPTION_VAL_240_ID },
         { "250%",          OPTION_VAL_250_ID },
         { "260%",          OPTION_VAL_260_ID },
         { "270%",          OPTION_VAL_270_ID },
         { "280%",          OPTION_VAL_280_ID },
         { "290%",          OPTION_VAL_290_ID },
         { "300%",          OPTION_VAL_300_ID },
         { "310%",          OPTION_VAL_310_ID },
         { "320%",          OPTION_VAL_320_ID },
         { "330%",          OPTION_VAL_330_ID },
         { "340%",          OPTION_VAL_340_ID },
         { "350%",          OPTION_VAL_350_ID },
         { "360%",          OPTION_VAL_360_ID },
         { "370%",          OPTION_VAL_370_ID },
         { "380%",          OPTION_VAL_380_ID },
         { "390%",          OPTION_VAL_390_ID },
         { "400%",          OPTION_VAL_400_ID },
         { "410%",          OPTION_VAL_410_ID },
         { "420%",          OPTION_VAL_420_ID },
         { "430%",          OPTION_VAL_430_ID },
         { "440%",          OPTION_VAL_440_ID },
         { "450%",          OPTION_VAL_450_ID },
         { "460%",          OPTION_VAL_460_ID },
         { "470%",          OPTION_VAL_470_ID },
         { "480%",          OPTION_VAL_480_ID },
         { "490%",          OPTION_VAL_490_ID },
         { "500%",          OPTION_VAL_500_ID },
         { "510%",          OPTION_VAL_510_ID },
         { "520%",          OPTION_VAL_520_ID },
         { "530%",          OPTION_VAL_530_ID },
         { "540%",          OPTION_VAL_540_ID },
         { "550%",          OPTION_VAL_550_ID },
         { "560%",          OPTION_VAL_560_ID },
         { "570%",          OPTION_VAL_570_ID },
         { "580%",          OPTION_VAL_580_ID },
         { "590%",          OPTION_VAL_590_ID },
         { "600%",          OPTION_VAL_600_ID },
         { "610%",          OPTION_VAL_610_ID },
         { "620%",          OPTION_VAL_620_ID },
         { "630%",          OPTION_VAL_630_ID },
         { "640%",          OPTION_VAL_640_ID },
         { "650%",          OPTION_VAL_650_ID },
         { "660%",          OPTION_VAL_660_ID },
         { "670%",          OPTION_VAL_670_ID },
         { "680%",          OPTION_VAL_680_ID },
         { "690%",          OPTION_VAL_690_ID },
         { "700%",          OPTION_VAL_700_ID },
         { "710%",          OPTION_VAL_710_ID },
         { "720%",          OPTION_VAL_720_ID },
         { "730%",          OPTION_VAL_730_ID },
         { "740%",          OPTION_VAL_740_ID },
         { "750%",          OPTION_VAL_750_ID },
         { NULL, NULL },
      },
      "100%"
   },
   {
      BEETLE_OPT(gpu_overclock),
      BEETLE_OPT_GPU_OVERCLOCK_LABEL_ID,
      NULL,
      BEETLE_OPT_GPU_OVERCLOCK_INFO_0_ID,
      NULL,
      "system",
      {
         { "1x(native)", OPTION_VAL_1X_NATIVE_ID },
         { "2x",         OPTION_VAL_2X_ID },
         { "4x",         OPTION_VAL_4X_ID },
         { "8x",         OPTION_VAL_8X_ID },
         { "16x",        OPTION_VAL_16X_ID },
         { "32x",        OPTION_VAL_32X_ID },
         { NULL, NULL },
      },
      "1x(native)"
   },
   {
      BEETLE_OPT(gte_overclock),
      BEETLE_OPT_GTE_OVERCLOCK_LABEL_ID,
      NULL,
      BEETLE_OPT_GTE_OVERCLOCK_INFO_0_ID,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
#ifndef EMSCRIPTEN
   {
      BEETLE_OPT(cd_access_method),
      BEETLE_OPT_CD_ACCESS_METHOD_LABEL_ID,
      NULL,
      BEETLE_OPT_CD_ACCESS_METHOD_INFO_0_ID,
      NULL,
      "system",
      {
         { "sync",     OPTION_VAL_SYNC_ID },
         { "async",    OPTION_VAL_ASYNC_ID },
         { "precache", OPTION_VAL_PRECACHE_ID },
         { NULL, NULL },
      },
      "sync"
   },
#endif
   {
      BEETLE_OPT(cd_fastload),
      BEETLE_OPT_CD_FASTLOAD_LABEL_ID,
      NULL,
      BEETLE_OPT_CD_FASTLOAD_INFO_0_ID,
      NULL,
      "system",
      {
         { "2x(native)", OPTION_VAL_2X_NATIVE_ID },
         { "4x",          OPTION_VAL_4X_ID },
         { "6x",          OPTION_VAL_6X_ID },
         { "8x",          OPTION_VAL_8X_ID },
         { "10x",         OPTION_VAL_10X_ID },
         { "12x",         OPTION_VAL_12X_ID },
         { "14x",         OPTION_VAL_14X_ID },
         { NULL, NULL },
      },
      "2x(native)"
   },
   {
      BEETLE_OPT(skip_bios),
      BEETLE_OPT_SKIP_BIOS_LABEL_ID,
      NULL,
      BEETLE_OPT_SKIP_BIOS_INFO_0_ID,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(override_bios),
      BEETLE_OPT_OVERRIDE_BIOS_LABEL_ID,
      NULL,
      BEETLE_OPT_OVERRIDE_BIOS_INFO_0_ID,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "psxonpsp", OPTION_VAL_PSXONPSP_ID },
         { "ps1_rom",  OPTION_VAL_PS1_ROM_ID },
         { "openbios", OPTION_VAL_OPENBIOS_ID     },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(region),
      BEETLE_OPT_REGION_LABEL_ID,
      NULL,
      BEETLE_OPT_REGION_INFO_0_ID,
      NULL,
      "system",
      {
         { "auto",   OPTION_VAL_AUTO_ID },
         { "ntsc-j", OPTION_VAL_NTSC_J_ID },
         { "ntsc-u", OPTION_VAL_NTSC_U_ID },
         { "pal",    OPTION_VAL_PAL_ID },
         { NULL, NULL },
      },
      "auto"
   },
   {
      BEETLE_OPT(internal_resolution),
      BEETLE_OPT_INTERNAL_RESOLUTION_LABEL_ID,
      NULL,
      BEETLE_OPT_INTERNAL_RESOLUTION_INFO_0_ID,
      NULL,
      "video",
      {
         { "1x(native)", OPTION_VAL_1X_NATIVE_ID },
         { "2x",         OPTION_VAL_2X_ID },
         { "4x",         OPTION_VAL_4X_ID },
         { "8x",         OPTION_VAL_8X_ID },
         { "16x",        OPTION_VAL_16X_ID },
         { NULL, NULL },
      },
      "1x(native)"
   },
#if defined(HAVE_OPENGL) || defined(HAVE_OPENGLES) || defined(HAVE_VULKAN)
   {
      BEETLE_OPT(renderer),
      BEETLE_OPT_RENDERER_LABEL_ID,
      NULL,
      BEETLE_OPT_RENDERER_INFO_0_ID,
      NULL,
      "video",
      {
         { "hardware",    OPTION_VAL_HARDWARE_ID },
#if defined(HAVE_OPENGL) || defined(HAVE_OPENGLES)
         { "hardware_gl", OPTION_VAL_HARDWARE_GL_ID },
#endif
#if defined(HAVE_VULKAN)
         { "hardware_vk", OPTION_VAL_HARDWARE_VK_ID },
#endif
         { "software",    OPTION_VAL_SOFTWARE_ID },
         { NULL, NULL },
      },
      "hardware"
   },
   {
      BEETLE_OPT(renderer_software_fb),
      BEETLE_OPT_RENDERER_SOFTWARE_FB_LABEL_ID,
      NULL,
      BEETLE_OPT_RENDERER_SOFTWARE_FB_INFO_0_ID,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled"
   },
#endif
#if defined(HAVE_OPENGL) || defined(HAVE_OPENGLES)
   {
      BEETLE_OPT(depth),
      BEETLE_OPT_DEPTH_LABEL_ID,
      NULL,
      BEETLE_OPT_DEPTH_INFO_0_ID,
      NULL,
      "video",
      {
         { "16bpp(native)", OPTION_VAL_16BPP_NATIVE_ID },
         { "32bpp",         OPTION_VAL_32BPP_ID },
         { NULL, NULL },
      },
      "16bpp(native)"
   },
   // Sort of, it's more like 15-bit high color and 24-bit true color for visible output. The alpha channel is used for mask bit. Vulkan renderer uses ABGR1555_555 for 31 bits internal? FMVs are always 24-bit on all renderers like original hardware (BGR888, no alpha)
#endif
   {
      BEETLE_OPT(dither_mode),
      BEETLE_OPT_DITHER_MODE_LABEL_ID,
      NULL,
      BEETLE_OPT_DITHER_MODE_INFO_0_ID,
      NULL,
      "video",
      {
         { "1x(native)",          OPTION_VAL_1X_NATIVE_ID },
         { "internal resolution", OPTION_VAL_INTERNAL_RESOLUTION_ID },
         { "disabled",            NULL },
         { NULL, NULL },
      },
      "1x(native)"
   },
#ifdef HAVE_VULKAN
   {
      BEETLE_OPT(scaled_uv_offset),
      BEETLE_OPT_SCALED_UV_OFFSET_LABEL_ID,
      NULL,
      BEETLE_OPT_SCALED_UV_OFFSET_INFO_0_ID,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled"
   },
#endif
#if defined(HAVE_OPENGL) || defined(HAVE_OPENGLES) || defined(HAVE_VULKAN)
   {
      BEETLE_OPT(filter),
      BEETLE_OPT_FILTER_LABEL_ID,
      NULL,
      BEETLE_OPT_FILTER_INFO_0_ID,
      NULL,
      "video",
      {
         { "nearest",  OPTION_VAL_NEAREST_ID },
         { "SABR",     OPTION_VAL_SABR_ID },
         { "xBR",      OPTION_VAL_XBR_ID },
         { "bilinear", OPTION_VAL_BILINEAR_ID },
         { "3-point",  OPTION_VAL_3_POINT_ID },
         { "JINC2",    OPTION_VAL_JINC2_ID },
         { NULL, NULL },
      },
      "nearest"
   },
#ifdef HAVE_VULKAN
   {
      BEETLE_OPT(filter_exclude_sprite),
      BEETLE_OPT_FILTER_EXCLUDE_SPRITE_LABEL_ID,
      NULL,
      BEETLE_OPT_FILTER_EXCLUDE_SPRITE_INFO_0_ID,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "opaque", OPTION_VAL_OPAQUE_ID },
         { "all", OPTION_VAL_ALL_ID },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(filter_exclude_2d_polygon),
      BEETLE_OPT_FILTER_EXCLUDE_2D_POLYGON_LABEL_ID,
      NULL,
      BEETLE_OPT_FILTER_EXCLUDE_2D_POLYGON_INFO_0_ID,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "opaque", OPTION_VAL_OPAQUE_ID },
         { "all", OPTION_VAL_ALL_ID },
         { NULL, NULL },
      },
      "disabled"
   },
#endif
#endif
#ifdef HAVE_VULKAN
   {
      BEETLE_OPT(adaptive_smoothing),
      BEETLE_OPT_ADAPTIVE_SMOOTHING_LABEL_ID,
      NULL,
      BEETLE_OPT_ADAPTIVE_SMOOTHING_INFO_0_ID,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(super_sampling),
      BEETLE_OPT_SUPER_SAMPLING_LABEL_ID,
      NULL,
      BEETLE_OPT_SUPER_SAMPLING_INFO_0_ID,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(msaa),
      BEETLE_OPT_MSAA_LABEL_ID,
      NULL,
      BEETLE_OPT_MSAA_INFO_0_ID,
      NULL,
      "video",
      {
         { "1x",  OPTION_VAL_1X_ID },
         { "2x",  OPTION_VAL_2X_ID },
         { "4x",  OPTION_VAL_4X_ID },
         { "8x",  OPTION_VAL_8X_ID },
         { "16x", OPTION_VAL_16X_ID },
         { NULL, NULL },
      },
      "1x"
   },
   {
      BEETLE_OPT(mdec_yuv),
      BEETLE_OPT_MDEC_YUV_LABEL_ID,
      NULL,
      BEETLE_OPT_MDEC_YUV_INFO_0_ID,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(track_textures),
      BEETLE_OPT_TRACK_TEXTURES_LABEL_ID,
      NULL,
      BEETLE_OPT_TRACK_TEXTURES_INFO_0_ID,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
#ifdef TEXTURE_DUMPING_ENABLED
   {
      BEETLE_OPT(dump_textures),
      BEETLE_OPT_DUMP_TEXTURES_LABEL_ID,
      NULL,
      BEETLE_OPT_DUMP_TEXTURES_INFO_0_ID,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
#endif
   {
      BEETLE_OPT(replace_textures),
      BEETLE_OPT_REPLACE_TEXTURES_LABEL_ID,
      NULL,
      BEETLE_OPT_REPLACE_TEXTURES_INFO_0_ID,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
#endif
   {
      BEETLE_OPT(pgxp_mode),
      BEETLE_OPT_PGXP_MODE_LABEL_ID,
      NULL,
      BEETLE_OPT_PGXP_MODE_INFO_0_ID,
      NULL,
      "pgxp",
      {
         { "disabled",     NULL },
         { "memory only",  OPTION_VAL_MEMORY_ONLY_ID },
         { "memory + CPU", OPTION_VAL_MEMORY_CPU_ID },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(pgxp_2d_tol),
      BEETLE_OPT_PGXP_2D_TOL_LABEL_ID,
      NULL,
      BEETLE_OPT_PGXP_2D_TOL_INFO_0_ID,
      NULL,
      "pgxp",
      {
         { "disabled", NULL },
         { "0px", OPTION_VAL_0PX_ID },
         { "1px", OPTION_VAL_1PX_ID },
         { "2px", OPTION_VAL_2PX_ID },
         { "3px", OPTION_VAL_3PX_ID },
         { "4px", OPTION_VAL_4PX_ID },
         { "5px", OPTION_VAL_5PX_ID },
         { "6px", OPTION_VAL_6PX_ID },
         { "7px", OPTION_VAL_7PX_ID },
         { "8px", OPTION_VAL_8PX_ID },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(pgxp_nclip),
      BEETLE_OPT_PGXP_NCLIP_LABEL_ID,
      NULL,
      BEETLE_OPT_PGXP_NCLIP_INFO_0_ID,
      NULL,
      "pgxp",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
#if defined(HAVE_OPENGL) || defined(HAVE_OPENGLES) || defined(HAVE_VULKAN)
   {
      BEETLE_OPT(pgxp_vertex),
      BEETLE_OPT_PGXP_VERTEX_LABEL_ID,
      NULL,
      BEETLE_OPT_PGXP_VERTEX_INFO_0_ID,
      NULL,
      "pgxp",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
#endif
   {
      BEETLE_OPT(pgxp_texture),
      BEETLE_OPT_PGXP_TEXTURE_LABEL_ID,
      NULL,
      BEETLE_OPT_PGXP_TEXTURE_INFO_0_ID,
      NULL,
      "pgxp",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(display_internal_fps),
      BEETLE_OPT_DISPLAY_INTERNAL_FPS_LABEL_ID,
      NULL,
      BEETLE_OPT_DISPLAY_INTERNAL_FPS_INFO_0_ID,
      NULL,
      "osd",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(display_osd),
      BEETLE_OPT_DISPLAY_OSD_LABEL_ID,
      NULL,
      BEETLE_OPT_DISPLAY_OSD_INFO_0_ID,
      NULL,
      "osd",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled"
   },
#if defined(HAVE_OPENGL) || defined(HAVE_OPENGLES) || defined(HAVE_VULKAN)
   {
      BEETLE_OPT(display_vram),
      BEETLE_OPT_DISPLAY_VRAM_LABEL_ID,
      NULL,
      BEETLE_OPT_DISPLAY_VRAM_INFO_0_ID,
      NULL,
      "osd",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
#endif
   {
      BEETLE_OPT(analog_calibration),
      BEETLE_OPT_ANALOG_CALIBRATION_LABEL_ID,
      NULL,
      BEETLE_OPT_ANALOG_CALIBRATION_INFO_0_ID,
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(analog_toggle),
      BEETLE_OPT_ANALOG_TOGGLE_LABEL_ID,
      NULL,
      BEETLE_OPT_ANALOG_TOGGLE_INFO_0_ID,
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { "enabled-analog", OPTION_VAL_ENABLED_ANALOG_ID },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(analog_toggle_combo),
      BEETLE_OPT_ANALOG_TOGGLE_COMBO_LABEL_ID,
      NULL,
      BEETLE_OPT_ANALOG_TOGGLE_COMBO_INFO_0_ID,
      NULL,
      "input",
      {
         { "l1+l2+r1+r2+start+select", OPTION_VAL_L1_L2_R1_R2_START_SELECT_ID },
         { "l1+r1+select",             OPTION_VAL_L1_R1_SELECT_ID },
         { "l1+r1+start",              OPTION_VAL_L1_R1_START_ID },
         { "l1+r1+l3",                 OPTION_VAL_L1_R1_L3_ID },
         { "l1+r1+r3",                 OPTION_VAL_L1_R1_R3_ID },
         { "l2+r2+select",             OPTION_VAL_L2_R2_SELECT_ID },
         { "l2+r2+start",              OPTION_VAL_L2_R2_START_ID },
         { "l2+r2+l3",                 OPTION_VAL_L2_R2_L3_ID },
         { "l2+r2+r3",                 OPTION_VAL_L2_R2_R3_ID },
         { "l3+r3",                    OPTION_VAL_L3_R3_ID },
         { NULL, NULL },
      },
      "l1+r1+select"
   },
   {
      BEETLE_OPT(analog_toggle_hold),
      BEETLE_OPT_ANALOG_TOGGLE_HOLD_LABEL_ID,
      NULL,
      BEETLE_OPT_ANALOG_TOGGLE_HOLD_INFO_0_ID,
      NULL,
      "input",
      {
         { "0", OPTION_VAL_0_ID },
         { "1", OPTION_VAL_1_ID },
         { "2", OPTION_VAL_2_ID },
         { "3", OPTION_VAL_3_ID },
         { "4", OPTION_VAL_4_ID },
         { "5", OPTION_VAL_5_ID },
         { NULL, NULL },
      },
      "1"
   },
   {
      BEETLE_OPT(enable_multitap_port1),
      BEETLE_OPT_ENABLE_MULTITAP_PORT1_LABEL_ID,
      NULL,
      BEETLE_OPT_ENABLE_MULTITAP_PORT1_INFO_0_ID,
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(enable_multitap_port2),
      BEETLE_OPT_ENABLE_MULTITAP_PORT2_LABEL_ID,
      NULL,
      BEETLE_OPT_ENABLE_MULTITAP_PORT2_INFO_0_ID,
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(gun_input_mode),
      BEETLE_OPT_GUN_INPUT_MODE_LABEL_ID,
      NULL,
      BEETLE_OPT_GUN_INPUT_MODE_INFO_0_ID,
      NULL,
      "input",
      {
         { "lightgun",    OPTION_VAL_LIGHTGUN_ID },
         { "touchscreen", OPTION_VAL_TOUCHSCREEN_ID },
         { NULL, NULL },
      },
      "lightgun"
   },
   // Shouldn't the gun_input_mode just be Mouse vs. Touchscreen?
   {
      BEETLE_OPT(gun_cursor),
      BEETLE_OPT_GUN_CURSOR_LABEL_ID,
      NULL,
      BEETLE_OPT_GUN_CURSOR_INFO_0_ID,
      NULL,
      "input",
      {
         { "off",   "disabled" },
         { "cross", OPTION_VAL_CROSS_ID },
         { "dot",   OPTION_VAL_DOT_ID },
         { NULL, NULL },
      },
      "cross"
   },
   {
      BEETLE_OPT(crosshair_color_p1),
      BEETLE_OPT_CROSSHAIR_COLOR_P1_LABEL_ID,
      NULL,
      BEETLE_OPT_CROSSHAIR_COLOR_P1_INFO_0_ID,
      NULL,
      "input",
      {
         { "red",    OPTION_VAL_RED_ID },
         { "blue",   OPTION_VAL_BLUE_ID },
         { "green",  OPTION_VAL_GREEN_ID },
         { "orange", OPTION_VAL_ORANGE_ID },
         { "yellow", OPTION_VAL_YELLOW_ID },
         { "cyan",   OPTION_VAL_CYAN_ID },
         { "pink",   OPTION_VAL_PINK_ID },
         { "purple", OPTION_VAL_PURPLE_ID },
         { "black",  OPTION_VAL_BLACK_ID },
         { "white",  OPTION_VAL_WHITE_ID },
         { NULL, NULL },
      },
      "red"
   },
   {
      BEETLE_OPT(crosshair_color_p2),
      BEETLE_OPT_CROSSHAIR_COLOR_P2_LABEL_ID,
      NULL,
      BEETLE_OPT_CROSSHAIR_COLOR_P2_INFO_0_ID,
      NULL,
      "input",
      {
         { "blue",   OPTION_VAL_BLUE_ID },
         { "red",    OPTION_VAL_RED_ID },
         { "green",  OPTION_VAL_GREEN_ID },
         { "orange", OPTION_VAL_ORANGE_ID },
         { "yellow", OPTION_VAL_YELLOW_ID },
         { "cyan",   OPTION_VAL_CYAN_ID },
         { "pink",   OPTION_VAL_PINK_ID },
         { "purple", OPTION_VAL_PURPLE_ID },
         { "black",  OPTION_VAL_BLACK_ID },
         { "white",  OPTION_VAL_WHITE_ID },
         { NULL, NULL },
      },
      "blue"
   },
   {
      BEETLE_OPT(mouse_sensitivity),
      BEETLE_OPT_MOUSE_SENSITIVITY_LABEL_ID,
      NULL,
      BEETLE_OPT_MOUSE_SENSITIVITY_INFO_0_ID,
      NULL,
      "input",
      {
         { "5%",   OPTION_VAL_5_O42_ID },
         { "10%",  OPTION_VAL_10_ID },
         { "15%",  OPTION_VAL_15_ID },
         { "20%",  OPTION_VAL_20_ID },
         { "25%",  OPTION_VAL_25_ID },
         { "30%",  OPTION_VAL_30_ID },
         { "35%",  OPTION_VAL_35_ID },
         { "40%",  OPTION_VAL_40_ID },
         { "45%",  OPTION_VAL_45_ID },
         { "50%",  OPTION_VAL_50_ID },
         { "55%",  OPTION_VAL_55_ID },
         { "60%",  OPTION_VAL_60_ID },
         { "65%",  OPTION_VAL_65_ID },
         { "70%",  OPTION_VAL_70_ID },
         { "75%",  OPTION_VAL_75_ID },
         { "80%",  OPTION_VAL_80_ID },
         { "85%",  OPTION_VAL_85_ID },
         { "90%",  OPTION_VAL_90_ID },
         { "95%",  OPTION_VAL_95_ID },
         { "100%", OPTION_VAL_100_O42_ID },
         { "105%", OPTION_VAL_105_ID },
         { "110%", OPTION_VAL_110_ID },
         { "115%", OPTION_VAL_115_ID },
         { "120%", OPTION_VAL_120_ID },
         { "125%", OPTION_VAL_125_ID },
         { "130%", OPTION_VAL_130_ID },
         { "135%", OPTION_VAL_135_ID },
         { "140%", OPTION_VAL_140_ID },
         { "145%", OPTION_VAL_145_ID },
         { "150%", OPTION_VAL_150_ID },
         { "155%", OPTION_VAL_155_ID },
         { "160%", OPTION_VAL_160_ID },
         { "165%", OPTION_VAL_165_ID },
         { "170%", OPTION_VAL_170_ID },
         { "175%", OPTION_VAL_175_ID },
         { "180%", OPTION_VAL_180_ID },
         { "185%", OPTION_VAL_185_ID },
         { "190%", OPTION_VAL_190_ID },
         { "195%", OPTION_VAL_195_ID },
         { "200%", OPTION_VAL_200_ID },
         { NULL, NULL },
      },
      "100%"
   },
   {
      BEETLE_OPT(negcon_response),
      BEETLE_OPT_NEGCON_RESPONSE_LABEL_ID,
      NULL,
      BEETLE_OPT_NEGCON_RESPONSE_INFO_0_ID,
      NULL,
      "input",
      {
         { "linear",    OPTION_VAL_LINEAR_ID },
         { "quadratic", OPTION_VAL_QUADRATIC_ID },
         { "cubic",     OPTION_VAL_CUBIC_ID },
         { NULL, NULL },
      },
      "linear"
   },
   {
      BEETLE_OPT(negcon_deadzone),
      BEETLE_OPT_NEGCON_DEADZONE_LABEL_ID,
      NULL,
      BEETLE_OPT_NEGCON_DEADZONE_INFO_0_ID,
      NULL,
      "input",
      {
         { "0%",  OPTION_VAL_0_O44_ID },
         { "5%",  OPTION_VAL_5_O42_ID },
         { "10%", OPTION_VAL_10_ID },
         { "15%", OPTION_VAL_15_ID },
         { "20%", OPTION_VAL_20_ID },
         { "25%", OPTION_VAL_25_ID },
         { "30%", OPTION_VAL_30_ID },
         { NULL, NULL },
      },
      "0%"
   },
   {
      BEETLE_OPT(use_mednafen_memcard0_method),
      BEETLE_OPT_USE_MEDNAFEN_MEMCARD0_METHOD_LABEL_ID,
      NULL,
      BEETLE_OPT_USE_MEDNAFEN_MEMCARD0_METHOD_INFO_0_ID,
      NULL,
      "memcards",
      {
         { "libretro", OPTION_VAL_LIBRETRO_ID },
         { "mednafen", OPTION_VAL_MEDNAFEN_ID },
         { NULL, NULL },
      },
      "libretro"
   },
   {
      BEETLE_OPT(enable_memcard1),
      BEETLE_OPT_ENABLE_MEMCARD1_LABEL_ID,
      NULL,
      BEETLE_OPT_ENABLE_MEMCARD1_INFO_0_ID,
      NULL,
      "memcards",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(shared_memory_cards),
      BEETLE_OPT_SHARED_MEMORY_CARDS_LABEL_ID,
      NULL,
      BEETLE_OPT_SHARED_MEMORY_CARDS_INFO_0_ID,
      NULL,
      "memcards",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(memcard_left_index),
      BEETLE_OPT_MEMCARD_LEFT_INDEX_LABEL_ID,
      NULL,
      BEETLE_OPT_MEMCARD_LEFT_INDEX_INFO_0_ID,
      NULL,
      "memcards",
      {
         { "0",  OPTION_VAL_0_O48_ID },
         { "1",  NULL },
         { "2",  NULL },
         { "3",  NULL },
         { "4",  NULL },
         { "5",  NULL },
         { "6",  NULL },
         { "7",  NULL },
         { "8",  NULL },
         { "9",  NULL },
         { "10", NULL },
         { "11", NULL },
         { "12", NULL },
         { "13", NULL },
         { "14", NULL },
         { "15", NULL },
         { "16", NULL },
         { "17", NULL },
         { "18", NULL },
         { "19", NULL },
         { "20", NULL },
         { "21", NULL },
         { "22", NULL },
         { "23", NULL },
         { "24", NULL },
         { "25", NULL },
         { "26", NULL },
         { "27", NULL },
         { "28", NULL },
         { "29", NULL },
         { "30", NULL },
         { "31", NULL },
         { "32", NULL },
         { "33", NULL },
         { "34", NULL },
         { "35", NULL },
         { "36", NULL },
         { "37", NULL },
         { "38", NULL },
         { "39", NULL },
         { "40", NULL },
         { "41", NULL },
         { "42", NULL },
         { "43", NULL },
         { "44", NULL },
         { "45", NULL },
         { "46", NULL },
         { "47", NULL },
         { "48", NULL },
         { "49", NULL },
         { "50", NULL },
         { "51", NULL },
         { "52", NULL },
         { "53", NULL },
         { "54", NULL },
         { "55", NULL },
         { "56", NULL },
         { "57", NULL },
         { "58", NULL },
         { "59", NULL },
         { "60", NULL },
         { "61", NULL },
         { "62", NULL },
         { "63", NULL },
         { NULL, NULL },
      },
      "0"
   },
   {
      BEETLE_OPT(memcard_right_index),
      BEETLE_OPT_MEMCARD_RIGHT_INDEX_LABEL_ID,
      NULL,
      BEETLE_OPT_MEMCARD_RIGHT_INDEX_INFO_0_ID,
      NULL,
      "memcards",
      {
         { "0",  NULL },
         { "1",  OPTION_VAL_1_O49_ID },
         { "2",  NULL },
         { "3",  NULL },
         { "4",  NULL },
         { "5",  NULL },
         { "6",  NULL },
         { "7",  NULL },
         { "8",  NULL },
         { "9",  NULL },
         { "10", NULL },
         { "11", NULL },
         { "12", NULL },
         { "13", NULL },
         { "14", NULL },
         { "15", NULL },
         { "16", NULL },
         { "17", NULL },
         { "18", NULL },
         { "19", NULL },
         { "20", NULL },
         { "21", NULL },
         { "22", NULL },
         { "23", NULL },
         { "24", NULL },
         { "25", NULL },
         { "26", NULL },
         { "27", NULL },
         { "28", NULL },
         { "29", NULL },
         { "30", NULL },
         { "31", NULL },
         { "32", NULL },
         { "33", NULL },
         { "34", NULL },
         { "35", NULL },
         { "36", NULL },
         { "37", NULL },
         { "38", NULL },
         { "39", NULL },
         { "40", NULL },
         { "41", NULL },
         { "42", NULL },
         { "43", NULL },
         { "44", NULL },
         { "45", NULL },
         { "46", NULL },
         { "47", NULL },
         { "48", NULL },
         { "49", NULL },
         { "50", NULL },
         { "51", NULL },
         { "52", NULL },
         { "53", NULL },
         { "54", NULL },
         { "55", NULL },
         { "56", NULL },
         { "57", NULL },
         { "58", NULL },
         { "59", NULL },
         { "60", NULL },
         { "61", NULL },
         { "62", NULL },
         { "63", NULL },
         { NULL, NULL },
      },
      "1"
   },
   {
      BEETLE_OPT(line_render),
      BEETLE_OPT_LINE_RENDER_LABEL_ID,
      NULL,
      BEETLE_OPT_LINE_RENDER_INFO_0_ID,
      NULL,
      "hacks",
      {
         { "default",    OPTION_VAL_DEFAULT_ID },
         { "aggressive", OPTION_VAL_AGGRESSIVE_ID },
         { "disabled",   NULL },
         { NULL, NULL },
      },
      "default"
   },
   {
      BEETLE_OPT(widescreen_hack),
      BEETLE_OPT_WIDESCREEN_HACK_LABEL_ID,
      NULL,
      BEETLE_OPT_WIDESCREEN_HACK_INFO_0_ID,
      NULL,
      "hacks",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(widescreen_hack_aspect_ratio),
      BEETLE_OPT_WIDESCREEN_HACK_ASPECT_RATIO_LABEL_ID,
      NULL,
      BEETLE_OPT_WIDESCREEN_HACK_ASPECT_RATIO_INFO_0_ID,
      NULL,
      "hacks",
      {
         { "16:9",  OPTION_VAL_16_9_ID },
         { "16:10", OPTION_VAL_16_10_ID },
         { "18:9",  OPTION_VAL_18_9_ID },
         { "19:9",  OPTION_VAL_19_9_ID },
         { "20:9",  OPTION_VAL_20_9_ID },
         { "21:9",  OPTION_VAL_21_9_ID }, // 64:27
         { "32:9",  OPTION_VAL_32_9_ID },
         { NULL,    NULL },
      },
      "16:9"
   },
#if defined(HAVE_LIGHTREC)
   {
      BEETLE_OPT(cpu_dynarec),
      BEETLE_OPT_CPU_DYNAREC_LABEL_ID,
      NULL,
      BEETLE_OPT_CPU_DYNAREC_INFO_0_ID,
      NULL,
      "hacks",
      {
         { "disabled",        OPTION_VAL_DISABLED_ID },
         { "execute",         OPTION_VAL_EXECUTE_ID },
         { "execute_one",     OPTION_VAL_EXECUTE_ONE_ID },
         { "run_interpreter", OPTION_VAL_RUN_INTERPRETER_ID },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(dynarec_invalidate),
      BEETLE_OPT_DYNAREC_INVALIDATE_LABEL_ID,
      NULL,
      BEETLE_OPT_DYNAREC_INVALIDATE_INFO_0_ID,
      NULL,
      "hacks",
      {
         { "full", OPTION_VAL_FULL_ID },
         { "dma",  OPTION_VAL_DMA_ID },
         { NULL, NULL },
      },
      "full"
   },
   {
      BEETLE_OPT(dynarec_eventcycles),
      BEETLE_OPT_DYNAREC_EVENTCYCLES_LABEL_ID,
      NULL,
      BEETLE_OPT_DYNAREC_EVENTCYCLES_INFO_0_ID,
      NULL,
      "hacks",
      {
         { "128",  OPTION_VAL_128_ID },
         { "256",  NULL },
         { "384",  NULL },
         { "512",  NULL },
         { "640",  NULL },
         { "768",  NULL },
         { "896",  NULL },
         { "1024", NULL },
         { "1152", NULL },
         { "1280", NULL },
         { "1408", NULL },
         { "1536", NULL },
         { "1664", NULL },
         { "1792", NULL },
         { "1920", NULL },
         { "2048", NULL },
         { NULL, NULL },
      },
      "128"
   },
   {
      BEETLE_OPT(dynarec_spu_samples),
      BEETLE_OPT_DYNAREC_SPU_SAMPLES_LABEL_ID,
      NULL,
      BEETLE_OPT_DYNAREC_SPU_SAMPLES_INFO_0_ID,
      NULL,
      "hacks",
      {
         { "1",  OPTION_VAL_1_O49_ID },
         { "4",  NULL },
         { "16", NULL },
         { NULL, NULL },
      },
      "1"
   },
#endif
   {
      BEETLE_OPT(core_timing_fps),
      BEETLE_OPT_CORE_TIMING_FPS_LABEL_ID,
      NULL,
      BEETLE_OPT_CORE_TIMING_FPS_INFO_0_ID,
      NULL,
      "video",
      {
         { "auto_toggle",       OPTION_VAL_AUTO_TOGGLE_ID },
         { "force_progressive", OPTION_VAL_FORCE_PROGRESSIVE_ID },
         { "force_interlaced",  OPTION_VAL_FORCE_INTERLACED_ID },
      },
      "auto_toggle"
   },
   {
      BEETLE_OPT(deinterlacer),
      BEETLE_OPT_DEINTERLACER_LABEL_ID,
      NULL,
      BEETLE_OPT_DEINTERLACER_INFO_0_ID,
      NULL,
      "video",
      {
         { "weave",      OPTION_VAL_WEAVE_ID },
         { "bob",        OPTION_VAL_BOB_ID },
         { "bob_offset", OPTION_VAL_BOB_OFFSET_ID },
         { "fastmad",    OPTION_VAL_FASTMAD_ID },
         { "off",        "Off" },
         { NULL, NULL },
      },
      "weave"
   },
   {
      BEETLE_OPT(frame_duping),
      BEETLE_OPT_FRAME_DUPING_LABEL_ID,
      NULL,
      BEETLE_OPT_FRAME_DUPING_INFO_0_ID,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(pal_video_timing_override),
      BEETLE_OPT_PAL_VIDEO_TIMING_OVERRIDE_LABEL_ID,
      NULL,
      BEETLE_OPT_PAL_VIDEO_TIMING_OVERRIDE_INFO_0_ID,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(aspect_ratio),
      BEETLE_OPT_ASPECT_RATIO_LABEL_ID,
      NULL,
      BEETLE_OPT_ASPECT_RATIO_INFO_0_ID,
      NULL,
      "video",
      {
         { "corrected", OPTION_VAL_CORRECTED_ID },
         { "uncorrected", OPTION_VAL_UNCORRECTED_ID },
         { "4:3",  OPTION_VAL_4_3_ID },
         { "ntsc", OPTION_VAL_NTSC_ID },
      },
      "corrected"
   },
   {
      BEETLE_OPT(crop_overscan),
      BEETLE_OPT_CROP_OVERSCAN_LABEL_ID,
      NULL,
      BEETLE_OPT_CROP_OVERSCAN_INFO_0_ID,
      NULL,
      "video",
      {
         { "disabled", "Disabled" },
         { "static",   OPTION_VAL_STATIC_ID },
         { "smart",    OPTION_VAL_SMART_ID },
         { NULL, NULL },
      },
      "smart"
   },
   {
      BEETLE_OPT(image_crop),
      BEETLE_OPT_IMAGE_CROP_LABEL_ID,
      NULL,
      BEETLE_OPT_IMAGE_CROP_INFO_0_ID,
      NULL,
      "video",
      {
         { "disabled", "0" },
         { "1px",      OPTION_VAL_1PX_ID },
         { "2px",      OPTION_VAL_2PX_ID },
         { "3px",      OPTION_VAL_3PX_ID },
         { "4px",      OPTION_VAL_4PX_ID },
         { "5px",      OPTION_VAL_5PX_ID },
         { "6px",      OPTION_VAL_6PX_ID },
         { "7px",      OPTION_VAL_7PX_ID },
         { "8px",      OPTION_VAL_8PX_ID },
         { "9px",      OPTION_VAL_9PX_ID },
         { "10px",     OPTION_VAL_10PX_ID },
         { "11px",     OPTION_VAL_11PX_ID },
         { "12px",     OPTION_VAL_12PX_ID },
         { "13px",     OPTION_VAL_13PX_ID },
         { "14px",     OPTION_VAL_14PX_ID },
         { "15px",     OPTION_VAL_15PX_ID },
         { "16px",     OPTION_VAL_16PX_ID },
         { "17px",     OPTION_VAL_17PX_ID },
         { "18px",     OPTION_VAL_18PX_ID },
         { "19px",     OPTION_VAL_19PX_ID },
         { "20px",     OPTION_VAL_20PX_ID },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(image_offset),
      BEETLE_OPT_IMAGE_OFFSET_LABEL_ID,
      NULL,
      BEETLE_OPT_IMAGE_OFFSET_INFO_0_ID,
      NULL,
      "video",
      {
         { "-12px",    OPTION_VAL_12PX_O64_ID },
         { "-11px",    OPTION_VAL_11PX_O64_ID },
         { "-10px",    OPTION_VAL_10PX_O64_ID },
         { "-9px",     OPTION_VAL_9PX_O64_ID },
         { "-8px",     OPTION_VAL_8PX_O64_ID },
         { "-7px",     OPTION_VAL_7PX_O64_ID },
         { "-6px",     OPTION_VAL_6PX_O64_ID },
         { "-5px",     OPTION_VAL_5PX_O64_ID },
         { "-4px",     OPTION_VAL_4PX_O64_ID },
         { "-3px",     OPTION_VAL_3PX_O64_ID },
         { "-2px",     OPTION_VAL_2PX_O64_ID },
         { "-1px",     OPTION_VAL_1PX_O64_ID },
         { "disabled", "0" },
         { "+1px",     OPTION_VAL_1PX_O64_0_ID },
         { "+2px",     OPTION_VAL_2PX_O64_0_ID },
         { "+3px",     OPTION_VAL_3PX_O64_0_ID },
         { "+4px",     OPTION_VAL_4PX_O64_0_ID },
         { "+5px",     OPTION_VAL_5PX_O64_0_ID },
         { "+6px",     OPTION_VAL_6PX_O64_0_ID },
         { "+7px",     OPTION_VAL_7PX_O64_0_ID },
         { "+8px",     OPTION_VAL_8PX_O64_0_ID },
         { "+9px",     OPTION_VAL_9PX_O64_0_ID },
         { "+10px",    OPTION_VAL_10PX_O64_0_ID },
         { "+11px",    OPTION_VAL_11PX_O64_0_ID },
         { "+12px",    OPTION_VAL_12PX_O64_0_ID },
         { NULL, NULL },
      },
      "disabled"
   },
#if defined(HAVE_OPENGL) || defined(HAVE_OPENGLES) || defined(HAVE_VULKAN)
   {
      BEETLE_OPT(image_offset_cycles),
      BEETLE_OPT_IMAGE_OFFSET_CYCLES_LABEL_ID,
      NULL,
      BEETLE_OPT_IMAGE_OFFSET_CYCLES_INFO_0_ID,
      NULL,
      "video",
      {
         { "-40",      NULL },
         { "-39",      NULL },
         { "-38",      NULL },
         { "-37",      NULL },
         { "-36",      NULL },
         { "-35",      NULL },
         { "-34",      NULL },
         { "-33",      NULL },
         { "-32",      NULL },
         { "-31",      NULL },
         { "-30",      NULL },
         { "-29",      NULL },
         { "-28",      NULL },
         { "-27",      NULL },
         { "-26",      NULL },
         { "-25",      NULL },
         { "-24",      NULL },
         { "-23",      NULL },
         { "-22",      NULL },
         { "-21",      NULL },
         { "-20",      NULL },
         { "-19",      NULL },
         { "-18",      NULL },
         { "-17",      NULL },
         { "-16",      NULL },
         { "-15",      NULL },
         { "-14",      NULL },
         { "-13",      NULL },
         { "-12",      NULL },
         { "-11",      NULL },
         { "-10",      NULL },
         { "-9",       NULL },
         { "-8",       NULL },
         { "-7",       NULL },
         { "-6",       NULL },
         { "-5",       NULL },
         { "-4",       NULL },
         { "-3",       NULL },
         { "-2",       NULL },
         { "-1",       NULL },
         { "0",        "0" },
         { "+1",       NULL },
         { "+2",       NULL },
         { "+3",       NULL },
         { "+4",       NULL },
         { "+5",       NULL },
         { "+6",       NULL },
         { "+7",       NULL },
         { "+8",       NULL },
         { "+9",       NULL },
         { "+10",      NULL },
         { "+11",      NULL },
         { "+12",      NULL },
         { "+13",      NULL },
         { "+14",      NULL },
         { "+15",      NULL },
         { "+16",      NULL },
         { "+17",      NULL },
         { "+18",      NULL },
         { "+19",      NULL },
         { "+20",      NULL },
         { "+21",      NULL },
         { "+22",      NULL },
         { "+23",      NULL },
         { "+24",      NULL },
         { "+25",      NULL },
         { "+26",      NULL },
         { "+27",      NULL },
         { "+28",      NULL },
         { "+29",      NULL },
         { "+30",      NULL },
         { "+31",      NULL },
         { "+32",      NULL },
         { "+33",      NULL },
         { "+34",      NULL },
         { "+35",      NULL },
         { "+36",      NULL },
         { "+37",      NULL },
         { "+38",      NULL },
         { "+39",      NULL },
         { "+40",      NULL },
         { NULL, NULL},
      },
      "0"
   },
#endif
   {
      BEETLE_OPT(initial_scanline),
      BEETLE_OPT_INITIAL_SCANLINE_LABEL_ID,
      NULL,
      BEETLE_OPT_INITIAL_SCANLINE_INFO_0_ID,
      NULL,
      "video",
      {
         { "0",  OPTION_VAL_0_O48_ID },
         { "1",  NULL },
         { "2",  NULL },
         { "3",  NULL },
         { "4",  NULL },
         { "5",  NULL },
         { "6",  NULL },
         { "7",  NULL },
         { "8",  NULL },
         { "9",  NULL },
         { "10", NULL },
         { "11", NULL },
         { "12", NULL },
         { "13", NULL },
         { "14", NULL },
         { "15", NULL },
         { "16", NULL },
         { "17", NULL },
         { "18", NULL },
         { "19", NULL },
         { "20", NULL },
         { "21", NULL },
         { "22", NULL },
         { "23", NULL },
         { "24", NULL },
         { "25", NULL },
         { "26", NULL },
         { "27", NULL },
         { "28", NULL },
         { "29", NULL },
         { "30", NULL },
         { "31", NULL },
         { "32", NULL },
         { "33", NULL },
         { "34", NULL },
         { "35", NULL },
         { "36", NULL },
         { "37", NULL },
         { "38", NULL },
         { "39", NULL },
         { "40", NULL },
         { NULL, NULL },
      },
      "0"
   },
   {
      BEETLE_OPT(last_scanline),
      BEETLE_OPT_LAST_SCANLINE_LABEL_ID,
      NULL,
      BEETLE_OPT_LAST_SCANLINE_INFO_0_ID,
      NULL,
      "video",
      {
         { "210", NULL },
         { "211", NULL },
         { "212", NULL },
         { "213", NULL },
         { "214", NULL },
         { "215", NULL },
         { "216", NULL },
         { "217", NULL },
         { "218", NULL },
         { "219", NULL },
         { "220", NULL },
         { "221", NULL },
         { "222", NULL },
         { "223", NULL },
         { "224", NULL },
         { "225", NULL },
         { "226", NULL },
         { "227", NULL },
         { "228", NULL },
         { "229", NULL },
         { "230", NULL },
         { "231", NULL },
         { "232", NULL },
         { "233", NULL },
         { "234", NULL },
         { "235", NULL },
         { "236", NULL },
         { "237", NULL },
         { "238", NULL },
         { "239", OPTION_VAL_239_ID },
         { NULL, NULL },
      },
      "239"
   },
   {
      BEETLE_OPT(initial_scanline_pal),
      BEETLE_OPT_INITIAL_SCANLINE_PAL_LABEL_ID,
      NULL,
      BEETLE_OPT_INITIAL_SCANLINE_PAL_INFO_0_ID,
      NULL,
      "video",
      {
         { "0",  OPTION_VAL_0_O48_ID },
         { "1",  NULL },
         { "2",  NULL },
         { "3",  NULL },
         { "4",  NULL },
         { "5",  NULL },
         { "6",  NULL },
         { "7",  NULL },
         { "8",  NULL },
         { "9",  NULL },
         { "10", NULL },
         { "11", NULL },
         { "12", NULL },
         { "13", NULL },
         { "14", NULL },
         { "15", NULL },
         { "16", NULL },
         { "17", NULL },
         { "18", NULL },
         { "19", NULL },
         { "20", NULL },
         { "21", NULL },
         { "22", NULL },
         { "23", NULL },
         { "24", NULL },
         { "25", NULL },
         { "26", NULL },
         { "27", NULL },
         { "28", NULL },
         { "29", NULL },
         { "30", NULL },
         { "31", NULL },
         { "32", NULL },
         { "33", NULL },
         { "34", NULL },
         { "35", NULL },
         { "36", NULL },
         { "37", NULL },
         { "38", NULL },
         { "39", NULL },
         { "40", NULL },
         { NULL, NULL },
      },
      "0"
   },
   {
      BEETLE_OPT(last_scanline_pal),
      BEETLE_OPT_LAST_SCANLINE_PAL_LABEL_ID,
      NULL,
      BEETLE_OPT_LAST_SCANLINE_PAL_INFO_0_ID,
      NULL,
      "video",
      {
         { "230", NULL },
         { "231", NULL },
         { "232", NULL },
         { "233", NULL },
         { "234", NULL },
         { "235", NULL },
         { "236", NULL },
         { "237", NULL },
         { "238", NULL },
         { "239", NULL },
         { "240", NULL },
         { "241", NULL },
         { "242", NULL },
         { "243", NULL },
         { "244", NULL },
         { "245", NULL },
         { "246", NULL },
         { "247", NULL },
         { "248", NULL },
         { "249", NULL },
         { "250", NULL },
         { "251", NULL },
         { "252", NULL },
         { "253", NULL },
         { "254", NULL },
         { "255", NULL },
         { "256", NULL },
         { "257", NULL },
         { "258", NULL },
         { "259", NULL },
         { "260", NULL },
         { "261", NULL },
         { "262", NULL },
         { "263", NULL },
         { "264", NULL },
         { "265", NULL },
         { "266", NULL },
         { "267", NULL },
         { "268", NULL },
         { "269", NULL },
         { "270", NULL },
         { "271", NULL },
         { "272", NULL },
         { "273", NULL },
         { "274", NULL },
         { "275", NULL },
         { "276", NULL },
         { "277", NULL },
         { "278", NULL },
         { "279", NULL },
         { "280", NULL },
         { "281", NULL },
         { "282", NULL },
         { "283", NULL },
         { "284", NULL },
         { "285", NULL },
         { "286", NULL },
         { "287", OPTION_VAL_287_ID },
         { NULL, NULL },
      },
      "287"
   },
   { NULL, NULL, NULL, NULL, NULL, NULL, {{0}}, NULL },
};
struct retro_core_options_v2 options_id = {
   option_cats_id,
   option_defs_id
};

/* RETRO_LANGUAGE_IT */

#define CATEGORY_SYSTEM_LABEL_IT "Sistema"
#define CATEGORY_SYSTEM_INFO_0_IT "Cambia sistema e opzioni di overclocking."
#define CATEGORY_VIDEO_LABEL_IT NULL
#define CATEGORY_VIDEO_INFO_0_IT "Cambia le proporzioni, il ritaglio, il filtro video e le impostazioni di salto dei fotogrammi."
#define CATEGORY_PGXP_LABEL_IT "Pgxp (Precision Geometry Transform Pipeline)"
#define CATEGORY_PGXP_INFO_0_IT "Queste opzioni controllano i miglioramenti che possono migliorare la grafica rispetto alla console originale. PGXP può sbarazzarsi di texture di deformazione e problemi di combattimento Z."
#define CATEGORY_OSD_LABEL_IT "Mostra sullo schermo"
#define CATEGORY_OSD_INFO_0_IT "Cambia le notifiche visualizzate sullo schermo."
#define CATEGORY_INPUT_LABEL_IT NULL
#define CATEGORY_INPUT_INFO_0_IT "Cambia le impostazioni della pistola ottica, del mouse e del neGcon."
#define CATEGORY_MEMCARDS_LABEL_IT "Scheda Di Memoria"
#define CATEGORY_MEMCARDS_INFO_0_IT "Cambia le impostazioni relative alla scheda di memoria virtuale."
#define CATEGORY_HACKS_LABEL_IT "Hack di Emulazione"
#define CATEGORY_HACKS_INFO_0_IT "Cambia le impostazioni di precisione di emulazione che influenzano prestazioni e compatibilità di basso livello."
#define BEETLE_OPT_CPU_FREQ_SCALE_LABEL_IT "Frequenza CPU (Overclock)"
#define BEETLE_OPT_CPU_FREQ_SCALE_INFO_0_IT "Overclock (o underclock) della CPU PSX emulata. Overclocking può eliminare il rallentamento e migliorare i frame rate in alcuni giochi a scapito di maggiori requisiti di prestazioni. Si noti che alcuni giochi hanno un limitatore interno di frame rate e potrebbero non beneficiare di overclocking. Può causare determinati effetti ad animarsi più velocemente di quanto previsto in alcuni titoli quando sovraorologiato."
#define OPTION_VAL_50_IT NULL
#define OPTION_VAL_60_IT NULL
#define OPTION_VAL_70_IT NULL
#define OPTION_VAL_80_IT NULL
#define OPTION_VAL_90_IT NULL
#define OPTION_VAL_100_IT "100% (Nativo)"
#define OPTION_VAL_110_IT NULL
#define OPTION_VAL_120_IT NULL
#define OPTION_VAL_130_IT NULL
#define OPTION_VAL_140_IT NULL
#define OPTION_VAL_150_IT NULL
#define OPTION_VAL_160_IT NULL
#define OPTION_VAL_170_IT NULL
#define OPTION_VAL_180_IT NULL
#define OPTION_VAL_190_IT NULL
#define OPTION_VAL_200_IT NULL
#define OPTION_VAL_210_IT NULL
#define OPTION_VAL_220_IT NULL
#define OPTION_VAL_230_IT NULL
#define OPTION_VAL_240_IT NULL
#define OPTION_VAL_250_IT NULL
#define OPTION_VAL_260_IT NULL
#define OPTION_VAL_270_IT NULL
#define OPTION_VAL_280_IT NULL
#define OPTION_VAL_290_IT NULL
#define OPTION_VAL_300_IT NULL
#define OPTION_VAL_310_IT NULL
#define OPTION_VAL_320_IT NULL
#define OPTION_VAL_330_IT NULL
#define OPTION_VAL_340_IT NULL
#define OPTION_VAL_350_IT NULL
#define OPTION_VAL_360_IT NULL
#define OPTION_VAL_370_IT NULL
#define OPTION_VAL_380_IT NULL
#define OPTION_VAL_390_IT NULL
#define OPTION_VAL_400_IT NULL
#define OPTION_VAL_410_IT NULL
#define OPTION_VAL_420_IT NULL
#define OPTION_VAL_430_IT NULL
#define OPTION_VAL_440_IT NULL
#define OPTION_VAL_450_IT NULL
#define OPTION_VAL_460_IT NULL
#define OPTION_VAL_470_IT NULL
#define OPTION_VAL_480_IT NULL
#define OPTION_VAL_490_IT NULL
#define OPTION_VAL_500_IT NULL
#define OPTION_VAL_510_IT NULL
#define OPTION_VAL_520_IT NULL
#define OPTION_VAL_530_IT NULL
#define OPTION_VAL_540_IT NULL
#define OPTION_VAL_550_IT NULL
#define OPTION_VAL_560_IT NULL
#define OPTION_VAL_570_IT NULL
#define OPTION_VAL_580_IT NULL
#define OPTION_VAL_590_IT NULL
#define OPTION_VAL_600_IT NULL
#define OPTION_VAL_610_IT NULL
#define OPTION_VAL_620_IT NULL
#define OPTION_VAL_630_IT NULL
#define OPTION_VAL_640_IT NULL
#define OPTION_VAL_650_IT NULL
#define OPTION_VAL_660_IT NULL
#define OPTION_VAL_670_IT NULL
#define OPTION_VAL_680_IT NULL
#define OPTION_VAL_690_IT NULL
#define OPTION_VAL_700_IT NULL
#define OPTION_VAL_710_IT NULL
#define OPTION_VAL_720_IT NULL
#define OPTION_VAL_730_IT NULL
#define OPTION_VAL_740_IT NULL
#define OPTION_VAL_750_IT NULL
#define BEETLE_OPT_GPU_OVERCLOCK_LABEL_IT "Overclock Rasterizzatore GPU"
#define BEETLE_OPT_GPU_OVERCLOCK_INFO_0_IT "Abilita l'overclock per il rasterizzatore 2D contenuto nella GPU del PSX emulato. Non migliora il rendering 3D e ha poco effetto in generale."
#define OPTION_VAL_1X_NATIVE_IT "1x (Nativo)"
#define OPTION_VAL_2X_IT NULL
#define OPTION_VAL_4X_IT NULL
#define OPTION_VAL_8X_IT NULL
#define OPTION_VAL_16X_IT NULL
#define OPTION_VAL_32X_IT NULL
#define BEETLE_OPT_GTE_OVERCLOCK_LABEL_IT "Overclock GTE"
#define BEETLE_OPT_GTE_OVERCLOCK_INFO_0_IT "Abbassa tutte le operazioni GTE emulate (CPU coprocessore per grafica 3D) ad una latenza costante di un ciclo. Per i giochi che fanno uso pesante del GTE, questo può migliorare notevolmente il frame rate e la stabilità del tempo del telaio."
#define BEETLE_OPT_CD_ACCESS_METHOD_LABEL_IT "Metodo di accesso CD"
#define BEETLE_OPT_CD_ACCESS_METHOD_INFO_0_IT "Decide come leggere i dati dalle immagini disco del contenuto. 'Sincrono' imita l'hardware originale, mentre 'Asincrono' riduce lo stuttering nei dispositivi con lettura dati lenta. 'Pre-Cache' carica l'intera immagine disco in memoria quando si avvia il contenuto, e può ridurre i tempi di caricamento in-game a costo di un avvio più lento. 'Pre-Cache' può causare problemi su dispositivi con poca memoria RAM, e ricade su 'Sincrono' per giochi fisici. È necessario il riavvio."
#define OPTION_VAL_SYNC_IT "Sincrono"
#define OPTION_VAL_ASYNC_IT "Asincrono"
#define OPTION_VAL_PRECACHE_IT NULL
#define BEETLE_OPT_CD_FASTLOAD_LABEL_IT "Velocità Caricamento CD"
#define BEETLE_OPT_CD_FASTLOAD_INFO_0_IT "Valori più alti di '2x (Nativo)' possono ridurre notevolmente i tempi di caricamento in-game, ma possono causare errori di timing. Alcuni giochi potrebbero non funzionare correttamente al di sopra di un certo valore."
#define OPTION_VAL_2X_NATIVE_IT "2x (Nativo)"
#define OPTION_VAL_6X_IT NULL
#define OPTION_VAL_10X_IT NULL
#define OPTION_VAL_12X_IT NULL
#define OPTION_VAL_14X_IT NULL
#define BEETLE_OPT_SKIP_BIOS_LABEL_IT "Salta BIOS"
#define BEETLE_OPT_SKIP_BIOS_INFO_0_IT "Salta l'animazione di avvio del BIOS di PlayStation normalmente visualizzata durante il caricamento del contenuto. Nota: Abilitare questo causa problemi di compatibilità con un certo numero di giochi (PAL copia giochi protetti, Saga Frontier, ecc.)."
#define BEETLE_OPT_OVERRIDE_BIOS_LABEL_IT "Sovrascrivi BIOS"
#define BEETLE_OPT_OVERRIDE_BIOS_INFO_0_IT "Sovrascrive il BIOS specifico della regione standard con uno senza regione se trovato. È necessario il riavvio."
#define OPTION_VAL_PSXONPSP_IT NULL
#define OPTION_VAL_PS1_ROM_IT NULL
#define OPTION_VAL_OPENBIOS_IT NULL
#define BEETLE_OPT_REGION_LABEL_IT "Regione Di Sistema"
#define BEETLE_OPT_REGION_INFO_0_IT NULL
#define OPTION_VAL_AUTO_IT NULL
#define OPTION_VAL_NTSC_J_IT NULL
#define OPTION_VAL_NTSC_U_IT NULL
#define OPTION_VAL_PAL_IT NULL
#define BEETLE_OPT_INTERNAL_RESOLUTION_LABEL_IT "Risoluzione GPU Interna"
#define BEETLE_OPT_INTERNAL_RESOLUTION_INFO_0_IT "Risoluzioni maggiori di '1x (Nativo)' migliorano la fedeltà dei modelli 3D a spese di requisiti di performance più alti. Elementi 2D sono generalmente non condizionati da questa impostazione."
#define BEETLE_OPT_RENDERER_LABEL_IT NULL
#define BEETLE_OPT_RENDERER_INFO_0_IT "Il renderer software è il più accurato, ma possiede alti requisiti di prestazione quando è eseguito a risoluzioni GPU interne aumentate. I renderer hardware, sebbene meno accurati, migliorano le prestazioni a risoluzioni interne aumentate, e consentono vari miglioramenti grafici. 'Hardware (Auto)' seleziona automaticamente il renderer Vulkan o OpenGL, a seconda del driver video front-end libretro corrente. Se il driver video fornito non è compatibile con Vulkan o OpenGL 3.3, allora il core tornerà al renderer software. È necessario il riavvio."
#define OPTION_VAL_HARDWARE_IT "Hardware"
#define OPTION_VAL_HARDWARE_GL_IT NULL
#define OPTION_VAL_HARDWARE_VK_IT NULL
#define OPTION_VAL_SOFTWARE_IT NULL
#define BEETLE_OPT_RENDERER_SOFTWARE_FB_LABEL_IT "Framebuffer Software"
#define BEETLE_OPT_RENDERER_SOFTWARE_FB_INFO_0_IT "Abilita l'emulazione accurata di effetti framebuffer (es: motion blur, FF7 battle swirl) quando si utilizza renderer hardware usando una copia del renderer software a risoluzione nativa in sottofondo. Se disabilitato, le operazioni precedenti sono tralasciate (OpenGL) oppure renderizzate dalla GPU (Vulkan). Disabilitare questa impostazione può migliorare le performance ma potrebbe causare errori grafici gravi. In caso di incertezza lasciare 'Abilitato'."
#define BEETLE_OPT_DEPTH_LABEL_IT "Profondità Colori Interni"
#define BEETLE_OPT_DEPTH_INFO_0_IT NULL
#define OPTION_VAL_16BPP_NATIVE_IT "16 bpp (Nativo)"
#define OPTION_VAL_32BPP_IT NULL
#define BEETLE_OPT_DITHER_MODE_LABEL_IT NULL
#define BEETLE_OPT_DITHER_MODE_INFO_0_IT NULL
#define OPTION_VAL_INTERNAL_RESOLUTION_IT "Risoluzione Interna"
#define BEETLE_OPT_SCALED_UV_OFFSET_LABEL_IT "Offset Di Texture UV"
#define BEETLE_OPT_SCALED_UV_OFFSET_INFO_0_IT "Texture di esempio per poligoni 3D a uno offset per una risoluzione interna superiore a 1x. Riduce le cuciture di texture ma può causare difetti grafici indesiderati."
#define BEETLE_OPT_FILTER_LABEL_IT "Filtro Texture"
#define BEETLE_OPT_FILTER_INFO_0_IT NULL
#define OPTION_VAL_NEAREST_IT "Più vicino"
#define OPTION_VAL_SABR_IT NULL
#define OPTION_VAL_XBR_IT NULL
#define OPTION_VAL_BILINEAR_IT "Bilineare"
#define OPTION_VAL_3_POINT_IT "3 Punti"
#define OPTION_VAL_JINC2_IT NULL
#define BEETLE_OPT_FILTER_EXCLUDE_SPRITE_LABEL_IT "Escludi Sprite dal Filtro"
#define BEETLE_OPT_FILTER_EXCLUDE_SPRITE_INFO_0_IT "Non applicare il filtraggio delle texture agli sprite. Previene cuciture in vari giochi con sfondi 2D renderizzati a sprite. Utilizzare insieme ad Adaptive Smoothing o ad un altro filtro post-elaborazione per il miglior effetto."
#define OPTION_VAL_OPAQUE_IT "Solo Opaco"
#define OPTION_VAL_ALL_IT "Opaco e semi-trasparente"
#define BEETLE_OPT_FILTER_EXCLUDE_2D_POLYGON_LABEL_IT "Escludi poligoni 2D dal filtro"
#define BEETLE_OPT_FILTER_EXCLUDE_2D_POLYGON_INFO_0_IT "Non applicare il filtraggio delle texture ai poligoni 2D. I poligoni 2D vengono rilevati con un euristico e potrebbero esserci dei difetti. Utilizzare insieme ad Adaptive Smoothing o ad un altro filtro post-elaborazione per il miglior effetto."
#define BEETLE_OPT_ADAPTIVE_SMOOTHING_LABEL_IT "Smoothing Addativo"
#define BEETLE_OPT_ADAPTIVE_SMOOTHING_INFO_0_IT "Liscia gli elementi di grafica 2D e interfaccia utente senza sfocare gli oggetti rendering 3D. Supportato solo dal renderer Vulkan."
#define BEETLE_OPT_SUPER_SAMPLING_LABEL_IT "Supersampling"
#define BEETLE_OPT_SUPER_SAMPLING_INFO_0_IT "Downsample contenuto renderizzato da risoluzione interna upscaled fino a risoluzione nativa. Combinando questo con i moltiplicatori di risoluzione interna superiore consente di visualizzare i giochi con oggetti 3D anti-aliasing a bassa risoluzione nativa. Produce i migliori risultati quando applicato ai titoli che mescolano elementi 2D e 3D (ad es. Caratteri 3D su sfondo pre-renderizzato), e funziona bene in combinazione con ombreggiatori CRT. Supportato solo dal renderer Vulkan. Nota: si consiglia di disabilitare il 'Dithering Pattern' quando si attiva questa opzione."
#define BEETLE_OPT_MSAA_LABEL_IT "Multisample Anti-Aliasing (MSAA)"
#define BEETLE_OPT_MSAA_INFO_0_IT NULL
#define OPTION_VAL_1X_IT NULL
#define BEETLE_OPT_MDEC_YUV_LABEL_IT "Filtro Chroma MDEC YUV Chroma"
#define BEETLE_OPT_MDEC_YUV_INFO_0_IT "Migliora la qualità della riproduzione FMV riducendo gli artefatti 'macroblocco' (squari/bordi frastagliati). supportati solo dal renderer Vulkan."
#define BEETLE_OPT_TRACK_TEXTURES_LABEL_IT "Traccia Texture"
#define BEETLE_OPT_TRACK_TEXTURES_INFO_0_IT NULL
#define BEETLE_OPT_DUMP_TEXTURES_LABEL_IT "Scarica Texture"
#define BEETLE_OPT_DUMP_TEXTURES_INFO_0_IT "Scarica le texture usate su <cd>-texture-dump/"
#define BEETLE_OPT_REPLACE_TEXTURES_LABEL_IT "Sostituisci Texture"
#define BEETLE_OPT_REPLACE_TEXTURES_INFO_0_IT "Sostituisci le texture con versioni HD da <cd>-texture-replacements/"
#define BEETLE_OPT_PGXP_MODE_LABEL_IT "Modalità Operazione PGXP"
#define BEETLE_OPT_PGXP_MODE_INFO_0_IT NULL
#define OPTION_VAL_MEMORY_ONLY_IT "Solo Memoria"
#define OPTION_VAL_MEMORY_CPU_IT "Memoria + CPU (Buggy)"
#define BEETLE_OPT_PGXP_2D_TOL_LABEL_IT "Tolerancia de geometría PGXP 2D"
#define BEETLE_OPT_PGXP_2D_TOL_INFO_0_IT "Ocultar errores más evidentes en las operaciones de PGXP: el valor especifica la tolerancia en la que se mantendrán los valores de PGXP en caso de geometrías sin la información de profundidad adecuada"
#define OPTION_VAL_0PX_IT NULL
#define OPTION_VAL_1PX_IT NULL
#define OPTION_VAL_2PX_IT NULL
#define OPTION_VAL_3PX_IT NULL
#define OPTION_VAL_4PX_IT NULL
#define OPTION_VAL_5PX_IT NULL
#define OPTION_VAL_6PX_IT NULL
#define OPTION_VAL_7PX_IT NULL
#define OPTION_VAL_8PX_IT NULL
#define BEETLE_OPT_PGXP_NCLIP_LABEL_IT "Eliminación primitiva de PGXP"
#define BEETLE_OPT_PGXP_NCLIP_INFO_0_IT "Utilice la implementación NCLIP de PGXP. Mejora la apariencia al reducir los agujeros en las geometrías con coordenadas PGXP. Se sabe que hace que algunos juegos se bloqueen en diversas circunstancias."
#define BEETLE_OPT_PGXP_VERTEX_LABEL_IT "Vertice Cache PGXP"
#define BEETLE_OPT_PGXP_VERTEX_INFO_0_IT "Posizioni verticali migliorate per Cache PGXP per il riutilizzo tra disegni poligonali. Può potenzialmente migliorare l'allineamento degli oggetti e ridurre le cuciture visibili durante il rendering delle texture, ma i falsi positivi quando si interroga la cache possono produrre glitch grafici. Al momento si raccomanda di lasciare disabilitata questa opzione. Questa opzione è applicata solo quando la Modalità Operazione PGXP è abilitata. Supporta solo i renderer hardware."
#define BEETLE_OPT_PGXP_TEXTURE_LABEL_IT "Prospettiva Texturing Corretta PGXP"
#define BEETLE_OPT_PGXP_TEXTURE_INFO_0_IT NULL
#define BEETLE_OPT_DISPLAY_INTERNAL_FPS_LABEL_IT "Visualizza Contatore FPS Interno"
#define BEETLE_OPT_DISPLAY_INTERNAL_FPS_INFO_0_IT "Visualizza il contatore del frame rate interno a cui viene renderizzato il contenuto del sistema PlayStation emulato. Nota: Richiede di essere abilitato l'opzione delle notifiche su schermo in libretro."
#define BEETLE_OPT_DISPLAY_OSD_LABEL_IT "Visualizza Messaggi OSD"
#define BEETLE_OPT_DISPLAY_OSD_INFO_0_IT "Visualizza i messaggi OSD generati dal nucleo."
#define BEETLE_OPT_DISPLAY_VRAM_LABEL_IT "Visualizza Full VRAM (Debug)"
#define BEETLE_OPT_DISPLAY_VRAM_INFO_0_IT "Visualizza l'intero VRAM della console emulata. Supportato solo dai renderer hardware OpenGL e Vulkan. Nota: Questo è per scopi di debug, e normalmente dovrebbe essere disabilitato."
#define BEETLE_OPT_ANALOG_CALIBRATION_LABEL_IT "Auto-Calibrazione Analogici"
#define BEETLE_OPT_ANALOG_CALIBRATION_INFO_0_IT NULL
#define BEETLE_OPT_ANALOG_TOGGLE_LABEL_IT "Abilita Modalità Toggle Dualshock Analogico"
#define BEETLE_OPT_ANALOG_TOGGLE_INFO_0_IT "Quando il tipo di dispositivo di ingresso è DualShock, questa opzione consente di attivare il DualShock emulato tra la modalità DIGITAL e ANALOG come l'hardware originale. Quando disabilitato, DualShock è bloccato in modalità ANALOG e quando abilitato, il DualShock può essere commutato tra la modalità DIGITAL e ANALOG utilizzando la combinazione di pulsanti selezionata."
#define OPTION_VAL_ENABLED_ANALOG_IT "Analogico di default"
#define BEETLE_OPT_ANALOG_TOGGLE_COMBO_LABEL_IT "Modalità Comibinata DualShock Analogico"
#define BEETLE_OPT_ANALOG_TOGGLE_COMBO_INFO_0_IT NULL
#define OPTION_VAL_L1_L2_R1_R2_START_SELECT_IT NULL
#define OPTION_VAL_L1_R1_SELECT_IT NULL
#define OPTION_VAL_L1_R1_START_IT NULL
#define OPTION_VAL_L1_R1_L3_IT NULL
#define OPTION_VAL_L1_R1_R3_IT NULL
#define OPTION_VAL_L2_R2_SELECT_IT NULL
#define OPTION_VAL_L2_R2_START_IT NULL
#define OPTION_VAL_L2_R2_L3_IT NULL
#define OPTION_VAL_L2_R2_R3_IT NULL
#define OPTION_VAL_L3_R3_IT NULL
#define BEETLE_OPT_ANALOG_TOGGLE_HOLD_LABEL_IT "Tempo di attesa per la combinazione della Modalità Dualshock Analogica"
#define BEETLE_OPT_ANALOG_TOGGLE_HOLD_INFO_0_IT NULL
#define OPTION_VAL_0_IT "0 Secondi"
#define OPTION_VAL_1_IT "1 Secondi"
#define OPTION_VAL_2_IT "2 Secondi"
#define OPTION_VAL_3_IT "3 Secondi"
#define OPTION_VAL_4_IT "4 Secondi"
#define OPTION_VAL_5_IT "5 Secondi"
#define BEETLE_OPT_ENABLE_MULTITAP_PORT1_LABEL_IT "Multitap sulla porta 1"
#define BEETLE_OPT_ENABLE_MULTITAP_PORT1_INFO_0_IT "Abilita la funzionalità multitap sulla porta 1."
#define BEETLE_OPT_ENABLE_MULTITAP_PORT2_LABEL_IT "Multitap sulla porta 2"
#define BEETLE_OPT_ENABLE_MULTITAP_PORT2_INFO_0_IT "Abilita la funzionalità multitap sulla porta 2."
#define BEETLE_OPT_GUN_INPUT_MODE_LABEL_IT "Modalità Input Pistola"
#define BEETLE_OPT_GUN_INPUT_MODE_INFO_0_IT NULL
#define OPTION_VAL_LIGHTGUN_IT "Pistola Ottica"
#define OPTION_VAL_TOUCHSCREEN_IT "Schermo tattile"
#define BEETLE_OPT_GUN_CURSOR_LABEL_IT "Cursore Pistola"
#define BEETLE_OPT_GUN_CURSOR_INFO_0_IT NULL
#define OPTION_VAL_CROSS_IT "Croce"
#define OPTION_VAL_DOT_IT "Puntino"
#define BEETLE_OPT_CROSSHAIR_COLOR_P1_LABEL_IT NULL
#define BEETLE_OPT_CROSSHAIR_COLOR_P1_INFO_0_IT NULL
#define OPTION_VAL_RED_IT "Rosso"
#define OPTION_VAL_BLUE_IT "Blu"
#define OPTION_VAL_GREEN_IT "Verde"
#define OPTION_VAL_ORANGE_IT "Arancione"
#define OPTION_VAL_YELLOW_IT "Giallo"
#define OPTION_VAL_CYAN_IT "Ciano"
#define OPTION_VAL_PINK_IT "Rosa"
#define OPTION_VAL_PURPLE_IT "Viola"
#define OPTION_VAL_BLACK_IT "Nero"
#define OPTION_VAL_WHITE_IT "Bianco"
#define BEETLE_OPT_CROSSHAIR_COLOR_P2_LABEL_IT NULL
#define BEETLE_OPT_CROSSHAIR_COLOR_P2_INFO_0_IT NULL
#define BEETLE_OPT_MOUSE_SENSITIVITY_LABEL_IT "Sensibilità Mouse"
#define BEETLE_OPT_MOUSE_SENSITIVITY_INFO_0_IT NULL
#define OPTION_VAL_5_O42_IT NULL
#define OPTION_VAL_10_IT NULL
#define OPTION_VAL_15_IT NULL
#define OPTION_VAL_20_IT NULL
#define OPTION_VAL_25_IT NULL
#define OPTION_VAL_30_IT NULL
#define OPTION_VAL_35_IT NULL
#define OPTION_VAL_40_IT NULL
#define OPTION_VAL_45_IT NULL
#define OPTION_VAL_55_IT NULL
#define OPTION_VAL_65_IT NULL
#define OPTION_VAL_75_IT NULL
#define OPTION_VAL_85_IT NULL
#define OPTION_VAL_95_IT NULL
#define OPTION_VAL_100_O42_IT NULL
#define OPTION_VAL_105_IT NULL
#define OPTION_VAL_115_IT NULL
#define OPTION_VAL_125_IT NULL
#define OPTION_VAL_135_IT NULL
#define OPTION_VAL_145_IT NULL
#define OPTION_VAL_155_IT NULL
#define OPTION_VAL_165_IT NULL
#define OPTION_VAL_175_IT NULL
#define OPTION_VAL_185_IT NULL
#define OPTION_VAL_195_IT NULL
#define BEETLE_OPT_NEGCON_RESPONSE_LABEL_IT "risposta della torsione neGcon"
#define BEETLE_OPT_NEGCON_RESPONSE_INFO_0_IT NULL
#define OPTION_VAL_LINEAR_IT "Lineare"
#define OPTION_VAL_QUADRATIC_IT "Quadratico"
#define OPTION_VAL_CUBIC_IT "Cubico"
#define BEETLE_OPT_NEGCON_DEADZONE_LABEL_IT "zona morta della torsione del neGcon"
#define BEETLE_OPT_NEGCON_DEADZONE_INFO_0_IT NULL
#define OPTION_VAL_0_O44_IT NULL
#define BEETLE_OPT_USE_MEDNAFEN_MEMCARD0_METHOD_LABEL_IT "Metodo Memory Card"
#define BEETLE_OPT_USE_MEDNAFEN_MEMCARD0_METHOD_INFO_0_IT NULL
#define OPTION_VAL_LIBRETRO_IT NULL
#define OPTION_VAL_MEDNAFEN_IT NULL
#define BEETLE_OPT_ENABLE_MEMCARD1_LABEL_IT "Ingresso Memory Card 2"
#define BEETLE_OPT_ENABLE_MEMCARD1_INFO_0_IT "Emula una seconda memory card nel secondo ingresso. Nota: Alcuni giochi richiedono che l'impostazione sia disabilitata per funzionare correttamente (es. Codename Tenka). Nota: L'ingresso Memory Card 2 usa il formato di salvataggio Mednafen (.mcr). È necessario il riavvio."
#define BEETLE_OPT_SHARED_MEMORY_CARDS_LABEL_IT "Memory Card condivise"
#define BEETLE_OPT_SHARED_MEMORY_CARDS_INFO_0_IT "Usa un'unica memory card invece di memory card separate per ogni contenuto. Nota: Se l'opzione 'Metodo Memory Card' è impostata su 'Libretro', quest'opzione è applicata solo nel secondo ingresso. È necessario il riavvio."
#define BEETLE_OPT_MEMCARD_LEFT_INDEX_LABEL_IT "Indice Ingresso Memory Card 1"
#define BEETLE_OPT_MEMCARD_LEFT_INDEX_INFO_0_IT NULL
#define OPTION_VAL_0_O48_IT NULL
#define BEETLE_OPT_MEMCARD_RIGHT_INDEX_LABEL_IT NULL
#define BEETLE_OPT_MEMCARD_RIGHT_INDEX_INFO_0_IT NULL
#define OPTION_VAL_1_O49_IT NULL
#define BEETLE_OPT_LINE_RENDER_LABEL_IT "Hack Line-to-Quad"
#define BEETLE_OPT_LINE_RENDER_INFO_0_IT NULL
#define OPTION_VAL_DEFAULT_IT "Predefinito"
#define OPTION_VAL_AGGRESSIVE_IT "Aggressivo"
#define BEETLE_OPT_WIDESCREEN_HACK_LABEL_IT "Hack Modalità Widescreen"
#define BEETLE_OPT_WIDESCREEN_HACK_INFO_0_IT "Render il contenuto 3D anamorficamente ed emette il framebuffer emulato ad un formato widescreen. Produce i migliori risultati con giochi completamente 3D. Gli elementi 2D saranno allungati orizzontalmente e potrebbero essere disallineati."
#define BEETLE_OPT_WIDESCREEN_HACK_ASPECT_RATIO_LABEL_IT "Rapporto Dimensioni dell'Hack per Modalità Widescreen"
#define BEETLE_OPT_WIDESCREEN_HACK_ASPECT_RATIO_INFO_0_IT NULL
#define OPTION_VAL_16_9_IT NULL
#define OPTION_VAL_16_10_IT NULL
#define OPTION_VAL_18_9_IT NULL
#define OPTION_VAL_19_9_IT NULL
#define OPTION_VAL_20_9_IT NULL
#define OPTION_VAL_21_9_IT NULL
#define OPTION_VAL_32_9_IT NULL
#define BEETLE_OPT_CPU_DYNAREC_LABEL_IT "Ricompilazione Dinamica CPU (Dynarec)"
#define BEETLE_OPT_CPU_DYNAREC_INFO_0_IT "Ricompila dinamicamente le istruzioni di CPU a istruzioni native. Molto più veloce dell'interpreter, ma i timing della CPU sono meno accurati e potrebbero avere bug."
#define OPTION_VAL_DISABLED_IT "Disabilitato (Beetle Interpreter)"
#define OPTION_VAL_EXECUTE_IT "Performance Massima"
#define OPTION_VAL_EXECUTE_ONE_IT "Controllo Tempi Ciclo"
#define OPTION_VAL_RUN_INTERPRETER_IT "Interprete Lightrec"
#define BEETLE_OPT_DYNAREC_INVALIDATE_LABEL_IT "Invalidamento Codice Dynarec"
#define BEETLE_OPT_DYNAREC_INVALIDATE_INFO_0_IT "Alcuni giochi richiedono 'Full' invalidazione, alcuni richiedono 'DMA soltanto'."
#define OPTION_VAL_FULL_IT "Pieno"
#define OPTION_VAL_DMA_IT "Solo DMA (Leggermente più Veloce)"
#define BEETLE_OPT_DYNAREC_EVENTCYCLES_LABEL_IT "Cicli Di Eventi Dynarec Dma/GPU/MDEC/Timer"
#define BEETLE_OPT_DYNAREC_EVENTCYCLES_INFO_0_IT "Max cicli eseguiti dalla CPU prima che un GPU/DMA/MDEC/Timer Update è controllato, numero più alto sarà più veloce, ha molto meno impatto sull'interprete coleottero di dynarec."
#define OPTION_VAL_128_IT "128 (Predefinito)"
#define BEETLE_OPT_DYNAREC_SPU_SAMPLES_LABEL_IT "Campioni Dynarec SPU"
#define BEETLE_OPT_DYNAREC_SPU_SAMPLES_INFO_0_IT "Max campioni SPU da eseguire prima che venga controllato un aggiornamento SPU, il numero più alto sarà più veloce, ma causerà problemi sonori in alcuni giochi con qualcosa di diverso da 1."
#define BEETLE_OPT_CORE_TIMING_FPS_LABEL_IT "Tempistica FPS segnalata dal core"
#define BEETLE_OPT_CORE_TIMING_FPS_INFO_0_IT NULL
#define OPTION_VAL_AUTO_TOGGLE_IT NULL
#define OPTION_VAL_FORCE_PROGRESSIVE_IT NULL
#define OPTION_VAL_FORCE_INTERLACED_IT "Forza Tasso Interlacciato"
#define BEETLE_OPT_DEINTERLACER_LABEL_IT "Metodo Deinterlacciamento"
#define BEETLE_OPT_DEINTERLACER_INFO_0_IT NULL
#define OPTION_VAL_WEAVE_IT "Intreccio"
#define OPTION_VAL_BOB_IT NULL
#define OPTION_VAL_BOB_OFFSET_IT NULL
#define OPTION_VAL_FASTMAD_IT NULL
#define BEETLE_OPT_FRAME_DUPING_LABEL_IT NULL
#define BEETLE_OPT_FRAME_DUPING_INFO_0_IT NULL
#define BEETLE_OPT_PAL_VIDEO_TIMING_OVERRIDE_LABEL_IT NULL
#define BEETLE_OPT_PAL_VIDEO_TIMING_OVERRIDE_INFO_0_IT NULL
#define BEETLE_OPT_ASPECT_RATIO_LABEL_IT "Rapporto d'aspetto Core"
#define BEETLE_OPT_ASPECT_RATIO_INFO_0_IT NULL
#define OPTION_VAL_CORRECTED_IT "Corretto"
#define OPTION_VAL_UNCORRECTED_IT "Non Corretto"
#define OPTION_VAL_4_3_IT "Forza 4:3"
#define OPTION_VAL_NTSC_IT "Forza NTSC"
#define BEETLE_OPT_CROP_OVERSCAN_LABEL_IT "Taglia Overscan Orizzontale"
#define BEETLE_OPT_CROP_OVERSCAN_INFO_0_IT NULL
#define OPTION_VAL_STATIC_IT "Orizzontale"
#define OPTION_VAL_SMART_IT NULL
#define BEETLE_OPT_IMAGE_CROP_LABEL_IT "Tagli Addizionali"
#define BEETLE_OPT_IMAGE_CROP_INFO_0_IT NULL
#define OPTION_VAL_9PX_IT NULL
#define OPTION_VAL_10PX_IT NULL
#define OPTION_VAL_11PX_IT NULL
#define OPTION_VAL_12PX_IT NULL
#define OPTION_VAL_13PX_IT NULL
#define OPTION_VAL_14PX_IT NULL
#define OPTION_VAL_15PX_IT NULL
#define OPTION_VAL_16PX_IT NULL
#define OPTION_VAL_17PX_IT NULL
#define OPTION_VAL_18PX_IT NULL
#define OPTION_VAL_19PX_IT NULL
#define OPTION_VAL_20PX_IT NULL
#define BEETLE_OPT_IMAGE_OFFSET_LABEL_IT "Offset Immagine Tagliata"
#define BEETLE_OPT_IMAGE_OFFSET_INFO_0_IT NULL
#define OPTION_VAL_12PX_O64_IT NULL
#define OPTION_VAL_11PX_O64_IT NULL
#define OPTION_VAL_10PX_O64_IT NULL
#define OPTION_VAL_9PX_O64_IT NULL
#define OPTION_VAL_8PX_O64_IT NULL
#define OPTION_VAL_7PX_O64_IT NULL
#define OPTION_VAL_6PX_O64_IT NULL
#define OPTION_VAL_5PX_O64_IT NULL
#define OPTION_VAL_4PX_O64_IT NULL
#define OPTION_VAL_3PX_O64_IT NULL
#define OPTION_VAL_2PX_O64_IT NULL
#define OPTION_VAL_1PX_O64_IT NULL
#define OPTION_VAL_1PX_O64_0_IT NULL
#define OPTION_VAL_2PX_O64_0_IT NULL
#define OPTION_VAL_3PX_O64_0_IT NULL
#define OPTION_VAL_4PX_O64_0_IT NULL
#define OPTION_VAL_5PX_O64_0_IT NULL
#define OPTION_VAL_6PX_O64_0_IT NULL
#define OPTION_VAL_7PX_O64_0_IT NULL
#define OPTION_VAL_8PX_O64_0_IT NULL
#define OPTION_VAL_9PX_O64_0_IT NULL
#define OPTION_VAL_10PX_O64_0_IT NULL
#define OPTION_VAL_11PX_O64_0_IT NULL
#define OPTION_VAL_12PX_O64_0_IT NULL
#define BEETLE_OPT_IMAGE_OFFSET_CYCLES_LABEL_IT "Offset Immagine Orizzontale (Cicli GPU)"
#define BEETLE_OPT_IMAGE_OFFSET_CYCLES_INFO_0_IT "I valori positivi muovono l'immagine a destra, i valori negativi muovono invece l'immagine a sinistra. Questa impostazione è supportata solo dai renderer hardware."
#define BEETLE_OPT_INITIAL_SCANLINE_LABEL_IT "Linea Iniziale Di Scansione - NTSC"
#define BEETLE_OPT_INITIAL_SCANLINE_INFO_0_IT NULL
#define BEETLE_OPT_LAST_SCANLINE_LABEL_IT "Ultima Linea Di Scansione - NTSC"
#define BEETLE_OPT_LAST_SCANLINE_INFO_0_IT NULL
#define OPTION_VAL_239_IT "239 (Predefinito)"
#define BEETLE_OPT_INITIAL_SCANLINE_PAL_LABEL_IT "Linea Iniziale Di Scansione - Pal"
#define BEETLE_OPT_INITIAL_SCANLINE_PAL_INFO_0_IT NULL
#define BEETLE_OPT_LAST_SCANLINE_PAL_LABEL_IT "Ultima Linea Di Scansione - Pal"
#define BEETLE_OPT_LAST_SCANLINE_PAL_INFO_0_IT NULL
#define OPTION_VAL_287_IT "287 (Predefinito)"

struct retro_core_option_v2_category option_cats_it[] = {
   {
      "system",
      CATEGORY_SYSTEM_LABEL_IT,
      CATEGORY_SYSTEM_INFO_0_IT
   },
   {
      "video",
      CATEGORY_VIDEO_LABEL_IT,
      CATEGORY_VIDEO_INFO_0_IT
   },
   {
      "pgxp",
      CATEGORY_PGXP_LABEL_IT,
      CATEGORY_PGXP_INFO_0_IT
   },
   {
      "osd",
      CATEGORY_OSD_LABEL_IT,
      CATEGORY_OSD_INFO_0_IT
   },
   {
      "input",
      CATEGORY_INPUT_LABEL_IT,
      CATEGORY_INPUT_INFO_0_IT
   },
   {
      "memcards",
      CATEGORY_MEMCARDS_LABEL_IT,
      CATEGORY_MEMCARDS_INFO_0_IT
   },
   {
      "hacks",
      CATEGORY_HACKS_LABEL_IT,
      CATEGORY_HACKS_INFO_0_IT
   },
   { NULL, NULL, NULL },
};
struct retro_core_option_v2_definition option_defs_it[] = {
   {
      BEETLE_OPT(cpu_freq_scale),
      BEETLE_OPT_CPU_FREQ_SCALE_LABEL_IT,
      NULL,
      BEETLE_OPT_CPU_FREQ_SCALE_INFO_0_IT,
      NULL,
      "system",
      {
         { "50%",           OPTION_VAL_50_IT },
         { "60%",           OPTION_VAL_60_IT },
         { "70%",           OPTION_VAL_70_IT },
         { "80%",           OPTION_VAL_80_IT },
         { "90%",           OPTION_VAL_90_IT },
         { "100%",          OPTION_VAL_100_IT },
         { "110%",          OPTION_VAL_110_IT },
         { "120%",          OPTION_VAL_120_IT },
         { "130%",          OPTION_VAL_130_IT },
         { "140%",          OPTION_VAL_140_IT },
         { "150%",          OPTION_VAL_150_IT },
         { "160%",          OPTION_VAL_160_IT },
         { "170%",          OPTION_VAL_170_IT },
         { "180%",          OPTION_VAL_180_IT },
         { "190%",          OPTION_VAL_190_IT },
         { "200%",          OPTION_VAL_200_IT },
         { "210%",          OPTION_VAL_210_IT },
         { "220%",          OPTION_VAL_220_IT },
         { "230%",          OPTION_VAL_230_IT },
         { "240%",          OPTION_VAL_240_IT },
         { "250%",          OPTION_VAL_250_IT },
         { "260%",          OPTION_VAL_260_IT },
         { "270%",          OPTION_VAL_270_IT },
         { "280%",          OPTION_VAL_280_IT },
         { "290%",          OPTION_VAL_290_IT },
         { "300%",          OPTION_VAL_300_IT },
         { "310%",          OPTION_VAL_310_IT },
         { "320%",          OPTION_VAL_320_IT },
         { "330%",          OPTION_VAL_330_IT },
         { "340%",          OPTION_VAL_340_IT },
         { "350%",          OPTION_VAL_350_IT },
         { "360%",          OPTION_VAL_360_IT },
         { "370%",          OPTION_VAL_370_IT },
         { "380%",          OPTION_VAL_380_IT },
         { "390%",          OPTION_VAL_390_IT },
         { "400%",          OPTION_VAL_400_IT },
         { "410%",          OPTION_VAL_410_IT },
         { "420%",          OPTION_VAL_420_IT },
         { "430%",          OPTION_VAL_430_IT },
         { "440%",          OPTION_VAL_440_IT },
         { "450%",          OPTION_VAL_450_IT },
         { "460%",          OPTION_VAL_460_IT },
         { "470%",          OPTION_VAL_470_IT },
         { "480%",          OPTION_VAL_480_IT },
         { "490%",          OPTION_VAL_490_IT },
         { "500%",          OPTION_VAL_500_IT },
         { "510%",          OPTION_VAL_510_IT },
         { "520%",          OPTION_VAL_520_IT },
         { "530%",          OPTION_VAL_530_IT },
         { "540%",          OPTION_VAL_540_IT },
         { "550%",          OPTION_VAL_550_IT },
         { "560%",          OPTION_VAL_560_IT },
         { "570%",          OPTION_VAL_570_IT },
         { "580%",          OPTION_VAL_580_IT },
         { "590%",          OPTION_VAL_590_IT },
         { "600%",          OPTION_VAL_600_IT },
         { "610%",          OPTION_VAL_610_IT },
         { "620%",          OPTION_VAL_620_IT },
         { "630%",          OPTION_VAL_630_IT },
         { "640%",          OPTION_VAL_640_IT },
         { "650%",          OPTION_VAL_650_IT },
         { "660%",          OPTION_VAL_660_IT },
         { "670%",          OPTION_VAL_670_IT },
         { "680%",          OPTION_VAL_680_IT },
         { "690%",          OPTION_VAL_690_IT },
         { "700%",          OPTION_VAL_700_IT },
         { "710%",          OPTION_VAL_710_IT },
         { "720%",          OPTION_VAL_720_IT },
         { "730%",          OPTION_VAL_730_IT },
         { "740%",          OPTION_VAL_740_IT },
         { "750%",          OPTION_VAL_750_IT },
         { NULL, NULL },
      },
      "100%"
   },
   {
      BEETLE_OPT(gpu_overclock),
      BEETLE_OPT_GPU_OVERCLOCK_LABEL_IT,
      NULL,
      BEETLE_OPT_GPU_OVERCLOCK_INFO_0_IT,
      NULL,
      "system",
      {
         { "1x(native)", OPTION_VAL_1X_NATIVE_IT },
         { "2x",         OPTION_VAL_2X_IT },
         { "4x",         OPTION_VAL_4X_IT },
         { "8x",         OPTION_VAL_8X_IT },
         { "16x",        OPTION_VAL_16X_IT },
         { "32x",        OPTION_VAL_32X_IT },
         { NULL, NULL },
      },
      "1x(native)"
   },
   {
      BEETLE_OPT(gte_overclock),
      BEETLE_OPT_GTE_OVERCLOCK_LABEL_IT,
      NULL,
      BEETLE_OPT_GTE_OVERCLOCK_INFO_0_IT,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
#ifndef EMSCRIPTEN
   {
      BEETLE_OPT(cd_access_method),
      BEETLE_OPT_CD_ACCESS_METHOD_LABEL_IT,
      NULL,
      BEETLE_OPT_CD_ACCESS_METHOD_INFO_0_IT,
      NULL,
      "system",
      {
         { "sync",     OPTION_VAL_SYNC_IT },
         { "async",    OPTION_VAL_ASYNC_IT },
         { "precache", OPTION_VAL_PRECACHE_IT },
         { NULL, NULL },
      },
      "sync"
   },
#endif
   {
      BEETLE_OPT(cd_fastload),
      BEETLE_OPT_CD_FASTLOAD_LABEL_IT,
      NULL,
      BEETLE_OPT_CD_FASTLOAD_INFO_0_IT,
      NULL,
      "system",
      {
         { "2x(native)", OPTION_VAL_2X_NATIVE_IT },
         { "4x",          OPTION_VAL_4X_IT },
         { "6x",          OPTION_VAL_6X_IT },
         { "8x",          OPTION_VAL_8X_IT },
         { "10x",         OPTION_VAL_10X_IT },
         { "12x",         OPTION_VAL_12X_IT },
         { "14x",         OPTION_VAL_14X_IT },
         { NULL, NULL },
      },
      "2x(native)"
   },
   {
      BEETLE_OPT(skip_bios),
      BEETLE_OPT_SKIP_BIOS_LABEL_IT,
      NULL,
      BEETLE_OPT_SKIP_BIOS_INFO_0_IT,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(override_bios),
      BEETLE_OPT_OVERRIDE_BIOS_LABEL_IT,
      NULL,
      BEETLE_OPT_OVERRIDE_BIOS_INFO_0_IT,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "psxonpsp", OPTION_VAL_PSXONPSP_IT },
         { "ps1_rom",  OPTION_VAL_PS1_ROM_IT },
         { "openbios", OPTION_VAL_OPENBIOS_IT     },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(region),
      BEETLE_OPT_REGION_LABEL_IT,
      NULL,
      BEETLE_OPT_REGION_INFO_0_IT,
      NULL,
      "system",
      {
         { "auto",   OPTION_VAL_AUTO_IT },
         { "ntsc-j", OPTION_VAL_NTSC_J_IT },
         { "ntsc-u", OPTION_VAL_NTSC_U_IT },
         { "pal",    OPTION_VAL_PAL_IT },
         { NULL, NULL },
      },
      "auto"
   },
   {
      BEETLE_OPT(internal_resolution),
      BEETLE_OPT_INTERNAL_RESOLUTION_LABEL_IT,
      NULL,
      BEETLE_OPT_INTERNAL_RESOLUTION_INFO_0_IT,
      NULL,
      "video",
      {
         { "1x(native)", OPTION_VAL_1X_NATIVE_IT },
         { "2x",         OPTION_VAL_2X_IT },
         { "4x",         OPTION_VAL_4X_IT },
         { "8x",         OPTION_VAL_8X_IT },
         { "16x",        OPTION_VAL_16X_IT },
         { NULL, NULL },
      },
      "1x(native)"
   },
#if defined(HAVE_OPENGL) || defined(HAVE_OPENGLES) || defined(HAVE_VULKAN)
   {
      BEETLE_OPT(renderer),
      BEETLE_OPT_RENDERER_LABEL_IT,
      NULL,
      BEETLE_OPT_RENDERER_INFO_0_IT,
      NULL,
      "video",
      {
         { "hardware",    OPTION_VAL_HARDWARE_IT },
#if defined(HAVE_OPENGL) || defined(HAVE_OPENGLES)
         { "hardware_gl", OPTION_VAL_HARDWARE_GL_IT },
#endif
#if defined(HAVE_VULKAN)
         { "hardware_vk", OPTION_VAL_HARDWARE_VK_IT },
#endif
         { "software",    OPTION_VAL_SOFTWARE_IT },
         { NULL, NULL },
      },
      "hardware"
   },
   {
      BEETLE_OPT(renderer_software_fb),
      BEETLE_OPT_RENDERER_SOFTWARE_FB_LABEL_IT,
      NULL,
      BEETLE_OPT_RENDERER_SOFTWARE_FB_INFO_0_IT,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled"
   },
#endif
#if defined(HAVE_OPENGL) || defined(HAVE_OPENGLES)
   {
      BEETLE_OPT(depth),
      BEETLE_OPT_DEPTH_LABEL_IT,
      NULL,
      BEETLE_OPT_DEPTH_INFO_0_IT,
      NULL,
      "video",
      {
         { "16bpp(native)", OPTION_VAL_16BPP_NATIVE_IT },
         { "32bpp",         OPTION_VAL_32BPP_IT },
         { NULL, NULL },
      },
      "16bpp(native)"
   },
   // Sort of, it's more like 15-bit high color and 24-bit true color for visible output. The alpha channel is used for mask bit. Vulkan renderer uses ABGR1555_555 for 31 bits internal? FMVs are always 24-bit on all renderers like original hardware (BGR888, no alpha)
#endif
   {
      BEETLE_OPT(dither_mode),
      BEETLE_OPT_DITHER_MODE_LABEL_IT,
      NULL,
      BEETLE_OPT_DITHER_MODE_INFO_0_IT,
      NULL,
      "video",
      {
         { "1x(native)",          OPTION_VAL_1X_NATIVE_IT },
         { "internal resolution", OPTION_VAL_INTERNAL_RESOLUTION_IT },
         { "disabled",            NULL },
         { NULL, NULL },
      },
      "1x(native)"
   },
#ifdef HAVE_VULKAN
   {
      BEETLE_OPT(scaled_uv_offset),
      BEETLE_OPT_SCALED_UV_OFFSET_LABEL_IT,
      NULL,
      BEETLE_OPT_SCALED_UV_OFFSET_INFO_0_IT,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled"
   },
#endif
#if defined(HAVE_OPENGL) || defined(HAVE_OPENGLES) || defined(HAVE_VULKAN)
   {
      BEETLE_OPT(filter),
      BEETLE_OPT_FILTER_LABEL_IT,
      NULL,
      BEETLE_OPT_FILTER_INFO_0_IT,
      NULL,
      "video",
      {
         { "nearest",  OPTION_VAL_NEAREST_IT },
         { "SABR",     OPTION_VAL_SABR_IT },
         { "xBR",      OPTION_VAL_XBR_IT },
         { "bilinear", OPTION_VAL_BILINEAR_IT },
         { "3-point",  OPTION_VAL_3_POINT_IT },
         { "JINC2",    OPTION_VAL_JINC2_IT },
         { NULL, NULL },
      },
      "nearest"
   },
#ifdef HAVE_VULKAN
   {
      BEETLE_OPT(filter_exclude_sprite),
      BEETLE_OPT_FILTER_EXCLUDE_SPRITE_LABEL_IT,
      NULL,
      BEETLE_OPT_FILTER_EXCLUDE_SPRITE_INFO_0_IT,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "opaque", OPTION_VAL_OPAQUE_IT },
         { "all", OPTION_VAL_ALL_IT },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(filter_exclude_2d_polygon),
      BEETLE_OPT_FILTER_EXCLUDE_2D_POLYGON_LABEL_IT,
      NULL,
      BEETLE_OPT_FILTER_EXCLUDE_2D_POLYGON_INFO_0_IT,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "opaque", OPTION_VAL_OPAQUE_IT },
         { "all", OPTION_VAL_ALL_IT },
         { NULL, NULL },
      },
      "disabled"
   },
#endif
#endif
#ifdef HAVE_VULKAN
   {
      BEETLE_OPT(adaptive_smoothing),
      BEETLE_OPT_ADAPTIVE_SMOOTHING_LABEL_IT,
      NULL,
      BEETLE_OPT_ADAPTIVE_SMOOTHING_INFO_0_IT,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(super_sampling),
      BEETLE_OPT_SUPER_SAMPLING_LABEL_IT,
      NULL,
      BEETLE_OPT_SUPER_SAMPLING_INFO_0_IT,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(msaa),
      BEETLE_OPT_MSAA_LABEL_IT,
      NULL,
      BEETLE_OPT_MSAA_INFO_0_IT,
      NULL,
      "video",
      {
         { "1x",  OPTION_VAL_1X_IT },
         { "2x",  OPTION_VAL_2X_IT },
         { "4x",  OPTION_VAL_4X_IT },
         { "8x",  OPTION_VAL_8X_IT },
         { "16x", OPTION_VAL_16X_IT },
         { NULL, NULL },
      },
      "1x"
   },
   {
      BEETLE_OPT(mdec_yuv),
      BEETLE_OPT_MDEC_YUV_LABEL_IT,
      NULL,
      BEETLE_OPT_MDEC_YUV_INFO_0_IT,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(track_textures),
      BEETLE_OPT_TRACK_TEXTURES_LABEL_IT,
      NULL,
      BEETLE_OPT_TRACK_TEXTURES_INFO_0_IT,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
#ifdef TEXTURE_DUMPING_ENABLED
   {
      BEETLE_OPT(dump_textures),
      BEETLE_OPT_DUMP_TEXTURES_LABEL_IT,
      NULL,
      BEETLE_OPT_DUMP_TEXTURES_INFO_0_IT,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
#endif
   {
      BEETLE_OPT(replace_textures),
      BEETLE_OPT_REPLACE_TEXTURES_LABEL_IT,
      NULL,
      BEETLE_OPT_REPLACE_TEXTURES_INFO_0_IT,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
#endif
   {
      BEETLE_OPT(pgxp_mode),
      BEETLE_OPT_PGXP_MODE_LABEL_IT,
      NULL,
      BEETLE_OPT_PGXP_MODE_INFO_0_IT,
      NULL,
      "pgxp",
      {
         { "disabled",     NULL },
         { "memory only",  OPTION_VAL_MEMORY_ONLY_IT },
         { "memory + CPU", OPTION_VAL_MEMORY_CPU_IT },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(pgxp_2d_tol),
      BEETLE_OPT_PGXP_2D_TOL_LABEL_IT,
      NULL,
      BEETLE_OPT_PGXP_2D_TOL_INFO_0_IT,
      NULL,
      "pgxp",
      {
         { "disabled", NULL },
         { "0px", OPTION_VAL_0PX_IT },
         { "1px", OPTION_VAL_1PX_IT },
         { "2px", OPTION_VAL_2PX_IT },
         { "3px", OPTION_VAL_3PX_IT },
         { "4px", OPTION_VAL_4PX_IT },
         { "5px", OPTION_VAL_5PX_IT },
         { "6px", OPTION_VAL_6PX_IT },
         { "7px", OPTION_VAL_7PX_IT },
         { "8px", OPTION_VAL_8PX_IT },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(pgxp_nclip),
      BEETLE_OPT_PGXP_NCLIP_LABEL_IT,
      NULL,
      BEETLE_OPT_PGXP_NCLIP_INFO_0_IT,
      NULL,
      "pgxp",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
#if defined(HAVE_OPENGL) || defined(HAVE_OPENGLES) || defined(HAVE_VULKAN)
   {
      BEETLE_OPT(pgxp_vertex),
      BEETLE_OPT_PGXP_VERTEX_LABEL_IT,
      NULL,
      BEETLE_OPT_PGXP_VERTEX_INFO_0_IT,
      NULL,
      "pgxp",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
#endif
   {
      BEETLE_OPT(pgxp_texture),
      BEETLE_OPT_PGXP_TEXTURE_LABEL_IT,
      NULL,
      BEETLE_OPT_PGXP_TEXTURE_INFO_0_IT,
      NULL,
      "pgxp",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(display_internal_fps),
      BEETLE_OPT_DISPLAY_INTERNAL_FPS_LABEL_IT,
      NULL,
      BEETLE_OPT_DISPLAY_INTERNAL_FPS_INFO_0_IT,
      NULL,
      "osd",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(display_osd),
      BEETLE_OPT_DISPLAY_OSD_LABEL_IT,
      NULL,
      BEETLE_OPT_DISPLAY_OSD_INFO_0_IT,
      NULL,
      "osd",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled"
   },
#if defined(HAVE_OPENGL) || defined(HAVE_OPENGLES) || defined(HAVE_VULKAN)
   {
      BEETLE_OPT(display_vram),
      BEETLE_OPT_DISPLAY_VRAM_LABEL_IT,
      NULL,
      BEETLE_OPT_DISPLAY_VRAM_INFO_0_IT,
      NULL,
      "osd",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
#endif
   {
      BEETLE_OPT(analog_calibration),
      BEETLE_OPT_ANALOG_CALIBRATION_LABEL_IT,
      NULL,
      BEETLE_OPT_ANALOG_CALIBRATION_INFO_0_IT,
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(analog_toggle),
      BEETLE_OPT_ANALOG_TOGGLE_LABEL_IT,
      NULL,
      BEETLE_OPT_ANALOG_TOGGLE_INFO_0_IT,
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { "enabled-analog", OPTION_VAL_ENABLED_ANALOG_IT },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(analog_toggle_combo),
      BEETLE_OPT_ANALOG_TOGGLE_COMBO_LABEL_IT,
      NULL,
      BEETLE_OPT_ANALOG_TOGGLE_COMBO_INFO_0_IT,
      NULL,
      "input",
      {
         { "l1+l2+r1+r2+start+select", OPTION_VAL_L1_L2_R1_R2_START_SELECT_IT },
         { "l1+r1+select",             OPTION_VAL_L1_R1_SELECT_IT },
         { "l1+r1+start",              OPTION_VAL_L1_R1_START_IT },
         { "l1+r1+l3",                 OPTION_VAL_L1_R1_L3_IT },
         { "l1+r1+r3",                 OPTION_VAL_L1_R1_R3_IT },
         { "l2+r2+select",             OPTION_VAL_L2_R2_SELECT_IT },
         { "l2+r2+start",              OPTION_VAL_L2_R2_START_IT },
         { "l2+r2+l3",                 OPTION_VAL_L2_R2_L3_IT },
         { "l2+r2+r3",                 OPTION_VAL_L2_R2_R3_IT },
         { "l3+r3",                    OPTION_VAL_L3_R3_IT },
         { NULL, NULL },
      },
      "l1+r1+select"
   },
   {
      BEETLE_OPT(analog_toggle_hold),
      BEETLE_OPT_ANALOG_TOGGLE_HOLD_LABEL_IT,
      NULL,
      BEETLE_OPT_ANALOG_TOGGLE_HOLD_INFO_0_IT,
      NULL,
      "input",
      {
         { "0", OPTION_VAL_0_IT },
         { "1", OPTION_VAL_1_IT },
         { "2", OPTION_VAL_2_IT },
         { "3", OPTION_VAL_3_IT },
         { "4", OPTION_VAL_4_IT },
         { "5", OPTION_VAL_5_IT },
         { NULL, NULL },
      },
      "1"
   },
   {
      BEETLE_OPT(enable_multitap_port1),
      BEETLE_OPT_ENABLE_MULTITAP_PORT1_LABEL_IT,
      NULL,
      BEETLE_OPT_ENABLE_MULTITAP_PORT1_INFO_0_IT,
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(enable_multitap_port2),
      BEETLE_OPT_ENABLE_MULTITAP_PORT2_LABEL_IT,
      NULL,
      BEETLE_OPT_ENABLE_MULTITAP_PORT2_INFO_0_IT,
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(gun_input_mode),
      BEETLE_OPT_GUN_INPUT_MODE_LABEL_IT,
      NULL,
      BEETLE_OPT_GUN_INPUT_MODE_INFO_0_IT,
      NULL,
      "input",
      {
         { "lightgun",    OPTION_VAL_LIGHTGUN_IT },
         { "touchscreen", OPTION_VAL_TOUCHSCREEN_IT },
         { NULL, NULL },
      },
      "lightgun"
   },
   // Shouldn't the gun_input_mode just be Mouse vs. Touchscreen?
   {
      BEETLE_OPT(gun_cursor),
      BEETLE_OPT_GUN_CURSOR_LABEL_IT,
      NULL,
      BEETLE_OPT_GUN_CURSOR_INFO_0_IT,
      NULL,
      "input",
      {
         { "off",   "disabled" },
         { "cross", OPTION_VAL_CROSS_IT },
         { "dot",   OPTION_VAL_DOT_IT },
         { NULL, NULL },
      },
      "cross"
   },
   {
      BEETLE_OPT(crosshair_color_p1),
      BEETLE_OPT_CROSSHAIR_COLOR_P1_LABEL_IT,
      NULL,
      BEETLE_OPT_CROSSHAIR_COLOR_P1_INFO_0_IT,
      NULL,
      "input",
      {
         { "red",    OPTION_VAL_RED_IT },
         { "blue",   OPTION_VAL_BLUE_IT },
         { "green",  OPTION_VAL_GREEN_IT },
         { "orange", OPTION_VAL_ORANGE_IT },
         { "yellow", OPTION_VAL_YELLOW_IT },
         { "cyan",   OPTION_VAL_CYAN_IT },
         { "pink",   OPTION_VAL_PINK_IT },
         { "purple", OPTION_VAL_PURPLE_IT },
         { "black",  OPTION_VAL_BLACK_IT },
         { "white",  OPTION_VAL_WHITE_IT },
         { NULL, NULL },
      },
      "red"
   },
   {
      BEETLE_OPT(crosshair_color_p2),
      BEETLE_OPT_CROSSHAIR_COLOR_P2_LABEL_IT,
      NULL,
      BEETLE_OPT_CROSSHAIR_COLOR_P2_INFO_0_IT,
      NULL,
      "input",
      {
         { "blue",   OPTION_VAL_BLUE_IT },
         { "red",    OPTION_VAL_RED_IT },
         { "green",  OPTION_VAL_GREEN_IT },
         { "orange", OPTION_VAL_ORANGE_IT },
         { "yellow", OPTION_VAL_YELLOW_IT },
         { "cyan",   OPTION_VAL_CYAN_IT },
         { "pink",   OPTION_VAL_PINK_IT },
         { "purple", OPTION_VAL_PURPLE_IT },
         { "black",  OPTION_VAL_BLACK_IT },
         { "white",  OPTION_VAL_WHITE_IT },
         { NULL, NULL },
      },
      "blue"
   },
   {
      BEETLE_OPT(mouse_sensitivity),
      BEETLE_OPT_MOUSE_SENSITIVITY_LABEL_IT,
      NULL,
      BEETLE_OPT_MOUSE_SENSITIVITY_INFO_0_IT,
      NULL,
      "input",
      {
         { "5%",   OPTION_VAL_5_O42_IT },
         { "10%",  OPTION_VAL_10_IT },
         { "15%",  OPTION_VAL_15_IT },
         { "20%",  OPTION_VAL_20_IT },
         { "25%",  OPTION_VAL_25_IT },
         { "30%",  OPTION_VAL_30_IT },
         { "35%",  OPTION_VAL_35_IT },
         { "40%",  OPTION_VAL_40_IT },
         { "45%",  OPTION_VAL_45_IT },
         { "50%",  OPTION_VAL_50_IT },
         { "55%",  OPTION_VAL_55_IT },
         { "60%",  OPTION_VAL_60_IT },
         { "65%",  OPTION_VAL_65_IT },
         { "70%",  OPTION_VAL_70_IT },
         { "75%",  OPTION_VAL_75_IT },
         { "80%",  OPTION_VAL_80_IT },
         { "85%",  OPTION_VAL_85_IT },
         { "90%",  OPTION_VAL_90_IT },
         { "95%",  OPTION_VAL_95_IT },
         { "100%", OPTION_VAL_100_O42_IT },
         { "105%", OPTION_VAL_105_IT },
         { "110%", OPTION_VAL_110_IT },
         { "115%", OPTION_VAL_115_IT },
         { "120%", OPTION_VAL_120_IT },
         { "125%", OPTION_VAL_125_IT },
         { "130%", OPTION_VAL_130_IT },
         { "135%", OPTION_VAL_135_IT },
         { "140%", OPTION_VAL_140_IT },
         { "145%", OPTION_VAL_145_IT },
         { "150%", OPTION_VAL_150_IT },
         { "155%", OPTION_VAL_155_IT },
         { "160%", OPTION_VAL_160_IT },
         { "165%", OPTION_VAL_165_IT },
         { "170%", OPTION_VAL_170_IT },
         { "175%", OPTION_VAL_175_IT },
         { "180%", OPTION_VAL_180_IT },
         { "185%", OPTION_VAL_185_IT },
         { "190%", OPTION_VAL_190_IT },
         { "195%", OPTION_VAL_195_IT },
         { "200%", OPTION_VAL_200_IT },
         { NULL, NULL },
      },
      "100%"
   },
   {
      BEETLE_OPT(negcon_response),
      BEETLE_OPT_NEGCON_RESPONSE_LABEL_IT,
      NULL,
      BEETLE_OPT_NEGCON_RESPONSE_INFO_0_IT,
      NULL,
      "input",
      {
         { "linear",    OPTION_VAL_LINEAR_IT },
         { "quadratic", OPTION_VAL_QUADRATIC_IT },
         { "cubic",     OPTION_VAL_CUBIC_IT },
         { NULL, NULL },
      },
      "linear"
   },
   {
      BEETLE_OPT(negcon_deadzone),
      BEETLE_OPT_NEGCON_DEADZONE_LABEL_IT,
      NULL,
      BEETLE_OPT_NEGCON_DEADZONE_INFO_0_IT,
      NULL,
      "input",
      {
         { "0%",  OPTION_VAL_0_O44_IT },
         { "5%",  OPTION_VAL_5_O42_IT },
         { "10%", OPTION_VAL_10_IT },
         { "15%", OPTION_VAL_15_IT },
         { "20%", OPTION_VAL_20_IT },
         { "25%", OPTION_VAL_25_IT },
         { "30%", OPTION_VAL_30_IT },
         { NULL, NULL },
      },
      "0%"
   },
   {
      BEETLE_OPT(use_mednafen_memcard0_method),
      BEETLE_OPT_USE_MEDNAFEN_MEMCARD0_METHOD_LABEL_IT,
      NULL,
      BEETLE_OPT_USE_MEDNAFEN_MEMCARD0_METHOD_INFO_0_IT,
      NULL,
      "memcards",
      {
         { "libretro", OPTION_VAL_LIBRETRO_IT },
         { "mednafen", OPTION_VAL_MEDNAFEN_IT },
         { NULL, NULL },
      },
      "libretro"
   },
   {
      BEETLE_OPT(enable_memcard1),
      BEETLE_OPT_ENABLE_MEMCARD1_LABEL_IT,
      NULL,
      BEETLE_OPT_ENABLE_MEMCARD1_INFO_0_IT,
      NULL,
      "memcards",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(shared_memory_cards),
      BEETLE_OPT_SHARED_MEMORY_CARDS_LABEL_IT,
      NULL,
      BEETLE_OPT_SHARED_MEMORY_CARDS_INFO_0_IT,
      NULL,
      "memcards",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(memcard_left_index),
      BEETLE_OPT_MEMCARD_LEFT_INDEX_LABEL_IT,
      NULL,
      BEETLE_OPT_MEMCARD_LEFT_INDEX_INFO_0_IT,
      NULL,
      "memcards",
      {
         { "0",  OPTION_VAL_0_O48_IT },
         { "1",  NULL },
         { "2",  NULL },
         { "3",  NULL },
         { "4",  NULL },
         { "5",  NULL },
         { "6",  NULL },
         { "7",  NULL },
         { "8",  NULL },
         { "9",  NULL },
         { "10", NULL },
         { "11", NULL },
         { "12", NULL },
         { "13", NULL },
         { "14", NULL },
         { "15", NULL },
         { "16", NULL },
         { "17", NULL },
         { "18", NULL },
         { "19", NULL },
         { "20", NULL },
         { "21", NULL },
         { "22", NULL },
         { "23", NULL },
         { "24", NULL },
         { "25", NULL },
         { "26", NULL },
         { "27", NULL },
         { "28", NULL },
         { "29", NULL },
         { "30", NULL },
         { "31", NULL },
         { "32", NULL },
         { "33", NULL },
         { "34", NULL },
         { "35", NULL },
         { "36", NULL },
         { "37", NULL },
         { "38", NULL },
         { "39", NULL },
         { "40", NULL },
         { "41", NULL },
         { "42", NULL },
         { "43", NULL },
         { "44", NULL },
         { "45", NULL },
         { "46", NULL },
         { "47", NULL },
         { "48", NULL },
         { "49", NULL },
         { "50", NULL },
         { "51", NULL },
         { "52", NULL },
         { "53", NULL },
         { "54", NULL },
         { "55", NULL },
         { "56", NULL },
         { "57", NULL },
         { "58", NULL },
         { "59", NULL },
         { "60", NULL },
         { "61", NULL },
         { "62", NULL },
         { "63", NULL },
         { NULL, NULL },
      },
      "0"
   },
   {
      BEETLE_OPT(memcard_right_index),
      BEETLE_OPT_MEMCARD_RIGHT_INDEX_LABEL_IT,
      NULL,
      BEETLE_OPT_MEMCARD_RIGHT_INDEX_INFO_0_IT,
      NULL,
      "memcards",
      {
         { "0",  NULL },
         { "1",  OPTION_VAL_1_O49_IT },
         { "2",  NULL },
         { "3",  NULL },
         { "4",  NULL },
         { "5",  NULL },
         { "6",  NULL },
         { "7",  NULL },
         { "8",  NULL },
         { "9",  NULL },
         { "10", NULL },
         { "11", NULL },
         { "12", NULL },
         { "13", NULL },
         { "14", NULL },
         { "15", NULL },
         { "16", NULL },
         { "17", NULL },
         { "18", NULL },
         { "19", NULL },
         { "20", NULL },
         { "21", NULL },
         { "22", NULL },
         { "23", NULL },
         { "24", NULL },
         { "25", NULL },
         { "26", NULL },
         { "27", NULL },
         { "28", NULL },
         { "29", NULL },
         { "30", NULL },
         { "31", NULL },
         { "32", NULL },
         { "33", NULL },
         { "34", NULL },
         { "35", NULL },
         { "36", NULL },
         { "37", NULL },
         { "38", NULL },
         { "39", NULL },
         { "40", NULL },
         { "41", NULL },
         { "42", NULL },
         { "43", NULL },
         { "44", NULL },
         { "45", NULL },
         { "46", NULL },
         { "47", NULL },
         { "48", NULL },
         { "49", NULL },
         { "50", NULL },
         { "51", NULL },
         { "52", NULL },
         { "53", NULL },
         { "54", NULL },
         { "55", NULL },
         { "56", NULL },
         { "57", NULL },
         { "58", NULL },
         { "59", NULL },
         { "60", NULL },
         { "61", NULL },
         { "62", NULL },
         { "63", NULL },
         { NULL, NULL },
      },
      "1"
   },
   {
      BEETLE_OPT(line_render),
      BEETLE_OPT_LINE_RENDER_LABEL_IT,
      NULL,
      BEETLE_OPT_LINE_RENDER_INFO_0_IT,
      NULL,
      "hacks",
      {
         { "default",    OPTION_VAL_DEFAULT_IT },
         { "aggressive", OPTION_VAL_AGGRESSIVE_IT },
         { "disabled",   NULL },
         { NULL, NULL },
      },
      "default"
   },
   {
      BEETLE_OPT(widescreen_hack),
      BEETLE_OPT_WIDESCREEN_HACK_LABEL_IT,
      NULL,
      BEETLE_OPT_WIDESCREEN_HACK_INFO_0_IT,
      NULL,
      "hacks",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(widescreen_hack_aspect_ratio),
      BEETLE_OPT_WIDESCREEN_HACK_ASPECT_RATIO_LABEL_IT,
      NULL,
      BEETLE_OPT_WIDESCREEN_HACK_ASPECT_RATIO_INFO_0_IT,
      NULL,
      "hacks",
      {
         { "16:9",  OPTION_VAL_16_9_IT },
         { "16:10", OPTION_VAL_16_10_IT },
         { "18:9",  OPTION_VAL_18_9_IT },
         { "19:9",  OPTION_VAL_19_9_IT },
         { "20:9",  OPTION_VAL_20_9_IT },
         { "21:9",  OPTION_VAL_21_9_IT }, // 64:27
         { "32:9",  OPTION_VAL_32_9_IT },
         { NULL,    NULL },
      },
      "16:9"
   },
#if defined(HAVE_LIGHTREC)
   {
      BEETLE_OPT(cpu_dynarec),
      BEETLE_OPT_CPU_DYNAREC_LABEL_IT,
      NULL,
      BEETLE_OPT_CPU_DYNAREC_INFO_0_IT,
      NULL,
      "hacks",
      {
         { "disabled",        OPTION_VAL_DISABLED_IT },
         { "execute",         OPTION_VAL_EXECUTE_IT },
         { "execute_one",     OPTION_VAL_EXECUTE_ONE_IT },
         { "run_interpreter", OPTION_VAL_RUN_INTERPRETER_IT },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(dynarec_invalidate),
      BEETLE_OPT_DYNAREC_INVALIDATE_LABEL_IT,
      NULL,
      BEETLE_OPT_DYNAREC_INVALIDATE_INFO_0_IT,
      NULL,
      "hacks",
      {
         { "full", OPTION_VAL_FULL_IT },
         { "dma",  OPTION_VAL_DMA_IT },
         { NULL, NULL },
      },
      "full"
   },
   {
      BEETLE_OPT(dynarec_eventcycles),
      BEETLE_OPT_DYNAREC_EVENTCYCLES_LABEL_IT,
      NULL,
      BEETLE_OPT_DYNAREC_EVENTCYCLES_INFO_0_IT,
      NULL,
      "hacks",
      {
         { "128",  OPTION_VAL_128_IT },
         { "256",  NULL },
         { "384",  NULL },
         { "512",  NULL },
         { "640",  NULL },
         { "768",  NULL },
         { "896",  NULL },
         { "1024", NULL },
         { "1152", NULL },
         { "1280", NULL },
         { "1408", NULL },
         { "1536", NULL },
         { "1664", NULL },
         { "1792", NULL },
         { "1920", NULL },
         { "2048", NULL },
         { NULL, NULL },
      },
      "128"
   },
   {
      BEETLE_OPT(dynarec_spu_samples),
      BEETLE_OPT_DYNAREC_SPU_SAMPLES_LABEL_IT,
      NULL,
      BEETLE_OPT_DYNAREC_SPU_SAMPLES_INFO_0_IT,
      NULL,
      "hacks",
      {
         { "1",  OPTION_VAL_1_O49_IT },
         { "4",  NULL },
         { "16", NULL },
         { NULL, NULL },
      },
      "1"
   },
#endif
   {
      BEETLE_OPT(core_timing_fps),
      BEETLE_OPT_CORE_TIMING_FPS_LABEL_IT,
      NULL,
      BEETLE_OPT_CORE_TIMING_FPS_INFO_0_IT,
      NULL,
      "video",
      {
         { "auto_toggle",       OPTION_VAL_AUTO_TOGGLE_IT },
         { "force_progressive", OPTION_VAL_FORCE_PROGRESSIVE_IT },
         { "force_interlaced",  OPTION_VAL_FORCE_INTERLACED_IT },
      },
      "auto_toggle"
   },
   {
      BEETLE_OPT(deinterlacer),
      BEETLE_OPT_DEINTERLACER_LABEL_IT,
      NULL,
      BEETLE_OPT_DEINTERLACER_INFO_0_IT,
      NULL,
      "video",
      {
         { "weave",      OPTION_VAL_WEAVE_IT },
         { "bob",        OPTION_VAL_BOB_IT },
         { "bob_offset", OPTION_VAL_BOB_OFFSET_IT },
         { "fastmad",    OPTION_VAL_FASTMAD_IT },
         { "off",        "Off" },
         { NULL, NULL },
      },
      "weave"
   },
   {
      BEETLE_OPT(frame_duping),
      BEETLE_OPT_FRAME_DUPING_LABEL_IT,
      NULL,
      BEETLE_OPT_FRAME_DUPING_INFO_0_IT,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(pal_video_timing_override),
      BEETLE_OPT_PAL_VIDEO_TIMING_OVERRIDE_LABEL_IT,
      NULL,
      BEETLE_OPT_PAL_VIDEO_TIMING_OVERRIDE_INFO_0_IT,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(aspect_ratio),
      BEETLE_OPT_ASPECT_RATIO_LABEL_IT,
      NULL,
      BEETLE_OPT_ASPECT_RATIO_INFO_0_IT,
      NULL,
      "video",
      {
         { "corrected", OPTION_VAL_CORRECTED_IT },
         { "uncorrected", OPTION_VAL_UNCORRECTED_IT },
         { "4:3",  OPTION_VAL_4_3_IT },
         { "ntsc", OPTION_VAL_NTSC_IT },
      },
      "corrected"
   },
   {
      BEETLE_OPT(crop_overscan),
      BEETLE_OPT_CROP_OVERSCAN_LABEL_IT,
      NULL,
      BEETLE_OPT_CROP_OVERSCAN_INFO_0_IT,
      NULL,
      "video",
      {
         { "disabled", "Disabled" },
         { "static",   OPTION_VAL_STATIC_IT },
         { "smart",    OPTION_VAL_SMART_IT },
         { NULL, NULL },
      },
      "smart"
   },
   {
      BEETLE_OPT(image_crop),
      BEETLE_OPT_IMAGE_CROP_LABEL_IT,
      NULL,
      BEETLE_OPT_IMAGE_CROP_INFO_0_IT,
      NULL,
      "video",
      {
         { "disabled", "0" },
         { "1px",      OPTION_VAL_1PX_IT },
         { "2px",      OPTION_VAL_2PX_IT },
         { "3px",      OPTION_VAL_3PX_IT },
         { "4px",      OPTION_VAL_4PX_IT },
         { "5px",      OPTION_VAL_5PX_IT },
         { "6px",      OPTION_VAL_6PX_IT },
         { "7px",      OPTION_VAL_7PX_IT },
         { "8px",      OPTION_VAL_8PX_IT },
         { "9px",      OPTION_VAL_9PX_IT },
         { "10px",     OPTION_VAL_10PX_IT },
         { "11px",     OPTION_VAL_11PX_IT },
         { "12px",     OPTION_VAL_12PX_IT },
         { "13px",     OPTION_VAL_13PX_IT },
         { "14px",     OPTION_VAL_14PX_IT },
         { "15px",     OPTION_VAL_15PX_IT },
         { "16px",     OPTION_VAL_16PX_IT },
         { "17px",     OPTION_VAL_17PX_IT },
         { "18px",     OPTION_VAL_18PX_IT },
         { "19px",     OPTION_VAL_19PX_IT },
         { "20px",     OPTION_VAL_20PX_IT },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(image_offset),
      BEETLE_OPT_IMAGE_OFFSET_LABEL_IT,
      NULL,
      BEETLE_OPT_IMAGE_OFFSET_INFO_0_IT,
      NULL,
      "video",
      {
         { "-12px",    OPTION_VAL_12PX_O64_IT },
         { "-11px",    OPTION_VAL_11PX_O64_IT },
         { "-10px",    OPTION_VAL_10PX_O64_IT },
         { "-9px",     OPTION_VAL_9PX_O64_IT },
         { "-8px",     OPTION_VAL_8PX_O64_IT },
         { "-7px",     OPTION_VAL_7PX_O64_IT },
         { "-6px",     OPTION_VAL_6PX_O64_IT },
         { "-5px",     OPTION_VAL_5PX_O64_IT },
         { "-4px",     OPTION_VAL_4PX_O64_IT },
         { "-3px",     OPTION_VAL_3PX_O64_IT },
         { "-2px",     OPTION_VAL_2PX_O64_IT },
         { "-1px",     OPTION_VAL_1PX_O64_IT },
         { "disabled", "0" },
         { "+1px",     OPTION_VAL_1PX_O64_0_IT },
         { "+2px",     OPTION_VAL_2PX_O64_0_IT },
         { "+3px",     OPTION_VAL_3PX_O64_0_IT },
         { "+4px",     OPTION_VAL_4PX_O64_0_IT },
         { "+5px",     OPTION_VAL_5PX_O64_0_IT },
         { "+6px",     OPTION_VAL_6PX_O64_0_IT },
         { "+7px",     OPTION_VAL_7PX_O64_0_IT },
         { "+8px",     OPTION_VAL_8PX_O64_0_IT },
         { "+9px",     OPTION_VAL_9PX_O64_0_IT },
         { "+10px",    OPTION_VAL_10PX_O64_0_IT },
         { "+11px",    OPTION_VAL_11PX_O64_0_IT },
         { "+12px",    OPTION_VAL_12PX_O64_0_IT },
         { NULL, NULL },
      },
      "disabled"
   },
#if defined(HAVE_OPENGL) || defined(HAVE_OPENGLES) || defined(HAVE_VULKAN)
   {
      BEETLE_OPT(image_offset_cycles),
      BEETLE_OPT_IMAGE_OFFSET_CYCLES_LABEL_IT,
      NULL,
      BEETLE_OPT_IMAGE_OFFSET_CYCLES_INFO_0_IT,
      NULL,
      "video",
      {
         { "-40",      NULL },
         { "-39",      NULL },
         { "-38",      NULL },
         { "-37",      NULL },
         { "-36",      NULL },
         { "-35",      NULL },
         { "-34",      NULL },
         { "-33",      NULL },
         { "-32",      NULL },
         { "-31",      NULL },
         { "-30",      NULL },
         { "-29",      NULL },
         { "-28",      NULL },
         { "-27",      NULL },
         { "-26",      NULL },
         { "-25",      NULL },
         { "-24",      NULL },
         { "-23",      NULL },
         { "-22",      NULL },
         { "-21",      NULL },
         { "-20",      NULL },
         { "-19",      NULL },
         { "-18",      NULL },
         { "-17",      NULL },
         { "-16",      NULL },
         { "-15",      NULL },
         { "-14",      NULL },
         { "-13",      NULL },
         { "-12",      NULL },
         { "-11",      NULL },
         { "-10",      NULL },
         { "-9",       NULL },
         { "-8",       NULL },
         { "-7",       NULL },
         { "-6",       NULL },
         { "-5",       NULL },
         { "-4",       NULL },
         { "-3",       NULL },
         { "-2",       NULL },
         { "-1",       NULL },
         { "0",        "0" },
         { "+1",       NULL },
         { "+2",       NULL },
         { "+3",       NULL },
         { "+4",       NULL },
         { "+5",       NULL },
         { "+6",       NULL },
         { "+7",       NULL },
         { "+8",       NULL },
         { "+9",       NULL },
         { "+10",      NULL },
         { "+11",      NULL },
         { "+12",      NULL },
         { "+13",      NULL },
         { "+14",      NULL },
         { "+15",      NULL },
         { "+16",      NULL },
         { "+17",      NULL },
         { "+18",      NULL },
         { "+19",      NULL },
         { "+20",      NULL },
         { "+21",      NULL },
         { "+22",      NULL },
         { "+23",      NULL },
         { "+24",      NULL },
         { "+25",      NULL },
         { "+26",      NULL },
         { "+27",      NULL },
         { "+28",      NULL },
         { "+29",      NULL },
         { "+30",      NULL },
         { "+31",      NULL },
         { "+32",      NULL },
         { "+33",      NULL },
         { "+34",      NULL },
         { "+35",      NULL },
         { "+36",      NULL },
         { "+37",      NULL },
         { "+38",      NULL },
         { "+39",      NULL },
         { "+40",      NULL },
         { NULL, NULL},
      },
      "0"
   },
#endif
   {
      BEETLE_OPT(initial_scanline),
      BEETLE_OPT_INITIAL_SCANLINE_LABEL_IT,
      NULL,
      BEETLE_OPT_INITIAL_SCANLINE_INFO_0_IT,
      NULL,
      "video",
      {
         { "0",  OPTION_VAL_0_O48_IT },
         { "1",  NULL },
         { "2",  NULL },
         { "3",  NULL },
         { "4",  NULL },
         { "5",  NULL },
         { "6",  NULL },
         { "7",  NULL },
         { "8",  NULL },
         { "9",  NULL },
         { "10", NULL },
         { "11", NULL },
         { "12", NULL },
         { "13", NULL },
         { "14", NULL },
         { "15", NULL },
         { "16", NULL },
         { "17", NULL },
         { "18", NULL },
         { "19", NULL },
         { "20", NULL },
         { "21", NULL },
         { "22", NULL },
         { "23", NULL },
         { "24", NULL },
         { "25", NULL },
         { "26", NULL },
         { "27", NULL },
         { "28", NULL },
         { "29", NULL },
         { "30", NULL },
         { "31", NULL },
         { "32", NULL },
         { "33", NULL },
         { "34", NULL },
         { "35", NULL },
         { "36", NULL },
         { "37", NULL },
         { "38", NULL },
         { "39", NULL },
         { "40", NULL },
         { NULL, NULL },
      },
      "0"
   },
   {
      BEETLE_OPT(last_scanline),
      BEETLE_OPT_LAST_SCANLINE_LABEL_IT,
      NULL,
      BEETLE_OPT_LAST_SCANLINE_INFO_0_IT,
      NULL,
      "video",
      {
         { "210", NULL },
         { "211", NULL },
         { "212", NULL },
         { "213", NULL },
         { "214", NULL },
         { "215", NULL },
         { "216", NULL },
         { "217", NULL },
         { "218", NULL },
         { "219", NULL },
         { "220", NULL },
         { "221", NULL },
         { "222", NULL },
         { "223", NULL },
         { "224", NULL },
         { "225", NULL },
         { "226", NULL },
         { "227", NULL },
         { "228", NULL },
         { "229", NULL },
         { "230", NULL },
         { "231", NULL },
         { "232", NULL },
         { "233", NULL },
         { "234", NULL },
         { "235", NULL },
         { "236", NULL },
         { "237", NULL },
         { "238", NULL },
         { "239", OPTION_VAL_239_IT },
         { NULL, NULL },
      },
      "239"
   },
   {
      BEETLE_OPT(initial_scanline_pal),
      BEETLE_OPT_INITIAL_SCANLINE_PAL_LABEL_IT,
      NULL,
      BEETLE_OPT_INITIAL_SCANLINE_PAL_INFO_0_IT,
      NULL,
      "video",
      {
         { "0",  OPTION_VAL_0_O48_IT },
         { "1",  NULL },
         { "2",  NULL },
         { "3",  NULL },
         { "4",  NULL },
         { "5",  NULL },
         { "6",  NULL },
         { "7",  NULL },
         { "8",  NULL },
         { "9",  NULL },
         { "10", NULL },
         { "11", NULL },
         { "12", NULL },
         { "13", NULL },
         { "14", NULL },
         { "15", NULL },
         { "16", NULL },
         { "17", NULL },
         { "18", NULL },
         { "19", NULL },
         { "20", NULL },
         { "21", NULL },
         { "22", NULL },
         { "23", NULL },
         { "24", NULL },
         { "25", NULL },
         { "26", NULL },
         { "27", NULL },
         { "28", NULL },
         { "29", NULL },
         { "30", NULL },
         { "31", NULL },
         { "32", NULL },
         { "33", NULL },
         { "34", NULL },
         { "35", NULL },
         { "36", NULL },
         { "37", NULL },
         { "38", NULL },
         { "39", NULL },
         { "40", NULL },
         { NULL, NULL },
      },
      "0"
   },
   {
      BEETLE_OPT(last_scanline_pal),
      BEETLE_OPT_LAST_SCANLINE_PAL_LABEL_IT,
      NULL,
      BEETLE_OPT_LAST_SCANLINE_PAL_INFO_0_IT,
      NULL,
      "video",
      {
         { "230", NULL },
         { "231", NULL },
         { "232", NULL },
         { "233", NULL },
         { "234", NULL },
         { "235", NULL },
         { "236", NULL },
         { "237", NULL },
         { "238", NULL },
         { "239", NULL },
         { "240", NULL },
         { "241", NULL },
         { "242", NULL },
         { "243", NULL },
         { "244", NULL },
         { "245", NULL },
         { "246", NULL },
         { "247", NULL },
         { "248", NULL },
         { "249", NULL },
         { "250", NULL },
         { "251", NULL },
         { "252", NULL },
         { "253", NULL },
         { "254", NULL },
         { "255", NULL },
         { "256", NULL },
         { "257", NULL },
         { "258", NULL },
         { "259", NULL },
         { "260", NULL },
         { "261", NULL },
         { "262", NULL },
         { "263", NULL },
         { "264", NULL },
         { "265", NULL },
         { "266", NULL },
         { "267", NULL },
         { "268", NULL },
         { "269", NULL },
         { "270", NULL },
         { "271", NULL },
         { "272", NULL },
         { "273", NULL },
         { "274", NULL },
         { "275", NULL },
         { "276", NULL },
         { "277", NULL },
         { "278", NULL },
         { "279", NULL },
         { "280", NULL },
         { "281", NULL },
         { "282", NULL },
         { "283", NULL },
         { "284", NULL },
         { "285", NULL },
         { "286", NULL },
         { "287", OPTION_VAL_287_IT },
         { NULL, NULL },
      },
      "287"
   },
   { NULL, NULL, NULL, NULL, NULL, NULL, {{0}}, NULL },
};
struct retro_core_options_v2 options_it = {
   option_cats_it,
   option_defs_it
};

/* RETRO_LANGUAGE_JA */

#define CATEGORY_SYSTEM_LABEL_JA "システム"
#define CATEGORY_SYSTEM_INFO_0_JA NULL
#define CATEGORY_VIDEO_LABEL_JA "ビデオ"
#define CATEGORY_VIDEO_INFO_0_JA "アスペクト比、画面のトリミング、ビデオフィルターおよびフレームスキップの設定を変更します。"
#define CATEGORY_PGXP_LABEL_JA "PGXP (精度ジオメトリ変換パイプライン)"
#define CATEGORY_PGXP_INFO_0_JA "これらのオプションは、実機と比較してグラフィックスを改善することができる拡張を制御します。PGXP は、テクスチャの歪みやZ ファイティングの問題を取り除くことができます。"
#define CATEGORY_OSD_LABEL_JA "OSD"
#define CATEGORY_OSD_INFO_0_JA "画面に表示される通知を変更します。"
#define CATEGORY_INPUT_LABEL_JA "入力"
#define CATEGORY_INPUT_INFO_0_JA "ライトガン、マウスおよびネジコンの設定を変更します。"
#define CATEGORY_MEMCARDS_LABEL_JA "メモリーカード"
#define CATEGORY_MEMCARDS_INFO_0_JA NULL
#define CATEGORY_HACKS_LABEL_JA "エミュレーションハック"
#define CATEGORY_HACKS_INFO_0_JA NULL
#define BEETLE_OPT_CPU_FREQ_SCALE_LABEL_JA "CPU 周波数倍率 (オーバークロック)"
#define BEETLE_OPT_CPU_FREQ_SCALE_INFO_0_JA "エミュレートされた PS1 CPU をオーバークロック (またはアンダークロック) します。システム要件が増加する代わりに、特定のゲームのフレームレートを向上させることができます。注意: 一部のゲームにはフレームレートリミッターがあり、オーバークロックの恩恵を受けない場合があります。また、意図されているものよりも高速に動作することもあります。"
#define OPTION_VAL_50_JA NULL
#define OPTION_VAL_60_JA NULL
#define OPTION_VAL_70_JA NULL
#define OPTION_VAL_80_JA NULL
#define OPTION_VAL_90_JA NULL
#define OPTION_VAL_100_JA NULL
#define OPTION_VAL_110_JA NULL
#define OPTION_VAL_120_JA NULL
#define OPTION_VAL_130_JA NULL
#define OPTION_VAL_140_JA NULL
#define OPTION_VAL_150_JA NULL
#define OPTION_VAL_160_JA NULL
#define OPTION_VAL_170_JA NULL
#define OPTION_VAL_180_JA NULL
#define OPTION_VAL_190_JA NULL
#define OPTION_VAL_200_JA NULL
#define OPTION_VAL_210_JA NULL
#define OPTION_VAL_220_JA NULL
#define OPTION_VAL_230_JA NULL
#define OPTION_VAL_240_JA NULL
#define OPTION_VAL_250_JA NULL
#define OPTION_VAL_260_JA NULL
#define OPTION_VAL_270_JA NULL
#define OPTION_VAL_280_JA NULL
#define OPTION_VAL_290_JA NULL
#define OPTION_VAL_300_JA NULL
#define OPTION_VAL_310_JA NULL
#define OPTION_VAL_320_JA NULL
#define OPTION_VAL_330_JA NULL
#define OPTION_VAL_340_JA NULL
#define OPTION_VAL_350_JA NULL
#define OPTION_VAL_360_JA NULL
#define OPTION_VAL_370_JA NULL
#define OPTION_VAL_380_JA NULL
#define OPTION_VAL_390_JA NULL
#define OPTION_VAL_400_JA NULL
#define OPTION_VAL_410_JA NULL
#define OPTION_VAL_420_JA NULL
#define OPTION_VAL_430_JA NULL
#define OPTION_VAL_440_JA NULL
#define OPTION_VAL_450_JA NULL
#define OPTION_VAL_460_JA NULL
#define OPTION_VAL_470_JA NULL
#define OPTION_VAL_480_JA NULL
#define OPTION_VAL_490_JA NULL
#define OPTION_VAL_500_JA NULL
#define OPTION_VAL_510_JA NULL
#define OPTION_VAL_520_JA NULL
#define OPTION_VAL_530_JA NULL
#define OPTION_VAL_540_JA NULL
#define OPTION_VAL_550_JA NULL
#define OPTION_VAL_560_JA NULL
#define OPTION_VAL_570_JA NULL
#define OPTION_VAL_580_JA NULL
#define OPTION_VAL_590_JA NULL
#define OPTION_VAL_600_JA NULL
#define OPTION_VAL_610_JA NULL
#define OPTION_VAL_620_JA NULL
#define OPTION_VAL_630_JA NULL
#define OPTION_VAL_640_JA NULL
#define OPTION_VAL_650_JA NULL
#define OPTION_VAL_660_JA NULL
#define OPTION_VAL_670_JA NULL
#define OPTION_VAL_680_JA NULL
#define OPTION_VAL_690_JA NULL
#define OPTION_VAL_700_JA NULL
#define OPTION_VAL_710_JA NULL
#define OPTION_VAL_720_JA NULL
#define OPTION_VAL_730_JA NULL
#define OPTION_VAL_740_JA NULL
#define OPTION_VAL_750_JA NULL
#define BEETLE_OPT_GPU_OVERCLOCK_LABEL_JA "GPU ラスタライザオーバークロック"
#define BEETLE_OPT_GPU_OVERCLOCK_INFO_0_JA "エミュレートされた PS1 の GPU に含まれる 2D ラスタライザのオーバークロックを有効にします。3D レンダリングは改善されず、一般的にはほとんど効果がありません。"
#define OPTION_VAL_1X_NATIVE_JA "1x (ネイティブ)"
#define OPTION_VAL_2X_JA NULL
#define OPTION_VAL_4X_JA NULL
#define OPTION_VAL_8X_JA NULL
#define OPTION_VAL_16X_JA NULL
#define OPTION_VAL_32X_JA NULL
#define BEETLE_OPT_GTE_OVERCLOCK_LABEL_JA "GTE オーバークロック"
#define BEETLE_OPT_GTE_OVERCLOCK_INFO_0_JA "すべてのエミュレートされた GTE (3D グラフィック用 CPU コプロセッサ) 動画を一定の 1 サイクルレイテンシに低減します。GTE を多用するゲームでフレームレートとフレームタイムの安定性が大幅に向上します。"
#define BEETLE_OPT_CD_ACCESS_METHOD_LABEL_JA NULL
#define BEETLE_OPT_CD_ACCESS_METHOD_INFO_0_JA NULL
#define OPTION_VAL_SYNC_JA "同期"
#define OPTION_VAL_ASYNC_JA "非同期"
#define OPTION_VAL_PRECACHE_JA "プリキャッシュ"
#define BEETLE_OPT_CD_FASTLOAD_LABEL_JA "CD 読み込み速度"
#define BEETLE_OPT_CD_FASTLOAD_INFO_0_JA NULL
#define OPTION_VAL_2X_NATIVE_JA "2x (ネイティブ)"
#define OPTION_VAL_6X_JA NULL
#define OPTION_VAL_10X_JA NULL
#define OPTION_VAL_12X_JA NULL
#define OPTION_VAL_14X_JA NULL
#define BEETLE_OPT_SKIP_BIOS_LABEL_JA "BIOS をスキップ"
#define BEETLE_OPT_SKIP_BIOS_INFO_0_JA "コンテンツのロード時に通常表示される PlayStation BIOS 起動アニメーションをスキップします。注意: これのオプションを有効にすると、多くのゲームで互換性の問題が発生します (PAL コピープロテクトが存在するゲーム、サガフロンティアなど)。"
#define BEETLE_OPT_OVERRIDE_BIOS_LABEL_JA NULL
#define BEETLE_OPT_OVERRIDE_BIOS_INFO_0_JA NULL
#define OPTION_VAL_PSXONPSP_JA NULL
#define OPTION_VAL_PS1_ROM_JA NULL
#define OPTION_VAL_OPENBIOS_JA NULL
#define BEETLE_OPT_REGION_LABEL_JA "システム地域"
#define BEETLE_OPT_REGION_INFO_0_JA NULL
#define OPTION_VAL_AUTO_JA "自動"
#define OPTION_VAL_NTSC_J_JA NULL
#define OPTION_VAL_NTSC_U_JA NULL
#define OPTION_VAL_PAL_JA NULL
#define BEETLE_OPT_INTERNAL_RESOLUTION_LABEL_JA "内部 GPU 解像度"
#define BEETLE_OPT_INTERNAL_RESOLUTION_INFO_0_JA NULL
#define BEETLE_OPT_RENDERER_LABEL_JA NULL
#define BEETLE_OPT_RENDERER_INFO_0_JA NULL
#define OPTION_VAL_HARDWARE_JA "ハードウェア (自動)"
#define OPTION_VAL_HARDWARE_GL_JA "ハードウェア (OpenGL)"
#define OPTION_VAL_HARDWARE_VK_JA "ハードウェア (Vulkan)"
#define OPTION_VAL_SOFTWARE_JA "ソフトウェア"
#define BEETLE_OPT_RENDERER_SOFTWARE_FB_LABEL_JA "ソフトウェアフレームバッファ"
#define BEETLE_OPT_RENDERER_SOFTWARE_FB_INFO_0_JA "ソフトウェアレンダラーのコピー (ネイティブ解像度) をバックグラウンドで実行することで、ハードウェアレンダラーを使用中にフレームバッファ効果の正確なエミュレーションを有効にします。無効にすると、これらの操作は省略される (OpenGL) か、GPU でレンダリング (Vulkan) されます。不明な場合は有効のままにしてください。"
#define BEETLE_OPT_DEPTH_LABEL_JA "内部色深度"
#define BEETLE_OPT_DEPTH_INFO_0_JA NULL
#define OPTION_VAL_16BPP_NATIVE_JA "16 bpp (ネイティブ)"
#define OPTION_VAL_32BPP_JA NULL
#define BEETLE_OPT_DITHER_MODE_LABEL_JA "ディザリングパターン"
#define BEETLE_OPT_DITHER_MODE_INFO_0_JA NULL
#define OPTION_VAL_INTERNAL_RESOLUTION_JA "内部解像度"
#define BEETLE_OPT_SCALED_UV_OFFSET_LABEL_JA "テクスチャ UV オフセット"
#define BEETLE_OPT_SCALED_UV_OFFSET_INFO_0_JA "内部解像度が 1x より高い場合、3D ポリゴンのテクスチャをオフセットでサンプリングします。テクスチャの継ぎ目の発生を軽減しますが、意図しないグラフィックの不具合を引き起こす可能性があります。"
#define BEETLE_OPT_FILTER_LABEL_JA "テクスチャフィルタリング"
#define BEETLE_OPT_FILTER_INFO_0_JA NULL
#define OPTION_VAL_NEAREST_JA "ニアレスト"
#define OPTION_VAL_SABR_JA NULL
#define OPTION_VAL_XBR_JA NULL
#define OPTION_VAL_BILINEAR_JA "バイリニア"
#define OPTION_VAL_3_POINT_JA NULL
#define OPTION_VAL_JINC2_JA NULL
#define BEETLE_OPT_FILTER_EXCLUDE_SPRITE_LABEL_JA "スプライトをフィルタリングから除外"
#define BEETLE_OPT_FILTER_EXCLUDE_SPRITE_INFO_0_JA "スプライトをテクスチャフィルタリングの対象から除外します。2D スプライトレンダリングされた背景を持つ様々なゲームで継ぎ目の発生を防止します。最適な効果を得るには、適応型スムージングまたは他のポストプロセスフィルターと併用してください。"
#define OPTION_VAL_OPAQUE_JA "不透明のみ"
#define OPTION_VAL_ALL_JA "不透明と半透明"
#define BEETLE_OPT_FILTER_EXCLUDE_2D_POLYGON_LABEL_JA "2D ポリゴンをフィルタリングから除外"
#define BEETLE_OPT_FILTER_EXCLUDE_2D_POLYGON_INFO_0_JA "2D ポリゴンをテクスチャフィルタリングの対象から除外します。2D ポリゴンはヒューステリックに検出され、不具合が発生する可能性があります。最適な効果を得るためには、適応型スムージングまたは他のポストプロセスフィルターと併せて使用してください。"
#define BEETLE_OPT_ADAPTIVE_SMOOTHING_LABEL_JA "適応型スムージング"
#define BEETLE_OPT_ADAPTIVE_SMOOTHING_INFO_0_JA "3D レンダリングされたオブジェクトをぼかすことなく、2D アートワークと UI 要素を滑らかにします。Vulkan レンダラーでのみサポートされています。"
#define BEETLE_OPT_SUPER_SAMPLING_LABEL_JA "スーパーサンプリング (ネイティブ解像度へのダウンサンプリング)"
#define BEETLE_OPT_SUPER_SAMPLING_INFO_0_JA "レンダリングされたコンテンツをアップスケールされた解像度からネイティブ解像度までダウンサンプリングします。このオプションをより高い内部解像度倍率と併用することで、ネイティブの低解像度で滑らかな 3D オブジェクトを表示することができます。CRT シェーダーを使用する場合に役立ちます。Vulkan レンダラーでのみサポートされています。"
#define BEETLE_OPT_MSAA_LABEL_JA "マルチサンプリングアンチエイリアシング (MSAA)"
#define BEETLE_OPT_MSAA_INFO_0_JA NULL
#define OPTION_VAL_1X_JA NULL
#define BEETLE_OPT_MDEC_YUV_LABEL_JA "MDEC YUV クロマフィルター"
#define BEETLE_OPT_MDEC_YUV_INFO_0_JA "マクロブロックアーティファクト (四角形/ギザギザのエッジ) を減らすことで FMV 再生の品質を向上させます。Vulkan レンダラーでのみサポートされています。"
#define BEETLE_OPT_TRACK_TEXTURES_LABEL_JA "テクスチャを追跡"
#define BEETLE_OPT_TRACK_TEXTURES_INFO_0_JA NULL
#define BEETLE_OPT_DUMP_TEXTURES_LABEL_JA "テクスチャをダンプ"
#define BEETLE_OPT_DUMP_TEXTURES_INFO_0_JA "使用されているテクスチャを <cd>-texture-dump/ にダンプします。"
#define BEETLE_OPT_REPLACE_TEXTURES_LABEL_JA "テクスチャを置換"
#define BEETLE_OPT_REPLACE_TEXTURES_INFO_0_JA "テクスチャを <cd>-texture-replacements/ からロードして置換します。"
#define BEETLE_OPT_PGXP_MODE_LABEL_JA "PGXP 動作モード"
#define BEETLE_OPT_PGXP_MODE_INFO_0_JA NULL
#define OPTION_VAL_MEMORY_ONLY_JA "メモリのみ"
#define OPTION_VAL_MEMORY_CPU_JA "メモリ + CPU (バグあり)"
#define BEETLE_OPT_PGXP_2D_TOL_LABEL_JA "PGXP 2D ジオメトリ許容範囲"
#define BEETLE_OPT_PGXP_2D_TOL_INFO_0_JA "PGXP 動作における顕著なエラーを隠します。この値は、ジオメトリが適切な深度情報を持たない場合に PGXP 値が保持される許容範囲を指定します。"
#define OPTION_VAL_0PX_JA NULL
#define OPTION_VAL_1PX_JA NULL
#define OPTION_VAL_2PX_JA NULL
#define OPTION_VAL_3PX_JA NULL
#define OPTION_VAL_4PX_JA NULL
#define OPTION_VAL_5PX_JA NULL
#define OPTION_VAL_6PX_JA NULL
#define OPTION_VAL_7PX_JA NULL
#define OPTION_VAL_8PX_JA NULL
#define BEETLE_OPT_PGXP_NCLIP_LABEL_JA "PGXP プリミティブカリング"
#define BEETLE_OPT_PGXP_NCLIP_INFO_0_JA "PGXP の NCLIP 実装を使用します。 PGXP 座標を持つジオメトリの穴を減らすことで外観を改善します。一部のゲームは様々な状況でフリーズすることが判明しています。"
#define BEETLE_OPT_PGXP_VERTEX_LABEL_JA "PGXP 頂点キャッシュ"
#define BEETLE_OPT_PGXP_VERTEX_INFO_0_JA "ポリゴン描画全体で再利用するため、PGXP 拡張頂点位置をキャッシュします。テクスチャのレンダリング時にオブジェクトのアライメントを改善し、表示される継ぎ目を減らすことができますが、キャッシュの照会時に誤検知が発生すると描画バグが発生する可能性があります。無効にすることをおすすめします。ハードウェアレンダラーでのみサポートされています。"
#define BEETLE_OPT_PGXP_TEXTURE_LABEL_JA "PGXP 視点修正テクスチャリング"
#define BEETLE_OPT_PGXP_TEXTURE_INFO_0_JA NULL
#define BEETLE_OPT_DISPLAY_INTERNAL_FPS_LABEL_JA "内部 FPS を表示"
#define BEETLE_OPT_DISPLAY_INTERNAL_FPS_INFO_0_JA "エミュレートされた PlayStation システムが描画しているコンテンツの内部フレームレートを表示します。注意: Libretro フロントエンドで OSD 通知を有効にする必要があります。"
#define BEETLE_OPT_DISPLAY_OSD_LABEL_JA NULL
#define BEETLE_OPT_DISPLAY_OSD_INFO_0_JA NULL
#define BEETLE_OPT_DISPLAY_VRAM_LABEL_JA "フル VRAM を表示 (デバッグ)"
#define BEETLE_OPT_DISPLAY_VRAM_INFO_0_JA "エミュレートされたコンソールの VRAM 全体を可視化します。OpenGL と Vulkan ハードウェアレンダラーでのみサポートされています。注意: これはデバッグ用であり、通常は無効にする必要があります。"
#define BEETLE_OPT_ANALOG_CALIBRATION_LABEL_JA "アナログ自動キャリブレーション"
#define BEETLE_OPT_ANALOG_CALIBRATION_INFO_0_JA NULL
#define BEETLE_OPT_ANALOG_TOGGLE_LABEL_JA "DUALSHOCK アナログモード切り替え"
#define BEETLE_OPT_ANALOG_TOGGLE_INFO_0_JA "入力デバイスの種類が DUALSHOCK の場合、このオプションを使用すると、エミュレートされた DUALSHOCK を元のハードウェアと同様にデジタル/アナログモードの間で切り替えることができます。選択したボタンの組み合わせでモードを切り替えることもできます。"
#define OPTION_VAL_ENABLED_ANALOG_JA "デフォルトアナログ"
#define BEETLE_OPT_ANALOG_TOGGLE_COMBO_LABEL_JA "DUALSHOCK アナログモード切り替えコンボ"
#define BEETLE_OPT_ANALOG_TOGGLE_COMBO_INFO_0_JA NULL
#define OPTION_VAL_L1_L2_R1_R2_START_SELECT_JA "L1 + L2 + R1 + R2 + スタート + セレクト"
#define OPTION_VAL_L1_R1_SELECT_JA "L1 + R1 + セレクト"
#define OPTION_VAL_L1_R1_START_JA "L1 + R1 + スタート"
#define OPTION_VAL_L1_R1_L3_JA NULL
#define OPTION_VAL_L1_R1_R3_JA NULL
#define OPTION_VAL_L2_R2_SELECT_JA "L2 + R2 + セレクト"
#define OPTION_VAL_L2_R2_START_JA "L2 + R2 + スタート"
#define OPTION_VAL_L2_R2_L3_JA NULL
#define OPTION_VAL_L2_R2_R3_JA NULL
#define OPTION_VAL_L3_R3_JA NULL
#define BEETLE_OPT_ANALOG_TOGGLE_HOLD_LABEL_JA "DUALSHOCK アナログモード切り替えコンボの長押し遅延"
#define BEETLE_OPT_ANALOG_TOGGLE_HOLD_INFO_0_JA NULL
#define OPTION_VAL_0_JA "0 秒遅延"
#define OPTION_VAL_1_JA "1 秒遅延"
#define OPTION_VAL_2_JA "2 秒遅延"
#define OPTION_VAL_3_JA "3 秒遅延"
#define OPTION_VAL_4_JA "4 秒遅延"
#define OPTION_VAL_5_JA "5 秒遅延"
#define BEETLE_OPT_ENABLE_MULTITAP_PORT1_LABEL_JA NULL
#define BEETLE_OPT_ENABLE_MULTITAP_PORT1_INFO_0_JA "マルチタップ機能をポート 1 で有効にします。"
#define BEETLE_OPT_ENABLE_MULTITAP_PORT2_LABEL_JA NULL
#define BEETLE_OPT_ENABLE_MULTITAP_PORT2_INFO_0_JA "マルチタップ機能をポート 2 で有効にします。"
#define BEETLE_OPT_GUN_INPUT_MODE_LABEL_JA "ガン入力モード"
#define BEETLE_OPT_GUN_INPUT_MODE_INFO_0_JA NULL
#define OPTION_VAL_LIGHTGUN_JA "ライトガン"
#define OPTION_VAL_TOUCHSCREEN_JA "タッチスクリーン"
#define BEETLE_OPT_GUN_CURSOR_LABEL_JA "ガンカーソル"
#define BEETLE_OPT_GUN_CURSOR_INFO_0_JA NULL
#define OPTION_VAL_CROSS_JA "クロス"
#define OPTION_VAL_DOT_JA "ドット"
#define BEETLE_OPT_CROSSHAIR_COLOR_P1_LABEL_JA NULL
#define BEETLE_OPT_CROSSHAIR_COLOR_P1_INFO_0_JA NULL
#define OPTION_VAL_RED_JA "レッド"
#define OPTION_VAL_BLUE_JA "ブルー"
#define OPTION_VAL_GREEN_JA "グリーン"
#define OPTION_VAL_ORANGE_JA "オレンジ"
#define OPTION_VAL_YELLOW_JA "イエロー"
#define OPTION_VAL_CYAN_JA "シアン"
#define OPTION_VAL_PINK_JA "ピンク"
#define OPTION_VAL_PURPLE_JA "紫"
#define OPTION_VAL_BLACK_JA "黒"
#define OPTION_VAL_WHITE_JA "白"
#define BEETLE_OPT_CROSSHAIR_COLOR_P2_LABEL_JA NULL
#define BEETLE_OPT_CROSSHAIR_COLOR_P2_INFO_0_JA NULL
#define BEETLE_OPT_MOUSE_SENSITIVITY_LABEL_JA "マウス感度"
#define BEETLE_OPT_MOUSE_SENSITIVITY_INFO_0_JA NULL
#define OPTION_VAL_5_O42_JA NULL
#define OPTION_VAL_10_JA NULL
#define OPTION_VAL_15_JA NULL
#define OPTION_VAL_20_JA NULL
#define OPTION_VAL_25_JA NULL
#define OPTION_VAL_30_JA NULL
#define OPTION_VAL_35_JA NULL
#define OPTION_VAL_40_JA NULL
#define OPTION_VAL_45_JA NULL
#define OPTION_VAL_55_JA NULL
#define OPTION_VAL_65_JA NULL
#define OPTION_VAL_75_JA NULL
#define OPTION_VAL_85_JA NULL
#define OPTION_VAL_95_JA NULL
#define OPTION_VAL_100_O42_JA NULL
#define OPTION_VAL_105_JA NULL
#define OPTION_VAL_115_JA NULL
#define OPTION_VAL_125_JA NULL
#define OPTION_VAL_135_JA NULL
#define OPTION_VAL_145_JA NULL
#define OPTION_VAL_155_JA NULL
#define OPTION_VAL_165_JA NULL
#define OPTION_VAL_175_JA NULL
#define OPTION_VAL_185_JA NULL
#define OPTION_VAL_195_JA NULL
#define BEETLE_OPT_NEGCON_RESPONSE_LABEL_JA "ネジコンねじり応答"
#define BEETLE_OPT_NEGCON_RESPONSE_INFO_0_JA NULL
#define OPTION_VAL_LINEAR_JA "線形"
#define OPTION_VAL_QUADRATIC_JA "平方"
#define OPTION_VAL_CUBIC_JA "立方"
#define BEETLE_OPT_NEGCON_DEADZONE_LABEL_JA "ネジコンねじりデッドゾーン"
#define BEETLE_OPT_NEGCON_DEADZONE_INFO_0_JA NULL
#define OPTION_VAL_0_O44_JA NULL
#define BEETLE_OPT_USE_MEDNAFEN_MEMCARD0_METHOD_LABEL_JA NULL
#define BEETLE_OPT_USE_MEDNAFEN_MEMCARD0_METHOD_INFO_0_JA NULL
#define OPTION_VAL_LIBRETRO_JA NULL
#define OPTION_VAL_MEDNAFEN_JA NULL
#define BEETLE_OPT_ENABLE_MEMCARD1_LABEL_JA NULL
#define BEETLE_OPT_ENABLE_MEMCARD1_INFO_0_JA NULL
#define BEETLE_OPT_SHARED_MEMORY_CARDS_LABEL_JA NULL
#define BEETLE_OPT_SHARED_MEMORY_CARDS_INFO_0_JA NULL
#define BEETLE_OPT_MEMCARD_LEFT_INDEX_LABEL_JA NULL
#define BEETLE_OPT_MEMCARD_LEFT_INDEX_INFO_0_JA NULL
#define OPTION_VAL_0_O48_JA NULL
#define BEETLE_OPT_MEMCARD_RIGHT_INDEX_LABEL_JA NULL
#define BEETLE_OPT_MEMCARD_RIGHT_INDEX_INFO_0_JA NULL
#define OPTION_VAL_1_O49_JA NULL
#define BEETLE_OPT_LINE_RENDER_LABEL_JA "線形二次ハック"
#define BEETLE_OPT_LINE_RENDER_INFO_0_JA NULL
#define OPTION_VAL_DEFAULT_JA "デフォルト"
#define OPTION_VAL_AGGRESSIVE_JA "アグレッシブ"
#define BEETLE_OPT_WIDESCREEN_HACK_LABEL_JA "ワイドスクリーンモードハック"
#define BEETLE_OPT_WIDESCREEN_HACK_INFO_0_JA "3D コンテンツをアナモルフィック状態でレンダリングし、エミュレートされたフレームバッファをワイドスクリーンアスペクト比で出力します。フルポリゴン 3D ゲームで最高の結果を生成します。2D 要素は水平方向に引き伸ばされ、配置がずれる可能性があります。"
#define BEETLE_OPT_WIDESCREEN_HACK_ASPECT_RATIO_LABEL_JA "ワイドスクリーンモードハックのアスペクト比"
#define BEETLE_OPT_WIDESCREEN_HACK_ASPECT_RATIO_INFO_0_JA NULL
#define OPTION_VAL_16_9_JA NULL
#define OPTION_VAL_16_10_JA NULL
#define OPTION_VAL_18_9_JA NULL
#define OPTION_VAL_19_9_JA NULL
#define OPTION_VAL_20_9_JA NULL
#define OPTION_VAL_21_9_JA NULL
#define OPTION_VAL_32_9_JA NULL
#define BEETLE_OPT_CPU_DYNAREC_LABEL_JA "CPU 動的再コンパイル"
#define BEETLE_OPT_CPU_DYNAREC_INFO_0_JA "CPU 命令をネイティブ命令に動的に再コンパイルします。インタプリタよりもはるかに高速ですが、CPU タイミングは正確ではなく、バグがある可能性があります。"
#define OPTION_VAL_DISABLED_JA "無効 (Beetle インタプリタ)"
#define OPTION_VAL_EXECUTE_JA "最大パフォーマンス"
#define OPTION_VAL_EXECUTE_ONE_JA "サイクルタイミングチェック"
#define OPTION_VAL_RUN_INTERPRETER_JA "Lightrec インタプリタ"
#define BEETLE_OPT_DYNAREC_INVALIDATE_LABEL_JA "動的再コンパイルコードの無効化"
#define BEETLE_OPT_DYNAREC_INVALIDATE_INFO_0_JA "ゲームによって [フル] 無効化を必要とするものと [DMA のみ] を必要とするものがあります。"
#define OPTION_VAL_FULL_JA "フル"
#define OPTION_VAL_DMA_JA "DMA のみ (やや早い)"
#define BEETLE_OPT_DYNAREC_EVENTCYCLES_LABEL_JA "動的再コンパイル DMA/GPU/MDEC/Timer イベントサイクル"
#define BEETLE_OPT_DYNAREC_EVENTCYCLES_INFO_0_JA "GPU/DMA/MDEC/Timer 更新がチェックされる前に CPU によって実行される最大サイクル数です。値を高くするほど高速になりますが、Beetle インタプリタにはほとんど影響を与えません。"
#define OPTION_VAL_128_JA "128 (デフォルト)"
#define BEETLE_OPT_DYNAREC_SPU_SAMPLES_LABEL_JA "動的再コンパイル SPU サンプル"
#define BEETLE_OPT_DYNAREC_SPU_SAMPLES_INFO_0_JA "SPU 更新がチェックされる前に実行される最大 SPU サンプル数です。値を高くするほど高速になりますが、1 以外に設定すると一部のゲームで音声バグが発生する可能性があります。"
#define BEETLE_OPT_CORE_TIMING_FPS_LABEL_JA "コア報告 FPS タイミング"
#define BEETLE_OPT_CORE_TIMING_FPS_INFO_0_JA NULL
#define OPTION_VAL_AUTO_TOGGLE_JA NULL
#define OPTION_VAL_FORCE_PROGRESSIVE_JA NULL
#define OPTION_VAL_FORCE_INTERLACED_JA "強制インターレースレート"
#define BEETLE_OPT_DEINTERLACER_LABEL_JA "インターレース解除方式"
#define BEETLE_OPT_DEINTERLACER_INFO_0_JA NULL
#define OPTION_VAL_WEAVE_JA NULL
#define OPTION_VAL_BOB_JA NULL
#define OPTION_VAL_BOB_OFFSET_JA NULL
#define OPTION_VAL_FASTMAD_JA NULL
#define BEETLE_OPT_FRAME_DUPING_LABEL_JA NULL
#define BEETLE_OPT_FRAME_DUPING_INFO_0_JA NULL
#define BEETLE_OPT_PAL_VIDEO_TIMING_OVERRIDE_LABEL_JA NULL
#define BEETLE_OPT_PAL_VIDEO_TIMING_OVERRIDE_INFO_0_JA NULL
#define BEETLE_OPT_ASPECT_RATIO_LABEL_JA "コアアスペクト比"
#define BEETLE_OPT_ASPECT_RATIO_INFO_0_JA NULL
#define OPTION_VAL_CORRECTED_JA "補正あり"
#define OPTION_VAL_UNCORRECTED_JA "補正なし"
#define OPTION_VAL_4_3_JA "強制 4:3"
#define OPTION_VAL_NTSC_JA "強制 NTSC"
#define BEETLE_OPT_CROP_OVERSCAN_LABEL_JA "オーバースキャンをトリミング"
#define BEETLE_OPT_CROP_OVERSCAN_INFO_0_JA NULL
#define OPTION_VAL_STATIC_JA "水平"
#define OPTION_VAL_SMART_JA NULL
#define BEETLE_OPT_IMAGE_CROP_LABEL_JA "追加トリミング"
#define BEETLE_OPT_IMAGE_CROP_INFO_0_JA NULL
#define OPTION_VAL_9PX_JA NULL
#define OPTION_VAL_10PX_JA NULL
#define OPTION_VAL_11PX_JA NULL
#define OPTION_VAL_12PX_JA NULL
#define OPTION_VAL_13PX_JA NULL
#define OPTION_VAL_14PX_JA NULL
#define OPTION_VAL_15PX_JA NULL
#define OPTION_VAL_16PX_JA NULL
#define OPTION_VAL_17PX_JA NULL
#define OPTION_VAL_18PX_JA NULL
#define OPTION_VAL_19PX_JA NULL
#define OPTION_VAL_20PX_JA NULL
#define BEETLE_OPT_IMAGE_OFFSET_LABEL_JA "トリミングされた画像のオフセット"
#define BEETLE_OPT_IMAGE_OFFSET_INFO_0_JA NULL
#define OPTION_VAL_12PX_O64_JA NULL
#define OPTION_VAL_11PX_O64_JA NULL
#define OPTION_VAL_10PX_O64_JA NULL
#define OPTION_VAL_9PX_O64_JA NULL
#define OPTION_VAL_8PX_O64_JA NULL
#define OPTION_VAL_7PX_O64_JA NULL
#define OPTION_VAL_6PX_O64_JA NULL
#define OPTION_VAL_5PX_O64_JA NULL
#define OPTION_VAL_4PX_O64_JA NULL
#define OPTION_VAL_3PX_O64_JA NULL
#define OPTION_VAL_2PX_O64_JA NULL
#define OPTION_VAL_1PX_O64_JA NULL
#define OPTION_VAL_1PX_O64_0_JA NULL
#define OPTION_VAL_2PX_O64_0_JA NULL
#define OPTION_VAL_3PX_O64_0_JA NULL
#define OPTION_VAL_4PX_O64_0_JA NULL
#define OPTION_VAL_5PX_O64_0_JA NULL
#define OPTION_VAL_6PX_O64_0_JA NULL
#define OPTION_VAL_7PX_O64_0_JA NULL
#define OPTION_VAL_8PX_O64_0_JA NULL
#define OPTION_VAL_9PX_O64_0_JA NULL
#define OPTION_VAL_10PX_O64_0_JA NULL
#define OPTION_VAL_11PX_O64_0_JA NULL
#define OPTION_VAL_12PX_O64_0_JA NULL
#define BEETLE_OPT_IMAGE_OFFSET_CYCLES_LABEL_JA "水平イメージオフセット (GPU サイクル)"
#define BEETLE_OPT_IMAGE_OFFSET_CYCLES_INFO_0_JA NULL
#define BEETLE_OPT_INITIAL_SCANLINE_LABEL_JA "開始スキャンライン - NTSC"
#define BEETLE_OPT_INITIAL_SCANLINE_INFO_0_JA NULL
#define BEETLE_OPT_LAST_SCANLINE_LABEL_JA "終了スキャンライン - NTSC"
#define BEETLE_OPT_LAST_SCANLINE_INFO_0_JA NULL
#define OPTION_VAL_239_JA "239 (デフォルト)"
#define BEETLE_OPT_INITIAL_SCANLINE_PAL_LABEL_JA "開始スキャンライン - PAL"
#define BEETLE_OPT_INITIAL_SCANLINE_PAL_INFO_0_JA NULL
#define BEETLE_OPT_LAST_SCANLINE_PAL_LABEL_JA "終了スキャンライン - PAL"
#define BEETLE_OPT_LAST_SCANLINE_PAL_INFO_0_JA NULL
#define OPTION_VAL_287_JA "287 (デフォルト)"

struct retro_core_option_v2_category option_cats_ja[] = {
   {
      "system",
      CATEGORY_SYSTEM_LABEL_JA,
      CATEGORY_SYSTEM_INFO_0_JA
   },
   {
      "video",
      CATEGORY_VIDEO_LABEL_JA,
      CATEGORY_VIDEO_INFO_0_JA
   },
   {
      "pgxp",
      CATEGORY_PGXP_LABEL_JA,
      CATEGORY_PGXP_INFO_0_JA
   },
   {
      "osd",
      CATEGORY_OSD_LABEL_JA,
      CATEGORY_OSD_INFO_0_JA
   },
   {
      "input",
      CATEGORY_INPUT_LABEL_JA,
      CATEGORY_INPUT_INFO_0_JA
   },
   {
      "memcards",
      CATEGORY_MEMCARDS_LABEL_JA,
      CATEGORY_MEMCARDS_INFO_0_JA
   },
   {
      "hacks",
      CATEGORY_HACKS_LABEL_JA,
      CATEGORY_HACKS_INFO_0_JA
   },
   { NULL, NULL, NULL },
};
struct retro_core_option_v2_definition option_defs_ja[] = {
   {
      BEETLE_OPT(cpu_freq_scale),
      BEETLE_OPT_CPU_FREQ_SCALE_LABEL_JA,
      NULL,
      BEETLE_OPT_CPU_FREQ_SCALE_INFO_0_JA,
      NULL,
      "system",
      {
         { "50%",           OPTION_VAL_50_JA },
         { "60%",           OPTION_VAL_60_JA },
         { "70%",           OPTION_VAL_70_JA },
         { "80%",           OPTION_VAL_80_JA },
         { "90%",           OPTION_VAL_90_JA },
         { "100%",          OPTION_VAL_100_JA },
         { "110%",          OPTION_VAL_110_JA },
         { "120%",          OPTION_VAL_120_JA },
         { "130%",          OPTION_VAL_130_JA },
         { "140%",          OPTION_VAL_140_JA },
         { "150%",          OPTION_VAL_150_JA },
         { "160%",          OPTION_VAL_160_JA },
         { "170%",          OPTION_VAL_170_JA },
         { "180%",          OPTION_VAL_180_JA },
         { "190%",          OPTION_VAL_190_JA },
         { "200%",          OPTION_VAL_200_JA },
         { "210%",          OPTION_VAL_210_JA },
         { "220%",          OPTION_VAL_220_JA },
         { "230%",          OPTION_VAL_230_JA },
         { "240%",          OPTION_VAL_240_JA },
         { "250%",          OPTION_VAL_250_JA },
         { "260%",          OPTION_VAL_260_JA },
         { "270%",          OPTION_VAL_270_JA },
         { "280%",          OPTION_VAL_280_JA },
         { "290%",          OPTION_VAL_290_JA },
         { "300%",          OPTION_VAL_300_JA },
         { "310%",          OPTION_VAL_310_JA },
         { "320%",          OPTION_VAL_320_JA },
         { "330%",          OPTION_VAL_330_JA },
         { "340%",          OPTION_VAL_340_JA },
         { "350%",          OPTION_VAL_350_JA },
         { "360%",          OPTION_VAL_360_JA },
         { "370%",          OPTION_VAL_370_JA },
         { "380%",          OPTION_VAL_380_JA },
         { "390%",          OPTION_VAL_390_JA },
         { "400%",          OPTION_VAL_400_JA },
         { "410%",          OPTION_VAL_410_JA },
         { "420%",          OPTION_VAL_420_JA },
         { "430%",          OPTION_VAL_430_JA },
         { "440%",          OPTION_VAL_440_JA },
         { "450%",          OPTION_VAL_450_JA },
         { "460%",          OPTION_VAL_460_JA },
         { "470%",          OPTION_VAL_470_JA },
         { "480%",          OPTION_VAL_480_JA },
         { "490%",          OPTION_VAL_490_JA },
         { "500%",          OPTION_VAL_500_JA },
         { "510%",          OPTION_VAL_510_JA },
         { "520%",          OPTION_VAL_520_JA },
         { "530%",          OPTION_VAL_530_JA },
         { "540%",          OPTION_VAL_540_JA },
         { "550%",          OPTION_VAL_550_JA },
         { "560%",          OPTION_VAL_560_JA },
         { "570%",          OPTION_VAL_570_JA },
         { "580%",          OPTION_VAL_580_JA },
         { "590%",          OPTION_VAL_590_JA },
         { "600%",          OPTION_VAL_600_JA },
         { "610%",          OPTION_VAL_610_JA },
         { "620%",          OPTION_VAL_620_JA },
         { "630%",          OPTION_VAL_630_JA },
         { "640%",          OPTION_VAL_640_JA },
         { "650%",          OPTION_VAL_650_JA },
         { "660%",          OPTION_VAL_660_JA },
         { "670%",          OPTION_VAL_670_JA },
         { "680%",          OPTION_VAL_680_JA },
         { "690%",          OPTION_VAL_690_JA },
         { "700%",          OPTION_VAL_700_JA },
         { "710%",          OPTION_VAL_710_JA },
         { "720%",          OPTION_VAL_720_JA },
         { "730%",          OPTION_VAL_730_JA },
         { "740%",          OPTION_VAL_740_JA },
         { "750%",          OPTION_VAL_750_JA },
         { NULL, NULL },
      },
      "100%"
   },
   {
      BEETLE_OPT(gpu_overclock),
      BEETLE_OPT_GPU_OVERCLOCK_LABEL_JA,
      NULL,
      BEETLE_OPT_GPU_OVERCLOCK_INFO_0_JA,
      NULL,
      "system",
      {
         { "1x(native)", OPTION_VAL_1X_NATIVE_JA },
         { "2x",         OPTION_VAL_2X_JA },
         { "4x",         OPTION_VAL_4X_JA },
         { "8x",         OPTION_VAL_8X_JA },
         { "16x",        OPTION_VAL_16X_JA },
         { "32x",        OPTION_VAL_32X_JA },
         { NULL, NULL },
      },
      "1x(native)"
   },
   {
      BEETLE_OPT(gte_overclock),
      BEETLE_OPT_GTE_OVERCLOCK_LABEL_JA,
      NULL,
      BEETLE_OPT_GTE_OVERCLOCK_INFO_0_JA,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
#ifndef EMSCRIPTEN
   {
      BEETLE_OPT(cd_access_method),
      BEETLE_OPT_CD_ACCESS_METHOD_LABEL_JA,
      NULL,
      BEETLE_OPT_CD_ACCESS_METHOD_INFO_0_JA,
      NULL,
      "system",
      {
         { "sync",     OPTION_VAL_SYNC_JA },
         { "async",    OPTION_VAL_ASYNC_JA },
         { "precache", OPTION_VAL_PRECACHE_JA },
         { NULL, NULL },
      },
      "sync"
   },
#endif
   {
      BEETLE_OPT(cd_fastload),
      BEETLE_OPT_CD_FASTLOAD_LABEL_JA,
      NULL,
      BEETLE_OPT_CD_FASTLOAD_INFO_0_JA,
      NULL,
      "system",
      {
         { "2x(native)", OPTION_VAL_2X_NATIVE_JA },
         { "4x",          OPTION_VAL_4X_JA },
         { "6x",          OPTION_VAL_6X_JA },
         { "8x",          OPTION_VAL_8X_JA },
         { "10x",         OPTION_VAL_10X_JA },
         { "12x",         OPTION_VAL_12X_JA },
         { "14x",         OPTION_VAL_14X_JA },
         { NULL, NULL },
      },
      "2x(native)"
   },
   {
      BEETLE_OPT(skip_bios),
      BEETLE_OPT_SKIP_BIOS_LABEL_JA,
      NULL,
      BEETLE_OPT_SKIP_BIOS_INFO_0_JA,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(override_bios),
      BEETLE_OPT_OVERRIDE_BIOS_LABEL_JA,
      NULL,
      BEETLE_OPT_OVERRIDE_BIOS_INFO_0_JA,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "psxonpsp", OPTION_VAL_PSXONPSP_JA },
         { "ps1_rom",  OPTION_VAL_PS1_ROM_JA },
         { "openbios", OPTION_VAL_OPENBIOS_JA     },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(region),
      BEETLE_OPT_REGION_LABEL_JA,
      NULL,
      BEETLE_OPT_REGION_INFO_0_JA,
      NULL,
      "system",
      {
         { "auto",   OPTION_VAL_AUTO_JA },
         { "ntsc-j", OPTION_VAL_NTSC_J_JA },
         { "ntsc-u", OPTION_VAL_NTSC_U_JA },
         { "pal",    OPTION_VAL_PAL_JA },
         { NULL, NULL },
      },
      "auto"
   },
   {
      BEETLE_OPT(internal_resolution),
      BEETLE_OPT_INTERNAL_RESOLUTION_LABEL_JA,
      NULL,
      BEETLE_OPT_INTERNAL_RESOLUTION_INFO_0_JA,
      NULL,
      "video",
      {
         { "1x(native)", OPTION_VAL_1X_NATIVE_JA },
         { "2x",         OPTION_VAL_2X_JA },
         { "4x",         OPTION_VAL_4X_JA },
         { "8x",         OPTION_VAL_8X_JA },
         { "16x",        OPTION_VAL_16X_JA },
         { NULL, NULL },
      },
      "1x(native)"
   },
#if defined(HAVE_OPENGL) || defined(HAVE_OPENGLES) || defined(HAVE_VULKAN)
   {
      BEETLE_OPT(renderer),
      BEETLE_OPT_RENDERER_LABEL_JA,
      NULL,
      BEETLE_OPT_RENDERER_INFO_0_JA,
      NULL,
      "video",
      {
         { "hardware",    OPTION_VAL_HARDWARE_JA },
#if defined(HAVE_OPENGL) || defined(HAVE_OPENGLES)
         { "hardware_gl", OPTION_VAL_HARDWARE_GL_JA },
#endif
#if defined(HAVE_VULKAN)
         { "hardware_vk", OPTION_VAL_HARDWARE_VK_JA },
#endif
         { "software",    OPTION_VAL_SOFTWARE_JA },
         { NULL, NULL },
      },
      "hardware"
   },
   {
      BEETLE_OPT(renderer_software_fb),
      BEETLE_OPT_RENDERER_SOFTWARE_FB_LABEL_JA,
      NULL,
      BEETLE_OPT_RENDERER_SOFTWARE_FB_INFO_0_JA,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled"
   },
#endif
#if defined(HAVE_OPENGL) || defined(HAVE_OPENGLES)
   {
      BEETLE_OPT(depth),
      BEETLE_OPT_DEPTH_LABEL_JA,
      NULL,
      BEETLE_OPT_DEPTH_INFO_0_JA,
      NULL,
      "video",
      {
         { "16bpp(native)", OPTION_VAL_16BPP_NATIVE_JA },
         { "32bpp",         OPTION_VAL_32BPP_JA },
         { NULL, NULL },
      },
      "16bpp(native)"
   },
   // Sort of, it's more like 15-bit high color and 24-bit true color for visible output. The alpha channel is used for mask bit. Vulkan renderer uses ABGR1555_555 for 31 bits internal? FMVs are always 24-bit on all renderers like original hardware (BGR888, no alpha)
#endif
   {
      BEETLE_OPT(dither_mode),
      BEETLE_OPT_DITHER_MODE_LABEL_JA,
      NULL,
      BEETLE_OPT_DITHER_MODE_INFO_0_JA,
      NULL,
      "video",
      {
         { "1x(native)",          OPTION_VAL_1X_NATIVE_JA },
         { "internal resolution", OPTION_VAL_INTERNAL_RESOLUTION_JA },
         { "disabled",            NULL },
         { NULL, NULL },
      },
      "1x(native)"
   },
#ifdef HAVE_VULKAN
   {
      BEETLE_OPT(scaled_uv_offset),
      BEETLE_OPT_SCALED_UV_OFFSET_LABEL_JA,
      NULL,
      BEETLE_OPT_SCALED_UV_OFFSET_INFO_0_JA,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled"
   },
#endif
#if defined(HAVE_OPENGL) || defined(HAVE_OPENGLES) || defined(HAVE_VULKAN)
   {
      BEETLE_OPT(filter),
      BEETLE_OPT_FILTER_LABEL_JA,
      NULL,
      BEETLE_OPT_FILTER_INFO_0_JA,
      NULL,
      "video",
      {
         { "nearest",  OPTION_VAL_NEAREST_JA },
         { "SABR",     OPTION_VAL_SABR_JA },
         { "xBR",      OPTION_VAL_XBR_JA },
         { "bilinear", OPTION_VAL_BILINEAR_JA },
         { "3-point",  OPTION_VAL_3_POINT_JA },
         { "JINC2",    OPTION_VAL_JINC2_JA },
         { NULL, NULL },
      },
      "nearest"
   },
#ifdef HAVE_VULKAN
   {
      BEETLE_OPT(filter_exclude_sprite),
      BEETLE_OPT_FILTER_EXCLUDE_SPRITE_LABEL_JA,
      NULL,
      BEETLE_OPT_FILTER_EXCLUDE_SPRITE_INFO_0_JA,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "opaque", OPTION_VAL_OPAQUE_JA },
         { "all", OPTION_VAL_ALL_JA },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(filter_exclude_2d_polygon),
      BEETLE_OPT_FILTER_EXCLUDE_2D_POLYGON_LABEL_JA,
      NULL,
      BEETLE_OPT_FILTER_EXCLUDE_2D_POLYGON_INFO_0_JA,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "opaque", OPTION_VAL_OPAQUE_JA },
         { "all", OPTION_VAL_ALL_JA },
         { NULL, NULL },
      },
      "disabled"
   },
#endif
#endif
#ifdef HAVE_VULKAN
   {
      BEETLE_OPT(adaptive_smoothing),
      BEETLE_OPT_ADAPTIVE_SMOOTHING_LABEL_JA,
      NULL,
      BEETLE_OPT_ADAPTIVE_SMOOTHING_INFO_0_JA,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(super_sampling),
      BEETLE_OPT_SUPER_SAMPLING_LABEL_JA,
      NULL,
      BEETLE_OPT_SUPER_SAMPLING_INFO_0_JA,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(msaa),
      BEETLE_OPT_MSAA_LABEL_JA,
      NULL,
      BEETLE_OPT_MSAA_INFO_0_JA,
      NULL,
      "video",
      {
         { "1x",  OPTION_VAL_1X_JA },
         { "2x",  OPTION_VAL_2X_JA },
         { "4x",  OPTION_VAL_4X_JA },
         { "8x",  OPTION_VAL_8X_JA },
         { "16x", OPTION_VAL_16X_JA },
         { NULL, NULL },
      },
      "1x"
   },
   {
      BEETLE_OPT(mdec_yuv),
      BEETLE_OPT_MDEC_YUV_LABEL_JA,
      NULL,
      BEETLE_OPT_MDEC_YUV_INFO_0_JA,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(track_textures),
      BEETLE_OPT_TRACK_TEXTURES_LABEL_JA,
      NULL,
      BEETLE_OPT_TRACK_TEXTURES_INFO_0_JA,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
#ifdef TEXTURE_DUMPING_ENABLED
   {
      BEETLE_OPT(dump_textures),
      BEETLE_OPT_DUMP_TEXTURES_LABEL_JA,
      NULL,
      BEETLE_OPT_DUMP_TEXTURES_INFO_0_JA,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
#endif
   {
      BEETLE_OPT(replace_textures),
      BEETLE_OPT_REPLACE_TEXTURES_LABEL_JA,
      NULL,
      BEETLE_OPT_REPLACE_TEXTURES_INFO_0_JA,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
#endif
   {
      BEETLE_OPT(pgxp_mode),
      BEETLE_OPT_PGXP_MODE_LABEL_JA,
      NULL,
      BEETLE_OPT_PGXP_MODE_INFO_0_JA,
      NULL,
      "pgxp",
      {
         { "disabled",     NULL },
         { "memory only",  OPTION_VAL_MEMORY_ONLY_JA },
         { "memory + CPU", OPTION_VAL_MEMORY_CPU_JA },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(pgxp_2d_tol),
      BEETLE_OPT_PGXP_2D_TOL_LABEL_JA,
      NULL,
      BEETLE_OPT_PGXP_2D_TOL_INFO_0_JA,
      NULL,
      "pgxp",
      {
         { "disabled", NULL },
         { "0px", OPTION_VAL_0PX_JA },
         { "1px", OPTION_VAL_1PX_JA },
         { "2px", OPTION_VAL_2PX_JA },
         { "3px", OPTION_VAL_3PX_JA },
         { "4px", OPTION_VAL_4PX_JA },
         { "5px", OPTION_VAL_5PX_JA },
         { "6px", OPTION_VAL_6PX_JA },
         { "7px", OPTION_VAL_7PX_JA },
         { "8px", OPTION_VAL_8PX_JA },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(pgxp_nclip),
      BEETLE_OPT_PGXP_NCLIP_LABEL_JA,
      NULL,
      BEETLE_OPT_PGXP_NCLIP_INFO_0_JA,
      NULL,
      "pgxp",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
#if defined(HAVE_OPENGL) || defined(HAVE_OPENGLES) || defined(HAVE_VULKAN)
   {
      BEETLE_OPT(pgxp_vertex),
      BEETLE_OPT_PGXP_VERTEX_LABEL_JA,
      NULL,
      BEETLE_OPT_PGXP_VERTEX_INFO_0_JA,
      NULL,
      "pgxp",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
#endif
   {
      BEETLE_OPT(pgxp_texture),
      BEETLE_OPT_PGXP_TEXTURE_LABEL_JA,
      NULL,
      BEETLE_OPT_PGXP_TEXTURE_INFO_0_JA,
      NULL,
      "pgxp",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(display_internal_fps),
      BEETLE_OPT_DISPLAY_INTERNAL_FPS_LABEL_JA,
      NULL,
      BEETLE_OPT_DISPLAY_INTERNAL_FPS_INFO_0_JA,
      NULL,
      "osd",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(display_osd),
      BEETLE_OPT_DISPLAY_OSD_LABEL_JA,
      NULL,
      BEETLE_OPT_DISPLAY_OSD_INFO_0_JA,
      NULL,
      "osd",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled"
   },
#if defined(HAVE_OPENGL) || defined(HAVE_OPENGLES) || defined(HAVE_VULKAN)
   {
      BEETLE_OPT(display_vram),
      BEETLE_OPT_DISPLAY_VRAM_LABEL_JA,
      NULL,
      BEETLE_OPT_DISPLAY_VRAM_INFO_0_JA,
      NULL,
      "osd",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
#endif
   {
      BEETLE_OPT(analog_calibration),
      BEETLE_OPT_ANALOG_CALIBRATION_LABEL_JA,
      NULL,
      BEETLE_OPT_ANALOG_CALIBRATION_INFO_0_JA,
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(analog_toggle),
      BEETLE_OPT_ANALOG_TOGGLE_LABEL_JA,
      NULL,
      BEETLE_OPT_ANALOG_TOGGLE_INFO_0_JA,
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { "enabled-analog", OPTION_VAL_ENABLED_ANALOG_JA },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(analog_toggle_combo),
      BEETLE_OPT_ANALOG_TOGGLE_COMBO_LABEL_JA,
      NULL,
      BEETLE_OPT_ANALOG_TOGGLE_COMBO_INFO_0_JA,
      NULL,
      "input",
      {
         { "l1+l2+r1+r2+start+select", OPTION_VAL_L1_L2_R1_R2_START_SELECT_JA },
         { "l1+r1+select",             OPTION_VAL_L1_R1_SELECT_JA },
         { "l1+r1+start",              OPTION_VAL_L1_R1_START_JA },
         { "l1+r1+l3",                 OPTION_VAL_L1_R1_L3_JA },
         { "l1+r1+r3",                 OPTION_VAL_L1_R1_R3_JA },
         { "l2+r2+select",             OPTION_VAL_L2_R2_SELECT_JA },
         { "l2+r2+start",              OPTION_VAL_L2_R2_START_JA },
         { "l2+r2+l3",                 OPTION_VAL_L2_R2_L3_JA },
         { "l2+r2+r3",                 OPTION_VAL_L2_R2_R3_JA },
         { "l3+r3",                    OPTION_VAL_L3_R3_JA },
         { NULL, NULL },
      },
      "l1+r1+select"
   },
   {
      BEETLE_OPT(analog_toggle_hold),
      BEETLE_OPT_ANALOG_TOGGLE_HOLD_LABEL_JA,
      NULL,
      BEETLE_OPT_ANALOG_TOGGLE_HOLD_INFO_0_JA,
      NULL,
      "input",
      {
         { "0", OPTION_VAL_0_JA },
         { "1", OPTION_VAL_1_JA },
         { "2", OPTION_VAL_2_JA },
         { "3", OPTION_VAL_3_JA },
         { "4", OPTION_VAL_4_JA },
         { "5", OPTION_VAL_5_JA },
         { NULL, NULL },
      },
      "1"
   },
   {
      BEETLE_OPT(enable_multitap_port1),
      BEETLE_OPT_ENABLE_MULTITAP_PORT1_LABEL_JA,
      NULL,
      BEETLE_OPT_ENABLE_MULTITAP_PORT1_INFO_0_JA,
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(enable_multitap_port2),
      BEETLE_OPT_ENABLE_MULTITAP_PORT2_LABEL_JA,
      NULL,
      BEETLE_OPT_ENABLE_MULTITAP_PORT2_INFO_0_JA,
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(gun_input_mode),
      BEETLE_OPT_GUN_INPUT_MODE_LABEL_JA,
      NULL,
      BEETLE_OPT_GUN_INPUT_MODE_INFO_0_JA,
      NULL,
      "input",
      {
         { "lightgun",    OPTION_VAL_LIGHTGUN_JA },
         { "touchscreen", OPTION_VAL_TOUCHSCREEN_JA },
         { NULL, NULL },
      },
      "lightgun"
   },
   // Shouldn't the gun_input_mode just be Mouse vs. Touchscreen?
   {
      BEETLE_OPT(gun_cursor),
      BEETLE_OPT_GUN_CURSOR_LABEL_JA,
      NULL,
      BEETLE_OPT_GUN_CURSOR_INFO_0_JA,
      NULL,
      "input",
      {
         { "off",   "disabled" },
         { "cross", OPTION_VAL_CROSS_JA },
         { "dot",   OPTION_VAL_DOT_JA },
         { NULL, NULL },
      },
      "cross"
   },
   {
      BEETLE_OPT(crosshair_color_p1),
      BEETLE_OPT_CROSSHAIR_COLOR_P1_LABEL_JA,
      NULL,
      BEETLE_OPT_CROSSHAIR_COLOR_P1_INFO_0_JA,
      NULL,
      "input",
      {
         { "red",    OPTION_VAL_RED_JA },
         { "blue",   OPTION_VAL_BLUE_JA },
         { "green",  OPTION_VAL_GREEN_JA },
         { "orange", OPTION_VAL_ORANGE_JA },
         { "yellow", OPTION_VAL_YELLOW_JA },
         { "cyan",   OPTION_VAL_CYAN_JA },
         { "pink",   OPTION_VAL_PINK_JA },
         { "purple", OPTION_VAL_PURPLE_JA },
         { "black",  OPTION_VAL_BLACK_JA },
         { "white",  OPTION_VAL_WHITE_JA },
         { NULL, NULL },
      },
      "red"
   },
   {
      BEETLE_OPT(crosshair_color_p2),
      BEETLE_OPT_CROSSHAIR_COLOR_P2_LABEL_JA,
      NULL,
      BEETLE_OPT_CROSSHAIR_COLOR_P2_INFO_0_JA,
      NULL,
      "input",
      {
         { "blue",   OPTION_VAL_BLUE_JA },
         { "red",    OPTION_VAL_RED_JA },
         { "green",  OPTION_VAL_GREEN_JA },
         { "orange", OPTION_VAL_ORANGE_JA },
         { "yellow", OPTION_VAL_YELLOW_JA },
         { "cyan",   OPTION_VAL_CYAN_JA },
         { "pink",   OPTION_VAL_PINK_JA },
         { "purple", OPTION_VAL_PURPLE_JA },
         { "black",  OPTION_VAL_BLACK_JA },
         { "white",  OPTION_VAL_WHITE_JA },
         { NULL, NULL },
      },
      "blue"
   },
   {
      BEETLE_OPT(mouse_sensitivity),
      BEETLE_OPT_MOUSE_SENSITIVITY_LABEL_JA,
      NULL,
      BEETLE_OPT_MOUSE_SENSITIVITY_INFO_0_JA,
      NULL,
      "input",
      {
         { "5%",   OPTION_VAL_5_O42_JA },
         { "10%",  OPTION_VAL_10_JA },
         { "15%",  OPTION_VAL_15_JA },
         { "20%",  OPTION_VAL_20_JA },
         { "25%",  OPTION_VAL_25_JA },
         { "30%",  OPTION_VAL_30_JA },
         { "35%",  OPTION_VAL_35_JA },
         { "40%",  OPTION_VAL_40_JA },
         { "45%",  OPTION_VAL_45_JA },
         { "50%",  OPTION_VAL_50_JA },
         { "55%",  OPTION_VAL_55_JA },
         { "60%",  OPTION_VAL_60_JA },
         { "65%",  OPTION_VAL_65_JA },
         { "70%",  OPTION_VAL_70_JA },
         { "75%",  OPTION_VAL_75_JA },
         { "80%",  OPTION_VAL_80_JA },
         { "85%",  OPTION_VAL_85_JA },
         { "90%",  OPTION_VAL_90_JA },
         { "95%",  OPTION_VAL_95_JA },
         { "100%", OPTION_VAL_100_O42_JA },
         { "105%", OPTION_VAL_105_JA },
         { "110%", OPTION_VAL_110_JA },
         { "115%", OPTION_VAL_115_JA },
         { "120%", OPTION_VAL_120_JA },
         { "125%", OPTION_VAL_125_JA },
         { "130%", OPTION_VAL_130_JA },
         { "135%", OPTION_VAL_135_JA },
         { "140%", OPTION_VAL_140_JA },
         { "145%", OPTION_VAL_145_JA },
         { "150%", OPTION_VAL_150_JA },
         { "155%", OPTION_VAL_155_JA },
         { "160%", OPTION_VAL_160_JA },
         { "165%", OPTION_VAL_165_JA },
         { "170%", OPTION_VAL_170_JA },
         { "175%", OPTION_VAL_175_JA },
         { "180%", OPTION_VAL_180_JA },
         { "185%", OPTION_VAL_185_JA },
         { "190%", OPTION_VAL_190_JA },
         { "195%", OPTION_VAL_195_JA },
         { "200%", OPTION_VAL_200_JA },
         { NULL, NULL },
      },
      "100%"
   },
   {
      BEETLE_OPT(negcon_response),
      BEETLE_OPT_NEGCON_RESPONSE_LABEL_JA,
      NULL,
      BEETLE_OPT_NEGCON_RESPONSE_INFO_0_JA,
      NULL,
      "input",
      {
         { "linear",    OPTION_VAL_LINEAR_JA },
         { "quadratic", OPTION_VAL_QUADRATIC_JA },
         { "cubic",     OPTION_VAL_CUBIC_JA },
         { NULL, NULL },
      },
      "linear"
   },
   {
      BEETLE_OPT(negcon_deadzone),
      BEETLE_OPT_NEGCON_DEADZONE_LABEL_JA,
      NULL,
      BEETLE_OPT_NEGCON_DEADZONE_INFO_0_JA,
      NULL,
      "input",
      {
         { "0%",  OPTION_VAL_0_O44_JA },
         { "5%",  OPTION_VAL_5_O42_JA },
         { "10%", OPTION_VAL_10_JA },
         { "15%", OPTION_VAL_15_JA },
         { "20%", OPTION_VAL_20_JA },
         { "25%", OPTION_VAL_25_JA },
         { "30%", OPTION_VAL_30_JA },
         { NULL, NULL },
      },
      "0%"
   },
   {
      BEETLE_OPT(use_mednafen_memcard0_method),
      BEETLE_OPT_USE_MEDNAFEN_MEMCARD0_METHOD_LABEL_JA,
      NULL,
      BEETLE_OPT_USE_MEDNAFEN_MEMCARD0_METHOD_INFO_0_JA,
      NULL,
      "memcards",
      {
         { "libretro", OPTION_VAL_LIBRETRO_JA },
         { "mednafen", OPTION_VAL_MEDNAFEN_JA },
         { NULL, NULL },
      },
      "libretro"
   },
   {
      BEETLE_OPT(enable_memcard1),
      BEETLE_OPT_ENABLE_MEMCARD1_LABEL_JA,
      NULL,
      BEETLE_OPT_ENABLE_MEMCARD1_INFO_0_JA,
      NULL,
      "memcards",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(shared_memory_cards),
      BEETLE_OPT_SHARED_MEMORY_CARDS_LABEL_JA,
      NULL,
      BEETLE_OPT_SHARED_MEMORY_CARDS_INFO_0_JA,
      NULL,
      "memcards",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(memcard_left_index),
      BEETLE_OPT_MEMCARD_LEFT_INDEX_LABEL_JA,
      NULL,
      BEETLE_OPT_MEMCARD_LEFT_INDEX_INFO_0_JA,
      NULL,
      "memcards",
      {
         { "0",  OPTION_VAL_0_O48_JA },
         { "1",  NULL },
         { "2",  NULL },
         { "3",  NULL },
         { "4",  NULL },
         { "5",  NULL },
         { "6",  NULL },
         { "7",  NULL },
         { "8",  NULL },
         { "9",  NULL },
         { "10", NULL },
         { "11", NULL },
         { "12", NULL },
         { "13", NULL },
         { "14", NULL },
         { "15", NULL },
         { "16", NULL },
         { "17", NULL },
         { "18", NULL },
         { "19", NULL },
         { "20", NULL },
         { "21", NULL },
         { "22", NULL },
         { "23", NULL },
         { "24", NULL },
         { "25", NULL },
         { "26", NULL },
         { "27", NULL },
         { "28", NULL },
         { "29", NULL },
         { "30", NULL },
         { "31", NULL },
         { "32", NULL },
         { "33", NULL },
         { "34", NULL },
         { "35", NULL },
         { "36", NULL },
         { "37", NULL },
         { "38", NULL },
         { "39", NULL },
         { "40", NULL },
         { "41", NULL },
         { "42", NULL },
         { "43", NULL },
         { "44", NULL },
         { "45", NULL },
         { "46", NULL },
         { "47", NULL },
         { "48", NULL },
         { "49", NULL },
         { "50", NULL },
         { "51", NULL },
         { "52", NULL },
         { "53", NULL },
         { "54", NULL },
         { "55", NULL },
         { "56", NULL },
         { "57", NULL },
         { "58", NULL },
         { "59", NULL },
         { "60", NULL },
         { "61", NULL },
         { "62", NULL },
         { "63", NULL },
         { NULL, NULL },
      },
      "0"
   },
   {
      BEETLE_OPT(memcard_right_index),
      BEETLE_OPT_MEMCARD_RIGHT_INDEX_LABEL_JA,
      NULL,
      BEETLE_OPT_MEMCARD_RIGHT_INDEX_INFO_0_JA,
      NULL,
      "memcards",
      {
         { "0",  NULL },
         { "1",  OPTION_VAL_1_O49_JA },
         { "2",  NULL },
         { "3",  NULL },
         { "4",  NULL },
         { "5",  NULL },
         { "6",  NULL },
         { "7",  NULL },
         { "8",  NULL },
         { "9",  NULL },
         { "10", NULL },
         { "11", NULL },
         { "12", NULL },
         { "13", NULL },
         { "14", NULL },
         { "15", NULL },
         { "16", NULL },
         { "17", NULL },
         { "18", NULL },
         { "19", NULL },
         { "20", NULL },
         { "21", NULL },
         { "22", NULL },
         { "23", NULL },
         { "24", NULL },
         { "25", NULL },
         { "26", NULL },
         { "27", NULL },
         { "28", NULL },
         { "29", NULL },
         { "30", NULL },
         { "31", NULL },
         { "32", NULL },
         { "33", NULL },
         { "34", NULL },
         { "35", NULL },
         { "36", NULL },
         { "37", NULL },
         { "38", NULL },
         { "39", NULL },
         { "40", NULL },
         { "41", NULL },
         { "42", NULL },
         { "43", NULL },
         { "44", NULL },
         { "45", NULL },
         { "46", NULL },
         { "47", NULL },
         { "48", NULL },
         { "49", NULL },
         { "50", NULL },
         { "51", NULL },
         { "52", NULL },
         { "53", NULL },
         { "54", NULL },
         { "55", NULL },
         { "56", NULL },
         { "57", NULL },
         { "58", NULL },
         { "59", NULL },
         { "60", NULL },
         { "61", NULL },
         { "62", NULL },
         { "63", NULL },
         { NULL, NULL },
      },
      "1"
   },
   {
      BEETLE_OPT(line_render),
      BEETLE_OPT_LINE_RENDER_LABEL_JA,
      NULL,
      BEETLE_OPT_LINE_RENDER_INFO_0_JA,
      NULL,
      "hacks",
      {
         { "default",    OPTION_VAL_DEFAULT_JA },
         { "aggressive", OPTION_VAL_AGGRESSIVE_JA },
         { "disabled",   NULL },
         { NULL, NULL },
      },
      "default"
   },
   {
      BEETLE_OPT(widescreen_hack),
      BEETLE_OPT_WIDESCREEN_HACK_LABEL_JA,
      NULL,
      BEETLE_OPT_WIDESCREEN_HACK_INFO_0_JA,
      NULL,
      "hacks",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(widescreen_hack_aspect_ratio),
      BEETLE_OPT_WIDESCREEN_HACK_ASPECT_RATIO_LABEL_JA,
      NULL,
      BEETLE_OPT_WIDESCREEN_HACK_ASPECT_RATIO_INFO_0_JA,
      NULL,
      "hacks",
      {
         { "16:9",  OPTION_VAL_16_9_JA },
         { "16:10", OPTION_VAL_16_10_JA },
         { "18:9",  OPTION_VAL_18_9_JA },
         { "19:9",  OPTION_VAL_19_9_JA },
         { "20:9",  OPTION_VAL_20_9_JA },
         { "21:9",  OPTION_VAL_21_9_JA }, // 64:27
         { "32:9",  OPTION_VAL_32_9_JA },
         { NULL,    NULL },
      },
      "16:9"
   },
#if defined(HAVE_LIGHTREC)
   {
      BEETLE_OPT(cpu_dynarec),
      BEETLE_OPT_CPU_DYNAREC_LABEL_JA,
      NULL,
      BEETLE_OPT_CPU_DYNAREC_INFO_0_JA,
      NULL,
      "hacks",
      {
         { "disabled",        OPTION_VAL_DISABLED_JA },
         { "execute",         OPTION_VAL_EXECUTE_JA },
         { "execute_one",     OPTION_VAL_EXECUTE_ONE_JA },
         { "run_interpreter", OPTION_VAL_RUN_INTERPRETER_JA },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(dynarec_invalidate),
      BEETLE_OPT_DYNAREC_INVALIDATE_LABEL_JA,
      NULL,
      BEETLE_OPT_DYNAREC_INVALIDATE_INFO_0_JA,
      NULL,
      "hacks",
      {
         { "full", OPTION_VAL_FULL_JA },
         { "dma",  OPTION_VAL_DMA_JA },
         { NULL, NULL },
      },
      "full"
   },
   {
      BEETLE_OPT(dynarec_eventcycles),
      BEETLE_OPT_DYNAREC_EVENTCYCLES_LABEL_JA,
      NULL,
      BEETLE_OPT_DYNAREC_EVENTCYCLES_INFO_0_JA,
      NULL,
      "hacks",
      {
         { "128",  OPTION_VAL_128_JA },
         { "256",  NULL },
         { "384",  NULL },
         { "512",  NULL },
         { "640",  NULL },
         { "768",  NULL },
         { "896",  NULL },
         { "1024", NULL },
         { "1152", NULL },
         { "1280", NULL },
         { "1408", NULL },
         { "1536", NULL },
         { "1664", NULL },
         { "1792", NULL },
         { "1920", NULL },
         { "2048", NULL },
         { NULL, NULL },
      },
      "128"
   },
   {
      BEETLE_OPT(dynarec_spu_samples),
      BEETLE_OPT_DYNAREC_SPU_SAMPLES_LABEL_JA,
      NULL,
      BEETLE_OPT_DYNAREC_SPU_SAMPLES_INFO_0_JA,
      NULL,
      "hacks",
      {
         { "1",  OPTION_VAL_1_O49_JA },
         { "4",  NULL },
         { "16", NULL },
         { NULL, NULL },
      },
      "1"
   },
#endif
   {
      BEETLE_OPT(core_timing_fps),
      BEETLE_OPT_CORE_TIMING_FPS_LABEL_JA,
      NULL,
      BEETLE_OPT_CORE_TIMING_FPS_INFO_0_JA,
      NULL,
      "video",
      {
         { "auto_toggle",       OPTION_VAL_AUTO_TOGGLE_JA },
         { "force_progressive", OPTION_VAL_FORCE_PROGRESSIVE_JA },
         { "force_interlaced",  OPTION_VAL_FORCE_INTERLACED_JA },
      },
      "auto_toggle"
   },
   {
      BEETLE_OPT(deinterlacer),
      BEETLE_OPT_DEINTERLACER_LABEL_JA,
      NULL,
      BEETLE_OPT_DEINTERLACER_INFO_0_JA,
      NULL,
      "video",
      {
         { "weave",      OPTION_VAL_WEAVE_JA },
         { "bob",        OPTION_VAL_BOB_JA },
         { "bob_offset", OPTION_VAL_BOB_OFFSET_JA },
         { "fastmad",    OPTION_VAL_FASTMAD_JA },
         { "off",        "Off" },
         { NULL, NULL },
      },
      "weave"
   },
   {
      BEETLE_OPT(frame_duping),
      BEETLE_OPT_FRAME_DUPING_LABEL_JA,
      NULL,
      BEETLE_OPT_FRAME_DUPING_INFO_0_JA,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(pal_video_timing_override),
      BEETLE_OPT_PAL_VIDEO_TIMING_OVERRIDE_LABEL_JA,
      NULL,
      BEETLE_OPT_PAL_VIDEO_TIMING_OVERRIDE_INFO_0_JA,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(aspect_ratio),
      BEETLE_OPT_ASPECT_RATIO_LABEL_JA,
      NULL,
      BEETLE_OPT_ASPECT_RATIO_INFO_0_JA,
      NULL,
      "video",
      {
         { "corrected", OPTION_VAL_CORRECTED_JA },
         { "uncorrected", OPTION_VAL_UNCORRECTED_JA },
         { "4:3",  OPTION_VAL_4_3_JA },
         { "ntsc", OPTION_VAL_NTSC_JA },
      },
      "corrected"
   },
   {
      BEETLE_OPT(crop_overscan),
      BEETLE_OPT_CROP_OVERSCAN_LABEL_JA,
      NULL,
      BEETLE_OPT_CROP_OVERSCAN_INFO_0_JA,
      NULL,
      "video",
      {
         { "disabled", "Disabled" },
         { "static",   OPTION_VAL_STATIC_JA },
         { "smart",    OPTION_VAL_SMART_JA },
         { NULL, NULL },
      },
      "smart"
   },
   {
      BEETLE_OPT(image_crop),
      BEETLE_OPT_IMAGE_CROP_LABEL_JA,
      NULL,
      BEETLE_OPT_IMAGE_CROP_INFO_0_JA,
      NULL,
      "video",
      {
         { "disabled", "0" },
         { "1px",      OPTION_VAL_1PX_JA },
         { "2px",      OPTION_VAL_2PX_JA },
         { "3px",      OPTION_VAL_3PX_JA },
         { "4px",      OPTION_VAL_4PX_JA },
         { "5px",      OPTION_VAL_5PX_JA },
         { "6px",      OPTION_VAL_6PX_JA },
         { "7px",      OPTION_VAL_7PX_JA },
         { "8px",      OPTION_VAL_8PX_JA },
         { "9px",      OPTION_VAL_9PX_JA },
         { "10px",     OPTION_VAL_10PX_JA },
         { "11px",     OPTION_VAL_11PX_JA },
         { "12px",     OPTION_VAL_12PX_JA },
         { "13px",     OPTION_VAL_13PX_JA },
         { "14px",     OPTION_VAL_14PX_JA },
         { "15px",     OPTION_VAL_15PX_JA },
         { "16px",     OPTION_VAL_16PX_JA },
         { "17px",     OPTION_VAL_17PX_JA },
         { "18px",     OPTION_VAL_18PX_JA },
         { "19px",     OPTION_VAL_19PX_JA },
         { "20px",     OPTION_VAL_20PX_JA },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(image_offset),
      BEETLE_OPT_IMAGE_OFFSET_LABEL_JA,
      NULL,
      BEETLE_OPT_IMAGE_OFFSET_INFO_0_JA,
      NULL,
      "video",
      {
         { "-12px",    OPTION_VAL_12PX_O64_JA },
         { "-11px",    OPTION_VAL_11PX_O64_JA },
         { "-10px",    OPTION_VAL_10PX_O64_JA },
         { "-9px",     OPTION_VAL_9PX_O64_JA },
         { "-8px",     OPTION_VAL_8PX_O64_JA },
         { "-7px",     OPTION_VAL_7PX_O64_JA },
         { "-6px",     OPTION_VAL_6PX_O64_JA },
         { "-5px",     OPTION_VAL_5PX_O64_JA },
         { "-4px",     OPTION_VAL_4PX_O64_JA },
         { "-3px",     OPTION_VAL_3PX_O64_JA },
         { "-2px",     OPTION_VAL_2PX_O64_JA },
         { "-1px",     OPTION_VAL_1PX_O64_JA },
         { "disabled", "0" },
         { "+1px",     OPTION_VAL_1PX_O64_0_JA },
         { "+2px",     OPTION_VAL_2PX_O64_0_JA },
         { "+3px",     OPTION_VAL_3PX_O64_0_JA },
         { "+4px",     OPTION_VAL_4PX_O64_0_JA },
         { "+5px",     OPTION_VAL_5PX_O64_0_JA },
         { "+6px",     OPTION_VAL_6PX_O64_0_JA },
         { "+7px",     OPTION_VAL_7PX_O64_0_JA },
         { "+8px",     OPTION_VAL_8PX_O64_0_JA },
         { "+9px",     OPTION_VAL_9PX_O64_0_JA },
         { "+10px",    OPTION_VAL_10PX_O64_0_JA },
         { "+11px",    OPTION_VAL_11PX_O64_0_JA },
         { "+12px",    OPTION_VAL_12PX_O64_0_JA },
         { NULL, NULL },
      },
      "disabled"
   },
#if defined(HAVE_OPENGL) || defined(HAVE_OPENGLES) || defined(HAVE_VULKAN)
   {
      BEETLE_OPT(image_offset_cycles),
      BEETLE_OPT_IMAGE_OFFSET_CYCLES_LABEL_JA,
      NULL,
      BEETLE_OPT_IMAGE_OFFSET_CYCLES_INFO_0_JA,
      NULL,
      "video",
      {
         { "-40",      NULL },
         { "-39",      NULL },
         { "-38",      NULL },
         { "-37",      NULL },
         { "-36",      NULL },
         { "-35",      NULL },
         { "-34",      NULL },
         { "-33",      NULL },
         { "-32",      NULL },
         { "-31",      NULL },
         { "-30",      NULL },
         { "-29",      NULL },
         { "-28",      NULL },
         { "-27",      NULL },
         { "-26",      NULL },
         { "-25",      NULL },
         { "-24",      NULL },
         { "-23",      NULL },
         { "-22",      NULL },
         { "-21",      NULL },
         { "-20",      NULL },
         { "-19",      NULL },
         { "-18",      NULL },
         { "-17",      NULL },
         { "-16",      NULL },
         { "-15",      NULL },
         { "-14",      NULL },
         { "-13",      NULL },
         { "-12",      NULL },
         { "-11",      NULL },
         { "-10",      NULL },
         { "-9",       NULL },
         { "-8",       NULL },
         { "-7",       NULL },
         { "-6",       NULL },
         { "-5",       NULL },
         { "-4",       NULL },
         { "-3",       NULL },
         { "-2",       NULL },
         { "-1",       NULL },
         { "0",        "0" },
         { "+1",       NULL },
         { "+2",       NULL },
         { "+3",       NULL },
         { "+4",       NULL },
         { "+5",       NULL },
         { "+6",       NULL },
         { "+7",       NULL },
         { "+8",       NULL },
         { "+9",       NULL },
         { "+10",      NULL },
         { "+11",      NULL },
         { "+12",      NULL },
         { "+13",      NULL },
         { "+14",      NULL },
         { "+15",      NULL },
         { "+16",      NULL },
         { "+17",      NULL },
         { "+18",      NULL },
         { "+19",      NULL },
         { "+20",      NULL },
         { "+21",      NULL },
         { "+22",      NULL },
         { "+23",      NULL },
         { "+24",      NULL },
         { "+25",      NULL },
         { "+26",      NULL },
         { "+27",      NULL },
         { "+28",      NULL },
         { "+29",      NULL },
         { "+30",      NULL },
         { "+31",      NULL },
         { "+32",      NULL },
         { "+33",      NULL },
         { "+34",      NULL },
         { "+35",      NULL },
         { "+36",      NULL },
         { "+37",      NULL },
         { "+38",      NULL },
         { "+39",      NULL },
         { "+40",      NULL },
         { NULL, NULL},
      },
      "0"
   },
#endif
   {
      BEETLE_OPT(initial_scanline),
      BEETLE_OPT_INITIAL_SCANLINE_LABEL_JA,
      NULL,
      BEETLE_OPT_INITIAL_SCANLINE_INFO_0_JA,
      NULL,
      "video",
      {
         { "0",  OPTION_VAL_0_O48_JA },
         { "1",  NULL },
         { "2",  NULL },
         { "3",  NULL },
         { "4",  NULL },
         { "5",  NULL },
         { "6",  NULL },
         { "7",  NULL },
         { "8",  NULL },
         { "9",  NULL },
         { "10", NULL },
         { "11", NULL },
         { "12", NULL },
         { "13", NULL },
         { "14", NULL },
         { "15", NULL },
         { "16", NULL },
         { "17", NULL },
         { "18", NULL },
         { "19", NULL },
         { "20", NULL },
         { "21", NULL },
         { "22", NULL },
         { "23", NULL },
         { "24", NULL },
         { "25", NULL },
         { "26", NULL },
         { "27", NULL },
         { "28", NULL },
         { "29", NULL },
         { "30", NULL },
         { "31", NULL },
         { "32", NULL },
         { "33", NULL },
         { "34", NULL },
         { "35", NULL },
         { "36", NULL },
         { "37", NULL },
         { "38", NULL },
         { "39", NULL },
         { "40", NULL },
         { NULL, NULL },
      },
      "0"
   },
   {
      BEETLE_OPT(last_scanline),
      BEETLE_OPT_LAST_SCANLINE_LABEL_JA,
      NULL,
      BEETLE_OPT_LAST_SCANLINE_INFO_0_JA,
      NULL,
      "video",
      {
         { "210", NULL },
         { "211", NULL },
         { "212", NULL },
         { "213", NULL },
         { "214", NULL },
         { "215", NULL },
         { "216", NULL },
         { "217", NULL },
         { "218", NULL },
         { "219", NULL },
         { "220", NULL },
         { "221", NULL },
         { "222", NULL },
         { "223", NULL },
         { "224", NULL },
         { "225", NULL },
         { "226", NULL },
         { "227", NULL },
         { "228", NULL },
         { "229", NULL },
         { "230", NULL },
         { "231", NULL },
         { "232", NULL },
         { "233", NULL },
         { "234", NULL },
         { "235", NULL },
         { "236", NULL },
         { "237", NULL },
         { "238", NULL },
         { "239", OPTION_VAL_239_JA },
         { NULL, NULL },
      },
      "239"
   },
   {
      BEETLE_OPT(initial_scanline_pal),
      BEETLE_OPT_INITIAL_SCANLINE_PAL_LABEL_JA,
      NULL,
      BEETLE_OPT_INITIAL_SCANLINE_PAL_INFO_0_JA,
      NULL,
      "video",
      {
         { "0",  OPTION_VAL_0_O48_JA },
         { "1",  NULL },
         { "2",  NULL },
         { "3",  NULL },
         { "4",  NULL },
         { "5",  NULL },
         { "6",  NULL },
         { "7",  NULL },
         { "8",  NULL },
         { "9",  NULL },
         { "10", NULL },
         { "11", NULL },
         { "12", NULL },
         { "13", NULL },
         { "14", NULL },
         { "15", NULL },
         { "16", NULL },
         { "17", NULL },
         { "18", NULL },
         { "19", NULL },
         { "20", NULL },
         { "21", NULL },
         { "22", NULL },
         { "23", NULL },
         { "24", NULL },
         { "25", NULL },
         { "26", NULL },
         { "27", NULL },
         { "28", NULL },
         { "29", NULL },
         { "30", NULL },
         { "31", NULL },
         { "32", NULL },
         { "33", NULL },
         { "34", NULL },
         { "35", NULL },
         { "36", NULL },
         { "37", NULL },
         { "38", NULL },
         { "39", NULL },
         { "40", NULL },
         { NULL, NULL },
      },
      "0"
   },
   {
      BEETLE_OPT(last_scanline_pal),
      BEETLE_OPT_LAST_SCANLINE_PAL_LABEL_JA,
      NULL,
      BEETLE_OPT_LAST_SCANLINE_PAL_INFO_0_JA,
      NULL,
      "video",
      {
         { "230", NULL },
         { "231", NULL },
         { "232", NULL },
         { "233", NULL },
         { "234", NULL },
         { "235", NULL },
         { "236", NULL },
         { "237", NULL },
         { "238", NULL },
         { "239", NULL },
         { "240", NULL },
         { "241", NULL },
         { "242", NULL },
         { "243", NULL },
         { "244", NULL },
         { "245", NULL },
         { "246", NULL },
         { "247", NULL },
         { "248", NULL },
         { "249", NULL },
         { "250", NULL },
         { "251", NULL },
         { "252", NULL },
         { "253", NULL },
         { "254", NULL },
         { "255", NULL },
         { "256", NULL },
         { "257", NULL },
         { "258", NULL },
         { "259", NULL },
         { "260", NULL },
         { "261", NULL },
         { "262", NULL },
         { "263", NULL },
         { "264", NULL },
         { "265", NULL },
         { "266", NULL },
         { "267", NULL },
         { "268", NULL },
         { "269", NULL },
         { "270", NULL },
         { "271", NULL },
         { "272", NULL },
         { "273", NULL },
         { "274", NULL },
         { "275", NULL },
         { "276", NULL },
         { "277", NULL },
         { "278", NULL },
         { "279", NULL },
         { "280", NULL },
         { "281", NULL },
         { "282", NULL },
         { "283", NULL },
         { "284", NULL },
         { "285", NULL },
         { "286", NULL },
         { "287", OPTION_VAL_287_JA },
         { NULL, NULL },
      },
      "287"
   },
   { NULL, NULL, NULL, NULL, NULL, NULL, {{0}}, NULL },
};
struct retro_core_options_v2 options_ja = {
   option_cats_ja,
   option_defs_ja
};

/* RETRO_LANGUAGE_KO */

#define CATEGORY_SYSTEM_LABEL_KO "시스템"
#define CATEGORY_SYSTEM_INFO_0_KO "시스템 및 오버클럭 옵션을 변경합니다."
#define CATEGORY_VIDEO_LABEL_KO "비디오"
#define CATEGORY_VIDEO_INFO_0_KO "화면비, 잘라내기, 비디오 필터, 프레임 스킵 등의 설정을 변경합니다."
#define CATEGORY_PGXP_LABEL_KO NULL
#define CATEGORY_PGXP_INFO_0_KO "그래픽을 원본 하드웨어보다 더 좋게 개선하는 설정을 조정합니다. PGXP는 텍스쳐 뒤틀림과 Z-파이팅 문제를 해결할 수 있습니다."
#define CATEGORY_OSD_LABEL_KO "온스크린 디스플레이"
#define CATEGORY_OSD_INFO_0_KO "화면에 표시되는 OSD 알림을 설정합니다."
#define CATEGORY_INPUT_LABEL_KO "입력"
#define CATEGORY_INPUT_INFO_0_KO "라이트 건, 마우스, neGcon 등을 설정합니다."
#define CATEGORY_MEMCARDS_LABEL_KO "메모리 카드"
#define CATEGORY_MEMCARDS_INFO_0_KO "가상 메모리 카드 관련 설정을 변경합니다."
#define CATEGORY_HACKS_LABEL_KO "에뮬레이션 핵"
#define CATEGORY_HACKS_INFO_0_KO "성능과 호환성에 영향을 끼치는 에뮬레이션 정확도 설정을 변경합니다."
#define BEETLE_OPT_CPU_FREQ_SCALE_LABEL_KO "CPU 속도 스케일링 (오버클럭)"
#define BEETLE_OPT_CPU_FREQ_SCALE_INFO_0_KO "에뮬레이트되는 PSX CPU를 오버클럭(언더클럭)합니다. 오버클러킹은 높은 사양을 요구하는 대신 일부 게임에서 느려짐 증상을 없애고 프레임레이트를 개선할 수 있습니다. 일부 게임은 프레임레이트 리미터가 내장되어 있으므로 오버클러킹의 이득이 없을 수도 있습니다. 일부 타이틀에서 오버클럭을 할 경우 일부 효과가 의도된 것보다 빠르게 재생될 수 있습니다."
#define OPTION_VAL_50_KO NULL
#define OPTION_VAL_60_KO NULL
#define OPTION_VAL_70_KO NULL
#define OPTION_VAL_80_KO NULL
#define OPTION_VAL_90_KO NULL
#define OPTION_VAL_100_KO "100% (네이티브)"
#define OPTION_VAL_110_KO NULL
#define OPTION_VAL_120_KO NULL
#define OPTION_VAL_130_KO NULL
#define OPTION_VAL_140_KO NULL
#define OPTION_VAL_150_KO NULL
#define OPTION_VAL_160_KO NULL
#define OPTION_VAL_170_KO NULL
#define OPTION_VAL_180_KO NULL
#define OPTION_VAL_190_KO NULL
#define OPTION_VAL_200_KO NULL
#define OPTION_VAL_210_KO NULL
#define OPTION_VAL_220_KO NULL
#define OPTION_VAL_230_KO NULL
#define OPTION_VAL_240_KO NULL
#define OPTION_VAL_250_KO NULL
#define OPTION_VAL_260_KO NULL
#define OPTION_VAL_270_KO NULL
#define OPTION_VAL_280_KO NULL
#define OPTION_VAL_290_KO NULL
#define OPTION_VAL_300_KO NULL
#define OPTION_VAL_310_KO NULL
#define OPTION_VAL_320_KO NULL
#define OPTION_VAL_330_KO NULL
#define OPTION_VAL_340_KO NULL
#define OPTION_VAL_350_KO NULL
#define OPTION_VAL_360_KO NULL
#define OPTION_VAL_370_KO NULL
#define OPTION_VAL_380_KO NULL
#define OPTION_VAL_390_KO NULL
#define OPTION_VAL_400_KO NULL
#define OPTION_VAL_410_KO NULL
#define OPTION_VAL_420_KO NULL
#define OPTION_VAL_430_KO NULL
#define OPTION_VAL_440_KO NULL
#define OPTION_VAL_450_KO NULL
#define OPTION_VAL_460_KO NULL
#define OPTION_VAL_470_KO NULL
#define OPTION_VAL_480_KO NULL
#define OPTION_VAL_490_KO NULL
#define OPTION_VAL_500_KO NULL
#define OPTION_VAL_510_KO NULL
#define OPTION_VAL_520_KO NULL
#define OPTION_VAL_530_KO NULL
#define OPTION_VAL_540_KO NULL
#define OPTION_VAL_550_KO NULL
#define OPTION_VAL_560_KO NULL
#define OPTION_VAL_570_KO NULL
#define OPTION_VAL_580_KO NULL
#define OPTION_VAL_590_KO NULL
#define OPTION_VAL_600_KO NULL
#define OPTION_VAL_610_KO NULL
#define OPTION_VAL_620_KO NULL
#define OPTION_VAL_630_KO NULL
#define OPTION_VAL_640_KO NULL
#define OPTION_VAL_650_KO NULL
#define OPTION_VAL_660_KO NULL
#define OPTION_VAL_670_KO NULL
#define OPTION_VAL_680_KO NULL
#define OPTION_VAL_690_KO NULL
#define OPTION_VAL_700_KO NULL
#define OPTION_VAL_710_KO NULL
#define OPTION_VAL_720_KO NULL
#define OPTION_VAL_730_KO NULL
#define OPTION_VAL_740_KO NULL
#define OPTION_VAL_750_KO NULL
#define BEETLE_OPT_GPU_OVERCLOCK_LABEL_KO "GPU 래스터라이저 오버클럭"
#define BEETLE_OPT_GPU_OVERCLOCK_INFO_0_KO "에뮬레이트되는 PSX GPU의 2D 래스터라이저를 오버클럭합니다. 3D 렌더링은 개선되지 않으며, 전반적으로 큰 효과는 없습니다."
#define OPTION_VAL_1X_NATIVE_KO "1x (네이티브)"
#define OPTION_VAL_2X_KO NULL
#define OPTION_VAL_4X_KO NULL
#define OPTION_VAL_8X_KO NULL
#define OPTION_VAL_16X_KO NULL
#define OPTION_VAL_32X_KO "32배"
#define BEETLE_OPT_GTE_OVERCLOCK_LABEL_KO "GTE 오버클럭"
#define BEETLE_OPT_GTE_OVERCLOCK_INFO_0_KO "에뮬레이트되는 GTE(3D 그래픽 보조 처리장치) 연산을 모두 1사이클 시간 내에 처리합니다. GTE를 많이 사용하는 게임에서 프레임레이트와 프레임 시간 안정성을 크게 향상시킬 수 있습니다."
#define BEETLE_OPT_CD_ACCESS_METHOD_LABEL_KO "CD 액세스 방식"
#define BEETLE_OPT_CD_ACCESS_METHOD_INFO_0_KO "'동기'는 원본 하드웨어의 동작을 흉내냅니다. '비동기'는 저장소가 느린 장치에서 버벅임을 줄일 수 있습니다. '미리 캐싱'은 컨텐츠를 실행할 때 전체 디스크 이미지를 메모리에 불러옵니다. 첫 시작 시 딜레이가 생기지만 인게임 로딩 시간이 줄어듭니다. '미리 캐싱'은 RAM이 적은 장치에서 문제를 일으킬 수 있으며, 물리적 미디어를 사용할 경우 '동기'가 사용됩니다."
#define OPTION_VAL_SYNC_KO "동기"
#define OPTION_VAL_ASYNC_KO "비동기"
#define OPTION_VAL_PRECACHE_KO "미리 캐싱"
#define BEETLE_OPT_CD_FASTLOAD_LABEL_KO "CD 로딩 속도"
#define BEETLE_OPT_CD_FASTLOAD_INFO_0_KO "'2배 (네이티브)'를 초과하는 값은 인게임 로딩 시간을 크게 줄일 수 있지만, 타이밍 오류를 유발하기도 합니다. 일부 게임은 일정 속도 이상에서 올바르게 작동하지 않을 수 있습니다."
#define OPTION_VAL_2X_NATIVE_KO "2배 (Native)"
#define OPTION_VAL_6X_KO NULL
#define OPTION_VAL_10X_KO "10배"
#define OPTION_VAL_12X_KO "12배"
#define OPTION_VAL_14X_KO "14배"
#define BEETLE_OPT_SKIP_BIOS_LABEL_KO "BIOS 건너뛰기"
#define BEETLE_OPT_SKIP_BIOS_INFO_0_KO "컨텐츠를 불러올 때 표시되는 PlayStation BIOS 부트 애니메이션을 건너뜁니다. 주의: 이 옵션을 사용하면 일부 게임(PAL 복사 방지 게임, Saga Frontier 등)에서 호환성 문제가 발생합니다."
#define BEETLE_OPT_OVERRIDE_BIOS_LABEL_KO "BIOS 오버라이드"
#define BEETLE_OPT_OVERRIDE_BIOS_INFO_0_KO "존재할 경우, 기존의 지역 제한 BIOS를 지역 제한이 없는 BIOS로 대체합니다. 재시작이 필요합니다."
#define OPTION_VAL_PSXONPSP_KO NULL
#define OPTION_VAL_PS1_ROM_KO NULL
#define OPTION_VAL_OPENBIOS_KO NULL
#define BEETLE_OPT_REGION_LABEL_KO "시스템 지역"
#define BEETLE_OPT_REGION_INFO_0_KO NULL
#define OPTION_VAL_AUTO_KO "자동"
#define OPTION_VAL_NTSC_J_KO NULL
#define OPTION_VAL_NTSC_U_KO NULL
#define OPTION_VAL_PAL_KO NULL
#define BEETLE_OPT_INTERNAL_RESOLUTION_LABEL_KO "내부 GPU 해상도"
#define BEETLE_OPT_INTERNAL_RESOLUTION_INFO_0_KO "'1x (네이티브)'보다 높은 값으로 설정할 경우 3D 모델의 품질이 개선되지만 요구되는 성능이 높아집니다. 2D 요소들은 일반적으로 이 설정에 아무런 영향을 받지 않습니다."
#define BEETLE_OPT_RENDERER_LABEL_KO "렌더러"
#define BEETLE_OPT_RENDERER_INFO_0_KO "소프트웨어 렌더러는 가장 정확하지만 내부 GPU 해상도를 높여 사용할 경우 요구 사양이 높아집니다. 하드웨어 렌더러는 조금 더 부정확하지만 높은 내부 GPU 해상도에서 소프트웨어 렌더러 대비 성능이 향상되며 다양한 그래픽 개선을 적용할 수 있습니다. '하드웨어 (자동)'은 현재 libretro 프론트엔드 비디오 드라이버에 따라 Vulkan 또는 OpenGL 중 적절한 렌더러를 선택합니다. 제공된 비디오 드라이버가 Vulkan 또는 OpenGL 3.3과 호환되지 않을 경우, 코어는 소프트웨어 렌더러를 사용할 것입니다. 재시작이 필요합니다."
#define OPTION_VAL_HARDWARE_KO "하드웨어 (자동)"
#define OPTION_VAL_HARDWARE_GL_KO "하드웨어 (OpenGL)"
#define OPTION_VAL_HARDWARE_VK_KO "하드웨어 (Vulkan)"
#define OPTION_VAL_SOFTWARE_KO "소프트웨어"
#define BEETLE_OPT_RENDERER_SOFTWARE_FB_LABEL_KO "소프트웨어 프레임버퍼"
#define BEETLE_OPT_RENDERER_SOFTWARE_FB_INFO_0_KO "하드웨어 렌더러 사용 시 1배 해상도의 소프트웨어 렌더러를 백그라운드에서 함께 구동하여 프레임버퍼 효과(모션 블러, FF7 전투 개시 효과 등)를 정확하게 에뮬레이트합니다. 비활성화할 경우, 이러한 효과는 무시되거나(OpenGL) GPU에서 렌더됩니다(Vulkan). 비활성화 시 성능이 개선되지만 심각한 그래픽 오류를 일으킬 수 있습니다. 불확실할 경우 활성으로 설정하십시오."
#define BEETLE_OPT_DEPTH_LABEL_KO "내부 색 깊이"
#define BEETLE_OPT_DEPTH_INFO_0_KO "높은 색 깊이는 디더링을 사용하지 않을 때 나타나는 색 밴딩 현상을 완화할 수 있습니다. 16 bpp는 원본 하드웨어를 에뮬레이트하며 디더링을 사용하지 않을 경우 눈에 띄는 색 밴딩이 나타날 수 있습니다. 이 옵션을 32 bpp로 설정할 경우 '디더링 패턴'을 비활성화할 것을 권장합니다."
#define OPTION_VAL_16BPP_NATIVE_KO "16 bpp (네이티브)"
#define OPTION_VAL_32BPP_KO NULL
#define BEETLE_OPT_DITHER_MODE_LABEL_KO "디더링 패턴"
#define BEETLE_OPT_DITHER_MODE_INFO_0_KO "'1x (네이티브)'는 원본 하드웨어의 저해상도 디더링을 에뮬레이트하여 네이티브 색 깊이에서 나타나는 색 밴딩 현상을 자연스럽게 숨깁니다. '내부 해상도'는 디더링을 내장 해상도에 맞춰 설정하여 깔끔한 결과물을 제공합니다. 32 bpp 색 깊이를 사용할 때는 비활성화할 것을 권장합니다. 주의: Vulcan에서 이 옵션을 사용하면 색 깊이가 네이티브 수치로 강제 다운샘플링되며 사용하지 않으면 자동으로 높은 색 깊이를 사용합니다."
#define OPTION_VAL_INTERNAL_RESOLUTION_KO "내부 해상도"
#define BEETLE_OPT_SCALED_UV_OFFSET_LABEL_KO "텍스쳐 UV 오프셋"
#define BEETLE_OPT_SCALED_UV_OFFSET_INFO_0_KO "3D 폴리곤을 높은 해상도에서 표시할 때 텍스쳐를 설정한 오프셋에서 샘플링합니다. 텍스쳐 사이에 틈이 나타나는 것을 줄여주지만 의도되지 않은 그래픽 문제를 일으킬 수 있습니다."
#define BEETLE_OPT_FILTER_LABEL_KO "텍스쳐 필터링"
#define BEETLE_OPT_FILTER_INFO_0_KO "'최근접 이웃'은 원본 하드웨어의 행동을 에뮬레이트합니다. '이중선형' 및 '3포인트'는 텍스쳐의 각짐을 줄이고 부드럽게 표시합니다. 'SABR', 'xBR', 'JINC2'는 업스케일링 필터로 텍스쳐의 품질/선명도를 개선할 수도 있지만 요구하는 성능이 더 높아집니다. 하드웨어 렌더러를 사용할 경우에만 적용됩니다."
#define OPTION_VAL_NEAREST_KO "최근접 이웃"
#define OPTION_VAL_SABR_KO NULL
#define OPTION_VAL_XBR_KO NULL
#define OPTION_VAL_BILINEAR_KO "이중선형"
#define OPTION_VAL_3_POINT_KO "3포인트"
#define OPTION_VAL_JINC2_KO NULL
#define BEETLE_OPT_FILTER_EXCLUDE_SPRITE_LABEL_KO "스프라이트 필터링 안 함"
#define BEETLE_OPT_FILTER_EXCLUDE_SPRITE_INFO_0_KO "스프라이트에는 텍스쳐 필터링을 적용하지 않습니다. 2D 스프라이트로 배경을 그리는 다양한 게임에서 틈이 생기는 것을 방지해줍니다. 최선의 효과를 얻으려면 선택적 스무딩 또는 다른 후처리 필터와 함께 사용하십시오."
#define OPTION_VAL_OPAQUE_KO "불투명 스프라이트만"
#define OPTION_VAL_ALL_KO "불투명 및 반투명 스프라이트"
#define BEETLE_OPT_FILTER_EXCLUDE_2D_POLYGON_LABEL_KO "2D 폴리곤 필터링 안 함"
#define BEETLE_OPT_FILTER_EXCLUDE_2D_POLYGON_INFO_0_KO "2D 폴리곤에는 텍스쳐 필터링을 적용하지 않습니다. 2D 폴리곤은 휴리스틱으로 감지되며 오류가 있을 수 있습니다. 최선의 효과를 얻으려면 선택적 스무딩 또는 다른 후처리 필터와 함께 사용하십시오."
#define BEETLE_OPT_ADAPTIVE_SMOOTHING_LABEL_KO "선택적 스무딩"
#define BEETLE_OPT_ADAPTIVE_SMOOTHING_INFO_0_KO "3D로 그려지는 개체를 제외한 2D 그림과 UI만 부드럽게 표시합니다. Vulkan 렌더러에서만 지원됩니다."
#define BEETLE_OPT_SUPER_SAMPLING_LABEL_KO "슈퍼샘플링 (네이티브 해상도로 다운샘플)"
#define BEETLE_OPT_SUPER_SAMPLING_INFO_0_KO "높은 내부 해상도로 렌더된 컨텐츠를 네이티브 해상도로 다운샘플링하여 표시합니다. 이 옵션을 높은 내부 해상도와 함께 사용할 경우 게임을 3D 개체에 안티에일리어싱 효과가 적용된 네이티브 해상도로 표시할 수 있습니다. 2D와 3D 요소를 섞어 사용하는 게임(예: 미리 렌더된 배경 위 3D 캐릭터)에 적용했을 때 최선의 효과를 얻을 수 있으며, CRT 셰이더와 함께 사용하기 좋습니다. Vulkan 렌더러에서만 지원됩니다. 주의: 이 옵션을 사용할 경우 '디더링 패턴'을 비활성화하는 것을 추천합니다."
#define BEETLE_OPT_MSAA_LABEL_KO "멀티 샘플 안티에일리어싱 (MSAA)"
#define BEETLE_OPT_MSAA_INFO_0_KO "MSAA는 3D 개체의 표현을 개선합니다. Vulkan 렌더러에서만 지원됩니다."
#define OPTION_VAL_1X_KO NULL
#define BEETLE_OPT_MDEC_YUV_LABEL_KO "MDEC YUV 크로마 필터"
#define BEETLE_OPT_MDEC_YUV_INFO_0_KO "'매크로블로킹' 아티팩트(사각형/날카로운 외곽선)를 감소시켜 FMV의 품질을 향상시킵니다. Vulkan 렌더러에서만 지원됩니다."
#define BEETLE_OPT_TRACK_TEXTURES_LABEL_KO "텍스쳐 추적"
#define BEETLE_OPT_TRACK_TEXTURES_INFO_0_KO "텍스쳐를 덤프하거나 교체하기 위해 필요합니다. 대부분의 게임에서 오류가 발생할 것입니다."
#define BEETLE_OPT_DUMP_TEXTURES_LABEL_KO "텍스쳐 덤프하기"
#define BEETLE_OPT_DUMP_TEXTURES_INFO_0_KO "사용된 텍스쳐를 <cd>-texture-dump/에 덤프합니다."
#define BEETLE_OPT_REPLACE_TEXTURES_LABEL_KO "텍스쳐 교체하기"
#define BEETLE_OPT_REPLACE_TEXTURES_INFO_0_KO "텍스쳐를 <cd>-texture-replacements/의 HD 버전으로 교체합니다."
#define BEETLE_OPT_PGXP_MODE_LABEL_KO "PGXP 작동 모드"
#define BEETLE_OPT_PGXP_MODE_INFO_0_KO "3D 개체를 서브픽셀 단위 정확도로 렌더링합니다. 이는 원본 하드웨어가 버텍스 좌표에 고정소수점을 사용하여 발생하는 3D 개체의 뒤틀림과 떨림 현상을 최소화합니다. '메모리만'은 호환성 문제가 적으며 일반적인 사용에 적합합니다. '메모리 + CPU (불안정)'은 떨림 현상을 더 많이 줄일 수 있지만 성능 요구사항이 높으며 다양한 그래픽 오류가 발생할 수 있습니다."
#define OPTION_VAL_MEMORY_ONLY_KO "메모리만"
#define OPTION_VAL_MEMORY_CPU_KO "메모리 + CPU (불안정)"
#define BEETLE_OPT_PGXP_2D_TOL_LABEL_KO "PGXP 2D 기하 허용오차"
#define BEETLE_OPT_PGXP_2D_TOL_INFO_0_KO "PGXP 작업에서 더 눈에 띄는 오류를 숨깁니다. 개체에 적절한 깊이 정보가 없는 경우 지정한 만큼의 PGXP 값이 오차로서 허용됩니다."
#define OPTION_VAL_0PX_KO NULL
#define OPTION_VAL_1PX_KO NULL
#define OPTION_VAL_2PX_KO NULL
#define OPTION_VAL_3PX_KO NULL
#define OPTION_VAL_4PX_KO NULL
#define OPTION_VAL_5PX_KO NULL
#define OPTION_VAL_6PX_KO NULL
#define OPTION_VAL_7PX_KO NULL
#define OPTION_VAL_8PX_KO NULL
#define BEETLE_OPT_PGXP_NCLIP_LABEL_KO "PGXP 프리미티브 컬링"
#define BEETLE_OPT_PGXP_NCLIP_INFO_0_KO "PGXP의 NCLIP 구현을 사용합니다. PGXP 좌표의 폴리곤에 나타나는 구멍을 줄여 외형을 개선합니다. 일부 게임이 다양한 상황에서 멈추는 부작용이 있습니다."
#define BEETLE_OPT_PGXP_VERTEX_LABEL_KO "PGXP 버텍스 캐시"
#define BEETLE_OPT_PGXP_VERTEX_INFO_0_KO "PGXP 개선된 버텍스 위치를 캐싱하여 다양한 폴리곤 그리기에 재사용합니다. 텍스쳐를 렌더링할 때 개체 정렬을 개선하고 틈을 줄이는 등의 개선 효과를 볼 수 있지만, 캐시 쿼리 시 오탐으로 인해 그래픽 오류가 발생할 수 있습니다. 현재는 사용하지 않는 것이 권장됩니다. 이 설정은 PGXP 작동 모드가 활성화되어 있어야 적용됩니다. 하드웨어 렌더러에서만 지원됩니다."
#define BEETLE_OPT_PGXP_TEXTURE_LABEL_KO "PGXP 원근 교정 텍스쳐링"
#define BEETLE_OPT_PGXP_TEXTURE_INFO_0_KO NULL
#define BEETLE_OPT_DISPLAY_INTERNAL_FPS_LABEL_KO "내부 FPS 표시"
#define BEETLE_OPT_DISPLAY_INTERNAL_FPS_INFO_0_KO "에뮬레이트된 PlayStation 시스템이 컨텐츠를 렌더링하는 내부 프레임 레이트를 표시합니다. 주의: libretro 프론트엔드에서 OSD 알림을 활성화해야 합니다."
#define BEETLE_OPT_DISPLAY_OSD_LABEL_KO "온스크린 알림으로 메시지 표시"
#define BEETLE_OPT_DISPLAY_OSD_INFO_0_KO "코어가 출력하는 메시지를 온스크린 알림으로 표시합니다."
#define BEETLE_OPT_DISPLAY_VRAM_LABEL_KO "전체 VRAM 표시 (디버그)"
#define BEETLE_OPT_DISPLAY_VRAM_INFO_0_KO "에뮬레이트되는 콘솔의 VRAM을 전부 표시합니다. OpenGL 및 Vulkan 하드웨어 렌더러에서만 지원됩니다. 주의: 이 옵션은 디버깅용으로, 일반적으로는 비활성화되어야 합니다."
#define BEETLE_OPT_ANALOG_CALIBRATION_LABEL_KO "아날로그 자가 교정"
#define BEETLE_OPT_ANALOG_CALIBRATION_INFO_0_KO "입력 장치가 DualShock, 아날로그 컨트롤러, 아날로그 조이스틱, neGcon으로 설정된 경우, 아날로그 입력을 동적으로 교정합니다. 입력되는 최대 값을 실시간으로 확인하여 에뮬레이터에 조정된 아날로그 수치를 전달합니다. 이 설정은 Mega Man Legends 2처럼 현대의 컨트롤러들이 제공하는 값보다 더 큰 값을 기대하는 게임을 플레이할 때 사용되어야 합니다. 최선의 효과를 위해서는 컨텐츠를 불러올 때마다 아날로그 스틱들을 최대한 끝까지 밀고 한 바퀴 돌리는 동작을 수행하여 교정 알고리즘을 조율해야 합니다."
#define BEETLE_OPT_ANALOG_TOGGLE_LABEL_KO "DualShock 아날로그 모드 토글"
#define BEETLE_OPT_ANALOG_TOGGLE_INFO_0_KO "입력 장치가 DualShock일 때, 에뮬레이트된 DualShock이 실제 하드웨어처럼 DIGITAL과 ANALOG 모드를 전환할 수 있게 합니다. 선택한 버튼 조합을 눌러 DIGITAL과 ANALOG 모드를 전환할 수 있습니다."
#define OPTION_VAL_ENABLED_ANALOG_KO "기본-아날로그"
#define BEETLE_OPT_ANALOG_TOGGLE_COMBO_LABEL_KO "DualShock 아날로그 모드 버튼 조합"
#define BEETLE_OPT_ANALOG_TOGGLE_COMBO_INFO_0_KO "에뮬레이트된 DualShock의 DIGITAL과 ANALOG 모드를 전환할 버튼 조합입니다."
#define OPTION_VAL_L1_L2_R1_R2_START_SELECT_KO "L1 + L2 + R1 + R2 + 시작 + 선택"
#define OPTION_VAL_L1_R1_SELECT_KO "L1 + R1 + 선택"
#define OPTION_VAL_L1_R1_START_KO "L1 + R1 + 시작"
#define OPTION_VAL_L1_R1_L3_KO NULL
#define OPTION_VAL_L1_R1_R3_KO NULL
#define OPTION_VAL_L2_R2_SELECT_KO "L2 + R2 + 선택"
#define OPTION_VAL_L2_R2_START_KO "L2 + R2 + 시작"
#define OPTION_VAL_L2_R2_L3_KO NULL
#define OPTION_VAL_L2_R2_R3_KO NULL
#define OPTION_VAL_L3_R3_KO NULL
#define BEETLE_OPT_ANALOG_TOGGLE_HOLD_LABEL_KO "DualShock 아날로그 모드 버튼 조합 홀드 지연"
#define BEETLE_OPT_ANALOG_TOGGLE_HOLD_INFO_0_KO "아날로그 모드 버튼 조합을 누르고 있어야 할 시간입니다."
#define OPTION_VAL_0_KO "0 초 지연"
#define OPTION_VAL_1_KO "1 초 지연"
#define OPTION_VAL_2_KO "2 초 지연"
#define OPTION_VAL_3_KO "3 초 지연"
#define OPTION_VAL_4_KO "4 초 지연"
#define OPTION_VAL_5_KO "5 초 지연"
#define BEETLE_OPT_ENABLE_MULTITAP_PORT1_LABEL_KO "1번 포트에서 멀티탭 사용"
#define BEETLE_OPT_ENABLE_MULTITAP_PORT1_INFO_0_KO "1번 포트에 멀티탭 기능을 활성화합니다."
#define BEETLE_OPT_ENABLE_MULTITAP_PORT2_LABEL_KO "2번 포트에서 멀티탭 사용"
#define BEETLE_OPT_ENABLE_MULTITAP_PORT2_INFO_0_KO "2번 포트에 멀티탭 기능을 활성화합니다."
#define BEETLE_OPT_GUN_INPUT_MODE_LABEL_KO "건 입력 모드"
#define BEETLE_OPT_GUN_INPUT_MODE_INFO_0_KO "입력 장치가 'Guncon/G-Con 45' 또는 'Justifier'로 설정되었을 때, 마우스로 조작되는 '라이트 건' 또는 '터치스크린' 입력을 사용합니다."
#define OPTION_VAL_LIGHTGUN_KO "라이트 건"
#define OPTION_VAL_TOUCHSCREEN_KO "터치스크린"
#define BEETLE_OPT_GUN_CURSOR_LABEL_KO "건 커서"
#define BEETLE_OPT_GUN_CURSOR_INFO_0_KO "'Guncon/G-Con 45' 또는 'Justifier' 입력 장치를 사용할 때 화면에 커서를 표시합니다."
#define OPTION_VAL_CROSS_KO "십자"
#define OPTION_VAL_DOT_KO "점"
#define BEETLE_OPT_CROSSHAIR_COLOR_P1_LABEL_KO "건 십자선 색: 포트 1"
#define BEETLE_OPT_CROSSHAIR_COLOR_P1_INFO_0_KO "1번 포트 라이트 건의 십자선 색입니다."
#define OPTION_VAL_RED_KO "빨강"
#define OPTION_VAL_BLUE_KO "파랑"
#define OPTION_VAL_GREEN_KO "초록"
#define OPTION_VAL_ORANGE_KO "오렌지"
#define OPTION_VAL_YELLOW_KO "노랑"
#define OPTION_VAL_CYAN_KO "청록"
#define OPTION_VAL_PINK_KO "핑크"
#define OPTION_VAL_PURPLE_KO "보라"
#define OPTION_VAL_BLACK_KO "검정"
#define OPTION_VAL_WHITE_KO "흰색"
#define BEETLE_OPT_CROSSHAIR_COLOR_P2_LABEL_KO "건 십자선 색: 포트 2"
#define BEETLE_OPT_CROSSHAIR_COLOR_P2_INFO_0_KO "2번 포트 라이트 건의 십자선 색입니다."
#define BEETLE_OPT_MOUSE_SENSITIVITY_LABEL_KO "마우스 감도"
#define BEETLE_OPT_MOUSE_SENSITIVITY_INFO_0_KO "'마우스' 형식 입력 장치의 반응 정도입니다."
#define OPTION_VAL_5_O42_KO NULL
#define OPTION_VAL_10_KO NULL
#define OPTION_VAL_15_KO NULL
#define OPTION_VAL_20_KO NULL
#define OPTION_VAL_25_KO NULL
#define OPTION_VAL_30_KO NULL
#define OPTION_VAL_35_KO NULL
#define OPTION_VAL_40_KO NULL
#define OPTION_VAL_45_KO NULL
#define OPTION_VAL_55_KO NULL
#define OPTION_VAL_65_KO NULL
#define OPTION_VAL_75_KO NULL
#define OPTION_VAL_85_KO NULL
#define OPTION_VAL_95_KO NULL
#define OPTION_VAL_100_O42_KO NULL
#define OPTION_VAL_105_KO NULL
#define OPTION_VAL_115_KO NULL
#define OPTION_VAL_125_KO NULL
#define OPTION_VAL_135_KO NULL
#define OPTION_VAL_145_KO NULL
#define OPTION_VAL_155_KO NULL
#define OPTION_VAL_165_KO NULL
#define OPTION_VAL_175_KO NULL
#define OPTION_VAL_185_KO NULL
#define OPTION_VAL_195_KO NULL
#define BEETLE_OPT_NEGCON_RESPONSE_LABEL_KO "neGcon 뒤틀기 감도"
#define BEETLE_OPT_NEGCON_RESPONSE_INFO_0_KO "RetroPad 왼쪽 아날로그 스틱으로 'neGcon' 입력 장치의 '뒤틀기' 입력을 흉내낼 때의 반응 형식입니다. 아날로그 스틱 입력은 neGcon 회전에 선형, 정방(제곱), 입방(세제곱)으로 매핑될 수 있습니다. '정방'은 작은 움직임을 '선형'보다 더 정확하게 입력할 수 있습니다. '입방'은 작은 움직임을 더욱 정확하게 감지하지만, 큰 움직임이 '과장'됩니다. 주의: '선형'은 레이싱 휠을 사용할 때만 권장합니다. 일반적인 컨트롤러의 아날로그 입력은 neGcon의 '비틀기' 입력과 전혀 다르게 설계되어 있으므로 선형 매핑은 작은 움직임을 과장하여 미세 조작을 힘들게 합니다. 일반적으로 '정방'이 실제 하드웨어와 가장 비슷한 느낌을 제공합니다."
#define OPTION_VAL_LINEAR_KO "선형"
#define OPTION_VAL_QUADRATIC_KO "정방"
#define OPTION_VAL_CUBIC_KO "입방"
#define BEETLE_OPT_NEGCON_DEADZONE_LABEL_KO "neGcon 뒤틀기 데드존"
#define BEETLE_OPT_NEGCON_DEADZONE_INFO_0_KO "RetroPad 왼쪽 아날로그 스틱으로 'neGcon' 입력 장치의 '뒤틀기' 입력을 흉내낼 때의 데드존입니다. 컨트롤러 쏠림 현장을 방지하기 위해 사용됩니다. 주의: 대부분의 neGcon 호환 타이틀은 게임 내에서 '뒤틀기' 데드존 값을 설정하는 것을 지원합니다. 가장 정확한 입력을 위해서는 게임 내 데드존 설정이 *반드시* 0으로 설정되어야 하며, 관련 조정은 *반드시* 이 코어 설정을 통해서만 이루어져야 합니다. 특히 'neGcon 뒤틀기 감도'가 '정방' 또는 '입방'일 경우에는 이렇게 하는 것이 중요합니다."
#define OPTION_VAL_0_O44_KO NULL
#define BEETLE_OPT_USE_MEDNAFEN_MEMCARD0_METHOD_LABEL_KO "메모리 카드 방식"
#define BEETLE_OPT_USE_MEDNAFEN_MEMCARD0_METHOD_INFO_0_KO "왼쪽 메모리 카드 슬롯의 저장 데이터 포맷입니다. 'Mednafen'은 단독 버전의 Mednafen과의 호환성을 위해 사용될 수 있습니다. Beetle PSX에서 사용할 경우, Libretro(.srm)와 Mednafen(.mcr) 저장 파일은 내부적으로 동일한 형식이며 파일 이름을 바꾸는 것으로 상호 변환할 수 있습니다."
#define OPTION_VAL_LIBRETRO_KO NULL
#define OPTION_VAL_MEDNAFEN_KO NULL
#define BEETLE_OPT_ENABLE_MEMCARD1_LABEL_KO "메모리 카드 슬롯 2"
#define BEETLE_OPT_ENABLE_MEMCARD1_INFO_0_KO "오른쪽 슬롯에 두 번째 메모리 카드를 에뮬레이트합니다. 주의: 일부 게임(예: Codename Tenka)은 이 옵션을 비활성화해야 정상 작동합니다. 주의: 메모리 카드 슬롯 2는 Mednafen(.mcr) 저장 포맷을 사용합니다. 재시작이 필요합니다."
#define BEETLE_OPT_SHARED_MEMORY_CARDS_LABEL_KO "메모리 카드 공유"
#define BEETLE_OPT_SHARED_MEMORY_CARDS_INFO_0_KO "컨텐츠마다 개별 메모리 카드가 아닌, 모든 컨텐츠가 공유하는 메모리 카드를 사용합니다. 주의: 만약 '메모리 카드 방식'이 'Libretro'로 설정된 경우, 오직 메모리 카드 슬롯 2만 이 설정의 영향을 받습니다. 재시작이 필요합니다."
#define BEETLE_OPT_MEMCARD_LEFT_INDEX_LABEL_KO "메모리 카드 슬롯 1 인덱스"
#define BEETLE_OPT_MEMCARD_LEFT_INDEX_INFO_0_KO "왼쪽 슬롯에 삽입될 메모리 카드입니다. 이 옵션은 메모리 카드 방식이 Mednafen으로 설정되어 있어야 작동합니다."
#define OPTION_VAL_0_O48_KO NULL
#define BEETLE_OPT_MEMCARD_RIGHT_INDEX_LABEL_KO "메모리 카드 슬롯 2 인덱스"
#define BEETLE_OPT_MEMCARD_RIGHT_INDEX_INFO_0_KO "오른쪽 슬롯에 삽입될 메모리 카드입니다. 이 옵션은 메모리 카드 슬롯 2가 활성화되어 있어야 작동합니다."
#define OPTION_VAL_1_O49_KO NULL
#define BEETLE_OPT_LINE_RENDER_LABEL_KO "Line-to-Quad 핵"
#define BEETLE_OPT_LINE_RENDER_INFO_0_KO "일부 게임(예: Doom, Hexen, Soul Blade 등)은 수평선을 원본 하드웨어에서 한 줄의 픽셀로 래스터화되는 1픽셀 높이의 가로로 긴 삼각형을 그리는 방식으로 구현합니다. 이 핵은 그러한 작은 삼각형을 탐지하여 하드웨어 렌더러에서 올바르게 그려지고 업스케일링도 할 수 있는 사각형으로 변환합니다. '과감하게'는 일부 타이틀(예: Dark Forces, Duke Nukem)의 구동에 필요하나 이외의 경우 그래픽 오류를 유발할 수 있습니다. 불확실할 경우 '기본'으로 설정하십시오."
#define OPTION_VAL_DEFAULT_KO "기본"
#define OPTION_VAL_AGGRESSIVE_KO "과감하게"
#define BEETLE_OPT_WIDESCREEN_HACK_LABEL_KO "와이드스크린 모드 핵"
#define BEETLE_OPT_WIDESCREEN_HACK_INFO_0_KO "3D 컨텐츠를 아나모픽하게 렌더링한 뒤 에뮬레이트되는 프레임버퍼에 와이드스크린 화면비로 출력합니다. 풀 3D 게임에서 최선의 효과를 얻을 수 있습니다. 2D 요소는 가로로 늘어나며 뒤틀릴 수 있습니다."
#define BEETLE_OPT_WIDESCREEN_HACK_ASPECT_RATIO_LABEL_KO "와이드스크린 모드 핵 화면비"
#define BEETLE_OPT_WIDESCREEN_HACK_ASPECT_RATIO_INFO_0_KO "와이드스크린 모드 핵에서 사용될 화면비입니다."
#define OPTION_VAL_16_9_KO NULL
#define OPTION_VAL_16_10_KO NULL
#define OPTION_VAL_18_9_KO NULL
#define OPTION_VAL_19_9_KO NULL
#define OPTION_VAL_20_9_KO NULL
#define OPTION_VAL_21_9_KO NULL
#define OPTION_VAL_32_9_KO NULL
#define BEETLE_OPT_CPU_DYNAREC_LABEL_KO "CPU 동적 리컴파일러"
#define BEETLE_OPT_CPU_DYNAREC_INFO_0_KO "CPU 명령어를 네이티브 명령어로 동적 리컴파일합니다. 인터프리터보다 훨씬 빠르지만, CPU 타이밍이 부정확해지며, 오류를 일으킬 수 있습니다."
#define OPTION_VAL_DISABLED_KO "비활성화 (Beetle 인터프리터)"
#define OPTION_VAL_EXECUTE_KO "최대 성능"
#define OPTION_VAL_EXECUTE_ONE_KO "사이클 타이밍 검사"
#define OPTION_VAL_RUN_INTERPRETER_KO "Lightrec 인터프리터"
#define BEETLE_OPT_DYNAREC_INVALIDATE_LABEL_KO "동적 리컴파일러 코드 무효화"
#define BEETLE_OPT_DYNAREC_INVALIDATE_INFO_0_KO "일부 게임은 '모두' 무효화를 필요로 하며, 일부는 'DMA만'을 필요로 합니다."
#define OPTION_VAL_FULL_KO "모두"
#define OPTION_VAL_DMA_KO "DMA만 (조금 더 빠름)"
#define BEETLE_OPT_DYNAREC_EVENTCYCLES_LABEL_KO "동적 리컴파일러 DMA/GPU/MDEC/타이머 이벤트 주기"
#define BEETLE_OPT_DYNAREC_EVENTCYCLES_INFO_0_KO "GPU/DMA/MDEC/타이머 업데이트 확인 전에 구동할 최대 CPU 사이클 수입니다. 숫자가 높을수록 더 빨라지며, 동적 리컴파일러(Dynarec)에 비해 Beetle 인터프리터에는 큰 효과가 없습니다."
#define OPTION_VAL_128_KO "128 (기본값)"
#define BEETLE_OPT_DYNAREC_SPU_SAMPLES_LABEL_KO "동적 리컴파일러 SPU 샘플"
#define BEETLE_OPT_DYNAREC_SPU_SAMPLES_INFO_0_KO "SPU 업데이트 확인 전 실행할 최대 SPU 샘플 수. 숫자가 높을수록 더 빨라지지만 값이 1이 아닌 경우 일부 게임에서 사운드 결함이 발생합니다."
#define BEETLE_OPT_CORE_TIMING_FPS_LABEL_KO "코어 보고 FPS 타이밍"
#define BEETLE_OPT_CORE_TIMING_FPS_INFO_0_KO "자동 전환을 활성화하면 코어가 주사 방식을 프로그레시브 또는 인터레이스로 전환할 수 있게 됩니다."
#define OPTION_VAL_AUTO_TOGGLE_KO "자동 전환"
#define OPTION_VAL_FORCE_PROGRESSIVE_KO "프로그레시브 주사 강제"
#define OPTION_VAL_FORCE_INTERLACED_KO "인터레이스 주사율 강제"
#define BEETLE_OPT_DEINTERLACER_LABEL_KO "디인터레이스 방식"
#define BEETLE_OPT_DEINTERLACER_INFO_0_KO NULL
#define OPTION_VAL_WEAVE_KO NULL
#define OPTION_VAL_BOB_KO NULL
#define OPTION_VAL_BOB_OFFSET_KO NULL
#define OPTION_VAL_FASTMAD_KO NULL
#define BEETLE_OPT_FRAME_DUPING_LABEL_KO "프레임 반복"
#define BEETLE_OPT_FRAME_DUPING_INFO_0_KO "새로 표시할 내용이 없을 경우 이전 프레임을 한 번 더 표시합니다."
#define BEETLE_OPT_PAL_VIDEO_TIMING_OVERRIDE_LABEL_KO "PAL 비디오 타이밍 오버라이드"
#define BEETLE_OPT_PAL_VIDEO_TIMING_OVERRIDE_INFO_0_KO "서로 다른 표준으로 인해 PAL 게임들은 미국 또는 일본의 NTSC 게임들에 비해 다소 느리게 구동되기도 합니다. 이 옵션은 PAL 타이밍을 오버라이드하여 NTSC 게임과 동일한 프레임레이트로 구동되게 합니다."
#define BEETLE_OPT_ASPECT_RATIO_LABEL_KO "코어 화면비"
#define BEETLE_OPT_ASPECT_RATIO_INFO_0_KO "이 설정은 와이드스크린 모드 핵 또는 전체 VRAM 표시 옵션이 활성화된 경우 무시됩니다."
#define OPTION_VAL_CORRECTED_KO "교정됨"
#define OPTION_VAL_UNCORRECTED_KO "교정 안 됨"
#define OPTION_VAL_4_3_KO "강제 4:3"
#define OPTION_VAL_NTSC_KO "강제 NTSC"
#define BEETLE_OPT_CROP_OVERSCAN_LABEL_KO "오버스캔 잘라내기"
#define BEETLE_OPT_CROP_OVERSCAN_INFO_0_KO "네이티브 전체를 표시할 지 또는 테두리를 제외하고 표시할 지 선택합니다."
#define OPTION_VAL_STATIC_KO "가로"
#define OPTION_VAL_SMART_KO "가로 + 세로"
#define BEETLE_OPT_IMAGE_CROP_LABEL_KO "추가 잘라내기"
#define BEETLE_OPT_IMAGE_CROP_INFO_0_KO "지정된 수의 픽셀만큼 이미지의 가로 너비를 잘라냅니다."
#define OPTION_VAL_9PX_KO NULL
#define OPTION_VAL_10PX_KO NULL
#define OPTION_VAL_11PX_KO NULL
#define OPTION_VAL_12PX_KO NULL
#define OPTION_VAL_13PX_KO NULL
#define OPTION_VAL_14PX_KO NULL
#define OPTION_VAL_15PX_KO NULL
#define OPTION_VAL_16PX_KO NULL
#define OPTION_VAL_17PX_KO NULL
#define OPTION_VAL_18PX_KO NULL
#define OPTION_VAL_19PX_KO NULL
#define OPTION_VAL_20PX_KO NULL
#define BEETLE_OPT_IMAGE_OFFSET_LABEL_KO "잘라낸 이미지 위치 조정"
#define BEETLE_OPT_IMAGE_OFFSET_INFO_0_KO "화면의 가로 위치를 지정된 픽셀 수만큼 오른쪽(양수) 또는 왼쪽(음수)으로 이동합니다. 정렬 문제를 해결할 수 있습니다. 소프트웨어 렌더러에서만 지원됩니다."
#define OPTION_VAL_12PX_O64_KO NULL
#define OPTION_VAL_11PX_O64_KO NULL
#define OPTION_VAL_10PX_O64_KO NULL
#define OPTION_VAL_9PX_O64_KO NULL
#define OPTION_VAL_8PX_O64_KO NULL
#define OPTION_VAL_7PX_O64_KO NULL
#define OPTION_VAL_6PX_O64_KO NULL
#define OPTION_VAL_5PX_O64_KO NULL
#define OPTION_VAL_4PX_O64_KO NULL
#define OPTION_VAL_3PX_O64_KO NULL
#define OPTION_VAL_2PX_O64_KO NULL
#define OPTION_VAL_1PX_O64_KO NULL
#define OPTION_VAL_1PX_O64_0_KO NULL
#define OPTION_VAL_2PX_O64_0_KO NULL
#define OPTION_VAL_3PX_O64_0_KO NULL
#define OPTION_VAL_4PX_O64_0_KO NULL
#define OPTION_VAL_5PX_O64_0_KO NULL
#define OPTION_VAL_6PX_O64_0_KO NULL
#define OPTION_VAL_7PX_O64_0_KO NULL
#define OPTION_VAL_8PX_O64_0_KO NULL
#define OPTION_VAL_9PX_O64_0_KO NULL
#define OPTION_VAL_10PX_O64_0_KO NULL
#define OPTION_VAL_11PX_O64_0_KO NULL
#define OPTION_VAL_12PX_O64_0_KO NULL
#define BEETLE_OPT_IMAGE_OFFSET_CYCLES_LABEL_KO "가로 이미지 위치 조정 (GPU 사이클)"
#define BEETLE_OPT_IMAGE_OFFSET_CYCLES_INFO_0_KO "양수는 이미지를 오른쪽으로, 음수는 왼쪽으로 이동시킵니다. 하드웨어 렌더러에서만 지원됩니다."
#define BEETLE_OPT_INITIAL_SCANLINE_LABEL_KO "첫 스캔라인 - NTSC"
#define BEETLE_OPT_INITIAL_SCANLINE_INFO_0_KO "NTSC 컨텐츠를 실행할 때 표시할 첫 번째 스캔라인입니다. 0보다 큰 값으로 설정할 경우 출력되는 이미지를 위에서부터 잘라내 높이를 줄입니다. 레터박스를 없애는 데 사용될 수 있습니다."
#define BEETLE_OPT_LAST_SCANLINE_LABEL_KO "마지막 스캔라인 - NTSC"
#define BEETLE_OPT_LAST_SCANLINE_INFO_0_KO "NTSC 컨텐츠를 실행할 때 표시할 마지막 스캔라인입니다. 239보다 작은 값으로 설정할 경우 출력되는 이미지를 아래에서부터 잘라내 높이를 줄입니다. 레터박스를 없애는 데 사용될 수 있습니다."
#define OPTION_VAL_239_KO "239 (기본)"
#define BEETLE_OPT_INITIAL_SCANLINE_PAL_LABEL_KO "첫 스캔라인 - PAL"
#define BEETLE_OPT_INITIAL_SCANLINE_PAL_INFO_0_KO "PAL 컨텐츠를 실행할 때 표시할 첫 번째 스캔라인입니다. 0보다 큰 값으로 설정할 경우 출력되는 이미지를 위에서부터 잘라내 높이를 줄입니다. 레터박스를 없애는 데 사용될 수 있습니다."
#define BEETLE_OPT_LAST_SCANLINE_PAL_LABEL_KO "마지막 스캔라인 - PAL"
#define BEETLE_OPT_LAST_SCANLINE_PAL_INFO_0_KO "PAL 컨텐츠를 실행할 때 표시할 마지막 스캔라인입니다. 287보다 작은 값으로 설정할 경우 출력되는 이미지를 아래에서부터 잘라내 높이를 줄입니다. 레터박스를 없애는 데 사용될 수 있습니다."
#define OPTION_VAL_287_KO "287 (기본)"

struct retro_core_option_v2_category option_cats_ko[] = {
   {
      "system",
      CATEGORY_SYSTEM_LABEL_KO,
      CATEGORY_SYSTEM_INFO_0_KO
   },
   {
      "video",
      CATEGORY_VIDEO_LABEL_KO,
      CATEGORY_VIDEO_INFO_0_KO
   },
   {
      "pgxp",
      CATEGORY_PGXP_LABEL_KO,
      CATEGORY_PGXP_INFO_0_KO
   },
   {
      "osd",
      CATEGORY_OSD_LABEL_KO,
      CATEGORY_OSD_INFO_0_KO
   },
   {
      "input",
      CATEGORY_INPUT_LABEL_KO,
      CATEGORY_INPUT_INFO_0_KO
   },
   {
      "memcards",
      CATEGORY_MEMCARDS_LABEL_KO,
      CATEGORY_MEMCARDS_INFO_0_KO
   },
   {
      "hacks",
      CATEGORY_HACKS_LABEL_KO,
      CATEGORY_HACKS_INFO_0_KO
   },
   { NULL, NULL, NULL },
};
struct retro_core_option_v2_definition option_defs_ko[] = {
   {
      BEETLE_OPT(cpu_freq_scale),
      BEETLE_OPT_CPU_FREQ_SCALE_LABEL_KO,
      NULL,
      BEETLE_OPT_CPU_FREQ_SCALE_INFO_0_KO,
      NULL,
      "system",
      {
         { "50%",           OPTION_VAL_50_KO },
         { "60%",           OPTION_VAL_60_KO },
         { "70%",           OPTION_VAL_70_KO },
         { "80%",           OPTION_VAL_80_KO },
         { "90%",           OPTION_VAL_90_KO },
         { "100%",          OPTION_VAL_100_KO },
         { "110%",          OPTION_VAL_110_KO },
         { "120%",          OPTION_VAL_120_KO },
         { "130%",          OPTION_VAL_130_KO },
         { "140%",          OPTION_VAL_140_KO },
         { "150%",          OPTION_VAL_150_KO },
         { "160%",          OPTION_VAL_160_KO },
         { "170%",          OPTION_VAL_170_KO },
         { "180%",          OPTION_VAL_180_KO },
         { "190%",          OPTION_VAL_190_KO },
         { "200%",          OPTION_VAL_200_KO },
         { "210%",          OPTION_VAL_210_KO },
         { "220%",          OPTION_VAL_220_KO },
         { "230%",          OPTION_VAL_230_KO },
         { "240%",          OPTION_VAL_240_KO },
         { "250%",          OPTION_VAL_250_KO },
         { "260%",          OPTION_VAL_260_KO },
         { "270%",          OPTION_VAL_270_KO },
         { "280%",          OPTION_VAL_280_KO },
         { "290%",          OPTION_VAL_290_KO },
         { "300%",          OPTION_VAL_300_KO },
         { "310%",          OPTION_VAL_310_KO },
         { "320%",          OPTION_VAL_320_KO },
         { "330%",          OPTION_VAL_330_KO },
         { "340%",          OPTION_VAL_340_KO },
         { "350%",          OPTION_VAL_350_KO },
         { "360%",          OPTION_VAL_360_KO },
         { "370%",          OPTION_VAL_370_KO },
         { "380%",          OPTION_VAL_380_KO },
         { "390%",          OPTION_VAL_390_KO },
         { "400%",          OPTION_VAL_400_KO },
         { "410%",          OPTION_VAL_410_KO },
         { "420%",          OPTION_VAL_420_KO },
         { "430%",          OPTION_VAL_430_KO },
         { "440%",          OPTION_VAL_440_KO },
         { "450%",          OPTION_VAL_450_KO },
         { "460%",          OPTION_VAL_460_KO },
         { "470%",          OPTION_VAL_470_KO },
         { "480%",          OPTION_VAL_480_KO },
         { "490%",          OPTION_VAL_490_KO },
         { "500%",          OPTION_VAL_500_KO },
         { "510%",          OPTION_VAL_510_KO },
         { "520%",          OPTION_VAL_520_KO },
         { "530%",          OPTION_VAL_530_KO },
         { "540%",          OPTION_VAL_540_KO },
         { "550%",          OPTION_VAL_550_KO },
         { "560%",          OPTION_VAL_560_KO },
         { "570%",          OPTION_VAL_570_KO },
         { "580%",          OPTION_VAL_580_KO },
         { "590%",          OPTION_VAL_590_KO },
         { "600%",          OPTION_VAL_600_KO },
         { "610%",          OPTION_VAL_610_KO },
         { "620%",          OPTION_VAL_620_KO },
         { "630%",          OPTION_VAL_630_KO },
         { "640%",          OPTION_VAL_640_KO },
         { "650%",          OPTION_VAL_650_KO },
         { "660%",          OPTION_VAL_660_KO },
         { "670%",          OPTION_VAL_670_KO },
         { "680%",          OPTION_VAL_680_KO },
         { "690%",          OPTION_VAL_690_KO },
         { "700%",          OPTION_VAL_700_KO },
         { "710%",          OPTION_VAL_710_KO },
         { "720%",          OPTION_VAL_720_KO },
         { "730%",          OPTION_VAL_730_KO },
         { "740%",          OPTION_VAL_740_KO },
         { "750%",          OPTION_VAL_750_KO },
         { NULL, NULL },
      },
      "100%"
   },
   {
      BEETLE_OPT(gpu_overclock),
      BEETLE_OPT_GPU_OVERCLOCK_LABEL_KO,
      NULL,
      BEETLE_OPT_GPU_OVERCLOCK_INFO_0_KO,
      NULL,
      "system",
      {
         { "1x(native)", OPTION_VAL_1X_NATIVE_KO },
         { "2x",         OPTION_VAL_2X_KO },
         { "4x",         OPTION_VAL_4X_KO },
         { "8x",         OPTION_VAL_8X_KO },
         { "16x",        OPTION_VAL_16X_KO },
         { "32x",        OPTION_VAL_32X_KO },
         { NULL, NULL },
      },
      "1x(native)"
   },
   {
      BEETLE_OPT(gte_overclock),
      BEETLE_OPT_GTE_OVERCLOCK_LABEL_KO,
      NULL,
      BEETLE_OPT_GTE_OVERCLOCK_INFO_0_KO,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
#ifndef EMSCRIPTEN
   {
      BEETLE_OPT(cd_access_method),
      BEETLE_OPT_CD_ACCESS_METHOD_LABEL_KO,
      NULL,
      BEETLE_OPT_CD_ACCESS_METHOD_INFO_0_KO,
      NULL,
      "system",
      {
         { "sync",     OPTION_VAL_SYNC_KO },
         { "async",    OPTION_VAL_ASYNC_KO },
         { "precache", OPTION_VAL_PRECACHE_KO },
         { NULL, NULL },
      },
      "sync"
   },
#endif
   {
      BEETLE_OPT(cd_fastload),
      BEETLE_OPT_CD_FASTLOAD_LABEL_KO,
      NULL,
      BEETLE_OPT_CD_FASTLOAD_INFO_0_KO,
      NULL,
      "system",
      {
         { "2x(native)", OPTION_VAL_2X_NATIVE_KO },
         { "4x",          OPTION_VAL_4X_KO },
         { "6x",          OPTION_VAL_6X_KO },
         { "8x",          OPTION_VAL_8X_KO },
         { "10x",         OPTION_VAL_10X_KO },
         { "12x",         OPTION_VAL_12X_KO },
         { "14x",         OPTION_VAL_14X_KO },
         { NULL, NULL },
      },
      "2x(native)"
   },
   {
      BEETLE_OPT(skip_bios),
      BEETLE_OPT_SKIP_BIOS_LABEL_KO,
      NULL,
      BEETLE_OPT_SKIP_BIOS_INFO_0_KO,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(override_bios),
      BEETLE_OPT_OVERRIDE_BIOS_LABEL_KO,
      NULL,
      BEETLE_OPT_OVERRIDE_BIOS_INFO_0_KO,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "psxonpsp", OPTION_VAL_PSXONPSP_KO },
         { "ps1_rom",  OPTION_VAL_PS1_ROM_KO },
         { "openbios", OPTION_VAL_OPENBIOS_KO     },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(region),
      BEETLE_OPT_REGION_LABEL_KO,
      NULL,
      BEETLE_OPT_REGION_INFO_0_KO,
      NULL,
      "system",
      {
         { "auto",   OPTION_VAL_AUTO_KO },
         { "ntsc-j", OPTION_VAL_NTSC_J_KO },
         { "ntsc-u", OPTION_VAL_NTSC_U_KO },
         { "pal",    OPTION_VAL_PAL_KO },
         { NULL, NULL },
      },
      "auto"
   },
   {
      BEETLE_OPT(internal_resolution),
      BEETLE_OPT_INTERNAL_RESOLUTION_LABEL_KO,
      NULL,
      BEETLE_OPT_INTERNAL_RESOLUTION_INFO_0_KO,
      NULL,
      "video",
      {
         { "1x(native)", OPTION_VAL_1X_NATIVE_KO },
         { "2x",         OPTION_VAL_2X_KO },
         { "4x",         OPTION_VAL_4X_KO },
         { "8x",         OPTION_VAL_8X_KO },
         { "16x",        OPTION_VAL_16X_KO },
         { NULL, NULL },
      },
      "1x(native)"
   },
#if defined(HAVE_OPENGL) || defined(HAVE_OPENGLES) || defined(HAVE_VULKAN)
   {
      BEETLE_OPT(renderer),
      BEETLE_OPT_RENDERER_LABEL_KO,
      NULL,
      BEETLE_OPT_RENDERER_INFO_0_KO,
      NULL,
      "video",
      {
         { "hardware",    OPTION_VAL_HARDWARE_KO },
#if defined(HAVE_OPENGL) || defined(HAVE_OPENGLES)
         { "hardware_gl", OPTION_VAL_HARDWARE_GL_KO },
#endif
#if defined(HAVE_VULKAN)
         { "hardware_vk", OPTION_VAL_HARDWARE_VK_KO },
#endif
         { "software",    OPTION_VAL_SOFTWARE_KO },
         { NULL, NULL },
      },
      "hardware"
   },
   {
      BEETLE_OPT(renderer_software_fb),
      BEETLE_OPT_RENDERER_SOFTWARE_FB_LABEL_KO,
      NULL,
      BEETLE_OPT_RENDERER_SOFTWARE_FB_INFO_0_KO,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled"
   },
#endif
#if defined(HAVE_OPENGL) || defined(HAVE_OPENGLES)
   {
      BEETLE_OPT(depth),
      BEETLE_OPT_DEPTH_LABEL_KO,
      NULL,
      BEETLE_OPT_DEPTH_INFO_0_KO,
      NULL,
      "video",
      {
         { "16bpp(native)", OPTION_VAL_16BPP_NATIVE_KO },
         { "32bpp",         OPTION_VAL_32BPP_KO },
         { NULL, NULL },
      },
      "16bpp(native)"
   },
   // Sort of, it's more like 15-bit high color and 24-bit true color for visible output. The alpha channel is used for mask bit. Vulkan renderer uses ABGR1555_555 for 31 bits internal? FMVs are always 24-bit on all renderers like original hardware (BGR888, no alpha)
#endif
   {
      BEETLE_OPT(dither_mode),
      BEETLE_OPT_DITHER_MODE_LABEL_KO,
      NULL,
      BEETLE_OPT_DITHER_MODE_INFO_0_KO,
      NULL,
      "video",
      {
         { "1x(native)",          OPTION_VAL_1X_NATIVE_KO },
         { "internal resolution", OPTION_VAL_INTERNAL_RESOLUTION_KO },
         { "disabled",            NULL },
         { NULL, NULL },
      },
      "1x(native)"
   },
#ifdef HAVE_VULKAN
   {
      BEETLE_OPT(scaled_uv_offset),
      BEETLE_OPT_SCALED_UV_OFFSET_LABEL_KO,
      NULL,
      BEETLE_OPT_SCALED_UV_OFFSET_INFO_0_KO,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled"
   },
#endif
#if defined(HAVE_OPENGL) || defined(HAVE_OPENGLES) || defined(HAVE_VULKAN)
   {
      BEETLE_OPT(filter),
      BEETLE_OPT_FILTER_LABEL_KO,
      NULL,
      BEETLE_OPT_FILTER_INFO_0_KO,
      NULL,
      "video",
      {
         { "nearest",  OPTION_VAL_NEAREST_KO },
         { "SABR",     OPTION_VAL_SABR_KO },
         { "xBR",      OPTION_VAL_XBR_KO },
         { "bilinear", OPTION_VAL_BILINEAR_KO },
         { "3-point",  OPTION_VAL_3_POINT_KO },
         { "JINC2",    OPTION_VAL_JINC2_KO },
         { NULL, NULL },
      },
      "nearest"
   },
#ifdef HAVE_VULKAN
   {
      BEETLE_OPT(filter_exclude_sprite),
      BEETLE_OPT_FILTER_EXCLUDE_SPRITE_LABEL_KO,
      NULL,
      BEETLE_OPT_FILTER_EXCLUDE_SPRITE_INFO_0_KO,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "opaque", OPTION_VAL_OPAQUE_KO },
         { "all", OPTION_VAL_ALL_KO },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(filter_exclude_2d_polygon),
      BEETLE_OPT_FILTER_EXCLUDE_2D_POLYGON_LABEL_KO,
      NULL,
      BEETLE_OPT_FILTER_EXCLUDE_2D_POLYGON_INFO_0_KO,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "opaque", OPTION_VAL_OPAQUE_KO },
         { "all", OPTION_VAL_ALL_KO },
         { NULL, NULL },
      },
      "disabled"
   },
#endif
#endif
#ifdef HAVE_VULKAN
   {
      BEETLE_OPT(adaptive_smoothing),
      BEETLE_OPT_ADAPTIVE_SMOOTHING_LABEL_KO,
      NULL,
      BEETLE_OPT_ADAPTIVE_SMOOTHING_INFO_0_KO,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(super_sampling),
      BEETLE_OPT_SUPER_SAMPLING_LABEL_KO,
      NULL,
      BEETLE_OPT_SUPER_SAMPLING_INFO_0_KO,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(msaa),
      BEETLE_OPT_MSAA_LABEL_KO,
      NULL,
      BEETLE_OPT_MSAA_INFO_0_KO,
      NULL,
      "video",
      {
         { "1x",  OPTION_VAL_1X_KO },
         { "2x",  OPTION_VAL_2X_KO },
         { "4x",  OPTION_VAL_4X_KO },
         { "8x",  OPTION_VAL_8X_KO },
         { "16x", OPTION_VAL_16X_KO },
         { NULL, NULL },
      },
      "1x"
   },
   {
      BEETLE_OPT(mdec_yuv),
      BEETLE_OPT_MDEC_YUV_LABEL_KO,
      NULL,
      BEETLE_OPT_MDEC_YUV_INFO_0_KO,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(track_textures),
      BEETLE_OPT_TRACK_TEXTURES_LABEL_KO,
      NULL,
      BEETLE_OPT_TRACK_TEXTURES_INFO_0_KO,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
#ifdef TEXTURE_DUMPING_ENABLED
   {
      BEETLE_OPT(dump_textures),
      BEETLE_OPT_DUMP_TEXTURES_LABEL_KO,
      NULL,
      BEETLE_OPT_DUMP_TEXTURES_INFO_0_KO,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
#endif
   {
      BEETLE_OPT(replace_textures),
      BEETLE_OPT_REPLACE_TEXTURES_LABEL_KO,
      NULL,
      BEETLE_OPT_REPLACE_TEXTURES_INFO_0_KO,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
#endif
   {
      BEETLE_OPT(pgxp_mode),
      BEETLE_OPT_PGXP_MODE_LABEL_KO,
      NULL,
      BEETLE_OPT_PGXP_MODE_INFO_0_KO,
      NULL,
      "pgxp",
      {
         { "disabled",     NULL },
         { "memory only",  OPTION_VAL_MEMORY_ONLY_KO },
         { "memory + CPU", OPTION_VAL_MEMORY_CPU_KO },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(pgxp_2d_tol),
      BEETLE_OPT_PGXP_2D_TOL_LABEL_KO,
      NULL,
      BEETLE_OPT_PGXP_2D_TOL_INFO_0_KO,
      NULL,
      "pgxp",
      {
         { "disabled", NULL },
         { "0px", OPTION_VAL_0PX_KO },
         { "1px", OPTION_VAL_1PX_KO },
         { "2px", OPTION_VAL_2PX_KO },
         { "3px", OPTION_VAL_3PX_KO },
         { "4px", OPTION_VAL_4PX_KO },
         { "5px", OPTION_VAL_5PX_KO },
         { "6px", OPTION_VAL_6PX_KO },
         { "7px", OPTION_VAL_7PX_KO },
         { "8px", OPTION_VAL_8PX_KO },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(pgxp_nclip),
      BEETLE_OPT_PGXP_NCLIP_LABEL_KO,
      NULL,
      BEETLE_OPT_PGXP_NCLIP_INFO_0_KO,
      NULL,
      "pgxp",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
#if defined(HAVE_OPENGL) || defined(HAVE_OPENGLES) || defined(HAVE_VULKAN)
   {
      BEETLE_OPT(pgxp_vertex),
      BEETLE_OPT_PGXP_VERTEX_LABEL_KO,
      NULL,
      BEETLE_OPT_PGXP_VERTEX_INFO_0_KO,
      NULL,
      "pgxp",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
#endif
   {
      BEETLE_OPT(pgxp_texture),
      BEETLE_OPT_PGXP_TEXTURE_LABEL_KO,
      NULL,
      BEETLE_OPT_PGXP_TEXTURE_INFO_0_KO,
      NULL,
      "pgxp",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(display_internal_fps),
      BEETLE_OPT_DISPLAY_INTERNAL_FPS_LABEL_KO,
      NULL,
      BEETLE_OPT_DISPLAY_INTERNAL_FPS_INFO_0_KO,
      NULL,
      "osd",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(display_osd),
      BEETLE_OPT_DISPLAY_OSD_LABEL_KO,
      NULL,
      BEETLE_OPT_DISPLAY_OSD_INFO_0_KO,
      NULL,
      "osd",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled"
   },
#if defined(HAVE_OPENGL) || defined(HAVE_OPENGLES) || defined(HAVE_VULKAN)
   {
      BEETLE_OPT(display_vram),
      BEETLE_OPT_DISPLAY_VRAM_LABEL_KO,
      NULL,
      BEETLE_OPT_DISPLAY_VRAM_INFO_0_KO,
      NULL,
      "osd",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
#endif
   {
      BEETLE_OPT(analog_calibration),
      BEETLE_OPT_ANALOG_CALIBRATION_LABEL_KO,
      NULL,
      BEETLE_OPT_ANALOG_CALIBRATION_INFO_0_KO,
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(analog_toggle),
      BEETLE_OPT_ANALOG_TOGGLE_LABEL_KO,
      NULL,
      BEETLE_OPT_ANALOG_TOGGLE_INFO_0_KO,
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { "enabled-analog", OPTION_VAL_ENABLED_ANALOG_KO },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(analog_toggle_combo),
      BEETLE_OPT_ANALOG_TOGGLE_COMBO_LABEL_KO,
      NULL,
      BEETLE_OPT_ANALOG_TOGGLE_COMBO_INFO_0_KO,
      NULL,
      "input",
      {
         { "l1+l2+r1+r2+start+select", OPTION_VAL_L1_L2_R1_R2_START_SELECT_KO },
         { "l1+r1+select",             OPTION_VAL_L1_R1_SELECT_KO },
         { "l1+r1+start",              OPTION_VAL_L1_R1_START_KO },
         { "l1+r1+l3",                 OPTION_VAL_L1_R1_L3_KO },
         { "l1+r1+r3",                 OPTION_VAL_L1_R1_R3_KO },
         { "l2+r2+select",             OPTION_VAL_L2_R2_SELECT_KO },
         { "l2+r2+start",              OPTION_VAL_L2_R2_START_KO },
         { "l2+r2+l3",                 OPTION_VAL_L2_R2_L3_KO },
         { "l2+r2+r3",                 OPTION_VAL_L2_R2_R3_KO },
         { "l3+r3",                    OPTION_VAL_L3_R3_KO },
         { NULL, NULL },
      },
      "l1+r1+select"
   },
   {
      BEETLE_OPT(analog_toggle_hold),
      BEETLE_OPT_ANALOG_TOGGLE_HOLD_LABEL_KO,
      NULL,
      BEETLE_OPT_ANALOG_TOGGLE_HOLD_INFO_0_KO,
      NULL,
      "input",
      {
         { "0", OPTION_VAL_0_KO },
         { "1", OPTION_VAL_1_KO },
         { "2", OPTION_VAL_2_KO },
         { "3", OPTION_VAL_3_KO },
         { "4", OPTION_VAL_4_KO },
         { "5", OPTION_VAL_5_KO },
         { NULL, NULL },
      },
      "1"
   },
   {
      BEETLE_OPT(enable_multitap_port1),
      BEETLE_OPT_ENABLE_MULTITAP_PORT1_LABEL_KO,
      NULL,
      BEETLE_OPT_ENABLE_MULTITAP_PORT1_INFO_0_KO,
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(enable_multitap_port2),
      BEETLE_OPT_ENABLE_MULTITAP_PORT2_LABEL_KO,
      NULL,
      BEETLE_OPT_ENABLE_MULTITAP_PORT2_INFO_0_KO,
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(gun_input_mode),
      BEETLE_OPT_GUN_INPUT_MODE_LABEL_KO,
      NULL,
      BEETLE_OPT_GUN_INPUT_MODE_INFO_0_KO,
      NULL,
      "input",
      {
         { "lightgun",    OPTION_VAL_LIGHTGUN_KO },
         { "touchscreen", OPTION_VAL_TOUCHSCREEN_KO },
         { NULL, NULL },
      },
      "lightgun"
   },
   // Shouldn't the gun_input_mode just be Mouse vs. Touchscreen?
   {
      BEETLE_OPT(gun_cursor),
      BEETLE_OPT_GUN_CURSOR_LABEL_KO,
      NULL,
      BEETLE_OPT_GUN_CURSOR_INFO_0_KO,
      NULL,
      "input",
      {
         { "off",   "disabled" },
         { "cross", OPTION_VAL_CROSS_KO },
         { "dot",   OPTION_VAL_DOT_KO },
         { NULL, NULL },
      },
      "cross"
   },
   {
      BEETLE_OPT(crosshair_color_p1),
      BEETLE_OPT_CROSSHAIR_COLOR_P1_LABEL_KO,
      NULL,
      BEETLE_OPT_CROSSHAIR_COLOR_P1_INFO_0_KO,
      NULL,
      "input",
      {
         { "red",    OPTION_VAL_RED_KO },
         { "blue",   OPTION_VAL_BLUE_KO },
         { "green",  OPTION_VAL_GREEN_KO },
         { "orange", OPTION_VAL_ORANGE_KO },
         { "yellow", OPTION_VAL_YELLOW_KO },
         { "cyan",   OPTION_VAL_CYAN_KO },
         { "pink",   OPTION_VAL_PINK_KO },
         { "purple", OPTION_VAL_PURPLE_KO },
         { "black",  OPTION_VAL_BLACK_KO },
         { "white",  OPTION_VAL_WHITE_KO },
         { NULL, NULL },
      },
      "red"
   },
   {
      BEETLE_OPT(crosshair_color_p2),
      BEETLE_OPT_CROSSHAIR_COLOR_P2_LABEL_KO,
      NULL,
      BEETLE_OPT_CROSSHAIR_COLOR_P2_INFO_0_KO,
      NULL,
      "input",
      {
         { "blue",   OPTION_VAL_BLUE_KO },
         { "red",    OPTION_VAL_RED_KO },
         { "green",  OPTION_VAL_GREEN_KO },
         { "orange", OPTION_VAL_ORANGE_KO },
         { "yellow", OPTION_VAL_YELLOW_KO },
         { "cyan",   OPTION_VAL_CYAN_KO },
         { "pink",   OPTION_VAL_PINK_KO },
         { "purple", OPTION_VAL_PURPLE_KO },
         { "black",  OPTION_VAL_BLACK_KO },
         { "white",  OPTION_VAL_WHITE_KO },
         { NULL, NULL },
      },
      "blue"
   },
   {
      BEETLE_OPT(mouse_sensitivity),
      BEETLE_OPT_MOUSE_SENSITIVITY_LABEL_KO,
      NULL,
      BEETLE_OPT_MOUSE_SENSITIVITY_INFO_0_KO,
      NULL,
      "input",
      {
         { "5%",   OPTION_VAL_5_O42_KO },
         { "10%",  OPTION_VAL_10_KO },
         { "15%",  OPTION_VAL_15_KO },
         { "20%",  OPTION_VAL_20_KO },
         { "25%",  OPTION_VAL_25_KO },
         { "30%",  OPTION_VAL_30_KO },
         { "35%",  OPTION_VAL_35_KO },
         { "40%",  OPTION_VAL_40_KO },
         { "45%",  OPTION_VAL_45_KO },
         { "50%",  OPTION_VAL_50_KO },
         { "55%",  OPTION_VAL_55_KO },
         { "60%",  OPTION_VAL_60_KO },
         { "65%",  OPTION_VAL_65_KO },
         { "70%",  OPTION_VAL_70_KO },
         { "75%",  OPTION_VAL_75_KO },
         { "80%",  OPTION_VAL_80_KO },
         { "85%",  OPTION_VAL_85_KO },
         { "90%",  OPTION_VAL_90_KO },
         { "95%",  OPTION_VAL_95_KO },
         { "100%", OPTION_VAL_100_O42_KO },
         { "105%", OPTION_VAL_105_KO },
         { "110%", OPTION_VAL_110_KO },
         { "115%", OPTION_VAL_115_KO },
         { "120%", OPTION_VAL_120_KO },
         { "125%", OPTION_VAL_125_KO },
         { "130%", OPTION_VAL_130_KO },
         { "135%", OPTION_VAL_135_KO },
         { "140%", OPTION_VAL_140_KO },
         { "145%", OPTION_VAL_145_KO },
         { "150%", OPTION_VAL_150_KO },
         { "155%", OPTION_VAL_155_KO },
         { "160%", OPTION_VAL_160_KO },
         { "165%", OPTION_VAL_165_KO },
         { "170%", OPTION_VAL_170_KO },
         { "175%", OPTION_VAL_175_KO },
         { "180%", OPTION_VAL_180_KO },
         { "185%", OPTION_VAL_185_KO },
         { "190%", OPTION_VAL_190_KO },
         { "195%", OPTION_VAL_195_KO },
         { "200%", OPTION_VAL_200_KO },
         { NULL, NULL },
      },
      "100%"
   },
   {
      BEETLE_OPT(negcon_response),
      BEETLE_OPT_NEGCON_RESPONSE_LABEL_KO,
      NULL,
      BEETLE_OPT_NEGCON_RESPONSE_INFO_0_KO,
      NULL,
      "input",
      {
         { "linear",    OPTION_VAL_LINEAR_KO },
         { "quadratic", OPTION_VAL_QUADRATIC_KO },
         { "cubic",     OPTION_VAL_CUBIC_KO },
         { NULL, NULL },
      },
      "linear"
   },
   {
      BEETLE_OPT(negcon_deadzone),
      BEETLE_OPT_NEGCON_DEADZONE_LABEL_KO,
      NULL,
      BEETLE_OPT_NEGCON_DEADZONE_INFO_0_KO,
      NULL,
      "input",
      {
         { "0%",  OPTION_VAL_0_O44_KO },
         { "5%",  OPTION_VAL_5_O42_KO },
         { "10%", OPTION_VAL_10_KO },
         { "15%", OPTION_VAL_15_KO },
         { "20%", OPTION_VAL_20_KO },
         { "25%", OPTION_VAL_25_KO },
         { "30%", OPTION_VAL_30_KO },
         { NULL, NULL },
      },
      "0%"
   },
   {
      BEETLE_OPT(use_mednafen_memcard0_method),
      BEETLE_OPT_USE_MEDNAFEN_MEMCARD0_METHOD_LABEL_KO,
      NULL,
      BEETLE_OPT_USE_MEDNAFEN_MEMCARD0_METHOD_INFO_0_KO,
      NULL,
      "memcards",
      {
         { "libretro", OPTION_VAL_LIBRETRO_KO },
         { "mednafen", OPTION_VAL_MEDNAFEN_KO },
         { NULL, NULL },
      },
      "libretro"
   },
   {
      BEETLE_OPT(enable_memcard1),
      BEETLE_OPT_ENABLE_MEMCARD1_LABEL_KO,
      NULL,
      BEETLE_OPT_ENABLE_MEMCARD1_INFO_0_KO,
      NULL,
      "memcards",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(shared_memory_cards),
      BEETLE_OPT_SHARED_MEMORY_CARDS_LABEL_KO,
      NULL,
      BEETLE_OPT_SHARED_MEMORY_CARDS_INFO_0_KO,
      NULL,
      "memcards",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(memcard_left_index),
      BEETLE_OPT_MEMCARD_LEFT_INDEX_LABEL_KO,
      NULL,
      BEETLE_OPT_MEMCARD_LEFT_INDEX_INFO_0_KO,
      NULL,
      "memcards",
      {
         { "0",  OPTION_VAL_0_O48_KO },
         { "1",  NULL },
         { "2",  NULL },
         { "3",  NULL },
         { "4",  NULL },
         { "5",  NULL },
         { "6",  NULL },
         { "7",  NULL },
         { "8",  NULL },
         { "9",  NULL },
         { "10", NULL },
         { "11", NULL },
         { "12", NULL },
         { "13", NULL },
         { "14", NULL },
         { "15", NULL },
         { "16", NULL },
         { "17", NULL },
         { "18", NULL },
         { "19", NULL },
         { "20", NULL },
         { "21", NULL },
         { "22", NULL },
         { "23", NULL },
         { "24", NULL },
         { "25", NULL },
         { "26", NULL },
         { "27", NULL },
         { "28", NULL },
         { "29", NULL },
         { "30", NULL },
         { "31", NULL },
         { "32", NULL },
         { "33", NULL },
         { "34", NULL },
         { "35", NULL },
         { "36", NULL },
         { "37", NULL },
         { "38", NULL },
         { "39", NULL },
         { "40", NULL },
         { "41", NULL },
         { "42", NULL },
         { "43", NULL },
         { "44", NULL },
         { "45", NULL },
         { "46", NULL },
         { "47", NULL },
         { "48", NULL },
         { "49", NULL },
         { "50", NULL },
         { "51", NULL },
         { "52", NULL },
         { "53", NULL },
         { "54", NULL },
         { "55", NULL },
         { "56", NULL },
         { "57", NULL },
         { "58", NULL },
         { "59", NULL },
         { "60", NULL },
         { "61", NULL },
         { "62", NULL },
         { "63", NULL },
         { NULL, NULL },
      },
      "0"
   },
   {
      BEETLE_OPT(memcard_right_index),
      BEETLE_OPT_MEMCARD_RIGHT_INDEX_LABEL_KO,
      NULL,
      BEETLE_OPT_MEMCARD_RIGHT_INDEX_INFO_0_KO,
      NULL,
      "memcards",
      {
         { "0",  NULL },
         { "1",  OPTION_VAL_1_O49_KO },
         { "2",  NULL },
         { "3",  NULL },
         { "4",  NULL },
         { "5",  NULL },
         { "6",  NULL },
         { "7",  NULL },
         { "8",  NULL },
         { "9",  NULL },
         { "10", NULL },
         { "11", NULL },
         { "12", NULL },
         { "13", NULL },
         { "14", NULL },
         { "15", NULL },
         { "16", NULL },
         { "17", NULL },
         { "18", NULL },
         { "19", NULL },
         { "20", NULL },
         { "21", NULL },
         { "22", NULL },
         { "23", NULL },
         { "24", NULL },
         { "25", NULL },
         { "26", NULL },
         { "27", NULL },
         { "28", NULL },
         { "29", NULL },
         { "30", NULL },
         { "31", NULL },
         { "32", NULL },
         { "33", NULL },
         { "34", NULL },
         { "35", NULL },
         { "36", NULL },
         { "37", NULL },
         { "38", NULL },
         { "39", NULL },
         { "40", NULL },
         { "41", NULL },
         { "42", NULL },
         { "43", NULL },
         { "44", NULL },
         { "45", NULL },
         { "46", NULL },
         { "47", NULL },
         { "48", NULL },
         { "49", NULL },
         { "50", NULL },
         { "51", NULL },
         { "52", NULL },
         { "53", NULL },
         { "54", NULL },
         { "55", NULL },
         { "56", NULL },
         { "57", NULL },
         { "58", NULL },
         { "59", NULL },
         { "60", NULL },
         { "61", NULL },
         { "62", NULL },
         { "63", NULL },
         { NULL, NULL },
      },
      "1"
   },
   {
      BEETLE_OPT(line_render),
      BEETLE_OPT_LINE_RENDER_LABEL_KO,
      NULL,
      BEETLE_OPT_LINE_RENDER_INFO_0_KO,
      NULL,
      "hacks",
      {
         { "default",    OPTION_VAL_DEFAULT_KO },
         { "aggressive", OPTION_VAL_AGGRESSIVE_KO },
         { "disabled",   NULL },
         { NULL, NULL },
      },
      "default"
   },
   {
      BEETLE_OPT(widescreen_hack),
      BEETLE_OPT_WIDESCREEN_HACK_LABEL_KO,
      NULL,
      BEETLE_OPT_WIDESCREEN_HACK_INFO_0_KO,
      NULL,
      "hacks",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(widescreen_hack_aspect_ratio),
      BEETLE_OPT_WIDESCREEN_HACK_ASPECT_RATIO_LABEL_KO,
      NULL,
      BEETLE_OPT_WIDESCREEN_HACK_ASPECT_RATIO_INFO_0_KO,
      NULL,
      "hacks",
      {
         { "16:9",  OPTION_VAL_16_9_KO },
         { "16:10", OPTION_VAL_16_10_KO },
         { "18:9",  OPTION_VAL_18_9_KO },
         { "19:9",  OPTION_VAL_19_9_KO },
         { "20:9",  OPTION_VAL_20_9_KO },
         { "21:9",  OPTION_VAL_21_9_KO }, // 64:27
         { "32:9",  OPTION_VAL_32_9_KO },
         { NULL,    NULL },
      },
      "16:9"
   },
#if defined(HAVE_LIGHTREC)
   {
      BEETLE_OPT(cpu_dynarec),
      BEETLE_OPT_CPU_DYNAREC_LABEL_KO,
      NULL,
      BEETLE_OPT_CPU_DYNAREC_INFO_0_KO,
      NULL,
      "hacks",
      {
         { "disabled",        OPTION_VAL_DISABLED_KO },
         { "execute",         OPTION_VAL_EXECUTE_KO },
         { "execute_one",     OPTION_VAL_EXECUTE_ONE_KO },
         { "run_interpreter", OPTION_VAL_RUN_INTERPRETER_KO },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(dynarec_invalidate),
      BEETLE_OPT_DYNAREC_INVALIDATE_LABEL_KO,
      NULL,
      BEETLE_OPT_DYNAREC_INVALIDATE_INFO_0_KO,
      NULL,
      "hacks",
      {
         { "full", OPTION_VAL_FULL_KO },
         { "dma",  OPTION_VAL_DMA_KO },
         { NULL, NULL },
      },
      "full"
   },
   {
      BEETLE_OPT(dynarec_eventcycles),
      BEETLE_OPT_DYNAREC_EVENTCYCLES_LABEL_KO,
      NULL,
      BEETLE_OPT_DYNAREC_EVENTCYCLES_INFO_0_KO,
      NULL,
      "hacks",
      {
         { "128",  OPTION_VAL_128_KO },
         { "256",  NULL },
         { "384",  NULL },
         { "512",  NULL },
         { "640",  NULL },
         { "768",  NULL },
         { "896",  NULL },
         { "1024", NULL },
         { "1152", NULL },
         { "1280", NULL },
         { "1408", NULL },
         { "1536", NULL },
         { "1664", NULL },
         { "1792", NULL },
         { "1920", NULL },
         { "2048", NULL },
         { NULL, NULL },
      },
      "128"
   },
   {
      BEETLE_OPT(dynarec_spu_samples),
      BEETLE_OPT_DYNAREC_SPU_SAMPLES_LABEL_KO,
      NULL,
      BEETLE_OPT_DYNAREC_SPU_SAMPLES_INFO_0_KO,
      NULL,
      "hacks",
      {
         { "1",  OPTION_VAL_1_O49_KO },
         { "4",  NULL },
         { "16", NULL },
         { NULL, NULL },
      },
      "1"
   },
#endif
   {
      BEETLE_OPT(core_timing_fps),
      BEETLE_OPT_CORE_TIMING_FPS_LABEL_KO,
      NULL,
      BEETLE_OPT_CORE_TIMING_FPS_INFO_0_KO,
      NULL,
      "video",
      {
         { "auto_toggle",       OPTION_VAL_AUTO_TOGGLE_KO },
         { "force_progressive", OPTION_VAL_FORCE_PROGRESSIVE_KO },
         { "force_interlaced",  OPTION_VAL_FORCE_INTERLACED_KO },
      },
      "auto_toggle"
   },
   {
      BEETLE_OPT(deinterlacer),
      BEETLE_OPT_DEINTERLACER_LABEL_KO,
      NULL,
      BEETLE_OPT_DEINTERLACER_INFO_0_KO,
      NULL,
      "video",
      {
         { "weave",      OPTION_VAL_WEAVE_KO },
         { "bob",        OPTION_VAL_BOB_KO },
         { "bob_offset", OPTION_VAL_BOB_OFFSET_KO },
         { "fastmad",    OPTION_VAL_FASTMAD_KO },
         { "off",        "Off" },
         { NULL, NULL },
      },
      "weave"
   },
   {
      BEETLE_OPT(frame_duping),
      BEETLE_OPT_FRAME_DUPING_LABEL_KO,
      NULL,
      BEETLE_OPT_FRAME_DUPING_INFO_0_KO,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(pal_video_timing_override),
      BEETLE_OPT_PAL_VIDEO_TIMING_OVERRIDE_LABEL_KO,
      NULL,
      BEETLE_OPT_PAL_VIDEO_TIMING_OVERRIDE_INFO_0_KO,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(aspect_ratio),
      BEETLE_OPT_ASPECT_RATIO_LABEL_KO,
      NULL,
      BEETLE_OPT_ASPECT_RATIO_INFO_0_KO,
      NULL,
      "video",
      {
         { "corrected", OPTION_VAL_CORRECTED_KO },
         { "uncorrected", OPTION_VAL_UNCORRECTED_KO },
         { "4:3",  OPTION_VAL_4_3_KO },
         { "ntsc", OPTION_VAL_NTSC_KO },
      },
      "corrected"
   },
   {
      BEETLE_OPT(crop_overscan),
      BEETLE_OPT_CROP_OVERSCAN_LABEL_KO,
      NULL,
      BEETLE_OPT_CROP_OVERSCAN_INFO_0_KO,
      NULL,
      "video",
      {
         { "disabled", "Disabled" },
         { "static",   OPTION_VAL_STATIC_KO },
         { "smart",    OPTION_VAL_SMART_KO },
         { NULL, NULL },
      },
      "smart"
   },
   {
      BEETLE_OPT(image_crop),
      BEETLE_OPT_IMAGE_CROP_LABEL_KO,
      NULL,
      BEETLE_OPT_IMAGE_CROP_INFO_0_KO,
      NULL,
      "video",
      {
         { "disabled", "0" },
         { "1px",      OPTION_VAL_1PX_KO },
         { "2px",      OPTION_VAL_2PX_KO },
         { "3px",      OPTION_VAL_3PX_KO },
         { "4px",      OPTION_VAL_4PX_KO },
         { "5px",      OPTION_VAL_5PX_KO },
         { "6px",      OPTION_VAL_6PX_KO },
         { "7px",      OPTION_VAL_7PX_KO },
         { "8px",      OPTION_VAL_8PX_KO },
         { "9px",      OPTION_VAL_9PX_KO },
         { "10px",     OPTION_VAL_10PX_KO },
         { "11px",     OPTION_VAL_11PX_KO },
         { "12px",     OPTION_VAL_12PX_KO },
         { "13px",     OPTION_VAL_13PX_KO },
         { "14px",     OPTION_VAL_14PX_KO },
         { "15px",     OPTION_VAL_15PX_KO },
         { "16px",     OPTION_VAL_16PX_KO },
         { "17px",     OPTION_VAL_17PX_KO },
         { "18px",     OPTION_VAL_18PX_KO },
         { "19px",     OPTION_VAL_19PX_KO },
         { "20px",     OPTION_VAL_20PX_KO },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(image_offset),
      BEETLE_OPT_IMAGE_OFFSET_LABEL_KO,
      NULL,
      BEETLE_OPT_IMAGE_OFFSET_INFO_0_KO,
      NULL,
      "video",
      {
         { "-12px",    OPTION_VAL_12PX_O64_KO },
         { "-11px",    OPTION_VAL_11PX_O64_KO },
         { "-10px",    OPTION_VAL_10PX_O64_KO },
         { "-9px",     OPTION_VAL_9PX_O64_KO },
         { "-8px",     OPTION_VAL_8PX_O64_KO },
         { "-7px",     OPTION_VAL_7PX_O64_KO },
         { "-6px",     OPTION_VAL_6PX_O64_KO },
         { "-5px",     OPTION_VAL_5PX_O64_KO },
         { "-4px",     OPTION_VAL_4PX_O64_KO },
         { "-3px",     OPTION_VAL_3PX_O64_KO },
         { "-2px",     OPTION_VAL_2PX_O64_KO },
         { "-1px",     OPTION_VAL_1PX_O64_KO },
         { "disabled", "0" },
         { "+1px",     OPTION_VAL_1PX_O64_0_KO },
         { "+2px",     OPTION_VAL_2PX_O64_0_KO },
         { "+3px",     OPTION_VAL_3PX_O64_0_KO },
         { "+4px",     OPTION_VAL_4PX_O64_0_KO },
         { "+5px",     OPTION_VAL_5PX_O64_0_KO },
         { "+6px",     OPTION_VAL_6PX_O64_0_KO },
         { "+7px",     OPTION_VAL_7PX_O64_0_KO },
         { "+8px",     OPTION_VAL_8PX_O64_0_KO },
         { "+9px",     OPTION_VAL_9PX_O64_0_KO },
         { "+10px",    OPTION_VAL_10PX_O64_0_KO },
         { "+11px",    OPTION_VAL_11PX_O64_0_KO },
         { "+12px",    OPTION_VAL_12PX_O64_0_KO },
         { NULL, NULL },
      },
      "disabled"
   },
#if defined(HAVE_OPENGL) || defined(HAVE_OPENGLES) || defined(HAVE_VULKAN)
   {
      BEETLE_OPT(image_offset_cycles),
      BEETLE_OPT_IMAGE_OFFSET_CYCLES_LABEL_KO,
      NULL,
      BEETLE_OPT_IMAGE_OFFSET_CYCLES_INFO_0_KO,
      NULL,
      "video",
      {
         { "-40",      NULL },
         { "-39",      NULL },
         { "-38",      NULL },
         { "-37",      NULL },
         { "-36",      NULL },
         { "-35",      NULL },
         { "-34",      NULL },
         { "-33",      NULL },
         { "-32",      NULL },
         { "-31",      NULL },
         { "-30",      NULL },
         { "-29",      NULL },
         { "-28",      NULL },
         { "-27",      NULL },
         { "-26",      NULL },
         { "-25",      NULL },
         { "-24",      NULL },
         { "-23",      NULL },
         { "-22",      NULL },
         { "-21",      NULL },
         { "-20",      NULL },
         { "-19",      NULL },
         { "-18",      NULL },
         { "-17",      NULL },
         { "-16",      NULL },
         { "-15",      NULL },
         { "-14",      NULL },
         { "-13",      NULL },
         { "-12",      NULL },
         { "-11",      NULL },
         { "-10",      NULL },
         { "-9",       NULL },
         { "-8",       NULL },
         { "-7",       NULL },
         { "-6",       NULL },
         { "-5",       NULL },
         { "-4",       NULL },
         { "-3",       NULL },
         { "-2",       NULL },
         { "-1",       NULL },
         { "0",        "0" },
         { "+1",       NULL },
         { "+2",       NULL },
         { "+3",       NULL },
         { "+4",       NULL },
         { "+5",       NULL },
         { "+6",       NULL },
         { "+7",       NULL },
         { "+8",       NULL },
         { "+9",       NULL },
         { "+10",      NULL },
         { "+11",      NULL },
         { "+12",      NULL },
         { "+13",      NULL },
         { "+14",      NULL },
         { "+15",      NULL },
         { "+16",      NULL },
         { "+17",      NULL },
         { "+18",      NULL },
         { "+19",      NULL },
         { "+20",      NULL },
         { "+21",      NULL },
         { "+22",      NULL },
         { "+23",      NULL },
         { "+24",      NULL },
         { "+25",      NULL },
         { "+26",      NULL },
         { "+27",      NULL },
         { "+28",      NULL },
         { "+29",      NULL },
         { "+30",      NULL },
         { "+31",      NULL },
         { "+32",      NULL },
         { "+33",      NULL },
         { "+34",      NULL },
         { "+35",      NULL },
         { "+36",      NULL },
         { "+37",      NULL },
         { "+38",      NULL },
         { "+39",      NULL },
         { "+40",      NULL },
         { NULL, NULL},
      },
      "0"
   },
#endif
   {
      BEETLE_OPT(initial_scanline),
      BEETLE_OPT_INITIAL_SCANLINE_LABEL_KO,
      NULL,
      BEETLE_OPT_INITIAL_SCANLINE_INFO_0_KO,
      NULL,
      "video",
      {
         { "0",  OPTION_VAL_0_O48_KO },
         { "1",  NULL },
         { "2",  NULL },
         { "3",  NULL },
         { "4",  NULL },
         { "5",  NULL },
         { "6",  NULL },
         { "7",  NULL },
         { "8",  NULL },
         { "9",  NULL },
         { "10", NULL },
         { "11", NULL },
         { "12", NULL },
         { "13", NULL },
         { "14", NULL },
         { "15", NULL },
         { "16", NULL },
         { "17", NULL },
         { "18", NULL },
         { "19", NULL },
         { "20", NULL },
         { "21", NULL },
         { "22", NULL },
         { "23", NULL },
         { "24", NULL },
         { "25", NULL },
         { "26", NULL },
         { "27", NULL },
         { "28", NULL },
         { "29", NULL },
         { "30", NULL },
         { "31", NULL },
         { "32", NULL },
         { "33", NULL },
         { "34", NULL },
         { "35", NULL },
         { "36", NULL },
         { "37", NULL },
         { "38", NULL },
         { "39", NULL },
         { "40", NULL },
         { NULL, NULL },
      },
      "0"
   },
   {
      BEETLE_OPT(last_scanline),
      BEETLE_OPT_LAST_SCANLINE_LABEL_KO,
      NULL,
      BEETLE_OPT_LAST_SCANLINE_INFO_0_KO,
      NULL,
      "video",
      {
         { "210", NULL },
         { "211", NULL },
         { "212", NULL },
         { "213", NULL },
         { "214", NULL },
         { "215", NULL },
         { "216", NULL },
         { "217", NULL },
         { "218", NULL },
         { "219", NULL },
         { "220", NULL },
         { "221", NULL },
         { "222", NULL },
         { "223", NULL },
         { "224", NULL },
         { "225", NULL },
         { "226", NULL },
         { "227", NULL },
         { "228", NULL },
         { "229", NULL },
         { "230", NULL },
         { "231", NULL },
         { "232", NULL },
         { "233", NULL },
         { "234", NULL },
         { "235", NULL },
         { "236", NULL },
         { "237", NULL },
         { "238", NULL },
         { "239", OPTION_VAL_239_KO },
         { NULL, NULL },
      },
      "239"
   },
   {
      BEETLE_OPT(initial_scanline_pal),
      BEETLE_OPT_INITIAL_SCANLINE_PAL_LABEL_KO,
      NULL,
      BEETLE_OPT_INITIAL_SCANLINE_PAL_INFO_0_KO,
      NULL,
      "video",
      {
         { "0",  OPTION_VAL_0_O48_KO },
         { "1",  NULL },
         { "2",  NULL },
         { "3",  NULL },
         { "4",  NULL },
         { "5",  NULL },
         { "6",  NULL },
         { "7",  NULL },
         { "8",  NULL },
         { "9",  NULL },
         { "10", NULL },
         { "11", NULL },
         { "12", NULL },
         { "13", NULL },
         { "14", NULL },
         { "15", NULL },
         { "16", NULL },
         { "17", NULL },
         { "18", NULL },
         { "19", NULL },
         { "20", NULL },
         { "21", NULL },
         { "22", NULL },
         { "23", NULL },
         { "24", NULL },
         { "25", NULL },
         { "26", NULL },
         { "27", NULL },
         { "28", NULL },
         { "29", NULL },
         { "30", NULL },
         { "31", NULL },
         { "32", NULL },
         { "33", NULL },
         { "34", NULL },
         { "35", NULL },
         { "36", NULL },
         { "37", NULL },
         { "38", NULL },
         { "39", NULL },
         { "40", NULL },
         { NULL, NULL },
      },
      "0"
   },
   {
      BEETLE_OPT(last_scanline_pal),
      BEETLE_OPT_LAST_SCANLINE_PAL_LABEL_KO,
      NULL,
      BEETLE_OPT_LAST_SCANLINE_PAL_INFO_0_KO,
      NULL,
      "video",
      {
         { "230", NULL },
         { "231", NULL },
         { "232", NULL },
         { "233", NULL },
         { "234", NULL },
         { "235", NULL },
         { "236", NULL },
         { "237", NULL },
         { "238", NULL },
         { "239", NULL },
         { "240", NULL },
         { "241", NULL },
         { "242", NULL },
         { "243", NULL },
         { "244", NULL },
         { "245", NULL },
         { "246", NULL },
         { "247", NULL },
         { "248", NULL },
         { "249", NULL },
         { "250", NULL },
         { "251", NULL },
         { "252", NULL },
         { "253", NULL },
         { "254", NULL },
         { "255", NULL },
         { "256", NULL },
         { "257", NULL },
         { "258", NULL },
         { "259", NULL },
         { "260", NULL },
         { "261", NULL },
         { "262", NULL },
         { "263", NULL },
         { "264", NULL },
         { "265", NULL },
         { "266", NULL },
         { "267", NULL },
         { "268", NULL },
         { "269", NULL },
         { "270", NULL },
         { "271", NULL },
         { "272", NULL },
         { "273", NULL },
         { "274", NULL },
         { "275", NULL },
         { "276", NULL },
         { "277", NULL },
         { "278", NULL },
         { "279", NULL },
         { "280", NULL },
         { "281", NULL },
         { "282", NULL },
         { "283", NULL },
         { "284", NULL },
         { "285", NULL },
         { "286", NULL },
         { "287", OPTION_VAL_287_KO },
         { NULL, NULL },
      },
      "287"
   },
   { NULL, NULL, NULL, NULL, NULL, NULL, {{0}}, NULL },
};
struct retro_core_options_v2 options_ko = {
   option_cats_ko,
   option_defs_ko
};

/* RETRO_LANGUAGE_NL */

#define CATEGORY_SYSTEM_LABEL_NL "Systeem"
#define CATEGORY_SYSTEM_INFO_0_NL NULL
#define CATEGORY_VIDEO_LABEL_NL NULL
#define CATEGORY_VIDEO_INFO_0_NL NULL
#define CATEGORY_PGXP_LABEL_NL NULL
#define CATEGORY_PGXP_INFO_0_NL NULL
#define CATEGORY_OSD_LABEL_NL "Onscreen Weergave"
#define CATEGORY_OSD_INFO_0_NL NULL
#define CATEGORY_INPUT_LABEL_NL "Invoer"
#define CATEGORY_INPUT_INFO_0_NL NULL
#define CATEGORY_MEMCARDS_LABEL_NL NULL
#define CATEGORY_MEMCARDS_INFO_0_NL NULL
#define CATEGORY_HACKS_LABEL_NL NULL
#define CATEGORY_HACKS_INFO_0_NL NULL
#define BEETLE_OPT_CPU_FREQ_SCALE_LABEL_NL NULL
#define BEETLE_OPT_CPU_FREQ_SCALE_INFO_0_NL NULL
#define OPTION_VAL_50_NL NULL
#define OPTION_VAL_60_NL NULL
#define OPTION_VAL_70_NL NULL
#define OPTION_VAL_80_NL NULL
#define OPTION_VAL_90_NL NULL
#define OPTION_VAL_100_NL NULL
#define OPTION_VAL_110_NL NULL
#define OPTION_VAL_120_NL NULL
#define OPTION_VAL_130_NL NULL
#define OPTION_VAL_140_NL NULL
#define OPTION_VAL_150_NL NULL
#define OPTION_VAL_160_NL NULL
#define OPTION_VAL_170_NL NULL
#define OPTION_VAL_180_NL NULL
#define OPTION_VAL_190_NL NULL
#define OPTION_VAL_200_NL NULL
#define OPTION_VAL_210_NL NULL
#define OPTION_VAL_220_NL NULL
#define OPTION_VAL_230_NL NULL
#define OPTION_VAL_240_NL NULL
#define OPTION_VAL_250_NL NULL
#define OPTION_VAL_260_NL NULL
#define OPTION_VAL_270_NL NULL
#define OPTION_VAL_280_NL NULL
#define OPTION_VAL_290_NL NULL
#define OPTION_VAL_300_NL NULL
#define OPTION_VAL_310_NL NULL
#define OPTION_VAL_320_NL NULL
#define OPTION_VAL_330_NL NULL
#define OPTION_VAL_340_NL NULL
#define OPTION_VAL_350_NL NULL
#define OPTION_VAL_360_NL NULL
#define OPTION_VAL_370_NL NULL
#define OPTION_VAL_380_NL NULL
#define OPTION_VAL_390_NL NULL
#define OPTION_VAL_400_NL NULL
#define OPTION_VAL_410_NL NULL
#define OPTION_VAL_420_NL NULL
#define OPTION_VAL_430_NL NULL
#define OPTION_VAL_440_NL NULL
#define OPTION_VAL_450_NL NULL
#define OPTION_VAL_460_NL NULL
#define OPTION_VAL_470_NL NULL
#define OPTION_VAL_480_NL NULL
#define OPTION_VAL_490_NL NULL
#define OPTION_VAL_500_NL NULL
#define OPTION_VAL_510_NL NULL
#define OPTION_VAL_520_NL NULL
#define OPTION_VAL_530_NL NULL
#define OPTION_VAL_540_NL NULL
#define OPTION_VAL_550_NL NULL
#define OPTION_VAL_560_NL NULL
#define OPTION_VAL_570_NL NULL
#define OPTION_VAL_580_NL NULL
#define OPTION_VAL_590_NL NULL
#define OPTION_VAL_600_NL NULL
#define OPTION_VAL_610_NL NULL
#define OPTION_VAL_620_NL NULL
#define OPTION_VAL_630_NL NULL
#define OPTION_VAL_640_NL NULL
#define OPTION_VAL_650_NL NULL
#define OPTION_VAL_660_NL NULL
#define OPTION_VAL_670_NL NULL
#define OPTION_VAL_680_NL NULL
#define OPTION_VAL_690_NL NULL
#define OPTION_VAL_700_NL NULL
#define OPTION_VAL_710_NL NULL
#define OPTION_VAL_720_NL NULL
#define OPTION_VAL_730_NL NULL
#define OPTION_VAL_740_NL NULL
#define OPTION_VAL_750_NL NULL
#define BEETLE_OPT_GPU_OVERCLOCK_LABEL_NL NULL
#define BEETLE_OPT_GPU_OVERCLOCK_INFO_0_NL NULL
#define OPTION_VAL_1X_NATIVE_NL NULL
#define OPTION_VAL_2X_NL NULL
#define OPTION_VAL_4X_NL NULL
#define OPTION_VAL_8X_NL NULL
#define OPTION_VAL_16X_NL NULL
#define OPTION_VAL_32X_NL NULL
#define BEETLE_OPT_GTE_OVERCLOCK_LABEL_NL NULL
#define BEETLE_OPT_GTE_OVERCLOCK_INFO_0_NL "Verlaag alle geëmuleerde GTE-bewerkingen (CPU-coprocessor voor 3D-graphics) tot een constante latentie van één cyclus. Voor games die intensief gebruik maken van de GTE, kan dit de framesnelheid en frametijdstabiliteit aanzienlijk verbeteren."
#define BEETLE_OPT_CD_ACCESS_METHOD_LABEL_NL NULL
#define BEETLE_OPT_CD_ACCESS_METHOD_INFO_0_NL NULL
#define OPTION_VAL_SYNC_NL NULL
#define OPTION_VAL_ASYNC_NL NULL
#define OPTION_VAL_PRECACHE_NL NULL
#define BEETLE_OPT_CD_FASTLOAD_LABEL_NL NULL
#define BEETLE_OPT_CD_FASTLOAD_INFO_0_NL NULL
#define OPTION_VAL_2X_NATIVE_NL NULL
#define OPTION_VAL_6X_NL NULL
#define OPTION_VAL_10X_NL NULL
#define OPTION_VAL_12X_NL NULL
#define OPTION_VAL_14X_NL NULL
#define BEETLE_OPT_SKIP_BIOS_LABEL_NL NULL
#define BEETLE_OPT_SKIP_BIOS_INFO_0_NL NULL
#define BEETLE_OPT_OVERRIDE_BIOS_LABEL_NL NULL
#define BEETLE_OPT_OVERRIDE_BIOS_INFO_0_NL NULL
#define OPTION_VAL_PSXONPSP_NL NULL
#define OPTION_VAL_PS1_ROM_NL NULL
#define OPTION_VAL_OPENBIOS_NL NULL
#define BEETLE_OPT_REGION_LABEL_NL NULL
#define BEETLE_OPT_REGION_INFO_0_NL NULL
#define OPTION_VAL_AUTO_NL "Automatisch"
#define OPTION_VAL_NTSC_J_NL NULL
#define OPTION_VAL_NTSC_U_NL NULL
#define OPTION_VAL_PAL_NL NULL
#define BEETLE_OPT_INTERNAL_RESOLUTION_LABEL_NL NULL
#define BEETLE_OPT_INTERNAL_RESOLUTION_INFO_0_NL NULL
#define BEETLE_OPT_RENDERER_LABEL_NL NULL
#define BEETLE_OPT_RENDERER_INFO_0_NL NULL
#define OPTION_VAL_HARDWARE_NL NULL
#define OPTION_VAL_HARDWARE_GL_NL NULL
#define OPTION_VAL_HARDWARE_VK_NL NULL
#define OPTION_VAL_SOFTWARE_NL NULL
#define BEETLE_OPT_RENDERER_SOFTWARE_FB_LABEL_NL NULL
#define BEETLE_OPT_RENDERER_SOFTWARE_FB_INFO_0_NL NULL
#define BEETLE_OPT_DEPTH_LABEL_NL NULL
#define BEETLE_OPT_DEPTH_INFO_0_NL NULL
#define OPTION_VAL_16BPP_NATIVE_NL NULL
#define OPTION_VAL_32BPP_NL NULL
#define BEETLE_OPT_DITHER_MODE_LABEL_NL NULL
#define BEETLE_OPT_DITHER_MODE_INFO_0_NL NULL
#define OPTION_VAL_INTERNAL_RESOLUTION_NL NULL
#define BEETLE_OPT_SCALED_UV_OFFSET_LABEL_NL NULL
#define BEETLE_OPT_SCALED_UV_OFFSET_INFO_0_NL NULL
#define BEETLE_OPT_FILTER_LABEL_NL NULL
#define BEETLE_OPT_FILTER_INFO_0_NL NULL
#define OPTION_VAL_NEAREST_NL "Naaste"
#define OPTION_VAL_SABR_NL NULL
#define OPTION_VAL_XBR_NL NULL
#define OPTION_VAL_BILINEAR_NL "Bilineair"
#define OPTION_VAL_3_POINT_NL NULL
#define OPTION_VAL_JINC2_NL NULL
#define BEETLE_OPT_FILTER_EXCLUDE_SPRITE_LABEL_NL NULL
#define BEETLE_OPT_FILTER_EXCLUDE_SPRITE_INFO_0_NL NULL
#define OPTION_VAL_OPAQUE_NL NULL
#define OPTION_VAL_ALL_NL NULL
#define BEETLE_OPT_FILTER_EXCLUDE_2D_POLYGON_LABEL_NL NULL
#define BEETLE_OPT_FILTER_EXCLUDE_2D_POLYGON_INFO_0_NL NULL
#define BEETLE_OPT_ADAPTIVE_SMOOTHING_LABEL_NL NULL
#define BEETLE_OPT_ADAPTIVE_SMOOTHING_INFO_0_NL NULL
#define BEETLE_OPT_SUPER_SAMPLING_LABEL_NL NULL
#define BEETLE_OPT_SUPER_SAMPLING_INFO_0_NL NULL
#define BEETLE_OPT_MSAA_LABEL_NL NULL
#define BEETLE_OPT_MSAA_INFO_0_NL NULL
#define OPTION_VAL_1X_NL NULL
#define BEETLE_OPT_MDEC_YUV_LABEL_NL NULL
#define BEETLE_OPT_MDEC_YUV_INFO_0_NL NULL
#define BEETLE_OPT_TRACK_TEXTURES_LABEL_NL NULL
#define BEETLE_OPT_TRACK_TEXTURES_INFO_0_NL NULL
#define BEETLE_OPT_DUMP_TEXTURES_LABEL_NL NULL
#define BEETLE_OPT_DUMP_TEXTURES_INFO_0_NL NULL
#define BEETLE_OPT_REPLACE_TEXTURES_LABEL_NL NULL
#define BEETLE_OPT_REPLACE_TEXTURES_INFO_0_NL NULL
#define BEETLE_OPT_PGXP_MODE_LABEL_NL NULL
#define BEETLE_OPT_PGXP_MODE_INFO_0_NL NULL
#define OPTION_VAL_MEMORY_ONLY_NL NULL
#define OPTION_VAL_MEMORY_CPU_NL NULL
#define BEETLE_OPT_PGXP_2D_TOL_LABEL_NL NULL
#define BEETLE_OPT_PGXP_2D_TOL_INFO_0_NL NULL
#define OPTION_VAL_0PX_NL NULL
#define OPTION_VAL_1PX_NL NULL
#define OPTION_VAL_2PX_NL NULL
#define OPTION_VAL_3PX_NL NULL
#define OPTION_VAL_4PX_NL NULL
#define OPTION_VAL_5PX_NL NULL
#define OPTION_VAL_6PX_NL NULL
#define OPTION_VAL_7PX_NL NULL
#define OPTION_VAL_8PX_NL NULL
#define BEETLE_OPT_PGXP_NCLIP_LABEL_NL NULL
#define BEETLE_OPT_PGXP_NCLIP_INFO_0_NL NULL
#define BEETLE_OPT_PGXP_VERTEX_LABEL_NL NULL
#define BEETLE_OPT_PGXP_VERTEX_INFO_0_NL NULL
#define BEETLE_OPT_PGXP_TEXTURE_LABEL_NL NULL
#define BEETLE_OPT_PGXP_TEXTURE_INFO_0_NL NULL
#define BEETLE_OPT_DISPLAY_INTERNAL_FPS_LABEL_NL NULL
#define BEETLE_OPT_DISPLAY_INTERNAL_FPS_INFO_0_NL NULL
#define BEETLE_OPT_DISPLAY_OSD_LABEL_NL NULL
#define BEETLE_OPT_DISPLAY_OSD_INFO_0_NL NULL
#define BEETLE_OPT_DISPLAY_VRAM_LABEL_NL NULL
#define BEETLE_OPT_DISPLAY_VRAM_INFO_0_NL NULL
#define BEETLE_OPT_ANALOG_CALIBRATION_LABEL_NL NULL
#define BEETLE_OPT_ANALOG_CALIBRATION_INFO_0_NL NULL
#define BEETLE_OPT_ANALOG_TOGGLE_LABEL_NL NULL
#define BEETLE_OPT_ANALOG_TOGGLE_INFO_0_NL NULL
#define OPTION_VAL_ENABLED_ANALOG_NL NULL
#define BEETLE_OPT_ANALOG_TOGGLE_COMBO_LABEL_NL NULL
#define BEETLE_OPT_ANALOG_TOGGLE_COMBO_INFO_0_NL NULL
#define OPTION_VAL_L1_L2_R1_R2_START_SELECT_NL NULL
#define OPTION_VAL_L1_R1_SELECT_NL NULL
#define OPTION_VAL_L1_R1_START_NL NULL
#define OPTION_VAL_L1_R1_L3_NL NULL
#define OPTION_VAL_L1_R1_R3_NL NULL
#define OPTION_VAL_L2_R2_SELECT_NL NULL
#define OPTION_VAL_L2_R2_START_NL NULL
#define OPTION_VAL_L2_R2_L3_NL NULL
#define OPTION_VAL_L2_R2_R3_NL NULL
#define OPTION_VAL_L3_R3_NL NULL
#define BEETLE_OPT_ANALOG_TOGGLE_HOLD_LABEL_NL NULL
#define BEETLE_OPT_ANALOG_TOGGLE_HOLD_INFO_0_NL NULL
#define OPTION_VAL_0_NL NULL
#define OPTION_VAL_1_NL NULL
#define OPTION_VAL_2_NL NULL
#define OPTION_VAL_3_NL NULL
#define OPTION_VAL_4_NL NULL
#define OPTION_VAL_5_NL NULL
#define BEETLE_OPT_ENABLE_MULTITAP_PORT1_LABEL_NL NULL
#define BEETLE_OPT_ENABLE_MULTITAP_PORT1_INFO_0_NL NULL
#define BEETLE_OPT_ENABLE_MULTITAP_PORT2_LABEL_NL NULL
#define BEETLE_OPT_ENABLE_MULTITAP_PORT2_INFO_0_NL NULL
#define BEETLE_OPT_GUN_INPUT_MODE_LABEL_NL NULL
#define BEETLE_OPT_GUN_INPUT_MODE_INFO_0_NL NULL
#define OPTION_VAL_LIGHTGUN_NL NULL
#define OPTION_VAL_TOUCHSCREEN_NL NULL
#define BEETLE_OPT_GUN_CURSOR_LABEL_NL NULL
#define BEETLE_OPT_GUN_CURSOR_INFO_0_NL NULL
#define OPTION_VAL_CROSS_NL NULL
#define OPTION_VAL_DOT_NL NULL
#define BEETLE_OPT_CROSSHAIR_COLOR_P1_LABEL_NL NULL
#define BEETLE_OPT_CROSSHAIR_COLOR_P1_INFO_0_NL NULL
#define OPTION_VAL_RED_NL "Rood"
#define OPTION_VAL_BLUE_NL "Blauw"
#define OPTION_VAL_GREEN_NL "Groen"
#define OPTION_VAL_ORANGE_NL NULL
#define OPTION_VAL_YELLOW_NL "Geel"
#define OPTION_VAL_CYAN_NL NULL
#define OPTION_VAL_PINK_NL NULL
#define OPTION_VAL_PURPLE_NL NULL
#define OPTION_VAL_BLACK_NL NULL
#define OPTION_VAL_WHITE_NL NULL
#define BEETLE_OPT_CROSSHAIR_COLOR_P2_LABEL_NL NULL
#define BEETLE_OPT_CROSSHAIR_COLOR_P2_INFO_0_NL NULL
#define BEETLE_OPT_MOUSE_SENSITIVITY_LABEL_NL NULL
#define BEETLE_OPT_MOUSE_SENSITIVITY_INFO_0_NL NULL
#define OPTION_VAL_5_O42_NL NULL
#define OPTION_VAL_10_NL NULL
#define OPTION_VAL_15_NL NULL
#define OPTION_VAL_20_NL NULL
#define OPTION_VAL_25_NL NULL
#define OPTION_VAL_30_NL NULL
#define OPTION_VAL_35_NL NULL
#define OPTION_VAL_40_NL NULL
#define OPTION_VAL_45_NL NULL
#define OPTION_VAL_55_NL NULL
#define OPTION_VAL_65_NL NULL
#define OPTION_VAL_75_NL NULL
#define OPTION_VAL_85_NL NULL
#define OPTION_VAL_95_NL NULL
#define OPTION_VAL_100_O42_NL NULL
#define OPTION_VAL_105_NL NULL
#define OPTION_VAL_115_NL NULL
#define OPTION_VAL_125_NL NULL
#define OPTION_VAL_135_NL NULL
#define OPTION_VAL_145_NL NULL
#define OPTION_VAL_155_NL NULL
#define OPTION_VAL_165_NL NULL
#define OPTION_VAL_175_NL NULL
#define OPTION_VAL_185_NL NULL
#define OPTION_VAL_195_NL NULL
#define BEETLE_OPT_NEGCON_RESPONSE_LABEL_NL NULL
#define BEETLE_OPT_NEGCON_RESPONSE_INFO_0_NL NULL
#define OPTION_VAL_LINEAR_NL "Lineair"
#define OPTION_VAL_QUADRATIC_NL NULL
#define OPTION_VAL_CUBIC_NL NULL
#define BEETLE_OPT_NEGCON_DEADZONE_LABEL_NL NULL
#define BEETLE_OPT_NEGCON_DEADZONE_INFO_0_NL NULL
#define OPTION_VAL_0_O44_NL NULL
#define BEETLE_OPT_USE_MEDNAFEN_MEMCARD0_METHOD_LABEL_NL NULL
#define BEETLE_OPT_USE_MEDNAFEN_MEMCARD0_METHOD_INFO_0_NL NULL
#define OPTION_VAL_LIBRETRO_NL NULL
#define OPTION_VAL_MEDNAFEN_NL NULL
#define BEETLE_OPT_ENABLE_MEMCARD1_LABEL_NL NULL
#define BEETLE_OPT_ENABLE_MEMCARD1_INFO_0_NL NULL
#define BEETLE_OPT_SHARED_MEMORY_CARDS_LABEL_NL NULL
#define BEETLE_OPT_SHARED_MEMORY_CARDS_INFO_0_NL NULL
#define BEETLE_OPT_MEMCARD_LEFT_INDEX_LABEL_NL NULL
#define BEETLE_OPT_MEMCARD_LEFT_INDEX_INFO_0_NL NULL
#define OPTION_VAL_0_O48_NL NULL
#define BEETLE_OPT_MEMCARD_RIGHT_INDEX_LABEL_NL NULL
#define BEETLE_OPT_MEMCARD_RIGHT_INDEX_INFO_0_NL NULL
#define OPTION_VAL_1_O49_NL NULL
#define BEETLE_OPT_LINE_RENDER_LABEL_NL NULL
#define BEETLE_OPT_LINE_RENDER_INFO_0_NL NULL
#define OPTION_VAL_DEFAULT_NL "Standaard"
#define OPTION_VAL_AGGRESSIVE_NL NULL
#define BEETLE_OPT_WIDESCREEN_HACK_LABEL_NL NULL
#define BEETLE_OPT_WIDESCREEN_HACK_INFO_0_NL NULL
#define BEETLE_OPT_WIDESCREEN_HACK_ASPECT_RATIO_LABEL_NL NULL
#define BEETLE_OPT_WIDESCREEN_HACK_ASPECT_RATIO_INFO_0_NL NULL
#define OPTION_VAL_16_9_NL NULL
#define OPTION_VAL_16_10_NL NULL
#define OPTION_VAL_18_9_NL NULL
#define OPTION_VAL_19_9_NL NULL
#define OPTION_VAL_20_9_NL NULL
#define OPTION_VAL_21_9_NL NULL
#define OPTION_VAL_32_9_NL NULL
#define BEETLE_OPT_CPU_DYNAREC_LABEL_NL NULL
#define BEETLE_OPT_CPU_DYNAREC_INFO_0_NL NULL
#define OPTION_VAL_DISABLED_NL NULL
#define OPTION_VAL_EXECUTE_NL NULL
#define OPTION_VAL_EXECUTE_ONE_NL NULL
#define OPTION_VAL_RUN_INTERPRETER_NL NULL
#define BEETLE_OPT_DYNAREC_INVALIDATE_LABEL_NL NULL
#define BEETLE_OPT_DYNAREC_INVALIDATE_INFO_0_NL NULL
#define OPTION_VAL_FULL_NL NULL
#define OPTION_VAL_DMA_NL NULL
#define BEETLE_OPT_DYNAREC_EVENTCYCLES_LABEL_NL NULL
#define BEETLE_OPT_DYNAREC_EVENTCYCLES_INFO_0_NL NULL
#define OPTION_VAL_128_NL NULL
#define BEETLE_OPT_DYNAREC_SPU_SAMPLES_LABEL_NL NULL
#define BEETLE_OPT_DYNAREC_SPU_SAMPLES_INFO_0_NL NULL
#define BEETLE_OPT_CORE_TIMING_FPS_LABEL_NL NULL
#define BEETLE_OPT_CORE_TIMING_FPS_INFO_0_NL NULL
#define OPTION_VAL_AUTO_TOGGLE_NL NULL
#define OPTION_VAL_FORCE_PROGRESSIVE_NL NULL
#define OPTION_VAL_FORCE_INTERLACED_NL NULL
#define BEETLE_OPT_DEINTERLACER_LABEL_NL NULL
#define BEETLE_OPT_DEINTERLACER_INFO_0_NL NULL
#define OPTION_VAL_WEAVE_NL NULL
#define OPTION_VAL_BOB_NL NULL
#define OPTION_VAL_BOB_OFFSET_NL NULL
#define OPTION_VAL_FASTMAD_NL NULL
#define BEETLE_OPT_FRAME_DUPING_LABEL_NL NULL
#define BEETLE_OPT_FRAME_DUPING_INFO_0_NL NULL
#define BEETLE_OPT_PAL_VIDEO_TIMING_OVERRIDE_LABEL_NL NULL
#define BEETLE_OPT_PAL_VIDEO_TIMING_OVERRIDE_INFO_0_NL NULL
#define BEETLE_OPT_ASPECT_RATIO_LABEL_NL NULL
#define BEETLE_OPT_ASPECT_RATIO_INFO_0_NL NULL
#define OPTION_VAL_CORRECTED_NL NULL
#define OPTION_VAL_UNCORRECTED_NL NULL
#define OPTION_VAL_4_3_NL NULL
#define OPTION_VAL_NTSC_NL NULL
#define BEETLE_OPT_CROP_OVERSCAN_LABEL_NL NULL
#define BEETLE_OPT_CROP_OVERSCAN_INFO_0_NL NULL
#define OPTION_VAL_STATIC_NL NULL
#define OPTION_VAL_SMART_NL NULL
#define BEETLE_OPT_IMAGE_CROP_LABEL_NL NULL
#define BEETLE_OPT_IMAGE_CROP_INFO_0_NL NULL
#define OPTION_VAL_9PX_NL NULL
#define OPTION_VAL_10PX_NL NULL
#define OPTION_VAL_11PX_NL NULL
#define OPTION_VAL_12PX_NL NULL
#define OPTION_VAL_13PX_NL NULL
#define OPTION_VAL_14PX_NL NULL
#define OPTION_VAL_15PX_NL NULL
#define OPTION_VAL_16PX_NL NULL
#define OPTION_VAL_17PX_NL NULL
#define OPTION_VAL_18PX_NL NULL
#define OPTION_VAL_19PX_NL NULL
#define OPTION_VAL_20PX_NL NULL
#define BEETLE_OPT_IMAGE_OFFSET_LABEL_NL NULL
#define BEETLE_OPT_IMAGE_OFFSET_INFO_0_NL NULL
#define OPTION_VAL_12PX_O64_NL NULL
#define OPTION_VAL_11PX_O64_NL NULL
#define OPTION_VAL_10PX_O64_NL NULL
#define OPTION_VAL_9PX_O64_NL NULL
#define OPTION_VAL_8PX_O64_NL NULL
#define OPTION_VAL_7PX_O64_NL NULL
#define OPTION_VAL_6PX_O64_NL NULL
#define OPTION_VAL_5PX_O64_NL NULL
#define OPTION_VAL_4PX_O64_NL NULL
#define OPTION_VAL_3PX_O64_NL NULL
#define OPTION_VAL_2PX_O64_NL NULL
#define OPTION_VAL_1PX_O64_NL NULL
#define OPTION_VAL_1PX_O64_0_NL NULL
#define OPTION_VAL_2PX_O64_0_NL NULL
#define OPTION_VAL_3PX_O64_0_NL NULL
#define OPTION_VAL_4PX_O64_0_NL NULL
#define OPTION_VAL_5PX_O64_0_NL NULL
#define OPTION_VAL_6PX_O64_0_NL NULL
#define OPTION_VAL_7PX_O64_0_NL NULL
#define OPTION_VAL_8PX_O64_0_NL NULL
#define OPTION_VAL_9PX_O64_0_NL NULL
#define OPTION_VAL_10PX_O64_0_NL NULL
#define OPTION_VAL_11PX_O64_0_NL NULL
#define OPTION_VAL_12PX_O64_0_NL NULL
#define BEETLE_OPT_IMAGE_OFFSET_CYCLES_LABEL_NL NULL
#define BEETLE_OPT_IMAGE_OFFSET_CYCLES_INFO_0_NL NULL
#define BEETLE_OPT_INITIAL_SCANLINE_LABEL_NL NULL
#define BEETLE_OPT_INITIAL_SCANLINE_INFO_0_NL NULL
#define BEETLE_OPT_LAST_SCANLINE_LABEL_NL NULL
#define BEETLE_OPT_LAST_SCANLINE_INFO_0_NL NULL
#define OPTION_VAL_239_NL NULL
#define BEETLE_OPT_INITIAL_SCANLINE_PAL_LABEL_NL NULL
#define BEETLE_OPT_INITIAL_SCANLINE_PAL_INFO_0_NL NULL
#define BEETLE_OPT_LAST_SCANLINE_PAL_LABEL_NL NULL
#define BEETLE_OPT_LAST_SCANLINE_PAL_INFO_0_NL NULL
#define OPTION_VAL_287_NL NULL

struct retro_core_option_v2_category option_cats_nl[] = {
   {
      "system",
      CATEGORY_SYSTEM_LABEL_NL,
      CATEGORY_SYSTEM_INFO_0_NL
   },
   {
      "video",
      CATEGORY_VIDEO_LABEL_NL,
      CATEGORY_VIDEO_INFO_0_NL
   },
   {
      "pgxp",
      CATEGORY_PGXP_LABEL_NL,
      CATEGORY_PGXP_INFO_0_NL
   },
   {
      "osd",
      CATEGORY_OSD_LABEL_NL,
      CATEGORY_OSD_INFO_0_NL
   },
   {
      "input",
      CATEGORY_INPUT_LABEL_NL,
      CATEGORY_INPUT_INFO_0_NL
   },
   {
      "memcards",
      CATEGORY_MEMCARDS_LABEL_NL,
      CATEGORY_MEMCARDS_INFO_0_NL
   },
   {
      "hacks",
      CATEGORY_HACKS_LABEL_NL,
      CATEGORY_HACKS_INFO_0_NL
   },
   { NULL, NULL, NULL },
};
struct retro_core_option_v2_definition option_defs_nl[] = {
   {
      BEETLE_OPT(cpu_freq_scale),
      BEETLE_OPT_CPU_FREQ_SCALE_LABEL_NL,
      NULL,
      BEETLE_OPT_CPU_FREQ_SCALE_INFO_0_NL,
      NULL,
      "system",
      {
         { "50%",           OPTION_VAL_50_NL },
         { "60%",           OPTION_VAL_60_NL },
         { "70%",           OPTION_VAL_70_NL },
         { "80%",           OPTION_VAL_80_NL },
         { "90%",           OPTION_VAL_90_NL },
         { "100%",          OPTION_VAL_100_NL },
         { "110%",          OPTION_VAL_110_NL },
         { "120%",          OPTION_VAL_120_NL },
         { "130%",          OPTION_VAL_130_NL },
         { "140%",          OPTION_VAL_140_NL },
         { "150%",          OPTION_VAL_150_NL },
         { "160%",          OPTION_VAL_160_NL },
         { "170%",          OPTION_VAL_170_NL },
         { "180%",          OPTION_VAL_180_NL },
         { "190%",          OPTION_VAL_190_NL },
         { "200%",          OPTION_VAL_200_NL },
         { "210%",          OPTION_VAL_210_NL },
         { "220%",          OPTION_VAL_220_NL },
         { "230%",          OPTION_VAL_230_NL },
         { "240%",          OPTION_VAL_240_NL },
         { "250%",          OPTION_VAL_250_NL },
         { "260%",          OPTION_VAL_260_NL },
         { "270%",          OPTION_VAL_270_NL },
         { "280%",          OPTION_VAL_280_NL },
         { "290%",          OPTION_VAL_290_NL },
         { "300%",          OPTION_VAL_300_NL },
         { "310%",          OPTION_VAL_310_NL },
         { "320%",          OPTION_VAL_320_NL },
         { "330%",          OPTION_VAL_330_NL },
         { "340%",          OPTION_VAL_340_NL },
         { "350%",          OPTION_VAL_350_NL },
         { "360%",          OPTION_VAL_360_NL },
         { "370%",          OPTION_VAL_370_NL },
         { "380%",          OPTION_VAL_380_NL },
         { "390%",          OPTION_VAL_390_NL },
         { "400%",          OPTION_VAL_400_NL },
         { "410%",          OPTION_VAL_410_NL },
         { "420%",          OPTION_VAL_420_NL },
         { "430%",          OPTION_VAL_430_NL },
         { "440%",          OPTION_VAL_440_NL },
         { "450%",          OPTION_VAL_450_NL },
         { "460%",          OPTION_VAL_460_NL },
         { "470%",          OPTION_VAL_470_NL },
         { "480%",          OPTION_VAL_480_NL },
         { "490%",          OPTION_VAL_490_NL },
         { "500%",          OPTION_VAL_500_NL },
         { "510%",          OPTION_VAL_510_NL },
         { "520%",          OPTION_VAL_520_NL },
         { "530%",          OPTION_VAL_530_NL },
         { "540%",          OPTION_VAL_540_NL },
         { "550%",          OPTION_VAL_550_NL },
         { "560%",          OPTION_VAL_560_NL },
         { "570%",          OPTION_VAL_570_NL },
         { "580%",          OPTION_VAL_580_NL },
         { "590%",          OPTION_VAL_590_NL },
         { "600%",          OPTION_VAL_600_NL },
         { "610%",          OPTION_VAL_610_NL },
         { "620%",          OPTION_VAL_620_NL },
         { "630%",          OPTION_VAL_630_NL },
         { "640%",          OPTION_VAL_640_NL },
         { "650%",          OPTION_VAL_650_NL },
         { "660%",          OPTION_VAL_660_NL },
         { "670%",          OPTION_VAL_670_NL },
         { "680%",          OPTION_VAL_680_NL },
         { "690%",          OPTION_VAL_690_NL },
         { "700%",          OPTION_VAL_700_NL },
         { "710%",          OPTION_VAL_710_NL },
         { "720%",          OPTION_VAL_720_NL },
         { "730%",          OPTION_VAL_730_NL },
         { "740%",          OPTION_VAL_740_NL },
         { "750%",          OPTION_VAL_750_NL },
         { NULL, NULL },
      },
      "100%"
   },
   {
      BEETLE_OPT(gpu_overclock),
      BEETLE_OPT_GPU_OVERCLOCK_LABEL_NL,
      NULL,
      BEETLE_OPT_GPU_OVERCLOCK_INFO_0_NL,
      NULL,
      "system",
      {
         { "1x(native)", OPTION_VAL_1X_NATIVE_NL },
         { "2x",         OPTION_VAL_2X_NL },
         { "4x",         OPTION_VAL_4X_NL },
         { "8x",         OPTION_VAL_8X_NL },
         { "16x",        OPTION_VAL_16X_NL },
         { "32x",        OPTION_VAL_32X_NL },
         { NULL, NULL },
      },
      "1x(native)"
   },
   {
      BEETLE_OPT(gte_overclock),
      BEETLE_OPT_GTE_OVERCLOCK_LABEL_NL,
      NULL,
      BEETLE_OPT_GTE_OVERCLOCK_INFO_0_NL,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
#ifndef EMSCRIPTEN
   {
      BEETLE_OPT(cd_access_method),
      BEETLE_OPT_CD_ACCESS_METHOD_LABEL_NL,
      NULL,
      BEETLE_OPT_CD_ACCESS_METHOD_INFO_0_NL,
      NULL,
      "system",
      {
         { "sync",     OPTION_VAL_SYNC_NL },
         { "async",    OPTION_VAL_ASYNC_NL },
         { "precache", OPTION_VAL_PRECACHE_NL },
         { NULL, NULL },
      },
      "sync"
   },
#endif
   {
      BEETLE_OPT(cd_fastload),
      BEETLE_OPT_CD_FASTLOAD_LABEL_NL,
      NULL,
      BEETLE_OPT_CD_FASTLOAD_INFO_0_NL,
      NULL,
      "system",
      {
         { "2x(native)", OPTION_VAL_2X_NATIVE_NL },
         { "4x",          OPTION_VAL_4X_NL },
         { "6x",          OPTION_VAL_6X_NL },
         { "8x",          OPTION_VAL_8X_NL },
         { "10x",         OPTION_VAL_10X_NL },
         { "12x",         OPTION_VAL_12X_NL },
         { "14x",         OPTION_VAL_14X_NL },
         { NULL, NULL },
      },
      "2x(native)"
   },
   {
      BEETLE_OPT(skip_bios),
      BEETLE_OPT_SKIP_BIOS_LABEL_NL,
      NULL,
      BEETLE_OPT_SKIP_BIOS_INFO_0_NL,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(override_bios),
      BEETLE_OPT_OVERRIDE_BIOS_LABEL_NL,
      NULL,
      BEETLE_OPT_OVERRIDE_BIOS_INFO_0_NL,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "psxonpsp", OPTION_VAL_PSXONPSP_NL },
         { "ps1_rom",  OPTION_VAL_PS1_ROM_NL },
         { "openbios", OPTION_VAL_OPENBIOS_NL     },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(region),
      BEETLE_OPT_REGION_LABEL_NL,
      NULL,
      BEETLE_OPT_REGION_INFO_0_NL,
      NULL,
      "system",
      {
         { "auto",   OPTION_VAL_AUTO_NL },
         { "ntsc-j", OPTION_VAL_NTSC_J_NL },
         { "ntsc-u", OPTION_VAL_NTSC_U_NL },
         { "pal",    OPTION_VAL_PAL_NL },
         { NULL, NULL },
      },
      "auto"
   },
   {
      BEETLE_OPT(internal_resolution),
      BEETLE_OPT_INTERNAL_RESOLUTION_LABEL_NL,
      NULL,
      BEETLE_OPT_INTERNAL_RESOLUTION_INFO_0_NL,
      NULL,
      "video",
      {
         { "1x(native)", OPTION_VAL_1X_NATIVE_NL },
         { "2x",         OPTION_VAL_2X_NL },
         { "4x",         OPTION_VAL_4X_NL },
         { "8x",         OPTION_VAL_8X_NL },
         { "16x",        OPTION_VAL_16X_NL },
         { NULL, NULL },
      },
      "1x(native)"
   },
#if defined(HAVE_OPENGL) || defined(HAVE_OPENGLES) || defined(HAVE_VULKAN)
   {
      BEETLE_OPT(renderer),
      BEETLE_OPT_RENDERER_LABEL_NL,
      NULL,
      BEETLE_OPT_RENDERER_INFO_0_NL,
      NULL,
      "video",
      {
         { "hardware",    OPTION_VAL_HARDWARE_NL },
#if defined(HAVE_OPENGL) || defined(HAVE_OPENGLES)
         { "hardware_gl", OPTION_VAL_HARDWARE_GL_NL },
#endif
#if defined(HAVE_VULKAN)
         { "hardware_vk", OPTION_VAL_HARDWARE_VK_NL },
#endif
         { "software",    OPTION_VAL_SOFTWARE_NL },
         { NULL, NULL },
      },
      "hardware"
   },
   {
      BEETLE_OPT(renderer_software_fb),
      BEETLE_OPT_RENDERER_SOFTWARE_FB_LABEL_NL,
      NULL,
      BEETLE_OPT_RENDERER_SOFTWARE_FB_INFO_0_NL,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled"
   },
#endif
#if defined(HAVE_OPENGL) || defined(HAVE_OPENGLES)
   {
      BEETLE_OPT(depth),
      BEETLE_OPT_DEPTH_LABEL_NL,
      NULL,
      BEETLE_OPT_DEPTH_INFO_0_NL,
      NULL,
      "video",
      {
         { "16bpp(native)", OPTION_VAL_16BPP_NATIVE_NL },
         { "32bpp",         OPTION_VAL_32BPP_NL },
         { NULL, NULL },
      },
      "16bpp(native)"
   },
   // Sort of, it's more like 15-bit high color and 24-bit true color for visible output. The alpha channel is used for mask bit. Vulkan renderer uses ABGR1555_555 for 31 bits internal? FMVs are always 24-bit on all renderers like original hardware (BGR888, no alpha)
#endif
   {
      BEETLE_OPT(dither_mode),
      BEETLE_OPT_DITHER_MODE_LABEL_NL,
      NULL,
      BEETLE_OPT_DITHER_MODE_INFO_0_NL,
      NULL,
      "video",
      {
         { "1x(native)",          OPTION_VAL_1X_NATIVE_NL },
         { "internal resolution", OPTION_VAL_INTERNAL_RESOLUTION_NL },
         { "disabled",            NULL },
         { NULL, NULL },
      },
      "1x(native)"
   },
#ifdef HAVE_VULKAN
   {
      BEETLE_OPT(scaled_uv_offset),
      BEETLE_OPT_SCALED_UV_OFFSET_LABEL_NL,
      NULL,
      BEETLE_OPT_SCALED_UV_OFFSET_INFO_0_NL,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled"
   },
#endif
#if defined(HAVE_OPENGL) || defined(HAVE_OPENGLES) || defined(HAVE_VULKAN)
   {
      BEETLE_OPT(filter),
      BEETLE_OPT_FILTER_LABEL_NL,
      NULL,
      BEETLE_OPT_FILTER_INFO_0_NL,
      NULL,
      "video",
      {
         { "nearest",  OPTION_VAL_NEAREST_NL },
         { "SABR",     OPTION_VAL_SABR_NL },
         { "xBR",      OPTION_VAL_XBR_NL },
         { "bilinear", OPTION_VAL_BILINEAR_NL },
         { "3-point",  OPTION_VAL_3_POINT_NL },
         { "JINC2",    OPTION_VAL_JINC2_NL },
         { NULL, NULL },
      },
      "nearest"
   },
#ifdef HAVE_VULKAN
   {
      BEETLE_OPT(filter_exclude_sprite),
      BEETLE_OPT_FILTER_EXCLUDE_SPRITE_LABEL_NL,
      NULL,
      BEETLE_OPT_FILTER_EXCLUDE_SPRITE_INFO_0_NL,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "opaque", OPTION_VAL_OPAQUE_NL },
         { "all", OPTION_VAL_ALL_NL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(filter_exclude_2d_polygon),
      BEETLE_OPT_FILTER_EXCLUDE_2D_POLYGON_LABEL_NL,
      NULL,
      BEETLE_OPT_FILTER_EXCLUDE_2D_POLYGON_INFO_0_NL,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "opaque", OPTION_VAL_OPAQUE_NL },
         { "all", OPTION_VAL_ALL_NL },
         { NULL, NULL },
      },
      "disabled"
   },
#endif
#endif
#ifdef HAVE_VULKAN
   {
      BEETLE_OPT(adaptive_smoothing),
      BEETLE_OPT_ADAPTIVE_SMOOTHING_LABEL_NL,
      NULL,
      BEETLE_OPT_ADAPTIVE_SMOOTHING_INFO_0_NL,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(super_sampling),
      BEETLE_OPT_SUPER_SAMPLING_LABEL_NL,
      NULL,
      BEETLE_OPT_SUPER_SAMPLING_INFO_0_NL,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(msaa),
      BEETLE_OPT_MSAA_LABEL_NL,
      NULL,
      BEETLE_OPT_MSAA_INFO_0_NL,
      NULL,
      "video",
      {
         { "1x",  OPTION_VAL_1X_NL },
         { "2x",  OPTION_VAL_2X_NL },
         { "4x",  OPTION_VAL_4X_NL },
         { "8x",  OPTION_VAL_8X_NL },
         { "16x", OPTION_VAL_16X_NL },
         { NULL, NULL },
      },
      "1x"
   },
   {
      BEETLE_OPT(mdec_yuv),
      BEETLE_OPT_MDEC_YUV_LABEL_NL,
      NULL,
      BEETLE_OPT_MDEC_YUV_INFO_0_NL,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(track_textures),
      BEETLE_OPT_TRACK_TEXTURES_LABEL_NL,
      NULL,
      BEETLE_OPT_TRACK_TEXTURES_INFO_0_NL,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
#ifdef TEXTURE_DUMPING_ENABLED
   {
      BEETLE_OPT(dump_textures),
      BEETLE_OPT_DUMP_TEXTURES_LABEL_NL,
      NULL,
      BEETLE_OPT_DUMP_TEXTURES_INFO_0_NL,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
#endif
   {
      BEETLE_OPT(replace_textures),
      BEETLE_OPT_REPLACE_TEXTURES_LABEL_NL,
      NULL,
      BEETLE_OPT_REPLACE_TEXTURES_INFO_0_NL,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
#endif
   {
      BEETLE_OPT(pgxp_mode),
      BEETLE_OPT_PGXP_MODE_LABEL_NL,
      NULL,
      BEETLE_OPT_PGXP_MODE_INFO_0_NL,
      NULL,
      "pgxp",
      {
         { "disabled",     NULL },
         { "memory only",  OPTION_VAL_MEMORY_ONLY_NL },
         { "memory + CPU", OPTION_VAL_MEMORY_CPU_NL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(pgxp_2d_tol),
      BEETLE_OPT_PGXP_2D_TOL_LABEL_NL,
      NULL,
      BEETLE_OPT_PGXP_2D_TOL_INFO_0_NL,
      NULL,
      "pgxp",
      {
         { "disabled", NULL },
         { "0px", OPTION_VAL_0PX_NL },
         { "1px", OPTION_VAL_1PX_NL },
         { "2px", OPTION_VAL_2PX_NL },
         { "3px", OPTION_VAL_3PX_NL },
         { "4px", OPTION_VAL_4PX_NL },
         { "5px", OPTION_VAL_5PX_NL },
         { "6px", OPTION_VAL_6PX_NL },
         { "7px", OPTION_VAL_7PX_NL },
         { "8px", OPTION_VAL_8PX_NL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(pgxp_nclip),
      BEETLE_OPT_PGXP_NCLIP_LABEL_NL,
      NULL,
      BEETLE_OPT_PGXP_NCLIP_INFO_0_NL,
      NULL,
      "pgxp",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
#if defined(HAVE_OPENGL) || defined(HAVE_OPENGLES) || defined(HAVE_VULKAN)
   {
      BEETLE_OPT(pgxp_vertex),
      BEETLE_OPT_PGXP_VERTEX_LABEL_NL,
      NULL,
      BEETLE_OPT_PGXP_VERTEX_INFO_0_NL,
      NULL,
      "pgxp",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
#endif
   {
      BEETLE_OPT(pgxp_texture),
      BEETLE_OPT_PGXP_TEXTURE_LABEL_NL,
      NULL,
      BEETLE_OPT_PGXP_TEXTURE_INFO_0_NL,
      NULL,
      "pgxp",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(display_internal_fps),
      BEETLE_OPT_DISPLAY_INTERNAL_FPS_LABEL_NL,
      NULL,
      BEETLE_OPT_DISPLAY_INTERNAL_FPS_INFO_0_NL,
      NULL,
      "osd",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(display_osd),
      BEETLE_OPT_DISPLAY_OSD_LABEL_NL,
      NULL,
      BEETLE_OPT_DISPLAY_OSD_INFO_0_NL,
      NULL,
      "osd",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled"
   },
#if defined(HAVE_OPENGL) || defined(HAVE_OPENGLES) || defined(HAVE_VULKAN)
   {
      BEETLE_OPT(display_vram),
      BEETLE_OPT_DISPLAY_VRAM_LABEL_NL,
      NULL,
      BEETLE_OPT_DISPLAY_VRAM_INFO_0_NL,
      NULL,
      "osd",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
#endif
   {
      BEETLE_OPT(analog_calibration),
      BEETLE_OPT_ANALOG_CALIBRATION_LABEL_NL,
      NULL,
      BEETLE_OPT_ANALOG_CALIBRATION_INFO_0_NL,
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(analog_toggle),
      BEETLE_OPT_ANALOG_TOGGLE_LABEL_NL,
      NULL,
      BEETLE_OPT_ANALOG_TOGGLE_INFO_0_NL,
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { "enabled-analog", OPTION_VAL_ENABLED_ANALOG_NL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(analog_toggle_combo),
      BEETLE_OPT_ANALOG_TOGGLE_COMBO_LABEL_NL,
      NULL,
      BEETLE_OPT_ANALOG_TOGGLE_COMBO_INFO_0_NL,
      NULL,
      "input",
      {
         { "l1+l2+r1+r2+start+select", OPTION_VAL_L1_L2_R1_R2_START_SELECT_NL },
         { "l1+r1+select",             OPTION_VAL_L1_R1_SELECT_NL },
         { "l1+r1+start",              OPTION_VAL_L1_R1_START_NL },
         { "l1+r1+l3",                 OPTION_VAL_L1_R1_L3_NL },
         { "l1+r1+r3",                 OPTION_VAL_L1_R1_R3_NL },
         { "l2+r2+select",             OPTION_VAL_L2_R2_SELECT_NL },
         { "l2+r2+start",              OPTION_VAL_L2_R2_START_NL },
         { "l2+r2+l3",                 OPTION_VAL_L2_R2_L3_NL },
         { "l2+r2+r3",                 OPTION_VAL_L2_R2_R3_NL },
         { "l3+r3",                    OPTION_VAL_L3_R3_NL },
         { NULL, NULL },
      },
      "l1+r1+select"
   },
   {
      BEETLE_OPT(analog_toggle_hold),
      BEETLE_OPT_ANALOG_TOGGLE_HOLD_LABEL_NL,
      NULL,
      BEETLE_OPT_ANALOG_TOGGLE_HOLD_INFO_0_NL,
      NULL,
      "input",
      {
         { "0", OPTION_VAL_0_NL },
         { "1", OPTION_VAL_1_NL },
         { "2", OPTION_VAL_2_NL },
         { "3", OPTION_VAL_3_NL },
         { "4", OPTION_VAL_4_NL },
         { "5", OPTION_VAL_5_NL },
         { NULL, NULL },
      },
      "1"
   },
   {
      BEETLE_OPT(enable_multitap_port1),
      BEETLE_OPT_ENABLE_MULTITAP_PORT1_LABEL_NL,
      NULL,
      BEETLE_OPT_ENABLE_MULTITAP_PORT1_INFO_0_NL,
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(enable_multitap_port2),
      BEETLE_OPT_ENABLE_MULTITAP_PORT2_LABEL_NL,
      NULL,
      BEETLE_OPT_ENABLE_MULTITAP_PORT2_INFO_0_NL,
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(gun_input_mode),
      BEETLE_OPT_GUN_INPUT_MODE_LABEL_NL,
      NULL,
      BEETLE_OPT_GUN_INPUT_MODE_INFO_0_NL,
      NULL,
      "input",
      {
         { "lightgun",    OPTION_VAL_LIGHTGUN_NL },
         { "touchscreen", OPTION_VAL_TOUCHSCREEN_NL },
         { NULL, NULL },
      },
      "lightgun"
   },
   // Shouldn't the gun_input_mode just be Mouse vs. Touchscreen?
   {
      BEETLE_OPT(gun_cursor),
      BEETLE_OPT_GUN_CURSOR_LABEL_NL,
      NULL,
      BEETLE_OPT_GUN_CURSOR_INFO_0_NL,
      NULL,
      "input",
      {
         { "off",   "disabled" },
         { "cross", OPTION_VAL_CROSS_NL },
         { "dot",   OPTION_VAL_DOT_NL },
         { NULL, NULL },
      },
      "cross"
   },
   {
      BEETLE_OPT(crosshair_color_p1),
      BEETLE_OPT_CROSSHAIR_COLOR_P1_LABEL_NL,
      NULL,
      BEETLE_OPT_CROSSHAIR_COLOR_P1_INFO_0_NL,
      NULL,
      "input",
      {
         { "red",    OPTION_VAL_RED_NL },
         { "blue",   OPTION_VAL_BLUE_NL },
         { "green",  OPTION_VAL_GREEN_NL },
         { "orange", OPTION_VAL_ORANGE_NL },
         { "yellow", OPTION_VAL_YELLOW_NL },
         { "cyan",   OPTION_VAL_CYAN_NL },
         { "pink",   OPTION_VAL_PINK_NL },
         { "purple", OPTION_VAL_PURPLE_NL },
         { "black",  OPTION_VAL_BLACK_NL },
         { "white",  OPTION_VAL_WHITE_NL },
         { NULL, NULL },
      },
      "red"
   },
   {
      BEETLE_OPT(crosshair_color_p2),
      BEETLE_OPT_CROSSHAIR_COLOR_P2_LABEL_NL,
      NULL,
      BEETLE_OPT_CROSSHAIR_COLOR_P2_INFO_0_NL,
      NULL,
      "input",
      {
         { "blue",   OPTION_VAL_BLUE_NL },
         { "red",    OPTION_VAL_RED_NL },
         { "green",  OPTION_VAL_GREEN_NL },
         { "orange", OPTION_VAL_ORANGE_NL },
         { "yellow", OPTION_VAL_YELLOW_NL },
         { "cyan",   OPTION_VAL_CYAN_NL },
         { "pink",   OPTION_VAL_PINK_NL },
         { "purple", OPTION_VAL_PURPLE_NL },
         { "black",  OPTION_VAL_BLACK_NL },
         { "white",  OPTION_VAL_WHITE_NL },
         { NULL, NULL },
      },
      "blue"
   },
   {
      BEETLE_OPT(mouse_sensitivity),
      BEETLE_OPT_MOUSE_SENSITIVITY_LABEL_NL,
      NULL,
      BEETLE_OPT_MOUSE_SENSITIVITY_INFO_0_NL,
      NULL,
      "input",
      {
         { "5%",   OPTION_VAL_5_O42_NL },
         { "10%",  OPTION_VAL_10_NL },
         { "15%",  OPTION_VAL_15_NL },
         { "20%",  OPTION_VAL_20_NL },
         { "25%",  OPTION_VAL_25_NL },
         { "30%",  OPTION_VAL_30_NL },
         { "35%",  OPTION_VAL_35_NL },
         { "40%",  OPTION_VAL_40_NL },
         { "45%",  OPTION_VAL_45_NL },
         { "50%",  OPTION_VAL_50_NL },
         { "55%",  OPTION_VAL_55_NL },
         { "60%",  OPTION_VAL_60_NL },
         { "65%",  OPTION_VAL_65_NL },
         { "70%",  OPTION_VAL_70_NL },
         { "75%",  OPTION_VAL_75_NL },
         { "80%",  OPTION_VAL_80_NL },
         { "85%",  OPTION_VAL_85_NL },
         { "90%",  OPTION_VAL_90_NL },
         { "95%",  OPTION_VAL_95_NL },
         { "100%", OPTION_VAL_100_O42_NL },
         { "105%", OPTION_VAL_105_NL },
         { "110%", OPTION_VAL_110_NL },
         { "115%", OPTION_VAL_115_NL },
         { "120%", OPTION_VAL_120_NL },
         { "125%", OPTION_VAL_125_NL },
         { "130%", OPTION_VAL_130_NL },
         { "135%", OPTION_VAL_135_NL },
         { "140%", OPTION_VAL_140_NL },
         { "145%", OPTION_VAL_145_NL },
         { "150%", OPTION_VAL_150_NL },
         { "155%", OPTION_VAL_155_NL },
         { "160%", OPTION_VAL_160_NL },
         { "165%", OPTION_VAL_165_NL },
         { "170%", OPTION_VAL_170_NL },
         { "175%", OPTION_VAL_175_NL },
         { "180%", OPTION_VAL_180_NL },
         { "185%", OPTION_VAL_185_NL },
         { "190%", OPTION_VAL_190_NL },
         { "195%", OPTION_VAL_195_NL },
         { "200%", OPTION_VAL_200_NL },
         { NULL, NULL },
      },
      "100%"
   },
   {
      BEETLE_OPT(negcon_response),
      BEETLE_OPT_NEGCON_RESPONSE_LABEL_NL,
      NULL,
      BEETLE_OPT_NEGCON_RESPONSE_INFO_0_NL,
      NULL,
      "input",
      {
         { "linear",    OPTION_VAL_LINEAR_NL },
         { "quadratic", OPTION_VAL_QUADRATIC_NL },
         { "cubic",     OPTION_VAL_CUBIC_NL },
         { NULL, NULL },
      },
      "linear"
   },
   {
      BEETLE_OPT(negcon_deadzone),
      BEETLE_OPT_NEGCON_DEADZONE_LABEL_NL,
      NULL,
      BEETLE_OPT_NEGCON_DEADZONE_INFO_0_NL,
      NULL,
      "input",
      {
         { "0%",  OPTION_VAL_0_O44_NL },
         { "5%",  OPTION_VAL_5_O42_NL },
         { "10%", OPTION_VAL_10_NL },
         { "15%", OPTION_VAL_15_NL },
         { "20%", OPTION_VAL_20_NL },
         { "25%", OPTION_VAL_25_NL },
         { "30%", OPTION_VAL_30_NL },
         { NULL, NULL },
      },
      "0%"
   },
   {
      BEETLE_OPT(use_mednafen_memcard0_method),
      BEETLE_OPT_USE_MEDNAFEN_MEMCARD0_METHOD_LABEL_NL,
      NULL,
      BEETLE_OPT_USE_MEDNAFEN_MEMCARD0_METHOD_INFO_0_NL,
      NULL,
      "memcards",
      {
         { "libretro", OPTION_VAL_LIBRETRO_NL },
         { "mednafen", OPTION_VAL_MEDNAFEN_NL },
         { NULL, NULL },
      },
      "libretro"
   },
   {
      BEETLE_OPT(enable_memcard1),
      BEETLE_OPT_ENABLE_MEMCARD1_LABEL_NL,
      NULL,
      BEETLE_OPT_ENABLE_MEMCARD1_INFO_0_NL,
      NULL,
      "memcards",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(shared_memory_cards),
      BEETLE_OPT_SHARED_MEMORY_CARDS_LABEL_NL,
      NULL,
      BEETLE_OPT_SHARED_MEMORY_CARDS_INFO_0_NL,
      NULL,
      "memcards",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(memcard_left_index),
      BEETLE_OPT_MEMCARD_LEFT_INDEX_LABEL_NL,
      NULL,
      BEETLE_OPT_MEMCARD_LEFT_INDEX_INFO_0_NL,
      NULL,
      "memcards",
      {
         { "0",  OPTION_VAL_0_O48_NL },
         { "1",  NULL },
         { "2",  NULL },
         { "3",  NULL },
         { "4",  NULL },
         { "5",  NULL },
         { "6",  NULL },
         { "7",  NULL },
         { "8",  NULL },
         { "9",  NULL },
         { "10", NULL },
         { "11", NULL },
         { "12", NULL },
         { "13", NULL },
         { "14", NULL },
         { "15", NULL },
         { "16", NULL },
         { "17", NULL },
         { "18", NULL },
         { "19", NULL },
         { "20", NULL },
         { "21", NULL },
         { "22", NULL },
         { "23", NULL },
         { "24", NULL },
         { "25", NULL },
         { "26", NULL },
         { "27", NULL },
         { "28", NULL },
         { "29", NULL },
         { "30", NULL },
         { "31", NULL },
         { "32", NULL },
         { "33", NULL },
         { "34", NULL },
         { "35", NULL },
         { "36", NULL },
         { "37", NULL },
         { "38", NULL },
         { "39", NULL },
         { "40", NULL },
         { "41", NULL },
         { "42", NULL },
         { "43", NULL },
         { "44", NULL },
         { "45", NULL },
         { "46", NULL },
         { "47", NULL },
         { "48", NULL },
         { "49", NULL },
         { "50", NULL },
         { "51", NULL },
         { "52", NULL },
         { "53", NULL },
         { "54", NULL },
         { "55", NULL },
         { "56", NULL },
         { "57", NULL },
         { "58", NULL },
         { "59", NULL },
         { "60", NULL },
         { "61", NULL },
         { "62", NULL },
         { "63", NULL },
         { NULL, NULL },
      },
      "0"
   },
   {
      BEETLE_OPT(memcard_right_index),
      BEETLE_OPT_MEMCARD_RIGHT_INDEX_LABEL_NL,
      NULL,
      BEETLE_OPT_MEMCARD_RIGHT_INDEX_INFO_0_NL,
      NULL,
      "memcards",
      {
         { "0",  NULL },
         { "1",  OPTION_VAL_1_O49_NL },
         { "2",  NULL },
         { "3",  NULL },
         { "4",  NULL },
         { "5",  NULL },
         { "6",  NULL },
         { "7",  NULL },
         { "8",  NULL },
         { "9",  NULL },
         { "10", NULL },
         { "11", NULL },
         { "12", NULL },
         { "13", NULL },
         { "14", NULL },
         { "15", NULL },
         { "16", NULL },
         { "17", NULL },
         { "18", NULL },
         { "19", NULL },
         { "20", NULL },
         { "21", NULL },
         { "22", NULL },
         { "23", NULL },
         { "24", NULL },
         { "25", NULL },
         { "26", NULL },
         { "27", NULL },
         { "28", NULL },
         { "29", NULL },
         { "30", NULL },
         { "31", NULL },
         { "32", NULL },
         { "33", NULL },
         { "34", NULL },
         { "35", NULL },
         { "36", NULL },
         { "37", NULL },
         { "38", NULL },
         { "39", NULL },
         { "40", NULL },
         { "41", NULL },
         { "42", NULL },
         { "43", NULL },
         { "44", NULL },
         { "45", NULL },
         { "46", NULL },
         { "47", NULL },
         { "48", NULL },
         { "49", NULL },
         { "50", NULL },
         { "51", NULL },
         { "52", NULL },
         { "53", NULL },
         { "54", NULL },
         { "55", NULL },
         { "56", NULL },
         { "57", NULL },
         { "58", NULL },
         { "59", NULL },
         { "60", NULL },
         { "61", NULL },
         { "62", NULL },
         { "63", NULL },
         { NULL, NULL },
      },
      "1"
   },
   {
      BEETLE_OPT(line_render),
      BEETLE_OPT_LINE_RENDER_LABEL_NL,
      NULL,
      BEETLE_OPT_LINE_RENDER_INFO_0_NL,
      NULL,
      "hacks",
      {
         { "default",    OPTION_VAL_DEFAULT_NL },
         { "aggressive", OPTION_VAL_AGGRESSIVE_NL },
         { "disabled",   NULL },
         { NULL, NULL },
      },
      "default"
   },
   {
      BEETLE_OPT(widescreen_hack),
      BEETLE_OPT_WIDESCREEN_HACK_LABEL_NL,
      NULL,
      BEETLE_OPT_WIDESCREEN_HACK_INFO_0_NL,
      NULL,
      "hacks",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(widescreen_hack_aspect_ratio),
      BEETLE_OPT_WIDESCREEN_HACK_ASPECT_RATIO_LABEL_NL,
      NULL,
      BEETLE_OPT_WIDESCREEN_HACK_ASPECT_RATIO_INFO_0_NL,
      NULL,
      "hacks",
      {
         { "16:9",  OPTION_VAL_16_9_NL },
         { "16:10", OPTION_VAL_16_10_NL },
         { "18:9",  OPTION_VAL_18_9_NL },
         { "19:9",  OPTION_VAL_19_9_NL },
         { "20:9",  OPTION_VAL_20_9_NL },
         { "21:9",  OPTION_VAL_21_9_NL }, // 64:27
         { "32:9",  OPTION_VAL_32_9_NL },
         { NULL,    NULL },
      },
      "16:9"
   },
#if defined(HAVE_LIGHTREC)
   {
      BEETLE_OPT(cpu_dynarec),
      BEETLE_OPT_CPU_DYNAREC_LABEL_NL,
      NULL,
      BEETLE_OPT_CPU_DYNAREC_INFO_0_NL,
      NULL,
      "hacks",
      {
         { "disabled",        OPTION_VAL_DISABLED_NL },
         { "execute",         OPTION_VAL_EXECUTE_NL },
         { "execute_one",     OPTION_VAL_EXECUTE_ONE_NL },
         { "run_interpreter", OPTION_VAL_RUN_INTERPRETER_NL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(dynarec_invalidate),
      BEETLE_OPT_DYNAREC_INVALIDATE_LABEL_NL,
      NULL,
      BEETLE_OPT_DYNAREC_INVALIDATE_INFO_0_NL,
      NULL,
      "hacks",
      {
         { "full", OPTION_VAL_FULL_NL },
         { "dma",  OPTION_VAL_DMA_NL },
         { NULL, NULL },
      },
      "full"
   },
   {
      BEETLE_OPT(dynarec_eventcycles),
      BEETLE_OPT_DYNAREC_EVENTCYCLES_LABEL_NL,
      NULL,
      BEETLE_OPT_DYNAREC_EVENTCYCLES_INFO_0_NL,
      NULL,
      "hacks",
      {
         { "128",  OPTION_VAL_128_NL },
         { "256",  NULL },
         { "384",  NULL },
         { "512",  NULL },
         { "640",  NULL },
         { "768",  NULL },
         { "896",  NULL },
         { "1024", NULL },
         { "1152", NULL },
         { "1280", NULL },
         { "1408", NULL },
         { "1536", NULL },
         { "1664", NULL },
         { "1792", NULL },
         { "1920", NULL },
         { "2048", NULL },
         { NULL, NULL },
      },
      "128"
   },
   {
      BEETLE_OPT(dynarec_spu_samples),
      BEETLE_OPT_DYNAREC_SPU_SAMPLES_LABEL_NL,
      NULL,
      BEETLE_OPT_DYNAREC_SPU_SAMPLES_INFO_0_NL,
      NULL,
      "hacks",
      {
         { "1",  OPTION_VAL_1_O49_NL },
         { "4",  NULL },
         { "16", NULL },
         { NULL, NULL },
      },
      "1"
   },
#endif
   {
      BEETLE_OPT(core_timing_fps),
      BEETLE_OPT_CORE_TIMING_FPS_LABEL_NL,
      NULL,
      BEETLE_OPT_CORE_TIMING_FPS_INFO_0_NL,
      NULL,
      "video",
      {
         { "auto_toggle",       OPTION_VAL_AUTO_TOGGLE_NL },
         { "force_progressive", OPTION_VAL_FORCE_PROGRESSIVE_NL },
         { "force_interlaced",  OPTION_VAL_FORCE_INTERLACED_NL },
      },
      "auto_toggle"
   },
   {
      BEETLE_OPT(deinterlacer),
      BEETLE_OPT_DEINTERLACER_LABEL_NL,
      NULL,
      BEETLE_OPT_DEINTERLACER_INFO_0_NL,
      NULL,
      "video",
      {
         { "weave",      OPTION_VAL_WEAVE_NL },
         { "bob",        OPTION_VAL_BOB_NL },
         { "bob_offset", OPTION_VAL_BOB_OFFSET_NL },
         { "fastmad",    OPTION_VAL_FASTMAD_NL },
         { "off",        "Off" },
         { NULL, NULL },
      },
      "weave"
   },
   {
      BEETLE_OPT(frame_duping),
      BEETLE_OPT_FRAME_DUPING_LABEL_NL,
      NULL,
      BEETLE_OPT_FRAME_DUPING_INFO_0_NL,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(pal_video_timing_override),
      BEETLE_OPT_PAL_VIDEO_TIMING_OVERRIDE_LABEL_NL,
      NULL,
      BEETLE_OPT_PAL_VIDEO_TIMING_OVERRIDE_INFO_0_NL,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(aspect_ratio),
      BEETLE_OPT_ASPECT_RATIO_LABEL_NL,
      NULL,
      BEETLE_OPT_ASPECT_RATIO_INFO_0_NL,
      NULL,
      "video",
      {
         { "corrected", OPTION_VAL_CORRECTED_NL },
         { "uncorrected", OPTION_VAL_UNCORRECTED_NL },
         { "4:3",  OPTION_VAL_4_3_NL },
         { "ntsc", OPTION_VAL_NTSC_NL },
      },
      "corrected"
   },
   {
      BEETLE_OPT(crop_overscan),
      BEETLE_OPT_CROP_OVERSCAN_LABEL_NL,
      NULL,
      BEETLE_OPT_CROP_OVERSCAN_INFO_0_NL,
      NULL,
      "video",
      {
         { "disabled", "Disabled" },
         { "static",   OPTION_VAL_STATIC_NL },
         { "smart",    OPTION_VAL_SMART_NL },
         { NULL, NULL },
      },
      "smart"
   },
   {
      BEETLE_OPT(image_crop),
      BEETLE_OPT_IMAGE_CROP_LABEL_NL,
      NULL,
      BEETLE_OPT_IMAGE_CROP_INFO_0_NL,
      NULL,
      "video",
      {
         { "disabled", "0" },
         { "1px",      OPTION_VAL_1PX_NL },
         { "2px",      OPTION_VAL_2PX_NL },
         { "3px",      OPTION_VAL_3PX_NL },
         { "4px",      OPTION_VAL_4PX_NL },
         { "5px",      OPTION_VAL_5PX_NL },
         { "6px",      OPTION_VAL_6PX_NL },
         { "7px",      OPTION_VAL_7PX_NL },
         { "8px",      OPTION_VAL_8PX_NL },
         { "9px",      OPTION_VAL_9PX_NL },
         { "10px",     OPTION_VAL_10PX_NL },
         { "11px",     OPTION_VAL_11PX_NL },
         { "12px",     OPTION_VAL_12PX_NL },
         { "13px",     OPTION_VAL_13PX_NL },
         { "14px",     OPTION_VAL_14PX_NL },
         { "15px",     OPTION_VAL_15PX_NL },
         { "16px",     OPTION_VAL_16PX_NL },
         { "17px",     OPTION_VAL_17PX_NL },
         { "18px",     OPTION_VAL_18PX_NL },
         { "19px",     OPTION_VAL_19PX_NL },
         { "20px",     OPTION_VAL_20PX_NL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(image_offset),
      BEETLE_OPT_IMAGE_OFFSET_LABEL_NL,
      NULL,
      BEETLE_OPT_IMAGE_OFFSET_INFO_0_NL,
      NULL,
      "video",
      {
         { "-12px",    OPTION_VAL_12PX_O64_NL },
         { "-11px",    OPTION_VAL_11PX_O64_NL },
         { "-10px",    OPTION_VAL_10PX_O64_NL },
         { "-9px",     OPTION_VAL_9PX_O64_NL },
         { "-8px",     OPTION_VAL_8PX_O64_NL },
         { "-7px",     OPTION_VAL_7PX_O64_NL },
         { "-6px",     OPTION_VAL_6PX_O64_NL },
         { "-5px",     OPTION_VAL_5PX_O64_NL },
         { "-4px",     OPTION_VAL_4PX_O64_NL },
         { "-3px",     OPTION_VAL_3PX_O64_NL },
         { "-2px",     OPTION_VAL_2PX_O64_NL },
         { "-1px",     OPTION_VAL_1PX_O64_NL },
         { "disabled", "0" },
         { "+1px",     OPTION_VAL_1PX_O64_0_NL },
         { "+2px",     OPTION_VAL_2PX_O64_0_NL },
         { "+3px",     OPTION_VAL_3PX_O64_0_NL },
         { "+4px",     OPTION_VAL_4PX_O64_0_NL },
         { "+5px",     OPTION_VAL_5PX_O64_0_NL },
         { "+6px",     OPTION_VAL_6PX_O64_0_NL },
         { "+7px",     OPTION_VAL_7PX_O64_0_NL },
         { "+8px",     OPTION_VAL_8PX_O64_0_NL },
         { "+9px",     OPTION_VAL_9PX_O64_0_NL },
         { "+10px",    OPTION_VAL_10PX_O64_0_NL },
         { "+11px",    OPTION_VAL_11PX_O64_0_NL },
         { "+12px",    OPTION_VAL_12PX_O64_0_NL },
         { NULL, NULL },
      },
      "disabled"
   },
#if defined(HAVE_OPENGL) || defined(HAVE_OPENGLES) || defined(HAVE_VULKAN)
   {
      BEETLE_OPT(image_offset_cycles),
      BEETLE_OPT_IMAGE_OFFSET_CYCLES_LABEL_NL,
      NULL,
      BEETLE_OPT_IMAGE_OFFSET_CYCLES_INFO_0_NL,
      NULL,
      "video",
      {
         { "-40",      NULL },
         { "-39",      NULL },
         { "-38",      NULL },
         { "-37",      NULL },
         { "-36",      NULL },
         { "-35",      NULL },
         { "-34",      NULL },
         { "-33",      NULL },
         { "-32",      NULL },
         { "-31",      NULL },
         { "-30",      NULL },
         { "-29",      NULL },
         { "-28",      NULL },
         { "-27",      NULL },
         { "-26",      NULL },
         { "-25",      NULL },
         { "-24",      NULL },
         { "-23",      NULL },
         { "-22",      NULL },
         { "-21",      NULL },
         { "-20",      NULL },
         { "-19",      NULL },
         { "-18",      NULL },
         { "-17",      NULL },
         { "-16",      NULL },
         { "-15",      NULL },
         { "-14",      NULL },
         { "-13",      NULL },
         { "-12",      NULL },
         { "-11",      NULL },
         { "-10",      NULL },
         { "-9",       NULL },
         { "-8",       NULL },
         { "-7",       NULL },
         { "-6",       NULL },
         { "-5",       NULL },
         { "-4",       NULL },
         { "-3",       NULL },
         { "-2",       NULL },
         { "-1",       NULL },
         { "0",        "0" },
         { "+1",       NULL },
         { "+2",       NULL },
         { "+3",       NULL },
         { "+4",       NULL },
         { "+5",       NULL },
         { "+6",       NULL },
         { "+7",       NULL },
         { "+8",       NULL },
         { "+9",       NULL },
         { "+10",      NULL },
         { "+11",      NULL },
         { "+12",      NULL },
         { "+13",      NULL },
         { "+14",      NULL },
         { "+15",      NULL },
         { "+16",      NULL },
         { "+17",      NULL },
         { "+18",      NULL },
         { "+19",      NULL },
         { "+20",      NULL },
         { "+21",      NULL },
         { "+22",      NULL },
         { "+23",      NULL },
         { "+24",      NULL },
         { "+25",      NULL },
         { "+26",      NULL },
         { "+27",      NULL },
         { "+28",      NULL },
         { "+29",      NULL },
         { "+30",      NULL },
         { "+31",      NULL },
         { "+32",      NULL },
         { "+33",      NULL },
         { "+34",      NULL },
         { "+35",      NULL },
         { "+36",      NULL },
         { "+37",      NULL },
         { "+38",      NULL },
         { "+39",      NULL },
         { "+40",      NULL },
         { NULL, NULL},
      },
      "0"
   },
#endif
   {
      BEETLE_OPT(initial_scanline),
      BEETLE_OPT_INITIAL_SCANLINE_LABEL_NL,
      NULL,
      BEETLE_OPT_INITIAL_SCANLINE_INFO_0_NL,
      NULL,
      "video",
      {
         { "0",  OPTION_VAL_0_O48_NL },
         { "1",  NULL },
         { "2",  NULL },
         { "3",  NULL },
         { "4",  NULL },
         { "5",  NULL },
         { "6",  NULL },
         { "7",  NULL },
         { "8",  NULL },
         { "9",  NULL },
         { "10", NULL },
         { "11", NULL },
         { "12", NULL },
         { "13", NULL },
         { "14", NULL },
         { "15", NULL },
         { "16", NULL },
         { "17", NULL },
         { "18", NULL },
         { "19", NULL },
         { "20", NULL },
         { "21", NULL },
         { "22", NULL },
         { "23", NULL },
         { "24", NULL },
         { "25", NULL },
         { "26", NULL },
         { "27", NULL },
         { "28", NULL },
         { "29", NULL },
         { "30", NULL },
         { "31", NULL },
         { "32", NULL },
         { "33", NULL },
         { "34", NULL },
         { "35", NULL },
         { "36", NULL },
         { "37", NULL },
         { "38", NULL },
         { "39", NULL },
         { "40", NULL },
         { NULL, NULL },
      },
      "0"
   },
   {
      BEETLE_OPT(last_scanline),
      BEETLE_OPT_LAST_SCANLINE_LABEL_NL,
      NULL,
      BEETLE_OPT_LAST_SCANLINE_INFO_0_NL,
      NULL,
      "video",
      {
         { "210", NULL },
         { "211", NULL },
         { "212", NULL },
         { "213", NULL },
         { "214", NULL },
         { "215", NULL },
         { "216", NULL },
         { "217", NULL },
         { "218", NULL },
         { "219", NULL },
         { "220", NULL },
         { "221", NULL },
         { "222", NULL },
         { "223", NULL },
         { "224", NULL },
         { "225", NULL },
         { "226", NULL },
         { "227", NULL },
         { "228", NULL },
         { "229", NULL },
         { "230", NULL },
         { "231", NULL },
         { "232", NULL },
         { "233", NULL },
         { "234", NULL },
         { "235", NULL },
         { "236", NULL },
         { "237", NULL },
         { "238", NULL },
         { "239", OPTION_VAL_239_NL },
         { NULL, NULL },
      },
      "239"
   },
   {
      BEETLE_OPT(initial_scanline_pal),
      BEETLE_OPT_INITIAL_SCANLINE_PAL_LABEL_NL,
      NULL,
      BEETLE_OPT_INITIAL_SCANLINE_PAL_INFO_0_NL,
      NULL,
      "video",
      {
         { "0",  OPTION_VAL_0_O48_NL },
         { "1",  NULL },
         { "2",  NULL },
         { "3",  NULL },
         { "4",  NULL },
         { "5",  NULL },
         { "6",  NULL },
         { "7",  NULL },
         { "8",  NULL },
         { "9",  NULL },
         { "10", NULL },
         { "11", NULL },
         { "12", NULL },
         { "13", NULL },
         { "14", NULL },
         { "15", NULL },
         { "16", NULL },
         { "17", NULL },
         { "18", NULL },
         { "19", NULL },
         { "20", NULL },
         { "21", NULL },
         { "22", NULL },
         { "23", NULL },
         { "24", NULL },
         { "25", NULL },
         { "26", NULL },
         { "27", NULL },
         { "28", NULL },
         { "29", NULL },
         { "30", NULL },
         { "31", NULL },
         { "32", NULL },
         { "33", NULL },
         { "34", NULL },
         { "35", NULL },
         { "36", NULL },
         { "37", NULL },
         { "38", NULL },
         { "39", NULL },
         { "40", NULL },
         { NULL, NULL },
      },
      "0"
   },
   {
      BEETLE_OPT(last_scanline_pal),
      BEETLE_OPT_LAST_SCANLINE_PAL_LABEL_NL,
      NULL,
      BEETLE_OPT_LAST_SCANLINE_PAL_INFO_0_NL,
      NULL,
      "video",
      {
         { "230", NULL },
         { "231", NULL },
         { "232", NULL },
         { "233", NULL },
         { "234", NULL },
         { "235", NULL },
         { "236", NULL },
         { "237", NULL },
         { "238", NULL },
         { "239", NULL },
         { "240", NULL },
         { "241", NULL },
         { "242", NULL },
         { "243", NULL },
         { "244", NULL },
         { "245", NULL },
         { "246", NULL },
         { "247", NULL },
         { "248", NULL },
         { "249", NULL },
         { "250", NULL },
         { "251", NULL },
         { "252", NULL },
         { "253", NULL },
         { "254", NULL },
         { "255", NULL },
         { "256", NULL },
         { "257", NULL },
         { "258", NULL },
         { "259", NULL },
         { "260", NULL },
         { "261", NULL },
         { "262", NULL },
         { "263", NULL },
         { "264", NULL },
         { "265", NULL },
         { "266", NULL },
         { "267", NULL },
         { "268", NULL },
         { "269", NULL },
         { "270", NULL },
         { "271", NULL },
         { "272", NULL },
         { "273", NULL },
         { "274", NULL },
         { "275", NULL },
         { "276", NULL },
         { "277", NULL },
         { "278", NULL },
         { "279", NULL },
         { "280", NULL },
         { "281", NULL },
         { "282", NULL },
         { "283", NULL },
         { "284", NULL },
         { "285", NULL },
         { "286", NULL },
         { "287", OPTION_VAL_287_NL },
         { NULL, NULL },
      },
      "287"
   },
   { NULL, NULL, NULL, NULL, NULL, NULL, {{0}}, NULL },
};
struct retro_core_options_v2 options_nl = {
   option_cats_nl,
   option_defs_nl
};

/* RETRO_LANGUAGE_NO */

#define CATEGORY_SYSTEM_LABEL_NO NULL
#define CATEGORY_SYSTEM_INFO_0_NO NULL
#define CATEGORY_VIDEO_LABEL_NO NULL
#define CATEGORY_VIDEO_INFO_0_NO NULL
#define CATEGORY_PGXP_LABEL_NO NULL
#define CATEGORY_PGXP_INFO_0_NO NULL
#define CATEGORY_OSD_LABEL_NO "Skjermvisning"
#define CATEGORY_OSD_INFO_0_NO NULL
#define CATEGORY_INPUT_LABEL_NO "Inndata"
#define CATEGORY_INPUT_INFO_0_NO NULL
#define CATEGORY_MEMCARDS_LABEL_NO NULL
#define CATEGORY_MEMCARDS_INFO_0_NO NULL
#define CATEGORY_HACKS_LABEL_NO NULL
#define CATEGORY_HACKS_INFO_0_NO NULL
#define BEETLE_OPT_CPU_FREQ_SCALE_LABEL_NO NULL
#define BEETLE_OPT_CPU_FREQ_SCALE_INFO_0_NO NULL
#define OPTION_VAL_50_NO NULL
#define OPTION_VAL_60_NO NULL
#define OPTION_VAL_70_NO NULL
#define OPTION_VAL_80_NO NULL
#define OPTION_VAL_90_NO NULL
#define OPTION_VAL_100_NO NULL
#define OPTION_VAL_110_NO NULL
#define OPTION_VAL_120_NO NULL
#define OPTION_VAL_130_NO NULL
#define OPTION_VAL_140_NO NULL
#define OPTION_VAL_150_NO NULL
#define OPTION_VAL_160_NO NULL
#define OPTION_VAL_170_NO NULL
#define OPTION_VAL_180_NO NULL
#define OPTION_VAL_190_NO NULL
#define OPTION_VAL_200_NO NULL
#define OPTION_VAL_210_NO NULL
#define OPTION_VAL_220_NO NULL
#define OPTION_VAL_230_NO NULL
#define OPTION_VAL_240_NO NULL
#define OPTION_VAL_250_NO NULL
#define OPTION_VAL_260_NO NULL
#define OPTION_VAL_270_NO NULL
#define OPTION_VAL_280_NO NULL
#define OPTION_VAL_290_NO NULL
#define OPTION_VAL_300_NO NULL
#define OPTION_VAL_310_NO NULL
#define OPTION_VAL_320_NO NULL
#define OPTION_VAL_330_NO NULL
#define OPTION_VAL_340_NO NULL
#define OPTION_VAL_350_NO NULL
#define OPTION_VAL_360_NO NULL
#define OPTION_VAL_370_NO NULL
#define OPTION_VAL_380_NO NULL
#define OPTION_VAL_390_NO NULL
#define OPTION_VAL_400_NO NULL
#define OPTION_VAL_410_NO NULL
#define OPTION_VAL_420_NO NULL
#define OPTION_VAL_430_NO NULL
#define OPTION_VAL_440_NO NULL
#define OPTION_VAL_450_NO NULL
#define OPTION_VAL_460_NO NULL
#define OPTION_VAL_470_NO NULL
#define OPTION_VAL_480_NO NULL
#define OPTION_VAL_490_NO NULL
#define OPTION_VAL_500_NO NULL
#define OPTION_VAL_510_NO NULL
#define OPTION_VAL_520_NO NULL
#define OPTION_VAL_530_NO NULL
#define OPTION_VAL_540_NO NULL
#define OPTION_VAL_550_NO NULL
#define OPTION_VAL_560_NO NULL
#define OPTION_VAL_570_NO NULL
#define OPTION_VAL_580_NO NULL
#define OPTION_VAL_590_NO NULL
#define OPTION_VAL_600_NO NULL
#define OPTION_VAL_610_NO NULL
#define OPTION_VAL_620_NO NULL
#define OPTION_VAL_630_NO NULL
#define OPTION_VAL_640_NO NULL
#define OPTION_VAL_650_NO NULL
#define OPTION_VAL_660_NO NULL
#define OPTION_VAL_670_NO NULL
#define OPTION_VAL_680_NO NULL
#define OPTION_VAL_690_NO NULL
#define OPTION_VAL_700_NO NULL
#define OPTION_VAL_710_NO NULL
#define OPTION_VAL_720_NO NULL
#define OPTION_VAL_730_NO NULL
#define OPTION_VAL_740_NO NULL
#define OPTION_VAL_750_NO NULL
#define BEETLE_OPT_GPU_OVERCLOCK_LABEL_NO NULL
#define BEETLE_OPT_GPU_OVERCLOCK_INFO_0_NO NULL
#define OPTION_VAL_1X_NATIVE_NO NULL
#define OPTION_VAL_2X_NO NULL
#define OPTION_VAL_4X_NO NULL
#define OPTION_VAL_8X_NO NULL
#define OPTION_VAL_16X_NO NULL
#define OPTION_VAL_32X_NO NULL
#define BEETLE_OPT_GTE_OVERCLOCK_LABEL_NO NULL
#define BEETLE_OPT_GTE_OVERCLOCK_INFO_0_NO NULL
#define BEETLE_OPT_CD_ACCESS_METHOD_LABEL_NO NULL
#define BEETLE_OPT_CD_ACCESS_METHOD_INFO_0_NO NULL
#define OPTION_VAL_SYNC_NO NULL
#define OPTION_VAL_ASYNC_NO NULL
#define OPTION_VAL_PRECACHE_NO NULL
#define BEETLE_OPT_CD_FASTLOAD_LABEL_NO NULL
#define BEETLE_OPT_CD_FASTLOAD_INFO_0_NO NULL
#define OPTION_VAL_2X_NATIVE_NO NULL
#define OPTION_VAL_6X_NO NULL
#define OPTION_VAL_10X_NO NULL
#define OPTION_VAL_12X_NO NULL
#define OPTION_VAL_14X_NO NULL
#define BEETLE_OPT_SKIP_BIOS_LABEL_NO NULL
#define BEETLE_OPT_SKIP_BIOS_INFO_0_NO NULL
#define BEETLE_OPT_OVERRIDE_BIOS_LABEL_NO NULL
#define BEETLE_OPT_OVERRIDE_BIOS_INFO_0_NO NULL
#define OPTION_VAL_PSXONPSP_NO NULL
#define OPTION_VAL_PS1_ROM_NO NULL
#define OPTION_VAL_OPENBIOS_NO NULL
#define BEETLE_OPT_REGION_LABEL_NO NULL
#define BEETLE_OPT_REGION_INFO_0_NO NULL
#define OPTION_VAL_AUTO_NO NULL
#define OPTION_VAL_NTSC_J_NO NULL
#define OPTION_VAL_NTSC_U_NO NULL
#define OPTION_VAL_PAL_NO NULL
#define BEETLE_OPT_INTERNAL_RESOLUTION_LABEL_NO NULL
#define BEETLE_OPT_INTERNAL_RESOLUTION_INFO_0_NO NULL
#define BEETLE_OPT_RENDERER_LABEL_NO NULL
#define BEETLE_OPT_RENDERER_INFO_0_NO NULL
#define OPTION_VAL_HARDWARE_NO NULL
#define OPTION_VAL_HARDWARE_GL_NO NULL
#define OPTION_VAL_HARDWARE_VK_NO NULL
#define OPTION_VAL_SOFTWARE_NO NULL
#define BEETLE_OPT_RENDERER_SOFTWARE_FB_LABEL_NO NULL
#define BEETLE_OPT_RENDERER_SOFTWARE_FB_INFO_0_NO NULL
#define BEETLE_OPT_DEPTH_LABEL_NO NULL
#define BEETLE_OPT_DEPTH_INFO_0_NO NULL
#define OPTION_VAL_16BPP_NATIVE_NO NULL
#define OPTION_VAL_32BPP_NO NULL
#define BEETLE_OPT_DITHER_MODE_LABEL_NO NULL
#define BEETLE_OPT_DITHER_MODE_INFO_0_NO NULL
#define OPTION_VAL_INTERNAL_RESOLUTION_NO "Intern oppløsning"
#define BEETLE_OPT_SCALED_UV_OFFSET_LABEL_NO NULL
#define BEETLE_OPT_SCALED_UV_OFFSET_INFO_0_NO NULL
#define BEETLE_OPT_FILTER_LABEL_NO NULL
#define BEETLE_OPT_FILTER_INFO_0_NO NULL
#define OPTION_VAL_NEAREST_NO NULL
#define OPTION_VAL_SABR_NO NULL
#define OPTION_VAL_XBR_NO NULL
#define OPTION_VAL_BILINEAR_NO "Bilineær"
#define OPTION_VAL_3_POINT_NO NULL
#define OPTION_VAL_JINC2_NO NULL
#define BEETLE_OPT_FILTER_EXCLUDE_SPRITE_LABEL_NO NULL
#define BEETLE_OPT_FILTER_EXCLUDE_SPRITE_INFO_0_NO NULL
#define OPTION_VAL_OPAQUE_NO NULL
#define OPTION_VAL_ALL_NO NULL
#define BEETLE_OPT_FILTER_EXCLUDE_2D_POLYGON_LABEL_NO NULL
#define BEETLE_OPT_FILTER_EXCLUDE_2D_POLYGON_INFO_0_NO NULL
#define BEETLE_OPT_ADAPTIVE_SMOOTHING_LABEL_NO NULL
#define BEETLE_OPT_ADAPTIVE_SMOOTHING_INFO_0_NO NULL
#define BEETLE_OPT_SUPER_SAMPLING_LABEL_NO NULL
#define BEETLE_OPT_SUPER_SAMPLING_INFO_0_NO NULL
#define BEETLE_OPT_MSAA_LABEL_NO NULL
#define BEETLE_OPT_MSAA_INFO_0_NO NULL
#define OPTION_VAL_1X_NO NULL
#define BEETLE_OPT_MDEC_YUV_LABEL_NO NULL
#define BEETLE_OPT_MDEC_YUV_INFO_0_NO NULL
#define BEETLE_OPT_TRACK_TEXTURES_LABEL_NO NULL
#define BEETLE_OPT_TRACK_TEXTURES_INFO_0_NO NULL
#define BEETLE_OPT_DUMP_TEXTURES_LABEL_NO NULL
#define BEETLE_OPT_DUMP_TEXTURES_INFO_0_NO NULL
#define BEETLE_OPT_REPLACE_TEXTURES_LABEL_NO NULL
#define BEETLE_OPT_REPLACE_TEXTURES_INFO_0_NO NULL
#define BEETLE_OPT_PGXP_MODE_LABEL_NO NULL
#define BEETLE_OPT_PGXP_MODE_INFO_0_NO NULL
#define OPTION_VAL_MEMORY_ONLY_NO NULL
#define OPTION_VAL_MEMORY_CPU_NO NULL
#define BEETLE_OPT_PGXP_2D_TOL_LABEL_NO NULL
#define BEETLE_OPT_PGXP_2D_TOL_INFO_0_NO NULL
#define OPTION_VAL_0PX_NO NULL
#define OPTION_VAL_1PX_NO NULL
#define OPTION_VAL_2PX_NO NULL
#define OPTION_VAL_3PX_NO NULL
#define OPTION_VAL_4PX_NO NULL
#define OPTION_VAL_5PX_NO NULL
#define OPTION_VAL_6PX_NO NULL
#define OPTION_VAL_7PX_NO NULL
#define OPTION_VAL_8PX_NO NULL
#define BEETLE_OPT_PGXP_NCLIP_LABEL_NO NULL
#define BEETLE_OPT_PGXP_NCLIP_INFO_0_NO NULL
#define BEETLE_OPT_PGXP_VERTEX_LABEL_NO NULL
#define BEETLE_OPT_PGXP_VERTEX_INFO_0_NO NULL
#define BEETLE_OPT_PGXP_TEXTURE_LABEL_NO NULL
#define BEETLE_OPT_PGXP_TEXTURE_INFO_0_NO NULL
#define BEETLE_OPT_DISPLAY_INTERNAL_FPS_LABEL_NO NULL
#define BEETLE_OPT_DISPLAY_INTERNAL_FPS_INFO_0_NO NULL
#define BEETLE_OPT_DISPLAY_OSD_LABEL_NO NULL
#define BEETLE_OPT_DISPLAY_OSD_INFO_0_NO NULL
#define BEETLE_OPT_DISPLAY_VRAM_LABEL_NO NULL
#define BEETLE_OPT_DISPLAY_VRAM_INFO_0_NO NULL
#define BEETLE_OPT_ANALOG_CALIBRATION_LABEL_NO NULL
#define BEETLE_OPT_ANALOG_CALIBRATION_INFO_0_NO NULL
#define BEETLE_OPT_ANALOG_TOGGLE_LABEL_NO NULL
#define BEETLE_OPT_ANALOG_TOGGLE_INFO_0_NO NULL
#define OPTION_VAL_ENABLED_ANALOG_NO NULL
#define BEETLE_OPT_ANALOG_TOGGLE_COMBO_LABEL_NO NULL
#define BEETLE_OPT_ANALOG_TOGGLE_COMBO_INFO_0_NO NULL
#define OPTION_VAL_L1_L2_R1_R2_START_SELECT_NO NULL
#define OPTION_VAL_L1_R1_SELECT_NO NULL
#define OPTION_VAL_L1_R1_START_NO NULL
#define OPTION_VAL_L1_R1_L3_NO NULL
#define OPTION_VAL_L1_R1_R3_NO NULL
#define OPTION_VAL_L2_R2_SELECT_NO NULL
#define OPTION_VAL_L2_R2_START_NO NULL
#define OPTION_VAL_L2_R2_L3_NO NULL
#define OPTION_VAL_L2_R2_R3_NO NULL
#define OPTION_VAL_L3_R3_NO NULL
#define BEETLE_OPT_ANALOG_TOGGLE_HOLD_LABEL_NO NULL
#define BEETLE_OPT_ANALOG_TOGGLE_HOLD_INFO_0_NO NULL
#define OPTION_VAL_0_NO NULL
#define OPTION_VAL_1_NO NULL
#define OPTION_VAL_2_NO NULL
#define OPTION_VAL_3_NO NULL
#define OPTION_VAL_4_NO NULL
#define OPTION_VAL_5_NO NULL
#define BEETLE_OPT_ENABLE_MULTITAP_PORT1_LABEL_NO NULL
#define BEETLE_OPT_ENABLE_MULTITAP_PORT1_INFO_0_NO NULL
#define BEETLE_OPT_ENABLE_MULTITAP_PORT2_LABEL_NO NULL
#define BEETLE_OPT_ENABLE_MULTITAP_PORT2_INFO_0_NO NULL
#define BEETLE_OPT_GUN_INPUT_MODE_LABEL_NO NULL
#define BEETLE_OPT_GUN_INPUT_MODE_INFO_0_NO NULL
#define OPTION_VAL_LIGHTGUN_NO NULL
#define OPTION_VAL_TOUCHSCREEN_NO NULL
#define BEETLE_OPT_GUN_CURSOR_LABEL_NO NULL
#define BEETLE_OPT_GUN_CURSOR_INFO_0_NO NULL
#define OPTION_VAL_CROSS_NO NULL
#define OPTION_VAL_DOT_NO NULL
#define BEETLE_OPT_CROSSHAIR_COLOR_P1_LABEL_NO NULL
#define BEETLE_OPT_CROSSHAIR_COLOR_P1_INFO_0_NO NULL
#define OPTION_VAL_RED_NO "Rød"
#define OPTION_VAL_BLUE_NO "Blå"
#define OPTION_VAL_GREEN_NO "Grønn"
#define OPTION_VAL_ORANGE_NO NULL
#define OPTION_VAL_YELLOW_NO "Gul"
#define OPTION_VAL_CYAN_NO NULL
#define OPTION_VAL_PINK_NO NULL
#define OPTION_VAL_PURPLE_NO NULL
#define OPTION_VAL_BLACK_NO NULL
#define OPTION_VAL_WHITE_NO NULL
#define BEETLE_OPT_CROSSHAIR_COLOR_P2_LABEL_NO NULL
#define BEETLE_OPT_CROSSHAIR_COLOR_P2_INFO_0_NO NULL
#define BEETLE_OPT_MOUSE_SENSITIVITY_LABEL_NO NULL
#define BEETLE_OPT_MOUSE_SENSITIVITY_INFO_0_NO NULL
#define OPTION_VAL_5_O42_NO NULL
#define OPTION_VAL_10_NO NULL
#define OPTION_VAL_15_NO NULL
#define OPTION_VAL_20_NO NULL
#define OPTION_VAL_25_NO NULL
#define OPTION_VAL_30_NO NULL
#define OPTION_VAL_35_NO NULL
#define OPTION_VAL_40_NO NULL
#define OPTION_VAL_45_NO NULL
#define OPTION_VAL_55_NO NULL
#define OPTION_VAL_65_NO NULL
#define OPTION_VAL_75_NO NULL
#define OPTION_VAL_85_NO NULL
#define OPTION_VAL_95_NO NULL
#define OPTION_VAL_100_O42_NO NULL
#define OPTION_VAL_105_NO NULL
#define OPTION_VAL_115_NO NULL
#define OPTION_VAL_125_NO NULL
#define OPTION_VAL_135_NO NULL
#define OPTION_VAL_145_NO NULL
#define OPTION_VAL_155_NO NULL
#define OPTION_VAL_165_NO NULL
#define OPTION_VAL_175_NO NULL
#define OPTION_VAL_185_NO NULL
#define OPTION_VAL_195_NO NULL
#define BEETLE_OPT_NEGCON_RESPONSE_LABEL_NO NULL
#define BEETLE_OPT_NEGCON_RESPONSE_INFO_0_NO NULL
#define OPTION_VAL_LINEAR_NO NULL
#define OPTION_VAL_QUADRATIC_NO NULL
#define OPTION_VAL_CUBIC_NO NULL
#define BEETLE_OPT_NEGCON_DEADZONE_LABEL_NO NULL
#define BEETLE_OPT_NEGCON_DEADZONE_INFO_0_NO NULL
#define OPTION_VAL_0_O44_NO NULL
#define BEETLE_OPT_USE_MEDNAFEN_MEMCARD0_METHOD_LABEL_NO NULL
#define BEETLE_OPT_USE_MEDNAFEN_MEMCARD0_METHOD_INFO_0_NO NULL
#define OPTION_VAL_LIBRETRO_NO NULL
#define OPTION_VAL_MEDNAFEN_NO NULL
#define BEETLE_OPT_ENABLE_MEMCARD1_LABEL_NO NULL
#define BEETLE_OPT_ENABLE_MEMCARD1_INFO_0_NO NULL
#define BEETLE_OPT_SHARED_MEMORY_CARDS_LABEL_NO NULL
#define BEETLE_OPT_SHARED_MEMORY_CARDS_INFO_0_NO NULL
#define BEETLE_OPT_MEMCARD_LEFT_INDEX_LABEL_NO NULL
#define BEETLE_OPT_MEMCARD_LEFT_INDEX_INFO_0_NO NULL
#define OPTION_VAL_0_O48_NO NULL
#define BEETLE_OPT_MEMCARD_RIGHT_INDEX_LABEL_NO NULL
#define BEETLE_OPT_MEMCARD_RIGHT_INDEX_INFO_0_NO NULL
#define OPTION_VAL_1_O49_NO NULL
#define BEETLE_OPT_LINE_RENDER_LABEL_NO NULL
#define BEETLE_OPT_LINE_RENDER_INFO_0_NO NULL
#define OPTION_VAL_DEFAULT_NO NULL
#define OPTION_VAL_AGGRESSIVE_NO NULL
#define BEETLE_OPT_WIDESCREEN_HACK_LABEL_NO NULL
#define BEETLE_OPT_WIDESCREEN_HACK_INFO_0_NO NULL
#define BEETLE_OPT_WIDESCREEN_HACK_ASPECT_RATIO_LABEL_NO NULL
#define BEETLE_OPT_WIDESCREEN_HACK_ASPECT_RATIO_INFO_0_NO NULL
#define OPTION_VAL_16_9_NO NULL
#define OPTION_VAL_16_10_NO NULL
#define OPTION_VAL_18_9_NO NULL
#define OPTION_VAL_19_9_NO NULL
#define OPTION_VAL_20_9_NO NULL
#define OPTION_VAL_21_9_NO NULL
#define OPTION_VAL_32_9_NO NULL
#define BEETLE_OPT_CPU_DYNAREC_LABEL_NO NULL
#define BEETLE_OPT_CPU_DYNAREC_INFO_0_NO NULL
#define OPTION_VAL_DISABLED_NO NULL
#define OPTION_VAL_EXECUTE_NO NULL
#define OPTION_VAL_EXECUTE_ONE_NO NULL
#define OPTION_VAL_RUN_INTERPRETER_NO NULL
#define BEETLE_OPT_DYNAREC_INVALIDATE_LABEL_NO NULL
#define BEETLE_OPT_DYNAREC_INVALIDATE_INFO_0_NO NULL
#define OPTION_VAL_FULL_NO NULL
#define OPTION_VAL_DMA_NO NULL
#define BEETLE_OPT_DYNAREC_EVENTCYCLES_LABEL_NO NULL
#define BEETLE_OPT_DYNAREC_EVENTCYCLES_INFO_0_NO NULL
#define OPTION_VAL_128_NO NULL
#define BEETLE_OPT_DYNAREC_SPU_SAMPLES_LABEL_NO NULL
#define BEETLE_OPT_DYNAREC_SPU_SAMPLES_INFO_0_NO NULL
#define BEETLE_OPT_CORE_TIMING_FPS_LABEL_NO NULL
#define BEETLE_OPT_CORE_TIMING_FPS_INFO_0_NO NULL
#define OPTION_VAL_AUTO_TOGGLE_NO NULL
#define OPTION_VAL_FORCE_PROGRESSIVE_NO NULL
#define OPTION_VAL_FORCE_INTERLACED_NO NULL
#define BEETLE_OPT_DEINTERLACER_LABEL_NO NULL
#define BEETLE_OPT_DEINTERLACER_INFO_0_NO NULL
#define OPTION_VAL_WEAVE_NO NULL
#define OPTION_VAL_BOB_NO NULL
#define OPTION_VAL_BOB_OFFSET_NO NULL
#define OPTION_VAL_FASTMAD_NO NULL
#define BEETLE_OPT_FRAME_DUPING_LABEL_NO NULL
#define BEETLE_OPT_FRAME_DUPING_INFO_0_NO NULL
#define BEETLE_OPT_PAL_VIDEO_TIMING_OVERRIDE_LABEL_NO NULL
#define BEETLE_OPT_PAL_VIDEO_TIMING_OVERRIDE_INFO_0_NO NULL
#define BEETLE_OPT_ASPECT_RATIO_LABEL_NO NULL
#define BEETLE_OPT_ASPECT_RATIO_INFO_0_NO NULL
#define OPTION_VAL_CORRECTED_NO NULL
#define OPTION_VAL_UNCORRECTED_NO NULL
#define OPTION_VAL_4_3_NO NULL
#define OPTION_VAL_NTSC_NO NULL
#define BEETLE_OPT_CROP_OVERSCAN_LABEL_NO NULL
#define BEETLE_OPT_CROP_OVERSCAN_INFO_0_NO NULL
#define OPTION_VAL_STATIC_NO NULL
#define OPTION_VAL_SMART_NO NULL
#define BEETLE_OPT_IMAGE_CROP_LABEL_NO NULL
#define BEETLE_OPT_IMAGE_CROP_INFO_0_NO NULL
#define OPTION_VAL_9PX_NO NULL
#define OPTION_VAL_10PX_NO NULL
#define OPTION_VAL_11PX_NO NULL
#define OPTION_VAL_12PX_NO NULL
#define OPTION_VAL_13PX_NO NULL
#define OPTION_VAL_14PX_NO NULL
#define OPTION_VAL_15PX_NO NULL
#define OPTION_VAL_16PX_NO NULL
#define OPTION_VAL_17PX_NO NULL
#define OPTION_VAL_18PX_NO NULL
#define OPTION_VAL_19PX_NO NULL
#define OPTION_VAL_20PX_NO NULL
#define BEETLE_OPT_IMAGE_OFFSET_LABEL_NO NULL
#define BEETLE_OPT_IMAGE_OFFSET_INFO_0_NO NULL
#define OPTION_VAL_12PX_O64_NO NULL
#define OPTION_VAL_11PX_O64_NO NULL
#define OPTION_VAL_10PX_O64_NO NULL
#define OPTION_VAL_9PX_O64_NO NULL
#define OPTION_VAL_8PX_O64_NO NULL
#define OPTION_VAL_7PX_O64_NO NULL
#define OPTION_VAL_6PX_O64_NO NULL
#define OPTION_VAL_5PX_O64_NO NULL
#define OPTION_VAL_4PX_O64_NO NULL
#define OPTION_VAL_3PX_O64_NO NULL
#define OPTION_VAL_2PX_O64_NO NULL
#define OPTION_VAL_1PX_O64_NO NULL
#define OPTION_VAL_1PX_O64_0_NO NULL
#define OPTION_VAL_2PX_O64_0_NO NULL
#define OPTION_VAL_3PX_O64_0_NO NULL
#define OPTION_VAL_4PX_O64_0_NO NULL
#define OPTION_VAL_5PX_O64_0_NO NULL
#define OPTION_VAL_6PX_O64_0_NO NULL
#define OPTION_VAL_7PX_O64_0_NO NULL
#define OPTION_VAL_8PX_O64_0_NO NULL
#define OPTION_VAL_9PX_O64_0_NO NULL
#define OPTION_VAL_10PX_O64_0_NO NULL
#define OPTION_VAL_11PX_O64_0_NO NULL
#define OPTION_VAL_12PX_O64_0_NO NULL
#define BEETLE_OPT_IMAGE_OFFSET_CYCLES_LABEL_NO NULL
#define BEETLE_OPT_IMAGE_OFFSET_CYCLES_INFO_0_NO NULL
#define BEETLE_OPT_INITIAL_SCANLINE_LABEL_NO NULL
#define BEETLE_OPT_INITIAL_SCANLINE_INFO_0_NO NULL
#define BEETLE_OPT_LAST_SCANLINE_LABEL_NO NULL
#define BEETLE_OPT_LAST_SCANLINE_INFO_0_NO NULL
#define OPTION_VAL_239_NO NULL
#define BEETLE_OPT_INITIAL_SCANLINE_PAL_LABEL_NO NULL
#define BEETLE_OPT_INITIAL_SCANLINE_PAL_INFO_0_NO NULL
#define BEETLE_OPT_LAST_SCANLINE_PAL_LABEL_NO NULL
#define BEETLE_OPT_LAST_SCANLINE_PAL_INFO_0_NO NULL
#define OPTION_VAL_287_NO NULL

struct retro_core_option_v2_category option_cats_no[] = {
   {
      "system",
      CATEGORY_SYSTEM_LABEL_NO,
      CATEGORY_SYSTEM_INFO_0_NO
   },
   {
      "video",
      CATEGORY_VIDEO_LABEL_NO,
      CATEGORY_VIDEO_INFO_0_NO
   },
   {
      "pgxp",
      CATEGORY_PGXP_LABEL_NO,
      CATEGORY_PGXP_INFO_0_NO
   },
   {
      "osd",
      CATEGORY_OSD_LABEL_NO,
      CATEGORY_OSD_INFO_0_NO
   },
   {
      "input",
      CATEGORY_INPUT_LABEL_NO,
      CATEGORY_INPUT_INFO_0_NO
   },
   {
      "memcards",
      CATEGORY_MEMCARDS_LABEL_NO,
      CATEGORY_MEMCARDS_INFO_0_NO
   },
   {
      "hacks",
      CATEGORY_HACKS_LABEL_NO,
      CATEGORY_HACKS_INFO_0_NO
   },
   { NULL, NULL, NULL },
};
struct retro_core_option_v2_definition option_defs_no[] = {
   {
      BEETLE_OPT(cpu_freq_scale),
      BEETLE_OPT_CPU_FREQ_SCALE_LABEL_NO,
      NULL,
      BEETLE_OPT_CPU_FREQ_SCALE_INFO_0_NO,
      NULL,
      "system",
      {
         { "50%",           OPTION_VAL_50_NO },
         { "60%",           OPTION_VAL_60_NO },
         { "70%",           OPTION_VAL_70_NO },
         { "80%",           OPTION_VAL_80_NO },
         { "90%",           OPTION_VAL_90_NO },
         { "100%",          OPTION_VAL_100_NO },
         { "110%",          OPTION_VAL_110_NO },
         { "120%",          OPTION_VAL_120_NO },
         { "130%",          OPTION_VAL_130_NO },
         { "140%",          OPTION_VAL_140_NO },
         { "150%",          OPTION_VAL_150_NO },
         { "160%",          OPTION_VAL_160_NO },
         { "170%",          OPTION_VAL_170_NO },
         { "180%",          OPTION_VAL_180_NO },
         { "190%",          OPTION_VAL_190_NO },
         { "200%",          OPTION_VAL_200_NO },
         { "210%",          OPTION_VAL_210_NO },
         { "220%",          OPTION_VAL_220_NO },
         { "230%",          OPTION_VAL_230_NO },
         { "240%",          OPTION_VAL_240_NO },
         { "250%",          OPTION_VAL_250_NO },
         { "260%",          OPTION_VAL_260_NO },
         { "270%",          OPTION_VAL_270_NO },
         { "280%",          OPTION_VAL_280_NO },
         { "290%",          OPTION_VAL_290_NO },
         { "300%",          OPTION_VAL_300_NO },
         { "310%",          OPTION_VAL_310_NO },
         { "320%",          OPTION_VAL_320_NO },
         { "330%",          OPTION_VAL_330_NO },
         { "340%",          OPTION_VAL_340_NO },
         { "350%",          OPTION_VAL_350_NO },
         { "360%",          OPTION_VAL_360_NO },
         { "370%",          OPTION_VAL_370_NO },
         { "380%",          OPTION_VAL_380_NO },
         { "390%",          OPTION_VAL_390_NO },
         { "400%",          OPTION_VAL_400_NO },
         { "410%",          OPTION_VAL_410_NO },
         { "420%",          OPTION_VAL_420_NO },
         { "430%",          OPTION_VAL_430_NO },
         { "440%",          OPTION_VAL_440_NO },
         { "450%",          OPTION_VAL_450_NO },
         { "460%",          OPTION_VAL_460_NO },
         { "470%",          OPTION_VAL_470_NO },
         { "480%",          OPTION_VAL_480_NO },
         { "490%",          OPTION_VAL_490_NO },
         { "500%",          OPTION_VAL_500_NO },
         { "510%",          OPTION_VAL_510_NO },
         { "520%",          OPTION_VAL_520_NO },
         { "530%",          OPTION_VAL_530_NO },
         { "540%",          OPTION_VAL_540_NO },
         { "550%",          OPTION_VAL_550_NO },
         { "560%",          OPTION_VAL_560_NO },
         { "570%",          OPTION_VAL_570_NO },
         { "580%",          OPTION_VAL_580_NO },
         { "590%",          OPTION_VAL_590_NO },
         { "600%",          OPTION_VAL_600_NO },
         { "610%",          OPTION_VAL_610_NO },
         { "620%",          OPTION_VAL_620_NO },
         { "630%",          OPTION_VAL_630_NO },
         { "640%",          OPTION_VAL_640_NO },
         { "650%",          OPTION_VAL_650_NO },
         { "660%",          OPTION_VAL_660_NO },
         { "670%",          OPTION_VAL_670_NO },
         { "680%",          OPTION_VAL_680_NO },
         { "690%",          OPTION_VAL_690_NO },
         { "700%",          OPTION_VAL_700_NO },
         { "710%",          OPTION_VAL_710_NO },
         { "720%",          OPTION_VAL_720_NO },
         { "730%",          OPTION_VAL_730_NO },
         { "740%",          OPTION_VAL_740_NO },
         { "750%",          OPTION_VAL_750_NO },
         { NULL, NULL },
      },
      "100%"
   },
   {
      BEETLE_OPT(gpu_overclock),
      BEETLE_OPT_GPU_OVERCLOCK_LABEL_NO,
      NULL,
      BEETLE_OPT_GPU_OVERCLOCK_INFO_0_NO,
      NULL,
      "system",
      {
         { "1x(native)", OPTION_VAL_1X_NATIVE_NO },
         { "2x",         OPTION_VAL_2X_NO },
         { "4x",         OPTION_VAL_4X_NO },
         { "8x",         OPTION_VAL_8X_NO },
         { "16x",        OPTION_VAL_16X_NO },
         { "32x",        OPTION_VAL_32X_NO },
         { NULL, NULL },
      },
      "1x(native)"
   },
   {
      BEETLE_OPT(gte_overclock),
      BEETLE_OPT_GTE_OVERCLOCK_LABEL_NO,
      NULL,
      BEETLE_OPT_GTE_OVERCLOCK_INFO_0_NO,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
#ifndef EMSCRIPTEN
   {
      BEETLE_OPT(cd_access_method),
      BEETLE_OPT_CD_ACCESS_METHOD_LABEL_NO,
      NULL,
      BEETLE_OPT_CD_ACCESS_METHOD_INFO_0_NO,
      NULL,
      "system",
      {
         { "sync",     OPTION_VAL_SYNC_NO },
         { "async",    OPTION_VAL_ASYNC_NO },
         { "precache", OPTION_VAL_PRECACHE_NO },
         { NULL, NULL },
      },
      "sync"
   },
#endif
   {
      BEETLE_OPT(cd_fastload),
      BEETLE_OPT_CD_FASTLOAD_LABEL_NO,
      NULL,
      BEETLE_OPT_CD_FASTLOAD_INFO_0_NO,
      NULL,
      "system",
      {
         { "2x(native)", OPTION_VAL_2X_NATIVE_NO },
         { "4x",          OPTION_VAL_4X_NO },
         { "6x",          OPTION_VAL_6X_NO },
         { "8x",          OPTION_VAL_8X_NO },
         { "10x",         OPTION_VAL_10X_NO },
         { "12x",         OPTION_VAL_12X_NO },
         { "14x",         OPTION_VAL_14X_NO },
         { NULL, NULL },
      },
      "2x(native)"
   },
   {
      BEETLE_OPT(skip_bios),
      BEETLE_OPT_SKIP_BIOS_LABEL_NO,
      NULL,
      BEETLE_OPT_SKIP_BIOS_INFO_0_NO,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(override_bios),
      BEETLE_OPT_OVERRIDE_BIOS_LABEL_NO,
      NULL,
      BEETLE_OPT_OVERRIDE_BIOS_INFO_0_NO,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "psxonpsp", OPTION_VAL_PSXONPSP_NO },
         { "ps1_rom",  OPTION_VAL_PS1_ROM_NO },
         { "openbios", OPTION_VAL_OPENBIOS_NO     },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(region),
      BEETLE_OPT_REGION_LABEL_NO,
      NULL,
      BEETLE_OPT_REGION_INFO_0_NO,
      NULL,
      "system",
      {
         { "auto",   OPTION_VAL_AUTO_NO },
         { "ntsc-j", OPTION_VAL_NTSC_J_NO },
         { "ntsc-u", OPTION_VAL_NTSC_U_NO },
         { "pal",    OPTION_VAL_PAL_NO },
         { NULL, NULL },
      },
      "auto"
   },
   {
      BEETLE_OPT(internal_resolution),
      BEETLE_OPT_INTERNAL_RESOLUTION_LABEL_NO,
      NULL,
      BEETLE_OPT_INTERNAL_RESOLUTION_INFO_0_NO,
      NULL,
      "video",
      {
         { "1x(native)", OPTION_VAL_1X_NATIVE_NO },
         { "2x",         OPTION_VAL_2X_NO },
         { "4x",         OPTION_VAL_4X_NO },
         { "8x",         OPTION_VAL_8X_NO },
         { "16x",        OPTION_VAL_16X_NO },
         { NULL, NULL },
      },
      "1x(native)"
   },
#if defined(HAVE_OPENGL) || defined(HAVE_OPENGLES) || defined(HAVE_VULKAN)
   {
      BEETLE_OPT(renderer),
      BEETLE_OPT_RENDERER_LABEL_NO,
      NULL,
      BEETLE_OPT_RENDERER_INFO_0_NO,
      NULL,
      "video",
      {
         { "hardware",    OPTION_VAL_HARDWARE_NO },
#if defined(HAVE_OPENGL) || defined(HAVE_OPENGLES)
         { "hardware_gl", OPTION_VAL_HARDWARE_GL_NO },
#endif
#if defined(HAVE_VULKAN)
         { "hardware_vk", OPTION_VAL_HARDWARE_VK_NO },
#endif
         { "software",    OPTION_VAL_SOFTWARE_NO },
         { NULL, NULL },
      },
      "hardware"
   },
   {
      BEETLE_OPT(renderer_software_fb),
      BEETLE_OPT_RENDERER_SOFTWARE_FB_LABEL_NO,
      NULL,
      BEETLE_OPT_RENDERER_SOFTWARE_FB_INFO_0_NO,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled"
   },
#endif
#if defined(HAVE_OPENGL) || defined(HAVE_OPENGLES)
   {
      BEETLE_OPT(depth),
      BEETLE_OPT_DEPTH_LABEL_NO,
      NULL,
      BEETLE_OPT_DEPTH_INFO_0_NO,
      NULL,
      "video",
      {
         { "16bpp(native)", OPTION_VAL_16BPP_NATIVE_NO },
         { "32bpp",         OPTION_VAL_32BPP_NO },
         { NULL, NULL },
      },
      "16bpp(native)"
   },
   // Sort of, it's more like 15-bit high color and 24-bit true color for visible output. The alpha channel is used for mask bit. Vulkan renderer uses ABGR1555_555 for 31 bits internal? FMVs are always 24-bit on all renderers like original hardware (BGR888, no alpha)
#endif
   {
      BEETLE_OPT(dither_mode),
      BEETLE_OPT_DITHER_MODE_LABEL_NO,
      NULL,
      BEETLE_OPT_DITHER_MODE_INFO_0_NO,
      NULL,
      "video",
      {
         { "1x(native)",          OPTION_VAL_1X_NATIVE_NO },
         { "internal resolution", OPTION_VAL_INTERNAL_RESOLUTION_NO },
         { "disabled",            NULL },
         { NULL, NULL },
      },
      "1x(native)"
   },
#ifdef HAVE_VULKAN
   {
      BEETLE_OPT(scaled_uv_offset),
      BEETLE_OPT_SCALED_UV_OFFSET_LABEL_NO,
      NULL,
      BEETLE_OPT_SCALED_UV_OFFSET_INFO_0_NO,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled"
   },
#endif
#if defined(HAVE_OPENGL) || defined(HAVE_OPENGLES) || defined(HAVE_VULKAN)
   {
      BEETLE_OPT(filter),
      BEETLE_OPT_FILTER_LABEL_NO,
      NULL,
      BEETLE_OPT_FILTER_INFO_0_NO,
      NULL,
      "video",
      {
         { "nearest",  OPTION_VAL_NEAREST_NO },
         { "SABR",     OPTION_VAL_SABR_NO },
         { "xBR",      OPTION_VAL_XBR_NO },
         { "bilinear", OPTION_VAL_BILINEAR_NO },
         { "3-point",  OPTION_VAL_3_POINT_NO },
         { "JINC2",    OPTION_VAL_JINC2_NO },
         { NULL, NULL },
      },
      "nearest"
   },
#ifdef HAVE_VULKAN
   {
      BEETLE_OPT(filter_exclude_sprite),
      BEETLE_OPT_FILTER_EXCLUDE_SPRITE_LABEL_NO,
      NULL,
      BEETLE_OPT_FILTER_EXCLUDE_SPRITE_INFO_0_NO,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "opaque", OPTION_VAL_OPAQUE_NO },
         { "all", OPTION_VAL_ALL_NO },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(filter_exclude_2d_polygon),
      BEETLE_OPT_FILTER_EXCLUDE_2D_POLYGON_LABEL_NO,
      NULL,
      BEETLE_OPT_FILTER_EXCLUDE_2D_POLYGON_INFO_0_NO,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "opaque", OPTION_VAL_OPAQUE_NO },
         { "all", OPTION_VAL_ALL_NO },
         { NULL, NULL },
      },
      "disabled"
   },
#endif
#endif
#ifdef HAVE_VULKAN
   {
      BEETLE_OPT(adaptive_smoothing),
      BEETLE_OPT_ADAPTIVE_SMOOTHING_LABEL_NO,
      NULL,
      BEETLE_OPT_ADAPTIVE_SMOOTHING_INFO_0_NO,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(super_sampling),
      BEETLE_OPT_SUPER_SAMPLING_LABEL_NO,
      NULL,
      BEETLE_OPT_SUPER_SAMPLING_INFO_0_NO,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(msaa),
      BEETLE_OPT_MSAA_LABEL_NO,
      NULL,
      BEETLE_OPT_MSAA_INFO_0_NO,
      NULL,
      "video",
      {
         { "1x",  OPTION_VAL_1X_NO },
         { "2x",  OPTION_VAL_2X_NO },
         { "4x",  OPTION_VAL_4X_NO },
         { "8x",  OPTION_VAL_8X_NO },
         { "16x", OPTION_VAL_16X_NO },
         { NULL, NULL },
      },
      "1x"
   },
   {
      BEETLE_OPT(mdec_yuv),
      BEETLE_OPT_MDEC_YUV_LABEL_NO,
      NULL,
      BEETLE_OPT_MDEC_YUV_INFO_0_NO,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(track_textures),
      BEETLE_OPT_TRACK_TEXTURES_LABEL_NO,
      NULL,
      BEETLE_OPT_TRACK_TEXTURES_INFO_0_NO,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
#ifdef TEXTURE_DUMPING_ENABLED
   {
      BEETLE_OPT(dump_textures),
      BEETLE_OPT_DUMP_TEXTURES_LABEL_NO,
      NULL,
      BEETLE_OPT_DUMP_TEXTURES_INFO_0_NO,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
#endif
   {
      BEETLE_OPT(replace_textures),
      BEETLE_OPT_REPLACE_TEXTURES_LABEL_NO,
      NULL,
      BEETLE_OPT_REPLACE_TEXTURES_INFO_0_NO,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
#endif
   {
      BEETLE_OPT(pgxp_mode),
      BEETLE_OPT_PGXP_MODE_LABEL_NO,
      NULL,
      BEETLE_OPT_PGXP_MODE_INFO_0_NO,
      NULL,
      "pgxp",
      {
         { "disabled",     NULL },
         { "memory only",  OPTION_VAL_MEMORY_ONLY_NO },
         { "memory + CPU", OPTION_VAL_MEMORY_CPU_NO },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(pgxp_2d_tol),
      BEETLE_OPT_PGXP_2D_TOL_LABEL_NO,
      NULL,
      BEETLE_OPT_PGXP_2D_TOL_INFO_0_NO,
      NULL,
      "pgxp",
      {
         { "disabled", NULL },
         { "0px", OPTION_VAL_0PX_NO },
         { "1px", OPTION_VAL_1PX_NO },
         { "2px", OPTION_VAL_2PX_NO },
         { "3px", OPTION_VAL_3PX_NO },
         { "4px", OPTION_VAL_4PX_NO },
         { "5px", OPTION_VAL_5PX_NO },
         { "6px", OPTION_VAL_6PX_NO },
         { "7px", OPTION_VAL_7PX_NO },
         { "8px", OPTION_VAL_8PX_NO },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(pgxp_nclip),
      BEETLE_OPT_PGXP_NCLIP_LABEL_NO,
      NULL,
      BEETLE_OPT_PGXP_NCLIP_INFO_0_NO,
      NULL,
      "pgxp",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
#if defined(HAVE_OPENGL) || defined(HAVE_OPENGLES) || defined(HAVE_VULKAN)
   {
      BEETLE_OPT(pgxp_vertex),
      BEETLE_OPT_PGXP_VERTEX_LABEL_NO,
      NULL,
      BEETLE_OPT_PGXP_VERTEX_INFO_0_NO,
      NULL,
      "pgxp",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
#endif
   {
      BEETLE_OPT(pgxp_texture),
      BEETLE_OPT_PGXP_TEXTURE_LABEL_NO,
      NULL,
      BEETLE_OPT_PGXP_TEXTURE_INFO_0_NO,
      NULL,
      "pgxp",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(display_internal_fps),
      BEETLE_OPT_DISPLAY_INTERNAL_FPS_LABEL_NO,
      NULL,
      BEETLE_OPT_DISPLAY_INTERNAL_FPS_INFO_0_NO,
      NULL,
      "osd",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(display_osd),
      BEETLE_OPT_DISPLAY_OSD_LABEL_NO,
      NULL,
      BEETLE_OPT_DISPLAY_OSD_INFO_0_NO,
      NULL,
      "osd",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled"
   },
#if defined(HAVE_OPENGL) || defined(HAVE_OPENGLES) || defined(HAVE_VULKAN)
   {
      BEETLE_OPT(display_vram),
      BEETLE_OPT_DISPLAY_VRAM_LABEL_NO,
      NULL,
      BEETLE_OPT_DISPLAY_VRAM_INFO_0_NO,
      NULL,
      "osd",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
#endif
   {
      BEETLE_OPT(analog_calibration),
      BEETLE_OPT_ANALOG_CALIBRATION_LABEL_NO,
      NULL,
      BEETLE_OPT_ANALOG_CALIBRATION_INFO_0_NO,
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(analog_toggle),
      BEETLE_OPT_ANALOG_TOGGLE_LABEL_NO,
      NULL,
      BEETLE_OPT_ANALOG_TOGGLE_INFO_0_NO,
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { "enabled-analog", OPTION_VAL_ENABLED_ANALOG_NO },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(analog_toggle_combo),
      BEETLE_OPT_ANALOG_TOGGLE_COMBO_LABEL_NO,
      NULL,
      BEETLE_OPT_ANALOG_TOGGLE_COMBO_INFO_0_NO,
      NULL,
      "input",
      {
         { "l1+l2+r1+r2+start+select", OPTION_VAL_L1_L2_R1_R2_START_SELECT_NO },
         { "l1+r1+select",             OPTION_VAL_L1_R1_SELECT_NO },
         { "l1+r1+start",              OPTION_VAL_L1_R1_START_NO },
         { "l1+r1+l3",                 OPTION_VAL_L1_R1_L3_NO },
         { "l1+r1+r3",                 OPTION_VAL_L1_R1_R3_NO },
         { "l2+r2+select",             OPTION_VAL_L2_R2_SELECT_NO },
         { "l2+r2+start",              OPTION_VAL_L2_R2_START_NO },
         { "l2+r2+l3",                 OPTION_VAL_L2_R2_L3_NO },
         { "l2+r2+r3",                 OPTION_VAL_L2_R2_R3_NO },
         { "l3+r3",                    OPTION_VAL_L3_R3_NO },
         { NULL, NULL },
      },
      "l1+r1+select"
   },
   {
      BEETLE_OPT(analog_toggle_hold),
      BEETLE_OPT_ANALOG_TOGGLE_HOLD_LABEL_NO,
      NULL,
      BEETLE_OPT_ANALOG_TOGGLE_HOLD_INFO_0_NO,
      NULL,
      "input",
      {
         { "0", OPTION_VAL_0_NO },
         { "1", OPTION_VAL_1_NO },
         { "2", OPTION_VAL_2_NO },
         { "3", OPTION_VAL_3_NO },
         { "4", OPTION_VAL_4_NO },
         { "5", OPTION_VAL_5_NO },
         { NULL, NULL },
      },
      "1"
   },
   {
      BEETLE_OPT(enable_multitap_port1),
      BEETLE_OPT_ENABLE_MULTITAP_PORT1_LABEL_NO,
      NULL,
      BEETLE_OPT_ENABLE_MULTITAP_PORT1_INFO_0_NO,
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(enable_multitap_port2),
      BEETLE_OPT_ENABLE_MULTITAP_PORT2_LABEL_NO,
      NULL,
      BEETLE_OPT_ENABLE_MULTITAP_PORT2_INFO_0_NO,
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(gun_input_mode),
      BEETLE_OPT_GUN_INPUT_MODE_LABEL_NO,
      NULL,
      BEETLE_OPT_GUN_INPUT_MODE_INFO_0_NO,
      NULL,
      "input",
      {
         { "lightgun",    OPTION_VAL_LIGHTGUN_NO },
         { "touchscreen", OPTION_VAL_TOUCHSCREEN_NO },
         { NULL, NULL },
      },
      "lightgun"
   },
   // Shouldn't the gun_input_mode just be Mouse vs. Touchscreen?
   {
      BEETLE_OPT(gun_cursor),
      BEETLE_OPT_GUN_CURSOR_LABEL_NO,
      NULL,
      BEETLE_OPT_GUN_CURSOR_INFO_0_NO,
      NULL,
      "input",
      {
         { "off",   "disabled" },
         { "cross", OPTION_VAL_CROSS_NO },
         { "dot",   OPTION_VAL_DOT_NO },
         { NULL, NULL },
      },
      "cross"
   },
   {
      BEETLE_OPT(crosshair_color_p1),
      BEETLE_OPT_CROSSHAIR_COLOR_P1_LABEL_NO,
      NULL,
      BEETLE_OPT_CROSSHAIR_COLOR_P1_INFO_0_NO,
      NULL,
      "input",
      {
         { "red",    OPTION_VAL_RED_NO },
         { "blue",   OPTION_VAL_BLUE_NO },
         { "green",  OPTION_VAL_GREEN_NO },
         { "orange", OPTION_VAL_ORANGE_NO },
         { "yellow", OPTION_VAL_YELLOW_NO },
         { "cyan",   OPTION_VAL_CYAN_NO },
         { "pink",   OPTION_VAL_PINK_NO },
         { "purple", OPTION_VAL_PURPLE_NO },
         { "black",  OPTION_VAL_BLACK_NO },
         { "white",  OPTION_VAL_WHITE_NO },
         { NULL, NULL },
      },
      "red"
   },
   {
      BEETLE_OPT(crosshair_color_p2),
      BEETLE_OPT_CROSSHAIR_COLOR_P2_LABEL_NO,
      NULL,
      BEETLE_OPT_CROSSHAIR_COLOR_P2_INFO_0_NO,
      NULL,
      "input",
      {
         { "blue",   OPTION_VAL_BLUE_NO },
         { "red",    OPTION_VAL_RED_NO },
         { "green",  OPTION_VAL_GREEN_NO },
         { "orange", OPTION_VAL_ORANGE_NO },
         { "yellow", OPTION_VAL_YELLOW_NO },
         { "cyan",   OPTION_VAL_CYAN_NO },
         { "pink",   OPTION_VAL_PINK_NO },
         { "purple", OPTION_VAL_PURPLE_NO },
         { "black",  OPTION_VAL_BLACK_NO },
         { "white",  OPTION_VAL_WHITE_NO },
         { NULL, NULL },
      },
      "blue"
   },
   {
      BEETLE_OPT(mouse_sensitivity),
      BEETLE_OPT_MOUSE_SENSITIVITY_LABEL_NO,
      NULL,
      BEETLE_OPT_MOUSE_SENSITIVITY_INFO_0_NO,
      NULL,
      "input",
      {
         { "5%",   OPTION_VAL_5_O42_NO },
         { "10%",  OPTION_VAL_10_NO },
         { "15%",  OPTION_VAL_15_NO },
         { "20%",  OPTION_VAL_20_NO },
         { "25%",  OPTION_VAL_25_NO },
         { "30%",  OPTION_VAL_30_NO },
         { "35%",  OPTION_VAL_35_NO },
         { "40%",  OPTION_VAL_40_NO },
         { "45%",  OPTION_VAL_45_NO },
         { "50%",  OPTION_VAL_50_NO },
         { "55%",  OPTION_VAL_55_NO },
         { "60%",  OPTION_VAL_60_NO },
         { "65%",  OPTION_VAL_65_NO },
         { "70%",  OPTION_VAL_70_NO },
         { "75%",  OPTION_VAL_75_NO },
         { "80%",  OPTION_VAL_80_NO },
         { "85%",  OPTION_VAL_85_NO },
         { "90%",  OPTION_VAL_90_NO },
         { "95%",  OPTION_VAL_95_NO },
         { "100%", OPTION_VAL_100_O42_NO },
         { "105%", OPTION_VAL_105_NO },
         { "110%", OPTION_VAL_110_NO },
         { "115%", OPTION_VAL_115_NO },
         { "120%", OPTION_VAL_120_NO },
         { "125%", OPTION_VAL_125_NO },
         { "130%", OPTION_VAL_130_NO },
         { "135%", OPTION_VAL_135_NO },
         { "140%", OPTION_VAL_140_NO },
         { "145%", OPTION_VAL_145_NO },
         { "150%", OPTION_VAL_150_NO },
         { "155%", OPTION_VAL_155_NO },
         { "160%", OPTION_VAL_160_NO },
         { "165%", OPTION_VAL_165_NO },
         { "170%", OPTION_VAL_170_NO },
         { "175%", OPTION_VAL_175_NO },
         { "180%", OPTION_VAL_180_NO },
         { "185%", OPTION_VAL_185_NO },
         { "190%", OPTION_VAL_190_NO },
         { "195%", OPTION_VAL_195_NO },
         { "200%", OPTION_VAL_200_NO },
         { NULL, NULL },
      },
      "100%"
   },
   {
      BEETLE_OPT(negcon_response),
      BEETLE_OPT_NEGCON_RESPONSE_LABEL_NO,
      NULL,
      BEETLE_OPT_NEGCON_RESPONSE_INFO_0_NO,
      NULL,
      "input",
      {
         { "linear",    OPTION_VAL_LINEAR_NO },
         { "quadratic", OPTION_VAL_QUADRATIC_NO },
         { "cubic",     OPTION_VAL_CUBIC_NO },
         { NULL, NULL },
      },
      "linear"
   },
   {
      BEETLE_OPT(negcon_deadzone),
      BEETLE_OPT_NEGCON_DEADZONE_LABEL_NO,
      NULL,
      BEETLE_OPT_NEGCON_DEADZONE_INFO_0_NO,
      NULL,
      "input",
      {
         { "0%",  OPTION_VAL_0_O44_NO },
         { "5%",  OPTION_VAL_5_O42_NO },
         { "10%", OPTION_VAL_10_NO },
         { "15%", OPTION_VAL_15_NO },
         { "20%", OPTION_VAL_20_NO },
         { "25%", OPTION_VAL_25_NO },
         { "30%", OPTION_VAL_30_NO },
         { NULL, NULL },
      },
      "0%"
   },
   {
      BEETLE_OPT(use_mednafen_memcard0_method),
      BEETLE_OPT_USE_MEDNAFEN_MEMCARD0_METHOD_LABEL_NO,
      NULL,
      BEETLE_OPT_USE_MEDNAFEN_MEMCARD0_METHOD_INFO_0_NO,
      NULL,
      "memcards",
      {
         { "libretro", OPTION_VAL_LIBRETRO_NO },
         { "mednafen", OPTION_VAL_MEDNAFEN_NO },
         { NULL, NULL },
      },
      "libretro"
   },
   {
      BEETLE_OPT(enable_memcard1),
      BEETLE_OPT_ENABLE_MEMCARD1_LABEL_NO,
      NULL,
      BEETLE_OPT_ENABLE_MEMCARD1_INFO_0_NO,
      NULL,
      "memcards",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(shared_memory_cards),
      BEETLE_OPT_SHARED_MEMORY_CARDS_LABEL_NO,
      NULL,
      BEETLE_OPT_SHARED_MEMORY_CARDS_INFO_0_NO,
      NULL,
      "memcards",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(memcard_left_index),
      BEETLE_OPT_MEMCARD_LEFT_INDEX_LABEL_NO,
      NULL,
      BEETLE_OPT_MEMCARD_LEFT_INDEX_INFO_0_NO,
      NULL,
      "memcards",
      {
         { "0",  OPTION_VAL_0_O48_NO },
         { "1",  NULL },
         { "2",  NULL },
         { "3",  NULL },
         { "4",  NULL },
         { "5",  NULL },
         { "6",  NULL },
         { "7",  NULL },
         { "8",  NULL },
         { "9",  NULL },
         { "10", NULL },
         { "11", NULL },
         { "12", NULL },
         { "13", NULL },
         { "14", NULL },
         { "15", NULL },
         { "16", NULL },
         { "17", NULL },
         { "18", NULL },
         { "19", NULL },
         { "20", NULL },
         { "21", NULL },
         { "22", NULL },
         { "23", NULL },
         { "24", NULL },
         { "25", NULL },
         { "26", NULL },
         { "27", NULL },
         { "28", NULL },
         { "29", NULL },
         { "30", NULL },
         { "31", NULL },
         { "32", NULL },
         { "33", NULL },
         { "34", NULL },
         { "35", NULL },
         { "36", NULL },
         { "37", NULL },
         { "38", NULL },
         { "39", NULL },
         { "40", NULL },
         { "41", NULL },
         { "42", NULL },
         { "43", NULL },
         { "44", NULL },
         { "45", NULL },
         { "46", NULL },
         { "47", NULL },
         { "48", NULL },
         { "49", NULL },
         { "50", NULL },
         { "51", NULL },
         { "52", NULL },
         { "53", NULL },
         { "54", NULL },
         { "55", NULL },
         { "56", NULL },
         { "57", NULL },
         { "58", NULL },
         { "59", NULL },
         { "60", NULL },
         { "61", NULL },
         { "62", NULL },
         { "63", NULL },
         { NULL, NULL },
      },
      "0"
   },
   {
      BEETLE_OPT(memcard_right_index),
      BEETLE_OPT_MEMCARD_RIGHT_INDEX_LABEL_NO,
      NULL,
      BEETLE_OPT_MEMCARD_RIGHT_INDEX_INFO_0_NO,
      NULL,
      "memcards",
      {
         { "0",  NULL },
         { "1",  OPTION_VAL_1_O49_NO },
         { "2",  NULL },
         { "3",  NULL },
         { "4",  NULL },
         { "5",  NULL },
         { "6",  NULL },
         { "7",  NULL },
         { "8",  NULL },
         { "9",  NULL },
         { "10", NULL },
         { "11", NULL },
         { "12", NULL },
         { "13", NULL },
         { "14", NULL },
         { "15", NULL },
         { "16", NULL },
         { "17", NULL },
         { "18", NULL },
         { "19", NULL },
         { "20", NULL },
         { "21", NULL },
         { "22", NULL },
         { "23", NULL },
         { "24", NULL },
         { "25", NULL },
         { "26", NULL },
         { "27", NULL },
         { "28", NULL },
         { "29", NULL },
         { "30", NULL },
         { "31", NULL },
         { "32", NULL },
         { "33", NULL },
         { "34", NULL },
         { "35", NULL },
         { "36", NULL },
         { "37", NULL },
         { "38", NULL },
         { "39", NULL },
         { "40", NULL },
         { "41", NULL },
         { "42", NULL },
         { "43", NULL },
         { "44", NULL },
         { "45", NULL },
         { "46", NULL },
         { "47", NULL },
         { "48", NULL },
         { "49", NULL },
         { "50", NULL },
         { "51", NULL },
         { "52", NULL },
         { "53", NULL },
         { "54", NULL },
         { "55", NULL },
         { "56", NULL },
         { "57", NULL },
         { "58", NULL },
         { "59", NULL },
         { "60", NULL },
         { "61", NULL },
         { "62", NULL },
         { "63", NULL },
         { NULL, NULL },
      },
      "1"
   },
   {
      BEETLE_OPT(line_render),
      BEETLE_OPT_LINE_RENDER_LABEL_NO,
      NULL,
      BEETLE_OPT_LINE_RENDER_INFO_0_NO,
      NULL,
      "hacks",
      {
         { "default",    OPTION_VAL_DEFAULT_NO },
         { "aggressive", OPTION_VAL_AGGRESSIVE_NO },
         { "disabled",   NULL },
         { NULL, NULL },
      },
      "default"
   },
   {
      BEETLE_OPT(widescreen_hack),
      BEETLE_OPT_WIDESCREEN_HACK_LABEL_NO,
      NULL,
      BEETLE_OPT_WIDESCREEN_HACK_INFO_0_NO,
      NULL,
      "hacks",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(widescreen_hack_aspect_ratio),
      BEETLE_OPT_WIDESCREEN_HACK_ASPECT_RATIO_LABEL_NO,
      NULL,
      BEETLE_OPT_WIDESCREEN_HACK_ASPECT_RATIO_INFO_0_NO,
      NULL,
      "hacks",
      {
         { "16:9",  OPTION_VAL_16_9_NO },
         { "16:10", OPTION_VAL_16_10_NO },
         { "18:9",  OPTION_VAL_18_9_NO },
         { "19:9",  OPTION_VAL_19_9_NO },
         { "20:9",  OPTION_VAL_20_9_NO },
         { "21:9",  OPTION_VAL_21_9_NO }, // 64:27
         { "32:9",  OPTION_VAL_32_9_NO },
         { NULL,    NULL },
      },
      "16:9"
   },
#if defined(HAVE_LIGHTREC)
   {
      BEETLE_OPT(cpu_dynarec),
      BEETLE_OPT_CPU_DYNAREC_LABEL_NO,
      NULL,
      BEETLE_OPT_CPU_DYNAREC_INFO_0_NO,
      NULL,
      "hacks",
      {
         { "disabled",        OPTION_VAL_DISABLED_NO },
         { "execute",         OPTION_VAL_EXECUTE_NO },
         { "execute_one",     OPTION_VAL_EXECUTE_ONE_NO },
         { "run_interpreter", OPTION_VAL_RUN_INTERPRETER_NO },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(dynarec_invalidate),
      BEETLE_OPT_DYNAREC_INVALIDATE_LABEL_NO,
      NULL,
      BEETLE_OPT_DYNAREC_INVALIDATE_INFO_0_NO,
      NULL,
      "hacks",
      {
         { "full", OPTION_VAL_FULL_NO },
         { "dma",  OPTION_VAL_DMA_NO },
         { NULL, NULL },
      },
      "full"
   },
   {
      BEETLE_OPT(dynarec_eventcycles),
      BEETLE_OPT_DYNAREC_EVENTCYCLES_LABEL_NO,
      NULL,
      BEETLE_OPT_DYNAREC_EVENTCYCLES_INFO_0_NO,
      NULL,
      "hacks",
      {
         { "128",  OPTION_VAL_128_NO },
         { "256",  NULL },
         { "384",  NULL },
         { "512",  NULL },
         { "640",  NULL },
         { "768",  NULL },
         { "896",  NULL },
         { "1024", NULL },
         { "1152", NULL },
         { "1280", NULL },
         { "1408", NULL },
         { "1536", NULL },
         { "1664", NULL },
         { "1792", NULL },
         { "1920", NULL },
         { "2048", NULL },
         { NULL, NULL },
      },
      "128"
   },
   {
      BEETLE_OPT(dynarec_spu_samples),
      BEETLE_OPT_DYNAREC_SPU_SAMPLES_LABEL_NO,
      NULL,
      BEETLE_OPT_DYNAREC_SPU_SAMPLES_INFO_0_NO,
      NULL,
      "hacks",
      {
         { "1",  OPTION_VAL_1_O49_NO },
         { "4",  NULL },
         { "16", NULL },
         { NULL, NULL },
      },
      "1"
   },
#endif
   {
      BEETLE_OPT(core_timing_fps),
      BEETLE_OPT_CORE_TIMING_FPS_LABEL_NO,
      NULL,
      BEETLE_OPT_CORE_TIMING_FPS_INFO_0_NO,
      NULL,
      "video",
      {
         { "auto_toggle",       OPTION_VAL_AUTO_TOGGLE_NO },
         { "force_progressive", OPTION_VAL_FORCE_PROGRESSIVE_NO },
         { "force_interlaced",  OPTION_VAL_FORCE_INTERLACED_NO },
      },
      "auto_toggle"
   },
   {
      BEETLE_OPT(deinterlacer),
      BEETLE_OPT_DEINTERLACER_LABEL_NO,
      NULL,
      BEETLE_OPT_DEINTERLACER_INFO_0_NO,
      NULL,
      "video",
      {
         { "weave",      OPTION_VAL_WEAVE_NO },
         { "bob",        OPTION_VAL_BOB_NO },
         { "bob_offset", OPTION_VAL_BOB_OFFSET_NO },
         { "fastmad",    OPTION_VAL_FASTMAD_NO },
         { "off",        "Off" },
         { NULL, NULL },
      },
      "weave"
   },
   {
      BEETLE_OPT(frame_duping),
      BEETLE_OPT_FRAME_DUPING_LABEL_NO,
      NULL,
      BEETLE_OPT_FRAME_DUPING_INFO_0_NO,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(pal_video_timing_override),
      BEETLE_OPT_PAL_VIDEO_TIMING_OVERRIDE_LABEL_NO,
      NULL,
      BEETLE_OPT_PAL_VIDEO_TIMING_OVERRIDE_INFO_0_NO,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(aspect_ratio),
      BEETLE_OPT_ASPECT_RATIO_LABEL_NO,
      NULL,
      BEETLE_OPT_ASPECT_RATIO_INFO_0_NO,
      NULL,
      "video",
      {
         { "corrected", OPTION_VAL_CORRECTED_NO },
         { "uncorrected", OPTION_VAL_UNCORRECTED_NO },
         { "4:3",  OPTION_VAL_4_3_NO },
         { "ntsc", OPTION_VAL_NTSC_NO },
      },
      "corrected"
   },
   {
      BEETLE_OPT(crop_overscan),
      BEETLE_OPT_CROP_OVERSCAN_LABEL_NO,
      NULL,
      BEETLE_OPT_CROP_OVERSCAN_INFO_0_NO,
      NULL,
      "video",
      {
         { "disabled", "Disabled" },
         { "static",   OPTION_VAL_STATIC_NO },
         { "smart",    OPTION_VAL_SMART_NO },
         { NULL, NULL },
      },
      "smart"
   },
   {
      BEETLE_OPT(image_crop),
      BEETLE_OPT_IMAGE_CROP_LABEL_NO,
      NULL,
      BEETLE_OPT_IMAGE_CROP_INFO_0_NO,
      NULL,
      "video",
      {
         { "disabled", "0" },
         { "1px",      OPTION_VAL_1PX_NO },
         { "2px",      OPTION_VAL_2PX_NO },
         { "3px",      OPTION_VAL_3PX_NO },
         { "4px",      OPTION_VAL_4PX_NO },
         { "5px",      OPTION_VAL_5PX_NO },
         { "6px",      OPTION_VAL_6PX_NO },
         { "7px",      OPTION_VAL_7PX_NO },
         { "8px",      OPTION_VAL_8PX_NO },
         { "9px",      OPTION_VAL_9PX_NO },
         { "10px",     OPTION_VAL_10PX_NO },
         { "11px",     OPTION_VAL_11PX_NO },
         { "12px",     OPTION_VAL_12PX_NO },
         { "13px",     OPTION_VAL_13PX_NO },
         { "14px",     OPTION_VAL_14PX_NO },
         { "15px",     OPTION_VAL_15PX_NO },
         { "16px",     OPTION_VAL_16PX_NO },
         { "17px",     OPTION_VAL_17PX_NO },
         { "18px",     OPTION_VAL_18PX_NO },
         { "19px",     OPTION_VAL_19PX_NO },
         { "20px",     OPTION_VAL_20PX_NO },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(image_offset),
      BEETLE_OPT_IMAGE_OFFSET_LABEL_NO,
      NULL,
      BEETLE_OPT_IMAGE_OFFSET_INFO_0_NO,
      NULL,
      "video",
      {
         { "-12px",    OPTION_VAL_12PX_O64_NO },
         { "-11px",    OPTION_VAL_11PX_O64_NO },
         { "-10px",    OPTION_VAL_10PX_O64_NO },
         { "-9px",     OPTION_VAL_9PX_O64_NO },
         { "-8px",     OPTION_VAL_8PX_O64_NO },
         { "-7px",     OPTION_VAL_7PX_O64_NO },
         { "-6px",     OPTION_VAL_6PX_O64_NO },
         { "-5px",     OPTION_VAL_5PX_O64_NO },
         { "-4px",     OPTION_VAL_4PX_O64_NO },
         { "-3px",     OPTION_VAL_3PX_O64_NO },
         { "-2px",     OPTION_VAL_2PX_O64_NO },
         { "-1px",     OPTION_VAL_1PX_O64_NO },
         { "disabled", "0" },
         { "+1px",     OPTION_VAL_1PX_O64_0_NO },
         { "+2px",     OPTION_VAL_2PX_O64_0_NO },
         { "+3px",     OPTION_VAL_3PX_O64_0_NO },
         { "+4px",     OPTION_VAL_4PX_O64_0_NO },
         { "+5px",     OPTION_VAL_5PX_O64_0_NO },
         { "+6px",     OPTION_VAL_6PX_O64_0_NO },
         { "+7px",     OPTION_VAL_7PX_O64_0_NO },
         { "+8px",     OPTION_VAL_8PX_O64_0_NO },
         { "+9px",     OPTION_VAL_9PX_O64_0_NO },
         { "+10px",    OPTION_VAL_10PX_O64_0_NO },
         { "+11px",    OPTION_VAL_11PX_O64_0_NO },
         { "+12px",    OPTION_VAL_12PX_O64_0_NO },
         { NULL, NULL },
      },
      "disabled"
   },
#if defined(HAVE_OPENGL) || defined(HAVE_OPENGLES) || defined(HAVE_VULKAN)
   {
      BEETLE_OPT(image_offset_cycles),
      BEETLE_OPT_IMAGE_OFFSET_CYCLES_LABEL_NO,
      NULL,
      BEETLE_OPT_IMAGE_OFFSET_CYCLES_INFO_0_NO,
      NULL,
      "video",
      {
         { "-40",      NULL },
         { "-39",      NULL },
         { "-38",      NULL },
         { "-37",      NULL },
         { "-36",      NULL },
         { "-35",      NULL },
         { "-34",      NULL },
         { "-33",      NULL },
         { "-32",      NULL },
         { "-31",      NULL },
         { "-30",      NULL },
         { "-29",      NULL },
         { "-28",      NULL },
         { "-27",      NULL },
         { "-26",      NULL },
         { "-25",      NULL },
         { "-24",      NULL },
         { "-23",      NULL },
         { "-22",      NULL },
         { "-21",      NULL },
         { "-20",      NULL },
         { "-19",      NULL },
         { "-18",      NULL },
         { "-17",      NULL },
         { "-16",      NULL },
         { "-15",      NULL },
         { "-14",      NULL },
         { "-13",      NULL },
         { "-12",      NULL },
         { "-11",      NULL },
         { "-10",      NULL },
         { "-9",       NULL },
         { "-8",       NULL },
         { "-7",       NULL },
         { "-6",       NULL },
         { "-5",       NULL },
         { "-4",       NULL },
         { "-3",       NULL },
         { "-2",       NULL },
         { "-1",       NULL },
         { "0",        "0" },
         { "+1",       NULL },
         { "+2",       NULL },
         { "+3",       NULL },
         { "+4",       NULL },
         { "+5",       NULL },
         { "+6",       NULL },
         { "+7",       NULL },
         { "+8",       NULL },
         { "+9",       NULL },
         { "+10",      NULL },
         { "+11",      NULL },
         { "+12",      NULL },
         { "+13",      NULL },
         { "+14",      NULL },
         { "+15",      NULL },
         { "+16",      NULL },
         { "+17",      NULL },
         { "+18",      NULL },
         { "+19",      NULL },
         { "+20",      NULL },
         { "+21",      NULL },
         { "+22",      NULL },
         { "+23",      NULL },
         { "+24",      NULL },
         { "+25",      NULL },
         { "+26",      NULL },
         { "+27",      NULL },
         { "+28",      NULL },
         { "+29",      NULL },
         { "+30",      NULL },
         { "+31",      NULL },
         { "+32",      NULL },
         { "+33",      NULL },
         { "+34",      NULL },
         { "+35",      NULL },
         { "+36",      NULL },
         { "+37",      NULL },
         { "+38",      NULL },
         { "+39",      NULL },
         { "+40",      NULL },
         { NULL, NULL},
      },
      "0"
   },
#endif
   {
      BEETLE_OPT(initial_scanline),
      BEETLE_OPT_INITIAL_SCANLINE_LABEL_NO,
      NULL,
      BEETLE_OPT_INITIAL_SCANLINE_INFO_0_NO,
      NULL,
      "video",
      {
         { "0",  OPTION_VAL_0_O48_NO },
         { "1",  NULL },
         { "2",  NULL },
         { "3",  NULL },
         { "4",  NULL },
         { "5",  NULL },
         { "6",  NULL },
         { "7",  NULL },
         { "8",  NULL },
         { "9",  NULL },
         { "10", NULL },
         { "11", NULL },
         { "12", NULL },
         { "13", NULL },
         { "14", NULL },
         { "15", NULL },
         { "16", NULL },
         { "17", NULL },
         { "18", NULL },
         { "19", NULL },
         { "20", NULL },
         { "21", NULL },
         { "22", NULL },
         { "23", NULL },
         { "24", NULL },
         { "25", NULL },
         { "26", NULL },
         { "27", NULL },
         { "28", NULL },
         { "29", NULL },
         { "30", NULL },
         { "31", NULL },
         { "32", NULL },
         { "33", NULL },
         { "34", NULL },
         { "35", NULL },
         { "36", NULL },
         { "37", NULL },
         { "38", NULL },
         { "39", NULL },
         { "40", NULL },
         { NULL, NULL },
      },
      "0"
   },
   {
      BEETLE_OPT(last_scanline),
      BEETLE_OPT_LAST_SCANLINE_LABEL_NO,
      NULL,
      BEETLE_OPT_LAST_SCANLINE_INFO_0_NO,
      NULL,
      "video",
      {
         { "210", NULL },
         { "211", NULL },
         { "212", NULL },
         { "213", NULL },
         { "214", NULL },
         { "215", NULL },
         { "216", NULL },
         { "217", NULL },
         { "218", NULL },
         { "219", NULL },
         { "220", NULL },
         { "221", NULL },
         { "222", NULL },
         { "223", NULL },
         { "224", NULL },
         { "225", NULL },
         { "226", NULL },
         { "227", NULL },
         { "228", NULL },
         { "229", NULL },
         { "230", NULL },
         { "231", NULL },
         { "232", NULL },
         { "233", NULL },
         { "234", NULL },
         { "235", NULL },
         { "236", NULL },
         { "237", NULL },
         { "238", NULL },
         { "239", OPTION_VAL_239_NO },
         { NULL, NULL },
      },
      "239"
   },
   {
      BEETLE_OPT(initial_scanline_pal),
      BEETLE_OPT_INITIAL_SCANLINE_PAL_LABEL_NO,
      NULL,
      BEETLE_OPT_INITIAL_SCANLINE_PAL_INFO_0_NO,
      NULL,
      "video",
      {
         { "0",  OPTION_VAL_0_O48_NO },
         { "1",  NULL },
         { "2",  NULL },
         { "3",  NULL },
         { "4",  NULL },
         { "5",  NULL },
         { "6",  NULL },
         { "7",  NULL },
         { "8",  NULL },
         { "9",  NULL },
         { "10", NULL },
         { "11", NULL },
         { "12", NULL },
         { "13", NULL },
         { "14", NULL },
         { "15", NULL },
         { "16", NULL },
         { "17", NULL },
         { "18", NULL },
         { "19", NULL },
         { "20", NULL },
         { "21", NULL },
         { "22", NULL },
         { "23", NULL },
         { "24", NULL },
         { "25", NULL },
         { "26", NULL },
         { "27", NULL },
         { "28", NULL },
         { "29", NULL },
         { "30", NULL },
         { "31", NULL },
         { "32", NULL },
         { "33", NULL },
         { "34", NULL },
         { "35", NULL },
         { "36", NULL },
         { "37", NULL },
         { "38", NULL },
         { "39", NULL },
         { "40", NULL },
         { NULL, NULL },
      },
      "0"
   },
   {
      BEETLE_OPT(last_scanline_pal),
      BEETLE_OPT_LAST_SCANLINE_PAL_LABEL_NO,
      NULL,
      BEETLE_OPT_LAST_SCANLINE_PAL_INFO_0_NO,
      NULL,
      "video",
      {
         { "230", NULL },
         { "231", NULL },
         { "232", NULL },
         { "233", NULL },
         { "234", NULL },
         { "235", NULL },
         { "236", NULL },
         { "237", NULL },
         { "238", NULL },
         { "239", NULL },
         { "240", NULL },
         { "241", NULL },
         { "242", NULL },
         { "243", NULL },
         { "244", NULL },
         { "245", NULL },
         { "246", NULL },
         { "247", NULL },
         { "248", NULL },
         { "249", NULL },
         { "250", NULL },
         { "251", NULL },
         { "252", NULL },
         { "253", NULL },
         { "254", NULL },
         { "255", NULL },
         { "256", NULL },
         { "257", NULL },
         { "258", NULL },
         { "259", NULL },
         { "260", NULL },
         { "261", NULL },
         { "262", NULL },
         { "263", NULL },
         { "264", NULL },
         { "265", NULL },
         { "266", NULL },
         { "267", NULL },
         { "268", NULL },
         { "269", NULL },
         { "270", NULL },
         { "271", NULL },
         { "272", NULL },
         { "273", NULL },
         { "274", NULL },
         { "275", NULL },
         { "276", NULL },
         { "277", NULL },
         { "278", NULL },
         { "279", NULL },
         { "280", NULL },
         { "281", NULL },
         { "282", NULL },
         { "283", NULL },
         { "284", NULL },
         { "285", NULL },
         { "286", NULL },
         { "287", OPTION_VAL_287_NO },
         { NULL, NULL },
      },
      "287"
   },
   { NULL, NULL, NULL, NULL, NULL, NULL, {{0}}, NULL },
};
struct retro_core_options_v2 options_no = {
   option_cats_no,
   option_defs_no
};

/* RETRO_LANGUAGE_OR */

#define CATEGORY_SYSTEM_LABEL_OR NULL
#define CATEGORY_SYSTEM_INFO_0_OR NULL
#define CATEGORY_VIDEO_LABEL_OR "ଵିଡ଼ିଓ"
#define CATEGORY_VIDEO_INFO_0_OR NULL
#define CATEGORY_PGXP_LABEL_OR NULL
#define CATEGORY_PGXP_INFO_0_OR NULL
#define CATEGORY_OSD_LABEL_OR NULL
#define CATEGORY_OSD_INFO_0_OR NULL
#define CATEGORY_INPUT_LABEL_OR NULL
#define CATEGORY_INPUT_INFO_0_OR NULL
#define CATEGORY_MEMCARDS_LABEL_OR NULL
#define CATEGORY_MEMCARDS_INFO_0_OR NULL
#define CATEGORY_HACKS_LABEL_OR NULL
#define CATEGORY_HACKS_INFO_0_OR NULL
#define BEETLE_OPT_CPU_FREQ_SCALE_LABEL_OR NULL
#define BEETLE_OPT_CPU_FREQ_SCALE_INFO_0_OR NULL
#define OPTION_VAL_50_OR NULL
#define OPTION_VAL_60_OR NULL
#define OPTION_VAL_70_OR NULL
#define OPTION_VAL_80_OR NULL
#define OPTION_VAL_90_OR NULL
#define OPTION_VAL_100_OR NULL
#define OPTION_VAL_110_OR NULL
#define OPTION_VAL_120_OR NULL
#define OPTION_VAL_130_OR NULL
#define OPTION_VAL_140_OR NULL
#define OPTION_VAL_150_OR NULL
#define OPTION_VAL_160_OR NULL
#define OPTION_VAL_170_OR NULL
#define OPTION_VAL_180_OR NULL
#define OPTION_VAL_190_OR NULL
#define OPTION_VAL_200_OR NULL
#define OPTION_VAL_210_OR NULL
#define OPTION_VAL_220_OR NULL
#define OPTION_VAL_230_OR NULL
#define OPTION_VAL_240_OR NULL
#define OPTION_VAL_250_OR NULL
#define OPTION_VAL_260_OR NULL
#define OPTION_VAL_270_OR NULL
#define OPTION_VAL_280_OR NULL
#define OPTION_VAL_290_OR NULL
#define OPTION_VAL_300_OR NULL
#define OPTION_VAL_310_OR NULL
#define OPTION_VAL_320_OR NULL
#define OPTION_VAL_330_OR NULL
#define OPTION_VAL_340_OR NULL
#define OPTION_VAL_350_OR NULL
#define OPTION_VAL_360_OR NULL
#define OPTION_VAL_370_OR NULL
#define OPTION_VAL_380_OR NULL
#define OPTION_VAL_390_OR NULL
#define OPTION_VAL_400_OR NULL
#define OPTION_VAL_410_OR NULL
#define OPTION_VAL_420_OR NULL
#define OPTION_VAL_430_OR NULL
#define OPTION_VAL_440_OR NULL
#define OPTION_VAL_450_OR NULL
#define OPTION_VAL_460_OR NULL
#define OPTION_VAL_470_OR NULL
#define OPTION_VAL_480_OR NULL
#define OPTION_VAL_490_OR NULL
#define OPTION_VAL_500_OR NULL
#define OPTION_VAL_510_OR NULL
#define OPTION_VAL_520_OR NULL
#define OPTION_VAL_530_OR NULL
#define OPTION_VAL_540_OR NULL
#define OPTION_VAL_550_OR NULL
#define OPTION_VAL_560_OR NULL
#define OPTION_VAL_570_OR NULL
#define OPTION_VAL_580_OR NULL
#define OPTION_VAL_590_OR NULL
#define OPTION_VAL_600_OR NULL
#define OPTION_VAL_610_OR NULL
#define OPTION_VAL_620_OR NULL
#define OPTION_VAL_630_OR NULL
#define OPTION_VAL_640_OR NULL
#define OPTION_VAL_650_OR NULL
#define OPTION_VAL_660_OR NULL
#define OPTION_VAL_670_OR NULL
#define OPTION_VAL_680_OR NULL
#define OPTION_VAL_690_OR NULL
#define OPTION_VAL_700_OR NULL
#define OPTION_VAL_710_OR NULL
#define OPTION_VAL_720_OR NULL
#define OPTION_VAL_730_OR NULL
#define OPTION_VAL_740_OR NULL
#define OPTION_VAL_750_OR NULL
#define BEETLE_OPT_GPU_OVERCLOCK_LABEL_OR NULL
#define BEETLE_OPT_GPU_OVERCLOCK_INFO_0_OR NULL
#define OPTION_VAL_1X_NATIVE_OR NULL
#define OPTION_VAL_2X_OR NULL
#define OPTION_VAL_4X_OR NULL
#define OPTION_VAL_8X_OR NULL
#define OPTION_VAL_16X_OR NULL
#define OPTION_VAL_32X_OR NULL
#define BEETLE_OPT_GTE_OVERCLOCK_LABEL_OR NULL
#define BEETLE_OPT_GTE_OVERCLOCK_INFO_0_OR NULL
#define BEETLE_OPT_CD_ACCESS_METHOD_LABEL_OR NULL
#define BEETLE_OPT_CD_ACCESS_METHOD_INFO_0_OR NULL
#define OPTION_VAL_SYNC_OR NULL
#define OPTION_VAL_ASYNC_OR NULL
#define OPTION_VAL_PRECACHE_OR NULL
#define BEETLE_OPT_CD_FASTLOAD_LABEL_OR NULL
#define BEETLE_OPT_CD_FASTLOAD_INFO_0_OR NULL
#define OPTION_VAL_2X_NATIVE_OR NULL
#define OPTION_VAL_6X_OR NULL
#define OPTION_VAL_10X_OR NULL
#define OPTION_VAL_12X_OR NULL
#define OPTION_VAL_14X_OR NULL
#define BEETLE_OPT_SKIP_BIOS_LABEL_OR NULL
#define BEETLE_OPT_SKIP_BIOS_INFO_0_OR NULL
#define BEETLE_OPT_OVERRIDE_BIOS_LABEL_OR NULL
#define BEETLE_OPT_OVERRIDE_BIOS_INFO_0_OR NULL
#define OPTION_VAL_PSXONPSP_OR NULL
#define OPTION_VAL_PS1_ROM_OR NULL
#define OPTION_VAL_OPENBIOS_OR NULL
#define BEETLE_OPT_REGION_LABEL_OR NULL
#define BEETLE_OPT_REGION_INFO_0_OR NULL
#define OPTION_VAL_AUTO_OR NULL
#define OPTION_VAL_NTSC_J_OR NULL
#define OPTION_VAL_NTSC_U_OR NULL
#define OPTION_VAL_PAL_OR NULL
#define BEETLE_OPT_INTERNAL_RESOLUTION_LABEL_OR NULL
#define BEETLE_OPT_INTERNAL_RESOLUTION_INFO_0_OR NULL
#define BEETLE_OPT_RENDERER_LABEL_OR NULL
#define BEETLE_OPT_RENDERER_INFO_0_OR NULL
#define OPTION_VAL_HARDWARE_OR NULL
#define OPTION_VAL_HARDWARE_GL_OR NULL
#define OPTION_VAL_HARDWARE_VK_OR NULL
#define OPTION_VAL_SOFTWARE_OR NULL
#define BEETLE_OPT_RENDERER_SOFTWARE_FB_LABEL_OR NULL
#define BEETLE_OPT_RENDERER_SOFTWARE_FB_INFO_0_OR NULL
#define BEETLE_OPT_DEPTH_LABEL_OR NULL
#define BEETLE_OPT_DEPTH_INFO_0_OR NULL
#define OPTION_VAL_16BPP_NATIVE_OR NULL
#define OPTION_VAL_32BPP_OR NULL
#define BEETLE_OPT_DITHER_MODE_LABEL_OR NULL
#define BEETLE_OPT_DITHER_MODE_INFO_0_OR NULL
#define OPTION_VAL_INTERNAL_RESOLUTION_OR NULL
#define BEETLE_OPT_SCALED_UV_OFFSET_LABEL_OR NULL
#define BEETLE_OPT_SCALED_UV_OFFSET_INFO_0_OR NULL
#define BEETLE_OPT_FILTER_LABEL_OR NULL
#define BEETLE_OPT_FILTER_INFO_0_OR NULL
#define OPTION_VAL_NEAREST_OR NULL
#define OPTION_VAL_SABR_OR NULL
#define OPTION_VAL_XBR_OR NULL
#define OPTION_VAL_BILINEAR_OR NULL
#define OPTION_VAL_3_POINT_OR NULL
#define OPTION_VAL_JINC2_OR NULL
#define BEETLE_OPT_FILTER_EXCLUDE_SPRITE_LABEL_OR NULL
#define BEETLE_OPT_FILTER_EXCLUDE_SPRITE_INFO_0_OR NULL
#define OPTION_VAL_OPAQUE_OR NULL
#define OPTION_VAL_ALL_OR NULL
#define BEETLE_OPT_FILTER_EXCLUDE_2D_POLYGON_LABEL_OR NULL
#define BEETLE_OPT_FILTER_EXCLUDE_2D_POLYGON_INFO_0_OR NULL
#define BEETLE_OPT_ADAPTIVE_SMOOTHING_LABEL_OR NULL
#define BEETLE_OPT_ADAPTIVE_SMOOTHING_INFO_0_OR NULL
#define BEETLE_OPT_SUPER_SAMPLING_LABEL_OR NULL
#define BEETLE_OPT_SUPER_SAMPLING_INFO_0_OR NULL
#define BEETLE_OPT_MSAA_LABEL_OR NULL
#define BEETLE_OPT_MSAA_INFO_0_OR NULL
#define OPTION_VAL_1X_OR NULL
#define BEETLE_OPT_MDEC_YUV_LABEL_OR NULL
#define BEETLE_OPT_MDEC_YUV_INFO_0_OR NULL
#define BEETLE_OPT_TRACK_TEXTURES_LABEL_OR NULL
#define BEETLE_OPT_TRACK_TEXTURES_INFO_0_OR NULL
#define BEETLE_OPT_DUMP_TEXTURES_LABEL_OR NULL
#define BEETLE_OPT_DUMP_TEXTURES_INFO_0_OR NULL
#define BEETLE_OPT_REPLACE_TEXTURES_LABEL_OR NULL
#define BEETLE_OPT_REPLACE_TEXTURES_INFO_0_OR NULL
#define BEETLE_OPT_PGXP_MODE_LABEL_OR NULL
#define BEETLE_OPT_PGXP_MODE_INFO_0_OR NULL
#define OPTION_VAL_MEMORY_ONLY_OR NULL
#define OPTION_VAL_MEMORY_CPU_OR NULL
#define BEETLE_OPT_PGXP_2D_TOL_LABEL_OR NULL
#define BEETLE_OPT_PGXP_2D_TOL_INFO_0_OR NULL
#define OPTION_VAL_0PX_OR NULL
#define OPTION_VAL_1PX_OR NULL
#define OPTION_VAL_2PX_OR NULL
#define OPTION_VAL_3PX_OR NULL
#define OPTION_VAL_4PX_OR NULL
#define OPTION_VAL_5PX_OR NULL
#define OPTION_VAL_6PX_OR NULL
#define OPTION_VAL_7PX_OR NULL
#define OPTION_VAL_8PX_OR NULL
#define BEETLE_OPT_PGXP_NCLIP_LABEL_OR NULL
#define BEETLE_OPT_PGXP_NCLIP_INFO_0_OR NULL
#define BEETLE_OPT_PGXP_VERTEX_LABEL_OR NULL
#define BEETLE_OPT_PGXP_VERTEX_INFO_0_OR NULL
#define BEETLE_OPT_PGXP_TEXTURE_LABEL_OR NULL
#define BEETLE_OPT_PGXP_TEXTURE_INFO_0_OR NULL
#define BEETLE_OPT_DISPLAY_INTERNAL_FPS_LABEL_OR NULL
#define BEETLE_OPT_DISPLAY_INTERNAL_FPS_INFO_0_OR NULL
#define BEETLE_OPT_DISPLAY_OSD_LABEL_OR NULL
#define BEETLE_OPT_DISPLAY_OSD_INFO_0_OR NULL
#define BEETLE_OPT_DISPLAY_VRAM_LABEL_OR NULL
#define BEETLE_OPT_DISPLAY_VRAM_INFO_0_OR NULL
#define BEETLE_OPT_ANALOG_CALIBRATION_LABEL_OR NULL
#define BEETLE_OPT_ANALOG_CALIBRATION_INFO_0_OR NULL
#define BEETLE_OPT_ANALOG_TOGGLE_LABEL_OR NULL
#define BEETLE_OPT_ANALOG_TOGGLE_INFO_0_OR NULL
#define OPTION_VAL_ENABLED_ANALOG_OR NULL
#define BEETLE_OPT_ANALOG_TOGGLE_COMBO_LABEL_OR NULL
#define BEETLE_OPT_ANALOG_TOGGLE_COMBO_INFO_0_OR NULL
#define OPTION_VAL_L1_L2_R1_R2_START_SELECT_OR NULL
#define OPTION_VAL_L1_R1_SELECT_OR NULL
#define OPTION_VAL_L1_R1_START_OR NULL
#define OPTION_VAL_L1_R1_L3_OR NULL
#define OPTION_VAL_L1_R1_R3_OR NULL
#define OPTION_VAL_L2_R2_SELECT_OR NULL
#define OPTION_VAL_L2_R2_START_OR NULL
#define OPTION_VAL_L2_R2_L3_OR NULL
#define OPTION_VAL_L2_R2_R3_OR NULL
#define OPTION_VAL_L3_R3_OR NULL
#define BEETLE_OPT_ANALOG_TOGGLE_HOLD_LABEL_OR NULL
#define BEETLE_OPT_ANALOG_TOGGLE_HOLD_INFO_0_OR NULL
#define OPTION_VAL_0_OR NULL
#define OPTION_VAL_1_OR NULL
#define OPTION_VAL_2_OR NULL
#define OPTION_VAL_3_OR NULL
#define OPTION_VAL_4_OR NULL
#define OPTION_VAL_5_OR NULL
#define BEETLE_OPT_ENABLE_MULTITAP_PORT1_LABEL_OR NULL
#define BEETLE_OPT_ENABLE_MULTITAP_PORT1_INFO_0_OR NULL
#define BEETLE_OPT_ENABLE_MULTITAP_PORT2_LABEL_OR NULL
#define BEETLE_OPT_ENABLE_MULTITAP_PORT2_INFO_0_OR NULL
#define BEETLE_OPT_GUN_INPUT_MODE_LABEL_OR NULL
#define BEETLE_OPT_GUN_INPUT_MODE_INFO_0_OR NULL
#define OPTION_VAL_LIGHTGUN_OR NULL
#define OPTION_VAL_TOUCHSCREEN_OR NULL
#define BEETLE_OPT_GUN_CURSOR_LABEL_OR NULL
#define BEETLE_OPT_GUN_CURSOR_INFO_0_OR NULL
#define OPTION_VAL_CROSS_OR NULL
#define OPTION_VAL_DOT_OR NULL
#define BEETLE_OPT_CROSSHAIR_COLOR_P1_LABEL_OR NULL
#define BEETLE_OPT_CROSSHAIR_COLOR_P1_INFO_0_OR NULL
#define OPTION_VAL_RED_OR NULL
#define OPTION_VAL_BLUE_OR NULL
#define OPTION_VAL_GREEN_OR NULL
#define OPTION_VAL_ORANGE_OR NULL
#define OPTION_VAL_YELLOW_OR NULL
#define OPTION_VAL_CYAN_OR NULL
#define OPTION_VAL_PINK_OR NULL
#define OPTION_VAL_PURPLE_OR NULL
#define OPTION_VAL_BLACK_OR NULL
#define OPTION_VAL_WHITE_OR NULL
#define BEETLE_OPT_CROSSHAIR_COLOR_P2_LABEL_OR NULL
#define BEETLE_OPT_CROSSHAIR_COLOR_P2_INFO_0_OR NULL
#define BEETLE_OPT_MOUSE_SENSITIVITY_LABEL_OR NULL
#define BEETLE_OPT_MOUSE_SENSITIVITY_INFO_0_OR NULL
#define OPTION_VAL_5_O42_OR NULL
#define OPTION_VAL_10_OR NULL
#define OPTION_VAL_15_OR NULL
#define OPTION_VAL_20_OR NULL
#define OPTION_VAL_25_OR NULL
#define OPTION_VAL_30_OR NULL
#define OPTION_VAL_35_OR NULL
#define OPTION_VAL_40_OR NULL
#define OPTION_VAL_45_OR NULL
#define OPTION_VAL_55_OR NULL
#define OPTION_VAL_65_OR NULL
#define OPTION_VAL_75_OR NULL
#define OPTION_VAL_85_OR NULL
#define OPTION_VAL_95_OR NULL
#define OPTION_VAL_100_O42_OR NULL
#define OPTION_VAL_105_OR NULL
#define OPTION_VAL_115_OR NULL
#define OPTION_VAL_125_OR NULL
#define OPTION_VAL_135_OR NULL
#define OPTION_VAL_145_OR NULL
#define OPTION_VAL_155_OR NULL
#define OPTION_VAL_165_OR NULL
#define OPTION_VAL_175_OR NULL
#define OPTION_VAL_185_OR NULL
#define OPTION_VAL_195_OR NULL
#define BEETLE_OPT_NEGCON_RESPONSE_LABEL_OR NULL
#define BEETLE_OPT_NEGCON_RESPONSE_INFO_0_OR NULL
#define OPTION_VAL_LINEAR_OR NULL
#define OPTION_VAL_QUADRATIC_OR NULL
#define OPTION_VAL_CUBIC_OR NULL
#define BEETLE_OPT_NEGCON_DEADZONE_LABEL_OR NULL
#define BEETLE_OPT_NEGCON_DEADZONE_INFO_0_OR NULL
#define OPTION_VAL_0_O44_OR NULL
#define BEETLE_OPT_USE_MEDNAFEN_MEMCARD0_METHOD_LABEL_OR NULL
#define BEETLE_OPT_USE_MEDNAFEN_MEMCARD0_METHOD_INFO_0_OR NULL
#define OPTION_VAL_LIBRETRO_OR NULL
#define OPTION_VAL_MEDNAFEN_OR NULL
#define BEETLE_OPT_ENABLE_MEMCARD1_LABEL_OR NULL
#define BEETLE_OPT_ENABLE_MEMCARD1_INFO_0_OR NULL
#define BEETLE_OPT_SHARED_MEMORY_CARDS_LABEL_OR NULL
#define BEETLE_OPT_SHARED_MEMORY_CARDS_INFO_0_OR NULL
#define BEETLE_OPT_MEMCARD_LEFT_INDEX_LABEL_OR NULL
#define BEETLE_OPT_MEMCARD_LEFT_INDEX_INFO_0_OR NULL
#define OPTION_VAL_0_O48_OR NULL
#define BEETLE_OPT_MEMCARD_RIGHT_INDEX_LABEL_OR NULL
#define BEETLE_OPT_MEMCARD_RIGHT_INDEX_INFO_0_OR NULL
#define OPTION_VAL_1_O49_OR NULL
#define BEETLE_OPT_LINE_RENDER_LABEL_OR NULL
#define BEETLE_OPT_LINE_RENDER_INFO_0_OR NULL
#define OPTION_VAL_DEFAULT_OR NULL
#define OPTION_VAL_AGGRESSIVE_OR NULL
#define BEETLE_OPT_WIDESCREEN_HACK_LABEL_OR NULL
#define BEETLE_OPT_WIDESCREEN_HACK_INFO_0_OR NULL
#define BEETLE_OPT_WIDESCREEN_HACK_ASPECT_RATIO_LABEL_OR NULL
#define BEETLE_OPT_WIDESCREEN_HACK_ASPECT_RATIO_INFO_0_OR NULL
#define OPTION_VAL_16_9_OR NULL
#define OPTION_VAL_16_10_OR NULL
#define OPTION_VAL_18_9_OR NULL
#define OPTION_VAL_19_9_OR NULL
#define OPTION_VAL_20_9_OR NULL
#define OPTION_VAL_21_9_OR NULL
#define OPTION_VAL_32_9_OR NULL
#define BEETLE_OPT_CPU_DYNAREC_LABEL_OR NULL
#define BEETLE_OPT_CPU_DYNAREC_INFO_0_OR NULL
#define OPTION_VAL_DISABLED_OR NULL
#define OPTION_VAL_EXECUTE_OR NULL
#define OPTION_VAL_EXECUTE_ONE_OR NULL
#define OPTION_VAL_RUN_INTERPRETER_OR NULL
#define BEETLE_OPT_DYNAREC_INVALIDATE_LABEL_OR NULL
#define BEETLE_OPT_DYNAREC_INVALIDATE_INFO_0_OR NULL
#define OPTION_VAL_FULL_OR NULL
#define OPTION_VAL_DMA_OR NULL
#define BEETLE_OPT_DYNAREC_EVENTCYCLES_LABEL_OR NULL
#define BEETLE_OPT_DYNAREC_EVENTCYCLES_INFO_0_OR NULL
#define OPTION_VAL_128_OR NULL
#define BEETLE_OPT_DYNAREC_SPU_SAMPLES_LABEL_OR NULL
#define BEETLE_OPT_DYNAREC_SPU_SAMPLES_INFO_0_OR NULL
#define BEETLE_OPT_CORE_TIMING_FPS_LABEL_OR NULL
#define BEETLE_OPT_CORE_TIMING_FPS_INFO_0_OR NULL
#define OPTION_VAL_AUTO_TOGGLE_OR NULL
#define OPTION_VAL_FORCE_PROGRESSIVE_OR NULL
#define OPTION_VAL_FORCE_INTERLACED_OR NULL
#define BEETLE_OPT_DEINTERLACER_LABEL_OR NULL
#define BEETLE_OPT_DEINTERLACER_INFO_0_OR NULL
#define OPTION_VAL_WEAVE_OR NULL
#define OPTION_VAL_BOB_OR NULL
#define OPTION_VAL_BOB_OFFSET_OR NULL
#define OPTION_VAL_FASTMAD_OR NULL
#define BEETLE_OPT_FRAME_DUPING_LABEL_OR NULL
#define BEETLE_OPT_FRAME_DUPING_INFO_0_OR NULL
#define BEETLE_OPT_PAL_VIDEO_TIMING_OVERRIDE_LABEL_OR NULL
#define BEETLE_OPT_PAL_VIDEO_TIMING_OVERRIDE_INFO_0_OR NULL
#define BEETLE_OPT_ASPECT_RATIO_LABEL_OR NULL
#define BEETLE_OPT_ASPECT_RATIO_INFO_0_OR NULL
#define OPTION_VAL_CORRECTED_OR NULL
#define OPTION_VAL_UNCORRECTED_OR NULL
#define OPTION_VAL_4_3_OR NULL
#define OPTION_VAL_NTSC_OR NULL
#define BEETLE_OPT_CROP_OVERSCAN_LABEL_OR NULL
#define BEETLE_OPT_CROP_OVERSCAN_INFO_0_OR NULL
#define OPTION_VAL_STATIC_OR NULL
#define OPTION_VAL_SMART_OR NULL
#define BEETLE_OPT_IMAGE_CROP_LABEL_OR NULL
#define BEETLE_OPT_IMAGE_CROP_INFO_0_OR NULL
#define OPTION_VAL_9PX_OR NULL
#define OPTION_VAL_10PX_OR NULL
#define OPTION_VAL_11PX_OR NULL
#define OPTION_VAL_12PX_OR NULL
#define OPTION_VAL_13PX_OR NULL
#define OPTION_VAL_14PX_OR NULL
#define OPTION_VAL_15PX_OR NULL
#define OPTION_VAL_16PX_OR NULL
#define OPTION_VAL_17PX_OR NULL
#define OPTION_VAL_18PX_OR NULL
#define OPTION_VAL_19PX_OR NULL
#define OPTION_VAL_20PX_OR NULL
#define BEETLE_OPT_IMAGE_OFFSET_LABEL_OR NULL
#define BEETLE_OPT_IMAGE_OFFSET_INFO_0_OR NULL
#define OPTION_VAL_12PX_O64_OR NULL
#define OPTION_VAL_11PX_O64_OR NULL
#define OPTION_VAL_10PX_O64_OR NULL
#define OPTION_VAL_9PX_O64_OR NULL
#define OPTION_VAL_8PX_O64_OR NULL
#define OPTION_VAL_7PX_O64_OR NULL
#define OPTION_VAL_6PX_O64_OR NULL
#define OPTION_VAL_5PX_O64_OR NULL
#define OPTION_VAL_4PX_O64_OR NULL
#define OPTION_VAL_3PX_O64_OR NULL
#define OPTION_VAL_2PX_O64_OR NULL
#define OPTION_VAL_1PX_O64_OR NULL
#define OPTION_VAL_1PX_O64_0_OR NULL
#define OPTION_VAL_2PX_O64_0_OR NULL
#define OPTION_VAL_3PX_O64_0_OR NULL
#define OPTION_VAL_4PX_O64_0_OR NULL
#define OPTION_VAL_5PX_O64_0_OR NULL
#define OPTION_VAL_6PX_O64_0_OR NULL
#define OPTION_VAL_7PX_O64_0_OR NULL
#define OPTION_VAL_8PX_O64_0_OR NULL
#define OPTION_VAL_9PX_O64_0_OR NULL
#define OPTION_VAL_10PX_O64_0_OR NULL
#define OPTION_VAL_11PX_O64_0_OR NULL
#define OPTION_VAL_12PX_O64_0_OR NULL
#define BEETLE_OPT_IMAGE_OFFSET_CYCLES_LABEL_OR NULL
#define BEETLE_OPT_IMAGE_OFFSET_CYCLES_INFO_0_OR NULL
#define BEETLE_OPT_INITIAL_SCANLINE_LABEL_OR NULL
#define BEETLE_OPT_INITIAL_SCANLINE_INFO_0_OR NULL
#define BEETLE_OPT_LAST_SCANLINE_LABEL_OR NULL
#define BEETLE_OPT_LAST_SCANLINE_INFO_0_OR NULL
#define OPTION_VAL_239_OR NULL
#define BEETLE_OPT_INITIAL_SCANLINE_PAL_LABEL_OR NULL
#define BEETLE_OPT_INITIAL_SCANLINE_PAL_INFO_0_OR NULL
#define BEETLE_OPT_LAST_SCANLINE_PAL_LABEL_OR NULL
#define BEETLE_OPT_LAST_SCANLINE_PAL_INFO_0_OR NULL
#define OPTION_VAL_287_OR NULL

struct retro_core_option_v2_category option_cats_or[] = {
   {
      "system",
      CATEGORY_SYSTEM_LABEL_OR,
      CATEGORY_SYSTEM_INFO_0_OR
   },
   {
      "video",
      CATEGORY_VIDEO_LABEL_OR,
      CATEGORY_VIDEO_INFO_0_OR
   },
   {
      "pgxp",
      CATEGORY_PGXP_LABEL_OR,
      CATEGORY_PGXP_INFO_0_OR
   },
   {
      "osd",
      CATEGORY_OSD_LABEL_OR,
      CATEGORY_OSD_INFO_0_OR
   },
   {
      "input",
      CATEGORY_INPUT_LABEL_OR,
      CATEGORY_INPUT_INFO_0_OR
   },
   {
      "memcards",
      CATEGORY_MEMCARDS_LABEL_OR,
      CATEGORY_MEMCARDS_INFO_0_OR
   },
   {
      "hacks",
      CATEGORY_HACKS_LABEL_OR,
      CATEGORY_HACKS_INFO_0_OR
   },
   { NULL, NULL, NULL },
};
struct retro_core_option_v2_definition option_defs_or[] = {
   {
      BEETLE_OPT(cpu_freq_scale),
      BEETLE_OPT_CPU_FREQ_SCALE_LABEL_OR,
      NULL,
      BEETLE_OPT_CPU_FREQ_SCALE_INFO_0_OR,
      NULL,
      "system",
      {
         { "50%",           OPTION_VAL_50_OR },
         { "60%",           OPTION_VAL_60_OR },
         { "70%",           OPTION_VAL_70_OR },
         { "80%",           OPTION_VAL_80_OR },
         { "90%",           OPTION_VAL_90_OR },
         { "100%",          OPTION_VAL_100_OR },
         { "110%",          OPTION_VAL_110_OR },
         { "120%",          OPTION_VAL_120_OR },
         { "130%",          OPTION_VAL_130_OR },
         { "140%",          OPTION_VAL_140_OR },
         { "150%",          OPTION_VAL_150_OR },
         { "160%",          OPTION_VAL_160_OR },
         { "170%",          OPTION_VAL_170_OR },
         { "180%",          OPTION_VAL_180_OR },
         { "190%",          OPTION_VAL_190_OR },
         { "200%",          OPTION_VAL_200_OR },
         { "210%",          OPTION_VAL_210_OR },
         { "220%",          OPTION_VAL_220_OR },
         { "230%",          OPTION_VAL_230_OR },
         { "240%",          OPTION_VAL_240_OR },
         { "250%",          OPTION_VAL_250_OR },
         { "260%",          OPTION_VAL_260_OR },
         { "270%",          OPTION_VAL_270_OR },
         { "280%",          OPTION_VAL_280_OR },
         { "290%",          OPTION_VAL_290_OR },
         { "300%",          OPTION_VAL_300_OR },
         { "310%",          OPTION_VAL_310_OR },
         { "320%",          OPTION_VAL_320_OR },
         { "330%",          OPTION_VAL_330_OR },
         { "340%",          OPTION_VAL_340_OR },
         { "350%",          OPTION_VAL_350_OR },
         { "360%",          OPTION_VAL_360_OR },
         { "370%",          OPTION_VAL_370_OR },
         { "380%",          OPTION_VAL_380_OR },
         { "390%",          OPTION_VAL_390_OR },
         { "400%",          OPTION_VAL_400_OR },
         { "410%",          OPTION_VAL_410_OR },
         { "420%",          OPTION_VAL_420_OR },
         { "430%",          OPTION_VAL_430_OR },
         { "440%",          OPTION_VAL_440_OR },
         { "450%",          OPTION_VAL_450_OR },
         { "460%",          OPTION_VAL_460_OR },
         { "470%",          OPTION_VAL_470_OR },
         { "480%",          OPTION_VAL_480_OR },
         { "490%",          OPTION_VAL_490_OR },
         { "500%",          OPTION_VAL_500_OR },
         { "510%",          OPTION_VAL_510_OR },
         { "520%",          OPTION_VAL_520_OR },
         { "530%",          OPTION_VAL_530_OR },
         { "540%",          OPTION_VAL_540_OR },
         { "550%",          OPTION_VAL_550_OR },
         { "560%",          OPTION_VAL_560_OR },
         { "570%",          OPTION_VAL_570_OR },
         { "580%",          OPTION_VAL_580_OR },
         { "590%",          OPTION_VAL_590_OR },
         { "600%",          OPTION_VAL_600_OR },
         { "610%",          OPTION_VAL_610_OR },
         { "620%",          OPTION_VAL_620_OR },
         { "630%",          OPTION_VAL_630_OR },
         { "640%",          OPTION_VAL_640_OR },
         { "650%",          OPTION_VAL_650_OR },
         { "660%",          OPTION_VAL_660_OR },
         { "670%",          OPTION_VAL_670_OR },
         { "680%",          OPTION_VAL_680_OR },
         { "690%",          OPTION_VAL_690_OR },
         { "700%",          OPTION_VAL_700_OR },
         { "710%",          OPTION_VAL_710_OR },
         { "720%",          OPTION_VAL_720_OR },
         { "730%",          OPTION_VAL_730_OR },
         { "740%",          OPTION_VAL_740_OR },
         { "750%",          OPTION_VAL_750_OR },
         { NULL, NULL },
      },
      "100%"
   },
   {
      BEETLE_OPT(gpu_overclock),
      BEETLE_OPT_GPU_OVERCLOCK_LABEL_OR,
      NULL,
      BEETLE_OPT_GPU_OVERCLOCK_INFO_0_OR,
      NULL,
      "system",
      {
         { "1x(native)", OPTION_VAL_1X_NATIVE_OR },
         { "2x",         OPTION_VAL_2X_OR },
         { "4x",         OPTION_VAL_4X_OR },
         { "8x",         OPTION_VAL_8X_OR },
         { "16x",        OPTION_VAL_16X_OR },
         { "32x",        OPTION_VAL_32X_OR },
         { NULL, NULL },
      },
      "1x(native)"
   },
   {
      BEETLE_OPT(gte_overclock),
      BEETLE_OPT_GTE_OVERCLOCK_LABEL_OR,
      NULL,
      BEETLE_OPT_GTE_OVERCLOCK_INFO_0_OR,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
#ifndef EMSCRIPTEN
   {
      BEETLE_OPT(cd_access_method),
      BEETLE_OPT_CD_ACCESS_METHOD_LABEL_OR,
      NULL,
      BEETLE_OPT_CD_ACCESS_METHOD_INFO_0_OR,
      NULL,
      "system",
      {
         { "sync",     OPTION_VAL_SYNC_OR },
         { "async",    OPTION_VAL_ASYNC_OR },
         { "precache", OPTION_VAL_PRECACHE_OR },
         { NULL, NULL },
      },
      "sync"
   },
#endif
   {
      BEETLE_OPT(cd_fastload),
      BEETLE_OPT_CD_FASTLOAD_LABEL_OR,
      NULL,
      BEETLE_OPT_CD_FASTLOAD_INFO_0_OR,
      NULL,
      "system",
      {
         { "2x(native)", OPTION_VAL_2X_NATIVE_OR },
         { "4x",          OPTION_VAL_4X_OR },
         { "6x",          OPTION_VAL_6X_OR },
         { "8x",          OPTION_VAL_8X_OR },
         { "10x",         OPTION_VAL_10X_OR },
         { "12x",         OPTION_VAL_12X_OR },
         { "14x",         OPTION_VAL_14X_OR },
         { NULL, NULL },
      },
      "2x(native)"
   },
   {
      BEETLE_OPT(skip_bios),
      BEETLE_OPT_SKIP_BIOS_LABEL_OR,
      NULL,
      BEETLE_OPT_SKIP_BIOS_INFO_0_OR,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(override_bios),
      BEETLE_OPT_OVERRIDE_BIOS_LABEL_OR,
      NULL,
      BEETLE_OPT_OVERRIDE_BIOS_INFO_0_OR,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "psxonpsp", OPTION_VAL_PSXONPSP_OR },
         { "ps1_rom",  OPTION_VAL_PS1_ROM_OR },
         { "openbios", OPTION_VAL_OPENBIOS_OR     },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(region),
      BEETLE_OPT_REGION_LABEL_OR,
      NULL,
      BEETLE_OPT_REGION_INFO_0_OR,
      NULL,
      "system",
      {
         { "auto",   OPTION_VAL_AUTO_OR },
         { "ntsc-j", OPTION_VAL_NTSC_J_OR },
         { "ntsc-u", OPTION_VAL_NTSC_U_OR },
         { "pal",    OPTION_VAL_PAL_OR },
         { NULL, NULL },
      },
      "auto"
   },
   {
      BEETLE_OPT(internal_resolution),
      BEETLE_OPT_INTERNAL_RESOLUTION_LABEL_OR,
      NULL,
      BEETLE_OPT_INTERNAL_RESOLUTION_INFO_0_OR,
      NULL,
      "video",
      {
         { "1x(native)", OPTION_VAL_1X_NATIVE_OR },
         { "2x",         OPTION_VAL_2X_OR },
         { "4x",         OPTION_VAL_4X_OR },
         { "8x",         OPTION_VAL_8X_OR },
         { "16x",        OPTION_VAL_16X_OR },
         { NULL, NULL },
      },
      "1x(native)"
   },
#if defined(HAVE_OPENGL) || defined(HAVE_OPENGLES) || defined(HAVE_VULKAN)
   {
      BEETLE_OPT(renderer),
      BEETLE_OPT_RENDERER_LABEL_OR,
      NULL,
      BEETLE_OPT_RENDERER_INFO_0_OR,
      NULL,
      "video",
      {
         { "hardware",    OPTION_VAL_HARDWARE_OR },
#if defined(HAVE_OPENGL) || defined(HAVE_OPENGLES)
         { "hardware_gl", OPTION_VAL_HARDWARE_GL_OR },
#endif
#if defined(HAVE_VULKAN)
         { "hardware_vk", OPTION_VAL_HARDWARE_VK_OR },
#endif
         { "software",    OPTION_VAL_SOFTWARE_OR },
         { NULL, NULL },
      },
      "hardware"
   },
   {
      BEETLE_OPT(renderer_software_fb),
      BEETLE_OPT_RENDERER_SOFTWARE_FB_LABEL_OR,
      NULL,
      BEETLE_OPT_RENDERER_SOFTWARE_FB_INFO_0_OR,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled"
   },
#endif
#if defined(HAVE_OPENGL) || defined(HAVE_OPENGLES)
   {
      BEETLE_OPT(depth),
      BEETLE_OPT_DEPTH_LABEL_OR,
      NULL,
      BEETLE_OPT_DEPTH_INFO_0_OR,
      NULL,
      "video",
      {
         { "16bpp(native)", OPTION_VAL_16BPP_NATIVE_OR },
         { "32bpp",         OPTION_VAL_32BPP_OR },
         { NULL, NULL },
      },
      "16bpp(native)"
   },
   // Sort of, it's more like 15-bit high color and 24-bit true color for visible output. The alpha channel is used for mask bit. Vulkan renderer uses ABGR1555_555 for 31 bits internal? FMVs are always 24-bit on all renderers like original hardware (BGR888, no alpha)
#endif
   {
      BEETLE_OPT(dither_mode),
      BEETLE_OPT_DITHER_MODE_LABEL_OR,
      NULL,
      BEETLE_OPT_DITHER_MODE_INFO_0_OR,
      NULL,
      "video",
      {
         { "1x(native)",          OPTION_VAL_1X_NATIVE_OR },
         { "internal resolution", OPTION_VAL_INTERNAL_RESOLUTION_OR },
         { "disabled",            NULL },
         { NULL, NULL },
      },
      "1x(native)"
   },
#ifdef HAVE_VULKAN
   {
      BEETLE_OPT(scaled_uv_offset),
      BEETLE_OPT_SCALED_UV_OFFSET_LABEL_OR,
      NULL,
      BEETLE_OPT_SCALED_UV_OFFSET_INFO_0_OR,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled"
   },
#endif
#if defined(HAVE_OPENGL) || defined(HAVE_OPENGLES) || defined(HAVE_VULKAN)
   {
      BEETLE_OPT(filter),
      BEETLE_OPT_FILTER_LABEL_OR,
      NULL,
      BEETLE_OPT_FILTER_INFO_0_OR,
      NULL,
      "video",
      {
         { "nearest",  OPTION_VAL_NEAREST_OR },
         { "SABR",     OPTION_VAL_SABR_OR },
         { "xBR",      OPTION_VAL_XBR_OR },
         { "bilinear", OPTION_VAL_BILINEAR_OR },
         { "3-point",  OPTION_VAL_3_POINT_OR },
         { "JINC2",    OPTION_VAL_JINC2_OR },
         { NULL, NULL },
      },
      "nearest"
   },
#ifdef HAVE_VULKAN
   {
      BEETLE_OPT(filter_exclude_sprite),
      BEETLE_OPT_FILTER_EXCLUDE_SPRITE_LABEL_OR,
      NULL,
      BEETLE_OPT_FILTER_EXCLUDE_SPRITE_INFO_0_OR,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "opaque", OPTION_VAL_OPAQUE_OR },
         { "all", OPTION_VAL_ALL_OR },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(filter_exclude_2d_polygon),
      BEETLE_OPT_FILTER_EXCLUDE_2D_POLYGON_LABEL_OR,
      NULL,
      BEETLE_OPT_FILTER_EXCLUDE_2D_POLYGON_INFO_0_OR,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "opaque", OPTION_VAL_OPAQUE_OR },
         { "all", OPTION_VAL_ALL_OR },
         { NULL, NULL },
      },
      "disabled"
   },
#endif
#endif
#ifdef HAVE_VULKAN
   {
      BEETLE_OPT(adaptive_smoothing),
      BEETLE_OPT_ADAPTIVE_SMOOTHING_LABEL_OR,
      NULL,
      BEETLE_OPT_ADAPTIVE_SMOOTHING_INFO_0_OR,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(super_sampling),
      BEETLE_OPT_SUPER_SAMPLING_LABEL_OR,
      NULL,
      BEETLE_OPT_SUPER_SAMPLING_INFO_0_OR,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(msaa),
      BEETLE_OPT_MSAA_LABEL_OR,
      NULL,
      BEETLE_OPT_MSAA_INFO_0_OR,
      NULL,
      "video",
      {
         { "1x",  OPTION_VAL_1X_OR },
         { "2x",  OPTION_VAL_2X_OR },
         { "4x",  OPTION_VAL_4X_OR },
         { "8x",  OPTION_VAL_8X_OR },
         { "16x", OPTION_VAL_16X_OR },
         { NULL, NULL },
      },
      "1x"
   },
   {
      BEETLE_OPT(mdec_yuv),
      BEETLE_OPT_MDEC_YUV_LABEL_OR,
      NULL,
      BEETLE_OPT_MDEC_YUV_INFO_0_OR,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(track_textures),
      BEETLE_OPT_TRACK_TEXTURES_LABEL_OR,
      NULL,
      BEETLE_OPT_TRACK_TEXTURES_INFO_0_OR,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
#ifdef TEXTURE_DUMPING_ENABLED
   {
      BEETLE_OPT(dump_textures),
      BEETLE_OPT_DUMP_TEXTURES_LABEL_OR,
      NULL,
      BEETLE_OPT_DUMP_TEXTURES_INFO_0_OR,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
#endif
   {
      BEETLE_OPT(replace_textures),
      BEETLE_OPT_REPLACE_TEXTURES_LABEL_OR,
      NULL,
      BEETLE_OPT_REPLACE_TEXTURES_INFO_0_OR,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
#endif
   {
      BEETLE_OPT(pgxp_mode),
      BEETLE_OPT_PGXP_MODE_LABEL_OR,
      NULL,
      BEETLE_OPT_PGXP_MODE_INFO_0_OR,
      NULL,
      "pgxp",
      {
         { "disabled",     NULL },
         { "memory only",  OPTION_VAL_MEMORY_ONLY_OR },
         { "memory + CPU", OPTION_VAL_MEMORY_CPU_OR },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(pgxp_2d_tol),
      BEETLE_OPT_PGXP_2D_TOL_LABEL_OR,
      NULL,
      BEETLE_OPT_PGXP_2D_TOL_INFO_0_OR,
      NULL,
      "pgxp",
      {
         { "disabled", NULL },
         { "0px", OPTION_VAL_0PX_OR },
         { "1px", OPTION_VAL_1PX_OR },
         { "2px", OPTION_VAL_2PX_OR },
         { "3px", OPTION_VAL_3PX_OR },
         { "4px", OPTION_VAL_4PX_OR },
         { "5px", OPTION_VAL_5PX_OR },
         { "6px", OPTION_VAL_6PX_OR },
         { "7px", OPTION_VAL_7PX_OR },
         { "8px", OPTION_VAL_8PX_OR },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(pgxp_nclip),
      BEETLE_OPT_PGXP_NCLIP_LABEL_OR,
      NULL,
      BEETLE_OPT_PGXP_NCLIP_INFO_0_OR,
      NULL,
      "pgxp",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
#if defined(HAVE_OPENGL) || defined(HAVE_OPENGLES) || defined(HAVE_VULKAN)
   {
      BEETLE_OPT(pgxp_vertex),
      BEETLE_OPT_PGXP_VERTEX_LABEL_OR,
      NULL,
      BEETLE_OPT_PGXP_VERTEX_INFO_0_OR,
      NULL,
      "pgxp",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
#endif
   {
      BEETLE_OPT(pgxp_texture),
      BEETLE_OPT_PGXP_TEXTURE_LABEL_OR,
      NULL,
      BEETLE_OPT_PGXP_TEXTURE_INFO_0_OR,
      NULL,
      "pgxp",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(display_internal_fps),
      BEETLE_OPT_DISPLAY_INTERNAL_FPS_LABEL_OR,
      NULL,
      BEETLE_OPT_DISPLAY_INTERNAL_FPS_INFO_0_OR,
      NULL,
      "osd",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(display_osd),
      BEETLE_OPT_DISPLAY_OSD_LABEL_OR,
      NULL,
      BEETLE_OPT_DISPLAY_OSD_INFO_0_OR,
      NULL,
      "osd",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled"
   },
#if defined(HAVE_OPENGL) || defined(HAVE_OPENGLES) || defined(HAVE_VULKAN)
   {
      BEETLE_OPT(display_vram),
      BEETLE_OPT_DISPLAY_VRAM_LABEL_OR,
      NULL,
      BEETLE_OPT_DISPLAY_VRAM_INFO_0_OR,
      NULL,
      "osd",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
#endif
   {
      BEETLE_OPT(analog_calibration),
      BEETLE_OPT_ANALOG_CALIBRATION_LABEL_OR,
      NULL,
      BEETLE_OPT_ANALOG_CALIBRATION_INFO_0_OR,
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(analog_toggle),
      BEETLE_OPT_ANALOG_TOGGLE_LABEL_OR,
      NULL,
      BEETLE_OPT_ANALOG_TOGGLE_INFO_0_OR,
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { "enabled-analog", OPTION_VAL_ENABLED_ANALOG_OR },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(analog_toggle_combo),
      BEETLE_OPT_ANALOG_TOGGLE_COMBO_LABEL_OR,
      NULL,
      BEETLE_OPT_ANALOG_TOGGLE_COMBO_INFO_0_OR,
      NULL,
      "input",
      {
         { "l1+l2+r1+r2+start+select", OPTION_VAL_L1_L2_R1_R2_START_SELECT_OR },
         { "l1+r1+select",             OPTION_VAL_L1_R1_SELECT_OR },
         { "l1+r1+start",              OPTION_VAL_L1_R1_START_OR },
         { "l1+r1+l3",                 OPTION_VAL_L1_R1_L3_OR },
         { "l1+r1+r3",                 OPTION_VAL_L1_R1_R3_OR },
         { "l2+r2+select",             OPTION_VAL_L2_R2_SELECT_OR },
         { "l2+r2+start",              OPTION_VAL_L2_R2_START_OR },
         { "l2+r2+l3",                 OPTION_VAL_L2_R2_L3_OR },
         { "l2+r2+r3",                 OPTION_VAL_L2_R2_R3_OR },
         { "l3+r3",                    OPTION_VAL_L3_R3_OR },
         { NULL, NULL },
      },
      "l1+r1+select"
   },
   {
      BEETLE_OPT(analog_toggle_hold),
      BEETLE_OPT_ANALOG_TOGGLE_HOLD_LABEL_OR,
      NULL,
      BEETLE_OPT_ANALOG_TOGGLE_HOLD_INFO_0_OR,
      NULL,
      "input",
      {
         { "0", OPTION_VAL_0_OR },
         { "1", OPTION_VAL_1_OR },
         { "2", OPTION_VAL_2_OR },
         { "3", OPTION_VAL_3_OR },
         { "4", OPTION_VAL_4_OR },
         { "5", OPTION_VAL_5_OR },
         { NULL, NULL },
      },
      "1"
   },
   {
      BEETLE_OPT(enable_multitap_port1),
      BEETLE_OPT_ENABLE_MULTITAP_PORT1_LABEL_OR,
      NULL,
      BEETLE_OPT_ENABLE_MULTITAP_PORT1_INFO_0_OR,
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(enable_multitap_port2),
      BEETLE_OPT_ENABLE_MULTITAP_PORT2_LABEL_OR,
      NULL,
      BEETLE_OPT_ENABLE_MULTITAP_PORT2_INFO_0_OR,
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(gun_input_mode),
      BEETLE_OPT_GUN_INPUT_MODE_LABEL_OR,
      NULL,
      BEETLE_OPT_GUN_INPUT_MODE_INFO_0_OR,
      NULL,
      "input",
      {
         { "lightgun",    OPTION_VAL_LIGHTGUN_OR },
         { "touchscreen", OPTION_VAL_TOUCHSCREEN_OR },
         { NULL, NULL },
      },
      "lightgun"
   },
   // Shouldn't the gun_input_mode just be Mouse vs. Touchscreen?
   {
      BEETLE_OPT(gun_cursor),
      BEETLE_OPT_GUN_CURSOR_LABEL_OR,
      NULL,
      BEETLE_OPT_GUN_CURSOR_INFO_0_OR,
      NULL,
      "input",
      {
         { "off",   "disabled" },
         { "cross", OPTION_VAL_CROSS_OR },
         { "dot",   OPTION_VAL_DOT_OR },
         { NULL, NULL },
      },
      "cross"
   },
   {
      BEETLE_OPT(crosshair_color_p1),
      BEETLE_OPT_CROSSHAIR_COLOR_P1_LABEL_OR,
      NULL,
      BEETLE_OPT_CROSSHAIR_COLOR_P1_INFO_0_OR,
      NULL,
      "input",
      {
         { "red",    OPTION_VAL_RED_OR },
         { "blue",   OPTION_VAL_BLUE_OR },
         { "green",  OPTION_VAL_GREEN_OR },
         { "orange", OPTION_VAL_ORANGE_OR },
         { "yellow", OPTION_VAL_YELLOW_OR },
         { "cyan",   OPTION_VAL_CYAN_OR },
         { "pink",   OPTION_VAL_PINK_OR },
         { "purple", OPTION_VAL_PURPLE_OR },
         { "black",  OPTION_VAL_BLACK_OR },
         { "white",  OPTION_VAL_WHITE_OR },
         { NULL, NULL },
      },
      "red"
   },
   {
      BEETLE_OPT(crosshair_color_p2),
      BEETLE_OPT_CROSSHAIR_COLOR_P2_LABEL_OR,
      NULL,
      BEETLE_OPT_CROSSHAIR_COLOR_P2_INFO_0_OR,
      NULL,
      "input",
      {
         { "blue",   OPTION_VAL_BLUE_OR },
         { "red",    OPTION_VAL_RED_OR },
         { "green",  OPTION_VAL_GREEN_OR },
         { "orange", OPTION_VAL_ORANGE_OR },
         { "yellow", OPTION_VAL_YELLOW_OR },
         { "cyan",   OPTION_VAL_CYAN_OR },
         { "pink",   OPTION_VAL_PINK_OR },
         { "purple", OPTION_VAL_PURPLE_OR },
         { "black",  OPTION_VAL_BLACK_OR },
         { "white",  OPTION_VAL_WHITE_OR },
         { NULL, NULL },
      },
      "blue"
   },
   {
      BEETLE_OPT(mouse_sensitivity),
      BEETLE_OPT_MOUSE_SENSITIVITY_LABEL_OR,
      NULL,
      BEETLE_OPT_MOUSE_SENSITIVITY_INFO_0_OR,
      NULL,
      "input",
      {
         { "5%",   OPTION_VAL_5_O42_OR },
         { "10%",  OPTION_VAL_10_OR },
         { "15%",  OPTION_VAL_15_OR },
         { "20%",  OPTION_VAL_20_OR },
         { "25%",  OPTION_VAL_25_OR },
         { "30%",  OPTION_VAL_30_OR },
         { "35%",  OPTION_VAL_35_OR },
         { "40%",  OPTION_VAL_40_OR },
         { "45%",  OPTION_VAL_45_OR },
         { "50%",  OPTION_VAL_50_OR },
         { "55%",  OPTION_VAL_55_OR },
         { "60%",  OPTION_VAL_60_OR },
         { "65%",  OPTION_VAL_65_OR },
         { "70%",  OPTION_VAL_70_OR },
         { "75%",  OPTION_VAL_75_OR },
         { "80%",  OPTION_VAL_80_OR },
         { "85%",  OPTION_VAL_85_OR },
         { "90%",  OPTION_VAL_90_OR },
         { "95%",  OPTION_VAL_95_OR },
         { "100%", OPTION_VAL_100_O42_OR },
         { "105%", OPTION_VAL_105_OR },
         { "110%", OPTION_VAL_110_OR },
         { "115%", OPTION_VAL_115_OR },
         { "120%", OPTION_VAL_120_OR },
         { "125%", OPTION_VAL_125_OR },
         { "130%", OPTION_VAL_130_OR },
         { "135%", OPTION_VAL_135_OR },
         { "140%", OPTION_VAL_140_OR },
         { "145%", OPTION_VAL_145_OR },
         { "150%", OPTION_VAL_150_OR },
         { "155%", OPTION_VAL_155_OR },
         { "160%", OPTION_VAL_160_OR },
         { "165%", OPTION_VAL_165_OR },
         { "170%", OPTION_VAL_170_OR },
         { "175%", OPTION_VAL_175_OR },
         { "180%", OPTION_VAL_180_OR },
         { "185%", OPTION_VAL_185_OR },
         { "190%", OPTION_VAL_190_OR },
         { "195%", OPTION_VAL_195_OR },
         { "200%", OPTION_VAL_200_OR },
         { NULL, NULL },
      },
      "100%"
   },
   {
      BEETLE_OPT(negcon_response),
      BEETLE_OPT_NEGCON_RESPONSE_LABEL_OR,
      NULL,
      BEETLE_OPT_NEGCON_RESPONSE_INFO_0_OR,
      NULL,
      "input",
      {
         { "linear",    OPTION_VAL_LINEAR_OR },
         { "quadratic", OPTION_VAL_QUADRATIC_OR },
         { "cubic",     OPTION_VAL_CUBIC_OR },
         { NULL, NULL },
      },
      "linear"
   },
   {
      BEETLE_OPT(negcon_deadzone),
      BEETLE_OPT_NEGCON_DEADZONE_LABEL_OR,
      NULL,
      BEETLE_OPT_NEGCON_DEADZONE_INFO_0_OR,
      NULL,
      "input",
      {
         { "0%",  OPTION_VAL_0_O44_OR },
         { "5%",  OPTION_VAL_5_O42_OR },
         { "10%", OPTION_VAL_10_OR },
         { "15%", OPTION_VAL_15_OR },
         { "20%", OPTION_VAL_20_OR },
         { "25%", OPTION_VAL_25_OR },
         { "30%", OPTION_VAL_30_OR },
         { NULL, NULL },
      },
      "0%"
   },
   {
      BEETLE_OPT(use_mednafen_memcard0_method),
      BEETLE_OPT_USE_MEDNAFEN_MEMCARD0_METHOD_LABEL_OR,
      NULL,
      BEETLE_OPT_USE_MEDNAFEN_MEMCARD0_METHOD_INFO_0_OR,
      NULL,
      "memcards",
      {
         { "libretro", OPTION_VAL_LIBRETRO_OR },
         { "mednafen", OPTION_VAL_MEDNAFEN_OR },
         { NULL, NULL },
      },
      "libretro"
   },
   {
      BEETLE_OPT(enable_memcard1),
      BEETLE_OPT_ENABLE_MEMCARD1_LABEL_OR,
      NULL,
      BEETLE_OPT_ENABLE_MEMCARD1_INFO_0_OR,
      NULL,
      "memcards",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(shared_memory_cards),
      BEETLE_OPT_SHARED_MEMORY_CARDS_LABEL_OR,
      NULL,
      BEETLE_OPT_SHARED_MEMORY_CARDS_INFO_0_OR,
      NULL,
      "memcards",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(memcard_left_index),
      BEETLE_OPT_MEMCARD_LEFT_INDEX_LABEL_OR,
      NULL,
      BEETLE_OPT_MEMCARD_LEFT_INDEX_INFO_0_OR,
      NULL,
      "memcards",
      {
         { "0",  OPTION_VAL_0_O48_OR },
         { "1",  NULL },
         { "2",  NULL },
         { "3",  NULL },
         { "4",  NULL },
         { "5",  NULL },
         { "6",  NULL },
         { "7",  NULL },
         { "8",  NULL },
         { "9",  NULL },
         { "10", NULL },
         { "11", NULL },
         { "12", NULL },
         { "13", NULL },
         { "14", NULL },
         { "15", NULL },
         { "16", NULL },
         { "17", NULL },
         { "18", NULL },
         { "19", NULL },
         { "20", NULL },
         { "21", NULL },
         { "22", NULL },
         { "23", NULL },
         { "24", NULL },
         { "25", NULL },
         { "26", NULL },
         { "27", NULL },
         { "28", NULL },
         { "29", NULL },
         { "30", NULL },
         { "31", NULL },
         { "32", NULL },
         { "33", NULL },
         { "34", NULL },
         { "35", NULL },
         { "36", NULL },
         { "37", NULL },
         { "38", NULL },
         { "39", NULL },
         { "40", NULL },
         { "41", NULL },
         { "42", NULL },
         { "43", NULL },
         { "44", NULL },
         { "45", NULL },
         { "46", NULL },
         { "47", NULL },
         { "48", NULL },
         { "49", NULL },
         { "50", NULL },
         { "51", NULL },
         { "52", NULL },
         { "53", NULL },
         { "54", NULL },
         { "55", NULL },
         { "56", NULL },
         { "57", NULL },
         { "58", NULL },
         { "59", NULL },
         { "60", NULL },
         { "61", NULL },
         { "62", NULL },
         { "63", NULL },
         { NULL, NULL },
      },
      "0"
   },
   {
      BEETLE_OPT(memcard_right_index),
      BEETLE_OPT_MEMCARD_RIGHT_INDEX_LABEL_OR,
      NULL,
      BEETLE_OPT_MEMCARD_RIGHT_INDEX_INFO_0_OR,
      NULL,
      "memcards",
      {
         { "0",  NULL },
         { "1",  OPTION_VAL_1_O49_OR },
         { "2",  NULL },
         { "3",  NULL },
         { "4",  NULL },
         { "5",  NULL },
         { "6",  NULL },
         { "7",  NULL },
         { "8",  NULL },
         { "9",  NULL },
         { "10", NULL },
         { "11", NULL },
         { "12", NULL },
         { "13", NULL },
         { "14", NULL },
         { "15", NULL },
         { "16", NULL },
         { "17", NULL },
         { "18", NULL },
         { "19", NULL },
         { "20", NULL },
         { "21", NULL },
         { "22", NULL },
         { "23", NULL },
         { "24", NULL },
         { "25", NULL },
         { "26", NULL },
         { "27", NULL },
         { "28", NULL },
         { "29", NULL },
         { "30", NULL },
         { "31", NULL },
         { "32", NULL },
         { "33", NULL },
         { "34", NULL },
         { "35", NULL },
         { "36", NULL },
         { "37", NULL },
         { "38", NULL },
         { "39", NULL },
         { "40", NULL },
         { "41", NULL },
         { "42", NULL },
         { "43", NULL },
         { "44", NULL },
         { "45", NULL },
         { "46", NULL },
         { "47", NULL },
         { "48", NULL },
         { "49", NULL },
         { "50", NULL },
         { "51", NULL },
         { "52", NULL },
         { "53", NULL },
         { "54", NULL },
         { "55", NULL },
         { "56", NULL },
         { "57", NULL },
         { "58", NULL },
         { "59", NULL },
         { "60", NULL },
         { "61", NULL },
         { "62", NULL },
         { "63", NULL },
         { NULL, NULL },
      },
      "1"
   },
   {
      BEETLE_OPT(line_render),
      BEETLE_OPT_LINE_RENDER_LABEL_OR,
      NULL,
      BEETLE_OPT_LINE_RENDER_INFO_0_OR,
      NULL,
      "hacks",
      {
         { "default",    OPTION_VAL_DEFAULT_OR },
         { "aggressive", OPTION_VAL_AGGRESSIVE_OR },
         { "disabled",   NULL },
         { NULL, NULL },
      },
      "default"
   },
   {
      BEETLE_OPT(widescreen_hack),
      BEETLE_OPT_WIDESCREEN_HACK_LABEL_OR,
      NULL,
      BEETLE_OPT_WIDESCREEN_HACK_INFO_0_OR,
      NULL,
      "hacks",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(widescreen_hack_aspect_ratio),
      BEETLE_OPT_WIDESCREEN_HACK_ASPECT_RATIO_LABEL_OR,
      NULL,
      BEETLE_OPT_WIDESCREEN_HACK_ASPECT_RATIO_INFO_0_OR,
      NULL,
      "hacks",
      {
         { "16:9",  OPTION_VAL_16_9_OR },
         { "16:10", OPTION_VAL_16_10_OR },
         { "18:9",  OPTION_VAL_18_9_OR },
         { "19:9",  OPTION_VAL_19_9_OR },
         { "20:9",  OPTION_VAL_20_9_OR },
         { "21:9",  OPTION_VAL_21_9_OR }, // 64:27
         { "32:9",  OPTION_VAL_32_9_OR },
         { NULL,    NULL },
      },
      "16:9"
   },
#if defined(HAVE_LIGHTREC)
   {
      BEETLE_OPT(cpu_dynarec),
      BEETLE_OPT_CPU_DYNAREC_LABEL_OR,
      NULL,
      BEETLE_OPT_CPU_DYNAREC_INFO_0_OR,
      NULL,
      "hacks",
      {
         { "disabled",        OPTION_VAL_DISABLED_OR },
         { "execute",         OPTION_VAL_EXECUTE_OR },
         { "execute_one",     OPTION_VAL_EXECUTE_ONE_OR },
         { "run_interpreter", OPTION_VAL_RUN_INTERPRETER_OR },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(dynarec_invalidate),
      BEETLE_OPT_DYNAREC_INVALIDATE_LABEL_OR,
      NULL,
      BEETLE_OPT_DYNAREC_INVALIDATE_INFO_0_OR,
      NULL,
      "hacks",
      {
         { "full", OPTION_VAL_FULL_OR },
         { "dma",  OPTION_VAL_DMA_OR },
         { NULL, NULL },
      },
      "full"
   },
   {
      BEETLE_OPT(dynarec_eventcycles),
      BEETLE_OPT_DYNAREC_EVENTCYCLES_LABEL_OR,
      NULL,
      BEETLE_OPT_DYNAREC_EVENTCYCLES_INFO_0_OR,
      NULL,
      "hacks",
      {
         { "128",  OPTION_VAL_128_OR },
         { "256",  NULL },
         { "384",  NULL },
         { "512",  NULL },
         { "640",  NULL },
         { "768",  NULL },
         { "896",  NULL },
         { "1024", NULL },
         { "1152", NULL },
         { "1280", NULL },
         { "1408", NULL },
         { "1536", NULL },
         { "1664", NULL },
         { "1792", NULL },
         { "1920", NULL },
         { "2048", NULL },
         { NULL, NULL },
      },
      "128"
   },
   {
      BEETLE_OPT(dynarec_spu_samples),
      BEETLE_OPT_DYNAREC_SPU_SAMPLES_LABEL_OR,
      NULL,
      BEETLE_OPT_DYNAREC_SPU_SAMPLES_INFO_0_OR,
      NULL,
      "hacks",
      {
         { "1",  OPTION_VAL_1_O49_OR },
         { "4",  NULL },
         { "16", NULL },
         { NULL, NULL },
      },
      "1"
   },
#endif
   {
      BEETLE_OPT(core_timing_fps),
      BEETLE_OPT_CORE_TIMING_FPS_LABEL_OR,
      NULL,
      BEETLE_OPT_CORE_TIMING_FPS_INFO_0_OR,
      NULL,
      "video",
      {
         { "auto_toggle",       OPTION_VAL_AUTO_TOGGLE_OR },
         { "force_progressive", OPTION_VAL_FORCE_PROGRESSIVE_OR },
         { "force_interlaced",  OPTION_VAL_FORCE_INTERLACED_OR },
      },
      "auto_toggle"
   },
   {
      BEETLE_OPT(deinterlacer),
      BEETLE_OPT_DEINTERLACER_LABEL_OR,
      NULL,
      BEETLE_OPT_DEINTERLACER_INFO_0_OR,
      NULL,
      "video",
      {
         { "weave",      OPTION_VAL_WEAVE_OR },
         { "bob",        OPTION_VAL_BOB_OR },
         { "bob_offset", OPTION_VAL_BOB_OFFSET_OR },
         { "fastmad",    OPTION_VAL_FASTMAD_OR },
         { "off",        "Off" },
         { NULL, NULL },
      },
      "weave"
   },
   {
      BEETLE_OPT(frame_duping),
      BEETLE_OPT_FRAME_DUPING_LABEL_OR,
      NULL,
      BEETLE_OPT_FRAME_DUPING_INFO_0_OR,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(pal_video_timing_override),
      BEETLE_OPT_PAL_VIDEO_TIMING_OVERRIDE_LABEL_OR,
      NULL,
      BEETLE_OPT_PAL_VIDEO_TIMING_OVERRIDE_INFO_0_OR,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(aspect_ratio),
      BEETLE_OPT_ASPECT_RATIO_LABEL_OR,
      NULL,
      BEETLE_OPT_ASPECT_RATIO_INFO_0_OR,
      NULL,
      "video",
      {
         { "corrected", OPTION_VAL_CORRECTED_OR },
         { "uncorrected", OPTION_VAL_UNCORRECTED_OR },
         { "4:3",  OPTION_VAL_4_3_OR },
         { "ntsc", OPTION_VAL_NTSC_OR },
      },
      "corrected"
   },
   {
      BEETLE_OPT(crop_overscan),
      BEETLE_OPT_CROP_OVERSCAN_LABEL_OR,
      NULL,
      BEETLE_OPT_CROP_OVERSCAN_INFO_0_OR,
      NULL,
      "video",
      {
         { "disabled", "Disabled" },
         { "static",   OPTION_VAL_STATIC_OR },
         { "smart",    OPTION_VAL_SMART_OR },
         { NULL, NULL },
      },
      "smart"
   },
   {
      BEETLE_OPT(image_crop),
      BEETLE_OPT_IMAGE_CROP_LABEL_OR,
      NULL,
      BEETLE_OPT_IMAGE_CROP_INFO_0_OR,
      NULL,
      "video",
      {
         { "disabled", "0" },
         { "1px",      OPTION_VAL_1PX_OR },
         { "2px",      OPTION_VAL_2PX_OR },
         { "3px",      OPTION_VAL_3PX_OR },
         { "4px",      OPTION_VAL_4PX_OR },
         { "5px",      OPTION_VAL_5PX_OR },
         { "6px",      OPTION_VAL_6PX_OR },
         { "7px",      OPTION_VAL_7PX_OR },
         { "8px",      OPTION_VAL_8PX_OR },
         { "9px",      OPTION_VAL_9PX_OR },
         { "10px",     OPTION_VAL_10PX_OR },
         { "11px",     OPTION_VAL_11PX_OR },
         { "12px",     OPTION_VAL_12PX_OR },
         { "13px",     OPTION_VAL_13PX_OR },
         { "14px",     OPTION_VAL_14PX_OR },
         { "15px",     OPTION_VAL_15PX_OR },
         { "16px",     OPTION_VAL_16PX_OR },
         { "17px",     OPTION_VAL_17PX_OR },
         { "18px",     OPTION_VAL_18PX_OR },
         { "19px",     OPTION_VAL_19PX_OR },
         { "20px",     OPTION_VAL_20PX_OR },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(image_offset),
      BEETLE_OPT_IMAGE_OFFSET_LABEL_OR,
      NULL,
      BEETLE_OPT_IMAGE_OFFSET_INFO_0_OR,
      NULL,
      "video",
      {
         { "-12px",    OPTION_VAL_12PX_O64_OR },
         { "-11px",    OPTION_VAL_11PX_O64_OR },
         { "-10px",    OPTION_VAL_10PX_O64_OR },
         { "-9px",     OPTION_VAL_9PX_O64_OR },
         { "-8px",     OPTION_VAL_8PX_O64_OR },
         { "-7px",     OPTION_VAL_7PX_O64_OR },
         { "-6px",     OPTION_VAL_6PX_O64_OR },
         { "-5px",     OPTION_VAL_5PX_O64_OR },
         { "-4px",     OPTION_VAL_4PX_O64_OR },
         { "-3px",     OPTION_VAL_3PX_O64_OR },
         { "-2px",     OPTION_VAL_2PX_O64_OR },
         { "-1px",     OPTION_VAL_1PX_O64_OR },
         { "disabled", "0" },
         { "+1px",     OPTION_VAL_1PX_O64_0_OR },
         { "+2px",     OPTION_VAL_2PX_O64_0_OR },
         { "+3px",     OPTION_VAL_3PX_O64_0_OR },
         { "+4px",     OPTION_VAL_4PX_O64_0_OR },
         { "+5px",     OPTION_VAL_5PX_O64_0_OR },
         { "+6px",     OPTION_VAL_6PX_O64_0_OR },
         { "+7px",     OPTION_VAL_7PX_O64_0_OR },
         { "+8px",     OPTION_VAL_8PX_O64_0_OR },
         { "+9px",     OPTION_VAL_9PX_O64_0_OR },
         { "+10px",    OPTION_VAL_10PX_O64_0_OR },
         { "+11px",    OPTION_VAL_11PX_O64_0_OR },
         { "+12px",    OPTION_VAL_12PX_O64_0_OR },
         { NULL, NULL },
      },
      "disabled"
   },
#if defined(HAVE_OPENGL) || defined(HAVE_OPENGLES) || defined(HAVE_VULKAN)
   {
      BEETLE_OPT(image_offset_cycles),
      BEETLE_OPT_IMAGE_OFFSET_CYCLES_LABEL_OR,
      NULL,
      BEETLE_OPT_IMAGE_OFFSET_CYCLES_INFO_0_OR,
      NULL,
      "video",
      {
         { "-40",      NULL },
         { "-39",      NULL },
         { "-38",      NULL },
         { "-37",      NULL },
         { "-36",      NULL },
         { "-35",      NULL },
         { "-34",      NULL },
         { "-33",      NULL },
         { "-32",      NULL },
         { "-31",      NULL },
         { "-30",      NULL },
         { "-29",      NULL },
         { "-28",      NULL },
         { "-27",      NULL },
         { "-26",      NULL },
         { "-25",      NULL },
         { "-24",      NULL },
         { "-23",      NULL },
         { "-22",      NULL },
         { "-21",      NULL },
         { "-20",      NULL },
         { "-19",      NULL },
         { "-18",      NULL },
         { "-17",      NULL },
         { "-16",      NULL },
         { "-15",      NULL },
         { "-14",      NULL },
         { "-13",      NULL },
         { "-12",      NULL },
         { "-11",      NULL },
         { "-10",      NULL },
         { "-9",       NULL },
         { "-8",       NULL },
         { "-7",       NULL },
         { "-6",       NULL },
         { "-5",       NULL },
         { "-4",       NULL },
         { "-3",       NULL },
         { "-2",       NULL },
         { "-1",       NULL },
         { "0",        "0" },
         { "+1",       NULL },
         { "+2",       NULL },
         { "+3",       NULL },
         { "+4",       NULL },
         { "+5",       NULL },
         { "+6",       NULL },
         { "+7",       NULL },
         { "+8",       NULL },
         { "+9",       NULL },
         { "+10",      NULL },
         { "+11",      NULL },
         { "+12",      NULL },
         { "+13",      NULL },
         { "+14",      NULL },
         { "+15",      NULL },
         { "+16",      NULL },
         { "+17",      NULL },
         { "+18",      NULL },
         { "+19",      NULL },
         { "+20",      NULL },
         { "+21",      NULL },
         { "+22",      NULL },
         { "+23",      NULL },
         { "+24",      NULL },
         { "+25",      NULL },
         { "+26",      NULL },
         { "+27",      NULL },
         { "+28",      NULL },
         { "+29",      NULL },
         { "+30",      NULL },
         { "+31",      NULL },
         { "+32",      NULL },
         { "+33",      NULL },
         { "+34",      NULL },
         { "+35",      NULL },
         { "+36",      NULL },
         { "+37",      NULL },
         { "+38",      NULL },
         { "+39",      NULL },
         { "+40",      NULL },
         { NULL, NULL},
      },
      "0"
   },
#endif
   {
      BEETLE_OPT(initial_scanline),
      BEETLE_OPT_INITIAL_SCANLINE_LABEL_OR,
      NULL,
      BEETLE_OPT_INITIAL_SCANLINE_INFO_0_OR,
      NULL,
      "video",
      {
         { "0",  OPTION_VAL_0_O48_OR },
         { "1",  NULL },
         { "2",  NULL },
         { "3",  NULL },
         { "4",  NULL },
         { "5",  NULL },
         { "6",  NULL },
         { "7",  NULL },
         { "8",  NULL },
         { "9",  NULL },
         { "10", NULL },
         { "11", NULL },
         { "12", NULL },
         { "13", NULL },
         { "14", NULL },
         { "15", NULL },
         { "16", NULL },
         { "17", NULL },
         { "18", NULL },
         { "19", NULL },
         { "20", NULL },
         { "21", NULL },
         { "22", NULL },
         { "23", NULL },
         { "24", NULL },
         { "25", NULL },
         { "26", NULL },
         { "27", NULL },
         { "28", NULL },
         { "29", NULL },
         { "30", NULL },
         { "31", NULL },
         { "32", NULL },
         { "33", NULL },
         { "34", NULL },
         { "35", NULL },
         { "36", NULL },
         { "37", NULL },
         { "38", NULL },
         { "39", NULL },
         { "40", NULL },
         { NULL, NULL },
      },
      "0"
   },
   {
      BEETLE_OPT(last_scanline),
      BEETLE_OPT_LAST_SCANLINE_LABEL_OR,
      NULL,
      BEETLE_OPT_LAST_SCANLINE_INFO_0_OR,
      NULL,
      "video",
      {
         { "210", NULL },
         { "211", NULL },
         { "212", NULL },
         { "213", NULL },
         { "214", NULL },
         { "215", NULL },
         { "216", NULL },
         { "217", NULL },
         { "218", NULL },
         { "219", NULL },
         { "220", NULL },
         { "221", NULL },
         { "222", NULL },
         { "223", NULL },
         { "224", NULL },
         { "225", NULL },
         { "226", NULL },
         { "227", NULL },
         { "228", NULL },
         { "229", NULL },
         { "230", NULL },
         { "231", NULL },
         { "232", NULL },
         { "233", NULL },
         { "234", NULL },
         { "235", NULL },
         { "236", NULL },
         { "237", NULL },
         { "238", NULL },
         { "239", OPTION_VAL_239_OR },
         { NULL, NULL },
      },
      "239"
   },
   {
      BEETLE_OPT(initial_scanline_pal),
      BEETLE_OPT_INITIAL_SCANLINE_PAL_LABEL_OR,
      NULL,
      BEETLE_OPT_INITIAL_SCANLINE_PAL_INFO_0_OR,
      NULL,
      "video",
      {
         { "0",  OPTION_VAL_0_O48_OR },
         { "1",  NULL },
         { "2",  NULL },
         { "3",  NULL },
         { "4",  NULL },
         { "5",  NULL },
         { "6",  NULL },
         { "7",  NULL },
         { "8",  NULL },
         { "9",  NULL },
         { "10", NULL },
         { "11", NULL },
         { "12", NULL },
         { "13", NULL },
         { "14", NULL },
         { "15", NULL },
         { "16", NULL },
         { "17", NULL },
         { "18", NULL },
         { "19", NULL },
         { "20", NULL },
         { "21", NULL },
         { "22", NULL },
         { "23", NULL },
         { "24", NULL },
         { "25", NULL },
         { "26", NULL },
         { "27", NULL },
         { "28", NULL },
         { "29", NULL },
         { "30", NULL },
         { "31", NULL },
         { "32", NULL },
         { "33", NULL },
         { "34", NULL },
         { "35", NULL },
         { "36", NULL },
         { "37", NULL },
         { "38", NULL },
         { "39", NULL },
         { "40", NULL },
         { NULL, NULL },
      },
      "0"
   },
   {
      BEETLE_OPT(last_scanline_pal),
      BEETLE_OPT_LAST_SCANLINE_PAL_LABEL_OR,
      NULL,
      BEETLE_OPT_LAST_SCANLINE_PAL_INFO_0_OR,
      NULL,
      "video",
      {
         { "230", NULL },
         { "231", NULL },
         { "232", NULL },
         { "233", NULL },
         { "234", NULL },
         { "235", NULL },
         { "236", NULL },
         { "237", NULL },
         { "238", NULL },
         { "239", NULL },
         { "240", NULL },
         { "241", NULL },
         { "242", NULL },
         { "243", NULL },
         { "244", NULL },
         { "245", NULL },
         { "246", NULL },
         { "247", NULL },
         { "248", NULL },
         { "249", NULL },
         { "250", NULL },
         { "251", NULL },
         { "252", NULL },
         { "253", NULL },
         { "254", NULL },
         { "255", NULL },
         { "256", NULL },
         { "257", NULL },
         { "258", NULL },
         { "259", NULL },
         { "260", NULL },
         { "261", NULL },
         { "262", NULL },
         { "263", NULL },
         { "264", NULL },
         { "265", NULL },
         { "266", NULL },
         { "267", NULL },
         { "268", NULL },
         { "269", NULL },
         { "270", NULL },
         { "271", NULL },
         { "272", NULL },
         { "273", NULL },
         { "274", NULL },
         { "275", NULL },
         { "276", NULL },
         { "277", NULL },
         { "278", NULL },
         { "279", NULL },
         { "280", NULL },
         { "281", NULL },
         { "282", NULL },
         { "283", NULL },
         { "284", NULL },
         { "285", NULL },
         { "286", NULL },
         { "287", OPTION_VAL_287_OR },
         { NULL, NULL },
      },
      "287"
   },
   { NULL, NULL, NULL, NULL, NULL, NULL, {{0}}, NULL },
};
struct retro_core_options_v2 options_or = {
   option_cats_or,
   option_defs_or
};

/* RETRO_LANGUAGE_PL */

#define CATEGORY_SYSTEM_LABEL_PL NULL
#define CATEGORY_SYSTEM_INFO_0_PL "Zmień system i opcje przetaktowania."
#define CATEGORY_VIDEO_LABEL_PL "Wideo"
#define CATEGORY_VIDEO_INFO_0_PL "Zmień proporcje obrazu, przycięcie obrazu, filtry wideo i ustawienia pomijania klatek."
#define CATEGORY_PGXP_LABEL_PL "PGXP (Precyzyjne Przekształcanie Geometrii)"
#define CATEGORY_PGXP_INFO_0_PL "Te opcje wpływają na ulepszenia, które mogą poprawić grafikę w porównaniu z oryginalną konsolą. PGXP pomaga pozbyć się pływających tekstur i przenikania obiektów."
#define CATEGORY_OSD_LABEL_PL "Wyświetlane na ekranie"
#define CATEGORY_OSD_INFO_0_PL "Zmień powiadomienia wyświetlane na ekranie."
#define CATEGORY_INPUT_LABEL_PL "Urządzenia sterowania"
#define CATEGORY_INPUT_INFO_0_PL "Zmień ustawienia pistoletu, myszy i neGcon."
#define CATEGORY_MEMCARDS_LABEL_PL "Karta pamięci"
#define CATEGORY_MEMCARDS_INFO_0_PL "Zmień ustawienia związane z kartą pamięci wirtualnej."
#define CATEGORY_HACKS_LABEL_PL "Hacki emulacyjne"
#define CATEGORY_HACKS_INFO_0_PL "Zmiana ustawień dokładności emulacji, mających wpływ na niskopoziomową wydajność i kompatybilność."
#define BEETLE_OPT_CPU_FREQ_SCALE_LABEL_PL "Skalowanie częstotliwości procesora (Overclock)"
#define BEETLE_OPT_CPU_FREQ_SCALE_INFO_0_PL "Overclock (lub underclock) emulowanego procesora PSX. Przełączanie może wyeliminować spowolnienie i poprawić częstotliwość klatek w niektórych grach kosztem zwiększonych wymagań w zakresie wydajności. Zauważ, że niektóre gry mają wewnętrzny ogranicznik przepustowości klatek i mogą nie odnosić korzyści z przełączenia. Może powodować przyspieszoną animację niektórych efektów - szybciej niż jest to przewidziane w niektórych tytułach po przełączeniu."
#define OPTION_VAL_50_PL "50 %"
#define OPTION_VAL_60_PL "60 %"
#define OPTION_VAL_70_PL NULL
#define OPTION_VAL_80_PL "80 %"
#define OPTION_VAL_90_PL NULL
#define OPTION_VAL_100_PL "100% (Natywna)"
#define OPTION_VAL_110_PL NULL
#define OPTION_VAL_120_PL NULL
#define OPTION_VAL_130_PL NULL
#define OPTION_VAL_140_PL NULL
#define OPTION_VAL_150_PL NULL
#define OPTION_VAL_160_PL NULL
#define OPTION_VAL_170_PL NULL
#define OPTION_VAL_180_PL NULL
#define OPTION_VAL_190_PL NULL
#define OPTION_VAL_200_PL NULL
#define OPTION_VAL_210_PL NULL
#define OPTION_VAL_220_PL NULL
#define OPTION_VAL_230_PL NULL
#define OPTION_VAL_240_PL NULL
#define OPTION_VAL_250_PL NULL
#define OPTION_VAL_260_PL NULL
#define OPTION_VAL_270_PL NULL
#define OPTION_VAL_280_PL NULL
#define OPTION_VAL_290_PL NULL
#define OPTION_VAL_300_PL NULL
#define OPTION_VAL_310_PL NULL
#define OPTION_VAL_320_PL NULL
#define OPTION_VAL_330_PL NULL
#define OPTION_VAL_340_PL NULL
#define OPTION_VAL_350_PL NULL
#define OPTION_VAL_360_PL NULL
#define OPTION_VAL_370_PL NULL
#define OPTION_VAL_380_PL NULL
#define OPTION_VAL_390_PL NULL
#define OPTION_VAL_400_PL NULL
#define OPTION_VAL_410_PL NULL
#define OPTION_VAL_420_PL NULL
#define OPTION_VAL_430_PL NULL
#define OPTION_VAL_440_PL NULL
#define OPTION_VAL_450_PL NULL
#define OPTION_VAL_460_PL NULL
#define OPTION_VAL_470_PL NULL
#define OPTION_VAL_480_PL NULL
#define OPTION_VAL_490_PL NULL
#define OPTION_VAL_500_PL NULL
#define OPTION_VAL_510_PL NULL
#define OPTION_VAL_520_PL NULL
#define OPTION_VAL_530_PL NULL
#define OPTION_VAL_540_PL NULL
#define OPTION_VAL_550_PL NULL
#define OPTION_VAL_560_PL NULL
#define OPTION_VAL_570_PL NULL
#define OPTION_VAL_580_PL NULL
#define OPTION_VAL_590_PL NULL
#define OPTION_VAL_600_PL NULL
#define OPTION_VAL_610_PL NULL
#define OPTION_VAL_620_PL NULL
#define OPTION_VAL_630_PL NULL
#define OPTION_VAL_640_PL NULL
#define OPTION_VAL_650_PL NULL
#define OPTION_VAL_660_PL NULL
#define OPTION_VAL_670_PL NULL
#define OPTION_VAL_680_PL NULL
#define OPTION_VAL_690_PL NULL
#define OPTION_VAL_700_PL NULL
#define OPTION_VAL_710_PL NULL
#define OPTION_VAL_720_PL NULL
#define OPTION_VAL_730_PL NULL
#define OPTION_VAL_740_PL NULL
#define OPTION_VAL_750_PL NULL
#define BEETLE_OPT_GPU_OVERCLOCK_LABEL_PL NULL
#define BEETLE_OPT_GPU_OVERCLOCK_INFO_0_PL NULL
#define OPTION_VAL_1X_NATIVE_PL "1x (Natywna)"
#define OPTION_VAL_2X_PL NULL
#define OPTION_VAL_4X_PL NULL
#define OPTION_VAL_8X_PL NULL
#define OPTION_VAL_16X_PL NULL
#define OPTION_VAL_32X_PL NULL
#define BEETLE_OPT_GTE_OVERCLOCK_LABEL_PL NULL
#define BEETLE_OPT_GTE_OVERCLOCK_INFO_0_PL NULL
#define BEETLE_OPT_CD_ACCESS_METHOD_LABEL_PL NULL
#define BEETLE_OPT_CD_ACCESS_METHOD_INFO_0_PL NULL
#define OPTION_VAL_SYNC_PL NULL
#define OPTION_VAL_ASYNC_PL NULL
#define OPTION_VAL_PRECACHE_PL NULL
#define BEETLE_OPT_CD_FASTLOAD_LABEL_PL NULL
#define BEETLE_OPT_CD_FASTLOAD_INFO_0_PL NULL
#define OPTION_VAL_2X_NATIVE_PL NULL
#define OPTION_VAL_6X_PL NULL
#define OPTION_VAL_10X_PL NULL
#define OPTION_VAL_12X_PL NULL
#define OPTION_VAL_14X_PL NULL
#define BEETLE_OPT_SKIP_BIOS_LABEL_PL NULL
#define BEETLE_OPT_SKIP_BIOS_INFO_0_PL NULL
#define BEETLE_OPT_OVERRIDE_BIOS_LABEL_PL NULL
#define BEETLE_OPT_OVERRIDE_BIOS_INFO_0_PL NULL
#define OPTION_VAL_PSXONPSP_PL NULL
#define OPTION_VAL_PS1_ROM_PL NULL
#define OPTION_VAL_OPENBIOS_PL NULL
#define BEETLE_OPT_REGION_LABEL_PL "Region systemowy"
#define BEETLE_OPT_REGION_INFO_0_PL NULL
#define OPTION_VAL_AUTO_PL "Automatyczny"
#define OPTION_VAL_NTSC_J_PL NULL
#define OPTION_VAL_NTSC_U_PL NULL
#define OPTION_VAL_PAL_PL NULL
#define BEETLE_OPT_INTERNAL_RESOLUTION_LABEL_PL "Wewnętrzna rozdzielczość GPU"
#define BEETLE_OPT_INTERNAL_RESOLUTION_INFO_0_PL NULL
#define BEETLE_OPT_RENDERER_LABEL_PL NULL
#define BEETLE_OPT_RENDERER_INFO_0_PL NULL
#define OPTION_VAL_HARDWARE_PL NULL
#define OPTION_VAL_HARDWARE_GL_PL NULL
#define OPTION_VAL_HARDWARE_VK_PL NULL
#define OPTION_VAL_SOFTWARE_PL NULL
#define BEETLE_OPT_RENDERER_SOFTWARE_FB_LABEL_PL NULL
#define BEETLE_OPT_RENDERER_SOFTWARE_FB_INFO_0_PL "Włącz dokładną emulację efektów framebufora (np. ruch rozmyty, wir bitwy FF7) podczas używania sprzętowego renderowania poprzez uruchomienie kopii software render w natywnej rozdzielczości w tle. Jeśli wyłączone, te operacje są pomijane (OpenGL) lub renderowane na GPU (Vulkan). Wyłączenie może poprawić wydajność, ale może powodować poważne błędy graficzne. Pozostaw włączone, jeśli nie ma pewności."
#define BEETLE_OPT_DEPTH_LABEL_PL "Wewnętrzna głębia koloru"
#define BEETLE_OPT_DEPTH_INFO_0_PL "Wybierz wewnętrzną głębię kolorów. Wyższa głębia kolorów może zredukować efekt bandingu bez użycia ditheringu. 16 bpp emuluje oryginalny sprzęt, ale może mieć widoczny banding, jeśli dithering nie jest włączony. Zaleca się wyłączenie opcji \"Wzorzec ditheringu\", gdy ta opcja jest ustawiona na 32 bpp."
#define OPTION_VAL_16BPP_NATIVE_PL "16 bpp (natywna)"
#define OPTION_VAL_32BPP_PL NULL
#define BEETLE_OPT_DITHER_MODE_LABEL_PL "Wzorzec ditheringu"
#define BEETLE_OPT_DITHER_MODE_INFO_0_PL "Wybierz konfigurację wzorca ditheringu. \"1x (Natywny)\" emuluje natywny dithering o niskiej rozdzielczości używany przez oryginalny sprzęt do wygładzania artefaktów bandingu kolorów widocznych przy natywnej głębi kolorów. \"Wewnętrzna rozdzielczość\" skaluje ziarnistość ditheringu do skonfigurowanej rozdzielczości wewnętrznej dla czystszych rezultatów. Zaleca się wyłączenie tej funkcji, gdy pracujemy z głębią kolorów 32 bpp. Uwaga: W przypadku Vulkan, włączenie tej opcji wymusi downsampling do natywnej głębi kolorów, natomiast wyłączenie spowoduje automatyczne włączenie wyjścia przy wyższej głębi kolorów."
#define OPTION_VAL_INTERNAL_RESOLUTION_PL "Wewnętrzna rozdzielczość"
#define BEETLE_OPT_SCALED_UV_OFFSET_LABEL_PL "Przesunięcie UV tekstur"
#define BEETLE_OPT_SCALED_UV_OFFSET_INFO_0_PL "Próbkowanie tekstur dla wielokątów 3D z przesunięciem dla wyższej niż 1x wewnętrznej rozdzielczości. Redukuje przeszycia tekstur, ale może powodować niezamierzone błędy graficzne."
#define BEETLE_OPT_FILTER_LABEL_PL "Filtrowanie tekstur"
#define BEETLE_OPT_FILTER_INFO_0_PL "Wybierz metodę filtrowania tekstur. \"Najbliższe\" emuluje oryginalny sprzęt. Filtry \"Dwuliniowy\" i \"Trzypunktowy\" to filtry wygładzające, które redukują pikselizację poprzez rozmycie. SABR, xBR i JINC2 to filtry skalujące, które mogą poprawić wierność/ostrość tekstur kosztem zwiększonych wymagań wydajnościowych. Obsługiwane tylko przez renderery sprzętowe."
#define OPTION_VAL_NEAREST_PL "Najbliższe"
#define OPTION_VAL_SABR_PL NULL
#define OPTION_VAL_XBR_PL NULL
#define OPTION_VAL_BILINEAR_PL "Dwuliniowy"
#define OPTION_VAL_3_POINT_PL "Trzypunktowy"
#define OPTION_VAL_JINC2_PL NULL
#define BEETLE_OPT_FILTER_EXCLUDE_SPRITE_LABEL_PL "Wyklucz sprite'y z filtrowania"
#define BEETLE_OPT_FILTER_EXCLUDE_SPRITE_INFO_0_PL "Nie stosuje filtrowania tekstur do sprite'ów. Zapobiega powstawaniu przeszyć w różnych grach z 2D tłem renderowanym za pomocą sprite'ów. Użyj razem z Wygładzaniem adaptacyjnym lub innym filtrem post-processingowym, aby uzyskać najlepszy efekt."
#define OPTION_VAL_OPAQUE_PL "Tylko nieprzezroczyste"
#define OPTION_VAL_ALL_PL "Nieprzezroczyste i półprzezroczyste"
#define BEETLE_OPT_FILTER_EXCLUDE_2D_POLYGON_LABEL_PL "Wyklucz wielokąty 2D z filtrowania"
#define BEETLE_OPT_FILTER_EXCLUDE_2D_POLYGON_INFO_0_PL "Nie nakłada filtrowania tekstur na wielokąty 2D. Wielokąty 2D są wykrywane za pomocą heurystyki i mogą występować błędy. Aby uzyskać najlepszy efekt, zastosuj filtr Wygładzanie adaptacyjne lub inny filtr post-processingu."
#define BEETLE_OPT_ADAPTIVE_SMOOTHING_LABEL_PL "Wygładzanie adaptacyjne"
#define BEETLE_OPT_ADAPTIVE_SMOOTHING_INFO_0_PL "Wygładza grafiki 2D i elementy UI bez rozmycia renderowanych obiektów 3D. Obsługiwane tylko przez renderer Vulkan."
#define BEETLE_OPT_SUPER_SAMPLING_LABEL_PL "Supersampling (Downsampling do natywnej rozdzielczości)"
#define BEETLE_OPT_SUPER_SAMPLING_INFO_0_PL "Obniżanie próbkowania renderowanej zawartości z podniesionej rozdzielczości wewnętrznej do rozdzielczości natywnej. Połączenie tego z wyższymi mnożnikami rozdzielczości wewnętrznej pozwala na wyświetlanie gier z antyaliasowanymi obiektami 3D w natywnej niskiej rozdzielczości. Najlepsze rezultaty uzyskuje się w grach, w których mieszają się elementy 2D i 3D (np. postacie 3D na wstępnie wyrenderowanych tłach), a także w połączeniu z shaderami CRT. Obsługiwane tylko przez renderer Vulkan.UWAGA: Zaleca się wyłączenie opcji 'Wzorzec ditheringu' przy włączaniu tej opcji."
#define BEETLE_OPT_MSAA_LABEL_PL "Antyaliasing wielopróbkowy (MSAA)"
#define BEETLE_OPT_MSAA_INFO_0_PL "MSAA poprawia wygląd obiektów 3D. Obsługiwane tylko przez render Vulkanem."
#define OPTION_VAL_1X_PL NULL
#define BEETLE_OPT_MDEC_YUV_LABEL_PL "Filtr chromatograficzny MDEC YUV"
#define BEETLE_OPT_MDEC_YUV_INFO_0_PL "Poprawa jakości odtwarzania FMV poprzez redukcję artefaktów \"makroblokowania\" (kwadraty/poszarpane krawędzie). Obsługiwane tylko przez renderer Vulkan."
#define BEETLE_OPT_TRACK_TEXTURES_LABEL_PL NULL
#define BEETLE_OPT_TRACK_TEXTURES_INFO_0_PL NULL
#define BEETLE_OPT_DUMP_TEXTURES_LABEL_PL "Zgraj tekstury"
#define BEETLE_OPT_DUMP_TEXTURES_INFO_0_PL "Zrzuć użyte tekstury do <cd>-texture-dump/"
#define BEETLE_OPT_REPLACE_TEXTURES_LABEL_PL "Zastąp tekstury"
#define BEETLE_OPT_REPLACE_TEXTURES_INFO_0_PL "Zastąp tekstury za pomocą wersji HD z <cd>-texture-replacements/"
#define BEETLE_OPT_PGXP_MODE_LABEL_PL NULL
#define BEETLE_OPT_PGXP_MODE_INFO_0_PL NULL
#define OPTION_VAL_MEMORY_ONLY_PL "Tylko pamięć"
#define OPTION_VAL_MEMORY_CPU_PL "Pamięć + CPU (błędy)"
#define BEETLE_OPT_PGXP_2D_TOL_LABEL_PL "Tolerancja geometrii PGXP 2D"
#define BEETLE_OPT_PGXP_2D_TOL_INFO_0_PL NULL
#define OPTION_VAL_0PX_PL NULL
#define OPTION_VAL_1PX_PL NULL
#define OPTION_VAL_2PX_PL NULL
#define OPTION_VAL_3PX_PL NULL
#define OPTION_VAL_4PX_PL NULL
#define OPTION_VAL_5PX_PL NULL
#define OPTION_VAL_6PX_PL NULL
#define OPTION_VAL_7PX_PL NULL
#define OPTION_VAL_8PX_PL NULL
#define BEETLE_OPT_PGXP_NCLIP_LABEL_PL NULL
#define BEETLE_OPT_PGXP_NCLIP_INFO_0_PL NULL
#define BEETLE_OPT_PGXP_VERTEX_LABEL_PL NULL
#define BEETLE_OPT_PGXP_VERTEX_INFO_0_PL NULL
#define BEETLE_OPT_PGXP_TEXTURE_LABEL_PL NULL
#define BEETLE_OPT_PGXP_TEXTURE_INFO_0_PL NULL
#define BEETLE_OPT_DISPLAY_INTERNAL_FPS_LABEL_PL "Wyświetl wewnętrzne FPS"
#define BEETLE_OPT_DISPLAY_INTERNAL_FPS_INFO_0_PL "Wyświetlanie wewnętrznej szybkości klatek wyświetlania treści emulowanego systemu PlayStation. Uwaga: Wymagane jest włączanie powiadomień na ekranie w front-endzie libretro."
#define BEETLE_OPT_DISPLAY_OSD_LABEL_PL "Wyświetlaj wiadomości OSD"
#define BEETLE_OPT_DISPLAY_OSD_INFO_0_PL "Wyświetlaj wiadomości OSD generowane przez rdzeń."
#define BEETLE_OPT_DISPLAY_VRAM_LABEL_PL "Wyświetl pełną pamięć VRAM (Debug)"
#define BEETLE_OPT_DISPLAY_VRAM_INFO_0_PL NULL
#define BEETLE_OPT_ANALOG_CALIBRATION_LABEL_PL NULL
#define BEETLE_OPT_ANALOG_CALIBRATION_INFO_0_PL NULL
#define BEETLE_OPT_ANALOG_TOGGLE_LABEL_PL "Przełącznik trybu analogowego DualShock"
#define BEETLE_OPT_ANALOG_TOGGLE_INFO_0_PL "Gdy typ urządzenia wejściowego to DualShock, wtedy ta opcja pozwala na przełączanie emulowanej funkcji DualShock pomiędzy trybem DIGITAL i ANALOG, jak oryginalny sprzęt. Tryb może być przełączony przez wybraną kombinację przycisków."
#define OPTION_VAL_ENABLED_ANALOG_PL "Domyślny analogowy"
#define BEETLE_OPT_ANALOG_TOGGLE_COMBO_LABEL_PL NULL
#define BEETLE_OPT_ANALOG_TOGGLE_COMBO_INFO_0_PL NULL
#define OPTION_VAL_L1_L2_R1_R2_START_SELECT_PL NULL
#define OPTION_VAL_L1_R1_SELECT_PL NULL
#define OPTION_VAL_L1_R1_START_PL NULL
#define OPTION_VAL_L1_R1_L3_PL NULL
#define OPTION_VAL_L1_R1_R3_PL NULL
#define OPTION_VAL_L2_R2_SELECT_PL NULL
#define OPTION_VAL_L2_R2_START_PL NULL
#define OPTION_VAL_L2_R2_L3_PL NULL
#define OPTION_VAL_L2_R2_R3_PL NULL
#define OPTION_VAL_L3_R3_PL NULL
#define BEETLE_OPT_ANALOG_TOGGLE_HOLD_LABEL_PL NULL
#define BEETLE_OPT_ANALOG_TOGGLE_HOLD_INFO_0_PL NULL
#define OPTION_VAL_0_PL "0 sekund opóźnienia"
#define OPTION_VAL_1_PL "1 sekunda opóźnienia"
#define OPTION_VAL_2_PL "2 sekundy opóźnienia"
#define OPTION_VAL_3_PL "3 sekundy opóźnienia"
#define OPTION_VAL_4_PL "4 sekundy opóźnienia"
#define OPTION_VAL_5_PL "5 sekund opóźnienia"
#define BEETLE_OPT_ENABLE_MULTITAP_PORT1_LABEL_PL NULL
#define BEETLE_OPT_ENABLE_MULTITAP_PORT1_INFO_0_PL NULL
#define BEETLE_OPT_ENABLE_MULTITAP_PORT2_LABEL_PL NULL
#define BEETLE_OPT_ENABLE_MULTITAP_PORT2_INFO_0_PL NULL
#define BEETLE_OPT_GUN_INPUT_MODE_LABEL_PL NULL
#define BEETLE_OPT_GUN_INPUT_MODE_INFO_0_PL NULL
#define OPTION_VAL_LIGHTGUN_PL "Pistolet Świetlny"
#define OPTION_VAL_TOUCHSCREEN_PL "Ekran dotykowy"
#define BEETLE_OPT_GUN_CURSOR_LABEL_PL NULL
#define BEETLE_OPT_GUN_CURSOR_INFO_0_PL NULL
#define OPTION_VAL_CROSS_PL NULL
#define OPTION_VAL_DOT_PL NULL
#define BEETLE_OPT_CROSSHAIR_COLOR_P1_LABEL_PL NULL
#define BEETLE_OPT_CROSSHAIR_COLOR_P1_INFO_0_PL NULL
#define OPTION_VAL_RED_PL "Czerwony"
#define OPTION_VAL_BLUE_PL "Niebieski"
#define OPTION_VAL_GREEN_PL "Zielony"
#define OPTION_VAL_ORANGE_PL NULL
#define OPTION_VAL_YELLOW_PL "Żółty"
#define OPTION_VAL_CYAN_PL NULL
#define OPTION_VAL_PINK_PL NULL
#define OPTION_VAL_PURPLE_PL NULL
#define OPTION_VAL_BLACK_PL NULL
#define OPTION_VAL_WHITE_PL NULL
#define BEETLE_OPT_CROSSHAIR_COLOR_P2_LABEL_PL NULL
#define BEETLE_OPT_CROSSHAIR_COLOR_P2_INFO_0_PL NULL
#define BEETLE_OPT_MOUSE_SENSITIVITY_LABEL_PL "Czułość myszy"
#define BEETLE_OPT_MOUSE_SENSITIVITY_INFO_0_PL NULL
#define OPTION_VAL_5_O42_PL NULL
#define OPTION_VAL_10_PL NULL
#define OPTION_VAL_15_PL NULL
#define OPTION_VAL_20_PL NULL
#define OPTION_VAL_25_PL NULL
#define OPTION_VAL_30_PL "30 %"
#define OPTION_VAL_35_PL NULL
#define OPTION_VAL_40_PL NULL
#define OPTION_VAL_45_PL NULL
#define OPTION_VAL_55_PL NULL
#define OPTION_VAL_65_PL NULL
#define OPTION_VAL_75_PL NULL
#define OPTION_VAL_85_PL NULL
#define OPTION_VAL_95_PL NULL
#define OPTION_VAL_100_O42_PL NULL
#define OPTION_VAL_105_PL NULL
#define OPTION_VAL_115_PL NULL
#define OPTION_VAL_125_PL NULL
#define OPTION_VAL_135_PL NULL
#define OPTION_VAL_145_PL NULL
#define OPTION_VAL_155_PL NULL
#define OPTION_VAL_165_PL NULL
#define OPTION_VAL_175_PL NULL
#define OPTION_VAL_185_PL NULL
#define OPTION_VAL_195_PL NULL
#define BEETLE_OPT_NEGCON_RESPONSE_LABEL_PL "odpowiedź neGcon Twist"
#define BEETLE_OPT_NEGCON_RESPONSE_INFO_0_PL NULL
#define OPTION_VAL_LINEAR_PL "Liniowy"
#define OPTION_VAL_QUADRATIC_PL NULL
#define OPTION_VAL_CUBIC_PL NULL
#define BEETLE_OPT_NEGCON_DEADZONE_LABEL_PL NULL
#define BEETLE_OPT_NEGCON_DEADZONE_INFO_0_PL NULL
#define OPTION_VAL_0_O44_PL NULL
#define BEETLE_OPT_USE_MEDNAFEN_MEMCARD0_METHOD_LABEL_PL NULL
#define BEETLE_OPT_USE_MEDNAFEN_MEMCARD0_METHOD_INFO_0_PL NULL
#define OPTION_VAL_LIBRETRO_PL NULL
#define OPTION_VAL_MEDNAFEN_PL NULL
#define BEETLE_OPT_ENABLE_MEMCARD1_LABEL_PL NULL
#define BEETLE_OPT_ENABLE_MEMCARD1_INFO_0_PL NULL
#define BEETLE_OPT_SHARED_MEMORY_CARDS_LABEL_PL NULL
#define BEETLE_OPT_SHARED_MEMORY_CARDS_INFO_0_PL NULL
#define BEETLE_OPT_MEMCARD_LEFT_INDEX_LABEL_PL NULL
#define BEETLE_OPT_MEMCARD_LEFT_INDEX_INFO_0_PL NULL
#define OPTION_VAL_0_O48_PL NULL
#define BEETLE_OPT_MEMCARD_RIGHT_INDEX_LABEL_PL NULL
#define BEETLE_OPT_MEMCARD_RIGHT_INDEX_INFO_0_PL NULL
#define OPTION_VAL_1_O49_PL NULL
#define BEETLE_OPT_LINE_RENDER_LABEL_PL NULL
#define BEETLE_OPT_LINE_RENDER_INFO_0_PL NULL
#define OPTION_VAL_DEFAULT_PL "Domyślny"
#define OPTION_VAL_AGGRESSIVE_PL NULL
#define BEETLE_OPT_WIDESCREEN_HACK_LABEL_PL NULL
#define BEETLE_OPT_WIDESCREEN_HACK_INFO_0_PL NULL
#define BEETLE_OPT_WIDESCREEN_HACK_ASPECT_RATIO_LABEL_PL NULL
#define BEETLE_OPT_WIDESCREEN_HACK_ASPECT_RATIO_INFO_0_PL NULL
#define OPTION_VAL_16_9_PL NULL
#define OPTION_VAL_16_10_PL NULL
#define OPTION_VAL_18_9_PL NULL
#define OPTION_VAL_19_9_PL NULL
#define OPTION_VAL_20_9_PL NULL
#define OPTION_VAL_21_9_PL NULL
#define OPTION_VAL_32_9_PL NULL
#define BEETLE_OPT_CPU_DYNAREC_LABEL_PL NULL
#define BEETLE_OPT_CPU_DYNAREC_INFO_0_PL NULL
#define OPTION_VAL_DISABLED_PL NULL
#define OPTION_VAL_EXECUTE_PL NULL
#define OPTION_VAL_EXECUTE_ONE_PL NULL
#define OPTION_VAL_RUN_INTERPRETER_PL NULL
#define BEETLE_OPT_DYNAREC_INVALIDATE_LABEL_PL NULL
#define BEETLE_OPT_DYNAREC_INVALIDATE_INFO_0_PL NULL
#define OPTION_VAL_FULL_PL "Pełny"
#define OPTION_VAL_DMA_PL NULL
#define BEETLE_OPT_DYNAREC_EVENTCYCLES_LABEL_PL NULL
#define BEETLE_OPT_DYNAREC_EVENTCYCLES_INFO_0_PL NULL
#define OPTION_VAL_128_PL NULL
#define BEETLE_OPT_DYNAREC_SPU_SAMPLES_LABEL_PL NULL
#define BEETLE_OPT_DYNAREC_SPU_SAMPLES_INFO_0_PL NULL
#define BEETLE_OPT_CORE_TIMING_FPS_LABEL_PL NULL
#define BEETLE_OPT_CORE_TIMING_FPS_INFO_0_PL NULL
#define OPTION_VAL_AUTO_TOGGLE_PL NULL
#define OPTION_VAL_FORCE_PROGRESSIVE_PL NULL
#define OPTION_VAL_FORCE_INTERLACED_PL NULL
#define BEETLE_OPT_DEINTERLACER_LABEL_PL NULL
#define BEETLE_OPT_DEINTERLACER_INFO_0_PL NULL
#define OPTION_VAL_WEAVE_PL NULL
#define OPTION_VAL_BOB_PL NULL
#define OPTION_VAL_BOB_OFFSET_PL NULL
#define OPTION_VAL_FASTMAD_PL NULL
#define BEETLE_OPT_FRAME_DUPING_LABEL_PL NULL
#define BEETLE_OPT_FRAME_DUPING_INFO_0_PL NULL
#define BEETLE_OPT_PAL_VIDEO_TIMING_OVERRIDE_LABEL_PL NULL
#define BEETLE_OPT_PAL_VIDEO_TIMING_OVERRIDE_INFO_0_PL NULL
#define BEETLE_OPT_ASPECT_RATIO_LABEL_PL NULL
#define BEETLE_OPT_ASPECT_RATIO_INFO_0_PL NULL
#define OPTION_VAL_CORRECTED_PL NULL
#define OPTION_VAL_UNCORRECTED_PL "Niepoprawione"
#define OPTION_VAL_4_3_PL NULL
#define OPTION_VAL_NTSC_PL NULL
#define BEETLE_OPT_CROP_OVERSCAN_LABEL_PL "Przytnij overscan"
#define BEETLE_OPT_CROP_OVERSCAN_INFO_0_PL NULL
#define OPTION_VAL_STATIC_PL NULL
#define OPTION_VAL_SMART_PL NULL
#define BEETLE_OPT_IMAGE_CROP_LABEL_PL NULL
#define BEETLE_OPT_IMAGE_CROP_INFO_0_PL NULL
#define OPTION_VAL_9PX_PL NULL
#define OPTION_VAL_10PX_PL NULL
#define OPTION_VAL_11PX_PL NULL
#define OPTION_VAL_12PX_PL NULL
#define OPTION_VAL_13PX_PL NULL
#define OPTION_VAL_14PX_PL NULL
#define OPTION_VAL_15PX_PL NULL
#define OPTION_VAL_16PX_PL NULL
#define OPTION_VAL_17PX_PL NULL
#define OPTION_VAL_18PX_PL NULL
#define OPTION_VAL_19PX_PL NULL
#define OPTION_VAL_20PX_PL NULL
#define BEETLE_OPT_IMAGE_OFFSET_LABEL_PL NULL
#define BEETLE_OPT_IMAGE_OFFSET_INFO_0_PL NULL
#define OPTION_VAL_12PX_O64_PL NULL
#define OPTION_VAL_11PX_O64_PL NULL
#define OPTION_VAL_10PX_O64_PL NULL
#define OPTION_VAL_9PX_O64_PL NULL
#define OPTION_VAL_8PX_O64_PL NULL
#define OPTION_VAL_7PX_O64_PL NULL
#define OPTION_VAL_6PX_O64_PL NULL
#define OPTION_VAL_5PX_O64_PL NULL
#define OPTION_VAL_4PX_O64_PL NULL
#define OPTION_VAL_3PX_O64_PL NULL
#define OPTION_VAL_2PX_O64_PL NULL
#define OPTION_VAL_1PX_O64_PL NULL
#define OPTION_VAL_1PX_O64_0_PL NULL
#define OPTION_VAL_2PX_O64_0_PL NULL
#define OPTION_VAL_3PX_O64_0_PL NULL
#define OPTION_VAL_4PX_O64_0_PL NULL
#define OPTION_VAL_5PX_O64_0_PL NULL
#define OPTION_VAL_6PX_O64_0_PL NULL
#define OPTION_VAL_7PX_O64_0_PL NULL
#define OPTION_VAL_8PX_O64_0_PL NULL
#define OPTION_VAL_9PX_O64_0_PL NULL
#define OPTION_VAL_10PX_O64_0_PL NULL
#define OPTION_VAL_11PX_O64_0_PL NULL
#define OPTION_VAL_12PX_O64_0_PL NULL
#define BEETLE_OPT_IMAGE_OFFSET_CYCLES_LABEL_PL NULL
#define BEETLE_OPT_IMAGE_OFFSET_CYCLES_INFO_0_PL NULL
#define BEETLE_OPT_INITIAL_SCANLINE_LABEL_PL NULL
#define BEETLE_OPT_INITIAL_SCANLINE_INFO_0_PL NULL
#define BEETLE_OPT_LAST_SCANLINE_LABEL_PL NULL
#define BEETLE_OPT_LAST_SCANLINE_INFO_0_PL NULL
#define OPTION_VAL_239_PL "239 (domyślnie)"
#define BEETLE_OPT_INITIAL_SCANLINE_PAL_LABEL_PL NULL
#define BEETLE_OPT_INITIAL_SCANLINE_PAL_INFO_0_PL NULL
#define BEETLE_OPT_LAST_SCANLINE_PAL_LABEL_PL NULL
#define BEETLE_OPT_LAST_SCANLINE_PAL_INFO_0_PL NULL
#define OPTION_VAL_287_PL "287 (domyślnie)"

struct retro_core_option_v2_category option_cats_pl[] = {
   {
      "system",
      CATEGORY_SYSTEM_LABEL_PL,
      CATEGORY_SYSTEM_INFO_0_PL
   },
   {
      "video",
      CATEGORY_VIDEO_LABEL_PL,
      CATEGORY_VIDEO_INFO_0_PL
   },
   {
      "pgxp",
      CATEGORY_PGXP_LABEL_PL,
      CATEGORY_PGXP_INFO_0_PL
   },
   {
      "osd",
      CATEGORY_OSD_LABEL_PL,
      CATEGORY_OSD_INFO_0_PL
   },
   {
      "input",
      CATEGORY_INPUT_LABEL_PL,
      CATEGORY_INPUT_INFO_0_PL
   },
   {
      "memcards",
      CATEGORY_MEMCARDS_LABEL_PL,
      CATEGORY_MEMCARDS_INFO_0_PL
   },
   {
      "hacks",
      CATEGORY_HACKS_LABEL_PL,
      CATEGORY_HACKS_INFO_0_PL
   },
   { NULL, NULL, NULL },
};
struct retro_core_option_v2_definition option_defs_pl[] = {
   {
      BEETLE_OPT(cpu_freq_scale),
      BEETLE_OPT_CPU_FREQ_SCALE_LABEL_PL,
      NULL,
      BEETLE_OPT_CPU_FREQ_SCALE_INFO_0_PL,
      NULL,
      "system",
      {
         { "50%",           OPTION_VAL_50_PL },
         { "60%",           OPTION_VAL_60_PL },
         { "70%",           OPTION_VAL_70_PL },
         { "80%",           OPTION_VAL_80_PL },
         { "90%",           OPTION_VAL_90_PL },
         { "100%",          OPTION_VAL_100_PL },
         { "110%",          OPTION_VAL_110_PL },
         { "120%",          OPTION_VAL_120_PL },
         { "130%",          OPTION_VAL_130_PL },
         { "140%",          OPTION_VAL_140_PL },
         { "150%",          OPTION_VAL_150_PL },
         { "160%",          OPTION_VAL_160_PL },
         { "170%",          OPTION_VAL_170_PL },
         { "180%",          OPTION_VAL_180_PL },
         { "190%",          OPTION_VAL_190_PL },
         { "200%",          OPTION_VAL_200_PL },
         { "210%",          OPTION_VAL_210_PL },
         { "220%",          OPTION_VAL_220_PL },
         { "230%",          OPTION_VAL_230_PL },
         { "240%",          OPTION_VAL_240_PL },
         { "250%",          OPTION_VAL_250_PL },
         { "260%",          OPTION_VAL_260_PL },
         { "270%",          OPTION_VAL_270_PL },
         { "280%",          OPTION_VAL_280_PL },
         { "290%",          OPTION_VAL_290_PL },
         { "300%",          OPTION_VAL_300_PL },
         { "310%",          OPTION_VAL_310_PL },
         { "320%",          OPTION_VAL_320_PL },
         { "330%",          OPTION_VAL_330_PL },
         { "340%",          OPTION_VAL_340_PL },
         { "350%",          OPTION_VAL_350_PL },
         { "360%",          OPTION_VAL_360_PL },
         { "370%",          OPTION_VAL_370_PL },
         { "380%",          OPTION_VAL_380_PL },
         { "390%",          OPTION_VAL_390_PL },
         { "400%",          OPTION_VAL_400_PL },
         { "410%",          OPTION_VAL_410_PL },
         { "420%",          OPTION_VAL_420_PL },
         { "430%",          OPTION_VAL_430_PL },
         { "440%",          OPTION_VAL_440_PL },
         { "450%",          OPTION_VAL_450_PL },
         { "460%",          OPTION_VAL_460_PL },
         { "470%",          OPTION_VAL_470_PL },
         { "480%",          OPTION_VAL_480_PL },
         { "490%",          OPTION_VAL_490_PL },
         { "500%",          OPTION_VAL_500_PL },
         { "510%",          OPTION_VAL_510_PL },
         { "520%",          OPTION_VAL_520_PL },
         { "530%",          OPTION_VAL_530_PL },
         { "540%",          OPTION_VAL_540_PL },
         { "550%",          OPTION_VAL_550_PL },
         { "560%",          OPTION_VAL_560_PL },
         { "570%",          OPTION_VAL_570_PL },
         { "580%",          OPTION_VAL_580_PL },
         { "590%",          OPTION_VAL_590_PL },
         { "600%",          OPTION_VAL_600_PL },
         { "610%",          OPTION_VAL_610_PL },
         { "620%",          OPTION_VAL_620_PL },
         { "630%",          OPTION_VAL_630_PL },
         { "640%",          OPTION_VAL_640_PL },
         { "650%",          OPTION_VAL_650_PL },
         { "660%",          OPTION_VAL_660_PL },
         { "670%",          OPTION_VAL_670_PL },
         { "680%",          OPTION_VAL_680_PL },
         { "690%",          OPTION_VAL_690_PL },
         { "700%",          OPTION_VAL_700_PL },
         { "710%",          OPTION_VAL_710_PL },
         { "720%",          OPTION_VAL_720_PL },
         { "730%",          OPTION_VAL_730_PL },
         { "740%",          OPTION_VAL_740_PL },
         { "750%",          OPTION_VAL_750_PL },
         { NULL, NULL },
      },
      "100%"
   },
   {
      BEETLE_OPT(gpu_overclock),
      BEETLE_OPT_GPU_OVERCLOCK_LABEL_PL,
      NULL,
      BEETLE_OPT_GPU_OVERCLOCK_INFO_0_PL,
      NULL,
      "system",
      {
         { "1x(native)", OPTION_VAL_1X_NATIVE_PL },
         { "2x",         OPTION_VAL_2X_PL },
         { "4x",         OPTION_VAL_4X_PL },
         { "8x",         OPTION_VAL_8X_PL },
         { "16x",        OPTION_VAL_16X_PL },
         { "32x",        OPTION_VAL_32X_PL },
         { NULL, NULL },
      },
      "1x(native)"
   },
   {
      BEETLE_OPT(gte_overclock),
      BEETLE_OPT_GTE_OVERCLOCK_LABEL_PL,
      NULL,
      BEETLE_OPT_GTE_OVERCLOCK_INFO_0_PL,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
#ifndef EMSCRIPTEN
   {
      BEETLE_OPT(cd_access_method),
      BEETLE_OPT_CD_ACCESS_METHOD_LABEL_PL,
      NULL,
      BEETLE_OPT_CD_ACCESS_METHOD_INFO_0_PL,
      NULL,
      "system",
      {
         { "sync",     OPTION_VAL_SYNC_PL },
         { "async",    OPTION_VAL_ASYNC_PL },
         { "precache", OPTION_VAL_PRECACHE_PL },
         { NULL, NULL },
      },
      "sync"
   },
#endif
   {
      BEETLE_OPT(cd_fastload),
      BEETLE_OPT_CD_FASTLOAD_LABEL_PL,
      NULL,
      BEETLE_OPT_CD_FASTLOAD_INFO_0_PL,
      NULL,
      "system",
      {
         { "2x(native)", OPTION_VAL_2X_NATIVE_PL },
         { "4x",          OPTION_VAL_4X_PL },
         { "6x",          OPTION_VAL_6X_PL },
         { "8x",          OPTION_VAL_8X_PL },
         { "10x",         OPTION_VAL_10X_PL },
         { "12x",         OPTION_VAL_12X_PL },
         { "14x",         OPTION_VAL_14X_PL },
         { NULL, NULL },
      },
      "2x(native)"
   },
   {
      BEETLE_OPT(skip_bios),
      BEETLE_OPT_SKIP_BIOS_LABEL_PL,
      NULL,
      BEETLE_OPT_SKIP_BIOS_INFO_0_PL,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(override_bios),
      BEETLE_OPT_OVERRIDE_BIOS_LABEL_PL,
      NULL,
      BEETLE_OPT_OVERRIDE_BIOS_INFO_0_PL,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "psxonpsp", OPTION_VAL_PSXONPSP_PL },
         { "ps1_rom",  OPTION_VAL_PS1_ROM_PL },
         { "openbios", OPTION_VAL_OPENBIOS_PL     },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(region),
      BEETLE_OPT_REGION_LABEL_PL,
      NULL,
      BEETLE_OPT_REGION_INFO_0_PL,
      NULL,
      "system",
      {
         { "auto",   OPTION_VAL_AUTO_PL },
         { "ntsc-j", OPTION_VAL_NTSC_J_PL },
         { "ntsc-u", OPTION_VAL_NTSC_U_PL },
         { "pal",    OPTION_VAL_PAL_PL },
         { NULL, NULL },
      },
      "auto"
   },
   {
      BEETLE_OPT(internal_resolution),
      BEETLE_OPT_INTERNAL_RESOLUTION_LABEL_PL,
      NULL,
      BEETLE_OPT_INTERNAL_RESOLUTION_INFO_0_PL,
      NULL,
      "video",
      {
         { "1x(native)", OPTION_VAL_1X_NATIVE_PL },
         { "2x",         OPTION_VAL_2X_PL },
         { "4x",         OPTION_VAL_4X_PL },
         { "8x",         OPTION_VAL_8X_PL },
         { "16x",        OPTION_VAL_16X_PL },
         { NULL, NULL },
      },
      "1x(native)"
   },
#if defined(HAVE_OPENGL) || defined(HAVE_OPENGLES) || defined(HAVE_VULKAN)
   {
      BEETLE_OPT(renderer),
      BEETLE_OPT_RENDERER_LABEL_PL,
      NULL,
      BEETLE_OPT_RENDERER_INFO_0_PL,
      NULL,
      "video",
      {
         { "hardware",    OPTION_VAL_HARDWARE_PL },
#if defined(HAVE_OPENGL) || defined(HAVE_OPENGLES)
         { "hardware_gl", OPTION_VAL_HARDWARE_GL_PL },
#endif
#if defined(HAVE_VULKAN)
         { "hardware_vk", OPTION_VAL_HARDWARE_VK_PL },
#endif
         { "software",    OPTION_VAL_SOFTWARE_PL },
         { NULL, NULL },
      },
      "hardware"
   },
   {
      BEETLE_OPT(renderer_software_fb),
      BEETLE_OPT_RENDERER_SOFTWARE_FB_LABEL_PL,
      NULL,
      BEETLE_OPT_RENDERER_SOFTWARE_FB_INFO_0_PL,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled"
   },
#endif
#if defined(HAVE_OPENGL) || defined(HAVE_OPENGLES)
   {
      BEETLE_OPT(depth),
      BEETLE_OPT_DEPTH_LABEL_PL,
      NULL,
      BEETLE_OPT_DEPTH_INFO_0_PL,
      NULL,
      "video",
      {
         { "16bpp(native)", OPTION_VAL_16BPP_NATIVE_PL },
         { "32bpp",         OPTION_VAL_32BPP_PL },
         { NULL, NULL },
      },
      "16bpp(native)"
   },
   // Sort of, it's more like 15-bit high color and 24-bit true color for visible output. The alpha channel is used for mask bit. Vulkan renderer uses ABGR1555_555 for 31 bits internal? FMVs are always 24-bit on all renderers like original hardware (BGR888, no alpha)
#endif
   {
      BEETLE_OPT(dither_mode),
      BEETLE_OPT_DITHER_MODE_LABEL_PL,
      NULL,
      BEETLE_OPT_DITHER_MODE_INFO_0_PL,
      NULL,
      "video",
      {
         { "1x(native)",          OPTION_VAL_1X_NATIVE_PL },
         { "internal resolution", OPTION_VAL_INTERNAL_RESOLUTION_PL },
         { "disabled",            NULL },
         { NULL, NULL },
      },
      "1x(native)"
   },
#ifdef HAVE_VULKAN
   {
      BEETLE_OPT(scaled_uv_offset),
      BEETLE_OPT_SCALED_UV_OFFSET_LABEL_PL,
      NULL,
      BEETLE_OPT_SCALED_UV_OFFSET_INFO_0_PL,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled"
   },
#endif
#if defined(HAVE_OPENGL) || defined(HAVE_OPENGLES) || defined(HAVE_VULKAN)
   {
      BEETLE_OPT(filter),
      BEETLE_OPT_FILTER_LABEL_PL,
      NULL,
      BEETLE_OPT_FILTER_INFO_0_PL,
      NULL,
      "video",
      {
         { "nearest",  OPTION_VAL_NEAREST_PL },
         { "SABR",     OPTION_VAL_SABR_PL },
         { "xBR",      OPTION_VAL_XBR_PL },
         { "bilinear", OPTION_VAL_BILINEAR_PL },
         { "3-point",  OPTION_VAL_3_POINT_PL },
         { "JINC2",    OPTION_VAL_JINC2_PL },
         { NULL, NULL },
      },
      "nearest"
   },
#ifdef HAVE_VULKAN
   {
      BEETLE_OPT(filter_exclude_sprite),
      BEETLE_OPT_FILTER_EXCLUDE_SPRITE_LABEL_PL,
      NULL,
      BEETLE_OPT_FILTER_EXCLUDE_SPRITE_INFO_0_PL,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "opaque", OPTION_VAL_OPAQUE_PL },
         { "all", OPTION_VAL_ALL_PL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(filter_exclude_2d_polygon),
      BEETLE_OPT_FILTER_EXCLUDE_2D_POLYGON_LABEL_PL,
      NULL,
      BEETLE_OPT_FILTER_EXCLUDE_2D_POLYGON_INFO_0_PL,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "opaque", OPTION_VAL_OPAQUE_PL },
         { "all", OPTION_VAL_ALL_PL },
         { NULL, NULL },
      },
      "disabled"
   },
#endif
#endif
#ifdef HAVE_VULKAN
   {
      BEETLE_OPT(adaptive_smoothing),
      BEETLE_OPT_ADAPTIVE_SMOOTHING_LABEL_PL,
      NULL,
      BEETLE_OPT_ADAPTIVE_SMOOTHING_INFO_0_PL,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(super_sampling),
      BEETLE_OPT_SUPER_SAMPLING_LABEL_PL,
      NULL,
      BEETLE_OPT_SUPER_SAMPLING_INFO_0_PL,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(msaa),
      BEETLE_OPT_MSAA_LABEL_PL,
      NULL,
      BEETLE_OPT_MSAA_INFO_0_PL,
      NULL,
      "video",
      {
         { "1x",  OPTION_VAL_1X_PL },
         { "2x",  OPTION_VAL_2X_PL },
         { "4x",  OPTION_VAL_4X_PL },
         { "8x",  OPTION_VAL_8X_PL },
         { "16x", OPTION_VAL_16X_PL },
         { NULL, NULL },
      },
      "1x"
   },
   {
      BEETLE_OPT(mdec_yuv),
      BEETLE_OPT_MDEC_YUV_LABEL_PL,
      NULL,
      BEETLE_OPT_MDEC_YUV_INFO_0_PL,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(track_textures),
      BEETLE_OPT_TRACK_TEXTURES_LABEL_PL,
      NULL,
      BEETLE_OPT_TRACK_TEXTURES_INFO_0_PL,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
#ifdef TEXTURE_DUMPING_ENABLED
   {
      BEETLE_OPT(dump_textures),
      BEETLE_OPT_DUMP_TEXTURES_LABEL_PL,
      NULL,
      BEETLE_OPT_DUMP_TEXTURES_INFO_0_PL,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
#endif
   {
      BEETLE_OPT(replace_textures),
      BEETLE_OPT_REPLACE_TEXTURES_LABEL_PL,
      NULL,
      BEETLE_OPT_REPLACE_TEXTURES_INFO_0_PL,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
#endif
   {
      BEETLE_OPT(pgxp_mode),
      BEETLE_OPT_PGXP_MODE_LABEL_PL,
      NULL,
      BEETLE_OPT_PGXP_MODE_INFO_0_PL,
      NULL,
      "pgxp",
      {
         { "disabled",     NULL },
         { "memory only",  OPTION_VAL_MEMORY_ONLY_PL },
         { "memory + CPU", OPTION_VAL_MEMORY_CPU_PL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(pgxp_2d_tol),
      BEETLE_OPT_PGXP_2D_TOL_LABEL_PL,
      NULL,
      BEETLE_OPT_PGXP_2D_TOL_INFO_0_PL,
      NULL,
      "pgxp",
      {
         { "disabled", NULL },
         { "0px", OPTION_VAL_0PX_PL },
         { "1px", OPTION_VAL_1PX_PL },
         { "2px", OPTION_VAL_2PX_PL },
         { "3px", OPTION_VAL_3PX_PL },
         { "4px", OPTION_VAL_4PX_PL },
         { "5px", OPTION_VAL_5PX_PL },
         { "6px", OPTION_VAL_6PX_PL },
         { "7px", OPTION_VAL_7PX_PL },
         { "8px", OPTION_VAL_8PX_PL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(pgxp_nclip),
      BEETLE_OPT_PGXP_NCLIP_LABEL_PL,
      NULL,
      BEETLE_OPT_PGXP_NCLIP_INFO_0_PL,
      NULL,
      "pgxp",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
#if defined(HAVE_OPENGL) || defined(HAVE_OPENGLES) || defined(HAVE_VULKAN)
   {
      BEETLE_OPT(pgxp_vertex),
      BEETLE_OPT_PGXP_VERTEX_LABEL_PL,
      NULL,
      BEETLE_OPT_PGXP_VERTEX_INFO_0_PL,
      NULL,
      "pgxp",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
#endif
   {
      BEETLE_OPT(pgxp_texture),
      BEETLE_OPT_PGXP_TEXTURE_LABEL_PL,
      NULL,
      BEETLE_OPT_PGXP_TEXTURE_INFO_0_PL,
      NULL,
      "pgxp",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(display_internal_fps),
      BEETLE_OPT_DISPLAY_INTERNAL_FPS_LABEL_PL,
      NULL,
      BEETLE_OPT_DISPLAY_INTERNAL_FPS_INFO_0_PL,
      NULL,
      "osd",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(display_osd),
      BEETLE_OPT_DISPLAY_OSD_LABEL_PL,
      NULL,
      BEETLE_OPT_DISPLAY_OSD_INFO_0_PL,
      NULL,
      "osd",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled"
   },
#if defined(HAVE_OPENGL) || defined(HAVE_OPENGLES) || defined(HAVE_VULKAN)
   {
      BEETLE_OPT(display_vram),
      BEETLE_OPT_DISPLAY_VRAM_LABEL_PL,
      NULL,
      BEETLE_OPT_DISPLAY_VRAM_INFO_0_PL,
      NULL,
      "osd",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
#endif
   {
      BEETLE_OPT(analog_calibration),
      BEETLE_OPT_ANALOG_CALIBRATION_LABEL_PL,
      NULL,
      BEETLE_OPT_ANALOG_CALIBRATION_INFO_0_PL,
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(analog_toggle),
      BEETLE_OPT_ANALOG_TOGGLE_LABEL_PL,
      NULL,
      BEETLE_OPT_ANALOG_TOGGLE_INFO_0_PL,
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { "enabled-analog", OPTION_VAL_ENABLED_ANALOG_PL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(analog_toggle_combo),
      BEETLE_OPT_ANALOG_TOGGLE_COMBO_LABEL_PL,
      NULL,
      BEETLE_OPT_ANALOG_TOGGLE_COMBO_INFO_0_PL,
      NULL,
      "input",
      {
         { "l1+l2+r1+r2+start+select", OPTION_VAL_L1_L2_R1_R2_START_SELECT_PL },
         { "l1+r1+select",             OPTION_VAL_L1_R1_SELECT_PL },
         { "l1+r1+start",              OPTION_VAL_L1_R1_START_PL },
         { "l1+r1+l3",                 OPTION_VAL_L1_R1_L3_PL },
         { "l1+r1+r3",                 OPTION_VAL_L1_R1_R3_PL },
         { "l2+r2+select",             OPTION_VAL_L2_R2_SELECT_PL },
         { "l2+r2+start",              OPTION_VAL_L2_R2_START_PL },
         { "l2+r2+l3",                 OPTION_VAL_L2_R2_L3_PL },
         { "l2+r2+r3",                 OPTION_VAL_L2_R2_R3_PL },
         { "l3+r3",                    OPTION_VAL_L3_R3_PL },
         { NULL, NULL },
      },
      "l1+r1+select"
   },
   {
      BEETLE_OPT(analog_toggle_hold),
      BEETLE_OPT_ANALOG_TOGGLE_HOLD_LABEL_PL,
      NULL,
      BEETLE_OPT_ANALOG_TOGGLE_HOLD_INFO_0_PL,
      NULL,
      "input",
      {
         { "0", OPTION_VAL_0_PL },
         { "1", OPTION_VAL_1_PL },
         { "2", OPTION_VAL_2_PL },
         { "3", OPTION_VAL_3_PL },
         { "4", OPTION_VAL_4_PL },
         { "5", OPTION_VAL_5_PL },
         { NULL, NULL },
      },
      "1"
   },
   {
      BEETLE_OPT(enable_multitap_port1),
      BEETLE_OPT_ENABLE_MULTITAP_PORT1_LABEL_PL,
      NULL,
      BEETLE_OPT_ENABLE_MULTITAP_PORT1_INFO_0_PL,
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(enable_multitap_port2),
      BEETLE_OPT_ENABLE_MULTITAP_PORT2_LABEL_PL,
      NULL,
      BEETLE_OPT_ENABLE_MULTITAP_PORT2_INFO_0_PL,
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(gun_input_mode),
      BEETLE_OPT_GUN_INPUT_MODE_LABEL_PL,
      NULL,
      BEETLE_OPT_GUN_INPUT_MODE_INFO_0_PL,
      NULL,
      "input",
      {
         { "lightgun",    OPTION_VAL_LIGHTGUN_PL },
         { "touchscreen", OPTION_VAL_TOUCHSCREEN_PL },
         { NULL, NULL },
      },
      "lightgun"
   },
   // Shouldn't the gun_input_mode just be Mouse vs. Touchscreen?
   {
      BEETLE_OPT(gun_cursor),
      BEETLE_OPT_GUN_CURSOR_LABEL_PL,
      NULL,
      BEETLE_OPT_GUN_CURSOR_INFO_0_PL,
      NULL,
      "input",
      {
         { "off",   "disabled" },
         { "cross", OPTION_VAL_CROSS_PL },
         { "dot",   OPTION_VAL_DOT_PL },
         { NULL, NULL },
      },
      "cross"
   },
   {
      BEETLE_OPT(crosshair_color_p1),
      BEETLE_OPT_CROSSHAIR_COLOR_P1_LABEL_PL,
      NULL,
      BEETLE_OPT_CROSSHAIR_COLOR_P1_INFO_0_PL,
      NULL,
      "input",
      {
         { "red",    OPTION_VAL_RED_PL },
         { "blue",   OPTION_VAL_BLUE_PL },
         { "green",  OPTION_VAL_GREEN_PL },
         { "orange", OPTION_VAL_ORANGE_PL },
         { "yellow", OPTION_VAL_YELLOW_PL },
         { "cyan",   OPTION_VAL_CYAN_PL },
         { "pink",   OPTION_VAL_PINK_PL },
         { "purple", OPTION_VAL_PURPLE_PL },
         { "black",  OPTION_VAL_BLACK_PL },
         { "white",  OPTION_VAL_WHITE_PL },
         { NULL, NULL },
      },
      "red"
   },
   {
      BEETLE_OPT(crosshair_color_p2),
      BEETLE_OPT_CROSSHAIR_COLOR_P2_LABEL_PL,
      NULL,
      BEETLE_OPT_CROSSHAIR_COLOR_P2_INFO_0_PL,
      NULL,
      "input",
      {
         { "blue",   OPTION_VAL_BLUE_PL },
         { "red",    OPTION_VAL_RED_PL },
         { "green",  OPTION_VAL_GREEN_PL },
         { "orange", OPTION_VAL_ORANGE_PL },
         { "yellow", OPTION_VAL_YELLOW_PL },
         { "cyan",   OPTION_VAL_CYAN_PL },
         { "pink",   OPTION_VAL_PINK_PL },
         { "purple", OPTION_VAL_PURPLE_PL },
         { "black",  OPTION_VAL_BLACK_PL },
         { "white",  OPTION_VAL_WHITE_PL },
         { NULL, NULL },
      },
      "blue"
   },
   {
      BEETLE_OPT(mouse_sensitivity),
      BEETLE_OPT_MOUSE_SENSITIVITY_LABEL_PL,
      NULL,
      BEETLE_OPT_MOUSE_SENSITIVITY_INFO_0_PL,
      NULL,
      "input",
      {
         { "5%",   OPTION_VAL_5_O42_PL },
         { "10%",  OPTION_VAL_10_PL },
         { "15%",  OPTION_VAL_15_PL },
         { "20%",  OPTION_VAL_20_PL },
         { "25%",  OPTION_VAL_25_PL },
         { "30%",  OPTION_VAL_30_PL },
         { "35%",  OPTION_VAL_35_PL },
         { "40%",  OPTION_VAL_40_PL },
         { "45%",  OPTION_VAL_45_PL },
         { "50%",  OPTION_VAL_50_PL },
         { "55%",  OPTION_VAL_55_PL },
         { "60%",  OPTION_VAL_60_PL },
         { "65%",  OPTION_VAL_65_PL },
         { "70%",  OPTION_VAL_70_PL },
         { "75%",  OPTION_VAL_75_PL },
         { "80%",  OPTION_VAL_80_PL },
         { "85%",  OPTION_VAL_85_PL },
         { "90%",  OPTION_VAL_90_PL },
         { "95%",  OPTION_VAL_95_PL },
         { "100%", OPTION_VAL_100_O42_PL },
         { "105%", OPTION_VAL_105_PL },
         { "110%", OPTION_VAL_110_PL },
         { "115%", OPTION_VAL_115_PL },
         { "120%", OPTION_VAL_120_PL },
         { "125%", OPTION_VAL_125_PL },
         { "130%", OPTION_VAL_130_PL },
         { "135%", OPTION_VAL_135_PL },
         { "140%", OPTION_VAL_140_PL },
         { "145%", OPTION_VAL_145_PL },
         { "150%", OPTION_VAL_150_PL },
         { "155%", OPTION_VAL_155_PL },
         { "160%", OPTION_VAL_160_PL },
         { "165%", OPTION_VAL_165_PL },
         { "170%", OPTION_VAL_170_PL },
         { "175%", OPTION_VAL_175_PL },
         { "180%", OPTION_VAL_180_PL },
         { "185%", OPTION_VAL_185_PL },
         { "190%", OPTION_VAL_190_PL },
         { "195%", OPTION_VAL_195_PL },
         { "200%", OPTION_VAL_200_PL },
         { NULL, NULL },
      },
      "100%"
   },
   {
      BEETLE_OPT(negcon_response),
      BEETLE_OPT_NEGCON_RESPONSE_LABEL_PL,
      NULL,
      BEETLE_OPT_NEGCON_RESPONSE_INFO_0_PL,
      NULL,
      "input",
      {
         { "linear",    OPTION_VAL_LINEAR_PL },
         { "quadratic", OPTION_VAL_QUADRATIC_PL },
         { "cubic",     OPTION_VAL_CUBIC_PL },
         { NULL, NULL },
      },
      "linear"
   },
   {
      BEETLE_OPT(negcon_deadzone),
      BEETLE_OPT_NEGCON_DEADZONE_LABEL_PL,
      NULL,
      BEETLE_OPT_NEGCON_DEADZONE_INFO_0_PL,
      NULL,
      "input",
      {
         { "0%",  OPTION_VAL_0_O44_PL },
         { "5%",  OPTION_VAL_5_O42_PL },
         { "10%", OPTION_VAL_10_PL },
         { "15%", OPTION_VAL_15_PL },
         { "20%", OPTION_VAL_20_PL },
         { "25%", OPTION_VAL_25_PL },
         { "30%", OPTION_VAL_30_PL },
         { NULL, NULL },
      },
      "0%"
   },
   {
      BEETLE_OPT(use_mednafen_memcard0_method),
      BEETLE_OPT_USE_MEDNAFEN_MEMCARD0_METHOD_LABEL_PL,
      NULL,
      BEETLE_OPT_USE_MEDNAFEN_MEMCARD0_METHOD_INFO_0_PL,
      NULL,
      "memcards",
      {
         { "libretro", OPTION_VAL_LIBRETRO_PL },
         { "mednafen", OPTION_VAL_MEDNAFEN_PL },
         { NULL, NULL },
      },
      "libretro"
   },
   {
      BEETLE_OPT(enable_memcard1),
      BEETLE_OPT_ENABLE_MEMCARD1_LABEL_PL,
      NULL,
      BEETLE_OPT_ENABLE_MEMCARD1_INFO_0_PL,
      NULL,
      "memcards",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(shared_memory_cards),
      BEETLE_OPT_SHARED_MEMORY_CARDS_LABEL_PL,
      NULL,
      BEETLE_OPT_SHARED_MEMORY_CARDS_INFO_0_PL,
      NULL,
      "memcards",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(memcard_left_index),
      BEETLE_OPT_MEMCARD_LEFT_INDEX_LABEL_PL,
      NULL,
      BEETLE_OPT_MEMCARD_LEFT_INDEX_INFO_0_PL,
      NULL,
      "memcards",
      {
         { "0",  OPTION_VAL_0_O48_PL },
         { "1",  NULL },
         { "2",  NULL },
         { "3",  NULL },
         { "4",  NULL },
         { "5",  NULL },
         { "6",  NULL },
         { "7",  NULL },
         { "8",  NULL },
         { "9",  NULL },
         { "10", NULL },
         { "11", NULL },
         { "12", NULL },
         { "13", NULL },
         { "14", NULL },
         { "15", NULL },
         { "16", NULL },
         { "17", NULL },
         { "18", NULL },
         { "19", NULL },
         { "20", NULL },
         { "21", NULL },
         { "22", NULL },
         { "23", NULL },
         { "24", NULL },
         { "25", NULL },
         { "26", NULL },
         { "27", NULL },
         { "28", NULL },
         { "29", NULL },
         { "30", NULL },
         { "31", NULL },
         { "32", NULL },
         { "33", NULL },
         { "34", NULL },
         { "35", NULL },
         { "36", NULL },
         { "37", NULL },
         { "38", NULL },
         { "39", NULL },
         { "40", NULL },
         { "41", NULL },
         { "42", NULL },
         { "43", NULL },
         { "44", NULL },
         { "45", NULL },
         { "46", NULL },
         { "47", NULL },
         { "48", NULL },
         { "49", NULL },
         { "50", NULL },
         { "51", NULL },
         { "52", NULL },
         { "53", NULL },
         { "54", NULL },
         { "55", NULL },
         { "56", NULL },
         { "57", NULL },
         { "58", NULL },
         { "59", NULL },
         { "60", NULL },
         { "61", NULL },
         { "62", NULL },
         { "63", NULL },
         { NULL, NULL },
      },
      "0"
   },
   {
      BEETLE_OPT(memcard_right_index),
      BEETLE_OPT_MEMCARD_RIGHT_INDEX_LABEL_PL,
      NULL,
      BEETLE_OPT_MEMCARD_RIGHT_INDEX_INFO_0_PL,
      NULL,
      "memcards",
      {
         { "0",  NULL },
         { "1",  OPTION_VAL_1_O49_PL },
         { "2",  NULL },
         { "3",  NULL },
         { "4",  NULL },
         { "5",  NULL },
         { "6",  NULL },
         { "7",  NULL },
         { "8",  NULL },
         { "9",  NULL },
         { "10", NULL },
         { "11", NULL },
         { "12", NULL },
         { "13", NULL },
         { "14", NULL },
         { "15", NULL },
         { "16", NULL },
         { "17", NULL },
         { "18", NULL },
         { "19", NULL },
         { "20", NULL },
         { "21", NULL },
         { "22", NULL },
         { "23", NULL },
         { "24", NULL },
         { "25", NULL },
         { "26", NULL },
         { "27", NULL },
         { "28", NULL },
         { "29", NULL },
         { "30", NULL },
         { "31", NULL },
         { "32", NULL },
         { "33", NULL },
         { "34", NULL },
         { "35", NULL },
         { "36", NULL },
         { "37", NULL },
         { "38", NULL },
         { "39", NULL },
         { "40", NULL },
         { "41", NULL },
         { "42", NULL },
         { "43", NULL },
         { "44", NULL },
         { "45", NULL },
         { "46", NULL },
         { "47", NULL },
         { "48", NULL },
         { "49", NULL },
         { "50", NULL },
         { "51", NULL },
         { "52", NULL },
         { "53", NULL },
         { "54", NULL },
         { "55", NULL },
         { "56", NULL },
         { "57", NULL },
         { "58", NULL },
         { "59", NULL },
         { "60", NULL },
         { "61", NULL },
         { "62", NULL },
         { "63", NULL },
         { NULL, NULL },
      },
      "1"
   },
   {
      BEETLE_OPT(line_render),
      BEETLE_OPT_LINE_RENDER_LABEL_PL,
      NULL,
      BEETLE_OPT_LINE_RENDER_INFO_0_PL,
      NULL,
      "hacks",
      {
         { "default",    OPTION_VAL_DEFAULT_PL },
         { "aggressive", OPTION_VAL_AGGRESSIVE_PL },
         { "disabled",   NULL },
         { NULL, NULL },
      },
      "default"
   },
   {
      BEETLE_OPT(widescreen_hack),
      BEETLE_OPT_WIDESCREEN_HACK_LABEL_PL,
      NULL,
      BEETLE_OPT_WIDESCREEN_HACK_INFO_0_PL,
      NULL,
      "hacks",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(widescreen_hack_aspect_ratio),
      BEETLE_OPT_WIDESCREEN_HACK_ASPECT_RATIO_LABEL_PL,
      NULL,
      BEETLE_OPT_WIDESCREEN_HACK_ASPECT_RATIO_INFO_0_PL,
      NULL,
      "hacks",
      {
         { "16:9",  OPTION_VAL_16_9_PL },
         { "16:10", OPTION_VAL_16_10_PL },
         { "18:9",  OPTION_VAL_18_9_PL },
         { "19:9",  OPTION_VAL_19_9_PL },
         { "20:9",  OPTION_VAL_20_9_PL },
         { "21:9",  OPTION_VAL_21_9_PL }, // 64:27
         { "32:9",  OPTION_VAL_32_9_PL },
         { NULL,    NULL },
      },
      "16:9"
   },
#if defined(HAVE_LIGHTREC)
   {
      BEETLE_OPT(cpu_dynarec),
      BEETLE_OPT_CPU_DYNAREC_LABEL_PL,
      NULL,
      BEETLE_OPT_CPU_DYNAREC_INFO_0_PL,
      NULL,
      "hacks",
      {
         { "disabled",        OPTION_VAL_DISABLED_PL },
         { "execute",         OPTION_VAL_EXECUTE_PL },
         { "execute_one",     OPTION_VAL_EXECUTE_ONE_PL },
         { "run_interpreter", OPTION_VAL_RUN_INTERPRETER_PL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(dynarec_invalidate),
      BEETLE_OPT_DYNAREC_INVALIDATE_LABEL_PL,
      NULL,
      BEETLE_OPT_DYNAREC_INVALIDATE_INFO_0_PL,
      NULL,
      "hacks",
      {
         { "full", OPTION_VAL_FULL_PL },
         { "dma",  OPTION_VAL_DMA_PL },
         { NULL, NULL },
      },
      "full"
   },
   {
      BEETLE_OPT(dynarec_eventcycles),
      BEETLE_OPT_DYNAREC_EVENTCYCLES_LABEL_PL,
      NULL,
      BEETLE_OPT_DYNAREC_EVENTCYCLES_INFO_0_PL,
      NULL,
      "hacks",
      {
         { "128",  OPTION_VAL_128_PL },
         { "256",  NULL },
         { "384",  NULL },
         { "512",  NULL },
         { "640",  NULL },
         { "768",  NULL },
         { "896",  NULL },
         { "1024", NULL },
         { "1152", NULL },
         { "1280", NULL },
         { "1408", NULL },
         { "1536", NULL },
         { "1664", NULL },
         { "1792", NULL },
         { "1920", NULL },
         { "2048", NULL },
         { NULL, NULL },
      },
      "128"
   },
   {
      BEETLE_OPT(dynarec_spu_samples),
      BEETLE_OPT_DYNAREC_SPU_SAMPLES_LABEL_PL,
      NULL,
      BEETLE_OPT_DYNAREC_SPU_SAMPLES_INFO_0_PL,
      NULL,
      "hacks",
      {
         { "1",  OPTION_VAL_1_O49_PL },
         { "4",  NULL },
         { "16", NULL },
         { NULL, NULL },
      },
      "1"
   },
#endif
   {
      BEETLE_OPT(core_timing_fps),
      BEETLE_OPT_CORE_TIMING_FPS_LABEL_PL,
      NULL,
      BEETLE_OPT_CORE_TIMING_FPS_INFO_0_PL,
      NULL,
      "video",
      {
         { "auto_toggle",       OPTION_VAL_AUTO_TOGGLE_PL },
         { "force_progressive", OPTION_VAL_FORCE_PROGRESSIVE_PL },
         { "force_interlaced",  OPTION_VAL_FORCE_INTERLACED_PL },
      },
      "auto_toggle"
   },
   {
      BEETLE_OPT(deinterlacer),
      BEETLE_OPT_DEINTERLACER_LABEL_PL,
      NULL,
      BEETLE_OPT_DEINTERLACER_INFO_0_PL,
      NULL,
      "video",
      {
         { "weave",      OPTION_VAL_WEAVE_PL },
         { "bob",        OPTION_VAL_BOB_PL },
         { "bob_offset", OPTION_VAL_BOB_OFFSET_PL },
         { "fastmad",    OPTION_VAL_FASTMAD_PL },
         { "off",        "Off" },
         { NULL, NULL },
      },
      "weave"
   },
   {
      BEETLE_OPT(frame_duping),
      BEETLE_OPT_FRAME_DUPING_LABEL_PL,
      NULL,
      BEETLE_OPT_FRAME_DUPING_INFO_0_PL,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(pal_video_timing_override),
      BEETLE_OPT_PAL_VIDEO_TIMING_OVERRIDE_LABEL_PL,
      NULL,
      BEETLE_OPT_PAL_VIDEO_TIMING_OVERRIDE_INFO_0_PL,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(aspect_ratio),
      BEETLE_OPT_ASPECT_RATIO_LABEL_PL,
      NULL,
      BEETLE_OPT_ASPECT_RATIO_INFO_0_PL,
      NULL,
      "video",
      {
         { "corrected", OPTION_VAL_CORRECTED_PL },
         { "uncorrected", OPTION_VAL_UNCORRECTED_PL },
         { "4:3",  OPTION_VAL_4_3_PL },
         { "ntsc", OPTION_VAL_NTSC_PL },
      },
      "corrected"
   },
   {
      BEETLE_OPT(crop_overscan),
      BEETLE_OPT_CROP_OVERSCAN_LABEL_PL,
      NULL,
      BEETLE_OPT_CROP_OVERSCAN_INFO_0_PL,
      NULL,
      "video",
      {
         { "disabled", "Disabled" },
         { "static",   OPTION_VAL_STATIC_PL },
         { "smart",    OPTION_VAL_SMART_PL },
         { NULL, NULL },
      },
      "smart"
   },
   {
      BEETLE_OPT(image_crop),
      BEETLE_OPT_IMAGE_CROP_LABEL_PL,
      NULL,
      BEETLE_OPT_IMAGE_CROP_INFO_0_PL,
      NULL,
      "video",
      {
         { "disabled", "0" },
         { "1px",      OPTION_VAL_1PX_PL },
         { "2px",      OPTION_VAL_2PX_PL },
         { "3px",      OPTION_VAL_3PX_PL },
         { "4px",      OPTION_VAL_4PX_PL },
         { "5px",      OPTION_VAL_5PX_PL },
         { "6px",      OPTION_VAL_6PX_PL },
         { "7px",      OPTION_VAL_7PX_PL },
         { "8px",      OPTION_VAL_8PX_PL },
         { "9px",      OPTION_VAL_9PX_PL },
         { "10px",     OPTION_VAL_10PX_PL },
         { "11px",     OPTION_VAL_11PX_PL },
         { "12px",     OPTION_VAL_12PX_PL },
         { "13px",     OPTION_VAL_13PX_PL },
         { "14px",     OPTION_VAL_14PX_PL },
         { "15px",     OPTION_VAL_15PX_PL },
         { "16px",     OPTION_VAL_16PX_PL },
         { "17px",     OPTION_VAL_17PX_PL },
         { "18px",     OPTION_VAL_18PX_PL },
         { "19px",     OPTION_VAL_19PX_PL },
         { "20px",     OPTION_VAL_20PX_PL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(image_offset),
      BEETLE_OPT_IMAGE_OFFSET_LABEL_PL,
      NULL,
      BEETLE_OPT_IMAGE_OFFSET_INFO_0_PL,
      NULL,
      "video",
      {
         { "-12px",    OPTION_VAL_12PX_O64_PL },
         { "-11px",    OPTION_VAL_11PX_O64_PL },
         { "-10px",    OPTION_VAL_10PX_O64_PL },
         { "-9px",     OPTION_VAL_9PX_O64_PL },
         { "-8px",     OPTION_VAL_8PX_O64_PL },
         { "-7px",     OPTION_VAL_7PX_O64_PL },
         { "-6px",     OPTION_VAL_6PX_O64_PL },
         { "-5px",     OPTION_VAL_5PX_O64_PL },
         { "-4px",     OPTION_VAL_4PX_O64_PL },
         { "-3px",     OPTION_VAL_3PX_O64_PL },
         { "-2px",     OPTION_VAL_2PX_O64_PL },
         { "-1px",     OPTION_VAL_1PX_O64_PL },
         { "disabled", "0" },
         { "+1px",     OPTION_VAL_1PX_O64_0_PL },
         { "+2px",     OPTION_VAL_2PX_O64_0_PL },
         { "+3px",     OPTION_VAL_3PX_O64_0_PL },
         { "+4px",     OPTION_VAL_4PX_O64_0_PL },
         { "+5px",     OPTION_VAL_5PX_O64_0_PL },
         { "+6px",     OPTION_VAL_6PX_O64_0_PL },
         { "+7px",     OPTION_VAL_7PX_O64_0_PL },
         { "+8px",     OPTION_VAL_8PX_O64_0_PL },
         { "+9px",     OPTION_VAL_9PX_O64_0_PL },
         { "+10px",    OPTION_VAL_10PX_O64_0_PL },
         { "+11px",    OPTION_VAL_11PX_O64_0_PL },
         { "+12px",    OPTION_VAL_12PX_O64_0_PL },
         { NULL, NULL },
      },
      "disabled"
   },
#if defined(HAVE_OPENGL) || defined(HAVE_OPENGLES) || defined(HAVE_VULKAN)
   {
      BEETLE_OPT(image_offset_cycles),
      BEETLE_OPT_IMAGE_OFFSET_CYCLES_LABEL_PL,
      NULL,
      BEETLE_OPT_IMAGE_OFFSET_CYCLES_INFO_0_PL,
      NULL,
      "video",
      {
         { "-40",      NULL },
         { "-39",      NULL },
         { "-38",      NULL },
         { "-37",      NULL },
         { "-36",      NULL },
         { "-35",      NULL },
         { "-34",      NULL },
         { "-33",      NULL },
         { "-32",      NULL },
         { "-31",      NULL },
         { "-30",      NULL },
         { "-29",      NULL },
         { "-28",      NULL },
         { "-27",      NULL },
         { "-26",      NULL },
         { "-25",      NULL },
         { "-24",      NULL },
         { "-23",      NULL },
         { "-22",      NULL },
         { "-21",      NULL },
         { "-20",      NULL },
         { "-19",      NULL },
         { "-18",      NULL },
         { "-17",      NULL },
         { "-16",      NULL },
         { "-15",      NULL },
         { "-14",      NULL },
         { "-13",      NULL },
         { "-12",      NULL },
         { "-11",      NULL },
         { "-10",      NULL },
         { "-9",       NULL },
         { "-8",       NULL },
         { "-7",       NULL },
         { "-6",       NULL },
         { "-5",       NULL },
         { "-4",       NULL },
         { "-3",       NULL },
         { "-2",       NULL },
         { "-1",       NULL },
         { "0",        "0" },
         { "+1",       NULL },
         { "+2",       NULL },
         { "+3",       NULL },
         { "+4",       NULL },
         { "+5",       NULL },
         { "+6",       NULL },
         { "+7",       NULL },
         { "+8",       NULL },
         { "+9",       NULL },
         { "+10",      NULL },
         { "+11",      NULL },
         { "+12",      NULL },
         { "+13",      NULL },
         { "+14",      NULL },
         { "+15",      NULL },
         { "+16",      NULL },
         { "+17",      NULL },
         { "+18",      NULL },
         { "+19",      NULL },
         { "+20",      NULL },
         { "+21",      NULL },
         { "+22",      NULL },
         { "+23",      NULL },
         { "+24",      NULL },
         { "+25",      NULL },
         { "+26",      NULL },
         { "+27",      NULL },
         { "+28",      NULL },
         { "+29",      NULL },
         { "+30",      NULL },
         { "+31",      NULL },
         { "+32",      NULL },
         { "+33",      NULL },
         { "+34",      NULL },
         { "+35",      NULL },
         { "+36",      NULL },
         { "+37",      NULL },
         { "+38",      NULL },
         { "+39",      NULL },
         { "+40",      NULL },
         { NULL, NULL},
      },
      "0"
   },
#endif
   {
      BEETLE_OPT(initial_scanline),
      BEETLE_OPT_INITIAL_SCANLINE_LABEL_PL,
      NULL,
      BEETLE_OPT_INITIAL_SCANLINE_INFO_0_PL,
      NULL,
      "video",
      {
         { "0",  OPTION_VAL_0_O48_PL },
         { "1",  NULL },
         { "2",  NULL },
         { "3",  NULL },
         { "4",  NULL },
         { "5",  NULL },
         { "6",  NULL },
         { "7",  NULL },
         { "8",  NULL },
         { "9",  NULL },
         { "10", NULL },
         { "11", NULL },
         { "12", NULL },
         { "13", NULL },
         { "14", NULL },
         { "15", NULL },
         { "16", NULL },
         { "17", NULL },
         { "18", NULL },
         { "19", NULL },
         { "20", NULL },
         { "21", NULL },
         { "22", NULL },
         { "23", NULL },
         { "24", NULL },
         { "25", NULL },
         { "26", NULL },
         { "27", NULL },
         { "28", NULL },
         { "29", NULL },
         { "30", NULL },
         { "31", NULL },
         { "32", NULL },
         { "33", NULL },
         { "34", NULL },
         { "35", NULL },
         { "36", NULL },
         { "37", NULL },
         { "38", NULL },
         { "39", NULL },
         { "40", NULL },
         { NULL, NULL },
      },
      "0"
   },
   {
      BEETLE_OPT(last_scanline),
      BEETLE_OPT_LAST_SCANLINE_LABEL_PL,
      NULL,
      BEETLE_OPT_LAST_SCANLINE_INFO_0_PL,
      NULL,
      "video",
      {
         { "210", NULL },
         { "211", NULL },
         { "212", NULL },
         { "213", NULL },
         { "214", NULL },
         { "215", NULL },
         { "216", NULL },
         { "217", NULL },
         { "218", NULL },
         { "219", NULL },
         { "220", NULL },
         { "221", NULL },
         { "222", NULL },
         { "223", NULL },
         { "224", NULL },
         { "225", NULL },
         { "226", NULL },
         { "227", NULL },
         { "228", NULL },
         { "229", NULL },
         { "230", NULL },
         { "231", NULL },
         { "232", NULL },
         { "233", NULL },
         { "234", NULL },
         { "235", NULL },
         { "236", NULL },
         { "237", NULL },
         { "238", NULL },
         { "239", OPTION_VAL_239_PL },
         { NULL, NULL },
      },
      "239"
   },
   {
      BEETLE_OPT(initial_scanline_pal),
      BEETLE_OPT_INITIAL_SCANLINE_PAL_LABEL_PL,
      NULL,
      BEETLE_OPT_INITIAL_SCANLINE_PAL_INFO_0_PL,
      NULL,
      "video",
      {
         { "0",  OPTION_VAL_0_O48_PL },
         { "1",  NULL },
         { "2",  NULL },
         { "3",  NULL },
         { "4",  NULL },
         { "5",  NULL },
         { "6",  NULL },
         { "7",  NULL },
         { "8",  NULL },
         { "9",  NULL },
         { "10", NULL },
         { "11", NULL },
         { "12", NULL },
         { "13", NULL },
         { "14", NULL },
         { "15", NULL },
         { "16", NULL },
         { "17", NULL },
         { "18", NULL },
         { "19", NULL },
         { "20", NULL },
         { "21", NULL },
         { "22", NULL },
         { "23", NULL },
         { "24", NULL },
         { "25", NULL },
         { "26", NULL },
         { "27", NULL },
         { "28", NULL },
         { "29", NULL },
         { "30", NULL },
         { "31", NULL },
         { "32", NULL },
         { "33", NULL },
         { "34", NULL },
         { "35", NULL },
         { "36", NULL },
         { "37", NULL },
         { "38", NULL },
         { "39", NULL },
         { "40", NULL },
         { NULL, NULL },
      },
      "0"
   },
   {
      BEETLE_OPT(last_scanline_pal),
      BEETLE_OPT_LAST_SCANLINE_PAL_LABEL_PL,
      NULL,
      BEETLE_OPT_LAST_SCANLINE_PAL_INFO_0_PL,
      NULL,
      "video",
      {
         { "230", NULL },
         { "231", NULL },
         { "232", NULL },
         { "233", NULL },
         { "234", NULL },
         { "235", NULL },
         { "236", NULL },
         { "237", NULL },
         { "238", NULL },
         { "239", NULL },
         { "240", NULL },
         { "241", NULL },
         { "242", NULL },
         { "243", NULL },
         { "244", NULL },
         { "245", NULL },
         { "246", NULL },
         { "247", NULL },
         { "248", NULL },
         { "249", NULL },
         { "250", NULL },
         { "251", NULL },
         { "252", NULL },
         { "253", NULL },
         { "254", NULL },
         { "255", NULL },
         { "256", NULL },
         { "257", NULL },
         { "258", NULL },
         { "259", NULL },
         { "260", NULL },
         { "261", NULL },
         { "262", NULL },
         { "263", NULL },
         { "264", NULL },
         { "265", NULL },
         { "266", NULL },
         { "267", NULL },
         { "268", NULL },
         { "269", NULL },
         { "270", NULL },
         { "271", NULL },
         { "272", NULL },
         { "273", NULL },
         { "274", NULL },
         { "275", NULL },
         { "276", NULL },
         { "277", NULL },
         { "278", NULL },
         { "279", NULL },
         { "280", NULL },
         { "281", NULL },
         { "282", NULL },
         { "283", NULL },
         { "284", NULL },
         { "285", NULL },
         { "286", NULL },
         { "287", OPTION_VAL_287_PL },
         { NULL, NULL },
      },
      "287"
   },
   { NULL, NULL, NULL, NULL, NULL, NULL, {{0}}, NULL },
};
struct retro_core_options_v2 options_pl = {
   option_cats_pl,
   option_defs_pl
};

/* RETRO_LANGUAGE_PT_BR */

#define CATEGORY_SYSTEM_LABEL_PT_BR "Sistema"
#define CATEGORY_SYSTEM_INFO_0_PT_BR "Altera as opções de sistema e overclock."
#define CATEGORY_VIDEO_LABEL_PT_BR "Vídeo"
#define CATEGORY_VIDEO_INFO_0_PT_BR "Altera as configurações da proporção de tela, corte de tela, filtro de vídeo e pulo de quadros."
#define CATEGORY_PGXP_LABEL_PT_BR "PGXP (Canal de Transformação de Geometria de Precisão)"
#define CATEGORY_PGXP_INFO_0_PT_BR "Estas opções controlam aprimoramentos que podem melhorar os gráficos em comparação com o console original. O PGXP pode se livrar de texturas deformadas e problemas de polígonos no eixo Z."
#define CATEGORY_OSD_LABEL_PT_BR "Informações na tela (OSD)"
#define CATEGORY_OSD_INFO_0_PT_BR "Altera a visibilidade de notificações e informações na tela."
#define CATEGORY_INPUT_LABEL_PT_BR "Entrada"
#define CATEGORY_INPUT_INFO_0_PT_BR "Altera as configurações das pistolas de luz, mouse e neGcon."
#define CATEGORY_MEMCARDS_LABEL_PT_BR NULL
#define CATEGORY_MEMCARDS_INFO_0_PT_BR "Altera as configurações relacionadas ao cartão de memória virtual."
#define CATEGORY_HACKS_LABEL_PT_BR "Hacks de emulação"
#define CATEGORY_HACKS_INFO_0_PT_BR "Altera as configurações de precisão de emulação que afetam o desempenho e a compatibilidade de baixo nível."
#define BEETLE_OPT_CPU_FREQ_SCALE_LABEL_PT_BR "Escala de frequência da CPU (overclock)"
#define BEETLE_OPT_CPU_FREQ_SCALE_INFO_0_PT_BR "Aumenta (ou diminui) a velocidade da CPU emulada do PSX. Uma maior velocidade pode eliminar lentidão e melhorar as taxas de quadros em certos jogos, ao custo de maiores exigências de desempenho. Note que certos jogos incluem um limitador de quadro interno e não obterão nenhum benefício com o aumento da velocidade. Pode fazer certos efeitos ficarem mais rápidos do que o normal em certos jogos."
#define OPTION_VAL_50_PT_BR NULL
#define OPTION_VAL_60_PT_BR NULL
#define OPTION_VAL_70_PT_BR NULL
#define OPTION_VAL_80_PT_BR NULL
#define OPTION_VAL_90_PT_BR NULL
#define OPTION_VAL_100_PT_BR "100% (Nativo)"
#define OPTION_VAL_110_PT_BR NULL
#define OPTION_VAL_120_PT_BR NULL
#define OPTION_VAL_130_PT_BR NULL
#define OPTION_VAL_140_PT_BR NULL
#define OPTION_VAL_150_PT_BR NULL
#define OPTION_VAL_160_PT_BR NULL
#define OPTION_VAL_170_PT_BR NULL
#define OPTION_VAL_180_PT_BR NULL
#define OPTION_VAL_190_PT_BR NULL
#define OPTION_VAL_200_PT_BR NULL
#define OPTION_VAL_210_PT_BR NULL
#define OPTION_VAL_220_PT_BR NULL
#define OPTION_VAL_230_PT_BR NULL
#define OPTION_VAL_240_PT_BR NULL
#define OPTION_VAL_250_PT_BR NULL
#define OPTION_VAL_260_PT_BR NULL
#define OPTION_VAL_270_PT_BR NULL
#define OPTION_VAL_280_PT_BR NULL
#define OPTION_VAL_290_PT_BR NULL
#define OPTION_VAL_300_PT_BR NULL
#define OPTION_VAL_310_PT_BR NULL
#define OPTION_VAL_320_PT_BR NULL
#define OPTION_VAL_330_PT_BR NULL
#define OPTION_VAL_340_PT_BR NULL
#define OPTION_VAL_350_PT_BR NULL
#define OPTION_VAL_360_PT_BR NULL
#define OPTION_VAL_370_PT_BR NULL
#define OPTION_VAL_380_PT_BR NULL
#define OPTION_VAL_390_PT_BR NULL
#define OPTION_VAL_400_PT_BR NULL
#define OPTION_VAL_410_PT_BR NULL
#define OPTION_VAL_420_PT_BR NULL
#define OPTION_VAL_430_PT_BR NULL
#define OPTION_VAL_440_PT_BR NULL
#define OPTION_VAL_450_PT_BR NULL
#define OPTION_VAL_460_PT_BR NULL
#define OPTION_VAL_470_PT_BR NULL
#define OPTION_VAL_480_PT_BR NULL
#define OPTION_VAL_490_PT_BR NULL
#define OPTION_VAL_500_PT_BR NULL
#define OPTION_VAL_510_PT_BR NULL
#define OPTION_VAL_520_PT_BR NULL
#define OPTION_VAL_530_PT_BR NULL
#define OPTION_VAL_540_PT_BR NULL
#define OPTION_VAL_550_PT_BR NULL
#define OPTION_VAL_560_PT_BR NULL
#define OPTION_VAL_570_PT_BR NULL
#define OPTION_VAL_580_PT_BR NULL
#define OPTION_VAL_590_PT_BR NULL
#define OPTION_VAL_600_PT_BR NULL
#define OPTION_VAL_610_PT_BR NULL
#define OPTION_VAL_620_PT_BR NULL
#define OPTION_VAL_630_PT_BR NULL
#define OPTION_VAL_640_PT_BR NULL
#define OPTION_VAL_650_PT_BR NULL
#define OPTION_VAL_660_PT_BR NULL
#define OPTION_VAL_670_PT_BR NULL
#define OPTION_VAL_680_PT_BR NULL
#define OPTION_VAL_690_PT_BR NULL
#define OPTION_VAL_700_PT_BR NULL
#define OPTION_VAL_710_PT_BR NULL
#define OPTION_VAL_720_PT_BR NULL
#define OPTION_VAL_730_PT_BR NULL
#define OPTION_VAL_740_PT_BR NULL
#define OPTION_VAL_750_PT_BR NULL
#define BEETLE_OPT_GPU_OVERCLOCK_LABEL_PT_BR "Overclock do rasterizador da GPU"
#define BEETLE_OPT_GPU_OVERCLOCK_INFO_0_PT_BR "Ativa o overclocking do rasterizador 2D contido dentro da GPU emulada do PSX. Não melhora a renderização 3D, e em geral tem pouco efeito."
#define OPTION_VAL_1X_NATIVE_PT_BR "1x (nativa)"
#define OPTION_VAL_2X_PT_BR NULL
#define OPTION_VAL_4X_PT_BR NULL
#define OPTION_VAL_8X_PT_BR NULL
#define OPTION_VAL_16X_PT_BR NULL
#define OPTION_VAL_32X_PT_BR NULL
#define BEETLE_OPT_GTE_OVERCLOCK_LABEL_PT_BR "Overclock do GTE"
#define BEETLE_OPT_GTE_OVERCLOCK_INFO_0_PT_BR "Reduz todas as operações do GTE (coprocessador da CPU para gráficos 3D) emuladas a uma latência constante de um ciclo. Para jogos que fazem uso pesado do GTE, isto pode melhorar muito a taxa de quadros e a estabilidade do tempo de quadros."
#define BEETLE_OPT_CD_ACCESS_METHOD_LABEL_PT_BR "Método de Acesso ao CD"
#define BEETLE_OPT_CD_ACCESS_METHOD_INFO_0_PT_BR "'Síncrono' imita o hardware original. 'Assíncrono' pode reduzir engasgos em dispositivos com armazenamento lento. 'Pré-cache' carrega toda a imagem do disco na memória ao iniciar o conteúdo, o que pode melhorar os tempos de carregamento dentro do jogo, ao custo de um atraso inicial na inicialização. 'Pré-cache' pode causar problemas em sistemas com pouca RAM e voltará automaticamente para 'Síncrono' ao usar mídia física.Reinício necessário."
#define OPTION_VAL_SYNC_PT_BR "Síncrono"
#define OPTION_VAL_ASYNC_PT_BR "Assíncrono"
#define OPTION_VAL_PRECACHE_PT_BR "Pré-cache"
#define BEETLE_OPT_CD_FASTLOAD_LABEL_PT_BR "Velocidade de carregamento do CD"
#define BEETLE_OPT_CD_FASTLOAD_INFO_0_PT_BR "Valores acima de '2x (Nativo)' podem reduzir bastante os tempos de carregamento dentro do jogo, mas podem introduzir erros de temporização. Alguns jogos podem não funcionar corretamente acima de determinados valores."
#define OPTION_VAL_2X_NATIVE_PT_BR "2x (nativa)"
#define OPTION_VAL_6X_PT_BR NULL
#define OPTION_VAL_10X_PT_BR NULL
#define OPTION_VAL_12X_PT_BR NULL
#define OPTION_VAL_14X_PT_BR NULL
#define BEETLE_OPT_SKIP_BIOS_LABEL_PT_BR "Pular BIOS"
#define BEETLE_OPT_SKIP_BIOS_INFO_0_PT_BR "Omite a animação de inicialização da BIOS do PlayStation normalmente exibida ao carregar um conteúdo. Nota: Ativar isso causa problemas de compatibilidade com vários jogos (jogos PAL com proteção anticópia, Saga Frontier, etc.)."
#define BEETLE_OPT_OVERRIDE_BIOS_LABEL_PT_BR "Substituir BIOS"
#define BEETLE_OPT_OVERRIDE_BIOS_INFO_0_PT_BR "Substitui a BIOS padrão específica da região por uma BIOS livre de região, se encontrada. Reinicialização necessária."
#define OPTION_VAL_PSXONPSP_PT_BR "BIOS de PS1 de PSP"
#define OPTION_VAL_PS1_ROM_PT_BR "BIOS de PS1 de PS3"
#define OPTION_VAL_OPENBIOS_PT_BR NULL
#define BEETLE_OPT_REGION_LABEL_PT_BR "Região do sistema"
#define BEETLE_OPT_REGION_INFO_0_PT_BR "Defina a região de fallback usada quando o conteúdo não puder ser detectado automaticamente (ex.: arquivos PS-X EXE brutos). Jogos em disco continuam usando a região gravada no disco. Requer reinicialização."
#define OPTION_VAL_AUTO_PT_BR "Automático"
#define OPTION_VAL_NTSC_J_PT_BR "NTSC-J (Japão)"
#define OPTION_VAL_NTSC_U_PT_BR "NTSC-U (América do Norte)"
#define OPTION_VAL_PAL_PT_BR "PAL (Europa)"
#define BEETLE_OPT_INTERNAL_RESOLUTION_LABEL_PT_BR "Resolução interna da GPU"
#define BEETLE_OPT_INTERNAL_RESOLUTION_INFO_0_PT_BR "Resoluções superiores a '1x (Nativa)' melhoram a fidelidade dos modelos 3D, ao custo de maiores requisitos de desempenho. Elementos 2D geralmente não são afetados por esta configuração."
#define BEETLE_OPT_RENDERER_LABEL_PT_BR "Renderizador"
#define BEETLE_OPT_RENDERER_INFO_0_PT_BR "O renderizador de software é o mais preciso, mas possui requisitos de desempenho elevados ao rodar em resoluções internas de GPU aumentadas. Os renderizadores de hardware, embora menos precisos, melhoram o desempenho em relação ao renderizador de software em resoluções internas maiores e permitem várias melhorias gráficas. 'Hardware (Auto)' seleciona automaticamente o renderizador Vulkan ou OpenGL, dependendo do driver de vídeo do frontend libretro atual. Se o driver de vídeo fornecido não for compatível com Vulkan ou OpenGL 3.3, o core usará o renderizador de software. Reinicialização necessária."
#define OPTION_VAL_HARDWARE_PT_BR "Hardware (Automático)"
#define OPTION_VAL_HARDWARE_GL_PT_BR NULL
#define OPTION_VAL_HARDWARE_VK_PT_BR NULL
#define OPTION_VAL_SOFTWARE_PT_BR NULL
#define BEETLE_OPT_RENDERER_SOFTWARE_FB_LABEL_PT_BR "Quadro de buffer por software"
#define BEETLE_OPT_RENDERER_SOFTWARE_FB_INFO_0_PT_BR "Permite emular com precisão os efeitos de quadro de buffer (por exemplo, desfoque de movimento, espiral de batalha FF7) ao usar renderizadores de hardware, executando uma cópia do renderizador de software com resolução nativa em segundo plano. A desativação desta opção irá omitir estas operações (OpenGL) ou serão renderizadas na GPU (Vulkan). A desativação desta opção pode melhorar o desempenho ao custo de erros gráficos graves. Em caso de dúvida, deixe esta opção ativada."
#define BEETLE_OPT_DEPTH_LABEL_PT_BR "Profundidade de cor interna"
#define BEETLE_OPT_DEPTH_INFO_0_PT_BR "Maior profundidade de cor pode reduzir efeitos de banding de cores sem o uso de dithering. 16 bpp emula o hardware original, mas pode apresentar banding visível se o dithering não estiver ativado. 'Padrão de Dithering' é recomendado ficar desativado quando esta opção estiver definida para 32 bpp."
#define OPTION_VAL_16BPP_NATIVE_PT_BR "16 bpp (nativo)"
#define OPTION_VAL_32BPP_PT_BR NULL
#define BEETLE_OPT_DITHER_MODE_LABEL_PT_BR "Pontilhamento"
#define BEETLE_OPT_DITHER_MODE_INFO_0_PT_BR "'1x (Nativa)' emula o dithering de baixa resolução nativa usado pelo hardware original para suavizar artefatos de banding de cores visíveis na profundidade de cor nativa. 'Resolução Interna' escala a granularidade do dithering conforme a resolução interna configurada, resultando em resultados mais limpos. É recomendado desativar quando estiver usando profundidade de cor de 32 bpp. Nota: No Vulkan, ativar esta opção forçará o downsampling para a profundidade de cor nativa, enquanto desativá-la ativará automaticamente a saída em profundidade de cor mais alta."
#define OPTION_VAL_INTERNAL_RESOLUTION_PT_BR "Resolução interna"
#define BEETLE_OPT_SCALED_UV_OFFSET_LABEL_PT_BR "Deslocamento UV da textura"
#define BEETLE_OPT_SCALED_UV_OFFSET_INFO_0_PT_BR "Amostra de texturas para polígonos 3D em um deslocamento para uma resolução interna superior a 1x. Reduz as emendas de textura, mas pode causar falhas gráficas não intencionais."
#define BEETLE_OPT_FILTER_LABEL_PT_BR "Filtragem de textura"
#define BEETLE_OPT_FILTER_INFO_0_PT_BR "'Nearest' emula o hardware original. 'Bilinear' e '3-Point' são filtros de suavização que reduzem a pixelização por meio de desfoque. 'SABR', 'xBR' e 'JINC2' são filtros de upscale que podem melhorar a fidelidade/nitidez das texturas, ao custo de maiores requisitos de desempenho. Compatíveis apenas com renderizadores por hardware."
#define OPTION_VAL_NEAREST_PT_BR NULL
#define OPTION_VAL_SABR_PT_BR NULL
#define OPTION_VAL_XBR_PT_BR NULL
#define OPTION_VAL_BILINEAR_PT_BR NULL
#define OPTION_VAL_3_POINT_PT_BR "3 pontos"
#define OPTION_VAL_JINC2_PT_BR NULL
#define BEETLE_OPT_FILTER_EXCLUDE_SPRITE_LABEL_PT_BR "Excluir sprites da filtragem"
#define BEETLE_OPT_FILTER_EXCLUDE_SPRITE_INFO_0_PT_BR "Não aplicará filtragem de textura aos sprites. Previne costuras em vários jogos com fundo 2D renderizado com sprite. Use junto com a Suavização adaptativa ou outro filtro de pós-processamento para obter o melhor efeito."
#define OPTION_VAL_OPAQUE_PT_BR "Apenas sprites opacos"
#define OPTION_VAL_ALL_PT_BR "Sprites opacos e semitransparente"
#define BEETLE_OPT_FILTER_EXCLUDE_2D_POLYGON_LABEL_PT_BR "Excluir polígonos 2D da filtragem"
#define BEETLE_OPT_FILTER_EXCLUDE_2D_POLYGON_INFO_0_PT_BR "Não aplicará filtragem de textura a polígonos 2D. Os polígonos 2D são detectados com uma heurística e pode haver falhas. Use junto com a Suavização adaptativa ou outro filtro de pós-processamento para obter o melhor efeito."
#define BEETLE_OPT_ADAPTIVE_SMOOTHING_LABEL_PT_BR "Suavização adaptativa"
#define BEETLE_OPT_ADAPTIVE_SMOOTHING_INFO_0_PT_BR "Suaviza os gráficos e elementos 2D da interface sem embaçar os objetos renderizados em 3D. Somente suportado pelo renderizador Vulkan."
#define BEETLE_OPT_SUPER_SAMPLING_LABEL_PT_BR "Superamostragem (redução para resolução nativa)"
#define BEETLE_OPT_SUPER_SAMPLING_INFO_0_PT_BR "Reduz o conteúdo renderizado de uma resolução interna superior a nativa. Ao combinar isso com multiplicadores de aumento de resolução interna permite que os jogos sejam exibidos com objetos 3D suavizados em baixa resolução nativa. Produz melhores resultados quando aplicado a títulos que misturam elementos 2D e 3D (por exemplo, personagens 3D em fundos pré-renderizados), e funciona bem em conjunto com shaders CRT. Somente suportado pelo renderizador Vulkan. Nota: É recomendado desativar o 'Pontilhamento' quando esta opção estiver ativada."
#define BEETLE_OPT_MSAA_LABEL_PT_BR "Anti-serrilhado MSAA"
#define BEETLE_OPT_MSAA_INFO_0_PT_BR "MSAA melhora a aparência de objetos 3D. Compatível apenas com o renderizador Vulkan."
#define OPTION_VAL_1X_PT_BR NULL
#define BEETLE_OPT_MDEC_YUV_LABEL_PT_BR "Filtro cromático YUV para MDEC"
#define BEETLE_OPT_MDEC_YUV_INFO_0_PT_BR "Melhora a qualidade da reprodução de FMV, reduzindo os artefatos dos 'macroblocos' (quadrados e arestas recortadas). Somente suportado pelo renderizador Vulkan."
#define BEETLE_OPT_TRACK_TEXTURES_LABEL_PT_BR "Rastreio de texturas"
#define BEETLE_OPT_TRACK_TEXTURES_INFO_0_PT_BR "Pré-requisito para extração e substituição de texturas. Provavelmente causará falhas na maioria dos jogos."
#define BEETLE_OPT_DUMP_TEXTURES_LABEL_PT_BR "Despejo de texturas"
#define BEETLE_OPT_DUMP_TEXTURES_INFO_0_PT_BR "Despeja as texturas usadas par <cd>-texture-dump/"
#define BEETLE_OPT_REPLACE_TEXTURES_LABEL_PT_BR "Substituir texturas"
#define BEETLE_OPT_REPLACE_TEXTURES_INFO_0_PT_BR "Substitui as texturas usando versões HD de <cd>-texture-replacements/"
#define BEETLE_OPT_PGXP_MODE_LABEL_PT_BR "Modo de operação do PGXP"
#define BEETLE_OPT_PGXP_MODE_INFO_0_PT_BR "Renderiza objetos 3D com precisão subpixel, minimizando distorções e jitter observados no hardware original devido ao uso de coordenadas de vértices em ponto fixo. O modo 'Apenas Memória' apresenta problemas mínimos de compatibilidade e é recomendado para uso geral. O modo 'Memória + CPU (Buggy)' pode reduzir ainda mais o jitter, mas possui altos requisitos de desempenho e pode causar diversos erros de geometria."
#define OPTION_VAL_MEMORY_ONLY_PT_BR "Apenas memória"
#define OPTION_VAL_MEMORY_CPU_PT_BR "Memória + CPU (instável)"
#define BEETLE_OPT_PGXP_2D_TOL_LABEL_PT_BR "Tolerância à geometria 2D do PGXP"
#define BEETLE_OPT_PGXP_2D_TOL_INFO_0_PT_BR "Oculta erros mais evidentes nas operações do PGXP: este valor especifica a tolerância em que os valores do PGXP serão mantidos no caso de geometrias sem informações de profundidade adequadas."
#define OPTION_VAL_0PX_PT_BR "0 px"
#define OPTION_VAL_1PX_PT_BR "1 px"
#define OPTION_VAL_2PX_PT_BR "2 px"
#define OPTION_VAL_3PX_PT_BR "3 px"
#define OPTION_VAL_4PX_PT_BR "4 px"
#define OPTION_VAL_5PX_PT_BR "5 px"
#define OPTION_VAL_6PX_PT_BR "6 px"
#define OPTION_VAL_7PX_PT_BR "7 px"
#define OPTION_VAL_8PX_PT_BR "8 px"
#define BEETLE_OPT_PGXP_NCLIP_LABEL_PT_BR "Seleção de primitivos do PGXP"
#define BEETLE_OPT_PGXP_NCLIP_INFO_0_PT_BR "Usa a implementação de NCLIP do PGXP. Melhora a aparência visual ao preencher os furos nas geometrias que tenham coordenadas PGXP. Provoca colisões em certos jogos e circunstâncias."
#define BEETLE_OPT_PGXP_VERTEX_LABEL_PT_BR "Cache de vértice do PGXP"
#define BEETLE_OPT_PGXP_VERTEX_INFO_0_PT_BR "Faz o cache das posições de vértices melhoradas pelo PGXP para reutilização ao desenhar polígonos. Pode melhorar o alinhamento dos objetos e reduzir bordas, mas os falsos positivos ao usar o cache podem produzir defeitos visuais. Atualmente é recomendado desativar esta opção. Só se aplicará quando o modo de operação PGXP estiver ativado. Funciona somente com renderizadores de hardware."
#define BEETLE_OPT_PGXP_TEXTURE_LABEL_PT_BR "Textura fiel à perspectiva do PGXP"
#define BEETLE_OPT_PGXP_TEXTURE_INFO_0_PT_BR NULL
#define BEETLE_OPT_DISPLAY_INTERNAL_FPS_LABEL_PT_BR "Mostrar QPS interno"
#define BEETLE_OPT_DISPLAY_INTERNAL_FPS_INFO_0_PT_BR "Exibir a taxa de quadros internos na qual o sistema PlayStation emulado está renderizando o conteúdo. Nota: Requer que as notificações na tela sejam habilitadas na interface libretro."
#define BEETLE_OPT_DISPLAY_OSD_LABEL_PT_BR "Mostrar mensagens na tela"
#define BEETLE_OPT_DISPLAY_OSD_INFO_0_PT_BR "Mostrar mensagens na tela geradas pelo núcleo."
#define BEETLE_OPT_DISPLAY_VRAM_LABEL_PT_BR "Exibir a VRAM completa (depuração)"
#define BEETLE_OPT_DISPLAY_VRAM_INFO_0_PT_BR "Visualiza toda a VRAM do console emulado. Somente suportada pelos renderizadores de hardware OpenGL e Vulkan. Nota: Isto é para fins de depuração, e normalmente deve estar desativado."
#define BEETLE_OPT_ANALOG_CALIBRATION_LABEL_PT_BR "Auto-calibração do analógico"
#define BEETLE_OPT_ANALOG_CALIBRATION_INFO_0_PT_BR "Ativa a calibração dinâmica das entradas analógicas quando o dispositivo estiver definido como DualShock, Analog Controller, Analog Joystick ou neGcon. Os valores máximos de entrada registrados são monitorados em tempo real e usados para escalar as coordenadas analógicas enviadas ao emulador. Isso deve ser usado para jogos como Mega Man Legends 2, que esperam valores maiores do que os fornecidos por controles modernos. Para obter melhores resultados, os analógicos devem ser girados até o limite máximo para ajustar o algoritmo de calibração sempre que o conteúdo for carregado."
#define BEETLE_OPT_ANALOG_TOGGLE_LABEL_PT_BR "Alternar o modo analógico do DualShock"
#define BEETLE_OPT_ANALOG_TOGGLE_INFO_0_PT_BR "Quando o tipo de dispositivo de entrada é o DualShock, esta opção permite que o DualShock emulado seja alternado entre os modos DIGITAL e ANALÓGICO, assim como no hardware original. O modo também pode ser alternado usando a combinação de botões selecionada."
#define OPTION_VAL_ENABLED_ANALOG_PT_BR "Analógico por padrão"
#define BEETLE_OPT_ANALOG_TOGGLE_COMBO_LABEL_PT_BR "Combinação de botões do modo analógico do DualShock"
#define BEETLE_OPT_ANALOG_TOGGLE_COMBO_INFO_0_PT_BR "A combinação de botões que será usada para alternar entre os modos DIGITAL e ANALÓGICO do DualShock emulado."
#define OPTION_VAL_L1_L2_R1_R2_START_SELECT_PT_BR NULL
#define OPTION_VAL_L1_R1_SELECT_PT_BR NULL
#define OPTION_VAL_L1_R1_START_PT_BR NULL
#define OPTION_VAL_L1_R1_L3_PT_BR NULL
#define OPTION_VAL_L1_R1_R3_PT_BR NULL
#define OPTION_VAL_L2_R2_SELECT_PT_BR NULL
#define OPTION_VAL_L2_R2_START_PT_BR NULL
#define OPTION_VAL_L2_R2_L3_PT_BR NULL
#define OPTION_VAL_L2_R2_R3_PT_BR NULL
#define OPTION_VAL_L3_R3_PT_BR NULL
#define BEETLE_OPT_ANALOG_TOGGLE_HOLD_LABEL_PT_BR "Tempo de espera para a combinação de botões do modo analógico do DualShock"
#define BEETLE_OPT_ANALOG_TOGGLE_HOLD_INFO_0_PT_BR "Tempo necessário para manter pressionada a combinação de botões do modo analógico."
#define OPTION_VAL_0_PT_BR "0 segundos"
#define OPTION_VAL_1_PT_BR "1 segundo"
#define OPTION_VAL_2_PT_BR "2 segundos"
#define OPTION_VAL_3_PT_BR "3 segundos"
#define OPTION_VAL_4_PT_BR "4 segundos"
#define OPTION_VAL_5_PT_BR "5 segundos"
#define BEETLE_OPT_ENABLE_MULTITAP_PORT1_LABEL_PT_BR "Multitap na Porta 1"
#define BEETLE_OPT_ENABLE_MULTITAP_PORT1_INFO_0_PT_BR "Ativa o sistema Multitap na porta 1."
#define BEETLE_OPT_ENABLE_MULTITAP_PORT2_LABEL_PT_BR "Multitap na Porta 2"
#define BEETLE_OPT_ENABLE_MULTITAP_PORT2_INFO_0_PT_BR "Ativa o sistema Multitap na porta 2."
#define BEETLE_OPT_GUN_INPUT_MODE_LABEL_PT_BR "Modo de entrada de pistolas de luz"
#define BEETLE_OPT_GUN_INPUT_MODE_INFO_0_PT_BR "Usa entrada de 'Light Gun' ou 'Tela Sensível ao Toque' controlada pelo mouse quando o tipo de dispositivo estiver definido como 'Guncon/G-Con 45' ou 'Justifier'."
#define OPTION_VAL_LIGHTGUN_PT_BR "Pistola de luz"
#define OPTION_VAL_TOUCHSCREEN_PT_BR "Tela sensível ao toque"
#define BEETLE_OPT_GUN_CURSOR_LABEL_PT_BR "Mira da arma"
#define BEETLE_OPT_GUN_CURSOR_INFO_0_PT_BR "Desenha um cursor na tela ao usar os tipos de dispositivo de entrada 'Guncon/G-Con 45' e 'Justifier'."
#define OPTION_VAL_CROSS_PT_BR "Mira de cruz"
#define OPTION_VAL_DOT_PT_BR "Mira de ponto"
#define BEETLE_OPT_CROSSHAIR_COLOR_P1_LABEL_PT_BR "Cor da Mira da Arma: Porta 1"
#define BEETLE_OPT_CROSSHAIR_COLOR_P1_INFO_0_PT_BR "Cor da mira da light gun para a porta 1."
#define OPTION_VAL_RED_PT_BR "Vermelha"
#define OPTION_VAL_BLUE_PT_BR "Azul"
#define OPTION_VAL_GREEN_PT_BR "Verde"
#define OPTION_VAL_ORANGE_PT_BR "Laranja"
#define OPTION_VAL_YELLOW_PT_BR "Amarela"
#define OPTION_VAL_CYAN_PT_BR "Azul-marinho"
#define OPTION_VAL_PINK_PT_BR "Rosa"
#define OPTION_VAL_PURPLE_PT_BR "Violeta"
#define OPTION_VAL_BLACK_PT_BR "Preta"
#define OPTION_VAL_WHITE_PT_BR "Branca"
#define BEETLE_OPT_CROSSHAIR_COLOR_P2_LABEL_PT_BR "Cor da Mira da Arma: Porta 2"
#define BEETLE_OPT_CROSSHAIR_COLOR_P2_INFO_0_PT_BR "Cor da mira da light gun para a porta 2."
#define BEETLE_OPT_MOUSE_SENSITIVITY_LABEL_PT_BR "Sensibilidade do mouse"
#define BEETLE_OPT_MOUSE_SENSITIVITY_INFO_0_PT_BR "Responsividade do tipo de dispositivo de entrada 'Mouse'."
#define OPTION_VAL_5_O42_PT_BR NULL
#define OPTION_VAL_10_PT_BR NULL
#define OPTION_VAL_15_PT_BR NULL
#define OPTION_VAL_20_PT_BR NULL
#define OPTION_VAL_25_PT_BR NULL
#define OPTION_VAL_30_PT_BR NULL
#define OPTION_VAL_35_PT_BR NULL
#define OPTION_VAL_40_PT_BR NULL
#define OPTION_VAL_45_PT_BR NULL
#define OPTION_VAL_55_PT_BR NULL
#define OPTION_VAL_65_PT_BR NULL
#define OPTION_VAL_75_PT_BR NULL
#define OPTION_VAL_85_PT_BR NULL
#define OPTION_VAL_95_PT_BR NULL
#define OPTION_VAL_100_O42_PT_BR NULL
#define OPTION_VAL_105_PT_BR NULL
#define OPTION_VAL_115_PT_BR NULL
#define OPTION_VAL_125_PT_BR NULL
#define OPTION_VAL_135_PT_BR NULL
#define OPTION_VAL_145_PT_BR NULL
#define OPTION_VAL_155_PT_BR NULL
#define OPTION_VAL_165_PT_BR NULL
#define OPTION_VAL_175_PT_BR NULL
#define OPTION_VAL_185_PT_BR NULL
#define OPTION_VAL_195_PT_BR NULL
#define BEETLE_OPT_NEGCON_RESPONSE_LABEL_PT_BR "Resposta ao torcer o neGcon"
#define BEETLE_OPT_NEGCON_RESPONSE_INFO_0_PT_BR "Tipo de resposta do analógico esquerdo ao simular a ação de 'torção' dos dispositivos de entrada 'neGcon' emulados. O deslocamento do analógico pode ser mapeado para o ângulo de rotação do neGcon de forma linear, quadrática ou cúbica. 'Quadrática' permite maior precisão do que 'Linear' ao realizar movimentos pequenos. 'Cúbica' aumenta ainda mais a precisão em movimentos pequenos, mas 'exagera' movimentos maiores. Nota: 'Linear' é recomendada apenas ao usar periféricos de volante. Controles convencionais implementam entrada analógica de forma fundamentalmente diferente do mecanismo de 'torção' do neGcon, de modo que o mapeamento linear amplifica excessivamente movimentos pequenos, prejudicando o controle fino. Geralmente, 'Quadrática' fornece a aproximação mais próxima do hardware real."
#define OPTION_VAL_LINEAR_PT_BR NULL
#define OPTION_VAL_QUADRATIC_PT_BR "Quadrático"
#define OPTION_VAL_CUBIC_PT_BR "Cúbico"
#define BEETLE_OPT_NEGCON_DEADZONE_LABEL_PT_BR "Zona morta ao torcer o NegCon"
#define BEETLE_OPT_NEGCON_DEADZONE_INFO_0_PT_BR "Zona morta do analógico esquerdo ao simular a ação de 'torção' dos dispositivos de entrada 'neGcon' emulados. Usada para eliminar o drift do controle. Nota: A maioria dos títulos compatíveis com neGcon oferece opções no jogo para definir um valor de zona morta da 'torção'. Para evitar perda de precisão, a zona morta no jogo deve sempre ser definida como zero. Quaisquer ajustes necessários devem ser aplicados somente por meio desta opção do núcleo. Isso é particularmente importante quando 'Resposta de Torção do neGcon' estiver definida como 'Quadrática' ou 'Cúbica'."
#define OPTION_VAL_0_O44_PT_BR NULL
#define BEETLE_OPT_USE_MEDNAFEN_MEMCARD0_METHOD_LABEL_PT_BR "Método do Memory Card"
#define BEETLE_OPT_USE_MEDNAFEN_MEMCARD0_METHOD_INFO_0_PT_BR "Formato de dados de salvamento usado para o slot esquerdo do memory card. 'Mednafen' pode ser usado para compatibilidade com a versão independente do Mednafen. Quando usado com o Beetle PSX, os saves Libretro (.srm) e Mednafen (.mcr) possuem formatos internamente idênticos e podem ser convertidos entre si por meio da renomeação. Reinicialização necessária."
#define OPTION_VAL_LIBRETRO_PT_BR NULL
#define OPTION_VAL_MEDNAFEN_PT_BR NULL
#define BEETLE_OPT_ENABLE_MEMCARD1_LABEL_PT_BR "Slot 2 do Memory Card"
#define BEETLE_OPT_ENABLE_MEMCARD1_INFO_0_PT_BR "Emula um segundo memory card no slot direito do memory card. Nota: Alguns jogos exigem que esta opção esteja desativada para funcionamento correto (por exemplo, Codename Tenka).Nota: O Slot 2 do Memory Card utiliza o formato de salvamento Mednafen (.mcr). Reinicialização necessária."
#define BEETLE_OPT_SHARED_MEMORY_CARDS_LABEL_PT_BR "Memory Cards Compartilhados"
#define BEETLE_OPT_SHARED_MEMORY_CARDS_INFO_0_PT_BR "Usa memory cards compartilhados em vez de cartões separados para cada conteúdo. Nota: Se o 'Método do Memory Card' estiver definido como 'Libretro', apenas o Slot 2 do Memory Card será afetado. Reinicialização necessária."
#define BEETLE_OPT_MEMCARD_LEFT_INDEX_LABEL_PT_BR "Índice do Slot 1 do Memory Card"
#define BEETLE_OPT_MEMCARD_LEFT_INDEX_INFO_0_PT_BR "Memory Card carregado atualmente no slot da esquerda. Esta opção só funcionará se o Método de Memory Card estiver definido como Mednafen."
#define OPTION_VAL_0_O48_PT_BR "0 (Padrão)"
#define BEETLE_OPT_MEMCARD_RIGHT_INDEX_LABEL_PT_BR "Índice do Slot 2 do Memory Card"
#define BEETLE_OPT_MEMCARD_RIGHT_INDEX_INFO_0_PT_BR "Memory Card carregado atualmente no slot da direita. Esta opção só funcionará se o Slot de Memory Card 2 estiver ativado."
#define OPTION_VAL_1_O49_PT_BR "1 (Padrão)"
#define BEETLE_OPT_LINE_RENDER_LABEL_PT_BR "Hack para o efeito 'Line-to-Quad'"
#define BEETLE_OPT_LINE_RENDER_INFO_0_PT_BR "Alguns jogos (por exemplo, Doom, Hexen, Soul Blade, etc.) desenham linhas horizontais esticando triângulos de um pixel de altura pela tela, que são rasterizados como uma linha de pixels no hardware original. Este hack detecta esses pequenos triângulos e os converte em quadriláteros conforme necessário, permitindo que sejam exibidos corretamente nos renderizadores de hardware e em resoluções internas aumentadas. A opção 'Agressivo' é necessária para alguns títulos (por exemplo, Dark Forces, Duke Nukem), mas pode causar falhas gráficas em outros casos. Deixe em 'Padrão' se não tiver certeza."
#define OPTION_VAL_DEFAULT_PT_BR "Padrão"
#define OPTION_VAL_AGGRESSIVE_PT_BR "Agressivo"
#define BEETLE_OPT_WIDESCREEN_HACK_LABEL_PT_BR "Hack de modificação de widescreen"
#define BEETLE_OPT_WIDESCREEN_HACK_INFO_0_PT_BR "Renderiza o conteúdo 3D em formato anamórfico e torna o quadro de buffer emulado em uma proporção de tela widescreen. Produz os melhores resultados com jogos feitos inteiramente em 3D. Os elementos 2D serão esticados horizontalmente e não poderão mais ser alinhados."
#define BEETLE_OPT_WIDESCREEN_HACK_ASPECT_RATIO_LABEL_PT_BR "Proporção da tela do hack de modificação de widescreen"
#define BEETLE_OPT_WIDESCREEN_HACK_ASPECT_RATIO_INFO_0_PT_BR "Proporção de tela a ser usada pelo Hack de Modo Widescreen."
#define OPTION_VAL_16_9_PT_BR NULL
#define OPTION_VAL_16_10_PT_BR NULL
#define OPTION_VAL_18_9_PT_BR NULL
#define OPTION_VAL_19_9_PT_BR NULL
#define OPTION_VAL_20_9_PT_BR NULL
#define OPTION_VAL_21_9_PT_BR NULL
#define OPTION_VAL_32_9_PT_BR NULL
#define BEETLE_OPT_CPU_DYNAREC_LABEL_PT_BR "Dynarec de CPU"
#define BEETLE_OPT_CPU_DYNAREC_INFO_0_PT_BR "Recompilar dinamicamente as instruções da CPU para instruções nativas. Muito mais rápido do que o intérprete, mas o tempo da CPU é menos preciso e pode ter erros."
#define OPTION_VAL_DISABLED_PT_BR "Desativada (interpretador do Beetle)"
#define OPTION_VAL_EXECUTE_PT_BR "Máximo desempenho"
#define OPTION_VAL_EXECUTE_ONE_PT_BR "Verificação de intervalo entre ciclos"
#define OPTION_VAL_RUN_INTERPRETER_PT_BR "Interpretador Lightrec"
#define BEETLE_OPT_DYNAREC_INVALIDATE_LABEL_PT_BR "Invalidação de código do dynarec"
#define BEETLE_OPT_DYNAREC_INVALIDATE_INFO_0_PT_BR "Alguns jogos requerem invalidação 'Completa', alguns requerem 'Apenas DMA'."
#define OPTION_VAL_FULL_PT_BR "Completa"
#define OPTION_VAL_DMA_PT_BR "Apenas DMA (um pouco mais rápido)"
#define BEETLE_OPT_DYNAREC_EVENTCYCLES_LABEL_PT_BR "Ciclos de eventos de DMA, GPU, MDEC e Temporizador do dynarec"
#define BEETLE_OPT_DYNAREC_EVENTCYCLES_INFO_0_PT_BR "Indica o número máximo de ciclos executados pela CPU antes de verificar atualizações na GPU, DMA, MDEC e no Temporizador, um número maior será mais rápido, tendo muito menos impacto no interpretador do Beetle do que o dynarec."
#define OPTION_VAL_128_PT_BR "128 (padrão)"
#define BEETLE_OPT_DYNAREC_SPU_SAMPLES_LABEL_PT_BR "Amostras de SPU do Dynarec"
#define BEETLE_OPT_DYNAREC_SPU_SAMPLES_INFO_0_PT_BR "Indica o número máximo de amostras de SPU para executar antes que uma atualização do SPU esteja selecionada, um número maior será mais rápido, mas um valor que não seja 1 causará defeitos sonoros em alguns jogos."
#define BEETLE_OPT_CORE_TIMING_FPS_LABEL_PT_BR "Tempos dos quadros reportado pelo núcleo"
#define BEETLE_OPT_CORE_TIMING_FPS_INFO_0_PT_BR "A Alternância Automática permite que o core mude entre taxas progressivas e entrelaçadas."
#define OPTION_VAL_AUTO_TOGGLE_PT_BR "Alternância Automática"
#define OPTION_VAL_FORCE_PROGRESSIVE_PT_BR "Forçar Taxa Progressiva"
#define OPTION_VAL_FORCE_INTERLACED_PT_BR "Forçar taxa entrelaçada"
#define BEETLE_OPT_DEINTERLACER_LABEL_PT_BR "Método de desentrelaçamento"
#define BEETLE_OPT_DEINTERLACER_INFO_0_PT_BR "'Weave' exibe ambos os campos no mesmo quadro para obter uma imagem estática em alta resolução total. 'Bob' sacrifica a clareza da resolução em favor da clareza de movimento. 'Bob (Offset)' é o Bob com o campo atual duplicado para a altura total da superfície com um deslocamento de uma linha entre os campos, preservando a resolução de saída do libretro enquanto mantém o movimento suave no estilo Bob. 'FastMAD (Adaptativo ao Movimento)' entrelaça pixels estáticos e interpola verticalmente pixels em movimento, pixel por pixel, mantendo a altura de saída total sem o efeito de pente em movimento do Weave nem a resolução reduzida pela metade do Bob. 'Off' faz com que o renderizador por software corresponda ao comportamento do renderizador por hardware (Vulkan/OpenGL), rasterizando todas as linhas da VRAM a cada quadro e adiando o scanout para o final do quadro, eliminando o efeito de pente em movimento (comb-on-motion) nos jogos entrelaçados. Nota: programas que fazem DMA para a VRAM durante o scanout (algumas demos que usam truques de wrap de framebuffer PAL com >512 linhas), ou que atualizam sprites em movimento via transferências de imagem da CPU para a VRAM vinculadas ao campo exibido (ex.: Pro Pinball), podem exibir linhas desatualizadas ou efeito fantasma (ghosting) nos sprites em 'Off'; use 'Weave', 'Bob', 'Bob (Offset)' ou 'FastMAD' para eles."
#define OPTION_VAL_WEAVE_PT_BR NULL
#define OPTION_VAL_BOB_PT_BR NULL
#define OPTION_VAL_BOB_OFFSET_PT_BR NULL
#define OPTION_VAL_FASTMAD_PT_BR "FastMAD (Adaptativo ao Movimento)"
#define BEETLE_OPT_FRAME_DUPING_LABEL_PT_BR "Duplicação de Frames"
#define BEETLE_OPT_FRAME_DUPING_INFO_0_PT_BR "Repete o frame anterior se o core não tiver nada novo para exibir."
#define BEETLE_OPT_PAL_VIDEO_TIMING_OVERRIDE_LABEL_PT_BR "Sobreposição de Sincronização de Vídeo PAL"
#define BEETLE_OPT_PAL_VIDEO_TIMING_OVERRIDE_INFO_0_PT_BR "Devido aos diferentes padrões, os jogos PAL geralmente parecem mais lentos em comparação com os lançamentos NTSC Americanos ou Japoneses. Esta opção pode ser usada para sobrepor as sincronizações PAL a fim de tentar executar esses jogos com a taxa de quadros NTSC."
#define BEETLE_OPT_ASPECT_RATIO_LABEL_PT_BR "Proporção de tela do núcleo"
#define BEETLE_OPT_ASPECT_RATIO_INFO_0_PT_BR "Esta configuração é ignorada quando o Hack de Modo Widescreen ou a opção Exibir VRAM Completa estão ativados."
#define OPTION_VAL_CORRECTED_PT_BR "Corrigida"
#define OPTION_VAL_UNCORRECTED_PT_BR "Não corrigida"
#define OPTION_VAL_4_3_PT_BR "Forçar 4:3"
#define OPTION_VAL_NTSC_PT_BR "Forçar NTSC"
#define BEETLE_OPT_CROP_OVERSCAN_LABEL_PT_BR "Cortar overscan"
#define BEETLE_OPT_CROP_OVERSCAN_INFO_0_PT_BR "Imagem nativa completa ou sem bordas."
#define OPTION_VAL_STATIC_PT_BR NULL
#define OPTION_VAL_SMART_PT_BR NULL
#define BEETLE_OPT_IMAGE_CROP_LABEL_PT_BR "Corte adicional"
#define BEETLE_OPT_IMAGE_CROP_INFO_0_PT_BR "Reduz a largura da imagem cortada pelo número de pixels especificado."
#define OPTION_VAL_9PX_PT_BR "9 px"
#define OPTION_VAL_10PX_PT_BR "10 px"
#define OPTION_VAL_11PX_PT_BR "11 px"
#define OPTION_VAL_12PX_PT_BR "12 px"
#define OPTION_VAL_13PX_PT_BR "13 px"
#define OPTION_VAL_14PX_PT_BR "14 px"
#define OPTION_VAL_15PX_PT_BR "15 px"
#define OPTION_VAL_16PX_PT_BR "16 px"
#define OPTION_VAL_17PX_PT_BR "17 px"
#define OPTION_VAL_18PX_PT_BR "18 px"
#define OPTION_VAL_19PX_PT_BR "19 px"
#define OPTION_VAL_20PX_PT_BR "20 px"
#define BEETLE_OPT_IMAGE_OFFSET_LABEL_PT_BR "Deslocar imagem cortada"
#define BEETLE_OPT_IMAGE_OFFSET_INFO_0_PT_BR "Desloca horizontalmente para a direita (positivo) ou esquerda (negativo) pelo número de pixels especificado. Pode ser usado para corrigir problemas de alinhamento. Suportado apenas pelo renderizador de software."
#define OPTION_VAL_12PX_O64_PT_BR NULL
#define OPTION_VAL_11PX_O64_PT_BR NULL
#define OPTION_VAL_10PX_O64_PT_BR NULL
#define OPTION_VAL_9PX_O64_PT_BR NULL
#define OPTION_VAL_8PX_O64_PT_BR NULL
#define OPTION_VAL_7PX_O64_PT_BR NULL
#define OPTION_VAL_6PX_O64_PT_BR NULL
#define OPTION_VAL_5PX_O64_PT_BR NULL
#define OPTION_VAL_4PX_O64_PT_BR NULL
#define OPTION_VAL_3PX_O64_PT_BR NULL
#define OPTION_VAL_2PX_O64_PT_BR NULL
#define OPTION_VAL_1PX_O64_PT_BR NULL
#define OPTION_VAL_1PX_O64_0_PT_BR NULL
#define OPTION_VAL_2PX_O64_0_PT_BR NULL
#define OPTION_VAL_3PX_O64_0_PT_BR NULL
#define OPTION_VAL_4PX_O64_0_PT_BR NULL
#define OPTION_VAL_5PX_O64_0_PT_BR NULL
#define OPTION_VAL_6PX_O64_0_PT_BR NULL
#define OPTION_VAL_7PX_O64_0_PT_BR NULL
#define OPTION_VAL_8PX_O64_0_PT_BR NULL
#define OPTION_VAL_9PX_O64_0_PT_BR NULL
#define OPTION_VAL_10PX_O64_0_PT_BR NULL
#define OPTION_VAL_11PX_O64_0_PT_BR NULL
#define OPTION_VAL_12PX_O64_0_PT_BR NULL
#define BEETLE_OPT_IMAGE_OFFSET_CYCLES_LABEL_PT_BR "Deslocar imagem horizontal (ciclos de GPU)"
#define BEETLE_OPT_IMAGE_OFFSET_CYCLES_INFO_0_PT_BR "Valores positivos movem a imagem para a direita, valores negativos movem a imagem para a esquerda. Compatível apenas com os renderizadores por hardware."
#define BEETLE_OPT_INITIAL_SCANLINE_LABEL_PT_BR "Linha de varredura inicial (NTSC)"
#define BEETLE_OPT_INITIAL_SCANLINE_INFO_0_PT_BR "Primeira linha de varredura exibida ao executar conteúdo NTSC. Valores maiores que zero reduzem a altura da imagem de saída ao cortar pixels da borda superior. Pode ser usado para compensar o efeito de barras pretas."
#define BEETLE_OPT_LAST_SCANLINE_LABEL_PT_BR "Linha de varredura final (NTSC)"
#define BEETLE_OPT_LAST_SCANLINE_INFO_0_PT_BR "Última linha de varredura exibida ao executar conteúdo NTSC. Valores menores que 239 reduzem a altura da imagem de saída ao cortar pixels da borda inferior. Pode ser usado para compensar o efeito de barras pretas."
#define OPTION_VAL_239_PT_BR "239 (padrão)"
#define BEETLE_OPT_INITIAL_SCANLINE_PAL_LABEL_PT_BR "Linha de varredura inicial (PAL)"
#define BEETLE_OPT_INITIAL_SCANLINE_PAL_INFO_0_PT_BR "Primeira linha de varredura exibida ao executar conteúdo PAL. Valores maiores que zero reduzem a altura da imagem de saída ao cortar pixels da borda superior. Pode ser usado para compensar o efeito de barras pretas."
#define BEETLE_OPT_LAST_SCANLINE_PAL_LABEL_PT_BR "Linha de varredura final (PAL)"
#define BEETLE_OPT_LAST_SCANLINE_PAL_INFO_0_PT_BR "Última linha de varredura exibida ao executar conteúdo PAL. Valores menores que 287 reduzem a altura da imagem de saída ao cortar pixels da borda inferior. Pode ser usado para compensar o efeito de barras pretas."
#define OPTION_VAL_287_PT_BR "287 (padrão)"

struct retro_core_option_v2_category option_cats_pt_br[] = {
   {
      "system",
      CATEGORY_SYSTEM_LABEL_PT_BR,
      CATEGORY_SYSTEM_INFO_0_PT_BR
   },
   {
      "video",
      CATEGORY_VIDEO_LABEL_PT_BR,
      CATEGORY_VIDEO_INFO_0_PT_BR
   },
   {
      "pgxp",
      CATEGORY_PGXP_LABEL_PT_BR,
      CATEGORY_PGXP_INFO_0_PT_BR
   },
   {
      "osd",
      CATEGORY_OSD_LABEL_PT_BR,
      CATEGORY_OSD_INFO_0_PT_BR
   },
   {
      "input",
      CATEGORY_INPUT_LABEL_PT_BR,
      CATEGORY_INPUT_INFO_0_PT_BR
   },
   {
      "memcards",
      CATEGORY_MEMCARDS_LABEL_PT_BR,
      CATEGORY_MEMCARDS_INFO_0_PT_BR
   },
   {
      "hacks",
      CATEGORY_HACKS_LABEL_PT_BR,
      CATEGORY_HACKS_INFO_0_PT_BR
   },
   { NULL, NULL, NULL },
};
struct retro_core_option_v2_definition option_defs_pt_br[] = {
   {
      BEETLE_OPT(cpu_freq_scale),
      BEETLE_OPT_CPU_FREQ_SCALE_LABEL_PT_BR,
      NULL,
      BEETLE_OPT_CPU_FREQ_SCALE_INFO_0_PT_BR,
      NULL,
      "system",
      {
         { "50%",           OPTION_VAL_50_PT_BR },
         { "60%",           OPTION_VAL_60_PT_BR },
         { "70%",           OPTION_VAL_70_PT_BR },
         { "80%",           OPTION_VAL_80_PT_BR },
         { "90%",           OPTION_VAL_90_PT_BR },
         { "100%",          OPTION_VAL_100_PT_BR },
         { "110%",          OPTION_VAL_110_PT_BR },
         { "120%",          OPTION_VAL_120_PT_BR },
         { "130%",          OPTION_VAL_130_PT_BR },
         { "140%",          OPTION_VAL_140_PT_BR },
         { "150%",          OPTION_VAL_150_PT_BR },
         { "160%",          OPTION_VAL_160_PT_BR },
         { "170%",          OPTION_VAL_170_PT_BR },
         { "180%",          OPTION_VAL_180_PT_BR },
         { "190%",          OPTION_VAL_190_PT_BR },
         { "200%",          OPTION_VAL_200_PT_BR },
         { "210%",          OPTION_VAL_210_PT_BR },
         { "220%",          OPTION_VAL_220_PT_BR },
         { "230%",          OPTION_VAL_230_PT_BR },
         { "240%",          OPTION_VAL_240_PT_BR },
         { "250%",          OPTION_VAL_250_PT_BR },
         { "260%",          OPTION_VAL_260_PT_BR },
         { "270%",          OPTION_VAL_270_PT_BR },
         { "280%",          OPTION_VAL_280_PT_BR },
         { "290%",          OPTION_VAL_290_PT_BR },
         { "300%",          OPTION_VAL_300_PT_BR },
         { "310%",          OPTION_VAL_310_PT_BR },
         { "320%",          OPTION_VAL_320_PT_BR },
         { "330%",          OPTION_VAL_330_PT_BR },
         { "340%",          OPTION_VAL_340_PT_BR },
         { "350%",          OPTION_VAL_350_PT_BR },
         { "360%",          OPTION_VAL_360_PT_BR },
         { "370%",          OPTION_VAL_370_PT_BR },
         { "380%",          OPTION_VAL_380_PT_BR },
         { "390%",          OPTION_VAL_390_PT_BR },
         { "400%",          OPTION_VAL_400_PT_BR },
         { "410%",          OPTION_VAL_410_PT_BR },
         { "420%",          OPTION_VAL_420_PT_BR },
         { "430%",          OPTION_VAL_430_PT_BR },
         { "440%",          OPTION_VAL_440_PT_BR },
         { "450%",          OPTION_VAL_450_PT_BR },
         { "460%",          OPTION_VAL_460_PT_BR },
         { "470%",          OPTION_VAL_470_PT_BR },
         { "480%",          OPTION_VAL_480_PT_BR },
         { "490%",          OPTION_VAL_490_PT_BR },
         { "500%",          OPTION_VAL_500_PT_BR },
         { "510%",          OPTION_VAL_510_PT_BR },
         { "520%",          OPTION_VAL_520_PT_BR },
         { "530%",          OPTION_VAL_530_PT_BR },
         { "540%",          OPTION_VAL_540_PT_BR },
         { "550%",          OPTION_VAL_550_PT_BR },
         { "560%",          OPTION_VAL_560_PT_BR },
         { "570%",          OPTION_VAL_570_PT_BR },
         { "580%",          OPTION_VAL_580_PT_BR },
         { "590%",          OPTION_VAL_590_PT_BR },
         { "600%",          OPTION_VAL_600_PT_BR },
         { "610%",          OPTION_VAL_610_PT_BR },
         { "620%",          OPTION_VAL_620_PT_BR },
         { "630%",          OPTION_VAL_630_PT_BR },
         { "640%",          OPTION_VAL_640_PT_BR },
         { "650%",          OPTION_VAL_650_PT_BR },
         { "660%",          OPTION_VAL_660_PT_BR },
         { "670%",          OPTION_VAL_670_PT_BR },
         { "680%",          OPTION_VAL_680_PT_BR },
         { "690%",          OPTION_VAL_690_PT_BR },
         { "700%",          OPTION_VAL_700_PT_BR },
         { "710%",          OPTION_VAL_710_PT_BR },
         { "720%",          OPTION_VAL_720_PT_BR },
         { "730%",          OPTION_VAL_730_PT_BR },
         { "740%",          OPTION_VAL_740_PT_BR },
         { "750%",          OPTION_VAL_750_PT_BR },
         { NULL, NULL },
      },
      "100%"
   },
   {
      BEETLE_OPT(gpu_overclock),
      BEETLE_OPT_GPU_OVERCLOCK_LABEL_PT_BR,
      NULL,
      BEETLE_OPT_GPU_OVERCLOCK_INFO_0_PT_BR,
      NULL,
      "system",
      {
         { "1x(native)", OPTION_VAL_1X_NATIVE_PT_BR },
         { "2x",         OPTION_VAL_2X_PT_BR },
         { "4x",         OPTION_VAL_4X_PT_BR },
         { "8x",         OPTION_VAL_8X_PT_BR },
         { "16x",        OPTION_VAL_16X_PT_BR },
         { "32x",        OPTION_VAL_32X_PT_BR },
         { NULL, NULL },
      },
      "1x(native)"
   },
   {
      BEETLE_OPT(gte_overclock),
      BEETLE_OPT_GTE_OVERCLOCK_LABEL_PT_BR,
      NULL,
      BEETLE_OPT_GTE_OVERCLOCK_INFO_0_PT_BR,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
#ifndef EMSCRIPTEN
   {
      BEETLE_OPT(cd_access_method),
      BEETLE_OPT_CD_ACCESS_METHOD_LABEL_PT_BR,
      NULL,
      BEETLE_OPT_CD_ACCESS_METHOD_INFO_0_PT_BR,
      NULL,
      "system",
      {
         { "sync",     OPTION_VAL_SYNC_PT_BR },
         { "async",    OPTION_VAL_ASYNC_PT_BR },
         { "precache", OPTION_VAL_PRECACHE_PT_BR },
         { NULL, NULL },
      },
      "sync"
   },
#endif
   {
      BEETLE_OPT(cd_fastload),
      BEETLE_OPT_CD_FASTLOAD_LABEL_PT_BR,
      NULL,
      BEETLE_OPT_CD_FASTLOAD_INFO_0_PT_BR,
      NULL,
      "system",
      {
         { "2x(native)", OPTION_VAL_2X_NATIVE_PT_BR },
         { "4x",          OPTION_VAL_4X_PT_BR },
         { "6x",          OPTION_VAL_6X_PT_BR },
         { "8x",          OPTION_VAL_8X_PT_BR },
         { "10x",         OPTION_VAL_10X_PT_BR },
         { "12x",         OPTION_VAL_12X_PT_BR },
         { "14x",         OPTION_VAL_14X_PT_BR },
         { NULL, NULL },
      },
      "2x(native)"
   },
   {
      BEETLE_OPT(skip_bios),
      BEETLE_OPT_SKIP_BIOS_LABEL_PT_BR,
      NULL,
      BEETLE_OPT_SKIP_BIOS_INFO_0_PT_BR,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(override_bios),
      BEETLE_OPT_OVERRIDE_BIOS_LABEL_PT_BR,
      NULL,
      BEETLE_OPT_OVERRIDE_BIOS_INFO_0_PT_BR,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "psxonpsp", OPTION_VAL_PSXONPSP_PT_BR },
         { "ps1_rom",  OPTION_VAL_PS1_ROM_PT_BR },
         { "openbios", OPTION_VAL_OPENBIOS_PT_BR     },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(region),
      BEETLE_OPT_REGION_LABEL_PT_BR,
      NULL,
      BEETLE_OPT_REGION_INFO_0_PT_BR,
      NULL,
      "system",
      {
         { "auto",   OPTION_VAL_AUTO_PT_BR },
         { "ntsc-j", OPTION_VAL_NTSC_J_PT_BR },
         { "ntsc-u", OPTION_VAL_NTSC_U_PT_BR },
         { "pal",    OPTION_VAL_PAL_PT_BR },
         { NULL, NULL },
      },
      "auto"
   },
   {
      BEETLE_OPT(internal_resolution),
      BEETLE_OPT_INTERNAL_RESOLUTION_LABEL_PT_BR,
      NULL,
      BEETLE_OPT_INTERNAL_RESOLUTION_INFO_0_PT_BR,
      NULL,
      "video",
      {
         { "1x(native)", OPTION_VAL_1X_NATIVE_PT_BR },
         { "2x",         OPTION_VAL_2X_PT_BR },
         { "4x",         OPTION_VAL_4X_PT_BR },
         { "8x",         OPTION_VAL_8X_PT_BR },
         { "16x",        OPTION_VAL_16X_PT_BR },
         { NULL, NULL },
      },
      "1x(native)"
   },
#if defined(HAVE_OPENGL) || defined(HAVE_OPENGLES) || defined(HAVE_VULKAN)
   {
      BEETLE_OPT(renderer),
      BEETLE_OPT_RENDERER_LABEL_PT_BR,
      NULL,
      BEETLE_OPT_RENDERER_INFO_0_PT_BR,
      NULL,
      "video",
      {
         { "hardware",    OPTION_VAL_HARDWARE_PT_BR },
#if defined(HAVE_OPENGL) || defined(HAVE_OPENGLES)
         { "hardware_gl", OPTION_VAL_HARDWARE_GL_PT_BR },
#endif
#if defined(HAVE_VULKAN)
         { "hardware_vk", OPTION_VAL_HARDWARE_VK_PT_BR },
#endif
         { "software",    OPTION_VAL_SOFTWARE_PT_BR },
         { NULL, NULL },
      },
      "hardware"
   },
   {
      BEETLE_OPT(renderer_software_fb),
      BEETLE_OPT_RENDERER_SOFTWARE_FB_LABEL_PT_BR,
      NULL,
      BEETLE_OPT_RENDERER_SOFTWARE_FB_INFO_0_PT_BR,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled"
   },
#endif
#if defined(HAVE_OPENGL) || defined(HAVE_OPENGLES)
   {
      BEETLE_OPT(depth),
      BEETLE_OPT_DEPTH_LABEL_PT_BR,
      NULL,
      BEETLE_OPT_DEPTH_INFO_0_PT_BR,
      NULL,
      "video",
      {
         { "16bpp(native)", OPTION_VAL_16BPP_NATIVE_PT_BR },
         { "32bpp",         OPTION_VAL_32BPP_PT_BR },
         { NULL, NULL },
      },
      "16bpp(native)"
   },
   // Sort of, it's more like 15-bit high color and 24-bit true color for visible output. The alpha channel is used for mask bit. Vulkan renderer uses ABGR1555_555 for 31 bits internal? FMVs are always 24-bit on all renderers like original hardware (BGR888, no alpha)
#endif
   {
      BEETLE_OPT(dither_mode),
      BEETLE_OPT_DITHER_MODE_LABEL_PT_BR,
      NULL,
      BEETLE_OPT_DITHER_MODE_INFO_0_PT_BR,
      NULL,
      "video",
      {
         { "1x(native)",          OPTION_VAL_1X_NATIVE_PT_BR },
         { "internal resolution", OPTION_VAL_INTERNAL_RESOLUTION_PT_BR },
         { "disabled",            NULL },
         { NULL, NULL },
      },
      "1x(native)"
   },
#ifdef HAVE_VULKAN
   {
      BEETLE_OPT(scaled_uv_offset),
      BEETLE_OPT_SCALED_UV_OFFSET_LABEL_PT_BR,
      NULL,
      BEETLE_OPT_SCALED_UV_OFFSET_INFO_0_PT_BR,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled"
   },
#endif
#if defined(HAVE_OPENGL) || defined(HAVE_OPENGLES) || defined(HAVE_VULKAN)
   {
      BEETLE_OPT(filter),
      BEETLE_OPT_FILTER_LABEL_PT_BR,
      NULL,
      BEETLE_OPT_FILTER_INFO_0_PT_BR,
      NULL,
      "video",
      {
         { "nearest",  OPTION_VAL_NEAREST_PT_BR },
         { "SABR",     OPTION_VAL_SABR_PT_BR },
         { "xBR",      OPTION_VAL_XBR_PT_BR },
         { "bilinear", OPTION_VAL_BILINEAR_PT_BR },
         { "3-point",  OPTION_VAL_3_POINT_PT_BR },
         { "JINC2",    OPTION_VAL_JINC2_PT_BR },
         { NULL, NULL },
      },
      "nearest"
   },
#ifdef HAVE_VULKAN
   {
      BEETLE_OPT(filter_exclude_sprite),
      BEETLE_OPT_FILTER_EXCLUDE_SPRITE_LABEL_PT_BR,
      NULL,
      BEETLE_OPT_FILTER_EXCLUDE_SPRITE_INFO_0_PT_BR,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "opaque", OPTION_VAL_OPAQUE_PT_BR },
         { "all", OPTION_VAL_ALL_PT_BR },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(filter_exclude_2d_polygon),
      BEETLE_OPT_FILTER_EXCLUDE_2D_POLYGON_LABEL_PT_BR,
      NULL,
      BEETLE_OPT_FILTER_EXCLUDE_2D_POLYGON_INFO_0_PT_BR,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "opaque", OPTION_VAL_OPAQUE_PT_BR },
         { "all", OPTION_VAL_ALL_PT_BR },
         { NULL, NULL },
      },
      "disabled"
   },
#endif
#endif
#ifdef HAVE_VULKAN
   {
      BEETLE_OPT(adaptive_smoothing),
      BEETLE_OPT_ADAPTIVE_SMOOTHING_LABEL_PT_BR,
      NULL,
      BEETLE_OPT_ADAPTIVE_SMOOTHING_INFO_0_PT_BR,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(super_sampling),
      BEETLE_OPT_SUPER_SAMPLING_LABEL_PT_BR,
      NULL,
      BEETLE_OPT_SUPER_SAMPLING_INFO_0_PT_BR,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(msaa),
      BEETLE_OPT_MSAA_LABEL_PT_BR,
      NULL,
      BEETLE_OPT_MSAA_INFO_0_PT_BR,
      NULL,
      "video",
      {
         { "1x",  OPTION_VAL_1X_PT_BR },
         { "2x",  OPTION_VAL_2X_PT_BR },
         { "4x",  OPTION_VAL_4X_PT_BR },
         { "8x",  OPTION_VAL_8X_PT_BR },
         { "16x", OPTION_VAL_16X_PT_BR },
         { NULL, NULL },
      },
      "1x"
   },
   {
      BEETLE_OPT(mdec_yuv),
      BEETLE_OPT_MDEC_YUV_LABEL_PT_BR,
      NULL,
      BEETLE_OPT_MDEC_YUV_INFO_0_PT_BR,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(track_textures),
      BEETLE_OPT_TRACK_TEXTURES_LABEL_PT_BR,
      NULL,
      BEETLE_OPT_TRACK_TEXTURES_INFO_0_PT_BR,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
#ifdef TEXTURE_DUMPING_ENABLED
   {
      BEETLE_OPT(dump_textures),
      BEETLE_OPT_DUMP_TEXTURES_LABEL_PT_BR,
      NULL,
      BEETLE_OPT_DUMP_TEXTURES_INFO_0_PT_BR,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
#endif
   {
      BEETLE_OPT(replace_textures),
      BEETLE_OPT_REPLACE_TEXTURES_LABEL_PT_BR,
      NULL,
      BEETLE_OPT_REPLACE_TEXTURES_INFO_0_PT_BR,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
#endif
   {
      BEETLE_OPT(pgxp_mode),
      BEETLE_OPT_PGXP_MODE_LABEL_PT_BR,
      NULL,
      BEETLE_OPT_PGXP_MODE_INFO_0_PT_BR,
      NULL,
      "pgxp",
      {
         { "disabled",     NULL },
         { "memory only",  OPTION_VAL_MEMORY_ONLY_PT_BR },
         { "memory + CPU", OPTION_VAL_MEMORY_CPU_PT_BR },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(pgxp_2d_tol),
      BEETLE_OPT_PGXP_2D_TOL_LABEL_PT_BR,
      NULL,
      BEETLE_OPT_PGXP_2D_TOL_INFO_0_PT_BR,
      NULL,
      "pgxp",
      {
         { "disabled", NULL },
         { "0px", OPTION_VAL_0PX_PT_BR },
         { "1px", OPTION_VAL_1PX_PT_BR },
         { "2px", OPTION_VAL_2PX_PT_BR },
         { "3px", OPTION_VAL_3PX_PT_BR },
         { "4px", OPTION_VAL_4PX_PT_BR },
         { "5px", OPTION_VAL_5PX_PT_BR },
         { "6px", OPTION_VAL_6PX_PT_BR },
         { "7px", OPTION_VAL_7PX_PT_BR },
         { "8px", OPTION_VAL_8PX_PT_BR },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(pgxp_nclip),
      BEETLE_OPT_PGXP_NCLIP_LABEL_PT_BR,
      NULL,
      BEETLE_OPT_PGXP_NCLIP_INFO_0_PT_BR,
      NULL,
      "pgxp",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
#if defined(HAVE_OPENGL) || defined(HAVE_OPENGLES) || defined(HAVE_VULKAN)
   {
      BEETLE_OPT(pgxp_vertex),
      BEETLE_OPT_PGXP_VERTEX_LABEL_PT_BR,
      NULL,
      BEETLE_OPT_PGXP_VERTEX_INFO_0_PT_BR,
      NULL,
      "pgxp",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
#endif
   {
      BEETLE_OPT(pgxp_texture),
      BEETLE_OPT_PGXP_TEXTURE_LABEL_PT_BR,
      NULL,
      BEETLE_OPT_PGXP_TEXTURE_INFO_0_PT_BR,
      NULL,
      "pgxp",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(display_internal_fps),
      BEETLE_OPT_DISPLAY_INTERNAL_FPS_LABEL_PT_BR,
      NULL,
      BEETLE_OPT_DISPLAY_INTERNAL_FPS_INFO_0_PT_BR,
      NULL,
      "osd",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(display_osd),
      BEETLE_OPT_DISPLAY_OSD_LABEL_PT_BR,
      NULL,
      BEETLE_OPT_DISPLAY_OSD_INFO_0_PT_BR,
      NULL,
      "osd",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled"
   },
#if defined(HAVE_OPENGL) || defined(HAVE_OPENGLES) || defined(HAVE_VULKAN)
   {
      BEETLE_OPT(display_vram),
      BEETLE_OPT_DISPLAY_VRAM_LABEL_PT_BR,
      NULL,
      BEETLE_OPT_DISPLAY_VRAM_INFO_0_PT_BR,
      NULL,
      "osd",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
#endif
   {
      BEETLE_OPT(analog_calibration),
      BEETLE_OPT_ANALOG_CALIBRATION_LABEL_PT_BR,
      NULL,
      BEETLE_OPT_ANALOG_CALIBRATION_INFO_0_PT_BR,
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(analog_toggle),
      BEETLE_OPT_ANALOG_TOGGLE_LABEL_PT_BR,
      NULL,
      BEETLE_OPT_ANALOG_TOGGLE_INFO_0_PT_BR,
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { "enabled-analog", OPTION_VAL_ENABLED_ANALOG_PT_BR },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(analog_toggle_combo),
      BEETLE_OPT_ANALOG_TOGGLE_COMBO_LABEL_PT_BR,
      NULL,
      BEETLE_OPT_ANALOG_TOGGLE_COMBO_INFO_0_PT_BR,
      NULL,
      "input",
      {
         { "l1+l2+r1+r2+start+select", OPTION_VAL_L1_L2_R1_R2_START_SELECT_PT_BR },
         { "l1+r1+select",             OPTION_VAL_L1_R1_SELECT_PT_BR },
         { "l1+r1+start",              OPTION_VAL_L1_R1_START_PT_BR },
         { "l1+r1+l3",                 OPTION_VAL_L1_R1_L3_PT_BR },
         { "l1+r1+r3",                 OPTION_VAL_L1_R1_R3_PT_BR },
         { "l2+r2+select",             OPTION_VAL_L2_R2_SELECT_PT_BR },
         { "l2+r2+start",              OPTION_VAL_L2_R2_START_PT_BR },
         { "l2+r2+l3",                 OPTION_VAL_L2_R2_L3_PT_BR },
         { "l2+r2+r3",                 OPTION_VAL_L2_R2_R3_PT_BR },
         { "l3+r3",                    OPTION_VAL_L3_R3_PT_BR },
         { NULL, NULL },
      },
      "l1+r1+select"
   },
   {
      BEETLE_OPT(analog_toggle_hold),
      BEETLE_OPT_ANALOG_TOGGLE_HOLD_LABEL_PT_BR,
      NULL,
      BEETLE_OPT_ANALOG_TOGGLE_HOLD_INFO_0_PT_BR,
      NULL,
      "input",
      {
         { "0", OPTION_VAL_0_PT_BR },
         { "1", OPTION_VAL_1_PT_BR },
         { "2", OPTION_VAL_2_PT_BR },
         { "3", OPTION_VAL_3_PT_BR },
         { "4", OPTION_VAL_4_PT_BR },
         { "5", OPTION_VAL_5_PT_BR },
         { NULL, NULL },
      },
      "1"
   },
   {
      BEETLE_OPT(enable_multitap_port1),
      BEETLE_OPT_ENABLE_MULTITAP_PORT1_LABEL_PT_BR,
      NULL,
      BEETLE_OPT_ENABLE_MULTITAP_PORT1_INFO_0_PT_BR,
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(enable_multitap_port2),
      BEETLE_OPT_ENABLE_MULTITAP_PORT2_LABEL_PT_BR,
      NULL,
      BEETLE_OPT_ENABLE_MULTITAP_PORT2_INFO_0_PT_BR,
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(gun_input_mode),
      BEETLE_OPT_GUN_INPUT_MODE_LABEL_PT_BR,
      NULL,
      BEETLE_OPT_GUN_INPUT_MODE_INFO_0_PT_BR,
      NULL,
      "input",
      {
         { "lightgun",    OPTION_VAL_LIGHTGUN_PT_BR },
         { "touchscreen", OPTION_VAL_TOUCHSCREEN_PT_BR },
         { NULL, NULL },
      },
      "lightgun"
   },
   // Shouldn't the gun_input_mode just be Mouse vs. Touchscreen?
   {
      BEETLE_OPT(gun_cursor),
      BEETLE_OPT_GUN_CURSOR_LABEL_PT_BR,
      NULL,
      BEETLE_OPT_GUN_CURSOR_INFO_0_PT_BR,
      NULL,
      "input",
      {
         { "off",   "disabled" },
         { "cross", OPTION_VAL_CROSS_PT_BR },
         { "dot",   OPTION_VAL_DOT_PT_BR },
         { NULL, NULL },
      },
      "cross"
   },
   {
      BEETLE_OPT(crosshair_color_p1),
      BEETLE_OPT_CROSSHAIR_COLOR_P1_LABEL_PT_BR,
      NULL,
      BEETLE_OPT_CROSSHAIR_COLOR_P1_INFO_0_PT_BR,
      NULL,
      "input",
      {
         { "red",    OPTION_VAL_RED_PT_BR },
         { "blue",   OPTION_VAL_BLUE_PT_BR },
         { "green",  OPTION_VAL_GREEN_PT_BR },
         { "orange", OPTION_VAL_ORANGE_PT_BR },
         { "yellow", OPTION_VAL_YELLOW_PT_BR },
         { "cyan",   OPTION_VAL_CYAN_PT_BR },
         { "pink",   OPTION_VAL_PINK_PT_BR },
         { "purple", OPTION_VAL_PURPLE_PT_BR },
         { "black",  OPTION_VAL_BLACK_PT_BR },
         { "white",  OPTION_VAL_WHITE_PT_BR },
         { NULL, NULL },
      },
      "red"
   },
   {
      BEETLE_OPT(crosshair_color_p2),
      BEETLE_OPT_CROSSHAIR_COLOR_P2_LABEL_PT_BR,
      NULL,
      BEETLE_OPT_CROSSHAIR_COLOR_P2_INFO_0_PT_BR,
      NULL,
      "input",
      {
         { "blue",   OPTION_VAL_BLUE_PT_BR },
         { "red",    OPTION_VAL_RED_PT_BR },
         { "green",  OPTION_VAL_GREEN_PT_BR },
         { "orange", OPTION_VAL_ORANGE_PT_BR },
         { "yellow", OPTION_VAL_YELLOW_PT_BR },
         { "cyan",   OPTION_VAL_CYAN_PT_BR },
         { "pink",   OPTION_VAL_PINK_PT_BR },
         { "purple", OPTION_VAL_PURPLE_PT_BR },
         { "black",  OPTION_VAL_BLACK_PT_BR },
         { "white",  OPTION_VAL_WHITE_PT_BR },
         { NULL, NULL },
      },
      "blue"
   },
   {
      BEETLE_OPT(mouse_sensitivity),
      BEETLE_OPT_MOUSE_SENSITIVITY_LABEL_PT_BR,
      NULL,
      BEETLE_OPT_MOUSE_SENSITIVITY_INFO_0_PT_BR,
      NULL,
      "input",
      {
         { "5%",   OPTION_VAL_5_O42_PT_BR },
         { "10%",  OPTION_VAL_10_PT_BR },
         { "15%",  OPTION_VAL_15_PT_BR },
         { "20%",  OPTION_VAL_20_PT_BR },
         { "25%",  OPTION_VAL_25_PT_BR },
         { "30%",  OPTION_VAL_30_PT_BR },
         { "35%",  OPTION_VAL_35_PT_BR },
         { "40%",  OPTION_VAL_40_PT_BR },
         { "45%",  OPTION_VAL_45_PT_BR },
         { "50%",  OPTION_VAL_50_PT_BR },
         { "55%",  OPTION_VAL_55_PT_BR },
         { "60%",  OPTION_VAL_60_PT_BR },
         { "65%",  OPTION_VAL_65_PT_BR },
         { "70%",  OPTION_VAL_70_PT_BR },
         { "75%",  OPTION_VAL_75_PT_BR },
         { "80%",  OPTION_VAL_80_PT_BR },
         { "85%",  OPTION_VAL_85_PT_BR },
         { "90%",  OPTION_VAL_90_PT_BR },
         { "95%",  OPTION_VAL_95_PT_BR },
         { "100%", OPTION_VAL_100_O42_PT_BR },
         { "105%", OPTION_VAL_105_PT_BR },
         { "110%", OPTION_VAL_110_PT_BR },
         { "115%", OPTION_VAL_115_PT_BR },
         { "120%", OPTION_VAL_120_PT_BR },
         { "125%", OPTION_VAL_125_PT_BR },
         { "130%", OPTION_VAL_130_PT_BR },
         { "135%", OPTION_VAL_135_PT_BR },
         { "140%", OPTION_VAL_140_PT_BR },
         { "145%", OPTION_VAL_145_PT_BR },
         { "150%", OPTION_VAL_150_PT_BR },
         { "155%", OPTION_VAL_155_PT_BR },
         { "160%", OPTION_VAL_160_PT_BR },
         { "165%", OPTION_VAL_165_PT_BR },
         { "170%", OPTION_VAL_170_PT_BR },
         { "175%", OPTION_VAL_175_PT_BR },
         { "180%", OPTION_VAL_180_PT_BR },
         { "185%", OPTION_VAL_185_PT_BR },
         { "190%", OPTION_VAL_190_PT_BR },
         { "195%", OPTION_VAL_195_PT_BR },
         { "200%", OPTION_VAL_200_PT_BR },
         { NULL, NULL },
      },
      "100%"
   },
   {
      BEETLE_OPT(negcon_response),
      BEETLE_OPT_NEGCON_RESPONSE_LABEL_PT_BR,
      NULL,
      BEETLE_OPT_NEGCON_RESPONSE_INFO_0_PT_BR,
      NULL,
      "input",
      {
         { "linear",    OPTION_VAL_LINEAR_PT_BR },
         { "quadratic", OPTION_VAL_QUADRATIC_PT_BR },
         { "cubic",     OPTION_VAL_CUBIC_PT_BR },
         { NULL, NULL },
      },
      "linear"
   },
   {
      BEETLE_OPT(negcon_deadzone),
      BEETLE_OPT_NEGCON_DEADZONE_LABEL_PT_BR,
      NULL,
      BEETLE_OPT_NEGCON_DEADZONE_INFO_0_PT_BR,
      NULL,
      "input",
      {
         { "0%",  OPTION_VAL_0_O44_PT_BR },
         { "5%",  OPTION_VAL_5_O42_PT_BR },
         { "10%", OPTION_VAL_10_PT_BR },
         { "15%", OPTION_VAL_15_PT_BR },
         { "20%", OPTION_VAL_20_PT_BR },
         { "25%", OPTION_VAL_25_PT_BR },
         { "30%", OPTION_VAL_30_PT_BR },
         { NULL, NULL },
      },
      "0%"
   },
   {
      BEETLE_OPT(use_mednafen_memcard0_method),
      BEETLE_OPT_USE_MEDNAFEN_MEMCARD0_METHOD_LABEL_PT_BR,
      NULL,
      BEETLE_OPT_USE_MEDNAFEN_MEMCARD0_METHOD_INFO_0_PT_BR,
      NULL,
      "memcards",
      {
         { "libretro", OPTION_VAL_LIBRETRO_PT_BR },
         { "mednafen", OPTION_VAL_MEDNAFEN_PT_BR },
         { NULL, NULL },
      },
      "libretro"
   },
   {
      BEETLE_OPT(enable_memcard1),
      BEETLE_OPT_ENABLE_MEMCARD1_LABEL_PT_BR,
      NULL,
      BEETLE_OPT_ENABLE_MEMCARD1_INFO_0_PT_BR,
      NULL,
      "memcards",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(shared_memory_cards),
      BEETLE_OPT_SHARED_MEMORY_CARDS_LABEL_PT_BR,
      NULL,
      BEETLE_OPT_SHARED_MEMORY_CARDS_INFO_0_PT_BR,
      NULL,
      "memcards",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(memcard_left_index),
      BEETLE_OPT_MEMCARD_LEFT_INDEX_LABEL_PT_BR,
      NULL,
      BEETLE_OPT_MEMCARD_LEFT_INDEX_INFO_0_PT_BR,
      NULL,
      "memcards",
      {
         { "0",  OPTION_VAL_0_O48_PT_BR },
         { "1",  NULL },
         { "2",  NULL },
         { "3",  NULL },
         { "4",  NULL },
         { "5",  NULL },
         { "6",  NULL },
         { "7",  NULL },
         { "8",  NULL },
         { "9",  NULL },
         { "10", NULL },
         { "11", NULL },
         { "12", NULL },
         { "13", NULL },
         { "14", NULL },
         { "15", NULL },
         { "16", NULL },
         { "17", NULL },
         { "18", NULL },
         { "19", NULL },
         { "20", NULL },
         { "21", NULL },
         { "22", NULL },
         { "23", NULL },
         { "24", NULL },
         { "25", NULL },
         { "26", NULL },
         { "27", NULL },
         { "28", NULL },
         { "29", NULL },
         { "30", NULL },
         { "31", NULL },
         { "32", NULL },
         { "33", NULL },
         { "34", NULL },
         { "35", NULL },
         { "36", NULL },
         { "37", NULL },
         { "38", NULL },
         { "39", NULL },
         { "40", NULL },
         { "41", NULL },
         { "42", NULL },
         { "43", NULL },
         { "44", NULL },
         { "45", NULL },
         { "46", NULL },
         { "47", NULL },
         { "48", NULL },
         { "49", NULL },
         { "50", NULL },
         { "51", NULL },
         { "52", NULL },
         { "53", NULL },
         { "54", NULL },
         { "55", NULL },
         { "56", NULL },
         { "57", NULL },
         { "58", NULL },
         { "59", NULL },
         { "60", NULL },
         { "61", NULL },
         { "62", NULL },
         { "63", NULL },
         { NULL, NULL },
      },
      "0"
   },
   {
      BEETLE_OPT(memcard_right_index),
      BEETLE_OPT_MEMCARD_RIGHT_INDEX_LABEL_PT_BR,
      NULL,
      BEETLE_OPT_MEMCARD_RIGHT_INDEX_INFO_0_PT_BR,
      NULL,
      "memcards",
      {
         { "0",  NULL },
         { "1",  OPTION_VAL_1_O49_PT_BR },
         { "2",  NULL },
         { "3",  NULL },
         { "4",  NULL },
         { "5",  NULL },
         { "6",  NULL },
         { "7",  NULL },
         { "8",  NULL },
         { "9",  NULL },
         { "10", NULL },
         { "11", NULL },
         { "12", NULL },
         { "13", NULL },
         { "14", NULL },
         { "15", NULL },
         { "16", NULL },
         { "17", NULL },
         { "18", NULL },
         { "19", NULL },
         { "20", NULL },
         { "21", NULL },
         { "22", NULL },
         { "23", NULL },
         { "24", NULL },
         { "25", NULL },
         { "26", NULL },
         { "27", NULL },
         { "28", NULL },
         { "29", NULL },
         { "30", NULL },
         { "31", NULL },
         { "32", NULL },
         { "33", NULL },
         { "34", NULL },
         { "35", NULL },
         { "36", NULL },
         { "37", NULL },
         { "38", NULL },
         { "39", NULL },
         { "40", NULL },
         { "41", NULL },
         { "42", NULL },
         { "43", NULL },
         { "44", NULL },
         { "45", NULL },
         { "46", NULL },
         { "47", NULL },
         { "48", NULL },
         { "49", NULL },
         { "50", NULL },
         { "51", NULL },
         { "52", NULL },
         { "53", NULL },
         { "54", NULL },
         { "55", NULL },
         { "56", NULL },
         { "57", NULL },
         { "58", NULL },
         { "59", NULL },
         { "60", NULL },
         { "61", NULL },
         { "62", NULL },
         { "63", NULL },
         { NULL, NULL },
      },
      "1"
   },
   {
      BEETLE_OPT(line_render),
      BEETLE_OPT_LINE_RENDER_LABEL_PT_BR,
      NULL,
      BEETLE_OPT_LINE_RENDER_INFO_0_PT_BR,
      NULL,
      "hacks",
      {
         { "default",    OPTION_VAL_DEFAULT_PT_BR },
         { "aggressive", OPTION_VAL_AGGRESSIVE_PT_BR },
         { "disabled",   NULL },
         { NULL, NULL },
      },
      "default"
   },
   {
      BEETLE_OPT(widescreen_hack),
      BEETLE_OPT_WIDESCREEN_HACK_LABEL_PT_BR,
      NULL,
      BEETLE_OPT_WIDESCREEN_HACK_INFO_0_PT_BR,
      NULL,
      "hacks",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(widescreen_hack_aspect_ratio),
      BEETLE_OPT_WIDESCREEN_HACK_ASPECT_RATIO_LABEL_PT_BR,
      NULL,
      BEETLE_OPT_WIDESCREEN_HACK_ASPECT_RATIO_INFO_0_PT_BR,
      NULL,
      "hacks",
      {
         { "16:9",  OPTION_VAL_16_9_PT_BR },
         { "16:10", OPTION_VAL_16_10_PT_BR },
         { "18:9",  OPTION_VAL_18_9_PT_BR },
         { "19:9",  OPTION_VAL_19_9_PT_BR },
         { "20:9",  OPTION_VAL_20_9_PT_BR },
         { "21:9",  OPTION_VAL_21_9_PT_BR }, // 64:27
         { "32:9",  OPTION_VAL_32_9_PT_BR },
         { NULL,    NULL },
      },
      "16:9"
   },
#if defined(HAVE_LIGHTREC)
   {
      BEETLE_OPT(cpu_dynarec),
      BEETLE_OPT_CPU_DYNAREC_LABEL_PT_BR,
      NULL,
      BEETLE_OPT_CPU_DYNAREC_INFO_0_PT_BR,
      NULL,
      "hacks",
      {
         { "disabled",        OPTION_VAL_DISABLED_PT_BR },
         { "execute",         OPTION_VAL_EXECUTE_PT_BR },
         { "execute_one",     OPTION_VAL_EXECUTE_ONE_PT_BR },
         { "run_interpreter", OPTION_VAL_RUN_INTERPRETER_PT_BR },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(dynarec_invalidate),
      BEETLE_OPT_DYNAREC_INVALIDATE_LABEL_PT_BR,
      NULL,
      BEETLE_OPT_DYNAREC_INVALIDATE_INFO_0_PT_BR,
      NULL,
      "hacks",
      {
         { "full", OPTION_VAL_FULL_PT_BR },
         { "dma",  OPTION_VAL_DMA_PT_BR },
         { NULL, NULL },
      },
      "full"
   },
   {
      BEETLE_OPT(dynarec_eventcycles),
      BEETLE_OPT_DYNAREC_EVENTCYCLES_LABEL_PT_BR,
      NULL,
      BEETLE_OPT_DYNAREC_EVENTCYCLES_INFO_0_PT_BR,
      NULL,
      "hacks",
      {
         { "128",  OPTION_VAL_128_PT_BR },
         { "256",  NULL },
         { "384",  NULL },
         { "512",  NULL },
         { "640",  NULL },
         { "768",  NULL },
         { "896",  NULL },
         { "1024", NULL },
         { "1152", NULL },
         { "1280", NULL },
         { "1408", NULL },
         { "1536", NULL },
         { "1664", NULL },
         { "1792", NULL },
         { "1920", NULL },
         { "2048", NULL },
         { NULL, NULL },
      },
      "128"
   },
   {
      BEETLE_OPT(dynarec_spu_samples),
      BEETLE_OPT_DYNAREC_SPU_SAMPLES_LABEL_PT_BR,
      NULL,
      BEETLE_OPT_DYNAREC_SPU_SAMPLES_INFO_0_PT_BR,
      NULL,
      "hacks",
      {
         { "1",  OPTION_VAL_1_O49_PT_BR },
         { "4",  NULL },
         { "16", NULL },
         { NULL, NULL },
      },
      "1"
   },
#endif
   {
      BEETLE_OPT(core_timing_fps),
      BEETLE_OPT_CORE_TIMING_FPS_LABEL_PT_BR,
      NULL,
      BEETLE_OPT_CORE_TIMING_FPS_INFO_0_PT_BR,
      NULL,
      "video",
      {
         { "auto_toggle",       OPTION_VAL_AUTO_TOGGLE_PT_BR },
         { "force_progressive", OPTION_VAL_FORCE_PROGRESSIVE_PT_BR },
         { "force_interlaced",  OPTION_VAL_FORCE_INTERLACED_PT_BR },
      },
      "auto_toggle"
   },
   {
      BEETLE_OPT(deinterlacer),
      BEETLE_OPT_DEINTERLACER_LABEL_PT_BR,
      NULL,
      BEETLE_OPT_DEINTERLACER_INFO_0_PT_BR,
      NULL,
      "video",
      {
         { "weave",      OPTION_VAL_WEAVE_PT_BR },
         { "bob",        OPTION_VAL_BOB_PT_BR },
         { "bob_offset", OPTION_VAL_BOB_OFFSET_PT_BR },
         { "fastmad",    OPTION_VAL_FASTMAD_PT_BR },
         { "off",        "Off" },
         { NULL, NULL },
      },
      "weave"
   },
   {
      BEETLE_OPT(frame_duping),
      BEETLE_OPT_FRAME_DUPING_LABEL_PT_BR,
      NULL,
      BEETLE_OPT_FRAME_DUPING_INFO_0_PT_BR,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(pal_video_timing_override),
      BEETLE_OPT_PAL_VIDEO_TIMING_OVERRIDE_LABEL_PT_BR,
      NULL,
      BEETLE_OPT_PAL_VIDEO_TIMING_OVERRIDE_INFO_0_PT_BR,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(aspect_ratio),
      BEETLE_OPT_ASPECT_RATIO_LABEL_PT_BR,
      NULL,
      BEETLE_OPT_ASPECT_RATIO_INFO_0_PT_BR,
      NULL,
      "video",
      {
         { "corrected", OPTION_VAL_CORRECTED_PT_BR },
         { "uncorrected", OPTION_VAL_UNCORRECTED_PT_BR },
         { "4:3",  OPTION_VAL_4_3_PT_BR },
         { "ntsc", OPTION_VAL_NTSC_PT_BR },
      },
      "corrected"
   },
   {
      BEETLE_OPT(crop_overscan),
      BEETLE_OPT_CROP_OVERSCAN_LABEL_PT_BR,
      NULL,
      BEETLE_OPT_CROP_OVERSCAN_INFO_0_PT_BR,
      NULL,
      "video",
      {
         { "disabled", "Disabled" },
         { "static",   OPTION_VAL_STATIC_PT_BR },
         { "smart",    OPTION_VAL_SMART_PT_BR },
         { NULL, NULL },
      },
      "smart"
   },
   {
      BEETLE_OPT(image_crop),
      BEETLE_OPT_IMAGE_CROP_LABEL_PT_BR,
      NULL,
      BEETLE_OPT_IMAGE_CROP_INFO_0_PT_BR,
      NULL,
      "video",
      {
         { "disabled", "0" },
         { "1px",      OPTION_VAL_1PX_PT_BR },
         { "2px",      OPTION_VAL_2PX_PT_BR },
         { "3px",      OPTION_VAL_3PX_PT_BR },
         { "4px",      OPTION_VAL_4PX_PT_BR },
         { "5px",      OPTION_VAL_5PX_PT_BR },
         { "6px",      OPTION_VAL_6PX_PT_BR },
         { "7px",      OPTION_VAL_7PX_PT_BR },
         { "8px",      OPTION_VAL_8PX_PT_BR },
         { "9px",      OPTION_VAL_9PX_PT_BR },
         { "10px",     OPTION_VAL_10PX_PT_BR },
         { "11px",     OPTION_VAL_11PX_PT_BR },
         { "12px",     OPTION_VAL_12PX_PT_BR },
         { "13px",     OPTION_VAL_13PX_PT_BR },
         { "14px",     OPTION_VAL_14PX_PT_BR },
         { "15px",     OPTION_VAL_15PX_PT_BR },
         { "16px",     OPTION_VAL_16PX_PT_BR },
         { "17px",     OPTION_VAL_17PX_PT_BR },
         { "18px",     OPTION_VAL_18PX_PT_BR },
         { "19px",     OPTION_VAL_19PX_PT_BR },
         { "20px",     OPTION_VAL_20PX_PT_BR },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(image_offset),
      BEETLE_OPT_IMAGE_OFFSET_LABEL_PT_BR,
      NULL,
      BEETLE_OPT_IMAGE_OFFSET_INFO_0_PT_BR,
      NULL,
      "video",
      {
         { "-12px",    OPTION_VAL_12PX_O64_PT_BR },
         { "-11px",    OPTION_VAL_11PX_O64_PT_BR },
         { "-10px",    OPTION_VAL_10PX_O64_PT_BR },
         { "-9px",     OPTION_VAL_9PX_O64_PT_BR },
         { "-8px",     OPTION_VAL_8PX_O64_PT_BR },
         { "-7px",     OPTION_VAL_7PX_O64_PT_BR },
         { "-6px",     OPTION_VAL_6PX_O64_PT_BR },
         { "-5px",     OPTION_VAL_5PX_O64_PT_BR },
         { "-4px",     OPTION_VAL_4PX_O64_PT_BR },
         { "-3px",     OPTION_VAL_3PX_O64_PT_BR },
         { "-2px",     OPTION_VAL_2PX_O64_PT_BR },
         { "-1px",     OPTION_VAL_1PX_O64_PT_BR },
         { "disabled", "0" },
         { "+1px",     OPTION_VAL_1PX_O64_0_PT_BR },
         { "+2px",     OPTION_VAL_2PX_O64_0_PT_BR },
         { "+3px",     OPTION_VAL_3PX_O64_0_PT_BR },
         { "+4px",     OPTION_VAL_4PX_O64_0_PT_BR },
         { "+5px",     OPTION_VAL_5PX_O64_0_PT_BR },
         { "+6px",     OPTION_VAL_6PX_O64_0_PT_BR },
         { "+7px",     OPTION_VAL_7PX_O64_0_PT_BR },
         { "+8px",     OPTION_VAL_8PX_O64_0_PT_BR },
         { "+9px",     OPTION_VAL_9PX_O64_0_PT_BR },
         { "+10px",    OPTION_VAL_10PX_O64_0_PT_BR },
         { "+11px",    OPTION_VAL_11PX_O64_0_PT_BR },
         { "+12px",    OPTION_VAL_12PX_O64_0_PT_BR },
         { NULL, NULL },
      },
      "disabled"
   },
#if defined(HAVE_OPENGL) || defined(HAVE_OPENGLES) || defined(HAVE_VULKAN)
   {
      BEETLE_OPT(image_offset_cycles),
      BEETLE_OPT_IMAGE_OFFSET_CYCLES_LABEL_PT_BR,
      NULL,
      BEETLE_OPT_IMAGE_OFFSET_CYCLES_INFO_0_PT_BR,
      NULL,
      "video",
      {
         { "-40",      NULL },
         { "-39",      NULL },
         { "-38",      NULL },
         { "-37",      NULL },
         { "-36",      NULL },
         { "-35",      NULL },
         { "-34",      NULL },
         { "-33",      NULL },
         { "-32",      NULL },
         { "-31",      NULL },
         { "-30",      NULL },
         { "-29",      NULL },
         { "-28",      NULL },
         { "-27",      NULL },
         { "-26",      NULL },
         { "-25",      NULL },
         { "-24",      NULL },
         { "-23",      NULL },
         { "-22",      NULL },
         { "-21",      NULL },
         { "-20",      NULL },
         { "-19",      NULL },
         { "-18",      NULL },
         { "-17",      NULL },
         { "-16",      NULL },
         { "-15",      NULL },
         { "-14",      NULL },
         { "-13",      NULL },
         { "-12",      NULL },
         { "-11",      NULL },
         { "-10",      NULL },
         { "-9",       NULL },
         { "-8",       NULL },
         { "-7",       NULL },
         { "-6",       NULL },
         { "-5",       NULL },
         { "-4",       NULL },
         { "-3",       NULL },
         { "-2",       NULL },
         { "-1",       NULL },
         { "0",        "0" },
         { "+1",       NULL },
         { "+2",       NULL },
         { "+3",       NULL },
         { "+4",       NULL },
         { "+5",       NULL },
         { "+6",       NULL },
         { "+7",       NULL },
         { "+8",       NULL },
         { "+9",       NULL },
         { "+10",      NULL },
         { "+11",      NULL },
         { "+12",      NULL },
         { "+13",      NULL },
         { "+14",      NULL },
         { "+15",      NULL },
         { "+16",      NULL },
         { "+17",      NULL },
         { "+18",      NULL },
         { "+19",      NULL },
         { "+20",      NULL },
         { "+21",      NULL },
         { "+22",      NULL },
         { "+23",      NULL },
         { "+24",      NULL },
         { "+25",      NULL },
         { "+26",      NULL },
         { "+27",      NULL },
         { "+28",      NULL },
         { "+29",      NULL },
         { "+30",      NULL },
         { "+31",      NULL },
         { "+32",      NULL },
         { "+33",      NULL },
         { "+34",      NULL },
         { "+35",      NULL },
         { "+36",      NULL },
         { "+37",      NULL },
         { "+38",      NULL },
         { "+39",      NULL },
         { "+40",      NULL },
         { NULL, NULL},
      },
      "0"
   },
#endif
   {
      BEETLE_OPT(initial_scanline),
      BEETLE_OPT_INITIAL_SCANLINE_LABEL_PT_BR,
      NULL,
      BEETLE_OPT_INITIAL_SCANLINE_INFO_0_PT_BR,
      NULL,
      "video",
      {
         { "0",  OPTION_VAL_0_O48_PT_BR },
         { "1",  NULL },
         { "2",  NULL },
         { "3",  NULL },
         { "4",  NULL },
         { "5",  NULL },
         { "6",  NULL },
         { "7",  NULL },
         { "8",  NULL },
         { "9",  NULL },
         { "10", NULL },
         { "11", NULL },
         { "12", NULL },
         { "13", NULL },
         { "14", NULL },
         { "15", NULL },
         { "16", NULL },
         { "17", NULL },
         { "18", NULL },
         { "19", NULL },
         { "20", NULL },
         { "21", NULL },
         { "22", NULL },
         { "23", NULL },
         { "24", NULL },
         { "25", NULL },
         { "26", NULL },
         { "27", NULL },
         { "28", NULL },
         { "29", NULL },
         { "30", NULL },
         { "31", NULL },
         { "32", NULL },
         { "33", NULL },
         { "34", NULL },
         { "35", NULL },
         { "36", NULL },
         { "37", NULL },
         { "38", NULL },
         { "39", NULL },
         { "40", NULL },
         { NULL, NULL },
      },
      "0"
   },
   {
      BEETLE_OPT(last_scanline),
      BEETLE_OPT_LAST_SCANLINE_LABEL_PT_BR,
      NULL,
      BEETLE_OPT_LAST_SCANLINE_INFO_0_PT_BR,
      NULL,
      "video",
      {
         { "210", NULL },
         { "211", NULL },
         { "212", NULL },
         { "213", NULL },
         { "214", NULL },
         { "215", NULL },
         { "216", NULL },
         { "217", NULL },
         { "218", NULL },
         { "219", NULL },
         { "220", NULL },
         { "221", NULL },
         { "222", NULL },
         { "223", NULL },
         { "224", NULL },
         { "225", NULL },
         { "226", NULL },
         { "227", NULL },
         { "228", NULL },
         { "229", NULL },
         { "230", NULL },
         { "231", NULL },
         { "232", NULL },
         { "233", NULL },
         { "234", NULL },
         { "235", NULL },
         { "236", NULL },
         { "237", NULL },
         { "238", NULL },
         { "239", OPTION_VAL_239_PT_BR },
         { NULL, NULL },
      },
      "239"
   },
   {
      BEETLE_OPT(initial_scanline_pal),
      BEETLE_OPT_INITIAL_SCANLINE_PAL_LABEL_PT_BR,
      NULL,
      BEETLE_OPT_INITIAL_SCANLINE_PAL_INFO_0_PT_BR,
      NULL,
      "video",
      {
         { "0",  OPTION_VAL_0_O48_PT_BR },
         { "1",  NULL },
         { "2",  NULL },
         { "3",  NULL },
         { "4",  NULL },
         { "5",  NULL },
         { "6",  NULL },
         { "7",  NULL },
         { "8",  NULL },
         { "9",  NULL },
         { "10", NULL },
         { "11", NULL },
         { "12", NULL },
         { "13", NULL },
         { "14", NULL },
         { "15", NULL },
         { "16", NULL },
         { "17", NULL },
         { "18", NULL },
         { "19", NULL },
         { "20", NULL },
         { "21", NULL },
         { "22", NULL },
         { "23", NULL },
         { "24", NULL },
         { "25", NULL },
         { "26", NULL },
         { "27", NULL },
         { "28", NULL },
         { "29", NULL },
         { "30", NULL },
         { "31", NULL },
         { "32", NULL },
         { "33", NULL },
         { "34", NULL },
         { "35", NULL },
         { "36", NULL },
         { "37", NULL },
         { "38", NULL },
         { "39", NULL },
         { "40", NULL },
         { NULL, NULL },
      },
      "0"
   },
   {
      BEETLE_OPT(last_scanline_pal),
      BEETLE_OPT_LAST_SCANLINE_PAL_LABEL_PT_BR,
      NULL,
      BEETLE_OPT_LAST_SCANLINE_PAL_INFO_0_PT_BR,
      NULL,
      "video",
      {
         { "230", NULL },
         { "231", NULL },
         { "232", NULL },
         { "233", NULL },
         { "234", NULL },
         { "235", NULL },
         { "236", NULL },
         { "237", NULL },
         { "238", NULL },
         { "239", NULL },
         { "240", NULL },
         { "241", NULL },
         { "242", NULL },
         { "243", NULL },
         { "244", NULL },
         { "245", NULL },
         { "246", NULL },
         { "247", NULL },
         { "248", NULL },
         { "249", NULL },
         { "250", NULL },
         { "251", NULL },
         { "252", NULL },
         { "253", NULL },
         { "254", NULL },
         { "255", NULL },
         { "256", NULL },
         { "257", NULL },
         { "258", NULL },
         { "259", NULL },
         { "260", NULL },
         { "261", NULL },
         { "262", NULL },
         { "263", NULL },
         { "264", NULL },
         { "265", NULL },
         { "266", NULL },
         { "267", NULL },
         { "268", NULL },
         { "269", NULL },
         { "270", NULL },
         { "271", NULL },
         { "272", NULL },
         { "273", NULL },
         { "274", NULL },
         { "275", NULL },
         { "276", NULL },
         { "277", NULL },
         { "278", NULL },
         { "279", NULL },
         { "280", NULL },
         { "281", NULL },
         { "282", NULL },
         { "283", NULL },
         { "284", NULL },
         { "285", NULL },
         { "286", NULL },
         { "287", OPTION_VAL_287_PT_BR },
         { NULL, NULL },
      },
      "287"
   },
   { NULL, NULL, NULL, NULL, NULL, NULL, {{0}}, NULL },
};
struct retro_core_options_v2 options_pt_br = {
   option_cats_pt_br,
   option_defs_pt_br
};

/* RETRO_LANGUAGE_PT_PT */

#define CATEGORY_SYSTEM_LABEL_PT_PT NULL
#define CATEGORY_SYSTEM_INFO_0_PT_PT NULL
#define CATEGORY_VIDEO_LABEL_PT_PT "Vídeo"
#define CATEGORY_VIDEO_INFO_0_PT_PT NULL
#define CATEGORY_PGXP_LABEL_PT_PT NULL
#define CATEGORY_PGXP_INFO_0_PT_PT NULL
#define CATEGORY_OSD_LABEL_PT_PT "Apresentação no ecrã"
#define CATEGORY_OSD_INFO_0_PT_PT NULL
#define CATEGORY_INPUT_LABEL_PT_PT "Entrada"
#define CATEGORY_INPUT_INFO_0_PT_PT NULL
#define CATEGORY_MEMCARDS_LABEL_PT_PT NULL
#define CATEGORY_MEMCARDS_INFO_0_PT_PT NULL
#define CATEGORY_HACKS_LABEL_PT_PT NULL
#define CATEGORY_HACKS_INFO_0_PT_PT NULL
#define BEETLE_OPT_CPU_FREQ_SCALE_LABEL_PT_PT NULL
#define BEETLE_OPT_CPU_FREQ_SCALE_INFO_0_PT_PT NULL
#define OPTION_VAL_50_PT_PT NULL
#define OPTION_VAL_60_PT_PT NULL
#define OPTION_VAL_70_PT_PT NULL
#define OPTION_VAL_80_PT_PT NULL
#define OPTION_VAL_90_PT_PT NULL
#define OPTION_VAL_100_PT_PT NULL
#define OPTION_VAL_110_PT_PT NULL
#define OPTION_VAL_120_PT_PT NULL
#define OPTION_VAL_130_PT_PT NULL
#define OPTION_VAL_140_PT_PT NULL
#define OPTION_VAL_150_PT_PT NULL
#define OPTION_VAL_160_PT_PT NULL
#define OPTION_VAL_170_PT_PT NULL
#define OPTION_VAL_180_PT_PT NULL
#define OPTION_VAL_190_PT_PT NULL
#define OPTION_VAL_200_PT_PT NULL
#define OPTION_VAL_210_PT_PT NULL
#define OPTION_VAL_220_PT_PT NULL
#define OPTION_VAL_230_PT_PT NULL
#define OPTION_VAL_240_PT_PT NULL
#define OPTION_VAL_250_PT_PT NULL
#define OPTION_VAL_260_PT_PT NULL
#define OPTION_VAL_270_PT_PT NULL
#define OPTION_VAL_280_PT_PT NULL
#define OPTION_VAL_290_PT_PT NULL
#define OPTION_VAL_300_PT_PT NULL
#define OPTION_VAL_310_PT_PT NULL
#define OPTION_VAL_320_PT_PT NULL
#define OPTION_VAL_330_PT_PT NULL
#define OPTION_VAL_340_PT_PT NULL
#define OPTION_VAL_350_PT_PT NULL
#define OPTION_VAL_360_PT_PT NULL
#define OPTION_VAL_370_PT_PT NULL
#define OPTION_VAL_380_PT_PT NULL
#define OPTION_VAL_390_PT_PT NULL
#define OPTION_VAL_400_PT_PT NULL
#define OPTION_VAL_410_PT_PT NULL
#define OPTION_VAL_420_PT_PT NULL
#define OPTION_VAL_430_PT_PT NULL
#define OPTION_VAL_440_PT_PT NULL
#define OPTION_VAL_450_PT_PT NULL
#define OPTION_VAL_460_PT_PT NULL
#define OPTION_VAL_470_PT_PT NULL
#define OPTION_VAL_480_PT_PT NULL
#define OPTION_VAL_490_PT_PT NULL
#define OPTION_VAL_500_PT_PT NULL
#define OPTION_VAL_510_PT_PT NULL
#define OPTION_VAL_520_PT_PT NULL
#define OPTION_VAL_530_PT_PT NULL
#define OPTION_VAL_540_PT_PT NULL
#define OPTION_VAL_550_PT_PT NULL
#define OPTION_VAL_560_PT_PT NULL
#define OPTION_VAL_570_PT_PT NULL
#define OPTION_VAL_580_PT_PT NULL
#define OPTION_VAL_590_PT_PT NULL
#define OPTION_VAL_600_PT_PT NULL
#define OPTION_VAL_610_PT_PT NULL
#define OPTION_VAL_620_PT_PT NULL
#define OPTION_VAL_630_PT_PT NULL
#define OPTION_VAL_640_PT_PT NULL
#define OPTION_VAL_650_PT_PT NULL
#define OPTION_VAL_660_PT_PT NULL
#define OPTION_VAL_670_PT_PT NULL
#define OPTION_VAL_680_PT_PT NULL
#define OPTION_VAL_690_PT_PT NULL
#define OPTION_VAL_700_PT_PT NULL
#define OPTION_VAL_710_PT_PT NULL
#define OPTION_VAL_720_PT_PT NULL
#define OPTION_VAL_730_PT_PT NULL
#define OPTION_VAL_740_PT_PT NULL
#define OPTION_VAL_750_PT_PT NULL
#define BEETLE_OPT_GPU_OVERCLOCK_LABEL_PT_PT NULL
#define BEETLE_OPT_GPU_OVERCLOCK_INFO_0_PT_PT NULL
#define OPTION_VAL_1X_NATIVE_PT_PT NULL
#define OPTION_VAL_2X_PT_PT NULL
#define OPTION_VAL_4X_PT_PT NULL
#define OPTION_VAL_8X_PT_PT NULL
#define OPTION_VAL_16X_PT_PT NULL
#define OPTION_VAL_32X_PT_PT NULL
#define BEETLE_OPT_GTE_OVERCLOCK_LABEL_PT_PT NULL
#define BEETLE_OPT_GTE_OVERCLOCK_INFO_0_PT_PT NULL
#define BEETLE_OPT_CD_ACCESS_METHOD_LABEL_PT_PT NULL
#define BEETLE_OPT_CD_ACCESS_METHOD_INFO_0_PT_PT NULL
#define OPTION_VAL_SYNC_PT_PT NULL
#define OPTION_VAL_ASYNC_PT_PT NULL
#define OPTION_VAL_PRECACHE_PT_PT NULL
#define BEETLE_OPT_CD_FASTLOAD_LABEL_PT_PT NULL
#define BEETLE_OPT_CD_FASTLOAD_INFO_0_PT_PT NULL
#define OPTION_VAL_2X_NATIVE_PT_PT NULL
#define OPTION_VAL_6X_PT_PT NULL
#define OPTION_VAL_10X_PT_PT NULL
#define OPTION_VAL_12X_PT_PT NULL
#define OPTION_VAL_14X_PT_PT NULL
#define BEETLE_OPT_SKIP_BIOS_LABEL_PT_PT NULL
#define BEETLE_OPT_SKIP_BIOS_INFO_0_PT_PT NULL
#define BEETLE_OPT_OVERRIDE_BIOS_LABEL_PT_PT NULL
#define BEETLE_OPT_OVERRIDE_BIOS_INFO_0_PT_PT NULL
#define OPTION_VAL_PSXONPSP_PT_PT NULL
#define OPTION_VAL_PS1_ROM_PT_PT NULL
#define OPTION_VAL_OPENBIOS_PT_PT NULL
#define BEETLE_OPT_REGION_LABEL_PT_PT NULL
#define BEETLE_OPT_REGION_INFO_0_PT_PT NULL
#define OPTION_VAL_AUTO_PT_PT "Automático"
#define OPTION_VAL_NTSC_J_PT_PT NULL
#define OPTION_VAL_NTSC_U_PT_PT NULL
#define OPTION_VAL_PAL_PT_PT NULL
#define BEETLE_OPT_INTERNAL_RESOLUTION_LABEL_PT_PT NULL
#define BEETLE_OPT_INTERNAL_RESOLUTION_INFO_0_PT_PT NULL
#define BEETLE_OPT_RENDERER_LABEL_PT_PT NULL
#define BEETLE_OPT_RENDERER_INFO_0_PT_PT NULL
#define OPTION_VAL_HARDWARE_PT_PT NULL
#define OPTION_VAL_HARDWARE_GL_PT_PT NULL
#define OPTION_VAL_HARDWARE_VK_PT_PT NULL
#define OPTION_VAL_SOFTWARE_PT_PT NULL
#define BEETLE_OPT_RENDERER_SOFTWARE_FB_LABEL_PT_PT NULL
#define BEETLE_OPT_RENDERER_SOFTWARE_FB_INFO_0_PT_PT NULL
#define BEETLE_OPT_DEPTH_LABEL_PT_PT NULL
#define BEETLE_OPT_DEPTH_INFO_0_PT_PT NULL
#define OPTION_VAL_16BPP_NATIVE_PT_PT NULL
#define OPTION_VAL_32BPP_PT_PT NULL
#define BEETLE_OPT_DITHER_MODE_LABEL_PT_PT NULL
#define BEETLE_OPT_DITHER_MODE_INFO_0_PT_PT NULL
#define OPTION_VAL_INTERNAL_RESOLUTION_PT_PT NULL
#define BEETLE_OPT_SCALED_UV_OFFSET_LABEL_PT_PT NULL
#define BEETLE_OPT_SCALED_UV_OFFSET_INFO_0_PT_PT NULL
#define BEETLE_OPT_FILTER_LABEL_PT_PT NULL
#define BEETLE_OPT_FILTER_INFO_0_PT_PT NULL
#define OPTION_VAL_NEAREST_PT_PT "O mais próximo"
#define OPTION_VAL_SABR_PT_PT NULL
#define OPTION_VAL_XBR_PT_PT NULL
#define OPTION_VAL_BILINEAR_PT_PT NULL
#define OPTION_VAL_3_POINT_PT_PT NULL
#define OPTION_VAL_JINC2_PT_PT NULL
#define BEETLE_OPT_FILTER_EXCLUDE_SPRITE_LABEL_PT_PT NULL
#define BEETLE_OPT_FILTER_EXCLUDE_SPRITE_INFO_0_PT_PT NULL
#define OPTION_VAL_OPAQUE_PT_PT NULL
#define OPTION_VAL_ALL_PT_PT NULL
#define BEETLE_OPT_FILTER_EXCLUDE_2D_POLYGON_LABEL_PT_PT NULL
#define BEETLE_OPT_FILTER_EXCLUDE_2D_POLYGON_INFO_0_PT_PT NULL
#define BEETLE_OPT_ADAPTIVE_SMOOTHING_LABEL_PT_PT NULL
#define BEETLE_OPT_ADAPTIVE_SMOOTHING_INFO_0_PT_PT NULL
#define BEETLE_OPT_SUPER_SAMPLING_LABEL_PT_PT NULL
#define BEETLE_OPT_SUPER_SAMPLING_INFO_0_PT_PT NULL
#define BEETLE_OPT_MSAA_LABEL_PT_PT NULL
#define BEETLE_OPT_MSAA_INFO_0_PT_PT NULL
#define OPTION_VAL_1X_PT_PT NULL
#define BEETLE_OPT_MDEC_YUV_LABEL_PT_PT NULL
#define BEETLE_OPT_MDEC_YUV_INFO_0_PT_PT NULL
#define BEETLE_OPT_TRACK_TEXTURES_LABEL_PT_PT NULL
#define BEETLE_OPT_TRACK_TEXTURES_INFO_0_PT_PT NULL
#define BEETLE_OPT_DUMP_TEXTURES_LABEL_PT_PT NULL
#define BEETLE_OPT_DUMP_TEXTURES_INFO_0_PT_PT NULL
#define BEETLE_OPT_REPLACE_TEXTURES_LABEL_PT_PT NULL
#define BEETLE_OPT_REPLACE_TEXTURES_INFO_0_PT_PT NULL
#define BEETLE_OPT_PGXP_MODE_LABEL_PT_PT NULL
#define BEETLE_OPT_PGXP_MODE_INFO_0_PT_PT NULL
#define OPTION_VAL_MEMORY_ONLY_PT_PT NULL
#define OPTION_VAL_MEMORY_CPU_PT_PT NULL
#define BEETLE_OPT_PGXP_2D_TOL_LABEL_PT_PT NULL
#define BEETLE_OPT_PGXP_2D_TOL_INFO_0_PT_PT NULL
#define OPTION_VAL_0PX_PT_PT NULL
#define OPTION_VAL_1PX_PT_PT NULL
#define OPTION_VAL_2PX_PT_PT NULL
#define OPTION_VAL_3PX_PT_PT NULL
#define OPTION_VAL_4PX_PT_PT NULL
#define OPTION_VAL_5PX_PT_PT NULL
#define OPTION_VAL_6PX_PT_PT NULL
#define OPTION_VAL_7PX_PT_PT NULL
#define OPTION_VAL_8PX_PT_PT NULL
#define BEETLE_OPT_PGXP_NCLIP_LABEL_PT_PT NULL
#define BEETLE_OPT_PGXP_NCLIP_INFO_0_PT_PT NULL
#define BEETLE_OPT_PGXP_VERTEX_LABEL_PT_PT NULL
#define BEETLE_OPT_PGXP_VERTEX_INFO_0_PT_PT NULL
#define BEETLE_OPT_PGXP_TEXTURE_LABEL_PT_PT NULL
#define BEETLE_OPT_PGXP_TEXTURE_INFO_0_PT_PT NULL
#define BEETLE_OPT_DISPLAY_INTERNAL_FPS_LABEL_PT_PT NULL
#define BEETLE_OPT_DISPLAY_INTERNAL_FPS_INFO_0_PT_PT NULL
#define BEETLE_OPT_DISPLAY_OSD_LABEL_PT_PT NULL
#define BEETLE_OPT_DISPLAY_OSD_INFO_0_PT_PT NULL
#define BEETLE_OPT_DISPLAY_VRAM_LABEL_PT_PT NULL
#define BEETLE_OPT_DISPLAY_VRAM_INFO_0_PT_PT NULL
#define BEETLE_OPT_ANALOG_CALIBRATION_LABEL_PT_PT NULL
#define BEETLE_OPT_ANALOG_CALIBRATION_INFO_0_PT_PT NULL
#define BEETLE_OPT_ANALOG_TOGGLE_LABEL_PT_PT "Alternar o modo analógico do DualShock"
#define BEETLE_OPT_ANALOG_TOGGLE_INFO_0_PT_PT "Quando o tipo de dispositivo de entrada é o DualShock, esta opção permite que o DualShock emulado seja alternado entre os modos DIGITAL e ANALÓGICO, tal como o hardware original. O modo também pode ser alternado utilizando a combinação de botões selecionada."
#define OPTION_VAL_ENABLED_ANALOG_PT_PT "Analógico por predefinição"
#define BEETLE_OPT_ANALOG_TOGGLE_COMBO_LABEL_PT_PT NULL
#define BEETLE_OPT_ANALOG_TOGGLE_COMBO_INFO_0_PT_PT NULL
#define OPTION_VAL_L1_L2_R1_R2_START_SELECT_PT_PT NULL
#define OPTION_VAL_L1_R1_SELECT_PT_PT NULL
#define OPTION_VAL_L1_R1_START_PT_PT NULL
#define OPTION_VAL_L1_R1_L3_PT_PT NULL
#define OPTION_VAL_L1_R1_R3_PT_PT NULL
#define OPTION_VAL_L2_R2_SELECT_PT_PT NULL
#define OPTION_VAL_L2_R2_START_PT_PT NULL
#define OPTION_VAL_L2_R2_L3_PT_PT NULL
#define OPTION_VAL_L2_R2_R3_PT_PT NULL
#define OPTION_VAL_L3_R3_PT_PT NULL
#define BEETLE_OPT_ANALOG_TOGGLE_HOLD_LABEL_PT_PT NULL
#define BEETLE_OPT_ANALOG_TOGGLE_HOLD_INFO_0_PT_PT NULL
#define OPTION_VAL_0_PT_PT NULL
#define OPTION_VAL_1_PT_PT NULL
#define OPTION_VAL_2_PT_PT NULL
#define OPTION_VAL_3_PT_PT NULL
#define OPTION_VAL_4_PT_PT NULL
#define OPTION_VAL_5_PT_PT NULL
#define BEETLE_OPT_ENABLE_MULTITAP_PORT1_LABEL_PT_PT NULL
#define BEETLE_OPT_ENABLE_MULTITAP_PORT1_INFO_0_PT_PT NULL
#define BEETLE_OPT_ENABLE_MULTITAP_PORT2_LABEL_PT_PT NULL
#define BEETLE_OPT_ENABLE_MULTITAP_PORT2_INFO_0_PT_PT NULL
#define BEETLE_OPT_GUN_INPUT_MODE_LABEL_PT_PT NULL
#define BEETLE_OPT_GUN_INPUT_MODE_INFO_0_PT_PT NULL
#define OPTION_VAL_LIGHTGUN_PT_PT NULL
#define OPTION_VAL_TOUCHSCREEN_PT_PT NULL
#define BEETLE_OPT_GUN_CURSOR_LABEL_PT_PT NULL
#define BEETLE_OPT_GUN_CURSOR_INFO_0_PT_PT NULL
#define OPTION_VAL_CROSS_PT_PT NULL
#define OPTION_VAL_DOT_PT_PT NULL
#define BEETLE_OPT_CROSSHAIR_COLOR_P1_LABEL_PT_PT NULL
#define BEETLE_OPT_CROSSHAIR_COLOR_P1_INFO_0_PT_PT NULL
#define OPTION_VAL_RED_PT_PT "Vermelho"
#define OPTION_VAL_BLUE_PT_PT "Azul"
#define OPTION_VAL_GREEN_PT_PT "Verde"
#define OPTION_VAL_ORANGE_PT_PT NULL
#define OPTION_VAL_YELLOW_PT_PT "Amarelo"
#define OPTION_VAL_CYAN_PT_PT NULL
#define OPTION_VAL_PINK_PT_PT NULL
#define OPTION_VAL_PURPLE_PT_PT NULL
#define OPTION_VAL_BLACK_PT_PT NULL
#define OPTION_VAL_WHITE_PT_PT NULL
#define BEETLE_OPT_CROSSHAIR_COLOR_P2_LABEL_PT_PT NULL
#define BEETLE_OPT_CROSSHAIR_COLOR_P2_INFO_0_PT_PT NULL
#define BEETLE_OPT_MOUSE_SENSITIVITY_LABEL_PT_PT NULL
#define BEETLE_OPT_MOUSE_SENSITIVITY_INFO_0_PT_PT NULL
#define OPTION_VAL_5_O42_PT_PT NULL
#define OPTION_VAL_10_PT_PT NULL
#define OPTION_VAL_15_PT_PT NULL
#define OPTION_VAL_20_PT_PT NULL
#define OPTION_VAL_25_PT_PT NULL
#define OPTION_VAL_30_PT_PT NULL
#define OPTION_VAL_35_PT_PT NULL
#define OPTION_VAL_40_PT_PT NULL
#define OPTION_VAL_45_PT_PT NULL
#define OPTION_VAL_55_PT_PT NULL
#define OPTION_VAL_65_PT_PT NULL
#define OPTION_VAL_75_PT_PT NULL
#define OPTION_VAL_85_PT_PT NULL
#define OPTION_VAL_95_PT_PT NULL
#define OPTION_VAL_100_O42_PT_PT NULL
#define OPTION_VAL_105_PT_PT NULL
#define OPTION_VAL_115_PT_PT NULL
#define OPTION_VAL_125_PT_PT NULL
#define OPTION_VAL_135_PT_PT NULL
#define OPTION_VAL_145_PT_PT NULL
#define OPTION_VAL_155_PT_PT NULL
#define OPTION_VAL_165_PT_PT NULL
#define OPTION_VAL_175_PT_PT NULL
#define OPTION_VAL_185_PT_PT NULL
#define OPTION_VAL_195_PT_PT NULL
#define BEETLE_OPT_NEGCON_RESPONSE_LABEL_PT_PT NULL
#define BEETLE_OPT_NEGCON_RESPONSE_INFO_0_PT_PT NULL
#define OPTION_VAL_LINEAR_PT_PT NULL
#define OPTION_VAL_QUADRATIC_PT_PT NULL
#define OPTION_VAL_CUBIC_PT_PT NULL
#define BEETLE_OPT_NEGCON_DEADZONE_LABEL_PT_PT NULL
#define BEETLE_OPT_NEGCON_DEADZONE_INFO_0_PT_PT NULL
#define OPTION_VAL_0_O44_PT_PT NULL
#define BEETLE_OPT_USE_MEDNAFEN_MEMCARD0_METHOD_LABEL_PT_PT NULL
#define BEETLE_OPT_USE_MEDNAFEN_MEMCARD0_METHOD_INFO_0_PT_PT NULL
#define OPTION_VAL_LIBRETRO_PT_PT NULL
#define OPTION_VAL_MEDNAFEN_PT_PT NULL
#define BEETLE_OPT_ENABLE_MEMCARD1_LABEL_PT_PT NULL
#define BEETLE_OPT_ENABLE_MEMCARD1_INFO_0_PT_PT NULL
#define BEETLE_OPT_SHARED_MEMORY_CARDS_LABEL_PT_PT NULL
#define BEETLE_OPT_SHARED_MEMORY_CARDS_INFO_0_PT_PT NULL
#define BEETLE_OPT_MEMCARD_LEFT_INDEX_LABEL_PT_PT NULL
#define BEETLE_OPT_MEMCARD_LEFT_INDEX_INFO_0_PT_PT NULL
#define OPTION_VAL_0_O48_PT_PT NULL
#define BEETLE_OPT_MEMCARD_RIGHT_INDEX_LABEL_PT_PT NULL
#define BEETLE_OPT_MEMCARD_RIGHT_INDEX_INFO_0_PT_PT NULL
#define OPTION_VAL_1_O49_PT_PT NULL
#define BEETLE_OPT_LINE_RENDER_LABEL_PT_PT NULL
#define BEETLE_OPT_LINE_RENDER_INFO_0_PT_PT NULL
#define OPTION_VAL_DEFAULT_PT_PT NULL
#define OPTION_VAL_AGGRESSIVE_PT_PT NULL
#define BEETLE_OPT_WIDESCREEN_HACK_LABEL_PT_PT NULL
#define BEETLE_OPT_WIDESCREEN_HACK_INFO_0_PT_PT NULL
#define BEETLE_OPT_WIDESCREEN_HACK_ASPECT_RATIO_LABEL_PT_PT NULL
#define BEETLE_OPT_WIDESCREEN_HACK_ASPECT_RATIO_INFO_0_PT_PT NULL
#define OPTION_VAL_16_9_PT_PT NULL
#define OPTION_VAL_16_10_PT_PT NULL
#define OPTION_VAL_18_9_PT_PT NULL
#define OPTION_VAL_19_9_PT_PT NULL
#define OPTION_VAL_20_9_PT_PT NULL
#define OPTION_VAL_21_9_PT_PT NULL
#define OPTION_VAL_32_9_PT_PT NULL
#define BEETLE_OPT_CPU_DYNAREC_LABEL_PT_PT NULL
#define BEETLE_OPT_CPU_DYNAREC_INFO_0_PT_PT NULL
#define OPTION_VAL_DISABLED_PT_PT NULL
#define OPTION_VAL_EXECUTE_PT_PT NULL
#define OPTION_VAL_EXECUTE_ONE_PT_PT NULL
#define OPTION_VAL_RUN_INTERPRETER_PT_PT NULL
#define BEETLE_OPT_DYNAREC_INVALIDATE_LABEL_PT_PT NULL
#define BEETLE_OPT_DYNAREC_INVALIDATE_INFO_0_PT_PT NULL
#define OPTION_VAL_FULL_PT_PT NULL
#define OPTION_VAL_DMA_PT_PT NULL
#define BEETLE_OPT_DYNAREC_EVENTCYCLES_LABEL_PT_PT NULL
#define BEETLE_OPT_DYNAREC_EVENTCYCLES_INFO_0_PT_PT NULL
#define OPTION_VAL_128_PT_PT NULL
#define BEETLE_OPT_DYNAREC_SPU_SAMPLES_LABEL_PT_PT NULL
#define BEETLE_OPT_DYNAREC_SPU_SAMPLES_INFO_0_PT_PT NULL
#define BEETLE_OPT_CORE_TIMING_FPS_LABEL_PT_PT NULL
#define BEETLE_OPT_CORE_TIMING_FPS_INFO_0_PT_PT NULL
#define OPTION_VAL_AUTO_TOGGLE_PT_PT NULL
#define OPTION_VAL_FORCE_PROGRESSIVE_PT_PT NULL
#define OPTION_VAL_FORCE_INTERLACED_PT_PT NULL
#define BEETLE_OPT_DEINTERLACER_LABEL_PT_PT NULL
#define BEETLE_OPT_DEINTERLACER_INFO_0_PT_PT NULL
#define OPTION_VAL_WEAVE_PT_PT NULL
#define OPTION_VAL_BOB_PT_PT NULL
#define OPTION_VAL_BOB_OFFSET_PT_PT NULL
#define OPTION_VAL_FASTMAD_PT_PT NULL
#define BEETLE_OPT_FRAME_DUPING_LABEL_PT_PT NULL
#define BEETLE_OPT_FRAME_DUPING_INFO_0_PT_PT NULL
#define BEETLE_OPT_PAL_VIDEO_TIMING_OVERRIDE_LABEL_PT_PT NULL
#define BEETLE_OPT_PAL_VIDEO_TIMING_OVERRIDE_INFO_0_PT_PT NULL
#define BEETLE_OPT_ASPECT_RATIO_LABEL_PT_PT NULL
#define BEETLE_OPT_ASPECT_RATIO_INFO_0_PT_PT NULL
#define OPTION_VAL_CORRECTED_PT_PT NULL
#define OPTION_VAL_UNCORRECTED_PT_PT NULL
#define OPTION_VAL_4_3_PT_PT NULL
#define OPTION_VAL_NTSC_PT_PT NULL
#define BEETLE_OPT_CROP_OVERSCAN_LABEL_PT_PT NULL
#define BEETLE_OPT_CROP_OVERSCAN_INFO_0_PT_PT NULL
#define OPTION_VAL_STATIC_PT_PT NULL
#define OPTION_VAL_SMART_PT_PT NULL
#define BEETLE_OPT_IMAGE_CROP_LABEL_PT_PT NULL
#define BEETLE_OPT_IMAGE_CROP_INFO_0_PT_PT NULL
#define OPTION_VAL_9PX_PT_PT NULL
#define OPTION_VAL_10PX_PT_PT NULL
#define OPTION_VAL_11PX_PT_PT NULL
#define OPTION_VAL_12PX_PT_PT NULL
#define OPTION_VAL_13PX_PT_PT NULL
#define OPTION_VAL_14PX_PT_PT NULL
#define OPTION_VAL_15PX_PT_PT NULL
#define OPTION_VAL_16PX_PT_PT NULL
#define OPTION_VAL_17PX_PT_PT NULL
#define OPTION_VAL_18PX_PT_PT NULL
#define OPTION_VAL_19PX_PT_PT NULL
#define OPTION_VAL_20PX_PT_PT NULL
#define BEETLE_OPT_IMAGE_OFFSET_LABEL_PT_PT NULL
#define BEETLE_OPT_IMAGE_OFFSET_INFO_0_PT_PT NULL
#define OPTION_VAL_12PX_O64_PT_PT NULL
#define OPTION_VAL_11PX_O64_PT_PT NULL
#define OPTION_VAL_10PX_O64_PT_PT NULL
#define OPTION_VAL_9PX_O64_PT_PT NULL
#define OPTION_VAL_8PX_O64_PT_PT NULL
#define OPTION_VAL_7PX_O64_PT_PT NULL
#define OPTION_VAL_6PX_O64_PT_PT NULL
#define OPTION_VAL_5PX_O64_PT_PT NULL
#define OPTION_VAL_4PX_O64_PT_PT NULL
#define OPTION_VAL_3PX_O64_PT_PT NULL
#define OPTION_VAL_2PX_O64_PT_PT NULL
#define OPTION_VAL_1PX_O64_PT_PT NULL
#define OPTION_VAL_1PX_O64_0_PT_PT NULL
#define OPTION_VAL_2PX_O64_0_PT_PT NULL
#define OPTION_VAL_3PX_O64_0_PT_PT NULL
#define OPTION_VAL_4PX_O64_0_PT_PT NULL
#define OPTION_VAL_5PX_O64_0_PT_PT NULL
#define OPTION_VAL_6PX_O64_0_PT_PT NULL
#define OPTION_VAL_7PX_O64_0_PT_PT NULL
#define OPTION_VAL_8PX_O64_0_PT_PT NULL
#define OPTION_VAL_9PX_O64_0_PT_PT NULL
#define OPTION_VAL_10PX_O64_0_PT_PT NULL
#define OPTION_VAL_11PX_O64_0_PT_PT NULL
#define OPTION_VAL_12PX_O64_0_PT_PT NULL
#define BEETLE_OPT_IMAGE_OFFSET_CYCLES_LABEL_PT_PT NULL
#define BEETLE_OPT_IMAGE_OFFSET_CYCLES_INFO_0_PT_PT NULL
#define BEETLE_OPT_INITIAL_SCANLINE_LABEL_PT_PT NULL
#define BEETLE_OPT_INITIAL_SCANLINE_INFO_0_PT_PT NULL
#define BEETLE_OPT_LAST_SCANLINE_LABEL_PT_PT NULL
#define BEETLE_OPT_LAST_SCANLINE_INFO_0_PT_PT NULL
#define OPTION_VAL_239_PT_PT "239 (Padrão)"
#define BEETLE_OPT_INITIAL_SCANLINE_PAL_LABEL_PT_PT NULL
#define BEETLE_OPT_INITIAL_SCANLINE_PAL_INFO_0_PT_PT NULL
#define BEETLE_OPT_LAST_SCANLINE_PAL_LABEL_PT_PT NULL
#define BEETLE_OPT_LAST_SCANLINE_PAL_INFO_0_PT_PT NULL
#define OPTION_VAL_287_PT_PT "287 (Padrão)"

struct retro_core_option_v2_category option_cats_pt_pt[] = {
   {
      "system",
      CATEGORY_SYSTEM_LABEL_PT_PT,
      CATEGORY_SYSTEM_INFO_0_PT_PT
   },
   {
      "video",
      CATEGORY_VIDEO_LABEL_PT_PT,
      CATEGORY_VIDEO_INFO_0_PT_PT
   },
   {
      "pgxp",
      CATEGORY_PGXP_LABEL_PT_PT,
      CATEGORY_PGXP_INFO_0_PT_PT
   },
   {
      "osd",
      CATEGORY_OSD_LABEL_PT_PT,
      CATEGORY_OSD_INFO_0_PT_PT
   },
   {
      "input",
      CATEGORY_INPUT_LABEL_PT_PT,
      CATEGORY_INPUT_INFO_0_PT_PT
   },
   {
      "memcards",
      CATEGORY_MEMCARDS_LABEL_PT_PT,
      CATEGORY_MEMCARDS_INFO_0_PT_PT
   },
   {
      "hacks",
      CATEGORY_HACKS_LABEL_PT_PT,
      CATEGORY_HACKS_INFO_0_PT_PT
   },
   { NULL, NULL, NULL },
};
struct retro_core_option_v2_definition option_defs_pt_pt[] = {
   {
      BEETLE_OPT(cpu_freq_scale),
      BEETLE_OPT_CPU_FREQ_SCALE_LABEL_PT_PT,
      NULL,
      BEETLE_OPT_CPU_FREQ_SCALE_INFO_0_PT_PT,
      NULL,
      "system",
      {
         { "50%",           OPTION_VAL_50_PT_PT },
         { "60%",           OPTION_VAL_60_PT_PT },
         { "70%",           OPTION_VAL_70_PT_PT },
         { "80%",           OPTION_VAL_80_PT_PT },
         { "90%",           OPTION_VAL_90_PT_PT },
         { "100%",          OPTION_VAL_100_PT_PT },
         { "110%",          OPTION_VAL_110_PT_PT },
         { "120%",          OPTION_VAL_120_PT_PT },
         { "130%",          OPTION_VAL_130_PT_PT },
         { "140%",          OPTION_VAL_140_PT_PT },
         { "150%",          OPTION_VAL_150_PT_PT },
         { "160%",          OPTION_VAL_160_PT_PT },
         { "170%",          OPTION_VAL_170_PT_PT },
         { "180%",          OPTION_VAL_180_PT_PT },
         { "190%",          OPTION_VAL_190_PT_PT },
         { "200%",          OPTION_VAL_200_PT_PT },
         { "210%",          OPTION_VAL_210_PT_PT },
         { "220%",          OPTION_VAL_220_PT_PT },
         { "230%",          OPTION_VAL_230_PT_PT },
         { "240%",          OPTION_VAL_240_PT_PT },
         { "250%",          OPTION_VAL_250_PT_PT },
         { "260%",          OPTION_VAL_260_PT_PT },
         { "270%",          OPTION_VAL_270_PT_PT },
         { "280%",          OPTION_VAL_280_PT_PT },
         { "290%",          OPTION_VAL_290_PT_PT },
         { "300%",          OPTION_VAL_300_PT_PT },
         { "310%",          OPTION_VAL_310_PT_PT },
         { "320%",          OPTION_VAL_320_PT_PT },
         { "330%",          OPTION_VAL_330_PT_PT },
         { "340%",          OPTION_VAL_340_PT_PT },
         { "350%",          OPTION_VAL_350_PT_PT },
         { "360%",          OPTION_VAL_360_PT_PT },
         { "370%",          OPTION_VAL_370_PT_PT },
         { "380%",          OPTION_VAL_380_PT_PT },
         { "390%",          OPTION_VAL_390_PT_PT },
         { "400%",          OPTION_VAL_400_PT_PT },
         { "410%",          OPTION_VAL_410_PT_PT },
         { "420%",          OPTION_VAL_420_PT_PT },
         { "430%",          OPTION_VAL_430_PT_PT },
         { "440%",          OPTION_VAL_440_PT_PT },
         { "450%",          OPTION_VAL_450_PT_PT },
         { "460%",          OPTION_VAL_460_PT_PT },
         { "470%",          OPTION_VAL_470_PT_PT },
         { "480%",          OPTION_VAL_480_PT_PT },
         { "490%",          OPTION_VAL_490_PT_PT },
         { "500%",          OPTION_VAL_500_PT_PT },
         { "510%",          OPTION_VAL_510_PT_PT },
         { "520%",          OPTION_VAL_520_PT_PT },
         { "530%",          OPTION_VAL_530_PT_PT },
         { "540%",          OPTION_VAL_540_PT_PT },
         { "550%",          OPTION_VAL_550_PT_PT },
         { "560%",          OPTION_VAL_560_PT_PT },
         { "570%",          OPTION_VAL_570_PT_PT },
         { "580%",          OPTION_VAL_580_PT_PT },
         { "590%",          OPTION_VAL_590_PT_PT },
         { "600%",          OPTION_VAL_600_PT_PT },
         { "610%",          OPTION_VAL_610_PT_PT },
         { "620%",          OPTION_VAL_620_PT_PT },
         { "630%",          OPTION_VAL_630_PT_PT },
         { "640%",          OPTION_VAL_640_PT_PT },
         { "650%",          OPTION_VAL_650_PT_PT },
         { "660%",          OPTION_VAL_660_PT_PT },
         { "670%",          OPTION_VAL_670_PT_PT },
         { "680%",          OPTION_VAL_680_PT_PT },
         { "690%",          OPTION_VAL_690_PT_PT },
         { "700%",          OPTION_VAL_700_PT_PT },
         { "710%",          OPTION_VAL_710_PT_PT },
         { "720%",          OPTION_VAL_720_PT_PT },
         { "730%",          OPTION_VAL_730_PT_PT },
         { "740%",          OPTION_VAL_740_PT_PT },
         { "750%",          OPTION_VAL_750_PT_PT },
         { NULL, NULL },
      },
      "100%"
   },
   {
      BEETLE_OPT(gpu_overclock),
      BEETLE_OPT_GPU_OVERCLOCK_LABEL_PT_PT,
      NULL,
      BEETLE_OPT_GPU_OVERCLOCK_INFO_0_PT_PT,
      NULL,
      "system",
      {
         { "1x(native)", OPTION_VAL_1X_NATIVE_PT_PT },
         { "2x",         OPTION_VAL_2X_PT_PT },
         { "4x",         OPTION_VAL_4X_PT_PT },
         { "8x",         OPTION_VAL_8X_PT_PT },
         { "16x",        OPTION_VAL_16X_PT_PT },
         { "32x",        OPTION_VAL_32X_PT_PT },
         { NULL, NULL },
      },
      "1x(native)"
   },
   {
      BEETLE_OPT(gte_overclock),
      BEETLE_OPT_GTE_OVERCLOCK_LABEL_PT_PT,
      NULL,
      BEETLE_OPT_GTE_OVERCLOCK_INFO_0_PT_PT,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
#ifndef EMSCRIPTEN
   {
      BEETLE_OPT(cd_access_method),
      BEETLE_OPT_CD_ACCESS_METHOD_LABEL_PT_PT,
      NULL,
      BEETLE_OPT_CD_ACCESS_METHOD_INFO_0_PT_PT,
      NULL,
      "system",
      {
         { "sync",     OPTION_VAL_SYNC_PT_PT },
         { "async",    OPTION_VAL_ASYNC_PT_PT },
         { "precache", OPTION_VAL_PRECACHE_PT_PT },
         { NULL, NULL },
      },
      "sync"
   },
#endif
   {
      BEETLE_OPT(cd_fastload),
      BEETLE_OPT_CD_FASTLOAD_LABEL_PT_PT,
      NULL,
      BEETLE_OPT_CD_FASTLOAD_INFO_0_PT_PT,
      NULL,
      "system",
      {
         { "2x(native)", OPTION_VAL_2X_NATIVE_PT_PT },
         { "4x",          OPTION_VAL_4X_PT_PT },
         { "6x",          OPTION_VAL_6X_PT_PT },
         { "8x",          OPTION_VAL_8X_PT_PT },
         { "10x",         OPTION_VAL_10X_PT_PT },
         { "12x",         OPTION_VAL_12X_PT_PT },
         { "14x",         OPTION_VAL_14X_PT_PT },
         { NULL, NULL },
      },
      "2x(native)"
   },
   {
      BEETLE_OPT(skip_bios),
      BEETLE_OPT_SKIP_BIOS_LABEL_PT_PT,
      NULL,
      BEETLE_OPT_SKIP_BIOS_INFO_0_PT_PT,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(override_bios),
      BEETLE_OPT_OVERRIDE_BIOS_LABEL_PT_PT,
      NULL,
      BEETLE_OPT_OVERRIDE_BIOS_INFO_0_PT_PT,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "psxonpsp", OPTION_VAL_PSXONPSP_PT_PT },
         { "ps1_rom",  OPTION_VAL_PS1_ROM_PT_PT },
         { "openbios", OPTION_VAL_OPENBIOS_PT_PT     },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(region),
      BEETLE_OPT_REGION_LABEL_PT_PT,
      NULL,
      BEETLE_OPT_REGION_INFO_0_PT_PT,
      NULL,
      "system",
      {
         { "auto",   OPTION_VAL_AUTO_PT_PT },
         { "ntsc-j", OPTION_VAL_NTSC_J_PT_PT },
         { "ntsc-u", OPTION_VAL_NTSC_U_PT_PT },
         { "pal",    OPTION_VAL_PAL_PT_PT },
         { NULL, NULL },
      },
      "auto"
   },
   {
      BEETLE_OPT(internal_resolution),
      BEETLE_OPT_INTERNAL_RESOLUTION_LABEL_PT_PT,
      NULL,
      BEETLE_OPT_INTERNAL_RESOLUTION_INFO_0_PT_PT,
      NULL,
      "video",
      {
         { "1x(native)", OPTION_VAL_1X_NATIVE_PT_PT },
         { "2x",         OPTION_VAL_2X_PT_PT },
         { "4x",         OPTION_VAL_4X_PT_PT },
         { "8x",         OPTION_VAL_8X_PT_PT },
         { "16x",        OPTION_VAL_16X_PT_PT },
         { NULL, NULL },
      },
      "1x(native)"
   },
#if defined(HAVE_OPENGL) || defined(HAVE_OPENGLES) || defined(HAVE_VULKAN)
   {
      BEETLE_OPT(renderer),
      BEETLE_OPT_RENDERER_LABEL_PT_PT,
      NULL,
      BEETLE_OPT_RENDERER_INFO_0_PT_PT,
      NULL,
      "video",
      {
         { "hardware",    OPTION_VAL_HARDWARE_PT_PT },
#if defined(HAVE_OPENGL) || defined(HAVE_OPENGLES)
         { "hardware_gl", OPTION_VAL_HARDWARE_GL_PT_PT },
#endif
#if defined(HAVE_VULKAN)
         { "hardware_vk", OPTION_VAL_HARDWARE_VK_PT_PT },
#endif
         { "software",    OPTION_VAL_SOFTWARE_PT_PT },
         { NULL, NULL },
      },
      "hardware"
   },
   {
      BEETLE_OPT(renderer_software_fb),
      BEETLE_OPT_RENDERER_SOFTWARE_FB_LABEL_PT_PT,
      NULL,
      BEETLE_OPT_RENDERER_SOFTWARE_FB_INFO_0_PT_PT,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled"
   },
#endif
#if defined(HAVE_OPENGL) || defined(HAVE_OPENGLES)
   {
      BEETLE_OPT(depth),
      BEETLE_OPT_DEPTH_LABEL_PT_PT,
      NULL,
      BEETLE_OPT_DEPTH_INFO_0_PT_PT,
      NULL,
      "video",
      {
         { "16bpp(native)", OPTION_VAL_16BPP_NATIVE_PT_PT },
         { "32bpp",         OPTION_VAL_32BPP_PT_PT },
         { NULL, NULL },
      },
      "16bpp(native)"
   },
   // Sort of, it's more like 15-bit high color and 24-bit true color for visible output. The alpha channel is used for mask bit. Vulkan renderer uses ABGR1555_555 for 31 bits internal? FMVs are always 24-bit on all renderers like original hardware (BGR888, no alpha)
#endif
   {
      BEETLE_OPT(dither_mode),
      BEETLE_OPT_DITHER_MODE_LABEL_PT_PT,
      NULL,
      BEETLE_OPT_DITHER_MODE_INFO_0_PT_PT,
      NULL,
      "video",
      {
         { "1x(native)",          OPTION_VAL_1X_NATIVE_PT_PT },
         { "internal resolution", OPTION_VAL_INTERNAL_RESOLUTION_PT_PT },
         { "disabled",            NULL },
         { NULL, NULL },
      },
      "1x(native)"
   },
#ifdef HAVE_VULKAN
   {
      BEETLE_OPT(scaled_uv_offset),
      BEETLE_OPT_SCALED_UV_OFFSET_LABEL_PT_PT,
      NULL,
      BEETLE_OPT_SCALED_UV_OFFSET_INFO_0_PT_PT,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled"
   },
#endif
#if defined(HAVE_OPENGL) || defined(HAVE_OPENGLES) || defined(HAVE_VULKAN)
   {
      BEETLE_OPT(filter),
      BEETLE_OPT_FILTER_LABEL_PT_PT,
      NULL,
      BEETLE_OPT_FILTER_INFO_0_PT_PT,
      NULL,
      "video",
      {
         { "nearest",  OPTION_VAL_NEAREST_PT_PT },
         { "SABR",     OPTION_VAL_SABR_PT_PT },
         { "xBR",      OPTION_VAL_XBR_PT_PT },
         { "bilinear", OPTION_VAL_BILINEAR_PT_PT },
         { "3-point",  OPTION_VAL_3_POINT_PT_PT },
         { "JINC2",    OPTION_VAL_JINC2_PT_PT },
         { NULL, NULL },
      },
      "nearest"
   },
#ifdef HAVE_VULKAN
   {
      BEETLE_OPT(filter_exclude_sprite),
      BEETLE_OPT_FILTER_EXCLUDE_SPRITE_LABEL_PT_PT,
      NULL,
      BEETLE_OPT_FILTER_EXCLUDE_SPRITE_INFO_0_PT_PT,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "opaque", OPTION_VAL_OPAQUE_PT_PT },
         { "all", OPTION_VAL_ALL_PT_PT },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(filter_exclude_2d_polygon),
      BEETLE_OPT_FILTER_EXCLUDE_2D_POLYGON_LABEL_PT_PT,
      NULL,
      BEETLE_OPT_FILTER_EXCLUDE_2D_POLYGON_INFO_0_PT_PT,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "opaque", OPTION_VAL_OPAQUE_PT_PT },
         { "all", OPTION_VAL_ALL_PT_PT },
         { NULL, NULL },
      },
      "disabled"
   },
#endif
#endif
#ifdef HAVE_VULKAN
   {
      BEETLE_OPT(adaptive_smoothing),
      BEETLE_OPT_ADAPTIVE_SMOOTHING_LABEL_PT_PT,
      NULL,
      BEETLE_OPT_ADAPTIVE_SMOOTHING_INFO_0_PT_PT,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(super_sampling),
      BEETLE_OPT_SUPER_SAMPLING_LABEL_PT_PT,
      NULL,
      BEETLE_OPT_SUPER_SAMPLING_INFO_0_PT_PT,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(msaa),
      BEETLE_OPT_MSAA_LABEL_PT_PT,
      NULL,
      BEETLE_OPT_MSAA_INFO_0_PT_PT,
      NULL,
      "video",
      {
         { "1x",  OPTION_VAL_1X_PT_PT },
         { "2x",  OPTION_VAL_2X_PT_PT },
         { "4x",  OPTION_VAL_4X_PT_PT },
         { "8x",  OPTION_VAL_8X_PT_PT },
         { "16x", OPTION_VAL_16X_PT_PT },
         { NULL, NULL },
      },
      "1x"
   },
   {
      BEETLE_OPT(mdec_yuv),
      BEETLE_OPT_MDEC_YUV_LABEL_PT_PT,
      NULL,
      BEETLE_OPT_MDEC_YUV_INFO_0_PT_PT,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(track_textures),
      BEETLE_OPT_TRACK_TEXTURES_LABEL_PT_PT,
      NULL,
      BEETLE_OPT_TRACK_TEXTURES_INFO_0_PT_PT,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
#ifdef TEXTURE_DUMPING_ENABLED
   {
      BEETLE_OPT(dump_textures),
      BEETLE_OPT_DUMP_TEXTURES_LABEL_PT_PT,
      NULL,
      BEETLE_OPT_DUMP_TEXTURES_INFO_0_PT_PT,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
#endif
   {
      BEETLE_OPT(replace_textures),
      BEETLE_OPT_REPLACE_TEXTURES_LABEL_PT_PT,
      NULL,
      BEETLE_OPT_REPLACE_TEXTURES_INFO_0_PT_PT,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
#endif
   {
      BEETLE_OPT(pgxp_mode),
      BEETLE_OPT_PGXP_MODE_LABEL_PT_PT,
      NULL,
      BEETLE_OPT_PGXP_MODE_INFO_0_PT_PT,
      NULL,
      "pgxp",
      {
         { "disabled",     NULL },
         { "memory only",  OPTION_VAL_MEMORY_ONLY_PT_PT },
         { "memory + CPU", OPTION_VAL_MEMORY_CPU_PT_PT },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(pgxp_2d_tol),
      BEETLE_OPT_PGXP_2D_TOL_LABEL_PT_PT,
      NULL,
      BEETLE_OPT_PGXP_2D_TOL_INFO_0_PT_PT,
      NULL,
      "pgxp",
      {
         { "disabled", NULL },
         { "0px", OPTION_VAL_0PX_PT_PT },
         { "1px", OPTION_VAL_1PX_PT_PT },
         { "2px", OPTION_VAL_2PX_PT_PT },
         { "3px", OPTION_VAL_3PX_PT_PT },
         { "4px", OPTION_VAL_4PX_PT_PT },
         { "5px", OPTION_VAL_5PX_PT_PT },
         { "6px", OPTION_VAL_6PX_PT_PT },
         { "7px", OPTION_VAL_7PX_PT_PT },
         { "8px", OPTION_VAL_8PX_PT_PT },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(pgxp_nclip),
      BEETLE_OPT_PGXP_NCLIP_LABEL_PT_PT,
      NULL,
      BEETLE_OPT_PGXP_NCLIP_INFO_0_PT_PT,
      NULL,
      "pgxp",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
#if defined(HAVE_OPENGL) || defined(HAVE_OPENGLES) || defined(HAVE_VULKAN)
   {
      BEETLE_OPT(pgxp_vertex),
      BEETLE_OPT_PGXP_VERTEX_LABEL_PT_PT,
      NULL,
      BEETLE_OPT_PGXP_VERTEX_INFO_0_PT_PT,
      NULL,
      "pgxp",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
#endif
   {
      BEETLE_OPT(pgxp_texture),
      BEETLE_OPT_PGXP_TEXTURE_LABEL_PT_PT,
      NULL,
      BEETLE_OPT_PGXP_TEXTURE_INFO_0_PT_PT,
      NULL,
      "pgxp",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(display_internal_fps),
      BEETLE_OPT_DISPLAY_INTERNAL_FPS_LABEL_PT_PT,
      NULL,
      BEETLE_OPT_DISPLAY_INTERNAL_FPS_INFO_0_PT_PT,
      NULL,
      "osd",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(display_osd),
      BEETLE_OPT_DISPLAY_OSD_LABEL_PT_PT,
      NULL,
      BEETLE_OPT_DISPLAY_OSD_INFO_0_PT_PT,
      NULL,
      "osd",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled"
   },
#if defined(HAVE_OPENGL) || defined(HAVE_OPENGLES) || defined(HAVE_VULKAN)
   {
      BEETLE_OPT(display_vram),
      BEETLE_OPT_DISPLAY_VRAM_LABEL_PT_PT,
      NULL,
      BEETLE_OPT_DISPLAY_VRAM_INFO_0_PT_PT,
      NULL,
      "osd",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
#endif
   {
      BEETLE_OPT(analog_calibration),
      BEETLE_OPT_ANALOG_CALIBRATION_LABEL_PT_PT,
      NULL,
      BEETLE_OPT_ANALOG_CALIBRATION_INFO_0_PT_PT,
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(analog_toggle),
      BEETLE_OPT_ANALOG_TOGGLE_LABEL_PT_PT,
      NULL,
      BEETLE_OPT_ANALOG_TOGGLE_INFO_0_PT_PT,
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { "enabled-analog", OPTION_VAL_ENABLED_ANALOG_PT_PT },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(analog_toggle_combo),
      BEETLE_OPT_ANALOG_TOGGLE_COMBO_LABEL_PT_PT,
      NULL,
      BEETLE_OPT_ANALOG_TOGGLE_COMBO_INFO_0_PT_PT,
      NULL,
      "input",
      {
         { "l1+l2+r1+r2+start+select", OPTION_VAL_L1_L2_R1_R2_START_SELECT_PT_PT },
         { "l1+r1+select",             OPTION_VAL_L1_R1_SELECT_PT_PT },
         { "l1+r1+start",              OPTION_VAL_L1_R1_START_PT_PT },
         { "l1+r1+l3",                 OPTION_VAL_L1_R1_L3_PT_PT },
         { "l1+r1+r3",                 OPTION_VAL_L1_R1_R3_PT_PT },
         { "l2+r2+select",             OPTION_VAL_L2_R2_SELECT_PT_PT },
         { "l2+r2+start",              OPTION_VAL_L2_R2_START_PT_PT },
         { "l2+r2+l3",                 OPTION_VAL_L2_R2_L3_PT_PT },
         { "l2+r2+r3",                 OPTION_VAL_L2_R2_R3_PT_PT },
         { "l3+r3",                    OPTION_VAL_L3_R3_PT_PT },
         { NULL, NULL },
      },
      "l1+r1+select"
   },
   {
      BEETLE_OPT(analog_toggle_hold),
      BEETLE_OPT_ANALOG_TOGGLE_HOLD_LABEL_PT_PT,
      NULL,
      BEETLE_OPT_ANALOG_TOGGLE_HOLD_INFO_0_PT_PT,
      NULL,
      "input",
      {
         { "0", OPTION_VAL_0_PT_PT },
         { "1", OPTION_VAL_1_PT_PT },
         { "2", OPTION_VAL_2_PT_PT },
         { "3", OPTION_VAL_3_PT_PT },
         { "4", OPTION_VAL_4_PT_PT },
         { "5", OPTION_VAL_5_PT_PT },
         { NULL, NULL },
      },
      "1"
   },
   {
      BEETLE_OPT(enable_multitap_port1),
      BEETLE_OPT_ENABLE_MULTITAP_PORT1_LABEL_PT_PT,
      NULL,
      BEETLE_OPT_ENABLE_MULTITAP_PORT1_INFO_0_PT_PT,
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(enable_multitap_port2),
      BEETLE_OPT_ENABLE_MULTITAP_PORT2_LABEL_PT_PT,
      NULL,
      BEETLE_OPT_ENABLE_MULTITAP_PORT2_INFO_0_PT_PT,
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(gun_input_mode),
      BEETLE_OPT_GUN_INPUT_MODE_LABEL_PT_PT,
      NULL,
      BEETLE_OPT_GUN_INPUT_MODE_INFO_0_PT_PT,
      NULL,
      "input",
      {
         { "lightgun",    OPTION_VAL_LIGHTGUN_PT_PT },
         { "touchscreen", OPTION_VAL_TOUCHSCREEN_PT_PT },
         { NULL, NULL },
      },
      "lightgun"
   },
   // Shouldn't the gun_input_mode just be Mouse vs. Touchscreen?
   {
      BEETLE_OPT(gun_cursor),
      BEETLE_OPT_GUN_CURSOR_LABEL_PT_PT,
      NULL,
      BEETLE_OPT_GUN_CURSOR_INFO_0_PT_PT,
      NULL,
      "input",
      {
         { "off",   "disabled" },
         { "cross", OPTION_VAL_CROSS_PT_PT },
         { "dot",   OPTION_VAL_DOT_PT_PT },
         { NULL, NULL },
      },
      "cross"
   },
   {
      BEETLE_OPT(crosshair_color_p1),
      BEETLE_OPT_CROSSHAIR_COLOR_P1_LABEL_PT_PT,
      NULL,
      BEETLE_OPT_CROSSHAIR_COLOR_P1_INFO_0_PT_PT,
      NULL,
      "input",
      {
         { "red",    OPTION_VAL_RED_PT_PT },
         { "blue",   OPTION_VAL_BLUE_PT_PT },
         { "green",  OPTION_VAL_GREEN_PT_PT },
         { "orange", OPTION_VAL_ORANGE_PT_PT },
         { "yellow", OPTION_VAL_YELLOW_PT_PT },
         { "cyan",   OPTION_VAL_CYAN_PT_PT },
         { "pink",   OPTION_VAL_PINK_PT_PT },
         { "purple", OPTION_VAL_PURPLE_PT_PT },
         { "black",  OPTION_VAL_BLACK_PT_PT },
         { "white",  OPTION_VAL_WHITE_PT_PT },
         { NULL, NULL },
      },
      "red"
   },
   {
      BEETLE_OPT(crosshair_color_p2),
      BEETLE_OPT_CROSSHAIR_COLOR_P2_LABEL_PT_PT,
      NULL,
      BEETLE_OPT_CROSSHAIR_COLOR_P2_INFO_0_PT_PT,
      NULL,
      "input",
      {
         { "blue",   OPTION_VAL_BLUE_PT_PT },
         { "red",    OPTION_VAL_RED_PT_PT },
         { "green",  OPTION_VAL_GREEN_PT_PT },
         { "orange", OPTION_VAL_ORANGE_PT_PT },
         { "yellow", OPTION_VAL_YELLOW_PT_PT },
         { "cyan",   OPTION_VAL_CYAN_PT_PT },
         { "pink",   OPTION_VAL_PINK_PT_PT },
         { "purple", OPTION_VAL_PURPLE_PT_PT },
         { "black",  OPTION_VAL_BLACK_PT_PT },
         { "white",  OPTION_VAL_WHITE_PT_PT },
         { NULL, NULL },
      },
      "blue"
   },
   {
      BEETLE_OPT(mouse_sensitivity),
      BEETLE_OPT_MOUSE_SENSITIVITY_LABEL_PT_PT,
      NULL,
      BEETLE_OPT_MOUSE_SENSITIVITY_INFO_0_PT_PT,
      NULL,
      "input",
      {
         { "5%",   OPTION_VAL_5_O42_PT_PT },
         { "10%",  OPTION_VAL_10_PT_PT },
         { "15%",  OPTION_VAL_15_PT_PT },
         { "20%",  OPTION_VAL_20_PT_PT },
         { "25%",  OPTION_VAL_25_PT_PT },
         { "30%",  OPTION_VAL_30_PT_PT },
         { "35%",  OPTION_VAL_35_PT_PT },
         { "40%",  OPTION_VAL_40_PT_PT },
         { "45%",  OPTION_VAL_45_PT_PT },
         { "50%",  OPTION_VAL_50_PT_PT },
         { "55%",  OPTION_VAL_55_PT_PT },
         { "60%",  OPTION_VAL_60_PT_PT },
         { "65%",  OPTION_VAL_65_PT_PT },
         { "70%",  OPTION_VAL_70_PT_PT },
         { "75%",  OPTION_VAL_75_PT_PT },
         { "80%",  OPTION_VAL_80_PT_PT },
         { "85%",  OPTION_VAL_85_PT_PT },
         { "90%",  OPTION_VAL_90_PT_PT },
         { "95%",  OPTION_VAL_95_PT_PT },
         { "100%", OPTION_VAL_100_O42_PT_PT },
         { "105%", OPTION_VAL_105_PT_PT },
         { "110%", OPTION_VAL_110_PT_PT },
         { "115%", OPTION_VAL_115_PT_PT },
         { "120%", OPTION_VAL_120_PT_PT },
         { "125%", OPTION_VAL_125_PT_PT },
         { "130%", OPTION_VAL_130_PT_PT },
         { "135%", OPTION_VAL_135_PT_PT },
         { "140%", OPTION_VAL_140_PT_PT },
         { "145%", OPTION_VAL_145_PT_PT },
         { "150%", OPTION_VAL_150_PT_PT },
         { "155%", OPTION_VAL_155_PT_PT },
         { "160%", OPTION_VAL_160_PT_PT },
         { "165%", OPTION_VAL_165_PT_PT },
         { "170%", OPTION_VAL_170_PT_PT },
         { "175%", OPTION_VAL_175_PT_PT },
         { "180%", OPTION_VAL_180_PT_PT },
         { "185%", OPTION_VAL_185_PT_PT },
         { "190%", OPTION_VAL_190_PT_PT },
         { "195%", OPTION_VAL_195_PT_PT },
         { "200%", OPTION_VAL_200_PT_PT },
         { NULL, NULL },
      },
      "100%"
   },
   {
      BEETLE_OPT(negcon_response),
      BEETLE_OPT_NEGCON_RESPONSE_LABEL_PT_PT,
      NULL,
      BEETLE_OPT_NEGCON_RESPONSE_INFO_0_PT_PT,
      NULL,
      "input",
      {
         { "linear",    OPTION_VAL_LINEAR_PT_PT },
         { "quadratic", OPTION_VAL_QUADRATIC_PT_PT },
         { "cubic",     OPTION_VAL_CUBIC_PT_PT },
         { NULL, NULL },
      },
      "linear"
   },
   {
      BEETLE_OPT(negcon_deadzone),
      BEETLE_OPT_NEGCON_DEADZONE_LABEL_PT_PT,
      NULL,
      BEETLE_OPT_NEGCON_DEADZONE_INFO_0_PT_PT,
      NULL,
      "input",
      {
         { "0%",  OPTION_VAL_0_O44_PT_PT },
         { "5%",  OPTION_VAL_5_O42_PT_PT },
         { "10%", OPTION_VAL_10_PT_PT },
         { "15%", OPTION_VAL_15_PT_PT },
         { "20%", OPTION_VAL_20_PT_PT },
         { "25%", OPTION_VAL_25_PT_PT },
         { "30%", OPTION_VAL_30_PT_PT },
         { NULL, NULL },
      },
      "0%"
   },
   {
      BEETLE_OPT(use_mednafen_memcard0_method),
      BEETLE_OPT_USE_MEDNAFEN_MEMCARD0_METHOD_LABEL_PT_PT,
      NULL,
      BEETLE_OPT_USE_MEDNAFEN_MEMCARD0_METHOD_INFO_0_PT_PT,
      NULL,
      "memcards",
      {
         { "libretro", OPTION_VAL_LIBRETRO_PT_PT },
         { "mednafen", OPTION_VAL_MEDNAFEN_PT_PT },
         { NULL, NULL },
      },
      "libretro"
   },
   {
      BEETLE_OPT(enable_memcard1),
      BEETLE_OPT_ENABLE_MEMCARD1_LABEL_PT_PT,
      NULL,
      BEETLE_OPT_ENABLE_MEMCARD1_INFO_0_PT_PT,
      NULL,
      "memcards",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(shared_memory_cards),
      BEETLE_OPT_SHARED_MEMORY_CARDS_LABEL_PT_PT,
      NULL,
      BEETLE_OPT_SHARED_MEMORY_CARDS_INFO_0_PT_PT,
      NULL,
      "memcards",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(memcard_left_index),
      BEETLE_OPT_MEMCARD_LEFT_INDEX_LABEL_PT_PT,
      NULL,
      BEETLE_OPT_MEMCARD_LEFT_INDEX_INFO_0_PT_PT,
      NULL,
      "memcards",
      {
         { "0",  OPTION_VAL_0_O48_PT_PT },
         { "1",  NULL },
         { "2",  NULL },
         { "3",  NULL },
         { "4",  NULL },
         { "5",  NULL },
         { "6",  NULL },
         { "7",  NULL },
         { "8",  NULL },
         { "9",  NULL },
         { "10", NULL },
         { "11", NULL },
         { "12", NULL },
         { "13", NULL },
         { "14", NULL },
         { "15", NULL },
         { "16", NULL },
         { "17", NULL },
         { "18", NULL },
         { "19", NULL },
         { "20", NULL },
         { "21", NULL },
         { "22", NULL },
         { "23", NULL },
         { "24", NULL },
         { "25", NULL },
         { "26", NULL },
         { "27", NULL },
         { "28", NULL },
         { "29", NULL },
         { "30", NULL },
         { "31", NULL },
         { "32", NULL },
         { "33", NULL },
         { "34", NULL },
         { "35", NULL },
         { "36", NULL },
         { "37", NULL },
         { "38", NULL },
         { "39", NULL },
         { "40", NULL },
         { "41", NULL },
         { "42", NULL },
         { "43", NULL },
         { "44", NULL },
         { "45", NULL },
         { "46", NULL },
         { "47", NULL },
         { "48", NULL },
         { "49", NULL },
         { "50", NULL },
         { "51", NULL },
         { "52", NULL },
         { "53", NULL },
         { "54", NULL },
         { "55", NULL },
         { "56", NULL },
         { "57", NULL },
         { "58", NULL },
         { "59", NULL },
         { "60", NULL },
         { "61", NULL },
         { "62", NULL },
         { "63", NULL },
         { NULL, NULL },
      },
      "0"
   },
   {
      BEETLE_OPT(memcard_right_index),
      BEETLE_OPT_MEMCARD_RIGHT_INDEX_LABEL_PT_PT,
      NULL,
      BEETLE_OPT_MEMCARD_RIGHT_INDEX_INFO_0_PT_PT,
      NULL,
      "memcards",
      {
         { "0",  NULL },
         { "1",  OPTION_VAL_1_O49_PT_PT },
         { "2",  NULL },
         { "3",  NULL },
         { "4",  NULL },
         { "5",  NULL },
         { "6",  NULL },
         { "7",  NULL },
         { "8",  NULL },
         { "9",  NULL },
         { "10", NULL },
         { "11", NULL },
         { "12", NULL },
         { "13", NULL },
         { "14", NULL },
         { "15", NULL },
         { "16", NULL },
         { "17", NULL },
         { "18", NULL },
         { "19", NULL },
         { "20", NULL },
         { "21", NULL },
         { "22", NULL },
         { "23", NULL },
         { "24", NULL },
         { "25", NULL },
         { "26", NULL },
         { "27", NULL },
         { "28", NULL },
         { "29", NULL },
         { "30", NULL },
         { "31", NULL },
         { "32", NULL },
         { "33", NULL },
         { "34", NULL },
         { "35", NULL },
         { "36", NULL },
         { "37", NULL },
         { "38", NULL },
         { "39", NULL },
         { "40", NULL },
         { "41", NULL },
         { "42", NULL },
         { "43", NULL },
         { "44", NULL },
         { "45", NULL },
         { "46", NULL },
         { "47", NULL },
         { "48", NULL },
         { "49", NULL },
         { "50", NULL },
         { "51", NULL },
         { "52", NULL },
         { "53", NULL },
         { "54", NULL },
         { "55", NULL },
         { "56", NULL },
         { "57", NULL },
         { "58", NULL },
         { "59", NULL },
         { "60", NULL },
         { "61", NULL },
         { "62", NULL },
         { "63", NULL },
         { NULL, NULL },
      },
      "1"
   },
   {
      BEETLE_OPT(line_render),
      BEETLE_OPT_LINE_RENDER_LABEL_PT_PT,
      NULL,
      BEETLE_OPT_LINE_RENDER_INFO_0_PT_PT,
      NULL,
      "hacks",
      {
         { "default",    OPTION_VAL_DEFAULT_PT_PT },
         { "aggressive", OPTION_VAL_AGGRESSIVE_PT_PT },
         { "disabled",   NULL },
         { NULL, NULL },
      },
      "default"
   },
   {
      BEETLE_OPT(widescreen_hack),
      BEETLE_OPT_WIDESCREEN_HACK_LABEL_PT_PT,
      NULL,
      BEETLE_OPT_WIDESCREEN_HACK_INFO_0_PT_PT,
      NULL,
      "hacks",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(widescreen_hack_aspect_ratio),
      BEETLE_OPT_WIDESCREEN_HACK_ASPECT_RATIO_LABEL_PT_PT,
      NULL,
      BEETLE_OPT_WIDESCREEN_HACK_ASPECT_RATIO_INFO_0_PT_PT,
      NULL,
      "hacks",
      {
         { "16:9",  OPTION_VAL_16_9_PT_PT },
         { "16:10", OPTION_VAL_16_10_PT_PT },
         { "18:9",  OPTION_VAL_18_9_PT_PT },
         { "19:9",  OPTION_VAL_19_9_PT_PT },
         { "20:9",  OPTION_VAL_20_9_PT_PT },
         { "21:9",  OPTION_VAL_21_9_PT_PT }, // 64:27
         { "32:9",  OPTION_VAL_32_9_PT_PT },
         { NULL,    NULL },
      },
      "16:9"
   },
#if defined(HAVE_LIGHTREC)
   {
      BEETLE_OPT(cpu_dynarec),
      BEETLE_OPT_CPU_DYNAREC_LABEL_PT_PT,
      NULL,
      BEETLE_OPT_CPU_DYNAREC_INFO_0_PT_PT,
      NULL,
      "hacks",
      {
         { "disabled",        OPTION_VAL_DISABLED_PT_PT },
         { "execute",         OPTION_VAL_EXECUTE_PT_PT },
         { "execute_one",     OPTION_VAL_EXECUTE_ONE_PT_PT },
         { "run_interpreter", OPTION_VAL_RUN_INTERPRETER_PT_PT },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(dynarec_invalidate),
      BEETLE_OPT_DYNAREC_INVALIDATE_LABEL_PT_PT,
      NULL,
      BEETLE_OPT_DYNAREC_INVALIDATE_INFO_0_PT_PT,
      NULL,
      "hacks",
      {
         { "full", OPTION_VAL_FULL_PT_PT },
         { "dma",  OPTION_VAL_DMA_PT_PT },
         { NULL, NULL },
      },
      "full"
   },
   {
      BEETLE_OPT(dynarec_eventcycles),
      BEETLE_OPT_DYNAREC_EVENTCYCLES_LABEL_PT_PT,
      NULL,
      BEETLE_OPT_DYNAREC_EVENTCYCLES_INFO_0_PT_PT,
      NULL,
      "hacks",
      {
         { "128",  OPTION_VAL_128_PT_PT },
         { "256",  NULL },
         { "384",  NULL },
         { "512",  NULL },
         { "640",  NULL },
         { "768",  NULL },
         { "896",  NULL },
         { "1024", NULL },
         { "1152", NULL },
         { "1280", NULL },
         { "1408", NULL },
         { "1536", NULL },
         { "1664", NULL },
         { "1792", NULL },
         { "1920", NULL },
         { "2048", NULL },
         { NULL, NULL },
      },
      "128"
   },
   {
      BEETLE_OPT(dynarec_spu_samples),
      BEETLE_OPT_DYNAREC_SPU_SAMPLES_LABEL_PT_PT,
      NULL,
      BEETLE_OPT_DYNAREC_SPU_SAMPLES_INFO_0_PT_PT,
      NULL,
      "hacks",
      {
         { "1",  OPTION_VAL_1_O49_PT_PT },
         { "4",  NULL },
         { "16", NULL },
         { NULL, NULL },
      },
      "1"
   },
#endif
   {
      BEETLE_OPT(core_timing_fps),
      BEETLE_OPT_CORE_TIMING_FPS_LABEL_PT_PT,
      NULL,
      BEETLE_OPT_CORE_TIMING_FPS_INFO_0_PT_PT,
      NULL,
      "video",
      {
         { "auto_toggle",       OPTION_VAL_AUTO_TOGGLE_PT_PT },
         { "force_progressive", OPTION_VAL_FORCE_PROGRESSIVE_PT_PT },
         { "force_interlaced",  OPTION_VAL_FORCE_INTERLACED_PT_PT },
      },
      "auto_toggle"
   },
   {
      BEETLE_OPT(deinterlacer),
      BEETLE_OPT_DEINTERLACER_LABEL_PT_PT,
      NULL,
      BEETLE_OPT_DEINTERLACER_INFO_0_PT_PT,
      NULL,
      "video",
      {
         { "weave",      OPTION_VAL_WEAVE_PT_PT },
         { "bob",        OPTION_VAL_BOB_PT_PT },
         { "bob_offset", OPTION_VAL_BOB_OFFSET_PT_PT },
         { "fastmad",    OPTION_VAL_FASTMAD_PT_PT },
         { "off",        "Off" },
         { NULL, NULL },
      },
      "weave"
   },
   {
      BEETLE_OPT(frame_duping),
      BEETLE_OPT_FRAME_DUPING_LABEL_PT_PT,
      NULL,
      BEETLE_OPT_FRAME_DUPING_INFO_0_PT_PT,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(pal_video_timing_override),
      BEETLE_OPT_PAL_VIDEO_TIMING_OVERRIDE_LABEL_PT_PT,
      NULL,
      BEETLE_OPT_PAL_VIDEO_TIMING_OVERRIDE_INFO_0_PT_PT,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(aspect_ratio),
      BEETLE_OPT_ASPECT_RATIO_LABEL_PT_PT,
      NULL,
      BEETLE_OPT_ASPECT_RATIO_INFO_0_PT_PT,
      NULL,
      "video",
      {
         { "corrected", OPTION_VAL_CORRECTED_PT_PT },
         { "uncorrected", OPTION_VAL_UNCORRECTED_PT_PT },
         { "4:3",  OPTION_VAL_4_3_PT_PT },
         { "ntsc", OPTION_VAL_NTSC_PT_PT },
      },
      "corrected"
   },
   {
      BEETLE_OPT(crop_overscan),
      BEETLE_OPT_CROP_OVERSCAN_LABEL_PT_PT,
      NULL,
      BEETLE_OPT_CROP_OVERSCAN_INFO_0_PT_PT,
      NULL,
      "video",
      {
         { "disabled", "Disabled" },
         { "static",   OPTION_VAL_STATIC_PT_PT },
         { "smart",    OPTION_VAL_SMART_PT_PT },
         { NULL, NULL },
      },
      "smart"
   },
   {
      BEETLE_OPT(image_crop),
      BEETLE_OPT_IMAGE_CROP_LABEL_PT_PT,
      NULL,
      BEETLE_OPT_IMAGE_CROP_INFO_0_PT_PT,
      NULL,
      "video",
      {
         { "disabled", "0" },
         { "1px",      OPTION_VAL_1PX_PT_PT },
         { "2px",      OPTION_VAL_2PX_PT_PT },
         { "3px",      OPTION_VAL_3PX_PT_PT },
         { "4px",      OPTION_VAL_4PX_PT_PT },
         { "5px",      OPTION_VAL_5PX_PT_PT },
         { "6px",      OPTION_VAL_6PX_PT_PT },
         { "7px",      OPTION_VAL_7PX_PT_PT },
         { "8px",      OPTION_VAL_8PX_PT_PT },
         { "9px",      OPTION_VAL_9PX_PT_PT },
         { "10px",     OPTION_VAL_10PX_PT_PT },
         { "11px",     OPTION_VAL_11PX_PT_PT },
         { "12px",     OPTION_VAL_12PX_PT_PT },
         { "13px",     OPTION_VAL_13PX_PT_PT },
         { "14px",     OPTION_VAL_14PX_PT_PT },
         { "15px",     OPTION_VAL_15PX_PT_PT },
         { "16px",     OPTION_VAL_16PX_PT_PT },
         { "17px",     OPTION_VAL_17PX_PT_PT },
         { "18px",     OPTION_VAL_18PX_PT_PT },
         { "19px",     OPTION_VAL_19PX_PT_PT },
         { "20px",     OPTION_VAL_20PX_PT_PT },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(image_offset),
      BEETLE_OPT_IMAGE_OFFSET_LABEL_PT_PT,
      NULL,
      BEETLE_OPT_IMAGE_OFFSET_INFO_0_PT_PT,
      NULL,
      "video",
      {
         { "-12px",    OPTION_VAL_12PX_O64_PT_PT },
         { "-11px",    OPTION_VAL_11PX_O64_PT_PT },
         { "-10px",    OPTION_VAL_10PX_O64_PT_PT },
         { "-9px",     OPTION_VAL_9PX_O64_PT_PT },
         { "-8px",     OPTION_VAL_8PX_O64_PT_PT },
         { "-7px",     OPTION_VAL_7PX_O64_PT_PT },
         { "-6px",     OPTION_VAL_6PX_O64_PT_PT },
         { "-5px",     OPTION_VAL_5PX_O64_PT_PT },
         { "-4px",     OPTION_VAL_4PX_O64_PT_PT },
         { "-3px",     OPTION_VAL_3PX_O64_PT_PT },
         { "-2px",     OPTION_VAL_2PX_O64_PT_PT },
         { "-1px",     OPTION_VAL_1PX_O64_PT_PT },
         { "disabled", "0" },
         { "+1px",     OPTION_VAL_1PX_O64_0_PT_PT },
         { "+2px",     OPTION_VAL_2PX_O64_0_PT_PT },
         { "+3px",     OPTION_VAL_3PX_O64_0_PT_PT },
         { "+4px",     OPTION_VAL_4PX_O64_0_PT_PT },
         { "+5px",     OPTION_VAL_5PX_O64_0_PT_PT },
         { "+6px",     OPTION_VAL_6PX_O64_0_PT_PT },
         { "+7px",     OPTION_VAL_7PX_O64_0_PT_PT },
         { "+8px",     OPTION_VAL_8PX_O64_0_PT_PT },
         { "+9px",     OPTION_VAL_9PX_O64_0_PT_PT },
         { "+10px",    OPTION_VAL_10PX_O64_0_PT_PT },
         { "+11px",    OPTION_VAL_11PX_O64_0_PT_PT },
         { "+12px",    OPTION_VAL_12PX_O64_0_PT_PT },
         { NULL, NULL },
      },
      "disabled"
   },
#if defined(HAVE_OPENGL) || defined(HAVE_OPENGLES) || defined(HAVE_VULKAN)
   {
      BEETLE_OPT(image_offset_cycles),
      BEETLE_OPT_IMAGE_OFFSET_CYCLES_LABEL_PT_PT,
      NULL,
      BEETLE_OPT_IMAGE_OFFSET_CYCLES_INFO_0_PT_PT,
      NULL,
      "video",
      {
         { "-40",      NULL },
         { "-39",      NULL },
         { "-38",      NULL },
         { "-37",      NULL },
         { "-36",      NULL },
         { "-35",      NULL },
         { "-34",      NULL },
         { "-33",      NULL },
         { "-32",      NULL },
         { "-31",      NULL },
         { "-30",      NULL },
         { "-29",      NULL },
         { "-28",      NULL },
         { "-27",      NULL },
         { "-26",      NULL },
         { "-25",      NULL },
         { "-24",      NULL },
         { "-23",      NULL },
         { "-22",      NULL },
         { "-21",      NULL },
         { "-20",      NULL },
         { "-19",      NULL },
         { "-18",      NULL },
         { "-17",      NULL },
         { "-16",      NULL },
         { "-15",      NULL },
         { "-14",      NULL },
         { "-13",      NULL },
         { "-12",      NULL },
         { "-11",      NULL },
         { "-10",      NULL },
         { "-9",       NULL },
         { "-8",       NULL },
         { "-7",       NULL },
         { "-6",       NULL },
         { "-5",       NULL },
         { "-4",       NULL },
         { "-3",       NULL },
         { "-2",       NULL },
         { "-1",       NULL },
         { "0",        "0" },
         { "+1",       NULL },
         { "+2",       NULL },
         { "+3",       NULL },
         { "+4",       NULL },
         { "+5",       NULL },
         { "+6",       NULL },
         { "+7",       NULL },
         { "+8",       NULL },
         { "+9",       NULL },
         { "+10",      NULL },
         { "+11",      NULL },
         { "+12",      NULL },
         { "+13",      NULL },
         { "+14",      NULL },
         { "+15",      NULL },
         { "+16",      NULL },
         { "+17",      NULL },
         { "+18",      NULL },
         { "+19",      NULL },
         { "+20",      NULL },
         { "+21",      NULL },
         { "+22",      NULL },
         { "+23",      NULL },
         { "+24",      NULL },
         { "+25",      NULL },
         { "+26",      NULL },
         { "+27",      NULL },
         { "+28",      NULL },
         { "+29",      NULL },
         { "+30",      NULL },
         { "+31",      NULL },
         { "+32",      NULL },
         { "+33",      NULL },
         { "+34",      NULL },
         { "+35",      NULL },
         { "+36",      NULL },
         { "+37",      NULL },
         { "+38",      NULL },
         { "+39",      NULL },
         { "+40",      NULL },
         { NULL, NULL},
      },
      "0"
   },
#endif
   {
      BEETLE_OPT(initial_scanline),
      BEETLE_OPT_INITIAL_SCANLINE_LABEL_PT_PT,
      NULL,
      BEETLE_OPT_INITIAL_SCANLINE_INFO_0_PT_PT,
      NULL,
      "video",
      {
         { "0",  OPTION_VAL_0_O48_PT_PT },
         { "1",  NULL },
         { "2",  NULL },
         { "3",  NULL },
         { "4",  NULL },
         { "5",  NULL },
         { "6",  NULL },
         { "7",  NULL },
         { "8",  NULL },
         { "9",  NULL },
         { "10", NULL },
         { "11", NULL },
         { "12", NULL },
         { "13", NULL },
         { "14", NULL },
         { "15", NULL },
         { "16", NULL },
         { "17", NULL },
         { "18", NULL },
         { "19", NULL },
         { "20", NULL },
         { "21", NULL },
         { "22", NULL },
         { "23", NULL },
         { "24", NULL },
         { "25", NULL },
         { "26", NULL },
         { "27", NULL },
         { "28", NULL },
         { "29", NULL },
         { "30", NULL },
         { "31", NULL },
         { "32", NULL },
         { "33", NULL },
         { "34", NULL },
         { "35", NULL },
         { "36", NULL },
         { "37", NULL },
         { "38", NULL },
         { "39", NULL },
         { "40", NULL },
         { NULL, NULL },
      },
      "0"
   },
   {
      BEETLE_OPT(last_scanline),
      BEETLE_OPT_LAST_SCANLINE_LABEL_PT_PT,
      NULL,
      BEETLE_OPT_LAST_SCANLINE_INFO_0_PT_PT,
      NULL,
      "video",
      {
         { "210", NULL },
         { "211", NULL },
         { "212", NULL },
         { "213", NULL },
         { "214", NULL },
         { "215", NULL },
         { "216", NULL },
         { "217", NULL },
         { "218", NULL },
         { "219", NULL },
         { "220", NULL },
         { "221", NULL },
         { "222", NULL },
         { "223", NULL },
         { "224", NULL },
         { "225", NULL },
         { "226", NULL },
         { "227", NULL },
         { "228", NULL },
         { "229", NULL },
         { "230", NULL },
         { "231", NULL },
         { "232", NULL },
         { "233", NULL },
         { "234", NULL },
         { "235", NULL },
         { "236", NULL },
         { "237", NULL },
         { "238", NULL },
         { "239", OPTION_VAL_239_PT_PT },
         { NULL, NULL },
      },
      "239"
   },
   {
      BEETLE_OPT(initial_scanline_pal),
      BEETLE_OPT_INITIAL_SCANLINE_PAL_LABEL_PT_PT,
      NULL,
      BEETLE_OPT_INITIAL_SCANLINE_PAL_INFO_0_PT_PT,
      NULL,
      "video",
      {
         { "0",  OPTION_VAL_0_O48_PT_PT },
         { "1",  NULL },
         { "2",  NULL },
         { "3",  NULL },
         { "4",  NULL },
         { "5",  NULL },
         { "6",  NULL },
         { "7",  NULL },
         { "8",  NULL },
         { "9",  NULL },
         { "10", NULL },
         { "11", NULL },
         { "12", NULL },
         { "13", NULL },
         { "14", NULL },
         { "15", NULL },
         { "16", NULL },
         { "17", NULL },
         { "18", NULL },
         { "19", NULL },
         { "20", NULL },
         { "21", NULL },
         { "22", NULL },
         { "23", NULL },
         { "24", NULL },
         { "25", NULL },
         { "26", NULL },
         { "27", NULL },
         { "28", NULL },
         { "29", NULL },
         { "30", NULL },
         { "31", NULL },
         { "32", NULL },
         { "33", NULL },
         { "34", NULL },
         { "35", NULL },
         { "36", NULL },
         { "37", NULL },
         { "38", NULL },
         { "39", NULL },
         { "40", NULL },
         { NULL, NULL },
      },
      "0"
   },
   {
      BEETLE_OPT(last_scanline_pal),
      BEETLE_OPT_LAST_SCANLINE_PAL_LABEL_PT_PT,
      NULL,
      BEETLE_OPT_LAST_SCANLINE_PAL_INFO_0_PT_PT,
      NULL,
      "video",
      {
         { "230", NULL },
         { "231", NULL },
         { "232", NULL },
         { "233", NULL },
         { "234", NULL },
         { "235", NULL },
         { "236", NULL },
         { "237", NULL },
         { "238", NULL },
         { "239", NULL },
         { "240", NULL },
         { "241", NULL },
         { "242", NULL },
         { "243", NULL },
         { "244", NULL },
         { "245", NULL },
         { "246", NULL },
         { "247", NULL },
         { "248", NULL },
         { "249", NULL },
         { "250", NULL },
         { "251", NULL },
         { "252", NULL },
         { "253", NULL },
         { "254", NULL },
         { "255", NULL },
         { "256", NULL },
         { "257", NULL },
         { "258", NULL },
         { "259", NULL },
         { "260", NULL },
         { "261", NULL },
         { "262", NULL },
         { "263", NULL },
         { "264", NULL },
         { "265", NULL },
         { "266", NULL },
         { "267", NULL },
         { "268", NULL },
         { "269", NULL },
         { "270", NULL },
         { "271", NULL },
         { "272", NULL },
         { "273", NULL },
         { "274", NULL },
         { "275", NULL },
         { "276", NULL },
         { "277", NULL },
         { "278", NULL },
         { "279", NULL },
         { "280", NULL },
         { "281", NULL },
         { "282", NULL },
         { "283", NULL },
         { "284", NULL },
         { "285", NULL },
         { "286", NULL },
         { "287", OPTION_VAL_287_PT_PT },
         { NULL, NULL },
      },
      "287"
   },
   { NULL, NULL, NULL, NULL, NULL, NULL, {{0}}, NULL },
};
struct retro_core_options_v2 options_pt_pt = {
   option_cats_pt_pt,
   option_defs_pt_pt
};

/* RETRO_LANGUAGE_RU */

#define CATEGORY_SYSTEM_LABEL_RU "Система"
#define CATEGORY_SYSTEM_INFO_0_RU "Изменение параметров системы и разгона."
#define CATEGORY_VIDEO_LABEL_RU "Видео"
#define CATEGORY_VIDEO_INFO_0_RU "Настройки соотношения сторон, кадрирования, графических фильтров и пропуска кадров."
#define CATEGORY_PGXP_LABEL_RU NULL
#define CATEGORY_PGXP_INFO_0_RU "Настройки для улучшения качества графики выше уровня оригинальной консоли. PGXP помогает избавиться от Z-конфликтов и деформации текстур."
#define CATEGORY_OSD_LABEL_RU "Индикация на экране"
#define CATEGORY_OSD_INFO_0_RU "Настройка отображаемых на экране уведомлений."
#define CATEGORY_INPUT_LABEL_RU "Ввод"
#define CATEGORY_INPUT_INFO_0_RU "Настройки светового пистолета, мыши и neGcon."
#define CATEGORY_MEMCARDS_LABEL_RU "Карты памяти"
#define CATEGORY_MEMCARDS_INFO_0_RU "Изменение настроек виртуальной карты памяти."
#define CATEGORY_HACKS_LABEL_RU "Хаки эмуляции"
#define CATEGORY_HACKS_INFO_0_RU "Изменение настроек точности эмуляции, влияющих на низкую производительность и совместимость."
#define BEETLE_OPT_CPU_FREQ_SCALE_LABEL_RU "Масштабирование частоты CPU (разгон)"
#define BEETLE_OPT_CPU_FREQ_SCALE_INFO_0_RU "Разгон (или занижение) частоты CPU эмулируемой PSX. В некоторых играх может устранить замедления и выровнять частоту кадров, но повышает нагрузку. Часть игр не получит преимуществ из-за наличия внутр. ограничителей частоты. Также может привести к завышению скорости анимаций."
#define OPTION_VAL_50_RU NULL
#define OPTION_VAL_60_RU NULL
#define OPTION_VAL_70_RU NULL
#define OPTION_VAL_80_RU NULL
#define OPTION_VAL_90_RU NULL
#define OPTION_VAL_100_RU "100% (нативно)"
#define OPTION_VAL_110_RU NULL
#define OPTION_VAL_120_RU NULL
#define OPTION_VAL_130_RU NULL
#define OPTION_VAL_140_RU NULL
#define OPTION_VAL_150_RU NULL
#define OPTION_VAL_160_RU NULL
#define OPTION_VAL_170_RU NULL
#define OPTION_VAL_180_RU NULL
#define OPTION_VAL_190_RU NULL
#define OPTION_VAL_200_RU NULL
#define OPTION_VAL_210_RU NULL
#define OPTION_VAL_220_RU NULL
#define OPTION_VAL_230_RU NULL
#define OPTION_VAL_240_RU NULL
#define OPTION_VAL_250_RU NULL
#define OPTION_VAL_260_RU NULL
#define OPTION_VAL_270_RU NULL
#define OPTION_VAL_280_RU NULL
#define OPTION_VAL_290_RU NULL
#define OPTION_VAL_300_RU NULL
#define OPTION_VAL_310_RU NULL
#define OPTION_VAL_320_RU NULL
#define OPTION_VAL_330_RU NULL
#define OPTION_VAL_340_RU NULL
#define OPTION_VAL_350_RU NULL
#define OPTION_VAL_360_RU NULL
#define OPTION_VAL_370_RU NULL
#define OPTION_VAL_380_RU NULL
#define OPTION_VAL_390_RU NULL
#define OPTION_VAL_400_RU NULL
#define OPTION_VAL_410_RU NULL
#define OPTION_VAL_420_RU NULL
#define OPTION_VAL_430_RU NULL
#define OPTION_VAL_440_RU NULL
#define OPTION_VAL_450_RU NULL
#define OPTION_VAL_460_RU NULL
#define OPTION_VAL_470_RU NULL
#define OPTION_VAL_480_RU NULL
#define OPTION_VAL_490_RU NULL
#define OPTION_VAL_500_RU NULL
#define OPTION_VAL_510_RU NULL
#define OPTION_VAL_520_RU NULL
#define OPTION_VAL_530_RU NULL
#define OPTION_VAL_540_RU NULL
#define OPTION_VAL_550_RU NULL
#define OPTION_VAL_560_RU NULL
#define OPTION_VAL_570_RU NULL
#define OPTION_VAL_580_RU NULL
#define OPTION_VAL_590_RU NULL
#define OPTION_VAL_600_RU NULL
#define OPTION_VAL_610_RU NULL
#define OPTION_VAL_620_RU NULL
#define OPTION_VAL_630_RU NULL
#define OPTION_VAL_640_RU NULL
#define OPTION_VAL_650_RU NULL
#define OPTION_VAL_660_RU NULL
#define OPTION_VAL_670_RU NULL
#define OPTION_VAL_680_RU NULL
#define OPTION_VAL_690_RU NULL
#define OPTION_VAL_700_RU NULL
#define OPTION_VAL_710_RU NULL
#define OPTION_VAL_720_RU NULL
#define OPTION_VAL_730_RU NULL
#define OPTION_VAL_740_RU NULL
#define OPTION_VAL_750_RU NULL
#define BEETLE_OPT_GPU_OVERCLOCK_LABEL_RU "Разгон GPU-растеризатора"
#define BEETLE_OPT_GPU_OVERCLOCK_INFO_0_RU "Разгоняет 2D-растеризатор внутри GPU эмулируемой PSX. Не влияет на качество 3D-рендеринга и в целом не оказывает значимый эффект."
#define OPTION_VAL_1X_NATIVE_RU "1x (нативно)"
#define OPTION_VAL_2X_RU NULL
#define OPTION_VAL_4X_RU NULL
#define OPTION_VAL_8X_RU NULL
#define OPTION_VAL_16X_RU NULL
#define OPTION_VAL_32X_RU NULL
#define BEETLE_OPT_GTE_OVERCLOCK_LABEL_RU "Разгон GTE"
#define BEETLE_OPT_GTE_OVERCLOCK_INFO_0_RU "Понижает задержку выполнения операций эмулируемого GTE (сопроцессора для 3D-графики) до одного цикла. Поднимает частоту кадров и стабилизирует время кадра в играх, интенсивно использующих GTE."
#define BEETLE_OPT_CD_ACCESS_METHOD_LABEL_RU "Метод доступа CD"
#define BEETLE_OPT_CD_ACCESS_METHOD_INFO_0_RU NULL
#define OPTION_VAL_SYNC_RU "Синхронный"
#define OPTION_VAL_ASYNC_RU "Асинхронный"
#define OPTION_VAL_PRECACHE_RU "Предзагрузка"
#define BEETLE_OPT_CD_FASTLOAD_LABEL_RU "Скорость чтения CD"
#define BEETLE_OPT_CD_FASTLOAD_INFO_0_RU NULL
#define OPTION_VAL_2X_NATIVE_RU "2x (нативно)"
#define OPTION_VAL_6X_RU NULL
#define OPTION_VAL_10X_RU NULL
#define OPTION_VAL_12X_RU NULL
#define OPTION_VAL_14X_RU NULL
#define BEETLE_OPT_SKIP_BIOS_LABEL_RU "Пропускать BIOS"
#define BEETLE_OPT_SKIP_BIOS_INFO_0_RU "Пропуск начальной анимации BIOS PlayStation, отображаемой при запуске игр. При включении может влиять на совместимость с некоторыми играми (Saga Frontier, PAL-игры с защитой от копирования и др.)."
#define BEETLE_OPT_OVERRIDE_BIOS_LABEL_RU "Переопределение настроек BIOS"
#define BEETLE_OPT_OVERRIDE_BIOS_INFO_0_RU "Заменить стандартный региональный BIOS на образ без региональной привязки, если он будет обнаружен. Требуется перезагрузка."
#define OPTION_VAL_PSXONPSP_RU NULL
#define OPTION_VAL_PS1_ROM_RU NULL
#define OPTION_VAL_OPENBIOS_RU NULL
#define BEETLE_OPT_REGION_LABEL_RU "Регион системы"
#define BEETLE_OPT_REGION_INFO_0_RU NULL
#define OPTION_VAL_AUTO_RU "Автоматически"
#define OPTION_VAL_NTSC_J_RU "NTSC-J (Япония)"
#define OPTION_VAL_NTSC_U_RU "NTSC-U (Северная Америка)"
#define OPTION_VAL_PAL_RU "PAL (Европа)"
#define BEETLE_OPT_INTERNAL_RESOLUTION_LABEL_RU "Внутреннее разрешение GPU"
#define BEETLE_OPT_INTERNAL_RESOLUTION_INFO_0_RU NULL
#define BEETLE_OPT_RENDERER_LABEL_RU "Рендер"
#define BEETLE_OPT_RENDERER_INFO_0_RU NULL
#define OPTION_VAL_HARDWARE_RU "Аппаратный (авто)"
#define OPTION_VAL_HARDWARE_GL_RU "Аппаратный (OpenGL)"
#define OPTION_VAL_HARDWARE_VK_RU "Аппаратный (Vulkan)"
#define OPTION_VAL_SOFTWARE_RU "Программный"
#define BEETLE_OPT_RENDERER_SOFTWARE_FB_LABEL_RU "Программный фреймбуфер"
#define BEETLE_OPT_RENDERER_SOFTWARE_FB_INFO_0_RU "Обеспечивает для аппаратных рендеров точную эмуляцию эффектов фреймбуфера (напр. размытие движения, закручивание перед боем в FF7), запуская фоном копию программного рендера в исходном разрешении. При откл. данные эффекты пропускаются (OpenGL) или выполняются посредством GPU (Vulkan). Отключение повышает скорость, но вызывает ошибки графики. Если не уверены, оставьте вкл."
#define BEETLE_OPT_DEPTH_LABEL_RU "Внутренняя глубина цвета"
#define BEETLE_OPT_DEPTH_INFO_0_RU NULL
#define OPTION_VAL_16BPP_NATIVE_RU "16 бит/пикс (нативно)"
#define OPTION_VAL_32BPP_RU "32 бит/пикс"
#define BEETLE_OPT_DITHER_MODE_LABEL_RU "Паттерн дизеринга"
#define BEETLE_OPT_DITHER_MODE_INFO_0_RU NULL
#define OPTION_VAL_INTERNAL_RESOLUTION_RU "Внутреннее разрешение"
#define BEETLE_OPT_SCALED_UV_OFFSET_LABEL_RU "Смещение текстур по UV"
#define BEETLE_OPT_SCALED_UV_OFFSET_INFO_0_RU "Сэмплировать текстуры для 3D-полигонов со смещением при внутреннем разрешении выше 1x. Уменьшает видимость швов текстур, но может вызывать произвольные проблемы с графикой."
#define BEETLE_OPT_FILTER_LABEL_RU "Сглаживание текстур"
#define BEETLE_OPT_FILTER_INFO_0_RU NULL
#define OPTION_VAL_NEAREST_RU "Ближайший сосед"
#define OPTION_VAL_SABR_RU NULL
#define OPTION_VAL_XBR_RU NULL
#define OPTION_VAL_BILINEAR_RU "Билинейное"
#define OPTION_VAL_3_POINT_RU "По трём точкам"
#define OPTION_VAL_JINC2_RU NULL
#define BEETLE_OPT_FILTER_EXCLUDE_SPRITE_LABEL_RU "Исключать спрайты из фильтрации"
#define BEETLE_OPT_FILTER_EXCLUDE_SPRITE_INFO_0_RU "Не применять фильтрацию текстур к спрайтам. Предотвращает появление швов в играх с задним фоном, отрисованным 2D-спрайтами. Для макс. эффективности используйте совместно с адаптивным сглаживанием или иным фильтром постобработки."
#define OPTION_VAL_OPAQUE_RU "Только непрозрачные"
#define OPTION_VAL_ALL_RU "Непрозрачные и полупрозрачные"
#define BEETLE_OPT_FILTER_EXCLUDE_2D_POLYGON_LABEL_RU "Исключать 2D-полигоны из фильтрации"
#define BEETLE_OPT_FILTER_EXCLUDE_2D_POLYGON_INFO_0_RU "Не применять фильтрацию текстур к 2D-полигонам. 2D-полигоны определяются эвристически, поэтому возможны искажения графики. Для лучшего эффекта используйте вместе с опцией 'Адаптивное сглаживание' или другим фильтром пост-обработки."
#define BEETLE_OPT_ADAPTIVE_SMOOTHING_LABEL_RU "Адаптивное сглаживание"
#define BEETLE_OPT_ADAPTIVE_SMOOTHING_INFO_0_RU "Сглаживает 2D-элементы графики и интерфейса без потери чёткости объектов в 3D. Только для рендеринга на Vulkan."
#define BEETLE_OPT_SUPER_SAMPLING_LABEL_RU "Суперсэмплинг (понижать к исходному разрешению)"
#define BEETLE_OPT_SUPER_SAMPLING_INFO_0_RU "Понижает качество рендеринга с высокого разрешения до исходного. Совместно с повышением разрешения обеспечивает сглаженное 3D, но в ориг. разрешении. Наиболее эффективно для игр с 2D и 3D-элементами (напр. 3D-персонажи на отрисованном фоне), а также в связке с CRT-шейдерами. Поддерживается только рендером Vulkan. Примечание: при вкл. рекомендуется отключить опцию 'Паттерн дизеринга'."
#define BEETLE_OPT_MSAA_LABEL_RU "Множественная выборка сглаживания"
#define BEETLE_OPT_MSAA_INFO_0_RU "MSAA улучшает внешний вид 3D-объектов. Поддерживается только рендерингом на Vulkan."
#define OPTION_VAL_1X_RU NULL
#define BEETLE_OPT_MDEC_YUV_LABEL_RU "Сглаживание YUV MDEC"
#define BEETLE_OPT_MDEC_YUV_INFO_0_RU "Улучшает качество видеозаставок путём уменьшения артефактов макроблокирования (квадраты/неровные края). Только для рендеринга на Vulkan."
#define BEETLE_OPT_TRACK_TEXTURES_LABEL_RU "Отслеживать текстуры"
#define BEETLE_OPT_TRACK_TEXTURES_INFO_0_RU "Требуется для выгрузки и замещения текстур. В большинстве игр будет приводить к вылетам."
#define BEETLE_OPT_DUMP_TEXTURES_LABEL_RU "Снимать дамп текстур"
#define BEETLE_OPT_DUMP_TEXTURES_INFO_0_RU "Выгружать используемые текстуры в <cd>-texture-dump/"
#define BEETLE_OPT_REPLACE_TEXTURES_LABEL_RU "Замещать текстуры"
#define BEETLE_OPT_REPLACE_TEXTURES_INFO_0_RU "Заменять текстуры HD-копиями из <cd>-texture-replacements/"
#define BEETLE_OPT_PGXP_MODE_LABEL_RU "Режим выполнения PGXP"
#define BEETLE_OPT_PGXP_MODE_INFO_0_RU NULL
#define OPTION_VAL_MEMORY_ONLY_RU "Только память"
#define OPTION_VAL_MEMORY_CPU_RU "Память + CPU (нестабильно)"
#define BEETLE_OPT_PGXP_2D_TOL_LABEL_RU "Допуск PGXP для 2D-геометрии"
#define BEETLE_OPT_PGXP_2D_TOL_INFO_0_RU "Скрывает больше выраженных ошибок PGXP: настройка устанавливает допустимое отклонение значений PGXP для геометрии без точных данных о глубине."
#define OPTION_VAL_0PX_RU "0 пикс."
#define OPTION_VAL_1PX_RU "1 пикс."
#define OPTION_VAL_2PX_RU "2 пикс."
#define OPTION_VAL_3PX_RU "3 пикс."
#define OPTION_VAL_4PX_RU "4 пикс."
#define OPTION_VAL_5PX_RU "5 пикс."
#define OPTION_VAL_6PX_RU "6 пикс."
#define OPTION_VAL_7PX_RU "7 пикс."
#define OPTION_VAL_8PX_RU "8 пикс."
#define BEETLE_OPT_PGXP_NCLIP_LABEL_RU "Отбраковка примитивов PGXP"
#define BEETLE_OPT_PGXP_NCLIP_INFO_0_RU "Использовать PGXP-реализацию NCLIP. Улучшает графику за счёт уменьшения провалов в геометрии с помощью PGXP-координат. В некоторых случаях может приводить к зависанию игр."
#define BEETLE_OPT_PGXP_VERTEX_LABEL_RU "Вершинный кэш PGXP"
#define BEETLE_OPT_PGXP_VERTEX_INFO_0_RU "Кэшировать координаты вершин, улучшенных с PGXP для повторного исп. при отрисовке полигонов. Улучшает выравнивание объектов и уменьшает видимость швов при рендеринге текстур, но из-за ошибок при опросе кэша может вызывать ошибки графики. Рекомендуется оставлять выкл. Применяется только если вкл. 'Режим выполнения PGXP'. Поддерживается только аппаратными рендерами."
#define BEETLE_OPT_PGXP_TEXTURE_LABEL_RU "Перспективно-корректное текстурирование PGXP"
#define BEETLE_OPT_PGXP_TEXTURE_INFO_0_RU NULL
#define BEETLE_OPT_DISPLAY_INTERNAL_FPS_LABEL_RU "Показывать частоту кадров"
#define BEETLE_OPT_DISPLAY_INTERNAL_FPS_INFO_0_RU "Отображать внутреннюю частоту кадров, с которой эмулируемая PlayStation производит рендеринг. Примечание: требуется включение уведомлений в настройках фронтенда."
#define BEETLE_OPT_DISPLAY_OSD_LABEL_RU "Показывать уведомления"
#define BEETLE_OPT_DISPLAY_OSD_INFO_0_RU "Выводить на экран сообщения, генерируемые ядром."
#define BEETLE_OPT_DISPLAY_VRAM_LABEL_RU "Отображать всю VRAM (отладка)"
#define BEETLE_OPT_DISPLAY_VRAM_INFO_0_RU "Выводить на экран полное содержимое видеопамяти эмулируемой консоли. Поддерживается только для аппаратных рендеров OpenGL и Vulkan. Используется для отладки и в обычном режиме должно быть отключено."
#define BEETLE_OPT_ANALOG_CALIBRATION_LABEL_RU "Автокалибровка аналогового ввода"
#define BEETLE_OPT_ANALOG_CALIBRATION_INFO_0_RU NULL
#define BEETLE_OPT_ANALOG_TOGGLE_LABEL_RU "Переключатель режима DualShock"
#define BEETLE_OPT_ANALOG_TOGGLE_INFO_0_RU "При выборе устройства ввода типа DualShock данная опция позволяет переключать эмулируемый DualShock между цифровым и аналоговым режимами, как на реальной консоли. Режим также может быть изменён одной из комбинаций кнопок."
#define OPTION_VAL_ENABLED_ANALOG_RU "По умолчанию-аналоговый"
#define BEETLE_OPT_ANALOG_TOGGLE_COMBO_LABEL_RU "Связка для смены режима DualShock"
#define BEETLE_OPT_ANALOG_TOGGLE_COMBO_INFO_0_RU NULL
#define OPTION_VAL_L1_L2_R1_R2_START_SELECT_RU NULL
#define OPTION_VAL_L1_R1_SELECT_RU NULL
#define OPTION_VAL_L1_R1_START_RU NULL
#define OPTION_VAL_L1_R1_L3_RU NULL
#define OPTION_VAL_L1_R1_R3_RU NULL
#define OPTION_VAL_L2_R2_SELECT_RU NULL
#define OPTION_VAL_L2_R2_START_RU NULL
#define OPTION_VAL_L2_R2_L3_RU NULL
#define OPTION_VAL_L2_R2_R3_RU NULL
#define OPTION_VAL_L3_R3_RU NULL
#define BEETLE_OPT_ANALOG_TOGGLE_HOLD_LABEL_RU "Задержка нажатия для смены режима DualShock"
#define BEETLE_OPT_ANALOG_TOGGLE_HOLD_INFO_0_RU NULL
#define OPTION_VAL_0_RU "0 секунд"
#define OPTION_VAL_1_RU "1 секунда"
#define OPTION_VAL_2_RU "2 секунды"
#define OPTION_VAL_3_RU "3 секунды"
#define OPTION_VAL_4_RU "4 секунды"
#define OPTION_VAL_5_RU "5 секунд"
#define BEETLE_OPT_ENABLE_MULTITAP_PORT1_LABEL_RU "Мультитап на порт 1"
#define BEETLE_OPT_ENABLE_MULTITAP_PORT1_INFO_0_RU "Включает расширение мультитапа на порту 1."
#define BEETLE_OPT_ENABLE_MULTITAP_PORT2_LABEL_RU "Мультитап на порт 2"
#define BEETLE_OPT_ENABLE_MULTITAP_PORT2_INFO_0_RU "Включает расширение мультитапа на порту 2."
#define BEETLE_OPT_GUN_INPUT_MODE_LABEL_RU "Режим ввода светового пистолета"
#define BEETLE_OPT_GUN_INPUT_MODE_INFO_0_RU NULL
#define OPTION_VAL_LIGHTGUN_RU "Световой пистолет"
#define OPTION_VAL_TOUCHSCREEN_RU "Сенсорный экран"
#define BEETLE_OPT_GUN_CURSOR_LABEL_RU "Курсор пистолета"
#define BEETLE_OPT_GUN_CURSOR_INFO_0_RU NULL
#define OPTION_VAL_CROSS_RU "Перекрестье"
#define OPTION_VAL_DOT_RU "Точка"
#define BEETLE_OPT_CROSSHAIR_COLOR_P1_LABEL_RU "Цвет прицела оружия: Порт 1"
#define BEETLE_OPT_CROSSHAIR_COLOR_P1_INFO_0_RU "Цвет прицела для светового пистолета в порте 1."
#define OPTION_VAL_RED_RU "Красный"
#define OPTION_VAL_BLUE_RU "Голубой"
#define OPTION_VAL_GREEN_RU "Зелёный"
#define OPTION_VAL_ORANGE_RU "Оранжевый"
#define OPTION_VAL_YELLOW_RU "Жёлтый"
#define OPTION_VAL_CYAN_RU "Сине-зелёный"
#define OPTION_VAL_PINK_RU "Розовый"
#define OPTION_VAL_PURPLE_RU "Фиолетовый"
#define OPTION_VAL_BLACK_RU "Чёрный"
#define OPTION_VAL_WHITE_RU "Белый"
#define BEETLE_OPT_CROSSHAIR_COLOR_P2_LABEL_RU "Цвет прицела оружия: Порт 2"
#define BEETLE_OPT_CROSSHAIR_COLOR_P2_INFO_0_RU "Цвет прицела для светового пистолета в порте 2."
#define BEETLE_OPT_MOUSE_SENSITIVITY_LABEL_RU "Чувствительность мыши"
#define BEETLE_OPT_MOUSE_SENSITIVITY_INFO_0_RU "Чувствительность для устройства ввода типа \"Мышь\"."
#define OPTION_VAL_5_O42_RU NULL
#define OPTION_VAL_10_RU NULL
#define OPTION_VAL_15_RU NULL
#define OPTION_VAL_20_RU NULL
#define OPTION_VAL_25_RU NULL
#define OPTION_VAL_30_RU NULL
#define OPTION_VAL_35_RU NULL
#define OPTION_VAL_40_RU NULL
#define OPTION_VAL_45_RU NULL
#define OPTION_VAL_55_RU NULL
#define OPTION_VAL_65_RU NULL
#define OPTION_VAL_75_RU NULL
#define OPTION_VAL_85_RU NULL
#define OPTION_VAL_95_RU NULL
#define OPTION_VAL_100_O42_RU NULL
#define OPTION_VAL_105_RU NULL
#define OPTION_VAL_115_RU NULL
#define OPTION_VAL_125_RU NULL
#define OPTION_VAL_135_RU NULL
#define OPTION_VAL_145_RU NULL
#define OPTION_VAL_155_RU NULL
#define OPTION_VAL_165_RU NULL
#define OPTION_VAL_175_RU NULL
#define OPTION_VAL_185_RU NULL
#define OPTION_VAL_195_RU NULL
#define BEETLE_OPT_NEGCON_RESPONSE_LABEL_RU "Отклик скручивания neGcon"
#define BEETLE_OPT_NEGCON_RESPONSE_INFO_0_RU NULL
#define OPTION_VAL_LINEAR_RU "Линейный"
#define OPTION_VAL_QUADRATIC_RU "Квадратичный"
#define OPTION_VAL_CUBIC_RU "Кубический"
#define BEETLE_OPT_NEGCON_DEADZONE_LABEL_RU "Мёртвая зона скручивания neGcon"
#define BEETLE_OPT_NEGCON_DEADZONE_INFO_0_RU NULL
#define OPTION_VAL_0_O44_RU NULL
#define BEETLE_OPT_USE_MEDNAFEN_MEMCARD0_METHOD_LABEL_RU "Метод карты памяти"
#define BEETLE_OPT_USE_MEDNAFEN_MEMCARD0_METHOD_INFO_0_RU NULL
#define OPTION_VAL_LIBRETRO_RU NULL
#define OPTION_VAL_MEDNAFEN_RU NULL
#define BEETLE_OPT_ENABLE_MEMCARD1_LABEL_RU "Слот карты памяти 2"
#define BEETLE_OPT_ENABLE_MEMCARD1_INFO_0_RU NULL
#define BEETLE_OPT_SHARED_MEMORY_CARDS_LABEL_RU "Общие карты памяти"
#define BEETLE_OPT_SHARED_MEMORY_CARDS_INFO_0_RU NULL
#define BEETLE_OPT_MEMCARD_LEFT_INDEX_LABEL_RU NULL
#define BEETLE_OPT_MEMCARD_LEFT_INDEX_INFO_0_RU NULL
#define OPTION_VAL_0_O48_RU "0 (по умолчанию)"
#define BEETLE_OPT_MEMCARD_RIGHT_INDEX_LABEL_RU NULL
#define BEETLE_OPT_MEMCARD_RIGHT_INDEX_INFO_0_RU NULL
#define OPTION_VAL_1_O49_RU "1 (по умолчанию)"
#define BEETLE_OPT_LINE_RENDER_LABEL_RU "Преобразование линий в полигоны"
#define BEETLE_OPT_LINE_RENDER_INFO_0_RU NULL
#define OPTION_VAL_DEFAULT_RU "По умолчанию"
#define OPTION_VAL_AGGRESSIVE_RU "Агрессивно"
#define BEETLE_OPT_WIDESCREEN_HACK_LABEL_RU "Широкоэкранный режим"
#define BEETLE_OPT_WIDESCREEN_HACK_INFO_0_RU "Анаморфный 3D-рендеринг и вывод фреймбуфера в широкоформатном соотношении сторон. Наилучший результат достигается для игр полностью в 3D. Элементы 2D могут смещаться и будут вытянуты по горизонтали."
#define BEETLE_OPT_WIDESCREEN_HACK_ASPECT_RATIO_LABEL_RU "Пропорции широкоэкранного режима"
#define BEETLE_OPT_WIDESCREEN_HACK_ASPECT_RATIO_INFO_0_RU NULL
#define OPTION_VAL_16_9_RU NULL
#define OPTION_VAL_16_10_RU NULL
#define OPTION_VAL_18_9_RU NULL
#define OPTION_VAL_19_9_RU NULL
#define OPTION_VAL_20_9_RU NULL
#define OPTION_VAL_21_9_RU NULL
#define OPTION_VAL_32_9_RU NULL
#define BEETLE_OPT_CPU_DYNAREC_LABEL_RU "Динам. рекомпилятор CPU"
#define BEETLE_OPT_CPU_DYNAREC_INFO_0_RU "Динамически преобразовывает инструкции CPU в нативные. Существенно быстрее интерпретатора, но обеспечивает меньшую точность таймингов CPU и может вызывать ошибки."
#define OPTION_VAL_DISABLED_RU "Откл. (интерпретатор Beetle)"
#define OPTION_VAL_EXECUTE_RU "Макс. производительность"
#define OPTION_VAL_EXECUTE_ONE_RU "Проверка таймингов такта"
#define OPTION_VAL_RUN_INTERPRETER_RU "Интерпретатор Lightrec"
#define BEETLE_OPT_DYNAREC_INVALIDATE_LABEL_RU "Инвалидация кода dynarec"
#define BEETLE_OPT_DYNAREC_INVALIDATE_INFO_0_RU "Для некоторых игр требуется 'Полная' инвалидация, для других 'Только DMA'."
#define OPTION_VAL_FULL_RU "Полная"
#define OPTION_VAL_DMA_RU "Только DMA (немного быстрее)"
#define BEETLE_OPT_DYNAREC_EVENTCYCLES_LABEL_RU "Циклы событий DMA/MDEC/GPU в dynarec"
#define BEETLE_OPT_DYNAREC_EVENTCYCLES_INFO_0_RU "Макс. циклов, выполняемых процессором до проверки обновления таймеров/GPU/DMA/MDEC. Более высокие значения повышают скорость. Для интерпретатора beetle эффект менее выражен, чем для dynarec."
#define OPTION_VAL_128_RU "128 (по умолчанию)"
#define BEETLE_OPT_DYNAREC_SPU_SAMPLES_LABEL_RU "Сэмплы SPU в dynarec"
#define BEETLE_OPT_DYNAREC_SPU_SAMPLES_INFO_0_RU "Макс. сэмплов, обрабатываемых до проверки обновления SPU. Более высокие значения повышают скорость, но в некоторых играх отклонение от 1 приводит к искажению звука."
#define BEETLE_OPT_CORE_TIMING_FPS_LABEL_RU "Тайминг частоты кадров ядра"
#define BEETLE_OPT_CORE_TIMING_FPS_INFO_0_RU NULL
#define OPTION_VAL_AUTO_TOGGLE_RU "Автоматическое переключение"
#define OPTION_VAL_FORCE_PROGRESSIVE_RU "Принудительная прогрессивная скорость"
#define OPTION_VAL_FORCE_INTERLACED_RU "Принуд. чересстрочная"
#define BEETLE_OPT_DEINTERLACER_LABEL_RU "Метод деинтерлейсинга"
#define BEETLE_OPT_DEINTERLACER_INFO_0_RU NULL
#define OPTION_VAL_WEAVE_RU NULL
#define OPTION_VAL_BOB_RU NULL
#define OPTION_VAL_BOB_OFFSET_RU NULL
#define OPTION_VAL_FASTMAD_RU NULL
#define BEETLE_OPT_FRAME_DUPING_LABEL_RU "Дамп кадров"
#define BEETLE_OPT_FRAME_DUPING_INFO_0_RU NULL
#define BEETLE_OPT_PAL_VIDEO_TIMING_OVERRIDE_LABEL_RU NULL
#define BEETLE_OPT_PAL_VIDEO_TIMING_OVERRIDE_INFO_0_RU NULL
#define BEETLE_OPT_ASPECT_RATIO_LABEL_RU "Соотношение сторон ядра"
#define BEETLE_OPT_ASPECT_RATIO_INFO_0_RU NULL
#define OPTION_VAL_CORRECTED_RU "Скорректированное"
#define OPTION_VAL_UNCORRECTED_RU "Нескорректированное"
#define OPTION_VAL_4_3_RU "Принуд. 4:3"
#define OPTION_VAL_NTSC_RU "Принуд. NTSC"
#define BEETLE_OPT_CROP_OVERSCAN_LABEL_RU "Кадрирование вылетов развёртки"
#define BEETLE_OPT_CROP_OVERSCAN_INFO_0_RU "Полное нативное изображение или без рамок."
#define OPTION_VAL_STATIC_RU "Горизонтальная"
#define OPTION_VAL_SMART_RU "Горизонтальный + вертикальный"
#define BEETLE_OPT_IMAGE_CROP_LABEL_RU "Добавочное кадрирование"
#define BEETLE_OPT_IMAGE_CROP_INFO_0_RU NULL
#define OPTION_VAL_9PX_RU "9 пикс."
#define OPTION_VAL_10PX_RU "10 пикс."
#define OPTION_VAL_11PX_RU "11  пикс."
#define OPTION_VAL_12PX_RU "12 пикс."
#define OPTION_VAL_13PX_RU "13 пикс."
#define OPTION_VAL_14PX_RU "14 пикс."
#define OPTION_VAL_15PX_RU "15 пикс."
#define OPTION_VAL_16PX_RU "16 пикс."
#define OPTION_VAL_17PX_RU "17 пикс."
#define OPTION_VAL_18PX_RU "18 пикс."
#define OPTION_VAL_19PX_RU "19 пикс."
#define OPTION_VAL_20PX_RU "20 пикс."
#define BEETLE_OPT_IMAGE_OFFSET_LABEL_RU "Сдвиг изображения после обрезки"
#define BEETLE_OPT_IMAGE_OFFSET_INFO_0_RU NULL
#define OPTION_VAL_12PX_O64_RU "-12 пикс."
#define OPTION_VAL_11PX_O64_RU "-11 пикс."
#define OPTION_VAL_10PX_O64_RU "-10 пикс."
#define OPTION_VAL_9PX_O64_RU "-9 пикс."
#define OPTION_VAL_8PX_O64_RU NULL
#define OPTION_VAL_7PX_O64_RU NULL
#define OPTION_VAL_6PX_O64_RU NULL
#define OPTION_VAL_5PX_O64_RU NULL
#define OPTION_VAL_4PX_O64_RU NULL
#define OPTION_VAL_3PX_O64_RU NULL
#define OPTION_VAL_2PX_O64_RU NULL
#define OPTION_VAL_1PX_O64_RU NULL
#define OPTION_VAL_1PX_O64_0_RU NULL
#define OPTION_VAL_2PX_O64_0_RU NULL
#define OPTION_VAL_3PX_O64_0_RU NULL
#define OPTION_VAL_4PX_O64_0_RU NULL
#define OPTION_VAL_5PX_O64_0_RU NULL
#define OPTION_VAL_6PX_O64_0_RU "+6 пикс."
#define OPTION_VAL_7PX_O64_0_RU "+7 пикс."
#define OPTION_VAL_8PX_O64_0_RU "+8 пикс."
#define OPTION_VAL_9PX_O64_0_RU "+9 пикс."
#define OPTION_VAL_10PX_O64_0_RU "+10 пикс."
#define OPTION_VAL_11PX_O64_0_RU "+11 пикс."
#define OPTION_VAL_12PX_O64_0_RU "+12 пикс."
#define BEETLE_OPT_IMAGE_OFFSET_CYCLES_LABEL_RU "Сдвиг изображения по горизонтали (в тактах GPU)"
#define BEETLE_OPT_IMAGE_OFFSET_CYCLES_INFO_0_RU NULL
#define BEETLE_OPT_INITIAL_SCANLINE_LABEL_RU "Начальная строка развёртки - NTSC"
#define BEETLE_OPT_INITIAL_SCANLINE_INFO_0_RU NULL
#define BEETLE_OPT_LAST_SCANLINE_LABEL_RU "Последняя строка развёртки - NTSC"
#define BEETLE_OPT_LAST_SCANLINE_INFO_0_RU NULL
#define OPTION_VAL_239_RU "239 (по умолчанию)"
#define BEETLE_OPT_INITIAL_SCANLINE_PAL_LABEL_RU "Начальная строка развёртки - PAL"
#define BEETLE_OPT_INITIAL_SCANLINE_PAL_INFO_0_RU NULL
#define BEETLE_OPT_LAST_SCANLINE_PAL_LABEL_RU "Последняя строка развёртки - PAL"
#define BEETLE_OPT_LAST_SCANLINE_PAL_INFO_0_RU NULL
#define OPTION_VAL_287_RU "287 (по умолчанию)"

struct retro_core_option_v2_category option_cats_ru[] = {
   {
      "system",
      CATEGORY_SYSTEM_LABEL_RU,
      CATEGORY_SYSTEM_INFO_0_RU
   },
   {
      "video",
      CATEGORY_VIDEO_LABEL_RU,
      CATEGORY_VIDEO_INFO_0_RU
   },
   {
      "pgxp",
      CATEGORY_PGXP_LABEL_RU,
      CATEGORY_PGXP_INFO_0_RU
   },
   {
      "osd",
      CATEGORY_OSD_LABEL_RU,
      CATEGORY_OSD_INFO_0_RU
   },
   {
      "input",
      CATEGORY_INPUT_LABEL_RU,
      CATEGORY_INPUT_INFO_0_RU
   },
   {
      "memcards",
      CATEGORY_MEMCARDS_LABEL_RU,
      CATEGORY_MEMCARDS_INFO_0_RU
   },
   {
      "hacks",
      CATEGORY_HACKS_LABEL_RU,
      CATEGORY_HACKS_INFO_0_RU
   },
   { NULL, NULL, NULL },
};
struct retro_core_option_v2_definition option_defs_ru[] = {
   {
      BEETLE_OPT(cpu_freq_scale),
      BEETLE_OPT_CPU_FREQ_SCALE_LABEL_RU,
      NULL,
      BEETLE_OPT_CPU_FREQ_SCALE_INFO_0_RU,
      NULL,
      "system",
      {
         { "50%",           OPTION_VAL_50_RU },
         { "60%",           OPTION_VAL_60_RU },
         { "70%",           OPTION_VAL_70_RU },
         { "80%",           OPTION_VAL_80_RU },
         { "90%",           OPTION_VAL_90_RU },
         { "100%",          OPTION_VAL_100_RU },
         { "110%",          OPTION_VAL_110_RU },
         { "120%",          OPTION_VAL_120_RU },
         { "130%",          OPTION_VAL_130_RU },
         { "140%",          OPTION_VAL_140_RU },
         { "150%",          OPTION_VAL_150_RU },
         { "160%",          OPTION_VAL_160_RU },
         { "170%",          OPTION_VAL_170_RU },
         { "180%",          OPTION_VAL_180_RU },
         { "190%",          OPTION_VAL_190_RU },
         { "200%",          OPTION_VAL_200_RU },
         { "210%",          OPTION_VAL_210_RU },
         { "220%",          OPTION_VAL_220_RU },
         { "230%",          OPTION_VAL_230_RU },
         { "240%",          OPTION_VAL_240_RU },
         { "250%",          OPTION_VAL_250_RU },
         { "260%",          OPTION_VAL_260_RU },
         { "270%",          OPTION_VAL_270_RU },
         { "280%",          OPTION_VAL_280_RU },
         { "290%",          OPTION_VAL_290_RU },
         { "300%",          OPTION_VAL_300_RU },
         { "310%",          OPTION_VAL_310_RU },
         { "320%",          OPTION_VAL_320_RU },
         { "330%",          OPTION_VAL_330_RU },
         { "340%",          OPTION_VAL_340_RU },
         { "350%",          OPTION_VAL_350_RU },
         { "360%",          OPTION_VAL_360_RU },
         { "370%",          OPTION_VAL_370_RU },
         { "380%",          OPTION_VAL_380_RU },
         { "390%",          OPTION_VAL_390_RU },
         { "400%",          OPTION_VAL_400_RU },
         { "410%",          OPTION_VAL_410_RU },
         { "420%",          OPTION_VAL_420_RU },
         { "430%",          OPTION_VAL_430_RU },
         { "440%",          OPTION_VAL_440_RU },
         { "450%",          OPTION_VAL_450_RU },
         { "460%",          OPTION_VAL_460_RU },
         { "470%",          OPTION_VAL_470_RU },
         { "480%",          OPTION_VAL_480_RU },
         { "490%",          OPTION_VAL_490_RU },
         { "500%",          OPTION_VAL_500_RU },
         { "510%",          OPTION_VAL_510_RU },
         { "520%",          OPTION_VAL_520_RU },
         { "530%",          OPTION_VAL_530_RU },
         { "540%",          OPTION_VAL_540_RU },
         { "550%",          OPTION_VAL_550_RU },
         { "560%",          OPTION_VAL_560_RU },
         { "570%",          OPTION_VAL_570_RU },
         { "580%",          OPTION_VAL_580_RU },
         { "590%",          OPTION_VAL_590_RU },
         { "600%",          OPTION_VAL_600_RU },
         { "610%",          OPTION_VAL_610_RU },
         { "620%",          OPTION_VAL_620_RU },
         { "630%",          OPTION_VAL_630_RU },
         { "640%",          OPTION_VAL_640_RU },
         { "650%",          OPTION_VAL_650_RU },
         { "660%",          OPTION_VAL_660_RU },
         { "670%",          OPTION_VAL_670_RU },
         { "680%",          OPTION_VAL_680_RU },
         { "690%",          OPTION_VAL_690_RU },
         { "700%",          OPTION_VAL_700_RU },
         { "710%",          OPTION_VAL_710_RU },
         { "720%",          OPTION_VAL_720_RU },
         { "730%",          OPTION_VAL_730_RU },
         { "740%",          OPTION_VAL_740_RU },
         { "750%",          OPTION_VAL_750_RU },
         { NULL, NULL },
      },
      "100%"
   },
   {
      BEETLE_OPT(gpu_overclock),
      BEETLE_OPT_GPU_OVERCLOCK_LABEL_RU,
      NULL,
      BEETLE_OPT_GPU_OVERCLOCK_INFO_0_RU,
      NULL,
      "system",
      {
         { "1x(native)", OPTION_VAL_1X_NATIVE_RU },
         { "2x",         OPTION_VAL_2X_RU },
         { "4x",         OPTION_VAL_4X_RU },
         { "8x",         OPTION_VAL_8X_RU },
         { "16x",        OPTION_VAL_16X_RU },
         { "32x",        OPTION_VAL_32X_RU },
         { NULL, NULL },
      },
      "1x(native)"
   },
   {
      BEETLE_OPT(gte_overclock),
      BEETLE_OPT_GTE_OVERCLOCK_LABEL_RU,
      NULL,
      BEETLE_OPT_GTE_OVERCLOCK_INFO_0_RU,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
#ifndef EMSCRIPTEN
   {
      BEETLE_OPT(cd_access_method),
      BEETLE_OPT_CD_ACCESS_METHOD_LABEL_RU,
      NULL,
      BEETLE_OPT_CD_ACCESS_METHOD_INFO_0_RU,
      NULL,
      "system",
      {
         { "sync",     OPTION_VAL_SYNC_RU },
         { "async",    OPTION_VAL_ASYNC_RU },
         { "precache", OPTION_VAL_PRECACHE_RU },
         { NULL, NULL },
      },
      "sync"
   },
#endif
   {
      BEETLE_OPT(cd_fastload),
      BEETLE_OPT_CD_FASTLOAD_LABEL_RU,
      NULL,
      BEETLE_OPT_CD_FASTLOAD_INFO_0_RU,
      NULL,
      "system",
      {
         { "2x(native)", OPTION_VAL_2X_NATIVE_RU },
         { "4x",          OPTION_VAL_4X_RU },
         { "6x",          OPTION_VAL_6X_RU },
         { "8x",          OPTION_VAL_8X_RU },
         { "10x",         OPTION_VAL_10X_RU },
         { "12x",         OPTION_VAL_12X_RU },
         { "14x",         OPTION_VAL_14X_RU },
         { NULL, NULL },
      },
      "2x(native)"
   },
   {
      BEETLE_OPT(skip_bios),
      BEETLE_OPT_SKIP_BIOS_LABEL_RU,
      NULL,
      BEETLE_OPT_SKIP_BIOS_INFO_0_RU,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(override_bios),
      BEETLE_OPT_OVERRIDE_BIOS_LABEL_RU,
      NULL,
      BEETLE_OPT_OVERRIDE_BIOS_INFO_0_RU,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "psxonpsp", OPTION_VAL_PSXONPSP_RU },
         { "ps1_rom",  OPTION_VAL_PS1_ROM_RU },
         { "openbios", OPTION_VAL_OPENBIOS_RU     },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(region),
      BEETLE_OPT_REGION_LABEL_RU,
      NULL,
      BEETLE_OPT_REGION_INFO_0_RU,
      NULL,
      "system",
      {
         { "auto",   OPTION_VAL_AUTO_RU },
         { "ntsc-j", OPTION_VAL_NTSC_J_RU },
         { "ntsc-u", OPTION_VAL_NTSC_U_RU },
         { "pal",    OPTION_VAL_PAL_RU },
         { NULL, NULL },
      },
      "auto"
   },
   {
      BEETLE_OPT(internal_resolution),
      BEETLE_OPT_INTERNAL_RESOLUTION_LABEL_RU,
      NULL,
      BEETLE_OPT_INTERNAL_RESOLUTION_INFO_0_RU,
      NULL,
      "video",
      {
         { "1x(native)", OPTION_VAL_1X_NATIVE_RU },
         { "2x",         OPTION_VAL_2X_RU },
         { "4x",         OPTION_VAL_4X_RU },
         { "8x",         OPTION_VAL_8X_RU },
         { "16x",        OPTION_VAL_16X_RU },
         { NULL, NULL },
      },
      "1x(native)"
   },
#if defined(HAVE_OPENGL) || defined(HAVE_OPENGLES) || defined(HAVE_VULKAN)
   {
      BEETLE_OPT(renderer),
      BEETLE_OPT_RENDERER_LABEL_RU,
      NULL,
      BEETLE_OPT_RENDERER_INFO_0_RU,
      NULL,
      "video",
      {
         { "hardware",    OPTION_VAL_HARDWARE_RU },
#if defined(HAVE_OPENGL) || defined(HAVE_OPENGLES)
         { "hardware_gl", OPTION_VAL_HARDWARE_GL_RU },
#endif
#if defined(HAVE_VULKAN)
         { "hardware_vk", OPTION_VAL_HARDWARE_VK_RU },
#endif
         { "software",    OPTION_VAL_SOFTWARE_RU },
         { NULL, NULL },
      },
      "hardware"
   },
   {
      BEETLE_OPT(renderer_software_fb),
      BEETLE_OPT_RENDERER_SOFTWARE_FB_LABEL_RU,
      NULL,
      BEETLE_OPT_RENDERER_SOFTWARE_FB_INFO_0_RU,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled"
   },
#endif
#if defined(HAVE_OPENGL) || defined(HAVE_OPENGLES)
   {
      BEETLE_OPT(depth),
      BEETLE_OPT_DEPTH_LABEL_RU,
      NULL,
      BEETLE_OPT_DEPTH_INFO_0_RU,
      NULL,
      "video",
      {
         { "16bpp(native)", OPTION_VAL_16BPP_NATIVE_RU },
         { "32bpp",         OPTION_VAL_32BPP_RU },
         { NULL, NULL },
      },
      "16bpp(native)"
   },
   // Sort of, it's more like 15-bit high color and 24-bit true color for visible output. The alpha channel is used for mask bit. Vulkan renderer uses ABGR1555_555 for 31 bits internal? FMVs are always 24-bit on all renderers like original hardware (BGR888, no alpha)
#endif
   {
      BEETLE_OPT(dither_mode),
      BEETLE_OPT_DITHER_MODE_LABEL_RU,
      NULL,
      BEETLE_OPT_DITHER_MODE_INFO_0_RU,
      NULL,
      "video",
      {
         { "1x(native)",          OPTION_VAL_1X_NATIVE_RU },
         { "internal resolution", OPTION_VAL_INTERNAL_RESOLUTION_RU },
         { "disabled",            NULL },
         { NULL, NULL },
      },
      "1x(native)"
   },
#ifdef HAVE_VULKAN
   {
      BEETLE_OPT(scaled_uv_offset),
      BEETLE_OPT_SCALED_UV_OFFSET_LABEL_RU,
      NULL,
      BEETLE_OPT_SCALED_UV_OFFSET_INFO_0_RU,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled"
   },
#endif
#if defined(HAVE_OPENGL) || defined(HAVE_OPENGLES) || defined(HAVE_VULKAN)
   {
      BEETLE_OPT(filter),
      BEETLE_OPT_FILTER_LABEL_RU,
      NULL,
      BEETLE_OPT_FILTER_INFO_0_RU,
      NULL,
      "video",
      {
         { "nearest",  OPTION_VAL_NEAREST_RU },
         { "SABR",     OPTION_VAL_SABR_RU },
         { "xBR",      OPTION_VAL_XBR_RU },
         { "bilinear", OPTION_VAL_BILINEAR_RU },
         { "3-point",  OPTION_VAL_3_POINT_RU },
         { "JINC2",    OPTION_VAL_JINC2_RU },
         { NULL, NULL },
      },
      "nearest"
   },
#ifdef HAVE_VULKAN
   {
      BEETLE_OPT(filter_exclude_sprite),
      BEETLE_OPT_FILTER_EXCLUDE_SPRITE_LABEL_RU,
      NULL,
      BEETLE_OPT_FILTER_EXCLUDE_SPRITE_INFO_0_RU,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "opaque", OPTION_VAL_OPAQUE_RU },
         { "all", OPTION_VAL_ALL_RU },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(filter_exclude_2d_polygon),
      BEETLE_OPT_FILTER_EXCLUDE_2D_POLYGON_LABEL_RU,
      NULL,
      BEETLE_OPT_FILTER_EXCLUDE_2D_POLYGON_INFO_0_RU,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "opaque", OPTION_VAL_OPAQUE_RU },
         { "all", OPTION_VAL_ALL_RU },
         { NULL, NULL },
      },
      "disabled"
   },
#endif
#endif
#ifdef HAVE_VULKAN
   {
      BEETLE_OPT(adaptive_smoothing),
      BEETLE_OPT_ADAPTIVE_SMOOTHING_LABEL_RU,
      NULL,
      BEETLE_OPT_ADAPTIVE_SMOOTHING_INFO_0_RU,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(super_sampling),
      BEETLE_OPT_SUPER_SAMPLING_LABEL_RU,
      NULL,
      BEETLE_OPT_SUPER_SAMPLING_INFO_0_RU,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(msaa),
      BEETLE_OPT_MSAA_LABEL_RU,
      NULL,
      BEETLE_OPT_MSAA_INFO_0_RU,
      NULL,
      "video",
      {
         { "1x",  OPTION_VAL_1X_RU },
         { "2x",  OPTION_VAL_2X_RU },
         { "4x",  OPTION_VAL_4X_RU },
         { "8x",  OPTION_VAL_8X_RU },
         { "16x", OPTION_VAL_16X_RU },
         { NULL, NULL },
      },
      "1x"
   },
   {
      BEETLE_OPT(mdec_yuv),
      BEETLE_OPT_MDEC_YUV_LABEL_RU,
      NULL,
      BEETLE_OPT_MDEC_YUV_INFO_0_RU,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(track_textures),
      BEETLE_OPT_TRACK_TEXTURES_LABEL_RU,
      NULL,
      BEETLE_OPT_TRACK_TEXTURES_INFO_0_RU,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
#ifdef TEXTURE_DUMPING_ENABLED
   {
      BEETLE_OPT(dump_textures),
      BEETLE_OPT_DUMP_TEXTURES_LABEL_RU,
      NULL,
      BEETLE_OPT_DUMP_TEXTURES_INFO_0_RU,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
#endif
   {
      BEETLE_OPT(replace_textures),
      BEETLE_OPT_REPLACE_TEXTURES_LABEL_RU,
      NULL,
      BEETLE_OPT_REPLACE_TEXTURES_INFO_0_RU,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
#endif
   {
      BEETLE_OPT(pgxp_mode),
      BEETLE_OPT_PGXP_MODE_LABEL_RU,
      NULL,
      BEETLE_OPT_PGXP_MODE_INFO_0_RU,
      NULL,
      "pgxp",
      {
         { "disabled",     NULL },
         { "memory only",  OPTION_VAL_MEMORY_ONLY_RU },
         { "memory + CPU", OPTION_VAL_MEMORY_CPU_RU },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(pgxp_2d_tol),
      BEETLE_OPT_PGXP_2D_TOL_LABEL_RU,
      NULL,
      BEETLE_OPT_PGXP_2D_TOL_INFO_0_RU,
      NULL,
      "pgxp",
      {
         { "disabled", NULL },
         { "0px", OPTION_VAL_0PX_RU },
         { "1px", OPTION_VAL_1PX_RU },
         { "2px", OPTION_VAL_2PX_RU },
         { "3px", OPTION_VAL_3PX_RU },
         { "4px", OPTION_VAL_4PX_RU },
         { "5px", OPTION_VAL_5PX_RU },
         { "6px", OPTION_VAL_6PX_RU },
         { "7px", OPTION_VAL_7PX_RU },
         { "8px", OPTION_VAL_8PX_RU },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(pgxp_nclip),
      BEETLE_OPT_PGXP_NCLIP_LABEL_RU,
      NULL,
      BEETLE_OPT_PGXP_NCLIP_INFO_0_RU,
      NULL,
      "pgxp",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
#if defined(HAVE_OPENGL) || defined(HAVE_OPENGLES) || defined(HAVE_VULKAN)
   {
      BEETLE_OPT(pgxp_vertex),
      BEETLE_OPT_PGXP_VERTEX_LABEL_RU,
      NULL,
      BEETLE_OPT_PGXP_VERTEX_INFO_0_RU,
      NULL,
      "pgxp",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
#endif
   {
      BEETLE_OPT(pgxp_texture),
      BEETLE_OPT_PGXP_TEXTURE_LABEL_RU,
      NULL,
      BEETLE_OPT_PGXP_TEXTURE_INFO_0_RU,
      NULL,
      "pgxp",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(display_internal_fps),
      BEETLE_OPT_DISPLAY_INTERNAL_FPS_LABEL_RU,
      NULL,
      BEETLE_OPT_DISPLAY_INTERNAL_FPS_INFO_0_RU,
      NULL,
      "osd",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(display_osd),
      BEETLE_OPT_DISPLAY_OSD_LABEL_RU,
      NULL,
      BEETLE_OPT_DISPLAY_OSD_INFO_0_RU,
      NULL,
      "osd",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled"
   },
#if defined(HAVE_OPENGL) || defined(HAVE_OPENGLES) || defined(HAVE_VULKAN)
   {
      BEETLE_OPT(display_vram),
      BEETLE_OPT_DISPLAY_VRAM_LABEL_RU,
      NULL,
      BEETLE_OPT_DISPLAY_VRAM_INFO_0_RU,
      NULL,
      "osd",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
#endif
   {
      BEETLE_OPT(analog_calibration),
      BEETLE_OPT_ANALOG_CALIBRATION_LABEL_RU,
      NULL,
      BEETLE_OPT_ANALOG_CALIBRATION_INFO_0_RU,
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(analog_toggle),
      BEETLE_OPT_ANALOG_TOGGLE_LABEL_RU,
      NULL,
      BEETLE_OPT_ANALOG_TOGGLE_INFO_0_RU,
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { "enabled-analog", OPTION_VAL_ENABLED_ANALOG_RU },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(analog_toggle_combo),
      BEETLE_OPT_ANALOG_TOGGLE_COMBO_LABEL_RU,
      NULL,
      BEETLE_OPT_ANALOG_TOGGLE_COMBO_INFO_0_RU,
      NULL,
      "input",
      {
         { "l1+l2+r1+r2+start+select", OPTION_VAL_L1_L2_R1_R2_START_SELECT_RU },
         { "l1+r1+select",             OPTION_VAL_L1_R1_SELECT_RU },
         { "l1+r1+start",              OPTION_VAL_L1_R1_START_RU },
         { "l1+r1+l3",                 OPTION_VAL_L1_R1_L3_RU },
         { "l1+r1+r3",                 OPTION_VAL_L1_R1_R3_RU },
         { "l2+r2+select",             OPTION_VAL_L2_R2_SELECT_RU },
         { "l2+r2+start",              OPTION_VAL_L2_R2_START_RU },
         { "l2+r2+l3",                 OPTION_VAL_L2_R2_L3_RU },
         { "l2+r2+r3",                 OPTION_VAL_L2_R2_R3_RU },
         { "l3+r3",                    OPTION_VAL_L3_R3_RU },
         { NULL, NULL },
      },
      "l1+r1+select"
   },
   {
      BEETLE_OPT(analog_toggle_hold),
      BEETLE_OPT_ANALOG_TOGGLE_HOLD_LABEL_RU,
      NULL,
      BEETLE_OPT_ANALOG_TOGGLE_HOLD_INFO_0_RU,
      NULL,
      "input",
      {
         { "0", OPTION_VAL_0_RU },
         { "1", OPTION_VAL_1_RU },
         { "2", OPTION_VAL_2_RU },
         { "3", OPTION_VAL_3_RU },
         { "4", OPTION_VAL_4_RU },
         { "5", OPTION_VAL_5_RU },
         { NULL, NULL },
      },
      "1"
   },
   {
      BEETLE_OPT(enable_multitap_port1),
      BEETLE_OPT_ENABLE_MULTITAP_PORT1_LABEL_RU,
      NULL,
      BEETLE_OPT_ENABLE_MULTITAP_PORT1_INFO_0_RU,
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(enable_multitap_port2),
      BEETLE_OPT_ENABLE_MULTITAP_PORT2_LABEL_RU,
      NULL,
      BEETLE_OPT_ENABLE_MULTITAP_PORT2_INFO_0_RU,
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(gun_input_mode),
      BEETLE_OPT_GUN_INPUT_MODE_LABEL_RU,
      NULL,
      BEETLE_OPT_GUN_INPUT_MODE_INFO_0_RU,
      NULL,
      "input",
      {
         { "lightgun",    OPTION_VAL_LIGHTGUN_RU },
         { "touchscreen", OPTION_VAL_TOUCHSCREEN_RU },
         { NULL, NULL },
      },
      "lightgun"
   },
   // Shouldn't the gun_input_mode just be Mouse vs. Touchscreen?
   {
      BEETLE_OPT(gun_cursor),
      BEETLE_OPT_GUN_CURSOR_LABEL_RU,
      NULL,
      BEETLE_OPT_GUN_CURSOR_INFO_0_RU,
      NULL,
      "input",
      {
         { "off",   "disabled" },
         { "cross", OPTION_VAL_CROSS_RU },
         { "dot",   OPTION_VAL_DOT_RU },
         { NULL, NULL },
      },
      "cross"
   },
   {
      BEETLE_OPT(crosshair_color_p1),
      BEETLE_OPT_CROSSHAIR_COLOR_P1_LABEL_RU,
      NULL,
      BEETLE_OPT_CROSSHAIR_COLOR_P1_INFO_0_RU,
      NULL,
      "input",
      {
         { "red",    OPTION_VAL_RED_RU },
         { "blue",   OPTION_VAL_BLUE_RU },
         { "green",  OPTION_VAL_GREEN_RU },
         { "orange", OPTION_VAL_ORANGE_RU },
         { "yellow", OPTION_VAL_YELLOW_RU },
         { "cyan",   OPTION_VAL_CYAN_RU },
         { "pink",   OPTION_VAL_PINK_RU },
         { "purple", OPTION_VAL_PURPLE_RU },
         { "black",  OPTION_VAL_BLACK_RU },
         { "white",  OPTION_VAL_WHITE_RU },
         { NULL, NULL },
      },
      "red"
   },
   {
      BEETLE_OPT(crosshair_color_p2),
      BEETLE_OPT_CROSSHAIR_COLOR_P2_LABEL_RU,
      NULL,
      BEETLE_OPT_CROSSHAIR_COLOR_P2_INFO_0_RU,
      NULL,
      "input",
      {
         { "blue",   OPTION_VAL_BLUE_RU },
         { "red",    OPTION_VAL_RED_RU },
         { "green",  OPTION_VAL_GREEN_RU },
         { "orange", OPTION_VAL_ORANGE_RU },
         { "yellow", OPTION_VAL_YELLOW_RU },
         { "cyan",   OPTION_VAL_CYAN_RU },
         { "pink",   OPTION_VAL_PINK_RU },
         { "purple", OPTION_VAL_PURPLE_RU },
         { "black",  OPTION_VAL_BLACK_RU },
         { "white",  OPTION_VAL_WHITE_RU },
         { NULL, NULL },
      },
      "blue"
   },
   {
      BEETLE_OPT(mouse_sensitivity),
      BEETLE_OPT_MOUSE_SENSITIVITY_LABEL_RU,
      NULL,
      BEETLE_OPT_MOUSE_SENSITIVITY_INFO_0_RU,
      NULL,
      "input",
      {
         { "5%",   OPTION_VAL_5_O42_RU },
         { "10%",  OPTION_VAL_10_RU },
         { "15%",  OPTION_VAL_15_RU },
         { "20%",  OPTION_VAL_20_RU },
         { "25%",  OPTION_VAL_25_RU },
         { "30%",  OPTION_VAL_30_RU },
         { "35%",  OPTION_VAL_35_RU },
         { "40%",  OPTION_VAL_40_RU },
         { "45%",  OPTION_VAL_45_RU },
         { "50%",  OPTION_VAL_50_RU },
         { "55%",  OPTION_VAL_55_RU },
         { "60%",  OPTION_VAL_60_RU },
         { "65%",  OPTION_VAL_65_RU },
         { "70%",  OPTION_VAL_70_RU },
         { "75%",  OPTION_VAL_75_RU },
         { "80%",  OPTION_VAL_80_RU },
         { "85%",  OPTION_VAL_85_RU },
         { "90%",  OPTION_VAL_90_RU },
         { "95%",  OPTION_VAL_95_RU },
         { "100%", OPTION_VAL_100_O42_RU },
         { "105%", OPTION_VAL_105_RU },
         { "110%", OPTION_VAL_110_RU },
         { "115%", OPTION_VAL_115_RU },
         { "120%", OPTION_VAL_120_RU },
         { "125%", OPTION_VAL_125_RU },
         { "130%", OPTION_VAL_130_RU },
         { "135%", OPTION_VAL_135_RU },
         { "140%", OPTION_VAL_140_RU },
         { "145%", OPTION_VAL_145_RU },
         { "150%", OPTION_VAL_150_RU },
         { "155%", OPTION_VAL_155_RU },
         { "160%", OPTION_VAL_160_RU },
         { "165%", OPTION_VAL_165_RU },
         { "170%", OPTION_VAL_170_RU },
         { "175%", OPTION_VAL_175_RU },
         { "180%", OPTION_VAL_180_RU },
         { "185%", OPTION_VAL_185_RU },
         { "190%", OPTION_VAL_190_RU },
         { "195%", OPTION_VAL_195_RU },
         { "200%", OPTION_VAL_200_RU },
         { NULL, NULL },
      },
      "100%"
   },
   {
      BEETLE_OPT(negcon_response),
      BEETLE_OPT_NEGCON_RESPONSE_LABEL_RU,
      NULL,
      BEETLE_OPT_NEGCON_RESPONSE_INFO_0_RU,
      NULL,
      "input",
      {
         { "linear",    OPTION_VAL_LINEAR_RU },
         { "quadratic", OPTION_VAL_QUADRATIC_RU },
         { "cubic",     OPTION_VAL_CUBIC_RU },
         { NULL, NULL },
      },
      "linear"
   },
   {
      BEETLE_OPT(negcon_deadzone),
      BEETLE_OPT_NEGCON_DEADZONE_LABEL_RU,
      NULL,
      BEETLE_OPT_NEGCON_DEADZONE_INFO_0_RU,
      NULL,
      "input",
      {
         { "0%",  OPTION_VAL_0_O44_RU },
         { "5%",  OPTION_VAL_5_O42_RU },
         { "10%", OPTION_VAL_10_RU },
         { "15%", OPTION_VAL_15_RU },
         { "20%", OPTION_VAL_20_RU },
         { "25%", OPTION_VAL_25_RU },
         { "30%", OPTION_VAL_30_RU },
         { NULL, NULL },
      },
      "0%"
   },
   {
      BEETLE_OPT(use_mednafen_memcard0_method),
      BEETLE_OPT_USE_MEDNAFEN_MEMCARD0_METHOD_LABEL_RU,
      NULL,
      BEETLE_OPT_USE_MEDNAFEN_MEMCARD0_METHOD_INFO_0_RU,
      NULL,
      "memcards",
      {
         { "libretro", OPTION_VAL_LIBRETRO_RU },
         { "mednafen", OPTION_VAL_MEDNAFEN_RU },
         { NULL, NULL },
      },
      "libretro"
   },
   {
      BEETLE_OPT(enable_memcard1),
      BEETLE_OPT_ENABLE_MEMCARD1_LABEL_RU,
      NULL,
      BEETLE_OPT_ENABLE_MEMCARD1_INFO_0_RU,
      NULL,
      "memcards",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(shared_memory_cards),
      BEETLE_OPT_SHARED_MEMORY_CARDS_LABEL_RU,
      NULL,
      BEETLE_OPT_SHARED_MEMORY_CARDS_INFO_0_RU,
      NULL,
      "memcards",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(memcard_left_index),
      BEETLE_OPT_MEMCARD_LEFT_INDEX_LABEL_RU,
      NULL,
      BEETLE_OPT_MEMCARD_LEFT_INDEX_INFO_0_RU,
      NULL,
      "memcards",
      {
         { "0",  OPTION_VAL_0_O48_RU },
         { "1",  NULL },
         { "2",  NULL },
         { "3",  NULL },
         { "4",  NULL },
         { "5",  NULL },
         { "6",  NULL },
         { "7",  NULL },
         { "8",  NULL },
         { "9",  NULL },
         { "10", NULL },
         { "11", NULL },
         { "12", NULL },
         { "13", NULL },
         { "14", NULL },
         { "15", NULL },
         { "16", NULL },
         { "17", NULL },
         { "18", NULL },
         { "19", NULL },
         { "20", NULL },
         { "21", NULL },
         { "22", NULL },
         { "23", NULL },
         { "24", NULL },
         { "25", NULL },
         { "26", NULL },
         { "27", NULL },
         { "28", NULL },
         { "29", NULL },
         { "30", NULL },
         { "31", NULL },
         { "32", NULL },
         { "33", NULL },
         { "34", NULL },
         { "35", NULL },
         { "36", NULL },
         { "37", NULL },
         { "38", NULL },
         { "39", NULL },
         { "40", NULL },
         { "41", NULL },
         { "42", NULL },
         { "43", NULL },
         { "44", NULL },
         { "45", NULL },
         { "46", NULL },
         { "47", NULL },
         { "48", NULL },
         { "49", NULL },
         { "50", NULL },
         { "51", NULL },
         { "52", NULL },
         { "53", NULL },
         { "54", NULL },
         { "55", NULL },
         { "56", NULL },
         { "57", NULL },
         { "58", NULL },
         { "59", NULL },
         { "60", NULL },
         { "61", NULL },
         { "62", NULL },
         { "63", NULL },
         { NULL, NULL },
      },
      "0"
   },
   {
      BEETLE_OPT(memcard_right_index),
      BEETLE_OPT_MEMCARD_RIGHT_INDEX_LABEL_RU,
      NULL,
      BEETLE_OPT_MEMCARD_RIGHT_INDEX_INFO_0_RU,
      NULL,
      "memcards",
      {
         { "0",  NULL },
         { "1",  OPTION_VAL_1_O49_RU },
         { "2",  NULL },
         { "3",  NULL },
         { "4",  NULL },
         { "5",  NULL },
         { "6",  NULL },
         { "7",  NULL },
         { "8",  NULL },
         { "9",  NULL },
         { "10", NULL },
         { "11", NULL },
         { "12", NULL },
         { "13", NULL },
         { "14", NULL },
         { "15", NULL },
         { "16", NULL },
         { "17", NULL },
         { "18", NULL },
         { "19", NULL },
         { "20", NULL },
         { "21", NULL },
         { "22", NULL },
         { "23", NULL },
         { "24", NULL },
         { "25", NULL },
         { "26", NULL },
         { "27", NULL },
         { "28", NULL },
         { "29", NULL },
         { "30", NULL },
         { "31", NULL },
         { "32", NULL },
         { "33", NULL },
         { "34", NULL },
         { "35", NULL },
         { "36", NULL },
         { "37", NULL },
         { "38", NULL },
         { "39", NULL },
         { "40", NULL },
         { "41", NULL },
         { "42", NULL },
         { "43", NULL },
         { "44", NULL },
         { "45", NULL },
         { "46", NULL },
         { "47", NULL },
         { "48", NULL },
         { "49", NULL },
         { "50", NULL },
         { "51", NULL },
         { "52", NULL },
         { "53", NULL },
         { "54", NULL },
         { "55", NULL },
         { "56", NULL },
         { "57", NULL },
         { "58", NULL },
         { "59", NULL },
         { "60", NULL },
         { "61", NULL },
         { "62", NULL },
         { "63", NULL },
         { NULL, NULL },
      },
      "1"
   },
   {
      BEETLE_OPT(line_render),
      BEETLE_OPT_LINE_RENDER_LABEL_RU,
      NULL,
      BEETLE_OPT_LINE_RENDER_INFO_0_RU,
      NULL,
      "hacks",
      {
         { "default",    OPTION_VAL_DEFAULT_RU },
         { "aggressive", OPTION_VAL_AGGRESSIVE_RU },
         { "disabled",   NULL },
         { NULL, NULL },
      },
      "default"
   },
   {
      BEETLE_OPT(widescreen_hack),
      BEETLE_OPT_WIDESCREEN_HACK_LABEL_RU,
      NULL,
      BEETLE_OPT_WIDESCREEN_HACK_INFO_0_RU,
      NULL,
      "hacks",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(widescreen_hack_aspect_ratio),
      BEETLE_OPT_WIDESCREEN_HACK_ASPECT_RATIO_LABEL_RU,
      NULL,
      BEETLE_OPT_WIDESCREEN_HACK_ASPECT_RATIO_INFO_0_RU,
      NULL,
      "hacks",
      {
         { "16:9",  OPTION_VAL_16_9_RU },
         { "16:10", OPTION_VAL_16_10_RU },
         { "18:9",  OPTION_VAL_18_9_RU },
         { "19:9",  OPTION_VAL_19_9_RU },
         { "20:9",  OPTION_VAL_20_9_RU },
         { "21:9",  OPTION_VAL_21_9_RU }, // 64:27
         { "32:9",  OPTION_VAL_32_9_RU },
         { NULL,    NULL },
      },
      "16:9"
   },
#if defined(HAVE_LIGHTREC)
   {
      BEETLE_OPT(cpu_dynarec),
      BEETLE_OPT_CPU_DYNAREC_LABEL_RU,
      NULL,
      BEETLE_OPT_CPU_DYNAREC_INFO_0_RU,
      NULL,
      "hacks",
      {
         { "disabled",        OPTION_VAL_DISABLED_RU },
         { "execute",         OPTION_VAL_EXECUTE_RU },
         { "execute_one",     OPTION_VAL_EXECUTE_ONE_RU },
         { "run_interpreter", OPTION_VAL_RUN_INTERPRETER_RU },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(dynarec_invalidate),
      BEETLE_OPT_DYNAREC_INVALIDATE_LABEL_RU,
      NULL,
      BEETLE_OPT_DYNAREC_INVALIDATE_INFO_0_RU,
      NULL,
      "hacks",
      {
         { "full", OPTION_VAL_FULL_RU },
         { "dma",  OPTION_VAL_DMA_RU },
         { NULL, NULL },
      },
      "full"
   },
   {
      BEETLE_OPT(dynarec_eventcycles),
      BEETLE_OPT_DYNAREC_EVENTCYCLES_LABEL_RU,
      NULL,
      BEETLE_OPT_DYNAREC_EVENTCYCLES_INFO_0_RU,
      NULL,
      "hacks",
      {
         { "128",  OPTION_VAL_128_RU },
         { "256",  NULL },
         { "384",  NULL },
         { "512",  NULL },
         { "640",  NULL },
         { "768",  NULL },
         { "896",  NULL },
         { "1024", NULL },
         { "1152", NULL },
         { "1280", NULL },
         { "1408", NULL },
         { "1536", NULL },
         { "1664", NULL },
         { "1792", NULL },
         { "1920", NULL },
         { "2048", NULL },
         { NULL, NULL },
      },
      "128"
   },
   {
      BEETLE_OPT(dynarec_spu_samples),
      BEETLE_OPT_DYNAREC_SPU_SAMPLES_LABEL_RU,
      NULL,
      BEETLE_OPT_DYNAREC_SPU_SAMPLES_INFO_0_RU,
      NULL,
      "hacks",
      {
         { "1",  OPTION_VAL_1_O49_RU },
         { "4",  NULL },
         { "16", NULL },
         { NULL, NULL },
      },
      "1"
   },
#endif
   {
      BEETLE_OPT(core_timing_fps),
      BEETLE_OPT_CORE_TIMING_FPS_LABEL_RU,
      NULL,
      BEETLE_OPT_CORE_TIMING_FPS_INFO_0_RU,
      NULL,
      "video",
      {
         { "auto_toggle",       OPTION_VAL_AUTO_TOGGLE_RU },
         { "force_progressive", OPTION_VAL_FORCE_PROGRESSIVE_RU },
         { "force_interlaced",  OPTION_VAL_FORCE_INTERLACED_RU },
      },
      "auto_toggle"
   },
   {
      BEETLE_OPT(deinterlacer),
      BEETLE_OPT_DEINTERLACER_LABEL_RU,
      NULL,
      BEETLE_OPT_DEINTERLACER_INFO_0_RU,
      NULL,
      "video",
      {
         { "weave",      OPTION_VAL_WEAVE_RU },
         { "bob",        OPTION_VAL_BOB_RU },
         { "bob_offset", OPTION_VAL_BOB_OFFSET_RU },
         { "fastmad",    OPTION_VAL_FASTMAD_RU },
         { "off",        "Off" },
         { NULL, NULL },
      },
      "weave"
   },
   {
      BEETLE_OPT(frame_duping),
      BEETLE_OPT_FRAME_DUPING_LABEL_RU,
      NULL,
      BEETLE_OPT_FRAME_DUPING_INFO_0_RU,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(pal_video_timing_override),
      BEETLE_OPT_PAL_VIDEO_TIMING_OVERRIDE_LABEL_RU,
      NULL,
      BEETLE_OPT_PAL_VIDEO_TIMING_OVERRIDE_INFO_0_RU,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(aspect_ratio),
      BEETLE_OPT_ASPECT_RATIO_LABEL_RU,
      NULL,
      BEETLE_OPT_ASPECT_RATIO_INFO_0_RU,
      NULL,
      "video",
      {
         { "corrected", OPTION_VAL_CORRECTED_RU },
         { "uncorrected", OPTION_VAL_UNCORRECTED_RU },
         { "4:3",  OPTION_VAL_4_3_RU },
         { "ntsc", OPTION_VAL_NTSC_RU },
      },
      "corrected"
   },
   {
      BEETLE_OPT(crop_overscan),
      BEETLE_OPT_CROP_OVERSCAN_LABEL_RU,
      NULL,
      BEETLE_OPT_CROP_OVERSCAN_INFO_0_RU,
      NULL,
      "video",
      {
         { "disabled", "Disabled" },
         { "static",   OPTION_VAL_STATIC_RU },
         { "smart",    OPTION_VAL_SMART_RU },
         { NULL, NULL },
      },
      "smart"
   },
   {
      BEETLE_OPT(image_crop),
      BEETLE_OPT_IMAGE_CROP_LABEL_RU,
      NULL,
      BEETLE_OPT_IMAGE_CROP_INFO_0_RU,
      NULL,
      "video",
      {
         { "disabled", "0" },
         { "1px",      OPTION_VAL_1PX_RU },
         { "2px",      OPTION_VAL_2PX_RU },
         { "3px",      OPTION_VAL_3PX_RU },
         { "4px",      OPTION_VAL_4PX_RU },
         { "5px",      OPTION_VAL_5PX_RU },
         { "6px",      OPTION_VAL_6PX_RU },
         { "7px",      OPTION_VAL_7PX_RU },
         { "8px",      OPTION_VAL_8PX_RU },
         { "9px",      OPTION_VAL_9PX_RU },
         { "10px",     OPTION_VAL_10PX_RU },
         { "11px",     OPTION_VAL_11PX_RU },
         { "12px",     OPTION_VAL_12PX_RU },
         { "13px",     OPTION_VAL_13PX_RU },
         { "14px",     OPTION_VAL_14PX_RU },
         { "15px",     OPTION_VAL_15PX_RU },
         { "16px",     OPTION_VAL_16PX_RU },
         { "17px",     OPTION_VAL_17PX_RU },
         { "18px",     OPTION_VAL_18PX_RU },
         { "19px",     OPTION_VAL_19PX_RU },
         { "20px",     OPTION_VAL_20PX_RU },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(image_offset),
      BEETLE_OPT_IMAGE_OFFSET_LABEL_RU,
      NULL,
      BEETLE_OPT_IMAGE_OFFSET_INFO_0_RU,
      NULL,
      "video",
      {
         { "-12px",    OPTION_VAL_12PX_O64_RU },
         { "-11px",    OPTION_VAL_11PX_O64_RU },
         { "-10px",    OPTION_VAL_10PX_O64_RU },
         { "-9px",     OPTION_VAL_9PX_O64_RU },
         { "-8px",     OPTION_VAL_8PX_O64_RU },
         { "-7px",     OPTION_VAL_7PX_O64_RU },
         { "-6px",     OPTION_VAL_6PX_O64_RU },
         { "-5px",     OPTION_VAL_5PX_O64_RU },
         { "-4px",     OPTION_VAL_4PX_O64_RU },
         { "-3px",     OPTION_VAL_3PX_O64_RU },
         { "-2px",     OPTION_VAL_2PX_O64_RU },
         { "-1px",     OPTION_VAL_1PX_O64_RU },
         { "disabled", "0" },
         { "+1px",     OPTION_VAL_1PX_O64_0_RU },
         { "+2px",     OPTION_VAL_2PX_O64_0_RU },
         { "+3px",     OPTION_VAL_3PX_O64_0_RU },
         { "+4px",     OPTION_VAL_4PX_O64_0_RU },
         { "+5px",     OPTION_VAL_5PX_O64_0_RU },
         { "+6px",     OPTION_VAL_6PX_O64_0_RU },
         { "+7px",     OPTION_VAL_7PX_O64_0_RU },
         { "+8px",     OPTION_VAL_8PX_O64_0_RU },
         { "+9px",     OPTION_VAL_9PX_O64_0_RU },
         { "+10px",    OPTION_VAL_10PX_O64_0_RU },
         { "+11px",    OPTION_VAL_11PX_O64_0_RU },
         { "+12px",    OPTION_VAL_12PX_O64_0_RU },
         { NULL, NULL },
      },
      "disabled"
   },
#if defined(HAVE_OPENGL) || defined(HAVE_OPENGLES) || defined(HAVE_VULKAN)
   {
      BEETLE_OPT(image_offset_cycles),
      BEETLE_OPT_IMAGE_OFFSET_CYCLES_LABEL_RU,
      NULL,
      BEETLE_OPT_IMAGE_OFFSET_CYCLES_INFO_0_RU,
      NULL,
      "video",
      {
         { "-40",      NULL },
         { "-39",      NULL },
         { "-38",      NULL },
         { "-37",      NULL },
         { "-36",      NULL },
         { "-35",      NULL },
         { "-34",      NULL },
         { "-33",      NULL },
         { "-32",      NULL },
         { "-31",      NULL },
         { "-30",      NULL },
         { "-29",      NULL },
         { "-28",      NULL },
         { "-27",      NULL },
         { "-26",      NULL },
         { "-25",      NULL },
         { "-24",      NULL },
         { "-23",      NULL },
         { "-22",      NULL },
         { "-21",      NULL },
         { "-20",      NULL },
         { "-19",      NULL },
         { "-18",      NULL },
         { "-17",      NULL },
         { "-16",      NULL },
         { "-15",      NULL },
         { "-14",      NULL },
         { "-13",      NULL },
         { "-12",      NULL },
         { "-11",      NULL },
         { "-10",      NULL },
         { "-9",       NULL },
         { "-8",       NULL },
         { "-7",       NULL },
         { "-6",       NULL },
         { "-5",       NULL },
         { "-4",       NULL },
         { "-3",       NULL },
         { "-2",       NULL },
         { "-1",       NULL },
         { "0",        "0" },
         { "+1",       NULL },
         { "+2",       NULL },
         { "+3",       NULL },
         { "+4",       NULL },
         { "+5",       NULL },
         { "+6",       NULL },
         { "+7",       NULL },
         { "+8",       NULL },
         { "+9",       NULL },
         { "+10",      NULL },
         { "+11",      NULL },
         { "+12",      NULL },
         { "+13",      NULL },
         { "+14",      NULL },
         { "+15",      NULL },
         { "+16",      NULL },
         { "+17",      NULL },
         { "+18",      NULL },
         { "+19",      NULL },
         { "+20",      NULL },
         { "+21",      NULL },
         { "+22",      NULL },
         { "+23",      NULL },
         { "+24",      NULL },
         { "+25",      NULL },
         { "+26",      NULL },
         { "+27",      NULL },
         { "+28",      NULL },
         { "+29",      NULL },
         { "+30",      NULL },
         { "+31",      NULL },
         { "+32",      NULL },
         { "+33",      NULL },
         { "+34",      NULL },
         { "+35",      NULL },
         { "+36",      NULL },
         { "+37",      NULL },
         { "+38",      NULL },
         { "+39",      NULL },
         { "+40",      NULL },
         { NULL, NULL},
      },
      "0"
   },
#endif
   {
      BEETLE_OPT(initial_scanline),
      BEETLE_OPT_INITIAL_SCANLINE_LABEL_RU,
      NULL,
      BEETLE_OPT_INITIAL_SCANLINE_INFO_0_RU,
      NULL,
      "video",
      {
         { "0",  OPTION_VAL_0_O48_RU },
         { "1",  NULL },
         { "2",  NULL },
         { "3",  NULL },
         { "4",  NULL },
         { "5",  NULL },
         { "6",  NULL },
         { "7",  NULL },
         { "8",  NULL },
         { "9",  NULL },
         { "10", NULL },
         { "11", NULL },
         { "12", NULL },
         { "13", NULL },
         { "14", NULL },
         { "15", NULL },
         { "16", NULL },
         { "17", NULL },
         { "18", NULL },
         { "19", NULL },
         { "20", NULL },
         { "21", NULL },
         { "22", NULL },
         { "23", NULL },
         { "24", NULL },
         { "25", NULL },
         { "26", NULL },
         { "27", NULL },
         { "28", NULL },
         { "29", NULL },
         { "30", NULL },
         { "31", NULL },
         { "32", NULL },
         { "33", NULL },
         { "34", NULL },
         { "35", NULL },
         { "36", NULL },
         { "37", NULL },
         { "38", NULL },
         { "39", NULL },
         { "40", NULL },
         { NULL, NULL },
      },
      "0"
   },
   {
      BEETLE_OPT(last_scanline),
      BEETLE_OPT_LAST_SCANLINE_LABEL_RU,
      NULL,
      BEETLE_OPT_LAST_SCANLINE_INFO_0_RU,
      NULL,
      "video",
      {
         { "210", NULL },
         { "211", NULL },
         { "212", NULL },
         { "213", NULL },
         { "214", NULL },
         { "215", NULL },
         { "216", NULL },
         { "217", NULL },
         { "218", NULL },
         { "219", NULL },
         { "220", NULL },
         { "221", NULL },
         { "222", NULL },
         { "223", NULL },
         { "224", NULL },
         { "225", NULL },
         { "226", NULL },
         { "227", NULL },
         { "228", NULL },
         { "229", NULL },
         { "230", NULL },
         { "231", NULL },
         { "232", NULL },
         { "233", NULL },
         { "234", NULL },
         { "235", NULL },
         { "236", NULL },
         { "237", NULL },
         { "238", NULL },
         { "239", OPTION_VAL_239_RU },
         { NULL, NULL },
      },
      "239"
   },
   {
      BEETLE_OPT(initial_scanline_pal),
      BEETLE_OPT_INITIAL_SCANLINE_PAL_LABEL_RU,
      NULL,
      BEETLE_OPT_INITIAL_SCANLINE_PAL_INFO_0_RU,
      NULL,
      "video",
      {
         { "0",  OPTION_VAL_0_O48_RU },
         { "1",  NULL },
         { "2",  NULL },
         { "3",  NULL },
         { "4",  NULL },
         { "5",  NULL },
         { "6",  NULL },
         { "7",  NULL },
         { "8",  NULL },
         { "9",  NULL },
         { "10", NULL },
         { "11", NULL },
         { "12", NULL },
         { "13", NULL },
         { "14", NULL },
         { "15", NULL },
         { "16", NULL },
         { "17", NULL },
         { "18", NULL },
         { "19", NULL },
         { "20", NULL },
         { "21", NULL },
         { "22", NULL },
         { "23", NULL },
         { "24", NULL },
         { "25", NULL },
         { "26", NULL },
         { "27", NULL },
         { "28", NULL },
         { "29", NULL },
         { "30", NULL },
         { "31", NULL },
         { "32", NULL },
         { "33", NULL },
         { "34", NULL },
         { "35", NULL },
         { "36", NULL },
         { "37", NULL },
         { "38", NULL },
         { "39", NULL },
         { "40", NULL },
         { NULL, NULL },
      },
      "0"
   },
   {
      BEETLE_OPT(last_scanline_pal),
      BEETLE_OPT_LAST_SCANLINE_PAL_LABEL_RU,
      NULL,
      BEETLE_OPT_LAST_SCANLINE_PAL_INFO_0_RU,
      NULL,
      "video",
      {
         { "230", NULL },
         { "231", NULL },
         { "232", NULL },
         { "233", NULL },
         { "234", NULL },
         { "235", NULL },
         { "236", NULL },
         { "237", NULL },
         { "238", NULL },
         { "239", NULL },
         { "240", NULL },
         { "241", NULL },
         { "242", NULL },
         { "243", NULL },
         { "244", NULL },
         { "245", NULL },
         { "246", NULL },
         { "247", NULL },
         { "248", NULL },
         { "249", NULL },
         { "250", NULL },
         { "251", NULL },
         { "252", NULL },
         { "253", NULL },
         { "254", NULL },
         { "255", NULL },
         { "256", NULL },
         { "257", NULL },
         { "258", NULL },
         { "259", NULL },
         { "260", NULL },
         { "261", NULL },
         { "262", NULL },
         { "263", NULL },
         { "264", NULL },
         { "265", NULL },
         { "266", NULL },
         { "267", NULL },
         { "268", NULL },
         { "269", NULL },
         { "270", NULL },
         { "271", NULL },
         { "272", NULL },
         { "273", NULL },
         { "274", NULL },
         { "275", NULL },
         { "276", NULL },
         { "277", NULL },
         { "278", NULL },
         { "279", NULL },
         { "280", NULL },
         { "281", NULL },
         { "282", NULL },
         { "283", NULL },
         { "284", NULL },
         { "285", NULL },
         { "286", NULL },
         { "287", OPTION_VAL_287_RU },
         { NULL, NULL },
      },
      "287"
   },
   { NULL, NULL, NULL, NULL, NULL, NULL, {{0}}, NULL },
};
struct retro_core_options_v2 options_ru = {
   option_cats_ru,
   option_defs_ru
};

/* RETRO_LANGUAGE_SK */

#define CATEGORY_SYSTEM_LABEL_SK "Systém"
#define CATEGORY_SYSTEM_INFO_0_SK "Zmeňte možnosti systému a pretaktovania."
#define CATEGORY_VIDEO_LABEL_SK NULL
#define CATEGORY_VIDEO_INFO_0_SK "Zmeňte nastavenia pomeru strán, orezania obrazu, video filtra a vynechávania snímok."
#define CATEGORY_PGXP_LABEL_SK NULL
#define CATEGORY_PGXP_INFO_0_SK "Tieto možnosti riadia vylepšenia, ktoré môžu zlepšiť grafiku oproti pôvodnej konzole. PGXP dokáže odstrániť deformácie textúr a problémy so Z-fightingom."
#define CATEGORY_OSD_LABEL_SK "Upraviť zobrazenie"
#define CATEGORY_OSD_INFO_0_SK "Zmeňte upozornenia zobrazované na obrazovke."
#define CATEGORY_INPUT_LABEL_SK "Vstup"
#define CATEGORY_INPUT_INFO_0_SK "Zmeňte nastavenia svetelnej pištole, myši a ovládača neGcon."
#define CATEGORY_MEMCARDS_LABEL_SK "Pamäťová karta"
#define CATEGORY_MEMCARDS_INFO_0_SK "Zmeňte nastavenia týkajúce sa virtuálnej pamäťovej karty."
#define CATEGORY_HACKS_LABEL_SK "Emulačné hacky"
#define CATEGORY_HACKS_INFO_0_SK "Zmeňte nastavenia presnosti emulácie ovplyvňujúce nízkoúrovňový výkon a kompatibilitu."
#define BEETLE_OPT_CPU_FREQ_SCALE_LABEL_SK "Škálovanie frekvencie CPU (pretaktovanie)"
#define BEETLE_OPT_CPU_FREQ_SCALE_INFO_0_SK "Pretaktujte (alebo podtaktujte) emulovaný PSX CPU. Pretaktovanie môže odstrániť spomalenia a zvýšiť snímkovú frekvenciu v niektorých hrách na úkor zvýšených nárokov na výkon. Niektoré hry majú vstavaný obmedzovač snímkovej frekvencie a pretaktovanie im nemusí pomôcť. V niektorých tituloch môžu pri pretaktovaní určité efekty animovať rýchlejšie, než bolo zamýšľané."
#define OPTION_VAL_50_SK NULL
#define OPTION_VAL_60_SK NULL
#define OPTION_VAL_70_SK NULL
#define OPTION_VAL_80_SK NULL
#define OPTION_VAL_90_SK NULL
#define OPTION_VAL_100_SK "100% (natívne)"
#define OPTION_VAL_110_SK NULL
#define OPTION_VAL_120_SK NULL
#define OPTION_VAL_130_SK NULL
#define OPTION_VAL_140_SK NULL
#define OPTION_VAL_150_SK NULL
#define OPTION_VAL_160_SK NULL
#define OPTION_VAL_170_SK NULL
#define OPTION_VAL_180_SK NULL
#define OPTION_VAL_190_SK NULL
#define OPTION_VAL_200_SK NULL
#define OPTION_VAL_210_SK NULL
#define OPTION_VAL_220_SK NULL
#define OPTION_VAL_230_SK NULL
#define OPTION_VAL_240_SK NULL
#define OPTION_VAL_250_SK NULL
#define OPTION_VAL_260_SK NULL
#define OPTION_VAL_270_SK NULL
#define OPTION_VAL_280_SK NULL
#define OPTION_VAL_290_SK NULL
#define OPTION_VAL_300_SK NULL
#define OPTION_VAL_310_SK NULL
#define OPTION_VAL_320_SK NULL
#define OPTION_VAL_330_SK NULL
#define OPTION_VAL_340_SK NULL
#define OPTION_VAL_350_SK NULL
#define OPTION_VAL_360_SK NULL
#define OPTION_VAL_370_SK NULL
#define OPTION_VAL_380_SK NULL
#define OPTION_VAL_390_SK NULL
#define OPTION_VAL_400_SK NULL
#define OPTION_VAL_410_SK NULL
#define OPTION_VAL_420_SK NULL
#define OPTION_VAL_430_SK NULL
#define OPTION_VAL_440_SK NULL
#define OPTION_VAL_450_SK NULL
#define OPTION_VAL_460_SK NULL
#define OPTION_VAL_470_SK NULL
#define OPTION_VAL_480_SK NULL
#define OPTION_VAL_490_SK NULL
#define OPTION_VAL_500_SK NULL
#define OPTION_VAL_510_SK NULL
#define OPTION_VAL_520_SK NULL
#define OPTION_VAL_530_SK NULL
#define OPTION_VAL_540_SK NULL
#define OPTION_VAL_550_SK NULL
#define OPTION_VAL_560_SK NULL
#define OPTION_VAL_570_SK NULL
#define OPTION_VAL_580_SK NULL
#define OPTION_VAL_590_SK NULL
#define OPTION_VAL_600_SK NULL
#define OPTION_VAL_610_SK NULL
#define OPTION_VAL_620_SK NULL
#define OPTION_VAL_630_SK NULL
#define OPTION_VAL_640_SK NULL
#define OPTION_VAL_650_SK NULL
#define OPTION_VAL_660_SK NULL
#define OPTION_VAL_670_SK NULL
#define OPTION_VAL_680_SK NULL
#define OPTION_VAL_690_SK NULL
#define OPTION_VAL_700_SK NULL
#define OPTION_VAL_710_SK NULL
#define OPTION_VAL_720_SK NULL
#define OPTION_VAL_730_SK NULL
#define OPTION_VAL_740_SK NULL
#define OPTION_VAL_750_SK NULL
#define BEETLE_OPT_GPU_OVERCLOCK_LABEL_SK "Pretaktovanie GPU rasterizéra"
#define BEETLE_OPT_GPU_OVERCLOCK_INFO_0_SK "Povolí pretaktovanie 2D rasterizéra v emulovanom GPU PSX. Nezlepšuje 3D vykresľovanie a vo všeobecnosti má malý vplyv."
#define OPTION_VAL_1X_NATIVE_SK "1x (natívne)"
#define OPTION_VAL_2X_SK NULL
#define OPTION_VAL_4X_SK NULL
#define OPTION_VAL_8X_SK NULL
#define OPTION_VAL_16X_SK NULL
#define OPTION_VAL_32X_SK NULL
#define BEETLE_OPT_GTE_OVERCLOCK_LABEL_SK "Pretaktovanie GTE"
#define BEETLE_OPT_GTE_OVERCLOCK_INFO_0_SK "Zníži všetky emulované operácie GTE (CPU koprocesor pre 3D grafiku) na konštantnú latenciu jedného cyklu. V hrách, ktoré GTE intenzívne využívajú, to môže výrazne zlepšiť snímkovú frekvenciu a stabilitu časovania snímok."
#define BEETLE_OPT_CD_ACCESS_METHOD_LABEL_SK "Metóda prístupu k CD"
#define BEETLE_OPT_CD_ACCESS_METHOD_INFO_0_SK "'Synchronous' napodobňuje pôvodný hardvér. 'Asynchronous' môže redukovať trhanie na zariadeniach s pomalým úložiskom. 'Pre-Cache' načíta celý obraz disku do pamäte pri spustení obsahu, čo môže skrátiť načítavanie v hre za cenu úvodného zdržania pri štarte. 'Pre-Cache' môže spôsobiť problémy na systémoch s malou RAM a pri fyzických médiách sa vráti k 'Synchronous'. Vyžaduje reštart."
#define OPTION_VAL_SYNC_SK "Synchrónne"
#define OPTION_VAL_ASYNC_SK "Asynchrónne"
#define OPTION_VAL_PRECACHE_SK "Prednačítanie"
#define BEETLE_OPT_CD_FASTLOAD_LABEL_SK "Rýchlosť čítania CD"
#define BEETLE_OPT_CD_FASTLOAD_INFO_0_SK "Hodnoty vyššie ako '2x (Native)' môžu výrazne skrátiť načítavanie v hre, ale môžu spôsobiť chyby časovania. Niektoré hry nemusia nad určitou hodnotou fungovať správne."
#define OPTION_VAL_2X_NATIVE_SK "2x (natívne)"
#define OPTION_VAL_6X_SK NULL
#define OPTION_VAL_10X_SK NULL
#define OPTION_VAL_12X_SK NULL
#define OPTION_VAL_14X_SK NULL
#define BEETLE_OPT_SKIP_BIOS_LABEL_SK "Preskočiť BIOS"
#define BEETLE_OPT_SKIP_BIOS_INFO_0_SK "Preskočí bootovaciu animáciu BIOSu PlayStation, ktorá sa bežne zobrazuje pri načítavaní obsahu. Poznámka: povolenie spôsobuje problémy kompatibility u viacerých hier (PAL hry s ochranou proti kopírovaniu, Saga Frontier, atď.)."
#define BEETLE_OPT_OVERRIDE_BIOS_LABEL_SK "Prepísať BIOS"
#define BEETLE_OPT_OVERRIDE_BIOS_INFO_0_SK "Prepíše štandardný regionálny BIOS bezregionálnym, ak je dostupný. Vyžaduje reštart."
#define OPTION_VAL_PSXONPSP_SK NULL
#define OPTION_VAL_PS1_ROM_SK NULL
#define OPTION_VAL_OPENBIOS_SK NULL
#define BEETLE_OPT_REGION_LABEL_SK "Systémový región"
#define BEETLE_OPT_REGION_INFO_0_SK "Nastaví náhradný región, ktorý sa použije v prípade, že obsah nie je možné automaticky rozpoznať (napr. surové súbory PS-X EXE). Hry na diskoch naďalej používajú región zakódovaný na disku. Je potrebné reštartovať systém."
#define OPTION_VAL_AUTO_SK NULL
#define OPTION_VAL_NTSC_J_SK "NTSC-J (Japonsko)"
#define OPTION_VAL_NTSC_U_SK "NTSC-U (Severná Amerika)"
#define OPTION_VAL_PAL_SK "PAL (Európa)"
#define BEETLE_OPT_INTERNAL_RESOLUTION_LABEL_SK "Vnútorné rozlíšenie GPU"
#define BEETLE_OPT_INTERNAL_RESOLUTION_INFO_0_SK "Rozlíšenia vyššie ako '1x (Native)' zlepšujú vernosť 3D modelov za cenu vyšších nárokov na výkon. 2D prvky sú týmto nastavením vo všeobecnosti neovplyvnené."
#define BEETLE_OPT_RENDERER_LABEL_SK "Vykreslenie"
#define BEETLE_OPT_RENDERER_INFO_0_SK "Softvérový renderer je najpresnejší, ale pri zvýšených vnútorných rozlíšeniach GPU má vysoké nároky na výkon. Hardvérové renderery sú síce menej presné, no pri zvýšených vnútorných rozlíšeniach majú lepší výkon než softvérový renderer a umožňujú rôzne grafické vylepšenia. 'Hardware (Auto)' automaticky zvolí renderer Vulkan alebo OpenGL podľa aktuálneho video ovládača libretro frontendu. Ak poskytnutý video ovládač nie je Vulkan alebo OpenGL 3.3-kompatibilný, jadro sa vráti k softvérovému rendereru. Vyžaduje reštart."
#define OPTION_VAL_HARDWARE_SK "Hardvér (Auto)"
#define OPTION_VAL_HARDWARE_GL_SK "Hardvér (OpenGL)"
#define OPTION_VAL_HARDWARE_VK_SK "Hardvér (Vulkan)"
#define OPTION_VAL_SOFTWARE_SK "Softvér"
#define BEETLE_OPT_RENDERER_SOFTWARE_FB_LABEL_SK "Softvérový framebuffer"
#define BEETLE_OPT_RENDERER_SOFTWARE_FB_INFO_0_SK "Povolí presnú emuláciu efektov framebuffera (napr. motion blur, battle swirl vo FF7) pri použití hardvérových rendererov tak, že na pozadí beží kópia softvérového rendereru v natívnom rozlíšení. Pri vypnutí sa tieto operácie vynechajú (OpenGL) alebo vykreslia na GPU (Vulkan). Vypnutie môže zlepšiť výkon, ale môže spôsobiť závažné grafické chyby. Pri pochybnostiach nechajte zapnuté."
#define BEETLE_OPT_DEPTH_LABEL_SK "Vnútorná farebná hĺbka"
#define BEETLE_OPT_DEPTH_INFO_0_SK "Vyššia farebná hĺbka môže redukovať farebné pásy bez použitia ditheringu. 16 bpp emuluje pôvodný hardvér, ale ak nie je dithering povolený, môžu byť pásy viditeľné. Pri nastavení 32 bpp sa odporúča vypnúť 'Vzor ditheringu'."
#define OPTION_VAL_16BPP_NATIVE_SK "16 bpp (natívne)"
#define OPTION_VAL_32BPP_SK NULL
#define BEETLE_OPT_DITHER_MODE_LABEL_SK "Vzor ditheringu"
#define BEETLE_OPT_DITHER_MODE_INFO_0_SK "'1x (Native)' emuluje natívny nízkorozlišujúci dithering, ktorý pôvodný hardvér používa na zjemnenie farebných pásov viditeľných pri natívnej farebnej hĺbke. 'Internal Resolution' škáluje granularitu ditheringu na nastavené vnútorné rozlíšenie pre čistejší výsledok. Pri 32 bpp farebnej hĺbke sa odporúča vypnúť. Poznámka: vo Vulkane povolenie vynúti downsampling na natívnu farebnú hĺbku, kým vypnutie automaticky povolí výstup vo vyššej farebnej hĺbke."
#define OPTION_VAL_INTERNAL_RESOLUTION_SK "Vnútorné rozlíšenie"
#define BEETLE_OPT_SCALED_UV_OFFSET_LABEL_SK "Posun UV textúr"
#define BEETLE_OPT_SCALED_UV_OFFSET_INFO_0_SK "Vzorkovať textúry pre 3D polygóny s posunom pri internom rozlíšení vyššom ako 1x. Redukuje švy textúr, ale môže spôsobiť neúmyselné grafické chyby."
#define BEETLE_OPT_FILTER_LABEL_SK "Filtrovanie textúr"
#define BEETLE_OPT_FILTER_INFO_0_SK "'Nearest' emuluje pôvodný hardvér. 'Bilinear' a '3-Point' sú vyhladzovacie filtre, ktoré redukujú pixelizáciu rozmazaním. 'SABR', 'xBR' a 'JINC2' sú upscaling filtre, ktoré môžu zlepšiť vernosť/ostrosť textúr za cenu vyšších nárokov na výkon. Podporované len hardvérovými rendermi."
#define OPTION_VAL_NEAREST_SK "Najbližšie"
#define OPTION_VAL_SABR_SK NULL
#define OPTION_VAL_XBR_SK NULL
#define OPTION_VAL_BILINEAR_SK "Bilineárna"
#define OPTION_VAL_3_POINT_SK "3-bodové"
#define OPTION_VAL_JINC2_SK NULL
#define BEETLE_OPT_FILTER_EXCLUDE_SPRITE_LABEL_SK "Vylúčiť sprity z filtrovania"
#define BEETLE_OPT_FILTER_EXCLUDE_SPRITE_INFO_0_SK "Neaplikuje filtrovanie textúr na sprity. Zabraňuje vzniku švov v hrách s 2D pozadím tvoreným spritmi. Pre najlepší efekt použite spolu s Adaptívnym vyhladzovaním alebo iným post-processing filtrom."
#define OPTION_VAL_OPAQUE_SK "Iba nepriehľadné"
#define OPTION_VAL_ALL_SK "Nepriehľadné a polopriehľadné"
#define BEETLE_OPT_FILTER_EXCLUDE_2D_POLYGON_LABEL_SK "Vylúčiť 2D polygóny z filtrovania"
#define BEETLE_OPT_FILTER_EXCLUDE_2D_POLYGON_INFO_0_SK "Neaplikuje filtrovanie textúr na 2D polygóny. 2D polygóny sa detegujú heuristikou, môže dôjsť k chybám. Pre najlepší efekt použite spolu s Adaptívnym vyhladzovaním alebo iným post-processing filtrom."
#define BEETLE_OPT_ADAPTIVE_SMOOTHING_LABEL_SK "Adaptívne vyhladzovanie"
#define BEETLE_OPT_ADAPTIVE_SMOOTHING_INFO_0_SK "Vyhladí 2D grafiku a UI prvky bez rozmazania vykresľovaných 3D objektov. Podporované len rendererom Vulkan."
#define BEETLE_OPT_SUPER_SAMPLING_LABEL_SK "Supersampling (zníženie vzorkovania na natívne rozlíšenie)"
#define BEETLE_OPT_SUPER_SAMPLING_INFO_0_SK "Zníži vzorkovanie vykresleného obsahu z navýšeného vnútorného rozlíšenia späť na natívne rozlíšenie. Kombinácia s vyššími násobiteľmi vnútorného rozlíšenia umožňuje zobrazenie hier s anti-aliasovanými 3D objektmi v natívnom nízkom rozlíšení. Najlepšie výsledky dosahuje pri tituloch kombinujúcich 2D a 3D prvky (napr. 3D postavy na pred-renderovaných pozadiach) a funguje dobre v spojení s CRT shadermi. Podporované len rendererom Vulkan. Poznámka: pri povolení tejto možnosti sa odporúča vypnúť 'Vzor ditheringu'."
#define BEETLE_OPT_MSAA_LABEL_SK "Multi-sampled anti aliasing"
#define BEETLE_OPT_MSAA_INFO_0_SK "MSAA zlepšuje vzhľad 3D objektov. Podporované len rendererom Vulkan."
#define OPTION_VAL_1X_SK NULL
#define BEETLE_OPT_MDEC_YUV_LABEL_SK "MDEC YUV chroma filter"
#define BEETLE_OPT_MDEC_YUV_INFO_0_SK "Zlepší kvalitu prehrávania FMV redukciou 'macroblocking' artefaktov (štvorce/zubaté hrany). Podporované len rendererom Vulkan."
#define BEETLE_OPT_TRACK_TEXTURES_LABEL_SK "Sledovať textúry"
#define BEETLE_OPT_TRACK_TEXTURES_INFO_0_SK "Predpoklad pre dumpovanie a nahrádzanie textúr. Vo väčšine hier pravdepodobne spadne."
#define BEETLE_OPT_DUMP_TEXTURES_LABEL_SK "Dumpovať textúry"
#define BEETLE_OPT_DUMP_TEXTURES_INFO_0_SK "Dumpovať použité textúry do <cd>-texture-dump/"
#define BEETLE_OPT_REPLACE_TEXTURES_LABEL_SK "Nahradiť textúry"
#define BEETLE_OPT_REPLACE_TEXTURES_INFO_0_SK "Nahradiť textúry HD verziami z <cd>-texture-replacements/"
#define BEETLE_OPT_PGXP_MODE_LABEL_SK "Režim PGXP"
#define BEETLE_OPT_PGXP_MODE_INFO_0_SK "Vykresľuje 3D objekty so subpixelovou presnosťou, čo minimalizuje skreslenie a chvenie 3D objektov pozorované na pôvodnom hardvéri kvôli použitiu súradníc vrcholov s pevnou rádovou čiarkou. Režim 'Memory Only' má minimálne problémy s kompatibilitou a odporúča sa na bežné použitie. Režim 'Memory + CPU (Buggy)' môže chvenie znížiť ešte viac, no má vysoké nároky na výkon a môže spôsobiť rôzne chyby geometrie."
#define OPTION_VAL_MEMORY_ONLY_SK "Iba pamäť"
#define OPTION_VAL_MEMORY_CPU_SK "Pamäť + CPU (chybové)"
#define BEETLE_OPT_PGXP_2D_TOL_LABEL_SK "PGXP tolerancia 2D geometrie"
#define BEETLE_OPT_PGXP_2D_TOL_INFO_0_SK "Skryje výraznejšie chyby v operáciách PGXP: hodnota určuje toleranciu, pri ktorej sa PGXP hodnoty zachovajú aj pri geometriách bez správnej hĺbkovej informácie."
#define OPTION_VAL_0PX_SK NULL
#define OPTION_VAL_1PX_SK NULL
#define OPTION_VAL_2PX_SK NULL
#define OPTION_VAL_3PX_SK NULL
#define OPTION_VAL_4PX_SK NULL
#define OPTION_VAL_5PX_SK NULL
#define OPTION_VAL_6PX_SK NULL
#define OPTION_VAL_7PX_SK NULL
#define OPTION_VAL_8PX_SK NULL
#define BEETLE_OPT_PGXP_NCLIP_LABEL_SK "PGXP eliminácia primitív"
#define BEETLE_OPT_PGXP_NCLIP_INFO_0_SK "Použiť implementáciu NCLIP od PGXP. Zlepšuje vzhľad redukciou dier v geometriách so súradnicami PGXP. V niektorých hrách spôsobuje zamrznutie za rôznych okolností."
#define BEETLE_OPT_PGXP_VERTEX_LABEL_SK "PGXP cache vrcholov"
#define BEETLE_OPT_PGXP_VERTEX_INFO_0_SK "Ukladá pozície vrcholov vylepšených PGXP do cache pre opakované použitie naprieč vykresľovaním polygónov. Môže potenciálne zlepšiť zarovnanie objektov a redukovať viditeľné švy pri vykresľovaní textúr, no chybné zhody pri dotazovaní cache môžu spôsobiť grafické chyby. V súčasnosti sa odporúča nechať túto možnosť vypnutú. Možnosť sa aplikuje len ak je povolený PGXP režim. Podporované len hardvérovými rendermi."
#define BEETLE_OPT_PGXP_TEXTURE_LABEL_SK "PGXP perspektívne korektné texturovanie"
#define BEETLE_OPT_PGXP_TEXTURE_INFO_0_SK "Nahradiť pôvodné afínne mapovanie textúr PSX mapovaním textúr s korekciou perspektívy. Eliminuje skreslenie a deformáciu textúr závislé od polohy, čím sa dosiahne správne zarovnanie textúr. Táto možnosť sa uplatňuje len v prípade, ak je zapnutý prevádzkový režim PGXP. Podporujú ju hardvérové aj softvérové renderery."
#define BEETLE_OPT_DISPLAY_INTERNAL_FPS_LABEL_SK "Zobraziť vnútorné FPS"
#define BEETLE_OPT_DISPLAY_INTERNAL_FPS_INFO_0_SK "Zobrazí vnútornú frekvenciu snímok, ktorou emulovaný systém PlayStation vykresľuje obsah. Poznámka: vyžaduje povolené onscreen oznámenia v libretro frontende."
#define BEETLE_OPT_DISPLAY_OSD_LABEL_SK "Zobraziť OSD správy"
#define BEETLE_OPT_DISPLAY_OSD_INFO_0_SK "Zobrazí OSD správy generované jadrom."
#define BEETLE_OPT_DISPLAY_VRAM_LABEL_SK "Zobraziť celú VRAM (debug)"
#define BEETLE_OPT_DISPLAY_VRAM_INFO_0_SK "Vizualizuje celú VRAM emulovanej konzoly. Podporované len hardvérovými rendererami OpenGL a Vulkan. Poznámka: slúži na ladenie a malo by byť bežne vypnuté."
#define BEETLE_OPT_ANALOG_CALIBRATION_LABEL_SK "Auto-kalibrácia analógu"
#define BEETLE_OPT_ANALOG_CALIBRATION_INFO_0_SK "Povolí dynamickú kalibráciu analógových vstupov, keď je zariadenie nastavené na DualShock, Analog Controller, Analog Joystick alebo neGcon. Maximálne zaznamenané hodnoty vstupu sa sledujú v reálnom čase a používajú sa na škálovanie analógových súradníc odovzdávaných emulátoru. Toto by sa malo používať pri hrách ako Mega Man Legends 2, ktoré očakávajú väčšie hodnoty než poskytujú moderné ovládače. Pre najlepšie výsledky by sa analógové páčky pri každom načítaní obsahu mali otočiť na maximum, aby sa naladil kalibračný algoritmus."
#define BEETLE_OPT_ANALOG_TOGGLE_LABEL_SK "DualShock prepínanie analógového režimu"
#define BEETLE_OPT_ANALOG_TOGGLE_INFO_0_SK "Keď je typ vstupného zariadenia DualShock, táto možnosť umožňuje prepínať emulovaný DualShock medzi režimom DIGITAL a ANALOG ako na pôvodnom hardvéri. Režim možno prepnúť aj pomocou zvolenej kombinácie tlačidiel."
#define OPTION_VAL_ENABLED_ANALOG_SK "Predvolene-Analóg"
#define BEETLE_OPT_ANALOG_TOGGLE_COMBO_LABEL_SK "DualShock kombinácia analógového režimu"
#define BEETLE_OPT_ANALOG_TOGGLE_COMBO_INFO_0_SK "Kombinácia tlačidiel použitá na prepínanie medzi režimom DIGITAL a ANALOG emulovaného DualShocku."
#define OPTION_VAL_L1_L2_R1_R2_START_SELECT_SK NULL
#define OPTION_VAL_L1_R1_SELECT_SK NULL
#define OPTION_VAL_L1_R1_START_SK NULL
#define OPTION_VAL_L1_R1_L3_SK NULL
#define OPTION_VAL_L1_R1_R3_SK NULL
#define OPTION_VAL_L2_R2_SELECT_SK NULL
#define OPTION_VAL_L2_R2_START_SK NULL
#define OPTION_VAL_L2_R2_L3_SK NULL
#define OPTION_VAL_L2_R2_R3_SK NULL
#define OPTION_VAL_L3_R3_SK NULL
#define BEETLE_OPT_ANALOG_TOGGLE_HOLD_LABEL_SK "Oneskorenie podržania DualShock analóg kombinácie"
#define BEETLE_OPT_ANALOG_TOGGLE_HOLD_INFO_0_SK "Čas podržania tlačidiel kombinácie analógového režimu."
#define OPTION_VAL_0_SK "Trvanie 0 sekúnd"
#define OPTION_VAL_1_SK "Trvanie 1 sekunda"
#define OPTION_VAL_2_SK "Trvanie 2 sekundy"
#define OPTION_VAL_3_SK "Trvanie 3 sekundy"
#define OPTION_VAL_4_SK "Trvanie 4 sekundy"
#define OPTION_VAL_5_SK "Trvanie 5 sekúnd"
#define BEETLE_OPT_ENABLE_MULTITAP_PORT1_LABEL_SK "Multitap na porte 1"
#define BEETLE_OPT_ENABLE_MULTITAP_PORT1_INFO_0_SK "Povolí funkciu multitap na porte 1."
#define BEETLE_OPT_ENABLE_MULTITAP_PORT2_LABEL_SK "Multitap na porte 2"
#define BEETLE_OPT_ENABLE_MULTITAP_PORT2_INFO_0_SK "Povolí funkciu multitap na porte 2."
#define BEETLE_OPT_GUN_INPUT_MODE_LABEL_SK "Režim svetelnej zbrane"
#define BEETLE_OPT_GUN_INPUT_MODE_INFO_0_SK "Použije vstup 'Light Gun' alebo 'Touchscreen' ovládaný myšou, keď je typ zariadenia nastavený na 'Guncon/G-Con 45' alebo 'Justifier'."
#define OPTION_VAL_LIGHTGUN_SK "Svetelná pištoľ"
#define OPTION_VAL_TOUCHSCREEN_SK "Dotykový displej"
#define BEETLE_OPT_GUN_CURSOR_LABEL_SK "Kurzor zbrane"
#define BEETLE_OPT_GUN_CURSOR_INFO_0_SK "Vykreslí kurzor na obrazovke pri použití typov vstupného zariadenia 'Guncon/G-Con 45' a 'Justifier'."
#define OPTION_VAL_CROSS_SK "Kríž"
#define OPTION_VAL_DOT_SK "Bod"
#define BEETLE_OPT_CROSSHAIR_COLOR_P1_LABEL_SK "Farba zameriavača zbrane: Port 1"
#define BEETLE_OPT_CROSSHAIR_COLOR_P1_INFO_0_SK "Farba zameriavača svetelnej zbrane pre port 1."
#define OPTION_VAL_RED_SK "Červená"
#define OPTION_VAL_BLUE_SK "Modrá"
#define OPTION_VAL_GREEN_SK "Zelená"
#define OPTION_VAL_ORANGE_SK "Oranžová"
#define OPTION_VAL_YELLOW_SK "Žltá"
#define OPTION_VAL_CYAN_SK "Azúrová"
#define OPTION_VAL_PINK_SK "Ružová"
#define OPTION_VAL_PURPLE_SK "Fialová"
#define OPTION_VAL_BLACK_SK "Čierna"
#define OPTION_VAL_WHITE_SK "Biela"
#define BEETLE_OPT_CROSSHAIR_COLOR_P2_LABEL_SK "Farba zameriavača zbrane: Port 2"
#define BEETLE_OPT_CROSSHAIR_COLOR_P2_INFO_0_SK "Farba zameriavača svetelnej zbrane pre port 2."
#define BEETLE_OPT_MOUSE_SENSITIVITY_LABEL_SK "Citlivosť myši"
#define BEETLE_OPT_MOUSE_SENSITIVITY_INFO_0_SK "Citlivosť vstupného zariadenia typu 'Mouse'."
#define OPTION_VAL_5_O42_SK NULL
#define OPTION_VAL_10_SK NULL
#define OPTION_VAL_15_SK NULL
#define OPTION_VAL_20_SK NULL
#define OPTION_VAL_25_SK NULL
#define OPTION_VAL_30_SK NULL
#define OPTION_VAL_35_SK NULL
#define OPTION_VAL_40_SK NULL
#define OPTION_VAL_45_SK NULL
#define OPTION_VAL_55_SK NULL
#define OPTION_VAL_65_SK NULL
#define OPTION_VAL_75_SK NULL
#define OPTION_VAL_85_SK NULL
#define OPTION_VAL_95_SK NULL
#define OPTION_VAL_100_O42_SK NULL
#define OPTION_VAL_105_SK NULL
#define OPTION_VAL_115_SK NULL
#define OPTION_VAL_125_SK NULL
#define OPTION_VAL_135_SK NULL
#define OPTION_VAL_145_SK NULL
#define OPTION_VAL_155_SK NULL
#define OPTION_VAL_165_SK NULL
#define OPTION_VAL_175_SK NULL
#define OPTION_VAL_185_SK NULL
#define OPTION_VAL_195_SK NULL
#define BEETLE_OPT_NEGCON_RESPONSE_LABEL_SK "neGcon Twist - odozva"
#define BEETLE_OPT_NEGCON_RESPONSE_INFO_0_SK "Typ odozvy ľavej analógovej páčky pri simulácii akcie 'twist' emulovaných vstupných zariadení 'neGcon'. Posunutie analógovej páčky možno mapovať na uhol rotácie neGcon lineárne, kvadraticky alebo kubicky. 'Quadratic' poskytuje pri malých pohyboch väčšiu presnosť než 'Linear'. 'Cubic' ešte viac zvyšuje presnosť malých pohybov, no 'preexponuje' väčšie pohyby. Poznámka: 'Linear' sa odporúča iba pri použití pretekárskych volantov. Konvenčné ovládače implementujú analógový vstup zásadne odlišne od mechanizmu 'twist' neGcon, takže lineárne mapovanie príliš zosilňuje malé pohyby, čo zhoršuje jemnú kontrolu. Vo väčšine prípadov 'Quadratic' poskytuje najbližšiu aproximáciu skutočného hardvéru."
#define OPTION_VAL_LINEAR_SK "Lineárny"
#define OPTION_VAL_QUADRATIC_SK "Kvadratický"
#define OPTION_VAL_CUBIC_SK "Kubické"
#define BEETLE_OPT_NEGCON_DEADZONE_LABEL_SK "neGcon Twist - mŕtva zóna"
#define BEETLE_OPT_NEGCON_DEADZONE_INFO_0_SK "Mŕtva zóna ľavej analógovej páčky pri simulácii akcie 'twist' emulovaných vstupných zariadení 'neGcon'. Používa sa na elimináciu driftu ovládača. Poznámka: väčšina titulov kompatibilných s neGcon poskytuje v hre možnosti na nastavenie hodnoty mŕtvej zóny 'twist'. Pre zachovanie presnosti by mŕtva zóna v hre mala byť *vždy* nastavená na nulu. Akékoľvek potrebné úpravy by sa mali aplikovať *iba* cez túto možnosť jadra. Je to obzvlášť dôležité, keď je 'neGcon Twist Response' nastavené na 'Quadratic' alebo 'Cubic'."
#define OPTION_VAL_0_O44_SK NULL
#define BEETLE_OPT_USE_MEDNAFEN_MEMCARD0_METHOD_LABEL_SK "Metóda pamäťovej karty"
#define BEETLE_OPT_USE_MEDNAFEN_MEMCARD0_METHOD_INFO_0_SK "Formát uložených dát použitý pre ľavý slot pamäťovej karty. 'Mednafen' možno použiť pre kompatibilitu so samostatnou verziou Mednafen. Pri použití s Beetle PSX majú uloženia Libretro (.srm) a Mednafen (.mcr) interne identický formát a možno ich medzi sebou konvertovať premenovaním. Vyžaduje reštart."
#define OPTION_VAL_LIBRETRO_SK NULL
#define OPTION_VAL_MEDNAFEN_SK NULL
#define BEETLE_OPT_ENABLE_MEMCARD1_LABEL_SK "Slot pamäťovej karty 2"
#define BEETLE_OPT_ENABLE_MEMCARD1_INFO_0_SK "Emuluje druhú pamäťovú kartu v pravom slote pamäťovej karty. Poznámka: niektoré hry vyžadujú vypnutie tejto možnosti pre správnu funkciu (napr. Codename Tenka). Poznámka: Slot pamäťovej karty 2 používa formát uloženia Mednafen (.mcr). Vyžaduje reštart."
#define BEETLE_OPT_SHARED_MEMORY_CARDS_LABEL_SK "Zdieľané pamäťové karty"
#define BEETLE_OPT_SHARED_MEMORY_CARDS_INFO_0_SK "Použije zdieľané pamäťové karty namiesto samostatných pre každý obsah. Poznámka: ak je 'Memory Card Method' nastavená na 'Libretro', ovplyvní to len slot pamäťovej karty 2. Vyžaduje reštart."
#define BEETLE_OPT_MEMCARD_LEFT_INDEX_LABEL_SK "Index slotu pamäťovej karty 1"
#define BEETLE_OPT_MEMCARD_LEFT_INDEX_INFO_0_SK "Pamäťová karta aktuálne načítaná v ľavom slote. Možnosť funguje len ak je 'Metóda pamäťovej karty' nastavená na Mednafen."
#define OPTION_VAL_0_O48_SK "0 (predvolené)"
#define BEETLE_OPT_MEMCARD_RIGHT_INDEX_LABEL_SK "Index slotu pamäťovej karty 2"
#define BEETLE_OPT_MEMCARD_RIGHT_INDEX_INFO_0_SK "Pamäťová karta aktuálne načítaná v pravom slote. Možnosť funguje len ak je povolený Slot pamäťovej karty 2."
#define OPTION_VAL_1_O49_SK "1 (predvolené)"
#define BEETLE_OPT_LINE_RENDER_LABEL_SK "Hack čiara-na-quad"
#define BEETLE_OPT_LINE_RENDER_INFO_0_SK "Niektoré hry (napr. Doom, Hexen, Soul Blade, atď.) kreslia horizontálne čiary natiahnutím trojuholníkov vysokých jeden pixel cez obrazovku, ktoré sa na pôvodnom hardvéri rasterizujú ako riadok pixelov. Tento hack tieto malé trojuholníky detekuje a podľa potreby ich prevedie na quady, čo umožňuje ich správne zobrazenie hardvérovými rendermi a vo vyšších vnútorných rozlíšeniach. 'Aggressive' je vyžadované pre niektoré tituly (napr. Dark Forces, Duke Nukem), no môže inde spôsobiť grafické chyby. Pri pochybnostiach nechajte na 'Default'."
#define OPTION_VAL_DEFAULT_SK "Predvolené"
#define OPTION_VAL_AGGRESSIVE_SK "Agresívne"
#define BEETLE_OPT_WIDESCREEN_HACK_LABEL_SK "Hack širokouhlého režimu"
#define BEETLE_OPT_WIDESCREEN_HACK_INFO_0_SK "Vykreslí 3D obsah anamorficky a výstupný emulovaný framebuffer poskytne v širokouhlom pomere strán. Najlepšie výsledky dosahuje pri plne 3D hrách. 2D prvky budú horizontálne natiahnuté a môžu byť zle zarovnané."
#define BEETLE_OPT_WIDESCREEN_HACK_ASPECT_RATIO_LABEL_SK "Pomer strán hacku širokouhlého režimu"
#define BEETLE_OPT_WIDESCREEN_HACK_ASPECT_RATIO_INFO_0_SK "Pomer strán použitý pre Hack širokouhlého režimu."
#define OPTION_VAL_16_9_SK NULL
#define OPTION_VAL_16_10_SK NULL
#define OPTION_VAL_18_9_SK NULL
#define OPTION_VAL_19_9_SK NULL
#define OPTION_VAL_20_9_SK NULL
#define OPTION_VAL_21_9_SK NULL
#define OPTION_VAL_32_9_SK NULL
#define BEETLE_OPT_CPU_DYNAREC_LABEL_SK NULL
#define BEETLE_OPT_CPU_DYNAREC_INFO_0_SK "Dynamicky rekompiluje inštrukcie CPU na natívne inštrukcie. Oveľa rýchlejšie ako interpreter, ale časovanie CPU je menej presné a môžu sa vyskytnúť chyby."
#define OPTION_VAL_DISABLED_SK "Vypnuté (Beetle Interpreter)"
#define OPTION_VAL_EXECUTE_SK "Maximálny výkon"
#define OPTION_VAL_EXECUTE_ONE_SK "Kontrola časovania cyklov"
#define OPTION_VAL_RUN_INTERPRETER_SK "Lightrec interpreter"
#define BEETLE_OPT_DYNAREC_INVALIDATE_LABEL_SK "Invalidácia dynarec kódu"
#define BEETLE_OPT_DYNAREC_INVALIDATE_INFO_0_SK "Niektoré hry vyžadujú 'Full' invalidáciu, iné 'DMA Only'."
#define OPTION_VAL_FULL_SK "Plné"
#define OPTION_VAL_DMA_SK "Iba DMA (mierne rýchlejšie)"
#define BEETLE_OPT_DYNAREC_EVENTCYCLES_LABEL_SK "Dynarec cykly udalostí DMA/GPU/MDEC/časovača"
#define BEETLE_OPT_DYNAREC_EVENTCYCLES_INFO_0_SK "Maximálny počet cyklov vykonaných CPU pred kontrolou aktualizácie GPU/DMA/MDEC/časovača. Vyššia hodnota je rýchlejšia, na beetle interpreter má oveľa menší vplyv než na dynarec."
#define OPTION_VAL_128_SK "128 (predvolené)"
#define BEETLE_OPT_DYNAREC_SPU_SAMPLES_LABEL_SK "Dynarec SPU vzorky"
#define BEETLE_OPT_DYNAREC_SPU_SAMPLES_INFO_0_SK "Maximálny počet SPU vzoriek pred kontrolou aktualizácie SPU. Vyššia hodnota je rýchlejšia, no pri iných hodnotách než 1 spôsobí v niektorých hrách zvukové chyby."
#define BEETLE_OPT_CORE_TIMING_FPS_LABEL_SK "Časovanie FPS hlásené jadrom"
#define BEETLE_OPT_CORE_TIMING_FPS_INFO_0_SK "Automatické prepínanie umožňuje jadru prepínať medzi progresívnou a prekladanou frekvenciou."
#define OPTION_VAL_AUTO_TOGGLE_SK "Automatické prepínanie"
#define OPTION_VAL_FORCE_PROGRESSIVE_SK "Vynútiť progresívnu frekvenciu"
#define OPTION_VAL_FORCE_INTERLACED_SK "Vynútiť prekladanú frekvenciu"
#define BEETLE_OPT_DEINTERLACER_LABEL_SK "Metóda deinterlace"
#define BEETLE_OPT_DEINTERLACER_INFO_0_SK "Režim „Weave“ zobrazuje obe polia v jednom snímku, čím sa dosahuje statický obraz v plnom vysokom rozlíšení, zatiaľ čo režim „Bob“ uprednostňuje plynulosť pohybu na úkor ostrosti obrazu. „Bob (Offset)“ je Bob s aktuálnym poľom zdvojeným na plnú výšku povrchu s posunom jednej línie medzi poľami, čím sa zachováva výstupné rozlíšenie libretro a zároveň sa zachová plynulý pohyb v štýle Bob. „FastMAD (Motion Adaptive)“ preplieta statické pixely a vertikálne interpoluje pohybujúce sa pixely na pixel, s plnou výškou výstupu bez hrebeňového efektu pri pohybe ani zníženého rozlíšenia Bob. „Off“ zosúlaďuje správanie softvérového renderera so správaním hardvérového (Vulkan/OpenGL) renderera rasterizáciou do všetkých riadkov VRAM v každom snímku a odložením výstupu na koniec snímku, čím sa eliminuje hrebeňový efekt pri pohybe v prekladaných hrách. Poznámka: programy, ktoré počas skenovania používajú DMA do VRAM (niektoré dema využívajúce triky s obalením framebufferu PAL s viac ako 512 riadkami), alebo ktoré aktualizujú pohybujúce sa sprajty prostredníctvom prenosov obrazu z CPU do VRAM viazaných na zobrazené pole (napr. Pro Pinball), môžu pri nastavení „Off“ zobrazovať neaktuálne riadky alebo duchovité obrazy sprajtov; pre tieto prípady použite „Weave“, „Bob“, „Bob (Offset)“ alebo „FastMAD“."
#define OPTION_VAL_WEAVE_SK "Výplet"
#define OPTION_VAL_BOB_SK NULL
#define OPTION_VAL_BOB_OFFSET_SK NULL
#define OPTION_VAL_FASTMAD_SK NULL
#define BEETLE_OPT_FRAME_DUPING_LABEL_SK "Duplikovanie snímok"
#define BEETLE_OPT_FRAME_DUPING_INFO_0_SK "Zopakuje predošlú snímku, ak jadro nemá čo nové zobraziť."
#define BEETLE_OPT_PAL_VIDEO_TIMING_OVERRIDE_LABEL_SK "Prepísanie PAL video časovania"
#define BEETLE_OPT_PAL_VIDEO_TIMING_OVERRIDE_INFO_0_SK "Kvôli odlišným štandardom sa PAL hry oproti americkým alebo japonským NTSC vydaniam často zdajú spomalené. Táto možnosť umožňuje prepísať PAL časovanie a pokúsiť sa tieto hry spustiť s NTSC snímkovou frekvenciou."
#define BEETLE_OPT_ASPECT_RATIO_LABEL_SK "Pomer strán jadra"
#define BEETLE_OPT_ASPECT_RATIO_INFO_0_SK "Toto nastavenie sa ignoruje, keď je povolený Hack širokouhlého režimu alebo Zobraziť celú VRAM."
#define OPTION_VAL_CORRECTED_SK "Opravené"
#define OPTION_VAL_UNCORRECTED_SK "Neopravené"
#define OPTION_VAL_4_3_SK "Vynútiť 4:3"
#define OPTION_VAL_NTSC_SK "Vynútiť NTSC"
#define BEETLE_OPT_CROP_OVERSCAN_LABEL_SK "Orezať overscan"
#define BEETLE_OPT_CROP_OVERSCAN_INFO_0_SK "Plný natívny obraz alebo bez okrajov."
#define OPTION_VAL_STATIC_SK "Vodorovne"
#define OPTION_VAL_SMART_SK "Horizontálne + vertikálne"
#define BEETLE_OPT_IMAGE_CROP_LABEL_SK "Dodatočné orezanie"
#define BEETLE_OPT_IMAGE_CROP_INFO_0_SK "Zníži šírku orezaného obrazu o zadaný počet pixelov."
#define OPTION_VAL_9PX_SK NULL
#define OPTION_VAL_10PX_SK NULL
#define OPTION_VAL_11PX_SK NULL
#define OPTION_VAL_12PX_SK NULL
#define OPTION_VAL_13PX_SK NULL
#define OPTION_VAL_14PX_SK NULL
#define OPTION_VAL_15PX_SK NULL
#define OPTION_VAL_16PX_SK NULL
#define OPTION_VAL_17PX_SK NULL
#define OPTION_VAL_18PX_SK NULL
#define OPTION_VAL_19PX_SK NULL
#define OPTION_VAL_20PX_SK NULL
#define BEETLE_OPT_IMAGE_OFFSET_LABEL_SK "Posun orezaného obrazu"
#define BEETLE_OPT_IMAGE_OFFSET_INFO_0_SK "Posunie horizontálne vpravo (kladná hodnota) alebo vľavo (záporná hodnota) o zadaný počet pixelov. Možno použiť na korekciu zarovnania. Podporované len softvérovým rendererom."
#define OPTION_VAL_12PX_O64_SK NULL
#define OPTION_VAL_11PX_O64_SK NULL
#define OPTION_VAL_10PX_O64_SK NULL
#define OPTION_VAL_9PX_O64_SK NULL
#define OPTION_VAL_8PX_O64_SK NULL
#define OPTION_VAL_7PX_O64_SK NULL
#define OPTION_VAL_6PX_O64_SK NULL
#define OPTION_VAL_5PX_O64_SK NULL
#define OPTION_VAL_4PX_O64_SK NULL
#define OPTION_VAL_3PX_O64_SK NULL
#define OPTION_VAL_2PX_O64_SK NULL
#define OPTION_VAL_1PX_O64_SK NULL
#define OPTION_VAL_1PX_O64_0_SK NULL
#define OPTION_VAL_2PX_O64_0_SK NULL
#define OPTION_VAL_3PX_O64_0_SK NULL
#define OPTION_VAL_4PX_O64_0_SK NULL
#define OPTION_VAL_5PX_O64_0_SK NULL
#define OPTION_VAL_6PX_O64_0_SK NULL
#define OPTION_VAL_7PX_O64_0_SK NULL
#define OPTION_VAL_8PX_O64_0_SK NULL
#define OPTION_VAL_9PX_O64_0_SK NULL
#define OPTION_VAL_10PX_O64_0_SK NULL
#define OPTION_VAL_11PX_O64_0_SK NULL
#define OPTION_VAL_12PX_O64_0_SK NULL
#define BEETLE_OPT_IMAGE_OFFSET_CYCLES_LABEL_SK "Horizontálny posun obrazu (GPU cykly)"
#define BEETLE_OPT_IMAGE_OFFSET_CYCLES_INFO_0_SK "Kladné hodnoty posúvajú obraz vpravo, záporné vľavo. Podporované len hardvérovými rendermi."
#define BEETLE_OPT_INITIAL_SCANLINE_LABEL_SK "Prvý riadok - NTSC"
#define BEETLE_OPT_INITIAL_SCANLINE_INFO_0_SK "Prvý zobrazený riadok pri spustení NTSC obsahu. Hodnoty väčšie ako nula znížia výšku výstupného obrazu orezaním pixelov od horného okraja. Možno použiť na potlačenie letterboxingu."
#define BEETLE_OPT_LAST_SCANLINE_LABEL_SK "Posledný riadok - NTSC"
#define BEETLE_OPT_LAST_SCANLINE_INFO_0_SK "Posledný zobrazený riadok pri spustení NTSC obsahu. Hodnoty pod 239 znížia výšku výstupného obrazu orezaním pixelov od spodného okraja. Možno použiť na potlačenie letterboxingu."
#define OPTION_VAL_239_SK "239 (predvolené)"
#define BEETLE_OPT_INITIAL_SCANLINE_PAL_LABEL_SK "Prvý riadok - PAL"
#define BEETLE_OPT_INITIAL_SCANLINE_PAL_INFO_0_SK "Prvý zobrazený riadok pri spustení PAL obsahu. Hodnoty väčšie ako nula znížia výšku výstupného obrazu orezaním pixelov od horného okraja. Možno použiť na potlačenie letterboxingu."
#define BEETLE_OPT_LAST_SCANLINE_PAL_LABEL_SK "Posledný riadok - PAL"
#define BEETLE_OPT_LAST_SCANLINE_PAL_INFO_0_SK "Posledný zobrazený riadok pri spustení PAL obsahu. Hodnoty pod 287 znížia výšku výstupného obrazu orezaním pixelov od spodného okraja. Možno použiť na potlačenie letterboxingu."
#define OPTION_VAL_287_SK "287 (predvolené)"

struct retro_core_option_v2_category option_cats_sk[] = {
   {
      "system",
      CATEGORY_SYSTEM_LABEL_SK,
      CATEGORY_SYSTEM_INFO_0_SK
   },
   {
      "video",
      CATEGORY_VIDEO_LABEL_SK,
      CATEGORY_VIDEO_INFO_0_SK
   },
   {
      "pgxp",
      CATEGORY_PGXP_LABEL_SK,
      CATEGORY_PGXP_INFO_0_SK
   },
   {
      "osd",
      CATEGORY_OSD_LABEL_SK,
      CATEGORY_OSD_INFO_0_SK
   },
   {
      "input",
      CATEGORY_INPUT_LABEL_SK,
      CATEGORY_INPUT_INFO_0_SK
   },
   {
      "memcards",
      CATEGORY_MEMCARDS_LABEL_SK,
      CATEGORY_MEMCARDS_INFO_0_SK
   },
   {
      "hacks",
      CATEGORY_HACKS_LABEL_SK,
      CATEGORY_HACKS_INFO_0_SK
   },
   { NULL, NULL, NULL },
};
struct retro_core_option_v2_definition option_defs_sk[] = {
   {
      BEETLE_OPT(cpu_freq_scale),
      BEETLE_OPT_CPU_FREQ_SCALE_LABEL_SK,
      NULL,
      BEETLE_OPT_CPU_FREQ_SCALE_INFO_0_SK,
      NULL,
      "system",
      {
         { "50%",           OPTION_VAL_50_SK },
         { "60%",           OPTION_VAL_60_SK },
         { "70%",           OPTION_VAL_70_SK },
         { "80%",           OPTION_VAL_80_SK },
         { "90%",           OPTION_VAL_90_SK },
         { "100%",          OPTION_VAL_100_SK },
         { "110%",          OPTION_VAL_110_SK },
         { "120%",          OPTION_VAL_120_SK },
         { "130%",          OPTION_VAL_130_SK },
         { "140%",          OPTION_VAL_140_SK },
         { "150%",          OPTION_VAL_150_SK },
         { "160%",          OPTION_VAL_160_SK },
         { "170%",          OPTION_VAL_170_SK },
         { "180%",          OPTION_VAL_180_SK },
         { "190%",          OPTION_VAL_190_SK },
         { "200%",          OPTION_VAL_200_SK },
         { "210%",          OPTION_VAL_210_SK },
         { "220%",          OPTION_VAL_220_SK },
         { "230%",          OPTION_VAL_230_SK },
         { "240%",          OPTION_VAL_240_SK },
         { "250%",          OPTION_VAL_250_SK },
         { "260%",          OPTION_VAL_260_SK },
         { "270%",          OPTION_VAL_270_SK },
         { "280%",          OPTION_VAL_280_SK },
         { "290%",          OPTION_VAL_290_SK },
         { "300%",          OPTION_VAL_300_SK },
         { "310%",          OPTION_VAL_310_SK },
         { "320%",          OPTION_VAL_320_SK },
         { "330%",          OPTION_VAL_330_SK },
         { "340%",          OPTION_VAL_340_SK },
         { "350%",          OPTION_VAL_350_SK },
         { "360%",          OPTION_VAL_360_SK },
         { "370%",          OPTION_VAL_370_SK },
         { "380%",          OPTION_VAL_380_SK },
         { "390%",          OPTION_VAL_390_SK },
         { "400%",          OPTION_VAL_400_SK },
         { "410%",          OPTION_VAL_410_SK },
         { "420%",          OPTION_VAL_420_SK },
         { "430%",          OPTION_VAL_430_SK },
         { "440%",          OPTION_VAL_440_SK },
         { "450%",          OPTION_VAL_450_SK },
         { "460%",          OPTION_VAL_460_SK },
         { "470%",          OPTION_VAL_470_SK },
         { "480%",          OPTION_VAL_480_SK },
         { "490%",          OPTION_VAL_490_SK },
         { "500%",          OPTION_VAL_500_SK },
         { "510%",          OPTION_VAL_510_SK },
         { "520%",          OPTION_VAL_520_SK },
         { "530%",          OPTION_VAL_530_SK },
         { "540%",          OPTION_VAL_540_SK },
         { "550%",          OPTION_VAL_550_SK },
         { "560%",          OPTION_VAL_560_SK },
         { "570%",          OPTION_VAL_570_SK },
         { "580%",          OPTION_VAL_580_SK },
         { "590%",          OPTION_VAL_590_SK },
         { "600%",          OPTION_VAL_600_SK },
         { "610%",          OPTION_VAL_610_SK },
         { "620%",          OPTION_VAL_620_SK },
         { "630%",          OPTION_VAL_630_SK },
         { "640%",          OPTION_VAL_640_SK },
         { "650%",          OPTION_VAL_650_SK },
         { "660%",          OPTION_VAL_660_SK },
         { "670%",          OPTION_VAL_670_SK },
         { "680%",          OPTION_VAL_680_SK },
         { "690%",          OPTION_VAL_690_SK },
         { "700%",          OPTION_VAL_700_SK },
         { "710%",          OPTION_VAL_710_SK },
         { "720%",          OPTION_VAL_720_SK },
         { "730%",          OPTION_VAL_730_SK },
         { "740%",          OPTION_VAL_740_SK },
         { "750%",          OPTION_VAL_750_SK },
         { NULL, NULL },
      },
      "100%"
   },
   {
      BEETLE_OPT(gpu_overclock),
      BEETLE_OPT_GPU_OVERCLOCK_LABEL_SK,
      NULL,
      BEETLE_OPT_GPU_OVERCLOCK_INFO_0_SK,
      NULL,
      "system",
      {
         { "1x(native)", OPTION_VAL_1X_NATIVE_SK },
         { "2x",         OPTION_VAL_2X_SK },
         { "4x",         OPTION_VAL_4X_SK },
         { "8x",         OPTION_VAL_8X_SK },
         { "16x",        OPTION_VAL_16X_SK },
         { "32x",        OPTION_VAL_32X_SK },
         { NULL, NULL },
      },
      "1x(native)"
   },
   {
      BEETLE_OPT(gte_overclock),
      BEETLE_OPT_GTE_OVERCLOCK_LABEL_SK,
      NULL,
      BEETLE_OPT_GTE_OVERCLOCK_INFO_0_SK,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
#ifndef EMSCRIPTEN
   {
      BEETLE_OPT(cd_access_method),
      BEETLE_OPT_CD_ACCESS_METHOD_LABEL_SK,
      NULL,
      BEETLE_OPT_CD_ACCESS_METHOD_INFO_0_SK,
      NULL,
      "system",
      {
         { "sync",     OPTION_VAL_SYNC_SK },
         { "async",    OPTION_VAL_ASYNC_SK },
         { "precache", OPTION_VAL_PRECACHE_SK },
         { NULL, NULL },
      },
      "sync"
   },
#endif
   {
      BEETLE_OPT(cd_fastload),
      BEETLE_OPT_CD_FASTLOAD_LABEL_SK,
      NULL,
      BEETLE_OPT_CD_FASTLOAD_INFO_0_SK,
      NULL,
      "system",
      {
         { "2x(native)", OPTION_VAL_2X_NATIVE_SK },
         { "4x",          OPTION_VAL_4X_SK },
         { "6x",          OPTION_VAL_6X_SK },
         { "8x",          OPTION_VAL_8X_SK },
         { "10x",         OPTION_VAL_10X_SK },
         { "12x",         OPTION_VAL_12X_SK },
         { "14x",         OPTION_VAL_14X_SK },
         { NULL, NULL },
      },
      "2x(native)"
   },
   {
      BEETLE_OPT(skip_bios),
      BEETLE_OPT_SKIP_BIOS_LABEL_SK,
      NULL,
      BEETLE_OPT_SKIP_BIOS_INFO_0_SK,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(override_bios),
      BEETLE_OPT_OVERRIDE_BIOS_LABEL_SK,
      NULL,
      BEETLE_OPT_OVERRIDE_BIOS_INFO_0_SK,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "psxonpsp", OPTION_VAL_PSXONPSP_SK },
         { "ps1_rom",  OPTION_VAL_PS1_ROM_SK },
         { "openbios", OPTION_VAL_OPENBIOS_SK     },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(region),
      BEETLE_OPT_REGION_LABEL_SK,
      NULL,
      BEETLE_OPT_REGION_INFO_0_SK,
      NULL,
      "system",
      {
         { "auto",   OPTION_VAL_AUTO_SK },
         { "ntsc-j", OPTION_VAL_NTSC_J_SK },
         { "ntsc-u", OPTION_VAL_NTSC_U_SK },
         { "pal",    OPTION_VAL_PAL_SK },
         { NULL, NULL },
      },
      "auto"
   },
   {
      BEETLE_OPT(internal_resolution),
      BEETLE_OPT_INTERNAL_RESOLUTION_LABEL_SK,
      NULL,
      BEETLE_OPT_INTERNAL_RESOLUTION_INFO_0_SK,
      NULL,
      "video",
      {
         { "1x(native)", OPTION_VAL_1X_NATIVE_SK },
         { "2x",         OPTION_VAL_2X_SK },
         { "4x",         OPTION_VAL_4X_SK },
         { "8x",         OPTION_VAL_8X_SK },
         { "16x",        OPTION_VAL_16X_SK },
         { NULL, NULL },
      },
      "1x(native)"
   },
#if defined(HAVE_OPENGL) || defined(HAVE_OPENGLES) || defined(HAVE_VULKAN)
   {
      BEETLE_OPT(renderer),
      BEETLE_OPT_RENDERER_LABEL_SK,
      NULL,
      BEETLE_OPT_RENDERER_INFO_0_SK,
      NULL,
      "video",
      {
         { "hardware",    OPTION_VAL_HARDWARE_SK },
#if defined(HAVE_OPENGL) || defined(HAVE_OPENGLES)
         { "hardware_gl", OPTION_VAL_HARDWARE_GL_SK },
#endif
#if defined(HAVE_VULKAN)
         { "hardware_vk", OPTION_VAL_HARDWARE_VK_SK },
#endif
         { "software",    OPTION_VAL_SOFTWARE_SK },
         { NULL, NULL },
      },
      "hardware"
   },
   {
      BEETLE_OPT(renderer_software_fb),
      BEETLE_OPT_RENDERER_SOFTWARE_FB_LABEL_SK,
      NULL,
      BEETLE_OPT_RENDERER_SOFTWARE_FB_INFO_0_SK,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled"
   },
#endif
#if defined(HAVE_OPENGL) || defined(HAVE_OPENGLES)
   {
      BEETLE_OPT(depth),
      BEETLE_OPT_DEPTH_LABEL_SK,
      NULL,
      BEETLE_OPT_DEPTH_INFO_0_SK,
      NULL,
      "video",
      {
         { "16bpp(native)", OPTION_VAL_16BPP_NATIVE_SK },
         { "32bpp",         OPTION_VAL_32BPP_SK },
         { NULL, NULL },
      },
      "16bpp(native)"
   },
   // Sort of, it's more like 15-bit high color and 24-bit true color for visible output. The alpha channel is used for mask bit. Vulkan renderer uses ABGR1555_555 for 31 bits internal? FMVs are always 24-bit on all renderers like original hardware (BGR888, no alpha)
#endif
   {
      BEETLE_OPT(dither_mode),
      BEETLE_OPT_DITHER_MODE_LABEL_SK,
      NULL,
      BEETLE_OPT_DITHER_MODE_INFO_0_SK,
      NULL,
      "video",
      {
         { "1x(native)",          OPTION_VAL_1X_NATIVE_SK },
         { "internal resolution", OPTION_VAL_INTERNAL_RESOLUTION_SK },
         { "disabled",            NULL },
         { NULL, NULL },
      },
      "1x(native)"
   },
#ifdef HAVE_VULKAN
   {
      BEETLE_OPT(scaled_uv_offset),
      BEETLE_OPT_SCALED_UV_OFFSET_LABEL_SK,
      NULL,
      BEETLE_OPT_SCALED_UV_OFFSET_INFO_0_SK,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled"
   },
#endif
#if defined(HAVE_OPENGL) || defined(HAVE_OPENGLES) || defined(HAVE_VULKAN)
   {
      BEETLE_OPT(filter),
      BEETLE_OPT_FILTER_LABEL_SK,
      NULL,
      BEETLE_OPT_FILTER_INFO_0_SK,
      NULL,
      "video",
      {
         { "nearest",  OPTION_VAL_NEAREST_SK },
         { "SABR",     OPTION_VAL_SABR_SK },
         { "xBR",      OPTION_VAL_XBR_SK },
         { "bilinear", OPTION_VAL_BILINEAR_SK },
         { "3-point",  OPTION_VAL_3_POINT_SK },
         { "JINC2",    OPTION_VAL_JINC2_SK },
         { NULL, NULL },
      },
      "nearest"
   },
#ifdef HAVE_VULKAN
   {
      BEETLE_OPT(filter_exclude_sprite),
      BEETLE_OPT_FILTER_EXCLUDE_SPRITE_LABEL_SK,
      NULL,
      BEETLE_OPT_FILTER_EXCLUDE_SPRITE_INFO_0_SK,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "opaque", OPTION_VAL_OPAQUE_SK },
         { "all", OPTION_VAL_ALL_SK },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(filter_exclude_2d_polygon),
      BEETLE_OPT_FILTER_EXCLUDE_2D_POLYGON_LABEL_SK,
      NULL,
      BEETLE_OPT_FILTER_EXCLUDE_2D_POLYGON_INFO_0_SK,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "opaque", OPTION_VAL_OPAQUE_SK },
         { "all", OPTION_VAL_ALL_SK },
         { NULL, NULL },
      },
      "disabled"
   },
#endif
#endif
#ifdef HAVE_VULKAN
   {
      BEETLE_OPT(adaptive_smoothing),
      BEETLE_OPT_ADAPTIVE_SMOOTHING_LABEL_SK,
      NULL,
      BEETLE_OPT_ADAPTIVE_SMOOTHING_INFO_0_SK,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(super_sampling),
      BEETLE_OPT_SUPER_SAMPLING_LABEL_SK,
      NULL,
      BEETLE_OPT_SUPER_SAMPLING_INFO_0_SK,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(msaa),
      BEETLE_OPT_MSAA_LABEL_SK,
      NULL,
      BEETLE_OPT_MSAA_INFO_0_SK,
      NULL,
      "video",
      {
         { "1x",  OPTION_VAL_1X_SK },
         { "2x",  OPTION_VAL_2X_SK },
         { "4x",  OPTION_VAL_4X_SK },
         { "8x",  OPTION_VAL_8X_SK },
         { "16x", OPTION_VAL_16X_SK },
         { NULL, NULL },
      },
      "1x"
   },
   {
      BEETLE_OPT(mdec_yuv),
      BEETLE_OPT_MDEC_YUV_LABEL_SK,
      NULL,
      BEETLE_OPT_MDEC_YUV_INFO_0_SK,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(track_textures),
      BEETLE_OPT_TRACK_TEXTURES_LABEL_SK,
      NULL,
      BEETLE_OPT_TRACK_TEXTURES_INFO_0_SK,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
#ifdef TEXTURE_DUMPING_ENABLED
   {
      BEETLE_OPT(dump_textures),
      BEETLE_OPT_DUMP_TEXTURES_LABEL_SK,
      NULL,
      BEETLE_OPT_DUMP_TEXTURES_INFO_0_SK,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
#endif
   {
      BEETLE_OPT(replace_textures),
      BEETLE_OPT_REPLACE_TEXTURES_LABEL_SK,
      NULL,
      BEETLE_OPT_REPLACE_TEXTURES_INFO_0_SK,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
#endif
   {
      BEETLE_OPT(pgxp_mode),
      BEETLE_OPT_PGXP_MODE_LABEL_SK,
      NULL,
      BEETLE_OPT_PGXP_MODE_INFO_0_SK,
      NULL,
      "pgxp",
      {
         { "disabled",     NULL },
         { "memory only",  OPTION_VAL_MEMORY_ONLY_SK },
         { "memory + CPU", OPTION_VAL_MEMORY_CPU_SK },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(pgxp_2d_tol),
      BEETLE_OPT_PGXP_2D_TOL_LABEL_SK,
      NULL,
      BEETLE_OPT_PGXP_2D_TOL_INFO_0_SK,
      NULL,
      "pgxp",
      {
         { "disabled", NULL },
         { "0px", OPTION_VAL_0PX_SK },
         { "1px", OPTION_VAL_1PX_SK },
         { "2px", OPTION_VAL_2PX_SK },
         { "3px", OPTION_VAL_3PX_SK },
         { "4px", OPTION_VAL_4PX_SK },
         { "5px", OPTION_VAL_5PX_SK },
         { "6px", OPTION_VAL_6PX_SK },
         { "7px", OPTION_VAL_7PX_SK },
         { "8px", OPTION_VAL_8PX_SK },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(pgxp_nclip),
      BEETLE_OPT_PGXP_NCLIP_LABEL_SK,
      NULL,
      BEETLE_OPT_PGXP_NCLIP_INFO_0_SK,
      NULL,
      "pgxp",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
#if defined(HAVE_OPENGL) || defined(HAVE_OPENGLES) || defined(HAVE_VULKAN)
   {
      BEETLE_OPT(pgxp_vertex),
      BEETLE_OPT_PGXP_VERTEX_LABEL_SK,
      NULL,
      BEETLE_OPT_PGXP_VERTEX_INFO_0_SK,
      NULL,
      "pgxp",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
#endif
   {
      BEETLE_OPT(pgxp_texture),
      BEETLE_OPT_PGXP_TEXTURE_LABEL_SK,
      NULL,
      BEETLE_OPT_PGXP_TEXTURE_INFO_0_SK,
      NULL,
      "pgxp",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(display_internal_fps),
      BEETLE_OPT_DISPLAY_INTERNAL_FPS_LABEL_SK,
      NULL,
      BEETLE_OPT_DISPLAY_INTERNAL_FPS_INFO_0_SK,
      NULL,
      "osd",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(display_osd),
      BEETLE_OPT_DISPLAY_OSD_LABEL_SK,
      NULL,
      BEETLE_OPT_DISPLAY_OSD_INFO_0_SK,
      NULL,
      "osd",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled"
   },
#if defined(HAVE_OPENGL) || defined(HAVE_OPENGLES) || defined(HAVE_VULKAN)
   {
      BEETLE_OPT(display_vram),
      BEETLE_OPT_DISPLAY_VRAM_LABEL_SK,
      NULL,
      BEETLE_OPT_DISPLAY_VRAM_INFO_0_SK,
      NULL,
      "osd",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
#endif
   {
      BEETLE_OPT(analog_calibration),
      BEETLE_OPT_ANALOG_CALIBRATION_LABEL_SK,
      NULL,
      BEETLE_OPT_ANALOG_CALIBRATION_INFO_0_SK,
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(analog_toggle),
      BEETLE_OPT_ANALOG_TOGGLE_LABEL_SK,
      NULL,
      BEETLE_OPT_ANALOG_TOGGLE_INFO_0_SK,
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { "enabled-analog", OPTION_VAL_ENABLED_ANALOG_SK },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(analog_toggle_combo),
      BEETLE_OPT_ANALOG_TOGGLE_COMBO_LABEL_SK,
      NULL,
      BEETLE_OPT_ANALOG_TOGGLE_COMBO_INFO_0_SK,
      NULL,
      "input",
      {
         { "l1+l2+r1+r2+start+select", OPTION_VAL_L1_L2_R1_R2_START_SELECT_SK },
         { "l1+r1+select",             OPTION_VAL_L1_R1_SELECT_SK },
         { "l1+r1+start",              OPTION_VAL_L1_R1_START_SK },
         { "l1+r1+l3",                 OPTION_VAL_L1_R1_L3_SK },
         { "l1+r1+r3",                 OPTION_VAL_L1_R1_R3_SK },
         { "l2+r2+select",             OPTION_VAL_L2_R2_SELECT_SK },
         { "l2+r2+start",              OPTION_VAL_L2_R2_START_SK },
         { "l2+r2+l3",                 OPTION_VAL_L2_R2_L3_SK },
         { "l2+r2+r3",                 OPTION_VAL_L2_R2_R3_SK },
         { "l3+r3",                    OPTION_VAL_L3_R3_SK },
         { NULL, NULL },
      },
      "l1+r1+select"
   },
   {
      BEETLE_OPT(analog_toggle_hold),
      BEETLE_OPT_ANALOG_TOGGLE_HOLD_LABEL_SK,
      NULL,
      BEETLE_OPT_ANALOG_TOGGLE_HOLD_INFO_0_SK,
      NULL,
      "input",
      {
         { "0", OPTION_VAL_0_SK },
         { "1", OPTION_VAL_1_SK },
         { "2", OPTION_VAL_2_SK },
         { "3", OPTION_VAL_3_SK },
         { "4", OPTION_VAL_4_SK },
         { "5", OPTION_VAL_5_SK },
         { NULL, NULL },
      },
      "1"
   },
   {
      BEETLE_OPT(enable_multitap_port1),
      BEETLE_OPT_ENABLE_MULTITAP_PORT1_LABEL_SK,
      NULL,
      BEETLE_OPT_ENABLE_MULTITAP_PORT1_INFO_0_SK,
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(enable_multitap_port2),
      BEETLE_OPT_ENABLE_MULTITAP_PORT2_LABEL_SK,
      NULL,
      BEETLE_OPT_ENABLE_MULTITAP_PORT2_INFO_0_SK,
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(gun_input_mode),
      BEETLE_OPT_GUN_INPUT_MODE_LABEL_SK,
      NULL,
      BEETLE_OPT_GUN_INPUT_MODE_INFO_0_SK,
      NULL,
      "input",
      {
         { "lightgun",    OPTION_VAL_LIGHTGUN_SK },
         { "touchscreen", OPTION_VAL_TOUCHSCREEN_SK },
         { NULL, NULL },
      },
      "lightgun"
   },
   // Shouldn't the gun_input_mode just be Mouse vs. Touchscreen?
   {
      BEETLE_OPT(gun_cursor),
      BEETLE_OPT_GUN_CURSOR_LABEL_SK,
      NULL,
      BEETLE_OPT_GUN_CURSOR_INFO_0_SK,
      NULL,
      "input",
      {
         { "off",   "disabled" },
         { "cross", OPTION_VAL_CROSS_SK },
         { "dot",   OPTION_VAL_DOT_SK },
         { NULL, NULL },
      },
      "cross"
   },
   {
      BEETLE_OPT(crosshair_color_p1),
      BEETLE_OPT_CROSSHAIR_COLOR_P1_LABEL_SK,
      NULL,
      BEETLE_OPT_CROSSHAIR_COLOR_P1_INFO_0_SK,
      NULL,
      "input",
      {
         { "red",    OPTION_VAL_RED_SK },
         { "blue",   OPTION_VAL_BLUE_SK },
         { "green",  OPTION_VAL_GREEN_SK },
         { "orange", OPTION_VAL_ORANGE_SK },
         { "yellow", OPTION_VAL_YELLOW_SK },
         { "cyan",   OPTION_VAL_CYAN_SK },
         { "pink",   OPTION_VAL_PINK_SK },
         { "purple", OPTION_VAL_PURPLE_SK },
         { "black",  OPTION_VAL_BLACK_SK },
         { "white",  OPTION_VAL_WHITE_SK },
         { NULL, NULL },
      },
      "red"
   },
   {
      BEETLE_OPT(crosshair_color_p2),
      BEETLE_OPT_CROSSHAIR_COLOR_P2_LABEL_SK,
      NULL,
      BEETLE_OPT_CROSSHAIR_COLOR_P2_INFO_0_SK,
      NULL,
      "input",
      {
         { "blue",   OPTION_VAL_BLUE_SK },
         { "red",    OPTION_VAL_RED_SK },
         { "green",  OPTION_VAL_GREEN_SK },
         { "orange", OPTION_VAL_ORANGE_SK },
         { "yellow", OPTION_VAL_YELLOW_SK },
         { "cyan",   OPTION_VAL_CYAN_SK },
         { "pink",   OPTION_VAL_PINK_SK },
         { "purple", OPTION_VAL_PURPLE_SK },
         { "black",  OPTION_VAL_BLACK_SK },
         { "white",  OPTION_VAL_WHITE_SK },
         { NULL, NULL },
      },
      "blue"
   },
   {
      BEETLE_OPT(mouse_sensitivity),
      BEETLE_OPT_MOUSE_SENSITIVITY_LABEL_SK,
      NULL,
      BEETLE_OPT_MOUSE_SENSITIVITY_INFO_0_SK,
      NULL,
      "input",
      {
         { "5%",   OPTION_VAL_5_O42_SK },
         { "10%",  OPTION_VAL_10_SK },
         { "15%",  OPTION_VAL_15_SK },
         { "20%",  OPTION_VAL_20_SK },
         { "25%",  OPTION_VAL_25_SK },
         { "30%",  OPTION_VAL_30_SK },
         { "35%",  OPTION_VAL_35_SK },
         { "40%",  OPTION_VAL_40_SK },
         { "45%",  OPTION_VAL_45_SK },
         { "50%",  OPTION_VAL_50_SK },
         { "55%",  OPTION_VAL_55_SK },
         { "60%",  OPTION_VAL_60_SK },
         { "65%",  OPTION_VAL_65_SK },
         { "70%",  OPTION_VAL_70_SK },
         { "75%",  OPTION_VAL_75_SK },
         { "80%",  OPTION_VAL_80_SK },
         { "85%",  OPTION_VAL_85_SK },
         { "90%",  OPTION_VAL_90_SK },
         { "95%",  OPTION_VAL_95_SK },
         { "100%", OPTION_VAL_100_O42_SK },
         { "105%", OPTION_VAL_105_SK },
         { "110%", OPTION_VAL_110_SK },
         { "115%", OPTION_VAL_115_SK },
         { "120%", OPTION_VAL_120_SK },
         { "125%", OPTION_VAL_125_SK },
         { "130%", OPTION_VAL_130_SK },
         { "135%", OPTION_VAL_135_SK },
         { "140%", OPTION_VAL_140_SK },
         { "145%", OPTION_VAL_145_SK },
         { "150%", OPTION_VAL_150_SK },
         { "155%", OPTION_VAL_155_SK },
         { "160%", OPTION_VAL_160_SK },
         { "165%", OPTION_VAL_165_SK },
         { "170%", OPTION_VAL_170_SK },
         { "175%", OPTION_VAL_175_SK },
         { "180%", OPTION_VAL_180_SK },
         { "185%", OPTION_VAL_185_SK },
         { "190%", OPTION_VAL_190_SK },
         { "195%", OPTION_VAL_195_SK },
         { "200%", OPTION_VAL_200_SK },
         { NULL, NULL },
      },
      "100%"
   },
   {
      BEETLE_OPT(negcon_response),
      BEETLE_OPT_NEGCON_RESPONSE_LABEL_SK,
      NULL,
      BEETLE_OPT_NEGCON_RESPONSE_INFO_0_SK,
      NULL,
      "input",
      {
         { "linear",    OPTION_VAL_LINEAR_SK },
         { "quadratic", OPTION_VAL_QUADRATIC_SK },
         { "cubic",     OPTION_VAL_CUBIC_SK },
         { NULL, NULL },
      },
      "linear"
   },
   {
      BEETLE_OPT(negcon_deadzone),
      BEETLE_OPT_NEGCON_DEADZONE_LABEL_SK,
      NULL,
      BEETLE_OPT_NEGCON_DEADZONE_INFO_0_SK,
      NULL,
      "input",
      {
         { "0%",  OPTION_VAL_0_O44_SK },
         { "5%",  OPTION_VAL_5_O42_SK },
         { "10%", OPTION_VAL_10_SK },
         { "15%", OPTION_VAL_15_SK },
         { "20%", OPTION_VAL_20_SK },
         { "25%", OPTION_VAL_25_SK },
         { "30%", OPTION_VAL_30_SK },
         { NULL, NULL },
      },
      "0%"
   },
   {
      BEETLE_OPT(use_mednafen_memcard0_method),
      BEETLE_OPT_USE_MEDNAFEN_MEMCARD0_METHOD_LABEL_SK,
      NULL,
      BEETLE_OPT_USE_MEDNAFEN_MEMCARD0_METHOD_INFO_0_SK,
      NULL,
      "memcards",
      {
         { "libretro", OPTION_VAL_LIBRETRO_SK },
         { "mednafen", OPTION_VAL_MEDNAFEN_SK },
         { NULL, NULL },
      },
      "libretro"
   },
   {
      BEETLE_OPT(enable_memcard1),
      BEETLE_OPT_ENABLE_MEMCARD1_LABEL_SK,
      NULL,
      BEETLE_OPT_ENABLE_MEMCARD1_INFO_0_SK,
      NULL,
      "memcards",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(shared_memory_cards),
      BEETLE_OPT_SHARED_MEMORY_CARDS_LABEL_SK,
      NULL,
      BEETLE_OPT_SHARED_MEMORY_CARDS_INFO_0_SK,
      NULL,
      "memcards",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(memcard_left_index),
      BEETLE_OPT_MEMCARD_LEFT_INDEX_LABEL_SK,
      NULL,
      BEETLE_OPT_MEMCARD_LEFT_INDEX_INFO_0_SK,
      NULL,
      "memcards",
      {
         { "0",  OPTION_VAL_0_O48_SK },
         { "1",  NULL },
         { "2",  NULL },
         { "3",  NULL },
         { "4",  NULL },
         { "5",  NULL },
         { "6",  NULL },
         { "7",  NULL },
         { "8",  NULL },
         { "9",  NULL },
         { "10", NULL },
         { "11", NULL },
         { "12", NULL },
         { "13", NULL },
         { "14", NULL },
         { "15", NULL },
         { "16", NULL },
         { "17", NULL },
         { "18", NULL },
         { "19", NULL },
         { "20", NULL },
         { "21", NULL },
         { "22", NULL },
         { "23", NULL },
         { "24", NULL },
         { "25", NULL },
         { "26", NULL },
         { "27", NULL },
         { "28", NULL },
         { "29", NULL },
         { "30", NULL },
         { "31", NULL },
         { "32", NULL },
         { "33", NULL },
         { "34", NULL },
         { "35", NULL },
         { "36", NULL },
         { "37", NULL },
         { "38", NULL },
         { "39", NULL },
         { "40", NULL },
         { "41", NULL },
         { "42", NULL },
         { "43", NULL },
         { "44", NULL },
         { "45", NULL },
         { "46", NULL },
         { "47", NULL },
         { "48", NULL },
         { "49", NULL },
         { "50", NULL },
         { "51", NULL },
         { "52", NULL },
         { "53", NULL },
         { "54", NULL },
         { "55", NULL },
         { "56", NULL },
         { "57", NULL },
         { "58", NULL },
         { "59", NULL },
         { "60", NULL },
         { "61", NULL },
         { "62", NULL },
         { "63", NULL },
         { NULL, NULL },
      },
      "0"
   },
   {
      BEETLE_OPT(memcard_right_index),
      BEETLE_OPT_MEMCARD_RIGHT_INDEX_LABEL_SK,
      NULL,
      BEETLE_OPT_MEMCARD_RIGHT_INDEX_INFO_0_SK,
      NULL,
      "memcards",
      {
         { "0",  NULL },
         { "1",  OPTION_VAL_1_O49_SK },
         { "2",  NULL },
         { "3",  NULL },
         { "4",  NULL },
         { "5",  NULL },
         { "6",  NULL },
         { "7",  NULL },
         { "8",  NULL },
         { "9",  NULL },
         { "10", NULL },
         { "11", NULL },
         { "12", NULL },
         { "13", NULL },
         { "14", NULL },
         { "15", NULL },
         { "16", NULL },
         { "17", NULL },
         { "18", NULL },
         { "19", NULL },
         { "20", NULL },
         { "21", NULL },
         { "22", NULL },
         { "23", NULL },
         { "24", NULL },
         { "25", NULL },
         { "26", NULL },
         { "27", NULL },
         { "28", NULL },
         { "29", NULL },
         { "30", NULL },
         { "31", NULL },
         { "32", NULL },
         { "33", NULL },
         { "34", NULL },
         { "35", NULL },
         { "36", NULL },
         { "37", NULL },
         { "38", NULL },
         { "39", NULL },
         { "40", NULL },
         { "41", NULL },
         { "42", NULL },
         { "43", NULL },
         { "44", NULL },
         { "45", NULL },
         { "46", NULL },
         { "47", NULL },
         { "48", NULL },
         { "49", NULL },
         { "50", NULL },
         { "51", NULL },
         { "52", NULL },
         { "53", NULL },
         { "54", NULL },
         { "55", NULL },
         { "56", NULL },
         { "57", NULL },
         { "58", NULL },
         { "59", NULL },
         { "60", NULL },
         { "61", NULL },
         { "62", NULL },
         { "63", NULL },
         { NULL, NULL },
      },
      "1"
   },
   {
      BEETLE_OPT(line_render),
      BEETLE_OPT_LINE_RENDER_LABEL_SK,
      NULL,
      BEETLE_OPT_LINE_RENDER_INFO_0_SK,
      NULL,
      "hacks",
      {
         { "default",    OPTION_VAL_DEFAULT_SK },
         { "aggressive", OPTION_VAL_AGGRESSIVE_SK },
         { "disabled",   NULL },
         { NULL, NULL },
      },
      "default"
   },
   {
      BEETLE_OPT(widescreen_hack),
      BEETLE_OPT_WIDESCREEN_HACK_LABEL_SK,
      NULL,
      BEETLE_OPT_WIDESCREEN_HACK_INFO_0_SK,
      NULL,
      "hacks",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(widescreen_hack_aspect_ratio),
      BEETLE_OPT_WIDESCREEN_HACK_ASPECT_RATIO_LABEL_SK,
      NULL,
      BEETLE_OPT_WIDESCREEN_HACK_ASPECT_RATIO_INFO_0_SK,
      NULL,
      "hacks",
      {
         { "16:9",  OPTION_VAL_16_9_SK },
         { "16:10", OPTION_VAL_16_10_SK },
         { "18:9",  OPTION_VAL_18_9_SK },
         { "19:9",  OPTION_VAL_19_9_SK },
         { "20:9",  OPTION_VAL_20_9_SK },
         { "21:9",  OPTION_VAL_21_9_SK }, // 64:27
         { "32:9",  OPTION_VAL_32_9_SK },
         { NULL,    NULL },
      },
      "16:9"
   },
#if defined(HAVE_LIGHTREC)
   {
      BEETLE_OPT(cpu_dynarec),
      BEETLE_OPT_CPU_DYNAREC_LABEL_SK,
      NULL,
      BEETLE_OPT_CPU_DYNAREC_INFO_0_SK,
      NULL,
      "hacks",
      {
         { "disabled",        OPTION_VAL_DISABLED_SK },
         { "execute",         OPTION_VAL_EXECUTE_SK },
         { "execute_one",     OPTION_VAL_EXECUTE_ONE_SK },
         { "run_interpreter", OPTION_VAL_RUN_INTERPRETER_SK },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(dynarec_invalidate),
      BEETLE_OPT_DYNAREC_INVALIDATE_LABEL_SK,
      NULL,
      BEETLE_OPT_DYNAREC_INVALIDATE_INFO_0_SK,
      NULL,
      "hacks",
      {
         { "full", OPTION_VAL_FULL_SK },
         { "dma",  OPTION_VAL_DMA_SK },
         { NULL, NULL },
      },
      "full"
   },
   {
      BEETLE_OPT(dynarec_eventcycles),
      BEETLE_OPT_DYNAREC_EVENTCYCLES_LABEL_SK,
      NULL,
      BEETLE_OPT_DYNAREC_EVENTCYCLES_INFO_0_SK,
      NULL,
      "hacks",
      {
         { "128",  OPTION_VAL_128_SK },
         { "256",  NULL },
         { "384",  NULL },
         { "512",  NULL },
         { "640",  NULL },
         { "768",  NULL },
         { "896",  NULL },
         { "1024", NULL },
         { "1152", NULL },
         { "1280", NULL },
         { "1408", NULL },
         { "1536", NULL },
         { "1664", NULL },
         { "1792", NULL },
         { "1920", NULL },
         { "2048", NULL },
         { NULL, NULL },
      },
      "128"
   },
   {
      BEETLE_OPT(dynarec_spu_samples),
      BEETLE_OPT_DYNAREC_SPU_SAMPLES_LABEL_SK,
      NULL,
      BEETLE_OPT_DYNAREC_SPU_SAMPLES_INFO_0_SK,
      NULL,
      "hacks",
      {
         { "1",  OPTION_VAL_1_O49_SK },
         { "4",  NULL },
         { "16", NULL },
         { NULL, NULL },
      },
      "1"
   },
#endif
   {
      BEETLE_OPT(core_timing_fps),
      BEETLE_OPT_CORE_TIMING_FPS_LABEL_SK,
      NULL,
      BEETLE_OPT_CORE_TIMING_FPS_INFO_0_SK,
      NULL,
      "video",
      {
         { "auto_toggle",       OPTION_VAL_AUTO_TOGGLE_SK },
         { "force_progressive", OPTION_VAL_FORCE_PROGRESSIVE_SK },
         { "force_interlaced",  OPTION_VAL_FORCE_INTERLACED_SK },
      },
      "auto_toggle"
   },
   {
      BEETLE_OPT(deinterlacer),
      BEETLE_OPT_DEINTERLACER_LABEL_SK,
      NULL,
      BEETLE_OPT_DEINTERLACER_INFO_0_SK,
      NULL,
      "video",
      {
         { "weave",      OPTION_VAL_WEAVE_SK },
         { "bob",        OPTION_VAL_BOB_SK },
         { "bob_offset", OPTION_VAL_BOB_OFFSET_SK },
         { "fastmad",    OPTION_VAL_FASTMAD_SK },
         { "off",        "Off" },
         { NULL, NULL },
      },
      "weave"
   },
   {
      BEETLE_OPT(frame_duping),
      BEETLE_OPT_FRAME_DUPING_LABEL_SK,
      NULL,
      BEETLE_OPT_FRAME_DUPING_INFO_0_SK,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(pal_video_timing_override),
      BEETLE_OPT_PAL_VIDEO_TIMING_OVERRIDE_LABEL_SK,
      NULL,
      BEETLE_OPT_PAL_VIDEO_TIMING_OVERRIDE_INFO_0_SK,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(aspect_ratio),
      BEETLE_OPT_ASPECT_RATIO_LABEL_SK,
      NULL,
      BEETLE_OPT_ASPECT_RATIO_INFO_0_SK,
      NULL,
      "video",
      {
         { "corrected", OPTION_VAL_CORRECTED_SK },
         { "uncorrected", OPTION_VAL_UNCORRECTED_SK },
         { "4:3",  OPTION_VAL_4_3_SK },
         { "ntsc", OPTION_VAL_NTSC_SK },
      },
      "corrected"
   },
   {
      BEETLE_OPT(crop_overscan),
      BEETLE_OPT_CROP_OVERSCAN_LABEL_SK,
      NULL,
      BEETLE_OPT_CROP_OVERSCAN_INFO_0_SK,
      NULL,
      "video",
      {
         { "disabled", "Disabled" },
         { "static",   OPTION_VAL_STATIC_SK },
         { "smart",    OPTION_VAL_SMART_SK },
         { NULL, NULL },
      },
      "smart"
   },
   {
      BEETLE_OPT(image_crop),
      BEETLE_OPT_IMAGE_CROP_LABEL_SK,
      NULL,
      BEETLE_OPT_IMAGE_CROP_INFO_0_SK,
      NULL,
      "video",
      {
         { "disabled", "0" },
         { "1px",      OPTION_VAL_1PX_SK },
         { "2px",      OPTION_VAL_2PX_SK },
         { "3px",      OPTION_VAL_3PX_SK },
         { "4px",      OPTION_VAL_4PX_SK },
         { "5px",      OPTION_VAL_5PX_SK },
         { "6px",      OPTION_VAL_6PX_SK },
         { "7px",      OPTION_VAL_7PX_SK },
         { "8px",      OPTION_VAL_8PX_SK },
         { "9px",      OPTION_VAL_9PX_SK },
         { "10px",     OPTION_VAL_10PX_SK },
         { "11px",     OPTION_VAL_11PX_SK },
         { "12px",     OPTION_VAL_12PX_SK },
         { "13px",     OPTION_VAL_13PX_SK },
         { "14px",     OPTION_VAL_14PX_SK },
         { "15px",     OPTION_VAL_15PX_SK },
         { "16px",     OPTION_VAL_16PX_SK },
         { "17px",     OPTION_VAL_17PX_SK },
         { "18px",     OPTION_VAL_18PX_SK },
         { "19px",     OPTION_VAL_19PX_SK },
         { "20px",     OPTION_VAL_20PX_SK },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(image_offset),
      BEETLE_OPT_IMAGE_OFFSET_LABEL_SK,
      NULL,
      BEETLE_OPT_IMAGE_OFFSET_INFO_0_SK,
      NULL,
      "video",
      {
         { "-12px",    OPTION_VAL_12PX_O64_SK },
         { "-11px",    OPTION_VAL_11PX_O64_SK },
         { "-10px",    OPTION_VAL_10PX_O64_SK },
         { "-9px",     OPTION_VAL_9PX_O64_SK },
         { "-8px",     OPTION_VAL_8PX_O64_SK },
         { "-7px",     OPTION_VAL_7PX_O64_SK },
         { "-6px",     OPTION_VAL_6PX_O64_SK },
         { "-5px",     OPTION_VAL_5PX_O64_SK },
         { "-4px",     OPTION_VAL_4PX_O64_SK },
         { "-3px",     OPTION_VAL_3PX_O64_SK },
         { "-2px",     OPTION_VAL_2PX_O64_SK },
         { "-1px",     OPTION_VAL_1PX_O64_SK },
         { "disabled", "0" },
         { "+1px",     OPTION_VAL_1PX_O64_0_SK },
         { "+2px",     OPTION_VAL_2PX_O64_0_SK },
         { "+3px",     OPTION_VAL_3PX_O64_0_SK },
         { "+4px",     OPTION_VAL_4PX_O64_0_SK },
         { "+5px",     OPTION_VAL_5PX_O64_0_SK },
         { "+6px",     OPTION_VAL_6PX_O64_0_SK },
         { "+7px",     OPTION_VAL_7PX_O64_0_SK },
         { "+8px",     OPTION_VAL_8PX_O64_0_SK },
         { "+9px",     OPTION_VAL_9PX_O64_0_SK },
         { "+10px",    OPTION_VAL_10PX_O64_0_SK },
         { "+11px",    OPTION_VAL_11PX_O64_0_SK },
         { "+12px",    OPTION_VAL_12PX_O64_0_SK },
         { NULL, NULL },
      },
      "disabled"
   },
#if defined(HAVE_OPENGL) || defined(HAVE_OPENGLES) || defined(HAVE_VULKAN)
   {
      BEETLE_OPT(image_offset_cycles),
      BEETLE_OPT_IMAGE_OFFSET_CYCLES_LABEL_SK,
      NULL,
      BEETLE_OPT_IMAGE_OFFSET_CYCLES_INFO_0_SK,
      NULL,
      "video",
      {
         { "-40",      NULL },
         { "-39",      NULL },
         { "-38",      NULL },
         { "-37",      NULL },
         { "-36",      NULL },
         { "-35",      NULL },
         { "-34",      NULL },
         { "-33",      NULL },
         { "-32",      NULL },
         { "-31",      NULL },
         { "-30",      NULL },
         { "-29",      NULL },
         { "-28",      NULL },
         { "-27",      NULL },
         { "-26",      NULL },
         { "-25",      NULL },
         { "-24",      NULL },
         { "-23",      NULL },
         { "-22",      NULL },
         { "-21",      NULL },
         { "-20",      NULL },
         { "-19",      NULL },
         { "-18",      NULL },
         { "-17",      NULL },
         { "-16",      NULL },
         { "-15",      NULL },
         { "-14",      NULL },
         { "-13",      NULL },
         { "-12",      NULL },
         { "-11",      NULL },
         { "-10",      NULL },
         { "-9",       NULL },
         { "-8",       NULL },
         { "-7",       NULL },
         { "-6",       NULL },
         { "-5",       NULL },
         { "-4",       NULL },
         { "-3",       NULL },
         { "-2",       NULL },
         { "-1",       NULL },
         { "0",        "0" },
         { "+1",       NULL },
         { "+2",       NULL },
         { "+3",       NULL },
         { "+4",       NULL },
         { "+5",       NULL },
         { "+6",       NULL },
         { "+7",       NULL },
         { "+8",       NULL },
         { "+9",       NULL },
         { "+10",      NULL },
         { "+11",      NULL },
         { "+12",      NULL },
         { "+13",      NULL },
         { "+14",      NULL },
         { "+15",      NULL },
         { "+16",      NULL },
         { "+17",      NULL },
         { "+18",      NULL },
         { "+19",      NULL },
         { "+20",      NULL },
         { "+21",      NULL },
         { "+22",      NULL },
         { "+23",      NULL },
         { "+24",      NULL },
         { "+25",      NULL },
         { "+26",      NULL },
         { "+27",      NULL },
         { "+28",      NULL },
         { "+29",      NULL },
         { "+30",      NULL },
         { "+31",      NULL },
         { "+32",      NULL },
         { "+33",      NULL },
         { "+34",      NULL },
         { "+35",      NULL },
         { "+36",      NULL },
         { "+37",      NULL },
         { "+38",      NULL },
         { "+39",      NULL },
         { "+40",      NULL },
         { NULL, NULL},
      },
      "0"
   },
#endif
   {
      BEETLE_OPT(initial_scanline),
      BEETLE_OPT_INITIAL_SCANLINE_LABEL_SK,
      NULL,
      BEETLE_OPT_INITIAL_SCANLINE_INFO_0_SK,
      NULL,
      "video",
      {
         { "0",  OPTION_VAL_0_O48_SK },
         { "1",  NULL },
         { "2",  NULL },
         { "3",  NULL },
         { "4",  NULL },
         { "5",  NULL },
         { "6",  NULL },
         { "7",  NULL },
         { "8",  NULL },
         { "9",  NULL },
         { "10", NULL },
         { "11", NULL },
         { "12", NULL },
         { "13", NULL },
         { "14", NULL },
         { "15", NULL },
         { "16", NULL },
         { "17", NULL },
         { "18", NULL },
         { "19", NULL },
         { "20", NULL },
         { "21", NULL },
         { "22", NULL },
         { "23", NULL },
         { "24", NULL },
         { "25", NULL },
         { "26", NULL },
         { "27", NULL },
         { "28", NULL },
         { "29", NULL },
         { "30", NULL },
         { "31", NULL },
         { "32", NULL },
         { "33", NULL },
         { "34", NULL },
         { "35", NULL },
         { "36", NULL },
         { "37", NULL },
         { "38", NULL },
         { "39", NULL },
         { "40", NULL },
         { NULL, NULL },
      },
      "0"
   },
   {
      BEETLE_OPT(last_scanline),
      BEETLE_OPT_LAST_SCANLINE_LABEL_SK,
      NULL,
      BEETLE_OPT_LAST_SCANLINE_INFO_0_SK,
      NULL,
      "video",
      {
         { "210", NULL },
         { "211", NULL },
         { "212", NULL },
         { "213", NULL },
         { "214", NULL },
         { "215", NULL },
         { "216", NULL },
         { "217", NULL },
         { "218", NULL },
         { "219", NULL },
         { "220", NULL },
         { "221", NULL },
         { "222", NULL },
         { "223", NULL },
         { "224", NULL },
         { "225", NULL },
         { "226", NULL },
         { "227", NULL },
         { "228", NULL },
         { "229", NULL },
         { "230", NULL },
         { "231", NULL },
         { "232", NULL },
         { "233", NULL },
         { "234", NULL },
         { "235", NULL },
         { "236", NULL },
         { "237", NULL },
         { "238", NULL },
         { "239", OPTION_VAL_239_SK },
         { NULL, NULL },
      },
      "239"
   },
   {
      BEETLE_OPT(initial_scanline_pal),
      BEETLE_OPT_INITIAL_SCANLINE_PAL_LABEL_SK,
      NULL,
      BEETLE_OPT_INITIAL_SCANLINE_PAL_INFO_0_SK,
      NULL,
      "video",
      {
         { "0",  OPTION_VAL_0_O48_SK },
         { "1",  NULL },
         { "2",  NULL },
         { "3",  NULL },
         { "4",  NULL },
         { "5",  NULL },
         { "6",  NULL },
         { "7",  NULL },
         { "8",  NULL },
         { "9",  NULL },
         { "10", NULL },
         { "11", NULL },
         { "12", NULL },
         { "13", NULL },
         { "14", NULL },
         { "15", NULL },
         { "16", NULL },
         { "17", NULL },
         { "18", NULL },
         { "19", NULL },
         { "20", NULL },
         { "21", NULL },
         { "22", NULL },
         { "23", NULL },
         { "24", NULL },
         { "25", NULL },
         { "26", NULL },
         { "27", NULL },
         { "28", NULL },
         { "29", NULL },
         { "30", NULL },
         { "31", NULL },
         { "32", NULL },
         { "33", NULL },
         { "34", NULL },
         { "35", NULL },
         { "36", NULL },
         { "37", NULL },
         { "38", NULL },
         { "39", NULL },
         { "40", NULL },
         { NULL, NULL },
      },
      "0"
   },
   {
      BEETLE_OPT(last_scanline_pal),
      BEETLE_OPT_LAST_SCANLINE_PAL_LABEL_SK,
      NULL,
      BEETLE_OPT_LAST_SCANLINE_PAL_INFO_0_SK,
      NULL,
      "video",
      {
         { "230", NULL },
         { "231", NULL },
         { "232", NULL },
         { "233", NULL },
         { "234", NULL },
         { "235", NULL },
         { "236", NULL },
         { "237", NULL },
         { "238", NULL },
         { "239", NULL },
         { "240", NULL },
         { "241", NULL },
         { "242", NULL },
         { "243", NULL },
         { "244", NULL },
         { "245", NULL },
         { "246", NULL },
         { "247", NULL },
         { "248", NULL },
         { "249", NULL },
         { "250", NULL },
         { "251", NULL },
         { "252", NULL },
         { "253", NULL },
         { "254", NULL },
         { "255", NULL },
         { "256", NULL },
         { "257", NULL },
         { "258", NULL },
         { "259", NULL },
         { "260", NULL },
         { "261", NULL },
         { "262", NULL },
         { "263", NULL },
         { "264", NULL },
         { "265", NULL },
         { "266", NULL },
         { "267", NULL },
         { "268", NULL },
         { "269", NULL },
         { "270", NULL },
         { "271", NULL },
         { "272", NULL },
         { "273", NULL },
         { "274", NULL },
         { "275", NULL },
         { "276", NULL },
         { "277", NULL },
         { "278", NULL },
         { "279", NULL },
         { "280", NULL },
         { "281", NULL },
         { "282", NULL },
         { "283", NULL },
         { "284", NULL },
         { "285", NULL },
         { "286", NULL },
         { "287", OPTION_VAL_287_SK },
         { NULL, NULL },
      },
      "287"
   },
   { NULL, NULL, NULL, NULL, NULL, NULL, {{0}}, NULL },
};
struct retro_core_options_v2 options_sk = {
   option_cats_sk,
   option_defs_sk
};

/* RETRO_LANGUAGE_SR */

#define CATEGORY_SYSTEM_LABEL_SR NULL
#define CATEGORY_SYSTEM_INFO_0_SR NULL
#define CATEGORY_VIDEO_LABEL_SR NULL
#define CATEGORY_VIDEO_INFO_0_SR NULL
#define CATEGORY_PGXP_LABEL_SR NULL
#define CATEGORY_PGXP_INFO_0_SR NULL
#define CATEGORY_OSD_LABEL_SR "Informacije na ekranu"
#define CATEGORY_OSD_INFO_0_SR NULL
#define CATEGORY_INPUT_LABEL_SR "Ulaz"
#define CATEGORY_INPUT_INFO_0_SR NULL
#define CATEGORY_MEMCARDS_LABEL_SR NULL
#define CATEGORY_MEMCARDS_INFO_0_SR NULL
#define CATEGORY_HACKS_LABEL_SR NULL
#define CATEGORY_HACKS_INFO_0_SR NULL
#define BEETLE_OPT_CPU_FREQ_SCALE_LABEL_SR NULL
#define BEETLE_OPT_CPU_FREQ_SCALE_INFO_0_SR NULL
#define OPTION_VAL_50_SR NULL
#define OPTION_VAL_60_SR NULL
#define OPTION_VAL_70_SR NULL
#define OPTION_VAL_80_SR NULL
#define OPTION_VAL_90_SR NULL
#define OPTION_VAL_100_SR NULL
#define OPTION_VAL_110_SR NULL
#define OPTION_VAL_120_SR NULL
#define OPTION_VAL_130_SR NULL
#define OPTION_VAL_140_SR NULL
#define OPTION_VAL_150_SR NULL
#define OPTION_VAL_160_SR NULL
#define OPTION_VAL_170_SR NULL
#define OPTION_VAL_180_SR NULL
#define OPTION_VAL_190_SR NULL
#define OPTION_VAL_200_SR NULL
#define OPTION_VAL_210_SR NULL
#define OPTION_VAL_220_SR NULL
#define OPTION_VAL_230_SR NULL
#define OPTION_VAL_240_SR NULL
#define OPTION_VAL_250_SR NULL
#define OPTION_VAL_260_SR NULL
#define OPTION_VAL_270_SR NULL
#define OPTION_VAL_280_SR NULL
#define OPTION_VAL_290_SR NULL
#define OPTION_VAL_300_SR NULL
#define OPTION_VAL_310_SR NULL
#define OPTION_VAL_320_SR NULL
#define OPTION_VAL_330_SR NULL
#define OPTION_VAL_340_SR NULL
#define OPTION_VAL_350_SR NULL
#define OPTION_VAL_360_SR NULL
#define OPTION_VAL_370_SR NULL
#define OPTION_VAL_380_SR NULL
#define OPTION_VAL_390_SR NULL
#define OPTION_VAL_400_SR NULL
#define OPTION_VAL_410_SR NULL
#define OPTION_VAL_420_SR NULL
#define OPTION_VAL_430_SR NULL
#define OPTION_VAL_440_SR NULL
#define OPTION_VAL_450_SR NULL
#define OPTION_VAL_460_SR NULL
#define OPTION_VAL_470_SR NULL
#define OPTION_VAL_480_SR NULL
#define OPTION_VAL_490_SR NULL
#define OPTION_VAL_500_SR NULL
#define OPTION_VAL_510_SR NULL
#define OPTION_VAL_520_SR NULL
#define OPTION_VAL_530_SR NULL
#define OPTION_VAL_540_SR NULL
#define OPTION_VAL_550_SR NULL
#define OPTION_VAL_560_SR NULL
#define OPTION_VAL_570_SR NULL
#define OPTION_VAL_580_SR NULL
#define OPTION_VAL_590_SR NULL
#define OPTION_VAL_600_SR NULL
#define OPTION_VAL_610_SR NULL
#define OPTION_VAL_620_SR NULL
#define OPTION_VAL_630_SR NULL
#define OPTION_VAL_640_SR NULL
#define OPTION_VAL_650_SR NULL
#define OPTION_VAL_660_SR NULL
#define OPTION_VAL_670_SR NULL
#define OPTION_VAL_680_SR NULL
#define OPTION_VAL_690_SR NULL
#define OPTION_VAL_700_SR NULL
#define OPTION_VAL_710_SR NULL
#define OPTION_VAL_720_SR NULL
#define OPTION_VAL_730_SR NULL
#define OPTION_VAL_740_SR NULL
#define OPTION_VAL_750_SR NULL
#define BEETLE_OPT_GPU_OVERCLOCK_LABEL_SR NULL
#define BEETLE_OPT_GPU_OVERCLOCK_INFO_0_SR NULL
#define OPTION_VAL_1X_NATIVE_SR NULL
#define OPTION_VAL_2X_SR NULL
#define OPTION_VAL_4X_SR NULL
#define OPTION_VAL_8X_SR NULL
#define OPTION_VAL_16X_SR NULL
#define OPTION_VAL_32X_SR NULL
#define BEETLE_OPT_GTE_OVERCLOCK_LABEL_SR NULL
#define BEETLE_OPT_GTE_OVERCLOCK_INFO_0_SR NULL
#define BEETLE_OPT_CD_ACCESS_METHOD_LABEL_SR NULL
#define BEETLE_OPT_CD_ACCESS_METHOD_INFO_0_SR NULL
#define OPTION_VAL_SYNC_SR NULL
#define OPTION_VAL_ASYNC_SR NULL
#define OPTION_VAL_PRECACHE_SR NULL
#define BEETLE_OPT_CD_FASTLOAD_LABEL_SR NULL
#define BEETLE_OPT_CD_FASTLOAD_INFO_0_SR NULL
#define OPTION_VAL_2X_NATIVE_SR NULL
#define OPTION_VAL_6X_SR NULL
#define OPTION_VAL_10X_SR NULL
#define OPTION_VAL_12X_SR NULL
#define OPTION_VAL_14X_SR NULL
#define BEETLE_OPT_SKIP_BIOS_LABEL_SR NULL
#define BEETLE_OPT_SKIP_BIOS_INFO_0_SR NULL
#define BEETLE_OPT_OVERRIDE_BIOS_LABEL_SR NULL
#define BEETLE_OPT_OVERRIDE_BIOS_INFO_0_SR NULL
#define OPTION_VAL_PSXONPSP_SR NULL
#define OPTION_VAL_PS1_ROM_SR NULL
#define OPTION_VAL_OPENBIOS_SR NULL
#define BEETLE_OPT_REGION_LABEL_SR NULL
#define BEETLE_OPT_REGION_INFO_0_SR NULL
#define OPTION_VAL_AUTO_SR NULL
#define OPTION_VAL_NTSC_J_SR NULL
#define OPTION_VAL_NTSC_U_SR NULL
#define OPTION_VAL_PAL_SR NULL
#define BEETLE_OPT_INTERNAL_RESOLUTION_LABEL_SR NULL
#define BEETLE_OPT_INTERNAL_RESOLUTION_INFO_0_SR NULL
#define BEETLE_OPT_RENDERER_LABEL_SR NULL
#define BEETLE_OPT_RENDERER_INFO_0_SR NULL
#define OPTION_VAL_HARDWARE_SR NULL
#define OPTION_VAL_HARDWARE_GL_SR NULL
#define OPTION_VAL_HARDWARE_VK_SR NULL
#define OPTION_VAL_SOFTWARE_SR NULL
#define BEETLE_OPT_RENDERER_SOFTWARE_FB_LABEL_SR NULL
#define BEETLE_OPT_RENDERER_SOFTWARE_FB_INFO_0_SR NULL
#define BEETLE_OPT_DEPTH_LABEL_SR NULL
#define BEETLE_OPT_DEPTH_INFO_0_SR NULL
#define OPTION_VAL_16BPP_NATIVE_SR NULL
#define OPTION_VAL_32BPP_SR NULL
#define BEETLE_OPT_DITHER_MODE_LABEL_SR NULL
#define BEETLE_OPT_DITHER_MODE_INFO_0_SR NULL
#define OPTION_VAL_INTERNAL_RESOLUTION_SR NULL
#define BEETLE_OPT_SCALED_UV_OFFSET_LABEL_SR NULL
#define BEETLE_OPT_SCALED_UV_OFFSET_INFO_0_SR NULL
#define BEETLE_OPT_FILTER_LABEL_SR NULL
#define BEETLE_OPT_FILTER_INFO_0_SR NULL
#define OPTION_VAL_NEAREST_SR NULL
#define OPTION_VAL_SABR_SR NULL
#define OPTION_VAL_XBR_SR NULL
#define OPTION_VAL_BILINEAR_SR "Bilinearno"
#define OPTION_VAL_3_POINT_SR NULL
#define OPTION_VAL_JINC2_SR NULL
#define BEETLE_OPT_FILTER_EXCLUDE_SPRITE_LABEL_SR NULL
#define BEETLE_OPT_FILTER_EXCLUDE_SPRITE_INFO_0_SR NULL
#define OPTION_VAL_OPAQUE_SR NULL
#define OPTION_VAL_ALL_SR NULL
#define BEETLE_OPT_FILTER_EXCLUDE_2D_POLYGON_LABEL_SR NULL
#define BEETLE_OPT_FILTER_EXCLUDE_2D_POLYGON_INFO_0_SR NULL
#define BEETLE_OPT_ADAPTIVE_SMOOTHING_LABEL_SR NULL
#define BEETLE_OPT_ADAPTIVE_SMOOTHING_INFO_0_SR NULL
#define BEETLE_OPT_SUPER_SAMPLING_LABEL_SR NULL
#define BEETLE_OPT_SUPER_SAMPLING_INFO_0_SR NULL
#define BEETLE_OPT_MSAA_LABEL_SR NULL
#define BEETLE_OPT_MSAA_INFO_0_SR NULL
#define OPTION_VAL_1X_SR NULL
#define BEETLE_OPT_MDEC_YUV_LABEL_SR NULL
#define BEETLE_OPT_MDEC_YUV_INFO_0_SR NULL
#define BEETLE_OPT_TRACK_TEXTURES_LABEL_SR NULL
#define BEETLE_OPT_TRACK_TEXTURES_INFO_0_SR NULL
#define BEETLE_OPT_DUMP_TEXTURES_LABEL_SR NULL
#define BEETLE_OPT_DUMP_TEXTURES_INFO_0_SR NULL
#define BEETLE_OPT_REPLACE_TEXTURES_LABEL_SR NULL
#define BEETLE_OPT_REPLACE_TEXTURES_INFO_0_SR NULL
#define BEETLE_OPT_PGXP_MODE_LABEL_SR NULL
#define BEETLE_OPT_PGXP_MODE_INFO_0_SR NULL
#define OPTION_VAL_MEMORY_ONLY_SR NULL
#define OPTION_VAL_MEMORY_CPU_SR NULL
#define BEETLE_OPT_PGXP_2D_TOL_LABEL_SR NULL
#define BEETLE_OPT_PGXP_2D_TOL_INFO_0_SR NULL
#define OPTION_VAL_0PX_SR NULL
#define OPTION_VAL_1PX_SR NULL
#define OPTION_VAL_2PX_SR NULL
#define OPTION_VAL_3PX_SR NULL
#define OPTION_VAL_4PX_SR NULL
#define OPTION_VAL_5PX_SR NULL
#define OPTION_VAL_6PX_SR NULL
#define OPTION_VAL_7PX_SR NULL
#define OPTION_VAL_8PX_SR NULL
#define BEETLE_OPT_PGXP_NCLIP_LABEL_SR NULL
#define BEETLE_OPT_PGXP_NCLIP_INFO_0_SR NULL
#define BEETLE_OPT_PGXP_VERTEX_LABEL_SR NULL
#define BEETLE_OPT_PGXP_VERTEX_INFO_0_SR NULL
#define BEETLE_OPT_PGXP_TEXTURE_LABEL_SR NULL
#define BEETLE_OPT_PGXP_TEXTURE_INFO_0_SR NULL
#define BEETLE_OPT_DISPLAY_INTERNAL_FPS_LABEL_SR NULL
#define BEETLE_OPT_DISPLAY_INTERNAL_FPS_INFO_0_SR NULL
#define BEETLE_OPT_DISPLAY_OSD_LABEL_SR NULL
#define BEETLE_OPT_DISPLAY_OSD_INFO_0_SR NULL
#define BEETLE_OPT_DISPLAY_VRAM_LABEL_SR NULL
#define BEETLE_OPT_DISPLAY_VRAM_INFO_0_SR NULL
#define BEETLE_OPT_ANALOG_CALIBRATION_LABEL_SR NULL
#define BEETLE_OPT_ANALOG_CALIBRATION_INFO_0_SR NULL
#define BEETLE_OPT_ANALOG_TOGGLE_LABEL_SR NULL
#define BEETLE_OPT_ANALOG_TOGGLE_INFO_0_SR NULL
#define OPTION_VAL_ENABLED_ANALOG_SR NULL
#define BEETLE_OPT_ANALOG_TOGGLE_COMBO_LABEL_SR NULL
#define BEETLE_OPT_ANALOG_TOGGLE_COMBO_INFO_0_SR NULL
#define OPTION_VAL_L1_L2_R1_R2_START_SELECT_SR NULL
#define OPTION_VAL_L1_R1_SELECT_SR NULL
#define OPTION_VAL_L1_R1_START_SR NULL
#define OPTION_VAL_L1_R1_L3_SR NULL
#define OPTION_VAL_L1_R1_R3_SR NULL
#define OPTION_VAL_L2_R2_SELECT_SR NULL
#define OPTION_VAL_L2_R2_START_SR NULL
#define OPTION_VAL_L2_R2_L3_SR NULL
#define OPTION_VAL_L2_R2_R3_SR NULL
#define OPTION_VAL_L3_R3_SR NULL
#define BEETLE_OPT_ANALOG_TOGGLE_HOLD_LABEL_SR NULL
#define BEETLE_OPT_ANALOG_TOGGLE_HOLD_INFO_0_SR NULL
#define OPTION_VAL_0_SR NULL
#define OPTION_VAL_1_SR NULL
#define OPTION_VAL_2_SR NULL
#define OPTION_VAL_3_SR NULL
#define OPTION_VAL_4_SR NULL
#define OPTION_VAL_5_SR NULL
#define BEETLE_OPT_ENABLE_MULTITAP_PORT1_LABEL_SR NULL
#define BEETLE_OPT_ENABLE_MULTITAP_PORT1_INFO_0_SR NULL
#define BEETLE_OPT_ENABLE_MULTITAP_PORT2_LABEL_SR NULL
#define BEETLE_OPT_ENABLE_MULTITAP_PORT2_INFO_0_SR NULL
#define BEETLE_OPT_GUN_INPUT_MODE_LABEL_SR NULL
#define BEETLE_OPT_GUN_INPUT_MODE_INFO_0_SR NULL
#define OPTION_VAL_LIGHTGUN_SR NULL
#define OPTION_VAL_TOUCHSCREEN_SR NULL
#define BEETLE_OPT_GUN_CURSOR_LABEL_SR NULL
#define BEETLE_OPT_GUN_CURSOR_INFO_0_SR NULL
#define OPTION_VAL_CROSS_SR NULL
#define OPTION_VAL_DOT_SR NULL
#define BEETLE_OPT_CROSSHAIR_COLOR_P1_LABEL_SR NULL
#define BEETLE_OPT_CROSSHAIR_COLOR_P1_INFO_0_SR NULL
#define OPTION_VAL_RED_SR NULL
#define OPTION_VAL_BLUE_SR NULL
#define OPTION_VAL_GREEN_SR NULL
#define OPTION_VAL_ORANGE_SR NULL
#define OPTION_VAL_YELLOW_SR NULL
#define OPTION_VAL_CYAN_SR NULL
#define OPTION_VAL_PINK_SR NULL
#define OPTION_VAL_PURPLE_SR NULL
#define OPTION_VAL_BLACK_SR NULL
#define OPTION_VAL_WHITE_SR NULL
#define BEETLE_OPT_CROSSHAIR_COLOR_P2_LABEL_SR NULL
#define BEETLE_OPT_CROSSHAIR_COLOR_P2_INFO_0_SR NULL
#define BEETLE_OPT_MOUSE_SENSITIVITY_LABEL_SR NULL
#define BEETLE_OPT_MOUSE_SENSITIVITY_INFO_0_SR NULL
#define OPTION_VAL_5_O42_SR NULL
#define OPTION_VAL_10_SR NULL
#define OPTION_VAL_15_SR NULL
#define OPTION_VAL_20_SR NULL
#define OPTION_VAL_25_SR NULL
#define OPTION_VAL_30_SR NULL
#define OPTION_VAL_35_SR NULL
#define OPTION_VAL_40_SR NULL
#define OPTION_VAL_45_SR NULL
#define OPTION_VAL_55_SR NULL
#define OPTION_VAL_65_SR NULL
#define OPTION_VAL_75_SR NULL
#define OPTION_VAL_85_SR NULL
#define OPTION_VAL_95_SR NULL
#define OPTION_VAL_100_O42_SR NULL
#define OPTION_VAL_105_SR NULL
#define OPTION_VAL_115_SR NULL
#define OPTION_VAL_125_SR NULL
#define OPTION_VAL_135_SR NULL
#define OPTION_VAL_145_SR NULL
#define OPTION_VAL_155_SR NULL
#define OPTION_VAL_165_SR NULL
#define OPTION_VAL_175_SR NULL
#define OPTION_VAL_185_SR NULL
#define OPTION_VAL_195_SR NULL
#define BEETLE_OPT_NEGCON_RESPONSE_LABEL_SR NULL
#define BEETLE_OPT_NEGCON_RESPONSE_INFO_0_SR NULL
#define OPTION_VAL_LINEAR_SR NULL
#define OPTION_VAL_QUADRATIC_SR NULL
#define OPTION_VAL_CUBIC_SR NULL
#define BEETLE_OPT_NEGCON_DEADZONE_LABEL_SR NULL
#define BEETLE_OPT_NEGCON_DEADZONE_INFO_0_SR NULL
#define OPTION_VAL_0_O44_SR NULL
#define BEETLE_OPT_USE_MEDNAFEN_MEMCARD0_METHOD_LABEL_SR NULL
#define BEETLE_OPT_USE_MEDNAFEN_MEMCARD0_METHOD_INFO_0_SR NULL
#define OPTION_VAL_LIBRETRO_SR NULL
#define OPTION_VAL_MEDNAFEN_SR NULL
#define BEETLE_OPT_ENABLE_MEMCARD1_LABEL_SR NULL
#define BEETLE_OPT_ENABLE_MEMCARD1_INFO_0_SR NULL
#define BEETLE_OPT_SHARED_MEMORY_CARDS_LABEL_SR NULL
#define BEETLE_OPT_SHARED_MEMORY_CARDS_INFO_0_SR NULL
#define BEETLE_OPT_MEMCARD_LEFT_INDEX_LABEL_SR NULL
#define BEETLE_OPT_MEMCARD_LEFT_INDEX_INFO_0_SR NULL
#define OPTION_VAL_0_O48_SR NULL
#define BEETLE_OPT_MEMCARD_RIGHT_INDEX_LABEL_SR NULL
#define BEETLE_OPT_MEMCARD_RIGHT_INDEX_INFO_0_SR NULL
#define OPTION_VAL_1_O49_SR NULL
#define BEETLE_OPT_LINE_RENDER_LABEL_SR NULL
#define BEETLE_OPT_LINE_RENDER_INFO_0_SR NULL
#define OPTION_VAL_DEFAULT_SR NULL
#define OPTION_VAL_AGGRESSIVE_SR NULL
#define BEETLE_OPT_WIDESCREEN_HACK_LABEL_SR NULL
#define BEETLE_OPT_WIDESCREEN_HACK_INFO_0_SR NULL
#define BEETLE_OPT_WIDESCREEN_HACK_ASPECT_RATIO_LABEL_SR NULL
#define BEETLE_OPT_WIDESCREEN_HACK_ASPECT_RATIO_INFO_0_SR NULL
#define OPTION_VAL_16_9_SR NULL
#define OPTION_VAL_16_10_SR NULL
#define OPTION_VAL_18_9_SR NULL
#define OPTION_VAL_19_9_SR NULL
#define OPTION_VAL_20_9_SR NULL
#define OPTION_VAL_21_9_SR NULL
#define OPTION_VAL_32_9_SR NULL
#define BEETLE_OPT_CPU_DYNAREC_LABEL_SR NULL
#define BEETLE_OPT_CPU_DYNAREC_INFO_0_SR NULL
#define OPTION_VAL_DISABLED_SR NULL
#define OPTION_VAL_EXECUTE_SR NULL
#define OPTION_VAL_EXECUTE_ONE_SR NULL
#define OPTION_VAL_RUN_INTERPRETER_SR NULL
#define BEETLE_OPT_DYNAREC_INVALIDATE_LABEL_SR NULL
#define BEETLE_OPT_DYNAREC_INVALIDATE_INFO_0_SR NULL
#define OPTION_VAL_FULL_SR NULL
#define OPTION_VAL_DMA_SR NULL
#define BEETLE_OPT_DYNAREC_EVENTCYCLES_LABEL_SR NULL
#define BEETLE_OPT_DYNAREC_EVENTCYCLES_INFO_0_SR NULL
#define OPTION_VAL_128_SR NULL
#define BEETLE_OPT_DYNAREC_SPU_SAMPLES_LABEL_SR NULL
#define BEETLE_OPT_DYNAREC_SPU_SAMPLES_INFO_0_SR NULL
#define BEETLE_OPT_CORE_TIMING_FPS_LABEL_SR NULL
#define BEETLE_OPT_CORE_TIMING_FPS_INFO_0_SR NULL
#define OPTION_VAL_AUTO_TOGGLE_SR NULL
#define OPTION_VAL_FORCE_PROGRESSIVE_SR NULL
#define OPTION_VAL_FORCE_INTERLACED_SR NULL
#define BEETLE_OPT_DEINTERLACER_LABEL_SR NULL
#define BEETLE_OPT_DEINTERLACER_INFO_0_SR NULL
#define OPTION_VAL_WEAVE_SR NULL
#define OPTION_VAL_BOB_SR NULL
#define OPTION_VAL_BOB_OFFSET_SR NULL
#define OPTION_VAL_FASTMAD_SR NULL
#define BEETLE_OPT_FRAME_DUPING_LABEL_SR NULL
#define BEETLE_OPT_FRAME_DUPING_INFO_0_SR NULL
#define BEETLE_OPT_PAL_VIDEO_TIMING_OVERRIDE_LABEL_SR NULL
#define BEETLE_OPT_PAL_VIDEO_TIMING_OVERRIDE_INFO_0_SR NULL
#define BEETLE_OPT_ASPECT_RATIO_LABEL_SR NULL
#define BEETLE_OPT_ASPECT_RATIO_INFO_0_SR NULL
#define OPTION_VAL_CORRECTED_SR NULL
#define OPTION_VAL_UNCORRECTED_SR NULL
#define OPTION_VAL_4_3_SR NULL
#define OPTION_VAL_NTSC_SR NULL
#define BEETLE_OPT_CROP_OVERSCAN_LABEL_SR NULL
#define BEETLE_OPT_CROP_OVERSCAN_INFO_0_SR NULL
#define OPTION_VAL_STATIC_SR NULL
#define OPTION_VAL_SMART_SR NULL
#define BEETLE_OPT_IMAGE_CROP_LABEL_SR NULL
#define BEETLE_OPT_IMAGE_CROP_INFO_0_SR NULL
#define OPTION_VAL_9PX_SR NULL
#define OPTION_VAL_10PX_SR NULL
#define OPTION_VAL_11PX_SR NULL
#define OPTION_VAL_12PX_SR NULL
#define OPTION_VAL_13PX_SR NULL
#define OPTION_VAL_14PX_SR NULL
#define OPTION_VAL_15PX_SR NULL
#define OPTION_VAL_16PX_SR NULL
#define OPTION_VAL_17PX_SR NULL
#define OPTION_VAL_18PX_SR NULL
#define OPTION_VAL_19PX_SR NULL
#define OPTION_VAL_20PX_SR NULL
#define BEETLE_OPT_IMAGE_OFFSET_LABEL_SR NULL
#define BEETLE_OPT_IMAGE_OFFSET_INFO_0_SR NULL
#define OPTION_VAL_12PX_O64_SR NULL
#define OPTION_VAL_11PX_O64_SR NULL
#define OPTION_VAL_10PX_O64_SR NULL
#define OPTION_VAL_9PX_O64_SR NULL
#define OPTION_VAL_8PX_O64_SR NULL
#define OPTION_VAL_7PX_O64_SR NULL
#define OPTION_VAL_6PX_O64_SR NULL
#define OPTION_VAL_5PX_O64_SR NULL
#define OPTION_VAL_4PX_O64_SR NULL
#define OPTION_VAL_3PX_O64_SR NULL
#define OPTION_VAL_2PX_O64_SR NULL
#define OPTION_VAL_1PX_O64_SR NULL
#define OPTION_VAL_1PX_O64_0_SR NULL
#define OPTION_VAL_2PX_O64_0_SR NULL
#define OPTION_VAL_3PX_O64_0_SR NULL
#define OPTION_VAL_4PX_O64_0_SR NULL
#define OPTION_VAL_5PX_O64_0_SR NULL
#define OPTION_VAL_6PX_O64_0_SR NULL
#define OPTION_VAL_7PX_O64_0_SR NULL
#define OPTION_VAL_8PX_O64_0_SR NULL
#define OPTION_VAL_9PX_O64_0_SR NULL
#define OPTION_VAL_10PX_O64_0_SR NULL
#define OPTION_VAL_11PX_O64_0_SR NULL
#define OPTION_VAL_12PX_O64_0_SR NULL
#define BEETLE_OPT_IMAGE_OFFSET_CYCLES_LABEL_SR NULL
#define BEETLE_OPT_IMAGE_OFFSET_CYCLES_INFO_0_SR NULL
#define BEETLE_OPT_INITIAL_SCANLINE_LABEL_SR NULL
#define BEETLE_OPT_INITIAL_SCANLINE_INFO_0_SR NULL
#define BEETLE_OPT_LAST_SCANLINE_LABEL_SR NULL
#define BEETLE_OPT_LAST_SCANLINE_INFO_0_SR NULL
#define OPTION_VAL_239_SR NULL
#define BEETLE_OPT_INITIAL_SCANLINE_PAL_LABEL_SR NULL
#define BEETLE_OPT_INITIAL_SCANLINE_PAL_INFO_0_SR NULL
#define BEETLE_OPT_LAST_SCANLINE_PAL_LABEL_SR NULL
#define BEETLE_OPT_LAST_SCANLINE_PAL_INFO_0_SR NULL
#define OPTION_VAL_287_SR NULL

struct retro_core_option_v2_category option_cats_sr[] = {
   {
      "system",
      CATEGORY_SYSTEM_LABEL_SR,
      CATEGORY_SYSTEM_INFO_0_SR
   },
   {
      "video",
      CATEGORY_VIDEO_LABEL_SR,
      CATEGORY_VIDEO_INFO_0_SR
   },
   {
      "pgxp",
      CATEGORY_PGXP_LABEL_SR,
      CATEGORY_PGXP_INFO_0_SR
   },
   {
      "osd",
      CATEGORY_OSD_LABEL_SR,
      CATEGORY_OSD_INFO_0_SR
   },
   {
      "input",
      CATEGORY_INPUT_LABEL_SR,
      CATEGORY_INPUT_INFO_0_SR
   },
   {
      "memcards",
      CATEGORY_MEMCARDS_LABEL_SR,
      CATEGORY_MEMCARDS_INFO_0_SR
   },
   {
      "hacks",
      CATEGORY_HACKS_LABEL_SR,
      CATEGORY_HACKS_INFO_0_SR
   },
   { NULL, NULL, NULL },
};
struct retro_core_option_v2_definition option_defs_sr[] = {
   {
      BEETLE_OPT(cpu_freq_scale),
      BEETLE_OPT_CPU_FREQ_SCALE_LABEL_SR,
      NULL,
      BEETLE_OPT_CPU_FREQ_SCALE_INFO_0_SR,
      NULL,
      "system",
      {
         { "50%",           OPTION_VAL_50_SR },
         { "60%",           OPTION_VAL_60_SR },
         { "70%",           OPTION_VAL_70_SR },
         { "80%",           OPTION_VAL_80_SR },
         { "90%",           OPTION_VAL_90_SR },
         { "100%",          OPTION_VAL_100_SR },
         { "110%",          OPTION_VAL_110_SR },
         { "120%",          OPTION_VAL_120_SR },
         { "130%",          OPTION_VAL_130_SR },
         { "140%",          OPTION_VAL_140_SR },
         { "150%",          OPTION_VAL_150_SR },
         { "160%",          OPTION_VAL_160_SR },
         { "170%",          OPTION_VAL_170_SR },
         { "180%",          OPTION_VAL_180_SR },
         { "190%",          OPTION_VAL_190_SR },
         { "200%",          OPTION_VAL_200_SR },
         { "210%",          OPTION_VAL_210_SR },
         { "220%",          OPTION_VAL_220_SR },
         { "230%",          OPTION_VAL_230_SR },
         { "240%",          OPTION_VAL_240_SR },
         { "250%",          OPTION_VAL_250_SR },
         { "260%",          OPTION_VAL_260_SR },
         { "270%",          OPTION_VAL_270_SR },
         { "280%",          OPTION_VAL_280_SR },
         { "290%",          OPTION_VAL_290_SR },
         { "300%",          OPTION_VAL_300_SR },
         { "310%",          OPTION_VAL_310_SR },
         { "320%",          OPTION_VAL_320_SR },
         { "330%",          OPTION_VAL_330_SR },
         { "340%",          OPTION_VAL_340_SR },
         { "350%",          OPTION_VAL_350_SR },
         { "360%",          OPTION_VAL_360_SR },
         { "370%",          OPTION_VAL_370_SR },
         { "380%",          OPTION_VAL_380_SR },
         { "390%",          OPTION_VAL_390_SR },
         { "400%",          OPTION_VAL_400_SR },
         { "410%",          OPTION_VAL_410_SR },
         { "420%",          OPTION_VAL_420_SR },
         { "430%",          OPTION_VAL_430_SR },
         { "440%",          OPTION_VAL_440_SR },
         { "450%",          OPTION_VAL_450_SR },
         { "460%",          OPTION_VAL_460_SR },
         { "470%",          OPTION_VAL_470_SR },
         { "480%",          OPTION_VAL_480_SR },
         { "490%",          OPTION_VAL_490_SR },
         { "500%",          OPTION_VAL_500_SR },
         { "510%",          OPTION_VAL_510_SR },
         { "520%",          OPTION_VAL_520_SR },
         { "530%",          OPTION_VAL_530_SR },
         { "540%",          OPTION_VAL_540_SR },
         { "550%",          OPTION_VAL_550_SR },
         { "560%",          OPTION_VAL_560_SR },
         { "570%",          OPTION_VAL_570_SR },
         { "580%",          OPTION_VAL_580_SR },
         { "590%",          OPTION_VAL_590_SR },
         { "600%",          OPTION_VAL_600_SR },
         { "610%",          OPTION_VAL_610_SR },
         { "620%",          OPTION_VAL_620_SR },
         { "630%",          OPTION_VAL_630_SR },
         { "640%",          OPTION_VAL_640_SR },
         { "650%",          OPTION_VAL_650_SR },
         { "660%",          OPTION_VAL_660_SR },
         { "670%",          OPTION_VAL_670_SR },
         { "680%",          OPTION_VAL_680_SR },
         { "690%",          OPTION_VAL_690_SR },
         { "700%",          OPTION_VAL_700_SR },
         { "710%",          OPTION_VAL_710_SR },
         { "720%",          OPTION_VAL_720_SR },
         { "730%",          OPTION_VAL_730_SR },
         { "740%",          OPTION_VAL_740_SR },
         { "750%",          OPTION_VAL_750_SR },
         { NULL, NULL },
      },
      "100%"
   },
   {
      BEETLE_OPT(gpu_overclock),
      BEETLE_OPT_GPU_OVERCLOCK_LABEL_SR,
      NULL,
      BEETLE_OPT_GPU_OVERCLOCK_INFO_0_SR,
      NULL,
      "system",
      {
         { "1x(native)", OPTION_VAL_1X_NATIVE_SR },
         { "2x",         OPTION_VAL_2X_SR },
         { "4x",         OPTION_VAL_4X_SR },
         { "8x",         OPTION_VAL_8X_SR },
         { "16x",        OPTION_VAL_16X_SR },
         { "32x",        OPTION_VAL_32X_SR },
         { NULL, NULL },
      },
      "1x(native)"
   },
   {
      BEETLE_OPT(gte_overclock),
      BEETLE_OPT_GTE_OVERCLOCK_LABEL_SR,
      NULL,
      BEETLE_OPT_GTE_OVERCLOCK_INFO_0_SR,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
#ifndef EMSCRIPTEN
   {
      BEETLE_OPT(cd_access_method),
      BEETLE_OPT_CD_ACCESS_METHOD_LABEL_SR,
      NULL,
      BEETLE_OPT_CD_ACCESS_METHOD_INFO_0_SR,
      NULL,
      "system",
      {
         { "sync",     OPTION_VAL_SYNC_SR },
         { "async",    OPTION_VAL_ASYNC_SR },
         { "precache", OPTION_VAL_PRECACHE_SR },
         { NULL, NULL },
      },
      "sync"
   },
#endif
   {
      BEETLE_OPT(cd_fastload),
      BEETLE_OPT_CD_FASTLOAD_LABEL_SR,
      NULL,
      BEETLE_OPT_CD_FASTLOAD_INFO_0_SR,
      NULL,
      "system",
      {
         { "2x(native)", OPTION_VAL_2X_NATIVE_SR },
         { "4x",          OPTION_VAL_4X_SR },
         { "6x",          OPTION_VAL_6X_SR },
         { "8x",          OPTION_VAL_8X_SR },
         { "10x",         OPTION_VAL_10X_SR },
         { "12x",         OPTION_VAL_12X_SR },
         { "14x",         OPTION_VAL_14X_SR },
         { NULL, NULL },
      },
      "2x(native)"
   },
   {
      BEETLE_OPT(skip_bios),
      BEETLE_OPT_SKIP_BIOS_LABEL_SR,
      NULL,
      BEETLE_OPT_SKIP_BIOS_INFO_0_SR,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(override_bios),
      BEETLE_OPT_OVERRIDE_BIOS_LABEL_SR,
      NULL,
      BEETLE_OPT_OVERRIDE_BIOS_INFO_0_SR,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "psxonpsp", OPTION_VAL_PSXONPSP_SR },
         { "ps1_rom",  OPTION_VAL_PS1_ROM_SR },
         { "openbios", OPTION_VAL_OPENBIOS_SR     },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(region),
      BEETLE_OPT_REGION_LABEL_SR,
      NULL,
      BEETLE_OPT_REGION_INFO_0_SR,
      NULL,
      "system",
      {
         { "auto",   OPTION_VAL_AUTO_SR },
         { "ntsc-j", OPTION_VAL_NTSC_J_SR },
         { "ntsc-u", OPTION_VAL_NTSC_U_SR },
         { "pal",    OPTION_VAL_PAL_SR },
         { NULL, NULL },
      },
      "auto"
   },
   {
      BEETLE_OPT(internal_resolution),
      BEETLE_OPT_INTERNAL_RESOLUTION_LABEL_SR,
      NULL,
      BEETLE_OPT_INTERNAL_RESOLUTION_INFO_0_SR,
      NULL,
      "video",
      {
         { "1x(native)", OPTION_VAL_1X_NATIVE_SR },
         { "2x",         OPTION_VAL_2X_SR },
         { "4x",         OPTION_VAL_4X_SR },
         { "8x",         OPTION_VAL_8X_SR },
         { "16x",        OPTION_VAL_16X_SR },
         { NULL, NULL },
      },
      "1x(native)"
   },
#if defined(HAVE_OPENGL) || defined(HAVE_OPENGLES) || defined(HAVE_VULKAN)
   {
      BEETLE_OPT(renderer),
      BEETLE_OPT_RENDERER_LABEL_SR,
      NULL,
      BEETLE_OPT_RENDERER_INFO_0_SR,
      NULL,
      "video",
      {
         { "hardware",    OPTION_VAL_HARDWARE_SR },
#if defined(HAVE_OPENGL) || defined(HAVE_OPENGLES)
         { "hardware_gl", OPTION_VAL_HARDWARE_GL_SR },
#endif
#if defined(HAVE_VULKAN)
         { "hardware_vk", OPTION_VAL_HARDWARE_VK_SR },
#endif
         { "software",    OPTION_VAL_SOFTWARE_SR },
         { NULL, NULL },
      },
      "hardware"
   },
   {
      BEETLE_OPT(renderer_software_fb),
      BEETLE_OPT_RENDERER_SOFTWARE_FB_LABEL_SR,
      NULL,
      BEETLE_OPT_RENDERER_SOFTWARE_FB_INFO_0_SR,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled"
   },
#endif
#if defined(HAVE_OPENGL) || defined(HAVE_OPENGLES)
   {
      BEETLE_OPT(depth),
      BEETLE_OPT_DEPTH_LABEL_SR,
      NULL,
      BEETLE_OPT_DEPTH_INFO_0_SR,
      NULL,
      "video",
      {
         { "16bpp(native)", OPTION_VAL_16BPP_NATIVE_SR },
         { "32bpp",         OPTION_VAL_32BPP_SR },
         { NULL, NULL },
      },
      "16bpp(native)"
   },
   // Sort of, it's more like 15-bit high color and 24-bit true color for visible output. The alpha channel is used for mask bit. Vulkan renderer uses ABGR1555_555 for 31 bits internal? FMVs are always 24-bit on all renderers like original hardware (BGR888, no alpha)
#endif
   {
      BEETLE_OPT(dither_mode),
      BEETLE_OPT_DITHER_MODE_LABEL_SR,
      NULL,
      BEETLE_OPT_DITHER_MODE_INFO_0_SR,
      NULL,
      "video",
      {
         { "1x(native)",          OPTION_VAL_1X_NATIVE_SR },
         { "internal resolution", OPTION_VAL_INTERNAL_RESOLUTION_SR },
         { "disabled",            NULL },
         { NULL, NULL },
      },
      "1x(native)"
   },
#ifdef HAVE_VULKAN
   {
      BEETLE_OPT(scaled_uv_offset),
      BEETLE_OPT_SCALED_UV_OFFSET_LABEL_SR,
      NULL,
      BEETLE_OPT_SCALED_UV_OFFSET_INFO_0_SR,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled"
   },
#endif
#if defined(HAVE_OPENGL) || defined(HAVE_OPENGLES) || defined(HAVE_VULKAN)
   {
      BEETLE_OPT(filter),
      BEETLE_OPT_FILTER_LABEL_SR,
      NULL,
      BEETLE_OPT_FILTER_INFO_0_SR,
      NULL,
      "video",
      {
         { "nearest",  OPTION_VAL_NEAREST_SR },
         { "SABR",     OPTION_VAL_SABR_SR },
         { "xBR",      OPTION_VAL_XBR_SR },
         { "bilinear", OPTION_VAL_BILINEAR_SR },
         { "3-point",  OPTION_VAL_3_POINT_SR },
         { "JINC2",    OPTION_VAL_JINC2_SR },
         { NULL, NULL },
      },
      "nearest"
   },
#ifdef HAVE_VULKAN
   {
      BEETLE_OPT(filter_exclude_sprite),
      BEETLE_OPT_FILTER_EXCLUDE_SPRITE_LABEL_SR,
      NULL,
      BEETLE_OPT_FILTER_EXCLUDE_SPRITE_INFO_0_SR,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "opaque", OPTION_VAL_OPAQUE_SR },
         { "all", OPTION_VAL_ALL_SR },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(filter_exclude_2d_polygon),
      BEETLE_OPT_FILTER_EXCLUDE_2D_POLYGON_LABEL_SR,
      NULL,
      BEETLE_OPT_FILTER_EXCLUDE_2D_POLYGON_INFO_0_SR,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "opaque", OPTION_VAL_OPAQUE_SR },
         { "all", OPTION_VAL_ALL_SR },
         { NULL, NULL },
      },
      "disabled"
   },
#endif
#endif
#ifdef HAVE_VULKAN
   {
      BEETLE_OPT(adaptive_smoothing),
      BEETLE_OPT_ADAPTIVE_SMOOTHING_LABEL_SR,
      NULL,
      BEETLE_OPT_ADAPTIVE_SMOOTHING_INFO_0_SR,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(super_sampling),
      BEETLE_OPT_SUPER_SAMPLING_LABEL_SR,
      NULL,
      BEETLE_OPT_SUPER_SAMPLING_INFO_0_SR,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(msaa),
      BEETLE_OPT_MSAA_LABEL_SR,
      NULL,
      BEETLE_OPT_MSAA_INFO_0_SR,
      NULL,
      "video",
      {
         { "1x",  OPTION_VAL_1X_SR },
         { "2x",  OPTION_VAL_2X_SR },
         { "4x",  OPTION_VAL_4X_SR },
         { "8x",  OPTION_VAL_8X_SR },
         { "16x", OPTION_VAL_16X_SR },
         { NULL, NULL },
      },
      "1x"
   },
   {
      BEETLE_OPT(mdec_yuv),
      BEETLE_OPT_MDEC_YUV_LABEL_SR,
      NULL,
      BEETLE_OPT_MDEC_YUV_INFO_0_SR,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(track_textures),
      BEETLE_OPT_TRACK_TEXTURES_LABEL_SR,
      NULL,
      BEETLE_OPT_TRACK_TEXTURES_INFO_0_SR,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
#ifdef TEXTURE_DUMPING_ENABLED
   {
      BEETLE_OPT(dump_textures),
      BEETLE_OPT_DUMP_TEXTURES_LABEL_SR,
      NULL,
      BEETLE_OPT_DUMP_TEXTURES_INFO_0_SR,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
#endif
   {
      BEETLE_OPT(replace_textures),
      BEETLE_OPT_REPLACE_TEXTURES_LABEL_SR,
      NULL,
      BEETLE_OPT_REPLACE_TEXTURES_INFO_0_SR,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
#endif
   {
      BEETLE_OPT(pgxp_mode),
      BEETLE_OPT_PGXP_MODE_LABEL_SR,
      NULL,
      BEETLE_OPT_PGXP_MODE_INFO_0_SR,
      NULL,
      "pgxp",
      {
         { "disabled",     NULL },
         { "memory only",  OPTION_VAL_MEMORY_ONLY_SR },
         { "memory + CPU", OPTION_VAL_MEMORY_CPU_SR },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(pgxp_2d_tol),
      BEETLE_OPT_PGXP_2D_TOL_LABEL_SR,
      NULL,
      BEETLE_OPT_PGXP_2D_TOL_INFO_0_SR,
      NULL,
      "pgxp",
      {
         { "disabled", NULL },
         { "0px", OPTION_VAL_0PX_SR },
         { "1px", OPTION_VAL_1PX_SR },
         { "2px", OPTION_VAL_2PX_SR },
         { "3px", OPTION_VAL_3PX_SR },
         { "4px", OPTION_VAL_4PX_SR },
         { "5px", OPTION_VAL_5PX_SR },
         { "6px", OPTION_VAL_6PX_SR },
         { "7px", OPTION_VAL_7PX_SR },
         { "8px", OPTION_VAL_8PX_SR },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(pgxp_nclip),
      BEETLE_OPT_PGXP_NCLIP_LABEL_SR,
      NULL,
      BEETLE_OPT_PGXP_NCLIP_INFO_0_SR,
      NULL,
      "pgxp",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
#if defined(HAVE_OPENGL) || defined(HAVE_OPENGLES) || defined(HAVE_VULKAN)
   {
      BEETLE_OPT(pgxp_vertex),
      BEETLE_OPT_PGXP_VERTEX_LABEL_SR,
      NULL,
      BEETLE_OPT_PGXP_VERTEX_INFO_0_SR,
      NULL,
      "pgxp",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
#endif
   {
      BEETLE_OPT(pgxp_texture),
      BEETLE_OPT_PGXP_TEXTURE_LABEL_SR,
      NULL,
      BEETLE_OPT_PGXP_TEXTURE_INFO_0_SR,
      NULL,
      "pgxp",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(display_internal_fps),
      BEETLE_OPT_DISPLAY_INTERNAL_FPS_LABEL_SR,
      NULL,
      BEETLE_OPT_DISPLAY_INTERNAL_FPS_INFO_0_SR,
      NULL,
      "osd",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(display_osd),
      BEETLE_OPT_DISPLAY_OSD_LABEL_SR,
      NULL,
      BEETLE_OPT_DISPLAY_OSD_INFO_0_SR,
      NULL,
      "osd",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled"
   },
#if defined(HAVE_OPENGL) || defined(HAVE_OPENGLES) || defined(HAVE_VULKAN)
   {
      BEETLE_OPT(display_vram),
      BEETLE_OPT_DISPLAY_VRAM_LABEL_SR,
      NULL,
      BEETLE_OPT_DISPLAY_VRAM_INFO_0_SR,
      NULL,
      "osd",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
#endif
   {
      BEETLE_OPT(analog_calibration),
      BEETLE_OPT_ANALOG_CALIBRATION_LABEL_SR,
      NULL,
      BEETLE_OPT_ANALOG_CALIBRATION_INFO_0_SR,
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(analog_toggle),
      BEETLE_OPT_ANALOG_TOGGLE_LABEL_SR,
      NULL,
      BEETLE_OPT_ANALOG_TOGGLE_INFO_0_SR,
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { "enabled-analog", OPTION_VAL_ENABLED_ANALOG_SR },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(analog_toggle_combo),
      BEETLE_OPT_ANALOG_TOGGLE_COMBO_LABEL_SR,
      NULL,
      BEETLE_OPT_ANALOG_TOGGLE_COMBO_INFO_0_SR,
      NULL,
      "input",
      {
         { "l1+l2+r1+r2+start+select", OPTION_VAL_L1_L2_R1_R2_START_SELECT_SR },
         { "l1+r1+select",             OPTION_VAL_L1_R1_SELECT_SR },
         { "l1+r1+start",              OPTION_VAL_L1_R1_START_SR },
         { "l1+r1+l3",                 OPTION_VAL_L1_R1_L3_SR },
         { "l1+r1+r3",                 OPTION_VAL_L1_R1_R3_SR },
         { "l2+r2+select",             OPTION_VAL_L2_R2_SELECT_SR },
         { "l2+r2+start",              OPTION_VAL_L2_R2_START_SR },
         { "l2+r2+l3",                 OPTION_VAL_L2_R2_L3_SR },
         { "l2+r2+r3",                 OPTION_VAL_L2_R2_R3_SR },
         { "l3+r3",                    OPTION_VAL_L3_R3_SR },
         { NULL, NULL },
      },
      "l1+r1+select"
   },
   {
      BEETLE_OPT(analog_toggle_hold),
      BEETLE_OPT_ANALOG_TOGGLE_HOLD_LABEL_SR,
      NULL,
      BEETLE_OPT_ANALOG_TOGGLE_HOLD_INFO_0_SR,
      NULL,
      "input",
      {
         { "0", OPTION_VAL_0_SR },
         { "1", OPTION_VAL_1_SR },
         { "2", OPTION_VAL_2_SR },
         { "3", OPTION_VAL_3_SR },
         { "4", OPTION_VAL_4_SR },
         { "5", OPTION_VAL_5_SR },
         { NULL, NULL },
      },
      "1"
   },
   {
      BEETLE_OPT(enable_multitap_port1),
      BEETLE_OPT_ENABLE_MULTITAP_PORT1_LABEL_SR,
      NULL,
      BEETLE_OPT_ENABLE_MULTITAP_PORT1_INFO_0_SR,
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(enable_multitap_port2),
      BEETLE_OPT_ENABLE_MULTITAP_PORT2_LABEL_SR,
      NULL,
      BEETLE_OPT_ENABLE_MULTITAP_PORT2_INFO_0_SR,
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(gun_input_mode),
      BEETLE_OPT_GUN_INPUT_MODE_LABEL_SR,
      NULL,
      BEETLE_OPT_GUN_INPUT_MODE_INFO_0_SR,
      NULL,
      "input",
      {
         { "lightgun",    OPTION_VAL_LIGHTGUN_SR },
         { "touchscreen", OPTION_VAL_TOUCHSCREEN_SR },
         { NULL, NULL },
      },
      "lightgun"
   },
   // Shouldn't the gun_input_mode just be Mouse vs. Touchscreen?
   {
      BEETLE_OPT(gun_cursor),
      BEETLE_OPT_GUN_CURSOR_LABEL_SR,
      NULL,
      BEETLE_OPT_GUN_CURSOR_INFO_0_SR,
      NULL,
      "input",
      {
         { "off",   "disabled" },
         { "cross", OPTION_VAL_CROSS_SR },
         { "dot",   OPTION_VAL_DOT_SR },
         { NULL, NULL },
      },
      "cross"
   },
   {
      BEETLE_OPT(crosshair_color_p1),
      BEETLE_OPT_CROSSHAIR_COLOR_P1_LABEL_SR,
      NULL,
      BEETLE_OPT_CROSSHAIR_COLOR_P1_INFO_0_SR,
      NULL,
      "input",
      {
         { "red",    OPTION_VAL_RED_SR },
         { "blue",   OPTION_VAL_BLUE_SR },
         { "green",  OPTION_VAL_GREEN_SR },
         { "orange", OPTION_VAL_ORANGE_SR },
         { "yellow", OPTION_VAL_YELLOW_SR },
         { "cyan",   OPTION_VAL_CYAN_SR },
         { "pink",   OPTION_VAL_PINK_SR },
         { "purple", OPTION_VAL_PURPLE_SR },
         { "black",  OPTION_VAL_BLACK_SR },
         { "white",  OPTION_VAL_WHITE_SR },
         { NULL, NULL },
      },
      "red"
   },
   {
      BEETLE_OPT(crosshair_color_p2),
      BEETLE_OPT_CROSSHAIR_COLOR_P2_LABEL_SR,
      NULL,
      BEETLE_OPT_CROSSHAIR_COLOR_P2_INFO_0_SR,
      NULL,
      "input",
      {
         { "blue",   OPTION_VAL_BLUE_SR },
         { "red",    OPTION_VAL_RED_SR },
         { "green",  OPTION_VAL_GREEN_SR },
         { "orange", OPTION_VAL_ORANGE_SR },
         { "yellow", OPTION_VAL_YELLOW_SR },
         { "cyan",   OPTION_VAL_CYAN_SR },
         { "pink",   OPTION_VAL_PINK_SR },
         { "purple", OPTION_VAL_PURPLE_SR },
         { "black",  OPTION_VAL_BLACK_SR },
         { "white",  OPTION_VAL_WHITE_SR },
         { NULL, NULL },
      },
      "blue"
   },
   {
      BEETLE_OPT(mouse_sensitivity),
      BEETLE_OPT_MOUSE_SENSITIVITY_LABEL_SR,
      NULL,
      BEETLE_OPT_MOUSE_SENSITIVITY_INFO_0_SR,
      NULL,
      "input",
      {
         { "5%",   OPTION_VAL_5_O42_SR },
         { "10%",  OPTION_VAL_10_SR },
         { "15%",  OPTION_VAL_15_SR },
         { "20%",  OPTION_VAL_20_SR },
         { "25%",  OPTION_VAL_25_SR },
         { "30%",  OPTION_VAL_30_SR },
         { "35%",  OPTION_VAL_35_SR },
         { "40%",  OPTION_VAL_40_SR },
         { "45%",  OPTION_VAL_45_SR },
         { "50%",  OPTION_VAL_50_SR },
         { "55%",  OPTION_VAL_55_SR },
         { "60%",  OPTION_VAL_60_SR },
         { "65%",  OPTION_VAL_65_SR },
         { "70%",  OPTION_VAL_70_SR },
         { "75%",  OPTION_VAL_75_SR },
         { "80%",  OPTION_VAL_80_SR },
         { "85%",  OPTION_VAL_85_SR },
         { "90%",  OPTION_VAL_90_SR },
         { "95%",  OPTION_VAL_95_SR },
         { "100%", OPTION_VAL_100_O42_SR },
         { "105%", OPTION_VAL_105_SR },
         { "110%", OPTION_VAL_110_SR },
         { "115%", OPTION_VAL_115_SR },
         { "120%", OPTION_VAL_120_SR },
         { "125%", OPTION_VAL_125_SR },
         { "130%", OPTION_VAL_130_SR },
         { "135%", OPTION_VAL_135_SR },
         { "140%", OPTION_VAL_140_SR },
         { "145%", OPTION_VAL_145_SR },
         { "150%", OPTION_VAL_150_SR },
         { "155%", OPTION_VAL_155_SR },
         { "160%", OPTION_VAL_160_SR },
         { "165%", OPTION_VAL_165_SR },
         { "170%", OPTION_VAL_170_SR },
         { "175%", OPTION_VAL_175_SR },
         { "180%", OPTION_VAL_180_SR },
         { "185%", OPTION_VAL_185_SR },
         { "190%", OPTION_VAL_190_SR },
         { "195%", OPTION_VAL_195_SR },
         { "200%", OPTION_VAL_200_SR },
         { NULL, NULL },
      },
      "100%"
   },
   {
      BEETLE_OPT(negcon_response),
      BEETLE_OPT_NEGCON_RESPONSE_LABEL_SR,
      NULL,
      BEETLE_OPT_NEGCON_RESPONSE_INFO_0_SR,
      NULL,
      "input",
      {
         { "linear",    OPTION_VAL_LINEAR_SR },
         { "quadratic", OPTION_VAL_QUADRATIC_SR },
         { "cubic",     OPTION_VAL_CUBIC_SR },
         { NULL, NULL },
      },
      "linear"
   },
   {
      BEETLE_OPT(negcon_deadzone),
      BEETLE_OPT_NEGCON_DEADZONE_LABEL_SR,
      NULL,
      BEETLE_OPT_NEGCON_DEADZONE_INFO_0_SR,
      NULL,
      "input",
      {
         { "0%",  OPTION_VAL_0_O44_SR },
         { "5%",  OPTION_VAL_5_O42_SR },
         { "10%", OPTION_VAL_10_SR },
         { "15%", OPTION_VAL_15_SR },
         { "20%", OPTION_VAL_20_SR },
         { "25%", OPTION_VAL_25_SR },
         { "30%", OPTION_VAL_30_SR },
         { NULL, NULL },
      },
      "0%"
   },
   {
      BEETLE_OPT(use_mednafen_memcard0_method),
      BEETLE_OPT_USE_MEDNAFEN_MEMCARD0_METHOD_LABEL_SR,
      NULL,
      BEETLE_OPT_USE_MEDNAFEN_MEMCARD0_METHOD_INFO_0_SR,
      NULL,
      "memcards",
      {
         { "libretro", OPTION_VAL_LIBRETRO_SR },
         { "mednafen", OPTION_VAL_MEDNAFEN_SR },
         { NULL, NULL },
      },
      "libretro"
   },
   {
      BEETLE_OPT(enable_memcard1),
      BEETLE_OPT_ENABLE_MEMCARD1_LABEL_SR,
      NULL,
      BEETLE_OPT_ENABLE_MEMCARD1_INFO_0_SR,
      NULL,
      "memcards",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(shared_memory_cards),
      BEETLE_OPT_SHARED_MEMORY_CARDS_LABEL_SR,
      NULL,
      BEETLE_OPT_SHARED_MEMORY_CARDS_INFO_0_SR,
      NULL,
      "memcards",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(memcard_left_index),
      BEETLE_OPT_MEMCARD_LEFT_INDEX_LABEL_SR,
      NULL,
      BEETLE_OPT_MEMCARD_LEFT_INDEX_INFO_0_SR,
      NULL,
      "memcards",
      {
         { "0",  OPTION_VAL_0_O48_SR },
         { "1",  NULL },
         { "2",  NULL },
         { "3",  NULL },
         { "4",  NULL },
         { "5",  NULL },
         { "6",  NULL },
         { "7",  NULL },
         { "8",  NULL },
         { "9",  NULL },
         { "10", NULL },
         { "11", NULL },
         { "12", NULL },
         { "13", NULL },
         { "14", NULL },
         { "15", NULL },
         { "16", NULL },
         { "17", NULL },
         { "18", NULL },
         { "19", NULL },
         { "20", NULL },
         { "21", NULL },
         { "22", NULL },
         { "23", NULL },
         { "24", NULL },
         { "25", NULL },
         { "26", NULL },
         { "27", NULL },
         { "28", NULL },
         { "29", NULL },
         { "30", NULL },
         { "31", NULL },
         { "32", NULL },
         { "33", NULL },
         { "34", NULL },
         { "35", NULL },
         { "36", NULL },
         { "37", NULL },
         { "38", NULL },
         { "39", NULL },
         { "40", NULL },
         { "41", NULL },
         { "42", NULL },
         { "43", NULL },
         { "44", NULL },
         { "45", NULL },
         { "46", NULL },
         { "47", NULL },
         { "48", NULL },
         { "49", NULL },
         { "50", NULL },
         { "51", NULL },
         { "52", NULL },
         { "53", NULL },
         { "54", NULL },
         { "55", NULL },
         { "56", NULL },
         { "57", NULL },
         { "58", NULL },
         { "59", NULL },
         { "60", NULL },
         { "61", NULL },
         { "62", NULL },
         { "63", NULL },
         { NULL, NULL },
      },
      "0"
   },
   {
      BEETLE_OPT(memcard_right_index),
      BEETLE_OPT_MEMCARD_RIGHT_INDEX_LABEL_SR,
      NULL,
      BEETLE_OPT_MEMCARD_RIGHT_INDEX_INFO_0_SR,
      NULL,
      "memcards",
      {
         { "0",  NULL },
         { "1",  OPTION_VAL_1_O49_SR },
         { "2",  NULL },
         { "3",  NULL },
         { "4",  NULL },
         { "5",  NULL },
         { "6",  NULL },
         { "7",  NULL },
         { "8",  NULL },
         { "9",  NULL },
         { "10", NULL },
         { "11", NULL },
         { "12", NULL },
         { "13", NULL },
         { "14", NULL },
         { "15", NULL },
         { "16", NULL },
         { "17", NULL },
         { "18", NULL },
         { "19", NULL },
         { "20", NULL },
         { "21", NULL },
         { "22", NULL },
         { "23", NULL },
         { "24", NULL },
         { "25", NULL },
         { "26", NULL },
         { "27", NULL },
         { "28", NULL },
         { "29", NULL },
         { "30", NULL },
         { "31", NULL },
         { "32", NULL },
         { "33", NULL },
         { "34", NULL },
         { "35", NULL },
         { "36", NULL },
         { "37", NULL },
         { "38", NULL },
         { "39", NULL },
         { "40", NULL },
         { "41", NULL },
         { "42", NULL },
         { "43", NULL },
         { "44", NULL },
         { "45", NULL },
         { "46", NULL },
         { "47", NULL },
         { "48", NULL },
         { "49", NULL },
         { "50", NULL },
         { "51", NULL },
         { "52", NULL },
         { "53", NULL },
         { "54", NULL },
         { "55", NULL },
         { "56", NULL },
         { "57", NULL },
         { "58", NULL },
         { "59", NULL },
         { "60", NULL },
         { "61", NULL },
         { "62", NULL },
         { "63", NULL },
         { NULL, NULL },
      },
      "1"
   },
   {
      BEETLE_OPT(line_render),
      BEETLE_OPT_LINE_RENDER_LABEL_SR,
      NULL,
      BEETLE_OPT_LINE_RENDER_INFO_0_SR,
      NULL,
      "hacks",
      {
         { "default",    OPTION_VAL_DEFAULT_SR },
         { "aggressive", OPTION_VAL_AGGRESSIVE_SR },
         { "disabled",   NULL },
         { NULL, NULL },
      },
      "default"
   },
   {
      BEETLE_OPT(widescreen_hack),
      BEETLE_OPT_WIDESCREEN_HACK_LABEL_SR,
      NULL,
      BEETLE_OPT_WIDESCREEN_HACK_INFO_0_SR,
      NULL,
      "hacks",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(widescreen_hack_aspect_ratio),
      BEETLE_OPT_WIDESCREEN_HACK_ASPECT_RATIO_LABEL_SR,
      NULL,
      BEETLE_OPT_WIDESCREEN_HACK_ASPECT_RATIO_INFO_0_SR,
      NULL,
      "hacks",
      {
         { "16:9",  OPTION_VAL_16_9_SR },
         { "16:10", OPTION_VAL_16_10_SR },
         { "18:9",  OPTION_VAL_18_9_SR },
         { "19:9",  OPTION_VAL_19_9_SR },
         { "20:9",  OPTION_VAL_20_9_SR },
         { "21:9",  OPTION_VAL_21_9_SR }, // 64:27
         { "32:9",  OPTION_VAL_32_9_SR },
         { NULL,    NULL },
      },
      "16:9"
   },
#if defined(HAVE_LIGHTREC)
   {
      BEETLE_OPT(cpu_dynarec),
      BEETLE_OPT_CPU_DYNAREC_LABEL_SR,
      NULL,
      BEETLE_OPT_CPU_DYNAREC_INFO_0_SR,
      NULL,
      "hacks",
      {
         { "disabled",        OPTION_VAL_DISABLED_SR },
         { "execute",         OPTION_VAL_EXECUTE_SR },
         { "execute_one",     OPTION_VAL_EXECUTE_ONE_SR },
         { "run_interpreter", OPTION_VAL_RUN_INTERPRETER_SR },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(dynarec_invalidate),
      BEETLE_OPT_DYNAREC_INVALIDATE_LABEL_SR,
      NULL,
      BEETLE_OPT_DYNAREC_INVALIDATE_INFO_0_SR,
      NULL,
      "hacks",
      {
         { "full", OPTION_VAL_FULL_SR },
         { "dma",  OPTION_VAL_DMA_SR },
         { NULL, NULL },
      },
      "full"
   },
   {
      BEETLE_OPT(dynarec_eventcycles),
      BEETLE_OPT_DYNAREC_EVENTCYCLES_LABEL_SR,
      NULL,
      BEETLE_OPT_DYNAREC_EVENTCYCLES_INFO_0_SR,
      NULL,
      "hacks",
      {
         { "128",  OPTION_VAL_128_SR },
         { "256",  NULL },
         { "384",  NULL },
         { "512",  NULL },
         { "640",  NULL },
         { "768",  NULL },
         { "896",  NULL },
         { "1024", NULL },
         { "1152", NULL },
         { "1280", NULL },
         { "1408", NULL },
         { "1536", NULL },
         { "1664", NULL },
         { "1792", NULL },
         { "1920", NULL },
         { "2048", NULL },
         { NULL, NULL },
      },
      "128"
   },
   {
      BEETLE_OPT(dynarec_spu_samples),
      BEETLE_OPT_DYNAREC_SPU_SAMPLES_LABEL_SR,
      NULL,
      BEETLE_OPT_DYNAREC_SPU_SAMPLES_INFO_0_SR,
      NULL,
      "hacks",
      {
         { "1",  OPTION_VAL_1_O49_SR },
         { "4",  NULL },
         { "16", NULL },
         { NULL, NULL },
      },
      "1"
   },
#endif
   {
      BEETLE_OPT(core_timing_fps),
      BEETLE_OPT_CORE_TIMING_FPS_LABEL_SR,
      NULL,
      BEETLE_OPT_CORE_TIMING_FPS_INFO_0_SR,
      NULL,
      "video",
      {
         { "auto_toggle",       OPTION_VAL_AUTO_TOGGLE_SR },
         { "force_progressive", OPTION_VAL_FORCE_PROGRESSIVE_SR },
         { "force_interlaced",  OPTION_VAL_FORCE_INTERLACED_SR },
      },
      "auto_toggle"
   },
   {
      BEETLE_OPT(deinterlacer),
      BEETLE_OPT_DEINTERLACER_LABEL_SR,
      NULL,
      BEETLE_OPT_DEINTERLACER_INFO_0_SR,
      NULL,
      "video",
      {
         { "weave",      OPTION_VAL_WEAVE_SR },
         { "bob",        OPTION_VAL_BOB_SR },
         { "bob_offset", OPTION_VAL_BOB_OFFSET_SR },
         { "fastmad",    OPTION_VAL_FASTMAD_SR },
         { "off",        "Off" },
         { NULL, NULL },
      },
      "weave"
   },
   {
      BEETLE_OPT(frame_duping),
      BEETLE_OPT_FRAME_DUPING_LABEL_SR,
      NULL,
      BEETLE_OPT_FRAME_DUPING_INFO_0_SR,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(pal_video_timing_override),
      BEETLE_OPT_PAL_VIDEO_TIMING_OVERRIDE_LABEL_SR,
      NULL,
      BEETLE_OPT_PAL_VIDEO_TIMING_OVERRIDE_INFO_0_SR,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(aspect_ratio),
      BEETLE_OPT_ASPECT_RATIO_LABEL_SR,
      NULL,
      BEETLE_OPT_ASPECT_RATIO_INFO_0_SR,
      NULL,
      "video",
      {
         { "corrected", OPTION_VAL_CORRECTED_SR },
         { "uncorrected", OPTION_VAL_UNCORRECTED_SR },
         { "4:3",  OPTION_VAL_4_3_SR },
         { "ntsc", OPTION_VAL_NTSC_SR },
      },
      "corrected"
   },
   {
      BEETLE_OPT(crop_overscan),
      BEETLE_OPT_CROP_OVERSCAN_LABEL_SR,
      NULL,
      BEETLE_OPT_CROP_OVERSCAN_INFO_0_SR,
      NULL,
      "video",
      {
         { "disabled", "Disabled" },
         { "static",   OPTION_VAL_STATIC_SR },
         { "smart",    OPTION_VAL_SMART_SR },
         { NULL, NULL },
      },
      "smart"
   },
   {
      BEETLE_OPT(image_crop),
      BEETLE_OPT_IMAGE_CROP_LABEL_SR,
      NULL,
      BEETLE_OPT_IMAGE_CROP_INFO_0_SR,
      NULL,
      "video",
      {
         { "disabled", "0" },
         { "1px",      OPTION_VAL_1PX_SR },
         { "2px",      OPTION_VAL_2PX_SR },
         { "3px",      OPTION_VAL_3PX_SR },
         { "4px",      OPTION_VAL_4PX_SR },
         { "5px",      OPTION_VAL_5PX_SR },
         { "6px",      OPTION_VAL_6PX_SR },
         { "7px",      OPTION_VAL_7PX_SR },
         { "8px",      OPTION_VAL_8PX_SR },
         { "9px",      OPTION_VAL_9PX_SR },
         { "10px",     OPTION_VAL_10PX_SR },
         { "11px",     OPTION_VAL_11PX_SR },
         { "12px",     OPTION_VAL_12PX_SR },
         { "13px",     OPTION_VAL_13PX_SR },
         { "14px",     OPTION_VAL_14PX_SR },
         { "15px",     OPTION_VAL_15PX_SR },
         { "16px",     OPTION_VAL_16PX_SR },
         { "17px",     OPTION_VAL_17PX_SR },
         { "18px",     OPTION_VAL_18PX_SR },
         { "19px",     OPTION_VAL_19PX_SR },
         { "20px",     OPTION_VAL_20PX_SR },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(image_offset),
      BEETLE_OPT_IMAGE_OFFSET_LABEL_SR,
      NULL,
      BEETLE_OPT_IMAGE_OFFSET_INFO_0_SR,
      NULL,
      "video",
      {
         { "-12px",    OPTION_VAL_12PX_O64_SR },
         { "-11px",    OPTION_VAL_11PX_O64_SR },
         { "-10px",    OPTION_VAL_10PX_O64_SR },
         { "-9px",     OPTION_VAL_9PX_O64_SR },
         { "-8px",     OPTION_VAL_8PX_O64_SR },
         { "-7px",     OPTION_VAL_7PX_O64_SR },
         { "-6px",     OPTION_VAL_6PX_O64_SR },
         { "-5px",     OPTION_VAL_5PX_O64_SR },
         { "-4px",     OPTION_VAL_4PX_O64_SR },
         { "-3px",     OPTION_VAL_3PX_O64_SR },
         { "-2px",     OPTION_VAL_2PX_O64_SR },
         { "-1px",     OPTION_VAL_1PX_O64_SR },
         { "disabled", "0" },
         { "+1px",     OPTION_VAL_1PX_O64_0_SR },
         { "+2px",     OPTION_VAL_2PX_O64_0_SR },
         { "+3px",     OPTION_VAL_3PX_O64_0_SR },
         { "+4px",     OPTION_VAL_4PX_O64_0_SR },
         { "+5px",     OPTION_VAL_5PX_O64_0_SR },
         { "+6px",     OPTION_VAL_6PX_O64_0_SR },
         { "+7px",     OPTION_VAL_7PX_O64_0_SR },
         { "+8px",     OPTION_VAL_8PX_O64_0_SR },
         { "+9px",     OPTION_VAL_9PX_O64_0_SR },
         { "+10px",    OPTION_VAL_10PX_O64_0_SR },
         { "+11px",    OPTION_VAL_11PX_O64_0_SR },
         { "+12px",    OPTION_VAL_12PX_O64_0_SR },
         { NULL, NULL },
      },
      "disabled"
   },
#if defined(HAVE_OPENGL) || defined(HAVE_OPENGLES) || defined(HAVE_VULKAN)
   {
      BEETLE_OPT(image_offset_cycles),
      BEETLE_OPT_IMAGE_OFFSET_CYCLES_LABEL_SR,
      NULL,
      BEETLE_OPT_IMAGE_OFFSET_CYCLES_INFO_0_SR,
      NULL,
      "video",
      {
         { "-40",      NULL },
         { "-39",      NULL },
         { "-38",      NULL },
         { "-37",      NULL },
         { "-36",      NULL },
         { "-35",      NULL },
         { "-34",      NULL },
         { "-33",      NULL },
         { "-32",      NULL },
         { "-31",      NULL },
         { "-30",      NULL },
         { "-29",      NULL },
         { "-28",      NULL },
         { "-27",      NULL },
         { "-26",      NULL },
         { "-25",      NULL },
         { "-24",      NULL },
         { "-23",      NULL },
         { "-22",      NULL },
         { "-21",      NULL },
         { "-20",      NULL },
         { "-19",      NULL },
         { "-18",      NULL },
         { "-17",      NULL },
         { "-16",      NULL },
         { "-15",      NULL },
         { "-14",      NULL },
         { "-13",      NULL },
         { "-12",      NULL },
         { "-11",      NULL },
         { "-10",      NULL },
         { "-9",       NULL },
         { "-8",       NULL },
         { "-7",       NULL },
         { "-6",       NULL },
         { "-5",       NULL },
         { "-4",       NULL },
         { "-3",       NULL },
         { "-2",       NULL },
         { "-1",       NULL },
         { "0",        "0" },
         { "+1",       NULL },
         { "+2",       NULL },
         { "+3",       NULL },
         { "+4",       NULL },
         { "+5",       NULL },
         { "+6",       NULL },
         { "+7",       NULL },
         { "+8",       NULL },
         { "+9",       NULL },
         { "+10",      NULL },
         { "+11",      NULL },
         { "+12",      NULL },
         { "+13",      NULL },
         { "+14",      NULL },
         { "+15",      NULL },
         { "+16",      NULL },
         { "+17",      NULL },
         { "+18",      NULL },
         { "+19",      NULL },
         { "+20",      NULL },
         { "+21",      NULL },
         { "+22",      NULL },
         { "+23",      NULL },
         { "+24",      NULL },
         { "+25",      NULL },
         { "+26",      NULL },
         { "+27",      NULL },
         { "+28",      NULL },
         { "+29",      NULL },
         { "+30",      NULL },
         { "+31",      NULL },
         { "+32",      NULL },
         { "+33",      NULL },
         { "+34",      NULL },
         { "+35",      NULL },
         { "+36",      NULL },
         { "+37",      NULL },
         { "+38",      NULL },
         { "+39",      NULL },
         { "+40",      NULL },
         { NULL, NULL},
      },
      "0"
   },
#endif
   {
      BEETLE_OPT(initial_scanline),
      BEETLE_OPT_INITIAL_SCANLINE_LABEL_SR,
      NULL,
      BEETLE_OPT_INITIAL_SCANLINE_INFO_0_SR,
      NULL,
      "video",
      {
         { "0",  OPTION_VAL_0_O48_SR },
         { "1",  NULL },
         { "2",  NULL },
         { "3",  NULL },
         { "4",  NULL },
         { "5",  NULL },
         { "6",  NULL },
         { "7",  NULL },
         { "8",  NULL },
         { "9",  NULL },
         { "10", NULL },
         { "11", NULL },
         { "12", NULL },
         { "13", NULL },
         { "14", NULL },
         { "15", NULL },
         { "16", NULL },
         { "17", NULL },
         { "18", NULL },
         { "19", NULL },
         { "20", NULL },
         { "21", NULL },
         { "22", NULL },
         { "23", NULL },
         { "24", NULL },
         { "25", NULL },
         { "26", NULL },
         { "27", NULL },
         { "28", NULL },
         { "29", NULL },
         { "30", NULL },
         { "31", NULL },
         { "32", NULL },
         { "33", NULL },
         { "34", NULL },
         { "35", NULL },
         { "36", NULL },
         { "37", NULL },
         { "38", NULL },
         { "39", NULL },
         { "40", NULL },
         { NULL, NULL },
      },
      "0"
   },
   {
      BEETLE_OPT(last_scanline),
      BEETLE_OPT_LAST_SCANLINE_LABEL_SR,
      NULL,
      BEETLE_OPT_LAST_SCANLINE_INFO_0_SR,
      NULL,
      "video",
      {
         { "210", NULL },
         { "211", NULL },
         { "212", NULL },
         { "213", NULL },
         { "214", NULL },
         { "215", NULL },
         { "216", NULL },
         { "217", NULL },
         { "218", NULL },
         { "219", NULL },
         { "220", NULL },
         { "221", NULL },
         { "222", NULL },
         { "223", NULL },
         { "224", NULL },
         { "225", NULL },
         { "226", NULL },
         { "227", NULL },
         { "228", NULL },
         { "229", NULL },
         { "230", NULL },
         { "231", NULL },
         { "232", NULL },
         { "233", NULL },
         { "234", NULL },
         { "235", NULL },
         { "236", NULL },
         { "237", NULL },
         { "238", NULL },
         { "239", OPTION_VAL_239_SR },
         { NULL, NULL },
      },
      "239"
   },
   {
      BEETLE_OPT(initial_scanline_pal),
      BEETLE_OPT_INITIAL_SCANLINE_PAL_LABEL_SR,
      NULL,
      BEETLE_OPT_INITIAL_SCANLINE_PAL_INFO_0_SR,
      NULL,
      "video",
      {
         { "0",  OPTION_VAL_0_O48_SR },
         { "1",  NULL },
         { "2",  NULL },
         { "3",  NULL },
         { "4",  NULL },
         { "5",  NULL },
         { "6",  NULL },
         { "7",  NULL },
         { "8",  NULL },
         { "9",  NULL },
         { "10", NULL },
         { "11", NULL },
         { "12", NULL },
         { "13", NULL },
         { "14", NULL },
         { "15", NULL },
         { "16", NULL },
         { "17", NULL },
         { "18", NULL },
         { "19", NULL },
         { "20", NULL },
         { "21", NULL },
         { "22", NULL },
         { "23", NULL },
         { "24", NULL },
         { "25", NULL },
         { "26", NULL },
         { "27", NULL },
         { "28", NULL },
         { "29", NULL },
         { "30", NULL },
         { "31", NULL },
         { "32", NULL },
         { "33", NULL },
         { "34", NULL },
         { "35", NULL },
         { "36", NULL },
         { "37", NULL },
         { "38", NULL },
         { "39", NULL },
         { "40", NULL },
         { NULL, NULL },
      },
      "0"
   },
   {
      BEETLE_OPT(last_scanline_pal),
      BEETLE_OPT_LAST_SCANLINE_PAL_LABEL_SR,
      NULL,
      BEETLE_OPT_LAST_SCANLINE_PAL_INFO_0_SR,
      NULL,
      "video",
      {
         { "230", NULL },
         { "231", NULL },
         { "232", NULL },
         { "233", NULL },
         { "234", NULL },
         { "235", NULL },
         { "236", NULL },
         { "237", NULL },
         { "238", NULL },
         { "239", NULL },
         { "240", NULL },
         { "241", NULL },
         { "242", NULL },
         { "243", NULL },
         { "244", NULL },
         { "245", NULL },
         { "246", NULL },
         { "247", NULL },
         { "248", NULL },
         { "249", NULL },
         { "250", NULL },
         { "251", NULL },
         { "252", NULL },
         { "253", NULL },
         { "254", NULL },
         { "255", NULL },
         { "256", NULL },
         { "257", NULL },
         { "258", NULL },
         { "259", NULL },
         { "260", NULL },
         { "261", NULL },
         { "262", NULL },
         { "263", NULL },
         { "264", NULL },
         { "265", NULL },
         { "266", NULL },
         { "267", NULL },
         { "268", NULL },
         { "269", NULL },
         { "270", NULL },
         { "271", NULL },
         { "272", NULL },
         { "273", NULL },
         { "274", NULL },
         { "275", NULL },
         { "276", NULL },
         { "277", NULL },
         { "278", NULL },
         { "279", NULL },
         { "280", NULL },
         { "281", NULL },
         { "282", NULL },
         { "283", NULL },
         { "284", NULL },
         { "285", NULL },
         { "286", NULL },
         { "287", OPTION_VAL_287_SR },
         { NULL, NULL },
      },
      "287"
   },
   { NULL, NULL, NULL, NULL, NULL, NULL, {{0}}, NULL },
};
struct retro_core_options_v2 options_sr = {
   option_cats_sr,
   option_defs_sr
};

/* RETRO_LANGUAGE_SV */

#define CATEGORY_SYSTEM_LABEL_SV NULL
#define CATEGORY_SYSTEM_INFO_0_SV "Ändra alternativ för system och överklockning."
#define CATEGORY_VIDEO_LABEL_SV NULL
#define CATEGORY_VIDEO_INFO_0_SV "Ändra bildformat, beskärning, videofilter och inställningar för bildruteskippning."
#define CATEGORY_PGXP_LABEL_SV "PGXP (Precision Geometrisk Transformering Pipeline)"
#define CATEGORY_PGXP_INFO_0_SV "Dessa alternativ styr förbättringar som kan förbättra grafiken jämfört med den ursprungliga spelkonsolen. PGXP kan eliminera förvrängda texturer och Z-fighting problem."
#define CATEGORY_OSD_LABEL_SV "Skärmvisning"
#define CATEGORY_OSD_INFO_0_SV "Ändra aviseringar som visas på skärmen."
#define CATEGORY_INPUT_LABEL_SV "Inmatning"
#define CATEGORY_INPUT_INFO_0_SV "Ändra inställningar för ljuspistol, mus och neGcon."
#define CATEGORY_MEMCARDS_LABEL_SV "Minneskort"
#define CATEGORY_MEMCARDS_INFO_0_SV "Ändra inställningar relaterade till virtuella minneskort."
#define CATEGORY_HACKS_LABEL_SV "Emulatorhack"
#define CATEGORY_HACKS_INFO_0_SV "Ändra inställningarna för emuleringsnoggrannhet som påverkar prestanda och kompatibilitet på låg nivå."
#define BEETLE_OPT_CPU_FREQ_SCALE_LABEL_SV "CPU-frekvens Skalning (Överklockning)"
#define BEETLE_OPT_CPU_FREQ_SCALE_INFO_0_SV "Överklocka (eller underklocka) den emulerade PSX-processorn. Överklockning kan eliminera inbromsningar och förbättra bildfrekvensen i vissa spel på bekostnad av ökade prestandakrav. Observera att vissa spel har en intern bildfrekvensbegränsare och kanske inte gynnas av överklockning. Kan leda till att vissa effekter animeras snabbare än avsett i vissa titlar vid överklockning."
#define OPTION_VAL_50_SV NULL
#define OPTION_VAL_60_SV NULL
#define OPTION_VAL_70_SV NULL
#define OPTION_VAL_80_SV NULL
#define OPTION_VAL_90_SV NULL
#define OPTION_VAL_100_SV "100% (Inbyggd)"
#define OPTION_VAL_110_SV NULL
#define OPTION_VAL_120_SV NULL
#define OPTION_VAL_130_SV NULL
#define OPTION_VAL_140_SV NULL
#define OPTION_VAL_150_SV NULL
#define OPTION_VAL_160_SV NULL
#define OPTION_VAL_170_SV NULL
#define OPTION_VAL_180_SV NULL
#define OPTION_VAL_190_SV NULL
#define OPTION_VAL_200_SV NULL
#define OPTION_VAL_210_SV NULL
#define OPTION_VAL_220_SV NULL
#define OPTION_VAL_230_SV NULL
#define OPTION_VAL_240_SV NULL
#define OPTION_VAL_250_SV NULL
#define OPTION_VAL_260_SV NULL
#define OPTION_VAL_270_SV NULL
#define OPTION_VAL_280_SV NULL
#define OPTION_VAL_290_SV NULL
#define OPTION_VAL_300_SV NULL
#define OPTION_VAL_310_SV NULL
#define OPTION_VAL_320_SV NULL
#define OPTION_VAL_330_SV NULL
#define OPTION_VAL_340_SV NULL
#define OPTION_VAL_350_SV NULL
#define OPTION_VAL_360_SV NULL
#define OPTION_VAL_370_SV NULL
#define OPTION_VAL_380_SV NULL
#define OPTION_VAL_390_SV NULL
#define OPTION_VAL_400_SV NULL
#define OPTION_VAL_410_SV NULL
#define OPTION_VAL_420_SV NULL
#define OPTION_VAL_430_SV NULL
#define OPTION_VAL_440_SV NULL
#define OPTION_VAL_450_SV NULL
#define OPTION_VAL_460_SV NULL
#define OPTION_VAL_470_SV NULL
#define OPTION_VAL_480_SV NULL
#define OPTION_VAL_490_SV NULL
#define OPTION_VAL_500_SV NULL
#define OPTION_VAL_510_SV NULL
#define OPTION_VAL_520_SV NULL
#define OPTION_VAL_530_SV NULL
#define OPTION_VAL_540_SV NULL
#define OPTION_VAL_550_SV NULL
#define OPTION_VAL_560_SV NULL
#define OPTION_VAL_570_SV NULL
#define OPTION_VAL_580_SV NULL
#define OPTION_VAL_590_SV NULL
#define OPTION_VAL_600_SV NULL
#define OPTION_VAL_610_SV NULL
#define OPTION_VAL_620_SV NULL
#define OPTION_VAL_630_SV NULL
#define OPTION_VAL_640_SV NULL
#define OPTION_VAL_650_SV NULL
#define OPTION_VAL_660_SV NULL
#define OPTION_VAL_670_SV NULL
#define OPTION_VAL_680_SV NULL
#define OPTION_VAL_690_SV NULL
#define OPTION_VAL_700_SV NULL
#define OPTION_VAL_710_SV NULL
#define OPTION_VAL_720_SV NULL
#define OPTION_VAL_730_SV NULL
#define OPTION_VAL_740_SV NULL
#define OPTION_VAL_750_SV NULL
#define BEETLE_OPT_GPU_OVERCLOCK_LABEL_SV "Överklockning för GPU-rasterisering"
#define BEETLE_OPT_GPU_OVERCLOCK_INFO_0_SV "Möjliggör överklockning av den 2D-rasterizer som finns i den emulerade PSX: ens GPU. Förbättrar inte 3D-rendering och har i allmänhet liten effekt."
#define OPTION_VAL_1X_NATIVE_SV "1x (Nativ)"
#define OPTION_VAL_2X_SV NULL
#define OPTION_VAL_4X_SV NULL
#define OPTION_VAL_8X_SV NULL
#define OPTION_VAL_16X_SV NULL
#define OPTION_VAL_32X_SV NULL
#define BEETLE_OPT_GTE_OVERCLOCK_LABEL_SV "GTE Överklockning"
#define BEETLE_OPT_GTE_OVERCLOCK_INFO_0_SV "Sänk alla emulerade GTE-operationer (CPU-koprocessor för 3D-grafik) till en konstant latens på en cykel. För spel som använder GTE i stor utsträckning kan detta förbättra bildfrekvensen och stabiliteten i bildtid avsevärt."
#define BEETLE_OPT_CD_ACCESS_METHOD_LABEL_SV "CD-åtkomstmetod"
#define BEETLE_OPT_CD_ACCESS_METHOD_INFO_0_SV "”Synkron” efterliknar originalhårdvaran. ”Asynkron” kan minska hackande på enheter med långsam lagring. ”Förhandsinläsning” läser in hela diskavbildningen i minnet när innehållet startas, vilket kan förbättra inläsningstiderna i spelet på bekostnad av en initial fördröjning vid start. ’Förhandsinläsning’ kan orsaka problem på system med lite RAM-minne och faller tillbaka till ”Synkron” för fysiska medier. Omstart krävs."
#define OPTION_VAL_SYNC_SV "Synkron"
#define OPTION_VAL_ASYNC_SV "Asynkron"
#define OPTION_VAL_PRECACHE_SV "Förhandsinläsning"
#define BEETLE_OPT_CD_FASTLOAD_LABEL_SV "CD-laddningshastighet"
#define BEETLE_OPT_CD_FASTLOAD_INFO_0_SV "Värden högre än ”2x (Inbyggd)” kan avsevärt minska inläsningstiderna i spelet, men kan orsaka tidsfel. Vissa spel kanske inte fungerar korrekt över ett visst värde."
#define OPTION_VAL_2X_NATIVE_SV "2x (Standard)"
#define OPTION_VAL_6X_SV NULL
#define OPTION_VAL_10X_SV NULL
#define OPTION_VAL_12X_SV NULL
#define OPTION_VAL_14X_SV NULL
#define BEETLE_OPT_SKIP_BIOS_LABEL_SV "Hoppa över BIOS"
#define BEETLE_OPT_SKIP_BIOS_INFO_0_SV "Hoppa över PlayStation BIOS-startanimationen som normalt visas när innehåll laddas. Obs: Om du aktiverar detta uppstår kompatibilitetsproblem med ett antal spel (PAL-kopieringsskyddade spel, Saga Frontier etc.)."
#define BEETLE_OPT_OVERRIDE_BIOS_LABEL_SV "Åsidosätt BIOS"
#define BEETLE_OPT_OVERRIDE_BIOS_INFO_0_SV "Åsidosätt det vanliga regionsspecifika BIOS med ett regionsfritt BIOS om sådant finns. Omstart krävs."
#define OPTION_VAL_PSXONPSP_SV NULL
#define OPTION_VAL_PS1_ROM_SV NULL
#define OPTION_VAL_OPENBIOS_SV NULL
#define BEETLE_OPT_REGION_LABEL_SV "Systemregion"
#define BEETLE_OPT_REGION_INFO_0_SV NULL
#define OPTION_VAL_AUTO_SV NULL
#define OPTION_VAL_NTSC_J_SV NULL
#define OPTION_VAL_NTSC_U_SV NULL
#define OPTION_VAL_PAL_SV NULL
#define BEETLE_OPT_INTERNAL_RESOLUTION_LABEL_SV "Intern GPU-upplösning"
#define BEETLE_OPT_INTERNAL_RESOLUTION_INFO_0_SV "Upplösningar högre än ”1x (Inbyggd)” förbättrar 3D-modellernas trohet, men kräver högre prestanda. 2D-element påverkas i allmänhet inte av denna inställning."
#define BEETLE_OPT_RENDERER_LABEL_SV "Renderare"
#define BEETLE_OPT_RENDERER_INFO_0_SV "Programvarurenderaren är den mest exakta men har höga prestandakrav när den körs med högre interna GPU-upplösningar. Hårdvarurenderarna är mindre exakta, men förbättrar prestandan jämfört med programvarurenderaren vid högre interna upplösningar och möjliggör olika grafiska förbättringar. ”Hårdvara (Auto)” väljer automatiskt Vulkan- eller OpenGL-renderaren, beroende på den aktuella videodrivrutinen för libretro-frontend. Om den medföljande videodrivrutinen inte är kompatibel med Vulkan eller OpenGL 3.3, faller kärnan tillbaka till programvarurenderaren. Omstart krävs."
#define OPTION_VAL_HARDWARE_SV "Hårdvara (Auto)"
#define OPTION_VAL_HARDWARE_GL_SV "Hårdvara (OpenGL)"
#define OPTION_VAL_HARDWARE_VK_SV NULL
#define OPTION_VAL_SOFTWARE_SV "Mjukvara"
#define BEETLE_OPT_RENDERER_SOFTWARE_FB_LABEL_SV "Bildbuffer mjukvara"
#define BEETLE_OPT_RENDERER_SOFTWARE_FB_INFO_0_SV "Möjliggör korrekt emulering av framebuffereffekter (t.ex. rörelseoskärpa, FF7-stridsvirvel) när du använder hårdvarurendering genom att köra en kopia av programvarurenderingen med inbyggd upplösning i bakgrunden. Om funktionen inaktiveras utelämnas dessa operationer (OpenGL) eller renderas på GPU:n (Vulkan). Inaktivering kan förbättra prestandan men kan orsaka allvarliga grafiska fel. Lämna aktiverat om du är osäker."
#define BEETLE_OPT_DEPTH_LABEL_SV "Intern färgdjup"
#define BEETLE_OPT_DEPTH_INFO_0_SV "Högre färgdjup kan minska färgbandningseffekter utan användning av dithering. 16 bpp emulerar originalhårdvaran men kan ge synlig bandning om dithering inte är aktiverat. Vi rekommenderar att du inaktiverar ”Dithering-mönster” när detta alternativ är inställt på 32 bpp."
#define OPTION_VAL_16BPP_NATIVE_SV "16 bpp (Nativ)"
#define OPTION_VAL_32BPP_SV NULL
#define BEETLE_OPT_DITHER_MODE_LABEL_SV "Dithering-mönster"
#define BEETLE_OPT_DITHER_MODE_INFO_0_SV "”1x (Inbyggd)” emulerar den inbyggda dithering med låg upplösning som används av originalhårdvaran för att jämna ut färgbandningsartefakter som syns vid inbyggd färgdjup. ”Internupplösning” skalar dithering-granulariteten till den konfigurerade interna upplösningen för renare resultat. Rekommenderas att inaktiveras vid körning med 32 bpp färgdjup. Observera! På Vulkan tvingar aktivering av detta nedskalning till naturligt färgdjup, medan inaktivering automatiskt aktiverar utdata med högre färgdjup."
#define OPTION_VAL_INTERNAL_RESOLUTION_SV "Intern upplösning"
#define BEETLE_OPT_SCALED_UV_OFFSET_LABEL_SV "Textur UV-förskjutning"
#define BEETLE_OPT_SCALED_UV_OFFSET_INFO_0_SV "Texturer för 3D-polygoner med en förskjutning för högre än 1x intern upplösning. Minskar textursekvenser men kan orsaka oavsiktliga grafiska problem."
#define BEETLE_OPT_FILTER_LABEL_SV "Texturfiltrering"
#define BEETLE_OPT_FILTER_INFO_0_SV "”Närmaste” emulerar originalhårdvaran. ”Bilinjär” och ”3-punkt” är utjämningsfilter som minskar pixelering genom att sudda ut bilden. ”SABR”, ’xBR’ och ”JINC2” är uppskalningsfilter som kan förbättra texturens trohet/skärpa på bekostnad av ökade prestandakrav. Stöds endast av hårdvarurenderare."
#define OPTION_VAL_NEAREST_SV "Närmsta"
#define OPTION_VAL_SABR_SV NULL
#define OPTION_VAL_XBR_SV NULL
#define OPTION_VAL_BILINEAR_SV "Bilinjär"
#define OPTION_VAL_3_POINT_SV "3-Punkt"
#define OPTION_VAL_JINC2_SV NULL
#define BEETLE_OPT_FILTER_EXCLUDE_SPRITE_LABEL_SV "Exkludera blockfigurer från filtrering"
#define BEETLE_OPT_FILTER_EXCLUDE_SPRITE_INFO_0_SV "Tillämpa inte texturfiltrering på blockfigurer. Förhindrar skarvar i olika spel med 2D-blockfigur-renderade bakgrunder. Använd tillsammans med adaptiv jämnhet eller en annan efterbehandlingsfilter för bästa effekt."
#define OPTION_VAL_OPAQUE_SV "Endast opaka"
#define OPTION_VAL_ALL_SV "Opaka och halvgenomskinlig"
#define BEETLE_OPT_FILTER_EXCLUDE_2D_POLYGON_LABEL_SV "Undanta 2D-polygoner från filtrering"
#define BEETLE_OPT_FILTER_EXCLUDE_2D_POLYGON_INFO_0_SV "Använd inte texturfiltrering på 2D-polygoner. 2D-polygoner upptäcks med en heuristik och det kan uppstå grafiska fel. Använd tillsammans med adaptiv jämnhet eller en annan efterbehandlingsfiltrering för bästa effekt."
#define BEETLE_OPT_ADAPTIVE_SMOOTHING_LABEL_SV "Adaptiv jämnhet"
#define BEETLE_OPT_ADAPTIVE_SMOOTHING_INFO_0_SV "Jämna ut 2D-konstverk och användargränssnittselement utan att sudda ut 3D-renderade objekt. Stöds endast av Vulkan-renderaren."
#define BEETLE_OPT_SUPER_SAMPLING_LABEL_SV "Supersampling (Nedsampling till nativ upplösning)"
#define BEETLE_OPT_SUPER_SAMPLING_INFO_0_SV "Minska uppskalat internt innehåll till nativ upplösning genom nedsampling. Kombinera detta med högre interna upplösningmultiplikatorer för att tillåta spel med antialiaserade 3D-objekt att visas vid nativ låg upplösning. Ger bäst resultat när det tillämpas på titlar som blandar 2D- och 3D-element (t.ex. 3D-figurer på förrenderade bakgrunder) och fungerar bra tillsammans med CRT-filter. Stöds endast av Vulkan-renderaren. Obs!: \"Dithering-mönster\" rekommenderas att vara inaktiverat när du aktiverar denna inställning."
#define BEETLE_OPT_MSAA_LABEL_SV "Multisampling för kantutjämning (MSAA)"
#define BEETLE_OPT_MSAA_INFO_0_SV "MSAA förbättrar utseendet på 3D-objekt. Stöds endast av Vulkan-renderaren."
#define OPTION_VAL_1X_SV NULL
#define BEETLE_OPT_MDEC_YUV_LABEL_SV "MDEC YUV Chroma-filter"
#define BEETLE_OPT_MDEC_YUV_INFO_0_SV "Förbättra kvaliteten på uppspelning av fullrörlig video (FMV) genom att minska \"makroblockerings\"-artefakter (fyrkanter/sneda kanter). Stöds endast av Vulkan-renderaren."
#define BEETLE_OPT_TRACK_TEXTURES_LABEL_SV "Spåra texturer"
#define BEETLE_OPT_TRACK_TEXTURES_INFO_0_SV "Förutsättning för texturdumpning och ersättning. Kommer troligen att krascha i de flesta spel."
#define BEETLE_OPT_DUMP_TEXTURES_LABEL_SV "Dumpa texturer"
#define BEETLE_OPT_DUMP_TEXTURES_INFO_0_SV "Dumpa använda texturer till <cd>-texture-dump/"
#define BEETLE_OPT_REPLACE_TEXTURES_LABEL_SV "Ersätt texturer"
#define BEETLE_OPT_REPLACE_TEXTURES_INFO_0_SV "Ersätt texturer med HD-versioner från <cd>-texture-replacements/"
#define BEETLE_OPT_PGXP_MODE_LABEL_SV "PGXP Driftläge"
#define BEETLE_OPT_PGXP_MODE_INFO_0_SV "Rendera 3D-objekt med subpixelprecision, vilket minimerar distorsion och flimmer hos 3D-objekt som visas på originalhårdvara på grund av användningen av fasta punktkoordinater. Läget ”Endast minne” har minimala kompatibilitetsproblem och rekommenderas för allmänt bruk. Läget ”Minne + CPU (Buggig)” kan minska flimmer ännu mer, men har höga prestandakrav och kan orsaka olika geometrifel."
#define OPTION_VAL_MEMORY_ONLY_SV "Endast minne"
#define OPTION_VAL_MEMORY_CPU_SV "Minne + CPU (Buggig)"
#define BEETLE_OPT_PGXP_2D_TOL_LABEL_SV NULL
#define BEETLE_OPT_PGXP_2D_TOL_INFO_0_SV "Dölj fler uppenbara fel i PGXP-operationer: värdet anger toleransen inom vilken PGXP-värden kommer att behållas i fall av geometrier utan korrekt djupinformation."
#define OPTION_VAL_0PX_SV NULL
#define OPTION_VAL_1PX_SV NULL
#define OPTION_VAL_2PX_SV NULL
#define OPTION_VAL_3PX_SV NULL
#define OPTION_VAL_4PX_SV NULL
#define OPTION_VAL_5PX_SV NULL
#define OPTION_VAL_6PX_SV NULL
#define OPTION_VAL_7PX_SV NULL
#define OPTION_VAL_8PX_SV NULL
#define BEETLE_OPT_PGXP_NCLIP_LABEL_SV NULL
#define BEETLE_OPT_PGXP_NCLIP_INFO_0_SV "Använd PGXP: s NCLIP-implementering. Förbättrar utseendet genom att minska hål i geometrier med PGXP-koordinater. Känd för att orsaka att vissa spel låser sig under olika omständigheter."
#define BEETLE_OPT_PGXP_VERTEX_LABEL_SV NULL
#define BEETLE_OPT_PGXP_VERTEX_INFO_0_SV "Cachar PGXP-förbättrade vertexpositioner för återanvändning vid polygonritningar. Kan potentiellt förbättra objektinriktningen och minska synliga sömmar vid rendering av texturer, men falska positiva resultat vid förfrågan till cacheminnet kan ge grafiska störningar. För närvarande rekommenderas att detta alternativ är avaktiverat. Detta alternativ tillämpas endast när PGXP Operation Mode är aktiverat. Stöds endast av hårdvarurenderingarna."
#define BEETLE_OPT_PGXP_TEXTURE_LABEL_SV "Korrekt texturering för PGXP-perspektiv"
#define BEETLE_OPT_PGXP_TEXTURE_INFO_0_SV NULL
#define BEETLE_OPT_DISPLAY_INTERNAL_FPS_LABEL_SV "Visa interna bilder per sekund (FPS)"
#define BEETLE_OPT_DISPLAY_INTERNAL_FPS_INFO_0_SV "Visa den interna bildrutefrekvensen (FPS) som det emulerade PlayStation-systemet renderar innehåll i. Obs!: Kräver att meddelanden på skärmen är aktiverade i libretro-frontenden."
#define BEETLE_OPT_DISPLAY_OSD_LABEL_SV "Visa OSD-meddelanden"
#define BEETLE_OPT_DISPLAY_OSD_INFO_0_SV "Visa OSD-meddelanden genererade av kärnan."
#define BEETLE_OPT_DISPLAY_VRAM_LABEL_SV "Visa fullständig VRAM (Felsökning)"
#define BEETLE_OPT_DISPLAY_VRAM_INFO_0_SV "Visualisera hela den emulerade konsolens VRAM. Stöds endast av OpenGL- och Vulkan-hårdvarurenderare. Obs!: Detta är för felsökningsändamål och bör normalt sett vara inaktiverat."
#define BEETLE_OPT_ANALOG_CALIBRATION_LABEL_SV "Självkalibrera analoga spakar"
#define BEETLE_OPT_ANALOG_CALIBRATION_INFO_0_SV "Aktivera dynamisk kalibrering av analoga ingångar när enheten är inställd på DualShock, Analog Controller, Analog Joystick eller neGcon. Maximala registrerade ingångsvärden övervakas i realtid och används för att skala analoga koordinater som skickas till emulatorn. Detta bör användas för spel som Mega Man Legends 2 som förväntar sig större värden än vad moderna kontroller tillhandahåller. För bästa resultat bör analoga spakar roteras i full utsträckning för att justera kalibreringsalgoritmen varje gång innehållet laddas."
#define BEETLE_OPT_ANALOG_TOGGLE_LABEL_SV "Växla analogt läge för DualShock"
#define BEETLE_OPT_ANALOG_TOGGLE_INFO_0_SV "När typen av inmatningsenhet är DualShock gör det här alternativet att den emulerade DualShock kan växlas mellan DIGITAL- och ANALOG-läge som den ursprungliga hårdvaran. Läget kan också växlas genom att använda den valda knappkombinationen."
#define OPTION_VAL_ENABLED_ANALOG_SV "Standard-analog"
#define BEETLE_OPT_ANALOG_TOGGLE_COMBO_LABEL_SV "DualShock-analogt läge för kombinationer"
#define BEETLE_OPT_ANALOG_TOGGLE_COMBO_INFO_0_SV "Knappkombinationen som används för att växla mellan DIGITAL- och ANALOG-läge för den emulerade DualShock."
#define OPTION_VAL_L1_L2_R1_R2_START_SELECT_SV NULL
#define OPTION_VAL_L1_R1_SELECT_SV NULL
#define OPTION_VAL_L1_R1_START_SV NULL
#define OPTION_VAL_L1_R1_L3_SV NULL
#define OPTION_VAL_L1_R1_R3_SV NULL
#define OPTION_VAL_L2_R2_SELECT_SV NULL
#define OPTION_VAL_L2_R2_START_SV NULL
#define OPTION_VAL_L2_R2_L3_SV NULL
#define OPTION_VAL_L2_R2_R3_SV NULL
#define OPTION_VAL_L3_R3_SV NULL
#define BEETLE_OPT_ANALOG_TOGGLE_HOLD_LABEL_SV "Hållningsfördröjning för DualShock analogt läges-kombo"
#define BEETLE_OPT_ANALOG_TOGGLE_HOLD_INFO_0_SV "Hållningstid för de analoga kombinationsknapparna."
#define OPTION_VAL_0_SV "0 sekunds fördröjning"
#define OPTION_VAL_1_SV "1 sekunds fördröjning"
#define OPTION_VAL_2_SV "2 sekunders fördröjning"
#define OPTION_VAL_3_SV "3 sekunders fördröjning"
#define OPTION_VAL_4_SV "4 sekunders fördröjning"
#define OPTION_VAL_5_SV "5 sekunders fördröjning"
#define BEETLE_OPT_ENABLE_MULTITAP_PORT1_LABEL_SV "Multitap på Port 1"
#define BEETLE_OPT_ENABLE_MULTITAP_PORT1_INFO_0_SV "Aktivera multitap-funktionalitet på port 1."
#define BEETLE_OPT_ENABLE_MULTITAP_PORT2_LABEL_SV "Multitap på Port 2"
#define BEETLE_OPT_ENABLE_MULTITAP_PORT2_INFO_0_SV "Aktivera multitap-funktionalitet på port 2."
#define BEETLE_OPT_GUN_INPUT_MODE_LABEL_SV "Inmatningsläge för pistol"
#define BEETLE_OPT_GUN_INPUT_MODE_INFO_0_SV "Använd musstyrd \"Ljuspistol\" eller \"Pekskärm\" när enhetstypen är inställd på \"Guncon/G-Con 45\" eller \"Justifier\"."
#define OPTION_VAL_LIGHTGUN_SV "Ljuspistol"
#define OPTION_VAL_TOUCHSCREEN_SV "Pekskärm"
#define BEETLE_OPT_GUN_CURSOR_LABEL_SV "Pistolmarkör"
#define BEETLE_OPT_GUN_CURSOR_INFO_0_SV "Rita en markör på skärmen när du använder enhetstyper för inmatning som \"Guncon/G-Con 45\" och \"Justifier\"."
#define OPTION_VAL_CROSS_SV "Kors"
#define OPTION_VAL_DOT_SV "Punkt"
#define BEETLE_OPT_CROSSHAIR_COLOR_P1_LABEL_SV "Färg för pistolhårkors: Port 1"
#define BEETLE_OPT_CROSSHAIR_COLOR_P1_INFO_0_SV "Färg på ljuspistolens hårkors för port 1."
#define OPTION_VAL_RED_SV "Röd"
#define OPTION_VAL_BLUE_SV "Blå"
#define OPTION_VAL_GREEN_SV "Grön"
#define OPTION_VAL_ORANGE_SV NULL
#define OPTION_VAL_YELLOW_SV "Gul"
#define OPTION_VAL_CYAN_SV NULL
#define OPTION_VAL_PINK_SV "Rosa"
#define OPTION_VAL_PURPLE_SV "Lila"
#define OPTION_VAL_BLACK_SV "Svart"
#define OPTION_VAL_WHITE_SV "Vit"
#define BEETLE_OPT_CROSSHAIR_COLOR_P2_LABEL_SV "Färg för pistolhårkors: Port 2"
#define BEETLE_OPT_CROSSHAIR_COLOR_P2_INFO_0_SV "Färg på ljuspistolens hårkors för port 2."
#define BEETLE_OPT_MOUSE_SENSITIVITY_LABEL_SV "Muskänslighet"
#define BEETLE_OPT_MOUSE_SENSITIVITY_INFO_0_SV "Responsivitet för inmatningsenheter av typen \"Mus\"."
#define OPTION_VAL_5_O42_SV NULL
#define OPTION_VAL_10_SV NULL
#define OPTION_VAL_15_SV NULL
#define OPTION_VAL_20_SV NULL
#define OPTION_VAL_25_SV NULL
#define OPTION_VAL_30_SV NULL
#define OPTION_VAL_35_SV NULL
#define OPTION_VAL_40_SV NULL
#define OPTION_VAL_45_SV NULL
#define OPTION_VAL_55_SV NULL
#define OPTION_VAL_65_SV NULL
#define OPTION_VAL_75_SV NULL
#define OPTION_VAL_85_SV NULL
#define OPTION_VAL_95_SV NULL
#define OPTION_VAL_100_O42_SV NULL
#define OPTION_VAL_105_SV NULL
#define OPTION_VAL_115_SV NULL
#define OPTION_VAL_125_SV NULL
#define OPTION_VAL_135_SV NULL
#define OPTION_VAL_145_SV NULL
#define OPTION_VAL_155_SV NULL
#define OPTION_VAL_165_SV NULL
#define OPTION_VAL_175_SV NULL
#define OPTION_VAL_185_SV NULL
#define OPTION_VAL_195_SV NULL
#define BEETLE_OPT_NEGCON_RESPONSE_LABEL_SV "Vridningsrespons för neGcon"
#define BEETLE_OPT_NEGCON_RESPONSE_INFO_0_SV "Vänster analogspaks responstyp när man simulerar ”vridningsåtgärden” hos emulerade ”neGcon”-inmatningsenheter. Analoga spakars förskjutning kan mappas till neGcon-rotationsvinkeln antingen linjärt, kvadratiskt eller kubiskt. ”Kvadratisk” möjliggör större precision än ”Linjär” när man gör små rörelser. ’Kubisk’ ökar precisionen för små rörelser ytterligare, men ”överdriver” större rörelser. Observera: ”Linjär” rekommenderas endast vid användning av bilspelsrattar. Konventionella kontroller implementerar analog inmatning på ett sätt som skiljer sig fundamentalt från neGcons ’vridningsmekanism’, så att linjär mappning överförstärker små rörelser och försämrar finstyrningen. I de flesta fall ger ”kvadratisk” den närmaste approximationen av verklig hårdvara."
#define OPTION_VAL_LINEAR_SV "Linjär"
#define OPTION_VAL_QUADRATIC_SV "Kvadratisk"
#define OPTION_VAL_CUBIC_SV "Kubisk"
#define BEETLE_OPT_NEGCON_DEADZONE_LABEL_SV "Dödzon för neGcon-vridning"
#define BEETLE_OPT_NEGCON_DEADZONE_INFO_0_SV "Vänster analogspaks dödzon vid simulering av ”vridningsåtgärden\" hos emulerade ”neGcon”-inmatningsenheter. Används för att eliminera kontrollerdrift. Observera: De flesta neGcon-kompatibla titlar har alternativ i spelet för att ställa in ett värde för dödzonen för ”vridning” (twist). För att undvika förlust av precision bör dödzonen i spelet *alltid* ställas in på noll. Eventuella justeringar bör *endast* göras via detta grundläggande alternativ. Detta är särskilt viktigt när ”Vridningsrespons för neGcon” är inställt på ’Kvadratisk’ eller ”Kubisk”."
#define OPTION_VAL_0_O44_SV NULL
#define BEETLE_OPT_USE_MEDNAFEN_MEMCARD0_METHOD_LABEL_SV "Metod för minneskort"
#define BEETLE_OPT_USE_MEDNAFEN_MEMCARD0_METHOD_INFO_0_SV "Format för sparat data som används för det vänstra minneskortfacket. ’Mednafen’ kan användas för kompatibilitet med den fristående versionen av Mednafen. När det används med Beetle PSX har Libretro (.srm) och Mednafen (.mcr) internt identiska format och kan konverteras mellan varandra genom att byta namn. Omstart krävs."
#define OPTION_VAL_LIBRETRO_SV NULL
#define OPTION_VAL_MEDNAFEN_SV NULL
#define BEETLE_OPT_ENABLE_MEMCARD1_LABEL_SV "Minneskortsplats 2"
#define BEETLE_OPT_ENABLE_MEMCARD1_INFO_0_SV "Emulera ett andra minneskort i det högra minneskortfacket. Observera: Vissa spel kräver att detta alternativ är inaktiverat för att fungera korrekt (t.ex. Codename Tenka). Observera: Minneskortfack 2 använder Mednafen (.mcr) som sparningsformat. Omstart krävs."
#define BEETLE_OPT_SHARED_MEMORY_CARDS_LABEL_SV "Delade minneskort"
#define BEETLE_OPT_SHARED_MEMORY_CARDS_INFO_0_SV "Använd delade minneskort istället för separata kort för varje innehåll. Observera: Om ”Metod för minneskort” är inställt på ”Libretro” påverkas endast minneskortplats 2. Omstart krävs."
#define BEETLE_OPT_MEMCARD_LEFT_INDEX_LABEL_SV "Index för minneskortsplats 1"
#define BEETLE_OPT_MEMCARD_LEFT_INDEX_INFO_0_SV "Minneskortet är för närvarande isatt i den vänstra kortplatsen. Det här alternativet fungerar bara om minneskortsmetoden är inställd på Mednafen."
#define OPTION_VAL_0_O48_SV NULL
#define BEETLE_OPT_MEMCARD_RIGHT_INDEX_LABEL_SV "Index för minneskortsplats 2"
#define BEETLE_OPT_MEMCARD_RIGHT_INDEX_INFO_0_SV "Minneskortet är för närvarande isatt i den högra kortplatsen. Det här alternativet fungerar bara om minneskortsplats 2 är aktiverad."
#define OPTION_VAL_1_O49_SV NULL
#define BEETLE_OPT_LINE_RENDER_LABEL_SV NULL
#define BEETLE_OPT_LINE_RENDER_INFO_0_SV "Vissa spel (t.ex. Doom, Hexen, Soul Blade, etc.) ritar horisontella linjer genom att sträcka ut trianglar med en bildpunktshöjd över hela skärmen, som rasteriseras som en rad bildpunkter på originalhårdvaran. Detta hack upptäcker dessa små trianglar och konverterar dem till fyrkanter efter behov, så att de kan visas korrekt på hårdvarurenderarna och vid uppskalade interna upplösningar. ”Aggressiv” krävs för vissa titlar (t.ex. Dark Forces, Duke Nukem), men kan annars orsaka grafiska fel. Lämna inställningen på ”Standard” om du är osäker."
#define OPTION_VAL_DEFAULT_SV "Standard"
#define OPTION_VAL_AGGRESSIVE_SV "Aggressiv"
#define BEETLE_OPT_WIDESCREEN_HACK_LABEL_SV "Hack för bredbildsläge"
#define BEETLE_OPT_WIDESCREEN_HACK_INFO_0_SV "Rendera 3D-innehåll anamorfiskt och mata ut den emulerade framebufferten i ett bredbildsformat. Ger bäst resultat med spel som är helt i 3D. 2D-element kommer att sträckas ut horisontellt och kan vara feljusterade."
#define BEETLE_OPT_WIDESCREEN_HACK_ASPECT_RATIO_LABEL_SV "Bildförhållande för bredbildshack"
#define BEETLE_OPT_WIDESCREEN_HACK_ASPECT_RATIO_INFO_0_SV "Bildförhållande som ska användas av Hack för bredbildsläge."
#define OPTION_VAL_16_9_SV NULL
#define OPTION_VAL_16_10_SV NULL
#define OPTION_VAL_18_9_SV NULL
#define OPTION_VAL_19_9_SV NULL
#define OPTION_VAL_20_9_SV NULL
#define OPTION_VAL_21_9_SV NULL
#define OPTION_VAL_32_9_SV NULL
#define BEETLE_OPT_CPU_DYNAREC_LABEL_SV NULL
#define BEETLE_OPT_CPU_DYNAREC_INFO_0_SV "Dynamisk omkompilering av CPU-instruktioner till inbyggda instruktioner. Mycket snabbare än tolken, men CPU-timingen är inte lika exakt och kan innehålla buggar."
#define OPTION_VAL_DISABLED_SV "Inaktiverad (Beetle Interpreter)"
#define OPTION_VAL_EXECUTE_SV "Max prestanda"
#define OPTION_VAL_EXECUTE_ONE_SV NULL
#define OPTION_VAL_RUN_INTERPRETER_SV "Lightrec-tolk"
#define BEETLE_OPT_DYNAREC_INVALIDATE_LABEL_SV NULL
#define BEETLE_OPT_DYNAREC_INVALIDATE_INFO_0_SV "Vissa spel kräver \"Fullständig\" invalidering, vissa kräver \"Endast DMA\"."
#define OPTION_VAL_FULL_SV NULL
#define OPTION_VAL_DMA_SV "Endast DMA (lite snabbare)"
#define BEETLE_OPT_DYNAREC_EVENTCYCLES_LABEL_SV NULL
#define BEETLE_OPT_DYNAREC_EVENTCYCLES_INFO_0_SV "Max cykler som körs av CPU innan en GPU/DMA/MDEC/Timer-uppdatering kontrolleras, högre tal blir snabbare, har mycket mindre inverkan på Beetle-tolken än dynarec."
#define OPTION_VAL_128_SV "128 (Standard)"
#define BEETLE_OPT_DYNAREC_SPU_SAMPLES_LABEL_SV "Dynarec SPU-samplingar"
#define BEETLE_OPT_DYNAREC_SPU_SAMPLES_INFO_0_SV "Max SPU-samplingar som ska köras innan en SPU-uppdatering kontrolleras, högre tal blir snabbare, men orsakar ljudstörningar i vissa spel med något annat än 1."
#define BEETLE_OPT_CORE_TIMING_FPS_LABEL_SV "Kärnrapporterad bildfrekvenstiming"
#define BEETLE_OPT_CORE_TIMING_FPS_INFO_0_SV "Automatisk växling gör det möjligt för kärnan att växla mellan progressiva och interlaced frekvenser."
#define OPTION_VAL_AUTO_TOGGLE_SV "Automatisk växling"
#define OPTION_VAL_FORCE_PROGRESSIVE_SV "Tvinga progressiv frekvens"
#define OPTION_VAL_FORCE_INTERLACED_SV "Tvinga frekvens med interlace"
#define BEETLE_OPT_DEINTERLACER_LABEL_SV "Deinterlace-metod"
#define BEETLE_OPT_DEINTERLACER_INFO_0_SV NULL
#define OPTION_VAL_WEAVE_SV NULL
#define OPTION_VAL_BOB_SV NULL
#define OPTION_VAL_BOB_OFFSET_SV NULL
#define OPTION_VAL_FASTMAD_SV NULL
#define BEETLE_OPT_FRAME_DUPING_LABEL_SV "Bildruteduplicering"
#define BEETLE_OPT_FRAME_DUPING_INFO_0_SV "Upprepa föregående bildruta om kärnan inte har något nytt att visa."
#define BEETLE_OPT_PAL_VIDEO_TIMING_OVERRIDE_LABEL_SV "Åsidosätt PAL-videotiming"
#define BEETLE_OPT_PAL_VIDEO_TIMING_OVERRIDE_INFO_0_SV "På grund av olika standarder verkar PAL-spel ofta långsammare jämfört med amerikanska eller japanska NTSC-utgåvor. Det här alternativet kan användas för att åsidosätta PAL-timings för att försöka köra dessa spel med NTSC-bildfrekvens."
#define BEETLE_OPT_ASPECT_RATIO_LABEL_SV "Kärnspecifikt bildformat"
#define BEETLE_OPT_ASPECT_RATIO_INFO_0_SV "Denna inställning ignoreras när alternativen Hack för bredbildsläge eller Visa fullständig VRAM är aktiverade."
#define OPTION_VAL_CORRECTED_SV "Korrigerad"
#define OPTION_VAL_UNCORRECTED_SV "Okorrigerad"
#define OPTION_VAL_4_3_SV "Tvinga 4:3"
#define OPTION_VAL_NTSC_SV "Tvinga NTSC"
#define BEETLE_OPT_CROP_OVERSCAN_LABEL_SV "Beskär Overscan"
#define BEETLE_OPT_CROP_OVERSCAN_INFO_0_SV "Fullständig inbyggd bild eller inga ramar."
#define OPTION_VAL_STATIC_SV "Horisontell"
#define OPTION_VAL_SMART_SV "Horisontell + Vertikal"
#define BEETLE_OPT_IMAGE_CROP_LABEL_SV "Ytterligare beskärning"
#define BEETLE_OPT_IMAGE_CROP_INFO_0_SV "Minska bredden på den beskärda bilden med det angivna antalet bildpunkter."
#define OPTION_VAL_9PX_SV NULL
#define OPTION_VAL_10PX_SV NULL
#define OPTION_VAL_11PX_SV NULL
#define OPTION_VAL_12PX_SV NULL
#define OPTION_VAL_13PX_SV NULL
#define OPTION_VAL_14PX_SV NULL
#define OPTION_VAL_15PX_SV NULL
#define OPTION_VAL_16PX_SV NULL
#define OPTION_VAL_17PX_SV NULL
#define OPTION_VAL_18PX_SV NULL
#define OPTION_VAL_19PX_SV NULL
#define OPTION_VAL_20PX_SV NULL
#define BEETLE_OPT_IMAGE_OFFSET_LABEL_SV "Förskjutning beskuren bild"
#define BEETLE_OPT_IMAGE_OFFSET_INFO_0_SV "Förskjut horisontellt åt höger (positivt) eller vänster (negativt) med det angivna antalet bildpunkter. Kan användas för att korrigera justeringsproblem. Stöds endast av programvarurenderaren."
#define OPTION_VAL_12PX_O64_SV NULL
#define OPTION_VAL_11PX_O64_SV NULL
#define OPTION_VAL_10PX_O64_SV NULL
#define OPTION_VAL_9PX_O64_SV NULL
#define OPTION_VAL_8PX_O64_SV NULL
#define OPTION_VAL_7PX_O64_SV NULL
#define OPTION_VAL_6PX_O64_SV NULL
#define OPTION_VAL_5PX_O64_SV NULL
#define OPTION_VAL_4PX_O64_SV NULL
#define OPTION_VAL_3PX_O64_SV NULL
#define OPTION_VAL_2PX_O64_SV NULL
#define OPTION_VAL_1PX_O64_SV NULL
#define OPTION_VAL_1PX_O64_0_SV NULL
#define OPTION_VAL_2PX_O64_0_SV NULL
#define OPTION_VAL_3PX_O64_0_SV NULL
#define OPTION_VAL_4PX_O64_0_SV NULL
#define OPTION_VAL_5PX_O64_0_SV NULL
#define OPTION_VAL_6PX_O64_0_SV NULL
#define OPTION_VAL_7PX_O64_0_SV NULL
#define OPTION_VAL_8PX_O64_0_SV NULL
#define OPTION_VAL_9PX_O64_0_SV NULL
#define OPTION_VAL_10PX_O64_0_SV NULL
#define OPTION_VAL_11PX_O64_0_SV NULL
#define OPTION_VAL_12PX_O64_0_SV NULL
#define BEETLE_OPT_IMAGE_OFFSET_CYCLES_LABEL_SV "Horisontell bildförskjutning (GPU-cykler)"
#define BEETLE_OPT_IMAGE_OFFSET_CYCLES_INFO_0_SV "Positiva värden flyttar bilden åt höger, negativa värden flyttar bilden åt vänster. Stöds endast av hårdvarurenderare."
#define BEETLE_OPT_INITIAL_SCANLINE_LABEL_SV "Initial skanningslinje - NTSC"
#define BEETLE_OPT_INITIAL_SCANLINE_INFO_0_SV "Första visade skanningslinjen vid uppspelning av NTSC-innehåll. Värden större än noll minskar höjden på utgångsbilderna genom att bildpunkter beskärs från den övre kanten. Kan användas för att motverka letterboxing."
#define BEETLE_OPT_LAST_SCANLINE_LABEL_SV "Sista skanningslinje - NTSC"
#define BEETLE_OPT_LAST_SCANLINE_INFO_0_SV "Senast visade skanningslinje vid uppspelning av NTSC-innehåll. Värden under 239 minskar höjden på utgångsbilderna genom att bildpunkter beskärs från den nedre kanten. Kan användas för att motverka letterboxing."
#define OPTION_VAL_239_SV "239 (Standard)"
#define BEETLE_OPT_INITIAL_SCANLINE_PAL_LABEL_SV "Initial skanningslinje - PAL"
#define BEETLE_OPT_INITIAL_SCANLINE_PAL_INFO_0_SV "Första visade skanningslinjen vid uppspelning av PAL-innehåll. Värden större än noll minskar höjden på utgångsbilderna genom att bildpunkter beskärs från den översta kanten. Kan användas för att motverka letterboxing."
#define BEETLE_OPT_LAST_SCANLINE_PAL_LABEL_SV "Senast visade skanningslinje - PAL"
#define BEETLE_OPT_LAST_SCANLINE_PAL_INFO_0_SV "Senast visade skanningslinje vid uppspelning av PAL-innehåll. Värden under 287 minskar höjden på utgångsbilderna genom att bildpunkter beskärs från den nedre kanten. Kan användas för att motverka letterboxing."
#define OPTION_VAL_287_SV "287 (Standard)"

struct retro_core_option_v2_category option_cats_sv[] = {
   {
      "system",
      CATEGORY_SYSTEM_LABEL_SV,
      CATEGORY_SYSTEM_INFO_0_SV
   },
   {
      "video",
      CATEGORY_VIDEO_LABEL_SV,
      CATEGORY_VIDEO_INFO_0_SV
   },
   {
      "pgxp",
      CATEGORY_PGXP_LABEL_SV,
      CATEGORY_PGXP_INFO_0_SV
   },
   {
      "osd",
      CATEGORY_OSD_LABEL_SV,
      CATEGORY_OSD_INFO_0_SV
   },
   {
      "input",
      CATEGORY_INPUT_LABEL_SV,
      CATEGORY_INPUT_INFO_0_SV
   },
   {
      "memcards",
      CATEGORY_MEMCARDS_LABEL_SV,
      CATEGORY_MEMCARDS_INFO_0_SV
   },
   {
      "hacks",
      CATEGORY_HACKS_LABEL_SV,
      CATEGORY_HACKS_INFO_0_SV
   },
   { NULL, NULL, NULL },
};
struct retro_core_option_v2_definition option_defs_sv[] = {
   {
      BEETLE_OPT(cpu_freq_scale),
      BEETLE_OPT_CPU_FREQ_SCALE_LABEL_SV,
      NULL,
      BEETLE_OPT_CPU_FREQ_SCALE_INFO_0_SV,
      NULL,
      "system",
      {
         { "50%",           OPTION_VAL_50_SV },
         { "60%",           OPTION_VAL_60_SV },
         { "70%",           OPTION_VAL_70_SV },
         { "80%",           OPTION_VAL_80_SV },
         { "90%",           OPTION_VAL_90_SV },
         { "100%",          OPTION_VAL_100_SV },
         { "110%",          OPTION_VAL_110_SV },
         { "120%",          OPTION_VAL_120_SV },
         { "130%",          OPTION_VAL_130_SV },
         { "140%",          OPTION_VAL_140_SV },
         { "150%",          OPTION_VAL_150_SV },
         { "160%",          OPTION_VAL_160_SV },
         { "170%",          OPTION_VAL_170_SV },
         { "180%",          OPTION_VAL_180_SV },
         { "190%",          OPTION_VAL_190_SV },
         { "200%",          OPTION_VAL_200_SV },
         { "210%",          OPTION_VAL_210_SV },
         { "220%",          OPTION_VAL_220_SV },
         { "230%",          OPTION_VAL_230_SV },
         { "240%",          OPTION_VAL_240_SV },
         { "250%",          OPTION_VAL_250_SV },
         { "260%",          OPTION_VAL_260_SV },
         { "270%",          OPTION_VAL_270_SV },
         { "280%",          OPTION_VAL_280_SV },
         { "290%",          OPTION_VAL_290_SV },
         { "300%",          OPTION_VAL_300_SV },
         { "310%",          OPTION_VAL_310_SV },
         { "320%",          OPTION_VAL_320_SV },
         { "330%",          OPTION_VAL_330_SV },
         { "340%",          OPTION_VAL_340_SV },
         { "350%",          OPTION_VAL_350_SV },
         { "360%",          OPTION_VAL_360_SV },
         { "370%",          OPTION_VAL_370_SV },
         { "380%",          OPTION_VAL_380_SV },
         { "390%",          OPTION_VAL_390_SV },
         { "400%",          OPTION_VAL_400_SV },
         { "410%",          OPTION_VAL_410_SV },
         { "420%",          OPTION_VAL_420_SV },
         { "430%",          OPTION_VAL_430_SV },
         { "440%",          OPTION_VAL_440_SV },
         { "450%",          OPTION_VAL_450_SV },
         { "460%",          OPTION_VAL_460_SV },
         { "470%",          OPTION_VAL_470_SV },
         { "480%",          OPTION_VAL_480_SV },
         { "490%",          OPTION_VAL_490_SV },
         { "500%",          OPTION_VAL_500_SV },
         { "510%",          OPTION_VAL_510_SV },
         { "520%",          OPTION_VAL_520_SV },
         { "530%",          OPTION_VAL_530_SV },
         { "540%",          OPTION_VAL_540_SV },
         { "550%",          OPTION_VAL_550_SV },
         { "560%",          OPTION_VAL_560_SV },
         { "570%",          OPTION_VAL_570_SV },
         { "580%",          OPTION_VAL_580_SV },
         { "590%",          OPTION_VAL_590_SV },
         { "600%",          OPTION_VAL_600_SV },
         { "610%",          OPTION_VAL_610_SV },
         { "620%",          OPTION_VAL_620_SV },
         { "630%",          OPTION_VAL_630_SV },
         { "640%",          OPTION_VAL_640_SV },
         { "650%",          OPTION_VAL_650_SV },
         { "660%",          OPTION_VAL_660_SV },
         { "670%",          OPTION_VAL_670_SV },
         { "680%",          OPTION_VAL_680_SV },
         { "690%",          OPTION_VAL_690_SV },
         { "700%",          OPTION_VAL_700_SV },
         { "710%",          OPTION_VAL_710_SV },
         { "720%",          OPTION_VAL_720_SV },
         { "730%",          OPTION_VAL_730_SV },
         { "740%",          OPTION_VAL_740_SV },
         { "750%",          OPTION_VAL_750_SV },
         { NULL, NULL },
      },
      "100%"
   },
   {
      BEETLE_OPT(gpu_overclock),
      BEETLE_OPT_GPU_OVERCLOCK_LABEL_SV,
      NULL,
      BEETLE_OPT_GPU_OVERCLOCK_INFO_0_SV,
      NULL,
      "system",
      {
         { "1x(native)", OPTION_VAL_1X_NATIVE_SV },
         { "2x",         OPTION_VAL_2X_SV },
         { "4x",         OPTION_VAL_4X_SV },
         { "8x",         OPTION_VAL_8X_SV },
         { "16x",        OPTION_VAL_16X_SV },
         { "32x",        OPTION_VAL_32X_SV },
         { NULL, NULL },
      },
      "1x(native)"
   },
   {
      BEETLE_OPT(gte_overclock),
      BEETLE_OPT_GTE_OVERCLOCK_LABEL_SV,
      NULL,
      BEETLE_OPT_GTE_OVERCLOCK_INFO_0_SV,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
#ifndef EMSCRIPTEN
   {
      BEETLE_OPT(cd_access_method),
      BEETLE_OPT_CD_ACCESS_METHOD_LABEL_SV,
      NULL,
      BEETLE_OPT_CD_ACCESS_METHOD_INFO_0_SV,
      NULL,
      "system",
      {
         { "sync",     OPTION_VAL_SYNC_SV },
         { "async",    OPTION_VAL_ASYNC_SV },
         { "precache", OPTION_VAL_PRECACHE_SV },
         { NULL, NULL },
      },
      "sync"
   },
#endif
   {
      BEETLE_OPT(cd_fastload),
      BEETLE_OPT_CD_FASTLOAD_LABEL_SV,
      NULL,
      BEETLE_OPT_CD_FASTLOAD_INFO_0_SV,
      NULL,
      "system",
      {
         { "2x(native)", OPTION_VAL_2X_NATIVE_SV },
         { "4x",          OPTION_VAL_4X_SV },
         { "6x",          OPTION_VAL_6X_SV },
         { "8x",          OPTION_VAL_8X_SV },
         { "10x",         OPTION_VAL_10X_SV },
         { "12x",         OPTION_VAL_12X_SV },
         { "14x",         OPTION_VAL_14X_SV },
         { NULL, NULL },
      },
      "2x(native)"
   },
   {
      BEETLE_OPT(skip_bios),
      BEETLE_OPT_SKIP_BIOS_LABEL_SV,
      NULL,
      BEETLE_OPT_SKIP_BIOS_INFO_0_SV,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(override_bios),
      BEETLE_OPT_OVERRIDE_BIOS_LABEL_SV,
      NULL,
      BEETLE_OPT_OVERRIDE_BIOS_INFO_0_SV,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "psxonpsp", OPTION_VAL_PSXONPSP_SV },
         { "ps1_rom",  OPTION_VAL_PS1_ROM_SV },
         { "openbios", OPTION_VAL_OPENBIOS_SV     },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(region),
      BEETLE_OPT_REGION_LABEL_SV,
      NULL,
      BEETLE_OPT_REGION_INFO_0_SV,
      NULL,
      "system",
      {
         { "auto",   OPTION_VAL_AUTO_SV },
         { "ntsc-j", OPTION_VAL_NTSC_J_SV },
         { "ntsc-u", OPTION_VAL_NTSC_U_SV },
         { "pal",    OPTION_VAL_PAL_SV },
         { NULL, NULL },
      },
      "auto"
   },
   {
      BEETLE_OPT(internal_resolution),
      BEETLE_OPT_INTERNAL_RESOLUTION_LABEL_SV,
      NULL,
      BEETLE_OPT_INTERNAL_RESOLUTION_INFO_0_SV,
      NULL,
      "video",
      {
         { "1x(native)", OPTION_VAL_1X_NATIVE_SV },
         { "2x",         OPTION_VAL_2X_SV },
         { "4x",         OPTION_VAL_4X_SV },
         { "8x",         OPTION_VAL_8X_SV },
         { "16x",        OPTION_VAL_16X_SV },
         { NULL, NULL },
      },
      "1x(native)"
   },
#if defined(HAVE_OPENGL) || defined(HAVE_OPENGLES) || defined(HAVE_VULKAN)
   {
      BEETLE_OPT(renderer),
      BEETLE_OPT_RENDERER_LABEL_SV,
      NULL,
      BEETLE_OPT_RENDERER_INFO_0_SV,
      NULL,
      "video",
      {
         { "hardware",    OPTION_VAL_HARDWARE_SV },
#if defined(HAVE_OPENGL) || defined(HAVE_OPENGLES)
         { "hardware_gl", OPTION_VAL_HARDWARE_GL_SV },
#endif
#if defined(HAVE_VULKAN)
         { "hardware_vk", OPTION_VAL_HARDWARE_VK_SV },
#endif
         { "software",    OPTION_VAL_SOFTWARE_SV },
         { NULL, NULL },
      },
      "hardware"
   },
   {
      BEETLE_OPT(renderer_software_fb),
      BEETLE_OPT_RENDERER_SOFTWARE_FB_LABEL_SV,
      NULL,
      BEETLE_OPT_RENDERER_SOFTWARE_FB_INFO_0_SV,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled"
   },
#endif
#if defined(HAVE_OPENGL) || defined(HAVE_OPENGLES)
   {
      BEETLE_OPT(depth),
      BEETLE_OPT_DEPTH_LABEL_SV,
      NULL,
      BEETLE_OPT_DEPTH_INFO_0_SV,
      NULL,
      "video",
      {
         { "16bpp(native)", OPTION_VAL_16BPP_NATIVE_SV },
         { "32bpp",         OPTION_VAL_32BPP_SV },
         { NULL, NULL },
      },
      "16bpp(native)"
   },
   // Sort of, it's more like 15-bit high color and 24-bit true color for visible output. The alpha channel is used for mask bit. Vulkan renderer uses ABGR1555_555 for 31 bits internal? FMVs are always 24-bit on all renderers like original hardware (BGR888, no alpha)
#endif
   {
      BEETLE_OPT(dither_mode),
      BEETLE_OPT_DITHER_MODE_LABEL_SV,
      NULL,
      BEETLE_OPT_DITHER_MODE_INFO_0_SV,
      NULL,
      "video",
      {
         { "1x(native)",          OPTION_VAL_1X_NATIVE_SV },
         { "internal resolution", OPTION_VAL_INTERNAL_RESOLUTION_SV },
         { "disabled",            NULL },
         { NULL, NULL },
      },
      "1x(native)"
   },
#ifdef HAVE_VULKAN
   {
      BEETLE_OPT(scaled_uv_offset),
      BEETLE_OPT_SCALED_UV_OFFSET_LABEL_SV,
      NULL,
      BEETLE_OPT_SCALED_UV_OFFSET_INFO_0_SV,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled"
   },
#endif
#if defined(HAVE_OPENGL) || defined(HAVE_OPENGLES) || defined(HAVE_VULKAN)
   {
      BEETLE_OPT(filter),
      BEETLE_OPT_FILTER_LABEL_SV,
      NULL,
      BEETLE_OPT_FILTER_INFO_0_SV,
      NULL,
      "video",
      {
         { "nearest",  OPTION_VAL_NEAREST_SV },
         { "SABR",     OPTION_VAL_SABR_SV },
         { "xBR",      OPTION_VAL_XBR_SV },
         { "bilinear", OPTION_VAL_BILINEAR_SV },
         { "3-point",  OPTION_VAL_3_POINT_SV },
         { "JINC2",    OPTION_VAL_JINC2_SV },
         { NULL, NULL },
      },
      "nearest"
   },
#ifdef HAVE_VULKAN
   {
      BEETLE_OPT(filter_exclude_sprite),
      BEETLE_OPT_FILTER_EXCLUDE_SPRITE_LABEL_SV,
      NULL,
      BEETLE_OPT_FILTER_EXCLUDE_SPRITE_INFO_0_SV,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "opaque", OPTION_VAL_OPAQUE_SV },
         { "all", OPTION_VAL_ALL_SV },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(filter_exclude_2d_polygon),
      BEETLE_OPT_FILTER_EXCLUDE_2D_POLYGON_LABEL_SV,
      NULL,
      BEETLE_OPT_FILTER_EXCLUDE_2D_POLYGON_INFO_0_SV,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "opaque", OPTION_VAL_OPAQUE_SV },
         { "all", OPTION_VAL_ALL_SV },
         { NULL, NULL },
      },
      "disabled"
   },
#endif
#endif
#ifdef HAVE_VULKAN
   {
      BEETLE_OPT(adaptive_smoothing),
      BEETLE_OPT_ADAPTIVE_SMOOTHING_LABEL_SV,
      NULL,
      BEETLE_OPT_ADAPTIVE_SMOOTHING_INFO_0_SV,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(super_sampling),
      BEETLE_OPT_SUPER_SAMPLING_LABEL_SV,
      NULL,
      BEETLE_OPT_SUPER_SAMPLING_INFO_0_SV,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(msaa),
      BEETLE_OPT_MSAA_LABEL_SV,
      NULL,
      BEETLE_OPT_MSAA_INFO_0_SV,
      NULL,
      "video",
      {
         { "1x",  OPTION_VAL_1X_SV },
         { "2x",  OPTION_VAL_2X_SV },
         { "4x",  OPTION_VAL_4X_SV },
         { "8x",  OPTION_VAL_8X_SV },
         { "16x", OPTION_VAL_16X_SV },
         { NULL, NULL },
      },
      "1x"
   },
   {
      BEETLE_OPT(mdec_yuv),
      BEETLE_OPT_MDEC_YUV_LABEL_SV,
      NULL,
      BEETLE_OPT_MDEC_YUV_INFO_0_SV,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(track_textures),
      BEETLE_OPT_TRACK_TEXTURES_LABEL_SV,
      NULL,
      BEETLE_OPT_TRACK_TEXTURES_INFO_0_SV,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
#ifdef TEXTURE_DUMPING_ENABLED
   {
      BEETLE_OPT(dump_textures),
      BEETLE_OPT_DUMP_TEXTURES_LABEL_SV,
      NULL,
      BEETLE_OPT_DUMP_TEXTURES_INFO_0_SV,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
#endif
   {
      BEETLE_OPT(replace_textures),
      BEETLE_OPT_REPLACE_TEXTURES_LABEL_SV,
      NULL,
      BEETLE_OPT_REPLACE_TEXTURES_INFO_0_SV,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
#endif
   {
      BEETLE_OPT(pgxp_mode),
      BEETLE_OPT_PGXP_MODE_LABEL_SV,
      NULL,
      BEETLE_OPT_PGXP_MODE_INFO_0_SV,
      NULL,
      "pgxp",
      {
         { "disabled",     NULL },
         { "memory only",  OPTION_VAL_MEMORY_ONLY_SV },
         { "memory + CPU", OPTION_VAL_MEMORY_CPU_SV },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(pgxp_2d_tol),
      BEETLE_OPT_PGXP_2D_TOL_LABEL_SV,
      NULL,
      BEETLE_OPT_PGXP_2D_TOL_INFO_0_SV,
      NULL,
      "pgxp",
      {
         { "disabled", NULL },
         { "0px", OPTION_VAL_0PX_SV },
         { "1px", OPTION_VAL_1PX_SV },
         { "2px", OPTION_VAL_2PX_SV },
         { "3px", OPTION_VAL_3PX_SV },
         { "4px", OPTION_VAL_4PX_SV },
         { "5px", OPTION_VAL_5PX_SV },
         { "6px", OPTION_VAL_6PX_SV },
         { "7px", OPTION_VAL_7PX_SV },
         { "8px", OPTION_VAL_8PX_SV },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(pgxp_nclip),
      BEETLE_OPT_PGXP_NCLIP_LABEL_SV,
      NULL,
      BEETLE_OPT_PGXP_NCLIP_INFO_0_SV,
      NULL,
      "pgxp",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
#if defined(HAVE_OPENGL) || defined(HAVE_OPENGLES) || defined(HAVE_VULKAN)
   {
      BEETLE_OPT(pgxp_vertex),
      BEETLE_OPT_PGXP_VERTEX_LABEL_SV,
      NULL,
      BEETLE_OPT_PGXP_VERTEX_INFO_0_SV,
      NULL,
      "pgxp",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
#endif
   {
      BEETLE_OPT(pgxp_texture),
      BEETLE_OPT_PGXP_TEXTURE_LABEL_SV,
      NULL,
      BEETLE_OPT_PGXP_TEXTURE_INFO_0_SV,
      NULL,
      "pgxp",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(display_internal_fps),
      BEETLE_OPT_DISPLAY_INTERNAL_FPS_LABEL_SV,
      NULL,
      BEETLE_OPT_DISPLAY_INTERNAL_FPS_INFO_0_SV,
      NULL,
      "osd",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(display_osd),
      BEETLE_OPT_DISPLAY_OSD_LABEL_SV,
      NULL,
      BEETLE_OPT_DISPLAY_OSD_INFO_0_SV,
      NULL,
      "osd",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled"
   },
#if defined(HAVE_OPENGL) || defined(HAVE_OPENGLES) || defined(HAVE_VULKAN)
   {
      BEETLE_OPT(display_vram),
      BEETLE_OPT_DISPLAY_VRAM_LABEL_SV,
      NULL,
      BEETLE_OPT_DISPLAY_VRAM_INFO_0_SV,
      NULL,
      "osd",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
#endif
   {
      BEETLE_OPT(analog_calibration),
      BEETLE_OPT_ANALOG_CALIBRATION_LABEL_SV,
      NULL,
      BEETLE_OPT_ANALOG_CALIBRATION_INFO_0_SV,
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(analog_toggle),
      BEETLE_OPT_ANALOG_TOGGLE_LABEL_SV,
      NULL,
      BEETLE_OPT_ANALOG_TOGGLE_INFO_0_SV,
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { "enabled-analog", OPTION_VAL_ENABLED_ANALOG_SV },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(analog_toggle_combo),
      BEETLE_OPT_ANALOG_TOGGLE_COMBO_LABEL_SV,
      NULL,
      BEETLE_OPT_ANALOG_TOGGLE_COMBO_INFO_0_SV,
      NULL,
      "input",
      {
         { "l1+l2+r1+r2+start+select", OPTION_VAL_L1_L2_R1_R2_START_SELECT_SV },
         { "l1+r1+select",             OPTION_VAL_L1_R1_SELECT_SV },
         { "l1+r1+start",              OPTION_VAL_L1_R1_START_SV },
         { "l1+r1+l3",                 OPTION_VAL_L1_R1_L3_SV },
         { "l1+r1+r3",                 OPTION_VAL_L1_R1_R3_SV },
         { "l2+r2+select",             OPTION_VAL_L2_R2_SELECT_SV },
         { "l2+r2+start",              OPTION_VAL_L2_R2_START_SV },
         { "l2+r2+l3",                 OPTION_VAL_L2_R2_L3_SV },
         { "l2+r2+r3",                 OPTION_VAL_L2_R2_R3_SV },
         { "l3+r3",                    OPTION_VAL_L3_R3_SV },
         { NULL, NULL },
      },
      "l1+r1+select"
   },
   {
      BEETLE_OPT(analog_toggle_hold),
      BEETLE_OPT_ANALOG_TOGGLE_HOLD_LABEL_SV,
      NULL,
      BEETLE_OPT_ANALOG_TOGGLE_HOLD_INFO_0_SV,
      NULL,
      "input",
      {
         { "0", OPTION_VAL_0_SV },
         { "1", OPTION_VAL_1_SV },
         { "2", OPTION_VAL_2_SV },
         { "3", OPTION_VAL_3_SV },
         { "4", OPTION_VAL_4_SV },
         { "5", OPTION_VAL_5_SV },
         { NULL, NULL },
      },
      "1"
   },
   {
      BEETLE_OPT(enable_multitap_port1),
      BEETLE_OPT_ENABLE_MULTITAP_PORT1_LABEL_SV,
      NULL,
      BEETLE_OPT_ENABLE_MULTITAP_PORT1_INFO_0_SV,
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(enable_multitap_port2),
      BEETLE_OPT_ENABLE_MULTITAP_PORT2_LABEL_SV,
      NULL,
      BEETLE_OPT_ENABLE_MULTITAP_PORT2_INFO_0_SV,
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(gun_input_mode),
      BEETLE_OPT_GUN_INPUT_MODE_LABEL_SV,
      NULL,
      BEETLE_OPT_GUN_INPUT_MODE_INFO_0_SV,
      NULL,
      "input",
      {
         { "lightgun",    OPTION_VAL_LIGHTGUN_SV },
         { "touchscreen", OPTION_VAL_TOUCHSCREEN_SV },
         { NULL, NULL },
      },
      "lightgun"
   },
   // Shouldn't the gun_input_mode just be Mouse vs. Touchscreen?
   {
      BEETLE_OPT(gun_cursor),
      BEETLE_OPT_GUN_CURSOR_LABEL_SV,
      NULL,
      BEETLE_OPT_GUN_CURSOR_INFO_0_SV,
      NULL,
      "input",
      {
         { "off",   "disabled" },
         { "cross", OPTION_VAL_CROSS_SV },
         { "dot",   OPTION_VAL_DOT_SV },
         { NULL, NULL },
      },
      "cross"
   },
   {
      BEETLE_OPT(crosshair_color_p1),
      BEETLE_OPT_CROSSHAIR_COLOR_P1_LABEL_SV,
      NULL,
      BEETLE_OPT_CROSSHAIR_COLOR_P1_INFO_0_SV,
      NULL,
      "input",
      {
         { "red",    OPTION_VAL_RED_SV },
         { "blue",   OPTION_VAL_BLUE_SV },
         { "green",  OPTION_VAL_GREEN_SV },
         { "orange", OPTION_VAL_ORANGE_SV },
         { "yellow", OPTION_VAL_YELLOW_SV },
         { "cyan",   OPTION_VAL_CYAN_SV },
         { "pink",   OPTION_VAL_PINK_SV },
         { "purple", OPTION_VAL_PURPLE_SV },
         { "black",  OPTION_VAL_BLACK_SV },
         { "white",  OPTION_VAL_WHITE_SV },
         { NULL, NULL },
      },
      "red"
   },
   {
      BEETLE_OPT(crosshair_color_p2),
      BEETLE_OPT_CROSSHAIR_COLOR_P2_LABEL_SV,
      NULL,
      BEETLE_OPT_CROSSHAIR_COLOR_P2_INFO_0_SV,
      NULL,
      "input",
      {
         { "blue",   OPTION_VAL_BLUE_SV },
         { "red",    OPTION_VAL_RED_SV },
         { "green",  OPTION_VAL_GREEN_SV },
         { "orange", OPTION_VAL_ORANGE_SV },
         { "yellow", OPTION_VAL_YELLOW_SV },
         { "cyan",   OPTION_VAL_CYAN_SV },
         { "pink",   OPTION_VAL_PINK_SV },
         { "purple", OPTION_VAL_PURPLE_SV },
         { "black",  OPTION_VAL_BLACK_SV },
         { "white",  OPTION_VAL_WHITE_SV },
         { NULL, NULL },
      },
      "blue"
   },
   {
      BEETLE_OPT(mouse_sensitivity),
      BEETLE_OPT_MOUSE_SENSITIVITY_LABEL_SV,
      NULL,
      BEETLE_OPT_MOUSE_SENSITIVITY_INFO_0_SV,
      NULL,
      "input",
      {
         { "5%",   OPTION_VAL_5_O42_SV },
         { "10%",  OPTION_VAL_10_SV },
         { "15%",  OPTION_VAL_15_SV },
         { "20%",  OPTION_VAL_20_SV },
         { "25%",  OPTION_VAL_25_SV },
         { "30%",  OPTION_VAL_30_SV },
         { "35%",  OPTION_VAL_35_SV },
         { "40%",  OPTION_VAL_40_SV },
         { "45%",  OPTION_VAL_45_SV },
         { "50%",  OPTION_VAL_50_SV },
         { "55%",  OPTION_VAL_55_SV },
         { "60%",  OPTION_VAL_60_SV },
         { "65%",  OPTION_VAL_65_SV },
         { "70%",  OPTION_VAL_70_SV },
         { "75%",  OPTION_VAL_75_SV },
         { "80%",  OPTION_VAL_80_SV },
         { "85%",  OPTION_VAL_85_SV },
         { "90%",  OPTION_VAL_90_SV },
         { "95%",  OPTION_VAL_95_SV },
         { "100%", OPTION_VAL_100_O42_SV },
         { "105%", OPTION_VAL_105_SV },
         { "110%", OPTION_VAL_110_SV },
         { "115%", OPTION_VAL_115_SV },
         { "120%", OPTION_VAL_120_SV },
         { "125%", OPTION_VAL_125_SV },
         { "130%", OPTION_VAL_130_SV },
         { "135%", OPTION_VAL_135_SV },
         { "140%", OPTION_VAL_140_SV },
         { "145%", OPTION_VAL_145_SV },
         { "150%", OPTION_VAL_150_SV },
         { "155%", OPTION_VAL_155_SV },
         { "160%", OPTION_VAL_160_SV },
         { "165%", OPTION_VAL_165_SV },
         { "170%", OPTION_VAL_170_SV },
         { "175%", OPTION_VAL_175_SV },
         { "180%", OPTION_VAL_180_SV },
         { "185%", OPTION_VAL_185_SV },
         { "190%", OPTION_VAL_190_SV },
         { "195%", OPTION_VAL_195_SV },
         { "200%", OPTION_VAL_200_SV },
         { NULL, NULL },
      },
      "100%"
   },
   {
      BEETLE_OPT(negcon_response),
      BEETLE_OPT_NEGCON_RESPONSE_LABEL_SV,
      NULL,
      BEETLE_OPT_NEGCON_RESPONSE_INFO_0_SV,
      NULL,
      "input",
      {
         { "linear",    OPTION_VAL_LINEAR_SV },
         { "quadratic", OPTION_VAL_QUADRATIC_SV },
         { "cubic",     OPTION_VAL_CUBIC_SV },
         { NULL, NULL },
      },
      "linear"
   },
   {
      BEETLE_OPT(negcon_deadzone),
      BEETLE_OPT_NEGCON_DEADZONE_LABEL_SV,
      NULL,
      BEETLE_OPT_NEGCON_DEADZONE_INFO_0_SV,
      NULL,
      "input",
      {
         { "0%",  OPTION_VAL_0_O44_SV },
         { "5%",  OPTION_VAL_5_O42_SV },
         { "10%", OPTION_VAL_10_SV },
         { "15%", OPTION_VAL_15_SV },
         { "20%", OPTION_VAL_20_SV },
         { "25%", OPTION_VAL_25_SV },
         { "30%", OPTION_VAL_30_SV },
         { NULL, NULL },
      },
      "0%"
   },
   {
      BEETLE_OPT(use_mednafen_memcard0_method),
      BEETLE_OPT_USE_MEDNAFEN_MEMCARD0_METHOD_LABEL_SV,
      NULL,
      BEETLE_OPT_USE_MEDNAFEN_MEMCARD0_METHOD_INFO_0_SV,
      NULL,
      "memcards",
      {
         { "libretro", OPTION_VAL_LIBRETRO_SV },
         { "mednafen", OPTION_VAL_MEDNAFEN_SV },
         { NULL, NULL },
      },
      "libretro"
   },
   {
      BEETLE_OPT(enable_memcard1),
      BEETLE_OPT_ENABLE_MEMCARD1_LABEL_SV,
      NULL,
      BEETLE_OPT_ENABLE_MEMCARD1_INFO_0_SV,
      NULL,
      "memcards",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(shared_memory_cards),
      BEETLE_OPT_SHARED_MEMORY_CARDS_LABEL_SV,
      NULL,
      BEETLE_OPT_SHARED_MEMORY_CARDS_INFO_0_SV,
      NULL,
      "memcards",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(memcard_left_index),
      BEETLE_OPT_MEMCARD_LEFT_INDEX_LABEL_SV,
      NULL,
      BEETLE_OPT_MEMCARD_LEFT_INDEX_INFO_0_SV,
      NULL,
      "memcards",
      {
         { "0",  OPTION_VAL_0_O48_SV },
         { "1",  NULL },
         { "2",  NULL },
         { "3",  NULL },
         { "4",  NULL },
         { "5",  NULL },
         { "6",  NULL },
         { "7",  NULL },
         { "8",  NULL },
         { "9",  NULL },
         { "10", NULL },
         { "11", NULL },
         { "12", NULL },
         { "13", NULL },
         { "14", NULL },
         { "15", NULL },
         { "16", NULL },
         { "17", NULL },
         { "18", NULL },
         { "19", NULL },
         { "20", NULL },
         { "21", NULL },
         { "22", NULL },
         { "23", NULL },
         { "24", NULL },
         { "25", NULL },
         { "26", NULL },
         { "27", NULL },
         { "28", NULL },
         { "29", NULL },
         { "30", NULL },
         { "31", NULL },
         { "32", NULL },
         { "33", NULL },
         { "34", NULL },
         { "35", NULL },
         { "36", NULL },
         { "37", NULL },
         { "38", NULL },
         { "39", NULL },
         { "40", NULL },
         { "41", NULL },
         { "42", NULL },
         { "43", NULL },
         { "44", NULL },
         { "45", NULL },
         { "46", NULL },
         { "47", NULL },
         { "48", NULL },
         { "49", NULL },
         { "50", NULL },
         { "51", NULL },
         { "52", NULL },
         { "53", NULL },
         { "54", NULL },
         { "55", NULL },
         { "56", NULL },
         { "57", NULL },
         { "58", NULL },
         { "59", NULL },
         { "60", NULL },
         { "61", NULL },
         { "62", NULL },
         { "63", NULL },
         { NULL, NULL },
      },
      "0"
   },
   {
      BEETLE_OPT(memcard_right_index),
      BEETLE_OPT_MEMCARD_RIGHT_INDEX_LABEL_SV,
      NULL,
      BEETLE_OPT_MEMCARD_RIGHT_INDEX_INFO_0_SV,
      NULL,
      "memcards",
      {
         { "0",  NULL },
         { "1",  OPTION_VAL_1_O49_SV },
         { "2",  NULL },
         { "3",  NULL },
         { "4",  NULL },
         { "5",  NULL },
         { "6",  NULL },
         { "7",  NULL },
         { "8",  NULL },
         { "9",  NULL },
         { "10", NULL },
         { "11", NULL },
         { "12", NULL },
         { "13", NULL },
         { "14", NULL },
         { "15", NULL },
         { "16", NULL },
         { "17", NULL },
         { "18", NULL },
         { "19", NULL },
         { "20", NULL },
         { "21", NULL },
         { "22", NULL },
         { "23", NULL },
         { "24", NULL },
         { "25", NULL },
         { "26", NULL },
         { "27", NULL },
         { "28", NULL },
         { "29", NULL },
         { "30", NULL },
         { "31", NULL },
         { "32", NULL },
         { "33", NULL },
         { "34", NULL },
         { "35", NULL },
         { "36", NULL },
         { "37", NULL },
         { "38", NULL },
         { "39", NULL },
         { "40", NULL },
         { "41", NULL },
         { "42", NULL },
         { "43", NULL },
         { "44", NULL },
         { "45", NULL },
         { "46", NULL },
         { "47", NULL },
         { "48", NULL },
         { "49", NULL },
         { "50", NULL },
         { "51", NULL },
         { "52", NULL },
         { "53", NULL },
         { "54", NULL },
         { "55", NULL },
         { "56", NULL },
         { "57", NULL },
         { "58", NULL },
         { "59", NULL },
         { "60", NULL },
         { "61", NULL },
         { "62", NULL },
         { "63", NULL },
         { NULL, NULL },
      },
      "1"
   },
   {
      BEETLE_OPT(line_render),
      BEETLE_OPT_LINE_RENDER_LABEL_SV,
      NULL,
      BEETLE_OPT_LINE_RENDER_INFO_0_SV,
      NULL,
      "hacks",
      {
         { "default",    OPTION_VAL_DEFAULT_SV },
         { "aggressive", OPTION_VAL_AGGRESSIVE_SV },
         { "disabled",   NULL },
         { NULL, NULL },
      },
      "default"
   },
   {
      BEETLE_OPT(widescreen_hack),
      BEETLE_OPT_WIDESCREEN_HACK_LABEL_SV,
      NULL,
      BEETLE_OPT_WIDESCREEN_HACK_INFO_0_SV,
      NULL,
      "hacks",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(widescreen_hack_aspect_ratio),
      BEETLE_OPT_WIDESCREEN_HACK_ASPECT_RATIO_LABEL_SV,
      NULL,
      BEETLE_OPT_WIDESCREEN_HACK_ASPECT_RATIO_INFO_0_SV,
      NULL,
      "hacks",
      {
         { "16:9",  OPTION_VAL_16_9_SV },
         { "16:10", OPTION_VAL_16_10_SV },
         { "18:9",  OPTION_VAL_18_9_SV },
         { "19:9",  OPTION_VAL_19_9_SV },
         { "20:9",  OPTION_VAL_20_9_SV },
         { "21:9",  OPTION_VAL_21_9_SV }, // 64:27
         { "32:9",  OPTION_VAL_32_9_SV },
         { NULL,    NULL },
      },
      "16:9"
   },
#if defined(HAVE_LIGHTREC)
   {
      BEETLE_OPT(cpu_dynarec),
      BEETLE_OPT_CPU_DYNAREC_LABEL_SV,
      NULL,
      BEETLE_OPT_CPU_DYNAREC_INFO_0_SV,
      NULL,
      "hacks",
      {
         { "disabled",        OPTION_VAL_DISABLED_SV },
         { "execute",         OPTION_VAL_EXECUTE_SV },
         { "execute_one",     OPTION_VAL_EXECUTE_ONE_SV },
         { "run_interpreter", OPTION_VAL_RUN_INTERPRETER_SV },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(dynarec_invalidate),
      BEETLE_OPT_DYNAREC_INVALIDATE_LABEL_SV,
      NULL,
      BEETLE_OPT_DYNAREC_INVALIDATE_INFO_0_SV,
      NULL,
      "hacks",
      {
         { "full", OPTION_VAL_FULL_SV },
         { "dma",  OPTION_VAL_DMA_SV },
         { NULL, NULL },
      },
      "full"
   },
   {
      BEETLE_OPT(dynarec_eventcycles),
      BEETLE_OPT_DYNAREC_EVENTCYCLES_LABEL_SV,
      NULL,
      BEETLE_OPT_DYNAREC_EVENTCYCLES_INFO_0_SV,
      NULL,
      "hacks",
      {
         { "128",  OPTION_VAL_128_SV },
         { "256",  NULL },
         { "384",  NULL },
         { "512",  NULL },
         { "640",  NULL },
         { "768",  NULL },
         { "896",  NULL },
         { "1024", NULL },
         { "1152", NULL },
         { "1280", NULL },
         { "1408", NULL },
         { "1536", NULL },
         { "1664", NULL },
         { "1792", NULL },
         { "1920", NULL },
         { "2048", NULL },
         { NULL, NULL },
      },
      "128"
   },
   {
      BEETLE_OPT(dynarec_spu_samples),
      BEETLE_OPT_DYNAREC_SPU_SAMPLES_LABEL_SV,
      NULL,
      BEETLE_OPT_DYNAREC_SPU_SAMPLES_INFO_0_SV,
      NULL,
      "hacks",
      {
         { "1",  OPTION_VAL_1_O49_SV },
         { "4",  NULL },
         { "16", NULL },
         { NULL, NULL },
      },
      "1"
   },
#endif
   {
      BEETLE_OPT(core_timing_fps),
      BEETLE_OPT_CORE_TIMING_FPS_LABEL_SV,
      NULL,
      BEETLE_OPT_CORE_TIMING_FPS_INFO_0_SV,
      NULL,
      "video",
      {
         { "auto_toggle",       OPTION_VAL_AUTO_TOGGLE_SV },
         { "force_progressive", OPTION_VAL_FORCE_PROGRESSIVE_SV },
         { "force_interlaced",  OPTION_VAL_FORCE_INTERLACED_SV },
      },
      "auto_toggle"
   },
   {
      BEETLE_OPT(deinterlacer),
      BEETLE_OPT_DEINTERLACER_LABEL_SV,
      NULL,
      BEETLE_OPT_DEINTERLACER_INFO_0_SV,
      NULL,
      "video",
      {
         { "weave",      OPTION_VAL_WEAVE_SV },
         { "bob",        OPTION_VAL_BOB_SV },
         { "bob_offset", OPTION_VAL_BOB_OFFSET_SV },
         { "fastmad",    OPTION_VAL_FASTMAD_SV },
         { "off",        "Off" },
         { NULL, NULL },
      },
      "weave"
   },
   {
      BEETLE_OPT(frame_duping),
      BEETLE_OPT_FRAME_DUPING_LABEL_SV,
      NULL,
      BEETLE_OPT_FRAME_DUPING_INFO_0_SV,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(pal_video_timing_override),
      BEETLE_OPT_PAL_VIDEO_TIMING_OVERRIDE_LABEL_SV,
      NULL,
      BEETLE_OPT_PAL_VIDEO_TIMING_OVERRIDE_INFO_0_SV,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(aspect_ratio),
      BEETLE_OPT_ASPECT_RATIO_LABEL_SV,
      NULL,
      BEETLE_OPT_ASPECT_RATIO_INFO_0_SV,
      NULL,
      "video",
      {
         { "corrected", OPTION_VAL_CORRECTED_SV },
         { "uncorrected", OPTION_VAL_UNCORRECTED_SV },
         { "4:3",  OPTION_VAL_4_3_SV },
         { "ntsc", OPTION_VAL_NTSC_SV },
      },
      "corrected"
   },
   {
      BEETLE_OPT(crop_overscan),
      BEETLE_OPT_CROP_OVERSCAN_LABEL_SV,
      NULL,
      BEETLE_OPT_CROP_OVERSCAN_INFO_0_SV,
      NULL,
      "video",
      {
         { "disabled", "Disabled" },
         { "static",   OPTION_VAL_STATIC_SV },
         { "smart",    OPTION_VAL_SMART_SV },
         { NULL, NULL },
      },
      "smart"
   },
   {
      BEETLE_OPT(image_crop),
      BEETLE_OPT_IMAGE_CROP_LABEL_SV,
      NULL,
      BEETLE_OPT_IMAGE_CROP_INFO_0_SV,
      NULL,
      "video",
      {
         { "disabled", "0" },
         { "1px",      OPTION_VAL_1PX_SV },
         { "2px",      OPTION_VAL_2PX_SV },
         { "3px",      OPTION_VAL_3PX_SV },
         { "4px",      OPTION_VAL_4PX_SV },
         { "5px",      OPTION_VAL_5PX_SV },
         { "6px",      OPTION_VAL_6PX_SV },
         { "7px",      OPTION_VAL_7PX_SV },
         { "8px",      OPTION_VAL_8PX_SV },
         { "9px",      OPTION_VAL_9PX_SV },
         { "10px",     OPTION_VAL_10PX_SV },
         { "11px",     OPTION_VAL_11PX_SV },
         { "12px",     OPTION_VAL_12PX_SV },
         { "13px",     OPTION_VAL_13PX_SV },
         { "14px",     OPTION_VAL_14PX_SV },
         { "15px",     OPTION_VAL_15PX_SV },
         { "16px",     OPTION_VAL_16PX_SV },
         { "17px",     OPTION_VAL_17PX_SV },
         { "18px",     OPTION_VAL_18PX_SV },
         { "19px",     OPTION_VAL_19PX_SV },
         { "20px",     OPTION_VAL_20PX_SV },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(image_offset),
      BEETLE_OPT_IMAGE_OFFSET_LABEL_SV,
      NULL,
      BEETLE_OPT_IMAGE_OFFSET_INFO_0_SV,
      NULL,
      "video",
      {
         { "-12px",    OPTION_VAL_12PX_O64_SV },
         { "-11px",    OPTION_VAL_11PX_O64_SV },
         { "-10px",    OPTION_VAL_10PX_O64_SV },
         { "-9px",     OPTION_VAL_9PX_O64_SV },
         { "-8px",     OPTION_VAL_8PX_O64_SV },
         { "-7px",     OPTION_VAL_7PX_O64_SV },
         { "-6px",     OPTION_VAL_6PX_O64_SV },
         { "-5px",     OPTION_VAL_5PX_O64_SV },
         { "-4px",     OPTION_VAL_4PX_O64_SV },
         { "-3px",     OPTION_VAL_3PX_O64_SV },
         { "-2px",     OPTION_VAL_2PX_O64_SV },
         { "-1px",     OPTION_VAL_1PX_O64_SV },
         { "disabled", "0" },
         { "+1px",     OPTION_VAL_1PX_O64_0_SV },
         { "+2px",     OPTION_VAL_2PX_O64_0_SV },
         { "+3px",     OPTION_VAL_3PX_O64_0_SV },
         { "+4px",     OPTION_VAL_4PX_O64_0_SV },
         { "+5px",     OPTION_VAL_5PX_O64_0_SV },
         { "+6px",     OPTION_VAL_6PX_O64_0_SV },
         { "+7px",     OPTION_VAL_7PX_O64_0_SV },
         { "+8px",     OPTION_VAL_8PX_O64_0_SV },
         { "+9px",     OPTION_VAL_9PX_O64_0_SV },
         { "+10px",    OPTION_VAL_10PX_O64_0_SV },
         { "+11px",    OPTION_VAL_11PX_O64_0_SV },
         { "+12px",    OPTION_VAL_12PX_O64_0_SV },
         { NULL, NULL },
      },
      "disabled"
   },
#if defined(HAVE_OPENGL) || defined(HAVE_OPENGLES) || defined(HAVE_VULKAN)
   {
      BEETLE_OPT(image_offset_cycles),
      BEETLE_OPT_IMAGE_OFFSET_CYCLES_LABEL_SV,
      NULL,
      BEETLE_OPT_IMAGE_OFFSET_CYCLES_INFO_0_SV,
      NULL,
      "video",
      {
         { "-40",      NULL },
         { "-39",      NULL },
         { "-38",      NULL },
         { "-37",      NULL },
         { "-36",      NULL },
         { "-35",      NULL },
         { "-34",      NULL },
         { "-33",      NULL },
         { "-32",      NULL },
         { "-31",      NULL },
         { "-30",      NULL },
         { "-29",      NULL },
         { "-28",      NULL },
         { "-27",      NULL },
         { "-26",      NULL },
         { "-25",      NULL },
         { "-24",      NULL },
         { "-23",      NULL },
         { "-22",      NULL },
         { "-21",      NULL },
         { "-20",      NULL },
         { "-19",      NULL },
         { "-18",      NULL },
         { "-17",      NULL },
         { "-16",      NULL },
         { "-15",      NULL },
         { "-14",      NULL },
         { "-13",      NULL },
         { "-12",      NULL },
         { "-11",      NULL },
         { "-10",      NULL },
         { "-9",       NULL },
         { "-8",       NULL },
         { "-7",       NULL },
         { "-6",       NULL },
         { "-5",       NULL },
         { "-4",       NULL },
         { "-3",       NULL },
         { "-2",       NULL },
         { "-1",       NULL },
         { "0",        "0" },
         { "+1",       NULL },
         { "+2",       NULL },
         { "+3",       NULL },
         { "+4",       NULL },
         { "+5",       NULL },
         { "+6",       NULL },
         { "+7",       NULL },
         { "+8",       NULL },
         { "+9",       NULL },
         { "+10",      NULL },
         { "+11",      NULL },
         { "+12",      NULL },
         { "+13",      NULL },
         { "+14",      NULL },
         { "+15",      NULL },
         { "+16",      NULL },
         { "+17",      NULL },
         { "+18",      NULL },
         { "+19",      NULL },
         { "+20",      NULL },
         { "+21",      NULL },
         { "+22",      NULL },
         { "+23",      NULL },
         { "+24",      NULL },
         { "+25",      NULL },
         { "+26",      NULL },
         { "+27",      NULL },
         { "+28",      NULL },
         { "+29",      NULL },
         { "+30",      NULL },
         { "+31",      NULL },
         { "+32",      NULL },
         { "+33",      NULL },
         { "+34",      NULL },
         { "+35",      NULL },
         { "+36",      NULL },
         { "+37",      NULL },
         { "+38",      NULL },
         { "+39",      NULL },
         { "+40",      NULL },
         { NULL, NULL},
      },
      "0"
   },
#endif
   {
      BEETLE_OPT(initial_scanline),
      BEETLE_OPT_INITIAL_SCANLINE_LABEL_SV,
      NULL,
      BEETLE_OPT_INITIAL_SCANLINE_INFO_0_SV,
      NULL,
      "video",
      {
         { "0",  OPTION_VAL_0_O48_SV },
         { "1",  NULL },
         { "2",  NULL },
         { "3",  NULL },
         { "4",  NULL },
         { "5",  NULL },
         { "6",  NULL },
         { "7",  NULL },
         { "8",  NULL },
         { "9",  NULL },
         { "10", NULL },
         { "11", NULL },
         { "12", NULL },
         { "13", NULL },
         { "14", NULL },
         { "15", NULL },
         { "16", NULL },
         { "17", NULL },
         { "18", NULL },
         { "19", NULL },
         { "20", NULL },
         { "21", NULL },
         { "22", NULL },
         { "23", NULL },
         { "24", NULL },
         { "25", NULL },
         { "26", NULL },
         { "27", NULL },
         { "28", NULL },
         { "29", NULL },
         { "30", NULL },
         { "31", NULL },
         { "32", NULL },
         { "33", NULL },
         { "34", NULL },
         { "35", NULL },
         { "36", NULL },
         { "37", NULL },
         { "38", NULL },
         { "39", NULL },
         { "40", NULL },
         { NULL, NULL },
      },
      "0"
   },
   {
      BEETLE_OPT(last_scanline),
      BEETLE_OPT_LAST_SCANLINE_LABEL_SV,
      NULL,
      BEETLE_OPT_LAST_SCANLINE_INFO_0_SV,
      NULL,
      "video",
      {
         { "210", NULL },
         { "211", NULL },
         { "212", NULL },
         { "213", NULL },
         { "214", NULL },
         { "215", NULL },
         { "216", NULL },
         { "217", NULL },
         { "218", NULL },
         { "219", NULL },
         { "220", NULL },
         { "221", NULL },
         { "222", NULL },
         { "223", NULL },
         { "224", NULL },
         { "225", NULL },
         { "226", NULL },
         { "227", NULL },
         { "228", NULL },
         { "229", NULL },
         { "230", NULL },
         { "231", NULL },
         { "232", NULL },
         { "233", NULL },
         { "234", NULL },
         { "235", NULL },
         { "236", NULL },
         { "237", NULL },
         { "238", NULL },
         { "239", OPTION_VAL_239_SV },
         { NULL, NULL },
      },
      "239"
   },
   {
      BEETLE_OPT(initial_scanline_pal),
      BEETLE_OPT_INITIAL_SCANLINE_PAL_LABEL_SV,
      NULL,
      BEETLE_OPT_INITIAL_SCANLINE_PAL_INFO_0_SV,
      NULL,
      "video",
      {
         { "0",  OPTION_VAL_0_O48_SV },
         { "1",  NULL },
         { "2",  NULL },
         { "3",  NULL },
         { "4",  NULL },
         { "5",  NULL },
         { "6",  NULL },
         { "7",  NULL },
         { "8",  NULL },
         { "9",  NULL },
         { "10", NULL },
         { "11", NULL },
         { "12", NULL },
         { "13", NULL },
         { "14", NULL },
         { "15", NULL },
         { "16", NULL },
         { "17", NULL },
         { "18", NULL },
         { "19", NULL },
         { "20", NULL },
         { "21", NULL },
         { "22", NULL },
         { "23", NULL },
         { "24", NULL },
         { "25", NULL },
         { "26", NULL },
         { "27", NULL },
         { "28", NULL },
         { "29", NULL },
         { "30", NULL },
         { "31", NULL },
         { "32", NULL },
         { "33", NULL },
         { "34", NULL },
         { "35", NULL },
         { "36", NULL },
         { "37", NULL },
         { "38", NULL },
         { "39", NULL },
         { "40", NULL },
         { NULL, NULL },
      },
      "0"
   },
   {
      BEETLE_OPT(last_scanline_pal),
      BEETLE_OPT_LAST_SCANLINE_PAL_LABEL_SV,
      NULL,
      BEETLE_OPT_LAST_SCANLINE_PAL_INFO_0_SV,
      NULL,
      "video",
      {
         { "230", NULL },
         { "231", NULL },
         { "232", NULL },
         { "233", NULL },
         { "234", NULL },
         { "235", NULL },
         { "236", NULL },
         { "237", NULL },
         { "238", NULL },
         { "239", NULL },
         { "240", NULL },
         { "241", NULL },
         { "242", NULL },
         { "243", NULL },
         { "244", NULL },
         { "245", NULL },
         { "246", NULL },
         { "247", NULL },
         { "248", NULL },
         { "249", NULL },
         { "250", NULL },
         { "251", NULL },
         { "252", NULL },
         { "253", NULL },
         { "254", NULL },
         { "255", NULL },
         { "256", NULL },
         { "257", NULL },
         { "258", NULL },
         { "259", NULL },
         { "260", NULL },
         { "261", NULL },
         { "262", NULL },
         { "263", NULL },
         { "264", NULL },
         { "265", NULL },
         { "266", NULL },
         { "267", NULL },
         { "268", NULL },
         { "269", NULL },
         { "270", NULL },
         { "271", NULL },
         { "272", NULL },
         { "273", NULL },
         { "274", NULL },
         { "275", NULL },
         { "276", NULL },
         { "277", NULL },
         { "278", NULL },
         { "279", NULL },
         { "280", NULL },
         { "281", NULL },
         { "282", NULL },
         { "283", NULL },
         { "284", NULL },
         { "285", NULL },
         { "286", NULL },
         { "287", OPTION_VAL_287_SV },
         { NULL, NULL },
      },
      "287"
   },
   { NULL, NULL, NULL, NULL, NULL, NULL, {{0}}, NULL },
};
struct retro_core_options_v2 options_sv = {
   option_cats_sv,
   option_defs_sv
};

/* RETRO_LANGUAGE_TH */

#define CATEGORY_SYSTEM_LABEL_TH "ระบบ"
#define CATEGORY_SYSTEM_INFO_0_TH "เปลี่ยนตัวเลือกของระบบและการโอเวอร์คล็อก"
#define CATEGORY_VIDEO_LABEL_TH "วิดีโอ"
#define CATEGORY_VIDEO_INFO_0_TH "เปลี่ยนอัตราส่วนภาพ, การครอบตัดการแสดงผล, ตัวกรองวิดีโอ และการข้ามเฟรม"
#define CATEGORY_PGXP_LABEL_TH NULL
#define CATEGORY_PGXP_INFO_0_TH "ตัวเลือกเหล่านี้ใช้ควบคุมส่วนเสริมที่สามารถปรับปรุงงานภาพให้ดีกว่าเครื่องคอนโซลดั้งเดิม PGXP สามารถขจัดปัญหาการบิดเบี้ยวของพื้นผิวและปัญหาภาพซ้อนทับกันได้"
#define CATEGORY_OSD_LABEL_TH "การแสดงผลบนหน้าจอ"
#define CATEGORY_OSD_INFO_0_TH "เปลี่ยนการแจ้งเตือนที่แสดงบนหน้าจอ"
#define CATEGORY_INPUT_LABEL_TH "การควบคุม"
#define CATEGORY_INPUT_INFO_0_TH "เปลี่ยนการตั้งค่า Light gun, เมาส์ และ neGcon"
#define CATEGORY_MEMCARDS_LABEL_TH "เม็มโมรี่การ์ด"
#define CATEGORY_MEMCARDS_INFO_0_TH "เปลี่ยนการตั้งค่าที่เกี่ยวข้องกับเมมโมรี่การ์ดจำลอง"
#define CATEGORY_HACKS_LABEL_TH NULL
#define CATEGORY_HACKS_INFO_0_TH "เปลี่ยนการตั้งค่าความแม่นยำของอีมูเลชันที่ส่งผลต่อประสิทธิภาพระดับล่างและความเข้ากันได้"
#define BEETLE_OPT_CPU_FREQ_SCALE_LABEL_TH "การปรับขนาดความถี่ CPU (โอเวอร์คล็อก)"
#define BEETLE_OPT_CPU_FREQ_SCALE_INFO_0_TH "โอเวอร์คล็อก (หรืออันเดอร์คล็อก) CPU ของ PSX ที่จำลองขึ้น การโอเวอร์คล็อกสามารถขจัดอาการหน่วงและปรับปรุงเฟรมเรตในบางเกมได้โดยแลกกับการใช้ทรัพยากรที่สูงขึ้น โปรดทราบว่าบางเกมมีการจำกัดเฟรมเรตภายในและอาจไม่ได้รับประโยชน์จากการโอเวอร์คล็อก อาจทำให้เอฟเฟกต์บางอย่างเคลื่อนไหวเร็วกว่าที่ควรจะเป็นในบางเกมเมื่อทำการโอเวอร์คล็อก"
#define OPTION_VAL_50_TH NULL
#define OPTION_VAL_60_TH NULL
#define OPTION_VAL_70_TH NULL
#define OPTION_VAL_80_TH NULL
#define OPTION_VAL_90_TH NULL
#define OPTION_VAL_100_TH "100% (ดั้งเดิม)"
#define OPTION_VAL_110_TH NULL
#define OPTION_VAL_120_TH NULL
#define OPTION_VAL_130_TH NULL
#define OPTION_VAL_140_TH NULL
#define OPTION_VAL_150_TH NULL
#define OPTION_VAL_160_TH NULL
#define OPTION_VAL_170_TH NULL
#define OPTION_VAL_180_TH NULL
#define OPTION_VAL_190_TH NULL
#define OPTION_VAL_200_TH NULL
#define OPTION_VAL_210_TH NULL
#define OPTION_VAL_220_TH NULL
#define OPTION_VAL_230_TH NULL
#define OPTION_VAL_240_TH NULL
#define OPTION_VAL_250_TH NULL
#define OPTION_VAL_260_TH NULL
#define OPTION_VAL_270_TH NULL
#define OPTION_VAL_280_TH NULL
#define OPTION_VAL_290_TH NULL
#define OPTION_VAL_300_TH NULL
#define OPTION_VAL_310_TH NULL
#define OPTION_VAL_320_TH NULL
#define OPTION_VAL_330_TH NULL
#define OPTION_VAL_340_TH NULL
#define OPTION_VAL_350_TH NULL
#define OPTION_VAL_360_TH NULL
#define OPTION_VAL_370_TH NULL
#define OPTION_VAL_380_TH NULL
#define OPTION_VAL_390_TH NULL
#define OPTION_VAL_400_TH NULL
#define OPTION_VAL_410_TH NULL
#define OPTION_VAL_420_TH NULL
#define OPTION_VAL_430_TH NULL
#define OPTION_VAL_440_TH NULL
#define OPTION_VAL_450_TH NULL
#define OPTION_VAL_460_TH NULL
#define OPTION_VAL_470_TH NULL
#define OPTION_VAL_480_TH NULL
#define OPTION_VAL_490_TH NULL
#define OPTION_VAL_500_TH NULL
#define OPTION_VAL_510_TH NULL
#define OPTION_VAL_520_TH NULL
#define OPTION_VAL_530_TH NULL
#define OPTION_VAL_540_TH NULL
#define OPTION_VAL_550_TH NULL
#define OPTION_VAL_560_TH NULL
#define OPTION_VAL_570_TH NULL
#define OPTION_VAL_580_TH NULL
#define OPTION_VAL_590_TH NULL
#define OPTION_VAL_600_TH NULL
#define OPTION_VAL_610_TH NULL
#define OPTION_VAL_620_TH NULL
#define OPTION_VAL_630_TH NULL
#define OPTION_VAL_640_TH NULL
#define OPTION_VAL_650_TH NULL
#define OPTION_VAL_660_TH NULL
#define OPTION_VAL_670_TH NULL
#define OPTION_VAL_680_TH NULL
#define OPTION_VAL_690_TH NULL
#define OPTION_VAL_700_TH NULL
#define OPTION_VAL_710_TH NULL
#define OPTION_VAL_720_TH NULL
#define OPTION_VAL_730_TH NULL
#define OPTION_VAL_740_TH NULL
#define OPTION_VAL_750_TH NULL
#define BEETLE_OPT_GPU_OVERCLOCK_LABEL_TH "โอเวอร์คล็อก GPU Rasterizer"
#define BEETLE_OPT_GPU_OVERCLOCK_INFO_0_TH "เปิดใช้งานการโอเวอร์คล็อก 2D rasterizer ภายใน GPU ของ PSX ที่จำลองขึ้น ไม่ช่วยปรับปรุงการเรนเดอร์ 3D และโดยทั่วไปมีผลเพียงเล็กน้อย"
#define OPTION_VAL_1X_NATIVE_TH "1x (ดั้งเดิม)"
#define OPTION_VAL_2X_TH NULL
#define OPTION_VAL_4X_TH NULL
#define OPTION_VAL_8X_TH NULL
#define OPTION_VAL_16X_TH NULL
#define OPTION_VAL_32X_TH NULL
#define BEETLE_OPT_GTE_OVERCLOCK_LABEL_TH "โอเวอร์คล็อก GTE"
#define BEETLE_OPT_GTE_OVERCLOCK_INFO_0_TH "ลดระยะเวลาการทำงานของ GTE (หน่วยประมวลผลร่วมของ CPU สำหรับกราฟิก 3D) ที่จำลองขึ้นทั้งหมดให้เหลือเพียงรอบสัญญาณนาฬิกาเดียว สำหรับเกมที่ใช้งาน GTE อย่างหนัก สิ่งนี้สามารถช่วยปรับปรุงเฟรมเรตและความเสถียรของเวลาในแต่ละเฟรมได้อย่างมาก"
#define BEETLE_OPT_CD_ACCESS_METHOD_LABEL_TH "วิธีการเข้าถึง CD"
#define BEETLE_OPT_CD_ACCESS_METHOD_INFO_0_TH "'Synchronous' เลียนแบบฮาร์ดแวร์ดั้งเดิม 'Asynchronous' ลดอาการกระตุกสำหรับเครื่องที่อ่านข้อมูลช้า 'Pre-Cache' โหลดไฟล์เกมลงแรมทั้งหมดเพื่อลดเวลาโหลดในเกม แต่อาจเริ่มเกมช้าลงและมีปัญหากับเครื่องที่ RAM น้อย (จะสลับเป็น 'Synchronous' หากใช้แผ่นจริง) ต้องรีสตาร์ทเพื่อเริ่มผลใหม่"
#define OPTION_VAL_SYNC_TH NULL
#define OPTION_VAL_ASYNC_TH NULL
#define OPTION_VAL_PRECACHE_TH NULL
#define BEETLE_OPT_CD_FASTLOAD_LABEL_TH "ความเร็วการโหลด CD"
#define BEETLE_OPT_CD_FASTLOAD_INFO_0_TH "ค่าที่สูงกว่า '2x (ดั้งเดิม)' ช่วยลดเวลาโหลดในเกมได้มาก แต่อาจทำให้จังหวะเวลาผิดพลาด บางเกมอาจทำงานผิดปกติหากใช้ค่าที่สูงเกินไป"
#define OPTION_VAL_2X_NATIVE_TH "2x (ดั้งเดิม)"
#define OPTION_VAL_6X_TH NULL
#define OPTION_VAL_10X_TH NULL
#define OPTION_VAL_12X_TH NULL
#define OPTION_VAL_14X_TH NULL
#define BEETLE_OPT_SKIP_BIOS_LABEL_TH "ข้าม BIOS"
#define BEETLE_OPT_SKIP_BIOS_INFO_0_TH "ข้ามแอนิเมชันการบูต BIOS ของ PlayStation ที่ปกติจะแสดงเมื่อโหลดเนื้อหา หมายเหตุ: การเปิดใช้งานสิ่งนี้จะทำให้เกิดปัญหาความเข้ากันได้กับหลายเกม (เกมที่มีการป้องกันการก๊อปปี้โซน PAL, Saga Frontier และอื่นๆ)"
#define BEETLE_OPT_OVERRIDE_BIOS_LABEL_TH "เขียนทับ BIOS"
#define BEETLE_OPT_OVERRIDE_BIOS_INFO_0_TH "เขียนทับ BIOS มาตรฐานเฉพาะโซนด้วยเวอร์ชัน Region-free หากตรวจพบ จำเป็นต้องเริ่มระบบใหม่"
#define OPTION_VAL_PSXONPSP_TH NULL
#define OPTION_VAL_PS1_ROM_TH NULL
#define OPTION_VAL_OPENBIOS_TH "Open Source BIOS"
#define BEETLE_OPT_REGION_LABEL_TH "ภูมิภาคระบบ"
#define BEETLE_OPT_REGION_INFO_0_TH NULL
#define OPTION_VAL_AUTO_TH "อัตโนมัติ"
#define OPTION_VAL_NTSC_J_TH NULL
#define OPTION_VAL_NTSC_U_TH NULL
#define OPTION_VAL_PAL_TH NULL
#define BEETLE_OPT_INTERNAL_RESOLUTION_LABEL_TH "ความละเอียด GPU ภายใน"
#define BEETLE_OPT_INTERNAL_RESOLUTION_INFO_0_TH "ความละเอียดที่สูงกว่า '1x (ดั้งเดิม)' จะช่วยเพิ่มความคมชัดของโมเดล 3D แต่ต้องแลกมาด้วยการใช้ทรัพยากรเครื่องที่สูงขึ้น โดยทั่วไปแล้วองค์ประกอบที่เป็น 2D จะไม่ได้รับผลกระทบจากการตั้งค่านี้"
#define BEETLE_OPT_RENDERER_LABEL_TH "การเรนเดอร์"
#define BEETLE_OPT_RENDERER_INFO_0_TH "การเรนเดอร์ด้วยซอฟต์แวร์มีความแม่นยำที่สุด แต่ก็ต้องการประสิทธิภาพเครื่องที่สูงมากเมื่อรันที่ความละเอียด GPU ภายในที่เพิ่มขึ้น ในขณะที่การเรนเดอร์ด้วยฮาร์ดแวร์แม้จะมีความแม่นยำน้อยกว่า แต่จะช่วยเพิ่มประสิทธิภาพได้ดีกว่าการเรนเดอร์ด้วยซอฟต์แวร์เมื่อเพิ่มความละเอียด และยังช่วยให้สามารถปรับแต่งกราฟิกต่างๆ ได้อีกด้วย ตัวเลือก 'Hardware (อัตโนมัติ)' จะเลือกการเรนเดอร์ด้วย Vulkan หรือ OpenGL ให้โดยอัตโนมัติ ขึ้นอยู่กับไดรเวอร์วิดีโอของ libretro frontend ที่ใช้อยู่ในขณะนั้น หากไดรเวอร์วิดีโอที่ใช้ไม่รองรับ Vulkan หรือ OpenGL 3.3 คอร์จะสลับกลับไปใช้การเรนเดอร์ด้วยซอฟต์แวร์แทน จำเป็นต้องเริ่มระบบใหม่"
#define OPTION_VAL_HARDWARE_TH "ฮาร์ดแวร์ (อัตโนมัติ)"
#define OPTION_VAL_HARDWARE_GL_TH "ฮาร์ดแวร์ (OpenGL)"
#define OPTION_VAL_HARDWARE_VK_TH "ฮาร์ดแวร์ (Vulkan)"
#define OPTION_VAL_SOFTWARE_TH "ซอฟต์แวร์"
#define BEETLE_OPT_RENDERER_SOFTWARE_FB_LABEL_TH "เฟรมบัฟเฟอร์ซอฟต์แวร์"
#define BEETLE_OPT_RENDERER_SOFTWARE_FB_INFO_0_TH "เปิดใช้งานการจำลองเอฟเฟกต์เฟรมบัฟเฟอร์ที่แม่นยำ (เช่น ภาพเบลอจากการเคลื่อนไหว, เอฟเฟกต์วงล้อมก่อนตัดเข้าฉากต่อสู้ใน FF7) เมื่อใช้การเรนเดอร์ด้วยฮาร์ดแวร์ โดยการรันการเรนเดอร์ด้วยซอฟต์แวร์ควบคู่ไปในพื้นหลังที่ความละเอียดดั้งเดิม หากปิดใช้งาน การทำงานเหล่านี้จะถูกข้ามไป (สำหรับ OpenGL) หรือถูกเรนเดอร์บน GPU (สำหรับ Vulkan) การปิดใช้งานอาจช่วยเพิ่มประสิทธิภาพได้แต่อาจทำให้เกิดข้อผิดพลาดทางกราฟิกที่รุนแรง หากไม่แน่ใจให้เปิดทิ้งไว้"
#define BEETLE_OPT_DEPTH_LABEL_TH "ความลึกสีภายใน"
#define BEETLE_OPT_DEPTH_INFO_0_TH NULL
#define OPTION_VAL_16BPP_NATIVE_TH "16 bpp (ดั้งเดิม)"
#define OPTION_VAL_32BPP_TH NULL
#define BEETLE_OPT_DITHER_MODE_LABEL_TH "รูปแบบการกระจายเม็ดสี"
#define BEETLE_OPT_DITHER_MODE_INFO_0_TH "'1x (ดั้งเดิม)' จะจำลองการกระจายเม็ดสีความละเอียดต่ำแบบดั้งเดิมที่เครื่องเกมต้นฉบับใช้ เพื่อช่วยลดรอยต่อของสี (Color Banding) ที่เห็นได้ชัดในความลึกสีระดับดั้งเดิม ส่วน 'ความละเอียดภายใน' จะปรับขนาดความละเอียดของการกระจายเม็ดสีตามค่าความละเอียดภายในที่ตั้งไว้เพื่อให้ได้ผลลัพธ์ที่ดูสะอาดตาขึ้น แนะนำให้ปิดการใช้งานเมื่อรันที่ความลึกสี 32 bpp"
#define OPTION_VAL_INTERNAL_RESOLUTION_TH "ความละเอียดภายใน"
#define BEETLE_OPT_SCALED_UV_OFFSET_LABEL_TH "ระยะห่าง UV ของพื้นผิว"
#define BEETLE_OPT_SCALED_UV_OFFSET_INFO_0_TH "สุ่มตัวอย่างตำแหน่งพื้นผิว สำหรับโพลีกอน 3D แบบขยับระยะ เมื่อใช้งานความละเอียดภายในที่สูงกว่า 1x ช่วยลดรอยต่อของพื้นผิว แต่อาจทำให้เกิดข้อผิดพลาดทางกราฟิกที่ไม่พึงประสงค์ได้"
#define BEETLE_OPT_FILTER_LABEL_TH "การกรองพื้นผิว"
#define BEETLE_OPT_FILTER_INFO_0_TH "'Nearest' จำลองตามเครื่องต้นฉบับ 'Bilinear' และ '3-Point' เป็นฟิลเตอร์ลดรอยหยักโดยการเบลอภาพ ส่วน 'SABR', 'xBR' และ 'JINC2' เป็นฟิลเตอร์ขยายภาพที่ช่วยเพิ่มความคมชัดของพื้นผิวแต่กินทรัพยากรเครื่องมากขึ้น รองรับเฉพาะการเรนเดอร์ด้วยฮาร์ดแวร์เท่านั้น"
#define OPTION_VAL_NEAREST_TH "ใกล้ที่สุด"
#define OPTION_VAL_SABR_TH NULL
#define OPTION_VAL_XBR_TH NULL
#define OPTION_VAL_BILINEAR_TH NULL
#define OPTION_VAL_3_POINT_TH NULL
#define OPTION_VAL_JINC2_TH NULL
#define BEETLE_OPT_FILTER_EXCLUDE_SPRITE_LABEL_TH "ยกเว้นสไปรท์จากการกรองภาพ"
#define BEETLE_OPT_FILTER_EXCLUDE_SPRITE_INFO_0_TH "ไม่ต้องใช้การกรองพื้นผิวกับสไปรท์ ช่วยป้องกันรอยแยกของภาพในหลายเกมที่ใช้พื้นหลังแบบสไปรท์ 2D แนะนำให้ใช้ร่วมกับ 'การลดรอยหยักแบบปรับตามภาพ' (Adaptive Smoothing) หรือฟิลเตอร์ปรับแต่งภาพอื่น ๆ เพื่อผลลัพธ์ที่ดีที่สุด"
#define OPTION_VAL_OPAQUE_TH "เฉพาะทึบแสง"
#define OPTION_VAL_ALL_TH "ทึบแสงและกึ่งโปร่งใส"
#define BEETLE_OPT_FILTER_EXCLUDE_2D_POLYGON_LABEL_TH "ยกเว้นโพลีกอน 2D จากการกรองภาพ"
#define BEETLE_OPT_FILTER_EXCLUDE_2D_POLYGON_INFO_0_TH "ไม่ต้องใช้การกรองพื้นผิวกับโพลีกอน 2D ซึ่งตรวจหาด้วยระบบการคาดคะเนและอาจเกิดข้อผิดพลาดได้ แนะนำให้ใช้ร่วมกับ 'การลดรอยหยักแบบปรับตามภาพ' หรือฟิลเตอร์ปรับแต่งภาพอื่น ๆ เพื่อผลลัพธ์ที่ดีที่สุด"
#define BEETLE_OPT_ADAPTIVE_SMOOTHING_LABEL_TH "การลดรอยหยักแบบปรับตามภาพ"
#define BEETLE_OPT_ADAPTIVE_SMOOTHING_INFO_0_TH "ช่วยลดรอยหยักของภาพ 2D และเมนู (UI) ให้เนียนตาขึ้นโดยไม่ทำให้วัตถุ 3D เบลอ รองรับเฉพาะการเรนเดอร์ด้วย Vulkan เท่านั้น"
#define BEETLE_OPT_SUPER_SAMPLING_LABEL_TH "การสุ่มตัวอย่างระดับสูง (ลดความละเอียดลงมาที่ระดับดั้งเดิม)"
#define BEETLE_OPT_SUPER_SAMPLING_INFO_0_TH "ลดระดับสัญญาณภาพจากการเพิ่มความละเอียดภายในกลับมาเป็นความละเอียดดั้งเดิม การใช้ตัวเลือกนี้ร่วมกับตัวคูณความละเอียดภายในที่สูงขึ้น จะช่วยให้แสดงผลวัตถุ 3D แบบลบรอยหยักในความละเอียดต่ำแบบดั้งเดิมได้ ให้ผลลัพธ์ดีที่สุดกับเกมที่ผสมองค์ประกอบ 2D และ 3D (เช่น ตัวละคร 3D บนพื้นหลังภาพนิ่ง) และใช้งานร่วมกับ CRT Shader ได้ดี รองรับเฉพาะการเรนเดอร์ด้วย Vulkan เท่านั้น"
#define BEETLE_OPT_MSAA_LABEL_TH "การลบรอยหยักแบบสุ่มตัวอย่างหลายตำแหน่ง"
#define BEETLE_OPT_MSAA_INFO_0_TH "MSAA ช่วยปรับปรุงลักษณะของวัตถุ 3D ให้ดูดีขึ้น รองรับเฉพาะการเรนเดอร์ด้วย Vulkan เท่านั้น"
#define OPTION_VAL_1X_TH NULL
#define BEETLE_OPT_MDEC_YUV_LABEL_TH "ฟิลเตอร์จัดการความอิ่มตัวของสี"
#define BEETLE_OPT_MDEC_YUV_INFO_0_TH "ปรับปรุงคุณภาพการเล่นวิดีโอ (FMV) โดยการลดความผิดเพี้ยนของภาพแบบ 'Macroblocking' (รอยเหลี่ยมหรือขอบหยัก) รองรับเฉพาะการเรนเดอร์ด้วย Vulkan เท่านั้น"
#define BEETLE_OPT_TRACK_TEXTURES_LABEL_TH "ติดตามพื้นผิว"
#define BEETLE_OPT_TRACK_TEXTURES_INFO_0_TH "เงื่อนไขเบื้องต้นสำหรับการ Dump และการแทนที่ Texture แต่อาจทำให้เกมส่วนใหญ่ค้างได้"
#define BEETLE_OPT_DUMP_TEXTURES_LABEL_TH NULL
#define BEETLE_OPT_DUMP_TEXTURES_INFO_0_TH "ดัมป์พื้นผิวที่ใช้ลงใน <cd>-texture-dump/"
#define BEETLE_OPT_REPLACE_TEXTURES_LABEL_TH "แทนที่พื้นผิว (Texture)"
#define BEETLE_OPT_REPLACE_TEXTURES_INFO_0_TH "แทนที่พื้นผิว (Texture) โดยใช้เวอร์ชัน HD จาก <cd>-texture-replacements/"
#define BEETLE_OPT_PGXP_MODE_LABEL_TH "โหมดการทำงาน PGXP"
#define BEETLE_OPT_PGXP_MODE_INFO_0_TH "เรนเดอร์วัตถุ 3D ด้วยความละเอียดระดับซับพิกเซล เพื่อลดการบิดเบี้ยวและอาการสั่นของวัตถุ 3D ที่มักพบในฮาร์ดแวร์ดั้งเดิมเนื่องจากการใช้พิกัดจุดยอดแบบเลขฐานคงที่ โหมด 'Memory Only' มีปัญหาความเข้ากันได้น้อยที่สุดและแนะนำสำหรับการใช้งานทั่วไป ส่วนโหมด 'Memory + CPU (Buggy)' สามารถลดอาการสั่นได้ดียิ่งขึ้นแต่ต้องการประสิทธิภาพเครื่องสูงและอาจทำให้เกิดข้อผิดพลาดของรูปทรงเรขาคณิตได้หลากหลายรูปแบบ"
#define OPTION_VAL_MEMORY_ONLY_TH "หน่วยความจำเท่านั้น"
#define OPTION_VAL_MEMORY_CPU_TH "หน่วยความจำ + CPU (ไม่เสถียร)"
#define BEETLE_OPT_PGXP_2D_TOL_LABEL_TH "พิกัดความต่างรูปทรง 2D ของ PGXP"
#define BEETLE_OPT_PGXP_2D_TOL_INFO_0_TH "ซ่อนข้อผิดพลาดที่ชัดเจนในการทำงานของ PGXP: ค่านี้กำหนดระยะพิกัดความต่างที่จะยอมให้ใช้ค่า PGXP ได้ ในกรณีที่รูปทรงเรขาคณิตไม่มีข้อมูลความลึกที่เหมาะสม"
#define OPTION_VAL_0PX_TH NULL
#define OPTION_VAL_1PX_TH NULL
#define OPTION_VAL_2PX_TH NULL
#define OPTION_VAL_3PX_TH NULL
#define OPTION_VAL_4PX_TH NULL
#define OPTION_VAL_5PX_TH NULL
#define OPTION_VAL_6PX_TH NULL
#define OPTION_VAL_7PX_TH NULL
#define OPTION_VAL_8PX_TH NULL
#define BEETLE_OPT_PGXP_NCLIP_LABEL_TH "การคัดกรองโพลีกอน PGXP"
#define BEETLE_OPT_PGXP_NCLIP_INFO_0_TH "ใช้การประมวลผล NCLIP ของ PGXP; ช่วยปรับปรุงการแสดงผลโดยลดช่องโหว่ในรูปทรงเรขาคณิตที่มีพิกัด PGXP แต่อาจทำให้บางเกมค้างได้ในหลายสถานการณ์"
#define BEETLE_OPT_PGXP_VERTEX_LABEL_TH NULL
#define BEETLE_OPT_PGXP_VERTEX_INFO_0_TH "แคชตำแหน่งจุดยอด (Vertex) ที่ปรับปรุงด้วย PGXP เพื่อนำกลับมาใช้ใหม่ในการวาดโพลีกอน; ช่วยปรับปรุงการจัดวางวัตถุและลดรอยต่อที่มองเห็นได้เมื่อเรนเดอร์พื้นผิว แต่หากมีการดึงข้อมูลจากแคชผิดพลาดอาจทำให้เกิดอาการภาพค้างหรือกราฟิกผิดเพี้ยนได้ ในขณะนี้แนะนำให้ปิดตัวเลือกนี้ไว้ โดยตัวเลือกนี้จะมีผลเมื่อเปิดใช้งานโหมดการทำงาน PGXP เท่านั้น และรองรับเฉพาะตัวเรนเดอร์ฮาร์ดแวร์"
#define BEETLE_OPT_PGXP_TEXTURE_LABEL_TH "การปรับปรุงพื้นผิวให้ถูกต้องตามมิติภาพ"
#define BEETLE_OPT_PGXP_TEXTURE_INFO_0_TH NULL
#define BEETLE_OPT_DISPLAY_INTERNAL_FPS_LABEL_TH "แสดง FPS ภายใน"
#define BEETLE_OPT_DISPLAY_INTERNAL_FPS_INFO_0_TH "แสดงอัตราเฟรมเรตภายใน (FPS) ที่ระบบ PlayStation ที่จำลองอยู่กำลังประมวลผล หมายเหตุ: จำเป็นต้องเปิดใช้งานการแจ้งเตือนบนหน้าจอ ในส่วนหน้าของ libretro (Frontend) ก่อนจึงจะแสดงผลได้"
#define BEETLE_OPT_DISPLAY_OSD_LABEL_TH "แสดงข้อความแจ้งเตือนบนหน้าจอ"
#define BEETLE_OPT_DISPLAY_OSD_INFO_0_TH "แสดงข้อความ OSD ที่สร้างโดย Core"
#define BEETLE_OPT_DISPLAY_VRAM_LABEL_TH "แสดง VRAM ทั้งหมด (Debug)"
#define BEETLE_OPT_DISPLAY_VRAM_INFO_0_TH "แสดงภาพ VRAM ของระบบคอนโซลที่จำลองอยู่ทั้งหมด รองรับเฉพาะตัวเรนเดอร์ฮาร์ดแวร์ OpenGL และ Vulkan เท่านั้น หมายเหตุ: ใช้สำหรับการ Debug และควรปิดไว้ในการใช้งานปกติ"
#define BEETLE_OPT_ANALOG_CALIBRATION_LABEL_TH "การปรับเทียบ Analog อัตโนมัติ"
#define BEETLE_OPT_ANALOG_CALIBRATION_INFO_0_TH "เปิดใช้งานการปรับเทียบอินพุต Analog แบบไดนามิกเมื่อตั้งค่าอุปกรณ์เป็น DualShock, Analog Controller, Analog Joystick หรือ neGcon ค่าอินพุตสูงสุดที่บันทึกได้จะถูกตรวจสอบแบบเรียลไทม์และใช้ในการปรับขนาดพิกัด Analog ที่ส่งไปยังโปรแกรมจำลอง ควรใช้ตัวเลือกนี้สำหรับเกมอย่าง Mega Man Legends 2 ที่ต้องการค่าอินพุตที่มากกว่าที่คอนโทรลเลอร์สมัยใหม่จะให้ได้ เพื่อผลลัพธ์ที่ดีที่สุด ควรหมุนก้าน Analog ให้สุดทุกทิศทางเพื่อจูนอัลกอริทึมการปรับเทียบทุกครั้งที่โหลดเนื้อหา"
#define BEETLE_OPT_ANALOG_TOGGLE_LABEL_TH "เปิด-ปิด โหมด Analog ของ DualShock"
#define BEETLE_OPT_ANALOG_TOGGLE_INFO_0_TH "เมื่อประเภทอุปกรณ์อินพุตคือ DualShock ตัวเลือกนี้จะอนุญาตให้เปิด-ปิดการสลับระหว่างโหมด DIGITAL และ ANALOG ของ DualShock ที่จำลองอยู่ได้เหมือนกับฮาร์ดแวร์ดั้งเดิม นอกจากนี้ยังสามารถเปิด-ปิดโหมดได้โดยใช้การกดปุ่มผสมตามที่เลือกไว้ด้วย"
#define OPTION_VAL_ENABLED_ANALOG_TH "เริ่มต้น-Analog"
#define BEETLE_OPT_ANALOG_TOGGLE_COMBO_LABEL_TH "ปุ่มคอมโบโหมดอนาล็อก DualShock"
#define BEETLE_OPT_ANALOG_TOGGLE_COMBO_INFO_0_TH "คอมโบปุ่มที่ใช้สำหรับสลับระหว่างโหมด DIGITAL และ ANALOG ของจอย DualShock ที่จำลองอยู่"
#define OPTION_VAL_L1_L2_R1_R2_START_SELECT_TH NULL
#define OPTION_VAL_L1_R1_SELECT_TH NULL
#define OPTION_VAL_L1_R1_START_TH NULL
#define OPTION_VAL_L1_R1_L3_TH NULL
#define OPTION_VAL_L1_R1_R3_TH NULL
#define OPTION_VAL_L2_R2_SELECT_TH NULL
#define OPTION_VAL_L2_R2_START_TH NULL
#define OPTION_VAL_L2_R2_L3_TH NULL
#define OPTION_VAL_L2_R2_R3_TH NULL
#define OPTION_VAL_L3_R3_TH NULL
#define BEETLE_OPT_ANALOG_TOGGLE_HOLD_LABEL_TH "ระยะหน่วงการกดคอมโบค้างเพื่อเข้าโหมด Analog ของ DualShock"
#define BEETLE_OPT_ANALOG_TOGGLE_HOLD_INFO_0_TH "ระยะเวลากดคอมโบปุ่มค้างเพื่อเข้าโหมด Analog"
#define OPTION_VAL_0_TH "0 วินาที (ไม่มีดีเลย์)"
#define OPTION_VAL_1_TH "1 วินาที (ดีเลย์)"
#define OPTION_VAL_2_TH "2 วินาที (ดีเลย์)"
#define OPTION_VAL_3_TH "3 วินาที (ดีเลย์)"
#define OPTION_VAL_4_TH "4 วินาที (ดีเลย์)"
#define OPTION_VAL_5_TH "5 วินาที (ดีเลย์)"
#define BEETLE_OPT_ENABLE_MULTITAP_PORT1_LABEL_TH "มัลติแทปบนพอร์ต 1"
#define BEETLE_OPT_ENABLE_MULTITAP_PORT1_INFO_0_TH "เปิดใช้งานมัลติแทปบนพอร์ต 1"
#define BEETLE_OPT_ENABLE_MULTITAP_PORT2_LABEL_TH "มัลติแทปบนพอร์ต 2"
#define BEETLE_OPT_ENABLE_MULTITAP_PORT2_INFO_0_TH "เปิดใช้งานมัลติแทปบนพอร์ต 2"
#define BEETLE_OPT_GUN_INPUT_MODE_LABEL_TH "โหมดอินพุตปืน"
#define BEETLE_OPT_GUN_INPUT_MODE_INFO_0_TH "ใช้อินพุต 'ปืนแสง' (Light Gun) ที่ควบคุมด้วยเมาส์ หรืออินพุต 'หน้าจอสัมผัส' เมื่อตั้งค่าประเภทอุปกรณ์เป็น 'Guncon/G-Con 45' หรือ 'Justifier'"
#define OPTION_VAL_LIGHTGUN_TH "ปืนแสง"
#define OPTION_VAL_TOUCHSCREEN_TH "หน้าจอสัมผัส"
#define BEETLE_OPT_GUN_CURSOR_LABEL_TH "เป้าเล็งปืน"
#define BEETLE_OPT_GUN_CURSOR_INFO_0_TH "แสดงเป้าเล็งบนหน้าจอขณะใช้อุปกรณ์อินพุตประเภท 'Guncon/G-Con 45' และ 'Justifier'"
#define OPTION_VAL_CROSS_TH "กากบาท"
#define OPTION_VAL_DOT_TH "จุด"
#define BEETLE_OPT_CROSSHAIR_COLOR_P1_LABEL_TH "สีของเป้ากลางปืนผู้เล่น: พอร์ต 1"
#define BEETLE_OPT_CROSSHAIR_COLOR_P1_INFO_0_TH "สีของเป้ากลางปืนผู้เล่น พอร์ต 1"
#define OPTION_VAL_RED_TH NULL
#define OPTION_VAL_BLUE_TH "น้ำเงิน"
#define OPTION_VAL_GREEN_TH "เขียว"
#define OPTION_VAL_ORANGE_TH "ส้ม"
#define OPTION_VAL_YELLOW_TH NULL
#define OPTION_VAL_CYAN_TH "ฟ้าอมเขียว"
#define OPTION_VAL_PINK_TH "ชมพู"
#define OPTION_VAL_PURPLE_TH "ม่วง"
#define OPTION_VAL_BLACK_TH "ดำ"
#define OPTION_VAL_WHITE_TH "ขาว"
#define BEETLE_OPT_CROSSHAIR_COLOR_P2_LABEL_TH "สีของเป้ากลางปืนผู้เล่น: พอร์ต 2"
#define BEETLE_OPT_CROSSHAIR_COLOR_P2_INFO_0_TH "สีของเป้ากลางปืนผู้เล่น พอร์ต 2"
#define BEETLE_OPT_MOUSE_SENSITIVITY_LABEL_TH "ความไวเมาส์"
#define BEETLE_OPT_MOUSE_SENSITIVITY_INFO_0_TH "การตอบสนองของอุปกรณ์อินพุต 'เมาส์'"
#define OPTION_VAL_5_O42_TH NULL
#define OPTION_VAL_10_TH NULL
#define OPTION_VAL_15_TH NULL
#define OPTION_VAL_20_TH NULL
#define OPTION_VAL_25_TH NULL
#define OPTION_VAL_30_TH NULL
#define OPTION_VAL_35_TH NULL
#define OPTION_VAL_40_TH NULL
#define OPTION_VAL_45_TH NULL
#define OPTION_VAL_55_TH NULL
#define OPTION_VAL_65_TH NULL
#define OPTION_VAL_75_TH NULL
#define OPTION_VAL_85_TH NULL
#define OPTION_VAL_95_TH NULL
#define OPTION_VAL_100_O42_TH NULL
#define OPTION_VAL_105_TH NULL
#define OPTION_VAL_115_TH NULL
#define OPTION_VAL_125_TH NULL
#define OPTION_VAL_135_TH NULL
#define OPTION_VAL_145_TH NULL
#define OPTION_VAL_155_TH NULL
#define OPTION_VAL_165_TH NULL
#define OPTION_VAL_175_TH NULL
#define OPTION_VAL_185_TH NULL
#define OPTION_VAL_195_TH NULL
#define BEETLE_OPT_NEGCON_RESPONSE_LABEL_TH "การตอบสนองต่อการบิดของ neGcon"
#define BEETLE_OPT_NEGCON_RESPONSE_INFO_0_TH "รูปแบบการตอบสนองของอนาล็อกซ้ายเมื่อจำลองการ \"บิด\" ของอุปกรณ์อินพุต 'neGcon' ที่ถูกจำลองขึ้น การเลื่อนของก้านอนาล็อกอาจถูกจับคู่เข้ากับมุมหมุนของ neGcon ได้ทั้งแบบ Linear, Quadratic หรือ Cubic 'Quadratic' ช่วยให้มีความแม่นยำมากกว่า 'Linear' เมื่อทำการเคลื่อนไหวเพียงเล็กน้อย 'Cubic' จะยิ่งเพิ่มความแม่นยำในการเคลื่อนไหวเพียงเล็กน้อย แต่จะทำให้การเคลื่อนไหวในช่วงที่กว้างขึ้นดู 'เกินจริง' หมายเหตุ: แนะนำให้ใช้ 'Linear' เฉพาะเมื่อใช้อุปกรณ์เสริมประเภทพวงมาลัยเท่านั้น คอนโทรลเลอร์ทั่วไปมีการส่งค่าอินพุตแบบอนาล็อกในลักษณะที่แตกต่างจากกลไกการ 'บิด' ของ neGcon โดยสิ้นเชิง ดังนั้นการจับคู่แบบ Linear จะขยายการตอบสนองของการเคลื่อนไหวเพียงเล็กน้อยมากเกินไปจนทำให้เสียการควบคุมที่ละเอียดอ่อน ในกรณีส่วนใหญ่ 'Quadratic' จะให้ผลลัพธ์ที่ใกล้เคียงกับฮาร์ดแวร์จริงมากที่สุด"
#define OPTION_VAL_LINEAR_TH "เส้นตรง"
#define OPTION_VAL_QUADRATIC_TH NULL
#define OPTION_VAL_CUBIC_TH NULL
#define BEETLE_OPT_NEGCON_DEADZONE_LABEL_TH "ระยะฟรีของการบิด neGcon"
#define BEETLE_OPT_NEGCON_DEADZONE_INFO_0_TH "ระยะฟรีของก้านอนาล็อกซ้ายเมื่อจำลองการ \"บิด\" ของอุปกรณ์อินพุต 'neGcon' ที่ถูกจำลองขึ้น ใช้เพื่อแก้ปัญหาจอยเดินเอง (Controller Drift) หมายเหตุ: เกมที่รองรับ neGcon ส่วนใหญ่จะมีตัวเลือกในเกมสำหรับตั้งค่าระยะฟรีของการบิดอยู่แล้ว เพื่อป้องกันการสูญเสียความแม่นยำ ควรตั้งค่าระยะฟรีในเกมเป็นศูนย์ *เสมอ* และควรปรับแต่งค่าที่จำเป็นผ่านตัวเลือกของ Core นี้ *เท่านั้น* ซึ่งถือเป็นเรื่องสำคัญอย่างยิ่งเมื่อตั้งค่า 'การตอบสนองต่อการบิดของ neGcon' เป็นแบบ 'Quadratic' หรือ 'Cubic'"
#define OPTION_VAL_0_O44_TH NULL
#define BEETLE_OPT_USE_MEDNAFEN_MEMCARD0_METHOD_LABEL_TH "รูปแบบเมมโมรี่การ์ด"
#define BEETLE_OPT_USE_MEDNAFEN_MEMCARD0_METHOD_INFO_0_TH "รูปแบบการเซฟข้อมูลสำหรับช่องเมมโมรี่การ์ดซ้าย 'Mednafen' ใช้เพื่อให้เข้ากับ Mednafen รุ่นปกติ เมื่อใช้กับ Beetle PSX ไฟล์เซฟ Libretro (.srm) และ Mednafen (.mcr) จะมีรูปแบบภายในเหมือนกันและเปลี่ยนชื่อเพื่อแปลงไฟล์ได้ ต้องรีสตาร์ทเพื่อเริ่มผล"
#define OPTION_VAL_LIBRETRO_TH NULL
#define OPTION_VAL_MEDNAFEN_TH NULL
#define BEETLE_OPT_ENABLE_MEMCARD1_LABEL_TH "ช่องเสียบเมมโมรี่การ์ด 2"
#define BEETLE_OPT_ENABLE_MEMCARD1_INFO_0_TH "จำลองเมมโมรี่การ์ดใบที่สองในช่องเสียบด้านขวา หมายเหตุ: บางเกมจำเป็นต้องปิดตัวเลือกนี้เพื่อให้ทำงานได้ถูกต้อง (เช่น Codename Tenka) หมายเหตุ: ช่องเสียบเมมโมรี่การ์ด 2 จะใช้รูปแบบการเซฟแบบ Mednafen (.mcr) ต้องรีสตาร์ทเพื่อเริ่มผล"
#define BEETLE_OPT_SHARED_MEMORY_CARDS_LABEL_TH "แชร์เมมโมรี่การ์ด"
#define BEETLE_OPT_SHARED_MEMORY_CARDS_INFO_0_TH "ใช้เมมโมรี่การ์ดร่วมกันแทนการแยกไฟล์ตามรายชื่อเกม หมายเหตุ: หากตั้งค่า 'รูปแบบเมมโมรี่การ์ด' เป็น 'Libretro' จะมีผลเฉพาะช่องเสียบที่ 2 เท่านั้น ต้องรีสตาร์ทเพื่อเริ่มผล"
#define BEETLE_OPT_MEMCARD_LEFT_INDEX_LABEL_TH "ช่องเสียบเมมโมรี่การ์ด 1"
#define BEETLE_OPT_MEMCARD_LEFT_INDEX_INFO_0_TH "เมมโมรี่การ์ดที่โหลดอยู่ในช่องซ้าย ตัวเลือกนี้จะใช้งานได้ต่อเมื่อตั้งค่ารูปแบบเมมโมรี่การ์ดเป็น Mednafen เท่านั้น"
#define OPTION_VAL_0_O48_TH NULL
#define BEETLE_OPT_MEMCARD_RIGHT_INDEX_LABEL_TH "ช่องเสียบเมมโมรี่การ์ด 2"
#define BEETLE_OPT_MEMCARD_RIGHT_INDEX_INFO_0_TH "เมมโมรี่การ์ดที่โหลดอยู่ในช่องขวา ตัวเลือกนี้จะใช้งานได้ต่อเมื่อเปิดใช้งานช่องเสียบเมมโมรี่การ์ด 2 เท่านั้น"
#define OPTION_VAL_1_O49_TH NULL
#define BEETLE_OPT_LINE_RENDER_LABEL_TH NULL
#define BEETLE_OPT_LINE_RENDER_INFO_0_TH "บางเกม (เช่น Doom, Hexen, Soul Blade เป็นต้น) วาดเส้นแนวนอนด้วยการยืดรูปสามเหลี่ยมที่มีความสูงเพียงหนึ่งพิกเซลไปทั่วหน้าจอ ซึ่งบนฮาร์ดแวร์ดั้งเดิมจะถูกประมวลผลเป็นแถวพิกเซล แฮกนี้จะตรวจจับสามเหลี่ยมขนาดเล็กเหล่านี้และแปลงเป็นรูปสี่เหลี่ยม (Quads) ตามความจำเป็น เพื่อให้แสดงผลได้อย่างถูกต้องบน Hardware Renderer และเมื่อเพิ่มความละเอียดภายใน (Upscale) โดยตัวเลือก 'แบบรุก' จำเป็นสำหรับบางเกม (เช่น Dark Forces, Duke Nukem) แต่อาจทำให้เกิดปัญหาภาพผิดเพี้ยนในเกมอื่น หากไม่แน่ใจให้เลือก 'Default' ไว้ตามเดิม"
#define OPTION_VAL_DEFAULT_TH "ค่าเริ่มต้น"
#define OPTION_VAL_AGGRESSIVE_TH "แบบรุก (Aggressive)"
#define BEETLE_OPT_WIDESCREEN_HACK_LABEL_TH "การแฮ็กโหมดจอกว้าง"
#define BEETLE_OPT_WIDESCREEN_HACK_INFO_0_TH "เรนเดอร์เนื้อหา 3D แบบอนามอร์ฟิก และแสดงผลเฟรมบัฟเฟอร์ที่จำลองขึ้นในอัตราส่วนภาพแบบจอกว้าง ให้ผลลัพธ์ที่ดีที่สุดกับเกมที่เป็น 3D เต็มรูปแบบ ส่วนองค์ประกอบที่เป็น 2D จะถูกยืดออกในแนวนอนและอาจเกิดการวางตำแหน่งที่ผิดเพี้ยนได้"
#define BEETLE_OPT_WIDESCREEN_HACK_ASPECT_RATIO_LABEL_TH "อัตราส่วนภาพของการแฮ็กโหมดจอกว้าง"
#define BEETLE_OPT_WIDESCREEN_HACK_ASPECT_RATIO_INFO_0_TH "อัตราส่วนภาพที่จะใช้กับการแฮ็กโหมดจอกว้าง"
#define OPTION_VAL_16_9_TH NULL
#define OPTION_VAL_16_10_TH NULL
#define OPTION_VAL_18_9_TH NULL
#define OPTION_VAL_19_9_TH NULL
#define OPTION_VAL_20_9_TH NULL
#define OPTION_VAL_21_9_TH NULL
#define OPTION_VAL_32_9_TH NULL
#define BEETLE_OPT_CPU_DYNAREC_LABEL_TH NULL
#define BEETLE_OPT_CPU_DYNAREC_INFO_0_TH "รีคอมไพล์คำสั่งของ CPU เป็นคำสั่งแบบ Native โดยอัตโนมัติ ซึ่งมีความเร็วสูงกว่า Interpreter มาก แต่การกำหนดจังหวะเวลา ของ CPU จะมีความแม่นยำน้อยกว่า และอาจพบข้อผิดพลาด ได้ในบางกรณี"
#define OPTION_VAL_DISABLED_TH "ปิดการใช้งาน (Beetle Interpreter)"
#define OPTION_VAL_EXECUTE_TH "ประสิทธิภาพสูงสุด"
#define OPTION_VAL_EXECUTE_ONE_TH "การตรวจสอบจังหวะเวลาของรอบ"
#define OPTION_VAL_RUN_INTERPRETER_TH NULL
#define BEETLE_OPT_DYNAREC_INVALIDATE_LABEL_TH NULL
#define BEETLE_OPT_DYNAREC_INVALIDATE_INFO_0_TH "บางเกมจำเป็นต้องใช้การล้างข้อมูลแบบ 'ทั้งหมด' บางเกมจำเป็นต้องใช้แบบ 'DMA เท่านั้น'"
#define OPTION_VAL_FULL_TH "เต็ม"
#define OPTION_VAL_DMA_TH "DMA เท่านั้น (เร็วขึ้นเล็กน้อย)"
#define BEETLE_OPT_DYNAREC_EVENTCYCLES_LABEL_TH NULL
#define BEETLE_OPT_DYNAREC_EVENTCYCLES_INFO_0_TH "จำนวนรอบสูงสุดที่ CPU จะประมวลผลก่อนที่จะมีการตรวจสอบการอัปเดต GPU/DMA/MDEC/Timer โดยค่าที่สูงขึ้นจะช่วยให้ทำงานได้เร็วขึ้น แต่จะส่งผลต่อ beetle interpreter น้อยกว่าแบบ dynarec มาก"
#define OPTION_VAL_128_TH "128 (ค่าเริ่มต้น)"
#define BEETLE_OPT_DYNAREC_SPU_SAMPLES_LABEL_TH NULL
#define BEETLE_OPT_DYNAREC_SPU_SAMPLES_INFO_0_TH "จำนวนตัวอย่าง SPU สูงสุดที่จะประมวลผลก่อนการตรวจสอบการอัปเดต SPU ยิ่งค่าสูงจะยิ่งทำงานได้เร็วขึ้น แต่อาจทำให้เกิดข้อผิดพลาดของเสียงในบางเกม หากตั้งค่าเป็นค่าอื่นที่ไม่ใช่ 1"
#define BEETLE_OPT_CORE_TIMING_FPS_LABEL_TH "การกำหนดจังหวะ FPS ตามที่คอร์รายงาน"
#define BEETLE_OPT_CORE_TIMING_FPS_INFO_0_TH "ระบบสลับอัตโนมัติช่วยให้คอร์สลับระหว่างโหมด Progressive และ Interlaced ได้เอง"
#define OPTION_VAL_AUTO_TOGGLE_TH "สลับอัตโนมัติ"
#define OPTION_VAL_FORCE_PROGRESSIVE_TH "บังคับโหมด Progressive"
#define OPTION_VAL_FORCE_INTERLACED_TH "บังคับโหมด Interlaced"
#define BEETLE_OPT_DEINTERLACER_LABEL_TH "วิธีการ Deinterlace"
#define BEETLE_OPT_DEINTERLACER_INFO_0_TH NULL
#define OPTION_VAL_WEAVE_TH NULL
#define OPTION_VAL_BOB_TH NULL
#define OPTION_VAL_BOB_OFFSET_TH NULL
#define OPTION_VAL_FASTMAD_TH NULL
#define BEETLE_OPT_FRAME_DUPING_LABEL_TH "การทำซ้ำเฟรม"
#define BEETLE_OPT_FRAME_DUPING_INFO_0_TH "แสดงเฟรมเดิมซ้ำหากไม่มีข้อมูลใหม่จากคอร์"
#define BEETLE_OPT_PAL_VIDEO_TIMING_OVERRIDE_LABEL_TH "เขียนทับจังหวะวิดีโอ PAL"
#define BEETLE_OPT_PAL_VIDEO_TIMING_OVERRIDE_INFO_0_TH "เนื่องจากมาตรฐานที่ต่างกัน เกม PAL มักจะช้ากว่าเกม NTSC ของอเมริกาหรือญี่ปุ่น ตัวเลือกนี้ใช้เขียนทับจังหวะเวลาของ PAL เพื่อพยายามรันเกมด้วยเฟรมเรตแบบ NTSC"
#define BEETLE_OPT_ASPECT_RATIO_LABEL_TH "อัตราส่วนภาพของคอร์"
#define BEETLE_OPT_ASPECT_RATIO_INFO_0_TH "จะข้ามการตั้งค่านี้ หากเปิดใช้งาน Widescreen Mode Hack หรือ Display Full VRAM อยู่"
#define OPTION_VAL_CORRECTED_TH "แก้ไขแล้ว"
#define OPTION_VAL_UNCORRECTED_TH "ยังไม่แก้ไข"
#define OPTION_VAL_4_3_TH "บังคับ 4:3"
#define OPTION_VAL_NTSC_TH "บังคับ NTSC"
#define BEETLE_OPT_CROP_OVERSCAN_LABEL_TH "ตัดขอบภาพ"
#define BEETLE_OPT_CROP_OVERSCAN_INFO_0_TH "แสดงภาพต้นฉบับเต็มหรือตัดขอบดำออก"
#define OPTION_VAL_STATIC_TH "แนวนอน"
#define OPTION_VAL_SMART_TH "แนวนอน + แนวตั้ง"
#define BEETLE_OPT_IMAGE_CROP_LABEL_TH "ตัดขอบภาพเพิ่มเติม"
#define BEETLE_OPT_IMAGE_CROP_INFO_0_TH "ลดความกว้างของภาพที่ตัดตามจำนวนพิกเซลที่กำหนด"
#define OPTION_VAL_9PX_TH NULL
#define OPTION_VAL_10PX_TH NULL
#define OPTION_VAL_11PX_TH NULL
#define OPTION_VAL_12PX_TH NULL
#define OPTION_VAL_13PX_TH NULL
#define OPTION_VAL_14PX_TH NULL
#define OPTION_VAL_15PX_TH NULL
#define OPTION_VAL_16PX_TH NULL
#define OPTION_VAL_17PX_TH NULL
#define OPTION_VAL_18PX_TH NULL
#define OPTION_VAL_19PX_TH NULL
#define OPTION_VAL_20PX_TH NULL
#define BEETLE_OPT_IMAGE_OFFSET_LABEL_TH "ชดเชยการครอปภาพ"
#define BEETLE_OPT_IMAGE_OFFSET_INFO_0_TH "ชดเชยแนวนอนไปทางขวา (บวก) หรือซ้าย (ลบ) ตามจำนวนพิกเซลที่กำหนด อาจใช้เพื่อแก้ไขปัญหาการจัดวางภาพ รองรับเฉพาะซอฟต์แวร์เรนเดอร์เลอร์เท่านั้น"
#define OPTION_VAL_12PX_O64_TH NULL
#define OPTION_VAL_11PX_O64_TH NULL
#define OPTION_VAL_10PX_O64_TH NULL
#define OPTION_VAL_9PX_O64_TH NULL
#define OPTION_VAL_8PX_O64_TH NULL
#define OPTION_VAL_7PX_O64_TH NULL
#define OPTION_VAL_6PX_O64_TH NULL
#define OPTION_VAL_5PX_O64_TH NULL
#define OPTION_VAL_4PX_O64_TH NULL
#define OPTION_VAL_3PX_O64_TH NULL
#define OPTION_VAL_2PX_O64_TH NULL
#define OPTION_VAL_1PX_O64_TH NULL
#define OPTION_VAL_1PX_O64_0_TH NULL
#define OPTION_VAL_2PX_O64_0_TH NULL
#define OPTION_VAL_3PX_O64_0_TH NULL
#define OPTION_VAL_4PX_O64_0_TH NULL
#define OPTION_VAL_5PX_O64_0_TH NULL
#define OPTION_VAL_6PX_O64_0_TH NULL
#define OPTION_VAL_7PX_O64_0_TH NULL
#define OPTION_VAL_8PX_O64_0_TH NULL
#define OPTION_VAL_9PX_O64_0_TH NULL
#define OPTION_VAL_10PX_O64_0_TH NULL
#define OPTION_VAL_11PX_O64_0_TH NULL
#define OPTION_VAL_12PX_O64_0_TH NULL
#define BEETLE_OPT_IMAGE_OFFSET_CYCLES_LABEL_TH "การชดเชยภาพแนวนอน (GPU Cycles)"
#define BEETLE_OPT_IMAGE_OFFSET_CYCLES_INFO_0_TH "ค่าบวกจะเลื่อนภาพไปทางขวา ค่าลบจะเลื่อนภาพไปทางซ้าย รองรับเฉพาะฮาร์ดแวร์เรนเดอร์เลอร์เท่านั้น"
#define BEETLE_OPT_INITIAL_SCANLINE_LABEL_TH NULL
#define BEETLE_OPT_INITIAL_SCANLINE_INFO_0_TH "บรรทัดสแกนแรกที่แสดงผลเมื่อรันเนื้อหา NTSC ค่าที่มากกว่าศูนย์จะลดความสูงของภาพที่แสดงผลโดยการครอปพิกเซลจากขอบบนสุด อาจใช้เพื่อแก้ปัญหาแถบดำด้านบนและล่าง (letterboxing)"
#define BEETLE_OPT_LAST_SCANLINE_LABEL_TH NULL
#define BEETLE_OPT_LAST_SCANLINE_INFO_0_TH "บรรทัดสแกนสุดท้ายที่แสดงผลเมื่อรันเนื้อหา NTSC ค่าที่น้อยกว่า 239 จะลดความสูงของภาพที่แสดงผลโดยการครอปพิกเซลจากขอบล่างสุด อาจใช้เพื่อแก้ปัญหาแถบดำด้านบนและล่าง (letterboxing)"
#define OPTION_VAL_239_TH "239 (ค่าเริ่มต้น)"
#define BEETLE_OPT_INITIAL_SCANLINE_PAL_LABEL_TH NULL
#define BEETLE_OPT_INITIAL_SCANLINE_PAL_INFO_0_TH "บรรทัดสแกนแรกที่แสดงผลเมื่อรันเนื้อหา PAL ค่าที่มากกว่าศูนย์จะลดความสูงของภาพที่แสดงผลโดยการครอปพิกเซลจากขอบบนสุด อาจใช้เพื่อแก้ปัญหาแถบดำด้านบนและล่าง (letterboxing)"
#define BEETLE_OPT_LAST_SCANLINE_PAL_LABEL_TH NULL
#define BEETLE_OPT_LAST_SCANLINE_PAL_INFO_0_TH "บรรทัดสแกนสุดท้ายที่แสดงผลเมื่อรันเนื้อหา PAL ค่าที่น้อยกว่า 287 จะลดความสูงของภาพที่แสดงผลโดยการครอปพิกเซลจากขอบล่างสุด อาจใช้เพื่อแก้ปัญหาแถบดำด้านบนและล่าง (letterboxing)"
#define OPTION_VAL_287_TH "287 (ค่าเริ่มต้น)"

struct retro_core_option_v2_category option_cats_th[] = {
   {
      "system",
      CATEGORY_SYSTEM_LABEL_TH,
      CATEGORY_SYSTEM_INFO_0_TH
   },
   {
      "video",
      CATEGORY_VIDEO_LABEL_TH,
      CATEGORY_VIDEO_INFO_0_TH
   },
   {
      "pgxp",
      CATEGORY_PGXP_LABEL_TH,
      CATEGORY_PGXP_INFO_0_TH
   },
   {
      "osd",
      CATEGORY_OSD_LABEL_TH,
      CATEGORY_OSD_INFO_0_TH
   },
   {
      "input",
      CATEGORY_INPUT_LABEL_TH,
      CATEGORY_INPUT_INFO_0_TH
   },
   {
      "memcards",
      CATEGORY_MEMCARDS_LABEL_TH,
      CATEGORY_MEMCARDS_INFO_0_TH
   },
   {
      "hacks",
      CATEGORY_HACKS_LABEL_TH,
      CATEGORY_HACKS_INFO_0_TH
   },
   { NULL, NULL, NULL },
};
struct retro_core_option_v2_definition option_defs_th[] = {
   {
      BEETLE_OPT(cpu_freq_scale),
      BEETLE_OPT_CPU_FREQ_SCALE_LABEL_TH,
      NULL,
      BEETLE_OPT_CPU_FREQ_SCALE_INFO_0_TH,
      NULL,
      "system",
      {
         { "50%",           OPTION_VAL_50_TH },
         { "60%",           OPTION_VAL_60_TH },
         { "70%",           OPTION_VAL_70_TH },
         { "80%",           OPTION_VAL_80_TH },
         { "90%",           OPTION_VAL_90_TH },
         { "100%",          OPTION_VAL_100_TH },
         { "110%",          OPTION_VAL_110_TH },
         { "120%",          OPTION_VAL_120_TH },
         { "130%",          OPTION_VAL_130_TH },
         { "140%",          OPTION_VAL_140_TH },
         { "150%",          OPTION_VAL_150_TH },
         { "160%",          OPTION_VAL_160_TH },
         { "170%",          OPTION_VAL_170_TH },
         { "180%",          OPTION_VAL_180_TH },
         { "190%",          OPTION_VAL_190_TH },
         { "200%",          OPTION_VAL_200_TH },
         { "210%",          OPTION_VAL_210_TH },
         { "220%",          OPTION_VAL_220_TH },
         { "230%",          OPTION_VAL_230_TH },
         { "240%",          OPTION_VAL_240_TH },
         { "250%",          OPTION_VAL_250_TH },
         { "260%",          OPTION_VAL_260_TH },
         { "270%",          OPTION_VAL_270_TH },
         { "280%",          OPTION_VAL_280_TH },
         { "290%",          OPTION_VAL_290_TH },
         { "300%",          OPTION_VAL_300_TH },
         { "310%",          OPTION_VAL_310_TH },
         { "320%",          OPTION_VAL_320_TH },
         { "330%",          OPTION_VAL_330_TH },
         { "340%",          OPTION_VAL_340_TH },
         { "350%",          OPTION_VAL_350_TH },
         { "360%",          OPTION_VAL_360_TH },
         { "370%",          OPTION_VAL_370_TH },
         { "380%",          OPTION_VAL_380_TH },
         { "390%",          OPTION_VAL_390_TH },
         { "400%",          OPTION_VAL_400_TH },
         { "410%",          OPTION_VAL_410_TH },
         { "420%",          OPTION_VAL_420_TH },
         { "430%",          OPTION_VAL_430_TH },
         { "440%",          OPTION_VAL_440_TH },
         { "450%",          OPTION_VAL_450_TH },
         { "460%",          OPTION_VAL_460_TH },
         { "470%",          OPTION_VAL_470_TH },
         { "480%",          OPTION_VAL_480_TH },
         { "490%",          OPTION_VAL_490_TH },
         { "500%",          OPTION_VAL_500_TH },
         { "510%",          OPTION_VAL_510_TH },
         { "520%",          OPTION_VAL_520_TH },
         { "530%",          OPTION_VAL_530_TH },
         { "540%",          OPTION_VAL_540_TH },
         { "550%",          OPTION_VAL_550_TH },
         { "560%",          OPTION_VAL_560_TH },
         { "570%",          OPTION_VAL_570_TH },
         { "580%",          OPTION_VAL_580_TH },
         { "590%",          OPTION_VAL_590_TH },
         { "600%",          OPTION_VAL_600_TH },
         { "610%",          OPTION_VAL_610_TH },
         { "620%",          OPTION_VAL_620_TH },
         { "630%",          OPTION_VAL_630_TH },
         { "640%",          OPTION_VAL_640_TH },
         { "650%",          OPTION_VAL_650_TH },
         { "660%",          OPTION_VAL_660_TH },
         { "670%",          OPTION_VAL_670_TH },
         { "680%",          OPTION_VAL_680_TH },
         { "690%",          OPTION_VAL_690_TH },
         { "700%",          OPTION_VAL_700_TH },
         { "710%",          OPTION_VAL_710_TH },
         { "720%",          OPTION_VAL_720_TH },
         { "730%",          OPTION_VAL_730_TH },
         { "740%",          OPTION_VAL_740_TH },
         { "750%",          OPTION_VAL_750_TH },
         { NULL, NULL },
      },
      "100%"
   },
   {
      BEETLE_OPT(gpu_overclock),
      BEETLE_OPT_GPU_OVERCLOCK_LABEL_TH,
      NULL,
      BEETLE_OPT_GPU_OVERCLOCK_INFO_0_TH,
      NULL,
      "system",
      {
         { "1x(native)", OPTION_VAL_1X_NATIVE_TH },
         { "2x",         OPTION_VAL_2X_TH },
         { "4x",         OPTION_VAL_4X_TH },
         { "8x",         OPTION_VAL_8X_TH },
         { "16x",        OPTION_VAL_16X_TH },
         { "32x",        OPTION_VAL_32X_TH },
         { NULL, NULL },
      },
      "1x(native)"
   },
   {
      BEETLE_OPT(gte_overclock),
      BEETLE_OPT_GTE_OVERCLOCK_LABEL_TH,
      NULL,
      BEETLE_OPT_GTE_OVERCLOCK_INFO_0_TH,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
#ifndef EMSCRIPTEN
   {
      BEETLE_OPT(cd_access_method),
      BEETLE_OPT_CD_ACCESS_METHOD_LABEL_TH,
      NULL,
      BEETLE_OPT_CD_ACCESS_METHOD_INFO_0_TH,
      NULL,
      "system",
      {
         { "sync",     OPTION_VAL_SYNC_TH },
         { "async",    OPTION_VAL_ASYNC_TH },
         { "precache", OPTION_VAL_PRECACHE_TH },
         { NULL, NULL },
      },
      "sync"
   },
#endif
   {
      BEETLE_OPT(cd_fastload),
      BEETLE_OPT_CD_FASTLOAD_LABEL_TH,
      NULL,
      BEETLE_OPT_CD_FASTLOAD_INFO_0_TH,
      NULL,
      "system",
      {
         { "2x(native)", OPTION_VAL_2X_NATIVE_TH },
         { "4x",          OPTION_VAL_4X_TH },
         { "6x",          OPTION_VAL_6X_TH },
         { "8x",          OPTION_VAL_8X_TH },
         { "10x",         OPTION_VAL_10X_TH },
         { "12x",         OPTION_VAL_12X_TH },
         { "14x",         OPTION_VAL_14X_TH },
         { NULL, NULL },
      },
      "2x(native)"
   },
   {
      BEETLE_OPT(skip_bios),
      BEETLE_OPT_SKIP_BIOS_LABEL_TH,
      NULL,
      BEETLE_OPT_SKIP_BIOS_INFO_0_TH,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(override_bios),
      BEETLE_OPT_OVERRIDE_BIOS_LABEL_TH,
      NULL,
      BEETLE_OPT_OVERRIDE_BIOS_INFO_0_TH,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "psxonpsp", OPTION_VAL_PSXONPSP_TH },
         { "ps1_rom",  OPTION_VAL_PS1_ROM_TH },
         { "openbios", OPTION_VAL_OPENBIOS_TH     },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(region),
      BEETLE_OPT_REGION_LABEL_TH,
      NULL,
      BEETLE_OPT_REGION_INFO_0_TH,
      NULL,
      "system",
      {
         { "auto",   OPTION_VAL_AUTO_TH },
         { "ntsc-j", OPTION_VAL_NTSC_J_TH },
         { "ntsc-u", OPTION_VAL_NTSC_U_TH },
         { "pal",    OPTION_VAL_PAL_TH },
         { NULL, NULL },
      },
      "auto"
   },
   {
      BEETLE_OPT(internal_resolution),
      BEETLE_OPT_INTERNAL_RESOLUTION_LABEL_TH,
      NULL,
      BEETLE_OPT_INTERNAL_RESOLUTION_INFO_0_TH,
      NULL,
      "video",
      {
         { "1x(native)", OPTION_VAL_1X_NATIVE_TH },
         { "2x",         OPTION_VAL_2X_TH },
         { "4x",         OPTION_VAL_4X_TH },
         { "8x",         OPTION_VAL_8X_TH },
         { "16x",        OPTION_VAL_16X_TH },
         { NULL, NULL },
      },
      "1x(native)"
   },
#if defined(HAVE_OPENGL) || defined(HAVE_OPENGLES) || defined(HAVE_VULKAN)
   {
      BEETLE_OPT(renderer),
      BEETLE_OPT_RENDERER_LABEL_TH,
      NULL,
      BEETLE_OPT_RENDERER_INFO_0_TH,
      NULL,
      "video",
      {
         { "hardware",    OPTION_VAL_HARDWARE_TH },
#if defined(HAVE_OPENGL) || defined(HAVE_OPENGLES)
         { "hardware_gl", OPTION_VAL_HARDWARE_GL_TH },
#endif
#if defined(HAVE_VULKAN)
         { "hardware_vk", OPTION_VAL_HARDWARE_VK_TH },
#endif
         { "software",    OPTION_VAL_SOFTWARE_TH },
         { NULL, NULL },
      },
      "hardware"
   },
   {
      BEETLE_OPT(renderer_software_fb),
      BEETLE_OPT_RENDERER_SOFTWARE_FB_LABEL_TH,
      NULL,
      BEETLE_OPT_RENDERER_SOFTWARE_FB_INFO_0_TH,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled"
   },
#endif
#if defined(HAVE_OPENGL) || defined(HAVE_OPENGLES)
   {
      BEETLE_OPT(depth),
      BEETLE_OPT_DEPTH_LABEL_TH,
      NULL,
      BEETLE_OPT_DEPTH_INFO_0_TH,
      NULL,
      "video",
      {
         { "16bpp(native)", OPTION_VAL_16BPP_NATIVE_TH },
         { "32bpp",         OPTION_VAL_32BPP_TH },
         { NULL, NULL },
      },
      "16bpp(native)"
   },
   // Sort of, it's more like 15-bit high color and 24-bit true color for visible output. The alpha channel is used for mask bit. Vulkan renderer uses ABGR1555_555 for 31 bits internal? FMVs are always 24-bit on all renderers like original hardware (BGR888, no alpha)
#endif
   {
      BEETLE_OPT(dither_mode),
      BEETLE_OPT_DITHER_MODE_LABEL_TH,
      NULL,
      BEETLE_OPT_DITHER_MODE_INFO_0_TH,
      NULL,
      "video",
      {
         { "1x(native)",          OPTION_VAL_1X_NATIVE_TH },
         { "internal resolution", OPTION_VAL_INTERNAL_RESOLUTION_TH },
         { "disabled",            NULL },
         { NULL, NULL },
      },
      "1x(native)"
   },
#ifdef HAVE_VULKAN
   {
      BEETLE_OPT(scaled_uv_offset),
      BEETLE_OPT_SCALED_UV_OFFSET_LABEL_TH,
      NULL,
      BEETLE_OPT_SCALED_UV_OFFSET_INFO_0_TH,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled"
   },
#endif
#if defined(HAVE_OPENGL) || defined(HAVE_OPENGLES) || defined(HAVE_VULKAN)
   {
      BEETLE_OPT(filter),
      BEETLE_OPT_FILTER_LABEL_TH,
      NULL,
      BEETLE_OPT_FILTER_INFO_0_TH,
      NULL,
      "video",
      {
         { "nearest",  OPTION_VAL_NEAREST_TH },
         { "SABR",     OPTION_VAL_SABR_TH },
         { "xBR",      OPTION_VAL_XBR_TH },
         { "bilinear", OPTION_VAL_BILINEAR_TH },
         { "3-point",  OPTION_VAL_3_POINT_TH },
         { "JINC2",    OPTION_VAL_JINC2_TH },
         { NULL, NULL },
      },
      "nearest"
   },
#ifdef HAVE_VULKAN
   {
      BEETLE_OPT(filter_exclude_sprite),
      BEETLE_OPT_FILTER_EXCLUDE_SPRITE_LABEL_TH,
      NULL,
      BEETLE_OPT_FILTER_EXCLUDE_SPRITE_INFO_0_TH,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "opaque", OPTION_VAL_OPAQUE_TH },
         { "all", OPTION_VAL_ALL_TH },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(filter_exclude_2d_polygon),
      BEETLE_OPT_FILTER_EXCLUDE_2D_POLYGON_LABEL_TH,
      NULL,
      BEETLE_OPT_FILTER_EXCLUDE_2D_POLYGON_INFO_0_TH,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "opaque", OPTION_VAL_OPAQUE_TH },
         { "all", OPTION_VAL_ALL_TH },
         { NULL, NULL },
      },
      "disabled"
   },
#endif
#endif
#ifdef HAVE_VULKAN
   {
      BEETLE_OPT(adaptive_smoothing),
      BEETLE_OPT_ADAPTIVE_SMOOTHING_LABEL_TH,
      NULL,
      BEETLE_OPT_ADAPTIVE_SMOOTHING_INFO_0_TH,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(super_sampling),
      BEETLE_OPT_SUPER_SAMPLING_LABEL_TH,
      NULL,
      BEETLE_OPT_SUPER_SAMPLING_INFO_0_TH,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(msaa),
      BEETLE_OPT_MSAA_LABEL_TH,
      NULL,
      BEETLE_OPT_MSAA_INFO_0_TH,
      NULL,
      "video",
      {
         { "1x",  OPTION_VAL_1X_TH },
         { "2x",  OPTION_VAL_2X_TH },
         { "4x",  OPTION_VAL_4X_TH },
         { "8x",  OPTION_VAL_8X_TH },
         { "16x", OPTION_VAL_16X_TH },
         { NULL, NULL },
      },
      "1x"
   },
   {
      BEETLE_OPT(mdec_yuv),
      BEETLE_OPT_MDEC_YUV_LABEL_TH,
      NULL,
      BEETLE_OPT_MDEC_YUV_INFO_0_TH,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(track_textures),
      BEETLE_OPT_TRACK_TEXTURES_LABEL_TH,
      NULL,
      BEETLE_OPT_TRACK_TEXTURES_INFO_0_TH,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
#ifdef TEXTURE_DUMPING_ENABLED
   {
      BEETLE_OPT(dump_textures),
      BEETLE_OPT_DUMP_TEXTURES_LABEL_TH,
      NULL,
      BEETLE_OPT_DUMP_TEXTURES_INFO_0_TH,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
#endif
   {
      BEETLE_OPT(replace_textures),
      BEETLE_OPT_REPLACE_TEXTURES_LABEL_TH,
      NULL,
      BEETLE_OPT_REPLACE_TEXTURES_INFO_0_TH,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
#endif
   {
      BEETLE_OPT(pgxp_mode),
      BEETLE_OPT_PGXP_MODE_LABEL_TH,
      NULL,
      BEETLE_OPT_PGXP_MODE_INFO_0_TH,
      NULL,
      "pgxp",
      {
         { "disabled",     NULL },
         { "memory only",  OPTION_VAL_MEMORY_ONLY_TH },
         { "memory + CPU", OPTION_VAL_MEMORY_CPU_TH },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(pgxp_2d_tol),
      BEETLE_OPT_PGXP_2D_TOL_LABEL_TH,
      NULL,
      BEETLE_OPT_PGXP_2D_TOL_INFO_0_TH,
      NULL,
      "pgxp",
      {
         { "disabled", NULL },
         { "0px", OPTION_VAL_0PX_TH },
         { "1px", OPTION_VAL_1PX_TH },
         { "2px", OPTION_VAL_2PX_TH },
         { "3px", OPTION_VAL_3PX_TH },
         { "4px", OPTION_VAL_4PX_TH },
         { "5px", OPTION_VAL_5PX_TH },
         { "6px", OPTION_VAL_6PX_TH },
         { "7px", OPTION_VAL_7PX_TH },
         { "8px", OPTION_VAL_8PX_TH },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(pgxp_nclip),
      BEETLE_OPT_PGXP_NCLIP_LABEL_TH,
      NULL,
      BEETLE_OPT_PGXP_NCLIP_INFO_0_TH,
      NULL,
      "pgxp",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
#if defined(HAVE_OPENGL) || defined(HAVE_OPENGLES) || defined(HAVE_VULKAN)
   {
      BEETLE_OPT(pgxp_vertex),
      BEETLE_OPT_PGXP_VERTEX_LABEL_TH,
      NULL,
      BEETLE_OPT_PGXP_VERTEX_INFO_0_TH,
      NULL,
      "pgxp",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
#endif
   {
      BEETLE_OPT(pgxp_texture),
      BEETLE_OPT_PGXP_TEXTURE_LABEL_TH,
      NULL,
      BEETLE_OPT_PGXP_TEXTURE_INFO_0_TH,
      NULL,
      "pgxp",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(display_internal_fps),
      BEETLE_OPT_DISPLAY_INTERNAL_FPS_LABEL_TH,
      NULL,
      BEETLE_OPT_DISPLAY_INTERNAL_FPS_INFO_0_TH,
      NULL,
      "osd",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(display_osd),
      BEETLE_OPT_DISPLAY_OSD_LABEL_TH,
      NULL,
      BEETLE_OPT_DISPLAY_OSD_INFO_0_TH,
      NULL,
      "osd",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled"
   },
#if defined(HAVE_OPENGL) || defined(HAVE_OPENGLES) || defined(HAVE_VULKAN)
   {
      BEETLE_OPT(display_vram),
      BEETLE_OPT_DISPLAY_VRAM_LABEL_TH,
      NULL,
      BEETLE_OPT_DISPLAY_VRAM_INFO_0_TH,
      NULL,
      "osd",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
#endif
   {
      BEETLE_OPT(analog_calibration),
      BEETLE_OPT_ANALOG_CALIBRATION_LABEL_TH,
      NULL,
      BEETLE_OPT_ANALOG_CALIBRATION_INFO_0_TH,
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(analog_toggle),
      BEETLE_OPT_ANALOG_TOGGLE_LABEL_TH,
      NULL,
      BEETLE_OPT_ANALOG_TOGGLE_INFO_0_TH,
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { "enabled-analog", OPTION_VAL_ENABLED_ANALOG_TH },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(analog_toggle_combo),
      BEETLE_OPT_ANALOG_TOGGLE_COMBO_LABEL_TH,
      NULL,
      BEETLE_OPT_ANALOG_TOGGLE_COMBO_INFO_0_TH,
      NULL,
      "input",
      {
         { "l1+l2+r1+r2+start+select", OPTION_VAL_L1_L2_R1_R2_START_SELECT_TH },
         { "l1+r1+select",             OPTION_VAL_L1_R1_SELECT_TH },
         { "l1+r1+start",              OPTION_VAL_L1_R1_START_TH },
         { "l1+r1+l3",                 OPTION_VAL_L1_R1_L3_TH },
         { "l1+r1+r3",                 OPTION_VAL_L1_R1_R3_TH },
         { "l2+r2+select",             OPTION_VAL_L2_R2_SELECT_TH },
         { "l2+r2+start",              OPTION_VAL_L2_R2_START_TH },
         { "l2+r2+l3",                 OPTION_VAL_L2_R2_L3_TH },
         { "l2+r2+r3",                 OPTION_VAL_L2_R2_R3_TH },
         { "l3+r3",                    OPTION_VAL_L3_R3_TH },
         { NULL, NULL },
      },
      "l1+r1+select"
   },
   {
      BEETLE_OPT(analog_toggle_hold),
      BEETLE_OPT_ANALOG_TOGGLE_HOLD_LABEL_TH,
      NULL,
      BEETLE_OPT_ANALOG_TOGGLE_HOLD_INFO_0_TH,
      NULL,
      "input",
      {
         { "0", OPTION_VAL_0_TH },
         { "1", OPTION_VAL_1_TH },
         { "2", OPTION_VAL_2_TH },
         { "3", OPTION_VAL_3_TH },
         { "4", OPTION_VAL_4_TH },
         { "5", OPTION_VAL_5_TH },
         { NULL, NULL },
      },
      "1"
   },
   {
      BEETLE_OPT(enable_multitap_port1),
      BEETLE_OPT_ENABLE_MULTITAP_PORT1_LABEL_TH,
      NULL,
      BEETLE_OPT_ENABLE_MULTITAP_PORT1_INFO_0_TH,
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(enable_multitap_port2),
      BEETLE_OPT_ENABLE_MULTITAP_PORT2_LABEL_TH,
      NULL,
      BEETLE_OPT_ENABLE_MULTITAP_PORT2_INFO_0_TH,
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(gun_input_mode),
      BEETLE_OPT_GUN_INPUT_MODE_LABEL_TH,
      NULL,
      BEETLE_OPT_GUN_INPUT_MODE_INFO_0_TH,
      NULL,
      "input",
      {
         { "lightgun",    OPTION_VAL_LIGHTGUN_TH },
         { "touchscreen", OPTION_VAL_TOUCHSCREEN_TH },
         { NULL, NULL },
      },
      "lightgun"
   },
   // Shouldn't the gun_input_mode just be Mouse vs. Touchscreen?
   {
      BEETLE_OPT(gun_cursor),
      BEETLE_OPT_GUN_CURSOR_LABEL_TH,
      NULL,
      BEETLE_OPT_GUN_CURSOR_INFO_0_TH,
      NULL,
      "input",
      {
         { "off",   "disabled" },
         { "cross", OPTION_VAL_CROSS_TH },
         { "dot",   OPTION_VAL_DOT_TH },
         { NULL, NULL },
      },
      "cross"
   },
   {
      BEETLE_OPT(crosshair_color_p1),
      BEETLE_OPT_CROSSHAIR_COLOR_P1_LABEL_TH,
      NULL,
      BEETLE_OPT_CROSSHAIR_COLOR_P1_INFO_0_TH,
      NULL,
      "input",
      {
         { "red",    OPTION_VAL_RED_TH },
         { "blue",   OPTION_VAL_BLUE_TH },
         { "green",  OPTION_VAL_GREEN_TH },
         { "orange", OPTION_VAL_ORANGE_TH },
         { "yellow", OPTION_VAL_YELLOW_TH },
         { "cyan",   OPTION_VAL_CYAN_TH },
         { "pink",   OPTION_VAL_PINK_TH },
         { "purple", OPTION_VAL_PURPLE_TH },
         { "black",  OPTION_VAL_BLACK_TH },
         { "white",  OPTION_VAL_WHITE_TH },
         { NULL, NULL },
      },
      "red"
   },
   {
      BEETLE_OPT(crosshair_color_p2),
      BEETLE_OPT_CROSSHAIR_COLOR_P2_LABEL_TH,
      NULL,
      BEETLE_OPT_CROSSHAIR_COLOR_P2_INFO_0_TH,
      NULL,
      "input",
      {
         { "blue",   OPTION_VAL_BLUE_TH },
         { "red",    OPTION_VAL_RED_TH },
         { "green",  OPTION_VAL_GREEN_TH },
         { "orange", OPTION_VAL_ORANGE_TH },
         { "yellow", OPTION_VAL_YELLOW_TH },
         { "cyan",   OPTION_VAL_CYAN_TH },
         { "pink",   OPTION_VAL_PINK_TH },
         { "purple", OPTION_VAL_PURPLE_TH },
         { "black",  OPTION_VAL_BLACK_TH },
         { "white",  OPTION_VAL_WHITE_TH },
         { NULL, NULL },
      },
      "blue"
   },
   {
      BEETLE_OPT(mouse_sensitivity),
      BEETLE_OPT_MOUSE_SENSITIVITY_LABEL_TH,
      NULL,
      BEETLE_OPT_MOUSE_SENSITIVITY_INFO_0_TH,
      NULL,
      "input",
      {
         { "5%",   OPTION_VAL_5_O42_TH },
         { "10%",  OPTION_VAL_10_TH },
         { "15%",  OPTION_VAL_15_TH },
         { "20%",  OPTION_VAL_20_TH },
         { "25%",  OPTION_VAL_25_TH },
         { "30%",  OPTION_VAL_30_TH },
         { "35%",  OPTION_VAL_35_TH },
         { "40%",  OPTION_VAL_40_TH },
         { "45%",  OPTION_VAL_45_TH },
         { "50%",  OPTION_VAL_50_TH },
         { "55%",  OPTION_VAL_55_TH },
         { "60%",  OPTION_VAL_60_TH },
         { "65%",  OPTION_VAL_65_TH },
         { "70%",  OPTION_VAL_70_TH },
         { "75%",  OPTION_VAL_75_TH },
         { "80%",  OPTION_VAL_80_TH },
         { "85%",  OPTION_VAL_85_TH },
         { "90%",  OPTION_VAL_90_TH },
         { "95%",  OPTION_VAL_95_TH },
         { "100%", OPTION_VAL_100_O42_TH },
         { "105%", OPTION_VAL_105_TH },
         { "110%", OPTION_VAL_110_TH },
         { "115%", OPTION_VAL_115_TH },
         { "120%", OPTION_VAL_120_TH },
         { "125%", OPTION_VAL_125_TH },
         { "130%", OPTION_VAL_130_TH },
         { "135%", OPTION_VAL_135_TH },
         { "140%", OPTION_VAL_140_TH },
         { "145%", OPTION_VAL_145_TH },
         { "150%", OPTION_VAL_150_TH },
         { "155%", OPTION_VAL_155_TH },
         { "160%", OPTION_VAL_160_TH },
         { "165%", OPTION_VAL_165_TH },
         { "170%", OPTION_VAL_170_TH },
         { "175%", OPTION_VAL_175_TH },
         { "180%", OPTION_VAL_180_TH },
         { "185%", OPTION_VAL_185_TH },
         { "190%", OPTION_VAL_190_TH },
         { "195%", OPTION_VAL_195_TH },
         { "200%", OPTION_VAL_200_TH },
         { NULL, NULL },
      },
      "100%"
   },
   {
      BEETLE_OPT(negcon_response),
      BEETLE_OPT_NEGCON_RESPONSE_LABEL_TH,
      NULL,
      BEETLE_OPT_NEGCON_RESPONSE_INFO_0_TH,
      NULL,
      "input",
      {
         { "linear",    OPTION_VAL_LINEAR_TH },
         { "quadratic", OPTION_VAL_QUADRATIC_TH },
         { "cubic",     OPTION_VAL_CUBIC_TH },
         { NULL, NULL },
      },
      "linear"
   },
   {
      BEETLE_OPT(negcon_deadzone),
      BEETLE_OPT_NEGCON_DEADZONE_LABEL_TH,
      NULL,
      BEETLE_OPT_NEGCON_DEADZONE_INFO_0_TH,
      NULL,
      "input",
      {
         { "0%",  OPTION_VAL_0_O44_TH },
         { "5%",  OPTION_VAL_5_O42_TH },
         { "10%", OPTION_VAL_10_TH },
         { "15%", OPTION_VAL_15_TH },
         { "20%", OPTION_VAL_20_TH },
         { "25%", OPTION_VAL_25_TH },
         { "30%", OPTION_VAL_30_TH },
         { NULL, NULL },
      },
      "0%"
   },
   {
      BEETLE_OPT(use_mednafen_memcard0_method),
      BEETLE_OPT_USE_MEDNAFEN_MEMCARD0_METHOD_LABEL_TH,
      NULL,
      BEETLE_OPT_USE_MEDNAFEN_MEMCARD0_METHOD_INFO_0_TH,
      NULL,
      "memcards",
      {
         { "libretro", OPTION_VAL_LIBRETRO_TH },
         { "mednafen", OPTION_VAL_MEDNAFEN_TH },
         { NULL, NULL },
      },
      "libretro"
   },
   {
      BEETLE_OPT(enable_memcard1),
      BEETLE_OPT_ENABLE_MEMCARD1_LABEL_TH,
      NULL,
      BEETLE_OPT_ENABLE_MEMCARD1_INFO_0_TH,
      NULL,
      "memcards",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(shared_memory_cards),
      BEETLE_OPT_SHARED_MEMORY_CARDS_LABEL_TH,
      NULL,
      BEETLE_OPT_SHARED_MEMORY_CARDS_INFO_0_TH,
      NULL,
      "memcards",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(memcard_left_index),
      BEETLE_OPT_MEMCARD_LEFT_INDEX_LABEL_TH,
      NULL,
      BEETLE_OPT_MEMCARD_LEFT_INDEX_INFO_0_TH,
      NULL,
      "memcards",
      {
         { "0",  OPTION_VAL_0_O48_TH },
         { "1",  NULL },
         { "2",  NULL },
         { "3",  NULL },
         { "4",  NULL },
         { "5",  NULL },
         { "6",  NULL },
         { "7",  NULL },
         { "8",  NULL },
         { "9",  NULL },
         { "10", NULL },
         { "11", NULL },
         { "12", NULL },
         { "13", NULL },
         { "14", NULL },
         { "15", NULL },
         { "16", NULL },
         { "17", NULL },
         { "18", NULL },
         { "19", NULL },
         { "20", NULL },
         { "21", NULL },
         { "22", NULL },
         { "23", NULL },
         { "24", NULL },
         { "25", NULL },
         { "26", NULL },
         { "27", NULL },
         { "28", NULL },
         { "29", NULL },
         { "30", NULL },
         { "31", NULL },
         { "32", NULL },
         { "33", NULL },
         { "34", NULL },
         { "35", NULL },
         { "36", NULL },
         { "37", NULL },
         { "38", NULL },
         { "39", NULL },
         { "40", NULL },
         { "41", NULL },
         { "42", NULL },
         { "43", NULL },
         { "44", NULL },
         { "45", NULL },
         { "46", NULL },
         { "47", NULL },
         { "48", NULL },
         { "49", NULL },
         { "50", NULL },
         { "51", NULL },
         { "52", NULL },
         { "53", NULL },
         { "54", NULL },
         { "55", NULL },
         { "56", NULL },
         { "57", NULL },
         { "58", NULL },
         { "59", NULL },
         { "60", NULL },
         { "61", NULL },
         { "62", NULL },
         { "63", NULL },
         { NULL, NULL },
      },
      "0"
   },
   {
      BEETLE_OPT(memcard_right_index),
      BEETLE_OPT_MEMCARD_RIGHT_INDEX_LABEL_TH,
      NULL,
      BEETLE_OPT_MEMCARD_RIGHT_INDEX_INFO_0_TH,
      NULL,
      "memcards",
      {
         { "0",  NULL },
         { "1",  OPTION_VAL_1_O49_TH },
         { "2",  NULL },
         { "3",  NULL },
         { "4",  NULL },
         { "5",  NULL },
         { "6",  NULL },
         { "7",  NULL },
         { "8",  NULL },
         { "9",  NULL },
         { "10", NULL },
         { "11", NULL },
         { "12", NULL },
         { "13", NULL },
         { "14", NULL },
         { "15", NULL },
         { "16", NULL },
         { "17", NULL },
         { "18", NULL },
         { "19", NULL },
         { "20", NULL },
         { "21", NULL },
         { "22", NULL },
         { "23", NULL },
         { "24", NULL },
         { "25", NULL },
         { "26", NULL },
         { "27", NULL },
         { "28", NULL },
         { "29", NULL },
         { "30", NULL },
         { "31", NULL },
         { "32", NULL },
         { "33", NULL },
         { "34", NULL },
         { "35", NULL },
         { "36", NULL },
         { "37", NULL },
         { "38", NULL },
         { "39", NULL },
         { "40", NULL },
         { "41", NULL },
         { "42", NULL },
         { "43", NULL },
         { "44", NULL },
         { "45", NULL },
         { "46", NULL },
         { "47", NULL },
         { "48", NULL },
         { "49", NULL },
         { "50", NULL },
         { "51", NULL },
         { "52", NULL },
         { "53", NULL },
         { "54", NULL },
         { "55", NULL },
         { "56", NULL },
         { "57", NULL },
         { "58", NULL },
         { "59", NULL },
         { "60", NULL },
         { "61", NULL },
         { "62", NULL },
         { "63", NULL },
         { NULL, NULL },
      },
      "1"
   },
   {
      BEETLE_OPT(line_render),
      BEETLE_OPT_LINE_RENDER_LABEL_TH,
      NULL,
      BEETLE_OPT_LINE_RENDER_INFO_0_TH,
      NULL,
      "hacks",
      {
         { "default",    OPTION_VAL_DEFAULT_TH },
         { "aggressive", OPTION_VAL_AGGRESSIVE_TH },
         { "disabled",   NULL },
         { NULL, NULL },
      },
      "default"
   },
   {
      BEETLE_OPT(widescreen_hack),
      BEETLE_OPT_WIDESCREEN_HACK_LABEL_TH,
      NULL,
      BEETLE_OPT_WIDESCREEN_HACK_INFO_0_TH,
      NULL,
      "hacks",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(widescreen_hack_aspect_ratio),
      BEETLE_OPT_WIDESCREEN_HACK_ASPECT_RATIO_LABEL_TH,
      NULL,
      BEETLE_OPT_WIDESCREEN_HACK_ASPECT_RATIO_INFO_0_TH,
      NULL,
      "hacks",
      {
         { "16:9",  OPTION_VAL_16_9_TH },
         { "16:10", OPTION_VAL_16_10_TH },
         { "18:9",  OPTION_VAL_18_9_TH },
         { "19:9",  OPTION_VAL_19_9_TH },
         { "20:9",  OPTION_VAL_20_9_TH },
         { "21:9",  OPTION_VAL_21_9_TH }, // 64:27
         { "32:9",  OPTION_VAL_32_9_TH },
         { NULL,    NULL },
      },
      "16:9"
   },
#if defined(HAVE_LIGHTREC)
   {
      BEETLE_OPT(cpu_dynarec),
      BEETLE_OPT_CPU_DYNAREC_LABEL_TH,
      NULL,
      BEETLE_OPT_CPU_DYNAREC_INFO_0_TH,
      NULL,
      "hacks",
      {
         { "disabled",        OPTION_VAL_DISABLED_TH },
         { "execute",         OPTION_VAL_EXECUTE_TH },
         { "execute_one",     OPTION_VAL_EXECUTE_ONE_TH },
         { "run_interpreter", OPTION_VAL_RUN_INTERPRETER_TH },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(dynarec_invalidate),
      BEETLE_OPT_DYNAREC_INVALIDATE_LABEL_TH,
      NULL,
      BEETLE_OPT_DYNAREC_INVALIDATE_INFO_0_TH,
      NULL,
      "hacks",
      {
         { "full", OPTION_VAL_FULL_TH },
         { "dma",  OPTION_VAL_DMA_TH },
         { NULL, NULL },
      },
      "full"
   },
   {
      BEETLE_OPT(dynarec_eventcycles),
      BEETLE_OPT_DYNAREC_EVENTCYCLES_LABEL_TH,
      NULL,
      BEETLE_OPT_DYNAREC_EVENTCYCLES_INFO_0_TH,
      NULL,
      "hacks",
      {
         { "128",  OPTION_VAL_128_TH },
         { "256",  NULL },
         { "384",  NULL },
         { "512",  NULL },
         { "640",  NULL },
         { "768",  NULL },
         { "896",  NULL },
         { "1024", NULL },
         { "1152", NULL },
         { "1280", NULL },
         { "1408", NULL },
         { "1536", NULL },
         { "1664", NULL },
         { "1792", NULL },
         { "1920", NULL },
         { "2048", NULL },
         { NULL, NULL },
      },
      "128"
   },
   {
      BEETLE_OPT(dynarec_spu_samples),
      BEETLE_OPT_DYNAREC_SPU_SAMPLES_LABEL_TH,
      NULL,
      BEETLE_OPT_DYNAREC_SPU_SAMPLES_INFO_0_TH,
      NULL,
      "hacks",
      {
         { "1",  OPTION_VAL_1_O49_TH },
         { "4",  NULL },
         { "16", NULL },
         { NULL, NULL },
      },
      "1"
   },
#endif
   {
      BEETLE_OPT(core_timing_fps),
      BEETLE_OPT_CORE_TIMING_FPS_LABEL_TH,
      NULL,
      BEETLE_OPT_CORE_TIMING_FPS_INFO_0_TH,
      NULL,
      "video",
      {
         { "auto_toggle",       OPTION_VAL_AUTO_TOGGLE_TH },
         { "force_progressive", OPTION_VAL_FORCE_PROGRESSIVE_TH },
         { "force_interlaced",  OPTION_VAL_FORCE_INTERLACED_TH },
      },
      "auto_toggle"
   },
   {
      BEETLE_OPT(deinterlacer),
      BEETLE_OPT_DEINTERLACER_LABEL_TH,
      NULL,
      BEETLE_OPT_DEINTERLACER_INFO_0_TH,
      NULL,
      "video",
      {
         { "weave",      OPTION_VAL_WEAVE_TH },
         { "bob",        OPTION_VAL_BOB_TH },
         { "bob_offset", OPTION_VAL_BOB_OFFSET_TH },
         { "fastmad",    OPTION_VAL_FASTMAD_TH },
         { "off",        "Off" },
         { NULL, NULL },
      },
      "weave"
   },
   {
      BEETLE_OPT(frame_duping),
      BEETLE_OPT_FRAME_DUPING_LABEL_TH,
      NULL,
      BEETLE_OPT_FRAME_DUPING_INFO_0_TH,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(pal_video_timing_override),
      BEETLE_OPT_PAL_VIDEO_TIMING_OVERRIDE_LABEL_TH,
      NULL,
      BEETLE_OPT_PAL_VIDEO_TIMING_OVERRIDE_INFO_0_TH,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(aspect_ratio),
      BEETLE_OPT_ASPECT_RATIO_LABEL_TH,
      NULL,
      BEETLE_OPT_ASPECT_RATIO_INFO_0_TH,
      NULL,
      "video",
      {
         { "corrected", OPTION_VAL_CORRECTED_TH },
         { "uncorrected", OPTION_VAL_UNCORRECTED_TH },
         { "4:3",  OPTION_VAL_4_3_TH },
         { "ntsc", OPTION_VAL_NTSC_TH },
      },
      "corrected"
   },
   {
      BEETLE_OPT(crop_overscan),
      BEETLE_OPT_CROP_OVERSCAN_LABEL_TH,
      NULL,
      BEETLE_OPT_CROP_OVERSCAN_INFO_0_TH,
      NULL,
      "video",
      {
         { "disabled", "Disabled" },
         { "static",   OPTION_VAL_STATIC_TH },
         { "smart",    OPTION_VAL_SMART_TH },
         { NULL, NULL },
      },
      "smart"
   },
   {
      BEETLE_OPT(image_crop),
      BEETLE_OPT_IMAGE_CROP_LABEL_TH,
      NULL,
      BEETLE_OPT_IMAGE_CROP_INFO_0_TH,
      NULL,
      "video",
      {
         { "disabled", "0" },
         { "1px",      OPTION_VAL_1PX_TH },
         { "2px",      OPTION_VAL_2PX_TH },
         { "3px",      OPTION_VAL_3PX_TH },
         { "4px",      OPTION_VAL_4PX_TH },
         { "5px",      OPTION_VAL_5PX_TH },
         { "6px",      OPTION_VAL_6PX_TH },
         { "7px",      OPTION_VAL_7PX_TH },
         { "8px",      OPTION_VAL_8PX_TH },
         { "9px",      OPTION_VAL_9PX_TH },
         { "10px",     OPTION_VAL_10PX_TH },
         { "11px",     OPTION_VAL_11PX_TH },
         { "12px",     OPTION_VAL_12PX_TH },
         { "13px",     OPTION_VAL_13PX_TH },
         { "14px",     OPTION_VAL_14PX_TH },
         { "15px",     OPTION_VAL_15PX_TH },
         { "16px",     OPTION_VAL_16PX_TH },
         { "17px",     OPTION_VAL_17PX_TH },
         { "18px",     OPTION_VAL_18PX_TH },
         { "19px",     OPTION_VAL_19PX_TH },
         { "20px",     OPTION_VAL_20PX_TH },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(image_offset),
      BEETLE_OPT_IMAGE_OFFSET_LABEL_TH,
      NULL,
      BEETLE_OPT_IMAGE_OFFSET_INFO_0_TH,
      NULL,
      "video",
      {
         { "-12px",    OPTION_VAL_12PX_O64_TH },
         { "-11px",    OPTION_VAL_11PX_O64_TH },
         { "-10px",    OPTION_VAL_10PX_O64_TH },
         { "-9px",     OPTION_VAL_9PX_O64_TH },
         { "-8px",     OPTION_VAL_8PX_O64_TH },
         { "-7px",     OPTION_VAL_7PX_O64_TH },
         { "-6px",     OPTION_VAL_6PX_O64_TH },
         { "-5px",     OPTION_VAL_5PX_O64_TH },
         { "-4px",     OPTION_VAL_4PX_O64_TH },
         { "-3px",     OPTION_VAL_3PX_O64_TH },
         { "-2px",     OPTION_VAL_2PX_O64_TH },
         { "-1px",     OPTION_VAL_1PX_O64_TH },
         { "disabled", "0" },
         { "+1px",     OPTION_VAL_1PX_O64_0_TH },
         { "+2px",     OPTION_VAL_2PX_O64_0_TH },
         { "+3px",     OPTION_VAL_3PX_O64_0_TH },
         { "+4px",     OPTION_VAL_4PX_O64_0_TH },
         { "+5px",     OPTION_VAL_5PX_O64_0_TH },
         { "+6px",     OPTION_VAL_6PX_O64_0_TH },
         { "+7px",     OPTION_VAL_7PX_O64_0_TH },
         { "+8px",     OPTION_VAL_8PX_O64_0_TH },
         { "+9px",     OPTION_VAL_9PX_O64_0_TH },
         { "+10px",    OPTION_VAL_10PX_O64_0_TH },
         { "+11px",    OPTION_VAL_11PX_O64_0_TH },
         { "+12px",    OPTION_VAL_12PX_O64_0_TH },
         { NULL, NULL },
      },
      "disabled"
   },
#if defined(HAVE_OPENGL) || defined(HAVE_OPENGLES) || defined(HAVE_VULKAN)
   {
      BEETLE_OPT(image_offset_cycles),
      BEETLE_OPT_IMAGE_OFFSET_CYCLES_LABEL_TH,
      NULL,
      BEETLE_OPT_IMAGE_OFFSET_CYCLES_INFO_0_TH,
      NULL,
      "video",
      {
         { "-40",      NULL },
         { "-39",      NULL },
         { "-38",      NULL },
         { "-37",      NULL },
         { "-36",      NULL },
         { "-35",      NULL },
         { "-34",      NULL },
         { "-33",      NULL },
         { "-32",      NULL },
         { "-31",      NULL },
         { "-30",      NULL },
         { "-29",      NULL },
         { "-28",      NULL },
         { "-27",      NULL },
         { "-26",      NULL },
         { "-25",      NULL },
         { "-24",      NULL },
         { "-23",      NULL },
         { "-22",      NULL },
         { "-21",      NULL },
         { "-20",      NULL },
         { "-19",      NULL },
         { "-18",      NULL },
         { "-17",      NULL },
         { "-16",      NULL },
         { "-15",      NULL },
         { "-14",      NULL },
         { "-13",      NULL },
         { "-12",      NULL },
         { "-11",      NULL },
         { "-10",      NULL },
         { "-9",       NULL },
         { "-8",       NULL },
         { "-7",       NULL },
         { "-6",       NULL },
         { "-5",       NULL },
         { "-4",       NULL },
         { "-3",       NULL },
         { "-2",       NULL },
         { "-1",       NULL },
         { "0",        "0" },
         { "+1",       NULL },
         { "+2",       NULL },
         { "+3",       NULL },
         { "+4",       NULL },
         { "+5",       NULL },
         { "+6",       NULL },
         { "+7",       NULL },
         { "+8",       NULL },
         { "+9",       NULL },
         { "+10",      NULL },
         { "+11",      NULL },
         { "+12",      NULL },
         { "+13",      NULL },
         { "+14",      NULL },
         { "+15",      NULL },
         { "+16",      NULL },
         { "+17",      NULL },
         { "+18",      NULL },
         { "+19",      NULL },
         { "+20",      NULL },
         { "+21",      NULL },
         { "+22",      NULL },
         { "+23",      NULL },
         { "+24",      NULL },
         { "+25",      NULL },
         { "+26",      NULL },
         { "+27",      NULL },
         { "+28",      NULL },
         { "+29",      NULL },
         { "+30",      NULL },
         { "+31",      NULL },
         { "+32",      NULL },
         { "+33",      NULL },
         { "+34",      NULL },
         { "+35",      NULL },
         { "+36",      NULL },
         { "+37",      NULL },
         { "+38",      NULL },
         { "+39",      NULL },
         { "+40",      NULL },
         { NULL, NULL},
      },
      "0"
   },
#endif
   {
      BEETLE_OPT(initial_scanline),
      BEETLE_OPT_INITIAL_SCANLINE_LABEL_TH,
      NULL,
      BEETLE_OPT_INITIAL_SCANLINE_INFO_0_TH,
      NULL,
      "video",
      {
         { "0",  OPTION_VAL_0_O48_TH },
         { "1",  NULL },
         { "2",  NULL },
         { "3",  NULL },
         { "4",  NULL },
         { "5",  NULL },
         { "6",  NULL },
         { "7",  NULL },
         { "8",  NULL },
         { "9",  NULL },
         { "10", NULL },
         { "11", NULL },
         { "12", NULL },
         { "13", NULL },
         { "14", NULL },
         { "15", NULL },
         { "16", NULL },
         { "17", NULL },
         { "18", NULL },
         { "19", NULL },
         { "20", NULL },
         { "21", NULL },
         { "22", NULL },
         { "23", NULL },
         { "24", NULL },
         { "25", NULL },
         { "26", NULL },
         { "27", NULL },
         { "28", NULL },
         { "29", NULL },
         { "30", NULL },
         { "31", NULL },
         { "32", NULL },
         { "33", NULL },
         { "34", NULL },
         { "35", NULL },
         { "36", NULL },
         { "37", NULL },
         { "38", NULL },
         { "39", NULL },
         { "40", NULL },
         { NULL, NULL },
      },
      "0"
   },
   {
      BEETLE_OPT(last_scanline),
      BEETLE_OPT_LAST_SCANLINE_LABEL_TH,
      NULL,
      BEETLE_OPT_LAST_SCANLINE_INFO_0_TH,
      NULL,
      "video",
      {
         { "210", NULL },
         { "211", NULL },
         { "212", NULL },
         { "213", NULL },
         { "214", NULL },
         { "215", NULL },
         { "216", NULL },
         { "217", NULL },
         { "218", NULL },
         { "219", NULL },
         { "220", NULL },
         { "221", NULL },
         { "222", NULL },
         { "223", NULL },
         { "224", NULL },
         { "225", NULL },
         { "226", NULL },
         { "227", NULL },
         { "228", NULL },
         { "229", NULL },
         { "230", NULL },
         { "231", NULL },
         { "232", NULL },
         { "233", NULL },
         { "234", NULL },
         { "235", NULL },
         { "236", NULL },
         { "237", NULL },
         { "238", NULL },
         { "239", OPTION_VAL_239_TH },
         { NULL, NULL },
      },
      "239"
   },
   {
      BEETLE_OPT(initial_scanline_pal),
      BEETLE_OPT_INITIAL_SCANLINE_PAL_LABEL_TH,
      NULL,
      BEETLE_OPT_INITIAL_SCANLINE_PAL_INFO_0_TH,
      NULL,
      "video",
      {
         { "0",  OPTION_VAL_0_O48_TH },
         { "1",  NULL },
         { "2",  NULL },
         { "3",  NULL },
         { "4",  NULL },
         { "5",  NULL },
         { "6",  NULL },
         { "7",  NULL },
         { "8",  NULL },
         { "9",  NULL },
         { "10", NULL },
         { "11", NULL },
         { "12", NULL },
         { "13", NULL },
         { "14", NULL },
         { "15", NULL },
         { "16", NULL },
         { "17", NULL },
         { "18", NULL },
         { "19", NULL },
         { "20", NULL },
         { "21", NULL },
         { "22", NULL },
         { "23", NULL },
         { "24", NULL },
         { "25", NULL },
         { "26", NULL },
         { "27", NULL },
         { "28", NULL },
         { "29", NULL },
         { "30", NULL },
         { "31", NULL },
         { "32", NULL },
         { "33", NULL },
         { "34", NULL },
         { "35", NULL },
         { "36", NULL },
         { "37", NULL },
         { "38", NULL },
         { "39", NULL },
         { "40", NULL },
         { NULL, NULL },
      },
      "0"
   },
   {
      BEETLE_OPT(last_scanline_pal),
      BEETLE_OPT_LAST_SCANLINE_PAL_LABEL_TH,
      NULL,
      BEETLE_OPT_LAST_SCANLINE_PAL_INFO_0_TH,
      NULL,
      "video",
      {
         { "230", NULL },
         { "231", NULL },
         { "232", NULL },
         { "233", NULL },
         { "234", NULL },
         { "235", NULL },
         { "236", NULL },
         { "237", NULL },
         { "238", NULL },
         { "239", NULL },
         { "240", NULL },
         { "241", NULL },
         { "242", NULL },
         { "243", NULL },
         { "244", NULL },
         { "245", NULL },
         { "246", NULL },
         { "247", NULL },
         { "248", NULL },
         { "249", NULL },
         { "250", NULL },
         { "251", NULL },
         { "252", NULL },
         { "253", NULL },
         { "254", NULL },
         { "255", NULL },
         { "256", NULL },
         { "257", NULL },
         { "258", NULL },
         { "259", NULL },
         { "260", NULL },
         { "261", NULL },
         { "262", NULL },
         { "263", NULL },
         { "264", NULL },
         { "265", NULL },
         { "266", NULL },
         { "267", NULL },
         { "268", NULL },
         { "269", NULL },
         { "270", NULL },
         { "271", NULL },
         { "272", NULL },
         { "273", NULL },
         { "274", NULL },
         { "275", NULL },
         { "276", NULL },
         { "277", NULL },
         { "278", NULL },
         { "279", NULL },
         { "280", NULL },
         { "281", NULL },
         { "282", NULL },
         { "283", NULL },
         { "284", NULL },
         { "285", NULL },
         { "286", NULL },
         { "287", OPTION_VAL_287_TH },
         { NULL, NULL },
      },
      "287"
   },
   { NULL, NULL, NULL, NULL, NULL, NULL, {{0}}, NULL },
};
struct retro_core_options_v2 options_th = {
   option_cats_th,
   option_defs_th
};

/* RETRO_LANGUAGE_TR */

#define CATEGORY_SYSTEM_LABEL_TR "Sistem"
#define CATEGORY_SYSTEM_INFO_0_TR "Sistem ve hız aşırtma seçeneklerini değiştirin."
#define CATEGORY_VIDEO_LABEL_TR NULL
#define CATEGORY_VIDEO_INFO_0_TR "En boy oranını, ekran kırpmayı, video filtresini ve kare atlama ayarlarını değiştirin."
#define CATEGORY_PGXP_LABEL_TR "PGXP (Hassas Geometri Dönüşüm Hattı)"
#define CATEGORY_PGXP_INFO_0_TR "Bu seçenekler, orijinal konsola kıyasla grafikleri iyileştirebilecek geliştirmeleri kontrol eder. PGXP, çarpık dokulardan ve Z-fighting sorunlarından kurtulabilir."
#define CATEGORY_OSD_LABEL_TR "Ekran Görünümü"
#define CATEGORY_OSD_INFO_0_TR "Ekranda gösterilen bildimleri değiştir."
#define CATEGORY_INPUT_LABEL_TR "Girdi"
#define CATEGORY_INPUT_INFO_0_TR "Hafif silah, fare ve neGcon ayarlarını değiştirin."
#define CATEGORY_MEMCARDS_LABEL_TR "Bellek Kartı"
#define CATEGORY_MEMCARDS_INFO_0_TR "Sanal hafıza kartıyla ilgili ayarları değiştirin."
#define CATEGORY_HACKS_LABEL_TR "Emülatör Geliştirmeleri"
#define CATEGORY_HACKS_INFO_0_TR "Emülatör doğruluğu ayarlarını değiştirmek düşük seviyeli performans ve uyumluluk sorunu çıkarabilir."
#define BEETLE_OPT_CPU_FREQ_SCALE_LABEL_TR "CPU Frekans Ölçeği (Hız Aşırtma)"
#define BEETLE_OPT_CPU_FREQ_SCALE_INFO_0_TR "Taklit edilmiş PSX CPU hız aşırtma. Hız aşırtma, artan performans gereksinimleri pahasına belirli oyunlarda yavaşlamayı ortadan kaldırabilir ve kare hızlarını iyileştirebilir. Bazı oyunların dahili bir kare hızı sınırlayıcısı olduğunu ve hız aşırtmadan faydalanmayabileceğini unutmayın. Bazı oyunlarda hız aşırtıldığında bazı efektlerin amaçlanandan daha hızlı canlandırılmasına neden olabilir."
#define OPTION_VAL_50_TR "%50"
#define OPTION_VAL_60_TR "%60"
#define OPTION_VAL_70_TR "%70"
#define OPTION_VAL_80_TR "%80"
#define OPTION_VAL_90_TR "%90"
#define OPTION_VAL_100_TR "%100 (Doğal)"
#define OPTION_VAL_110_TR "%110"
#define OPTION_VAL_120_TR "%120"
#define OPTION_VAL_130_TR "%130"
#define OPTION_VAL_140_TR "%140"
#define OPTION_VAL_150_TR "%150"
#define OPTION_VAL_160_TR "%160"
#define OPTION_VAL_170_TR "%170"
#define OPTION_VAL_180_TR "%180"
#define OPTION_VAL_190_TR "%190"
#define OPTION_VAL_200_TR "%200"
#define OPTION_VAL_210_TR "%210"
#define OPTION_VAL_220_TR "%220"
#define OPTION_VAL_230_TR "%230"
#define OPTION_VAL_240_TR "%240"
#define OPTION_VAL_250_TR "%250"
#define OPTION_VAL_260_TR "%260"
#define OPTION_VAL_270_TR "%270"
#define OPTION_VAL_280_TR "%280"
#define OPTION_VAL_290_TR "%290"
#define OPTION_VAL_300_TR "%300"
#define OPTION_VAL_310_TR "%310"
#define OPTION_VAL_320_TR "%320"
#define OPTION_VAL_330_TR "%330"
#define OPTION_VAL_340_TR "%340"
#define OPTION_VAL_350_TR "%350"
#define OPTION_VAL_360_TR "%360"
#define OPTION_VAL_370_TR "%370"
#define OPTION_VAL_380_TR "%380"
#define OPTION_VAL_390_TR "%390"
#define OPTION_VAL_400_TR "%400"
#define OPTION_VAL_410_TR "%410"
#define OPTION_VAL_420_TR "%420"
#define OPTION_VAL_430_TR "%430"
#define OPTION_VAL_440_TR "%440"
#define OPTION_VAL_450_TR "%450"
#define OPTION_VAL_460_TR "%460"
#define OPTION_VAL_470_TR "%470"
#define OPTION_VAL_480_TR "%480"
#define OPTION_VAL_490_TR "%490"
#define OPTION_VAL_500_TR "%500"
#define OPTION_VAL_510_TR "%510"
#define OPTION_VAL_520_TR "%520"
#define OPTION_VAL_530_TR "%530"
#define OPTION_VAL_540_TR "%540"
#define OPTION_VAL_550_TR "%550"
#define OPTION_VAL_560_TR "%560"
#define OPTION_VAL_570_TR "%570"
#define OPTION_VAL_580_TR "%580"
#define OPTION_VAL_590_TR "%590"
#define OPTION_VAL_600_TR "%600"
#define OPTION_VAL_610_TR "%610"
#define OPTION_VAL_620_TR "%620"
#define OPTION_VAL_630_TR "%630"
#define OPTION_VAL_640_TR "%640"
#define OPTION_VAL_650_TR "%650"
#define OPTION_VAL_660_TR "%660"
#define OPTION_VAL_670_TR "%670"
#define OPTION_VAL_680_TR "%680"
#define OPTION_VAL_690_TR "%690"
#define OPTION_VAL_700_TR "%700"
#define OPTION_VAL_710_TR "%710"
#define OPTION_VAL_720_TR "%720"
#define OPTION_VAL_730_TR "%730"
#define OPTION_VAL_740_TR "%740"
#define OPTION_VAL_750_TR "%750"
#define BEETLE_OPT_GPU_OVERCLOCK_LABEL_TR "GPU Rasterleştirici Hız Aşırtma"
#define BEETLE_OPT_GPU_OVERCLOCK_INFO_0_TR "Taklit edilmiş PSX GPU üstünde bulunan 2D rasterleştiricinin hız aşırtmasını etkinleştirin. 3B oluşturmayı iyileştirmez ve genel olarak çok az etkisi vardır."
#define OPTION_VAL_1X_NATIVE_TR "1x (Doğal)"
#define OPTION_VAL_2X_TR NULL
#define OPTION_VAL_4X_TR NULL
#define OPTION_VAL_8X_TR NULL
#define OPTION_VAL_16X_TR NULL
#define OPTION_VAL_32X_TR NULL
#define BEETLE_OPT_GTE_OVERCLOCK_LABEL_TR "GTE Hız Aşırtma"
#define BEETLE_OPT_GTE_OVERCLOCK_INFO_0_TR "Tüm taklit edilmiş GTE (3D grafikler için CPU yardımcı işlemcisi) işlemlerini sabit bir döngü gecikmesine indirin. GTE yoğun olarak kullanan oyunlar için bu kare hızını ve kare süresi kararlılığını büyük ölçüde iyileştirebilir."
#define BEETLE_OPT_CD_ACCESS_METHOD_LABEL_TR NULL
#define BEETLE_OPT_CD_ACCESS_METHOD_INFO_0_TR NULL
#define OPTION_VAL_SYNC_TR "Eşzamanlı"
#define OPTION_VAL_ASYNC_TR "Eşzamansız"
#define OPTION_VAL_PRECACHE_TR "Önbellek"
#define BEETLE_OPT_CD_FASTLOAD_LABEL_TR "CD Yükleme Hızı"
#define BEETLE_OPT_CD_FASTLOAD_INFO_0_TR NULL
#define OPTION_VAL_2X_NATIVE_TR "2x (Doğal)"
#define OPTION_VAL_6X_TR NULL
#define OPTION_VAL_10X_TR NULL
#define OPTION_VAL_12X_TR NULL
#define OPTION_VAL_14X_TR NULL
#define BEETLE_OPT_SKIP_BIOS_LABEL_TR "BIOS'u Atla"
#define BEETLE_OPT_SKIP_BIOS_INFO_0_TR "İçerik yüklenirken normalde görüntülenen PlayStation BIOS önyükleme animasyonunu atlayın. Not: Bunun etkinleştirilmesi, bazı oyunlarda (PAL kopya korumalı oyunlar, Saga Frontier, vb.) uyumluluk sorunlarına neden olur."
#define BEETLE_OPT_OVERRIDE_BIOS_LABEL_TR NULL
#define BEETLE_OPT_OVERRIDE_BIOS_INFO_0_TR NULL
#define OPTION_VAL_PSXONPSP_TR NULL
#define OPTION_VAL_PS1_ROM_TR NULL
#define OPTION_VAL_OPENBIOS_TR NULL
#define BEETLE_OPT_REGION_LABEL_TR "Sistem Bölgesi"
#define BEETLE_OPT_REGION_INFO_0_TR NULL
#define OPTION_VAL_AUTO_TR "Otomatik"
#define OPTION_VAL_NTSC_J_TR NULL
#define OPTION_VAL_NTSC_U_TR NULL
#define OPTION_VAL_PAL_TR NULL
#define BEETLE_OPT_INTERNAL_RESOLUTION_LABEL_TR "Dahili GPU Çözünürlüğü"
#define BEETLE_OPT_INTERNAL_RESOLUTION_INFO_0_TR NULL
#define BEETLE_OPT_RENDERER_LABEL_TR NULL
#define BEETLE_OPT_RENDERER_INFO_0_TR NULL
#define OPTION_VAL_HARDWARE_TR "Donanım (Otomatik)"
#define OPTION_VAL_HARDWARE_GL_TR "Donanım (OpenGL)"
#define OPTION_VAL_HARDWARE_VK_TR "Donanım (Vulkan)"
#define OPTION_VAL_SOFTWARE_TR "Yazılım"
#define BEETLE_OPT_RENDERER_SOFTWARE_FB_LABEL_TR "Yazılım Kare Önbelleği"
#define BEETLE_OPT_RENDERER_SOFTWARE_FB_INFO_0_TR "Yazılımsal işleyicinin bir kopyasını arka planda yerel çözünürlükte çalıştırarak, donanım oluşturucuları kullanırken kare önbelleği efektlerinin (ör. hareket bulanıklığı, FF7 battle swirl) doğru taklidini sağlayın. Devre dışı bırakılırsa, bu işlemler atlanır (OpenGL) veya GPU üstünde (Vulkan) oluşturulur. Devre dışı bırakmak performansı artırabilir ama ciddi grafik hatalarına neden olabilir. Emin değilseniz etkin bırakın."
#define BEETLE_OPT_DEPTH_LABEL_TR "Dahili Renk Derinliği"
#define BEETLE_OPT_DEPTH_INFO_0_TR NULL
#define OPTION_VAL_16BPP_NATIVE_TR "16 bpp (Doğal)"
#define OPTION_VAL_32BPP_TR NULL
#define BEETLE_OPT_DITHER_MODE_LABEL_TR "Titreşim Kalıbı"
#define BEETLE_OPT_DITHER_MODE_INFO_0_TR NULL
#define OPTION_VAL_INTERNAL_RESOLUTION_TR "Dahili Çözünürlük"
#define BEETLE_OPT_SCALED_UV_OFFSET_LABEL_TR "Doku UV Dengesi"
#define BEETLE_OPT_SCALED_UV_OFFSET_INFO_0_TR "1x çözünürlükten daha yüksek dahili çözünürlük için dengeli 3B çokgenler için örnek dokular. Doku bozulmaları azaltır ancak istenmeyen grafik hatalarına neden olabilir."
#define BEETLE_OPT_FILTER_LABEL_TR "Doku Filtreleme"
#define BEETLE_OPT_FILTER_INFO_0_TR NULL
#define OPTION_VAL_NEAREST_TR "En Yakın"
#define OPTION_VAL_SABR_TR NULL
#define OPTION_VAL_XBR_TR NULL
#define OPTION_VAL_BILINEAR_TR "İkili Doğrusal"
#define OPTION_VAL_3_POINT_TR NULL
#define OPTION_VAL_JINC2_TR NULL
#define BEETLE_OPT_FILTER_EXCLUDE_SPRITE_LABEL_TR "Hareketli Grafikleri Filtrelemeden Çıkar"
#define BEETLE_OPT_FILTER_EXCLUDE_SPRITE_INFO_0_TR "Hareketli görüntülere doku filtrelemesi uygulamayın. 2B hareketli grafikle oluşturulmuş arka planlara sahip çeşitli oyunlarda dikişleri önler. En iyi etki için Uyarlamalı Yumuşatma veya başka bir işleme sonrası filtre ile birlikte kullanın."
#define OPTION_VAL_OPAQUE_TR "Sadece Mat"
#define OPTION_VAL_ALL_TR "Mat ve Yarı Şeffaf"
#define BEETLE_OPT_FILTER_EXCLUDE_2D_POLYGON_LABEL_TR "2B Çokgenleri Filtrelemeden Çıkar"
#define BEETLE_OPT_FILTER_EXCLUDE_2D_POLYGON_INFO_0_TR "2B çokgenlere doku filtrelemesi uygulamayın. 2B çokgenler bir buluşsal yöntem ile algılanır ve aksaklıklar olabilir. En iyi etki için Uyarlamalı Yumuşatma veya başka bir işleme sonrası filtre ile birlikte kullanın."
#define BEETLE_OPT_ADAPTIVE_SMOOTHING_LABEL_TR "Uyarlanabilir Yumuşatma"
#define BEETLE_OPT_ADAPTIVE_SMOOTHING_INFO_0_TR "3B işlenmiş nesneleri bulanıklaştırmadan 2B resim ve arayüz ögelerini düzeltin. Yalnızca Vulkan işleyici tarafından desteklenir."
#define BEETLE_OPT_SUPER_SAMPLING_LABEL_TR "Süper Örnekleme (Yerel Çözünürlüğe Alt Örnekleme)"
#define BEETLE_OPT_SUPER_SAMPLING_INFO_0_TR "İşlenen içeriği, yükseltilmiş dahili çözünürlükten yerel çözünürlüğe kadar aşağı örnekleyin. Bunu daha yüksek dahili çözünürlük çarpanlarıyla birleştirmek, oyunların kenar yumuşatılmış 3B nesnelerle yerel düşük çözünürlükte görüntülenmesine olanak tanır. 2B ve 3B öğeleri karıştıran başlıklara uygulandığında en iyi sonuçları verir (örneğin, önceden oluşturulmuş arka planlarda 3B karakterler) ve CRT gölgelendiricilerle birlikte iyi çalışır. Yalnızca Vulkan işleyici tarafından desteklenir. Not: Bu seçenek etkinleştirilirken 'Titreşim Kalıbı'nın devre dışı bırakılması önerilir."
#define BEETLE_OPT_MSAA_LABEL_TR "Çoklu Örnek Örtüşme Önleme"
#define BEETLE_OPT_MSAA_INFO_0_TR NULL
#define OPTION_VAL_1X_TR NULL
#define BEETLE_OPT_MDEC_YUV_LABEL_TR "MDEC YUV Chroma Filtresi"
#define BEETLE_OPT_MDEC_YUV_INFO_0_TR "'makro blokaj' eserlerini (kareler/çentikli kenarlar) azaltarak FMV oynatım kalitesini iyileştirin. Yalnızca Vulkan işleyici tarafından desteklenir."
#define BEETLE_OPT_TRACK_TEXTURES_LABEL_TR "Dokuları İzle"
#define BEETLE_OPT_TRACK_TEXTURES_INFO_0_TR NULL
#define BEETLE_OPT_DUMP_TEXTURES_LABEL_TR "Dokuları Sakla"
#define BEETLE_OPT_DUMP_TEXTURES_INFO_0_TR "Kullanılan dokuları <cd>-texture-dump/ dizinine saklayın"
#define BEETLE_OPT_REPLACE_TEXTURES_LABEL_TR "Dokuları Değiştir"
#define BEETLE_OPT_REPLACE_TEXTURES_INFO_0_TR "<cd>-texture-replacements/ adresindeki HD sürümlerini kullanarak dokuları değiştirin"
#define BEETLE_OPT_PGXP_MODE_LABEL_TR "PGXP Çalışma Kipi"
#define BEETLE_OPT_PGXP_MODE_INFO_0_TR NULL
#define OPTION_VAL_MEMORY_ONLY_TR "Yalnızca Bellek"
#define OPTION_VAL_MEMORY_CPU_TR "Bellek + Çip (Buggy)"
#define BEETLE_OPT_PGXP_2D_TOL_LABEL_TR "PGXP 2D Geometri Toleransı"
#define BEETLE_OPT_PGXP_2D_TOL_INFO_0_TR "PGXP işlemlerinde daha fazla göze çarpan hataları gizleyin: değer, uygun derinlik bilgisi olmayan geometriler olması durumunda PGXP değerlerinin tutulacağı toleransı belirtir."
#define OPTION_VAL_0PX_TR NULL
#define OPTION_VAL_1PX_TR NULL
#define OPTION_VAL_2PX_TR NULL
#define OPTION_VAL_3PX_TR NULL
#define OPTION_VAL_4PX_TR NULL
#define OPTION_VAL_5PX_TR NULL
#define OPTION_VAL_6PX_TR NULL
#define OPTION_VAL_7PX_TR NULL
#define OPTION_VAL_8PX_TR NULL
#define BEETLE_OPT_PGXP_NCLIP_LABEL_TR "PGXP Eskiliği Kaldır"
#define BEETLE_OPT_PGXP_NCLIP_INFO_0_TR "PGXP NCLIP uygulamasını kullanın. PGXP koordinatlarıyla geometrilerdeki delikleri azaltarak görünümü iyileştirir. Bazı oyunların çeşitli durumlarda kilitlenmesine neden olduğu bilinmektedir."
#define BEETLE_OPT_PGXP_VERTEX_LABEL_TR "PGXP Köşe Önbelleği"
#define BEETLE_OPT_PGXP_VERTEX_INFO_0_TR "Çokgen çizimlerinde yeniden kullanım için PGXP ile geliştirilmiş köşe konumlarını önbelleğe alın. Dokuları işlerken potansiyel olarak nesne hizalamasını iyileştirebilir ve görünür tırtıkları azaltabilir, ancak önbelleği sorgularken yanlış pozitifler grafik arızalarına neden olabilir. Şu anda bu seçeneği devre dışı bırakmanız önerilir. Bu seçenek yalnızca PGXP Çalışma Kipi etkinleştirildiğinde uygulanır. Yalnızca donanım işleyiciler tarafından desteklenir."
#define BEETLE_OPT_PGXP_TEXTURE_LABEL_TR "PGXP Perspektif Doğru Doku"
#define BEETLE_OPT_PGXP_TEXTURE_INFO_0_TR NULL
#define BEETLE_OPT_DISPLAY_INTERNAL_FPS_LABEL_TR "Dahili FPS'yi Görüntüle"
#define BEETLE_OPT_DISPLAY_INTERNAL_FPS_INFO_0_TR "Taklit edilmiş PlayStation sisteminin dahili kare hızını görüntüleyin. Not: Libretro ön ucunda etkinleştirilecek ekran bildirimlerini gerektirir."
#define BEETLE_OPT_DISPLAY_OSD_LABEL_TR "OSD Mesajlarını Görüntüle"
#define BEETLE_OPT_DISPLAY_OSD_INFO_0_TR "Çekirdek tarafından oluşturulan OSD mesajlarını görüntüleyin."
#define BEETLE_OPT_DISPLAY_VRAM_LABEL_TR "Tam VRAM Görüntüle (Hata Ayıklama)"
#define BEETLE_OPT_DISPLAY_VRAM_INFO_0_TR "Taklit edilmiş konsolun VRAM miktarının tamamını görselleştirin. Yalnızca OpenGL ve Vulkan donanım işleyicileri tarafından desteklenir. Not: Bu, hata ayıklama amaçlıdır ve normalde devre dışı bırakılmalıdır."
#define BEETLE_OPT_ANALOG_CALIBRATION_LABEL_TR "Analog Otomatik Kalibrasyon"
#define BEETLE_OPT_ANALOG_CALIBRATION_INFO_0_TR NULL
#define BEETLE_OPT_ANALOG_TOGGLE_LABEL_TR "DualShock Analog Mod Değiştir"
#define BEETLE_OPT_ANALOG_TOGGLE_INFO_0_TR "Giriş cihazı türü DualShock olduğunda, bu seçenek taklit edilen DualShock'un orijinal donanım gibi DİJİTAL ve ANALOG mod arasında geçiş yapmasına olanak tanır. Mod, seçilen düğme kombinasyonu kullanılarak da değiştirilebilir."
#define OPTION_VAL_ENABLED_ANALOG_TR "Varsayılan-Analog"
#define BEETLE_OPT_ANALOG_TOGGLE_COMBO_LABEL_TR "DualShock Analog Mod Kombinasyonu"
#define BEETLE_OPT_ANALOG_TOGGLE_COMBO_INFO_0_TR NULL
#define OPTION_VAL_L1_L2_R1_R2_START_SELECT_TR NULL
#define OPTION_VAL_L1_R1_SELECT_TR NULL
#define OPTION_VAL_L1_R1_START_TR NULL
#define OPTION_VAL_L1_R1_L3_TR NULL
#define OPTION_VAL_L1_R1_R3_TR NULL
#define OPTION_VAL_L2_R2_SELECT_TR NULL
#define OPTION_VAL_L2_R2_START_TR NULL
#define OPTION_VAL_L2_R2_L3_TR NULL
#define OPTION_VAL_L2_R2_R3_TR NULL
#define OPTION_VAL_L3_R3_TR NULL
#define BEETLE_OPT_ANALOG_TOGGLE_HOLD_LABEL_TR "DualShock Analog Mod Kombinasyonu Basma Gecikmesi"
#define BEETLE_OPT_ANALOG_TOGGLE_HOLD_INFO_0_TR NULL
#define OPTION_VAL_0_TR "0 Saniye Gecikme"
#define OPTION_VAL_1_TR "1 Saniye Gecikme"
#define OPTION_VAL_2_TR "2 Saniye Gecikme"
#define OPTION_VAL_3_TR "3 Saniye Gecikme"
#define OPTION_VAL_4_TR "4 Saniye Gecikme"
#define OPTION_VAL_5_TR "5 Saniye Gecikme"
#define BEETLE_OPT_ENABLE_MULTITAP_PORT1_LABEL_TR NULL
#define BEETLE_OPT_ENABLE_MULTITAP_PORT1_INFO_0_TR "Port 1 üstünde multilab işlevini etkinleştirin."
#define BEETLE_OPT_ENABLE_MULTITAP_PORT2_LABEL_TR NULL
#define BEETLE_OPT_ENABLE_MULTITAP_PORT2_INFO_0_TR "Port 2 üstünde multilab işlevini etkinleştirin."
#define BEETLE_OPT_GUN_INPUT_MODE_LABEL_TR "Tabanca Girdi Kipi"
#define BEETLE_OPT_GUN_INPUT_MODE_INFO_0_TR NULL
#define OPTION_VAL_LIGHTGUN_TR NULL
#define OPTION_VAL_TOUCHSCREEN_TR "Dokunmatik Ekran"
#define BEETLE_OPT_GUN_CURSOR_LABEL_TR "Silah İmleci"
#define BEETLE_OPT_GUN_CURSOR_INFO_0_TR NULL
#define OPTION_VAL_CROSS_TR "Çarpı"
#define OPTION_VAL_DOT_TR "Nokta"
#define BEETLE_OPT_CROSSHAIR_COLOR_P1_LABEL_TR NULL
#define BEETLE_OPT_CROSSHAIR_COLOR_P1_INFO_0_TR NULL
#define OPTION_VAL_RED_TR "Kırmızı"
#define OPTION_VAL_BLUE_TR "Mavi"
#define OPTION_VAL_GREEN_TR "Yeşil"
#define OPTION_VAL_ORANGE_TR "Turuncu"
#define OPTION_VAL_YELLOW_TR "Sarı"
#define OPTION_VAL_CYAN_TR "Camgöbeği"
#define OPTION_VAL_PINK_TR "Pembe"
#define OPTION_VAL_PURPLE_TR "Mor"
#define OPTION_VAL_BLACK_TR "Siyah"
#define OPTION_VAL_WHITE_TR "Beyaz"
#define BEETLE_OPT_CROSSHAIR_COLOR_P2_LABEL_TR NULL
#define BEETLE_OPT_CROSSHAIR_COLOR_P2_INFO_0_TR NULL
#define BEETLE_OPT_MOUSE_SENSITIVITY_LABEL_TR "Fare Hassasiyeti"
#define BEETLE_OPT_MOUSE_SENSITIVITY_INFO_0_TR NULL
#define OPTION_VAL_5_O42_TR "%5"
#define OPTION_VAL_10_TR "%10"
#define OPTION_VAL_15_TR "%15"
#define OPTION_VAL_20_TR "%20"
#define OPTION_VAL_25_TR "%25"
#define OPTION_VAL_30_TR "%30"
#define OPTION_VAL_35_TR "%35"
#define OPTION_VAL_40_TR "%40"
#define OPTION_VAL_45_TR "%45"
#define OPTION_VAL_55_TR "%55"
#define OPTION_VAL_65_TR "%65"
#define OPTION_VAL_75_TR "%75"
#define OPTION_VAL_85_TR "%85"
#define OPTION_VAL_95_TR "%95"
#define OPTION_VAL_100_O42_TR "%100"
#define OPTION_VAL_105_TR "%105"
#define OPTION_VAL_115_TR "%115"
#define OPTION_VAL_125_TR "%125"
#define OPTION_VAL_135_TR "%135"
#define OPTION_VAL_145_TR "%145"
#define OPTION_VAL_155_TR "%155"
#define OPTION_VAL_165_TR "%165"
#define OPTION_VAL_175_TR "%175"
#define OPTION_VAL_185_TR "%185"
#define OPTION_VAL_195_TR "%195"
#define BEETLE_OPT_NEGCON_RESPONSE_LABEL_TR "neGcon Büküm Tepkisi"
#define BEETLE_OPT_NEGCON_RESPONSE_INFO_0_TR NULL
#define OPTION_VAL_LINEAR_TR "Doğrusal"
#define OPTION_VAL_QUADRATIC_TR "Karesel"
#define OPTION_VAL_CUBIC_TR "Kübik"
#define BEETLE_OPT_NEGCON_DEADZONE_LABEL_TR "neGcon Büküm Ölü Bölge"
#define BEETLE_OPT_NEGCON_DEADZONE_INFO_0_TR NULL
#define OPTION_VAL_0_O44_TR "%0"
#define BEETLE_OPT_USE_MEDNAFEN_MEMCARD0_METHOD_LABEL_TR NULL
#define BEETLE_OPT_USE_MEDNAFEN_MEMCARD0_METHOD_INFO_0_TR NULL
#define OPTION_VAL_LIBRETRO_TR NULL
#define OPTION_VAL_MEDNAFEN_TR NULL
#define BEETLE_OPT_ENABLE_MEMCARD1_LABEL_TR NULL
#define BEETLE_OPT_ENABLE_MEMCARD1_INFO_0_TR NULL
#define BEETLE_OPT_SHARED_MEMORY_CARDS_LABEL_TR NULL
#define BEETLE_OPT_SHARED_MEMORY_CARDS_INFO_0_TR NULL
#define BEETLE_OPT_MEMCARD_LEFT_INDEX_LABEL_TR NULL
#define BEETLE_OPT_MEMCARD_LEFT_INDEX_INFO_0_TR NULL
#define OPTION_VAL_0_O48_TR "0 (Varsayılan)"
#define BEETLE_OPT_MEMCARD_RIGHT_INDEX_LABEL_TR NULL
#define BEETLE_OPT_MEMCARD_RIGHT_INDEX_INFO_0_TR NULL
#define OPTION_VAL_1_O49_TR "1 (Varsayılan)"
#define BEETLE_OPT_LINE_RENDER_LABEL_TR "Dörtlü Hat Geliştirmesi"
#define BEETLE_OPT_LINE_RENDER_INFO_0_TR NULL
#define OPTION_VAL_DEFAULT_TR "Varsayılan"
#define OPTION_VAL_AGGRESSIVE_TR "Agresif"
#define BEETLE_OPT_WIDESCREEN_HACK_LABEL_TR "Geniş Ekran Geliştirmesi"
#define BEETLE_OPT_WIDESCREEN_HACK_INFO_0_TR "3D içeriği anamorfik olarak işleyin ve taklit edilmiş çerçeve arabelleğini geniş ekran en boy oranında yükseltin. 3D oyunlarla en iyi sonuçları üretir. 2D ögeler yatay olarak gerilecek ve yanlış hizalanmış olabilir."
#define BEETLE_OPT_WIDESCREEN_HACK_ASPECT_RATIO_LABEL_TR "Geniş Ekran Kipi Geliştirmesi En Boy Oranı"
#define BEETLE_OPT_WIDESCREEN_HACK_ASPECT_RATIO_INFO_0_TR NULL
#define OPTION_VAL_16_9_TR NULL
#define OPTION_VAL_16_10_TR NULL
#define OPTION_VAL_18_9_TR NULL
#define OPTION_VAL_19_9_TR NULL
#define OPTION_VAL_20_9_TR NULL
#define OPTION_VAL_21_9_TR NULL
#define OPTION_VAL_32_9_TR NULL
#define BEETLE_OPT_CPU_DYNAREC_LABEL_TR "CPU Dinamik Derleyici"
#define BEETLE_OPT_CPU_DYNAREC_INFO_0_TR "CPU talimatlarını yerel talimatlara dinamik olarak yeniden derleyin. Yorumlayıcıdan çok daha hızlıdır, ancak CPU zamanlaması daha az doğrudur ve hataları olabilir."
#define OPTION_VAL_DISABLED_TR "Kapalı (Beetle Interpreter)"
#define OPTION_VAL_EXECUTE_TR "Azami Performans"
#define OPTION_VAL_EXECUTE_ONE_TR "Döngü Zaman Kontrolü"
#define OPTION_VAL_RUN_INTERPRETER_TR NULL
#define BEETLE_OPT_DYNAREC_INVALIDATE_LABEL_TR "Dynarec Kodu Geçerliliği"
#define BEETLE_OPT_DYNAREC_INVALIDATE_INFO_0_TR "Bazı oyunlar 'Tam' geçersizlik gerektirir, bazıları 'Yalnızca DMA' gerektirir."
#define OPTION_VAL_FULL_TR "Tam"
#define OPTION_VAL_DMA_TR "Sadece DMA (Biraz Daha Hızlı)"
#define BEETLE_OPT_DYNAREC_EVENTCYCLES_LABEL_TR "Dynarec DMA/GPU/MDEC/Zamanlayıcı Olay Döngüleri"
#define BEETLE_OPT_DYNAREC_EVENTCYCLES_INFO_0_TR "Bir GPU/DMA/MDEC/Zamanlayıcı Güncellemesi kontrol edilmeden önce CPU tarafından çalıştırılan en fazla döngü sayısı, daha yüksek sayı daha hızlı olacaktır, beetle derleyiciı üzerinde dynarec için çok daha az etkiye sahiptir."
#define OPTION_VAL_128_TR "128 (Varsayılan)"
#define BEETLE_OPT_DYNAREC_SPU_SAMPLES_LABEL_TR "Dynarec SPU Örnekleri"
#define BEETLE_OPT_DYNAREC_SPU_SAMPLES_INFO_0_TR "Bir SPU Güncellemesi kontrol edilmeden önce çalıştırılacak en fazla SPU örnekleri, daha yüksek sayı daha hızlı olacaktır, ancak bazı oyunlarda 1'den farklı herhangi bir değerde ses aksaklıklarına neden olacaktır."
#define BEETLE_OPT_CORE_TIMING_FPS_LABEL_TR "Çekirdek Raporlu FPS Zamanlaması"
#define BEETLE_OPT_CORE_TIMING_FPS_INFO_0_TR NULL
#define OPTION_VAL_AUTO_TOGGLE_TR NULL
#define OPTION_VAL_FORCE_PROGRESSIVE_TR NULL
#define OPTION_VAL_FORCE_INTERLACED_TR "Geçişli Oran Zorla"
#define BEETLE_OPT_DEINTERLACER_LABEL_TR "Taramasızlaştırma Yöntemi"
#define BEETLE_OPT_DEINTERLACER_INFO_0_TR NULL
#define OPTION_VAL_WEAVE_TR NULL
#define OPTION_VAL_BOB_TR NULL
#define OPTION_VAL_BOB_OFFSET_TR NULL
#define OPTION_VAL_FASTMAD_TR NULL
#define BEETLE_OPT_FRAME_DUPING_LABEL_TR NULL
#define BEETLE_OPT_FRAME_DUPING_INFO_0_TR NULL
#define BEETLE_OPT_PAL_VIDEO_TIMING_OVERRIDE_LABEL_TR NULL
#define BEETLE_OPT_PAL_VIDEO_TIMING_OVERRIDE_INFO_0_TR NULL
#define BEETLE_OPT_ASPECT_RATIO_LABEL_TR "Çekirdek En Boy Oranı"
#define BEETLE_OPT_ASPECT_RATIO_INFO_0_TR NULL
#define OPTION_VAL_CORRECTED_TR "Düzeltilmiş"
#define OPTION_VAL_UNCORRECTED_TR "Düzeltilmemiş"
#define OPTION_VAL_4_3_TR "4:3 İçin Zorla"
#define OPTION_VAL_NTSC_TR "NTSC İçin Zorla"
#define BEETLE_OPT_CROP_OVERSCAN_LABEL_TR "Aşırı Taramayı Kırp"
#define BEETLE_OPT_CROP_OVERSCAN_INFO_0_TR NULL
#define OPTION_VAL_STATIC_TR "Yatay"
#define OPTION_VAL_SMART_TR NULL
#define BEETLE_OPT_IMAGE_CROP_LABEL_TR "Ek Kırpma"
#define BEETLE_OPT_IMAGE_CROP_INFO_0_TR NULL
#define OPTION_VAL_9PX_TR NULL
#define OPTION_VAL_10PX_TR NULL
#define OPTION_VAL_11PX_TR NULL
#define OPTION_VAL_12PX_TR NULL
#define OPTION_VAL_13PX_TR NULL
#define OPTION_VAL_14PX_TR NULL
#define OPTION_VAL_15PX_TR NULL
#define OPTION_VAL_16PX_TR NULL
#define OPTION_VAL_17PX_TR NULL
#define OPTION_VAL_18PX_TR NULL
#define OPTION_VAL_19PX_TR NULL
#define OPTION_VAL_20PX_TR NULL
#define BEETLE_OPT_IMAGE_OFFSET_LABEL_TR "Merkeze Kırpılmış Görüntü"
#define BEETLE_OPT_IMAGE_OFFSET_INFO_0_TR NULL
#define OPTION_VAL_12PX_O64_TR NULL
#define OPTION_VAL_11PX_O64_TR NULL
#define OPTION_VAL_10PX_O64_TR NULL
#define OPTION_VAL_9PX_O64_TR NULL
#define OPTION_VAL_8PX_O64_TR NULL
#define OPTION_VAL_7PX_O64_TR NULL
#define OPTION_VAL_6PX_O64_TR NULL
#define OPTION_VAL_5PX_O64_TR NULL
#define OPTION_VAL_4PX_O64_TR NULL
#define OPTION_VAL_3PX_O64_TR NULL
#define OPTION_VAL_2PX_O64_TR NULL
#define OPTION_VAL_1PX_O64_TR NULL
#define OPTION_VAL_1PX_O64_0_TR NULL
#define OPTION_VAL_2PX_O64_0_TR NULL
#define OPTION_VAL_3PX_O64_0_TR NULL
#define OPTION_VAL_4PX_O64_0_TR NULL
#define OPTION_VAL_5PX_O64_0_TR NULL
#define OPTION_VAL_6PX_O64_0_TR NULL
#define OPTION_VAL_7PX_O64_0_TR NULL
#define OPTION_VAL_8PX_O64_0_TR NULL
#define OPTION_VAL_9PX_O64_0_TR NULL
#define OPTION_VAL_10PX_O64_0_TR NULL
#define OPTION_VAL_11PX_O64_0_TR NULL
#define OPTION_VAL_12PX_O64_0_TR NULL
#define BEETLE_OPT_IMAGE_OFFSET_CYCLES_LABEL_TR "Yatay Görüntü Dengesi (GPU Döngüleri)"
#define BEETLE_OPT_IMAGE_OFFSET_CYCLES_INFO_0_TR NULL
#define BEETLE_OPT_INITIAL_SCANLINE_LABEL_TR "İlk Tarama Hattı - NTSC"
#define BEETLE_OPT_INITIAL_SCANLINE_INFO_0_TR NULL
#define BEETLE_OPT_LAST_SCANLINE_LABEL_TR "Son Tarama Hattı - NTSC"
#define BEETLE_OPT_LAST_SCANLINE_INFO_0_TR NULL
#define OPTION_VAL_239_TR "239 (Varsayılan)"
#define BEETLE_OPT_INITIAL_SCANLINE_PAL_LABEL_TR "İlk Tarama Hattı - PAL"
#define BEETLE_OPT_INITIAL_SCANLINE_PAL_INFO_0_TR NULL
#define BEETLE_OPT_LAST_SCANLINE_PAL_LABEL_TR "Son Tarama Hattı - PAL"
#define BEETLE_OPT_LAST_SCANLINE_PAL_INFO_0_TR NULL
#define OPTION_VAL_287_TR "287 (Varsayılan)"

struct retro_core_option_v2_category option_cats_tr[] = {
   {
      "system",
      CATEGORY_SYSTEM_LABEL_TR,
      CATEGORY_SYSTEM_INFO_0_TR
   },
   {
      "video",
      CATEGORY_VIDEO_LABEL_TR,
      CATEGORY_VIDEO_INFO_0_TR
   },
   {
      "pgxp",
      CATEGORY_PGXP_LABEL_TR,
      CATEGORY_PGXP_INFO_0_TR
   },
   {
      "osd",
      CATEGORY_OSD_LABEL_TR,
      CATEGORY_OSD_INFO_0_TR
   },
   {
      "input",
      CATEGORY_INPUT_LABEL_TR,
      CATEGORY_INPUT_INFO_0_TR
   },
   {
      "memcards",
      CATEGORY_MEMCARDS_LABEL_TR,
      CATEGORY_MEMCARDS_INFO_0_TR
   },
   {
      "hacks",
      CATEGORY_HACKS_LABEL_TR,
      CATEGORY_HACKS_INFO_0_TR
   },
   { NULL, NULL, NULL },
};
struct retro_core_option_v2_definition option_defs_tr[] = {
   {
      BEETLE_OPT(cpu_freq_scale),
      BEETLE_OPT_CPU_FREQ_SCALE_LABEL_TR,
      NULL,
      BEETLE_OPT_CPU_FREQ_SCALE_INFO_0_TR,
      NULL,
      "system",
      {
         { "50%",           OPTION_VAL_50_TR },
         { "60%",           OPTION_VAL_60_TR },
         { "70%",           OPTION_VAL_70_TR },
         { "80%",           OPTION_VAL_80_TR },
         { "90%",           OPTION_VAL_90_TR },
         { "100%",          OPTION_VAL_100_TR },
         { "110%",          OPTION_VAL_110_TR },
         { "120%",          OPTION_VAL_120_TR },
         { "130%",          OPTION_VAL_130_TR },
         { "140%",          OPTION_VAL_140_TR },
         { "150%",          OPTION_VAL_150_TR },
         { "160%",          OPTION_VAL_160_TR },
         { "170%",          OPTION_VAL_170_TR },
         { "180%",          OPTION_VAL_180_TR },
         { "190%",          OPTION_VAL_190_TR },
         { "200%",          OPTION_VAL_200_TR },
         { "210%",          OPTION_VAL_210_TR },
         { "220%",          OPTION_VAL_220_TR },
         { "230%",          OPTION_VAL_230_TR },
         { "240%",          OPTION_VAL_240_TR },
         { "250%",          OPTION_VAL_250_TR },
         { "260%",          OPTION_VAL_260_TR },
         { "270%",          OPTION_VAL_270_TR },
         { "280%",          OPTION_VAL_280_TR },
         { "290%",          OPTION_VAL_290_TR },
         { "300%",          OPTION_VAL_300_TR },
         { "310%",          OPTION_VAL_310_TR },
         { "320%",          OPTION_VAL_320_TR },
         { "330%",          OPTION_VAL_330_TR },
         { "340%",          OPTION_VAL_340_TR },
         { "350%",          OPTION_VAL_350_TR },
         { "360%",          OPTION_VAL_360_TR },
         { "370%",          OPTION_VAL_370_TR },
         { "380%",          OPTION_VAL_380_TR },
         { "390%",          OPTION_VAL_390_TR },
         { "400%",          OPTION_VAL_400_TR },
         { "410%",          OPTION_VAL_410_TR },
         { "420%",          OPTION_VAL_420_TR },
         { "430%",          OPTION_VAL_430_TR },
         { "440%",          OPTION_VAL_440_TR },
         { "450%",          OPTION_VAL_450_TR },
         { "460%",          OPTION_VAL_460_TR },
         { "470%",          OPTION_VAL_470_TR },
         { "480%",          OPTION_VAL_480_TR },
         { "490%",          OPTION_VAL_490_TR },
         { "500%",          OPTION_VAL_500_TR },
         { "510%",          OPTION_VAL_510_TR },
         { "520%",          OPTION_VAL_520_TR },
         { "530%",          OPTION_VAL_530_TR },
         { "540%",          OPTION_VAL_540_TR },
         { "550%",          OPTION_VAL_550_TR },
         { "560%",          OPTION_VAL_560_TR },
         { "570%",          OPTION_VAL_570_TR },
         { "580%",          OPTION_VAL_580_TR },
         { "590%",          OPTION_VAL_590_TR },
         { "600%",          OPTION_VAL_600_TR },
         { "610%",          OPTION_VAL_610_TR },
         { "620%",          OPTION_VAL_620_TR },
         { "630%",          OPTION_VAL_630_TR },
         { "640%",          OPTION_VAL_640_TR },
         { "650%",          OPTION_VAL_650_TR },
         { "660%",          OPTION_VAL_660_TR },
         { "670%",          OPTION_VAL_670_TR },
         { "680%",          OPTION_VAL_680_TR },
         { "690%",          OPTION_VAL_690_TR },
         { "700%",          OPTION_VAL_700_TR },
         { "710%",          OPTION_VAL_710_TR },
         { "720%",          OPTION_VAL_720_TR },
         { "730%",          OPTION_VAL_730_TR },
         { "740%",          OPTION_VAL_740_TR },
         { "750%",          OPTION_VAL_750_TR },
         { NULL, NULL },
      },
      "100%"
   },
   {
      BEETLE_OPT(gpu_overclock),
      BEETLE_OPT_GPU_OVERCLOCK_LABEL_TR,
      NULL,
      BEETLE_OPT_GPU_OVERCLOCK_INFO_0_TR,
      NULL,
      "system",
      {
         { "1x(native)", OPTION_VAL_1X_NATIVE_TR },
         { "2x",         OPTION_VAL_2X_TR },
         { "4x",         OPTION_VAL_4X_TR },
         { "8x",         OPTION_VAL_8X_TR },
         { "16x",        OPTION_VAL_16X_TR },
         { "32x",        OPTION_VAL_32X_TR },
         { NULL, NULL },
      },
      "1x(native)"
   },
   {
      BEETLE_OPT(gte_overclock),
      BEETLE_OPT_GTE_OVERCLOCK_LABEL_TR,
      NULL,
      BEETLE_OPT_GTE_OVERCLOCK_INFO_0_TR,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
#ifndef EMSCRIPTEN
   {
      BEETLE_OPT(cd_access_method),
      BEETLE_OPT_CD_ACCESS_METHOD_LABEL_TR,
      NULL,
      BEETLE_OPT_CD_ACCESS_METHOD_INFO_0_TR,
      NULL,
      "system",
      {
         { "sync",     OPTION_VAL_SYNC_TR },
         { "async",    OPTION_VAL_ASYNC_TR },
         { "precache", OPTION_VAL_PRECACHE_TR },
         { NULL, NULL },
      },
      "sync"
   },
#endif
   {
      BEETLE_OPT(cd_fastload),
      BEETLE_OPT_CD_FASTLOAD_LABEL_TR,
      NULL,
      BEETLE_OPT_CD_FASTLOAD_INFO_0_TR,
      NULL,
      "system",
      {
         { "2x(native)", OPTION_VAL_2X_NATIVE_TR },
         { "4x",          OPTION_VAL_4X_TR },
         { "6x",          OPTION_VAL_6X_TR },
         { "8x",          OPTION_VAL_8X_TR },
         { "10x",         OPTION_VAL_10X_TR },
         { "12x",         OPTION_VAL_12X_TR },
         { "14x",         OPTION_VAL_14X_TR },
         { NULL, NULL },
      },
      "2x(native)"
   },
   {
      BEETLE_OPT(skip_bios),
      BEETLE_OPT_SKIP_BIOS_LABEL_TR,
      NULL,
      BEETLE_OPT_SKIP_BIOS_INFO_0_TR,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(override_bios),
      BEETLE_OPT_OVERRIDE_BIOS_LABEL_TR,
      NULL,
      BEETLE_OPT_OVERRIDE_BIOS_INFO_0_TR,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "psxonpsp", OPTION_VAL_PSXONPSP_TR },
         { "ps1_rom",  OPTION_VAL_PS1_ROM_TR },
         { "openbios", OPTION_VAL_OPENBIOS_TR     },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(region),
      BEETLE_OPT_REGION_LABEL_TR,
      NULL,
      BEETLE_OPT_REGION_INFO_0_TR,
      NULL,
      "system",
      {
         { "auto",   OPTION_VAL_AUTO_TR },
         { "ntsc-j", OPTION_VAL_NTSC_J_TR },
         { "ntsc-u", OPTION_VAL_NTSC_U_TR },
         { "pal",    OPTION_VAL_PAL_TR },
         { NULL, NULL },
      },
      "auto"
   },
   {
      BEETLE_OPT(internal_resolution),
      BEETLE_OPT_INTERNAL_RESOLUTION_LABEL_TR,
      NULL,
      BEETLE_OPT_INTERNAL_RESOLUTION_INFO_0_TR,
      NULL,
      "video",
      {
         { "1x(native)", OPTION_VAL_1X_NATIVE_TR },
         { "2x",         OPTION_VAL_2X_TR },
         { "4x",         OPTION_VAL_4X_TR },
         { "8x",         OPTION_VAL_8X_TR },
         { "16x",        OPTION_VAL_16X_TR },
         { NULL, NULL },
      },
      "1x(native)"
   },
#if defined(HAVE_OPENGL) || defined(HAVE_OPENGLES) || defined(HAVE_VULKAN)
   {
      BEETLE_OPT(renderer),
      BEETLE_OPT_RENDERER_LABEL_TR,
      NULL,
      BEETLE_OPT_RENDERER_INFO_0_TR,
      NULL,
      "video",
      {
         { "hardware",    OPTION_VAL_HARDWARE_TR },
#if defined(HAVE_OPENGL) || defined(HAVE_OPENGLES)
         { "hardware_gl", OPTION_VAL_HARDWARE_GL_TR },
#endif
#if defined(HAVE_VULKAN)
         { "hardware_vk", OPTION_VAL_HARDWARE_VK_TR },
#endif
         { "software",    OPTION_VAL_SOFTWARE_TR },
         { NULL, NULL },
      },
      "hardware"
   },
   {
      BEETLE_OPT(renderer_software_fb),
      BEETLE_OPT_RENDERER_SOFTWARE_FB_LABEL_TR,
      NULL,
      BEETLE_OPT_RENDERER_SOFTWARE_FB_INFO_0_TR,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled"
   },
#endif
#if defined(HAVE_OPENGL) || defined(HAVE_OPENGLES)
   {
      BEETLE_OPT(depth),
      BEETLE_OPT_DEPTH_LABEL_TR,
      NULL,
      BEETLE_OPT_DEPTH_INFO_0_TR,
      NULL,
      "video",
      {
         { "16bpp(native)", OPTION_VAL_16BPP_NATIVE_TR },
         { "32bpp",         OPTION_VAL_32BPP_TR },
         { NULL, NULL },
      },
      "16bpp(native)"
   },
   // Sort of, it's more like 15-bit high color and 24-bit true color for visible output. The alpha channel is used for mask bit. Vulkan renderer uses ABGR1555_555 for 31 bits internal? FMVs are always 24-bit on all renderers like original hardware (BGR888, no alpha)
#endif
   {
      BEETLE_OPT(dither_mode),
      BEETLE_OPT_DITHER_MODE_LABEL_TR,
      NULL,
      BEETLE_OPT_DITHER_MODE_INFO_0_TR,
      NULL,
      "video",
      {
         { "1x(native)",          OPTION_VAL_1X_NATIVE_TR },
         { "internal resolution", OPTION_VAL_INTERNAL_RESOLUTION_TR },
         { "disabled",            NULL },
         { NULL, NULL },
      },
      "1x(native)"
   },
#ifdef HAVE_VULKAN
   {
      BEETLE_OPT(scaled_uv_offset),
      BEETLE_OPT_SCALED_UV_OFFSET_LABEL_TR,
      NULL,
      BEETLE_OPT_SCALED_UV_OFFSET_INFO_0_TR,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled"
   },
#endif
#if defined(HAVE_OPENGL) || defined(HAVE_OPENGLES) || defined(HAVE_VULKAN)
   {
      BEETLE_OPT(filter),
      BEETLE_OPT_FILTER_LABEL_TR,
      NULL,
      BEETLE_OPT_FILTER_INFO_0_TR,
      NULL,
      "video",
      {
         { "nearest",  OPTION_VAL_NEAREST_TR },
         { "SABR",     OPTION_VAL_SABR_TR },
         { "xBR",      OPTION_VAL_XBR_TR },
         { "bilinear", OPTION_VAL_BILINEAR_TR },
         { "3-point",  OPTION_VAL_3_POINT_TR },
         { "JINC2",    OPTION_VAL_JINC2_TR },
         { NULL, NULL },
      },
      "nearest"
   },
#ifdef HAVE_VULKAN
   {
      BEETLE_OPT(filter_exclude_sprite),
      BEETLE_OPT_FILTER_EXCLUDE_SPRITE_LABEL_TR,
      NULL,
      BEETLE_OPT_FILTER_EXCLUDE_SPRITE_INFO_0_TR,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "opaque", OPTION_VAL_OPAQUE_TR },
         { "all", OPTION_VAL_ALL_TR },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(filter_exclude_2d_polygon),
      BEETLE_OPT_FILTER_EXCLUDE_2D_POLYGON_LABEL_TR,
      NULL,
      BEETLE_OPT_FILTER_EXCLUDE_2D_POLYGON_INFO_0_TR,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "opaque", OPTION_VAL_OPAQUE_TR },
         { "all", OPTION_VAL_ALL_TR },
         { NULL, NULL },
      },
      "disabled"
   },
#endif
#endif
#ifdef HAVE_VULKAN
   {
      BEETLE_OPT(adaptive_smoothing),
      BEETLE_OPT_ADAPTIVE_SMOOTHING_LABEL_TR,
      NULL,
      BEETLE_OPT_ADAPTIVE_SMOOTHING_INFO_0_TR,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(super_sampling),
      BEETLE_OPT_SUPER_SAMPLING_LABEL_TR,
      NULL,
      BEETLE_OPT_SUPER_SAMPLING_INFO_0_TR,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(msaa),
      BEETLE_OPT_MSAA_LABEL_TR,
      NULL,
      BEETLE_OPT_MSAA_INFO_0_TR,
      NULL,
      "video",
      {
         { "1x",  OPTION_VAL_1X_TR },
         { "2x",  OPTION_VAL_2X_TR },
         { "4x",  OPTION_VAL_4X_TR },
         { "8x",  OPTION_VAL_8X_TR },
         { "16x", OPTION_VAL_16X_TR },
         { NULL, NULL },
      },
      "1x"
   },
   {
      BEETLE_OPT(mdec_yuv),
      BEETLE_OPT_MDEC_YUV_LABEL_TR,
      NULL,
      BEETLE_OPT_MDEC_YUV_INFO_0_TR,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(track_textures),
      BEETLE_OPT_TRACK_TEXTURES_LABEL_TR,
      NULL,
      BEETLE_OPT_TRACK_TEXTURES_INFO_0_TR,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
#ifdef TEXTURE_DUMPING_ENABLED
   {
      BEETLE_OPT(dump_textures),
      BEETLE_OPT_DUMP_TEXTURES_LABEL_TR,
      NULL,
      BEETLE_OPT_DUMP_TEXTURES_INFO_0_TR,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
#endif
   {
      BEETLE_OPT(replace_textures),
      BEETLE_OPT_REPLACE_TEXTURES_LABEL_TR,
      NULL,
      BEETLE_OPT_REPLACE_TEXTURES_INFO_0_TR,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
#endif
   {
      BEETLE_OPT(pgxp_mode),
      BEETLE_OPT_PGXP_MODE_LABEL_TR,
      NULL,
      BEETLE_OPT_PGXP_MODE_INFO_0_TR,
      NULL,
      "pgxp",
      {
         { "disabled",     NULL },
         { "memory only",  OPTION_VAL_MEMORY_ONLY_TR },
         { "memory + CPU", OPTION_VAL_MEMORY_CPU_TR },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(pgxp_2d_tol),
      BEETLE_OPT_PGXP_2D_TOL_LABEL_TR,
      NULL,
      BEETLE_OPT_PGXP_2D_TOL_INFO_0_TR,
      NULL,
      "pgxp",
      {
         { "disabled", NULL },
         { "0px", OPTION_VAL_0PX_TR },
         { "1px", OPTION_VAL_1PX_TR },
         { "2px", OPTION_VAL_2PX_TR },
         { "3px", OPTION_VAL_3PX_TR },
         { "4px", OPTION_VAL_4PX_TR },
         { "5px", OPTION_VAL_5PX_TR },
         { "6px", OPTION_VAL_6PX_TR },
         { "7px", OPTION_VAL_7PX_TR },
         { "8px", OPTION_VAL_8PX_TR },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(pgxp_nclip),
      BEETLE_OPT_PGXP_NCLIP_LABEL_TR,
      NULL,
      BEETLE_OPT_PGXP_NCLIP_INFO_0_TR,
      NULL,
      "pgxp",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
#if defined(HAVE_OPENGL) || defined(HAVE_OPENGLES) || defined(HAVE_VULKAN)
   {
      BEETLE_OPT(pgxp_vertex),
      BEETLE_OPT_PGXP_VERTEX_LABEL_TR,
      NULL,
      BEETLE_OPT_PGXP_VERTEX_INFO_0_TR,
      NULL,
      "pgxp",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
#endif
   {
      BEETLE_OPT(pgxp_texture),
      BEETLE_OPT_PGXP_TEXTURE_LABEL_TR,
      NULL,
      BEETLE_OPT_PGXP_TEXTURE_INFO_0_TR,
      NULL,
      "pgxp",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(display_internal_fps),
      BEETLE_OPT_DISPLAY_INTERNAL_FPS_LABEL_TR,
      NULL,
      BEETLE_OPT_DISPLAY_INTERNAL_FPS_INFO_0_TR,
      NULL,
      "osd",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(display_osd),
      BEETLE_OPT_DISPLAY_OSD_LABEL_TR,
      NULL,
      BEETLE_OPT_DISPLAY_OSD_INFO_0_TR,
      NULL,
      "osd",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled"
   },
#if defined(HAVE_OPENGL) || defined(HAVE_OPENGLES) || defined(HAVE_VULKAN)
   {
      BEETLE_OPT(display_vram),
      BEETLE_OPT_DISPLAY_VRAM_LABEL_TR,
      NULL,
      BEETLE_OPT_DISPLAY_VRAM_INFO_0_TR,
      NULL,
      "osd",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
#endif
   {
      BEETLE_OPT(analog_calibration),
      BEETLE_OPT_ANALOG_CALIBRATION_LABEL_TR,
      NULL,
      BEETLE_OPT_ANALOG_CALIBRATION_INFO_0_TR,
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(analog_toggle),
      BEETLE_OPT_ANALOG_TOGGLE_LABEL_TR,
      NULL,
      BEETLE_OPT_ANALOG_TOGGLE_INFO_0_TR,
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { "enabled-analog", OPTION_VAL_ENABLED_ANALOG_TR },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(analog_toggle_combo),
      BEETLE_OPT_ANALOG_TOGGLE_COMBO_LABEL_TR,
      NULL,
      BEETLE_OPT_ANALOG_TOGGLE_COMBO_INFO_0_TR,
      NULL,
      "input",
      {
         { "l1+l2+r1+r2+start+select", OPTION_VAL_L1_L2_R1_R2_START_SELECT_TR },
         { "l1+r1+select",             OPTION_VAL_L1_R1_SELECT_TR },
         { "l1+r1+start",              OPTION_VAL_L1_R1_START_TR },
         { "l1+r1+l3",                 OPTION_VAL_L1_R1_L3_TR },
         { "l1+r1+r3",                 OPTION_VAL_L1_R1_R3_TR },
         { "l2+r2+select",             OPTION_VAL_L2_R2_SELECT_TR },
         { "l2+r2+start",              OPTION_VAL_L2_R2_START_TR },
         { "l2+r2+l3",                 OPTION_VAL_L2_R2_L3_TR },
         { "l2+r2+r3",                 OPTION_VAL_L2_R2_R3_TR },
         { "l3+r3",                    OPTION_VAL_L3_R3_TR },
         { NULL, NULL },
      },
      "l1+r1+select"
   },
   {
      BEETLE_OPT(analog_toggle_hold),
      BEETLE_OPT_ANALOG_TOGGLE_HOLD_LABEL_TR,
      NULL,
      BEETLE_OPT_ANALOG_TOGGLE_HOLD_INFO_0_TR,
      NULL,
      "input",
      {
         { "0", OPTION_VAL_0_TR },
         { "1", OPTION_VAL_1_TR },
         { "2", OPTION_VAL_2_TR },
         { "3", OPTION_VAL_3_TR },
         { "4", OPTION_VAL_4_TR },
         { "5", OPTION_VAL_5_TR },
         { NULL, NULL },
      },
      "1"
   },
   {
      BEETLE_OPT(enable_multitap_port1),
      BEETLE_OPT_ENABLE_MULTITAP_PORT1_LABEL_TR,
      NULL,
      BEETLE_OPT_ENABLE_MULTITAP_PORT1_INFO_0_TR,
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(enable_multitap_port2),
      BEETLE_OPT_ENABLE_MULTITAP_PORT2_LABEL_TR,
      NULL,
      BEETLE_OPT_ENABLE_MULTITAP_PORT2_INFO_0_TR,
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(gun_input_mode),
      BEETLE_OPT_GUN_INPUT_MODE_LABEL_TR,
      NULL,
      BEETLE_OPT_GUN_INPUT_MODE_INFO_0_TR,
      NULL,
      "input",
      {
         { "lightgun",    OPTION_VAL_LIGHTGUN_TR },
         { "touchscreen", OPTION_VAL_TOUCHSCREEN_TR },
         { NULL, NULL },
      },
      "lightgun"
   },
   // Shouldn't the gun_input_mode just be Mouse vs. Touchscreen?
   {
      BEETLE_OPT(gun_cursor),
      BEETLE_OPT_GUN_CURSOR_LABEL_TR,
      NULL,
      BEETLE_OPT_GUN_CURSOR_INFO_0_TR,
      NULL,
      "input",
      {
         { "off",   "disabled" },
         { "cross", OPTION_VAL_CROSS_TR },
         { "dot",   OPTION_VAL_DOT_TR },
         { NULL, NULL },
      },
      "cross"
   },
   {
      BEETLE_OPT(crosshair_color_p1),
      BEETLE_OPT_CROSSHAIR_COLOR_P1_LABEL_TR,
      NULL,
      BEETLE_OPT_CROSSHAIR_COLOR_P1_INFO_0_TR,
      NULL,
      "input",
      {
         { "red",    OPTION_VAL_RED_TR },
         { "blue",   OPTION_VAL_BLUE_TR },
         { "green",  OPTION_VAL_GREEN_TR },
         { "orange", OPTION_VAL_ORANGE_TR },
         { "yellow", OPTION_VAL_YELLOW_TR },
         { "cyan",   OPTION_VAL_CYAN_TR },
         { "pink",   OPTION_VAL_PINK_TR },
         { "purple", OPTION_VAL_PURPLE_TR },
         { "black",  OPTION_VAL_BLACK_TR },
         { "white",  OPTION_VAL_WHITE_TR },
         { NULL, NULL },
      },
      "red"
   },
   {
      BEETLE_OPT(crosshair_color_p2),
      BEETLE_OPT_CROSSHAIR_COLOR_P2_LABEL_TR,
      NULL,
      BEETLE_OPT_CROSSHAIR_COLOR_P2_INFO_0_TR,
      NULL,
      "input",
      {
         { "blue",   OPTION_VAL_BLUE_TR },
         { "red",    OPTION_VAL_RED_TR },
         { "green",  OPTION_VAL_GREEN_TR },
         { "orange", OPTION_VAL_ORANGE_TR },
         { "yellow", OPTION_VAL_YELLOW_TR },
         { "cyan",   OPTION_VAL_CYAN_TR },
         { "pink",   OPTION_VAL_PINK_TR },
         { "purple", OPTION_VAL_PURPLE_TR },
         { "black",  OPTION_VAL_BLACK_TR },
         { "white",  OPTION_VAL_WHITE_TR },
         { NULL, NULL },
      },
      "blue"
   },
   {
      BEETLE_OPT(mouse_sensitivity),
      BEETLE_OPT_MOUSE_SENSITIVITY_LABEL_TR,
      NULL,
      BEETLE_OPT_MOUSE_SENSITIVITY_INFO_0_TR,
      NULL,
      "input",
      {
         { "5%",   OPTION_VAL_5_O42_TR },
         { "10%",  OPTION_VAL_10_TR },
         { "15%",  OPTION_VAL_15_TR },
         { "20%",  OPTION_VAL_20_TR },
         { "25%",  OPTION_VAL_25_TR },
         { "30%",  OPTION_VAL_30_TR },
         { "35%",  OPTION_VAL_35_TR },
         { "40%",  OPTION_VAL_40_TR },
         { "45%",  OPTION_VAL_45_TR },
         { "50%",  OPTION_VAL_50_TR },
         { "55%",  OPTION_VAL_55_TR },
         { "60%",  OPTION_VAL_60_TR },
         { "65%",  OPTION_VAL_65_TR },
         { "70%",  OPTION_VAL_70_TR },
         { "75%",  OPTION_VAL_75_TR },
         { "80%",  OPTION_VAL_80_TR },
         { "85%",  OPTION_VAL_85_TR },
         { "90%",  OPTION_VAL_90_TR },
         { "95%",  OPTION_VAL_95_TR },
         { "100%", OPTION_VAL_100_O42_TR },
         { "105%", OPTION_VAL_105_TR },
         { "110%", OPTION_VAL_110_TR },
         { "115%", OPTION_VAL_115_TR },
         { "120%", OPTION_VAL_120_TR },
         { "125%", OPTION_VAL_125_TR },
         { "130%", OPTION_VAL_130_TR },
         { "135%", OPTION_VAL_135_TR },
         { "140%", OPTION_VAL_140_TR },
         { "145%", OPTION_VAL_145_TR },
         { "150%", OPTION_VAL_150_TR },
         { "155%", OPTION_VAL_155_TR },
         { "160%", OPTION_VAL_160_TR },
         { "165%", OPTION_VAL_165_TR },
         { "170%", OPTION_VAL_170_TR },
         { "175%", OPTION_VAL_175_TR },
         { "180%", OPTION_VAL_180_TR },
         { "185%", OPTION_VAL_185_TR },
         { "190%", OPTION_VAL_190_TR },
         { "195%", OPTION_VAL_195_TR },
         { "200%", OPTION_VAL_200_TR },
         { NULL, NULL },
      },
      "100%"
   },
   {
      BEETLE_OPT(negcon_response),
      BEETLE_OPT_NEGCON_RESPONSE_LABEL_TR,
      NULL,
      BEETLE_OPT_NEGCON_RESPONSE_INFO_0_TR,
      NULL,
      "input",
      {
         { "linear",    OPTION_VAL_LINEAR_TR },
         { "quadratic", OPTION_VAL_QUADRATIC_TR },
         { "cubic",     OPTION_VAL_CUBIC_TR },
         { NULL, NULL },
      },
      "linear"
   },
   {
      BEETLE_OPT(negcon_deadzone),
      BEETLE_OPT_NEGCON_DEADZONE_LABEL_TR,
      NULL,
      BEETLE_OPT_NEGCON_DEADZONE_INFO_0_TR,
      NULL,
      "input",
      {
         { "0%",  OPTION_VAL_0_O44_TR },
         { "5%",  OPTION_VAL_5_O42_TR },
         { "10%", OPTION_VAL_10_TR },
         { "15%", OPTION_VAL_15_TR },
         { "20%", OPTION_VAL_20_TR },
         { "25%", OPTION_VAL_25_TR },
         { "30%", OPTION_VAL_30_TR },
         { NULL, NULL },
      },
      "0%"
   },
   {
      BEETLE_OPT(use_mednafen_memcard0_method),
      BEETLE_OPT_USE_MEDNAFEN_MEMCARD0_METHOD_LABEL_TR,
      NULL,
      BEETLE_OPT_USE_MEDNAFEN_MEMCARD0_METHOD_INFO_0_TR,
      NULL,
      "memcards",
      {
         { "libretro", OPTION_VAL_LIBRETRO_TR },
         { "mednafen", OPTION_VAL_MEDNAFEN_TR },
         { NULL, NULL },
      },
      "libretro"
   },
   {
      BEETLE_OPT(enable_memcard1),
      BEETLE_OPT_ENABLE_MEMCARD1_LABEL_TR,
      NULL,
      BEETLE_OPT_ENABLE_MEMCARD1_INFO_0_TR,
      NULL,
      "memcards",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(shared_memory_cards),
      BEETLE_OPT_SHARED_MEMORY_CARDS_LABEL_TR,
      NULL,
      BEETLE_OPT_SHARED_MEMORY_CARDS_INFO_0_TR,
      NULL,
      "memcards",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(memcard_left_index),
      BEETLE_OPT_MEMCARD_LEFT_INDEX_LABEL_TR,
      NULL,
      BEETLE_OPT_MEMCARD_LEFT_INDEX_INFO_0_TR,
      NULL,
      "memcards",
      {
         { "0",  OPTION_VAL_0_O48_TR },
         { "1",  NULL },
         { "2",  NULL },
         { "3",  NULL },
         { "4",  NULL },
         { "5",  NULL },
         { "6",  NULL },
         { "7",  NULL },
         { "8",  NULL },
         { "9",  NULL },
         { "10", NULL },
         { "11", NULL },
         { "12", NULL },
         { "13", NULL },
         { "14", NULL },
         { "15", NULL },
         { "16", NULL },
         { "17", NULL },
         { "18", NULL },
         { "19", NULL },
         { "20", NULL },
         { "21", NULL },
         { "22", NULL },
         { "23", NULL },
         { "24", NULL },
         { "25", NULL },
         { "26", NULL },
         { "27", NULL },
         { "28", NULL },
         { "29", NULL },
         { "30", NULL },
         { "31", NULL },
         { "32", NULL },
         { "33", NULL },
         { "34", NULL },
         { "35", NULL },
         { "36", NULL },
         { "37", NULL },
         { "38", NULL },
         { "39", NULL },
         { "40", NULL },
         { "41", NULL },
         { "42", NULL },
         { "43", NULL },
         { "44", NULL },
         { "45", NULL },
         { "46", NULL },
         { "47", NULL },
         { "48", NULL },
         { "49", NULL },
         { "50", NULL },
         { "51", NULL },
         { "52", NULL },
         { "53", NULL },
         { "54", NULL },
         { "55", NULL },
         { "56", NULL },
         { "57", NULL },
         { "58", NULL },
         { "59", NULL },
         { "60", NULL },
         { "61", NULL },
         { "62", NULL },
         { "63", NULL },
         { NULL, NULL },
      },
      "0"
   },
   {
      BEETLE_OPT(memcard_right_index),
      BEETLE_OPT_MEMCARD_RIGHT_INDEX_LABEL_TR,
      NULL,
      BEETLE_OPT_MEMCARD_RIGHT_INDEX_INFO_0_TR,
      NULL,
      "memcards",
      {
         { "0",  NULL },
         { "1",  OPTION_VAL_1_O49_TR },
         { "2",  NULL },
         { "3",  NULL },
         { "4",  NULL },
         { "5",  NULL },
         { "6",  NULL },
         { "7",  NULL },
         { "8",  NULL },
         { "9",  NULL },
         { "10", NULL },
         { "11", NULL },
         { "12", NULL },
         { "13", NULL },
         { "14", NULL },
         { "15", NULL },
         { "16", NULL },
         { "17", NULL },
         { "18", NULL },
         { "19", NULL },
         { "20", NULL },
         { "21", NULL },
         { "22", NULL },
         { "23", NULL },
         { "24", NULL },
         { "25", NULL },
         { "26", NULL },
         { "27", NULL },
         { "28", NULL },
         { "29", NULL },
         { "30", NULL },
         { "31", NULL },
         { "32", NULL },
         { "33", NULL },
         { "34", NULL },
         { "35", NULL },
         { "36", NULL },
         { "37", NULL },
         { "38", NULL },
         { "39", NULL },
         { "40", NULL },
         { "41", NULL },
         { "42", NULL },
         { "43", NULL },
         { "44", NULL },
         { "45", NULL },
         { "46", NULL },
         { "47", NULL },
         { "48", NULL },
         { "49", NULL },
         { "50", NULL },
         { "51", NULL },
         { "52", NULL },
         { "53", NULL },
         { "54", NULL },
         { "55", NULL },
         { "56", NULL },
         { "57", NULL },
         { "58", NULL },
         { "59", NULL },
         { "60", NULL },
         { "61", NULL },
         { "62", NULL },
         { "63", NULL },
         { NULL, NULL },
      },
      "1"
   },
   {
      BEETLE_OPT(line_render),
      BEETLE_OPT_LINE_RENDER_LABEL_TR,
      NULL,
      BEETLE_OPT_LINE_RENDER_INFO_0_TR,
      NULL,
      "hacks",
      {
         { "default",    OPTION_VAL_DEFAULT_TR },
         { "aggressive", OPTION_VAL_AGGRESSIVE_TR },
         { "disabled",   NULL },
         { NULL, NULL },
      },
      "default"
   },
   {
      BEETLE_OPT(widescreen_hack),
      BEETLE_OPT_WIDESCREEN_HACK_LABEL_TR,
      NULL,
      BEETLE_OPT_WIDESCREEN_HACK_INFO_0_TR,
      NULL,
      "hacks",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(widescreen_hack_aspect_ratio),
      BEETLE_OPT_WIDESCREEN_HACK_ASPECT_RATIO_LABEL_TR,
      NULL,
      BEETLE_OPT_WIDESCREEN_HACK_ASPECT_RATIO_INFO_0_TR,
      NULL,
      "hacks",
      {
         { "16:9",  OPTION_VAL_16_9_TR },
         { "16:10", OPTION_VAL_16_10_TR },
         { "18:9",  OPTION_VAL_18_9_TR },
         { "19:9",  OPTION_VAL_19_9_TR },
         { "20:9",  OPTION_VAL_20_9_TR },
         { "21:9",  OPTION_VAL_21_9_TR }, // 64:27
         { "32:9",  OPTION_VAL_32_9_TR },
         { NULL,    NULL },
      },
      "16:9"
   },
#if defined(HAVE_LIGHTREC)
   {
      BEETLE_OPT(cpu_dynarec),
      BEETLE_OPT_CPU_DYNAREC_LABEL_TR,
      NULL,
      BEETLE_OPT_CPU_DYNAREC_INFO_0_TR,
      NULL,
      "hacks",
      {
         { "disabled",        OPTION_VAL_DISABLED_TR },
         { "execute",         OPTION_VAL_EXECUTE_TR },
         { "execute_one",     OPTION_VAL_EXECUTE_ONE_TR },
         { "run_interpreter", OPTION_VAL_RUN_INTERPRETER_TR },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(dynarec_invalidate),
      BEETLE_OPT_DYNAREC_INVALIDATE_LABEL_TR,
      NULL,
      BEETLE_OPT_DYNAREC_INVALIDATE_INFO_0_TR,
      NULL,
      "hacks",
      {
         { "full", OPTION_VAL_FULL_TR },
         { "dma",  OPTION_VAL_DMA_TR },
         { NULL, NULL },
      },
      "full"
   },
   {
      BEETLE_OPT(dynarec_eventcycles),
      BEETLE_OPT_DYNAREC_EVENTCYCLES_LABEL_TR,
      NULL,
      BEETLE_OPT_DYNAREC_EVENTCYCLES_INFO_0_TR,
      NULL,
      "hacks",
      {
         { "128",  OPTION_VAL_128_TR },
         { "256",  NULL },
         { "384",  NULL },
         { "512",  NULL },
         { "640",  NULL },
         { "768",  NULL },
         { "896",  NULL },
         { "1024", NULL },
         { "1152", NULL },
         { "1280", NULL },
         { "1408", NULL },
         { "1536", NULL },
         { "1664", NULL },
         { "1792", NULL },
         { "1920", NULL },
         { "2048", NULL },
         { NULL, NULL },
      },
      "128"
   },
   {
      BEETLE_OPT(dynarec_spu_samples),
      BEETLE_OPT_DYNAREC_SPU_SAMPLES_LABEL_TR,
      NULL,
      BEETLE_OPT_DYNAREC_SPU_SAMPLES_INFO_0_TR,
      NULL,
      "hacks",
      {
         { "1",  OPTION_VAL_1_O49_TR },
         { "4",  NULL },
         { "16", NULL },
         { NULL, NULL },
      },
      "1"
   },
#endif
   {
      BEETLE_OPT(core_timing_fps),
      BEETLE_OPT_CORE_TIMING_FPS_LABEL_TR,
      NULL,
      BEETLE_OPT_CORE_TIMING_FPS_INFO_0_TR,
      NULL,
      "video",
      {
         { "auto_toggle",       OPTION_VAL_AUTO_TOGGLE_TR },
         { "force_progressive", OPTION_VAL_FORCE_PROGRESSIVE_TR },
         { "force_interlaced",  OPTION_VAL_FORCE_INTERLACED_TR },
      },
      "auto_toggle"
   },
   {
      BEETLE_OPT(deinterlacer),
      BEETLE_OPT_DEINTERLACER_LABEL_TR,
      NULL,
      BEETLE_OPT_DEINTERLACER_INFO_0_TR,
      NULL,
      "video",
      {
         { "weave",      OPTION_VAL_WEAVE_TR },
         { "bob",        OPTION_VAL_BOB_TR },
         { "bob_offset", OPTION_VAL_BOB_OFFSET_TR },
         { "fastmad",    OPTION_VAL_FASTMAD_TR },
         { "off",        "Off" },
         { NULL, NULL },
      },
      "weave"
   },
   {
      BEETLE_OPT(frame_duping),
      BEETLE_OPT_FRAME_DUPING_LABEL_TR,
      NULL,
      BEETLE_OPT_FRAME_DUPING_INFO_0_TR,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(pal_video_timing_override),
      BEETLE_OPT_PAL_VIDEO_TIMING_OVERRIDE_LABEL_TR,
      NULL,
      BEETLE_OPT_PAL_VIDEO_TIMING_OVERRIDE_INFO_0_TR,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(aspect_ratio),
      BEETLE_OPT_ASPECT_RATIO_LABEL_TR,
      NULL,
      BEETLE_OPT_ASPECT_RATIO_INFO_0_TR,
      NULL,
      "video",
      {
         { "corrected", OPTION_VAL_CORRECTED_TR },
         { "uncorrected", OPTION_VAL_UNCORRECTED_TR },
         { "4:3",  OPTION_VAL_4_3_TR },
         { "ntsc", OPTION_VAL_NTSC_TR },
      },
      "corrected"
   },
   {
      BEETLE_OPT(crop_overscan),
      BEETLE_OPT_CROP_OVERSCAN_LABEL_TR,
      NULL,
      BEETLE_OPT_CROP_OVERSCAN_INFO_0_TR,
      NULL,
      "video",
      {
         { "disabled", "Disabled" },
         { "static",   OPTION_VAL_STATIC_TR },
         { "smart",    OPTION_VAL_SMART_TR },
         { NULL, NULL },
      },
      "smart"
   },
   {
      BEETLE_OPT(image_crop),
      BEETLE_OPT_IMAGE_CROP_LABEL_TR,
      NULL,
      BEETLE_OPT_IMAGE_CROP_INFO_0_TR,
      NULL,
      "video",
      {
         { "disabled", "0" },
         { "1px",      OPTION_VAL_1PX_TR },
         { "2px",      OPTION_VAL_2PX_TR },
         { "3px",      OPTION_VAL_3PX_TR },
         { "4px",      OPTION_VAL_4PX_TR },
         { "5px",      OPTION_VAL_5PX_TR },
         { "6px",      OPTION_VAL_6PX_TR },
         { "7px",      OPTION_VAL_7PX_TR },
         { "8px",      OPTION_VAL_8PX_TR },
         { "9px",      OPTION_VAL_9PX_TR },
         { "10px",     OPTION_VAL_10PX_TR },
         { "11px",     OPTION_VAL_11PX_TR },
         { "12px",     OPTION_VAL_12PX_TR },
         { "13px",     OPTION_VAL_13PX_TR },
         { "14px",     OPTION_VAL_14PX_TR },
         { "15px",     OPTION_VAL_15PX_TR },
         { "16px",     OPTION_VAL_16PX_TR },
         { "17px",     OPTION_VAL_17PX_TR },
         { "18px",     OPTION_VAL_18PX_TR },
         { "19px",     OPTION_VAL_19PX_TR },
         { "20px",     OPTION_VAL_20PX_TR },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(image_offset),
      BEETLE_OPT_IMAGE_OFFSET_LABEL_TR,
      NULL,
      BEETLE_OPT_IMAGE_OFFSET_INFO_0_TR,
      NULL,
      "video",
      {
         { "-12px",    OPTION_VAL_12PX_O64_TR },
         { "-11px",    OPTION_VAL_11PX_O64_TR },
         { "-10px",    OPTION_VAL_10PX_O64_TR },
         { "-9px",     OPTION_VAL_9PX_O64_TR },
         { "-8px",     OPTION_VAL_8PX_O64_TR },
         { "-7px",     OPTION_VAL_7PX_O64_TR },
         { "-6px",     OPTION_VAL_6PX_O64_TR },
         { "-5px",     OPTION_VAL_5PX_O64_TR },
         { "-4px",     OPTION_VAL_4PX_O64_TR },
         { "-3px",     OPTION_VAL_3PX_O64_TR },
         { "-2px",     OPTION_VAL_2PX_O64_TR },
         { "-1px",     OPTION_VAL_1PX_O64_TR },
         { "disabled", "0" },
         { "+1px",     OPTION_VAL_1PX_O64_0_TR },
         { "+2px",     OPTION_VAL_2PX_O64_0_TR },
         { "+3px",     OPTION_VAL_3PX_O64_0_TR },
         { "+4px",     OPTION_VAL_4PX_O64_0_TR },
         { "+5px",     OPTION_VAL_5PX_O64_0_TR },
         { "+6px",     OPTION_VAL_6PX_O64_0_TR },
         { "+7px",     OPTION_VAL_7PX_O64_0_TR },
         { "+8px",     OPTION_VAL_8PX_O64_0_TR },
         { "+9px",     OPTION_VAL_9PX_O64_0_TR },
         { "+10px",    OPTION_VAL_10PX_O64_0_TR },
         { "+11px",    OPTION_VAL_11PX_O64_0_TR },
         { "+12px",    OPTION_VAL_12PX_O64_0_TR },
         { NULL, NULL },
      },
      "disabled"
   },
#if defined(HAVE_OPENGL) || defined(HAVE_OPENGLES) || defined(HAVE_VULKAN)
   {
      BEETLE_OPT(image_offset_cycles),
      BEETLE_OPT_IMAGE_OFFSET_CYCLES_LABEL_TR,
      NULL,
      BEETLE_OPT_IMAGE_OFFSET_CYCLES_INFO_0_TR,
      NULL,
      "video",
      {
         { "-40",      NULL },
         { "-39",      NULL },
         { "-38",      NULL },
         { "-37",      NULL },
         { "-36",      NULL },
         { "-35",      NULL },
         { "-34",      NULL },
         { "-33",      NULL },
         { "-32",      NULL },
         { "-31",      NULL },
         { "-30",      NULL },
         { "-29",      NULL },
         { "-28",      NULL },
         { "-27",      NULL },
         { "-26",      NULL },
         { "-25",      NULL },
         { "-24",      NULL },
         { "-23",      NULL },
         { "-22",      NULL },
         { "-21",      NULL },
         { "-20",      NULL },
         { "-19",      NULL },
         { "-18",      NULL },
         { "-17",      NULL },
         { "-16",      NULL },
         { "-15",      NULL },
         { "-14",      NULL },
         { "-13",      NULL },
         { "-12",      NULL },
         { "-11",      NULL },
         { "-10",      NULL },
         { "-9",       NULL },
         { "-8",       NULL },
         { "-7",       NULL },
         { "-6",       NULL },
         { "-5",       NULL },
         { "-4",       NULL },
         { "-3",       NULL },
         { "-2",       NULL },
         { "-1",       NULL },
         { "0",        "0" },
         { "+1",       NULL },
         { "+2",       NULL },
         { "+3",       NULL },
         { "+4",       NULL },
         { "+5",       NULL },
         { "+6",       NULL },
         { "+7",       NULL },
         { "+8",       NULL },
         { "+9",       NULL },
         { "+10",      NULL },
         { "+11",      NULL },
         { "+12",      NULL },
         { "+13",      NULL },
         { "+14",      NULL },
         { "+15",      NULL },
         { "+16",      NULL },
         { "+17",      NULL },
         { "+18",      NULL },
         { "+19",      NULL },
         { "+20",      NULL },
         { "+21",      NULL },
         { "+22",      NULL },
         { "+23",      NULL },
         { "+24",      NULL },
         { "+25",      NULL },
         { "+26",      NULL },
         { "+27",      NULL },
         { "+28",      NULL },
         { "+29",      NULL },
         { "+30",      NULL },
         { "+31",      NULL },
         { "+32",      NULL },
         { "+33",      NULL },
         { "+34",      NULL },
         { "+35",      NULL },
         { "+36",      NULL },
         { "+37",      NULL },
         { "+38",      NULL },
         { "+39",      NULL },
         { "+40",      NULL },
         { NULL, NULL},
      },
      "0"
   },
#endif
   {
      BEETLE_OPT(initial_scanline),
      BEETLE_OPT_INITIAL_SCANLINE_LABEL_TR,
      NULL,
      BEETLE_OPT_INITIAL_SCANLINE_INFO_0_TR,
      NULL,
      "video",
      {
         { "0",  OPTION_VAL_0_O48_TR },
         { "1",  NULL },
         { "2",  NULL },
         { "3",  NULL },
         { "4",  NULL },
         { "5",  NULL },
         { "6",  NULL },
         { "7",  NULL },
         { "8",  NULL },
         { "9",  NULL },
         { "10", NULL },
         { "11", NULL },
         { "12", NULL },
         { "13", NULL },
         { "14", NULL },
         { "15", NULL },
         { "16", NULL },
         { "17", NULL },
         { "18", NULL },
         { "19", NULL },
         { "20", NULL },
         { "21", NULL },
         { "22", NULL },
         { "23", NULL },
         { "24", NULL },
         { "25", NULL },
         { "26", NULL },
         { "27", NULL },
         { "28", NULL },
         { "29", NULL },
         { "30", NULL },
         { "31", NULL },
         { "32", NULL },
         { "33", NULL },
         { "34", NULL },
         { "35", NULL },
         { "36", NULL },
         { "37", NULL },
         { "38", NULL },
         { "39", NULL },
         { "40", NULL },
         { NULL, NULL },
      },
      "0"
   },
   {
      BEETLE_OPT(last_scanline),
      BEETLE_OPT_LAST_SCANLINE_LABEL_TR,
      NULL,
      BEETLE_OPT_LAST_SCANLINE_INFO_0_TR,
      NULL,
      "video",
      {
         { "210", NULL },
         { "211", NULL },
         { "212", NULL },
         { "213", NULL },
         { "214", NULL },
         { "215", NULL },
         { "216", NULL },
         { "217", NULL },
         { "218", NULL },
         { "219", NULL },
         { "220", NULL },
         { "221", NULL },
         { "222", NULL },
         { "223", NULL },
         { "224", NULL },
         { "225", NULL },
         { "226", NULL },
         { "227", NULL },
         { "228", NULL },
         { "229", NULL },
         { "230", NULL },
         { "231", NULL },
         { "232", NULL },
         { "233", NULL },
         { "234", NULL },
         { "235", NULL },
         { "236", NULL },
         { "237", NULL },
         { "238", NULL },
         { "239", OPTION_VAL_239_TR },
         { NULL, NULL },
      },
      "239"
   },
   {
      BEETLE_OPT(initial_scanline_pal),
      BEETLE_OPT_INITIAL_SCANLINE_PAL_LABEL_TR,
      NULL,
      BEETLE_OPT_INITIAL_SCANLINE_PAL_INFO_0_TR,
      NULL,
      "video",
      {
         { "0",  OPTION_VAL_0_O48_TR },
         { "1",  NULL },
         { "2",  NULL },
         { "3",  NULL },
         { "4",  NULL },
         { "5",  NULL },
         { "6",  NULL },
         { "7",  NULL },
         { "8",  NULL },
         { "9",  NULL },
         { "10", NULL },
         { "11", NULL },
         { "12", NULL },
         { "13", NULL },
         { "14", NULL },
         { "15", NULL },
         { "16", NULL },
         { "17", NULL },
         { "18", NULL },
         { "19", NULL },
         { "20", NULL },
         { "21", NULL },
         { "22", NULL },
         { "23", NULL },
         { "24", NULL },
         { "25", NULL },
         { "26", NULL },
         { "27", NULL },
         { "28", NULL },
         { "29", NULL },
         { "30", NULL },
         { "31", NULL },
         { "32", NULL },
         { "33", NULL },
         { "34", NULL },
         { "35", NULL },
         { "36", NULL },
         { "37", NULL },
         { "38", NULL },
         { "39", NULL },
         { "40", NULL },
         { NULL, NULL },
      },
      "0"
   },
   {
      BEETLE_OPT(last_scanline_pal),
      BEETLE_OPT_LAST_SCANLINE_PAL_LABEL_TR,
      NULL,
      BEETLE_OPT_LAST_SCANLINE_PAL_INFO_0_TR,
      NULL,
      "video",
      {
         { "230", NULL },
         { "231", NULL },
         { "232", NULL },
         { "233", NULL },
         { "234", NULL },
         { "235", NULL },
         { "236", NULL },
         { "237", NULL },
         { "238", NULL },
         { "239", NULL },
         { "240", NULL },
         { "241", NULL },
         { "242", NULL },
         { "243", NULL },
         { "244", NULL },
         { "245", NULL },
         { "246", NULL },
         { "247", NULL },
         { "248", NULL },
         { "249", NULL },
         { "250", NULL },
         { "251", NULL },
         { "252", NULL },
         { "253", NULL },
         { "254", NULL },
         { "255", NULL },
         { "256", NULL },
         { "257", NULL },
         { "258", NULL },
         { "259", NULL },
         { "260", NULL },
         { "261", NULL },
         { "262", NULL },
         { "263", NULL },
         { "264", NULL },
         { "265", NULL },
         { "266", NULL },
         { "267", NULL },
         { "268", NULL },
         { "269", NULL },
         { "270", NULL },
         { "271", NULL },
         { "272", NULL },
         { "273", NULL },
         { "274", NULL },
         { "275", NULL },
         { "276", NULL },
         { "277", NULL },
         { "278", NULL },
         { "279", NULL },
         { "280", NULL },
         { "281", NULL },
         { "282", NULL },
         { "283", NULL },
         { "284", NULL },
         { "285", NULL },
         { "286", NULL },
         { "287", OPTION_VAL_287_TR },
         { NULL, NULL },
      },
      "287"
   },
   { NULL, NULL, NULL, NULL, NULL, NULL, {{0}}, NULL },
};
struct retro_core_options_v2 options_tr = {
   option_cats_tr,
   option_defs_tr
};

/* RETRO_LANGUAGE_TT */

#define CATEGORY_SYSTEM_LABEL_TT NULL
#define CATEGORY_SYSTEM_INFO_0_TT NULL
#define CATEGORY_VIDEO_LABEL_TT "Видео"
#define CATEGORY_VIDEO_INFO_0_TT NULL
#define CATEGORY_PGXP_LABEL_TT NULL
#define CATEGORY_PGXP_INFO_0_TT NULL
#define CATEGORY_OSD_LABEL_TT NULL
#define CATEGORY_OSD_INFO_0_TT NULL
#define CATEGORY_INPUT_LABEL_TT NULL
#define CATEGORY_INPUT_INFO_0_TT NULL
#define CATEGORY_MEMCARDS_LABEL_TT NULL
#define CATEGORY_MEMCARDS_INFO_0_TT NULL
#define CATEGORY_HACKS_LABEL_TT NULL
#define CATEGORY_HACKS_INFO_0_TT NULL
#define BEETLE_OPT_CPU_FREQ_SCALE_LABEL_TT NULL
#define BEETLE_OPT_CPU_FREQ_SCALE_INFO_0_TT NULL
#define OPTION_VAL_50_TT NULL
#define OPTION_VAL_60_TT NULL
#define OPTION_VAL_70_TT NULL
#define OPTION_VAL_80_TT NULL
#define OPTION_VAL_90_TT NULL
#define OPTION_VAL_100_TT NULL
#define OPTION_VAL_110_TT NULL
#define OPTION_VAL_120_TT NULL
#define OPTION_VAL_130_TT NULL
#define OPTION_VAL_140_TT NULL
#define OPTION_VAL_150_TT NULL
#define OPTION_VAL_160_TT NULL
#define OPTION_VAL_170_TT NULL
#define OPTION_VAL_180_TT NULL
#define OPTION_VAL_190_TT NULL
#define OPTION_VAL_200_TT NULL
#define OPTION_VAL_210_TT NULL
#define OPTION_VAL_220_TT NULL
#define OPTION_VAL_230_TT NULL
#define OPTION_VAL_240_TT NULL
#define OPTION_VAL_250_TT NULL
#define OPTION_VAL_260_TT NULL
#define OPTION_VAL_270_TT NULL
#define OPTION_VAL_280_TT NULL
#define OPTION_VAL_290_TT NULL
#define OPTION_VAL_300_TT NULL
#define OPTION_VAL_310_TT NULL
#define OPTION_VAL_320_TT NULL
#define OPTION_VAL_330_TT NULL
#define OPTION_VAL_340_TT NULL
#define OPTION_VAL_350_TT NULL
#define OPTION_VAL_360_TT NULL
#define OPTION_VAL_370_TT NULL
#define OPTION_VAL_380_TT NULL
#define OPTION_VAL_390_TT NULL
#define OPTION_VAL_400_TT NULL
#define OPTION_VAL_410_TT NULL
#define OPTION_VAL_420_TT NULL
#define OPTION_VAL_430_TT NULL
#define OPTION_VAL_440_TT NULL
#define OPTION_VAL_450_TT NULL
#define OPTION_VAL_460_TT NULL
#define OPTION_VAL_470_TT NULL
#define OPTION_VAL_480_TT NULL
#define OPTION_VAL_490_TT NULL
#define OPTION_VAL_500_TT NULL
#define OPTION_VAL_510_TT NULL
#define OPTION_VAL_520_TT NULL
#define OPTION_VAL_530_TT NULL
#define OPTION_VAL_540_TT NULL
#define OPTION_VAL_550_TT NULL
#define OPTION_VAL_560_TT NULL
#define OPTION_VAL_570_TT NULL
#define OPTION_VAL_580_TT NULL
#define OPTION_VAL_590_TT NULL
#define OPTION_VAL_600_TT NULL
#define OPTION_VAL_610_TT NULL
#define OPTION_VAL_620_TT NULL
#define OPTION_VAL_630_TT NULL
#define OPTION_VAL_640_TT NULL
#define OPTION_VAL_650_TT NULL
#define OPTION_VAL_660_TT NULL
#define OPTION_VAL_670_TT NULL
#define OPTION_VAL_680_TT NULL
#define OPTION_VAL_690_TT NULL
#define OPTION_VAL_700_TT NULL
#define OPTION_VAL_710_TT NULL
#define OPTION_VAL_720_TT NULL
#define OPTION_VAL_730_TT NULL
#define OPTION_VAL_740_TT NULL
#define OPTION_VAL_750_TT NULL
#define BEETLE_OPT_GPU_OVERCLOCK_LABEL_TT NULL
#define BEETLE_OPT_GPU_OVERCLOCK_INFO_0_TT NULL
#define OPTION_VAL_1X_NATIVE_TT NULL
#define OPTION_VAL_2X_TT NULL
#define OPTION_VAL_4X_TT NULL
#define OPTION_VAL_8X_TT NULL
#define OPTION_VAL_16X_TT NULL
#define OPTION_VAL_32X_TT NULL
#define BEETLE_OPT_GTE_OVERCLOCK_LABEL_TT NULL
#define BEETLE_OPT_GTE_OVERCLOCK_INFO_0_TT NULL
#define BEETLE_OPT_CD_ACCESS_METHOD_LABEL_TT NULL
#define BEETLE_OPT_CD_ACCESS_METHOD_INFO_0_TT NULL
#define OPTION_VAL_SYNC_TT NULL
#define OPTION_VAL_ASYNC_TT NULL
#define OPTION_VAL_PRECACHE_TT NULL
#define BEETLE_OPT_CD_FASTLOAD_LABEL_TT NULL
#define BEETLE_OPT_CD_FASTLOAD_INFO_0_TT NULL
#define OPTION_VAL_2X_NATIVE_TT NULL
#define OPTION_VAL_6X_TT NULL
#define OPTION_VAL_10X_TT NULL
#define OPTION_VAL_12X_TT NULL
#define OPTION_VAL_14X_TT NULL
#define BEETLE_OPT_SKIP_BIOS_LABEL_TT NULL
#define BEETLE_OPT_SKIP_BIOS_INFO_0_TT NULL
#define BEETLE_OPT_OVERRIDE_BIOS_LABEL_TT NULL
#define BEETLE_OPT_OVERRIDE_BIOS_INFO_0_TT NULL
#define OPTION_VAL_PSXONPSP_TT NULL
#define OPTION_VAL_PS1_ROM_TT NULL
#define OPTION_VAL_OPENBIOS_TT NULL
#define BEETLE_OPT_REGION_LABEL_TT NULL
#define BEETLE_OPT_REGION_INFO_0_TT NULL
#define OPTION_VAL_AUTO_TT NULL
#define OPTION_VAL_NTSC_J_TT NULL
#define OPTION_VAL_NTSC_U_TT NULL
#define OPTION_VAL_PAL_TT NULL
#define BEETLE_OPT_INTERNAL_RESOLUTION_LABEL_TT NULL
#define BEETLE_OPT_INTERNAL_RESOLUTION_INFO_0_TT NULL
#define BEETLE_OPT_RENDERER_LABEL_TT NULL
#define BEETLE_OPT_RENDERER_INFO_0_TT NULL
#define OPTION_VAL_HARDWARE_TT NULL
#define OPTION_VAL_HARDWARE_GL_TT NULL
#define OPTION_VAL_HARDWARE_VK_TT NULL
#define OPTION_VAL_SOFTWARE_TT NULL
#define BEETLE_OPT_RENDERER_SOFTWARE_FB_LABEL_TT NULL
#define BEETLE_OPT_RENDERER_SOFTWARE_FB_INFO_0_TT NULL
#define BEETLE_OPT_DEPTH_LABEL_TT NULL
#define BEETLE_OPT_DEPTH_INFO_0_TT NULL
#define OPTION_VAL_16BPP_NATIVE_TT NULL
#define OPTION_VAL_32BPP_TT NULL
#define BEETLE_OPT_DITHER_MODE_LABEL_TT NULL
#define BEETLE_OPT_DITHER_MODE_INFO_0_TT NULL
#define OPTION_VAL_INTERNAL_RESOLUTION_TT NULL
#define BEETLE_OPT_SCALED_UV_OFFSET_LABEL_TT NULL
#define BEETLE_OPT_SCALED_UV_OFFSET_INFO_0_TT NULL
#define BEETLE_OPT_FILTER_LABEL_TT NULL
#define BEETLE_OPT_FILTER_INFO_0_TT NULL
#define OPTION_VAL_NEAREST_TT NULL
#define OPTION_VAL_SABR_TT NULL
#define OPTION_VAL_XBR_TT NULL
#define OPTION_VAL_BILINEAR_TT NULL
#define OPTION_VAL_3_POINT_TT NULL
#define OPTION_VAL_JINC2_TT NULL
#define BEETLE_OPT_FILTER_EXCLUDE_SPRITE_LABEL_TT NULL
#define BEETLE_OPT_FILTER_EXCLUDE_SPRITE_INFO_0_TT NULL
#define OPTION_VAL_OPAQUE_TT NULL
#define OPTION_VAL_ALL_TT NULL
#define BEETLE_OPT_FILTER_EXCLUDE_2D_POLYGON_LABEL_TT NULL
#define BEETLE_OPT_FILTER_EXCLUDE_2D_POLYGON_INFO_0_TT NULL
#define BEETLE_OPT_ADAPTIVE_SMOOTHING_LABEL_TT NULL
#define BEETLE_OPT_ADAPTIVE_SMOOTHING_INFO_0_TT NULL
#define BEETLE_OPT_SUPER_SAMPLING_LABEL_TT NULL
#define BEETLE_OPT_SUPER_SAMPLING_INFO_0_TT NULL
#define BEETLE_OPT_MSAA_LABEL_TT NULL
#define BEETLE_OPT_MSAA_INFO_0_TT NULL
#define OPTION_VAL_1X_TT NULL
#define BEETLE_OPT_MDEC_YUV_LABEL_TT NULL
#define BEETLE_OPT_MDEC_YUV_INFO_0_TT NULL
#define BEETLE_OPT_TRACK_TEXTURES_LABEL_TT NULL
#define BEETLE_OPT_TRACK_TEXTURES_INFO_0_TT NULL
#define BEETLE_OPT_DUMP_TEXTURES_LABEL_TT NULL
#define BEETLE_OPT_DUMP_TEXTURES_INFO_0_TT NULL
#define BEETLE_OPT_REPLACE_TEXTURES_LABEL_TT NULL
#define BEETLE_OPT_REPLACE_TEXTURES_INFO_0_TT NULL
#define BEETLE_OPT_PGXP_MODE_LABEL_TT NULL
#define BEETLE_OPT_PGXP_MODE_INFO_0_TT NULL
#define OPTION_VAL_MEMORY_ONLY_TT NULL
#define OPTION_VAL_MEMORY_CPU_TT NULL
#define BEETLE_OPT_PGXP_2D_TOL_LABEL_TT NULL
#define BEETLE_OPT_PGXP_2D_TOL_INFO_0_TT NULL
#define OPTION_VAL_0PX_TT NULL
#define OPTION_VAL_1PX_TT NULL
#define OPTION_VAL_2PX_TT NULL
#define OPTION_VAL_3PX_TT NULL
#define OPTION_VAL_4PX_TT NULL
#define OPTION_VAL_5PX_TT NULL
#define OPTION_VAL_6PX_TT NULL
#define OPTION_VAL_7PX_TT NULL
#define OPTION_VAL_8PX_TT NULL
#define BEETLE_OPT_PGXP_NCLIP_LABEL_TT NULL
#define BEETLE_OPT_PGXP_NCLIP_INFO_0_TT NULL
#define BEETLE_OPT_PGXP_VERTEX_LABEL_TT NULL
#define BEETLE_OPT_PGXP_VERTEX_INFO_0_TT NULL
#define BEETLE_OPT_PGXP_TEXTURE_LABEL_TT NULL
#define BEETLE_OPT_PGXP_TEXTURE_INFO_0_TT NULL
#define BEETLE_OPT_DISPLAY_INTERNAL_FPS_LABEL_TT NULL
#define BEETLE_OPT_DISPLAY_INTERNAL_FPS_INFO_0_TT NULL
#define BEETLE_OPT_DISPLAY_OSD_LABEL_TT NULL
#define BEETLE_OPT_DISPLAY_OSD_INFO_0_TT NULL
#define BEETLE_OPT_DISPLAY_VRAM_LABEL_TT NULL
#define BEETLE_OPT_DISPLAY_VRAM_INFO_0_TT NULL
#define BEETLE_OPT_ANALOG_CALIBRATION_LABEL_TT NULL
#define BEETLE_OPT_ANALOG_CALIBRATION_INFO_0_TT NULL
#define BEETLE_OPT_ANALOG_TOGGLE_LABEL_TT NULL
#define BEETLE_OPT_ANALOG_TOGGLE_INFO_0_TT NULL
#define OPTION_VAL_ENABLED_ANALOG_TT NULL
#define BEETLE_OPT_ANALOG_TOGGLE_COMBO_LABEL_TT NULL
#define BEETLE_OPT_ANALOG_TOGGLE_COMBO_INFO_0_TT NULL
#define OPTION_VAL_L1_L2_R1_R2_START_SELECT_TT NULL
#define OPTION_VAL_L1_R1_SELECT_TT NULL
#define OPTION_VAL_L1_R1_START_TT NULL
#define OPTION_VAL_L1_R1_L3_TT NULL
#define OPTION_VAL_L1_R1_R3_TT NULL
#define OPTION_VAL_L2_R2_SELECT_TT NULL
#define OPTION_VAL_L2_R2_START_TT NULL
#define OPTION_VAL_L2_R2_L3_TT NULL
#define OPTION_VAL_L2_R2_R3_TT NULL
#define OPTION_VAL_L3_R3_TT NULL
#define BEETLE_OPT_ANALOG_TOGGLE_HOLD_LABEL_TT NULL
#define BEETLE_OPT_ANALOG_TOGGLE_HOLD_INFO_0_TT NULL
#define OPTION_VAL_0_TT NULL
#define OPTION_VAL_1_TT NULL
#define OPTION_VAL_2_TT NULL
#define OPTION_VAL_3_TT NULL
#define OPTION_VAL_4_TT NULL
#define OPTION_VAL_5_TT NULL
#define BEETLE_OPT_ENABLE_MULTITAP_PORT1_LABEL_TT NULL
#define BEETLE_OPT_ENABLE_MULTITAP_PORT1_INFO_0_TT NULL
#define BEETLE_OPT_ENABLE_MULTITAP_PORT2_LABEL_TT NULL
#define BEETLE_OPT_ENABLE_MULTITAP_PORT2_INFO_0_TT NULL
#define BEETLE_OPT_GUN_INPUT_MODE_LABEL_TT NULL
#define BEETLE_OPT_GUN_INPUT_MODE_INFO_0_TT NULL
#define OPTION_VAL_LIGHTGUN_TT NULL
#define OPTION_VAL_TOUCHSCREEN_TT NULL
#define BEETLE_OPT_GUN_CURSOR_LABEL_TT NULL
#define BEETLE_OPT_GUN_CURSOR_INFO_0_TT NULL
#define OPTION_VAL_CROSS_TT NULL
#define OPTION_VAL_DOT_TT "Нокта"
#define BEETLE_OPT_CROSSHAIR_COLOR_P1_LABEL_TT NULL
#define BEETLE_OPT_CROSSHAIR_COLOR_P1_INFO_0_TT NULL
#define OPTION_VAL_RED_TT "Кызыл"
#define OPTION_VAL_BLUE_TT NULL
#define OPTION_VAL_GREEN_TT NULL
#define OPTION_VAL_ORANGE_TT NULL
#define OPTION_VAL_YELLOW_TT NULL
#define OPTION_VAL_CYAN_TT NULL
#define OPTION_VAL_PINK_TT "Алсу"
#define OPTION_VAL_PURPLE_TT NULL
#define OPTION_VAL_BLACK_TT "Кара"
#define OPTION_VAL_WHITE_TT "Ак"
#define BEETLE_OPT_CROSSHAIR_COLOR_P2_LABEL_TT NULL
#define BEETLE_OPT_CROSSHAIR_COLOR_P2_INFO_0_TT NULL
#define BEETLE_OPT_MOUSE_SENSITIVITY_LABEL_TT NULL
#define BEETLE_OPT_MOUSE_SENSITIVITY_INFO_0_TT NULL
#define OPTION_VAL_5_O42_TT NULL
#define OPTION_VAL_10_TT NULL
#define OPTION_VAL_15_TT NULL
#define OPTION_VAL_20_TT NULL
#define OPTION_VAL_25_TT NULL
#define OPTION_VAL_30_TT NULL
#define OPTION_VAL_35_TT NULL
#define OPTION_VAL_40_TT NULL
#define OPTION_VAL_45_TT NULL
#define OPTION_VAL_55_TT NULL
#define OPTION_VAL_65_TT NULL
#define OPTION_VAL_75_TT NULL
#define OPTION_VAL_85_TT NULL
#define OPTION_VAL_95_TT NULL
#define OPTION_VAL_100_O42_TT NULL
#define OPTION_VAL_105_TT NULL
#define OPTION_VAL_115_TT NULL
#define OPTION_VAL_125_TT NULL
#define OPTION_VAL_135_TT NULL
#define OPTION_VAL_145_TT NULL
#define OPTION_VAL_155_TT NULL
#define OPTION_VAL_165_TT NULL
#define OPTION_VAL_175_TT NULL
#define OPTION_VAL_185_TT NULL
#define OPTION_VAL_195_TT NULL
#define BEETLE_OPT_NEGCON_RESPONSE_LABEL_TT NULL
#define BEETLE_OPT_NEGCON_RESPONSE_INFO_0_TT NULL
#define OPTION_VAL_LINEAR_TT NULL
#define OPTION_VAL_QUADRATIC_TT NULL
#define OPTION_VAL_CUBIC_TT NULL
#define BEETLE_OPT_NEGCON_DEADZONE_LABEL_TT NULL
#define BEETLE_OPT_NEGCON_DEADZONE_INFO_0_TT NULL
#define OPTION_VAL_0_O44_TT NULL
#define BEETLE_OPT_USE_MEDNAFEN_MEMCARD0_METHOD_LABEL_TT NULL
#define BEETLE_OPT_USE_MEDNAFEN_MEMCARD0_METHOD_INFO_0_TT NULL
#define OPTION_VAL_LIBRETRO_TT NULL
#define OPTION_VAL_MEDNAFEN_TT NULL
#define BEETLE_OPT_ENABLE_MEMCARD1_LABEL_TT NULL
#define BEETLE_OPT_ENABLE_MEMCARD1_INFO_0_TT NULL
#define BEETLE_OPT_SHARED_MEMORY_CARDS_LABEL_TT NULL
#define BEETLE_OPT_SHARED_MEMORY_CARDS_INFO_0_TT NULL
#define BEETLE_OPT_MEMCARD_LEFT_INDEX_LABEL_TT NULL
#define BEETLE_OPT_MEMCARD_LEFT_INDEX_INFO_0_TT NULL
#define OPTION_VAL_0_O48_TT NULL
#define BEETLE_OPT_MEMCARD_RIGHT_INDEX_LABEL_TT NULL
#define BEETLE_OPT_MEMCARD_RIGHT_INDEX_INFO_0_TT NULL
#define OPTION_VAL_1_O49_TT NULL
#define BEETLE_OPT_LINE_RENDER_LABEL_TT NULL
#define BEETLE_OPT_LINE_RENDER_INFO_0_TT NULL
#define OPTION_VAL_DEFAULT_TT NULL
#define OPTION_VAL_AGGRESSIVE_TT NULL
#define BEETLE_OPT_WIDESCREEN_HACK_LABEL_TT NULL
#define BEETLE_OPT_WIDESCREEN_HACK_INFO_0_TT NULL
#define BEETLE_OPT_WIDESCREEN_HACK_ASPECT_RATIO_LABEL_TT NULL
#define BEETLE_OPT_WIDESCREEN_HACK_ASPECT_RATIO_INFO_0_TT NULL
#define OPTION_VAL_16_9_TT NULL
#define OPTION_VAL_16_10_TT NULL
#define OPTION_VAL_18_9_TT NULL
#define OPTION_VAL_19_9_TT NULL
#define OPTION_VAL_20_9_TT NULL
#define OPTION_VAL_21_9_TT NULL
#define OPTION_VAL_32_9_TT NULL
#define BEETLE_OPT_CPU_DYNAREC_LABEL_TT NULL
#define BEETLE_OPT_CPU_DYNAREC_INFO_0_TT NULL
#define OPTION_VAL_DISABLED_TT NULL
#define OPTION_VAL_EXECUTE_TT NULL
#define OPTION_VAL_EXECUTE_ONE_TT NULL
#define OPTION_VAL_RUN_INTERPRETER_TT NULL
#define BEETLE_OPT_DYNAREC_INVALIDATE_LABEL_TT NULL
#define BEETLE_OPT_DYNAREC_INVALIDATE_INFO_0_TT NULL
#define OPTION_VAL_FULL_TT NULL
#define OPTION_VAL_DMA_TT NULL
#define BEETLE_OPT_DYNAREC_EVENTCYCLES_LABEL_TT NULL
#define BEETLE_OPT_DYNAREC_EVENTCYCLES_INFO_0_TT NULL
#define OPTION_VAL_128_TT NULL
#define BEETLE_OPT_DYNAREC_SPU_SAMPLES_LABEL_TT NULL
#define BEETLE_OPT_DYNAREC_SPU_SAMPLES_INFO_0_TT NULL
#define BEETLE_OPT_CORE_TIMING_FPS_LABEL_TT NULL
#define BEETLE_OPT_CORE_TIMING_FPS_INFO_0_TT NULL
#define OPTION_VAL_AUTO_TOGGLE_TT NULL
#define OPTION_VAL_FORCE_PROGRESSIVE_TT NULL
#define OPTION_VAL_FORCE_INTERLACED_TT NULL
#define BEETLE_OPT_DEINTERLACER_LABEL_TT NULL
#define BEETLE_OPT_DEINTERLACER_INFO_0_TT NULL
#define OPTION_VAL_WEAVE_TT NULL
#define OPTION_VAL_BOB_TT NULL
#define OPTION_VAL_BOB_OFFSET_TT NULL
#define OPTION_VAL_FASTMAD_TT NULL
#define BEETLE_OPT_FRAME_DUPING_LABEL_TT NULL
#define BEETLE_OPT_FRAME_DUPING_INFO_0_TT NULL
#define BEETLE_OPT_PAL_VIDEO_TIMING_OVERRIDE_LABEL_TT NULL
#define BEETLE_OPT_PAL_VIDEO_TIMING_OVERRIDE_INFO_0_TT NULL
#define BEETLE_OPT_ASPECT_RATIO_LABEL_TT NULL
#define BEETLE_OPT_ASPECT_RATIO_INFO_0_TT NULL
#define OPTION_VAL_CORRECTED_TT NULL
#define OPTION_VAL_UNCORRECTED_TT NULL
#define OPTION_VAL_4_3_TT NULL
#define OPTION_VAL_NTSC_TT NULL
#define BEETLE_OPT_CROP_OVERSCAN_LABEL_TT NULL
#define BEETLE_OPT_CROP_OVERSCAN_INFO_0_TT NULL
#define OPTION_VAL_STATIC_TT NULL
#define OPTION_VAL_SMART_TT NULL
#define BEETLE_OPT_IMAGE_CROP_LABEL_TT NULL
#define BEETLE_OPT_IMAGE_CROP_INFO_0_TT NULL
#define OPTION_VAL_9PX_TT NULL
#define OPTION_VAL_10PX_TT NULL
#define OPTION_VAL_11PX_TT NULL
#define OPTION_VAL_12PX_TT NULL
#define OPTION_VAL_13PX_TT NULL
#define OPTION_VAL_14PX_TT NULL
#define OPTION_VAL_15PX_TT NULL
#define OPTION_VAL_16PX_TT NULL
#define OPTION_VAL_17PX_TT NULL
#define OPTION_VAL_18PX_TT NULL
#define OPTION_VAL_19PX_TT NULL
#define OPTION_VAL_20PX_TT NULL
#define BEETLE_OPT_IMAGE_OFFSET_LABEL_TT NULL
#define BEETLE_OPT_IMAGE_OFFSET_INFO_0_TT NULL
#define OPTION_VAL_12PX_O64_TT NULL
#define OPTION_VAL_11PX_O64_TT NULL
#define OPTION_VAL_10PX_O64_TT NULL
#define OPTION_VAL_9PX_O64_TT NULL
#define OPTION_VAL_8PX_O64_TT NULL
#define OPTION_VAL_7PX_O64_TT NULL
#define OPTION_VAL_6PX_O64_TT NULL
#define OPTION_VAL_5PX_O64_TT NULL
#define OPTION_VAL_4PX_O64_TT NULL
#define OPTION_VAL_3PX_O64_TT NULL
#define OPTION_VAL_2PX_O64_TT NULL
#define OPTION_VAL_1PX_O64_TT NULL
#define OPTION_VAL_1PX_O64_0_TT NULL
#define OPTION_VAL_2PX_O64_0_TT NULL
#define OPTION_VAL_3PX_O64_0_TT NULL
#define OPTION_VAL_4PX_O64_0_TT NULL
#define OPTION_VAL_5PX_O64_0_TT NULL
#define OPTION_VAL_6PX_O64_0_TT NULL
#define OPTION_VAL_7PX_O64_0_TT NULL
#define OPTION_VAL_8PX_O64_0_TT NULL
#define OPTION_VAL_9PX_O64_0_TT NULL
#define OPTION_VAL_10PX_O64_0_TT NULL
#define OPTION_VAL_11PX_O64_0_TT NULL
#define OPTION_VAL_12PX_O64_0_TT NULL
#define BEETLE_OPT_IMAGE_OFFSET_CYCLES_LABEL_TT NULL
#define BEETLE_OPT_IMAGE_OFFSET_CYCLES_INFO_0_TT NULL
#define BEETLE_OPT_INITIAL_SCANLINE_LABEL_TT NULL
#define BEETLE_OPT_INITIAL_SCANLINE_INFO_0_TT NULL
#define BEETLE_OPT_LAST_SCANLINE_LABEL_TT NULL
#define BEETLE_OPT_LAST_SCANLINE_INFO_0_TT NULL
#define OPTION_VAL_239_TT NULL
#define BEETLE_OPT_INITIAL_SCANLINE_PAL_LABEL_TT NULL
#define BEETLE_OPT_INITIAL_SCANLINE_PAL_INFO_0_TT NULL
#define BEETLE_OPT_LAST_SCANLINE_PAL_LABEL_TT NULL
#define BEETLE_OPT_LAST_SCANLINE_PAL_INFO_0_TT NULL
#define OPTION_VAL_287_TT NULL

struct retro_core_option_v2_category option_cats_tt[] = {
   {
      "system",
      CATEGORY_SYSTEM_LABEL_TT,
      CATEGORY_SYSTEM_INFO_0_TT
   },
   {
      "video",
      CATEGORY_VIDEO_LABEL_TT,
      CATEGORY_VIDEO_INFO_0_TT
   },
   {
      "pgxp",
      CATEGORY_PGXP_LABEL_TT,
      CATEGORY_PGXP_INFO_0_TT
   },
   {
      "osd",
      CATEGORY_OSD_LABEL_TT,
      CATEGORY_OSD_INFO_0_TT
   },
   {
      "input",
      CATEGORY_INPUT_LABEL_TT,
      CATEGORY_INPUT_INFO_0_TT
   },
   {
      "memcards",
      CATEGORY_MEMCARDS_LABEL_TT,
      CATEGORY_MEMCARDS_INFO_0_TT
   },
   {
      "hacks",
      CATEGORY_HACKS_LABEL_TT,
      CATEGORY_HACKS_INFO_0_TT
   },
   { NULL, NULL, NULL },
};
struct retro_core_option_v2_definition option_defs_tt[] = {
   {
      BEETLE_OPT(cpu_freq_scale),
      BEETLE_OPT_CPU_FREQ_SCALE_LABEL_TT,
      NULL,
      BEETLE_OPT_CPU_FREQ_SCALE_INFO_0_TT,
      NULL,
      "system",
      {
         { "50%",           OPTION_VAL_50_TT },
         { "60%",           OPTION_VAL_60_TT },
         { "70%",           OPTION_VAL_70_TT },
         { "80%",           OPTION_VAL_80_TT },
         { "90%",           OPTION_VAL_90_TT },
         { "100%",          OPTION_VAL_100_TT },
         { "110%",          OPTION_VAL_110_TT },
         { "120%",          OPTION_VAL_120_TT },
         { "130%",          OPTION_VAL_130_TT },
         { "140%",          OPTION_VAL_140_TT },
         { "150%",          OPTION_VAL_150_TT },
         { "160%",          OPTION_VAL_160_TT },
         { "170%",          OPTION_VAL_170_TT },
         { "180%",          OPTION_VAL_180_TT },
         { "190%",          OPTION_VAL_190_TT },
         { "200%",          OPTION_VAL_200_TT },
         { "210%",          OPTION_VAL_210_TT },
         { "220%",          OPTION_VAL_220_TT },
         { "230%",          OPTION_VAL_230_TT },
         { "240%",          OPTION_VAL_240_TT },
         { "250%",          OPTION_VAL_250_TT },
         { "260%",          OPTION_VAL_260_TT },
         { "270%",          OPTION_VAL_270_TT },
         { "280%",          OPTION_VAL_280_TT },
         { "290%",          OPTION_VAL_290_TT },
         { "300%",          OPTION_VAL_300_TT },
         { "310%",          OPTION_VAL_310_TT },
         { "320%",          OPTION_VAL_320_TT },
         { "330%",          OPTION_VAL_330_TT },
         { "340%",          OPTION_VAL_340_TT },
         { "350%",          OPTION_VAL_350_TT },
         { "360%",          OPTION_VAL_360_TT },
         { "370%",          OPTION_VAL_370_TT },
         { "380%",          OPTION_VAL_380_TT },
         { "390%",          OPTION_VAL_390_TT },
         { "400%",          OPTION_VAL_400_TT },
         { "410%",          OPTION_VAL_410_TT },
         { "420%",          OPTION_VAL_420_TT },
         { "430%",          OPTION_VAL_430_TT },
         { "440%",          OPTION_VAL_440_TT },
         { "450%",          OPTION_VAL_450_TT },
         { "460%",          OPTION_VAL_460_TT },
         { "470%",          OPTION_VAL_470_TT },
         { "480%",          OPTION_VAL_480_TT },
         { "490%",          OPTION_VAL_490_TT },
         { "500%",          OPTION_VAL_500_TT },
         { "510%",          OPTION_VAL_510_TT },
         { "520%",          OPTION_VAL_520_TT },
         { "530%",          OPTION_VAL_530_TT },
         { "540%",          OPTION_VAL_540_TT },
         { "550%",          OPTION_VAL_550_TT },
         { "560%",          OPTION_VAL_560_TT },
         { "570%",          OPTION_VAL_570_TT },
         { "580%",          OPTION_VAL_580_TT },
         { "590%",          OPTION_VAL_590_TT },
         { "600%",          OPTION_VAL_600_TT },
         { "610%",          OPTION_VAL_610_TT },
         { "620%",          OPTION_VAL_620_TT },
         { "630%",          OPTION_VAL_630_TT },
         { "640%",          OPTION_VAL_640_TT },
         { "650%",          OPTION_VAL_650_TT },
         { "660%",          OPTION_VAL_660_TT },
         { "670%",          OPTION_VAL_670_TT },
         { "680%",          OPTION_VAL_680_TT },
         { "690%",          OPTION_VAL_690_TT },
         { "700%",          OPTION_VAL_700_TT },
         { "710%",          OPTION_VAL_710_TT },
         { "720%",          OPTION_VAL_720_TT },
         { "730%",          OPTION_VAL_730_TT },
         { "740%",          OPTION_VAL_740_TT },
         { "750%",          OPTION_VAL_750_TT },
         { NULL, NULL },
      },
      "100%"
   },
   {
      BEETLE_OPT(gpu_overclock),
      BEETLE_OPT_GPU_OVERCLOCK_LABEL_TT,
      NULL,
      BEETLE_OPT_GPU_OVERCLOCK_INFO_0_TT,
      NULL,
      "system",
      {
         { "1x(native)", OPTION_VAL_1X_NATIVE_TT },
         { "2x",         OPTION_VAL_2X_TT },
         { "4x",         OPTION_VAL_4X_TT },
         { "8x",         OPTION_VAL_8X_TT },
         { "16x",        OPTION_VAL_16X_TT },
         { "32x",        OPTION_VAL_32X_TT },
         { NULL, NULL },
      },
      "1x(native)"
   },
   {
      BEETLE_OPT(gte_overclock),
      BEETLE_OPT_GTE_OVERCLOCK_LABEL_TT,
      NULL,
      BEETLE_OPT_GTE_OVERCLOCK_INFO_0_TT,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
#ifndef EMSCRIPTEN
   {
      BEETLE_OPT(cd_access_method),
      BEETLE_OPT_CD_ACCESS_METHOD_LABEL_TT,
      NULL,
      BEETLE_OPT_CD_ACCESS_METHOD_INFO_0_TT,
      NULL,
      "system",
      {
         { "sync",     OPTION_VAL_SYNC_TT },
         { "async",    OPTION_VAL_ASYNC_TT },
         { "precache", OPTION_VAL_PRECACHE_TT },
         { NULL, NULL },
      },
      "sync"
   },
#endif
   {
      BEETLE_OPT(cd_fastload),
      BEETLE_OPT_CD_FASTLOAD_LABEL_TT,
      NULL,
      BEETLE_OPT_CD_FASTLOAD_INFO_0_TT,
      NULL,
      "system",
      {
         { "2x(native)", OPTION_VAL_2X_NATIVE_TT },
         { "4x",          OPTION_VAL_4X_TT },
         { "6x",          OPTION_VAL_6X_TT },
         { "8x",          OPTION_VAL_8X_TT },
         { "10x",         OPTION_VAL_10X_TT },
         { "12x",         OPTION_VAL_12X_TT },
         { "14x",         OPTION_VAL_14X_TT },
         { NULL, NULL },
      },
      "2x(native)"
   },
   {
      BEETLE_OPT(skip_bios),
      BEETLE_OPT_SKIP_BIOS_LABEL_TT,
      NULL,
      BEETLE_OPT_SKIP_BIOS_INFO_0_TT,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(override_bios),
      BEETLE_OPT_OVERRIDE_BIOS_LABEL_TT,
      NULL,
      BEETLE_OPT_OVERRIDE_BIOS_INFO_0_TT,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "psxonpsp", OPTION_VAL_PSXONPSP_TT },
         { "ps1_rom",  OPTION_VAL_PS1_ROM_TT },
         { "openbios", OPTION_VAL_OPENBIOS_TT     },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(region),
      BEETLE_OPT_REGION_LABEL_TT,
      NULL,
      BEETLE_OPT_REGION_INFO_0_TT,
      NULL,
      "system",
      {
         { "auto",   OPTION_VAL_AUTO_TT },
         { "ntsc-j", OPTION_VAL_NTSC_J_TT },
         { "ntsc-u", OPTION_VAL_NTSC_U_TT },
         { "pal",    OPTION_VAL_PAL_TT },
         { NULL, NULL },
      },
      "auto"
   },
   {
      BEETLE_OPT(internal_resolution),
      BEETLE_OPT_INTERNAL_RESOLUTION_LABEL_TT,
      NULL,
      BEETLE_OPT_INTERNAL_RESOLUTION_INFO_0_TT,
      NULL,
      "video",
      {
         { "1x(native)", OPTION_VAL_1X_NATIVE_TT },
         { "2x",         OPTION_VAL_2X_TT },
         { "4x",         OPTION_VAL_4X_TT },
         { "8x",         OPTION_VAL_8X_TT },
         { "16x",        OPTION_VAL_16X_TT },
         { NULL, NULL },
      },
      "1x(native)"
   },
#if defined(HAVE_OPENGL) || defined(HAVE_OPENGLES) || defined(HAVE_VULKAN)
   {
      BEETLE_OPT(renderer),
      BEETLE_OPT_RENDERER_LABEL_TT,
      NULL,
      BEETLE_OPT_RENDERER_INFO_0_TT,
      NULL,
      "video",
      {
         { "hardware",    OPTION_VAL_HARDWARE_TT },
#if defined(HAVE_OPENGL) || defined(HAVE_OPENGLES)
         { "hardware_gl", OPTION_VAL_HARDWARE_GL_TT },
#endif
#if defined(HAVE_VULKAN)
         { "hardware_vk", OPTION_VAL_HARDWARE_VK_TT },
#endif
         { "software",    OPTION_VAL_SOFTWARE_TT },
         { NULL, NULL },
      },
      "hardware"
   },
   {
      BEETLE_OPT(renderer_software_fb),
      BEETLE_OPT_RENDERER_SOFTWARE_FB_LABEL_TT,
      NULL,
      BEETLE_OPT_RENDERER_SOFTWARE_FB_INFO_0_TT,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled"
   },
#endif
#if defined(HAVE_OPENGL) || defined(HAVE_OPENGLES)
   {
      BEETLE_OPT(depth),
      BEETLE_OPT_DEPTH_LABEL_TT,
      NULL,
      BEETLE_OPT_DEPTH_INFO_0_TT,
      NULL,
      "video",
      {
         { "16bpp(native)", OPTION_VAL_16BPP_NATIVE_TT },
         { "32bpp",         OPTION_VAL_32BPP_TT },
         { NULL, NULL },
      },
      "16bpp(native)"
   },
   // Sort of, it's more like 15-bit high color and 24-bit true color for visible output. The alpha channel is used for mask bit. Vulkan renderer uses ABGR1555_555 for 31 bits internal? FMVs are always 24-bit on all renderers like original hardware (BGR888, no alpha)
#endif
   {
      BEETLE_OPT(dither_mode),
      BEETLE_OPT_DITHER_MODE_LABEL_TT,
      NULL,
      BEETLE_OPT_DITHER_MODE_INFO_0_TT,
      NULL,
      "video",
      {
         { "1x(native)",          OPTION_VAL_1X_NATIVE_TT },
         { "internal resolution", OPTION_VAL_INTERNAL_RESOLUTION_TT },
         { "disabled",            NULL },
         { NULL, NULL },
      },
      "1x(native)"
   },
#ifdef HAVE_VULKAN
   {
      BEETLE_OPT(scaled_uv_offset),
      BEETLE_OPT_SCALED_UV_OFFSET_LABEL_TT,
      NULL,
      BEETLE_OPT_SCALED_UV_OFFSET_INFO_0_TT,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled"
   },
#endif
#if defined(HAVE_OPENGL) || defined(HAVE_OPENGLES) || defined(HAVE_VULKAN)
   {
      BEETLE_OPT(filter),
      BEETLE_OPT_FILTER_LABEL_TT,
      NULL,
      BEETLE_OPT_FILTER_INFO_0_TT,
      NULL,
      "video",
      {
         { "nearest",  OPTION_VAL_NEAREST_TT },
         { "SABR",     OPTION_VAL_SABR_TT },
         { "xBR",      OPTION_VAL_XBR_TT },
         { "bilinear", OPTION_VAL_BILINEAR_TT },
         { "3-point",  OPTION_VAL_3_POINT_TT },
         { "JINC2",    OPTION_VAL_JINC2_TT },
         { NULL, NULL },
      },
      "nearest"
   },
#ifdef HAVE_VULKAN
   {
      BEETLE_OPT(filter_exclude_sprite),
      BEETLE_OPT_FILTER_EXCLUDE_SPRITE_LABEL_TT,
      NULL,
      BEETLE_OPT_FILTER_EXCLUDE_SPRITE_INFO_0_TT,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "opaque", OPTION_VAL_OPAQUE_TT },
         { "all", OPTION_VAL_ALL_TT },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(filter_exclude_2d_polygon),
      BEETLE_OPT_FILTER_EXCLUDE_2D_POLYGON_LABEL_TT,
      NULL,
      BEETLE_OPT_FILTER_EXCLUDE_2D_POLYGON_INFO_0_TT,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "opaque", OPTION_VAL_OPAQUE_TT },
         { "all", OPTION_VAL_ALL_TT },
         { NULL, NULL },
      },
      "disabled"
   },
#endif
#endif
#ifdef HAVE_VULKAN
   {
      BEETLE_OPT(adaptive_smoothing),
      BEETLE_OPT_ADAPTIVE_SMOOTHING_LABEL_TT,
      NULL,
      BEETLE_OPT_ADAPTIVE_SMOOTHING_INFO_0_TT,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(super_sampling),
      BEETLE_OPT_SUPER_SAMPLING_LABEL_TT,
      NULL,
      BEETLE_OPT_SUPER_SAMPLING_INFO_0_TT,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(msaa),
      BEETLE_OPT_MSAA_LABEL_TT,
      NULL,
      BEETLE_OPT_MSAA_INFO_0_TT,
      NULL,
      "video",
      {
         { "1x",  OPTION_VAL_1X_TT },
         { "2x",  OPTION_VAL_2X_TT },
         { "4x",  OPTION_VAL_4X_TT },
         { "8x",  OPTION_VAL_8X_TT },
         { "16x", OPTION_VAL_16X_TT },
         { NULL, NULL },
      },
      "1x"
   },
   {
      BEETLE_OPT(mdec_yuv),
      BEETLE_OPT_MDEC_YUV_LABEL_TT,
      NULL,
      BEETLE_OPT_MDEC_YUV_INFO_0_TT,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(track_textures),
      BEETLE_OPT_TRACK_TEXTURES_LABEL_TT,
      NULL,
      BEETLE_OPT_TRACK_TEXTURES_INFO_0_TT,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
#ifdef TEXTURE_DUMPING_ENABLED
   {
      BEETLE_OPT(dump_textures),
      BEETLE_OPT_DUMP_TEXTURES_LABEL_TT,
      NULL,
      BEETLE_OPT_DUMP_TEXTURES_INFO_0_TT,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
#endif
   {
      BEETLE_OPT(replace_textures),
      BEETLE_OPT_REPLACE_TEXTURES_LABEL_TT,
      NULL,
      BEETLE_OPT_REPLACE_TEXTURES_INFO_0_TT,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
#endif
   {
      BEETLE_OPT(pgxp_mode),
      BEETLE_OPT_PGXP_MODE_LABEL_TT,
      NULL,
      BEETLE_OPT_PGXP_MODE_INFO_0_TT,
      NULL,
      "pgxp",
      {
         { "disabled",     NULL },
         { "memory only",  OPTION_VAL_MEMORY_ONLY_TT },
         { "memory + CPU", OPTION_VAL_MEMORY_CPU_TT },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(pgxp_2d_tol),
      BEETLE_OPT_PGXP_2D_TOL_LABEL_TT,
      NULL,
      BEETLE_OPT_PGXP_2D_TOL_INFO_0_TT,
      NULL,
      "pgxp",
      {
         { "disabled", NULL },
         { "0px", OPTION_VAL_0PX_TT },
         { "1px", OPTION_VAL_1PX_TT },
         { "2px", OPTION_VAL_2PX_TT },
         { "3px", OPTION_VAL_3PX_TT },
         { "4px", OPTION_VAL_4PX_TT },
         { "5px", OPTION_VAL_5PX_TT },
         { "6px", OPTION_VAL_6PX_TT },
         { "7px", OPTION_VAL_7PX_TT },
         { "8px", OPTION_VAL_8PX_TT },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(pgxp_nclip),
      BEETLE_OPT_PGXP_NCLIP_LABEL_TT,
      NULL,
      BEETLE_OPT_PGXP_NCLIP_INFO_0_TT,
      NULL,
      "pgxp",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
#if defined(HAVE_OPENGL) || defined(HAVE_OPENGLES) || defined(HAVE_VULKAN)
   {
      BEETLE_OPT(pgxp_vertex),
      BEETLE_OPT_PGXP_VERTEX_LABEL_TT,
      NULL,
      BEETLE_OPT_PGXP_VERTEX_INFO_0_TT,
      NULL,
      "pgxp",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
#endif
   {
      BEETLE_OPT(pgxp_texture),
      BEETLE_OPT_PGXP_TEXTURE_LABEL_TT,
      NULL,
      BEETLE_OPT_PGXP_TEXTURE_INFO_0_TT,
      NULL,
      "pgxp",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(display_internal_fps),
      BEETLE_OPT_DISPLAY_INTERNAL_FPS_LABEL_TT,
      NULL,
      BEETLE_OPT_DISPLAY_INTERNAL_FPS_INFO_0_TT,
      NULL,
      "osd",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(display_osd),
      BEETLE_OPT_DISPLAY_OSD_LABEL_TT,
      NULL,
      BEETLE_OPT_DISPLAY_OSD_INFO_0_TT,
      NULL,
      "osd",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled"
   },
#if defined(HAVE_OPENGL) || defined(HAVE_OPENGLES) || defined(HAVE_VULKAN)
   {
      BEETLE_OPT(display_vram),
      BEETLE_OPT_DISPLAY_VRAM_LABEL_TT,
      NULL,
      BEETLE_OPT_DISPLAY_VRAM_INFO_0_TT,
      NULL,
      "osd",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
#endif
   {
      BEETLE_OPT(analog_calibration),
      BEETLE_OPT_ANALOG_CALIBRATION_LABEL_TT,
      NULL,
      BEETLE_OPT_ANALOG_CALIBRATION_INFO_0_TT,
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(analog_toggle),
      BEETLE_OPT_ANALOG_TOGGLE_LABEL_TT,
      NULL,
      BEETLE_OPT_ANALOG_TOGGLE_INFO_0_TT,
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { "enabled-analog", OPTION_VAL_ENABLED_ANALOG_TT },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(analog_toggle_combo),
      BEETLE_OPT_ANALOG_TOGGLE_COMBO_LABEL_TT,
      NULL,
      BEETLE_OPT_ANALOG_TOGGLE_COMBO_INFO_0_TT,
      NULL,
      "input",
      {
         { "l1+l2+r1+r2+start+select", OPTION_VAL_L1_L2_R1_R2_START_SELECT_TT },
         { "l1+r1+select",             OPTION_VAL_L1_R1_SELECT_TT },
         { "l1+r1+start",              OPTION_VAL_L1_R1_START_TT },
         { "l1+r1+l3",                 OPTION_VAL_L1_R1_L3_TT },
         { "l1+r1+r3",                 OPTION_VAL_L1_R1_R3_TT },
         { "l2+r2+select",             OPTION_VAL_L2_R2_SELECT_TT },
         { "l2+r2+start",              OPTION_VAL_L2_R2_START_TT },
         { "l2+r2+l3",                 OPTION_VAL_L2_R2_L3_TT },
         { "l2+r2+r3",                 OPTION_VAL_L2_R2_R3_TT },
         { "l3+r3",                    OPTION_VAL_L3_R3_TT },
         { NULL, NULL },
      },
      "l1+r1+select"
   },
   {
      BEETLE_OPT(analog_toggle_hold),
      BEETLE_OPT_ANALOG_TOGGLE_HOLD_LABEL_TT,
      NULL,
      BEETLE_OPT_ANALOG_TOGGLE_HOLD_INFO_0_TT,
      NULL,
      "input",
      {
         { "0", OPTION_VAL_0_TT },
         { "1", OPTION_VAL_1_TT },
         { "2", OPTION_VAL_2_TT },
         { "3", OPTION_VAL_3_TT },
         { "4", OPTION_VAL_4_TT },
         { "5", OPTION_VAL_5_TT },
         { NULL, NULL },
      },
      "1"
   },
   {
      BEETLE_OPT(enable_multitap_port1),
      BEETLE_OPT_ENABLE_MULTITAP_PORT1_LABEL_TT,
      NULL,
      BEETLE_OPT_ENABLE_MULTITAP_PORT1_INFO_0_TT,
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(enable_multitap_port2),
      BEETLE_OPT_ENABLE_MULTITAP_PORT2_LABEL_TT,
      NULL,
      BEETLE_OPT_ENABLE_MULTITAP_PORT2_INFO_0_TT,
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(gun_input_mode),
      BEETLE_OPT_GUN_INPUT_MODE_LABEL_TT,
      NULL,
      BEETLE_OPT_GUN_INPUT_MODE_INFO_0_TT,
      NULL,
      "input",
      {
         { "lightgun",    OPTION_VAL_LIGHTGUN_TT },
         { "touchscreen", OPTION_VAL_TOUCHSCREEN_TT },
         { NULL, NULL },
      },
      "lightgun"
   },
   // Shouldn't the gun_input_mode just be Mouse vs. Touchscreen?
   {
      BEETLE_OPT(gun_cursor),
      BEETLE_OPT_GUN_CURSOR_LABEL_TT,
      NULL,
      BEETLE_OPT_GUN_CURSOR_INFO_0_TT,
      NULL,
      "input",
      {
         { "off",   "disabled" },
         { "cross", OPTION_VAL_CROSS_TT },
         { "dot",   OPTION_VAL_DOT_TT },
         { NULL, NULL },
      },
      "cross"
   },
   {
      BEETLE_OPT(crosshair_color_p1),
      BEETLE_OPT_CROSSHAIR_COLOR_P1_LABEL_TT,
      NULL,
      BEETLE_OPT_CROSSHAIR_COLOR_P1_INFO_0_TT,
      NULL,
      "input",
      {
         { "red",    OPTION_VAL_RED_TT },
         { "blue",   OPTION_VAL_BLUE_TT },
         { "green",  OPTION_VAL_GREEN_TT },
         { "orange", OPTION_VAL_ORANGE_TT },
         { "yellow", OPTION_VAL_YELLOW_TT },
         { "cyan",   OPTION_VAL_CYAN_TT },
         { "pink",   OPTION_VAL_PINK_TT },
         { "purple", OPTION_VAL_PURPLE_TT },
         { "black",  OPTION_VAL_BLACK_TT },
         { "white",  OPTION_VAL_WHITE_TT },
         { NULL, NULL },
      },
      "red"
   },
   {
      BEETLE_OPT(crosshair_color_p2),
      BEETLE_OPT_CROSSHAIR_COLOR_P2_LABEL_TT,
      NULL,
      BEETLE_OPT_CROSSHAIR_COLOR_P2_INFO_0_TT,
      NULL,
      "input",
      {
         { "blue",   OPTION_VAL_BLUE_TT },
         { "red",    OPTION_VAL_RED_TT },
         { "green",  OPTION_VAL_GREEN_TT },
         { "orange", OPTION_VAL_ORANGE_TT },
         { "yellow", OPTION_VAL_YELLOW_TT },
         { "cyan",   OPTION_VAL_CYAN_TT },
         { "pink",   OPTION_VAL_PINK_TT },
         { "purple", OPTION_VAL_PURPLE_TT },
         { "black",  OPTION_VAL_BLACK_TT },
         { "white",  OPTION_VAL_WHITE_TT },
         { NULL, NULL },
      },
      "blue"
   },
   {
      BEETLE_OPT(mouse_sensitivity),
      BEETLE_OPT_MOUSE_SENSITIVITY_LABEL_TT,
      NULL,
      BEETLE_OPT_MOUSE_SENSITIVITY_INFO_0_TT,
      NULL,
      "input",
      {
         { "5%",   OPTION_VAL_5_O42_TT },
         { "10%",  OPTION_VAL_10_TT },
         { "15%",  OPTION_VAL_15_TT },
         { "20%",  OPTION_VAL_20_TT },
         { "25%",  OPTION_VAL_25_TT },
         { "30%",  OPTION_VAL_30_TT },
         { "35%",  OPTION_VAL_35_TT },
         { "40%",  OPTION_VAL_40_TT },
         { "45%",  OPTION_VAL_45_TT },
         { "50%",  OPTION_VAL_50_TT },
         { "55%",  OPTION_VAL_55_TT },
         { "60%",  OPTION_VAL_60_TT },
         { "65%",  OPTION_VAL_65_TT },
         { "70%",  OPTION_VAL_70_TT },
         { "75%",  OPTION_VAL_75_TT },
         { "80%",  OPTION_VAL_80_TT },
         { "85%",  OPTION_VAL_85_TT },
         { "90%",  OPTION_VAL_90_TT },
         { "95%",  OPTION_VAL_95_TT },
         { "100%", OPTION_VAL_100_O42_TT },
         { "105%", OPTION_VAL_105_TT },
         { "110%", OPTION_VAL_110_TT },
         { "115%", OPTION_VAL_115_TT },
         { "120%", OPTION_VAL_120_TT },
         { "125%", OPTION_VAL_125_TT },
         { "130%", OPTION_VAL_130_TT },
         { "135%", OPTION_VAL_135_TT },
         { "140%", OPTION_VAL_140_TT },
         { "145%", OPTION_VAL_145_TT },
         { "150%", OPTION_VAL_150_TT },
         { "155%", OPTION_VAL_155_TT },
         { "160%", OPTION_VAL_160_TT },
         { "165%", OPTION_VAL_165_TT },
         { "170%", OPTION_VAL_170_TT },
         { "175%", OPTION_VAL_175_TT },
         { "180%", OPTION_VAL_180_TT },
         { "185%", OPTION_VAL_185_TT },
         { "190%", OPTION_VAL_190_TT },
         { "195%", OPTION_VAL_195_TT },
         { "200%", OPTION_VAL_200_TT },
         { NULL, NULL },
      },
      "100%"
   },
   {
      BEETLE_OPT(negcon_response),
      BEETLE_OPT_NEGCON_RESPONSE_LABEL_TT,
      NULL,
      BEETLE_OPT_NEGCON_RESPONSE_INFO_0_TT,
      NULL,
      "input",
      {
         { "linear",    OPTION_VAL_LINEAR_TT },
         { "quadratic", OPTION_VAL_QUADRATIC_TT },
         { "cubic",     OPTION_VAL_CUBIC_TT },
         { NULL, NULL },
      },
      "linear"
   },
   {
      BEETLE_OPT(negcon_deadzone),
      BEETLE_OPT_NEGCON_DEADZONE_LABEL_TT,
      NULL,
      BEETLE_OPT_NEGCON_DEADZONE_INFO_0_TT,
      NULL,
      "input",
      {
         { "0%",  OPTION_VAL_0_O44_TT },
         { "5%",  OPTION_VAL_5_O42_TT },
         { "10%", OPTION_VAL_10_TT },
         { "15%", OPTION_VAL_15_TT },
         { "20%", OPTION_VAL_20_TT },
         { "25%", OPTION_VAL_25_TT },
         { "30%", OPTION_VAL_30_TT },
         { NULL, NULL },
      },
      "0%"
   },
   {
      BEETLE_OPT(use_mednafen_memcard0_method),
      BEETLE_OPT_USE_MEDNAFEN_MEMCARD0_METHOD_LABEL_TT,
      NULL,
      BEETLE_OPT_USE_MEDNAFEN_MEMCARD0_METHOD_INFO_0_TT,
      NULL,
      "memcards",
      {
         { "libretro", OPTION_VAL_LIBRETRO_TT },
         { "mednafen", OPTION_VAL_MEDNAFEN_TT },
         { NULL, NULL },
      },
      "libretro"
   },
   {
      BEETLE_OPT(enable_memcard1),
      BEETLE_OPT_ENABLE_MEMCARD1_LABEL_TT,
      NULL,
      BEETLE_OPT_ENABLE_MEMCARD1_INFO_0_TT,
      NULL,
      "memcards",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(shared_memory_cards),
      BEETLE_OPT_SHARED_MEMORY_CARDS_LABEL_TT,
      NULL,
      BEETLE_OPT_SHARED_MEMORY_CARDS_INFO_0_TT,
      NULL,
      "memcards",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(memcard_left_index),
      BEETLE_OPT_MEMCARD_LEFT_INDEX_LABEL_TT,
      NULL,
      BEETLE_OPT_MEMCARD_LEFT_INDEX_INFO_0_TT,
      NULL,
      "memcards",
      {
         { "0",  OPTION_VAL_0_O48_TT },
         { "1",  NULL },
         { "2",  NULL },
         { "3",  NULL },
         { "4",  NULL },
         { "5",  NULL },
         { "6",  NULL },
         { "7",  NULL },
         { "8",  NULL },
         { "9",  NULL },
         { "10", NULL },
         { "11", NULL },
         { "12", NULL },
         { "13", NULL },
         { "14", NULL },
         { "15", NULL },
         { "16", NULL },
         { "17", NULL },
         { "18", NULL },
         { "19", NULL },
         { "20", NULL },
         { "21", NULL },
         { "22", NULL },
         { "23", NULL },
         { "24", NULL },
         { "25", NULL },
         { "26", NULL },
         { "27", NULL },
         { "28", NULL },
         { "29", NULL },
         { "30", NULL },
         { "31", NULL },
         { "32", NULL },
         { "33", NULL },
         { "34", NULL },
         { "35", NULL },
         { "36", NULL },
         { "37", NULL },
         { "38", NULL },
         { "39", NULL },
         { "40", NULL },
         { "41", NULL },
         { "42", NULL },
         { "43", NULL },
         { "44", NULL },
         { "45", NULL },
         { "46", NULL },
         { "47", NULL },
         { "48", NULL },
         { "49", NULL },
         { "50", NULL },
         { "51", NULL },
         { "52", NULL },
         { "53", NULL },
         { "54", NULL },
         { "55", NULL },
         { "56", NULL },
         { "57", NULL },
         { "58", NULL },
         { "59", NULL },
         { "60", NULL },
         { "61", NULL },
         { "62", NULL },
         { "63", NULL },
         { NULL, NULL },
      },
      "0"
   },
   {
      BEETLE_OPT(memcard_right_index),
      BEETLE_OPT_MEMCARD_RIGHT_INDEX_LABEL_TT,
      NULL,
      BEETLE_OPT_MEMCARD_RIGHT_INDEX_INFO_0_TT,
      NULL,
      "memcards",
      {
         { "0",  NULL },
         { "1",  OPTION_VAL_1_O49_TT },
         { "2",  NULL },
         { "3",  NULL },
         { "4",  NULL },
         { "5",  NULL },
         { "6",  NULL },
         { "7",  NULL },
         { "8",  NULL },
         { "9",  NULL },
         { "10", NULL },
         { "11", NULL },
         { "12", NULL },
         { "13", NULL },
         { "14", NULL },
         { "15", NULL },
         { "16", NULL },
         { "17", NULL },
         { "18", NULL },
         { "19", NULL },
         { "20", NULL },
         { "21", NULL },
         { "22", NULL },
         { "23", NULL },
         { "24", NULL },
         { "25", NULL },
         { "26", NULL },
         { "27", NULL },
         { "28", NULL },
         { "29", NULL },
         { "30", NULL },
         { "31", NULL },
         { "32", NULL },
         { "33", NULL },
         { "34", NULL },
         { "35", NULL },
         { "36", NULL },
         { "37", NULL },
         { "38", NULL },
         { "39", NULL },
         { "40", NULL },
         { "41", NULL },
         { "42", NULL },
         { "43", NULL },
         { "44", NULL },
         { "45", NULL },
         { "46", NULL },
         { "47", NULL },
         { "48", NULL },
         { "49", NULL },
         { "50", NULL },
         { "51", NULL },
         { "52", NULL },
         { "53", NULL },
         { "54", NULL },
         { "55", NULL },
         { "56", NULL },
         { "57", NULL },
         { "58", NULL },
         { "59", NULL },
         { "60", NULL },
         { "61", NULL },
         { "62", NULL },
         { "63", NULL },
         { NULL, NULL },
      },
      "1"
   },
   {
      BEETLE_OPT(line_render),
      BEETLE_OPT_LINE_RENDER_LABEL_TT,
      NULL,
      BEETLE_OPT_LINE_RENDER_INFO_0_TT,
      NULL,
      "hacks",
      {
         { "default",    OPTION_VAL_DEFAULT_TT },
         { "aggressive", OPTION_VAL_AGGRESSIVE_TT },
         { "disabled",   NULL },
         { NULL, NULL },
      },
      "default"
   },
   {
      BEETLE_OPT(widescreen_hack),
      BEETLE_OPT_WIDESCREEN_HACK_LABEL_TT,
      NULL,
      BEETLE_OPT_WIDESCREEN_HACK_INFO_0_TT,
      NULL,
      "hacks",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(widescreen_hack_aspect_ratio),
      BEETLE_OPT_WIDESCREEN_HACK_ASPECT_RATIO_LABEL_TT,
      NULL,
      BEETLE_OPT_WIDESCREEN_HACK_ASPECT_RATIO_INFO_0_TT,
      NULL,
      "hacks",
      {
         { "16:9",  OPTION_VAL_16_9_TT },
         { "16:10", OPTION_VAL_16_10_TT },
         { "18:9",  OPTION_VAL_18_9_TT },
         { "19:9",  OPTION_VAL_19_9_TT },
         { "20:9",  OPTION_VAL_20_9_TT },
         { "21:9",  OPTION_VAL_21_9_TT }, // 64:27
         { "32:9",  OPTION_VAL_32_9_TT },
         { NULL,    NULL },
      },
      "16:9"
   },
#if defined(HAVE_LIGHTREC)
   {
      BEETLE_OPT(cpu_dynarec),
      BEETLE_OPT_CPU_DYNAREC_LABEL_TT,
      NULL,
      BEETLE_OPT_CPU_DYNAREC_INFO_0_TT,
      NULL,
      "hacks",
      {
         { "disabled",        OPTION_VAL_DISABLED_TT },
         { "execute",         OPTION_VAL_EXECUTE_TT },
         { "execute_one",     OPTION_VAL_EXECUTE_ONE_TT },
         { "run_interpreter", OPTION_VAL_RUN_INTERPRETER_TT },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(dynarec_invalidate),
      BEETLE_OPT_DYNAREC_INVALIDATE_LABEL_TT,
      NULL,
      BEETLE_OPT_DYNAREC_INVALIDATE_INFO_0_TT,
      NULL,
      "hacks",
      {
         { "full", OPTION_VAL_FULL_TT },
         { "dma",  OPTION_VAL_DMA_TT },
         { NULL, NULL },
      },
      "full"
   },
   {
      BEETLE_OPT(dynarec_eventcycles),
      BEETLE_OPT_DYNAREC_EVENTCYCLES_LABEL_TT,
      NULL,
      BEETLE_OPT_DYNAREC_EVENTCYCLES_INFO_0_TT,
      NULL,
      "hacks",
      {
         { "128",  OPTION_VAL_128_TT },
         { "256",  NULL },
         { "384",  NULL },
         { "512",  NULL },
         { "640",  NULL },
         { "768",  NULL },
         { "896",  NULL },
         { "1024", NULL },
         { "1152", NULL },
         { "1280", NULL },
         { "1408", NULL },
         { "1536", NULL },
         { "1664", NULL },
         { "1792", NULL },
         { "1920", NULL },
         { "2048", NULL },
         { NULL, NULL },
      },
      "128"
   },
   {
      BEETLE_OPT(dynarec_spu_samples),
      BEETLE_OPT_DYNAREC_SPU_SAMPLES_LABEL_TT,
      NULL,
      BEETLE_OPT_DYNAREC_SPU_SAMPLES_INFO_0_TT,
      NULL,
      "hacks",
      {
         { "1",  OPTION_VAL_1_O49_TT },
         { "4",  NULL },
         { "16", NULL },
         { NULL, NULL },
      },
      "1"
   },
#endif
   {
      BEETLE_OPT(core_timing_fps),
      BEETLE_OPT_CORE_TIMING_FPS_LABEL_TT,
      NULL,
      BEETLE_OPT_CORE_TIMING_FPS_INFO_0_TT,
      NULL,
      "video",
      {
         { "auto_toggle",       OPTION_VAL_AUTO_TOGGLE_TT },
         { "force_progressive", OPTION_VAL_FORCE_PROGRESSIVE_TT },
         { "force_interlaced",  OPTION_VAL_FORCE_INTERLACED_TT },
      },
      "auto_toggle"
   },
   {
      BEETLE_OPT(deinterlacer),
      BEETLE_OPT_DEINTERLACER_LABEL_TT,
      NULL,
      BEETLE_OPT_DEINTERLACER_INFO_0_TT,
      NULL,
      "video",
      {
         { "weave",      OPTION_VAL_WEAVE_TT },
         { "bob",        OPTION_VAL_BOB_TT },
         { "bob_offset", OPTION_VAL_BOB_OFFSET_TT },
         { "fastmad",    OPTION_VAL_FASTMAD_TT },
         { "off",        "Off" },
         { NULL, NULL },
      },
      "weave"
   },
   {
      BEETLE_OPT(frame_duping),
      BEETLE_OPT_FRAME_DUPING_LABEL_TT,
      NULL,
      BEETLE_OPT_FRAME_DUPING_INFO_0_TT,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(pal_video_timing_override),
      BEETLE_OPT_PAL_VIDEO_TIMING_OVERRIDE_LABEL_TT,
      NULL,
      BEETLE_OPT_PAL_VIDEO_TIMING_OVERRIDE_INFO_0_TT,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(aspect_ratio),
      BEETLE_OPT_ASPECT_RATIO_LABEL_TT,
      NULL,
      BEETLE_OPT_ASPECT_RATIO_INFO_0_TT,
      NULL,
      "video",
      {
         { "corrected", OPTION_VAL_CORRECTED_TT },
         { "uncorrected", OPTION_VAL_UNCORRECTED_TT },
         { "4:3",  OPTION_VAL_4_3_TT },
         { "ntsc", OPTION_VAL_NTSC_TT },
      },
      "corrected"
   },
   {
      BEETLE_OPT(crop_overscan),
      BEETLE_OPT_CROP_OVERSCAN_LABEL_TT,
      NULL,
      BEETLE_OPT_CROP_OVERSCAN_INFO_0_TT,
      NULL,
      "video",
      {
         { "disabled", "Disabled" },
         { "static",   OPTION_VAL_STATIC_TT },
         { "smart",    OPTION_VAL_SMART_TT },
         { NULL, NULL },
      },
      "smart"
   },
   {
      BEETLE_OPT(image_crop),
      BEETLE_OPT_IMAGE_CROP_LABEL_TT,
      NULL,
      BEETLE_OPT_IMAGE_CROP_INFO_0_TT,
      NULL,
      "video",
      {
         { "disabled", "0" },
         { "1px",      OPTION_VAL_1PX_TT },
         { "2px",      OPTION_VAL_2PX_TT },
         { "3px",      OPTION_VAL_3PX_TT },
         { "4px",      OPTION_VAL_4PX_TT },
         { "5px",      OPTION_VAL_5PX_TT },
         { "6px",      OPTION_VAL_6PX_TT },
         { "7px",      OPTION_VAL_7PX_TT },
         { "8px",      OPTION_VAL_8PX_TT },
         { "9px",      OPTION_VAL_9PX_TT },
         { "10px",     OPTION_VAL_10PX_TT },
         { "11px",     OPTION_VAL_11PX_TT },
         { "12px",     OPTION_VAL_12PX_TT },
         { "13px",     OPTION_VAL_13PX_TT },
         { "14px",     OPTION_VAL_14PX_TT },
         { "15px",     OPTION_VAL_15PX_TT },
         { "16px",     OPTION_VAL_16PX_TT },
         { "17px",     OPTION_VAL_17PX_TT },
         { "18px",     OPTION_VAL_18PX_TT },
         { "19px",     OPTION_VAL_19PX_TT },
         { "20px",     OPTION_VAL_20PX_TT },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(image_offset),
      BEETLE_OPT_IMAGE_OFFSET_LABEL_TT,
      NULL,
      BEETLE_OPT_IMAGE_OFFSET_INFO_0_TT,
      NULL,
      "video",
      {
         { "-12px",    OPTION_VAL_12PX_O64_TT },
         { "-11px",    OPTION_VAL_11PX_O64_TT },
         { "-10px",    OPTION_VAL_10PX_O64_TT },
         { "-9px",     OPTION_VAL_9PX_O64_TT },
         { "-8px",     OPTION_VAL_8PX_O64_TT },
         { "-7px",     OPTION_VAL_7PX_O64_TT },
         { "-6px",     OPTION_VAL_6PX_O64_TT },
         { "-5px",     OPTION_VAL_5PX_O64_TT },
         { "-4px",     OPTION_VAL_4PX_O64_TT },
         { "-3px",     OPTION_VAL_3PX_O64_TT },
         { "-2px",     OPTION_VAL_2PX_O64_TT },
         { "-1px",     OPTION_VAL_1PX_O64_TT },
         { "disabled", "0" },
         { "+1px",     OPTION_VAL_1PX_O64_0_TT },
         { "+2px",     OPTION_VAL_2PX_O64_0_TT },
         { "+3px",     OPTION_VAL_3PX_O64_0_TT },
         { "+4px",     OPTION_VAL_4PX_O64_0_TT },
         { "+5px",     OPTION_VAL_5PX_O64_0_TT },
         { "+6px",     OPTION_VAL_6PX_O64_0_TT },
         { "+7px",     OPTION_VAL_7PX_O64_0_TT },
         { "+8px",     OPTION_VAL_8PX_O64_0_TT },
         { "+9px",     OPTION_VAL_9PX_O64_0_TT },
         { "+10px",    OPTION_VAL_10PX_O64_0_TT },
         { "+11px",    OPTION_VAL_11PX_O64_0_TT },
         { "+12px",    OPTION_VAL_12PX_O64_0_TT },
         { NULL, NULL },
      },
      "disabled"
   },
#if defined(HAVE_OPENGL) || defined(HAVE_OPENGLES) || defined(HAVE_VULKAN)
   {
      BEETLE_OPT(image_offset_cycles),
      BEETLE_OPT_IMAGE_OFFSET_CYCLES_LABEL_TT,
      NULL,
      BEETLE_OPT_IMAGE_OFFSET_CYCLES_INFO_0_TT,
      NULL,
      "video",
      {
         { "-40",      NULL },
         { "-39",      NULL },
         { "-38",      NULL },
         { "-37",      NULL },
         { "-36",      NULL },
         { "-35",      NULL },
         { "-34",      NULL },
         { "-33",      NULL },
         { "-32",      NULL },
         { "-31",      NULL },
         { "-30",      NULL },
         { "-29",      NULL },
         { "-28",      NULL },
         { "-27",      NULL },
         { "-26",      NULL },
         { "-25",      NULL },
         { "-24",      NULL },
         { "-23",      NULL },
         { "-22",      NULL },
         { "-21",      NULL },
         { "-20",      NULL },
         { "-19",      NULL },
         { "-18",      NULL },
         { "-17",      NULL },
         { "-16",      NULL },
         { "-15",      NULL },
         { "-14",      NULL },
         { "-13",      NULL },
         { "-12",      NULL },
         { "-11",      NULL },
         { "-10",      NULL },
         { "-9",       NULL },
         { "-8",       NULL },
         { "-7",       NULL },
         { "-6",       NULL },
         { "-5",       NULL },
         { "-4",       NULL },
         { "-3",       NULL },
         { "-2",       NULL },
         { "-1",       NULL },
         { "0",        "0" },
         { "+1",       NULL },
         { "+2",       NULL },
         { "+3",       NULL },
         { "+4",       NULL },
         { "+5",       NULL },
         { "+6",       NULL },
         { "+7",       NULL },
         { "+8",       NULL },
         { "+9",       NULL },
         { "+10",      NULL },
         { "+11",      NULL },
         { "+12",      NULL },
         { "+13",      NULL },
         { "+14",      NULL },
         { "+15",      NULL },
         { "+16",      NULL },
         { "+17",      NULL },
         { "+18",      NULL },
         { "+19",      NULL },
         { "+20",      NULL },
         { "+21",      NULL },
         { "+22",      NULL },
         { "+23",      NULL },
         { "+24",      NULL },
         { "+25",      NULL },
         { "+26",      NULL },
         { "+27",      NULL },
         { "+28",      NULL },
         { "+29",      NULL },
         { "+30",      NULL },
         { "+31",      NULL },
         { "+32",      NULL },
         { "+33",      NULL },
         { "+34",      NULL },
         { "+35",      NULL },
         { "+36",      NULL },
         { "+37",      NULL },
         { "+38",      NULL },
         { "+39",      NULL },
         { "+40",      NULL },
         { NULL, NULL},
      },
      "0"
   },
#endif
   {
      BEETLE_OPT(initial_scanline),
      BEETLE_OPT_INITIAL_SCANLINE_LABEL_TT,
      NULL,
      BEETLE_OPT_INITIAL_SCANLINE_INFO_0_TT,
      NULL,
      "video",
      {
         { "0",  OPTION_VAL_0_O48_TT },
         { "1",  NULL },
         { "2",  NULL },
         { "3",  NULL },
         { "4",  NULL },
         { "5",  NULL },
         { "6",  NULL },
         { "7",  NULL },
         { "8",  NULL },
         { "9",  NULL },
         { "10", NULL },
         { "11", NULL },
         { "12", NULL },
         { "13", NULL },
         { "14", NULL },
         { "15", NULL },
         { "16", NULL },
         { "17", NULL },
         { "18", NULL },
         { "19", NULL },
         { "20", NULL },
         { "21", NULL },
         { "22", NULL },
         { "23", NULL },
         { "24", NULL },
         { "25", NULL },
         { "26", NULL },
         { "27", NULL },
         { "28", NULL },
         { "29", NULL },
         { "30", NULL },
         { "31", NULL },
         { "32", NULL },
         { "33", NULL },
         { "34", NULL },
         { "35", NULL },
         { "36", NULL },
         { "37", NULL },
         { "38", NULL },
         { "39", NULL },
         { "40", NULL },
         { NULL, NULL },
      },
      "0"
   },
   {
      BEETLE_OPT(last_scanline),
      BEETLE_OPT_LAST_SCANLINE_LABEL_TT,
      NULL,
      BEETLE_OPT_LAST_SCANLINE_INFO_0_TT,
      NULL,
      "video",
      {
         { "210", NULL },
         { "211", NULL },
         { "212", NULL },
         { "213", NULL },
         { "214", NULL },
         { "215", NULL },
         { "216", NULL },
         { "217", NULL },
         { "218", NULL },
         { "219", NULL },
         { "220", NULL },
         { "221", NULL },
         { "222", NULL },
         { "223", NULL },
         { "224", NULL },
         { "225", NULL },
         { "226", NULL },
         { "227", NULL },
         { "228", NULL },
         { "229", NULL },
         { "230", NULL },
         { "231", NULL },
         { "232", NULL },
         { "233", NULL },
         { "234", NULL },
         { "235", NULL },
         { "236", NULL },
         { "237", NULL },
         { "238", NULL },
         { "239", OPTION_VAL_239_TT },
         { NULL, NULL },
      },
      "239"
   },
   {
      BEETLE_OPT(initial_scanline_pal),
      BEETLE_OPT_INITIAL_SCANLINE_PAL_LABEL_TT,
      NULL,
      BEETLE_OPT_INITIAL_SCANLINE_PAL_INFO_0_TT,
      NULL,
      "video",
      {
         { "0",  OPTION_VAL_0_O48_TT },
         { "1",  NULL },
         { "2",  NULL },
         { "3",  NULL },
         { "4",  NULL },
         { "5",  NULL },
         { "6",  NULL },
         { "7",  NULL },
         { "8",  NULL },
         { "9",  NULL },
         { "10", NULL },
         { "11", NULL },
         { "12", NULL },
         { "13", NULL },
         { "14", NULL },
         { "15", NULL },
         { "16", NULL },
         { "17", NULL },
         { "18", NULL },
         { "19", NULL },
         { "20", NULL },
         { "21", NULL },
         { "22", NULL },
         { "23", NULL },
         { "24", NULL },
         { "25", NULL },
         { "26", NULL },
         { "27", NULL },
         { "28", NULL },
         { "29", NULL },
         { "30", NULL },
         { "31", NULL },
         { "32", NULL },
         { "33", NULL },
         { "34", NULL },
         { "35", NULL },
         { "36", NULL },
         { "37", NULL },
         { "38", NULL },
         { "39", NULL },
         { "40", NULL },
         { NULL, NULL },
      },
      "0"
   },
   {
      BEETLE_OPT(last_scanline_pal),
      BEETLE_OPT_LAST_SCANLINE_PAL_LABEL_TT,
      NULL,
      BEETLE_OPT_LAST_SCANLINE_PAL_INFO_0_TT,
      NULL,
      "video",
      {
         { "230", NULL },
         { "231", NULL },
         { "232", NULL },
         { "233", NULL },
         { "234", NULL },
         { "235", NULL },
         { "236", NULL },
         { "237", NULL },
         { "238", NULL },
         { "239", NULL },
         { "240", NULL },
         { "241", NULL },
         { "242", NULL },
         { "243", NULL },
         { "244", NULL },
         { "245", NULL },
         { "246", NULL },
         { "247", NULL },
         { "248", NULL },
         { "249", NULL },
         { "250", NULL },
         { "251", NULL },
         { "252", NULL },
         { "253", NULL },
         { "254", NULL },
         { "255", NULL },
         { "256", NULL },
         { "257", NULL },
         { "258", NULL },
         { "259", NULL },
         { "260", NULL },
         { "261", NULL },
         { "262", NULL },
         { "263", NULL },
         { "264", NULL },
         { "265", NULL },
         { "266", NULL },
         { "267", NULL },
         { "268", NULL },
         { "269", NULL },
         { "270", NULL },
         { "271", NULL },
         { "272", NULL },
         { "273", NULL },
         { "274", NULL },
         { "275", NULL },
         { "276", NULL },
         { "277", NULL },
         { "278", NULL },
         { "279", NULL },
         { "280", NULL },
         { "281", NULL },
         { "282", NULL },
         { "283", NULL },
         { "284", NULL },
         { "285", NULL },
         { "286", NULL },
         { "287", OPTION_VAL_287_TT },
         { NULL, NULL },
      },
      "287"
   },
   { NULL, NULL, NULL, NULL, NULL, NULL, {{0}}, NULL },
};
struct retro_core_options_v2 options_tt = {
   option_cats_tt,
   option_defs_tt
};

/* RETRO_LANGUAGE_UK */

#define CATEGORY_SYSTEM_LABEL_UK "Система"
#define CATEGORY_SYSTEM_INFO_0_UK NULL
#define CATEGORY_VIDEO_LABEL_UK "Відео"
#define CATEGORY_VIDEO_INFO_0_UK "Налаштування співвідношення сторін / кадрування / відеофільтрів / пропуску кадрів."
#define CATEGORY_PGXP_LABEL_UK NULL
#define CATEGORY_PGXP_INFO_0_UK "Цей параметр може покращувати графіку в порівнянні з оригінальною консолею. PGXP може виправити викривлення текстур і проблем з Z-fighting."
#define CATEGORY_OSD_LABEL_UK "Зображення на екрані"
#define CATEGORY_OSD_INFO_0_UK "Змінити сповіщення, що показуються на екрані."
#define CATEGORY_INPUT_LABEL_UK "Ввід"
#define CATEGORY_INPUT_INFO_0_UK "Зміна налаштувань світлового пістолету, миші та neGcon."
#define CATEGORY_MEMCARDS_LABEL_UK "Карта пам'яті"
#define CATEGORY_MEMCARDS_INFO_0_UK NULL
#define CATEGORY_HACKS_LABEL_UK "Гаки емуляції"
#define CATEGORY_HACKS_INFO_0_UK NULL
#define BEETLE_OPT_CPU_FREQ_SCALE_LABEL_UK "Масштабування частоти процесора (замінити)"
#define BEETLE_OPT_CPU_FREQ_SCALE_INFO_0_UK "За розкладом (або підкресленням) емульований ЦП PSX. Перемикання може усунути уповільнення та покращити швидкість кадрів у певних іграх за рахунок збільшення показників. Зверніть увагу, що деякі ігри мають внутрішній ліміт частоти кадрів і не можуть отримати користь від блокування. Може викликати певні ефекти швидше, ніж передбачено в деяких назвах при перевищенні."
#define OPTION_VAL_50_UK NULL
#define OPTION_VAL_60_UK NULL
#define OPTION_VAL_70_UK NULL
#define OPTION_VAL_80_UK NULL
#define OPTION_VAL_90_UK NULL
#define OPTION_VAL_100_UK NULL
#define OPTION_VAL_110_UK NULL
#define OPTION_VAL_120_UK NULL
#define OPTION_VAL_130_UK NULL
#define OPTION_VAL_140_UK NULL
#define OPTION_VAL_150_UK NULL
#define OPTION_VAL_160_UK NULL
#define OPTION_VAL_170_UK NULL
#define OPTION_VAL_180_UK NULL
#define OPTION_VAL_190_UK NULL
#define OPTION_VAL_200_UK NULL
#define OPTION_VAL_210_UK NULL
#define OPTION_VAL_220_UK NULL
#define OPTION_VAL_230_UK NULL
#define OPTION_VAL_240_UK NULL
#define OPTION_VAL_250_UK NULL
#define OPTION_VAL_260_UK NULL
#define OPTION_VAL_270_UK NULL
#define OPTION_VAL_280_UK NULL
#define OPTION_VAL_290_UK NULL
#define OPTION_VAL_300_UK NULL
#define OPTION_VAL_310_UK NULL
#define OPTION_VAL_320_UK NULL
#define OPTION_VAL_330_UK NULL
#define OPTION_VAL_340_UK NULL
#define OPTION_VAL_350_UK NULL
#define OPTION_VAL_360_UK NULL
#define OPTION_VAL_370_UK NULL
#define OPTION_VAL_380_UK NULL
#define OPTION_VAL_390_UK NULL
#define OPTION_VAL_400_UK NULL
#define OPTION_VAL_410_UK NULL
#define OPTION_VAL_420_UK NULL
#define OPTION_VAL_430_UK NULL
#define OPTION_VAL_440_UK NULL
#define OPTION_VAL_450_UK NULL
#define OPTION_VAL_460_UK NULL
#define OPTION_VAL_470_UK NULL
#define OPTION_VAL_480_UK NULL
#define OPTION_VAL_490_UK NULL
#define OPTION_VAL_500_UK NULL
#define OPTION_VAL_510_UK NULL
#define OPTION_VAL_520_UK NULL
#define OPTION_VAL_530_UK NULL
#define OPTION_VAL_540_UK NULL
#define OPTION_VAL_550_UK NULL
#define OPTION_VAL_560_UK NULL
#define OPTION_VAL_570_UK NULL
#define OPTION_VAL_580_UK NULL
#define OPTION_VAL_590_UK NULL
#define OPTION_VAL_600_UK NULL
#define OPTION_VAL_610_UK NULL
#define OPTION_VAL_620_UK NULL
#define OPTION_VAL_630_UK NULL
#define OPTION_VAL_640_UK NULL
#define OPTION_VAL_650_UK NULL
#define OPTION_VAL_660_UK NULL
#define OPTION_VAL_670_UK NULL
#define OPTION_VAL_680_UK NULL
#define OPTION_VAL_690_UK NULL
#define OPTION_VAL_700_UK NULL
#define OPTION_VAL_710_UK NULL
#define OPTION_VAL_720_UK NULL
#define OPTION_VAL_730_UK NULL
#define OPTION_VAL_740_UK NULL
#define OPTION_VAL_750_UK NULL
#define BEETLE_OPT_GPU_OVERCLOCK_LABEL_UK "Перевизначення частоти GPU Rasterizer"
#define BEETLE_OPT_GPU_OVERCLOCK_INFO_0_UK "Увімкнути блокування 2D растеризатора, що міститься в емулятивному графічному режимі PSX. Не покращує 3D-рендеринг і загалом мало ефекту."
#define OPTION_VAL_1X_NATIVE_UK "1x (рідна)"
#define OPTION_VAL_2X_UK NULL
#define OPTION_VAL_4X_UK NULL
#define OPTION_VAL_8X_UK NULL
#define OPTION_VAL_16X_UK NULL
#define OPTION_VAL_32X_UK "32х"
#define BEETLE_OPT_GTE_OVERCLOCK_LABEL_UK NULL
#define BEETLE_OPT_GTE_OVERCLOCK_INFO_0_UK "Понизити всі емульовані GTE (процесорний роцетор для 3D графіки) операції з постійною затримкою в один цикл. Для ігор, які сильно використовують GTE, це може значно покращити частоту кадрів та стабільність часу проведення рамів."
#define BEETLE_OPT_CD_ACCESS_METHOD_LABEL_UK NULL
#define BEETLE_OPT_CD_ACCESS_METHOD_INFO_0_UK NULL
#define OPTION_VAL_SYNC_UK "Синхронний"
#define OPTION_VAL_ASYNC_UK "Асинхронний"
#define OPTION_VAL_PRECACHE_UK "Попередній кеш"
#define BEETLE_OPT_CD_FASTLOAD_LABEL_UK "Швидкість завантаження CD"
#define BEETLE_OPT_CD_FASTLOAD_INFO_0_UK NULL
#define OPTION_VAL_2X_NATIVE_UK "2x (рідна)"
#define OPTION_VAL_6X_UK NULL
#define OPTION_VAL_10X_UK "10х"
#define OPTION_VAL_12X_UK "12 х"
#define OPTION_VAL_14X_UK NULL
#define BEETLE_OPT_SKIP_BIOS_LABEL_UK "Пропустити BIOS"
#define BEETLE_OPT_SKIP_BIOS_INFO_0_UK "Пропустити завантажувальну анімацію PlayStation BIOS при завантаженні вмісту. Примітка: Увімкнення цієї проблеми щодо сумісності з рядом ігор (PAL копіює захищені ігри, Saga Frontier та інше)."
#define BEETLE_OPT_OVERRIDE_BIOS_LABEL_UK NULL
#define BEETLE_OPT_OVERRIDE_BIOS_INFO_0_UK NULL
#define OPTION_VAL_PSXONPSP_UK NULL
#define OPTION_VAL_PS1_ROM_UK "PS3 BIOS"
#define OPTION_VAL_OPENBIOS_UK NULL
#define BEETLE_OPT_REGION_LABEL_UK "Регіон системи"
#define BEETLE_OPT_REGION_INFO_0_UK NULL
#define OPTION_VAL_AUTO_UK "Автоматично"
#define OPTION_VAL_NTSC_J_UK NULL
#define OPTION_VAL_NTSC_U_UK NULL
#define OPTION_VAL_PAL_UK NULL
#define BEETLE_OPT_INTERNAL_RESOLUTION_LABEL_UK "Роздільна здатність GPU"
#define BEETLE_OPT_INTERNAL_RESOLUTION_INFO_0_UK NULL
#define BEETLE_OPT_RENDERER_LABEL_UK NULL
#define BEETLE_OPT_RENDERER_INFO_0_UK NULL
#define OPTION_VAL_HARDWARE_UK "Апаратний (автоматично)"
#define OPTION_VAL_HARDWARE_GL_UK "Апаратна (OpenGL)"
#define OPTION_VAL_HARDWARE_VK_UK NULL
#define OPTION_VAL_SOFTWARE_UK "Програмне забезпечення"
#define BEETLE_OPT_RENDERER_SOFTWARE_FB_LABEL_UK "Програмний фреймбуфер"
#define BEETLE_OPT_RENDERER_SOFTWARE_FB_INFO_0_UK "Увімкнути точну емуляцію ефектів framebuffer (наприклад, рух blur, FF7 battle swirl) при використанні візуалізації обладнання запустивши копію програмного рендеру в рідній роздільній здатності у фоновому режимі. Якщо вимкнено, ці операції не виконуються (OpenGL) або будуть показані на GPU (Vulkan). Вимкнення може підвищити продуктивність, але може призвести до серйозних графічних помилок. Якщо не впевнені, вимкніть цю функцію."
#define BEETLE_OPT_DEPTH_LABEL_UK "Внутрішня глибина кольору"
#define BEETLE_OPT_DEPTH_INFO_0_UK NULL
#define OPTION_VAL_16BPP_NATIVE_UK "16 bpp (Рідне)"
#define OPTION_VAL_32BPP_UK NULL
#define BEETLE_OPT_DITHER_MODE_LABEL_UK "Шаблон дизерингу"
#define BEETLE_OPT_DITHER_MODE_INFO_0_UK NULL
#define OPTION_VAL_INTERNAL_RESOLUTION_UK "Роздільна здатність"
#define BEETLE_OPT_SCALED_UV_OFFSET_LABEL_UK "Зміщення текстури UV"
#define BEETLE_OPT_SCALED_UV_OFFSET_INFO_0_UK "Зразок текстур для 3D багатокутників з зсуву для більш ніж 1x внутрішнього розв'язання. Знижує морські текстури, але може спричинити ненавмисні графічні клешки."
#define BEETLE_OPT_FILTER_LABEL_UK "Фільтрація текстур"
#define BEETLE_OPT_FILTER_INFO_0_UK NULL
#define OPTION_VAL_NEAREST_UK "Найближчі"
#define OPTION_VAL_SABR_UK NULL
#define OPTION_VAL_XBR_UK NULL
#define OPTION_VAL_BILINEAR_UK "Білінійна"
#define OPTION_VAL_3_POINT_UK "3-точка"
#define OPTION_VAL_JINC2_UK "ПІДПИСКА 2"
#define BEETLE_OPT_FILTER_EXCLUDE_SPRITE_LABEL_UK "Виключити спрайти з фільтрації"
#define BEETLE_OPT_FILTER_EXCLUDE_SPRITE_INFO_0_UK "Не застосовувати фільтрування текстур до спрайтів. Запобігає мореплавцям у різних іграх з 2D фонами рендерного спрайту. Використовуйте разом з Адаптивним Згладжуванням або іншими пост-обробками для найкращого ефекту."
#define OPTION_VAL_OPAQUE_UK "Лише непрозора"
#define OPTION_VAL_ALL_UK "Непрозорий і напівпрозорий"
#define BEETLE_OPT_FILTER_EXCLUDE_2D_POLYGON_LABEL_UK "Виключити 2D Полігони з фільтрації"
#define BEETLE_OPT_FILTER_EXCLUDE_2D_POLYGON_INFO_0_UK "Не застосовувати фільтрування текстур до 2D багатокутників. 2D багатокутники виявлені з евристичним і там можуть бути відблиски. Використовуйте разом з Адаптивним Згладжуванням або іншими пост-обробками для найкращого ефекту."
#define BEETLE_OPT_ADAPTIVE_SMOOTHING_LABEL_UK "Адаптивне згладжування"
#define BEETLE_OPT_ADAPTIVE_SMOOTHING_INFO_0_UK "Згладити 2D файли та елементи інтерфейсу без розмиття 3D об’єктів. Підтримується лише програмою Vulkan renderer."
#define BEETLE_OPT_SUPER_SAMPLING_LABEL_UK "Супервибірка (Знижка до корінної роздільної здатності)"
#define BEETLE_OPT_SUPER_SAMPLING_INFO_0_UK "Знижувати показовий вміст з вбудованої внутрішньої роздільної здатності до нативної роздільності. Поєднання цього з більшими мультиплікаторами внутрішньої роздільної здатності дозволяє ігри відображати анти-відчутні 3D об'єкти на рідній низькій роздільності. Робить кращі результати при застосуванні до заголовків, які змішують 2D та 3D елементи (наприклад, 3D символи на попередньо візуальному фоні) і добре працює разом з шейдерами CRT. Підтримується лише Vulkan renderer. Примітка: 'Dithering Pattern' рекомендується при увімкненні цієї опції."
#define BEETLE_OPT_MSAA_LABEL_UK "Багатовибіркове Аліазингу Anti Aliasing"
#define BEETLE_OPT_MSAA_INFO_0_UK NULL
#define OPTION_VAL_1X_UK NULL
#define BEETLE_OPT_MDEC_YUV_LABEL_UK "Фільтр MDEC YUV Chroma"
#define BEETLE_OPT_MDEC_YUV_INFO_0_UK "Покращити якість відтворення FMV за рахунок зменшення артефактів макроблокінга (квадратних). Тільки за підтримки візуального рендеру Vulkan renderer."
#define BEETLE_OPT_TRACK_TEXTURES_LABEL_UK "Відслідковувати текстури"
#define BEETLE_OPT_TRACK_TEXTURES_INFO_0_UK NULL
#define BEETLE_OPT_DUMP_TEXTURES_LABEL_UK "Дамп текстур"
#define BEETLE_OPT_DUMP_TEXTURES_INFO_0_UK "Дамп використаних текстур для <cd>-texture-dump/"
#define BEETLE_OPT_REPLACE_TEXTURES_LABEL_UK "Замінити текстури"
#define BEETLE_OPT_REPLACE_TEXTURES_INFO_0_UK "Замінити текстури за допомогою HD версій з <cd>-texture-replacements/"
#define BEETLE_OPT_PGXP_MODE_LABEL_UK "Режим операції PGXP"
#define BEETLE_OPT_PGXP_MODE_INFO_0_UK NULL
#define OPTION_VAL_MEMORY_ONLY_UK "Лише пам’ять"
#define OPTION_VAL_MEMORY_CPU_UK "Пам'ять + ЦП (довгий)"
#define BEETLE_OPT_PGXP_2D_TOL_LABEL_UK "Похибка в PGXP 2D геометрії"
#define BEETLE_OPT_PGXP_2D_TOL_INFO_0_UK "Приховати більше кривих помилок в операціях PGXP: значення вказує на толерантність до яких значення PGXP будуть зберігатися у випадку геометрій без правильної детальної інформації."
#define OPTION_VAL_0PX_UK NULL
#define OPTION_VAL_1PX_UK NULL
#define OPTION_VAL_2PX_UK NULL
#define OPTION_VAL_3PX_UK NULL
#define OPTION_VAL_4PX_UK NULL
#define OPTION_VAL_5PX_UK NULL
#define OPTION_VAL_6PX_UK NULL
#define OPTION_VAL_7PX_UK NULL
#define OPTION_VAL_8PX_UK NULL
#define BEETLE_OPT_PGXP_NCLIP_LABEL_UK "Примітивний корпус PGXP"
#define BEETLE_OPT_PGXP_NCLIP_INFO_0_UK "Використовуйте реалізацію PGXP NCLIP. Покращує появу за рахунок зменшення отворів в геометріях з координатами PGXP. Відомо, що деякі ігри зачиняються в різних обставинах."
#define BEETLE_OPT_PGXP_VERTEX_LABEL_UK "Кеш вершини PGXP"
#define BEETLE_OPT_PGXP_VERTEX_INFO_0_UK "Кешувати вершинні позиції PGXP-для повторного використання з допомогою багатокутника. потенційно може покращити вирівнювання об'єктів та зменшити видимі трави під час рендерингу текстур, але помилкові розміри при запиті кеш може призвести до графічних відблисків. Зараз рекомендуємо залишити цю опцію відключеною. Цей параметр застосовується тільки коли увімкнено режим PGXP операції. Підтримується тільки візуальні засоби."
#define BEETLE_OPT_PGXP_TEXTURE_LABEL_UK "Виправлення текстури PGXP"
#define BEETLE_OPT_PGXP_TEXTURE_INFO_0_UK NULL
#define BEETLE_OPT_DISPLAY_INTERNAL_FPS_LABEL_UK "Показувати внутрішню FPS"
#define BEETLE_OPT_DISPLAY_INTERNAL_FPS_INFO_0_UK "Відображення інтенсивності кадрів за вказаним вмістом PlayStation системи. Примітка: Потрібні сповіщення для увімкненого в лібро-інтерфейсі."
#define BEETLE_OPT_DISPLAY_OSD_LABEL_UK "Відображати повідомлення OSD"
#define BEETLE_OPT_DISPLAY_OSD_INFO_0_UK "Відображати повідомлень OSD, згенеровані в ядрі."
#define BEETLE_OPT_DISPLAY_VRAM_LABEL_UK "Відображати повну версію VRAM (налагодження)"
#define BEETLE_OPT_DISPLAY_VRAM_INFO_0_UK "Візуалізувати всю емульовану версію VRAM. Підтримуються лише програмісти OpenGL і Vulkan апаратних рендерів. Примітка: Це для потреб відладки, і, як правило, має бути вимкнено."
#define BEETLE_OPT_ANALOG_CALIBRATION_LABEL_UK "Аналоговий самокалібрування"
#define BEETLE_OPT_ANALOG_CALIBRATION_INFO_0_UK NULL
#define BEETLE_OPT_ANALOG_TOGGLE_LABEL_UK "Увімкнути/вимкнути аналоговий режим DualShock"
#define BEETLE_OPT_ANALOG_TOGGLE_INFO_0_UK "Якщо тип пристрою введення DualShock, ця опція дозволяє емуляції DualShock перемикатися між ЦИФРОВИМ і АНАЛОГОВИМ режимами, як на оригінальному обладнанні. Якщо опцію вимкнено, DualShock буде заблоковано у режимі АНАЛОГ, а якщо увімкнено, DualShock можна перемикати між ЦИФРОВИМ та АНАЛОГОВИМ режимами за допомогою вибраної комбінації кнопок."
#define OPTION_VAL_ENABLED_ANALOG_UK "Типовий аналог"
#define BEETLE_OPT_ANALOG_TOGGLE_COMBO_LABEL_UK "Комбінація Аналогового Режиму DualShock"
#define BEETLE_OPT_ANALOG_TOGGLE_COMBO_INFO_0_UK NULL
#define OPTION_VAL_L1_L2_R1_R2_START_SELECT_UK "L1 + L2 + R1 + R2 + Старт + Вибрати"
#define OPTION_VAL_L1_R1_SELECT_UK "L1 + R1 + Вибрати"
#define OPTION_VAL_L1_R1_START_UK "L1 + R1 + Старт"
#define OPTION_VAL_L1_R1_L3_UK NULL
#define OPTION_VAL_L1_R1_R3_UK NULL
#define OPTION_VAL_L2_R2_SELECT_UK "L2 + R2 + Вибрати"
#define OPTION_VAL_L2_R2_START_UK "L2 + R2 + Старт"
#define OPTION_VAL_L2_R2_L3_UK NULL
#define OPTION_VAL_L2_R2_R3_UK NULL
#define OPTION_VAL_L3_R3_UK NULL
#define BEETLE_OPT_ANALOG_TOGGLE_HOLD_LABEL_UK "Комбінація Аналогового Режиму DualShock"
#define BEETLE_OPT_ANALOG_TOGGLE_HOLD_INFO_0_UK NULL
#define OPTION_VAL_0_UK "0 Секунд Затримки"
#define OPTION_VAL_1_UK "1 Секунда Затримки"
#define OPTION_VAL_2_UK "2 Секунди Затримки"
#define OPTION_VAL_3_UK "3 Секунди Затримки"
#define OPTION_VAL_4_UK "4 Секунди Затримки"
#define OPTION_VAL_5_UK "5 Секунд Затримки"
#define BEETLE_OPT_ENABLE_MULTITAP_PORT1_LABEL_UK NULL
#define BEETLE_OPT_ENABLE_MULTITAP_PORT1_INFO_0_UK "Увімкнути функціональність мультікарті на порту 1."
#define BEETLE_OPT_ENABLE_MULTITAP_PORT2_LABEL_UK NULL
#define BEETLE_OPT_ENABLE_MULTITAP_PORT2_INFO_0_UK "Увімкнути функціональність мультикарти на порту 2."
#define BEETLE_OPT_GUN_INPUT_MODE_LABEL_UK "Режим введення пістолета"
#define BEETLE_OPT_GUN_INPUT_MODE_INFO_0_UK NULL
#define OPTION_VAL_LIGHTGUN_UK "Легкий пістолет"
#define OPTION_VAL_TOUCHSCREEN_UK "Сенсорний екран"
#define BEETLE_OPT_GUN_CURSOR_LABEL_UK "Курсор гармати"
#define BEETLE_OPT_GUN_CURSOR_INFO_0_UK NULL
#define OPTION_VAL_CROSS_UK "Кросс"
#define OPTION_VAL_DOT_UK "Крапка"
#define BEETLE_OPT_CROSSHAIR_COLOR_P1_LABEL_UK NULL
#define BEETLE_OPT_CROSSHAIR_COLOR_P1_INFO_0_UK NULL
#define OPTION_VAL_RED_UK "Червоний"
#define OPTION_VAL_BLUE_UK "Синій"
#define OPTION_VAL_GREEN_UK "Зелений"
#define OPTION_VAL_ORANGE_UK "Помаранчевий"
#define OPTION_VAL_YELLOW_UK "Жовтий"
#define OPTION_VAL_CYAN_UK "Бірюзовий"
#define OPTION_VAL_PINK_UK "Рожевий"
#define OPTION_VAL_PURPLE_UK "Пурпурний"
#define OPTION_VAL_BLACK_UK "Чорний колір"
#define OPTION_VAL_WHITE_UK "Білі"
#define BEETLE_OPT_CROSSHAIR_COLOR_P2_LABEL_UK NULL
#define BEETLE_OPT_CROSSHAIR_COLOR_P2_INFO_0_UK NULL
#define BEETLE_OPT_MOUSE_SENSITIVITY_LABEL_UK "Чутливість миші"
#define BEETLE_OPT_MOUSE_SENSITIVITY_INFO_0_UK NULL
#define OPTION_VAL_5_O42_UK NULL
#define OPTION_VAL_10_UK NULL
#define OPTION_VAL_15_UK NULL
#define OPTION_VAL_20_UK NULL
#define OPTION_VAL_25_UK NULL
#define OPTION_VAL_30_UK NULL
#define OPTION_VAL_35_UK NULL
#define OPTION_VAL_40_UK NULL
#define OPTION_VAL_45_UK NULL
#define OPTION_VAL_55_UK NULL
#define OPTION_VAL_65_UK NULL
#define OPTION_VAL_75_UK NULL
#define OPTION_VAL_85_UK NULL
#define OPTION_VAL_95_UK NULL
#define OPTION_VAL_100_O42_UK NULL
#define OPTION_VAL_105_UK NULL
#define OPTION_VAL_115_UK NULL
#define OPTION_VAL_125_UK NULL
#define OPTION_VAL_135_UK NULL
#define OPTION_VAL_145_UK NULL
#define OPTION_VAL_155_UK NULL
#define OPTION_VAL_165_UK NULL
#define OPTION_VAL_175_UK NULL
#define OPTION_VAL_185_UK NULL
#define OPTION_VAL_195_UK NULL
#define BEETLE_OPT_NEGCON_RESPONSE_LABEL_UK "відповідь неGcon Twist"
#define BEETLE_OPT_NEGCON_RESPONSE_INFO_0_UK NULL
#define OPTION_VAL_LINEAR_UK "Лінійна"
#define OPTION_VAL_QUADRATIC_UK "Квадратичний"
#define OPTION_VAL_CUBIC_UK "Кубічний"
#define BEETLE_OPT_NEGCON_DEADZONE_LABEL_UK "неGcon Твіст Деадцоне"
#define BEETLE_OPT_NEGCON_DEADZONE_INFO_0_UK NULL
#define OPTION_VAL_0_O44_UK NULL
#define BEETLE_OPT_USE_MEDNAFEN_MEMCARD0_METHOD_LABEL_UK NULL
#define BEETLE_OPT_USE_MEDNAFEN_MEMCARD0_METHOD_INFO_0_UK NULL
#define OPTION_VAL_LIBRETRO_UK "Лібретро"
#define OPTION_VAL_MEDNAFEN_UK NULL
#define BEETLE_OPT_ENABLE_MEMCARD1_LABEL_UK NULL
#define BEETLE_OPT_ENABLE_MEMCARD1_INFO_0_UK NULL
#define BEETLE_OPT_SHARED_MEMORY_CARDS_LABEL_UK NULL
#define BEETLE_OPT_SHARED_MEMORY_CARDS_INFO_0_UK NULL
#define BEETLE_OPT_MEMCARD_LEFT_INDEX_LABEL_UK NULL
#define BEETLE_OPT_MEMCARD_LEFT_INDEX_INFO_0_UK NULL
#define OPTION_VAL_0_O48_UK NULL
#define BEETLE_OPT_MEMCARD_RIGHT_INDEX_LABEL_UK NULL
#define BEETLE_OPT_MEMCARD_RIGHT_INDEX_INFO_0_UK NULL
#define OPTION_VAL_1_O49_UK NULL
#define BEETLE_OPT_LINE_RENDER_LABEL_UK "Вирівняти хак"
#define BEETLE_OPT_LINE_RENDER_INFO_0_UK NULL
#define OPTION_VAL_DEFAULT_UK "За замовчуванням"
#define OPTION_VAL_AGGRESSIVE_UK "Агресивний"
#define BEETLE_OPT_WIDESCREEN_HACK_LABEL_UK "Hack для Широкого екрану"
#define BEETLE_OPT_WIDESCREEN_HACK_INFO_0_UK "Візуалізація 3D-контенту анаморфічно та виведення випроміненого фреймбурга при широко поширеному співвідношенні сторін. Виробляє найкращі результати із повністю 3D іграми. 2D елементи будуть горизонтально розтягнуті і можуть бути неправильно вирівнювані."
#define BEETLE_OPT_WIDESCREEN_HACK_ASPECT_RATIO_LABEL_UK "Співвідношення сторін широкого екрану"
#define BEETLE_OPT_WIDESCREEN_HACK_ASPECT_RATIO_INFO_0_UK NULL
#define OPTION_VAL_16_9_UK NULL
#define OPTION_VAL_16_10_UK NULL
#define OPTION_VAL_18_9_UK NULL
#define OPTION_VAL_19_9_UK NULL
#define OPTION_VAL_20_9_UK NULL
#define OPTION_VAL_21_9_UK NULL
#define OPTION_VAL_32_9_UK NULL
#define BEETLE_OPT_CPU_DYNAREC_LABEL_UK NULL
#define BEETLE_OPT_CPU_DYNAREC_INFO_0_UK "Динамічно перекомпільовані інструкції процесора з нативних інструкцій. Значно швидше, ніж інтерпретатор, але таймінг ЦП менш точний і може містити помилки."
#define OPTION_VAL_DISABLED_UK "Вимкнено (Перетин жуків)"
#define OPTION_VAL_EXECUTE_UK "Максимальна продуктивність"
#define OPTION_VAL_EXECUTE_ONE_UK "Перевірити таймінг циклу"
#define OPTION_VAL_RUN_INTERPRETER_UK "Інтерпретатор Lightrec"
#define BEETLE_OPT_DYNAREC_INVALIDATE_LABEL_UK "Перевірка коду Dynarec"
#define BEETLE_OPT_DYNAREC_INVALIDATE_INFO_0_UK "Деякі ігри потребують \"Full\" перевірки, деякі вимагають \"DMA Only\"."
#define OPTION_VAL_FULL_UK "Повна"
#define OPTION_VAL_DMA_UK "DMA лише (трохи швидше)"
#define BEETLE_OPT_DYNAREC_EVENTCYCLES_LABEL_UK "Dynarec DMA/GPU/MDEC/Timer циклів події"
#define BEETLE_OPT_DYNAREC_EVENTCYCLES_INFO_0_UK "Максимальна кількість циклів, що виконуються процесором перед встановленням оновлення GPU/DMA/MDEC/Timer, більше число буде швидшим, проте значно менше впливає на інтерпретатор Beetle, ніж dynarec."
#define OPTION_VAL_128_UK "128 (за замовчуванням)"
#define BEETLE_OPT_DYNAREC_SPU_SAMPLES_LABEL_UK "Динарек SPU зразки"
#define BEETLE_OPT_DYNAREC_SPU_SAMPLES_INFO_0_UK "Максимум зразків SPU для запуску перед тим, як буде встановлено оновлення SPU, вище число буде швидшим, але призведе до падіння звуку в деяких іграх із іншими, окрім 1."
#define BEETLE_OPT_CORE_TIMING_FPS_LABEL_UK "Таймінг з ядром та повідомлення на FPS"
#define BEETLE_OPT_CORE_TIMING_FPS_INFO_0_UK NULL
#define OPTION_VAL_AUTO_TOGGLE_UK NULL
#define OPTION_VAL_FORCE_PROGRESSIVE_UK NULL
#define OPTION_VAL_FORCE_INTERLACED_UK "Примусова швидкість перетину"
#define BEETLE_OPT_DEINTERLACER_LABEL_UK "Метод Деінтерлейса"
#define BEETLE_OPT_DEINTERLACER_INFO_0_UK NULL
#define OPTION_VAL_WEAVE_UK "Візерунок"
#define OPTION_VAL_BOB_UK "Боб"
#define OPTION_VAL_BOB_OFFSET_UK NULL
#define OPTION_VAL_FASTMAD_UK NULL
#define BEETLE_OPT_FRAME_DUPING_LABEL_UK NULL
#define BEETLE_OPT_FRAME_DUPING_INFO_0_UK NULL
#define BEETLE_OPT_PAL_VIDEO_TIMING_OVERRIDE_LABEL_UK NULL
#define BEETLE_OPT_PAL_VIDEO_TIMING_OVERRIDE_INFO_0_UK NULL
#define BEETLE_OPT_ASPECT_RATIO_LABEL_UK "Співвідношення сторін ядра"
#define BEETLE_OPT_ASPECT_RATIO_INFO_0_UK NULL
#define OPTION_VAL_CORRECTED_UK "Правильно"
#define OPTION_VAL_UNCORRECTED_UK "Не коректне"
#define OPTION_VAL_4_3_UK "Примусово 4:3"
#define OPTION_VAL_NTSC_UK "Форс НТК"
#define BEETLE_OPT_CROP_OVERSCAN_LABEL_UK "Обрізати змію"
#define BEETLE_OPT_CROP_OVERSCAN_INFO_0_UK NULL
#define OPTION_VAL_STATIC_UK "Горизонтально"
#define OPTION_VAL_SMART_UK NULL
#define BEETLE_OPT_IMAGE_CROP_LABEL_UK "Додаткове обрізання"
#define BEETLE_OPT_IMAGE_CROP_INFO_0_UK NULL
#define OPTION_VAL_9PX_UK "9пікс"
#define OPTION_VAL_10PX_UK NULL
#define OPTION_VAL_11PX_UK NULL
#define OPTION_VAL_12PX_UK "12пікс"
#define OPTION_VAL_13PX_UK "13пікс"
#define OPTION_VAL_14PX_UK NULL
#define OPTION_VAL_15PX_UK NULL
#define OPTION_VAL_16PX_UK NULL
#define OPTION_VAL_17PX_UK NULL
#define OPTION_VAL_18PX_UK NULL
#define OPTION_VAL_19PX_UK "19пікс"
#define OPTION_VAL_20PX_UK NULL
#define BEETLE_OPT_IMAGE_OFFSET_LABEL_UK "Обрізане зображення зміщення"
#define BEETLE_OPT_IMAGE_OFFSET_INFO_0_UK NULL
#define OPTION_VAL_12PX_O64_UK NULL
#define OPTION_VAL_11PX_O64_UK NULL
#define OPTION_VAL_10PX_O64_UK NULL
#define OPTION_VAL_9PX_O64_UK NULL
#define OPTION_VAL_8PX_O64_UK NULL
#define OPTION_VAL_7PX_O64_UK NULL
#define OPTION_VAL_6PX_O64_UK NULL
#define OPTION_VAL_5PX_O64_UK NULL
#define OPTION_VAL_4PX_O64_UK NULL
#define OPTION_VAL_3PX_O64_UK NULL
#define OPTION_VAL_2PX_O64_UK NULL
#define OPTION_VAL_1PX_O64_UK NULL
#define OPTION_VAL_1PX_O64_0_UK NULL
#define OPTION_VAL_2PX_O64_0_UK NULL
#define OPTION_VAL_3PX_O64_0_UK NULL
#define OPTION_VAL_4PX_O64_0_UK NULL
#define OPTION_VAL_5PX_O64_0_UK NULL
#define OPTION_VAL_6PX_O64_0_UK NULL
#define OPTION_VAL_7PX_O64_0_UK NULL
#define OPTION_VAL_8PX_O64_0_UK NULL
#define OPTION_VAL_9PX_O64_0_UK NULL
#define OPTION_VAL_10PX_O64_0_UK NULL
#define OPTION_VAL_11PX_O64_0_UK NULL
#define OPTION_VAL_12PX_O64_0_UK NULL
#define BEETLE_OPT_IMAGE_OFFSET_CYCLES_LABEL_UK "Горизонтальний зсув зображення (GPU цикл)"
#define BEETLE_OPT_IMAGE_OFFSET_CYCLES_INFO_0_UK NULL
#define BEETLE_OPT_INITIAL_SCANLINE_LABEL_UK "Початковий рядок сканування - NTSC"
#define BEETLE_OPT_INITIAL_SCANLINE_INFO_0_UK NULL
#define BEETLE_OPT_LAST_SCANLINE_LABEL_UK "Останній скан рядок - NTSC"
#define BEETLE_OPT_LAST_SCANLINE_INFO_0_UK NULL
#define OPTION_VAL_239_UK "239 (Типово)"
#define BEETLE_OPT_INITIAL_SCANLINE_PAL_LABEL_UK "Початковий рядок сканування - PAL"
#define BEETLE_OPT_INITIAL_SCANLINE_PAL_INFO_0_UK NULL
#define BEETLE_OPT_LAST_SCANLINE_PAL_LABEL_UK "Останнє сканування рядка - PAL"
#define BEETLE_OPT_LAST_SCANLINE_PAL_INFO_0_UK NULL
#define OPTION_VAL_287_UK "287 (Типово)"

struct retro_core_option_v2_category option_cats_uk[] = {
   {
      "system",
      CATEGORY_SYSTEM_LABEL_UK,
      CATEGORY_SYSTEM_INFO_0_UK
   },
   {
      "video",
      CATEGORY_VIDEO_LABEL_UK,
      CATEGORY_VIDEO_INFO_0_UK
   },
   {
      "pgxp",
      CATEGORY_PGXP_LABEL_UK,
      CATEGORY_PGXP_INFO_0_UK
   },
   {
      "osd",
      CATEGORY_OSD_LABEL_UK,
      CATEGORY_OSD_INFO_0_UK
   },
   {
      "input",
      CATEGORY_INPUT_LABEL_UK,
      CATEGORY_INPUT_INFO_0_UK
   },
   {
      "memcards",
      CATEGORY_MEMCARDS_LABEL_UK,
      CATEGORY_MEMCARDS_INFO_0_UK
   },
   {
      "hacks",
      CATEGORY_HACKS_LABEL_UK,
      CATEGORY_HACKS_INFO_0_UK
   },
   { NULL, NULL, NULL },
};
struct retro_core_option_v2_definition option_defs_uk[] = {
   {
      BEETLE_OPT(cpu_freq_scale),
      BEETLE_OPT_CPU_FREQ_SCALE_LABEL_UK,
      NULL,
      BEETLE_OPT_CPU_FREQ_SCALE_INFO_0_UK,
      NULL,
      "system",
      {
         { "50%",           OPTION_VAL_50_UK },
         { "60%",           OPTION_VAL_60_UK },
         { "70%",           OPTION_VAL_70_UK },
         { "80%",           OPTION_VAL_80_UK },
         { "90%",           OPTION_VAL_90_UK },
         { "100%",          OPTION_VAL_100_UK },
         { "110%",          OPTION_VAL_110_UK },
         { "120%",          OPTION_VAL_120_UK },
         { "130%",          OPTION_VAL_130_UK },
         { "140%",          OPTION_VAL_140_UK },
         { "150%",          OPTION_VAL_150_UK },
         { "160%",          OPTION_VAL_160_UK },
         { "170%",          OPTION_VAL_170_UK },
         { "180%",          OPTION_VAL_180_UK },
         { "190%",          OPTION_VAL_190_UK },
         { "200%",          OPTION_VAL_200_UK },
         { "210%",          OPTION_VAL_210_UK },
         { "220%",          OPTION_VAL_220_UK },
         { "230%",          OPTION_VAL_230_UK },
         { "240%",          OPTION_VAL_240_UK },
         { "250%",          OPTION_VAL_250_UK },
         { "260%",          OPTION_VAL_260_UK },
         { "270%",          OPTION_VAL_270_UK },
         { "280%",          OPTION_VAL_280_UK },
         { "290%",          OPTION_VAL_290_UK },
         { "300%",          OPTION_VAL_300_UK },
         { "310%",          OPTION_VAL_310_UK },
         { "320%",          OPTION_VAL_320_UK },
         { "330%",          OPTION_VAL_330_UK },
         { "340%",          OPTION_VAL_340_UK },
         { "350%",          OPTION_VAL_350_UK },
         { "360%",          OPTION_VAL_360_UK },
         { "370%",          OPTION_VAL_370_UK },
         { "380%",          OPTION_VAL_380_UK },
         { "390%",          OPTION_VAL_390_UK },
         { "400%",          OPTION_VAL_400_UK },
         { "410%",          OPTION_VAL_410_UK },
         { "420%",          OPTION_VAL_420_UK },
         { "430%",          OPTION_VAL_430_UK },
         { "440%",          OPTION_VAL_440_UK },
         { "450%",          OPTION_VAL_450_UK },
         { "460%",          OPTION_VAL_460_UK },
         { "470%",          OPTION_VAL_470_UK },
         { "480%",          OPTION_VAL_480_UK },
         { "490%",          OPTION_VAL_490_UK },
         { "500%",          OPTION_VAL_500_UK },
         { "510%",          OPTION_VAL_510_UK },
         { "520%",          OPTION_VAL_520_UK },
         { "530%",          OPTION_VAL_530_UK },
         { "540%",          OPTION_VAL_540_UK },
         { "550%",          OPTION_VAL_550_UK },
         { "560%",          OPTION_VAL_560_UK },
         { "570%",          OPTION_VAL_570_UK },
         { "580%",          OPTION_VAL_580_UK },
         { "590%",          OPTION_VAL_590_UK },
         { "600%",          OPTION_VAL_600_UK },
         { "610%",          OPTION_VAL_610_UK },
         { "620%",          OPTION_VAL_620_UK },
         { "630%",          OPTION_VAL_630_UK },
         { "640%",          OPTION_VAL_640_UK },
         { "650%",          OPTION_VAL_650_UK },
         { "660%",          OPTION_VAL_660_UK },
         { "670%",          OPTION_VAL_670_UK },
         { "680%",          OPTION_VAL_680_UK },
         { "690%",          OPTION_VAL_690_UK },
         { "700%",          OPTION_VAL_700_UK },
         { "710%",          OPTION_VAL_710_UK },
         { "720%",          OPTION_VAL_720_UK },
         { "730%",          OPTION_VAL_730_UK },
         { "740%",          OPTION_VAL_740_UK },
         { "750%",          OPTION_VAL_750_UK },
         { NULL, NULL },
      },
      "100%"
   },
   {
      BEETLE_OPT(gpu_overclock),
      BEETLE_OPT_GPU_OVERCLOCK_LABEL_UK,
      NULL,
      BEETLE_OPT_GPU_OVERCLOCK_INFO_0_UK,
      NULL,
      "system",
      {
         { "1x(native)", OPTION_VAL_1X_NATIVE_UK },
         { "2x",         OPTION_VAL_2X_UK },
         { "4x",         OPTION_VAL_4X_UK },
         { "8x",         OPTION_VAL_8X_UK },
         { "16x",        OPTION_VAL_16X_UK },
         { "32x",        OPTION_VAL_32X_UK },
         { NULL, NULL },
      },
      "1x(native)"
   },
   {
      BEETLE_OPT(gte_overclock),
      BEETLE_OPT_GTE_OVERCLOCK_LABEL_UK,
      NULL,
      BEETLE_OPT_GTE_OVERCLOCK_INFO_0_UK,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
#ifndef EMSCRIPTEN
   {
      BEETLE_OPT(cd_access_method),
      BEETLE_OPT_CD_ACCESS_METHOD_LABEL_UK,
      NULL,
      BEETLE_OPT_CD_ACCESS_METHOD_INFO_0_UK,
      NULL,
      "system",
      {
         { "sync",     OPTION_VAL_SYNC_UK },
         { "async",    OPTION_VAL_ASYNC_UK },
         { "precache", OPTION_VAL_PRECACHE_UK },
         { NULL, NULL },
      },
      "sync"
   },
#endif
   {
      BEETLE_OPT(cd_fastload),
      BEETLE_OPT_CD_FASTLOAD_LABEL_UK,
      NULL,
      BEETLE_OPT_CD_FASTLOAD_INFO_0_UK,
      NULL,
      "system",
      {
         { "2x(native)", OPTION_VAL_2X_NATIVE_UK },
         { "4x",          OPTION_VAL_4X_UK },
         { "6x",          OPTION_VAL_6X_UK },
         { "8x",          OPTION_VAL_8X_UK },
         { "10x",         OPTION_VAL_10X_UK },
         { "12x",         OPTION_VAL_12X_UK },
         { "14x",         OPTION_VAL_14X_UK },
         { NULL, NULL },
      },
      "2x(native)"
   },
   {
      BEETLE_OPT(skip_bios),
      BEETLE_OPT_SKIP_BIOS_LABEL_UK,
      NULL,
      BEETLE_OPT_SKIP_BIOS_INFO_0_UK,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(override_bios),
      BEETLE_OPT_OVERRIDE_BIOS_LABEL_UK,
      NULL,
      BEETLE_OPT_OVERRIDE_BIOS_INFO_0_UK,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "psxonpsp", OPTION_VAL_PSXONPSP_UK },
         { "ps1_rom",  OPTION_VAL_PS1_ROM_UK },
         { "openbios", OPTION_VAL_OPENBIOS_UK     },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(region),
      BEETLE_OPT_REGION_LABEL_UK,
      NULL,
      BEETLE_OPT_REGION_INFO_0_UK,
      NULL,
      "system",
      {
         { "auto",   OPTION_VAL_AUTO_UK },
         { "ntsc-j", OPTION_VAL_NTSC_J_UK },
         { "ntsc-u", OPTION_VAL_NTSC_U_UK },
         { "pal",    OPTION_VAL_PAL_UK },
         { NULL, NULL },
      },
      "auto"
   },
   {
      BEETLE_OPT(internal_resolution),
      BEETLE_OPT_INTERNAL_RESOLUTION_LABEL_UK,
      NULL,
      BEETLE_OPT_INTERNAL_RESOLUTION_INFO_0_UK,
      NULL,
      "video",
      {
         { "1x(native)", OPTION_VAL_1X_NATIVE_UK },
         { "2x",         OPTION_VAL_2X_UK },
         { "4x",         OPTION_VAL_4X_UK },
         { "8x",         OPTION_VAL_8X_UK },
         { "16x",        OPTION_VAL_16X_UK },
         { NULL, NULL },
      },
      "1x(native)"
   },
#if defined(HAVE_OPENGL) || defined(HAVE_OPENGLES) || defined(HAVE_VULKAN)
   {
      BEETLE_OPT(renderer),
      BEETLE_OPT_RENDERER_LABEL_UK,
      NULL,
      BEETLE_OPT_RENDERER_INFO_0_UK,
      NULL,
      "video",
      {
         { "hardware",    OPTION_VAL_HARDWARE_UK },
#if defined(HAVE_OPENGL) || defined(HAVE_OPENGLES)
         { "hardware_gl", OPTION_VAL_HARDWARE_GL_UK },
#endif
#if defined(HAVE_VULKAN)
         { "hardware_vk", OPTION_VAL_HARDWARE_VK_UK },
#endif
         { "software",    OPTION_VAL_SOFTWARE_UK },
         { NULL, NULL },
      },
      "hardware"
   },
   {
      BEETLE_OPT(renderer_software_fb),
      BEETLE_OPT_RENDERER_SOFTWARE_FB_LABEL_UK,
      NULL,
      BEETLE_OPT_RENDERER_SOFTWARE_FB_INFO_0_UK,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled"
   },
#endif
#if defined(HAVE_OPENGL) || defined(HAVE_OPENGLES)
   {
      BEETLE_OPT(depth),
      BEETLE_OPT_DEPTH_LABEL_UK,
      NULL,
      BEETLE_OPT_DEPTH_INFO_0_UK,
      NULL,
      "video",
      {
         { "16bpp(native)", OPTION_VAL_16BPP_NATIVE_UK },
         { "32bpp",         OPTION_VAL_32BPP_UK },
         { NULL, NULL },
      },
      "16bpp(native)"
   },
   // Sort of, it's more like 15-bit high color and 24-bit true color for visible output. The alpha channel is used for mask bit. Vulkan renderer uses ABGR1555_555 for 31 bits internal? FMVs are always 24-bit on all renderers like original hardware (BGR888, no alpha)
#endif
   {
      BEETLE_OPT(dither_mode),
      BEETLE_OPT_DITHER_MODE_LABEL_UK,
      NULL,
      BEETLE_OPT_DITHER_MODE_INFO_0_UK,
      NULL,
      "video",
      {
         { "1x(native)",          OPTION_VAL_1X_NATIVE_UK },
         { "internal resolution", OPTION_VAL_INTERNAL_RESOLUTION_UK },
         { "disabled",            NULL },
         { NULL, NULL },
      },
      "1x(native)"
   },
#ifdef HAVE_VULKAN
   {
      BEETLE_OPT(scaled_uv_offset),
      BEETLE_OPT_SCALED_UV_OFFSET_LABEL_UK,
      NULL,
      BEETLE_OPT_SCALED_UV_OFFSET_INFO_0_UK,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled"
   },
#endif
#if defined(HAVE_OPENGL) || defined(HAVE_OPENGLES) || defined(HAVE_VULKAN)
   {
      BEETLE_OPT(filter),
      BEETLE_OPT_FILTER_LABEL_UK,
      NULL,
      BEETLE_OPT_FILTER_INFO_0_UK,
      NULL,
      "video",
      {
         { "nearest",  OPTION_VAL_NEAREST_UK },
         { "SABR",     OPTION_VAL_SABR_UK },
         { "xBR",      OPTION_VAL_XBR_UK },
         { "bilinear", OPTION_VAL_BILINEAR_UK },
         { "3-point",  OPTION_VAL_3_POINT_UK },
         { "JINC2",    OPTION_VAL_JINC2_UK },
         { NULL, NULL },
      },
      "nearest"
   },
#ifdef HAVE_VULKAN
   {
      BEETLE_OPT(filter_exclude_sprite),
      BEETLE_OPT_FILTER_EXCLUDE_SPRITE_LABEL_UK,
      NULL,
      BEETLE_OPT_FILTER_EXCLUDE_SPRITE_INFO_0_UK,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "opaque", OPTION_VAL_OPAQUE_UK },
         { "all", OPTION_VAL_ALL_UK },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(filter_exclude_2d_polygon),
      BEETLE_OPT_FILTER_EXCLUDE_2D_POLYGON_LABEL_UK,
      NULL,
      BEETLE_OPT_FILTER_EXCLUDE_2D_POLYGON_INFO_0_UK,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "opaque", OPTION_VAL_OPAQUE_UK },
         { "all", OPTION_VAL_ALL_UK },
         { NULL, NULL },
      },
      "disabled"
   },
#endif
#endif
#ifdef HAVE_VULKAN
   {
      BEETLE_OPT(adaptive_smoothing),
      BEETLE_OPT_ADAPTIVE_SMOOTHING_LABEL_UK,
      NULL,
      BEETLE_OPT_ADAPTIVE_SMOOTHING_INFO_0_UK,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(super_sampling),
      BEETLE_OPT_SUPER_SAMPLING_LABEL_UK,
      NULL,
      BEETLE_OPT_SUPER_SAMPLING_INFO_0_UK,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(msaa),
      BEETLE_OPT_MSAA_LABEL_UK,
      NULL,
      BEETLE_OPT_MSAA_INFO_0_UK,
      NULL,
      "video",
      {
         { "1x",  OPTION_VAL_1X_UK },
         { "2x",  OPTION_VAL_2X_UK },
         { "4x",  OPTION_VAL_4X_UK },
         { "8x",  OPTION_VAL_8X_UK },
         { "16x", OPTION_VAL_16X_UK },
         { NULL, NULL },
      },
      "1x"
   },
   {
      BEETLE_OPT(mdec_yuv),
      BEETLE_OPT_MDEC_YUV_LABEL_UK,
      NULL,
      BEETLE_OPT_MDEC_YUV_INFO_0_UK,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(track_textures),
      BEETLE_OPT_TRACK_TEXTURES_LABEL_UK,
      NULL,
      BEETLE_OPT_TRACK_TEXTURES_INFO_0_UK,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
#ifdef TEXTURE_DUMPING_ENABLED
   {
      BEETLE_OPT(dump_textures),
      BEETLE_OPT_DUMP_TEXTURES_LABEL_UK,
      NULL,
      BEETLE_OPT_DUMP_TEXTURES_INFO_0_UK,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
#endif
   {
      BEETLE_OPT(replace_textures),
      BEETLE_OPT_REPLACE_TEXTURES_LABEL_UK,
      NULL,
      BEETLE_OPT_REPLACE_TEXTURES_INFO_0_UK,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
#endif
   {
      BEETLE_OPT(pgxp_mode),
      BEETLE_OPT_PGXP_MODE_LABEL_UK,
      NULL,
      BEETLE_OPT_PGXP_MODE_INFO_0_UK,
      NULL,
      "pgxp",
      {
         { "disabled",     NULL },
         { "memory only",  OPTION_VAL_MEMORY_ONLY_UK },
         { "memory + CPU", OPTION_VAL_MEMORY_CPU_UK },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(pgxp_2d_tol),
      BEETLE_OPT_PGXP_2D_TOL_LABEL_UK,
      NULL,
      BEETLE_OPT_PGXP_2D_TOL_INFO_0_UK,
      NULL,
      "pgxp",
      {
         { "disabled", NULL },
         { "0px", OPTION_VAL_0PX_UK },
         { "1px", OPTION_VAL_1PX_UK },
         { "2px", OPTION_VAL_2PX_UK },
         { "3px", OPTION_VAL_3PX_UK },
         { "4px", OPTION_VAL_4PX_UK },
         { "5px", OPTION_VAL_5PX_UK },
         { "6px", OPTION_VAL_6PX_UK },
         { "7px", OPTION_VAL_7PX_UK },
         { "8px", OPTION_VAL_8PX_UK },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(pgxp_nclip),
      BEETLE_OPT_PGXP_NCLIP_LABEL_UK,
      NULL,
      BEETLE_OPT_PGXP_NCLIP_INFO_0_UK,
      NULL,
      "pgxp",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
#if defined(HAVE_OPENGL) || defined(HAVE_OPENGLES) || defined(HAVE_VULKAN)
   {
      BEETLE_OPT(pgxp_vertex),
      BEETLE_OPT_PGXP_VERTEX_LABEL_UK,
      NULL,
      BEETLE_OPT_PGXP_VERTEX_INFO_0_UK,
      NULL,
      "pgxp",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
#endif
   {
      BEETLE_OPT(pgxp_texture),
      BEETLE_OPT_PGXP_TEXTURE_LABEL_UK,
      NULL,
      BEETLE_OPT_PGXP_TEXTURE_INFO_0_UK,
      NULL,
      "pgxp",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(display_internal_fps),
      BEETLE_OPT_DISPLAY_INTERNAL_FPS_LABEL_UK,
      NULL,
      BEETLE_OPT_DISPLAY_INTERNAL_FPS_INFO_0_UK,
      NULL,
      "osd",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(display_osd),
      BEETLE_OPT_DISPLAY_OSD_LABEL_UK,
      NULL,
      BEETLE_OPT_DISPLAY_OSD_INFO_0_UK,
      NULL,
      "osd",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled"
   },
#if defined(HAVE_OPENGL) || defined(HAVE_OPENGLES) || defined(HAVE_VULKAN)
   {
      BEETLE_OPT(display_vram),
      BEETLE_OPT_DISPLAY_VRAM_LABEL_UK,
      NULL,
      BEETLE_OPT_DISPLAY_VRAM_INFO_0_UK,
      NULL,
      "osd",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
#endif
   {
      BEETLE_OPT(analog_calibration),
      BEETLE_OPT_ANALOG_CALIBRATION_LABEL_UK,
      NULL,
      BEETLE_OPT_ANALOG_CALIBRATION_INFO_0_UK,
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(analog_toggle),
      BEETLE_OPT_ANALOG_TOGGLE_LABEL_UK,
      NULL,
      BEETLE_OPT_ANALOG_TOGGLE_INFO_0_UK,
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { "enabled-analog", OPTION_VAL_ENABLED_ANALOG_UK },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(analog_toggle_combo),
      BEETLE_OPT_ANALOG_TOGGLE_COMBO_LABEL_UK,
      NULL,
      BEETLE_OPT_ANALOG_TOGGLE_COMBO_INFO_0_UK,
      NULL,
      "input",
      {
         { "l1+l2+r1+r2+start+select", OPTION_VAL_L1_L2_R1_R2_START_SELECT_UK },
         { "l1+r1+select",             OPTION_VAL_L1_R1_SELECT_UK },
         { "l1+r1+start",              OPTION_VAL_L1_R1_START_UK },
         { "l1+r1+l3",                 OPTION_VAL_L1_R1_L3_UK },
         { "l1+r1+r3",                 OPTION_VAL_L1_R1_R3_UK },
         { "l2+r2+select",             OPTION_VAL_L2_R2_SELECT_UK },
         { "l2+r2+start",              OPTION_VAL_L2_R2_START_UK },
         { "l2+r2+l3",                 OPTION_VAL_L2_R2_L3_UK },
         { "l2+r2+r3",                 OPTION_VAL_L2_R2_R3_UK },
         { "l3+r3",                    OPTION_VAL_L3_R3_UK },
         { NULL, NULL },
      },
      "l1+r1+select"
   },
   {
      BEETLE_OPT(analog_toggle_hold),
      BEETLE_OPT_ANALOG_TOGGLE_HOLD_LABEL_UK,
      NULL,
      BEETLE_OPT_ANALOG_TOGGLE_HOLD_INFO_0_UK,
      NULL,
      "input",
      {
         { "0", OPTION_VAL_0_UK },
         { "1", OPTION_VAL_1_UK },
         { "2", OPTION_VAL_2_UK },
         { "3", OPTION_VAL_3_UK },
         { "4", OPTION_VAL_4_UK },
         { "5", OPTION_VAL_5_UK },
         { NULL, NULL },
      },
      "1"
   },
   {
      BEETLE_OPT(enable_multitap_port1),
      BEETLE_OPT_ENABLE_MULTITAP_PORT1_LABEL_UK,
      NULL,
      BEETLE_OPT_ENABLE_MULTITAP_PORT1_INFO_0_UK,
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(enable_multitap_port2),
      BEETLE_OPT_ENABLE_MULTITAP_PORT2_LABEL_UK,
      NULL,
      BEETLE_OPT_ENABLE_MULTITAP_PORT2_INFO_0_UK,
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(gun_input_mode),
      BEETLE_OPT_GUN_INPUT_MODE_LABEL_UK,
      NULL,
      BEETLE_OPT_GUN_INPUT_MODE_INFO_0_UK,
      NULL,
      "input",
      {
         { "lightgun",    OPTION_VAL_LIGHTGUN_UK },
         { "touchscreen", OPTION_VAL_TOUCHSCREEN_UK },
         { NULL, NULL },
      },
      "lightgun"
   },
   // Shouldn't the gun_input_mode just be Mouse vs. Touchscreen?
   {
      BEETLE_OPT(gun_cursor),
      BEETLE_OPT_GUN_CURSOR_LABEL_UK,
      NULL,
      BEETLE_OPT_GUN_CURSOR_INFO_0_UK,
      NULL,
      "input",
      {
         { "off",   "disabled" },
         { "cross", OPTION_VAL_CROSS_UK },
         { "dot",   OPTION_VAL_DOT_UK },
         { NULL, NULL },
      },
      "cross"
   },
   {
      BEETLE_OPT(crosshair_color_p1),
      BEETLE_OPT_CROSSHAIR_COLOR_P1_LABEL_UK,
      NULL,
      BEETLE_OPT_CROSSHAIR_COLOR_P1_INFO_0_UK,
      NULL,
      "input",
      {
         { "red",    OPTION_VAL_RED_UK },
         { "blue",   OPTION_VAL_BLUE_UK },
         { "green",  OPTION_VAL_GREEN_UK },
         { "orange", OPTION_VAL_ORANGE_UK },
         { "yellow", OPTION_VAL_YELLOW_UK },
         { "cyan",   OPTION_VAL_CYAN_UK },
         { "pink",   OPTION_VAL_PINK_UK },
         { "purple", OPTION_VAL_PURPLE_UK },
         { "black",  OPTION_VAL_BLACK_UK },
         { "white",  OPTION_VAL_WHITE_UK },
         { NULL, NULL },
      },
      "red"
   },
   {
      BEETLE_OPT(crosshair_color_p2),
      BEETLE_OPT_CROSSHAIR_COLOR_P2_LABEL_UK,
      NULL,
      BEETLE_OPT_CROSSHAIR_COLOR_P2_INFO_0_UK,
      NULL,
      "input",
      {
         { "blue",   OPTION_VAL_BLUE_UK },
         { "red",    OPTION_VAL_RED_UK },
         { "green",  OPTION_VAL_GREEN_UK },
         { "orange", OPTION_VAL_ORANGE_UK },
         { "yellow", OPTION_VAL_YELLOW_UK },
         { "cyan",   OPTION_VAL_CYAN_UK },
         { "pink",   OPTION_VAL_PINK_UK },
         { "purple", OPTION_VAL_PURPLE_UK },
         { "black",  OPTION_VAL_BLACK_UK },
         { "white",  OPTION_VAL_WHITE_UK },
         { NULL, NULL },
      },
      "blue"
   },
   {
      BEETLE_OPT(mouse_sensitivity),
      BEETLE_OPT_MOUSE_SENSITIVITY_LABEL_UK,
      NULL,
      BEETLE_OPT_MOUSE_SENSITIVITY_INFO_0_UK,
      NULL,
      "input",
      {
         { "5%",   OPTION_VAL_5_O42_UK },
         { "10%",  OPTION_VAL_10_UK },
         { "15%",  OPTION_VAL_15_UK },
         { "20%",  OPTION_VAL_20_UK },
         { "25%",  OPTION_VAL_25_UK },
         { "30%",  OPTION_VAL_30_UK },
         { "35%",  OPTION_VAL_35_UK },
         { "40%",  OPTION_VAL_40_UK },
         { "45%",  OPTION_VAL_45_UK },
         { "50%",  OPTION_VAL_50_UK },
         { "55%",  OPTION_VAL_55_UK },
         { "60%",  OPTION_VAL_60_UK },
         { "65%",  OPTION_VAL_65_UK },
         { "70%",  OPTION_VAL_70_UK },
         { "75%",  OPTION_VAL_75_UK },
         { "80%",  OPTION_VAL_80_UK },
         { "85%",  OPTION_VAL_85_UK },
         { "90%",  OPTION_VAL_90_UK },
         { "95%",  OPTION_VAL_95_UK },
         { "100%", OPTION_VAL_100_O42_UK },
         { "105%", OPTION_VAL_105_UK },
         { "110%", OPTION_VAL_110_UK },
         { "115%", OPTION_VAL_115_UK },
         { "120%", OPTION_VAL_120_UK },
         { "125%", OPTION_VAL_125_UK },
         { "130%", OPTION_VAL_130_UK },
         { "135%", OPTION_VAL_135_UK },
         { "140%", OPTION_VAL_140_UK },
         { "145%", OPTION_VAL_145_UK },
         { "150%", OPTION_VAL_150_UK },
         { "155%", OPTION_VAL_155_UK },
         { "160%", OPTION_VAL_160_UK },
         { "165%", OPTION_VAL_165_UK },
         { "170%", OPTION_VAL_170_UK },
         { "175%", OPTION_VAL_175_UK },
         { "180%", OPTION_VAL_180_UK },
         { "185%", OPTION_VAL_185_UK },
         { "190%", OPTION_VAL_190_UK },
         { "195%", OPTION_VAL_195_UK },
         { "200%", OPTION_VAL_200_UK },
         { NULL, NULL },
      },
      "100%"
   },
   {
      BEETLE_OPT(negcon_response),
      BEETLE_OPT_NEGCON_RESPONSE_LABEL_UK,
      NULL,
      BEETLE_OPT_NEGCON_RESPONSE_INFO_0_UK,
      NULL,
      "input",
      {
         { "linear",    OPTION_VAL_LINEAR_UK },
         { "quadratic", OPTION_VAL_QUADRATIC_UK },
         { "cubic",     OPTION_VAL_CUBIC_UK },
         { NULL, NULL },
      },
      "linear"
   },
   {
      BEETLE_OPT(negcon_deadzone),
      BEETLE_OPT_NEGCON_DEADZONE_LABEL_UK,
      NULL,
      BEETLE_OPT_NEGCON_DEADZONE_INFO_0_UK,
      NULL,
      "input",
      {
         { "0%",  OPTION_VAL_0_O44_UK },
         { "5%",  OPTION_VAL_5_O42_UK },
         { "10%", OPTION_VAL_10_UK },
         { "15%", OPTION_VAL_15_UK },
         { "20%", OPTION_VAL_20_UK },
         { "25%", OPTION_VAL_25_UK },
         { "30%", OPTION_VAL_30_UK },
         { NULL, NULL },
      },
      "0%"
   },
   {
      BEETLE_OPT(use_mednafen_memcard0_method),
      BEETLE_OPT_USE_MEDNAFEN_MEMCARD0_METHOD_LABEL_UK,
      NULL,
      BEETLE_OPT_USE_MEDNAFEN_MEMCARD0_METHOD_INFO_0_UK,
      NULL,
      "memcards",
      {
         { "libretro", OPTION_VAL_LIBRETRO_UK },
         { "mednafen", OPTION_VAL_MEDNAFEN_UK },
         { NULL, NULL },
      },
      "libretro"
   },
   {
      BEETLE_OPT(enable_memcard1),
      BEETLE_OPT_ENABLE_MEMCARD1_LABEL_UK,
      NULL,
      BEETLE_OPT_ENABLE_MEMCARD1_INFO_0_UK,
      NULL,
      "memcards",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(shared_memory_cards),
      BEETLE_OPT_SHARED_MEMORY_CARDS_LABEL_UK,
      NULL,
      BEETLE_OPT_SHARED_MEMORY_CARDS_INFO_0_UK,
      NULL,
      "memcards",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(memcard_left_index),
      BEETLE_OPT_MEMCARD_LEFT_INDEX_LABEL_UK,
      NULL,
      BEETLE_OPT_MEMCARD_LEFT_INDEX_INFO_0_UK,
      NULL,
      "memcards",
      {
         { "0",  OPTION_VAL_0_O48_UK },
         { "1",  NULL },
         { "2",  NULL },
         { "3",  NULL },
         { "4",  NULL },
         { "5",  NULL },
         { "6",  NULL },
         { "7",  NULL },
         { "8",  NULL },
         { "9",  NULL },
         { "10", NULL },
         { "11", NULL },
         { "12", NULL },
         { "13", NULL },
         { "14", NULL },
         { "15", NULL },
         { "16", NULL },
         { "17", NULL },
         { "18", NULL },
         { "19", NULL },
         { "20", NULL },
         { "21", NULL },
         { "22", NULL },
         { "23", NULL },
         { "24", NULL },
         { "25", NULL },
         { "26", NULL },
         { "27", NULL },
         { "28", NULL },
         { "29", NULL },
         { "30", NULL },
         { "31", NULL },
         { "32", NULL },
         { "33", NULL },
         { "34", NULL },
         { "35", NULL },
         { "36", NULL },
         { "37", NULL },
         { "38", NULL },
         { "39", NULL },
         { "40", NULL },
         { "41", NULL },
         { "42", NULL },
         { "43", NULL },
         { "44", NULL },
         { "45", NULL },
         { "46", NULL },
         { "47", NULL },
         { "48", NULL },
         { "49", NULL },
         { "50", NULL },
         { "51", NULL },
         { "52", NULL },
         { "53", NULL },
         { "54", NULL },
         { "55", NULL },
         { "56", NULL },
         { "57", NULL },
         { "58", NULL },
         { "59", NULL },
         { "60", NULL },
         { "61", NULL },
         { "62", NULL },
         { "63", NULL },
         { NULL, NULL },
      },
      "0"
   },
   {
      BEETLE_OPT(memcard_right_index),
      BEETLE_OPT_MEMCARD_RIGHT_INDEX_LABEL_UK,
      NULL,
      BEETLE_OPT_MEMCARD_RIGHT_INDEX_INFO_0_UK,
      NULL,
      "memcards",
      {
         { "0",  NULL },
         { "1",  OPTION_VAL_1_O49_UK },
         { "2",  NULL },
         { "3",  NULL },
         { "4",  NULL },
         { "5",  NULL },
         { "6",  NULL },
         { "7",  NULL },
         { "8",  NULL },
         { "9",  NULL },
         { "10", NULL },
         { "11", NULL },
         { "12", NULL },
         { "13", NULL },
         { "14", NULL },
         { "15", NULL },
         { "16", NULL },
         { "17", NULL },
         { "18", NULL },
         { "19", NULL },
         { "20", NULL },
         { "21", NULL },
         { "22", NULL },
         { "23", NULL },
         { "24", NULL },
         { "25", NULL },
         { "26", NULL },
         { "27", NULL },
         { "28", NULL },
         { "29", NULL },
         { "30", NULL },
         { "31", NULL },
         { "32", NULL },
         { "33", NULL },
         { "34", NULL },
         { "35", NULL },
         { "36", NULL },
         { "37", NULL },
         { "38", NULL },
         { "39", NULL },
         { "40", NULL },
         { "41", NULL },
         { "42", NULL },
         { "43", NULL },
         { "44", NULL },
         { "45", NULL },
         { "46", NULL },
         { "47", NULL },
         { "48", NULL },
         { "49", NULL },
         { "50", NULL },
         { "51", NULL },
         { "52", NULL },
         { "53", NULL },
         { "54", NULL },
         { "55", NULL },
         { "56", NULL },
         { "57", NULL },
         { "58", NULL },
         { "59", NULL },
         { "60", NULL },
         { "61", NULL },
         { "62", NULL },
         { "63", NULL },
         { NULL, NULL },
      },
      "1"
   },
   {
      BEETLE_OPT(line_render),
      BEETLE_OPT_LINE_RENDER_LABEL_UK,
      NULL,
      BEETLE_OPT_LINE_RENDER_INFO_0_UK,
      NULL,
      "hacks",
      {
         { "default",    OPTION_VAL_DEFAULT_UK },
         { "aggressive", OPTION_VAL_AGGRESSIVE_UK },
         { "disabled",   NULL },
         { NULL, NULL },
      },
      "default"
   },
   {
      BEETLE_OPT(widescreen_hack),
      BEETLE_OPT_WIDESCREEN_HACK_LABEL_UK,
      NULL,
      BEETLE_OPT_WIDESCREEN_HACK_INFO_0_UK,
      NULL,
      "hacks",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(widescreen_hack_aspect_ratio),
      BEETLE_OPT_WIDESCREEN_HACK_ASPECT_RATIO_LABEL_UK,
      NULL,
      BEETLE_OPT_WIDESCREEN_HACK_ASPECT_RATIO_INFO_0_UK,
      NULL,
      "hacks",
      {
         { "16:9",  OPTION_VAL_16_9_UK },
         { "16:10", OPTION_VAL_16_10_UK },
         { "18:9",  OPTION_VAL_18_9_UK },
         { "19:9",  OPTION_VAL_19_9_UK },
         { "20:9",  OPTION_VAL_20_9_UK },
         { "21:9",  OPTION_VAL_21_9_UK }, // 64:27
         { "32:9",  OPTION_VAL_32_9_UK },
         { NULL,    NULL },
      },
      "16:9"
   },
#if defined(HAVE_LIGHTREC)
   {
      BEETLE_OPT(cpu_dynarec),
      BEETLE_OPT_CPU_DYNAREC_LABEL_UK,
      NULL,
      BEETLE_OPT_CPU_DYNAREC_INFO_0_UK,
      NULL,
      "hacks",
      {
         { "disabled",        OPTION_VAL_DISABLED_UK },
         { "execute",         OPTION_VAL_EXECUTE_UK },
         { "execute_one",     OPTION_VAL_EXECUTE_ONE_UK },
         { "run_interpreter", OPTION_VAL_RUN_INTERPRETER_UK },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(dynarec_invalidate),
      BEETLE_OPT_DYNAREC_INVALIDATE_LABEL_UK,
      NULL,
      BEETLE_OPT_DYNAREC_INVALIDATE_INFO_0_UK,
      NULL,
      "hacks",
      {
         { "full", OPTION_VAL_FULL_UK },
         { "dma",  OPTION_VAL_DMA_UK },
         { NULL, NULL },
      },
      "full"
   },
   {
      BEETLE_OPT(dynarec_eventcycles),
      BEETLE_OPT_DYNAREC_EVENTCYCLES_LABEL_UK,
      NULL,
      BEETLE_OPT_DYNAREC_EVENTCYCLES_INFO_0_UK,
      NULL,
      "hacks",
      {
         { "128",  OPTION_VAL_128_UK },
         { "256",  NULL },
         { "384",  NULL },
         { "512",  NULL },
         { "640",  NULL },
         { "768",  NULL },
         { "896",  NULL },
         { "1024", NULL },
         { "1152", NULL },
         { "1280", NULL },
         { "1408", NULL },
         { "1536", NULL },
         { "1664", NULL },
         { "1792", NULL },
         { "1920", NULL },
         { "2048", NULL },
         { NULL, NULL },
      },
      "128"
   },
   {
      BEETLE_OPT(dynarec_spu_samples),
      BEETLE_OPT_DYNAREC_SPU_SAMPLES_LABEL_UK,
      NULL,
      BEETLE_OPT_DYNAREC_SPU_SAMPLES_INFO_0_UK,
      NULL,
      "hacks",
      {
         { "1",  OPTION_VAL_1_O49_UK },
         { "4",  NULL },
         { "16", NULL },
         { NULL, NULL },
      },
      "1"
   },
#endif
   {
      BEETLE_OPT(core_timing_fps),
      BEETLE_OPT_CORE_TIMING_FPS_LABEL_UK,
      NULL,
      BEETLE_OPT_CORE_TIMING_FPS_INFO_0_UK,
      NULL,
      "video",
      {
         { "auto_toggle",       OPTION_VAL_AUTO_TOGGLE_UK },
         { "force_progressive", OPTION_VAL_FORCE_PROGRESSIVE_UK },
         { "force_interlaced",  OPTION_VAL_FORCE_INTERLACED_UK },
      },
      "auto_toggle"
   },
   {
      BEETLE_OPT(deinterlacer),
      BEETLE_OPT_DEINTERLACER_LABEL_UK,
      NULL,
      BEETLE_OPT_DEINTERLACER_INFO_0_UK,
      NULL,
      "video",
      {
         { "weave",      OPTION_VAL_WEAVE_UK },
         { "bob",        OPTION_VAL_BOB_UK },
         { "bob_offset", OPTION_VAL_BOB_OFFSET_UK },
         { "fastmad",    OPTION_VAL_FASTMAD_UK },
         { "off",        "Off" },
         { NULL, NULL },
      },
      "weave"
   },
   {
      BEETLE_OPT(frame_duping),
      BEETLE_OPT_FRAME_DUPING_LABEL_UK,
      NULL,
      BEETLE_OPT_FRAME_DUPING_INFO_0_UK,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(pal_video_timing_override),
      BEETLE_OPT_PAL_VIDEO_TIMING_OVERRIDE_LABEL_UK,
      NULL,
      BEETLE_OPT_PAL_VIDEO_TIMING_OVERRIDE_INFO_0_UK,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(aspect_ratio),
      BEETLE_OPT_ASPECT_RATIO_LABEL_UK,
      NULL,
      BEETLE_OPT_ASPECT_RATIO_INFO_0_UK,
      NULL,
      "video",
      {
         { "corrected", OPTION_VAL_CORRECTED_UK },
         { "uncorrected", OPTION_VAL_UNCORRECTED_UK },
         { "4:3",  OPTION_VAL_4_3_UK },
         { "ntsc", OPTION_VAL_NTSC_UK },
      },
      "corrected"
   },
   {
      BEETLE_OPT(crop_overscan),
      BEETLE_OPT_CROP_OVERSCAN_LABEL_UK,
      NULL,
      BEETLE_OPT_CROP_OVERSCAN_INFO_0_UK,
      NULL,
      "video",
      {
         { "disabled", "Disabled" },
         { "static",   OPTION_VAL_STATIC_UK },
         { "smart",    OPTION_VAL_SMART_UK },
         { NULL, NULL },
      },
      "smart"
   },
   {
      BEETLE_OPT(image_crop),
      BEETLE_OPT_IMAGE_CROP_LABEL_UK,
      NULL,
      BEETLE_OPT_IMAGE_CROP_INFO_0_UK,
      NULL,
      "video",
      {
         { "disabled", "0" },
         { "1px",      OPTION_VAL_1PX_UK },
         { "2px",      OPTION_VAL_2PX_UK },
         { "3px",      OPTION_VAL_3PX_UK },
         { "4px",      OPTION_VAL_4PX_UK },
         { "5px",      OPTION_VAL_5PX_UK },
         { "6px",      OPTION_VAL_6PX_UK },
         { "7px",      OPTION_VAL_7PX_UK },
         { "8px",      OPTION_VAL_8PX_UK },
         { "9px",      OPTION_VAL_9PX_UK },
         { "10px",     OPTION_VAL_10PX_UK },
         { "11px",     OPTION_VAL_11PX_UK },
         { "12px",     OPTION_VAL_12PX_UK },
         { "13px",     OPTION_VAL_13PX_UK },
         { "14px",     OPTION_VAL_14PX_UK },
         { "15px",     OPTION_VAL_15PX_UK },
         { "16px",     OPTION_VAL_16PX_UK },
         { "17px",     OPTION_VAL_17PX_UK },
         { "18px",     OPTION_VAL_18PX_UK },
         { "19px",     OPTION_VAL_19PX_UK },
         { "20px",     OPTION_VAL_20PX_UK },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(image_offset),
      BEETLE_OPT_IMAGE_OFFSET_LABEL_UK,
      NULL,
      BEETLE_OPT_IMAGE_OFFSET_INFO_0_UK,
      NULL,
      "video",
      {
         { "-12px",    OPTION_VAL_12PX_O64_UK },
         { "-11px",    OPTION_VAL_11PX_O64_UK },
         { "-10px",    OPTION_VAL_10PX_O64_UK },
         { "-9px",     OPTION_VAL_9PX_O64_UK },
         { "-8px",     OPTION_VAL_8PX_O64_UK },
         { "-7px",     OPTION_VAL_7PX_O64_UK },
         { "-6px",     OPTION_VAL_6PX_O64_UK },
         { "-5px",     OPTION_VAL_5PX_O64_UK },
         { "-4px",     OPTION_VAL_4PX_O64_UK },
         { "-3px",     OPTION_VAL_3PX_O64_UK },
         { "-2px",     OPTION_VAL_2PX_O64_UK },
         { "-1px",     OPTION_VAL_1PX_O64_UK },
         { "disabled", "0" },
         { "+1px",     OPTION_VAL_1PX_O64_0_UK },
         { "+2px",     OPTION_VAL_2PX_O64_0_UK },
         { "+3px",     OPTION_VAL_3PX_O64_0_UK },
         { "+4px",     OPTION_VAL_4PX_O64_0_UK },
         { "+5px",     OPTION_VAL_5PX_O64_0_UK },
         { "+6px",     OPTION_VAL_6PX_O64_0_UK },
         { "+7px",     OPTION_VAL_7PX_O64_0_UK },
         { "+8px",     OPTION_VAL_8PX_O64_0_UK },
         { "+9px",     OPTION_VAL_9PX_O64_0_UK },
         { "+10px",    OPTION_VAL_10PX_O64_0_UK },
         { "+11px",    OPTION_VAL_11PX_O64_0_UK },
         { "+12px",    OPTION_VAL_12PX_O64_0_UK },
         { NULL, NULL },
      },
      "disabled"
   },
#if defined(HAVE_OPENGL) || defined(HAVE_OPENGLES) || defined(HAVE_VULKAN)
   {
      BEETLE_OPT(image_offset_cycles),
      BEETLE_OPT_IMAGE_OFFSET_CYCLES_LABEL_UK,
      NULL,
      BEETLE_OPT_IMAGE_OFFSET_CYCLES_INFO_0_UK,
      NULL,
      "video",
      {
         { "-40",      NULL },
         { "-39",      NULL },
         { "-38",      NULL },
         { "-37",      NULL },
         { "-36",      NULL },
         { "-35",      NULL },
         { "-34",      NULL },
         { "-33",      NULL },
         { "-32",      NULL },
         { "-31",      NULL },
         { "-30",      NULL },
         { "-29",      NULL },
         { "-28",      NULL },
         { "-27",      NULL },
         { "-26",      NULL },
         { "-25",      NULL },
         { "-24",      NULL },
         { "-23",      NULL },
         { "-22",      NULL },
         { "-21",      NULL },
         { "-20",      NULL },
         { "-19",      NULL },
         { "-18",      NULL },
         { "-17",      NULL },
         { "-16",      NULL },
         { "-15",      NULL },
         { "-14",      NULL },
         { "-13",      NULL },
         { "-12",      NULL },
         { "-11",      NULL },
         { "-10",      NULL },
         { "-9",       NULL },
         { "-8",       NULL },
         { "-7",       NULL },
         { "-6",       NULL },
         { "-5",       NULL },
         { "-4",       NULL },
         { "-3",       NULL },
         { "-2",       NULL },
         { "-1",       NULL },
         { "0",        "0" },
         { "+1",       NULL },
         { "+2",       NULL },
         { "+3",       NULL },
         { "+4",       NULL },
         { "+5",       NULL },
         { "+6",       NULL },
         { "+7",       NULL },
         { "+8",       NULL },
         { "+9",       NULL },
         { "+10",      NULL },
         { "+11",      NULL },
         { "+12",      NULL },
         { "+13",      NULL },
         { "+14",      NULL },
         { "+15",      NULL },
         { "+16",      NULL },
         { "+17",      NULL },
         { "+18",      NULL },
         { "+19",      NULL },
         { "+20",      NULL },
         { "+21",      NULL },
         { "+22",      NULL },
         { "+23",      NULL },
         { "+24",      NULL },
         { "+25",      NULL },
         { "+26",      NULL },
         { "+27",      NULL },
         { "+28",      NULL },
         { "+29",      NULL },
         { "+30",      NULL },
         { "+31",      NULL },
         { "+32",      NULL },
         { "+33",      NULL },
         { "+34",      NULL },
         { "+35",      NULL },
         { "+36",      NULL },
         { "+37",      NULL },
         { "+38",      NULL },
         { "+39",      NULL },
         { "+40",      NULL },
         { NULL, NULL},
      },
      "0"
   },
#endif
   {
      BEETLE_OPT(initial_scanline),
      BEETLE_OPT_INITIAL_SCANLINE_LABEL_UK,
      NULL,
      BEETLE_OPT_INITIAL_SCANLINE_INFO_0_UK,
      NULL,
      "video",
      {
         { "0",  OPTION_VAL_0_O48_UK },
         { "1",  NULL },
         { "2",  NULL },
         { "3",  NULL },
         { "4",  NULL },
         { "5",  NULL },
         { "6",  NULL },
         { "7",  NULL },
         { "8",  NULL },
         { "9",  NULL },
         { "10", NULL },
         { "11", NULL },
         { "12", NULL },
         { "13", NULL },
         { "14", NULL },
         { "15", NULL },
         { "16", NULL },
         { "17", NULL },
         { "18", NULL },
         { "19", NULL },
         { "20", NULL },
         { "21", NULL },
         { "22", NULL },
         { "23", NULL },
         { "24", NULL },
         { "25", NULL },
         { "26", NULL },
         { "27", NULL },
         { "28", NULL },
         { "29", NULL },
         { "30", NULL },
         { "31", NULL },
         { "32", NULL },
         { "33", NULL },
         { "34", NULL },
         { "35", NULL },
         { "36", NULL },
         { "37", NULL },
         { "38", NULL },
         { "39", NULL },
         { "40", NULL },
         { NULL, NULL },
      },
      "0"
   },
   {
      BEETLE_OPT(last_scanline),
      BEETLE_OPT_LAST_SCANLINE_LABEL_UK,
      NULL,
      BEETLE_OPT_LAST_SCANLINE_INFO_0_UK,
      NULL,
      "video",
      {
         { "210", NULL },
         { "211", NULL },
         { "212", NULL },
         { "213", NULL },
         { "214", NULL },
         { "215", NULL },
         { "216", NULL },
         { "217", NULL },
         { "218", NULL },
         { "219", NULL },
         { "220", NULL },
         { "221", NULL },
         { "222", NULL },
         { "223", NULL },
         { "224", NULL },
         { "225", NULL },
         { "226", NULL },
         { "227", NULL },
         { "228", NULL },
         { "229", NULL },
         { "230", NULL },
         { "231", NULL },
         { "232", NULL },
         { "233", NULL },
         { "234", NULL },
         { "235", NULL },
         { "236", NULL },
         { "237", NULL },
         { "238", NULL },
         { "239", OPTION_VAL_239_UK },
         { NULL, NULL },
      },
      "239"
   },
   {
      BEETLE_OPT(initial_scanline_pal),
      BEETLE_OPT_INITIAL_SCANLINE_PAL_LABEL_UK,
      NULL,
      BEETLE_OPT_INITIAL_SCANLINE_PAL_INFO_0_UK,
      NULL,
      "video",
      {
         { "0",  OPTION_VAL_0_O48_UK },
         { "1",  NULL },
         { "2",  NULL },
         { "3",  NULL },
         { "4",  NULL },
         { "5",  NULL },
         { "6",  NULL },
         { "7",  NULL },
         { "8",  NULL },
         { "9",  NULL },
         { "10", NULL },
         { "11", NULL },
         { "12", NULL },
         { "13", NULL },
         { "14", NULL },
         { "15", NULL },
         { "16", NULL },
         { "17", NULL },
         { "18", NULL },
         { "19", NULL },
         { "20", NULL },
         { "21", NULL },
         { "22", NULL },
         { "23", NULL },
         { "24", NULL },
         { "25", NULL },
         { "26", NULL },
         { "27", NULL },
         { "28", NULL },
         { "29", NULL },
         { "30", NULL },
         { "31", NULL },
         { "32", NULL },
         { "33", NULL },
         { "34", NULL },
         { "35", NULL },
         { "36", NULL },
         { "37", NULL },
         { "38", NULL },
         { "39", NULL },
         { "40", NULL },
         { NULL, NULL },
      },
      "0"
   },
   {
      BEETLE_OPT(last_scanline_pal),
      BEETLE_OPT_LAST_SCANLINE_PAL_LABEL_UK,
      NULL,
      BEETLE_OPT_LAST_SCANLINE_PAL_INFO_0_UK,
      NULL,
      "video",
      {
         { "230", NULL },
         { "231", NULL },
         { "232", NULL },
         { "233", NULL },
         { "234", NULL },
         { "235", NULL },
         { "236", NULL },
         { "237", NULL },
         { "238", NULL },
         { "239", NULL },
         { "240", NULL },
         { "241", NULL },
         { "242", NULL },
         { "243", NULL },
         { "244", NULL },
         { "245", NULL },
         { "246", NULL },
         { "247", NULL },
         { "248", NULL },
         { "249", NULL },
         { "250", NULL },
         { "251", NULL },
         { "252", NULL },
         { "253", NULL },
         { "254", NULL },
         { "255", NULL },
         { "256", NULL },
         { "257", NULL },
         { "258", NULL },
         { "259", NULL },
         { "260", NULL },
         { "261", NULL },
         { "262", NULL },
         { "263", NULL },
         { "264", NULL },
         { "265", NULL },
         { "266", NULL },
         { "267", NULL },
         { "268", NULL },
         { "269", NULL },
         { "270", NULL },
         { "271", NULL },
         { "272", NULL },
         { "273", NULL },
         { "274", NULL },
         { "275", NULL },
         { "276", NULL },
         { "277", NULL },
         { "278", NULL },
         { "279", NULL },
         { "280", NULL },
         { "281", NULL },
         { "282", NULL },
         { "283", NULL },
         { "284", NULL },
         { "285", NULL },
         { "286", NULL },
         { "287", OPTION_VAL_287_UK },
         { NULL, NULL },
      },
      "287"
   },
   { NULL, NULL, NULL, NULL, NULL, NULL, {{0}}, NULL },
};
struct retro_core_options_v2 options_uk = {
   option_cats_uk,
   option_defs_uk
};

/* RETRO_LANGUAGE_VAL */

#define CATEGORY_SYSTEM_LABEL_VAL "Sistema"
#define CATEGORY_SYSTEM_INFO_0_VAL NULL
#define CATEGORY_VIDEO_LABEL_VAL "Vídeo"
#define CATEGORY_VIDEO_INFO_0_VAL NULL
#define CATEGORY_PGXP_LABEL_VAL NULL
#define CATEGORY_PGXP_INFO_0_VAL NULL
#define CATEGORY_OSD_LABEL_VAL "Visualitzacions a la pantalla"
#define CATEGORY_OSD_INFO_0_VAL NULL
#define CATEGORY_INPUT_LABEL_VAL "Entrada"
#define CATEGORY_INPUT_INFO_0_VAL NULL
#define CATEGORY_MEMCARDS_LABEL_VAL NULL
#define CATEGORY_MEMCARDS_INFO_0_VAL NULL
#define CATEGORY_HACKS_LABEL_VAL NULL
#define CATEGORY_HACKS_INFO_0_VAL NULL
#define BEETLE_OPT_CPU_FREQ_SCALE_LABEL_VAL NULL
#define BEETLE_OPT_CPU_FREQ_SCALE_INFO_0_VAL NULL
#define OPTION_VAL_50_VAL NULL
#define OPTION_VAL_60_VAL NULL
#define OPTION_VAL_70_VAL NULL
#define OPTION_VAL_80_VAL NULL
#define OPTION_VAL_90_VAL NULL
#define OPTION_VAL_100_VAL NULL
#define OPTION_VAL_110_VAL NULL
#define OPTION_VAL_120_VAL NULL
#define OPTION_VAL_130_VAL NULL
#define OPTION_VAL_140_VAL NULL
#define OPTION_VAL_150_VAL NULL
#define OPTION_VAL_160_VAL NULL
#define OPTION_VAL_170_VAL NULL
#define OPTION_VAL_180_VAL NULL
#define OPTION_VAL_190_VAL NULL
#define OPTION_VAL_200_VAL NULL
#define OPTION_VAL_210_VAL NULL
#define OPTION_VAL_220_VAL NULL
#define OPTION_VAL_230_VAL NULL
#define OPTION_VAL_240_VAL NULL
#define OPTION_VAL_250_VAL NULL
#define OPTION_VAL_260_VAL NULL
#define OPTION_VAL_270_VAL NULL
#define OPTION_VAL_280_VAL NULL
#define OPTION_VAL_290_VAL NULL
#define OPTION_VAL_300_VAL NULL
#define OPTION_VAL_310_VAL NULL
#define OPTION_VAL_320_VAL NULL
#define OPTION_VAL_330_VAL NULL
#define OPTION_VAL_340_VAL NULL
#define OPTION_VAL_350_VAL NULL
#define OPTION_VAL_360_VAL NULL
#define OPTION_VAL_370_VAL NULL
#define OPTION_VAL_380_VAL NULL
#define OPTION_VAL_390_VAL NULL
#define OPTION_VAL_400_VAL NULL
#define OPTION_VAL_410_VAL NULL
#define OPTION_VAL_420_VAL NULL
#define OPTION_VAL_430_VAL NULL
#define OPTION_VAL_440_VAL NULL
#define OPTION_VAL_450_VAL NULL
#define OPTION_VAL_460_VAL NULL
#define OPTION_VAL_470_VAL NULL
#define OPTION_VAL_480_VAL NULL
#define OPTION_VAL_490_VAL NULL
#define OPTION_VAL_500_VAL NULL
#define OPTION_VAL_510_VAL NULL
#define OPTION_VAL_520_VAL NULL
#define OPTION_VAL_530_VAL NULL
#define OPTION_VAL_540_VAL NULL
#define OPTION_VAL_550_VAL NULL
#define OPTION_VAL_560_VAL NULL
#define OPTION_VAL_570_VAL NULL
#define OPTION_VAL_580_VAL NULL
#define OPTION_VAL_590_VAL NULL
#define OPTION_VAL_600_VAL NULL
#define OPTION_VAL_610_VAL NULL
#define OPTION_VAL_620_VAL NULL
#define OPTION_VAL_630_VAL NULL
#define OPTION_VAL_640_VAL NULL
#define OPTION_VAL_650_VAL NULL
#define OPTION_VAL_660_VAL NULL
#define OPTION_VAL_670_VAL NULL
#define OPTION_VAL_680_VAL NULL
#define OPTION_VAL_690_VAL NULL
#define OPTION_VAL_700_VAL NULL
#define OPTION_VAL_710_VAL NULL
#define OPTION_VAL_720_VAL NULL
#define OPTION_VAL_730_VAL NULL
#define OPTION_VAL_740_VAL NULL
#define OPTION_VAL_750_VAL NULL
#define BEETLE_OPT_GPU_OVERCLOCK_LABEL_VAL NULL
#define BEETLE_OPT_GPU_OVERCLOCK_INFO_0_VAL NULL
#define OPTION_VAL_1X_NATIVE_VAL NULL
#define OPTION_VAL_2X_VAL NULL
#define OPTION_VAL_4X_VAL NULL
#define OPTION_VAL_8X_VAL NULL
#define OPTION_VAL_16X_VAL NULL
#define OPTION_VAL_32X_VAL NULL
#define BEETLE_OPT_GTE_OVERCLOCK_LABEL_VAL NULL
#define BEETLE_OPT_GTE_OVERCLOCK_INFO_0_VAL NULL
#define BEETLE_OPT_CD_ACCESS_METHOD_LABEL_VAL NULL
#define BEETLE_OPT_CD_ACCESS_METHOD_INFO_0_VAL NULL
#define OPTION_VAL_SYNC_VAL NULL
#define OPTION_VAL_ASYNC_VAL NULL
#define OPTION_VAL_PRECACHE_VAL NULL
#define BEETLE_OPT_CD_FASTLOAD_LABEL_VAL NULL
#define BEETLE_OPT_CD_FASTLOAD_INFO_0_VAL NULL
#define OPTION_VAL_2X_NATIVE_VAL NULL
#define OPTION_VAL_6X_VAL NULL
#define OPTION_VAL_10X_VAL NULL
#define OPTION_VAL_12X_VAL NULL
#define OPTION_VAL_14X_VAL NULL
#define BEETLE_OPT_SKIP_BIOS_LABEL_VAL NULL
#define BEETLE_OPT_SKIP_BIOS_INFO_0_VAL NULL
#define BEETLE_OPT_OVERRIDE_BIOS_LABEL_VAL NULL
#define BEETLE_OPT_OVERRIDE_BIOS_INFO_0_VAL NULL
#define OPTION_VAL_PSXONPSP_VAL NULL
#define OPTION_VAL_PS1_ROM_VAL NULL
#define OPTION_VAL_OPENBIOS_VAL NULL
#define BEETLE_OPT_REGION_LABEL_VAL NULL
#define BEETLE_OPT_REGION_INFO_0_VAL NULL
#define OPTION_VAL_AUTO_VAL "Selecció automàtica"
#define OPTION_VAL_NTSC_J_VAL NULL
#define OPTION_VAL_NTSC_U_VAL NULL
#define OPTION_VAL_PAL_VAL NULL
#define BEETLE_OPT_INTERNAL_RESOLUTION_LABEL_VAL NULL
#define BEETLE_OPT_INTERNAL_RESOLUTION_INFO_0_VAL NULL
#define BEETLE_OPT_RENDERER_LABEL_VAL NULL
#define BEETLE_OPT_RENDERER_INFO_0_VAL NULL
#define OPTION_VAL_HARDWARE_VAL NULL
#define OPTION_VAL_HARDWARE_GL_VAL NULL
#define OPTION_VAL_HARDWARE_VK_VAL NULL
#define OPTION_VAL_SOFTWARE_VAL NULL
#define BEETLE_OPT_RENDERER_SOFTWARE_FB_LABEL_VAL NULL
#define BEETLE_OPT_RENDERER_SOFTWARE_FB_INFO_0_VAL NULL
#define BEETLE_OPT_DEPTH_LABEL_VAL NULL
#define BEETLE_OPT_DEPTH_INFO_0_VAL NULL
#define OPTION_VAL_16BPP_NATIVE_VAL NULL
#define OPTION_VAL_32BPP_VAL NULL
#define BEETLE_OPT_DITHER_MODE_LABEL_VAL NULL
#define BEETLE_OPT_DITHER_MODE_INFO_0_VAL NULL
#define OPTION_VAL_INTERNAL_RESOLUTION_VAL NULL
#define BEETLE_OPT_SCALED_UV_OFFSET_LABEL_VAL NULL
#define BEETLE_OPT_SCALED_UV_OFFSET_INFO_0_VAL NULL
#define BEETLE_OPT_FILTER_LABEL_VAL NULL
#define BEETLE_OPT_FILTER_INFO_0_VAL NULL
#define OPTION_VAL_NEAREST_VAL NULL
#define OPTION_VAL_SABR_VAL NULL
#define OPTION_VAL_XBR_VAL NULL
#define OPTION_VAL_BILINEAR_VAL "Bilineal"
#define OPTION_VAL_3_POINT_VAL NULL
#define OPTION_VAL_JINC2_VAL NULL
#define BEETLE_OPT_FILTER_EXCLUDE_SPRITE_LABEL_VAL NULL
#define BEETLE_OPT_FILTER_EXCLUDE_SPRITE_INFO_0_VAL NULL
#define OPTION_VAL_OPAQUE_VAL NULL
#define OPTION_VAL_ALL_VAL NULL
#define BEETLE_OPT_FILTER_EXCLUDE_2D_POLYGON_LABEL_VAL NULL
#define BEETLE_OPT_FILTER_EXCLUDE_2D_POLYGON_INFO_0_VAL NULL
#define BEETLE_OPT_ADAPTIVE_SMOOTHING_LABEL_VAL NULL
#define BEETLE_OPT_ADAPTIVE_SMOOTHING_INFO_0_VAL NULL
#define BEETLE_OPT_SUPER_SAMPLING_LABEL_VAL NULL
#define BEETLE_OPT_SUPER_SAMPLING_INFO_0_VAL NULL
#define BEETLE_OPT_MSAA_LABEL_VAL NULL
#define BEETLE_OPT_MSAA_INFO_0_VAL NULL
#define OPTION_VAL_1X_VAL NULL
#define BEETLE_OPT_MDEC_YUV_LABEL_VAL NULL
#define BEETLE_OPT_MDEC_YUV_INFO_0_VAL NULL
#define BEETLE_OPT_TRACK_TEXTURES_LABEL_VAL NULL
#define BEETLE_OPT_TRACK_TEXTURES_INFO_0_VAL NULL
#define BEETLE_OPT_DUMP_TEXTURES_LABEL_VAL NULL
#define BEETLE_OPT_DUMP_TEXTURES_INFO_0_VAL NULL
#define BEETLE_OPT_REPLACE_TEXTURES_LABEL_VAL NULL
#define BEETLE_OPT_REPLACE_TEXTURES_INFO_0_VAL NULL
#define BEETLE_OPT_PGXP_MODE_LABEL_VAL NULL
#define BEETLE_OPT_PGXP_MODE_INFO_0_VAL NULL
#define OPTION_VAL_MEMORY_ONLY_VAL NULL
#define OPTION_VAL_MEMORY_CPU_VAL NULL
#define BEETLE_OPT_PGXP_2D_TOL_LABEL_VAL NULL
#define BEETLE_OPT_PGXP_2D_TOL_INFO_0_VAL NULL
#define OPTION_VAL_0PX_VAL NULL
#define OPTION_VAL_1PX_VAL NULL
#define OPTION_VAL_2PX_VAL NULL
#define OPTION_VAL_3PX_VAL NULL
#define OPTION_VAL_4PX_VAL NULL
#define OPTION_VAL_5PX_VAL NULL
#define OPTION_VAL_6PX_VAL NULL
#define OPTION_VAL_7PX_VAL NULL
#define OPTION_VAL_8PX_VAL NULL
#define BEETLE_OPT_PGXP_NCLIP_LABEL_VAL NULL
#define BEETLE_OPT_PGXP_NCLIP_INFO_0_VAL NULL
#define BEETLE_OPT_PGXP_VERTEX_LABEL_VAL NULL
#define BEETLE_OPT_PGXP_VERTEX_INFO_0_VAL NULL
#define BEETLE_OPT_PGXP_TEXTURE_LABEL_VAL NULL
#define BEETLE_OPT_PGXP_TEXTURE_INFO_0_VAL NULL
#define BEETLE_OPT_DISPLAY_INTERNAL_FPS_LABEL_VAL NULL
#define BEETLE_OPT_DISPLAY_INTERNAL_FPS_INFO_0_VAL NULL
#define BEETLE_OPT_DISPLAY_OSD_LABEL_VAL NULL
#define BEETLE_OPT_DISPLAY_OSD_INFO_0_VAL NULL
#define BEETLE_OPT_DISPLAY_VRAM_LABEL_VAL NULL
#define BEETLE_OPT_DISPLAY_VRAM_INFO_0_VAL NULL
#define BEETLE_OPT_ANALOG_CALIBRATION_LABEL_VAL NULL
#define BEETLE_OPT_ANALOG_CALIBRATION_INFO_0_VAL NULL
#define BEETLE_OPT_ANALOG_TOGGLE_LABEL_VAL NULL
#define BEETLE_OPT_ANALOG_TOGGLE_INFO_0_VAL NULL
#define OPTION_VAL_ENABLED_ANALOG_VAL NULL
#define BEETLE_OPT_ANALOG_TOGGLE_COMBO_LABEL_VAL NULL
#define BEETLE_OPT_ANALOG_TOGGLE_COMBO_INFO_0_VAL NULL
#define OPTION_VAL_L1_L2_R1_R2_START_SELECT_VAL NULL
#define OPTION_VAL_L1_R1_SELECT_VAL NULL
#define OPTION_VAL_L1_R1_START_VAL NULL
#define OPTION_VAL_L1_R1_L3_VAL NULL
#define OPTION_VAL_L1_R1_R3_VAL NULL
#define OPTION_VAL_L2_R2_SELECT_VAL NULL
#define OPTION_VAL_L2_R2_START_VAL NULL
#define OPTION_VAL_L2_R2_L3_VAL NULL
#define OPTION_VAL_L2_R2_R3_VAL NULL
#define OPTION_VAL_L3_R3_VAL NULL
#define BEETLE_OPT_ANALOG_TOGGLE_HOLD_LABEL_VAL NULL
#define BEETLE_OPT_ANALOG_TOGGLE_HOLD_INFO_0_VAL NULL
#define OPTION_VAL_0_VAL NULL
#define OPTION_VAL_1_VAL NULL
#define OPTION_VAL_2_VAL NULL
#define OPTION_VAL_3_VAL NULL
#define OPTION_VAL_4_VAL NULL
#define OPTION_VAL_5_VAL NULL
#define BEETLE_OPT_ENABLE_MULTITAP_PORT1_LABEL_VAL NULL
#define BEETLE_OPT_ENABLE_MULTITAP_PORT1_INFO_0_VAL NULL
#define BEETLE_OPT_ENABLE_MULTITAP_PORT2_LABEL_VAL NULL
#define BEETLE_OPT_ENABLE_MULTITAP_PORT2_INFO_0_VAL NULL
#define BEETLE_OPT_GUN_INPUT_MODE_LABEL_VAL NULL
#define BEETLE_OPT_GUN_INPUT_MODE_INFO_0_VAL NULL
#define OPTION_VAL_LIGHTGUN_VAL NULL
#define OPTION_VAL_TOUCHSCREEN_VAL NULL
#define BEETLE_OPT_GUN_CURSOR_LABEL_VAL NULL
#define BEETLE_OPT_GUN_CURSOR_INFO_0_VAL NULL
#define OPTION_VAL_CROSS_VAL NULL
#define OPTION_VAL_DOT_VAL NULL
#define BEETLE_OPT_CROSSHAIR_COLOR_P1_LABEL_VAL NULL
#define BEETLE_OPT_CROSSHAIR_COLOR_P1_INFO_0_VAL NULL
#define OPTION_VAL_RED_VAL NULL
#define OPTION_VAL_BLUE_VAL NULL
#define OPTION_VAL_GREEN_VAL NULL
#define OPTION_VAL_ORANGE_VAL NULL
#define OPTION_VAL_YELLOW_VAL NULL
#define OPTION_VAL_CYAN_VAL NULL
#define OPTION_VAL_PINK_VAL NULL
#define OPTION_VAL_PURPLE_VAL NULL
#define OPTION_VAL_BLACK_VAL NULL
#define OPTION_VAL_WHITE_VAL NULL
#define BEETLE_OPT_CROSSHAIR_COLOR_P2_LABEL_VAL NULL
#define BEETLE_OPT_CROSSHAIR_COLOR_P2_INFO_0_VAL NULL
#define BEETLE_OPT_MOUSE_SENSITIVITY_LABEL_VAL NULL
#define BEETLE_OPT_MOUSE_SENSITIVITY_INFO_0_VAL NULL
#define OPTION_VAL_5_O42_VAL NULL
#define OPTION_VAL_10_VAL NULL
#define OPTION_VAL_15_VAL NULL
#define OPTION_VAL_20_VAL NULL
#define OPTION_VAL_25_VAL NULL
#define OPTION_VAL_30_VAL NULL
#define OPTION_VAL_35_VAL NULL
#define OPTION_VAL_40_VAL NULL
#define OPTION_VAL_45_VAL NULL
#define OPTION_VAL_55_VAL NULL
#define OPTION_VAL_65_VAL NULL
#define OPTION_VAL_75_VAL NULL
#define OPTION_VAL_85_VAL NULL
#define OPTION_VAL_95_VAL NULL
#define OPTION_VAL_100_O42_VAL NULL
#define OPTION_VAL_105_VAL NULL
#define OPTION_VAL_115_VAL NULL
#define OPTION_VAL_125_VAL NULL
#define OPTION_VAL_135_VAL NULL
#define OPTION_VAL_145_VAL NULL
#define OPTION_VAL_155_VAL NULL
#define OPTION_VAL_165_VAL NULL
#define OPTION_VAL_175_VAL NULL
#define OPTION_VAL_185_VAL NULL
#define OPTION_VAL_195_VAL NULL
#define BEETLE_OPT_NEGCON_RESPONSE_LABEL_VAL NULL
#define BEETLE_OPT_NEGCON_RESPONSE_INFO_0_VAL NULL
#define OPTION_VAL_LINEAR_VAL NULL
#define OPTION_VAL_QUADRATIC_VAL NULL
#define OPTION_VAL_CUBIC_VAL NULL
#define BEETLE_OPT_NEGCON_DEADZONE_LABEL_VAL NULL
#define BEETLE_OPT_NEGCON_DEADZONE_INFO_0_VAL NULL
#define OPTION_VAL_0_O44_VAL NULL
#define BEETLE_OPT_USE_MEDNAFEN_MEMCARD0_METHOD_LABEL_VAL NULL
#define BEETLE_OPT_USE_MEDNAFEN_MEMCARD0_METHOD_INFO_0_VAL NULL
#define OPTION_VAL_LIBRETRO_VAL NULL
#define OPTION_VAL_MEDNAFEN_VAL NULL
#define BEETLE_OPT_ENABLE_MEMCARD1_LABEL_VAL NULL
#define BEETLE_OPT_ENABLE_MEMCARD1_INFO_0_VAL NULL
#define BEETLE_OPT_SHARED_MEMORY_CARDS_LABEL_VAL NULL
#define BEETLE_OPT_SHARED_MEMORY_CARDS_INFO_0_VAL NULL
#define BEETLE_OPT_MEMCARD_LEFT_INDEX_LABEL_VAL NULL
#define BEETLE_OPT_MEMCARD_LEFT_INDEX_INFO_0_VAL NULL
#define OPTION_VAL_0_O48_VAL NULL
#define BEETLE_OPT_MEMCARD_RIGHT_INDEX_LABEL_VAL NULL
#define BEETLE_OPT_MEMCARD_RIGHT_INDEX_INFO_0_VAL NULL
#define OPTION_VAL_1_O49_VAL NULL
#define BEETLE_OPT_LINE_RENDER_LABEL_VAL NULL
#define BEETLE_OPT_LINE_RENDER_INFO_0_VAL NULL
#define OPTION_VAL_DEFAULT_VAL NULL
#define OPTION_VAL_AGGRESSIVE_VAL NULL
#define BEETLE_OPT_WIDESCREEN_HACK_LABEL_VAL NULL
#define BEETLE_OPT_WIDESCREEN_HACK_INFO_0_VAL NULL
#define BEETLE_OPT_WIDESCREEN_HACK_ASPECT_RATIO_LABEL_VAL NULL
#define BEETLE_OPT_WIDESCREEN_HACK_ASPECT_RATIO_INFO_0_VAL NULL
#define OPTION_VAL_16_9_VAL NULL
#define OPTION_VAL_16_10_VAL NULL
#define OPTION_VAL_18_9_VAL NULL
#define OPTION_VAL_19_9_VAL NULL
#define OPTION_VAL_20_9_VAL NULL
#define OPTION_VAL_21_9_VAL NULL
#define OPTION_VAL_32_9_VAL NULL
#define BEETLE_OPT_CPU_DYNAREC_LABEL_VAL NULL
#define BEETLE_OPT_CPU_DYNAREC_INFO_0_VAL NULL
#define OPTION_VAL_DISABLED_VAL NULL
#define OPTION_VAL_EXECUTE_VAL NULL
#define OPTION_VAL_EXECUTE_ONE_VAL NULL
#define OPTION_VAL_RUN_INTERPRETER_VAL NULL
#define BEETLE_OPT_DYNAREC_INVALIDATE_LABEL_VAL NULL
#define BEETLE_OPT_DYNAREC_INVALIDATE_INFO_0_VAL NULL
#define OPTION_VAL_FULL_VAL NULL
#define OPTION_VAL_DMA_VAL NULL
#define BEETLE_OPT_DYNAREC_EVENTCYCLES_LABEL_VAL NULL
#define BEETLE_OPT_DYNAREC_EVENTCYCLES_INFO_0_VAL NULL
#define OPTION_VAL_128_VAL NULL
#define BEETLE_OPT_DYNAREC_SPU_SAMPLES_LABEL_VAL NULL
#define BEETLE_OPT_DYNAREC_SPU_SAMPLES_INFO_0_VAL NULL
#define BEETLE_OPT_CORE_TIMING_FPS_LABEL_VAL NULL
#define BEETLE_OPT_CORE_TIMING_FPS_INFO_0_VAL NULL
#define OPTION_VAL_AUTO_TOGGLE_VAL NULL
#define OPTION_VAL_FORCE_PROGRESSIVE_VAL NULL
#define OPTION_VAL_FORCE_INTERLACED_VAL NULL
#define BEETLE_OPT_DEINTERLACER_LABEL_VAL NULL
#define BEETLE_OPT_DEINTERLACER_INFO_0_VAL NULL
#define OPTION_VAL_WEAVE_VAL NULL
#define OPTION_VAL_BOB_VAL NULL
#define OPTION_VAL_BOB_OFFSET_VAL NULL
#define OPTION_VAL_FASTMAD_VAL NULL
#define BEETLE_OPT_FRAME_DUPING_LABEL_VAL NULL
#define BEETLE_OPT_FRAME_DUPING_INFO_0_VAL NULL
#define BEETLE_OPT_PAL_VIDEO_TIMING_OVERRIDE_LABEL_VAL NULL
#define BEETLE_OPT_PAL_VIDEO_TIMING_OVERRIDE_INFO_0_VAL NULL
#define BEETLE_OPT_ASPECT_RATIO_LABEL_VAL NULL
#define BEETLE_OPT_ASPECT_RATIO_INFO_0_VAL NULL
#define OPTION_VAL_CORRECTED_VAL NULL
#define OPTION_VAL_UNCORRECTED_VAL NULL
#define OPTION_VAL_4_3_VAL NULL
#define OPTION_VAL_NTSC_VAL NULL
#define BEETLE_OPT_CROP_OVERSCAN_LABEL_VAL NULL
#define BEETLE_OPT_CROP_OVERSCAN_INFO_0_VAL NULL
#define OPTION_VAL_STATIC_VAL NULL
#define OPTION_VAL_SMART_VAL NULL
#define BEETLE_OPT_IMAGE_CROP_LABEL_VAL NULL
#define BEETLE_OPT_IMAGE_CROP_INFO_0_VAL NULL
#define OPTION_VAL_9PX_VAL NULL
#define OPTION_VAL_10PX_VAL NULL
#define OPTION_VAL_11PX_VAL NULL
#define OPTION_VAL_12PX_VAL NULL
#define OPTION_VAL_13PX_VAL NULL
#define OPTION_VAL_14PX_VAL NULL
#define OPTION_VAL_15PX_VAL NULL
#define OPTION_VAL_16PX_VAL NULL
#define OPTION_VAL_17PX_VAL NULL
#define OPTION_VAL_18PX_VAL NULL
#define OPTION_VAL_19PX_VAL NULL
#define OPTION_VAL_20PX_VAL NULL
#define BEETLE_OPT_IMAGE_OFFSET_LABEL_VAL NULL
#define BEETLE_OPT_IMAGE_OFFSET_INFO_0_VAL NULL
#define OPTION_VAL_12PX_O64_VAL NULL
#define OPTION_VAL_11PX_O64_VAL NULL
#define OPTION_VAL_10PX_O64_VAL NULL
#define OPTION_VAL_9PX_O64_VAL NULL
#define OPTION_VAL_8PX_O64_VAL NULL
#define OPTION_VAL_7PX_O64_VAL NULL
#define OPTION_VAL_6PX_O64_VAL NULL
#define OPTION_VAL_5PX_O64_VAL NULL
#define OPTION_VAL_4PX_O64_VAL NULL
#define OPTION_VAL_3PX_O64_VAL NULL
#define OPTION_VAL_2PX_O64_VAL NULL
#define OPTION_VAL_1PX_O64_VAL NULL
#define OPTION_VAL_1PX_O64_0_VAL NULL
#define OPTION_VAL_2PX_O64_0_VAL NULL
#define OPTION_VAL_3PX_O64_0_VAL NULL
#define OPTION_VAL_4PX_O64_0_VAL NULL
#define OPTION_VAL_5PX_O64_0_VAL NULL
#define OPTION_VAL_6PX_O64_0_VAL NULL
#define OPTION_VAL_7PX_O64_0_VAL NULL
#define OPTION_VAL_8PX_O64_0_VAL NULL
#define OPTION_VAL_9PX_O64_0_VAL NULL
#define OPTION_VAL_10PX_O64_0_VAL NULL
#define OPTION_VAL_11PX_O64_0_VAL NULL
#define OPTION_VAL_12PX_O64_0_VAL NULL
#define BEETLE_OPT_IMAGE_OFFSET_CYCLES_LABEL_VAL NULL
#define BEETLE_OPT_IMAGE_OFFSET_CYCLES_INFO_0_VAL NULL
#define BEETLE_OPT_INITIAL_SCANLINE_LABEL_VAL NULL
#define BEETLE_OPT_INITIAL_SCANLINE_INFO_0_VAL NULL
#define BEETLE_OPT_LAST_SCANLINE_LABEL_VAL NULL
#define BEETLE_OPT_LAST_SCANLINE_INFO_0_VAL NULL
#define OPTION_VAL_239_VAL NULL
#define BEETLE_OPT_INITIAL_SCANLINE_PAL_LABEL_VAL NULL
#define BEETLE_OPT_INITIAL_SCANLINE_PAL_INFO_0_VAL NULL
#define BEETLE_OPT_LAST_SCANLINE_PAL_LABEL_VAL NULL
#define BEETLE_OPT_LAST_SCANLINE_PAL_INFO_0_VAL NULL
#define OPTION_VAL_287_VAL NULL

struct retro_core_option_v2_category option_cats_val[] = {
   {
      "system",
      CATEGORY_SYSTEM_LABEL_VAL,
      CATEGORY_SYSTEM_INFO_0_VAL
   },
   {
      "video",
      CATEGORY_VIDEO_LABEL_VAL,
      CATEGORY_VIDEO_INFO_0_VAL
   },
   {
      "pgxp",
      CATEGORY_PGXP_LABEL_VAL,
      CATEGORY_PGXP_INFO_0_VAL
   },
   {
      "osd",
      CATEGORY_OSD_LABEL_VAL,
      CATEGORY_OSD_INFO_0_VAL
   },
   {
      "input",
      CATEGORY_INPUT_LABEL_VAL,
      CATEGORY_INPUT_INFO_0_VAL
   },
   {
      "memcards",
      CATEGORY_MEMCARDS_LABEL_VAL,
      CATEGORY_MEMCARDS_INFO_0_VAL
   },
   {
      "hacks",
      CATEGORY_HACKS_LABEL_VAL,
      CATEGORY_HACKS_INFO_0_VAL
   },
   { NULL, NULL, NULL },
};
struct retro_core_option_v2_definition option_defs_val[] = {
   {
      BEETLE_OPT(cpu_freq_scale),
      BEETLE_OPT_CPU_FREQ_SCALE_LABEL_VAL,
      NULL,
      BEETLE_OPT_CPU_FREQ_SCALE_INFO_0_VAL,
      NULL,
      "system",
      {
         { "50%",           OPTION_VAL_50_VAL },
         { "60%",           OPTION_VAL_60_VAL },
         { "70%",           OPTION_VAL_70_VAL },
         { "80%",           OPTION_VAL_80_VAL },
         { "90%",           OPTION_VAL_90_VAL },
         { "100%",          OPTION_VAL_100_VAL },
         { "110%",          OPTION_VAL_110_VAL },
         { "120%",          OPTION_VAL_120_VAL },
         { "130%",          OPTION_VAL_130_VAL },
         { "140%",          OPTION_VAL_140_VAL },
         { "150%",          OPTION_VAL_150_VAL },
         { "160%",          OPTION_VAL_160_VAL },
         { "170%",          OPTION_VAL_170_VAL },
         { "180%",          OPTION_VAL_180_VAL },
         { "190%",          OPTION_VAL_190_VAL },
         { "200%",          OPTION_VAL_200_VAL },
         { "210%",          OPTION_VAL_210_VAL },
         { "220%",          OPTION_VAL_220_VAL },
         { "230%",          OPTION_VAL_230_VAL },
         { "240%",          OPTION_VAL_240_VAL },
         { "250%",          OPTION_VAL_250_VAL },
         { "260%",          OPTION_VAL_260_VAL },
         { "270%",          OPTION_VAL_270_VAL },
         { "280%",          OPTION_VAL_280_VAL },
         { "290%",          OPTION_VAL_290_VAL },
         { "300%",          OPTION_VAL_300_VAL },
         { "310%",          OPTION_VAL_310_VAL },
         { "320%",          OPTION_VAL_320_VAL },
         { "330%",          OPTION_VAL_330_VAL },
         { "340%",          OPTION_VAL_340_VAL },
         { "350%",          OPTION_VAL_350_VAL },
         { "360%",          OPTION_VAL_360_VAL },
         { "370%",          OPTION_VAL_370_VAL },
         { "380%",          OPTION_VAL_380_VAL },
         { "390%",          OPTION_VAL_390_VAL },
         { "400%",          OPTION_VAL_400_VAL },
         { "410%",          OPTION_VAL_410_VAL },
         { "420%",          OPTION_VAL_420_VAL },
         { "430%",          OPTION_VAL_430_VAL },
         { "440%",          OPTION_VAL_440_VAL },
         { "450%",          OPTION_VAL_450_VAL },
         { "460%",          OPTION_VAL_460_VAL },
         { "470%",          OPTION_VAL_470_VAL },
         { "480%",          OPTION_VAL_480_VAL },
         { "490%",          OPTION_VAL_490_VAL },
         { "500%",          OPTION_VAL_500_VAL },
         { "510%",          OPTION_VAL_510_VAL },
         { "520%",          OPTION_VAL_520_VAL },
         { "530%",          OPTION_VAL_530_VAL },
         { "540%",          OPTION_VAL_540_VAL },
         { "550%",          OPTION_VAL_550_VAL },
         { "560%",          OPTION_VAL_560_VAL },
         { "570%",          OPTION_VAL_570_VAL },
         { "580%",          OPTION_VAL_580_VAL },
         { "590%",          OPTION_VAL_590_VAL },
         { "600%",          OPTION_VAL_600_VAL },
         { "610%",          OPTION_VAL_610_VAL },
         { "620%",          OPTION_VAL_620_VAL },
         { "630%",          OPTION_VAL_630_VAL },
         { "640%",          OPTION_VAL_640_VAL },
         { "650%",          OPTION_VAL_650_VAL },
         { "660%",          OPTION_VAL_660_VAL },
         { "670%",          OPTION_VAL_670_VAL },
         { "680%",          OPTION_VAL_680_VAL },
         { "690%",          OPTION_VAL_690_VAL },
         { "700%",          OPTION_VAL_700_VAL },
         { "710%",          OPTION_VAL_710_VAL },
         { "720%",          OPTION_VAL_720_VAL },
         { "730%",          OPTION_VAL_730_VAL },
         { "740%",          OPTION_VAL_740_VAL },
         { "750%",          OPTION_VAL_750_VAL },
         { NULL, NULL },
      },
      "100%"
   },
   {
      BEETLE_OPT(gpu_overclock),
      BEETLE_OPT_GPU_OVERCLOCK_LABEL_VAL,
      NULL,
      BEETLE_OPT_GPU_OVERCLOCK_INFO_0_VAL,
      NULL,
      "system",
      {
         { "1x(native)", OPTION_VAL_1X_NATIVE_VAL },
         { "2x",         OPTION_VAL_2X_VAL },
         { "4x",         OPTION_VAL_4X_VAL },
         { "8x",         OPTION_VAL_8X_VAL },
         { "16x",        OPTION_VAL_16X_VAL },
         { "32x",        OPTION_VAL_32X_VAL },
         { NULL, NULL },
      },
      "1x(native)"
   },
   {
      BEETLE_OPT(gte_overclock),
      BEETLE_OPT_GTE_OVERCLOCK_LABEL_VAL,
      NULL,
      BEETLE_OPT_GTE_OVERCLOCK_INFO_0_VAL,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
#ifndef EMSCRIPTEN
   {
      BEETLE_OPT(cd_access_method),
      BEETLE_OPT_CD_ACCESS_METHOD_LABEL_VAL,
      NULL,
      BEETLE_OPT_CD_ACCESS_METHOD_INFO_0_VAL,
      NULL,
      "system",
      {
         { "sync",     OPTION_VAL_SYNC_VAL },
         { "async",    OPTION_VAL_ASYNC_VAL },
         { "precache", OPTION_VAL_PRECACHE_VAL },
         { NULL, NULL },
      },
      "sync"
   },
#endif
   {
      BEETLE_OPT(cd_fastload),
      BEETLE_OPT_CD_FASTLOAD_LABEL_VAL,
      NULL,
      BEETLE_OPT_CD_FASTLOAD_INFO_0_VAL,
      NULL,
      "system",
      {
         { "2x(native)", OPTION_VAL_2X_NATIVE_VAL },
         { "4x",          OPTION_VAL_4X_VAL },
         { "6x",          OPTION_VAL_6X_VAL },
         { "8x",          OPTION_VAL_8X_VAL },
         { "10x",         OPTION_VAL_10X_VAL },
         { "12x",         OPTION_VAL_12X_VAL },
         { "14x",         OPTION_VAL_14X_VAL },
         { NULL, NULL },
      },
      "2x(native)"
   },
   {
      BEETLE_OPT(skip_bios),
      BEETLE_OPT_SKIP_BIOS_LABEL_VAL,
      NULL,
      BEETLE_OPT_SKIP_BIOS_INFO_0_VAL,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(override_bios),
      BEETLE_OPT_OVERRIDE_BIOS_LABEL_VAL,
      NULL,
      BEETLE_OPT_OVERRIDE_BIOS_INFO_0_VAL,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "psxonpsp", OPTION_VAL_PSXONPSP_VAL },
         { "ps1_rom",  OPTION_VAL_PS1_ROM_VAL },
         { "openbios", OPTION_VAL_OPENBIOS_VAL     },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(region),
      BEETLE_OPT_REGION_LABEL_VAL,
      NULL,
      BEETLE_OPT_REGION_INFO_0_VAL,
      NULL,
      "system",
      {
         { "auto",   OPTION_VAL_AUTO_VAL },
         { "ntsc-j", OPTION_VAL_NTSC_J_VAL },
         { "ntsc-u", OPTION_VAL_NTSC_U_VAL },
         { "pal",    OPTION_VAL_PAL_VAL },
         { NULL, NULL },
      },
      "auto"
   },
   {
      BEETLE_OPT(internal_resolution),
      BEETLE_OPT_INTERNAL_RESOLUTION_LABEL_VAL,
      NULL,
      BEETLE_OPT_INTERNAL_RESOLUTION_INFO_0_VAL,
      NULL,
      "video",
      {
         { "1x(native)", OPTION_VAL_1X_NATIVE_VAL },
         { "2x",         OPTION_VAL_2X_VAL },
         { "4x",         OPTION_VAL_4X_VAL },
         { "8x",         OPTION_VAL_8X_VAL },
         { "16x",        OPTION_VAL_16X_VAL },
         { NULL, NULL },
      },
      "1x(native)"
   },
#if defined(HAVE_OPENGL) || defined(HAVE_OPENGLES) || defined(HAVE_VULKAN)
   {
      BEETLE_OPT(renderer),
      BEETLE_OPT_RENDERER_LABEL_VAL,
      NULL,
      BEETLE_OPT_RENDERER_INFO_0_VAL,
      NULL,
      "video",
      {
         { "hardware",    OPTION_VAL_HARDWARE_VAL },
#if defined(HAVE_OPENGL) || defined(HAVE_OPENGLES)
         { "hardware_gl", OPTION_VAL_HARDWARE_GL_VAL },
#endif
#if defined(HAVE_VULKAN)
         { "hardware_vk", OPTION_VAL_HARDWARE_VK_VAL },
#endif
         { "software",    OPTION_VAL_SOFTWARE_VAL },
         { NULL, NULL },
      },
      "hardware"
   },
   {
      BEETLE_OPT(renderer_software_fb),
      BEETLE_OPT_RENDERER_SOFTWARE_FB_LABEL_VAL,
      NULL,
      BEETLE_OPT_RENDERER_SOFTWARE_FB_INFO_0_VAL,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled"
   },
#endif
#if defined(HAVE_OPENGL) || defined(HAVE_OPENGLES)
   {
      BEETLE_OPT(depth),
      BEETLE_OPT_DEPTH_LABEL_VAL,
      NULL,
      BEETLE_OPT_DEPTH_INFO_0_VAL,
      NULL,
      "video",
      {
         { "16bpp(native)", OPTION_VAL_16BPP_NATIVE_VAL },
         { "32bpp",         OPTION_VAL_32BPP_VAL },
         { NULL, NULL },
      },
      "16bpp(native)"
   },
   // Sort of, it's more like 15-bit high color and 24-bit true color for visible output. The alpha channel is used for mask bit. Vulkan renderer uses ABGR1555_555 for 31 bits internal? FMVs are always 24-bit on all renderers like original hardware (BGR888, no alpha)
#endif
   {
      BEETLE_OPT(dither_mode),
      BEETLE_OPT_DITHER_MODE_LABEL_VAL,
      NULL,
      BEETLE_OPT_DITHER_MODE_INFO_0_VAL,
      NULL,
      "video",
      {
         { "1x(native)",          OPTION_VAL_1X_NATIVE_VAL },
         { "internal resolution", OPTION_VAL_INTERNAL_RESOLUTION_VAL },
         { "disabled",            NULL },
         { NULL, NULL },
      },
      "1x(native)"
   },
#ifdef HAVE_VULKAN
   {
      BEETLE_OPT(scaled_uv_offset),
      BEETLE_OPT_SCALED_UV_OFFSET_LABEL_VAL,
      NULL,
      BEETLE_OPT_SCALED_UV_OFFSET_INFO_0_VAL,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled"
   },
#endif
#if defined(HAVE_OPENGL) || defined(HAVE_OPENGLES) || defined(HAVE_VULKAN)
   {
      BEETLE_OPT(filter),
      BEETLE_OPT_FILTER_LABEL_VAL,
      NULL,
      BEETLE_OPT_FILTER_INFO_0_VAL,
      NULL,
      "video",
      {
         { "nearest",  OPTION_VAL_NEAREST_VAL },
         { "SABR",     OPTION_VAL_SABR_VAL },
         { "xBR",      OPTION_VAL_XBR_VAL },
         { "bilinear", OPTION_VAL_BILINEAR_VAL },
         { "3-point",  OPTION_VAL_3_POINT_VAL },
         { "JINC2",    OPTION_VAL_JINC2_VAL },
         { NULL, NULL },
      },
      "nearest"
   },
#ifdef HAVE_VULKAN
   {
      BEETLE_OPT(filter_exclude_sprite),
      BEETLE_OPT_FILTER_EXCLUDE_SPRITE_LABEL_VAL,
      NULL,
      BEETLE_OPT_FILTER_EXCLUDE_SPRITE_INFO_0_VAL,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "opaque", OPTION_VAL_OPAQUE_VAL },
         { "all", OPTION_VAL_ALL_VAL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(filter_exclude_2d_polygon),
      BEETLE_OPT_FILTER_EXCLUDE_2D_POLYGON_LABEL_VAL,
      NULL,
      BEETLE_OPT_FILTER_EXCLUDE_2D_POLYGON_INFO_0_VAL,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "opaque", OPTION_VAL_OPAQUE_VAL },
         { "all", OPTION_VAL_ALL_VAL },
         { NULL, NULL },
      },
      "disabled"
   },
#endif
#endif
#ifdef HAVE_VULKAN
   {
      BEETLE_OPT(adaptive_smoothing),
      BEETLE_OPT_ADAPTIVE_SMOOTHING_LABEL_VAL,
      NULL,
      BEETLE_OPT_ADAPTIVE_SMOOTHING_INFO_0_VAL,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(super_sampling),
      BEETLE_OPT_SUPER_SAMPLING_LABEL_VAL,
      NULL,
      BEETLE_OPT_SUPER_SAMPLING_INFO_0_VAL,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(msaa),
      BEETLE_OPT_MSAA_LABEL_VAL,
      NULL,
      BEETLE_OPT_MSAA_INFO_0_VAL,
      NULL,
      "video",
      {
         { "1x",  OPTION_VAL_1X_VAL },
         { "2x",  OPTION_VAL_2X_VAL },
         { "4x",  OPTION_VAL_4X_VAL },
         { "8x",  OPTION_VAL_8X_VAL },
         { "16x", OPTION_VAL_16X_VAL },
         { NULL, NULL },
      },
      "1x"
   },
   {
      BEETLE_OPT(mdec_yuv),
      BEETLE_OPT_MDEC_YUV_LABEL_VAL,
      NULL,
      BEETLE_OPT_MDEC_YUV_INFO_0_VAL,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(track_textures),
      BEETLE_OPT_TRACK_TEXTURES_LABEL_VAL,
      NULL,
      BEETLE_OPT_TRACK_TEXTURES_INFO_0_VAL,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
#ifdef TEXTURE_DUMPING_ENABLED
   {
      BEETLE_OPT(dump_textures),
      BEETLE_OPT_DUMP_TEXTURES_LABEL_VAL,
      NULL,
      BEETLE_OPT_DUMP_TEXTURES_INFO_0_VAL,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
#endif
   {
      BEETLE_OPT(replace_textures),
      BEETLE_OPT_REPLACE_TEXTURES_LABEL_VAL,
      NULL,
      BEETLE_OPT_REPLACE_TEXTURES_INFO_0_VAL,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
#endif
   {
      BEETLE_OPT(pgxp_mode),
      BEETLE_OPT_PGXP_MODE_LABEL_VAL,
      NULL,
      BEETLE_OPT_PGXP_MODE_INFO_0_VAL,
      NULL,
      "pgxp",
      {
         { "disabled",     NULL },
         { "memory only",  OPTION_VAL_MEMORY_ONLY_VAL },
         { "memory + CPU", OPTION_VAL_MEMORY_CPU_VAL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(pgxp_2d_tol),
      BEETLE_OPT_PGXP_2D_TOL_LABEL_VAL,
      NULL,
      BEETLE_OPT_PGXP_2D_TOL_INFO_0_VAL,
      NULL,
      "pgxp",
      {
         { "disabled", NULL },
         { "0px", OPTION_VAL_0PX_VAL },
         { "1px", OPTION_VAL_1PX_VAL },
         { "2px", OPTION_VAL_2PX_VAL },
         { "3px", OPTION_VAL_3PX_VAL },
         { "4px", OPTION_VAL_4PX_VAL },
         { "5px", OPTION_VAL_5PX_VAL },
         { "6px", OPTION_VAL_6PX_VAL },
         { "7px", OPTION_VAL_7PX_VAL },
         { "8px", OPTION_VAL_8PX_VAL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(pgxp_nclip),
      BEETLE_OPT_PGXP_NCLIP_LABEL_VAL,
      NULL,
      BEETLE_OPT_PGXP_NCLIP_INFO_0_VAL,
      NULL,
      "pgxp",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
#if defined(HAVE_OPENGL) || defined(HAVE_OPENGLES) || defined(HAVE_VULKAN)
   {
      BEETLE_OPT(pgxp_vertex),
      BEETLE_OPT_PGXP_VERTEX_LABEL_VAL,
      NULL,
      BEETLE_OPT_PGXP_VERTEX_INFO_0_VAL,
      NULL,
      "pgxp",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
#endif
   {
      BEETLE_OPT(pgxp_texture),
      BEETLE_OPT_PGXP_TEXTURE_LABEL_VAL,
      NULL,
      BEETLE_OPT_PGXP_TEXTURE_INFO_0_VAL,
      NULL,
      "pgxp",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(display_internal_fps),
      BEETLE_OPT_DISPLAY_INTERNAL_FPS_LABEL_VAL,
      NULL,
      BEETLE_OPT_DISPLAY_INTERNAL_FPS_INFO_0_VAL,
      NULL,
      "osd",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(display_osd),
      BEETLE_OPT_DISPLAY_OSD_LABEL_VAL,
      NULL,
      BEETLE_OPT_DISPLAY_OSD_INFO_0_VAL,
      NULL,
      "osd",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled"
   },
#if defined(HAVE_OPENGL) || defined(HAVE_OPENGLES) || defined(HAVE_VULKAN)
   {
      BEETLE_OPT(display_vram),
      BEETLE_OPT_DISPLAY_VRAM_LABEL_VAL,
      NULL,
      BEETLE_OPT_DISPLAY_VRAM_INFO_0_VAL,
      NULL,
      "osd",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
#endif
   {
      BEETLE_OPT(analog_calibration),
      BEETLE_OPT_ANALOG_CALIBRATION_LABEL_VAL,
      NULL,
      BEETLE_OPT_ANALOG_CALIBRATION_INFO_0_VAL,
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(analog_toggle),
      BEETLE_OPT_ANALOG_TOGGLE_LABEL_VAL,
      NULL,
      BEETLE_OPT_ANALOG_TOGGLE_INFO_0_VAL,
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { "enabled-analog", OPTION_VAL_ENABLED_ANALOG_VAL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(analog_toggle_combo),
      BEETLE_OPT_ANALOG_TOGGLE_COMBO_LABEL_VAL,
      NULL,
      BEETLE_OPT_ANALOG_TOGGLE_COMBO_INFO_0_VAL,
      NULL,
      "input",
      {
         { "l1+l2+r1+r2+start+select", OPTION_VAL_L1_L2_R1_R2_START_SELECT_VAL },
         { "l1+r1+select",             OPTION_VAL_L1_R1_SELECT_VAL },
         { "l1+r1+start",              OPTION_VAL_L1_R1_START_VAL },
         { "l1+r1+l3",                 OPTION_VAL_L1_R1_L3_VAL },
         { "l1+r1+r3",                 OPTION_VAL_L1_R1_R3_VAL },
         { "l2+r2+select",             OPTION_VAL_L2_R2_SELECT_VAL },
         { "l2+r2+start",              OPTION_VAL_L2_R2_START_VAL },
         { "l2+r2+l3",                 OPTION_VAL_L2_R2_L3_VAL },
         { "l2+r2+r3",                 OPTION_VAL_L2_R2_R3_VAL },
         { "l3+r3",                    OPTION_VAL_L3_R3_VAL },
         { NULL, NULL },
      },
      "l1+r1+select"
   },
   {
      BEETLE_OPT(analog_toggle_hold),
      BEETLE_OPT_ANALOG_TOGGLE_HOLD_LABEL_VAL,
      NULL,
      BEETLE_OPT_ANALOG_TOGGLE_HOLD_INFO_0_VAL,
      NULL,
      "input",
      {
         { "0", OPTION_VAL_0_VAL },
         { "1", OPTION_VAL_1_VAL },
         { "2", OPTION_VAL_2_VAL },
         { "3", OPTION_VAL_3_VAL },
         { "4", OPTION_VAL_4_VAL },
         { "5", OPTION_VAL_5_VAL },
         { NULL, NULL },
      },
      "1"
   },
   {
      BEETLE_OPT(enable_multitap_port1),
      BEETLE_OPT_ENABLE_MULTITAP_PORT1_LABEL_VAL,
      NULL,
      BEETLE_OPT_ENABLE_MULTITAP_PORT1_INFO_0_VAL,
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(enable_multitap_port2),
      BEETLE_OPT_ENABLE_MULTITAP_PORT2_LABEL_VAL,
      NULL,
      BEETLE_OPT_ENABLE_MULTITAP_PORT2_INFO_0_VAL,
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(gun_input_mode),
      BEETLE_OPT_GUN_INPUT_MODE_LABEL_VAL,
      NULL,
      BEETLE_OPT_GUN_INPUT_MODE_INFO_0_VAL,
      NULL,
      "input",
      {
         { "lightgun",    OPTION_VAL_LIGHTGUN_VAL },
         { "touchscreen", OPTION_VAL_TOUCHSCREEN_VAL },
         { NULL, NULL },
      },
      "lightgun"
   },
   // Shouldn't the gun_input_mode just be Mouse vs. Touchscreen?
   {
      BEETLE_OPT(gun_cursor),
      BEETLE_OPT_GUN_CURSOR_LABEL_VAL,
      NULL,
      BEETLE_OPT_GUN_CURSOR_INFO_0_VAL,
      NULL,
      "input",
      {
         { "off",   "disabled" },
         { "cross", OPTION_VAL_CROSS_VAL },
         { "dot",   OPTION_VAL_DOT_VAL },
         { NULL, NULL },
      },
      "cross"
   },
   {
      BEETLE_OPT(crosshair_color_p1),
      BEETLE_OPT_CROSSHAIR_COLOR_P1_LABEL_VAL,
      NULL,
      BEETLE_OPT_CROSSHAIR_COLOR_P1_INFO_0_VAL,
      NULL,
      "input",
      {
         { "red",    OPTION_VAL_RED_VAL },
         { "blue",   OPTION_VAL_BLUE_VAL },
         { "green",  OPTION_VAL_GREEN_VAL },
         { "orange", OPTION_VAL_ORANGE_VAL },
         { "yellow", OPTION_VAL_YELLOW_VAL },
         { "cyan",   OPTION_VAL_CYAN_VAL },
         { "pink",   OPTION_VAL_PINK_VAL },
         { "purple", OPTION_VAL_PURPLE_VAL },
         { "black",  OPTION_VAL_BLACK_VAL },
         { "white",  OPTION_VAL_WHITE_VAL },
         { NULL, NULL },
      },
      "red"
   },
   {
      BEETLE_OPT(crosshair_color_p2),
      BEETLE_OPT_CROSSHAIR_COLOR_P2_LABEL_VAL,
      NULL,
      BEETLE_OPT_CROSSHAIR_COLOR_P2_INFO_0_VAL,
      NULL,
      "input",
      {
         { "blue",   OPTION_VAL_BLUE_VAL },
         { "red",    OPTION_VAL_RED_VAL },
         { "green",  OPTION_VAL_GREEN_VAL },
         { "orange", OPTION_VAL_ORANGE_VAL },
         { "yellow", OPTION_VAL_YELLOW_VAL },
         { "cyan",   OPTION_VAL_CYAN_VAL },
         { "pink",   OPTION_VAL_PINK_VAL },
         { "purple", OPTION_VAL_PURPLE_VAL },
         { "black",  OPTION_VAL_BLACK_VAL },
         { "white",  OPTION_VAL_WHITE_VAL },
         { NULL, NULL },
      },
      "blue"
   },
   {
      BEETLE_OPT(mouse_sensitivity),
      BEETLE_OPT_MOUSE_SENSITIVITY_LABEL_VAL,
      NULL,
      BEETLE_OPT_MOUSE_SENSITIVITY_INFO_0_VAL,
      NULL,
      "input",
      {
         { "5%",   OPTION_VAL_5_O42_VAL },
         { "10%",  OPTION_VAL_10_VAL },
         { "15%",  OPTION_VAL_15_VAL },
         { "20%",  OPTION_VAL_20_VAL },
         { "25%",  OPTION_VAL_25_VAL },
         { "30%",  OPTION_VAL_30_VAL },
         { "35%",  OPTION_VAL_35_VAL },
         { "40%",  OPTION_VAL_40_VAL },
         { "45%",  OPTION_VAL_45_VAL },
         { "50%",  OPTION_VAL_50_VAL },
         { "55%",  OPTION_VAL_55_VAL },
         { "60%",  OPTION_VAL_60_VAL },
         { "65%",  OPTION_VAL_65_VAL },
         { "70%",  OPTION_VAL_70_VAL },
         { "75%",  OPTION_VAL_75_VAL },
         { "80%",  OPTION_VAL_80_VAL },
         { "85%",  OPTION_VAL_85_VAL },
         { "90%",  OPTION_VAL_90_VAL },
         { "95%",  OPTION_VAL_95_VAL },
         { "100%", OPTION_VAL_100_O42_VAL },
         { "105%", OPTION_VAL_105_VAL },
         { "110%", OPTION_VAL_110_VAL },
         { "115%", OPTION_VAL_115_VAL },
         { "120%", OPTION_VAL_120_VAL },
         { "125%", OPTION_VAL_125_VAL },
         { "130%", OPTION_VAL_130_VAL },
         { "135%", OPTION_VAL_135_VAL },
         { "140%", OPTION_VAL_140_VAL },
         { "145%", OPTION_VAL_145_VAL },
         { "150%", OPTION_VAL_150_VAL },
         { "155%", OPTION_VAL_155_VAL },
         { "160%", OPTION_VAL_160_VAL },
         { "165%", OPTION_VAL_165_VAL },
         { "170%", OPTION_VAL_170_VAL },
         { "175%", OPTION_VAL_175_VAL },
         { "180%", OPTION_VAL_180_VAL },
         { "185%", OPTION_VAL_185_VAL },
         { "190%", OPTION_VAL_190_VAL },
         { "195%", OPTION_VAL_195_VAL },
         { "200%", OPTION_VAL_200_VAL },
         { NULL, NULL },
      },
      "100%"
   },
   {
      BEETLE_OPT(negcon_response),
      BEETLE_OPT_NEGCON_RESPONSE_LABEL_VAL,
      NULL,
      BEETLE_OPT_NEGCON_RESPONSE_INFO_0_VAL,
      NULL,
      "input",
      {
         { "linear",    OPTION_VAL_LINEAR_VAL },
         { "quadratic", OPTION_VAL_QUADRATIC_VAL },
         { "cubic",     OPTION_VAL_CUBIC_VAL },
         { NULL, NULL },
      },
      "linear"
   },
   {
      BEETLE_OPT(negcon_deadzone),
      BEETLE_OPT_NEGCON_DEADZONE_LABEL_VAL,
      NULL,
      BEETLE_OPT_NEGCON_DEADZONE_INFO_0_VAL,
      NULL,
      "input",
      {
         { "0%",  OPTION_VAL_0_O44_VAL },
         { "5%",  OPTION_VAL_5_O42_VAL },
         { "10%", OPTION_VAL_10_VAL },
         { "15%", OPTION_VAL_15_VAL },
         { "20%", OPTION_VAL_20_VAL },
         { "25%", OPTION_VAL_25_VAL },
         { "30%", OPTION_VAL_30_VAL },
         { NULL, NULL },
      },
      "0%"
   },
   {
      BEETLE_OPT(use_mednafen_memcard0_method),
      BEETLE_OPT_USE_MEDNAFEN_MEMCARD0_METHOD_LABEL_VAL,
      NULL,
      BEETLE_OPT_USE_MEDNAFEN_MEMCARD0_METHOD_INFO_0_VAL,
      NULL,
      "memcards",
      {
         { "libretro", OPTION_VAL_LIBRETRO_VAL },
         { "mednafen", OPTION_VAL_MEDNAFEN_VAL },
         { NULL, NULL },
      },
      "libretro"
   },
   {
      BEETLE_OPT(enable_memcard1),
      BEETLE_OPT_ENABLE_MEMCARD1_LABEL_VAL,
      NULL,
      BEETLE_OPT_ENABLE_MEMCARD1_INFO_0_VAL,
      NULL,
      "memcards",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(shared_memory_cards),
      BEETLE_OPT_SHARED_MEMORY_CARDS_LABEL_VAL,
      NULL,
      BEETLE_OPT_SHARED_MEMORY_CARDS_INFO_0_VAL,
      NULL,
      "memcards",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(memcard_left_index),
      BEETLE_OPT_MEMCARD_LEFT_INDEX_LABEL_VAL,
      NULL,
      BEETLE_OPT_MEMCARD_LEFT_INDEX_INFO_0_VAL,
      NULL,
      "memcards",
      {
         { "0",  OPTION_VAL_0_O48_VAL },
         { "1",  NULL },
         { "2",  NULL },
         { "3",  NULL },
         { "4",  NULL },
         { "5",  NULL },
         { "6",  NULL },
         { "7",  NULL },
         { "8",  NULL },
         { "9",  NULL },
         { "10", NULL },
         { "11", NULL },
         { "12", NULL },
         { "13", NULL },
         { "14", NULL },
         { "15", NULL },
         { "16", NULL },
         { "17", NULL },
         { "18", NULL },
         { "19", NULL },
         { "20", NULL },
         { "21", NULL },
         { "22", NULL },
         { "23", NULL },
         { "24", NULL },
         { "25", NULL },
         { "26", NULL },
         { "27", NULL },
         { "28", NULL },
         { "29", NULL },
         { "30", NULL },
         { "31", NULL },
         { "32", NULL },
         { "33", NULL },
         { "34", NULL },
         { "35", NULL },
         { "36", NULL },
         { "37", NULL },
         { "38", NULL },
         { "39", NULL },
         { "40", NULL },
         { "41", NULL },
         { "42", NULL },
         { "43", NULL },
         { "44", NULL },
         { "45", NULL },
         { "46", NULL },
         { "47", NULL },
         { "48", NULL },
         { "49", NULL },
         { "50", NULL },
         { "51", NULL },
         { "52", NULL },
         { "53", NULL },
         { "54", NULL },
         { "55", NULL },
         { "56", NULL },
         { "57", NULL },
         { "58", NULL },
         { "59", NULL },
         { "60", NULL },
         { "61", NULL },
         { "62", NULL },
         { "63", NULL },
         { NULL, NULL },
      },
      "0"
   },
   {
      BEETLE_OPT(memcard_right_index),
      BEETLE_OPT_MEMCARD_RIGHT_INDEX_LABEL_VAL,
      NULL,
      BEETLE_OPT_MEMCARD_RIGHT_INDEX_INFO_0_VAL,
      NULL,
      "memcards",
      {
         { "0",  NULL },
         { "1",  OPTION_VAL_1_O49_VAL },
         { "2",  NULL },
         { "3",  NULL },
         { "4",  NULL },
         { "5",  NULL },
         { "6",  NULL },
         { "7",  NULL },
         { "8",  NULL },
         { "9",  NULL },
         { "10", NULL },
         { "11", NULL },
         { "12", NULL },
         { "13", NULL },
         { "14", NULL },
         { "15", NULL },
         { "16", NULL },
         { "17", NULL },
         { "18", NULL },
         { "19", NULL },
         { "20", NULL },
         { "21", NULL },
         { "22", NULL },
         { "23", NULL },
         { "24", NULL },
         { "25", NULL },
         { "26", NULL },
         { "27", NULL },
         { "28", NULL },
         { "29", NULL },
         { "30", NULL },
         { "31", NULL },
         { "32", NULL },
         { "33", NULL },
         { "34", NULL },
         { "35", NULL },
         { "36", NULL },
         { "37", NULL },
         { "38", NULL },
         { "39", NULL },
         { "40", NULL },
         { "41", NULL },
         { "42", NULL },
         { "43", NULL },
         { "44", NULL },
         { "45", NULL },
         { "46", NULL },
         { "47", NULL },
         { "48", NULL },
         { "49", NULL },
         { "50", NULL },
         { "51", NULL },
         { "52", NULL },
         { "53", NULL },
         { "54", NULL },
         { "55", NULL },
         { "56", NULL },
         { "57", NULL },
         { "58", NULL },
         { "59", NULL },
         { "60", NULL },
         { "61", NULL },
         { "62", NULL },
         { "63", NULL },
         { NULL, NULL },
      },
      "1"
   },
   {
      BEETLE_OPT(line_render),
      BEETLE_OPT_LINE_RENDER_LABEL_VAL,
      NULL,
      BEETLE_OPT_LINE_RENDER_INFO_0_VAL,
      NULL,
      "hacks",
      {
         { "default",    OPTION_VAL_DEFAULT_VAL },
         { "aggressive", OPTION_VAL_AGGRESSIVE_VAL },
         { "disabled",   NULL },
         { NULL, NULL },
      },
      "default"
   },
   {
      BEETLE_OPT(widescreen_hack),
      BEETLE_OPT_WIDESCREEN_HACK_LABEL_VAL,
      NULL,
      BEETLE_OPT_WIDESCREEN_HACK_INFO_0_VAL,
      NULL,
      "hacks",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(widescreen_hack_aspect_ratio),
      BEETLE_OPT_WIDESCREEN_HACK_ASPECT_RATIO_LABEL_VAL,
      NULL,
      BEETLE_OPT_WIDESCREEN_HACK_ASPECT_RATIO_INFO_0_VAL,
      NULL,
      "hacks",
      {
         { "16:9",  OPTION_VAL_16_9_VAL },
         { "16:10", OPTION_VAL_16_10_VAL },
         { "18:9",  OPTION_VAL_18_9_VAL },
         { "19:9",  OPTION_VAL_19_9_VAL },
         { "20:9",  OPTION_VAL_20_9_VAL },
         { "21:9",  OPTION_VAL_21_9_VAL }, // 64:27
         { "32:9",  OPTION_VAL_32_9_VAL },
         { NULL,    NULL },
      },
      "16:9"
   },
#if defined(HAVE_LIGHTREC)
   {
      BEETLE_OPT(cpu_dynarec),
      BEETLE_OPT_CPU_DYNAREC_LABEL_VAL,
      NULL,
      BEETLE_OPT_CPU_DYNAREC_INFO_0_VAL,
      NULL,
      "hacks",
      {
         { "disabled",        OPTION_VAL_DISABLED_VAL },
         { "execute",         OPTION_VAL_EXECUTE_VAL },
         { "execute_one",     OPTION_VAL_EXECUTE_ONE_VAL },
         { "run_interpreter", OPTION_VAL_RUN_INTERPRETER_VAL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(dynarec_invalidate),
      BEETLE_OPT_DYNAREC_INVALIDATE_LABEL_VAL,
      NULL,
      BEETLE_OPT_DYNAREC_INVALIDATE_INFO_0_VAL,
      NULL,
      "hacks",
      {
         { "full", OPTION_VAL_FULL_VAL },
         { "dma",  OPTION_VAL_DMA_VAL },
         { NULL, NULL },
      },
      "full"
   },
   {
      BEETLE_OPT(dynarec_eventcycles),
      BEETLE_OPT_DYNAREC_EVENTCYCLES_LABEL_VAL,
      NULL,
      BEETLE_OPT_DYNAREC_EVENTCYCLES_INFO_0_VAL,
      NULL,
      "hacks",
      {
         { "128",  OPTION_VAL_128_VAL },
         { "256",  NULL },
         { "384",  NULL },
         { "512",  NULL },
         { "640",  NULL },
         { "768",  NULL },
         { "896",  NULL },
         { "1024", NULL },
         { "1152", NULL },
         { "1280", NULL },
         { "1408", NULL },
         { "1536", NULL },
         { "1664", NULL },
         { "1792", NULL },
         { "1920", NULL },
         { "2048", NULL },
         { NULL, NULL },
      },
      "128"
   },
   {
      BEETLE_OPT(dynarec_spu_samples),
      BEETLE_OPT_DYNAREC_SPU_SAMPLES_LABEL_VAL,
      NULL,
      BEETLE_OPT_DYNAREC_SPU_SAMPLES_INFO_0_VAL,
      NULL,
      "hacks",
      {
         { "1",  OPTION_VAL_1_O49_VAL },
         { "4",  NULL },
         { "16", NULL },
         { NULL, NULL },
      },
      "1"
   },
#endif
   {
      BEETLE_OPT(core_timing_fps),
      BEETLE_OPT_CORE_TIMING_FPS_LABEL_VAL,
      NULL,
      BEETLE_OPT_CORE_TIMING_FPS_INFO_0_VAL,
      NULL,
      "video",
      {
         { "auto_toggle",       OPTION_VAL_AUTO_TOGGLE_VAL },
         { "force_progressive", OPTION_VAL_FORCE_PROGRESSIVE_VAL },
         { "force_interlaced",  OPTION_VAL_FORCE_INTERLACED_VAL },
      },
      "auto_toggle"
   },
   {
      BEETLE_OPT(deinterlacer),
      BEETLE_OPT_DEINTERLACER_LABEL_VAL,
      NULL,
      BEETLE_OPT_DEINTERLACER_INFO_0_VAL,
      NULL,
      "video",
      {
         { "weave",      OPTION_VAL_WEAVE_VAL },
         { "bob",        OPTION_VAL_BOB_VAL },
         { "bob_offset", OPTION_VAL_BOB_OFFSET_VAL },
         { "fastmad",    OPTION_VAL_FASTMAD_VAL },
         { "off",        "Off" },
         { NULL, NULL },
      },
      "weave"
   },
   {
      BEETLE_OPT(frame_duping),
      BEETLE_OPT_FRAME_DUPING_LABEL_VAL,
      NULL,
      BEETLE_OPT_FRAME_DUPING_INFO_0_VAL,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(pal_video_timing_override),
      BEETLE_OPT_PAL_VIDEO_TIMING_OVERRIDE_LABEL_VAL,
      NULL,
      BEETLE_OPT_PAL_VIDEO_TIMING_OVERRIDE_INFO_0_VAL,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(aspect_ratio),
      BEETLE_OPT_ASPECT_RATIO_LABEL_VAL,
      NULL,
      BEETLE_OPT_ASPECT_RATIO_INFO_0_VAL,
      NULL,
      "video",
      {
         { "corrected", OPTION_VAL_CORRECTED_VAL },
         { "uncorrected", OPTION_VAL_UNCORRECTED_VAL },
         { "4:3",  OPTION_VAL_4_3_VAL },
         { "ntsc", OPTION_VAL_NTSC_VAL },
      },
      "corrected"
   },
   {
      BEETLE_OPT(crop_overscan),
      BEETLE_OPT_CROP_OVERSCAN_LABEL_VAL,
      NULL,
      BEETLE_OPT_CROP_OVERSCAN_INFO_0_VAL,
      NULL,
      "video",
      {
         { "disabled", "Disabled" },
         { "static",   OPTION_VAL_STATIC_VAL },
         { "smart",    OPTION_VAL_SMART_VAL },
         { NULL, NULL },
      },
      "smart"
   },
   {
      BEETLE_OPT(image_crop),
      BEETLE_OPT_IMAGE_CROP_LABEL_VAL,
      NULL,
      BEETLE_OPT_IMAGE_CROP_INFO_0_VAL,
      NULL,
      "video",
      {
         { "disabled", "0" },
         { "1px",      OPTION_VAL_1PX_VAL },
         { "2px",      OPTION_VAL_2PX_VAL },
         { "3px",      OPTION_VAL_3PX_VAL },
         { "4px",      OPTION_VAL_4PX_VAL },
         { "5px",      OPTION_VAL_5PX_VAL },
         { "6px",      OPTION_VAL_6PX_VAL },
         { "7px",      OPTION_VAL_7PX_VAL },
         { "8px",      OPTION_VAL_8PX_VAL },
         { "9px",      OPTION_VAL_9PX_VAL },
         { "10px",     OPTION_VAL_10PX_VAL },
         { "11px",     OPTION_VAL_11PX_VAL },
         { "12px",     OPTION_VAL_12PX_VAL },
         { "13px",     OPTION_VAL_13PX_VAL },
         { "14px",     OPTION_VAL_14PX_VAL },
         { "15px",     OPTION_VAL_15PX_VAL },
         { "16px",     OPTION_VAL_16PX_VAL },
         { "17px",     OPTION_VAL_17PX_VAL },
         { "18px",     OPTION_VAL_18PX_VAL },
         { "19px",     OPTION_VAL_19PX_VAL },
         { "20px",     OPTION_VAL_20PX_VAL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(image_offset),
      BEETLE_OPT_IMAGE_OFFSET_LABEL_VAL,
      NULL,
      BEETLE_OPT_IMAGE_OFFSET_INFO_0_VAL,
      NULL,
      "video",
      {
         { "-12px",    OPTION_VAL_12PX_O64_VAL },
         { "-11px",    OPTION_VAL_11PX_O64_VAL },
         { "-10px",    OPTION_VAL_10PX_O64_VAL },
         { "-9px",     OPTION_VAL_9PX_O64_VAL },
         { "-8px",     OPTION_VAL_8PX_O64_VAL },
         { "-7px",     OPTION_VAL_7PX_O64_VAL },
         { "-6px",     OPTION_VAL_6PX_O64_VAL },
         { "-5px",     OPTION_VAL_5PX_O64_VAL },
         { "-4px",     OPTION_VAL_4PX_O64_VAL },
         { "-3px",     OPTION_VAL_3PX_O64_VAL },
         { "-2px",     OPTION_VAL_2PX_O64_VAL },
         { "-1px",     OPTION_VAL_1PX_O64_VAL },
         { "disabled", "0" },
         { "+1px",     OPTION_VAL_1PX_O64_0_VAL },
         { "+2px",     OPTION_VAL_2PX_O64_0_VAL },
         { "+3px",     OPTION_VAL_3PX_O64_0_VAL },
         { "+4px",     OPTION_VAL_4PX_O64_0_VAL },
         { "+5px",     OPTION_VAL_5PX_O64_0_VAL },
         { "+6px",     OPTION_VAL_6PX_O64_0_VAL },
         { "+7px",     OPTION_VAL_7PX_O64_0_VAL },
         { "+8px",     OPTION_VAL_8PX_O64_0_VAL },
         { "+9px",     OPTION_VAL_9PX_O64_0_VAL },
         { "+10px",    OPTION_VAL_10PX_O64_0_VAL },
         { "+11px",    OPTION_VAL_11PX_O64_0_VAL },
         { "+12px",    OPTION_VAL_12PX_O64_0_VAL },
         { NULL, NULL },
      },
      "disabled"
   },
#if defined(HAVE_OPENGL) || defined(HAVE_OPENGLES) || defined(HAVE_VULKAN)
   {
      BEETLE_OPT(image_offset_cycles),
      BEETLE_OPT_IMAGE_OFFSET_CYCLES_LABEL_VAL,
      NULL,
      BEETLE_OPT_IMAGE_OFFSET_CYCLES_INFO_0_VAL,
      NULL,
      "video",
      {
         { "-40",      NULL },
         { "-39",      NULL },
         { "-38",      NULL },
         { "-37",      NULL },
         { "-36",      NULL },
         { "-35",      NULL },
         { "-34",      NULL },
         { "-33",      NULL },
         { "-32",      NULL },
         { "-31",      NULL },
         { "-30",      NULL },
         { "-29",      NULL },
         { "-28",      NULL },
         { "-27",      NULL },
         { "-26",      NULL },
         { "-25",      NULL },
         { "-24",      NULL },
         { "-23",      NULL },
         { "-22",      NULL },
         { "-21",      NULL },
         { "-20",      NULL },
         { "-19",      NULL },
         { "-18",      NULL },
         { "-17",      NULL },
         { "-16",      NULL },
         { "-15",      NULL },
         { "-14",      NULL },
         { "-13",      NULL },
         { "-12",      NULL },
         { "-11",      NULL },
         { "-10",      NULL },
         { "-9",       NULL },
         { "-8",       NULL },
         { "-7",       NULL },
         { "-6",       NULL },
         { "-5",       NULL },
         { "-4",       NULL },
         { "-3",       NULL },
         { "-2",       NULL },
         { "-1",       NULL },
         { "0",        "0" },
         { "+1",       NULL },
         { "+2",       NULL },
         { "+3",       NULL },
         { "+4",       NULL },
         { "+5",       NULL },
         { "+6",       NULL },
         { "+7",       NULL },
         { "+8",       NULL },
         { "+9",       NULL },
         { "+10",      NULL },
         { "+11",      NULL },
         { "+12",      NULL },
         { "+13",      NULL },
         { "+14",      NULL },
         { "+15",      NULL },
         { "+16",      NULL },
         { "+17",      NULL },
         { "+18",      NULL },
         { "+19",      NULL },
         { "+20",      NULL },
         { "+21",      NULL },
         { "+22",      NULL },
         { "+23",      NULL },
         { "+24",      NULL },
         { "+25",      NULL },
         { "+26",      NULL },
         { "+27",      NULL },
         { "+28",      NULL },
         { "+29",      NULL },
         { "+30",      NULL },
         { "+31",      NULL },
         { "+32",      NULL },
         { "+33",      NULL },
         { "+34",      NULL },
         { "+35",      NULL },
         { "+36",      NULL },
         { "+37",      NULL },
         { "+38",      NULL },
         { "+39",      NULL },
         { "+40",      NULL },
         { NULL, NULL},
      },
      "0"
   },
#endif
   {
      BEETLE_OPT(initial_scanline),
      BEETLE_OPT_INITIAL_SCANLINE_LABEL_VAL,
      NULL,
      BEETLE_OPT_INITIAL_SCANLINE_INFO_0_VAL,
      NULL,
      "video",
      {
         { "0",  OPTION_VAL_0_O48_VAL },
         { "1",  NULL },
         { "2",  NULL },
         { "3",  NULL },
         { "4",  NULL },
         { "5",  NULL },
         { "6",  NULL },
         { "7",  NULL },
         { "8",  NULL },
         { "9",  NULL },
         { "10", NULL },
         { "11", NULL },
         { "12", NULL },
         { "13", NULL },
         { "14", NULL },
         { "15", NULL },
         { "16", NULL },
         { "17", NULL },
         { "18", NULL },
         { "19", NULL },
         { "20", NULL },
         { "21", NULL },
         { "22", NULL },
         { "23", NULL },
         { "24", NULL },
         { "25", NULL },
         { "26", NULL },
         { "27", NULL },
         { "28", NULL },
         { "29", NULL },
         { "30", NULL },
         { "31", NULL },
         { "32", NULL },
         { "33", NULL },
         { "34", NULL },
         { "35", NULL },
         { "36", NULL },
         { "37", NULL },
         { "38", NULL },
         { "39", NULL },
         { "40", NULL },
         { NULL, NULL },
      },
      "0"
   },
   {
      BEETLE_OPT(last_scanline),
      BEETLE_OPT_LAST_SCANLINE_LABEL_VAL,
      NULL,
      BEETLE_OPT_LAST_SCANLINE_INFO_0_VAL,
      NULL,
      "video",
      {
         { "210", NULL },
         { "211", NULL },
         { "212", NULL },
         { "213", NULL },
         { "214", NULL },
         { "215", NULL },
         { "216", NULL },
         { "217", NULL },
         { "218", NULL },
         { "219", NULL },
         { "220", NULL },
         { "221", NULL },
         { "222", NULL },
         { "223", NULL },
         { "224", NULL },
         { "225", NULL },
         { "226", NULL },
         { "227", NULL },
         { "228", NULL },
         { "229", NULL },
         { "230", NULL },
         { "231", NULL },
         { "232", NULL },
         { "233", NULL },
         { "234", NULL },
         { "235", NULL },
         { "236", NULL },
         { "237", NULL },
         { "238", NULL },
         { "239", OPTION_VAL_239_VAL },
         { NULL, NULL },
      },
      "239"
   },
   {
      BEETLE_OPT(initial_scanline_pal),
      BEETLE_OPT_INITIAL_SCANLINE_PAL_LABEL_VAL,
      NULL,
      BEETLE_OPT_INITIAL_SCANLINE_PAL_INFO_0_VAL,
      NULL,
      "video",
      {
         { "0",  OPTION_VAL_0_O48_VAL },
         { "1",  NULL },
         { "2",  NULL },
         { "3",  NULL },
         { "4",  NULL },
         { "5",  NULL },
         { "6",  NULL },
         { "7",  NULL },
         { "8",  NULL },
         { "9",  NULL },
         { "10", NULL },
         { "11", NULL },
         { "12", NULL },
         { "13", NULL },
         { "14", NULL },
         { "15", NULL },
         { "16", NULL },
         { "17", NULL },
         { "18", NULL },
         { "19", NULL },
         { "20", NULL },
         { "21", NULL },
         { "22", NULL },
         { "23", NULL },
         { "24", NULL },
         { "25", NULL },
         { "26", NULL },
         { "27", NULL },
         { "28", NULL },
         { "29", NULL },
         { "30", NULL },
         { "31", NULL },
         { "32", NULL },
         { "33", NULL },
         { "34", NULL },
         { "35", NULL },
         { "36", NULL },
         { "37", NULL },
         { "38", NULL },
         { "39", NULL },
         { "40", NULL },
         { NULL, NULL },
      },
      "0"
   },
   {
      BEETLE_OPT(last_scanline_pal),
      BEETLE_OPT_LAST_SCANLINE_PAL_LABEL_VAL,
      NULL,
      BEETLE_OPT_LAST_SCANLINE_PAL_INFO_0_VAL,
      NULL,
      "video",
      {
         { "230", NULL },
         { "231", NULL },
         { "232", NULL },
         { "233", NULL },
         { "234", NULL },
         { "235", NULL },
         { "236", NULL },
         { "237", NULL },
         { "238", NULL },
         { "239", NULL },
         { "240", NULL },
         { "241", NULL },
         { "242", NULL },
         { "243", NULL },
         { "244", NULL },
         { "245", NULL },
         { "246", NULL },
         { "247", NULL },
         { "248", NULL },
         { "249", NULL },
         { "250", NULL },
         { "251", NULL },
         { "252", NULL },
         { "253", NULL },
         { "254", NULL },
         { "255", NULL },
         { "256", NULL },
         { "257", NULL },
         { "258", NULL },
         { "259", NULL },
         { "260", NULL },
         { "261", NULL },
         { "262", NULL },
         { "263", NULL },
         { "264", NULL },
         { "265", NULL },
         { "266", NULL },
         { "267", NULL },
         { "268", NULL },
         { "269", NULL },
         { "270", NULL },
         { "271", NULL },
         { "272", NULL },
         { "273", NULL },
         { "274", NULL },
         { "275", NULL },
         { "276", NULL },
         { "277", NULL },
         { "278", NULL },
         { "279", NULL },
         { "280", NULL },
         { "281", NULL },
         { "282", NULL },
         { "283", NULL },
         { "284", NULL },
         { "285", NULL },
         { "286", NULL },
         { "287", OPTION_VAL_287_VAL },
         { NULL, NULL },
      },
      "287"
   },
   { NULL, NULL, NULL, NULL, NULL, NULL, {{0}}, NULL },
};
struct retro_core_options_v2 options_val = {
   option_cats_val,
   option_defs_val
};

/* RETRO_LANGUAGE_VN */

#define CATEGORY_SYSTEM_LABEL_VN "Hệ thống"
#define CATEGORY_SYSTEM_INFO_0_VN "Thay đổi các tùy chọn hệ thống và ép xung."
#define CATEGORY_VIDEO_LABEL_VN NULL
#define CATEGORY_VIDEO_INFO_0_VN "Thay đổi tỷ lệ khung hình, cắt xén màn hình, bộ lọc video và cài đặt bỏ qua khung hình."
#define CATEGORY_PGXP_LABEL_VN "PGXP (Đường ống biến đổi hình học chính xác)"
#define CATEGORY_PGXP_INFO_0_VN "Các tùy chọn này kiểm soát các cải tiến có thể cải thiện đồ họa so với máy chơi game gốc. PGXP có thể loại bỏ hiện tượng cong vênh kết cấu và lỗi Z-fighting."
#define CATEGORY_OSD_LABEL_VN "Hiển thị trên màn hình"
#define CATEGORY_OSD_INFO_0_VN "Thay đổi thông báo hiển thị trên màn hình."
#define CATEGORY_INPUT_LABEL_VN "Đều khiển"
#define CATEGORY_INPUT_INFO_0_VN "Thay đổi cài đặt súng ánh sáng, chuột và neGcon."
#define CATEGORY_MEMCARDS_LABEL_VN "Thẻ nhớ"
#define CATEGORY_MEMCARDS_INFO_0_VN "Thay đổi các cài đặt liên quan đến thẻ nhớ ảo."
#define CATEGORY_HACKS_LABEL_VN "Thủ thuật giả lập"
#define CATEGORY_HACKS_INFO_0_VN "Thay đổi các cài đặt độ chính xác giả lập, ảnh hưởng đến hiệu năng mức thấp và khả năng tương thích."
#define BEETLE_OPT_CPU_FREQ_SCALE_LABEL_VN "Điều chỉnh tần số CPU (Ép xung)"
#define BEETLE_OPT_CPU_FREQ_SCALE_INFO_0_VN "Ép xung (hoặc giảm xung) CPU của PSX giả lập. Ép xung có thể loại bỏ hiện tượng chậm và cải thiện tốc độ khung hình trong một số trò chơi, đổi lại yêu cầu hiệu năng cao hơn. Lưu ý rằng một số trò chơi có bộ giới hạn khung hình nội bộ và có thể không hưởng lợi từ việc ép xung. Có thể khiến một số hiệu ứng trong một số trò chơi chạy nhanh hơn so với dự định khi ép xung."
#define OPTION_VAL_50_VN NULL
#define OPTION_VAL_60_VN NULL
#define OPTION_VAL_70_VN NULL
#define OPTION_VAL_80_VN NULL
#define OPTION_VAL_90_VN NULL
#define OPTION_VAL_100_VN "100% (Gốc)"
#define OPTION_VAL_110_VN NULL
#define OPTION_VAL_120_VN NULL
#define OPTION_VAL_130_VN NULL
#define OPTION_VAL_140_VN NULL
#define OPTION_VAL_150_VN NULL
#define OPTION_VAL_160_VN NULL
#define OPTION_VAL_170_VN NULL
#define OPTION_VAL_180_VN NULL
#define OPTION_VAL_190_VN NULL
#define OPTION_VAL_200_VN NULL
#define OPTION_VAL_210_VN NULL
#define OPTION_VAL_220_VN NULL
#define OPTION_VAL_230_VN NULL
#define OPTION_VAL_240_VN NULL
#define OPTION_VAL_250_VN NULL
#define OPTION_VAL_260_VN NULL
#define OPTION_VAL_270_VN NULL
#define OPTION_VAL_280_VN NULL
#define OPTION_VAL_290_VN NULL
#define OPTION_VAL_300_VN NULL
#define OPTION_VAL_310_VN NULL
#define OPTION_VAL_320_VN NULL
#define OPTION_VAL_330_VN NULL
#define OPTION_VAL_340_VN NULL
#define OPTION_VAL_350_VN NULL
#define OPTION_VAL_360_VN NULL
#define OPTION_VAL_370_VN NULL
#define OPTION_VAL_380_VN NULL
#define OPTION_VAL_390_VN NULL
#define OPTION_VAL_400_VN NULL
#define OPTION_VAL_410_VN NULL
#define OPTION_VAL_420_VN NULL
#define OPTION_VAL_430_VN NULL
#define OPTION_VAL_440_VN NULL
#define OPTION_VAL_450_VN NULL
#define OPTION_VAL_460_VN NULL
#define OPTION_VAL_470_VN NULL
#define OPTION_VAL_480_VN NULL
#define OPTION_VAL_490_VN NULL
#define OPTION_VAL_500_VN NULL
#define OPTION_VAL_510_VN NULL
#define OPTION_VAL_520_VN NULL
#define OPTION_VAL_530_VN NULL
#define OPTION_VAL_540_VN NULL
#define OPTION_VAL_550_VN NULL
#define OPTION_VAL_560_VN NULL
#define OPTION_VAL_570_VN NULL
#define OPTION_VAL_580_VN NULL
#define OPTION_VAL_590_VN NULL
#define OPTION_VAL_600_VN NULL
#define OPTION_VAL_610_VN NULL
#define OPTION_VAL_620_VN NULL
#define OPTION_VAL_630_VN NULL
#define OPTION_VAL_640_VN NULL
#define OPTION_VAL_650_VN NULL
#define OPTION_VAL_660_VN NULL
#define OPTION_VAL_670_VN NULL
#define OPTION_VAL_680_VN NULL
#define OPTION_VAL_690_VN NULL
#define OPTION_VAL_700_VN NULL
#define OPTION_VAL_710_VN NULL
#define OPTION_VAL_720_VN NULL
#define OPTION_VAL_730_VN NULL
#define OPTION_VAL_740_VN NULL
#define OPTION_VAL_750_VN NULL
#define BEETLE_OPT_GPU_OVERCLOCK_LABEL_VN "Ép xung bộ xử lý đồ họa (GPU Rasterizer)"
#define BEETLE_OPT_GPU_OVERCLOCK_INFO_0_VN "Bật ép xung bộ raster 2D bên trong GPU giả lập của PSX. Không cải thiện việc hiển thị 3D và nhìn chung tác động rất nhỏ."
#define OPTION_VAL_1X_NATIVE_VN "1x (gốc)"
#define OPTION_VAL_2X_VN NULL
#define OPTION_VAL_4X_VN NULL
#define OPTION_VAL_8X_VN NULL
#define OPTION_VAL_16X_VN NULL
#define OPTION_VAL_32X_VN NULL
#define BEETLE_OPT_GTE_OVERCLOCK_LABEL_VN "Ép xung GTE"
#define BEETLE_OPT_GTE_OVERCLOCK_INFO_0_VN "Giảm tất cả các thao tác GTE giả lập (bộ coprocesor CPU cho đồ họa 3D) xuống độ trễ cố định một chu kỳ. Đối với các game sử dụng nhiều GTE, điều này có thể cải thiện đáng kể tốc độ khung hình và độ ổn định thời gian khung hình."
#define BEETLE_OPT_CD_ACCESS_METHOD_LABEL_VN "Phương thức truy cập đĩa CD"
#define BEETLE_OPT_CD_ACCESS_METHOD_INFO_0_VN "Đồng bộ: mô phỏng chính xác hành vi của phần cứng gốc.Bất đồng bộ: có thể giảm hiện tượng giật hình trên các thiết bị có bộ nhớ lưu trữ chậm.Nạp sẵn: tải toàn bộ nội dung đĩa vào bộ nhớ khi khởi chạy trò chơi, có thể cải thiện thời gian tải trong game nhưng sẽ gây chậm lúc khởi động ban đầu. Chế độ này có thể gây vấn đề trên các hệ thống có ít RAM và sẽ tự động quay về chế độ đồng bộ khi dùng đĩa vật lý.Cần khởi động lại để áp dụng."
#define OPTION_VAL_SYNC_VN "Đồng bộ"
#define OPTION_VAL_ASYNC_VN "Không đồng bộ"
#define OPTION_VAL_PRECACHE_VN "Tải trước vào bộ nhớ"
#define BEETLE_OPT_CD_FASTLOAD_LABEL_VN "Tốc độ tải CD"
#define BEETLE_OPT_CD_FASTLOAD_INFO_0_VN "Các giá trị cao hơn “2x (Gốc)” có thể giảm đáng kể thời gian tải trong game, nhưng có thể gây lỗi về thời gian xử lý. Một số trò chơi có thể không hoạt động đúng khi vượt quá một mức nhất định."
#define OPTION_VAL_2X_NATIVE_VN "2x (Nguyên bản)"
#define OPTION_VAL_6X_VN NULL
#define OPTION_VAL_10X_VN NULL
#define OPTION_VAL_12X_VN NULL
#define OPTION_VAL_14X_VN NULL
#define BEETLE_OPT_SKIP_BIOS_LABEL_VN "Bỏ qua BIOS"
#define BEETLE_OPT_SKIP_BIOS_INFO_0_VN "Bỏ qua hoạt ảnh khởi động BIOS của PlayStation thường hiển thị khi tải nội dung. Lưu ý: Bật tùy chọn này có thể gây vấn đề tương thích với một số game (game PAL có bảo vệ bản quyền, Saga Frontier, v.v.)."
#define BEETLE_OPT_OVERRIDE_BIOS_LABEL_VN "Ghi đè BIOS"
#define BEETLE_OPT_OVERRIDE_BIOS_INFO_0_VN "Ghi đè BIOS theo vùng tiêu chuẩn bằng BIOS không phân vùng nếu được tìm thấy.Cần khởi động lại."
#define OPTION_VAL_PSXONPSP_VN "BIOS PS1 cho PSP"
#define OPTION_VAL_PS1_ROM_VN "BIOS PS1 cho PS3"
#define OPTION_VAL_OPENBIOS_VN "Bios mã nguồn mở"
#define BEETLE_OPT_REGION_LABEL_VN "Vùng Hệ thống"
#define BEETLE_OPT_REGION_INFO_0_VN "Thiết lập khu vực dự phòng được sử dụng khi nội dung không thể tự động nhận diện (ví dụ: file PS-X EXE thô/raw). Các trò chơi chạy bằng đĩa vẫn sẽ sử dụng region được ghi sẵn trên đĩa. Cần khởi động lại để áp dụng."
#define OPTION_VAL_AUTO_VN "Tự động"
#define OPTION_VAL_NTSC_J_VN "NTSC-J (Nhật Bản)"
#define OPTION_VAL_NTSC_U_VN "NTSC-U (Bắc Mỹ)"
#define OPTION_VAL_PAL_VN "PAL (Châu Âu)"
#define BEETLE_OPT_INTERNAL_RESOLUTION_LABEL_VN "Độ phân giải GPU nội bộ"
#define BEETLE_OPT_INTERNAL_RESOLUTION_INFO_0_VN "Độ phân giải cao hơn “1x (Native)” sẽ cải thiện độ chi tiết của các mô hình 3D nhưng yêu cầu hiệu năng cao hơn. Các thành phần 2D nhìn chung không bị ảnh hưởng bởi thiết lập này."
#define BEETLE_OPT_RENDERER_LABEL_VN "Renderer (Trình kết xuất)"
#define BEETLE_OPT_RENDERER_INFO_0_VN "Trình kết xuất phần mềm cho độ chính xác cao nhất nhưng yêu cầu hiệu năng rất lớn khi tăng độ phân giải GPU nội bộ.Các trình kết xuất phần cứng tuy kém chính xác hơn nhưng cải thiện hiệu năng so với phần mềm khi chạy ở độ phân giải nội bộ cao, đồng thời cho phép nhiều cải tiến đồ họa.“Hardware (Auto)” sẽ tự động chọn trình kết xuất Vulkan hoặc OpenGL tùy theo driver video của frontend libretro hiện tại. Nếu driver không hỗ trợ Vulkan hoặc OpenGL 3.3, core sẽ quay về trình kết xuất phần mềm.Cần khởi động lại để áp dụng."
#define OPTION_VAL_HARDWARE_VN "Phần cứng (Tự động)"
#define OPTION_VAL_HARDWARE_GL_VN "Phần cứng (OpenGL)"
#define OPTION_VAL_HARDWARE_VK_VN "Phần cứng (Vulkan)"
#define OPTION_VAL_SOFTWARE_VN "Phần mềm"
#define BEETLE_OPT_RENDERER_SOFTWARE_FB_LABEL_VN "Bộ nhớ khung phần mềm"
#define BEETLE_OPT_RENDERER_SOFTWARE_FB_INFO_0_VN "Bật mô phỏng chính xác các hiệu ứng bộ nhớ khung (ví dụ: mờ chuyển động, xoáy chiến đấu FF7) khi sử dụng bộ kết xuất phần cứng bằng cách chạy một bản sao bộ kết xuất phần mềm ở độ phân giải gốc trong nền. Nếu tắt, các thao tác này sẽ bị bỏ qua (OpenGL) hoặc được kết xuất trên GPU (Vulkan). Tắt có thể cải thiện hiệu năng nhưng có thể gây lỗi đồ họa nghiêm trọng. Giữ bật nếu không chắc chắn."
#define BEETLE_OPT_DEPTH_LABEL_VN "Độ sâu màu nội bộ"
#define BEETLE_OPT_DEPTH_INFO_0_VN "Độ sâu màu cao hơn có thể giảm hiện tượng dải màu (color banding) mà không cần dùng dithering. 16 bpp mô phỏng phần cứng gốc nhưng có thể xuất hiện dải màu rõ nếu không bật dithering. Khuyến nghị tắt “Dithering Pattern” khi đặt tùy chọn này ở 32 bpp."
#define OPTION_VAL_16BPP_NATIVE_VN "16 bpp (Gốc)"
#define OPTION_VAL_32BPP_VN NULL
#define BEETLE_OPT_DITHER_MODE_LABEL_VN "Mẫu dither"
#define BEETLE_OPT_DITHER_MODE_INFO_0_VN "“1x (Native)” mô phỏng dithering độ phân giải thấp nguyên bản của phần cứng gốc để làm mượt các artefact dải màu ở độ sâu màu gốc. “Internal Resolution” sẽ mở rộng độ chi tiết của dithering theo độ phân giải nội bộ đã cấu hình để cho kết quả sạch hơn. Khuyến nghị tắt khi chạy ở độ sâu màu 32 bpp.Lưu ý: Trên Vulkan, bật tùy chọn này sẽ buộc giảm mẫu (downsampling) về độ sâu màu gốc; tắt thì sẽ tự động bật xuất hình ở độ sâu màu cao hơn."
#define OPTION_VAL_INTERNAL_RESOLUTION_VN "Độ phân giải nội bộ"
#define BEETLE_OPT_SCALED_UV_OFFSET_LABEL_VN "Độ lệch UV kết cấu"
#define BEETLE_OPT_SCALED_UV_OFFSET_INFO_0_VN "Mẫu kết cấu cho đa giác 3D với độ lệch khi độ phân giải nội bộ cao hơn 1x. Giảm các đường nối kết cấu nhưng có thể gây ra lỗi đồ họa không mong muốn."
#define BEETLE_OPT_FILTER_LABEL_VN "Lọc kết cấu"
#define BEETLE_OPT_FILTER_INFO_0_VN "“Nearest” mô phỏng phần cứng gốc.“Bilinear” và “3-Point” là các bộ lọc làm mượt, giảm vỡ điểm ảnh bằng cách làm mờ.“SABR”, “xBR”, “JINC2” là các bộ lọc upscale có thể cải thiện độ chi tiết/độ sắc nét của texture nhưng đòi hỏi hiệu năng cao hơn. Chỉ được hỗ trợ bởi các trình kết xuất phần cứng."
#define OPTION_VAL_NEAREST_VN "Gần nhất"
#define OPTION_VAL_SABR_VN NULL
#define OPTION_VAL_XBR_VN NULL
#define OPTION_VAL_BILINEAR_VN "Lọc song tuyến"
#define OPTION_VAL_3_POINT_VN NULL
#define OPTION_VAL_JINC2_VN NULL
#define BEETLE_OPT_FILTER_EXCLUDE_SPRITE_LABEL_VN "Loại trừ Sprites khỏi bộ lọc"
#define BEETLE_OPT_FILTER_EXCLUDE_SPRITE_INFO_0_VN "Không áp dụng lọc kết cấu cho sprites. Ngăn các đường nối trong nhiều trò chơi với nền được vẽ bằng sprite 2D. Sử dụng cùng với Làm mượt thích ứng hoặc bộ lọc xử lý hậu kỳ khác để đạt hiệu quả tốt nhất."
#define OPTION_VAL_OPAQUE_VN "Chỉ mờ đục"
#define OPTION_VAL_ALL_VN "Mờ đục và bán trong suốt"
#define BEETLE_OPT_FILTER_EXCLUDE_2D_POLYGON_LABEL_VN "Loại trừ đa giác 2D khỏi Lọc"
#define BEETLE_OPT_FILTER_EXCLUDE_2D_POLYGON_INFO_0_VN "Không áp dụng lọc kết cấu cho đa giác 2D. Các đa giác 2D được phát hiện bằng phương pháp heuristic và có thể xảy ra lỗi. Sử dụng cùng với Làm mượt thích ứng hoặc bộ lọc xử lý hậu kỳ khác để đạt hiệu quả tốt nhất."
#define BEETLE_OPT_ADAPTIVE_SMOOTHING_LABEL_VN "Làm mượt thích ứng"
#define BEETLE_OPT_ADAPTIVE_SMOOTHING_INFO_0_VN "Làm mượt đồ họa 2D và các phần tử giao diện mà không làm mờ các đối tượng 3D được dựng hình. Chỉ được hỗ trợ bởi bộ kết xuất Vulkan."
#define BEETLE_OPT_SUPER_SAMPLING_LABEL_VN "Siêu lấy mẫu (về độ phân giải gốc)"
#define BEETLE_OPT_SUPER_SAMPLING_INFO_0_VN "Giảm mẫu nội dung đã dựng từ độ phân giải nội bộ được nâng cấp xuống độ phân giải gốc. Kết hợp với hệ số độ phân giải nội bộ cao hơn cho phép trò chơi hiển thị các đối tượng 3D được làm mịn răng cưa ở độ phân giải gốc thấp. Cho kết quả tốt nhất khi áp dụng cho các tựa game kết hợp cả 2D và 3D (ví dụ: nhân vật 3D trên nền dựng sẵn), và hoạt động tốt với shader CRT.Chỉ được hỗ trợ bởi bộ kết xuất Vulkan. Lưu ý: nên tắt “Mẫu tô bóng Dithering” khi bật tùy chọn này."
#define BEETLE_OPT_MSAA_LABEL_VN "Khử răng cưa đa mẫu (Msaa)"
#define BEETLE_OPT_MSAA_INFO_0_VN "MSAA cải thiện hiển thị của các đối tượng 3D. Chỉ được hỗ trợ bởi trình kết xuất Vulkan."
#define OPTION_VAL_1X_VN NULL
#define BEETLE_OPT_MDEC_YUV_LABEL_VN "Bộ lọc MDEC YUV Chroma"
#define BEETLE_OPT_MDEC_YUV_INFO_0_VN "Cải thiện chất lượng phát lại FMV bằng cách giảm hiện tượng khối vuông (macroblocking) và răng cưa. Chỉ được hỗ trợ bởi bộ kết xuất Vulkan."
#define BEETLE_OPT_TRACK_TEXTURES_LABEL_VN "Theo dõi kết cấu"
#define BEETLE_OPT_TRACK_TEXTURES_INFO_0_VN "Điều kiện bắt buộc để trích xuất (dump) và thay thế texture. Rất có thể gây crash trong hầu hết các trò chơi."
#define BEETLE_OPT_DUMP_TEXTURES_LABEL_VN "Xuất kết cấu"
#define BEETLE_OPT_DUMP_TEXTURES_INFO_0_VN "Xuất các kết cấu đang dùng vào thư mục <cd>-texture-dump/"
#define BEETLE_OPT_REPLACE_TEXTURES_LABEL_VN "Thay thế kết cấu"
#define BEETLE_OPT_REPLACE_TEXTURES_INFO_0_VN "Thay thế kết cấu bằng các phiên bản HD từ thư mục <cd>-texture-replacements/"
#define BEETLE_OPT_PGXP_MODE_LABEL_VN "Chế độ hoạt động PGXP"
#define BEETLE_OPT_PGXP_MODE_INFO_0_VN "Kết xuất đối tượng 3D với độ chính xác subpixel, giảm méo hình và rung (jitter) vốn xuất hiện trên phần cứng gốc do dùng tọa độ đỉnh fixed-point.Chế độ “Memory Only” có ít vấn đề tương thích và được khuyến nghị dùng chung.Chế độ “Memory + CPU (Buggy)” có thể giảm jitter hơn nữa nhưng yêu cầu hiệu năng rất cao và có thể gây nhiều lỗi hình học."
#define OPTION_VAL_MEMORY_ONLY_VN "Chỉ bộ nhớ"
#define OPTION_VAL_MEMORY_CPU_VN "Bộ nhớ + CPU (Có lỗi)"
#define BEETLE_OPT_PGXP_2D_TOL_LABEL_VN "Ngưỡng hình học 2D PGXP"
#define BEETLE_OPT_PGXP_2D_TOL_INFO_0_VN "Ẩn những lỗi hiển nhiên hơn trong hoạt động PGXP: giá trị này xác định độ dung sai mà các giá trị PGXP sẽ được giữ trong trường hợp hình học không có thông tin độ sâu phù hợp."
#define OPTION_VAL_0PX_VN NULL
#define OPTION_VAL_1PX_VN NULL
#define OPTION_VAL_2PX_VN NULL
#define OPTION_VAL_3PX_VN NULL
#define OPTION_VAL_4PX_VN NULL
#define OPTION_VAL_5PX_VN NULL
#define OPTION_VAL_6PX_VN NULL
#define OPTION_VAL_7PX_VN NULL
#define OPTION_VAL_8PX_VN NULL
#define BEETLE_OPT_PGXP_NCLIP_LABEL_VN "Loại bỏ hình nguyên PGXP"
#define BEETLE_OPT_PGXP_NCLIP_INFO_0_VN "Sử dụng triển khai NCLIP của PGXP. Cải thiện hình ảnh bằng cách giảm các lỗ trong hình học có tọa độ PGXP. Được biết có thể khiến một số trò chơi bị treo trong một số trường hợp."
#define BEETLE_OPT_PGXP_VERTEX_LABEL_VN "Bộ nhớ đệm đỉnh PGXP"
#define BEETLE_OPT_PGXP_VERTEX_INFO_0_VN "Lưu trữ tạm thời các vị trí đỉnh được nâng cấp bởi PGXP để tái sử dụng khi vẽ đa giác. Có thể cải thiện sự căn chỉnh đối tượng và giảm các đường nối hiển thị khi kết xuất kết cấu, nhưng kết quả sai khi truy vấn bộ nhớ đệm có thể gây ra lỗi đồ họa. Hiện tại khuyến nghị giữ tùy chọn này tắt. Tùy chọn này chỉ áp dụng khi chế độ hoạt động PGXP được bật. Chỉ được hỗ trợ bởi các bộ kết xuất phần cứng."
#define BEETLE_OPT_PGXP_TEXTURE_LABEL_VN "Kết cấu theo phối cảnh PGXP"
#define BEETLE_OPT_PGXP_TEXTURE_INFO_0_VN "Thay thế ánh xạ texture affine nguyên bản của PSX bằng ánh xạ texture hiệu chỉnh phối cảnh (perspective correct). Loại bỏ hiện tượng méo và biến dạng texture phụ thuộc vào vị trí, giúp texture được căn chỉnh chính xác hơn. Tùy chọn này chỉ được áp dụng khi bật chế độ PGXP Operation Mode. Được hỗ trợ bởi trình dựng phần cứng và cả trình dựng phần mềm."
#define BEETLE_OPT_DISPLAY_INTERNAL_FPS_LABEL_VN "Hiển thị FPS nội bộ"
#define BEETLE_OPT_DISPLAY_INTERNAL_FPS_INFO_0_VN "Hiển thị tốc độ khung hình nội bộ mà hệ thống PlayStation giả lập đang render nội dung. Lưu ý: Cần bật thông báo trên màn hình trong frontend libretro."
#define BEETLE_OPT_DISPLAY_OSD_LABEL_VN "Hiển thị thông báo OSD"
#define BEETLE_OPT_DISPLAY_OSD_INFO_0_VN "Hiển thị các thông báo OSD do core tạo ra."
#define BEETLE_OPT_DISPLAY_VRAM_LABEL_VN "Hiển thị toàn bộ VRAM (Gỡ lỗi)"
#define BEETLE_OPT_DISPLAY_VRAM_INFO_0_VN "Hiển thị toàn bộ VRAM của console giả lập. Chỉ được hỗ trợ bởi bộ xử lý đồ họa OpenGL và Vulkan. Lưu ý: Chế độ này chỉ dùng để gỡ lỗi và thường nên tắt."
#define BEETLE_OPT_ANALOG_CALIBRATION_LABEL_VN "Tự hiệu chuẩn Analog"
#define BEETLE_OPT_ANALOG_CALIBRATION_INFO_0_VN "Bật hiệu chỉnh động cho đầu vào analog khi thiết bị được đặt là DualShock, Analog Controller, Analog Joystick hoặc neGcon. Giá trị đầu vào tối đa được ghi nhận theo thời gian thực và dùng để scale tọa độ analog truyền vào giả lập. Nên dùng cho các game như Mega Man Legends 2 vốn yêu cầu giá trị lớn hơn so với tay cầm hiện đại. Để đạt kết quả tốt nhất, hãy xoay cần analog hết biên để tinh chỉnh thuật toán hiệu chỉnh mỗi khi tải nội dung."
#define BEETLE_OPT_ANALOG_TOGGLE_LABEL_VN "Chuyển chế độ Analog DualShock"
#define BEETLE_OPT_ANALOG_TOGGLE_INFO_0_VN "Khi loại thiết bị điều khiển là DualShock, tùy chọn này cho phép DualShock giả lập chuyển đổi giữa chế độ KỸ THUẬT SỐ và CHUẨN ANALOG như phần cứng gốc. Chế độ cũng có thể được chuyển bằng tổ hợp nút đã chọn."
#define OPTION_VAL_ENABLED_ANALOG_VN "Analog Mặc định"
#define BEETLE_OPT_ANALOG_TOGGLE_COMBO_LABEL_VN "Tổ hợp chế độ Analog của DualShock"
#define BEETLE_OPT_ANALOG_TOGGLE_COMBO_INFO_0_VN "Tổ hợp nút dùng để chuyển đổi giữa chế độ DIGITAL và ANALOG cho DualShock giả lập."
#define OPTION_VAL_L1_L2_R1_R2_START_SELECT_VN NULL
#define OPTION_VAL_L1_R1_SELECT_VN NULL
#define OPTION_VAL_L1_R1_START_VN NULL
#define OPTION_VAL_L1_R1_L3_VN NULL
#define OPTION_VAL_L1_R1_R3_VN NULL
#define OPTION_VAL_L2_R2_SELECT_VN NULL
#define OPTION_VAL_L2_R2_START_VN NULL
#define OPTION_VAL_L2_R2_L3_VN NULL
#define OPTION_VAL_L2_R2_R3_VN NULL
#define OPTION_VAL_L3_R3_VN NULL
#define BEETLE_OPT_ANALOG_TOGGLE_HOLD_LABEL_VN "Độ trễ giữ tổ hợp chế độ Analog của DualShock"
#define BEETLE_OPT_ANALOG_TOGGLE_HOLD_INFO_0_VN "Thời gian giữ tổ hợp nút chuyển chế độ analog."
#define OPTION_VAL_0_VN "Độ trễ 0 giây"
#define OPTION_VAL_1_VN "Trễ 1 giây"
#define OPTION_VAL_2_VN "Trễ 2 giây"
#define OPTION_VAL_3_VN "Trễ 3 giây"
#define OPTION_VAL_4_VN "Trễ 4 giây"
#define OPTION_VAL_5_VN "Trễ 5 giây"
#define BEETLE_OPT_ENABLE_MULTITAP_PORT1_LABEL_VN "Multitap ở Cổng 1"
#define BEETLE_OPT_ENABLE_MULTITAP_PORT1_INFO_0_VN "Bật chức năng đa tay cầm trên cổng 1."
#define BEETLE_OPT_ENABLE_MULTITAP_PORT2_LABEL_VN "Multitap ở Cổng 2"
#define BEETLE_OPT_ENABLE_MULTITAP_PORT2_INFO_0_VN "Bật chức năng đa tay cầm trên cổng 2."
#define BEETLE_OPT_GUN_INPUT_MODE_LABEL_VN "Chế độ nhập liệu súng"
#define BEETLE_OPT_GUN_INPUT_MODE_INFO_0_VN "Sử dụng chuột điều khiển “Light Gun” hoặc “Touchscreen” khi loại thiết bị được đặt là “Guncon/G-Con 45” hoặc “Justifier”."
#define OPTION_VAL_LIGHTGUN_VN "Súng ánh sáng"
#define OPTION_VAL_TOUCHSCREEN_VN "Màn hình cảm ứng"
#define BEETLE_OPT_GUN_CURSOR_LABEL_VN "Con trỏ súng"
#define BEETLE_OPT_GUN_CURSOR_INFO_0_VN "Vẽ con trỏ trên màn hình khi sử dụng các thiết bị nhập “Guncon/G-Con 45” và “Justifier”."
#define OPTION_VAL_CROSS_VN "Dấu cộng"
#define OPTION_VAL_DOT_VN "Dấu chấm"
#define BEETLE_OPT_CROSSHAIR_COLOR_P1_LABEL_VN "Màu tâm ngắm súng (Crosshair): Cổng 1"
#define BEETLE_OPT_CROSSHAIR_COLOR_P1_INFO_0_VN "Màu tâm ngắm súng ánh sáng cho cổng 1."
#define OPTION_VAL_RED_VN "Đỏ"
#define OPTION_VAL_BLUE_VN "Xanh dương"
#define OPTION_VAL_GREEN_VN "Xanh lá"
#define OPTION_VAL_ORANGE_VN "Cam"
#define OPTION_VAL_YELLOW_VN "Vàng"
#define OPTION_VAL_CYAN_VN "Lục lam"
#define OPTION_VAL_PINK_VN "Hồng"
#define OPTION_VAL_PURPLE_VN "Tím"
#define OPTION_VAL_BLACK_VN "Đen"
#define OPTION_VAL_WHITE_VN "Trắng"
#define BEETLE_OPT_CROSSHAIR_COLOR_P2_LABEL_VN "Màu tâm ngắm súng: Cổng 2"
#define BEETLE_OPT_CROSSHAIR_COLOR_P2_INFO_0_VN "Màu tâm ngắm súng ánh sáng cho cổng 2."
#define BEETLE_OPT_MOUSE_SENSITIVITY_LABEL_VN "Độ nhạy Chuột"
#define BEETLE_OPT_MOUSE_SENSITIVITY_INFO_0_VN "Độ nhạy phản hồi của loại thiết bị nhập Chuột."
#define OPTION_VAL_5_O42_VN NULL
#define OPTION_VAL_10_VN NULL
#define OPTION_VAL_15_VN NULL
#define OPTION_VAL_20_VN NULL
#define OPTION_VAL_25_VN NULL
#define OPTION_VAL_30_VN NULL
#define OPTION_VAL_35_VN NULL
#define OPTION_VAL_40_VN NULL
#define OPTION_VAL_45_VN NULL
#define OPTION_VAL_55_VN NULL
#define OPTION_VAL_65_VN NULL
#define OPTION_VAL_75_VN NULL
#define OPTION_VAL_85_VN NULL
#define OPTION_VAL_95_VN NULL
#define OPTION_VAL_100_O42_VN NULL
#define OPTION_VAL_105_VN NULL
#define OPTION_VAL_115_VN NULL
#define OPTION_VAL_125_VN NULL
#define OPTION_VAL_135_VN NULL
#define OPTION_VAL_145_VN NULL
#define OPTION_VAL_155_VN NULL
#define OPTION_VAL_165_VN NULL
#define OPTION_VAL_175_VN NULL
#define OPTION_VAL_185_VN NULL
#define OPTION_VAL_195_VN NULL
#define BEETLE_OPT_NEGCON_RESPONSE_LABEL_VN "phản hồi xoay neGcon"
#define BEETLE_OPT_NEGCON_RESPONSE_INFO_0_VN "Kiểu phản hồi của cần analog trái khi mô phỏng thao tác xoay (twist) của thiết bị nhập neGcon giả lập.Độ lệch của cần analog có thể được ánh xạ sang góc xoay của neGcon theo dạng tuyến tính, bậc hai hoặc bậc ba.Bậc hai cho độ chính xác cao hơn tuyến tính khi thực hiện các chuyển động nhỏ. Bậc ba còn tăng độ chính xác cho chuyển động nhỏ hơn nữa, nhưng sẽ khuếch đại quá mức các chuyển động lớn. Lưu ý: Tuyến tính chỉ nên dùng khi sử dụng vô lăng đua xe. Tay cầm thông thường xử lý analog theo cách hoàn toàn khác với cơ chế xoay của neGcon, nên ánh xạ tuyến tính sẽ làm khuếch đại quá mức các chuyển động nhỏ, gây khó kiểm soát chính xác. Trong hầu hết trường hợp, bậc hai cho cảm giác gần với phần cứng thật nhất."
#define OPTION_VAL_LINEAR_VN "Tuyến tính"
#define OPTION_VAL_QUADRATIC_VN "Bậc hai"
#define OPTION_VAL_CUBIC_VN "Bậc ba"
#define BEETLE_OPT_NEGCON_DEADZONE_LABEL_VN "vùng chết xoay neGcon"
#define BEETLE_OPT_NEGCON_DEADZONE_INFO_0_VN "Vùng chết của cần analog trái khi mô phỏng thao tác xoay của neGcon giả lập. Dùng để loại bỏ hiện tượng trôi cần.Lưu ý: Hầu hết các trò chơi hỗ trợ neGcon đều có tùy chọn trong game để đặt giá trị vùng chết cho thao tác xoay. Để tránh mất độ chính xác, vùng chết trong game luôn phải đặt về 0. Mọi điều chỉnh cần thiết chỉ nên thực hiện tại tùy chọn này của lõi giả lập. Điều này đặc biệt quan trọng khi kiểu phản hồi neGcon được đặt là bậc hai hoặc bậc ba."
#define OPTION_VAL_0_O44_VN NULL
#define BEETLE_OPT_USE_MEDNAFEN_MEMCARD0_METHOD_LABEL_VN "Phương thức thẻ nhớ"
#define BEETLE_OPT_USE_MEDNAFEN_MEMCARD0_METHOD_INFO_0_VN "Định dạng dữ liệu lưu dùng cho khe thẻ nhớ bên trái. Có thể dùng định dạng Mednafen để tương thích với phiên bản Mednafen độc lập. Khi sử dụng với Beetle PSX, dữ liệu lưu của Libretro và Mednafen có định dạng nội bộ giống hệt nhau và có thể chuyển đổi qua lại chỉ bằng cách đổi tên tệp. Cần khởi động lại."
#define OPTION_VAL_LIBRETRO_VN NULL
#define OPTION_VAL_MEDNAFEN_VN NULL
#define BEETLE_OPT_ENABLE_MEMCARD1_LABEL_VN "Khe thẻ nhớ 2"
#define BEETLE_OPT_ENABLE_MEMCARD1_INFO_0_VN "Giả lập thẻ nhớ thứ hai ở khe thẻ nhớ bên phải.Lưu ý: Một số trò chơi yêu cầu tắt tùy chọn này để hoạt động đúng (ví dụ: Codename Tenka).Lưu ý: Khe thẻ nhớ 2 sử dụng định dạng lưu của Mednafen. Cần khởi động lại."
#define BEETLE_OPT_SHARED_MEMORY_CARDS_LABEL_VN "Thẻ nhớ dùng chung"
#define BEETLE_OPT_SHARED_MEMORY_CARDS_INFO_0_VN "Sử dụng thẻ nhớ dùng chung thay vì mỗi trò chơi một thẻ riêng.Lưu ý: Nếu phương thức thẻ nhớ được đặt là Libretro, chỉ khe thẻ nhớ số 2 bị ảnh hưởng. Cần khởi động lại."
#define BEETLE_OPT_MEMCARD_LEFT_INDEX_LABEL_VN "Chỉ số khe thẻ nhớ 1"
#define BEETLE_OPT_MEMCARD_LEFT_INDEX_INFO_0_VN "Thẻ nhớ hiện đang được nạp ở khe bên trái. Tùy chọn này chỉ hoạt động khi Memory Card Method được đặt là Mednafen."
#define OPTION_VAL_0_O48_VN "0 (Mặc định)"
#define BEETLE_OPT_MEMCARD_RIGHT_INDEX_LABEL_VN "Chỉ số Khe Thẻ Nhớ 2"
#define BEETLE_OPT_MEMCARD_RIGHT_INDEX_INFO_0_VN "Thẻ nhớ hiện đang được nạp ở khe bên phải. Tùy chọn này chỉ hoạt động khi Memory Card Slot 2 được bật."
#define OPTION_VAL_1_O49_VN "1 (Mặc định)"
#define BEETLE_OPT_LINE_RENDER_LABEL_VN "Mẹo chuyển dòng thành hình tứ giác"
#define BEETLE_OPT_LINE_RENDER_INFO_0_VN "Một số game (ví dụ: Doom, Hexen, Soul Blade, v.v.) vẽ các đường ngang bằng cách kéo giãn các tam giác chỉ cao 1 pixel trên toàn màn hình, vốn trên phần cứng gốc sẽ được raster hóa thành một hàng pixel. Bản vá này phát hiện các tam giác nhỏ đó và chuyển chúng thành tứ giác (quad) khi cần, giúp hiển thị đúng trên các trình kết xuất phần cứng và ở độ phân giải nội bộ được upscale.Mức “Aggressive” là cần thiết cho một số tựa game (ví dụ: Dark Forces, Duke Nukem) nhưng có thể gây lỗi đồ họa ở các game khác. Nếu không chắc, hãy để “Mặc định”."
#define OPTION_VAL_DEFAULT_VN "Mặc định"
#define OPTION_VAL_AGGRESSIVE_VN "Mạnh"
#define BEETLE_OPT_WIDESCREEN_HACK_LABEL_VN "Mẹo Chế độ Màn hình Rộng"
#define BEETLE_OPT_WIDESCREEN_HACK_INFO_0_VN "Hiển thị nội dung 3D theo dạng anamorphic và xuất framebuffer giả lập ở tỷ lệ màn hình rộng. Cho kết quả tốt nhất với các trò chơi hoàn toàn 3D. Các yếu tố 2D sẽ bị kéo ngang và có thể không căn chỉnh đúng."
#define BEETLE_OPT_WIDESCREEN_HACK_ASPECT_RATIO_LABEL_VN "Tỷ lệ khung hình của Mẹo Chế độ Màn hình Rộng"
#define BEETLE_OPT_WIDESCREEN_HACK_ASPECT_RATIO_INFO_0_VN "Tỷ lệ khung hình được sử dụng cho bản vá Widescreen Mode Hack."
#define OPTION_VAL_16_9_VN NULL
#define OPTION_VAL_16_10_VN NULL
#define OPTION_VAL_18_9_VN NULL
#define OPTION_VAL_19_9_VN NULL
#define OPTION_VAL_20_9_VN NULL
#define OPTION_VAL_21_9_VN NULL
#define OPTION_VAL_32_9_VN NULL
#define BEETLE_OPT_CPU_DYNAREC_LABEL_VN "Biên dịch động CPU"
#define BEETLE_OPT_CPU_DYNAREC_INFO_0_VN "Biên dịch động các lệnh CPU thành lệnh gốc. Nhanh hơn nhiều so với thông dịch, nhưng thời gian CPU kém chính xác và có thể có lỗi."
#define OPTION_VAL_DISABLED_VN "Tắt (Thông dịch Beetle)"
#define OPTION_VAL_EXECUTE_VN "Hiệu năng tối đa"
#define OPTION_VAL_EXECUTE_ONE_VN "Kiểm tra thời gian chu kỳ"
#define OPTION_VAL_RUN_INTERPRETER_VN "Thông dịch Lightrec"
#define BEETLE_OPT_DYNAREC_INVALIDATE_LABEL_VN "Vô hiệu hóa mã Biên dịch động"
#define BEETLE_OPT_DYNAREC_INVALIDATE_INFO_0_VN "Một số trò chơi yêu cầu vô hiệu hóa ‘Toàn bộ’, một số yêu cầu ‘Chỉ DMA’."
#define OPTION_VAL_FULL_VN "Toàn bộ"
#define OPTION_VAL_DMA_VN "Chỉ DMA (Nhanh hơn một chút)"
#define BEETLE_OPT_DYNAREC_EVENTCYCLES_LABEL_VN "Chu kỳ sự kiện Dynarec DMA/GPU/MDEC/Bộ hẹn giờ"
#define BEETLE_OPT_DYNAREC_EVENTCYCLES_INFO_0_VN "Số chu kỳ tối đa CPU chạy trước khi kiểm tra cập nhật GPU/DMA/MDEC/Bộ hẹn giờ."
#define OPTION_VAL_128_VN "128 (Mặc định)"
#define BEETLE_OPT_DYNAREC_SPU_SAMPLES_LABEL_VN "Mẫu SPU Dynarec"
#define BEETLE_OPT_DYNAREC_SPU_SAMPLES_INFO_0_VN "Số mẫu SPU tối đa để chạy trước khi kiểm tra Cập nhật SPU, số cao hơn sẽ nhanh hơn, nhưng có thể gây lỗi âm thanh trong một số game nếu khác 1."
#define BEETLE_OPT_CORE_TIMING_FPS_LABEL_VN "Thời gian FPS do Core báo cáo"
#define BEETLE_OPT_CORE_TIMING_FPS_INFO_0_VN "Tự động chuyển đổi cho phép core chuyển giữa chế độ quét progressive và interlaced."
#define OPTION_VAL_AUTO_TOGGLE_VN "Tự động chuyển đổi"
#define OPTION_VAL_FORCE_PROGRESSIVE_VN "Buộc tốc độ quét liên tục"
#define OPTION_VAL_FORCE_INTERLACED_VN "Ép tốc độ xen kẽ"
#define BEETLE_OPT_DEINTERLACER_LABEL_VN "Phương pháp khử răng cưa"
#define BEETLE_OPT_DEINTERLACER_INFO_0_VN "‘Weave’ hiển thị cả hai trường trên cùng một khung hình để tạo ảnh tĩnh có độ phân giải đầy đủ, còn ‘Bob’ hy sinh độ nét độ phân giải để tăng độ mượt chuyển động. ‘Bob (Offset)’ là Bob nhưng nhân đôi trường hiện tại lên toàn bộ chiều cao hiển thị với độ lệch một dòng giữa các trường, giữ nguyên độ phân giải đầu ra libretro trong khi vẫn duy trì chuyển động mượt kiểu Bob. ‘FastMAD (Motion Adaptive)’ ghép các pixel tĩnh và nội suy theo chiều dọc các pixel đang chuyển động trên từng pixel, cho chiều cao đầu ra đầy đủ mà không gặp hiện tượng răng cưa khi chuyển động như weave hay giảm nửa độ phân giải như bob. ‘Off’ làm cho trình dựng phần mềm hoạt động giống trình dựng phần cứng (Vulkan/OpenGL) bằng cách rasterize lên tất cả các dòng VRAM mỗi khung hình và hoãn quét xuất hình đến cuối khung hình, loại bỏ hiện tượng răng cưa khi chuyển động trong game xen kẽ dòng (interlaced). Lưu ý: các chương trình DMA tới VRAM trong lúc quét xuất hình (một số demo dùng thủ thuật framebuffer-wrap PAL hơn 512 dòng), hoặc cập nhật sprite chuyển động qua truyền ảnh CPU-to-VRAM gắn với trường đang hiển thị (ví dụ Pro Pinball), có thể xuất hiện dòng hình cũ hoặc bóng mờ sprite khi dùng ‘Off’; hãy dùng ‘Weave’, ‘Bob’, ‘Bob (Offset)’ hoặc ‘FastMAD’ trong các trường hợp đó."
#define OPTION_VAL_WEAVE_VN NULL
#define OPTION_VAL_BOB_VN NULL
#define OPTION_VAL_BOB_OFFSET_VN "Bob (Lệch dòng)"
#define OPTION_VAL_FASTMAD_VN "FastMAD (Thích ứng chuyển động)"
#define BEETLE_OPT_FRAME_DUPING_LABEL_VN "Nhân bản khung hình"
#define BEETLE_OPT_FRAME_DUPING_INFO_0_VN "Lặp lại khung hình trước đó nếu lõi giả lập không có nội dung mới để hiển thị."
#define BEETLE_OPT_PAL_VIDEO_TIMING_OVERRIDE_LABEL_VN "Ghi đè thời gian video PAL"
#define BEETLE_OPT_PAL_VIDEO_TIMING_OVERRIDE_INFO_0_VN "Do sự khác biệt về tiêu chuẩn, các game PAL thường chạy chậm hơn so với các bản NTSC của Mỹ hoặc Nhật. Tùy chọn này cho phép ghi đè thời gian PAL để cố gắng chạy các game này với tốc độ khung hình NTSC."
#define BEETLE_OPT_ASPECT_RATIO_LABEL_VN "Tỷ lệ khung hình của Core"
#define BEETLE_OPT_ASPECT_RATIO_INFO_0_VN "Thiết lập này sẽ bị bỏ qua khi bản vá chế độ màn hình rộng hoặc tùy chọn hiển thị toàn bộ bộ nhớ video được bật."
#define OPTION_VAL_CORRECTED_VN "Đã hiệu chỉnh"
#define OPTION_VAL_UNCORRECTED_VN "Chưa hiệu chỉnh"
#define OPTION_VAL_4_3_VN "Ép 4:3"
#define OPTION_VAL_NTSC_VN "Ép NTSC"
#define BEETLE_OPT_CROP_OVERSCAN_LABEL_VN "Cắt vùng dư màn hình"
#define BEETLE_OPT_CROP_OVERSCAN_INFO_0_VN "Hiển thị toàn bộ hình ảnh gốc hoặc không có viền đen."
#define OPTION_VAL_STATIC_VN "Ngang"
#define OPTION_VAL_SMART_VN "Ngang + Dọc"
#define BEETLE_OPT_IMAGE_CROP_LABEL_VN "Cắt bổ sung"
#define BEETLE_OPT_IMAGE_CROP_INFO_0_VN "Giảm chiều rộng của hình ảnh đã cắt theo số lượng điểm ảnh được chỉ định."
#define OPTION_VAL_9PX_VN NULL
#define OPTION_VAL_10PX_VN NULL
#define OPTION_VAL_11PX_VN NULL
#define OPTION_VAL_12PX_VN NULL
#define OPTION_VAL_13PX_VN NULL
#define OPTION_VAL_14PX_VN NULL
#define OPTION_VAL_15PX_VN NULL
#define OPTION_VAL_16PX_VN NULL
#define OPTION_VAL_17PX_VN NULL
#define OPTION_VAL_18PX_VN NULL
#define OPTION_VAL_19PX_VN NULL
#define OPTION_VAL_20PX_VN NULL
#define BEETLE_OPT_IMAGE_OFFSET_LABEL_VN "Dịch ảnh bị cắt"
#define BEETLE_OPT_IMAGE_OFFSET_INFO_0_VN "Dịch chuyển hình ảnh theo chiều ngang sang phải (giá trị dương) hoặc sang trái (giá trị âm) theo số lượng điểm ảnh đã chỉ định. Có thể dùng để sửa các vấn đề lệch khung hình. Chỉ được hỗ trợ bởi trình kết xuất phần mềm."
#define OPTION_VAL_12PX_O64_VN NULL
#define OPTION_VAL_11PX_O64_VN NULL
#define OPTION_VAL_10PX_O64_VN NULL
#define OPTION_VAL_9PX_O64_VN NULL
#define OPTION_VAL_8PX_O64_VN NULL
#define OPTION_VAL_7PX_O64_VN NULL
#define OPTION_VAL_6PX_O64_VN NULL
#define OPTION_VAL_5PX_O64_VN NULL
#define OPTION_VAL_4PX_O64_VN NULL
#define OPTION_VAL_3PX_O64_VN NULL
#define OPTION_VAL_2PX_O64_VN NULL
#define OPTION_VAL_1PX_O64_VN NULL
#define OPTION_VAL_1PX_O64_0_VN NULL
#define OPTION_VAL_2PX_O64_0_VN NULL
#define OPTION_VAL_3PX_O64_0_VN NULL
#define OPTION_VAL_4PX_O64_0_VN NULL
#define OPTION_VAL_5PX_O64_0_VN NULL
#define OPTION_VAL_6PX_O64_0_VN NULL
#define OPTION_VAL_7PX_O64_0_VN NULL
#define OPTION_VAL_8PX_O64_0_VN NULL
#define OPTION_VAL_9PX_O64_0_VN NULL
#define OPTION_VAL_10PX_O64_0_VN NULL
#define OPTION_VAL_11PX_O64_0_VN NULL
#define OPTION_VAL_12PX_O64_0_VN NULL
#define BEETLE_OPT_IMAGE_OFFSET_CYCLES_LABEL_VN "Dịch ngang hình ảnh (Chu kỳ GPU)"
#define BEETLE_OPT_IMAGE_OFFSET_CYCLES_INFO_0_VN "Giá trị dương sẽ dịch hình ảnh sang bên phải, giá trị âm sẽ dịch hình ảnh sang bên trái.Tùy chọn này chỉ được hỗ trợ bởi các trình kết xuất phần cứng."
#define BEETLE_OPT_INITIAL_SCANLINE_LABEL_VN "Dòng quét đầu tiên - NTSC"
#define BEETLE_OPT_INITIAL_SCANLINE_INFO_0_VN "Dòng quét đầu tiên được hiển thị khi chạy nội dung NTSC.Các giá trị lớn hơn không sẽ làm giảm chiều cao hình ảnh đầu ra bằng cách cắt bớt các điểm ảnh ở mép trên cùng. Có thể dùng để khắc phục hiện tượng viền đen."
#define BEETLE_OPT_LAST_SCANLINE_LABEL_VN "Dòng quét cuối cùng - NTSC"
#define BEETLE_OPT_LAST_SCANLINE_INFO_0_VN "Dòng quét cuối cùng được hiển thị khi chạy nội dung NTSC.Các giá trị nhỏ hơn 239 sẽ làm giảm chiều cao hình ảnh đầu ra bằng cách cắt bớt các điểm ảnh ở mép dưới cùng. Có thể dùng để khắc phục hiện tượng viền đen (letterbox)."
#define OPTION_VAL_239_VN "239 (Mặc định)"
#define BEETLE_OPT_INITIAL_SCANLINE_PAL_LABEL_VN "Dòng quét đầu tiên - PAL"
#define BEETLE_OPT_INITIAL_SCANLINE_PAL_INFO_0_VN "Dòng quét đầu tiên được hiển thị khi chạy nội dung PAL.Các giá trị lớn hơn 0 sẽ làm giảm chiều cao hình ảnh đầu ra bằng cách cắt bớt các điểm ảnh ở mép trên cùng. Có thể dùng để khắc phục hiện tượng viền đen."
#define BEETLE_OPT_LAST_SCANLINE_PAL_LABEL_VN "Dòng quét cuối cùng - PAL"
#define BEETLE_OPT_LAST_SCANLINE_PAL_INFO_0_VN "Dòng quét cuối cùng được hiển thị khi chạy nội dung PAL.Các giá trị nhỏ hơn 287 sẽ làm giảm chiều cao hình ảnh đầu ra bằng cách cắt bớt các điểm ảnh ở mép dưới cùng. Có thể dùng để khắc phục hiện tượng viền đen."
#define OPTION_VAL_287_VN "287 (Mặc định)"

struct retro_core_option_v2_category option_cats_vn[] = {
   {
      "system",
      CATEGORY_SYSTEM_LABEL_VN,
      CATEGORY_SYSTEM_INFO_0_VN
   },
   {
      "video",
      CATEGORY_VIDEO_LABEL_VN,
      CATEGORY_VIDEO_INFO_0_VN
   },
   {
      "pgxp",
      CATEGORY_PGXP_LABEL_VN,
      CATEGORY_PGXP_INFO_0_VN
   },
   {
      "osd",
      CATEGORY_OSD_LABEL_VN,
      CATEGORY_OSD_INFO_0_VN
   },
   {
      "input",
      CATEGORY_INPUT_LABEL_VN,
      CATEGORY_INPUT_INFO_0_VN
   },
   {
      "memcards",
      CATEGORY_MEMCARDS_LABEL_VN,
      CATEGORY_MEMCARDS_INFO_0_VN
   },
   {
      "hacks",
      CATEGORY_HACKS_LABEL_VN,
      CATEGORY_HACKS_INFO_0_VN
   },
   { NULL, NULL, NULL },
};
struct retro_core_option_v2_definition option_defs_vn[] = {
   {
      BEETLE_OPT(cpu_freq_scale),
      BEETLE_OPT_CPU_FREQ_SCALE_LABEL_VN,
      NULL,
      BEETLE_OPT_CPU_FREQ_SCALE_INFO_0_VN,
      NULL,
      "system",
      {
         { "50%",           OPTION_VAL_50_VN },
         { "60%",           OPTION_VAL_60_VN },
         { "70%",           OPTION_VAL_70_VN },
         { "80%",           OPTION_VAL_80_VN },
         { "90%",           OPTION_VAL_90_VN },
         { "100%",          OPTION_VAL_100_VN },
         { "110%",          OPTION_VAL_110_VN },
         { "120%",          OPTION_VAL_120_VN },
         { "130%",          OPTION_VAL_130_VN },
         { "140%",          OPTION_VAL_140_VN },
         { "150%",          OPTION_VAL_150_VN },
         { "160%",          OPTION_VAL_160_VN },
         { "170%",          OPTION_VAL_170_VN },
         { "180%",          OPTION_VAL_180_VN },
         { "190%",          OPTION_VAL_190_VN },
         { "200%",          OPTION_VAL_200_VN },
         { "210%",          OPTION_VAL_210_VN },
         { "220%",          OPTION_VAL_220_VN },
         { "230%",          OPTION_VAL_230_VN },
         { "240%",          OPTION_VAL_240_VN },
         { "250%",          OPTION_VAL_250_VN },
         { "260%",          OPTION_VAL_260_VN },
         { "270%",          OPTION_VAL_270_VN },
         { "280%",          OPTION_VAL_280_VN },
         { "290%",          OPTION_VAL_290_VN },
         { "300%",          OPTION_VAL_300_VN },
         { "310%",          OPTION_VAL_310_VN },
         { "320%",          OPTION_VAL_320_VN },
         { "330%",          OPTION_VAL_330_VN },
         { "340%",          OPTION_VAL_340_VN },
         { "350%",          OPTION_VAL_350_VN },
         { "360%",          OPTION_VAL_360_VN },
         { "370%",          OPTION_VAL_370_VN },
         { "380%",          OPTION_VAL_380_VN },
         { "390%",          OPTION_VAL_390_VN },
         { "400%",          OPTION_VAL_400_VN },
         { "410%",          OPTION_VAL_410_VN },
         { "420%",          OPTION_VAL_420_VN },
         { "430%",          OPTION_VAL_430_VN },
         { "440%",          OPTION_VAL_440_VN },
         { "450%",          OPTION_VAL_450_VN },
         { "460%",          OPTION_VAL_460_VN },
         { "470%",          OPTION_VAL_470_VN },
         { "480%",          OPTION_VAL_480_VN },
         { "490%",          OPTION_VAL_490_VN },
         { "500%",          OPTION_VAL_500_VN },
         { "510%",          OPTION_VAL_510_VN },
         { "520%",          OPTION_VAL_520_VN },
         { "530%",          OPTION_VAL_530_VN },
         { "540%",          OPTION_VAL_540_VN },
         { "550%",          OPTION_VAL_550_VN },
         { "560%",          OPTION_VAL_560_VN },
         { "570%",          OPTION_VAL_570_VN },
         { "580%",          OPTION_VAL_580_VN },
         { "590%",          OPTION_VAL_590_VN },
         { "600%",          OPTION_VAL_600_VN },
         { "610%",          OPTION_VAL_610_VN },
         { "620%",          OPTION_VAL_620_VN },
         { "630%",          OPTION_VAL_630_VN },
         { "640%",          OPTION_VAL_640_VN },
         { "650%",          OPTION_VAL_650_VN },
         { "660%",          OPTION_VAL_660_VN },
         { "670%",          OPTION_VAL_670_VN },
         { "680%",          OPTION_VAL_680_VN },
         { "690%",          OPTION_VAL_690_VN },
         { "700%",          OPTION_VAL_700_VN },
         { "710%",          OPTION_VAL_710_VN },
         { "720%",          OPTION_VAL_720_VN },
         { "730%",          OPTION_VAL_730_VN },
         { "740%",          OPTION_VAL_740_VN },
         { "750%",          OPTION_VAL_750_VN },
         { NULL, NULL },
      },
      "100%"
   },
   {
      BEETLE_OPT(gpu_overclock),
      BEETLE_OPT_GPU_OVERCLOCK_LABEL_VN,
      NULL,
      BEETLE_OPT_GPU_OVERCLOCK_INFO_0_VN,
      NULL,
      "system",
      {
         { "1x(native)", OPTION_VAL_1X_NATIVE_VN },
         { "2x",         OPTION_VAL_2X_VN },
         { "4x",         OPTION_VAL_4X_VN },
         { "8x",         OPTION_VAL_8X_VN },
         { "16x",        OPTION_VAL_16X_VN },
         { "32x",        OPTION_VAL_32X_VN },
         { NULL, NULL },
      },
      "1x(native)"
   },
   {
      BEETLE_OPT(gte_overclock),
      BEETLE_OPT_GTE_OVERCLOCK_LABEL_VN,
      NULL,
      BEETLE_OPT_GTE_OVERCLOCK_INFO_0_VN,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
#ifndef EMSCRIPTEN
   {
      BEETLE_OPT(cd_access_method),
      BEETLE_OPT_CD_ACCESS_METHOD_LABEL_VN,
      NULL,
      BEETLE_OPT_CD_ACCESS_METHOD_INFO_0_VN,
      NULL,
      "system",
      {
         { "sync",     OPTION_VAL_SYNC_VN },
         { "async",    OPTION_VAL_ASYNC_VN },
         { "precache", OPTION_VAL_PRECACHE_VN },
         { NULL, NULL },
      },
      "sync"
   },
#endif
   {
      BEETLE_OPT(cd_fastload),
      BEETLE_OPT_CD_FASTLOAD_LABEL_VN,
      NULL,
      BEETLE_OPT_CD_FASTLOAD_INFO_0_VN,
      NULL,
      "system",
      {
         { "2x(native)", OPTION_VAL_2X_NATIVE_VN },
         { "4x",          OPTION_VAL_4X_VN },
         { "6x",          OPTION_VAL_6X_VN },
         { "8x",          OPTION_VAL_8X_VN },
         { "10x",         OPTION_VAL_10X_VN },
         { "12x",         OPTION_VAL_12X_VN },
         { "14x",         OPTION_VAL_14X_VN },
         { NULL, NULL },
      },
      "2x(native)"
   },
   {
      BEETLE_OPT(skip_bios),
      BEETLE_OPT_SKIP_BIOS_LABEL_VN,
      NULL,
      BEETLE_OPT_SKIP_BIOS_INFO_0_VN,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(override_bios),
      BEETLE_OPT_OVERRIDE_BIOS_LABEL_VN,
      NULL,
      BEETLE_OPT_OVERRIDE_BIOS_INFO_0_VN,
      NULL,
      "system",
      {
         { "disabled", NULL },
         { "psxonpsp", OPTION_VAL_PSXONPSP_VN },
         { "ps1_rom",  OPTION_VAL_PS1_ROM_VN },
         { "openbios", OPTION_VAL_OPENBIOS_VN     },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(region),
      BEETLE_OPT_REGION_LABEL_VN,
      NULL,
      BEETLE_OPT_REGION_INFO_0_VN,
      NULL,
      "system",
      {
         { "auto",   OPTION_VAL_AUTO_VN },
         { "ntsc-j", OPTION_VAL_NTSC_J_VN },
         { "ntsc-u", OPTION_VAL_NTSC_U_VN },
         { "pal",    OPTION_VAL_PAL_VN },
         { NULL, NULL },
      },
      "auto"
   },
   {
      BEETLE_OPT(internal_resolution),
      BEETLE_OPT_INTERNAL_RESOLUTION_LABEL_VN,
      NULL,
      BEETLE_OPT_INTERNAL_RESOLUTION_INFO_0_VN,
      NULL,
      "video",
      {
         { "1x(native)", OPTION_VAL_1X_NATIVE_VN },
         { "2x",         OPTION_VAL_2X_VN },
         { "4x",         OPTION_VAL_4X_VN },
         { "8x",         OPTION_VAL_8X_VN },
         { "16x",        OPTION_VAL_16X_VN },
         { NULL, NULL },
      },
      "1x(native)"
   },
#if defined(HAVE_OPENGL) || defined(HAVE_OPENGLES) || defined(HAVE_VULKAN)
   {
      BEETLE_OPT(renderer),
      BEETLE_OPT_RENDERER_LABEL_VN,
      NULL,
      BEETLE_OPT_RENDERER_INFO_0_VN,
      NULL,
      "video",
      {
         { "hardware",    OPTION_VAL_HARDWARE_VN },
#if defined(HAVE_OPENGL) || defined(HAVE_OPENGLES)
         { "hardware_gl", OPTION_VAL_HARDWARE_GL_VN },
#endif
#if defined(HAVE_VULKAN)
         { "hardware_vk", OPTION_VAL_HARDWARE_VK_VN },
#endif
         { "software",    OPTION_VAL_SOFTWARE_VN },
         { NULL, NULL },
      },
      "hardware"
   },
   {
      BEETLE_OPT(renderer_software_fb),
      BEETLE_OPT_RENDERER_SOFTWARE_FB_LABEL_VN,
      NULL,
      BEETLE_OPT_RENDERER_SOFTWARE_FB_INFO_0_VN,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled"
   },
#endif
#if defined(HAVE_OPENGL) || defined(HAVE_OPENGLES)
   {
      BEETLE_OPT(depth),
      BEETLE_OPT_DEPTH_LABEL_VN,
      NULL,
      BEETLE_OPT_DEPTH_INFO_0_VN,
      NULL,
      "video",
      {
         { "16bpp(native)", OPTION_VAL_16BPP_NATIVE_VN },
         { "32bpp",         OPTION_VAL_32BPP_VN },
         { NULL, NULL },
      },
      "16bpp(native)"
   },
   // Sort of, it's more like 15-bit high color and 24-bit true color for visible output. The alpha channel is used for mask bit. Vulkan renderer uses ABGR1555_555 for 31 bits internal? FMVs are always 24-bit on all renderers like original hardware (BGR888, no alpha)
#endif
   {
      BEETLE_OPT(dither_mode),
      BEETLE_OPT_DITHER_MODE_LABEL_VN,
      NULL,
      BEETLE_OPT_DITHER_MODE_INFO_0_VN,
      NULL,
      "video",
      {
         { "1x(native)",          OPTION_VAL_1X_NATIVE_VN },
         { "internal resolution", OPTION_VAL_INTERNAL_RESOLUTION_VN },
         { "disabled",            NULL },
         { NULL, NULL },
      },
      "1x(native)"
   },
#ifdef HAVE_VULKAN
   {
      BEETLE_OPT(scaled_uv_offset),
      BEETLE_OPT_SCALED_UV_OFFSET_LABEL_VN,
      NULL,
      BEETLE_OPT_SCALED_UV_OFFSET_INFO_0_VN,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled"
   },
#endif
#if defined(HAVE_OPENGL) || defined(HAVE_OPENGLES) || defined(HAVE_VULKAN)
   {
      BEETLE_OPT(filter),
      BEETLE_OPT_FILTER_LABEL_VN,
      NULL,
      BEETLE_OPT_FILTER_INFO_0_VN,
      NULL,
      "video",
      {
         { "nearest",  OPTION_VAL_NEAREST_VN },
         { "SABR",     OPTION_VAL_SABR_VN },
         { "xBR",      OPTION_VAL_XBR_VN },
         { "bilinear", OPTION_VAL_BILINEAR_VN },
         { "3-point",  OPTION_VAL_3_POINT_VN },
         { "JINC2",    OPTION_VAL_JINC2_VN },
         { NULL, NULL },
      },
      "nearest"
   },
#ifdef HAVE_VULKAN
   {
      BEETLE_OPT(filter_exclude_sprite),
      BEETLE_OPT_FILTER_EXCLUDE_SPRITE_LABEL_VN,
      NULL,
      BEETLE_OPT_FILTER_EXCLUDE_SPRITE_INFO_0_VN,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "opaque", OPTION_VAL_OPAQUE_VN },
         { "all", OPTION_VAL_ALL_VN },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(filter_exclude_2d_polygon),
      BEETLE_OPT_FILTER_EXCLUDE_2D_POLYGON_LABEL_VN,
      NULL,
      BEETLE_OPT_FILTER_EXCLUDE_2D_POLYGON_INFO_0_VN,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "opaque", OPTION_VAL_OPAQUE_VN },
         { "all", OPTION_VAL_ALL_VN },
         { NULL, NULL },
      },
      "disabled"
   },
#endif
#endif
#ifdef HAVE_VULKAN
   {
      BEETLE_OPT(adaptive_smoothing),
      BEETLE_OPT_ADAPTIVE_SMOOTHING_LABEL_VN,
      NULL,
      BEETLE_OPT_ADAPTIVE_SMOOTHING_INFO_0_VN,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(super_sampling),
      BEETLE_OPT_SUPER_SAMPLING_LABEL_VN,
      NULL,
      BEETLE_OPT_SUPER_SAMPLING_INFO_0_VN,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(msaa),
      BEETLE_OPT_MSAA_LABEL_VN,
      NULL,
      BEETLE_OPT_MSAA_INFO_0_VN,
      NULL,
      "video",
      {
         { "1x",  OPTION_VAL_1X_VN },
         { "2x",  OPTION_VAL_2X_VN },
         { "4x",  OPTION_VAL_4X_VN },
         { "8x",  OPTION_VAL_8X_VN },
         { "16x", OPTION_VAL_16X_VN },
         { NULL, NULL },
      },
      "1x"
   },
   {
      BEETLE_OPT(mdec_yuv),
      BEETLE_OPT_MDEC_YUV_LABEL_VN,
      NULL,
      BEETLE_OPT_MDEC_YUV_INFO_0_VN,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(track_textures),
      BEETLE_OPT_TRACK_TEXTURES_LABEL_VN,
      NULL,
      BEETLE_OPT_TRACK_TEXTURES_INFO_0_VN,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
#ifdef TEXTURE_DUMPING_ENABLED
   {
      BEETLE_OPT(dump_textures),
      BEETLE_OPT_DUMP_TEXTURES_LABEL_VN,
      NULL,
      BEETLE_OPT_DUMP_TEXTURES_INFO_0_VN,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
#endif
   {
      BEETLE_OPT(replace_textures),
      BEETLE_OPT_REPLACE_TEXTURES_LABEL_VN,
      NULL,
      BEETLE_OPT_REPLACE_TEXTURES_INFO_0_VN,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
#endif
   {
      BEETLE_OPT(pgxp_mode),
      BEETLE_OPT_PGXP_MODE_LABEL_VN,
      NULL,
      BEETLE_OPT_PGXP_MODE_INFO_0_VN,
      NULL,
      "pgxp",
      {
         { "disabled",     NULL },
         { "memory only",  OPTION_VAL_MEMORY_ONLY_VN },
         { "memory + CPU", OPTION_VAL_MEMORY_CPU_VN },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(pgxp_2d_tol),
      BEETLE_OPT_PGXP_2D_TOL_LABEL_VN,
      NULL,
      BEETLE_OPT_PGXP_2D_TOL_INFO_0_VN,
      NULL,
      "pgxp",
      {
         { "disabled", NULL },
         { "0px", OPTION_VAL_0PX_VN },
         { "1px", OPTION_VAL_1PX_VN },
         { "2px", OPTION_VAL_2PX_VN },
         { "3px", OPTION_VAL_3PX_VN },
         { "4px", OPTION_VAL_4PX_VN },
         { "5px", OPTION_VAL_5PX_VN },
         { "6px", OPTION_VAL_6PX_VN },
         { "7px", OPTION_VAL_7PX_VN },
         { "8px", OPTION_VAL_8PX_VN },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(pgxp_nclip),
      BEETLE_OPT_PGXP_NCLIP_LABEL_VN,
      NULL,
      BEETLE_OPT_PGXP_NCLIP_INFO_0_VN,
      NULL,
      "pgxp",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
#if defined(HAVE_OPENGL) || defined(HAVE_OPENGLES) || defined(HAVE_VULKAN)
   {
      BEETLE_OPT(pgxp_vertex),
      BEETLE_OPT_PGXP_VERTEX_LABEL_VN,
      NULL,
      BEETLE_OPT_PGXP_VERTEX_INFO_0_VN,
      NULL,
      "pgxp",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
#endif
   {
      BEETLE_OPT(pgxp_texture),
      BEETLE_OPT_PGXP_TEXTURE_LABEL_VN,
      NULL,
      BEETLE_OPT_PGXP_TEXTURE_INFO_0_VN,
      NULL,
      "pgxp",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(display_internal_fps),
      BEETLE_OPT_DISPLAY_INTERNAL_FPS_LABEL_VN,
      NULL,
      BEETLE_OPT_DISPLAY_INTERNAL_FPS_INFO_0_VN,
      NULL,
      "osd",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(display_osd),
      BEETLE_OPT_DISPLAY_OSD_LABEL_VN,
      NULL,
      BEETLE_OPT_DISPLAY_OSD_INFO_0_VN,
      NULL,
      "osd",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "enabled"
   },
#if defined(HAVE_OPENGL) || defined(HAVE_OPENGLES) || defined(HAVE_VULKAN)
   {
      BEETLE_OPT(display_vram),
      BEETLE_OPT_DISPLAY_VRAM_LABEL_VN,
      NULL,
      BEETLE_OPT_DISPLAY_VRAM_INFO_0_VN,
      NULL,
      "osd",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
#endif
   {
      BEETLE_OPT(analog_calibration),
      BEETLE_OPT_ANALOG_CALIBRATION_LABEL_VN,
      NULL,
      BEETLE_OPT_ANALOG_CALIBRATION_INFO_0_VN,
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(analog_toggle),
      BEETLE_OPT_ANALOG_TOGGLE_LABEL_VN,
      NULL,
      BEETLE_OPT_ANALOG_TOGGLE_INFO_0_VN,
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { "enabled-analog", OPTION_VAL_ENABLED_ANALOG_VN },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(analog_toggle_combo),
      BEETLE_OPT_ANALOG_TOGGLE_COMBO_LABEL_VN,
      NULL,
      BEETLE_OPT_ANALOG_TOGGLE_COMBO_INFO_0_VN,
      NULL,
      "input",
      {
         { "l1+l2+r1+r2+start+select", OPTION_VAL_L1_L2_R1_R2_START_SELECT_VN },
         { "l1+r1+select",             OPTION_VAL_L1_R1_SELECT_VN },
         { "l1+r1+start",              OPTION_VAL_L1_R1_START_VN },
         { "l1+r1+l3",                 OPTION_VAL_L1_R1_L3_VN },
         { "l1+r1+r3",                 OPTION_VAL_L1_R1_R3_VN },
         { "l2+r2+select",             OPTION_VAL_L2_R2_SELECT_VN },
         { "l2+r2+start",              OPTION_VAL_L2_R2_START_VN },
         { "l2+r2+l3",                 OPTION_VAL_L2_R2_L3_VN },
         { "l2+r2+r3",                 OPTION_VAL_L2_R2_R3_VN },
         { "l3+r3",                    OPTION_VAL_L3_R3_VN },
         { NULL, NULL },
      },
      "l1+r1+select"
   },
   {
      BEETLE_OPT(analog_toggle_hold),
      BEETLE_OPT_ANALOG_TOGGLE_HOLD_LABEL_VN,
      NULL,
      BEETLE_OPT_ANALOG_TOGGLE_HOLD_INFO_0_VN,
      NULL,
      "input",
      {
         { "0", OPTION_VAL_0_VN },
         { "1", OPTION_VAL_1_VN },
         { "2", OPTION_VAL_2_VN },
         { "3", OPTION_VAL_3_VN },
         { "4", OPTION_VAL_4_VN },
         { "5", OPTION_VAL_5_VN },
         { NULL, NULL },
      },
      "1"
   },
   {
      BEETLE_OPT(enable_multitap_port1),
      BEETLE_OPT_ENABLE_MULTITAP_PORT1_LABEL_VN,
      NULL,
      BEETLE_OPT_ENABLE_MULTITAP_PORT1_INFO_0_VN,
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(enable_multitap_port2),
      BEETLE_OPT_ENABLE_MULTITAP_PORT2_LABEL_VN,
      NULL,
      BEETLE_OPT_ENABLE_MULTITAP_PORT2_INFO_0_VN,
      NULL,
      "input",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(gun_input_mode),
      BEETLE_OPT_GUN_INPUT_MODE_LABEL_VN,
      NULL,
      BEETLE_OPT_GUN_INPUT_MODE_INFO_0_VN,
      NULL,
      "input",
      {
         { "lightgun",    OPTION_VAL_LIGHTGUN_VN },
         { "touchscreen", OPTION_VAL_TOUCHSCREEN_VN },
         { NULL, NULL },
      },
      "lightgun"
   },
   // Shouldn't the gun_input_mode just be Mouse vs. Touchscreen?
   {
      BEETLE_OPT(gun_cursor),
      BEETLE_OPT_GUN_CURSOR_LABEL_VN,
      NULL,
      BEETLE_OPT_GUN_CURSOR_INFO_0_VN,
      NULL,
      "input",
      {
         { "off",   "disabled" },
         { "cross", OPTION_VAL_CROSS_VN },
         { "dot",   OPTION_VAL_DOT_VN },
         { NULL, NULL },
      },
      "cross"
   },
   {
      BEETLE_OPT(crosshair_color_p1),
      BEETLE_OPT_CROSSHAIR_COLOR_P1_LABEL_VN,
      NULL,
      BEETLE_OPT_CROSSHAIR_COLOR_P1_INFO_0_VN,
      NULL,
      "input",
      {
         { "red",    OPTION_VAL_RED_VN },
         { "blue",   OPTION_VAL_BLUE_VN },
         { "green",  OPTION_VAL_GREEN_VN },
         { "orange", OPTION_VAL_ORANGE_VN },
         { "yellow", OPTION_VAL_YELLOW_VN },
         { "cyan",   OPTION_VAL_CYAN_VN },
         { "pink",   OPTION_VAL_PINK_VN },
         { "purple", OPTION_VAL_PURPLE_VN },
         { "black",  OPTION_VAL_BLACK_VN },
         { "white",  OPTION_VAL_WHITE_VN },
         { NULL, NULL },
      },
      "red"
   },
   {
      BEETLE_OPT(crosshair_color_p2),
      BEETLE_OPT_CROSSHAIR_COLOR_P2_LABEL_VN,
      NULL,
      BEETLE_OPT_CROSSHAIR_COLOR_P2_INFO_0_VN,
      NULL,
      "input",
      {
         { "blue",   OPTION_VAL_BLUE_VN },
         { "red",    OPTION_VAL_RED_VN },
         { "green",  OPTION_VAL_GREEN_VN },
         { "orange", OPTION_VAL_ORANGE_VN },
         { "yellow", OPTION_VAL_YELLOW_VN },
         { "cyan",   OPTION_VAL_CYAN_VN },
         { "pink",   OPTION_VAL_PINK_VN },
         { "purple", OPTION_VAL_PURPLE_VN },
         { "black",  OPTION_VAL_BLACK_VN },
         { "white",  OPTION_VAL_WHITE_VN },
         { NULL, NULL },
      },
      "blue"
   },
   {
      BEETLE_OPT(mouse_sensitivity),
      BEETLE_OPT_MOUSE_SENSITIVITY_LABEL_VN,
      NULL,
      BEETLE_OPT_MOUSE_SENSITIVITY_INFO_0_VN,
      NULL,
      "input",
      {
         { "5%",   OPTION_VAL_5_O42_VN },
         { "10%",  OPTION_VAL_10_VN },
         { "15%",  OPTION_VAL_15_VN },
         { "20%",  OPTION_VAL_20_VN },
         { "25%",  OPTION_VAL_25_VN },
         { "30%",  OPTION_VAL_30_VN },
         { "35%",  OPTION_VAL_35_VN },
         { "40%",  OPTION_VAL_40_VN },
         { "45%",  OPTION_VAL_45_VN },
         { "50%",  OPTION_VAL_50_VN },
         { "55%",  OPTION_VAL_55_VN },
         { "60%",  OPTION_VAL_60_VN },
         { "65%",  OPTION_VAL_65_VN },
         { "70%",  OPTION_VAL_70_VN },
         { "75%",  OPTION_VAL_75_VN },
         { "80%",  OPTION_VAL_80_VN },
         { "85%",  OPTION_VAL_85_VN },
         { "90%",  OPTION_VAL_90_VN },
         { "95%",  OPTION_VAL_95_VN },
         { "100%", OPTION_VAL_100_O42_VN },
         { "105%", OPTION_VAL_105_VN },
         { "110%", OPTION_VAL_110_VN },
         { "115%", OPTION_VAL_115_VN },
         { "120%", OPTION_VAL_120_VN },
         { "125%", OPTION_VAL_125_VN },
         { "130%", OPTION_VAL_130_VN },
         { "135%", OPTION_VAL_135_VN },
         { "140%", OPTION_VAL_140_VN },
         { "145%", OPTION_VAL_145_VN },
         { "150%", OPTION_VAL_150_VN },
         { "155%", OPTION_VAL_155_VN },
         { "160%", OPTION_VAL_160_VN },
         { "165%", OPTION_VAL_165_VN },
         { "170%", OPTION_VAL_170_VN },
         { "175%", OPTION_VAL_175_VN },
         { "180%", OPTION_VAL_180_VN },
         { "185%", OPTION_VAL_185_VN },
         { "190%", OPTION_VAL_190_VN },
         { "195%", OPTION_VAL_195_VN },
         { "200%", OPTION_VAL_200_VN },
         { NULL, NULL },
      },
      "100%"
   },
   {
      BEETLE_OPT(negcon_response),
      BEETLE_OPT_NEGCON_RESPONSE_LABEL_VN,
      NULL,
      BEETLE_OPT_NEGCON_RESPONSE_INFO_0_VN,
      NULL,
      "input",
      {
         { "linear",    OPTION_VAL_LINEAR_VN },
         { "quadratic", OPTION_VAL_QUADRATIC_VN },
         { "cubic",     OPTION_VAL_CUBIC_VN },
         { NULL, NULL },
      },
      "linear"
   },
   {
      BEETLE_OPT(negcon_deadzone),
      BEETLE_OPT_NEGCON_DEADZONE_LABEL_VN,
      NULL,
      BEETLE_OPT_NEGCON_DEADZONE_INFO_0_VN,
      NULL,
      "input",
      {
         { "0%",  OPTION_VAL_0_O44_VN },
         { "5%",  OPTION_VAL_5_O42_VN },
         { "10%", OPTION_VAL_10_VN },
         { "15%", OPTION_VAL_15_VN },
         { "20%", OPTION_VAL_20_VN },
         { "25%", OPTION_VAL_25_VN },
         { "30%", OPTION_VAL_30_VN },
         { NULL, NULL },
      },
      "0%"
   },
   {
      BEETLE_OPT(use_mednafen_memcard0_method),
      BEETLE_OPT_USE_MEDNAFEN_MEMCARD0_METHOD_LABEL_VN,
      NULL,
      BEETLE_OPT_USE_MEDNAFEN_MEMCARD0_METHOD_INFO_0_VN,
      NULL,
      "memcards",
      {
         { "libretro", OPTION_VAL_LIBRETRO_VN },
         { "mednafen", OPTION_VAL_MEDNAFEN_VN },
         { NULL, NULL },
      },
      "libretro"
   },
   {
      BEETLE_OPT(enable_memcard1),
      BEETLE_OPT_ENABLE_MEMCARD1_LABEL_VN,
      NULL,
      BEETLE_OPT_ENABLE_MEMCARD1_INFO_0_VN,
      NULL,
      "memcards",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(shared_memory_cards),
      BEETLE_OPT_SHARED_MEMORY_CARDS_LABEL_VN,
      NULL,
      BEETLE_OPT_SHARED_MEMORY_CARDS_INFO_0_VN,
      NULL,
      "memcards",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(memcard_left_index),
      BEETLE_OPT_MEMCARD_LEFT_INDEX_LABEL_VN,
      NULL,
      BEETLE_OPT_MEMCARD_LEFT_INDEX_INFO_0_VN,
      NULL,
      "memcards",
      {
         { "0",  OPTION_VAL_0_O48_VN },
         { "1",  NULL },
         { "2",  NULL },
         { "3",  NULL },
         { "4",  NULL },
         { "5",  NULL },
         { "6",  NULL },
         { "7",  NULL },
         { "8",  NULL },
         { "9",  NULL },
         { "10", NULL },
         { "11", NULL },
         { "12", NULL },
         { "13", NULL },
         { "14", NULL },
         { "15", NULL },
         { "16", NULL },
         { "17", NULL },
         { "18", NULL },
         { "19", NULL },
         { "20", NULL },
         { "21", NULL },
         { "22", NULL },
         { "23", NULL },
         { "24", NULL },
         { "25", NULL },
         { "26", NULL },
         { "27", NULL },
         { "28", NULL },
         { "29", NULL },
         { "30", NULL },
         { "31", NULL },
         { "32", NULL },
         { "33", NULL },
         { "34", NULL },
         { "35", NULL },
         { "36", NULL },
         { "37", NULL },
         { "38", NULL },
         { "39", NULL },
         { "40", NULL },
         { "41", NULL },
         { "42", NULL },
         { "43", NULL },
         { "44", NULL },
         { "45", NULL },
         { "46", NULL },
         { "47", NULL },
         { "48", NULL },
         { "49", NULL },
         { "50", NULL },
         { "51", NULL },
         { "52", NULL },
         { "53", NULL },
         { "54", NULL },
         { "55", NULL },
         { "56", NULL },
         { "57", NULL },
         { "58", NULL },
         { "59", NULL },
         { "60", NULL },
         { "61", NULL },
         { "62", NULL },
         { "63", NULL },
         { NULL, NULL },
      },
      "0"
   },
   {
      BEETLE_OPT(memcard_right_index),
      BEETLE_OPT_MEMCARD_RIGHT_INDEX_LABEL_VN,
      NULL,
      BEETLE_OPT_MEMCARD_RIGHT_INDEX_INFO_0_VN,
      NULL,
      "memcards",
      {
         { "0",  NULL },
         { "1",  OPTION_VAL_1_O49_VN },
         { "2",  NULL },
         { "3",  NULL },
         { "4",  NULL },
         { "5",  NULL },
         { "6",  NULL },
         { "7",  NULL },
         { "8",  NULL },
         { "9",  NULL },
         { "10", NULL },
         { "11", NULL },
         { "12", NULL },
         { "13", NULL },
         { "14", NULL },
         { "15", NULL },
         { "16", NULL },
         { "17", NULL },
         { "18", NULL },
         { "19", NULL },
         { "20", NULL },
         { "21", NULL },
         { "22", NULL },
         { "23", NULL },
         { "24", NULL },
         { "25", NULL },
         { "26", NULL },
         { "27", NULL },
         { "28", NULL },
         { "29", NULL },
         { "30", NULL },
         { "31", NULL },
         { "32", NULL },
         { "33", NULL },
         { "34", NULL },
         { "35", NULL },
         { "36", NULL },
         { "37", NULL },
         { "38", NULL },
         { "39", NULL },
         { "40", NULL },
         { "41", NULL },
         { "42", NULL },
         { "43", NULL },
         { "44", NULL },
         { "45", NULL },
         { "46", NULL },
         { "47", NULL },
         { "48", NULL },
         { "49", NULL },
         { "50", NULL },
         { "51", NULL },
         { "52", NULL },
         { "53", NULL },
         { "54", NULL },
         { "55", NULL },
         { "56", NULL },
         { "57", NULL },
         { "58", NULL },
         { "59", NULL },
         { "60", NULL },
         { "61", NULL },
         { "62", NULL },
         { "63", NULL },
         { NULL, NULL },
      },
      "1"
   },
   {
      BEETLE_OPT(line_render),
      BEETLE_OPT_LINE_RENDER_LABEL_VN,
      NULL,
      BEETLE_OPT_LINE_RENDER_INFO_0_VN,
      NULL,
      "hacks",
      {
         { "default",    OPTION_VAL_DEFAULT_VN },
         { "aggressive", OPTION_VAL_AGGRESSIVE_VN },
         { "disabled",   NULL },
         { NULL, NULL },
      },
      "default"
   },
   {
      BEETLE_OPT(widescreen_hack),
      BEETLE_OPT_WIDESCREEN_HACK_LABEL_VN,
      NULL,
      BEETLE_OPT_WIDESCREEN_HACK_INFO_0_VN,
      NULL,
      "hacks",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(widescreen_hack_aspect_ratio),
      BEETLE_OPT_WIDESCREEN_HACK_ASPECT_RATIO_LABEL_VN,
      NULL,
      BEETLE_OPT_WIDESCREEN_HACK_ASPECT_RATIO_INFO_0_VN,
      NULL,
      "hacks",
      {
         { "16:9",  OPTION_VAL_16_9_VN },
         { "16:10", OPTION_VAL_16_10_VN },
         { "18:9",  OPTION_VAL_18_9_VN },
         { "19:9",  OPTION_VAL_19_9_VN },
         { "20:9",  OPTION_VAL_20_9_VN },
         { "21:9",  OPTION_VAL_21_9_VN }, // 64:27
         { "32:9",  OPTION_VAL_32_9_VN },
         { NULL,    NULL },
      },
      "16:9"
   },
#if defined(HAVE_LIGHTREC)
   {
      BEETLE_OPT(cpu_dynarec),
      BEETLE_OPT_CPU_DYNAREC_LABEL_VN,
      NULL,
      BEETLE_OPT_CPU_DYNAREC_INFO_0_VN,
      NULL,
      "hacks",
      {
         { "disabled",        OPTION_VAL_DISABLED_VN },
         { "execute",         OPTION_VAL_EXECUTE_VN },
         { "execute_one",     OPTION_VAL_EXECUTE_ONE_VN },
         { "run_interpreter", OPTION_VAL_RUN_INTERPRETER_VN },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(dynarec_invalidate),
      BEETLE_OPT_DYNAREC_INVALIDATE_LABEL_VN,
      NULL,
      BEETLE_OPT_DYNAREC_INVALIDATE_INFO_0_VN,
      NULL,
      "hacks",
      {
         { "full", OPTION_VAL_FULL_VN },
         { "dma",  OPTION_VAL_DMA_VN },
         { NULL, NULL },
      },
      "full"
   },
   {
      BEETLE_OPT(dynarec_eventcycles),
      BEETLE_OPT_DYNAREC_EVENTCYCLES_LABEL_VN,
      NULL,
      BEETLE_OPT_DYNAREC_EVENTCYCLES_INFO_0_VN,
      NULL,
      "hacks",
      {
         { "128",  OPTION_VAL_128_VN },
         { "256",  NULL },
         { "384",  NULL },
         { "512",  NULL },
         { "640",  NULL },
         { "768",  NULL },
         { "896",  NULL },
         { "1024", NULL },
         { "1152", NULL },
         { "1280", NULL },
         { "1408", NULL },
         { "1536", NULL },
         { "1664", NULL },
         { "1792", NULL },
         { "1920", NULL },
         { "2048", NULL },
         { NULL, NULL },
      },
      "128"
   },
   {
      BEETLE_OPT(dynarec_spu_samples),
      BEETLE_OPT_DYNAREC_SPU_SAMPLES_LABEL_VN,
      NULL,
      BEETLE_OPT_DYNAREC_SPU_SAMPLES_INFO_0_VN,
      NULL,
      "hacks",
      {
         { "1",  OPTION_VAL_1_O49_VN },
         { "4",  NULL },
         { "16", NULL },
         { NULL, NULL },
      },
      "1"
   },
#endif
   {
      BEETLE_OPT(core_timing_fps),
      BEETLE_OPT_CORE_TIMING_FPS_LABEL_VN,
      NULL,
      BEETLE_OPT_CORE_TIMING_FPS_INFO_0_VN,
      NULL,
      "video",
      {
         { "auto_toggle",       OPTION_VAL_AUTO_TOGGLE_VN },
         { "force_progressive", OPTION_VAL_FORCE_PROGRESSIVE_VN },
         { "force_interlaced",  OPTION_VAL_FORCE_INTERLACED_VN },
      },
      "auto_toggle"
   },
   {
      BEETLE_OPT(deinterlacer),
      BEETLE_OPT_DEINTERLACER_LABEL_VN,
      NULL,
      BEETLE_OPT_DEINTERLACER_INFO_0_VN,
      NULL,
      "video",
      {
         { "weave",      OPTION_VAL_WEAVE_VN },
         { "bob",        OPTION_VAL_BOB_VN },
         { "bob_offset", OPTION_VAL_BOB_OFFSET_VN },
         { "fastmad",    OPTION_VAL_FASTMAD_VN },
         { "off",        "Off" },
         { NULL, NULL },
      },
      "weave"
   },
   {
      BEETLE_OPT(frame_duping),
      BEETLE_OPT_FRAME_DUPING_LABEL_VN,
      NULL,
      BEETLE_OPT_FRAME_DUPING_INFO_0_VN,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(pal_video_timing_override),
      BEETLE_OPT_PAL_VIDEO_TIMING_OVERRIDE_LABEL_VN,
      NULL,
      BEETLE_OPT_PAL_VIDEO_TIMING_OVERRIDE_INFO_0_VN,
      NULL,
      "video",
      {
         { "disabled", NULL },
         { "enabled",  NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(aspect_ratio),
      BEETLE_OPT_ASPECT_RATIO_LABEL_VN,
      NULL,
      BEETLE_OPT_ASPECT_RATIO_INFO_0_VN,
      NULL,
      "video",
      {
         { "corrected", OPTION_VAL_CORRECTED_VN },
         { "uncorrected", OPTION_VAL_UNCORRECTED_VN },
         { "4:3",  OPTION_VAL_4_3_VN },
         { "ntsc", OPTION_VAL_NTSC_VN },
      },
      "corrected"
   },
   {
      BEETLE_OPT(crop_overscan),
      BEETLE_OPT_CROP_OVERSCAN_LABEL_VN,
      NULL,
      BEETLE_OPT_CROP_OVERSCAN_INFO_0_VN,
      NULL,
      "video",
      {
         { "disabled", "Disabled" },
         { "static",   OPTION_VAL_STATIC_VN },
         { "smart",    OPTION_VAL_SMART_VN },
         { NULL, NULL },
      },
      "smart"
   },
   {
      BEETLE_OPT(image_crop),
      BEETLE_OPT_IMAGE_CROP_LABEL_VN,
      NULL,
      BEETLE_OPT_IMAGE_CROP_INFO_0_VN,
      NULL,
      "video",
      {
         { "disabled", "0" },
         { "1px",      OPTION_VAL_1PX_VN },
         { "2px",      OPTION_VAL_2PX_VN },
         { "3px",      OPTION_VAL_3PX_VN },
         { "4px",      OPTION_VAL_4PX_VN },
         { "5px",      OPTION_VAL_5PX_VN },
         { "6px",      OPTION_VAL_6PX_VN },
         { "7px",      OPTION_VAL_7PX_VN },
         { "8px",      OPTION_VAL_8PX_VN },
         { "9px",      OPTION_VAL_9PX_VN },
         { "10px",     OPTION_VAL_10PX_VN },
         { "11px",     OPTION_VAL_11PX_VN },
         { "12px",     OPTION_VAL_12PX_VN },
         { "13px",     OPTION_VAL_13PX_VN },
         { "14px",     OPTION_VAL_14PX_VN },
         { "15px",     OPTION_VAL_15PX_VN },
         { "16px",     OPTION_VAL_16PX_VN },
         { "17px",     OPTION_VAL_17PX_VN },
         { "18px",     OPTION_VAL_18PX_VN },
         { "19px",     OPTION_VAL_19PX_VN },
         { "20px",     OPTION_VAL_20PX_VN },
         { NULL, NULL },
      },
      "disabled"
   },
   {
      BEETLE_OPT(image_offset),
      BEETLE_OPT_IMAGE_OFFSET_LABEL_VN,
      NULL,
      BEETLE_OPT_IMAGE_OFFSET_INFO_0_VN,
      NULL,
      "video",
      {
         { "-12px",    OPTION_VAL_12PX_O64_VN },
         { "-11px",    OPTION_VAL_11PX_O64_VN },
         { "-10px",    OPTION_VAL_10PX_O64_VN },
         { "-9px",     OPTION_VAL_9PX_O64_VN },
         { "-8px",     OPTION_VAL_8PX_O64_VN },
         { "-7px",     OPTION_VAL_7PX_O64_VN },
         { "-6px",     OPTION_VAL_6PX_O64_VN },
         { "-5px",     OPTION_VAL_5PX_O64_VN },
         { "-4px",     OPTION_VAL_4PX_O64_VN },
         { "-3px",     OPTION_VAL_3PX_O64_VN },
         { "-2px",     OPTION_VAL_2PX_O64_VN },
         { "-1px",     OPTION_VAL_1PX_O64_VN },
         { "disabled", "0" },
         { "+1px",     OPTION_VAL_1PX_O64_0_VN },
         { "+2px",     OPTION_VAL_2PX_O64_0_VN },
         { "+3px",     OPTION_VAL_3PX_O64_0_VN },
         { "+4px",     OPTION_VAL_4PX_O64_0_VN },
         { "+5px",     OPTION_VAL_5PX_O64_0_VN },
         { "+6px",     OPTION_VAL_6PX_O64_0_VN },
         { "+7px",     OPTION_VAL_7PX_O64_0_VN },
         { "+8px",     OPTION_VAL_8PX_O64_0_VN },
         { "+9px",     OPTION_VAL_9PX_O64_0_VN },
         { "+10px",    OPTION_VAL_10PX_O64_0_VN },
         { "+11px",    OPTION_VAL_11PX_O64_0_VN },
         { "+12px",    OPTION_VAL_12PX_O64_0_VN },
         { NULL, NULL },
      },
      "disabled"
   },
#if defined(HAVE_OPENGL) || defined(HAVE_OPENGLES) || defined(HAVE_VULKAN)
   {
      BEETLE_OPT(image_offset_cycles),
      BEETLE_OPT_IMAGE_OFFSET_CYCLES_LABEL_VN,
      NULL,
      BEETLE_OPT_IMAGE_OFFSET_CYCLES_INFO_0_VN,
      NULL,
      "video",
      {
         { "-40",      NULL },
         { "-39",      NULL },
         { "-38",      NULL },
         { "-37",      NULL },
         { "-36",      NULL },
         { "-35",      NULL },
         { "-34",      NULL },
         { "-33",      NULL },
         { "-32",      NULL },
         { "-31",      NULL },
         { "-30",      NULL },
         { "-29",      NULL },
         { "-28",      NULL },
         { "-27",      NULL },
         { "-26",      NULL },
         { "-25",      NULL },
         { "-24",      NULL },
         { "-23",      NULL },
         { "-22",      NULL },
         { "-21",      NULL },
         { "-20",      NULL },
         { "-19",      NULL },
         { "-18",      NULL },
         { "-17",      NULL },
         { "-16",      NULL },
         { "-15",      NULL },
         { "-14",      NULL },
         { "-13",      NULL },
         { "-12",      NULL },
         { "-11",      NULL },
         { "-10",      NULL },
         { "-9",       NULL },
         { "-8",       NULL },
         { "-7",       NULL },
         { "-6",       NULL },
         { "-5",       NULL },
         { "-4",       NULL },
         { "-3",       NULL },
         { "-2",       NULL },
         { "-1",       NULL },
         { "0",        "0" },
         { "+1",       NULL },
         { "+2",       NULL },
         { "+3",       NULL },
         { "+4",       NULL },
         { "+5",       NULL },
         { "+6",       NULL },
         { "+7",       NULL },
         { "+8",       NULL },
         { "+9",       NULL },
         { "+10",      NULL },
         { "+11",      NULL },
         { "+12",      NULL },
         { "+13",      NULL },
         { "+14",      NULL },
         { "+15",      NULL },
         { "+16",      NULL },
         { "+17",      NULL },
         { "+18",      NULL },
         { "+19",      NULL },
         { "+20",      NULL },
         { "+21",      NULL },
         { "+22",      NULL },
         { "+23",      NULL },
         { "+24",      NULL },
         { "+25",      NULL },
         { "+26",      NULL },
         { "+27",      NULL },
         { "+28",      NULL },
         { "+29",      NULL },
         { "+30",      NULL },
         { "+31",      NULL },
         { "+32",      NULL },
         { "+33",      NULL },
         { "+34",      NULL },
         { "+35",      NULL },
         { "+36",      NULL },
         { "+37",      NULL },
         { "+38",      NULL },
         { "+39",      NULL },
         { "+40",      NULL },
         { NULL, NULL},
      },
      "0"
   },
#endif
   {
      BEETLE_OPT(initial_scanline),
      BEETLE_OPT_INITIAL_SCANLINE_LABEL_VN,
      NULL,
      BEETLE_OPT_INITIAL_SCANLINE_INFO_0_VN,
      NULL,
      "video",
      {
         { "0",  OPTION_VAL_0_O48_VN },
         { "1",  NULL },
         { "2",  NULL },
         { "3",  NULL },
         { "4",  NULL },
         { "5",  NULL },
         { "6",  NULL },
         { "7",  NULL },
         { "8",  NULL },
         { "9",  NULL },
         { "10", NULL },
         { "11", NULL },
         { "12", NULL },
         { "13", NULL },
         { "14", NULL },
         { "15", NULL },
         { "16", NULL },
         { "17", NULL },
         { "18", NULL },
         { "19", NULL },
         { "20", NULL },
         { "21", NULL },
         { "22", NULL },
         { "23", NULL },
         { "24", NULL },
         { "25", NULL },
         { "26", NULL },
         { "27", NULL },
         { "28", NULL },
         { "29", NULL },
         { "30", NULL },
         { "31", NULL },
         { "32", NULL },
         { "33", NULL },
         { "34", NULL },
         { "35", NULL },
         { "36", NULL },
         { "37", NULL },
         { "38", NULL },
         { "39", NULL },
         { "40", NULL },
         { NULL, NULL },
      },
      "0"
   },
   {
      BEETLE_OPT(last_scanline),
      BEETLE_OPT_LAST_SCANLINE_LABEL_VN,
      NULL,
      BEETLE_OPT_LAST_SCANLINE_INFO_0_VN,
      NULL,
      "video",
      {
         { "210", NULL },
         { "211", NULL },
         { "212", NULL },
         { "213", NULL },
         { "214", NULL },
         { "215", NULL },
         { "216", NULL },
         { "217", NULL },
         { "218", NULL },
         { "219", NULL },
         { "220", NULL },
         { "221", NULL },
         { "222", NULL },
         { "223", NULL },
         { "224", NULL },
         { "225", NULL },
         { "226", NULL },
         { "227", NULL },
         { "228", NULL },
         { "229", NULL },
         { "230", NULL },
         { "231", NULL },
         { "232", NULL },
         { "233", NULL },
         { "234", NULL },
         { "235", NULL },
         { "236", NULL },
         { "237", NULL },
         { "238", NULL },
         { "239", OPTION_VAL_239_VN },
         { NULL, NULL },
      },
      "239"
   },
   {
      BEETLE_OPT(initial_scanline_pal),
      BEETLE_OPT_INITIAL_SCANLINE_PAL_LABEL_VN,
      NULL,
      BEETLE_OPT_INITIAL_SCANLINE_PAL_INFO_0_VN,
      NULL,
      "video",
      {
         { "0",  OPTION_VAL_0_O48_VN },
         { "1",  NULL },
         { "2",  NULL },
         { "3",  NULL },
         { "4",  NULL },
         { "5",  NULL },
         { "6",  NULL },
         { "7",  NULL },
         { "8",  NULL },
         { "9",  NULL },
         { "10", NULL },
         { "11", NULL },
         { "12", NULL },
         { "13", NULL },
         { "14", NULL },
         { "15", NULL },
         { "16", NULL },
         { "17", NULL },
         { "18", NULL },
         { "19", NULL },
         { "20", NULL },
         { "21", NULL },
         { "22", NULL },
         { "23", NULL },
         { "24", NULL },
         { "25", NULL },
         { "26", NULL },
         { "27", NULL },
         { "28", NULL },
         { "29", NULL },
         { "30", NULL },
         { "31", NULL },
         { "32", NULL },
         { "33", NULL },
         { "34", NULL },
         { "35", NULL },
         { "36", NULL },
         { "37", NULL },
         { "38", NULL },
         { "39", NULL },
         { "40", NULL },
         { NULL, NULL },
      },
      "0"
   },
   {
      BEETLE_OPT(last_scanline_pal),
      BEETLE_OPT_LAST_SCANLINE_PAL_LABEL_VN,
      NULL,
      BEETLE_OPT_LAST_SCANLINE_PAL_INFO_0_VN,
      NULL,
      "video",
      {
         { "230", NULL },
         { "231", NULL },
         { "232", NULL },
         { "233", NULL },
         { "234", NULL },
         { "235", NULL },
         { "236", NULL },
         { "237", NULL },
         { "238", NULL },
         { "239", NULL },
         { "240", NULL },
         { "241", NULL },
         { "242", NULL },
         { "243", NULL },
         { "244", NULL },
         { "245", NULL },
         { "246", NULL },
         { "247", NULL },
         { "248", NULL },
         { "249", NULL },
         { "250", NULL },
         { "251", NULL },
         { "252", NULL },
         { "253", NULL },
         { "254", NULL },
         { "255", NULL },
         { "256", NULL },
         { "257", NULL },
         { "258", NULL },
         { "259", NULL },
         { "260", NULL },
         { "261", NULL },
         { "262", NULL },
         { "263", NULL },
         { "264", NULL },
         { "265", NULL },
         { "266", NULL },
         { "267", NULL },
         { "268", NULL },
         { "269", NULL },
         { "270", NULL },
         { "271", NULL },
         { "272", NULL },
         { "273", NULL },
         { "274", NULL },
         { "275", NULL },
         { "276", NULL },
         { "277", NULL },
         { "278", NULL },
         { "279", NULL },
         { "280", NULL },
         { "281", NULL },
         { "282", NULL },
         { "283", NULL },
         { "284", NULL },
         { "285", NULL },
         { "286", NULL },
         { "287", OPTION_VAL_287_VN },
         { NULL, NULL },
      },
      "287"
   },
   { NULL, NULL, NULL, NULL, NULL, NULL, {{0}}, NULL },
};
struct retro_core_options_v2 options_vn = {
   option_cats_vn,
   option_defs_vn
};


#ifdef __cplusplus
}
#endif

#endif